/*!
 * jQuery JavaScript Library v1.11.3
 * http://jquery.com/
 *
 * Includes Sizzle.js
 * http://sizzlejs.com/
 *
 * Copyright 2005, 2014 jQuery Foundation, Inc. and other contributors
 * Released under the MIT license
 * http://jquery.org/license
 *
 * Date: 2015-04-28T16:19Z
 */

(function( global, factory ) {

	if ( typeof module === "object" && typeof module.exports === "object" ) {
		// For CommonJS and CommonJS-like environments where a proper window is present,
		// execute the factory and get jQuery
		// For environments that do not inherently posses a window with a document
		// (such as Node.js), expose a jQuery-making factory as module.exports
		// This accentuates the need for the creation of a real window
		// e.g. var jQuery = require("jquery")(window);
		// See ticket #14549 for more info
		module.exports = global.document ?
			factory( global, true ) :
			function( w ) {
				if ( !w.document ) {
					throw new Error( "jQuery requires a window with a document" );
				}
				return factory( w );
			};
	} else {
		factory( global );
	}

// Pass this if window is not defined yet
}(typeof window !== "undefined" ? window : this, function( window, noGlobal ) {

// Can't do this because several apps including ASP.NET trace
// the stack via arguments.caller.callee and Firefox dies if
// you try to trace through "use strict" call chains. (#13335)
// Support: Firefox 18+
//

var deletedIds = [];

var slice = deletedIds.slice;

var concat = deletedIds.concat;

var push = deletedIds.push;

var indexOf = deletedIds.indexOf;

var class2type = {};

var toString = class2type.toString;

var hasOwn = class2type.hasOwnProperty;

var support = {};



var
	version = "1.11.3",

	// Define a local copy of jQuery
	jQuery = function( selector, context ) {
		// The jQuery object is actually just the init constructor 'enhanced'
		// Need init if jQuery is called (just allow error to be thrown if not included)
		return new jQuery.fn.init( selector, context );
	},

	// Support: Android<4.1, IE<9
	// Make sure we trim BOM and NBSP
	rtrim = /^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,

	// Matches dashed string for camelizing
	rmsPrefix = /^-ms-/,
	rdashAlpha = /-([\da-z])/gi,

	// Used by jQuery.camelCase as callback to replace()
	fcamelCase = function( all, letter ) {
		return letter.toUpperCase();
	};

jQuery.fn = jQuery.prototype = {
	// The current version of jQuery being used
	jquery: version,

	constructor: jQuery,

	// Start with an empty selector
	selector: "",

	// The default length of a jQuery object is 0
	length: 0,

	toArray: function() {
		return slice.call( this );
	},

	// Get the Nth element in the matched element set OR
	// Get the whole matched element set as a clean array
	get: function( num ) {
		return num != null ?

			// Return just the one element from the set
			( num < 0 ? this[ num + this.length ] : this[ num ] ) :

			// Return all the elements in a clean array
			slice.call( this );
	},

	// Take an array of elements and push it onto the stack
	// (returning the new matched element set)
	pushStack: function( elems ) {

		// Build a new jQuery matched element set
		var ret = jQuery.merge( this.constructor(), elems );

		// Add the old object onto the stack (as a reference)
		ret.prevObject = this;
		ret.context = this.context;

		// Return the newly-formed element set
		return ret;
	},

	// Execute a callback for every element in the matched set.
	// (You can seed the arguments with an array of args, but this is
	// only used internally.)
	each: function( callback, args ) {
		return jQuery.each( this, callback, args );
	},

	map: function( callback ) {
		return this.pushStack( jQuery.map(this, function( elem, i ) {
			return callback.call( elem, i, elem );
		}));
	},

	slice: function() {
		return this.pushStack( slice.apply( this, arguments ) );
	},

	first: function() {
		return this.eq( 0 );
	},

	last: function() {
		return this.eq( -1 );
	},

	eq: function( i ) {
		var len = this.length,
			j = +i + ( i < 0 ? len : 0 );
		return this.pushStack( j >= 0 && j < len ? [ this[j] ] : [] );
	},

	end: function() {
		return this.prevObject || this.constructor(null);
	},

	// For internal use only.
	// Behaves like an Array's method, not like a jQuery method.
	push: push,
	sort: deletedIds.sort,
	splice: deletedIds.splice
};

jQuery.extend = jQuery.fn.extend = function() {
	var src, copyIsArray, copy, name, options, clone,
		target = arguments[0] || {},
		i = 1,
		length = arguments.length,
		deep = false;

	// Handle a deep copy situation
	if ( typeof target === "boolean" ) {
		deep = target;

		// skip the boolean and the target
		target = arguments[ i ] || {};
		i++;
	}

	// Handle case when target is a string or something (possible in deep copy)
	if ( typeof target !== "object" && !jQuery.isFunction(target) ) {
		target = {};
	}

	// extend jQuery itself if only one argument is passed
	if ( i === length ) {
		target = this;
		i--;
	}

	for ( ; i < length; i++ ) {
		// Only deal with non-null/undefined values
		if ( (options = arguments[ i ]) != null ) {
			// Extend the base object
			for ( name in options ) {
				src = target[ name ];
				copy = options[ name ];

				// Prevent never-ending loop
				if ( target === copy ) {
					continue;
				}

				// Recurse if we're merging plain objects or arrays
				if ( deep && copy && ( jQuery.isPlainObject(copy) || (copyIsArray = jQuery.isArray(copy)) ) ) {
					if ( copyIsArray ) {
						copyIsArray = false;
						clone = src && jQuery.isArray(src) ? src : [];

					} else {
						clone = src && jQuery.isPlainObject(src) ? src : {};
					}

					// Never move original objects, clone them
					target[ name ] = jQuery.extend( deep, clone, copy );

				// Don't bring in undefined values
				} else if ( copy !== undefined ) {
					target[ name ] = copy;
				}
			}
		}
	}

	// Return the modified object
	return target;
};

jQuery.extend({
	// Unique for each copy of jQuery on the page
	expando: "jQuery" + ( version + Math.random() ).replace( /\D/g, "" ),

	// Assume jQuery is ready without the ready module
	isReady: true,

	error: function( msg ) {
		throw new Error( msg );
	},

	noop: function() {},

	// See test/unit/core.js for details concerning isFunction.
	// Since version 1.3, DOM methods and functions like alert
	// aren't supported. They return false on IE (#2968).
	isFunction: function( obj ) {
		return jQuery.type(obj) === "function";
	},

	isArray: Array.isArray || function( obj ) {
		return jQuery.type(obj) === "array";
	},

	isWindow: function( obj ) {
		/* jshint eqeqeq: false */
		return obj != null && obj == obj.window;
	},

	isNumeric: function( obj ) {
		// parseFloat NaNs numeric-cast false positives (null|true|false|"")
		// ...but misinterprets leading-number strings, particularly hex literals ("0x...")
		// subtraction forces infinities to NaN
		// adding 1 corrects loss of precision from parseFloat (#15100)
		return !jQuery.isArray( obj ) && (obj - parseFloat( obj ) + 1) >= 0;
	},

	isEmptyObject: function( obj ) {
		var name;
		for ( name in obj ) {
			return false;
		}
		return true;
	},

	isPlainObject: function( obj ) {
		var key;

		// Must be an Object.
		// Because of IE, we also have to check the presence of the constructor property.
		// Make sure that DOM nodes and window objects don't pass through, as well
		if ( !obj || jQuery.type(obj) !== "object" || obj.nodeType || jQuery.isWindow( obj ) ) {
			return false;
		}

		try {
			// Not own constructor property must be Object
			if ( obj.constructor &&
				!hasOwn.call(obj, "constructor") &&
				!hasOwn.call(obj.constructor.prototype, "isPrototypeOf") ) {
				return false;
			}
		} catch ( e ) {
			// IE8,9 Will throw exceptions on certain host objects #9897
			return false;
		}

		// Support: IE<9
		// Handle iteration over inherited properties before own properties.
		if ( support.ownLast ) {
			for ( key in obj ) {
				return hasOwn.call( obj, key );
			}
		}

		// Own properties are enumerated firstly, so to speed up,
		// if last one is own, then all properties are own.
		for ( key in obj ) {}

		return key === undefined || hasOwn.call( obj, key );
	},

	type: function( obj ) {
		if ( obj == null ) {
			return obj + "";
		}
		return typeof obj === "object" || typeof obj === "function" ?
			class2type[ toString.call(obj) ] || "object" :
			typeof obj;
	},

	// Evaluates a script in a global context
	// Workarounds based on findings by Jim Driscoll
	// http://weblogs.java.net/blog/driscoll/archive/2009/09/08/eval-javascript-global-context
	globalEval: function( data ) {
		if ( data && jQuery.trim( data ) ) {
			// We use execScript on Internet Explorer
			// We use an anonymous function so that context is window
			// rather than jQuery in Firefox
			( window.execScript || function( data ) {
				window[ "eval" ].call( window, data );
			} )( data );
		}
	},

	// Convert dashed to camelCase; used by the css and data modules
	// Microsoft forgot to hump their vendor prefix (#9572)
	camelCase: function( string ) {
		return string.replace( rmsPrefix, "ms-" ).replace( rdashAlpha, fcamelCase );
	},

	nodeName: function( elem, name ) {
		return elem.nodeName && elem.nodeName.toLowerCase() === name.toLowerCase();
	},

	// args is for internal usage only
	each: function( obj, callback, args ) {
		var value,
			i = 0,
			length = obj.length,
			isArray = isArraylike( obj );

		if ( args ) {
			if ( isArray ) {
				for ( ; i < length; i++ ) {
					value = callback.apply( obj[ i ], args );

					if ( value === false ) {
						break;
					}
				}
			} else {
				for ( i in obj ) {
					value = callback.apply( obj[ i ], args );

					if ( value === false ) {
						break;
					}
				}
			}

		// A special, fast, case for the most common use of each
		} else {
			if ( isArray ) {
				for ( ; i < length; i++ ) {
					value = callback.call( obj[ i ], i, obj[ i ] );

					if ( value === false ) {
						break;
					}
				}
			} else {
				for ( i in obj ) {
					value = callback.call( obj[ i ], i, obj[ i ] );

					if ( value === false ) {
						break;
					}
				}
			}
		}

		return obj;
	},

	// Support: Android<4.1, IE<9
	trim: function( text ) {
		return text == null ?
			"" :
			( text + "" ).replace( rtrim, "" );
	},

	// results is for internal usage only
	makeArray: function( arr, results ) {
		var ret = results || [];

		if ( arr != null ) {
			if ( isArraylike( Object(arr) ) ) {
				jQuery.merge( ret,
					typeof arr === "string" ?
					[ arr ] : arr
				);
			} else {
				push.call( ret, arr );
			}
		}

		return ret;
	},

	inArray: function( elem, arr, i ) {
		var len;

		if ( arr ) {
			if ( indexOf ) {
				return indexOf.call( arr, elem, i );
			}

			len = arr.length;
			i = i ? i < 0 ? Math.max( 0, len + i ) : i : 0;

			for ( ; i < len; i++ ) {
				// Skip accessing in sparse arrays
				if ( i in arr && arr[ i ] === elem ) {
					return i;
				}
			}
		}

		return -1;
	},

	merge: function( first, second ) {
		var len = +second.length,
			j = 0,
			i = first.length;

		while ( j < len ) {
			first[ i++ ] = second[ j++ ];
		}

		// Support: IE<9
		// Workaround casting of .length to NaN on otherwise arraylike objects (e.g., NodeLists)
		if ( len !== len ) {
			while ( second[j] !== undefined ) {
				first[ i++ ] = second[ j++ ];
			}
		}

		first.length = i;

		return first;
	},

	grep: function( elems, callback, invert ) {
		var callbackInverse,
			matches = [],
			i = 0,
			length = elems.length,
			callbackExpect = !invert;

		// Go through the array, only saving the items
		// that pass the validator function
		for ( ; i < length; i++ ) {
			callbackInverse = !callback( elems[ i ], i );
			if ( callbackInverse !== callbackExpect ) {
				matches.push( elems[ i ] );
			}
		}

		return matches;
	},

	// arg is for internal usage only
	map: function( elems, callback, arg ) {
		var value,
			i = 0,
			length = elems.length,
			isArray = isArraylike( elems ),
			ret = [];

		// Go through the array, translating each of the items to their new values
		if ( isArray ) {
			for ( ; i < length; i++ ) {
				value = callback( elems[ i ], i, arg );

				if ( value != null ) {
					ret.push( value );
				}
			}

		// Go through every key on the object,
		} else {
			for ( i in elems ) {
				value = callback( elems[ i ], i, arg );

				if ( value != null ) {
					ret.push( value );
				}
			}
		}

		// Flatten any nested arrays
		return concat.apply( [], ret );
	},

	// A global GUID counter for objects
	guid: 1,

	// Bind a function to a context, optionally partially applying any
	// arguments.
	proxy: function( fn, context ) {
		var args, proxy, tmp;

		if ( typeof context === "string" ) {
			tmp = fn[ context ];
			context = fn;
			fn = tmp;
		}

		// Quick check to determine if target is callable, in the spec
		// this throws a TypeError, but we will just return undefined.
		if ( !jQuery.isFunction( fn ) ) {
			return undefined;
		}

		// Simulated bind
		args = slice.call( arguments, 2 );
		proxy = function() {
			return fn.apply( context || this, args.concat( slice.call( arguments ) ) );
		};

		// Set the guid of unique handler to the same of original handler, so it can be removed
		proxy.guid = fn.guid = fn.guid || jQuery.guid++;

		return proxy;
	},

	now: function() {
		return +( new Date() );
	},

	// jQuery.support is not used in Core but other projects attach their
	// properties to it so it needs to exist.
	support: support
});

// Populate the class2type map
jQuery.each("Boolean Number String Function Array Date RegExp Object Error".split(" "), function(i, name) {
	class2type[ "[object " + name + "]" ] = name.toLowerCase();
});

function isArraylike( obj ) {

	// Support: iOS 8.2 (not reproducible in simulator)
	// `in` check used to prevent JIT error (gh-2145)
	// hasOwn isn't used here due to false negatives
	// regarding Nodelist length in IE
	var length = "length" in obj && obj.length,
		type = jQuery.type( obj );

	if ( type === "function" || jQuery.isWindow( obj ) ) {
		return false;
	}

	if ( obj.nodeType === 1 && length ) {
		return true;
	}

	return type === "array" || length === 0 ||
		typeof length === "number" && length > 0 && ( length - 1 ) in obj;
}
var Sizzle =
/*!
 * Sizzle CSS Selector Engine v2.2.0-pre
 * http://sizzlejs.com/
 *
 * Copyright 2008, 2014 jQuery Foundation, Inc. and other contributors
 * Released under the MIT license
 * http://jquery.org/license
 *
 * Date: 2014-12-16
 */
(function( window ) {

var i,
	support,
	Expr,
	getText,
	isXML,
	tokenize,
	compile,
	select,
	outermostContext,
	sortInput,
	hasDuplicate,

	// Local document vars
	setDocument,
	document,
	docElem,
	documentIsHTML,
	rbuggyQSA,
	rbuggyMatches,
	matches,
	contains,

	// Instance-specific data
	expando = "sizzle" + 1 * new Date(),
	preferredDoc = window.document,
	dirruns = 0,
	done = 0,
	classCache = createCache(),
	tokenCache = createCache(),
	compilerCache = createCache(),
	sortOrder = function( a, b ) {
		if ( a === b ) {
			hasDuplicate = true;
		}
		return 0;
	},

	// General-purpose constants
	MAX_NEGATIVE = 1 << 31,

	// Instance methods
	hasOwn = ({}).hasOwnProperty,
	arr = [],
	pop = arr.pop,
	push_native = arr.push,
	push = arr.push,
	slice = arr.slice,
	// Use a stripped-down indexOf as it's faster than native
	// http://jsperf.com/thor-indexof-vs-for/5
	indexOf = function( list, elem ) {
		var i = 0,
			len = list.length;
		for ( ; i < len; i++ ) {
			if ( list[i] === elem ) {
				return i;
			}
		}
		return -1;
	},

	booleans = "checked|selected|async|autofocus|autoplay|controls|defer|disabled|hidden|ismap|loop|multiple|open|readonly|required|scoped",

	// Regular expressions

	// Whitespace characters http://www.w3.org/TR/css3-selectors/#whitespace
	whitespace = "[\\x20\\t\\r\\n\\f]",
	// http://www.w3.org/TR/css3-syntax/#characters
	characterEncoding = "(?:\\\\.|[\\w-]|[^\\x00-\\xa0])+",

	// Loosely modeled on CSS identifier characters
	// An unquoted value should be a CSS identifier http://www.w3.org/TR/css3-selectors/#attribute-selectors
	// Proper syntax: http://www.w3.org/TR/CSS21/syndata.html#value-def-identifier
	identifier = characterEncoding.replace( "w", "w#" ),

	// Attribute selectors: http://www.w3.org/TR/selectors/#attribute-selectors
	attributes = "\\[" + whitespace + "*(" + characterEncoding + ")(?:" + whitespace +
		// Operator (capture 2)
		"*([*^$|!~]?=)" + whitespace +
		// "Attribute values must be CSS identifiers [capture 5] or strings [capture 3 or capture 4]"
		"*(?:'((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\"|(" + identifier + "))|)" + whitespace +
		"*\\]",

	pseudos = ":(" + characterEncoding + ")(?:\\((" +
		// To reduce the number of selectors needing tokenize in the preFilter, prefer arguments:
		// 1. quoted (capture 3; capture 4 or capture 5)
		"('((?:\\\\.|[^\\\\'])*)'|\"((?:\\\\.|[^\\\\\"])*)\")|" +
		// 2. simple (capture 6)
		"((?:\\\\.|[^\\\\()[\\]]|" + attributes + ")*)|" +
		// 3. anything else (capture 2)
		".*" +
		")\\)|)",

	// Leading and non-escaped trailing whitespace, capturing some non-whitespace characters preceding the latter
	rwhitespace = new RegExp( whitespace + "+", "g" ),
	rtrim = new RegExp( "^" + whitespace + "+|((?:^|[^\\\\])(?:\\\\.)*)" + whitespace + "+$", "g" ),

	rcomma = new RegExp( "^" + whitespace + "*," + whitespace + "*" ),
	rcombinators = new RegExp( "^" + whitespace + "*([>+~]|" + whitespace + ")" + whitespace + "*" ),

	rattributeQuotes = new RegExp( "=" + whitespace + "*([^\\]'\"]*?)" + whitespace + "*\\]", "g" ),

	rpseudo = new RegExp( pseudos ),
	ridentifier = new RegExp( "^" + identifier + "$" ),

	matchExpr = {
		"ID": new RegExp( "^#(" + characterEncoding + ")" ),
		"CLASS": new RegExp( "^\\.(" + characterEncoding + ")" ),
		"TAG": new RegExp( "^(" + characterEncoding.replace( "w", "w*" ) + ")" ),
		"ATTR": new RegExp( "^" + attributes ),
		"PSEUDO": new RegExp( "^" + pseudos ),
		"CHILD": new RegExp( "^:(only|first|last|nth|nth-last)-(child|of-type)(?:\\(" + whitespace +
			"*(even|odd|(([+-]|)(\\d*)n|)" + whitespace + "*(?:([+-]|)" + whitespace +
			"*(\\d+)|))" + whitespace + "*\\)|)", "i" ),
		"bool": new RegExp( "^(?:" + booleans + ")$", "i" ),
		// For use in libraries implementing .is()
		// We use this for POS matching in `select`
		"needsContext": new RegExp( "^" + whitespace + "*[>+~]|:(even|odd|eq|gt|lt|nth|first|last)(?:\\(" +
			whitespace + "*((?:-\\d)?\\d*)" + whitespace + "*\\)|)(?=[^-]|$)", "i" )
	},

	rinputs = /^(?:input|select|textarea|button)$/i,
	rheader = /^h\d$/i,

	rnative = /^[^{]+\{\s*\[native \w/,

	// Easily-parseable/retrievable ID or TAG or CLASS selectors
	rquickExpr = /^(?:#([\w-]+)|(\w+)|\.([\w-]+))$/,

	rsibling = /[+~]/,
	rescape = /'|\\/g,

	// CSS escapes http://www.w3.org/TR/CSS21/syndata.html#escaped-characters
	runescape = new RegExp( "\\\\([\\da-f]{1,6}" + whitespace + "?|(" + whitespace + ")|.)", "ig" ),
	funescape = function( _, escaped, escapedWhitespace ) {
		var high = "0x" + escaped - 0x10000;
		// NaN means non-codepoint
		// Support: Firefox<24
		// Workaround erroneous numeric interpretation of +"0x"
		return high !== high || escapedWhitespace ?
			escaped :
			high < 0 ?
				// BMP codepoint
				String.fromCharCode( high + 0x10000 ) :
				// Supplemental Plane codepoint (surrogate pair)
				String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
	},

	// Used for iframes
	// See setDocument()
	// Removing the function wrapper causes a "Permission Denied"
	// error in IE
	unloadHandler = function() {
		setDocument();
	};

// Optimize for push.apply( _, NodeList )
try {
	push.apply(
		(arr = slice.call( preferredDoc.childNodes )),
		preferredDoc.childNodes
	);
	// Support: Android<4.0
	// Detect silently failing push.apply
	arr[ preferredDoc.childNodes.length ].nodeType;
} catch ( e ) {
	push = { apply: arr.length ?

		// Leverage slice if possible
		function( target, els ) {
			push_native.apply( target, slice.call(els) );
		} :

		// Support: IE<9
		// Otherwise append directly
		function( target, els ) {
			var j = target.length,
				i = 0;
			// Can't trust NodeList.length
			while ( (target[j++] = els[i++]) ) {}
			target.length = j - 1;
		}
	};
}

function Sizzle( selector, context, results, seed ) {
	var match, elem, m, nodeType,
		// QSA vars
		i, groups, old, nid, newContext, newSelector;

	if ( ( context ? context.ownerDocument || context : preferredDoc ) !== document ) {
		setDocument( context );
	}

	context = context || document;
	results = results || [];
	nodeType = context.nodeType;

	if ( typeof selector !== "string" || !selector ||
		nodeType !== 1 && nodeType !== 9 && nodeType !== 11 ) {

		return results;
	}

	if ( !seed && documentIsHTML ) {

		// Try to shortcut find operations when possible (e.g., not under DocumentFragment)
		if ( nodeType !== 11 && (match = rquickExpr.exec( selector )) ) {
			// Speed-up: Sizzle("#ID")
			if ( (m = match[1]) ) {
				if ( nodeType === 9 ) {
					elem = context.getElementById( m );
					// Check parentNode to catch when Blackberry 4.6 returns
					// nodes that are no longer in the document (jQuery #6963)
					if ( elem && elem.parentNode ) {
						// Handle the case where IE, Opera, and Webkit return items
						// by name instead of ID
						if ( elem.id === m ) {
							results.push( elem );
							return results;
						}
					} else {
						return results;
					}
				} else {
					// Context is not a document
					if ( context.ownerDocument && (elem = context.ownerDocument.getElementById( m )) &&
						contains( context, elem ) && elem.id === m ) {
						results.push( elem );
						return results;
					}
				}

			// Speed-up: Sizzle("TAG")
			} else if ( match[2] ) {
				push.apply( results, context.getElementsByTagName( selector ) );
				return results;

			// Speed-up: Sizzle(".CLASS")
			} else if ( (m = match[3]) && support.getElementsByClassName ) {
				push.apply( results, context.getElementsByClassName( m ) );
				return results;
			}
		}

		// QSA path
		if ( support.qsa && (!rbuggyQSA || !rbuggyQSA.test( selector )) ) {
			nid = old = expando;
			newContext = context;
			newSelector = nodeType !== 1 && selector;

			// qSA works strangely on Element-rooted queries
			// We can work around this by specifying an extra ID on the root
			// and working up from there (Thanks to Andrew Dupont for the technique)
			// IE 8 doesn't work on object elements
			if ( nodeType === 1 && context.nodeName.toLowerCase() !== "object" ) {
				groups = tokenize( selector );

				if ( (old = context.getAttribute("id")) ) {
					nid = old.replace( rescape, "\\$&" );
				} else {
					context.setAttribute( "id", nid );
				}
				nid = "[id='" + nid + "'] ";

				i = groups.length;
				while ( i-- ) {
					groups[i] = nid + toSelector( groups[i] );
				}
				newContext = rsibling.test( selector ) && testContext( context.parentNode ) || context;
				newSelector = groups.join(",");
			}

			if ( newSelector ) {
				try {
					push.apply( results,
						newContext.querySelectorAll( newSelector )
					);
					return results;
				} catch(qsaError) {
				} finally {
					if ( !old ) {
						context.removeAttribute("id");
					}
				}
			}
		}
	}

	// All others
	return select( selector.replace( rtrim, "$1" ), context, results, seed );
}

/**
 * Create key-value caches of limited size
 * @returns {Function(string, Object)} Returns the Object data after storing it on itself with
 *	property name the (space-suffixed) string and (if the cache is larger than Expr.cacheLength)
 *	deleting the oldest entry
 */
function createCache() {
	var keys = [];

	function cache( key, value ) {
		// Use (key + " ") to avoid collision with native prototype properties (see Issue #157)
		if ( keys.push( key + " " ) > Expr.cacheLength ) {
			// Only keep the most recent entries
			delete cache[ keys.shift() ];
		}
		return (cache[ key + " " ] = value);
	}
	return cache;
}

/**
 * Mark a function for special use by Sizzle
 * @param {Function} fn The function to mark
 */
function markFunction( fn ) {
	fn[ expando ] = true;
	return fn;
}

/**
 * Support testing using an element
 * @param {Function} fn Passed the created div and expects a boolean result
 */
function assert( fn ) {
	var div = document.createElement("div");

	try {
		return !!fn( div );
	} catch (e) {
		return false;
	} finally {
		// Remove from its parent by default
		if ( div.parentNode ) {
			div.parentNode.removeChild( div );
		}
		// release memory in IE
		div = null;
	}
}

/**
 * Adds the same handler for all of the specified attrs
 * @param {String} attrs Pipe-separated list of attributes
 * @param {Function} handler The method that will be applied
 */
function addHandle( attrs, handler ) {
	var arr = attrs.split("|"),
		i = attrs.length;

	while ( i-- ) {
		Expr.attrHandle[ arr[i] ] = handler;
	}
}

/**
 * Checks document order of two siblings
 * @param {Element} a
 * @param {Element} b
 * @returns {Number} Returns less than 0 if a precedes b, greater than 0 if a follows b
 */
function siblingCheck( a, b ) {
	var cur = b && a,
		diff = cur && a.nodeType === 1 && b.nodeType === 1 &&
			( ~b.sourceIndex || MAX_NEGATIVE ) -
			( ~a.sourceIndex || MAX_NEGATIVE );

	// Use IE sourceIndex if available on both nodes
	if ( diff ) {
		return diff;
	}

	// Check if b follows a
	if ( cur ) {
		while ( (cur = cur.nextSibling) ) {
			if ( cur === b ) {
				return -1;
			}
		}
	}

	return a ? 1 : -1;
}

/**
 * Returns a function to use in pseudos for input types
 * @param {String} type
 */
function createInputPseudo( type ) {
	return function( elem ) {
		var name = elem.nodeName.toLowerCase();
		return name === "input" && elem.type === type;
	};
}

/**
 * Returns a function to use in pseudos for buttons
 * @param {String} type
 */
function createButtonPseudo( type ) {
	return function( elem ) {
		var name = elem.nodeName.toLowerCase();
		return (name === "input" || name === "button") && elem.type === type;
	};
}

/**
 * Returns a function to use in pseudos for positionals
 * @param {Function} fn
 */
function createPositionalPseudo( fn ) {
	return markFunction(function( argument ) {
		argument = +argument;
		return markFunction(function( seed, matches ) {
			var j,
				matchIndexes = fn( [], seed.length, argument ),
				i = matchIndexes.length;

			// Match elements found at the specified indexes
			while ( i-- ) {
				if ( seed[ (j = matchIndexes[i]) ] ) {
					seed[j] = !(matches[j] = seed[j]);
				}
			}
		});
	});
}

/**
 * Checks a node for validity as a Sizzle context
 * @param {Element|Object=} context
 * @returns {Element|Object|Boolean} The input node if acceptable, otherwise a falsy value
 */
function testContext( context ) {
	return context && typeof context.getElementsByTagName !== "undefined" && context;
}

// Expose support vars for convenience
support = Sizzle.support = {};

/**
 * Detects XML nodes
 * @param {Element|Object} elem An element or a document
 * @returns {Boolean} True iff elem is a non-HTML XML node
 */
isXML = Sizzle.isXML = function( elem ) {
	// documentElement is verified for cases where it doesn't yet exist
	// (such as loading iframes in IE - #4833)
	var documentElement = elem && (elem.ownerDocument || elem).documentElement;
	return documentElement ? documentElement.nodeName !== "HTML" : false;
};

/**
 * Sets document-related variables once based on the current document
 * @param {Element|Object} [doc] An element or document object to use to set the document
 * @returns {Object} Returns the current document
 */
setDocument = Sizzle.setDocument = function( node ) {
	var hasCompare, parent,
		doc = node ? node.ownerDocument || node : preferredDoc;

	// If no document and documentElement is available, return
	if ( doc === document || doc.nodeType !== 9 || !doc.documentElement ) {
		return document;
	}

	// Set our document
	document = doc;
	docElem = doc.documentElement;
	parent = doc.defaultView;

	// Support: IE>8
	// If iframe document is assigned to "document" variable and if iframe has been reloaded,
	// IE will throw "permission denied" error when accessing "document" variable, see jQuery #13936
	// IE6-8 do not support the defaultView property so parent will be undefined
	if ( parent && parent !== parent.top ) {
		// IE11 does not have attachEvent, so all must suffer
		if ( parent.addEventListener ) {
			parent.addEventListener( "unload", unloadHandler, false );
		} else if ( parent.attachEvent ) {
			parent.attachEvent( "onunload", unloadHandler );
		}
	}

	/* Support tests
	---------------------------------------------------------------------- */
	documentIsHTML = !isXML( doc );

	/* Attributes
	---------------------------------------------------------------------- */

	// Support: IE<8
	// Verify that getAttribute really returns attributes and not properties
	// (excepting IE8 booleans)
	support.attributes = assert(function( div ) {
		div.className = "i";
		return !div.getAttribute("className");
	});

	/* getElement(s)By*
	---------------------------------------------------------------------- */

	// Check if getElementsByTagName("*") returns only elements
	support.getElementsByTagName = assert(function( div ) {
		div.appendChild( doc.createComment("") );
		return !div.getElementsByTagName("*").length;
	});

	// Support: IE<9
	support.getElementsByClassName = rnative.test( doc.getElementsByClassName );

	// Support: IE<10
	// Check if getElementById returns elements by name
	// The broken getElementById methods don't pick up programatically-set names,
	// so use a roundabout getElementsByName test
	support.getById = assert(function( div ) {
		docElem.appendChild( div ).id = expando;
		return !doc.getElementsByName || !doc.getElementsByName( expando ).length;
	});

	// ID find and filter
	if ( support.getById ) {
		Expr.find["ID"] = function( id, context ) {
			if ( typeof context.getElementById !== "undefined" && documentIsHTML ) {
				var m = context.getElementById( id );
				// Check parentNode to catch when Blackberry 4.6 returns
				// nodes that are no longer in the document #6963
				return m && m.parentNode ? [ m ] : [];
			}
		};
		Expr.filter["ID"] = function( id ) {
			var attrId = id.replace( runescape, funescape );
			return function( elem ) {
				return elem.getAttribute("id") === attrId;
			};
		};
	} else {
		// Support: IE6/7
		// getElementById is not reliable as a find shortcut
		delete Expr.find["ID"];

		Expr.filter["ID"] =  function( id ) {
			var attrId = id.replace( runescape, funescape );
			return function( elem ) {
				var node = typeof elem.getAttributeNode !== "undefined" && elem.getAttributeNode("id");
				return node && node.value === attrId;
			};
		};
	}

	// Tag
	Expr.find["TAG"] = support.getElementsByTagName ?
		function( tag, context ) {
			if ( typeof context.getElementsByTagName !== "undefined" ) {
				return context.getElementsByTagName( tag );

			// DocumentFragment nodes don't have gEBTN
			} else if ( support.qsa ) {
				return context.querySelectorAll( tag );
			}
		} :

		function( tag, context ) {
			var elem,
				tmp = [],
				i = 0,
				// By happy coincidence, a (broken) gEBTN appears on DocumentFragment nodes too
				results = context.getElementsByTagName( tag );

			// Filter out possible comments
			if ( tag === "*" ) {
				while ( (elem = results[i++]) ) {
					if ( elem.nodeType === 1 ) {
						tmp.push( elem );
					}
				}

				return tmp;
			}
			return results;
		};

	// Class
	Expr.find["CLASS"] = support.getElementsByClassName && function( className, context ) {
		if ( documentIsHTML ) {
			return context.getElementsByClassName( className );
		}
	};

	/* QSA/matchesSelector
	---------------------------------------------------------------------- */

	// QSA and matchesSelector support

	// matchesSelector(:active) reports false when true (IE9/Opera 11.5)
	rbuggyMatches = [];

	// qSa(:focus) reports false when true (Chrome 21)
	// We allow this because of a bug in IE8/9 that throws an error
	// whenever `document.activeElement` is accessed on an iframe
	// So, we allow :focus to pass through QSA all the time to avoid the IE error
	// See http://bugs.jquery.com/ticket/13378
	rbuggyQSA = [];

	if ( (support.qsa = rnative.test( doc.querySelectorAll )) ) {
		// Build QSA regex
		// Regex strategy adopted from Diego Perini
		assert(function( div ) {
			// Select is set to empty string on purpose
			// This is to test IE's treatment of not explicitly
			// setting a boolean content attribute,
			// since its presence should be enough
			// http://bugs.jquery.com/ticket/12359
			docElem.appendChild( div ).innerHTML = "<a id='" + expando + "'></a>" +
				"<select id='" + expando + "-\f]' msallowcapture=''>" +
				"<option selected=''></option></select>";

			// Support: IE8, Opera 11-12.16
			// Nothing should be selected when empty strings follow ^= or $= or *=
			// The test attribute must be unknown in Opera but "safe" for WinRT
			// http://msdn.microsoft.com/en-us/library/ie/hh465388.aspx#attribute_section
			if ( div.querySelectorAll("[msallowcapture^='']").length ) {
				rbuggyQSA.push( "[*^$]=" + whitespace + "*(?:''|\"\")" );
			}

			// Support: IE8
			// Boolean attributes and "value" are not treated correctly
			if ( !div.querySelectorAll("[selected]").length ) {
				rbuggyQSA.push( "\\[" + whitespace + "*(?:value|" + booleans + ")" );
			}

			// Support: Chrome<29, Android<4.2+, Safari<7.0+, iOS<7.0+, PhantomJS<1.9.7+
			if ( !div.querySelectorAll( "[id~=" + expando + "-]" ).length ) {
				rbuggyQSA.push("~=");
			}

			// Webkit/Opera - :checked should return selected option elements
			// http://www.w3.org/TR/2011/REC-css3-selectors-20110929/#checked
			// IE8 throws error here and will not see later tests
			if ( !div.querySelectorAll(":checked").length ) {
				rbuggyQSA.push(":checked");
			}

			// Support: Safari 8+, iOS 8+
			// https://bugs.webkit.org/show_bug.cgi?id=136851
			// In-page `selector#id sibing-combinator selector` fails
			if ( !div.querySelectorAll( "a#" + expando + "+*" ).length ) {
				rbuggyQSA.push(".#.+[+~]");
			}
		});

		assert(function( div ) {
			// Support: Windows 8 Native Apps
			// The type and name attributes are restricted during .innerHTML assignment
			var input = doc.createElement("input");
			input.setAttribute( "type", "hidden" );
			div.appendChild( input ).setAttribute( "name", "D" );

			// Support: IE8
			// Enforce case-sensitivity of name attribute
			if ( div.querySelectorAll("[name=d]").length ) {
				rbuggyQSA.push( "name" + whitespace + "*[*^$|!~]?=" );
			}

			// FF 3.5 - :enabled/:disabled and hidden elements (hidden elements are still enabled)
			// IE8 throws error here and will not see later tests
			if ( !div.querySelectorAll(":enabled").length ) {
				rbuggyQSA.push( ":enabled", ":disabled" );
			}

			// Opera 10-11 does not throw on post-comma invalid pseudos
			div.querySelectorAll("*,:x");
			rbuggyQSA.push(",.*:");
		});
	}

	if ( (support.matchesSelector = rnative.test( (matches = docElem.matches ||
		docElem.webkitMatchesSelector ||
		docElem.mozMatchesSelector ||
		docElem.oMatchesSelector ||
		docElem.msMatchesSelector) )) ) {

		assert(function( div ) {
			// Check to see if it's possible to do matchesSelector
			// on a disconnected node (IE 9)
			support.disconnectedMatch = matches.call( div, "div" );

			// This should fail with an exception
			// Gecko does not error, returns false instead
			matches.call( div, "[s!='']:x" );
			rbuggyMatches.push( "!=", pseudos );
		});
	}

	rbuggyQSA = rbuggyQSA.length && new RegExp( rbuggyQSA.join("|") );
	rbuggyMatches = rbuggyMatches.length && new RegExp( rbuggyMatches.join("|") );

	/* Contains
	---------------------------------------------------------------------- */
	hasCompare = rnative.test( docElem.compareDocumentPosition );

	// Element contains another
	// Purposefully does not implement inclusive descendent
	// As in, an element does not contain itself
	contains = hasCompare || rnative.test( docElem.contains ) ?
		function( a, b ) {
			var adown = a.nodeType === 9 ? a.documentElement : a,
				bup = b && b.parentNode;
			return a === bup || !!( bup && bup.nodeType === 1 && (
				adown.contains ?
					adown.contains( bup ) :
					a.compareDocumentPosition && a.compareDocumentPosition( bup ) & 16
			));
		} :
		function( a, b ) {
			if ( b ) {
				while ( (b = b.parentNode) ) {
					if ( b === a ) {
						return true;
					}
				}
			}
			return false;
		};

	/* Sorting
	---------------------------------------------------------------------- */

	// Document order sorting
	sortOrder = hasCompare ?
	function( a, b ) {

		// Flag for duplicate removal
		if ( a === b ) {
			hasDuplicate = true;
			return 0;
		}

		// Sort on method existence if only one input has compareDocumentPosition
		var compare = !a.compareDocumentPosition - !b.compareDocumentPosition;
		if ( compare ) {
			return compare;
		}

		// Calculate position if both inputs belong to the same document
		compare = ( a.ownerDocument || a ) === ( b.ownerDocument || b ) ?
			a.compareDocumentPosition( b ) :

			// Otherwise we know they are disconnected
			1;

		// Disconnected nodes
		if ( compare & 1 ||
			(!support.sortDetached && b.compareDocumentPosition( a ) === compare) ) {

			// Choose the first element that is related to our preferred document
			if ( a === doc || a.ownerDocument === preferredDoc && contains(preferredDoc, a) ) {
				return -1;
			}
			if ( b === doc || b.ownerDocument === preferredDoc && contains(preferredDoc, b) ) {
				return 1;
			}

			// Maintain original order
			return sortInput ?
				( indexOf( sortInput, a ) - indexOf( sortInput, b ) ) :
				0;
		}

		return compare & 4 ? -1 : 1;
	} :
	function( a, b ) {
		// Exit early if the nodes are identical
		if ( a === b ) {
			hasDuplicate = true;
			return 0;
		}

		var cur,
			i = 0,
			aup = a.parentNode,
			bup = b.parentNode,
			ap = [ a ],
			bp = [ b ];

		// Parentless nodes are either documents or disconnected
		if ( !aup || !bup ) {
			return a === doc ? -1 :
				b === doc ? 1 :
				aup ? -1 :
				bup ? 1 :
				sortInput ?
				( indexOf( sortInput, a ) - indexOf( sortInput, b ) ) :
				0;

		// If the nodes are siblings, we can do a quick check
		} else if ( aup === bup ) {
			return siblingCheck( a, b );
		}

		// Otherwise we need full lists of their ancestors for comparison
		cur = a;
		while ( (cur = cur.parentNode) ) {
			ap.unshift( cur );
		}
		cur = b;
		while ( (cur = cur.parentNode) ) {
			bp.unshift( cur );
		}

		// Walk down the tree looking for a discrepancy
		while ( ap[i] === bp[i] ) {
			i++;
		}

		return i ?
			// Do a sibling check if the nodes have a common ancestor
			siblingCheck( ap[i], bp[i] ) :

			// Otherwise nodes in our document sort first
			ap[i] === preferredDoc ? -1 :
			bp[i] === preferredDoc ? 1 :
			0;
	};

	return doc;
};

Sizzle.matches = function( expr, elements ) {
	return Sizzle( expr, null, null, elements );
};

Sizzle.matchesSelector = function( elem, expr ) {
	// Set document vars if needed
	if ( ( elem.ownerDocument || elem ) !== document ) {
		setDocument( elem );
	}

	// Make sure that attribute selectors are quoted
	expr = expr.replace( rattributeQuotes, "='$1']" );

	if ( support.matchesSelector && documentIsHTML &&
		( !rbuggyMatches || !rbuggyMatches.test( expr ) ) &&
		( !rbuggyQSA     || !rbuggyQSA.test( expr ) ) ) {

		try {
			var ret = matches.call( elem, expr );

			// IE 9's matchesSelector returns false on disconnected nodes
			if ( ret || support.disconnectedMatch ||
					// As well, disconnected nodes are said to be in a document
					// fragment in IE 9
					elem.document && elem.document.nodeType !== 11 ) {
				return ret;
			}
		} catch (e) {}
	}

	return Sizzle( expr, document, null, [ elem ] ).length > 0;
};

Sizzle.contains = function( context, elem ) {
	// Set document vars if needed
	if ( ( context.ownerDocument || context ) !== document ) {
		setDocument( context );
	}
	return contains( context, elem );
};

Sizzle.attr = function( elem, name ) {
	// Set document vars if needed
	if ( ( elem.ownerDocument || elem ) !== document ) {
		setDocument( elem );
	}

	var fn = Expr.attrHandle[ name.toLowerCase() ],
		// Don't get fooled by Object.prototype properties (jQuery #13807)
		val = fn && hasOwn.call( Expr.attrHandle, name.toLowerCase() ) ?
			fn( elem, name, !documentIsHTML ) :
			undefined;

	return val !== undefined ?
		val :
		support.attributes || !documentIsHTML ?
			elem.getAttribute( name ) :
			(val = elem.getAttributeNode(name)) && val.specified ?
				val.value :
				null;
};

Sizzle.error = function( msg ) {
	throw new Error( "Syntax error, unrecognized expression: " + msg );
};

/**
 * Document sorting and removing duplicates
 * @param {ArrayLike} results
 */
Sizzle.uniqueSort = function( results ) {
	var elem,
		duplicates = [],
		j = 0,
		i = 0;

	// Unless we *know* we can detect duplicates, assume their presence
	hasDuplicate = !support.detectDuplicates;
	sortInput = !support.sortStable && results.slice( 0 );
	results.sort( sortOrder );

	if ( hasDuplicate ) {
		while ( (elem = results[i++]) ) {
			if ( elem === results[ i ] ) {
				j = duplicates.push( i );
			}
		}
		while ( j-- ) {
			results.splice( duplicates[ j ], 1 );
		}
	}

	// Clear input after sorting to release objects
	// See https://github.com/jquery/sizzle/pull/225
	sortInput = null;

	return results;
};

/**
 * Utility function for retrieving the text value of an array of DOM nodes
 * @param {Array|Element} elem
 */
getText = Sizzle.getText = function( elem ) {
	var node,
		ret = "",
		i = 0,
		nodeType = elem.nodeType;

	if ( !nodeType ) {
		// If no nodeType, this is expected to be an array
		while ( (node = elem[i++]) ) {
			// Do not traverse comment nodes
			ret += getText( node );
		}
	} else if ( nodeType === 1 || nodeType === 9 || nodeType === 11 ) {
		// Use textContent for elements
		// innerText usage removed for consistency of new lines (jQuery #11153)
		if ( typeof elem.textContent === "string" ) {
			return elem.textContent;
		} else {
			// Traverse its children
			for ( elem = elem.firstChild; elem; elem = elem.nextSibling ) {
				ret += getText( elem );
			}
		}
	} else if ( nodeType === 3 || nodeType === 4 ) {
		return elem.nodeValue;
	}
	// Do not include comment or processing instruction nodes

	return ret;
};

Expr = Sizzle.selectors = {

	// Can be adjusted by the user
	cacheLength: 50,

	createPseudo: markFunction,

	match: matchExpr,

	attrHandle: {},

	find: {},

	relative: {
		">": { dir: "parentNode", first: true },
		" ": { dir: "parentNode" },
		"+": { dir: "previousSibling", first: true },
		"~": { dir: "previousSibling" }
	},

	preFilter: {
		"ATTR": function( match ) {
			match[1] = match[1].replace( runescape, funescape );

			// Move the given value to match[3] whether quoted or unquoted
			match[3] = ( match[3] || match[4] || match[5] || "" ).replace( runescape, funescape );

			if ( match[2] === "~=" ) {
				match[3] = " " + match[3] + " ";
			}

			return match.slice( 0, 4 );
		},

		"CHILD": function( match ) {
			/* matches from matchExpr["CHILD"]
				1 type (only|nth|...)
				2 what (child|of-type)
				3 argument (even|odd|\d*|\d*n([+-]\d+)?|...)
				4 xn-component of xn+y argument ([+-]?\d*n|)
				5 sign of xn-component
				6 x of xn-component
				7 sign of y-component
				8 y of y-component
			*/
			match[1] = match[1].toLowerCase();

			if ( match[1].slice( 0, 3 ) === "nth" ) {
				// nth-* requires argument
				if ( !match[3] ) {
					Sizzle.error( match[0] );
				}

				// numeric x and y parameters for Expr.filter.CHILD
				// remember that false/true cast respectively to 0/1
				match[4] = +( match[4] ? match[5] + (match[6] || 1) : 2 * ( match[3] === "even" || match[3] === "odd" ) );
				match[5] = +( ( match[7] + match[8] ) || match[3] === "odd" );

			// other types prohibit arguments
			} else if ( match[3] ) {
				Sizzle.error( match[0] );
			}

			return match;
		},

		"PSEUDO": function( match ) {
			var excess,
				unquoted = !match[6] && match[2];

			if ( matchExpr["CHILD"].test( match[0] ) ) {
				return null;
			}

			// Accept quoted arguments as-is
			if ( match[3] ) {
				match[2] = match[4] || match[5] || "";

			// Strip excess characters from unquoted arguments
			} else if ( unquoted && rpseudo.test( unquoted ) &&
				// Get excess from tokenize (recursively)
				(excess = tokenize( unquoted, true )) &&
				// advance to the next closing parenthesis
				(excess = unquoted.indexOf( ")", unquoted.length - excess ) - unquoted.length) ) {

				// excess is a negative index
				match[0] = match[0].slice( 0, excess );
				match[2] = unquoted.slice( 0, excess );
			}

			// Return only captures needed by the pseudo filter method (type and argument)
			return match.slice( 0, 3 );
		}
	},

	filter: {

		"TAG": function( nodeNameSelector ) {
			var nodeName = nodeNameSelector.replace( runescape, funescape ).toLowerCase();
			return nodeNameSelector === "*" ?
				function() { return true; } :
				function( elem ) {
					return elem.nodeName && elem.nodeName.toLowerCase() === nodeName;
				};
		},

		"CLASS": function( className ) {
			var pattern = classCache[ className + " " ];

			return pattern ||
				(pattern = new RegExp( "(^|" + whitespace + ")" + className + "(" + whitespace + "|$)" )) &&
				classCache( className, function( elem ) {
					return pattern.test( typeof elem.className === "string" && elem.className || typeof elem.getAttribute !== "undefined" && elem.getAttribute("class") || "" );
				});
		},

		"ATTR": function( name, operator, check ) {
			return function( elem ) {
				var result = Sizzle.attr( elem, name );

				if ( result == null ) {
					return operator === "!=";
				}
				if ( !operator ) {
					return true;
				}

				result += "";

				return operator === "=" ? result === check :
					operator === "!=" ? result !== check :
					operator === "^=" ? check && result.indexOf( check ) === 0 :
					operator === "*=" ? check && result.indexOf( check ) > -1 :
					operator === "$=" ? check && result.slice( -check.length ) === check :
					operator === "~=" ? ( " " + result.replace( rwhitespace, " " ) + " " ).indexOf( check ) > -1 :
					operator === "|=" ? result === check || result.slice( 0, check.length + 1 ) === check + "-" :
					false;
			};
		},

		"CHILD": function( type, what, argument, first, last ) {
			var simple = type.slice( 0, 3 ) !== "nth",
				forward = type.slice( -4 ) !== "last",
				ofType = what === "of-type";

			return first === 1 && last === 0 ?

				// Shortcut for :nth-*(n)
				function( elem ) {
					return !!elem.parentNode;
				} :

				function( elem, context, xml ) {
					var cache, outerCache, node, diff, nodeIndex, start,
						dir = simple !== forward ? "nextSibling" : "previousSibling",
						parent = elem.parentNode,
						name = ofType && elem.nodeName.toLowerCase(),
						useCache = !xml && !ofType;

					if ( parent ) {

						// :(first|last|only)-(child|of-type)
						if ( simple ) {
							while ( dir ) {
								node = elem;
								while ( (node = node[ dir ]) ) {
									if ( ofType ? node.nodeName.toLowerCase() === name : node.nodeType === 1 ) {
										return false;
									}
								}
								// Reverse direction for :only-* (if we haven't yet done so)
								start = dir = type === "only" && !start && "nextSibling";
							}
							return true;
						}

						start = [ forward ? parent.firstChild : parent.lastChild ];

						// non-xml :nth-child(...) stores cache data on `parent`
						if ( forward && useCache ) {
							// Seek `elem` from a previously-cached index
							outerCache = parent[ expando ] || (parent[ expando ] = {});
							cache = outerCache[ type ] || [];
							nodeIndex = cache[0] === dirruns && cache[1];
							diff = cache[0] === dirruns && cache[2];
							node = nodeIndex && parent.childNodes[ nodeIndex ];

							while ( (node = ++nodeIndex && node && node[ dir ] ||

								// Fallback to seeking `elem` from the start
								(diff = nodeIndex = 0) || start.pop()) ) {

								// When found, cache indexes on `parent` and break
								if ( node.nodeType === 1 && ++diff && node === elem ) {
									outerCache[ type ] = [ dirruns, nodeIndex, diff ];
									break;
								}
							}

						// Use previously-cached element index if available
						} else if ( useCache && (cache = (elem[ expando ] || (elem[ expando ] = {}))[ type ]) && cache[0] === dirruns ) {
							diff = cache[1];

						// xml :nth-child(...) or :nth-last-child(...) or :nth(-last)?-of-type(...)
						} else {
							// Use the same loop as above to seek `elem` from the start
							while ( (node = ++nodeIndex && node && node[ dir ] ||
								(diff = nodeIndex = 0) || start.pop()) ) {

								if ( ( ofType ? node.nodeName.toLowerCase() === name : node.nodeType === 1 ) && ++diff ) {
									// Cache the index of each encountered element
									if ( useCache ) {
										(node[ expando ] || (node[ expando ] = {}))[ type ] = [ dirruns, diff ];
									}

									if ( node === elem ) {
										break;
									}
								}
							}
						}

						// Incorporate the offset, then check against cycle size
						diff -= last;
						return diff === first || ( diff % first === 0 && diff / first >= 0 );
					}
				};
		},

		"PSEUDO": function( pseudo, argument ) {
			// pseudo-class names are case-insensitive
			// http://www.w3.org/TR/selectors/#pseudo-classes
			// Prioritize by case sensitivity in case custom pseudos are added with uppercase letters
			// Remember that setFilters inherits from pseudos
			var args,
				fn = Expr.pseudos[ pseudo ] || Expr.setFilters[ pseudo.toLowerCase() ] ||
					Sizzle.error( "unsupported pseudo: " + pseudo );

			// The user may use createPseudo to indicate that
			// arguments are needed to create the filter function
			// just as Sizzle does
			if ( fn[ expando ] ) {
				return fn( argument );
			}

			// But maintain support for old signatures
			if ( fn.length > 1 ) {
				args = [ pseudo, pseudo, "", argument ];
				return Expr.setFilters.hasOwnProperty( pseudo.toLowerCase() ) ?
					markFunction(function( seed, matches ) {
						var idx,
							matched = fn( seed, argument ),
							i = matched.length;
						while ( i-- ) {
							idx = indexOf( seed, matched[i] );
							seed[ idx ] = !( matches[ idx ] = matched[i] );
						}
					}) :
					function( elem ) {
						return fn( elem, 0, args );
					};
			}

			return fn;
		}
	},

	pseudos: {
		// Potentially complex pseudos
		"not": markFunction(function( selector ) {
			// Trim the selector passed to compile
			// to avoid treating leading and trailing
			// spaces as combinators
			var input = [],
				results = [],
				matcher = compile( selector.replace( rtrim, "$1" ) );

			return matcher[ expando ] ?
				markFunction(function( seed, matches, context, xml ) {
					var elem,
						unmatched = matcher( seed, null, xml, [] ),
						i = seed.length;

					// Match elements unmatched by `matcher`
					while ( i-- ) {
						if ( (elem = unmatched[i]) ) {
							seed[i] = !(matches[i] = elem);
						}
					}
				}) :
				function( elem, context, xml ) {
					input[0] = elem;
					matcher( input, null, xml, results );
					// Don't keep the element (issue #299)
					input[0] = null;
					return !results.pop();
				};
		}),

		"has": markFunction(function( selector ) {
			return function( elem ) {
				return Sizzle( selector, elem ).length > 0;
			};
		}),

		"contains": markFunction(function( text ) {
			text = text.replace( runescape, funescape );
			return function( elem ) {
				return ( elem.textContent || elem.innerText || getText( elem ) ).indexOf( text ) > -1;
			};
		}),

		// "Whether an element is represented by a :lang() selector
		// is based solely on the element's language value
		// being equal to the identifier C,
		// or beginning with the identifier C immediately followed by "-".
		// The matching of C against the element's language value is performed case-insensitively.
		// The identifier C does not have to be a valid language name."
		// http://www.w3.org/TR/selectors/#lang-pseudo
		"lang": markFunction( function( lang ) {
			// lang value must be a valid identifier
			if ( !ridentifier.test(lang || "") ) {
				Sizzle.error( "unsupported lang: " + lang );
			}
			lang = lang.replace( runescape, funescape ).toLowerCase();
			return function( elem ) {
				var elemLang;
				do {
					if ( (elemLang = documentIsHTML ?
						elem.lang :
						elem.getAttribute("xml:lang") || elem.getAttribute("lang")) ) {

						elemLang = elemLang.toLowerCase();
						return elemLang === lang || elemLang.indexOf( lang + "-" ) === 0;
					}
				} while ( (elem = elem.parentNode) && elem.nodeType === 1 );
				return false;
			};
		}),

		// Miscellaneous
		"target": function( elem ) {
			var hash = window.location && window.location.hash;
			return hash && hash.slice( 1 ) === elem.id;
		},

		"root": function( elem ) {
			return elem === docElem;
		},

		"focus": function( elem ) {
			return elem === document.activeElement && (!document.hasFocus || document.hasFocus()) && !!(elem.type || elem.href || ~elem.tabIndex);
		},

		// Boolean properties
		"enabled": function( elem ) {
			return elem.disabled === false;
		},

		"disabled": function( elem ) {
			return elem.disabled === true;
		},

		"checked": function( elem ) {
			// In CSS3, :checked should return both checked and selected elements
			// http://www.w3.org/TR/2011/REC-css3-selectors-20110929/#checked
			var nodeName = elem.nodeName.toLowerCase();
			return (nodeName === "input" && !!elem.checked) || (nodeName === "option" && !!elem.selected);
		},

		"selected": function( elem ) {
			// Accessing this property makes selected-by-default
			// options in Safari work properly
			if ( elem.parentNode ) {
				elem.parentNode.selectedIndex;
			}

			return elem.selected === true;
		},

		// Contents
		"empty": function( elem ) {
			// http://www.w3.org/TR/selectors/#empty-pseudo
			// :empty is negated by element (1) or content nodes (text: 3; cdata: 4; entity ref: 5),
			//   but not by others (comment: 8; processing instruction: 7; etc.)
			// nodeType < 6 works because attributes (2) do not appear as children
			for ( elem = elem.firstChild; elem; elem = elem.nextSibling ) {
				if ( elem.nodeType < 6 ) {
					return false;
				}
			}
			return true;
		},

		"parent": function( elem ) {
			return !Expr.pseudos["empty"]( elem );
		},

		// Element/input types
		"header": function( elem ) {
			return rheader.test( elem.nodeName );
		},

		"input": function( elem ) {
			return rinputs.test( elem.nodeName );
		},

		"button": function( elem ) {
			var name = elem.nodeName.toLowerCase();
			return name === "input" && elem.type === "button" || name === "button";
		},

		"text": function( elem ) {
			var attr;
			return elem.nodeName.toLowerCase() === "input" &&
				elem.type === "text" &&

				// Support: IE<8
				// New HTML5 attribute values (e.g., "search") appear with elem.type === "text"
				( (attr = elem.getAttribute("type")) == null || attr.toLowerCase() === "text" );
		},

		// Position-in-collection
		"first": createPositionalPseudo(function() {
			return [ 0 ];
		}),

		"last": createPositionalPseudo(function( matchIndexes, length ) {
			return [ length - 1 ];
		}),

		"eq": createPositionalPseudo(function( matchIndexes, length, argument ) {
			return [ argument < 0 ? argument + length : argument ];
		}),

		"even": createPositionalPseudo(function( matchIndexes, length ) {
			var i = 0;
			for ( ; i < length; i += 2 ) {
				matchIndexes.push( i );
			}
			return matchIndexes;
		}),

		"odd": createPositionalPseudo(function( matchIndexes, length ) {
			var i = 1;
			for ( ; i < length; i += 2 ) {
				matchIndexes.push( i );
			}
			return matchIndexes;
		}),

		"lt": createPositionalPseudo(function( matchIndexes, length, argument ) {
			var i = argument < 0 ? argument + length : argument;
			for ( ; --i >= 0; ) {
				matchIndexes.push( i );
			}
			return matchIndexes;
		}),

		"gt": createPositionalPseudo(function( matchIndexes, length, argument ) {
			var i = argument < 0 ? argument + length : argument;
			for ( ; ++i < length; ) {
				matchIndexes.push( i );
			}
			return matchIndexes;
		})
	}
};

Expr.pseudos["nth"] = Expr.pseudos["eq"];

// Add button/input type pseudos
for ( i in { radio: true, checkbox: true, file: true, password: true, image: true } ) {
	Expr.pseudos[ i ] = createInputPseudo( i );
}
for ( i in { submit: true, reset: true } ) {
	Expr.pseudos[ i ] = createButtonPseudo( i );
}

// Easy API for creating new setFilters
function setFilters() {}
setFilters.prototype = Expr.filters = Expr.pseudos;
Expr.setFilters = new setFilters();

tokenize = Sizzle.tokenize = function( selector, parseOnly ) {
	var matched, match, tokens, type,
		soFar, groups, preFilters,
		cached = tokenCache[ selector + " " ];

	if ( cached ) {
		return parseOnly ? 0 : cached.slice( 0 );
	}

	soFar = selector;
	groups = [];
	preFilters = Expr.preFilter;

	while ( soFar ) {

		// Comma and first run
		if ( !matched || (match = rcomma.exec( soFar )) ) {
			if ( match ) {
				// Don't consume trailing commas as valid
				soFar = soFar.slice( match[0].length ) || soFar;
			}
			groups.push( (tokens = []) );
		}

		matched = false;

		// Combinators
		if ( (match = rcombinators.exec( soFar )) ) {
			matched = match.shift();
			tokens.push({
				value: matched,
				// Cast descendant combinators to space
				type: match[0].replace( rtrim, " " )
			});
			soFar = soFar.slice( matched.length );
		}

		// Filters
		for ( type in Expr.filter ) {
			if ( (match = matchExpr[ type ].exec( soFar )) && (!preFilters[ type ] ||
				(match = preFilters[ type ]( match ))) ) {
				matched = match.shift();
				tokens.push({
					value: matched,
					type: type,
					matches: match
				});
				soFar = soFar.slice( matched.length );
			}
		}

		if ( !matched ) {
			break;
		}
	}

	// Return the length of the invalid excess
	// if we're just parsing
	// Otherwise, throw an error or return tokens
	return parseOnly ?
		soFar.length :
		soFar ?
			Sizzle.error( selector ) :
			// Cache the tokens
			tokenCache( selector, groups ).slice( 0 );
};

function toSelector( tokens ) {
	var i = 0,
		len = tokens.length,
		selector = "";
	for ( ; i < len; i++ ) {
		selector += tokens[i].value;
	}
	return selector;
}

function addCombinator( matcher, combinator, base ) {
	var dir = combinator.dir,
		checkNonElements = base && dir === "parentNode",
		doneName = done++;

	return combinator.first ?
		// Check against closest ancestor/preceding element
		function( elem, context, xml ) {
			while ( (elem = elem[ dir ]) ) {
				if ( elem.nodeType === 1 || checkNonElements ) {
					return matcher( elem, context, xml );
				}
			}
		} :

		// Check against all ancestor/preceding elements
		function( elem, context, xml ) {
			var oldCache, outerCache,
				newCache = [ dirruns, doneName ];

			// We can't set arbitrary data on XML nodes, so they don't benefit from dir caching
			if ( xml ) {
				while ( (elem = elem[ dir ]) ) {
					if ( elem.nodeType === 1 || checkNonElements ) {
						if ( matcher( elem, context, xml ) ) {
							return true;
						}
					}
				}
			} else {
				while ( (elem = elem[ dir ]) ) {
					if ( elem.nodeType === 1 || checkNonElements ) {
						outerCache = elem[ expando ] || (elem[ expando ] = {});
						if ( (oldCache = outerCache[ dir ]) &&
							oldCache[ 0 ] === dirruns && oldCache[ 1 ] === doneName ) {

							// Assign to newCache so results back-propagate to previous elements
							return (newCache[ 2 ] = oldCache[ 2 ]);
						} else {
							// Reuse newcache so results back-propagate to previous elements
							outerCache[ dir ] = newCache;

							// A match means we're done; a fail means we have to keep checking
							if ( (newCache[ 2 ] = matcher( elem, context, xml )) ) {
								return true;
							}
						}
					}
				}
			}
		};
}

function elementMatcher( matchers ) {
	return matchers.length > 1 ?
		function( elem, context, xml ) {
			var i = matchers.length;
			while ( i-- ) {
				if ( !matchers[i]( elem, context, xml ) ) {
					return false;
				}
			}
			return true;
		} :
		matchers[0];
}

function multipleContexts( selector, contexts, results ) {
	var i = 0,
		len = contexts.length;
	for ( ; i < len; i++ ) {
		Sizzle( selector, contexts[i], results );
	}
	return results;
}

function condense( unmatched, map, filter, context, xml ) {
	var elem,
		newUnmatched = [],
		i = 0,
		len = unmatched.length,
		mapped = map != null;

	for ( ; i < len; i++ ) {
		if ( (elem = unmatched[i]) ) {
			if ( !filter || filter( elem, context, xml ) ) {
				newUnmatched.push( elem );
				if ( mapped ) {
					map.push( i );
				}
			}
		}
	}

	return newUnmatched;
}

function setMatcher( preFilter, selector, matcher, postFilter, postFinder, postSelector ) {
	if ( postFilter && !postFilter[ expando ] ) {
		postFilter = setMatcher( postFilter );
	}
	if ( postFinder && !postFinder[ expando ] ) {
		postFinder = setMatcher( postFinder, postSelector );
	}
	return markFunction(function( seed, results, context, xml ) {
		var temp, i, elem,
			preMap = [],
			postMap = [],
			preexisting = results.length,

			// Get initial elements from seed or context
			elems = seed || multipleContexts( selector || "*", context.nodeType ? [ context ] : context, [] ),

			// Prefilter to get matcher input, preserving a map for seed-results synchronization
			matcherIn = preFilter && ( seed || !selector ) ?
				condense( elems, preMap, preFilter, context, xml ) :
				elems,

			matcherOut = matcher ?
				// If we have a postFinder, or filtered seed, or non-seed postFilter or preexisting results,
				postFinder || ( seed ? preFilter : preexisting || postFilter ) ?

					// ...intermediate processing is necessary
					[] :

					// ...otherwise use results directly
					results :
				matcherIn;

		// Find primary matches
		if ( matcher ) {
			matcher( matcherIn, matcherOut, context, xml );
		}

		// Apply postFilter
		if ( postFilter ) {
			temp = condense( matcherOut, postMap );
			postFilter( temp, [], context, xml );

			// Un-match failing elements by moving them back to matcherIn
			i = temp.length;
			while ( i-- ) {
				if ( (elem = temp[i]) ) {
					matcherOut[ postMap[i] ] = !(matcherIn[ postMap[i] ] = elem);
				}
			}
		}

		if ( seed ) {
			if ( postFinder || preFilter ) {
				if ( postFinder ) {
					// Get the final matcherOut by condensing this intermediate into postFinder contexts
					temp = [];
					i = matcherOut.length;
					while ( i-- ) {
						if ( (elem = matcherOut[i]) ) {
							// Restore matcherIn since elem is not yet a final match
							temp.push( (matcherIn[i] = elem) );
						}
					}
					postFinder( null, (matcherOut = []), temp, xml );
				}

				// Move matched elements from seed to results to keep them synchronized
				i = matcherOut.length;
				while ( i-- ) {
					if ( (elem = matcherOut[i]) &&
						(temp = postFinder ? indexOf( seed, elem ) : preMap[i]) > -1 ) {

						seed[temp] = !(results[temp] = elem);
					}
				}
			}

		// Add elements to results, through postFinder if defined
		} else {
			matcherOut = condense(
				matcherOut === results ?
					matcherOut.splice( preexisting, matcherOut.length ) :
					matcherOut
			);
			if ( postFinder ) {
				postFinder( null, results, matcherOut, xml );
			} else {
				push.apply( results, matcherOut );
			}
		}
	});
}

function matcherFromTokens( tokens ) {
	var checkContext, matcher, j,
		len = tokens.length,
		leadingRelative = Expr.relative[ tokens[0].type ],
		implicitRelative = leadingRelative || Expr.relative[" "],
		i = leadingRelative ? 1 : 0,

		// The foundational matcher ensures that elements are reachable from top-level context(s)
		matchContext = addCombinator( function( elem ) {
			return elem === checkContext;
		}, implicitRelative, true ),
		matchAnyContext = addCombinator( function( elem ) {
			return indexOf( checkContext, elem ) > -1;
		}, implicitRelative, true ),
		matchers = [ function( elem, context, xml ) {
			var ret = ( !leadingRelative && ( xml || context !== outermostContext ) ) || (
				(checkContext = context).nodeType ?
					matchContext( elem, context, xml ) :
					matchAnyContext( elem, context, xml ) );
			// Avoid hanging onto element (issue #299)
			checkContext = null;
			return ret;
		} ];

	for ( ; i < len; i++ ) {
		if ( (matcher = Expr.relative[ tokens[i].type ]) ) {
			matchers = [ addCombinator(elementMatcher( matchers ), matcher) ];
		} else {
			matcher = Expr.filter[ tokens[i].type ].apply( null, tokens[i].matches );

			// Return special upon seeing a positional matcher
			if ( matcher[ expando ] ) {
				// Find the next relative operator (if any) for proper handling
				j = ++i;
				for ( ; j < len; j++ ) {
					if ( Expr.relative[ tokens[j].type ] ) {
						break;
					}
				}
				return setMatcher(
					i > 1 && elementMatcher( matchers ),
					i > 1 && toSelector(
						// If the preceding token was a descendant combinator, insert an implicit any-element `*`
						tokens.slice( 0, i - 1 ).concat({ value: tokens[ i - 2 ].type === " " ? "*" : "" })
					).replace( rtrim, "$1" ),
					matcher,
					i < j && matcherFromTokens( tokens.slice( i, j ) ),
					j < len && matcherFromTokens( (tokens = tokens.slice( j )) ),
					j < len && toSelector( tokens )
				);
			}
			matchers.push( matcher );
		}
	}

	return elementMatcher( matchers );
}

function matcherFromGroupMatchers( elementMatchers, setMatchers ) {
	var bySet = setMatchers.length > 0,
		byElement = elementMatchers.length > 0,
		superMatcher = function( seed, context, xml, results, outermost ) {
			var elem, j, matcher,
				matchedCount = 0,
				i = "0",
				unmatched = seed && [],
				setMatched = [],
				contextBackup = outermostContext,
				// We must always have either seed elements or outermost context
				elems = seed || byElement && Expr.find["TAG"]( "*", outermost ),
				// Use integer dirruns iff this is the outermost matcher
				dirrunsUnique = (dirruns += contextBackup == null ? 1 : Math.random() || 0.1),
				len = elems.length;

			if ( outermost ) {
				outermostContext = context !== document && context;
			}

			// Add elements passing elementMatchers directly to results
			// Keep `i` a string if there are no elements so `matchedCount` will be "00" below
			// Support: IE<9, Safari
			// Tolerate NodeList properties (IE: "length"; Safari: <number>) matching elements by id
			for ( ; i !== len && (elem = elems[i]) != null; i++ ) {
				if ( byElement && elem ) {
					j = 0;
					while ( (matcher = elementMatchers[j++]) ) {
						if ( matcher( elem, context, xml ) ) {
							results.push( elem );
							break;
						}
					}
					if ( outermost ) {
						dirruns = dirrunsUnique;
					}
				}

				// Track unmatched elements for set filters
				if ( bySet ) {
					// They will have gone through all possible matchers
					if ( (elem = !matcher && elem) ) {
						matchedCount--;
					}

					// Lengthen the array for every element, matched or not
					if ( seed ) {
						unmatched.push( elem );
					}
				}
			}

			// Apply set filters to unmatched elements
			matchedCount += i;
			if ( bySet && i !== matchedCount ) {
				j = 0;
				while ( (matcher = setMatchers[j++]) ) {
					matcher( unmatched, setMatched, context, xml );
				}

				if ( seed ) {
					// Reintegrate element matches to eliminate the need for sorting
					if ( matchedCount > 0 ) {
						while ( i-- ) {
							if ( !(unmatched[i] || setMatched[i]) ) {
								setMatched[i] = pop.call( results );
							}
						}
					}

					// Discard index placeholder values to get only actual matches
					setMatched = condense( setMatched );
				}

				// Add matches to results
				push.apply( results, setMatched );

				// Seedless set matches succeeding multiple successful matchers stipulate sorting
				if ( outermost && !seed && setMatched.length > 0 &&
					( matchedCount + setMatchers.length ) > 1 ) {

					Sizzle.uniqueSort( results );
				}
			}

			// Override manipulation of globals by nested matchers
			if ( outermost ) {
				dirruns = dirrunsUnique;
				outermostContext = contextBackup;
			}

			return unmatched;
		};

	return bySet ?
		markFunction( superMatcher ) :
		superMatcher;
}

compile = Sizzle.compile = function( selector, match /* Internal Use Only */ ) {
	var i,
		setMatchers = [],
		elementMatchers = [],
		cached = compilerCache[ selector + " " ];

	if ( !cached ) {
		// Generate a function of recursive functions that can be used to check each element
		if ( !match ) {
			match = tokenize( selector );
		}
		i = match.length;
		while ( i-- ) {
			cached = matcherFromTokens( match[i] );
			if ( cached[ expando ] ) {
				setMatchers.push( cached );
			} else {
				elementMatchers.push( cached );
			}
		}

		// Cache the compiled function
		cached = compilerCache( selector, matcherFromGroupMatchers( elementMatchers, setMatchers ) );

		// Save selector and tokenization
		cached.selector = selector;
	}
	return cached;
};

/**
 * A low-level selection function that works with Sizzle's compiled
 *  selector functions
 * @param {String|Function} selector A selector or a pre-compiled
 *  selector function built with Sizzle.compile
 * @param {Element} context
 * @param {Array} [results]
 * @param {Array} [seed] A set of elements to match against
 */
select = Sizzle.select = function( selector, context, results, seed ) {
	var i, tokens, token, type, find,
		compiled = typeof selector === "function" && selector,
		match = !seed && tokenize( (selector = compiled.selector || selector) );

	results = results || [];

	// Try to minimize operations if there is no seed and only one group
	if ( match.length === 1 ) {

		// Take a shortcut and set the context if the root selector is an ID
		tokens = match[0] = match[0].slice( 0 );
		if ( tokens.length > 2 && (token = tokens[0]).type === "ID" &&
				support.getById && context.nodeType === 9 && documentIsHTML &&
				Expr.relative[ tokens[1].type ] ) {

			context = ( Expr.find["ID"]( token.matches[0].replace(runescape, funescape), context ) || [] )[0];
			if ( !context ) {
				return results;

			// Precompiled matchers will still verify ancestry, so step up a level
			} else if ( compiled ) {
				context = context.parentNode;
			}

			selector = selector.slice( tokens.shift().value.length );
		}

		// Fetch a seed set for right-to-left matching
		i = matchExpr["needsContext"].test( selector ) ? 0 : tokens.length;
		while ( i-- ) {
			token = tokens[i];

			// Abort if we hit a combinator
			if ( Expr.relative[ (type = token.type) ] ) {
				break;
			}
			if ( (find = Expr.find[ type ]) ) {
				// Search, expanding context for leading sibling combinators
				if ( (seed = find(
					token.matches[0].replace( runescape, funescape ),
					rsibling.test( tokens[0].type ) && testContext( context.parentNode ) || context
				)) ) {

					// If seed is empty or no tokens remain, we can return early
					tokens.splice( i, 1 );
					selector = seed.length && toSelector( tokens );
					if ( !selector ) {
						push.apply( results, seed );
						return results;
					}

					break;
				}
			}
		}
	}

	// Compile and execute a filtering function if one is not provided
	// Provide `match` to avoid retokenization if we modified the selector above
	( compiled || compile( selector, match ) )(
		seed,
		context,
		!documentIsHTML,
		results,
		rsibling.test( selector ) && testContext( context.parentNode ) || context
	);
	return results;
};

// One-time assignments

// Sort stability
support.sortStable = expando.split("").sort( sortOrder ).join("") === expando;

// Support: Chrome 14-35+
// Always assume duplicates if they aren't passed to the comparison function
support.detectDuplicates = !!hasDuplicate;

// Initialize against the default document
setDocument();

// Support: Webkit<537.32 - Safari 6.0.3/Chrome 25 (fixed in Chrome 27)
// Detached nodes confoundingly follow *each other*
support.sortDetached = assert(function( div1 ) {
	// Should return 1, but returns 4 (following)
	return div1.compareDocumentPosition( document.createElement("div") ) & 1;
});

// Support: IE<8
// Prevent attribute/property "interpolation"
// http://msdn.microsoft.com/en-us/library/ms536429%28VS.85%29.aspx
if ( !assert(function( div ) {
	div.innerHTML = "<a href='#'></a>";
	return div.firstChild.getAttribute("href") === "#" ;
}) ) {
	addHandle( "type|href|height|width", function( elem, name, isXML ) {
		if ( !isXML ) {
			return elem.getAttribute( name, name.toLowerCase() === "type" ? 1 : 2 );
		}
	});
}

// Support: IE<9
// Use defaultValue in place of getAttribute("value")
if ( !support.attributes || !assert(function( div ) {
	div.innerHTML = "<input/>";
	div.firstChild.setAttribute( "value", "" );
	return div.firstChild.getAttribute( "value" ) === "";
}) ) {
	addHandle( "value", function( elem, name, isXML ) {
		if ( !isXML && elem.nodeName.toLowerCase() === "input" ) {
			return elem.defaultValue;
		}
	});
}

// Support: IE<9
// Use getAttributeNode to fetch booleans when getAttribute lies
if ( !assert(function( div ) {
	return div.getAttribute("disabled") == null;
}) ) {
	addHandle( booleans, function( elem, name, isXML ) {
		var val;
		if ( !isXML ) {
			return elem[ name ] === true ? name.toLowerCase() :
					(val = elem.getAttributeNode( name )) && val.specified ?
					val.value :
				null;
		}
	});
}

return Sizzle;

})( window );



jQuery.find = Sizzle;
jQuery.expr = Sizzle.selectors;
jQuery.expr[":"] = jQuery.expr.pseudos;
jQuery.unique = Sizzle.uniqueSort;
jQuery.text = Sizzle.getText;
jQuery.isXMLDoc = Sizzle.isXML;
jQuery.contains = Sizzle.contains;



var rneedsContext = jQuery.expr.match.needsContext;

var rsingleTag = (/^<(\w+)\s*\/?>(?:<\/\1>|)$/);



var risSimple = /^.[^:#\[\.,]*$/;

// Implement the identical functionality for filter and not
function winnow( elements, qualifier, not ) {
	if ( jQuery.isFunction( qualifier ) ) {
		return jQuery.grep( elements, function( elem, i ) {
			/* jshint -W018 */
			return !!qualifier.call( elem, i, elem ) !== not;
		});

	}

	if ( qualifier.nodeType ) {
		return jQuery.grep( elements, function( elem ) {
			return ( elem === qualifier ) !== not;
		});

	}

	if ( typeof qualifier === "string" ) {
		if ( risSimple.test( qualifier ) ) {
			return jQuery.filter( qualifier, elements, not );
		}

		qualifier = jQuery.filter( qualifier, elements );
	}

	return jQuery.grep( elements, function( elem ) {
		return ( jQuery.inArray( elem, qualifier ) >= 0 ) !== not;
	});
}

jQuery.filter = function( expr, elems, not ) {
	var elem = elems[ 0 ];

	if ( not ) {
		expr = ":not(" + expr + ")";
	}

	return elems.length === 1 && elem.nodeType === 1 ?
		jQuery.find.matchesSelector( elem, expr ) ? [ elem ] : [] :
		jQuery.find.matches( expr, jQuery.grep( elems, function( elem ) {
			return elem.nodeType === 1;
		}));
};

jQuery.fn.extend({
	find: function( selector ) {
		var i,
			ret = [],
			self = this,
			len = self.length;

		if ( typeof selector !== "string" ) {
			return this.pushStack( jQuery( selector ).filter(function() {
				for ( i = 0; i < len; i++ ) {
					if ( jQuery.contains( self[ i ], this ) ) {
						return true;
					}
				}
			}) );
		}

		for ( i = 0; i < len; i++ ) {
			jQuery.find( selector, self[ i ], ret );
		}

		// Needed because $( selector, context ) becomes $( context ).find( selector )
		ret = this.pushStack( len > 1 ? jQuery.unique( ret ) : ret );
		ret.selector = this.selector ? this.selector + " " + selector : selector;
		return ret;
	},
	filter: function( selector ) {
		return this.pushStack( winnow(this, selector || [], false) );
	},
	not: function( selector ) {
		return this.pushStack( winnow(this, selector || [], true) );
	},
	is: function( selector ) {
		return !!winnow(
			this,

			// If this is a positional/relative selector, check membership in the returned set
			// so $("p:first").is("p:last") won't return true for a doc with two "p".
			typeof selector === "string" && rneedsContext.test( selector ) ?
				jQuery( selector ) :
				selector || [],
			false
		).length;
	}
});


// Initialize a jQuery object


// A central reference to the root jQuery(document)
var rootjQuery,

	// Use the correct document accordingly with window argument (sandbox)
	document = window.document,

	// A simple way to check for HTML strings
	// Prioritize #id over <tag> to avoid XSS via location.hash (#9521)
	// Strict HTML recognition (#11290: must start with <)
	rquickExpr = /^(?:\s*(<[\w\W]+>)[^>]*|#([\w-]*))$/,

	init = jQuery.fn.init = function( selector, context ) {
		var match, elem;

		// HANDLE: $(""), $(null), $(undefined), $(false)
		if ( !selector ) {
			return this;
		}

		// Handle HTML strings
		if ( typeof selector === "string" ) {
			if ( selector.charAt(0) === "<" && selector.charAt( selector.length - 1 ) === ">" && selector.length >= 3 ) {
				// Assume that strings that start and end with <> are HTML and skip the regex check
				match = [ null, selector, null ];

			} else {
				match = rquickExpr.exec( selector );
			}

			// Match html or make sure no context is specified for #id
			if ( match && (match[1] || !context) ) {

				// HANDLE: $(html) -> $(array)
				if ( match[1] ) {
					context = context instanceof jQuery ? context[0] : context;

					// scripts is true for back-compat
					// Intentionally let the error be thrown if parseHTML is not present
					jQuery.merge( this, jQuery.parseHTML(
						match[1],
						context && context.nodeType ? context.ownerDocument || context : document,
						true
					) );

					// HANDLE: $(html, props)
					if ( rsingleTag.test( match[1] ) && jQuery.isPlainObject( context ) ) {
						for ( match in context ) {
							// Properties of context are called as methods if possible
							if ( jQuery.isFunction( this[ match ] ) ) {
								this[ match ]( context[ match ] );

							// ...and otherwise set as attributes
							} else {
								this.attr( match, context[ match ] );
							}
						}
					}

					return this;

				// HANDLE: $(#id)
				} else {
					elem = document.getElementById( match[2] );

					// Check parentNode to catch when Blackberry 4.6 returns
					// nodes that are no longer in the document #6963
					if ( elem && elem.parentNode ) {
						// Handle the case where IE and Opera return items
						// by name instead of ID
						if ( elem.id !== match[2] ) {
							return rootjQuery.find( selector );
						}

						// Otherwise, we inject the element directly into the jQuery object
						this.length = 1;
						this[0] = elem;
					}

					this.context = document;
					this.selector = selector;
					return this;
				}

			// HANDLE: $(expr, $(...))
			} else if ( !context || context.jquery ) {
				return ( context || rootjQuery ).find( selector );

			// HANDLE: $(expr, context)
			// (which is just equivalent to: $(context).find(expr)
			} else {
				return this.constructor( context ).find( selector );
			}

		// HANDLE: $(DOMElement)
		} else if ( selector.nodeType ) {
			this.context = this[0] = selector;
			this.length = 1;
			return this;

		// HANDLE: $(function)
		// Shortcut for document ready
		} else if ( jQuery.isFunction( selector ) ) {
			return typeof rootjQuery.ready !== "undefined" ?
				rootjQuery.ready( selector ) :
				// Execute immediately if ready is not present
				selector( jQuery );
		}

		if ( selector.selector !== undefined ) {
			this.selector = selector.selector;
			this.context = selector.context;
		}

		return jQuery.makeArray( selector, this );
	};

// Give the init function the jQuery prototype for later instantiation
init.prototype = jQuery.fn;

// Initialize central reference
rootjQuery = jQuery( document );


var rparentsprev = /^(?:parents|prev(?:Until|All))/,
	// methods guaranteed to produce a unique set when starting from a unique set
	guaranteedUnique = {
		children: true,
		contents: true,
		next: true,
		prev: true
	};

jQuery.extend({
	dir: function( elem, dir, until ) {
		var matched = [],
			cur = elem[ dir ];

		while ( cur && cur.nodeType !== 9 && (until === undefined || cur.nodeType !== 1 || !jQuery( cur ).is( until )) ) {
			if ( cur.nodeType === 1 ) {
				matched.push( cur );
			}
			cur = cur[dir];
		}
		return matched;
	},

	sibling: function( n, elem ) {
		var r = [];

		for ( ; n; n = n.nextSibling ) {
			if ( n.nodeType === 1 && n !== elem ) {
				r.push( n );
			}
		}

		return r;
	}
});

jQuery.fn.extend({
	has: function( target ) {
		var i,
			targets = jQuery( target, this ),
			len = targets.length;

		return this.filter(function() {
			for ( i = 0; i < len; i++ ) {
				if ( jQuery.contains( this, targets[i] ) ) {
					return true;
				}
			}
		});
	},

	closest: function( selectors, context ) {
		var cur,
			i = 0,
			l = this.length,
			matched = [],
			pos = rneedsContext.test( selectors ) || typeof selectors !== "string" ?
				jQuery( selectors, context || this.context ) :
				0;

		for ( ; i < l; i++ ) {
			for ( cur = this[i]; cur && cur !== context; cur = cur.parentNode ) {
				// Always skip document fragments
				if ( cur.nodeType < 11 && (pos ?
					pos.index(cur) > -1 :

					// Don't pass non-elements to Sizzle
					cur.nodeType === 1 &&
						jQuery.find.matchesSelector(cur, selectors)) ) {

					matched.push( cur );
					break;
				}
			}
		}

		return this.pushStack( matched.length > 1 ? jQuery.unique( matched ) : matched );
	},

	// Determine the position of an element within
	// the matched set of elements
	index: function( elem ) {

		// No argument, return index in parent
		if ( !elem ) {
			return ( this[0] && this[0].parentNode ) ? this.first().prevAll().length : -1;
		}

		// index in selector
		if ( typeof elem === "string" ) {
			return jQuery.inArray( this[0], jQuery( elem ) );
		}

		// Locate the position of the desired element
		return jQuery.inArray(
			// If it receives a jQuery object, the first element is used
			elem.jquery ? elem[0] : elem, this );
	},

	add: function( selector, context ) {
		return this.pushStack(
			jQuery.unique(
				jQuery.merge( this.get(), jQuery( selector, context ) )
			)
		);
	},

	addBack: function( selector ) {
		return this.add( selector == null ?
			this.prevObject : this.prevObject.filter(selector)
		);
	}
});

function sibling( cur, dir ) {
	do {
		cur = cur[ dir ];
	} while ( cur && cur.nodeType !== 1 );

	return cur;
}

jQuery.each({
	parent: function( elem ) {
		var parent = elem.parentNode;
		return parent && parent.nodeType !== 11 ? parent : null;
	},
	parents: function( elem ) {
		return jQuery.dir( elem, "parentNode" );
	},
	parentsUntil: function( elem, i, until ) {
		return jQuery.dir( elem, "parentNode", until );
	},
	next: function( elem ) {
		return sibling( elem, "nextSibling" );
	},
	prev: function( elem ) {
		return sibling( elem, "previousSibling" );
	},
	nextAll: function( elem ) {
		return jQuery.dir( elem, "nextSibling" );
	},
	prevAll: function( elem ) {
		return jQuery.dir( elem, "previousSibling" );
	},
	nextUntil: function( elem, i, until ) {
		return jQuery.dir( elem, "nextSibling", until );
	},
	prevUntil: function( elem, i, until ) {
		return jQuery.dir( elem, "previousSibling", until );
	},
	siblings: function( elem ) {
		return jQuery.sibling( ( elem.parentNode || {} ).firstChild, elem );
	},
	children: function( elem ) {
		return jQuery.sibling( elem.firstChild );
	},
	contents: function( elem ) {
		return jQuery.nodeName( elem, "iframe" ) ?
			elem.contentDocument || elem.contentWindow.document :
			jQuery.merge( [], elem.childNodes );
	}
}, function( name, fn ) {
	jQuery.fn[ name ] = function( until, selector ) {
		var ret = jQuery.map( this, fn, until );

		if ( name.slice( -5 ) !== "Until" ) {
			selector = until;
		}

		if ( selector && typeof selector === "string" ) {
			ret = jQuery.filter( selector, ret );
		}

		if ( this.length > 1 ) {
			// Remove duplicates
			if ( !guaranteedUnique[ name ] ) {
				ret = jQuery.unique( ret );
			}

			// Reverse order for parents* and prev-derivatives
			if ( rparentsprev.test( name ) ) {
				ret = ret.reverse();
			}
		}

		return this.pushStack( ret );
	};
});
var rnotwhite = (/\S+/g);



// String to Object options format cache
var optionsCache = {};

// Convert String-formatted options into Object-formatted ones and store in cache
function createOptions( options ) {
	var object = optionsCache[ options ] = {};
	jQuery.each( options.match( rnotwhite ) || [], function( _, flag ) {
		object[ flag ] = true;
	});
	return object;
}

/*
 * Create a callback list using the following parameters:
 *
 *	options: an optional list of space-separated options that will change how
 *			the callback list behaves or a more traditional option object
 *
 * By default a callback list will act like an event callback list and can be
 * "fired" multiple times.
 *
 * Possible options:
 *
 *	once:			will ensure the callback list can only be fired once (like a Deferred)
 *
 *	memory:			will keep track of previous values and will call any callback added
 *					after the list has been fired right away with the latest "memorized"
 *					values (like a Deferred)
 *
 *	unique:			will ensure a callback can only be added once (no duplicate in the list)
 *
 *	stopOnFalse:	interrupt callings when a callback returns false
 *
 */
jQuery.Callbacks = function( options ) {

	// Convert options from String-formatted to Object-formatted if needed
	// (we check in cache first)
	options = typeof options === "string" ?
		( optionsCache[ options ] || createOptions( options ) ) :
		jQuery.extend( {}, options );

	var // Flag to know if list is currently firing
		firing,
		// Last fire value (for non-forgettable lists)
		memory,
		// Flag to know if list was already fired
		fired,
		// End of the loop when firing
		firingLength,
		// Index of currently firing callback (modified by remove if needed)
		firingIndex,
		// First callback to fire (used internally by add and fireWith)
		firingStart,
		// Actual callback list
		list = [],
		// Stack of fire calls for repeatable lists
		stack = !options.once && [],
		// Fire callbacks
		fire = function( data ) {
			memory = options.memory && data;
			fired = true;
			firingIndex = firingStart || 0;
			firingStart = 0;
			firingLength = list.length;
			firing = true;
			for ( ; list && firingIndex < firingLength; firingIndex++ ) {
				if ( list[ firingIndex ].apply( data[ 0 ], data[ 1 ] ) === false && options.stopOnFalse ) {
					memory = false; // To prevent further calls using add
					break;
				}
			}
			firing = false;
			if ( list ) {
				if ( stack ) {
					if ( stack.length ) {
						fire( stack.shift() );
					}
				} else if ( memory ) {
					list = [];
				} else {
					self.disable();
				}
			}
		},
		// Actual Callbacks object
		self = {
			// Add a callback or a collection of callbacks to the list
			add: function() {
				if ( list ) {
					// First, we save the current length
					var start = list.length;
					(function add( args ) {
						jQuery.each( args, function( _, arg ) {
							var type = jQuery.type( arg );
							if ( type === "function" ) {
								if ( !options.unique || !self.has( arg ) ) {
									list.push( arg );
								}
							} else if ( arg && arg.length && type !== "string" ) {
								// Inspect recursively
								add( arg );
							}
						});
					})( arguments );
					// Do we need to add the callbacks to the
					// current firing batch?
					if ( firing ) {
						firingLength = list.length;
					// With memory, if we're not firing then
					// we should call right away
					} else if ( memory ) {
						firingStart = start;
						fire( memory );
					}
				}
				return this;
			},
			// Remove a callback from the list
			remove: function() {
				if ( list ) {
					jQuery.each( arguments, function( _, arg ) {
						var index;
						while ( ( index = jQuery.inArray( arg, list, index ) ) > -1 ) {
							list.splice( index, 1 );
							// Handle firing indexes
							if ( firing ) {
								if ( index <= firingLength ) {
									firingLength--;
								}
								if ( index <= firingIndex ) {
									firingIndex--;
								}
							}
						}
					});
				}
				return this;
			},
			// Check if a given callback is in the list.
			// If no argument is given, return whether or not list has callbacks attached.
			has: function( fn ) {
				return fn ? jQuery.inArray( fn, list ) > -1 : !!( list && list.length );
			},
			// Remove all callbacks from the list
			empty: function() {
				list = [];
				firingLength = 0;
				return this;
			},
			// Have the list do nothing anymore
			disable: function() {
				list = stack = memory = undefined;
				return this;
			},
			// Is it disabled?
			disabled: function() {
				return !list;
			},
			// Lock the list in its current state
			lock: function() {
				stack = undefined;
				if ( !memory ) {
					self.disable();
				}
				return this;
			},
			// Is it locked?
			locked: function() {
				return !stack;
			},
			// Call all callbacks with the given context and arguments
			fireWith: function( context, args ) {
				if ( list && ( !fired || stack ) ) {
					args = args || [];
					args = [ context, args.slice ? args.slice() : args ];
					if ( firing ) {
						stack.push( args );
					} else {
						fire( args );
					}
				}
				return this;
			},
			// Call all the callbacks with the given arguments
			fire: function() {
				self.fireWith( this, arguments );
				return this;
			},
			// To know if the callbacks have already been called at least once
			fired: function() {
				return !!fired;
			}
		};

	return self;
};


jQuery.extend({

	Deferred: function( func ) {
		var tuples = [
				// action, add listener, listener list, final state
				[ "resolve", "done", jQuery.Callbacks("once memory"), "resolved" ],
				[ "reject", "fail", jQuery.Callbacks("once memory"), "rejected" ],
				[ "notify", "progress", jQuery.Callbacks("memory") ]
			],
			state = "pending",
			promise = {
				state: function() {
					return state;
				},
				always: function() {
					deferred.done( arguments ).fail( arguments );
					return this;
				},
				then: function( /* fnDone, fnFail, fnProgress */ ) {
					var fns = arguments;
					return jQuery.Deferred(function( newDefer ) {
						jQuery.each( tuples, function( i, tuple ) {
							var fn = jQuery.isFunction( fns[ i ] ) && fns[ i ];
							// deferred[ done | fail | progress ] for forwarding actions to newDefer
							deferred[ tuple[1] ](function() {
								var returned = fn && fn.apply( this, arguments );
								if ( returned && jQuery.isFunction( returned.promise ) ) {
									returned.promise()
										.done( newDefer.resolve )
										.fail( newDefer.reject )
										.progress( newDefer.notify );
								} else {
									newDefer[ tuple[ 0 ] + "With" ]( this === promise ? newDefer.promise() : this, fn ? [ returned ] : arguments );
								}
							});
						});
						fns = null;
					}).promise();
				},
				// Get a promise for this deferred
				// If obj is provided, the promise aspect is added to the object
				promise: function( obj ) {
					return obj != null ? jQuery.extend( obj, promise ) : promise;
				}
			},
			deferred = {};

		// Keep pipe for back-compat
		promise.pipe = promise.then;

		// Add list-specific methods
		jQuery.each( tuples, function( i, tuple ) {
			var list = tuple[ 2 ],
				stateString = tuple[ 3 ];

			// promise[ done | fail | progress ] = list.add
			promise[ tuple[1] ] = list.add;

			// Handle state
			if ( stateString ) {
				list.add(function() {
					// state = [ resolved | rejected ]
					state = stateString;

				// [ reject_list | resolve_list ].disable; progress_list.lock
				}, tuples[ i ^ 1 ][ 2 ].disable, tuples[ 2 ][ 2 ].lock );
			}

			// deferred[ resolve | reject | notify ]
			deferred[ tuple[0] ] = function() {
				deferred[ tuple[0] + "With" ]( this === deferred ? promise : this, arguments );
				return this;
			};
			deferred[ tuple[0] + "With" ] = list.fireWith;
		});

		// Make the deferred a promise
		promise.promise( deferred );

		// Call given func if any
		if ( func ) {
			func.call( deferred, deferred );
		}

		// All done!
		return deferred;
	},

	// Deferred helper
	when: function( subordinate /* , ..., subordinateN */ ) {
		var i = 0,
			resolveValues = slice.call( arguments ),
			length = resolveValues.length,

			// the count of uncompleted subordinates
			remaining = length !== 1 || ( subordinate && jQuery.isFunction( subordinate.promise ) ) ? length : 0,

			// the master Deferred. If resolveValues consist of only a single Deferred, just use that.
			deferred = remaining === 1 ? subordinate : jQuery.Deferred(),

			// Update function for both resolve and progress values
			updateFunc = function( i, contexts, values ) {
				return function( value ) {
					contexts[ i ] = this;
					values[ i ] = arguments.length > 1 ? slice.call( arguments ) : value;
					if ( values === progressValues ) {
						deferred.notifyWith( contexts, values );

					} else if ( !(--remaining) ) {
						deferred.resolveWith( contexts, values );
					}
				};
			},

			progressValues, progressContexts, resolveContexts;

		// add listeners to Deferred subordinates; treat others as resolved
		if ( length > 1 ) {
			progressValues = new Array( length );
			progressContexts = new Array( length );
			resolveContexts = new Array( length );
			for ( ; i < length; i++ ) {
				if ( resolveValues[ i ] && jQuery.isFunction( resolveValues[ i ].promise ) ) {
					resolveValues[ i ].promise()
						.done( updateFunc( i, resolveContexts, resolveValues ) )
						.fail( deferred.reject )
						.progress( updateFunc( i, progressContexts, progressValues ) );
				} else {
					--remaining;
				}
			}
		}

		// if we're not waiting on anything, resolve the master
		if ( !remaining ) {
			deferred.resolveWith( resolveContexts, resolveValues );
		}

		return deferred.promise();
	}
});


// The deferred used on DOM ready
var readyList;

jQuery.fn.ready = function( fn ) {
	// Add the callback
	jQuery.ready.promise().done( fn );

	return this;
};

jQuery.extend({
	// Is the DOM ready to be used? Set to true once it occurs.
	isReady: false,

	// A counter to track how many items to wait for before
	// the ready event fires. See #6781
	readyWait: 1,

	// Hold (or release) the ready event
	holdReady: function( hold ) {
		if ( hold ) {
			jQuery.readyWait++;
		} else {
			jQuery.ready( true );
		}
	},

	// Handle when the DOM is ready
	ready: function( wait ) {

		// Abort if there are pending holds or we're already ready
		if ( wait === true ? --jQuery.readyWait : jQuery.isReady ) {
			return;
		}

		// Make sure body exists, at least, in case IE gets a little overzealous (ticket #5443).
		if ( !document.body ) {
			return setTimeout( jQuery.ready );
		}

		// Remember that the DOM is ready
		jQuery.isReady = true;

		// If a normal DOM Ready event fired, decrement, and wait if need be
		if ( wait !== true && --jQuery.readyWait > 0 ) {
			return;
		}

		// If there are functions bound, to execute
		readyList.resolveWith( document, [ jQuery ] );

		// Trigger any bound ready events
		if ( jQuery.fn.triggerHandler ) {
			jQuery( document ).triggerHandler( "ready" );
			jQuery( document ).off( "ready" );
		}
	}
});

/**
 * Clean-up method for dom ready events
 */
function detach() {
	if ( document.addEventListener ) {
		document.removeEventListener( "DOMContentLoaded", completed, false );
		window.removeEventListener( "load", completed, false );

	} else {
		document.detachEvent( "onreadystatechange", completed );
		window.detachEvent( "onload", completed );
	}
}

/**
 * The ready event handler and self cleanup method
 */
function completed() {
	// readyState === "complete" is good enough for us to call the dom ready in oldIE
	if ( document.addEventListener || event.type === "load" || document.readyState === "complete" ) {
		detach();
		jQuery.ready();
	}
}

jQuery.ready.promise = function( obj ) {
	if ( !readyList ) {

		readyList = jQuery.Deferred();

		// Catch cases where $(document).ready() is called after the browser event has already occurred.
		// we once tried to use readyState "interactive" here, but it caused issues like the one
		// discovered by ChrisS here: http://bugs.jquery.com/ticket/12282#comment:15
		if ( document.readyState === "complete" ) {
			// Handle it asynchronously to allow scripts the opportunity to delay ready
			setTimeout( jQuery.ready );

		// Standards-based browsers support DOMContentLoaded
		} else if ( document.addEventListener ) {
			// Use the handy event callback
			document.addEventListener( "DOMContentLoaded", completed, false );

			// A fallback to window.onload, that will always work
			window.addEventListener( "load", completed, false );

		// If IE event model is used
		} else {
			// Ensure firing before onload, maybe late but safe also for iframes
			document.attachEvent( "onreadystatechange", completed );

			// A fallback to window.onload, that will always work
			window.attachEvent( "onload", completed );

			// If IE and not a frame
			// continually check to see if the document is ready
			var top = false;

			try {
				top = window.frameElement == null && document.documentElement;
			} catch(e) {}

			if ( top && top.doScroll ) {
				(function doScrollCheck() {
					if ( !jQuery.isReady ) {

						try {
							// Use the trick by Diego Perini
							// http://javascript.nwbox.com/IEContentLoaded/
							top.doScroll("left");
						} catch(e) {
							return setTimeout( doScrollCheck, 50 );
						}

						// detach all dom ready events
						detach();

						// and execute any waiting functions
						jQuery.ready();
					}
				})();
			}
		}
	}
	return readyList.promise( obj );
};


var strundefined = typeof undefined;



// Support: IE<9
// Iteration over object's inherited properties before its own
var i;
for ( i in jQuery( support ) ) {
	break;
}
support.ownLast = i !== "0";

// Note: most support tests are defined in their respective modules.
// false until the test is run
support.inlineBlockNeedsLayout = false;

// Execute ASAP in case we need to set body.style.zoom
jQuery(function() {
	// Minified: var a,b,c,d
	var val, div, body, container;

	body = document.getElementsByTagName( "body" )[ 0 ];
	if ( !body || !body.style ) {
		// Return for frameset docs that don't have a body
		return;
	}

	// Setup
	div = document.createElement( "div" );
	container = document.createElement( "div" );
	container.style.cssText = "position:absolute;border:0;width:0;height:0;top:0;left:-9999px";
	body.appendChild( container ).appendChild( div );

	if ( typeof div.style.zoom !== strundefined ) {
		// Support: IE<8
		// Check if natively block-level elements act like inline-block
		// elements when setting their display to 'inline' and giving
		// them layout
		div.style.cssText = "display:inline;margin:0;border:0;padding:1px;width:1px;zoom:1";

		support.inlineBlockNeedsLayout = val = div.offsetWidth === 3;
		if ( val ) {
			// Prevent IE 6 from affecting layout for positioned elements #11048
			// Prevent IE from shrinking the body in IE 7 mode #12869
			// Support: IE<8
			body.style.zoom = 1;
		}
	}

	body.removeChild( container );
});




(function() {
	var div = document.createElement( "div" );

	// Execute the test only if not already executed in another module.
	if (support.deleteExpando == null) {
		// Support: IE<9
		support.deleteExpando = true;
		try {
			delete div.test;
		} catch( e ) {
			support.deleteExpando = false;
		}
	}

	// Null elements to avoid leaks in IE.
	div = null;
})();


/**
 * Determines whether an object can have data
 */
jQuery.acceptData = function( elem ) {
	var noData = jQuery.noData[ (elem.nodeName + " ").toLowerCase() ],
		nodeType = +elem.nodeType || 1;

	// Do not set data on non-element DOM nodes because it will not be cleared (#8335).
	return nodeType !== 1 && nodeType !== 9 ?
		false :

		// Nodes accept data unless otherwise specified; rejection can be conditional
		!noData || noData !== true && elem.getAttribute("classid") === noData;
};


var rbrace = /^(?:\{[\w\W]*\}|\[[\w\W]*\])$/,
	rmultiDash = /([A-Z])/g;

function dataAttr( elem, key, data ) {
	// If nothing was found internally, try to fetch any
	// data from the HTML5 data-* attribute
	if ( data === undefined && elem.nodeType === 1 ) {

		var name = "data-" + key.replace( rmultiDash, "-$1" ).toLowerCase();

		data = elem.getAttribute( name );

		if ( typeof data === "string" ) {
			try {
				data = data === "true" ? true :
					data === "false" ? false :
					data === "null" ? null :
					// Only convert to a number if it doesn't change the string
					+data + "" === data ? +data :
					rbrace.test( data ) ? jQuery.parseJSON( data ) :
					data;
			} catch( e ) {}

			// Make sure we set the data so it isn't changed later
			jQuery.data( elem, key, data );

		} else {
			data = undefined;
		}
	}

	return data;
}

// checks a cache object for emptiness
function isEmptyDataObject( obj ) {
	var name;
	for ( name in obj ) {

		// if the public data object is empty, the private is still empty
		if ( name === "data" && jQuery.isEmptyObject( obj[name] ) ) {
			continue;
		}
		if ( name !== "toJSON" ) {
			return false;
		}
	}

	return true;
}

function internalData( elem, name, data, pvt /* Internal Use Only */ ) {
	if ( !jQuery.acceptData( elem ) ) {
		return;
	}

	var ret, thisCache,
		internalKey = jQuery.expando,

		// We have to handle DOM nodes and JS objects differently because IE6-7
		// can't GC object references properly across the DOM-JS boundary
		isNode = elem.nodeType,

		// Only DOM nodes need the global jQuery cache; JS object data is
		// attached directly to the object so GC can occur automatically
		cache = isNode ? jQuery.cache : elem,

		// Only defining an ID for JS objects if its cache already exists allows
		// the code to shortcut on the same path as a DOM node with no cache
		id = isNode ? elem[ internalKey ] : elem[ internalKey ] && internalKey;

	// Avoid doing any more work than we need to when trying to get data on an
	// object that has no data at all
	if ( (!id || !cache[id] || (!pvt && !cache[id].data)) && data === undefined && typeof name === "string" ) {
		return;
	}

	if ( !id ) {
		// Only DOM nodes need a new unique ID for each element since their data
		// ends up in the global cache
		if ( isNode ) {
			id = elem[ internalKey ] = deletedIds.pop() || jQuery.guid++;
		} else {
			id = internalKey;
		}
	}

	if ( !cache[ id ] ) {
		// Avoid exposing jQuery metadata on plain JS objects when the object
		// is serialized using JSON.stringify
		cache[ id ] = isNode ? {} : { toJSON: jQuery.noop };
	}

	// An object can be passed to jQuery.data instead of a key/value pair; this gets
	// shallow copied over onto the existing cache
	if ( typeof name === "object" || typeof name === "function" ) {
		if ( pvt ) {
			cache[ id ] = jQuery.extend( cache[ id ], name );
		} else {
			cache[ id ].data = jQuery.extend( cache[ id ].data, name );
		}
	}

	thisCache = cache[ id ];

	// jQuery data() is stored in a separate object inside the object's internal data
	// cache in order to avoid key collisions between internal data and user-defined
	// data.
	if ( !pvt ) {
		if ( !thisCache.data ) {
			thisCache.data = {};
		}

		thisCache = thisCache.data;
	}

	if ( data !== undefined ) {
		thisCache[ jQuery.camelCase( name ) ] = data;
	}

	// Check for both converted-to-camel and non-converted data property names
	// If a data property was specified
	if ( typeof name === "string" ) {

		// First Try to find as-is property data
		ret = thisCache[ name ];

		// Test for null|undefined property data
		if ( ret == null ) {

			// Try to find the camelCased property
			ret = thisCache[ jQuery.camelCase( name ) ];
		}
	} else {
		ret = thisCache;
	}

	return ret;
}

function internalRemoveData( elem, name, pvt ) {
	if ( !jQuery.acceptData( elem ) ) {
		return;
	}

	var thisCache, i,
		isNode = elem.nodeType,

		// See jQuery.data for more information
		cache = isNode ? jQuery.cache : elem,
		id = isNode ? elem[ jQuery.expando ] : jQuery.expando;

	// If there is already no cache entry for this object, there is no
	// purpose in continuing
	if ( !cache[ id ] ) {
		return;
	}

	if ( name ) {

		thisCache = pvt ? cache[ id ] : cache[ id ].data;

		if ( thisCache ) {

			// Support array or space separated string names for data keys
			if ( !jQuery.isArray( name ) ) {

				// try the string as a key before any manipulation
				if ( name in thisCache ) {
					name = [ name ];
				} else {

					// split the camel cased version by spaces unless a key with the spaces exists
					name = jQuery.camelCase( name );
					if ( name in thisCache ) {
						name = [ name ];
					} else {
						name = name.split(" ");
					}
				}
			} else {
				// If "name" is an array of keys...
				// When data is initially created, via ("key", "val") signature,
				// keys will be converted to camelCase.
				// Since there is no way to tell _how_ a key was added, remove
				// both plain key and camelCase key. #12786
				// This will only penalize the array argument path.
				name = name.concat( jQuery.map( name, jQuery.camelCase ) );
			}

			i = name.length;
			while ( i-- ) {
				delete thisCache[ name[i] ];
			}

			// If there is no data left in the cache, we want to continue
			// and let the cache object itself get destroyed
			if ( pvt ? !isEmptyDataObject(thisCache) : !jQuery.isEmptyObject(thisCache) ) {
				return;
			}
		}
	}

	// See jQuery.data for more information
	if ( !pvt ) {
		delete cache[ id ].data;

		// Don't destroy the parent cache unless the internal data object
		// had been the only thing left in it
		if ( !isEmptyDataObject( cache[ id ] ) ) {
			return;
		}
	}

	// Destroy the cache
	if ( isNode ) {
		jQuery.cleanData( [ elem ], true );

	// Use delete when supported for expandos or `cache` is not a window per isWindow (#10080)
	/* jshint eqeqeq: false */
	} else if ( support.deleteExpando || cache != cache.window ) {
		/* jshint eqeqeq: true */
		delete cache[ id ];

	// When all else fails, null
	} else {
		cache[ id ] = null;
	}
}

jQuery.extend({
	cache: {},

	// The following elements (space-suffixed to avoid Object.prototype collisions)
	// throw uncatchable exceptions if you attempt to set expando properties
	noData: {
		"applet ": true,
		"embed ": true,
		// ...but Flash objects (which have this classid) *can* handle expandos
		"object ": "clsid:D27CDB6E-AE6D-11cf-96B8-444553540000"
	},

	hasData: function( elem ) {
		elem = elem.nodeType ? jQuery.cache[ elem[jQuery.expando] ] : elem[ jQuery.expando ];
		return !!elem && !isEmptyDataObject( elem );
	},

	data: function( elem, name, data ) {
		return internalData( elem, name, data );
	},

	removeData: function( elem, name ) {
		return internalRemoveData( elem, name );
	},

	// For internal use only.
	_data: function( elem, name, data ) {
		return internalData( elem, name, data, true );
	},

	_removeData: function( elem, name ) {
		return internalRemoveData( elem, name, true );
	}
});

jQuery.fn.extend({
	data: function( key, value ) {
		var i, name, data,
			elem = this[0],
			attrs = elem && elem.attributes;

		// Special expections of .data basically thwart jQuery.access,
		// so implement the relevant behavior ourselves

		// Gets all values
		if ( key === undefined ) {
			if ( this.length ) {
				data = jQuery.data( elem );

				if ( elem.nodeType === 1 && !jQuery._data( elem, "parsedAttrs" ) ) {
					i = attrs.length;
					while ( i-- ) {

						// Support: IE11+
						// The attrs elements can be null (#14894)
						if ( attrs[ i ] ) {
							name = attrs[ i ].name;
							if ( name.indexOf( "data-" ) === 0 ) {
								name = jQuery.camelCase( name.slice(5) );
								dataAttr( elem, name, data[ name ] );
							}
						}
					}
					jQuery._data( elem, "parsedAttrs", true );
				}
			}

			return data;
		}

		// Sets multiple values
		if ( typeof key === "object" ) {
			return this.each(function() {
				jQuery.data( this, key );
			});
		}

		return arguments.length > 1 ?

			// Sets one value
			this.each(function() {
				jQuery.data( this, key, value );
			}) :

			// Gets one value
			// Try to fetch any internally stored data first
			elem ? dataAttr( elem, key, jQuery.data( elem, key ) ) : undefined;
	},

	removeData: function( key ) {
		return this.each(function() {
			jQuery.removeData( this, key );
		});
	}
});


jQuery.extend({
	queue: function( elem, type, data ) {
		var queue;

		if ( elem ) {
			type = ( type || "fx" ) + "queue";
			queue = jQuery._data( elem, type );

			// Speed up dequeue by getting out quickly if this is just a lookup
			if ( data ) {
				if ( !queue || jQuery.isArray(data) ) {
					queue = jQuery._data( elem, type, jQuery.makeArray(data) );
				} else {
					queue.push( data );
				}
			}
			return queue || [];
		}
	},

	dequeue: function( elem, type ) {
		type = type || "fx";

		var queue = jQuery.queue( elem, type ),
			startLength = queue.length,
			fn = queue.shift(),
			hooks = jQuery._queueHooks( elem, type ),
			next = function() {
				jQuery.dequeue( elem, type );
			};

		// If the fx queue is dequeued, always remove the progress sentinel
		if ( fn === "inprogress" ) {
			fn = queue.shift();
			startLength--;
		}

		if ( fn ) {

			// Add a progress sentinel to prevent the fx queue from being
			// automatically dequeued
			if ( type === "fx" ) {
				queue.unshift( "inprogress" );
			}

			// clear up the last queue stop function
			delete hooks.stop;
			fn.call( elem, next, hooks );
		}

		if ( !startLength && hooks ) {
			hooks.empty.fire();
		}
	},

	// not intended for public consumption - generates a queueHooks object, or returns the current one
	_queueHooks: function( elem, type ) {
		var key = type + "queueHooks";
		return jQuery._data( elem, key ) || jQuery._data( elem, key, {
			empty: jQuery.Callbacks("once memory").add(function() {
				jQuery._removeData( elem, type + "queue" );
				jQuery._removeData( elem, key );
			})
		});
	}
});

jQuery.fn.extend({
	queue: function( type, data ) {
		var setter = 2;

		if ( typeof type !== "string" ) {
			data = type;
			type = "fx";
			setter--;
		}

		if ( arguments.length < setter ) {
			return jQuery.queue( this[0], type );
		}

		return data === undefined ?
			this :
			this.each(function() {
				var queue = jQuery.queue( this, type, data );

				// ensure a hooks for this queue
				jQuery._queueHooks( this, type );

				if ( type === "fx" && queue[0] !== "inprogress" ) {
					jQuery.dequeue( this, type );
				}
			});
	},
	dequeue: function( type ) {
		return this.each(function() {
			jQuery.dequeue( this, type );
		});
	},
	clearQueue: function( type ) {
		return this.queue( type || "fx", [] );
	},
	// Get a promise resolved when queues of a certain type
	// are emptied (fx is the type by default)
	promise: function( type, obj ) {
		var tmp,
			count = 1,
			defer = jQuery.Deferred(),
			elements = this,
			i = this.length,
			resolve = function() {
				if ( !( --count ) ) {
					defer.resolveWith( elements, [ elements ] );
				}
			};

		if ( typeof type !== "string" ) {
			obj = type;
			type = undefined;
		}
		type = type || "fx";

		while ( i-- ) {
			tmp = jQuery._data( elements[ i ], type + "queueHooks" );
			if ( tmp && tmp.empty ) {
				count++;
				tmp.empty.add( resolve );
			}
		}
		resolve();
		return defer.promise( obj );
	}
});
var pnum = (/[+-]?(?:\d*\.|)\d+(?:[eE][+-]?\d+|)/).source;

var cssExpand = [ "Top", "Right", "Bottom", "Left" ];

var isHidden = function( elem, el ) {
		// isHidden might be called from jQuery#filter function;
		// in that case, element will be second argument
		elem = el || elem;
		return jQuery.css( elem, "display" ) === "none" || !jQuery.contains( elem.ownerDocument, elem );
	};



// Multifunctional method to get and set values of a collection
// The value/s can optionally be executed if it's a function
var access = jQuery.access = function( elems, fn, key, value, chainable, emptyGet, raw ) {
	var i = 0,
		length = elems.length,
		bulk = key == null;

	// Sets many values
	if ( jQuery.type( key ) === "object" ) {
		chainable = true;
		for ( i in key ) {
			jQuery.access( elems, fn, i, key[i], true, emptyGet, raw );
		}

	// Sets one value
	} else if ( value !== undefined ) {
		chainable = true;

		if ( !jQuery.isFunction( value ) ) {
			raw = true;
		}

		if ( bulk ) {
			// Bulk operations run against the entire set
			if ( raw ) {
				fn.call( elems, value );
				fn = null;

			// ...except when executing function values
			} else {
				bulk = fn;
				fn = function( elem, key, value ) {
					return bulk.call( jQuery( elem ), value );
				};
			}
		}

		if ( fn ) {
			for ( ; i < length; i++ ) {
				fn( elems[i], key, raw ? value : value.call( elems[i], i, fn( elems[i], key ) ) );
			}
		}
	}

	return chainable ?
		elems :

		// Gets
		bulk ?
			fn.call( elems ) :
			length ? fn( elems[0], key ) : emptyGet;
};
var rcheckableType = (/^(?:checkbox|radio)$/i);



(function() {
	// Minified: var a,b,c
	var input = document.createElement( "input" ),
		div = document.createElement( "div" ),
		fragment = document.createDocumentFragment();

	// Setup
	div.innerHTML = "  <link/><table></table><a href='/a'>a</a><input type='checkbox'/>";

	// IE strips leading whitespace when .innerHTML is used
	support.leadingWhitespace = div.firstChild.nodeType === 3;

	// Make sure that tbody elements aren't automatically inserted
	// IE will insert them into empty tables
	support.tbody = !div.getElementsByTagName( "tbody" ).length;

	// Make sure that link elements get serialized correctly by innerHTML
	// This requires a wrapper element in IE
	support.htmlSerialize = !!div.getElementsByTagName( "link" ).length;

	// Makes sure cloning an html5 element does not cause problems
	// Where outerHTML is undefined, this still works
	support.html5Clone =
		document.createElement( "nav" ).cloneNode( true ).outerHTML !== "<:nav></:nav>";

	// Check if a disconnected checkbox will retain its checked
	// value of true after appended to the DOM (IE6/7)
	input.type = "checkbox";
	input.checked = true;
	fragment.appendChild( input );
	support.appendChecked = input.checked;

	// Make sure textarea (and checkbox) defaultValue is properly cloned
	// Support: IE6-IE11+
	div.innerHTML = "<textarea>x</textarea>";
	support.noCloneChecked = !!div.cloneNode( true ).lastChild.defaultValue;

	// #11217 - WebKit loses check when the name is after the checked attribute
	fragment.appendChild( div );
	div.innerHTML = "<input type='radio' checked='checked' name='t'/>";

	// Support: Safari 5.1, iOS 5.1, Android 4.x, Android 2.3
	// old WebKit doesn't clone checked state correctly in fragments
	support.checkClone = div.cloneNode( true ).cloneNode( true ).lastChild.checked;

	// Support: IE<9
	// Opera does not clone events (and typeof div.attachEvent === undefined).
	// IE9-10 clones events bound via attachEvent, but they don't trigger with .click()
	support.noCloneEvent = true;
	if ( div.attachEvent ) {
		div.attachEvent( "onclick", function() {
			support.noCloneEvent = false;
		});

		div.cloneNode( true ).click();
	}

	// Execute the test only if not already executed in another module.
	if (support.deleteExpando == null) {
		// Support: IE<9
		support.deleteExpando = true;
		try {
			delete div.test;
		} catch( e ) {
			support.deleteExpando = false;
		}
	}
})();


(function() {
	var i, eventName,
		div = document.createElement( "div" );

	// Support: IE<9 (lack submit/change bubble), Firefox 23+ (lack focusin event)
	for ( i in { submit: true, change: true, focusin: true }) {
		eventName = "on" + i;

		if ( !(support[ i + "Bubbles" ] = eventName in window) ) {
			// Beware of CSP restrictions (https://developer.mozilla.org/en/Security/CSP)
			div.setAttribute( eventName, "t" );
			support[ i + "Bubbles" ] = div.attributes[ eventName ].expando === false;
		}
	}

	// Null elements to avoid leaks in IE.
	div = null;
})();


var rformElems = /^(?:input|select|textarea)$/i,
	rkeyEvent = /^key/,
	rmouseEvent = /^(?:mouse|pointer|contextmenu)|click/,
	rfocusMorph = /^(?:focusinfocus|focusoutblur)$/,
	rtypenamespace = /^([^.]*)(?:\.(.+)|)$/;

function returnTrue() {
	return true;
}

function returnFalse() {
	return false;
}

function safeActiveElement() {
	try {
		return document.activeElement;
	} catch ( err ) { }
}

/*
 * Helper functions for managing events -- not part of the public interface.
 * Props to Dean Edwards' addEvent library for many of the ideas.
 */
jQuery.event = {

	global: {},

	add: function( elem, types, handler, data, selector ) {
		var tmp, events, t, handleObjIn,
			special, eventHandle, handleObj,
			handlers, type, namespaces, origType,
			elemData = jQuery._data( elem );

		// Don't attach events to noData or text/comment nodes (but allow plain objects)
		if ( !elemData ) {
			return;
		}

		// Caller can pass in an object of custom data in lieu of the handler
		if ( handler.handler ) {
			handleObjIn = handler;
			handler = handleObjIn.handler;
			selector = handleObjIn.selector;
		}

		// Make sure that the handler has a unique ID, used to find/remove it later
		if ( !handler.guid ) {
			handler.guid = jQuery.guid++;
		}

		// Init the element's event structure and main handler, if this is the first
		if ( !(events = elemData.events) ) {
			events = elemData.events = {};
		}
		if ( !(eventHandle = elemData.handle) ) {
			eventHandle = elemData.handle = function( e ) {
				// Discard the second event of a jQuery.event.trigger() and
				// when an event is called after a page has unloaded
				return typeof jQuery !== strundefined && (!e || jQuery.event.triggered !== e.type) ?
					jQuery.event.dispatch.apply( eventHandle.elem, arguments ) :
					undefined;
			};
			// Add elem as a property of the handle fn to prevent a memory leak with IE non-native events
			eventHandle.elem = elem;
		}

		// Handle multiple events separated by a space
		types = ( types || "" ).match( rnotwhite ) || [ "" ];
		t = types.length;
		while ( t-- ) {
			tmp = rtypenamespace.exec( types[t] ) || [];
			type = origType = tmp[1];
			namespaces = ( tmp[2] || "" ).split( "." ).sort();

			// There *must* be a type, no attaching namespace-only handlers
			if ( !type ) {
				continue;
			}

			// If event changes its type, use the special event handlers for the changed type
			special = jQuery.event.special[ type ] || {};

			// If selector defined, determine special event api type, otherwise given type
			type = ( selector ? special.delegateType : special.bindType ) || type;

			// Update special based on newly reset type
			special = jQuery.event.special[ type ] || {};

			// handleObj is passed to all event handlers
			handleObj = jQuery.extend({
				type: type,
				origType: origType,
				data: data,
				handler: handler,
				guid: handler.guid,
				selector: selector,
				needsContext: selector && jQuery.expr.match.needsContext.test( selector ),
				namespace: namespaces.join(".")
			}, handleObjIn );

			// Init the event handler queue if we're the first
			if ( !(handlers = events[ type ]) ) {
				handlers = events[ type ] = [];
				handlers.delegateCount = 0;

				// Only use addEventListener/attachEvent if the special events handler returns false
				if ( !special.setup || special.setup.call( elem, data, namespaces, eventHandle ) === false ) {
					// Bind the global event handler to the element
					if ( elem.addEventListener ) {
						elem.addEventListener( type, eventHandle, false );

					} else if ( elem.attachEvent ) {
						elem.attachEvent( "on" + type, eventHandle );
					}
				}
			}

			if ( special.add ) {
				special.add.call( elem, handleObj );

				if ( !handleObj.handler.guid ) {
					handleObj.handler.guid = handler.guid;
				}
			}

			// Add to the element's handler list, delegates in front
			if ( selector ) {
				handlers.splice( handlers.delegateCount++, 0, handleObj );
			} else {
				handlers.push( handleObj );
			}

			// Keep track of which events have ever been used, for event optimization
			jQuery.event.global[ type ] = true;
		}

		// Nullify elem to prevent memory leaks in IE
		elem = null;
	},

	// Detach an event or set of events from an element
	remove: function( elem, types, handler, selector, mappedTypes ) {
		var j, handleObj, tmp,
			origCount, t, events,
			special, handlers, type,
			namespaces, origType,
			elemData = jQuery.hasData( elem ) && jQuery._data( elem );

		if ( !elemData || !(events = elemData.events) ) {
			return;
		}

		// Once for each type.namespace in types; type may be omitted
		types = ( types || "" ).match( rnotwhite ) || [ "" ];
		t = types.length;
		while ( t-- ) {
			tmp = rtypenamespace.exec( types[t] ) || [];
			type = origType = tmp[1];
			namespaces = ( tmp[2] || "" ).split( "." ).sort();

			// Unbind all events (on this namespace, if provided) for the element
			if ( !type ) {
				for ( type in events ) {
					jQuery.event.remove( elem, type + types[ t ], handler, selector, true );
				}
				continue;
			}

			special = jQuery.event.special[ type ] || {};
			type = ( selector ? special.delegateType : special.bindType ) || type;
			handlers = events[ type ] || [];
			tmp = tmp[2] && new RegExp( "(^|\\.)" + namespaces.join("\\.(?:.*\\.|)") + "(\\.|$)" );

			// Remove matching events
			origCount = j = handlers.length;
			while ( j-- ) {
				handleObj = handlers[ j ];

				if ( ( mappedTypes || origType === handleObj.origType ) &&
					( !handler || handler.guid === handleObj.guid ) &&
					( !tmp || tmp.test( handleObj.namespace ) ) &&
					( !selector || selector === handleObj.selector || selector === "**" && handleObj.selector ) ) {
					handlers.splice( j, 1 );

					if ( handleObj.selector ) {
						handlers.delegateCount--;
					}
					if ( special.remove ) {
						special.remove.call( elem, handleObj );
					}
				}
			}

			// Remove generic event handler if we removed something and no more handlers exist
			// (avoids potential for endless recursion during removal of special event handlers)
			if ( origCount && !handlers.length ) {
				if ( !special.teardown || special.teardown.call( elem, namespaces, elemData.handle ) === false ) {
					jQuery.removeEvent( elem, type, elemData.handle );
				}

				delete events[ type ];
			}
		}

		// Remove the expando if it's no longer used
		if ( jQuery.isEmptyObject( events ) ) {
			delete elemData.handle;

			// removeData also checks for emptiness and clears the expando if empty
			// so use it instead of delete
			jQuery._removeData( elem, "events" );
		}
	},

	trigger: function( event, data, elem, onlyHandlers ) {
		var handle, ontype, cur,
			bubbleType, special, tmp, i,
			eventPath = [ elem || document ],
			type = hasOwn.call( event, "type" ) ? event.type : event,
			namespaces = hasOwn.call( event, "namespace" ) ? event.namespace.split(".") : [];

		cur = tmp = elem = elem || document;

		// Don't do events on text and comment nodes
		if ( elem.nodeType === 3 || elem.nodeType === 8 ) {
			return;
		}

		// focus/blur morphs to focusin/out; ensure we're not firing them right now
		if ( rfocusMorph.test( type + jQuery.event.triggered ) ) {
			return;
		}

		if ( type.indexOf(".") >= 0 ) {
			// Namespaced trigger; create a regexp to match event type in handle()
			namespaces = type.split(".");
			type = namespaces.shift();
			namespaces.sort();
		}
		ontype = type.indexOf(":") < 0 && "on" + type;

		// Caller can pass in a jQuery.Event object, Object, or just an event type string
		event = event[ jQuery.expando ] ?
			event :
			new jQuery.Event( type, typeof event === "object" && event );

		// Trigger bitmask: & 1 for native handlers; & 2 for jQuery (always true)
		event.isTrigger = onlyHandlers ? 2 : 3;
		event.namespace = namespaces.join(".");
		event.namespace_re = event.namespace ?
			new RegExp( "(^|\\.)" + namespaces.join("\\.(?:.*\\.|)") + "(\\.|$)" ) :
			null;

		// Clean up the event in case it is being reused
		event.result = undefined;
		if ( !event.target ) {
			event.target = elem;
		}

		// Clone any incoming data and prepend the event, creating the handler arg list
		data = data == null ?
			[ event ] :
			jQuery.makeArray( data, [ event ] );

		// Allow special events to draw outside the lines
		special = jQuery.event.special[ type ] || {};
		if ( !onlyHandlers && special.trigger && special.trigger.apply( elem, data ) === false ) {
			return;
		}

		// Determine event propagation path in advance, per W3C events spec (#9951)
		// Bubble up to document, then to window; watch for a global ownerDocument var (#9724)
		if ( !onlyHandlers && !special.noBubble && !jQuery.isWindow( elem ) ) {

			bubbleType = special.delegateType || type;
			if ( !rfocusMorph.test( bubbleType + type ) ) {
				cur = cur.parentNode;
			}
			for ( ; cur; cur = cur.parentNode ) {
				eventPath.push( cur );
				tmp = cur;
			}

			// Only add window if we got to document (e.g., not plain obj or detached DOM)
			if ( tmp === (elem.ownerDocument || document) ) {
				eventPath.push( tmp.defaultView || tmp.parentWindow || window );
			}
		}

		// Fire handlers on the event path
		i = 0;
		while ( (cur = eventPath[i++]) && !event.isPropagationStopped() ) {

			event.type = i > 1 ?
				bubbleType :
				special.bindType || type;

			// jQuery handler
			handle = ( jQuery._data( cur, "events" ) || {} )[ event.type ] && jQuery._data( cur, "handle" );
			if ( handle ) {
				handle.apply( cur, data );
			}

			// Native handler
			handle = ontype && cur[ ontype ];
			if ( handle && handle.apply && jQuery.acceptData( cur ) ) {
				event.result = handle.apply( cur, data );
				if ( event.result === false ) {
					event.preventDefault();
				}
			}
		}
		event.type = type;

		// If nobody prevented the default action, do it now
		if ( !onlyHandlers && !event.isDefaultPrevented() ) {

			if ( (!special._default || special._default.apply( eventPath.pop(), data ) === false) &&
				jQuery.acceptData( elem ) ) {

				// Call a native DOM method on the target with the same name name as the event.
				// Can't use an .isFunction() check here because IE6/7 fails that test.
				// Don't do default actions on window, that's where global variables be (#6170)
				if ( ontype && elem[ type ] && !jQuery.isWindow( elem ) ) {

					// Don't re-trigger an onFOO event when we call its FOO() method
					tmp = elem[ ontype ];

					if ( tmp ) {
						elem[ ontype ] = null;
					}

					// Prevent re-triggering of the same event, since we already bubbled it above
					jQuery.event.triggered = type;
					try {
						elem[ type ]();
					} catch ( e ) {
						// IE<9 dies on focus/blur to hidden element (#1486,#12518)
						// only reproducible on winXP IE8 native, not IE9 in IE8 mode
					}
					jQuery.event.triggered = undefined;

					if ( tmp ) {
						elem[ ontype ] = tmp;
					}
				}
			}
		}

		return event.result;
	},

	dispatch: function( event ) {

		// Make a writable jQuery.Event from the native event object
		event = jQuery.event.fix( event );

		var i, ret, handleObj, matched, j,
			handlerQueue = [],
			args = slice.call( arguments ),
			handlers = ( jQuery._data( this, "events" ) || {} )[ event.type ] || [],
			special = jQuery.event.special[ event.type ] || {};

		// Use the fix-ed jQuery.Event rather than the (read-only) native event
		args[0] = event;
		event.delegateTarget = this;

		// Call the preDispatch hook for the mapped type, and let it bail if desired
		if ( special.preDispatch && special.preDispatch.call( this, event ) === false ) {
			return;
		}

		// Determine handlers
		handlerQueue = jQuery.event.handlers.call( this, event, handlers );

		// Run delegates first; they may want to stop propagation beneath us
		i = 0;
		while ( (matched = handlerQueue[ i++ ]) && !event.isPropagationStopped() ) {
			event.currentTarget = matched.elem;

			j = 0;
			while ( (handleObj = matched.handlers[ j++ ]) && !event.isImmediatePropagationStopped() ) {

				// Triggered event must either 1) have no namespace, or
				// 2) have namespace(s) a subset or equal to those in the bound event (both can have no namespace).
				if ( !event.namespace_re || event.namespace_re.test( handleObj.namespace ) ) {

					event.handleObj = handleObj;
					event.data = handleObj.data;

					ret = ( (jQuery.event.special[ handleObj.origType ] || {}).handle || handleObj.handler )
							.apply( matched.elem, args );

					if ( ret !== undefined ) {
						if ( (event.result = ret) === false ) {
							event.preventDefault();
							event.stopPropagation();
						}
					}
				}
			}
		}

		// Call the postDispatch hook for the mapped type
		if ( special.postDispatch ) {
			special.postDispatch.call( this, event );
		}

		return event.result;
	},

	handlers: function( event, handlers ) {
		var sel, handleObj, matches, i,
			handlerQueue = [],
			delegateCount = handlers.delegateCount,
			cur = event.target;

		// Find delegate handlers
		// Black-hole SVG <use> instance trees (#13180)
		// Avoid non-left-click bubbling in Firefox (#3861)
		if ( delegateCount && cur.nodeType && (!event.button || event.type !== "click") ) {

			/* jshint eqeqeq: false */
			for ( ; cur != this; cur = cur.parentNode || this ) {
				/* jshint eqeqeq: true */

				// Don't check non-elements (#13208)
				// Don't process clicks on disabled elements (#6911, #8165, #11382, #11764)
				if ( cur.nodeType === 1 && (cur.disabled !== true || event.type !== "click") ) {
					matches = [];
					for ( i = 0; i < delegateCount; i++ ) {
						handleObj = handlers[ i ];

						// Don't conflict with Object.prototype properties (#13203)
						sel = handleObj.selector + " ";

						if ( matches[ sel ] === undefined ) {
							matches[ sel ] = handleObj.needsContext ?
								jQuery( sel, this ).index( cur ) >= 0 :
								jQuery.find( sel, this, null, [ cur ] ).length;
						}
						if ( matches[ sel ] ) {
							matches.push( handleObj );
						}
					}
					if ( matches.length ) {
						handlerQueue.push({ elem: cur, handlers: matches });
					}
				}
			}
		}

		// Add the remaining (directly-bound) handlers
		if ( delegateCount < handlers.length ) {
			handlerQueue.push({ elem: this, handlers: handlers.slice( delegateCount ) });
		}

		return handlerQueue;
	},

	fix: function( event ) {
		if ( event[ jQuery.expando ] ) {
			return event;
		}

		// Create a writable copy of the event object and normalize some properties
		var i, prop, copy,
			type = event.type,
			originalEvent = event,
			fixHook = this.fixHooks[ type ];

		if ( !fixHook ) {
			this.fixHooks[ type ] = fixHook =
				rmouseEvent.test( type ) ? this.mouseHooks :
				rkeyEvent.test( type ) ? this.keyHooks :
				{};
		}
		copy = fixHook.props ? this.props.concat( fixHook.props ) : this.props;

		event = new jQuery.Event( originalEvent );

		i = copy.length;
		while ( i-- ) {
			prop = copy[ i ];
			event[ prop ] = originalEvent[ prop ];
		}

		// Support: IE<9
		// Fix target property (#1925)
		if ( !event.target ) {
			event.target = originalEvent.srcElement || document;
		}

		// Support: Chrome 23+, Safari?
		// Target should not be a text node (#504, #13143)
		if ( event.target.nodeType === 3 ) {
			event.target = event.target.parentNode;
		}

		// Support: IE<9
		// For mouse/key events, metaKey==false if it's undefined (#3368, #11328)
		event.metaKey = !!event.metaKey;

		return fixHook.filter ? fixHook.filter( event, originalEvent ) : event;
	},

	// Includes some event props shared by KeyEvent and MouseEvent
	props: "altKey bubbles cancelable ctrlKey currentTarget eventPhase metaKey relatedTarget shiftKey target timeStamp view which".split(" "),

	fixHooks: {},

	keyHooks: {
		props: "char charCode key keyCode".split(" "),
		filter: function( event, original ) {

			// Add which for key events
			if ( event.which == null ) {
				event.which = original.charCode != null ? original.charCode : original.keyCode;
			}

			return event;
		}
	},

	mouseHooks: {
		props: "button buttons clientX clientY fromElement offsetX offsetY pageX pageY screenX screenY toElement".split(" "),
		filter: function( event, original ) {
			var body, eventDoc, doc,
				button = original.button,
				fromElement = original.fromElement;

			// Calculate pageX/Y if missing and clientX/Y available
			if ( event.pageX == null && original.clientX != null ) {
				eventDoc = event.target.ownerDocument || document;
				doc = eventDoc.documentElement;
				body = eventDoc.body;

				event.pageX = original.clientX + ( doc && doc.scrollLeft || body && body.scrollLeft || 0 ) - ( doc && doc.clientLeft || body && body.clientLeft || 0 );
				event.pageY = original.clientY + ( doc && doc.scrollTop  || body && body.scrollTop  || 0 ) - ( doc && doc.clientTop  || body && body.clientTop  || 0 );
			}

			// Add relatedTarget, if necessary
			if ( !event.relatedTarget && fromElement ) {
				event.relatedTarget = fromElement === event.target ? original.toElement : fromElement;
			}

			// Add which for click: 1 === left; 2 === middle; 3 === right
			// Note: button is not normalized, so don't use it
			if ( !event.which && button !== undefined ) {
				event.which = ( button & 1 ? 1 : ( button & 2 ? 3 : ( button & 4 ? 2 : 0 ) ) );
			}

			return event;
		}
	},

	special: {
		load: {
			// Prevent triggered image.load events from bubbling to window.load
			noBubble: true
		},
		focus: {
			// Fire native event if possible so blur/focus sequence is correct
			trigger: function() {
				if ( this !== safeActiveElement() && this.focus ) {
					try {
						this.focus();
						return false;
					} catch ( e ) {
						// Support: IE<9
						// If we error on focus to hidden element (#1486, #12518),
						// let .trigger() run the handlers
					}
				}
			},
			delegateType: "focusin"
		},
		blur: {
			trigger: function() {
				if ( this === safeActiveElement() && this.blur ) {
					this.blur();
					return false;
				}
			},
			delegateType: "focusout"
		},
		click: {
			// For checkbox, fire native event so checked state will be right
			trigger: function() {
				if ( jQuery.nodeName( this, "input" ) && this.type === "checkbox" && this.click ) {
					this.click();
					return false;
				}
			},

			// For cross-browser consistency, don't fire native .click() on links
			_default: function( event ) {
				return jQuery.nodeName( event.target, "a" );
			}
		},

		beforeunload: {
			postDispatch: function( event ) {

				// Support: Firefox 20+
				// Firefox doesn't alert if the returnValue field is not set.
				if ( event.result !== undefined && event.originalEvent ) {
					event.originalEvent.returnValue = event.result;
				}
			}
		}
	},

	simulate: function( type, elem, event, bubble ) {
		// Piggyback on a donor event to simulate a different one.
		// Fake originalEvent to avoid donor's stopPropagation, but if the
		// simulated event prevents default then we do the same on the donor.
		var e = jQuery.extend(
			new jQuery.Event(),
			event,
			{
				type: type,
				isSimulated: true,
				originalEvent: {}
			}
		);
		if ( bubble ) {
			jQuery.event.trigger( e, null, elem );
		} else {
			jQuery.event.dispatch.call( elem, e );
		}
		if ( e.isDefaultPrevented() ) {
			event.preventDefault();
		}
	}
};

jQuery.removeEvent = document.removeEventListener ?
	function( elem, type, handle ) {
		if ( elem.removeEventListener ) {
			elem.removeEventListener( type, handle, false );
		}
	} :
	function( elem, type, handle ) {
		var name = "on" + type;

		if ( elem.detachEvent ) {

			// #8545, #7054, preventing memory leaks for custom events in IE6-8
			// detachEvent needed property on element, by name of that event, to properly expose it to GC
			if ( typeof elem[ name ] === strundefined ) {
				elem[ name ] = null;
			}

			elem.detachEvent( name, handle );
		}
	};

jQuery.Event = function( src, props ) {
	// Allow instantiation without the 'new' keyword
	if ( !(this instanceof jQuery.Event) ) {
		return new jQuery.Event( src, props );
	}

	// Event object
	if ( src && src.type ) {
		this.originalEvent = src;
		this.type = src.type;

		// Events bubbling up the document may have been marked as prevented
		// by a handler lower down the tree; reflect the correct value.
		this.isDefaultPrevented = src.defaultPrevented ||
				src.defaultPrevented === undefined &&
				// Support: IE < 9, Android < 4.0
				src.returnValue === false ?
			returnTrue :
			returnFalse;

	// Event type
	} else {
		this.type = src;
	}

	// Put explicitly provided properties onto the event object
	if ( props ) {
		jQuery.extend( this, props );
	}

	// Create a timestamp if incoming event doesn't have one
	this.timeStamp = src && src.timeStamp || jQuery.now();

	// Mark it as fixed
	this[ jQuery.expando ] = true;
};

// jQuery.Event is based on DOM3 Events as specified by the ECMAScript Language Binding
// http://www.w3.org/TR/2003/WD-DOM-Level-3-Events-20030331/ecma-script-binding.html
jQuery.Event.prototype = {
	isDefaultPrevented: returnFalse,
	isPropagationStopped: returnFalse,
	isImmediatePropagationStopped: returnFalse,

	preventDefault: function() {
		var e = this.originalEvent;

		this.isDefaultPrevented = returnTrue;
		if ( !e ) {
			return;
		}

		// If preventDefault exists, run it on the original event
		if ( e.preventDefault ) {
			e.preventDefault();

		// Support: IE
		// Otherwise set the returnValue property of the original event to false
		} else {
			e.returnValue = false;
		}
	},
	stopPropagation: function() {
		var e = this.originalEvent;

		this.isPropagationStopped = returnTrue;
		if ( !e ) {
			return;
		}
		// If stopPropagation exists, run it on the original event
		if ( e.stopPropagation ) {
			e.stopPropagation();
		}

		// Support: IE
		// Set the cancelBubble property of the original event to true
		e.cancelBubble = true;
	},
	stopImmediatePropagation: function() {
		var e = this.originalEvent;

		this.isImmediatePropagationStopped = returnTrue;

		if ( e && e.stopImmediatePropagation ) {
			e.stopImmediatePropagation();
		}

		this.stopPropagation();
	}
};

// Create mouseenter/leave events using mouseover/out and event-time checks
jQuery.each({
	mouseenter: "mouseover",
	mouseleave: "mouseout",
	pointerenter: "pointerover",
	pointerleave: "pointerout"
}, function( orig, fix ) {
	jQuery.event.special[ orig ] = {
		delegateType: fix,
		bindType: fix,

		handle: function( event ) {
			var ret,
				target = this,
				related = event.relatedTarget,
				handleObj = event.handleObj;

			// For mousenter/leave call the handler if related is outside the target.
			// NB: No relatedTarget if the mouse left/entered the browser window
			if ( !related || (related !== target && !jQuery.contains( target, related )) ) {
				event.type = handleObj.origType;
				ret = handleObj.handler.apply( this, arguments );
				event.type = fix;
			}
			return ret;
		}
	};
});

// IE submit delegation
if ( !support.submitBubbles ) {

	jQuery.event.special.submit = {
		setup: function() {
			// Only need this for delegated form submit events
			if ( jQuery.nodeName( this, "form" ) ) {
				return false;
			}

			// Lazy-add a submit handler when a descendant form may potentially be submitted
			jQuery.event.add( this, "click._submit keypress._submit", function( e ) {
				// Node name check avoids a VML-related crash in IE (#9807)
				var elem = e.target,
					form = jQuery.nodeName( elem, "input" ) || jQuery.nodeName( elem, "button" ) ? elem.form : undefined;
				if ( form && !jQuery._data( form, "submitBubbles" ) ) {
					jQuery.event.add( form, "submit._submit", function( event ) {
						event._submit_bubble = true;
					});
					jQuery._data( form, "submitBubbles", true );
				}
			});
			// return undefined since we don't need an event listener
		},

		postDispatch: function( event ) {
			// If form was submitted by the user, bubble the event up the tree
			if ( event._submit_bubble ) {
				delete event._submit_bubble;
				if ( this.parentNode && !event.isTrigger ) {
					jQuery.event.simulate( "submit", this.parentNode, event, true );
				}
			}
		},

		teardown: function() {
			// Only need this for delegated form submit events
			if ( jQuery.nodeName( this, "form" ) ) {
				return false;
			}

			// Remove delegated handlers; cleanData eventually reaps submit handlers attached above
			jQuery.event.remove( this, "._submit" );
		}
	};
}

// IE change delegation and checkbox/radio fix
if ( !support.changeBubbles ) {

	jQuery.event.special.change = {

		setup: function() {

			if ( rformElems.test( this.nodeName ) ) {
				// IE doesn't fire change on a check/radio until blur; trigger it on click
				// after a propertychange. Eat the blur-change in special.change.handle.
				// This still fires onchange a second time for check/radio after blur.
				if ( this.type === "checkbox" || this.type === "radio" ) {
					jQuery.event.add( this, "propertychange._change", function( event ) {
						if ( event.originalEvent.propertyName === "checked" ) {
							this._just_changed = true;
						}
					});
					jQuery.event.add( this, "click._change", function( event ) {
						if ( this._just_changed && !event.isTrigger ) {
							this._just_changed = false;
						}
						// Allow triggered, simulated change events (#11500)
						jQuery.event.simulate( "change", this, event, true );
					});
				}
				return false;
			}
			// Delegated event; lazy-add a change handler on descendant inputs
			jQuery.event.add( this, "beforeactivate._change", function( e ) {
				var elem = e.target;

				if ( rformElems.test( elem.nodeName ) && !jQuery._data( elem, "changeBubbles" ) ) {
					jQuery.event.add( elem, "change._change", function( event ) {
						if ( this.parentNode && !event.isSimulated && !event.isTrigger ) {
							jQuery.event.simulate( "change", this.parentNode, event, true );
						}
					});
					jQuery._data( elem, "changeBubbles", true );
				}
			});
		},

		handle: function( event ) {
			var elem = event.target;

			// Swallow native change events from checkbox/radio, we already triggered them above
			if ( this !== elem || event.isSimulated || event.isTrigger || (elem.type !== "radio" && elem.type !== "checkbox") ) {
				return event.handleObj.handler.apply( this, arguments );
			}
		},

		teardown: function() {
			jQuery.event.remove( this, "._change" );

			return !rformElems.test( this.nodeName );
		}
	};
}

// Create "bubbling" focus and blur events
if ( !support.focusinBubbles ) {
	jQuery.each({ focus: "focusin", blur: "focusout" }, function( orig, fix ) {

		// Attach a single capturing handler on the document while someone wants focusin/focusout
		var handler = function( event ) {
				jQuery.event.simulate( fix, event.target, jQuery.event.fix( event ), true );
			};

		jQuery.event.special[ fix ] = {
			setup: function() {
				var doc = this.ownerDocument || this,
					attaches = jQuery._data( doc, fix );

				if ( !attaches ) {
					doc.addEventListener( orig, handler, true );
				}
				jQuery._data( doc, fix, ( attaches || 0 ) + 1 );
			},
			teardown: function() {
				var doc = this.ownerDocument || this,
					attaches = jQuery._data( doc, fix ) - 1;

				if ( !attaches ) {
					doc.removeEventListener( orig, handler, true );
					jQuery._removeData( doc, fix );
				} else {
					jQuery._data( doc, fix, attaches );
				}
			}
		};
	});
}

jQuery.fn.extend({

	on: function( types, selector, data, fn, /*INTERNAL*/ one ) {
		var type, origFn;

		// Types can be a map of types/handlers
		if ( typeof types === "object" ) {
			// ( types-Object, selector, data )
			if ( typeof selector !== "string" ) {
				// ( types-Object, data )
				data = data || selector;
				selector = undefined;
			}
			for ( type in types ) {
				this.on( type, selector, data, types[ type ], one );
			}
			return this;
		}

		if ( data == null && fn == null ) {
			// ( types, fn )
			fn = selector;
			data = selector = undefined;
		} else if ( fn == null ) {
			if ( typeof selector === "string" ) {
				// ( types, selector, fn )
				fn = data;
				data = undefined;
			} else {
				// ( types, data, fn )
				fn = data;
				data = selector;
				selector = undefined;
			}
		}
		if ( fn === false ) {
			fn = returnFalse;
		} else if ( !fn ) {
			return this;
		}

		if ( one === 1 ) {
			origFn = fn;
			fn = function( event ) {
				// Can use an empty set, since event contains the info
				jQuery().off( event );
				return origFn.apply( this, arguments );
			};
			// Use same guid so caller can remove using origFn
			fn.guid = origFn.guid || ( origFn.guid = jQuery.guid++ );
		}
		return this.each( function() {
			jQuery.event.add( this, types, fn, data, selector );
		});
	},
	one: function( types, selector, data, fn ) {
		return this.on( types, selector, data, fn, 1 );
	},
	off: function( types, selector, fn ) {
		var handleObj, type;
		if ( types && types.preventDefault && types.handleObj ) {
			// ( event )  dispatched jQuery.Event
			handleObj = types.handleObj;
			jQuery( types.delegateTarget ).off(
				handleObj.namespace ? handleObj.origType + "." + handleObj.namespace : handleObj.origType,
				handleObj.selector,
				handleObj.handler
			);
			return this;
		}
		if ( typeof types === "object" ) {
			// ( types-object [, selector] )
			for ( type in types ) {
				this.off( type, selector, types[ type ] );
			}
			return this;
		}
		if ( selector === false || typeof selector === "function" ) {
			// ( types [, fn] )
			fn = selector;
			selector = undefined;
		}
		if ( fn === false ) {
			fn = returnFalse;
		}
		return this.each(function() {
			jQuery.event.remove( this, types, fn, selector );
		});
	},

	trigger: function( type, data ) {
		return this.each(function() {
			jQuery.event.trigger( type, data, this );
		});
	},
	triggerHandler: function( type, data ) {
		var elem = this[0];
		if ( elem ) {
			return jQuery.event.trigger( type, data, elem, true );
		}
	}
});


function createSafeFragment( document ) {
	var list = nodeNames.split( "|" ),
		safeFrag = document.createDocumentFragment();

	if ( safeFrag.createElement ) {
		while ( list.length ) {
			safeFrag.createElement(
				list.pop()
			);
		}
	}
	return safeFrag;
}

var nodeNames = "abbr|article|aside|audio|bdi|canvas|data|datalist|details|figcaption|figure|footer|" +
		"header|hgroup|mark|meter|nav|output|progress|section|summary|time|video",
	rinlinejQuery = / jQuery\d+="(?:null|\d+)"/g,
	rnoshimcache = new RegExp("<(?:" + nodeNames + ")[\\s/>]", "i"),
	rleadingWhitespace = /^\s+/,
	rxhtmlTag = /<(?!area|br|col|embed|hr|img|input|link|meta|param)(([\w:]+)[^>]*)\/>/gi,
	rtagName = /<([\w:]+)/,
	rtbody = /<tbody/i,
	rhtml = /<|&#?\w+;/,
	rnoInnerhtml = /<(?:script|style|link)/i,
	// checked="checked" or checked
	rchecked = /checked\s*(?:[^=]|=\s*.checked.)/i,
	rscriptType = /^$|\/(?:java|ecma)script/i,
	rscriptTypeMasked = /^true\/(.*)/,
	rcleanScript = /^\s*<!(?:\[CDATA\[|--)|(?:\]\]|--)>\s*$/g,

	// We have to close these tags to support XHTML (#13200)
	wrapMap = {
		option: [ 1, "<select multiple='multiple'>", "</select>" ],
		legend: [ 1, "<fieldset>", "</fieldset>" ],
		area: [ 1, "<map>", "</map>" ],
		param: [ 1, "<object>", "</object>" ],
		thead: [ 1, "<table>", "</table>" ],
		tr: [ 2, "<table><tbody>", "</tbody></table>" ],
		col: [ 2, "<table><tbody></tbody><colgroup>", "</colgroup></table>" ],
		td: [ 3, "<table><tbody><tr>", "</tr></tbody></table>" ],

		// IE6-8 can't serialize link, script, style, or any html5 (NoScope) tags,
		// unless wrapped in a div with non-breaking characters in front of it.
		_default: support.htmlSerialize ? [ 0, "", "" ] : [ 1, "X<div>", "</div>"  ]
	},
	safeFragment = createSafeFragment( document ),
	fragmentDiv = safeFragment.appendChild( document.createElement("div") );

wrapMap.optgroup = wrapMap.option;
wrapMap.tbody = wrapMap.tfoot = wrapMap.colgroup = wrapMap.caption = wrapMap.thead;
wrapMap.th = wrapMap.td;

function getAll( context, tag ) {
	var elems, elem,
		i = 0,
		found = typeof context.getElementsByTagName !== strundefined ? context.getElementsByTagName( tag || "*" ) :
			typeof context.querySelectorAll !== strundefined ? context.querySelectorAll( tag || "*" ) :
			undefined;

	if ( !found ) {
		for ( found = [], elems = context.childNodes || context; (elem = elems[i]) != null; i++ ) {
			if ( !tag || jQuery.nodeName( elem, tag ) ) {
				found.push( elem );
			} else {
				jQuery.merge( found, getAll( elem, tag ) );
			}
		}
	}

	return tag === undefined || tag && jQuery.nodeName( context, tag ) ?
		jQuery.merge( [ context ], found ) :
		found;
}

// Used in buildFragment, fixes the defaultChecked property
function fixDefaultChecked( elem ) {
	if ( rcheckableType.test( elem.type ) ) {
		elem.defaultChecked = elem.checked;
	}
}

// Support: IE<8
// Manipulating tables requires a tbody
function manipulationTarget( elem, content ) {
	return jQuery.nodeName( elem, "table" ) &&
		jQuery.nodeName( content.nodeType !== 11 ? content : content.firstChild, "tr" ) ?

		elem.getElementsByTagName("tbody")[0] ||
			elem.appendChild( elem.ownerDocument.createElement("tbody") ) :
		elem;
}

// Replace/restore the type attribute of script elements for safe DOM manipulation
function disableScript( elem ) {
	elem.type = (jQuery.find.attr( elem, "type" ) !== null) + "/" + elem.type;
	return elem;
}
function restoreScript( elem ) {
	var match = rscriptTypeMasked.exec( elem.type );
	if ( match ) {
		elem.type = match[1];
	} else {
		elem.removeAttribute("type");
	}
	return elem;
}

// Mark scripts as having already been evaluated
function setGlobalEval( elems, refElements ) {
	var elem,
		i = 0;
	for ( ; (elem = elems[i]) != null; i++ ) {
		jQuery._data( elem, "globalEval", !refElements || jQuery._data( refElements[i], "globalEval" ) );
	}
}

function cloneCopyEvent( src, dest ) {

	if ( dest.nodeType !== 1 || !jQuery.hasData( src ) ) {
		return;
	}

	var type, i, l,
		oldData = jQuery._data( src ),
		curData = jQuery._data( dest, oldData ),
		events = oldData.events;

	if ( events ) {
		delete curData.handle;
		curData.events = {};

		for ( type in events ) {
			for ( i = 0, l = events[ type ].length; i < l; i++ ) {
				jQuery.event.add( dest, type, events[ type ][ i ] );
			}
		}
	}

	// make the cloned public data object a copy from the original
	if ( curData.data ) {
		curData.data = jQuery.extend( {}, curData.data );
	}
}

function fixCloneNodeIssues( src, dest ) {
	var nodeName, e, data;

	// We do not need to do anything for non-Elements
	if ( dest.nodeType !== 1 ) {
		return;
	}

	nodeName = dest.nodeName.toLowerCase();

	// IE6-8 copies events bound via attachEvent when using cloneNode.
	if ( !support.noCloneEvent && dest[ jQuery.expando ] ) {
		data = jQuery._data( dest );

		for ( e in data.events ) {
			jQuery.removeEvent( dest, e, data.handle );
		}

		// Event data gets referenced instead of copied if the expando gets copied too
		dest.removeAttribute( jQuery.expando );
	}

	// IE blanks contents when cloning scripts, and tries to evaluate newly-set text
	if ( nodeName === "script" && dest.text !== src.text ) {
		disableScript( dest ).text = src.text;
		restoreScript( dest );

	// IE6-10 improperly clones children of object elements using classid.
	// IE10 throws NoModificationAllowedError if parent is null, #12132.
	} else if ( nodeName === "object" ) {
		if ( dest.parentNode ) {
			dest.outerHTML = src.outerHTML;
		}

		// This path appears unavoidable for IE9. When cloning an object
		// element in IE9, the outerHTML strategy above is not sufficient.
		// If the src has innerHTML and the destination does not,
		// copy the src.innerHTML into the dest.innerHTML. #10324
		if ( support.html5Clone && ( src.innerHTML && !jQuery.trim(dest.innerHTML) ) ) {
			dest.innerHTML = src.innerHTML;
		}

	} else if ( nodeName === "input" && rcheckableType.test( src.type ) ) {
		// IE6-8 fails to persist the checked state of a cloned checkbox
		// or radio button. Worse, IE6-7 fail to give the cloned element
		// a checked appearance if the defaultChecked value isn't also set

		dest.defaultChecked = dest.checked = src.checked;

		// IE6-7 get confused and end up setting the value of a cloned
		// checkbox/radio button to an empty string instead of "on"
		if ( dest.value !== src.value ) {
			dest.value = src.value;
		}

	// IE6-8 fails to return the selected option to the default selected
	// state when cloning options
	} else if ( nodeName === "option" ) {
		dest.defaultSelected = dest.selected = src.defaultSelected;

	// IE6-8 fails to set the defaultValue to the correct value when
	// cloning other types of input fields
	} else if ( nodeName === "input" || nodeName === "textarea" ) {
		dest.defaultValue = src.defaultValue;
	}
}

jQuery.extend({
	clone: function( elem, dataAndEvents, deepDataAndEvents ) {
		var destElements, node, clone, i, srcElements,
			inPage = jQuery.contains( elem.ownerDocument, elem );

		if ( support.html5Clone || jQuery.isXMLDoc(elem) || !rnoshimcache.test( "<" + elem.nodeName + ">" ) ) {
			clone = elem.cloneNode( true );

		// IE<=8 does not properly clone detached, unknown element nodes
		} else {
			fragmentDiv.innerHTML = elem.outerHTML;
			fragmentDiv.removeChild( clone = fragmentDiv.firstChild );
		}

		if ( (!support.noCloneEvent || !support.noCloneChecked) &&
				(elem.nodeType === 1 || elem.nodeType === 11) && !jQuery.isXMLDoc(elem) ) {

			// We eschew Sizzle here for performance reasons: http://jsperf.com/getall-vs-sizzle/2
			destElements = getAll( clone );
			srcElements = getAll( elem );

			// Fix all IE cloning issues
			for ( i = 0; (node = srcElements[i]) != null; ++i ) {
				// Ensure that the destination node is not null; Fixes #9587
				if ( destElements[i] ) {
					fixCloneNodeIssues( node, destElements[i] );
				}
			}
		}

		// Copy the events from the original to the clone
		if ( dataAndEvents ) {
			if ( deepDataAndEvents ) {
				srcElements = srcElements || getAll( elem );
				destElements = destElements || getAll( clone );

				for ( i = 0; (node = srcElements[i]) != null; i++ ) {
					cloneCopyEvent( node, destElements[i] );
				}
			} else {
				cloneCopyEvent( elem, clone );
			}
		}

		// Preserve script evaluation history
		destElements = getAll( clone, "script" );
		if ( destElements.length > 0 ) {
			setGlobalEval( destElements, !inPage && getAll( elem, "script" ) );
		}

		destElements = srcElements = node = null;

		// Return the cloned set
		return clone;
	},

	buildFragment: function( elems, context, scripts, selection ) {
		var j, elem, contains,
			tmp, tag, tbody, wrap,
			l = elems.length,

			// Ensure a safe fragment
			safe = createSafeFragment( context ),

			nodes = [],
			i = 0;

		for ( ; i < l; i++ ) {
			elem = elems[ i ];

			if ( elem || elem === 0 ) {

				// Add nodes directly
				if ( jQuery.type( elem ) === "object" ) {
					jQuery.merge( nodes, elem.nodeType ? [ elem ] : elem );

				// Convert non-html into a text node
				} else if ( !rhtml.test( elem ) ) {
					nodes.push( context.createTextNode( elem ) );

				// Convert html into DOM nodes
				} else {
					tmp = tmp || safe.appendChild( context.createElement("div") );

					// Deserialize a standard representation
					tag = (rtagName.exec( elem ) || [ "", "" ])[ 1 ].toLowerCase();
					wrap = wrapMap[ tag ] || wrapMap._default;

					tmp.innerHTML = wrap[1] + elem.replace( rxhtmlTag, "<$1></$2>" ) + wrap[2];

					// Descend through wrappers to the right content
					j = wrap[0];
					while ( j-- ) {
						tmp = tmp.lastChild;
					}

					// Manually add leading whitespace removed by IE
					if ( !support.leadingWhitespace && rleadingWhitespace.test( elem ) ) {
						nodes.push( context.createTextNode( rleadingWhitespace.exec( elem )[0] ) );
					}

					// Remove IE's autoinserted <tbody> from table fragments
					if ( !support.tbody ) {

						// String was a <table>, *may* have spurious <tbody>
						elem = tag === "table" && !rtbody.test( elem ) ?
							tmp.firstChild :

							// String was a bare <thead> or <tfoot>
							wrap[1] === "<table>" && !rtbody.test( elem ) ?
								tmp :
								0;

						j = elem && elem.childNodes.length;
						while ( j-- ) {
							if ( jQuery.nodeName( (tbody = elem.childNodes[j]), "tbody" ) && !tbody.childNodes.length ) {
								elem.removeChild( tbody );
							}
						}
					}

					jQuery.merge( nodes, tmp.childNodes );

					// Fix #12392 for WebKit and IE > 9
					tmp.textContent = "";

					// Fix #12392 for oldIE
					while ( tmp.firstChild ) {
						tmp.removeChild( tmp.firstChild );
					}

					// Remember the top-level container for proper cleanup
					tmp = safe.lastChild;
				}
			}
		}

		// Fix #11356: Clear elements from fragment
		if ( tmp ) {
			safe.removeChild( tmp );
		}

		// Reset defaultChecked for any radios and checkboxes
		// about to be appended to the DOM in IE 6/7 (#8060)
		if ( !support.appendChecked ) {
			jQuery.grep( getAll( nodes, "input" ), fixDefaultChecked );
		}

		i = 0;
		while ( (elem = nodes[ i++ ]) ) {

			// #4087 - If origin and destination elements are the same, and this is
			// that element, do not do anything
			if ( selection && jQuery.inArray( elem, selection ) !== -1 ) {
				continue;
			}

			contains = jQuery.contains( elem.ownerDocument, elem );

			// Append to fragment
			tmp = getAll( safe.appendChild( elem ), "script" );

			// Preserve script evaluation history
			if ( contains ) {
				setGlobalEval( tmp );
			}

			// Capture executables
			if ( scripts ) {
				j = 0;
				while ( (elem = tmp[ j++ ]) ) {
					if ( rscriptType.test( elem.type || "" ) ) {
						scripts.push( elem );
					}
				}
			}
		}

		tmp = null;

		return safe;
	},

	cleanData: function( elems, /* internal */ acceptData ) {
		var elem, type, id, data,
			i = 0,
			internalKey = jQuery.expando,
			cache = jQuery.cache,
			deleteExpando = support.deleteExpando,
			special = jQuery.event.special;

		for ( ; (elem = elems[i]) != null; i++ ) {
			if ( acceptData || jQuery.acceptData( elem ) ) {

				id = elem[ internalKey ];
				data = id && cache[ id ];

				if ( data ) {
					if ( data.events ) {
						for ( type in data.events ) {
							if ( special[ type ] ) {
								jQuery.event.remove( elem, type );

							// This is a shortcut to avoid jQuery.event.remove's overhead
							} else {
								jQuery.removeEvent( elem, type, data.handle );
							}
						}
					}

					// Remove cache only if it was not already removed by jQuery.event.remove
					if ( cache[ id ] ) {

						delete cache[ id ];

						// IE does not allow us to delete expando properties from nodes,
						// nor does it have a removeAttribute function on Document nodes;
						// we must handle all of these cases
						if ( deleteExpando ) {
							delete elem[ internalKey ];

						} else if ( typeof elem.removeAttribute !== strundefined ) {
							elem.removeAttribute( internalKey );

						} else {
							elem[ internalKey ] = null;
						}

						deletedIds.push( id );
					}
				}
			}
		}
	}
});

jQuery.fn.extend({
	text: function( value ) {
		return access( this, function( value ) {
			return value === undefined ?
				jQuery.text( this ) :
				this.empty().append( ( this[0] && this[0].ownerDocument || document ).createTextNode( value ) );
		}, null, value, arguments.length );
	},

	append: function() {
		return this.domManip( arguments, function( elem ) {
			if ( this.nodeType === 1 || this.nodeType === 11 || this.nodeType === 9 ) {
				var target = manipulationTarget( this, elem );
				target.appendChild( elem );
			}
		});
	},

	prepend: function() {
		return this.domManip( arguments, function( elem ) {
			if ( this.nodeType === 1 || this.nodeType === 11 || this.nodeType === 9 ) {
				var target = manipulationTarget( this, elem );
				target.insertBefore( elem, target.firstChild );
			}
		});
	},

	before: function() {
		return this.domManip( arguments, function( elem ) {
			if ( this.parentNode ) {
				this.parentNode.insertBefore( elem, this );
			}
		});
	},

	after: function() {
		return this.domManip( arguments, function( elem ) {
			if ( this.parentNode ) {
				this.parentNode.insertBefore( elem, this.nextSibling );
			}
		});
	},

	remove: function( selector, keepData /* Internal Use Only */ ) {
		var elem,
			elems = selector ? jQuery.filter( selector, this ) : this,
			i = 0;

		for ( ; (elem = elems[i]) != null; i++ ) {

			if ( !keepData && elem.nodeType === 1 ) {
				jQuery.cleanData( getAll( elem ) );
			}

			if ( elem.parentNode ) {
				if ( keepData && jQuery.contains( elem.ownerDocument, elem ) ) {
					setGlobalEval( getAll( elem, "script" ) );
				}
				elem.parentNode.removeChild( elem );
			}
		}

		return this;
	},

	empty: function() {
		var elem,
			i = 0;

		for ( ; (elem = this[i]) != null; i++ ) {
			// Remove element nodes and prevent memory leaks
			if ( elem.nodeType === 1 ) {
				jQuery.cleanData( getAll( elem, false ) );
			}

			// Remove any remaining nodes
			while ( elem.firstChild ) {
				elem.removeChild( elem.firstChild );
			}

			// If this is a select, ensure that it displays empty (#12336)
			// Support: IE<9
			if ( elem.options && jQuery.nodeName( elem, "select" ) ) {
				elem.options.length = 0;
			}
		}

		return this;
	},

	clone: function( dataAndEvents, deepDataAndEvents ) {
		dataAndEvents = dataAndEvents == null ? false : dataAndEvents;
		deepDataAndEvents = deepDataAndEvents == null ? dataAndEvents : deepDataAndEvents;

		return this.map(function() {
			return jQuery.clone( this, dataAndEvents, deepDataAndEvents );
		});
	},

	html: function( value ) {
		return access( this, function( value ) {
			var elem = this[ 0 ] || {},
				i = 0,
				l = this.length;

			if ( value === undefined ) {
				return elem.nodeType === 1 ?
					elem.innerHTML.replace( rinlinejQuery, "" ) :
					undefined;
			}

			// See if we can take a shortcut and just use innerHTML
			if ( typeof value === "string" && !rnoInnerhtml.test( value ) &&
				( support.htmlSerialize || !rnoshimcache.test( value )  ) &&
				( support.leadingWhitespace || !rleadingWhitespace.test( value ) ) &&
				!wrapMap[ (rtagName.exec( value ) || [ "", "" ])[ 1 ].toLowerCase() ] ) {

				value = value.replace( rxhtmlTag, "<$1></$2>" );

				try {
					for (; i < l; i++ ) {
						// Remove element nodes and prevent memory leaks
						elem = this[i] || {};
						if ( elem.nodeType === 1 ) {
							jQuery.cleanData( getAll( elem, false ) );
							elem.innerHTML = value;
						}
					}

					elem = 0;

				// If using innerHTML throws an exception, use the fallback method
				} catch(e) {}
			}

			if ( elem ) {
				this.empty().append( value );
			}
		}, null, value, arguments.length );
	},

	replaceWith: function() {
		var arg = arguments[ 0 ];

		// Make the changes, replacing each context element with the new content
		this.domManip( arguments, function( elem ) {
			arg = this.parentNode;

			jQuery.cleanData( getAll( this ) );

			if ( arg ) {
				arg.replaceChild( elem, this );
			}
		});

		// Force removal if there was no new content (e.g., from empty arguments)
		return arg && (arg.length || arg.nodeType) ? this : this.remove();
	},

	detach: function( selector ) {
		return this.remove( selector, true );
	},

	domManip: function( args, callback ) {

		// Flatten any nested arrays
		args = concat.apply( [], args );

		var first, node, hasScripts,
			scripts, doc, fragment,
			i = 0,
			l = this.length,
			set = this,
			iNoClone = l - 1,
			value = args[0],
			isFunction = jQuery.isFunction( value );

		// We can't cloneNode fragments that contain checked, in WebKit
		if ( isFunction ||
				( l > 1 && typeof value === "string" &&
					!support.checkClone && rchecked.test( value ) ) ) {
			return this.each(function( index ) {
				var self = set.eq( index );
				if ( isFunction ) {
					args[0] = value.call( this, index, self.html() );
				}
				self.domManip( args, callback );
			});
		}

		if ( l ) {
			fragment = jQuery.buildFragment( args, this[ 0 ].ownerDocument, false, this );
			first = fragment.firstChild;

			if ( fragment.childNodes.length === 1 ) {
				fragment = first;
			}

			if ( first ) {
				scripts = jQuery.map( getAll( fragment, "script" ), disableScript );
				hasScripts = scripts.length;

				// Use the original fragment for the last item instead of the first because it can end up
				// being emptied incorrectly in certain situations (#8070).
				for ( ; i < l; i++ ) {
					node = fragment;

					if ( i !== iNoClone ) {
						node = jQuery.clone( node, true, true );

						// Keep references to cloned scripts for later restoration
						if ( hasScripts ) {
							jQuery.merge( scripts, getAll( node, "script" ) );
						}
					}

					callback.call( this[i], node, i );
				}

				if ( hasScripts ) {
					doc = scripts[ scripts.length - 1 ].ownerDocument;

					// Reenable scripts
					jQuery.map( scripts, restoreScript );

					// Evaluate executable scripts on first document insertion
					for ( i = 0; i < hasScripts; i++ ) {
						node = scripts[ i ];
						if ( rscriptType.test( node.type || "" ) &&
							!jQuery._data( node, "globalEval" ) && jQuery.contains( doc, node ) ) {

							if ( node.src ) {
								// Optional AJAX dependency, but won't run scripts if not present
								if ( jQuery._evalUrl ) {
									jQuery._evalUrl( node.src );
								}
							} else {
								jQuery.globalEval( ( node.text || node.textContent || node.innerHTML || "" ).replace( rcleanScript, "" ) );
							}
						}
					}
				}

				// Fix #11809: Avoid leaking memory
				fragment = first = null;
			}
		}

		return this;
	}
});

jQuery.each({
	appendTo: "append",
	prependTo: "prepend",
	insertBefore: "before",
	insertAfter: "after",
	replaceAll: "replaceWith"
}, function( name, original ) {
	jQuery.fn[ name ] = function( selector ) {
		var elems,
			i = 0,
			ret = [],
			insert = jQuery( selector ),
			last = insert.length - 1;

		for ( ; i <= last; i++ ) {
			elems = i === last ? this : this.clone(true);
			jQuery( insert[i] )[ original ]( elems );

			// Modern browsers can apply jQuery collections as arrays, but oldIE needs a .get()
			push.apply( ret, elems.get() );
		}

		return this.pushStack( ret );
	};
});


var iframe,
	elemdisplay = {};

/**
 * Retrieve the actual display of a element
 * @param {String} name nodeName of the element
 * @param {Object} doc Document object
 */
// Called only from within defaultDisplay
function actualDisplay( name, doc ) {
	var style,
		elem = jQuery( doc.createElement( name ) ).appendTo( doc.body ),

		// getDefaultComputedStyle might be reliably used only on attached element
		display = window.getDefaultComputedStyle && ( style = window.getDefaultComputedStyle( elem[ 0 ] ) ) ?

			// Use of this method is a temporary fix (more like optmization) until something better comes along,
			// since it was removed from specification and supported only in FF
			style.display : jQuery.css( elem[ 0 ], "display" );

	// We don't have any data stored on the element,
	// so use "detach" method as fast way to get rid of the element
	elem.detach();

	return display;
}

/**
 * Try to determine the default display value of an element
 * @param {String} nodeName
 */
function defaultDisplay( nodeName ) {
	var doc = document,
		display = elemdisplay[ nodeName ];

	if ( !display ) {
		display = actualDisplay( nodeName, doc );

		// If the simple way fails, read from inside an iframe
		if ( display === "none" || !display ) {

			// Use the already-created iframe if possible
			iframe = (iframe || jQuery( "<iframe frameborder='0' width='0' height='0'/>" )).appendTo( doc.documentElement );

			// Always write a new HTML skeleton so Webkit and Firefox don't choke on reuse
			doc = ( iframe[ 0 ].contentWindow || iframe[ 0 ].contentDocument ).document;

			// Support: IE
			doc.write();
			doc.close();

			display = actualDisplay( nodeName, doc );
			iframe.detach();
		}

		// Store the correct default display
		elemdisplay[ nodeName ] = display;
	}

	return display;
}


(function() {
	var shrinkWrapBlocksVal;

	support.shrinkWrapBlocks = function() {
		if ( shrinkWrapBlocksVal != null ) {
			return shrinkWrapBlocksVal;
		}

		// Will be changed later if needed.
		shrinkWrapBlocksVal = false;

		// Minified: var b,c,d
		var div, body, container;

		body = document.getElementsByTagName( "body" )[ 0 ];
		if ( !body || !body.style ) {
			// Test fired too early or in an unsupported environment, exit.
			return;
		}

		// Setup
		div = document.createElement( "div" );
		container = document.createElement( "div" );
		container.style.cssText = "position:absolute;border:0;width:0;height:0;top:0;left:-9999px";
		body.appendChild( container ).appendChild( div );

		// Support: IE6
		// Check if elements with layout shrink-wrap their children
		if ( typeof div.style.zoom !== strundefined ) {
			// Reset CSS: box-sizing; display; margin; border
			div.style.cssText =
				// Support: Firefox<29, Android 2.3
				// Vendor-prefix box-sizing
				"-webkit-box-sizing:content-box;-moz-box-sizing:content-box;" +
				"box-sizing:content-box;display:block;margin:0;border:0;" +
				"padding:1px;width:1px;zoom:1";
			div.appendChild( document.createElement( "div" ) ).style.width = "5px";
			shrinkWrapBlocksVal = div.offsetWidth !== 3;
		}

		body.removeChild( container );

		return shrinkWrapBlocksVal;
	};

})();
var rmargin = (/^margin/);

var rnumnonpx = new RegExp( "^(" + pnum + ")(?!px)[a-z%]+$", "i" );



var getStyles, curCSS,
	rposition = /^(top|right|bottom|left)$/;

if ( window.getComputedStyle ) {
	getStyles = function( elem ) {
		// Support: IE<=11+, Firefox<=30+ (#15098, #14150)
		// IE throws on elements created in popups
		// FF meanwhile throws on frame elements through "defaultView.getComputedStyle"
		if ( elem.ownerDocument.defaultView.opener ) {
			return elem.ownerDocument.defaultView.getComputedStyle( elem, null );
		}

		return window.getComputedStyle( elem, null );
	};

	curCSS = function( elem, name, computed ) {
		var width, minWidth, maxWidth, ret,
			style = elem.style;

		computed = computed || getStyles( elem );

		// getPropertyValue is only needed for .css('filter') in IE9, see #12537
		ret = computed ? computed.getPropertyValue( name ) || computed[ name ] : undefined;

		if ( computed ) {

			if ( ret === "" && !jQuery.contains( elem.ownerDocument, elem ) ) {
				ret = jQuery.style( elem, name );
			}

			// A tribute to the "awesome hack by Dean Edwards"
			// Chrome < 17 and Safari 5.0 uses "computed value" instead of "used value" for margin-right
			// Safari 5.1.7 (at least) returns percentage for a larger set of values, but width seems to be reliably pixels
			// this is against the CSSOM draft spec: http://dev.w3.org/csswg/cssom/#resolved-values
			if ( rnumnonpx.test( ret ) && rmargin.test( name ) ) {

				// Remember the original values
				width = style.width;
				minWidth = style.minWidth;
				maxWidth = style.maxWidth;

				// Put in the new values to get a computed value out
				style.minWidth = style.maxWidth = style.width = ret;
				ret = computed.width;

				// Revert the changed values
				style.width = width;
				style.minWidth = minWidth;
				style.maxWidth = maxWidth;
			}
		}

		// Support: IE
		// IE returns zIndex value as an integer.
		return ret === undefined ?
			ret :
			ret + "";
	};
} else if ( document.documentElement.currentStyle ) {
	getStyles = function( elem ) {
		return elem.currentStyle;
	};

	curCSS = function( elem, name, computed ) {
		var left, rs, rsLeft, ret,
			style = elem.style;

		computed = computed || getStyles( elem );
		ret = computed ? computed[ name ] : undefined;

		// Avoid setting ret to empty string here
		// so we don't default to auto
		if ( ret == null && style && style[ name ] ) {
			ret = style[ name ];
		}

		// From the awesome hack by Dean Edwards
		// http://erik.eae.net/archives/2007/07/27/18.54.15/#comment-102291

		// If we're not dealing with a regular pixel number
		// but a number that has a weird ending, we need to convert it to pixels
		// but not position css attributes, as those are proportional to the parent element instead
		// and we can't measure the parent instead because it might trigger a "stacking dolls" problem
		if ( rnumnonpx.test( ret ) && !rposition.test( name ) ) {

			// Remember the original values
			left = style.left;
			rs = elem.runtimeStyle;
			rsLeft = rs && rs.left;

			// Put in the new values to get a computed value out
			if ( rsLeft ) {
				rs.left = elem.currentStyle.left;
			}
			style.left = name === "fontSize" ? "1em" : ret;
			ret = style.pixelLeft + "px";

			// Revert the changed values
			style.left = left;
			if ( rsLeft ) {
				rs.left = rsLeft;
			}
		}

		// Support: IE
		// IE returns zIndex value as an integer.
		return ret === undefined ?
			ret :
			ret + "" || "auto";
	};
}




function addGetHookIf( conditionFn, hookFn ) {
	// Define the hook, we'll check on the first run if it's really needed.
	return {
		get: function() {
			var condition = conditionFn();

			if ( condition == null ) {
				// The test was not ready at this point; screw the hook this time
				// but check again when needed next time.
				return;
			}

			if ( condition ) {
				// Hook not needed (or it's not possible to use it due to missing dependency),
				// remove it.
				// Since there are no other hooks for marginRight, remove the whole object.
				delete this.get;
				return;
			}

			// Hook needed; redefine it so that the support test is not executed again.

			return (this.get = hookFn).apply( this, arguments );
		}
	};
}


(function() {
	// Minified: var b,c,d,e,f,g, h,i
	var div, style, a, pixelPositionVal, boxSizingReliableVal,
		reliableHiddenOffsetsVal, reliableMarginRightVal;

	// Setup
	div = document.createElement( "div" );
	div.innerHTML = "  <link/><table></table><a href='/a'>a</a><input type='checkbox'/>";
	a = div.getElementsByTagName( "a" )[ 0 ];
	style = a && a.style;

	// Finish early in limited (non-browser) environments
	if ( !style ) {
		return;
	}

	style.cssText = "float:left;opacity:.5";

	// Support: IE<9
	// Make sure that element opacity exists (as opposed to filter)
	support.opacity = style.opacity === "0.5";

	// Verify style float existence
	// (IE uses styleFloat instead of cssFloat)
	support.cssFloat = !!style.cssFloat;

	div.style.backgroundClip = "content-box";
	div.cloneNode( true ).style.backgroundClip = "";
	support.clearCloneStyle = div.style.backgroundClip === "content-box";

	// Support: Firefox<29, Android 2.3
	// Vendor-prefix box-sizing
	support.boxSizing = style.boxSizing === "" || style.MozBoxSizing === "" ||
		style.WebkitBoxSizing === "";

	jQuery.extend(support, {
		reliableHiddenOffsets: function() {
			if ( reliableHiddenOffsetsVal == null ) {
				computeStyleTests();
			}
			return reliableHiddenOffsetsVal;
		},

		boxSizingReliable: function() {
			if ( boxSizingReliableVal == null ) {
				computeStyleTests();
			}
			return boxSizingReliableVal;
		},

		pixelPosition: function() {
			if ( pixelPositionVal == null ) {
				computeStyleTests();
			}
			return pixelPositionVal;
		},

		// Support: Android 2.3
		reliableMarginRight: function() {
			if ( reliableMarginRightVal == null ) {
				computeStyleTests();
			}
			return reliableMarginRightVal;
		}
	});

	function computeStyleTests() {
		// Minified: var b,c,d,j
		var div, body, container, contents;

		body = document.getElementsByTagName( "body" )[ 0 ];
		if ( !body || !body.style ) {
			// Test fired too early or in an unsupported environment, exit.
			return;
		}

		// Setup
		div = document.createElement( "div" );
		container = document.createElement( "div" );
		container.style.cssText = "position:absolute;border:0;width:0;height:0;top:0;left:-9999px";
		body.appendChild( container ).appendChild( div );

		div.style.cssText =
			// Support: Firefox<29, Android 2.3
			// Vendor-prefix box-sizing
			"-webkit-box-sizing:border-box;-moz-box-sizing:border-box;" +
			"box-sizing:border-box;display:block;margin-top:1%;top:1%;" +
			"border:1px;padding:1px;width:4px;position:absolute";

		// Support: IE<9
		// Assume reasonable values in the absence of getComputedStyle
		pixelPositionVal = boxSizingReliableVal = false;
		reliableMarginRightVal = true;

		// Check for getComputedStyle so that this code is not run in IE<9.
		if ( window.getComputedStyle ) {
			pixelPositionVal = ( window.getComputedStyle( div, null ) || {} ).top !== "1%";
			boxSizingReliableVal =
				( window.getComputedStyle( div, null ) || { width: "4px" } ).width === "4px";

			// Support: Android 2.3
			// Div with explicit width and no margin-right incorrectly
			// gets computed margin-right based on width of container (#3333)
			// WebKit Bug 13343 - getComputedStyle returns wrong value for margin-right
			contents = div.appendChild( document.createElement( "div" ) );

			// Reset CSS: box-sizing; display; margin; border; padding
			contents.style.cssText = div.style.cssText =
				// Support: Firefox<29, Android 2.3
				// Vendor-prefix box-sizing
				"-webkit-box-sizing:content-box;-moz-box-sizing:content-box;" +
				"box-sizing:content-box;display:block;margin:0;border:0;padding:0";
			contents.style.marginRight = contents.style.width = "0";
			div.style.width = "1px";

			reliableMarginRightVal =
				!parseFloat( ( window.getComputedStyle( contents, null ) || {} ).marginRight );

			div.removeChild( contents );
		}

		// Support: IE8
		// Check if table cells still have offsetWidth/Height when they are set
		// to display:none and there are still other visible table cells in a
		// table row; if so, offsetWidth/Height are not reliable for use when
		// determining if an element has been hidden directly using
		// display:none (it is still safe to use offsets if a parent element is
		// hidden; don safety goggles and see bug #4512 for more information).
		div.innerHTML = "<table><tr><td></td><td>t</td></tr></table>";
		contents = div.getElementsByTagName( "td" );
		contents[ 0 ].style.cssText = "margin:0;border:0;padding:0;display:none";
		reliableHiddenOffsetsVal = contents[ 0 ].offsetHeight === 0;
		if ( reliableHiddenOffsetsVal ) {
			contents[ 0 ].style.display = "";
			contents[ 1 ].style.display = "none";
			reliableHiddenOffsetsVal = contents[ 0 ].offsetHeight === 0;
		}

		body.removeChild( container );
	}

})();


// A method for quickly swapping in/out CSS properties to get correct calculations.
jQuery.swap = function( elem, options, callback, args ) {
	var ret, name,
		old = {};

	// Remember the old values, and insert the new ones
	for ( name in options ) {
		old[ name ] = elem.style[ name ];
		elem.style[ name ] = options[ name ];
	}

	ret = callback.apply( elem, args || [] );

	// Revert the old values
	for ( name in options ) {
		elem.style[ name ] = old[ name ];
	}

	return ret;
};


var
		ralpha = /alpha\([^)]*\)/i,
	ropacity = /opacity\s*=\s*([^)]*)/,

	// swappable if display is none or starts with table except "table", "table-cell", or "table-caption"
	// see here for display values: https://developer.mozilla.org/en-US/docs/CSS/display
	rdisplayswap = /^(none|table(?!-c[ea]).+)/,
	rnumsplit = new RegExp( "^(" + pnum + ")(.*)$", "i" ),
	rrelNum = new RegExp( "^([+-])=(" + pnum + ")", "i" ),

	cssShow = { position: "absolute", visibility: "hidden", display: "block" },
	cssNormalTransform = {
		letterSpacing: "0",
		fontWeight: "400"
	},

	cssPrefixes = [ "Webkit", "O", "Moz", "ms" ];


// return a css property mapped to a potentially vendor prefixed property
function vendorPropName( style, name ) {

	// shortcut for names that are not vendor prefixed
	if ( name in style ) {
		return name;
	}

	// check for vendor prefixed names
	var capName = name.charAt(0).toUpperCase() + name.slice(1),
		origName = name,
		i = cssPrefixes.length;

	while ( i-- ) {
		name = cssPrefixes[ i ] + capName;
		if ( name in style ) {
			return name;
		}
	}

	return origName;
}

function showHide( elements, show ) {
	var display, elem, hidden,
		values = [],
		index = 0,
		length = elements.length;

	for ( ; index < length; index++ ) {
		elem = elements[ index ];
		if ( !elem.style ) {
			continue;
		}

		values[ index ] = jQuery._data( elem, "olddisplay" );
		display = elem.style.display;
		if ( show ) {
			// Reset the inline display of this element to learn if it is
			// being hidden by cascaded rules or not
			if ( !values[ index ] && display === "none" ) {
				elem.style.display = "";
			}

			// Set elements which have been overridden with display: none
			// in a stylesheet to whatever the default browser style is
			// for such an element
			if ( elem.style.display === "" && isHidden( elem ) ) {
				values[ index ] = jQuery._data( elem, "olddisplay", defaultDisplay(elem.nodeName) );
			}
		} else {
			hidden = isHidden( elem );

			if ( display && display !== "none" || !hidden ) {
				jQuery._data( elem, "olddisplay", hidden ? display : jQuery.css( elem, "display" ) );
			}
		}
	}

	// Set the display of most of the elements in a second loop
	// to avoid the constant reflow
	for ( index = 0; index < length; index++ ) {
		elem = elements[ index ];
		if ( !elem.style ) {
			continue;
		}
		if ( !show || elem.style.display === "none" || elem.style.display === "" ) {
			elem.style.display = show ? values[ index ] || "" : "none";
		}
	}

	return elements;
}

function setPositiveNumber( elem, value, subtract ) {
	var matches = rnumsplit.exec( value );
	return matches ?
		// Guard against undefined "subtract", e.g., when used as in cssHooks
		Math.max( 0, matches[ 1 ] - ( subtract || 0 ) ) + ( matches[ 2 ] || "px" ) :
		value;
}

function augmentWidthOrHeight( elem, name, extra, isBorderBox, styles ) {
	var i = extra === ( isBorderBox ? "border" : "content" ) ?
		// If we already have the right measurement, avoid augmentation
		4 :
		// Otherwise initialize for horizontal or vertical properties
		name === "width" ? 1 : 0,

		val = 0;

	for ( ; i < 4; i += 2 ) {
		// both box models exclude margin, so add it if we want it
		if ( extra === "margin" ) {
			val += jQuery.css( elem, extra + cssExpand[ i ], true, styles );
		}

		if ( isBorderBox ) {
			// border-box includes padding, so remove it if we want content
			if ( extra === "content" ) {
				val -= jQuery.css( elem, "padding" + cssExpand[ i ], true, styles );
			}

			// at this point, extra isn't border nor margin, so remove border
			if ( extra !== "margin" ) {
				val -= jQuery.css( elem, "border" + cssExpand[ i ] + "Width", true, styles );
			}
		} else {
			// at this point, extra isn't content, so add padding
			val += jQuery.css( elem, "padding" + cssExpand[ i ], true, styles );

			// at this point, extra isn't content nor padding, so add border
			if ( extra !== "padding" ) {
				val += jQuery.css( elem, "border" + cssExpand[ i ] + "Width", true, styles );
			}
		}
	}

	return val;
}

function getWidthOrHeight( elem, name, extra ) {

	// Start with offset property, which is equivalent to the border-box value
	var valueIsBorderBox = true,
		val = name === "width" ? elem.offsetWidth : elem.offsetHeight,
		styles = getStyles( elem ),
		isBorderBox = support.boxSizing && jQuery.css( elem, "boxSizing", false, styles ) === "border-box";

	// some non-html elements return undefined for offsetWidth, so check for null/undefined
	// svg - https://bugzilla.mozilla.org/show_bug.cgi?id=649285
	// MathML - https://bugzilla.mozilla.org/show_bug.cgi?id=491668
	if ( val <= 0 || val == null ) {
		// Fall back to computed then uncomputed css if necessary
		val = curCSS( elem, name, styles );
		if ( val < 0 || val == null ) {
			val = elem.style[ name ];
		}

		// Computed unit is not pixels. Stop here and return.
		if ( rnumnonpx.test(val) ) {
			return val;
		}

		// we need the check for style in case a browser which returns unreliable values
		// for getComputedStyle silently falls back to the reliable elem.style
		valueIsBorderBox = isBorderBox && ( support.boxSizingReliable() || val === elem.style[ name ] );

		// Normalize "", auto, and prepare for extra
		val = parseFloat( val ) || 0;
	}

	// use the active box-sizing model to add/subtract irrelevant styles
	return ( val +
		augmentWidthOrHeight(
			elem,
			name,
			extra || ( isBorderBox ? "border" : "content" ),
			valueIsBorderBox,
			styles
		)
	) + "px";
}

jQuery.extend({
	// Add in style property hooks for overriding the default
	// behavior of getting and setting a style property
	cssHooks: {
		opacity: {
			get: function( elem, computed ) {
				if ( computed ) {
					// We should always get a number back from opacity
					var ret = curCSS( elem, "opacity" );
					return ret === "" ? "1" : ret;
				}
			}
		}
	},

	// Don't automatically add "px" to these possibly-unitless properties
	cssNumber: {
		"columnCount": true,
		"fillOpacity": true,
		"flexGrow": true,
		"flexShrink": true,
		"fontWeight": true,
		"lineHeight": true,
		"opacity": true,
		"order": true,
		"orphans": true,
		"widows": true,
		"zIndex": true,
		"zoom": true
	},

	// Add in properties whose names you wish to fix before
	// setting or getting the value
	cssProps: {
		// normalize float css property
		"float": support.cssFloat ? "cssFloat" : "styleFloat"
	},

	// Get and set the style property on a DOM Node
	style: function( elem, name, value, extra ) {
		// Don't set styles on text and comment nodes
		if ( !elem || elem.nodeType === 3 || elem.nodeType === 8 || !elem.style ) {
			return;
		}

		// Make sure that we're working with the right name
		var ret, type, hooks,
			origName = jQuery.camelCase( name ),
			style = elem.style;

		name = jQuery.cssProps[ origName ] || ( jQuery.cssProps[ origName ] = vendorPropName( style, origName ) );

		// gets hook for the prefixed version
		// followed by the unprefixed version
		hooks = jQuery.cssHooks[ name ] || jQuery.cssHooks[ origName ];

		// Check if we're setting a value
		if ( value !== undefined ) {
			type = typeof value;

			// convert relative number strings (+= or -=) to relative numbers. #7345
			if ( type === "string" && (ret = rrelNum.exec( value )) ) {
				value = ( ret[1] + 1 ) * ret[2] + parseFloat( jQuery.css( elem, name ) );
				// Fixes bug #9237
				type = "number";
			}

			// Make sure that null and NaN values aren't set. See: #7116
			if ( value == null || value !== value ) {
				return;
			}

			// If a number was passed in, add 'px' to the (except for certain CSS properties)
			if ( type === "number" && !jQuery.cssNumber[ origName ] ) {
				value += "px";
			}

			// Fixes #8908, it can be done more correctly by specifing setters in cssHooks,
			// but it would mean to define eight (for every problematic property) identical functions
			if ( !support.clearCloneStyle && value === "" && name.indexOf("background") === 0 ) {
				style[ name ] = "inherit";
			}

			// If a hook was provided, use that value, otherwise just set the specified value
			if ( !hooks || !("set" in hooks) || (value = hooks.set( elem, value, extra )) !== undefined ) {

				// Support: IE
				// Swallow errors from 'invalid' CSS values (#5509)
				try {
					style[ name ] = value;
				} catch(e) {}
			}

		} else {
			// If a hook was provided get the non-computed value from there
			if ( hooks && "get" in hooks && (ret = hooks.get( elem, false, extra )) !== undefined ) {
				return ret;
			}

			// Otherwise just get the value from the style object
			return style[ name ];
		}
	},

	css: function( elem, name, extra, styles ) {
		var num, val, hooks,
			origName = jQuery.camelCase( name );

		// Make sure that we're working with the right name
		name = jQuery.cssProps[ origName ] || ( jQuery.cssProps[ origName ] = vendorPropName( elem.style, origName ) );

		// gets hook for the prefixed version
		// followed by the unprefixed version
		hooks = jQuery.cssHooks[ name ] || jQuery.cssHooks[ origName ];

		// If a hook was provided get the computed value from there
		if ( hooks && "get" in hooks ) {
			val = hooks.get( elem, true, extra );
		}

		// Otherwise, if a way to get the computed value exists, use that
		if ( val === undefined ) {
			val = curCSS( elem, name, styles );
		}

		//convert "normal" to computed value
		if ( val === "normal" && name in cssNormalTransform ) {
			val = cssNormalTransform[ name ];
		}

		// Return, converting to number if forced or a qualifier was provided and val looks numeric
		if ( extra === "" || extra ) {
			num = parseFloat( val );
			return extra === true || jQuery.isNumeric( num ) ? num || 0 : val;
		}
		return val;
	}
});

jQuery.each([ "height", "width" ], function( i, name ) {
	jQuery.cssHooks[ name ] = {
		get: function( elem, computed, extra ) {
			if ( computed ) {
				// certain elements can have dimension info if we invisibly show them
				// however, it must have a current display style that would benefit from this
				return rdisplayswap.test( jQuery.css( elem, "display" ) ) && elem.offsetWidth === 0 ?
					jQuery.swap( elem, cssShow, function() {
						return getWidthOrHeight( elem, name, extra );
					}) :
					getWidthOrHeight( elem, name, extra );
			}
		},

		set: function( elem, value, extra ) {
			var styles = extra && getStyles( elem );
			return setPositiveNumber( elem, value, extra ?
				augmentWidthOrHeight(
					elem,
					name,
					extra,
					support.boxSizing && jQuery.css( elem, "boxSizing", false, styles ) === "border-box",
					styles
				) : 0
			);
		}
	};
});

if ( !support.opacity ) {
	jQuery.cssHooks.opacity = {
		get: function( elem, computed ) {
			// IE uses filters for opacity
			return ropacity.test( (computed && elem.currentStyle ? elem.currentStyle.filter : elem.style.filter) || "" ) ?
				( 0.01 * parseFloat( RegExp.$1 ) ) + "" :
				computed ? "1" : "";
		},

		set: function( elem, value ) {
			var style = elem.style,
				currentStyle = elem.currentStyle,
				opacity = jQuery.isNumeric( value ) ? "alpha(opacity=" + value * 100 + ")" : "",
				filter = currentStyle && currentStyle.filter || style.filter || "";

			// IE has trouble with opacity if it does not have layout
			// Force it by setting the zoom level
			style.zoom = 1;

			// if setting opacity to 1, and no other filters exist - attempt to remove filter attribute #6652
			// if value === "", then remove inline opacity #12685
			if ( ( value >= 1 || value === "" ) &&
					jQuery.trim( filter.replace( ralpha, "" ) ) === "" &&
					style.removeAttribute ) {

				// Setting style.filter to null, "" & " " still leave "filter:" in the cssText
				// if "filter:" is present at all, clearType is disabled, we want to avoid this
				// style.removeAttribute is IE Only, but so apparently is this code path...
				style.removeAttribute( "filter" );

				// if there is no filter style applied in a css rule or unset inline opacity, we are done
				if ( value === "" || currentStyle && !currentStyle.filter ) {
					return;
				}
			}

			// otherwise, set new filter values
			style.filter = ralpha.test( filter ) ?
				filter.replace( ralpha, opacity ) :
				filter + " " + opacity;
		}
	};
}

jQuery.cssHooks.marginRight = addGetHookIf( support.reliableMarginRight,
	function( elem, computed ) {
		if ( computed ) {
			// WebKit Bug 13343 - getComputedStyle returns wrong value for margin-right
			// Work around by temporarily setting element display to inline-block
			return jQuery.swap( elem, { "display": "inline-block" },
				curCSS, [ elem, "marginRight" ] );
		}
	}
);

// These hooks are used by animate to expand properties
jQuery.each({
	margin: "",
	padding: "",
	border: "Width"
}, function( prefix, suffix ) {
	jQuery.cssHooks[ prefix + suffix ] = {
		expand: function( value ) {
			var i = 0,
				expanded = {},

				// assumes a single number if not a string
				parts = typeof value === "string" ? value.split(" ") : [ value ];

			for ( ; i < 4; i++ ) {
				expanded[ prefix + cssExpand[ i ] + suffix ] =
					parts[ i ] || parts[ i - 2 ] || parts[ 0 ];
			}

			return expanded;
		}
	};

	if ( !rmargin.test( prefix ) ) {
		jQuery.cssHooks[ prefix + suffix ].set = setPositiveNumber;
	}
});

jQuery.fn.extend({
	css: function( name, value ) {
		return access( this, function( elem, name, value ) {
			var styles, len,
				map = {},
				i = 0;

			if ( jQuery.isArray( name ) ) {
				styles = getStyles( elem );
				len = name.length;

				for ( ; i < len; i++ ) {
					map[ name[ i ] ] = jQuery.css( elem, name[ i ], false, styles );
				}

				return map;
			}

			return value !== undefined ?
				jQuery.style( elem, name, value ) :
				jQuery.css( elem, name );
		}, name, value, arguments.length > 1 );
	},
	show: function() {
		return showHide( this, true );
	},
	hide: function() {
		return showHide( this );
	},
	toggle: function( state ) {
		if ( typeof state === "boolean" ) {
			return state ? this.show() : this.hide();
		}

		return this.each(function() {
			if ( isHidden( this ) ) {
				jQuery( this ).show();
			} else {
				jQuery( this ).hide();
			}
		});
	}
});


function Tween( elem, options, prop, end, easing ) {
	return new Tween.prototype.init( elem, options, prop, end, easing );
}
jQuery.Tween = Tween;

Tween.prototype = {
	constructor: Tween,
	init: function( elem, options, prop, end, easing, unit ) {
		this.elem = elem;
		this.prop = prop;
		this.easing = easing || "swing";
		this.options = options;
		this.start = this.now = this.cur();
		this.end = end;
		this.unit = unit || ( jQuery.cssNumber[ prop ] ? "" : "px" );
	},
	cur: function() {
		var hooks = Tween.propHooks[ this.prop ];

		return hooks && hooks.get ?
			hooks.get( this ) :
			Tween.propHooks._default.get( this );
	},
	run: function( percent ) {
		var eased,
			hooks = Tween.propHooks[ this.prop ];

		if ( this.options.duration ) {
			this.pos = eased = jQuery.easing[ this.easing ](
				percent, this.options.duration * percent, 0, 1, this.options.duration
			);
		} else {
			this.pos = eased = percent;
		}
		this.now = ( this.end - this.start ) * eased + this.start;

		if ( this.options.step ) {
			this.options.step.call( this.elem, this.now, this );
		}

		if ( hooks && hooks.set ) {
			hooks.set( this );
		} else {
			Tween.propHooks._default.set( this );
		}
		return this;
	}
};

Tween.prototype.init.prototype = Tween.prototype;

Tween.propHooks = {
	_default: {
		get: function( tween ) {
			var result;

			if ( tween.elem[ tween.prop ] != null &&
				(!tween.elem.style || tween.elem.style[ tween.prop ] == null) ) {
				return tween.elem[ tween.prop ];
			}

			// passing an empty string as a 3rd parameter to .css will automatically
			// attempt a parseFloat and fallback to a string if the parse fails
			// so, simple values such as "10px" are parsed to Float.
			// complex values such as "rotate(1rad)" are returned as is.
			result = jQuery.css( tween.elem, tween.prop, "" );
			// Empty strings, null, undefined and "auto" are converted to 0.
			return !result || result === "auto" ? 0 : result;
		},
		set: function( tween ) {
			// use step hook for back compat - use cssHook if its there - use .style if its
			// available and use plain properties where available
			if ( jQuery.fx.step[ tween.prop ] ) {
				jQuery.fx.step[ tween.prop ]( tween );
			} else if ( tween.elem.style && ( tween.elem.style[ jQuery.cssProps[ tween.prop ] ] != null || jQuery.cssHooks[ tween.prop ] ) ) {
				jQuery.style( tween.elem, tween.prop, tween.now + tween.unit );
			} else {
				tween.elem[ tween.prop ] = tween.now;
			}
		}
	}
};

// Support: IE <=9
// Panic based approach to setting things on disconnected nodes

Tween.propHooks.scrollTop = Tween.propHooks.scrollLeft = {
	set: function( tween ) {
		if ( tween.elem.nodeType && tween.elem.parentNode ) {
			tween.elem[ tween.prop ] = tween.now;
		}
	}
};

jQuery.easing = {
	linear: function( p ) {
		return p;
	},
	swing: function( p ) {
		return 0.5 - Math.cos( p * Math.PI ) / 2;
	}
};

jQuery.fx = Tween.prototype.init;

// Back Compat <1.8 extension point
jQuery.fx.step = {};




var
	fxNow, timerId,
	rfxtypes = /^(?:toggle|show|hide)$/,
	rfxnum = new RegExp( "^(?:([+-])=|)(" + pnum + ")([a-z%]*)$", "i" ),
	rrun = /queueHooks$/,
	animationPrefilters = [ defaultPrefilter ],
	tweeners = {
		"*": [ function( prop, value ) {
			var tween = this.createTween( prop, value ),
				target = tween.cur(),
				parts = rfxnum.exec( value ),
				unit = parts && parts[ 3 ] || ( jQuery.cssNumber[ prop ] ? "" : "px" ),

				// Starting value computation is required for potential unit mismatches
				start = ( jQuery.cssNumber[ prop ] || unit !== "px" && +target ) &&
					rfxnum.exec( jQuery.css( tween.elem, prop ) ),
				scale = 1,
				maxIterations = 20;

			if ( start && start[ 3 ] !== unit ) {
				// Trust units reported by jQuery.css
				unit = unit || start[ 3 ];

				// Make sure we update the tween properties later on
				parts = parts || [];

				// Iteratively approximate from a nonzero starting point
				start = +target || 1;

				do {
					// If previous iteration zeroed out, double until we get *something*
					// Use a string for doubling factor so we don't accidentally see scale as unchanged below
					scale = scale || ".5";

					// Adjust and apply
					start = start / scale;
					jQuery.style( tween.elem, prop, start + unit );

				// Update scale, tolerating zero or NaN from tween.cur()
				// And breaking the loop if scale is unchanged or perfect, or if we've just had enough
				} while ( scale !== (scale = tween.cur() / target) && scale !== 1 && --maxIterations );
			}

			// Update tween properties
			if ( parts ) {
				start = tween.start = +start || +target || 0;
				tween.unit = unit;
				// If a +=/-= token was provided, we're doing a relative animation
				tween.end = parts[ 1 ] ?
					start + ( parts[ 1 ] + 1 ) * parts[ 2 ] :
					+parts[ 2 ];
			}

			return tween;
		} ]
	};

// Animations created synchronously will run synchronously
function createFxNow() {
	setTimeout(function() {
		fxNow = undefined;
	});
	return ( fxNow = jQuery.now() );
}

// Generate parameters to create a standard animation
function genFx( type, includeWidth ) {
	var which,
		attrs = { height: type },
		i = 0;

	// if we include width, step value is 1 to do all cssExpand values,
	// if we don't include width, step value is 2 to skip over Left and Right
	includeWidth = includeWidth ? 1 : 0;
	for ( ; i < 4 ; i += 2 - includeWidth ) {
		which = cssExpand[ i ];
		attrs[ "margin" + which ] = attrs[ "padding" + which ] = type;
	}

	if ( includeWidth ) {
		attrs.opacity = attrs.width = type;
	}

	return attrs;
}

function createTween( value, prop, animation ) {
	var tween,
		collection = ( tweeners[ prop ] || [] ).concat( tweeners[ "*" ] ),
		index = 0,
		length = collection.length;
	for ( ; index < length; index++ ) {
		if ( (tween = collection[ index ].call( animation, prop, value )) ) {

			// we're done with this property
			return tween;
		}
	}
}

function defaultPrefilter( elem, props, opts ) {
	/* jshint validthis: true */
	var prop, value, toggle, tween, hooks, oldfire, display, checkDisplay,
		anim = this,
		orig = {},
		style = elem.style,
		hidden = elem.nodeType && isHidden( elem ),
		dataShow = jQuery._data( elem, "fxshow" );

	// handle queue: false promises
	if ( !opts.queue ) {
		hooks = jQuery._queueHooks( elem, "fx" );
		if ( hooks.unqueued == null ) {
			hooks.unqueued = 0;
			oldfire = hooks.empty.fire;
			hooks.empty.fire = function() {
				if ( !hooks.unqueued ) {
					oldfire();
				}
			};
		}
		hooks.unqueued++;

		anim.always(function() {
			// doing this makes sure that the complete handler will be called
			// before this completes
			anim.always(function() {
				hooks.unqueued--;
				if ( !jQuery.queue( elem, "fx" ).length ) {
					hooks.empty.fire();
				}
			});
		});
	}

	// height/width overflow pass
	if ( elem.nodeType === 1 && ( "height" in props || "width" in props ) ) {
		// Make sure that nothing sneaks out
		// Record all 3 overflow attributes because IE does not
		// change the overflow attribute when overflowX and
		// overflowY are set to the same value
		opts.overflow = [ style.overflow, style.overflowX, style.overflowY ];

		// Set display property to inline-block for height/width
		// animations on inline elements that are having width/height animated
		display = jQuery.css( elem, "display" );

		// Test default display if display is currently "none"
		checkDisplay = display === "none" ?
			jQuery._data( elem, "olddisplay" ) || defaultDisplay( elem.nodeName ) : display;

		if ( checkDisplay === "inline" && jQuery.css( elem, "float" ) === "none" ) {

			// inline-level elements accept inline-block;
			// block-level elements need to be inline with layout
			if ( !support.inlineBlockNeedsLayout || defaultDisplay( elem.nodeName ) === "inline" ) {
				style.display = "inline-block";
			} else {
				style.zoom = 1;
			}
		}
	}

	if ( opts.overflow ) {
		style.overflow = "hidden";
		if ( !support.shrinkWrapBlocks() ) {
			anim.always(function() {
				style.overflow = opts.overflow[ 0 ];
				style.overflowX = opts.overflow[ 1 ];
				style.overflowY = opts.overflow[ 2 ];
			});
		}
	}

	// show/hide pass
	for ( prop in props ) {
		value = props[ prop ];
		if ( rfxtypes.exec( value ) ) {
			delete props[ prop ];
			toggle = toggle || value === "toggle";
			if ( value === ( hidden ? "hide" : "show" ) ) {

				// If there is dataShow left over from a stopped hide or show and we are going to proceed with show, we should pretend to be hidden
				if ( value === "show" && dataShow && dataShow[ prop ] !== undefined ) {
					hidden = true;
				} else {
					continue;
				}
			}
			orig[ prop ] = dataShow && dataShow[ prop ] || jQuery.style( elem, prop );

		// Any non-fx value stops us from restoring the original display value
		} else {
			display = undefined;
		}
	}

	if ( !jQuery.isEmptyObject( orig ) ) {
		if ( dataShow ) {
			if ( "hidden" in dataShow ) {
				hidden = dataShow.hidden;
			}
		} else {
			dataShow = jQuery._data( elem, "fxshow", {} );
		}

		// store state if its toggle - enables .stop().toggle() to "reverse"
		if ( toggle ) {
			dataShow.hidden = !hidden;
		}
		if ( hidden ) {
			jQuery( elem ).show();
		} else {
			anim.done(function() {
				jQuery( elem ).hide();
			});
		}
		anim.done(function() {
			var prop;
			jQuery._removeData( elem, "fxshow" );
			for ( prop in orig ) {
				jQuery.style( elem, prop, orig[ prop ] );
			}
		});
		for ( prop in orig ) {
			tween = createTween( hidden ? dataShow[ prop ] : 0, prop, anim );

			if ( !( prop in dataShow ) ) {
				dataShow[ prop ] = tween.start;
				if ( hidden ) {
					tween.end = tween.start;
					tween.start = prop === "width" || prop === "height" ? 1 : 0;
				}
			}
		}

	// If this is a noop like .hide().hide(), restore an overwritten display value
	} else if ( (display === "none" ? defaultDisplay( elem.nodeName ) : display) === "inline" ) {
		style.display = display;
	}
}

function propFilter( props, specialEasing ) {
	var index, name, easing, value, hooks;

	// camelCase, specialEasing and expand cssHook pass
	for ( index in props ) {
		name = jQuery.camelCase( index );
		easing = specialEasing[ name ];
		value = props[ index ];
		if ( jQuery.isArray( value ) ) {
			easing = value[ 1 ];
			value = props[ index ] = value[ 0 ];
		}

		if ( index !== name ) {
			props[ name ] = value;
			delete props[ index ];
		}

		hooks = jQuery.cssHooks[ name ];
		if ( hooks && "expand" in hooks ) {
			value = hooks.expand( value );
			delete props[ name ];

			// not quite $.extend, this wont overwrite keys already present.
			// also - reusing 'index' from above because we have the correct "name"
			for ( index in value ) {
				if ( !( index in props ) ) {
					props[ index ] = value[ index ];
					specialEasing[ index ] = easing;
				}
			}
		} else {
			specialEasing[ name ] = easing;
		}
	}
}

function Animation( elem, properties, options ) {
	var result,
		stopped,
		index = 0,
		length = animationPrefilters.length,
		deferred = jQuery.Deferred().always( function() {
			// don't match elem in the :animated selector
			delete tick.elem;
		}),
		tick = function() {
			if ( stopped ) {
				return false;
			}
			var currentTime = fxNow || createFxNow(),
				remaining = Math.max( 0, animation.startTime + animation.duration - currentTime ),
				// archaic crash bug won't allow us to use 1 - ( 0.5 || 0 ) (#12497)
				temp = remaining / animation.duration || 0,
				percent = 1 - temp,
				index = 0,
				length = animation.tweens.length;

			for ( ; index < length ; index++ ) {
				animation.tweens[ index ].run( percent );
			}

			deferred.notifyWith( elem, [ animation, percent, remaining ]);

			if ( percent < 1 && length ) {
				return remaining;
			} else {
				deferred.resolveWith( elem, [ animation ] );
				return false;
			}
		},
		animation = deferred.promise({
			elem: elem,
			props: jQuery.extend( {}, properties ),
			opts: jQuery.extend( true, { specialEasing: {} }, options ),
			originalProperties: properties,
			originalOptions: options,
			startTime: fxNow || createFxNow(),
			duration: options.duration,
			tweens: [],
			createTween: function( prop, end ) {
				var tween = jQuery.Tween( elem, animation.opts, prop, end,
						animation.opts.specialEasing[ prop ] || animation.opts.easing );
				animation.tweens.push( tween );
				return tween;
			},
			stop: function( gotoEnd ) {
				var index = 0,
					// if we are going to the end, we want to run all the tweens
					// otherwise we skip this part
					length = gotoEnd ? animation.tweens.length : 0;
				if ( stopped ) {
					return this;
				}
				stopped = true;
				for ( ; index < length ; index++ ) {
					animation.tweens[ index ].run( 1 );
				}

				// resolve when we played the last frame
				// otherwise, reject
				if ( gotoEnd ) {
					deferred.resolveWith( elem, [ animation, gotoEnd ] );
				} else {
					deferred.rejectWith( elem, [ animation, gotoEnd ] );
				}
				return this;
			}
		}),
		props = animation.props;

	propFilter( props, animation.opts.specialEasing );

	for ( ; index < length ; index++ ) {
		result = animationPrefilters[ index ].call( animation, elem, props, animation.opts );
		if ( result ) {
			return result;
		}
	}

	jQuery.map( props, createTween, animation );

	if ( jQuery.isFunction( animation.opts.start ) ) {
		animation.opts.start.call( elem, animation );
	}

	jQuery.fx.timer(
		jQuery.extend( tick, {
			elem: elem,
			anim: animation,
			queue: animation.opts.queue
		})
	);

	// attach callbacks from options
	return animation.progress( animation.opts.progress )
		.done( animation.opts.done, animation.opts.complete )
		.fail( animation.opts.fail )
		.always( animation.opts.always );
}

jQuery.Animation = jQuery.extend( Animation, {
	tweener: function( props, callback ) {
		if ( jQuery.isFunction( props ) ) {
			callback = props;
			props = [ "*" ];
		} else {
			props = props.split(" ");
		}

		var prop,
			index = 0,
			length = props.length;

		for ( ; index < length ; index++ ) {
			prop = props[ index ];
			tweeners[ prop ] = tweeners[ prop ] || [];
			tweeners[ prop ].unshift( callback );
		}
	},

	prefilter: function( callback, prepend ) {
		if ( prepend ) {
			animationPrefilters.unshift( callback );
		} else {
			animationPrefilters.push( callback );
		}
	}
});

jQuery.speed = function( speed, easing, fn ) {
	var opt = speed && typeof speed === "object" ? jQuery.extend( {}, speed ) : {
		complete: fn || !fn && easing ||
			jQuery.isFunction( speed ) && speed,
		duration: speed,
		easing: fn && easing || easing && !jQuery.isFunction( easing ) && easing
	};

	opt.duration = jQuery.fx.off ? 0 : typeof opt.duration === "number" ? opt.duration :
		opt.duration in jQuery.fx.speeds ? jQuery.fx.speeds[ opt.duration ] : jQuery.fx.speeds._default;

	// normalize opt.queue - true/undefined/null -> "fx"
	if ( opt.queue == null || opt.queue === true ) {
		opt.queue = "fx";
	}

	// Queueing
	opt.old = opt.complete;

	opt.complete = function() {
		if ( jQuery.isFunction( opt.old ) ) {
			opt.old.call( this );
		}

		if ( opt.queue ) {
			jQuery.dequeue( this, opt.queue );
		}
	};

	return opt;
};

jQuery.fn.extend({
	fadeTo: function( speed, to, easing, callback ) {

		// show any hidden elements after setting opacity to 0
		return this.filter( isHidden ).css( "opacity", 0 ).show()

			// animate to the value specified
			.end().animate({ opacity: to }, speed, easing, callback );
	},
	animate: function( prop, speed, easing, callback ) {
		var empty = jQuery.isEmptyObject( prop ),
			optall = jQuery.speed( speed, easing, callback ),
			doAnimation = function() {
				// Operate on a copy of prop so per-property easing won't be lost
				var anim = Animation( this, jQuery.extend( {}, prop ), optall );

				// Empty animations, or finishing resolves immediately
				if ( empty || jQuery._data( this, "finish" ) ) {
					anim.stop( true );
				}
			};
			doAnimation.finish = doAnimation;

		return empty || optall.queue === false ?
			this.each( doAnimation ) :
			this.queue( optall.queue, doAnimation );
	},
	stop: function( type, clearQueue, gotoEnd ) {
		var stopQueue = function( hooks ) {
			var stop = hooks.stop;
			delete hooks.stop;
			stop( gotoEnd );
		};

		if ( typeof type !== "string" ) {
			gotoEnd = clearQueue;
			clearQueue = type;
			type = undefined;
		}
		if ( clearQueue && type !== false ) {
			this.queue( type || "fx", [] );
		}

		return this.each(function() {
			var dequeue = true,
				index = type != null && type + "queueHooks",
				timers = jQuery.timers,
				data = jQuery._data( this );

			if ( index ) {
				if ( data[ index ] && data[ index ].stop ) {
					stopQueue( data[ index ] );
				}
			} else {
				for ( index in data ) {
					if ( data[ index ] && data[ index ].stop && rrun.test( index ) ) {
						stopQueue( data[ index ] );
					}
				}
			}

			for ( index = timers.length; index--; ) {
				if ( timers[ index ].elem === this && (type == null || timers[ index ].queue === type) ) {
					timers[ index ].anim.stop( gotoEnd );
					dequeue = false;
					timers.splice( index, 1 );
				}
			}

			// start the next in the queue if the last step wasn't forced
			// timers currently will call their complete callbacks, which will dequeue
			// but only if they were gotoEnd
			if ( dequeue || !gotoEnd ) {
				jQuery.dequeue( this, type );
			}
		});
	},
	finish: function( type ) {
		if ( type !== false ) {
			type = type || "fx";
		}
		return this.each(function() {
			var index,
				data = jQuery._data( this ),
				queue = data[ type + "queue" ],
				hooks = data[ type + "queueHooks" ],
				timers = jQuery.timers,
				length = queue ? queue.length : 0;

			// enable finishing flag on private data
			data.finish = true;

			// empty the queue first
			jQuery.queue( this, type, [] );

			if ( hooks && hooks.stop ) {
				hooks.stop.call( this, true );
			}

			// look for any active animations, and finish them
			for ( index = timers.length; index--; ) {
				if ( timers[ index ].elem === this && timers[ index ].queue === type ) {
					timers[ index ].anim.stop( true );
					timers.splice( index, 1 );
				}
			}

			// look for any animations in the old queue and finish them
			for ( index = 0; index < length; index++ ) {
				if ( queue[ index ] && queue[ index ].finish ) {
					queue[ index ].finish.call( this );
				}
			}

			// turn off finishing flag
			delete data.finish;
		});
	}
});

jQuery.each([ "toggle", "show", "hide" ], function( i, name ) {
	var cssFn = jQuery.fn[ name ];
	jQuery.fn[ name ] = function( speed, easing, callback ) {
		return speed == null || typeof speed === "boolean" ?
			cssFn.apply( this, arguments ) :
			this.animate( genFx( name, true ), speed, easing, callback );
	};
});

// Generate shortcuts for custom animations
jQuery.each({
	slideDown: genFx("show"),
	slideUp: genFx("hide"),
	slideToggle: genFx("toggle"),
	fadeIn: { opacity: "show" },
	fadeOut: { opacity: "hide" },
	fadeToggle: { opacity: "toggle" }
}, function( name, props ) {
	jQuery.fn[ name ] = function( speed, easing, callback ) {
		return this.animate( props, speed, easing, callback );
	};
});

jQuery.timers = [];
jQuery.fx.tick = function() {
	var timer,
		timers = jQuery.timers,
		i = 0;

	fxNow = jQuery.now();

	for ( ; i < timers.length; i++ ) {
		timer = timers[ i ];
		// Checks the timer has not already been removed
		if ( !timer() && timers[ i ] === timer ) {
			timers.splice( i--, 1 );
		}
	}

	if ( !timers.length ) {
		jQuery.fx.stop();
	}
	fxNow = undefined;
};

jQuery.fx.timer = function( timer ) {
	jQuery.timers.push( timer );
	if ( timer() ) {
		jQuery.fx.start();
	} else {
		jQuery.timers.pop();
	}
};

jQuery.fx.interval = 13;

jQuery.fx.start = function() {
	if ( !timerId ) {
		timerId = setInterval( jQuery.fx.tick, jQuery.fx.interval );
	}
};

jQuery.fx.stop = function() {
	clearInterval( timerId );
	timerId = null;
};

jQuery.fx.speeds = {
	slow: 600,
	fast: 200,
	// Default speed
	_default: 400
};


// Based off of the plugin by Clint Helfers, with permission.
// http://blindsignals.com/index.php/2009/07/jquery-delay/
jQuery.fn.delay = function( time, type ) {
	time = jQuery.fx ? jQuery.fx.speeds[ time ] || time : time;
	type = type || "fx";

	return this.queue( type, function( next, hooks ) {
		var timeout = setTimeout( next, time );
		hooks.stop = function() {
			clearTimeout( timeout );
		};
	});
};


(function() {
	// Minified: var a,b,c,d,e
	var input, div, select, a, opt;

	// Setup
	div = document.createElement( "div" );
	div.setAttribute( "className", "t" );
	div.innerHTML = "  <link/><table></table><a href='/a'>a</a><input type='checkbox'/>";
	a = div.getElementsByTagName("a")[ 0 ];

	// First batch of tests.
	select = document.createElement("select");
	opt = select.appendChild( document.createElement("option") );
	input = div.getElementsByTagName("input")[ 0 ];

	a.style.cssText = "top:1px";

	// Test setAttribute on camelCase class. If it works, we need attrFixes when doing get/setAttribute (ie6/7)
	support.getSetAttribute = div.className !== "t";

	// Get the style information from getAttribute
	// (IE uses .cssText instead)
	support.style = /top/.test( a.getAttribute("style") );

	// Make sure that URLs aren't manipulated
	// (IE normalizes it by default)
	support.hrefNormalized = a.getAttribute("href") === "/a";

	// Check the default checkbox/radio value ("" on WebKit; "on" elsewhere)
	support.checkOn = !!input.value;

	// Make sure that a selected-by-default option has a working selected property.
	// (WebKit defaults to false instead of true, IE too, if it's in an optgroup)
	support.optSelected = opt.selected;

	// Tests for enctype support on a form (#6743)
	support.enctype = !!document.createElement("form").enctype;

	// Make sure that the options inside disabled selects aren't marked as disabled
	// (WebKit marks them as disabled)
	select.disabled = true;
	support.optDisabled = !opt.disabled;

	// Support: IE8 only
	// Check if we can trust getAttribute("value")
	input = document.createElement( "input" );
	input.setAttribute( "value", "" );
	support.input = input.getAttribute( "value" ) === "";

	// Check if an input maintains its value after becoming a radio
	input.value = "t";
	input.setAttribute( "type", "radio" );
	support.radioValue = input.value === "t";
})();


var rreturn = /\r/g;

jQuery.fn.extend({
	val: function( value ) {
		var hooks, ret, isFunction,
			elem = this[0];

		if ( !arguments.length ) {
			if ( elem ) {
				hooks = jQuery.valHooks[ elem.type ] || jQuery.valHooks[ elem.nodeName.toLowerCase() ];

				if ( hooks && "get" in hooks && (ret = hooks.get( elem, "value" )) !== undefined ) {
					return ret;
				}

				ret = elem.value;

				return typeof ret === "string" ?
					// handle most common string cases
					ret.replace(rreturn, "") :
					// handle cases where value is null/undef or number
					ret == null ? "" : ret;
			}

			return;
		}

		isFunction = jQuery.isFunction( value );

		return this.each(function( i ) {
			var val;

			if ( this.nodeType !== 1 ) {
				return;
			}

			if ( isFunction ) {
				val = value.call( this, i, jQuery( this ).val() );
			} else {
				val = value;
			}

			// Treat null/undefined as ""; convert numbers to string
			if ( val == null ) {
				val = "";
			} else if ( typeof val === "number" ) {
				val += "";
			} else if ( jQuery.isArray( val ) ) {
				val = jQuery.map( val, function( value ) {
					return value == null ? "" : value + "";
				});
			}

			hooks = jQuery.valHooks[ this.type ] || jQuery.valHooks[ this.nodeName.toLowerCase() ];

			// If set returns undefined, fall back to normal setting
			if ( !hooks || !("set" in hooks) || hooks.set( this, val, "value" ) === undefined ) {
				this.value = val;
			}
		});
	}
});

jQuery.extend({
	valHooks: {
		option: {
			get: function( elem ) {
				var val = jQuery.find.attr( elem, "value" );
				return val != null ?
					val :
					// Support: IE10-11+
					// option.text throws exceptions (#14686, #14858)
					jQuery.trim( jQuery.text( elem ) );
			}
		},
		select: {
			get: function( elem ) {
				var value, option,
					options = elem.options,
					index = elem.selectedIndex,
					one = elem.type === "select-one" || index < 0,
					values = one ? null : [],
					max = one ? index + 1 : options.length,
					i = index < 0 ?
						max :
						one ? index : 0;

				// Loop through all the selected options
				for ( ; i < max; i++ ) {
					option = options[ i ];

					// oldIE doesn't update selected after form reset (#2551)
					if ( ( option.selected || i === index ) &&
							// Don't return options that are disabled or in a disabled optgroup
							( support.optDisabled ? !option.disabled : option.getAttribute("disabled") === null ) &&
							( !option.parentNode.disabled || !jQuery.nodeName( option.parentNode, "optgroup" ) ) ) {

						// Get the specific value for the option
						value = jQuery( option ).val();

						// We don't need an array for one selects
						if ( one ) {
							return value;
						}

						// Multi-Selects return an array
						values.push( value );
					}
				}

				return values;
			},

			set: function( elem, value ) {
				var optionSet, option,
					options = elem.options,
					values = jQuery.makeArray( value ),
					i = options.length;

				while ( i-- ) {
					option = options[ i ];

					if ( jQuery.inArray( jQuery.valHooks.option.get( option ), values ) >= 0 ) {

						// Support: IE6
						// When new option element is added to select box we need to
						// force reflow of newly added node in order to workaround delay
						// of initialization properties
						try {
							option.selected = optionSet = true;

						} catch ( _ ) {

							// Will be executed only in IE6
							option.scrollHeight;
						}

					} else {
						option.selected = false;
					}
				}

				// Force browsers to behave consistently when non-matching value is set
				if ( !optionSet ) {
					elem.selectedIndex = -1;
				}

				return options;
			}
		}
	}
});

// Radios and checkboxes getter/setter
jQuery.each([ "radio", "checkbox" ], function() {
	jQuery.valHooks[ this ] = {
		set: function( elem, value ) {
			if ( jQuery.isArray( value ) ) {
				return ( elem.checked = jQuery.inArray( jQuery(elem).val(), value ) >= 0 );
			}
		}
	};
	if ( !support.checkOn ) {
		jQuery.valHooks[ this ].get = function( elem ) {
			// Support: Webkit
			// "" is returned instead of "on" if a value isn't specified
			return elem.getAttribute("value") === null ? "on" : elem.value;
		};
	}
});




var nodeHook, boolHook,
	attrHandle = jQuery.expr.attrHandle,
	ruseDefault = /^(?:checked|selected)$/i,
	getSetAttribute = support.getSetAttribute,
	getSetInput = support.input;

jQuery.fn.extend({
	attr: function( name, value ) {
		return access( this, jQuery.attr, name, value, arguments.length > 1 );
	},

	removeAttr: function( name ) {
		return this.each(function() {
			jQuery.removeAttr( this, name );
		});
	}
});

jQuery.extend({
	attr: function( elem, name, value ) {
		var hooks, ret,
			nType = elem.nodeType;

		// don't get/set attributes on text, comment and attribute nodes
		if ( !elem || nType === 3 || nType === 8 || nType === 2 ) {
			return;
		}

		// Fallback to prop when attributes are not supported
		if ( typeof elem.getAttribute === strundefined ) {
			return jQuery.prop( elem, name, value );
		}

		// All attributes are lowercase
		// Grab necessary hook if one is defined
		if ( nType !== 1 || !jQuery.isXMLDoc( elem ) ) {
			name = name.toLowerCase();
			hooks = jQuery.attrHooks[ name ] ||
				( jQuery.expr.match.bool.test( name ) ? boolHook : nodeHook );
		}

		if ( value !== undefined ) {

			if ( value === null ) {
				jQuery.removeAttr( elem, name );

			} else if ( hooks && "set" in hooks && (ret = hooks.set( elem, value, name )) !== undefined ) {
				return ret;

			} else {
				elem.setAttribute( name, value + "" );
				return value;
			}

		} else if ( hooks && "get" in hooks && (ret = hooks.get( elem, name )) !== null ) {
			return ret;

		} else {
			ret = jQuery.find.attr( elem, name );

			// Non-existent attributes return null, we normalize to undefined
			return ret == null ?
				undefined :
				ret;
		}
	},

	removeAttr: function( elem, value ) {
		var name, propName,
			i = 0,
			attrNames = value && value.match( rnotwhite );

		if ( attrNames && elem.nodeType === 1 ) {
			while ( (name = attrNames[i++]) ) {
				propName = jQuery.propFix[ name ] || name;

				// Boolean attributes get special treatment (#10870)
				if ( jQuery.expr.match.bool.test( name ) ) {
					// Set corresponding property to false
					if ( getSetInput && getSetAttribute || !ruseDefault.test( name ) ) {
						elem[ propName ] = false;
					// Support: IE<9
					// Also clear defaultChecked/defaultSelected (if appropriate)
					} else {
						elem[ jQuery.camelCase( "default-" + name ) ] =
							elem[ propName ] = false;
					}

				// See #9699 for explanation of this approach (setting first, then removal)
				} else {
					jQuery.attr( elem, name, "" );
				}

				elem.removeAttribute( getSetAttribute ? name : propName );
			}
		}
	},

	attrHooks: {
		type: {
			set: function( elem, value ) {
				if ( !support.radioValue && value === "radio" && jQuery.nodeName(elem, "input") ) {
					// Setting the type on a radio button after the value resets the value in IE6-9
					// Reset value to default in case type is set after value during creation
					var val = elem.value;
					elem.setAttribute( "type", value );
					if ( val ) {
						elem.value = val;
					}
					return value;
				}
			}
		}
	}
});

// Hook for boolean attributes
boolHook = {
	set: function( elem, value, name ) {
		if ( value === false ) {
			// Remove boolean attributes when set to false
			jQuery.removeAttr( elem, name );
		} else if ( getSetInput && getSetAttribute || !ruseDefault.test( name ) ) {
			// IE<8 needs the *property* name
			elem.setAttribute( !getSetAttribute && jQuery.propFix[ name ] || name, name );

		// Use defaultChecked and defaultSelected for oldIE
		} else {
			elem[ jQuery.camelCase( "default-" + name ) ] = elem[ name ] = true;
		}

		return name;
	}
};

// Retrieve booleans specially
jQuery.each( jQuery.expr.match.bool.source.match( /\w+/g ), function( i, name ) {

	var getter = attrHandle[ name ] || jQuery.find.attr;

	attrHandle[ name ] = getSetInput && getSetAttribute || !ruseDefault.test( name ) ?
		function( elem, name, isXML ) {
			var ret, handle;
			if ( !isXML ) {
				// Avoid an infinite loop by temporarily removing this function from the getter
				handle = attrHandle[ name ];
				attrHandle[ name ] = ret;
				ret = getter( elem, name, isXML ) != null ?
					name.toLowerCase() :
					null;
				attrHandle[ name ] = handle;
			}
			return ret;
		} :
		function( elem, name, isXML ) {
			if ( !isXML ) {
				return elem[ jQuery.camelCase( "default-" + name ) ] ?
					name.toLowerCase() :
					null;
			}
		};
});

// fix oldIE attroperties
if ( !getSetInput || !getSetAttribute ) {
	jQuery.attrHooks.value = {
		set: function( elem, value, name ) {
			if ( jQuery.nodeName( elem, "input" ) ) {
				// Does not return so that setAttribute is also used
				elem.defaultValue = value;
			} else {
				// Use nodeHook if defined (#1954); otherwise setAttribute is fine
				return nodeHook && nodeHook.set( elem, value, name );
			}
		}
	};
}

// IE6/7 do not support getting/setting some attributes with get/setAttribute
if ( !getSetAttribute ) {

	// Use this for any attribute in IE6/7
	// This fixes almost every IE6/7 issue
	nodeHook = {
		set: function( elem, value, name ) {
			// Set the existing or create a new attribute node
			var ret = elem.getAttributeNode( name );
			if ( !ret ) {
				elem.setAttributeNode(
					(ret = elem.ownerDocument.createAttribute( name ))
				);
			}

			ret.value = value += "";

			// Break association with cloned elements by also using setAttribute (#9646)
			if ( name === "value" || value === elem.getAttribute( name ) ) {
				return value;
			}
		}
	};

	// Some attributes are constructed with empty-string values when not defined
	attrHandle.id = attrHandle.name = attrHandle.coords =
		function( elem, name, isXML ) {
			var ret;
			if ( !isXML ) {
				return (ret = elem.getAttributeNode( name )) && ret.value !== "" ?
					ret.value :
					null;
			}
		};

	// Fixing value retrieval on a button requires this module
	jQuery.valHooks.button = {
		get: function( elem, name ) {
			var ret = elem.getAttributeNode( name );
			if ( ret && ret.specified ) {
				return ret.value;
			}
		},
		set: nodeHook.set
	};

	// Set contenteditable to false on removals(#10429)
	// Setting to empty string throws an error as an invalid value
	jQuery.attrHooks.contenteditable = {
		set: function( elem, value, name ) {
			nodeHook.set( elem, value === "" ? false : value, name );
		}
	};

	// Set width and height to auto instead of 0 on empty string( Bug #8150 )
	// This is for removals
	jQuery.each([ "width", "height" ], function( i, name ) {
		jQuery.attrHooks[ name ] = {
			set: function( elem, value ) {
				if ( value === "" ) {
					elem.setAttribute( name, "auto" );
					return value;
				}
			}
		};
	});
}

if ( !support.style ) {
	jQuery.attrHooks.style = {
		get: function( elem ) {
			// Return undefined in the case of empty string
			// Note: IE uppercases css property names, but if we were to .toLowerCase()
			// .cssText, that would destroy case senstitivity in URL's, like in "background"
			return elem.style.cssText || undefined;
		},
		set: function( elem, value ) {
			return ( elem.style.cssText = value + "" );
		}
	};
}




var rfocusable = /^(?:input|select|textarea|button|object)$/i,
	rclickable = /^(?:a|area)$/i;

jQuery.fn.extend({
	prop: function( name, value ) {
		return access( this, jQuery.prop, name, value, arguments.length > 1 );
	},

	removeProp: function( name ) {
		name = jQuery.propFix[ name ] || name;
		return this.each(function() {
			// try/catch handles cases where IE balks (such as removing a property on window)
			try {
				this[ name ] = undefined;
				delete this[ name ];
			} catch( e ) {}
		});
	}
});

jQuery.extend({
	propFix: {
		"for": "htmlFor",
		"class": "className"
	},

	prop: function( elem, name, value ) {
		var ret, hooks, notxml,
			nType = elem.nodeType;

		// don't get/set properties on text, comment and attribute nodes
		if ( !elem || nType === 3 || nType === 8 || nType === 2 ) {
			return;
		}

		notxml = nType !== 1 || !jQuery.isXMLDoc( elem );

		if ( notxml ) {
			// Fix name and attach hooks
			name = jQuery.propFix[ name ] || name;
			hooks = jQuery.propHooks[ name ];
		}

		if ( value !== undefined ) {
			return hooks && "set" in hooks && (ret = hooks.set( elem, value, name )) !== undefined ?
				ret :
				( elem[ name ] = value );

		} else {
			return hooks && "get" in hooks && (ret = hooks.get( elem, name )) !== null ?
				ret :
				elem[ name ];
		}
	},

	propHooks: {
		tabIndex: {
			get: function( elem ) {
				// elem.tabIndex doesn't always return the correct value when it hasn't been explicitly set
				// http://fluidproject.org/blog/2008/01/09/getting-setting-and-removing-tabindex-values-with-javascript/
				// Use proper attribute retrieval(#12072)
				var tabindex = jQuery.find.attr( elem, "tabindex" );

				return tabindex ?
					parseInt( tabindex, 10 ) :
					rfocusable.test( elem.nodeName ) || rclickable.test( elem.nodeName ) && elem.href ?
						0 :
						-1;
			}
		}
	}
});

// Some attributes require a special call on IE
// http://msdn.microsoft.com/en-us/library/ms536429%28VS.85%29.aspx
if ( !support.hrefNormalized ) {
	// href/src property should get the full normalized URL (#10299/#12915)
	jQuery.each([ "href", "src" ], function( i, name ) {
		jQuery.propHooks[ name ] = {
			get: function( elem ) {
				return elem.getAttribute( name, 4 );
			}
		};
	});
}

// Support: Safari, IE9+
// mis-reports the default selected property of an option
// Accessing the parent's selectedIndex property fixes it
if ( !support.optSelected ) {
	jQuery.propHooks.selected = {
		get: function( elem ) {
			var parent = elem.parentNode;

			if ( parent ) {
				parent.selectedIndex;

				// Make sure that it also works with optgroups, see #5701
				if ( parent.parentNode ) {
					parent.parentNode.selectedIndex;
				}
			}
			return null;
		}
	};
}

jQuery.each([
	"tabIndex",
	"readOnly",
	"maxLength",
	"cellSpacing",
	"cellPadding",
	"rowSpan",
	"colSpan",
	"useMap",
	"frameBorder",
	"contentEditable"
], function() {
	jQuery.propFix[ this.toLowerCase() ] = this;
});

// IE6/7 call enctype encoding
if ( !support.enctype ) {
	jQuery.propFix.enctype = "encoding";
}




var rclass = /[\t\r\n\f]/g;

jQuery.fn.extend({
	addClass: function( value ) {
		var classes, elem, cur, clazz, j, finalValue,
			i = 0,
			len = this.length,
			proceed = typeof value === "string" && value;

		if ( jQuery.isFunction( value ) ) {
			return this.each(function( j ) {
				jQuery( this ).addClass( value.call( this, j, this.className ) );
			});
		}

		if ( proceed ) {
			// The disjunction here is for better compressibility (see removeClass)
			classes = ( value || "" ).match( rnotwhite ) || [];

			for ( ; i < len; i++ ) {
				elem = this[ i ];
				cur = elem.nodeType === 1 && ( elem.className ?
					( " " + elem.className + " " ).replace( rclass, " " ) :
					" "
				);

				if ( cur ) {
					j = 0;
					while ( (clazz = classes[j++]) ) {
						if ( cur.indexOf( " " + clazz + " " ) < 0 ) {
							cur += clazz + " ";
						}
					}

					// only assign if different to avoid unneeded rendering.
					finalValue = jQuery.trim( cur );
					if ( elem.className !== finalValue ) {
						elem.className = finalValue;
					}
				}
			}
		}

		return this;
	},

	removeClass: function( value ) {
		var classes, elem, cur, clazz, j, finalValue,
			i = 0,
			len = this.length,
			proceed = arguments.length === 0 || typeof value === "string" && value;

		if ( jQuery.isFunction( value ) ) {
			return this.each(function( j ) {
				jQuery( this ).removeClass( value.call( this, j, this.className ) );
			});
		}
		if ( proceed ) {
			classes = ( value || "" ).match( rnotwhite ) || [];

			for ( ; i < len; i++ ) {
				elem = this[ i ];
				// This expression is here for better compressibility (see addClass)
				cur = elem.nodeType === 1 && ( elem.className ?
					( " " + elem.className + " " ).replace( rclass, " " ) :
					""
				);

				if ( cur ) {
					j = 0;
					while ( (clazz = classes[j++]) ) {
						// Remove *all* instances
						while ( cur.indexOf( " " + clazz + " " ) >= 0 ) {
							cur = cur.replace( " " + clazz + " ", " " );
						}
					}

					// only assign if different to avoid unneeded rendering.
					finalValue = value ? jQuery.trim( cur ) : "";
					if ( elem.className !== finalValue ) {
						elem.className = finalValue;
					}
				}
			}
		}

		return this;
	},

	toggleClass: function( value, stateVal ) {
		var type = typeof value;

		if ( typeof stateVal === "boolean" && type === "string" ) {
			return stateVal ? this.addClass( value ) : this.removeClass( value );
		}

		if ( jQuery.isFunction( value ) ) {
			return this.each(function( i ) {
				jQuery( this ).toggleClass( value.call(this, i, this.className, stateVal), stateVal );
			});
		}

		return this.each(function() {
			if ( type === "string" ) {
				// toggle individual class names
				var className,
					i = 0,
					self = jQuery( this ),
					classNames = value.match( rnotwhite ) || [];

				while ( (className = classNames[ i++ ]) ) {
					// check each className given, space separated list
					if ( self.hasClass( className ) ) {
						self.removeClass( className );
					} else {
						self.addClass( className );
					}
				}

			// Toggle whole class name
			} else if ( type === strundefined || type === "boolean" ) {
				if ( this.className ) {
					// store className if set
					jQuery._data( this, "__className__", this.className );
				}

				// If the element has a class name or if we're passed "false",
				// then remove the whole classname (if there was one, the above saved it).
				// Otherwise bring back whatever was previously saved (if anything),
				// falling back to the empty string if nothing was stored.
				this.className = this.className || value === false ? "" : jQuery._data( this, "__className__" ) || "";
			}
		});
	},

	hasClass: function( selector ) {
		var className = " " + selector + " ",
			i = 0,
			l = this.length;
		for ( ; i < l; i++ ) {
			if ( this[i].nodeType === 1 && (" " + this[i].className + " ").replace(rclass, " ").indexOf( className ) >= 0 ) {
				return true;
			}
		}

		return false;
	}
});




// Return jQuery for attributes-only inclusion


jQuery.each( ("blur focus focusin focusout load resize scroll unload click dblclick " +
	"mousedown mouseup mousemove mouseover mouseout mouseenter mouseleave " +
	"change select submit keydown keypress keyup error contextmenu").split(" "), function( i, name ) {

	// Handle event binding
	jQuery.fn[ name ] = function( data, fn ) {
		return arguments.length > 0 ?
			this.on( name, null, data, fn ) :
			this.trigger( name );
	};
});

jQuery.fn.extend({
	hover: function( fnOver, fnOut ) {
		return this.mouseenter( fnOver ).mouseleave( fnOut || fnOver );
	},

	bind: function( types, data, fn ) {
		return this.on( types, null, data, fn );
	},
	unbind: function( types, fn ) {
		return this.off( types, null, fn );
	},

	delegate: function( selector, types, data, fn ) {
		return this.on( types, selector, data, fn );
	},
	undelegate: function( selector, types, fn ) {
		// ( namespace ) or ( selector, types [, fn] )
		return arguments.length === 1 ? this.off( selector, "**" ) : this.off( types, selector || "**", fn );
	}
});


var nonce = jQuery.now();

var rquery = (/\?/);



var rvalidtokens = /(,)|(\[|{)|(}|])|"(?:[^"\\\r\n]|\\["\\\/bfnrt]|\\u[\da-fA-F]{4})*"\s*:?|true|false|null|-?(?!0\d)\d+(?:\.\d+|)(?:[eE][+-]?\d+|)/g;

jQuery.parseJSON = function( data ) {
	// Attempt to parse using the native JSON parser first
	if ( window.JSON && window.JSON.parse ) {
		// Support: Android 2.3
		// Workaround failure to string-cast null input
		return window.JSON.parse( data + "" );
	}

	var requireNonComma,
		depth = null,
		str = jQuery.trim( data + "" );

	// Guard against invalid (and possibly dangerous) input by ensuring that nothing remains
	// after removing valid tokens
	return str && !jQuery.trim( str.replace( rvalidtokens, function( token, comma, open, close ) {

		// Force termination if we see a misplaced comma
		if ( requireNonComma && comma ) {
			depth = 0;
		}

		// Perform no more replacements after returning to outermost depth
		if ( depth === 0 ) {
			return token;
		}

		// Commas must not follow "[", "{", or ","
		requireNonComma = open || comma;

		// Determine new depth
		// array/object open ("[" or "{"): depth += true - false (increment)
		// array/object close ("]" or "}"): depth += false - true (decrement)
		// other cases ("," or primitive): depth += true - true (numeric cast)
		depth += !close - !open;

		// Remove this token
		return "";
	}) ) ?
		( Function( "return " + str ) )() :
		jQuery.error( "Invalid JSON: " + data );
};


// Cross-browser xml parsing
jQuery.parseXML = function( data ) {
	var xml, tmp;
	if ( !data || typeof data !== "string" ) {
		return null;
	}
	try {
		if ( window.DOMParser ) { // Standard
			tmp = new DOMParser();
			xml = tmp.parseFromString( data, "text/xml" );
		} else { // IE
			xml = new ActiveXObject( "Microsoft.XMLDOM" );
			xml.async = "false";
			xml.loadXML( data );
		}
	} catch( e ) {
		xml = undefined;
	}
	if ( !xml || !xml.documentElement || xml.getElementsByTagName( "parsererror" ).length ) {
		jQuery.error( "Invalid XML: " + data );
	}
	return xml;
};


var
	// Document location
	ajaxLocParts,
	ajaxLocation,

	rhash = /#.*$/,
	rts = /([?&])_=[^&]*/,
	rheaders = /^(.*?):[ \t]*([^\r\n]*)\r?$/mg, // IE leaves an \r character at EOL
	// #7653, #8125, #8152: local protocol detection
	rlocalProtocol = /^(?:about|app|app-storage|.+-extension|file|res|widget):$/,
	rnoContent = /^(?:GET|HEAD)$/,
	rprotocol = /^\/\//,
	rurl = /^([\w.+-]+:)(?:\/\/(?:[^\/?#]*@|)([^\/?#:]*)(?::(\d+)|)|)/,

	/* Prefilters
	 * 1) They are useful to introduce custom dataTypes (see ajax/jsonp.js for an example)
	 * 2) These are called:
	 *    - BEFORE asking for a transport
	 *    - AFTER param serialization (s.data is a string if s.processData is true)
	 * 3) key is the dataType
	 * 4) the catchall symbol "*" can be used
	 * 5) execution will start with transport dataType and THEN continue down to "*" if needed
	 */
	prefilters = {},

	/* Transports bindings
	 * 1) key is the dataType
	 * 2) the catchall symbol "*" can be used
	 * 3) selection will start with transport dataType and THEN go to "*" if needed
	 */
	transports = {},

	// Avoid comment-prolog char sequence (#10098); must appease lint and evade compression
	allTypes = "*/".concat("*");

// #8138, IE may throw an exception when accessing
// a field from window.location if document.domain has been set
try {
	ajaxLocation = location.href;
} catch( e ) {
	// Use the href attribute of an A element
	// since IE will modify it given document.location
	ajaxLocation = document.createElement( "a" );
	ajaxLocation.href = "";
	ajaxLocation = ajaxLocation.href;
}

// Segment location into parts
ajaxLocParts = rurl.exec( ajaxLocation.toLowerCase() ) || [];

// Base "constructor" for jQuery.ajaxPrefilter and jQuery.ajaxTransport
function addToPrefiltersOrTransports( structure ) {

	// dataTypeExpression is optional and defaults to "*"
	return function( dataTypeExpression, func ) {

		if ( typeof dataTypeExpression !== "string" ) {
			func = dataTypeExpression;
			dataTypeExpression = "*";
		}

		var dataType,
			i = 0,
			dataTypes = dataTypeExpression.toLowerCase().match( rnotwhite ) || [];

		if ( jQuery.isFunction( func ) ) {
			// For each dataType in the dataTypeExpression
			while ( (dataType = dataTypes[i++]) ) {
				// Prepend if requested
				if ( dataType.charAt( 0 ) === "+" ) {
					dataType = dataType.slice( 1 ) || "*";
					(structure[ dataType ] = structure[ dataType ] || []).unshift( func );

				// Otherwise append
				} else {
					(structure[ dataType ] = structure[ dataType ] || []).push( func );
				}
			}
		}
	};
}

// Base inspection function for prefilters and transports
function inspectPrefiltersOrTransports( structure, options, originalOptions, jqXHR ) {

	var inspected = {},
		seekingTransport = ( structure === transports );

	function inspect( dataType ) {
		var selected;
		inspected[ dataType ] = true;
		jQuery.each( structure[ dataType ] || [], function( _, prefilterOrFactory ) {
			var dataTypeOrTransport = prefilterOrFactory( options, originalOptions, jqXHR );
			if ( typeof dataTypeOrTransport === "string" && !seekingTransport && !inspected[ dataTypeOrTransport ] ) {
				options.dataTypes.unshift( dataTypeOrTransport );
				inspect( dataTypeOrTransport );
				return false;
			} else if ( seekingTransport ) {
				return !( selected = dataTypeOrTransport );
			}
		});
		return selected;
	}

	return inspect( options.dataTypes[ 0 ] ) || !inspected[ "*" ] && inspect( "*" );
}

// A special extend for ajax options
// that takes "flat" options (not to be deep extended)
// Fixes #9887
function ajaxExtend( target, src ) {
	var deep, key,
		flatOptions = jQuery.ajaxSettings.flatOptions || {};

	for ( key in src ) {
		if ( src[ key ] !== undefined ) {
			( flatOptions[ key ] ? target : ( deep || (deep = {}) ) )[ key ] = src[ key ];
		}
	}
	if ( deep ) {
		jQuery.extend( true, target, deep );
	}

	return target;
}

/* Handles responses to an ajax request:
 * - finds the right dataType (mediates between content-type and expected dataType)
 * - returns the corresponding response
 */
function ajaxHandleResponses( s, jqXHR, responses ) {
	var firstDataType, ct, finalDataType, type,
		contents = s.contents,
		dataTypes = s.dataTypes;

	// Remove auto dataType and get content-type in the process
	while ( dataTypes[ 0 ] === "*" ) {
		dataTypes.shift();
		if ( ct === undefined ) {
			ct = s.mimeType || jqXHR.getResponseHeader("Content-Type");
		}
	}

	// Check if we're dealing with a known content-type
	if ( ct ) {
		for ( type in contents ) {
			if ( contents[ type ] && contents[ type ].test( ct ) ) {
				dataTypes.unshift( type );
				break;
			}
		}
	}

	// Check to see if we have a response for the expected dataType
	if ( dataTypes[ 0 ] in responses ) {
		finalDataType = dataTypes[ 0 ];
	} else {
		// Try convertible dataTypes
		for ( type in responses ) {
			if ( !dataTypes[ 0 ] || s.converters[ type + " " + dataTypes[0] ] ) {
				finalDataType = type;
				break;
			}
			if ( !firstDataType ) {
				firstDataType = type;
			}
		}
		// Or just use first one
		finalDataType = finalDataType || firstDataType;
	}

	// If we found a dataType
	// We add the dataType to the list if needed
	// and return the corresponding response
	if ( finalDataType ) {
		if ( finalDataType !== dataTypes[ 0 ] ) {
			dataTypes.unshift( finalDataType );
		}
		return responses[ finalDataType ];
	}
}

/* Chain conversions given the request and the original response
 * Also sets the responseXXX fields on the jqXHR instance
 */
function ajaxConvert( s, response, jqXHR, isSuccess ) {
	var conv2, current, conv, tmp, prev,
		converters = {},
		// Work with a copy of dataTypes in case we need to modify it for conversion
		dataTypes = s.dataTypes.slice();

	// Create converters map with lowercased keys
	if ( dataTypes[ 1 ] ) {
		for ( conv in s.converters ) {
			converters[ conv.toLowerCase() ] = s.converters[ conv ];
		}
	}

	current = dataTypes.shift();

	// Convert to each sequential dataType
	while ( current ) {

		if ( s.responseFields[ current ] ) {
			jqXHR[ s.responseFields[ current ] ] = response;
		}

		// Apply the dataFilter if provided
		if ( !prev && isSuccess && s.dataFilter ) {
			response = s.dataFilter( response, s.dataType );
		}

		prev = current;
		current = dataTypes.shift();

		if ( current ) {

			// There's only work to do if current dataType is non-auto
			if ( current === "*" ) {

				current = prev;

			// Convert response if prev dataType is non-auto and differs from current
			} else if ( prev !== "*" && prev !== current ) {

				// Seek a direct converter
				conv = converters[ prev + " " + current ] || converters[ "* " + current ];

				// If none found, seek a pair
				if ( !conv ) {
					for ( conv2 in converters ) {

						// If conv2 outputs current
						tmp = conv2.split( " " );
						if ( tmp[ 1 ] === current ) {

							// If prev can be converted to accepted input
							conv = converters[ prev + " " + tmp[ 0 ] ] ||
								converters[ "* " + tmp[ 0 ] ];
							if ( conv ) {
								// Condense equivalence converters
								if ( conv === true ) {
									conv = converters[ conv2 ];

								// Otherwise, insert the intermediate dataType
								} else if ( converters[ conv2 ] !== true ) {
									current = tmp[ 0 ];
									dataTypes.unshift( tmp[ 1 ] );
								}
								break;
							}
						}
					}
				}

				// Apply converter (if not an equivalence)
				if ( conv !== true ) {

					// Unless errors are allowed to bubble, catch and return them
					if ( conv && s[ "throws" ] ) {
						response = conv( response );
					} else {
						try {
							response = conv( response );
						} catch ( e ) {
							return { state: "parsererror", error: conv ? e : "No conversion from " + prev + " to " + current };
						}
					}
				}
			}
		}
	}

	return { state: "success", data: response };
}

jQuery.extend({

	// Counter for holding the number of active queries
	active: 0,

	// Last-Modified header cache for next request
	lastModified: {},
	etag: {},

	ajaxSettings: {
		url: ajaxLocation,
		type: "GET",
		isLocal: rlocalProtocol.test( ajaxLocParts[ 1 ] ),
		global: true,
		processData: true,
		async: true,
		contentType: "application/x-www-form-urlencoded; charset=UTF-8",
		/*
		timeout: 0,
		data: null,
		dataType: null,
		username: null,
		password: null,
		cache: null,
		throws: false,
		traditional: false,
		headers: {},
		*/

		accepts: {
			"*": allTypes,
			text: "text/plain",
			html: "text/html",
			xml: "application/xml, text/xml",
			json: "application/json, text/javascript"
		},

		contents: {
			xml: /xml/,
			html: /html/,
			json: /json/
		},

		responseFields: {
			xml: "responseXML",
			text: "responseText",
			json: "responseJSON"
		},

		// Data converters
		// Keys separate source (or catchall "*") and destination types with a single space
		converters: {

			// Convert anything to text
			"* text": String,

			// Text to html (true = no transformation)
			"text html": true,

			// Evaluate text as a json expression
			"text json": jQuery.parseJSON,

			// Parse text as xml
			"text xml": jQuery.parseXML
		},

		// For options that shouldn't be deep extended:
		// you can add your own custom options here if
		// and when you create one that shouldn't be
		// deep extended (see ajaxExtend)
		flatOptions: {
			url: true,
			context: true
		}
	},

	// Creates a full fledged settings object into target
	// with both ajaxSettings and settings fields.
	// If target is omitted, writes into ajaxSettings.
	ajaxSetup: function( target, settings ) {
		return settings ?

			// Building a settings object
			ajaxExtend( ajaxExtend( target, jQuery.ajaxSettings ), settings ) :

			// Extending ajaxSettings
			ajaxExtend( jQuery.ajaxSettings, target );
	},

	ajaxPrefilter: addToPrefiltersOrTransports( prefilters ),
	ajaxTransport: addToPrefiltersOrTransports( transports ),

	// Main method
	ajax: function( url, options ) {

		// If url is an object, simulate pre-1.5 signature
		if ( typeof url === "object" ) {
			options = url;
			url = undefined;
		}

		// Force options to be an object
		options = options || {};

		var // Cross-domain detection vars
			parts,
			// Loop variable
			i,
			// URL without anti-cache param
			cacheURL,
			// Response headers as string
			responseHeadersString,
			// timeout handle
			timeoutTimer,

			// To know if global events are to be dispatched
			fireGlobals,

			transport,
			// Response headers
			responseHeaders,
			// Create the final options object
			s = jQuery.ajaxSetup( {}, options ),
			// Callbacks context
			callbackContext = s.context || s,
			// Context for global events is callbackContext if it is a DOM node or jQuery collection
			globalEventContext = s.context && ( callbackContext.nodeType || callbackContext.jquery ) ?
				jQuery( callbackContext ) :
				jQuery.event,
			// Deferreds
			deferred = jQuery.Deferred(),
			completeDeferred = jQuery.Callbacks("once memory"),
			// Status-dependent callbacks
			statusCode = s.statusCode || {},
			// Headers (they are sent all at once)
			requestHeaders = {},
			requestHeadersNames = {},
			// The jqXHR state
			state = 0,
			// Default abort message
			strAbort = "canceled",
			// Fake xhr
			jqXHR = {
				readyState: 0,

				// Builds headers hashtable if needed
				getResponseHeader: function( key ) {
					var match;
					if ( state === 2 ) {
						if ( !responseHeaders ) {
							responseHeaders = {};
							while ( (match = rheaders.exec( responseHeadersString )) ) {
								responseHeaders[ match[1].toLowerCase() ] = match[ 2 ];
							}
						}
						match = responseHeaders[ key.toLowerCase() ];
					}
					return match == null ? null : match;
				},

				// Raw string
				getAllResponseHeaders: function() {
					return state === 2 ? responseHeadersString : null;
				},

				// Caches the header
				setRequestHeader: function( name, value ) {
					var lname = name.toLowerCase();
					if ( !state ) {
						name = requestHeadersNames[ lname ] = requestHeadersNames[ lname ] || name;
						requestHeaders[ name ] = value;
					}
					return this;
				},

				// Overrides response content-type header
				overrideMimeType: function( type ) {
					if ( !state ) {
						s.mimeType = type;
					}
					return this;
				},

				// Status-dependent callbacks
				statusCode: function( map ) {
					var code;
					if ( map ) {
						if ( state < 2 ) {
							for ( code in map ) {
								// Lazy-add the new callback in a way that preserves old ones
								statusCode[ code ] = [ statusCode[ code ], map[ code ] ];
							}
						} else {
							// Execute the appropriate callbacks
							jqXHR.always( map[ jqXHR.status ] );
						}
					}
					return this;
				},

				// Cancel the request
				abort: function( statusText ) {
					var finalText = statusText || strAbort;
					if ( transport ) {
						transport.abort( finalText );
					}
					done( 0, finalText );
					return this;
				}
			};

		// Attach deferreds
		deferred.promise( jqXHR ).complete = completeDeferred.add;
		jqXHR.success = jqXHR.done;
		jqXHR.error = jqXHR.fail;

		// Remove hash character (#7531: and string promotion)
		// Add protocol if not provided (#5866: IE7 issue with protocol-less urls)
		// Handle falsy url in the settings object (#10093: consistency with old signature)
		// We also use the url parameter if available
		s.url = ( ( url || s.url || ajaxLocation ) + "" ).replace( rhash, "" ).replace( rprotocol, ajaxLocParts[ 1 ] + "//" );

		// Alias method option to type as per ticket #12004
		s.type = options.method || options.type || s.method || s.type;

		// Extract dataTypes list
		s.dataTypes = jQuery.trim( s.dataType || "*" ).toLowerCase().match( rnotwhite ) || [ "" ];

		// A cross-domain request is in order when we have a protocol:host:port mismatch
		if ( s.crossDomain == null ) {
			parts = rurl.exec( s.url.toLowerCase() );
			s.crossDomain = !!( parts &&
				( parts[ 1 ] !== ajaxLocParts[ 1 ] || parts[ 2 ] !== ajaxLocParts[ 2 ] ||
					( parts[ 3 ] || ( parts[ 1 ] === "http:" ? "80" : "443" ) ) !==
						( ajaxLocParts[ 3 ] || ( ajaxLocParts[ 1 ] === "http:" ? "80" : "443" ) ) )
			);
		}

		// Convert data if not already a string
		if ( s.data && s.processData && typeof s.data !== "string" ) {
			s.data = jQuery.param( s.data, s.traditional );
		}

		// Apply prefilters
		inspectPrefiltersOrTransports( prefilters, s, options, jqXHR );

		// If request was aborted inside a prefilter, stop there
		if ( state === 2 ) {
			return jqXHR;
		}

		// We can fire global events as of now if asked to
		// Don't fire events if jQuery.event is undefined in an AMD-usage scenario (#15118)
		fireGlobals = jQuery.event && s.global;

		// Watch for a new set of requests
		if ( fireGlobals && jQuery.active++ === 0 ) {
			jQuery.event.trigger("ajaxStart");
		}

		// Uppercase the type
		s.type = s.type.toUpperCase();

		// Determine if request has content
		s.hasContent = !rnoContent.test( s.type );

		// Save the URL in case we're toying with the If-Modified-Since
		// and/or If-None-Match header later on
		cacheURL = s.url;

		// More options handling for requests with no content
		if ( !s.hasContent ) {

			// If data is available, append data to url
			if ( s.data ) {
				cacheURL = ( s.url += ( rquery.test( cacheURL ) ? "&" : "?" ) + s.data );
				// #9682: remove data so that it's not used in an eventual retry
				delete s.data;
			}

			// Add anti-cache in url if needed
			if ( s.cache === false ) {
				s.url = rts.test( cacheURL ) ?

					// If there is already a '_' parameter, set its value
					cacheURL.replace( rts, "$1_=" + nonce++ ) :

					// Otherwise add one to the end
					cacheURL + ( rquery.test( cacheURL ) ? "&" : "?" ) + "_=" + nonce++;
			}
		}

		// Set the If-Modified-Since and/or If-None-Match header, if in ifModified mode.
		if ( s.ifModified ) {
			if ( jQuery.lastModified[ cacheURL ] ) {
				jqXHR.setRequestHeader( "If-Modified-Since", jQuery.lastModified[ cacheURL ] );
			}
			if ( jQuery.etag[ cacheURL ] ) {
				jqXHR.setRequestHeader( "If-None-Match", jQuery.etag[ cacheURL ] );
			}
		}

		// Set the correct header, if data is being sent
		if ( s.data && s.hasContent && s.contentType !== false || options.contentType ) {
			jqXHR.setRequestHeader( "Content-Type", s.contentType );
		}

		// Set the Accepts header for the server, depending on the dataType
		jqXHR.setRequestHeader(
			"Accept",
			s.dataTypes[ 0 ] && s.accepts[ s.dataTypes[0] ] ?
				s.accepts[ s.dataTypes[0] ] + ( s.dataTypes[ 0 ] !== "*" ? ", " + allTypes + "; q=0.01" : "" ) :
				s.accepts[ "*" ]
		);

		// Check for headers option
		for ( i in s.headers ) {
			jqXHR.setRequestHeader( i, s.headers[ i ] );
		}

		// Allow custom headers/mimetypes and early abort
		if ( s.beforeSend && ( s.beforeSend.call( callbackContext, jqXHR, s ) === false || state === 2 ) ) {
			// Abort if not done already and return
			return jqXHR.abort();
		}

		// aborting is no longer a cancellation
		strAbort = "abort";

		// Install callbacks on deferreds
		for ( i in { success: 1, error: 1, complete: 1 } ) {
			jqXHR[ i ]( s[ i ] );
		}

		// Get transport
		transport = inspectPrefiltersOrTransports( transports, s, options, jqXHR );

		// If no transport, we auto-abort
		if ( !transport ) {
			done( -1, "No Transport" );
		} else {
			jqXHR.readyState = 1;

			// Send global event
			if ( fireGlobals ) {
				globalEventContext.trigger( "ajaxSend", [ jqXHR, s ] );
			}
			// Timeout
			if ( s.async && s.timeout > 0 ) {
				timeoutTimer = setTimeout(function() {
					jqXHR.abort("timeout");
				}, s.timeout );
			}

			try {
				state = 1;
				transport.send( requestHeaders, done );
			} catch ( e ) {
				// Propagate exception as error if not done
				if ( state < 2 ) {
					done( -1, e );
				// Simply rethrow otherwise
				} else {
					throw e;
				}
			}
		}

		// Callback for when everything is done
		function done( status, nativeStatusText, responses, headers ) {
			var isSuccess, success, error, response, modified,
				statusText = nativeStatusText;

			// Called once
			if ( state === 2 ) {
				return;
			}

			// State is "done" now
			state = 2;

			// Clear timeout if it exists
			if ( timeoutTimer ) {
				clearTimeout( timeoutTimer );
			}

			// Dereference transport for early garbage collection
			// (no matter how long the jqXHR object will be used)
			transport = undefined;

			// Cache response headers
			responseHeadersString = headers || "";

			// Set readyState
			jqXHR.readyState = status > 0 ? 4 : 0;

			// Determine if successful
			isSuccess = status >= 200 && status < 300 || status === 304;

			// Get response data
			if ( responses ) {
				response = ajaxHandleResponses( s, jqXHR, responses );
			}

			// Convert no matter what (that way responseXXX fields are always set)
			response = ajaxConvert( s, response, jqXHR, isSuccess );

			// If successful, handle type chaining
			if ( isSuccess ) {

				// Set the If-Modified-Since and/or If-None-Match header, if in ifModified mode.
				if ( s.ifModified ) {
					modified = jqXHR.getResponseHeader("Last-Modified");
					if ( modified ) {
						jQuery.lastModified[ cacheURL ] = modified;
					}
					modified = jqXHR.getResponseHeader("etag");
					if ( modified ) {
						jQuery.etag[ cacheURL ] = modified;
					}
				}

				// if no content
				if ( status === 204 || s.type === "HEAD" ) {
					statusText = "nocontent";

				// if not modified
				} else if ( status === 304 ) {
					statusText = "notmodified";

				// If we have data, let's convert it
				} else {
					statusText = response.state;
					success = response.data;
					error = response.error;
					isSuccess = !error;
				}
			} else {
				// We extract error from statusText
				// then normalize statusText and status for non-aborts
				error = statusText;
				if ( status || !statusText ) {
					statusText = "error";
					if ( status < 0 ) {
						status = 0;
					}
				}
			}

			// Set data for the fake xhr object
			jqXHR.status = status;
			jqXHR.statusText = ( nativeStatusText || statusText ) + "";

			// Success/Error
			if ( isSuccess ) {
				deferred.resolveWith( callbackContext, [ success, statusText, jqXHR ] );
			} else {
				deferred.rejectWith( callbackContext, [ jqXHR, statusText, error ] );
			}

			// Status-dependent callbacks
			jqXHR.statusCode( statusCode );
			statusCode = undefined;

			if ( fireGlobals ) {
				globalEventContext.trigger( isSuccess ? "ajaxSuccess" : "ajaxError",
					[ jqXHR, s, isSuccess ? success : error ] );
			}

			// Complete
			completeDeferred.fireWith( callbackContext, [ jqXHR, statusText ] );

			if ( fireGlobals ) {
				globalEventContext.trigger( "ajaxComplete", [ jqXHR, s ] );
				// Handle the global AJAX counter
				if ( !( --jQuery.active ) ) {
					jQuery.event.trigger("ajaxStop");
				}
			}
		}

		return jqXHR;
	},

	getJSON: function( url, data, callback ) {
		return jQuery.get( url, data, callback, "json" );
	},

	getScript: function( url, callback ) {
		return jQuery.get( url, undefined, callback, "script" );
	}
});

jQuery.each( [ "get", "post" ], function( i, method ) {
	jQuery[ method ] = function( url, data, callback, type ) {
		// shift arguments if data argument was omitted
		if ( jQuery.isFunction( data ) ) {
			type = type || callback;
			callback = data;
			data = undefined;
		}

		return jQuery.ajax({
			url: url,
			type: method,
			dataType: type,
			data: data,
			success: callback
		});
	};
});


jQuery._evalUrl = function( url ) {
	return jQuery.ajax({
		url: url,
		type: "GET",
		dataType: "script",
		async: false,
		global: false,
		"throws": true
	});
};


jQuery.fn.extend({
	wrapAll: function( html ) {
		if ( jQuery.isFunction( html ) ) {
			return this.each(function(i) {
				jQuery(this).wrapAll( html.call(this, i) );
			});
		}

		if ( this[0] ) {
			// The elements to wrap the target around
			var wrap = jQuery( html, this[0].ownerDocument ).eq(0).clone(true);

			if ( this[0].parentNode ) {
				wrap.insertBefore( this[0] );
			}

			wrap.map(function() {
				var elem = this;

				while ( elem.firstChild && elem.firstChild.nodeType === 1 ) {
					elem = elem.firstChild;
				}

				return elem;
			}).append( this );
		}

		return this;
	},

	wrapInner: function( html ) {
		if ( jQuery.isFunction( html ) ) {
			return this.each(function(i) {
				jQuery(this).wrapInner( html.call(this, i) );
			});
		}

		return this.each(function() {
			var self = jQuery( this ),
				contents = self.contents();

			if ( contents.length ) {
				contents.wrapAll( html );

			} else {
				self.append( html );
			}
		});
	},

	wrap: function( html ) {
		var isFunction = jQuery.isFunction( html );

		return this.each(function(i) {
			jQuery( this ).wrapAll( isFunction ? html.call(this, i) : html );
		});
	},

	unwrap: function() {
		return this.parent().each(function() {
			if ( !jQuery.nodeName( this, "body" ) ) {
				jQuery( this ).replaceWith( this.childNodes );
			}
		}).end();
	}
});


jQuery.expr.filters.hidden = function( elem ) {
	// Support: Opera <= 12.12
	// Opera reports offsetWidths and offsetHeights less than zero on some elements
	return elem.offsetWidth <= 0 && elem.offsetHeight <= 0 ||
		(!support.reliableHiddenOffsets() &&
			((elem.style && elem.style.display) || jQuery.css( elem, "display" )) === "none");
};

jQuery.expr.filters.visible = function( elem ) {
	return !jQuery.expr.filters.hidden( elem );
};




var r20 = /%20/g,
	rbracket = /\[\]$/,
	rCRLF = /\r?\n/g,
	rsubmitterTypes = /^(?:submit|button|image|reset|file)$/i,
	rsubmittable = /^(?:input|select|textarea|keygen)/i;

function buildParams( prefix, obj, traditional, add ) {
	var name;

	if ( jQuery.isArray( obj ) ) {
		// Serialize array item.
		jQuery.each( obj, function( i, v ) {
			if ( traditional || rbracket.test( prefix ) ) {
				// Treat each array item as a scalar.
				add( prefix, v );

			} else {
				// Item is non-scalar (array or object), encode its numeric index.
				buildParams( prefix + "[" + ( typeof v === "object" ? i : "" ) + "]", v, traditional, add );
			}
		});

	} else if ( !traditional && jQuery.type( obj ) === "object" ) {
		// Serialize object item.
		for ( name in obj ) {
			buildParams( prefix + "[" + name + "]", obj[ name ], traditional, add );
		}

	} else {
		// Serialize scalar item.
		add( prefix, obj );
	}
}

// Serialize an array of form elements or a set of
// key/values into a query string
jQuery.param = function( a, traditional ) {
	var prefix,
		s = [],
		add = function( key, value ) {
			// If value is a function, invoke it and return its value
			value = jQuery.isFunction( value ) ? value() : ( value == null ? "" : value );
			s[ s.length ] = encodeURIComponent( key ) + "=" + encodeURIComponent( value );
		};

	// Set traditional to true for jQuery <= 1.3.2 behavior.
	if ( traditional === undefined ) {
		traditional = jQuery.ajaxSettings && jQuery.ajaxSettings.traditional;
	}

	// If an array was passed in, assume that it is an array of form elements.
	if ( jQuery.isArray( a ) || ( a.jquery && !jQuery.isPlainObject( a ) ) ) {
		// Serialize the form elements
		jQuery.each( a, function() {
			add( this.name, this.value );
		});

	} else {
		// If traditional, encode the "old" way (the way 1.3.2 or older
		// did it), otherwise encode params recursively.
		for ( prefix in a ) {
			buildParams( prefix, a[ prefix ], traditional, add );
		}
	}

	// Return the resulting serialization
	return s.join( "&" ).replace( r20, "+" );
};

jQuery.fn.extend({
	serialize: function() {
		return jQuery.param( this.serializeArray() );
	},
	serializeArray: function() {
		return this.map(function() {
			// Can add propHook for "elements" to filter or add form elements
			var elements = jQuery.prop( this, "elements" );
			return elements ? jQuery.makeArray( elements ) : this;
		})
		.filter(function() {
			var type = this.type;
			// Use .is(":disabled") so that fieldset[disabled] works
			return this.name && !jQuery( this ).is( ":disabled" ) &&
				rsubmittable.test( this.nodeName ) && !rsubmitterTypes.test( type ) &&
				( this.checked || !rcheckableType.test( type ) );
		})
		.map(function( i, elem ) {
			var val = jQuery( this ).val();

			return val == null ?
				null :
				jQuery.isArray( val ) ?
					jQuery.map( val, function( val ) {
						return { name: elem.name, value: val.replace( rCRLF, "\r\n" ) };
					}) :
					{ name: elem.name, value: val.replace( rCRLF, "\r\n" ) };
		}).get();
	}
});


// Create the request object
// (This is still attached to ajaxSettings for backward compatibility)
jQuery.ajaxSettings.xhr = window.ActiveXObject !== undefined ?
	// Support: IE6+
	function() {

		// XHR cannot access local files, always use ActiveX for that case
		return !this.isLocal &&

			// Support: IE7-8
			// oldIE XHR does not support non-RFC2616 methods (#13240)
			// See http://msdn.microsoft.com/en-us/library/ie/ms536648(v=vs.85).aspx
			// and http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html#sec9
			// Although this check for six methods instead of eight
			// since IE also does not support "trace" and "connect"
			/^(get|post|head|put|delete|options)$/i.test( this.type ) &&

			createStandardXHR() || createActiveXHR();
	} :
	// For all other browsers, use the standard XMLHttpRequest object
	createStandardXHR;

var xhrId = 0,
	xhrCallbacks = {},
	xhrSupported = jQuery.ajaxSettings.xhr();

// Support: IE<10
// Open requests must be manually aborted on unload (#5280)
// See https://support.microsoft.com/kb/2856746 for more info
if ( window.attachEvent ) {
	window.attachEvent( "onunload", function() {
		for ( var key in xhrCallbacks ) {
			xhrCallbacks[ key ]( undefined, true );
		}
	});
}

// Determine support properties
support.cors = !!xhrSupported && ( "withCredentials" in xhrSupported );
xhrSupported = support.ajax = !!xhrSupported;

// Create transport if the browser can provide an xhr
if ( xhrSupported ) {

	jQuery.ajaxTransport(function( options ) {
		// Cross domain only allowed if supported through XMLHttpRequest
		if ( !options.crossDomain || support.cors ) {

			var callback;

			return {
				send: function( headers, complete ) {
					var i,
						xhr = options.xhr(),
						id = ++xhrId;

					// Open the socket
					xhr.open( options.type, options.url, options.async, options.username, options.password );

					// Apply custom fields if provided
					if ( options.xhrFields ) {
						for ( i in options.xhrFields ) {
							xhr[ i ] = options.xhrFields[ i ];
						}
					}

					// Override mime type if needed
					if ( options.mimeType && xhr.overrideMimeType ) {
						xhr.overrideMimeType( options.mimeType );
					}

					// X-Requested-With header
					// For cross-domain requests, seeing as conditions for a preflight are
					// akin to a jigsaw puzzle, we simply never set it to be sure.
					// (it can always be set on a per-request basis or even using ajaxSetup)
					// For same-domain requests, won't change header if already provided.
					if ( !options.crossDomain && !headers["X-Requested-With"] ) {
						headers["X-Requested-With"] = "XMLHttpRequest";
					}

					// Set headers
					for ( i in headers ) {
						// Support: IE<9
						// IE's ActiveXObject throws a 'Type Mismatch' exception when setting
						// request header to a null-value.
						//
						// To keep consistent with other XHR implementations, cast the value
						// to string and ignore `undefined`.
						if ( headers[ i ] !== undefined ) {
							xhr.setRequestHeader( i, headers[ i ] + "" );
						}
					}

					// Do send the request
					// This may raise an exception which is actually
					// handled in jQuery.ajax (so no try/catch here)
					xhr.send( ( options.hasContent && options.data ) || null );

					// Listener
					callback = function( _, isAbort ) {
						var status, statusText, responses;

						// Was never called and is aborted or complete
						if ( callback && ( isAbort || xhr.readyState === 4 ) ) {
							// Clean up
							delete xhrCallbacks[ id ];
							callback = undefined;
							xhr.onreadystatechange = jQuery.noop;

							// Abort manually if needed
							if ( isAbort ) {
								if ( xhr.readyState !== 4 ) {
									xhr.abort();
								}
							} else {
								responses = {};
								status = xhr.status;

								// Support: IE<10
								// Accessing binary-data responseText throws an exception
								// (#11426)
								if ( typeof xhr.responseText === "string" ) {
									responses.text = xhr.responseText;
								}

								// Firefox throws an exception when accessing
								// statusText for faulty cross-domain requests
								try {
									statusText = xhr.statusText;
								} catch( e ) {
									// We normalize with Webkit giving an empty statusText
									statusText = "";
								}

								// Filter status for non standard behaviors

								// If the request is local and we have data: assume a success
								// (success with no data won't get notified, that's the best we
								// can do given current implementations)
								if ( !status && options.isLocal && !options.crossDomain ) {
									status = responses.text ? 200 : 404;
								// IE - #1450: sometimes returns 1223 when it should be 204
								} else if ( status === 1223 ) {
									status = 204;
								}
							}
						}

						// Call complete if needed
						if ( responses ) {
							complete( status, statusText, responses, xhr.getAllResponseHeaders() );
						}
					};

					if ( !options.async ) {
						// if we're in sync mode we fire the callback
						callback();
					} else if ( xhr.readyState === 4 ) {
						// (IE6 & IE7) if it's in cache and has been
						// retrieved directly we need to fire the callback
						setTimeout( callback );
					} else {
						// Add to the list of active xhr callbacks
						xhr.onreadystatechange = xhrCallbacks[ id ] = callback;
					}
				},

				abort: function() {
					if ( callback ) {
						callback( undefined, true );
					}
				}
			};
		}
	});
}

// Functions to create xhrs
function createStandardXHR() {
	try {
		return new window.XMLHttpRequest();
	} catch( e ) {}
}

function createActiveXHR() {
	try {
		return new window.ActiveXObject( "Microsoft.XMLHTTP" );
	} catch( e ) {}
}




// Install script dataType
jQuery.ajaxSetup({
	accepts: {
		script: "text/javascript, application/javascript, application/ecmascript, application/x-ecmascript"
	},
	contents: {
		script: /(?:java|ecma)script/
	},
	converters: {
		"text script": function( text ) {
			jQuery.globalEval( text );
			return text;
		}
	}
});

// Handle cache's special case and global
jQuery.ajaxPrefilter( "script", function( s ) {
	if ( s.cache === undefined ) {
		s.cache = false;
	}
	if ( s.crossDomain ) {
		s.type = "GET";
		s.global = false;
	}
});

// Bind script tag hack transport
jQuery.ajaxTransport( "script", function(s) {

	// This transport only deals with cross domain requests
	if ( s.crossDomain ) {

		var script,
			head = document.head || jQuery("head")[0] || document.documentElement;

		return {

			send: function( _, callback ) {

				script = document.createElement("script");

				script.async = true;

				if ( s.scriptCharset ) {
					script.charset = s.scriptCharset;
				}

				script.src = s.url;

				// Attach handlers for all browsers
				script.onload = script.onreadystatechange = function( _, isAbort ) {

					if ( isAbort || !script.readyState || /loaded|complete/.test( script.readyState ) ) {

						// Handle memory leak in IE
						script.onload = script.onreadystatechange = null;

						// Remove the script
						if ( script.parentNode ) {
							script.parentNode.removeChild( script );
						}

						// Dereference the script
						script = null;

						// Callback if not abort
						if ( !isAbort ) {
							callback( 200, "success" );
						}
					}
				};

				// Circumvent IE6 bugs with base elements (#2709 and #4378) by prepending
				// Use native DOM manipulation to avoid our domManip AJAX trickery
				head.insertBefore( script, head.firstChild );
			},

			abort: function() {
				if ( script ) {
					script.onload( undefined, true );
				}
			}
		};
	}
});




var oldCallbacks = [],
	rjsonp = /(=)\?(?=&|$)|\?\?/;

// Default jsonp settings
jQuery.ajaxSetup({
	jsonp: "callback",
	jsonpCallback: function() {
		var callback = oldCallbacks.pop() || ( jQuery.expando + "_" + ( nonce++ ) );
		this[ callback ] = true;
		return callback;
	}
});

// Detect, normalize options and install callbacks for jsonp requests
jQuery.ajaxPrefilter( "json jsonp", function( s, originalSettings, jqXHR ) {

	var callbackName, overwritten, responseContainer,
		jsonProp = s.jsonp !== false && ( rjsonp.test( s.url ) ?
			"url" :
			typeof s.data === "string" && !( s.contentType || "" ).indexOf("application/x-www-form-urlencoded") && rjsonp.test( s.data ) && "data"
		);

	// Handle iff the expected data type is "jsonp" or we have a parameter to set
	if ( jsonProp || s.dataTypes[ 0 ] === "jsonp" ) {

		// Get callback name, remembering preexisting value associated with it
		callbackName = s.jsonpCallback = jQuery.isFunction( s.jsonpCallback ) ?
			s.jsonpCallback() :
			s.jsonpCallback;

		// Insert callback into url or form data
		if ( jsonProp ) {
			s[ jsonProp ] = s[ jsonProp ].replace( rjsonp, "$1" + callbackName );
		} else if ( s.jsonp !== false ) {
			s.url += ( rquery.test( s.url ) ? "&" : "?" ) + s.jsonp + "=" + callbackName;
		}

		// Use data converter to retrieve json after script execution
		s.converters["script json"] = function() {
			if ( !responseContainer ) {
				jQuery.error( callbackName + " was not called" );
			}
			return responseContainer[ 0 ];
		};

		// force json dataType
		s.dataTypes[ 0 ] = "json";

		// Install callback
		overwritten = window[ callbackName ];
		window[ callbackName ] = function() {
			responseContainer = arguments;
		};

		// Clean-up function (fires after converters)
		jqXHR.always(function() {
			// Restore preexisting value
			window[ callbackName ] = overwritten;

			// Save back as free
			if ( s[ callbackName ] ) {
				// make sure that re-using the options doesn't screw things around
				s.jsonpCallback = originalSettings.jsonpCallback;

				// save the callback name for future use
				oldCallbacks.push( callbackName );
			}

			// Call if it was a function and we have a response
			if ( responseContainer && jQuery.isFunction( overwritten ) ) {
				overwritten( responseContainer[ 0 ] );
			}

			responseContainer = overwritten = undefined;
		});

		// Delegate to script
		return "script";
	}
});




// data: string of html
// context (optional): If specified, the fragment will be created in this context, defaults to document
// keepScripts (optional): If true, will include scripts passed in the html string
jQuery.parseHTML = function( data, context, keepScripts ) {
	if ( !data || typeof data !== "string" ) {
		return null;
	}
	if ( typeof context === "boolean" ) {
		keepScripts = context;
		context = false;
	}
	context = context || document;

	var parsed = rsingleTag.exec( data ),
		scripts = !keepScripts && [];

	// Single tag
	if ( parsed ) {
		return [ context.createElement( parsed[1] ) ];
	}

	parsed = jQuery.buildFragment( [ data ], context, scripts );

	if ( scripts && scripts.length ) {
		jQuery( scripts ).remove();
	}

	return jQuery.merge( [], parsed.childNodes );
};


// Keep a copy of the old load method
var _load = jQuery.fn.load;

/**
 * Load a url into a page
 */
jQuery.fn.load = function( url, params, callback ) {
	if ( typeof url !== "string" && _load ) {
		return _load.apply( this, arguments );
	}

	var selector, response, type,
		self = this,
		off = url.indexOf(" ");

	if ( off >= 0 ) {
		selector = jQuery.trim( url.slice( off, url.length ) );
		url = url.slice( 0, off );
	}

	// If it's a function
	if ( jQuery.isFunction( params ) ) {

		// We assume that it's the callback
		callback = params;
		params = undefined;

	// Otherwise, build a param string
	} else if ( params && typeof params === "object" ) {
		type = "POST";
	}

	// If we have elements to modify, make the request
	if ( self.length > 0 ) {
		jQuery.ajax({
			url: url,

			// if "type" variable is undefined, then "GET" method will be used
			type: type,
			dataType: "html",
			data: params
		}).done(function( responseText ) {

			// Save response for use in complete callback
			response = arguments;

			self.html( selector ?

				// If a selector was specified, locate the right elements in a dummy div
				// Exclude scripts to avoid IE 'Permission Denied' errors
				jQuery("<div>").append( jQuery.parseHTML( responseText ) ).find( selector ) :

				// Otherwise use the full result
				responseText );

		}).complete( callback && function( jqXHR, status ) {
			self.each( callback, response || [ jqXHR.responseText, status, jqXHR ] );
		});
	}

	return this;
};




// Attach a bunch of functions for handling common AJAX events
jQuery.each( [ "ajaxStart", "ajaxStop", "ajaxComplete", "ajaxError", "ajaxSuccess", "ajaxSend" ], function( i, type ) {
	jQuery.fn[ type ] = function( fn ) {
		return this.on( type, fn );
	};
});




jQuery.expr.filters.animated = function( elem ) {
	return jQuery.grep(jQuery.timers, function( fn ) {
		return elem === fn.elem;
	}).length;
};





var docElem = window.document.documentElement;

/**
 * Gets a window from an element
 */
function getWindow( elem ) {
	return jQuery.isWindow( elem ) ?
		elem :
		elem.nodeType === 9 ?
			elem.defaultView || elem.parentWindow :
			false;
}

jQuery.offset = {
	setOffset: function( elem, options, i ) {
		var curPosition, curLeft, curCSSTop, curTop, curOffset, curCSSLeft, calculatePosition,
			position = jQuery.css( elem, "position" ),
			curElem = jQuery( elem ),
			props = {};

		// set position first, in-case top/left are set even on static elem
		if ( position === "static" ) {
			elem.style.position = "relative";
		}

		curOffset = curElem.offset();
		curCSSTop = jQuery.css( elem, "top" );
		curCSSLeft = jQuery.css( elem, "left" );
		calculatePosition = ( position === "absolute" || position === "fixed" ) &&
			jQuery.inArray("auto", [ curCSSTop, curCSSLeft ] ) > -1;

		// need to be able to calculate position if either top or left is auto and position is either absolute or fixed
		if ( calculatePosition ) {
			curPosition = curElem.position();
			curTop = curPosition.top;
			curLeft = curPosition.left;
		} else {
			curTop = parseFloat( curCSSTop ) || 0;
			curLeft = parseFloat( curCSSLeft ) || 0;
		}

		if ( jQuery.isFunction( options ) ) {
			options = options.call( elem, i, curOffset );
		}

		if ( options.top != null ) {
			props.top = ( options.top - curOffset.top ) + curTop;
		}
		if ( options.left != null ) {
			props.left = ( options.left - curOffset.left ) + curLeft;
		}

		if ( "using" in options ) {
			options.using.call( elem, props );
		} else {
			curElem.css( props );
		}
	}
};

jQuery.fn.extend({
	offset: function( options ) {
		if ( arguments.length ) {
			return options === undefined ?
				this :
				this.each(function( i ) {
					jQuery.offset.setOffset( this, options, i );
				});
		}

		var docElem, win,
			box = { top: 0, left: 0 },
			elem = this[ 0 ],
			doc = elem && elem.ownerDocument;

		if ( !doc ) {
			return;
		}

		docElem = doc.documentElement;

		// Make sure it's not a disconnected DOM node
		if ( !jQuery.contains( docElem, elem ) ) {
			return box;
		}

		// If we don't have gBCR, just use 0,0 rather than error
		// BlackBerry 5, iOS 3 (original iPhone)
		if ( typeof elem.getBoundingClientRect !== strundefined ) {
			box = elem.getBoundingClientRect();
		}
		win = getWindow( doc );
		return {
			top: box.top  + ( win.pageYOffset || docElem.scrollTop )  - ( docElem.clientTop  || 0 ),
			left: box.left + ( win.pageXOffset || docElem.scrollLeft ) - ( docElem.clientLeft || 0 )
		};
	},

	position: function() {
		if ( !this[ 0 ] ) {
			return;
		}

		var offsetParent, offset,
			parentOffset = { top: 0, left: 0 },
			elem = this[ 0 ];

		// fixed elements are offset from window (parentOffset = {top:0, left: 0}, because it is its only offset parent
		if ( jQuery.css( elem, "position" ) === "fixed" ) {
			// we assume that getBoundingClientRect is available when computed position is fixed
			offset = elem.getBoundingClientRect();
		} else {
			// Get *real* offsetParent
			offsetParent = this.offsetParent();

			// Get correct offsets
			offset = this.offset();
			if ( !jQuery.nodeName( offsetParent[ 0 ], "html" ) ) {
				parentOffset = offsetParent.offset();
			}

			// Add offsetParent borders
			parentOffset.top  += jQuery.css( offsetParent[ 0 ], "borderTopWidth", true );
			parentOffset.left += jQuery.css( offsetParent[ 0 ], "borderLeftWidth", true );
		}

		// Subtract parent offsets and element margins
		// note: when an element has margin: auto the offsetLeft and marginLeft
		// are the same in Safari causing offset.left to incorrectly be 0
		return {
			top:  offset.top  - parentOffset.top - jQuery.css( elem, "marginTop", true ),
			left: offset.left - parentOffset.left - jQuery.css( elem, "marginLeft", true)
		};
	},

	offsetParent: function() {
		return this.map(function() {
			var offsetParent = this.offsetParent || docElem;

			while ( offsetParent && ( !jQuery.nodeName( offsetParent, "html" ) && jQuery.css( offsetParent, "position" ) === "static" ) ) {
				offsetParent = offsetParent.offsetParent;
			}
			return offsetParent || docElem;
		});
	}
});

// Create scrollLeft and scrollTop methods
jQuery.each( { scrollLeft: "pageXOffset", scrollTop: "pageYOffset" }, function( method, prop ) {
	var top = /Y/.test( prop );

	jQuery.fn[ method ] = function( val ) {
		return access( this, function( elem, method, val ) {
			var win = getWindow( elem );

			if ( val === undefined ) {
				return win ? (prop in win) ? win[ prop ] :
					win.document.documentElement[ method ] :
					elem[ method ];
			}

			if ( win ) {
				win.scrollTo(
					!top ? val : jQuery( win ).scrollLeft(),
					top ? val : jQuery( win ).scrollTop()
				);

			} else {
				elem[ method ] = val;
			}
		}, method, val, arguments.length, null );
	};
});

// Add the top/left cssHooks using jQuery.fn.position
// Webkit bug: https://bugs.webkit.org/show_bug.cgi?id=29084
// getComputedStyle returns percent when specified for top/left/bottom/right
// rather than make the css module depend on the offset module, we just check for it here
jQuery.each( [ "top", "left" ], function( i, prop ) {
	jQuery.cssHooks[ prop ] = addGetHookIf( support.pixelPosition,
		function( elem, computed ) {
			if ( computed ) {
				computed = curCSS( elem, prop );
				// if curCSS returns percentage, fallback to offset
				return rnumnonpx.test( computed ) ?
					jQuery( elem ).position()[ prop ] + "px" :
					computed;
			}
		}
	);
});


// Create innerHeight, innerWidth, height, width, outerHeight and outerWidth methods
jQuery.each( { Height: "height", Width: "width" }, function( name, type ) {
	jQuery.each( { padding: "inner" + name, content: type, "": "outer" + name }, function( defaultExtra, funcName ) {
		// margin is only for outerHeight, outerWidth
		jQuery.fn[ funcName ] = function( margin, value ) {
			var chainable = arguments.length && ( defaultExtra || typeof margin !== "boolean" ),
				extra = defaultExtra || ( margin === true || value === true ? "margin" : "border" );

			return access( this, function( elem, type, value ) {
				var doc;

				if ( jQuery.isWindow( elem ) ) {
					// As of 5/8/2012 this will yield incorrect results for Mobile Safari, but there
					// isn't a whole lot we can do. See pull request at this URL for discussion:
					// https://github.com/jquery/jquery/pull/764
					return elem.document.documentElement[ "client" + name ];
				}

				// Get document width or height
				if ( elem.nodeType === 9 ) {
					doc = elem.documentElement;

					// Either scroll[Width/Height] or offset[Width/Height] or client[Width/Height], whichever is greatest
					// unfortunately, this causes bug #3838 in IE6/8 only, but there is currently no good, small way to fix it.
					return Math.max(
						elem.body[ "scroll" + name ], doc[ "scroll" + name ],
						elem.body[ "offset" + name ], doc[ "offset" + name ],
						doc[ "client" + name ]
					);
				}

				return value === undefined ?
					// Get width or height on the element, requesting but not forcing parseFloat
					jQuery.css( elem, type, extra ) :

					// Set width or height on the element
					jQuery.style( elem, type, value, extra );
			}, type, chainable ? margin : undefined, chainable, null );
		};
	});
});


// The number of elements contained in the matched element set
jQuery.fn.size = function() {
	return this.length;
};

jQuery.fn.andSelf = jQuery.fn.addBack;




// Register as a named AMD module, since jQuery can be concatenated with other
// files that may use define, but not via a proper concatenation script that
// understands anonymous AMD modules. A named AMD is safest and most robust
// way to register. Lowercase jquery is used because AMD module names are
// derived from file names, and jQuery is normally delivered in a lowercase
// file name. Do this after creating the global so that if an AMD module wants
// to call noConflict to hide this version of jQuery, it will work.

// Note that for maximum portability, libraries that are not jQuery should
// declare themselves as anonymous modules, and avoid setting a global if an
// AMD loader is present. jQuery is a special case. For more information, see
// https://github.com/jrburke/requirejs/wiki/Updating-existing-libraries#wiki-anon

if ( typeof define === "function" && define.amd ) {
	define( "jquery", [], function() {
		return jQuery;
	});
}




var
	// Map over jQuery in case of overwrite
	_jQuery = window.jQuery,

	// Map over the $ in case of overwrite
	_$ = window.$;

jQuery.noConflict = function( deep ) {
	if ( window.$ === jQuery ) {
		window.$ = _$;
	}

	if ( deep && window.jQuery === jQuery ) {
		window.jQuery = _jQuery;
	}

	return jQuery;
};

// Expose jQuery and $ identifiers, even in
// AMD (#7102#comment:10, https://github.com/jquery/jquery/pull/557)
// and CommonJS for browser emulators (#13566)
if ( typeof noGlobal === strundefined ) {
	window.jQuery = window.$ = jQuery;
}




return jQuery;

}));
/*! jQuery UI - v1.11.4 - 2016-02-27
* http://jqueryui.com
* Includes: core.js, widget.js, mouse.js, position.js, draggable.js, droppable.js, resizable.js, selectable.js, sortable.js
* Copyright jQuery Foundation and other contributors; Licensed MIT */

(function( factory ) {
	if ( typeof define === "function" && define.amd ) {

		// AMD. Register as an anonymous module.
		define([ "jquery" ], factory );
	} else {

		// Browser globals
		factory( jQuery );
	}
}(function( $ ) {
/*!
 * jQuery UI Core 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/category/ui-core/
 */


// $.ui might exist from components with no dependencies, e.g., $.ui.position
$.ui = $.ui || {};

$.extend( $.ui, {
	version: "1.11.4",

	keyCode: {
		BACKSPACE: 8,
		COMMA: 188,
		DELETE: 46,
		DOWN: 40,
		END: 35,
		ENTER: 13,
		ESCAPE: 27,
		HOME: 36,
		LEFT: 37,
		PAGE_DOWN: 34,
		PAGE_UP: 33,
		PERIOD: 190,
		RIGHT: 39,
		SPACE: 32,
		TAB: 9,
		UP: 38
	}
});

// plugins
$.fn.extend({
	scrollParent: function( includeHidden ) {
		var position = this.css( "position" ),
			excludeStaticParent = position === "absolute",
			overflowRegex = includeHidden ? /(auto|scroll|hidden)/ : /(auto|scroll)/,
			scrollParent = this.parents().filter( function() {
				var parent = $( this );
				if ( excludeStaticParent && parent.css( "position" ) === "static" ) {
					return false;
				}
				return overflowRegex.test( parent.css( "overflow" ) + parent.css( "overflow-y" ) + parent.css( "overflow-x" ) );
			}).eq( 0 );

		return position === "fixed" || !scrollParent.length ? $( this[ 0 ].ownerDocument || document ) : scrollParent;
	},

	uniqueId: (function() {
		var uuid = 0;

		return function() {
			return this.each(function() {
				if ( !this.id ) {
					this.id = "ui-id-" + ( ++uuid );
				}
			});
		};
	})(),

	removeUniqueId: function() {
		return this.each(function() {
			if ( /^ui-id-\d+$/.test( this.id ) ) {
				$( this ).removeAttr( "id" );
			}
		});
	}
});

// selectors
function focusable( element, isTabIndexNotNaN ) {
	var map, mapName, img,
		nodeName = element.nodeName.toLowerCase();
	if ( "area" === nodeName ) {
		map = element.parentNode;
		mapName = map.name;
		if ( !element.href || !mapName || map.nodeName.toLowerCase() !== "map" ) {
			return false;
		}
		img = $( "img[usemap='#" + mapName + "']" )[ 0 ];
		return !!img && visible( img );
	}
	return ( /^(input|select|textarea|button|object)$/.test( nodeName ) ?
		!element.disabled :
		"a" === nodeName ?
			element.href || isTabIndexNotNaN :
			isTabIndexNotNaN) &&
		// the element and all of its ancestors must be visible
		visible( element );
}

function visible( element ) {
	return $.expr.filters.visible( element ) &&
		!$( element ).parents().addBack().filter(function() {
			return $.css( this, "visibility" ) === "hidden";
		}).length;
}

$.extend( $.expr[ ":" ], {
	data: $.expr.createPseudo ?
		$.expr.createPseudo(function( dataName ) {
			return function( elem ) {
				return !!$.data( elem, dataName );
			};
		}) :
		// support: jQuery <1.8
		function( elem, i, match ) {
			return !!$.data( elem, match[ 3 ] );
		},

	focusable: function( element ) {
		return focusable( element, !isNaN( $.attr( element, "tabindex" ) ) );
	},

	tabbable: function( element ) {
		var tabIndex = $.attr( element, "tabindex" ),
			isTabIndexNaN = isNaN( tabIndex );
		return ( isTabIndexNaN || tabIndex >= 0 ) && focusable( element, !isTabIndexNaN );
	}
});

// support: jQuery <1.8
if ( !$( "<a>" ).outerWidth( 1 ).jquery ) {
	$.each( [ "Width", "Height" ], function( i, name ) {
		var side = name === "Width" ? [ "Left", "Right" ] : [ "Top", "Bottom" ],
			type = name.toLowerCase(),
			orig = {
				innerWidth: $.fn.innerWidth,
				innerHeight: $.fn.innerHeight,
				outerWidth: $.fn.outerWidth,
				outerHeight: $.fn.outerHeight
			};

		function reduce( elem, size, border, margin ) {
			$.each( side, function() {
				size -= parseFloat( $.css( elem, "padding" + this ) ) || 0;
				if ( border ) {
					size -= parseFloat( $.css( elem, "border" + this + "Width" ) ) || 0;
				}
				if ( margin ) {
					size -= parseFloat( $.css( elem, "margin" + this ) ) || 0;
				}
			});
			return size;
		}

		$.fn[ "inner" + name ] = function( size ) {
			if ( size === undefined ) {
				return orig[ "inner" + name ].call( this );
			}

			return this.each(function() {
				$( this ).css( type, reduce( this, size ) + "px" );
			});
		};

		$.fn[ "outer" + name] = function( size, margin ) {
			if ( typeof size !== "number" ) {
				return orig[ "outer" + name ].call( this, size );
			}

			return this.each(function() {
				$( this).css( type, reduce( this, size, true, margin ) + "px" );
			});
		};
	});
}

// support: jQuery <1.8
if ( !$.fn.addBack ) {
	$.fn.addBack = function( selector ) {
		return this.add( selector == null ?
			this.prevObject : this.prevObject.filter( selector )
		);
	};
}

// support: jQuery 1.6.1, 1.6.2 (http://bugs.jquery.com/ticket/9413)
if ( $( "<a>" ).data( "a-b", "a" ).removeData( "a-b" ).data( "a-b" ) ) {
	$.fn.removeData = (function( removeData ) {
		return function( key ) {
			if ( arguments.length ) {
				return removeData.call( this, $.camelCase( key ) );
			} else {
				return removeData.call( this );
			}
		};
	})( $.fn.removeData );
}

// deprecated
$.ui.ie = !!/msie [\w.]+/.exec( navigator.userAgent.toLowerCase() );

$.fn.extend({
	focus: (function( orig ) {
		return function( delay, fn ) {
			return typeof delay === "number" ?
				this.each(function() {
					var elem = this;
					setTimeout(function() {
						$( elem ).focus();
						if ( fn ) {
							fn.call( elem );
						}
					}, delay );
				}) :
				orig.apply( this, arguments );
		};
	})( $.fn.focus ),

	disableSelection: (function() {
		var eventType = "onselectstart" in document.createElement( "div" ) ?
			"selectstart" :
			"mousedown";

		return function() {
			return this.bind( eventType + ".ui-disableSelection", function( event ) {
				event.preventDefault();
			});
		};
	})(),

	enableSelection: function() {
		return this.unbind( ".ui-disableSelection" );
	},

	zIndex: function( zIndex ) {
		if ( zIndex !== undefined ) {
			return this.css( "zIndex", zIndex );
		}

		if ( this.length ) {
			var elem = $( this[ 0 ] ), position, value;
			while ( elem.length && elem[ 0 ] !== document ) {
				// Ignore z-index if position is set to a value where z-index is ignored by the browser
				// This makes behavior of this function consistent across browsers
				// WebKit always returns auto if the element is positioned
				position = elem.css( "position" );
				if ( position === "absolute" || position === "relative" || position === "fixed" ) {
					// IE returns 0 when zIndex is not specified
					// other browsers return a string
					// we ignore the case of nested elements with an explicit value of 0
					// <div style="z-index: -10;"><div style="z-index: 0;"></div></div>
					value = parseInt( elem.css( "zIndex" ), 10 );
					if ( !isNaN( value ) && value !== 0 ) {
						return value;
					}
				}
				elem = elem.parent();
			}
		}

		return 0;
	}
});

// $.ui.plugin is deprecated. Use $.widget() extensions instead.
$.ui.plugin = {
	add: function( module, option, set ) {
		var i,
			proto = $.ui[ module ].prototype;
		for ( i in set ) {
			proto.plugins[ i ] = proto.plugins[ i ] || [];
			proto.plugins[ i ].push( [ option, set[ i ] ] );
		}
	},
	call: function( instance, name, args, allowDisconnected ) {
		var i,
			set = instance.plugins[ name ];

		if ( !set ) {
			return;
		}

		if ( !allowDisconnected && ( !instance.element[ 0 ].parentNode || instance.element[ 0 ].parentNode.nodeType === 11 ) ) {
			return;
		}

		for ( i = 0; i < set.length; i++ ) {
			if ( instance.options[ set[ i ][ 0 ] ] ) {
				set[ i ][ 1 ].apply( instance.element, args );
			}
		}
	}
};


/*!
 * jQuery UI Widget 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/jQuery.widget/
 */


var widget_uuid = 0,
	widget_slice = Array.prototype.slice;

$.cleanData = (function( orig ) {
	return function( elems ) {
		var events, elem, i;
		for ( i = 0; (elem = elems[i]) != null; i++ ) {
			try {

				// Only trigger remove when necessary to save time
				events = $._data( elem, "events" );
				if ( events && events.remove ) {
					$( elem ).triggerHandler( "remove" );
				}

			// http://bugs.jquery.com/ticket/8235
			} catch ( e ) {}
		}
		orig( elems );
	};
})( $.cleanData );

$.widget = function( name, base, prototype ) {
	var fullName, existingConstructor, constructor, basePrototype,
		// proxiedPrototype allows the provided prototype to remain unmodified
		// so that it can be used as a mixin for multiple widgets (#8876)
		proxiedPrototype = {},
		namespace = name.split( "." )[ 0 ];

	name = name.split( "." )[ 1 ];
	fullName = namespace + "-" + name;

	if ( !prototype ) {
		prototype = base;
		base = $.Widget;
	}

	// create selector for plugin
	$.expr[ ":" ][ fullName.toLowerCase() ] = function( elem ) {
		return !!$.data( elem, fullName );
	};

	$[ namespace ] = $[ namespace ] || {};
	existingConstructor = $[ namespace ][ name ];
	constructor = $[ namespace ][ name ] = function( options, element ) {
		// allow instantiation without "new" keyword
		if ( !this._createWidget ) {
			return new constructor( options, element );
		}

		// allow instantiation without initializing for simple inheritance
		// must use "new" keyword (the code above always passes args)
		if ( arguments.length ) {
			this._createWidget( options, element );
		}
	};
	// extend with the existing constructor to carry over any static properties
	$.extend( constructor, existingConstructor, {
		version: prototype.version,
		// copy the object used to create the prototype in case we need to
		// redefine the widget later
		_proto: $.extend( {}, prototype ),
		// track widgets that inherit from this widget in case this widget is
		// redefined after a widget inherits from it
		_childConstructors: []
	});

	basePrototype = new base();
	// we need to make the options hash a property directly on the new instance
	// otherwise we'll modify the options hash on the prototype that we're
	// inheriting from
	basePrototype.options = $.widget.extend( {}, basePrototype.options );
	$.each( prototype, function( prop, value ) {
		if ( !$.isFunction( value ) ) {
			proxiedPrototype[ prop ] = value;
			return;
		}
		proxiedPrototype[ prop ] = (function() {
			var _super = function() {
					return base.prototype[ prop ].apply( this, arguments );
				},
				_superApply = function( args ) {
					return base.prototype[ prop ].apply( this, args );
				};
			return function() {
				var __super = this._super,
					__superApply = this._superApply,
					returnValue;

				this._super = _super;
				this._superApply = _superApply;

				returnValue = value.apply( this, arguments );

				this._super = __super;
				this._superApply = __superApply;

				return returnValue;
			};
		})();
	});
	constructor.prototype = $.widget.extend( basePrototype, {
		// TODO: remove support for widgetEventPrefix
		// always use the name + a colon as the prefix, e.g., draggable:start
		// don't prefix for widgets that aren't DOM-based
		widgetEventPrefix: existingConstructor ? (basePrototype.widgetEventPrefix || name) : name
	}, proxiedPrototype, {
		constructor: constructor,
		namespace: namespace,
		widgetName: name,
		widgetFullName: fullName
	});

	// If this widget is being redefined then we need to find all widgets that
	// are inheriting from it and redefine all of them so that they inherit from
	// the new version of this widget. We're essentially trying to replace one
	// level in the prototype chain.
	if ( existingConstructor ) {
		$.each( existingConstructor._childConstructors, function( i, child ) {
			var childPrototype = child.prototype;

			// redefine the child widget using the same prototype that was
			// originally used, but inherit from the new version of the base
			$.widget( childPrototype.namespace + "." + childPrototype.widgetName, constructor, child._proto );
		});
		// remove the list of existing child constructors from the old constructor
		// so the old child constructors can be garbage collected
		delete existingConstructor._childConstructors;
	} else {
		base._childConstructors.push( constructor );
	}

	$.widget.bridge( name, constructor );

	return constructor;
};

$.widget.extend = function( target ) {
	var input = widget_slice.call( arguments, 1 ),
		inputIndex = 0,
		inputLength = input.length,
		key,
		value;
	for ( ; inputIndex < inputLength; inputIndex++ ) {
		for ( key in input[ inputIndex ] ) {
			value = input[ inputIndex ][ key ];
			if ( input[ inputIndex ].hasOwnProperty( key ) && value !== undefined ) {
				// Clone objects
				if ( $.isPlainObject( value ) ) {
					target[ key ] = $.isPlainObject( target[ key ] ) ?
						$.widget.extend( {}, target[ key ], value ) :
						// Don't extend strings, arrays, etc. with objects
						$.widget.extend( {}, value );
				// Copy everything else by reference
				} else {
					target[ key ] = value;
				}
			}
		}
	}
	return target;
};

$.widget.bridge = function( name, object ) {
	var fullName = object.prototype.widgetFullName || name;
	$.fn[ name ] = function( options ) {
		var isMethodCall = typeof options === "string",
			args = widget_slice.call( arguments, 1 ),
			returnValue = this;

		if ( isMethodCall ) {
			this.each(function() {
				var methodValue,
					instance = $.data( this, fullName );
				if ( options === "instance" ) {
					returnValue = instance;
					return false;
				}
				if ( !instance ) {
					return $.error( "cannot call methods on " + name + " prior to initialization; " +
						"attempted to call method '" + options + "'" );
				}
				if ( !$.isFunction( instance[options] ) || options.charAt( 0 ) === "_" ) {
					return $.error( "no such method '" + options + "' for " + name + " widget instance" );
				}
				methodValue = instance[ options ].apply( instance, args );
				if ( methodValue !== instance && methodValue !== undefined ) {
					returnValue = methodValue && methodValue.jquery ?
						returnValue.pushStack( methodValue.get() ) :
						methodValue;
					return false;
				}
			});
		} else {

			// Allow multiple hashes to be passed on init
			if ( args.length ) {
				options = $.widget.extend.apply( null, [ options ].concat(args) );
			}

			this.each(function() {
				var instance = $.data( this, fullName );
				if ( instance ) {
					instance.option( options || {} );
					if ( instance._init ) {
						instance._init();
					}
				} else {
					$.data( this, fullName, new object( options, this ) );
				}
			});
		}

		return returnValue;
	};
};

$.Widget = function( /* options, element */ ) {};
$.Widget._childConstructors = [];

$.Widget.prototype = {
	widgetName: "widget",
	widgetEventPrefix: "",
	defaultElement: "<div>",
	options: {
		disabled: false,

		// callbacks
		create: null
	},
	_createWidget: function( options, element ) {
		element = $( element || this.defaultElement || this )[ 0 ];
		this.element = $( element );
		this.uuid = widget_uuid++;
		this.eventNamespace = "." + this.widgetName + this.uuid;

		this.bindings = $();
		this.hoverable = $();
		this.focusable = $();

		if ( element !== this ) {
			$.data( element, this.widgetFullName, this );
			this._on( true, this.element, {
				remove: function( event ) {
					if ( event.target === element ) {
						this.destroy();
					}
				}
			});
			this.document = $( element.style ?
				// element within the document
				element.ownerDocument :
				// element is window or document
				element.document || element );
			this.window = $( this.document[0].defaultView || this.document[0].parentWindow );
		}

		this.options = $.widget.extend( {},
			this.options,
			this._getCreateOptions(),
			options );

		this._create();
		this._trigger( "create", null, this._getCreateEventData() );
		this._init();
	},
	_getCreateOptions: $.noop,
	_getCreateEventData: $.noop,
	_create: $.noop,
	_init: $.noop,

	destroy: function() {
		this._destroy();
		// we can probably remove the unbind calls in 2.0
		// all event bindings should go through this._on()
		this.element
			.unbind( this.eventNamespace )
			.removeData( this.widgetFullName )
			// support: jquery <1.6.3
			// http://bugs.jquery.com/ticket/9413
			.removeData( $.camelCase( this.widgetFullName ) );
		this.widget()
			.unbind( this.eventNamespace )
			.removeAttr( "aria-disabled" )
			.removeClass(
				this.widgetFullName + "-disabled " +
				"ui-state-disabled" );

		// clean up events and states
		this.bindings.unbind( this.eventNamespace );
		this.hoverable.removeClass( "ui-state-hover" );
		this.focusable.removeClass( "ui-state-focus" );
	},
	_destroy: $.noop,

	widget: function() {
		return this.element;
	},

	option: function( key, value ) {
		var options = key,
			parts,
			curOption,
			i;

		if ( arguments.length === 0 ) {
			// don't return a reference to the internal hash
			return $.widget.extend( {}, this.options );
		}

		if ( typeof key === "string" ) {
			// handle nested keys, e.g., "foo.bar" => { foo: { bar: ___ } }
			options = {};
			parts = key.split( "." );
			key = parts.shift();
			if ( parts.length ) {
				curOption = options[ key ] = $.widget.extend( {}, this.options[ key ] );
				for ( i = 0; i < parts.length - 1; i++ ) {
					curOption[ parts[ i ] ] = curOption[ parts[ i ] ] || {};
					curOption = curOption[ parts[ i ] ];
				}
				key = parts.pop();
				if ( arguments.length === 1 ) {
					return curOption[ key ] === undefined ? null : curOption[ key ];
				}
				curOption[ key ] = value;
			} else {
				if ( arguments.length === 1 ) {
					return this.options[ key ] === undefined ? null : this.options[ key ];
				}
				options[ key ] = value;
			}
		}

		this._setOptions( options );

		return this;
	},
	_setOptions: function( options ) {
		var key;

		for ( key in options ) {
			this._setOption( key, options[ key ] );
		}

		return this;
	},
	_setOption: function( key, value ) {
		this.options[ key ] = value;

		if ( key === "disabled" ) {
			this.widget()
				.toggleClass( this.widgetFullName + "-disabled", !!value );

			// If the widget is becoming disabled, then nothing is interactive
			if ( value ) {
				this.hoverable.removeClass( "ui-state-hover" );
				this.focusable.removeClass( "ui-state-focus" );
			}
		}

		return this;
	},

	enable: function() {
		return this._setOptions({ disabled: false });
	},
	disable: function() {
		return this._setOptions({ disabled: true });
	},

	_on: function( suppressDisabledCheck, element, handlers ) {
		var delegateElement,
			instance = this;

		// no suppressDisabledCheck flag, shuffle arguments
		if ( typeof suppressDisabledCheck !== "boolean" ) {
			handlers = element;
			element = suppressDisabledCheck;
			suppressDisabledCheck = false;
		}

		// no element argument, shuffle and use this.element
		if ( !handlers ) {
			handlers = element;
			element = this.element;
			delegateElement = this.widget();
		} else {
			element = delegateElement = $( element );
			this.bindings = this.bindings.add( element );
		}

		$.each( handlers, function( event, handler ) {
			function handlerProxy() {
				// allow widgets to customize the disabled handling
				// - disabled as an array instead of boolean
				// - disabled class as method for disabling individual parts
				if ( !suppressDisabledCheck &&
						( instance.options.disabled === true ||
							$( this ).hasClass( "ui-state-disabled" ) ) ) {
					return;
				}
				return ( typeof handler === "string" ? instance[ handler ] : handler )
					.apply( instance, arguments );
			}

			// copy the guid so direct unbinding works
			if ( typeof handler !== "string" ) {
				handlerProxy.guid = handler.guid =
					handler.guid || handlerProxy.guid || $.guid++;
			}

			var match = event.match( /^([\w:-]*)\s*(.*)$/ ),
				eventName = match[1] + instance.eventNamespace,
				selector = match[2];
			if ( selector ) {
				delegateElement.delegate( selector, eventName, handlerProxy );
			} else {
				element.bind( eventName, handlerProxy );
			}
		});
	},

	_off: function( element, eventName ) {
		eventName = (eventName || "").split( " " ).join( this.eventNamespace + " " ) +
			this.eventNamespace;
		element.unbind( eventName ).undelegate( eventName );

		// Clear the stack to avoid memory leaks (#10056)
		this.bindings = $( this.bindings.not( element ).get() );
		this.focusable = $( this.focusable.not( element ).get() );
		this.hoverable = $( this.hoverable.not( element ).get() );
	},

	_delay: function( handler, delay ) {
		function handlerProxy() {
			return ( typeof handler === "string" ? instance[ handler ] : handler )
				.apply( instance, arguments );
		}
		var instance = this;
		return setTimeout( handlerProxy, delay || 0 );
	},

	_hoverable: function( element ) {
		this.hoverable = this.hoverable.add( element );
		this._on( element, {
			mouseenter: function( event ) {
				$( event.currentTarget ).addClass( "ui-state-hover" );
			},
			mouseleave: function( event ) {
				$( event.currentTarget ).removeClass( "ui-state-hover" );
			}
		});
	},

	_focusable: function( element ) {
		this.focusable = this.focusable.add( element );
		this._on( element, {
			focusin: function( event ) {
				$( event.currentTarget ).addClass( "ui-state-focus" );
			},
			focusout: function( event ) {
				$( event.currentTarget ).removeClass( "ui-state-focus" );
			}
		});
	},

	_trigger: function( type, event, data ) {
		var prop, orig,
			callback = this.options[ type ];

		data = data || {};
		event = $.Event( event );
		event.type = ( type === this.widgetEventPrefix ?
			type :
			this.widgetEventPrefix + type ).toLowerCase();
		// the original event may come from any element
		// so we need to reset the target on the new event
		event.target = this.element[ 0 ];

		// copy original event properties over to the new event
		orig = event.originalEvent;
		if ( orig ) {
			for ( prop in orig ) {
				if ( !( prop in event ) ) {
					event[ prop ] = orig[ prop ];
				}
			}
		}

		this.element.trigger( event, data );
		return !( $.isFunction( callback ) &&
			callback.apply( this.element[0], [ event ].concat( data ) ) === false ||
			event.isDefaultPrevented() );
	}
};

$.each( { show: "fadeIn", hide: "fadeOut" }, function( method, defaultEffect ) {
	$.Widget.prototype[ "_" + method ] = function( element, options, callback ) {
		if ( typeof options === "string" ) {
			options = { effect: options };
		}
		var hasOptions,
			effectName = !options ?
				method :
				options === true || typeof options === "number" ?
					defaultEffect :
					options.effect || defaultEffect;
		options = options || {};
		if ( typeof options === "number" ) {
			options = { duration: options };
		}
		hasOptions = !$.isEmptyObject( options );
		options.complete = callback;
		if ( options.delay ) {
			element.delay( options.delay );
		}
		if ( hasOptions && $.effects && $.effects.effect[ effectName ] ) {
			element[ method ]( options );
		} else if ( effectName !== method && element[ effectName ] ) {
			element[ effectName ]( options.duration, options.easing, callback );
		} else {
			element.queue(function( next ) {
				$( this )[ method ]();
				if ( callback ) {
					callback.call( element[ 0 ] );
				}
				next();
			});
		}
	};
});

var widget = $.widget;


/*!
 * jQuery UI Mouse 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/mouse/
 */


var mouseHandled = false;
$( document ).mouseup( function() {
	mouseHandled = false;
});

var mouse = $.widget("ui.mouse", {
	version: "1.11.4",
	options: {
		cancel: "input,textarea,button,select,option",
		distance: 1,
		delay: 0
	},
	_mouseInit: function() {
		var that = this;

		this.element
			.bind("mousedown." + this.widgetName, function(event) {
				return that._mouseDown(event);
			})
			.bind("click." + this.widgetName, function(event) {
				if (true === $.data(event.target, that.widgetName + ".preventClickEvent")) {
					$.removeData(event.target, that.widgetName + ".preventClickEvent");
					event.stopImmediatePropagation();
					return false;
				}
			});

		this.started = false;
	},

	// TODO: make sure destroying one instance of mouse doesn't mess with
	// other instances of mouse
	_mouseDestroy: function() {
		this.element.unbind("." + this.widgetName);
		if ( this._mouseMoveDelegate ) {
			this.document
				.unbind("mousemove." + this.widgetName, this._mouseMoveDelegate)
				.unbind("mouseup." + this.widgetName, this._mouseUpDelegate);
		}
	},

	_mouseDown: function(event) {
		// don't let more than one widget handle mouseStart
		if ( mouseHandled ) {
			return;
		}

		this._mouseMoved = false;

		// we may have missed mouseup (out of window)
		(this._mouseStarted && this._mouseUp(event));

		this._mouseDownEvent = event;

		var that = this,
			btnIsLeft = (event.which === 1),
			// event.target.nodeName works around a bug in IE 8 with
			// disabled inputs (#7620)
			elIsCancel = (typeof this.options.cancel === "string" && event.target.nodeName ? $(event.target).closest(this.options.cancel).length : false);
		if (!btnIsLeft || elIsCancel || !this._mouseCapture(event)) {
			return true;
		}

		this.mouseDelayMet = !this.options.delay;
		if (!this.mouseDelayMet) {
			this._mouseDelayTimer = setTimeout(function() {
				that.mouseDelayMet = true;
			}, this.options.delay);
		}

		if (this._mouseDistanceMet(event) && this._mouseDelayMet(event)) {
			this._mouseStarted = (this._mouseStart(event) !== false);
			if (!this._mouseStarted) {
				event.preventDefault();
				return true;
			}
		}

		// Click event may never have fired (Gecko & Opera)
		if (true === $.data(event.target, this.widgetName + ".preventClickEvent")) {
			$.removeData(event.target, this.widgetName + ".preventClickEvent");
		}

		// these delegates are required to keep context
		this._mouseMoveDelegate = function(event) {
			return that._mouseMove(event);
		};
		this._mouseUpDelegate = function(event) {
			return that._mouseUp(event);
		};

		this.document
			.bind( "mousemove." + this.widgetName, this._mouseMoveDelegate )
			.bind( "mouseup." + this.widgetName, this._mouseUpDelegate );

		event.preventDefault();

		mouseHandled = true;
		return true;
	},

	_mouseMove: function(event) {
		// Only check for mouseups outside the document if you've moved inside the document
		// at least once. This prevents the firing of mouseup in the case of IE<9, which will
		// fire a mousemove event if content is placed under the cursor. See #7778
		// Support: IE <9
		if ( this._mouseMoved ) {
			// IE mouseup check - mouseup happened when mouse was out of window
			if ($.ui.ie && ( !document.documentMode || document.documentMode < 9 ) && !event.button) {
				return this._mouseUp(event);

			// Iframe mouseup check - mouseup occurred in another document
			} else if ( !event.which ) {
				return this._mouseUp( event );
			}
		}

		if ( event.which || event.button ) {
			this._mouseMoved = true;
		}

		if (this._mouseStarted) {
			this._mouseDrag(event);
			return event.preventDefault();
		}

		if (this._mouseDistanceMet(event) && this._mouseDelayMet(event)) {
			this._mouseStarted =
				(this._mouseStart(this._mouseDownEvent, event) !== false);
			(this._mouseStarted ? this._mouseDrag(event) : this._mouseUp(event));
		}

		return !this._mouseStarted;
	},

	_mouseUp: function(event) {
		this.document
			.unbind( "mousemove." + this.widgetName, this._mouseMoveDelegate )
			.unbind( "mouseup." + this.widgetName, this._mouseUpDelegate );

		if (this._mouseStarted) {
			this._mouseStarted = false;

			if (event.target === this._mouseDownEvent.target) {
				$.data(event.target, this.widgetName + ".preventClickEvent", true);
			}

			this._mouseStop(event);
		}

		mouseHandled = false;
		return false;
	},

	_mouseDistanceMet: function(event) {
		return (Math.max(
				Math.abs(this._mouseDownEvent.pageX - event.pageX),
				Math.abs(this._mouseDownEvent.pageY - event.pageY)
			) >= this.options.distance
		);
	},

	_mouseDelayMet: function(/* event */) {
		return this.mouseDelayMet;
	},

	// These are placeholder methods, to be overriden by extending plugin
	_mouseStart: function(/* event */) {},
	_mouseDrag: function(/* event */) {},
	_mouseStop: function(/* event */) {},
	_mouseCapture: function(/* event */) { return true; }
});


/*!
 * jQuery UI Position 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/position/
 */

(function() {

$.ui = $.ui || {};

var cachedScrollbarWidth, supportsOffsetFractions,
	max = Math.max,
	abs = Math.abs,
	round = Math.round,
	rhorizontal = /left|center|right/,
	rvertical = /top|center|bottom/,
	roffset = /[\+\-]\d+(\.[\d]+)?%?/,
	rposition = /^\w+/,
	rpercent = /%$/,
	_position = $.fn.position;

function getOffsets( offsets, width, height ) {
	return [
		parseFloat( offsets[ 0 ] ) * ( rpercent.test( offsets[ 0 ] ) ? width / 100 : 1 ),
		parseFloat( offsets[ 1 ] ) * ( rpercent.test( offsets[ 1 ] ) ? height / 100 : 1 )
	];
}

function parseCss( element, property ) {
	return parseInt( $.css( element, property ), 10 ) || 0;
}

function getDimensions( elem ) {
	var raw = elem[0];
	if ( raw.nodeType === 9 ) {
		return {
			width: elem.width(),
			height: elem.height(),
			offset: { top: 0, left: 0 }
		};
	}
	if ( $.isWindow( raw ) ) {
		return {
			width: elem.width(),
			height: elem.height(),
			offset: { top: elem.scrollTop(), left: elem.scrollLeft() }
		};
	}
	if ( raw.preventDefault ) {
		return {
			width: 0,
			height: 0,
			offset: { top: raw.pageY, left: raw.pageX }
		};
	}
	return {
		width: elem.outerWidth(),
		height: elem.outerHeight(),
		offset: elem.offset()
	};
}

$.position = {
	scrollbarWidth: function() {
		if ( cachedScrollbarWidth !== undefined ) {
			return cachedScrollbarWidth;
		}
		var w1, w2,
			div = $( "<div style='display:block;position:absolute;width:50px;height:50px;overflow:hidden;'><div style='height:100px;width:auto;'></div></div>" ),
			innerDiv = div.children()[0];

		$( "body" ).append( div );
		w1 = innerDiv.offsetWidth;
		div.css( "overflow", "scroll" );

		w2 = innerDiv.offsetWidth;

		if ( w1 === w2 ) {
			w2 = div[0].clientWidth;
		}

		div.remove();

		return (cachedScrollbarWidth = w1 - w2);
	},
	getScrollInfo: function( within ) {
		var overflowX = within.isWindow || within.isDocument ? "" :
				within.element.css( "overflow-x" ),
			overflowY = within.isWindow || within.isDocument ? "" :
				within.element.css( "overflow-y" ),
			hasOverflowX = overflowX === "scroll" ||
				( overflowX === "auto" && within.width < within.element[0].scrollWidth ),
			hasOverflowY = overflowY === "scroll" ||
				( overflowY === "auto" && within.height < within.element[0].scrollHeight );
		return {
			width: hasOverflowY ? $.position.scrollbarWidth() : 0,
			height: hasOverflowX ? $.position.scrollbarWidth() : 0
		};
	},
	getWithinInfo: function( element ) {
		var withinElement = $( element || window ),
			isWindow = $.isWindow( withinElement[0] ),
			isDocument = !!withinElement[ 0 ] && withinElement[ 0 ].nodeType === 9;
		return {
			element: withinElement,
			isWindow: isWindow,
			isDocument: isDocument,
			offset: withinElement.offset() || { left: 0, top: 0 },
			scrollLeft: withinElement.scrollLeft(),
			scrollTop: withinElement.scrollTop(),

			// support: jQuery 1.6.x
			// jQuery 1.6 doesn't support .outerWidth/Height() on documents or windows
			width: isWindow || isDocument ? withinElement.width() : withinElement.outerWidth(),
			height: isWindow || isDocument ? withinElement.height() : withinElement.outerHeight()
		};
	}
};

$.fn.position = function( options ) {
	if ( !options || !options.of ) {
		return _position.apply( this, arguments );
	}

	// make a copy, we don't want to modify arguments
	options = $.extend( {}, options );

	var atOffset, targetWidth, targetHeight, targetOffset, basePosition, dimensions,
		target = $( options.of ),
		within = $.position.getWithinInfo( options.within ),
		scrollInfo = $.position.getScrollInfo( within ),
		collision = ( options.collision || "flip" ).split( " " ),
		offsets = {};

	dimensions = getDimensions( target );
	if ( target[0].preventDefault ) {
		// force left top to allow flipping
		options.at = "left top";
	}
	targetWidth = dimensions.width;
	targetHeight = dimensions.height;
	targetOffset = dimensions.offset;
	// clone to reuse original targetOffset later
	basePosition = $.extend( {}, targetOffset );

	// force my and at to have valid horizontal and vertical positions
	// if a value is missing or invalid, it will be converted to center
	$.each( [ "my", "at" ], function() {
		var pos = ( options[ this ] || "" ).split( " " ),
			horizontalOffset,
			verticalOffset;

		if ( pos.length === 1) {
			pos = rhorizontal.test( pos[ 0 ] ) ?
				pos.concat( [ "center" ] ) :
				rvertical.test( pos[ 0 ] ) ?
					[ "center" ].concat( pos ) :
					[ "center", "center" ];
		}
		pos[ 0 ] = rhorizontal.test( pos[ 0 ] ) ? pos[ 0 ] : "center";
		pos[ 1 ] = rvertical.test( pos[ 1 ] ) ? pos[ 1 ] : "center";

		// calculate offsets
		horizontalOffset = roffset.exec( pos[ 0 ] );
		verticalOffset = roffset.exec( pos[ 1 ] );
		offsets[ this ] = [
			horizontalOffset ? horizontalOffset[ 0 ] : 0,
			verticalOffset ? verticalOffset[ 0 ] : 0
		];

		// reduce to just the positions without the offsets
		options[ this ] = [
			rposition.exec( pos[ 0 ] )[ 0 ],
			rposition.exec( pos[ 1 ] )[ 0 ]
		];
	});

	// normalize collision option
	if ( collision.length === 1 ) {
		collision[ 1 ] = collision[ 0 ];
	}

	if ( options.at[ 0 ] === "right" ) {
		basePosition.left += targetWidth;
	} else if ( options.at[ 0 ] === "center" ) {
		basePosition.left += targetWidth / 2;
	}

	if ( options.at[ 1 ] === "bottom" ) {
		basePosition.top += targetHeight;
	} else if ( options.at[ 1 ] === "center" ) {
		basePosition.top += targetHeight / 2;
	}

	atOffset = getOffsets( offsets.at, targetWidth, targetHeight );
	basePosition.left += atOffset[ 0 ];
	basePosition.top += atOffset[ 1 ];

	return this.each(function() {
		var collisionPosition, using,
			elem = $( this ),
			elemWidth = elem.outerWidth(),
			elemHeight = elem.outerHeight(),
			marginLeft = parseCss( this, "marginLeft" ),
			marginTop = parseCss( this, "marginTop" ),
			collisionWidth = elemWidth + marginLeft + parseCss( this, "marginRight" ) + scrollInfo.width,
			collisionHeight = elemHeight + marginTop + parseCss( this, "marginBottom" ) + scrollInfo.height,
			position = $.extend( {}, basePosition ),
			myOffset = getOffsets( offsets.my, elem.outerWidth(), elem.outerHeight() );

		if ( options.my[ 0 ] === "right" ) {
			position.left -= elemWidth;
		} else if ( options.my[ 0 ] === "center" ) {
			position.left -= elemWidth / 2;
		}

		if ( options.my[ 1 ] === "bottom" ) {
			position.top -= elemHeight;
		} else if ( options.my[ 1 ] === "center" ) {
			position.top -= elemHeight / 2;
		}

		position.left += myOffset[ 0 ];
		position.top += myOffset[ 1 ];

		// if the browser doesn't support fractions, then round for consistent results
		if ( !supportsOffsetFractions ) {
			position.left = round( position.left );
			position.top = round( position.top );
		}

		collisionPosition = {
			marginLeft: marginLeft,
			marginTop: marginTop
		};

		$.each( [ "left", "top" ], function( i, dir ) {
			if ( $.ui.position[ collision[ i ] ] ) {
				$.ui.position[ collision[ i ] ][ dir ]( position, {
					targetWidth: targetWidth,
					targetHeight: targetHeight,
					elemWidth: elemWidth,
					elemHeight: elemHeight,
					collisionPosition: collisionPosition,
					collisionWidth: collisionWidth,
					collisionHeight: collisionHeight,
					offset: [ atOffset[ 0 ] + myOffset[ 0 ], atOffset [ 1 ] + myOffset[ 1 ] ],
					my: options.my,
					at: options.at,
					within: within,
					elem: elem
				});
			}
		});

		if ( options.using ) {
			// adds feedback as second argument to using callback, if present
			using = function( props ) {
				var left = targetOffset.left - position.left,
					right = left + targetWidth - elemWidth,
					top = targetOffset.top - position.top,
					bottom = top + targetHeight - elemHeight,
					feedback = {
						target: {
							element: target,
							left: targetOffset.left,
							top: targetOffset.top,
							width: targetWidth,
							height: targetHeight
						},
						element: {
							element: elem,
							left: position.left,
							top: position.top,
							width: elemWidth,
							height: elemHeight
						},
						horizontal: right < 0 ? "left" : left > 0 ? "right" : "center",
						vertical: bottom < 0 ? "top" : top > 0 ? "bottom" : "middle"
					};
				if ( targetWidth < elemWidth && abs( left + right ) < targetWidth ) {
					feedback.horizontal = "center";
				}
				if ( targetHeight < elemHeight && abs( top + bottom ) < targetHeight ) {
					feedback.vertical = "middle";
				}
				if ( max( abs( left ), abs( right ) ) > max( abs( top ), abs( bottom ) ) ) {
					feedback.important = "horizontal";
				} else {
					feedback.important = "vertical";
				}
				options.using.call( this, props, feedback );
			};
		}

		elem.offset( $.extend( position, { using: using } ) );
	});
};

$.ui.position = {
	fit: {
		left: function( position, data ) {
			var within = data.within,
				withinOffset = within.isWindow ? within.scrollLeft : within.offset.left,
				outerWidth = within.width,
				collisionPosLeft = position.left - data.collisionPosition.marginLeft,
				overLeft = withinOffset - collisionPosLeft,
				overRight = collisionPosLeft + data.collisionWidth - outerWidth - withinOffset,
				newOverRight;

			// element is wider than within
			if ( data.collisionWidth > outerWidth ) {
				// element is initially over the left side of within
				if ( overLeft > 0 && overRight <= 0 ) {
					newOverRight = position.left + overLeft + data.collisionWidth - outerWidth - withinOffset;
					position.left += overLeft - newOverRight;
				// element is initially over right side of within
				} else if ( overRight > 0 && overLeft <= 0 ) {
					position.left = withinOffset;
				// element is initially over both left and right sides of within
				} else {
					if ( overLeft > overRight ) {
						position.left = withinOffset + outerWidth - data.collisionWidth;
					} else {
						position.left = withinOffset;
					}
				}
			// too far left -> align with left edge
			} else if ( overLeft > 0 ) {
				position.left += overLeft;
			// too far right -> align with right edge
			} else if ( overRight > 0 ) {
				position.left -= overRight;
			// adjust based on position and margin
			} else {
				position.left = max( position.left - collisionPosLeft, position.left );
			}
		},
		top: function( position, data ) {
			var within = data.within,
				withinOffset = within.isWindow ? within.scrollTop : within.offset.top,
				outerHeight = data.within.height,
				collisionPosTop = position.top - data.collisionPosition.marginTop,
				overTop = withinOffset - collisionPosTop,
				overBottom = collisionPosTop + data.collisionHeight - outerHeight - withinOffset,
				newOverBottom;

			// element is taller than within
			if ( data.collisionHeight > outerHeight ) {
				// element is initially over the top of within
				if ( overTop > 0 && overBottom <= 0 ) {
					newOverBottom = position.top + overTop + data.collisionHeight - outerHeight - withinOffset;
					position.top += overTop - newOverBottom;
				// element is initially over bottom of within
				} else if ( overBottom > 0 && overTop <= 0 ) {
					position.top = withinOffset;
				// element is initially over both top and bottom of within
				} else {
					if ( overTop > overBottom ) {
						position.top = withinOffset + outerHeight - data.collisionHeight;
					} else {
						position.top = withinOffset;
					}
				}
			// too far up -> align with top
			} else if ( overTop > 0 ) {
				position.top += overTop;
			// too far down -> align with bottom edge
			} else if ( overBottom > 0 ) {
				position.top -= overBottom;
			// adjust based on position and margin
			} else {
				position.top = max( position.top - collisionPosTop, position.top );
			}
		}
	},
	flip: {
		left: function( position, data ) {
			var within = data.within,
				withinOffset = within.offset.left + within.scrollLeft,
				outerWidth = within.width,
				offsetLeft = within.isWindow ? within.scrollLeft : within.offset.left,
				collisionPosLeft = position.left - data.collisionPosition.marginLeft,
				overLeft = collisionPosLeft - offsetLeft,
				overRight = collisionPosLeft + data.collisionWidth - outerWidth - offsetLeft,
				myOffset = data.my[ 0 ] === "left" ?
					-data.elemWidth :
					data.my[ 0 ] === "right" ?
						data.elemWidth :
						0,
				atOffset = data.at[ 0 ] === "left" ?
					data.targetWidth :
					data.at[ 0 ] === "right" ?
						-data.targetWidth :
						0,
				offset = -2 * data.offset[ 0 ],
				newOverRight,
				newOverLeft;

			if ( overLeft < 0 ) {
				newOverRight = position.left + myOffset + atOffset + offset + data.collisionWidth - outerWidth - withinOffset;
				if ( newOverRight < 0 || newOverRight < abs( overLeft ) ) {
					position.left += myOffset + atOffset + offset;
				}
			} else if ( overRight > 0 ) {
				newOverLeft = position.left - data.collisionPosition.marginLeft + myOffset + atOffset + offset - offsetLeft;
				if ( newOverLeft > 0 || abs( newOverLeft ) < overRight ) {
					position.left += myOffset + atOffset + offset;
				}
			}
		},
		top: function( position, data ) {
			var within = data.within,
				withinOffset = within.offset.top + within.scrollTop,
				outerHeight = within.height,
				offsetTop = within.isWindow ? within.scrollTop : within.offset.top,
				collisionPosTop = position.top - data.collisionPosition.marginTop,
				overTop = collisionPosTop - offsetTop,
				overBottom = collisionPosTop + data.collisionHeight - outerHeight - offsetTop,
				top = data.my[ 1 ] === "top",
				myOffset = top ?
					-data.elemHeight :
					data.my[ 1 ] === "bottom" ?
						data.elemHeight :
						0,
				atOffset = data.at[ 1 ] === "top" ?
					data.targetHeight :
					data.at[ 1 ] === "bottom" ?
						-data.targetHeight :
						0,
				offset = -2 * data.offset[ 1 ],
				newOverTop,
				newOverBottom;
			if ( overTop < 0 ) {
				newOverBottom = position.top + myOffset + atOffset + offset + data.collisionHeight - outerHeight - withinOffset;
				if ( newOverBottom < 0 || newOverBottom < abs( overTop ) ) {
					position.top += myOffset + atOffset + offset;
				}
			} else if ( overBottom > 0 ) {
				newOverTop = position.top - data.collisionPosition.marginTop + myOffset + atOffset + offset - offsetTop;
				if ( newOverTop > 0 || abs( newOverTop ) < overBottom ) {
					position.top += myOffset + atOffset + offset;
				}
			}
		}
	},
	flipfit: {
		left: function() {
			$.ui.position.flip.left.apply( this, arguments );
			$.ui.position.fit.left.apply( this, arguments );
		},
		top: function() {
			$.ui.position.flip.top.apply( this, arguments );
			$.ui.position.fit.top.apply( this, arguments );
		}
	}
};

// fraction support test
(function() {
	var testElement, testElementParent, testElementStyle, offsetLeft, i,
		body = document.getElementsByTagName( "body" )[ 0 ],
		div = document.createElement( "div" );

	//Create a "fake body" for testing based on method used in jQuery.support
	testElement = document.createElement( body ? "div" : "body" );
	testElementStyle = {
		visibility: "hidden",
		width: 0,
		height: 0,
		border: 0,
		margin: 0,
		background: "none"
	};
	if ( body ) {
		$.extend( testElementStyle, {
			position: "absolute",
			left: "-1000px",
			top: "-1000px"
		});
	}
	for ( i in testElementStyle ) {
		testElement.style[ i ] = testElementStyle[ i ];
	}
	testElement.appendChild( div );
	testElementParent = body || document.documentElement;
	testElementParent.insertBefore( testElement, testElementParent.firstChild );

	div.style.cssText = "position: absolute; left: 10.7432222px;";

	offsetLeft = $( div ).offset().left;
	supportsOffsetFractions = offsetLeft > 10 && offsetLeft < 11;

	testElement.innerHTML = "";
	testElementParent.removeChild( testElement );
})();

})();

var position = $.ui.position;


/*!
 * jQuery UI Draggable 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/draggable/
 */


$.widget("ui.draggable", $.ui.mouse, {
	version: "1.11.4",
	widgetEventPrefix: "drag",
	options: {
		addClasses: true,
		appendTo: "parent",
		axis: false,
		connectToSortable: false,
		containment: false,
		cursor: "auto",
		cursorAt: false,
		grid: false,
		handle: false,
		helper: "original",
		iframeFix: false,
		opacity: false,
		refreshPositions: false,
		revert: false,
		revertDuration: 500,
		scope: "default",
		scroll: true,
		scrollSensitivity: 20,
		scrollSpeed: 20,
		snap: false,
		snapMode: "both",
		snapTolerance: 20,
		stack: false,
		zIndex: false,

		// callbacks
		drag: null,
		start: null,
		stop: null
	},
	_create: function() {

		if ( this.options.helper === "original" ) {
			this._setPositionRelative();
		}
		if (this.options.addClasses){
			this.element.addClass("ui-draggable");
		}
		if (this.options.disabled){
			this.element.addClass("ui-draggable-disabled");
		}
		this._setHandleClassName();

		this._mouseInit();
	},

	_setOption: function( key, value ) {
		this._super( key, value );
		if ( key === "handle" ) {
			this._removeHandleClassName();
			this._setHandleClassName();
		}
	},

	_destroy: function() {
		if ( ( this.helper || this.element ).is( ".ui-draggable-dragging" ) ) {
			this.destroyOnClear = true;
			return;
		}
		this.element.removeClass( "ui-draggable ui-draggable-dragging ui-draggable-disabled" );
		this._removeHandleClassName();
		this._mouseDestroy();
	},

	_mouseCapture: function(event) {
		var o = this.options;

		this._blurActiveElement( event );

		// among others, prevent a drag on a resizable-handle
		if (this.helper || o.disabled || $(event.target).closest(".ui-resizable-handle").length > 0) {
			return false;
		}

		//Quit if we're not on a valid handle
		this.handle = this._getHandle(event);
		if (!this.handle) {
			return false;
		}

		this._blockFrames( o.iframeFix === true ? "iframe" : o.iframeFix );

		return true;

	},

	_blockFrames: function( selector ) {
		this.iframeBlocks = this.document.find( selector ).map(function() {
			var iframe = $( this );

			return $( "<div>" )
				.css( "position", "absolute" )
				.appendTo( iframe.parent() )
				.outerWidth( iframe.outerWidth() )
				.outerHeight( iframe.outerHeight() )
				.offset( iframe.offset() )[ 0 ];
		});
	},

	_unblockFrames: function() {
		if ( this.iframeBlocks ) {
			this.iframeBlocks.remove();
			delete this.iframeBlocks;
		}
	},

	_blurActiveElement: function( event ) {
		var document = this.document[ 0 ];

		// Only need to blur if the event occurred on the draggable itself, see #10527
		if ( !this.handleElement.is( event.target ) ) {
			return;
		}

		// support: IE9
		// IE9 throws an "Unspecified error" accessing document.activeElement from an <iframe>
		try {

			// Support: IE9, IE10
			// If the <body> is blurred, IE will switch windows, see #9520
			if ( document.activeElement && document.activeElement.nodeName.toLowerCase() !== "body" ) {

				// Blur any element that currently has focus, see #4261
				$( document.activeElement ).blur();
			}
		} catch ( error ) {}
	},

	_mouseStart: function(event) {

		var o = this.options;

		//Create and append the visible helper
		this.helper = this._createHelper(event);

		this.helper.addClass("ui-draggable-dragging");

		//Cache the helper size
		this._cacheHelperProportions();

		//If ddmanager is used for droppables, set the global draggable
		if ($.ui.ddmanager) {
			$.ui.ddmanager.current = this;
		}

		/*
		 * - Position generation -
		 * This block generates everything position related - it's the core of draggables.
		 */

		//Cache the margins of the original element
		this._cacheMargins();

		//Store the helper's css position
		this.cssPosition = this.helper.css( "position" );
		this.scrollParent = this.helper.scrollParent( true );
		this.offsetParent = this.helper.offsetParent();
		this.hasFixedAncestor = this.helper.parents().filter(function() {
				return $( this ).css( "position" ) === "fixed";
			}).length > 0;

		//The element's absolute position on the page minus margins
		this.positionAbs = this.element.offset();
		this._refreshOffsets( event );

		//Generate the original position
		this.originalPosition = this.position = this._generatePosition( event, false );
		this.originalPageX = event.pageX;
		this.originalPageY = event.pageY;

		//Adjust the mouse offset relative to the helper if "cursorAt" is supplied
		(o.cursorAt && this._adjustOffsetFromHelper(o.cursorAt));

		//Set a containment if given in the options
		this._setContainment();

		//Trigger event + callbacks
		if (this._trigger("start", event) === false) {
			this._clear();
			return false;
		}

		//Recache the helper size
		this._cacheHelperProportions();

		//Prepare the droppable offsets
		if ($.ui.ddmanager && !o.dropBehaviour) {
			$.ui.ddmanager.prepareOffsets(this, event);
		}

		// Reset helper's right/bottom css if they're set and set explicit width/height instead
		// as this prevents resizing of elements with right/bottom set (see #7772)
		this._normalizeRightBottom();

		this._mouseDrag(event, true); //Execute the drag once - this causes the helper not to be visible before getting its correct position

		//If the ddmanager is used for droppables, inform the manager that dragging has started (see #5003)
		if ( $.ui.ddmanager ) {
			$.ui.ddmanager.dragStart(this, event);
		}

		return true;
	},

	_refreshOffsets: function( event ) {
		this.offset = {
			top: this.positionAbs.top - this.margins.top,
			left: this.positionAbs.left - this.margins.left,
			scroll: false,
			parent: this._getParentOffset(),
			relative: this._getRelativeOffset()
		};

		this.offset.click = {
			left: event.pageX - this.offset.left,
			top: event.pageY - this.offset.top
		};
	},

	_mouseDrag: function(event, noPropagation) {
		// reset any necessary cached properties (see #5009)
		if ( this.hasFixedAncestor ) {
			this.offset.parent = this._getParentOffset();
		}

		//Compute the helpers position
		this.position = this._generatePosition( event, true );
		this.positionAbs = this._convertPositionTo("absolute");

		//Call plugins and callbacks and use the resulting position if something is returned
		if (!noPropagation) {
			var ui = this._uiHash();
			if (this._trigger("drag", event, ui) === false) {
				this._mouseUp({});
				return false;
			}
			this.position = ui.position;
		}

		this.helper[ 0 ].style.left = this.position.left + "px";
		this.helper[ 0 ].style.top = this.position.top + "px";

		if ($.ui.ddmanager) {
			$.ui.ddmanager.drag(this, event);
		}

		return false;
	},

	_mouseStop: function(event) {

		//If we are using droppables, inform the manager about the drop
		var that = this,
			dropped = false;
		if ($.ui.ddmanager && !this.options.dropBehaviour) {
			dropped = $.ui.ddmanager.drop(this, event);
		}

		//if a drop comes from outside (a sortable)
		if (this.dropped) {
			dropped = this.dropped;
			this.dropped = false;
		}

		if ((this.options.revert === "invalid" && !dropped) || (this.options.revert === "valid" && dropped) || this.options.revert === true || ($.isFunction(this.options.revert) && this.options.revert.call(this.element, dropped))) {
			$(this.helper).animate(this.originalPosition, parseInt(this.options.revertDuration, 10), function() {
				if (that._trigger("stop", event) !== false) {
					that._clear();
				}
			});
		} else {
			if (this._trigger("stop", event) !== false) {
				this._clear();
			}
		}

		return false;
	},

	_mouseUp: function( event ) {
		this._unblockFrames();

		//If the ddmanager is used for droppables, inform the manager that dragging has stopped (see #5003)
		if ( $.ui.ddmanager ) {
			$.ui.ddmanager.dragStop(this, event);
		}

		// Only need to focus if the event occurred on the draggable itself, see #10527
		if ( this.handleElement.is( event.target ) ) {
			// The interaction is over; whether or not the click resulted in a drag, focus the element
			this.element.focus();
		}

		return $.ui.mouse.prototype._mouseUp.call(this, event);
	},

	cancel: function() {

		if (this.helper.is(".ui-draggable-dragging")) {
			this._mouseUp({});
		} else {
			this._clear();
		}

		return this;

	},

	_getHandle: function(event) {
		return this.options.handle ?
			!!$( event.target ).closest( this.element.find( this.options.handle ) ).length :
			true;
	},

	_setHandleClassName: function() {
		this.handleElement = this.options.handle ?
			this.element.find( this.options.handle ) : this.element;
		this.handleElement.addClass( "ui-draggable-handle" );
	},

	_removeHandleClassName: function() {
		this.handleElement.removeClass( "ui-draggable-handle" );
	},

	_createHelper: function(event) {

		var o = this.options,
			helperIsFunction = $.isFunction( o.helper ),
			helper = helperIsFunction ?
				$( o.helper.apply( this.element[ 0 ], [ event ] ) ) :
				( o.helper === "clone" ?
					this.element.clone().removeAttr( "id" ) :
					this.element );

		if (!helper.parents("body").length) {
			helper.appendTo((o.appendTo === "parent" ? this.element[0].parentNode : o.appendTo));
		}

		// http://bugs.jqueryui.com/ticket/9446
		// a helper function can return the original element
		// which wouldn't have been set to relative in _create
		if ( helperIsFunction && helper[ 0 ] === this.element[ 0 ] ) {
			this._setPositionRelative();
		}

		if (helper[0] !== this.element[0] && !(/(fixed|absolute)/).test(helper.css("position"))) {
			helper.css("position", "absolute");
		}

		return helper;

	},

	_setPositionRelative: function() {
		if ( !( /^(?:r|a|f)/ ).test( this.element.css( "position" ) ) ) {
			this.element[ 0 ].style.position = "relative";
		}
	},

	_adjustOffsetFromHelper: function(obj) {
		if (typeof obj === "string") {
			obj = obj.split(" ");
		}
		if ($.isArray(obj)) {
			obj = { left: +obj[0], top: +obj[1] || 0 };
		}
		if ("left" in obj) {
			this.offset.click.left = obj.left + this.margins.left;
		}
		if ("right" in obj) {
			this.offset.click.left = this.helperProportions.width - obj.right + this.margins.left;
		}
		if ("top" in obj) {
			this.offset.click.top = obj.top + this.margins.top;
		}
		if ("bottom" in obj) {
			this.offset.click.top = this.helperProportions.height - obj.bottom + this.margins.top;
		}
	},

	_isRootNode: function( element ) {
		return ( /(html|body)/i ).test( element.tagName ) || element === this.document[ 0 ];
	},

	_getParentOffset: function() {

		//Get the offsetParent and cache its position
		var po = this.offsetParent.offset(),
			document = this.document[ 0 ];

		// This is a special case where we need to modify a offset calculated on start, since the following happened:
		// 1. The position of the helper is absolute, so it's position is calculated based on the next positioned parent
		// 2. The actual offset parent is a child of the scroll parent, and the scroll parent isn't the document, which means that
		//    the scroll is included in the initial calculation of the offset of the parent, and never recalculated upon drag
		if (this.cssPosition === "absolute" && this.scrollParent[0] !== document && $.contains(this.scrollParent[0], this.offsetParent[0])) {
			po.left += this.scrollParent.scrollLeft();
			po.top += this.scrollParent.scrollTop();
		}

		if ( this._isRootNode( this.offsetParent[ 0 ] ) ) {
			po = { top: 0, left: 0 };
		}

		return {
			top: po.top + (parseInt(this.offsetParent.css("borderTopWidth"), 10) || 0),
			left: po.left + (parseInt(this.offsetParent.css("borderLeftWidth"), 10) || 0)
		};

	},

	_getRelativeOffset: function() {
		if ( this.cssPosition !== "relative" ) {
			return { top: 0, left: 0 };
		}

		var p = this.element.position(),
			scrollIsRootNode = this._isRootNode( this.scrollParent[ 0 ] );

		return {
			top: p.top - ( parseInt(this.helper.css( "top" ), 10) || 0 ) + ( !scrollIsRootNode ? this.scrollParent.scrollTop() : 0 ),
			left: p.left - ( parseInt(this.helper.css( "left" ), 10) || 0 ) + ( !scrollIsRootNode ? this.scrollParent.scrollLeft() : 0 )
		};

	},

	_cacheMargins: function() {
		this.margins = {
			left: (parseInt(this.element.css("marginLeft"), 10) || 0),
			top: (parseInt(this.element.css("marginTop"), 10) || 0),
			right: (parseInt(this.element.css("marginRight"), 10) || 0),
			bottom: (parseInt(this.element.css("marginBottom"), 10) || 0)
		};
	},

	_cacheHelperProportions: function() {
		this.helperProportions = {
			width: this.helper.outerWidth(),
			height: this.helper.outerHeight()
		};
	},

	_setContainment: function() {

		var isUserScrollable, c, ce,
			o = this.options,
			document = this.document[ 0 ];

		this.relativeContainer = null;

		if ( !o.containment ) {
			this.containment = null;
			return;
		}

		if ( o.containment === "window" ) {
			this.containment = [
				$( window ).scrollLeft() - this.offset.relative.left - this.offset.parent.left,
				$( window ).scrollTop() - this.offset.relative.top - this.offset.parent.top,
				$( window ).scrollLeft() + $( window ).width() - this.helperProportions.width - this.margins.left,
				$( window ).scrollTop() + ( $( window ).height() || document.body.parentNode.scrollHeight ) - this.helperProportions.height - this.margins.top
			];
			return;
		}

		if ( o.containment === "document") {
			this.containment = [
				0,
				0,
				$( document ).width() - this.helperProportions.width - this.margins.left,
				( $( document ).height() || document.body.parentNode.scrollHeight ) - this.helperProportions.height - this.margins.top
			];
			return;
		}

		if ( o.containment.constructor === Array ) {
			this.containment = o.containment;
			return;
		}

		if ( o.containment === "parent" ) {
			o.containment = this.helper[ 0 ].parentNode;
		}

		c = $( o.containment );
		ce = c[ 0 ];

		if ( !ce ) {
			return;
		}

		isUserScrollable = /(scroll|auto)/.test( c.css( "overflow" ) );

		this.containment = [
			( parseInt( c.css( "borderLeftWidth" ), 10 ) || 0 ) + ( parseInt( c.css( "paddingLeft" ), 10 ) || 0 ),
			( parseInt( c.css( "borderTopWidth" ), 10 ) || 0 ) + ( parseInt( c.css( "paddingTop" ), 10 ) || 0 ),
			( isUserScrollable ? Math.max( ce.scrollWidth, ce.offsetWidth ) : ce.offsetWidth ) -
				( parseInt( c.css( "borderRightWidth" ), 10 ) || 0 ) -
				( parseInt( c.css( "paddingRight" ), 10 ) || 0 ) -
				this.helperProportions.width -
				this.margins.left -
				this.margins.right,
			( isUserScrollable ? Math.max( ce.scrollHeight, ce.offsetHeight ) : ce.offsetHeight ) -
				( parseInt( c.css( "borderBottomWidth" ), 10 ) || 0 ) -
				( parseInt( c.css( "paddingBottom" ), 10 ) || 0 ) -
				this.helperProportions.height -
				this.margins.top -
				this.margins.bottom
		];
		this.relativeContainer = c;
	},

	_convertPositionTo: function(d, pos) {

		if (!pos) {
			pos = this.position;
		}

		var mod = d === "absolute" ? 1 : -1,
			scrollIsRootNode = this._isRootNode( this.scrollParent[ 0 ] );

		return {
			top: (
				pos.top	+																// The absolute mouse position
				this.offset.relative.top * mod +										// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.top * mod -										// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.offset.scroll.top : ( scrollIsRootNode ? 0 : this.offset.scroll.top ) ) * mod)
			),
			left: (
				pos.left +																// The absolute mouse position
				this.offset.relative.left * mod +										// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.left * mod	-										// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.offset.scroll.left : ( scrollIsRootNode ? 0 : this.offset.scroll.left ) ) * mod)
			)
		};

	},

	_generatePosition: function( event, constrainPosition ) {

		var containment, co, top, left,
			o = this.options,
			scrollIsRootNode = this._isRootNode( this.scrollParent[ 0 ] ),
			pageX = event.pageX,
			pageY = event.pageY;

		// Cache the scroll
		if ( !scrollIsRootNode || !this.offset.scroll ) {
			this.offset.scroll = {
				top: this.scrollParent.scrollTop(),
				left: this.scrollParent.scrollLeft()
			};
		}

		/*
		 * - Position constraining -
		 * Constrain the position to a mix of grid, containment.
		 */

		// If we are not dragging yet, we won't check for options
		if ( constrainPosition ) {
			if ( this.containment ) {
				if ( this.relativeContainer ){
					co = this.relativeContainer.offset();
					containment = [
						this.containment[ 0 ] + co.left,
						this.containment[ 1 ] + co.top,
						this.containment[ 2 ] + co.left,
						this.containment[ 3 ] + co.top
					];
				} else {
					containment = this.containment;
				}

				if (event.pageX - this.offset.click.left < containment[0]) {
					pageX = containment[0] + this.offset.click.left;
				}
				if (event.pageY - this.offset.click.top < containment[1]) {
					pageY = containment[1] + this.offset.click.top;
				}
				if (event.pageX - this.offset.click.left > containment[2]) {
					pageX = containment[2] + this.offset.click.left;
				}
				if (event.pageY - this.offset.click.top > containment[3]) {
					pageY = containment[3] + this.offset.click.top;
				}
			}

			if (o.grid) {
				//Check for grid elements set to 0 to prevent divide by 0 error causing invalid argument errors in IE (see ticket #6950)
				top = o.grid[1] ? this.originalPageY + Math.round((pageY - this.originalPageY) / o.grid[1]) * o.grid[1] : this.originalPageY;
				pageY = containment ? ((top - this.offset.click.top >= containment[1] || top - this.offset.click.top > containment[3]) ? top : ((top - this.offset.click.top >= containment[1]) ? top - o.grid[1] : top + o.grid[1])) : top;

				left = o.grid[0] ? this.originalPageX + Math.round((pageX - this.originalPageX) / o.grid[0]) * o.grid[0] : this.originalPageX;
				pageX = containment ? ((left - this.offset.click.left >= containment[0] || left - this.offset.click.left > containment[2]) ? left : ((left - this.offset.click.left >= containment[0]) ? left - o.grid[0] : left + o.grid[0])) : left;
			}

			if ( o.axis === "y" ) {
				pageX = this.originalPageX;
			}

			if ( o.axis === "x" ) {
				pageY = this.originalPageY;
			}
		}

		return {
			top: (
				pageY -																	// The absolute mouse position
				this.offset.click.top	-												// Click offset (relative to the element)
				this.offset.relative.top -												// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.top +												// The offsetParent's offset without borders (offset + border)
				( this.cssPosition === "fixed" ? -this.offset.scroll.top : ( scrollIsRootNode ? 0 : this.offset.scroll.top ) )
			),
			left: (
				pageX -																	// The absolute mouse position
				this.offset.click.left -												// Click offset (relative to the element)
				this.offset.relative.left -												// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.left +												// The offsetParent's offset without borders (offset + border)
				( this.cssPosition === "fixed" ? -this.offset.scroll.left : ( scrollIsRootNode ? 0 : this.offset.scroll.left ) )
			)
		};

	},

	_clear: function() {
		this.helper.removeClass("ui-draggable-dragging");
		if (this.helper[0] !== this.element[0] && !this.cancelHelperRemoval) {
			this.helper.remove();
		}
		this.helper = null;
		this.cancelHelperRemoval = false;
		if ( this.destroyOnClear ) {
			this.destroy();
		}
	},

	_normalizeRightBottom: function() {
		if ( this.options.axis !== "y" && this.helper.css( "right" ) !== "auto" ) {
			this.helper.width( this.helper.width() );
			this.helper.css( "right", "auto" );
		}
		if ( this.options.axis !== "x" && this.helper.css( "bottom" ) !== "auto" ) {
			this.helper.height( this.helper.height() );
			this.helper.css( "bottom", "auto" );
		}
	},

	// From now on bulk stuff - mainly helpers

	_trigger: function( type, event, ui ) {
		ui = ui || this._uiHash();
		$.ui.plugin.call( this, type, [ event, ui, this ], true );

		// Absolute position and offset (see #6884 ) have to be recalculated after plugins
		if ( /^(drag|start|stop)/.test( type ) ) {
			this.positionAbs = this._convertPositionTo( "absolute" );
			ui.offset = this.positionAbs;
		}
		return $.Widget.prototype._trigger.call( this, type, event, ui );
	},

	plugins: {},

	_uiHash: function() {
		return {
			helper: this.helper,
			position: this.position,
			originalPosition: this.originalPosition,
			offset: this.positionAbs
		};
	}

});

$.ui.plugin.add( "draggable", "connectToSortable", {
	start: function( event, ui, draggable ) {
		var uiSortable = $.extend( {}, ui, {
			item: draggable.element
		});

		draggable.sortables = [];
		$( draggable.options.connectToSortable ).each(function() {
			var sortable = $( this ).sortable( "instance" );

			if ( sortable && !sortable.options.disabled ) {
				draggable.sortables.push( sortable );

				// refreshPositions is called at drag start to refresh the containerCache
				// which is used in drag. This ensures it's initialized and synchronized
				// with any changes that might have happened on the page since initialization.
				sortable.refreshPositions();
				sortable._trigger("activate", event, uiSortable);
			}
		});
	},
	stop: function( event, ui, draggable ) {
		var uiSortable = $.extend( {}, ui, {
			item: draggable.element
		});

		draggable.cancelHelperRemoval = false;

		$.each( draggable.sortables, function() {
			var sortable = this;

			if ( sortable.isOver ) {
				sortable.isOver = 0;

				// Allow this sortable to handle removing the helper
				draggable.cancelHelperRemoval = true;
				sortable.cancelHelperRemoval = false;

				// Use _storedCSS To restore properties in the sortable,
				// as this also handles revert (#9675) since the draggable
				// may have modified them in unexpected ways (#8809)
				sortable._storedCSS = {
					position: sortable.placeholder.css( "position" ),
					top: sortable.placeholder.css( "top" ),
					left: sortable.placeholder.css( "left" )
				};

				sortable._mouseStop(event);

				// Once drag has ended, the sortable should return to using
				// its original helper, not the shared helper from draggable
				sortable.options.helper = sortable.options._helper;
			} else {
				// Prevent this Sortable from removing the helper.
				// However, don't set the draggable to remove the helper
				// either as another connected Sortable may yet handle the removal.
				sortable.cancelHelperRemoval = true;

				sortable._trigger( "deactivate", event, uiSortable );
			}
		});
	},
	drag: function( event, ui, draggable ) {
		$.each( draggable.sortables, function() {
			var innermostIntersecting = false,
				sortable = this;

			// Copy over variables that sortable's _intersectsWith uses
			sortable.positionAbs = draggable.positionAbs;
			sortable.helperProportions = draggable.helperProportions;
			sortable.offset.click = draggable.offset.click;

			if ( sortable._intersectsWith( sortable.containerCache ) ) {
				innermostIntersecting = true;

				$.each( draggable.sortables, function() {
					// Copy over variables that sortable's _intersectsWith uses
					this.positionAbs = draggable.positionAbs;
					this.helperProportions = draggable.helperProportions;
					this.offset.click = draggable.offset.click;

					if ( this !== sortable &&
							this._intersectsWith( this.containerCache ) &&
							$.contains( sortable.element[ 0 ], this.element[ 0 ] ) ) {
						innermostIntersecting = false;
					}

					return innermostIntersecting;
				});
			}

			if ( innermostIntersecting ) {
				// If it intersects, we use a little isOver variable and set it once,
				// so that the move-in stuff gets fired only once.
				if ( !sortable.isOver ) {
					sortable.isOver = 1;

					// Store draggable's parent in case we need to reappend to it later.
					draggable._parent = ui.helper.parent();

					sortable.currentItem = ui.helper
						.appendTo( sortable.element )
						.data( "ui-sortable-item", true );

					// Store helper option to later restore it
					sortable.options._helper = sortable.options.helper;

					sortable.options.helper = function() {
						return ui.helper[ 0 ];
					};

					// Fire the start events of the sortable with our passed browser event,
					// and our own helper (so it doesn't create a new one)
					event.target = sortable.currentItem[ 0 ];
					sortable._mouseCapture( event, true );
					sortable._mouseStart( event, true, true );

					// Because the browser event is way off the new appended portlet,
					// modify necessary variables to reflect the changes
					sortable.offset.click.top = draggable.offset.click.top;
					sortable.offset.click.left = draggable.offset.click.left;
					sortable.offset.parent.left -= draggable.offset.parent.left -
						sortable.offset.parent.left;
					sortable.offset.parent.top -= draggable.offset.parent.top -
						sortable.offset.parent.top;

					draggable._trigger( "toSortable", event );

					// Inform draggable that the helper is in a valid drop zone,
					// used solely in the revert option to handle "valid/invalid".
					draggable.dropped = sortable.element;

					// Need to refreshPositions of all sortables in the case that
					// adding to one sortable changes the location of the other sortables (#9675)
					$.each( draggable.sortables, function() {
						this.refreshPositions();
					});

					// hack so receive/update callbacks work (mostly)
					draggable.currentItem = draggable.element;
					sortable.fromOutside = draggable;
				}

				if ( sortable.currentItem ) {
					sortable._mouseDrag( event );
					// Copy the sortable's position because the draggable's can potentially reflect
					// a relative position, while sortable is always absolute, which the dragged
					// element has now become. (#8809)
					ui.position = sortable.position;
				}
			} else {
				// If it doesn't intersect with the sortable, and it intersected before,
				// we fake the drag stop of the sortable, but make sure it doesn't remove
				// the helper by using cancelHelperRemoval.
				if ( sortable.isOver ) {

					sortable.isOver = 0;
					sortable.cancelHelperRemoval = true;

					// Calling sortable's mouseStop would trigger a revert,
					// so revert must be temporarily false until after mouseStop is called.
					sortable.options._revert = sortable.options.revert;
					sortable.options.revert = false;

					sortable._trigger( "out", event, sortable._uiHash( sortable ) );
					sortable._mouseStop( event, true );

					// restore sortable behaviors that were modfied
					// when the draggable entered the sortable area (#9481)
					sortable.options.revert = sortable.options._revert;
					sortable.options.helper = sortable.options._helper;

					if ( sortable.placeholder ) {
						sortable.placeholder.remove();
					}

					// Restore and recalculate the draggable's offset considering the sortable
					// may have modified them in unexpected ways. (#8809, #10669)
					ui.helper.appendTo( draggable._parent );
					draggable._refreshOffsets( event );
					ui.position = draggable._generatePosition( event, true );

					draggable._trigger( "fromSortable", event );

					// Inform draggable that the helper is no longer in a valid drop zone
					draggable.dropped = false;

					// Need to refreshPositions of all sortables just in case removing
					// from one sortable changes the location of other sortables (#9675)
					$.each( draggable.sortables, function() {
						this.refreshPositions();
					});
				}
			}
		});
	}
});

$.ui.plugin.add("draggable", "cursor", {
	start: function( event, ui, instance ) {
		var t = $( "body" ),
			o = instance.options;

		if (t.css("cursor")) {
			o._cursor = t.css("cursor");
		}
		t.css("cursor", o.cursor);
	},
	stop: function( event, ui, instance ) {
		var o = instance.options;
		if (o._cursor) {
			$("body").css("cursor", o._cursor);
		}
	}
});

$.ui.plugin.add("draggable", "opacity", {
	start: function( event, ui, instance ) {
		var t = $( ui.helper ),
			o = instance.options;
		if (t.css("opacity")) {
			o._opacity = t.css("opacity");
		}
		t.css("opacity", o.opacity);
	},
	stop: function( event, ui, instance ) {
		var o = instance.options;
		if (o._opacity) {
			$(ui.helper).css("opacity", o._opacity);
		}
	}
});

$.ui.plugin.add("draggable", "scroll", {
	start: function( event, ui, i ) {
		if ( !i.scrollParentNotHidden ) {
			i.scrollParentNotHidden = i.helper.scrollParent( false );
		}

		if ( i.scrollParentNotHidden[ 0 ] !== i.document[ 0 ] && i.scrollParentNotHidden[ 0 ].tagName !== "HTML" ) {
			i.overflowOffset = i.scrollParentNotHidden.offset();
		}
	},
	drag: function( event, ui, i  ) {

		var o = i.options,
			scrolled = false,
			scrollParent = i.scrollParentNotHidden[ 0 ],
			document = i.document[ 0 ];

		if ( scrollParent !== document && scrollParent.tagName !== "HTML" ) {
			if ( !o.axis || o.axis !== "x" ) {
				if ( ( i.overflowOffset.top + scrollParent.offsetHeight ) - event.pageY < o.scrollSensitivity ) {
					scrollParent.scrollTop = scrolled = scrollParent.scrollTop + o.scrollSpeed;
				} else if ( event.pageY - i.overflowOffset.top < o.scrollSensitivity ) {
					scrollParent.scrollTop = scrolled = scrollParent.scrollTop - o.scrollSpeed;
				}
			}

			if ( !o.axis || o.axis !== "y" ) {
				if ( ( i.overflowOffset.left + scrollParent.offsetWidth ) - event.pageX < o.scrollSensitivity ) {
					scrollParent.scrollLeft = scrolled = scrollParent.scrollLeft + o.scrollSpeed;
				} else if ( event.pageX - i.overflowOffset.left < o.scrollSensitivity ) {
					scrollParent.scrollLeft = scrolled = scrollParent.scrollLeft - o.scrollSpeed;
				}
			}

		} else {

			if (!o.axis || o.axis !== "x") {
				if (event.pageY - $(document).scrollTop() < o.scrollSensitivity) {
					scrolled = $(document).scrollTop($(document).scrollTop() - o.scrollSpeed);
				} else if ($(window).height() - (event.pageY - $(document).scrollTop()) < o.scrollSensitivity) {
					scrolled = $(document).scrollTop($(document).scrollTop() + o.scrollSpeed);
				}
			}

			if (!o.axis || o.axis !== "y") {
				if (event.pageX - $(document).scrollLeft() < o.scrollSensitivity) {
					scrolled = $(document).scrollLeft($(document).scrollLeft() - o.scrollSpeed);
				} else if ($(window).width() - (event.pageX - $(document).scrollLeft()) < o.scrollSensitivity) {
					scrolled = $(document).scrollLeft($(document).scrollLeft() + o.scrollSpeed);
				}
			}

		}

		if (scrolled !== false && $.ui.ddmanager && !o.dropBehaviour) {
			$.ui.ddmanager.prepareOffsets(i, event);
		}

	}
});

$.ui.plugin.add("draggable", "snap", {
	start: function( event, ui, i ) {

		var o = i.options;

		i.snapElements = [];

		$(o.snap.constructor !== String ? ( o.snap.items || ":data(ui-draggable)" ) : o.snap).each(function() {
			var $t = $(this),
				$o = $t.offset();
			if (this !== i.element[0]) {
				i.snapElements.push({
					item: this,
					width: $t.outerWidth(), height: $t.outerHeight(),
					top: $o.top, left: $o.left
				});
			}
		});

	},
	drag: function( event, ui, inst ) {

		var ts, bs, ls, rs, l, r, t, b, i, first,
			o = inst.options,
			d = o.snapTolerance,
			x1 = ui.offset.left, x2 = x1 + inst.helperProportions.width,
			y1 = ui.offset.top, y2 = y1 + inst.helperProportions.height;

		for (i = inst.snapElements.length - 1; i >= 0; i--){

			l = inst.snapElements[i].left - inst.margins.left;
			r = l + inst.snapElements[i].width;
			t = inst.snapElements[i].top - inst.margins.top;
			b = t + inst.snapElements[i].height;

			if ( x2 < l - d || x1 > r + d || y2 < t - d || y1 > b + d || !$.contains( inst.snapElements[ i ].item.ownerDocument, inst.snapElements[ i ].item ) ) {
				if (inst.snapElements[i].snapping) {
					(inst.options.snap.release && inst.options.snap.release.call(inst.element, event, $.extend(inst._uiHash(), { snapItem: inst.snapElements[i].item })));
				}
				inst.snapElements[i].snapping = false;
				continue;
			}

			if (o.snapMode !== "inner") {
				ts = Math.abs(t - y2) <= d;
				bs = Math.abs(b - y1) <= d;
				ls = Math.abs(l - x2) <= d;
				rs = Math.abs(r - x1) <= d;
				if (ts) {
					ui.position.top = inst._convertPositionTo("relative", { top: t - inst.helperProportions.height, left: 0 }).top;
				}
				if (bs) {
					ui.position.top = inst._convertPositionTo("relative", { top: b, left: 0 }).top;
				}
				if (ls) {
					ui.position.left = inst._convertPositionTo("relative", { top: 0, left: l - inst.helperProportions.width }).left;
				}
				if (rs) {
					ui.position.left = inst._convertPositionTo("relative", { top: 0, left: r }).left;
				}
			}

			first = (ts || bs || ls || rs);

			if (o.snapMode !== "outer") {
				ts = Math.abs(t - y1) <= d;
				bs = Math.abs(b - y2) <= d;
				ls = Math.abs(l - x1) <= d;
				rs = Math.abs(r - x2) <= d;
				if (ts) {
					ui.position.top = inst._convertPositionTo("relative", { top: t, left: 0 }).top;
				}
				if (bs) {
					ui.position.top = inst._convertPositionTo("relative", { top: b - inst.helperProportions.height, left: 0 }).top;
				}
				if (ls) {
					ui.position.left = inst._convertPositionTo("relative", { top: 0, left: l }).left;
				}
				if (rs) {
					ui.position.left = inst._convertPositionTo("relative", { top: 0, left: r - inst.helperProportions.width }).left;
				}
			}

			if (!inst.snapElements[i].snapping && (ts || bs || ls || rs || first)) {
				(inst.options.snap.snap && inst.options.snap.snap.call(inst.element, event, $.extend(inst._uiHash(), { snapItem: inst.snapElements[i].item })));
			}
			inst.snapElements[i].snapping = (ts || bs || ls || rs || first);

		}

	}
});

$.ui.plugin.add("draggable", "stack", {
	start: function( event, ui, instance ) {
		var min,
			o = instance.options,
			group = $.makeArray($(o.stack)).sort(function(a, b) {
				return (parseInt($(a).css("zIndex"), 10) || 0) - (parseInt($(b).css("zIndex"), 10) || 0);
			});

		if (!group.length) { return; }

		min = parseInt($(group[0]).css("zIndex"), 10) || 0;
		$(group).each(function(i) {
			$(this).css("zIndex", min + i);
		});
		this.css("zIndex", (min + group.length));
	}
});

$.ui.plugin.add("draggable", "zIndex", {
	start: function( event, ui, instance ) {
		var t = $( ui.helper ),
			o = instance.options;

		if (t.css("zIndex")) {
			o._zIndex = t.css("zIndex");
		}
		t.css("zIndex", o.zIndex);
	},
	stop: function( event, ui, instance ) {
		var o = instance.options;

		if (o._zIndex) {
			$(ui.helper).css("zIndex", o._zIndex);
		}
	}
});

var draggable = $.ui.draggable;


/*!
 * jQuery UI Droppable 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/droppable/
 */


$.widget( "ui.droppable", {
	version: "1.11.4",
	widgetEventPrefix: "drop",
	options: {
		accept: "*",
		activeClass: false,
		addClasses: true,
		greedy: false,
		hoverClass: false,
		scope: "default",
		tolerance: "intersect",

		// callbacks
		activate: null,
		deactivate: null,
		drop: null,
		out: null,
		over: null
	},
	_create: function() {

		var proportions,
			o = this.options,
			accept = o.accept;

		this.isover = false;
		this.isout = true;

		this.accept = $.isFunction( accept ) ? accept : function( d ) {
			return d.is( accept );
		};

		this.proportions = function( /* valueToWrite */ ) {
			if ( arguments.length ) {
				// Store the droppable's proportions
				proportions = arguments[ 0 ];
			} else {
				// Retrieve or derive the droppable's proportions
				return proportions ?
					proportions :
					proportions = {
						width: this.element[ 0 ].offsetWidth,
						height: this.element[ 0 ].offsetHeight
					};
			}
		};

		this._addToManager( o.scope );

		o.addClasses && this.element.addClass( "ui-droppable" );

	},

	_addToManager: function( scope ) {
		// Add the reference and positions to the manager
		$.ui.ddmanager.droppables[ scope ] = $.ui.ddmanager.droppables[ scope ] || [];
		$.ui.ddmanager.droppables[ scope ].push( this );
	},

	_splice: function( drop ) {
		var i = 0;
		for ( ; i < drop.length; i++ ) {
			if ( drop[ i ] === this ) {
				drop.splice( i, 1 );
			}
		}
	},

	_destroy: function() {
		var drop = $.ui.ddmanager.droppables[ this.options.scope ];

		this._splice( drop );

		this.element.removeClass( "ui-droppable ui-droppable-disabled" );
	},

	_setOption: function( key, value ) {

		if ( key === "accept" ) {
			this.accept = $.isFunction( value ) ? value : function( d ) {
				return d.is( value );
			};
		} else if ( key === "scope" ) {
			var drop = $.ui.ddmanager.droppables[ this.options.scope ];

			this._splice( drop );
			this._addToManager( value );
		}

		this._super( key, value );
	},

	_activate: function( event ) {
		var draggable = $.ui.ddmanager.current;
		if ( this.options.activeClass ) {
			this.element.addClass( this.options.activeClass );
		}
		if ( draggable ){
			this._trigger( "activate", event, this.ui( draggable ) );
		}
	},

	_deactivate: function( event ) {
		var draggable = $.ui.ddmanager.current;
		if ( this.options.activeClass ) {
			this.element.removeClass( this.options.activeClass );
		}
		if ( draggable ){
			this._trigger( "deactivate", event, this.ui( draggable ) );
		}
	},

	_over: function( event ) {

		var draggable = $.ui.ddmanager.current;

		// Bail if draggable and droppable are same element
		if ( !draggable || ( draggable.currentItem || draggable.element )[ 0 ] === this.element[ 0 ] ) {
			return;
		}

		if ( this.accept.call( this.element[ 0 ], ( draggable.currentItem || draggable.element ) ) ) {
			if ( this.options.hoverClass ) {
				this.element.addClass( this.options.hoverClass );
			}
			this._trigger( "over", event, this.ui( draggable ) );
		}

	},

	_out: function( event ) {

		var draggable = $.ui.ddmanager.current;

		// Bail if draggable and droppable are same element
		if ( !draggable || ( draggable.currentItem || draggable.element )[ 0 ] === this.element[ 0 ] ) {
			return;
		}

		if ( this.accept.call( this.element[ 0 ], ( draggable.currentItem || draggable.element ) ) ) {
			if ( this.options.hoverClass ) {
				this.element.removeClass( this.options.hoverClass );
			}
			this._trigger( "out", event, this.ui( draggable ) );
		}

	},

	_drop: function( event, custom ) {

		var draggable = custom || $.ui.ddmanager.current,
			childrenIntersection = false;

		// Bail if draggable and droppable are same element
		if ( !draggable || ( draggable.currentItem || draggable.element )[ 0 ] === this.element[ 0 ] ) {
			return false;
		}

		this.element.find( ":data(ui-droppable)" ).not( ".ui-draggable-dragging" ).each(function() {
			var inst = $( this ).droppable( "instance" );
			if (
				inst.options.greedy &&
				!inst.options.disabled &&
				inst.options.scope === draggable.options.scope &&
				inst.accept.call( inst.element[ 0 ], ( draggable.currentItem || draggable.element ) ) &&
				$.ui.intersect( draggable, $.extend( inst, { offset: inst.element.offset() } ), inst.options.tolerance, event )
			) { childrenIntersection = true; return false; }
		});
		if ( childrenIntersection ) {
			return false;
		}

		if ( this.accept.call( this.element[ 0 ], ( draggable.currentItem || draggable.element ) ) ) {
			if ( this.options.activeClass ) {
				this.element.removeClass( this.options.activeClass );
			}
			if ( this.options.hoverClass ) {
				this.element.removeClass( this.options.hoverClass );
			}
			this._trigger( "drop", event, this.ui( draggable ) );
			return this.element;
		}

		return false;

	},

	ui: function( c ) {
		return {
			draggable: ( c.currentItem || c.element ),
			helper: c.helper,
			position: c.position,
			offset: c.positionAbs
		};
	}

});

$.ui.intersect = (function() {
	function isOverAxis( x, reference, size ) {
		return ( x >= reference ) && ( x < ( reference + size ) );
	}

	return function( draggable, droppable, toleranceMode, event ) {

		if ( !droppable.offset ) {
			return false;
		}

		var x1 = ( draggable.positionAbs || draggable.position.absolute ).left + draggable.margins.left,
			y1 = ( draggable.positionAbs || draggable.position.absolute ).top + draggable.margins.top,
			x2 = x1 + draggable.helperProportions.width,
			y2 = y1 + draggable.helperProportions.height,
			l = droppable.offset.left,
			t = droppable.offset.top,
			r = l + droppable.proportions().width,
			b = t + droppable.proportions().height;

		switch ( toleranceMode ) {
		case "fit":
			return ( l <= x1 && x2 <= r && t <= y1 && y2 <= b );
		case "intersect":
			return ( l < x1 + ( draggable.helperProportions.width / 2 ) && // Right Half
				x2 - ( draggable.helperProportions.width / 2 ) < r && // Left Half
				t < y1 + ( draggable.helperProportions.height / 2 ) && // Bottom Half
				y2 - ( draggable.helperProportions.height / 2 ) < b ); // Top Half
		case "pointer":
			return isOverAxis( event.pageY, t, droppable.proportions().height ) && isOverAxis( event.pageX, l, droppable.proportions().width );
		case "touch":
			return (
				( y1 >= t && y1 <= b ) || // Top edge touching
				( y2 >= t && y2 <= b ) || // Bottom edge touching
				( y1 < t && y2 > b ) // Surrounded vertically
			) && (
				( x1 >= l && x1 <= r ) || // Left edge touching
				( x2 >= l && x2 <= r ) || // Right edge touching
				( x1 < l && x2 > r ) // Surrounded horizontally
			);
		default:
			return false;
		}
	};
})();

/*
	This manager tracks offsets of draggables and droppables
*/
$.ui.ddmanager = {
	current: null,
	droppables: { "default": [] },
	prepareOffsets: function( t, event ) {

		var i, j,
			m = $.ui.ddmanager.droppables[ t.options.scope ] || [],
			type = event ? event.type : null, // workaround for #2317
			list = ( t.currentItem || t.element ).find( ":data(ui-droppable)" ).addBack();

		droppablesLoop: for ( i = 0; i < m.length; i++ ) {

			// No disabled and non-accepted
			if ( m[ i ].options.disabled || ( t && !m[ i ].accept.call( m[ i ].element[ 0 ], ( t.currentItem || t.element ) ) ) ) {
				continue;
			}

			// Filter out elements in the current dragged item
			for ( j = 0; j < list.length; j++ ) {
				if ( list[ j ] === m[ i ].element[ 0 ] ) {
					m[ i ].proportions().height = 0;
					continue droppablesLoop;
				}
			}

			m[ i ].visible = m[ i ].element.css( "display" ) !== "none";
			if ( !m[ i ].visible ) {
				continue;
			}

			// Activate the droppable if used directly from draggables
			if ( type === "mousedown" ) {
				m[ i ]._activate.call( m[ i ], event );
			}

			m[ i ].offset = m[ i ].element.offset();
			m[ i ].proportions({ width: m[ i ].element[ 0 ].offsetWidth, height: m[ i ].element[ 0 ].offsetHeight });

		}

	},
	drop: function( draggable, event ) {

		var dropped = false;
		// Create a copy of the droppables in case the list changes during the drop (#9116)
		$.each( ( $.ui.ddmanager.droppables[ draggable.options.scope ] || [] ).slice(), function() {

			if ( !this.options ) {
				return;
			}
			if ( !this.options.disabled && this.visible && $.ui.intersect( draggable, this, this.options.tolerance, event ) ) {
				dropped = this._drop.call( this, event ) || dropped;
			}

			if ( !this.options.disabled && this.visible && this.accept.call( this.element[ 0 ], ( draggable.currentItem || draggable.element ) ) ) {
				this.isout = true;
				this.isover = false;
				this._deactivate.call( this, event );
			}

		});
		return dropped;

	},
	dragStart: function( draggable, event ) {
		// Listen for scrolling so that if the dragging causes scrolling the position of the droppables can be recalculated (see #5003)
		draggable.element.parentsUntil( "body" ).bind( "scroll.droppable", function() {
			if ( !draggable.options.refreshPositions ) {
				$.ui.ddmanager.prepareOffsets( draggable, event );
			}
		});
	},
	drag: function( draggable, event ) {

		// If you have a highly dynamic page, you might try this option. It renders positions every time you move the mouse.
		if ( draggable.options.refreshPositions ) {
			$.ui.ddmanager.prepareOffsets( draggable, event );
		}

		// Run through all droppables and check their positions based on specific tolerance options
		$.each( $.ui.ddmanager.droppables[ draggable.options.scope ] || [], function() {

			if ( this.options.disabled || this.greedyChild || !this.visible ) {
				return;
			}

			var parentInstance, scope, parent,
				intersects = $.ui.intersect( draggable, this, this.options.tolerance, event ),
				c = !intersects && this.isover ? "isout" : ( intersects && !this.isover ? "isover" : null );
			if ( !c ) {
				return;
			}

			if ( this.options.greedy ) {
				// find droppable parents with same scope
				scope = this.options.scope;
				parent = this.element.parents( ":data(ui-droppable)" ).filter(function() {
					return $( this ).droppable( "instance" ).options.scope === scope;
				});

				if ( parent.length ) {
					parentInstance = $( parent[ 0 ] ).droppable( "instance" );
					parentInstance.greedyChild = ( c === "isover" );
				}
			}

			// we just moved into a greedy child
			if ( parentInstance && c === "isover" ) {
				parentInstance.isover = false;
				parentInstance.isout = true;
				parentInstance._out.call( parentInstance, event );
			}

			this[ c ] = true;
			this[c === "isout" ? "isover" : "isout"] = false;
			this[c === "isover" ? "_over" : "_out"].call( this, event );

			// we just moved out of a greedy child
			if ( parentInstance && c === "isout" ) {
				parentInstance.isout = false;
				parentInstance.isover = true;
				parentInstance._over.call( parentInstance, event );
			}
		});

	},
	dragStop: function( draggable, event ) {
		draggable.element.parentsUntil( "body" ).unbind( "scroll.droppable" );
		// Call prepareOffsets one final time since IE does not fire return scroll events when overflow was caused by drag (see #5003)
		if ( !draggable.options.refreshPositions ) {
			$.ui.ddmanager.prepareOffsets( draggable, event );
		}
	}
};

var droppable = $.ui.droppable;


/*!
 * jQuery UI Resizable 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/resizable/
 */


$.widget("ui.resizable", $.ui.mouse, {
	version: "1.11.4",
	widgetEventPrefix: "resize",
	options: {
		alsoResize: false,
		animate: false,
		animateDuration: "slow",
		animateEasing: "swing",
		aspectRatio: false,
		autoHide: false,
		containment: false,
		ghost: false,
		grid: false,
		handles: "e,s,se",
		helper: false,
		maxHeight: null,
		maxWidth: null,
		minHeight: 10,
		minWidth: 10,
		// See #7960
		zIndex: 90,

		// callbacks
		resize: null,
		start: null,
		stop: null
	},

	_num: function( value ) {
		return parseInt( value, 10 ) || 0;
	},

	_isNumber: function( value ) {
		return !isNaN( parseInt( value, 10 ) );
	},

	_hasScroll: function( el, a ) {

		if ( $( el ).css( "overflow" ) === "hidden") {
			return false;
		}

		var scroll = ( a && a === "left" ) ? "scrollLeft" : "scrollTop",
			has = false;

		if ( el[ scroll ] > 0 ) {
			return true;
		}

		// TODO: determine which cases actually cause this to happen
		// if the element doesn't have the scroll set, see if it's possible to
		// set the scroll
		el[ scroll ] = 1;
		has = ( el[ scroll ] > 0 );
		el[ scroll ] = 0;
		return has;
	},

	_create: function() {

		var n, i, handle, axis, hname,
			that = this,
			o = this.options;
		this.element.addClass("ui-resizable");

		$.extend(this, {
			_aspectRatio: !!(o.aspectRatio),
			aspectRatio: o.aspectRatio,
			originalElement: this.element,
			_proportionallyResizeElements: [],
			_helper: o.helper || o.ghost || o.animate ? o.helper || "ui-resizable-helper" : null
		});

		// Wrap the element if it cannot hold child nodes
		if (this.element[0].nodeName.match(/^(canvas|textarea|input|select|button|img)$/i)) {

			this.element.wrap(
				$("<div class='ui-wrapper' style='overflow: hidden;'></div>").css({
					position: this.element.css("position"),
					width: this.element.outerWidth(),
					height: this.element.outerHeight(),
					top: this.element.css("top"),
					left: this.element.css("left")
				})
			);

			this.element = this.element.parent().data(
				"ui-resizable", this.element.resizable( "instance" )
			);

			this.elementIsWrapper = true;

			this.element.css({
				marginLeft: this.originalElement.css("marginLeft"),
				marginTop: this.originalElement.css("marginTop"),
				marginRight: this.originalElement.css("marginRight"),
				marginBottom: this.originalElement.css("marginBottom")
			});
			this.originalElement.css({
				marginLeft: 0,
				marginTop: 0,
				marginRight: 0,
				marginBottom: 0
			});
			// support: Safari
			// Prevent Safari textarea resize
			this.originalResizeStyle = this.originalElement.css("resize");
			this.originalElement.css("resize", "none");

			this._proportionallyResizeElements.push( this.originalElement.css({
				position: "static",
				zoom: 1,
				display: "block"
			}) );

			// support: IE9
			// avoid IE jump (hard set the margin)
			this.originalElement.css({ margin: this.originalElement.css("margin") });

			this._proportionallyResize();
		}

		this.handles = o.handles ||
			( !$(".ui-resizable-handle", this.element).length ?
				"e,s,se" : {
					n: ".ui-resizable-n",
					e: ".ui-resizable-e",
					s: ".ui-resizable-s",
					w: ".ui-resizable-w",
					se: ".ui-resizable-se",
					sw: ".ui-resizable-sw",
					ne: ".ui-resizable-ne",
					nw: ".ui-resizable-nw"
				} );

		this._handles = $();
		if ( this.handles.constructor === String ) {

			if ( this.handles === "all") {
				this.handles = "n,e,s,w,se,sw,ne,nw";
			}

			n = this.handles.split(",");
			this.handles = {};

			for (i = 0; i < n.length; i++) {

				handle = $.trim(n[i]);
				hname = "ui-resizable-" + handle;
				axis = $("<div class='ui-resizable-handle " + hname + "'></div>");

				axis.css({ zIndex: o.zIndex });

				// TODO : What's going on here?
				if ("se" === handle) {
					axis.addClass("ui-icon ui-icon-gripsmall-diagonal-se");
				}

				this.handles[handle] = ".ui-resizable-" + handle;
				this.element.append(axis);
			}

		}

		this._renderAxis = function(target) {

			var i, axis, padPos, padWrapper;

			target = target || this.element;

			for (i in this.handles) {

				if (this.handles[i].constructor === String) {
					this.handles[i] = this.element.children( this.handles[ i ] ).first().show();
				} else if ( this.handles[ i ].jquery || this.handles[ i ].nodeType ) {
					this.handles[ i ] = $( this.handles[ i ] );
					this._on( this.handles[ i ], { "mousedown": that._mouseDown });
				}

				if (this.elementIsWrapper && this.originalElement[0].nodeName.match(/^(textarea|input|select|button)$/i)) {

					axis = $(this.handles[i], this.element);

					padWrapper = /sw|ne|nw|se|n|s/.test(i) ? axis.outerHeight() : axis.outerWidth();

					padPos = [ "padding",
						/ne|nw|n/.test(i) ? "Top" :
						/se|sw|s/.test(i) ? "Bottom" :
						/^e$/.test(i) ? "Right" : "Left" ].join("");

					target.css(padPos, padWrapper);

					this._proportionallyResize();
				}

				this._handles = this._handles.add( this.handles[ i ] );
			}
		};

		// TODO: make renderAxis a prototype function
		this._renderAxis(this.element);

		this._handles = this._handles.add( this.element.find( ".ui-resizable-handle" ) );
		this._handles.disableSelection();

		this._handles.mouseover(function() {
			if (!that.resizing) {
				if (this.className) {
					axis = this.className.match(/ui-resizable-(se|sw|ne|nw|n|e|s|w)/i);
				}
				that.axis = axis && axis[1] ? axis[1] : "se";
			}
		});

		if (o.autoHide) {
			this._handles.hide();
			$(this.element)
				.addClass("ui-resizable-autohide")
				.mouseenter(function() {
					if (o.disabled) {
						return;
					}
					$(this).removeClass("ui-resizable-autohide");
					that._handles.show();
				})
				.mouseleave(function() {
					if (o.disabled) {
						return;
					}
					if (!that.resizing) {
						$(this).addClass("ui-resizable-autohide");
						that._handles.hide();
					}
				});
		}

		this._mouseInit();
	},

	_destroy: function() {

		this._mouseDestroy();

		var wrapper,
			_destroy = function(exp) {
				$(exp)
					.removeClass("ui-resizable ui-resizable-disabled ui-resizable-resizing")
					.removeData("resizable")
					.removeData("ui-resizable")
					.unbind(".resizable")
					.find(".ui-resizable-handle")
						.remove();
			};

		// TODO: Unwrap at same DOM position
		if (this.elementIsWrapper) {
			_destroy(this.element);
			wrapper = this.element;
			this.originalElement.css({
				position: wrapper.css("position"),
				width: wrapper.outerWidth(),
				height: wrapper.outerHeight(),
				top: wrapper.css("top"),
				left: wrapper.css("left")
			}).insertAfter( wrapper );
			wrapper.remove();
		}

		this.originalElement.css("resize", this.originalResizeStyle);
		_destroy(this.originalElement);

		return this;
	},

	_mouseCapture: function(event) {
		var i, handle,
			capture = false;

		for (i in this.handles) {
			handle = $(this.handles[i])[0];
			if (handle === event.target || $.contains(handle, event.target)) {
				capture = true;
			}
		}

		return !this.options.disabled && capture;
	},

	_mouseStart: function(event) {

		var curleft, curtop, cursor,
			o = this.options,
			el = this.element;

		this.resizing = true;

		this._renderProxy();

		curleft = this._num(this.helper.css("left"));
		curtop = this._num(this.helper.css("top"));

		if (o.containment) {
			curleft += $(o.containment).scrollLeft() || 0;
			curtop += $(o.containment).scrollTop() || 0;
		}

		this.offset = this.helper.offset();
		this.position = { left: curleft, top: curtop };

		this.size = this._helper ? {
				width: this.helper.width(),
				height: this.helper.height()
			} : {
				width: el.width(),
				height: el.height()
			};

		this.originalSize = this._helper ? {
				width: el.outerWidth(),
				height: el.outerHeight()
			} : {
				width: el.width(),
				height: el.height()
			};

		this.sizeDiff = {
			width: el.outerWidth() - el.width(),
			height: el.outerHeight() - el.height()
		};

		this.originalPosition = { left: curleft, top: curtop };
		this.originalMousePosition = { left: event.pageX, top: event.pageY };

		this.aspectRatio = (typeof o.aspectRatio === "number") ?
			o.aspectRatio :
			((this.originalSize.width / this.originalSize.height) || 1);

		cursor = $(".ui-resizable-" + this.axis).css("cursor");
		$("body").css("cursor", cursor === "auto" ? this.axis + "-resize" : cursor);

		el.addClass("ui-resizable-resizing");
		this._propagate("start", event);
		return true;
	},

	_mouseDrag: function(event) {

		var data, props,
			smp = this.originalMousePosition,
			a = this.axis,
			dx = (event.pageX - smp.left) || 0,
			dy = (event.pageY - smp.top) || 0,
			trigger = this._change[a];

		this._updatePrevProperties();

		if (!trigger) {
			return false;
		}

		data = trigger.apply(this, [ event, dx, dy ]);

		this._updateVirtualBoundaries(event.shiftKey);
		if (this._aspectRatio || event.shiftKey) {
			data = this._updateRatio(data, event);
		}

		data = this._respectSize(data, event);

		this._updateCache(data);

		this._propagate("resize", event);

		props = this._applyChanges();

		if ( !this._helper && this._proportionallyResizeElements.length ) {
			this._proportionallyResize();
		}

		if ( !$.isEmptyObject( props ) ) {
			this._updatePrevProperties();
			this._trigger( "resize", event, this.ui() );
			this._applyChanges();
		}

		return false;
	},

	_mouseStop: function(event) {

		this.resizing = false;
		var pr, ista, soffseth, soffsetw, s, left, top,
			o = this.options, that = this;

		if (this._helper) {

			pr = this._proportionallyResizeElements;
			ista = pr.length && (/textarea/i).test(pr[0].nodeName);
			soffseth = ista && this._hasScroll(pr[0], "left") ? 0 : that.sizeDiff.height;
			soffsetw = ista ? 0 : that.sizeDiff.width;

			s = {
				width: (that.helper.width()  - soffsetw),
				height: (that.helper.height() - soffseth)
			};
			left = (parseInt(that.element.css("left"), 10) +
				(that.position.left - that.originalPosition.left)) || null;
			top = (parseInt(that.element.css("top"), 10) +
				(that.position.top - that.originalPosition.top)) || null;

			if (!o.animate) {
				this.element.css($.extend(s, { top: top, left: left }));
			}

			that.helper.height(that.size.height);
			that.helper.width(that.size.width);

			if (this._helper && !o.animate) {
				this._proportionallyResize();
			}
		}

		$("body").css("cursor", "auto");

		this.element.removeClass("ui-resizable-resizing");

		this._propagate("stop", event);

		if (this._helper) {
			this.helper.remove();
		}

		return false;

	},

	_updatePrevProperties: function() {
		this.prevPosition = {
			top: this.position.top,
			left: this.position.left
		};
		this.prevSize = {
			width: this.size.width,
			height: this.size.height
		};
	},

	_applyChanges: function() {
		var props = {};

		if ( this.position.top !== this.prevPosition.top ) {
			props.top = this.position.top + "px";
		}
		if ( this.position.left !== this.prevPosition.left ) {
			props.left = this.position.left + "px";
		}
		if ( this.size.width !== this.prevSize.width ) {
			props.width = this.size.width + "px";
		}
		if ( this.size.height !== this.prevSize.height ) {
			props.height = this.size.height + "px";
		}

		this.helper.css( props );

		return props;
	},

	_updateVirtualBoundaries: function(forceAspectRatio) {
		var pMinWidth, pMaxWidth, pMinHeight, pMaxHeight, b,
			o = this.options;

		b = {
			minWidth: this._isNumber(o.minWidth) ? o.minWidth : 0,
			maxWidth: this._isNumber(o.maxWidth) ? o.maxWidth : Infinity,
			minHeight: this._isNumber(o.minHeight) ? o.minHeight : 0,
			maxHeight: this._isNumber(o.maxHeight) ? o.maxHeight : Infinity
		};

		if (this._aspectRatio || forceAspectRatio) {
			pMinWidth = b.minHeight * this.aspectRatio;
			pMinHeight = b.minWidth / this.aspectRatio;
			pMaxWidth = b.maxHeight * this.aspectRatio;
			pMaxHeight = b.maxWidth / this.aspectRatio;

			if (pMinWidth > b.minWidth) {
				b.minWidth = pMinWidth;
			}
			if (pMinHeight > b.minHeight) {
				b.minHeight = pMinHeight;
			}
			if (pMaxWidth < b.maxWidth) {
				b.maxWidth = pMaxWidth;
			}
			if (pMaxHeight < b.maxHeight) {
				b.maxHeight = pMaxHeight;
			}
		}
		this._vBoundaries = b;
	},

	_updateCache: function(data) {
		this.offset = this.helper.offset();
		if (this._isNumber(data.left)) {
			this.position.left = data.left;
		}
		if (this._isNumber(data.top)) {
			this.position.top = data.top;
		}
		if (this._isNumber(data.height)) {
			this.size.height = data.height;
		}
		if (this._isNumber(data.width)) {
			this.size.width = data.width;
		}
	},

	_updateRatio: function( data ) {

		var cpos = this.position,
			csize = this.size,
			a = this.axis;

		if (this._isNumber(data.height)) {
			data.width = (data.height * this.aspectRatio);
		} else if (this._isNumber(data.width)) {
			data.height = (data.width / this.aspectRatio);
		}

		if (a === "sw") {
			data.left = cpos.left + (csize.width - data.width);
			data.top = null;
		}
		if (a === "nw") {
			data.top = cpos.top + (csize.height - data.height);
			data.left = cpos.left + (csize.width - data.width);
		}

		return data;
	},

	_respectSize: function( data ) {

		var o = this._vBoundaries,
			a = this.axis,
			ismaxw = this._isNumber(data.width) && o.maxWidth && (o.maxWidth < data.width),
			ismaxh = this._isNumber(data.height) && o.maxHeight && (o.maxHeight < data.height),
			isminw = this._isNumber(data.width) && o.minWidth && (o.minWidth > data.width),
			isminh = this._isNumber(data.height) && o.minHeight && (o.minHeight > data.height),
			dw = this.originalPosition.left + this.originalSize.width,
			dh = this.position.top + this.size.height,
			cw = /sw|nw|w/.test(a), ch = /nw|ne|n/.test(a);
		if (isminw) {
			data.width = o.minWidth;
		}
		if (isminh) {
			data.height = o.minHeight;
		}
		if (ismaxw) {
			data.width = o.maxWidth;
		}
		if (ismaxh) {
			data.height = o.maxHeight;
		}

		if (isminw && cw) {
			data.left = dw - o.minWidth;
		}
		if (ismaxw && cw) {
			data.left = dw - o.maxWidth;
		}
		if (isminh && ch) {
			data.top = dh - o.minHeight;
		}
		if (ismaxh && ch) {
			data.top = dh - o.maxHeight;
		}

		// Fixing jump error on top/left - bug #2330
		if (!data.width && !data.height && !data.left && data.top) {
			data.top = null;
		} else if (!data.width && !data.height && !data.top && data.left) {
			data.left = null;
		}

		return data;
	},

	_getPaddingPlusBorderDimensions: function( element ) {
		var i = 0,
			widths = [],
			borders = [
				element.css( "borderTopWidth" ),
				element.css( "borderRightWidth" ),
				element.css( "borderBottomWidth" ),
				element.css( "borderLeftWidth" )
			],
			paddings = [
				element.css( "paddingTop" ),
				element.css( "paddingRight" ),
				element.css( "paddingBottom" ),
				element.css( "paddingLeft" )
			];

		for ( ; i < 4; i++ ) {
			widths[ i ] = ( parseInt( borders[ i ], 10 ) || 0 );
			widths[ i ] += ( parseInt( paddings[ i ], 10 ) || 0 );
		}

		return {
			height: widths[ 0 ] + widths[ 2 ],
			width: widths[ 1 ] + widths[ 3 ]
		};
	},

	_proportionallyResize: function() {

		if (!this._proportionallyResizeElements.length) {
			return;
		}

		var prel,
			i = 0,
			element = this.helper || this.element;

		for ( ; i < this._proportionallyResizeElements.length; i++) {

			prel = this._proportionallyResizeElements[i];

			// TODO: Seems like a bug to cache this.outerDimensions
			// considering that we are in a loop.
			if (!this.outerDimensions) {
				this.outerDimensions = this._getPaddingPlusBorderDimensions( prel );
			}

			prel.css({
				height: (element.height() - this.outerDimensions.height) || 0,
				width: (element.width() - this.outerDimensions.width) || 0
			});

		}

	},

	_renderProxy: function() {

		var el = this.element, o = this.options;
		this.elementOffset = el.offset();

		if (this._helper) {

			this.helper = this.helper || $("<div style='overflow:hidden;'></div>");

			this.helper.addClass(this._helper).css({
				width: this.element.outerWidth() - 1,
				height: this.element.outerHeight() - 1,
				position: "absolute",
				left: this.elementOffset.left + "px",
				top: this.elementOffset.top + "px",
				zIndex: ++o.zIndex //TODO: Don't modify option
			});

			this.helper
				.appendTo("body")
				.disableSelection();

		} else {
			this.helper = this.element;
		}

	},

	_change: {
		e: function(event, dx) {
			return { width: this.originalSize.width + dx };
		},
		w: function(event, dx) {
			var cs = this.originalSize, sp = this.originalPosition;
			return { left: sp.left + dx, width: cs.width - dx };
		},
		n: function(event, dx, dy) {
			var cs = this.originalSize, sp = this.originalPosition;
			return { top: sp.top + dy, height: cs.height - dy };
		},
		s: function(event, dx, dy) {
			return { height: this.originalSize.height + dy };
		},
		se: function(event, dx, dy) {
			return $.extend(this._change.s.apply(this, arguments),
				this._change.e.apply(this, [ event, dx, dy ]));
		},
		sw: function(event, dx, dy) {
			return $.extend(this._change.s.apply(this, arguments),
				this._change.w.apply(this, [ event, dx, dy ]));
		},
		ne: function(event, dx, dy) {
			return $.extend(this._change.n.apply(this, arguments),
				this._change.e.apply(this, [ event, dx, dy ]));
		},
		nw: function(event, dx, dy) {
			return $.extend(this._change.n.apply(this, arguments),
				this._change.w.apply(this, [ event, dx, dy ]));
		}
	},

	_propagate: function(n, event) {
		$.ui.plugin.call(this, n, [ event, this.ui() ]);
		(n !== "resize" && this._trigger(n, event, this.ui()));
	},

	plugins: {},

	ui: function() {
		return {
			originalElement: this.originalElement,
			element: this.element,
			helper: this.helper,
			position: this.position,
			size: this.size,
			originalSize: this.originalSize,
			originalPosition: this.originalPosition
		};
	}

});

/*
 * Resizable Extensions
 */

$.ui.plugin.add("resizable", "animate", {

	stop: function( event ) {
		var that = $(this).resizable( "instance" ),
			o = that.options,
			pr = that._proportionallyResizeElements,
			ista = pr.length && (/textarea/i).test(pr[0].nodeName),
			soffseth = ista && that._hasScroll(pr[0], "left") ? 0 : that.sizeDiff.height,
			soffsetw = ista ? 0 : that.sizeDiff.width,
			style = { width: (that.size.width - soffsetw), height: (that.size.height - soffseth) },
			left = (parseInt(that.element.css("left"), 10) +
				(that.position.left - that.originalPosition.left)) || null,
			top = (parseInt(that.element.css("top"), 10) +
				(that.position.top - that.originalPosition.top)) || null;

		that.element.animate(
			$.extend(style, top && left ? { top: top, left: left } : {}), {
				duration: o.animateDuration,
				easing: o.animateEasing,
				step: function() {

					var data = {
						width: parseInt(that.element.css("width"), 10),
						height: parseInt(that.element.css("height"), 10),
						top: parseInt(that.element.css("top"), 10),
						left: parseInt(that.element.css("left"), 10)
					};

					if (pr && pr.length) {
						$(pr[0]).css({ width: data.width, height: data.height });
					}

					// propagating resize, and updating values for each animation step
					that._updateCache(data);
					that._propagate("resize", event);

				}
			}
		);
	}

});

$.ui.plugin.add( "resizable", "containment", {

	start: function() {
		var element, p, co, ch, cw, width, height,
			that = $( this ).resizable( "instance" ),
			o = that.options,
			el = that.element,
			oc = o.containment,
			ce = ( oc instanceof $ ) ? oc.get( 0 ) : ( /parent/.test( oc ) ) ? el.parent().get( 0 ) : oc;

		if ( !ce ) {
			return;
		}

		that.containerElement = $( ce );

		if ( /document/.test( oc ) || oc === document ) {
			that.containerOffset = {
				left: 0,
				top: 0
			};
			that.containerPosition = {
				left: 0,
				top: 0
			};

			that.parentData = {
				element: $( document ),
				left: 0,
				top: 0,
				width: $( document ).width(),
				height: $( document ).height() || document.body.parentNode.scrollHeight
			};
		} else {
			element = $( ce );
			p = [];
			$([ "Top", "Right", "Left", "Bottom" ]).each(function( i, name ) {
				p[ i ] = that._num( element.css( "padding" + name ) );
			});

			that.containerOffset = element.offset();
			that.containerPosition = element.position();
			that.containerSize = {
				height: ( element.innerHeight() - p[ 3 ] ),
				width: ( element.innerWidth() - p[ 1 ] )
			};

			co = that.containerOffset;
			ch = that.containerSize.height;
			cw = that.containerSize.width;
			width = ( that._hasScroll ( ce, "left" ) ? ce.scrollWidth : cw );
			height = ( that._hasScroll ( ce ) ? ce.scrollHeight : ch ) ;

			that.parentData = {
				element: ce,
				left: co.left,
				top: co.top,
				width: width,
				height: height
			};
		}
	},

	resize: function( event ) {
		var woset, hoset, isParent, isOffsetRelative,
			that = $( this ).resizable( "instance" ),
			o = that.options,
			co = that.containerOffset,
			cp = that.position,
			pRatio = that._aspectRatio || event.shiftKey,
			cop = {
				top: 0,
				left: 0
			},
			ce = that.containerElement,
			continueResize = true;

		if ( ce[ 0 ] !== document && ( /static/ ).test( ce.css( "position" ) ) ) {
			cop = co;
		}

		if ( cp.left < ( that._helper ? co.left : 0 ) ) {
			that.size.width = that.size.width +
				( that._helper ?
					( that.position.left - co.left ) :
					( that.position.left - cop.left ) );

			if ( pRatio ) {
				that.size.height = that.size.width / that.aspectRatio;
				continueResize = false;
			}
			that.position.left = o.helper ? co.left : 0;
		}

		if ( cp.top < ( that._helper ? co.top : 0 ) ) {
			that.size.height = that.size.height +
				( that._helper ?
					( that.position.top - co.top ) :
					that.position.top );

			if ( pRatio ) {
				that.size.width = that.size.height * that.aspectRatio;
				continueResize = false;
			}
			that.position.top = that._helper ? co.top : 0;
		}

		isParent = that.containerElement.get( 0 ) === that.element.parent().get( 0 );
		isOffsetRelative = /relative|absolute/.test( that.containerElement.css( "position" ) );

		if ( isParent && isOffsetRelative ) {
			that.offset.left = that.parentData.left + that.position.left;
			that.offset.top = that.parentData.top + that.position.top;
		} else {
			that.offset.left = that.element.offset().left;
			that.offset.top = that.element.offset().top;
		}

		woset = Math.abs( that.sizeDiff.width +
			(that._helper ?
				that.offset.left - cop.left :
				(that.offset.left - co.left)) );

		hoset = Math.abs( that.sizeDiff.height +
			(that._helper ?
				that.offset.top - cop.top :
				(that.offset.top - co.top)) );

		if ( woset + that.size.width >= that.parentData.width ) {
			that.size.width = that.parentData.width - woset;
			if ( pRatio ) {
				that.size.height = that.size.width / that.aspectRatio;
				continueResize = false;
			}
		}

		if ( hoset + that.size.height >= that.parentData.height ) {
			that.size.height = that.parentData.height - hoset;
			if ( pRatio ) {
				that.size.width = that.size.height * that.aspectRatio;
				continueResize = false;
			}
		}

		if ( !continueResize ) {
			that.position.left = that.prevPosition.left;
			that.position.top = that.prevPosition.top;
			that.size.width = that.prevSize.width;
			that.size.height = that.prevSize.height;
		}
	},

	stop: function() {
		var that = $( this ).resizable( "instance" ),
			o = that.options,
			co = that.containerOffset,
			cop = that.containerPosition,
			ce = that.containerElement,
			helper = $( that.helper ),
			ho = helper.offset(),
			w = helper.outerWidth() - that.sizeDiff.width,
			h = helper.outerHeight() - that.sizeDiff.height;

		if ( that._helper && !o.animate && ( /relative/ ).test( ce.css( "position" ) ) ) {
			$( this ).css({
				left: ho.left - cop.left - co.left,
				width: w,
				height: h
			});
		}

		if ( that._helper && !o.animate && ( /static/ ).test( ce.css( "position" ) ) ) {
			$( this ).css({
				left: ho.left - cop.left - co.left,
				width: w,
				height: h
			});
		}
	}
});

$.ui.plugin.add("resizable", "alsoResize", {

	start: function() {
		var that = $(this).resizable( "instance" ),
			o = that.options;

		$(o.alsoResize).each(function() {
			var el = $(this);
			el.data("ui-resizable-alsoresize", {
				width: parseInt(el.width(), 10), height: parseInt(el.height(), 10),
				left: parseInt(el.css("left"), 10), top: parseInt(el.css("top"), 10)
			});
		});
	},

	resize: function(event, ui) {
		var that = $(this).resizable( "instance" ),
			o = that.options,
			os = that.originalSize,
			op = that.originalPosition,
			delta = {
				height: (that.size.height - os.height) || 0,
				width: (that.size.width - os.width) || 0,
				top: (that.position.top - op.top) || 0,
				left: (that.position.left - op.left) || 0
			};

			$(o.alsoResize).each(function() {
				var el = $(this), start = $(this).data("ui-resizable-alsoresize"), style = {},
					css = el.parents(ui.originalElement[0]).length ?
							[ "width", "height" ] :
							[ "width", "height", "top", "left" ];

				$.each(css, function(i, prop) {
					var sum = (start[prop] || 0) + (delta[prop] || 0);
					if (sum && sum >= 0) {
						style[prop] = sum || null;
					}
				});

				el.css(style);
			});
	},

	stop: function() {
		$(this).removeData("resizable-alsoresize");
	}
});

$.ui.plugin.add("resizable", "ghost", {

	start: function() {

		var that = $(this).resizable( "instance" ), o = that.options, cs = that.size;

		that.ghost = that.originalElement.clone();
		that.ghost
			.css({
				opacity: 0.25,
				display: "block",
				position: "relative",
				height: cs.height,
				width: cs.width,
				margin: 0,
				left: 0,
				top: 0
			})
			.addClass("ui-resizable-ghost")
			.addClass(typeof o.ghost === "string" ? o.ghost : "");

		that.ghost.appendTo(that.helper);

	},

	resize: function() {
		var that = $(this).resizable( "instance" );
		if (that.ghost) {
			that.ghost.css({
				position: "relative",
				height: that.size.height,
				width: that.size.width
			});
		}
	},

	stop: function() {
		var that = $(this).resizable( "instance" );
		if (that.ghost && that.helper) {
			that.helper.get(0).removeChild(that.ghost.get(0));
		}
	}

});

$.ui.plugin.add("resizable", "grid", {

	resize: function() {
		var outerDimensions,
			that = $(this).resizable( "instance" ),
			o = that.options,
			cs = that.size,
			os = that.originalSize,
			op = that.originalPosition,
			a = that.axis,
			grid = typeof o.grid === "number" ? [ o.grid, o.grid ] : o.grid,
			gridX = (grid[0] || 1),
			gridY = (grid[1] || 1),
			ox = Math.round((cs.width - os.width) / gridX) * gridX,
			oy = Math.round((cs.height - os.height) / gridY) * gridY,
			newWidth = os.width + ox,
			newHeight = os.height + oy,
			isMaxWidth = o.maxWidth && (o.maxWidth < newWidth),
			isMaxHeight = o.maxHeight && (o.maxHeight < newHeight),
			isMinWidth = o.minWidth && (o.minWidth > newWidth),
			isMinHeight = o.minHeight && (o.minHeight > newHeight);

		o.grid = grid;

		if (isMinWidth) {
			newWidth += gridX;
		}
		if (isMinHeight) {
			newHeight += gridY;
		}
		if (isMaxWidth) {
			newWidth -= gridX;
		}
		if (isMaxHeight) {
			newHeight -= gridY;
		}

		if (/^(se|s|e)$/.test(a)) {
			that.size.width = newWidth;
			that.size.height = newHeight;
		} else if (/^(ne)$/.test(a)) {
			that.size.width = newWidth;
			that.size.height = newHeight;
			that.position.top = op.top - oy;
		} else if (/^(sw)$/.test(a)) {
			that.size.width = newWidth;
			that.size.height = newHeight;
			that.position.left = op.left - ox;
		} else {
			if ( newHeight - gridY <= 0 || newWidth - gridX <= 0) {
				outerDimensions = that._getPaddingPlusBorderDimensions( this );
			}

			if ( newHeight - gridY > 0 ) {
				that.size.height = newHeight;
				that.position.top = op.top - oy;
			} else {
				newHeight = gridY - outerDimensions.height;
				that.size.height = newHeight;
				that.position.top = op.top + os.height - newHeight;
			}
			if ( newWidth - gridX > 0 ) {
				that.size.width = newWidth;
				that.position.left = op.left - ox;
			} else {
				newWidth = gridX - outerDimensions.width;
				that.size.width = newWidth;
				that.position.left = op.left + os.width - newWidth;
			}
		}
	}

});

var resizable = $.ui.resizable;


/*!
 * jQuery UI Selectable 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/selectable/
 */


var selectable = $.widget("ui.selectable", $.ui.mouse, {
	version: "1.11.4",
	options: {
		appendTo: "body",
		autoRefresh: true,
		distance: 0,
		filter: "*",
		tolerance: "touch",

		// callbacks
		selected: null,
		selecting: null,
		start: null,
		stop: null,
		unselected: null,
		unselecting: null
	},
	_create: function() {
		var selectees,
			that = this;

		this.element.addClass("ui-selectable");

		this.dragged = false;

		// cache selectee children based on filter
		this.refresh = function() {
			selectees = $(that.options.filter, that.element[0]);
			selectees.addClass("ui-selectee");
			selectees.each(function() {
				var $this = $(this),
					pos = $this.offset();
				$.data(this, "selectable-item", {
					element: this,
					$element: $this,
					left: pos.left,
					top: pos.top,
					right: pos.left + $this.outerWidth(),
					bottom: pos.top + $this.outerHeight(),
					startselected: false,
					selected: $this.hasClass("ui-selected"),
					selecting: $this.hasClass("ui-selecting"),
					unselecting: $this.hasClass("ui-unselecting")
				});
			});
		};
		this.refresh();

		this.selectees = selectees.addClass("ui-selectee");

		this._mouseInit();

		this.helper = $("<div class='ui-selectable-helper'></div>");
	},

	_destroy: function() {
		this.selectees
			.removeClass("ui-selectee")
			.removeData("selectable-item");
		this.element
			.removeClass("ui-selectable ui-selectable-disabled");
		this._mouseDestroy();
	},

	_mouseStart: function(event) {
		var that = this,
			options = this.options;

		this.opos = [ event.pageX, event.pageY ];

		if (this.options.disabled) {
			return;
		}

		this.selectees = $(options.filter, this.element[0]);

		this._trigger("start", event);

		$(options.appendTo).append(this.helper);
		// position helper (lasso)
		this.helper.css({
			"left": event.pageX,
			"top": event.pageY,
			"width": 0,
			"height": 0
		});

		if (options.autoRefresh) {
			this.refresh();
		}

		this.selectees.filter(".ui-selected").each(function() {
			var selectee = $.data(this, "selectable-item");
			selectee.startselected = true;
			if (!event.metaKey && !event.ctrlKey) {
				selectee.$element.removeClass("ui-selected");
				selectee.selected = false;
				selectee.$element.addClass("ui-unselecting");
				selectee.unselecting = true;
				// selectable UNSELECTING callback
				that._trigger("unselecting", event, {
					unselecting: selectee.element
				});
			}
		});

		$(event.target).parents().addBack().each(function() {
			var doSelect,
				selectee = $.data(this, "selectable-item");
			if (selectee) {
				doSelect = (!event.metaKey && !event.ctrlKey) || !selectee.$element.hasClass("ui-selected");
				selectee.$element
					.removeClass(doSelect ? "ui-unselecting" : "ui-selected")
					.addClass(doSelect ? "ui-selecting" : "ui-unselecting");
				selectee.unselecting = !doSelect;
				selectee.selecting = doSelect;
				selectee.selected = doSelect;
				// selectable (UN)SELECTING callback
				if (doSelect) {
					that._trigger("selecting", event, {
						selecting: selectee.element
					});
				} else {
					that._trigger("unselecting", event, {
						unselecting: selectee.element
					});
				}
				return false;
			}
		});

	},

	_mouseDrag: function(event) {

		this.dragged = true;

		if (this.options.disabled) {
			return;
		}

		var tmp,
			that = this,
			options = this.options,
			x1 = this.opos[0],
			y1 = this.opos[1],
			x2 = event.pageX,
			y2 = event.pageY;

		if (x1 > x2) { tmp = x2; x2 = x1; x1 = tmp; }
		if (y1 > y2) { tmp = y2; y2 = y1; y1 = tmp; }
		this.helper.css({ left: x1, top: y1, width: x2 - x1, height: y2 - y1 });

		this.selectees.each(function() {
			var selectee = $.data(this, "selectable-item"),
				hit = false;

			//prevent helper from being selected if appendTo: selectable
			if (!selectee || selectee.element === that.element[0]) {
				return;
			}

			if (options.tolerance === "touch") {
				hit = ( !(selectee.left > x2 || selectee.right < x1 || selectee.top > y2 || selectee.bottom < y1) );
			} else if (options.tolerance === "fit") {
				hit = (selectee.left > x1 && selectee.right < x2 && selectee.top > y1 && selectee.bottom < y2);
			}

			if (hit) {
				// SELECT
				if (selectee.selected) {
					selectee.$element.removeClass("ui-selected");
					selectee.selected = false;
				}
				if (selectee.unselecting) {
					selectee.$element.removeClass("ui-unselecting");
					selectee.unselecting = false;
				}
				if (!selectee.selecting) {
					selectee.$element.addClass("ui-selecting");
					selectee.selecting = true;
					// selectable SELECTING callback
					that._trigger("selecting", event, {
						selecting: selectee.element
					});
				}
			} else {
				// UNSELECT
				if (selectee.selecting) {
					if ((event.metaKey || event.ctrlKey) && selectee.startselected) {
						selectee.$element.removeClass("ui-selecting");
						selectee.selecting = false;
						selectee.$element.addClass("ui-selected");
						selectee.selected = true;
					} else {
						selectee.$element.removeClass("ui-selecting");
						selectee.selecting = false;
						if (selectee.startselected) {
							selectee.$element.addClass("ui-unselecting");
							selectee.unselecting = true;
						}
						// selectable UNSELECTING callback
						that._trigger("unselecting", event, {
							unselecting: selectee.element
						});
					}
				}
				if (selectee.selected) {
					if (!event.metaKey && !event.ctrlKey && !selectee.startselected) {
						selectee.$element.removeClass("ui-selected");
						selectee.selected = false;

						selectee.$element.addClass("ui-unselecting");
						selectee.unselecting = true;
						// selectable UNSELECTING callback
						that._trigger("unselecting", event, {
							unselecting: selectee.element
						});
					}
				}
			}
		});

		return false;
	},

	_mouseStop: function(event) {
		var that = this;

		this.dragged = false;

		$(".ui-unselecting", this.element[0]).each(function() {
			var selectee = $.data(this, "selectable-item");
			selectee.$element.removeClass("ui-unselecting");
			selectee.unselecting = false;
			selectee.startselected = false;
			that._trigger("unselected", event, {
				unselected: selectee.element
			});
		});
		$(".ui-selecting", this.element[0]).each(function() {
			var selectee = $.data(this, "selectable-item");
			selectee.$element.removeClass("ui-selecting").addClass("ui-selected");
			selectee.selecting = false;
			selectee.selected = true;
			selectee.startselected = true;
			that._trigger("selected", event, {
				selected: selectee.element
			});
		});
		this._trigger("stop", event);

		this.helper.remove();

		return false;
	}

});


/*!
 * jQuery UI Sortable 1.11.4
 * http://jqueryui.com
 *
 * Copyright jQuery Foundation and other contributors
 * Released under the MIT license.
 * http://jquery.org/license
 *
 * http://api.jqueryui.com/sortable/
 */


var sortable = $.widget("ui.sortable", $.ui.mouse, {
	version: "1.11.4",
	widgetEventPrefix: "sort",
	ready: false,
	options: {
		appendTo: "parent",
		axis: false,
		connectWith: false,
		containment: false,
		cursor: "auto",
		cursorAt: false,
		dropOnEmpty: true,
		forcePlaceholderSize: false,
		forceHelperSize: false,
		grid: false,
		handle: false,
		helper: "original",
		items: "> *",
		opacity: false,
		placeholder: false,
		revert: false,
		scroll: true,
		scrollSensitivity: 20,
		scrollSpeed: 20,
		scope: "default",
		tolerance: "intersect",
		zIndex: 1000,

		// callbacks
		activate: null,
		beforeStop: null,
		change: null,
		deactivate: null,
		out: null,
		over: null,
		receive: null,
		remove: null,
		sort: null,
		start: null,
		stop: null,
		update: null
	},

	_isOverAxis: function( x, reference, size ) {
		return ( x >= reference ) && ( x < ( reference + size ) );
	},

	_isFloating: function( item ) {
		return (/left|right/).test(item.css("float")) || (/inline|table-cell/).test(item.css("display"));
	},

	_create: function() {
		this.containerCache = {};
		this.element.addClass("ui-sortable");

		//Get the items
		this.refresh();

		//Let's determine the parent's offset
		this.offset = this.element.offset();

		//Initialize mouse events for interaction
		this._mouseInit();

		this._setHandleClassName();

		//We're ready to go
		this.ready = true;

	},

	_setOption: function( key, value ) {
		this._super( key, value );

		if ( key === "handle" ) {
			this._setHandleClassName();
		}
	},

	_setHandleClassName: function() {
		this.element.find( ".ui-sortable-handle" ).removeClass( "ui-sortable-handle" );
		$.each( this.items, function() {
			( this.instance.options.handle ?
				this.item.find( this.instance.options.handle ) : this.item )
				.addClass( "ui-sortable-handle" );
		});
	},

	_destroy: function() {
		this.element
			.removeClass( "ui-sortable ui-sortable-disabled" )
			.find( ".ui-sortable-handle" )
				.removeClass( "ui-sortable-handle" );
		this._mouseDestroy();

		for ( var i = this.items.length - 1; i >= 0; i-- ) {
			this.items[i].item.removeData(this.widgetName + "-item");
		}

		return this;
	},

	_mouseCapture: function(event, overrideHandle) {
		var currentItem = null,
			validHandle = false,
			that = this;

		if (this.reverting) {
			return false;
		}

		if(this.options.disabled || this.options.type === "static") {
			return false;
		}

		//We have to refresh the items data once first
		this._refreshItems(event);

		//Find out if the clicked node (or one of its parents) is a actual item in this.items
		$(event.target).parents().each(function() {
			if($.data(this, that.widgetName + "-item") === that) {
				currentItem = $(this);
				return false;
			}
		});
		if($.data(event.target, that.widgetName + "-item") === that) {
			currentItem = $(event.target);
		}

		if(!currentItem) {
			return false;
		}
		if(this.options.handle && !overrideHandle) {
			$(this.options.handle, currentItem).find("*").addBack().each(function() {
				if(this === event.target) {
					validHandle = true;
				}
			});
			if(!validHandle) {
				return false;
			}
		}

		this.currentItem = currentItem;
		this._removeCurrentsFromItems();
		return true;

	},

	_mouseStart: function(event, overrideHandle, noActivation) {

		var i, body,
			o = this.options;

		this.currentContainer = this;

		//We only need to call refreshPositions, because the refreshItems call has been moved to mouseCapture
		this.refreshPositions();

		//Create and append the visible helper
		this.helper = this._createHelper(event);

		//Cache the helper size
		this._cacheHelperProportions();

		/*
		 * - Position generation -
		 * This block generates everything position related - it's the core of draggables.
		 */

		//Cache the margins of the original element
		this._cacheMargins();

		//Get the next scrolling parent
		this.scrollParent = this.helper.scrollParent();

		//The element's absolute position on the page minus margins
		this.offset = this.currentItem.offset();
		this.offset = {
			top: this.offset.top - this.margins.top,
			left: this.offset.left - this.margins.left
		};

		$.extend(this.offset, {
			click: { //Where the click happened, relative to the element
				left: event.pageX - this.offset.left,
				top: event.pageY - this.offset.top
			},
			parent: this._getParentOffset(),
			relative: this._getRelativeOffset() //This is a relative to absolute position minus the actual position calculation - only used for relative positioned helper
		});

		// Only after we got the offset, we can change the helper's position to absolute
		// TODO: Still need to figure out a way to make relative sorting possible
		this.helper.css("position", "absolute");
		this.cssPosition = this.helper.css("position");

		//Generate the original position
		this.originalPosition = this._generatePosition(event);
		this.originalPageX = event.pageX;
		this.originalPageY = event.pageY;

		//Adjust the mouse offset relative to the helper if "cursorAt" is supplied
		(o.cursorAt && this._adjustOffsetFromHelper(o.cursorAt));

		//Cache the former DOM position
		this.domPosition = { prev: this.currentItem.prev()[0], parent: this.currentItem.parent()[0] };

		//If the helper is not the original, hide the original so it's not playing any role during the drag, won't cause anything bad this way
		if(this.helper[0] !== this.currentItem[0]) {
			this.currentItem.hide();
		}

		//Create the placeholder
		this._createPlaceholder();

		//Set a containment if given in the options
		if(o.containment) {
			this._setContainment();
		}

		if( o.cursor && o.cursor !== "auto" ) { // cursor option
			body = this.document.find( "body" );

			// support: IE
			this.storedCursor = body.css( "cursor" );
			body.css( "cursor", o.cursor );

			this.storedStylesheet = $( "<style>*{ cursor: "+o.cursor+" !important; }</style>" ).appendTo( body );
		}

		if(o.opacity) { // opacity option
			if (this.helper.css("opacity")) {
				this._storedOpacity = this.helper.css("opacity");
			}
			this.helper.css("opacity", o.opacity);
		}

		if(o.zIndex) { // zIndex option
			if (this.helper.css("zIndex")) {
				this._storedZIndex = this.helper.css("zIndex");
			}
			this.helper.css("zIndex", o.zIndex);
		}

		//Prepare scrolling
		if(this.scrollParent[0] !== this.document[0] && this.scrollParent[0].tagName !== "HTML") {
			this.overflowOffset = this.scrollParent.offset();
		}

		//Call callbacks
		this._trigger("start", event, this._uiHash());

		//Recache the helper size
		if(!this._preserveHelperProportions) {
			this._cacheHelperProportions();
		}


		//Post "activate" events to possible containers
		if( !noActivation ) {
			for ( i = this.containers.length - 1; i >= 0; i-- ) {
				this.containers[ i ]._trigger( "activate", event, this._uiHash( this ) );
			}
		}

		//Prepare possible droppables
		if($.ui.ddmanager) {
			$.ui.ddmanager.current = this;
		}

		if ($.ui.ddmanager && !o.dropBehaviour) {
			$.ui.ddmanager.prepareOffsets(this, event);
		}

		this.dragging = true;

		this.helper.addClass("ui-sortable-helper");
		this._mouseDrag(event); //Execute the drag once - this causes the helper not to be visible before getting its correct position
		return true;

	},

	_mouseDrag: function(event) {
		var i, item, itemElement, intersection,
			o = this.options,
			scrolled = false;

		//Compute the helpers position
		this.position = this._generatePosition(event);
		this.positionAbs = this._convertPositionTo("absolute");

		if (!this.lastPositionAbs) {
			this.lastPositionAbs = this.positionAbs;
		}

		//Do scrolling
		if(this.options.scroll) {
			if(this.scrollParent[0] !== this.document[0] && this.scrollParent[0].tagName !== "HTML") {

				if((this.overflowOffset.top + this.scrollParent[0].offsetHeight) - event.pageY < o.scrollSensitivity) {
					this.scrollParent[0].scrollTop = scrolled = this.scrollParent[0].scrollTop + o.scrollSpeed;
				} else if(event.pageY - this.overflowOffset.top < o.scrollSensitivity) {
					this.scrollParent[0].scrollTop = scrolled = this.scrollParent[0].scrollTop - o.scrollSpeed;
				}

				if((this.overflowOffset.left + this.scrollParent[0].offsetWidth) - event.pageX < o.scrollSensitivity) {
					this.scrollParent[0].scrollLeft = scrolled = this.scrollParent[0].scrollLeft + o.scrollSpeed;
				} else if(event.pageX - this.overflowOffset.left < o.scrollSensitivity) {
					this.scrollParent[0].scrollLeft = scrolled = this.scrollParent[0].scrollLeft - o.scrollSpeed;
				}

			} else {

				if(event.pageY - this.document.scrollTop() < o.scrollSensitivity) {
					scrolled = this.document.scrollTop(this.document.scrollTop() - o.scrollSpeed);
				} else if(this.window.height() - (event.pageY - this.document.scrollTop()) < o.scrollSensitivity) {
					scrolled = this.document.scrollTop(this.document.scrollTop() + o.scrollSpeed);
				}

				if(event.pageX - this.document.scrollLeft() < o.scrollSensitivity) {
					scrolled = this.document.scrollLeft(this.document.scrollLeft() - o.scrollSpeed);
				} else if(this.window.width() - (event.pageX - this.document.scrollLeft()) < o.scrollSensitivity) {
					scrolled = this.document.scrollLeft(this.document.scrollLeft() + o.scrollSpeed);
				}

			}

			if(scrolled !== false && $.ui.ddmanager && !o.dropBehaviour) {
				$.ui.ddmanager.prepareOffsets(this, event);
			}
		}

		//Regenerate the absolute position used for position checks
		this.positionAbs = this._convertPositionTo("absolute");

		//Set the helper position
		if(!this.options.axis || this.options.axis !== "y") {
			this.helper[0].style.left = this.position.left+"px";
		}
		if(!this.options.axis || this.options.axis !== "x") {
			this.helper[0].style.top = this.position.top+"px";
		}

		//Rearrange
		for (i = this.items.length - 1; i >= 0; i--) {

			//Cache variables and intersection, continue if no intersection
			item = this.items[i];
			itemElement = item.item[0];
			intersection = this._intersectsWithPointer(item);
			if (!intersection) {
				continue;
			}

			// Only put the placeholder inside the current Container, skip all
			// items from other containers. This works because when moving
			// an item from one container to another the
			// currentContainer is switched before the placeholder is moved.
			//
			// Without this, moving items in "sub-sortables" can cause
			// the placeholder to jitter between the outer and inner container.
			if (item.instance !== this.currentContainer) {
				continue;
			}

			// cannot intersect with itself
			// no useless actions that have been done before
			// no action if the item moved is the parent of the item checked
			if (itemElement !== this.currentItem[0] &&
				this.placeholder[intersection === 1 ? "next" : "prev"]()[0] !== itemElement &&
				!$.contains(this.placeholder[0], itemElement) &&
				(this.options.type === "semi-dynamic" ? !$.contains(this.element[0], itemElement) : true)
			) {

				this.direction = intersection === 1 ? "down" : "up";

				if (this.options.tolerance === "pointer" || this._intersectsWithSides(item)) {
					this._rearrange(event, item);
				} else {
					break;
				}

				this._trigger("change", event, this._uiHash());
				break;
			}
		}

		//Post events to containers
		this._contactContainers(event);

		//Interconnect with droppables
		if($.ui.ddmanager) {
			$.ui.ddmanager.drag(this, event);
		}

		//Call callbacks
		this._trigger("sort", event, this._uiHash());

		this.lastPositionAbs = this.positionAbs;
		return false;

	},

	_mouseStop: function(event, noPropagation) {

		if(!event) {
			return;
		}

		//If we are using droppables, inform the manager about the drop
		if ($.ui.ddmanager && !this.options.dropBehaviour) {
			$.ui.ddmanager.drop(this, event);
		}

		if(this.options.revert) {
			var that = this,
				cur = this.placeholder.offset(),
				axis = this.options.axis,
				animation = {};

			if ( !axis || axis === "x" ) {
				animation.left = cur.left - this.offset.parent.left - this.margins.left + (this.offsetParent[0] === this.document[0].body ? 0 : this.offsetParent[0].scrollLeft);
			}
			if ( !axis || axis === "y" ) {
				animation.top = cur.top - this.offset.parent.top - this.margins.top + (this.offsetParent[0] === this.document[0].body ? 0 : this.offsetParent[0].scrollTop);
			}
			this.reverting = true;
			$(this.helper).animate( animation, parseInt(this.options.revert, 10) || 500, function() {
				that._clear(event);
			});
		} else {
			this._clear(event, noPropagation);
		}

		return false;

	},

	cancel: function() {

		if(this.dragging) {

			this._mouseUp({ target: null });

			if(this.options.helper === "original") {
				this.currentItem.css(this._storedCSS).removeClass("ui-sortable-helper");
			} else {
				this.currentItem.show();
			}

			//Post deactivating events to containers
			for (var i = this.containers.length - 1; i >= 0; i--){
				this.containers[i]._trigger("deactivate", null, this._uiHash(this));
				if(this.containers[i].containerCache.over) {
					this.containers[i]._trigger("out", null, this._uiHash(this));
					this.containers[i].containerCache.over = 0;
				}
			}

		}

		if (this.placeholder) {
			//$(this.placeholder[0]).remove(); would have been the jQuery way - unfortunately, it unbinds ALL events from the original node!
			if(this.placeholder[0].parentNode) {
				this.placeholder[0].parentNode.removeChild(this.placeholder[0]);
			}
			if(this.options.helper !== "original" && this.helper && this.helper[0].parentNode) {
				this.helper.remove();
			}

			$.extend(this, {
				helper: null,
				dragging: false,
				reverting: false,
				_noFinalSort: null
			});

			if(this.domPosition.prev) {
				$(this.domPosition.prev).after(this.currentItem);
			} else {
				$(this.domPosition.parent).prepend(this.currentItem);
			}
		}

		return this;

	},

	serialize: function(o) {

		var items = this._getItemsAsjQuery(o && o.connected),
			str = [];
		o = o || {};

		$(items).each(function() {
			var res = ($(o.item || this).attr(o.attribute || "id") || "").match(o.expression || (/(.+)[\-=_](.+)/));
			if (res) {
				str.push((o.key || res[1]+"[]")+"="+(o.key && o.expression ? res[1] : res[2]));
			}
		});

		if(!str.length && o.key) {
			str.push(o.key + "=");
		}

		return str.join("&");

	},

	toArray: function(o) {

		var items = this._getItemsAsjQuery(o && o.connected),
			ret = [];

		o = o || {};

		items.each(function() { ret.push($(o.item || this).attr(o.attribute || "id") || ""); });
		return ret;

	},

	/* Be careful with the following core functions */
	_intersectsWith: function(item) {

		var x1 = this.positionAbs.left,
			x2 = x1 + this.helperProportions.width,
			y1 = this.positionAbs.top,
			y2 = y1 + this.helperProportions.height,
			l = item.left,
			r = l + item.width,
			t = item.top,
			b = t + item.height,
			dyClick = this.offset.click.top,
			dxClick = this.offset.click.left,
			isOverElementHeight = ( this.options.axis === "x" ) || ( ( y1 + dyClick ) > t && ( y1 + dyClick ) < b ),
			isOverElementWidth = ( this.options.axis === "y" ) || ( ( x1 + dxClick ) > l && ( x1 + dxClick ) < r ),
			isOverElement = isOverElementHeight && isOverElementWidth;

		if ( this.options.tolerance === "pointer" ||
			this.options.forcePointerForContainers ||
			(this.options.tolerance !== "pointer" && this.helperProportions[this.floating ? "width" : "height"] > item[this.floating ? "width" : "height"])
		) {
			return isOverElement;
		} else {

			return (l < x1 + (this.helperProportions.width / 2) && // Right Half
				x2 - (this.helperProportions.width / 2) < r && // Left Half
				t < y1 + (this.helperProportions.height / 2) && // Bottom Half
				y2 - (this.helperProportions.height / 2) < b ); // Top Half

		}
	},

	_intersectsWithPointer: function(item) {

		var isOverElementHeight = (this.options.axis === "x") || this._isOverAxis(this.positionAbs.top + this.offset.click.top, item.top, item.height),
			isOverElementWidth = (this.options.axis === "y") || this._isOverAxis(this.positionAbs.left + this.offset.click.left, item.left, item.width),
			isOverElement = isOverElementHeight && isOverElementWidth,
			verticalDirection = this._getDragVerticalDirection(),
			horizontalDirection = this._getDragHorizontalDirection();

		if (!isOverElement) {
			return false;
		}

		return this.floating ?
			( ((horizontalDirection && horizontalDirection === "right") || verticalDirection === "down") ? 2 : 1 )
			: ( verticalDirection && (verticalDirection === "down" ? 2 : 1) );

	},

	_intersectsWithSides: function(item) {

		var isOverBottomHalf = this._isOverAxis(this.positionAbs.top + this.offset.click.top, item.top + (item.height/2), item.height),
			isOverRightHalf = this._isOverAxis(this.positionAbs.left + this.offset.click.left, item.left + (item.width/2), item.width),
			verticalDirection = this._getDragVerticalDirection(),
			horizontalDirection = this._getDragHorizontalDirection();

		if (this.floating && horizontalDirection) {
			return ((horizontalDirection === "right" && isOverRightHalf) || (horizontalDirection === "left" && !isOverRightHalf));
		} else {
			return verticalDirection && ((verticalDirection === "down" && isOverBottomHalf) || (verticalDirection === "up" && !isOverBottomHalf));
		}

	},

	_getDragVerticalDirection: function() {
		var delta = this.positionAbs.top - this.lastPositionAbs.top;
		return delta !== 0 && (delta > 0 ? "down" : "up");
	},

	_getDragHorizontalDirection: function() {
		var delta = this.positionAbs.left - this.lastPositionAbs.left;
		return delta !== 0 && (delta > 0 ? "right" : "left");
	},

	refresh: function(event) {
		this._refreshItems(event);
		this._setHandleClassName();
		this.refreshPositions();
		return this;
	},

	_connectWith: function() {
		var options = this.options;
		return options.connectWith.constructor === String ? [options.connectWith] : options.connectWith;
	},

	_getItemsAsjQuery: function(connected) {

		var i, j, cur, inst,
			items = [],
			queries = [],
			connectWith = this._connectWith();

		if(connectWith && connected) {
			for (i = connectWith.length - 1; i >= 0; i--){
				cur = $(connectWith[i], this.document[0]);
				for ( j = cur.length - 1; j >= 0; j--){
					inst = $.data(cur[j], this.widgetFullName);
					if(inst && inst !== this && !inst.options.disabled) {
						queries.push([$.isFunction(inst.options.items) ? inst.options.items.call(inst.element) : $(inst.options.items, inst.element).not(".ui-sortable-helper").not(".ui-sortable-placeholder"), inst]);
					}
				}
			}
		}

		queries.push([$.isFunction(this.options.items) ? this.options.items.call(this.element, null, { options: this.options, item: this.currentItem }) : $(this.options.items, this.element).not(".ui-sortable-helper").not(".ui-sortable-placeholder"), this]);

		function addItems() {
			items.push( this );
		}
		for (i = queries.length - 1; i >= 0; i--){
			queries[i][0].each( addItems );
		}

		return $(items);

	},

	_removeCurrentsFromItems: function() {

		var list = this.currentItem.find(":data(" + this.widgetName + "-item)");

		this.items = $.grep(this.items, function (item) {
			for (var j=0; j < list.length; j++) {
				if(list[j] === item.item[0]) {
					return false;
				}
			}
			return true;
		});

	},

	_refreshItems: function(event) {

		this.items = [];
		this.containers = [this];

		var i, j, cur, inst, targetData, _queries, item, queriesLength,
			items = this.items,
			queries = [[$.isFunction(this.options.items) ? this.options.items.call(this.element[0], event, { item: this.currentItem }) : $(this.options.items, this.element), this]],
			connectWith = this._connectWith();

		if(connectWith && this.ready) { //Shouldn't be run the first time through due to massive slow-down
			for (i = connectWith.length - 1; i >= 0; i--){
				cur = $(connectWith[i], this.document[0]);
				for (j = cur.length - 1; j >= 0; j--){
					inst = $.data(cur[j], this.widgetFullName);
					if(inst && inst !== this && !inst.options.disabled) {
						queries.push([$.isFunction(inst.options.items) ? inst.options.items.call(inst.element[0], event, { item: this.currentItem }) : $(inst.options.items, inst.element), inst]);
						this.containers.push(inst);
					}
				}
			}
		}

		for (i = queries.length - 1; i >= 0; i--) {
			targetData = queries[i][1];
			_queries = queries[i][0];

			for (j=0, queriesLength = _queries.length; j < queriesLength; j++) {
				item = $(_queries[j]);

				item.data(this.widgetName + "-item", targetData); // Data for target checking (mouse manager)

				items.push({
					item: item,
					instance: targetData,
					width: 0, height: 0,
					left: 0, top: 0
				});
			}
		}

	},

	refreshPositions: function(fast) {

		// Determine whether items are being displayed horizontally
		this.floating = this.items.length ?
			this.options.axis === "x" || this._isFloating( this.items[ 0 ].item ) :
			false;

		//This has to be redone because due to the item being moved out/into the offsetParent, the offsetParent's position will change
		if(this.offsetParent && this.helper) {
			this.offset.parent = this._getParentOffset();
		}

		var i, item, t, p;

		for (i = this.items.length - 1; i >= 0; i--){
			item = this.items[i];

			//We ignore calculating positions of all connected containers when we're not over them
			if(item.instance !== this.currentContainer && this.currentContainer && item.item[0] !== this.currentItem[0]) {
				continue;
			}

			t = this.options.toleranceElement ? $(this.options.toleranceElement, item.item) : item.item;

			if (!fast) {
				item.width = t.outerWidth();
				item.height = t.outerHeight();
			}

			p = t.offset();
			item.left = p.left;
			item.top = p.top;
		}

		if(this.options.custom && this.options.custom.refreshContainers) {
			this.options.custom.refreshContainers.call(this);
		} else {
			for (i = this.containers.length - 1; i >= 0; i--){
				p = this.containers[i].element.offset();
				this.containers[i].containerCache.left = p.left;
				this.containers[i].containerCache.top = p.top;
				this.containers[i].containerCache.width = this.containers[i].element.outerWidth();
				this.containers[i].containerCache.height = this.containers[i].element.outerHeight();
			}
		}

		return this;
	},

	_createPlaceholder: function(that) {
		that = that || this;
		var className,
			o = that.options;

		if(!o.placeholder || o.placeholder.constructor === String) {
			className = o.placeholder;
			o.placeholder = {
				element: function() {

					var nodeName = that.currentItem[0].nodeName.toLowerCase(),
						element = $( "<" + nodeName + ">", that.document[0] )
							.addClass(className || that.currentItem[0].className+" ui-sortable-placeholder")
							.removeClass("ui-sortable-helper");

					if ( nodeName === "tbody" ) {
						that._createTrPlaceholder(
							that.currentItem.find( "tr" ).eq( 0 ),
							$( "<tr>", that.document[ 0 ] ).appendTo( element )
						);
					} else if ( nodeName === "tr" ) {
						that._createTrPlaceholder( that.currentItem, element );
					} else if ( nodeName === "img" ) {
						element.attr( "src", that.currentItem.attr( "src" ) );
					}

					if ( !className ) {
						element.css( "visibility", "hidden" );
					}

					return element;
				},
				update: function(container, p) {

					// 1. If a className is set as 'placeholder option, we don't force sizes - the class is responsible for that
					// 2. The option 'forcePlaceholderSize can be enabled to force it even if a class name is specified
					if(className && !o.forcePlaceholderSize) {
						return;
					}

					//If the element doesn't have a actual height by itself (without styles coming from a stylesheet), it receives the inline height from the dragged item
					if(!p.height()) { p.height(that.currentItem.innerHeight() - parseInt(that.currentItem.css("paddingTop")||0, 10) - parseInt(that.currentItem.css("paddingBottom")||0, 10)); }
					if(!p.width()) { p.width(that.currentItem.innerWidth() - parseInt(that.currentItem.css("paddingLeft")||0, 10) - parseInt(that.currentItem.css("paddingRight")||0, 10)); }
				}
			};
		}

		//Create the placeholder
		that.placeholder = $(o.placeholder.element.call(that.element, that.currentItem));

		//Append it after the actual current item
		that.currentItem.after(that.placeholder);

		//Update the size of the placeholder (TODO: Logic to fuzzy, see line 316/317)
		o.placeholder.update(that, that.placeholder);

	},

	_createTrPlaceholder: function( sourceTr, targetTr ) {
		var that = this;

		sourceTr.children().each(function() {
			$( "<td>&#160;</td>", that.document[ 0 ] )
				.attr( "colspan", $( this ).attr( "colspan" ) || 1 )
				.appendTo( targetTr );
		});
	},

	_contactContainers: function(event) {
		var i, j, dist, itemWithLeastDistance, posProperty, sizeProperty, cur, nearBottom, floating, axis,
			innermostContainer = null,
			innermostIndex = null;

		// get innermost container that intersects with item
		for (i = this.containers.length - 1; i >= 0; i--) {

			// never consider a container that's located within the item itself
			if($.contains(this.currentItem[0], this.containers[i].element[0])) {
				continue;
			}

			if(this._intersectsWith(this.containers[i].containerCache)) {

				// if we've already found a container and it's more "inner" than this, then continue
				if(innermostContainer && $.contains(this.containers[i].element[0], innermostContainer.element[0])) {
					continue;
				}

				innermostContainer = this.containers[i];
				innermostIndex = i;

			} else {
				// container doesn't intersect. trigger "out" event if necessary
				if(this.containers[i].containerCache.over) {
					this.containers[i]._trigger("out", event, this._uiHash(this));
					this.containers[i].containerCache.over = 0;
				}
			}

		}

		// if no intersecting containers found, return
		if(!innermostContainer) {
			return;
		}

		// move the item into the container if it's not there already
		if(this.containers.length === 1) {
			if (!this.containers[innermostIndex].containerCache.over) {
				this.containers[innermostIndex]._trigger("over", event, this._uiHash(this));
				this.containers[innermostIndex].containerCache.over = 1;
			}
		} else {

			//When entering a new container, we will find the item with the least distance and append our item near it
			dist = 10000;
			itemWithLeastDistance = null;
			floating = innermostContainer.floating || this._isFloating(this.currentItem);
			posProperty = floating ? "left" : "top";
			sizeProperty = floating ? "width" : "height";
			axis = floating ? "clientX" : "clientY";

			for (j = this.items.length - 1; j >= 0; j--) {
				if(!$.contains(this.containers[innermostIndex].element[0], this.items[j].item[0])) {
					continue;
				}
				if(this.items[j].item[0] === this.currentItem[0]) {
					continue;
				}

				cur = this.items[j].item.offset()[posProperty];
				nearBottom = false;
				if ( event[ axis ] - cur > this.items[ j ][ sizeProperty ] / 2 ) {
					nearBottom = true;
				}

				if ( Math.abs( event[ axis ] - cur ) < dist ) {
					dist = Math.abs( event[ axis ] - cur );
					itemWithLeastDistance = this.items[ j ];
					this.direction = nearBottom ? "up": "down";
				}
			}

			//Check if dropOnEmpty is enabled
			if(!itemWithLeastDistance && !this.options.dropOnEmpty) {
				return;
			}

			if(this.currentContainer === this.containers[innermostIndex]) {
				if ( !this.currentContainer.containerCache.over ) {
					this.containers[ innermostIndex ]._trigger( "over", event, this._uiHash() );
					this.currentContainer.containerCache.over = 1;
				}
				return;
			}

			itemWithLeastDistance ? this._rearrange(event, itemWithLeastDistance, null, true) : this._rearrange(event, null, this.containers[innermostIndex].element, true);
			this._trigger("change", event, this._uiHash());
			this.containers[innermostIndex]._trigger("change", event, this._uiHash(this));
			this.currentContainer = this.containers[innermostIndex];

			//Update the placeholder
			this.options.placeholder.update(this.currentContainer, this.placeholder);

			this.containers[innermostIndex]._trigger("over", event, this._uiHash(this));
			this.containers[innermostIndex].containerCache.over = 1;
		}


	},

	_createHelper: function(event) {

		var o = this.options,
			helper = $.isFunction(o.helper) ? $(o.helper.apply(this.element[0], [event, this.currentItem])) : (o.helper === "clone" ? this.currentItem.clone() : this.currentItem);

		//Add the helper to the DOM if that didn't happen already
		if(!helper.parents("body").length) {
			$(o.appendTo !== "parent" ? o.appendTo : this.currentItem[0].parentNode)[0].appendChild(helper[0]);
		}

		if(helper[0] === this.currentItem[0]) {
			this._storedCSS = { width: this.currentItem[0].style.width, height: this.currentItem[0].style.height, position: this.currentItem.css("position"), top: this.currentItem.css("top"), left: this.currentItem.css("left") };
		}

		if(!helper[0].style.width || o.forceHelperSize) {
			helper.width(this.currentItem.width());
		}
		if(!helper[0].style.height || o.forceHelperSize) {
			helper.height(this.currentItem.height());
		}

		return helper;

	},

	_adjustOffsetFromHelper: function(obj) {
		if (typeof obj === "string") {
			obj = obj.split(" ");
		}
		if ($.isArray(obj)) {
			obj = {left: +obj[0], top: +obj[1] || 0};
		}
		if ("left" in obj) {
			this.offset.click.left = obj.left + this.margins.left;
		}
		if ("right" in obj) {
			this.offset.click.left = this.helperProportions.width - obj.right + this.margins.left;
		}
		if ("top" in obj) {
			this.offset.click.top = obj.top + this.margins.top;
		}
		if ("bottom" in obj) {
			this.offset.click.top = this.helperProportions.height - obj.bottom + this.margins.top;
		}
	},

	_getParentOffset: function() {


		//Get the offsetParent and cache its position
		this.offsetParent = this.helper.offsetParent();
		var po = this.offsetParent.offset();

		// This is a special case where we need to modify a offset calculated on start, since the following happened:
		// 1. The position of the helper is absolute, so it's position is calculated based on the next positioned parent
		// 2. The actual offset parent is a child of the scroll parent, and the scroll parent isn't the document, which means that
		//    the scroll is included in the initial calculation of the offset of the parent, and never recalculated upon drag
		if(this.cssPosition === "absolute" && this.scrollParent[0] !== this.document[0] && $.contains(this.scrollParent[0], this.offsetParent[0])) {
			po.left += this.scrollParent.scrollLeft();
			po.top += this.scrollParent.scrollTop();
		}

		// This needs to be actually done for all browsers, since pageX/pageY includes this information
		// with an ugly IE fix
		if( this.offsetParent[0] === this.document[0].body || (this.offsetParent[0].tagName && this.offsetParent[0].tagName.toLowerCase() === "html" && $.ui.ie)) {
			po = { top: 0, left: 0 };
		}

		return {
			top: po.top + (parseInt(this.offsetParent.css("borderTopWidth"),10) || 0),
			left: po.left + (parseInt(this.offsetParent.css("borderLeftWidth"),10) || 0)
		};

	},

	_getRelativeOffset: function() {

		if(this.cssPosition === "relative") {
			var p = this.currentItem.position();
			return {
				top: p.top - (parseInt(this.helper.css("top"),10) || 0) + this.scrollParent.scrollTop(),
				left: p.left - (parseInt(this.helper.css("left"),10) || 0) + this.scrollParent.scrollLeft()
			};
		} else {
			return { top: 0, left: 0 };
		}

	},

	_cacheMargins: function() {
		this.margins = {
			left: (parseInt(this.currentItem.css("marginLeft"),10) || 0),
			top: (parseInt(this.currentItem.css("marginTop"),10) || 0)
		};
	},

	_cacheHelperProportions: function() {
		this.helperProportions = {
			width: this.helper.outerWidth(),
			height: this.helper.outerHeight()
		};
	},

	_setContainment: function() {

		var ce, co, over,
			o = this.options;
		if(o.containment === "parent") {
			o.containment = this.helper[0].parentNode;
		}
		if(o.containment === "document" || o.containment === "window") {
			this.containment = [
				0 - this.offset.relative.left - this.offset.parent.left,
				0 - this.offset.relative.top - this.offset.parent.top,
				o.containment === "document" ? this.document.width() : this.window.width() - this.helperProportions.width - this.margins.left,
				(o.containment === "document" ? this.document.width() : this.window.height() || this.document[0].body.parentNode.scrollHeight) - this.helperProportions.height - this.margins.top
			];
		}

		if(!(/^(document|window|parent)$/).test(o.containment)) {
			ce = $(o.containment)[0];
			co = $(o.containment).offset();
			over = ($(ce).css("overflow") !== "hidden");

			this.containment = [
				co.left + (parseInt($(ce).css("borderLeftWidth"),10) || 0) + (parseInt($(ce).css("paddingLeft"),10) || 0) - this.margins.left,
				co.top + (parseInt($(ce).css("borderTopWidth"),10) || 0) + (parseInt($(ce).css("paddingTop"),10) || 0) - this.margins.top,
				co.left+(over ? Math.max(ce.scrollWidth,ce.offsetWidth) : ce.offsetWidth) - (parseInt($(ce).css("borderLeftWidth"),10) || 0) - (parseInt($(ce).css("paddingRight"),10) || 0) - this.helperProportions.width - this.margins.left,
				co.top+(over ? Math.max(ce.scrollHeight,ce.offsetHeight) : ce.offsetHeight) - (parseInt($(ce).css("borderTopWidth"),10) || 0) - (parseInt($(ce).css("paddingBottom"),10) || 0) - this.helperProportions.height - this.margins.top
			];
		}

	},

	_convertPositionTo: function(d, pos) {

		if(!pos) {
			pos = this.position;
		}
		var mod = d === "absolute" ? 1 : -1,
			scroll = this.cssPosition === "absolute" && !(this.scrollParent[0] !== this.document[0] && $.contains(this.scrollParent[0], this.offsetParent[0])) ? this.offsetParent : this.scrollParent,
			scrollIsRootNode = (/(html|body)/i).test(scroll[0].tagName);

		return {
			top: (
				pos.top	+																// The absolute mouse position
				this.offset.relative.top * mod +										// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.top * mod -											// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.scrollParent.scrollTop() : ( scrollIsRootNode ? 0 : scroll.scrollTop() ) ) * mod)
			),
			left: (
				pos.left +																// The absolute mouse position
				this.offset.relative.left * mod +										// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.left * mod	-										// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.scrollParent.scrollLeft() : scrollIsRootNode ? 0 : scroll.scrollLeft() ) * mod)
			)
		};

	},

	_generatePosition: function(event) {

		var top, left,
			o = this.options,
			pageX = event.pageX,
			pageY = event.pageY,
			scroll = this.cssPosition === "absolute" && !(this.scrollParent[0] !== this.document[0] && $.contains(this.scrollParent[0], this.offsetParent[0])) ? this.offsetParent : this.scrollParent, scrollIsRootNode = (/(html|body)/i).test(scroll[0].tagName);

		// This is another very weird special case that only happens for relative elements:
		// 1. If the css position is relative
		// 2. and the scroll parent is the document or similar to the offset parent
		// we have to refresh the relative offset during the scroll so there are no jumps
		if(this.cssPosition === "relative" && !(this.scrollParent[0] !== this.document[0] && this.scrollParent[0] !== this.offsetParent[0])) {
			this.offset.relative = this._getRelativeOffset();
		}

		/*
		 * - Position constraining -
		 * Constrain the position to a mix of grid, containment.
		 */

		if(this.originalPosition) { //If we are not dragging yet, we won't check for options

			if(this.containment) {
				if(event.pageX - this.offset.click.left < this.containment[0]) {
					pageX = this.containment[0] + this.offset.click.left;
				}
				if(event.pageY - this.offset.click.top < this.containment[1]) {
					pageY = this.containment[1] + this.offset.click.top;
				}
				if(event.pageX - this.offset.click.left > this.containment[2]) {
					pageX = this.containment[2] + this.offset.click.left;
				}
				if(event.pageY - this.offset.click.top > this.containment[3]) {
					pageY = this.containment[3] + this.offset.click.top;
				}
			}

			if(o.grid) {
				top = this.originalPageY + Math.round((pageY - this.originalPageY) / o.grid[1]) * o.grid[1];
				pageY = this.containment ? ( (top - this.offset.click.top >= this.containment[1] && top - this.offset.click.top <= this.containment[3]) ? top : ((top - this.offset.click.top >= this.containment[1]) ? top - o.grid[1] : top + o.grid[1])) : top;

				left = this.originalPageX + Math.round((pageX - this.originalPageX) / o.grid[0]) * o.grid[0];
				pageX = this.containment ? ( (left - this.offset.click.left >= this.containment[0] && left - this.offset.click.left <= this.containment[2]) ? left : ((left - this.offset.click.left >= this.containment[0]) ? left - o.grid[0] : left + o.grid[0])) : left;
			}

		}

		return {
			top: (
				pageY -																// The absolute mouse position
				this.offset.click.top -													// Click offset (relative to the element)
				this.offset.relative.top	-											// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.top +												// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.scrollParent.scrollTop() : ( scrollIsRootNode ? 0 : scroll.scrollTop() ) ))
			),
			left: (
				pageX -																// The absolute mouse position
				this.offset.click.left -												// Click offset (relative to the element)
				this.offset.relative.left	-											// Only for relative positioned nodes: Relative offset from element to offset parent
				this.offset.parent.left +												// The offsetParent's offset without borders (offset + border)
				( ( this.cssPosition === "fixed" ? -this.scrollParent.scrollLeft() : scrollIsRootNode ? 0 : scroll.scrollLeft() ))
			)
		};

	},

	_rearrange: function(event, i, a, hardRefresh) {

		a ? a[0].appendChild(this.placeholder[0]) : i.item[0].parentNode.insertBefore(this.placeholder[0], (this.direction === "down" ? i.item[0] : i.item[0].nextSibling));

		//Various things done here to improve the performance:
		// 1. we create a setTimeout, that calls refreshPositions
		// 2. on the instance, we have a counter variable, that get's higher after every append
		// 3. on the local scope, we copy the counter variable, and check in the timeout, if it's still the same
		// 4. this lets only the last addition to the timeout stack through
		this.counter = this.counter ? ++this.counter : 1;
		var counter = this.counter;

		this._delay(function() {
			if(counter === this.counter) {
				this.refreshPositions(!hardRefresh); //Precompute after each DOM insertion, NOT on mousemove
			}
		});

	},

	_clear: function(event, noPropagation) {

		this.reverting = false;
		// We delay all events that have to be triggered to after the point where the placeholder has been removed and
		// everything else normalized again
		var i,
			delayedTriggers = [];

		// We first have to update the dom position of the actual currentItem
		// Note: don't do it if the current item is already removed (by a user), or it gets reappended (see #4088)
		if(!this._noFinalSort && this.currentItem.parent().length) {
			this.placeholder.before(this.currentItem);
		}
		this._noFinalSort = null;

		if(this.helper[0] === this.currentItem[0]) {
			for(i in this._storedCSS) {
				if(this._storedCSS[i] === "auto" || this._storedCSS[i] === "static") {
					this._storedCSS[i] = "";
				}
			}
			this.currentItem.css(this._storedCSS).removeClass("ui-sortable-helper");
		} else {
			this.currentItem.show();
		}

		if(this.fromOutside && !noPropagation) {
			delayedTriggers.push(function(event) { this._trigger("receive", event, this._uiHash(this.fromOutside)); });
		}
		if((this.fromOutside || this.domPosition.prev !== this.currentItem.prev().not(".ui-sortable-helper")[0] || this.domPosition.parent !== this.currentItem.parent()[0]) && !noPropagation) {
			delayedTriggers.push(function(event) { this._trigger("update", event, this._uiHash()); }); //Trigger update callback if the DOM position has changed
		}

		// Check if the items Container has Changed and trigger appropriate
		// events.
		if (this !== this.currentContainer) {
			if(!noPropagation) {
				delayedTriggers.push(function(event) { this._trigger("remove", event, this._uiHash()); });
				delayedTriggers.push((function(c) { return function(event) { c._trigger("receive", event, this._uiHash(this)); };  }).call(this, this.currentContainer));
				delayedTriggers.push((function(c) { return function(event) { c._trigger("update", event, this._uiHash(this));  }; }).call(this, this.currentContainer));
			}
		}


		//Post events to containers
		function delayEvent( type, instance, container ) {
			return function( event ) {
				container._trigger( type, event, instance._uiHash( instance ) );
			};
		}
		for (i = this.containers.length - 1; i >= 0; i--){
			if (!noPropagation) {
				delayedTriggers.push( delayEvent( "deactivate", this, this.containers[ i ] ) );
			}
			if(this.containers[i].containerCache.over) {
				delayedTriggers.push( delayEvent( "out", this, this.containers[ i ] ) );
				this.containers[i].containerCache.over = 0;
			}
		}

		//Do what was originally in plugins
		if ( this.storedCursor ) {
			this.document.find( "body" ).css( "cursor", this.storedCursor );
			this.storedStylesheet.remove();
		}
		if(this._storedOpacity) {
			this.helper.css("opacity", this._storedOpacity);
		}
		if(this._storedZIndex) {
			this.helper.css("zIndex", this._storedZIndex === "auto" ? "" : this._storedZIndex);
		}

		this.dragging = false;

		if(!noPropagation) {
			this._trigger("beforeStop", event, this._uiHash());
		}

		//$(this.placeholder[0]).remove(); would have been the jQuery way - unfortunately, it unbinds ALL events from the original node!
		this.placeholder[0].parentNode.removeChild(this.placeholder[0]);

		if ( !this.cancelHelperRemoval ) {
			if ( this.helper[ 0 ] !== this.currentItem[ 0 ] ) {
				this.helper.remove();
			}
			this.helper = null;
		}

		if(!noPropagation) {
			for (i=0; i < delayedTriggers.length; i++) {
				delayedTriggers[i].call(this, event);
			} //Trigger all delayed events
			this._trigger("stop", event, this._uiHash());
		}

		this.fromOutside = false;
		return !this.cancelHelperRemoval;

	},

	_trigger: function() {
		if ($.Widget.prototype._trigger.apply(this, arguments) === false) {
			this.cancel();
		}
	},

	_uiHash: function(_inst) {
		var inst = _inst || this;
		return {
			helper: inst.helper,
			placeholder: inst.placeholder || $([]),
			position: inst.position,
			originalPosition: inst.originalPosition,
			offset: inst.positionAbs,
			item: inst.currentItem,
			sender: _inst ? _inst.element : null
		};
	}

});



}));/*!
 * jQuery contextMenu v2.2.1-dev - Plugin for simple contextMenu handling
 *
 * Version: v2.2.1-dev
 *
 * Authors: Björn Brala (SWIS.nl), Rodney Rehm, Addy Osmani (patches for FF)
 * Web: http://swisnl.github.io/jQuery-contextMenu/
 *
 * Copyright (c) 2011-2016 SWIS BV and contributors
 *
 * Licensed under
 *   MIT License http://www.opensource.org/licenses/mit-license
 *   GPL v3 http://opensource.org/licenses/GPL-3.0
 *
 * Date: 2016-06-13T14:00:33.638Z
 */

(function (factory) {
    if (typeof define === 'function' && define.amd) {
        // AMD. Register as anonymous module.
        define(['jquery'], factory);
    } else if (typeof exports === 'object') {
        // Node / CommonJS
        factory(require('jquery'));
    } else {
        // Browser globals.
        factory(jQuery);
    }
})(function ($) {

    'use strict';

    // TODO: -
    // ARIA stuff: menuitem, menuitemcheckbox und menuitemradio
    // create <menu> structure if $.support[htmlCommand || htmlMenuitem] and !opt.disableNative

    // determine html5 compatibility
    $.support.htmlMenuitem = ('HTMLMenuItemElement' in window);
    $.support.htmlCommand = ('HTMLCommandElement' in window);
    $.support.eventSelectstart = ('onselectstart' in document.documentElement);
    /* // should the need arise, test for css user-select
     $.support.cssUserSelect = (function(){
     var t = false,
     e = document.createElement('div');

     $.each('Moz|Webkit|Khtml|O|ms|Icab|'.split('|'), function(i, prefix) {
     var propCC = prefix + (prefix ? 'U' : 'u') + 'serSelect',
     prop = (prefix ? ('-' + prefix.toLowerCase() + '-') : '') + 'user-select';

     e.style.cssText = prop + ': text;';
     if (e.style[propCC] == 'text') {
     t = true;
     return false;
     }

     return true;
     });

     return t;
     })();
     */

    /* jshint ignore:start */
    if (!$.ui || !$.widget) {
        // duck punch $.cleanData like jQueryUI does to get that remove event
        $.cleanData = (function (orig) {
            return function (elems) {
                var events, elem, i;
                for (i = 0; elems[i] != null; i++) {
                    elem = elems[i];
                    try {
                        // Only trigger remove when necessary to save time
                        events = $._data(elem, 'events');
                        if (events && events.remove) {
                            $(elem).triggerHandler('remove');
                        }

                        // Http://bugs.jquery.com/ticket/8235
                    } catch (e) {}
                }
                orig(elems);
            };
        })($.cleanData);
    }
    /* jshint ignore:end */

    var // currently active contextMenu trigger
        $currentTrigger = null,
    // is contextMenu initialized with at least one menu?
        initialized = false,
    // window handle
        $win = $(window),
    // number of registered menus
        counter = 0,
    // mapping selector to namespace
        namespaces = {},
    // mapping namespace to options
        menus = {},
    // custom command type handlers
        types = {},
    // default values
        defaults = {
            // selector of contextMenu trigger
            selector: null,
            // where to append the menu to
            appendTo: null,
            // method to trigger context menu ["right", "left", "hover"]
            trigger: 'right',
            // hide menu when mouse leaves trigger / menu elements
            autoHide: false,
            // ms to wait before showing a hover-triggered context menu
            delay: 200,
            // flag denoting if a second trigger should simply move (true) or rebuild (false) an open menu
            // as long as the trigger happened on one of the trigger-element's child nodes
            reposition: true,

            // Default classname configuration to be able avoid conflicts in frameworks
            classNames : {

                hover: 'context-menu-hover', // Item hover
                disabled: 'context-menu-disabled', // Item disabled
                visible: 'context-menu-visible', // Item visible
                notSelectable: 'context-menu-not-selectable', // Item not selectable

                icon: 'context-menu-icon',
                iconEdit: 'context-menu-icon-edit',
                iconCut: 'context-menu-icon-cut',
                iconCopy: 'context-menu-icon-copy',
                iconPaste: 'context-menu-icon-paste',
                iconDelete: 'context-menu-icon-delete',
                iconAdd: 'context-menu-icon-add',
                iconQuit: 'context-menu-icon-quit'
            },

            // determine position to show menu at
            determinePosition: function ($menu) {
                // position to the lower middle of the trigger element
                if ($.ui && $.ui.position) {
                    // .position() is provided as a jQuery UI utility
                    // (...and it won't work on hidden elements)
                    $menu.css('display', 'block').position({
                        my: 'center top',
                        at: 'center bottom',
                        of: this,
                        offset: '0 5',
                        collision: 'fit'
                    }).css('display', 'none');
                } else {
                    // determine contextMenu position
                    var offset = this.offset();
                    offset.top += this.outerHeight();
                    offset.left += this.outerWidth() / 2 - $menu.outerWidth() / 2;
                    $menu.css(offset);
                }
            },
            // position menu
            position: function (opt, x, y) {
                var offset;
                // determine contextMenu position
                if (!x && !y) {
                    opt.determinePosition.call(this, opt.$menu);
                    return;
                } else if (x === 'maintain' && y === 'maintain') {
                    // x and y must not be changed (after re-show on command click)
                    offset = opt.$menu.position();
                } else {
                    // x and y are given (by mouse event)
                    offset = {top: y, left: x};
                }

                // correct offset if viewport demands it
                var bottom = $win.scrollTop() + $win.height(),
                    right = $win.scrollLeft() + $win.width(),
                    height = opt.$menu.outerHeight(),
                    width = opt.$menu.outerWidth();

                if (offset.top + height > bottom) {
                    offset.top -= height;
                }

                if (offset.top < 0) {
                    offset.top = 0;
                }

                if (offset.left + width > right) {
                    offset.left -= width;
                }

                if (offset.left < 0) {
                    offset.left = 0;
                }

                opt.$menu.css(offset);
            },
            // position the sub-menu
            positionSubmenu: function ($menu) {
                if ($.ui && $.ui.position) {
                    // .position() is provided as a jQuery UI utility
                    // (...and it won't work on hidden elements)
                    $menu.css('display', 'block').position({
                        my: 'left top',
                        at: 'right top',
                        of: this,
                        collision: 'flipfit fit'
                    }).css('display', '');
                } else {
                    // determine contextMenu position
                    var offset = {
                        top: 0,
                        left: this.outerWidth()
                    };
                    $menu.css(offset);
                }
            },
            // offset to add to zIndex
            zIndex: 1,
            // show hide animation settings
            animation: {
                duration: 50,
                show: 'slideDown',
                hide: 'slideUp'
            },
            // events
            events: {
                show: $.noop,
                hide: $.noop
            },
            // default callback
            callback: null,
            // list of contextMenu items
            items: {}
        },
    // mouse position for hover activation
        hoveract = {
            timer: null,
            pageX: null,
            pageY: null
        },
    // determine zIndex
        zindex = function ($t) {
            var zin = 0,
                $tt = $t;

            while (true) {
                zin = Math.max(zin, parseInt($tt.css('z-index'), 10) || 0);
                $tt = $tt.parent();
                if (!$tt || !$tt.length || 'html body'.indexOf($tt.prop('nodeName').toLowerCase()) > -1) {
                    break;
                }
            }
            return zin;
        },
    // event handlers
        handle = {
            // abort anything
            abortevent: function (e) {
                e.preventDefault();
                e.stopImmediatePropagation();
            },
            // contextmenu show dispatcher
            contextmenu: function (e) {
                var $this = $(this);

                // disable actual context-menu if we are using the right mouse button as the trigger
                if (e.data.trigger === 'right') {
                    e.preventDefault();
                    e.stopImmediatePropagation();
                }

                // abort native-triggered events unless we're triggering on right click
                if ((e.data.trigger !== 'right' && e.data.trigger !== 'demand') && e.originalEvent) {
                    return;
                }

                // Let the current contextmenu decide if it should show or not based on its own trigger settings
                if (e.mouseButton !== undefined && e.data) {
                    if (!(e.data.trigger === 'left' && e.mouseButton === 0) && !(e.data.trigger === 'right' && e.mouseButton === 2)) {
                        // Mouse click is not valid.
                        return;
                    }
                }

                // abort event if menu is visible for this trigger
                if ($this.hasClass('context-menu-active')) {
                    return;
                }

                if (!$this.hasClass('context-menu-disabled')) {
                    // theoretically need to fire a show event at <menu>
                    // http://www.whatwg.org/specs/web-apps/current-work/multipage/interactive-elements.html#context-menus
                    // var evt = jQuery.Event("show", { data: data, pageX: e.pageX, pageY: e.pageY, relatedTarget: this });
                    // e.data.$menu.trigger(evt);

                    $currentTrigger = $this;
                    if (e.data.build) {
                        var built = e.data.build($currentTrigger, e);
                        // abort if build() returned false
                        if (built === false) {
                            return;
                        }

                        // dynamically build menu on invocation
                        e.data = $.extend(true, {}, defaults, e.data, built || {});

                        // abort if there are no items to display
                        if (!e.data.items || $.isEmptyObject(e.data.items)) {
                            // Note: jQuery captures and ignores errors from event handlers
                            if (window.console) {
                                (console.error || console.log).call(console, 'No items specified to show in contextMenu');
                            }

                            throw new Error('No Items specified');
                        }

                        // backreference for custom command type creation
                        e.data.$trigger = $currentTrigger;

                        op.create(e.data);
                    }
                    var showMenu = false;
                    for (var item in e.data.items) {
                        if (e.data.items.hasOwnProperty(item)) {
                            var visible;
                            if ($.isFunction(e.data.items[item].visible)) {
                                visible = e.data.items[item].visible.call($(e.currentTarget), item, e.data);
                            } else if (typeof item.visible !== 'undefined') {
                                visible = e.data.items[item].visible === true;
                            } else {
                                visible = true;
                            }
                            if (visible) {
                                showMenu = true;
                            }
                        }
                    }
                    if (showMenu) {
                        // show menu
		                var menuContainer = (e.data.appendTo === null ? $('body') : $(e.data.appendTo));
		                var srcElement = e.target || e.srcElement || e.originalTarget;
                    op.show.call($this, e.data, e.pageX, e.pageY);
                    }
                }
            },
            // contextMenu left-click trigger
            click: function (e) {
                e.preventDefault();
                e.stopImmediatePropagation();
                $(this).trigger($.Event('contextmenu', {data: e.data, pageX: e.pageX, pageY: e.pageY}));
            },
            // contextMenu right-click trigger
            mousedown: function (e) {
                // register mouse down
                var $this = $(this);

                // hide any previous menus
                if ($currentTrigger && $currentTrigger.length && !$currentTrigger.is($this)) {
                    $currentTrigger.data('contextMenu').$menu.trigger('contextmenu:hide');
                }

                // activate on right click
                if (e.button === 2) {
                    $currentTrigger = $this.data('contextMenuActive', true);
                }
            },
            // contextMenu right-click trigger
            mouseup: function (e) {
                // show menu
                var $this = $(this);
                if ($this.data('contextMenuActive') && $currentTrigger && $currentTrigger.length && $currentTrigger.is($this) && !$this.hasClass('context-menu-disabled')) {
                    e.preventDefault();
                    e.stopImmediatePropagation();
                    $currentTrigger = $this;
                    $this.trigger($.Event('contextmenu', {data: e.data, pageX: e.pageX, pageY: e.pageY}));
                }

                $this.removeData('contextMenuActive');
            },
            // contextMenu hover trigger
            mouseenter: function (e) {
                var $this = $(this),
                    $related = $(e.relatedTarget),
                    $document = $(document);

                // abort if we're coming from a menu
                if ($related.is('.context-menu-list') || $related.closest('.context-menu-list').length) {
                    return;
                }

                // abort if a menu is shown
                if ($currentTrigger && $currentTrigger.length) {
                    return;
                }

                hoveract.pageX = e.pageX;
                hoveract.pageY = e.pageY;
                hoveract.data = e.data;
                $document.on('mousemove.contextMenuShow', handle.mousemove);
                hoveract.timer = setTimeout(function () {
                    hoveract.timer = null;
                    $document.off('mousemove.contextMenuShow');
                    $currentTrigger = $this;
                    $this.trigger($.Event('contextmenu', {
                        data: hoveract.data,
                        pageX: hoveract.pageX,
                        pageY: hoveract.pageY
                    }));
                }, e.data.delay);
            },
            // contextMenu hover trigger
            mousemove: function (e) {
                hoveract.pageX = e.pageX;
                hoveract.pageY = e.pageY;
            },
            // contextMenu hover trigger
            mouseleave: function (e) {
                // abort if we're leaving for a menu
                var $related = $(e.relatedTarget);
                if ($related.is('.context-menu-list') || $related.closest('.context-menu-list').length) {
                    return;
                }

                try {
                    clearTimeout(hoveract.timer);
                } catch (e) {
                }

                hoveract.timer = null;
            },
            // click on layer to hide contextMenu
            layerClick: function (e) {
                var $this = $(this),
                    root = $this.data('contextMenuRoot'),
                    button = e.button,
                    x = e.pageX,
                    y = e.pageY,
                    target,
                    offset;

                e.preventDefault();
                e.stopImmediatePropagation();

                setTimeout(function () {
                    var $window;
                    var triggerAction = ((root.trigger === 'left' && button === 0) || (root.trigger === 'right' && button === 2));

                    // find the element that would've been clicked, wasn't the layer in the way
                    if (document.elementFromPoint && root.$layer) {
                        root.$layer.hide();
                        target = document.elementFromPoint(x - $win.scrollLeft(), y - $win.scrollTop());
                        root.$layer.show();
                    }

                    if (root.reposition && triggerAction) {
                        if (document.elementFromPoint) {
                            if (root.$trigger.is(target) || root.$trigger.has(target).length) {
                                root.position.call(root.$trigger, root, x, y);
                                return;
                            }
                        } else {
                            offset = root.$trigger.offset();
                            $window = $(window);
                            // while this looks kinda awful, it's the best way to avoid
                            // unnecessarily calculating any positions
                            offset.top += $window.scrollTop();
                            if (offset.top <= e.pageY) {
                                offset.left += $window.scrollLeft();
                                if (offset.left <= e.pageX) {
                                    offset.bottom = offset.top + root.$trigger.outerHeight();
                                    if (offset.bottom >= e.pageY) {
                                        offset.right = offset.left + root.$trigger.outerWidth();
                                        if (offset.right >= e.pageX) {
                                            // reposition
                                            root.position.call(root.$trigger, root, x, y);
                                            return;
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if (target && triggerAction) {
                        root.$trigger.one('contextmenu:hidden', function () {
                            $(target).contextMenu({ x: x, y: y, button: button });
                        });
                    }

                    root.$menu.trigger('contextmenu:hide');
                }, 50);
            },
            // key handled :hover
            keyStop: function (e, opt) {
                if (!opt.isInput) {
                    e.preventDefault();
                }

                e.stopPropagation();
            },
            key: function (e) {

                var opt = {};

                // Only get the data from $currentTrigger if it exists
                if ($currentTrigger) {
                    opt = $currentTrigger.data('contextMenu') || {};
                }
                // If the trigger happen on a element that are above the contextmenu do this
                if (opt.zIndex === undefined) {
                    opt.zIndex = 0;
				}
                var targetZIndex = 0;
                var getZIndexOfTriggerTarget = function (target) {
					if (target.style.zIndex !== '') {
						targetZIndex = target.style.zIndex;
					} else {
						if (target.offsetParent !== null && target.offsetParent !== undefined) {
							getZIndexOfTriggerTarget(target.offsetParent);
						}
						else if (target.parentElement !== null && target.parentElement !== undefined) {
							getZIndexOfTriggerTarget(target.parentElement);
						}
					}
                };
                getZIndexOfTriggerTarget(e.target);
                // If targetZIndex is heigher then opt.zIndex dont progress any futher.
                // This is used to make sure that if you are using a dialog with a input / textarea / contenteditable div
                // and its above the contextmenu it wont steal keys events
                if (targetZIndex > opt.zIndex) {
                    return;
				}
                switch (e.keyCode) {
                    case 9:
                    case 38: // up
                        handle.keyStop(e, opt);
                        // if keyCode is [38 (up)] or [9 (tab) with shift]
                        if (opt.isInput) {
                            if (e.keyCode === 9 && e.shiftKey) {
                                e.preventDefault();
                                if(opt.$selected) {
                                    opt.$selected.find('input, textarea, select').blur();
                                }
                                opt.$menu.trigger('prevcommand');
                                return;
                            } else if (e.keyCode === 38 && opt.$selected.find('input, textarea, select').prop('type') === 'checkbox') {
                                // checkboxes don't capture this key
                                e.preventDefault();
                                return;
                            }
                        } else if (e.keyCode !== 9 || e.shiftKey) {
                            opt.$menu.trigger('prevcommand');
                            return;
                        }
                        break;
                    // omitting break;
                    // case 9: // tab - reached through omitted break;
                    case 40: // down
                        handle.keyStop(e, opt);
                        if (opt.isInput) {
                            if (e.keyCode === 9) {
                                e.preventDefault();
                                if(opt.$selected) {
                                    opt.$selected.find('input, textarea, select').blur();
                                }
                                opt.$menu.trigger('nextcommand');
                                return;
                            } else if (e.keyCode === 40 && opt.$selected.find('input, textarea, select').prop('type') === 'checkbox') {
                                // checkboxes don't capture this key
                                e.preventDefault();
                                return;
                            }
                        } else {
                            opt.$menu.trigger('nextcommand');
                            return;
                        }
                        break;

                    case 37: // left
                        handle.keyStop(e, opt);
                        if (opt.isInput || !opt.$selected || !opt.$selected.length) {
                            break;
                        }

                        if (!opt.$selected.parent().hasClass('context-menu-root')) {
                            var $parent = opt.$selected.parent().parent();
                            opt.$selected.trigger('contextmenu:blur');
                            opt.$selected = $parent;
                            return;
                        }
                        break;

                    case 39: // right
                        handle.keyStop(e, opt);
                        if (opt.isInput || !opt.$selected || !opt.$selected.length) {
                            break;
                        }

                        var itemdata = opt.$selected.data('contextMenu') || {};
                        if (itemdata.$menu && opt.$selected.hasClass('context-menu-submenu')) {
                            opt.$selected = null;
                            itemdata.$selected = null;
                            itemdata.$menu.trigger('nextcommand');
                            return;
                        }
                        break;

                    case 35: // end
                    case 36: // home
                        if (opt.$selected && opt.$selected.find('input, textarea, select').length) {
                            return;
                        } else {
                            (opt.$selected && opt.$selected.parent() || opt.$menu)
                                .children(':not(.' + opt.classNames.disabled + ', .' + opt.classNames.notSelectable + ')')[e.keyCode === 36 ? 'first' : 'last']()
                                .trigger('contextmenu:focus');
                            e.preventDefault();
                            return;
                        }
                        break;

                    case 13: // enter
                        handle.keyStop(e, opt);
                        if (opt.isInput) {
                            if (opt.$selected && !opt.$selected.is('textarea, select')) {
                                e.preventDefault();
                                return;
                            }
                            break;
                        }
                        if (typeof opt.$selected !== 'undefined' && opt.$selected !== null) {
                            opt.$selected.trigger('mouseup');
                        }
                        return;

                    case 32: // space
                    case 33: // page up
                    case 34: // page down
                        // prevent browser from scrolling down while menu is visible
                        handle.keyStop(e, opt);
                        return;

                    case 27: // esc
                        handle.keyStop(e, opt);
                        opt.$menu.trigger('contextmenu:hide');
                        return;

                    default: // 0-9, a-z
                        var k = (String.fromCharCode(e.keyCode)).toUpperCase();
                        if (opt.accesskeys && opt.accesskeys[k]) {
                            // according to the specs accesskeys must be invoked immediately
                            opt.accesskeys[k].$node.trigger(opt.accesskeys[k].$menu ? 'contextmenu:focus' : 'mouseup');
                            return;
                        }
                        break;
                }
                // pass event to selected item,
                // stop propagation to avoid endless recursion
                e.stopPropagation();
                if (typeof opt.$selected !== 'undefined' && opt.$selected !== null) {
                    opt.$selected.trigger(e);
                }
            },
            // select previous possible command in menu
            prevItem: function (e) {
                e.stopPropagation();
                var opt = $(this).data('contextMenu') || {};
                var root = $(this).data('contextMenuRoot') || {};

                // obtain currently selected menu
                if (opt.$selected) {
                    var $s = opt.$selected;
                    opt = opt.$selected.parent().data('contextMenu') || {};
                    opt.$selected = $s;
                }

                var $children = opt.$menu.children(),
                    $prev = !opt.$selected || !opt.$selected.prev().length ? $children.last() : opt.$selected.prev(),
                    $round = $prev;

                // skip disabled or hidden elements
                while ($prev.hasClass(root.classNames.disabled) || $prev.hasClass(root.classNames.notSelectable) || $prev.is(':hidden')) {
                    if ($prev.prev().length) {
                        $prev = $prev.prev();
                    } else {
                        $prev = $children.last();
                    }
                    if ($prev.is($round)) {
                        // break endless loop
                        return;
                    }
                }

                // leave current
                if (opt.$selected) {
                    handle.itemMouseleave.call(opt.$selected.get(0), e);
                }

                // activate next
                handle.itemMouseenter.call($prev.get(0), e);

                // focus input
                var $input = $prev.find('input, textarea, select');
                if ($input.length) {
                    $input.focus();
                }
            },
            // select next possible command in menu
            nextItem: function (e) {
                e.stopPropagation();
                var opt = $(this).data('contextMenu') || {};
                var root = $(this).data('contextMenuRoot') || {};

                // obtain currently selected menu
                if (opt.$selected) {
                    var $s = opt.$selected;
                    opt = opt.$selected.parent().data('contextMenu') || {};
                    opt.$selected = $s;
                }

                var $children = opt.$menu.children(),
                    $next = !opt.$selected || !opt.$selected.next().length ? $children.first() : opt.$selected.next(),
                    $round = $next;

                // skip disabled
                while ($next.hasClass(root.classNames.disabled) || $next.hasClass(root.classNames.notSelectable) || $next.is(':hidden')) {
                    if ($next.next().length) {
                        $next = $next.next();
                    } else {
                        $next = $children.first();
                    }
                    if ($next.is($round)) {
                        // break endless loop
                        return;
                    }
                }

                // leave current
                if (opt.$selected) {
                    handle.itemMouseleave.call(opt.$selected.get(0), e);
                }

                // activate next
                handle.itemMouseenter.call($next.get(0), e);

                // focus input
                var $input = $next.find('input, textarea, select');
                if ($input.length) {
                    $input.focus();
                }
            },
            // flag that we're inside an input so the key handler can act accordingly
            focusInput: function () {
                var $this = $(this).closest('.context-menu-item'),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                root.$selected = opt.$selected = $this;
                root.isInput = opt.isInput = true;
            },
            // flag that we're inside an input so the key handler can act accordingly
            blurInput: function () {
                var $this = $(this).closest('.context-menu-item'),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                root.isInput = opt.isInput = false;
            },
            // :hover on menu
            menuMouseenter: function () {
                var root = $(this).data().contextMenuRoot;
                root.hovering = true;
            },
            // :hover on menu
            menuMouseleave: function (e) {
                var root = $(this).data().contextMenuRoot;
                if (root.$layer && root.$layer.is(e.relatedTarget)) {
                    root.hovering = false;
                }
            },
            // :hover done manually so key handling is possible
            itemMouseenter: function (e) {
                var $this = $(this),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                root.hovering = true;

                // abort if we're re-entering
                if (e && root.$layer && root.$layer.is(e.relatedTarget)) {
                    e.preventDefault();
                    e.stopImmediatePropagation();
                }

                // make sure only one item is selected
                (opt.$menu ? opt : root).$menu
                    .children('.' + root.classNames.hover).trigger('contextmenu:blur')
                    .children('.hover').trigger('contextmenu:blur');

                if ($this.hasClass(root.classNames.disabled) || $this.hasClass(root.classNames.notSelectable)) {
                    opt.$selected = null;
                    return;
                }

                $this.trigger('contextmenu:focus');
            },
            // :hover done manually so key handling is possible
            itemMouseleave: function (e) {
                var $this = $(this),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                if (root !== opt && root.$layer && root.$layer.is(e.relatedTarget)) {
                    if (typeof root.$selected !== 'undefined' && root.$selected !== null) {
                        root.$selected.trigger('contextmenu:blur');
                    }
                    e.preventDefault();
                    e.stopImmediatePropagation();
                    root.$selected = opt.$selected = opt.$node;
                    return;
                }

                $this.trigger('contextmenu:blur');
            },
            // contextMenu item click
            itemClick: function (e) {
                var $this = $(this),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot,
                    key = data.contextMenuKey,
                    callback;

                // abort if the key is unknown or disabled or is a menu
                if (!opt.items[key] || $this.is('.' + root.classNames.disabled + ', .context-menu-submenu, .context-menu-separator, .' + root.classNames.notSelectable)) {
                    return;
                }

                e.preventDefault();
                e.stopImmediatePropagation();

                if ($.isFunction(root.callbacks[key]) && Object.prototype.hasOwnProperty.call(root.callbacks, key)) {
                    // item-specific callback
                    callback = root.callbacks[key];
                } else if ($.isFunction(root.callback)) {
                    // default callback
                    callback = root.callback;
                } else {
                    // no callback, no action
                    return;
                }

                // hide menu if callback doesn't stop that
                if (callback.call(root.$trigger, key, root) !== false) {
                    root.$menu.trigger('contextmenu:hide');
                } else if (root.$menu.parent().length) {
                    op.update.call(root.$trigger, root);
                }
            },
            // ignore click events on input elements
            inputClick: function (e) {
                e.stopImmediatePropagation();
            },
            // hide <menu>
            hideMenu: function (e, data) {
                var root = $(this).data('contextMenuRoot');
                op.hide.call(root.$trigger, root, data && data.force);
            },
            // focus <command>
            focusItem: function (e) {
                e.stopPropagation();
                var $this = $(this),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                if ($this.hasClass(root.classNames.disabled) || $this.hasClass(root.classNames.notSelectable)) {
                    return;
                }

                $this
                    .addClass([root.classNames.hover, root.classNames.visible].join(' '))
                    // select other items and included items
                    .parent().find('.context-menu-item').not($this)
                    .removeClass(root.classNames.visible)
                    .filter('.' + root.classNames.hover)
                    .trigger('contextmenu:blur');

                // remember selected
                opt.$selected = root.$selected = $this;

                // position sub-menu - do after show so dumb $.ui.position can keep up
                if (opt.$node) {
                    root.positionSubmenu.call(opt.$node, opt.$menu);
                }
            },
            // blur <command>
            blurItem: function (e) {
                e.stopPropagation();
                var $this = $(this),
                    data = $this.data(),
                    opt = data.contextMenu,
                    root = data.contextMenuRoot;

                if (opt.autoHide) { // for tablets and touch screens this needs to remain
                    $this.removeClass(root.classNames.visible);
                }
                $this.removeClass(root.classNames.hover);
                opt.$selected = null;
            }
        },
    // operations
        op = {
            show: function (opt, x, y) {
                var $trigger = $(this),
                    css = {};

                // hide any open menus
                $('#context-menu-layer').trigger('mousedown');

                // backreference for callbacks
                opt.$trigger = $trigger;

                // show event
                if (opt.events.show.call($trigger, opt) === false) {
                    $currentTrigger = null;
                    return;
                }

                // create or update context menu
                op.update.call($trigger, opt);

                // position menu
                opt.position.call($trigger, opt, x, y);

                // make sure we're in front
                if (opt.zIndex) {
                  var additionalZValue = opt.zIndex;
                  // If opt.zIndex is a function, call the function to get the right zIndex.
                  if (typeof opt.zIndex === 'function') {
                      additionalZValue = opt.zIndex.call($trigger, opt);
                  }
                  css.zIndex = zindex($trigger) + additionalZValue;
                }

                // add layer
                op.layer.call(opt.$menu, opt, css.zIndex);

                // adjust sub-menu zIndexes
                opt.$menu.find('ul').css('zIndex', css.zIndex + 1);

                // position and show context menu
                opt.$menu.css(css)[opt.animation.show](opt.animation.duration, function () {
                    $trigger.trigger('contextmenu:visible');
                });
                // make options available and set state
                $trigger
                    .data('contextMenu', opt)
                    .addClass('context-menu-active');

                // register key handler
                $(document).off('keydown.contextMenu').on('keydown.contextMenu', handle.key);
                // register autoHide handler
                if (opt.autoHide) {
                    // mouse position handler
                    $(document).on('mousemove.contextMenuAutoHide', function (e) {
                        // need to capture the offset on mousemove,
                        // since the page might've been scrolled since activation
                        var pos = $trigger.offset();
                        pos.right = pos.left + $trigger.outerWidth();
                        pos.bottom = pos.top + $trigger.outerHeight();

                        if (opt.$layer && !opt.hovering && (!(e.pageX >= pos.left && e.pageX <= pos.right) || !(e.pageY >= pos.top && e.pageY <= pos.bottom))) {
                            /* Additional hover check after short time, you might just miss the edge of the menu */
                            setTimeout(function () {
                                if (!opt.hovering) { opt.$menu.trigger('contextmenu:hide'); }
                            }, 50);
                        }
                    });
                }
            },
            hide: function (opt, force) {
                var $trigger = $(this);
                if (!opt) {
                    opt = $trigger.data('contextMenu') || {};
                }

                // hide event
                if (!force && opt.events && opt.events.hide.call($trigger, opt) === false) {
                    return;
                }

                // remove options and revert state
                $trigger
                    .removeData('contextMenu')
                    .removeClass('context-menu-active');

                if (opt.$layer) {
                    // keep layer for a bit so the contextmenu event can be aborted properly by opera
                    setTimeout((function ($layer) {
                        return function () {
                            $layer.remove();
                        };
                    })(opt.$layer), 10);

                    try {
                        delete opt.$layer;
                    } catch (e) {
                        opt.$layer = null;
                    }
                }

                // remove handle
                $currentTrigger = null;
                // remove selected
                opt.$menu.find('.' + opt.classNames.hover).trigger('contextmenu:blur');
                opt.$selected = null;
                // collapse all submenus
                opt.$menu.find('.' + opt.classNames.visible).removeClass(opt.classNames.visible);
                // unregister key and mouse handlers
                // $(document).off('.contextMenuAutoHide keydown.contextMenu'); // http://bugs.jquery.com/ticket/10705
                $(document).off('.contextMenuAutoHide').off('keydown.contextMenu');
                // hide menu
                if(opt.$menu){
                    opt.$menu[opt.animation.hide](opt.animation.duration, function () {
                        // tear down dynamically built menu after animation is completed.
                        if (opt.build) {
                            opt.$menu.remove();
                            $.each(opt, function (key) {
                                switch (key) {
                                    case 'ns':
                                    case 'selector':
                                    case 'build':
                                    case 'trigger':
                                        return true;

                                    default:
                                        opt[key] = undefined;
                                        try {
                                            delete opt[key];
                                        } catch (e) {
                                        }
                                        return true;
                                }
                            });
                        }

                        setTimeout(function () {
                            $trigger.trigger('contextmenu:hidden');
                        }, 10);
                    });
                }
            },
            create: function (opt, root) {
                if (root === undefined) {
                    root = opt;
                }
                // create contextMenu
                opt.$menu = $('<ul class="context-menu-list"></ul>').addClass(opt.className || '').data({
                    'contextMenu': opt,
                    'contextMenuRoot': root
                });

                $.each(['callbacks', 'commands', 'inputs'], function (i, k) {
                    opt[k] = {};
                    if (!root[k]) {
                        root[k] = {};
                    }
                });

                if(!root.accesskeys){
                    root.accesskeys = {};
                }

                function createNameNode(item) {
                    var $name = $('<span></span>');
                    if (item._accesskey) {
                        if (item._beforeAccesskey) {
                            $name.append(document.createTextNode(item._beforeAccesskey));
                        }
                        $('<span></span>')
                            .addClass('context-menu-accesskey')
                            .text(item._accesskey)
                            .appendTo($name);
                        if (item._afterAccesskey) {
                            $name.append(document.createTextNode(item._afterAccesskey));
                        }
                    } else {
                        if (item.isHtmlName) {
                            // restrict use with access keys
                            if (typeof item.accesskey !== 'undefined') {
                                throw new Error('accesskeys are not compatible with HTML names and cannot be used together in the same item');
                            }
                            $name.html(item.name);
                        } else {
                            $name.text(item.name);
                        }
                    }
                    return $name;
                }

                // create contextMenu items
                $.each(opt.items, function (key, item) {
                    var $t = $('<li class="context-menu-item"></li>').addClass(item.className || ''),
                        $label = null,
                        $input = null;

                    // iOS needs to see a click-event bound to an element to actually
                    // have the TouchEvents infrastructure trigger the click event
                    $t.on('click', $.noop);

                    // Make old school string seperator a real item so checks wont be
                    // akward later.
                    if (typeof item === 'string') {
                        item = { type : 'cm_seperator' };
                    }

                    item.$node = $t.data({
                        'contextMenu': opt,
                        'contextMenuRoot': root,
                        'contextMenuKey': key
                    });

                    // register accesskey
                    // NOTE: the accesskey attribute should be applicable to any element, but Safari5 and Chrome13 still can't do that
                    if (typeof item.accesskey !== 'undefined') {
                        var aks = splitAccesskey(item.accesskey);
                        for (var i = 0, ak; ak = aks[i]; i++) {
                            if (!root.accesskeys[ak]) {
                                root.accesskeys[ak] = item;
                                var matched = item.name.match(new RegExp('^(.*?)(' + ak + ')(.*)$', 'i'));
                                if (matched) {
                                    item._beforeAccesskey = matched[1];
                                    item._accesskey = matched[2];
                                    item._afterAccesskey = matched[3];
                                }
                                break;
                            }
                        }
                    }

                    if (item.type && types[item.type]) {
                        // run custom type handler
                        types[item.type].call($t, item, opt, root);
                        // register commands
                        $.each([opt, root], function (i, k) {
                            k.commands[key] = item;
                            if ($.isFunction(item.callback)) {
                                k.callbacks[key] = item.callback;
                            }
                        });
                    } else {
                        // add label for input
                        if (item.type === 'cm_seperator') {
                            $t.addClass('context-menu-separator ' + root.classNames.notSelectable);
                        } else if (item.type === 'html') {
                            $t.addClass('context-menu-html ' + root.classNames.notSelectable);
                        } else if (item.type) {
                            $label = $('<label></label>').appendTo($t);
                            createNameNode(item).appendTo($label);

                            $t.addClass('context-menu-input');
                            opt.hasTypes = true;
                            $.each([opt, root], function (i, k) {
                                k.commands[key] = item;
                                k.inputs[key] = item;
                            });
                        } else if (item.items) {
                            item.type = 'sub';
                        }

                        switch (item.type) {
                            case 'cm_seperator':
                                break;

                            case 'text':
                                $input = $('<input type="text" value="1" name="" value="">')
                                    .attr('name', 'context-menu-input-' + key)
                                    .val(item.value || '')
                                    .appendTo($label);
                                break;

                            case 'textarea':
                                $input = $('<textarea name=""></textarea>')
                                    .attr('name', 'context-menu-input-' + key)
                                    .val(item.value || '')
                                    .appendTo($label);

                                if (item.height) {
                                    $input.height(item.height);
                                }
                                break;

                            case 'checkbox':
                                $input = $('<input type="checkbox" value="1" name="" value="">')
                                    .attr('name', 'context-menu-input-' + key)
                                    .val(item.value || '')
                                    .prop('checked', !!item.selected)
                                    .prependTo($label);
                                break;

                            case 'radio':
                                $input = $('<input type="radio" value="1" name="" value="">')
                                    .attr('name', 'context-menu-input-' + item.radio)
                                    .val(item.value || '')
                                    .prop('checked', !!item.selected)
                                    .prependTo($label);
                                break;

                            case 'select':
                                $input = $('<select name="">')
                                    .attr('name', 'context-menu-input-' + key)
                                    .appendTo($label);
                                if (item.options) {
                                    $.each(item.options, function (value, text) {
                                        $('<option></option>').val(value).text(text).appendTo($input);
                                    });
                                    $input.val(item.selected);
                                }
                                break;

                            case 'sub':
                                createNameNode(item).appendTo($t);

                                item.appendTo = item.$node;
                                op.create(item, root);
                                $t.data('contextMenu', item).addClass('context-menu-submenu');
                                item.callback = null;
                                break;

                            case 'html':
                                $(item.html).appendTo($t);
                                break;

                            default:
                                $.each([opt, root], function (i, k) {
                                    k.commands[key] = item;
                                    if ($.isFunction(item.callback)) {
                                        k.callbacks[key] = item.callback;
                                    }
                                });
                                createNameNode(item).appendTo($t);
                                break;
                        }

                        // disable key listener in <input>
                        if (item.type && item.type !== 'sub' && item.type !== 'html' && item.type !== 'cm_seperator') {
                            $input
                                .on('focus', handle.focusInput)
                                .on('blur', handle.blurInput);

                            if (item.events) {
                                $input.on(item.events, opt);
                            }
                        }

                        // add icons
                        if (item.icon) {
                            if ($.isFunction(item.icon)) {
                                item._icon = item.icon.call(this, this, $t, key, item);
                            } else {
                                if ( typeof(item.icon) === 'string' && item.icon.substring(0,3) == 'fa-' ) {
                                    // to enable font awesome
                                    item._icon = root.classNames.icon + ' ' + root.classNames.icon + '--fa fa ' + item.icon;
                                } else {
                                    item._icon = root.classNames.icon + ' ' + root.classNames.icon + '-' + item.icon;
                                }
                            }
                            $t.addClass(item._icon);
                        }
                    }

                    // cache contained elements
                    item.$input = $input;
                    item.$label = $label;

                    // attach item to menu
                    $t.appendTo(opt.$menu);

                    // Disable text selection
                    if (!opt.hasTypes && $.support.eventSelectstart) {
                        // browsers support user-select: none,
                        // IE has a special event for text-selection
                        // browsers supporting neither will not be preventing text-selection
                        $t.on('selectstart.disableTextSelect', handle.abortevent);
                    }
                });
                // attach contextMenu to <body> (to bypass any possible overflow:hidden issues on parents of the trigger element)
                if (!opt.$node) {
                    opt.$menu.css('display', 'none').addClass('context-menu-root');
                }
                opt.$menu.appendTo(opt.appendTo || document.body);
            },
            resize: function ($menu, nested) {
                var domMenu;
                // determine widths of submenus, as CSS won't grow them automatically
                // position:absolute within position:absolute; min-width:100; max-width:200; results in width: 100;
                // kinda sucks hard...

                // determine width of absolutely positioned element
                $menu.css({position: 'absolute', display: 'block'});
                // don't apply yet, because that would break nested elements' widths
                $menu.data('width',
                    (domMenu = $menu.get(0)).getBoundingClientRect ?
                        Math.ceil(domMenu.getBoundingClientRect().width) :
                        $menu.outerWidth() + 1); // outerWidth() returns rounded pixels
                // reset styles so they allow nested elements to grow/shrink naturally
                $menu.css({
                    position: 'static',
                    minWidth: '0px',
                    maxWidth: '100000px'
                });
                // identify width of nested menus
                $menu.find('> li > ul').each(function () {
                    op.resize($(this), true);
                });
                // reset and apply changes in the end because nested
                // elements' widths wouldn't be calculatable otherwise
                if (!nested) {
                    $menu.find('ul').addBack().css({
                        position: '',
                        display: '',
                        minWidth: '',
                        maxWidth: ''
                    }).outerWidth(function () {
                        return $(this).data('width');
                    });
                }
            },
            update: function (opt, root) {
                var $trigger = this;
                if (root === undefined) {
                    root = opt;
                    op.resize(opt.$menu);
                }
                // re-check disabled for each item
                opt.$menu.children().each(function () {
                    var $item = $(this),
                        key = $item.data('contextMenuKey'),
                        item = opt.items[key],
                        disabled = ($.isFunction(item.disabled) && item.disabled.call($trigger, key, root)) || item.disabled === true,
                        visible;
                    if ($.isFunction(item.visible)) {
                        visible = item.visible.call($trigger, key, root);
                    } else if (typeof item.visible !== 'undefined') {
                        visible = item.visible === true;
                    } else {
                        visible = true;
                    }
                    $item[visible ? 'show' : 'hide']();

                    // dis- / enable item
                    $item[disabled ? 'addClass' : 'removeClass'](root.classNames.disabled);

                    if ($.isFunction(item.icon)) {
                        $item.removeClass(item._icon);
                        item._icon = item.icon.call(this, $trigger, $item, key, item);
                        $item.addClass(item._icon);
                    }

                    if (item.type) {
                        // dis- / enable input elements
                        $item.find('input, select, textarea').prop('disabled', disabled);

                        // update input states
                        switch (item.type) {
                            case 'text':
                            case 'textarea':
                                item.$input.val(item.value || '');
                                break;

                            case 'checkbox':
                            case 'radio':
                                item.$input.val(item.value || '').prop('checked', !!item.selected);
                                break;

                            case 'select':
                                item.$input.val(item.selected || '');
                                break;
                        }
                    }

                    if (item.$menu) {
                        // update sub-menu
                        op.update.call($trigger, item, root);
                    }
                });
            },
            layer: function (opt, zIndex) {
                // add transparent layer for click area
                // filter and background for Internet Explorer, Issue #23
                var $layer = opt.$layer = $('<div id="context-menu-layer" style="position:fixed; z-index:' + zIndex + '; top:0; left:0; opacity: 0; filter: alpha(opacity=0); background-color: #000;"></div>')
                    .css({height: $win.height(), width: $win.width(), display: 'block'})
                    .data('contextMenuRoot', opt)
                    .insertBefore(this)
                    .on('contextmenu', handle.abortevent)
                    .on('mousedown', handle.layerClick);

                // IE6 doesn't know position:fixed;
                if (document.body.style.maxWidth === undefined) { // IE6 doesn't support maxWidth
                    $layer.css({
                        'position': 'absolute',
                        'height': $(document).height()
                    });
                }

                return $layer;
            }
        };

    // split accesskey according to http://www.whatwg.org/specs/web-apps/current-work/multipage/editing.html#assigned-access-key
    function splitAccesskey(val) {
        var t = val.split(/\s+/),
            keys = [];

        for (var i = 0, k; k = t[i]; i++) {
            k = k.charAt(0).toUpperCase(); // first character only
            // theoretically non-accessible characters should be ignored, but different systems, different keyboard layouts, ... screw it.
            // a map to look up already used access keys would be nice
            keys.push(k);
        }

        return keys;
    }

// handle contextMenu triggers
    $.fn.contextMenu = function (operation) {
        var $t = this, $o = operation;
        if (this.length > 0) {  // this is not a build on demand menu
            if (operation === undefined) {
                this.first().trigger('contextmenu');
            } else if (operation.x !== undefined && operation.y !== undefined) {
                this.first().trigger($.Event('contextmenu', { pageX: operation.x, pageY: operation.y, mouseButton: operation.button }));
            } else if (operation === 'hide') {
                var $menu = this.first().data('contextMenu') ? this.first().data('contextMenu').$menu : null;
                if($menu){
                    $menu.trigger('contextmenu:hide');
                }
            } else if (operation === 'destroy') {
                $.contextMenu('destroy', {context: this});
            } else if ($.isPlainObject(operation)) {
                operation.context = this;
                $.contextMenu('create', operation);
            } else if (operation) {
                this.removeClass('context-menu-disabled');
            } else if (!operation) {
                this.addClass('context-menu-disabled');
            }
        } else {
            $.each(menus, function () {
                if (this.selector === $t.selector) {
                    $o.data = this;

                    $.extend($o.data, {trigger: 'demand'});
                }
            });

            handle.contextmenu.call($o.target, $o);
        }

        return this;
    };

    // manage contextMenu instances
    $.contextMenu = function (operation, options) {
        if (typeof operation !== 'string') {
            options = operation;
            operation = 'create';
        }

        if (typeof options === 'string') {
            options = {selector: options};
        } else if (options === undefined) {
            options = {};
        }

        // merge with default options
        var o = $.extend(true, {}, defaults, options || {});
        var $document = $(document);
        var $context = $document;
        var _hasContext = false;

        if (!o.context || !o.context.length) {
            o.context = document;
        } else {
            // you never know what they throw at you...
            $context = $(o.context).first();
            o.context = $context.get(0);
            _hasContext = o.context !== document;
        }

        switch (operation) {
            case 'create':
                // no selector no joy
                if (!o.selector) {
                    throw new Error('No selector specified');
                }
                // make sure internal classes are not bound to
                if (o.selector.match(/.context-menu-(list|item|input)($|\s)/)) {
                    throw new Error('Cannot bind to selector "' + o.selector + '" as it contains a reserved className');
                }
                if (!o.build && (!o.items || $.isEmptyObject(o.items))) {
                    throw new Error('No Items specified');
                }
                counter++;
                o.ns = '.contextMenu' + counter;
                if (!_hasContext) {
                    namespaces[o.selector] = o.ns;
                }
                menus[o.ns] = o;

                // default to right click
                if (!o.trigger) {
                    o.trigger = 'right';
                }

                if (!initialized) {
                    var itemClick = o.itemClickEvent === 'click' ? 'click.contextMenu' : 'mouseup.contextMenu'
                    var contextMenuItemObj = {
                            // 'mouseup.contextMenu': handle.itemClick,
                            // 'click.contextMenu': handle.itemClick,
                            'contextmenu:focus.contextMenu': handle.focusItem,
                            'contextmenu:blur.contextMenu': handle.blurItem,
                            'contextmenu.contextMenu': handle.abortevent,
                            'mouseenter.contextMenu': handle.itemMouseenter,
                            'mouseleave.contextMenu': handle.itemMouseleave
                        };
                    contextMenuItemObj[itemClick] = handle.itemClick
                    // make sure item click is registered first
                    $document
                        .on({
                            'contextmenu:hide.contextMenu': handle.hideMenu,
                            'prevcommand.contextMenu': handle.prevItem,
                            'nextcommand.contextMenu': handle.nextItem,
                            'contextmenu.contextMenu': handle.abortevent,
                            'mouseenter.contextMenu': handle.menuMouseenter,
                            'mouseleave.contextMenu': handle.menuMouseleave
                        }, '.context-menu-list')
                        .on('mouseup.contextMenu', '.context-menu-input', handle.inputClick)
                        .on(contextMenuItemObj, '.context-menu-item');

                    initialized = true;
                }

                // engage native contextmenu event
                $context
                    .on('contextmenu' + o.ns, o.selector, o, handle.contextmenu);

                if (_hasContext) {
                    // add remove hook, just in case
                    $context.on('remove' + o.ns, function () {
                        $(this).contextMenu('destroy');
                    });
                }

                switch (o.trigger) {
                    case 'hover':
                        $context
                            .on('mouseenter' + o.ns, o.selector, o, handle.mouseenter)
                            .on('mouseleave' + o.ns, o.selector, o, handle.mouseleave);
                        break;

                    case 'left':
                        $context.on('click' + o.ns, o.selector, o, handle.click);
                        break;
                    /*
                     default:
                     // http://www.quirksmode.org/dom/events/contextmenu.html
                     $document
                     .on('mousedown' + o.ns, o.selector, o, handle.mousedown)
                     .on('mouseup' + o.ns, o.selector, o, handle.mouseup);
                     break;
                     */
                }

                // create menu
                if (!o.build) {
                    op.create(o);
                }
                break;

            case 'destroy':
                var $visibleMenu;
                if (_hasContext) {
                    // get proper options
                    var context = o.context;
                    $.each(menus, function (ns, o) {
                        if (o.context !== context) {
                            return true;
                        }

                        $visibleMenu = $('.context-menu-list').filter(':visible');
                        if ($visibleMenu.length && $visibleMenu.data().contextMenuRoot.$trigger.is($(o.context).find(o.selector))) {
                            $visibleMenu.trigger('contextmenu:hide', {force: true});
                        }

                        try {
                            if (menus[o.ns].$menu) {
                                menus[o.ns].$menu.remove();
                            }

                            delete menus[o.ns];
                        } catch (e) {
                            menus[o.ns] = null;
                        }

                        $(o.context).off(o.ns);

                        return true;
                    });
                } else if (!o.selector) {
                    $document.off('.contextMenu .contextMenuAutoHide');
                    $.each(menus, function (ns, o) {
                        $(o.context).off(o.ns);
                    });

                    namespaces = {};
                    menus = {};
                    counter = 0;
                    initialized = false;

                    $('#context-menu-layer, .context-menu-list').remove();
                } else if (namespaces[o.selector]) {
                    $visibleMenu = $('.context-menu-list').filter(':visible');
                    if ($visibleMenu.length && $visibleMenu.data().contextMenuRoot.$trigger.is(o.selector)) {
                        $visibleMenu.trigger('contextmenu:hide', {force: true});
                    }

                    try {
                        if (menus[namespaces[o.selector]].$menu) {
                            menus[namespaces[o.selector]].$menu.remove();
                        }

                        delete menus[namespaces[o.selector]];
                    } catch (e) {
                        menus[namespaces[o.selector]] = null;
                    }

                    $document.off(namespaces[o.selector]);
                }
                break;

            case 'html5':
                // if <command> or <menuitem> are not handled by the browser,
                // or options was a bool true,
                // initialize $.contextMenu for them
                if ((!$.support.htmlCommand && !$.support.htmlMenuitem) || (typeof options === 'boolean' && options)) {
                    $('menu[type="context"]').each(function () {
                        if (this.id) {
                            $.contextMenu({
                                selector: '[contextmenu=' + this.id + ']',
                                items: $.contextMenu.fromMenu(this)
                            });
                        }
                    }).css('display', 'none');
                }
                break;

            default:
                throw new Error('Unknown operation "' + operation + '"');
        }

        return this;
    };

// import values into <input> commands
    $.contextMenu.setInputValues = function (opt, data) {
        if (data === undefined) {
            data = {};
        }

        $.each(opt.inputs, function (key, item) {
            switch (item.type) {
                case 'text':
                case 'textarea':
                    item.value = data[key] || '';
                    break;

                case 'checkbox':
                    item.selected = data[key] ? true : false;
                    break;

                case 'radio':
                    item.selected = (data[item.radio] || '') === item.value;
                    break;

                case 'select':
                    item.selected = data[key] || '';
                    break;
            }
        });
    };

// export values from <input> commands
    $.contextMenu.getInputValues = function (opt, data) {
        if (data === undefined) {
            data = {};
        }

        $.each(opt.inputs, function (key, item) {
            switch (item.type) {
                case 'text':
                case 'textarea':
                case 'select':
                    data[key] = item.$input.val();
                    break;

                case 'checkbox':
                    data[key] = item.$input.prop('checked');
                    break;

                case 'radio':
                    if (item.$input.prop('checked')) {
                        data[item.radio] = item.value;
                    }
                    break;
            }
        });

        return data;
    };

// find <label for="xyz">
    function inputLabel(node) {
        return (node.id && $('label[for="' + node.id + '"]').val()) || node.name;
    }

// convert <menu> to items object
    function menuChildren(items, $children, counter) {
        if (!counter) {
            counter = 0;
        }

        $children.each(function () {
            var $node = $(this),
                node = this,
                nodeName = this.nodeName.toLowerCase(),
                label,
                item;

            // extract <label><input>
            if (nodeName === 'label' && $node.find('input, textarea, select').length) {
                label = $node.text();
                $node = $node.children().first();
                node = $node.get(0);
                nodeName = node.nodeName.toLowerCase();
            }

            /*
             * <menu> accepts flow-content as children. that means <embed>, <canvas> and such are valid menu items.
             * Not being the sadistic kind, $.contextMenu only accepts:
             * <command>, <menuitem>, <hr>, <span>, <p> <input [text, radio, checkbox]>, <textarea>, <select> and of course <menu>.
             * Everything else will be imported as an html node, which is not interfaced with contextMenu.
             */

            // http://www.whatwg.org/specs/web-apps/current-work/multipage/commands.html#concept-command
            switch (nodeName) {
                // http://www.whatwg.org/specs/web-apps/current-work/multipage/interactive-elements.html#the-menu-element
                case 'menu':
                    item = {name: $node.attr('label'), items: {}};
                    counter = menuChildren(item.items, $node.children(), counter);
                    break;

                // http://www.whatwg.org/specs/web-apps/current-work/multipage/commands.html#using-the-a-element-to-define-a-command
                case 'a':
                // http://www.whatwg.org/specs/web-apps/current-work/multipage/commands.html#using-the-button-element-to-define-a-command
                case 'button':
                    item = {
                        name: $node.text(),
                        disabled: !!$node.attr('disabled'),
                        callback: (function () {
                            return function () {
                                $node.click();
                            };
                        })()
                    };
                    break;

                // http://www.whatwg.org/specs/web-apps/current-work/multipage/commands.html#using-the-command-element-to-define-a-command

                case 'menuitem':
                case 'command':
                    switch ($node.attr('type')) {
                        case undefined:
                        case 'command':
                        case 'menuitem':
                            item = {
                                name: $node.attr('label'),
                                disabled: !!$node.attr('disabled'),
                                icon: $node.attr('icon'),
                                callback: (function () {
                                    return function () {
                                        $node.click();
                                    };
                                })()
                            };
                            break;

                        case 'checkbox':
                            item = {
                                type: 'checkbox',
                                disabled: !!$node.attr('disabled'),
                                name: $node.attr('label'),
                                selected: !!$node.attr('checked')
                            };
                            break;
                        case 'radio':
                            item = {
                                type: 'radio',
                                disabled: !!$node.attr('disabled'),
                                name: $node.attr('label'),
                                radio: $node.attr('radiogroup'),
                                value: $node.attr('id'),
                                selected: !!$node.attr('checked')
                            };
                            break;

                        default:
                            item = undefined;
                    }
                    break;

                case 'hr':
                    item = '-------';
                    break;

                case 'input':
                    switch ($node.attr('type')) {
                        case 'text':
                            item = {
                                type: 'text',
                                name: label || inputLabel(node),
                                disabled: !!$node.attr('disabled'),
                                value: $node.val()
                            };
                            break;

                        case 'checkbox':
                            item = {
                                type: 'checkbox',
                                name: label || inputLabel(node),
                                disabled: !!$node.attr('disabled'),
                                selected: !!$node.attr('checked')
                            };
                            break;

                        case 'radio':
                            item = {
                                type: 'radio',
                                name: label || inputLabel(node),
                                disabled: !!$node.attr('disabled'),
                                radio: !!$node.attr('name'),
                                value: $node.val(),
                                selected: !!$node.attr('checked')
                            };
                            break;

                        default:
                            item = undefined;
                            break;
                    }
                    break;

                case 'select':
                    item = {
                        type: 'select',
                        name: label || inputLabel(node),
                        disabled: !!$node.attr('disabled'),
                        selected: $node.val(),
                        options: {}
                    };
                    $node.children().each(function () {
                        item.options[this.value] = $(this).text();
                    });
                    break;

                case 'textarea':
                    item = {
                        type: 'textarea',
                        name: label || inputLabel(node),
                        disabled: !!$node.attr('disabled'),
                        value: $node.val()
                    };
                    break;

                case 'label':
                    break;

                default:
                    item = {type: 'html', html: $node.clone(true)};
                    break;
            }

            if (item) {
                counter++;
                items['key' + counter] = item;
            }
        });

        return counter;
    }

// convert html5 menu
    $.contextMenu.fromMenu = function (element) {
        var $this = $(element),
            items = {};

        menuChildren(items, $this.children());

        return items;
    };

// make defaults accessible
    $.contextMenu.defaults = defaults;
    $.contextMenu.types = types;
// export internal functions - undocumented, for hacking only!
    $.contextMenu.handle = handle;
    $.contextMenu.op = op;
    $.contextMenu.menus = menus;


});
/* Javascript plotting library for jQuery, version 0.8.3.

Copyright (c) 2007-2014 IOLA and Ole Laursen.
Licensed under the MIT license.

*/

// first an inline dependency, jquery.colorhelpers.js, we inline it here
// for convenience

/* Plugin for jQuery for working with colors.
 *
 * Version 1.1.
 *
 * Inspiration from jQuery color animation plugin by John Resig.
 *
 * Released under the MIT license by Ole Laursen, October 2009.
 *
 * Examples:
 *
 *   $.color.parse("#fff").scale('rgb', 0.25).add('a', -0.5).toString()
 *   var c = $.color.extract($("#mydiv"), 'background-color');
 *   console.log(c.r, c.g, c.b, c.a);
 *   $.color.make(100, 50, 25, 0.4).toString() // returns "rgba(100,50,25,0.4)"
 *
 * Note that .scale() and .add() return the same modified object
 * instead of making a new one.
 *
 * V. 1.1: Fix error handling so e.g. parsing an empty string does
 * produce a color rather than just crashing.
 */
(function($){$.color={};$.color.make=function(r,g,b,a){var o={};o.r=r||0;o.g=g||0;o.b=b||0;o.a=a!=null?a:1;o.add=function(c,d){for(var i=0;i<c.length;++i)o[c.charAt(i)]+=d;return o.normalize()};o.scale=function(c,f){for(var i=0;i<c.length;++i)o[c.charAt(i)]*=f;return o.normalize()};o.toString=function(){if(o.a>=1){return"rgb("+[o.r,o.g,o.b].join(",")+")"}else{return"rgba("+[o.r,o.g,o.b,o.a].join(",")+")"}};o.normalize=function(){function clamp(min,value,max){return value<min?min:value>max?max:value}o.r=clamp(0,parseInt(o.r),255);o.g=clamp(0,parseInt(o.g),255);o.b=clamp(0,parseInt(o.b),255);o.a=clamp(0,o.a,1);return o};o.clone=function(){return $.color.make(o.r,o.b,o.g,o.a)};return o.normalize()};$.color.extract=function(elem,css){var c;do{c=elem.css(css).toLowerCase();if(c!=""&&c!="transparent")break;elem=elem.parent()}while(elem.length&&!$.nodeName(elem.get(0),"body"));if(c=="rgba(0, 0, 0, 0)")c="transparent";return $.color.parse(c)};$.color.parse=function(str){var res,m=$.color.make;if(res=/rgb\(\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*\)/.exec(str))return m(parseInt(res[1],10),parseInt(res[2],10),parseInt(res[3],10));if(res=/rgba\(\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*,\s*([0-9]+(?:\.[0-9]+)?)\s*\)/.exec(str))return m(parseInt(res[1],10),parseInt(res[2],10),parseInt(res[3],10),parseFloat(res[4]));if(res=/rgb\(\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*\)/.exec(str))return m(parseFloat(res[1])*2.55,parseFloat(res[2])*2.55,parseFloat(res[3])*2.55);if(res=/rgba\(\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\%\s*,\s*([0-9]+(?:\.[0-9]+)?)\s*\)/.exec(str))return m(parseFloat(res[1])*2.55,parseFloat(res[2])*2.55,parseFloat(res[3])*2.55,parseFloat(res[4]));if(res=/#([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})/.exec(str))return m(parseInt(res[1],16),parseInt(res[2],16),parseInt(res[3],16));if(res=/#([a-fA-F0-9])([a-fA-F0-9])([a-fA-F0-9])/.exec(str))return m(parseInt(res[1]+res[1],16),parseInt(res[2]+res[2],16),parseInt(res[3]+res[3],16));var name=$.trim(str).toLowerCase();if(name=="transparent")return m(255,255,255,0);else{res=lookupColors[name]||[0,0,0];return m(res[0],res[1],res[2])}};var lookupColors={aqua:[0,255,255],azure:[240,255,255],beige:[245,245,220],black:[0,0,0],blue:[0,0,255],brown:[165,42,42],cyan:[0,255,255],darkblue:[0,0,139],darkcyan:[0,139,139],darkgrey:[169,169,169],darkgreen:[0,100,0],darkkhaki:[189,183,107],darkmagenta:[139,0,139],darkolivegreen:[85,107,47],darkorange:[255,140,0],darkorchid:[153,50,204],darkred:[139,0,0],darksalmon:[233,150,122],darkviolet:[148,0,211],fuchsia:[255,0,255],gold:[255,215,0],green:[0,128,0],indigo:[75,0,130],khaki:[240,230,140],lightblue:[173,216,230],lightcyan:[224,255,255],lightgreen:[144,238,144],lightgrey:[211,211,211],lightpink:[255,182,193],lightyellow:[255,255,224],lime:[0,255,0],magenta:[255,0,255],maroon:[128,0,0],navy:[0,0,128],olive:[128,128,0],orange:[255,165,0],pink:[255,192,203],purple:[128,0,128],violet:[128,0,128],red:[255,0,0],silver:[192,192,192],white:[255,255,255],yellow:[255,255,0]}})(jQuery);

// the actual Flot code
(function($) {

	// Cache the prototype hasOwnProperty for faster access

	var hasOwnProperty = Object.prototype.hasOwnProperty;

    // A shim to provide 'detach' to jQuery versions prior to 1.4.  Using a DOM
    // operation produces the same effect as detach, i.e. removing the element
    // without touching its jQuery data.

    // Do not merge this into Flot 0.9, since it requires jQuery 1.4.4+.

    if (!$.fn.detach) {
        $.fn.detach = function() {
            return this.each(function() {
                if (this.parentNode) {
                    this.parentNode.removeChild( this );
                }
            });
        };
    }

	///////////////////////////////////////////////////////////////////////////
	// The Canvas object is a wrapper around an HTML5 <canvas> tag.
	//
	// @constructor
	// @param {string} cls List of classes to apply to the canvas.
	// @param {element} container Element onto which to append the canvas.
	//
	// Requiring a container is a little iffy, but unfortunately canvas
	// operations don't work unless the canvas is attached to the DOM.

	function Canvas(cls, container) {

		var element = container.children("." + cls)[0];

		if (element == null) {

			element = document.createElement("canvas");
			element.className = cls;

			$(element).css({ direction: "ltr", position: "absolute", left: 0, top: 0 })
				.appendTo(container);

			// If HTML5 Canvas isn't available, fall back to [Ex|Flash]canvas

			if (!element.getContext) {
				if (window.G_vmlCanvasManager) {
					element = window.G_vmlCanvasManager.initElement(element);
				} else {
					throw new Error("Canvas is not available. If you're using IE with a fall-back such as Excanvas, then there's either a mistake in your conditional include, or the page has no DOCTYPE and is rendering in Quirks Mode.");
				}
			}
		}

		this.element = element;

		var context = this.context = element.getContext("2d");

		// Determine the screen's ratio of physical to device-independent
		// pixels.  This is the ratio between the canvas width that the browser
		// advertises and the number of pixels actually present in that space.

		// The iPhone 4, for example, has a device-independent width of 320px,
		// but its screen is actually 640px wide.  It therefore has a pixel
		// ratio of 2, while most normal devices have a ratio of 1.

		var devicePixelRatio = window.devicePixelRatio || 1,
			backingStoreRatio =
				context.webkitBackingStorePixelRatio ||
				context.mozBackingStorePixelRatio ||
				context.msBackingStorePixelRatio ||
				context.oBackingStorePixelRatio ||
				context.backingStorePixelRatio || 1;

		this.pixelRatio = devicePixelRatio / backingStoreRatio;

		// Size the canvas to match the internal dimensions of its container

		this.resize(container.width(), container.height());

		// Collection of HTML div layers for text overlaid onto the canvas

		this.textContainer = null;
		this.text = {};

		// Cache of text fragments and metrics, so we can avoid expensively
		// re-calculating them when the plot is re-rendered in a loop.

		this._textCache = {};
	}

	// Resizes the canvas to the given dimensions.
	//
	// @param {number} width New width of the canvas, in pixels.
	// @param {number} width New height of the canvas, in pixels.

	Canvas.prototype.resize = function(width, height) {

		if (width <= 0 || height <= 0) {
			throw new Error("Invalid dimensions for plot, width = " + width + ", height = " + height);
		}

		var element = this.element,
			context = this.context,
			pixelRatio = this.pixelRatio;

		// Resize the canvas, increasing its density based on the display's
		// pixel ratio; basically giving it more pixels without increasing the
		// size of its element, to take advantage of the fact that retina
		// displays have that many more pixels in the same advertised space.

		// Resizing should reset the state (excanvas seems to be buggy though)

		if (this.width != width) {
			element.width = width * pixelRatio;
			element.style.width = width + "px";
			this.width = width;
		}

		if (this.height != height) {
			element.height = height * pixelRatio;
			element.style.height = height + "px";
			this.height = height;
		}

		// Save the context, so we can reset in case we get replotted.  The
		// restore ensure that we're really back at the initial state, and
		// should be safe even if we haven't saved the initial state yet.

		context.restore();
		context.save();

		// Scale the coordinate space to match the display density; so even though we
		// may have twice as many pixels, we still want lines and other drawing to
		// appear at the same size; the extra pixels will just make them crisper.

		context.scale(pixelRatio, pixelRatio);
	};

	// Clears the entire canvas area, not including any overlaid HTML text

	Canvas.prototype.clear = function() {
		this.context.clearRect(0, 0, this.width, this.height);
	};

	// Finishes rendering the canvas, including managing the text overlay.

	Canvas.prototype.render = function() {

		var cache = this._textCache;

		// For each text layer, add elements marked as active that haven't
		// already been rendered, and remove those that are no longer active.

		for (var layerKey in cache) {
			if (hasOwnProperty.call(cache, layerKey)) {

				var layer = this.getTextLayer(layerKey),
					layerCache = cache[layerKey];

				layer.hide();

				for (var styleKey in layerCache) {
					if (hasOwnProperty.call(layerCache, styleKey)) {
						var styleCache = layerCache[styleKey];
						for (var key in styleCache) {
							if (hasOwnProperty.call(styleCache, key)) {

								var positions = styleCache[key].positions;

								for (var i = 0, position; position = positions[i]; i++) {
									if (position.active) {
										if (!position.rendered) {
											layer.append(position.element);
											position.rendered = true;
										}
									} else {
										positions.splice(i--, 1);
										if (position.rendered) {
											position.element.detach();
										}
									}
								}

								if (positions.length == 0) {
									delete styleCache[key];
								}
							}
						}
					}
				}

				layer.show();
			}
		}
	};

	// Creates (if necessary) and returns the text overlay container.
	//
	// @param {string} classes String of space-separated CSS classes used to
	//     uniquely identify the text layer.
	// @return {object} The jQuery-wrapped text-layer div.

	Canvas.prototype.getTextLayer = function(classes) {

		var layer = this.text[classes];

		// Create the text layer if it doesn't exist

		if (layer == null) {

			// Create the text layer container, if it doesn't exist

			if (this.textContainer == null) {
				this.textContainer = $("<div class='flot-text'></div>")
					.css({
						position: "absolute",
						top: 0,
						left: 0,
						bottom: 0,
						right: 0,
						'font-size': "smaller",
						color: "#545454"
					})
					.insertAfter(this.element);
			}

			layer = this.text[classes] = $("<div></div>")
				.addClass(classes)
				.css({
					position: "absolute",
					top: 0,
					left: 0,
					bottom: 0,
					right: 0
				})
				.appendTo(this.textContainer);
		}

		return layer;
	};

	// Creates (if necessary) and returns a text info object.
	//
	// The object looks like this:
	//
	// {
	//     width: Width of the text's wrapper div.
	//     height: Height of the text's wrapper div.
	//     element: The jQuery-wrapped HTML div containing the text.
	//     positions: Array of positions at which this text is drawn.
	// }
	//
	// The positions array contains objects that look like this:
	//
	// {
	//     active: Flag indicating whether the text should be visible.
	//     rendered: Flag indicating whether the text is currently visible.
	//     element: The jQuery-wrapped HTML div containing the text.
	//     x: X coordinate at which to draw the text.
	//     y: Y coordinate at which to draw the text.
	// }
	//
	// Each position after the first receives a clone of the original element.
	//
	// The idea is that that the width, height, and general 'identity' of the
	// text is constant no matter where it is placed; the placements are a
	// secondary property.
	//
	// Canvas maintains a cache of recently-used text info objects; getTextInfo
	// either returns the cached element or creates a new entry.
	//
	// @param {string} layer A string of space-separated CSS classes uniquely
	//     identifying the layer containing this text.
	// @param {string} text Text string to retrieve info for.
	// @param {(string|object)=} font Either a string of space-separated CSS
	//     classes or a font-spec object, defining the text's font and style.
	// @param {number=} angle Angle at which to rotate the text, in degrees.
	//     Angle is currently unused, it will be implemented in the future.
	// @param {number=} width Maximum width of the text before it wraps.
	// @return {object} a text info object.

	Canvas.prototype.getTextInfo = function(layer, text, font, angle, width) {

		var textStyle, layerCache, styleCache, info;

		// Cast the value to a string, in case we were given a number or such

		text = "" + text;

		// If the font is a font-spec object, generate a CSS font definition

		if (typeof font === "object") {
			textStyle = font.style + " " + font.variant + " " + font.weight + " " + font.size + "px/" + font.lineHeight + "px " + font.family;
		} else {
			textStyle = font;
		}

		// Retrieve (or create) the cache for the text's layer and styles

		layerCache = this._textCache[layer];

		if (layerCache == null) {
			layerCache = this._textCache[layer] = {};
		}

		styleCache = layerCache[textStyle];

		if (styleCache == null) {
			styleCache = layerCache[textStyle] = {};
		}

		info = styleCache[text];

		// If we can't find a matching element in our cache, create a new one

		if (info == null) {

			var element = $("<div></div>").html(text)
				.css({
					position: "absolute",
					'max-width': width,
					top: -9999
				})
				.appendTo(this.getTextLayer(layer));

			if (typeof font === "object") {
				element.css({
					font: textStyle,
					color: font.color
				});
			} else if (typeof font === "string") {
				element.addClass(font);
			}

			info = styleCache[text] = {
				width: element.outerWidth(true),
				height: element.outerHeight(true),
				element: element,
				positions: []
			};

			element.detach();
		}

		return info;
	};

	// Adds a text string to the canvas text overlay.
	//
	// The text isn't drawn immediately; it is marked as rendering, which will
	// result in its addition to the canvas on the next render pass.
	//
	// @param {string} layer A string of space-separated CSS classes uniquely
	//     identifying the layer containing this text.
	// @param {number} x X coordinate at which to draw the text.
	// @param {number} y Y coordinate at which to draw the text.
	// @param {string} text Text string to draw.
	// @param {(string|object)=} font Either a string of space-separated CSS
	//     classes or a font-spec object, defining the text's font and style.
	// @param {number=} angle Angle at which to rotate the text, in degrees.
	//     Angle is currently unused, it will be implemented in the future.
	// @param {number=} width Maximum width of the text before it wraps.
	// @param {string=} halign Horizontal alignment of the text; either "left",
	//     "center" or "right".
	// @param {string=} valign Vertical alignment of the text; either "top",
	//     "middle" or "bottom".

	Canvas.prototype.addText = function(layer, x, y, text, font, angle, width, halign, valign) {

		var info = this.getTextInfo(layer, text, font, angle, width),
			positions = info.positions;

		// Tweak the div's position to match the text's alignment

		if (halign == "center") {
			x -= info.width / 2;
		} else if (halign == "right") {
			x -= info.width;
		}

		if (valign == "middle") {
			y -= info.height / 2;
		} else if (valign == "bottom") {
			y -= info.height;
		}

		// Determine whether this text already exists at this position.
		// If so, mark it for inclusion in the next render pass.

		for (var i = 0, position; position = positions[i]; i++) {
			if (position.x == x && position.y == y) {
				position.active = true;
				return;
			}
		}

		// If the text doesn't exist at this position, create a new entry

		// For the very first position we'll re-use the original element,
		// while for subsequent ones we'll clone it.

		position = {
			active: true,
			rendered: false,
			element: positions.length ? info.element.clone() : info.element,
			x: x,
			y: y
		};

		positions.push(position);

		// Move the element to its final position within the container

		position.element.css({
			top: Math.round(y),
			left: Math.round(x),
			'text-align': halign	// In case the text wraps
		});
	};

	// Removes one or more text strings from the canvas text overlay.
	//
	// If no parameters are given, all text within the layer is removed.
	//
	// Note that the text is not immediately removed; it is simply marked as
	// inactive, which will result in its removal on the next render pass.
	// This avoids the performance penalty for 'clear and redraw' behavior,
	// where we potentially get rid of all text on a layer, but will likely
	// add back most or all of it later, as when redrawing axes, for example.
	//
	// @param {string} layer A string of space-separated CSS classes uniquely
	//     identifying the layer containing this text.
	// @param {number=} x X coordinate of the text.
	// @param {number=} y Y coordinate of the text.
	// @param {string=} text Text string to remove.
	// @param {(string|object)=} font Either a string of space-separated CSS
	//     classes or a font-spec object, defining the text's font and style.
	// @param {number=} angle Angle at which the text is rotated, in degrees.
	//     Angle is currently unused, it will be implemented in the future.

	Canvas.prototype.removeText = function(layer, x, y, text, font, angle) {
		if (text == null) {
			var layerCache = this._textCache[layer];
			if (layerCache != null) {
				for (var styleKey in layerCache) {
					if (hasOwnProperty.call(layerCache, styleKey)) {
						var styleCache = layerCache[styleKey];
						for (var key in styleCache) {
							if (hasOwnProperty.call(styleCache, key)) {
								var positions = styleCache[key].positions;
								for (var i = 0, position; position = positions[i]; i++) {
									position.active = false;
								}
							}
						}
					}
				}
			}
		} else {
			var positions = this.getTextInfo(layer, text, font, angle).positions;
			for (var i = 0, position; position = positions[i]; i++) {
				if (position.x == x && position.y == y) {
					position.active = false;
				}
			}
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// The top-level container for the entire plot.

    function Plot(placeholder, data_, options_, plugins) {
        // data is on the form:
        //   [ series1, series2 ... ]
        // where series is either just the data as [ [x1, y1], [x2, y2], ... ]
        // or { data: [ [x1, y1], [x2, y2], ... ], label: "some label", ... }

        var series = [],
            options = {
                // the color theme used for graphs
                colors: ["#edc240", "#afd8f8", "#cb4b4b", "#4da74d", "#9440ed"],
                legend: {
                    show: true,
                    noColumns: 1, // number of colums in legend table
                    labelFormatter: null, // fn: string -> string
                    labelBoxBorderColor: "#ccc", // border color for the little label boxes
                    container: null, // container (as jQuery object) to put legend in, null means default on top of graph
                    position: "ne", // position of default legend container within plot
                    margin: 5, // distance from grid edge to default legend container within plot
                    backgroundColor: null, // null means auto-detect
                    backgroundOpacity: 0.85, // set to 0 to avoid background
                    sorted: null    // default to no legend sorting
                },
                xaxis: {
                    show: null, // null = auto-detect, true = always, false = never
                    position: "bottom", // or "top"
                    mode: null, // null or "time"
                    font: null, // null (derived from CSS in placeholder) or object like { size: 11, lineHeight: 13, style: "italic", weight: "bold", family: "sans-serif", variant: "small-caps" }
                    color: null, // base color, labels, ticks
                    tickColor: null, // possibly different color of ticks, e.g. "rgba(0,0,0,0.15)"
                    transform: null, // null or f: number -> number to transform axis
                    inverseTransform: null, // if transform is set, this should be the inverse function
                    min: null, // min. value to show, null means set automatically
                    max: null, // max. value to show, null means set automatically
                    autoscaleMargin: null, // margin in % to add if auto-setting min/max
                    ticks: null, // either [1, 3] or [[1, "a"], 3] or (fn: axis info -> ticks) or app. number of ticks for auto-ticks
                    tickFormatter: null, // fn: number -> string
                    labelWidth: null, // size of tick labels in pixels
                    labelHeight: null,
                    reserveSpace: null, // whether to reserve space even if axis isn't shown
                    tickLength: null, // size in pixels of ticks, or "full" for whole line
                    alignTicksWithAxis: null, // axis number or null for no sync
                    tickDecimals: null, // no. of decimals, null means auto
                    tickSize: null, // number or [number, "unit"]
                    minTickSize: null // number or [number, "unit"]
                },
                yaxis: {
                    autoscaleMargin: 0.02,
                    position: "left" // or "right"
                },
                xaxes: [],
                yaxes: [],
                series: {
                    points: {
                        show: false,
                        radius: 3,
                        lineWidth: 2, // in pixels
                        fill: true,
                        fillColor: "#ffffff",
                        symbol: "circle" // or callback
                    },
                    lines: {
                        // we don't put in show: false so we can see
                        // whether lines were actively disabled
                        lineWidth: 2, // in pixels
                        fill: false,
                        fillColor: null,
                        steps: false
                        // Omit 'zero', so we can later default its value to
                        // match that of the 'fill' option.
                    },
                    bars: {
                        show: false,
                        lineWidth: 2, // in pixels
                        barWidth: 1, // in units of the x axis
                        fill: true,
                        fillColor: null,
                        align: "left", // "left", "right", or "center"
                        horizontal: false,
                        zero: true
                    },
                    shadowSize: 3,
                    highlightColor: null
                },
                grid: {
                    show: true,
                    aboveData: false,
                    color: "#545454", // primary color used for outline and labels
                    backgroundColor: null, // null for transparent, else color
                    borderColor: null, // set if different from the grid color
                    tickColor: null, // color for the ticks, e.g. "rgba(0,0,0,0.15)"
                    margin: 0, // distance from the canvas edge to the grid
                    labelMargin: 5, // in pixels
                    axisMargin: 8, // in pixels
                    borderWidth: 2, // in pixels
                    minBorderMargin: null, // in pixels, null means taken from points radius
                    markings: null, // array of ranges or fn: axes -> array of ranges
                    markingsColor: "#f4f4f4",
                    markingsLineWidth: 2,
                    // interactive stuff
                    clickable: false,
                    hoverable: false,
                    autoHighlight: true, // highlight in case mouse is near
                    mouseActiveRadius: 10 // how far the mouse can be away to activate an item
                },
                interaction: {
                    redrawOverlayInterval: 1000/60 // time between updates, -1 means in same flow
                },
                hooks: {}
            },
        surface = null,     // the canvas for the plot itself
        overlay = null,     // canvas for interactive stuff on top of plot
        eventHolder = null, // jQuery object that events should be bound to
        ctx = null, octx = null,
        xaxes = [], yaxes = [],
        plotOffset = { left: 0, right: 0, top: 0, bottom: 0},
        plotWidth = 0, plotHeight = 0,
        hooks = {
            processOptions: [],
            processRawData: [],
            processDatapoints: [],
            processOffset: [],
            drawBackground: [],
            drawSeries: [],
            draw: [],
            bindEvents: [],
            drawOverlay: [],
            shutdown: []
        },
        plot = this;

        // public functions
        plot.setData = setData;
        plot.setupGrid = setupGrid;
        plot.draw = draw;
        plot.getPlaceholder = function() { return placeholder; };
        plot.getCanvas = function() { return surface.element; };
        plot.getPlotOffset = function() { return plotOffset; };
        plot.width = function () { return plotWidth; };
        plot.height = function () { return plotHeight; };
        plot.offset = function () {
            var o = eventHolder.offset();
            o.left += plotOffset.left;
            o.top += plotOffset.top;
            return o;
        };
        plot.getData = function () { return series; };
        plot.getAxes = function () {
            var res = {}, i;
            $.each(xaxes.concat(yaxes), function (_, axis) {
                if (axis)
                    res[axis.direction + (axis.n != 1 ? axis.n : "") + "axis"] = axis;
            });
            return res;
        };
        plot.getXAxes = function () { return xaxes; };
        plot.getYAxes = function () { return yaxes; };
        plot.c2p = canvasToAxisCoords;
        plot.p2c = axisToCanvasCoords;
        plot.getOptions = function () { return options; };
        plot.highlight = highlight;
        plot.unhighlight = unhighlight;
        plot.triggerRedrawOverlay = triggerRedrawOverlay;
        plot.pointOffset = function(point) {
            return {
                left: parseInt(xaxes[axisNumber(point, "x") - 1].p2c(+point.x) + plotOffset.left, 10),
                top: parseInt(yaxes[axisNumber(point, "y") - 1].p2c(+point.y) + plotOffset.top, 10)
            };
        };
        plot.shutdown = shutdown;
        plot.destroy = function () {
            shutdown();
            placeholder.removeData("plot").empty();

            series = [];
            options = null;
            surface = null;
            overlay = null;
            eventHolder = null;
            ctx = null;
            octx = null;
            xaxes = [];
            yaxes = [];
            hooks = null;
            highlights = [];
            plot = null;
        };
        plot.resize = function () {
        	var width = placeholder.width(),
        		height = placeholder.height();
            surface.resize(width, height);
            overlay.resize(width, height);
        };

        // public attributes
        plot.hooks = hooks;

        // initialize
        initPlugins(plot);
        parseOptions(options_);
        setupCanvases();
        setData(data_);
        setupGrid();
        draw();
        bindEvents();


        function executeHooks(hook, args) {
            args = [plot].concat(args);
            for (var i = 0; i < hook.length; ++i)
                hook[i].apply(this, args);
        }

        function initPlugins() {

            // References to key classes, allowing plugins to modify them

            var classes = {
                Canvas: Canvas
            };

            for (var i = 0; i < plugins.length; ++i) {
                var p = plugins[i];
                p.init(plot, classes);
                if (p.options)
                    $.extend(true, options, p.options);
            }
        }

        function parseOptions(opts) {

            $.extend(true, options, opts);

            // $.extend merges arrays, rather than replacing them.  When less
            // colors are provided than the size of the default palette, we
            // end up with those colors plus the remaining defaults, which is
            // not expected behavior; avoid it by replacing them here.

            if (opts && opts.colors) {
            	options.colors = opts.colors;
            }

            if (options.xaxis.color == null)
                options.xaxis.color = $.color.parse(options.grid.color).scale('a', 0.22).toString();
            if (options.yaxis.color == null)
                options.yaxis.color = $.color.parse(options.grid.color).scale('a', 0.22).toString();

            if (options.xaxis.tickColor == null) // grid.tickColor for back-compatibility
                options.xaxis.tickColor = options.grid.tickColor || options.xaxis.color;
            if (options.yaxis.tickColor == null) // grid.tickColor for back-compatibility
                options.yaxis.tickColor = options.grid.tickColor || options.yaxis.color;

            if (options.grid.borderColor == null)
                options.grid.borderColor = options.grid.color;
            if (options.grid.tickColor == null)
                options.grid.tickColor = $.color.parse(options.grid.color).scale('a', 0.22).toString();

            // Fill in defaults for axis options, including any unspecified
            // font-spec fields, if a font-spec was provided.

            // If no x/y axis options were provided, create one of each anyway,
            // since the rest of the code assumes that they exist.

            var i, axisOptions, axisCount,
                fontSize = placeholder.css("font-size"),
                fontSizeDefault = fontSize ? +fontSize.replace("px", "") : 13,
                fontDefaults = {
                    style: placeholder.css("font-style"),
                    size: Math.round(0.8 * fontSizeDefault),
                    variant: placeholder.css("font-variant"),
                    weight: placeholder.css("font-weight"),
                    family: placeholder.css("font-family")
                };

            axisCount = options.xaxes.length || 1;
            for (i = 0; i < axisCount; ++i) {

                axisOptions = options.xaxes[i];
                if (axisOptions && !axisOptions.tickColor) {
                    axisOptions.tickColor = axisOptions.color;
                }

                axisOptions = $.extend(true, {}, options.xaxis, axisOptions);
                options.xaxes[i] = axisOptions;

                if (axisOptions.font) {
                    axisOptions.font = $.extend({}, fontDefaults, axisOptions.font);
                    if (!axisOptions.font.color) {
                        axisOptions.font.color = axisOptions.color;
                    }
                    if (!axisOptions.font.lineHeight) {
                        axisOptions.font.lineHeight = Math.round(axisOptions.font.size * 1.15);
                    }
                }
            }

            axisCount = options.yaxes.length || 1;
            for (i = 0; i < axisCount; ++i) {

                axisOptions = options.yaxes[i];
                if (axisOptions && !axisOptions.tickColor) {
                    axisOptions.tickColor = axisOptions.color;
                }

                axisOptions = $.extend(true, {}, options.yaxis, axisOptions);
                options.yaxes[i] = axisOptions;

                if (axisOptions.font) {
                    axisOptions.font = $.extend({}, fontDefaults, axisOptions.font);
                    if (!axisOptions.font.color) {
                        axisOptions.font.color = axisOptions.color;
                    }
                    if (!axisOptions.font.lineHeight) {
                        axisOptions.font.lineHeight = Math.round(axisOptions.font.size * 1.15);
                    }
                }
            }

            // backwards compatibility, to be removed in future
            if (options.xaxis.noTicks && options.xaxis.ticks == null)
                options.xaxis.ticks = options.xaxis.noTicks;
            if (options.yaxis.noTicks && options.yaxis.ticks == null)
                options.yaxis.ticks = options.yaxis.noTicks;
            if (options.x2axis) {
                options.xaxes[1] = $.extend(true, {}, options.xaxis, options.x2axis);
                options.xaxes[1].position = "top";
                // Override the inherit to allow the axis to auto-scale
                if (options.x2axis.min == null) {
                    options.xaxes[1].min = null;
                }
                if (options.x2axis.max == null) {
                    options.xaxes[1].max = null;
                }
            }
            if (options.y2axis) {
                options.yaxes[1] = $.extend(true, {}, options.yaxis, options.y2axis);
                options.yaxes[1].position = "right";
                // Override the inherit to allow the axis to auto-scale
                if (options.y2axis.min == null) {
                    options.yaxes[1].min = null;
                }
                if (options.y2axis.max == null) {
                    options.yaxes[1].max = null;
                }
            }
            if (options.grid.coloredAreas)
                options.grid.markings = options.grid.coloredAreas;
            if (options.grid.coloredAreasColor)
                options.grid.markingsColor = options.grid.coloredAreasColor;
            if (options.lines)
                $.extend(true, options.series.lines, options.lines);
            if (options.points)
                $.extend(true, options.series.points, options.points);
            if (options.bars)
                $.extend(true, options.series.bars, options.bars);
            if (options.shadowSize != null)
                options.series.shadowSize = options.shadowSize;
            if (options.highlightColor != null)
                options.series.highlightColor = options.highlightColor;

            // save options on axes for future reference
            for (i = 0; i < options.xaxes.length; ++i)
                getOrCreateAxis(xaxes, i + 1).options = options.xaxes[i];
            for (i = 0; i < options.yaxes.length; ++i)
                getOrCreateAxis(yaxes, i + 1).options = options.yaxes[i];

            // add hooks from options
            for (var n in hooks)
                if (options.hooks[n] && options.hooks[n].length)
                    hooks[n] = hooks[n].concat(options.hooks[n]);

            executeHooks(hooks.processOptions, [options]);
        }

        function setData(d) {
            series = parseData(d);
            fillInSeriesOptions();
            processData();
        }

        function parseData(d) {
            var res = [];
            for (var i = 0; i < d.length; ++i) {
                var s = $.extend(true, {}, options.series);

                if (d[i].data != null) {
                    s.data = d[i].data; // move the data instead of deep-copy
                    delete d[i].data;

                    $.extend(true, s, d[i]);

                    d[i].data = s.data;
                }
                else
                    s.data = d[i];
                res.push(s);
            }

            return res;
        }

        function axisNumber(obj, coord) {
            var a = obj[coord + "axis"];
            if (typeof a == "object") // if we got a real axis, extract number
                a = a.n;
            if (typeof a != "number")
                a = 1; // default to first axis
            return a;
        }

        function allAxes() {
            // return flat array without annoying null entries
            return $.grep(xaxes.concat(yaxes), function (a) { return a; });
        }

        function canvasToAxisCoords(pos) {
            // return an object with x/y corresponding to all used axes
            var res = {}, i, axis;
            for (i = 0; i < xaxes.length; ++i) {
                axis = xaxes[i];
                if (axis && axis.used)
                    res["x" + axis.n] = axis.c2p(pos.left);
            }

            for (i = 0; i < yaxes.length; ++i) {
                axis = yaxes[i];
                if (axis && axis.used)
                    res["y" + axis.n] = axis.c2p(pos.top);
            }

            if (res.x1 !== undefined)
                res.x = res.x1;
            if (res.y1 !== undefined)
                res.y = res.y1;

            return res;
        }

        function axisToCanvasCoords(pos) {
            // get canvas coords from the first pair of x/y found in pos
            var res = {}, i, axis, key;

            for (i = 0; i < xaxes.length; ++i) {
                axis = xaxes[i];
                if (axis && axis.used) {
                    key = "x" + axis.n;
                    if (pos[key] == null && axis.n == 1)
                        key = "x";

                    if (pos[key] != null) {
                        res.left = axis.p2c(pos[key]);
                        break;
                    }
                }
            }

            for (i = 0; i < yaxes.length; ++i) {
                axis = yaxes[i];
                if (axis && axis.used) {
                    key = "y" + axis.n;
                    if (pos[key] == null && axis.n == 1)
                        key = "y";

                    if (pos[key] != null) {
                        res.top = axis.p2c(pos[key]);
                        break;
                    }
                }
            }

            return res;
        }

        function getOrCreateAxis(axes, number) {
            if (!axes[number - 1])
                axes[number - 1] = {
                    n: number, // save the number for future reference
                    direction: axes == xaxes ? "x" : "y",
                    options: $.extend(true, {}, axes == xaxes ? options.xaxis : options.yaxis)
                };

            return axes[number - 1];
        }

        function fillInSeriesOptions() {

            var neededColors = series.length, maxIndex = -1, i;

            // Subtract the number of series that already have fixed colors or
            // color indexes from the number that we still need to generate.

            for (i = 0; i < series.length; ++i) {
                var sc = series[i].color;
                if (sc != null) {
                    neededColors--;
                    if (typeof sc == "number" && sc > maxIndex) {
                        maxIndex = sc;
                    }
                }
            }

            // If any of the series have fixed color indexes, then we need to
            // generate at least as many colors as the highest index.

            if (neededColors <= maxIndex) {
                neededColors = maxIndex + 1;
            }

            // Generate all the colors, using first the option colors and then
            // variations on those colors once they're exhausted.

            var c, colors = [], colorPool = options.colors,
                colorPoolSize = colorPool.length, variation = 0;

            for (i = 0; i < neededColors; i++) {

                c = $.color.parse(colorPool[i % colorPoolSize] || "#666");

                // Each time we exhaust the colors in the pool we adjust
                // a scaling factor used to produce more variations on
                // those colors. The factor alternates negative/positive
                // to produce lighter/darker colors.

                // Reset the variation after every few cycles, or else
                // it will end up producing only white or black colors.

                if (i % colorPoolSize == 0 && i) {
                    if (variation >= 0) {
                        if (variation < 0.5) {
                            variation = -variation - 0.2;
                        } else variation = 0;
                    } else variation = -variation;
                }

                colors[i] = c.scale('rgb', 1 + variation);
            }

            // Finalize the series options, filling in their colors

            var colori = 0, s;
            for (i = 0; i < series.length; ++i) {
                s = series[i];

                // assign colors
                if (s.color == null) {
                    s.color = colors[colori].toString();
                    ++colori;
                }
                else if (typeof s.color == "number")
                    s.color = colors[s.color].toString();

                // turn on lines automatically in case nothing is set
                if (s.lines.show == null) {
                    var v, show = true;
                    for (v in s)
                        if (s[v] && s[v].show) {
                            show = false;
                            break;
                        }
                    if (show)
                        s.lines.show = true;
                }

                // If nothing was provided for lines.zero, default it to match
                // lines.fill, since areas by default should extend to zero.

                if (s.lines.zero == null) {
                    s.lines.zero = !!s.lines.fill;
                }

                // setup axes
                s.xaxis = getOrCreateAxis(xaxes, axisNumber(s, "x"));
                s.yaxis = getOrCreateAxis(yaxes, axisNumber(s, "y"));
            }
        }

        function processData() {
            var topSentry = Number.POSITIVE_INFINITY,
                bottomSentry = Number.NEGATIVE_INFINITY,
                fakeInfinity = Number.MAX_VALUE,
                i, j, k, m, length,
                s, points, ps, x, y, axis, val, f, p,
                data, format;

            function updateAxis(axis, min, max) {
                if (min < axis.datamin && min != -fakeInfinity)
                    axis.datamin = min;
                if (max > axis.datamax && max != fakeInfinity)
                    axis.datamax = max;
            }

            $.each(allAxes(), function (_, axis) {
                // init axis
                axis.datamin = topSentry;
                axis.datamax = bottomSentry;
                axis.used = false;
            });

            for (i = 0; i < series.length; ++i) {
                s = series[i];
                s.datapoints = { points: [] };

                executeHooks(hooks.processRawData, [ s, s.data, s.datapoints ]);
            }

            // first pass: clean and copy data
            for (i = 0; i < series.length; ++i) {
                s = series[i];

                data = s.data;
                format = s.datapoints.format;

                if (!format) {
                    format = [];
                    // find out how to copy
                    format.push({ x: true, number: true, required: true });
                    format.push({ y: true, number: true, required: true });

                    if (s.bars.show || (s.lines.show && s.lines.fill)) {
                        var autoscale = !!((s.bars.show && s.bars.zero) || (s.lines.show && s.lines.zero));
                        format.push({ y: true, number: true, required: false, defaultValue: 0, autoscale: autoscale });
                        if (s.bars.horizontal) {
                            delete format[format.length - 1].y;
                            format[format.length - 1].x = true;
                        }
                    }

                    s.datapoints.format = format;
                }

                if (s.datapoints.pointsize != null)
                    continue; // already filled in

                s.datapoints.pointsize = format.length;

                ps = s.datapoints.pointsize;
                points = s.datapoints.points;

                var insertSteps = s.lines.show && s.lines.steps;
                s.xaxis.used = s.yaxis.used = true;

                for (j = k = 0; j < data.length; ++j, k += ps) {
                    p = data[j];

                    var nullify = p == null;
                    if (!nullify) {
                        for (m = 0; m < ps; ++m) {
                            val = p[m];
                            f = format[m];

                            if (f) {
                                if (f.number && val != null) {
                                    val = +val; // convert to number
                                    if (isNaN(val))
                                        val = null;
                                    else if (val == Infinity)
                                        val = fakeInfinity;
                                    else if (val == -Infinity)
                                        val = -fakeInfinity;
                                }

                                if (val == null) {
                                    if (f.required)
                                        nullify = true;

                                    if (f.defaultValue != null)
                                        val = f.defaultValue;
                                }
                            }

                            points[k + m] = val;
                        }
                    }

                    if (nullify) {
                        for (m = 0; m < ps; ++m) {
                            val = points[k + m];
                            if (val != null) {
                                f = format[m];
                                // extract min/max info
                                if (f.autoscale !== false) {
                                    if (f.x) {
                                        updateAxis(s.xaxis, val, val);
                                    }
                                    if (f.y) {
                                        updateAxis(s.yaxis, val, val);
                                    }
                                }
                            }
                            points[k + m] = null;
                        }
                    }
                    else {
                        // a little bit of line specific stuff that
                        // perhaps shouldn't be here, but lacking
                        // better means...
                        if (insertSteps && k > 0
                            && points[k - ps] != null
                            && points[k - ps] != points[k]
                            && points[k - ps + 1] != points[k + 1]) {
                            // copy the point to make room for a middle point
                            for (m = 0; m < ps; ++m)
                                points[k + ps + m] = points[k + m];

                            // middle point has same y
                            points[k + 1] = points[k - ps + 1];

                            // we've added a point, better reflect that
                            k += ps;
                        }
                    }
                }
            }

            // give the hooks a chance to run
            for (i = 0; i < series.length; ++i) {
                s = series[i];

                executeHooks(hooks.processDatapoints, [ s, s.datapoints]);
            }

            // second pass: find datamax/datamin for auto-scaling
            for (i = 0; i < series.length; ++i) {
                s = series[i];
                points = s.datapoints.points;
                ps = s.datapoints.pointsize;
                format = s.datapoints.format;

                var xmin = topSentry, ymin = topSentry,
                    xmax = bottomSentry, ymax = bottomSentry;

                for (j = 0; j < points.length; j += ps) {
                    if (points[j] == null)
                        continue;

                    for (m = 0; m < ps; ++m) {
                        val = points[j + m];
                        f = format[m];
                        if (!f || f.autoscale === false || val == fakeInfinity || val == -fakeInfinity)
                            continue;

                        if (f.x) {
                            if (val < xmin)
                                xmin = val;
                            if (val > xmax)
                                xmax = val;
                        }
                        if (f.y) {
                            if (val < ymin)
                                ymin = val;
                            if (val > ymax)
                                ymax = val;
                        }
                    }
                }

                if (s.bars.show) {
                    // make sure we got room for the bar on the dancing floor
                    var delta;

                    switch (s.bars.align) {
                        case "left":
                            delta = 0;
                            break;
                        case "right":
                            delta = -s.bars.barWidth;
                            break;
                        default:
                            delta = -s.bars.barWidth / 2;
                    }

                    if (s.bars.horizontal) {
                        ymin += delta;
                        ymax += delta + s.bars.barWidth;
                    }
                    else {
                        xmin += delta;
                        xmax += delta + s.bars.barWidth;
                    }
                }

                updateAxis(s.xaxis, xmin, xmax);
                updateAxis(s.yaxis, ymin, ymax);
            }

            $.each(allAxes(), function (_, axis) {
                if (axis.datamin == topSentry)
                    axis.datamin = null;
                if (axis.datamax == bottomSentry)
                    axis.datamax = null;
            });
        }

        function setupCanvases() {

            // Make sure the placeholder is clear of everything except canvases
            // from a previous plot in this container that we'll try to re-use.

            placeholder.css("padding", 0) // padding messes up the positioning
                .children().filter(function(){
                    return !$(this).hasClass("flot-overlay") && !$(this).hasClass('flot-base');
                }).remove();

            if (placeholder.css("position") == 'static')
                placeholder.css("position", "relative"); // for positioning labels and overlay

            surface = new Canvas("flot-base", placeholder);
            overlay = new Canvas("flot-overlay", placeholder); // overlay canvas for interactive features

            ctx = surface.context;
            octx = overlay.context;

            // define which element we're listening for events on
            eventHolder = $(overlay.element).unbind();

            // If we're re-using a plot object, shut down the old one

            var existing = placeholder.data("plot");

            if (existing) {
                existing.shutdown();
                overlay.clear();
            }

            // save in case we get replotted
            placeholder.data("plot", plot);
        }

        function bindEvents() {
            // bind events
            if (options.grid.hoverable) {
                eventHolder.mousemove(onMouseMove);

                // Use bind, rather than .mouseleave, because we officially
                // still support jQuery 1.2.6, which doesn't define a shortcut
                // for mouseenter or mouseleave.  This was a bug/oversight that
                // was fixed somewhere around 1.3.x.  We can return to using
                // .mouseleave when we drop support for 1.2.6.

                eventHolder.bind("mouseleave", onMouseLeave);
            }

            if (options.grid.clickable)
                eventHolder.click(onClick);

            executeHooks(hooks.bindEvents, [eventHolder]);
        }

        function shutdown() {
            if (redrawTimeout)
                clearTimeout(redrawTimeout);

            eventHolder.unbind("mousemove", onMouseMove);
            eventHolder.unbind("mouseleave", onMouseLeave);
            eventHolder.unbind("click", onClick);

            executeHooks(hooks.shutdown, [eventHolder]);
        }

        function setTransformationHelpers(axis) {
            // set helper functions on the axis, assumes plot area
            // has been computed already

            function identity(x) { return x; }

            var s, m, t = axis.options.transform || identity,
                it = axis.options.inverseTransform;

            // precompute how much the axis is scaling a point
            // in canvas space
            if (axis.direction == "x") {
                s = axis.scale = plotWidth / Math.abs(t(axis.max) - t(axis.min));
                m = Math.min(t(axis.max), t(axis.min));
            }
            else {
                s = axis.scale = plotHeight / Math.abs(t(axis.max) - t(axis.min));
                s = -s;
                m = Math.max(t(axis.max), t(axis.min));
            }

            // data point to canvas coordinate
            if (t == identity) // slight optimization
                axis.p2c = function (p) { return (p - m) * s; };
            else
                axis.p2c = function (p) { return (t(p) - m) * s; };
            // canvas coordinate to data point
            if (!it)
                axis.c2p = function (c) { return m + c / s; };
            else
                axis.c2p = function (c) { return it(m + c / s); };
        }

        function measureTickLabels(axis) {

            var opts = axis.options,
                ticks = axis.ticks || [],
                labelWidth = opts.labelWidth || 0,
                labelHeight = opts.labelHeight || 0,
                maxWidth = labelWidth || (axis.direction == "x" ? Math.floor(surface.width / (ticks.length || 1)) : null),
                legacyStyles = axis.direction + "Axis " + axis.direction + axis.n + "Axis",
                layer = "flot-" + axis.direction + "-axis flot-" + axis.direction + axis.n + "-axis " + legacyStyles,
                font = opts.font || "flot-tick-label tickLabel";

            for (var i = 0; i < ticks.length; ++i) {

                var t = ticks[i];

                if (!t.label)
                    continue;

                var info = surface.getTextInfo(layer, t.label, font, null, maxWidth);

                labelWidth = Math.max(labelWidth, info.width);
                labelHeight = Math.max(labelHeight, info.height);
            }

            axis.labelWidth = opts.labelWidth || labelWidth;
            axis.labelHeight = opts.labelHeight || labelHeight;
        }

        function allocateAxisBoxFirstPhase(axis) {
            // find the bounding box of the axis by looking at label
            // widths/heights and ticks, make room by diminishing the
            // plotOffset; this first phase only looks at one
            // dimension per axis, the other dimension depends on the
            // other axes so will have to wait

            var lw = axis.labelWidth,
                lh = axis.labelHeight,
                pos = axis.options.position,
                isXAxis = axis.direction === "x",
                tickLength = axis.options.tickLength,
                axisMargin = options.grid.axisMargin,
                padding = options.grid.labelMargin,
                innermost = true,
                outermost = true,
                first = true,
                found = false;

            // Determine the axis's position in its direction and on its side

            $.each(isXAxis ? xaxes : yaxes, function(i, a) {
                if (a && (a.show || a.reserveSpace)) {
                    if (a === axis) {
                        found = true;
                    } else if (a.options.position === pos) {
                        if (found) {
                            outermost = false;
                        } else {
                            innermost = false;
                        }
                    }
                    if (!found) {
                        first = false;
                    }
                }
            });

            // The outermost axis on each side has no margin

            if (outermost) {
                axisMargin = 0;
            }

            // The ticks for the first axis in each direction stretch across

            if (tickLength == null) {
                tickLength = first ? "full" : 5;
            }

            if (!isNaN(+tickLength))
                padding += +tickLength;

            if (isXAxis) {
                lh += padding;

                if (pos == "bottom") {
                    plotOffset.bottom += lh + axisMargin;
                    axis.box = { top: surface.height - plotOffset.bottom, height: lh };
                }
                else {
                    axis.box = { top: plotOffset.top + axisMargin, height: lh };
                    plotOffset.top += lh + axisMargin;
                }
            }
            else {
                lw += padding;

                if (pos == "left") {
                    axis.box = { left: plotOffset.left + axisMargin, width: lw };
                    plotOffset.left += lw + axisMargin;
                }
                else {
                    plotOffset.right += lw + axisMargin;
                    axis.box = { left: surface.width - plotOffset.right, width: lw };
                }
            }

             // save for future reference
            axis.position = pos;
            axis.tickLength = tickLength;
            axis.box.padding = padding;
            axis.innermost = innermost;
        }

        function allocateAxisBoxSecondPhase(axis) {
            // now that all axis boxes have been placed in one
            // dimension, we can set the remaining dimension coordinates
            if (axis.direction == "x") {
                axis.box.left = plotOffset.left - axis.labelWidth / 2;
                axis.box.width = surface.width - plotOffset.left - plotOffset.right + axis.labelWidth;
            }
            else {
                axis.box.top = plotOffset.top - axis.labelHeight / 2;
                axis.box.height = surface.height - plotOffset.bottom - plotOffset.top + axis.labelHeight;
            }
        }

        function adjustLayoutForThingsStickingOut() {
            // possibly adjust plot offset to ensure everything stays
            // inside the canvas and isn't clipped off

            var minMargin = options.grid.minBorderMargin,
                axis, i;

            // check stuff from the plot (FIXME: this should just read
            // a value from the series, otherwise it's impossible to
            // customize)
            if (minMargin == null) {
                minMargin = 0;
                for (i = 0; i < series.length; ++i)
                    minMargin = Math.max(minMargin, 2 * (series[i].points.radius + series[i].points.lineWidth/2));
            }

            var margins = {
                left: minMargin,
                right: minMargin,
                top: minMargin,
                bottom: minMargin
            };

            // check axis labels, note we don't check the actual
            // labels but instead use the overall width/height to not
            // jump as much around with replots
            $.each(allAxes(), function (_, axis) {
                if (axis.reserveSpace && axis.ticks && axis.ticks.length) {
                    if (axis.direction === "x") {
                        margins.left = Math.max(margins.left, axis.labelWidth / 2);
                        margins.right = Math.max(margins.right, axis.labelWidth / 2);
                    } else {
                        margins.bottom = Math.max(margins.bottom, axis.labelHeight / 2);
                        margins.top = Math.max(margins.top, axis.labelHeight / 2);
                    }
                }
            });

            plotOffset.left = Math.ceil(Math.max(margins.left, plotOffset.left));
            plotOffset.right = Math.ceil(Math.max(margins.right, plotOffset.right));
            plotOffset.top = Math.ceil(Math.max(margins.top, plotOffset.top));
            plotOffset.bottom = Math.ceil(Math.max(margins.bottom, plotOffset.bottom));
        }

        function setupGrid() {
            var i, axes = allAxes(), showGrid = options.grid.show;

            // Initialize the plot's offset from the edge of the canvas

            for (var a in plotOffset) {
                var margin = options.grid.margin || 0;
                plotOffset[a] = typeof margin == "number" ? margin : margin[a] || 0;
            }

            executeHooks(hooks.processOffset, [plotOffset]);

            // If the grid is visible, add its border width to the offset

            for (var a in plotOffset) {
                if(typeof(options.grid.borderWidth) == "object") {
                    plotOffset[a] += showGrid ? options.grid.borderWidth[a] : 0;
                }
                else {
                    plotOffset[a] += showGrid ? options.grid.borderWidth : 0;
                }
            }

            $.each(axes, function (_, axis) {
                var axisOpts = axis.options;
                axis.show = axisOpts.show == null ? axis.used : axisOpts.show;
                axis.reserveSpace = axisOpts.reserveSpace == null ? axis.show : axisOpts.reserveSpace;
                setRange(axis);
            });

            if (showGrid) {

                var allocatedAxes = $.grep(axes, function (axis) {
                    return axis.show || axis.reserveSpace;
                });

                $.each(allocatedAxes, function (_, axis) {
                    // make the ticks
                    setupTickGeneration(axis);
                    setTicks(axis);
                    snapRangeToTicks(axis, axis.ticks);
                    // find labelWidth/Height for axis
                    measureTickLabels(axis);
                });

                // with all dimensions calculated, we can compute the
                // axis bounding boxes, start from the outside
                // (reverse order)
                for (i = allocatedAxes.length - 1; i >= 0; --i)
                    allocateAxisBoxFirstPhase(allocatedAxes[i]);

                // make sure we've got enough space for things that
                // might stick out
                adjustLayoutForThingsStickingOut();

                $.each(allocatedAxes, function (_, axis) {
                    allocateAxisBoxSecondPhase(axis);
                });
            }

            plotWidth = surface.width - plotOffset.left - plotOffset.right;
            plotHeight = surface.height - plotOffset.bottom - plotOffset.top;

            // now we got the proper plot dimensions, we can compute the scaling
            $.each(axes, function (_, axis) {
                setTransformationHelpers(axis);
            });

            if (showGrid) {
                drawAxisLabels();
            }

            insertLegend();
        }

        function setRange(axis) {
            var opts = axis.options,
                min = +(opts.min != null ? opts.min : axis.datamin),
                max = +(opts.max != null ? opts.max : axis.datamax),
                delta = max - min;

            if (delta == 0.0) {
                // degenerate case
                var widen = max == 0 ? 1 : 0.01;

                if (opts.min == null)
                    min -= widen;
                // always widen max if we couldn't widen min to ensure we
                // don't fall into min == max which doesn't work
                if (opts.max == null || opts.min != null)
                    max += widen;
            }
            else {
                // consider autoscaling
                var margin = opts.autoscaleMargin;
                if (margin != null) {
                    if (opts.min == null) {
                        min -= delta * margin;
                        // make sure we don't go below zero if all values
                        // are positive
                        if (min < 0 && axis.datamin != null && axis.datamin >= 0)
                            min = 0;
                    }
                    if (opts.max == null) {
                        max += delta * margin;
                        if (max > 0 && axis.datamax != null && axis.datamax <= 0)
                            max = 0;
                    }
                }
            }
            axis.min = min;
            axis.max = max;
        }

        function setupTickGeneration(axis) {
            var opts = axis.options;

            // estimate number of ticks
            var noTicks;
            if (typeof opts.ticks == "number" && opts.ticks > 0)
                noTicks = opts.ticks;
            else
                // heuristic based on the model a*sqrt(x) fitted to
                // some data points that seemed reasonable
                noTicks = 0.3 * Math.sqrt(axis.direction == "x" ? surface.width : surface.height);

            var delta = (axis.max - axis.min) / noTicks,
                dec = -Math.floor(Math.log(delta) / Math.LN10),
                maxDec = opts.tickDecimals;

            if (maxDec != null && dec > maxDec) {
                dec = maxDec;
            }

            var magn = Math.pow(10, -dec),
                norm = delta / magn, // norm is between 1.0 and 10.0
                size;

            if (norm < 1.5) {
                size = 1;
            } else if (norm < 3) {
                size = 2;
                // special case for 2.5, requires an extra decimal
                if (norm > 2.25 && (maxDec == null || dec + 1 <= maxDec)) {
                    size = 2.5;
                    ++dec;
                }
            } else if (norm < 7.5) {
                size = 5;
            } else {
                size = 10;
            }

            size *= magn;

            if (opts.minTickSize != null && size < opts.minTickSize) {
                size = opts.minTickSize;
            }

            axis.delta = delta;
            axis.tickDecimals = Math.max(0, maxDec != null ? maxDec : dec);
            axis.tickSize = opts.tickSize || size;

            // Time mode was moved to a plug-in in 0.8, and since so many people use it
            // we'll add an especially friendly reminder to make sure they included it.

            if (opts.mode == "time" && !axis.tickGenerator) {
                throw new Error("Time mode requires the flot.time plugin.");
            }

            // Flot supports base-10 axes; any other mode else is handled by a plug-in,
            // like flot.time.js.

            if (!axis.tickGenerator) {

                axis.tickGenerator = function (axis) {

                    var ticks = [],
                        start = floorInBase(axis.min, axis.tickSize),
                        i = 0,
                        v = Number.NaN,
                        prev;

                    do {
                        prev = v;
                        v = start + i * axis.tickSize;
                        ticks.push(v);
                        ++i;
                    } while (v < axis.max && v != prev);
                    return ticks;
                };

				axis.tickFormatter = function (value, axis) {

					var factor = axis.tickDecimals ? Math.pow(10, axis.tickDecimals) : 1;
					var formatted = "" + Math.round(value * factor) / factor;

					// If tickDecimals was specified, ensure that we have exactly that
					// much precision; otherwise default to the value's own precision.

					if (axis.tickDecimals != null) {
						var decimal = formatted.indexOf(".");
						var precision = decimal == -1 ? 0 : formatted.length - decimal - 1;
						if (precision < axis.tickDecimals) {
							return (precision ? formatted : formatted + ".") + ("" + factor).substr(1, axis.tickDecimals - precision);
						}
					}

                    return formatted;
                };
            }

            if ($.isFunction(opts.tickFormatter))
                axis.tickFormatter = function (v, axis) { return "" + opts.tickFormatter(v, axis); };

            if (opts.alignTicksWithAxis != null) {
                var otherAxis = (axis.direction == "x" ? xaxes : yaxes)[opts.alignTicksWithAxis - 1];
                if (otherAxis && otherAxis.used && otherAxis != axis) {
                    // consider snapping min/max to outermost nice ticks
                    var niceTicks = axis.tickGenerator(axis);
                    if (niceTicks.length > 0) {
                        if (opts.min == null)
                            axis.min = Math.min(axis.min, niceTicks[0]);
                        if (opts.max == null && niceTicks.length > 1)
                            axis.max = Math.max(axis.max, niceTicks[niceTicks.length - 1]);
                    }

                    axis.tickGenerator = function (axis) {
                        // copy ticks, scaled to this axis
                        var ticks = [], v, i;
                        for (i = 0; i < otherAxis.ticks.length; ++i) {
                            v = (otherAxis.ticks[i].v - otherAxis.min) / (otherAxis.max - otherAxis.min);
                            v = axis.min + v * (axis.max - axis.min);
                            ticks.push(v);
                        }
                        return ticks;
                    };

                    // we might need an extra decimal since forced
                    // ticks don't necessarily fit naturally
                    if (!axis.mode && opts.tickDecimals == null) {
                        var extraDec = Math.max(0, -Math.floor(Math.log(axis.delta) / Math.LN10) + 1),
                            ts = axis.tickGenerator(axis);

                        // only proceed if the tick interval rounded
                        // with an extra decimal doesn't give us a
                        // zero at end
                        if (!(ts.length > 1 && /\..*0$/.test((ts[1] - ts[0]).toFixed(extraDec))))
                            axis.tickDecimals = extraDec;
                    }
                }
            }
        }

        function setTicks(axis) {
            var oticks = axis.options.ticks, ticks = [];
            if (oticks == null || (typeof oticks == "number" && oticks > 0))
                ticks = axis.tickGenerator(axis);
            else if (oticks) {
                if ($.isFunction(oticks))
                    // generate the ticks
                    ticks = oticks(axis);
                else
                    ticks = oticks;
            }

            // clean up/labelify the supplied ticks, copy them over
            var i, v;
            axis.ticks = [];
            for (i = 0; i < ticks.length; ++i) {
                var label = null;
                var t = ticks[i];
                if (typeof t == "object") {
                    v = +t[0];
                    if (t.length > 1)
                        label = t[1];
                }
                else
                    v = +t;
                if (label == null)
                    label = axis.tickFormatter(v, axis);
                if (!isNaN(v))
                    axis.ticks.push({ v: v, label: label });
            }
        }

        function snapRangeToTicks(axis, ticks) {
            if (axis.options.autoscaleMargin && ticks.length > 0) {
                // snap to ticks
                if (axis.options.min == null)
                    axis.min = Math.min(axis.min, ticks[0].v);
                if (axis.options.max == null && ticks.length > 1)
                    axis.max = Math.max(axis.max, ticks[ticks.length - 1].v);
            }
        }

        function draw() {

            surface.clear();

            executeHooks(hooks.drawBackground, [ctx]);

            var grid = options.grid;

            // draw background, if any
            if (grid.show && grid.backgroundColor)
                drawBackground();

            if (grid.show && !grid.aboveData) {
                drawGrid();
            }

            for (var i = 0; i < series.length; ++i) {
                executeHooks(hooks.drawSeries, [ctx, series[i]]);
                drawSeries(series[i]);
            }

            executeHooks(hooks.draw, [ctx]);

            if (grid.show && grid.aboveData) {
                drawGrid();
            }

            surface.render();

            // A draw implies that either the axes or data have changed, so we
            // should probably update the overlay highlights as well.

            triggerRedrawOverlay();
        }

        function extractRange(ranges, coord) {
            var axis, from, to, key, axes = allAxes();

            for (var i = 0; i < axes.length; ++i) {
                axis = axes[i];
                if (axis.direction == coord) {
                    key = coord + axis.n + "axis";
                    if (!ranges[key] && axis.n == 1)
                        key = coord + "axis"; // support x1axis as xaxis
                    if (ranges[key]) {
                        from = ranges[key].from;
                        to = ranges[key].to;
                        break;
                    }
                }
            }

            // backwards-compat stuff - to be removed in future
            if (!ranges[key]) {
                axis = coord == "x" ? xaxes[0] : yaxes[0];
                from = ranges[coord + "1"];
                to = ranges[coord + "2"];
            }

            // auto-reverse as an added bonus
            if (from != null && to != null && from > to) {
                var tmp = from;
                from = to;
                to = tmp;
            }

            return { from: from, to: to, axis: axis };
        }

        function drawBackground() {
            ctx.save();
            ctx.translate(plotOffset.left, plotOffset.top);

            ctx.fillStyle = getColorOrGradient(options.grid.backgroundColor, plotHeight, 0, "rgba(255, 255, 255, 0)");
            ctx.fillRect(0, 0, plotWidth, plotHeight);
            ctx.restore();
        }

        function drawGrid() {
            var i, axes, bw, bc;

            ctx.save();
            ctx.translate(plotOffset.left, plotOffset.top);

            // draw markings
            var markings = options.grid.markings;
            if (markings) {
                if ($.isFunction(markings)) {
                    axes = plot.getAxes();
                    // xmin etc. is backwards compatibility, to be
                    // removed in the future
                    axes.xmin = axes.xaxis.min;
                    axes.xmax = axes.xaxis.max;
                    axes.ymin = axes.yaxis.min;
                    axes.ymax = axes.yaxis.max;

                    markings = markings(axes);
                }

                for (i = 0; i < markings.length; ++i) {
                    var m = markings[i],
                        xrange = extractRange(m, "x"),
                        yrange = extractRange(m, "y");

                    // fill in missing
                    if (xrange.from == null)
                        xrange.from = xrange.axis.min;
                    if (xrange.to == null)
                        xrange.to = xrange.axis.max;
                    if (yrange.from == null)
                        yrange.from = yrange.axis.min;
                    if (yrange.to == null)
                        yrange.to = yrange.axis.max;

                    // clip
                    if (xrange.to < xrange.axis.min || xrange.from > xrange.axis.max ||
                        yrange.to < yrange.axis.min || yrange.from > yrange.axis.max)
                        continue;

                    xrange.from = Math.max(xrange.from, xrange.axis.min);
                    xrange.to = Math.min(xrange.to, xrange.axis.max);
                    yrange.from = Math.max(yrange.from, yrange.axis.min);
                    yrange.to = Math.min(yrange.to, yrange.axis.max);

                    var xequal = xrange.from === xrange.to,
                        yequal = yrange.from === yrange.to;

                    if (xequal && yequal) {
                        continue;
                    }

                    // then draw
                    xrange.from = Math.floor(xrange.axis.p2c(xrange.from));
                    xrange.to = Math.floor(xrange.axis.p2c(xrange.to));
                    yrange.from = Math.floor(yrange.axis.p2c(yrange.from));
                    yrange.to = Math.floor(yrange.axis.p2c(yrange.to));

                    if (xequal || yequal) {
                        var lineWidth = m.lineWidth || options.grid.markingsLineWidth,
                            subPixel = lineWidth % 2 ? 0.5 : 0;
                        ctx.beginPath();
                        ctx.strokeStyle = m.color || options.grid.markingsColor;
                        ctx.lineWidth = lineWidth;
                        if (xequal) {
                            ctx.moveTo(xrange.to + subPixel, yrange.from);
                            ctx.lineTo(xrange.to + subPixel, yrange.to);
                        } else {
                            ctx.moveTo(xrange.from, yrange.to + subPixel);
                            ctx.lineTo(xrange.to, yrange.to + subPixel);                            
                        }
                        ctx.stroke();
                    } else {
                        ctx.fillStyle = m.color || options.grid.markingsColor;
                        ctx.fillRect(xrange.from, yrange.to,
                                     xrange.to - xrange.from,
                                     yrange.from - yrange.to);
                    }
                }
            }

            // draw the ticks
            axes = allAxes();
            bw = options.grid.borderWidth;

            for (var j = 0; j < axes.length; ++j) {
                var axis = axes[j], box = axis.box,
                    t = axis.tickLength, x, y, xoff, yoff;
                if (!axis.show || axis.ticks.length == 0)
                    continue;

                ctx.lineWidth = 1;

                // find the edges
                if (axis.direction == "x") {
                    x = 0;
                    if (t == "full")
                        y = (axis.position == "top" ? 0 : plotHeight);
                    else
                        y = box.top - plotOffset.top + (axis.position == "top" ? box.height : 0);
                }
                else {
                    y = 0;
                    if (t == "full")
                        x = (axis.position == "left" ? 0 : plotWidth);
                    else
                        x = box.left - plotOffset.left + (axis.position == "left" ? box.width : 0);
                }

                // draw tick bar
                if (!axis.innermost) {
                    ctx.strokeStyle = axis.options.color;
                    ctx.beginPath();
                    xoff = yoff = 0;
                    if (axis.direction == "x")
                        xoff = plotWidth + 1;
                    else
                        yoff = plotHeight + 1;

                    if (ctx.lineWidth == 1) {
                        if (axis.direction == "x") {
                            y = Math.floor(y) + 0.5;
                        } else {
                            x = Math.floor(x) + 0.5;
                        }
                    }

                    ctx.moveTo(x, y);
                    ctx.lineTo(x + xoff, y + yoff);
                    ctx.stroke();
                }

                // draw ticks

                ctx.strokeStyle = axis.options.tickColor;

                ctx.beginPath();
                for (i = 0; i < axis.ticks.length; ++i) {
                    var v = axis.ticks[i].v;

                    xoff = yoff = 0;

                    if (isNaN(v) || v < axis.min || v > axis.max
                        // skip those lying on the axes if we got a border
                        || (t == "full"
                            && ((typeof bw == "object" && bw[axis.position] > 0) || bw > 0)
                            && (v == axis.min || v == axis.max)))
                        continue;

                    if (axis.direction == "x") {
                        x = axis.p2c(v);
                        yoff = t == "full" ? -plotHeight : t;

                        if (axis.position == "top")
                            yoff = -yoff;
                    }
                    else {
                        y = axis.p2c(v);
                        xoff = t == "full" ? -plotWidth : t;

                        if (axis.position == "left")
                            xoff = -xoff;
                    }

                    if (ctx.lineWidth == 1) {
                        if (axis.direction == "x")
                            x = Math.floor(x) + 0.5;
                        else
                            y = Math.floor(y) + 0.5;
                    }

                    ctx.moveTo(x, y);
                    ctx.lineTo(x + xoff, y + yoff);
                }

                ctx.stroke();
            }


            // draw border
            if (bw) {
                // If either borderWidth or borderColor is an object, then draw the border
                // line by line instead of as one rectangle
                bc = options.grid.borderColor;
                if(typeof bw == "object" || typeof bc == "object") {
                    if (typeof bw !== "object") {
                        bw = {top: bw, right: bw, bottom: bw, left: bw};
                    }
                    if (typeof bc !== "object") {
                        bc = {top: bc, right: bc, bottom: bc, left: bc};
                    }

                    if (bw.top > 0) {
                        ctx.strokeStyle = bc.top;
                        ctx.lineWidth = bw.top;
                        ctx.beginPath();
                        ctx.moveTo(0 - bw.left, 0 - bw.top/2);
                        ctx.lineTo(plotWidth, 0 - bw.top/2);
                        ctx.stroke();
                    }

                    if (bw.right > 0) {
                        ctx.strokeStyle = bc.right;
                        ctx.lineWidth = bw.right;
                        ctx.beginPath();
                        ctx.moveTo(plotWidth + bw.right / 2, 0 - bw.top);
                        ctx.lineTo(plotWidth + bw.right / 2, plotHeight);
                        ctx.stroke();
                    }

                    if (bw.bottom > 0) {
                        ctx.strokeStyle = bc.bottom;
                        ctx.lineWidth = bw.bottom;
                        ctx.beginPath();
                        ctx.moveTo(plotWidth + bw.right, plotHeight + bw.bottom / 2);
                        ctx.lineTo(0, plotHeight + bw.bottom / 2);
                        ctx.stroke();
                    }

                    if (bw.left > 0) {
                        ctx.strokeStyle = bc.left;
                        ctx.lineWidth = bw.left;
                        ctx.beginPath();
                        ctx.moveTo(0 - bw.left/2, plotHeight + bw.bottom);
                        ctx.lineTo(0- bw.left/2, 0);
                        ctx.stroke();
                    }
                }
                else {
                    ctx.lineWidth = bw;
                    ctx.strokeStyle = options.grid.borderColor;
                    ctx.strokeRect(-bw/2, -bw/2, plotWidth + bw, plotHeight + bw);
                }
            }

            ctx.restore();
        }

        function drawAxisLabels() {

            $.each(allAxes(), function (_, axis) {
                var box = axis.box,
                    legacyStyles = axis.direction + "Axis " + axis.direction + axis.n + "Axis",
                    layer = "flot-" + axis.direction + "-axis flot-" + axis.direction + axis.n + "-axis " + legacyStyles,
                    font = axis.options.font || "flot-tick-label tickLabel",
                    tick, x, y, halign, valign;

                // Remove text before checking for axis.show and ticks.length;
                // otherwise plugins, like flot-tickrotor, that draw their own
                // tick labels will end up with both theirs and the defaults.

                surface.removeText(layer);

                if (!axis.show || axis.ticks.length == 0)
                    return;

                for (var i = 0; i < axis.ticks.length; ++i) {

                    tick = axis.ticks[i];
                    if (!tick.label || tick.v < axis.min || tick.v > axis.max)
                        continue;

                    if (axis.direction == "x") {
                        halign = "center";
                        x = plotOffset.left + axis.p2c(tick.v);
                        if (axis.position == "bottom") {
                            y = box.top + box.padding;
                        } else {
                            y = box.top + box.height - box.padding;
                            valign = "bottom";
                        }
                    } else {
                        valign = "middle";
                        y = plotOffset.top + axis.p2c(tick.v);
                        if (axis.position == "left") {
                            x = box.left + box.width - box.padding;
                            halign = "right";
                        } else {
                            x = box.left + box.padding;
                        }
                    }

                    surface.addText(layer, x, y, tick.label, font, null, null, halign, valign);
                }
            });
        }

        function drawSeries(series) {
            if (series.lines.show)
                drawSeriesLines(series);
            if (series.bars.show)
                drawSeriesBars(series);
            if (series.points.show)
                drawSeriesPoints(series);
        }

        function drawSeriesLines(series) {
            function plotLine(datapoints, xoffset, yoffset, axisx, axisy) {
                var points = datapoints.points,
                    ps = datapoints.pointsize,
                    prevx = null, prevy = null;

                ctx.beginPath();
                for (var i = ps; i < points.length; i += ps) {
                    var x1 = points[i - ps], y1 = points[i - ps + 1],
                        x2 = points[i], y2 = points[i + 1];

                    if (x1 == null || x2 == null)
                        continue;

                    // clip with ymin
                    if (y1 <= y2 && y1 < axisy.min) {
                        if (y2 < axisy.min)
                            continue;   // line segment is outside
                        // compute new intersection point
                        x1 = (axisy.min - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y1 = axisy.min;
                    }
                    else if (y2 <= y1 && y2 < axisy.min) {
                        if (y1 < axisy.min)
                            continue;
                        x2 = (axisy.min - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y2 = axisy.min;
                    }

                    // clip with ymax
                    if (y1 >= y2 && y1 > axisy.max) {
                        if (y2 > axisy.max)
                            continue;
                        x1 = (axisy.max - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y1 = axisy.max;
                    }
                    else if (y2 >= y1 && y2 > axisy.max) {
                        if (y1 > axisy.max)
                            continue;
                        x2 = (axisy.max - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y2 = axisy.max;
                    }

                    // clip with xmin
                    if (x1 <= x2 && x1 < axisx.min) {
                        if (x2 < axisx.min)
                            continue;
                        y1 = (axisx.min - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x1 = axisx.min;
                    }
                    else if (x2 <= x1 && x2 < axisx.min) {
                        if (x1 < axisx.min)
                            continue;
                        y2 = (axisx.min - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x2 = axisx.min;
                    }

                    // clip with xmax
                    if (x1 >= x2 && x1 > axisx.max) {
                        if (x2 > axisx.max)
                            continue;
                        y1 = (axisx.max - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x1 = axisx.max;
                    }
                    else if (x2 >= x1 && x2 > axisx.max) {
                        if (x1 > axisx.max)
                            continue;
                        y2 = (axisx.max - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x2 = axisx.max;
                    }

                    if (x1 != prevx || y1 != prevy)
                        ctx.moveTo(axisx.p2c(x1) + xoffset, axisy.p2c(y1) + yoffset);

                    prevx = x2;
                    prevy = y2;
                    ctx.lineTo(axisx.p2c(x2) + xoffset, axisy.p2c(y2) + yoffset);
                }
                ctx.stroke();
            }

            function plotLineArea(datapoints, axisx, axisy) {
                var points = datapoints.points,
                    ps = datapoints.pointsize,
                    bottom = Math.min(Math.max(0, axisy.min), axisy.max),
                    i = 0, top, areaOpen = false,
                    ypos = 1, segmentStart = 0, segmentEnd = 0;

                // we process each segment in two turns, first forward
                // direction to sketch out top, then once we hit the
                // end we go backwards to sketch the bottom
                while (true) {
                    if (ps > 0 && i > points.length + ps)
                        break;

                    i += ps; // ps is negative if going backwards

                    var x1 = points[i - ps],
                        y1 = points[i - ps + ypos],
                        x2 = points[i], y2 = points[i + ypos];

                    if (areaOpen) {
                        if (ps > 0 && x1 != null && x2 == null) {
                            // at turning point
                            segmentEnd = i;
                            ps = -ps;
                            ypos = 2;
                            continue;
                        }

                        if (ps < 0 && i == segmentStart + ps) {
                            // done with the reverse sweep
                            ctx.fill();
                            areaOpen = false;
                            ps = -ps;
                            ypos = 1;
                            i = segmentStart = segmentEnd + ps;
                            continue;
                        }
                    }

                    if (x1 == null || x2 == null)
                        continue;

                    // clip x values

                    // clip with xmin
                    if (x1 <= x2 && x1 < axisx.min) {
                        if (x2 < axisx.min)
                            continue;
                        y1 = (axisx.min - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x1 = axisx.min;
                    }
                    else if (x2 <= x1 && x2 < axisx.min) {
                        if (x1 < axisx.min)
                            continue;
                        y2 = (axisx.min - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x2 = axisx.min;
                    }

                    // clip with xmax
                    if (x1 >= x2 && x1 > axisx.max) {
                        if (x2 > axisx.max)
                            continue;
                        y1 = (axisx.max - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x1 = axisx.max;
                    }
                    else if (x2 >= x1 && x2 > axisx.max) {
                        if (x1 > axisx.max)
                            continue;
                        y2 = (axisx.max - x1) / (x2 - x1) * (y2 - y1) + y1;
                        x2 = axisx.max;
                    }

                    if (!areaOpen) {
                        // open area
                        ctx.beginPath();
                        ctx.moveTo(axisx.p2c(x1), axisy.p2c(bottom));
                        areaOpen = true;
                    }

                    // now first check the case where both is outside
                    if (y1 >= axisy.max && y2 >= axisy.max) {
                        ctx.lineTo(axisx.p2c(x1), axisy.p2c(axisy.max));
                        ctx.lineTo(axisx.p2c(x2), axisy.p2c(axisy.max));
                        continue;
                    }
                    else if (y1 <= axisy.min && y2 <= axisy.min) {
                        ctx.lineTo(axisx.p2c(x1), axisy.p2c(axisy.min));
                        ctx.lineTo(axisx.p2c(x2), axisy.p2c(axisy.min));
                        continue;
                    }

                    // else it's a bit more complicated, there might
                    // be a flat maxed out rectangle first, then a
                    // triangular cutout or reverse; to find these
                    // keep track of the current x values
                    var x1old = x1, x2old = x2;

                    // clip the y values, without shortcutting, we
                    // go through all cases in turn

                    // clip with ymin
                    if (y1 <= y2 && y1 < axisy.min && y2 >= axisy.min) {
                        x1 = (axisy.min - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y1 = axisy.min;
                    }
                    else if (y2 <= y1 && y2 < axisy.min && y1 >= axisy.min) {
                        x2 = (axisy.min - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y2 = axisy.min;
                    }

                    // clip with ymax
                    if (y1 >= y2 && y1 > axisy.max && y2 <= axisy.max) {
                        x1 = (axisy.max - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y1 = axisy.max;
                    }
                    else if (y2 >= y1 && y2 > axisy.max && y1 <= axisy.max) {
                        x2 = (axisy.max - y1) / (y2 - y1) * (x2 - x1) + x1;
                        y2 = axisy.max;
                    }

                    // if the x value was changed we got a rectangle
                    // to fill
                    if (x1 != x1old) {
                        ctx.lineTo(axisx.p2c(x1old), axisy.p2c(y1));
                        // it goes to (x1, y1), but we fill that below
                    }

                    // fill triangular section, this sometimes result
                    // in redundant points if (x1, y1) hasn't changed
                    // from previous line to, but we just ignore that
                    ctx.lineTo(axisx.p2c(x1), axisy.p2c(y1));
                    ctx.lineTo(axisx.p2c(x2), axisy.p2c(y2));

                    // fill the other rectangle if it's there
                    if (x2 != x2old) {
                        ctx.lineTo(axisx.p2c(x2), axisy.p2c(y2));
                        ctx.lineTo(axisx.p2c(x2old), axisy.p2c(y2));
                    }
                }
            }

            ctx.save();
            ctx.translate(plotOffset.left, plotOffset.top);
            ctx.lineJoin = "round";

            var lw = series.lines.lineWidth,
                sw = series.shadowSize;
            // FIXME: consider another form of shadow when filling is turned on
            if (lw > 0 && sw > 0) {
                // draw shadow as a thick and thin line with transparency
                ctx.lineWidth = sw;
                ctx.strokeStyle = "rgba(0,0,0,0.1)";
                // position shadow at angle from the mid of line
                var angle = Math.PI/18;
                plotLine(series.datapoints, Math.sin(angle) * (lw/2 + sw/2), Math.cos(angle) * (lw/2 + sw/2), series.xaxis, series.yaxis);
                ctx.lineWidth = sw/2;
                plotLine(series.datapoints, Math.sin(angle) * (lw/2 + sw/4), Math.cos(angle) * (lw/2 + sw/4), series.xaxis, series.yaxis);
            }

            ctx.lineWidth = lw;
            ctx.strokeStyle = series.color;
            var fillStyle = getFillStyle(series.lines, series.color, 0, plotHeight);
            if (fillStyle) {
                ctx.fillStyle = fillStyle;
                plotLineArea(series.datapoints, series.xaxis, series.yaxis);
            }

            if (lw > 0)
                plotLine(series.datapoints, 0, 0, series.xaxis, series.yaxis);
            ctx.restore();
        }

        function drawSeriesPoints(series) {
            function plotPoints(datapoints, radius, fillStyle, offset, shadow, axisx, axisy, symbol) {
                var points = datapoints.points, ps = datapoints.pointsize;

                for (var i = 0; i < points.length; i += ps) {
                    var x = points[i], y = points[i + 1];
                    if (x == null || x < axisx.min || x > axisx.max || y < axisy.min || y > axisy.max)
                        continue;

                    ctx.beginPath();
                    x = axisx.p2c(x);
                    y = axisy.p2c(y) + offset;
                    if (symbol == "circle")
                        ctx.arc(x, y, radius, 0, shadow ? Math.PI : Math.PI * 2, false);
                    else
                        symbol(ctx, x, y, radius, shadow);
                    ctx.closePath();

                    if (fillStyle) {
                        ctx.fillStyle = fillStyle;
                        ctx.fill();
                    }
                    ctx.stroke();
                }
            }

            ctx.save();
            ctx.translate(plotOffset.left, plotOffset.top);

            var lw = series.points.lineWidth,
                sw = series.shadowSize,
                radius = series.points.radius,
                symbol = series.points.symbol;

            // If the user sets the line width to 0, we change it to a very 
            // small value. A line width of 0 seems to force the default of 1.
            // Doing the conditional here allows the shadow setting to still be 
            // optional even with a lineWidth of 0.

            if( lw == 0 )
                lw = 0.0001;

            if (lw > 0 && sw > 0) {
                // draw shadow in two steps
                var w = sw / 2;
                ctx.lineWidth = w;
                ctx.strokeStyle = "rgba(0,0,0,0.1)";
                plotPoints(series.datapoints, radius, null, w + w/2, true,
                           series.xaxis, series.yaxis, symbol);

                ctx.strokeStyle = "rgba(0,0,0,0.2)";
                plotPoints(series.datapoints, radius, null, w/2, true,
                           series.xaxis, series.yaxis, symbol);
            }

            ctx.lineWidth = lw;
            ctx.strokeStyle = series.color;
            plotPoints(series.datapoints, radius,
                       getFillStyle(series.points, series.color), 0, false,
                       series.xaxis, series.yaxis, symbol);
            ctx.restore();
        }

        function drawBar(x, y, b, barLeft, barRight, fillStyleCallback, axisx, axisy, c, horizontal, lineWidth) {
            var left, right, bottom, top,
                drawLeft, drawRight, drawTop, drawBottom,
                tmp;

            // in horizontal mode, we start the bar from the left
            // instead of from the bottom so it appears to be
            // horizontal rather than vertical
            if (horizontal) {
                drawBottom = drawRight = drawTop = true;
                drawLeft = false;
                left = b;
                right = x;
                top = y + barLeft;
                bottom = y + barRight;

                // account for negative bars
                if (right < left) {
                    tmp = right;
                    right = left;
                    left = tmp;
                    drawLeft = true;
                    drawRight = false;
                }
            }
            else {
                drawLeft = drawRight = drawTop = true;
                drawBottom = false;
                left = x + barLeft;
                right = x + barRight;
                bottom = b;
                top = y;

                // account for negative bars
                if (top < bottom) {
                    tmp = top;
                    top = bottom;
                    bottom = tmp;
                    drawBottom = true;
                    drawTop = false;
                }
            }

            // clip
            if (right < axisx.min || left > axisx.max ||
                top < axisy.min || bottom > axisy.max)
                return;

            if (left < axisx.min) {
                left = axisx.min;
                drawLeft = false;
            }

            if (right > axisx.max) {
                right = axisx.max;
                drawRight = false;
            }

            if (bottom < axisy.min) {
                bottom = axisy.min;
                drawBottom = false;
            }

            if (top > axisy.max) {
                top = axisy.max;
                drawTop = false;
            }

            left = axisx.p2c(left);
            bottom = axisy.p2c(bottom);
            right = axisx.p2c(right);
            top = axisy.p2c(top);

            // fill the bar
            if (fillStyleCallback) {
                c.fillStyle = fillStyleCallback(bottom, top);
                c.fillRect(left, top, right - left, bottom - top)
            }

            // draw outline
            if (lineWidth > 0 && (drawLeft || drawRight || drawTop || drawBottom)) {
                c.beginPath();

                // FIXME: inline moveTo is buggy with excanvas
                c.moveTo(left, bottom);
                if (drawLeft)
                    c.lineTo(left, top);
                else
                    c.moveTo(left, top);
                if (drawTop)
                    c.lineTo(right, top);
                else
                    c.moveTo(right, top);
                if (drawRight)
                    c.lineTo(right, bottom);
                else
                    c.moveTo(right, bottom);
                if (drawBottom)
                    c.lineTo(left, bottom);
                else
                    c.moveTo(left, bottom);
                c.stroke();
            }
        }

        function drawSeriesBars(series) {
            function plotBars(datapoints, barLeft, barRight, fillStyleCallback, axisx, axisy) {
                var points = datapoints.points, ps = datapoints.pointsize;

                for (var i = 0; i < points.length; i += ps) {
                    if (points[i] == null)
                        continue;
                    drawBar(points[i], points[i + 1], points[i + 2], barLeft, barRight, fillStyleCallback, axisx, axisy, ctx, series.bars.horizontal, series.bars.lineWidth);
                }
            }

            ctx.save();
            ctx.translate(plotOffset.left, plotOffset.top);

            // FIXME: figure out a way to add shadows (for instance along the right edge)
            ctx.lineWidth = series.bars.lineWidth;
            ctx.strokeStyle = series.color;

            var barLeft;

            switch (series.bars.align) {
                case "left":
                    barLeft = 0;
                    break;
                case "right":
                    barLeft = -series.bars.barWidth;
                    break;
                default:
                    barLeft = -series.bars.barWidth / 2;
            }

            var fillStyleCallback = series.bars.fill ? function (bottom, top) { return getFillStyle(series.bars, series.color, bottom, top); } : null;
            plotBars(series.datapoints, barLeft, barLeft + series.bars.barWidth, fillStyleCallback, series.xaxis, series.yaxis);
            ctx.restore();
        }

        function getFillStyle(filloptions, seriesColor, bottom, top) {
            var fill = filloptions.fill;
            if (!fill)
                return null;

            if (filloptions.fillColor)
                return getColorOrGradient(filloptions.fillColor, bottom, top, seriesColor);

            var c = $.color.parse(seriesColor);
            c.a = typeof fill == "number" ? fill : 0.4;
            c.normalize();
            return c.toString();
        }

        function insertLegend() {

            if (options.legend.container != null) {
                $(options.legend.container).html("");
            } else {
                placeholder.find(".legend").remove();
            }

            if (!options.legend.show) {
                return;
            }

            var fragments = [], entries = [], rowStarted = false,
                lf = options.legend.labelFormatter, s, label;

            // Build a list of legend entries, with each having a label and a color

            for (var i = 0; i < series.length; ++i) {
                s = series[i];
                if (s.label) {
                    label = lf ? lf(s.label, s) : s.label;
                    if (label) {
                        entries.push({
                            label: label,
                            color: s.color
                        });
                    }
                }
            }

            // Sort the legend using either the default or a custom comparator

            if (options.legend.sorted) {
                if ($.isFunction(options.legend.sorted)) {
                    entries.sort(options.legend.sorted);
                } else if (options.legend.sorted == "reverse") {
                	entries.reverse();
                } else {
                    var ascending = options.legend.sorted != "descending";
                    entries.sort(function(a, b) {
                        return a.label == b.label ? 0 : (
                            (a.label < b.label) != ascending ? 1 : -1   // Logical XOR
                        );
                    });
                }
            }

            // Generate markup for the list of entries, in their final order

            for (var i = 0; i < entries.length; ++i) {

                var entry = entries[i];

                if (i % options.legend.noColumns == 0) {
                    if (rowStarted)
                        fragments.push('</tr>');
                    fragments.push('<tr>');
                    rowStarted = true;
                }

                fragments.push(
                    '<td class="legendColorBox"><div style="border:1px solid ' + options.legend.labelBoxBorderColor + ';padding:1px"><div style="width:4px;height:0;border:5px solid ' + entry.color + ';overflow:hidden"></div></div></td>' +
                    '<td class="legendLabel">' + entry.label + '</td>'
                );
            }

            if (rowStarted)
                fragments.push('</tr>');

            if (fragments.length == 0)
                return;

            var table = '<table style="font-size:smaller;color:' + options.grid.color + '">' + fragments.join("") + '</table>';
            if (options.legend.container != null)
                $(options.legend.container).html(table);
            else {
                var pos = "",
                    p = options.legend.position,
                    m = options.legend.margin;
                if (m[0] == null)
                    m = [m, m];
                if (p.charAt(0) == "n")
                    pos += 'top:' + (m[1] + plotOffset.top) + 'px;';
                else if (p.charAt(0) == "s")
                    pos += 'bottom:' + (m[1] + plotOffset.bottom) + 'px;';
                if (p.charAt(1) == "e")
                    pos += 'right:' + (m[0] + plotOffset.right) + 'px;';
                else if (p.charAt(1) == "w")
                    pos += 'left:' + (m[0] + plotOffset.left) + 'px;';
                var legend = $('<div class="legend">' + table.replace('style="', 'style="position:absolute;' + pos +';') + '</div>').appendTo(placeholder);
                if (options.legend.backgroundOpacity != 0.0) {
                    // put in the transparent background
                    // separately to avoid blended labels and
                    // label boxes
                    var c = options.legend.backgroundColor;
                    if (c == null) {
                        c = options.grid.backgroundColor;
                        if (c && typeof c == "string")
                            c = $.color.parse(c);
                        else
                            c = $.color.extract(legend, 'background-color');
                        c.a = 1;
                        c = c.toString();
                    }
                    var div = legend.children();
                    $('<div style="position:absolute;width:' + div.width() + 'px;height:' + div.height() + 'px;' + pos +'background-color:' + c + ';"> </div>').prependTo(legend).css('opacity', options.legend.backgroundOpacity);
                }
            }
        }


        // interactive features

        var highlights = [],
            redrawTimeout = null;

        // returns the data item the mouse is over, or null if none is found
        function findNearbyItem(mouseX, mouseY, seriesFilter) {
            var maxDistance = options.grid.mouseActiveRadius,
                smallestDistance = maxDistance * maxDistance + 1,
                item = null, foundPoint = false, i, j, ps;

            for (i = series.length - 1; i >= 0; --i) {
                if (!seriesFilter(series[i]))
                    continue;

                var s = series[i],
                    axisx = s.xaxis,
                    axisy = s.yaxis,
                    points = s.datapoints.points,
                    mx = axisx.c2p(mouseX), // precompute some stuff to make the loop faster
                    my = axisy.c2p(mouseY),
                    maxx = maxDistance / axisx.scale,
                    maxy = maxDistance / axisy.scale;

                ps = s.datapoints.pointsize;
                // with inverse transforms, we can't use the maxx/maxy
                // optimization, sadly
                if (axisx.options.inverseTransform)
                    maxx = Number.MAX_VALUE;
                if (axisy.options.inverseTransform)
                    maxy = Number.MAX_VALUE;

                if (s.lines.show || s.points.show) {
                    for (j = 0; j < points.length; j += ps) {
                        var x = points[j], y = points[j + 1];
                        if (x == null)
                            continue;

                        // For points and lines, the cursor must be within a
                        // certain distance to the data point
                        if (x - mx > maxx || x - mx < -maxx ||
                            y - my > maxy || y - my < -maxy)
                            continue;

                        // We have to calculate distances in pixels, not in
                        // data units, because the scales of the axes may be different
                        var dx = Math.abs(axisx.p2c(x) - mouseX),
                            dy = Math.abs(axisy.p2c(y) - mouseY),
                            dist = dx * dx + dy * dy; // we save the sqrt

                        // use <= to ensure last point takes precedence
                        // (last generally means on top of)
                        if (dist < smallestDistance) {
                            smallestDistance = dist;
                            item = [i, j / ps];
                        }
                    }
                }

                if (s.bars.show && !item) { // no other point can be nearby

                    var barLeft, barRight;

                    switch (s.bars.align) {
                        case "left":
                            barLeft = 0;
                            break;
                        case "right":
                            barLeft = -s.bars.barWidth;
                            break;
                        default:
                            barLeft = -s.bars.barWidth / 2;
                    }

                    barRight = barLeft + s.bars.barWidth;

                    for (j = 0; j < points.length; j += ps) {
                        var x = points[j], y = points[j + 1], b = points[j + 2];
                        if (x == null)
                            continue;

                        // for a bar graph, the cursor must be inside the bar
                        if (series[i].bars.horizontal ?
                            (mx <= Math.max(b, x) && mx >= Math.min(b, x) &&
                             my >= y + barLeft && my <= y + barRight) :
                            (mx >= x + barLeft && mx <= x + barRight &&
                             my >= Math.min(b, y) && my <= Math.max(b, y)))
                                item = [i, j / ps];
                    }
                }
            }

            if (item) {
                i = item[0];
                j = item[1];
                ps = series[i].datapoints.pointsize;

                return { datapoint: series[i].datapoints.points.slice(j * ps, (j + 1) * ps),
                         dataIndex: j,
                         series: series[i],
                         seriesIndex: i };
            }

            return null;
        }

        function onMouseMove(e) {
            if (options.grid.hoverable)
                triggerClickHoverEvent("plothover", e,
                                       function (s) { return s["hoverable"] != false; });
        }

        function onMouseLeave(e) {
            if (options.grid.hoverable)
                triggerClickHoverEvent("plothover", e,
                                       function (s) { return false; });
        }

        function onClick(e) {
            triggerClickHoverEvent("plotclick", e,
                                   function (s) { return s["clickable"] != false; });
        }

        // trigger click or hover event (they send the same parameters
        // so we share their code)
        function triggerClickHoverEvent(eventname, event, seriesFilter) {
            var offset = eventHolder.offset(),
                canvasX = event.pageX - offset.left - plotOffset.left,
                canvasY = event.pageY - offset.top - plotOffset.top,
            pos = canvasToAxisCoords({ left: canvasX, top: canvasY });

            pos.pageX = event.pageX;
            pos.pageY = event.pageY;

            var item = findNearbyItem(canvasX, canvasY, seriesFilter);

            if (item) {
                // fill in mouse pos for any listeners out there
                item.pageX = parseInt(item.series.xaxis.p2c(item.datapoint[0]) + offset.left + plotOffset.left, 10);
                item.pageY = parseInt(item.series.yaxis.p2c(item.datapoint[1]) + offset.top + plotOffset.top, 10);
            }

            if (options.grid.autoHighlight) {
                // clear auto-highlights
                for (var i = 0; i < highlights.length; ++i) {
                    var h = highlights[i];
                    if (h.auto == eventname &&
                        !(item && h.series == item.series &&
                          h.point[0] == item.datapoint[0] &&
                          h.point[1] == item.datapoint[1]))
                        unhighlight(h.series, h.point);
                }

                if (item)
                    highlight(item.series, item.datapoint, eventname);
            }

            placeholder.trigger(eventname, [ pos, item ]);
        }

        function triggerRedrawOverlay() {
            var t = options.interaction.redrawOverlayInterval;
            if (t == -1) {      // skip event queue
                drawOverlay();
                return;
            }

            if (!redrawTimeout)
                redrawTimeout = setTimeout(drawOverlay, t);
        }

        function drawOverlay() {
            redrawTimeout = null;

            // draw highlights
            octx.save();
            overlay.clear();
            octx.translate(plotOffset.left, plotOffset.top);

            var i, hi;
            for (i = 0; i < highlights.length; ++i) {
                hi = highlights[i];

                if (hi.series.bars.show)
                    drawBarHighlight(hi.series, hi.point);
                else
                    drawPointHighlight(hi.series, hi.point);
            }
            octx.restore();

            executeHooks(hooks.drawOverlay, [octx]);
        }

        function highlight(s, point, auto) {
            if (typeof s == "number")
                s = series[s];

            if (typeof point == "number") {
                var ps = s.datapoints.pointsize;
                point = s.datapoints.points.slice(ps * point, ps * (point + 1));
            }

            var i = indexOfHighlight(s, point);
            if (i == -1) {
                highlights.push({ series: s, point: point, auto: auto });

                triggerRedrawOverlay();
            }
            else if (!auto)
                highlights[i].auto = false;
        }

        function unhighlight(s, point) {
            if (s == null && point == null) {
                highlights = [];
                triggerRedrawOverlay();
                return;
            }

            if (typeof s == "number")
                s = series[s];

            if (typeof point == "number") {
                var ps = s.datapoints.pointsize;
                point = s.datapoints.points.slice(ps * point, ps * (point + 1));
            }

            var i = indexOfHighlight(s, point);
            if (i != -1) {
                highlights.splice(i, 1);

                triggerRedrawOverlay();
            }
        }

        function indexOfHighlight(s, p) {
            for (var i = 0; i < highlights.length; ++i) {
                var h = highlights[i];
                if (h.series == s && h.point[0] == p[0]
                    && h.point[1] == p[1])
                    return i;
            }
            return -1;
        }

        function drawPointHighlight(series, point) {
            var x = point[0], y = point[1],
                axisx = series.xaxis, axisy = series.yaxis,
                highlightColor = (typeof series.highlightColor === "string") ? series.highlightColor : $.color.parse(series.color).scale('a', 0.5).toString();

            if (x < axisx.min || x > axisx.max || y < axisy.min || y > axisy.max)
                return;

            var pointRadius = series.points.radius + series.points.lineWidth / 2;
            octx.lineWidth = pointRadius;
            octx.strokeStyle = highlightColor;
            var radius = 1.5 * pointRadius;
            x = axisx.p2c(x);
            y = axisy.p2c(y);

            octx.beginPath();
            if (series.points.symbol == "circle")
                octx.arc(x, y, radius, 0, 2 * Math.PI, false);
            else
                series.points.symbol(octx, x, y, radius, false);
            octx.closePath();
            octx.stroke();
        }

        function drawBarHighlight(series, point) {
            var highlightColor = (typeof series.highlightColor === "string") ? series.highlightColor : $.color.parse(series.color).scale('a', 0.5).toString(),
                fillStyle = highlightColor,
                barLeft;

            switch (series.bars.align) {
                case "left":
                    barLeft = 0;
                    break;
                case "right":
                    barLeft = -series.bars.barWidth;
                    break;
                default:
                    barLeft = -series.bars.barWidth / 2;
            }

            octx.lineWidth = series.bars.lineWidth;
            octx.strokeStyle = highlightColor;

            drawBar(point[0], point[1], point[2] || 0, barLeft, barLeft + series.bars.barWidth,
                    function () { return fillStyle; }, series.xaxis, series.yaxis, octx, series.bars.horizontal, series.bars.lineWidth);
        }

        function getColorOrGradient(spec, bottom, top, defaultColor) {
            if (typeof spec == "string")
                return spec;
            else {
                // assume this is a gradient spec; IE currently only
                // supports a simple vertical gradient properly, so that's
                // what we support too
                var gradient = ctx.createLinearGradient(0, top, 0, bottom);

                for (var i = 0, l = spec.colors.length; i < l; ++i) {
                    var c = spec.colors[i];
                    if (typeof c != "string") {
                        var co = $.color.parse(defaultColor);
                        if (c.brightness != null)
                            co = co.scale('rgb', c.brightness);
                        if (c.opacity != null)
                            co.a *= c.opacity;
                        c = co.toString();
                    }
                    gradient.addColorStop(i / (l - 1), c);
                }

                return gradient;
            }
        }
    }

    // Add the plot function to the top level of the jQuery object

    $.plot = function(placeholder, data, options) {
        //var t0 = new Date();
        var plot = new Plot($(placeholder), data, options, $.plot.plugins);
        //(window.console ? console.log : alert)("time used (msecs): " + ((new Date()).getTime() - t0.getTime()));
        return plot;
    };

    $.plot.version = "0.8.3";

    $.plot.plugins = [];

    // Also add the plot function as a chainable property

    $.fn.plot = function(data, options) {
        return this.each(function() {
            $.plot(this, data, options);
        });
    };

    // round to nearby lower multiple of base
    function floorInBase(n, base) {
        return base * Math.floor(n / base);
    }

})(jQuery);
/* Flot plugin for plotting error bars.

Copyright (c) 2007-2014 IOLA and Ole Laursen.
Licensed under the MIT license.

Error bars are used to show standard deviation and other statistical
properties in a plot.

* Created by Rui Pereira  -  rui (dot) pereira (at) gmail (dot) com

This plugin allows you to plot error-bars over points. Set "errorbars" inside
the points series to the axis name over which there will be error values in
your data array (*even* if you do not intend to plot them later, by setting
"show: null" on xerr/yerr).

The plugin supports these options:

	series: {
		points: {
			errorbars: "x" or "y" or "xy",
			xerr: {
				show: null/false or true,
				asymmetric: null/false or true,
				upperCap: null or "-" or function,
				lowerCap: null or "-" or function,
				color: null or color,
				radius: null or number
			},
			yerr: { same options as xerr }
		}
	}

Each data point array is expected to be of the type:

	"x"  [ x, y, xerr ]
	"y"  [ x, y, yerr ]
	"xy" [ x, y, xerr, yerr ]

Where xerr becomes xerr_lower,xerr_upper for the asymmetric error case, and
equivalently for yerr. Eg., a datapoint for the "xy" case with symmetric
error-bars on X and asymmetric on Y would be:

	[ x, y, xerr, yerr_lower, yerr_upper ]

By default no end caps are drawn. Setting upperCap and/or lowerCap to "-" will
draw a small cap perpendicular to the error bar. They can also be set to a
user-defined drawing function, with (ctx, x, y, radius) as parameters, as eg.

	function drawSemiCircle( ctx, x, y, radius ) {
		ctx.beginPath();
		ctx.arc( x, y, radius, 0, Math.PI, false );
		ctx.moveTo( x - radius, y );
		ctx.lineTo( x + radius, y );
		ctx.stroke();
	}

Color and radius both default to the same ones of the points series if not
set. The independent radius parameter on xerr/yerr is useful for the case when
we may want to add error-bars to a line, without showing the interconnecting
points (with radius: 0), and still showing end caps on the error-bars.
shadowSize and lineWidth are derived as well from the points series.

*/

(function ($) {
    var options = {
        series: {
            points: {
                errorbars: null, //should be 'x', 'y' or 'xy'
                xerr: { err: 'x', show: null, asymmetric: null, upperCap: null, lowerCap: null, color: null, radius: null},
                yerr: { err: 'y', show: null, asymmetric: null, upperCap: null, lowerCap: null, color: null, radius: null}
            }
        }
    };

    function processRawData(plot, series, data, datapoints){
        if (!series.points.errorbars)
            return;

        // x,y values
        var format = [
            { x: true, number: true, required: true },
            { y: true, number: true, required: true }
        ];

        var errors = series.points.errorbars;
        // error bars - first X then Y
        if (errors == 'x' || errors == 'xy') {
            // lower / upper error
            if (series.points.xerr.asymmetric) {
                format.push({ x: true, number: true, required: true });
                format.push({ x: true, number: true, required: true });
            } else
                format.push({ x: true, number: true, required: true });
        }
        if (errors == 'y' || errors == 'xy') {
            // lower / upper error
            if (series.points.yerr.asymmetric) {
                format.push({ y: true, number: true, required: true });
                format.push({ y: true, number: true, required: true });
            } else
                format.push({ y: true, number: true, required: true });
        }
        datapoints.format = format;
    }

    function parseErrors(series, i){

        var points = series.datapoints.points;

        // read errors from points array
        var exl = null,
                exu = null,
                eyl = null,
                eyu = null;
        var xerr = series.points.xerr,
                yerr = series.points.yerr;

        var eb = series.points.errorbars;
        // error bars - first X
        if (eb == 'x' || eb == 'xy') {
            if (xerr.asymmetric) {
                exl = points[i + 2];
                exu = points[i + 3];
                if (eb == 'xy')
                    if (yerr.asymmetric){
                        eyl = points[i + 4];
                        eyu = points[i + 5];
                    } else eyl = points[i + 4];
            } else {
                exl = points[i + 2];
                if (eb == 'xy')
                    if (yerr.asymmetric) {
                        eyl = points[i + 3];
                        eyu = points[i + 4];
                    } else eyl = points[i + 3];
            }
        // only Y
        } else if (eb == 'y')
            if (yerr.asymmetric) {
                eyl = points[i + 2];
                eyu = points[i + 3];
            } else eyl = points[i + 2];

        // symmetric errors?
        if (exu == null) exu = exl;
        if (eyu == null) eyu = eyl;

        var errRanges = [exl, exu, eyl, eyu];
        // nullify if not showing
        if (!xerr.show){
            errRanges[0] = null;
            errRanges[1] = null;
        }
        if (!yerr.show){
            errRanges[2] = null;
            errRanges[3] = null;
        }
        return errRanges;
    }

    function drawSeriesErrors(plot, ctx, s){

        var points = s.datapoints.points,
                ps = s.datapoints.pointsize,
                ax = [s.xaxis, s.yaxis],
                radius = s.points.radius,
                err = [s.points.xerr, s.points.yerr];

        //sanity check, in case some inverted axis hack is applied to flot
        var invertX = false;
        if (ax[0].p2c(ax[0].max) < ax[0].p2c(ax[0].min)) {
            invertX = true;
            var tmp = err[0].lowerCap;
            err[0].lowerCap = err[0].upperCap;
            err[0].upperCap = tmp;
        }

        var invertY = false;
        if (ax[1].p2c(ax[1].min) < ax[1].p2c(ax[1].max)) {
            invertY = true;
            var tmp = err[1].lowerCap;
            err[1].lowerCap = err[1].upperCap;
            err[1].upperCap = tmp;
        }

        for (var i = 0; i < s.datapoints.points.length; i += ps) {

            //parse
            var errRanges = parseErrors(s, i);

            //cycle xerr & yerr
            for (var e = 0; e < err.length; e++){

                var minmax = [ax[e].min, ax[e].max];

                //draw this error?
                if (errRanges[e * err.length]){

                    //data coordinates
                    var x = points[i],
                        y = points[i + 1];

                    //errorbar ranges
                    var upper = [x, y][e] + errRanges[e * err.length + 1],
                        lower = [x, y][e] - errRanges[e * err.length];

                    //points outside of the canvas
                    if (err[e].err == 'x')
                        if (y > ax[1].max || y < ax[1].min || upper < ax[0].min || lower > ax[0].max)
                            continue;
                    if (err[e].err == 'y')
                        if (x > ax[0].max || x < ax[0].min || upper < ax[1].min || lower > ax[1].max)
                            continue;

                    // prevent errorbars getting out of the canvas
                    var drawUpper = true,
                        drawLower = true;

                    if (upper > minmax[1]) {
                        drawUpper = false;
                        upper = minmax[1];
                    }
                    if (lower < minmax[0]) {
                        drawLower = false;
                        lower = minmax[0];
                    }

                    //sanity check, in case some inverted axis hack is applied to flot
                    if ((err[e].err == 'x' && invertX) || (err[e].err == 'y' && invertY)) {
                        //swap coordinates
                        var tmp = lower;
                        lower = upper;
                        upper = tmp;
                        tmp = drawLower;
                        drawLower = drawUpper;
                        drawUpper = tmp;
                        tmp = minmax[0];
                        minmax[0] = minmax[1];
                        minmax[1] = tmp;
                    }

                    // convert to pixels
                    x = ax[0].p2c(x),
                        y = ax[1].p2c(y),
                        upper = ax[e].p2c(upper);
                    lower = ax[e].p2c(lower);
                    minmax[0] = ax[e].p2c(minmax[0]);
                    minmax[1] = ax[e].p2c(minmax[1]);

                    //same style as points by default
                    var lw = err[e].lineWidth ? err[e].lineWidth : s.points.lineWidth,
                        sw = s.points.shadowSize != null ? s.points.shadowSize : s.shadowSize;

                    //shadow as for points
                    if (lw > 0 && sw > 0) {
                        var w = sw / 2;
                        ctx.lineWidth = w;
                        ctx.strokeStyle = "rgba(0,0,0,0.1)";
                        drawError(ctx, err[e], x, y, upper, lower, drawUpper, drawLower, radius, w + w/2, minmax);

                        ctx.strokeStyle = "rgba(0,0,0,0.2)";
                        drawError(ctx, err[e], x, y, upper, lower, drawUpper, drawLower, radius, w/2, minmax);
                    }

                    ctx.strokeStyle = err[e].color? err[e].color: s.color;
                    ctx.lineWidth = lw;
                    //draw it
                    drawError(ctx, err[e], x, y, upper, lower, drawUpper, drawLower, radius, 0, minmax);
                }
            }
        }
    }

    function drawError(ctx,err,x,y,upper,lower,drawUpper,drawLower,radius,offset,minmax){

        //shadow offset
        y += offset;
        upper += offset;
        lower += offset;

        // error bar - avoid plotting over circles
        if (err.err == 'x'){
            if (upper > x + radius) drawPath(ctx, [[upper,y],[Math.max(x + radius,minmax[0]),y]]);
            else drawUpper = false;
            if (lower < x - radius) drawPath(ctx, [[Math.min(x - radius,minmax[1]),y],[lower,y]] );
            else drawLower = false;
        }
        else {
            if (upper < y - radius) drawPath(ctx, [[x,upper],[x,Math.min(y - radius,minmax[0])]] );
            else drawUpper = false;
            if (lower > y + radius) drawPath(ctx, [[x,Math.max(y + radius,minmax[1])],[x,lower]] );
            else drawLower = false;
        }

        //internal radius value in errorbar, allows to plot radius 0 points and still keep proper sized caps
        //this is a way to get errorbars on lines without visible connecting dots
        radius = err.radius != null? err.radius: radius;

        // upper cap
        if (drawUpper) {
            if (err.upperCap == '-'){
                if (err.err=='x') drawPath(ctx, [[upper,y - radius],[upper,y + radius]] );
                else drawPath(ctx, [[x - radius,upper],[x + radius,upper]] );
            } else if ($.isFunction(err.upperCap)){
                if (err.err=='x') err.upperCap(ctx, upper, y, radius);
                else err.upperCap(ctx, x, upper, radius);
            }
        }
        // lower cap
        if (drawLower) {
            if (err.lowerCap == '-'){
                if (err.err=='x') drawPath(ctx, [[lower,y - radius],[lower,y + radius]] );
                else drawPath(ctx, [[x - radius,lower],[x + radius,lower]] );
            } else if ($.isFunction(err.lowerCap)){
                if (err.err=='x') err.lowerCap(ctx, lower, y, radius);
                else err.lowerCap(ctx, x, lower, radius);
            }
        }
    }

    function drawPath(ctx, pts){
        ctx.beginPath();
        ctx.moveTo(pts[0][0], pts[0][1]);
        for (var p=1; p < pts.length; p++)
            ctx.lineTo(pts[p][0], pts[p][1]);
        ctx.stroke();
    }

    function draw(plot, ctx){
        var plotOffset = plot.getPlotOffset();

        ctx.save();
        ctx.translate(plotOffset.left, plotOffset.top);
        $.each(plot.getData(), function (i, s) {
            if (s.points.errorbars && (s.points.xerr.show || s.points.yerr.show))
                drawSeriesErrors(plot, ctx, s);
        });
        ctx.restore();
    }

    function init(plot) {
        plot.hooks.processRawData.push(processRawData);
        plot.hooks.draw.push(draw);
    }

    $.plot.plugins.push({
                init: init,
                options: options,
                name: 'errorbars',
                version: '1.0'
            });
})(jQuery);
/* Flot plugin for adding the ability to pan and zoom the plot.

Copyright (c) 2007-2014 IOLA and Ole Laursen.
Licensed under the MIT license.

The default behaviour is double click and scrollwheel up/down to zoom in, drag
to pan. The plugin defines plot.zoom({ center }), plot.zoomOut() and
plot.pan( offset ) so you easily can add custom controls. It also fires
"plotpan" and "plotzoom" events, useful for synchronizing plots.

The plugin supports these options:

	zoom: {
		interactive: false
		trigger: "dblclick" // or "click" for single click
		amount: 1.5         // 2 = 200% (zoom in), 0.5 = 50% (zoom out)
	}

	pan: {
		interactive: false
		cursor: "move"      // CSS mouse cursor value used when dragging, e.g. "pointer"
		frameRate: 20
	}

	xaxis, yaxis, x2axis, y2axis: {
		zoomRange: null  // or [ number, number ] (min range, max range) or false
		panRange: null   // or [ number, number ] (min, max) or false
	}

"interactive" enables the built-in drag/click behaviour. If you enable
interactive for pan, then you'll have a basic plot that supports moving
around; the same for zoom.

"amount" specifies the default amount to zoom in (so 1.5 = 150%) relative to
the current viewport.

"cursor" is a standard CSS mouse cursor string used for visual feedback to the
user when dragging.

"frameRate" specifies the maximum number of times per second the plot will
update itself while the user is panning around on it (set to null to disable
intermediate pans, the plot will then not update until the mouse button is
released).

"zoomRange" is the interval in which zooming can happen, e.g. with zoomRange:
[1, 100] the zoom will never scale the axis so that the difference between min
and max is smaller than 1 or larger than 100. You can set either end to null
to ignore, e.g. [1, null]. If you set zoomRange to false, zooming on that axis
will be disabled.

"panRange" confines the panning to stay within a range, e.g. with panRange:
[-10, 20] panning stops at -10 in one end and at 20 in the other. Either can
be null, e.g. [-10, null]. If you set panRange to false, panning on that axis
will be disabled.

Example API usage:

	plot = $.plot(...);

	// zoom default amount in on the pixel ( 10, 20 )
	plot.zoom({ center: { left: 10, top: 20 } });

	// zoom out again
	plot.zoomOut({ center: { left: 10, top: 20 } });

	// zoom 200% in on the pixel (10, 20)
	plot.zoom({ amount: 2, center: { left: 10, top: 20 } });

	// pan 100 pixels to the left and 20 down
	plot.pan({ left: -100, top: 20 })

Here, "center" specifies where the center of the zooming should happen. Note
that this is defined in pixel space, not the space of the data points (you can
use the p2c helpers on the axes in Flot to help you convert between these).

"amount" is the amount to zoom the viewport relative to the current range, so
1 is 100% (i.e. no change), 1.5 is 150% (zoom in), 0.7 is 70% (zoom out). You
can set the default in the options.

*/

// First two dependencies, jquery.event.drag.js and
// jquery.mousewheel.js, we put them inline here to save people the
// effort of downloading them.

/*
jquery.event.drag.js ~ v1.5 ~ Copyright (c) 2008, Three Dub Media (http://threedubmedia.com)
Licensed under the MIT License ~ http://threedubmedia.googlecode.com/files/MIT-LICENSE.txt
*/
(function(a){function e(h){var k,j=this,l=h.data||{};if(l.elem)j=h.dragTarget=l.elem,h.dragProxy=d.proxy||j,h.cursorOffsetX=l.pageX-l.left,h.cursorOffsetY=l.pageY-l.top,h.offsetX=h.pageX-h.cursorOffsetX,h.offsetY=h.pageY-h.cursorOffsetY;else if(d.dragging||l.which>0&&h.which!=l.which||a(h.target).is(l.not))return;switch(h.type){case"mousedown":return a.extend(l,a(j).offset(),{elem:j,target:h.target,pageX:h.pageX,pageY:h.pageY}),b.add(document,"mousemove mouseup",e,l),i(j,!1),d.dragging=null,!1;case!d.dragging&&"mousemove":if(g(h.pageX-l.pageX)+g(h.pageY-l.pageY)<l.distance)break;h.target=l.target,k=f(h,"dragstart",j),k!==!1&&(d.dragging=j,d.proxy=h.dragProxy=a(k||j)[0]);case"mousemove":if(d.dragging){if(k=f(h,"drag",j),c.drop&&(c.drop.allowed=k!==!1,c.drop.handler(h)),k!==!1)break;h.type="mouseup"}case"mouseup":b.remove(document,"mousemove mouseup",e),d.dragging&&(c.drop&&c.drop.handler(h),f(h,"dragend",j)),i(j,!0),d.dragging=d.proxy=l.elem=!1}return!0}function f(b,c,d){b.type=c;var e=a.event.dispatch.call(d,b);return e===!1?!1:e||b.result}function g(a){return Math.pow(a,2)}function h(){return d.dragging===!1}function i(a,b){a&&(a.unselectable=b?"off":"on",a.onselectstart=function(){return b},a.style&&(a.style.MozUserSelect=b?"":"none"))}a.fn.drag=function(a,b,c){return b&&this.bind("dragstart",a),c&&this.bind("dragend",c),a?this.bind("drag",b?b:a):this.trigger("drag")};var b=a.event,c=b.special,d=c.drag={not:":input",distance:0,which:1,dragging:!1,setup:function(c){c=a.extend({distance:d.distance,which:d.which,not:d.not},c||{}),c.distance=g(c.distance),b.add(this,"mousedown",e,c),this.attachEvent&&this.attachEvent("ondragstart",h)},teardown:function(){b.remove(this,"mousedown",e),this===d.dragging&&(d.dragging=d.proxy=!1),i(this,!0),this.detachEvent&&this.detachEvent("ondragstart",h)}};c.dragstart=c.dragend={setup:function(){},teardown:function(){}}})(jQuery);

/* jquery.mousewheel.min.js
 * Copyright (c) 2011 Brandon Aaron (http://brandonaaron.net)
 * Licensed under the MIT License (LICENSE.txt).
 * Thanks to: http://adomas.org/javascript-mouse-wheel/ for some pointers.
 * Thanks to: Mathias Bank(http://www.mathias-bank.de) for a scope bug fix.
 * Thanks to: Seamus Leahy for adding deltaX and deltaY
 *
 * Version: 3.0.6
 *
 * Requires: 1.2.2+
 */
(function(d){function e(a){var b=a||window.event,c=[].slice.call(arguments,1),f=0,e=0,g=0,a=d.event.fix(b);a.type="mousewheel";b.wheelDelta&&(f=b.wheelDelta/120);b.detail&&(f=-b.detail/3);g=f;void 0!==b.axis&&b.axis===b.HORIZONTAL_AXIS&&(g=0,e=-1*f);void 0!==b.wheelDeltaY&&(g=b.wheelDeltaY/120);void 0!==b.wheelDeltaX&&(e=-1*b.wheelDeltaX/120);c.unshift(a,f,e,g);return(d.event.dispatch||d.event.handle).apply(this,c)}var c=["DOMMouseScroll","mousewheel"];if(d.event.fixHooks)for(var h=c.length;h;)d.event.fixHooks[c[--h]]=d.event.mouseHooks;d.event.special.mousewheel={setup:function(){if(this.addEventListener)for(var a=c.length;a;)this.addEventListener(c[--a],e,!1);else this.onmousewheel=e},teardown:function(){if(this.removeEventListener)for(var a=c.length;a;)this.removeEventListener(c[--a],e,!1);else this.onmousewheel=null}};d.fn.extend({mousewheel:function(a){return a?this.bind("mousewheel",a):this.trigger("mousewheel")},unmousewheel:function(a){return this.unbind("mousewheel",a)}})})(jQuery);




(function ($) {
    var options = {
        xaxis: {
            zoomRange: null, // or [number, number] (min range, max range)
            panRange: null // or [number, number] (min, max)
        },
        zoom: {
            interactive: false,
            trigger: "dblclick", // or "click" for single click
            amount: 1.5 // how much to zoom relative to current position, 2 = 200% (zoom in), 0.5 = 50% (zoom out)
        },
        pan: {
            interactive: false,
            cursor: "move",
            frameRate: 20
        }
    };

    function init(plot) {
        function onZoomClick(e, zoomOut) {
            var c = plot.offset();
            c.left = e.pageX - c.left;
            c.top = e.pageY - c.top;
            if (zoomOut)
                plot.zoomOut({ center: c });
            else
                plot.zoom({ center: c });
        }

        function onMouseWheel(e, delta) {
            e.preventDefault();
            onZoomClick(e, delta < 0);
            return false;
        }
        
        var prevCursor = 'default', prevPageX = 0, prevPageY = 0,
            panTimeout = null;

        function onDragStart(e) {
            if (e.which != 1)  // only accept left-click
                return false;
            var c = plot.getPlaceholder().css('cursor');
            if (c)
                prevCursor = c;
            plot.getPlaceholder().css('cursor', plot.getOptions().pan.cursor);
            prevPageX = e.pageX;
            prevPageY = e.pageY;
        }
        
        function onDrag(e) {
            var frameRate = plot.getOptions().pan.frameRate;
            if (panTimeout || !frameRate)
                return;

            panTimeout = setTimeout(function () {
                plot.pan({ left: prevPageX - e.pageX,
                           top: prevPageY - e.pageY });
                prevPageX = e.pageX;
                prevPageY = e.pageY;
                                                    
                panTimeout = null;
            }, 1 / frameRate * 1000);
        }

        function onDragEnd(e) {
            if (panTimeout) {
                clearTimeout(panTimeout);
                panTimeout = null;
            }
                    
            plot.getPlaceholder().css('cursor', prevCursor);
            plot.pan({ left: prevPageX - e.pageX,
                       top: prevPageY - e.pageY });
        }
        
        function bindEvents(plot, eventHolder) {
            var o = plot.getOptions();
            if (o.zoom.interactive) {
                eventHolder[o.zoom.trigger](onZoomClick);
                eventHolder.mousewheel(onMouseWheel);
            }

            if (o.pan.interactive) {
                eventHolder.bind("dragstart", { distance: 10 }, onDragStart);
                eventHolder.bind("drag", onDrag);
                eventHolder.bind("dragend", onDragEnd);
            }
        }

        plot.zoomOut = function (args) {
            if (!args)
                args = {};
            
            if (!args.amount)
                args.amount = plot.getOptions().zoom.amount;

            args.amount = 1 / args.amount;
            plot.zoom(args);
        };
        
        plot.zoom = function (args) {
            if (!args)
                args = {};
            
            var c = args.center,
                amount = args.amount || plot.getOptions().zoom.amount,
                w = plot.width(), h = plot.height();

            if (!c)
                c = { left: w / 2, top: h / 2 };
                
            var xf = c.left / w,
                yf = c.top / h,
                minmax = {
                    x: {
                        min: c.left - xf * w / amount,
                        max: c.left + (1 - xf) * w / amount
                    },
                    y: {
                        min: c.top - yf * h / amount,
                        max: c.top + (1 - yf) * h / amount
                    }
                };

            $.each(plot.getAxes(), function(_, axis) {
                var opts = axis.options,
                    min = minmax[axis.direction].min,
                    max = minmax[axis.direction].max,
                    zr = opts.zoomRange,
                    pr = opts.panRange;

                if (zr === false) // no zooming on this axis
                    return;
                    
                min = axis.c2p(min);
                max = axis.c2p(max);
                if (min > max) {
                    // make sure min < max
                    var tmp = min;
                    min = max;
                    max = tmp;
                }

                //Check that we are in panRange
                if (pr) {
                    if (pr[0] != null && min < pr[0]) {
                        min = pr[0];
                    }
                    if (pr[1] != null && max > pr[1]) {
                        max = pr[1];
                    }
                }

                var range = max - min;
                if (zr &&
                    ((zr[0] != null && range < zr[0] && amount >1) ||
                     (zr[1] != null && range > zr[1] && amount <1)))
                    return;
            
                opts.min = min;
                opts.max = max;
            });
            
            plot.setupGrid();
            plot.draw();
            
            if (!args.preventEvent)
                plot.getPlaceholder().trigger("plotzoom", [ plot, args ]);
        };

        plot.pan = function (args) {
            var delta = {
                x: +args.left,
                y: +args.top
            };

            if (isNaN(delta.x))
                delta.x = 0;
            if (isNaN(delta.y))
                delta.y = 0;

            $.each(plot.getAxes(), function (_, axis) {
                var opts = axis.options,
                    min, max, d = delta[axis.direction];

                min = axis.c2p(axis.p2c(axis.min) + d),
                max = axis.c2p(axis.p2c(axis.max) + d);

                var pr = opts.panRange;
                if (pr === false) // no panning on this axis
                    return;
                
                if (pr) {
                    // check whether we hit the wall
                    if (pr[0] != null && pr[0] > min) {
                        d = pr[0] - min;
                        min += d;
                        max += d;
                    }
                    
                    if (pr[1] != null && pr[1] < max) {
                        d = pr[1] - max;
                        min += d;
                        max += d;
                    }
                }
                
                opts.min = min;
                opts.max = max;
            });
            
            plot.setupGrid();
            plot.draw();
            
            if (!args.preventEvent)
                plot.getPlaceholder().trigger("plotpan", [ plot, args ]);
        };

        function shutdown(plot, eventHolder) {
            eventHolder.unbind(plot.getOptions().zoom.trigger, onZoomClick);
            eventHolder.unbind("mousewheel", onMouseWheel);
            eventHolder.unbind("dragstart", onDragStart);
            eventHolder.unbind("drag", onDrag);
            eventHolder.unbind("dragend", onDragEnd);
            if (panTimeout)
                clearTimeout(panTimeout);
        }
        
        plot.hooks.bindEvents.push(bindEvents);
        plot.hooks.shutdown.push(shutdown);
    }
    
    $.plot.plugins.push({
        init: init,
        options: options,
        name: 'navigate',
        version: '1.3'
    });
})(jQuery);
/* Flot plugin for selecting regions of a plot.

Copyright (c) 2007-2014 IOLA and Ole Laursen.
Licensed under the MIT license.

The plugin supports these options:

selection: {
	mode: null or "x" or "y" or "xy",
	color: color,
	shape: "round" or "miter" or "bevel",
	minSize: number of pixels
}

Selection support is enabled by setting the mode to one of "x", "y" or "xy".
In "x" mode, the user will only be able to specify the x range, similarly for
"y" mode. For "xy", the selection becomes a rectangle where both ranges can be
specified. "color" is color of the selection (if you need to change the color
later on, you can get to it with plot.getOptions().selection.color). "shape"
is the shape of the corners of the selection.

"minSize" is the minimum size a selection can be in pixels. This value can
be customized to determine the smallest size a selection can be and still
have the selection rectangle be displayed. When customizing this value, the
fact that it refers to pixels, not axis units must be taken into account.
Thus, for example, if there is a bar graph in time mode with BarWidth set to 1
minute, setting "minSize" to 1 will not make the minimum selection size 1
minute, but rather 1 pixel. Note also that setting "minSize" to 0 will prevent
"plotunselected" events from being fired when the user clicks the mouse without
dragging.

When selection support is enabled, a "plotselected" event will be emitted on
the DOM element you passed into the plot function. The event handler gets a
parameter with the ranges selected on the axes, like this:

	placeholder.bind( "plotselected", function( event, ranges ) {
		alert("You selected " + ranges.xaxis.from + " to " + ranges.xaxis.to)
		// similar for yaxis - with multiple axes, the extra ones are in
		// x2axis, x3axis, ...
	});

The "plotselected" event is only fired when the user has finished making the
selection. A "plotselecting" event is fired during the process with the same
parameters as the "plotselected" event, in case you want to know what's
happening while it's happening,

A "plotunselected" event with no arguments is emitted when the user clicks the
mouse to remove the selection. As stated above, setting "minSize" to 0 will
destroy this behavior.

The plugin allso adds the following methods to the plot object:

- setSelection( ranges, preventEvent )

  Set the selection rectangle. The passed in ranges is on the same form as
  returned in the "plotselected" event. If the selection mode is "x", you
  should put in either an xaxis range, if the mode is "y" you need to put in
  an yaxis range and both xaxis and yaxis if the selection mode is "xy", like
  this:

	setSelection({ xaxis: { from: 0, to: 10 }, yaxis: { from: 40, to: 60 } });

  setSelection will trigger the "plotselected" event when called. If you don't
  want that to happen, e.g. if you're inside a "plotselected" handler, pass
  true as the second parameter. If you are using multiple axes, you can
  specify the ranges on any of those, e.g. as x2axis/x3axis/... instead of
  xaxis, the plugin picks the first one it sees.

- clearSelection( preventEvent )

  Clear the selection rectangle. Pass in true to avoid getting a
  "plotunselected" event.

- getSelection()

  Returns the current selection in the same format as the "plotselected"
  event. If there's currently no selection, the function returns null.

*/

(function ($) {
    function init(plot) {
        var selection = {
                first: { x: -1, y: -1}, second: { x: -1, y: -1},
                show: false,
                active: false
            };

        // FIXME: The drag handling implemented here should be
        // abstracted out, there's some similar code from a library in
        // the navigation plugin, this should be massaged a bit to fit
        // the Flot cases here better and reused. Doing this would
        // make this plugin much slimmer.
        var savedhandlers = {};

        var mouseUpHandler = null;
        
        function onMouseMove(e) {
            if (selection.active) {
                updateSelection(e);
                
                plot.getPlaceholder().trigger("plotselecting", [ getSelection() ]);
            }
        }

        function onMouseDown(e) {
            if (e.which != 1)  // only accept left-click
                return;
            
            // cancel out any text selections
            document.body.focus();

            // prevent text selection and drag in old-school browsers
            if (document.onselectstart !== undefined && savedhandlers.onselectstart == null) {
                savedhandlers.onselectstart = document.onselectstart;
                document.onselectstart = function () { return false; };
            }
            if (document.ondrag !== undefined && savedhandlers.ondrag == null) {
                savedhandlers.ondrag = document.ondrag;
                document.ondrag = function () { return false; };
            }

            setSelectionPos(selection.first, e);

            selection.active = true;

            // this is a bit silly, but we have to use a closure to be
            // able to whack the same handler again
            mouseUpHandler = function (e) { onMouseUp(e); };
            
            $(document).one("mouseup", mouseUpHandler);
        }

        function onMouseUp(e) {
            mouseUpHandler = null;
            
            // revert drag stuff for old-school browsers
            if (document.onselectstart !== undefined)
                document.onselectstart = savedhandlers.onselectstart;
            if (document.ondrag !== undefined)
                document.ondrag = savedhandlers.ondrag;

            // no more dragging
            selection.active = false;
            updateSelection(e);

            if (selectionIsSane())
                triggerSelectedEvent();
            else {
                // this counts as a clear
                plot.getPlaceholder().trigger("plotunselected", [ ]);
                plot.getPlaceholder().trigger("plotselecting", [ null ]);
            }

            return false;
        }

        function getSelection() {
            if (!selectionIsSane())
                return null;
            
            if (!selection.show) return null;

            var r = {}, c1 = selection.first, c2 = selection.second;
            $.each(plot.getAxes(), function (name, axis) {
                if (axis.used) {
                    var p1 = axis.c2p(c1[axis.direction]), p2 = axis.c2p(c2[axis.direction]); 
                    r[name] = { from: Math.min(p1, p2), to: Math.max(p1, p2) };
                }
            });
            return r;
        }

        function triggerSelectedEvent() {
            var r = getSelection();

            plot.getPlaceholder().trigger("plotselected", [ r ]);

            // backwards-compat stuff, to be removed in future
            if (r.xaxis && r.yaxis)
                plot.getPlaceholder().trigger("selected", [ { x1: r.xaxis.from, y1: r.yaxis.from, x2: r.xaxis.to, y2: r.yaxis.to } ]);
        }

        function clamp(min, value, max) {
            return value < min ? min: (value > max ? max: value);
        }

        function setSelectionPos(pos, e) {
            var o = plot.getOptions();
            var offset = plot.getPlaceholder().offset();
            var plotOffset = plot.getPlotOffset();
            pos.x = clamp(0, e.pageX - offset.left - plotOffset.left, plot.width());
            pos.y = clamp(0, e.pageY - offset.top - plotOffset.top, plot.height());

            if (o.selection.mode == "y")
                pos.x = pos == selection.first ? 0 : plot.width();

            if (o.selection.mode == "x")
                pos.y = pos == selection.first ? 0 : plot.height();
        }

        function updateSelection(pos) {
            if (pos.pageX == null)
                return;

            setSelectionPos(selection.second, pos);
            if (selectionIsSane()) {
                selection.show = true;
                plot.triggerRedrawOverlay();
            }
            else
                clearSelection(true);
        }

        function clearSelection(preventEvent) {
            if (selection.show) {
                selection.show = false;
                plot.triggerRedrawOverlay();
                if (!preventEvent)
                    plot.getPlaceholder().trigger("plotunselected", [ ]);
            }
        }

        // function taken from markings support in Flot
        function extractRange(ranges, coord) {
            var axis, from, to, key, axes = plot.getAxes();

            for (var k in axes) {
                axis = axes[k];
                if (axis.direction == coord) {
                    key = coord + axis.n + "axis";
                    if (!ranges[key] && axis.n == 1)
                        key = coord + "axis"; // support x1axis as xaxis
                    if (ranges[key]) {
                        from = ranges[key].from;
                        to = ranges[key].to;
                        break;
                    }
                }
            }

            // backwards-compat stuff - to be removed in future
            if (!ranges[key]) {
                axis = coord == "x" ? plot.getXAxes()[0] : plot.getYAxes()[0];
                from = ranges[coord + "1"];
                to = ranges[coord + "2"];
            }

            // auto-reverse as an added bonus
            if (from != null && to != null && from > to) {
                var tmp = from;
                from = to;
                to = tmp;
            }
            
            return { from: from, to: to, axis: axis };
        }
        
        function setSelection(ranges, preventEvent) {
            var axis, range, o = plot.getOptions();

            if (o.selection.mode == "y") {
                selection.first.x = 0;
                selection.second.x = plot.width();
            }
            else {
                range = extractRange(ranges, "x");

                selection.first.x = range.axis.p2c(range.from);
                selection.second.x = range.axis.p2c(range.to);
            }

            if (o.selection.mode == "x") {
                selection.first.y = 0;
                selection.second.y = plot.height();
            }
            else {
                range = extractRange(ranges, "y");

                selection.first.y = range.axis.p2c(range.from);
                selection.second.y = range.axis.p2c(range.to);
            }

            selection.show = true;
            plot.triggerRedrawOverlay();
            if (!preventEvent && selectionIsSane())
                triggerSelectedEvent();
        }

        function selectionIsSane() {
            var minSize = plot.getOptions().selection.minSize;
            return Math.abs(selection.second.x - selection.first.x) >= minSize &&
                Math.abs(selection.second.y - selection.first.y) >= minSize;
        }

        plot.clearSelection = clearSelection;
        plot.setSelection = setSelection;
        plot.getSelection = getSelection;

        plot.hooks.bindEvents.push(function(plot, eventHolder) {
            var o = plot.getOptions();
            if (o.selection.mode != null) {
                eventHolder.mousemove(onMouseMove);
                eventHolder.mousedown(onMouseDown);
            }
        });


        plot.hooks.drawOverlay.push(function (plot, ctx) {
            // draw selection
            if (selection.show && selectionIsSane()) {
                var plotOffset = plot.getPlotOffset();
                var o = plot.getOptions();

                ctx.save();
                ctx.translate(plotOffset.left, plotOffset.top);

                var c = $.color.parse(o.selection.color);

                ctx.strokeStyle = c.scale('a', 0.8).toString();
                ctx.lineWidth = 1;
                ctx.lineJoin = o.selection.shape;
                ctx.fillStyle = c.scale('a', 0.4).toString();

                var x = Math.min(selection.first.x, selection.second.x) + 0.5,
                    y = Math.min(selection.first.y, selection.second.y) + 0.5,
                    w = Math.abs(selection.second.x - selection.first.x) - 1,
                    h = Math.abs(selection.second.y - selection.first.y) - 1;

                ctx.fillRect(x, y, w, h);
                ctx.strokeRect(x, y, w, h);

                ctx.restore();
            }
        });
        
        plot.hooks.shutdown.push(function (plot, eventHolder) {
            eventHolder.unbind("mousemove", onMouseMove);
            eventHolder.unbind("mousedown", onMouseDown);
            
            if (mouseUpHandler)
                $(document).unbind("mouseup", mouseUpHandler);
        });

    }

    $.plot.plugins.push({
        init: init,
        options: {
            selection: {
                mode: null, // one of null, "x", "y" or "xy"
                color: "#e8cfac",
                shape: "round", // one of "round", "miter", or "bevel"
                minSize: 5 // minimum number of pixels
            }
        },
        name: 'selection',
        version: '1.1'
    });
})(jQuery);
/*jslint white: true, vars: true, plusplus: true, nomen: true, unparam: true */
/*globals $, JS9 */ 

"use strict";


(function($) {
    function zoomStackIn(plot, event, ranges, func) {
	var axes = plot.getAxes();
	var div  = plot.getPlaceholder();

	var r = {};
	r.xmin = axes.xaxis.min;
	r.xmax = axes.xaxis.max;
	r.ymin = axes.yaxis.min;
	r.ymax = axes.yaxis.max;

	plot.stack.push(r);

	axes.xaxis.options.min = ranges.xaxis.from;
	axes.xaxis.options.max = ranges.xaxis.to;
	axes.yaxis.options.min = ranges.yaxis.from;
	axes.yaxis.options.max = ranges.yaxis.to;

	plot.clearSelection(true);

	plot.setupGrid();
	plot.draw();

	$(div).find(".zoomout").css("visibility", "visible");

	if ( func !== undefined ) { func(plot, r); }
    }

    function zoomStackOut(plot, func) {
	var r    = plot.stack.pop();
	var div  = plot.getPlaceholder();

	if (  plot.stack.length === 0 ) {
	    $(div).find(".zoomout").css("visibility", "hidden");
	}

	plot.getAxes().xaxis.options.min = r.xmin;
	plot.getAxes().xaxis.options.max = r.xmax;
	plot.getAxes().yaxis.options.min = r.ymin;
	plot.getAxes().yaxis.options.max = r.ymax;

	plot.clearSelection(true);

	plot.setupGrid();
	plot.draw();

	if ( func !== undefined ) { func(plot, r); }
    }

    var enabled = 0;

    var options = {
	  zoomStack: 0
	, zoomFunc: undefined
    };

    function zoomStack(plot, ctx) {
	if ( enabled ) {
	    plot.stack = [];
	    var div = plot.getPlaceholder();
	    var apath = JS9.InstallDir("images/4arrow.png");
	    $(div).append("<div style='position:relative'><div style='z-index:100;position:absolute;right:12;top:12;z-index:2'>		\
		    <image class='zoomout'  src='"+apath+"'></div></div>");

	    $(div).bind("plotselected", function (event, ranges) { zoomStackIn (plot, event, ranges, options.zoomFunc); });
	    $(div).find(".zoomout").click(function ()            { zoomStackOut(plot, options.zoomFunc); });

	    $(div).find(".zoomout").css("visibility", "hidden");

	    enabled = 0;
	}
    }

    function zoomOptions(plot, opts) {
        // zoomStack can be true or it can be an object containing options
        if ( opts && opts.zoomStack ) {
            if( typeof opts.zoomStack === "object" ){
                options.zoomStack = opts.zoomStack.enabled;
                options.zoomFunc = opts.zoomStack.func;
            } else {
                options.zoomStack = opts.zoomStack;
            }
        }
        if( options.zoomStack ){
            enabled = 1;
	    plot.hooks.drawOverlay.push(zoomStack);
        }
    }

    function init(plot) {
	plot.hooks.processOptions.push(zoomOptions);
    }

    $.plot.plugins.push({ init: init, options: options, name: "zoomStack", version: 0.1 });
}($));

function sprintf () {
    // http://kevin.vanzonneveld.net
    // +   original by: Ash Searle (http://hexmen.com/blog/)
    // + namespaced by: Michael White (http://getsprink.com)
    // +    tweaked by: Jack
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // +      input by: Paulo Freitas
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // +      input by: Brett Zamir (http://brett-zamir.me)
    // +   improved by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
    // *     example 1: sprintf("%01.2f", 123.1);
    // *     returns 1: 123.10
    // *     example 2: sprintf("[%10s]", 'monkey');
    // *     returns 2: '[    monkey]'
    // *     example 3: sprintf("[%'#10s]", 'monkey');
    // *     returns 3: '[####monkey]'
    var regex = /%%|%(\d+\$)?([-+\'#0 ]*)(\*\d+\$|\*|\d+)?(\.(\*\d+\$|\*|\d+))?([scboxXuidfegEG])/g;
    var a = arguments,
        i = 0,
        format = a[i++];

    // pad()
    var pad = function (str, len, chr, leftJustify) {
        if (!chr) {
            chr = ' ';
        }
        var padding = (str.length >= len) ? '' : Array(1 + len - str.length >>> 0).join(chr);
        return leftJustify ? str + padding : padding + str;
    };

    // justify()
    var justify = function (value, prefix, leftJustify, minWidth, zeroPad, customPadChar) {
        var diff = minWidth - value.length;
        if (diff > 0) {
            if (leftJustify || !zeroPad) {
                value = pad(value, minWidth, customPadChar, leftJustify);
            } else {
                value = value.slice(0, prefix.length) + pad('', diff, '0', true) + value.slice(prefix.length);
            }
        }
        return value;
    };

    // formatBaseX()
    var formatBaseX = function (value, base, prefix, leftJustify, minWidth, precision, zeroPad) {
        // Note: casts negative numbers to positive ones
        var number = value >>> 0;
        prefix = prefix && number && {
            '2': '0b',
            '8': '0',
            '16': '0x'
        }[base] || '';
        value = prefix + pad(number.toString(base), precision || 0, '0', false);
        return justify(value, prefix, leftJustify, minWidth, zeroPad);
    };

    // formatString()
    var formatString = function (value, leftJustify, minWidth, precision, zeroPad, customPadChar) {
        if (precision != null) {
            value = value.slice(0, precision);
        }
        return justify(value, '', leftJustify, minWidth, zeroPad, customPadChar);
    };

    // doFormat()
    var doFormat = function (substring, valueIndex, flags, minWidth, _, precision, type) {
        var number;
        var prefix;
        var method;
        var textTransform;
        var value;

        if (substring == '%%') {
            return '%';
        }

        // parse flags
        var leftJustify = false,
            positivePrefix = '',
            zeroPad = false,
            prefixBaseX = false,
            customPadChar = ' ';
        var flagsl = flags.length;
        for (var j = 0; flags && j < flagsl; j++) {
            switch (flags.charAt(j)) {
            case ' ':
                positivePrefix = ' ';
                break;
            case '+':
                positivePrefix = '+';
                break;
            case '-':
                leftJustify = true;
                break;
            case "'":
                customPadChar = flags.charAt(j + 1);
                break;
            case '0':
                zeroPad = true;
                break;
            case '#':
                prefixBaseX = true;
                break;
            }
        }

        // parameters may be null, undefined, empty-string or real valued
        // we want to ignore null, undefined and empty-string values
        if (!minWidth) {
            minWidth = 0;
        } else if (minWidth == '*') {
            minWidth = +a[i++];
        } else if (minWidth.charAt(0) == '*') {
            minWidth = +a[minWidth.slice(1, -1)];
        } else {
            minWidth = +minWidth;
        }

        // Note: undocumented perl feature:
        if (minWidth < 0) {
            minWidth = -minWidth;
            leftJustify = true;
        }

        if (!isFinite(minWidth)) {
            throw new Error('sprintf: (minimum-)width must be finite');
        }

        if (!precision) {
            precision = 'fFeE'.indexOf(type) > -1 ? 6 : (type == 'd') ? 0 : undefined;
        } else if (precision == '*') {
            precision = +a[i++];
        } else if (precision.charAt(0) == '*') {
            precision = +a[precision.slice(1, -1)];
        } else {
            precision = +precision;
        }

        // grab value using valueIndex if required?
        value = valueIndex ? a[valueIndex.slice(0, -1)] : a[i++];

        switch (type) {
        case 's':
            return formatString(String(value), leftJustify, minWidth, precision, zeroPad, customPadChar);
        case 'c':
            return formatString(String.fromCharCode(+value), leftJustify, minWidth, precision, zeroPad);
        case 'b':
            return formatBaseX(value, 2, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'o':
            return formatBaseX(value, 8, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'x':
            return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'X':
            return formatBaseX(value, 16, prefixBaseX, leftJustify, minWidth, precision, zeroPad).toUpperCase();
        case 'u':
            return formatBaseX(value, 10, prefixBaseX, leftJustify, minWidth, precision, zeroPad);
        case 'i':
        case 'd':
            number = (+value) | 0;
            prefix = number < 0 ? '-' : positivePrefix;
            value = prefix + pad(String(Math.abs(number)), precision, '0', false);
            return justify(value, prefix, leftJustify, minWidth, zeroPad);
        case 'e':
        case 'E':
        case 'f':
        case 'F':
        case 'g':
        case 'G':
            number = +value;
            prefix = number < 0 ? '-' : positivePrefix;
            method = ['toExponential', 'toFixed', 'toPrecision']['efg'.indexOf(type.toLowerCase())];
            textTransform = ['toString', 'toUpperCase']['eEfFgG'.indexOf(type) % 2];
            value = prefix + Math.abs(number)[method](precision);
            return justify(value, prefix, leftJustify, minWidth, zeroPad)[textTransform]();
        default:
            return substring;
        }
    };

    return format.replace(regex, doFormat);
}
// -------------------------------------------------------------------
// DHTML Window Widget- By Dynamic Drive, available at: http://www.dynamicdrive.com
// v1.0: Script created Feb 15th, 07'
// v1.01: Feb 21th, 07' (see changelog.txt)
// v1.02: March 26th, 07' (see changelog.txt)
// v1.03: May 5th, 07' (see changelog.txt)
// v1.1:  Oct 29th, 07' (see changelog.txt)
// -------------------------------------------------------------------

var dhtmlwindow={
imagefiles:['windowfiles/min.gif', 'windowfiles/close.gif', 'windowfiles/restore.gif', 'windowfiles/resize.gif'], //Path to 4 images used by script, in that order
ajaxbustcache: true, //Bust caching when fetching a file via Ajax?
ajaxloadinghtml: '<b>Loading Page. Please wait...</b>', //HTML to show while window fetches Ajax Content?

minimizeorder: 0,
zIndexvalue:100,
tobjects: [], //object to contain references to dhtml window divs, for cleanup purposes
lastactivet: {}, //reference to last active DHTML window

init:function(t){
	var domwindow=document.createElement("div") //create dhtml window div
	domwindow.id=t
	domwindow.className="dhtmlwindow"
	var domwindowdata=''
	domwindowdata='<div class="drag-handle">'
	domwindowdata+='DHTML Window <div class="drag-controls"><img src="'+this.imagefiles[0]+'" title="Minimize" /><img src="'+this.imagefiles[1]+'" title="Close" /></div>'
	domwindowdata+='</div>'
	domwindowdata+='<div class="drag-contentarea"></div>'
	domwindowdata+='<div class="drag-statusarea"><div class="drag-resizearea" style="background: transparent url('+this.imagefiles[3]+') top right no-repeat;">&nbsp;</div></div>'
	domwindowdata+='</div>'
	domwindow.innerHTML=domwindowdata
	document.getElementById("dhtmlwindowholder").appendChild(domwindow)
	//this.zIndexvalue=(this.zIndexvalue)? this.zIndexvalue+1 : 100 //z-index value for DHTML window: starts at 0, increments whenever a window has focus
	var t=document.getElementById(t)
	var divs=t.getElementsByTagName("div")
	for (var i=0; i<divs.length; i++){ //go through divs inside dhtml window and extract all those with class="drag-" prefix
		if (/drag-/.test(divs[i].className))
			t[divs[i].className.replace(/drag-/, "")]=divs[i] //take out the "drag-" prefix for shorter access by name
	}
	//t.style.zIndex=this.zIndexvalue //set z-index of this dhtml window
	t.handle._parent=t //store back reference to dhtml window
	t.resizearea._parent=t //same
	t.controls._parent=t //same
	t.onclose=function(){return true} //custom event handler "onclose"
	t.onmousedown=function(){dhtmlwindow.setfocus(this)} //Increase z-index of window when focus is on it
	t.ontouchstart=function(){dhtmlwindow.setfocus(this)} //Increase z-index of window when focus is on it
	t.handle.onmousedown=dhtmlwindow.setupdrag //set up drag behavior when mouse down on handle div
	t.handle.ontouchstart=dhtmlwindow.setupdrag //set up drag behavior when mouse down on handle div
	t.resizearea.onmousedown=dhtmlwindow.setupdrag //set up drag behavior when mouse down on resize div
	t.resizearea.ontouchstart=dhtmlwindow.setupdrag //set up drag behavior when mouse down on resize div
	t.controls.onclick=dhtmlwindow.enablecontrols
	t.show=function(){dhtmlwindow.show(this)} //public function for showing dhtml window
	t.hide=function(){dhtmlwindow.hide(this)} //public function for hiding dhtml window
	t.close=function(){dhtmlwindow.close(this)} //public function for closing dhtml window (also empties DHTML window content)
	t.setSize=function(w, h){dhtmlwindow.setSize(this, w, h)} //public function for setting window dimensions
	t.moveTo=function(x, y){dhtmlwindow.moveTo(this, x, y)} //public function for moving dhtml window (relative to viewpoint)
	t.isResize=function(bol){dhtmlwindow.isResize(this, bol)} //public function for specifying if window is resizable
	t.isScrolling=function(bol){dhtmlwindow.isScrolling(this, bol)} //public function for specifying if window content contains scrollbars
	t.load=function(contenttype, contentsource, title){dhtmlwindow.load(this, contenttype, contentsource, title)} //public function for loading content into window
	this.tobjects[this.tobjects.length]=t
	return t //return reference to dhtml window div
},

open:function(t, contenttype, contentsource, title, attr, recalonload){
	var d=dhtmlwindow //reference dhtml window object
	function getValue(Name){
		var config=new RegExp(Name+"=([^,]+)", "i") //get name/value config pair (ie: width=400px,)
		return (config.test(attr))? parseInt(RegExp.$1) : 0 //return value portion (int), or 0 (false) if none found
	}
	if (document.getElementById(t)==null) //if window doesn't exist yet, create it
		t=this.init(t) //return reference to dhtml window div
	else
		t=document.getElementById(t)
	this.setfocus(t)
	t.setSize(getValue(("width")), (getValue("height"))) //Set dimensions of window
	var xpos=getValue("center")? "middle" : getValue("left") //Get x coord of window
	var ypos=getValue("center")? "middle" : getValue("top") //Get y coord of window
	//t.moveTo(xpos, ypos) //Position window
	if (typeof recalonload!="undefined" && recalonload=="recal" && this.scroll_top==0){ //reposition window when page fully loads with updated window viewpoints?
		if (window.attachEvent && !window.opera) //In IE, add another 400 milisecs on page load (viewpoint properties may return 0 b4 then)
			this.addEvent(window, function(){setTimeout(function(){t.moveTo(xpos, ypos)}, 400)}, "load")
		else
			this.addEvent(window, function(){t.moveTo(xpos, ypos)}, "load")
	}
	t.isResize(getValue("resize")) //Set whether window is resizable
	t.isScrolling(getValue("scrolling")) //Set whether window should contain scrollbars
	t.style.visibility="visible"
	t.style.display="block"
	t.contentarea.style.display="block"
	t.moveTo(xpos, ypos) //Position window
	t.load(contenttype, contentsource, title)
	if (t.state=="minimized" && t.controls.firstChild.title=="Restore"){ //If window exists and is currently minimized?
		t.controls.firstChild.setAttribute("src", dhtmlwindow.imagefiles[0]) //Change "restore" icon within window interface to "minimize" icon
		t.controls.firstChild.setAttribute("title", "Minimize")
		t.state="fullview" //indicate the state of the window as being "fullview"
	}
	return t
},

setSize:function(t, w, h){ //set window size (min is 150px wide by 100px tall)
	t.style.width=Math.max(parseInt(w), 150)+"px"
	t.contentarea.style.height=Math.max(parseInt(h), 20)+"px"
},

moveTo:function(t, x, y){ //move window. Position includes current viewpoint of document
	this.getviewpoint() //Get current viewpoint numbers
	t.style.left=(x=="middle")? this.scroll_left+(this.docwidth-t.offsetWidth)/2+"px" : this.scroll_left+parseInt(x)+"px"
	t.style.top=(y=="middle")? this.scroll_top+(this.docheight-t.offsetHeight)/2+"px" : this.scroll_top+parseInt(y)+"px"
},

isResize:function(t, bol){ //show or hide resize inteface (part of the status bar)
	t.statusarea.style.display=(bol)? "block" : "none"
	t.resizeBool=(bol)? 1 : 0
},

isScrolling:function(t, bol){ //set whether loaded content contains scrollbars
	t.contentarea.style.overflow=(bol)? "auto" : "hidden"
},

load:function(t, contenttype, contentsource, title){ //loads content into window plus set its title (3 content types: "inline", "iframe", or "ajax")
	if (t.isClosed){
		alert("DHTML Window has been closed, so no window to load contents into. Open/Create the window again.")
		return
	}
	var contenttype=contenttype.toLowerCase() //convert string to lower case
	if (typeof title!="undefined")
		t.handle.firstChild.nodeValue=title
	if (contenttype=="inline")
		t.contentarea.innerHTML=contentsource
	else if (contenttype=="div"){
		var inlinedivref=document.getElementById(contentsource)
		t.contentarea.innerHTML=(inlinedivref.defaultHTML || inlinedivref.innerHTML) //Populate window with contents of inline div on page
		if (!inlinedivref.defaultHTML)
			inlinedivref.defaultHTML=inlinedivref.innerHTML //save HTML within inline DIV
		inlinedivref.innerHTML="" //then, remove HTML within inline DIV (to prevent duplicate IDs, NAME attributes etc in contents of DHTML window
		inlinedivref.style.display="none" //hide that div
	}
	else if (contenttype=="iframe"){
		t.contentarea.style.overflow="hidden" //disable window scrollbars, as iframe already contains scrollbars
		if (!t.contentarea.firstChild || t.contentarea.firstChild.tagName!="IFRAME") //If iframe tag doesn't exist already, create it first
			t.contentarea.innerHTML='<iframe src="" style="margin:0; padding:0; width:100%; height: 100%" name="_iframe-'+t.id+'"></iframe>'
		window.frames["_iframe-"+t.id].location.replace(contentsource) //set location of iframe window to specified URL
		}
	else if (contenttype=="ajax"){
		this.ajax_connect(contentsource, t) //populate window with external contents fetched via Ajax
	}
	t.contentarea.datatype=contenttype //store contenttype of current window for future reference
},

setupdrag:function(e){
	var d=dhtmlwindow //reference dhtml window object
	var t=this._parent //reference dhtml window div
	d.etarget=this //remember div mouse is currently held down on ("handle" or "resize" div)
	var e=window.event || e
	// d.initmousex=e.clientX //store x position of mouse onmousedown
	// d.initmousey=e.clientY
        if( e.touches && e.touches.length ){
 	    d.initmousex = e.touches[0].clientX;
	    d.initmousey = e.touches[0].clientY;
        } else {
	    // mouse events
	    d.initmousex = e.clientX;
	    d.initmousey = e.clientY;
        }
	d.initx=parseInt(t.offsetLeft) //store offset x of window div onmousedown
	d.inity=parseInt(t.offsetTop)
	d.width=parseInt(t.offsetWidth) //store width of window div
	d.contentheight=parseInt(t.contentarea.offsetHeight) //store height of window div's content div
	if (t.contentarea.datatype=="iframe"){ //if content of this window div is "iframe"
		t.style.backgroundColor="#F8F8F8" //colorize and hide content div (while window is being dragged)
		t.contentarea.style.visibility="hidden"
	}
	document.onmousemove=d.getdistance //get distance travelled by mouse as it moves
	document.ontouchmove=d.getdistance //get distance travelled by mouse as it moves
	document.onmouseup=function(){
		if (t.contentarea.datatype=="iframe"){ //restore color and visibility of content div onmouseup
			t.contentarea.style.backgroundColor="white"
			t.contentarea.style.visibility="visible"
		}
		d.stop()
	}
	document.ontouchend=function(){
		if (t.contentarea.datatype=="iframe"){ //restore color and visibility of content div onmouseup
			t.contentarea.style.backgroundColor="white"
			t.contentarea.style.visibility="visible"
		}
		d.stop()
	}
	return false
},

getdistance:function(e){
	var d=dhtmlwindow
	var etarget=d.etarget
	var e=window.event || e
	// d.distancex=e.clientX-d.initmousex //horizontal distance travelled relative to starting point
	// d.distancey=e.clientY-d.initmousey
        if( e.touches && e.touches.length ){
	    d.distancex=e.touches[0].clientX-d.initmousex //horizontal distance travelled relative to starting point
	    d.distancey=e.touches[0].clientY-d.initmousey
        } else {
	    // mouse events
	    d.distancex=e.clientX-d.initmousex //horizontal distance travelled relative to starting point
	    d.distancey=e.clientY-d.initmousey
        }
	if (etarget.className=="drag-handle") //if target element is "handle" div
		d.move(etarget._parent, e)
	else if (etarget.className=="drag-resizearea") //if target element is "resize" div
		d.resize(etarget._parent, e)
	return false //cancel default dragging behavior
},

getviewpoint:function(){ //get window viewpoint numbers
	var ie=document.all && !window.opera
	var domclientWidth=document.documentElement && parseInt(document.documentElement.clientWidth) || 100000 //Preliminary doc width in non IE browsers
	this.standardbody=(document.compatMode=="CSS1Compat")? document.documentElement : document.body //create reference to common "body" across doctypes
	this.scroll_top=(ie)? this.standardbody.scrollTop : window.pageYOffset
	this.scroll_left=(ie)? this.standardbody.scrollLeft : window.pageXOffset
	this.docwidth=(ie)? this.standardbody.clientWidth : (/Safari/i.test(navigator.userAgent))? window.innerWidth : Math.min(domclientWidth, window.innerWidth-16)
	this.docheight=(ie)? this.standardbody.clientHeight: window.innerHeight
},

rememberattrs:function(t){ //remember certain attributes of the window when it's minimized or closed, such as dimensions, position on page
	this.getviewpoint() //Get current window viewpoint numbers
	t.lastx=parseInt((t.style.left || t.offsetLeft))-dhtmlwindow.scroll_left //store last known x coord of window just before minimizing
	t.lasty=parseInt((t.style.top || t.offsetTop))-dhtmlwindow.scroll_top
	t.lastwidth=parseInt(t.style.width) //store last known width of window just before minimizing/ closing
},

move:function(t, e){
	t.style.left=Math.max(0,dhtmlwindow.distancex+dhtmlwindow.initx)+"px"
	t.style.top=Math.max(0,dhtmlwindow.distancey+dhtmlwindow.inity)+"px"
},

resize:function(t, e){
	t.style.width=Math.max(dhtmlwindow.width+dhtmlwindow.distancex, 150)+"px"
	t.contentarea.style.height=Math.max(dhtmlwindow.contentheight+dhtmlwindow.distancey, 20)+"px"
},

enablecontrols:function(e){
	var d=dhtmlwindow
	var sourceobj=window.event? window.event.srcElement : e.target //Get element within "handle" div mouse is currently on (the controls)
	if (/Minimize/i.test(sourceobj.getAttribute("title"))) //if this is the "minimize" control
		d.minimize(sourceobj, this._parent)
	else if (/Restore/i.test(sourceobj.getAttribute("title"))) //if this is the "restore" control
		d.restore(sourceobj, this._parent)
	else if (/Close/i.test(sourceobj.getAttribute("title"))) //if this is the "close" control
		d.close(this._parent)
	return false
},

minimize:function(button, t){
	dhtmlwindow.rememberattrs(t)
	button.setAttribute("src", dhtmlwindow.imagefiles[2])
	button.setAttribute("title", "Restore")
	t.state="minimized" //indicate the state of the window as being "minimized"
	t.contentarea.style.display="none"
	t.statusarea.style.display="none"
	if (typeof t.minimizeorder=="undefined"){ //stack order of minmized window on screen relative to any other minimized windows
		dhtmlwindow.minimizeorder++ //increment order
		t.minimizeorder=dhtmlwindow.minimizeorder
	}
	t.style.left="10px" //left coord of minmized window
	t.style.width="200px"
	var windowspacing=t.minimizeorder*10 //spacing (gap) between each minmized window(s)
	t.style.top=dhtmlwindow.scroll_top+dhtmlwindow.docheight-(t.handle.offsetHeight*t.minimizeorder)-windowspacing+"px"
},

restore:function(button, t){
	dhtmlwindow.getviewpoint()
	button.setAttribute("src", dhtmlwindow.imagefiles[0])
	button.setAttribute("title", "Minimize")
	t.state="fullview" //indicate the state of the window as being "fullview"
	t.style.display="block"
	t.contentarea.style.display="block"
	if (t.resizeBool) //if this window is resizable, enable the resize icon
		t.statusarea.style.display="block"
	t.style.left=parseInt(t.lastx)+dhtmlwindow.scroll_left+"px" //position window to last known x coord just before minimizing
	t.style.top=parseInt(t.lasty)+dhtmlwindow.scroll_top+"px"
	t.style.width=parseInt(t.lastwidth)+"px"
},


close:function(t){
	try{
		var closewinbol=t.onclose()
	}
	catch(err){ //In non IE browsers, all errors are caught, so just run the below
		var closewinbol=true
 }
	finally{ //In IE, not all errors are caught, so check if variable isn't defined in IE in those cases
		if (typeof closewinbol=="undefined"){
			alert("An error has occured somwhere inside your \"onclose\" event handler")
			var closewinbol=true
		}
	}
	if (closewinbol){ //if custom event handler function returns true
		if (t.state!="minimized") //if this window isn't currently minimized
			dhtmlwindow.rememberattrs(t) //remember window's dimensions/position on the page before closing
		if (window.frames["_iframe-"+t.id]) //if this is an IFRAME DHTML window
			window.frames["_iframe-"+t.id].location.replace("about:blank")
		else
			t.contentarea.innerHTML=""
		t.style.display="none"
		t.isClosed=true //tell script this window is closed (for detection in t.show())
	}
	return closewinbol
},


setopacity:function(targetobject, value){ //Sets the opacity of targetobject based on the passed in value setting (0 to 1 and in between)
	if (!targetobject)
		return
	if (targetobject.filters && targetobject.filters[0]){ //IE syntax
		if (typeof targetobject.filters[0].opacity=="number") //IE6
			targetobject.filters[0].opacity=value*100
		else //IE 5.5
			targetobject.style.filter="alpha(opacity="+value*100+")"
		}
	else if (typeof targetobject.style.MozOpacity!="undefined") //Old Mozilla syntax
		targetobject.style.MozOpacity=value
	else if (typeof targetobject.style.opacity!="undefined") //Standard opacity syntax
		targetobject.style.opacity=value
},

setfocus:function(t){ //Sets focus to the currently active window
	this.zIndexvalue++
	t.style.zIndex=this.zIndexvalue
	t.isClosed=false //tell script this window isn't closed (for detection in t.show())
	this.setopacity(this.lastactivet.handle, 0.5) //unfocus last active window
	this.setopacity(t.handle, 1) //focus currently active window
	this.lastactivet=t //remember last active window
},


show:function(t){
	if (t.isClosed){
		alert("DHTML Window has been closed, so nothing to show. Open/Create the window again.")
		return
	}
	if (t.lastx) //If there exists previously stored information such as last x position on window attributes (meaning it's been minimized or closed)
		dhtmlwindow.restore(t.controls.firstChild, t) //restore the window using that info
	else
		t.style.display="block"
	this.setfocus(t)
	t.state="fullview" //indicate the state of the window as being "fullview"
},

hide:function(t){
	t.style.display="none"
},

ajax_connect:function(url, t){
	var page_request = false
	var bustcacheparameter=""
	if (window.XMLHttpRequest) // if Mozilla, IE7, Safari etc
		page_request = new XMLHttpRequest()
	else if (window.ActiveXObject){ // if IE6 or below
		try {
		page_request = new ActiveXObject("Msxml2.XMLHTTP")
		} 
		catch (e){
			try{
			page_request = new ActiveXObject("Microsoft.XMLHTTP")
			}
			catch (e){}
		}
	}
	else
		return false
	t.contentarea.innerHTML=this.ajaxloadinghtml
	page_request.onreadystatechange=function(){dhtmlwindow.ajax_loadpage(page_request, t)}
	if (this.ajaxbustcache) //if bust caching of external page
		bustcacheparameter=(url.indexOf("?")!=-1)? "&"+new Date().getTime() : "?"+new Date().getTime()
	page_request.open('GET', url+bustcacheparameter, true)
	page_request.send(null)
},

ajax_loadpage:function(page_request, t){
	if (page_request.readyState == 4 && (page_request.status==200 || window.location.href.indexOf("http")==-1)){
	t.contentarea.innerHTML=page_request.responseText
	}
},


stop:function(){
	dhtmlwindow.etarget=null //clean up
	dhtmlwindow.distancex=0 //clean up
	dhtmlwindow.distancey=0 //clean up
	document.onmousemove=null
	document.ontouchmove=null
	document.onmouseup=null
	document.ontouchend=null
},

addEvent:function(target, functionref, tasktype){ //assign a function to execute to an event handler (ie: onunload)
	var tasktype=(window.addEventListener)? tasktype : "on"+tasktype
	if (target.addEventListener)
		target.addEventListener(tasktype, functionref, false)
	else if (target.attachEvent)
		target.attachEvent(tasktype, functionref)
},

cleanup:function(){
	for (var i=0; i<dhtmlwindow.tobjects.length; i++){
		dhtmlwindow.tobjects[i].handle._parent=dhtmlwindow.tobjects[i].resizearea._parent=dhtmlwindow.tobjects[i].controls._parent=null
	}
	window.onload=null
}

} //End dhtmlwindow object

// document.write('<div id="dhtmlwindowholder"><span style="display:none">.</span></div>') //container that holds all dhtml window divs on page
window.onunload=dhtmlwindow.cleanup
  /***********************************************
  * DHTML Window Widget- © Dynamic Drive (www.dynamicdrive.com)
  * This notice must stay intact for legal use.
  * Visit http://www.dynamicdrive.com/ for full source code
  ***********************************************/
/* build: `node build.js modules=ALL exclude=json,gestures minifier=uglifyjs` */
/*! Fabric.js Copyright 2008-2015, Printio (Juriy Zaytsev, Maxim Chernyak) */

var fabric = fabric || { version: "1.6.3" };
if (typeof exports !== 'undefined') {
  exports.fabric = fabric;
}

if (typeof document !== 'undefined' && typeof window !== 'undefined') {
  fabric.document = document;
  fabric.window = window;
  // ensure globality even if entire library were function wrapped (as in Meteor.js packaging system)
  window.fabric = fabric;
}
else {
  // assume we're running under node.js when document/window are not present
  fabric.document = require("jsdom")
    .jsdom("<!DOCTYPE html><html><head></head><body></body></html>");

  if (fabric.document.createWindow) {
    fabric.window = fabric.document.createWindow();
  } else {
    fabric.window = fabric.document.parentWindow;
  }
}

/**
 * True when in environment that supports touch events
 * @type boolean
 */
fabric.isTouchSupported = "ontouchstart" in fabric.document.documentElement;

/**
 * True when in environment that's probably Node.js
 * @type boolean
 */
fabric.isLikelyNode = typeof Buffer !== 'undefined' &&
                      typeof window === 'undefined';

/* _FROM_SVG_START_ */
/**
 * Attributes parsed from all SVG elements
 * @type array
 */
fabric.SHARED_ATTRIBUTES = [
  "display",
  "transform",
  "fill", "fill-opacity", "fill-rule",
  "opacity",
  "stroke", "stroke-dasharray", "stroke-linecap",
  "stroke-linejoin", "stroke-miterlimit",
  "stroke-opacity", "stroke-width",
  "id"
];
/* _FROM_SVG_END_ */

/**
 * Pixel per Inch as a default value set to 96. Can be changed for more realistic conversion.
 */
fabric.DPI = 96;
fabric.reNum = '(?:[-+]?(?:\\d+|\\d*\\.\\d+)(?:e[-+]?\\d+)?)';
fabric.fontPaths = { };

/**
 * Cache Object for widths of chars in text rendering.
 */
fabric.charWidthsCache = { };

/**
 * Device Pixel Ratio
 * @see https://developer.apple.com/library/safari/documentation/AudioVideo/Conceptual/HTML-canvas-guide/SettingUptheCanvas/SettingUptheCanvas.html
 */
fabric.devicePixelRatio = fabric.window.devicePixelRatio ||
                          fabric.window.webkitDevicePixelRatio ||
                          fabric.window.mozDevicePixelRatio ||
                          1;


(function() {

  /**
   * @private
   * @param {String} eventName
   * @param {Function} handler
   */
  function _removeEventListener(eventName, handler) {
    if (!this.__eventListeners[eventName]) {
      return;
    }
    var eventListener = this.__eventListeners[eventName];
    if (handler) {
      eventListener[eventListener.indexOf(handler)] = false;
    }
    else {
      fabric.util.array.fill(eventListener, false);
    }
  }

  /**
   * Observes specified event
   * @deprecated `observe` deprecated since 0.8.34 (use `on` instead)
   * @memberOf fabric.Observable
   * @alias on
   * @param {String|Object} eventName Event name (eg. 'after:render') or object with key/value pairs (eg. {'after:render': handler, 'selection:cleared': handler})
   * @param {Function} handler Function that receives a notification when an event of the specified type occurs
   * @return {Self} thisArg
   * @chainable
   */
  function observe(eventName, handler) {
    if (!this.__eventListeners) {
      this.__eventListeners = { };
    }
    // one object with key/value pairs was passed
    if (arguments.length === 1) {
      for (var prop in eventName) {
        this.on(prop, eventName[prop]);
      }
    }
    else {
      if (!this.__eventListeners[eventName]) {
        this.__eventListeners[eventName] = [ ];
      }
      this.__eventListeners[eventName].push(handler);
    }
    return this;
  }

  /**
   * Stops event observing for a particular event handler. Calling this method
   * without arguments removes all handlers for all events
   * @deprecated `stopObserving` deprecated since 0.8.34 (use `off` instead)
   * @memberOf fabric.Observable
   * @alias off
   * @param {String|Object} eventName Event name (eg. 'after:render') or object with key/value pairs (eg. {'after:render': handler, 'selection:cleared': handler})
   * @param {Function} handler Function to be deleted from EventListeners
   * @return {Self} thisArg
   * @chainable
   */
  function stopObserving(eventName, handler) {
    if (!this.__eventListeners) {
      return;
    }

    // remove all key/value pairs (event name -> event handler)
    if (arguments.length === 0) {
      for (eventName in this.__eventListeners) {
        _removeEventListener.call(this, eventName);
      }
    }
    // one object with key/value pairs was passed
    else if (arguments.length === 1 && typeof arguments[0] === 'object') {
      for (var prop in eventName) {
        _removeEventListener.call(this, prop, eventName[prop]);
      }
    }
    else {
      _removeEventListener.call(this, eventName, handler);
    }
    return this;
  }

  /**
   * Fires event with an optional options object
   * @deprecated `fire` deprecated since 1.0.7 (use `trigger` instead)
   * @memberOf fabric.Observable
   * @alias trigger
   * @param {String} eventName Event name to fire
   * @param {Object} [options] Options object
   * @return {Self} thisArg
   * @chainable
   */
  function fire(eventName, options) {
    if (!this.__eventListeners) {
      return;
    }

    var listenersForEvent = this.__eventListeners[eventName];
    if (!listenersForEvent) {
      return;
    }

    for (var i = 0, len = listenersForEvent.length; i < len; i++) {
      listenersForEvent[i] && listenersForEvent[i].call(this, options || { });
    }
    this.__eventListeners[eventName] = listenersForEvent.filter(function(value) {
      return value !== false;
    });
    return this;
  }

  /**
   * @namespace fabric.Observable
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-2#events}
   * @see {@link http://fabricjs.com/events|Events demo}
   */
  fabric.Observable = {
    observe: observe,
    stopObserving: stopObserving,
    fire: fire,

    on: observe,
    off: stopObserving,
    trigger: fire
  };
})();


/**
 * @namespace fabric.Collection
 */
fabric.Collection = {

  /**
   * Adds objects to collection, then renders canvas (if `renderOnAddRemove` is not `false`)
   * Objects should be instances of (or inherit from) fabric.Object
   * @param {...fabric.Object} object Zero or more fabric instances
   * @return {Self} thisArg
   */
  add: function () {
    this._objects.push.apply(this._objects, arguments);
    for (var i = 0, length = arguments.length; i < length; i++) {
      this._onObjectAdded(arguments[i]);
    }
    this.renderOnAddRemove && this.renderAll();
    return this;
  },

  /**
   * Inserts an object into collection at specified index, then renders canvas (if `renderOnAddRemove` is not `false`)
   * An object should be an instance of (or inherit from) fabric.Object
   * @param {Object} object Object to insert
   * @param {Number} index Index to insert object at
   * @param {Boolean} nonSplicing When `true`, no splicing (shifting) of objects occurs
   * @return {Self} thisArg
   * @chainable
   */
  insertAt: function (object, index, nonSplicing) {
    var objects = this.getObjects();
    if (nonSplicing) {
      objects[index] = object;
    }
    else {
      objects.splice(index, 0, object);
    }
    this._onObjectAdded(object);
    this.renderOnAddRemove && this.renderAll();
    return this;
  },

  /**
   * Removes objects from a collection, then renders canvas (if `renderOnAddRemove` is not `false`)
   * @param {...fabric.Object} object Zero or more fabric instances
   * @return {Self} thisArg
   * @chainable
   */
  remove: function() {
    var objects = this.getObjects(),
        index;

    for (var i = 0, length = arguments.length; i < length; i++) {
      index = objects.indexOf(arguments[i]);

      // only call onObjectRemoved if an object was actually removed
      if (index !== -1) {
        objects.splice(index, 1);
        this._onObjectRemoved(arguments[i]);
      }
    }

    this.renderOnAddRemove && this.renderAll();
    return this;
  },

  /**
   * Executes given function for each object in this group
   * @param {Function} callback
   *                   Callback invoked with current object as first argument,
   *                   index - as second and an array of all objects - as third.
   *                   Iteration happens in reverse order (for performance reasons).
   *                   Callback is invoked in a context of Global Object (e.g. `window`)
   *                   when no `context` argument is given
   *
   * @param {Object} context Context (aka thisObject)
   * @return {Self} thisArg
   */
  forEachObject: function(callback, context) {
    var objects = this.getObjects(),
        i = objects.length;
    while (i--) {
      callback.call(context, objects[i], i, objects);
    }
    return this;
  },

  /**
   * Returns an array of children objects of this instance
   * Type parameter introduced in 1.3.10
   * @param {String} [type] When specified, only objects of this type are returned
   * @return {Array}
   */
  getObjects: function(type) {
    if (typeof type === 'undefined') {
      return this._objects;
    }
    return this._objects.filter(function(o) {
      return o.type === type;
    });
  },

  /**
   * Returns object at specified index
   * @param {Number} index
   * @return {Self} thisArg
   */
  item: function (index) {
    return this.getObjects()[index];
  },

  /**
   * Returns true if collection contains no objects
   * @return {Boolean} true if collection is empty
   */
  isEmpty: function () {
    return this.getObjects().length === 0;
  },

  /**
   * Returns a size of a collection (i.e: length of an array containing its objects)
   * @return {Number} Collection size
   */
  size: function() {
    return this.getObjects().length;
  },

  /**
   * Returns true if collection contains an object
   * @param {Object} object Object to check against
   * @return {Boolean} `true` if collection contains an object
   */
  contains: function(object) {
    return this.getObjects().indexOf(object) > -1;
  },

  /**
   * Returns number representation of a collection complexity
   * @return {Number} complexity
   */
  complexity: function () {
    return this.getObjects().reduce(function (memo, current) {
      memo += current.complexity ? current.complexity() : 0;
      return memo;
    }, 0);
  }
};


(function(global) {

  var sqrt = Math.sqrt,
      atan2 = Math.atan2,
      pow = Math.pow,
      abs = Math.abs,
      PiBy180 = Math.PI / 180;

  /**
   * @namespace fabric.util
   */
  fabric.util = {

    /**
     * Removes value from an array.
     * Presence of value (and its position in an array) is determined via `Array.prototype.indexOf`
     * @static
     * @memberOf fabric.util
     * @param {Array} array
     * @param {Any} value
     * @return {Array} original array
     */
    removeFromArray: function(array, value) {
      var idx = array.indexOf(value);
      if (idx !== -1) {
        array.splice(idx, 1);
      }
      return array;
    },

    /**
     * Returns random number between 2 specified ones.
     * @static
     * @memberOf fabric.util
     * @param {Number} min lower limit
     * @param {Number} max upper limit
     * @return {Number} random value (between min and max)
     */
    getRandomInt: function(min, max) {
      return Math.floor(Math.random() * (max - min + 1)) + min;
    },

    /**
     * Transforms degrees to radians.
     * @static
     * @memberOf fabric.util
     * @param {Number} degrees value in degrees
     * @return {Number} value in radians
     */
    degreesToRadians: function(degrees) {
      return degrees * PiBy180;
    },

    /**
     * Transforms radians to degrees.
     * @static
     * @memberOf fabric.util
     * @param {Number} radians value in radians
     * @return {Number} value in degrees
     */
    radiansToDegrees: function(radians) {
      return radians / PiBy180;
    },

    /**
     * Rotates `point` around `origin` with `radians`
     * @static
     * @memberOf fabric.util
     * @param {fabric.Point} point The point to rotate
     * @param {fabric.Point} origin The origin of the rotation
     * @param {Number} radians The radians of the angle for the rotation
     * @return {fabric.Point} The new rotated point
     */
    rotatePoint: function(point, origin, radians) {
      point.subtractEquals(origin);
      var v = fabric.util.rotateVector(point, radians);
      return new fabric.Point(v.x, v.y).addEquals(origin);
    },

    /**
     * Rotates `vector` with `radians`
     * @static
     * @memberOf fabric.util
     * @param {Object} vector The vector to rotate (x and y)
     * @param {Number} radians The radians of the angle for the rotation
     * @return {Object} The new rotated point
     */
    rotateVector: function(vector, radians) {
      var sin = Math.sin(radians),
          cos = Math.cos(radians),
          rx = vector.x * cos - vector.y * sin,
          ry = vector.x * sin + vector.y * cos;
      return {
        x: rx,
        y: ry
      };
    },

    /**
     * Apply transform t to point p
     * @static
     * @memberOf fabric.util
     * @param  {fabric.Point} p The point to transform
     * @param  {Array} t The transform
     * @param  {Boolean} [ignoreOffset] Indicates that the offset should not be applied
     * @return {fabric.Point} The transformed point
     */
    transformPoint: function(p, t, ignoreOffset) {
      if (ignoreOffset) {
        return new fabric.Point(
          t[0] * p.x + t[2] * p.y,
          t[1] * p.x + t[3] * p.y
        );
      }
      return new fabric.Point(
        t[0] * p.x + t[2] * p.y + t[4],
        t[1] * p.x + t[3] * p.y + t[5]
      );
    },

    /**
     * Returns coordinates of points's bounding rectangle (left, top, width, height)
     * @param {Array} points 4 points array
     * @return {Object} Object with left, top, width, height properties
     */
    makeBoundingBoxFromPoints: function(points) {
      var xPoints = [points[0].x, points[1].x, points[2].x, points[3].x],
          minX = fabric.util.array.min(xPoints),
          maxX = fabric.util.array.max(xPoints),
          width = Math.abs(minX - maxX),
          yPoints = [points[0].y, points[1].y, points[2].y, points[3].y],
          minY = fabric.util.array.min(yPoints),
          maxY = fabric.util.array.max(yPoints),
          height = Math.abs(minY - maxY);

      return {
        left: minX,
        top: minY,
        width: width,
        height: height
      };
    },

    /**
     * Invert transformation t
     * @static
     * @memberOf fabric.util
     * @param {Array} t The transform
     * @return {Array} The inverted transform
     */
    invertTransform: function(t) {
      var a = 1 / (t[0] * t[3] - t[1] * t[2]),
          r = [a * t[3], -a * t[1], -a * t[2], a * t[0]],
          o = fabric.util.transformPoint({ x: t[4], y: t[5] }, r, true);
      r[4] = -o.x;
      r[5] = -o.y;
      return r;
    },

    /**
     * A wrapper around Number#toFixed, which contrary to native method returns number, not string.
     * @static
     * @memberOf fabric.util
     * @param {Number|String} number number to operate on
     * @param {Number} fractionDigits number of fraction digits to "leave"
     * @return {Number}
     */
    toFixed: function(number, fractionDigits) {
      return parseFloat(Number(number).toFixed(fractionDigits));
    },

    /**
     * Converts from attribute value to pixel value if applicable.
     * Returns converted pixels or original value not converted.
     * @param {Number|String} value number to operate on
     * @return {Number|String}
     */
    parseUnit: function(value, fontSize) {
      var unit = /\D{0,2}$/.exec(value),
          number = parseFloat(value);
      if (!fontSize) {
        fontSize = fabric.Text.DEFAULT_SVG_FONT_SIZE;
      }
      switch (unit[0]) {
        case 'mm':
          return number * fabric.DPI / 25.4;

        case 'cm':
          return number * fabric.DPI / 2.54;

        case 'in':
          return number * fabric.DPI;

        case 'pt':
          return number * fabric.DPI / 72; // or * 4 / 3

        case 'pc':
          return number * fabric.DPI / 72 * 12; // or * 16

        case 'em':
          return number * fontSize;

        default:
          return number;
      }
    },

    /**
     * Function which always returns `false`.
     * @static
     * @memberOf fabric.util
     * @return {Boolean}
     */
    falseFunction: function() {
      return false;
    },

    /**
     * Returns klass "Class" object of given namespace
     * @memberOf fabric.util
     * @param {String} type Type of object (eg. 'circle')
     * @param {String} namespace Namespace to get klass "Class" object from
     * @return {Object} klass "Class"
     */
    getKlass: function(type, namespace) {
      // capitalize first letter only
      type = fabric.util.string.camelize(type.charAt(0).toUpperCase() + type.slice(1));
      return fabric.util.resolveNamespace(namespace)[type];
    },

    /**
     * Returns object of given namespace
     * @memberOf fabric.util
     * @param {String} namespace Namespace string e.g. 'fabric.Image.filter' or 'fabric'
     * @return {Object} Object for given namespace (default fabric)
     */
    resolveNamespace: function(namespace) {
      if (!namespace) {
        return fabric;
      }

      var parts = namespace.split('.'),
          len = parts.length,
          obj = global || fabric.window;

      for (var i = 0; i < len; ++i) {
        obj = obj[parts[i]];
      }

      return obj;
    },

    /**
     * Loads image element from given url and passes it to a callback
     * @memberOf fabric.util
     * @param {String} url URL representing an image
     * @param {Function} callback Callback; invoked with loaded image
     * @param {Any} [context] Context to invoke callback in
     * @param {Object} [crossOrigin] crossOrigin value to set image element to
     */
    loadImage: function(url, callback, context, crossOrigin) {
      if (!url) {
        callback && callback.call(context, url);
        return;
      }

      var img = fabric.util.createImage();

      /** @ignore */
      img.onload = function () {
        callback && callback.call(context, img);
        img = img.onload = img.onerror = null;
      };

      /** @ignore */
      img.onerror = function() {
        fabric.log('Error loading ' + img.src);
        callback && callback.call(context, null, true);
        img = img.onload = img.onerror = null;
      };

      // data-urls appear to be buggy with crossOrigin
      // https://github.com/kangax/fabric.js/commit/d0abb90f1cd5c5ef9d2a94d3fb21a22330da3e0a#commitcomment-4513767
      // see https://code.google.com/p/chromium/issues/detail?id=315152
      //     https://bugzilla.mozilla.org/show_bug.cgi?id=935069
      if (url.indexOf('data') !== 0 && crossOrigin) {
        img.crossOrigin = crossOrigin;
      }

      img.src = url;
    },

    /**
     * Creates corresponding fabric instances from their object representations
     * @static
     * @memberOf fabric.util
     * @param {Array} objects Objects to enliven
     * @param {Function} callback Callback to invoke when all objects are created
     * @param {String} namespace Namespace to get klass "Class" object from
     * @param {Function} reviver Method for further parsing of object elements,
     * called after each fabric object created.
     */
    enlivenObjects: function(objects, callback, namespace, reviver) {
      objects = objects || [ ];

      function onLoaded() {
        if (++numLoadedObjects === numTotalObjects) {
          callback && callback(enlivenedObjects);
        }
      }

      var enlivenedObjects = [ ],
          numLoadedObjects = 0,
          numTotalObjects = objects.length;

      if (!numTotalObjects) {
        callback && callback(enlivenedObjects);
        return;
      }

      objects.forEach(function (o, index) {
        // if sparse array
        if (!o || !o.type) {
          onLoaded();
          return;
        }
        var klass = fabric.util.getKlass(o.type, namespace);
        if (klass.async) {
          klass.fromObject(o, function (obj, error) {
            if (!error) {
              enlivenedObjects[index] = obj;
              reviver && reviver(o, enlivenedObjects[index]);
            }
            onLoaded();
          });
        }
        else {
          enlivenedObjects[index] = klass.fromObject(o);
          reviver && reviver(o, enlivenedObjects[index]);
          onLoaded();
        }
      });
    },

    /**
     * Groups SVG elements (usually those retrieved from SVG document)
     * @static
     * @memberOf fabric.util
     * @param {Array} elements SVG elements to group
     * @param {Object} [options] Options object
     * @return {fabric.Object|fabric.PathGroup}
     */
    groupSVGElements: function(elements, options, path) {
      var object;

      object = new fabric.PathGroup(elements, options);

      if (typeof path !== 'undefined') {
        object.setSourcePath(path);
      }
      return object;
    },

    /**
     * Populates an object with properties of another object
     * @static
     * @memberOf fabric.util
     * @param {Object} source Source object
     * @param {Object} destination Destination object
     * @return {Array} properties Propertie names to include
     */
    populateWithProperties: function(source, destination, properties) {
      if (properties && Object.prototype.toString.call(properties) === '[object Array]') {
        for (var i = 0, len = properties.length; i < len; i++) {
          if (properties[i] in source) {
            destination[properties[i]] = source[properties[i]];
          }
        }
      }
    },

    /**
     * Draws a dashed line between two points
     *
     * This method is used to draw dashed line around selection area.
     * See <a href="http://stackoverflow.com/questions/4576724/dotted-stroke-in-canvas">dotted stroke in canvas</a>
     *
     * @param {CanvasRenderingContext2D} ctx context
     * @param {Number} x  start x coordinate
     * @param {Number} y start y coordinate
     * @param {Number} x2 end x coordinate
     * @param {Number} y2 end y coordinate
     * @param {Array} da dash array pattern
     */
    drawDashedLine: function(ctx, x, y, x2, y2, da) {
      var dx = x2 - x,
          dy = y2 - y,
          len = sqrt(dx * dx + dy * dy),
          rot = atan2(dy, dx),
          dc = da.length,
          di = 0,
          draw = true;

      ctx.save();
      ctx.translate(x, y);
      ctx.moveTo(0, 0);
      ctx.rotate(rot);

      x = 0;
      while (len > x) {
        x += da[di++ % dc];
        if (x > len) {
          x = len;
        }
        ctx[draw ? 'lineTo' : 'moveTo'](x, 0);
        draw = !draw;
      }

      ctx.restore();
    },

    /**
     * Creates canvas element and initializes it via excanvas if necessary
     * @static
     * @memberOf fabric.util
     * @param {CanvasElement} [canvasEl] optional canvas element to initialize;
     * when not given, element is created implicitly
     * @return {CanvasElement} initialized canvas element
     */
    createCanvasElement: function(canvasEl) {
      canvasEl || (canvasEl = fabric.document.createElement('canvas'));
      //jscs:disable requireCamelCaseOrUpperCaseIdentifiers
      if (!canvasEl.getContext && typeof G_vmlCanvasManager !== 'undefined') {
        G_vmlCanvasManager.initElement(canvasEl);
      }
      //jscs:enable requireCamelCaseOrUpperCaseIdentifiers
      return canvasEl;
    },

    /**
     * Creates image element (works on client and node)
     * @static
     * @memberOf fabric.util
     * @return {HTMLImageElement} HTML image element
     */
    createImage: function() {
      return fabric.isLikelyNode
        ? new (require('canvas').Image)()
        : fabric.document.createElement('img');
    },

    /**
     * Creates accessors (getXXX, setXXX) for a "class", based on "stateProperties" array
     * @static
     * @memberOf fabric.util
     * @param {Object} klass "Class" to create accessors for
     */
    createAccessors: function(klass) {
      var proto = klass.prototype;

      for (var i = proto.stateProperties.length; i--; ) {

        var propName = proto.stateProperties[i],
            capitalizedPropName = propName.charAt(0).toUpperCase() + propName.slice(1),
            setterName = 'set' + capitalizedPropName,
            getterName = 'get' + capitalizedPropName;

        // using `new Function` for better introspection
        if (!proto[getterName]) {
          proto[getterName] = (function(property) {
            return new Function('return this.get("' + property + '")');
          })(propName);
        }
        if (!proto[setterName]) {
          proto[setterName] = (function(property) {
            return new Function('value', 'return this.set("' + property + '", value)');
          })(propName);
        }
      }
    },

    /**
     * @static
     * @memberOf fabric.util
     * @param {fabric.Object} receiver Object implementing `clipTo` method
     * @param {CanvasRenderingContext2D} ctx Context to clip
     */
    clipContext: function(receiver, ctx) {
      ctx.save();
      ctx.beginPath();
      receiver.clipTo(ctx);
      ctx.clip();
    },

    /**
     * Multiply matrix A by matrix B to nest transformations
     * @static
     * @memberOf fabric.util
     * @param  {Array} a First transformMatrix
     * @param  {Array} b Second transformMatrix
     * @param  {Boolean} is2x2 flag to multiply matrices as 2x2 matrices
     * @return {Array} The product of the two transform matrices
     */
    multiplyTransformMatrices: function(a, b, is2x2) {
      // Matrix multiply a * b
      return [
        a[0] * b[0] + a[2] * b[1],
        a[1] * b[0] + a[3] * b[1],
        a[0] * b[2] + a[2] * b[3],
        a[1] * b[2] + a[3] * b[3],
        is2x2 ? 0 : a[0] * b[4] + a[2] * b[5] + a[4],
        is2x2 ? 0 : a[1] * b[4] + a[3] * b[5] + a[5]
      ];
    },

    /**
     * Decomposes standard 2x2 matrix into transform componentes
     * @static
     * @memberOf fabric.util
     * @param  {Array} a transformMatrix
     * @return {Object} Components of transform
     */
    qrDecompose: function(a) {
      var angle = atan2(a[1], a[0]),
          denom = pow(a[0], 2) + pow(a[1], 2),
          scaleX = sqrt(denom),
          scaleY = (a[0] * a[3] - a[2] * a [1]) / scaleX,
          skewX = atan2(a[0] * a[2] + a[1] * a [3], denom);
      return {
        angle: angle  / PiBy180,
        scaleX: scaleX,
        scaleY: scaleY,
        skewX: skewX / PiBy180,
        skewY: 0,
        translateX: a[4],
        translateY: a[5]
      };
    },

    customTransformMatrix: function(scaleX, scaleY, skewX) {
      var skewMatrixX = [1, 0, abs(Math.tan(skewX * PiBy180)), 1],
          scaleMatrix = [abs(scaleX), 0, 0, abs(scaleY)];
      return fabric.util.multiplyTransformMatrices(scaleMatrix, skewMatrixX, true);
    },

    resetObjectTransform: function (target) {
      target.scaleX = 1;
      target.scaleY = 1;
      target.skewX = 0;
      target.skewY = 0;
      target.flipX = false;
      target.flipY = false;
      target.setAngle(0);
    },

    /**
     * Returns string representation of function body
     * @param {Function} fn Function to get body of
     * @return {String} Function body
     */
    getFunctionBody: function(fn) {
      return (String(fn).match(/function[^{]*\{([\s\S]*)\}/) || {})[1];
    },

    /**
     * Returns true if context has transparent pixel
     * at specified location (taking tolerance into account)
     * @param {CanvasRenderingContext2D} ctx context
     * @param {Number} x x coordinate
     * @param {Number} y y coordinate
     * @param {Number} tolerance Tolerance
     */
    isTransparent: function(ctx, x, y, tolerance) {

      // If tolerance is > 0 adjust start coords to take into account.
      // If moves off Canvas fix to 0
      if (tolerance > 0) {
        if (x > tolerance) {
          x -= tolerance;
        }
        else {
          x = 0;
        }
        if (y > tolerance) {
          y -= tolerance;
        }
        else {
          y = 0;
        }
      }

      var _isTransparent = true,
          imageData = ctx.getImageData(x, y, (tolerance * 2) || 1, (tolerance * 2) || 1);

      // Split image data - for tolerance > 1, pixelDataSize = 4;
      for (var i = 3, l = imageData.data.length; i < l; i += 4) {
        var temp = imageData.data[i];
        _isTransparent = temp <= 0;
        if (_isTransparent === false) {
          break; // Stop if colour found
        }
      }

      imageData = null;

      return _isTransparent;
    },

    /**
     * Parse preserveAspectRatio attribute from element
     * @param {string} attribute to be parsed
     * @return {Object} an object containing align and meetOrSlice attribute
     */
    parsePreserveAspectRatioAttribute: function(attribute) {
      var meetOrSlice = 'meet', alignX = 'Mid', alignY = 'Mid',
          aspectRatioAttrs = attribute.split(' '), align;

      if (aspectRatioAttrs && aspectRatioAttrs.length) {
        meetOrSlice = aspectRatioAttrs.pop();
        if (meetOrSlice !== 'meet' && meetOrSlice !== 'slice') {
          align = meetOrSlice;
          meetOrSlice = 'meet';
        }
        else if (aspectRatioAttrs.length) {
          align = aspectRatioAttrs.pop();
        }
      }
      //divide align in alignX and alignY
      alignX = align !== 'none' ? align.slice(1, 4) : 'none';
      alignY = align !== 'none' ? align.slice(5, 8) : 'none';
      return {
        meetOrSlice: meetOrSlice,
        alignX: alignX,
        alignY: alignY
      };
    },

    /**
     * Clear char widths cache for a font family.
     * @memberOf fabric.util
     * @param {String} fontFamily
     */
    clearFabricFontCache: function(fontFamily) {
      if (fabric.charWidthsCache[fontFamily]) {
        fabric.charWidthsCache[fontFamily] = { };
      }
    }
  };

})(typeof exports !== 'undefined' ? exports : this);


(function() {

  var arcToSegmentsCache = { },
      segmentToBezierCache = { },
      boundsOfCurveCache = { },
      _join = Array.prototype.join;

  /* Adapted from http://dxr.mozilla.org/mozilla-central/source/content/svg/content/src/nsSVGPathDataParser.cpp
   * by Andrea Bogazzi code is under MPL. if you don't have a copy of the license you can take it here
   * http://mozilla.org/MPL/2.0/
   */
  function arcToSegments(toX, toY, rx, ry, large, sweep, rotateX) {
    var argsString = _join.call(arguments);
    if (arcToSegmentsCache[argsString]) {
      return arcToSegmentsCache[argsString];
    }

    var PI = Math.PI, th = rotateX * PI / 180,
        sinTh = Math.sin(th),
        cosTh = Math.cos(th),
        fromX = 0, fromY = 0;

    rx = Math.abs(rx);
    ry = Math.abs(ry);

    var px = -cosTh * toX * 0.5 - sinTh * toY * 0.5,
        py = -cosTh * toY * 0.5 + sinTh * toX * 0.5,
        rx2 = rx * rx, ry2 = ry * ry, py2 = py * py, px2 = px * px,
        pl = rx2 * ry2 - rx2 * py2 - ry2 * px2,
        root = 0;

    if (pl < 0) {
      var s = Math.sqrt(1 - pl/(rx2 * ry2));
      rx *= s;
      ry *= s;
    }
    else {
      root = (large === sweep ? -1.0 : 1.0) *
              Math.sqrt( pl /(rx2 * py2 + ry2 * px2));
    }

    var cx = root * rx * py / ry,
        cy = -root * ry * px / rx,
        cx1 = cosTh * cx - sinTh * cy + toX * 0.5,
        cy1 = sinTh * cx + cosTh * cy + toY * 0.5,
        mTheta = calcVectorAngle(1, 0, (px - cx) / rx, (py - cy) / ry),
        dtheta = calcVectorAngle((px - cx) / rx, (py - cy) / ry, (-px - cx) / rx, (-py - cy) / ry);

    if (sweep === 0 && dtheta > 0) {
      dtheta -= 2 * PI;
    }
    else if (sweep === 1 && dtheta < 0) {
      dtheta += 2 * PI;
    }

    // Convert into cubic bezier segments <= 90deg
    var segments = Math.ceil(Math.abs(dtheta / PI * 2)),
        result = [], mDelta = dtheta / segments,
        mT = 8 / 3 * Math.sin(mDelta / 4) * Math.sin(mDelta / 4) / Math.sin(mDelta / 2),
        th3 = mTheta + mDelta;

    for (var i = 0; i < segments; i++) {
      result[i] = segmentToBezier(mTheta, th3, cosTh, sinTh, rx, ry, cx1, cy1, mT, fromX, fromY);
      fromX = result[i][4];
      fromY = result[i][5];
      mTheta = th3;
      th3 += mDelta;
    }
    arcToSegmentsCache[argsString] = result;
    return result;
  }

  function segmentToBezier(th2, th3, cosTh, sinTh, rx, ry, cx1, cy1, mT, fromX, fromY) {
    var argsString2 = _join.call(arguments);
    if (segmentToBezierCache[argsString2]) {
      return segmentToBezierCache[argsString2];
    }

    var costh2 = Math.cos(th2),
        sinth2 = Math.sin(th2),
        costh3 = Math.cos(th3),
        sinth3 = Math.sin(th3),
        toX = cosTh * rx * costh3 - sinTh * ry * sinth3 + cx1,
        toY = sinTh * rx * costh3 + cosTh * ry * sinth3 + cy1,
        cp1X = fromX + mT * ( - cosTh * rx * sinth2 - sinTh * ry * costh2),
        cp1Y = fromY + mT * ( - sinTh * rx * sinth2 + cosTh * ry * costh2),
        cp2X = toX + mT * ( cosTh * rx * sinth3 + sinTh * ry * costh3),
        cp2Y = toY + mT * ( sinTh * rx * sinth3 - cosTh * ry * costh3);

    segmentToBezierCache[argsString2] = [
      cp1X, cp1Y,
      cp2X, cp2Y,
      toX, toY
    ];
    return segmentToBezierCache[argsString2];
  }

  /*
   * Private
   */
  function calcVectorAngle(ux, uy, vx, vy) {
    var ta = Math.atan2(uy, ux),
        tb = Math.atan2(vy, vx);
    if (tb >= ta) {
      return tb - ta;
    }
    else {
      return 2 * Math.PI - (ta - tb);
    }
  }

  /**
   * Draws arc
   * @param {CanvasRenderingContext2D} ctx
   * @param {Number} fx
   * @param {Number} fy
   * @param {Array} coords
   */
  fabric.util.drawArc = function(ctx, fx, fy, coords) {
    var rx = coords[0],
        ry = coords[1],
        rot = coords[2],
        large = coords[3],
        sweep = coords[4],
        tx = coords[5],
        ty = coords[6],
        segs = [[ ], [ ], [ ], [ ]],
        segsNorm = arcToSegments(tx - fx, ty - fy, rx, ry, large, sweep, rot);

    for (var i = 0, len = segsNorm.length; i < len; i++) {
      segs[i][0] = segsNorm[i][0] + fx;
      segs[i][1] = segsNorm[i][1] + fy;
      segs[i][2] = segsNorm[i][2] + fx;
      segs[i][3] = segsNorm[i][3] + fy;
      segs[i][4] = segsNorm[i][4] + fx;
      segs[i][5] = segsNorm[i][5] + fy;
      ctx.bezierCurveTo.apply(ctx, segs[i]);
    }
  };

  /**
   * Calculate bounding box of a elliptic-arc
   * @param {Number} fx start point of arc
   * @param {Number} fy
   * @param {Number} rx horizontal radius
   * @param {Number} ry vertical radius
   * @param {Number} rot angle of horizontal axe
   * @param {Number} large 1 or 0, whatever the arc is the big or the small on the 2 points
   * @param {Number} sweep 1 or 0, 1 clockwise or counterclockwise direction
   * @param {Number} tx end point of arc
   * @param {Number} ty
   */
  fabric.util.getBoundsOfArc = function(fx, fy, rx, ry, rot, large, sweep, tx, ty) {

    var fromX = 0, fromY = 0, bound = [ ], bounds = [ ],
    segs = arcToSegments(tx - fx, ty - fy, rx, ry, large, sweep, rot),
    boundCopy = [[ ], [ ]];

    for (var i = 0, len = segs.length; i < len; i++) {
      bound = getBoundsOfCurve(fromX, fromY, segs[i][0], segs[i][1], segs[i][2], segs[i][3], segs[i][4], segs[i][5]);
      boundCopy[0].x = bound[0].x + fx;
      boundCopy[0].y = bound[0].y + fy;
      boundCopy[1].x = bound[1].x + fx;
      boundCopy[1].y = bound[1].y + fy;
      bounds.push(boundCopy[0]);
      bounds.push(boundCopy[1]);
      fromX = segs[i][4];
      fromY = segs[i][5];
    }
    return bounds;
  };

  /**
   * Calculate bounding box of a beziercurve
   * @param {Number} x0 starting point
   * @param {Number} y0
   * @param {Number} x1 first control point
   * @param {Number} y1
   * @param {Number} x2 secondo control point
   * @param {Number} y2
   * @param {Number} x3 end of beizer
   * @param {Number} y3
   */
  // taken from http://jsbin.com/ivomiq/56/edit  no credits available for that.
  function getBoundsOfCurve(x0, y0, x1, y1, x2, y2, x3, y3) {
    var argsString = _join.call(arguments);
    if (boundsOfCurveCache[argsString]) {
      return boundsOfCurveCache[argsString];
    }

    var sqrt = Math.sqrt,
        min = Math.min, max = Math.max,
        abs = Math.abs, tvalues = [ ],
        bounds = [[ ], [ ]],
        a, b, c, t, t1, t2, b2ac, sqrtb2ac;

    b = 6 * x0 - 12 * x1 + 6 * x2;
    a = -3 * x0 + 9 * x1 - 9 * x2 + 3 * x3;
    c = 3 * x1 - 3 * x0;

    for (var i = 0; i < 2; ++i) {
      if (i > 0) {
        b = 6 * y0 - 12 * y1 + 6 * y2;
        a = -3 * y0 + 9 * y1 - 9 * y2 + 3 * y3;
        c = 3 * y1 - 3 * y0;
      }

      if (abs(a) < 1e-12) {
        if (abs(b) < 1e-12) {
          continue;
        }
        t = -c / b;
        if (0 < t && t < 1) {
          tvalues.push(t);
        }
        continue;
      }
      b2ac = b * b - 4 * c * a;
      if (b2ac < 0) {
        continue;
      }
      sqrtb2ac = sqrt(b2ac);
      t1 = (-b + sqrtb2ac) / (2 * a);
      if (0 < t1 && t1 < 1) {
        tvalues.push(t1);
      }
      t2 = (-b - sqrtb2ac) / (2 * a);
      if (0 < t2 && t2 < 1) {
        tvalues.push(t2);
      }
    }

    var x, y, j = tvalues.length, jlen = j, mt;
    while (j--) {
      t = tvalues[j];
      mt = 1 - t;
      x = (mt * mt * mt * x0) + (3 * mt * mt * t * x1) + (3 * mt * t * t * x2) + (t * t * t * x3);
      bounds[0][j] = x;

      y = (mt * mt * mt * y0) + (3 * mt * mt * t * y1) + (3 * mt * t * t * y2) + (t * t * t * y3);
      bounds[1][j] = y;
    }

    bounds[0][jlen] = x0;
    bounds[1][jlen] = y0;
    bounds[0][jlen + 1] = x3;
    bounds[1][jlen + 1] = y3;
    var result = [
      {
        x: min.apply(null, bounds[0]),
        y: min.apply(null, bounds[1])
      },
      {
        x: max.apply(null, bounds[0]),
        y: max.apply(null, bounds[1])
      }
    ];
    boundsOfCurveCache[argsString] = result;
    return result;
  }

  fabric.util.getBoundsOfCurve = getBoundsOfCurve;

})();


(function() {

  var slice = Array.prototype.slice;

  /* _ES5_COMPAT_START_ */

  if (!Array.prototype.indexOf) {
    /**
     * Finds index of an element in an array
     * @param {Any} searchElement
     * @param {Number} [fromIndex]
     * @return {Number}
     */
    Array.prototype.indexOf = function (searchElement /*, fromIndex */ ) {
      if (this === void 0 || this === null) {
        throw new TypeError();
      }
      var t = Object(this), len = t.length >>> 0;
      if (len === 0) {
        return -1;
      }
      var n = 0;
      if (arguments.length > 0) {
        n = Number(arguments[1]);
        if (n !== n) { // shortcut for verifying if it's NaN
          n = 0;
        }
        else if (n !== 0 && n !== Number.POSITIVE_INFINITY && n !== Number.NEGATIVE_INFINITY) {
          n = (n > 0 || -1) * Math.floor(Math.abs(n));
        }
      }
      if (n >= len) {
        return -1;
      }
      var k = n >= 0 ? n : Math.max(len - Math.abs(n), 0);
      for (; k < len; k++) {
        if (k in t && t[k] === searchElement) {
          return k;
        }
      }
      return -1;
    };
  }

  if (!Array.prototype.forEach) {
    /**
     * Iterates an array, invoking callback for each element
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [context] Context to invoke callback in
     * @return {Array}
     */
    Array.prototype.forEach = function(fn, context) {
      for (var i = 0, len = this.length >>> 0; i < len; i++) {
        if (i in this) {
          fn.call(context, this[i], i, this);
        }
      }
    };
  }

  if (!Array.prototype.map) {
    /**
     * Returns a result of iterating over an array, invoking callback for each element
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [context] Context to invoke callback in
     * @return {Array}
     */
    Array.prototype.map = function(fn, context) {
      var result = [ ];
      for (var i = 0, len = this.length >>> 0; i < len; i++) {
        if (i in this) {
          result[i] = fn.call(context, this[i], i, this);
        }
      }
      return result;
    };
  }

  if (!Array.prototype.every) {
    /**
     * Returns true if a callback returns truthy value for all elements in an array
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [context] Context to invoke callback in
     * @return {Boolean}
     */
    Array.prototype.every = function(fn, context) {
      for (var i = 0, len = this.length >>> 0; i < len; i++) {
        if (i in this && !fn.call(context, this[i], i, this)) {
          return false;
        }
      }
      return true;
    };
  }

  if (!Array.prototype.some) {
    /**
     * Returns true if a callback returns truthy value for at least one element in an array
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [context] Context to invoke callback in
     * @return {Boolean}
     */
    Array.prototype.some = function(fn, context) {
      for (var i = 0, len = this.length >>> 0; i < len; i++) {
        if (i in this && fn.call(context, this[i], i, this)) {
          return true;
        }
      }
      return false;
    };
  }

  if (!Array.prototype.filter) {
    /**
     * Returns the result of iterating over elements in an array
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [context] Context to invoke callback in
     * @return {Array}
     */
    Array.prototype.filter = function(fn, context) {
      var result = [ ], val;
      for (var i = 0, len = this.length >>> 0; i < len; i++) {
        if (i in this) {
          val = this[i]; // in case fn mutates this
          if (fn.call(context, val, i, this)) {
            result.push(val);
          }
        }
      }
      return result;
    };
  }

  if (!Array.prototype.reduce) {
    /**
     * Returns "folded" (reduced) result of iterating over elements in an array
     * @param {Function} fn Callback to invoke for each element
     * @param {Object} [initial] Object to use as the first argument to the first call of the callback
     * @return {Any}
     */
    Array.prototype.reduce = function(fn /*, initial*/) {
      var len = this.length >>> 0,
          i = 0,
          rv;

      if (arguments.length > 1) {
        rv = arguments[1];
      }
      else {
        do {
          if (i in this) {
            rv = this[i++];
            break;
          }
          // if array contains no values, no initial value to return
          if (++i >= len) {
            throw new TypeError();
          }
        }
        while (true);
      }
      for (; i < len; i++) {
        if (i in this) {
          rv = fn.call(null, rv, this[i], i, this);
        }
      }
      return rv;
    };
  }

  /* _ES5_COMPAT_END_ */

  /**
   * Invokes method on all items in a given array
   * @memberOf fabric.util.array
   * @param {Array} array Array to iterate over
   * @param {String} method Name of a method to invoke
   * @return {Array}
   */
  function invoke(array, method) {
    var args = slice.call(arguments, 2), result = [ ];
    for (var i = 0, len = array.length; i < len; i++) {
      result[i] = args.length ? array[i][method].apply(array[i], args) : array[i][method].call(array[i]);
    }
    return result;
  }

  /**
   * Finds maximum value in array (not necessarily "first" one)
   * @memberOf fabric.util.array
   * @param {Array} array Array to iterate over
   * @param {String} byProperty
   * @return {Any}
   */
  function max(array, byProperty) {
    return find(array, byProperty, function(value1, value2) {
      return value1 >= value2;
    });
  }

  /**
   * Finds minimum value in array (not necessarily "first" one)
   * @memberOf fabric.util.array
   * @param {Array} array Array to iterate over
   * @param {String} byProperty
   * @return {Any}
   */
  function min(array, byProperty) {
    return find(array, byProperty, function(value1, value2) {
      return value1 < value2;
    });
  }

  /**
   * @private
   */
  function fill(array, value) {
    var k = array.length;
    while (k--) {
      array[k] = value;
    }
    return array;
  }

  /**
   * @private
   */
  function find(array, byProperty, condition) {
    if (!array || array.length === 0) {
      return;
    }

    var i = array.length - 1,
        result = byProperty ? array[i][byProperty] : array[i];
    if (byProperty) {
      while (i--) {
        if (condition(array[i][byProperty], result)) {
          result = array[i][byProperty];
        }
      }
    }
    else {
      while (i--) {
        if (condition(array[i], result)) {
          result = array[i];
        }
      }
    }
    return result;
  }

  /**
   * @namespace fabric.util.array
   */
  fabric.util.array = {
    fill: fill,
    invoke: invoke,
    min: min,
    max: max
  };

})();


(function() {

  /**
   * Copies all enumerable properties of one object to another
   * @memberOf fabric.util.object
   * @param {Object} destination Where to copy to
   * @param {Object} source Where to copy from
   * @return {Object}
   */
  function extend(destination, source) {
    // JScript DontEnum bug is not taken care of
    for (var property in source) {
      destination[property] = source[property];
    }
    return destination;
  }

  /**
   * Creates an empty object and copies all enumerable properties of another object to it
   * @memberOf fabric.util.object
   * @param {Object} object Object to clone
   * @return {Object}
   */
  function clone(object) {
    return extend({ }, object);
  }

  /** @namespace fabric.util.object */
  fabric.util.object = {
    extend: extend,
    clone: clone
  };

})();


(function() {

  /* _ES5_COMPAT_START_ */
  if (!String.prototype.trim) {
    /**
     * Trims a string (removing whitespace from the beginning and the end)
     * @function external:String#trim
     * @see <a href="https://developer.mozilla.org/en/JavaScript/Reference/Global_Objects/String/Trim">String#trim on MDN</a>
     */
    String.prototype.trim = function () {
      // this trim is not fully ES3 or ES5 compliant, but it should cover most cases for now
      return this.replace(/^[\s\xA0]+/, '').replace(/[\s\xA0]+$/, '');
    };
  }
  /* _ES5_COMPAT_END_ */

  /**
   * Camelizes a string
   * @memberOf fabric.util.string
   * @param {String} string String to camelize
   * @return {String} Camelized version of a string
   */
  function camelize(string) {
    return string.replace(/-+(.)?/g, function(match, character) {
      return character ? character.toUpperCase() : '';
    });
  }

  /**
   * Capitalizes a string
   * @memberOf fabric.util.string
   * @param {String} string String to capitalize
   * @param {Boolean} [firstLetterOnly] If true only first letter is capitalized
   * and other letters stay untouched, if false first letter is capitalized
   * and other letters are converted to lowercase.
   * @return {String} Capitalized version of a string
   */
  function capitalize(string, firstLetterOnly) {
    return string.charAt(0).toUpperCase() +
      (firstLetterOnly ? string.slice(1) : string.slice(1).toLowerCase());
  }

  /**
   * Escapes XML in a string
   * @memberOf fabric.util.string
   * @param {String} string String to escape
   * @return {String} Escaped version of a string
   */
  function escapeXml(string) {
    return string.replace(/&/g, '&amp;')
       .replace(/"/g, '&quot;')
       .replace(/'/g, '&apos;')
       .replace(/</g, '&lt;')
       .replace(/>/g, '&gt;');
  }

  /**
   * String utilities
   * @namespace fabric.util.string
   */
  fabric.util.string = {
    camelize: camelize,
    capitalize: capitalize,
    escapeXml: escapeXml
  };
}());


/* _ES5_COMPAT_START_ */
(function() {

  var slice = Array.prototype.slice,
      apply = Function.prototype.apply,
      Dummy = function() { };

  if (!Function.prototype.bind) {
    /**
     * Cross-browser approximation of ES5 Function.prototype.bind (not fully spec conforming)
     * @see <a href="https://developer.mozilla.org/en/JavaScript/Reference/Global_Objects/Function/bind">Function#bind on MDN</a>
     * @param {Object} thisArg Object to bind function to
     * @param {Any[]} Values to pass to a bound function
     * @return {Function}
     */
    Function.prototype.bind = function(thisArg) {
      var _this = this, args = slice.call(arguments, 1), bound;
      if (args.length) {
        bound = function() {
          return apply.call(_this, this instanceof Dummy ? this : thisArg, args.concat(slice.call(arguments)));
        };
      }
      else {
        /** @ignore */
        bound = function() {
          return apply.call(_this, this instanceof Dummy ? this : thisArg, arguments);
        };
      }
      Dummy.prototype = this.prototype;
      bound.prototype = new Dummy();

      return bound;
    };
  }

})();
/* _ES5_COMPAT_END_ */


(function() {

  var slice = Array.prototype.slice, emptyFunction = function() { },

      IS_DONTENUM_BUGGY = (function() {
        for (var p in { toString: 1 }) {
          if (p === 'toString') {
            return false;
          }
        }
        return true;
      })(),

      /** @ignore */
      addMethods = function(klass, source, parent) {
        for (var property in source) {

          if (property in klass.prototype &&
              typeof klass.prototype[property] === 'function' &&
              (source[property] + '').indexOf('callSuper') > -1) {

            klass.prototype[property] = (function(property) {
              return function() {

                var superclass = this.constructor.superclass;
                this.constructor.superclass = parent;
                var returnValue = source[property].apply(this, arguments);
                this.constructor.superclass = superclass;

                if (property !== 'initialize') {
                  return returnValue;
                }
              };
            })(property);
          }
          else {
            klass.prototype[property] = source[property];
          }

          if (IS_DONTENUM_BUGGY) {
            if (source.toString !== Object.prototype.toString) {
              klass.prototype.toString = source.toString;
            }
            if (source.valueOf !== Object.prototype.valueOf) {
              klass.prototype.valueOf = source.valueOf;
            }
          }
        }
      };

  function Subclass() { }

  function callSuper(methodName) {
    var fn = this.constructor.superclass.prototype[methodName];
    return (arguments.length > 1)
      ? fn.apply(this, slice.call(arguments, 1))
      : fn.call(this);
  }

  /**
   * Helper for creation of "classes".
   * @memberOf fabric.util
   * @param {Function} [parent] optional "Class" to inherit from
   * @param {Object} [properties] Properties shared by all instances of this class
   *                  (be careful modifying objects defined here as this would affect all instances)
   */
  function createClass() {
    var parent = null,
        properties = slice.call(arguments, 0);

    if (typeof properties[0] === 'function') {
      parent = properties.shift();
    }
    function klass() {
      this.initialize.apply(this, arguments);
    }

    klass.superclass = parent;
    klass.subclasses = [ ];

    if (parent) {
      Subclass.prototype = parent.prototype;
      klass.prototype = new Subclass();
      parent.subclasses.push(klass);
    }
    for (var i = 0, length = properties.length; i < length; i++) {
      addMethods(klass, properties[i], parent);
    }
    if (!klass.prototype.initialize) {
      klass.prototype.initialize = emptyFunction;
    }
    klass.prototype.constructor = klass;
    klass.prototype.callSuper = callSuper;
    return klass;
  }

  fabric.util.createClass = createClass;
})();


(function () {

  var unknown = 'unknown';

  /* EVENT HANDLING */

  function areHostMethods(object) {
    var methodNames = Array.prototype.slice.call(arguments, 1),
        t, i, len = methodNames.length;
    for (i = 0; i < len; i++) {
      t = typeof object[methodNames[i]];
      if (!(/^(?:function|object|unknown)$/).test(t)) {
        return false;
      }
    }
    return true;
  }

  /** @ignore */
  var getElement,
      setElement,
      getUniqueId = (function () {
        var uid = 0;
        return function (element) {
          return element.__uniqueID || (element.__uniqueID = 'uniqueID__' + uid++);
        };
      })();

  (function () {
    var elements = { };
    /** @ignore */
    getElement = function (uid) {
      return elements[uid];
    };
    /** @ignore */
    setElement = function (uid, element) {
      elements[uid] = element;
    };
  })();

  function createListener(uid, handler) {
    return {
      handler: handler,
      wrappedHandler: createWrappedHandler(uid, handler)
    };
  }

  function createWrappedHandler(uid, handler) {
    return function (e) {
      handler.call(getElement(uid), e || fabric.window.event);
    };
  }

  function createDispatcher(uid, eventName) {
    return function (e) {
      if (handlers[uid] && handlers[uid][eventName]) {
        var handlersForEvent = handlers[uid][eventName];
        for (var i = 0, len = handlersForEvent.length; i < len; i++) {
          handlersForEvent[i].call(this, e || fabric.window.event);
        }
      }
    };
  }

  var shouldUseAddListenerRemoveListener = (
        areHostMethods(fabric.document.documentElement, 'addEventListener', 'removeEventListener') &&
        areHostMethods(fabric.window, 'addEventListener', 'removeEventListener')),

      shouldUseAttachEventDetachEvent = (
        areHostMethods(fabric.document.documentElement, 'attachEvent', 'detachEvent') &&
        areHostMethods(fabric.window, 'attachEvent', 'detachEvent')),

      // IE branch
      listeners = { },

      // DOM L0 branch
      handlers = { },

      addListener, removeListener;

  if (shouldUseAddListenerRemoveListener) {
    /** @ignore */
    addListener = function (element, eventName, handler) {
      element.addEventListener(eventName, handler, false);
    };
    /** @ignore */
    removeListener = function (element, eventName, handler) {
      element.removeEventListener(eventName, handler, false);
    };
  }

  else if (shouldUseAttachEventDetachEvent) {
    /** @ignore */
    addListener = function (element, eventName, handler) {
      var uid = getUniqueId(element);
      setElement(uid, element);
      if (!listeners[uid]) {
        listeners[uid] = { };
      }
      if (!listeners[uid][eventName]) {
        listeners[uid][eventName] = [ ];

      }
      var listener = createListener(uid, handler);
      listeners[uid][eventName].push(listener);
      element.attachEvent('on' + eventName, listener.wrappedHandler);
    };
    /** @ignore */
    removeListener = function (element, eventName, handler) {
      var uid = getUniqueId(element), listener;
      if (listeners[uid] && listeners[uid][eventName]) {
        for (var i = 0, len = listeners[uid][eventName].length; i < len; i++) {
          listener = listeners[uid][eventName][i];
          if (listener && listener.handler === handler) {
            element.detachEvent('on' + eventName, listener.wrappedHandler);
            listeners[uid][eventName][i] = null;
          }
        }
      }
    };
  }
  else {
    /** @ignore */
    addListener = function (element, eventName, handler) {
      var uid = getUniqueId(element);
      if (!handlers[uid]) {
        handlers[uid] = { };
      }
      if (!handlers[uid][eventName]) {
        handlers[uid][eventName] = [ ];
        var existingHandler = element['on' + eventName];
        if (existingHandler) {
          handlers[uid][eventName].push(existingHandler);
        }
        element['on' + eventName] = createDispatcher(uid, eventName);
      }
      handlers[uid][eventName].push(handler);
    };
    /** @ignore */
    removeListener = function (element, eventName, handler) {
      var uid = getUniqueId(element);
      if (handlers[uid] && handlers[uid][eventName]) {
        var handlersForEvent = handlers[uid][eventName];
        for (var i = 0, len = handlersForEvent.length; i < len; i++) {
          if (handlersForEvent[i] === handler) {
            handlersForEvent.splice(i, 1);
          }
        }
      }
    };
  }

  /**
   * Adds an event listener to an element
   * @function
   * @memberOf fabric.util
   * @param {HTMLElement} element
   * @param {String} eventName
   * @param {Function} handler
   */
  fabric.util.addListener = addListener;

  /**
   * Removes an event listener from an element
   * @function
   * @memberOf fabric.util
   * @param {HTMLElement} element
   * @param {String} eventName
   * @param {Function} handler
   */
  fabric.util.removeListener = removeListener;

  /**
   * Cross-browser wrapper for getting event's coordinates
   * @memberOf fabric.util
   * @param {Event} event Event object
   */
  function getPointer(event) {
    event || (event = fabric.window.event);

    var element = event.target ||
                  (typeof event.srcElement !== unknown ? event.srcElement : null),

        scroll = fabric.util.getScrollLeftTop(element);

    return {
      x: pointerX(event) + scroll.left,
      y: pointerY(event) + scroll.top
    };
  }

  var pointerX = function(event) {
    // looks like in IE (<9) clientX at certain point (apparently when mouseup fires on VML element)
    // is represented as COM object, with all the consequences, like "unknown" type and error on [[Get]]
    // need to investigate later
    return (typeof event.clientX !== unknown ? event.clientX : 0);
  },

  pointerY = function(event) {
    return (typeof event.clientY !== unknown ? event.clientY : 0);
  };

  function _getPointer(event, pageProp, clientProp) {
    var touchProp = event.type === 'touchend' ? 'changedTouches' : 'touches';

    return (event[touchProp] && event[touchProp][0]
      ? (event[touchProp][0][pageProp] - (event[touchProp][0][pageProp] - event[touchProp][0][clientProp]))
        || event[clientProp]
      : event[clientProp]);
  }

  if (fabric.isTouchSupported) {
    pointerX = function(event) {
      return _getPointer(event, 'pageX', 'clientX');
    };
    pointerY = function(event) {
      return _getPointer(event, 'pageY', 'clientY');
    };
  }

  fabric.util.getPointer = getPointer;

  fabric.util.object.extend(fabric.util, fabric.Observable);

})();


(function () {

  /**
   * Cross-browser wrapper for setting element's style
   * @memberOf fabric.util
   * @param {HTMLElement} element
   * @param {Object} styles
   * @return {HTMLElement} Element that was passed as a first argument
   */
  function setStyle(element, styles) {
    var elementStyle = element.style;
    if (!elementStyle) {
      return element;
    }
    if (typeof styles === 'string') {
      element.style.cssText += ';' + styles;
      return styles.indexOf('opacity') > -1
        ? setOpacity(element, styles.match(/opacity:\s*(\d?\.?\d*)/)[1])
        : element;
    }
    for (var property in styles) {
      if (property === 'opacity') {
        setOpacity(element, styles[property]);
      }
      else {
        var normalizedProperty = (property === 'float' || property === 'cssFloat')
          ? (typeof elementStyle.styleFloat === 'undefined' ? 'cssFloat' : 'styleFloat')
          : property;
        elementStyle[normalizedProperty] = styles[property];
      }
    }
    return element;
  }

  var parseEl = fabric.document.createElement('div'),
      supportsOpacity = typeof parseEl.style.opacity === 'string',
      supportsFilters = typeof parseEl.style.filter === 'string',
      reOpacity = /alpha\s*\(\s*opacity\s*=\s*([^\)]+)\)/,

      /** @ignore */
      setOpacity = function (element) { return element; };

  if (supportsOpacity) {
    /** @ignore */
    setOpacity = function(element, value) {
      element.style.opacity = value;
      return element;
    };
  }
  else if (supportsFilters) {
    /** @ignore */
    setOpacity = function(element, value) {
      var es = element.style;
      if (element.currentStyle && !element.currentStyle.hasLayout) {
        es.zoom = 1;
      }
      if (reOpacity.test(es.filter)) {
        value = value >= 0.9999 ? '' : ('alpha(opacity=' + (value * 100) + ')');
        es.filter = es.filter.replace(reOpacity, value);
      }
      else {
        es.filter += ' alpha(opacity=' + (value * 100) + ')';
      }
      return element;
    };
  }

  fabric.util.setStyle = setStyle;

})();


(function() {

  var _slice = Array.prototype.slice;

  /**
   * Takes id and returns an element with that id (if one exists in a document)
   * @memberOf fabric.util
   * @param {String|HTMLElement} id
   * @return {HTMLElement|null}
   */
  function getById(id) {
    return typeof id === 'string' ? fabric.document.getElementById(id) : id;
  }

  var sliceCanConvertNodelists,
      /**
       * Converts an array-like object (e.g. arguments or NodeList) to an array
       * @memberOf fabric.util
       * @param {Object} arrayLike
       * @return {Array}
       */
      toArray = function(arrayLike) {
        return _slice.call(arrayLike, 0);
      };

  try {
    sliceCanConvertNodelists = toArray(fabric.document.childNodes) instanceof Array;
  }
  catch (err) { }

  if (!sliceCanConvertNodelists) {
    toArray = function(arrayLike) {
      var arr = new Array(arrayLike.length), i = arrayLike.length;
      while (i--) {
        arr[i] = arrayLike[i];
      }
      return arr;
    };
  }

  /**
   * Creates specified element with specified attributes
   * @memberOf fabric.util
   * @param {String} tagName Type of an element to create
   * @param {Object} [attributes] Attributes to set on an element
   * @return {HTMLElement} Newly created element
   */
  function makeElement(tagName, attributes) {
    var el = fabric.document.createElement(tagName);
    for (var prop in attributes) {
      if (prop === 'class') {
        el.className = attributes[prop];
      }
      else if (prop === 'for') {
        el.htmlFor = attributes[prop];
      }
      else {
        el.setAttribute(prop, attributes[prop]);
      }
    }
    return el;
  }

  /**
   * Adds class to an element
   * @memberOf fabric.util
   * @param {HTMLElement} element Element to add class to
   * @param {String} className Class to add to an element
   */
  function addClass(element, className) {
    if (element && (' ' + element.className + ' ').indexOf(' ' + className + ' ') === -1) {
      element.className += (element.className ? ' ' : '') + className;
    }
  }

  /**
   * Wraps element with another element
   * @memberOf fabric.util
   * @param {HTMLElement} element Element to wrap
   * @param {HTMLElement|String} wrapper Element to wrap with
   * @param {Object} [attributes] Attributes to set on a wrapper
   * @return {HTMLElement} wrapper
   */
  function wrapElement(element, wrapper, attributes) {
    if (typeof wrapper === 'string') {
      wrapper = makeElement(wrapper, attributes);
    }
    if (element.parentNode) {
      element.parentNode.replaceChild(wrapper, element);
    }
    wrapper.appendChild(element);
    return wrapper;
  }

  /**
   * Returns element scroll offsets
   * @memberOf fabric.util
   * @param {HTMLElement} element Element to operate on
   * @return {Object} Object with left/top values
   */
  function getScrollLeftTop(element) {

    var left = 0,
        top = 0,
        docElement = fabric.document.documentElement,
        body = fabric.document.body || {
          scrollLeft: 0, scrollTop: 0
        };

    // While loop checks (and then sets element to) .parentNode OR .host
    //  to account for ShadowDOM. We still want to traverse up out of ShadowDOM,
    //  but the .parentNode of a root ShadowDOM node will always be null, instead
    //  it should be accessed through .host. See http://stackoverflow.com/a/24765528/4383938
    while (element && (element.parentNode || element.host)) {

      // Set element to element parent, or 'host' in case of ShadowDOM
      element = element.parentNode || element.host;

      if (element === fabric.document) {
        left = body.scrollLeft || docElement.scrollLeft || 0;
        top = body.scrollTop ||  docElement.scrollTop || 0;
      }
      else {
        left += element.scrollLeft || 0;
        top += element.scrollTop || 0;
      }

      if (element.nodeType === 1 &&
          fabric.util.getElementStyle(element, 'position') === 'fixed') {
        break;
      }
    }

    return { left: left, top: top };
  }

  /**
   * Returns offset for a given element
   * @function
   * @memberOf fabric.util
   * @param {HTMLElement} element Element to get offset for
   * @return {Object} Object with "left" and "top" properties
   */
  function getElementOffset(element) {
    var docElem,
        doc = element && element.ownerDocument,
        box = { left: 0, top: 0 },
        offset = { left: 0, top: 0 },
        scrollLeftTop,
        offsetAttributes = {
          borderLeftWidth: 'left',
          borderTopWidth:  'top',
          paddingLeft:     'left',
          paddingTop:      'top'
        };

    if (!doc) {
      return offset;
    }

    for (var attr in offsetAttributes) {
      offset[offsetAttributes[attr]] += parseInt(getElementStyle(element, attr), 10) || 0;
    }

    docElem = doc.documentElement;
    if ( typeof element.getBoundingClientRect !== 'undefined' ) {
      box = element.getBoundingClientRect();
    }

    scrollLeftTop = getScrollLeftTop(element);

    return {
      left: box.left + scrollLeftTop.left - (docElem.clientLeft || 0) + offset.left,
      top: box.top + scrollLeftTop.top - (docElem.clientTop || 0)  + offset.top
    };
  }

  /**
   * Returns style attribute value of a given element
   * @memberOf fabric.util
   * @param {HTMLElement} element Element to get style attribute for
   * @param {String} attr Style attribute to get for element
   * @return {String} Style attribute value of the given element.
   */
  var getElementStyle;
  if (fabric.document.defaultView && fabric.document.defaultView.getComputedStyle) {
    getElementStyle = function(element, attr) {
      var style = fabric.document.defaultView.getComputedStyle(element, null);
      return style ? style[attr] : undefined;
    };
  }
  else {
    getElementStyle = function(element, attr) {
      var value = element.style[attr];
      if (!value && element.currentStyle) {
        value = element.currentStyle[attr];
      }
      return value;
    };
  }

  (function () {
    var style = fabric.document.documentElement.style,
        selectProp = 'userSelect' in style
          ? 'userSelect'
          : 'MozUserSelect' in style
            ? 'MozUserSelect'
            : 'WebkitUserSelect' in style
              ? 'WebkitUserSelect'
              : 'KhtmlUserSelect' in style
                ? 'KhtmlUserSelect'
                : '';

    /**
     * Makes element unselectable
     * @memberOf fabric.util
     * @param {HTMLElement} element Element to make unselectable
     * @return {HTMLElement} Element that was passed in
     */
    function makeElementUnselectable(element) {
      if (typeof element.onselectstart !== 'undefined') {
        element.onselectstart = fabric.util.falseFunction;
      }
      if (selectProp) {
        element.style[selectProp] = 'none';
      }
      else if (typeof element.unselectable === 'string') {
        element.unselectable = 'on';
      }
      return element;
    }

    /**
     * Makes element selectable
     * @memberOf fabric.util
     * @param {HTMLElement} element Element to make selectable
     * @return {HTMLElement} Element that was passed in
     */
    function makeElementSelectable(element) {
      if (typeof element.onselectstart !== 'undefined') {
        element.onselectstart = null;
      }
      if (selectProp) {
        element.style[selectProp] = '';
      }
      else if (typeof element.unselectable === 'string') {
        element.unselectable = '';
      }
      return element;
    }

    fabric.util.makeElementUnselectable = makeElementUnselectable;
    fabric.util.makeElementSelectable = makeElementSelectable;
  })();

  (function() {

    /**
     * Inserts a script element with a given url into a document; invokes callback, when that script is finished loading
     * @memberOf fabric.util
     * @param {String} url URL of a script to load
     * @param {Function} callback Callback to execute when script is finished loading
     */
    function getScript(url, callback) {
      var headEl = fabric.document.getElementsByTagName('head')[0],
          scriptEl = fabric.document.createElement('script'),
          loading = true;

      /** @ignore */
      scriptEl.onload = /** @ignore */ scriptEl.onreadystatechange = function(e) {
        if (loading) {
          if (typeof this.readyState === 'string' &&
              this.readyState !== 'loaded' &&
              this.readyState !== 'complete') {
            return;
          }
          loading = false;
          callback(e || fabric.window.event);
          scriptEl = scriptEl.onload = scriptEl.onreadystatechange = null;
        }
      };
      scriptEl.src = url;
      headEl.appendChild(scriptEl);
      // causes issue in Opera
      // headEl.removeChild(scriptEl);
    }

    fabric.util.getScript = getScript;
  })();

  fabric.util.getById = getById;
  fabric.util.toArray = toArray;
  fabric.util.makeElement = makeElement;
  fabric.util.addClass = addClass;
  fabric.util.wrapElement = wrapElement;
  fabric.util.getScrollLeftTop = getScrollLeftTop;
  fabric.util.getElementOffset = getElementOffset;
  fabric.util.getElementStyle = getElementStyle;

})();


(function() {

  function addParamToUrl(url, param) {
    return url + (/\?/.test(url) ? '&' : '?') + param;
  }

  var makeXHR = (function() {
    var factories = [
      function() { return new ActiveXObject('Microsoft.XMLHTTP'); },
      function() { return new ActiveXObject('Msxml2.XMLHTTP'); },
      function() { return new ActiveXObject('Msxml2.XMLHTTP.3.0'); },
      function() { return new XMLHttpRequest(); }
    ];
    for (var i = factories.length; i--; ) {
      try {
        var req = factories[i]();
        if (req) {
          return factories[i];
        }
      }
      catch (err) { }
    }
  })();

  function emptyFn() { }

  /**
   * Cross-browser abstraction for sending XMLHttpRequest
   * @memberOf fabric.util
   * @param {String} url URL to send XMLHttpRequest to
   * @param {Object} [options] Options object
   * @param {String} [options.method="GET"]
   * @param {Function} options.onComplete Callback to invoke when request is completed
   * @return {XMLHttpRequest} request
   */
  function request(url, options) {

    options || (options = { });

    var method = options.method ? options.method.toUpperCase() : 'GET',
        onComplete = options.onComplete || function() { },
        xhr = makeXHR(),
        body;

    /** @ignore */
    xhr.onreadystatechange = function() {
      if (xhr.readyState === 4) {
        onComplete(xhr);
        xhr.onreadystatechange = emptyFn;
      }
    };

    if (method === 'GET') {
      body = null;
      if (typeof options.parameters === 'string') {
        url = addParamToUrl(url, options.parameters);
      }
    }

    xhr.open(method, url, true);

    if (method === 'POST' || method === 'PUT') {
      xhr.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
    }

    xhr.send(body);
    return xhr;
  }

  fabric.util.request = request;
})();


/**
 * Wrapper around `console.log` (when available)
 * @param {Any} [values] Values to log
 */
fabric.log = function() { };

/**
 * Wrapper around `console.warn` (when available)
 * @param {Any} [values] Values to log as a warning
 */
fabric.warn = function() { };

/* jshint ignore:start */
if (typeof console !== 'undefined') {

  ['log', 'warn'].forEach(function(methodName) {

    if (typeof console[methodName] !== 'undefined' &&
        typeof console[methodName].apply === 'function') {

      fabric[methodName] = function() {
        return console[methodName].apply(console, arguments);
      };
    }
  });
}
/* jshint ignore:end */


(function() {

  /**
   * Changes value from one to another within certain period of time, invoking callbacks as value is being changed.
   * @memberOf fabric.util
   * @param {Object} [options] Animation options
   * @param {Function} [options.onChange] Callback; invoked on every value change
   * @param {Function} [options.onComplete] Callback; invoked when value change is completed
   * @param {Number} [options.startValue=0] Starting value
   * @param {Number} [options.endValue=100] Ending value
   * @param {Number} [options.byValue=100] Value to modify the property by
   * @param {Function} [options.easing] Easing function
   * @param {Number} [options.duration=500] Duration of change (in ms)
   */
  function animate(options) {

    requestAnimFrame(function(timestamp) {
      options || (options = { });

      var start = timestamp || +new Date(),
          duration = options.duration || 500,
          finish = start + duration, time,
          onChange = options.onChange || function() { },
          abort = options.abort || function() { return false; },
          easing = options.easing || function(t, b, c, d) {return -c * Math.cos(t / d * (Math.PI / 2)) + c + b;},
          startValue = 'startValue' in options ? options.startValue : 0,
          endValue = 'endValue' in options ? options.endValue : 100,
          byValue = options.byValue || endValue - startValue;

      options.onStart && options.onStart();

      (function tick(ticktime) {
        time = ticktime || +new Date();
        var currentTime = time > finish ? duration : (time - start);
        if (abort()) {
          options.onComplete && options.onComplete();
          return;
        }
        onChange(easing(currentTime, startValue, byValue, duration));
        if (time > finish) {
          options.onComplete && options.onComplete();
          return;
        }
        requestAnimFrame(tick);
      })(start);
    });

  }

  var _requestAnimFrame = fabric.window.requestAnimationFrame       ||
                          fabric.window.webkitRequestAnimationFrame ||
                          fabric.window.mozRequestAnimationFrame    ||
                          fabric.window.oRequestAnimationFrame      ||
                          fabric.window.msRequestAnimationFrame     ||
                          function(callback) {
                            fabric.window.setTimeout(callback, 1000 / 60);
                          };

  /**
   * requestAnimationFrame polyfill based on http://paulirish.com/2011/requestanimationframe-for-smart-animating/
   * In order to get a precise start time, `requestAnimFrame` should be called as an entry into the method
   * @memberOf fabric.util
   * @param {Function} callback Callback to invoke
   * @param {DOMElement} element optional Element to associate with animation
   */
  function requestAnimFrame() {
    return _requestAnimFrame.apply(fabric.window, arguments);
  }

  fabric.util.animate = animate;
  fabric.util.requestAnimFrame = requestAnimFrame;

})();


(function() {

  function normalize(a, c, p, s) {
    if (a < Math.abs(c)) {
      a = c;
      s = p / 4;
    }
    else {
      //handle the 0/0 case:
      if (c === 0 && a === 0) {
        s = p / (2 * Math.PI) * Math.asin(1);
      }
      else {
        s = p / (2 * Math.PI) * Math.asin(c / a);
      }
    }
    return { a: a, c: c, p: p, s: s };
  }

  function elastic(opts, t, d) {
    return opts.a *
      Math.pow(2, 10 * (t -= 1)) *
      Math.sin( (t * d - opts.s) * (2 * Math.PI) / opts.p );
  }

  /**
   * Cubic easing out
   * @memberOf fabric.util.ease
   */
  function easeOutCubic(t, b, c, d) {
    return c * ((t = t / d - 1) * t * t + 1) + b;
  }

  /**
   * Cubic easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutCubic(t, b, c, d) {
    t /= d/2;
    if (t < 1) {
      return c / 2 * t * t * t + b;
    }
    return c / 2 * ((t -= 2) * t * t + 2) + b;
  }

  /**
   * Quartic easing in
   * @memberOf fabric.util.ease
   */
  function easeInQuart(t, b, c, d) {
    return c * (t /= d) * t * t * t + b;
  }

  /**
   * Quartic easing out
   * @memberOf fabric.util.ease
   */
  function easeOutQuart(t, b, c, d) {
    return -c * ((t = t / d - 1) * t * t * t - 1) + b;
  }

  /**
   * Quartic easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutQuart(t, b, c, d) {
    t /= d / 2;
    if (t < 1) {
      return c / 2 * t * t * t * t + b;
    }
    return -c / 2 * ((t -= 2) * t * t * t - 2) + b;
  }

  /**
   * Quintic easing in
   * @memberOf fabric.util.ease
   */
  function easeInQuint(t, b, c, d) {
    return c * (t /= d) * t * t * t * t + b;
  }

  /**
   * Quintic easing out
   * @memberOf fabric.util.ease
   */
  function easeOutQuint(t, b, c, d) {
    return c * ((t = t / d - 1) * t * t * t * t + 1) + b;
  }

  /**
   * Quintic easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutQuint(t, b, c, d) {
    t /= d / 2;
    if (t < 1) {
      return c / 2 * t * t * t * t * t + b;
    }
    return c / 2 * ((t -= 2) * t * t * t * t + 2) + b;
  }

  /**
   * Sinusoidal easing in
   * @memberOf fabric.util.ease
   */
  function easeInSine(t, b, c, d) {
    return -c * Math.cos(t / d * (Math.PI / 2)) + c + b;
  }

  /**
   * Sinusoidal easing out
   * @memberOf fabric.util.ease
   */
  function easeOutSine(t, b, c, d) {
    return c * Math.sin(t / d * (Math.PI / 2)) + b;
  }

  /**
   * Sinusoidal easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutSine(t, b, c, d) {
    return -c / 2 * (Math.cos(Math.PI * t / d) - 1) + b;
  }

  /**
   * Exponential easing in
   * @memberOf fabric.util.ease
   */
  function easeInExpo(t, b, c, d) {
    return (t === 0) ? b : c * Math.pow(2, 10 * (t / d - 1)) + b;
  }

  /**
   * Exponential easing out
   * @memberOf fabric.util.ease
   */
  function easeOutExpo(t, b, c, d) {
    return (t === d) ? b + c : c * (-Math.pow(2, -10 * t / d) + 1) + b;
  }

  /**
   * Exponential easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutExpo(t, b, c, d) {
    if (t === 0) {
      return b;
    }
    if (t === d) {
      return b + c;
    }
    t /= d / 2;
    if (t < 1) {
      return c / 2 * Math.pow(2, 10 * (t - 1)) + b;
    }
    return c / 2 * (-Math.pow(2, -10 * --t) + 2) + b;
  }

  /**
   * Circular easing in
   * @memberOf fabric.util.ease
   */
  function easeInCirc(t, b, c, d) {
    return -c * (Math.sqrt(1 - (t /= d) * t) - 1) + b;
  }

  /**
   * Circular easing out
   * @memberOf fabric.util.ease
   */
  function easeOutCirc(t, b, c, d) {
    return c * Math.sqrt(1 - (t = t / d - 1) * t) + b;
  }

  /**
   * Circular easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutCirc(t, b, c, d) {
    t /= d / 2;
    if (t < 1) {
      return -c / 2 * (Math.sqrt(1 - t * t) - 1) + b;
    }
    return c / 2 * (Math.sqrt(1 - (t -= 2) * t) + 1) + b;
  }

  /**
   * Elastic easing in
   * @memberOf fabric.util.ease
   */
  function easeInElastic(t, b, c, d) {
    var s = 1.70158, p = 0, a = c;
    if (t === 0) {
      return b;
    }
    t /= d;
    if (t === 1) {
      return b + c;
    }
    if (!p) {
      p = d * 0.3;
    }
    var opts = normalize(a, c, p, s);
    return -elastic(opts, t, d) + b;
  }

  /**
   * Elastic easing out
   * @memberOf fabric.util.ease
   */
  function easeOutElastic(t, b, c, d) {
    var s = 1.70158, p = 0, a = c;
    if (t === 0) {
      return b;
    }
    t /= d;
    if (t === 1) {
      return b + c;
    }
    if (!p) {
      p = d * 0.3;
    }
    var opts = normalize(a, c, p, s);
    return opts.a * Math.pow(2, -10 * t) * Math.sin((t * d - opts.s) * (2 * Math.PI) / opts.p ) + opts.c + b;
  }

  /**
   * Elastic easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutElastic(t, b, c, d) {
    var s = 1.70158, p = 0, a = c;
    if (t === 0) {
      return b;
    }
    t /= d / 2;
    if (t === 2) {
      return b + c;
    }
    if (!p) {
      p = d * (0.3 * 1.5);
    }
    var opts = normalize(a, c, p, s);
    if (t < 1) {
      return -0.5 * elastic(opts, t, d) + b;
    }
    return opts.a * Math.pow(2, -10 * (t -= 1)) *
      Math.sin((t * d - opts.s) * (2 * Math.PI) / opts.p ) * 0.5 + opts.c + b;
  }

  /**
   * Backwards easing in
   * @memberOf fabric.util.ease
   */
  function easeInBack(t, b, c, d, s) {
    if (s === undefined) {
      s = 1.70158;
    }
    return c * (t /= d) * t * ((s + 1) * t - s) + b;
  }

  /**
   * Backwards easing out
   * @memberOf fabric.util.ease
   */
  function easeOutBack(t, b, c, d, s) {
    if (s === undefined) {
      s = 1.70158;
    }
    return c * ((t = t / d - 1) * t * ((s + 1) * t + s) + 1) + b;
  }

  /**
   * Backwards easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutBack(t, b, c, d, s) {
    if (s === undefined) {
      s = 1.70158;
    }
    t /= d / 2;
    if (t < 1) {
      return c / 2 * (t * t * (((s *= (1.525)) + 1) * t - s)) + b;
    }
    return c / 2 * ((t -= 2) * t * (((s *= (1.525)) + 1) * t + s) + 2) + b;
  }

  /**
   * Bouncing easing in
   * @memberOf fabric.util.ease
   */
  function easeInBounce(t, b, c, d) {
    return c - easeOutBounce (d - t, 0, c, d) + b;
  }

  /**
   * Bouncing easing out
   * @memberOf fabric.util.ease
   */
  function easeOutBounce(t, b, c, d) {
    if ((t /= d) < (1 / 2.75)) {
      return c * (7.5625 * t * t) + b;
    }
    else if (t < (2/2.75)) {
      return c * (7.5625 * (t -= (1.5 / 2.75)) * t + 0.75) + b;
    }
    else if (t < (2.5/2.75)) {
      return c * (7.5625 * (t -= (2.25 / 2.75)) * t + 0.9375) + b;
    }
    else {
      return c * (7.5625 * (t -= (2.625 / 2.75)) * t + 0.984375) + b;
    }
  }

  /**
   * Bouncing easing in and out
   * @memberOf fabric.util.ease
   */
  function easeInOutBounce(t, b, c, d) {
    if (t < d / 2) {
      return easeInBounce (t * 2, 0, c, d) * 0.5 + b;
    }
    return easeOutBounce(t * 2 - d, 0, c, d) * 0.5 + c * 0.5 + b;
  }

  /**
   * Easing functions
   * See <a href="http://gizma.com/easing/">Easing Equations by Robert Penner</a>
   * @namespace fabric.util.ease
   */
  fabric.util.ease = {

    /**
     * Quadratic easing in
     * @memberOf fabric.util.ease
     */
    easeInQuad: function(t, b, c, d) {
      return c * (t /= d) * t + b;
    },

    /**
     * Quadratic easing out
     * @memberOf fabric.util.ease
     */
    easeOutQuad: function(t, b, c, d) {
      return -c * (t /= d) * (t - 2) + b;
    },

    /**
     * Quadratic easing in and out
     * @memberOf fabric.util.ease
     */
    easeInOutQuad: function(t, b, c, d) {
      t /= (d / 2);
      if (t < 1) {
        return c / 2 * t * t + b;
      }
      return -c / 2 * ((--t) * (t - 2) - 1) + b;
    },

    /**
     * Cubic easing in
     * @memberOf fabric.util.ease
     */
    easeInCubic: function(t, b, c, d) {
      return c * (t /= d) * t * t + b;
    },

    easeOutCubic: easeOutCubic,
    easeInOutCubic: easeInOutCubic,
    easeInQuart: easeInQuart,
    easeOutQuart: easeOutQuart,
    easeInOutQuart: easeInOutQuart,
    easeInQuint: easeInQuint,
    easeOutQuint: easeOutQuint,
    easeInOutQuint: easeInOutQuint,
    easeInSine: easeInSine,
    easeOutSine: easeOutSine,
    easeInOutSine: easeInOutSine,
    easeInExpo: easeInExpo,
    easeOutExpo: easeOutExpo,
    easeInOutExpo: easeInOutExpo,
    easeInCirc: easeInCirc,
    easeOutCirc: easeOutCirc,
    easeInOutCirc: easeInOutCirc,
    easeInElastic: easeInElastic,
    easeOutElastic: easeOutElastic,
    easeInOutElastic: easeInOutElastic,
    easeInBack: easeInBack,
    easeOutBack: easeOutBack,
    easeInOutBack: easeInOutBack,
    easeInBounce: easeInBounce,
    easeOutBounce: easeOutBounce,
    easeInOutBounce: easeInOutBounce
  };

}());


(function(global) {

  'use strict';

  /**
   * @name fabric
   * @namespace
   */

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      capitalize = fabric.util.string.capitalize,
      clone = fabric.util.object.clone,
      toFixed = fabric.util.toFixed,
      parseUnit = fabric.util.parseUnit,
      multiplyTransformMatrices = fabric.util.multiplyTransformMatrices,

      reAllowedSVGTagNames = /^(path|circle|polygon|polyline|ellipse|rect|line|image|text)$/i,
      reViewBoxTagNames = /^(symbol|image|marker|pattern|view|svg)$/i,
      reNotAllowedAncestors = /^(?:pattern|defs|symbol|metadata)$/i,
      reAllowedParents = /^(symbol|g|a|svg)$/i,

      attributesMap = {
        cx:                   'left',
        x:                    'left',
        r:                    'radius',
        cy:                   'top',
        y:                    'top',
        display:              'visible',
        visibility:           'visible',
        transform:            'transformMatrix',
        'fill-opacity':       'fillOpacity',
        'fill-rule':          'fillRule',
        'font-family':        'fontFamily',
        'font-size':          'fontSize',
        'font-style':         'fontStyle',
        'font-weight':        'fontWeight',
        'stroke-dasharray':   'strokeDashArray',
        'stroke-linecap':     'strokeLineCap',
        'stroke-linejoin':    'strokeLineJoin',
        'stroke-miterlimit':  'strokeMiterLimit',
        'stroke-opacity':     'strokeOpacity',
        'stroke-width':       'strokeWidth',
        'text-decoration':    'textDecoration',
        'text-anchor':        'originX'
      },

      colorAttributes = {
        stroke: 'strokeOpacity',
        fill:   'fillOpacity'
      };

  fabric.cssRules = { };
  fabric.gradientDefs = { };

  function normalizeAttr(attr) {
    // transform attribute names
    if (attr in attributesMap) {
      return attributesMap[attr];
    }
    return attr;
  }

  function normalizeValue(attr, value, parentAttributes, fontSize) {
    var isArray = Object.prototype.toString.call(value) === '[object Array]',
        parsed;

    if ((attr === 'fill' || attr === 'stroke') && value === 'none') {
      value = '';
    }
    else if (attr === 'strokeDashArray') {
      value = value.replace(/,/g, ' ').split(/\s+/).map(function(n) {
        return parseFloat(n);
      });
    }
    else if (attr === 'transformMatrix') {
      if (parentAttributes && parentAttributes.transformMatrix) {
        value = multiplyTransformMatrices(
          parentAttributes.transformMatrix, fabric.parseTransformAttribute(value));
      }
      else {
        value = fabric.parseTransformAttribute(value);
      }
    }
    else if (attr === 'visible') {
      value = (value === 'none' || value === 'hidden') ? false : true;
      // display=none on parent element always takes precedence over child element
      if (parentAttributes && parentAttributes.visible === false) {
        value = false;
      }
    }
    else if (attr === 'originX' /* text-anchor */) {
      value = value === 'start' ? 'left' : value === 'end' ? 'right' : 'center';
    }
    else {
      parsed = isArray ? value.map(parseUnit) : parseUnit(value, fontSize);
    }

    return (!isArray && isNaN(parsed) ? value : parsed);
  }

  /**
   * @private
   * @param {Object} attributes Array of attributes to parse
   */
  function _setStrokeFillOpacity(attributes) {
    for (var attr in colorAttributes) {

      if (typeof attributes[colorAttributes[attr]] === 'undefined' || attributes[attr] === '') {
        continue;
      }

      if (typeof attributes[attr] === 'undefined') {
        if (!fabric.Object.prototype[attr]) {
          continue;
        }
        attributes[attr] = fabric.Object.prototype[attr];
      }

      if (attributes[attr].indexOf('url(') === 0) {
        continue;
      }

      var color = new fabric.Color(attributes[attr]);
      attributes[attr] = color.setAlpha(toFixed(color.getAlpha() * attributes[colorAttributes[attr]], 2)).toRgba();
    }
    return attributes;
  }

  /**
   * Parses "transform" attribute, returning an array of values
   * @static
   * @function
   * @memberOf fabric
   * @param {String} attributeValue String containing attribute value
   * @return {Array} Array of 6 elements representing transformation matrix
   */
  fabric.parseTransformAttribute = (function() {
    function rotateMatrix(matrix, args) {
      var angle = args[0],
          x = (args.length === 3) ? args[1] : 0,
          y = (args.length === 3) ? args[2] : 0;

      matrix[0] = Math.cos(angle);
      matrix[1] = Math.sin(angle);
      matrix[2] = -Math.sin(angle);
      matrix[3] = Math.cos(angle);
      matrix[4] = x - (matrix[0] * x + matrix[2] * y);
      matrix[5] = y - (matrix[1] * x + matrix[3] * y);
    }

    function scaleMatrix(matrix, args) {
      var multiplierX = args[0],
          multiplierY = (args.length === 2) ? args[1] : args[0];

      matrix[0] = multiplierX;
      matrix[3] = multiplierY;
    }

    function skewXMatrix(matrix, args) {
      matrix[2] = Math.tan(fabric.util.degreesToRadians(args[0]));
    }

    function skewYMatrix(matrix, args) {
      matrix[1] = Math.tan(fabric.util.degreesToRadians(args[0]));
    }

    function translateMatrix(matrix, args) {
      matrix[4] = args[0];
      if (args.length === 2) {
        matrix[5] = args[1];
      }
    }

    // identity matrix
    var iMatrix = [
          1, // a
          0, // b
          0, // c
          1, // d
          0, // e
          0  // f
        ],

        // == begin transform regexp
        number = fabric.reNum,

        commaWsp = '(?:\\s+,?\\s*|,\\s*)',

        skewX = '(?:(skewX)\\s*\\(\\s*(' + number + ')\\s*\\))',

        skewY = '(?:(skewY)\\s*\\(\\s*(' + number + ')\\s*\\))',

        rotate = '(?:(rotate)\\s*\\(\\s*(' + number + ')(?:' +
                    commaWsp + '(' + number + ')' +
                    commaWsp + '(' + number + '))?\\s*\\))',

        scale = '(?:(scale)\\s*\\(\\s*(' + number + ')(?:' +
                    commaWsp + '(' + number + '))?\\s*\\))',

        translate = '(?:(translate)\\s*\\(\\s*(' + number + ')(?:' +
                    commaWsp + '(' + number + '))?\\s*\\))',

        matrix = '(?:(matrix)\\s*\\(\\s*' +
                  '(' + number + ')' + commaWsp +
                  '(' + number + ')' + commaWsp +
                  '(' + number + ')' + commaWsp +
                  '(' + number + ')' + commaWsp +
                  '(' + number + ')' + commaWsp +
                  '(' + number + ')' +
                  '\\s*\\))',

        transform = '(?:' +
                    matrix + '|' +
                    translate + '|' +
                    scale + '|' +
                    rotate + '|' +
                    skewX + '|' +
                    skewY +
                    ')',

        transforms = '(?:' + transform + '(?:' + commaWsp + '*' + transform + ')*' + ')',

        transformList = '^\\s*(?:' + transforms + '?)\\s*$',

        // http://www.w3.org/TR/SVG/coords.html#TransformAttribute
        reTransformList = new RegExp(transformList),
        // == end transform regexp

        reTransform = new RegExp(transform, 'g');

    return function(attributeValue) {

      // start with identity matrix
      var matrix = iMatrix.concat(),
          matrices = [ ];

      // return if no argument was given or
      // an argument does not match transform attribute regexp
      if (!attributeValue || (attributeValue && !reTransformList.test(attributeValue))) {
        return matrix;
      }

      attributeValue.replace(reTransform, function(match) {

        var m = new RegExp(transform).exec(match).filter(function (match) {
              return (match !== '' && match != null);
            }),
            operation = m[1],
            args = m.slice(2).map(parseFloat);

        switch (operation) {
          case 'translate':
            translateMatrix(matrix, args);
            break;
          case 'rotate':
            args[0] = fabric.util.degreesToRadians(args[0]);
            rotateMatrix(matrix, args);
            break;
          case 'scale':
            scaleMatrix(matrix, args);
            break;
          case 'skewX':
            skewXMatrix(matrix, args);
            break;
          case 'skewY':
            skewYMatrix(matrix, args);
            break;
          case 'matrix':
            matrix = args;
            break;
        }

        // snapshot current matrix into matrices array
        matrices.push(matrix.concat());
        // reset
        matrix = iMatrix.concat();
      });

      var combinedMatrix = matrices[0];
      while (matrices.length > 1) {
        matrices.shift();
        combinedMatrix = fabric.util.multiplyTransformMatrices(combinedMatrix, matrices[0]);
      }
      return combinedMatrix;
    };
  })();

  /**
   * @private
   */
  function parseStyleString(style, oStyle) {
    var attr, value;
    style.replace(/;\s*$/, '').split(';').forEach(function (chunk) {
      var pair = chunk.split(':');

      attr = normalizeAttr(pair[0].trim().toLowerCase());
      value = normalizeValue(attr, pair[1].trim());

      oStyle[attr] = value;
    });
  }

  /**
   * @private
   */
  function parseStyleObject(style, oStyle) {
    var attr, value;
    for (var prop in style) {
      if (typeof style[prop] === 'undefined') {
        continue;
      }

      attr = normalizeAttr(prop.toLowerCase());
      value = normalizeValue(attr, style[prop]);

      oStyle[attr] = value;
    }
  }

  /**
   * @private
   */
  function getGlobalStylesForElement(element, svgUid) {
    var styles = { };
    for (var rule in fabric.cssRules[svgUid]) {
      if (elementMatchesRule(element, rule.split(' '))) {
        for (var property in fabric.cssRules[svgUid][rule]) {
          styles[property] = fabric.cssRules[svgUid][rule][property];
        }
      }
    }
    return styles;
  }

  /**
   * @private
   */
  function elementMatchesRule(element, selectors) {
    var firstMatching, parentMatching = true;
    //start from rightmost selector.
    firstMatching = selectorMatches(element, selectors.pop());
    if (firstMatching && selectors.length) {
      parentMatching = doesSomeParentMatch(element, selectors);
    }
    return firstMatching && parentMatching && (selectors.length === 0);
  }

  function doesSomeParentMatch(element, selectors) {
    var selector, parentMatching = true;
    while (element.parentNode && element.parentNode.nodeType === 1 && selectors.length) {
      if (parentMatching) {
        selector = selectors.pop();
      }
      element = element.parentNode;
      parentMatching = selectorMatches(element, selector);
    }
    return selectors.length === 0;
  }

  /**
   * @private
   */
  function selectorMatches(element, selector) {
    var nodeName = element.nodeName,
        classNames = element.getAttribute('class'),
        id = element.getAttribute('id'), matcher;
    // i check if a selector matches slicing away part from it.
    // if i get empty string i should match
    matcher = new RegExp('^' + nodeName, 'i');
    selector = selector.replace(matcher, '');
    if (id && selector.length) {
      matcher = new RegExp('#' + id + '(?![a-zA-Z\\-]+)', 'i');
      selector = selector.replace(matcher, '');
    }
    if (classNames && selector.length) {
      classNames = classNames.split(' ');
      for (var i = classNames.length; i--;) {
        matcher = new RegExp('\\.' + classNames[i] + '(?![a-zA-Z\\-]+)', 'i');
        selector = selector.replace(matcher, '');
      }
    }
    return selector.length === 0;
  }

  /**
   * @private
   * to support IE8 missing getElementById on SVGdocument
   */
  function elementById(doc, id) {
    var el;
    doc.getElementById && (el = doc.getElementById(id));
    if (el) {
      return el;
    }
    var node, i, nodelist = doc.getElementsByTagName('*');
    for (i = 0; i < nodelist.length; i++) {
      node = nodelist[i];
      if (id === node.getAttribute('id')) {
        return node;
      }
    }
  }

  /**
   * @private
   */
  function parseUseDirectives(doc) {
    var nodelist = doc.getElementsByTagName('use'), i = 0;
    while (nodelist.length && i < nodelist.length) {
      var el = nodelist[i],
          xlink = el.getAttribute('xlink:href').substr(1),
          x = el.getAttribute('x') || 0,
          y = el.getAttribute('y') || 0,
          el2 = elementById(doc, xlink).cloneNode(true),
          currentTrans = (el2.getAttribute('transform') || '') + ' translate(' + x + ', ' + y + ')',
          parentNode, oldLength = nodelist.length, attr, j, attrs, l;

      applyViewboxTransform(el2);
      if (/^svg$/i.test(el2.nodeName)) {
        var el3 = el2.ownerDocument.createElement('g');
        for (j = 0, attrs = el2.attributes, l = attrs.length; j < l; j++) {
          attr = attrs.item(j);
          el3.setAttribute(attr.nodeName, attr.nodeValue);
        }
        while (el2.firstChild != null) {
          el3.appendChild(el2.firstChild);
        }
        el2 = el3;
      }

      for (j = 0, attrs = el.attributes, l = attrs.length; j < l; j++) {
        attr = attrs.item(j);
        if (attr.nodeName === 'x' || attr.nodeName === 'y' || attr.nodeName === 'xlink:href') {
          continue;
        }

        if (attr.nodeName === 'transform') {
          currentTrans = attr.nodeValue + ' ' + currentTrans;
        }
        else {
          el2.setAttribute(attr.nodeName, attr.nodeValue);
        }
      }

      el2.setAttribute('transform', currentTrans);
      el2.setAttribute('instantiated_by_use', '1');
      el2.removeAttribute('id');
      parentNode = el.parentNode;
      parentNode.replaceChild(el2, el);
      // some browsers do not shorten nodelist after replaceChild (IE8)
      if (nodelist.length === oldLength) {
        i++;
      }
    }
  }

  // http://www.w3.org/TR/SVG/coords.html#ViewBoxAttribute
  // matches, e.g.: +14.56e-12, etc.
  var reViewBoxAttrValue = new RegExp(
    '^' +
    '\\s*(' + fabric.reNum + '+)\\s*,?' +
    '\\s*(' + fabric.reNum + '+)\\s*,?' +
    '\\s*(' + fabric.reNum + '+)\\s*,?' +
    '\\s*(' + fabric.reNum + '+)\\s*' +
    '$'
  );

  /**
   * Add a <g> element that envelop all child elements and makes the viewbox transformMatrix descend on all elements
   */
  function applyViewboxTransform(element) {

    var viewBoxAttr = element.getAttribute('viewBox'),
        scaleX = 1,
        scaleY = 1,
        minX = 0,
        minY = 0,
        viewBoxWidth, viewBoxHeight, matrix, el,
        widthAttr = element.getAttribute('width'),
        heightAttr = element.getAttribute('height'),
        x = element.getAttribute('x') || 0,
        y = element.getAttribute('y') || 0,
        preserveAspectRatio = element.getAttribute('preserveAspectRatio') || '',
        missingViewBox = (!viewBoxAttr || !reViewBoxTagNames.test(element.tagName)
                           || !(viewBoxAttr = viewBoxAttr.match(reViewBoxAttrValue))),
        missingDimAttr = (!widthAttr || !heightAttr || widthAttr === '100%' || heightAttr === '100%'),
        toBeParsed = missingViewBox && missingDimAttr,
        parsedDim = { }, translateMatrix = '';

    parsedDim.width = 0;
    parsedDim.height = 0;
    parsedDim.toBeParsed = toBeParsed;

    if (toBeParsed) {
      return parsedDim;
    }

    if (missingViewBox) {
      parsedDim.width = parseUnit(widthAttr);
      parsedDim.height = parseUnit(heightAttr);
      return parsedDim;
    }

    minX = -parseFloat(viewBoxAttr[1]),
    minY = -parseFloat(viewBoxAttr[2]),
    viewBoxWidth = parseFloat(viewBoxAttr[3]),
    viewBoxHeight = parseFloat(viewBoxAttr[4]);

    if (!missingDimAttr) {
      parsedDim.width = parseUnit(widthAttr);
      parsedDim.height = parseUnit(heightAttr);
      scaleX = parsedDim.width / viewBoxWidth;
      scaleY = parsedDim.height / viewBoxHeight;
    }
    else {
      parsedDim.width = viewBoxWidth;
      parsedDim.height = viewBoxHeight;
    }

    // default is to preserve aspect ratio
    preserveAspectRatio = fabric.util.parsePreserveAspectRatioAttribute(preserveAspectRatio);
    if (preserveAspectRatio.alignX !== 'none') {
      //translate all container for the effect of Mid, Min, Max
      scaleY = scaleX = (scaleX > scaleY ? scaleY : scaleX);
    }

    if (scaleX === 1 && scaleY === 1 && minX === 0 && minY === 0 && x === 0 && y === 0) {
      return parsedDim;
    }

    if (x || y) {
      translateMatrix = ' translate(' + parseUnit(x) + ' ' + parseUnit(y) + ') ';
    }

    matrix = translateMatrix + ' matrix(' + scaleX +
                  ' 0' +
                  ' 0 ' +
                  scaleY + ' ' +
                  (minX * scaleX) + ' ' +
                  (minY * scaleY) + ') ';

    if (element.tagName === 'svg') {
      el = element.ownerDocument.createElement('g');
      while (element.firstChild != null) {
        el.appendChild(element.firstChild);
      }
      element.appendChild(el);
    }
    else {
      el = element;
      matrix = el.getAttribute('transform') + matrix;
    }

    el.setAttribute('transform', matrix);
    return parsedDim;
  }

  /**
   * Parses an SVG document, converts it to an array of corresponding fabric.* instances and passes them to a callback
   * @static
   * @function
   * @memberOf fabric
   * @param {SVGDocument} doc SVG document to parse
   * @param {Function} callback Callback to call when parsing is finished; It's being passed an array of elements (parsed from a document).
   * @param {Function} [reviver] Method for further parsing of SVG elements, called after each fabric object created.
   */
  fabric.parseSVGDocument = (function() {

    function hasAncestorWithNodeName(element, nodeName) {
      while (element && (element = element.parentNode)) {
        if (nodeName.test(element.nodeName) && !element.getAttribute('instantiated_by_use')) {
          return true;
        }
      }
      return false;
    }

    return function(doc, callback, reviver) {
      if (!doc) {
        return;
      }

      parseUseDirectives(doc);

      var startTime = new Date(),
          svgUid =  fabric.Object.__uid++,
          options = applyViewboxTransform(doc),
          descendants = fabric.util.toArray(doc.getElementsByTagName('*'));

      options.svgUid = svgUid;

      if (descendants.length === 0 && fabric.isLikelyNode) {
        // we're likely in node, where "o3-xml" library fails to gEBTN("*")
        // https://github.com/ajaxorg/node-o3-xml/issues/21
        descendants = doc.selectNodes('//*[name(.)!="svg"]');
        var arr = [ ];
        for (var i = 0, len = descendants.length; i < len; i++) {
          arr[i] = descendants[i];
        }
        descendants = arr;
      }

      var elements = descendants.filter(function(el) {
        applyViewboxTransform(el);
        return reAllowedSVGTagNames.test(el.tagName) &&
              !hasAncestorWithNodeName(el, reNotAllowedAncestors); // http://www.w3.org/TR/SVG/struct.html#DefsElement
      });

      if (!elements || (elements && !elements.length)) {
        callback && callback([], {});
        return;
      }

      fabric.gradientDefs[svgUid] = fabric.getGradientDefs(doc);
      fabric.cssRules[svgUid] = fabric.getCSSRules(doc);
      // Precedence of rules:   style > class > attribute
      fabric.parseElements(elements, function(instances) {
        fabric.documentParsingTime = new Date() - startTime;
        if (callback) {
          callback(instances, options);
        }
      }, clone(options), reviver);
    };
  })();

  /**
   * Used for caching SVG documents (loaded via `fabric.Canvas#loadSVGFromURL`)
   * @namespace
   */
  var svgCache = {

    /**
     * @param {String} name
     * @param {Function} callback
     */
    has: function (name, callback) {
      callback(false);
    },

    get: function () {
      /* NOOP */
    },

    set: function () {
      /* NOOP */
    }
  };

  /**
   * @private
   */
  function _enlivenCachedObject(cachedObject) {

    var objects = cachedObject.objects,
        options = cachedObject.options;

    objects = objects.map(function (o) {
      return fabric[capitalize(o.type)].fromObject(o);
    });

    return ({ objects: objects, options: options });
  }

  /**
   * @private
   */
  function _createSVGPattern(markup, canvas, property) {
    if (canvas[property] && canvas[property].toSVG) {
      markup.push(
        '\t<pattern x="0" y="0" id="', property, 'Pattern" ',
          'width="', canvas[property].source.width,
          '" height="', canvas[property].source.height,
          '" patternUnits="userSpaceOnUse">\n',
        '\t\t<image x="0" y="0" ',
        'width="', canvas[property].source.width,
        '" height="', canvas[property].source.height,
        '" xlink:href="', canvas[property].source.src,
        '"></image>\n\t</pattern>\n'
      );
    }
  }

  var reFontDeclaration = new RegExp(
    '(normal|italic)?\\s*(normal|small-caps)?\\s*' +
    '(normal|bold|bolder|lighter|100|200|300|400|500|600|700|800|900)?\\s*(' +
      fabric.reNum +
    '(?:px|cm|mm|em|pt|pc|in)*)(?:\\/(normal|' + fabric.reNum + '))?\\s+(.*)');

  extend(fabric, {
    /**
     * Parses a short font declaration, building adding its properties to a style object
     * @static
     * @function
     * @memberOf fabric
     * @param {String} value font declaration
     * @param {Object} oStyle definition
     */
    parseFontDeclaration: function(value, oStyle) {
      var match = value.match(reFontDeclaration);

      if (!match) {
        return;
      }
      var fontStyle = match[1],
          // font variant is not used
          // fontVariant = match[2],
          fontWeight = match[3],
          fontSize = match[4],
          lineHeight = match[5],
          fontFamily = match[6];

      if (fontStyle) {
        oStyle.fontStyle = fontStyle;
      }
      if (fontWeight) {
        oStyle.fontWeight = isNaN(parseFloat(fontWeight)) ? fontWeight : parseFloat(fontWeight);
      }
      if (fontSize) {
        oStyle.fontSize = parseUnit(fontSize);
      }
      if (fontFamily) {
        oStyle.fontFamily = fontFamily;
      }
      if (lineHeight) {
        oStyle.lineHeight = lineHeight === 'normal' ? 1 : lineHeight;
      }
    },

    /**
     * Parses an SVG document, returning all of the gradient declarations found in it
     * @static
     * @function
     * @memberOf fabric
     * @param {SVGDocument} doc SVG document to parse
     * @return {Object} Gradient definitions; key corresponds to element id, value -- to gradient definition element
     */
    getGradientDefs: function(doc) {
      var linearGradientEls = doc.getElementsByTagName('linearGradient'),
          radialGradientEls = doc.getElementsByTagName('radialGradient'),
          el, i, j = 0, id, xlink, elList = [ ],
          gradientDefs = { }, idsToXlinkMap = { };

      elList.length = linearGradientEls.length + radialGradientEls.length;
      i = linearGradientEls.length;
      while (i--) {
        elList[j++] = linearGradientEls[i];
      }
      i = radialGradientEls.length;
      while (i--) {
        elList[j++] = radialGradientEls[i];
      }

      while (j--) {
        el = elList[j];
        xlink = el.getAttribute('xlink:href');
        id = el.getAttribute('id');
        if (xlink) {
          idsToXlinkMap[id] = xlink.substr(1);
        }
        gradientDefs[id] = el;
      }

      for (id in idsToXlinkMap) {
        var el2 = gradientDefs[idsToXlinkMap[id]].cloneNode(true);
        el = gradientDefs[id];
        while (el2.firstChild) {
          el.appendChild(el2.firstChild);
        }
      }
      return gradientDefs;
    },

    /**
     * Returns an object of attributes' name/value, given element and an array of attribute names;
     * Parses parent "g" nodes recursively upwards.
     * @static
     * @memberOf fabric
     * @param {DOMElement} element Element to parse
     * @param {Array} attributes Array of attributes to parse
     * @return {Object} object containing parsed attributes' names/values
     */
    parseAttributes: function(element, attributes, svgUid) {

      if (!element) {
        return;
      }

      var value,
          parentAttributes = { },
          fontSize;

      if (typeof svgUid === 'undefined') {
        svgUid = element.getAttribute('svgUid');
      }
      // if there's a parent container (`g` or `a` or `symbol` node), parse its attributes recursively upwards
      if (element.parentNode && reAllowedParents.test(element.parentNode.nodeName)) {
        parentAttributes = fabric.parseAttributes(element.parentNode, attributes, svgUid);
      }
      fontSize = (parentAttributes && parentAttributes.fontSize ) ||
                 element.getAttribute('font-size') || fabric.Text.DEFAULT_SVG_FONT_SIZE;

      var ownAttributes = attributes.reduce(function(memo, attr) {
        value = element.getAttribute(attr);
        if (value) {
          attr = normalizeAttr(attr);
          value = normalizeValue(attr, value, parentAttributes, fontSize);

          memo[attr] = value;
        }
        return memo;
      }, { });

      // add values parsed from style, which take precedence over attributes
      // (see: http://www.w3.org/TR/SVG/styling.html#UsingPresentationAttributes)
      ownAttributes = extend(ownAttributes,
        extend(getGlobalStylesForElement(element, svgUid), fabric.parseStyleAttribute(element)));
      if (ownAttributes.font) {
        fabric.parseFontDeclaration(ownAttributes.font, ownAttributes);
      }
      return _setStrokeFillOpacity(extend(parentAttributes, ownAttributes));
    },

    /**
     * Transforms an array of svg elements to corresponding fabric.* instances
     * @static
     * @memberOf fabric
     * @param {Array} elements Array of elements to parse
     * @param {Function} callback Being passed an array of fabric instances (transformed from SVG elements)
     * @param {Object} [options] Options object
     * @param {Function} [reviver] Method for further parsing of SVG elements, called after each fabric object created.
     */
    parseElements: function(elements, callback, options, reviver) {
      new fabric.ElementsParser(elements, callback, options, reviver).parse();
    },

    /**
     * Parses "style" attribute, retuning an object with values
     * @static
     * @memberOf fabric
     * @param {SVGElement} element Element to parse
     * @return {Object} Objects with values parsed from style attribute of an element
     */
    parseStyleAttribute: function(element) {
      var oStyle = { },
          style = element.getAttribute('style');

      if (!style) {
        return oStyle;
      }

      if (typeof style === 'string') {
        parseStyleString(style, oStyle);
      }
      else {
        parseStyleObject(style, oStyle);
      }

      return oStyle;
    },

    /**
     * Parses "points" attribute, returning an array of values
     * @static
     * @memberOf fabric
     * @param {String} points points attribute string
     * @return {Array} array of points
     */
    parsePointsAttribute: function(points) {

      // points attribute is required and must not be empty
      if (!points) {
        return null;
      }

      // replace commas with whitespace and remove bookending whitespace
      points = points.replace(/,/g, ' ').trim();

      points = points.split(/\s+/);
      var parsedPoints = [ ], i, len;

      i = 0;
      len = points.length;
      for (; i < len; i+=2) {
        parsedPoints.push({
          x: parseFloat(points[i]),
          y: parseFloat(points[i + 1])
        });
      }

      // odd number of points is an error
      // if (parsedPoints.length % 2 !== 0) {
      //   return null;
      // }

      return parsedPoints;
    },

    /**
     * Returns CSS rules for a given SVG document
     * @static
     * @function
     * @memberOf fabric
     * @param {SVGDocument} doc SVG document to parse
     * @return {Object} CSS rules of this document
     */
    getCSSRules: function(doc) {
      var styles = doc.getElementsByTagName('style'),
          allRules = { }, rules;

      // very crude parsing of style contents
      for (var i = 0, len = styles.length; i < len; i++) {
        // IE9 doesn't support textContent, but provides text instead.
        var styleContents = styles[i].textContent || styles[i].text;

        // remove comments
        styleContents = styleContents.replace(/\/\*[\s\S]*?\*\//g, '');
        if (styleContents.trim() === '') {
          continue;
        }
        rules = styleContents.match(/[^{]*\{[\s\S]*?\}/g);
        rules = rules.map(function(rule) { return rule.trim(); });

        rules.forEach(function(rule) {

          var match = rule.match(/([\s\S]*?)\s*\{([^}]*)\}/),
          ruleObj = { }, declaration = match[2].trim(),
          propertyValuePairs = declaration.replace(/;$/, '').split(/\s*;\s*/);

          for (var i = 0, len = propertyValuePairs.length; i < len; i++) {
            var pair = propertyValuePairs[i].split(/\s*:\s*/),
                property = normalizeAttr(pair[0]),
                value = normalizeValue(property, pair[1], pair[0]);
            ruleObj[property] = value;
          }
          rule = match[1];
          rule.split(',').forEach(function(_rule) {
            _rule = _rule.replace(/^svg/i, '').trim();
            if (_rule === '') {
              return;
            }
            allRules[_rule] = fabric.util.object.clone(ruleObj);
          });
        });
      }
      return allRules;
    },

    /**
     * Takes url corresponding to an SVG document, and parses it into a set of fabric objects. Note that SVG is fetched via XMLHttpRequest, so it needs to conform to SOP (Same Origin Policy)
     * @memberOf fabric
     * @param {String} url
     * @param {Function} callback
     * @param {Function} [reviver] Method for further parsing of SVG elements, called after each fabric object created.
     */
    loadSVGFromURL: function(url, callback, reviver) {

      url = url.replace(/^\n\s*/, '').trim();
      svgCache.has(url, function (hasUrl) {
        if (hasUrl) {
          svgCache.get(url, function (value) {
            var enlivedRecord = _enlivenCachedObject(value);
            callback(enlivedRecord.objects, enlivedRecord.options);
          });
        }
        else {
          new fabric.util.request(url, {
            method: 'get',
            onComplete: onComplete
          });
        }
      });

      function onComplete(r) {

        var xml = r.responseXML;
        if (xml && !xml.documentElement && fabric.window.ActiveXObject && r.responseText) {
          xml = new ActiveXObject('Microsoft.XMLDOM');
          xml.async = 'false';
          //IE chokes on DOCTYPE
          xml.loadXML(r.responseText.replace(/<!DOCTYPE[\s\S]*?(\[[\s\S]*\])*?>/i, ''));
        }
        if (!xml || !xml.documentElement) {
          callback && callback(null);
        }

        fabric.parseSVGDocument(xml.documentElement, function (results, options) {
          svgCache.set(url, {
            objects: fabric.util.array.invoke(results, 'toObject'),
            options: options
          });
          callback && callback(results, options);
        }, reviver);
      }
    },

    /**
     * Takes string corresponding to an SVG document, and parses it into a set of fabric objects
     * @memberOf fabric
     * @param {String} string
     * @param {Function} callback
     * @param {Function} [reviver] Method for further parsing of SVG elements, called after each fabric object created.
     */
    loadSVGFromString: function(string, callback, reviver) {
      string = string.trim();
      var doc;
      if (typeof DOMParser !== 'undefined') {
        var parser = new DOMParser();
        if (parser && parser.parseFromString) {
          doc = parser.parseFromString(string, 'text/xml');
        }
      }
      else if (fabric.window.ActiveXObject) {
        doc = new ActiveXObject('Microsoft.XMLDOM');
        doc.async = 'false';
        // IE chokes on DOCTYPE
        doc.loadXML(string.replace(/<!DOCTYPE[\s\S]*?(\[[\s\S]*\])*?>/i, ''));
      }

      fabric.parseSVGDocument(doc.documentElement, function (results, options) {
        callback(results, options);
      }, reviver);
    },

    /**
     * Creates markup containing SVG font faces,
     * font URLs for font faces must be collected by developers
     * and are not extracted from the DOM by fabricjs
     * @param {Array} objects Array of fabric objects
     * @return {String}
     */
    createSVGFontFacesMarkup: function(objects) {
      var markup = '', fontList = { }, obj, fontFamily,
          style, row, rowIndex, _char, charIndex,
          fontPaths = fabric.fontPaths;

      for (var i = 0, len = objects.length; i < len; i++) {
        obj = objects[i];
        fontFamily = obj.fontFamily;
        if (obj.type.indexOf('text') === -1 || fontList[fontFamily] || !fontPaths[fontFamily]) {
          continue;
        }
        fontList[fontFamily] = true;
        if (!obj.styles) {
          continue;
        }
        style = obj.styles;
        for (rowIndex in style) {
          row = style[rowIndex];
          for (charIndex in row) {
            _char = row[charIndex];
            fontFamily = _char.fontFamily;
            if (!fontList[fontFamily] && fontPaths[fontFamily]) {
              fontList[fontFamily] = true;
            }
          }
        }
      }

      for (var j in fontList) {
        markup += [
          //jscs:disable validateIndentation
          '\t\t@font-face {\n',
            '\t\t\tfont-family: \'', j, '\';\n',
            '\t\t\tsrc: url(\'', fontPaths[j], '\');\n',
          '\t\t}\n'
          //jscs:enable validateIndentation
        ].join('');
      }

      if (markup) {
        markup = [
          //jscs:disable validateIndentation
          '\t<style type="text/css">',
            '<![CDATA[\n',
              markup,
            ']]>',
          '</style>\n'
          //jscs:enable validateIndentation
        ].join('');
      }

      return markup;
    },

    /**
     * Creates markup containing SVG referenced elements like patterns, gradients etc.
     * @param {fabric.Canvas} canvas instance of fabric.Canvas
     * @return {String}
     */
    createSVGRefElementsMarkup: function(canvas) {
      var markup = [ ];

      _createSVGPattern(markup, canvas, 'backgroundColor');
      _createSVGPattern(markup, canvas, 'overlayColor');

      return markup.join('');
    }
  });

})(typeof exports !== 'undefined' ? exports : this);


fabric.ElementsParser = function(elements, callback, options, reviver) {
  this.elements = elements;
  this.callback = callback;
  this.options = options;
  this.reviver = reviver;
  this.svgUid = (options && options.svgUid) || 0;
};

fabric.ElementsParser.prototype.parse = function() {
  this.instances = new Array(this.elements.length);
  this.numElements = this.elements.length;

  this.createObjects();
};

fabric.ElementsParser.prototype.createObjects = function() {
  for (var i = 0, len = this.elements.length; i < len; i++) {
    this.elements[i].setAttribute('svgUid', this.svgUid);
    (function(_this, i) {
      setTimeout(function() {
        _this.createObject(_this.elements[i], i);
      }, 0);
    })(this, i);
  }
};

fabric.ElementsParser.prototype.createObject = function(el, index) {
  var klass = fabric[fabric.util.string.capitalize(el.tagName)];
  if (klass && klass.fromElement) {
    try {
      this._createObject(klass, el, index);
    }
    catch (err) {
      fabric.log(err);
    }
  }
  else {
    this.checkIfDone();
  }
};

fabric.ElementsParser.prototype._createObject = function(klass, el, index) {
  if (klass.async) {
    klass.fromElement(el, this.createCallback(index, el), this.options);
  }
  else {
    var obj = klass.fromElement(el, this.options);
    this.resolveGradient(obj, 'fill');
    this.resolveGradient(obj, 'stroke');
    this.reviver && this.reviver(el, obj);
    this.instances[index] = obj;
    this.checkIfDone();
  }
};

fabric.ElementsParser.prototype.createCallback = function(index, el) {
  var _this = this;
  return function(obj) {
    _this.resolveGradient(obj, 'fill');
    _this.resolveGradient(obj, 'stroke');
    _this.reviver && _this.reviver(el, obj);
    _this.instances[index] = obj;
    _this.checkIfDone();
  };
};

fabric.ElementsParser.prototype.resolveGradient = function(obj, property) {

  var instanceFillValue = obj.get(property);
  if (!(/^url\(/).test(instanceFillValue)) {
    return;
  }
  var gradientId = instanceFillValue.slice(5, instanceFillValue.length - 1);
  if (fabric.gradientDefs[this.svgUid][gradientId]) {
    obj.set(property,
      fabric.Gradient.fromElement(fabric.gradientDefs[this.svgUid][gradientId], obj));
  }
};

fabric.ElementsParser.prototype.checkIfDone = function() {
  if (--this.numElements === 0) {
    this.instances = this.instances.filter(function(el) {
      return el != null;
    });
    this.callback(this.instances);
  }
};


(function(global) {

  'use strict';

  /* Adaptation of work of Kevin Lindsey (kevin@kevlindev.com) */

  var fabric = global.fabric || (global.fabric = { });

  if (fabric.Point) {
    fabric.warn('fabric.Point is already defined');
    return;
  }

  fabric.Point = Point;

  /**
   * Point class
   * @class fabric.Point
   * @memberOf fabric
   * @constructor
   * @param {Number} x
   * @param {Number} y
   * @return {fabric.Point} thisArg
   */
  function Point(x, y) {
    this.x = x;
    this.y = y;
  }

  Point.prototype = /** @lends fabric.Point.prototype */ {

    constructor: Point,

    /**
     * Adds another point to this one and returns another one
     * @param {fabric.Point} that
     * @return {fabric.Point} new Point instance with added values
     */
    add: function (that) {
      return new Point(this.x + that.x, this.y + that.y);
    },

    /**
     * Adds another point to this one
     * @param {fabric.Point} that
     * @return {fabric.Point} thisArg
     */
    addEquals: function (that) {
      this.x += that.x;
      this.y += that.y;
      return this;
    },

    /**
     * Adds value to this point and returns a new one
     * @param {Number} scalar
     * @return {fabric.Point} new Point with added value
     */
    scalarAdd: function (scalar) {
      return new Point(this.x + scalar, this.y + scalar);
    },

    /**
     * Adds value to this point
     * @param {Number} scalar
     * @return {fabric.Point} thisArg
     */
    scalarAddEquals: function (scalar) {
      this.x += scalar;
      this.y += scalar;
      return this;
    },

    /**
     * Subtracts another point from this point and returns a new one
     * @param {fabric.Point} that
     * @return {fabric.Point} new Point object with subtracted values
     */
    subtract: function (that) {
      return new Point(this.x - that.x, this.y - that.y);
    },

    /**
     * Subtracts another point from this point
     * @param {fabric.Point} that
     * @return {fabric.Point} thisArg
     */
    subtractEquals: function (that) {
      this.x -= that.x;
      this.y -= that.y;
      return this;
    },

    /**
     * Subtracts value from this point and returns a new one
     * @param {Number} scalar
     * @return {fabric.Point}
     */
    scalarSubtract: function (scalar) {
      return new Point(this.x - scalar, this.y - scalar);
    },

    /**
     * Subtracts value from this point
     * @param {Number} scalar
     * @return {fabric.Point} thisArg
     */
    scalarSubtractEquals: function (scalar) {
      this.x -= scalar;
      this.y -= scalar;
      return this;
    },

    /**
     * Miltiplies this point by a value and returns a new one
     * @param {Number} scalar
     * @return {fabric.Point}
     */
    multiply: function (scalar) {
      return new Point(this.x * scalar, this.y * scalar);
    },

    /**
     * Miltiplies this point by a value
     * @param {Number} scalar
     * @return {fabric.Point} thisArg
     */
    multiplyEquals: function (scalar) {
      this.x *= scalar;
      this.y *= scalar;
      return this;
    },

    /**
     * Divides this point by a value and returns a new one
     * @param {Number} scalar
     * @return {fabric.Point}
     */
    divide: function (scalar) {
      return new Point(this.x / scalar, this.y / scalar);
    },

    /**
     * Divides this point by a value
     * @param {Number} scalar
     * @return {fabric.Point} thisArg
     */
    divideEquals: function (scalar) {
      this.x /= scalar;
      this.y /= scalar;
      return this;
    },

    /**
     * Returns true if this point is equal to another one
     * @param {fabric.Point} that
     * @return {Boolean}
     */
    eq: function (that) {
      return (this.x === that.x && this.y === that.y);
    },

    /**
     * Returns true if this point is less than another one
     * @param {fabric.Point} that
     * @return {Boolean}
     */
    lt: function (that) {
      return (this.x < that.x && this.y < that.y);
    },

    /**
     * Returns true if this point is less than or equal to another one
     * @param {fabric.Point} that
     * @return {Boolean}
     */
    lte: function (that) {
      return (this.x <= that.x && this.y <= that.y);
    },

    /**

     * Returns true if this point is greater another one
     * @param {fabric.Point} that
     * @return {Boolean}
     */
    gt: function (that) {
      return (this.x > that.x && this.y > that.y);
    },

    /**
     * Returns true if this point is greater than or equal to another one
     * @param {fabric.Point} that
     * @return {Boolean}
     */
    gte: function (that) {
      return (this.x >= that.x && this.y >= that.y);
    },

    /**
     * Returns new point which is the result of linear interpolation with this one and another one
     * @param {fabric.Point} that
     * @param {Number} t
     * @return {fabric.Point}
     */
    lerp: function (that, t) {
      return new Point(this.x + (that.x - this.x) * t, this.y + (that.y - this.y) * t);
    },

    /**
     * Returns distance from this point and another one
     * @param {fabric.Point} that
     * @return {Number}
     */
    distanceFrom: function (that) {
      var dx = this.x - that.x,
          dy = this.y - that.y;
      return Math.sqrt(dx * dx + dy * dy);
    },

    /**
     * Returns the point between this point and another one
     * @param {fabric.Point} that
     * @return {fabric.Point}
     */
    midPointFrom: function (that) {
      return new Point(this.x + (that.x - this.x)/2, this.y + (that.y - this.y)/2);
    },

    /**
     * Returns a new point which is the min of this and another one
     * @param {fabric.Point} that
     * @return {fabric.Point}
     */
    min: function (that) {
      return new Point(Math.min(this.x, that.x), Math.min(this.y, that.y));
    },

    /**
     * Returns a new point which is the max of this and another one
     * @param {fabric.Point} that
     * @return {fabric.Point}
     */
    max: function (that) {
      return new Point(Math.max(this.x, that.x), Math.max(this.y, that.y));
    },

    /**
     * Returns string representation of this point
     * @return {String}
     */
    toString: function () {
      return this.x + ',' + this.y;
    },

    /**
     * Sets x/y of this point
     * @param {Number} x
     * @param {Number} y
     */
    setXY: function (x, y) {
      this.x = x;
      this.y = y;
    },

    /**
     * Sets x/y of this point from another point
     * @param {fabric.Point} that
     */
    setFromPoint: function (that) {
      this.x = that.x;
      this.y = that.y;
    },

    /**
     * Swaps x/y of this point and another point
     * @param {fabric.Point} that
     */
    swap: function (that) {
      var x = this.x,
          y = this.y;
      this.x = that.x;
      this.y = that.y;
      that.x = x;
      that.y = y;
    }
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  /* Adaptation of work of Kevin Lindsey (kevin@kevlindev.com) */
  var fabric = global.fabric || (global.fabric = { });

  if (fabric.Intersection) {
    fabric.warn('fabric.Intersection is already defined');
    return;
  }

  /**
   * Intersection class
   * @class fabric.Intersection
   * @memberOf fabric
   * @constructor
   */
  function Intersection(status) {
    this.status = status;
    this.points = [];
  }

  fabric.Intersection = Intersection;

  fabric.Intersection.prototype = /** @lends fabric.Intersection.prototype */ {

    /**
     * Appends a point to intersection
     * @param {fabric.Point} point
     */
    appendPoint: function (point) {
      this.points.push(point);
    },

    /**
     * Appends points to intersection
     * @param {Array} points
     */
    appendPoints: function (points) {
      this.points = this.points.concat(points);
    }
  };

  /**
   * Checks if one line intersects another
   * @static
   * @param {fabric.Point} a1
   * @param {fabric.Point} a2
   * @param {fabric.Point} b1
   * @param {fabric.Point} b2
   * @return {fabric.Intersection}
   */
  fabric.Intersection.intersectLineLine = function (a1, a2, b1, b2) {
    var result,
        uaT = (b2.x - b1.x) * (a1.y - b1.y) - (b2.y - b1.y) * (a1.x - b1.x),
        ubT = (a2.x - a1.x) * (a1.y - b1.y) - (a2.y - a1.y) * (a1.x - b1.x),
        uB = (b2.y - b1.y) * (a2.x - a1.x) - (b2.x - b1.x) * (a2.y - a1.y);
    if (uB !== 0) {
      var ua = uaT / uB,
          ub = ubT / uB;
      if (0 <= ua && ua <= 1 && 0 <= ub && ub <= 1) {
        result = new Intersection('Intersection');
        result.points.push(new fabric.Point(a1.x + ua * (a2.x - a1.x), a1.y + ua * (a2.y - a1.y)));
      }
      else {
        result = new Intersection();
      }
    }
    else {
      if (uaT === 0 || ubT === 0) {
        result = new Intersection('Coincident');
      }
      else {
        result = new Intersection('Parallel');
      }
    }
    return result;
  };

  /**
   * Checks if line intersects polygon
   * @static
   * @param {fabric.Point} a1
   * @param {fabric.Point} a2
   * @param {Array} points
   * @return {fabric.Intersection}
   */
  fabric.Intersection.intersectLinePolygon = function(a1, a2, points) {
    var result = new Intersection(),
        length = points.length;

    for (var i = 0; i < length; i++) {
      var b1 = points[i],
          b2 = points[(i + 1) % length],
          inter = Intersection.intersectLineLine(a1, a2, b1, b2);

      result.appendPoints(inter.points);
    }
    if (result.points.length > 0) {
      result.status = 'Intersection';
    }
    return result;
  };

  /**
   * Checks if polygon intersects another polygon
   * @static
   * @param {Array} points1
   * @param {Array} points2
   * @return {fabric.Intersection}
   */
  fabric.Intersection.intersectPolygonPolygon = function (points1, points2) {
    var result = new Intersection(),
        length = points1.length;

    for (var i = 0; i < length; i++) {
      var a1 = points1[i],
          a2 = points1[(i + 1) % length],
          inter = Intersection.intersectLinePolygon(a1, a2, points2);

      result.appendPoints(inter.points);
    }
    if (result.points.length > 0) {
      result.status = 'Intersection';
    }
    return result;
  };

  /**
   * Checks if polygon intersects rectangle
   * @static
   * @param {Array} points
   * @param {Number} r1
   * @param {Number} r2
   * @return {fabric.Intersection}
   */
  fabric.Intersection.intersectPolygonRectangle = function (points, r1, r2) {
    var min = r1.min(r2),
        max = r1.max(r2),
        topRight = new fabric.Point(max.x, min.y),
        bottomLeft = new fabric.Point(min.x, max.y),
        inter1 = Intersection.intersectLinePolygon(min, topRight, points),
        inter2 = Intersection.intersectLinePolygon(topRight, max, points),
        inter3 = Intersection.intersectLinePolygon(max, bottomLeft, points),
        inter4 = Intersection.intersectLinePolygon(bottomLeft, min, points),
        result = new Intersection();

    result.appendPoints(inter1.points);
    result.appendPoints(inter2.points);
    result.appendPoints(inter3.points);
    result.appendPoints(inter4.points);

    if (result.points.length > 0) {
      result.status = 'Intersection';
    }
    return result;
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { });

  if (fabric.Color) {
    fabric.warn('fabric.Color is already defined.');
    return;
  }

  /**
   * Color class
   * The purpose of {@link fabric.Color} is to abstract and encapsulate common color operations;
   * {@link fabric.Color} is a constructor and creates instances of {@link fabric.Color} objects.
   *
   * @class fabric.Color
   * @param {String} color optional in hex or rgb(a) or hsl format or from known color list
   * @return {fabric.Color} thisArg
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-2/#colors}
   */
  function Color(color) {
    if (!color) {
      this.setSource([0, 0, 0, 1]);
    }
    else {
      this._tryParsingColor(color);
    }
  }

  fabric.Color = Color;

  fabric.Color.prototype = /** @lends fabric.Color.prototype */ {

    /**
     * @private
     * @param {String|Array} color Color value to parse
     */
    _tryParsingColor: function(color) {
      var source;

      if (color in Color.colorNameMap) {
        color = Color.colorNameMap[color];
      }

      if (color === 'transparent') {
        source = [255, 255, 255, 0];
      }

      if (!source) {
        source = Color.sourceFromHex(color);
      }
      if (!source) {
        source = Color.sourceFromRgb(color);
      }
      if (!source) {
        source = Color.sourceFromHsl(color);
      }
      if (!source) {
        //if color is not recognize let's make black as canvas does
        source = [0, 0, 0, 1];
      }
      if (source) {
        this.setSource(source);
      }
    },

    /**
     * Adapted from <a href="https://rawgithub.com/mjijackson/mjijackson.github.com/master/2008/02/rgb-to-hsl-and-rgb-to-hsv-color-model-conversion-algorithms-in-javascript.html">https://github.com/mjijackson</a>
     * @private
     * @param {Number} r Red color value
     * @param {Number} g Green color value
     * @param {Number} b Blue color value
     * @return {Array} Hsl color
     */
    _rgbToHsl: function(r, g, b) {
      r /= 255, g /= 255, b /= 255;

      var h, s, l,
          max = fabric.util.array.max([r, g, b]),
          min = fabric.util.array.min([r, g, b]);

      l = (max + min) / 2;

      if (max === min) {
        h = s = 0; // achromatic
      }
      else {
        var d = max - min;
        s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
        switch (max) {
          case r:
            h = (g - b) / d + (g < b ? 6 : 0);
            break;
          case g:
            h = (b - r) / d + 2;
            break;
          case b:
            h = (r - g) / d + 4;
            break;
        }
        h /= 6;
      }

      return [
        Math.round(h * 360),
        Math.round(s * 100),
        Math.round(l * 100)
      ];
    },

    /**
     * Returns source of this color (where source is an array representation; ex: [200, 200, 100, 1])
     * @return {Array}
     */
    getSource: function() {
      return this._source;
    },

    /**
     * Sets source of this color (where source is an array representation; ex: [200, 200, 100, 1])
     * @param {Array} source
     */
    setSource: function(source) {
      this._source = source;
    },

    /**
     * Returns color represenation in RGB format
     * @return {String} ex: rgb(0-255,0-255,0-255)
     */
    toRgb: function() {
      var source = this.getSource();
      return 'rgb(' + source[0] + ',' + source[1] + ',' + source[2] + ')';
    },

    /**
     * Returns color represenation in RGBA format
     * @return {String} ex: rgba(0-255,0-255,0-255,0-1)
     */
    toRgba: function() {
      var source = this.getSource();
      return 'rgba(' + source[0] + ',' + source[1] + ',' + source[2] + ',' + source[3] + ')';
    },

    /**
     * Returns color represenation in HSL format
     * @return {String} ex: hsl(0-360,0%-100%,0%-100%)
     */
    toHsl: function() {
      var source = this.getSource(),
          hsl = this._rgbToHsl(source[0], source[1], source[2]);

      return 'hsl(' + hsl[0] + ',' + hsl[1] + '%,' + hsl[2] + '%)';
    },

    /**
     * Returns color represenation in HSLA format
     * @return {String} ex: hsla(0-360,0%-100%,0%-100%,0-1)
     */
    toHsla: function() {
      var source = this.getSource(),
          hsl = this._rgbToHsl(source[0], source[1], source[2]);

      return 'hsla(' + hsl[0] + ',' + hsl[1] + '%,' + hsl[2] + '%,' + source[3] + ')';
    },

    /**
     * Returns color represenation in HEX format
     * @return {String} ex: FF5555
     */
    toHex: function() {
      var source = this.getSource(), r, g, b;

      r = source[0].toString(16);
      r = (r.length === 1) ? ('0' + r) : r;

      g = source[1].toString(16);
      g = (g.length === 1) ? ('0' + g) : g;

      b = source[2].toString(16);
      b = (b.length === 1) ? ('0' + b) : b;

      return r.toUpperCase() + g.toUpperCase() + b.toUpperCase();
    },

    /**
     * Gets value of alpha channel for this color
     * @return {Number} 0-1
     */
    getAlpha: function() {
      return this.getSource()[3];
    },

    /**
     * Sets value of alpha channel for this color
     * @param {Number} alpha Alpha value 0-1
     * @return {fabric.Color} thisArg
     */
    setAlpha: function(alpha) {
      var source = this.getSource();
      source[3] = alpha;
      this.setSource(source);
      return this;
    },

    /**
     * Transforms color to its grayscale representation
     * @return {fabric.Color} thisArg
     */
    toGrayscale: function() {
      var source = this.getSource(),
          average = parseInt((source[0] * 0.3 + source[1] * 0.59 + source[2] * 0.11).toFixed(0), 10),
          currentAlpha = source[3];
      this.setSource([average, average, average, currentAlpha]);
      return this;
    },

    /**
     * Transforms color to its black and white representation
     * @param {Number} threshold
     * @return {fabric.Color} thisArg
     */
    toBlackWhite: function(threshold) {
      var source = this.getSource(),
          average = (source[0] * 0.3 + source[1] * 0.59 + source[2] * 0.11).toFixed(0),
          currentAlpha = source[3];

      threshold = threshold || 127;

      average = (Number(average) < Number(threshold)) ? 0 : 255;
      this.setSource([average, average, average, currentAlpha]);
      return this;
    },

    /**
     * Overlays color with another color
     * @param {String|fabric.Color} otherColor
     * @return {fabric.Color} thisArg
     */
    overlayWith: function(otherColor) {
      if (!(otherColor instanceof Color)) {
        otherColor = new Color(otherColor);
      }

      var result = [],
          alpha = this.getAlpha(),
          otherAlpha = 0.5,
          source = this.getSource(),
          otherSource = otherColor.getSource();

      for (var i = 0; i < 3; i++) {
        result.push(Math.round((source[i] * (1 - otherAlpha)) + (otherSource[i] * otherAlpha)));
      }

      result[3] = alpha;
      this.setSource(result);
      return this;
    }
  };

  /**
   * Regex matching color in RGB or RGBA formats (ex: rgb(0, 0, 0), rgba(255, 100, 10, 0.5), rgba( 255 , 100 , 10 , 0.5 ), rgb(1,1,1), rgba(100%, 60%, 10%, 0.5))
   * @static
   * @field
   * @memberOf fabric.Color
   */
  fabric.Color.reRGBa = /^rgba?\(\s*(\d{1,3}(?:\.\d+)?\%?)\s*,\s*(\d{1,3}(?:\.\d+)?\%?)\s*,\s*(\d{1,3}(?:\.\d+)?\%?)\s*(?:\s*,\s*(\d+(?:\.\d+)?)\s*)?\)$/;

  /**
   * Regex matching color in HSL or HSLA formats (ex: hsl(200, 80%, 10%), hsla(300, 50%, 80%, 0.5), hsla( 300 , 50% , 80% , 0.5 ))
   * @static
   * @field
   * @memberOf fabric.Color
   */
  fabric.Color.reHSLa = /^hsla?\(\s*(\d{1,3})\s*,\s*(\d{1,3}\%)\s*,\s*(\d{1,3}\%)\s*(?:\s*,\s*(\d+(?:\.\d+)?)\s*)?\)$/;

  /**
   * Regex matching color in HEX format (ex: #FF5555, 010155, aff)
   * @static
   * @field
   * @memberOf fabric.Color
   */
  fabric.Color.reHex = /^#?([0-9a-f]{6}|[0-9a-f]{3})$/i;

  /**
   * Map of the 17 basic color names with HEX code
   * @static
   * @field
   * @memberOf fabric.Color
   * @see: http://www.w3.org/TR/CSS2/syndata.html#color-units
   */
  fabric.Color.colorNameMap = {
    aqua:    '#00FFFF',
    black:   '#000000',
    blue:    '#0000FF',
    fuchsia: '#FF00FF',
    gray:    '#808080',
    grey:    '#808080',
    green:   '#008000',
    lime:    '#00FF00',
    maroon:  '#800000',
    navy:    '#000080',
    olive:   '#808000',
    orange:  '#FFA500',
    purple:  '#800080',
    red:     '#FF0000',
    silver:  '#C0C0C0',
    teal:    '#008080',
    white:   '#FFFFFF',
    yellow:  '#FFFF00'
  };

  /**
   * @private
   * @param {Number} p
   * @param {Number} q
   * @param {Number} t
   * @return {Number}
   */
  function hue2rgb(p, q, t) {
    if (t < 0) {
      t += 1;
    }
    if (t > 1) {
      t -= 1;
    }
    if (t < 1/6) {
      return p + (q - p) * 6 * t;
    }
    if (t < 1/2) {
      return q;
    }
    if (t < 2/3) {
      return p + (q - p) * (2/3 - t) * 6;
    }
    return p;
  }

  /**
   * Returns new color object, when given a color in RGB format
   * @memberOf fabric.Color
   * @param {String} color Color value ex: rgb(0-255,0-255,0-255)
   * @return {fabric.Color}
   */
  fabric.Color.fromRgb = function(color) {
    return Color.fromSource(Color.sourceFromRgb(color));
  };

  /**
   * Returns array represenatation (ex: [100, 100, 200, 1]) of a color that's in RGB or RGBA format
   * @memberOf fabric.Color
   * @param {String} color Color value ex: rgb(0-255,0-255,0-255), rgb(0%-100%,0%-100%,0%-100%)
   * @return {Array} source
   */
  fabric.Color.sourceFromRgb = function(color) {
    var match = color.match(Color.reRGBa);
    if (match) {
      var r = parseInt(match[1], 10) / (/%$/.test(match[1]) ? 100 : 1) * (/%$/.test(match[1]) ? 255 : 1),
          g = parseInt(match[2], 10) / (/%$/.test(match[2]) ? 100 : 1) * (/%$/.test(match[2]) ? 255 : 1),
          b = parseInt(match[3], 10) / (/%$/.test(match[3]) ? 100 : 1) * (/%$/.test(match[3]) ? 255 : 1);

      return [
        parseInt(r, 10),
        parseInt(g, 10),
        parseInt(b, 10),
        match[4] ? parseFloat(match[4]) : 1
      ];
    }
  };

  /**
   * Returns new color object, when given a color in RGBA format
   * @static
   * @function
   * @memberOf fabric.Color
   * @param {String} color
   * @return {fabric.Color}
   */
  fabric.Color.fromRgba = Color.fromRgb;

  /**
   * Returns new color object, when given a color in HSL format
   * @param {String} color Color value ex: hsl(0-260,0%-100%,0%-100%)
   * @memberOf fabric.Color
   * @return {fabric.Color}
   */
  fabric.Color.fromHsl = function(color) {
    return Color.fromSource(Color.sourceFromHsl(color));
  };

  /**
   * Returns array represenatation (ex: [100, 100, 200, 1]) of a color that's in HSL or HSLA format.
   * Adapted from <a href="https://rawgithub.com/mjijackson/mjijackson.github.com/master/2008/02/rgb-to-hsl-and-rgb-to-hsv-color-model-conversion-algorithms-in-javascript.html">https://github.com/mjijackson</a>
   * @memberOf fabric.Color
   * @param {String} color Color value ex: hsl(0-360,0%-100%,0%-100%) or hsla(0-360,0%-100%,0%-100%, 0-1)
   * @return {Array} source
   * @see http://http://www.w3.org/TR/css3-color/#hsl-color
   */
  fabric.Color.sourceFromHsl = function(color) {
    var match = color.match(Color.reHSLa);
    if (!match) {
      return;
    }

    var h = (((parseFloat(match[1]) % 360) + 360) % 360) / 360,
        s = parseFloat(match[2]) / (/%$/.test(match[2]) ? 100 : 1),
        l = parseFloat(match[3]) / (/%$/.test(match[3]) ? 100 : 1),
        r, g, b;

    if (s === 0) {
      r = g = b = l;
    }
    else {
      var q = l <= 0.5 ? l * (s + 1) : l + s - l * s,
          p = l * 2 - q;

      r = hue2rgb(p, q, h + 1/3);
      g = hue2rgb(p, q, h);
      b = hue2rgb(p, q, h - 1/3);
    }

    return [
      Math.round(r * 255),
      Math.round(g * 255),
      Math.round(b * 255),
      match[4] ? parseFloat(match[4]) : 1
    ];
  };

  /**
   * Returns new color object, when given a color in HSLA format
   * @static
   * @function
   * @memberOf fabric.Color
   * @param {String} color
   * @return {fabric.Color}
   */
  fabric.Color.fromHsla = Color.fromHsl;

  /**
   * Returns new color object, when given a color in HEX format
   * @static
   * @memberOf fabric.Color
   * @param {String} color Color value ex: FF5555
   * @return {fabric.Color}
   */
  fabric.Color.fromHex = function(color) {
    return Color.fromSource(Color.sourceFromHex(color));
  };

  /**
   * Returns array represenatation (ex: [100, 100, 200, 1]) of a color that's in HEX format
   * @static
   * @memberOf fabric.Color
   * @param {String} color ex: FF5555
   * @return {Array} source
   */
  fabric.Color.sourceFromHex = function(color) {
    if (color.match(Color.reHex)) {
      var value = color.slice(color.indexOf('#') + 1),
          isShortNotation = (value.length === 3),
          r = isShortNotation ? (value.charAt(0) + value.charAt(0)) : value.substring(0, 2),
          g = isShortNotation ? (value.charAt(1) + value.charAt(1)) : value.substring(2, 4),
          b = isShortNotation ? (value.charAt(2) + value.charAt(2)) : value.substring(4, 6);

      return [
        parseInt(r, 16),
        parseInt(g, 16),
        parseInt(b, 16),
        1
      ];
    }
  };

  /**
   * Returns new color object, when given color in array representation (ex: [200, 100, 100, 0.5])
   * @static
   * @memberOf fabric.Color
   * @param {Array} source
   * @return {fabric.Color}
   */
  fabric.Color.fromSource = function(source) {
    var oColor = new Color();
    oColor.setSource(source);
    return oColor;
  };

})(typeof exports !== 'undefined' ? exports : this);


(function() {

  /* _FROM_SVG_START_ */
  function getColorStop(el) {
    var style = el.getAttribute('style'),
        offset = el.getAttribute('offset') || 0,
        color, colorAlpha, opacity;

    // convert percents to absolute values
    offset = parseFloat(offset) / (/%$/.test(offset) ? 100 : 1);
    offset = offset < 0 ? 0 : offset > 1 ? 1 : offset;
    if (style) {
      var keyValuePairs = style.split(/\s*;\s*/);

      if (keyValuePairs[keyValuePairs.length - 1] === '') {
        keyValuePairs.pop();
      }

      for (var i = keyValuePairs.length; i--; ) {

        var split = keyValuePairs[i].split(/\s*:\s*/),
            key = split[0].trim(),
            value = split[1].trim();

        if (key === 'stop-color') {
          color = value;
        }
        else if (key === 'stop-opacity') {
          opacity = value;
        }
      }
    }

    if (!color) {
      color = el.getAttribute('stop-color') || 'rgb(0,0,0)';
    }
    if (!opacity) {
      opacity = el.getAttribute('stop-opacity');
    }

    color = new fabric.Color(color);
    colorAlpha = color.getAlpha();
    opacity = isNaN(parseFloat(opacity)) ? 1 : parseFloat(opacity);
    opacity *= colorAlpha;

    return {
      offset: offset,
      color: color.toRgb(),
      opacity: opacity
    };
  }

  function getLinearCoords(el) {
    return {
      x1: el.getAttribute('x1') || 0,
      y1: el.getAttribute('y1') || 0,
      x2: el.getAttribute('x2') || '100%',
      y2: el.getAttribute('y2') || 0
    };
  }

  function getRadialCoords(el) {
    return {
      x1: el.getAttribute('fx') || el.getAttribute('cx') || '50%',
      y1: el.getAttribute('fy') || el.getAttribute('cy') || '50%',
      r1: 0,
      x2: el.getAttribute('cx') || '50%',
      y2: el.getAttribute('cy') || '50%',
      r2: el.getAttribute('r') || '50%'
    };
  }
  /* _FROM_SVG_END_ */

  /**
   * Gradient class
   * @class fabric.Gradient
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-2#gradients}
   * @see {@link fabric.Gradient#initialize} for constructor definition
   */
  fabric.Gradient = fabric.util.createClass(/** @lends fabric.Gradient.prototype */ {

    /**
     * Horizontal offset for aligning gradients coming from SVG when outside pathgroups
     * @type Number
     * @default 0
     */
    offsetX: 0,

    /**
     * Vertical offset for aligning gradients coming from SVG when outside pathgroups
     * @type Number
     * @default 0
     */
    offsetY: 0,

    /**
     * Constructor
     * @param {Object} [options] Options object with type, coords, gradientUnits and colorStops
     * @return {fabric.Gradient} thisArg
     */
    initialize: function(options) {
      options || (options = { });

      var coords = { };

      this.id = fabric.Object.__uid++;
      this.type = options.type || 'linear';

      coords = {
        x1: options.coords.x1 || 0,
        y1: options.coords.y1 || 0,
        x2: options.coords.x2 || 0,
        y2: options.coords.y2 || 0
      };

      if (this.type === 'radial') {
        coords.r1 = options.coords.r1 || 0;
        coords.r2 = options.coords.r2 || 0;
      }
      this.coords = coords;
      this.colorStops = options.colorStops.slice();
      if (options.gradientTransform) {
        this.gradientTransform = options.gradientTransform;
      }
      this.offsetX = options.offsetX || this.offsetX;
      this.offsetY = options.offsetY || this.offsetY;
    },

    /**
     * Adds another colorStop
     * @param {Object} colorStop Object with offset and color
     * @return {fabric.Gradient} thisArg
     */
    addColorStop: function(colorStop) {
      for (var position in colorStop) {
        var color = new fabric.Color(colorStop[position]);
        this.colorStops.push({
          offset: position,
          color: color.toRgb(),
          opacity: color.getAlpha()
        });
      }
      return this;
    },

    /**
     * Returns object representation of a gradient
     * @return {Object}
     */
    toObject: function() {
      return {
        type: this.type,
        coords: this.coords,
        colorStops: this.colorStops,
        offsetX: this.offsetX,
        offsetY: this.offsetY,
        gradientTransform: this.gradientTransform ? this.gradientTransform.concat() : this.gradientTransform
      };
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an gradient
     * @param {Object} object Object to create a gradient for
     * @param {Boolean} normalize Whether coords should be normalized
     * @return {String} SVG representation of an gradient (linear/radial)
     */
    toSVG: function(object) {
      var coords = fabric.util.object.clone(this.coords),
          markup, commonAttributes;

      // colorStops must be sorted ascending
      this.colorStops.sort(function(a, b) {
        return a.offset - b.offset;
      });

      if (!(object.group && object.group.type === 'path-group')) {
        for (var prop in coords) {
          if (prop === 'x1' || prop === 'x2' || prop === 'r2') {
            coords[prop] += this.offsetX - object.width / 2;
          }
          else if (prop === 'y1' || prop === 'y2') {
            coords[prop] += this.offsetY - object.height / 2;
          }
        }
      }

      commonAttributes = 'id="SVGID_' + this.id +
                     '" gradientUnits="userSpaceOnUse"';
      if (this.gradientTransform) {
        commonAttributes += ' gradientTransform="matrix(' + this.gradientTransform.join(' ') + ')" ';
      }
      if (this.type === 'linear') {
        markup = [
          //jscs:disable validateIndentation
          '<linearGradient ',
            commonAttributes,
            ' x1="', coords.x1,
            '" y1="', coords.y1,
            '" x2="', coords.x2,
            '" y2="', coords.y2,
          '">\n'
          //jscs:enable validateIndentation
        ];
      }
      else if (this.type === 'radial') {
        markup = [
          //jscs:disable validateIndentation
          '<radialGradient ',
            commonAttributes,
            ' cx="', coords.x2,
            '" cy="', coords.y2,
            '" r="', coords.r2,
            '" fx="', coords.x1,
            '" fy="', coords.y1,
          '">\n'
          //jscs:enable validateIndentation
        ];
      }

      for (var i = 0; i < this.colorStops.length; i++) {
        markup.push(
          //jscs:disable validateIndentation
          '<stop ',
            'offset="', (this.colorStops[i].offset * 100) + '%',
            '" style="stop-color:', this.colorStops[i].color,
            (this.colorStops[i].opacity != null ? ';stop-opacity: ' + this.colorStops[i].opacity : ';'),
          '"/>\n'
          //jscs:enable validateIndentation
        );
      }

      markup.push((this.type === 'linear' ? '</linearGradient>\n' : '</radialGradient>\n'));

      return markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns an instance of CanvasGradient
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @return {CanvasGradient}
     */
    toLive: function(ctx, object) {
      var gradient, prop, coords = fabric.util.object.clone(this.coords);

      if (!this.type) {
        return;
      }

      if (object.group && object.group.type === 'path-group') {
        for (prop in coords) {
          if (prop === 'x1' || prop === 'x2') {
            coords[prop] += -this.offsetX + object.width / 2;
          }
          else if (prop === 'y1' || prop === 'y2') {
            coords[prop] += -this.offsetY + object.height / 2;
          }
        }
      }

      if (this.type === 'linear') {
        gradient = ctx.createLinearGradient(
          coords.x1, coords.y1, coords.x2, coords.y2);
      }
      else if (this.type === 'radial') {
        gradient = ctx.createRadialGradient(
          coords.x1, coords.y1, coords.r1, coords.x2, coords.y2, coords.r2);
      }

      for (var i = 0, len = this.colorStops.length; i < len; i++) {
        var color = this.colorStops[i].color,
            opacity = this.colorStops[i].opacity,
            offset = this.colorStops[i].offset;

        if (typeof opacity !== 'undefined') {
          color = new fabric.Color(color).setAlpha(opacity).toRgba();
        }
        gradient.addColorStop(parseFloat(offset), color);
      }

      return gradient;
    }
  });

  fabric.util.object.extend(fabric.Gradient, {

    /* _FROM_SVG_START_ */
    /**
     * Returns {@link fabric.Gradient} instance from an SVG element
     * @static
     * @memberOf fabric.Gradient
     * @param {SVGGradientElement} el SVG gradient element
     * @param {fabric.Object} instance
     * @return {fabric.Gradient} Gradient instance
     * @see http://www.w3.org/TR/SVG/pservers.html#LinearGradientElement
     * @see http://www.w3.org/TR/SVG/pservers.html#RadialGradientElement
     */
    fromElement: function(el, instance) {

      /**
       *  @example:
       *
       *  <linearGradient id="linearGrad1">
       *    <stop offset="0%" stop-color="white"/>
       *    <stop offset="100%" stop-color="black"/>
       *  </linearGradient>
       *
       *  OR
       *
       *  <linearGradient id="linearGrad2">
       *    <stop offset="0" style="stop-color:rgb(255,255,255)"/>
       *    <stop offset="1" style="stop-color:rgb(0,0,0)"/>
       *  </linearGradient>
       *
       *  OR
       *
       *  <radialGradient id="radialGrad1">
       *    <stop offset="0%" stop-color="white" stop-opacity="1" />
       *    <stop offset="50%" stop-color="black" stop-opacity="0.5" />
       *    <stop offset="100%" stop-color="white" stop-opacity="1" />
       *  </radialGradient>
       *
       *  OR
       *
       *  <radialGradient id="radialGrad2">
       *    <stop offset="0" stop-color="rgb(255,255,255)" />
       *    <stop offset="0.5" stop-color="rgb(0,0,0)" />
       *    <stop offset="1" stop-color="rgb(255,255,255)" />
       *  </radialGradient>
       *
       */

      var colorStopEls = el.getElementsByTagName('stop'),
          type = (el.nodeName === 'linearGradient' ? 'linear' : 'radial'),
          gradientUnits = el.getAttribute('gradientUnits') || 'objectBoundingBox',
          gradientTransform = el.getAttribute('gradientTransform'),
          colorStops = [],
          coords = { }, ellipseMatrix;

      if (type === 'linear') {
        coords = getLinearCoords(el);
      }
      else if (type === 'radial') {
        coords = getRadialCoords(el);
      }

      for (var i = colorStopEls.length; i--; ) {
        colorStops.push(getColorStop(colorStopEls[i]));
      }

      ellipseMatrix = _convertPercentUnitsToValues(instance, coords, gradientUnits);

      var gradient = new fabric.Gradient({
        type: type,
        coords: coords,
        colorStops: colorStops,
        offsetX: -instance.left,
        offsetY: -instance.top
      });

      if (gradientTransform || ellipseMatrix !== '') {
        gradient.gradientTransform = fabric.parseTransformAttribute((gradientTransform || '') + ellipseMatrix);
      }
      return gradient;
    },
    /* _FROM_SVG_END_ */

    /**
     * Returns {@link fabric.Gradient} instance from its object representation
     * @static
     * @memberOf fabric.Gradient
     * @param {Object} obj
     * @param {Object} [options] Options object
     */
    forObject: function(obj, options) {
      options || (options = { });
      _convertPercentUnitsToValues(obj, options.coords, 'userSpaceOnUse');
      return new fabric.Gradient(options);
    }
  });

  /**
   * @private
   */
  function _convertPercentUnitsToValues(object, options, gradientUnits) {
    var propValue, addFactor = 0, multFactor = 1, ellipseMatrix = '';
    for (var prop in options) {
      propValue = parseFloat(options[prop], 10);
      if (typeof options[prop] === 'string' && /^\d+%$/.test(options[prop])) {
        multFactor = 0.01;
      }
      else {
        multFactor = 1;
      }
      if (prop === 'x1' || prop === 'x2' || prop === 'r2') {
        multFactor *= gradientUnits === 'objectBoundingBox' ? object.width : 1;
        addFactor = gradientUnits === 'objectBoundingBox' ? object.left || 0 : 0;
      }
      else if (prop === 'y1' || prop === 'y2') {
        multFactor *= gradientUnits === 'objectBoundingBox' ? object.height : 1;
        addFactor = gradientUnits === 'objectBoundingBox' ? object.top || 0 : 0;
      }
      options[prop] = propValue * multFactor + addFactor;
    }
    if (object.type === 'ellipse' &&
        options.r2 !== null &&
        gradientUnits === 'objectBoundingBox' &&
        object.rx !== object.ry) {

      var scaleFactor = object.ry/object.rx;
      ellipseMatrix = ' scale(1, ' + scaleFactor + ')';
      if (options.y1) {
        options.y1 /= scaleFactor;
      }
      if (options.y2) {
        options.y2 /= scaleFactor;
      }
    }
    return ellipseMatrix;
  }
})();


/**
 * Pattern class
 * @class fabric.Pattern
 * @see {@link http://fabricjs.com/patterns|Pattern demo}
 * @see {@link http://fabricjs.com/dynamic-patterns|DynamicPattern demo}
 * @see {@link fabric.Pattern#initialize} for constructor definition
 */
fabric.Pattern = fabric.util.createClass(/** @lends fabric.Pattern.prototype */ {

  /**
   * Repeat property of a pattern (one of repeat, repeat-x, repeat-y or no-repeat)
   * @type String
   * @default
   */
  repeat: 'repeat',

  /**
   * Pattern horizontal offset from object's left/top corner
   * @type Number
   * @default
   */
  offsetX: 0,

  /**
   * Pattern vertical offset from object's left/top corner
   * @type Number
   * @default
   */
  offsetY: 0,

  /**
   * Constructor
   * @param {Object} [options] Options object
   * @return {fabric.Pattern} thisArg
   */
  initialize: function(options) {
    options || (options = { });

    this.id = fabric.Object.__uid++;

    if (options.source) {
      if (typeof options.source === 'string') {
        // function string
        if (typeof fabric.util.getFunctionBody(options.source) !== 'undefined') {
          this.source = new Function(fabric.util.getFunctionBody(options.source));
        }
        else {
          // img src string
          var _this = this;
          this.source = fabric.util.createImage();
          fabric.util.loadImage(options.source, function(img) {
            _this.source = img;
          });
        }
      }
      else {
        // img element
        this.source = options.source;
      }
    }
    if (options.repeat) {
      this.repeat = options.repeat;
    }
    if (options.offsetX) {
      this.offsetX = options.offsetX;
    }
    if (options.offsetY) {
      this.offsetY = options.offsetY;
    }
  },

  /**
   * Returns object representation of a pattern
   * @return {Object} Object representation of a pattern instance
   */
  toObject: function() {

    var source;

    // callback
    if (typeof this.source === 'function') {
      source = String(this.source);
    }
    // <img> element
    else if (typeof this.source.src === 'string') {
      source = this.source.src;
    }
    // <canvas> element
    else if (typeof this.source === 'object' && this.source.toDataURL) {
      source = this.source.toDataURL();
    }

    return {
      source: source,
      repeat: this.repeat,
      offsetX: this.offsetX,
      offsetY: this.offsetY
    };
  },

  /* _TO_SVG_START_ */
  /**
   * Returns SVG representation of a pattern
   * @param {fabric.Object} object
   * @return {String} SVG representation of a pattern
   */
  toSVG: function(object) {
    var patternSource = typeof this.source === 'function' ? this.source() : this.source,
        patternWidth = patternSource.width / object.getWidth(),
        patternHeight = patternSource.height / object.getHeight(),
        patternOffsetX = this.offsetX / object.getWidth(),
        patternOffsetY = this.offsetY / object.getHeight(),
        patternImgSrc = '';
    if (this.repeat === 'repeat-x' || this.repeat === 'no-repeat') {
      patternHeight = 1;
    }
    if (this.repeat === 'repeat-y' || this.repeat === 'no-repeat') {
      patternWidth = 1;
    }
    if (patternSource.src) {
      patternImgSrc = patternSource.src;
    }
    else if (patternSource.toDataURL) {
      patternImgSrc = patternSource.toDataURL();
    }

    return '<pattern id="SVGID_' + this.id +
                  '" x="' + patternOffsetX +
                  '" y="' + patternOffsetY +
                  '" width="' + patternWidth +
                  '" height="' + patternHeight + '">\n' +
             '<image x="0" y="0"' +
                    ' width="' + patternSource.width +
                    '" height="' + patternSource.height +
                    '" xlink:href="' + patternImgSrc +
             '"></image>\n' +
           '</pattern>\n';
  },
  /* _TO_SVG_END_ */

  /**
   * Returns an instance of CanvasPattern
   * @param {CanvasRenderingContext2D} ctx Context to create pattern
   * @return {CanvasPattern}
   */
  toLive: function(ctx) {
    var source = typeof this.source === 'function'
      ? this.source()
      : this.source;

    // if the image failed to load, return, and allow rest to continue loading
    if (!source) {
      return '';
    }

    // if an image
    if (typeof source.src !== 'undefined') {
      if (!source.complete) {
        return '';
      }
      if (source.naturalWidth === 0 || source.naturalHeight === 0) {
        return '';
      }
    }
    return ctx.createPattern(source, this.repeat);
  }
});


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      toFixed = fabric.util.toFixed;

  if (fabric.Shadow) {
    fabric.warn('fabric.Shadow is already defined.');
    return;
  }

  /**
   * Shadow class
   * @class fabric.Shadow
   * @see {@link http://fabricjs.com/shadows|Shadow demo}
   * @see {@link fabric.Shadow#initialize} for constructor definition
   */
  fabric.Shadow = fabric.util.createClass(/** @lends fabric.Shadow.prototype */ {

    /**
     * Shadow color
     * @type String
     * @default
     */
    color: 'rgb(0,0,0)',

    /**
     * Shadow blur
     * @type Number
     */
    blur: 0,

    /**
     * Shadow horizontal offset
     * @type Number
     * @default
     */
    offsetX: 0,

    /**
     * Shadow vertical offset
     * @type Number
     * @default
     */
    offsetY: 0,

    /**
     * Whether the shadow should affect stroke operations
     * @type Boolean
     * @default
     */
    affectStroke: false,

    /**
     * Indicates whether toObject should include default values
     * @type Boolean
     * @default
     */
    includeDefaultValues: true,

    /**
     * Constructor
     * @param {Object|String} [options] Options object with any of color, blur, offsetX, offsetX properties or string (e.g. "rgba(0,0,0,0.2) 2px 2px 10px, "2px 2px 10px rgba(0,0,0,0.2)")
     * @return {fabric.Shadow} thisArg
     */
    initialize: function(options) {

      if (typeof options === 'string') {
        options = this._parseShadow(options);
      }

      for (var prop in options) {
        this[prop] = options[prop];
      }

      this.id = fabric.Object.__uid++;
    },

    /**
     * @private
     * @param {String} shadow Shadow value to parse
     * @return {Object} Shadow object with color, offsetX, offsetY and blur
     */
    _parseShadow: function(shadow) {
      var shadowStr = shadow.trim(),
          offsetsAndBlur = fabric.Shadow.reOffsetsAndBlur.exec(shadowStr) || [ ],
          color = shadowStr.replace(fabric.Shadow.reOffsetsAndBlur, '') || 'rgb(0,0,0)';

      return {
        color: color.trim(),
        offsetX: parseInt(offsetsAndBlur[1], 10) || 0,
        offsetY: parseInt(offsetsAndBlur[2], 10) || 0,
        blur: parseInt(offsetsAndBlur[3], 10) || 0
      };
    },

    /**
     * Returns a string representation of an instance
     * @see http://www.w3.org/TR/css-text-decor-3/#text-shadow
     * @return {String} Returns CSS3 text-shadow declaration
     */
    toString: function() {
      return [this.offsetX, this.offsetY, this.blur, this.color].join('px ');
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of a shadow
     * @param {fabric.Object} object
     * @return {String} SVG representation of a shadow
     */
    toSVG: function(object) {
      var fBoxX = 40, fBoxY = 40, NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS,
          offset = fabric.util.rotateVector(
            { x: this.offsetX, y: this.offsetY },
            fabric.util.degreesToRadians(-object.angle)),
          BLUR_BOX = 20;

      if (object.width && object.height) {
        //http://www.w3.org/TR/SVG/filters.html#FilterEffectsRegion
        // we add some extra space to filter box to contain the blur ( 20 )
        fBoxX = toFixed((Math.abs(offset.x) + this.blur) / object.width, NUM_FRACTION_DIGITS) * 100 + BLUR_BOX;
        fBoxY = toFixed((Math.abs(offset.y) + this.blur) / object.height, NUM_FRACTION_DIGITS) * 100 + BLUR_BOX;
      }
      if (object.flipX) {
        offset.x *= -1;
      }
      if (object.flipY) {
        offset.y *= -1;
      }
      return (
        '<filter id="SVGID_' + this.id + '" y="-' + fBoxY + '%" height="' + (100 + 2 * fBoxY) + '%" ' +
          'x="-' + fBoxX + '%" width="' + (100 + 2 * fBoxX) + '%" ' + '>\n' +
          '\t<feGaussianBlur in="SourceAlpha" stdDeviation="' +
            toFixed(this.blur ? this.blur / 2 : 0, NUM_FRACTION_DIGITS) + '"></feGaussianBlur>\n' +
          '\t<feOffset dx="' + toFixed(offset.x, NUM_FRACTION_DIGITS) +
          '" dy="' + toFixed(offset.y, NUM_FRACTION_DIGITS) + '" result="oBlur" ></feOffset>\n' +
          '\t<feFlood flood-color="' + this.color + '"/>\n' +
          '\t<feComposite in2="oBlur" operator="in" />\n' +
          '\t<feMerge>\n' +
            '\t\t<feMergeNode></feMergeNode>\n' +
            '\t\t<feMergeNode in="SourceGraphic"></feMergeNode>\n' +
          '\t</feMerge>\n' +
        '</filter>\n');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns object representation of a shadow
     * @return {Object} Object representation of a shadow instance
     */
    toObject: function() {
      if (this.includeDefaultValues) {
        return {
          color: this.color,
          blur: this.blur,
          offsetX: this.offsetX,
          offsetY: this.offsetY,
          affectStroke: this.affectStroke
        };
      }
      var obj = { }, proto = fabric.Shadow.prototype;

      ['color', 'blur', 'offsetX', 'offsetY', 'affectStroke'].forEach(function(prop) {
        if (this[prop] !== proto[prop]) {
          obj[prop] = this[prop];
        }
      }, this);

      return obj;
    }
  });

  /**
   * Regex matching shadow offsetX, offsetY and blur (ex: "2px 2px 10px rgba(0,0,0,0.2)", "rgb(0,255,0) 2px 2px")
   * @static
   * @field
   * @memberOf fabric.Shadow
   */
  fabric.Shadow.reOffsetsAndBlur = /(?:\s|^)(-?\d+(?:px)?(?:\s?|$))?(-?\d+(?:px)?(?:\s?|$))?(\d+(?:px)?)?(?:\s?|$)(?:$|\s)/;

})(typeof exports !== 'undefined' ? exports : this);


(function () {

  'use strict';

  if (fabric.StaticCanvas) {
    fabric.warn('fabric.StaticCanvas is already defined.');
    return;
  }

  // aliases for faster resolution
  var extend = fabric.util.object.extend,
      getElementOffset = fabric.util.getElementOffset,
      removeFromArray = fabric.util.removeFromArray,
      toFixed = fabric.util.toFixed,

      CANVAS_INIT_ERROR = new Error('Could not initialize `canvas` element');

  /**
   * Static canvas class
   * @class fabric.StaticCanvas
   * @mixes fabric.Collection
   * @mixes fabric.Observable
   * @see {@link http://fabricjs.com/static_canvas|StaticCanvas demo}
   * @see {@link fabric.StaticCanvas#initialize} for constructor definition
   * @fires before:render
   * @fires after:render
   * @fires canvas:cleared
   * @fires object:added
   * @fires object:removed
   */
  fabric.StaticCanvas = fabric.util.createClass(/** @lends fabric.StaticCanvas.prototype */ {

    /**
     * Constructor
     * @param {HTMLElement | String} el &lt;canvas> element to initialize instance on
     * @param {Object} [options] Options object
     * @return {Object} thisArg
     */
    initialize: function(el, options) {
      options || (options = { });

      this._initStatic(el, options);
    },

    /**
     * Background color of canvas instance.
     * Should be set via {@link fabric.StaticCanvas#setBackgroundColor}.
     * @type {(String|fabric.Pattern)}
     * @default
     */
    backgroundColor: '',

    /**
     * Background image of canvas instance.
     * Should be set via {@link fabric.StaticCanvas#setBackgroundImage}.
     * <b>Backwards incompatibility note:</b> The "backgroundImageOpacity"
     * and "backgroundImageStretch" properties are deprecated since 1.3.9.
     * Use {@link fabric.Image#opacity}, {@link fabric.Image#width} and {@link fabric.Image#height}.
     * @type fabric.Image
     * @default
     */
    backgroundImage: null,

    /**
     * Overlay color of canvas instance.
     * Should be set via {@link fabric.StaticCanvas#setOverlayColor}
     * @since 1.3.9
     * @type {(String|fabric.Pattern)}
     * @default
     */
    overlayColor: '',

    /**
     * Overlay image of canvas instance.
     * Should be set via {@link fabric.StaticCanvas#setOverlayImage}.
     * <b>Backwards incompatibility note:</b> The "overlayImageLeft"
     * and "overlayImageTop" properties are deprecated since 1.3.9.
     * Use {@link fabric.Image#left} and {@link fabric.Image#top}.
     * @type fabric.Image
     * @default
     */
    overlayImage: null,

    /**
     * Indicates whether toObject/toDatalessObject should include default values
     * @type Boolean
     * @default
     */
    includeDefaultValues: true,

    /**
     * Indicates whether objects' state should be saved
     * @type Boolean
     * @default
     */
    stateful: true,

    /**
     * Indicates whether {@link fabric.Collection.add}, {@link fabric.Collection.insertAt} and {@link fabric.Collection.remove} should also re-render canvas.
     * Disabling this option could give a great performance boost when adding/removing a lot of objects to/from canvas at once
     * (followed by a manual rendering after addition/deletion)
     * @type Boolean
     * @default
     */
    renderOnAddRemove: true,

    /**
     * Function that determines clipping of entire canvas area
     * Being passed context as first argument. See clipping canvas area in {@link https://github.com/kangax/fabric.js/wiki/FAQ}
     * @type Function
     * @default
     */
    clipTo: null,

    /**
     * Indicates whether object controls (borders/controls) are rendered above overlay image
     * @type Boolean
     * @default
     */
    controlsAboveOverlay: false,

    /**
     * Indicates whether the browser can be scrolled when using a touchscreen and dragging on the canvas
     * @type Boolean
     * @default
     */
    allowTouchScrolling: false,

    /**
     * Indicates whether this canvas will use image smoothing, this is on by default in browsers
     * @type Boolean
     * @default
     */
    imageSmoothingEnabled: true,

    /**
     * Indicates whether objects should remain in current stack position when selected. When false objects are brought to top and rendered as part of the selection group
     * @type Boolean
     * @default
     */
    preserveObjectStacking: false,

    /**
     * The transformation (in the format of Canvas transform) which focuses the viewport
     * @type Array
     * @default
     */
    viewportTransform: [1, 0, 0, 1, 0, 0],

    /**
     * if set to false background image is not affected by viewport transform
     * @since 1.6.3
     * @type Boolean
     * @default
     */
    backgroundVpt: true,

    /**
     * if set to false overlya image is not affected by viewport transform
     * @since 1.6.3
     * @type Boolean
     * @default
     */
    overlayVpt: true,

    /**
     * Callback; invoked right before object is about to be scaled/rotated
     */
    onBeforeScaleRotate: function () {
      /* NOOP */
    },

    /**
     * When true, canvas is scaled by devicePixelRatio for better rendering on retina screens
     */
    enableRetinaScaling: true,

    /**
     * @private
     * @param {HTMLElement | String} el &lt;canvas> element to initialize instance on
     * @param {Object} [options] Options object
     */
    _initStatic: function(el, options) {
      this._objects = [];

      this._createLowerCanvas(el);
      this._initOptions(options);
      this._setImageSmoothing();

      // only initialize retina scaling once
      if (!this.interactive) {
        this._initRetinaScaling();
      }

      if (options.overlayImage) {
        this.setOverlayImage(options.overlayImage, this.renderAll.bind(this));
      }
      if (options.backgroundImage) {
        this.setBackgroundImage(options.backgroundImage, this.renderAll.bind(this));
      }
      if (options.backgroundColor) {
        this.setBackgroundColor(options.backgroundColor, this.renderAll.bind(this));
      }
      if (options.overlayColor) {
        this.setOverlayColor(options.overlayColor, this.renderAll.bind(this));
      }
      this.calcOffset();
    },

    /**
     * @private
     */
    _isRetinaScaling: function() {
      return (fabric.devicePixelRatio !== 1 && this.enableRetinaScaling);
    },

    /**
     * @private
     */
    _initRetinaScaling: function() {
      if (!this._isRetinaScaling()) {
        return;
      }

      this.lowerCanvasEl.setAttribute('width', this.width * fabric.devicePixelRatio);
      this.lowerCanvasEl.setAttribute('height', this.height * fabric.devicePixelRatio);

      this.contextContainer.scale(fabric.devicePixelRatio, fabric.devicePixelRatio);
    },

    /**
     * Calculates canvas element offset relative to the document
     * This method is also attached as "resize" event handler of window
     * @return {fabric.Canvas} instance
     * @chainable
     */
    calcOffset: function () {
      this._offset = getElementOffset(this.lowerCanvasEl);
      return this;
    },

    /**
     * Sets {@link fabric.StaticCanvas#overlayImage|overlay image} for this canvas
     * @param {(fabric.Image|String)} image fabric.Image instance or URL of an image to set overlay to
     * @param {Function} callback callback to invoke when image is loaded and set as an overlay
     * @param {Object} [options] Optional options to set for the {@link fabric.Image|overlay image}.
     * @return {fabric.Canvas} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/MnzHT/|jsFiddle demo}
     * @example <caption>Normal overlayImage with left/top = 0</caption>
     * canvas.setOverlayImage('http://fabricjs.com/assets/jail_cell_bars.png', canvas.renderAll.bind(canvas), {
     *   // Needed to position overlayImage at 0/0
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>overlayImage with different properties</caption>
     * canvas.setOverlayImage('http://fabricjs.com/assets/jail_cell_bars.png', canvas.renderAll.bind(canvas), {
     *   opacity: 0.5,
     *   angle: 45,
     *   left: 400,
     *   top: 400,
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>Stretched overlayImage #1 - width/height correspond to canvas width/height</caption>
     * fabric.Image.fromURL('http://fabricjs.com/assets/jail_cell_bars.png', function(img) {
     *    img.set({width: canvas.width, height: canvas.height, originX: 'left', originY: 'top'});
     *    canvas.setOverlayImage(img, canvas.renderAll.bind(canvas));
     * });
     * @example <caption>Stretched overlayImage #2 - width/height correspond to canvas width/height</caption>
     * canvas.setOverlayImage('http://fabricjs.com/assets/jail_cell_bars.png', canvas.renderAll.bind(canvas), {
     *   width: canvas.width,
     *   height: canvas.height,
     *   // Needed to position overlayImage at 0/0
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>overlayImage loaded from cross-origin</caption>
     * canvas.setOverlayImage('http://fabricjs.com/assets/jail_cell_bars.png', canvas.renderAll.bind(canvas), {
     *   opacity: 0.5,
     *   angle: 45,
     *   left: 400,
     *   top: 400,
     *   originX: 'left',
     *   originY: 'top',
     *   crossOrigin: 'anonymous'
     * });
     */
    setOverlayImage: function (image, callback, options) {
      return this.__setBgOverlayImage('overlayImage', image, callback, options);
    },

    /**
     * Sets {@link fabric.StaticCanvas#backgroundImage|background image} for this canvas
     * @param {(fabric.Image|String)} image fabric.Image instance or URL of an image to set background to
     * @param {Function} callback Callback to invoke when image is loaded and set as background
     * @param {Object} [options] Optional options to set for the {@link fabric.Image|background image}.
     * @return {fabric.Canvas} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/YH9yD/|jsFiddle demo}
     * @example <caption>Normal backgroundImage with left/top = 0</caption>
     * canvas.setBackgroundImage('http://fabricjs.com/assets/honey_im_subtle.png', canvas.renderAll.bind(canvas), {
     *   // Needed to position backgroundImage at 0/0
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>backgroundImage with different properties</caption>
     * canvas.setBackgroundImage('http://fabricjs.com/assets/honey_im_subtle.png', canvas.renderAll.bind(canvas), {
     *   opacity: 0.5,
     *   angle: 45,
     *   left: 400,
     *   top: 400,
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>Stretched backgroundImage #1 - width/height correspond to canvas width/height</caption>
     * fabric.Image.fromURL('http://fabricjs.com/assets/honey_im_subtle.png', function(img) {
     *    img.set({width: canvas.width, height: canvas.height, originX: 'left', originY: 'top'});
     *    canvas.setBackgroundImage(img, canvas.renderAll.bind(canvas));
     * });
     * @example <caption>Stretched backgroundImage #2 - width/height correspond to canvas width/height</caption>
     * canvas.setBackgroundImage('http://fabricjs.com/assets/honey_im_subtle.png', canvas.renderAll.bind(canvas), {
     *   width: canvas.width,
     *   height: canvas.height,
     *   // Needed to position backgroundImage at 0/0
     *   originX: 'left',
     *   originY: 'top'
     * });
     * @example <caption>backgroundImage loaded from cross-origin</caption>
     * canvas.setBackgroundImage('http://fabricjs.com/assets/honey_im_subtle.png', canvas.renderAll.bind(canvas), {
     *   opacity: 0.5,
     *   angle: 45,
     *   left: 400,
     *   top: 400,
     *   originX: 'left',
     *   originY: 'top',
     *   crossOrigin: 'anonymous'
     * });
     */
    setBackgroundImage: function (image, callback, options) {
      return this.__setBgOverlayImage('backgroundImage', image, callback, options);
    },

    /**
     * Sets {@link fabric.StaticCanvas#overlayColor|background color} for this canvas
     * @param {(String|fabric.Pattern)} overlayColor Color or pattern to set background color to
     * @param {Function} callback Callback to invoke when background color is set
     * @return {fabric.Canvas} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/pB55h/|jsFiddle demo}
     * @example <caption>Normal overlayColor - color value</caption>
     * canvas.setOverlayColor('rgba(255, 73, 64, 0.6)', canvas.renderAll.bind(canvas));
     * @example <caption>fabric.Pattern used as overlayColor</caption>
     * canvas.setOverlayColor({
     *   source: 'http://fabricjs.com/assets/escheresque_ste.png'
     * }, canvas.renderAll.bind(canvas));
     * @example <caption>fabric.Pattern used as overlayColor with repeat and offset</caption>
     * canvas.setOverlayColor({
     *   source: 'http://fabricjs.com/assets/escheresque_ste.png',
     *   repeat: 'repeat',
     *   offsetX: 200,
     *   offsetY: 100
     * }, canvas.renderAll.bind(canvas));
     */
    setOverlayColor: function(overlayColor, callback) {
      return this.__setBgOverlayColor('overlayColor', overlayColor, callback);
    },

    /**
     * Sets {@link fabric.StaticCanvas#backgroundColor|background color} for this canvas
     * @param {(String|fabric.Pattern)} backgroundColor Color or pattern to set background color to
     * @param {Function} callback Callback to invoke when background color is set
     * @return {fabric.Canvas} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/hXzvk/|jsFiddle demo}
     * @example <caption>Normal backgroundColor - color value</caption>
     * canvas.setBackgroundColor('rgba(255, 73, 64, 0.6)', canvas.renderAll.bind(canvas));
     * @example <caption>fabric.Pattern used as backgroundColor</caption>
     * canvas.setBackgroundColor({
     *   source: 'http://fabricjs.com/assets/escheresque_ste.png'
     * }, canvas.renderAll.bind(canvas));
     * @example <caption>fabric.Pattern used as backgroundColor with repeat and offset</caption>
     * canvas.setBackgroundColor({
     *   source: 'http://fabricjs.com/assets/escheresque_ste.png',
     *   repeat: 'repeat',
     *   offsetX: 200,
     *   offsetY: 100
     * }, canvas.renderAll.bind(canvas));
     */
    setBackgroundColor: function(backgroundColor, callback) {
      return this.__setBgOverlayColor('backgroundColor', backgroundColor, callback);
    },

    /**
     * @private
     * @see {@link http://www.whatwg.org/specs/web-apps/current-work/multipage/the-canvas-element.html#dom-context-2d-imagesmoothingenabled|WhatWG Canvas Standard}
     */
    _setImageSmoothing: function() {
      var ctx = this.getContext();

      ctx.imageSmoothingEnabled = ctx.imageSmoothingEnabled || ctx.webkitImageSmoothingEnabled
        || ctx.mozImageSmoothingEnabled || ctx.msImageSmoothingEnabled || ctx.oImageSmoothingEnabled;
      ctx.imageSmoothingEnabled = this.imageSmoothingEnabled;
    },

    /**
     * @private
     * @param {String} property Property to set ({@link fabric.StaticCanvas#backgroundImage|backgroundImage}
     * or {@link fabric.StaticCanvas#overlayImage|overlayImage})
     * @param {(fabric.Image|String|null)} image fabric.Image instance, URL of an image or null to set background or overlay to
     * @param {Function} callback Callback to invoke when image is loaded and set as background or overlay
     * @param {Object} [options] Optional options to set for the {@link fabric.Image|image}.
     */
    __setBgOverlayImage: function(property, image, callback, options) {
      if (typeof image === 'string') {
        fabric.util.loadImage(image, function(img) {
          img && (this[property] = new fabric.Image(img, options));
          callback && callback(img);
        }, this, options && options.crossOrigin);
      }
      else {
        options && image.setOptions(options);
        this[property] = image;
        callback && callback(image);
      }

      return this;
    },

    /**
     * @private
     * @param {String} property Property to set ({@link fabric.StaticCanvas#backgroundColor|backgroundColor}
     * or {@link fabric.StaticCanvas#overlayColor|overlayColor})
     * @param {(Object|String|null)} color Object with pattern information, color value or null
     * @param {Function} [callback] Callback is invoked when color is set
     */
    __setBgOverlayColor: function(property, color, callback) {
      if (color && color.source) {
        var _this = this;
        fabric.util.loadImage(color.source, function(img) {
          _this[property] = new fabric.Pattern({
            source: img,
            repeat: color.repeat,
            offsetX: color.offsetX,
            offsetY: color.offsetY
          });
          callback && callback();
        });
      }
      else {
        this[property] = color;
        callback && callback();
      }

      return this;
    },

    /**
     * @private
     */
    _createCanvasElement: function() {
      var element = fabric.document.createElement('canvas');
      if (!element.style) {
        element.style = { };
      }
      if (!element) {
        throw CANVAS_INIT_ERROR;
      }
      this._initCanvasElement(element);
      return element;
    },

    /**
     * @private
     * @param {HTMLElement} element
     */
    _initCanvasElement: function(element) {
      fabric.util.createCanvasElement(element);

      if (typeof element.getContext === 'undefined') {
        throw CANVAS_INIT_ERROR;
      }
    },

    /**
     * @private
     * @param {Object} [options] Options object
     */
    _initOptions: function (options) {
      for (var prop in options) {
        this[prop] = options[prop];
      }

      this.width = this.width || parseInt(this.lowerCanvasEl.width, 10) || 0;
      this.height = this.height || parseInt(this.lowerCanvasEl.height, 10) || 0;

      if (!this.lowerCanvasEl.style) {
        return;
      }

      this.lowerCanvasEl.width = this.width;
      this.lowerCanvasEl.height = this.height;

      this.lowerCanvasEl.style.width = this.width + 'px';
      this.lowerCanvasEl.style.height = this.height + 'px';

      this.viewportTransform = this.viewportTransform.slice();
    },

    /**
     * Creates a bottom canvas
     * @private
     * @param {HTMLElement} [canvasEl]
     */
    _createLowerCanvas: function (canvasEl) {
      this.lowerCanvasEl = fabric.util.getById(canvasEl) || this._createCanvasElement();
      this._initCanvasElement(this.lowerCanvasEl);

      fabric.util.addClass(this.lowerCanvasEl, 'lower-canvas');

      if (this.interactive) {
        this._applyCanvasStyle(this.lowerCanvasEl);
      }

      this.contextContainer = this.lowerCanvasEl.getContext('2d');
    },

    /**
     * Returns canvas width (in px)
     * @return {Number}
     */
    getWidth: function () {
      return this.width;
    },

    /**
     * Returns canvas height (in px)
     * @return {Number}
     */
    getHeight: function () {
      return this.height;
    },

    /**
     * Sets width of this canvas instance
     * @param {Number|String} value                         Value to set width to
     * @param {Object}        [options]                     Options object
     * @param {Boolean}       [options.backstoreOnly=false] Set the given dimensions only as canvas backstore dimensions
     * @param {Boolean}       [options.cssOnly=false]       Set the given dimensions only as css dimensions
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    setWidth: function (value, options) {
      return this.setDimensions({ width: value }, options);
    },

    /**
     * Sets height of this canvas instance
     * @param {Number|String} value                         Value to set height to
     * @param {Object}        [options]                     Options object
     * @param {Boolean}       [options.backstoreOnly=false] Set the given dimensions only as canvas backstore dimensions
     * @param {Boolean}       [options.cssOnly=false]       Set the given dimensions only as css dimensions
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    setHeight: function (value, options) {
      return this.setDimensions({ height: value }, options);
    },

    /**
     * Sets dimensions (width, height) of this canvas instance. when options.cssOnly flag active you should also supply the unit of measure (px/%/em)
     * @param {Object}        dimensions                    Object with width/height properties
     * @param {Number|String} [dimensions.width]            Width of canvas element
     * @param {Number|String} [dimensions.height]           Height of canvas element
     * @param {Object}        [options]                     Options object
     * @param {Boolean}       [options.backstoreOnly=false] Set the given dimensions only as canvas backstore dimensions
     * @param {Boolean}       [options.cssOnly=false]       Set the given dimensions only as css dimensions
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    setDimensions: function (dimensions, options) {
      var cssValue;

      options = options || {};

      for (var prop in dimensions) {
        cssValue = dimensions[prop];

        if (!options.cssOnly) {
          this._setBackstoreDimension(prop, dimensions[prop]);
          cssValue += 'px';
        }

        if (!options.backstoreOnly) {
          this._setCssDimension(prop, cssValue);
        }
      }
      this._initRetinaScaling();
      this._setImageSmoothing();
      this.calcOffset();

      if (!options.cssOnly) {
        this.renderAll();
      }

      return this;
    },

    /**
     * Helper for setting width/height
     * @private
     * @param {String} prop property (width|height)
     * @param {Number} value value to set property to
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    _setBackstoreDimension: function (prop, value) {
      this.lowerCanvasEl[prop] = value;

      if (this.upperCanvasEl) {
        this.upperCanvasEl[prop] = value;
      }

      if (this.cacheCanvasEl) {
        this.cacheCanvasEl[prop] = value;
      }

      this[prop] = value;

      return this;
    },

    /**
     * Helper for setting css width/height
     * @private
     * @param {String} prop property (width|height)
     * @param {String} value value to set property to
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    _setCssDimension: function (prop, value) {
      this.lowerCanvasEl.style[prop] = value;

      if (this.upperCanvasEl) {
        this.upperCanvasEl.style[prop] = value;
      }

      if (this.wrapperEl) {
        this.wrapperEl.style[prop] = value;
      }

      return this;
    },

    /**
     * Returns canvas zoom level
     * @return {Number}
     */
    getZoom: function () {
      return Math.sqrt(this.viewportTransform[0] * this.viewportTransform[3]);
    },

    /**
     * Sets viewport transform of this canvas instance
     * @param {Array} vpt the transform in the form of context.transform
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    setViewportTransform: function (vpt) {
      var activeGroup = this.getActiveGroup();
      this.viewportTransform = vpt;
      this.renderAll();
      for (var i = 0, len = this._objects.length; i < len; i++) {
        this._objects[i].setCoords();
      }
      if (activeGroup) {
        activeGroup.setCoords();
      }
      return this;
    },

    /**
     * Sets zoom level of this canvas instance, zoom centered around point
     * @param {fabric.Point} point to zoom with respect to
     * @param {Number} value to set zoom to, less than 1 zooms out
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    zoomToPoint: function (point, value) {
      // TODO: just change the scale, preserve other transformations
      var before = point;
      point = fabric.util.transformPoint(point, fabric.util.invertTransform(this.viewportTransform));
      this.viewportTransform[0] = value;
      this.viewportTransform[3] = value;
      var after = fabric.util.transformPoint(point, this.viewportTransform);
      this.viewportTransform[4] += before.x - after.x;
      this.viewportTransform[5] += before.y - after.y;
      this.renderAll();
      for (var i = 0, len = this._objects.length; i < len; i++) {
        this._objects[i].setCoords();
      }
      return this;
    },

    /**
     * Sets zoom level of this canvas instance
     * @param {Number} value to set zoom to, less than 1 zooms out
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    setZoom: function (value) {
      this.zoomToPoint(new fabric.Point(0, 0), value);
      return this;
    },

    /**
     * Pan viewport so as to place point at top left corner of canvas
     * @param {fabric.Point} point to move to
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    absolutePan: function (point) {
      this.viewportTransform[4] = -point.x;
      this.viewportTransform[5] = -point.y;
      this.renderAll();
      for (var i = 0, len = this._objects.length; i < len; i++) {
        this._objects[i].setCoords();
      }
      return this;
    },

    /**
     * Pans viewpoint relatively
     * @param {fabric.Point} point (position vector) to move by
     * @return {fabric.Canvas} instance
     * @chainable true
     */
    relativePan: function (point) {
      return this.absolutePan(new fabric.Point(
        -point.x - this.viewportTransform[4],
        -point.y - this.viewportTransform[5]
      ));
    },

    /**
     * Returns &lt;canvas> element corresponding to this instance
     * @return {HTMLCanvasElement}
     */
    getElement: function () {
      return this.lowerCanvasEl;
    },

    /**
     * Returns currently selected object, if any
     * @return {fabric.Object}
     */
    getActiveObject: function() {
      return null;
    },

    /**
     * Returns currently selected group of object, if any
     * @return {fabric.Group}
     */
    getActiveGroup: function() {
      return null;
    },

    /**
     * @private
     * @param {fabric.Object} obj Object that was added
     */
    _onObjectAdded: function(obj) {
      this.stateful && obj.setupState();
      obj._set('canvas', this);
      obj.setCoords();
      this.fire('object:added', { target: obj });
      obj.fire('added');
    },

    /**
     * @private
     * @param {fabric.Object} obj Object that was removed
     */
    _onObjectRemoved: function(obj) {
      // removing active object should fire "selection:cleared" events
      if (this.getActiveObject() === obj) {
        this.fire('before:selection:cleared', { target: obj });
        this._discardActiveObject();
        this.fire('selection:cleared');
      }

      this.fire('object:removed', { target: obj });
      obj.fire('removed');
    },

    /**
     * Clears specified context of canvas element
     * @param {CanvasRenderingContext2D} ctx Context to clear
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    clearContext: function(ctx) {
      ctx.clearRect(0, 0, this.width, this.height);
      return this;
    },

    /**
     * Returns context of canvas where objects are drawn
     * @return {CanvasRenderingContext2D}
     */
    getContext: function () {
      return this.contextContainer;
    },

    /**
     * Clears all contexts (background, main, top) of an instance
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    clear: function () {
      this._objects.length = 0;
      if (this.discardActiveGroup) {
        this.discardActiveGroup();
      }
      if (this.discardActiveObject) {
        this.discardActiveObject();
      }
      this.clearContext(this.contextContainer);
      if (this.contextTop) {
        this.clearContext(this.contextTop);
      }
      this.fire('canvas:cleared');
      this.renderAll();
      return this;
    },

    /**
     * Divides objects in two groups, one to render immediately
     * and one to render as activeGroup.
     * return objects to render immediately and pushes the other in the activeGroup.
     */
    _chooseObjectsToRender: function() {
      var activeGroup = this.getActiveGroup(),
          activeObject = this.getActiveObject(),
          object, objsToRender = [ ], activeGroupObjects = [ ];

      if ((activeGroup || activeObject) && !this.preserveObjectStacking) {
        for (var i = 0, length = this._objects.length; i < length; i++) {
          object = this._objects[i];
          if ((!activeGroup || !activeGroup.contains(object)) && object !== activeObject) {
            objsToRender.push(object);
          }
          else {
            activeGroupObjects.push(object);
          }
        }
        activeGroup && activeGroup._set('_objects', activeGroupObjects);
      }
      else {
        objsToRender = this._objects;
      }
      return objsToRender;
    },

    /**
     * Renders both the top canvas and the secondary container canvas.
     * @param {Boolean} [allOnTop] Whether we want to force all images to be rendered on the top canvas
     * @return {fabric.Canvas} instance
     * @chainable
     */
    renderAll: function () {
      var canvasToDrawOn = this.contextContainer, objsToRender;

      if (this.contextTop && this.selection && !this._groupSelector && !this.isDrawingMode) {
        this.clearContext(this.contextTop);
      }

      this.clearContext(canvasToDrawOn);

      this.fire('before:render');

      if (this.clipTo) {
        fabric.util.clipContext(this, canvasToDrawOn);
      }
      this._renderBackground(canvasToDrawOn);

      canvasToDrawOn.save();
      objsToRender = this._chooseObjectsToRender();
      //apply viewport transform once for all rendering process
      canvasToDrawOn.transform.apply(canvasToDrawOn, this.viewportTransform);
      this._renderObjects(canvasToDrawOn, objsToRender);
      if (!this.preserveObjectStacking) {
        objsToRender = [this.getActiveGroup(), this.getActiveObject()];
        this._renderObjects(canvasToDrawOn, objsToRender);
      }
      canvasToDrawOn.restore();

      if (!this.controlsAboveOverlay && this.interactive) {
        this.drawControls(canvasToDrawOn);
      }
      if (this.clipTo) {
        canvasToDrawOn.restore();
      }
      this._renderOverlay(canvasToDrawOn);
      if (this.controlsAboveOverlay && this.interactive) {
        this.drawControls(canvasToDrawOn);
      }

      this.fire('after:render');
      return this;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Array} objects to render
     */
    _renderObjects: function(ctx, objects) {
      for (var i = 0, length = objects.length; i < length; ++i) {
        objects[i] && objects[i].render(ctx);
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {string} property 'background' or 'overlay'
     */
    _renderBackgroundOrOverlay: function(ctx, property) {
      var object = this[property + 'Color'];
      if (object) {
        ctx.fillStyle = object.toLive
          ? object.toLive(ctx)
          : object;

        ctx.fillRect(
          object.offsetX || 0,
          object.offsetY || 0,
          this.width,
          this.height);
      }
      object = this[property + 'Image'];
      if (object) {
        if (this[property + 'Vpt']) {
          ctx.save();
          ctx.transform.apply(ctx, this.viewportTransform);
        }
        object.render(ctx);
        this[property + 'Vpt'] && ctx.restore();
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderBackground: function(ctx) {
      this._renderBackgroundOrOverlay(ctx, 'background');
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderOverlay: function(ctx) {
      this._renderBackgroundOrOverlay(ctx, 'overlay');
    },

    /**
     * Method to render only the top canvas.
     * Also used to render the group selection box.
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    renderTop: function () {
      var ctx = this.contextTop || this.contextContainer;
      this.clearContext(ctx);

      // we render the top context - last object
      if (this.selection && this._groupSelector) {
        this._drawSelection();
      }

      this.fire('after:render');

      return this;
    },

    /**
     * Returns coordinates of a center of canvas.
     * Returned value is an object with top and left properties
     * @return {Object} object with "top" and "left" number values
     */
    getCenter: function () {
      return {
        top: this.getHeight() / 2,
        left: this.getWidth() / 2
      };
    },

    /**
     * Centers object horizontally in the canvas
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center horizontally
     * @return {fabric.Canvas} thisArg
     */
    centerObjectH: function (object) {
      return this._centerObject(object, new fabric.Point(this.getCenter().left, object.getCenterPoint().y));
    },

    /**
     * Centers object vertically in the canvas
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center vertically
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    centerObjectV: function (object) {
      return this._centerObject(object, new fabric.Point(object.getCenterPoint().x, this.getCenter().top));
    },

    /**
     * Centers object vertically and horizontally in the canvas
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center vertically and horizontally
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    centerObject: function(object) {
      var center = this.getCenter();

      return this._centerObject(object, new fabric.Point(center.left, center.top));
    },

    /**
     * Centers object vertically and horizontally in the viewport
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center vertically and horizontally
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    viewportCenterObject: function(object) {
      var vpCenter = this.getVpCenter();

      return this._centerObject(object, vpCenter);
    },

    /**
     * Centers object horizontally in the viewport, object.top is unchanged
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center vertically and horizontally
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    viewportCenterObjectH: function(object) {
      var vpCenter = this.getVpCenter();
      this._centerObject(object, new fabric.Point(vpCenter.x, object.getCenterPoint().y));
      return this;
    },

    /**
     * Centers object Vertically in the viewport, object.top is unchanged
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @param {fabric.Object} object Object to center vertically and horizontally
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    viewportCenterObjectV: function(object) {
      var vpCenter = this.getVpCenter();

      return this._centerObject(object, new fabric.Point(object.getCenterPoint().x, vpCenter.y));
    },

    /**
     * Calculate the point in canvas that correspond to the center of actual viewport.
     * @return {fabric.Point} vpCenter, viewport center
     * @chainable
     */
    getVpCenter: function() {
      var center = this.getCenter(),
          iVpt = fabric.util.invertTransform(this.viewportTransform);
      return fabric.util.transformPoint({ x: center.left, y: center.top }, iVpt);
    },

    /**
     * @private
     * @param {fabric.Object} object Object to center
     * @param {fabric.Point} center Center point
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    _centerObject: function(object, center) {
      object.setPositionByOrigin(center, 'center', 'center');
      this.renderAll();
      return this;
    },

    /**
     * Returs dataless JSON representation of canvas
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {String} json string
     */
    toDatalessJSON: function (propertiesToInclude) {
      return this.toDatalessObject(propertiesToInclude);
    },

    /**
     * Returns object representation of canvas
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function (propertiesToInclude) {
      return this._toObjectMethod('toObject', propertiesToInclude);
    },

    /**
     * Returns dataless object representation of canvas
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toDatalessObject: function (propertiesToInclude) {
      return this._toObjectMethod('toDatalessObject', propertiesToInclude);
    },

    /**
     * @private
     */
    _toObjectMethod: function (methodName, propertiesToInclude) {

      var data = {
        objects: this._toObjects(methodName, propertiesToInclude)
      };

      extend(data, this.__serializeBgOverlay());

      fabric.util.populateWithProperties(this, data, propertiesToInclude);

      return data;
    },

    /**
     * @private
     */
    _toObjects: function(methodName, propertiesToInclude) {
      return this.getObjects().filter(function(object) {
          return !object.excludeFromExport;
        }).map(function(instance) {
          return this._toObject(instance, methodName, propertiesToInclude);
        }, this);
    },

    /**
     * @private
     */
    _toObject: function(instance, methodName, propertiesToInclude) {
      var originalValue;

      if (!this.includeDefaultValues) {
        originalValue = instance.includeDefaultValues;
        instance.includeDefaultValues = false;
      }

      //If the object is part of the current selection group, it should
      //be transformed appropriately
      //i.e. it should be serialised as it would appear if the selection group
      //were to be destroyed.
      var originalProperties = this._realizeGroupTransformOnObject(instance),
          object = instance[methodName](propertiesToInclude);
      if (!this.includeDefaultValues) {
        instance.includeDefaultValues = originalValue;
      }

      //Undo the damage we did by changing all of its properties
      this._unwindGroupTransformOnObject(instance, originalProperties);

      return object;
    },

    /**
     * Realises an object's group transformation on it
     * @private
     * @param {fabric.Object} [instance] the object to transform (gets mutated)
     * @returns the original values of instance which were changed
     */
    _realizeGroupTransformOnObject: function(instance) {
      var layoutProps = ['angle', 'flipX', 'flipY', 'height', 'left', 'scaleX', 'scaleY', 'top', 'width'];
      if (instance.group && instance.group === this.getActiveGroup()) {
        //Copy all the positionally relevant properties across now
        var originalValues = {};
        layoutProps.forEach(function(prop) {
          originalValues[prop] = instance[prop];
        });
        this.getActiveGroup().realizeTransform(instance);
        return originalValues;
      }
      else {
        return null;
      }
    },

    /**
     * Restores the changed properties of instance
     * @private
     * @param {fabric.Object} [instance] the object to un-transform (gets mutated)
     * @param {Object} [originalValues] the original values of instance, as returned by _realizeGroupTransformOnObject
     */
    _unwindGroupTransformOnObject: function(instance, originalValues) {
      if (originalValues) {
        instance.set(originalValues);
      }
    },

    /**
     * @private
     */
    __serializeBgOverlay: function() {
      var data = {
        background: (this.backgroundColor && this.backgroundColor.toObject)
          ? this.backgroundColor.toObject()
          : this.backgroundColor
      };

      if (this.overlayColor) {
        data.overlay = this.overlayColor.toObject
          ? this.overlayColor.toObject()
          : this.overlayColor;
      }
      if (this.backgroundImage) {
        data.backgroundImage = this.backgroundImage.toObject();
      }
      if (this.overlayImage) {
        data.overlayImage = this.overlayImage.toObject();
      }

      return data;
    },

    /* _TO_SVG_START_ */
    /**
     * When true, getSvgTransform() will apply the StaticCanvas.viewportTransform to the SVG transformation. When true,
     * a zoomed canvas will then produce zoomed SVG output.
     * @type Boolean
     * @default
     */
    svgViewportTransformation: true,

    /**
     * Returns SVG representation of canvas
     * @function
     * @param {Object} [options] Options object for SVG output
     * @param {Boolean} [options.suppressPreamble=false] If true xml tag is not included
     * @param {Object} [options.viewBox] SVG viewbox object
     * @param {Number} [options.viewBox.x] x-cooridnate of viewbox
     * @param {Number} [options.viewBox.y] y-coordinate of viewbox
     * @param {Number} [options.viewBox.width] Width of viewbox
     * @param {Number} [options.viewBox.height] Height of viewbox
     * @param {String} [options.encoding=UTF-8] Encoding of SVG output
     * @param {String} [options.width] desired width of svg with or without units
     * @param {String} [options.height] desired height of svg with or without units
     * @param {Function} [reviver] Method for further parsing of svg elements, called after each fabric object converted into svg representation.
     * @return {String} SVG string
     * @tutorial {@link http://fabricjs.com/fabric-intro-part-3#serialization}
     * @see {@link http://jsfiddle.net/fabricjs/jQ3ZZ/|jsFiddle demo}
     * @example <caption>Normal SVG output</caption>
     * var svg = canvas.toSVG();
     * @example <caption>SVG output without preamble (without &lt;?xml ../>)</caption>
     * var svg = canvas.toSVG({suppressPreamble: true});
     * @example <caption>SVG output with viewBox attribute</caption>
     * var svg = canvas.toSVG({
     *   viewBox: {
     *     x: 100,
     *     y: 100,
     *     width: 200,
     *     height: 300
     *   }
     * });
     * @example <caption>SVG output with different encoding (default: UTF-8)</caption>
     * var svg = canvas.toSVG({encoding: 'ISO-8859-1'});
     * @example <caption>Modify SVG output with reviver function</caption>
     * var svg = canvas.toSVG(null, function(svg) {
     *   return svg.replace('stroke-dasharray: ; stroke-linecap: butt; stroke-linejoin: miter; stroke-miterlimit: 10; ', '');
     * });
     */
    toSVG: function(options, reviver) {
      options || (options = { });

      var markup = [];

      this._setSVGPreamble(markup, options);
      this._setSVGHeader(markup, options);

      this._setSVGBgOverlayColor(markup, 'backgroundColor');
      this._setSVGBgOverlayImage(markup, 'backgroundImage', reviver);

      this._setSVGObjects(markup, reviver);

      this._setSVGBgOverlayColor(markup, 'overlayColor');
      this._setSVGBgOverlayImage(markup, 'overlayImage', reviver);

      markup.push('</svg>');

      return markup.join('');
    },

    /**
     * @private
     */
    _setSVGPreamble: function(markup, options) {
      if (options.suppressPreamble) {
        return;
      }
      markup.push(
        '<?xml version="1.0" encoding="', (options.encoding || 'UTF-8'), '" standalone="no" ?>\n',
          '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" ',
            '"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n'
      );
    },

    /**
     * @private
     */
    _setSVGHeader: function(markup, options) {
      var width = options.width || this.width,
          height = options.height || this.height,
          vpt, viewBox = 'viewBox="0 0 ' + this.width + ' ' + this.height + '" ',
          NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS;

      if (options.viewBox) {
        viewBox = 'viewBox="' +
                options.viewBox.x + ' ' +
                options.viewBox.y + ' ' +
                options.viewBox.width + ' ' +
                options.viewBox.height + '" ';
      }
      else {
        if (this.svgViewportTransformation) {
          vpt = this.viewportTransform;
          viewBox = 'viewBox="' +
                  toFixed(-vpt[4] / vpt[0], NUM_FRACTION_DIGITS) + ' ' +
                  toFixed(-vpt[5] / vpt[3], NUM_FRACTION_DIGITS) + ' ' +
                  toFixed(this.width / vpt[0], NUM_FRACTION_DIGITS) + ' ' +
                  toFixed(this.height / vpt[3], NUM_FRACTION_DIGITS) + '" ';
        }
      }

      markup.push(
        '<svg ',
          'xmlns="http://www.w3.org/2000/svg" ',
          'xmlns:xlink="http://www.w3.org/1999/xlink" ',
          'version="1.1" ',
          'width="', width, '" ',
          'height="', height, '" ',
          (this.backgroundColor && !this.backgroundColor.toLive
            ? 'style="background-color: ' + this.backgroundColor + '" '
            : null),
          viewBox,
          'xml:space="preserve">\n',
        '<desc>Created with Fabric.js ', fabric.version, '</desc>\n',
        '<defs>',
          fabric.createSVGFontFacesMarkup(this.getObjects()),
          fabric.createSVGRefElementsMarkup(this),
        '</defs>\n'
      );
    },

    /**
     * @private
     */
    _setSVGObjects: function(markup, reviver) {
      var instance, originalProperties;
      for (var i = 0, objects = this.getObjects(), len = objects.length; i < len; i++) {
        instance = objects[i];
        if (instance.excludeFromExport) {
          continue;
        }
        //If the object is in a selection group, simulate what would happen to that
        //object when the group is deselected
        originalProperties = this._realizeGroupTransformOnObject(instance);
        markup.push(instance.toSVG(reviver));
        this._unwindGroupTransformOnObject(instance, originalProperties);
      }
    },

    /**
     * @private
     */
    _setSVGBgOverlayImage: function(markup, property, reviver) {
      if (this[property] && this[property].toSVG) {
        markup.push(this[property].toSVG(reviver));
      }
    },

    /**
     * @private
     */
    _setSVGBgOverlayColor: function(markup, property) {
      if (this[property] && this[property].source) {
        markup.push(
          '<rect x="', this[property].offsetX, '" y="', this[property].offsetY, '" ',
            'width="',
              (this[property].repeat === 'repeat-y' || this[property].repeat === 'no-repeat'
                ? this[property].source.width
                : this.width),
            '" height="',
              (this[property].repeat === 'repeat-x' || this[property].repeat === 'no-repeat'
                ? this[property].source.height
                : this.height),
            '" fill="url(#' + property + 'Pattern)"',
          '></rect>\n'
        );
      }
      else if (this[property] && property === 'overlayColor') {
        markup.push(
          '<rect x="0" y="0" ',
            'width="', this.width,
            '" height="', this.height,
            '" fill="', this[property], '"',
          '></rect>\n'
        );
      }
    },
    /* _TO_SVG_END_ */

    /**
     * Moves an object or the objects of a multiple selection
     * to the bottom of the stack of drawn objects
     * @param {fabric.Object} object Object to send to back
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    sendToBack: function (object) {
      if (!object) {
        return this;
      }
      var activeGroup = this.getActiveGroup ? this.getActiveGroup() : null,
          i, obj, objs;
      if (object === activeGroup) {
        objs = activeGroup._objects;
        for (i = objs.length; i--;) {
          obj = objs[i];
          removeFromArray(this._objects, obj);
          this._objects.unshift(obj);
        }
      }
      else {
        removeFromArray(this._objects, object);
        this._objects.unshift(object);
      }
      return this.renderAll && this.renderAll();
    },

    /**
     * Moves an object or the objects of a multiple selection
     * to the top of the stack of drawn objects
     * @param {fabric.Object} object Object to send
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    bringToFront: function (object) {
      if (!object) {
        return this;
      }
      var activeGroup = this.getActiveGroup ? this.getActiveGroup() : null,
          i, obj, objs;
      if (object === activeGroup) {
        objs = activeGroup._objects;
        for (i = 0; i < objs.length; i++) {
          obj = objs[i];
          removeFromArray(this._objects, obj);
          this._objects.push(obj);
        }
      }
      else {
        removeFromArray(this._objects, object);
        this._objects.push(object);
      }
      return this.renderAll && this.renderAll();
    },

    /**
     * Moves an object or a selection down in stack of drawn objects
     * @param {fabric.Object} object Object to send
     * @param {Boolean} [intersecting] If `true`, send object behind next lower intersecting object
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    sendBackwards: function (object, intersecting) {
      if (!object) {
        return this;
      }
      var activeGroup = this.getActiveGroup ? this.getActiveGroup() : null,
          i, obj, idx, newIdx, objs;

      if (object === activeGroup) {
        objs = activeGroup._objects;
        for (i = 0; i < objs.length; i++) {
          obj = objs[i];
          idx = this._objects.indexOf(obj);
          if (idx !== 0) {
            newIdx = idx - 1;
            removeFromArray(this._objects, obj);
            this._objects.splice(newIdx, 0, obj);
          }
        }
      }
      else {
        idx = this._objects.indexOf(object);
        if (idx !== 0) {
          // if object is not on the bottom of stack
          newIdx = this._findNewLowerIndex(object, idx, intersecting);
          removeFromArray(this._objects, object);
          this._objects.splice(newIdx, 0, object);
        }
      }
      this.renderAll && this.renderAll();
      return this;
    },

    /**
     * @private
     */
    _findNewLowerIndex: function(object, idx, intersecting) {
      var newIdx;

      if (intersecting) {
        newIdx = idx;

        // traverse down the stack looking for the nearest intersecting object
        for (var i = idx - 1; i >= 0; --i) {

          var isIntersecting = object.intersectsWithObject(this._objects[i]) ||
                               object.isContainedWithinObject(this._objects[i]) ||
                               this._objects[i].isContainedWithinObject(object);

          if (isIntersecting) {
            newIdx = i;
            break;
          }
        }
      }
      else {
        newIdx = idx - 1;
      }

      return newIdx;
    },

    /**
     * Moves an object or a selection up in stack of drawn objects
     * @param {fabric.Object} object Object to send
     * @param {Boolean} [intersecting] If `true`, send object in front of next upper intersecting object
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    bringForward: function (object, intersecting) {
      if (!object) {
        return this;
      }
      var activeGroup = this.getActiveGroup ? this.getActiveGroup() : null,
          i, obj, idx, newIdx, objs;

      if (object === activeGroup) {
        objs = activeGroup._objects;
        for (i = objs.length; i--;) {
          obj = objs[i];
          idx = this._objects.indexOf(obj);
          if (idx !== this._objects.length - 1) {
            newIdx = idx + 1;
            removeFromArray(this._objects, obj);
            this._objects.splice(newIdx, 0, obj);
          }
        }
      }
      else {
        idx = this._objects.indexOf(object);
        if (idx !== this._objects.length - 1) {
          // if object is not on top of stack (last item in an array)
          newIdx = this._findNewUpperIndex(object, idx, intersecting);
          removeFromArray(this._objects, object);
          this._objects.splice(newIdx, 0, object);
        }
      }
      this.renderAll && this.renderAll();
      return this;
    },

    /**
     * @private
     */
    _findNewUpperIndex: function(object, idx, intersecting) {
      var newIdx;

      if (intersecting) {
        newIdx = idx;

        // traverse up the stack looking for the nearest intersecting object
        for (var i = idx + 1; i < this._objects.length; ++i) {

          var isIntersecting = object.intersectsWithObject(this._objects[i]) ||
                               object.isContainedWithinObject(this._objects[i]) ||
                               this._objects[i].isContainedWithinObject(object);

          if (isIntersecting) {
            newIdx = i;
            break;
          }
        }
      }
      else {
        newIdx = idx + 1;
      }

      return newIdx;
    },

    /**
     * Moves an object to specified level in stack of drawn objects
     * @param {fabric.Object} object Object to send
     * @param {Number} index Position to move to
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    moveTo: function (object, index) {
      removeFromArray(this._objects, object);
      this._objects.splice(index, 0, object);
      return this.renderAll && this.renderAll();
    },

    /**
     * Clears a canvas element and removes all event listeners
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    dispose: function () {
      this.clear();
      return this;
    },

    /**
     * Returns a string representation of an instance
     * @return {String} string representation of an instance
     */
    toString: function () {
      return '#<fabric.Canvas (' + this.complexity() + '): ' +
               '{ objects: ' + this.getObjects().length + ' }>';
    }
  });

  extend(fabric.StaticCanvas.prototype, fabric.Observable);
  extend(fabric.StaticCanvas.prototype, fabric.Collection);
  extend(fabric.StaticCanvas.prototype, fabric.DataURLExporter);

  extend(fabric.StaticCanvas, /** @lends fabric.StaticCanvas */ {

    /**
     * @static
     * @type String
     * @default
     */
    EMPTY_JSON: '{"objects": [], "background": "white"}',

    /**
     * Provides a way to check support of some of the canvas methods
     * (either those of HTMLCanvasElement itself, or rendering context)
     *
     * @param {String} methodName Method to check support for;
     *                            Could be one of "getImageData", "toDataURL", "toDataURLWithQuality" or "setLineDash"
     * @return {Boolean | null} `true` if method is supported (or at least exists),
     *                          `null` if canvas element or context can not be initialized
     */
    supports: function (methodName) {
      var el = fabric.util.createCanvasElement();

      if (!el || !el.getContext) {
        return null;
      }

      var ctx = el.getContext('2d');
      if (!ctx) {
        return null;
      }

      switch (methodName) {

        case 'getImageData':
          return typeof ctx.getImageData !== 'undefined';

        case 'setLineDash':
          return typeof ctx.setLineDash !== 'undefined';

        case 'toDataURL':
          return typeof el.toDataURL !== 'undefined';

        case 'toDataURLWithQuality':
          try {
            el.toDataURL('image/jpeg', 0);
            return true;
          }
          catch (e) { }
          return false;

        default:
          return null;
      }
    }
  });

  /**
   * Returns JSON representation of canvas
   * @function
   * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
   * @return {String} JSON string
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-3#serialization}
   * @see {@link http://jsfiddle.net/fabricjs/pec86/|jsFiddle demo}
   * @example <caption>JSON without additional properties</caption>
   * var json = canvas.toJSON();
   * @example <caption>JSON with additional properties included</caption>
   * var json = canvas.toJSON(['lockMovementX', 'lockMovementY', 'lockRotation', 'lockScalingX', 'lockScalingY', 'lockUniScaling']);
   * @example <caption>JSON without default values</caption>
   * canvas.includeDefaultValues = false;
   * var json = canvas.toJSON();
   */
  fabric.StaticCanvas.prototype.toJSON = fabric.StaticCanvas.prototype.toObject;

})();


/**
 * BaseBrush class
 * @class fabric.BaseBrush
 * @see {@link http://fabricjs.com/freedrawing|Freedrawing demo}
 */
fabric.BaseBrush = fabric.util.createClass(/** @lends fabric.BaseBrush.prototype */ {

  /**
   * Color of a brush
   * @type String
   * @default
   */
  color:            'rgb(0, 0, 0)',

  /**
   * Width of a brush
   * @type Number
   * @default
   */
  width:            1,

  /**
   * Shadow object representing shadow of this shape.
   * <b>Backwards incompatibility note:</b> This property replaces "shadowColor" (String), "shadowOffsetX" (Number),
   * "shadowOffsetY" (Number) and "shadowBlur" (Number) since v1.2.12
   * @type fabric.Shadow
   * @default
   */
  shadow:           null,

  /**
   * Line endings style of a brush (one of "butt", "round", "square")
   * @type String
   * @default
   */
  strokeLineCap:    'round',

  /**
   * Corner style of a brush (one of "bevil", "round", "miter")
   * @type String
   * @default
   */
  strokeLineJoin:   'round',

  /**
   * Stroke Dash Array.
   * @type Array
   * @default
   */
  strokeDashArray:  null,

  /**
   * Sets shadow of an object
   * @param {Object|String} [options] Options object or string (e.g. "2px 2px 10px rgba(0,0,0,0.2)")
   * @return {fabric.Object} thisArg
   * @chainable
   */
  setShadow: function(options) {
    this.shadow = new fabric.Shadow(options);
    return this;
  },

  /**
   * Sets brush styles
   * @private
   */
  _setBrushStyles: function() {
    var ctx = this.canvas.contextTop;

    ctx.strokeStyle = this.color;
    ctx.lineWidth = this.width;
    ctx.lineCap = this.strokeLineCap;
    ctx.lineJoin = this.strokeLineJoin;
    if (this.strokeDashArray && fabric.StaticCanvas.supports('setLineDash')) {
      ctx.setLineDash(this.strokeDashArray);
    }
  },

  /**
   * Sets brush shadow styles
   * @private
   */
  _setShadow: function() {
    if (!this.shadow) {
      return;
    }

    var ctx = this.canvas.contextTop;

    ctx.shadowColor = this.shadow.color;
    ctx.shadowBlur = this.shadow.blur;
    ctx.shadowOffsetX = this.shadow.offsetX;
    ctx.shadowOffsetY = this.shadow.offsetY;
  },

  /**
   * Removes brush shadow styles
   * @private
   */
  _resetShadow: function() {
    var ctx = this.canvas.contextTop;

    ctx.shadowColor = '';
    ctx.shadowBlur = ctx.shadowOffsetX = ctx.shadowOffsetY = 0;
  }
});


(function() {

  /**
   * PencilBrush class
   * @class fabric.PencilBrush
   * @extends fabric.BaseBrush
   */
  fabric.PencilBrush = fabric.util.createClass(fabric.BaseBrush, /** @lends fabric.PencilBrush.prototype */ {

    /**
     * Constructor
     * @param {fabric.Canvas} canvas
     * @return {fabric.PencilBrush} Instance of a pencil brush
     */
    initialize: function(canvas) {
      this.canvas = canvas;
      this._points = [ ];
    },

    /**
     * Inovoked on mouse down
     * @param {Object} pointer
     */
    onMouseDown: function(pointer) {
      this._prepareForDrawing(pointer);
      // capture coordinates immediately
      // this allows to draw dots (when movement never occurs)
      this._captureDrawingPath(pointer);
      this._render();
    },

    /**
     * Inovoked on mouse move
     * @param {Object} pointer
     */
    onMouseMove: function(pointer) {
      this._captureDrawingPath(pointer);
      // redraw curve
      // clear top canvas
      this.canvas.clearContext(this.canvas.contextTop);
      this._render();
    },

    /**
     * Invoked on mouse up
     */
    onMouseUp: function() {
      this._finalizeAndAddPath();
    },

    /**
     * @private
     * @param {Object} pointer Actual mouse position related to the canvas.
     */
    _prepareForDrawing: function(pointer) {

      var p = new fabric.Point(pointer.x, pointer.y);

      this._reset();
      this._addPoint(p);

      this.canvas.contextTop.moveTo(p.x, p.y);
    },

    /**
     * @private
     * @param {fabric.Point} point Point to be added to points array
     */
    _addPoint: function(point) {
      this._points.push(point);
    },

    /**
     * Clear points array and set contextTop canvas style.
     * @private
     */
    _reset: function() {
      this._points.length = 0;

      this._setBrushStyles();
      this._setShadow();
    },

    /**
     * @private
     * @param {Object} pointer Actual mouse position related to the canvas.
     */
    _captureDrawingPath: function(pointer) {
      var pointerPoint = new fabric.Point(pointer.x, pointer.y);
      this._addPoint(pointerPoint);
    },

    /**
     * Draw a smooth path on the topCanvas using quadraticCurveTo
     * @private
     */
    _render: function() {
      var ctx  = this.canvas.contextTop,
          v = this.canvas.viewportTransform,
          p1 = this._points[0],
          p2 = this._points[1];

      ctx.save();
      ctx.transform(v[0], v[1], v[2], v[3], v[4], v[5]);
      ctx.beginPath();

      //if we only have 2 points in the path and they are the same
      //it means that the user only clicked the canvas without moving the mouse
      //then we should be drawing a dot. A path isn't drawn between two identical dots
      //that's why we set them apart a bit
      if (this._points.length === 2 && p1.x === p2.x && p1.y === p2.y) {
        p1.x -= 0.5;
        p2.x += 0.5;
      }
      ctx.moveTo(p1.x, p1.y);

      for (var i = 1, len = this._points.length; i < len; i++) {
        // we pick the point between pi + 1 & pi + 2 as the
        // end point and p1 as our control point.
        var midPoint = p1.midPointFrom(p2);
        ctx.quadraticCurveTo(p1.x, p1.y, midPoint.x, midPoint.y);

        p1 = this._points[i];
        p2 = this._points[i + 1];
      }
      // Draw last line as a straight line while
      // we wait for the next point to be able to calculate
      // the bezier control point
      ctx.lineTo(p1.x, p1.y);
      ctx.stroke();
      ctx.restore();
    },

    /**
     * Converts points to SVG path
     * @param {Array} points Array of points
     * @param {Number} minX
     * @param {Number} minY
     * @return {String} SVG path
     */
    convertPointsToSVGPath: function(points) {
      var path = [],
          p1 = new fabric.Point(points[0].x, points[0].y),
          p2 = new fabric.Point(points[1].x, points[1].y);

      path.push('M ', points[0].x, ' ', points[0].y, ' ');
      for (var i = 1, len = points.length; i < len; i++) {
        var midPoint = p1.midPointFrom(p2);
        // p1 is our bezier control point
        // midpoint is our endpoint
        // start point is p(i-1) value.
        path.push('Q ', p1.x, ' ', p1.y, ' ', midPoint.x, ' ', midPoint.y, ' ');
        p1 = new fabric.Point(points[i].x, points[i].y);
        if ((i + 1) < points.length) {
          p2 = new fabric.Point(points[i + 1].x, points[i + 1].y);
        }
      }
      path.push('L ', p1.x, ' ', p1.y, ' ');
      return path;
    },

    /**
     * Creates fabric.Path object to add on canvas
     * @param {String} pathData Path data
     * @return {fabric.Path} Path to add on canvas
     */
    createPath: function(pathData) {
      var path = new fabric.Path(pathData, {
                   fill: null,
                   stroke: this.color,
                   strokeWidth: this.width,
                   strokeLineCap: this.strokeLineCap,
                   strokeLineJoin: this.strokeLineJoin,
                   strokeDashArray: this.strokeDashArray,
                   originX: 'center',
                   originY: 'center'
                 });

      if (this.shadow) {
        this.shadow.affectStroke = true;
        path.setShadow(this.shadow);
      }

      return path;
    },

    /**
     * On mouseup after drawing the path on contextTop canvas
     * we use the points captured to create an new fabric path object
     * and add it to the fabric canvas.
     */
    _finalizeAndAddPath: function() {
      var ctx = this.canvas.contextTop;
      ctx.closePath();

      var pathData = this.convertPointsToSVGPath(this._points).join('');
      if (pathData === 'M 0 0 Q 0 0 0 0 L 0 0') {
        // do not create 0 width/height paths, as they are
        // rendered inconsistently across browsers
        // Firefox 4, for example, renders a dot,
        // whereas Chrome 10 renders nothing
        this.canvas.renderAll();
        return;
      }

      var path = this.createPath(pathData);

      this.canvas.add(path);
      path.setCoords();

      this.canvas.clearContext(this.canvas.contextTop);
      this._resetShadow();
      this.canvas.renderAll();

      // fire event 'path' created
      this.canvas.fire('path:created', { path: path });
    }
  });
})();


/**
 * CircleBrush class
 * @class fabric.CircleBrush
 */
fabric.CircleBrush = fabric.util.createClass(fabric.BaseBrush, /** @lends fabric.CircleBrush.prototype */ {

  /**
   * Width of a brush
   * @type Number
   * @default
   */
  width: 10,

  /**
   * Constructor
   * @param {fabric.Canvas} canvas
   * @return {fabric.CircleBrush} Instance of a circle brush
   */
  initialize: function(canvas) {
    this.canvas = canvas;
    this.points = [ ];
  },

  /**
   * Invoked inside on mouse down and mouse move
   * @param {Object} pointer
   */
  drawDot: function(pointer) {
    var point = this.addPoint(pointer),
        ctx = this.canvas.contextTop,
        v = this.canvas.viewportTransform;
    ctx.save();
    ctx.transform(v[0], v[1], v[2], v[3], v[4], v[5]);

    ctx.fillStyle = point.fill;
    ctx.beginPath();
    ctx.arc(point.x, point.y, point.radius, 0, Math.PI * 2, false);
    ctx.closePath();
    ctx.fill();

    ctx.restore();
  },

  /**
   * Invoked on mouse down
   */
  onMouseDown: function(pointer) {
    this.points.length = 0;
    this.canvas.clearContext(this.canvas.contextTop);
    this._setShadow();
    this.drawDot(pointer);
  },

  /**
   * Invoked on mouse move
   * @param {Object} pointer
   */
  onMouseMove: function(pointer) {
    this.drawDot(pointer);
  },

  /**
   * Invoked on mouse up
   */
  onMouseUp: function() {
    var originalRenderOnAddRemove = this.canvas.renderOnAddRemove;
    this.canvas.renderOnAddRemove = false;

    var circles = [ ];

    for (var i = 0, len = this.points.length; i < len; i++) {
      var point = this.points[i],
          circle = new fabric.Circle({
            radius: point.radius,
            left: point.x,
            top: point.y,
            originX: 'center',
            originY: 'center',
            fill: point.fill
          });

      this.shadow && circle.setShadow(this.shadow);

      circles.push(circle);
    }
    var group = new fabric.Group(circles, { originX: 'center', originY: 'center' });
    group.canvas = this.canvas;

    this.canvas.add(group);
    this.canvas.fire('path:created', { path: group });

    this.canvas.clearContext(this.canvas.contextTop);
    this._resetShadow();
    this.canvas.renderOnAddRemove = originalRenderOnAddRemove;
    this.canvas.renderAll();
  },

  /**
   * @param {Object} pointer
   * @return {fabric.Point} Just added pointer point
   */
  addPoint: function(pointer) {
    var pointerPoint = new fabric.Point(pointer.x, pointer.y),

        circleRadius = fabric.util.getRandomInt(
                        Math.max(0, this.width - 20), this.width + 20) / 2,

        circleColor = new fabric.Color(this.color)
                        .setAlpha(fabric.util.getRandomInt(0, 100) / 100)
                        .toRgba();

    pointerPoint.radius = circleRadius;
    pointerPoint.fill = circleColor;

    this.points.push(pointerPoint);

    return pointerPoint;
  }
});


/**
 * SprayBrush class
 * @class fabric.SprayBrush
 */
fabric.SprayBrush = fabric.util.createClass( fabric.BaseBrush, /** @lends fabric.SprayBrush.prototype */ {

  /**
   * Width of a spray
   * @type Number
   * @default
   */
  width:              10,

  /**
   * Density of a spray (number of dots per chunk)
   * @type Number
   * @default
   */
  density:            20,

  /**
   * Width of spray dots
   * @type Number
   * @default
   */
  dotWidth:           1,

  /**
   * Width variance of spray dots
   * @type Number
   * @default
   */
  dotWidthVariance:   1,

  /**
   * Whether opacity of a dot should be random
   * @type Boolean
   * @default
   */
  randomOpacity:        false,

  /**
   * Whether overlapping dots (rectangles) should be removed (for performance reasons)
   * @type Boolean
   * @default
   */
  optimizeOverlapping:  true,

  /**
   * Constructor
   * @param {fabric.Canvas} canvas
   * @return {fabric.SprayBrush} Instance of a spray brush
   */
  initialize: function(canvas) {
    this.canvas = canvas;
    this.sprayChunks = [ ];
  },

  /**
   * Invoked on mouse down
   * @param {Object} pointer
   */
  onMouseDown: function(pointer) {
    this.sprayChunks.length = 0;
    this.canvas.clearContext(this.canvas.contextTop);
    this._setShadow();

    this.addSprayChunk(pointer);
    this.render();
  },

  /**
   * Invoked on mouse move
   * @param {Object} pointer
   */
  onMouseMove: function(pointer) {
    this.addSprayChunk(pointer);
    this.render();
  },

  /**
   * Invoked on mouse up
   */
  onMouseUp: function() {
    var originalRenderOnAddRemove = this.canvas.renderOnAddRemove;
    this.canvas.renderOnAddRemove = false;

    var rects = [ ];

    for (var i = 0, ilen = this.sprayChunks.length; i < ilen; i++) {
      var sprayChunk = this.sprayChunks[i];

      for (var j = 0, jlen = sprayChunk.length; j < jlen; j++) {

        var rect = new fabric.Rect({
          width: sprayChunk[j].width,
          height: sprayChunk[j].width,
          left: sprayChunk[j].x + 1,
          top: sprayChunk[j].y + 1,
          originX: 'center',
          originY: 'center',
          fill: this.color
        });

        this.shadow && rect.setShadow(this.shadow);
        rects.push(rect);
      }
    }

    if (this.optimizeOverlapping) {
      rects = this._getOptimizedRects(rects);
    }

    var group = new fabric.Group(rects, { originX: 'center', originY: 'center' });
    group.canvas = this.canvas;

    this.canvas.add(group);
    this.canvas.fire('path:created', { path: group });

    this.canvas.clearContext(this.canvas.contextTop);
    this._resetShadow();
    this.canvas.renderOnAddRemove = originalRenderOnAddRemove;
    this.canvas.renderAll();
  },

  /**
   * @private
   * @param {Array} rects
   */
  _getOptimizedRects: function(rects) {

    // avoid creating duplicate rects at the same coordinates
    var uniqueRects = { }, key;

    for (var i = 0, len = rects.length; i < len; i++) {
      key = rects[i].left + '' + rects[i].top;
      if (!uniqueRects[key]) {
        uniqueRects[key] = rects[i];
      }
    }
    var uniqueRectsArray = [ ];
    for (key in uniqueRects) {
      uniqueRectsArray.push(uniqueRects[key]);
    }

    return uniqueRectsArray;
  },

  /**
   * Renders brush
   */
  render: function() {
    var ctx = this.canvas.contextTop;
    ctx.fillStyle = this.color;

    var v = this.canvas.viewportTransform;
    ctx.save();
    ctx.transform(v[0], v[1], v[2], v[3], v[4], v[5]);

    for (var i = 0, len = this.sprayChunkPoints.length; i < len; i++) {
      var point = this.sprayChunkPoints[i];
      if (typeof point.opacity !== 'undefined') {
        ctx.globalAlpha = point.opacity;
      }
      ctx.fillRect(point.x, point.y, point.width, point.width);
    }
    ctx.restore();
  },

  /**
   * @param {Object} pointer
   */
  addSprayChunk: function(pointer) {
    this.sprayChunkPoints = [ ];

    var x, y, width, radius = this.width / 2;

    for (var i = 0; i < this.density; i++) {

      x = fabric.util.getRandomInt(pointer.x - radius, pointer.x + radius);
      y = fabric.util.getRandomInt(pointer.y - radius, pointer.y + radius);

      if (this.dotWidthVariance) {
        width = fabric.util.getRandomInt(
          // bottom clamp width to 1
          Math.max(1, this.dotWidth - this.dotWidthVariance),
          this.dotWidth + this.dotWidthVariance);
      }
      else {
        width = this.dotWidth;
      }

      var point = new fabric.Point(x, y);
      point.width = width;

      if (this.randomOpacity) {
        point.opacity = fabric.util.getRandomInt(0, 100) / 100;
      }

      this.sprayChunkPoints.push(point);
    }

    this.sprayChunks.push(this.sprayChunkPoints);
  }
});


/**
 * PatternBrush class
 * @class fabric.PatternBrush
 * @extends fabric.BaseBrush
 */
fabric.PatternBrush = fabric.util.createClass(fabric.PencilBrush, /** @lends fabric.PatternBrush.prototype */ {

  getPatternSrc: function() {

    var dotWidth = 20,
        dotDistance = 5,
        patternCanvas = fabric.document.createElement('canvas'),
        patternCtx = patternCanvas.getContext('2d');

    patternCanvas.width = patternCanvas.height = dotWidth + dotDistance;

    patternCtx.fillStyle = this.color;
    patternCtx.beginPath();
    patternCtx.arc(dotWidth / 2, dotWidth / 2, dotWidth / 2, 0, Math.PI * 2, false);
    patternCtx.closePath();
    patternCtx.fill();

    return patternCanvas;
  },

  getPatternSrcFunction: function() {
    return String(this.getPatternSrc).replace('this.color', '"' + this.color + '"');
  },

  /**
   * Creates "pattern" instance property
   */
  getPattern: function() {
    return this.canvas.contextTop.createPattern(this.source || this.getPatternSrc(), 'repeat');
  },

  /**
   * Sets brush styles
   */
  _setBrushStyles: function() {
    this.callSuper('_setBrushStyles');
    this.canvas.contextTop.strokeStyle = this.getPattern();
  },

  /**
   * Creates path
   */
  createPath: function(pathData) {
    var path = this.callSuper('createPath', pathData);
    path.stroke = new fabric.Pattern({
      source: this.source || this.getPatternSrcFunction()
    });
    return path;
  }
});


(function() {

  var getPointer = fabric.util.getPointer,
      degreesToRadians = fabric.util.degreesToRadians,
      radiansToDegrees = fabric.util.radiansToDegrees,
      atan2 = Math.atan2,
      abs = Math.abs,

      STROKE_OFFSET = 0.5;

  /**
   * Canvas class
   * @class fabric.Canvas
   * @extends fabric.StaticCanvas
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-1#canvas}
   * @see {@link fabric.Canvas#initialize} for constructor definition
   *
   * @fires object:modified
   * @fires object:rotating
   * @fires object:scaling
   * @fires object:moving
   * @fires object:selected
   *
   * @fires before:selection:cleared
   * @fires selection:cleared
   * @fires selection:created
   *
   * @fires path:created
   * @fires mouse:down
   * @fires mouse:move
   * @fires mouse:up
   * @fires mouse:over
   * @fires mouse:out
   *
   */
  fabric.Canvas = fabric.util.createClass(fabric.StaticCanvas, /** @lends fabric.Canvas.prototype */ {

    /**
     * Constructor
     * @param {HTMLElement | String} el &lt;canvas> element to initialize instance on
     * @param {Object} [options] Options object
     * @return {Object} thisArg
     */
    initialize: function(el, options) {
      options || (options = { });

      this._initStatic(el, options);
      this._initInteractive();
      this._createCacheCanvas();
    },

    /**
     * When true, objects can be transformed by one side (unproportionally)
     * @type Boolean
     * @default
     */
    uniScaleTransform:      false,

    /**
     * Indicates which key enable unproportional scaling
     * values: altKey, shiftKey, ctrlKey
     * @since 1.6.2
     * @type String
     * @default
     */
    uniScaleKey:           'shiftKey',

    /**
     * When true, objects use center point as the origin of scale transformation.
     * <b>Backwards incompatibility note:</b> This property replaces "centerTransform" (Boolean).
     * @since 1.3.4
     * @type Boolean
     * @default
     */
    centeredScaling:        false,

    /**
     * When true, objects use center point as the origin of rotate transformation.
     * <b>Backwards incompatibility note:</b> This property replaces "centerTransform" (Boolean).
     * @since 1.3.4
     * @type Boolean
     * @default
     */
    centeredRotation:       false,

    /**
     * Indicates which key enable centered Transfrom
     * values: altKey, shiftKey, ctrlKey
     * @since 1.6.2
     * @type String
     * @default
     */
    centeredKey:           'altKey',

    /**
     * Indicates which key enable alternate action on corner
     * values: altKey, shiftKey, ctrlKey
     * @since 1.6.2
     * @type String
     * @default
     */
    altActionKey:           'shiftKey',

    /**
     * Indicates which key enable last rendered selection independently of stack position
     * values: altKey, shiftKey, ctrlKey
     * @since 1.6.3
     * @type String
     * @default
     */
    lastRenderedKey:        'altKey',

    /**
     * Indicates that canvas is interactive. This property should not be changed.
     * @type Boolean
     * @default
     */
    interactive:            true,

    /**
     * Indicates whether group selection should be enabled
     * @type Boolean
     * @default
     */
    selection:              true,

    /**
     * Indicates which key enable multiple click selection
     * values: altKey, shiftKey, ctrlKey
     * @since 1.6.2
     * @type String
     * @default
     */
    selectionKey:           'shiftKey',

    /**
     * Color of selection
     * @type String
     * @default
     */
    selectionColor:         'rgba(100, 100, 255, 0.3)', // blue

    /**
     * Default dash array pattern
     * If not empty the selection border is dashed
     * @type Array
     */
    selectionDashArray:     [ ],

    /**
     * Color of the border of selection (usually slightly darker than color of selection itself)
     * @type String
     * @default
     */
    selectionBorderColor:   'rgba(255, 255, 255, 0.3)',

    /**
     * Width of a line used in object/group selection
     * @type Number
     * @default
     */
    selectionLineWidth:     1,

    /**
     * Default cursor value used when hovering over an object on canvas
     * @type String
     * @default
     */
    hoverCursor:            'move',

    /**
     * Default cursor value used when moving an object on canvas
     * @type String
     * @default
     */
    moveCursor:             'move',

    /**
     * Default cursor value used for the entire canvas
     * @type String
     * @default
     */
    defaultCursor:          'default',

    /**
     * Cursor value used during free drawing
     * @type String
     * @default
     */
    freeDrawingCursor:      'crosshair',

    /**
     * Cursor value used for rotation point
     * @type String
     * @default
     */
    rotationCursor:         'crosshair',

    /**
     * Default element class that's given to wrapper (div) element of canvas
     * @type String
     * @default
     */
    containerClass:         'canvas-container',

    /**
     * When true, object detection happens on per-pixel basis rather than on per-bounding-box
     * @type Boolean
     * @default
     */
    perPixelTargetFind:     false,

    /**
     * Number of pixels around target pixel to tolerate (consider active) during object detection
     * @type Number
     * @default
     */
    targetFindTolerance:    0,

    /**
     * When true, target detection is skipped when hovering over canvas. This can be used to improve performance.
     * @type Boolean
     * @default
     */
    skipTargetFind:         false,

    /**
     * When true, mouse events on canvas (mousedown/mousemove/mouseup) result in free drawing.
     * After mousedown, mousemove creates a shape,
     * and then mouseup finalizes it and adds an instance of `fabric.Path` onto canvas.
     * @tutorial {@link http://fabricjs.com/fabric-intro-part-4#free_drawing}
     * @type Boolean
     * @default
     */
    isDrawingMode:          false,

    /**
     * @private
     */
    _initInteractive: function() {
      this._currentTransform = null;
      this._groupSelector = null;
      this._initWrapperElement();
      this._createUpperCanvas();
      this._initEventListeners();

      this._initRetinaScaling();

      this.freeDrawingBrush = fabric.PencilBrush && new fabric.PencilBrush(this);

      this.calcOffset();
    },

    /**
     * Resets the current transform to its original values and chooses the type of resizing based on the event
     * @private
     * @param {Event} e Event object fired on mousemove
     */
    _resetCurrentTransform: function() {
      var t = this._currentTransform;

      t.target.set({
        scaleX: t.original.scaleX,
        scaleY: t.original.scaleY,
        skewX: t.original.skewX,
        skewY: t.original.skewY,
        left: t.original.left,
        top: t.original.top
      });

      if (this._shouldCenterTransform(t.target)) {
        if (t.action === 'rotate') {
          this._setOriginToCenter(t.target);
        }
        else {
          if (t.originX !== 'center') {
            if (t.originX === 'right') {
              t.mouseXSign = -1;
            }
            else {
              t.mouseXSign = 1;
            }
          }
          if (t.originY !== 'center') {
            if (t.originY === 'bottom') {
              t.mouseYSign = -1;
            }
            else {
              t.mouseYSign = 1;
            }
          }

          t.originX = 'center';
          t.originY = 'center';
        }
      }
      else {
        t.originX = t.original.originX;
        t.originY = t.original.originY;
      }
    },

    /**
     * Checks if point is contained within an area of given object
     * @param {Event} e Event object
     * @param {fabric.Object} target Object to test against
     * @param {Object} [point] x,y object of point coordinates we want to check.
     * @return {Boolean} true if point is contained within an area of given object
     */
    containsPoint: function (e, target, point) {
      var pointer = point || this.getPointer(e, true),
          xy = this._normalizePointer(target, pointer);

      if (target.group && target.group === this.getActiveGroup()) {
        xy = this._normalizePointer(target.group, pointer);
      }
      else {
        xy = { x: pointer.x, y: pointer.y };
      }
      // http://www.geog.ubc.ca/courses/klink/gis.notes/ncgia/u32.html
      // http://idav.ucdavis.edu/~okreylos/TAship/Spring2000/PointInPolygon.html
      return (target.containsPoint(xy) || target._findTargetCorner(pointer));
    },

    /**
     * @private
     */
    _normalizePointer: function (object, pointer) {
      var lt, m;

      m = fabric.util.multiplyTransformMatrices(
            this.viewportTransform,
            object.calcTransformMatrix());

      m = fabric.util.invertTransform(m);
      pointer = fabric.util.transformPoint(pointer, m , false);
      lt = fabric.util.transformPoint(object.getCenterPoint(), m , false);
      pointer.x -= lt.x;
      pointer.y -= lt.y;
      return { x: pointer.x, y: pointer.y };
    },

    /**
     * Returns true if object is transparent at a certain location
     * @param {fabric.Object} target Object to check
     * @param {Number} x Left coordinate
     * @param {Number} y Top coordinate
     * @return {Boolean}
     */
    isTargetTransparent: function (target, x, y) {
      var hasBorders = target.hasBorders,
          transparentCorners = target.transparentCorners,
          ctx = this.contextCache,
          shouldTransform = target.group && target.group === this.getActiveGroup();

      target.hasBorders = target.transparentCorners = false;

      ctx.save();
      ctx.transform.apply(ctx, this.viewportTransform);
      if (shouldTransform) {
        ctx.transform.apply(ctx, target.group.calcTransformMatrix());
      }
      target.render(ctx);
      ctx.restore();

      target.active && target._renderControls(ctx);

      target.hasBorders = hasBorders;
      target.transparentCorners = transparentCorners;

      var isTransparent = fabric.util.isTransparent(
        ctx, x, y, this.targetFindTolerance);

      this.clearContext(ctx);

      return isTransparent;
    },

    /**
     * @private
     * @param {Event} e Event object
     * @param {fabric.Object} target
     */
    _shouldClearSelection: function (e, target) {
      var activeGroup = this.getActiveGroup(),
          activeObject = this.getActiveObject();

      return (
        !target
        ||
        (target &&
          activeGroup &&
          !activeGroup.contains(target) &&
          activeGroup !== target &&
          !e[this.selectionKey])
        ||
        (target && !target.evented)
        ||
        (target &&
          !target.selectable &&
          activeObject &&
          activeObject !== target)
      );
    },

    /**
     * @private
     * @param {fabric.Object} target
     */
    _shouldCenterTransform: function (target) {
      if (!target) {
        return;
      }

      var t = this._currentTransform,
          centerTransform;

      if (t.action === 'scale' || t.action === 'scaleX' || t.action === 'scaleY') {
        centerTransform = this.centeredScaling || target.centeredScaling;
      }
      else if (t.action === 'rotate') {
        centerTransform = this.centeredRotation || target.centeredRotation;
      }

      return centerTransform ? !t.altKey : t.altKey;
    },

    /**
     * @private
     */
    _getOriginFromCorner: function(target, corner) {
      var origin = {
        x: target.originX,
        y: target.originY
      };

      if (corner === 'ml' || corner === 'tl' || corner === 'bl') {
        origin.x = 'right';
      }
      else if (corner === 'mr' || corner === 'tr' || corner === 'br') {
        origin.x = 'left';
      }

      if (corner === 'tl' || corner === 'mt' || corner === 'tr') {
        origin.y = 'bottom';
      }
      else if (corner === 'bl' || corner === 'mb' || corner === 'br') {
        origin.y = 'top';
      }

      return origin;
    },

    /**
     * @private
     */
    _getActionFromCorner: function(target, corner, e) {
      if (!corner) {
        return 'drag';
      }

      switch (corner) {
        case 'mtr':
          return 'rotate';
        case 'ml':
        case 'mr':
          return e[this.altActionKey] ? 'skewY' : 'scaleX';
        case 'mt':
        case 'mb':
          return e[this.altActionKey] ? 'skewX' : 'scaleY';
        default:
          return 'scale';
      }
    },

    /**
     * @private
     * @param {Event} e Event object
     * @param {fabric.Object} target
     */
    _setupCurrentTransform: function (e, target) {
      if (!target) {
        return;
      }

      var pointer = this.getPointer(e),
          corner = target._findTargetCorner(this.getPointer(e, true)),
          action = this._getActionFromCorner(target, corner, e),
          origin = this._getOriginFromCorner(target, corner);

      this._currentTransform = {
        target: target,
        action: action,
        corner: corner,
        scaleX: target.scaleX,
        scaleY: target.scaleY,
        skewX: target.skewX,
        skewY: target.skewY,
        offsetX: pointer.x - target.left,
        offsetY: pointer.y - target.top,
        originX: origin.x,
        originY: origin.y,
        ex: pointer.x,
        ey: pointer.y,
        lastX: pointer.x,
        lastY: pointer.y,
        left: target.left,
        top: target.top,
        theta: degreesToRadians(target.angle),
        width: target.width * target.scaleX,
        mouseXSign: 1,
        mouseYSign: 1,
        shiftKey: e.shiftKey,
        altKey: e[this.centeredKey]
      };

      this._currentTransform.original = {
        left: target.left,
        top: target.top,
        scaleX: target.scaleX,
        scaleY: target.scaleY,
        skewX: target.skewX,
        skewY: target.skewY,
        originX: origin.x,
        originY: origin.y
      };

      this._resetCurrentTransform();
    },

    /**
     * Translates object by "setting" its left/top
     * @private
     * @param {Number} x pointer's x coordinate
     * @param {Number} y pointer's y coordinate
     * @return {Boolean} true if the translation occurred
     */
    _translateObject: function (x, y) {
      var transform = this._currentTransform,
          target = transform.target,
          newLeft = x - transform.offsetX,
          newTop = y - transform.offsetY,
          moveX = !target.get('lockMovementX') && target.left !== newLeft,
          moveY = !target.get('lockMovementY') && target.top !== newTop;

      moveX && target.set('left', newLeft);
      moveY && target.set('top', newTop);
      return moveX || moveY;
    },

    /**
     * Check if we are increasing a positive skew or lower it,
     * checking mouse direction and pressed corner.
     * @private
     */
    _changeSkewTransformOrigin: function(mouseMove, t, by) {
      var property = 'originX', origins = { 0: 'center' },
          skew = t.target.skewX, originA = 'left', originB = 'right',
          corner = t.corner === 'mt' || t.corner === 'ml' ? 1 : -1,
          flipSign = 1;

      mouseMove = mouseMove > 0 ? 1 : -1;
      if (by === 'y') {
        skew = t.target.skewY;
        originA = 'top';
        originB = 'bottom';
        property = 'originY';
      }
      origins[-1] = originA;
      origins[1] = originB;

      t.target.flipX && (flipSign *= -1);
      t.target.flipY && (flipSign *= -1);

      if (skew === 0) {
        t.skewSign = -corner * mouseMove * flipSign;
        t[property] = origins[-mouseMove];
      }
      else {
        skew = skew > 0 ? 1 : -1;
        t.skewSign = skew;
        t[property] = origins[skew * corner * flipSign];
      }
    },

    /**
     * Skew object by mouse events
     * @private
     * @param {Number} x pointer's x coordinate
     * @param {Number} y pointer's y coordinate
     * @param {String} by Either 'x' or 'y'
     * @return {Boolean} true if the skewing occurred
     */
    _skewObject: function (x, y, by) {
      var t = this._currentTransform,
          target = t.target, skewed = false,
          lockSkewingX = target.get('lockSkewingX'),
          lockSkewingY = target.get('lockSkewingY');

      if ((lockSkewingX && by === 'x') || (lockSkewingY && by === 'y')) {
        return false;
      }

      // Get the constraint point
      var center = target.getCenterPoint(),
          actualMouseByCenter = target.toLocalPoint(new fabric.Point(x, y), 'center', 'center')[by],
          lastMouseByCenter = target.toLocalPoint(new fabric.Point(t.lastX, t.lastY), 'center', 'center')[by],
          actualMouseByOrigin, constraintPosition, dim = target._getTransformedDimensions();

      this._changeSkewTransformOrigin(actualMouseByCenter - lastMouseByCenter, t, by);
      actualMouseByOrigin = target.toLocalPoint(new fabric.Point(x, y), t.originX, t.originY)[by],

      constraintPosition = target.translateToOriginPoint(center, t.originX, t.originY);
      // Actually skew the object
      skewed = this._setObjectSkew(actualMouseByOrigin, t, by, dim);
      t.lastX = x;
      t.lastY = y;
      // Make sure the constraints apply
      target.setPositionByOrigin(constraintPosition, t.originX, t.originY);
      return skewed;
    },

    /**
     * Set object skew
     * @private
     * @return {Boolean} true if the skewing occurred
     */
    _setObjectSkew: function(localMouse, transform, by, _dim) {
      var target = transform.target, newValue, skewed = false,
          skewSign = transform.skewSign, newDim, dimNoSkew,
          otherBy, _otherBy, _by, newDimMouse, skewX, skewY;

      if (by === 'x') {
        otherBy = 'y';
        _otherBy = 'Y';
        _by = 'X';
        skewX = 0;
        skewY = target.skewY;
      }
      else {
        otherBy = 'x';
        _otherBy = 'X';
        _by = 'Y';
        skewX = target.skewX;
        skewY = 0;
      }

      dimNoSkew = target._getTransformedDimensions(skewX, skewY);
      newDimMouse = 2 * Math.abs(localMouse) - dimNoSkew[by];
      if (newDimMouse <= 2) {
        newValue = 0;
      }
      else {
        newValue = skewSign * Math.atan((newDimMouse / target['scale' + _by]) /
                                        (dimNoSkew[otherBy] / target['scale' + _otherBy]));
        newValue = fabric.util.radiansToDegrees(newValue);
      }
      skewed = target['skew' + _by] !== newValue;
      target.set('skew' + _by, newValue);
      if (target['skew' + _otherBy] !== 0) {
        newDim = target._getTransformedDimensions();
        newValue = (_dim[otherBy] / newDim[otherBy]) * target['scale' + _otherBy];
        target.set('scale' + _otherBy, newValue);
      }
      return skewed;
    },

    /**
     * Scales object by invoking its scaleX/scaleY methods
     * @private
     * @param {Number} x pointer's x coordinate
     * @param {Number} y pointer's y coordinate
     * @param {String} by Either 'x' or 'y' - specifies dimension constraint by which to scale an object.
     *                    When not provided, an object is scaled by both dimensions equally
     * @return {Boolean} true if the scaling occurred
     */
    _scaleObject: function (x, y, by) {
      var t = this._currentTransform,
          target = t.target,
          lockScalingX = target.get('lockScalingX'),
          lockScalingY = target.get('lockScalingY'),
          lockScalingFlip = target.get('lockScalingFlip');

      if (lockScalingX && lockScalingY) {
        return false;
      }

      // Get the constraint point
      var constraintPosition = target.translateToOriginPoint(target.getCenterPoint(), t.originX, t.originY),
          localMouse = target.toLocalPoint(new fabric.Point(x, y), t.originX, t.originY),
          dim = target._getTransformedDimensions(), scaled = false;

      this._setLocalMouse(localMouse, t);

      // Actually scale the object
      scaled = this._setObjectScale(localMouse, t, lockScalingX, lockScalingY, by, lockScalingFlip, dim);

      // Make sure the constraints apply
      target.setPositionByOrigin(constraintPosition, t.originX, t.originY);
      return scaled;
    },

    /**
     * @private
     * @return {Boolean} true if the scaling occurred
     */
    _setObjectScale: function(localMouse, transform, lockScalingX, lockScalingY, by, lockScalingFlip, _dim) {
      var target = transform.target, forbidScalingX = false, forbidScalingY = false, scaled = false,
          changeX, changeY, scaleX, scaleY;

      scaleX = localMouse.x * target.scaleX / _dim.x;
      scaleY = localMouse.y * target.scaleY / _dim.y;
      changeX = target.scaleX !== scaleX;
      changeY = target.scaleY !== scaleY;

      if (lockScalingFlip && scaleX <= 0 && scaleX < target.scaleX) {
        forbidScalingX = true;
      }

      if (lockScalingFlip && scaleY <= 0 && scaleY < target.scaleY) {
        forbidScalingY = true;
      }

      if (by === 'equally' && !lockScalingX && !lockScalingY) {
        forbidScalingX || forbidScalingY || (scaled = this._scaleObjectEqually(localMouse, target, transform, _dim));
      }
      else if (!by) {
        forbidScalingX || lockScalingX || (target.set('scaleX', scaleX) && (scaled = scaled || changeX));
        forbidScalingY || lockScalingY || (target.set('scaleY', scaleY) && (scaled = scaled || changeY));
      }
      else if (by === 'x' && !target.get('lockUniScaling')) {
        forbidScalingX || lockScalingX || (target.set('scaleX', scaleX) && (scaled = scaled || changeX));
      }
      else if (by === 'y' && !target.get('lockUniScaling')) {
        forbidScalingY || lockScalingY || (target.set('scaleY', scaleY) && (scaled = scaled || changeY));
      }
      transform.newScaleX = scaleX;
      transform.newScaleY = scaleY;
      forbidScalingX || forbidScalingY || this._flipObject(transform, by);
      return scaled;
    },

    /**
     * @private
     * @return {Boolean} true if the scaling occurred
     */
    _scaleObjectEqually: function(localMouse, target, transform, _dim) {

      var dist = localMouse.y + localMouse.x,
          lastDist = _dim.y * transform.original.scaleY / target.scaleY +
                     _dim.x * transform.original.scaleX / target.scaleX,
          scaled;

      // We use transform.scaleX/Y instead of target.scaleX/Y
      // because the object may have a min scale and we'll loose the proportions
      transform.newScaleX = transform.original.scaleX * dist / lastDist;
      transform.newScaleY = transform.original.scaleY * dist / lastDist;
      scaled = transform.newScaleX !== target.scaleX || transform.newScaleY !== target.scaleY;
      target.set('scaleX', transform.newScaleX);
      target.set('scaleY', transform.newScaleY);
      return scaled;
    },

    /**
     * @private
     */
    _flipObject: function(transform, by) {
      if (transform.newScaleX < 0 && by !== 'y') {
        if (transform.originX === 'left') {
          transform.originX = 'right';
        }
        else if (transform.originX === 'right') {
          transform.originX = 'left';
        }
      }

      if (transform.newScaleY < 0 && by !== 'x') {
        if (transform.originY === 'top') {
          transform.originY = 'bottom';
        }
        else if (transform.originY === 'bottom') {
          transform.originY = 'top';
        }
      }
    },

    /**
     * @private
     */
    _setLocalMouse: function(localMouse, t) {
      var target = t.target;

      if (t.originX === 'right') {
        localMouse.x *= -1;
      }
      else if (t.originX === 'center') {
        localMouse.x *= t.mouseXSign * 2;

        if (localMouse.x < 0) {
          t.mouseXSign = -t.mouseXSign;
        }
      }

      if (t.originY === 'bottom') {
        localMouse.y *= -1;
      }
      else if (t.originY === 'center') {
        localMouse.y *= t.mouseYSign * 2;

        if (localMouse.y < 0) {
          t.mouseYSign = -t.mouseYSign;
        }
      }

      // adjust the mouse coordinates when dealing with padding
      if (abs(localMouse.x) > target.padding) {
        if (localMouse.x < 0) {
          localMouse.x += target.padding;
        }
        else {
          localMouse.x -= target.padding;
        }
      }
      else { // mouse is within the padding, set to 0
        localMouse.x = 0;
      }

      if (abs(localMouse.y) > target.padding) {
        if (localMouse.y < 0) {
          localMouse.y += target.padding;
        }
        else {
          localMouse.y -= target.padding;
        }
      }
      else {
        localMouse.y = 0;
      }
    },

    /**
     * Rotates object by invoking its rotate method
     * @private
     * @param {Number} x pointer's x coordinate
     * @param {Number} y pointer's y coordinate
     * @return {Boolean} true if the rotation occurred
     */
    _rotateObject: function (x, y) {

      var t = this._currentTransform;

      if (t.target.get('lockRotation')) {
        return false;
      }

      var lastAngle = atan2(t.ey - t.top, t.ex - t.left),
          curAngle = atan2(y - t.top, x - t.left),
          angle = radiansToDegrees(curAngle - lastAngle + t.theta);

      // normalize angle to positive value
      if (angle < 0) {
        angle = 360 + angle;
      }

      t.target.angle = angle % 360;
      return true;
    },

    /**
     * Set the cursor type of the canvas element
     * @param {String} value Cursor type of the canvas element.
     * @see http://www.w3.org/TR/css3-ui/#cursor
     */
    setCursor: function (value) {
      this.upperCanvasEl.style.cursor = value;
    },

    /**
     * @private
     */
    _resetObjectTransform: function (target) {
      target.scaleX = 1;
      target.scaleY = 1;
      target.skewX = 0;
      target.skewY = 0;
      target.setAngle(0);
    },

    /**
     * @private
     */
    _drawSelection: function () {
      var ctx = this.contextTop,
          groupSelector = this._groupSelector,
          left = groupSelector.left,
          top = groupSelector.top,
          aleft = abs(left),
          atop = abs(top);

      ctx.fillStyle = this.selectionColor;

      ctx.fillRect(
        groupSelector.ex - ((left > 0) ? 0 : -left),
        groupSelector.ey - ((top > 0) ? 0 : -top),
        aleft,
        atop
      );

      ctx.lineWidth = this.selectionLineWidth;
      ctx.strokeStyle = this.selectionBorderColor;

      // selection border
      if (this.selectionDashArray.length > 1) {

        var px = groupSelector.ex + STROKE_OFFSET - ((left > 0) ? 0: aleft),
            py = groupSelector.ey + STROKE_OFFSET - ((top > 0) ? 0: atop);

        ctx.beginPath();

        fabric.util.drawDashedLine(ctx, px, py, px + aleft, py, this.selectionDashArray);
        fabric.util.drawDashedLine(ctx, px, py + atop - 1, px + aleft, py + atop - 1, this.selectionDashArray);
        fabric.util.drawDashedLine(ctx, px, py, px, py + atop, this.selectionDashArray);
        fabric.util.drawDashedLine(ctx, px + aleft - 1, py, px + aleft - 1, py + atop, this.selectionDashArray);

        ctx.closePath();
        ctx.stroke();
      }
      else {
        ctx.strokeRect(
          groupSelector.ex + STROKE_OFFSET - ((left > 0) ? 0 : aleft),
          groupSelector.ey + STROKE_OFFSET - ((top > 0) ? 0 : atop),
          aleft,
          atop
        );
      }
    },

    /**
     * @private
     */
    _isLastRenderedObject: function(pointer, e) {
      var lastRendered = this.lastRenderedWithControls;
      return (
        (this.preserveObjectStacking || e[this.lastRenderedKey]) &&
        lastRendered &&
        lastRendered.visible &&
        (this.containsPoint(null, lastRendered, pointer) ||
        lastRendered._findTargetCorner(pointer)));
    },

    /**
     * Method that determines what object we are clicking on
     * @param {Event} e mouse event
     * @param {Boolean} skipGroup when true, activeGroup is skipped and only objects are traversed through
     */
    findTarget: function (e, skipGroup) {
      if (this.skipTargetFind) {
        return;
      }

      var pointer = this.getPointer(e, true),
      activeGroup = this.getActiveGroup();

      // first check current group (if one exists)
      // active group does not check sub targets like normal groups.
      // if active group just exits.
      if (activeGroup && !skipGroup && this._checkTarget(pointer, activeGroup)) {
        return activeGroup;
      }

      var objects = this._objects;
      this.targets = [ ];

      if (this._isLastRenderedObject(pointer, e)) {
        objects = [this.lastRenderedWithControls];
      }

      var target = this._searchPossibleTargets(objects, pointer);
      this._fireOverOutEvents(target, e);
      return target;
    },

    /**
     * @private
     */
    _fireOverOutEvents: function(target, e) {
      if (target) {
        if (this._hoveredTarget !== target) {
          if (this._hoveredTarget) {
            this.fire('mouse:out', { target: this._hoveredTarget, e: e });
            this._hoveredTarget.fire('mouseout');
          }
          this.fire('mouse:over', { target: target, e: e });
          target.fire('mouseover');
          this._hoveredTarget = target;
        }
      }
      else if (this._hoveredTarget) {
        this.fire('mouse:out', { target: this._hoveredTarget, e: e });
        this._hoveredTarget.fire('mouseout');
        this._hoveredTarget = null;
      }
    },

    /**
     * @private
     */
    _checkTarget: function(pointer, obj) {
      if (obj &&
          obj.visible &&
          obj.evented &&
          this.containsPoint(null, obj, pointer)){
        if ((this.perPixelTargetFind || obj.perPixelTargetFind) && !obj.isEditing) {
          var isTransparent = this.isTargetTransparent(obj, pointer.x, pointer.y);
          if (!isTransparent) {
            return true;
          }
        }
        else {
          return true;
        }
      }
    },

    /**
     * @private
     */
    _searchPossibleTargets: function(objects, pointer) {

      // Cache all targets where their bounding box contains point.
      var target, i = objects.length, normalizedPointer, subTarget;
      // Do not check for currently grouped objects, since we check the parent group itself.
      // untill we call this function specifically to search inside the activeGroup
      while (i--) {
        if (this._checkTarget(pointer, objects[i])) {
          target = objects[i];
          if (target.type === 'group' && target.subTargetCheck) {
            normalizedPointer = this._normalizePointer(target, pointer);
            subTarget = this._searchPossibleTargets(target._objects, normalizedPointer);
            subTarget && this.targets.push(subTarget);
          }
          break;
        }
      }
      return target;
    },

    /**
     * Returns pointer coordinates relative to canvas.
     * @param {Event} e
     * @return {Object} object with "x" and "y" number values
     */
    getPointer: function (e, ignoreZoom, upperCanvasEl) {
      if (!upperCanvasEl) {
        upperCanvasEl = this.upperCanvasEl;
      }
      var pointer = getPointer(e),
          bounds = upperCanvasEl.getBoundingClientRect(),
          boundsWidth = bounds.width || 0,
          boundsHeight = bounds.height || 0,
          cssScale;

      if (!boundsWidth || !boundsHeight ) {
        if ('top' in bounds && 'bottom' in bounds) {
          boundsHeight = Math.abs( bounds.top - bounds.bottom );
        }
        if ('right' in bounds && 'left' in bounds) {
          boundsWidth = Math.abs( bounds.right - bounds.left );
        }
      }

      this.calcOffset();

      pointer.x = pointer.x - this._offset.left;
      pointer.y = pointer.y - this._offset.top;
      if (!ignoreZoom) {
        pointer = fabric.util.transformPoint(
          pointer,
          fabric.util.invertTransform(this.viewportTransform)
        );
      }

      if (boundsWidth === 0 || boundsHeight === 0) {
        // If bounds are not available (i.e. not visible), do not apply scale.
        cssScale = { width: 1, height: 1 };
      }
      else {
        cssScale = {
          width: upperCanvasEl.width / boundsWidth,
          height: upperCanvasEl.height / boundsHeight
        };
      }

      return {
        x: pointer.x * cssScale.width,
        y: pointer.y * cssScale.height
      };
    },

    /**
     * @private
     * @throws {CANVAS_INIT_ERROR} If canvas can not be initialized
     */
    _createUpperCanvas: function () {
      var lowerCanvasClass = this.lowerCanvasEl.className.replace(/\s*lower-canvas\s*/, '');

      this.upperCanvasEl = this._createCanvasElement();
      fabric.util.addClass(this.upperCanvasEl, 'upper-canvas ' + lowerCanvasClass);

      this.wrapperEl.appendChild(this.upperCanvasEl);

      this._copyCanvasStyle(this.lowerCanvasEl, this.upperCanvasEl);
      this._applyCanvasStyle(this.upperCanvasEl);
      this.contextTop = this.upperCanvasEl.getContext('2d');
    },

    /**
     * @private
     */
    _createCacheCanvas: function () {
      this.cacheCanvasEl = this._createCanvasElement();
      this.cacheCanvasEl.setAttribute('width', this.width);
      this.cacheCanvasEl.setAttribute('height', this.height);
      this.contextCache = this.cacheCanvasEl.getContext('2d');
    },

    /**
     * @private
     */
    _initWrapperElement: function () {
      this.wrapperEl = fabric.util.wrapElement(this.lowerCanvasEl, 'div', {
        'class': this.containerClass
      });
      fabric.util.setStyle(this.wrapperEl, {
        width: this.getWidth() + 'px',
        height: this.getHeight() + 'px',
        position: 'relative'
      });
      fabric.util.makeElementUnselectable(this.wrapperEl);
    },

    /**
     * @private
     * @param {HTMLElement} element canvas element to apply styles on
     */
    _applyCanvasStyle: function (element) {
      var width = this.getWidth() || element.width,
          height = this.getHeight() || element.height;

      fabric.util.setStyle(element, {
        position: 'absolute',
        width: width + 'px',
        height: height + 'px',
        left: 0,
        top: 0
      });
      element.width = width;
      element.height = height;
      fabric.util.makeElementUnselectable(element);
    },

    /**
     * Copys the the entire inline style from one element (fromEl) to another (toEl)
     * @private
     * @param {Element} fromEl Element style is copied from
     * @param {Element} toEl Element copied style is applied to
     */
    _copyCanvasStyle: function (fromEl, toEl) {
      toEl.style.cssText = fromEl.style.cssText;
    },

    /**
     * Returns context of canvas where object selection is drawn
     * @return {CanvasRenderingContext2D}
     */
    getSelectionContext: function() {
      return this.contextTop;
    },

    /**
     * Returns &lt;canvas> element on which object selection is drawn
     * @return {HTMLCanvasElement}
     */
    getSelectionElement: function () {
      return this.upperCanvasEl;
    },

    /**
     * @private
     * @param {Object} object
     */
    _setActiveObject: function(object) {
      if (this._activeObject) {
        this._activeObject.set('active', false);
      }
      this._activeObject = object;
      object.set('active', true);
    },

    /**
     * Sets given object as the only active object on canvas
     * @param {fabric.Object} object Object to set as an active one
     * @param {Event} [e] Event (passed along when firing "object:selected")
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    setActiveObject: function (object, e) {
      this._setActiveObject(object);
      this.renderAll();
      this.fire('object:selected', { target: object, e: e });
      object.fire('selected', { e: e });
      return this;
    },

    /**
     * Returns currently active object
     * @return {fabric.Object} active object
     */
    getActiveObject: function () {
      return this._activeObject;
    },

    /**
     * @private
     */
    _discardActiveObject: function() {
      if (this._activeObject) {
        this._activeObject.set('active', false);
      }
      this._activeObject = null;
    },

    /**
     * Discards currently active object
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    discardActiveObject: function (e) {
      this._discardActiveObject();
      this.renderAll();
      this.fire('selection:cleared', { e: e });
      return this;
    },

    /**
     * @private
     * @param {fabric.Group} group
     */
    _setActiveGroup: function(group) {
      this._activeGroup = group;
      if (group) {
        group.set('active', true);
      }
    },

    /**
     * Sets active group to a specified one
     * @param {fabric.Group} group Group to set as a current one
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    setActiveGroup: function (group, e) {
      this._setActiveGroup(group);
      if (group) {
        this.fire('object:selected', { target: group, e: e });
        group.fire('selected', { e: e });
      }
      return this;
    },

    /**
     * Returns currently active group
     * @return {fabric.Group} Current group
     */
    getActiveGroup: function () {
      return this._activeGroup;
    },

    /**
     * @private
     */
    _discardActiveGroup: function() {
      var g = this.getActiveGroup();
      if (g) {
        g.destroy();
      }
      this.setActiveGroup(null);
    },

    /**
     * Discards currently active group
     * @return {fabric.Canvas} thisArg
     */
    discardActiveGroup: function (e) {
      this._discardActiveGroup();
      this.fire('selection:cleared', { e: e });
      return this;
    },

    /**
     * Deactivates all objects on canvas, removing any active group or object
     * @return {fabric.Canvas} thisArg
     */
    deactivateAll: function () {
      var allObjects = this.getObjects(),
          i = 0,
          len = allObjects.length;
      for ( ; i < len; i++) {
        allObjects[i].set('active', false);
      }
      this._discardActiveGroup();
      this._discardActiveObject();
      return this;
    },

    /**
     * Deactivates all objects and dispatches appropriate events
     * @return {fabric.Canvas} thisArg
     */
    deactivateAllWithDispatch: function (e) {
      var activeObject = this.getActiveGroup() || this.getActiveObject();
      if (activeObject) {
        this.fire('before:selection:cleared', { target: activeObject, e: e });
      }
      this.deactivateAll();
      if (activeObject) {
        this.fire('selection:cleared', { e: e });
      }
      return this;
    },

    /**
     * Clears a canvas element and removes all event listeners
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    dispose: function () {
      this.callSuper('dispose');
      var wrapper = this.wrapperEl;
      this.removeListeners();
      wrapper.removeChild(this.upperCanvasEl);
      wrapper.removeChild(this.lowerCanvasEl);
      delete this.upperCanvasEl;
      if (wrapper.parentNode) {
        wrapper.parentNode.replaceChild(this.lowerCanvasEl, this.wrapperEl);
      }
      delete this.wrapperEl;
      return this;
    },

    /**
     * Draws objects' controls (borders/controls)
     * @param {CanvasRenderingContext2D} ctx Context to render controls on
     */
    drawControls: function(ctx) {
      var activeGroup = this.getActiveGroup();

      if (activeGroup) {
        activeGroup._renderControls(ctx);
      }
      else {
        this._drawObjectsControls(ctx);
      }
    },

    /**
     * @private
     */
    _drawObjectsControls: function(ctx) {
      for (var i = 0, len = this._objects.length; i < len; ++i) {
        if (!this._objects[i] || !this._objects[i].active) {
          continue;
        }
        this._objects[i]._renderControls(ctx);
        this.lastRenderedWithControls = this._objects[i];
      }
    },

    /**
     * @private
     * @param {fabric.Object} obj Object that was removed
     */
    _onObjectRemoved: function(obj) {
      if (obj === this.lastRenderedWithControls) {
        delete this.lastRenderedWithControls;
      }
      this.callSuper('_onObjectRemoved', obj);
    },

    /**
     * Clears all contexts (background, main, top) of an instance
     * @return {fabric.Canvas} thisArg
     * @chainable
     */
    clear: function () {
      delete this.lastRenderedWithControls;
      return this.callSuper('clear');
    }
  });

  // copying static properties manually to work around Opera's bug,
  // where "prototype" property is enumerable and overrides existing prototype
  for (var prop in fabric.StaticCanvas) {
    if (prop !== 'prototype') {
      fabric.Canvas[prop] = fabric.StaticCanvas[prop];
    }
  }

  if (fabric.isTouchSupported) {
    /** @ignore */
    fabric.Canvas.prototype._setCursorFromEvent = function() { };
  }

  /**
   * @ignore
   * @class fabric.Element
   * @alias fabric.Canvas
   * @deprecated Use {@link fabric.Canvas} instead.
   * @constructor
   */
  fabric.Element = fabric.Canvas;
})();


(function() {

  var cursorOffset = {
    mt: 0, // n
    tr: 1, // ne
    mr: 2, // e
    br: 3, // se
    mb: 4, // s
    bl: 5, // sw
    ml: 6, // w
    tl: 7 // nw
  },
  addListener = fabric.util.addListener,
  removeListener = fabric.util.removeListener;

  fabric.util.object.extend(fabric.Canvas.prototype, /** @lends fabric.Canvas.prototype */ {

    /**
     * Map of cursor style values for each of the object controls
     * @private
     */
    cursorMap: [
      'n-resize',
      'ne-resize',
      'e-resize',
      'se-resize',
      's-resize',
      'sw-resize',
      'w-resize',
      'nw-resize'
    ],

    /**
     * Adds mouse listeners to canvas
     * @private
     */
    _initEventListeners: function () {

      this._bindEvents();

      addListener(fabric.window, 'resize', this._onResize);

      // mouse events
      addListener(this.upperCanvasEl, 'mousedown', this._onMouseDown);
      addListener(this.upperCanvasEl, 'mousemove', this._onMouseMove);
      addListener(this.upperCanvasEl, 'mousewheel', this._onMouseWheel);
      addListener(this.upperCanvasEl, 'mouseout', this._onMouseOut);

      // touch events
      addListener(this.upperCanvasEl, 'touchstart', this._onMouseDown);
      addListener(this.upperCanvasEl, 'touchmove', this._onMouseMove);

      if (typeof eventjs !== 'undefined' && 'add' in eventjs) {
        eventjs.add(this.upperCanvasEl, 'gesture', this._onGesture);
        eventjs.add(this.upperCanvasEl, 'drag', this._onDrag);
        eventjs.add(this.upperCanvasEl, 'orientation', this._onOrientationChange);
        eventjs.add(this.upperCanvasEl, 'shake', this._onShake);
        eventjs.add(this.upperCanvasEl, 'longpress', this._onLongPress);
      }
    },

    /**
     * @private
     */
    _bindEvents: function() {
      this._onMouseDown = this._onMouseDown.bind(this);
      this._onMouseMove = this._onMouseMove.bind(this);
      this._onMouseUp = this._onMouseUp.bind(this);
      this._onResize = this._onResize.bind(this);
      this._onGesture = this._onGesture.bind(this);
      this._onDrag = this._onDrag.bind(this);
      this._onShake = this._onShake.bind(this);
      this._onLongPress = this._onLongPress.bind(this);
      this._onOrientationChange = this._onOrientationChange.bind(this);
      this._onMouseWheel = this._onMouseWheel.bind(this);
      this._onMouseOut = this._onMouseOut.bind(this);
    },

    /**
     * Removes all event listeners
     */
    removeListeners: function() {
      removeListener(fabric.window, 'resize', this._onResize);

      removeListener(this.upperCanvasEl, 'mousedown', this._onMouseDown);
      removeListener(this.upperCanvasEl, 'mousemove', this._onMouseMove);
      removeListener(this.upperCanvasEl, 'mousewheel', this._onMouseWheel);
      removeListener(this.upperCanvasEl, 'mouseout', this._onMouseOut);

      removeListener(this.upperCanvasEl, 'touchstart', this._onMouseDown);
      removeListener(this.upperCanvasEl, 'touchmove', this._onMouseMove);

      if (typeof eventjs !== 'undefined' && 'remove' in eventjs) {
        eventjs.remove(this.upperCanvasEl, 'gesture', this._onGesture);
        eventjs.remove(this.upperCanvasEl, 'drag', this._onDrag);
        eventjs.remove(this.upperCanvasEl, 'orientation', this._onOrientationChange);
        eventjs.remove(this.upperCanvasEl, 'shake', this._onShake);
        eventjs.remove(this.upperCanvasEl, 'longpress', this._onLongPress);
      }
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js gesture
     * @param {Event} [self] Inner Event object
     */
    _onGesture: function(e, self) {
      this.__onTransformGesture && this.__onTransformGesture(e, self);
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js drag
     * @param {Event} [self] Inner Event object
     */
    _onDrag: function(e, self) {
      this.__onDrag && this.__onDrag(e, self);
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js wheel event
     * @param {Event} [self] Inner Event object
     */
    _onMouseWheel: function(e, self) {
      this.__onMouseWheel && this.__onMouseWheel(e, self);
    },

    /**
     * @private
     * @param {Event} e Event object fired on mousedown
     */
    _onMouseOut: function(e) {
      var target = this._hoveredTarget;
      this.fire('mouse:out', { target: target, e: e });
      this._hoveredTarget = null;
      target && target.fire('mouseout', { e: e });
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js orientation change
     * @param {Event} [self] Inner Event object
     */
    _onOrientationChange: function(e, self) {
      this.__onOrientationChange && this.__onOrientationChange(e, self);
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js shake
     * @param {Event} [self] Inner Event object
     */
    _onShake: function(e, self) {
      this.__onShake && this.__onShake(e, self);
    },

    /**
     * @private
     * @param {Event} [e] Event object fired on Event.js shake
     * @param {Event} [self] Inner Event object
     */
    _onLongPress: function(e, self) {
      this.__onLongPress && this.__onLongPress(e, self);
    },

    /**
     * @private
     * @param {Event} e Event object fired on mousedown
     */
    _onMouseDown: function (e) {
      this.__onMouseDown(e);

      addListener(fabric.document, 'touchend', this._onMouseUp);
      addListener(fabric.document, 'touchmove', this._onMouseMove);

      removeListener(this.upperCanvasEl, 'mousemove', this._onMouseMove);
      removeListener(this.upperCanvasEl, 'touchmove', this._onMouseMove);

      if (e.type === 'touchstart') {
        // Unbind mousedown to prevent double triggers from touch devices
        removeListener(this.upperCanvasEl, 'mousedown', this._onMouseDown);
      }
      else {
        addListener(fabric.document, 'mouseup', this._onMouseUp);
        addListener(fabric.document, 'mousemove', this._onMouseMove);
      }
    },

    /**
     * @private
     * @param {Event} e Event object fired on mouseup
     */
    _onMouseUp: function (e) {
      this.__onMouseUp(e);

      removeListener(fabric.document, 'mouseup', this._onMouseUp);
      removeListener(fabric.document, 'touchend', this._onMouseUp);

      removeListener(fabric.document, 'mousemove', this._onMouseMove);
      removeListener(fabric.document, 'touchmove', this._onMouseMove);

      addListener(this.upperCanvasEl, 'mousemove', this._onMouseMove);
      addListener(this.upperCanvasEl, 'touchmove', this._onMouseMove);

      if (e.type === 'touchend') {
        // Wait 400ms before rebinding mousedown to prevent double triggers
        // from touch devices
        var _this = this;
        setTimeout(function() {
          addListener(_this.upperCanvasEl, 'mousedown', _this._onMouseDown);
        }, 400);
      }
    },

    /**
     * @private
     * @param {Event} e Event object fired on mousemove
     */
    _onMouseMove: function (e) {
      !this.allowTouchScrolling && e.preventDefault && e.preventDefault();
      this.__onMouseMove(e);
    },

    /**
     * @private
     */
    _onResize: function () {
      this.calcOffset();
    },

    /**
     * Decides whether the canvas should be redrawn in mouseup and mousedown events.
     * @private
     * @param {Object} target
     * @param {Object} pointer
     */
    _shouldRender: function(target, pointer) {
      var activeObject = this.getActiveGroup() || this.getActiveObject();

      return !!(
        (target && (
          target.isMoving ||
          target !== activeObject))
        ||
        (!target && !!activeObject)
        ||
        (!target && !activeObject && !this._groupSelector)
        ||
        (pointer &&
          this._previousPointer &&
          this.selection && (
          pointer.x !== this._previousPointer.x ||
          pointer.y !== this._previousPointer.y))
      );
    },

    /**
     * Method that defines the actions when mouse is released on canvas.
     * The method resets the currentTransform parameters, store the image corner
     * position in the image object and render the canvas on top.
     * @private
     * @param {Event} e Event object fired on mouseup
     */
    __onMouseUp: function (e) {
      var target, searchTarget = true, transform = this._currentTransform,
          groupSelector = this._groupSelector,
          isClick = (!groupSelector || (groupSelector.left === 0 && groupSelector.top === 0));

      if (this.isDrawingMode && this._isCurrentlyDrawing) {
        this._onMouseUpInDrawingMode(e);
        return;
      }

      if (transform) {
        this._finalizeCurrentTransform();
        searchTarget = !transform.actionPerformed;
      }

      target = searchTarget ? this.findTarget(e, true) : transform.target;

      var shouldRender = this._shouldRender(target, this.getPointer(e));

      if (target || !isClick) {
        this._maybeGroupObjects(e);
      }
      else {
        // those are done by default on mouse up
        // by _maybeGroupObjects, we are skipping it in case of no target find
        this._groupSelector = null;
        this._currentTransform = null;
      }

      if (target) {
        target.isMoving = false;
      }

      this._handleCursorAndEvent(e, target, 'up');
      shouldRender && this.renderAll();
    },

    /**
     * set cursor for mouse up and handle mouseUp event
     * @param {Event} e event from mouse
     * @param {fabric.Object} target receiving event
     * @param {String} eventType event to fire (up, down or move)
     */
    _handleCursorAndEvent: function(e, target, eventType) {
      this._setCursorFromEvent(e, target);
      this._handleEvent(e, eventType, target ? target : null);
    },

    /**
     * Handle event firing for target and subtargets
     * @param {Event} e event from mouse
     * @param {String} eventType event to fire (up, down or move)
     * @param {fabric.Object} targetObj receiving event
     */
    _handleEvent: function(e, eventType, targetObj) {
      var target = typeof targetObj === undefined ? this.findTarget(e) : targetObj,
          targets = this.targets || [ ],
          options = { e: e, target: target, subTargets: targets };

      this.fire('mouse:' + eventType, options);
      target && target.fire('mouse' + eventType, options);
      for (var i = 0; i < targets.length; i++) {
        targets[i].fire('mouse' + eventType, options);
      }
    },

    /**
     * @private
     */
    _finalizeCurrentTransform: function() {

      var transform = this._currentTransform,
          target = transform.target;

      if (target._scaling) {
        target._scaling = false;
      }

      target.setCoords();
      this._restoreOriginXY(target);

      if (transform.actionPerformed || (this.stateful && target.hasStateChanged())) {
        this.fire('object:modified', { target: target });
        target.fire('modified');
      }
    },

    /**
     * @private
     * @param {Object} target Object to restore
     */
    _restoreOriginXY: function(target) {
      if (this._previousOriginX && this._previousOriginY) {

        var originPoint = target.translateToOriginPoint(
          target.getCenterPoint(),
          this._previousOriginX,
          this._previousOriginY);

        target.originX = this._previousOriginX;
        target.originY = this._previousOriginY;

        target.left = originPoint.x;
        target.top = originPoint.y;

        this._previousOriginX = null;
        this._previousOriginY = null;
      }
    },

    /**
     * @private
     * @param {Event} e Event object fired on mousedown
     */
    _onMouseDownInDrawingMode: function(e) {
      this._isCurrentlyDrawing = true;
      this.discardActiveObject(e).renderAll();
      if (this.clipTo) {
        fabric.util.clipContext(this, this.contextTop);
      }
      var ivt = fabric.util.invertTransform(this.viewportTransform),
          pointer = fabric.util.transformPoint(this.getPointer(e, true), ivt);
      this.freeDrawingBrush.onMouseDown(pointer);
      this._handleEvent(e, 'down');
    },

    /**
     * @private
     * @param {Event} e Event object fired on mousemove
     */
    _onMouseMoveInDrawingMode: function(e) {
      if (this._isCurrentlyDrawing) {
        var ivt = fabric.util.invertTransform(this.viewportTransform),
            pointer = fabric.util.transformPoint(this.getPointer(e, true), ivt);
        this.freeDrawingBrush.onMouseMove(pointer);
      }
      this.setCursor(this.freeDrawingCursor);
      this._handleEvent(e, 'move');
    },

    /**
     * @private
     * @param {Event} e Event object fired on mouseup
     */
    _onMouseUpInDrawingMode: function(e) {
      this._isCurrentlyDrawing = false;
      if (this.clipTo) {
        this.contextTop.restore();
      }
      this.freeDrawingBrush.onMouseUp();
      this._handleEvent(e, 'up');
    },

    /**
     * Method that defines the actions when mouse is clic ked on canvas.
     * The method inits the currentTransform parameters and renders all the
     * canvas so the current image can be placed on the top canvas and the rest
     * in on the container one.
     * @private
     * @param {Event} e Event object fired on mousedown
     */
    __onMouseDown: function (e) {

      // accept only left clicks
      var isLeftClick  = 'which' in e ? e.which === 1 : e.button === 0;
      if (!isLeftClick && !fabric.isTouchSupported) {
        return;
      }

      if (this.isDrawingMode) {
        this._onMouseDownInDrawingMode(e);
        return;
      }

      // ignore if some object is being transformed at this moment
      if (this._currentTransform) {
        return;
      }

      var target = this.findTarget(e),
          pointer = this.getPointer(e, true);

      // save pointer for check in __onMouseUp event
      this._previousPointer = pointer;

      var shouldRender = this._shouldRender(target, pointer),
          shouldGroup = this._shouldGroup(e, target);

      if (this._shouldClearSelection(e, target)) {
        this._clearSelection(e, target, pointer);
      }
      else if (shouldGroup) {
        this._handleGrouping(e, target);
        target = this.getActiveGroup();
      }

      if (target) {
        if (target.selectable && (target.__corner || !shouldGroup)) {
          this._beforeTransform(e, target);
          this._setupCurrentTransform(e, target);
        }

        if (target !== this.getActiveGroup() && target !== this.getActiveObject()) {
          this.deactivateAll();
          target.selectable && this.setActiveObject(target, e);
        }
      }
      this._handleEvent(e, 'down', target ? target : null);
      // we must renderAll so that active image is placed on the top canvas
      shouldRender && this.renderAll();
    },

    /**
     * @private
     */
    _beforeTransform: function(e, target) {
      this.stateful && target.saveState();

      // determine if it's a drag or rotate case
      if (target._findTargetCorner(this.getPointer(e))) {
        this.onBeforeScaleRotate(target);
      }

    },

    /**
     * @private
     */
    _clearSelection: function(e, target, pointer) {
      this.deactivateAllWithDispatch(e);

      if (target && target.selectable) {
        this.setActiveObject(target, e);
      }
      else if (this.selection) {
        this._groupSelector = {
          ex: pointer.x,
          ey: pointer.y,
          top: 0,
          left: 0
        };
      }
    },

    /**
     * @private
     * @param {Object} target Object for that origin is set to center
     */
    _setOriginToCenter: function(target) {
      this._previousOriginX = this._currentTransform.target.originX;
      this._previousOriginY = this._currentTransform.target.originY;

      var center = target.getCenterPoint();

      target.originX = 'center';
      target.originY = 'center';

      target.left = center.x;
      target.top = center.y;

      this._currentTransform.left = target.left;
      this._currentTransform.top = target.top;
    },

    /**
     * @private
     * @param {Object} target Object for that center is set to origin
     */
    _setCenterToOrigin: function(target) {
      var originPoint = target.translateToOriginPoint(
        target.getCenterPoint(),
        this._previousOriginX,
        this._previousOriginY);

      target.originX = this._previousOriginX;
      target.originY = this._previousOriginY;

      target.left = originPoint.x;
      target.top = originPoint.y;

      this._previousOriginX = null;
      this._previousOriginY = null;
    },

    /**
     * Method that defines the actions when mouse is hovering the canvas.
     * The currentTransform parameter will definde whether the user is rotating/scaling/translating
     * an image or neither of them (only hovering). A group selection is also possible and would cancel
     * all any other type of action.
     * In case of an image transformation only the top canvas will be rendered.
     * @private
     * @param {Event} e Event object fired on mousemove
     */
    __onMouseMove: function (e) {

      var target, pointer;

      if (this.isDrawingMode) {
        this._onMouseMoveInDrawingMode(e);
        return;
      }
      if (typeof e.touches !== 'undefined' && e.touches.length > 1) {
        return;
      }

      var groupSelector = this._groupSelector;

      // We initially clicked in an empty area, so we draw a box for multiple selection
      if (groupSelector) {
        pointer = this.getPointer(e, true);

        groupSelector.left = pointer.x - groupSelector.ex;
        groupSelector.top = pointer.y - groupSelector.ey;

        this.renderTop();
      }
      else if (!this._currentTransform) {
        target = this.findTarget(e);
        this._setCursorFromEvent(e, target);
      }
      else {
        this._transformObject(e);
      }
      this._handleEvent(e, 'move', target ? target : null);
    },

    /**
     * @private
     * @param {Event} e Event fired on mousemove
     */
    _transformObject: function(e) {
      var pointer = this.getPointer(e),
          transform = this._currentTransform;

      transform.reset = false,
      transform.target.isMoving = true;

      this._beforeScaleTransform(e, transform);
      this._performTransformAction(e, transform, pointer);

      this.renderAll();
    },

    /**
     * @private
     */
    _performTransformAction: function(e, transform, pointer) {
      var x = pointer.x,
          y = pointer.y,
          target = transform.target,
          action = transform.action,
          actionPerformed = false;

      if (action === 'rotate') {
        (actionPerformed = this._rotateObject(x, y)) && this._fire('rotating', target, e);
      }
      else if (action === 'scale') {
        (actionPerformed = this._onScale(e, transform, x, y)) && this._fire('scaling', target, e);
      }
      else if (action === 'scaleX') {
        (actionPerformed = this._scaleObject(x, y, 'x')) && this._fire('scaling', target, e);
      }
      else if (action === 'scaleY') {
        (actionPerformed = this._scaleObject(x, y, 'y')) && this._fire('scaling', target, e);
      }
      else if (action === 'skewX') {
        (actionPerformed = this._skewObject(x, y, 'x')) && this._fire('skewing', target, e);
      }
      else if (action === 'skewY') {
        (actionPerformed = this._skewObject(x, y, 'y')) && this._fire('skewing', target, e);
      }
      else {
        actionPerformed = this._translateObject(x, y);
        if (actionPerformed) {
          this._fire('moving', target, e);
          this.setCursor(target.moveCursor || this.moveCursor);
        }
      }
      transform.actionPerformed = actionPerformed;
    },

    /**
     * @private
     */
    _fire: function(eventName, target, e) {
      this.fire('object:' + eventName, { target: target, e: e });
      target.fire(eventName, { e: e });
    },

    /**
     * @private
     */
    _beforeScaleTransform: function(e, transform) {
      if (transform.action === 'scale' || transform.action === 'scaleX' || transform.action === 'scaleY') {
        var centerTransform = this._shouldCenterTransform(transform.target);

        // Switch from a normal resize to center-based
        if ((centerTransform && (transform.originX !== 'center' || transform.originY !== 'center')) ||
           // Switch from center-based resize to normal one
           (!centerTransform && transform.originX === 'center' && transform.originY === 'center')
        ) {
          this._resetCurrentTransform();
          transform.reset = true;
        }
      }
    },

    /**
     * @private
     * @return {Boolean} true if the scaling occurred
     */
    _onScale: function(e, transform, x, y) {
      // rotate object only if shift key is not pressed
      // and if it is not a group we are transforming
      if ((e[this.uniScaleKey] || this.uniScaleTransform) && !transform.target.get('lockUniScaling')) {
        transform.currentAction = 'scale';
        return this._scaleObject(x, y);
      }
      else {
        // Switch from a normal resize to proportional
        if (!transform.reset && transform.currentAction === 'scale') {
          this._resetCurrentTransform();
        }

        transform.currentAction = 'scaleEqually';
        return this._scaleObject(x, y, 'equally');
      }
    },

    /**
     * Sets the cursor depending on where the canvas is being hovered.
     * Note: very buggy in Opera
     * @param {Event} e Event object
     * @param {Object} target Object that the mouse is hovering, if so.
     */
    _setCursorFromEvent: function (e, target) {
      if (!target) {
        this.setCursor(this.defaultCursor);
        return false;
      }

      var hoverCursor = target.hoverCursor || this.hoverCursor;
      if (!target.selectable) {
        //let's skip _findTargetCorner if object is not selectable
        this.setCursor(hoverCursor);
      }
      else {
        var activeGroup = this.getActiveGroup(),
            // only show proper corner when group selection is not active
            corner = target._findTargetCorner
                      && (!activeGroup || !activeGroup.contains(target))
                      && target._findTargetCorner(this.getPointer(e, true));

        if (!corner) {
          this.setCursor(hoverCursor);
        }
        else {
          this._setCornerCursor(corner, target, e);
        }
      }
      //actually unclear why it should return something
      //is never evaluated
      return true;
    },

    /**
     * @private
     */
    _setCornerCursor: function(corner, target, e) {
      if (corner in cursorOffset) {
        this.setCursor(this._getRotatedCornerCursor(corner, target, e));
      }
      else if (corner === 'mtr' && target.hasRotatingPoint) {
        this.setCursor(this.rotationCursor);
      }
      else {
        this.setCursor(this.defaultCursor);
        return false;
      }
    },

    /**
     * @private
     */
    _getRotatedCornerCursor: function(corner, target, e) {
      var n = Math.round((target.getAngle() % 360) / 45);

      if (n < 0) {
        n += 8; // full circle ahead
      }
      n += cursorOffset[corner];
      if (e[this.altActionKey] && cursorOffset[corner] % 2 === 0) {
        //if we are holding shift and we are on a mx corner...
        n += 2;
      }
      // normalize n to be from 0 to 7
      n %= 8;

      return this.cursorMap[n];
    }
  });
})();


(function() {

  var min = Math.min,
      max = Math.max;

  fabric.util.object.extend(fabric.Canvas.prototype, /** @lends fabric.Canvas.prototype */ {

    /**
     * @private
     * @param {Event} e Event object
     * @param {fabric.Object} target
     * @return {Boolean}
     */
    _shouldGroup: function(e, target) {
      var activeObject = this.getActiveObject();
      return e[this.selectionKey] && target && target.selectable &&
            (this.getActiveGroup() || (activeObject && activeObject !== target))
            && this.selection;
    },

    /**
     * @private
     * @param {Event} e Event object
     * @param {fabric.Object} target
     */
    _handleGrouping: function (e, target) {
      var activeGroup = this.getActiveGroup();

      if (target === activeGroup) {
        // if it's a group, find target again, using activeGroup objects
        target = this.findTarget(e, true);
        // if even object is not found, bail out
        if (!target) {
          return;
        }
      }
      if (activeGroup) {
        this._updateActiveGroup(target, e);
      }
      else {
        this._createActiveGroup(target, e);
      }

      if (this._activeGroup) {
        this._activeGroup.saveCoords();
      }
    },

    /**
     * @private
     */
    _updateActiveGroup: function(target, e) {
      var activeGroup = this.getActiveGroup();

      if (activeGroup.contains(target)) {

        activeGroup.removeWithUpdate(target);
        target.set('active', false);

        if (activeGroup.size() === 1) {
          // remove group alltogether if after removal it only contains 1 object
          this.discardActiveGroup(e);
          // activate last remaining object
          this.setActiveObject(activeGroup.item(0));
          return;
        }
      }
      else {
        activeGroup.addWithUpdate(target);
      }
      this.fire('selection:created', { target: activeGroup, e: e });
      activeGroup.set('active', true);
    },

    /**
     * @private
     */
    _createActiveGroup: function(target, e) {

      if (this._activeObject && target !== this._activeObject) {

        var group = this._createGroup(target);
        group.addWithUpdate();

        this.setActiveGroup(group);
        this._activeObject = null;

        this.fire('selection:created', { target: group, e: e });
      }

      target.set('active', true);
    },

    /**
     * @private
     * @param {Object} target
     */
    _createGroup: function(target) {

      var objects = this.getObjects(),
          isActiveLower = objects.indexOf(this._activeObject) < objects.indexOf(target),
          groupObjects = isActiveLower
            ? [ this._activeObject, target ]
            : [ target, this._activeObject ];
      this._activeObject.isEditing && this._activeObject.exitEditing();
      return new fabric.Group(groupObjects, {
        canvas: this
      });
    },

    /**
     * @private
     * @param {Event} e mouse event
     */
    _groupSelectedObjects: function (e) {

      var group = this._collectObjects();

      // do not create group for 1 element only
      if (group.length === 1) {
        this.setActiveObject(group[0], e);
      }
      else if (group.length > 1) {
        group = new fabric.Group(group.reverse(), {
          canvas: this
        });
        group.addWithUpdate();
        this.setActiveGroup(group, e);
        group.saveCoords();
        this.fire('selection:created', { target: group });
        this.renderAll();
      }
    },

    /**
     * @private
     */
    _collectObjects: function() {
      var group = [ ],
          currentObject,
          x1 = this._groupSelector.ex,
          y1 = this._groupSelector.ey,
          x2 = x1 + this._groupSelector.left,
          y2 = y1 + this._groupSelector.top,
          selectionX1Y1 = new fabric.Point(min(x1, x2), min(y1, y2)),
          selectionX2Y2 = new fabric.Point(max(x1, x2), max(y1, y2)),
          isClick = x1 === x2 && y1 === y2;

      for (var i = this._objects.length; i--; ) {
        currentObject = this._objects[i];

        if (!currentObject || !currentObject.selectable || !currentObject.visible) {
          continue;
        }

        if (currentObject.intersectsWithRect(selectionX1Y1, selectionX2Y2) ||
            currentObject.isContainedWithinRect(selectionX1Y1, selectionX2Y2) ||
            currentObject.containsPoint(selectionX1Y1) ||
            currentObject.containsPoint(selectionX2Y2)
        ) {
          currentObject.set('active', true);
          group.push(currentObject);

          // only add one object if it's a click
          if (isClick) {
            break;
          }
        }
      }

      return group;
    },

    /**
     * @private
     */
    _maybeGroupObjects: function(e) {
      if (this.selection && this._groupSelector) {
        this._groupSelectedObjects(e);
      }

      var activeGroup = this.getActiveGroup();
      if (activeGroup) {
        activeGroup.setObjectsCoords().setCoords();
        activeGroup.isMoving = false;
        this.setCursor(this.defaultCursor);
      }

      // clear selection and current transformation
      this._groupSelector = null;
      this._currentTransform = null;
    }
  });

})();


fabric.util.object.extend(fabric.StaticCanvas.prototype, /** @lends fabric.StaticCanvas.prototype */ {

  /**
   * Exports canvas element to a dataurl image. Note that when multiplier is used, cropping is scaled appropriately
   * @param {Object} [options] Options object
   * @param {String} [options.format=png] The format of the output image. Either "jpeg" or "png"
   * @param {Number} [options.quality=1] Quality level (0..1). Only used for jpeg.
   * @param {Number} [options.multiplier=1] Multiplier to scale by
   * @param {Number} [options.left] Cropping left offset. Introduced in v1.2.14
   * @param {Number} [options.top] Cropping top offset. Introduced in v1.2.14
   * @param {Number} [options.width] Cropping width. Introduced in v1.2.14
   * @param {Number} [options.height] Cropping height. Introduced in v1.2.14
   * @return {String} Returns a data: URL containing a representation of the object in the format specified by options.format
   * @see {@link http://jsfiddle.net/fabricjs/NfZVb/|jsFiddle demo}
   * @example <caption>Generate jpeg dataURL with lower quality</caption>
   * var dataURL = canvas.toDataURL({
   *   format: 'jpeg',
   *   quality: 0.8
   * });
   * @example <caption>Generate cropped png dataURL (clipping of canvas)</caption>
   * var dataURL = canvas.toDataURL({
   *   format: 'png',
   *   left: 100,
   *   top: 100,
   *   width: 200,
   *   height: 200
   * });
   * @example <caption>Generate double scaled png dataURL</caption>
   * var dataURL = canvas.toDataURL({
   *   format: 'png',
   *   multiplier: 2
   * });
   */
  toDataURL: function (options) {
    options || (options = { });

    var format = options.format || 'png',
        quality = options.quality || 1,
        multiplier = options.multiplier || 1,
        cropping = {
          left: options.left,
          top: options.top,
          width: options.width,
          height: options.height
        };

    if (this._isRetinaScaling()) {
      multiplier *= fabric.devicePixelRatio;
    }

    if (multiplier !== 1) {
      return this.__toDataURLWithMultiplier(format, quality, cropping, multiplier);
    }
    else {
      return this.__toDataURL(format, quality, cropping);
    }
  },

  /**
   * @private
   */
  __toDataURL: function(format, quality, cropping) {

    this.renderAll();

    var canvasEl = this.contextContainer.canvas,
        croppedCanvasEl = this.__getCroppedCanvas(canvasEl, cropping);

    // to avoid common confusion https://github.com/kangax/fabric.js/issues/806
    if (format === 'jpg') {
      format = 'jpeg';
    }

    var data = (fabric.StaticCanvas.supports('toDataURLWithQuality'))
              ? (croppedCanvasEl || canvasEl).toDataURL('image/' + format, quality)
              : (croppedCanvasEl || canvasEl).toDataURL('image/' + format);

    if (croppedCanvasEl) {
      croppedCanvasEl = null;
    }

    return data;
  },

  /**
   * @private
   */
  __getCroppedCanvas: function(canvasEl, cropping) {

    var croppedCanvasEl,
        croppedCtx,
        shouldCrop = 'left' in cropping ||
                     'top' in cropping ||
                     'width' in cropping ||
                     'height' in cropping;

    if (shouldCrop) {

      croppedCanvasEl = fabric.util.createCanvasElement();
      croppedCtx = croppedCanvasEl.getContext('2d');

      croppedCanvasEl.width = cropping.width || this.width;
      croppedCanvasEl.height = cropping.height || this.height;

      croppedCtx.drawImage(canvasEl, -cropping.left || 0, -cropping.top || 0);
    }

    return croppedCanvasEl;
  },

  /**
   * @private
   */
  __toDataURLWithMultiplier: function(format, quality, cropping, multiplier) {

    var origWidth = this.getWidth(),
        origHeight = this.getHeight(),
        scaledWidth = origWidth * multiplier,
        scaledHeight = origHeight * multiplier,
        activeObject = this.getActiveObject(),
        activeGroup = this.getActiveGroup(),
        zoom = this.getZoom(),
        newZoom = zoom * multiplier / fabric.devicePixelRatio;

    if (multiplier > 1) {
      this.setDimensions({ width: scaledWidth, height: scaledHeight });
    }

    this.setZoom(newZoom);

    if (cropping.left) {
      cropping.left *= multiplier;
    }
    if (cropping.top) {
      cropping.top *= multiplier;
    }
    if (cropping.width) {
      cropping.width *= multiplier;
    }
    else if (multiplier < 1) {
      cropping.width = scaledWidth;
    }
    if (cropping.height) {
      cropping.height *= multiplier;
    }
    else if (multiplier < 1) {
      cropping.height = scaledHeight;
    }

    if (activeGroup) {
      // not removing group due to complications with restoring it with correct state afterwords
      this._tempRemoveBordersControlsFromGroup(activeGroup);
    }
    else if (activeObject && this.deactivateAll) {
      this.deactivateAll();
    }

    var data = this.__toDataURL(format, quality, cropping);
    if (activeGroup) {
      this._restoreBordersControlsOnGroup(activeGroup);
    }
    else if (activeObject && this.setActiveObject) {
      this.setActiveObject(activeObject);
    }
    this.setZoom(zoom);
    //setDimensions with no option object is taking care of:
    //this.width, this.height, this.renderAll()
    this.setDimensions({ width: origWidth, height: origHeight });

    return data;
  },

  /**
   * Exports canvas element to a dataurl image (allowing to change image size via multiplier).
   * @deprecated since 1.0.13
   * @param {String} format (png|jpeg)
   * @param {Number} multiplier
   * @param {Number} quality (0..1)
   * @return {String}
   */
  toDataURLWithMultiplier: function (format, multiplier, quality) {
    return this.toDataURL({
      format: format,
      multiplier: multiplier,
      quality: quality
    });
  },

  /**
   * @private
   */
  _tempRemoveBordersControlsFromGroup: function(group) {
    group.origHasControls = group.hasControls;
    group.origBorderColor = group.borderColor;

    group.hasControls = true;
    group.borderColor = 'rgba(0,0,0,0)';

    group.forEachObject(function(o) {
      o.origBorderColor = o.borderColor;
      o.borderColor = 'rgba(0,0,0,0)';
    });
  },

  /**
   * @private
   */
  _restoreBordersControlsOnGroup: function(group) {
    group.hideControls = group.origHideControls;
    group.borderColor = group.origBorderColor;

    group.forEachObject(function(o) {
      o.borderColor = o.origBorderColor;
      delete o.origBorderColor;
    });
  }
});


fabric.util.object.extend(fabric.StaticCanvas.prototype, /** @lends fabric.StaticCanvas.prototype */ {

  /**
   * Populates canvas with data from the specified dataless JSON.
   * JSON format must conform to the one of {@link fabric.Canvas#toDatalessJSON}
   * @deprecated since 1.2.2
   * @param {String|Object} json JSON string or object
   * @param {Function} callback Callback, invoked when json is parsed
   *                            and corresponding objects (e.g: {@link fabric.Image})
   *                            are initialized
   * @param {Function} [reviver] Method for further parsing of JSON elements, called after each fabric object created.
   * @return {fabric.Canvas} instance
   * @chainable
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-3#deserialization}
   */
  loadFromDatalessJSON: function (json, callback, reviver) {
    return this.loadFromJSON(json, callback, reviver);
  },

  /**
   * Populates canvas with data from the specified JSON.
   * JSON format must conform to the one of {@link fabric.Canvas#toJSON}
   * @param {String|Object} json JSON string or object
   * @param {Function} callback Callback, invoked when json is parsed
   *                            and corresponding objects (e.g: {@link fabric.Image})
   *                            are initialized
   * @param {Function} [reviver] Method for further parsing of JSON elements, called after each fabric object created.
   * @return {fabric.Canvas} instance
   * @chainable
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-3#deserialization}
   * @see {@link http://jsfiddle.net/fabricjs/fmgXt/|jsFiddle demo}
   * @example <caption>loadFromJSON</caption>
   * canvas.loadFromJSON(json, canvas.renderAll.bind(canvas));
   * @example <caption>loadFromJSON with reviver</caption>
   * canvas.loadFromJSON(json, canvas.renderAll.bind(canvas), function(o, object) {
   *   // `o` = json object
   *   // `object` = fabric.Object instance
   *   // ... do some stuff ...
   * });
   */
  loadFromJSON: function (json, callback, reviver) {
    if (!json) {
      return;
    }

    // serialize if it wasn't already
    var serialized = (typeof json === 'string')
      ? JSON.parse(json)
      : fabric.util.object.clone(json);

    this.clear();

    var _this = this;
    this._enlivenObjects(serialized.objects, function () {
      _this._setBgOverlay(serialized, function () {
        // remove parts i cannot set as options
        delete serialized.objects;
        delete serialized.backgroundImage;
        delete serialized.overlayImage;
        delete serialized.background;
        delete serialized.overlay;
        // this._initOptions does too many things to just
        // call it. Normally loading an Object from JSON
        // create the Object instance. Here the Canvas is
        // already an instance and we are just loading things over it
        for (var prop in serialized) {
          _this[prop] = serialized[prop];
        }
        callback && callback();
      });
    }, reviver);
    return this;
  },

  /**
   * @private
   * @param {Object} serialized Object with background and overlay information
   * @param {Function} callback Invoked after all background and overlay images/patterns loaded
   */
  _setBgOverlay: function(serialized, callback) {
    var _this = this,
        loaded = {
          backgroundColor: false,
          overlayColor: false,
          backgroundImage: false,
          overlayImage: false
        };

    if (!serialized.backgroundImage && !serialized.overlayImage && !serialized.background && !serialized.overlay) {
      callback && callback();
      return;
    }

    var cbIfLoaded = function () {
      if (loaded.backgroundImage && loaded.overlayImage && loaded.backgroundColor && loaded.overlayColor) {
        _this.renderAll();
        callback && callback();
      }
    };

    this.__setBgOverlay('backgroundImage', serialized.backgroundImage, loaded, cbIfLoaded);
    this.__setBgOverlay('overlayImage', serialized.overlayImage, loaded, cbIfLoaded);
    this.__setBgOverlay('backgroundColor', serialized.background, loaded, cbIfLoaded);
    this.__setBgOverlay('overlayColor', serialized.overlay, loaded, cbIfLoaded);

    cbIfLoaded();
  },

  /**
   * @private
   * @param {String} property Property to set (backgroundImage, overlayImage, backgroundColor, overlayColor)
   * @param {(Object|String)} value Value to set
   * @param {Object} loaded Set loaded property to true if property is set
   * @param {Object} callback Callback function to invoke after property is set
   */
  __setBgOverlay: function(property, value, loaded, callback) {
    var _this = this;

    if (!value) {
      loaded[property] = true;
      return;
    }

    if (property === 'backgroundImage' || property === 'overlayImage') {
      fabric.Image.fromObject(value, function(img) {
        _this[property] = img;
        loaded[property] = true;
        callback && callback();
      });
    }
    else {
      this['set' + fabric.util.string.capitalize(property, true)](value, function() {
        loaded[property] = true;
        callback && callback();
      });
    }
  },

  /**
   * @private
   * @param {Array} objects
   * @param {Function} callback
   * @param {Function} [reviver]
   */
  _enlivenObjects: function (objects, callback, reviver) {
    var _this = this;

    if (!objects || objects.length === 0) {
      callback && callback();
      return;
    }

    var renderOnAddRemove = this.renderOnAddRemove;
    this.renderOnAddRemove = false;

    fabric.util.enlivenObjects(objects, function(enlivenedObjects) {
      enlivenedObjects.forEach(function(obj, index) {
        _this.insertAt(obj, index, true);
      });

      _this.renderOnAddRemove = renderOnAddRemove;
      callback && callback();
    }, null, reviver);
  },

  /**
   * @private
   * @param {String} format
   * @param {Function} callback
   */
  _toDataURL: function (format, callback) {
    this.clone(function (clone) {
      callback(clone.toDataURL(format));
    });
  },

  /**
   * @private
   * @param {String} format
   * @param {Number} multiplier
   * @param {Function} callback
   */
  _toDataURLWithMultiplier: function (format, multiplier, callback) {
    this.clone(function (clone) {
      callback(clone.toDataURLWithMultiplier(format, multiplier));
    });
  },

  /**
   * Clones canvas instance
   * @param {Object} [callback] Receives cloned instance as a first argument
   * @param {Array} [properties] Array of properties to include in the cloned canvas and children
   */
  clone: function (callback, properties) {
    var data = JSON.stringify(this.toJSON(properties));
    this.cloneWithoutData(function(clone) {
      clone.loadFromJSON(data, function() {
        callback && callback(clone);
      });
    });
  },

  /**
   * Clones canvas instance without cloning existing data.
   * This essentially copies canvas dimensions, clipping properties, etc.
   * but leaves data empty (so that you can populate it with your own)
   * @param {Object} [callback] Receives cloned instance as a first argument
   */
  cloneWithoutData: function(callback) {
    var el = fabric.document.createElement('canvas');

    el.width = this.getWidth();
    el.height = this.getHeight();

    var clone = new fabric.Canvas(el);
    clone.clipTo = this.clipTo;
    if (this.backgroundImage) {
      clone.setBackgroundImage(this.backgroundImage.src, function() {
        clone.renderAll();
        callback && callback(clone);
      });
      clone.backgroundImageOpacity = this.backgroundImageOpacity;
      clone.backgroundImageStretch = this.backgroundImageStretch;
    }
    else {
      callback && callback(clone);
    }
  }
});


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      toFixed = fabric.util.toFixed,
      capitalize = fabric.util.string.capitalize,
      degreesToRadians = fabric.util.degreesToRadians,
      supportsLineDash = fabric.StaticCanvas.supports('setLineDash');

  if (fabric.Object) {
    return;
  }

  /**
   * Root object class from which all 2d shape classes inherit from
   * @class fabric.Object
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-1#objects}
   * @see {@link fabric.Object#initialize} for constructor definition
   *
   * @fires added
   * @fires removed
   *
   * @fires selected
   * @fires modified
   * @fires rotating
   * @fires scaling
   * @fires moving
   * @fires skewing
   *
   * @fires mousedown
   * @fires mouseup
   * @fires mouseover
   * @fires mouseout
   */
  fabric.Object = fabric.util.createClass(/** @lends fabric.Object.prototype */ {

    /**
     * Retrieves object's {@link fabric.Object#clipTo|clipping function}
     * @method getClipTo
     * @memberOf fabric.Object.prototype
     * @return {Function}
     */

    /**
     * Sets object's {@link fabric.Object#clipTo|clipping function}
     * @method setClipTo
     * @memberOf fabric.Object.prototype
     * @param {Function} clipTo Clipping function
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#transformMatrix|transformMatrix}
     * @method getTransformMatrix
     * @memberOf fabric.Object.prototype
     * @return {Array} transformMatrix
     */

    /**
     * Sets object's {@link fabric.Object#transformMatrix|transformMatrix}
     * @method setTransformMatrix
     * @memberOf fabric.Object.prototype
     * @param {Array} transformMatrix
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#visible|visible} state
     * @method getVisible
     * @memberOf fabric.Object.prototype
     * @return {Boolean} True if visible
     */

    /**
     * Sets object's {@link fabric.Object#visible|visible} state
     * @method setVisible
     * @memberOf fabric.Object.prototype
     * @param {Boolean} value visible value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#shadow|shadow}
     * @method getShadow
     * @memberOf fabric.Object.prototype
     * @return {Object} Shadow instance
     */

    /**
     * Retrieves object's {@link fabric.Object#stroke|stroke}
     * @method getStroke
     * @memberOf fabric.Object.prototype
     * @return {String} stroke value
     */

    /**
     * Sets object's {@link fabric.Object#stroke|stroke}
     * @method setStroke
     * @memberOf fabric.Object.prototype
     * @param {String} value stroke value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#strokeWidth|strokeWidth}
     * @method getStrokeWidth
     * @memberOf fabric.Object.prototype
     * @return {Number} strokeWidth value
     */

    /**
     * Sets object's {@link fabric.Object#strokeWidth|strokeWidth}
     * @method setStrokeWidth
     * @memberOf fabric.Object.prototype
     * @param {Number} value strokeWidth value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#originX|originX}
     * @method getOriginX
     * @memberOf fabric.Object.prototype
     * @return {String} originX value
     */

    /**
     * Sets object's {@link fabric.Object#originX|originX}
     * @method setOriginX
     * @memberOf fabric.Object.prototype
     * @param {String} value originX value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#originY|originY}
     * @method getOriginY
     * @memberOf fabric.Object.prototype
     * @return {String} originY value
     */

    /**
     * Sets object's {@link fabric.Object#originY|originY}
     * @method setOriginY
     * @memberOf fabric.Object.prototype
     * @param {String} value originY value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#fill|fill}
     * @method getFill
     * @memberOf fabric.Object.prototype
     * @return {String} Fill value
     */

    /**
     * Sets object's {@link fabric.Object#fill|fill}
     * @method setFill
     * @memberOf fabric.Object.prototype
     * @param {String} value Fill value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#opacity|opacity}
     * @method getOpacity
     * @memberOf fabric.Object.prototype
     * @return {Number} Opacity value (0-1)
     */

    /**
     * Sets object's {@link fabric.Object#opacity|opacity}
     * @method setOpacity
     * @memberOf fabric.Object.prototype
     * @param {Number} value Opacity value (0-1)
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#angle|angle} (in degrees)
     * @method getAngle
     * @memberOf fabric.Object.prototype
     * @return {Number}
     */

    /**
     * Retrieves object's {@link fabric.Object#top|top position}
     * @method getTop
     * @memberOf fabric.Object.prototype
     * @return {Number} Top value (in pixels)
     */

    /**
     * Sets object's {@link fabric.Object#top|top position}
     * @method setTop
     * @memberOf fabric.Object.prototype
     * @param {Number} value Top value (in pixels)
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#left|left position}
     * @method getLeft
     * @memberOf fabric.Object.prototype
     * @return {Number} Left value (in pixels)
     */

    /**
     * Sets object's {@link fabric.Object#left|left position}
     * @method setLeft
     * @memberOf fabric.Object.prototype
     * @param {Number} value Left value (in pixels)
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#scaleX|scaleX} value
     * @method getScaleX
     * @memberOf fabric.Object.prototype
     * @return {Number} scaleX value
     */

    /**
     * Sets object's {@link fabric.Object#scaleX|scaleX} value
     * @method setScaleX
     * @memberOf fabric.Object.prototype
     * @param {Number} value scaleX value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#scaleY|scaleY} value
     * @method getScaleY
     * @memberOf fabric.Object.prototype
     * @return {Number} scaleY value
     */

    /**
     * Sets object's {@link fabric.Object#scaleY|scaleY} value
     * @method setScaleY
     * @memberOf fabric.Object.prototype
     * @param {Number} value scaleY value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#flipX|flipX} value
     * @method getFlipX
     * @memberOf fabric.Object.prototype
     * @return {Boolean} flipX value
     */

    /**
     * Sets object's {@link fabric.Object#flipX|flipX} value
     * @method setFlipX
     * @memberOf fabric.Object.prototype
     * @param {Boolean} value flipX value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Retrieves object's {@link fabric.Object#flipY|flipY} value
     * @method getFlipY
     * @memberOf fabric.Object.prototype
     * @return {Boolean} flipY value
     */

    /**
     * Sets object's {@link fabric.Object#flipY|flipY} value
     * @method setFlipY
     * @memberOf fabric.Object.prototype
     * @param {Boolean} value flipY value
     * @return {fabric.Object} thisArg
     * @chainable
     */

    /**
     * Type of an object (rect, circle, path, etc.).
     * Note that this property is meant to be read-only and not meant to be modified.
     * If you modify, certain parts of Fabric (such as JSON loading) won't work correctly.
     * @type String
     * @default
     */
    type:                     'object',

    /**
     * Horizontal origin of transformation of an object (one of "left", "right", "center")
     * See http://jsfiddle.net/1ow02gea/40/ on how originX/originY affect objects in groups
     * @type String
     * @default
     */
    originX:                  'left',

    /**
     * Vertical origin of transformation of an object (one of "top", "bottom", "center")
     * See http://jsfiddle.net/1ow02gea/40/ on how originX/originY affect objects in groups
     * @type String
     * @default
     */
    originY:                  'top',

    /**
     * Top position of an object. Note that by default it's relative to object top. You can change this by setting originY={top/center/bottom}
     * @type Number
     * @default
     */
    top:                      0,

    /**
     * Left position of an object. Note that by default it's relative to object left. You can change this by setting originX={left/center/right}
     * @type Number
     * @default
     */
    left:                     0,

    /**
     * Object width
     * @type Number
     * @default
     */
    width:                    0,

    /**
     * Object height
     * @type Number
     * @default
     */
    height:                   0,

    /**
     * Object scale factor (horizontal)
     * @type Number
     * @default
     */
    scaleX:                   1,

    /**
     * Object scale factor (vertical)
     * @type Number
     * @default
     */
    scaleY:                   1,

    /**
     * When true, an object is rendered as flipped horizontally
     * @type Boolean
     * @default
     */
    flipX:                    false,

    /**
     * When true, an object is rendered as flipped vertically
     * @type Boolean
     * @default
     */
    flipY:                    false,

    /**
     * Opacity of an object
     * @type Number
     * @default
     */
    opacity:                  1,

    /**
     * Angle of rotation of an object (in degrees)
     * @type Number
     * @default
     */
    angle:                    0,

    /**
     * Angle of skew on x axes of an object (in degrees)
     * @type Number
     * @default
     */
    skewX:                    0,

    /**
     * Angle of skew on y axes of an object (in degrees)
     * @type Number
     * @default
     */
    skewY:                    0,

    /**
     * Size of object's controlling corners (in pixels)
     * @type Number
     * @default
     */
    cornerSize:               13,

    /**
     * When true, object's controlling corners are rendered as transparent inside (i.e. stroke instead of fill)
     * @type Boolean
     * @default
     */
    transparentCorners:       true,

    /**
     * Default cursor value used when hovering over this object on canvas
     * @type String
     * @default
     */
    hoverCursor:              null,

    /**
     * Default cursor value used when moving this object on canvas
     * @type String
     * @default
     */
    moveCursor:               null,

    /**
     * Padding between object and its controlling borders (in pixels)
     * @type Number
     * @default
     */
    padding:                  0,

    /**
     * Color of controlling borders of an object (when it's active)
     * @type String
     * @default
     */
    borderColor:              'rgba(102,153,255,0.75)',

    /**
     * Array specifying dash pattern of an object's borders (hasBorder must be true)
     * @since 1.6.2
     * @type Array
     */
    borderDashArray:          null,

    /**
     * Color of controlling corners of an object (when it's active)
     * @type String
     * @default
     */
    cornerColor:              'rgba(102,153,255,0.5)',

    /**
     * Color of controlling corners of an object (when it's active and transparentCorners false)
     * @since 1.6.2
     * @type String
     * @default
     */
    cornerStrokeColor:        null,

    /**
     * Specify style of control, 'rect' or 'circle'
     * @since 1.6.2
     * @type String
     */
    cornerStyle:          'rect',

    /**
     * Array specifying dash pattern of an object's control (hasBorder must be true)
     * @since 1.6.2
     * @type Array
     */
    cornerDashArray:          null,

    /**
     * When true, this object will use center point as the origin of transformation
     * when being scaled via the controls.
     * <b>Backwards incompatibility note:</b> This property replaces "centerTransform" (Boolean).
     * @since 1.3.4
     * @type Boolean
     * @default
     */
    centeredScaling:          false,

    /**
     * When true, this object will use center point as the origin of transformation
     * when being rotated via the controls.
     * <b>Backwards incompatibility note:</b> This property replaces "centerTransform" (Boolean).
     * @since 1.3.4
     * @type Boolean
     * @default
     */
    centeredRotation:         true,

    /**
     * Color of object's fill
     * @type String
     * @default
     */
    fill:                     'rgb(0,0,0)',

    /**
     * Fill rule used to fill an object
     * accepted values are nonzero, evenodd
     * <b>Backwards incompatibility note:</b> This property was used for setting globalCompositeOperation until v1.4.12 (use `fabric.Object#globalCompositeOperation` instead)
     * @type String
     * @default
     */
    fillRule:                 'nonzero',

    /**
     * Composite rule used for canvas globalCompositeOperation
     * @type String
     * @default
     */
    globalCompositeOperation: 'source-over',

    /**
     * Background color of an object. Only works with text objects at the moment.
     * @type String
     * @default
     */
    backgroundColor:          '',

    /**
     * Selection Background color of an object. colored layer behind the object when it is active.
     * does not mix good with globalCompositeOperation methods.
     * @type String
     * @default
     */
    selectionBackgroundColor:          '',

    /**
     * When defined, an object is rendered via stroke and this property specifies its color
     * @type String
     * @default
     */
    stroke:                   null,

    /**
     * Width of a stroke used to render this object
     * @type Number
     * @default
     */
    strokeWidth:              1,

    /**
     * Array specifying dash pattern of an object's stroke (stroke must be defined)
     * @type Array
     */
    strokeDashArray:          null,

    /**
     * Line endings style of an object's stroke (one of "butt", "round", "square")
     * @type String
     * @default
     */
    strokeLineCap:            'butt',

    /**
     * Corner style of an object's stroke (one of "bevil", "round", "miter")
     * @type String
     * @default
     */
    strokeLineJoin:           'miter',

    /**
     * Maximum miter length (used for strokeLineJoin = "miter") of an object's stroke
     * @type Number
     * @default
     */
    strokeMiterLimit:         10,

    /**
     * Shadow object representing shadow of this shape
     * @type fabric.Shadow
     * @default
     */
    shadow:                   null,

    /**
     * Opacity of object's controlling borders when object is active and moving
     * @type Number
     * @default
     */
    borderOpacityWhenMoving:  0.4,

    /**
     * Scale factor of object's controlling borders
     * @type Number
     * @default
     */
    borderScaleFactor:        1,

    /**
     * Transform matrix (similar to SVG's transform matrix)
     * @type Array
     */
    transformMatrix:          null,

    /**
     * Minimum allowed scale value of an object
     * @type Number
     * @default
     */
    minScaleLimit:            0.01,

    /**
     * When set to `false`, an object can not be selected for modification (using either point-click-based or group-based selection).
     * But events still fire on it.
     * @type Boolean
     * @default
     */
    selectable:               true,

    /**
     * When set to `false`, an object can not be a target of events. All events propagate through it. Introduced in v1.3.4
     * @type Boolean
     * @default
     */
    evented:                  true,

    /**
     * When set to `false`, an object is not rendered on canvas
     * @type Boolean
     * @default
     */
    visible:                  true,

    /**
     * When set to `false`, object's controls are not displayed and can not be used to manipulate object
     * @type Boolean
     * @default
     */
    hasControls:              true,

    /**
     * When set to `false`, object's controlling borders are not rendered
     * @type Boolean
     * @default
     */
    hasBorders:               true,

    /**
     * When set to `false`, object's controlling rotating point will not be visible or selectable
     * @type Boolean
     * @default
     */
    hasRotatingPoint:         true,

    /**
     * Offset for object's controlling rotating point (when enabled via `hasRotatingPoint`)
     * @type Number
     * @default
     */
    rotatingPointOffset:      40,

    /**
     * When set to `true`, objects are "found" on canvas on per-pixel basis rather than according to bounding box
     * @type Boolean
     * @default
     */
    perPixelTargetFind:       false,

    /**
     * When `false`, default object's values are not included in its serialization
     * @type Boolean
     * @default
     */
    includeDefaultValues:     true,

    /**
     * Function that determines clipping of an object (context is passed as a first argument)
     * Note that context origin is at the object's center point (not left/top corner)
     * @type Function
     */
    clipTo:                   null,

    /**
     * When `true`, object horizontal movement is locked
     * @type Boolean
     * @default
     */
    lockMovementX:            false,

    /**
     * When `true`, object vertical movement is locked
     * @type Boolean
     * @default
     */
    lockMovementY:            false,

    /**
     * When `true`, object rotation is locked
     * @type Boolean
     * @default
     */
    lockRotation:             false,

    /**
     * When `true`, object horizontal scaling is locked
     * @type Boolean
     * @default
     */
    lockScalingX:             false,

    /**
     * When `true`, object vertical scaling is locked
     * @type Boolean
     * @default
     */
    lockScalingY:             false,

    /**
     * When `true`, object non-uniform scaling is locked
     * @type Boolean
     * @default
     */
    lockUniScaling:           false,

    /**
     * When `true`, object horizontal skewing is locked
     * @type Boolean
     * @default
     */
    lockSkewingX:             false,

    /**
     * When `true`, object vertical skewing is locked
     * @type Boolean
     * @default
     */
    lockSkewingY:             false,

    /**
     * When `true`, object cannot be flipped by scaling into negative values
     * @type Boolean
     * @default
     */

    lockScalingFlip:          false,

    /**
     * When `true`, object is not exported in SVG or OBJECT/JSON
     * since 1.6.3
     * @type Boolean
     * @default
     */

    excludeFromExport:          false,

    /**
     * List of properties to consider when checking if state
     * of an object is changed (fabric.Object#hasStateChanged)
     * as well as for history (undo/redo) purposes
     * @type Array
     */
    stateProperties:  (
      'top left width height scaleX scaleY flipX flipY originX originY transformMatrix ' +
      'stroke strokeWidth strokeDashArray strokeLineCap strokeLineJoin strokeMiterLimit ' +
      'angle opacity fill fillRule globalCompositeOperation shadow clipTo visible backgroundColor ' +
      'alignX alignY meetOrSlice skewX skewY'
    ).split(' '),

    /**
     * Constructor
     * @param {Object} [options] Options object
     */
    initialize: function(options) {
      if (options) {
        this.setOptions(options);
      }
    },

    /**
     * @private
     * @param {Object} [options] Options object
     */
    _initGradient: function(options) {
      if (options.fill && options.fill.colorStops && !(options.fill instanceof fabric.Gradient)) {
        this.set('fill', new fabric.Gradient(options.fill));
      }
      if (options.stroke && options.stroke.colorStops && !(options.stroke instanceof fabric.Gradient)) {
        this.set('stroke', new fabric.Gradient(options.stroke));
      }
    },

    /**
     * @private
     * @param {Object} [options] Options object
     */
    _initPattern: function(options) {
      if (options.fill && options.fill.source && !(options.fill instanceof fabric.Pattern)) {
        this.set('fill', new fabric.Pattern(options.fill));
      }
      if (options.stroke && options.stroke.source && !(options.stroke instanceof fabric.Pattern)) {
        this.set('stroke', new fabric.Pattern(options.stroke));
      }
    },

    /**
     * @private
     * @param {Object} [options] Options object
     */
    _initClipping: function(options) {
      if (!options.clipTo || typeof options.clipTo !== 'string') {
        return;
      }

      var functionBody = fabric.util.getFunctionBody(options.clipTo);
      if (typeof functionBody !== 'undefined') {
        this.clipTo = new Function('ctx', functionBody);
      }
    },

    /**
     * Sets object's properties from options
     * @param {Object} [options] Options object
     */
    setOptions: function(options) {
      for (var prop in options) {
        this.set(prop, options[prop]);
      }
      this._initGradient(options);
      this._initPattern(options);
      this._initClipping(options);
    },

    /**
     * Transforms context when rendering an object
     * @param {CanvasRenderingContext2D} ctx Context
     * @param {Boolean} fromLeft When true, context is transformed to object's top/left corner. This is used when rendering text on Node
     */
    transform: function(ctx, fromLeft) {
      if (this.group && this.canvas.preserveObjectStacking && this.group === this.canvas._activeGroup) {
        this.group.transform(ctx);
      }
      var center = fromLeft ? this._getLeftTopCoords() : this.getCenterPoint();
      ctx.translate(center.x, center.y);
      ctx.rotate(degreesToRadians(this.angle));
      ctx.scale(
        this.scaleX * (this.flipX ? -1 : 1),
        this.scaleY * (this.flipY ? -1 : 1)
      );
      ctx.transform(1, 0, Math.tan(degreesToRadians(this.skewX)), 1, 0, 0);
      ctx.transform(1, Math.tan(degreesToRadians(this.skewY)), 0, 1, 0, 0);
    },

    /**
     * Returns an object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS,

          object = {
            type:                     this.type,
            originX:                  this.originX,
            originY:                  this.originY,
            left:                     toFixed(this.left, NUM_FRACTION_DIGITS),
            top:                      toFixed(this.top, NUM_FRACTION_DIGITS),
            width:                    toFixed(this.width, NUM_FRACTION_DIGITS),
            height:                   toFixed(this.height, NUM_FRACTION_DIGITS),
            fill:                     (this.fill && this.fill.toObject) ? this.fill.toObject() : this.fill,
            stroke:                   (this.stroke && this.stroke.toObject) ? this.stroke.toObject() : this.stroke,
            strokeWidth:              toFixed(this.strokeWidth, NUM_FRACTION_DIGITS),
            strokeDashArray:          this.strokeDashArray ? this.strokeDashArray.concat() : this.strokeDashArray,
            strokeLineCap:            this.strokeLineCap,
            strokeLineJoin:           this.strokeLineJoin,
            strokeMiterLimit:         toFixed(this.strokeMiterLimit, NUM_FRACTION_DIGITS),
            scaleX:                   toFixed(this.scaleX, NUM_FRACTION_DIGITS),
            scaleY:                   toFixed(this.scaleY, NUM_FRACTION_DIGITS),
            angle:                    toFixed(this.getAngle(), NUM_FRACTION_DIGITS),
            flipX:                    this.flipX,
            flipY:                    this.flipY,
            opacity:                  toFixed(this.opacity, NUM_FRACTION_DIGITS),
            shadow:                   (this.shadow && this.shadow.toObject) ? this.shadow.toObject() : this.shadow,
            visible:                  this.visible,
            clipTo:                   this.clipTo && String(this.clipTo),
            backgroundColor:          this.backgroundColor,
            fillRule:                 this.fillRule,
            globalCompositeOperation: this.globalCompositeOperation,
            transformMatrix:          this.transformMatrix ? this.transformMatrix.concat() : this.transformMatrix,
            skewX:                    toFixed(this.skewX, NUM_FRACTION_DIGITS),
            skewY:                    toFixed(this.skewY, NUM_FRACTION_DIGITS)
          };

      if (!this.includeDefaultValues) {
        object = this._removeDefaultValues(object);
      }

      fabric.util.populateWithProperties(this, object, propertiesToInclude);

      return object;
    },

    /**
     * Returns (dataless) object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toDatalessObject: function(propertiesToInclude) {
      // will be overwritten by subclasses
      return this.toObject(propertiesToInclude);
    },

    /**
     * @private
     * @param {Object} object
     */
    _removeDefaultValues: function(object) {
      var prototype = fabric.util.getKlass(object.type).prototype,
          stateProperties = prototype.stateProperties;

      stateProperties.forEach(function(prop) {
        if (object[prop] === prototype[prop]) {
          delete object[prop];
        }
        var isArray = Object.prototype.toString.call(object[prop]) === '[object Array]' &&
                      Object.prototype.toString.call(prototype[prop]) === '[object Array]';

        // basically a check for [] === []
        if (isArray && object[prop].length === 0 && prototype[prop].length === 0) {
          delete object[prop];
        }
      });

      return object;
    },

    /**
     * Returns a string representation of an instance
     * @return {String}
     */
    toString: function() {
      return '#<fabric.' + capitalize(this.type) + '>';
    },

    /**
     * Basic getter
     * @param {String} property Property name
     * @return {Any} value of a property
     */
    get: function(property) {
      return this[property];
    },

    /**
     * @private
     */
    _setObject: function(obj) {
      for (var prop in obj) {
        this._set(prop, obj[prop]);
      }
    },

    /**
     * Sets property to a given value. When changing position/dimension -related properties (left, top, scale, angle, etc.) `set` does not update position of object's borders/controls. If you need to update those, call `setCoords()`.
     * @param {String|Object} key Property name or object (if object, iterate over the object properties)
     * @param {Object|Function} value Property value (if function, the value is passed into it and its return value is used as a new one)
     * @return {fabric.Object} thisArg
     * @chainable
     */
    set: function(key, value) {
      if (typeof key === 'object') {
        this._setObject(key);
      }
      else {
        if (typeof value === 'function' && key !== 'clipTo') {
          this._set(key, value(this.get(key)));
        }
        else {
          this._set(key, value);
        }
      }
      return this;
    },

    /**
     * @private
     * @param {String} key
     * @param {Any} value
     * @return {fabric.Object} thisArg
     */
    _set: function(key, value) {
      var shouldConstrainValue = (key === 'scaleX' || key === 'scaleY');

      if (shouldConstrainValue) {
        value = this._constrainScale(value);
      }
      if (key === 'scaleX' && value < 0) {
        this.flipX = !this.flipX;
        value *= -1;
      }
      else if (key === 'scaleY' && value < 0) {
        this.flipY = !this.flipY;
        value *= -1;
      }
      else if (key === 'shadow' && value && !(value instanceof fabric.Shadow)) {
        value = new fabric.Shadow(value);
      }

      this[key] = value;

      if (key === 'width' || key === 'height') {
        this.minScaleLimit = Math.min(0.1, 1/Math.max(this.width, this.height));
      }

      return this;
    },

    /**
     * This callback function is called by the parent group of an object every
     * time a non-delegated property changes on the group. It is passed the key
     * and value as parameters. Not adding in this function's signature to avoid
     * Travis build error about unused variables.
     */
    setOnGroup: function() {
      // implemented by sub-classes, as needed.
    },

    /**
     * Toggles specified property from `true` to `false` or from `false` to `true`
     * @param {String} property Property to toggle
     * @return {fabric.Object} thisArg
     * @chainable
     */
    toggle: function(property) {
      var value = this.get(property);
      if (typeof value === 'boolean') {
        this.set(property, !value);
      }
      return this;
    },

    /**
     * Sets sourcePath of an object
     * @param {String} value Value to set sourcePath to
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setSourcePath: function(value) {
      this.sourcePath = value;
      return this;
    },

    /**
     * Retrieves viewportTransform from Object's canvas if possible
     * @method getViewportTransform
     * @memberOf fabric.Object.prototype
     * @return {Boolean} flipY value // TODO
     */
    getViewportTransform: function() {
      if (this.canvas && this.canvas.viewportTransform) {
        return this.canvas.viewportTransform;
      }
      return [1, 0, 0, 1, 0, 0];
    },

    /**
     * Renders an object on a specified context
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Boolean} [noTransform] When true, context is not transformed
     */
    render: function(ctx, noTransform) {
      // do not render if width/height are zeros or object is not visible
      if ((this.width === 0 && this.height === 0) || !this.visible) {
        return;
      }

      ctx.save();

      //setup fill rule for current object
      this._setupCompositeOperation(ctx);
      this.drawSelectionBackground(ctx);
      if (!noTransform) {
        this.transform(ctx);
      }
      this._setStrokeStyles(ctx);
      this._setFillStyles(ctx);
      if (this.transformMatrix) {
        ctx.transform.apply(ctx, this.transformMatrix);
      }
      this._setOpacity(ctx);
      this._setShadow(ctx);
      this.clipTo && fabric.util.clipContext(this, ctx);
      this._render(ctx, noTransform);
      this.clipTo && ctx.restore();

      ctx.restore();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _setOpacity: function(ctx) {
      if (this.group) {
        this.group._setOpacity(ctx);
      }
      ctx.globalAlpha *= this.opacity;
    },

    _setStrokeStyles: function(ctx) {
      if (this.stroke) {
        ctx.lineWidth = this.strokeWidth;
        ctx.lineCap = this.strokeLineCap;
        ctx.lineJoin = this.strokeLineJoin;
        ctx.miterLimit = this.strokeMiterLimit;
        ctx.strokeStyle = this.stroke.toLive
          ? this.stroke.toLive(ctx, this)
          : this.stroke;
      }
    },

    _setFillStyles: function(ctx) {
      if (this.fill) {
        ctx.fillStyle = this.fill.toLive
          ? this.fill.toLive(ctx, this)
          : this.fill;
      }
    },

    /**
     * @private
     * Sets line dash
     * @param {CanvasRenderingContext2D} ctx Context to set the dash line on
     * @param {Array} dashArray array representing dashes
     * @param {Function} alternative function to call if browaser does not support lineDash
     */
    _setLineDash: function(ctx, dashArray, alternative) {
      if (!dashArray) {
        return;
      }
      // Spec requires the concatenation of two copies the dash list when the number of elements is odd
      if (1 & dashArray.length) {
        dashArray.push.apply(dashArray, dashArray);
      }
      if (supportsLineDash) {
        ctx.setLineDash(dashArray);
      }
      else {
        alternative && alternative(ctx);
      }
    },

    /**
     * Renders controls and borders for the object
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Boolean} [noTransform] When true, context is not transformed
     */
    _renderControls: function(ctx, noTransform) {
      if (!this.active || noTransform
          || (this.group && this.group !== this.canvas.getActiveGroup())) {
        return;
      }

      var vpt = this.getViewportTransform(),
          matrix = this.calcTransformMatrix(),
          options;
      matrix = fabric.util.multiplyTransformMatrices(vpt, matrix);
      options = fabric.util.qrDecompose(matrix);

      ctx.save();
      ctx.translate(options.translateX, options.translateY);
      ctx.lineWidth = 1 / this.borderScaleFactor;
      ctx.globalAlpha = this.isMoving ? this.borderOpacityWhenMoving : 1;

      if (this.group && this.group === this.canvas.getActiveGroup()) {
        ctx.rotate(degreesToRadians(options.angle));
        this.drawBordersInGroup(ctx, options);
      }
      else {
        ctx.rotate(degreesToRadians(this.angle));
        this.drawBorders(ctx);
      }
      this.drawControls(ctx);
      ctx.restore();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _setShadow: function(ctx) {
      if (!this.shadow) {
        return;
      }

      var multX = (this.canvas && this.canvas.viewportTransform[0]) || 1,
          multY = (this.canvas && this.canvas.viewportTransform[3]) || 1;
      if (this.canvas && this.canvas._isRetinaScaling()) {
        multX *= fabric.devicePixelRatio;
        multY *= fabric.devicePixelRatio;
      }
      ctx.shadowColor = this.shadow.color;
      ctx.shadowBlur = this.shadow.blur * (multX + multY) * (this.scaleX + this.scaleY) / 4;
      ctx.shadowOffsetX = this.shadow.offsetX * multX * this.scaleX;
      ctx.shadowOffsetY = this.shadow.offsetY * multY * this.scaleY;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _removeShadow: function(ctx) {
      if (!this.shadow) {
        return;
      }

      ctx.shadowColor = '';
      ctx.shadowBlur = ctx.shadowOffsetX = ctx.shadowOffsetY = 0;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderFill: function(ctx) {
      if (!this.fill) {
        return;
      }

      ctx.save();
      if (this.fill.gradientTransform) {
        var g = this.fill.gradientTransform;
        ctx.transform.apply(ctx, g);
      }
      if (this.fill.toLive) {
        ctx.translate(
          -this.width / 2 + this.fill.offsetX || 0,
          -this.height / 2 + this.fill.offsetY || 0);
      }
      if (this.fillRule === 'evenodd') {
        ctx.fill('evenodd');
      }
      else {
        ctx.fill();
      }
      ctx.restore();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderStroke: function(ctx) {
      if (!this.stroke || this.strokeWidth === 0) {
        return;
      }

      if (this.shadow && !this.shadow.affectStroke) {
        this._removeShadow(ctx);
      }

      ctx.save();

      this._setLineDash(ctx, this.strokeDashArray, this._renderDashedStroke);
      if (this.stroke.gradientTransform) {
        var g = this.stroke.gradientTransform;
        ctx.transform.apply(ctx, g);
      }
      if (this.stroke.toLive) {
        ctx.translate(
          -this.width / 2 + this.stroke.offsetX || 0,
          -this.height / 2 + this.stroke.offsetY || 0);
      }
      ctx.stroke();
      ctx.restore();
    },

    /**
     * Clones an instance
     * @param {Function} callback Callback is invoked with a clone as a first argument
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {fabric.Object} clone of an instance
     */
    clone: function(callback, propertiesToInclude) {
      if (this.constructor.fromObject) {
        return this.constructor.fromObject(this.toObject(propertiesToInclude), callback);
      }
      return new fabric.Object(this.toObject(propertiesToInclude));
    },

    /**
     * Creates an instance of fabric.Image out of an object
     * @param {Function} callback callback, invoked with an instance as a first argument
     * @return {fabric.Object} thisArg
     */
    cloneAsImage: function(callback) {
      var dataUrl = this.toDataURL();
      fabric.util.loadImage(dataUrl, function(img) {
        if (callback) {
          callback(new fabric.Image(img));
        }
      });
      return this;
    },

    /**
     * Converts an object into a data-url-like string
     * @param {Object} options Options object
     * @param {String} [options.format=png] The format of the output image. Either "jpeg" or "png"
     * @param {Number} [options.quality=1] Quality level (0..1). Only used for jpeg.
     * @param {Number} [options.multiplier=1] Multiplier to scale by
     * @param {Number} [options.left] Cropping left offset. Introduced in v1.2.14
     * @param {Number} [options.top] Cropping top offset. Introduced in v1.2.14
     * @param {Number} [options.width] Cropping width. Introduced in v1.2.14
     * @param {Number} [options.height] Cropping height. Introduced in v1.2.14
     * @return {String} Returns a data: URL containing a representation of the object in the format specified by options.format
     */
    toDataURL: function(options) {
      options || (options = { });

      var el = fabric.util.createCanvasElement(),
          boundingRect = this.getBoundingRect();

      el.width = boundingRect.width;
      el.height = boundingRect.height;

      fabric.util.wrapElement(el, 'div');
      var canvas = new fabric.StaticCanvas(el);

      // to avoid common confusion https://github.com/kangax/fabric.js/issues/806
      if (options.format === 'jpg') {
        options.format = 'jpeg';
      }

      if (options.format === 'jpeg') {
        canvas.backgroundColor = '#fff';
      }

      var origParams = {
        active: this.get('active'),
        left: this.getLeft(),
        top: this.getTop()
      };

      this.set('active', false);
      this.setPositionByOrigin(new fabric.Point(canvas.getWidth() / 2, canvas.getHeight() / 2), 'center', 'center');

      var originalCanvas = this.canvas;
      canvas.add(this);
      var data = canvas.toDataURL(options);

      this.set(origParams).setCoords();
      this.canvas = originalCanvas;

      canvas.dispose();
      canvas = null;

      return data;
    },

    /**
     * Returns true if specified type is identical to the type of an instance
     * @param {String} type Type to check against
     * @return {Boolean}
     */
    isType: function(type) {
      return this.type === type;
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return 0;
    },

    /**
     * Returns a JSON representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} JSON
     */
    toJSON: function(propertiesToInclude) {
      // delegate, not alias
      return this.toObject(propertiesToInclude);
    },

    /**
     * Sets gradient (fill or stroke) of an object
     * <b>Backwards incompatibility note:</b> This method was named "setGradientFill" until v1.1.0
     * @param {String} property Property name 'stroke' or 'fill'
     * @param {Object} [options] Options object
     * @param {String} [options.type] Type of gradient 'radial' or 'linear'
     * @param {Number} [options.x1=0] x-coordinate of start point
     * @param {Number} [options.y1=0] y-coordinate of start point
     * @param {Number} [options.x2=0] x-coordinate of end point
     * @param {Number} [options.y2=0] y-coordinate of end point
     * @param {Number} [options.r1=0] Radius of start point (only for radial gradients)
     * @param {Number} [options.r2=0] Radius of end point (only for radial gradients)
     * @param {Object} [options.colorStops] Color stops object eg. {0: 'ff0000', 1: '000000'}
     * @param {Object} [options.gradientTransform] transforMatrix for gradient
     * @return {fabric.Object} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/58y8b/|jsFiddle demo}
     * @example <caption>Set linear gradient</caption>
     * object.setGradient('fill', {
     *   type: 'linear',
     *   x1: -object.width / 2,
     *   y1: 0,
     *   x2: object.width / 2,
     *   y2: 0,
     *   colorStops: {
     *     0: 'red',
     *     0.5: '#005555',
     *     1: 'rgba(0,0,255,0.5)'
     *   }
     * });
     * canvas.renderAll();
     * @example <caption>Set radial gradient</caption>
     * object.setGradient('fill', {
     *   type: 'radial',
     *   x1: 0,
     *   y1: 0,
     *   x2: 0,
     *   y2: 0,
     *   r1: object.width / 2,
     *   r2: 10,
     *   colorStops: {
     *     0: 'red',
     *     0.5: '#005555',
     *     1: 'rgba(0,0,255,0.5)'
     *   }
     * });
     * canvas.renderAll();
     */
    setGradient: function(property, options) {
      options || (options = { });

      var gradient = { colorStops: [] };

      gradient.type = options.type || (options.r1 || options.r2 ? 'radial' : 'linear');
      gradient.coords = {
        x1: options.x1,
        y1: options.y1,
        x2: options.x2,
        y2: options.y2
      };

      if (options.r1 || options.r2) {
        gradient.coords.r1 = options.r1;
        gradient.coords.r2 = options.r2;
      }

      options.gradientTransform && (gradient.gradientTransform = options.gradientTransform);

      for (var position in options.colorStops) {
        var color = new fabric.Color(options.colorStops[position]);
        gradient.colorStops.push({
          offset: position,
          color: color.toRgb(),
          opacity: color.getAlpha()
        });
      }

      return this.set(property, fabric.Gradient.forObject(this, gradient));
    },

    /**
     * Sets pattern fill of an object
     * @param {Object} options Options object
     * @param {(String|HTMLImageElement)} options.source Pattern source
     * @param {String} [options.repeat=repeat] Repeat property of a pattern (one of repeat, repeat-x, repeat-y or no-repeat)
     * @param {Number} [options.offsetX=0] Pattern horizontal offset from object's left/top corner
     * @param {Number} [options.offsetY=0] Pattern vertical offset from object's left/top corner
     * @return {fabric.Object} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/QT3pa/|jsFiddle demo}
     * @example <caption>Set pattern</caption>
     * fabric.util.loadImage('http://fabricjs.com/assets/escheresque_ste.png', function(img) {
     *   object.setPatternFill({
     *     source: img,
     *     repeat: 'repeat'
     *   });
     *   canvas.renderAll();
     * });
     */
    setPatternFill: function(options) {
      return this.set('fill', new fabric.Pattern(options));
    },

    /**
     * Sets {@link fabric.Object#shadow|shadow} of an object
     * @param {Object|String} [options] Options object or string (e.g. "2px 2px 10px rgba(0,0,0,0.2)")
     * @param {String} [options.color=rgb(0,0,0)] Shadow color
     * @param {Number} [options.blur=0] Shadow blur
     * @param {Number} [options.offsetX=0] Shadow horizontal offset
     * @param {Number} [options.offsetY=0] Shadow vertical offset
     * @return {fabric.Object} thisArg
     * @chainable
     * @see {@link http://jsfiddle.net/fabricjs/7gvJG/|jsFiddle demo}
     * @example <caption>Set shadow with string notation</caption>
     * object.setShadow('2px 2px 10px rgba(0,0,0,0.2)');
     * canvas.renderAll();
     * @example <caption>Set shadow with object notation</caption>
     * object.setShadow({
     *   color: 'red',
     *   blur: 10,
     *   offsetX: 20,
     *   offsetY: 20
     * });
     * canvas.renderAll();
     */
    setShadow: function(options) {
      return this.set('shadow', options ? new fabric.Shadow(options) : null);
    },

    /**
     * Sets "color" of an instance (alias of `set('fill', &hellip;)`)
     * @param {String} color Color value
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setColor: function(color) {
      this.set('fill', color);
      return this;
    },

    /**
     * Sets "angle" of an instance
     * @param {Number} angle Angle value (in degrees)
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setAngle: function(angle) {
      var shouldCenterOrigin = (this.originX !== 'center' || this.originY !== 'center') && this.centeredRotation;

      if (shouldCenterOrigin) {
        this._setOriginToCenter();
      }

      this.set('angle', angle);

      if (shouldCenterOrigin) {
        this._resetOrigin();
      }

      return this;
    },

    /**
     * Centers object horizontally on canvas to which it was added last.
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    centerH: function () {
      this.canvas && this.canvas.centerObjectH(this);
      return this;
    },

    /**
     * Centers object horizontally on current viewport of canvas to which it was added last.
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    viewportCenterH: function () {
      this.canvas && this.canvas.viewportCenterObjectH(this);
      return this;
    },

    /**
     * Centers object vertically on canvas to which it was added last.
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    centerV: function () {
      this.canvas && this.canvas.centerObjectV(this);
      return this;
    },

    /**
     * Centers object vertically on current viewport of canvas to which it was added last.
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    viewportCenterV: function () {
      this.canvas && this.canvas.viewportCenterObjectV(this);
      return this;
    },

    /**
     * Centers object vertically and horizontally on canvas to which is was added last
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    center: function () {
      this.canvas && this.canvas.centerObject(this);
      return this;
    },

    /**
     * Centers object on current viewport of canvas to which it was added last.
     * You might need to call `setCoords` on an object after centering, to update controls area.
     * @return {fabric.Object} thisArg
     * @chainable
     */
    viewportCenter: function () {
      this.canvas && this.canvas.viewportCenterObject(this);
      return this;
    },

    /**
     * Removes object from canvas to which it was added last
     * @return {fabric.Object} thisArg
     * @chainable
     */
    remove: function() {
      this.canvas && this.canvas.remove(this);
      return this;
    },

    /**
     * Returns coordinates of a pointer relative to an object
     * @param {Event} e Event to operate upon
     * @param {Object} [pointer] Pointer to operate upon (instead of event)
     * @return {Object} Coordinates of a pointer (x, y)
     */
    getLocalPointer: function(e, pointer) {
      pointer = pointer || this.canvas.getPointer(e);
      var pClicked = new fabric.Point(pointer.x, pointer.y),
          objectLeftTop = this._getLeftTopCoords();
      if (this.angle) {
        pClicked = fabric.util.rotatePoint(
          pClicked, objectLeftTop, fabric.util.degreesToRadians(-this.angle));
      }
      return {
        x: pClicked.x - objectLeftTop.x,
        y: pClicked.y - objectLeftTop.y
      };
    },

    /**
     * Sets canvas globalCompositeOperation for specific object
     * custom composition operation for the particular object can be specifed using globalCompositeOperation property
     * @param {CanvasRenderingContext2D} ctx Rendering canvas context
     */
    _setupCompositeOperation: function (ctx) {
      if (this.globalCompositeOperation) {
        ctx.globalCompositeOperation = this.globalCompositeOperation;
      }
    }
  });

  fabric.util.createAccessors(fabric.Object);

  /**
   * Alias for {@link fabric.Object.prototype.setAngle}
   * @alias rotate -> setAngle
   * @memberOf fabric.Object
   */
  fabric.Object.prototype.rotate = fabric.Object.prototype.setAngle;

  extend(fabric.Object.prototype, fabric.Observable);

  /**
   * Defines the number of fraction digits to use when serializing object values.
   * You can use it to increase/decrease precision of such values like left, top, scaleX, scaleY, etc.
   * @static
   * @memberOf fabric.Object
   * @constant
   * @type Number
   */
  fabric.Object.NUM_FRACTION_DIGITS = 2;

  /**
   * Unique id used internally when creating SVG elements
   * @static
   * @memberOf fabric.Object
   * @type Number
   */
  fabric.Object.__uid = 0;

})(typeof exports !== 'undefined' ? exports : this);


(function() {

  var degreesToRadians = fabric.util.degreesToRadians,
      originXOffset = {
        left: -0.5,
        center: 0,
        right: 0.5
      },
      originYOffset = {
        top: -0.5,
        center: 0,
        bottom: 0.5
      };

  fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

    /**
     * Translates the coordinates from origin to center coordinates (based on the object's dimensions)
     * @param {fabric.Point} point The point which corresponds to the originX and originY params
     * @param {String} fromOriginX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} fromOriginY Vertical origin: 'top', 'center' or 'bottom'
     * @param {String} toOriginX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} toOriginY Vertical origin: 'top', 'center' or 'bottom'
     * @return {fabric.Point}
     */
    translateToGivenOrigin: function(point, fromOriginX, fromOriginY, toOriginX, toOriginY) {
      var x = point.x,
          y = point.y,
          offsetX = originXOffset[toOriginX] - originXOffset[fromOriginX],
          offsetY = originYOffset[toOriginY] - originYOffset[fromOriginY],
          dim;
      if (offsetX || offsetY) {
        dim = this._getTransformedDimensions();
        x = point.x + offsetX * dim.x;
        y = point.y + offsetY * dim.y;
      }
      return new fabric.Point(x, y);
    },

    /**
     * Translates the coordinates from origin to center coordinates (based on the object's dimensions)
     * @param {fabric.Point} point The point which corresponds to the originX and originY params
     * @param {String} originX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} originY Vertical origin: 'top', 'center' or 'bottom'
     * @return {fabric.Point}
     */
    translateToCenterPoint: function(point, originX, originY) {
      var p = this.translateToGivenOrigin(point, originX, originY, 'center', 'center');
      if (this.angle) {
        return fabric.util.rotatePoint(p, point, degreesToRadians(this.angle));
      }
      return p;
    },

    /**
     * Translates the coordinates from center to origin coordinates (based on the object's dimensions)
     * @param {fabric.Point} center The point which corresponds to center of the object
     * @param {String} originX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} originY Vertical origin: 'top', 'center' or 'bottom'
     * @return {fabric.Point}
     */
    translateToOriginPoint: function(center, originX, originY) {
      var p = this.translateToGivenOrigin(center, 'center', 'center', originX, originY);
      if (this.angle) {
        return fabric.util.rotatePoint(p, center, degreesToRadians(this.angle));
      }
      return p;
    },

    /**
     * Returns the real center coordinates of the object
     * @return {fabric.Point}
     */
    getCenterPoint: function() {
      var leftTop = new fabric.Point(this.left, this.top);
      return this.translateToCenterPoint(leftTop, this.originX, this.originY);
    },

    /**
     * Returns the coordinates of the object based on center coordinates
     * @param {fabric.Point} point The point which corresponds to the originX and originY params
     * @return {fabric.Point}
     */
    // getOriginPoint: function(center) {
    //   return this.translateToOriginPoint(center, this.originX, this.originY);
    // },

    /**
     * Returns the coordinates of the object as if it has a different origin
     * @param {String} originX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} originY Vertical origin: 'top', 'center' or 'bottom'
     * @return {fabric.Point}
     */
    getPointByOrigin: function(originX, originY) {
      var center = this.getCenterPoint();
      return this.translateToOriginPoint(center, originX, originY);
    },

    /**
     * Returns the point in local coordinates
     * @param {fabric.Point} point The point relative to the global coordinate system
     * @param {String} originX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} originY Vertical origin: 'top', 'center' or 'bottom'
     * @return {fabric.Point}
     */
    toLocalPoint: function(point, originX, originY) {
      var center = this.getCenterPoint(),
          p, p2;

      if (originX && originY) {
        p = this.translateToGivenOrigin(center, 'center', 'center', originX, originY);
      }
      else {
        p = new fabric.Point(this.left, this.top);
      }

      p2 = new fabric.Point(point.x, point.y);
      if (this.angle) {
        p2 = fabric.util.rotatePoint(p2, center, -degreesToRadians(this.angle));
      }
      return p2.subtractEquals(p);
    },

    /**
     * Returns the point in global coordinates
     * @param {fabric.Point} The point relative to the local coordinate system
     * @return {fabric.Point}
     */
    // toGlobalPoint: function(point) {
    //   return fabric.util.rotatePoint(point, this.getCenterPoint(), degreesToRadians(this.angle)).addEquals(new fabric.Point(this.left, this.top));
    // },

    /**
     * Sets the position of the object taking into consideration the object's origin
     * @param {fabric.Point} pos The new position of the object
     * @param {String} originX Horizontal origin: 'left', 'center' or 'right'
     * @param {String} originY Vertical origin: 'top', 'center' or 'bottom'
     * @return {void}
     */
    setPositionByOrigin: function(pos, originX, originY) {
      var center = this.translateToCenterPoint(pos, originX, originY),
          position = this.translateToOriginPoint(center, this.originX, this.originY);

      this.set('left', position.x);
      this.set('top', position.y);
    },

    /**
     * @param {String} to One of 'left', 'center', 'right'
     */
    adjustPosition: function(to) {
      var angle = degreesToRadians(this.angle),
          hypotFull = this.getWidth(),
          xFull = Math.cos(angle) * hypotFull,
          yFull = Math.sin(angle) * hypotFull;

      //TODO: this function does not consider mixed situation like top, center.
      this.left += xFull * (originXOffset[to] - originXOffset[this.originX]);
      this.top += yFull * (originXOffset[to] - originXOffset[this.originX]);

      this.setCoords();
      this.originX = to;
    },

    /**
     * Sets the origin/position of the object to it's center point
     * @private
     * @return {void}
     */
    _setOriginToCenter: function() {
      this._originalOriginX = this.originX;
      this._originalOriginY = this.originY;

      var center = this.getCenterPoint();

      this.originX = 'center';
      this.originY = 'center';

      this.left = center.x;
      this.top = center.y;
    },

    /**
     * Resets the origin/position of the object to it's original origin
     * @private
     * @return {void}
     */
    _resetOrigin: function() {
      var originPoint = this.translateToOriginPoint(
        this.getCenterPoint(),
        this._originalOriginX,
        this._originalOriginY);

      this.originX = this._originalOriginX;
      this.originY = this._originalOriginY;

      this.left = originPoint.x;
      this.top = originPoint.y;

      this._originalOriginX = null;
      this._originalOriginY = null;
    },

    /**
     * @private
     */
    _getLeftTopCoords: function() {
      return this.translateToOriginPoint(this.getCenterPoint(), 'left', 'top');
    }
  });

})();


(function() {

  function getCoords(oCoords) {
    return [
      new fabric.Point(oCoords.tl.x, oCoords.tl.y),
      new fabric.Point(oCoords.tr.x, oCoords.tr.y),
      new fabric.Point(oCoords.br.x, oCoords.br.y),
      new fabric.Point(oCoords.bl.x, oCoords.bl.y)
    ];
  }

  var degreesToRadians = fabric.util.degreesToRadians,
      multiplyMatrices = fabric.util.multiplyTransformMatrices;

  fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

    /**
     * Object containing coordinates of object's controls
     * @type Object
     * @default
     */
    oCoords: null,

    /**
     * Checks if object intersects with an area formed by 2 points
     * @param {Object} pointTL top-left point of area
     * @param {Object} pointBR bottom-right point of area
     * @return {Boolean} true if object intersects with an area formed by 2 points
     */
    intersectsWithRect: function(pointTL, pointBR) {
      var oCoords = getCoords(this.oCoords),
          intersection = fabric.Intersection.intersectPolygonRectangle(
            oCoords,
            pointTL,
            pointBR
          );
      return intersection.status === 'Intersection';
    },

    /**
     * Checks if object intersects with another object
     * @param {Object} other Object to test
     * @return {Boolean} true if object intersects with another object
     */
    intersectsWithObject: function(other) {
      var intersection = fabric.Intersection.intersectPolygonPolygon(
            getCoords(this.oCoords),
            getCoords(other.oCoords)
          );

      return intersection.status === 'Intersection';
    },

    /**
     * Checks if object is fully contained within area of another object
     * @param {Object} other Object to test
     * @return {Boolean} true if object is fully contained within area of another object
     */
    isContainedWithinObject: function(other) {
      var boundingRect = other.getBoundingRect(),
          point1 = new fabric.Point(boundingRect.left, boundingRect.top),
          point2 = new fabric.Point(boundingRect.left + boundingRect.width, boundingRect.top + boundingRect.height);

      return this.isContainedWithinRect(point1, point2);
    },

    /**
     * Checks if object is fully contained within area formed by 2 points
     * @param {Object} pointTL top-left point of area
     * @param {Object} pointBR bottom-right point of area
     * @return {Boolean} true if object is fully contained within area formed by 2 points
     */
    isContainedWithinRect: function(pointTL, pointBR) {
      var boundingRect = this.getBoundingRect();

      return (
        boundingRect.left >= pointTL.x &&
        boundingRect.left + boundingRect.width <= pointBR.x &&
        boundingRect.top >= pointTL.y &&
        boundingRect.top + boundingRect.height <= pointBR.y
      );
    },

    /**
     * Checks if point is inside the object
     * @param {fabric.Point} point Point to check against
     * @return {Boolean} true if point is inside the object
     */
    containsPoint: function(point) {
      if (!this.oCoords) {
        this.setCoords();
      }
      var lines = this._getImageLines(this.oCoords),
          xPoints = this._findCrossPoints(point, lines);

      // if xPoints is odd then point is inside the object
      return (xPoints !== 0 && xPoints % 2 === 1);
    },

    /**
     * Method that returns an object with the object edges in it, given the coordinates of the corners
     * @private
     * @param {Object} oCoords Coordinates of the object corners
     */
    _getImageLines: function(oCoords) {
      return {
        topline: {
          o: oCoords.tl,
          d: oCoords.tr
        },
        rightline: {
          o: oCoords.tr,
          d: oCoords.br
        },
        bottomline: {
          o: oCoords.br,
          d: oCoords.bl
        },
        leftline: {
          o: oCoords.bl,
          d: oCoords.tl
        }
      };
    },

    /**
     * Helper method to determine how many cross points are between the 4 object edges
     * and the horizontal line determined by a point on canvas
     * @private
     * @param {fabric.Point} point Point to check
     * @param {Object} oCoords Coordinates of the object being evaluated
     */
    _findCrossPoints: function(point, oCoords) {
      var b1, b2, a1, a2, xi, yi,
          xcount = 0,
          iLine;

      for (var lineKey in oCoords) {
        iLine = oCoords[lineKey];
        // optimisation 1: line below point. no cross
        if ((iLine.o.y < point.y) && (iLine.d.y < point.y)) {
          continue;
        }
        // optimisation 2: line above point. no cross
        if ((iLine.o.y >= point.y) && (iLine.d.y >= point.y)) {
          continue;
        }
        // optimisation 3: vertical line case
        if ((iLine.o.x === iLine.d.x) && (iLine.o.x >= point.x)) {
          xi = iLine.o.x;
          yi = point.y;
        }
        // calculate the intersection point
        else {
          b1 = 0;
          b2 = (iLine.d.y - iLine.o.y) / (iLine.d.x - iLine.o.x);
          a1 = point.y - b1 * point.x;
          a2 = iLine.o.y - b2 * iLine.o.x;

          xi = - (a1 - a2) / (b1 - b2);
          yi = a1 + b1 * xi;
        }
        // dont count xi < point.x cases
        if (xi >= point.x) {
          xcount += 1;
        }
        // optimisation 4: specific for square images
        if (xcount === 2) {
          break;
        }
      }
      return xcount;
    },

    /**
     * Returns width of an object's bounding rectangle
     * @deprecated since 1.0.4
     * @return {Number} width value
     */
    getBoundingRectWidth: function() {
      return this.getBoundingRect().width;
    },

    /**
     * Returns height of an object's bounding rectangle
     * @deprecated since 1.0.4
     * @return {Number} height value
     */
    getBoundingRectHeight: function() {
      return this.getBoundingRect().height;
    },

    /**
     * Returns coordinates of object's bounding rectangle (left, top, width, height)
     * @return {Object} Object with left, top, width, height properties
     */
    getBoundingRect: function() {
      this.oCoords || this.setCoords();
      return fabric.util.makeBoundingBoxFromPoints([
        this.oCoords.tl,
        this.oCoords.tr,
        this.oCoords.br,
        this.oCoords.bl
      ]);
    },

    /**
     * Returns width of an object bounding box counting transformations
     * @return {Number} width value
     */
    getWidth: function() {
      return this._getTransformedDimensions().x;
    },

    /**
     * Returns height of an object bounding box counting transformations
     * to be renamed in 2.0
     * @return {Number} height value
     */
    getHeight: function() {
      return this._getTransformedDimensions().y;
    },

    /**
     * Makes sure the scale is valid and modifies it if necessary
     * @private
     * @param {Number} value
     * @return {Number}
     */
    _constrainScale: function(value) {
      if (Math.abs(value) < this.minScaleLimit) {
        if (value < 0) {
          return -this.minScaleLimit;
        }
        else {
          return this.minScaleLimit;
        }
      }
      return value;
    },

    /**
     * Scales an object (equally by x and y)
     * @param {Number} value Scale factor
     * @return {fabric.Object} thisArg
     * @chainable
     */
    scale: function(value) {
      value = this._constrainScale(value);

      if (value < 0) {
        this.flipX = !this.flipX;
        this.flipY = !this.flipY;
        value *= -1;
      }

      this.scaleX = value;
      this.scaleY = value;
      this.setCoords();
      return this;
    },

    /**
     * Scales an object to a given width, with respect to bounding box (scaling by x/y equally)
     * @param {Number} value New width value
     * @return {fabric.Object} thisArg
     * @chainable
     */
    scaleToWidth: function(value) {
      // adjust to bounding rect factor so that rotated shapes would fit as well
      var boundingRectFactor = this.getBoundingRect().width / this.getWidth();
      return this.scale(value / this.width / boundingRectFactor);
    },

    /**
     * Scales an object to a given height, with respect to bounding box (scaling by x/y equally)
     * @param {Number} value New height value
     * @return {fabric.Object} thisArg
     * @chainable
     */
    scaleToHeight: function(value) {
      // adjust to bounding rect factor so that rotated shapes would fit as well
      var boundingRectFactor = this.getBoundingRect().height / this.getHeight();
      return this.scale(value / this.height / boundingRectFactor);
    },

    /**
     * Sets corner position coordinates based on current angle, width and height
     * See https://github.com/kangax/fabric.js/wiki/When-to-call-setCoords
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setCoords: function() {
      var theta = degreesToRadians(this.angle),
          vpt = this.getViewportTransform(),
          dim = this._calculateCurrentDimensions(),
          currentWidth = dim.x, currentHeight = dim.y;

      // If width is negative, make postive. Fixes path selection issue
      if (currentWidth < 0) {
        currentWidth = Math.abs(currentWidth);
      }

      var sinTh = Math.sin(theta),
          cosTh = Math.cos(theta),
          _angle = currentWidth > 0 ? Math.atan(currentHeight / currentWidth) : 0,
          _hypotenuse = (currentWidth / Math.cos(_angle)) / 2,
          offsetX = Math.cos(_angle + theta) * _hypotenuse,
          offsetY = Math.sin(_angle + theta) * _hypotenuse,

          // offset added for rotate and scale actions
          coords = fabric.util.transformPoint(this.getCenterPoint(), vpt),
          tl  = new fabric.Point(coords.x - offsetX, coords.y - offsetY),
          tr  = new fabric.Point(tl.x + (currentWidth * cosTh), tl.y + (currentWidth * sinTh)),
          bl  = new fabric.Point(tl.x - (currentHeight * sinTh), tl.y + (currentHeight * cosTh)),
          br  = new fabric.Point(coords.x + offsetX, coords.y + offsetY),
          ml  = new fabric.Point((tl.x + bl.x)/2, (tl.y + bl.y)/2),
          mt  = new fabric.Point((tr.x + tl.x)/2, (tr.y + tl.y)/2),
          mr  = new fabric.Point((br.x + tr.x)/2, (br.y + tr.y)/2),
          mb  = new fabric.Point((br.x + bl.x)/2, (br.y + bl.y)/2),
          mtr = new fabric.Point(mt.x + sinTh * this.rotatingPointOffset, mt.y - cosTh * this.rotatingPointOffset);
      // debugging

      /* setTimeout(function() {
         canvas.contextTop.fillStyle = 'green';
         canvas.contextTop.fillRect(mb.x, mb.y, 3, 3);
         canvas.contextTop.fillRect(bl.x, bl.y, 3, 3);
         canvas.contextTop.fillRect(br.x, br.y, 3, 3);
         canvas.contextTop.fillRect(tl.x, tl.y, 3, 3);
         canvas.contextTop.fillRect(tr.x, tr.y, 3, 3);
         canvas.contextTop.fillRect(ml.x, ml.y, 3, 3);
         canvas.contextTop.fillRect(mr.x, mr.y, 3, 3);
         canvas.contextTop.fillRect(mt.x, mt.y, 3, 3);
         canvas.contextTop.fillRect(mtr.x, mtr.y, 3, 3);
       }, 50); */

      this.oCoords = {
        // corners
        tl: tl, tr: tr, br: br, bl: bl,
        // middle
        ml: ml, mt: mt, mr: mr, mb: mb,
        // rotating point
        mtr: mtr
      };

      // set coordinates of the draggable boxes in the corners used to scale/rotate the image
      this._setCornerCoords && this._setCornerCoords();

      return this;
    },

    _calcRotateMatrix: function() {
      if (this.angle) {
        var theta = degreesToRadians(this.angle), cos = Math.cos(theta), sin = Math.sin(theta);
        return [cos, sin, -sin, cos, 0, 0];
      }
      return [1, 0, 0, 1, 0, 0];
    },

    /*
     * calculate trasform Matrix that represent current transformation from
     * object properties.
     * @return {Array} matrix Transform Matrix for the object
     */
    calcTransformMatrix: function() {
      var center = this.getCenterPoint(),
          translateMatrix = [1, 0, 0, 1, center.x, center.y],
          rotateMatrix = this._calcRotateMatrix(),
          dimensionMatrix = this._calcDimensionsTransformMatrix(this.skewX, this.skewY, true),
          matrix = this.group ? this.group.calcTransformMatrix() : [1, 0, 0, 1, 0, 0];
      matrix = multiplyMatrices(matrix, translateMatrix);
      matrix = multiplyMatrices(matrix, rotateMatrix);
      matrix = multiplyMatrices(matrix, dimensionMatrix);
      return matrix;
    },

    _calcDimensionsTransformMatrix: function(skewX, skewY, flipping) {
      var skewMatrixX = [1, 0, Math.tan(degreesToRadians(skewX)), 1],
          skewMatrixY = [1, Math.tan(degreesToRadians(skewY)), 0, 1],
          scaleX = this.scaleX * (flipping && this.flipX ? -1 : 1),
          scaleY = this.scaleY * (flipping && this.flipY ? -1 : 1),
          scaleMatrix = [scaleX, 0, 0, scaleY],
          m = multiplyMatrices(scaleMatrix, skewMatrixX, true);
      return multiplyMatrices(m, skewMatrixY, true);
    }
  });
})();


fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

  /**
   * Moves an object to the bottom of the stack of drawn objects
   * @return {fabric.Object} thisArg
   * @chainable
   */
  sendToBack: function() {
    if (this.group) {
      fabric.StaticCanvas.prototype.sendToBack.call(this.group, this);
    }
    else {
      this.canvas.sendToBack(this);
    }
    return this;
  },

  /**
   * Moves an object to the top of the stack of drawn objects
   * @return {fabric.Object} thisArg
   * @chainable
   */
  bringToFront: function() {
    if (this.group) {
      fabric.StaticCanvas.prototype.bringToFront.call(this.group, this);
    }
    else {
      this.canvas.bringToFront(this);
    }
    return this;
  },

  /**
   * Moves an object down in stack of drawn objects
   * @param {Boolean} [intersecting] If `true`, send object behind next lower intersecting object
   * @return {fabric.Object} thisArg
   * @chainable
   */
  sendBackwards: function(intersecting) {
    if (this.group) {
      fabric.StaticCanvas.prototype.sendBackwards.call(this.group, this, intersecting);
    }
    else {
      this.canvas.sendBackwards(this, intersecting);
    }
    return this;
  },

  /**
   * Moves an object up in stack of drawn objects
   * @param {Boolean} [intersecting] If `true`, send object in front of next upper intersecting object
   * @return {fabric.Object} thisArg
   * @chainable
   */
  bringForward: function(intersecting) {
    if (this.group) {
      fabric.StaticCanvas.prototype.bringForward.call(this.group, this, intersecting);
    }
    else {
      this.canvas.bringForward(this, intersecting);
    }
    return this;
  },

  /**
   * Moves an object to specified level in stack of drawn objects
   * @param {Number} index New position of object
   * @return {fabric.Object} thisArg
   * @chainable
   */
  moveTo: function(index) {
    if (this.group) {
      fabric.StaticCanvas.prototype.moveTo.call(this.group, this, index);
    }
    else {
      this.canvas.moveTo(this, index);
    }
    return this;
  }
});


/* _TO_SVG_START_ */
(function() {

  function getSvgColorString(prop, value) {
    if (!value) {
      return prop + ': none; ';
    }
    else if (value.toLive) {
      return prop + ': url(#SVGID_' + value.id + '); ';
    }
    else {
      var color = new fabric.Color(value),
          str = prop + ': ' + color.toRgb() + '; ',
          opacity = color.getAlpha();
      if (opacity !== 1) {
        //change the color in rgb + opacity
        str += prop + '-opacity: ' + opacity.toString() + '; ';
      }
      return str;
    }
  }

  fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {
    /**
     * Returns styles-string for svg-export
     * @param {Boolean} skipShadow a boolean to skip shadow filter output
     * @return {String}
     */
    getSvgStyles: function(skipShadow) {

      var fillRule = this.fillRule,
          strokeWidth = this.strokeWidth ? this.strokeWidth : '0',
          strokeDashArray = this.strokeDashArray ? this.strokeDashArray.join(' ') : 'none',
          strokeLineCap = this.strokeLineCap ? this.strokeLineCap : 'butt',
          strokeLineJoin = this.strokeLineJoin ? this.strokeLineJoin : 'miter',
          strokeMiterLimit = this.strokeMiterLimit ? this.strokeMiterLimit : '4',
          opacity = typeof this.opacity !== 'undefined' ? this.opacity : '1',
          visibility = this.visible ? '' : ' visibility: hidden;',
          filter = skipShadow ? '' : this.getSvgFilter(),
          fill = getSvgColorString('fill', this.fill),
          stroke = getSvgColorString('stroke', this.stroke);

      return [
        stroke,
        'stroke-width: ', strokeWidth, '; ',
        'stroke-dasharray: ', strokeDashArray, '; ',
        'stroke-linecap: ', strokeLineCap, '; ',
        'stroke-linejoin: ', strokeLineJoin, '; ',
        'stroke-miterlimit: ', strokeMiterLimit, '; ',
        fill,
        'fill-rule: ', fillRule, '; ',
        'opacity: ', opacity, ';',
        filter,
        visibility
      ].join('');
    },

    /**
     * Returns filter for svg shadow
     * @return {String}
     */
    getSvgFilter: function() {
      return this.shadow ? 'filter: url(#SVGID_' + this.shadow.id + ');' : '';
    },

    /**
     * Returns id attribute for svg output
     * @return {String}
     */
    getSvgId: function() {
      return this.id ? 'id="' + this.id + '" ' : '';
    },

    /**
     * Returns transform-string for svg-export
     * @return {String}
     */
    getSvgTransform: function() {
      if (this.group && this.group.type === 'path-group') {
        return '';
      }
      var toFixed = fabric.util.toFixed,
          angle = this.getAngle(),
          skewX = (this.getSkewX() % 360),
          skewY = (this.getSkewY() % 360),
          center = this.getCenterPoint(),

          NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS,

          translatePart = this.type === 'path-group' ? '' : 'translate(' +
                            toFixed(center.x, NUM_FRACTION_DIGITS) +
                            ' ' +
                            toFixed(center.y, NUM_FRACTION_DIGITS) +
                          ')',

          anglePart = angle !== 0
            ? (' rotate(' + toFixed(angle, NUM_FRACTION_DIGITS) + ')')
            : '',

          scalePart = (this.scaleX === 1 && this.scaleY === 1)
            ? '' :
            (' scale(' +
              toFixed(this.scaleX, NUM_FRACTION_DIGITS) +
              ' ' +
              toFixed(this.scaleY, NUM_FRACTION_DIGITS) +
            ')'),

          skewXPart = skewX !== 0 ? ' skewX(' + toFixed(skewX, NUM_FRACTION_DIGITS) + ')' : '',

          skewYPart = skewY !== 0 ? ' skewY(' + toFixed(skewY, NUM_FRACTION_DIGITS) + ')' : '',

          addTranslateX = this.type === 'path-group' ? this.width : 0,

          flipXPart = this.flipX ? ' matrix(-1 0 0 1 ' + addTranslateX + ' 0) ' : '',

          addTranslateY = this.type === 'path-group' ? this.height : 0,

          flipYPart = this.flipY ? ' matrix(1 0 0 -1 0 ' + addTranslateY + ')' : '';

      return [
        translatePart, anglePart, scalePart, flipXPart, flipYPart, skewXPart, skewYPart
      ].join('');
    },

    /**
     * Returns transform-string for svg-export from the transform matrix of single elements
     * @return {String}
     */
    getSvgTransformMatrix: function() {
      return this.transformMatrix ? ' matrix(' + this.transformMatrix.join(' ') + ') ' : '';
    },

    /**
     * @private
     */
    _createBaseSVGMarkup: function() {
      var markup = [ ];

      if (this.fill && this.fill.toLive) {
        markup.push(this.fill.toSVG(this, false));
      }
      if (this.stroke && this.stroke.toLive) {
        markup.push(this.stroke.toSVG(this, false));
      }
      if (this.shadow) {
        markup.push(this.shadow.toSVG(this));
      }
      return markup;
    }
  });
})();
/* _TO_SVG_END_ */


/*
  Depends on `stateProperties`
*/
fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

  /**
   * Returns true if object state (one of its state properties) was changed
   * @return {Boolean} true if instance' state has changed since `{@link fabric.Object#saveState}` was called
   */
  hasStateChanged: function() {
    return this.stateProperties.some(function(prop) {
      return this.get(prop) !== this.originalState[prop];
    }, this);
  },

  /**
   * Saves state of an object
   * @param {Object} [options] Object with additional `stateProperties` array to include when saving state
   * @return {fabric.Object} thisArg
   */
  saveState: function(options) {
    this.stateProperties.forEach(function(prop) {
      this.originalState[prop] = this.get(prop);
    }, this);

    if (options && options.stateProperties) {
      options.stateProperties.forEach(function(prop) {
        this.originalState[prop] = this.get(prop);
      }, this);
    }

    return this;
  },

  /**
   * Setups state of an object
   * @return {fabric.Object} thisArg
   */
  setupState: function() {
    this.originalState = { };
    this.saveState();

    return this;
  }
});


(function() {

  var degreesToRadians = fabric.util.degreesToRadians,
      //jscs:disable requireCamelCaseOrUpperCaseIdentifiers
      isVML = function() { return typeof G_vmlCanvasManager !== 'undefined'; };
  //jscs:enable requireCamelCaseOrUpperCaseIdentifiers

  fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

    /**
     * The object interactivity controls.
     * @private
     */
    _controlsVisibility: null,

    /**
     * Determines which corner has been clicked
     * @private
     * @param {Object} pointer The pointer indicating the mouse position
     * @return {String|Boolean} corner code (tl, tr, bl, br, etc.), or false if nothing is found
     */
    _findTargetCorner: function(pointer) {
      if (!this.hasControls || !this.active) {
        return false;
      }

      var ex = pointer.x,
          ey = pointer.y,
          xPoints,
          lines;
      this.__corner = 0;
      for (var i in this.oCoords) {

        if (!this.isControlVisible(i)) {
          continue;
        }

        if (i === 'mtr' && !this.hasRotatingPoint) {
          continue;
        }

        if (this.get('lockUniScaling') &&
           (i === 'mt' || i === 'mr' || i === 'mb' || i === 'ml')) {
          continue;
        }

        lines = this._getImageLines(this.oCoords[i].corner);

        // debugging

        // canvas.contextTop.fillRect(lines.bottomline.d.x, lines.bottomline.d.y, 2, 2);
        // canvas.contextTop.fillRect(lines.bottomline.o.x, lines.bottomline.o.y, 2, 2);

        // canvas.contextTop.fillRect(lines.leftline.d.x, lines.leftline.d.y, 2, 2);
        // canvas.contextTop.fillRect(lines.leftline.o.x, lines.leftline.o.y, 2, 2);

        // canvas.contextTop.fillRect(lines.topline.d.x, lines.topline.d.y, 2, 2);
        // canvas.contextTop.fillRect(lines.topline.o.x, lines.topline.o.y, 2, 2);

        // canvas.contextTop.fillRect(lines.rightline.d.x, lines.rightline.d.y, 2, 2);
        // canvas.contextTop.fillRect(lines.rightline.o.x, lines.rightline.o.y, 2, 2);

        xPoints = this._findCrossPoints({ x: ex, y: ey }, lines);
        if (xPoints !== 0 && xPoints % 2 === 1) {
          this.__corner = i;
          return i;
        }
      }
      return false;
    },

    /**
     * Sets the coordinates of the draggable boxes in the corners of
     * the image used to scale/rotate it.
     * @private
     */
    _setCornerCoords: function() {
      var coords = this.oCoords,
          newTheta = degreesToRadians(45 - this.angle),
          /* Math.sqrt(2 * Math.pow(this.cornerSize, 2)) / 2, */
          /* 0.707106 stands for sqrt(2)/2 */
          cornerHypotenuse = this.cornerSize * 0.707106,
          cosHalfOffset = cornerHypotenuse * Math.cos(newTheta),
          sinHalfOffset = cornerHypotenuse * Math.sin(newTheta),
          x, y;

      for (var point in coords) {
        x = coords[point].x;
        y = coords[point].y;
        coords[point].corner = {
          tl: {
            x: x - sinHalfOffset,
            y: y - cosHalfOffset
          },
          tr: {
            x: x + cosHalfOffset,
            y: y - sinHalfOffset
          },
          bl: {
            x: x - cosHalfOffset,
            y: y + sinHalfOffset
          },
          br: {
            x: x + sinHalfOffset,
            y: y + cosHalfOffset
          }
        };
      }
    },

    /*
     * Calculate object dimensions from its properties
     * @private
     */
    _getNonTransformedDimensions: function() {
      var strokeWidth = this.strokeWidth,
          w = this.width,
          h = this.height,
          addStrokeToW = true,
          addStrokeToH = true;

      if (this.type === 'line' && this.strokeLineCap === 'butt') {
        addStrokeToH = w;
        addStrokeToW = h;
      }

      if (addStrokeToH) {
        h += h < 0 ? -strokeWidth : strokeWidth;
      }

      if (addStrokeToW) {
        w += w < 0 ? -strokeWidth : strokeWidth;
      }

      return { x: w, y: h };
    },

    /*
     * @private
     */
    _getTransformedDimensions: function(skewX, skewY) {
      if (typeof skewX === 'undefined') {
        skewX = this.skewX;
      }
      if (typeof skewY === 'undefined') {
        skewY = this.skewY;
      }
      var dimensions = this._getNonTransformedDimensions(),
          dimX = dimensions.x /2, dimY = dimensions.y / 2,
          points = [
          {
            x: -dimX,
            y: -dimY
          },
          {
            x: dimX,
            y: -dimY
          },
          {
            x: -dimX,
            y: dimY
          },
          {
            x: dimX,
            y: dimY
          }],
          i, transformMatrix = this._calcDimensionsTransformMatrix(skewX, skewY, false),
          bbox;
      for (i = 0; i < points.length; i++) {
        points[i] = fabric.util.transformPoint(points[i], transformMatrix);
      }
      bbox = fabric.util.makeBoundingBoxFromPoints(points);
      return { x: bbox.width, y: bbox.height };
    },

    /*
     * private
     */
    _calculateCurrentDimensions: function()  {
      var vpt = this.getViewportTransform(),
          dim = this._getTransformedDimensions(),
          w = dim.x, h = dim.y,
          p = fabric.util.transformPoint(new fabric.Point(w, h), vpt, true);

      return p.scalarAdd(2 * this.padding);
    },

    /**
     * Draws a colored layer behind the object, inside its selection borders.
     * Requires public options: padding, selectionBackgroundColor
     * this function is called when the context is transformed
     * @param {CanvasRenderingContext2D} ctx Context to draw on
     * @return {fabric.Object} thisArg
     * @chainable
     */
    drawSelectionBackground: function(ctx) {
      if (!this.selectionBackgroundColor || this.group
        || this !== this.canvas.getActiveObject()) {
        return this;
      }
      ctx.save();
      var center = this.getCenterPoint(), wh = this._calculateCurrentDimensions(),
          vpt = this.canvas.viewportTransform;
      ctx.translate(center.x, center.y);
      ctx.scale(1 / vpt[0], 1 / vpt[3]);
      ctx.rotate(degreesToRadians(this.angle));
      ctx.fillStyle = this.selectionBackgroundColor;
      ctx.fillRect(-wh.x/2, -wh.y/2, wh.x, wh.y);
      ctx.restore();
      return this;
    },

    /**
     * Draws borders of an object's bounding box.
     * Requires public properties: width, height
     * Requires public options: padding, borderColor
     * @param {CanvasRenderingContext2D} ctx Context to draw on
     * @return {fabric.Object} thisArg
     * @chainable
     */
    drawBorders: function(ctx) {
      if (!this.hasBorders) {
        return this;
      }

      var wh = this._calculateCurrentDimensions(),
          strokeWidth = 1 / this.borderScaleFactor,
          width = wh.x + strokeWidth,
          height = wh.y + strokeWidth;

      ctx.save();
      ctx.strokeStyle = this.borderColor;
      this._setLineDash(ctx, this.borderDashArray, null);

      ctx.strokeRect(
        -width / 2,
        -height / 2,
        width,
        height
      );

      if (this.hasRotatingPoint && this.isControlVisible('mtr') && !this.get('lockRotation') && this.hasControls) {

        var rotateHeight = -height / 2;

        ctx.beginPath();
        ctx.moveTo(0, rotateHeight);
        ctx.lineTo(0, rotateHeight - this.rotatingPointOffset);
        ctx.closePath();
        ctx.stroke();
      }

      ctx.restore();
      return this;
    },

    /**
     * Draws borders of an object's bounding box when it is inside a group.
     * Requires public properties: width, height
     * Requires public options: padding, borderColor
     * @param {CanvasRenderingContext2D} ctx Context to draw on
     * @param {object} options object representing current object parameters
     * @return {fabric.Object} thisArg
     * @chainable
     */
    drawBordersInGroup: function(ctx, options) {
      if (!this.hasBorders) {
        return this;
      }

      var p = this._getNonTransformedDimensions(),
          matrix = fabric.util.customTransformMatrix(options.scaleX, options.scaleY, options.skewX),
          wh = fabric.util.transformPoint(p, matrix),
          strokeWidth = 1 / this.borderScaleFactor,
          width = wh.x + strokeWidth + 2 * this.padding,
          height = wh.y + strokeWidth + 2 * this.padding;

      ctx.save();
      this._setLineDash(ctx, this.borderDashArray, null);
      ctx.strokeStyle = this.borderColor;

      ctx.strokeRect(
        -width / 2,
        -height / 2,
        width,
        height
      );

      ctx.restore();
      return this;
    },

    /**
     * Draws corners of an object's bounding box.
     * Requires public properties: width, height
     * Requires public options: cornerSize, padding
     * @param {CanvasRenderingContext2D} ctx Context to draw on
     * @return {fabric.Object} thisArg
     * @chainable
     */
    drawControls: function(ctx) {
      if (!this.hasControls) {
        return this;
      }

      var wh = this._calculateCurrentDimensions(),
          width = wh.x,
          height = wh.y,
          scaleOffset = this.cornerSize,
          left = -(width + scaleOffset) / 2,
          top = -(height + scaleOffset) / 2,
          methodName = this.transparentCorners ? 'stroke' : 'fill';

      ctx.save();
      ctx.strokeStyle = ctx.fillStyle = this.cornerColor;
      if (!this.transparentCorners) {
        ctx.strokeStyle = this.cornerStrokeColor;
      }
      this._setLineDash(ctx, this.cornerDashArray, null);

      // top-left
      this._drawControl('tl', ctx, methodName,
        left,
        top);

      // top-right
      this._drawControl('tr', ctx, methodName,
        left + width,
        top);

      // bottom-left
      this._drawControl('bl', ctx, methodName,
        left,
        top + height);

      // bottom-right
      this._drawControl('br', ctx, methodName,
        left + width,
        top + height);

      if (!this.get('lockUniScaling')) {

        // middle-top
        this._drawControl('mt', ctx, methodName,
          left + width/2,
          top);

        // middle-bottom
        this._drawControl('mb', ctx, methodName,
          left + width/2,
          top + height);

        // middle-right
        this._drawControl('mr', ctx, methodName,
          left + width,
          top + height/2);

        // middle-left
        this._drawControl('ml', ctx, methodName,
          left,
          top + height/2);
      }

      // middle-top-rotate
      if (this.hasRotatingPoint) {
        this._drawControl('mtr', ctx, methodName,
          left + width / 2,
          top - this.rotatingPointOffset);
      }

      ctx.restore();

      return this;
    },

    /**
     * @private
     */
    _drawControl: function(control, ctx, methodName, left, top) {
      if (!this.isControlVisible(control)) {
        return;
      }
      var size = this.cornerSize, stroke = !this.transparentCorners && this.cornerStrokeColor;
      switch (this.cornerStyle) {
        case 'circle':
          ctx.beginPath();
          ctx.arc(left + size/2, top + size/2, size/2, 0, 2 * Math.PI, false);
          ctx[methodName]();
          if (stroke) {
            ctx.stroke();
          }
          break;
        default:
          isVML() || this.transparentCorners || ctx.clearRect(left, top, size, size);
          ctx[methodName + 'Rect'](left, top, size, size);
          if (stroke) {
            ctx.strokeRect(left, top, size, size);
          }
      }
    },

    /**
     * Returns true if the specified control is visible, false otherwise.
     * @param {String} controlName The name of the control. Possible values are 'tl', 'tr', 'br', 'bl', 'ml', 'mt', 'mr', 'mb', 'mtr'.
     * @returns {Boolean} true if the specified control is visible, false otherwise
     */
    isControlVisible: function(controlName) {
      return this._getControlsVisibility()[controlName];
    },

    /**
     * Sets the visibility of the specified control.
     * @param {String} controlName The name of the control. Possible values are 'tl', 'tr', 'br', 'bl', 'ml', 'mt', 'mr', 'mb', 'mtr'.
     * @param {Boolean} visible true to set the specified control visible, false otherwise
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setControlVisible: function(controlName, visible) {
      this._getControlsVisibility()[controlName] = visible;
      return this;
    },

    /**
     * Sets the visibility state of object controls.
     * @param {Object} [options] Options object
     * @param {Boolean} [options.bl] true to enable the bottom-left control, false to disable it
     * @param {Boolean} [options.br] true to enable the bottom-right control, false to disable it
     * @param {Boolean} [options.mb] true to enable the middle-bottom control, false to disable it
     * @param {Boolean} [options.ml] true to enable the middle-left control, false to disable it
     * @param {Boolean} [options.mr] true to enable the middle-right control, false to disable it
     * @param {Boolean} [options.mt] true to enable the middle-top control, false to disable it
     * @param {Boolean} [options.tl] true to enable the top-left control, false to disable it
     * @param {Boolean} [options.tr] true to enable the top-right control, false to disable it
     * @param {Boolean} [options.mtr] true to enable the middle-top-rotate control, false to disable it
     * @return {fabric.Object} thisArg
     * @chainable
     */
    setControlsVisibility: function(options) {
      options || (options = { });

      for (var p in options) {
        this.setControlVisible(p, options[p]);
      }
      return this;
    },

    /**
     * Returns the instance of the control visibility set for this object.
     * @private
     * @returns {Object}
     */
    _getControlsVisibility: function() {
      if (!this._controlsVisibility) {
        this._controlsVisibility = {
          tl: true,
          tr: true,
          br: true,
          bl: true,
          ml: true,
          mt: true,
          mr: true,
          mb: true,
          mtr: true
        };
      }
      return this._controlsVisibility;
    }
  });
})();


fabric.util.object.extend(fabric.StaticCanvas.prototype, /** @lends fabric.StaticCanvas.prototype */ {

  /**
   * Animation duration (in ms) for fx* methods
   * @type Number
   * @default
   */
  FX_DURATION: 500,

  /**
   * Centers object horizontally with animation.
   * @param {fabric.Object} object Object to center
   * @param {Object} [callbacks] Callbacks object with optional "onComplete" and/or "onChange" properties
   * @param {Function} [callbacks.onComplete] Invoked on completion
   * @param {Function} [callbacks.onChange] Invoked on every step of animation
   * @return {fabric.Canvas} thisArg
   * @chainable
   */
  fxCenterObjectH: function (object, callbacks) {
    callbacks = callbacks || { };

    var empty = function() { },
        onComplete = callbacks.onComplete || empty,
        onChange = callbacks.onChange || empty,
        _this = this;

    fabric.util.animate({
      startValue: object.get('left'),
      endValue: this.getCenter().left,
      duration: this.FX_DURATION,
      onChange: function(value) {
        object.set('left', value);
        _this.renderAll();
        onChange();
      },
      onComplete: function() {
        object.setCoords();
        onComplete();
      }
    });

    return this;
  },

  /**
   * Centers object vertically with animation.
   * @param {fabric.Object} object Object to center
   * @param {Object} [callbacks] Callbacks object with optional "onComplete" and/or "onChange" properties
   * @param {Function} [callbacks.onComplete] Invoked on completion
   * @param {Function} [callbacks.onChange] Invoked on every step of animation
   * @return {fabric.Canvas} thisArg
   * @chainable
   */
  fxCenterObjectV: function (object, callbacks) {
    callbacks = callbacks || { };

    var empty = function() { },
        onComplete = callbacks.onComplete || empty,
        onChange = callbacks.onChange || empty,
        _this = this;

    fabric.util.animate({
      startValue: object.get('top'),
      endValue: this.getCenter().top,
      duration: this.FX_DURATION,
      onChange: function(value) {
        object.set('top', value);
        _this.renderAll();
        onChange();
      },
      onComplete: function() {
        object.setCoords();
        onComplete();
      }
    });

    return this;
  },

  /**
   * Same as `fabric.Canvas#remove` but animated
   * @param {fabric.Object} object Object to remove
   * @param {Object} [callbacks] Callbacks object with optional "onComplete" and/or "onChange" properties
   * @param {Function} [callbacks.onComplete] Invoked on completion
   * @param {Function} [callbacks.onChange] Invoked on every step of animation
   * @return {fabric.Canvas} thisArg
   * @chainable
   */
  fxRemove: function (object, callbacks) {
    callbacks = callbacks || { };

    var empty = function() { },
        onComplete = callbacks.onComplete || empty,
        onChange = callbacks.onChange || empty,
        _this = this;

    fabric.util.animate({
      startValue: object.get('opacity'),
      endValue: 0,
      duration: this.FX_DURATION,
      onStart: function() {
        object.set('active', false);
      },
      onChange: function(value) {
        object.set('opacity', value);
        _this.renderAll();
        onChange();
      },
      onComplete: function () {
        _this.remove(object);
        onComplete();
      }
    });

    return this;
  }
});

fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {
  /**
   * Animates object's properties
   * @param {String|Object} property Property to animate (if string) or properties to animate (if object)
   * @param {Number|Object} value Value to animate property to (if string was given first) or options object
   * @return {fabric.Object} thisArg
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-2#animation}
   * @chainable
   *
   * As object — multiple properties
   *
   * object.animate({ left: ..., top: ... });
   * object.animate({ left: ..., top: ... }, { duration: ... });
   *
   * As string — one property
   *
   * object.animate('left', ...);
   * object.animate('left', { duration: ... });
   *
   */
  animate: function() {
    if (arguments[0] && typeof arguments[0] === 'object') {
      var propsToAnimate = [ ], prop, skipCallbacks;
      for (prop in arguments[0]) {
        propsToAnimate.push(prop);
      }
      for (var i = 0, len = propsToAnimate.length; i < len; i++) {
        prop = propsToAnimate[i];
        skipCallbacks = i !== len - 1;
        this._animate(prop, arguments[0][prop], arguments[1], skipCallbacks);
      }
    }
    else {
      this._animate.apply(this, arguments);
    }
    return this;
  },

  /**
   * @private
   * @param {String} property Property to animate
   * @param {String} to Value to animate to
   * @param {Object} [options] Options object
   * @param {Boolean} [skipCallbacks] When true, callbacks like onchange and oncomplete are not invoked
   */
  _animate: function(property, to, options, skipCallbacks) {
    var _this = this, propPair;

    to = to.toString();

    if (!options) {
      options = { };
    }
    else {
      options = fabric.util.object.clone(options);
    }

    if (~property.indexOf('.')) {
      propPair = property.split('.');
    }

    var currentValue = propPair
      ? this.get(propPair[0])[propPair[1]]
      : this.get(property);

    if (!('from' in options)) {
      options.from = currentValue;
    }

    if (~to.indexOf('=')) {
      to = currentValue + parseFloat(to.replace('=', ''));
    }
    else {
      to = parseFloat(to);
    }

    fabric.util.animate({
      startValue: options.from,
      endValue: to,
      byValue: options.by,
      easing: options.easing,
      duration: options.duration,
      abort: options.abort && function() {
        return options.abort.call(_this);
      },
      onChange: function(value) {
        if (propPair) {
          _this[propPair[0]][propPair[1]] = value;
        }
        else {
          _this.set(property, value);
        }
        if (skipCallbacks) {
          return;
        }
        options.onChange && options.onChange();
      },
      onComplete: function() {
        if (skipCallbacks) {
          return;
        }

        _this.setCoords();
        options.onComplete && options.onComplete();
      }
    });
  }
});


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      coordProps = { x1: 1, x2: 1, y1: 1, y2: 1 },
      supportsLineDash = fabric.StaticCanvas.supports('setLineDash');

  if (fabric.Line) {
    fabric.warn('fabric.Line is already defined');
    return;
  }

  /**
   * Line class
   * @class fabric.Line
   * @extends fabric.Object
   * @see {@link fabric.Line#initialize} for constructor definition
   */
  fabric.Line = fabric.util.createClass(fabric.Object, /** @lends fabric.Line.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'line',

    /**
     * x value or first line edge
     * @type Number
     * @default
     */
    x1: 0,

    /**
     * y value or first line edge
     * @type Number
     * @default
     */
    y1: 0,

    /**
     * x value or second line edge
     * @type Number
     * @default
     */
    x2: 0,

    /**
     * y value or second line edge
     * @type Number
     * @default
     */
    y2: 0,

    /**
     * Constructor
     * @param {Array} [points] Array of points
     * @param {Object} [options] Options object
     * @return {fabric.Line} thisArg
     */
    initialize: function(points, options) {
      options = options || { };

      if (!points) {
        points = [0, 0, 0, 0];
      }

      this.callSuper('initialize', options);

      this.set('x1', points[0]);
      this.set('y1', points[1]);
      this.set('x2', points[2]);
      this.set('y2', points[3]);

      this._setWidthHeight(options);
    },

    /**
     * @private
     * @param {Object} [options] Options
     */
    _setWidthHeight: function(options) {
      options || (options = { });

      this.width = Math.abs(this.x2 - this.x1);
      this.height = Math.abs(this.y2 - this.y1);

      this.left = 'left' in options
        ? options.left
        : this._getLeftToOriginX();

      this.top = 'top' in options
        ? options.top
        : this._getTopToOriginY();
    },

    /**
     * @private
     * @param {String} key
     * @param {Any} value
     */
    _set: function(key, value) {
      this.callSuper('_set', key, value);
      if (typeof coordProps[key] !== 'undefined') {
        this._setWidthHeight();
      }
      return this;
    },

    /**
     * @private
     * @return {Number} leftToOriginX Distance from left edge of canvas to originX of Line.
     */
    _getLeftToOriginX: makeEdgeToOriginGetter(
      { // property names
        origin: 'originX',
        axis1: 'x1',
        axis2: 'x2',
        dimension: 'width'
      },
      { // possible values of origin
        nearest: 'left',
        center: 'center',
        farthest: 'right'
      }
    ),

    /**
     * @private
     * @return {Number} topToOriginY Distance from top edge of canvas to originY of Line.
     */
    _getTopToOriginY: makeEdgeToOriginGetter(
      { // property names
        origin: 'originY',
        axis1: 'y1',
        axis2: 'y2',
        dimension: 'height'
      },
      { // possible values of origin
        nearest: 'top',
        center: 'center',
        farthest: 'bottom'
      }
    ),

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx, noTransform) {
      ctx.beginPath();

      if (noTransform) {
        //  Line coords are distances from left-top of canvas to origin of line.
        //  To render line in a path-group, we need to translate them to
        //  distances from center of path-group to center of line.
        var cp = this.getCenterPoint();
        ctx.translate(
          cp.x - this.strokeWidth / 2,
          cp.y - this.strokeWidth / 2
        );
      }

      if (!this.strokeDashArray || this.strokeDashArray && supportsLineDash) {
        // move from center (of virtual box) to its left/top corner
        // we can't assume x1, y1 is top left and x2, y2 is bottom right
        var p = this.calcLinePoints();
        ctx.moveTo(p.x1, p.y1);
        ctx.lineTo(p.x2, p.y2);
      }

      ctx.lineWidth = this.strokeWidth;

      // TODO: test this
      // make sure setting "fill" changes color of a line
      // (by copying fillStyle to strokeStyle, since line is stroked, not filled)
      var origStrokeStyle = ctx.strokeStyle;
      ctx.strokeStyle = this.stroke || ctx.fillStyle;
      this.stroke && this._renderStroke(ctx);
      ctx.strokeStyle = origStrokeStyle;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      var p = this.calcLinePoints();

      ctx.beginPath();
      fabric.util.drawDashedLine(ctx, p.x1, p.y1, p.x2, p.y2, this.strokeDashArray);
      ctx.closePath();
    },

    /**
     * Returns object representation of an instance
     * @methd toObject
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return extend(this.callSuper('toObject', propertiesToInclude), this.calcLinePoints());
    },

    /**
     * Recalculates line points given width and height
     * @private
     */
    calcLinePoints: function() {
      var xMult = this.x1 <= this.x2 ? -1 : 1,
          yMult = this.y1 <= this.y2 ? -1 : 1,
          x1 = (xMult * this.width * 0.5),
          y1 = (yMult * this.height * 0.5),
          x2 = (xMult * this.width * -0.5),
          y2 = (yMult * this.height * -0.5);

      return {
        x1: x1,
        x2: x2,
        y1: y1,
        y2: y2
      };
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(),
          p = { x1: this.x1, x2: this.x2, y1: this.y1, y2: this.y2 };

      if (!(this.group && this.group.type === 'path-group')) {
        p = this.calcLinePoints();
      }
      markup.push(
        '<line ', this.getSvgId(),
          'x1="', p.x1,
          '" y1="', p.y1,
          '" x2="', p.x2,
          '" y2="', p.y2,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(),
          this.getSvgTransformMatrix(),
        '"/>\n'
      );

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns complexity of an instance
     * @return {Number} complexity
     */
    complexity: function() {
      return 1;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Line.fromElement})
   * @static
   * @memberOf fabric.Line
   * @see http://www.w3.org/TR/SVG/shapes.html#LineElement
   */
  fabric.Line.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat('x1 y1 x2 y2'.split(' '));

  /**
   * Returns fabric.Line instance from an SVG element
   * @static
   * @memberOf fabric.Line
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Line} instance of fabric.Line
   */
  fabric.Line.fromElement = function(element, options) {
    var parsedAttributes = fabric.parseAttributes(element, fabric.Line.ATTRIBUTE_NAMES),
        points = [
          parsedAttributes.x1 || 0,
          parsedAttributes.y1 || 0,
          parsedAttributes.x2 || 0,
          parsedAttributes.y2 || 0
        ];
    return new fabric.Line(points, extend(parsedAttributes, options));
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns fabric.Line instance from an object representation
   * @static
   * @memberOf fabric.Line
   * @param {Object} object Object to create an instance from
   * @return {fabric.Line} instance of fabric.Line
   */
  fabric.Line.fromObject = function(object) {
    var points = [object.x1, object.y1, object.x2, object.y2];
    return new fabric.Line(points, object);
  };

  /**
   * Produces a function that calculates distance from canvas edge to Line origin.
   */
  function makeEdgeToOriginGetter(propertyNames, originValues) {
    var origin = propertyNames.origin,
        axis1 = propertyNames.axis1,
        axis2 = propertyNames.axis2,
        dimension = propertyNames.dimension,
        nearest = originValues.nearest,
        center = originValues.center,
        farthest = originValues.farthest;

    return function() {
      switch (this.get(origin)) {
      case nearest:
        return Math.min(this.get(axis1), this.get(axis2));
      case center:
        return Math.min(this.get(axis1), this.get(axis2)) + (0.5 * this.get(dimension));
      case farthest:
        return Math.max(this.get(axis1), this.get(axis2));
      }
    };

  }

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      pi = Math.PI,
      extend = fabric.util.object.extend;

  if (fabric.Circle) {
    fabric.warn('fabric.Circle is already defined.');
    return;
  }

  /**
   * Circle class
   * @class fabric.Circle
   * @extends fabric.Object
   * @see {@link fabric.Circle#initialize} for constructor definition
   */
  fabric.Circle = fabric.util.createClass(fabric.Object, /** @lends fabric.Circle.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'circle',

    /**
     * Radius of this circle
     * @type Number
     * @default
     */
    radius: 0,

    /**
     * Start angle of the circle, moving clockwise
     * @type Number
     * @default 0
     */
    startAngle: 0,

    /**
     * End angle of the circle
     * @type Number
     * @default 2Pi
     */
    endAngle: pi * 2,

    /**
     * Constructor
     * @param {Object} [options] Options object
     * @return {fabric.Circle} thisArg
     */
    initialize: function(options) {
      options = options || { };

      this.callSuper('initialize', options);
      this.set('radius', options.radius || 0);

      this.startAngle = options.startAngle || this.startAngle;
      this.endAngle = options.endAngle || this.endAngle;
    },

    /**
     * @private
     * @param {String} key
     * @param {Any} value
     * @return {fabric.Circle} thisArg
     */
    _set: function(key, value) {
      this.callSuper('_set', key, value);

      if (key === 'radius') {
        this.setRadius(value);
      }

      return this;
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return extend(this.callSuper('toObject', propertiesToInclude), {
        radius: this.get('radius'),
        startAngle: this.startAngle,
        endAngle: this.endAngle
      });
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(), x = 0, y = 0,
      angle = (this.endAngle - this.startAngle) % ( 2 * pi);

      if (angle === 0) {
        if (this.group && this.group.type === 'path-group') {
          x = this.left + this.radius;
          y = this.top + this.radius;
        }
        markup.push(
          '<circle ', this.getSvgId(),
            'cx="' + x + '" cy="' + y + '" ',
            'r="', this.radius,
            '" style="', this.getSvgStyles(),
            '" transform="', this.getSvgTransform(),
            ' ', this.getSvgTransformMatrix(),
          '"/>\n'
        );
      }
      else {
        var startX = Math.cos(this.startAngle) * this.radius,
            startY = Math.sin(this.startAngle) * this.radius,
            endX = Math.cos(this.endAngle) * this.radius,
            endY = Math.sin(this.endAngle) * this.radius,
            largeFlag = angle > pi ? '1' : '0';

        markup.push(
          '<path d="M ' + startX + ' ' + startY,
          ' A ' + this.radius + ' ' + this.radius,
          ' 0 ', + largeFlag + ' 1', ' ' + endX + ' ' + endY,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(),
          ' ', this.getSvgTransformMatrix(),
          '"/>\n'
        );
      }

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx context to render on
     * @param {Boolean} [noTransform] When true, context is not transformed
     */
    _render: function(ctx, noTransform) {
      ctx.beginPath();
      ctx.arc(noTransform ? this.left + this.radius : 0,
              noTransform ? this.top + this.radius : 0,
              this.radius,
              this.startAngle,
              this.endAngle, false);
      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * Returns horizontal radius of an object (according to how an object is scaled)
     * @return {Number}
     */
    getRadiusX: function() {
      return this.get('radius') * this.get('scaleX');
    },

    /**
     * Returns vertical radius of an object (according to how an object is scaled)
     * @return {Number}
     */
    getRadiusY: function() {
      return this.get('radius') * this.get('scaleY');
    },

    /**
     * Sets radius of an object (and updates width accordingly)
     * @return {fabric.Circle} thisArg
     */
    setRadius: function(value) {
      this.radius = value;
      return this.set('width', value * 2).set('height', value * 2);
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return 1;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Circle.fromElement})
   * @static
   * @memberOf fabric.Circle
   * @see: http://www.w3.org/TR/SVG/shapes.html#CircleElement
   */
  fabric.Circle.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat('cx cy r'.split(' '));

  /**
   * Returns {@link fabric.Circle} instance from an SVG element
   * @static
   * @memberOf fabric.Circle
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @throws {Error} If value of `r` attribute is missing or invalid
   * @return {fabric.Circle} Instance of fabric.Circle
   */
  fabric.Circle.fromElement = function(element, options) {
    options || (options = { });

    var parsedAttributes = fabric.parseAttributes(element, fabric.Circle.ATTRIBUTE_NAMES);

    if (!isValidRadius(parsedAttributes)) {
      throw new Error('value of `r` attribute is required and can not be negative');
    }

    parsedAttributes.left = parsedAttributes.left || 0;
    parsedAttributes.top = parsedAttributes.top || 0;

    var obj = new fabric.Circle(extend(parsedAttributes, options));

    obj.left -= obj.radius;
    obj.top -= obj.radius;
    return obj;
  };

  /**
   * @private
   */
  function isValidRadius(attributes) {
    return (('radius' in attributes) && (attributes.radius >= 0));
  }
  /* _FROM_SVG_END_ */

  /**
   * Returns {@link fabric.Circle} instance from an object representation
   * @static
   * @memberOf fabric.Circle
   * @param {Object} object Object to create an instance from
   * @return {Object} Instance of fabric.Circle
   */
  fabric.Circle.fromObject = function(object) {
    return new fabric.Circle(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { });

  if (fabric.Triangle) {
    fabric.warn('fabric.Triangle is already defined');
    return;
  }

  /**
   * Triangle class
   * @class fabric.Triangle
   * @extends fabric.Object
   * @return {fabric.Triangle} thisArg
   * @see {@link fabric.Triangle#initialize} for constructor definition
   */
  fabric.Triangle = fabric.util.createClass(fabric.Object, /** @lends fabric.Triangle.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'triangle',

    /**
     * Constructor
     * @param {Object} [options] Options object
     * @return {Object} thisArg
     */
    initialize: function(options) {
      options = options || { };

      this.callSuper('initialize', options);

      this.set('width', options.width || 100)
          .set('height', options.height || 100);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx) {
      var widthBy2 = this.width / 2,
          heightBy2 = this.height / 2;

      ctx.beginPath();
      ctx.moveTo(-widthBy2, heightBy2);
      ctx.lineTo(0, -heightBy2);
      ctx.lineTo(widthBy2, heightBy2);
      ctx.closePath();

      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      var widthBy2 = this.width / 2,
          heightBy2 = this.height / 2;

      ctx.beginPath();
      fabric.util.drawDashedLine(ctx, -widthBy2, heightBy2, 0, -heightBy2, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, 0, -heightBy2, widthBy2, heightBy2, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, widthBy2, heightBy2, -widthBy2, heightBy2, this.strokeDashArray);
      ctx.closePath();
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(),
          widthBy2 = this.width / 2,
          heightBy2 = this.height / 2,
          points = [
            -widthBy2 + ' ' + heightBy2,
            '0 ' + -heightBy2,
            widthBy2 + ' ' + heightBy2
          ]
          .join(',');

      markup.push(
        '<polygon ', this.getSvgId(),
          'points="', points,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(),
        '"/>'
      );

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return 1;
    }
  });

  /**
   * Returns fabric.Triangle instance from an object representation
   * @static
   * @memberOf fabric.Triangle
   * @param {Object} object Object to create an instance from
   * @return {Object} instance of Canvas.Triangle
   */
  fabric.Triangle.fromObject = function(object) {
    return new fabric.Triangle(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      piBy2   = Math.PI * 2,
      extend = fabric.util.object.extend;

  if (fabric.Ellipse) {
    fabric.warn('fabric.Ellipse is already defined.');
    return;
  }

  /**
   * Ellipse class
   * @class fabric.Ellipse
   * @extends fabric.Object
   * @return {fabric.Ellipse} thisArg
   * @see {@link fabric.Ellipse#initialize} for constructor definition
   */
  fabric.Ellipse = fabric.util.createClass(fabric.Object, /** @lends fabric.Ellipse.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'ellipse',

    /**
     * Horizontal radius
     * @type Number
     * @default
     */
    rx:   0,

    /**
     * Vertical radius
     * @type Number
     * @default
     */
    ry:   0,

    /**
     * Constructor
     * @param {Object} [options] Options object
     * @return {fabric.Ellipse} thisArg
     */
    initialize: function(options) {
      options = options || { };

      this.callSuper('initialize', options);

      this.set('rx', options.rx || 0);
      this.set('ry', options.ry || 0);
    },

    /**
     * @private
     * @param {String} key
     * @param {Any} value
     * @return {fabric.Ellipse} thisArg
     */
    _set: function(key, value) {
      this.callSuper('_set', key, value);
      switch (key) {

        case 'rx':
          this.rx = value;
          this.set('width', value * 2);
          break;

        case 'ry':
          this.ry = value;
          this.set('height', value * 2);
          break;

      }
      return this;
    },

    /**
     * Returns horizontal radius of an object (according to how an object is scaled)
     * @return {Number}
     */
    getRx: function() {
      return this.get('rx') * this.get('scaleX');
    },

    /**
     * Returns Vertical radius of an object (according to how an object is scaled)
     * @return {Number}
     */
    getRy: function() {
      return this.get('ry') * this.get('scaleY');
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return extend(this.callSuper('toObject', propertiesToInclude), {
        rx: this.get('rx'),
        ry: this.get('ry')
      });
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(), x = 0, y = 0;
      if (this.group && this.group.type === 'path-group') {
        x = this.left + this.rx;
        y = this.top + this.ry;
      }
      markup.push(
        '<ellipse ', this.getSvgId(),
          'cx="', x, '" cy="', y, '" ',
          'rx="', this.rx,
          '" ry="', this.ry,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(),
          this.getSvgTransformMatrix(),
        '"/>\n'
      );

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx context to render on
     * @param {Boolean} [noTransform] When true, context is not transformed
     */
    _render: function(ctx, noTransform) {
      ctx.beginPath();
      ctx.save();
      ctx.transform(1, 0, 0, this.ry/this.rx, 0, 0);
      ctx.arc(
        noTransform ? this.left + this.rx : 0,
        noTransform ? (this.top + this.ry) * this.rx/this.ry : 0,
        this.rx,
        0,
        piBy2,
        false);
      ctx.restore();
      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity
     */
    complexity: function() {
      return 1;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Ellipse.fromElement})
   * @static
   * @memberOf fabric.Ellipse
   * @see http://www.w3.org/TR/SVG/shapes.html#EllipseElement
   */
  fabric.Ellipse.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat('cx cy rx ry'.split(' '));

  /**
   * Returns {@link fabric.Ellipse} instance from an SVG element
   * @static
   * @memberOf fabric.Ellipse
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Ellipse}
   */
  fabric.Ellipse.fromElement = function(element, options) {
    options || (options = { });

    var parsedAttributes = fabric.parseAttributes(element, fabric.Ellipse.ATTRIBUTE_NAMES);

    parsedAttributes.left = parsedAttributes.left || 0;
    parsedAttributes.top = parsedAttributes.top || 0;

    var ellipse = new fabric.Ellipse(extend(parsedAttributes, options));

    ellipse.top -= ellipse.ry;
    ellipse.left -= ellipse.rx;
    return ellipse;
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns {@link fabric.Ellipse} instance from an object representation
   * @static
   * @memberOf fabric.Ellipse
   * @param {Object} object Object to create an instance from
   * @return {fabric.Ellipse}
   */
  fabric.Ellipse.fromObject = function(object) {
    return new fabric.Ellipse(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  if (fabric.Rect) {
    fabric.warn('fabric.Rect is already defined');
    return;
  }

  var stateProperties = fabric.Object.prototype.stateProperties.concat();
  stateProperties.push('rx', 'ry', 'x', 'y');

  /**
   * Rectangle class
   * @class fabric.Rect
   * @extends fabric.Object
   * @return {fabric.Rect} thisArg
   * @see {@link fabric.Rect#initialize} for constructor definition
   */
  fabric.Rect = fabric.util.createClass(fabric.Object, /** @lends fabric.Rect.prototype */ {

    /**
     * List of properties to consider when checking if state of an object is changed ({@link fabric.Object#hasStateChanged})
     * as well as for history (undo/redo) purposes
     * @type Array
     */
    stateProperties: stateProperties,

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'rect',

    /**
     * Horizontal border radius
     * @type Number
     * @default
     */
    rx:   0,

    /**
     * Vertical border radius
     * @type Number
     * @default
     */
    ry:   0,

    /**
     * Used to specify dash pattern for stroke on this object
     * @type Array
     */
    strokeDashArray: null,

    /**
     * Constructor
     * @param {Object} [options] Options object
     * @return {Object} thisArg
     */
    initialize: function(options) {
      options = options || { };

      this.callSuper('initialize', options);
      this._initRxRy();

    },

    /**
     * Initializes rx/ry attributes
     * @private
     */
    _initRxRy: function() {
      if (this.rx && !this.ry) {
        this.ry = this.rx;
      }
      else if (this.ry && !this.rx) {
        this.rx = this.ry;
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx, noTransform) {

      // optimize 1x1 case (used in spray brush)
      if (this.width === 1 && this.height === 1) {
        ctx.fillRect(-0.5, -0.5, 1, 1);
        return;
      }

      var rx = this.rx ? Math.min(this.rx, this.width / 2) : 0,
          ry = this.ry ? Math.min(this.ry, this.height / 2) : 0,
          w = this.width,
          h = this.height,
          x = noTransform ? this.left : -this.width / 2,
          y = noTransform ? this.top : -this.height / 2,
          isRounded = rx !== 0 || ry !== 0,
          k = 1 - 0.5522847498 /* "magic number" for bezier approximations of arcs (http://itc.ktu.lt/itc354/Riskus354.pdf) */;

      ctx.beginPath();

      ctx.moveTo(x + rx, y);

      ctx.lineTo(x + w - rx, y);
      isRounded && ctx.bezierCurveTo(x + w - k * rx, y, x + w, y + k * ry, x + w, y + ry);

      ctx.lineTo(x + w, y + h - ry);
      isRounded && ctx.bezierCurveTo(x + w, y + h - k * ry, x + w - k * rx, y + h, x + w - rx, y + h);

      ctx.lineTo(x + rx, y + h);
      isRounded && ctx.bezierCurveTo(x + k * rx, y + h, x, y + h - k * ry, x, y + h - ry);

      ctx.lineTo(x, y + ry);
      isRounded && ctx.bezierCurveTo(x, y + k * ry, x + k * rx, y, x + rx, y);

      ctx.closePath();

      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      var x = -this.width / 2,
          y = -this.height / 2,
          w = this.width,
          h = this.height;

      ctx.beginPath();
      fabric.util.drawDashedLine(ctx, x, y, x + w, y, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x + w, y, x + w, y + h, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x + w, y + h, x, y + h, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x, y + h, x, y, this.strokeDashArray);
      ctx.closePath();
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var object = extend(this.callSuper('toObject', propertiesToInclude), {
        rx: this.get('rx') || 0,
        ry: this.get('ry') || 0
      });
      if (!this.includeDefaultValues) {
        this._removeDefaultValues(object);
      }
      return object;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(), x = this.left, y = this.top;
      if (!(this.group && this.group.type === 'path-group')) {
        x = -this.width / 2;
        y = -this.height / 2;
      }
      markup.push(
        '<rect ', this.getSvgId(),
          'x="', x, '" y="', y,
          '" rx="', this.get('rx'), '" ry="', this.get('ry'),
          '" width="', this.width, '" height="', this.height,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(),
          this.getSvgTransformMatrix(),
        '"/>\n');

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns complexity of an instance
     * @return {Number} complexity
     */
    complexity: function() {
      return 1;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by `fabric.Rect.fromElement`)
   * @static
   * @memberOf fabric.Rect
   * @see: http://www.w3.org/TR/SVG/shapes.html#RectElement
   */
  fabric.Rect.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat('x y rx ry width height'.split(' '));

  /**
   * Returns {@link fabric.Rect} instance from an SVG element
   * @static
   * @memberOf fabric.Rect
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Rect} Instance of fabric.Rect
   */
  fabric.Rect.fromElement = function(element, options) {
    if (!element) {
      return null;
    }
    options = options || { };

    var parsedAttributes = fabric.parseAttributes(element, fabric.Rect.ATTRIBUTE_NAMES);

    parsedAttributes.left = parsedAttributes.left || 0;
    parsedAttributes.top  = parsedAttributes.top  || 0;
    var rect = new fabric.Rect(extend((options ? fabric.util.object.clone(options) : { }), parsedAttributes));
    rect.visible = rect.width > 0 && rect.height > 0;
    return rect;
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns {@link fabric.Rect} instance from an object representation
   * @static
   * @memberOf fabric.Rect
   * @param {Object} object Object to create an instance from
   * @return {Object} instance of fabric.Rect
   */
  fabric.Rect.fromObject = function(object) {
    return new fabric.Rect(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { });

  if (fabric.Polyline) {
    fabric.warn('fabric.Polyline is already defined');
    return;
  }

  /**
   * Polyline class
   * @class fabric.Polyline
   * @extends fabric.Object
   * @see {@link fabric.Polyline#initialize} for constructor definition
   */
  fabric.Polyline = fabric.util.createClass(fabric.Object, /** @lends fabric.Polyline.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'polyline',

    /**
     * Points array
     * @type Array
     * @default
     */
    points: null,

    /**
     * Minimum X from points values, necessary to offset points
     * @type Number
     * @default
     */
    minX: 0,

    /**
     * Minimum Y from points values, necessary to offset points
     * @type Number
     * @default
     */
    minY: 0,

    /**
     * Constructor
     * @param {Array} points Array of points (where each point is an object with x and y)
     * @param {Object} [options] Options object
     * @param {Boolean} [skipOffset] Whether points offsetting should be skipped
     * @return {fabric.Polyline} thisArg
     * @example
     * var poly = new fabric.Polyline([
     *     { x: 10, y: 10 },
     *     { x: 50, y: 30 },
     *     { x: 40, y: 70 },
     *     { x: 60, y: 50 },
     *     { x: 100, y: 150 },
     *     { x: 40, y: 100 }
     *   ], {
     *   stroke: 'red',
     *   left: 100,
     *   top: 100
     * });
     */
    initialize: function(points, options) {
      return fabric.Polygon.prototype.initialize.call(this, points, options);
    },

    /**
     * @private
     */
    _calcDimensions: function() {
      return fabric.Polygon.prototype._calcDimensions.call(this);
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return fabric.Polygon.prototype.toObject.call(this, propertiesToInclude);
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      return fabric.Polygon.prototype.toSVG.call(this, reviver);
    },
    /* _TO_SVG_END_ */

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx, noTransform) {
      if (!fabric.Polygon.prototype.commonRender.call(this, ctx, noTransform)) {
        return;
      }
      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      var p1, p2;

      ctx.beginPath();
      for (var i = 0, len = this.points.length; i < len; i++) {
        p1 = this.points[i];
        p2 = this.points[i + 1] || p1;
        fabric.util.drawDashedLine(ctx, p1.x, p1.y, p2.x, p2.y, this.strokeDashArray);
      }
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return this.get('points').length;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Polyline.fromElement})
   * @static
   * @memberOf fabric.Polyline
   * @see: http://www.w3.org/TR/SVG/shapes.html#PolylineElement
   */
  fabric.Polyline.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat();

  /**
   * Returns fabric.Polyline instance from an SVG element
   * @static
   * @memberOf fabric.Polyline
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Polyline} Instance of fabric.Polyline
   */
  fabric.Polyline.fromElement = function(element, options) {
    if (!element) {
      return null;
    }
    options || (options = { });

    var points = fabric.parsePointsAttribute(element.getAttribute('points')),
        parsedAttributes = fabric.parseAttributes(element, fabric.Polyline.ATTRIBUTE_NAMES);

    return new fabric.Polyline(points, fabric.util.object.extend(parsedAttributes, options));
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns fabric.Polyline instance from an object representation
   * @static
   * @memberOf fabric.Polyline
   * @param {Object} object Object to create an instance from
   * @return {fabric.Polyline} Instance of fabric.Polyline
   */
  fabric.Polyline.fromObject = function(object) {
    var points = object.points;
    return new fabric.Polyline(points, object, true);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      min = fabric.util.array.min,
      max = fabric.util.array.max,
      toFixed = fabric.util.toFixed;

  if (fabric.Polygon) {
    fabric.warn('fabric.Polygon is already defined');
    return;
  }

  /**
   * Polygon class
   * @class fabric.Polygon
   * @extends fabric.Object
   * @see {@link fabric.Polygon#initialize} for constructor definition
   */
  fabric.Polygon = fabric.util.createClass(fabric.Object, /** @lends fabric.Polygon.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'polygon',

    /**
     * Points array
     * @type Array
     * @default
     */
    points: null,

    /**
     * Minimum X from points values, necessary to offset points
     * @type Number
     * @default
     */
    minX: 0,

    /**
     * Minimum Y from points values, necessary to offset points
     * @type Number
     * @default
     */
    minY: 0,

    /**
     * Constructor
     * @param {Array} points Array of points
     * @param {Object} [options] Options object
     * @return {fabric.Polygon} thisArg
     */
    initialize: function(points, options) {
      options = options || { };
      this.points = points || [ ];
      this.callSuper('initialize', options);
      this._calcDimensions();
      if (!('top' in options)) {
        this.top = this.minY;
      }
      if (!('left' in options)) {
        this.left = this.minX;
      }
      this.pathOffset = {
        x: this.minX + this.width / 2,
        y: this.minY + this.height / 2
      };
    },

    /**
     * @private
     */
    _calcDimensions: function() {

      var points = this.points,
          minX = min(points, 'x'),
          minY = min(points, 'y'),
          maxX = max(points, 'x'),
          maxY = max(points, 'y');

      this.width = (maxX - minX) || 0;
      this.height = (maxY - minY) || 0;

      this.minX = minX || 0,
      this.minY = minY || 0;
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return extend(this.callSuper('toObject', propertiesToInclude), {
        points: this.points.concat()
      });
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var points = [], addTransform,
          markup = this._createBaseSVGMarkup();

      for (var i = 0, len = this.points.length; i < len; i++) {
        points.push(toFixed(this.points[i].x, 2), ',', toFixed(this.points[i].y, 2), ' ');
      }
      if (!(this.group && this.group.type === 'path-group')) {
        addTransform = ' translate(' + (-this.pathOffset.x) + ', ' + (-this.pathOffset.y) + ') ';
      }
      markup.push(
        '<', this.type, ' ', this.getSvgId(),
          'points="', points.join(''),
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(), addTransform,
          ' ', this.getSvgTransformMatrix(),
        '"/>\n'
      );

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx, noTransform) {
      if (!this.commonRender(ctx, noTransform)) {
        return;
      }
      this._renderFill(ctx);
      if (this.stroke || this.strokeDashArray) {
        ctx.closePath();
        this._renderStroke(ctx);
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    commonRender: function(ctx, noTransform) {
      var point, len = this.points.length;

      if (!len || isNaN(this.points[len - 1].y)) {
        // do not draw if no points or odd points
        // NaN comes from parseFloat of a empty string in parser
        return false;
      }

      noTransform || ctx.translate(-this.pathOffset.x, -this.pathOffset.y);
      ctx.beginPath();
      ctx.moveTo(this.points[0].x, this.points[0].y);
      for (var i = 0; i < len; i++) {
        point = this.points[i];
        ctx.lineTo(point.x, point.y);
      }
      return true;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      fabric.Polyline.prototype._renderDashedStroke.call(this, ctx);
      ctx.closePath();
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return this.points.length;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by `fabric.Polygon.fromElement`)
   * @static
   * @memberOf fabric.Polygon
   * @see: http://www.w3.org/TR/SVG/shapes.html#PolygonElement
   */
  fabric.Polygon.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat();

  /**
   * Returns {@link fabric.Polygon} instance from an SVG element
   * @static
   * @memberOf fabric.Polygon
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Polygon} Instance of fabric.Polygon
   */
  fabric.Polygon.fromElement = function(element, options) {
    if (!element) {
      return null;
    }

    options || (options = { });

    var points = fabric.parsePointsAttribute(element.getAttribute('points')),
        parsedAttributes = fabric.parseAttributes(element, fabric.Polygon.ATTRIBUTE_NAMES);

    return new fabric.Polygon(points, extend(parsedAttributes, options));
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns fabric.Polygon instance from an object representation
   * @static
   * @memberOf fabric.Polygon
   * @param {Object} object Object to create an instance from
   * @return {fabric.Polygon} Instance of fabric.Polygon
   */
  fabric.Polygon.fromObject = function(object) {
    return new fabric.Polygon(object.points, object, true);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      min = fabric.util.array.min,
      max = fabric.util.array.max,
      extend = fabric.util.object.extend,
      _toString = Object.prototype.toString,
      drawArc = fabric.util.drawArc,
      commandLengths = {
        m: 2,
        l: 2,
        h: 1,
        v: 1,
        c: 6,
        s: 4,
        q: 4,
        t: 2,
        a: 7
      },
      repeatedCommands = {
        m: 'l',
        M: 'L'
      };

  if (fabric.Path) {
    fabric.warn('fabric.Path is already defined');
    return;
  }

  /**
   * Path class
   * @class fabric.Path
   * @extends fabric.Object
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-1#path_and_pathgroup}
   * @see {@link fabric.Path#initialize} for constructor definition
   */
  fabric.Path = fabric.util.createClass(fabric.Object, /** @lends fabric.Path.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'path',

    /**
     * Array of path points
     * @type Array
     * @default
     */
    path: null,

    /**
     * Minimum X from points values, necessary to offset points
     * @type Number
     * @default
     */
    minX: 0,

    /**
     * Minimum Y from points values, necessary to offset points
     * @type Number
     * @default
     */
    minY: 0,

    /**
     * Constructor
     * @param {Array|String} path Path data (sequence of coordinates and corresponding "command" tokens)
     * @param {Object} [options] Options object
     * @return {fabric.Path} thisArg
     */
    initialize: function(path, options) {
      options = options || { };

      this.setOptions(options);

      if (!path) {
        path = [ ];
      }

      var fromArray = _toString.call(path) === '[object Array]';

      this.path = fromArray
        ? path
        // one of commands (m,M,l,L,q,Q,c,C,etc.) followed by non-command characters (i.e. command values)
        : path.match && path.match(/[mzlhvcsqta][^mzlhvcsqta]*/gi);

      if (!this.path) {
        return;
      }

      if (!fromArray) {
        this.path = this._parsePath();
      }

      this._setPositionDimensions(options);

      if (options.sourcePath) {
        this.setSourcePath(options.sourcePath);
      }
    },

    /**
     * @private
     * @param {Object} options Options object
     */
    _setPositionDimensions: function(options) {
      var calcDim = this._parseDimensions();

      this.minX = calcDim.left;
      this.minY = calcDim.top;
      this.width = calcDim.width;
      this.height = calcDim.height;

      if (typeof options.left === 'undefined') {
        this.left = calcDim.left + (this.originX === 'center'
          ? this.width / 2
          : this.originX === 'right'
            ? this.width
            : 0);
      }

      if (typeof options.top === 'undefined') {
        this.top = calcDim.top + (this.originY === 'center'
          ? this.height / 2
          : this.originY === 'bottom'
            ? this.height
            : 0);
      }

      this.pathOffset = this.pathOffset || {
        x: this.minX + this.width / 2,
        y: this.minY + this.height / 2
      };
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx context to render path on
     */
    _render: function(ctx) {
      var current, // current instruction
          previous = null,
          subpathStartX = 0,
          subpathStartY = 0,
          x = 0, // current x
          y = 0, // current y
          controlX = 0, // current control point x
          controlY = 0, // current control point y
          tempX,
          tempY,
          l = -this.pathOffset.x,
          t = -this.pathOffset.y;

      if (this.group && this.group.type === 'path-group') {
        l = 0;
        t = 0;
      }

      ctx.beginPath();

      for (var i = 0, len = this.path.length; i < len; ++i) {

        current = this.path[i];

        switch (current[0]) { // first letter

          case 'l': // lineto, relative
            x += current[1];
            y += current[2];
            ctx.lineTo(x + l, y + t);
            break;

          case 'L': // lineto, absolute
            x = current[1];
            y = current[2];
            ctx.lineTo(x + l, y + t);
            break;

          case 'h': // horizontal lineto, relative
            x += current[1];
            ctx.lineTo(x + l, y + t);
            break;

          case 'H': // horizontal lineto, absolute
            x = current[1];
            ctx.lineTo(x + l, y + t);
            break;

          case 'v': // vertical lineto, relative
            y += current[1];
            ctx.lineTo(x + l, y + t);
            break;

          case 'V': // verical lineto, absolute
            y = current[1];
            ctx.lineTo(x + l, y + t);
            break;

          case 'm': // moveTo, relative
            x += current[1];
            y += current[2];
            subpathStartX = x;
            subpathStartY = y;
            ctx.moveTo(x + l, y + t);
            break;

          case 'M': // moveTo, absolute
            x = current[1];
            y = current[2];
            subpathStartX = x;
            subpathStartY = y;
            ctx.moveTo(x + l, y + t);
            break;

          case 'c': // bezierCurveTo, relative
            tempX = x + current[5];
            tempY = y + current[6];
            controlX = x + current[3];
            controlY = y + current[4];
            ctx.bezierCurveTo(
              x + current[1] + l, // x1
              y + current[2] + t, // y1
              controlX + l, // x2
              controlY + t, // y2
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;
            break;

          case 'C': // bezierCurveTo, absolute
            x = current[5];
            y = current[6];
            controlX = current[3];
            controlY = current[4];
            ctx.bezierCurveTo(
              current[1] + l,
              current[2] + t,
              controlX + l,
              controlY + t,
              x + l,
              y + t
            );
            break;

          case 's': // shorthand cubic bezierCurveTo, relative

            // transform to absolute x,y
            tempX = x + current[3];
            tempY = y + current[4];

            if (previous[0].match(/[CcSs]/) === null) {
              // If there is no previous command or if the previous command was not a C, c, S, or s,
              // the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control points
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }

            ctx.bezierCurveTo(
              controlX + l,
              controlY + t,
              x + current[1] + l,
              y + current[2] + t,
              tempX + l,
              tempY + t
            );
            // set control point to 2nd one of this command
            // "... the first control point is assumed to be
            // the reflection of the second control point on
            // the previous command relative to the current point."
            controlX = x + current[1];
            controlY = y + current[2];

            x = tempX;
            y = tempY;
            break;

          case 'S': // shorthand cubic bezierCurveTo, absolute
            tempX = current[3];
            tempY = current[4];
            if (previous[0].match(/[CcSs]/) === null) {
              // If there is no previous command or if the previous command was not a C, c, S, or s,
              // the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control points
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }
            ctx.bezierCurveTo(
              controlX + l,
              controlY + t,
              current[1] + l,
              current[2] + t,
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;

            // set control point to 2nd one of this command
            // "... the first control point is assumed to be
            // the reflection of the second control point on
            // the previous command relative to the current point."
            controlX = current[1];
            controlY = current[2];

            break;

          case 'q': // quadraticCurveTo, relative
            // transform to absolute x,y
            tempX = x + current[3];
            tempY = y + current[4];

            controlX = x + current[1];
            controlY = y + current[2];

            ctx.quadraticCurveTo(
              controlX + l,
              controlY + t,
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;
            break;

          case 'Q': // quadraticCurveTo, absolute
            tempX = current[3];
            tempY = current[4];

            ctx.quadraticCurveTo(
              current[1] + l,
              current[2] + t,
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;
            controlX = current[1];
            controlY = current[2];
            break;

          case 't': // shorthand quadraticCurveTo, relative

            // transform to absolute x,y
            tempX = x + current[1];
            tempY = y + current[2];

            if (previous[0].match(/[QqTt]/) === null) {
              // If there is no previous command or if the previous command was not a Q, q, T or t,
              // assume the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control point
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }

            ctx.quadraticCurveTo(
              controlX + l,
              controlY + t,
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;

            break;

          case 'T':
            tempX = current[1];
            tempY = current[2];

            if (previous[0].match(/[QqTt]/) === null) {
              // If there is no previous command or if the previous command was not a Q, q, T or t,
              // assume the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control point
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }
            ctx.quadraticCurveTo(
              controlX + l,
              controlY + t,
              tempX + l,
              tempY + t
            );
            x = tempX;
            y = tempY;
            break;

          case 'a':
            // TODO: optimize this
            drawArc(ctx, x + l, y + t, [
              current[1],
              current[2],
              current[3],
              current[4],
              current[5],
              current[6] + x + l,
              current[7] + y + t
            ]);
            x += current[6];
            y += current[7];
            break;

          case 'A':
            // TODO: optimize this
            drawArc(ctx, x + l, y + t, [
              current[1],
              current[2],
              current[3],
              current[4],
              current[5],
              current[6] + l,
              current[7] + t
            ]);
            x = current[6];
            y = current[7];
            break;

          case 'z':
          case 'Z':
            x = subpathStartX;
            y = subpathStartY;
            ctx.closePath();
            break;
        }
        previous = current;
      }
      this._renderFill(ctx);
      this._renderStroke(ctx);
    },

    /**
     * Returns string representation of an instance
     * @return {String} string representation of an instance
     */
    toString: function() {
      return '#<fabric.Path (' + this.complexity() +
        '): { "top": ' + this.top + ', "left": ' + this.left + ' }>';
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var o = extend(this.callSuper('toObject', propertiesToInclude), {
        path: this.path.map(function(item) { return item.slice() }),
        pathOffset: this.pathOffset
      });
      if (this.sourcePath) {
        o.sourcePath = this.sourcePath;
      }
      if (this.transformMatrix) {
        o.transformMatrix = this.transformMatrix;
      }
      return o;
    },

    /**
     * Returns dataless object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toDatalessObject: function(propertiesToInclude) {
      var o = this.toObject(propertiesToInclude);
      if (this.sourcePath) {
        o.path = this.sourcePath;
      }
      delete o.sourcePath;
      return o;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var chunks = [],
          markup = this._createBaseSVGMarkup(), addTransform = '';

      for (var i = 0, len = this.path.length; i < len; i++) {
        chunks.push(this.path[i].join(' '));
      }
      var path = chunks.join(' ');
      if (!(this.group && this.group.type === 'path-group')) {
        addTransform = ' translate(' + (-this.pathOffset.x) + ', ' + (-this.pathOffset.y) + ') ';
      }
      markup.push(
        //jscs:disable validateIndentation
        '<path ', this.getSvgId(),
          'd="', path,
          '" style="', this.getSvgStyles(),
          '" transform="', this.getSvgTransform(), addTransform,
          this.getSvgTransformMatrix(), '" stroke-linecap="round" ',
        '/>\n'
        //jscs:enable validateIndentation
      );

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns number representation of an instance complexity
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return this.path.length;
    },

    /**
     * @private
     */
    _parsePath: function() {
      var result = [ ],
          coords = [ ],
          currentPath,
          parsed,
          re = /([-+]?((\d+\.\d+)|((\d+)|(\.\d+)))(?:e[-+]?\d+)?)/ig,
          match,
          coordsStr;

      for (var i = 0, coordsParsed, len = this.path.length; i < len; i++) {
        currentPath = this.path[i];

        coordsStr = currentPath.slice(1).trim();
        coords.length = 0;

        while ((match = re.exec(coordsStr))) {
          coords.push(match[0]);
        }

        coordsParsed = [ currentPath.charAt(0) ];

        for (var j = 0, jlen = coords.length; j < jlen; j++) {
          parsed = parseFloat(coords[j]);
          if (!isNaN(parsed)) {
            coordsParsed.push(parsed);
          }
        }

        var command = coordsParsed[0],
            commandLength = commandLengths[command.toLowerCase()],
            repeatedCommand = repeatedCommands[command] || command;

        if (coordsParsed.length - 1 > commandLength) {
          for (var k = 1, klen = coordsParsed.length; k < klen; k += commandLength) {
            result.push([ command ].concat(coordsParsed.slice(k, k + commandLength)));
            command = repeatedCommand;
          }
        }
        else {
          result.push(coordsParsed);
        }
      }

      return result;
    },

    /**
     * @private
     */
    _parseDimensions: function() {

      var aX = [],
          aY = [],
          current, // current instruction
          previous = null,
          subpathStartX = 0,
          subpathStartY = 0,
          x = 0, // current x
          y = 0, // current y
          controlX = 0, // current control point x
          controlY = 0, // current control point y
          tempX,
          tempY,
          bounds;

      for (var i = 0, len = this.path.length; i < len; ++i) {

        current = this.path[i];

        switch (current[0]) { // first letter

          case 'l': // lineto, relative
            x += current[1];
            y += current[2];
            bounds = [ ];
            break;

          case 'L': // lineto, absolute
            x = current[1];
            y = current[2];
            bounds = [ ];
            break;

          case 'h': // horizontal lineto, relative
            x += current[1];
            bounds = [ ];
            break;

          case 'H': // horizontal lineto, absolute
            x = current[1];
            bounds = [ ];
            break;

          case 'v': // vertical lineto, relative
            y += current[1];
            bounds = [ ];
            break;

          case 'V': // verical lineto, absolute
            y = current[1];
            bounds = [ ];
            break;

          case 'm': // moveTo, relative
            x += current[1];
            y += current[2];
            subpathStartX = x;
            subpathStartY = y;
            bounds = [ ];
            break;

          case 'M': // moveTo, absolute
            x = current[1];
            y = current[2];
            subpathStartX = x;
            subpathStartY = y;
            bounds = [ ];
            break;

          case 'c': // bezierCurveTo, relative
            tempX = x + current[5];
            tempY = y + current[6];
            controlX = x + current[3];
            controlY = y + current[4];
            bounds = fabric.util.getBoundsOfCurve(x, y,
              x + current[1], // x1
              y + current[2], // y1
              controlX, // x2
              controlY, // y2
              tempX,
              tempY
            );
            x = tempX;
            y = tempY;
            break;

          case 'C': // bezierCurveTo, absolute
            x = current[5];
            y = current[6];
            controlX = current[3];
            controlY = current[4];
            bounds = fabric.util.getBoundsOfCurve(x, y,
              current[1],
              current[2],
              controlX,
              controlY,
              x,
              y
            );
            break;

          case 's': // shorthand cubic bezierCurveTo, relative

            // transform to absolute x,y
            tempX = x + current[3];
            tempY = y + current[4];

            if (previous[0].match(/[CcSs]/) === null) {
              // If there is no previous command or if the previous command was not a C, c, S, or s,
              // the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control points
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }

            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              x + current[1],
              y + current[2],
              tempX,
              tempY
            );
            // set control point to 2nd one of this command
            // "... the first control point is assumed to be
            // the reflection of the second control point on
            // the previous command relative to the current point."
            controlX = x + current[1];
            controlY = y + current[2];
            x = tempX;
            y = tempY;
            break;

          case 'S': // shorthand cubic bezierCurveTo, absolute
            tempX = current[3];
            tempY = current[4];
            if (previous[0].match(/[CcSs]/) === null) {
              // If there is no previous command or if the previous command was not a C, c, S, or s,
              // the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control points
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }
            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              current[1],
              current[2],
              tempX,
              tempY
            );
            x = tempX;
            y = tempY;
            // set control point to 2nd one of this command
            // "... the first control point is assumed to be
            // the reflection of the second control point on
            // the previous command relative to the current point."
            controlX = current[1];
            controlY = current[2];
            break;

          case 'q': // quadraticCurveTo, relative
            // transform to absolute x,y
            tempX = x + current[3];
            tempY = y + current[4];
            controlX = x + current[1];
            controlY = y + current[2];
            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              controlX,
              controlY,
              tempX,
              tempY
            );
            x = tempX;
            y = tempY;
            break;

          case 'Q': // quadraticCurveTo, absolute
            controlX = current[1];
            controlY = current[2];
            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              controlX,
              controlY,
              current[3],
              current[4]
            );
            x = current[3];
            y = current[4];
            break;

          case 't': // shorthand quadraticCurveTo, relative
            // transform to absolute x,y
            tempX = x + current[1];
            tempY = y + current[2];
            if (previous[0].match(/[QqTt]/) === null) {
              // If there is no previous command or if the previous command was not a Q, q, T or t,
              // assume the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control point
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }

            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              controlX,
              controlY,
              tempX,
              tempY
            );
            x = tempX;
            y = tempY;

            break;

          case 'T':
            tempX = current[1];
            tempY = current[2];

            if (previous[0].match(/[QqTt]/) === null) {
              // If there is no previous command or if the previous command was not a Q, q, T or t,
              // assume the control point is coincident with the current point
              controlX = x;
              controlY = y;
            }
            else {
              // calculate reflection of previous control point
              controlX = 2 * x - controlX;
              controlY = 2 * y - controlY;
            }
            bounds = fabric.util.getBoundsOfCurve(x, y,
              controlX,
              controlY,
              controlX,
              controlY,
              tempX,
              tempY
            );
            x = tempX;
            y = tempY;
            break;

          case 'a':
            // TODO: optimize this
            bounds = fabric.util.getBoundsOfArc(x, y,
              current[1],
              current[2],
              current[3],
              current[4],
              current[5],
              current[6] + x,
              current[7] + y
            );
            x += current[6];
            y += current[7];
            break;

          case 'A':
            // TODO: optimize this
            bounds = fabric.util.getBoundsOfArc(x, y,
              current[1],
              current[2],
              current[3],
              current[4],
              current[5],
              current[6],
              current[7]
            );
            x = current[6];
            y = current[7];
            break;

          case 'z':
          case 'Z':
            x = subpathStartX;
            y = subpathStartY;
            break;
        }
        previous = current;
        bounds.forEach(function (point) {
          aX.push(point.x);
          aY.push(point.y);
        });
        aX.push(x);
        aY.push(y);
      }

      var minX = min(aX) || 0,
          minY = min(aY) || 0,
          maxX = max(aX) || 0,
          maxY = max(aY) || 0,
          deltaX = maxX - minX,
          deltaY = maxY - minY,

          o = {
            left: minX,
            top: minY,
            width: deltaX,
            height: deltaY
          };

      return o;
    }
  });

  /**
   * Creates an instance of fabric.Path from an object
   * @static
   * @memberOf fabric.Path
   * @param {Object} object
   * @param {Function} callback Callback to invoke when an fabric.Path instance is created
   */
  fabric.Path.fromObject = function(object, callback) {
    if (typeof object.path === 'string') {
      fabric.loadSVGFromURL(object.path, function (elements) {
        var path = elements[0],
            pathUrl = object.path;

        delete object.path;

        fabric.util.object.extend(path, object);
        path.setSourcePath(pathUrl);

        callback(path);
      });
    }
    else {
      callback(new fabric.Path(object.path, object));
    }
  };

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by `fabric.Path.fromElement`)
   * @static
   * @memberOf fabric.Path
   * @see http://www.w3.org/TR/SVG/paths.html#PathElement
   */
  fabric.Path.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat(['d']);

  /**
   * Creates an instance of fabric.Path from an SVG <path> element
   * @static
   * @memberOf fabric.Path
   * @param {SVGElement} element to parse
   * @param {Function} callback Callback to invoke when an fabric.Path instance is created
   * @param {Object} [options] Options object
   */
  fabric.Path.fromElement = function(element, callback, options) {
    var parsedAttributes = fabric.parseAttributes(element, fabric.Path.ATTRIBUTE_NAMES);
    callback && callback(new fabric.Path(parsedAttributes.d, extend(parsedAttributes, options)));
  };
  /* _FROM_SVG_END_ */

  /**
   * Indicates that instances of this type are async
   * @static
   * @memberOf fabric.Path
   * @type Boolean
   * @default
   */
  fabric.Path.async = true;

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      invoke = fabric.util.array.invoke,
      parentToObject = fabric.Object.prototype.toObject;

  if (fabric.PathGroup) {
    fabric.warn('fabric.PathGroup is already defined');
    return;
  }

  /**
   * Path group class
   * @class fabric.PathGroup
   * @extends fabric.Path
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-1#path_and_pathgroup}
   * @see {@link fabric.PathGroup#initialize} for constructor definition
   */
  fabric.PathGroup = fabric.util.createClass(fabric.Path, /** @lends fabric.PathGroup.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'path-group',

    /**
     * Fill value
     * @type String
     * @default
     */
    fill: '',

    /**
     * Constructor
     * @param {Array} paths
     * @param {Object} [options] Options object
     * @return {fabric.PathGroup} thisArg
     */
    initialize: function(paths, options) {

      options = options || { };
      this.paths = paths || [ ];

      for (var i = this.paths.length; i--;) {
        this.paths[i].group = this;
      }

      if (options.toBeParsed) {
        this.parseDimensionsFromPaths(options);
        delete options.toBeParsed;
      }
      this.setOptions(options);
      this.setCoords();

      if (options.sourcePath) {
        this.setSourcePath(options.sourcePath);
      }
    },

    /**
     * Calculate width and height based on paths contained
     */
    parseDimensionsFromPaths: function(options) {
      var points, p, xC = [ ], yC = [ ], path, height, width,
          m;
      for (var j = this.paths.length; j--;) {
        path = this.paths[j];
        height = path.height + path.strokeWidth;
        width = path.width + path.strokeWidth;
        points = [
          { x: path.left, y: path.top },
          { x: path.left + width, y: path.top },
          { x: path.left, y: path.top + height },
          { x: path.left + width, y: path.top + height }
        ];
        m = this.paths[j].transformMatrix;
        for (var i = 0; i < points.length; i++) {
          p = points[i];
          if (m) {
            p = fabric.util.transformPoint(p, m, false);
          }
          xC.push(p.x);
          yC.push(p.y);
        }
      }
      options.width = Math.max.apply(null, xC);
      options.height = Math.max.apply(null, yC);
    },

    /**
     * Renders this group on a specified context
     * @param {CanvasRenderingContext2D} ctx Context to render this instance on
     */
    render: function(ctx) {
      // do not render if object is not visible
      if (!this.visible) {
        return;
      }

      ctx.save();

      if (this.transformMatrix) {
        ctx.transform.apply(ctx, this.transformMatrix);
      }
      this.transform(ctx);

      this._setShadow(ctx);
      this.clipTo && fabric.util.clipContext(this, ctx);
      ctx.translate(-this.width/2, -this.height/2);
      for (var i = 0, l = this.paths.length; i < l; ++i) {
        this.paths[i].render(ctx, true);
      }
      this.clipTo && ctx.restore();
      ctx.restore();
    },

    /**
     * Sets certain property to a certain value
     * @param {String} prop
     * @param {Any} value
     * @return {fabric.PathGroup} thisArg
     */
    _set: function(prop, value) {

      if (prop === 'fill' && value && this.isSameColor()) {
        var i = this.paths.length;
        while (i--) {
          this.paths[i]._set(prop, value);
        }
      }

      return this.callSuper('_set', prop, value);
    },

    /**
     * Returns object representation of this path group
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var o = extend(parentToObject.call(this, propertiesToInclude), {
        paths: invoke(this.getObjects(), 'toObject', propertiesToInclude)
      });
      if (this.sourcePath) {
        o.sourcePath = this.sourcePath;
      }
      return o;
    },

    /**
     * Returns dataless object representation of this path group
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} dataless object representation of an instance
     */
    toDatalessObject: function(propertiesToInclude) {
      var o = this.toObject(propertiesToInclude);
      if (this.sourcePath) {
        o.paths = this.sourcePath;
      }
      return o;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var objects = this.getObjects(),
          p = this.getPointByOrigin('left', 'top'),
          translatePart = 'translate(' + p.x + ' ' + p.y + ')',
          markup = this._createBaseSVGMarkup();
      markup.push(
        '<g ', this.getSvgId(),
        'style="', this.getSvgStyles(), '" ',
        'transform="', this.getSvgTransformMatrix(), translatePart, this.getSvgTransform(), '" ',
        '>\n'
      );

      for (var i = 0, len = objects.length; i < len; i++) {
        markup.push('\t', objects[i].toSVG(reviver));
      }
      markup.push('</g>\n');

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns a string representation of this path group
     * @return {String} string representation of an object
     */
    toString: function() {
      return '#<fabric.PathGroup (' + this.complexity() +
        '): { top: ' + this.top + ', left: ' + this.left + ' }>';
    },

    /**
     * Returns true if all paths in this group are of same color
     * @return {Boolean} true if all paths are of the same color (`fill`)
     */
    isSameColor: function() {
      var firstPathFill = this.getObjects()[0].get('fill') || '';
      if (typeof firstPathFill !== 'string') {
        return false;
      }
      firstPathFill = firstPathFill.toLowerCase();
      return this.getObjects().every(function(path) {
        var pathFill = path.get('fill') || '';
        return typeof pathFill === 'string' && (pathFill).toLowerCase() === firstPathFill;
      });
    },

    /**
     * Returns number representation of object's complexity
     * @return {Number} complexity
     */
    complexity: function() {
      return this.paths.reduce(function(total, path) {
        return total + ((path && path.complexity) ? path.complexity() : 0);
      }, 0);
    },

    /**
     * Returns all paths in this path group
     * @return {Array} array of path objects included in this path group
     */
    getObjects: function() {
      return this.paths;
    }
  });

  /**
   * Creates fabric.PathGroup instance from an object representation
   * @static
   * @memberOf fabric.PathGroup
   * @param {Object} object Object to create an instance from
   * @param {Function} callback Callback to invoke when an fabric.PathGroup instance is created
   */
  fabric.PathGroup.fromObject = function(object, callback) {
    if (typeof object.paths === 'string') {
      fabric.loadSVGFromURL(object.paths, function (elements) {

        var pathUrl = object.paths;
        delete object.paths;

        var pathGroup = fabric.util.groupSVGElements(elements, object, pathUrl);

        callback(pathGroup);
      });
    }
    else {
      fabric.util.enlivenObjects(object.paths, function(enlivenedObjects) {
        delete object.paths;
        callback(new fabric.PathGroup(enlivenedObjects, object));
      });
    }
  };

  /**
   * Indicates that instances of this type are async
   * @static
   * @memberOf fabric.PathGroup
   * @type Boolean
   * @default
   */
  fabric.PathGroup.async = true;

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      min = fabric.util.array.min,
      max = fabric.util.array.max,
      invoke = fabric.util.array.invoke;

  if (fabric.Group) {
    return;
  }

  // lock-related properties, for use in fabric.Group#get
  // to enable locking behavior on group
  // when one of its objects has lock-related properties set
  var _lockProperties = {
    lockMovementX:  true,
    lockMovementY:  true,
    lockRotation:   true,
    lockScalingX:   true,
    lockScalingY:   true,
    lockUniScaling: true
  };

  /**
   * Group class
   * @class fabric.Group
   * @extends fabric.Object
   * @mixes fabric.Collection
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-3#groups}
   * @see {@link fabric.Group#initialize} for constructor definition
   */
  fabric.Group = fabric.util.createClass(fabric.Object, fabric.Collection, /** @lends fabric.Group.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'group',

    /**
     * Width of stroke
     * @type Number
     * @default
     */
    strokeWidth: 0,

    /**
     * Indicates if click events should also check for subtargets
     * @type Boolean
     * @default
     */
    subTargetCheck: false,

    /**
     * Constructor
     * @param {Object} objects Group objects
     * @param {Object} [options] Options object
     * @param {Boolean} [isAlreadyGrouped] if true, objects have been grouped already.
     * @return {Object} thisArg
     */
    initialize: function(objects, options, isAlreadyGrouped) {
      options = options || { };

      this._objects = [];
      // if objects enclosed in a group have been grouped already,
      // we cannot change properties of objects.
      // Thus we need to set options to group without objects,
      // because delegatedProperties propagate to objects.
      isAlreadyGrouped && this.callSuper('initialize', options);

      this._objects = objects || [];
      for (var i = this._objects.length; i--; ) {
        this._objects[i].group = this;
      }

      this.originalState = { };

      if (options.originX) {
        this.originX = options.originX;
      }
      if (options.originY) {
        this.originY = options.originY;
      }

      if (isAlreadyGrouped) {
        // do not change coordinate of objects enclosed in a group,
        // because objects coordinate system have been group coodinate system already.
        this._updateObjectsCoords(true);
      }
      else {
        this._calcBounds();
        this._updateObjectsCoords();
        this.callSuper('initialize', options);
      }

      this.setCoords();
      this.saveCoords();
    },

    /**
     * @private
     * @param {Boolean} [skipCoordsChange] if true, coordinates of objects enclosed in a group do not change
     */
    _updateObjectsCoords: function(skipCoordsChange) {
      for (var i = this._objects.length; i--; ){
        this._updateObjectCoords(this._objects[i], skipCoordsChange);
      }
    },

    /**
     * @private
     * @param {Object} object
     * @param {Boolean} [skipCoordsChange] if true, coordinates of object dose not change
     */
    _updateObjectCoords: function(object, skipCoordsChange) {
      // do not display corners of objects enclosed in a group
      object.__origHasControls = object.hasControls;
      object.hasControls = false;

      if (skipCoordsChange) {
        return;
      }

      var objectLeft = object.getLeft(),
          objectTop = object.getTop(),
          center = this.getCenterPoint();

      object.set({
        originalLeft: objectLeft,
        originalTop: objectTop,
        left: objectLeft - center.x,
        top: objectTop - center.y
      });
      object.setCoords();
    },

    /**
     * Returns string represenation of a group
     * @return {String}
     */
    toString: function() {
      return '#<fabric.Group: (' + this.complexity() + ')>';
    },

    /**
     * Adds an object to a group; Then recalculates group's dimension, position.
     * @param {Object} object
     * @return {fabric.Group} thisArg
     * @chainable
     */
    addWithUpdate: function(object) {
      this._restoreObjectsState();
      fabric.util.resetObjectTransform(this);
      if (object) {
        this._objects.push(object);
        object.group = this;
        object._set('canvas', this.canvas);
      }
      // since _restoreObjectsState set objects inactive
      this.forEachObject(this._setObjectActive, this);
      this._calcBounds();
      this._updateObjectsCoords();
      return this;
    },

    /**
     * @private
     */
    _setObjectActive: function(object) {
      object.set('active', true);
      object.group = this;
    },

    /**
     * Removes an object from a group; Then recalculates group's dimension, position.
     * @param {Object} object
     * @return {fabric.Group} thisArg
     * @chainable
     */
    removeWithUpdate: function(object) {
      this._restoreObjectsState();
      fabric.util.resetObjectTransform(this);
      // since _restoreObjectsState set objects inactive
      this.forEachObject(this._setObjectActive, this);

      this.remove(object);
      this._calcBounds();
      this._updateObjectsCoords();

      return this;
    },

    /**
     * @private
     */
    _onObjectAdded: function(object) {
      object.group = this;
      object._set('canvas', this.canvas);
    },

    /**
     * @private
     */
    _onObjectRemoved: function(object) {
      delete object.group;
      object.set('active', false);
    },

    /**
     * Properties that are delegated to group objects when reading/writing
     * @param {Object} delegatedProperties
     */
    delegatedProperties: {
      fill:             true,
      stroke:           true,
      strokeWidth:      true,
      fontFamily:       true,
      fontWeight:       true,
      fontSize:         true,
      fontStyle:        true,
      lineHeight:       true,
      textDecoration:   true,
      textAlign:        true,
      backgroundColor:  true
    },

    /**
     * @private
     */
    _set: function(key, value) {
      var i = this._objects.length;

      if (this.delegatedProperties[key] || key === 'canvas') {
        while (i--) {
          this._objects[i].set(key, value);
        }
      }
      else {
        while (i--) {
          this._objects[i].setOnGroup(key, value);
        }
      }

      this.callSuper('_set', key, value);
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return extend(this.callSuper('toObject', propertiesToInclude), {
        objects: invoke(this._objects, 'toObject', propertiesToInclude)
      });
    },

    /**
     * Renders instance on a given context
     * @param {CanvasRenderingContext2D} ctx context to render instance on
     */
    render: function(ctx) {
      // do not render if object is not visible
      if (!this.visible) {
        return;
      }

      ctx.save();
      if (this.transformMatrix) {
        ctx.transform.apply(ctx, this.transformMatrix);
      }
      this.transform(ctx);
      this._setShadow(ctx);
      this.clipTo && fabric.util.clipContext(this, ctx);
      // the array is now sorted in order of highest first, so start from end
      for (var i = 0, len = this._objects.length; i < len; i++) {
        this._renderObject(this._objects[i], ctx);
      }

      this.clipTo && ctx.restore();

      ctx.restore();
    },

    /**
     * Renders controls and borders for the object
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Boolean} [noTransform] When true, context is not transformed
     */
    _renderControls: function(ctx, noTransform) {
      this.callSuper('_renderControls', ctx, noTransform);
      for (var i = 0, len = this._objects.length; i < len; i++) {
        this._objects[i]._renderControls(ctx);
      }
    },

    /**
     * @private
     */
    _renderObject: function(object, ctx) {
      // do not render if object is not visible
      if (!object.visible) {
        return;
      }

      var originalHasRotatingPoint = object.hasRotatingPoint;
      object.hasRotatingPoint = false;
      object.render(ctx);
      object.hasRotatingPoint = originalHasRotatingPoint;
    },

    /**
     * Retores original state of each of group objects (original state is that which was before group was created).
     * @private
     * @return {fabric.Group} thisArg
     * @chainable
     */
    _restoreObjectsState: function() {
      this._objects.forEach(this._restoreObjectState, this);
      return this;
    },

    /**
     * Realises the transform from this group onto the supplied object
     * i.e. it tells you what would happen if the supplied object was in
     * the group, and then the group was destroyed. It mutates the supplied
     * object.
     * @param {fabric.Object} object
     * @return {fabric.Object} transformedObject
     */
    realizeTransform: function(object) {
      var matrix = object.calcTransformMatrix(),
          options = fabric.util.qrDecompose(matrix),
          center = new fabric.Point(options.translateX, options.translateY);
      object.scaleX = options.scaleX;
      object.scaleY = options.scaleY;
      object.skewX = options.skewX;
      object.skewY = options.skewY;
      object.angle = options.angle;
      object.flipX = false;
      object.flipY = false;
      object.setPositionByOrigin(center, 'center', 'center');
      return object;
    },

    /**
     * Restores original state of a specified object in group
     * @private
     * @param {fabric.Object} object
     * @return {fabric.Group} thisArg
     */
    _restoreObjectState: function(object) {
      this.realizeTransform(object);
      object.setCoords();
      object.hasControls = object.__origHasControls;
      delete object.__origHasControls;
      object.set('active', false);
      delete object.group;

      return this;
    },

    /**
     * Destroys a group (restoring state of its objects)
     * @return {fabric.Group} thisArg
     * @chainable
     */
    destroy: function() {
      return this._restoreObjectsState();
    },

    /**
     * Saves coordinates of this instance (to be used together with `hasMoved`)
     * @saveCoords
     * @return {fabric.Group} thisArg
     * @chainable
     */
    saveCoords: function() {
      this._originalLeft = this.get('left');
      this._originalTop = this.get('top');
      return this;
    },

    /**
     * Checks whether this group was moved (since `saveCoords` was called last)
     * @return {Boolean} true if an object was moved (since fabric.Group#saveCoords was called)
     */
    hasMoved: function() {
      return this._originalLeft !== this.get('left') ||
             this._originalTop !== this.get('top');
    },

    /**
     * Sets coordinates of all group objects
     * @return {fabric.Group} thisArg
     * @chainable
     */
    setObjectsCoords: function() {
      this.forEachObject(function(object) {
        object.setCoords();
      });
      return this;
    },

    /**
     * @private
     */
    _calcBounds: function(onlyWidthHeight) {
      var aX = [],
          aY = [],
          o, prop,
          props = ['tr', 'br', 'bl', 'tl'],
          i = 0, iLen = this._objects.length,
          j, jLen = props.length;

      for ( ; i < iLen; ++i) {
        o = this._objects[i];
        o.setCoords();
        for (j = 0; j < jLen; j++) {
          prop = props[j];
          aX.push(o.oCoords[prop].x);
          aY.push(o.oCoords[prop].y);
        }
      }

      this.set(this._getBounds(aX, aY, onlyWidthHeight));
    },

    /**
     * @private
     */
    _getBounds: function(aX, aY, onlyWidthHeight) {
      var ivt = fabric.util.invertTransform(this.getViewportTransform()),
          minXY = fabric.util.transformPoint(new fabric.Point(min(aX), min(aY)), ivt),
          maxXY = fabric.util.transformPoint(new fabric.Point(max(aX), max(aY)), ivt),
          obj = {
            width: (maxXY.x - minXY.x) || 0,
            height: (maxXY.y - minXY.y) || 0
          };

      if (!onlyWidthHeight) {
        obj.left = minXY.x || 0;
        obj.top = minXY.y || 0;
        if (this.originX === 'center') {
          obj.left += obj.width / 2;
        }
        if (this.originX === 'right') {
          obj.left += obj.width;
        }
        if (this.originY === 'center') {
          obj.top += obj.height / 2;
        }
        if (this.originY === 'bottom') {
          obj.top += obj.height;
        }
      }
      return obj;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns svg representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup();
      markup.push(
        '<g ', this.getSvgId(), 'transform="',
        /* avoiding styles intentionally */
        this.getSvgTransform(),
        this.getSvgTransformMatrix(),
        '" style="',
        this.getSvgFilter(),
        '">\n'
      );

      for (var i = 0, len = this._objects.length; i < len; i++) {
        markup.push('\t', this._objects[i].toSVG(reviver));
      }

      markup.push('</g>\n');

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns requested property
     * @param {String} prop Property to get
     * @return {Any}
     */
    get: function(prop) {
      if (prop in _lockProperties) {
        if (this[prop]) {
          return this[prop];
        }
        else {
          for (var i = 0, len = this._objects.length; i < len; i++) {
            if (this._objects[i][prop]) {
              return true;
            }
          }
          return false;
        }
      }
      else {
        if (prop in this.delegatedProperties) {
          return this._objects[0] && this._objects[0].get(prop);
        }
        return this[prop];
      }
    }
  });

  /**
   * Returns {@link fabric.Group} instance from an object representation
   * @static
   * @memberOf fabric.Group
   * @param {Object} object Object to create a group from
   * @param {Function} [callback] Callback to invoke when an group instance is created
   * @return {fabric.Group} An instance of fabric.Group
   */
  fabric.Group.fromObject = function(object, callback) {
    fabric.util.enlivenObjects(object.objects, function(enlivenedObjects) {
      delete object.objects;
      callback && callback(new fabric.Group(enlivenedObjects, object, true));
    });
  };

  /**
   * Indicates that instances of this type are async
   * @static
   * @memberOf fabric.Group
   * @type Boolean
   * @default
   */
  fabric.Group.async = true;

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var extend = fabric.util.object.extend;

  if (!global.fabric) {
    global.fabric = { };
  }

  if (global.fabric.Image) {
    fabric.warn('fabric.Image is already defined.');
    return;
  }

  /**
   * Image class
   * @class fabric.Image
   * @extends fabric.Object
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-1#images}
   * @see {@link fabric.Image#initialize} for constructor definition
   */
  fabric.Image = fabric.util.createClass(fabric.Object, /** @lends fabric.Image.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'image',

    /**
     * crossOrigin value (one of "", "anonymous", "use-credentials")
     * @see https://developer.mozilla.org/en-US/docs/HTML/CORS_settings_attributes
     * @type String
     * @default
     */
    crossOrigin: '',

    /**
     * AlignX value, part of preserveAspectRatio (one of "none", "mid", "min", "max")
     * @see http://www.w3.org/TR/SVG/coords.html#PreserveAspectRatioAttribute
     * This parameter defines how the picture is aligned to its viewport when image element width differs from image width.
     * @type String
     * @default
     */
    alignX: 'none',

    /**
     * AlignY value, part of preserveAspectRatio (one of "none", "mid", "min", "max")
     * @see http://www.w3.org/TR/SVG/coords.html#PreserveAspectRatioAttribute
     * This parameter defines how the picture is aligned to its viewport when image element height differs from image height.
     * @type String
     * @default
     */
    alignY: 'none',

    /**
     * meetOrSlice value, part of preserveAspectRatio  (one of "meet", "slice").
     * if meet the image is always fully visibile, if slice the viewport is always filled with image.
     * @see http://www.w3.org/TR/SVG/coords.html#PreserveAspectRatioAttribute
     * @type String
     * @default
     */
    meetOrSlice: 'meet',

    /**
     * Width of a stroke.
     * For image quality a stroke multiple of 2 gives better results.
     * @type Number
     * @default
     */
    strokeWidth: 0,

    /**
     * private
     * contains last value of scaleX to detect
     * if the Image got resized after the last Render
     * @type Number
     */
    _lastScaleX: 1,

    /**
     * private
     * contains last value of scaleY to detect
     * if the Image got resized after the last Render
     * @type Number
     */
    _lastScaleY: 1,

    /**
     * Constructor
     * @param {HTMLImageElement | String} element Image element
     * @param {Object} [options] Options object
     * @return {fabric.Image} thisArg
     */
    initialize: function(element, options) {
      options || (options = { });
      this.filters = [ ];
      this.resizeFilters = [ ];
      this.callSuper('initialize', options);
      this._initElement(element, options);
    },

    /**
     * Returns image element which this instance if based on
     * @return {HTMLImageElement} Image element
     */
    getElement: function() {
      return this._element;
    },

    /**
     * Sets image element for this instance to a specified one.
     * If filters defined they are applied to new image.
     * You might need to call `canvas.renderAll` and `object.setCoords` after replacing, to render new image and update controls area.
     * @param {HTMLImageElement} element
     * @param {Function} [callback] Callback is invoked when all filters have been applied and new image is generated
     * @param {Object} [options] Options object
     * @return {fabric.Image} thisArg
     * @chainable
     */
    setElement: function(element, callback, options) {
      this._element = element;
      this._originalElement = element;
      this._initConfig(options);

      if (this.filters.length !== 0) {
        this.applyFilters(callback);
      }
      else if (callback) {
        callback();
      }

      return this;
    },

    /**
     * Sets crossOrigin value (on an instance and corresponding image element)
     * @return {fabric.Image} thisArg
     * @chainable
     */
    setCrossOrigin: function(value) {
      this.crossOrigin = value;
      this._element.crossOrigin = value;

      return this;
    },

    /**
     * Returns original size of an image
     * @return {Object} Object with "width" and "height" properties
     */
    getOriginalSize: function() {
      var element = this.getElement();
      return {
        width: element.width,
        height: element.height
      };
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _stroke: function(ctx) {
      if (!this.stroke || this.strokeWidth === 0) {
        return;
      }
      var w = this.width / 2, h = this.height / 2;
      ctx.beginPath();
      ctx.moveTo(-w, -h);
      ctx.lineTo(w, -h);
      ctx.lineTo(w, h);
      ctx.lineTo(-w, h);
      ctx.lineTo(-w, -h);
      ctx.closePath();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderDashedStroke: function(ctx) {
      var x = -this.width / 2,
          y = -this.height / 2,
          w = this.width,
          h = this.height;

      ctx.save();
      this._setStrokeStyles(ctx);

      ctx.beginPath();
      fabric.util.drawDashedLine(ctx, x, y, x + w, y, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x + w, y, x + w, y + h, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x + w, y + h, x, y + h, this.strokeDashArray);
      fabric.util.drawDashedLine(ctx, x, y + h, x, y, this.strokeDashArray);
      ctx.closePath();
      ctx.restore();
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var filters = [ ], resizeFilters = [ ],
          element = this._originalElement,
          scaleX = 1, scaleY = 1;

      this.filters.forEach(function(filterObj) {
        if (filterObj) {
          if (filterObj.type === 'Resize') {
            scaleX *= filterObj.scaleX;
            scaleY *= filterObj.scaleY;
          }
          filters.push(filterObj.toObject());
        }
      });

      this.resizeFilters.forEach(function(filterObj) {
        filterObj && resizeFilters.push(filterObj.toObject());
      });

      var object = extend(this.callSuper('toObject', propertiesToInclude), {
        src: element ? element.src || element._src : '',
        filters: filters,
        resizeFilters: resizeFilters,
        crossOrigin: this.crossOrigin,
        alignX: this.alignX,
        alignY: this.alignY,
        meetOrSlice: this.meetOrSlice
      });

      object.width /= scaleX;
      object.height /= scaleY;

      if (!this.includeDefaultValues) {
        this._removeDefaultValues(object);
      }

      return object;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(), x = -this.width / 2, y = -this.height / 2,
          preserveAspectRatio = 'none';
      if (this.group && this.group.type === 'path-group') {
        x = this.left;
        y = this.top;
      }
      if (this.alignX !== 'none' && this.alignY !== 'none') {
        preserveAspectRatio = 'x' + this.alignX + 'Y' + this.alignY + ' ' + this.meetOrSlice;
      }
      markup.push(
        '<g transform="', this.getSvgTransform(), this.getSvgTransformMatrix(), '">\n',
          '<image ', this.getSvgId(), 'xlink:href="', this.getSvgSrc(),
            '" x="', x, '" y="', y,
            '" style="', this.getSvgStyles(),
            // we're essentially moving origin of transformation from top/left corner to the center of the shape
            // by wrapping it in container <g> element with actual transformation, then offsetting object to the top/left
            // so that object's center aligns with container's left/top
            '" width="', this.width,
            '" height="', this.height,
            '" preserveAspectRatio="', preserveAspectRatio, '"',
          '></image>\n'
      );

      if (this.stroke || this.strokeDashArray) {
        var origFill = this.fill;
        this.fill = null;
        markup.push(
          '<rect ',
            'x="', x, '" y="', y,
            '" width="', this.width, '" height="', this.height,
            '" style="', this.getSvgStyles(),
          '"/>\n'
        );
        this.fill = origFill;
      }

      markup.push('</g>\n');

      return reviver ? reviver(markup.join('')) : markup.join('');
    },
    /* _TO_SVG_END_ */

    /**
     * Returns source of an image
     * @return {String} Source of an image
     */
    getSrc: function() {
      if (this.getElement()) {
        return this.getElement().src || this.getElement()._src;
      }
    },

    /**
     * Sets source of an image
     * @param {String} src Source string (URL)
     * @param {Function} [callback] Callback is invoked when image has been loaded (and all filters have been applied)
     * @param {Object} [options] Options object
     * @return {fabric.Image} thisArg
     * @chainable
     */
    setSrc: function(src, callback, options) {
      fabric.util.loadImage(src, function(img) {
        return this.setElement(img, callback, options);
      }, this, options && options.crossOrigin);
    },

    /**
     * Returns string representation of an instance
     * @return {String} String representation of an instance
     */
    toString: function() {
      return '#<fabric.Image: { src: "' + this.getSrc() + '" }>';
    },

    /**
     * Returns a clone of an instance
     * @param {Function} callback Callback is invoked with a clone as a first argument
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     */
    clone: function(callback, propertiesToInclude) {
      this.constructor.fromObject(this.toObject(propertiesToInclude), callback);
    },

    /**
     * Applies filters assigned to this image (from "filters" array)
     * @method applyFilters
     * @param {Function} callback Callback is invoked when all filters have been applied and new image is generated
     * @return {fabric.Image} thisArg
     * @chainable
     */
    applyFilters: function(callback, filters, imgElement, forResizing) {

      filters = filters || this.filters;
      imgElement = imgElement || this._originalElement;

      if (!imgElement) {
        return;
      }

      var imgEl = imgElement,
          canvasEl = fabric.util.createCanvasElement(),
          replacement = fabric.util.createImage(),
          _this = this;

      canvasEl.width = imgEl.width;
      canvasEl.height = imgEl.height;
      canvasEl.getContext('2d').drawImage(imgEl, 0, 0, imgEl.width, imgEl.height);

      if (filters.length === 0) {
        this._element = imgElement;
        callback && callback();
        return canvasEl;
      }
      filters.forEach(function(filter) {
        filter && filter.applyTo(canvasEl, filter.scaleX || _this.scaleX, filter.scaleY || _this.scaleY);
        if (!forResizing && filter && filter.type === 'Resize') {
          _this.width *= filter.scaleX;
          _this.height *= filter.scaleY;
        }
      });

      /** @ignore */
      replacement.width = canvasEl.width;
      replacement.height = canvasEl.height;

      if (fabric.isLikelyNode) {
        replacement.src = canvasEl.toBuffer(undefined, fabric.Image.pngCompression);
        // onload doesn't fire in some node versions, so we invoke callback manually
        _this._element = replacement;
        !forResizing && (_this._filteredEl = replacement);
        callback && callback();
      }
      else {
        replacement.onload = function() {
          _this._element = replacement;
          !forResizing && (_this._filteredEl = replacement);
          callback && callback();
          replacement.onload = canvasEl = imgEl = null;
        };
        replacement.src = canvasEl.toDataURL('image/png');
      }
      return canvasEl;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx, noTransform) {
      var x, y, imageMargins = this._findMargins(), elementToDraw;

      x = (noTransform ? this.left : -this.width / 2);
      y = (noTransform ? this.top : -this.height / 2);

      if (this.meetOrSlice === 'slice') {
        ctx.beginPath();
        ctx.rect(x, y, this.width, this.height);
        ctx.clip();
      }

      if (this.isMoving === false && this.resizeFilters.length && this._needsResize()) {
        this._lastScaleX = this.scaleX;
        this._lastScaleY = this.scaleY;
        elementToDraw = this.applyFilters(null, this.resizeFilters, this._filteredEl || this._originalElement, true);
      }
      else {
        elementToDraw = this._element;
      }
      elementToDraw && ctx.drawImage(elementToDraw,
                                     x + imageMargins.marginX,
                                     y + imageMargins.marginY,
                                     imageMargins.width,
                                     imageMargins.height
                                    );

      this._stroke(ctx);
      this._renderStroke(ctx);
    },

    /**
     * @private, needed to check if image needs resize
     */
    _needsResize: function() {
      return (this.scaleX !== this._lastScaleX || this.scaleY !== this._lastScaleY);
    },

    /**
     * @private
     */
    _findMargins: function() {
      var width = this.width, height = this.height, scales,
          scale, marginX = 0, marginY = 0;

      if (this.alignX !== 'none' || this.alignY !== 'none') {
        scales = [this.width / this._element.width, this.height / this._element.height];
        scale = this.meetOrSlice === 'meet'
                ? Math.min.apply(null, scales) : Math.max.apply(null, scales);
        width = this._element.width * scale;
        height = this._element.height * scale;
        if (this.alignX === 'Mid') {
          marginX = (this.width - width) / 2;
        }
        if (this.alignX === 'Max') {
          marginX = this.width - width;
        }
        if (this.alignY === 'Mid') {
          marginY = (this.height - height) / 2;
        }
        if (this.alignY === 'Max') {
          marginY = this.height - height;
        }
      }
      return {
        width:  width,
        height: height,
        marginX: marginX,
        marginY: marginY
      };
    },

    /**
     * @private
     */
    _resetWidthHeight: function() {
      var element = this.getElement();

      this.set('width', element.width);
      this.set('height', element.height);
    },

    /**
     * The Image class's initialization method. This method is automatically
     * called by the constructor.
     * @private
     * @param {HTMLImageElement|String} element The element representing the image
     * @param {Object} [options] Options object
     */
    _initElement: function(element, options) {
      this.setElement(fabric.util.getById(element), null, options);
      fabric.util.addClass(this.getElement(), fabric.Image.CSS_CANVAS);
    },

    /**
     * @private
     * @param {Object} [options] Options object
     */
    _initConfig: function(options) {
      options || (options = { });
      this.setOptions(options);
      this._setWidthHeight(options);
      if (this._element && this.crossOrigin) {
        this._element.crossOrigin = this.crossOrigin;
      }
    },

    /**
     * @private
     * @param {Array} filters to be initialized
     * @param {Function} callback Callback to invoke when all fabric.Image.filters instances are created
     */
    _initFilters: function(filters, callback) {
      if (filters && filters.length) {
        fabric.util.enlivenObjects(filters, function(enlivenedObjects) {
          callback && callback(enlivenedObjects);
        }, 'fabric.Image.filters');
      }
      else {
        callback && callback();
      }
    },

    /**
     * @private
     * @param {Object} [options] Object with width/height properties
     */
    _setWidthHeight: function(options) {
      this.width = 'width' in options
        ? options.width
        : (this.getElement()
            ? this.getElement().width || 0
            : 0);

      this.height = 'height' in options
        ? options.height
        : (this.getElement()
            ? this.getElement().height || 0
            : 0);
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity of this instance
     */
    complexity: function() {
      return 1;
    }
  });

  /**
   * Default CSS class name for canvas
   * @static
   * @type String
   * @default
   */
  fabric.Image.CSS_CANVAS = 'canvas-img';

  /**
   * Alias for getSrc
   * @static
   */
  fabric.Image.prototype.getSvgSrc = fabric.Image.prototype.getSrc;

  /**
   * Creates an instance of fabric.Image from its object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @param {Function} [callback] Callback to invoke when an image instance is created
   */
  fabric.Image.fromObject = function(object, callback) {
    fabric.util.loadImage(object.src, function(img) {
      fabric.Image.prototype._initFilters.call(object, object.filters, function(filters) {
        object.filters = filters || [ ];
        fabric.Image.prototype._initFilters.call(object, object.resizeFilters, function(resizeFilters) {
          object.resizeFilters = resizeFilters || [ ];
          var instance = new fabric.Image(img, object);
          callback && callback(instance);
        });
      });
    }, null, object.crossOrigin);
  };

  /**
   * Creates an instance of fabric.Image from an URL string
   * @static
   * @param {String} url URL to create an image from
   * @param {Function} [callback] Callback to invoke when image is created (newly created image is passed as a first argument)
   * @param {Object} [imgOptions] Options object
   */
  fabric.Image.fromURL = function(url, callback, imgOptions) {
    fabric.util.loadImage(url, function(img) {
      callback && callback(new fabric.Image(img, imgOptions));
    }, null, imgOptions && imgOptions.crossOrigin);
  };

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Image.fromElement})
   * @static
   * @see {@link http://www.w3.org/TR/SVG/struct.html#ImageElement}
   */
  fabric.Image.ATTRIBUTE_NAMES =
    fabric.SHARED_ATTRIBUTES.concat('x y width height preserveAspectRatio xlink:href'.split(' '));

  /**
   * Returns {@link fabric.Image} instance from an SVG element
   * @static
   * @param {SVGElement} element Element to parse
   * @param {Function} callback Callback to execute when fabric.Image object is created
   * @param {Object} [options] Options object
   * @return {fabric.Image} Instance of fabric.Image
   */
  fabric.Image.fromElement = function(element, callback, options) {
    var parsedAttributes = fabric.parseAttributes(element, fabric.Image.ATTRIBUTE_NAMES),
        preserveAR;

    if (parsedAttributes.preserveAspectRatio) {
      preserveAR = fabric.util.parsePreserveAspectRatioAttribute(parsedAttributes.preserveAspectRatio);
      extend(parsedAttributes, preserveAR);
    }

    fabric.Image.fromURL(parsedAttributes['xlink:href'], callback,
      extend((options ? fabric.util.object.clone(options) : { }), parsedAttributes));
  };
  /* _FROM_SVG_END_ */

  /**
   * Indicates that instances of this type are async
   * @static
   * @type Boolean
   * @default
   */
  fabric.Image.async = true;

  /**
   * Indicates compression level used when generating PNG under Node (in applyFilters). Any of 0-9
   * @static
   * @type Number
   * @default
   */
  fabric.Image.pngCompression = 1;

})(typeof exports !== 'undefined' ? exports : this);


fabric.util.object.extend(fabric.Object.prototype, /** @lends fabric.Object.prototype */ {

  /**
   * @private
   * @return {Number} angle value
   */
  _getAngleValueForStraighten: function() {
    var angle = this.getAngle() % 360;
    if (angle > 0) {
      return Math.round((angle - 1) / 90) * 90;
    }
    return Math.round(angle / 90) * 90;
  },

  /**
   * Straightens an object (rotating it from current angle to one of 0, 90, 180, 270, etc. depending on which is closer)
   * @return {fabric.Object} thisArg
   * @chainable
   */
  straighten: function() {
    this.setAngle(this._getAngleValueForStraighten());
    return this;
  },

  /**
   * Same as {@link fabric.Object.prototype.straighten} but with animation
   * @param {Object} callbacks Object with callback functions
   * @param {Function} [callbacks.onComplete] Invoked on completion
   * @param {Function} [callbacks.onChange] Invoked on every step of animation
   * @return {fabric.Object} thisArg
   * @chainable
   */
  fxStraighten: function(callbacks) {
    callbacks = callbacks || { };

    var empty = function() { },
        onComplete = callbacks.onComplete || empty,
        onChange = callbacks.onChange || empty,
        _this = this;

    fabric.util.animate({
      startValue: this.get('angle'),
      endValue: this._getAngleValueForStraighten(),
      duration: this.FX_DURATION,
      onChange: function(value) {
        _this.setAngle(value);
        onChange();
      },
      onComplete: function() {
        _this.setCoords();
        onComplete();
      },
      onStart: function() {
        _this.set('active', false);
      }
    });

    return this;
  }
});

fabric.util.object.extend(fabric.StaticCanvas.prototype, /** @lends fabric.StaticCanvas.prototype */ {

  /**
   * Straightens object, then rerenders canvas
   * @param {fabric.Object} object Object to straighten
   * @return {fabric.Canvas} thisArg
   * @chainable
   */
  straightenObject: function (object) {
    object.straighten();
    this.renderAll();
    return this;
  },

  /**
   * Same as {@link fabric.Canvas.prototype.straightenObject}, but animated
   * @param {fabric.Object} object Object to straighten
   * @return {fabric.Canvas} thisArg
   * @chainable
   */
  fxStraightenObject: function (object) {
    object.fxStraighten({
      onChange: this.renderAll.bind(this)
    });
    return this;
  }
});


/**
 * @namespace fabric.Image.filters
 * @memberOf fabric.Image
 * @tutorial {@link http://fabricjs.com/fabric-intro-part-2#image_filters}
 * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
 */
fabric.Image.filters = fabric.Image.filters || { };

/**
 * Root filter class from which all filter classes inherit from
 * @class fabric.Image.filters.BaseFilter
 * @memberOf fabric.Image.filters
 */
fabric.Image.filters.BaseFilter = fabric.util.createClass(/** @lends fabric.Image.filters.BaseFilter.prototype */ {

  /**
   * Filter type
   * @param {String} type
   * @default
   */
  type: 'BaseFilter',

  /**
   * Constructor
   * @param {Object} [options] Options object
   */
  initialize: function(options) {
    if (options) {
      this.setOptions(options);
    }
  },

  /**
   * Sets filter's properties from options
   * @param {Object} [options] Options object
   */
  setOptions: function(options) {
    for (var prop in options) {
      this[prop] = options[prop];
    }
  },

  /**
   * Returns object representation of an instance
   * @return {Object} Object representation of an instance
   */
  toObject: function() {
    return { type: this.type };
  },

  /**
   * Returns a JSON representation of an instance
   * @return {Object} JSON
   */
  toJSON: function() {
    // delegate, not alias
    return this.toObject();
  }
});


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Brightness filter class
   * @class fabric.Image.filters.Brightness
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Brightness#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Brightness({
   *   brightness: 200
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Brightness = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Brightness.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Brightness',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Brightness.prototype
     * @param {Object} [options] Options object
     * @param {Number} [options.brightness=0] Value to brighten the image up (0..255)
     */
    initialize: function(options) {
      options = options || { };
      this.brightness = options.brightness || 0;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          brightness = this.brightness;

      for (var i = 0, len = data.length; i < len; i += 4) {
        data[i] += brightness;
        data[i + 1] += brightness;
        data[i + 2] += brightness;
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        brightness: this.brightness
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Brightness} Instance of fabric.Image.filters.Brightness
   */
  fabric.Image.filters.Brightness.fromObject = function(object) {
    return new fabric.Image.filters.Brightness(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Adapted from <a href="http://www.html5rocks.com/en/tutorials/canvas/imagefilters/">html5rocks article</a>
   * @class fabric.Image.filters.Convolute
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Convolute#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example <caption>Sharpen filter</caption>
   * var filter = new fabric.Image.filters.Convolute({
   *   matrix: [ 0, -1,  0,
   *            -1,  5, -1,
   *             0, -1,  0 ]
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   * @example <caption>Blur filter</caption>
   * var filter = new fabric.Image.filters.Convolute({
   *   matrix: [ 1/9, 1/9, 1/9,
   *             1/9, 1/9, 1/9,
   *             1/9, 1/9, 1/9 ]
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   * @example <caption>Emboss filter</caption>
   * var filter = new fabric.Image.filters.Convolute({
   *   matrix: [ 1,   1,  1,
   *             1, 0.7, -1,
   *            -1,  -1, -1 ]
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   * @example <caption>Emboss filter with opaqueness</caption>
   * var filter = new fabric.Image.filters.Convolute({
   *   opaque: true,
   *   matrix: [ 1,   1,  1,
   *             1, 0.7, -1,
   *            -1,  -1, -1 ]
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Convolute = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Convolute.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Convolute',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Convolute.prototype
     * @param {Object} [options] Options object
     * @param {Boolean} [options.opaque=false] Opaque value (true/false)
     * @param {Array} [options.matrix] Filter matrix
     */
    initialize: function(options) {
      options = options || { };

      this.opaque = options.opaque;
      this.matrix = options.matrix || [
        0, 0, 0,
        0, 1, 0,
        0, 0, 0
      ];
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {

      var weights = this.matrix,
          context = canvasEl.getContext('2d'),
          pixels = context.getImageData(0, 0, canvasEl.width, canvasEl.height),

          side = Math.round(Math.sqrt(weights.length)),
          halfSide = Math.floor(side/2),
          src = pixels.data,
          sw = pixels.width,
          sh = pixels.height,
          output = context.createImageData(sw, sh),
          dst = output.data,
          // go through the destination image pixels
          alphaFac = this.opaque ? 1 : 0,
          r, g, b, a, dstOff,
          scx, scy, srcOff, wt;

      for (var y = 0; y < sh; y++) {
        for (var x = 0; x < sw; x++) {
          dstOff = (y * sw + x) * 4;
          // calculate the weighed sum of the source image pixels that
          // fall under the convolution matrix
          r = 0; g = 0; b = 0; a = 0;

          for (var cy = 0; cy < side; cy++) {
            for (var cx = 0; cx < side; cx++) {
              scy = y + cy - halfSide;
              scx = x + cx - halfSide;

              /* jshint maxdepth:5 */
              if (scy < 0 || scy > sh || scx < 0 || scx > sw) {
                continue;
              }

              srcOff = (scy * sw + scx) * 4;
              wt = weights[cy * side + cx];

              r += src[srcOff] * wt;
              g += src[srcOff + 1] * wt;
              b += src[srcOff + 2] * wt;
              a += src[srcOff + 3] * wt;
            }
          }
          dst[dstOff] = r;
          dst[dstOff + 1] = g;
          dst[dstOff + 2] = b;
          dst[dstOff + 3] = a + alphaFac * (255 - a);
        }
      }

      context.putImageData(output, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        opaque: this.opaque,
        matrix: this.matrix
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Convolute} Instance of fabric.Image.filters.Convolute
   */
  fabric.Image.filters.Convolute.fromObject = function(object) {
    return new fabric.Image.filters.Convolute(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * GradientTransparency filter class
   * @class fabric.Image.filters.GradientTransparency
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.GradientTransparency#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.GradientTransparency({
   *   threshold: 200
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.GradientTransparency = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.GradientTransparency.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'GradientTransparency',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.GradientTransparency.prototype
     * @param {Object} [options] Options object
     * @param {Number} [options.threshold=100] Threshold value
     */
    initialize: function(options) {
      options = options || { };
      this.threshold = options.threshold || 100;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          threshold = this.threshold,
          total = data.length;

      for (var i = 0, len = data.length; i < len; i += 4) {
        data[i + 3] = threshold + 255 * (total - i) / total;
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        threshold: this.threshold
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.GradientTransparency} Instance of fabric.Image.filters.GradientTransparency
   */
  fabric.Image.filters.GradientTransparency.fromObject = function(object) {
    return new fabric.Image.filters.GradientTransparency(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { });

  /**
   * Grayscale image filter class
   * @class fabric.Image.filters.Grayscale
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Grayscale();
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Grayscale = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Grayscale.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Grayscale',

    /**
     * Applies filter to canvas element
     * @memberOf fabric.Image.filters.Grayscale.prototype
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          len = imageData.width * imageData.height * 4,
          index = 0,
          average;

      while (index < len) {
        average = (data[index] + data[index + 1] + data[index + 2]) / 3;
        data[index]     = average;
        data[index + 1] = average;
        data[index + 2] = average;
        index += 4;
      }

      context.putImageData(imageData, 0, 0);
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @return {fabric.Image.filters.Grayscale} Instance of fabric.Image.filters.Grayscale
   */
  fabric.Image.filters.Grayscale.fromObject = function() {
    return new fabric.Image.filters.Grayscale();
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { });

  /**
   * Invert filter class
   * @class fabric.Image.filters.Invert
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Invert();
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Invert = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Invert.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Invert',

    /**
     * Applies filter to canvas element
     * @memberOf fabric.Image.filters.Invert.prototype
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = data.length, i;

      for (i = 0; i < iLen; i+=4) {
        data[i] = 255 - data[i];
        data[i + 1] = 255 - data[i + 1];
        data[i + 2] = 255 - data[i + 2];
      }

      context.putImageData(imageData, 0, 0);
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @return {fabric.Image.filters.Invert} Instance of fabric.Image.filters.Invert
   */
  fabric.Image.filters.Invert.fromObject = function() {
    return new fabric.Image.filters.Invert();
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Mask filter class
   * See http://resources.aleph-1.com/mask/
   * @class fabric.Image.filters.Mask
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Mask#initialize} for constructor definition
   */
  fabric.Image.filters.Mask = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Mask.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Mask',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Mask.prototype
     * @param {Object} [options] Options object
     * @param {fabric.Image} [options.mask] Mask image object
     * @param {Number} [options.channel=0] Rgb channel (0, 1, 2 or 3)
     */
    initialize: function(options) {
      options = options || { };

      this.mask = options.mask;
      this.channel = [ 0, 1, 2, 3 ].indexOf(options.channel) > -1 ? options.channel : 0;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      if (!this.mask) {
        return;
      }

      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          maskEl = this.mask.getElement(),
          maskCanvasEl = fabric.util.createCanvasElement(),
          channel = this.channel,
          i,
          iLen = imageData.width * imageData.height * 4;

      maskCanvasEl.width = canvasEl.width;
      maskCanvasEl.height = canvasEl.height;

      maskCanvasEl.getContext('2d').drawImage(maskEl, 0, 0, canvasEl.width, canvasEl.height);

      var maskImageData = maskCanvasEl.getContext('2d').getImageData(0, 0, canvasEl.width, canvasEl.height),
          maskData = maskImageData.data;

      for (i = 0; i < iLen; i += 4) {
        data[i + 3] = maskData[i + channel];
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        mask: this.mask.toObject(),
        channel: this.channel
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @param {Function} [callback] Callback to invoke when a mask filter instance is created
   */
  fabric.Image.filters.Mask.fromObject = function(object, callback) {
    fabric.util.loadImage(object.mask.src, function(img) {
      object.mask = new fabric.Image(img, object.mask);
      callback && callback(new fabric.Image.filters.Mask(object));
    });
  };

  /**
   * Indicates that instances of this type are async
   * @static
   * @type Boolean
   * @default
   */
  fabric.Image.filters.Mask.async = true;

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Noise filter class
   * @class fabric.Image.filters.Noise
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Noise#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Noise({
   *   noise: 700
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Noise = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Noise.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Noise',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Noise.prototype
     * @param {Object} [options] Options object
     * @param {Number} [options.noise=0] Noise value
     */
    initialize: function(options) {
      options = options || { };
      this.noise = options.noise || 0;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          noise = this.noise, rand;

      for (var i = 0, len = data.length; i < len; i += 4) {

        rand = (0.5 - Math.random()) * noise;

        data[i] += rand;
        data[i + 1] += rand;
        data[i + 2] += rand;
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        noise: this.noise
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Noise} Instance of fabric.Image.filters.Noise
   */
  fabric.Image.filters.Noise.fromObject = function(object) {
    return new fabric.Image.filters.Noise(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Pixelate filter class
   * @class fabric.Image.filters.Pixelate
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Pixelate#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Pixelate({
   *   blocksize: 8
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Pixelate = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Pixelate.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Pixelate',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Pixelate.prototype
     * @param {Object} [options] Options object
     * @param {Number} [options.blocksize=4] Blocksize for pixelate
     */
    initialize: function(options) {
      options = options || { };
      this.blocksize = options.blocksize || 4;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = imageData.height,
          jLen = imageData.width,
          index, i, j, r, g, b, a;

      for (i = 0; i < iLen; i += this.blocksize) {
        for (j = 0; j < jLen; j += this.blocksize) {

          index = (i * 4) * jLen + (j * 4);

          r = data[index];
          g = data[index + 1];
          b = data[index + 2];
          a = data[index + 3];

          /*
           blocksize: 4

           [1,x,x,x,1]
           [x,x,x,x,1]
           [x,x,x,x,1]
           [x,x,x,x,1]
           [1,1,1,1,1]
           */

          for (var _i = i, _ilen = i + this.blocksize; _i < _ilen; _i++) {
            for (var _j = j, _jlen = j + this.blocksize; _j < _jlen; _j++) {
              index = (_i * 4) * jLen + (_j * 4);
              data[index] = r;
              data[index + 1] = g;
              data[index + 2] = b;
              data[index + 3] = a;
            }
          }
        }
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        blocksize: this.blocksize
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Pixelate} Instance of fabric.Image.filters.Pixelate
   */
  fabric.Image.filters.Pixelate.fromObject = function(object) {
    return new fabric.Image.filters.Pixelate(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Remove white filter class
   * @class fabric.Image.filters.RemoveWhite
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.RemoveWhite#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.RemoveWhite({
   *   threshold: 40,
   *   distance: 140
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.RemoveWhite = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.RemoveWhite.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'RemoveWhite',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.RemoveWhite.prototype
     * @param {Object} [options] Options object
     * @param {Number} [options.threshold=30] Threshold value
     * @param {Number} [options.distance=20] Distance value
     */
    initialize: function(options) {
      options = options || { };
      this.threshold = options.threshold || 30;
      this.distance = options.distance || 20;
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          threshold = this.threshold,
          distance = this.distance,
          limit = 255 - threshold,
          abs = Math.abs,
          r, g, b;

      for (var i = 0, len = data.length; i < len; i += 4) {
        r = data[i];
        g = data[i + 1];
        b = data[i + 2];

        if (r > limit &&
            g > limit &&
            b > limit &&
            abs(r - g) < distance &&
            abs(r - b) < distance &&
            abs(g - b) < distance
        ) {
          data[i + 3] = 0;
        }
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        threshold: this.threshold,
        distance: this.distance
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.RemoveWhite} Instance of fabric.Image.filters.RemoveWhite
   */
  fabric.Image.filters.RemoveWhite.fromObject = function(object) {
    return new fabric.Image.filters.RemoveWhite(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { });

  /**
   * Sepia filter class
   * @class fabric.Image.filters.Sepia
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Sepia();
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Sepia = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Sepia.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Sepia',

    /**
     * Applies filter to canvas element
     * @memberOf fabric.Image.filters.Sepia.prototype
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = data.length, i, avg;

      for (i = 0; i < iLen; i+=4) {
        avg = 0.3  * data[i] + 0.59 * data[i + 1] + 0.11 * data[i + 2];
        data[i] = avg + 100;
        data[i + 1] = avg + 50;
        data[i + 2] = avg + 255;
      }

      context.putImageData(imageData, 0, 0);
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @return {fabric.Image.filters.Sepia} Instance of fabric.Image.filters.Sepia
   */
  fabric.Image.filters.Sepia.fromObject = function() {
    return new fabric.Image.filters.Sepia();
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { });

  /**
   * Sepia2 filter class
   * @class fabric.Image.filters.Sepia2
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Sepia2();
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Sepia2 = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Sepia2.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Sepia2',

    /**
     * Applies filter to canvas element
     * @memberOf fabric.Image.filters.Sepia.prototype
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = data.length, i, r, g, b;

      for (i = 0; i < iLen; i+=4) {
        r = data[i];
        g = data[i + 1];
        b = data[i + 2];

        data[i] = (r * 0.393 + g * 0.769 + b * 0.189 ) / 1.351;
        data[i + 1] = (r * 0.349 + g * 0.686 + b * 0.168 ) / 1.203;
        data[i + 2] = (r * 0.272 + g * 0.534 + b * 0.131 ) / 2.140;
      }

      context.putImageData(imageData, 0, 0);
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @return {fabric.Image.filters.Sepia2} Instance of fabric.Image.filters.Sepia2
   */
  fabric.Image.filters.Sepia2.fromObject = function() {
    return new fabric.Image.filters.Sepia2();
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Tint filter class
   * Adapted from <a href="https://github.com/mezzoblue/PaintbrushJS">https://github.com/mezzoblue/PaintbrushJS</a>
   * @class fabric.Image.filters.Tint
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link fabric.Image.filters.Tint#initialize} for constructor definition
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example <caption>Tint filter with hex color and opacity</caption>
   * var filter = new fabric.Image.filters.Tint({
   *   color: '#3513B0',
   *   opacity: 0.5
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   * @example <caption>Tint filter with rgba color</caption>
   * var filter = new fabric.Image.filters.Tint({
   *   color: 'rgba(53, 21, 176, 0.5)'
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Tint = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Tint.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Tint',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Tint.prototype
     * @param {Object} [options] Options object
     * @param {String} [options.color=#000000] Color to tint the image with
     * @param {Number} [options.opacity] Opacity value that controls the tint effect's transparency (0..1)
     */
    initialize: function(options) {
      options = options || { };

      this.color = options.color || '#000000';
      this.opacity = typeof options.opacity !== 'undefined'
                      ? options.opacity
                      : new fabric.Color(this.color).getAlpha();
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = data.length, i,
          tintR, tintG, tintB,
          r, g, b, alpha1,
          source;

      source = new fabric.Color(this.color).getSource();

      tintR = source[0] * this.opacity;
      tintG = source[1] * this.opacity;
      tintB = source[2] * this.opacity;

      alpha1 = 1 - this.opacity;

      for (i = 0; i < iLen; i+=4) {
        r = data[i];
        g = data[i + 1];
        b = data[i + 2];

        // alpha compositing
        data[i] = tintR + r * alpha1;
        data[i + 1] = tintG + g * alpha1;
        data[i + 2] = tintB + b * alpha1;
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        color: this.color,
        opacity: this.opacity
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Tint} Instance of fabric.Image.filters.Tint
   */
  fabric.Image.filters.Tint.fromObject = function(object) {
    return new fabric.Image.filters.Tint(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend;

  /**
   * Multiply filter class
   * Adapted from <a href="http://www.laurenscorijn.com/articles/colormath-basics">http://www.laurenscorijn.com/articles/colormath-basics</a>
   * @class fabric.Image.filters.Multiply
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @example <caption>Multiply filter with hex color</caption>
   * var filter = new fabric.Image.filters.Multiply({
   *   color: '#F0F'
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   * @example <caption>Multiply filter with rgb color</caption>
   * var filter = new fabric.Image.filters.Multiply({
   *   color: 'rgb(53, 21, 176)'
   * });
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Multiply = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Multiply.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Multiply',

    /**
     * Constructor
     * @memberOf fabric.Image.filters.Multiply.prototype
     * @param {Object} [options] Options object
     * @param {String} [options.color=#000000] Color to multiply the image pixels with
     */
    initialize: function(options) {
      options = options || { };

      this.color = options.color || '#000000';
    },

    /**
     * Applies filter to canvas element
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          iLen = data.length, i,
          source;

      source = new fabric.Color(this.color).getSource();

      for (i = 0; i < iLen; i+=4) {
        data[i] *= source[0] / 255;
        data[i + 1] *= source[1] / 255;
        data[i + 2] *= source[2] / 255;
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return extend(this.callSuper('toObject'), {
        color: this.color
      });
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @param {Object} object Object to create an instance from
   * @return {fabric.Image.filters.Multiply} Instance of fabric.Image.filters.Multiply
   */
  fabric.Image.filters.Multiply.fromObject = function(object) {
    return new fabric.Image.filters.Multiply(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {
  'use strict';

  var fabric = global.fabric;

  /**
   * Color Blend filter class
   * @class fabric.Image.filter.Blend
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @example
   * var filter = new fabric.Image.filters.Blend({
   *  color: '#000',
   *  mode: 'multiply'
   * });
   *
   * var filter = new fabric.Image.filters.Blend({
   *  image: fabricImageObject,
   *  mode: 'multiply',
   *  alpha: 0.5
   * });

   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Blend = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Blend.prototype */{
    type: 'Blend',

    initialize: function(options) {
      options = options || {};
      this.color = options.color || '#000';
      this.image = options.image || false;
      this.mode = options.mode || 'multiply';
      this.alpha = options.alpha || 1;
    },

    applyTo: function(canvasEl) {
      var context = canvasEl.getContext('2d'),
          imageData = context.getImageData(0, 0, canvasEl.width, canvasEl.height),
          data = imageData.data,
          tr, tg, tb,
          r, g, b,
          _r, _g, _b,
          source,
          isImage = false;

      if (this.image) {
        // Blend images
        isImage = true;

        var _el = fabric.util.createCanvasElement();
        _el.width = this.image.width;
        _el.height = this.image.height;

        var tmpCanvas = new fabric.StaticCanvas(_el);
        tmpCanvas.add(this.image);
        var context2 =  tmpCanvas.getContext('2d');
        source = context2.getImageData(0, 0, tmpCanvas.width, tmpCanvas.height).data;
      }
      else {
        // Blend color
        source = new fabric.Color(this.color).getSource();

        tr = source[0] * this.alpha;
        tg = source[1] * this.alpha;
        tb = source[2] * this.alpha;
      }

      for (var i = 0, len = data.length; i < len; i += 4) {

        r = data[i];
        g = data[i + 1];
        b = data[i + 2];

        if (isImage) {
          tr = source[i] * this.alpha;
          tg = source[i + 1] * this.alpha;
          tb = source[i + 2] * this.alpha;
        }

        switch (this.mode) {
          case 'multiply':
            data[i] = r * tr / 255;
            data[i + 1] = g * tg / 255;
            data[i + 2] = b * tb / 255;
            break;
          case 'screen':
            data[i] = 1 - (1 - r) * (1 - tr);
            data[i + 1] = 1 - (1 - g) * (1 - tg);
            data[i + 2] = 1 - (1 - b) * (1 - tb);
            break;
          case 'add':
            data[i] = Math.min(255, r + tr);
            data[i + 1] = Math.min(255, g + tg);
            data[i + 2] = Math.min(255, b + tb);
            break;
          case 'diff':
          case 'difference':
            data[i] = Math.abs(r - tr);
            data[i + 1] = Math.abs(g - tg);
            data[i + 2] = Math.abs(b - tb);
            break;
          case 'subtract':
            _r = r - tr;
            _g = g - tg;
            _b = b - tb;

            data[i] = (_r < 0) ? 0 : _r;
            data[i + 1] = (_g < 0) ? 0 : _g;
            data[i + 2] = (_b < 0) ? 0 : _b;
            break;
          case 'darken':
            data[i] = Math.min(r, tr);
            data[i + 1] = Math.min(g, tg);
            data[i + 2] = Math.min(b, tb);
            break;
          case 'lighten':
            data[i] = Math.max(r, tr);
            data[i + 1] = Math.max(g, tg);
            data[i + 2] = Math.max(b, tb);
            break;
        }
      }

      context.putImageData(imageData, 0, 0);
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return {
        color: this.color,
        image: this.image,
        mode: this.mode,
        alpha: this.alpha
      };
    }
  });

  fabric.Image.filters.Blend.fromObject = function(object) {
    return new fabric.Image.filters.Blend(object);
  };
})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric  = global.fabric || (global.fabric = { }), pow = Math.pow, floor = Math.floor,
      sqrt = Math.sqrt, abs = Math.abs, max = Math.max, round = Math.round, sin = Math.sin,
      ceil = Math.ceil;

  /**
   * Resize image filter class
   * @class fabric.Image.filters.Resize
   * @memberOf fabric.Image.filters
   * @extends fabric.Image.filters.BaseFilter
   * @see {@link http://fabricjs.com/image-filters|ImageFilters demo}
   * @example
   * var filter = new fabric.Image.filters.Resize();
   * object.filters.push(filter);
   * object.applyFilters(canvas.renderAll.bind(canvas));
   */
  fabric.Image.filters.Resize = fabric.util.createClass(fabric.Image.filters.BaseFilter, /** @lends fabric.Image.filters.Resize.prototype */ {

    /**
     * Filter type
     * @param {String} type
     * @default
     */
    type: 'Resize',

    /**
     * Resize type
     * @param {String} resizeType
     * @default
     */
    resizeType: 'hermite',

    /**
     * Scale factor for resizing, x axis
     * @param {Number} scaleX
     * @default
     */
    scaleX: 0,

    /**
     * Scale factor for resizing, y axis
     * @param {Number} scaleY
     * @default
     */
    scaleY: 0,

    /**
     * LanczosLobes parameter for lanczos filter
     * @param {Number} lanczosLobes
     * @default
     */
    lanczosLobes: 3,

    /**
     * Applies filter to canvas element
     * @memberOf fabric.Image.filters.Resize.prototype
     * @param {Object} canvasEl Canvas element to apply filter to
     */
    applyTo: function(canvasEl, scaleX, scaleY) {

      this.rcpScaleX = 1 / scaleX;
      this.rcpScaleY = 1 / scaleY;

      var oW = canvasEl.width, oH = canvasEl.height,
          dW = round(oW * scaleX), dH = round(oH * scaleY),
          imageData;

      if (this.resizeType === 'sliceHack') {
        imageData = this.sliceByTwo(canvasEl, oW, oH, dW, dH);
      }
      if (this.resizeType === 'hermite') {
        imageData = this.hermiteFastResize(canvasEl, oW, oH, dW, dH);
      }
      if (this.resizeType === 'bilinear') {
        imageData = this.bilinearFiltering(canvasEl, oW, oH, dW, dH);
      }
      if (this.resizeType === 'lanczos') {
        imageData = this.lanczosResize(canvasEl, oW, oH, dW, dH);
      }
      canvasEl.width = dW;
      canvasEl.height = dH;
      canvasEl.getContext('2d').putImageData(imageData, 0, 0);
    },

    sliceByTwo: function(canvasEl, width, height, newWidth, newHeight) {
      var context = canvasEl.getContext('2d'), imageData,
          multW = 0.5, multH = 0.5, signW = 1, signH = 1,
          doneW = false, doneH = false, stepW = width, stepH = height,
          tmpCanvas = fabric.util.createCanvasElement(),
          tmpCtx = tmpCanvas.getContext('2d');
      newWidth = floor(newWidth);
      newHeight = floor(newHeight);
      tmpCanvas.width = max(newWidth, width);
      tmpCanvas.height = max(newHeight, height);

      if (newWidth > width) {
        multW = 2;
        signW = -1;
      }
      if (newHeight > height) {
        multH = 2;
        signH = -1;
      }
      imageData = context.getImageData(0, 0, width, height);
      canvasEl.width = max(newWidth, width);
      canvasEl.height = max(newHeight, height);
      context.putImageData(imageData, 0, 0);

      while (!doneW || !doneH) {
        width = stepW;
        height = stepH;
        if (newWidth * signW < floor(stepW * multW * signW)) {
          stepW = floor(stepW * multW);
        }
        else {
          stepW = newWidth;
          doneW = true;
        }
        if (newHeight * signH < floor(stepH * multH * signH)) {
          stepH = floor(stepH * multH);
        }
        else {
          stepH = newHeight;
          doneH = true;
        }
        imageData = context.getImageData(0, 0, width, height);
        tmpCtx.putImageData(imageData, 0, 0);
        context.clearRect(0, 0, stepW, stepH);
        context.drawImage(tmpCanvas, 0, 0, width, height, 0, 0, stepW, stepH);
      }
      return context.getImageData(0, 0, newWidth, newHeight);
    },

    lanczosResize: function(canvasEl, oW, oH, dW, dH) {

      function lanczosCreate(lobes) {
        return function(x) {
          if (x > lobes) {
            return 0;
          }
          x *= Math.PI;
          if (abs(x) < 1e-16) {
            return 1;
          }
          var xx = x / lobes;
          return sin(x) * sin(xx) / x / xx;
        };
      }

      function process(u) {
        var v, i, weight, idx, a, red, green,
            blue, alpha, fX, fY;
        center.x = (u + 0.5) * ratioX;
        icenter.x = floor(center.x);
        for (v = 0; v < dH; v++) {
          center.y = (v + 0.5) * ratioY;
          icenter.y = floor(center.y);
          a = 0, red = 0, green = 0, blue = 0, alpha = 0;
          for (i = icenter.x - range2X; i <= icenter.x + range2X; i++) {
            if (i < 0 || i >= oW) {
              continue;
            }
            fX = floor(1000 * abs(i - center.x));
            if (!cacheLanc[fX]) {
              cacheLanc[fX] = { };
            }
            for (var j = icenter.y - range2Y; j <= icenter.y + range2Y; j++) {
              if (j < 0 || j >= oH) {
                continue;
              }
              fY = floor(1000 * abs(j - center.y));
              if (!cacheLanc[fX][fY]) {
                cacheLanc[fX][fY] = lanczos(sqrt(pow(fX * rcpRatioX, 2) + pow(fY * rcpRatioY, 2)) / 1000);
              }
              weight = cacheLanc[fX][fY];
              if (weight > 0) {
                idx = (j * oW + i) * 4;
                a += weight;
                red += weight * srcData[idx];
                green += weight * srcData[idx + 1];
                blue += weight * srcData[idx + 2];
                alpha += weight * srcData[idx + 3];
              }
            }
          }
          idx = (v * dW + u) * 4;
          destData[idx] = red / a;
          destData[idx + 1] = green / a;
          destData[idx + 2] = blue / a;
          destData[idx + 3] = alpha / a;
        }

        if (++u < dW) {
          return process(u);
        }
        else {
          return destImg;
        }
      }

      var context = canvasEl.getContext('2d'),
          srcImg = context.getImageData(0, 0, oW, oH),
          destImg = context.getImageData(0, 0, dW, dH),
          srcData = srcImg.data, destData = destImg.data,
          lanczos = lanczosCreate(this.lanczosLobes),
          ratioX = this.rcpScaleX, ratioY = this.rcpScaleY,
          rcpRatioX = 2 / this.rcpScaleX, rcpRatioY = 2 / this.rcpScaleY,
          range2X = ceil(ratioX * this.lanczosLobes / 2),
          range2Y = ceil(ratioY * this.lanczosLobes / 2),
          cacheLanc = { }, center = { }, icenter = { };

      return process(0);
    },

    bilinearFiltering: function(canvasEl, w, h, w2, h2) {
      var a, b, c, d, x, y, i, j, xDiff, yDiff, chnl,
          color, offset = 0, origPix, ratioX = this.rcpScaleX,
          ratioY = this.rcpScaleY, context = canvasEl.getContext('2d'),
          w4 = 4 * (w - 1), img = context.getImageData(0, 0, w, h),
          pixels = img.data, destImage = context.getImageData(0, 0, w2, h2),
          destPixels = destImage.data;
      for (i = 0; i < h2; i++) {
        for (j = 0; j < w2; j++) {
          x = floor(ratioX * j);
          y = floor(ratioY * i);
          xDiff = ratioX * j - x;
          yDiff = ratioY * i - y;
          origPix = 4 * (y * w + x);

          for (chnl = 0; chnl < 4; chnl++) {
            a = pixels[origPix + chnl];
            b = pixels[origPix + 4 + chnl];
            c = pixels[origPix + w4 + chnl];
            d = pixels[origPix + w4 + 4 + chnl];
            color = a * (1 - xDiff) * (1 - yDiff) + b * xDiff * (1 - yDiff) +
                    c * yDiff * (1 - xDiff) + d * xDiff * yDiff;
            destPixels[offset++] = color;
          }
        }
      }
      return destImage;
    },

    hermiteFastResize: function(canvasEl, oW, oH, dW, dH) {
      var ratioW = this.rcpScaleX, ratioH = this.rcpScaleY,
          ratioWHalf = ceil(ratioW / 2),
          ratioHHalf = ceil(ratioH / 2),
          context = canvasEl.getContext('2d'),
          img = context.getImageData(0, 0, oW, oH), data = img.data,
          img2 = context.getImageData(0, 0, dW, dH), data2 = img2.data;
      for (var j = 0; j < dH; j++) {
        for (var i = 0; i < dW; i++) {
          var x2 = (i + j * dW) * 4, weight = 0, weights = 0, weightsAlpha = 0,
              gxR = 0, gxG = 0, gxB = 0, gxA = 0, centerY = (j + 0.5) * ratioH;
          for (var yy = floor(j * ratioH); yy < (j + 1) * ratioH; yy++) {
            var dy = abs(centerY - (yy + 0.5)) / ratioHHalf,
                centerX = (i + 0.5) * ratioW, w0 = dy * dy;
            for (var xx = floor(i * ratioW); xx < (i + 1) * ratioW; xx++) {
              var dx = abs(centerX - (xx + 0.5)) / ratioWHalf,
                  w = sqrt(w0 + dx * dx);
              /*jshint maxdepth:5 */
              if (w > 1 && w < -1) {
                continue;
              }
              //hermite filter
              weight = 2 * w * w * w - 3 * w * w + 1;
              if (weight > 0) {
                dx = 4 * (xx + yy * oW);
                //alpha
                gxA += weight * data[dx + 3];
                weightsAlpha += weight;
                //colors
                /*jshint maxdepth:6 */
                if (data[dx + 3] < 255) {
                  weight = weight * data[dx + 3] / 250;
                }
                /*jshint maxdepth:5 */
                gxR += weight * data[dx];
                gxG += weight * data[dx + 1];
                gxB += weight * data[dx + 2];
                weights += weight;
              }
              /*jshint maxdepth:4 */
            }
          }
          data2[x2] = gxR / weights;
          data2[x2 + 1] = gxG / weights;
          data2[x2 + 2] = gxB / weights;
          data2[x2 + 3] = gxA / weightsAlpha;
        }
      }
      return img2;
    },

    /**
     * Returns object representation of an instance
     * @return {Object} Object representation of an instance
     */
    toObject: function() {
      return {
        type: this.type,
        scaleX: this.scaleX,
        scaleY: this.scaleY,
        resizeType: this.resizeType,
        lanczosLobes: this.lanczosLobes
      };
    }
  });

  /**
   * Returns filter instance from an object representation
   * @static
   * @return {fabric.Image.filters.Resize} Instance of fabric.Image.filters.Resize
   */
  fabric.Image.filters.Resize.fromObject = function(object) {
    return new fabric.Image.filters.Resize(object);
  };

})(typeof exports !== 'undefined' ? exports : this);


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = { }),
      extend = fabric.util.object.extend,
      clone = fabric.util.object.clone,
      toFixed = fabric.util.toFixed,
      supportsLineDash = fabric.StaticCanvas.supports('setLineDash'),
      NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS;

  if (fabric.Text) {
    fabric.warn('fabric.Text is already defined');
    return;
  }

  var stateProperties = fabric.Object.prototype.stateProperties.concat();
  stateProperties.push(
    'fontFamily',
    'fontWeight',
    'fontSize',
    'text',
    'textDecoration',
    'textAlign',
    'fontStyle',
    'lineHeight',
    'textBackgroundColor'
  );

  /**
   * Text class
   * @class fabric.Text
   * @extends fabric.Object
   * @return {fabric.Text} thisArg
   * @tutorial {@link http://fabricjs.com/fabric-intro-part-2#text}
   * @see {@link fabric.Text#initialize} for constructor definition
   */
  fabric.Text = fabric.util.createClass(fabric.Object, /** @lends fabric.Text.prototype */ {

    /**
     * Properties which when set cause object to change dimensions
     * @type Object
     * @private
     */
    _dimensionAffectingProps: {
      fontSize: true,
      fontWeight: true,
      fontFamily: true,
      fontStyle: true,
      lineHeight: true,
      stroke: true,
      strokeWidth: true,
      text: true,
      textAlign: true
    },

    /**
     * @private
     */
    _reNewline: /\r?\n/,

    /**
     * Use this regular expression to filter for whitespace that is not a new line.
     * Mostly used when text is 'justify' aligned.
     * @private
     */
    _reSpacesAndTabs: /[ \t\r]+/g,

    /**
     * Retrieves object's fontSize
     * @method getFontSize
     * @memberOf fabric.Text.prototype
     * @return {String} Font size (in pixels)
     */

    /**
     * Sets object's fontSize
     * Does not update the object .width and .height,
     * call ._initDimensions() to update the values.
     * @method setFontSize
     * @memberOf fabric.Text.prototype
     * @param {Number} fontSize Font size (in pixels)
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's fontWeight
     * @method getFontWeight
     * @memberOf fabric.Text.prototype
     * @return {(String|Number)} Font weight
     */

    /**
     * Sets object's fontWeight
     * Does not update the object .width and .height,
     * call ._initDimensions() to update the values.
     * @method setFontWeight
     * @memberOf fabric.Text.prototype
     * @param {(Number|String)} fontWeight Font weight
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's fontFamily
     * @method getFontFamily
     * @memberOf fabric.Text.prototype
     * @return {String} Font family
     */

    /**
     * Sets object's fontFamily
     * Does not update the object .width and .height,
     * call ._initDimensions() to update the values.
     * @method setFontFamily
     * @memberOf fabric.Text.prototype
     * @param {String} fontFamily Font family
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's text
     * @method getText
     * @memberOf fabric.Text.prototype
     * @return {String} text
     */

    /**
     * Sets object's text
     * Does not update the object .width and .height,
     * call ._initDimensions() to update the values.
     * @method setText
     * @memberOf fabric.Text.prototype
     * @param {String} text Text
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's textDecoration
     * @method getTextDecoration
     * @memberOf fabric.Text.prototype
     * @return {String} Text decoration
     */

    /**
     * Sets object's textDecoration
     * @method setTextDecoration
     * @memberOf fabric.Text.prototype
     * @param {String} textDecoration Text decoration
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's fontStyle
     * @method getFontStyle
     * @memberOf fabric.Text.prototype
     * @return {String} Font style
     */

    /**
     * Sets object's fontStyle
     * Does not update the object .width and .height,
     * call ._initDimensions() to update the values.
     * @method setFontStyle
     * @memberOf fabric.Text.prototype
     * @param {String} fontStyle Font style
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's lineHeight
     * @method getLineHeight
     * @memberOf fabric.Text.prototype
     * @return {Number} Line height
     */

    /**
     * Sets object's lineHeight
     * @method setLineHeight
     * @memberOf fabric.Text.prototype
     * @param {Number} lineHeight Line height
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's textAlign
     * @method getTextAlign
     * @memberOf fabric.Text.prototype
     * @return {String} Text alignment
     */

    /**
     * Sets object's textAlign
     * @method setTextAlign
     * @memberOf fabric.Text.prototype
     * @param {String} textAlign Text alignment
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Retrieves object's textBackgroundColor
     * @method getTextBackgroundColor
     * @memberOf fabric.Text.prototype
     * @return {String} Text background color
     */

    /**
     * Sets object's textBackgroundColor
     * @method setTextBackgroundColor
     * @memberOf fabric.Text.prototype
     * @param {String} textBackgroundColor Text background color
     * @return {fabric.Text}
     * @chainable
     */

    /**
     * Type of an object
     * @type String
     * @default
     */
    type:                 'text',

    /**
     * Font size (in pixels)
     * @type Number
     * @default
     */
    fontSize:             40,

    /**
     * Font weight (e.g. bold, normal, 400, 600, 800)
     * @type {(Number|String)}
     * @default
     */
    fontWeight:           'normal',

    /**
     * Font family
     * @type String
     * @default
     */
    fontFamily:           'Times New Roman',

    /**
     * Text decoration Possible values: "", "underline", "overline" or "line-through".
     * @type String
     * @default
     */
    textDecoration:       '',

    /**
     * Text alignment. Possible values: "left", "center", "right" or "justify".
     * @type String
     * @default
     */
    textAlign:            'left',

    /**
     * Font style . Possible values: "", "normal", "italic" or "oblique".
     * @type String
     * @default
     */
    fontStyle:            '',

    /**
     * Line height
     * @type Number
     * @default
     */
    lineHeight:           1.16,

    /**
     * Background color of text lines
     * @type String
     * @default
     */
    textBackgroundColor:  '',

    /**
     * List of properties to consider when checking if
     * state of an object is changed ({@link fabric.Object#hasStateChanged})
     * as well as for history (undo/redo) purposes
     * @type Array
     */
    stateProperties:      stateProperties,

    /**
     * When defined, an object is rendered via stroke and this property specifies its color.
     * <b>Backwards incompatibility note:</b> This property was named "strokeStyle" until v1.1.6
     * @type String
     * @default
     */
    stroke:               null,

    /**
     * Shadow object representing shadow of this shape.
     * <b>Backwards incompatibility note:</b> This property was named "textShadow" (String) until v1.2.11
     * @type fabric.Shadow
     * @default
     */
    shadow:               null,

    /**
     * @private
     */
    _fontSizeFraction: 0.25,

    /**
     * Text Line proportion to font Size (in pixels)
     * @type Number
     * @default
     */
    _fontSizeMult:             1.13,

    /**
     * Constructor
     * @param {String} text Text string
     * @param {Object} [options] Options object
     * @return {fabric.Text} thisArg
     */
    initialize: function(text, options) {
      options = options || { };
      this.text = text;
      this.__skipDimension = true;
      this.setOptions(options);
      this.__skipDimension = false;
      this._initDimensions();
    },

    /**
     * Initialize text dimensions. Render all text on given context
     * or on a offscreen canvas to get the text width with measureText.
     * Updates this.width and this.height with the proper values.
     * Does not return dimensions.
     * @param {CanvasRenderingContext2D} [ctx] Context to render on
     * @private
     */
    _initDimensions: function(ctx) {
      if (this.__skipDimension) {
        return;
      }
      if (!ctx) {
        ctx = fabric.util.createCanvasElement().getContext('2d');
        this._setTextStyles(ctx);
      }
      this._textLines = this._splitTextIntoLines();
      this._clearCache();
      this.width = this._getTextWidth(ctx);
      this.height = this._getTextHeight(ctx);
    },

    /**
     * Returns string representation of an instance
     * @return {String} String representation of text object
     */
    toString: function() {
      return '#<fabric.Text (' + this.complexity() +
        '): { "text": "' + this.text + '", "fontFamily": "' + this.fontFamily + '" }>';
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx) {
      this.clipTo && fabric.util.clipContext(this, ctx);
      this._setOpacity(ctx);
      this._setShadow(ctx);
      this._setupCompositeOperation(ctx);
      this._renderTextBackground(ctx);
      this._setStrokeStyles(ctx);
      this._setFillStyles(ctx);
      this._renderText(ctx);
      this._renderTextDecoration(ctx);
      this.clipTo && ctx.restore();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderText: function(ctx) {

      this._translateForTextAlign(ctx);
      this._renderTextFill(ctx);
      this._renderTextStroke(ctx);
      this._translateForTextAlign(ctx, true);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Boolean} back Indicates if translate back or forward
     */
    _translateForTextAlign: function(ctx, back) {
      if (this.textAlign !== 'left' && this.textAlign !== 'justify') {
        var sign = back ? -1 : 1;
        ctx.translate(this.textAlign === 'center' ? (sign * this.width / 2) : sign * this.width, 0);
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _setTextStyles: function(ctx) {
      ctx.textBaseline = 'alphabetic';
      if (!this.skipTextAlign) {
        ctx.textAlign = this.textAlign;
      }
      ctx.font = this._getFontDeclaration();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @return {Number} Height of fabric.Text object
     */
    _getTextHeight: function() {
      return this._getHeightOfSingleLine() + (this._textLines.length - 1) * this._getHeightOfLine();
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @return {Number} Maximum width of fabric.Text object
     */
    _getTextWidth: function(ctx) {
      var maxWidth = this._getLineWidth(ctx, 0);

      for (var i = 1, len = this._textLines.length; i < len; i++) {
        var currentLineWidth = this._getLineWidth(ctx, i);
        if (currentLineWidth > maxWidth) {
          maxWidth = currentLineWidth;
        }
      }
      return maxWidth;
    },

    /*
     * Calculate object dimensions from its properties
     * @override
     * @private
     */
    _getNonTransformedDimensions: function() {
      return { x: this.width, y: this.height };
    },

    /**
     * @private
     * @param {String} method Method name ("fillText" or "strokeText")
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {String} chars Chars to render
     * @param {Number} left Left position of text
     * @param {Number} top Top position of text
     */
    _renderChars: function(method, ctx, chars, left, top) {
      // remove Text word from method var
      var shortM = method.slice(0, -4);
      if (this[shortM].toLive) {
        var offsetX = -this.width / 2 + this[shortM].offsetX || 0,
            offsetY = -this.height / 2 + this[shortM].offsetY || 0;
        ctx.save();
        ctx.translate(offsetX, offsetY);
        left -= offsetX;
        top -= offsetY;
      }
      ctx[method](chars, left, top);
      this[shortM].toLive && ctx.restore();
    },

    /**
     * @private
     * @param {String} method Method name ("fillText" or "strokeText")
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {String} line Text to render
     * @param {Number} left Left position of text
     * @param {Number} top Top position of text
     * @param {Number} lineIndex Index of a line in a text
     */
    _renderTextLine: function(method, ctx, line, left, top, lineIndex) {
      // lift the line by quarter of fontSize
      top -= this.fontSize * this._fontSizeFraction;

      // short-circuit
      var lineWidth = this._getLineWidth(ctx, lineIndex);
      if (this.textAlign !== 'justify' || this.width < lineWidth) {
        this._renderChars(method, ctx, line, left, top, lineIndex);
        return;
      }

      // stretch the line
      var words = line.split(/\s+/),
          charOffset = 0,
          wordsWidth = this._getWidthOfWords(ctx, line, lineIndex, 0),
          widthDiff = this.width - wordsWidth,
          numSpaces = words.length - 1,
          spaceWidth = numSpaces > 0 ? widthDiff / numSpaces : 0,
          leftOffset = 0, word;

      for (var i = 0, len = words.length; i < len; i++) {
        while (line[charOffset] === ' ' && charOffset < line.length) {
          charOffset++;
        }
        word = words[i];
        this._renderChars(method, ctx, word, left + leftOffset, top, lineIndex, charOffset);
        leftOffset += this._getWidthOfWords(ctx, word, lineIndex, charOffset) + spaceWidth;
        charOffset += word.length;
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} line
     */
    _getWidthOfWords: function (ctx, line) {
      return ctx.measureText(line.replace(/\s+/g, '')).width;
    },

    /**
     * @private
     * @return {Number} Left offset
     */
    _getLeftOffset: function() {
      return -this.width / 2;
    },

    /**
     * @private
     * @return {Number} Top offset
     */
    _getTopOffset: function() {
      return -this.height / 2;
    },

    /**
     * Returns true because text has no style
     */
    isEmptyStyles: function() {
      return true;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextFill: function(ctx) {
      if (!this.fill && this.isEmptyStyles()) {
        return;
      }

      var lineHeights = 0;

      for (var i = 0, len = this._textLines.length; i < len; i++) {
        var heightOfLine = this._getHeightOfLine(ctx, i),
            maxHeight = heightOfLine / this.lineHeight;

        this._renderTextLine(
          'fillText',
          ctx,
          this._textLines[i],
          this._getLeftOffset(),
          this._getTopOffset() + lineHeights + maxHeight,
          i
        );
        lineHeights += heightOfLine;
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextStroke: function(ctx) {
      if ((!this.stroke || this.strokeWidth === 0) && this.isEmptyStyles()) {
        return;
      }

      var lineHeights = 0;

      if (this.shadow && !this.shadow.affectStroke) {
        this._removeShadow(ctx);
      }

      ctx.save();

      if (this.strokeDashArray) {
        // Spec requires the concatenation of two copies the dash list when the number of elements is odd
        if (1 & this.strokeDashArray.length) {
          this.strokeDashArray.push.apply(this.strokeDashArray, this.strokeDashArray);
        }
        supportsLineDash && ctx.setLineDash(this.strokeDashArray);
      }

      ctx.beginPath();
      for (var i = 0, len = this._textLines.length; i < len; i++) {
        var heightOfLine = this._getHeightOfLine(ctx, i),
            maxHeight = heightOfLine / this.lineHeight;

        this._renderTextLine(
          'strokeText',
          ctx,
          this._textLines[i],
          this._getLeftOffset(),
          this._getTopOffset() + lineHeights + maxHeight,
          i
        );
        lineHeights += heightOfLine;
      }
      ctx.closePath();
      ctx.restore();
    },

    /**
     * @private
     * @return {Number} height of line
     */
    _getHeightOfLine: function() {
      return this._getHeightOfSingleLine() * this.lineHeight;
    },

    /**
     * @private
     * @return {Number} height of line without lineHeight
     */
    _getHeightOfSingleLine: function() {
      return this.fontSize * this._fontSizeMult;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Array} textLines Array of all text lines
     */
    _renderTextBackground: function(ctx) {
      this._renderTextBoxBackground(ctx);
      this._renderTextLinesBackground(ctx);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextBoxBackground: function(ctx) {
      if (!this.backgroundColor) {
        return;
      }

      ctx.fillStyle = this.backgroundColor;

      ctx.fillRect(
        this._getLeftOffset(),
        this._getTopOffset(),
        this.width,
        this.height
      );
      // if there is background color no other shadows
      // should be casted
      this._removeShadow(ctx);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextLinesBackground: function(ctx) {
      if (!this.textBackgroundColor) {
        return;
      }
      var lineTopOffset = 0, heightOfLine,
          lineWidth, lineLeftOffset;

      ctx.fillStyle = this.textBackgroundColor;
      for (var i = 0, len = this._textLines.length; i < len; i++) {
        heightOfLine = this._getHeightOfLine(ctx, i);
        lineWidth = this._getLineWidth(ctx, i);
        if (lineWidth > 0) {
          lineLeftOffset = this._getLineLeftOffset(lineWidth);
          ctx.fillRect(
            this._getLeftOffset() + lineLeftOffset,
            this._getTopOffset() + lineTopOffset,
            lineWidth,
            heightOfLine / this.lineHeight
          );
        }
        lineTopOffset += heightOfLine;
      }
      // if there is text background color no
      // other shadows should be casted
      this._removeShadow(ctx);
    },

    /**
     * @private
     * @param {Number} lineWidth Width of text line
     * @return {Number} Line left offset
     */
    _getLineLeftOffset: function(lineWidth) {
      if (this.textAlign === 'center') {
        return (this.width - lineWidth) / 2;
      }
      if (this.textAlign === 'right') {
        return this.width - lineWidth;
      }
      return 0;
    },

    /**
     * @private
     */
    _clearCache: function() {
      this.__lineWidths = [ ];
      this.__lineHeights = [ ];
    },

    /**
     * @private
     */
    _shouldClearCache: function() {
      var shouldClear = false;
      if (this._forceClearCache) {
        this._forceClearCache = false;
        return true;
      }
      for (var prop in this._dimensionAffectingProps) {
        if (this['__' + prop] !== this[prop]) {
          this['__' + prop] = this[prop];
          shouldClear = true;
        }
      }
      return shouldClear;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex line number
     * @return {Number} Line width
     */
    _getLineWidth: function(ctx, lineIndex) {
      if (this.__lineWidths[lineIndex]) {
        return this.__lineWidths[lineIndex] === -1 ? this.width : this.__lineWidths[lineIndex];
      }

      var width, wordCount, line = this._textLines[lineIndex];

      if (line === '') {
        width = 0;
      }
      else {
        width = this._measureLine(ctx, lineIndex);
      }
      this.__lineWidths[lineIndex] = width;

      if (width && this.textAlign === 'justify') {
        wordCount = line.split(/\s+/);
        if (wordCount.length > 1) {
          this.__lineWidths[lineIndex] = -1;
        }
      }
      return width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex line number
     * @return {Number} Line width
     */
    _measureLine: function(ctx, lineIndex) {
      return ctx.measureText(this._textLines[lineIndex]).width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextDecoration: function(ctx) {
      if (!this.textDecoration) {
        return;
      }

      var halfOfVerticalBox = this.height / 2,
          _this = this, offsets = [];

      /** @ignore */
      function renderLinesAtOffset(offsets) {
        var i, lineHeight = 0, len, j, oLen, lineWidth,
            lineLeftOffset, heightOfLine;

        for (i = 0, len = _this._textLines.length; i < len; i++) {

          lineWidth = _this._getLineWidth(ctx, i),
          lineLeftOffset = _this._getLineLeftOffset(lineWidth),
          heightOfLine = _this._getHeightOfLine(ctx, i);

          for (j = 0, oLen = offsets.length; j < oLen; j++) {
            ctx.fillRect(
              _this._getLeftOffset() + lineLeftOffset,
              lineHeight + (_this._fontSizeMult - 1 + offsets[j] ) * _this.fontSize - halfOfVerticalBox,
              lineWidth,
              _this.fontSize / 15);
          }
          lineHeight += heightOfLine;
        }
      }

      if (this.textDecoration.indexOf('underline') > -1) {
        offsets.push(0.85); // 1 - 3/16
      }
      if (this.textDecoration.indexOf('line-through') > -1) {
        offsets.push(0.43);
      }
      if (this.textDecoration.indexOf('overline') > -1) {
        offsets.push(-0.12);
      }
      if (offsets.length > 0) {
        renderLinesAtOffset(offsets);
      }
    },

    /**
     * @private
     */
    _getFontDeclaration: function() {
      return [
        // node-canvas needs "weight style", while browsers need "style weight"
        (fabric.isLikelyNode ? this.fontWeight : this.fontStyle),
        (fabric.isLikelyNode ? this.fontStyle : this.fontWeight),
        this.fontSize + 'px',
        (fabric.isLikelyNode ? ('"' + this.fontFamily + '"') : this.fontFamily)
      ].join(' ');
    },

    /**
     * Renders text instance on a specified context
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    render: function(ctx, noTransform) {
      // do not render if object is not visible
      if (!this.visible) {
        return;
      }

      ctx.save();
      this._setTextStyles(ctx);

      if (this._shouldClearCache()) {
        this._initDimensions(ctx);
      }
      this.drawSelectionBackground(ctx);
      if (!noTransform) {
        this.transform(ctx);
      }
      if (this.transformMatrix) {
        ctx.transform.apply(ctx, this.transformMatrix);
      }
      if (this.group && this.group.type === 'path-group') {
        ctx.translate(this.left, this.top);
      }
      this._render(ctx);
      ctx.restore();
    },

    /**
     * Returns the text as an array of lines.
     * @returns {Array} Lines in the text
     */
    _splitTextIntoLines: function() {
      return this.text.split(this._reNewline);
    },

    /**
     * Returns object representation of an instance
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} Object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var object = extend(this.callSuper('toObject', propertiesToInclude), {
        text:                 this.text,
        fontSize:             this.fontSize,
        fontWeight:           this.fontWeight,
        fontFamily:           this.fontFamily,
        fontStyle:            this.fontStyle,
        lineHeight:           this.lineHeight,
        textDecoration:       this.textDecoration,
        textAlign:            this.textAlign,
        textBackgroundColor:  this.textBackgroundColor
      });
      if (!this.includeDefaultValues) {
        this._removeDefaultValues(object);
      }
      return object;
    },

    /* _TO_SVG_START_ */
    /**
     * Returns SVG representation of an instance
     * @param {Function} [reviver] Method for further parsing of svg representation.
     * @return {String} svg representation of an instance
     */
    toSVG: function(reviver) {
      var markup = this._createBaseSVGMarkup(),
          offsets = this._getSVGLeftTopOffsets(this.ctx),
          textAndBg = this._getSVGTextAndBg(offsets.textTop, offsets.textLeft);
      this._wrapSVGTextAndBg(markup, textAndBg);

      return reviver ? reviver(markup.join('')) : markup.join('');
    },

    /**
     * @private
     */
    _getSVGLeftTopOffsets: function(ctx) {
      var lineTop = this._getHeightOfLine(ctx, 0),
          textLeft = -this.width / 2,
          textTop = 0;

      return {
        textLeft: textLeft + (this.group && this.group.type === 'path-group' ? this.left : 0),
        textTop: textTop + (this.group && this.group.type === 'path-group' ? -this.top : 0),
        lineTop: lineTop
      };
    },

    /**
     * @private
     */
    _wrapSVGTextAndBg: function(markup, textAndBg) {
      var noShadow = true, filter = this.getSvgFilter(),
          style = filter === '' ? '' : ' style="' + filter + '"';

      markup.push(
        '\t<g ', this.getSvgId(), 'transform="', this.getSvgTransform(), this.getSvgTransformMatrix(), '"',
          style, '>\n',
          textAndBg.textBgRects.join(''),
          '\t\t<text ',
            (this.fontFamily ? 'font-family="' + this.fontFamily.replace(/"/g, '\'') + '" ': ''),
            (this.fontSize ? 'font-size="' + this.fontSize + '" ': ''),
            (this.fontStyle ? 'font-style="' + this.fontStyle + '" ': ''),
            (this.fontWeight ? 'font-weight="' + this.fontWeight + '" ': ''),
            (this.textDecoration ? 'text-decoration="' + this.textDecoration + '" ': ''),
            'style="', this.getSvgStyles(noShadow), '" >\n',
            textAndBg.textSpans.join(''),
          '\t\t</text>\n',
        '\t</g>\n'
      );
    },

    /**
     * @private
     * @param {Number} textTopOffset Text top offset
     * @param {Number} textLeftOffset Text left offset
     * @return {Object}
     */
    _getSVGTextAndBg: function(textTopOffset, textLeftOffset) {
      var textSpans = [ ],
          textBgRects = [ ],
          height = 0;
      // bounding-box background
      this._setSVGBg(textBgRects);

      // text and text-background
      for (var i = 0, len = this._textLines.length; i < len; i++) {
        if (this.textBackgroundColor) {
          this._setSVGTextLineBg(textBgRects, i, textLeftOffset, textTopOffset, height);
        }
        this._setSVGTextLineText(i, textSpans, height, textLeftOffset, textTopOffset, textBgRects);
        height += this._getHeightOfLine(this.ctx, i);
      }

      return {
        textSpans: textSpans,
        textBgRects: textBgRects
      };
    },

    _setSVGTextLineText: function(i, textSpans, height, textLeftOffset, textTopOffset) {
      var yPos = this.fontSize * (this._fontSizeMult - this._fontSizeFraction)
        - textTopOffset + height - this.height / 2;
      if (this.textAlign === 'justify') {
        // i call from here to do not intefere with IText
        this._setSVGTextLineJustifed(i, textSpans, yPos, textLeftOffset);
        return;
      }
      textSpans.push(
        '\t\t\t<tspan x="',
          toFixed(textLeftOffset + this._getLineLeftOffset(this._getLineWidth(this.ctx, i)), NUM_FRACTION_DIGITS), '" ',
          'y="',
          toFixed(yPos, NUM_FRACTION_DIGITS),
          '" ',
          // doing this on <tspan> elements since setting opacity
          // on containing <text> one doesn't work in Illustrator
          this._getFillAttributes(this.fill), '>',
          fabric.util.string.escapeXml(this._textLines[i]),
        '</tspan>\n'
      );
    },

    _setSVGTextLineJustifed: function(i, textSpans, yPos, textLeftOffset) {
      var ctx = fabric.util.createCanvasElement().getContext('2d');

      this._setTextStyles(ctx);

      var line = this._textLines[i],
          words = line.split(/\s+/),
          wordsWidth = this._getWidthOfWords(ctx, line),
          widthDiff = this.width - wordsWidth,
          numSpaces = words.length - 1,
          spaceWidth = numSpaces > 0 ? widthDiff / numSpaces : 0,
          word, attributes = this._getFillAttributes(this.fill),
          len;

      textLeftOffset += this._getLineLeftOffset(this._getLineWidth(ctx, i));

      for (i = 0, len = words.length; i < len; i++) {
        word = words[i];
        textSpans.push(
          '\t\t\t<tspan x="',
            toFixed(textLeftOffset, NUM_FRACTION_DIGITS), '" ',
            'y="',
            toFixed(yPos, NUM_FRACTION_DIGITS),
            '" ',
            // doing this on <tspan> elements since setting opacity
            // on containing <text> one doesn't work in Illustrator
            attributes, '>',
            fabric.util.string.escapeXml(word),
          '</tspan>\n'
        );
        textLeftOffset += this._getWidthOfWords(ctx, word) + spaceWidth;
      }
    },

    _setSVGTextLineBg: function(textBgRects, i, textLeftOffset, textTopOffset, height) {
      textBgRects.push(
        '\t\t<rect ',
          this._getFillAttributes(this.textBackgroundColor),
          ' x="',
          toFixed(textLeftOffset + this._getLineLeftOffset(this._getLineWidth(this.ctx, i)), NUM_FRACTION_DIGITS),
          '" y="',
          toFixed(height - this.height / 2, NUM_FRACTION_DIGITS),
          '" width="',
          toFixed(this._getLineWidth(this.ctx, i), NUM_FRACTION_DIGITS),
          '" height="',
          toFixed(this._getHeightOfLine(this.ctx, i) / this.lineHeight, NUM_FRACTION_DIGITS),
        '"></rect>\n');
    },

    _setSVGBg: function(textBgRects) {
      if (this.backgroundColor) {
        textBgRects.push(
          '\t\t<rect ',
            this._getFillAttributes(this.backgroundColor),
            ' x="',
            toFixed(-this.width / 2, NUM_FRACTION_DIGITS),
            '" y="',
            toFixed(-this.height / 2, NUM_FRACTION_DIGITS),
            '" width="',
            toFixed(this.width, NUM_FRACTION_DIGITS),
            '" height="',
            toFixed(this.height, NUM_FRACTION_DIGITS),
          '"></rect>\n');
      }
    },

    /**
     * Adobe Illustrator (at least CS5) is unable to render rgba()-based fill values
     * we work around it by "moving" alpha channel into opacity attribute and setting fill's alpha to 1
     *
     * @private
     * @param {Any} value
     * @return {String}
     */
    _getFillAttributes: function(value) {
      var fillColor = (value && typeof value === 'string') ? new fabric.Color(value) : '';
      if (!fillColor || !fillColor.getSource() || fillColor.getAlpha() === 1) {
        return 'fill="' + value + '"';
      }
      return 'opacity="' + fillColor.getAlpha() + '" fill="' + fillColor.setAlpha(1).toRgb() + '"';
    },
    /* _TO_SVG_END_ */

    /**
     * Sets specified property to a specified value
     * @param {String} key
     * @param {Any} value
     * @return {fabric.Text} thisArg
     * @chainable
     */
    _set: function(key, value) {
      this.callSuper('_set', key, value);

      if (key in this._dimensionAffectingProps) {
        this._initDimensions();
        this.setCoords();
      }
    },

    /**
     * Returns complexity of an instance
     * @return {Number} complexity
     */
    complexity: function() {
      return 1;
    }
  });

  /* _FROM_SVG_START_ */
  /**
   * List of attribute names to account for when parsing SVG element (used by {@link fabric.Text.fromElement})
   * @static
   * @memberOf fabric.Text
   * @see: http://www.w3.org/TR/SVG/text.html#TextElement
   */
  fabric.Text.ATTRIBUTE_NAMES = fabric.SHARED_ATTRIBUTES.concat(
    'x y dx dy font-family font-style font-weight font-size text-decoration text-anchor'.split(' '));

  /**
   * Default SVG font size
   * @static
   * @memberOf fabric.Text
   */
  fabric.Text.DEFAULT_SVG_FONT_SIZE = 16;

  /**
   * Returns fabric.Text instance from an SVG element (<b>not yet implemented</b>)
   * @static
   * @memberOf fabric.Text
   * @param {SVGElement} element Element to parse
   * @param {Object} [options] Options object
   * @return {fabric.Text} Instance of fabric.Text
   */
  fabric.Text.fromElement = function(element, options) {
    if (!element) {
      return null;
    }

    var parsedAttributes = fabric.parseAttributes(element, fabric.Text.ATTRIBUTE_NAMES);
    options = fabric.util.object.extend((options ? fabric.util.object.clone(options) : { }), parsedAttributes);

    options.top = options.top || 0;
    options.left = options.left || 0;
    if ('dx' in parsedAttributes) {
      options.left += parsedAttributes.dx;
    }
    if ('dy' in parsedAttributes) {
      options.top += parsedAttributes.dy;
    }
    if (!('fontSize' in options)) {
      options.fontSize = fabric.Text.DEFAULT_SVG_FONT_SIZE;
    }

    if (!options.originX) {
      options.originX = 'left';
    }

    var textContent = '';

    // The XML is not properly parsed in IE9 so a workaround to get
    // textContent is through firstChild.data. Another workaround would be
    // to convert XML loaded from a file to be converted using DOMParser (same way loadSVGFromString() does)
    if (!('textContent' in element)) {
      if ('firstChild' in element && element.firstChild !== null) {
        if ('data' in element.firstChild && element.firstChild.data !== null) {
          textContent = element.firstChild.data;
        }
      }
    }
    else {
      textContent = element.textContent;
    }

    textContent = textContent.replace(/^\s+|\s+$|\n+/g, '').replace(/\s+/g, ' ');

    var text = new fabric.Text(textContent, options),
        textHeightScaleFactor = text.getHeight() / text.height,
        lineHeightDiff = (text.height + text.strokeWidth) * text.lineHeight - text.height,
        scaledDiff = lineHeightDiff * textHeightScaleFactor,
        textHeight = text.getHeight() + scaledDiff,
        offX = 0;
    /*
      Adjust positioning:
        x/y attributes in SVG correspond to the bottom-left corner of text bounding box
        top/left properties in Fabric correspond to center point of text bounding box
    */
    if (text.originX === 'left') {
      offX = text.getWidth() / 2;
    }
    if (text.originX === 'right') {
      offX = -text.getWidth() / 2;
    }
    text.set({
      left: text.getLeft() + offX,
      top: text.getTop() - textHeight / 2 + text.fontSize * (0.18 + text._fontSizeFraction) / text.lineHeight /* 0.3 is the old lineHeight */
    });

    return text;
  };
  /* _FROM_SVG_END_ */

  /**
   * Returns fabric.Text instance from an object representation
   * @static
   * @memberOf fabric.Text
   * @param {Object} object Object to create an instance from
   * @return {fabric.Text} Instance of fabric.Text
   */
  fabric.Text.fromObject = function(object) {
    return new fabric.Text(object.text, clone(object));
  };

  fabric.util.createAccessors(fabric.Text);

})(typeof exports !== 'undefined' ? exports : this);


(function() {

  var clone = fabric.util.object.clone;

  /**
   * IText class (introduced in <b>v1.4</b>) Events are also fired with "text:"
   * prefix when observing canvas.
   * @class fabric.IText
   * @extends fabric.Text
   * @mixes fabric.Observable
   *
   * @fires changed
   * @fires selection:changed
   * @fires editing:entered
   * @fires editing:exited
   *
   * @return {fabric.IText} thisArg
   * @see {@link fabric.IText#initialize} for constructor definition
   *
   * <p>Supported key combinations:</p>
   * <pre>
   *   Move cursor:                    left, right, up, down
   *   Select character:               shift + left, shift + right
   *   Select text vertically:         shift + up, shift + down
   *   Move cursor by word:            alt + left, alt + right
   *   Select words:                   shift + alt + left, shift + alt + right
   *   Move cursor to line start/end:  cmd + left, cmd + right or home, end
   *   Select till start/end of line:  cmd + shift + left, cmd + shift + right or shift + home, shift + end
   *   Jump to start/end of text:      cmd + up, cmd + down
   *   Select till start/end of text:  cmd + shift + up, cmd + shift + down or shift + pgUp, shift + pgDown
   *   Delete character:               backspace
   *   Delete word:                    alt + backspace
   *   Delete line:                    cmd + backspace
   *   Forward delete:                 delete
   *   Copy text:                      ctrl/cmd + c
   *   Paste text:                     ctrl/cmd + v
   *   Cut text:                       ctrl/cmd + x
   *   Select entire text:             ctrl/cmd + a
   *   Quit editing                    tab or esc
   * </pre>
   *
   * <p>Supported mouse/touch combination</p>
   * <pre>
   *   Position cursor:                click/touch
   *   Create selection:               click/touch & drag
   *   Create selection:               click & shift + click
   *   Select word:                    double click
   *   Select line:                    triple click
   * </pre>
   */
  fabric.IText = fabric.util.createClass(fabric.Text, fabric.Observable, /** @lends fabric.IText.prototype */ {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'i-text',

    /**
     * Index where text selection starts (or where cursor is when there is no selection)
     * @type Number
     * @default
     */
    selectionStart: 0,

    /**
     * Index where text selection ends
     * @type Number
     * @default
     */
    selectionEnd: 0,

    /**
     * Color of text selection
     * @type String
     * @default
     */
    selectionColor: 'rgba(17,119,255,0.3)',

    /**
     * Indicates whether text is in editing mode
     * @type Boolean
     * @default
     */
    isEditing: false,

    /**
     * Indicates whether a text can be edited
     * @type Boolean
     * @default
     */
    editable: true,

    /**
     * Border color of text object while it's in editing mode
     * @type String
     * @default
     */
    editingBorderColor: 'rgba(102,153,255,0.25)',

    /**
     * Width of cursor (in px)
     * @type Number
     * @default
     */
    cursorWidth: 2,

    /**
     * Color of default cursor (when not overwritten by character style)
     * @type String
     * @default
     */
    cursorColor: '#333',

    /**
     * Delay between cursor blink (in ms)
     * @type Number
     * @default
     */
    cursorDelay: 1000,

    /**
     * Duration of cursor fadein (in ms)
     * @type Number
     * @default
     */
    cursorDuration: 600,

    /**
     * Object containing character styles
     * (where top-level properties corresponds to line number and 2nd-level properties -- to char number in a line)
     * @type Object
     * @default
     */
    styles: null,

    /**
     * Indicates whether internal text char widths can be cached
     * @type Boolean
     * @default
     */
    caching: true,

    /**
     * @private
     */
    _reSpace: /\s|\n/,

    /**
     * @private
     */
    _currentCursorOpacity: 0,

    /**
     * @private
     */
    _selectionDirection: null,

    /**
     * @private
     */
    _abortCursorAnimation: false,

    /**
     * @private
     */
    __widthOfSpace: [ ],

    /**
     * Constructor
     * @param {String} text Text string
     * @param {Object} [options] Options object
     * @return {fabric.IText} thisArg
     */
    initialize: function(text, options) {
      this.styles = options ? (options.styles || { }) : { };
      this.callSuper('initialize', text, options);
      this.initBehavior();
    },

    /**
     * @private
     */
    _clearCache: function() {
      this.callSuper('_clearCache');
      this.__widthOfSpace = [ ];
    },

    /**
     * Returns true if object has no styling
     */
    isEmptyStyles: function() {
      if (!this.styles) {
        return true;
      }
      var obj = this.styles;

      for (var p1 in obj) {
        for (var p2 in obj[p1]) {
          /*jshint unused:false */
          for (var p3 in obj[p1][p2]) {
            return false;
          }
        }
      }
      return true;
    },

    /**
     * Sets selection start (left boundary of a selection)
     * @param {Number} index Index to set selection start to
     */
    setSelectionStart: function(index) {
      index = Math.max(index, 0);
      if (this.selectionStart !== index) {
        this.fire('selection:changed');
        this.canvas && this.canvas.fire('text:selection:changed', { target: this });
        this.selectionStart = index;
      }
      this._updateTextarea();
    },

    /**
     * Sets selection end (right boundary of a selection)
     * @param {Number} index Index to set selection end to
     */
    setSelectionEnd: function(index) {
      index = Math.min(index, this.text.length);
      if (this.selectionEnd !== index) {
        this.fire('selection:changed');
        this.canvas && this.canvas.fire('text:selection:changed', { target: this });
        this.selectionEnd = index;
      }
      this._updateTextarea();
    },

    /**
     * Gets style of a current selection/cursor (at the start position)
     * @param {Number} [startIndex] Start index to get styles at
     * @param {Number} [endIndex] End index to get styles at
     * @return {Object} styles Style object at a specified (or current) index
     */
    getSelectionStyles: function(startIndex, endIndex) {

      if (arguments.length === 2) {
        var styles = [ ];
        for (var i = startIndex; i < endIndex; i++) {
          styles.push(this.getSelectionStyles(i));
        }
        return styles;
      }

      var loc = this.get2DCursorLocation(startIndex),
          style = this._getStyleDeclaration(loc.lineIndex, loc.charIndex);

      return style || {};
    },

    /**
     * Sets style of a current selection
     * @param {Object} [styles] Styles object
     * @return {fabric.IText} thisArg
     * @chainable
     */
    setSelectionStyles: function(styles) {
      if (this.selectionStart === this.selectionEnd) {
        this._extendStyles(this.selectionStart, styles);
      }
      else {
        for (var i = this.selectionStart; i < this.selectionEnd; i++) {
          this._extendStyles(i, styles);
        }
      }
      /* not included in _extendStyles to avoid clearing cache more than once */
      this._forceClearCache = true;
      return this;
    },

    /**
     * @private
     */
    _extendStyles: function(index, styles) {
      var loc = this.get2DCursorLocation(index);

      if (!this._getLineStyle(loc.lineIndex)) {
        this._setLineStyle(loc.lineIndex, {});
      }

      if (!this._getStyleDeclaration(loc.lineIndex, loc.charIndex)) {
        this._setStyleDeclaration(loc.lineIndex, loc.charIndex, {});
      }

      fabric.util.object.extend(this._getStyleDeclaration(loc.lineIndex, loc.charIndex), styles);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _render: function(ctx) {
      this.oldWidth = this.width;
      this.oldHeight = this.height;
      this.callSuper('_render', ctx);
      this.ctx = ctx;
      this.isEditing && this.renderCursorOrSelection();
    },

    /**
     * Renders cursor or selection (depending on what exists)
     */
    renderCursorOrSelection: function() {
      if (!this.active || !this.isEditing) {
        return;
      }

      var chars = this.text.split(''),
          boundaries, ctx;

      if (this.canvas.contextTop) {
        ctx = this.canvas.contextTop;
        ctx.save();
        ctx.transform.apply(ctx, this.canvas.viewportTransform);
        this.transform(ctx);
        this.transformMatrix && ctx.transform.apply(ctx, this.transformMatrix);
        this._clearTextArea(ctx);
      }
      else {
        ctx = this.ctx;
        ctx.save();
      }
      if (this.selectionStart === this.selectionEnd) {
        boundaries = this._getCursorBoundaries(chars, 'cursor');
        this.renderCursor(boundaries, ctx);
      }
      else {
        boundaries = this._getCursorBoundaries(chars, 'selection');
        this.renderSelection(chars, boundaries, ctx);
      }

      ctx.restore();
    },

    _clearTextArea: function(ctx) {
      // we add 4 pixel, to be sure to do not leave any pixel out
      var width = this.oldWidth + 4, height = this.oldHeight + 4;
      ctx.clearRect(-width / 2, -height / 2, width, height);
    },
    /**
     * Returns 2d representation (lineIndex and charIndex) of cursor (or selection start)
     * @param {Number} [selectionStart] Optional index. When not given, current selectionStart is used.
     */
    get2DCursorLocation: function(selectionStart) {
      if (typeof selectionStart === 'undefined') {
        selectionStart = this.selectionStart;
      }
      var len = this._textLines.length;
      for (var i = 0; i < len; i++) {
        if (selectionStart <= this._textLines[i].length) {
          return {
            lineIndex: i,
            charIndex: selectionStart
          };
        }
        selectionStart -= this._textLines[i].length + 1;
      }
      return {
        lineIndex: i - 1,
        charIndex: this._textLines[i - 1].length < selectionStart ? this._textLines[i - 1].length : selectionStart
      };
    },

    /**
     * Returns complete style of char at the current cursor
     * @param {Number} lineIndex Line index
     * @param {Number} charIndex Char index
     * @return {Object} Character style
     */
    getCurrentCharStyle: function(lineIndex, charIndex) {
      var style = this._getStyleDeclaration(lineIndex, charIndex === 0 ? 0 : charIndex - 1);

      return {
        fontSize: style && style.fontSize || this.fontSize,
        fill: style && style.fill || this.fill,
        textBackgroundColor: style && style.textBackgroundColor || this.textBackgroundColor,
        textDecoration: style && style.textDecoration || this.textDecoration,
        fontFamily: style && style.fontFamily || this.fontFamily,
        fontWeight: style && style.fontWeight || this.fontWeight,
        fontStyle: style && style.fontStyle || this.fontStyle,
        stroke: style && style.stroke || this.stroke,
        strokeWidth: style && style.strokeWidth || this.strokeWidth
      };
    },

    /**
     * Returns fontSize of char at the current cursor
     * @param {Number} lineIndex Line index
     * @param {Number} charIndex Char index
     * @return {Number} Character font size
     */
    getCurrentCharFontSize: function(lineIndex, charIndex) {
      var style = this._getStyleDeclaration(lineIndex, charIndex === 0 ? 0 : charIndex - 1);
      return style && style.fontSize ? style.fontSize : this.fontSize;
    },

    /**
     * Returns color (fill) of char at the current cursor
     * @param {Number} lineIndex Line index
     * @param {Number} charIndex Char index
     * @return {String} Character color (fill)
     */
    getCurrentCharColor: function(lineIndex, charIndex) {
      var style = this._getStyleDeclaration(lineIndex, charIndex === 0 ? 0 : charIndex - 1);
      return style && style.fill ? style.fill : this.cursorColor;
    },

    /**
     * Returns cursor boundaries (left, top, leftOffset, topOffset)
     * @private
     * @param {Array} chars Array of characters
     * @param {String} typeOfBoundaries
     */
    _getCursorBoundaries: function(chars, typeOfBoundaries) {

      // left/top are left/top of entire text box
      // leftOffset/topOffset are offset from that left/top point of a text box

      var left = Math.round(this._getLeftOffset()),
          top = this._getTopOffset(),

          offsets = this._getCursorBoundariesOffsets(
                      chars, typeOfBoundaries);

      return {
        left: left,
        top: top,
        leftOffset: offsets.left + offsets.lineLeft,
        topOffset: offsets.top
      };
    },

    /**
     * @private
     */
    _getCursorBoundariesOffsets: function(chars, typeOfBoundaries) {

      var lineLeftOffset = 0,

          lineIndex = 0,
          charIndex = 0,
          topOffset = 0,
          leftOffset = 0;

      for (var i = 0; i < this.selectionStart; i++) {
        if (chars[i] === '\n') {
          leftOffset = 0;
          topOffset += this._getHeightOfLine(this.ctx, lineIndex);

          lineIndex++;
          charIndex = 0;
        }
        else {
          leftOffset += this._getWidthOfChar(this.ctx, chars[i], lineIndex, charIndex);
          charIndex++;
        }

        lineLeftOffset = this._getLineLeftOffset(this._getLineWidth(this.ctx, lineIndex));
      }
      if (typeOfBoundaries === 'cursor') {
        topOffset += (1 - this._fontSizeFraction) * this._getHeightOfLine(this.ctx, lineIndex) / this.lineHeight
          - this.getCurrentCharFontSize(lineIndex, charIndex) * (1 - this._fontSizeFraction);
      }

      return {
        top: topOffset,
        left: leftOffset,
        lineLeft: lineLeftOffset
      };
    },

    /**
     * Renders cursor
     * @param {Object} boundaries
     * @param {CanvasRenderingContext2D} ctx transformed context to draw on
     */
    renderCursor: function(boundaries, ctx) {

      var cursorLocation = this.get2DCursorLocation(),
          lineIndex = cursorLocation.lineIndex,
          charIndex = cursorLocation.charIndex,
          charHeight = this.getCurrentCharFontSize(lineIndex, charIndex),
          leftOffset = (lineIndex === 0 && charIndex === 0)
                    ? this._getLineLeftOffset(this._getLineWidth(ctx, lineIndex))
                    : boundaries.leftOffset,
          multiplier = this.scaleX * this.canvas.getZoom(),
          cursorWidth = this.cursorWidth / multiplier;

      ctx.fillStyle = this.getCurrentCharColor(lineIndex, charIndex);
      ctx.globalAlpha = this.__isMousedown ? 1 : this._currentCursorOpacity;

      ctx.fillRect(
        boundaries.left + leftOffset - cursorWidth/2,
        boundaries.top + boundaries.topOffset,
        cursorWidth,
        charHeight);
    },

    /**
     * Renders text selection
     * @param {Array} chars Array of characters
     * @param {Object} boundaries Object with left/top/leftOffset/topOffset
     * @param {CanvasRenderingContext2D} ctx transformed context to draw on
     */
    renderSelection: function(chars, boundaries, ctx) {

      ctx.fillStyle = this.selectionColor;

      var start = this.get2DCursorLocation(this.selectionStart),
          end = this.get2DCursorLocation(this.selectionEnd),
          startLine = start.lineIndex,
          endLine = end.lineIndex;
      for (var i = startLine; i <= endLine; i++) {
        var lineOffset = this._getLineLeftOffset(this._getLineWidth(ctx, i)) || 0,
            lineHeight = this._getHeightOfLine(this.ctx, i),
            boxWidth = 0, line = this._textLines[i];

        if (i === startLine) {
          for (var j = 0, len = line.length; j < len; j++) {
            if (j >= start.charIndex && (i !== endLine || j < end.charIndex)) {
              boxWidth += this._getWidthOfChar(ctx, line[j], i, j);
            }
            if (j < start.charIndex) {
              lineOffset += this._getWidthOfChar(ctx, line[j], i, j);
            }
          }
        }
        else if (i > startLine && i < endLine) {
          boxWidth += this._getLineWidth(ctx, i) || 5;
        }
        else if (i === endLine) {
          for (var j2 = 0, j2len = end.charIndex; j2 < j2len; j2++) {
            boxWidth += this._getWidthOfChar(ctx, line[j2], i, j2);
          }
        }
        ctx.fillRect(
          boundaries.left + lineOffset,
          boundaries.top + boundaries.topOffset,
          boxWidth,
          lineHeight);

        boundaries.topOffset += lineHeight;
      }
    },

    /**
     * @private
     * @param {String} method
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderChars: function(method, ctx, line, left, top, lineIndex, charOffset) {

      if (this.isEmptyStyles()) {
        return this._renderCharsFast(method, ctx, line, left, top);
      }

      charOffset = charOffset || 0;
      this.skipTextAlign = true;

      // set proper box offset
      left -= this.textAlign === 'center'
        ? (this.width / 2)
        : (this.textAlign === 'right')
          ? this.width
          : 0;

      // set proper line offset
      var lineHeight = this._getHeightOfLine(ctx, lineIndex),
          lineLeftOffset = this._getLineLeftOffset(this._getLineWidth(ctx, lineIndex)),
          prevStyle,
          thisStyle,
          charsToRender = '';

      left += lineLeftOffset || 0;

      ctx.save();
      top -= lineHeight / this.lineHeight * this._fontSizeFraction;
      for (var i = charOffset, len = line.length + charOffset; i <= len; i++) {
        prevStyle = prevStyle || this.getCurrentCharStyle(lineIndex, i);
        thisStyle = this.getCurrentCharStyle(lineIndex, i + 1);

        if (this._hasStyleChanged(prevStyle, thisStyle) || i === len) {
          this._renderChar(method, ctx, lineIndex, i - 1, charsToRender, left, top, lineHeight);
          charsToRender = '';
          prevStyle = thisStyle;
        }
        charsToRender += line[i - charOffset];
      }
      ctx.restore();
    },

    /**
     * @private
     * @param {String} method
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {String} line Content of the line
     * @param {Number} left Left coordinate
     * @param {Number} top Top coordinate
     */
    _renderCharsFast: function(method, ctx, line, left, top) {
      this.skipTextAlign = false;

      if (method === 'fillText' && this.fill) {
        this.callSuper('_renderChars', method, ctx, line, left, top);
      }
      if (method === 'strokeText' && ((this.stroke && this.strokeWidth > 0) || this.skipFillStrokeCheck)) {
        this.callSuper('_renderChars', method, ctx, line, left, top);
      }
    },

    /**
     * @private
     * @param {String} method
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex
     * @param {Number} i
     * @param {String} _char
     * @param {Number} left Left coordinate
     * @param {Number} top Top coordinate
     * @param {Number} lineHeight Height of the line
     */
    _renderChar: function(method, ctx, lineIndex, i, _char, left, top, lineHeight) {
      var charWidth, charHeight, shouldFill, shouldStroke,
          decl = this._getStyleDeclaration(lineIndex, i),
          offset, textDecoration;

      if (decl) {
        charHeight = this._getHeightOfChar(ctx, _char, lineIndex, i);
        shouldStroke = decl.stroke;
        shouldFill = decl.fill;
        textDecoration = decl.textDecoration;
      }
      else {
        charHeight = this.fontSize;
      }

      shouldStroke = (shouldStroke || this.stroke) && method === 'strokeText';
      shouldFill = (shouldFill || this.fill) && method === 'fillText';

      decl && ctx.save();

      charWidth = this._applyCharStylesGetWidth(ctx, _char, lineIndex, i, decl || {});
      textDecoration = textDecoration || this.textDecoration;

      if (decl && decl.textBackgroundColor) {
        this._removeShadow(ctx);
      }
      shouldFill && ctx.fillText(_char, left, top);
      shouldStroke && ctx.strokeText(_char, left, top);

      if (textDecoration || textDecoration !== '') {
        offset = this._fontSizeFraction * lineHeight / this.lineHeight;
        this._renderCharDecoration(ctx, textDecoration, left, top, offset, charWidth, charHeight);
      }

      decl && ctx.restore();
      ctx.translate(charWidth, 0);
    },

    /**
     * @private
     * @param {Object} prevStyle
     * @param {Object} thisStyle
     */
    _hasStyleChanged: function(prevStyle, thisStyle) {
      return (prevStyle.fill !== thisStyle.fill ||
              prevStyle.fontSize !== thisStyle.fontSize ||
              prevStyle.textBackgroundColor !== thisStyle.textBackgroundColor ||
              prevStyle.textDecoration !== thisStyle.textDecoration ||
              prevStyle.fontFamily !== thisStyle.fontFamily ||
              prevStyle.fontWeight !== thisStyle.fontWeight ||
              prevStyle.fontStyle !== thisStyle.fontStyle ||
              prevStyle.stroke !== thisStyle.stroke ||
              prevStyle.strokeWidth !== thisStyle.strokeWidth
      );
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderCharDecoration: function(ctx, textDecoration, left, top, offset, charWidth, charHeight) {

      if (!textDecoration) {
        return;
      }

      var decorationWeight = charHeight / 15,
          positions = {
            underline: top + charHeight / 10,
            'line-through': top - charHeight * (this._fontSizeFraction + this._fontSizeMult - 1) + decorationWeight,
            overline: top - (this._fontSizeMult - this._fontSizeFraction) * charHeight
          },
          decorations = ['underline', 'line-through', 'overline'], i, decoration;

      for (i = 0; i < decorations.length; i++) {
        decoration = decorations[i];
        if (textDecoration.indexOf(decoration) > -1) {
          ctx.fillRect(left, positions[decoration], charWidth , decorationWeight);
        }
      }
    },

    /**
     * @private
     * @param {String} method
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {String} line
     */
    _renderTextLine: function(method, ctx, line, left, top, lineIndex) {
      // to "cancel" this.fontSize subtraction in fabric.Text#_renderTextLine
      // the adding 0.03 is just to align text with itext by overlap test
      if (!this.isEmptyStyles()) {
        top += this.fontSize * (this._fontSizeFraction + 0.03);
      }
      this.callSuper('_renderTextLine', method, ctx, line, left, top, lineIndex);
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextDecoration: function(ctx) {
      if (this.isEmptyStyles()) {
        return this.callSuper('_renderTextDecoration', ctx);
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _renderTextLinesBackground: function(ctx) {
      this.callSuper('_renderTextLinesBackground', ctx);

      var lineTopOffset = 0, heightOfLine,
          lineWidth, lineLeftOffset,
          leftOffset = this._getLeftOffset(),
          topOffset = this._getTopOffset(),
          line, _char, style;

      for (var i = 0, len = this._textLines.length; i < len; i++) {
        heightOfLine = this._getHeightOfLine(ctx, i);
        line = this._textLines[i];

        if (line === '' || !this.styles || !this._getLineStyle(i)) {
          lineTopOffset += heightOfLine;
          continue;
        }

        lineWidth = this._getLineWidth(ctx, i);
        lineLeftOffset = this._getLineLeftOffset(lineWidth);

        for (var j = 0, jlen = line.length; j < jlen; j++) {
          style = this._getStyleDeclaration(i, j);
          if (!style || !style.textBackgroundColor) {
            continue;
          }
          _char = line[j];

          ctx.fillStyle = style.textBackgroundColor;

          ctx.fillRect(
            leftOffset + lineLeftOffset + this._getWidthOfCharsAt(ctx, i, j),
            topOffset + lineTopOffset,
            this._getWidthOfChar(ctx, _char, i, j) + 1,
            heightOfLine / this.lineHeight
          );
        }
        lineTopOffset += heightOfLine;
      }
    },

    /**
     * @private
     */
    _getCacheProp: function(_char, styleDeclaration) {
      return _char +
             styleDeclaration.fontSize +
             styleDeclaration.fontWeight +
             styleDeclaration.fontStyle;
    },

    /**
     * @private
     * @param {String} fontFamily name
     * @return {Object} reference to cache
     */
    _getFontCache: function(fontFamily) {
      if (!fabric.charWidthsCache[fontFamily]) {
        fabric.charWidthsCache[fontFamily] = { };
      }
      return fabric.charWidthsCache[fontFamily];
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {String} _char
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @param {Object} [decl]
     */
    _applyCharStylesGetWidth: function(ctx, _char, lineIndex, charIndex, decl) {
      var charDecl = this._getStyleDeclaration(lineIndex, charIndex),
          styleDeclaration = (decl && clone(decl)) || clone(charDecl),
          width, cacheProp, charWidthsCache;

      this._applyFontStyles(styleDeclaration);
      charWidthsCache = this._getFontCache(styleDeclaration.fontFamily);
      cacheProp = this._getCacheProp(_char, styleDeclaration);

      // short-circuit if no styles for this char
      // global style from object is always applyed and handled by save and restore
      if (!charDecl && charWidthsCache[cacheProp] && this.caching) {
        return charWidthsCache[cacheProp];
      }

      if (typeof styleDeclaration.shadow === 'string') {
        styleDeclaration.shadow = new fabric.Shadow(styleDeclaration.shadow);
      }

      var fill = styleDeclaration.fill || this.fill;
      ctx.fillStyle = fill.toLive
        ? fill.toLive(ctx, this)
        : fill;

      if (styleDeclaration.stroke) {
        ctx.strokeStyle = (styleDeclaration.stroke && styleDeclaration.stroke.toLive)
          ? styleDeclaration.stroke.toLive(ctx, this)
          : styleDeclaration.stroke;
      }

      ctx.lineWidth = styleDeclaration.strokeWidth || this.strokeWidth;
      ctx.font = this._getFontDeclaration.call(styleDeclaration);

      //if we want this._setShadow.call to work with styleDeclarion
      //we have to add those references
      if (styleDeclaration.shadow) {
        styleDeclaration.scaleX = this.scaleX;
        styleDeclaration.scaleY = this.scaleY;
        styleDeclaration.canvas = this.canvas;
        this._setShadow.call(styleDeclaration, ctx);
      }

      if (!this.caching || !charWidthsCache[cacheProp]) {
        width = ctx.measureText(_char).width;
        this.caching && (charWidthsCache[cacheProp] = width);
        return width;
      }

      return charWidthsCache[cacheProp];
    },

    /**
     * @private
     * @param {Object} styleDeclaration
     */
    _applyFontStyles: function(styleDeclaration) {
      if (!styleDeclaration.fontFamily) {
        styleDeclaration.fontFamily = this.fontFamily;
      }
      if (!styleDeclaration.fontSize) {
        styleDeclaration.fontSize = this.fontSize;
      }
      if (!styleDeclaration.fontWeight) {
        styleDeclaration.fontWeight = this.fontWeight;
      }
      if (!styleDeclaration.fontStyle) {
        styleDeclaration.fontStyle = this.fontStyle;
      }
    },

    /**
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @param {Boolean} [returnCloneOrEmpty=false]
     * @private
     */
    _getStyleDeclaration: function(lineIndex, charIndex, returnCloneOrEmpty) {
      if (returnCloneOrEmpty) {
        return (this.styles[lineIndex] && this.styles[lineIndex][charIndex])
          ? clone(this.styles[lineIndex][charIndex])
          : { };
      }

      return this.styles[lineIndex] && this.styles[lineIndex][charIndex] ? this.styles[lineIndex][charIndex] : null;
    },

    /**
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @param {Object} style
     * @private
     */
    _setStyleDeclaration: function(lineIndex, charIndex, style) {
      this.styles[lineIndex][charIndex] = style;
    },

    /**
     *
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @private
     */
    _deleteStyleDeclaration: function(lineIndex, charIndex) {
      delete this.styles[lineIndex][charIndex];
    },

    /**
     * @param {Number} lineIndex
     * @private
     */
    _getLineStyle: function(lineIndex) {
      return this.styles[lineIndex];
    },

    /**
     * @param {Number} lineIndex
     * @param {Object} style
     * @private
     */
    _setLineStyle: function(lineIndex, style) {
      this.styles[lineIndex] = style;
    },

    /**
     * @param {Number} lineIndex
     * @private
     */
    _deleteLineStyle: function(lineIndex) {
      delete this.styles[lineIndex];
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _getWidthOfChar: function(ctx, _char, lineIndex, charIndex) {
      if (!this._isMeasuring && this.textAlign === 'justify' && this._reSpacesAndTabs.test(_char)) {
        return this._getWidthOfSpace(ctx, lineIndex);
      }
      var charWidthsCache, cacheProp,
          styleDeclaration = this._getStyleDeclaration(lineIndex, charIndex, true);
      this._applyFontStyles(styleDeclaration);
      charWidthsCache = this._getFontCache(styleDeclaration.fontFamily);
      cacheProp = this._getCacheProp(_char, styleDeclaration);

      if (charWidthsCache[cacheProp] && this.caching) {
        return charWidthsCache[cacheProp];
      }
      else if (ctx) {
        ctx.save();
        var width = this._applyCharStylesGetWidth(ctx, _char, lineIndex, charIndex);
        ctx.restore();
        return width;
      }
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _getHeightOfChar: function(ctx, lineIndex, charIndex) {
      var style = this._getStyleDeclaration(lineIndex, charIndex);
      return style && style.fontSize ? style.fontSize : this.fontSize;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex
     * @param {Number} charIndex
     */
    _getWidthOfCharsAt: function(ctx, lineIndex, charIndex) {
      var width = 0, i, _char;
      for (i = 0; i < charIndex; i++) {
        _char = this._textLines[lineIndex][i];
        width += this._getWidthOfChar(ctx, _char, lineIndex, i);
      }
      return width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex line number
     * @return {Number} Line width
     */
    _measureLine: function(ctx, lineIndex) {
      this._isMeasuring = true;
      var width = this._getWidthOfCharsAt(ctx, lineIndex, this._textLines[lineIndex].length);
      this._isMeasuring = false;
      return width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} lineIndex
     */
    _getWidthOfSpace: function (ctx, lineIndex) {
      if (this.__widthOfSpace[lineIndex]) {
        return this.__widthOfSpace[lineIndex];
      }
      var line = this._textLines[lineIndex],
          wordsWidth = this._getWidthOfWords(ctx, line, lineIndex, 0),
          widthDiff = this.width - wordsWidth,
          numSpaces = line.length - line.replace(this._reSpacesAndTabs, '').length,
          width = Math.max(widthDiff / numSpaces, ctx.measureText(' ').width);
      this.__widthOfSpace[lineIndex] = width;
      return width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     * @param {Number} line
     * @param {Number} lineIndex
     */
    _getWidthOfWords: function (ctx, line, lineIndex, charOffset) {
      var width = 0;

      for (var charIndex = 0; charIndex < line.length; charIndex++) {
        var _char = line[charIndex];

        if (!_char.match(/\s/)) {
          width += this._getWidthOfChar(ctx, _char, lineIndex, charIndex + charOffset);
        }
      }

      return width;
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _getHeightOfLine: function(ctx, lineIndex) {
      if (this.__lineHeights[lineIndex]) {
        return this.__lineHeights[lineIndex];
      }

      var line = this._textLines[lineIndex],
          maxHeight = this._getHeightOfChar(ctx, lineIndex, 0);

      for (var i = 1, len = line.length; i < len; i++) {
        var currentCharHeight = this._getHeightOfChar(ctx, lineIndex, i);
        if (currentCharHeight > maxHeight) {
          maxHeight = currentCharHeight;
        }
      }
      this.__lineHeights[lineIndex] = maxHeight * this.lineHeight * this._fontSizeMult;
      return this.__lineHeights[lineIndex];
    },

    /**
     * @private
     * @param {CanvasRenderingContext2D} ctx Context to render on
     */
    _getTextHeight: function(ctx) {
      var height = this._getHeightOfLine(ctx, 0) / this.lineHeight;
      for (var i = 1, len = this._textLines.length; i < len; i++) {
        height += this._getHeightOfLine(ctx, i);
      }
      return height;
    },

    /**
     * Returns object representation of an instance
     * @method toObject
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      var clonedStyles = { }, i, j, row;
      for (i in this.styles) {
        row = this.styles[i];
        clonedStyles[i] = { };
        for (j in row) {
          clonedStyles[i][j] = clone(row[j]);
        }
      }
      return fabric.util.object.extend(this.callSuper('toObject', propertiesToInclude), {
        styles: clonedStyles
      });
    }
  });

  /**
   * Returns fabric.IText instance from an object representation
   * @static
   * @memberOf fabric.IText
   * @param {Object} object Object to create an instance from
   * @return {fabric.IText} instance of fabric.IText
   */
  fabric.IText.fromObject = function(object) {
    return new fabric.IText(object.text, clone(object));
  };
})();


(function() {

  var clone = fabric.util.object.clone;

  fabric.util.object.extend(fabric.IText.prototype, /** @lends fabric.IText.prototype */ {

    /**
     * Initializes all the interactive behavior of IText
     */
    initBehavior: function() {
      this.initAddedHandler();
      this.initRemovedHandler();
      this.initCursorSelectionHandlers();
      this.initDoubleClickSimulation();
    },

    /**
     * Initializes "selected" event handler
     */
    initSelectedHandler: function() {
      this.on('selected', function() {

        var _this = this;
        setTimeout(function() {
          _this.selected = true;
        }, 100);
      });
    },

    /**
     * Initializes "added" event handler
     */
    initAddedHandler: function() {
      var _this = this;
      this.on('added', function() {
        if (this.canvas && !this.canvas._hasITextHandlers) {
          this.canvas._hasITextHandlers = true;
          this._initCanvasHandlers();
        }

        // Track IText instances per-canvas. Only register in this array once added
        // to a canvas; we don't want to leak a reference to the instance forever
        // simply because it existed at some point.
        // (Might be added to a collection, but not on a canvas.)
        if (_this.canvas) {
          _this.canvas._iTextInstances = _this.canvas._iTextInstances || [];
          _this.canvas._iTextInstances.push(_this);
        }
      });
    },

    initRemovedHandler: function() {
      var _this = this;
      this.on('removed', function() {
        // (Might be removed from a collection, but not on a canvas.)
        if (_this.canvas) {
          _this.canvas._iTextInstances = _this.canvas._iTextInstances || [];
          fabric.util.removeFromArray(_this.canvas._iTextInstances, _this);
        }
      });
    },

    /**
     * @private
     */
    _initCanvasHandlers: function() {
      var _this = this;

      this.canvas.on('selection:cleared', function() {
        fabric.IText.prototype.exitEditingOnOthers(_this.canvas);
      });

      this.canvas.on('mouse:up', function() {
        if (_this.canvas._iTextInstances) {
          _this.canvas._iTextInstances.forEach(function(obj) {
            obj.__isMousedown = false;
          });
        }
      });

      this.canvas.on('object:selected', function() {
        fabric.IText.prototype.exitEditingOnOthers(_this.canvas);
      });
    },

    /**
     * @private
     */
    _tick: function() {
      this._currentTickState = this._animateCursor(this, 1, this.cursorDuration, '_onTickComplete');
    },

    /**
     * @private
     */
    _animateCursor: function(obj, targetOpacity, duration, completeMethod) {

      var tickState;

      tickState = {
        isAborted: false,
        abort: function() {
          this.isAborted = true;
        },
      };

      obj.animate('_currentCursorOpacity', targetOpacity, {
        duration: duration,
        onComplete: function() {
          if (!tickState.isAborted) {
            obj[completeMethod]();
          }
        },
        onChange: function() {
          // we do not want to animate a selection, only cursor
          if (obj.canvas && obj.selectionStart === obj.selectionEnd) {
            obj.renderCursorOrSelection();
          }
        },
        abort: function() {
          return tickState.isAborted;
        }
      });
      return tickState;
    },

    /**
     * @private
     */
    _onTickComplete: function() {

      var _this = this;

      if (this._cursorTimeout1) {
        clearTimeout(this._cursorTimeout1);
      }
      this._cursorTimeout1 = setTimeout(function() {
        _this._currentTickCompleteState = _this._animateCursor(_this, 0, this.cursorDuration / 2, '_tick');
      }, 100);
    },

    /**
     * Initializes delayed cursor
     */
    initDelayedCursor: function(restart) {
      var _this = this,
          delay = restart ? 0 : this.cursorDelay;

      this.abortCursorAnimation();
      this._currentCursorOpacity = 1;
      this._cursorTimeout2 = setTimeout(function() {
        _this._tick();
      }, delay);
    },

    /**
     * Aborts cursor animation and clears all timeouts
     */
    abortCursorAnimation: function() {
      this._currentTickState && this._currentTickState.abort();
      this._currentTickCompleteState && this._currentTickCompleteState.abort();

      clearTimeout(this._cursorTimeout1);
      clearTimeout(this._cursorTimeout2);

      this._currentCursorOpacity = 0;
      // to clear just itext area we need to transform the context
      // it may not be worth it
      this.canvas && this.canvas.clearContext(this.canvas.contextTop || this.ctx);
    },

    /**
     * Selects entire text
     */
    selectAll: function() {
      this.setSelectionStart(0);
      this.setSelectionEnd(this.text.length);
    },

    /**
     * Returns selected text
     * @return {String}
     */
    getSelectedText: function() {
      return this.text.slice(this.selectionStart, this.selectionEnd);
    },

    /**
     * Find new selection index representing start of current word according to current selection index
     * @param {Number} startFrom Surrent selection index
     * @return {Number} New selection index
     */
    findWordBoundaryLeft: function(startFrom) {
      var offset = 0, index = startFrom - 1;

      // remove space before cursor first
      if (this._reSpace.test(this.text.charAt(index))) {
        while (this._reSpace.test(this.text.charAt(index))) {
          offset++;
          index--;
        }
      }
      while (/\S/.test(this.text.charAt(index)) && index > -1) {
        offset++;
        index--;
      }

      return startFrom - offset;
    },

    /**
     * Find new selection index representing end of current word according to current selection index
     * @param {Number} startFrom Current selection index
     * @return {Number} New selection index
     */
    findWordBoundaryRight: function(startFrom) {
      var offset = 0, index = startFrom;

      // remove space after cursor first
      if (this._reSpace.test(this.text.charAt(index))) {
        while (this._reSpace.test(this.text.charAt(index))) {
          offset++;
          index++;
        }
      }
      while (/\S/.test(this.text.charAt(index)) && index < this.text.length) {
        offset++;
        index++;
      }

      return startFrom + offset;
    },

    /**
     * Find new selection index representing start of current line according to current selection index
     * @param {Number} startFrom Current selection index
     * @return {Number} New selection index
     */
    findLineBoundaryLeft: function(startFrom) {
      var offset = 0, index = startFrom - 1;

      while (!/\n/.test(this.text.charAt(index)) && index > -1) {
        offset++;
        index--;
      }

      return startFrom - offset;
    },

    /**
     * Find new selection index representing end of current line according to current selection index
     * @param {Number} startFrom Current selection index
     * @return {Number} New selection index
     */
    findLineBoundaryRight: function(startFrom) {
      var offset = 0, index = startFrom;

      while (!/\n/.test(this.text.charAt(index)) && index < this.text.length) {
        offset++;
        index++;
      }

      return startFrom + offset;
    },

    /**
     * Returns number of newlines in selected text
     * @return {Number} Number of newlines in selected text
     */
    getNumNewLinesInSelectedText: function() {
      var selectedText = this.getSelectedText(),
          numNewLines  = 0;

      for (var i = 0, len = selectedText.length; i < len; i++) {
        if (selectedText[i] === '\n') {
          numNewLines++;
        }
      }
      return numNewLines;
    },

    /**
     * Finds index corresponding to beginning or end of a word
     * @param {Number} selectionStart Index of a character
     * @param {Number} direction 1 or -1
     * @return {Number} Index of the beginning or end of a word
     */
    searchWordBoundary: function(selectionStart, direction) {
      var index     = this._reSpace.test(this.text.charAt(selectionStart)) ? selectionStart - 1 : selectionStart,
          _char     = this.text.charAt(index),
          reNonWord = /[ \n\.,;!\?\-]/;

      while (!reNonWord.test(_char) && index > 0 && index < this.text.length) {
        index += direction;
        _char = this.text.charAt(index);
      }
      if (reNonWord.test(_char) && _char !== '\n') {
        index += direction === 1 ? 0 : 1;
      }
      return index;
    },

    /**
     * Selects a word based on the index
     * @param {Number} selectionStart Index of a character
     */
    selectWord: function(selectionStart) {
      var newSelectionStart = this.searchWordBoundary(selectionStart, -1), /* search backwards */
          newSelectionEnd   = this.searchWordBoundary(selectionStart, 1);
      /* search forward */

      this.setSelectionStart(newSelectionStart);
      this.setSelectionEnd(newSelectionEnd);
    },

    /**
     * Selects a line based on the index
     * @param {Number} selectionStart Index of a character
     */
    selectLine: function(selectionStart) {
      var newSelectionStart = this.findLineBoundaryLeft(selectionStart),
          newSelectionEnd   = this.findLineBoundaryRight(selectionStart);

      this.setSelectionStart(newSelectionStart);
      this.setSelectionEnd(newSelectionEnd);
    },

    /**
     * Enters editing state
     * @return {fabric.IText} thisArg
     * @chainable
     */
    enterEditing: function(e) {
      if (this.isEditing || !this.editable) {
        return;
      }

      if (this.canvas) {
        this.exitEditingOnOthers(this.canvas);
      }

      this.isEditing = true;

      this.initHiddenTextarea(e);
      this.hiddenTextarea.focus();
      this._updateTextarea();
      this._saveEditingProps();
      this._setEditingProps();
      this._textBeforeEdit = this.text;

      this._tick();
      this.fire('editing:entered');

      if (!this.canvas) {
        return this;
      }

      this.canvas.renderAll();
      this.canvas.fire('text:editing:entered', { target: this });
      this.initMouseMoveHandler();
      return this;
    },

    exitEditingOnOthers: function(canvas) {
      if (canvas._iTextInstances) {
        canvas._iTextInstances.forEach(function(obj) {
          obj.selected = false;
          if (obj.isEditing) {
            obj.exitEditing();
          }
        });
      }
    },

    /**
     * Initializes "mousemove" event handler
     */
    initMouseMoveHandler: function() {
      this.canvas.on('mouse:move', this.mouseMoveHandler.bind(this));
    },

    /**
     * @private
     */
    mouseMoveHandler: function(options) {
      if (!this.__isMousedown || !this.isEditing) {
        return;
      }

      var newSelectionStart = this.getSelectionStartFromPointer(options.e);
      if (newSelectionStart === this.__selectionStartOnMouseDown) {
        return;
      }
      if (newSelectionStart > this.__selectionStartOnMouseDown) {
        this.setSelectionStart(this.__selectionStartOnMouseDown);
        this.setSelectionEnd(newSelectionStart);
      }
      else {
        this.setSelectionStart(newSelectionStart);
        this.setSelectionEnd(this.__selectionStartOnMouseDown);
      }
      this.renderCursorOrSelection();
    },

    /**
     * @private
     */
    _setEditingProps: function() {
      this.hoverCursor = 'text';

      if (this.canvas) {
        this.canvas.defaultCursor = this.canvas.moveCursor = 'text';
      }

      this.borderColor = this.editingBorderColor;

      this.hasControls = this.selectable = false;
      this.lockMovementX = this.lockMovementY = true;
    },

    /**
     * @private
     */
    _updateTextarea: function() {
      if (!this.hiddenTextarea || this.inCompositionMode) {
        return;
      }
      this.hiddenTextarea.value = this.text;
      this.hiddenTextarea.selectionStart = this.selectionStart;
      this.hiddenTextarea.selectionEnd = this.selectionEnd;
      if (this.selectionStart === this.selectionEnd) {
        var style = this._calcTextareaPosition();
        this.hiddenTextarea.style.left = style.left;
        this.hiddenTextarea.style.top = style.top;
        this.hiddenTextarea.style.fontSize = style.fontSize;
      }
    },

    /**
     * @private
     * @return {Object} style contains style for hiddenTextarea
     */
    _calcTextareaPosition: function() {
      if (!this.canvas) {
        return { x: 1, y: 1 };
      }
      var chars = this.text.split(''),
          boundaries = this._getCursorBoundaries(chars, 'cursor'),
          cursorLocation = this.get2DCursorLocation(),
          lineIndex = cursorLocation.lineIndex,
          charIndex = cursorLocation.charIndex,
          charHeight = this.getCurrentCharFontSize(lineIndex, charIndex),
          leftOffset = (lineIndex === 0 && charIndex === 0)
                    ? this._getLineLeftOffset(this._getLineWidth(this.ctx, lineIndex))
                    : boundaries.leftOffset,
          m = this.calcTransformMatrix(),
          p = {
            x: boundaries.left + leftOffset,
            y: boundaries.top + boundaries.topOffset + charHeight
          },
          upperCanvas = this.canvas.upperCanvasEl,
          maxWidth = upperCanvas.width - charHeight,
          maxHeight = upperCanvas.height - charHeight;

      p = fabric.util.transformPoint(p, m);
      p = fabric.util.transformPoint(p, this.canvas.viewportTransform);

      if (p.x < 0) {
        p.x = 0;
      }
      if (p.x > maxWidth) {
        p.x = maxWidth;
      }
      if (p.y < 0) {
        p.y = 0;
      }
      if (p.y > maxHeight) {
        p.y = maxHeight;
      }

      return { left: p.x + 'px', top: p.y + 'px', fontSize: charHeight };
    },

    /**
     * @private
     */
    _saveEditingProps: function() {
      this._savedProps = {
        hasControls: this.hasControls,
        borderColor: this.borderColor,
        lockMovementX: this.lockMovementX,
        lockMovementY: this.lockMovementY,
        hoverCursor: this.hoverCursor,
        defaultCursor: this.canvas && this.canvas.defaultCursor,
        moveCursor: this.canvas && this.canvas.moveCursor
      };
    },

    /**
     * @private
     */
    _restoreEditingProps: function() {
      if (!this._savedProps) {
        return;
      }

      this.hoverCursor = this._savedProps.overCursor;
      this.hasControls = this._savedProps.hasControls;
      this.borderColor = this._savedProps.borderColor;
      this.lockMovementX = this._savedProps.lockMovementX;
      this.lockMovementY = this._savedProps.lockMovementY;

      if (this.canvas) {
        this.canvas.defaultCursor = this._savedProps.defaultCursor;
        this.canvas.moveCursor = this._savedProps.moveCursor;
      }
    },

    /**
     * Exits from editing state
     * @return {fabric.IText} thisArg
     * @chainable
     */
    exitEditing: function() {
      var isTextChanged = (this._textBeforeEdit !== this.text);
      this.selected = false;
      this.isEditing = false;
      this.selectable = true;

      this.selectionEnd = this.selectionStart;
      this.hiddenTextarea && this.canvas && this.hiddenTextarea.parentNode.removeChild(this.hiddenTextarea);
      this.hiddenTextarea = null;

      this.abortCursorAnimation();
      this._restoreEditingProps();
      this._currentCursorOpacity = 0;

      this.fire('editing:exited');
      isTextChanged && this.fire('modified');
      if (this.canvas) {
        this.canvas.off('mouse:move', this.mouseMoveHandler);
        this.canvas.fire('text:editing:exited', { target: this });
        isTextChanged && this.canvas.fire('object:modified', { target: this });
      }

      return this;
    },

    /**
     * @private
     */
    _removeExtraneousStyles: function() {
      for (var prop in this.styles) {
        if (!this._textLines[prop]) {
          delete this.styles[prop];
        }
      }
    },

    /**
     * @private
     */
    _removeCharsFromTo: function(start, end) {
      while (end !== start) {
        this._removeSingleCharAndStyle(start + 1);
        end--;
      }
      this.setSelectionStart(start);
    },

    _removeSingleCharAndStyle: function(index) {
      var isBeginningOfLine = this.text[index - 1] === '\n',
          indexStyle        = isBeginningOfLine ? index : index - 1;
      this.removeStyleObject(isBeginningOfLine, indexStyle);
      this.text = this.text.slice(0, index - 1) +
        this.text.slice(index);

      this._textLines = this._splitTextIntoLines();
    },

    /**
     * Inserts characters where cursor is (replacing selection if one exists)
     * @param {String} _chars Characters to insert
     * @param {Boolean} useCopiedStyle use fabric.copiedTextStyle
     */
    insertChars: function(_chars, useCopiedStyle) {
      var style;

      if (this.selectionEnd - this.selectionStart > 1) {
        this._removeCharsFromTo(this.selectionStart, this.selectionEnd);
        this.setSelectionEnd(this.selectionStart);
      }
      //short circuit for block paste
      if (!useCopiedStyle && this.isEmptyStyles()) {
        this.insertChar(_chars, false);
        return;
      }
      for (var i = 0, len = _chars.length; i < len; i++) {
        if (useCopiedStyle) {
          style = fabric.copiedTextStyle[i];
        }
        this.insertChar(_chars[i], i < len - 1, style);
      }
    },

    /**
     * Inserts a character where cursor is
     * @param {String} _char Characters to insert
     * @param {Boolean} skipUpdate trigger rendering and updates at the end of text insert
     * @param {Object} styleObject Style to be inserted for the new char
     */
    insertChar: function(_char, skipUpdate, styleObject) {
      var isEndOfLine = this.text[this.selectionStart] === '\n';
      this.text = this.text.slice(0, this.selectionStart) +
        _char + this.text.slice(this.selectionEnd);
      this._textLines = this._splitTextIntoLines();
      this.insertStyleObjects(_char, isEndOfLine, styleObject);
      this.selectionStart += _char.length;
      this.selectionEnd = this.selectionStart;
      if (skipUpdate) {
        return;
      }
      this._updateTextarea();
      this.canvas && this.canvas.renderAll();
      this.setCoords();
      this.fire('changed');
      this.canvas && this.canvas.fire('text:changed', { target: this });
    },

    /**
     * Inserts new style object
     * @param {Number} lineIndex Index of a line
     * @param {Number} charIndex Index of a char
     * @param {Boolean} isEndOfLine True if it's end of line
     */
    insertNewlineStyleObject: function(lineIndex, charIndex, isEndOfLine) {

      this.shiftLineStyles(lineIndex, +1);

      if (!this.styles[lineIndex + 1]) {
        this.styles[lineIndex + 1] = {};
      }

      var currentCharStyle = {},
          newLineStyles    = {};

      if (this.styles[lineIndex] && this.styles[lineIndex][charIndex - 1]) {
        currentCharStyle = this.styles[lineIndex][charIndex - 1];
      }

      // if there's nothing after cursor,
      // we clone current char style onto the next (otherwise empty) line
      if (isEndOfLine) {
        newLineStyles[0] = clone(currentCharStyle);
        this.styles[lineIndex + 1] = newLineStyles;
      }
      // otherwise we clone styles of all chars
      // after cursor onto the next line, from the beginning
      else {
        for (var index in this.styles[lineIndex]) {
          if (parseInt(index, 10) >= charIndex) {
            newLineStyles[parseInt(index, 10) - charIndex] = this.styles[lineIndex][index];
            // remove lines from the previous line since they're on a new line now
            delete this.styles[lineIndex][index];
          }
        }
        this.styles[lineIndex + 1] = newLineStyles;
      }
      this._forceClearCache = true;
    },

    /**
     * Inserts style object for a given line/char index
     * @param {Number} lineIndex Index of a line
     * @param {Number} charIndex Index of a char
     * @param {Object} [style] Style object to insert, if given
     */
    insertCharStyleObject: function(lineIndex, charIndex, style) {

      var currentLineStyles       = this.styles[lineIndex],
          currentLineStylesCloned = clone(currentLineStyles);

      if (charIndex === 0 && !style) {
        charIndex = 1;
      }

      // shift all char styles by 1 forward
      // 0,1,2,3 -> (charIndex=2) -> 0,1,3,4 -> (insert 2) -> 0,1,2,3,4
      for (var index in currentLineStylesCloned) {
        var numericIndex = parseInt(index, 10);

        if (numericIndex >= charIndex) {
          currentLineStyles[numericIndex + 1] = currentLineStylesCloned[numericIndex];

          // only delete the style if there was nothing moved there
          if (!currentLineStylesCloned[numericIndex - 1]) {
            delete currentLineStyles[numericIndex];
          }
        }
      }

      this.styles[lineIndex][charIndex] =
        style || clone(currentLineStyles[charIndex - 1]);
      this._forceClearCache = true;
    },

    /**
     * Inserts style object(s)
     * @param {String} _chars Characters at the location where style is inserted
     * @param {Boolean} isEndOfLine True if it's end of line
     * @param {Object} [styleObject] Style to insert
     */
    insertStyleObjects: function(_chars, isEndOfLine, styleObject) {
      // removed shortcircuit over isEmptyStyles

      var cursorLocation = this.get2DCursorLocation(),
          lineIndex      = cursorLocation.lineIndex,
          charIndex      = cursorLocation.charIndex;

      if (!this._getLineStyle(lineIndex)) {
        this._setLineStyle(lineIndex, {});
      }

      if (_chars === '\n') {
        this.insertNewlineStyleObject(lineIndex, charIndex, isEndOfLine);
      }
      else {
        this.insertCharStyleObject(lineIndex, charIndex, styleObject);
      }
    },

    /**
     * Shifts line styles up or down
     * @param {Number} lineIndex Index of a line
     * @param {Number} offset Can be -1 or +1
     */
    shiftLineStyles: function(lineIndex, offset) {
      // shift all line styles by 1 upward
      var clonedStyles = clone(this.styles);
      for (var line in this.styles) {
        var numericLine = parseInt(line, 10);
        if (numericLine > lineIndex) {
          this.styles[numericLine + offset] = clonedStyles[numericLine];
          if (!clonedStyles[numericLine - offset]) {
            delete this.styles[numericLine];
          }
        }
      }
      //TODO: evaluate if delete old style lines with offset -1
    },

    /**
     * Removes style object
     * @param {Boolean} isBeginningOfLine True if cursor is at the beginning of line
     * @param {Number} [index] Optional index. When not given, current selectionStart is used.
     */
    removeStyleObject: function(isBeginningOfLine, index) {

      var cursorLocation = this.get2DCursorLocation(index),
          lineIndex      = cursorLocation.lineIndex,
          charIndex      = cursorLocation.charIndex;

      this._removeStyleObject(isBeginningOfLine, cursorLocation, lineIndex, charIndex);
    },

    _getTextOnPreviousLine: function(lIndex) {
      return this._textLines[lIndex - 1];
    },

    _removeStyleObject: function(isBeginningOfLine, cursorLocation, lineIndex, charIndex) {

      if (isBeginningOfLine) {
        var textOnPreviousLine     = this._getTextOnPreviousLine(cursorLocation.lineIndex),
            newCharIndexOnPrevLine = textOnPreviousLine ? textOnPreviousLine.length : 0;

        if (!this.styles[lineIndex - 1]) {
          this.styles[lineIndex - 1] = {};
        }
        for (charIndex in this.styles[lineIndex]) {
          this.styles[lineIndex - 1][parseInt(charIndex, 10) + newCharIndexOnPrevLine]
            = this.styles[lineIndex][charIndex];
        }
        this.shiftLineStyles(cursorLocation.lineIndex, -1);
      }
      else {
        var currentLineStyles = this.styles[lineIndex];

        if (currentLineStyles) {
          delete currentLineStyles[charIndex];
        }
        var currentLineStylesCloned = clone(currentLineStyles);
        // shift all styles by 1 backwards
        for (var i in currentLineStylesCloned) {
          var numericIndex = parseInt(i, 10);
          if (numericIndex >= charIndex && numericIndex !== 0) {
            currentLineStyles[numericIndex - 1] = currentLineStylesCloned[numericIndex];
            delete currentLineStyles[numericIndex];
          }
        }
      }
    },

    /**
     * Inserts new line
     */
    insertNewline: function() {
      this.insertChars('\n');
    }
  });
})();


fabric.util.object.extend(fabric.IText.prototype, /** @lends fabric.IText.prototype */ {
  /**
   * Initializes "dbclick" event handler
   */
  initDoubleClickSimulation: function() {

    // for double click
    this.__lastClickTime = +new Date();

    // for triple click
    this.__lastLastClickTime = +new Date();

    this.__lastPointer = { };

    this.on('mousedown', this.onMouseDown.bind(this));
  },

  onMouseDown: function(options) {

    this.__newClickTime = +new Date();
    var newPointer = this.canvas.getPointer(options.e);

    if (this.isTripleClick(newPointer)) {
      this.fire('tripleclick', options);
      this._stopEvent(options.e);
    }
    else if (this.isDoubleClick(newPointer)) {
      this.fire('dblclick', options);
      this._stopEvent(options.e);
    }

    this.__lastLastClickTime = this.__lastClickTime;
    this.__lastClickTime = this.__newClickTime;
    this.__lastPointer = newPointer;
    this.__lastIsEditing = this.isEditing;
    this.__lastSelected = this.selected;
  },

  isDoubleClick: function(newPointer) {
    return this.__newClickTime - this.__lastClickTime < 500 &&
        this.__lastPointer.x === newPointer.x &&
        this.__lastPointer.y === newPointer.y && this.__lastIsEditing;
  },

  isTripleClick: function(newPointer) {
    return this.__newClickTime - this.__lastClickTime < 500 &&
        this.__lastClickTime - this.__lastLastClickTime < 500 &&
        this.__lastPointer.x === newPointer.x &&
        this.__lastPointer.y === newPointer.y;
  },

  /**
   * @private
   */
  _stopEvent: function(e) {
    e.preventDefault && e.preventDefault();
    e.stopPropagation && e.stopPropagation();
  },

  /**
   * Initializes event handlers related to cursor or selection
   */
  initCursorSelectionHandlers: function() {
    this.initSelectedHandler();
    this.initMousedownHandler();
    this.initMouseupHandler();
    this.initClicks();
  },

  /**
   * Initializes double and triple click event handlers
   */
  initClicks: function() {
    this.on('dblclick', function(options) {
      this.selectWord(this.getSelectionStartFromPointer(options.e));
    });
    this.on('tripleclick', function(options) {
      this.selectLine(this.getSelectionStartFromPointer(options.e));
    });
  },

  /**
   * Initializes "mousedown" event handler
   */
  initMousedownHandler: function() {
    this.on('mousedown', function(options) {
      if (!this.editable) {
        return;
      }
      var pointer = this.canvas.getPointer(options.e);

      this.__mousedownX = pointer.x;
      this.__mousedownY = pointer.y;
      this.__isMousedown = true;

      if (this.hiddenTextarea && this.canvas) {
        this.canvas.wrapperEl.appendChild(this.hiddenTextarea);
      }

      if (this.selected) {
        this.setCursorByClick(options.e);
      }

      if (this.isEditing) {
        this.__selectionStartOnMouseDown = this.selectionStart;
        this.initDelayedCursor(true);
      }
    });
  },

  /**
   * @private
   */
  _isObjectMoved: function(e) {
    var pointer = this.canvas.getPointer(e);

    return this.__mousedownX !== pointer.x ||
           this.__mousedownY !== pointer.y;
  },

  /**
   * Initializes "mouseup" event handler
   */
  initMouseupHandler: function() {
    this.on('mouseup', function(options) {
      this.__isMousedown = false;
      if (!this.editable || this._isObjectMoved(options.e)) {
        return;
      }

      if (this.__lastSelected && !this.__corner) {
        this.enterEditing(options.e);
        this.initDelayedCursor(true);
      }
      this.selected = true;
    });
  },

  /**
   * Changes cursor location in a text depending on passed pointer (x/y) object
   * @param {Event} e Event object
   */
  setCursorByClick: function(e) {
    var newSelectionStart = this.getSelectionStartFromPointer(e);

    if (e.shiftKey) {
      if (newSelectionStart < this.selectionStart) {
        this.setSelectionEnd(this.selectionStart);
        this.setSelectionStart(newSelectionStart);
      }
      else {
        this.setSelectionEnd(newSelectionStart);
      }
    }
    else {
      this.setSelectionStart(newSelectionStart);
      this.setSelectionEnd(newSelectionStart);
    }
  },

  /**
   * Returns index of a character corresponding to where an object was clicked
   * @param {Event} e Event object
   * @return {Number} Index of a character
   */
  getSelectionStartFromPointer: function(e) {
    var mouseOffset = this.getLocalPointer(e),
        prevWidth = 0,
        width = 0,
        height = 0,
        charIndex = 0,
        newSelectionStart,
        line;

    for (var i = 0, len = this._textLines.length; i < len; i++) {
      line = this._textLines[i];
      height += this._getHeightOfLine(this.ctx, i) * this.scaleY;

      var widthOfLine = this._getLineWidth(this.ctx, i),
          lineLeftOffset = this._getLineLeftOffset(widthOfLine);

      width = lineLeftOffset * this.scaleX;

      for (var j = 0, jlen = line.length; j < jlen; j++) {

        prevWidth = width;

        width += this._getWidthOfChar(this.ctx, line[j], i, this.flipX ? jlen - j : j) *
                 this.scaleX;

        if (height <= mouseOffset.y || width <= mouseOffset.x) {
          charIndex++;
          continue;
        }

        return this._getNewSelectionStartFromOffset(
          mouseOffset, prevWidth, width, charIndex + i, jlen);
      }

      if (mouseOffset.y < height) {
        //this happens just on end of lines.
        return this._getNewSelectionStartFromOffset(
          mouseOffset, prevWidth, width, charIndex + i - 1, jlen);
      }
    }

    // clicked somewhere after all chars, so set at the end
    if (typeof newSelectionStart === 'undefined') {
      return this.text.length;
    }
  },

  /**
   * @private
   */
  _getNewSelectionStartFromOffset: function(mouseOffset, prevWidth, width, index, jlen) {

    var distanceBtwLastCharAndCursor = mouseOffset.x - prevWidth,
        distanceBtwNextCharAndCursor = width - mouseOffset.x,
        offset = distanceBtwNextCharAndCursor > distanceBtwLastCharAndCursor ? 0 : 1,
        newSelectionStart = index + offset;

    // if object is horizontally flipped, mirror cursor location from the end
    if (this.flipX) {
      newSelectionStart = jlen - newSelectionStart;
    }

    if (newSelectionStart > this.text.length) {
      newSelectionStart = this.text.length;
    }

    return newSelectionStart;
  }
});


fabric.util.object.extend(fabric.IText.prototype, /** @lends fabric.IText.prototype */ {

  /**
   * Initializes hidden textarea (needed to bring up keyboard in iOS)
   */
  initHiddenTextarea: function() {
    this.hiddenTextarea = fabric.document.createElement('textarea');
    this.hiddenTextarea.setAttribute('autocapitalize', 'off');
    var style = this._calcTextareaPosition();
    this.hiddenTextarea.style.cssText = 'position: absolute; top: ' + style.top + '; left: ' + style.left + ';'
                                        + ' opacity: 0; width: 0px; height: 0px; z-index: -999;';
    if (this.canvas) {
      this.canvas.lowerCanvasEl.parentNode.appendChild(this.hiddenTextarea);
    }
    else {
      fabric.document.body.appendChild(this.hiddenTextarea);
    }

    fabric.util.addListener(this.hiddenTextarea, 'keydown', this.onKeyDown.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'keyup', this.onKeyUp.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'input', this.onInput.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'copy', this.copy.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'cut', this.cut.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'paste', this.paste.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'compositionstart', this.onCompositionStart.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'compositionupdate', this.onCompositionUpdate.bind(this));
    fabric.util.addListener(this.hiddenTextarea, 'compositionend', this.onCompositionEnd.bind(this));

    if (!this._clickHandlerInitialized && this.canvas) {
      fabric.util.addListener(this.canvas.upperCanvasEl, 'click', this.onClick.bind(this));
      this._clickHandlerInitialized = true;
    }
  },

  /**
   * @private
   */
  _keysMap: {
    8:  'removeChars',
    9:  'exitEditing',
    27: 'exitEditing',
    13: 'insertNewline',
    33: 'moveCursorUp',
    34: 'moveCursorDown',
    35: 'moveCursorRight',
    36: 'moveCursorLeft',
    37: 'moveCursorLeft',
    38: 'moveCursorUp',
    39: 'moveCursorRight',
    40: 'moveCursorDown',
    46: 'forwardDelete'
  },

  /**
   * @private
   */
  _ctrlKeysMapUp: {
    67: 'copy',
    88: 'cut'
  },

  /**
   * @private
   */
  _ctrlKeysMapDown: {
    65: 'selectAll'
  },

  onClick: function() {
    // No need to trigger click event here, focus is enough to have the keyboard appear on Android
    this.hiddenTextarea && this.hiddenTextarea.focus();
  },

  /**
   * Handles keyup event
   * @param {Event} e Event object
   */
  onKeyDown: function(e) {
    if (!this.isEditing) {
      return;
    }
    if (e.keyCode in this._keysMap) {
      this[this._keysMap[e.keyCode]](e);
    }
    else if ((e.keyCode in this._ctrlKeysMapDown) && (e.ctrlKey || e.metaKey)) {
      this[this._ctrlKeysMapDown[e.keyCode]](e);
    }
    else {
      return;
    }
    e.stopImmediatePropagation();
    e.preventDefault();
    this.canvas && this.canvas.renderAll();
  },

  /**
   * Handles keyup event
   * We handle KeyUp because ie11 and edge have difficulties copy/pasting
   * if a copy/cut event fired, keyup is dismissed
   * @param {Event} e Event object
   */
  onKeyUp: function(e) {
    if (!this.isEditing || this._copyDone) {
      this._copyDone = false;
      return;
    }
    if ((e.keyCode in this._ctrlKeysMapUp) && (e.ctrlKey || e.metaKey)) {
      this[this._ctrlKeysMapUp[e.keyCode]](e);
    }
    else {
      return;
    }
    e.stopImmediatePropagation();
    e.preventDefault();
    this.canvas && this.canvas.renderAll();
  },

  /**
   * Handles onInput event
   * @param {Event} e Event object
   */
  onInput: function(e) {
    if (!this.isEditing || this.inCompositionMode) {
      return;
    }
    var offset = this.selectionStart || 0,
        offsetEnd = this.selectionEnd || 0,
        textLength = this.text.length,
        newTextLength = this.hiddenTextarea.value.length,
        diff, charsToInsert, start;
    if (newTextLength > textLength) {
      //we added some character
      start = this._selectionDirection === 'left' ? offsetEnd : offset;
      diff = newTextLength - textLength;
      charsToInsert = this.hiddenTextarea.value.slice(start, start + diff);
    }
    else {
      //we selected a portion of text and then input something else.
      //Internet explorer does not trigger this else
      diff = newTextLength - textLength + offsetEnd - offset;
      charsToInsert = this.hiddenTextarea.value.slice(offset, offset + diff);
    }
    this.insertChars(charsToInsert);
    e.stopPropagation();
  },

  /**
   * Composition start
   */
  onCompositionStart: function() {
    this.inCompositionMode = true;
    this.prevCompositionLength = 0;
    this.compositionStart = this.selectionStart;
  },

  /**
   * Composition end
   */
  onCompositionEnd: function() {
    this.inCompositionMode = false;
  },

  /**
   * Composition update
   */
  onCompositionUpdate: function(e) {
    var data = e.data;
    this.selectionStart = this.compositionStart;
    this.selectionEnd = this.selectionEnd === this.selectionStart ?
      this.compositionStart + this.prevCompositionLength : this.selectionEnd;
    this.insertChars(data, false);
    this.prevCompositionLength = data.length;
  },

  /**
   * Forward delete
   */
  forwardDelete: function(e) {
    if (this.selectionStart === this.selectionEnd) {
      if (this.selectionStart === this.text.length) {
        return;
      }
      this.moveCursorRight(e);
    }
    this.removeChars(e);
  },

  /**
   * Copies selected text
   * @param {Event} e Event object
   */
  copy: function(e) {
    if (this.selectionStart === this.selectionEnd) {
      //do not cut-copy if no selection
      return;
    }
    var selectedText = this.getSelectedText(),
        clipboardData = this._getClipboardData(e);

    // Check for backward compatibility with old browsers
    if (clipboardData) {
      clipboardData.setData('text', selectedText);
    }

    fabric.copiedText = selectedText;
    fabric.copiedTextStyle = this.getSelectionStyles(
                          this.selectionStart,
                          this.selectionEnd);
    e.stopImmediatePropagation();
    e.preventDefault();
    this._copyDone = true;
  },

  /**
   * Pastes text
   * @param {Event} e Event object
   */
  paste: function(e) {
    var copiedText = null,
        clipboardData = this._getClipboardData(e),
        useCopiedStyle = true;

    // Check for backward compatibility with old browsers
    if (clipboardData) {
      copiedText = clipboardData.getData('text').replace(/\r/g, '');
      if (!fabric.copiedTextStyle || fabric.copiedText !== copiedText) {
        useCopiedStyle = false;
      }
    }
    else {
      copiedText = fabric.copiedText;
    }

    if (copiedText) {
      this.insertChars(copiedText, useCopiedStyle);
    }
    e.stopImmediatePropagation();
    e.preventDefault();
  },

  /**
   * Cuts text
   * @param {Event} e Event object
   */
  cut: function(e) {
    if (this.selectionStart === this.selectionEnd) {
      return;
    }

    this.copy(e);
    this.removeChars(e);
  },

  /**
   * @private
   * @param {Event} e Event object
   * @return {Object} Clipboard data object
   */
  _getClipboardData: function(e) {
    return (e && e.clipboardData) || fabric.window.clipboardData;
  },

  /**
   * Gets start offset of a selection
   * @param {Event} e Event object
   * @param {Boolean} isRight
   * @return {Number}
   */
  getDownCursorOffset: function(e, isRight) {
    var selectionProp = isRight ? this.selectionEnd : this.selectionStart,
        cursorLocation = this.get2DCursorLocation(selectionProp),
        _char, lineLeftOffset, lineIndex = cursorLocation.lineIndex,
        textOnSameLineBeforeCursor = this._textLines[lineIndex].slice(0, cursorLocation.charIndex),
        textOnSameLineAfterCursor = this._textLines[lineIndex].slice(cursorLocation.charIndex),
        textOnNextLine = this._textLines[lineIndex + 1] || '';

    // if on last line, down cursor goes to end of line
    if (lineIndex === this._textLines.length - 1 || e.metaKey || e.keyCode === 34) {

      // move to the end of a text
      return this.text.length - selectionProp;
    }

    var widthOfSameLineBeforeCursor = this._getLineWidth(this.ctx, lineIndex);
    lineLeftOffset = this._getLineLeftOffset(widthOfSameLineBeforeCursor);

    var widthOfCharsOnSameLineBeforeCursor = lineLeftOffset;

    for (var i = 0, len = textOnSameLineBeforeCursor.length; i < len; i++) {
      _char = textOnSameLineBeforeCursor[i];
      widthOfCharsOnSameLineBeforeCursor += this._getWidthOfChar(this.ctx, _char, lineIndex, i);
    }

    var indexOnNextLine = this._getIndexOnNextLine(
      cursorLocation, textOnNextLine, widthOfCharsOnSameLineBeforeCursor);

    return textOnSameLineAfterCursor.length + 1 + indexOnNextLine;
  },

  /**
   * @private
   */
  _getIndexOnNextLine: function(cursorLocation, textOnNextLine, widthOfCharsOnSameLineBeforeCursor) {
    var lineIndex = cursorLocation.lineIndex + 1,
        widthOfNextLine = this._getLineWidth(this.ctx, lineIndex),
        lineLeftOffset = this._getLineLeftOffset(widthOfNextLine),
        widthOfCharsOnNextLine = lineLeftOffset,
        indexOnNextLine = 0,
        foundMatch;

    for (var j = 0, jlen = textOnNextLine.length; j < jlen; j++) {

      var _char = textOnNextLine[j],
          widthOfChar = this._getWidthOfChar(this.ctx, _char, lineIndex, j);

      widthOfCharsOnNextLine += widthOfChar;

      if (widthOfCharsOnNextLine > widthOfCharsOnSameLineBeforeCursor) {

        foundMatch = true;

        var leftEdge = widthOfCharsOnNextLine - widthOfChar,
            rightEdge = widthOfCharsOnNextLine,
            offsetFromLeftEdge = Math.abs(leftEdge - widthOfCharsOnSameLineBeforeCursor),
            offsetFromRightEdge = Math.abs(rightEdge - widthOfCharsOnSameLineBeforeCursor);

        indexOnNextLine = offsetFromRightEdge < offsetFromLeftEdge ? j + 1 : j;

        break;
      }
    }

    // reached end
    if (!foundMatch) {
      indexOnNextLine = textOnNextLine.length;
    }

    return indexOnNextLine;
  },

  /**
   * Moves cursor down
   * @param {Event} e Event object
   */
  moveCursorDown: function(e) {
    this.abortCursorAnimation();
    this._currentCursorOpacity = 1;

    var offset = this.getDownCursorOffset(e, this._selectionDirection === 'right');

    if (e.shiftKey) {
      this.moveCursorDownWithShift(offset);
    }
    else {
      this.moveCursorDownWithoutShift(offset);
    }

    this.initDelayedCursor();
  },

  /**
   * Moves cursor down without keeping selection
   * @param {Number} offset
   */
  moveCursorDownWithoutShift: function(offset) {
    this._selectionDirection = 'right';
    this.setSelectionStart(this.selectionStart + offset);
    this.setSelectionEnd(this.selectionStart);
  },

  /**
   * private
   */
  swapSelectionPoints: function() {
    var swapSel = this.selectionEnd;
    this.setSelectionEnd(this.selectionStart);
    this.setSelectionStart(swapSel);
  },

  /**
   * Moves cursor down while keeping selection
   * @param {Number} offset
   */
  moveCursorDownWithShift: function(offset) {
    if (this.selectionEnd === this.selectionStart) {
      this._selectionDirection = 'right';
    }
    if (this._selectionDirection === 'right') {
      this.setSelectionEnd(this.selectionEnd + offset);
    }
    else {
      this.setSelectionStart(this.selectionStart + offset);
    }
    if (this.selectionEnd < this.selectionStart  && this._selectionDirection === 'left') {
      this.swapSelectionPoints();
      this._selectionDirection = 'right';
    }
    if (this.selectionEnd > this.text.length) {
      this.setSelectionEnd(this.text.length);
    }
  },

  /**
   * @param {Event} e Event object
   * @param {Boolean} isRight
   * @return {Number}
   */
  getUpCursorOffset: function(e, isRight) {
    var selectionProp = isRight ? this.selectionEnd : this.selectionStart,
        cursorLocation = this.get2DCursorLocation(selectionProp),
        lineIndex = cursorLocation.lineIndex;
    // if on first line, up cursor goes to start of line
    if (lineIndex === 0 || e.metaKey || e.keyCode === 33) {
      return selectionProp;
    }

    var textOnSameLineBeforeCursor = this._textLines[lineIndex].slice(0, cursorLocation.charIndex),
        textOnPreviousLine = this._textLines[lineIndex - 1] || '',
        _char,
        widthOfSameLineBeforeCursor = this._getLineWidth(this.ctx, cursorLocation.lineIndex),
        lineLeftOffset = this._getLineLeftOffset(widthOfSameLineBeforeCursor),
        widthOfCharsOnSameLineBeforeCursor = lineLeftOffset;

    for (var i = 0, len = textOnSameLineBeforeCursor.length; i < len; i++) {
      _char = textOnSameLineBeforeCursor[i];
      widthOfCharsOnSameLineBeforeCursor += this._getWidthOfChar(this.ctx, _char, lineIndex, i);
    }

    var indexOnPrevLine = this._getIndexOnPrevLine(
      cursorLocation, textOnPreviousLine, widthOfCharsOnSameLineBeforeCursor);

    return textOnPreviousLine.length - indexOnPrevLine + textOnSameLineBeforeCursor.length;
  },

  /**
   * @private
   */
  _getIndexOnPrevLine: function(cursorLocation, textOnPreviousLine, widthOfCharsOnSameLineBeforeCursor) {

    var lineIndex = cursorLocation.lineIndex - 1,
        widthOfPreviousLine = this._getLineWidth(this.ctx, lineIndex),
        lineLeftOffset = this._getLineLeftOffset(widthOfPreviousLine),
        widthOfCharsOnPreviousLine = lineLeftOffset,
        indexOnPrevLine = 0,
        foundMatch;

    for (var j = 0, jlen = textOnPreviousLine.length; j < jlen; j++) {

      var _char = textOnPreviousLine[j],
          widthOfChar = this._getWidthOfChar(this.ctx, _char, lineIndex, j);

      widthOfCharsOnPreviousLine += widthOfChar;

      if (widthOfCharsOnPreviousLine > widthOfCharsOnSameLineBeforeCursor) {

        foundMatch = true;

        var leftEdge = widthOfCharsOnPreviousLine - widthOfChar,
            rightEdge = widthOfCharsOnPreviousLine,
            offsetFromLeftEdge = Math.abs(leftEdge - widthOfCharsOnSameLineBeforeCursor),
            offsetFromRightEdge = Math.abs(rightEdge - widthOfCharsOnSameLineBeforeCursor);

        indexOnPrevLine = offsetFromRightEdge < offsetFromLeftEdge ? j : (j - 1);

        break;
      }
    }

    // reached end
    if (!foundMatch) {
      indexOnPrevLine = textOnPreviousLine.length - 1;
    }

    return indexOnPrevLine;
  },

  /**
   * Moves cursor up
   * @param {Event} e Event object
   */
  moveCursorUp: function(e) {

    this.abortCursorAnimation();
    this._currentCursorOpacity = 1;

    var offset = this.getUpCursorOffset(e, this._selectionDirection === 'right');
    if (e.shiftKey) {
      this.moveCursorUpWithShift(offset);
    }
    else {
      this.moveCursorUpWithoutShift(offset);
    }

    this.initDelayedCursor();
  },

  /**
   * Moves cursor up with shift
   * @param {Number} offset
   */
  moveCursorUpWithShift: function(offset) {
    if (this.selectionEnd === this.selectionStart) {
      this._selectionDirection = 'left';
    }
    if (this._selectionDirection === 'right') {
      this.setSelectionEnd(this.selectionEnd - offset);
    }
    else {
      this.setSelectionStart(this.selectionStart - offset);
    }
    if (this.selectionEnd < this.selectionStart && this._selectionDirection === 'right') {
      this.swapSelectionPoints();
      this._selectionDirection = 'left';
    }
  },

  /**
   * Moves cursor up without shift
   * @param {Number} offset
   */
  moveCursorUpWithoutShift: function(offset) {
    if (this.selectionStart === this.selectionEnd) {
      this.setSelectionStart(this.selectionStart - offset);
    }
    this.setSelectionEnd(this.selectionStart);

    this._selectionDirection = 'left';
  },

  /**
   * Moves cursor left
   * @param {Event} e Event object
   */
  moveCursorLeft: function(e) {
    if (this.selectionStart === 0 && this.selectionEnd === 0) {
      return;
    }

    this.abortCursorAnimation();
    this._currentCursorOpacity = 1;

    if (e.shiftKey) {
      this.moveCursorLeftWithShift(e);
    }
    else {
      this.moveCursorLeftWithoutShift(e);
    }

    this.initDelayedCursor();
  },

  /**
   * @private
   */
  _move: function(e, prop, direction) {
    var propMethod = (prop === 'selectionStart' ? 'setSelectionStart' : 'setSelectionEnd');
    if (e.altKey) {
      this[propMethod](this['findWordBoundary' + direction](this[prop]));
    }
    else if (e.metaKey || e.keyCode === 35 ||  e.keyCode === 36 ) {
      this[propMethod](this['findLineBoundary' + direction](this[prop]));
    }
    else {
      this[propMethod](this[prop] + (direction === 'Left' ? -1 : 1));
    }
  },

  /**
   * @private
   */
  _moveLeft: function(e, prop) {
    this._move(e, prop, 'Left');
  },

  /**
   * @private
   */
  _moveRight: function(e, prop) {
    this._move(e, prop, 'Right');
  },

  /**
   * Moves cursor left without keeping selection
   * @param {Event} e
   */
  moveCursorLeftWithoutShift: function(e) {
    this._selectionDirection = 'left';

    // only move cursor when there is no selection,
    // otherwise we discard it, and leave cursor on same place
    if (this.selectionEnd === this.selectionStart) {
      this._moveLeft(e, 'selectionStart');
    }
    this.setSelectionEnd(this.selectionStart);
  },

  /**
   * Moves cursor left while keeping selection
   * @param {Event} e
   */
  moveCursorLeftWithShift: function(e) {
    if (this._selectionDirection === 'right' && this.selectionStart !== this.selectionEnd) {
      this._moveLeft(e, 'selectionEnd');
    }
    else {
      this._selectionDirection = 'left';
      this._moveLeft(e, 'selectionStart');
    }
  },

  /**
   * Moves cursor right
   * @param {Event} e Event object
   */
  moveCursorRight: function(e) {
    if (this.selectionStart >= this.text.length && this.selectionEnd >= this.text.length) {
      return;
    }

    this.abortCursorAnimation();
    this._currentCursorOpacity = 1;

    if (e.shiftKey) {
      this.moveCursorRightWithShift(e);
    }
    else {
      this.moveCursorRightWithoutShift(e);
    }

    this.initDelayedCursor();
  },

  /**
   * Moves cursor right while keeping selection
   * @param {Event} e
   */
  moveCursorRightWithShift: function(e) {
    if (this._selectionDirection === 'left' && this.selectionStart !== this.selectionEnd) {
      this._moveRight(e, 'selectionStart');
    }
    else {
      this._selectionDirection = 'right';
      this._moveRight(e, 'selectionEnd');
    }
  },

  /**
   * Moves cursor right without keeping selection
   * @param {Event} e Event object
   */
  moveCursorRightWithoutShift: function(e) {
    this._selectionDirection = 'right';

    if (this.selectionStart === this.selectionEnd) {
      this._moveRight(e, 'selectionStart');
      this.setSelectionEnd(this.selectionStart);
    }
    else {
      this.setSelectionEnd(this.selectionEnd + this.getNumNewLinesInSelectedText());
      this.setSelectionStart(this.selectionEnd);
    }
  },

  /**
   * Removes characters selected by selection
   * @param {Event} e Event object
   */
  removeChars: function(e) {
    if (this.selectionStart === this.selectionEnd) {
      this._removeCharsNearCursor(e);
    }
    else {
      this._removeCharsFromTo(this.selectionStart, this.selectionEnd);
    }

    this.setSelectionEnd(this.selectionStart);

    this._removeExtraneousStyles();

    this.canvas && this.canvas.renderAll();

    this.setCoords();
    this.fire('changed');
    this.canvas && this.canvas.fire('text:changed', { target: this });
  },

  /**
   * @private
   * @param {Event} e Event object
   */
  _removeCharsNearCursor: function(e) {
    if (this.selectionStart === 0) {
      return;
    }
    if (e.metaKey) {
      // remove all till the start of current line
      var leftLineBoundary = this.findLineBoundaryLeft(this.selectionStart);

      this._removeCharsFromTo(leftLineBoundary, this.selectionStart);
      this.setSelectionStart(leftLineBoundary);
    }
    else if (e.altKey) {
      // remove all till the start of current word
      var leftWordBoundary = this.findWordBoundaryLeft(this.selectionStart);

      this._removeCharsFromTo(leftWordBoundary, this.selectionStart);
      this.setSelectionStart(leftWordBoundary);
    }
    else {
      this._removeSingleCharAndStyle(this.selectionStart);
      this.setSelectionStart(this.selectionStart - 1);
    }
  }
});


/* _TO_SVG_START_ */
(function() {
  var toFixed = fabric.util.toFixed,
      NUM_FRACTION_DIGITS = fabric.Object.NUM_FRACTION_DIGITS;

  fabric.util.object.extend(fabric.IText.prototype, /** @lends fabric.IText.prototype */ {

    /**
     * @private
     */
    _setSVGTextLineText: function(lineIndex, textSpans, height, textLeftOffset, textTopOffset, textBgRects) {
      if (!this._getLineStyle(lineIndex)) {
        fabric.Text.prototype._setSVGTextLineText.call(this,
          lineIndex, textSpans, height, textLeftOffset, textTopOffset);
      }
      else {
        this._setSVGTextLineChars(
          lineIndex, textSpans, height, textLeftOffset, textBgRects);
      }
    },

    /**
     * @private
     */
    _setSVGTextLineChars: function(lineIndex, textSpans, height, textLeftOffset, textBgRects) {

      var chars = this._textLines[lineIndex],
          charOffset = 0,
          lineLeftOffset = this._getLineLeftOffset(this._getLineWidth(this.ctx, lineIndex)) - this.width / 2,
          lineOffset = this._getSVGLineTopOffset(lineIndex),
          heightOfLine = this._getHeightOfLine(this.ctx, lineIndex);

      for (var i = 0, len = chars.length; i < len; i++) {
        var styleDecl = this._getStyleDeclaration(lineIndex, i) || { };

        textSpans.push(
          this._createTextCharSpan(
            chars[i], styleDecl, lineLeftOffset, lineOffset.lineTop + lineOffset.offset, charOffset));

        var charWidth = this._getWidthOfChar(this.ctx, chars[i], lineIndex, i);

        if (styleDecl.textBackgroundColor) {
          textBgRects.push(
            this._createTextCharBg(
              styleDecl, lineLeftOffset, lineOffset.lineTop, heightOfLine, charWidth, charOffset));
        }

        charOffset += charWidth;
      }
    },

    /**
     * @private
     */
    _getSVGLineTopOffset: function(lineIndex) {
      var lineTopOffset = 0, lastHeight = 0;
      for (var j = 0; j < lineIndex; j++) {
        lineTopOffset += this._getHeightOfLine(this.ctx, j);
      }
      lastHeight = this._getHeightOfLine(this.ctx, j);
      return {
        lineTop: lineTopOffset,
        offset: (this._fontSizeMult - this._fontSizeFraction) * lastHeight / (this.lineHeight * this._fontSizeMult)
      };
    },

    /**
     * @private
     */
    _createTextCharBg: function(styleDecl, lineLeftOffset, lineTopOffset, heightOfLine, charWidth, charOffset) {
      return [
        //jscs:disable validateIndentation
        '\t\t<rect fill="', styleDecl.textBackgroundColor,
        '" x="', toFixed(lineLeftOffset + charOffset, NUM_FRACTION_DIGITS),
        '" y="', toFixed(lineTopOffset - this.height/2, NUM_FRACTION_DIGITS),
        '" width="', toFixed(charWidth, NUM_FRACTION_DIGITS),
        '" height="', toFixed(heightOfLine / this.lineHeight, NUM_FRACTION_DIGITS),
        '"></rect>\n'
        //jscs:enable validateIndentation
      ].join('');
    },

    /**
     * @private
     */
    _createTextCharSpan: function(_char, styleDecl, lineLeftOffset, lineTopOffset, charOffset) {

      var fillStyles = this.getSvgStyles.call(fabric.util.object.extend({
        visible: true,
        fill: this.fill,
        stroke: this.stroke,
        type: 'text',
        getSvgFilter: fabric.Object.prototype.getSvgFilter
      }, styleDecl));

      return [
        //jscs:disable validateIndentation
        '\t\t\t<tspan x="', toFixed(lineLeftOffset + charOffset, NUM_FRACTION_DIGITS), '" y="',
          toFixed(lineTopOffset - this.height/2, NUM_FRACTION_DIGITS), '" ',
          (styleDecl.fontFamily ? 'font-family="' + styleDecl.fontFamily.replace(/"/g, '\'') + '" ': ''),
          (styleDecl.fontSize ? 'font-size="' + styleDecl.fontSize + '" ': ''),
          (styleDecl.fontStyle ? 'font-style="' + styleDecl.fontStyle + '" ': ''),
          (styleDecl.fontWeight ? 'font-weight="' + styleDecl.fontWeight + '" ': ''),
          (styleDecl.textDecoration ? 'text-decoration="' + styleDecl.textDecoration + '" ': ''),
          'style="', fillStyles, '">',
          fabric.util.string.escapeXml(_char),
        '</tspan>\n'
        //jscs:enable validateIndentation
      ].join('');
    }
  });
})();
/* _TO_SVG_END_ */


(function(global) {

  'use strict';

  var fabric = global.fabric || (global.fabric = {}),
      clone  = fabric.util.object.clone;

  /**
   * Textbox class, based on IText, allows the user to resize the text rectangle
   * and wraps lines automatically. Textboxes have their Y scaling locked, the
   * user can only change width. Height is adjusted automatically based on the
   * wrapping of lines.
   * @class fabric.Textbox
   * @extends fabric.IText
   * @mixes fabric.Observable
   * @return {fabric.Textbox} thisArg
   * @see {@link fabric.Textbox#initialize} for constructor definition
   */
  fabric.Textbox = fabric.util.createClass(fabric.IText, fabric.Observable, {

    /**
     * Type of an object
     * @type String
     * @default
     */
    type: 'textbox',

    /**
     * Minimum width of textbox, in pixels.
     * @type Number
     * @default
     */
    minWidth: 20,

    /**
     * Minimum calculated width of a textbox, in pixels.
     * @type Number
     * @default
     */
    dynamicMinWidth: 0,

    /**
     * Cached array of text wrapping.
     * @type Array
     */
    __cachedLines: null,

    /**
     * Override standard Object class values
     */
    lockScalingY: true,

    /**
     * Override standard Object class values
     */
    lockScalingFlip: true,

    /**
     * Constructor. Some scaling related property values are forced. Visibility
     * of controls is also fixed; only the rotation and width controls are
     * made available.
     * @param {String} text Text string
     * @param {Object} [options] Options object
     * @return {fabric.Textbox} thisArg
     */
    initialize: function(text, options) {
      this.ctx = fabric.util.createCanvasElement().getContext('2d');
      this.callSuper('initialize', text, options);
      this.setControlsVisibility(fabric.Textbox.getTextboxControlVisibility());

      // add width to this list of props that effect line wrapping.
      this._dimensionAffectingProps.width = true;
    },

    /**
     * Unlike superclass's version of this function, Textbox does not update
     * its width.
     * @param {CanvasRenderingContext2D} ctx Context to use for measurements
     * @private
     * @override
     */
    _initDimensions: function(ctx) {
      if (this.__skipDimension) {
        return;
      }

      if (!ctx) {
        ctx = fabric.util.createCanvasElement().getContext('2d');
        this._setTextStyles(ctx);
      }

      // clear dynamicMinWidth as it will be different after we re-wrap line
      this.dynamicMinWidth = 0;

      // wrap lines
      this._textLines = this._splitTextIntoLines();

      // if after wrapping, the width is smaller than dynamicMinWidth, change the width and re-wrap
      if (this.dynamicMinWidth > this.width) {
        this._set('width', this.dynamicMinWidth);
      }

      // clear cache and re-calculate height
      this._clearCache();
      this.height = this._getTextHeight(ctx);
    },

    /**
     * Generate an object that translates the style object so that it is
     * broken up by visual lines (new lines and automatic wrapping).
     * The original text styles object is broken up by actual lines (new lines only),
     * which is only sufficient for Text / IText
     * @private
     */
    _generateStyleMap: function() {
      var realLineCount     = 0,
          realLineCharCount = 0,
          charCount         = 0,
          map               = {};

      for (var i = 0; i < this._textLines.length; i++) {
        if (this.text[charCount] === '\n') {
          realLineCharCount = 0;
          charCount++;
          realLineCount++;
        }
        else if (this.text[charCount] === ' ') {
          // this case deals with space's that are removed from end of lines when wrapping
          realLineCharCount++;
          charCount++;
        }

        map[i] = { line: realLineCount, offset: realLineCharCount };

        charCount += this._textLines[i].length;
        realLineCharCount += this._textLines[i].length;
      }

      return map;
    },

    /**
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @param {Boolean} [returnCloneOrEmpty=false]
     * @private
     */
    _getStyleDeclaration: function(lineIndex, charIndex, returnCloneOrEmpty) {
      if (this._styleMap) {
        var map = this._styleMap[lineIndex];
        if (!map) {
          return returnCloneOrEmpty ? { } : null;
        }
        lineIndex = map.line;
        charIndex = map.offset + charIndex;
      }
      return this.callSuper('_getStyleDeclaration', lineIndex, charIndex, returnCloneOrEmpty);
    },

    /**
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @param {Object} style
     * @private
     */
    _setStyleDeclaration: function(lineIndex, charIndex, style) {
      var map = this._styleMap[lineIndex];
      lineIndex = map.line;
      charIndex = map.offset + charIndex;

      this.styles[lineIndex][charIndex] = style;
    },

    /**
     * @param {Number} lineIndex
     * @param {Number} charIndex
     * @private
     */
    _deleteStyleDeclaration: function(lineIndex, charIndex) {
      var map = this._styleMap[lineIndex];
      lineIndex = map.line;
      charIndex = map.offset + charIndex;

      delete this.styles[lineIndex][charIndex];
    },

    /**
     * @param {Number} lineIndex
     * @private
     */
    _getLineStyle: function(lineIndex) {
      var map = this._styleMap[lineIndex];
      return this.styles[map.line];
    },

    /**
     * @param {Number} lineIndex
     * @param {Object} style
     * @private
     */
    _setLineStyle: function(lineIndex, style) {
      var map = this._styleMap[lineIndex];
      this.styles[map.line] = style;
    },

    /**
     * @param {Number} lineIndex
     * @private
     */
    _deleteLineStyle: function(lineIndex) {
      var map = this._styleMap[lineIndex];
      delete this.styles[map.line];
    },

    /**
     * Wraps text using the 'width' property of Textbox. First this function
     * splits text on newlines, so we preserve newlines entered by the user.
     * Then it wraps each line using the width of the Textbox by calling
     * _wrapLine().
     * @param {CanvasRenderingContext2D} ctx Context to use for measurements
     * @param {String} text The string of text that is split into lines
     * @returns {Array} Array of lines
     */
    _wrapText: function(ctx, text) {
      var lines = text.split(this._reNewline), wrapped = [], i;

      for (i = 0; i < lines.length; i++) {
        wrapped = wrapped.concat(this._wrapLine(ctx, lines[i], i));
      }

      return wrapped;
    },

    /**
     * Helper function to measure a string of text, given its lineIndex and charIndex offset
     *
     * @param {CanvasRenderingContext2D} ctx
     * @param {String} text
     * @param {number} lineIndex
     * @param {number} charOffset
     * @returns {number}
     * @private
     */
    _measureText: function(ctx, text, lineIndex, charOffset) {
      var width = 0;
      charOffset = charOffset || 0;

      for (var i = 0, len = text.length; i < len; i++) {
        width += this._getWidthOfChar(ctx, text[i], lineIndex, i + charOffset);
      }

      return width;
    },

    /**
     * Wraps a line of text using the width of the Textbox and a context.
     * @param {CanvasRenderingContext2D} ctx Context to use for measurements
     * @param {String} text The string of text to split into lines
     * @param {Number} lineIndex
     * @returns {Array} Array of line(s) into which the given text is wrapped
     * to.
     */
    _wrapLine: function(ctx, text, lineIndex) {
      var lineWidth        = 0,
          lines            = [],
          line             = '',
          words            = text.split(' '),
          word             = '',
          offset           = 0,
          infix            = ' ',
          wordWidth        = 0,
          infixWidth       = 0,
          largestWordWidth = 0,
          lineJustStarted = true;

      for (var i = 0; i < words.length; i++) {
        word = words[i];
        wordWidth = this._measureText(ctx, word, lineIndex, offset);

        offset += word.length;

        lineWidth += infixWidth + wordWidth;

        if (lineWidth >= this.width && !lineJustStarted) {
          lines.push(line);
          line = '';
          lineWidth = wordWidth;
          lineJustStarted = true;
        }

        if (!lineJustStarted) {
          line += infix;
        }
        line += word;

        infixWidth = this._measureText(ctx, infix, lineIndex, offset);
        offset++;
        lineJustStarted = false;
        // keep track of largest word
        if (wordWidth > largestWordWidth) {
          largestWordWidth = wordWidth;
        }
      }

      i && lines.push(line);

      if (largestWordWidth > this.dynamicMinWidth) {
        this.dynamicMinWidth = largestWordWidth;
      }

      return lines;
    },
    /**
     * Gets lines of text to render in the Textbox. This function calculates
     * text wrapping on the fly everytime it is called.
     * @returns {Array} Array of lines in the Textbox.
     * @override
     */
    _splitTextIntoLines: function() {
      var originalAlign = this.textAlign;
      this.ctx.save();
      this._setTextStyles(this.ctx);
      this.textAlign = 'left';
      var lines = this._wrapText(this.ctx, this.text);
      this.textAlign = originalAlign;
      this.ctx.restore();
      this._textLines = lines;
      this._styleMap = this._generateStyleMap();
      return lines;
    },

    /**
     * When part of a group, we don't want the Textbox's scale to increase if
     * the group's increases. That's why we reduce the scale of the Textbox by
     * the amount that the group's increases. This is to maintain the effective
     * scale of the Textbox at 1, so that font-size values make sense. Otherwise
     * the same font-size value would result in different actual size depending
     * on the value of the scale.
     * @param {String} key
     * @param {Any} value
     */
    setOnGroup: function(key, value) {
      if (key === 'scaleX') {
        this.set('scaleX', Math.abs(1 / value));
        this.set('width', (this.get('width') * value) /
          (typeof this.__oldScaleX === 'undefined' ? 1 : this.__oldScaleX));
        this.__oldScaleX = value;
      }
    },

    /**
     * Returns 2d representation (lineIndex and charIndex) of cursor (or selection start).
     * Overrides the superclass function to take into account text wrapping.
     *
     * @param {Number} [selectionStart] Optional index. When not given, current selectionStart is used.
     */
    get2DCursorLocation: function(selectionStart) {
      if (typeof selectionStart === 'undefined') {
        selectionStart = this.selectionStart;
      }

      var numLines = this._textLines.length,
          removed  = 0;

      for (var i = 0; i < numLines; i++) {
        var line    = this._textLines[i],
            lineLen = line.length;

        if (selectionStart <= removed + lineLen) {
          return {
            lineIndex: i,
            charIndex: selectionStart - removed
          };
        }

        removed += lineLen;

        if (this.text[removed] === '\n' || this.text[removed] === ' ') {
          removed++;
        }
      }

      return {
        lineIndex: numLines - 1,
        charIndex: this._textLines[numLines - 1].length
      };
    },

    /**
     * Overrides superclass function and uses text wrapping data to get cursor
     * boundary offsets instead of the array of chars.
     * @param {Array} chars Unused
     * @param {String} typeOfBoundaries Can be 'cursor' or 'selection'
     * @returns {Object} Object with 'top', 'left', and 'lineLeft' properties set.
     */
    _getCursorBoundariesOffsets: function(chars, typeOfBoundaries) {
      var topOffset      = 0,
          leftOffset     = 0,
          cursorLocation = this.get2DCursorLocation(),
          lineChars      = this._textLines[cursorLocation.lineIndex].split(''),
          lineLeftOffset = this._getLineLeftOffset(this._getLineWidth(this.ctx, cursorLocation.lineIndex));

      for (var i = 0; i < cursorLocation.charIndex; i++) {
        leftOffset += this._getWidthOfChar(this.ctx, lineChars[i], cursorLocation.lineIndex, i);
      }

      for (i = 0; i < cursorLocation.lineIndex; i++) {
        topOffset += this._getHeightOfLine(this.ctx, i);
      }

      if (typeOfBoundaries === 'cursor') {
        topOffset += (1 - this._fontSizeFraction) * this._getHeightOfLine(this.ctx, cursorLocation.lineIndex)
          / this.lineHeight - this.getCurrentCharFontSize(cursorLocation.lineIndex, cursorLocation.charIndex)
          * (1 - this._fontSizeFraction);
      }

      return {
        top: topOffset,
        left: leftOffset,
        lineLeft: lineLeftOffset
      };
    },

    getMinWidth: function() {
      return Math.max(this.minWidth, this.dynamicMinWidth);
    },

    /**
     * Returns object representation of an instance
     * @method toObject
     * @param {Array} [propertiesToInclude] Any properties that you might want to additionally include in the output
     * @return {Object} object representation of an instance
     */
    toObject: function(propertiesToInclude) {
      return fabric.util.object.extend(this.callSuper('toObject', propertiesToInclude), {
        minWidth: this.minWidth
      });
    }
  });
  /**
   * Returns fabric.Textbox instance from an object representation
   * @static
   * @memberOf fabric.Textbox
   * @param {Object} object Object to create an instance from
   * @return {fabric.Textbox} instance of fabric.Textbox
   */
  fabric.Textbox.fromObject = function(object) {
    return new fabric.Textbox(object.text, clone(object));
  };
  /**
   * Returns the default controls visibility required for Textboxes.
   * @returns {Object}
   */
  fabric.Textbox.getTextboxControlVisibility = function() {
    return {
      tl: false,
      tr: false,
      br: false,
      bl: false,
      ml: true,
      mt: false,
      mr: true,
      mb: false,
      mtr: true
    };
  };

})(typeof exports !== 'undefined' ? exports : this);


(function() {

  /**
   * Override _setObjectScale and add Textbox specific resizing behavior. Resizing
   * a Textbox doesn't scale text, it only changes width and makes text wrap automatically.
   */
  var setObjectScaleOverridden = fabric.Canvas.prototype._setObjectScale;

  fabric.Canvas.prototype._setObjectScale = function(localMouse, transform,
                                                     lockScalingX, lockScalingY, by, lockScalingFlip, _dim) {

    var t = transform.target;
    if (t instanceof fabric.Textbox) {
      var w = t.width * ((localMouse.x / transform.scaleX) / (t.width + t.strokeWidth));
      if (w >= t.getMinWidth()) {
        t.set('width', w);
        return true;
      }
    }
    else {
      return setObjectScaleOverridden.call(fabric.Canvas.prototype, localMouse, transform,
        lockScalingX, lockScalingY, by, lockScalingFlip, _dim);
    }
  };

  /**
   * Sets controls of this group to the Textbox's special configuration if
   * one is present in the group. Deletes _controlsVisibility otherwise, so that
   * it gets initialized to default value at runtime.
   */
  fabric.Group.prototype._refreshControlsVisibility = function() {
    if (typeof fabric.Textbox === 'undefined') {
      return;
    }
    for (var i = this._objects.length; i--;) {
      if (this._objects[i] instanceof fabric.Textbox) {
        this.setControlsVisibility(fabric.Textbox.getTextboxControlVisibility());
        return;
      }
    }
  };

  var clone = fabric.util.object.clone;

  fabric.util.object.extend(fabric.Textbox.prototype, /** @lends fabric.IText.prototype */ {
    /**
     * @private
     */
    _removeExtraneousStyles: function() {
      for (var prop in this._styleMap) {
        if (!this._textLines[prop]) {
          delete this.styles[this._styleMap[prop].line];
        }
      }
    },

    /**
     * Inserts style object for a given line/char index
     * @param {Number} lineIndex Index of a line
     * @param {Number} charIndex Index of a char
     * @param {Object} [style] Style object to insert, if given
     */
    insertCharStyleObject: function(lineIndex, charIndex, style) {
      // adjust lineIndex and charIndex
      var map = this._styleMap[lineIndex];
      lineIndex = map.line;
      charIndex = map.offset + charIndex;

      fabric.IText.prototype.insertCharStyleObject.apply(this, [lineIndex, charIndex, style]);
    },

    /**
     * Inserts new style object
     * @param {Number} lineIndex Index of a line
     * @param {Number} charIndex Index of a char
     * @param {Boolean} isEndOfLine True if it's end of line
     */
    insertNewlineStyleObject: function(lineIndex, charIndex, isEndOfLine) {
      // adjust lineIndex and charIndex
      var map = this._styleMap[lineIndex];
      lineIndex = map.line;
      charIndex = map.offset + charIndex;

      fabric.IText.prototype.insertNewlineStyleObject.apply(this, [lineIndex, charIndex, isEndOfLine]);
    },

    /**
     * Shifts line styles up or down. This function is slightly different than the one in
     * itext_behaviour as it takes into account the styleMap.
     *
     * @param {Number} lineIndex Index of a line
     * @param {Number} offset Can be -1 or +1
     */
    shiftLineStyles: function(lineIndex, offset) {
      // shift all line styles by 1 upward
      var clonedStyles = clone(this.styles),
          map          = this._styleMap[lineIndex];

      // adjust line index
      lineIndex = map.line;

      for (var line in this.styles) {
        var numericLine = parseInt(line, 10);

        if (numericLine > lineIndex) {
          this.styles[numericLine + offset] = clonedStyles[numericLine];

          if (!clonedStyles[numericLine - offset]) {
            delete this.styles[numericLine];
          }
        }
      }
      //TODO: evaluate if delete old style lines with offset -1
    },

    /**
     * Figure out programatically the text on previous actual line (actual = separated by \n);
     *
     * @param {Number} lIndex
     * @returns {String}
     * @private
     */
    _getTextOnPreviousLine: function(lIndex) {
      var textOnPreviousLine = this._textLines[lIndex - 1];

      while (this._styleMap[lIndex - 2] && this._styleMap[lIndex - 2].line === this._styleMap[lIndex - 1].line) {
        textOnPreviousLine = this._textLines[lIndex - 2] + textOnPreviousLine;

        lIndex--;
      }

      return textOnPreviousLine;
    },

    /**
     * Removes style object
     * @param {Boolean} isBeginningOfLine True if cursor is at the beginning of line
     * @param {Number} [index] Optional index. When not given, current selectionStart is used.
     */
    removeStyleObject: function(isBeginningOfLine, index) {

      var cursorLocation = this.get2DCursorLocation(index),
          map            = this._styleMap[cursorLocation.lineIndex],
          lineIndex      = map.line,
          charIndex      = map.offset + cursorLocation.charIndex;
      this._removeStyleObject(isBeginningOfLine, cursorLocation, lineIndex, charIndex);
    }
  });
})();


(function() {
  var override = fabric.IText.prototype._getNewSelectionStartFromOffset;
  /**
   * Overrides the IText implementation and adjusts character index as there is not always a linebreak
   *
   * @param {Number} mouseOffset
   * @param {Number} prevWidth
   * @param {Number} width
   * @param {Number} index
   * @param {Number} jlen
   * @returns {Number}
   */
  fabric.IText.prototype._getNewSelectionStartFromOffset = function(mouseOffset, prevWidth, width, index, jlen) {
    index = override.call(this, mouseOffset, prevWidth, width, index, jlen);

    // the index passed into the function is padded by the amount of lines from _textLines (to account for \n)
    // we need to remove this padding, and pad it by actual lines, and / or spaces that are meant to be there
    var tmp     = 0,
        removed = 0;

    // account for removed characters
    for (var i = 0; i < this._textLines.length; i++) {
      tmp += this._textLines[i].length;

      if (tmp + removed >= index) {
        break;
      }

      if (this.text[tmp + removed] === '\n' || this.text[tmp + removed] === ' ') {
        removed++;
      }
    }

    return index - i + removed;
  };
})();


(function() {

  if (typeof document !== 'undefined' && typeof window !== 'undefined') {
    return;
  }

  var DOMParser = require('xmldom').DOMParser,
      URL = require('url'),
      HTTP = require('http'),
      HTTPS = require('https'),

      Canvas = require('canvas'),
      Image = require('canvas').Image;

  /** @private */
  function request(url, encoding, callback) {
    var oURL = URL.parse(url);

    // detect if http or https is used
    if ( !oURL.port ) {
      oURL.port = ( oURL.protocol.indexOf('https:') === 0 ) ? 443 : 80;
    }

    // assign request handler based on protocol
    var reqHandler = (oURL.protocol.indexOf('https:') === 0 ) ? HTTPS : HTTP,
        req = reqHandler.request({
          hostname: oURL.hostname,
          port: oURL.port,
          path: oURL.path,
          method: 'GET'
        }, function(response) {
          var body = '';
          if (encoding) {
            response.setEncoding(encoding);
          }
          response.on('end', function () {
            callback(body);
          });
          response.on('data', function (chunk) {
            if (response.statusCode === 200) {
              body += chunk;
            }
          });
        });

    req.on('error', function(err) {
      if (err.errno === process.ECONNREFUSED) {
        fabric.log('ECONNREFUSED: connection refused to ' + oURL.hostname + ':' + oURL.port);
      }
      else {
        fabric.log(err.message);
      }
      callback(null);
    });

    req.end();
  }

  /** @private */
  function requestFs(path, callback) {
    var fs = require('fs');
    fs.readFile(path, function (err, data) {
      if (err) {
        fabric.log(err);
        throw err;
      }
      else {
        callback(data);
      }
    });
  }

  fabric.util.loadImage = function(url, callback, context) {
    function createImageAndCallBack(data) {
      if (data) {
        img.src = new Buffer(data, 'binary');
        // preserving original url, which seems to be lost in node-canvas
        img._src = url;
        callback && callback.call(context, img);
      }
      else {
        img = null;
        callback && callback.call(context, null, true);
      }
    }
    var img = new Image();
    if (url && (url instanceof Buffer || url.indexOf('data') === 0)) {
      img.src = img._src = url;
      callback && callback.call(context, img);
    }
    else if (url && url.indexOf('http') !== 0) {
      requestFs(url, createImageAndCallBack);
    }
    else if (url) {
      request(url, 'binary', createImageAndCallBack);
    }
    else {
      callback && callback.call(context, url);
    }
  };

  fabric.loadSVGFromURL = function(url, callback, reviver) {
    url = url.replace(/^\n\s*/, '').replace(/\?.*$/, '').trim();
    if (url.indexOf('http') !== 0) {
      requestFs(url, function(body) {
        fabric.loadSVGFromString(body.toString(), callback, reviver);
      });
    }
    else {
      request(url, '', function(body) {
        fabric.loadSVGFromString(body, callback, reviver);
      });
    }
  };

  fabric.loadSVGFromString = function(string, callback, reviver) {
    var doc = new DOMParser().parseFromString(string);
    fabric.parseSVGDocument(doc.documentElement, function(results, options) {
      callback && callback(results, options);
    }, reviver);
  };

  fabric.util.getScript = function(url, callback) {
    request(url, '', function(body) {
      eval(body);
      callback && callback();
    });
  };

  fabric.Image.fromObject = function(object, callback) {
    fabric.util.loadImage(object.src, function(img) {
      var oImg = new fabric.Image(img);

      oImg._initConfig(object);
      oImg._initFilters(object.filters, function(filters) {
        oImg.filters = filters || [ ];
        oImg._initFilters(object.resizeFilters, function(resizeFilters) {
          oImg.resizeFilters = resizeFilters || [ ];
          callback && callback(oImg);
        });
      });
    });
  };
  /**
   * Only available when running fabric on node.js
   * @param {Number} width Canvas width
   * @param {Number} height Canvas height
   * @param {Object} [options] Options to pass to FabricCanvas.
   * @param {Object} [nodeCanvasOptions] Options to pass to NodeCanvas.
   * @return {Object} wrapped canvas instance
   */
  fabric.createCanvasForNode = function(width, height, options, nodeCanvasOptions) {
    nodeCanvasOptions = nodeCanvasOptions || options;

    var canvasEl = fabric.document.createElement('canvas'),
        nodeCanvas = new Canvas(width || 600, height || 600, nodeCanvasOptions);

    // jsdom doesn't create style on canvas element, so here be temp. workaround
    canvasEl.style = { };

    canvasEl.width = nodeCanvas.width;
    canvasEl.height = nodeCanvas.height;

    var FabricCanvas = fabric.Canvas || fabric.StaticCanvas,
        fabricCanvas = new FabricCanvas(canvasEl, options);

    fabricCanvas.contextContainer = nodeCanvas.getContext('2d');
    fabricCanvas.nodeCanvas = nodeCanvas;
    fabricCanvas.Font = Canvas.Font;

    return fabricCanvas;
  };

  /** @ignore */
  fabric.StaticCanvas.prototype.createPNGStream = function() {
    return this.nodeCanvas.createPNGStream();
  };

  fabric.StaticCanvas.prototype.createJPEGStream = function(opts) {
    return this.nodeCanvas.createJPEGStream(opts);
  };

  var origSetWidth = fabric.StaticCanvas.prototype.setWidth;
  fabric.StaticCanvas.prototype.setWidth = function(width, options) {
    origSetWidth.call(this, width, options);
    this.nodeCanvas.width = width;
    return this;
  };
  if (fabric.Canvas) {
    fabric.Canvas.prototype.setWidth = fabric.StaticCanvas.prototype.setWidth;
  }

  var origSetHeight = fabric.StaticCanvas.prototype.setHeight;
  fabric.StaticCanvas.prototype.setHeight = function(height, options) {
    origSetHeight.call(this, height, options);
    this.nodeCanvas.height = height;
    return this;
  };
  if (fabric.Canvas) {
    fabric.Canvas.prototype.setHeight = fabric.StaticCanvas.prototype.setHeight;
  }

})();

/* pako 0.2.3 nodeca/pako */!function(e){if("object"==typeof exports&&"undefined"!=typeof module)module.exports=e();else if("function"==typeof define&&define.amd)define([],e);else{var f;"undefined"!=typeof window?f=window:"undefined"!=typeof global?f=global:"undefined"!=typeof self&&(f=self),f.pako=e()}}(function(){var define,module,exports;return (function e(t,n,r){function s(o,u){if(!n[o]){if(!t[o]){var a=typeof require=="function"&&require;if(!u&&a)return a(o,!0);if(i)return i(o,!0);throw new Error("Cannot find module '"+o+"'")}var f=n[o]={exports:{}};t[o][0].call(f.exports,function(e){var n=t[o][1][e];return s(n?n:e)},f,f.exports,e,t,n,r)}return n[o].exports}var i=typeof require=="function"&&require;for(var o=0;o<r.length;o++)s(r[o]);return s})({1:[function(_dereq_,module,exports){
'use strict';


var zlib_inflate = _dereq_('./zlib/inflate.js');
var utils = _dereq_('./utils/common');
var strings = _dereq_('./utils/strings');
var c = _dereq_('./zlib/constants');
var msg = _dereq_('./zlib/messages');
var zstream = _dereq_('./zlib/zstream');
var gzheader = _dereq_('./zlib/gzheader');


/**
 * class Inflate
 *
 * Generic JS-style wrapper for zlib calls. If you don't need
 * streaming behaviour - use more simple functions: [[inflate]]
 * and [[inflateRaw]].
 **/

/* internal
 * inflate.chunks -> Array
 *
 * Chunks of output data, if [[Inflate#onData]] not overriden.
 **/

/**
 * Inflate.result -> Uint8Array|Array|String
 *
 * Uncompressed result, generated by default [[Inflate#onData]]
 * and [[Inflate#onEnd]] handlers. Filled after you push last chunk
 * (call [[Inflate#push]] with `Z_FINISH` / `true` param).
 **/

/**
 * Inflate.err -> Number
 *
 * Error code after inflate finished. 0 (Z_OK) on success.
 * Should be checked if broken data possible.
 **/

/**
 * Inflate.msg -> String
 *
 * Error message, if [[Inflate.err]] != 0
 **/


/**
 * new Inflate(options)
 * - options (Object): zlib inflate options.
 *
 * Creates new inflator instance with specified params. Throws exception
 * on bad params. Supported options:
 *
 * - `windowBits`
 *
 * [http://zlib.net/manual.html#Advanced](http://zlib.net/manual.html#Advanced)
 * for more information on these.
 *
 * Additional options, for internal needs:
 *
 * - `chunkSize` - size of generated data chunks (16K by default)
 * - `raw` (Boolean) - do raw inflate
 * - `to` (String) - if equal to 'string', then result will be converted
 *   from utf8 to utf16 (javascript) string. When string output requested,
 *   chunk length can differ from `chunkSize`, depending on content.
 *
 * By default, when no options set, autodetect deflate/gzip data format via
 * wrapper header.
 *
 * ##### Example:
 *
 * ```javascript
 * var pako = require('pako')
 *   , chunk1 = Uint8Array([1,2,3,4,5,6,7,8,9])
 *   , chunk2 = Uint8Array([10,11,12,13,14,15,16,17,18,19]);
 *
 * var inflate = new pako.Inflate({ level: 3});
 *
 * inflate.push(chunk1, false);
 * inflate.push(chunk2, true);  // true -> last chunk
 *
 * if (inflate.err) { throw new Error(inflate.err); }
 *
 * console.log(inflate.result);
 * ```
 **/
var Inflate = function(options) {

  this.options = utils.assign({
    chunkSize: 16384,
    windowBits: 0,
    to: ''
  }, options || {});

  var opt = this.options;

  // Force window size for `raw` data, if not set directly,
  // because we have no header for autodetect.
  if (opt.raw && (opt.windowBits >= 0) && (opt.windowBits < 16)) {
    opt.windowBits = -opt.windowBits;
    if (opt.windowBits === 0) { opt.windowBits = -15; }
  }

  // If `windowBits` not defined (and mode not raw) - set autodetect flag for gzip/deflate
  if ((opt.windowBits >= 0) && (opt.windowBits < 16) &&
      !(options && options.windowBits)) {
    opt.windowBits += 32;
  }

  // Gzip header has no info about windows size, we can do autodetect only
  // for deflate. So, if window size not set, force it to max when gzip possible
  if ((opt.windowBits > 15) && (opt.windowBits < 48)) {
    // bit 3 (16) -> gzipped data
    // bit 4 (32) -> autodetect gzip/deflate
    if ((opt.windowBits & 15) === 0) {
      opt.windowBits |= 15;
    }
  }

  this.err    = 0;      // error code, if happens (0 = Z_OK)
  this.msg    = '';     // error message
  this.ended  = false;  // used to avoid multiple onEnd() calls
  this.chunks = [];     // chunks of compressed data

  this.strm   = new zstream();
  this.strm.avail_out = 0;

  var status  = zlib_inflate.inflateInit2(
    this.strm,
    opt.windowBits
  );

  if (status !== c.Z_OK) {
    throw new Error(msg[status]);
  }

  this.header = new gzheader();

  zlib_inflate.inflateGetHeader(this.strm, this.header);
};

/**
 * Inflate#push(data[, mode]) -> Boolean
 * - data (Uint8Array|Array|String): input data
 * - mode (Number|Boolean): 0..6 for corresponding Z_NO_FLUSH..Z_TREE modes.
 *   See constants. Skipped or `false` means Z_NO_FLUSH, `true` meansh Z_FINISH.
 *
 * Sends input data to inflate pipe, generating [[Inflate#onData]] calls with
 * new output chunks. Returns `true` on success. The last data block must have
 * mode Z_FINISH (or `true`). That flush internal pending buffers and call
 * [[Inflate#onEnd]].
 *
 * On fail call [[Inflate#onEnd]] with error code and return false.
 *
 * We strongly recommend to use `Uint8Array` on input for best speed (output
 * format is detected automatically). Also, don't skip last param and always
 * use the same type in your code (boolean or number). That will improve JS speed.
 *
 * For regular `Array`-s make sure all elements are [0..255].
 *
 * ##### Example
 *
 * ```javascript
 * push(chunk, false); // push one of data chunks
 * ...
 * push(chunk, true);  // push last chunk
 * ```
 **/
Inflate.prototype.push = function(data, mode) {
  var strm = this.strm;
  var chunkSize = this.options.chunkSize;
  var status, _mode;
  var next_out_utf8, tail, utf8str;

  if (this.ended) { return false; }
  _mode = (mode === ~~mode) ? mode : ((mode === true) ? c.Z_FINISH : c.Z_NO_FLUSH);

  // Convert data if needed
  if (typeof data === 'string') {
    // Only binary strings can be decompressed on practice
    strm.input = strings.binstring2buf(data);
  } else {
    strm.input = data;
  }

  strm.next_in = 0;
  strm.avail_in = strm.input.length;

  do {
    if (strm.avail_out === 0) {
      strm.output = new utils.Buf8(chunkSize);
      strm.next_out = 0;
      strm.avail_out = chunkSize;
    }

    status = zlib_inflate.inflate(strm, c.Z_NO_FLUSH);    /* no bad return value */

    if (status !== c.Z_STREAM_END && status !== c.Z_OK) {
      this.onEnd(status);
      this.ended = true;
      return false;
    }

    if (strm.next_out) {
      if (strm.avail_out === 0 || status === c.Z_STREAM_END || (strm.avail_in === 0 && _mode === c.Z_FINISH)) {

        if (this.options.to === 'string') {

          next_out_utf8 = strings.utf8border(strm.output, strm.next_out);

          tail = strm.next_out - next_out_utf8;
          utf8str = strings.buf2string(strm.output, next_out_utf8);

          // move tail
          strm.next_out = tail;
          strm.avail_out = chunkSize - tail;
          if (tail) { utils.arraySet(strm.output, strm.output, next_out_utf8, tail, 0); }

          this.onData(utf8str);

        } else {
          this.onData(utils.shrinkBuf(strm.output, strm.next_out));
        }
      }
    }
  } while ((strm.avail_in > 0 || strm.avail_out === 0) && status !== c.Z_STREAM_END);

  if (status === c.Z_STREAM_END) {
    _mode = c.Z_FINISH;
  }
  // Finalize on the last chunk.
  if (_mode === c.Z_FINISH) {
    status = zlib_inflate.inflateEnd(this.strm);
    this.onEnd(status);
    this.ended = true;
    return status === c.Z_OK;
  }

  return true;
};


/**
 * Inflate#onData(chunk) -> Void
 * - chunk (Uint8Array|Array|String): ouput data. Type of array depends
 *   on js engine support. When string output requested, each chunk
 *   will be string.
 *
 * By default, stores data blocks in `chunks[]` property and glue
 * those in `onEnd`. Override this handler, if you need another behaviour.
 **/
Inflate.prototype.onData = function(chunk) {
  this.chunks.push(chunk);
};


/**
 * Inflate#onEnd(status) -> Void
 * - status (Number): inflate status. 0 (Z_OK) on success,
 *   other if not.
 *
 * Called once after you tell inflate that input stream complete
 * or error happenned. By default - join collected chunks,
 * free memory and fill `results` / `err` properties.
 **/
Inflate.prototype.onEnd = function(status) {
  // On success - join
  if (status === c.Z_OK) {
    if (this.options.to === 'string') {
      // Glue & convert here, until we teach pako to send
      // utf8 alligned strings to onData
      this.result = this.chunks.join('');
    } else {
      this.result = utils.flattenChunks(this.chunks);
    }
  }
  this.chunks = [];
  this.err = status;
  this.msg = this.strm.msg;
};


/**
 * inflate(data[, options]) -> Uint8Array|Array|String
 * - data (Uint8Array|Array|String): input data to decompress.
 * - options (Object): zlib inflate options.
 *
 * Decompress `data` with inflate/ungzip and `options`. Autodetect
 * format via wrapper header by default. That's why we don't provide
 * separate `ungzip` method.
 *
 * Supported options are:
 *
 * - windowBits
 *
 * [http://zlib.net/manual.html#Advanced](http://zlib.net/manual.html#Advanced)
 * for more information.
 *
 * Sugar (options):
 *
 * - `raw` (Boolean) - say that we work with raw stream, if you don't wish to specify
 *   negative windowBits implicitly.
 * - `to` (String) - if equal to 'string', then result will be converted
 *   from utf8 to utf16 (javascript) string. When string output requested,
 *   chunk length can differ from `chunkSize`, depending on content.
 *
 *
 * ##### Example:
 *
 * ```javascript
 * var pako = require('pako')
 *   , input = pako.deflate([1,2,3,4,5,6,7,8,9])
 *   , output;
 *
 * try {
 *   output = pako.inflate(input);
 * } catch (err)
 *   console.log(err);
 * }
 * ```
 **/
function inflate(input, options) {
  var inflator = new Inflate(options);

  inflator.push(input, true);

  // That will never happens, if you don't cheat with options :)
  if (inflator.err) { throw inflator.msg; }

  return inflator.result;
}


/**
 * inflateRaw(data[, options]) -> Uint8Array|Array|String
 * - data (Uint8Array|Array|String): input data to decompress.
 * - options (Object): zlib inflate options.
 *
 * The same as [[inflate]], but creates raw data, without wrapper
 * (header and adler32 crc).
 **/
function inflateRaw(input, options) {
  options = options || {};
  options.raw = true;
  return inflate(input, options);
}


/**
 * ungzip(data[, options]) -> Uint8Array|Array|String
 * - data (Uint8Array|Array|String): input data to decompress.
 * - options (Object): zlib inflate options.
 *
 * Just shortcut to [[inflate]], because it autodetects format
 * by header.content. Done for convenience.
 **/


exports.Inflate = Inflate;
exports.inflate = inflate;
exports.inflateRaw = inflateRaw;
exports.ungzip  = inflate;

},{"./utils/common":2,"./utils/strings":3,"./zlib/constants":5,"./zlib/gzheader":7,"./zlib/inflate.js":9,"./zlib/messages":11,"./zlib/zstream":12}],2:[function(_dereq_,module,exports){
'use strict';


var TYPED_OK =  (typeof Uint8Array !== 'undefined') &&
                (typeof Uint16Array !== 'undefined') &&
                (typeof Int32Array !== 'undefined');


exports.assign = function (obj /*from1, from2, from3, ...*/) {
  var sources = Array.prototype.slice.call(arguments, 1);
  while (sources.length) {
    var source = sources.shift();
    if (!source) { continue; }

    if (typeof(source) !== 'object') {
      throw new TypeError(source + 'must be non-object');
    }

    for (var p in source) {
      if (source.hasOwnProperty(p)) {
        obj[p] = source[p];
      }
    }
  }

  return obj;
};


// reduce buffer size, avoiding mem copy
exports.shrinkBuf = function (buf, size) {
  if (buf.length === size) { return buf; }
  if (buf.subarray) { return buf.subarray(0, size); }
  buf.length = size;
  return buf;
};


var fnTyped = {
  arraySet: function (dest, src, src_offs, len, dest_offs) {
    if (src.subarray && dest.subarray) {
      dest.set(src.subarray(src_offs, src_offs+len), dest_offs);
      return;
    }
    // Fallback to ordinary array
    for(var i=0; i<len; i++) {
      dest[dest_offs + i] = src[src_offs + i];
    }
  },
  // Join array of chunks to single array.
  flattenChunks: function(chunks) {
    var i, l, len, pos, chunk, result;

    // calculate data length
    len = 0;
    for (i=0, l=chunks.length; i<l; i++) {
      len += chunks[i].length;
    }

    // join chunks
    result = new Uint8Array(len);
    pos = 0;
    for (i=0, l=chunks.length; i<l; i++) {
      chunk = chunks[i];
      result.set(chunk, pos);
      pos += chunk.length;
    }

    return result;
  }
};

var fnUntyped = {
  arraySet: function (dest, src, src_offs, len, dest_offs) {
    for(var i=0; i<len; i++) {
      dest[dest_offs + i] = src[src_offs + i];
    }
  },
  // Join array of chunks to single array.
  flattenChunks: function(chunks) {
    return [].concat.apply([], chunks);
  }
};


// Enable/Disable typed arrays use, for testing
//
exports.setTyped = function (on) {
  if (on) {
    exports.Buf8  = Uint8Array;
    exports.Buf16 = Uint16Array;
    exports.Buf32 = Int32Array;
    exports.assign(exports, fnTyped);
  } else {
    exports.Buf8  = Array;
    exports.Buf16 = Array;
    exports.Buf32 = Array;
    exports.assign(exports, fnUntyped);
  }
};

exports.setTyped(TYPED_OK);
},{}],3:[function(_dereq_,module,exports){
// String encode/decode helpers
'use strict';


var utils = _dereq_('./common');


// Quick check if we can use fast array to bin string conversion
//
// - apply(Array) can fail on Android 2.2
// - apply(Uint8Array) can fail on iOS 5.1 Safary
//
var STR_APPLY_OK = true;
var STR_APPLY_UIA_OK = true;

try { String.fromCharCode.apply(null, [0]); } catch(__) { STR_APPLY_OK = false; }
try { String.fromCharCode.apply(null, new Uint8Array(1)); } catch(__) { STR_APPLY_UIA_OK = false; }


// Table with utf8 lengths (calculated by first byte of sequence)
// Note, that 5 & 6-byte values and some 4-byte values can not be represented in JS,
// because max possible codepoint is 0x10ffff
var _utf8len = new utils.Buf8(256);
for (var i=0; i<256; i++) {
  _utf8len[i] = (i >= 252 ? 6 : i >= 248 ? 5 : i >= 240 ? 4 : i >= 224 ? 3 : i >= 192 ? 2 : 1);
}
_utf8len[254]=_utf8len[254]=1; // Invalid sequence start


// convert string to array (typed, when possible)
exports.string2buf = function (str) {
  var buf, c, c2, m_pos, i, str_len = str.length, buf_len = 0;

  // count binary size
  for (m_pos = 0; m_pos < str_len; m_pos++) {
    c = str.charCodeAt(m_pos);
    if ((c & 0xfc00) === 0xd800 && (m_pos+1 < str_len)) {
      c2 = str.charCodeAt(m_pos+1);
      if ((c2 & 0xfc00) === 0xdc00) {
        c = 0x10000 + ((c - 0xd800) << 10) + (c2 - 0xdc00);
        m_pos++;
      }
    }
    buf_len += c < 0x80 ? 1 : c < 0x800 ? 2 : c < 0x10000 ? 3 : 4;
  }

  // allocate buffer
  buf = new utils.Buf8(buf_len);

  // convert
  for (i=0, m_pos = 0; i < buf_len; m_pos++) {
    c = str.charCodeAt(m_pos);
    if ((c & 0xfc00) === 0xd800 && (m_pos+1 < str_len)) {
      c2 = str.charCodeAt(m_pos+1);
      if ((c2 & 0xfc00) === 0xdc00) {
        c = 0x10000 + ((c - 0xd800) << 10) + (c2 - 0xdc00);
        m_pos++;
      }
    }
    if (c < 0x80) {
      /* one byte */
      buf[i++] = c;
    } else if (c < 0x800) {
      /* two bytes */
      buf[i++] = 0xC0 | (c >>> 6);
      buf[i++] = 0x80 | (c & 0x3f);
    } else if (c < 0x10000) {
      /* three bytes */
      buf[i++] = 0xE0 | (c >>> 12);
      buf[i++] = 0x80 | (c >>> 6 & 0x3f);
      buf[i++] = 0x80 | (c & 0x3f);
    } else {
      /* four bytes */
      buf[i++] = 0xf0 | (c >>> 18);
      buf[i++] = 0x80 | (c >>> 12 & 0x3f);
      buf[i++] = 0x80 | (c >>> 6 & 0x3f);
      buf[i++] = 0x80 | (c & 0x3f);
    }
  }

  return buf;
};

// Helper (used in 2 places)
function buf2binstring(buf, len) {
  // use fallback for big arrays to avoid stack overflow
  if (len < 65537) {
    if ((buf.subarray && STR_APPLY_UIA_OK) || (!buf.subarray && STR_APPLY_OK)) {
      return String.fromCharCode.apply(null, utils.shrinkBuf(buf, len));
    }
  }

  var result = '';
  for(var i=0; i < len; i++) {
    result += String.fromCharCode(buf[i]);
  }
  return result;
}


// Convert byte array to binary string
exports.buf2binstring = function(buf) {
  return buf2binstring(buf, buf.length);
};


// Convert binary string (typed, when possible)
exports.binstring2buf = function(str) {
  var buf = new utils.Buf8(str.length);
  for(var i=0, len=buf.length; i < len; i++) {
    buf[i] = str.charCodeAt(i);
  }
  return buf;
};


// convert array to string
exports.buf2string = function (buf, max) {
  var i, out, c, c_len;
  var len = max || buf.length;

  // Reserve max possible length (2 words per char)
  // NB: by unknown reasons, Array is significantly faster for
  //     String.fromCharCode.apply than Uint16Array.
  var utf16buf = new Array(len*2);

  for (out=0, i=0; i<len;) {
    c = buf[i++];
    // quick process ascii
    if (c < 0x80) { utf16buf[out++] = c; continue; }

    c_len = _utf8len[c];
    // skip 5 & 6 byte codes
    if (c_len > 4) { utf16buf[out++] = 0xfffd; i += c_len-1; continue; }

    // apply mask on first byte
    c &= c_len === 2 ? 0x1f : c_len === 3 ? 0x0f : 0x07;
    // join the rest
    while (c_len > 1 && i < len) {
      c = (c << 6) | (buf[i++] & 0x3f);
      c_len--;
    }

    // terminated by end of string?
    if (c_len > 1) { utf16buf[out++] = 0xfffd; continue; }

    if (c < 0x10000) {
      utf16buf[out++] = c;
    } else {
      c -= 0x10000;
      utf16buf[out++] = 0xd800 | ((c >> 10) & 0x3ff);
      utf16buf[out++] = 0xdc00 | (c & 0x3ff);
    }
  }

  return buf2binstring(utf16buf, out);
};


// Calculate max possible position in utf8 buffer,
// that will not break sequence. If that's not possible
// - (very small limits) return max size as is.
//
// buf[] - utf8 bytes array
// max   - length limit (mandatory);
exports.utf8border = function(buf, max) {
  var pos;

  max = max || buf.length;
  if (max > buf.length) { max = buf.length; }

  // go back from last position, until start of sequence found
  pos = max-1;
  while (pos >= 0 && (buf[pos] & 0xC0) === 0x80) { pos--; }

  // Fuckup - very small and broken sequence,
  // return max, because we should return something anyway.
  if (pos < 0) { return max; }

  // If we came to start of buffer - that means vuffer is too small,
  // return max too.
  if (pos === 0) { return max; }

  return (pos + _utf8len[buf[pos]] > max) ? pos : max;
};

},{"./common":2}],4:[function(_dereq_,module,exports){
'use strict';

// Note: adler32 takes 12% for level 0 and 2% for level 6.
// It doesn't worth to make additional optimizationa as in original.
// Small size is preferable.

function adler32(adler, buf, len, pos) {
  var s1 = (adler & 0xffff) |0
    , s2 = ((adler >>> 16) & 0xffff) |0
    , n = 0;

  while (len !== 0) {
    // Set limit ~ twice less than 5552, to keep
    // s2 in 31-bits, because we force signed ints.
    // in other case %= will fail.
    n = len > 2000 ? 2000 : len;
    len -= n;

    do {
      s1 = (s1 + buf[pos++]) |0;
      s2 = (s2 + s1) |0;
    } while (--n);

    s1 %= 65521;
    s2 %= 65521;
  }

  return (s1 | (s2 << 16)) |0;
}


module.exports = adler32;
},{}],5:[function(_dereq_,module,exports){
module.exports = {

  /* Allowed flush values; see deflate() and inflate() below for details */
  Z_NO_FLUSH:         0,
  Z_PARTIAL_FLUSH:    1,
  Z_SYNC_FLUSH:       2,
  Z_FULL_FLUSH:       3,
  Z_FINISH:           4,
  Z_BLOCK:            5,
  Z_TREES:            6,

  /* Return codes for the compression/decompression functions. Negative values
  * are errors, positive values are used for special but normal events.
  */
  Z_OK:               0,
  Z_STREAM_END:       1,
  Z_NEED_DICT:        2,
  Z_ERRNO:           -1,
  Z_STREAM_ERROR:    -2,
  Z_DATA_ERROR:      -3,
  //Z_MEM_ERROR:     -4,
  Z_BUF_ERROR:       -5,
  //Z_VERSION_ERROR: -6,

  /* compression levels */
  Z_NO_COMPRESSION:         0,
  Z_BEST_SPEED:             1,
  Z_BEST_COMPRESSION:       9,
  Z_DEFAULT_COMPRESSION:   -1,


  Z_FILTERED:               1,
  Z_HUFFMAN_ONLY:           2,
  Z_RLE:                    3,
  Z_FIXED:                  4,
  Z_DEFAULT_STRATEGY:       0,

  /* Possible values of the data_type field (though see inflate()) */
  Z_BINARY:                 0,
  Z_TEXT:                   1,
  //Z_ASCII:                1, // = Z_TEXT (deprecated)
  Z_UNKNOWN:                2,

  /* The deflate compression method */
  Z_DEFLATED:               8
  //Z_NULL:                 null // Use -1 or null inline, depending on var type
};
},{}],6:[function(_dereq_,module,exports){
'use strict';

// Note: we can't get significant speed boost here.
// So write code to minimize size - no pregenerated tables
// and array tools dependencies.


// Use ordinary array, since untyped makes no boost here
function makeTable() {
  var c, table = [];

  for(var n =0; n < 256; n++){
    c = n;
    for(var k =0; k < 8; k++){
      c = ((c&1) ? (0xEDB88320 ^ (c >>> 1)) : (c >>> 1));
    }
    table[n] = c;
  }

  return table;
}

// Create table on load. Just 255 signed longs. Not a problem.
var crcTable = makeTable();


function crc32(crc, buf, len, pos) {
  var t = crcTable
    , end = pos + len;

  crc = crc ^ (-1);

  for (var i = pos; i < end; i++ ) {
    crc = (crc >>> 8) ^ t[(crc ^ buf[i]) & 0xFF];
  }

  return (crc ^ (-1)); // >>> 0;
}


module.exports = crc32;
},{}],7:[function(_dereq_,module,exports){
'use strict';


function GZheader() {
  /* true if compressed data believed to be text */
  this.text       = 0;
  /* modification time */
  this.time       = 0;
  /* extra flags (not used when writing a gzip file) */
  this.xflags     = 0;
  /* operating system */
  this.os         = 0;
  /* pointer to extra field or Z_NULL if none */
  this.extra      = null;
  /* extra field length (valid if extra != Z_NULL) */
  this.extra_len  = 0; // Actually, we don't need it in JS,
                       // but leave for few code modifications

  //
  // Setup limits is not necessary because in js we should not preallocate memory 
  // for inflate use constant limit in 65536 bytes
  //

  /* space at extra (only when reading header) */
  // this.extra_max  = 0;
  /* pointer to zero-terminated file name or Z_NULL */
  this.name       = '';
  /* space at name (only when reading header) */
  // this.name_max   = 0;
  /* pointer to zero-terminated comment or Z_NULL */
  this.comment    = '';
  /* space at comment (only when reading header) */
  // this.comm_max   = 0;
  /* true if there was or will be a header crc */
  this.hcrc       = 0;
  /* true when done reading gzip header (not used when writing a gzip file) */
  this.done       = false;
}

module.exports = GZheader;
},{}],8:[function(_dereq_,module,exports){
'use strict';

// See state defs from inflate.js
var BAD = 30;       /* got a data error -- remain here until reset */
var TYPE = 12;      /* i: waiting for type bits, including last-flag bit */

/*
   Decode literal, length, and distance codes and write out the resulting
   literal and match bytes until either not enough input or output is
   available, an end-of-block is encountered, or a data error is encountered.
   When large enough input and output buffers are supplied to inflate(), for
   example, a 16K input buffer and a 64K output buffer, more than 95% of the
   inflate execution time is spent in this routine.

   Entry assumptions:

        state.mode === LEN
        strm.avail_in >= 6
        strm.avail_out >= 258
        start >= strm.avail_out
        state.bits < 8

   On return, state.mode is one of:

        LEN -- ran out of enough output space or enough available input
        TYPE -- reached end of block code, inflate() to interpret next block
        BAD -- error in block data

   Notes:

    - The maximum input bits used by a length/distance pair is 15 bits for the
      length code, 5 bits for the length extra, 15 bits for the distance code,
      and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
      Therefore if strm.avail_in >= 6, then there is enough input to avoid
      checking for available input while decoding.

    - The maximum bytes that a single length/distance pair can output is 258
      bytes, which is the maximum length that can be coded.  inflate_fast()
      requires strm.avail_out >= 258 for each loop to avoid checking for
      output space.
 */
module.exports = function inflate_fast(strm, start) {
  var state;
  var _in;                    /* local strm.input */
  var last;                   /* have enough input while in < last */
  var _out;                   /* local strm.output */
  var beg;                    /* inflate()'s initial strm.output */
  var end;                    /* while out < end, enough space available */
//#ifdef INFLATE_STRICT
  var dmax;                   /* maximum distance from zlib header */
//#endif
  var wsize;                  /* window size or zero if not using window */
  var whave;                  /* valid bytes in the window */
  var wnext;                  /* window write index */
  var window;                 /* allocated sliding window, if wsize != 0 */
  var hold;                   /* local strm.hold */
  var bits;                   /* local strm.bits */
  var lcode;                  /* local strm.lencode */
  var dcode;                  /* local strm.distcode */
  var lmask;                  /* mask for first level of length codes */
  var dmask;                  /* mask for first level of distance codes */
  var here;                   /* retrieved table entry */
  var op;                     /* code bits, operation, extra bits, or */
                              /*  window position, window bytes to copy */
  var len;                    /* match length, unused bytes */
  var dist;                   /* match distance */
  var from;                   /* where to copy match from */
  var from_source;


  var input, output; // JS specific, because we have no pointers

  /* copy state to local variables */
  state = strm.state;
  //here = state.here;
  _in = strm.next_in;
  input = strm.input;
  last = _in + (strm.avail_in - 5);
  _out = strm.next_out;
  output = strm.output;
  beg = _out - (start - strm.avail_out);
  end = _out + (strm.avail_out - 257);
//#ifdef INFLATE_STRICT
  dmax = state.dmax;
//#endif
  wsize = state.wsize;
  whave = state.whave;
  wnext = state.wnext;
  window = state.window;
  hold = state.hold;
  bits = state.bits;
  lcode = state.lencode;
  dcode = state.distcode;
  lmask = (1 << state.lenbits) - 1;
  dmask = (1 << state.distbits) - 1;


  /* decode literals and length/distances until end-of-block or not enough
     input data or output space */

  top:
  do {
    if (bits < 15) {
      hold += input[_in++] << bits;
      bits += 8;
      hold += input[_in++] << bits;
      bits += 8;
    }

    here = lcode[hold & lmask];

    dolen:
    for (;;) { // Goto emulation
      op = here >>> 24/*here.bits*/;
      hold >>>= op;
      bits -= op;
      op = (here >>> 16) & 0xff/*here.op*/;
      if (op === 0) {                          /* literal */
        //Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
        //        "inflate:         literal '%c'\n" :
        //        "inflate:         literal 0x%02x\n", here.val));
        output[_out++] = here & 0xffff/*here.val*/;
      }
      else if (op & 16) {                     /* length base */
        len = here & 0xffff/*here.val*/;
        op &= 15;                           /* number of extra bits */
        if (op) {
          if (bits < op) {
            hold += input[_in++] << bits;
            bits += 8;
          }
          len += hold & ((1 << op) - 1);
          hold >>>= op;
          bits -= op;
        }
        //Tracevv((stderr, "inflate:         length %u\n", len));
        if (bits < 15) {
          hold += input[_in++] << bits;
          bits += 8;
          hold += input[_in++] << bits;
          bits += 8;
        }
        here = dcode[hold & dmask];

        dodist:
        for (;;) { // goto emulation
          op = here >>> 24/*here.bits*/;
          hold >>>= op;
          bits -= op;
          op = (here >>> 16) & 0xff/*here.op*/;

          if (op & 16) {                      /* distance base */
            dist = here & 0xffff/*here.val*/;
            op &= 15;                       /* number of extra bits */
            if (bits < op) {
              hold += input[_in++] << bits;
              bits += 8;
              if (bits < op) {
                hold += input[_in++] << bits;
                bits += 8;
              }
            }
            dist += hold & ((1 << op) - 1);
//#ifdef INFLATE_STRICT
            if (dist > dmax) {
              strm.msg = 'invalid distance too far back';
              state.mode = BAD;
              break top;
            }
//#endif
            hold >>>= op;
            bits -= op;
            //Tracevv((stderr, "inflate:         distance %u\n", dist));
            op = _out - beg;                /* max distance in output */
            if (dist > op) {                /* see if copy from window */
              op = dist - op;               /* distance back in window */
              if (op > whave) {
                if (state.sane) {
                  strm.msg = 'invalid distance too far back';
                  state.mode = BAD;
                  break top;
                }

// (!) This block is disabled in zlib defailts,
// don't enable it for binary compatibility
//#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
//                if (len <= op - whave) {
//                  do {
//                    output[_out++] = 0;
//                  } while (--len);
//                  continue top;
//                }
//                len -= op - whave;
//                do {
//                  output[_out++] = 0;
//                } while (--op > whave);
//                if (op === 0) {
//                  from = _out - dist;
//                  do {
//                    output[_out++] = output[from++];
//                  } while (--len);
//                  continue top;
//                }
//#endif
              }
              from = 0; // window index
              from_source = window;
              if (wnext === 0) {           /* very common case */
                from += wsize - op;
                if (op < len) {         /* some from window */
                  len -= op;
                  do {
                    output[_out++] = window[from++];
                  } while (--op);
                  from = _out - dist;  /* rest from output */
                  from_source = output;
                }
              }
              else if (wnext < op) {      /* wrap around window */
                from += wsize + wnext - op;
                op -= wnext;
                if (op < len) {         /* some from end of window */
                  len -= op;
                  do {
                    output[_out++] = window[from++];
                  } while (--op);
                  from = 0;
                  if (wnext < len) {  /* some from start of window */
                    op = wnext;
                    len -= op;
                    do {
                      output[_out++] = window[from++];
                    } while (--op);
                    from = _out - dist;      /* rest from output */
                    from_source = output;
                  }
                }
              }
              else {                      /* contiguous in window */
                from += wnext - op;
                if (op < len) {         /* some from window */
                  len -= op;
                  do {
                    output[_out++] = window[from++];
                  } while (--op);
                  from = _out - dist;  /* rest from output */
                  from_source = output;
                }
              }
              while (len > 2) {
                output[_out++] = from_source[from++];
                output[_out++] = from_source[from++];
                output[_out++] = from_source[from++];
                len -= 3;
              }
              if (len) {
                output[_out++] = from_source[from++];
                if (len > 1) {
                  output[_out++] = from_source[from++];
                }
              }
            }
            else {
              from = _out - dist;          /* copy direct from output */
              do {                        /* minimum length is three */
                output[_out++] = output[from++];
                output[_out++] = output[from++];
                output[_out++] = output[from++];
                len -= 3;
              } while (len > 2);
              if (len) {
                output[_out++] = output[from++];
                if (len > 1) {
                  output[_out++] = output[from++];
                }
              }
            }
          }
          else if ((op & 64) === 0) {          /* 2nd level distance code */
            here = dcode[(here & 0xffff)/*here.val*/ + (hold & ((1 << op) - 1))];
            continue dodist;
          }
          else {
            strm.msg = 'invalid distance code';
            state.mode = BAD;
            break top;
          }

          break; // need to emulate goto via "continue"
        }
      }
      else if ((op & 64) === 0) {              /* 2nd level length code */
        here = lcode[(here & 0xffff)/*here.val*/ + (hold & ((1 << op) - 1))];
        continue dolen;
      }
      else if (op & 32) {                     /* end-of-block */
        //Tracevv((stderr, "inflate:         end of block\n"));
        state.mode = TYPE;
        break top;
      }
      else {
        strm.msg = 'invalid literal/length code';
        state.mode = BAD;
        break top;
      }

      break; // need to emulate goto via "continue"
    }
  } while (_in < last && _out < end);

  /* return unused bytes (on entry, bits < 8, so in won't go too far back) */
  len = bits >> 3;
  _in -= len;
  bits -= len << 3;
  hold &= (1 << bits) - 1;

  /* update state and return */
  strm.next_in = _in;
  strm.next_out = _out;
  strm.avail_in = (_in < last ? 5 + (last - _in) : 5 - (_in - last));
  strm.avail_out = (_out < end ? 257 + (end - _out) : 257 - (_out - end));
  state.hold = hold;
  state.bits = bits;
  return;
};

},{}],9:[function(_dereq_,module,exports){
'use strict';


var utils = _dereq_('../utils/common');
var adler32 = _dereq_('./adler32');
var crc32   = _dereq_('./crc32');
var inflate_fast = _dereq_('./inffast');
var inflate_table = _dereq_('./inftrees');

var CODES = 0;
var LENS = 1;
var DISTS = 2;

/* Public constants ==========================================================*/
/* ===========================================================================*/


/* Allowed flush values; see deflate() and inflate() below for details */
//var Z_NO_FLUSH      = 0;
//var Z_PARTIAL_FLUSH = 1;
//var Z_SYNC_FLUSH    = 2;
//var Z_FULL_FLUSH    = 3;
var Z_FINISH        = 4;
var Z_BLOCK         = 5;
var Z_TREES         = 6;


/* Return codes for the compression/decompression functions. Negative values
 * are errors, positive values are used for special but normal events.
 */
var Z_OK            = 0;
var Z_STREAM_END    = 1;
var Z_NEED_DICT     = 2;
//var Z_ERRNO         = -1;
var Z_STREAM_ERROR  = -2;
var Z_DATA_ERROR    = -3;
var Z_MEM_ERROR     = -4;
var Z_BUF_ERROR     = -5;
//var Z_VERSION_ERROR = -6;

/* The deflate compression method */
var Z_DEFLATED  = 8;


/* STATES ====================================================================*/
/* ===========================================================================*/


var    HEAD = 1;       /* i: waiting for magic header */
var    FLAGS = 2;      /* i: waiting for method and flags (gzip) */
var    TIME = 3;       /* i: waiting for modification time (gzip) */
var    OS = 4;         /* i: waiting for extra flags and operating system (gzip) */
var    EXLEN = 5;      /* i: waiting for extra length (gzip) */
var    EXTRA = 6;      /* i: waiting for extra bytes (gzip) */
var    NAME = 7;       /* i: waiting for end of file name (gzip) */
var    COMMENT = 8;    /* i: waiting for end of comment (gzip) */
var    HCRC = 9;       /* i: waiting for header crc (gzip) */
var    DICTID = 10;    /* i: waiting for dictionary check value */
var    DICT = 11;      /* waiting for inflateSetDictionary() call */
var        TYPE = 12;      /* i: waiting for type bits, including last-flag bit */
var        TYPEDO = 13;    /* i: same, but skip check to exit inflate on new block */
var        STORED = 14;    /* i: waiting for stored size (length and complement) */
var        COPY_ = 15;     /* i/o: same as COPY below, but only first time in */
var        COPY = 16;      /* i/o: waiting for input or output to copy stored block */
var        TABLE = 17;     /* i: waiting for dynamic block table lengths */
var        LENLENS = 18;   /* i: waiting for code length code lengths */
var        CODELENS = 19;  /* i: waiting for length/lit and distance code lengths */
var            LEN_ = 20;      /* i: same as LEN below, but only first time in */
var            LEN = 21;       /* i: waiting for length/lit/eob code */
var            LENEXT = 22;    /* i: waiting for length extra bits */
var            DIST = 23;      /* i: waiting for distance code */
var            DISTEXT = 24;   /* i: waiting for distance extra bits */
var            MATCH = 25;     /* o: waiting for output space to copy string */
var            LIT = 26;       /* o: waiting for output space to write literal */
var    CHECK = 27;     /* i: waiting for 32-bit check value */
var    LENGTH = 28;    /* i: waiting for 32-bit length (gzip) */
var    DONE = 29;      /* finished check, done -- remain here until reset */
var    BAD = 30;       /* got a data error -- remain here until reset */
var    MEM = 31;       /* got an inflate() memory error -- remain here until reset */
var    SYNC = 32;      /* looking for synchronization bytes to restart inflate() */

/* ===========================================================================*/



var ENOUGH_LENS = 852;
var ENOUGH_DISTS = 592;
//var ENOUGH =  (ENOUGH_LENS+ENOUGH_DISTS);

var MAX_WBITS = 15;
/* 32K LZ77 window */
var DEF_WBITS = MAX_WBITS;


function ZSWAP32(q) {
  return  (((q >>> 24) & 0xff) +
          ((q >>> 8) & 0xff00) +
          ((q & 0xff00) << 8) +
          ((q & 0xff) << 24));
}


function InflateState() {
  this.mode = 0;             /* current inflate mode */
  this.last = false;          /* true if processing last block */
  this.wrap = 0;              /* bit 0 true for zlib, bit 1 true for gzip */
  this.havedict = false;      /* true if dictionary provided */
  this.flags = 0;             /* gzip header method and flags (0 if zlib) */
  this.dmax = 0;              /* zlib header max distance (INFLATE_STRICT) */
  this.check = 0;             /* protected copy of check value */
  this.total = 0;             /* protected copy of output count */
  // TODO: may be {}
  this.head = null;           /* where to save gzip header information */

  /* sliding window */
  this.wbits = 0;             /* log base 2 of requested window size */
  this.wsize = 0;             /* window size or zero if not using window */
  this.whave = 0;             /* valid bytes in the window */
  this.wnext = 0;             /* window write index */
  this.window = null;         /* allocated sliding window, if needed */

  /* bit accumulator */
  this.hold = 0;              /* input bit accumulator */
  this.bits = 0;              /* number of bits in "in" */

  /* for string and stored block copying */
  this.length = 0;            /* literal or length of data to copy */
  this.offset = 0;            /* distance back to copy string from */

  /* for table and code decoding */
  this.extra = 0;             /* extra bits needed */

  /* fixed and dynamic code tables */
  this.lencode = null;          /* starting table for length/literal codes */
  this.distcode = null;         /* starting table for distance codes */
  this.lenbits = 0;           /* index bits for lencode */
  this.distbits = 0;          /* index bits for distcode */

  /* dynamic table building */
  this.ncode = 0;             /* number of code length code lengths */
  this.nlen = 0;              /* number of length code lengths */
  this.ndist = 0;             /* number of distance code lengths */
  this.have = 0;              /* number of code lengths in lens[] */
  this.next = null;              /* next available space in codes[] */

  this.lens = new utils.Buf16(320); /* temporary storage for code lengths */
  this.work = new utils.Buf16(288); /* work area for code table building */

  /*
   because we don't have pointers in js, we use lencode and distcode directly
   as buffers so we don't need codes
  */
  //this.codes = new utils.Buf32(ENOUGH);       /* space for code tables */
  this.lendyn = null;              /* dynamic table for length/literal codes (JS specific) */
  this.distdyn = null;             /* dynamic table for distance codes (JS specific) */
  this.sane = 0;                   /* if false, allow invalid distance too far */
  this.back = 0;                   /* bits back of last unprocessed length/lit */
  this.was = 0;                    /* initial length of match */
}

function inflateResetKeep(strm) {
  var state;

  if (!strm || !strm.state) { return Z_STREAM_ERROR; }
  state = strm.state;
  strm.total_in = strm.total_out = state.total = 0;
  strm.msg = ''; /*Z_NULL*/
  if (state.wrap) {       /* to support ill-conceived Java test suite */
    strm.adler = state.wrap & 1;
  }
  state.mode = HEAD;
  state.last = 0;
  state.havedict = 0;
  state.dmax = 32768;
  state.head = null/*Z_NULL*/;
  state.hold = 0;
  state.bits = 0;
  //state.lencode = state.distcode = state.next = state.codes;
  state.lencode = state.lendyn = new utils.Buf32(ENOUGH_LENS);
  state.distcode = state.distdyn = new utils.Buf32(ENOUGH_DISTS);

  state.sane = 1;
  state.back = -1;
  //Tracev((stderr, "inflate: reset\n"));
  return Z_OK;
}

function inflateReset(strm) {
  var state;

  if (!strm || !strm.state) { return Z_STREAM_ERROR; }
  state = strm.state;
  state.wsize = 0;
  state.whave = 0;
  state.wnext = 0;
  return inflateResetKeep(strm);

}

function inflateReset2(strm, windowBits) {
  var wrap;
  var state;

  /* get the state */
  if (!strm || !strm.state) { return Z_STREAM_ERROR; }
  state = strm.state;

  /* extract wrap request from windowBits parameter */
  if (windowBits < 0) {
    wrap = 0;
    windowBits = -windowBits;
  }
  else {
    wrap = (windowBits >> 4) + 1;
    if (windowBits < 48) {
      windowBits &= 15;
    }
  }

  /* set number of window bits, free window if different */
  if (windowBits && (windowBits < 8 || windowBits > 15)) {
    return Z_STREAM_ERROR;
  }
  if (state.window !== null && state.wbits !== windowBits) {
    state.window = null;
  }

  /* update state and reset the rest of it */
  state.wrap = wrap;
  state.wbits = windowBits;
  return inflateReset(strm);
}

function inflateInit2(strm, windowBits) {
  var ret;
  var state;

  if (!strm) { return Z_STREAM_ERROR; }
  //strm.msg = Z_NULL;                 /* in case we return an error */

  state = new InflateState();

  //if (state === Z_NULL) return Z_MEM_ERROR;
  //Tracev((stderr, "inflate: allocated\n"));
  strm.state = state;
  state.window = null/*Z_NULL*/;
  ret = inflateReset2(strm, windowBits);
  if (ret !== Z_OK) {
    strm.state = null/*Z_NULL*/;
  }
  return ret;
}

function inflateInit(strm) {
  return inflateInit2(strm, DEF_WBITS);
}


/*
 Return state with length and distance decoding tables and index sizes set to
 fixed code decoding.  Normally this returns fixed tables from inffixed.h.
 If BUILDFIXED is defined, then instead this routine builds the tables the
 first time it's called, and returns those tables the first time and
 thereafter.  This reduces the size of the code by about 2K bytes, in
 exchange for a little execution time.  However, BUILDFIXED should not be
 used for threaded applications, since the rewriting of the tables and virgin
 may not be thread-safe.
 */
var virgin = true;

var lenfix, distfix; // We have no pointers in JS, so keep tables separate

function fixedtables(state) {
  /* build fixed huffman tables if first call (may not be thread safe) */
  if (virgin) {
    var sym;

    lenfix = new utils.Buf32(512);
    distfix = new utils.Buf32(32);

    /* literal/length table */
    sym = 0;
    while (sym < 144) { state.lens[sym++] = 8; }
    while (sym < 256) { state.lens[sym++] = 9; }
    while (sym < 280) { state.lens[sym++] = 7; }
    while (sym < 288) { state.lens[sym++] = 8; }

    inflate_table(LENS,  state.lens, 0, 288, lenfix,   0, state.work, {bits: 9});

    /* distance table */
    sym = 0;
    while (sym < 32) { state.lens[sym++] = 5; }

    inflate_table(DISTS, state.lens, 0, 32,   distfix, 0, state.work, {bits: 5});

    /* do this just once */
    virgin = false;
  }

  state.lencode = lenfix;
  state.lenbits = 9;
  state.distcode = distfix;
  state.distbits = 5;
}


/*
 Update the window with the last wsize (normally 32K) bytes written before
 returning.  If window does not exist yet, create it.  This is only called
 when a window is already in use, or when output has been written during this
 inflate call, but the end of the deflate stream has not been reached yet.
 It is also called to create a window for dictionary data when a dictionary
 is loaded.

 Providing output buffers larger than 32K to inflate() should provide a speed
 advantage, since only the last 32K of output is copied to the sliding window
 upon return from inflate(), and since all distances after the first 32K of
 output will fall in the output data, making match copies simpler and faster.
 The advantage may be dependent on the size of the processor's data caches.
 */
function updatewindow(strm, src, end, copy) {
  var dist;
  var state = strm.state;

  /* if it hasn't been done already, allocate space for the window */
  if (state.window === null) {
    state.wsize = 1 << state.wbits;
    state.wnext = 0;
    state.whave = 0;

    state.window = new utils.Buf8(state.wsize);
  }

  /* copy state->wsize or less output bytes into the circular window */
  if (copy >= state.wsize) {
    utils.arraySet(state.window,src, end - state.wsize, state.wsize, 0);
    state.wnext = 0;
    state.whave = state.wsize;
  }
  else {
    dist = state.wsize - state.wnext;
    if (dist > copy) {
      dist = copy;
    }
    //zmemcpy(state->window + state->wnext, end - copy, dist);
    utils.arraySet(state.window,src, end - copy, dist, state.wnext);
    copy -= dist;
    if (copy) {
      //zmemcpy(state->window, end - copy, copy);
      utils.arraySet(state.window,src, end - copy, copy, 0);
      state.wnext = copy;
      state.whave = state.wsize;
    }
    else {
      state.wnext += dist;
      if (state.wnext === state.wsize) { state.wnext = 0; }
      if (state.whave < state.wsize) { state.whave += dist; }
    }
  }
  return 0;
}

function inflate(strm, flush) {
  var state;
  var input, output;          // input/output buffers
  var next;                   /* next input INDEX */
  var put;                    /* next output INDEX */
  var have, left;             /* available input and output */
  var hold;                   /* bit buffer */
  var bits;                   /* bits in bit buffer */
  var _in, _out;              /* save starting available input and output */
  var copy;                   /* number of stored or match bytes to copy */
  var from;                   /* where to copy match bytes from */
  var from_source;
  var here = 0;               /* current decoding table entry */
  var here_bits, here_op, here_val; // paked "here" denormalized (JS specific)
  //var last;                   /* parent table entry */
  var last_bits, last_op, last_val; // paked "last" denormalized (JS specific)
  var len;                    /* length to copy for repeats, bits to drop */
  var ret;                    /* return code */
  var hbuf = new utils.Buf8(4);    /* buffer for gzip header crc calculation */
  var opts;

  var n; // temporary var for NEED_BITS

  var order = /* permutation of code lengths */
    [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15];


  if (!strm || !strm.state || !strm.output ||
      (!strm.input && strm.avail_in !== 0)) {
    return Z_STREAM_ERROR;
  }

  state = strm.state;
  if (state.mode === TYPE) { state.mode = TYPEDO; }    /* skip check */


  //--- LOAD() ---
  put = strm.next_out;
  output = strm.output;
  left = strm.avail_out;
  next = strm.next_in;
  input = strm.input;
  have = strm.avail_in;
  hold = state.hold;
  bits = state.bits;
  //---

  _in = have;
  _out = left;
  ret = Z_OK;

  inf_leave: // goto emulation
  for (;;) {
    switch (state.mode) {
    case HEAD:
      if (state.wrap === 0) {
        state.mode = TYPEDO;
        break;
      }
      //=== NEEDBITS(16);
      while (bits < 16) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      if ((state.wrap & 2) && hold === 0x8b1f) {  /* gzip header */
        state.check = 0/*crc32(0L, Z_NULL, 0)*/;
        //=== CRC2(state.check, hold);
        hbuf[0] = hold & 0xff;
        hbuf[1] = (hold >>> 8) & 0xff;
        state.check = crc32(state.check, hbuf, 2, 0);
        //===//

        //=== INITBITS();
        hold = 0;
        bits = 0;
        //===//
        state.mode = FLAGS;
        break;
      }
      state.flags = 0;           /* expect zlib header */
      if (state.head) {
        state.head.done = false;
      }
      if (!(state.wrap & 1) ||   /* check if zlib header allowed */
        (((hold & 0xff)/*BITS(8)*/ << 8) + (hold >> 8)) % 31) {
        strm.msg = 'incorrect header check';
        state.mode = BAD;
        break;
      }
      if ((hold & 0x0f)/*BITS(4)*/ !== Z_DEFLATED) {
        strm.msg = 'unknown compression method';
        state.mode = BAD;
        break;
      }
      //--- DROPBITS(4) ---//
      hold >>>= 4;
      bits -= 4;
      //---//
      len = (hold & 0x0f)/*BITS(4)*/ + 8;
      if (state.wbits === 0) {
        state.wbits = len;
      }
      else if (len > state.wbits) {
        strm.msg = 'invalid window size';
        state.mode = BAD;
        break;
      }
      state.dmax = 1 << len;
      //Tracev((stderr, "inflate:   zlib header ok\n"));
      strm.adler = state.check = 1/*adler32(0L, Z_NULL, 0)*/;
      state.mode = hold & 0x200 ? DICTID : TYPE;
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      break;
    case FLAGS:
      //=== NEEDBITS(16); */
      while (bits < 16) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      state.flags = hold;
      if ((state.flags & 0xff) !== Z_DEFLATED) {
        strm.msg = 'unknown compression method';
        state.mode = BAD;
        break;
      }
      if (state.flags & 0xe000) {
        strm.msg = 'unknown header flags set';
        state.mode = BAD;
        break;
      }
      if (state.head) {
        state.head.text = ((hold >> 8) & 1);
      }
      if (state.flags & 0x0200) {
        //=== CRC2(state.check, hold);
        hbuf[0] = hold & 0xff;
        hbuf[1] = (hold >>> 8) & 0xff;
        state.check = crc32(state.check, hbuf, 2, 0);
        //===//
      }
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      state.mode = TIME;
      /* falls through */
    case TIME:
      //=== NEEDBITS(32); */
      while (bits < 32) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      if (state.head) {
        state.head.time = hold;
      }
      if (state.flags & 0x0200) {
        //=== CRC4(state.check, hold)
        hbuf[0] = hold & 0xff;
        hbuf[1] = (hold >>> 8) & 0xff;
        hbuf[2] = (hold >>> 16) & 0xff;
        hbuf[3] = (hold >>> 24) & 0xff;
        state.check = crc32(state.check, hbuf, 4, 0);
        //===
      }
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      state.mode = OS;
      /* falls through */
    case OS:
      //=== NEEDBITS(16); */
      while (bits < 16) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      if (state.head) {
        state.head.xflags = (hold & 0xff);
        state.head.os = (hold >> 8);
      }
      if (state.flags & 0x0200) {
        //=== CRC2(state.check, hold);
        hbuf[0] = hold & 0xff;
        hbuf[1] = (hold >>> 8) & 0xff;
        state.check = crc32(state.check, hbuf, 2, 0);
        //===//
      }
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      state.mode = EXLEN;
      /* falls through */
    case EXLEN:
      if (state.flags & 0x0400) {
        //=== NEEDBITS(16); */
        while (bits < 16) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        state.length = hold;
        if (state.head) {
          state.head.extra_len = hold;
        }
        if (state.flags & 0x0200) {
          //=== CRC2(state.check, hold);
          hbuf[0] = hold & 0xff;
          hbuf[1] = (hold >>> 8) & 0xff;
          state.check = crc32(state.check, hbuf, 2, 0);
          //===//
        }
        //=== INITBITS();
        hold = 0;
        bits = 0;
        //===//
      }
      else if (state.head) {
        state.head.extra = null/*Z_NULL*/;
      }
      state.mode = EXTRA;
      /* falls through */
    case EXTRA:
      if (state.flags & 0x0400) {
        copy = state.length;
        if (copy > have) { copy = have; }
        if (copy) {
          if (state.head) {
            len = state.head.extra_len - state.length;
            if (!state.head.extra) {
              // Use untyped array for more conveniend processing later
              state.head.extra = new Array(state.head.extra_len);
            }
            utils.arraySet(
              state.head.extra,
              input,
              next,
              // extra field is limited to 65536 bytes
              // - no need for additional size check
              copy,
              /*len + copy > state.head.extra_max - len ? state.head.extra_max : copy,*/
              len
            );
            //zmemcpy(state.head.extra + len, next,
            //        len + copy > state.head.extra_max ?
            //        state.head.extra_max - len : copy);
          }
          if (state.flags & 0x0200) {
            state.check = crc32(state.check, input, copy, next);
          }
          have -= copy;
          next += copy;
          state.length -= copy;
        }
        if (state.length) { break inf_leave; }
      }
      state.length = 0;
      state.mode = NAME;
      /* falls through */
    case NAME:
      if (state.flags & 0x0800) {
        if (have === 0) { break inf_leave; }
        copy = 0;
        do {
          // TODO: 2 or 1 bytes?
          len = input[next + copy++];
          /* use constant limit because in js we should not preallocate memory */
          if (state.head && len &&
              (state.length < 65536 /*state.head.name_max*/)) {
            state.head.name += String.fromCharCode(len);
          }
        } while (len && copy < have);

        if (state.flags & 0x0200) {
          state.check = crc32(state.check, input, copy, next);
        }
        have -= copy;
        next += copy;
        if (len) { break inf_leave; }
      }
      else if (state.head) {
        state.head.name = null;
      }
      state.length = 0;
      state.mode = COMMENT;
      /* falls through */
    case COMMENT:
      if (state.flags & 0x1000) {
        if (have === 0) { break inf_leave; }
        copy = 0;
        do {
          len = input[next + copy++];
          /* use constant limit because in js we should not preallocate memory */
          if (state.head && len &&
              (state.length < 65536 /*state.head.comm_max*/)) {
            state.head.comment += String.fromCharCode(len);
          }
        } while (len && copy < have);
        if (state.flags & 0x0200) {
          state.check = crc32(state.check, input, copy, next);
        }
        have -= copy;
        next += copy;
        if (len) { break inf_leave; }
      }
      else if (state.head) {
        state.head.comment = null;
      }
      state.mode = HCRC;
      /* falls through */
    case HCRC:
      if (state.flags & 0x0200) {
        //=== NEEDBITS(16); */
        while (bits < 16) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        if (hold !== (state.check & 0xffff)) {
          strm.msg = 'header crc mismatch';
          state.mode = BAD;
          break;
        }
        //=== INITBITS();
        hold = 0;
        bits = 0;
        //===//
      }
      if (state.head) {
        state.head.hcrc = ((state.flags >> 9) & 1);
        state.head.done = true;
      }
      strm.adler = state.check = 0 /*crc32(0L, Z_NULL, 0)*/;
      state.mode = TYPE;
      break;
    case DICTID:
      //=== NEEDBITS(32); */
      while (bits < 32) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      strm.adler = state.check = ZSWAP32(hold);
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      state.mode = DICT;
      /* falls through */
    case DICT:
      if (state.havedict === 0) {
        //--- RESTORE() ---
        strm.next_out = put;
        strm.avail_out = left;
        strm.next_in = next;
        strm.avail_in = have;
        state.hold = hold;
        state.bits = bits;
        //---
        return Z_NEED_DICT;
      }
      strm.adler = state.check = 1/*adler32(0L, Z_NULL, 0)*/;
      state.mode = TYPE;
      /* falls through */
    case TYPE:
      if (flush === Z_BLOCK || flush === Z_TREES) { break inf_leave; }
      /* falls through */
    case TYPEDO:
      if (state.last) {
        //--- BYTEBITS() ---//
        hold >>>= bits & 7;
        bits -= bits & 7;
        //---//
        state.mode = CHECK;
        break;
      }
      //=== NEEDBITS(3); */
      while (bits < 3) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      state.last = (hold & 0x01)/*BITS(1)*/;
      //--- DROPBITS(1) ---//
      hold >>>= 1;
      bits -= 1;
      //---//

      switch ((hold & 0x03)/*BITS(2)*/) {
      case 0:                             /* stored block */
        //Tracev((stderr, "inflate:     stored block%s\n",
        //        state.last ? " (last)" : ""));
        state.mode = STORED;
        break;
      case 1:                             /* fixed block */
        fixedtables(state);
        //Tracev((stderr, "inflate:     fixed codes block%s\n",
        //        state.last ? " (last)" : ""));
        state.mode = LEN_;             /* decode codes */
        if (flush === Z_TREES) {
          //--- DROPBITS(2) ---//
          hold >>>= 2;
          bits -= 2;
          //---//
          break inf_leave;
        }
        break;
      case 2:                             /* dynamic block */
        //Tracev((stderr, "inflate:     dynamic codes block%s\n",
        //        state.last ? " (last)" : ""));
        state.mode = TABLE;
        break;
      case 3:
        strm.msg = 'invalid block type';
        state.mode = BAD;
      }
      //--- DROPBITS(2) ---//
      hold >>>= 2;
      bits -= 2;
      //---//
      break;
    case STORED:
      //--- BYTEBITS() ---// /* go to byte boundary */
      hold >>>= bits & 7;
      bits -= bits & 7;
      //---//
      //=== NEEDBITS(32); */
      while (bits < 32) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      if ((hold & 0xffff) !== ((hold >>> 16) ^ 0xffff)) {
        strm.msg = 'invalid stored block lengths';
        state.mode = BAD;
        break;
      }
      state.length = hold & 0xffff;
      //Tracev((stderr, "inflate:       stored length %u\n",
      //        state.length));
      //=== INITBITS();
      hold = 0;
      bits = 0;
      //===//
      state.mode = COPY_;
      if (flush === Z_TREES) { break inf_leave; }
      /* falls through */
    case COPY_:
      state.mode = COPY;
      /* falls through */
    case COPY:
      copy = state.length;
      if (copy) {
        if (copy > have) { copy = have; }
        if (copy > left) { copy = left; }
        if (copy === 0) { break inf_leave; }
        //--- zmemcpy(put, next, copy); ---
        utils.arraySet(output, input, next, copy, put);
        //---//
        have -= copy;
        next += copy;
        left -= copy;
        put += copy;
        state.length -= copy;
        break;
      }
      //Tracev((stderr, "inflate:       stored end\n"));
      state.mode = TYPE;
      break;
    case TABLE:
      //=== NEEDBITS(14); */
      while (bits < 14) {
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
      }
      //===//
      state.nlen = (hold & 0x1f)/*BITS(5)*/ + 257;
      //--- DROPBITS(5) ---//
      hold >>>= 5;
      bits -= 5;
      //---//
      state.ndist = (hold & 0x1f)/*BITS(5)*/ + 1;
      //--- DROPBITS(5) ---//
      hold >>>= 5;
      bits -= 5;
      //---//
      state.ncode = (hold & 0x0f)/*BITS(4)*/ + 4;
      //--- DROPBITS(4) ---//
      hold >>>= 4;
      bits -= 4;
      //---//
//#ifndef PKZIP_BUG_WORKAROUND
      if (state.nlen > 286 || state.ndist > 30) {
        strm.msg = 'too many length or distance symbols';
        state.mode = BAD;
        break;
      }
//#endif
      //Tracev((stderr, "inflate:       table sizes ok\n"));
      state.have = 0;
      state.mode = LENLENS;
      /* falls through */
    case LENLENS:
      while (state.have < state.ncode) {
        //=== NEEDBITS(3);
        while (bits < 3) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        state.lens[order[state.have++]] = (hold & 0x07);//BITS(3);
        //--- DROPBITS(3) ---//
        hold >>>= 3;
        bits -= 3;
        //---//
      }
      while (state.have < 19) {
        state.lens[order[state.have++]] = 0;
      }
      // We have separate tables & no pointers. 2 commented lines below not needed.
      //state.next = state.codes;
      //state.lencode = state.next;
      // Switch to use dynamic table
      state.lencode = state.lendyn;
      state.lenbits = 7;

      opts = {bits: state.lenbits};
      ret = inflate_table(CODES, state.lens, 0, 19, state.lencode, 0, state.work, opts);
      state.lenbits = opts.bits;

      if (ret) {
        strm.msg = 'invalid code lengths set';
        state.mode = BAD;
        break;
      }
      //Tracev((stderr, "inflate:       code lengths ok\n"));
      state.have = 0;
      state.mode = CODELENS;
      /* falls through */
    case CODELENS:
      while (state.have < state.nlen + state.ndist) {
        for (;;) {
          here = state.lencode[hold & ((1 << state.lenbits) - 1)];/*BITS(state.lenbits)*/
          here_bits = here >>> 24;
          here_op = (here >>> 16) & 0xff;
          here_val = here & 0xffff;

          if ((here_bits) <= bits) { break; }
          //--- PULLBYTE() ---//
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
          //---//
        }
        if (here_val < 16) {
          //--- DROPBITS(here.bits) ---//
          hold >>>= here_bits;
          bits -= here_bits;
          //---//
          state.lens[state.have++] = here_val;
        }
        else {
          if (here_val === 16) {
            //=== NEEDBITS(here.bits + 2);
            n = here_bits + 2;
            while (bits < n) {
              if (have === 0) { break inf_leave; }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            //--- DROPBITS(here.bits) ---//
            hold >>>= here_bits;
            bits -= here_bits;
            //---//
            if (state.have === 0) {
              strm.msg = 'invalid bit length repeat';
              state.mode = BAD;
              break;
            }
            len = state.lens[state.have - 1];
            copy = 3 + (hold & 0x03);//BITS(2);
            //--- DROPBITS(2) ---//
            hold >>>= 2;
            bits -= 2;
            //---//
          }
          else if (here_val === 17) {
            //=== NEEDBITS(here.bits + 3);
            n = here_bits + 3;
            while (bits < n) {
              if (have === 0) { break inf_leave; }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            //--- DROPBITS(here.bits) ---//
            hold >>>= here_bits;
            bits -= here_bits;
            //---//
            len = 0;
            copy = 3 + (hold & 0x07);//BITS(3);
            //--- DROPBITS(3) ---//
            hold >>>= 3;
            bits -= 3;
            //---//
          }
          else {
            //=== NEEDBITS(here.bits + 7);
            n = here_bits + 7;
            while (bits < n) {
              if (have === 0) { break inf_leave; }
              have--;
              hold += input[next++] << bits;
              bits += 8;
            }
            //===//
            //--- DROPBITS(here.bits) ---//
            hold >>>= here_bits;
            bits -= here_bits;
            //---//
            len = 0;
            copy = 11 + (hold & 0x7f);//BITS(7);
            //--- DROPBITS(7) ---//
            hold >>>= 7;
            bits -= 7;
            //---//
          }
          if (state.have + copy > state.nlen + state.ndist) {
            strm.msg = 'invalid bit length repeat';
            state.mode = BAD;
            break;
          }
          while (copy--) {
            state.lens[state.have++] = len;
          }
        }
      }

      /* handle error breaks in while */
      if (state.mode === BAD) { break; }

      /* check for end-of-block code (better have one) */
      if (state.lens[256] === 0) {
        strm.msg = 'invalid code -- missing end-of-block';
        state.mode = BAD;
        break;
      }

      /* build code tables -- note: do not change the lenbits or distbits
         values here (9 and 6) without reading the comments in inftrees.h
         concerning the ENOUGH constants, which depend on those values */
      state.lenbits = 9;

      opts = {bits: state.lenbits};
      ret = inflate_table(LENS, state.lens, 0, state.nlen, state.lencode, 0, state.work, opts);
      // We have separate tables & no pointers. 2 commented lines below not needed.
      // state.next_index = opts.table_index;
      state.lenbits = opts.bits;
      // state.lencode = state.next;

      if (ret) {
        strm.msg = 'invalid literal/lengths set';
        state.mode = BAD;
        break;
      }

      state.distbits = 6;
      //state.distcode.copy(state.codes);
      // Switch to use dynamic table
      state.distcode = state.distdyn;
      opts = {bits: state.distbits};
      ret = inflate_table(DISTS, state.lens, state.nlen, state.ndist, state.distcode, 0, state.work, opts);
      // We have separate tables & no pointers. 2 commented lines below not needed.
      // state.next_index = opts.table_index;
      state.distbits = opts.bits;
      // state.distcode = state.next;

      if (ret) {
        strm.msg = 'invalid distances set';
        state.mode = BAD;
        break;
      }
      //Tracev((stderr, 'inflate:       codes ok\n'));
      state.mode = LEN_;
      if (flush === Z_TREES) { break inf_leave; }
      /* falls through */
    case LEN_:
      state.mode = LEN;
      /* falls through */
    case LEN:
      if (have >= 6 && left >= 258) {
        //--- RESTORE() ---
        strm.next_out = put;
        strm.avail_out = left;
        strm.next_in = next;
        strm.avail_in = have;
        state.hold = hold;
        state.bits = bits;
        //---
        inflate_fast(strm, _out);
        //--- LOAD() ---
        put = strm.next_out;
        output = strm.output;
        left = strm.avail_out;
        next = strm.next_in;
        input = strm.input;
        have = strm.avail_in;
        hold = state.hold;
        bits = state.bits;
        //---

        if (state.mode === TYPE) {
          state.back = -1;
        }
        break;
      }
      state.back = 0;
      for (;;) {
        here = state.lencode[hold & ((1 << state.lenbits) -1)];  /*BITS(state.lenbits)*/
        here_bits = here >>> 24;
        here_op = (here >>> 16) & 0xff;
        here_val = here & 0xffff;

        if (here_bits <= bits) { break; }
        //--- PULLBYTE() ---//
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
        //---//
      }
      if (here_op && (here_op & 0xf0) === 0) {
        last_bits = here_bits;
        last_op = here_op;
        last_val = here_val;
        for (;;) {
          here = state.lencode[last_val +
                  ((hold & ((1 << (last_bits + last_op)) -1))/*BITS(last.bits + last.op)*/ >> last_bits)];
          here_bits = here >>> 24;
          here_op = (here >>> 16) & 0xff;
          here_val = here & 0xffff;

          if ((last_bits + here_bits) <= bits) { break; }
          //--- PULLBYTE() ---//
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
          //---//
        }
        //--- DROPBITS(last.bits) ---//
        hold >>>= last_bits;
        bits -= last_bits;
        //---//
        state.back += last_bits;
      }
      //--- DROPBITS(here.bits) ---//
      hold >>>= here_bits;
      bits -= here_bits;
      //---//
      state.back += here_bits;
      state.length = here_val;
      if (here_op === 0) {
        //Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
        //        "inflate:         literal '%c'\n" :
        //        "inflate:         literal 0x%02x\n", here.val));
        state.mode = LIT;
        break;
      }
      if (here_op & 32) {
        //Tracevv((stderr, "inflate:         end of block\n"));
        state.back = -1;
        state.mode = TYPE;
        break;
      }
      if (here_op & 64) {
        strm.msg = 'invalid literal/length code';
        state.mode = BAD;
        break;
      }
      state.extra = here_op & 15;
      state.mode = LENEXT;
      /* falls through */
    case LENEXT:
      if (state.extra) {
        //=== NEEDBITS(state.extra);
        n = state.extra;
        while (bits < n) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        state.length += hold & ((1 << state.extra) -1)/*BITS(state.extra)*/;
        //--- DROPBITS(state.extra) ---//
        hold >>>= state.extra;
        bits -= state.extra;
        //---//
        state.back += state.extra;
      }
      //Tracevv((stderr, "inflate:         length %u\n", state.length));
      state.was = state.length;
      state.mode = DIST;
      /* falls through */
    case DIST:
      for (;;) {
        here = state.distcode[hold & ((1 << state.distbits) -1)];/*BITS(state.distbits)*/
        here_bits = here >>> 24;
        here_op = (here >>> 16) & 0xff;
        here_val = here & 0xffff;

        if ((here_bits) <= bits) { break; }
        //--- PULLBYTE() ---//
        if (have === 0) { break inf_leave; }
        have--;
        hold += input[next++] << bits;
        bits += 8;
        //---//
      }
      if ((here_op & 0xf0) === 0) {
        last_bits = here_bits;
        last_op = here_op;
        last_val = here_val;
        for (;;) {
          here = state.distcode[last_val +
                  ((hold & ((1 << (last_bits + last_op)) -1))/*BITS(last.bits + last.op)*/ >> last_bits)];
          here_bits = here >>> 24;
          here_op = (here >>> 16) & 0xff;
          here_val = here & 0xffff;

          if ((last_bits + here_bits) <= bits) { break; }
          //--- PULLBYTE() ---//
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
          //---//
        }
        //--- DROPBITS(last.bits) ---//
        hold >>>= last_bits;
        bits -= last_bits;
        //---//
        state.back += last_bits;
      }
      //--- DROPBITS(here.bits) ---//
      hold >>>= here_bits;
      bits -= here_bits;
      //---//
      state.back += here_bits;
      if (here_op & 64) {
        strm.msg = 'invalid distance code';
        state.mode = BAD;
        break;
      }
      state.offset = here_val;
      state.extra = (here_op) & 15;
      state.mode = DISTEXT;
      /* falls through */
    case DISTEXT:
      if (state.extra) {
        //=== NEEDBITS(state.extra);
        n = state.extra;
        while (bits < n) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        state.offset += hold & ((1 << state.extra) -1)/*BITS(state.extra)*/;
        //--- DROPBITS(state.extra) ---//
        hold >>>= state.extra;
        bits -= state.extra;
        //---//
        state.back += state.extra;
      }
//#ifdef INFLATE_STRICT
      if (state.offset > state.dmax) {
        strm.msg = 'invalid distance too far back';
        state.mode = BAD;
        break;
      }
//#endif
      //Tracevv((stderr, "inflate:         distance %u\n", state.offset));
      state.mode = MATCH;
      /* falls through */
    case MATCH:
      if (left === 0) { break inf_leave; }
      copy = _out - left;
      if (state.offset > copy) {         /* copy from window */
        copy = state.offset - copy;
        if (copy > state.whave) {
          if (state.sane) {
            strm.msg = 'invalid distance too far back';
            state.mode = BAD;
            break;
          }
// (!) This block is disabled in zlib defailts,
// don't enable it for binary compatibility
//#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
//          Trace((stderr, "inflate.c too far\n"));
//          copy -= state.whave;
//          if (copy > state.length) { copy = state.length; }
//          if (copy > left) { copy = left; }
//          left -= copy;
//          state.length -= copy;
//          do {
//            output[put++] = 0;
//          } while (--copy);
//          if (state.length === 0) { state.mode = LEN; }
//          break;
//#endif
        }
        if (copy > state.wnext) {
          copy -= state.wnext;
          from = state.wsize - copy;
        }
        else {
          from = state.wnext - copy;
        }
        if (copy > state.length) { copy = state.length; }
        from_source = state.window;
      }
      else {                              /* copy from output */
        from_source = output;
        from = put - state.offset;
        copy = state.length;
      }
      if (copy > left) { copy = left; }
      left -= copy;
      state.length -= copy;
      do {
        output[put++] = from_source[from++];
      } while (--copy);
      if (state.length === 0) { state.mode = LEN; }
      break;
    case LIT:
      if (left === 0) { break inf_leave; }
      output[put++] = state.length;
      left--;
      state.mode = LEN;
      break;
    case CHECK:
      if (state.wrap) {
        //=== NEEDBITS(32);
        while (bits < 32) {
          if (have === 0) { break inf_leave; }
          have--;
          // Use '|' insdead of '+' to make sure that result is signed
          hold |= input[next++] << bits;
          bits += 8;
        }
        //===//
        _out -= left;
        strm.total_out += _out;
        state.total += _out;
        if (_out) {
          strm.adler = state.check =
              /*UPDATE(state.check, put - _out, _out);*/
              (state.flags ? crc32(state.check, output, _out, put - _out) : adler32(state.check, output, _out, put - _out));

        }
        _out = left;
        // NB: crc32 stored as signed 32-bit int, ZSWAP32 returns signed too
        if ((state.flags ? hold : ZSWAP32(hold)) !== state.check) {
          strm.msg = 'incorrect data check';
          state.mode = BAD;
          break;
        }
        //=== INITBITS();
        hold = 0;
        bits = 0;
        //===//
        //Tracev((stderr, "inflate:   check matches trailer\n"));
      }
      state.mode = LENGTH;
      /* falls through */
    case LENGTH:
      if (state.wrap && state.flags) {
        //=== NEEDBITS(32);
        while (bits < 32) {
          if (have === 0) { break inf_leave; }
          have--;
          hold += input[next++] << bits;
          bits += 8;
        }
        //===//
        if (hold !== (state.total & 0xffffffff)) {
          strm.msg = 'incorrect length check';
          state.mode = BAD;
          break;
        }
        //=== INITBITS();
        hold = 0;
        bits = 0;
        //===//
        //Tracev((stderr, "inflate:   length matches trailer\n"));
      }
      state.mode = DONE;
      /* falls through */
    case DONE:
      ret = Z_STREAM_END;
      break inf_leave;
    case BAD:
      ret = Z_DATA_ERROR;
      break inf_leave;
    case MEM:
      return Z_MEM_ERROR;
    case SYNC:
      /* falls through */
    default:
      return Z_STREAM_ERROR;
    }
  }

  // inf_leave <- here is real place for "goto inf_leave", emulated via "break inf_leave"

  /*
     Return from inflate(), updating the total counts and the check value.
     If there was no progress during the inflate() call, return a buffer
     error.  Call updatewindow() to create and/or update the window state.
     Note: a memory error from inflate() is non-recoverable.
   */

  //--- RESTORE() ---
  strm.next_out = put;
  strm.avail_out = left;
  strm.next_in = next;
  strm.avail_in = have;
  state.hold = hold;
  state.bits = bits;
  //---

  if (state.wsize || (_out !== strm.avail_out && state.mode < BAD &&
                      (state.mode < CHECK || flush !== Z_FINISH))) {
    if (updatewindow(strm, strm.output, strm.next_out, _out - strm.avail_out)) {
      state.mode = MEM;
      return Z_MEM_ERROR;
    }
  }
  _in -= strm.avail_in;
  _out -= strm.avail_out;
  strm.total_in += _in;
  strm.total_out += _out;
  state.total += _out;
  if (state.wrap && _out) {
    strm.adler = state.check = /*UPDATE(state.check, strm.next_out - _out, _out);*/
      (state.flags ? crc32(state.check, output, _out, strm.next_out - _out) : adler32(state.check, output, _out, strm.next_out - _out));
  }
  strm.data_type = state.bits + (state.last ? 64 : 0) +
                    (state.mode === TYPE ? 128 : 0) +
                    (state.mode === LEN_ || state.mode === COPY_ ? 256 : 0);
  if (((_in === 0 && _out === 0) || flush === Z_FINISH) && ret === Z_OK) {
    ret = Z_BUF_ERROR;
  }
  return ret;
}

function inflateEnd(strm) {

  if (!strm || !strm.state /*|| strm->zfree == (free_func)0*/) {
    return Z_STREAM_ERROR;
  }

  var state = strm.state;
  if (state.window) {
    state.window = null;
  }
  strm.state = null;
  return Z_OK;
}

function inflateGetHeader(strm, head) {
  var state;

  /* check state */
  if (!strm || !strm.state) { return Z_STREAM_ERROR; }
  state = strm.state;
  if ((state.wrap & 2) === 0) { return Z_STREAM_ERROR; }

  /* save header structure */
  state.head = head;
  head.done = false;
  return Z_OK;
}


exports.inflateReset = inflateReset;
exports.inflateReset2 = inflateReset2;
exports.inflateResetKeep = inflateResetKeep;
exports.inflateInit = inflateInit;
exports.inflateInit2 = inflateInit2;
exports.inflate = inflate;
exports.inflateEnd = inflateEnd;
exports.inflateGetHeader = inflateGetHeader;
exports.inflateInfo = 'pako inflate (from Nodeca project)';

/* Not implemented
exports.inflateCopy = inflateCopy;
exports.inflateGetDictionary = inflateGetDictionary;
exports.inflateMark = inflateMark;
exports.inflatePrime = inflatePrime;
exports.inflateSetDictionary = inflateSetDictionary;
exports.inflateSync = inflateSync;
exports.inflateSyncPoint = inflateSyncPoint;
exports.inflateUndermine = inflateUndermine;
*/
},{"../utils/common":2,"./adler32":4,"./crc32":6,"./inffast":8,"./inftrees":10}],10:[function(_dereq_,module,exports){
'use strict';


var utils = _dereq_('../utils/common');

var MAXBITS = 15;
var ENOUGH_LENS = 852;
var ENOUGH_DISTS = 592;
//var ENOUGH = (ENOUGH_LENS+ENOUGH_DISTS);

var CODES = 0;
var LENS = 1;
var DISTS = 2;

var lbase = [ /* Length codes 257..285 base */
  3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
  35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0
];

var lext = [ /* Length codes 257..285 extra */
  16, 16, 16, 16, 16, 16, 16, 16, 17, 17, 17, 17, 18, 18, 18, 18,
  19, 19, 19, 19, 20, 20, 20, 20, 21, 21, 21, 21, 16, 72, 78
];

var dbase = [ /* Distance codes 0..29 base */
  1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
  257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
  8193, 12289, 16385, 24577, 0, 0
];

var dext = [ /* Distance codes 0..29 extra */
  16, 16, 16, 16, 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22,
  23, 23, 24, 24, 25, 25, 26, 26, 27, 27,
  28, 28, 29, 29, 64, 64
];

module.exports = function inflate_table(type, lens, lens_index, codes, table, table_index, work, opts)
{
  var bits = opts.bits;
      //here = opts.here; /* table entry for duplication */

  var len = 0;               /* a code's length in bits */
  var sym = 0;               /* index of code symbols */
  var min = 0, max = 0;          /* minimum and maximum code lengths */
  var root = 0;              /* number of index bits for root table */
  var curr = 0;              /* number of index bits for current table */
  var drop = 0;              /* code bits to drop for sub-table */
  var left = 0;                   /* number of prefix codes available */
  var used = 0;              /* code entries in table used */
  var huff = 0;              /* Huffman code */
  var incr;              /* for incrementing code, index */
  var fill;              /* index for replicating entries */
  var low;               /* low bits for current root entry */
  var mask;              /* mask for low root bits */
  var next;             /* next available space in table */
  var base = null;     /* base value table to use */
  var base_index = 0;
//  var shoextra;    /* extra bits table to use */
  var end;                    /* use base and extra for symbol > end */
  var count = new utils.Buf16(MAXBITS+1); //[MAXBITS+1];    /* number of codes of each length */
  var offs = new utils.Buf16(MAXBITS+1); //[MAXBITS+1];     /* offsets in table for each length */
  var extra = null;
  var extra_index = 0;

  var here_bits, here_op, here_val;

  /*
   Process a set of code lengths to create a canonical Huffman code.  The
   code lengths are lens[0..codes-1].  Each length corresponds to the
   symbols 0..codes-1.  The Huffman code is generated by first sorting the
   symbols by length from short to long, and retaining the symbol order
   for codes with equal lengths.  Then the code starts with all zero bits
   for the first code of the shortest length, and the codes are integer
   increments for the same length, and zeros are appended as the length
   increases.  For the deflate format, these bits are stored backwards
   from their more natural integer increment ordering, and so when the
   decoding tables are built in the large loop below, the integer codes
   are incremented backwards.

   This routine assumes, but does not check, that all of the entries in
   lens[] are in the range 0..MAXBITS.  The caller must assure this.
   1..MAXBITS is interpreted as that code length.  zero means that that
   symbol does not occur in this code.

   The codes are sorted by computing a count of codes for each length,
   creating from that a table of starting indices for each length in the
   sorted table, and then entering the symbols in order in the sorted
   table.  The sorted table is work[], with that space being provided by
   the caller.

   The length counts are used for other purposes as well, i.e. finding
   the minimum and maximum length codes, determining if there are any
   codes at all, checking for a valid set of lengths, and looking ahead
   at length counts to determine sub-table sizes when building the
   decoding tables.
   */

  /* accumulate lengths for codes (assumes lens[] all in 0..MAXBITS) */
  for (len = 0; len <= MAXBITS; len++) {
    count[len] = 0;
  }
  for (sym = 0; sym < codes; sym++) {
    count[lens[lens_index + sym]]++;
  }

  /* bound code lengths, force root to be within code lengths */
  root = bits;
  for (max = MAXBITS; max >= 1; max--) {
    if (count[max] !== 0) { break; }
  }
  if (root > max) {
    root = max;
  }
  if (max === 0) {                     /* no symbols to code at all */
    //table.op[opts.table_index] = 64;  //here.op = (var char)64;    /* invalid code marker */
    //table.bits[opts.table_index] = 1;   //here.bits = (var char)1;
    //table.val[opts.table_index++] = 0;   //here.val = (var short)0;
    table[table_index++] = (1 << 24) | (64 << 16) | 0;


    //table.op[opts.table_index] = 64;
    //table.bits[opts.table_index] = 1;
    //table.val[opts.table_index++] = 0;
    table[table_index++] = (1 << 24) | (64 << 16) | 0;

    opts.bits = 1;
    return 0;     /* no symbols, but wait for decoding to report error */
  }
  for (min = 1; min < max; min++) {
    if (count[min] !== 0) { break; }
  }
  if (root < min) {
    root = min;
  }

  /* check for an over-subscribed or incomplete set of lengths */
  left = 1;
  for (len = 1; len <= MAXBITS; len++) {
    left <<= 1;
    left -= count[len];
    if (left < 0) {
      return -1;
    }        /* over-subscribed */
  }
  if (left > 0 && (type === CODES || max !== 1)) {
    return -1;                      /* incomplete set */
  }

  /* generate offsets into symbol table for each length for sorting */
  offs[1] = 0;
  for (len = 1; len < MAXBITS; len++) {
    offs[len + 1] = offs[len] + count[len];
  }

  /* sort symbols by length, by symbol order within each length */
  for (sym = 0; sym < codes; sym++) {
    if (lens[lens_index + sym] !== 0) {
      work[offs[lens[lens_index + sym]]++] = sym;
    }
  }

  /*
   Create and fill in decoding tables.  In this loop, the table being
   filled is at next and has curr index bits.  The code being used is huff
   with length len.  That code is converted to an index by dropping drop
   bits off of the bottom.  For codes where len is less than drop + curr,
   those top drop + curr - len bits are incremented through all values to
   fill the table with replicated entries.

   root is the number of index bits for the root table.  When len exceeds
   root, sub-tables are created pointed to by the root entry with an index
   of the low root bits of huff.  This is saved in low to check for when a
   new sub-table should be started.  drop is zero when the root table is
   being filled, and drop is root when sub-tables are being filled.

   When a new sub-table is needed, it is necessary to look ahead in the
   code lengths to determine what size sub-table is needed.  The length
   counts are used for this, and so count[] is decremented as codes are
   entered in the tables.

   used keeps track of how many table entries have been allocated from the
   provided *table space.  It is checked for LENS and DIST tables against
   the constants ENOUGH_LENS and ENOUGH_DISTS to guard against changes in
   the initial root table size constants.  See the comments in inftrees.h
   for more information.

   sym increments through all symbols, and the loop terminates when
   all codes of length max, i.e. all codes, have been processed.  This
   routine permits incomplete codes, so another loop after this one fills
   in the rest of the decoding tables with invalid code markers.
   */

  /* set up for code type */
  switch (type) {
    case CODES:
      base = extra = work;    /* dummy value--not used */
      end = 19;
      break;
    case LENS:
      base = lbase;
      base_index -= 257;
      extra = lext;
      extra_index -= 257;
      end = 256;
      break;
    default:            /* DISTS */
      base = dbase;
      extra = dext;
      end = -1;
  }

  /* initialize opts for loop */
  huff = 0;                   /* starting code */
  sym = 0;                    /* starting code symbol */
  len = min;                  /* starting code length */
  next = table_index;              /* current table to fill in */
  curr = root;                /* current table index bits */
  drop = 0;                   /* current bits to drop from code for index */
  low = -1;                   /* trigger new sub-table when len > root */
  used = 1 << root;          /* use root table entries */
  mask = used - 1;            /* mask for comparing low */

  /* check available table space */
  if ((type === LENS && used > ENOUGH_LENS) ||
    (type === DISTS && used > ENOUGH_DISTS)) {
    return 1;
  }

  var i=0;
  /* process all codes and make table entries */
  for (;;) {
    i++;
    /* create table entry */
    here_bits = len - drop;
    if (work[sym] < end) {
      here_op = 0;
      here_val = work[sym];
    }
    else if (work[sym] > end) {
      here_op = extra[extra_index + work[sym]];
      here_val = base[base_index + work[sym]];
    }
    else {
      here_op = 32 + 64;         /* end of block */
      here_val = 0;
    }

    /* replicate for those indices with low len bits equal to huff */
    incr = 1 << (len - drop);
    fill = 1 << curr;
    min = fill;                 /* save offset to next table */
    do {
      fill -= incr;
      table[next + (huff >> drop) + fill] = (here_bits << 24) | (here_op << 16) | here_val |0;
    } while (fill !== 0);

    /* backwards increment the len-bit code huff */
    incr = 1 << (len - 1);
    while (huff & incr) {
      incr >>= 1;
    }
    if (incr !== 0) {
      huff &= incr - 1;
      huff += incr;
    } else {
      huff = 0;
    }

    /* go to next symbol, update count, len */
    sym++;
    if (--(count[len]) === 0) {
      if (len === max) { break; }
      len = lens[lens_index + work[sym]];
    }

    /* create new sub-table if needed */
    if (len > root && (huff & mask) !== low) {
      /* if first time, transition to sub-tables */
      if (drop === 0) {
        drop = root;
      }

      /* increment past last table */
      next += min;            /* here min is 1 << curr */

      /* determine length of next table */
      curr = len - drop;
      left = 1 << curr;
      while (curr + drop < max) {
        left -= count[curr + drop];
        if (left <= 0) { break; }
        curr++;
        left <<= 1;
      }

      /* check for enough space */
      used += 1 << curr;
      if ((type === LENS && used > ENOUGH_LENS) ||
        (type === DISTS && used > ENOUGH_DISTS)) {
        return 1;
      }

      /* point entry in root table to sub-table */
      low = huff & mask;
      /*table.op[low] = curr;
      table.bits[low] = root;
      table.val[low] = next - opts.table_index;*/
      table[low] = (root << 24) | (curr << 16) | (next - table_index) |0;
    }
  }

  /* fill in remaining table entry if code is incomplete (guaranteed to have
   at most one remaining entry, since if the code is incomplete, the
   maximum code length that was allowed to get this far is one bit) */
  if (huff !== 0) {
    //table.op[next + huff] = 64;            /* invalid code marker */
    //table.bits[next + huff] = len - drop;
    //table.val[next + huff] = 0;
    table[next + huff] = ((len - drop) << 24) | (64 << 16) |0;
  }

  /* set return parameters */
  //opts.table_index += used;
  opts.bits = root;
  return 0;
};
},{"../utils/common":2}],11:[function(_dereq_,module,exports){
'use strict';

module.exports = {
  '2':    'need dictionary',     /* Z_NEED_DICT       2  */
  '1':    'stream end',          /* Z_STREAM_END      1  */
  '0':    '',                    /* Z_OK              0  */
  '-1':   'file error',          /* Z_ERRNO         (-1) */
  '-2':   'stream error',        /* Z_STREAM_ERROR  (-2) */
  '-3':   'data error',          /* Z_DATA_ERROR    (-3) */
  '-4':   'insufficient memory', /* Z_MEM_ERROR     (-4) */
  '-5':   'buffer error',        /* Z_BUF_ERROR     (-5) */
  '-6':   'incompatible version' /* Z_VERSION_ERROR (-6) */
};
},{}],12:[function(_dereq_,module,exports){
'use strict';


function ZStream() {
  /* next input byte */
  this.input = null; // JS specific, because we have no pointers
  this.next_in = 0;
  /* number of bytes available at input */
  this.avail_in = 0;
  /* total number of input bytes read so far */
  this.total_in = 0;
  /* next output byte should be put there */
  this.output = null; // JS specific, because we have no pointers
  this.next_out = 0;
  /* remaining free space at output */
  this.avail_out = 0;
  /* total number of bytes output so far */
  this.total_out = 0;
  /* last error message, NULL if no error */
  this.msg = ''/*Z_NULL*/;
  /* not visible by applications */
  this.state = null;
  /* best guess about the data type: binary or text */
  this.data_type = 2/*Z_UNKNOWN*/;
  /* adler32 value of the uncompressed data */
  this.adler = 0;
}

module.exports = ZStream;
},{}]},{},[1])
(1)
});/* FileSaver.js
 * A saveAs() FileSaver implementation.
 * 2015-03-04
 *
 * By Eli Grey, http://eligrey.com
 * License: X11/MIT
 *   See https://github.com/eligrey/FileSaver.js/blob/master/LICENSE.md
 */

/*global self */
/*jslint bitwise: true, indent: 4, laxbreak: true, laxcomma: true, smarttabs: true, plusplus: true */

/*! @source http://purl.eligrey.com/github/FileSaver.js/blob/master/FileSaver.js */

var saveAs = saveAs
  // IE 10+ (native saveAs)
  || (typeof navigator !== "undefined" &&
      navigator.msSaveOrOpenBlob && navigator.msSaveOrOpenBlob.bind(navigator))
  // Everyone else
  || (function(view) {
	"use strict";
	// IE <10 is explicitly unsupported
	if (typeof navigator !== "undefined" &&
	    /MSIE [1-9]\./.test(navigator.userAgent)) {
		return;
	}
	var
		  doc = view.document
		  // only get URL when necessary in case Blob.js hasn't overridden it yet
		, get_URL = function() {
			return view.URL || view.webkitURL || view;
		}
		, save_link = doc.createElementNS("http://www.w3.org/1999/xhtml", "a")
		, can_use_save_link = "download" in save_link
		, click = function(node) {
			var event = doc.createEvent("MouseEvents");
			event.initMouseEvent(
				"click", true, false, view, 0, 0, 0, 0, 0
				, false, false, false, false, 0, null
			);
			node.dispatchEvent(event);
		}
		, webkit_req_fs = view.webkitRequestFileSystem
		, req_fs = view.requestFileSystem || webkit_req_fs || view.mozRequestFileSystem
		, throw_outside = function(ex) {
			(view.setImmediate || view.setTimeout)(function() {
				throw ex;
			}, 0);
		}
		, force_saveable_type = "application/octet-stream"
		, fs_min_size = 0
		// See https://code.google.com/p/chromium/issues/detail?id=375297#c7 and
		// https://github.com/eligrey/FileSaver.js/commit/485930a#commitcomment-8768047
		// for the reasoning behind the timeout and revocation flow
		, arbitrary_revoke_timeout = 500 // in ms
		, revoke = function(file) {
			var revoker = function() {
				if (typeof file === "string") { // file is an object URL
					get_URL().revokeObjectURL(file);
				} else { // file is a File
					file.remove();
				}
			};
			if (view.chrome) {
				revoker();
			} else {
				setTimeout(revoker, arbitrary_revoke_timeout);
			}
		}
		, dispatch = function(filesaver, event_types, event) {
			event_types = [].concat(event_types);
			var i = event_types.length;
			while (i--) {
				var listener = filesaver["on" + event_types[i]];
				if (typeof listener === "function") {
					try {
						listener.call(filesaver, event || filesaver);
					} catch (ex) {
						throw_outside(ex);
					}
				}
			}
		}
		, FileSaver = function(blob, name) {
			// First try a.download, then web filesystem, then object URLs
			var
				  filesaver = this
				, type = blob.type
				, blob_changed = false
				, object_url
				, target_view
				, dispatch_all = function() {
					dispatch(filesaver, "writestart progress write writeend".split(" "));
				}
				// on any filesys errors revert to saving with object URLs
				, fs_error = function() {
					// don't create more object URLs than needed
					if (blob_changed || !object_url) {
						object_url = get_URL().createObjectURL(blob);
					}
					if (target_view) {
						target_view.location.href = object_url;
					} else {
						var new_tab = view.open(object_url, "_blank");
						if (new_tab == undefined && typeof safari !== "undefined") {
							//Apple do not allow window.open, see http://bit.ly/1kZffRI
							view.location.href = object_url
						}
					}
					filesaver.readyState = filesaver.DONE;
					dispatch_all();
					revoke(object_url);
				}
				, abortable = function(func) {
					return function() {
						if (filesaver.readyState !== filesaver.DONE) {
							return func.apply(this, arguments);
						}
					};
				}
				, create_if_not_found = {create: true, exclusive: false}
				, slice
			;
			filesaver.readyState = filesaver.INIT;
			if (!name) {
				name = "download";
			}
			if (can_use_save_link) {
				object_url = get_URL().createObjectURL(blob);
				save_link.href = object_url;
				save_link.download = name;
				click(save_link);
				filesaver.readyState = filesaver.DONE;
				dispatch_all();
				revoke(object_url);
				return;
			}
			// prepend BOM for UTF-8 XML and text/plain types
			if (/^\s*(?:text\/(?:plain|xml)|application\/xml|\S*\/\S*\+xml)\s*;.*charset\s*=\s*utf-8/i.test(blob.type)) {
				blob = new Blob(["\ufeff", blob], {type: blob.type});
			}
			// Object and web filesystem URLs have a problem saving in Google Chrome when
			// viewed in a tab, so I force save with application/octet-stream
			// http://code.google.com/p/chromium/issues/detail?id=91158
			// Update: Google errantly closed 91158, I submitted it again:
			// https://code.google.com/p/chromium/issues/detail?id=389642
			if (view.chrome && type && type !== force_saveable_type) {
				slice = blob.slice || blob.webkitSlice;
				blob = slice.call(blob, 0, blob.size, force_saveable_type);
				blob_changed = true;
			}
			// Since I can't be sure that the guessed media type will trigger a download
			// in WebKit, I append .download to the filename.
			// https://bugs.webkit.org/show_bug.cgi?id=65440
			if (webkit_req_fs && name !== "download") {
				name += ".download";
			}
			if (type === force_saveable_type || webkit_req_fs) {
				target_view = view;
			}
			if (!req_fs) {
				fs_error();
				return;
			}
			fs_min_size += blob.size;
			req_fs(view.TEMPORARY, fs_min_size, abortable(function(fs) {
				fs.root.getDirectory("saved", create_if_not_found, abortable(function(dir) {
					var save = function() {
						dir.getFile(name, create_if_not_found, abortable(function(file) {
							file.createWriter(abortable(function(writer) {
								writer.onwriteend = function(event) {
									target_view.location.href = file.toURL();
									filesaver.readyState = filesaver.DONE;
									dispatch(filesaver, "writeend", event);
									revoke(file);
								};
								writer.onerror = function() {
									var error = writer.error;
									if (error.code !== error.ABORT_ERR) {
										fs_error();
									}
								};
								"writestart progress write abort".split(" ").forEach(function(event) {
									writer["on" + event] = filesaver["on" + event];
								});
								writer.write(blob);
								filesaver.abort = function() {
									writer.abort();
									filesaver.readyState = filesaver.DONE;
								};
								filesaver.readyState = filesaver.WRITING;
							}), fs_error);
						}), fs_error);
					};
					dir.getFile(name, {create: false}, abortable(function(file) {
						// delete file if it already exists
						file.remove();
						save();
					}), abortable(function(ex) {
						if (ex.code === ex.NOT_FOUND_ERR) {
							save();
						} else {
							fs_error();
						}
					}));
				}), fs_error);
			}), fs_error);
		}
		, FS_proto = FileSaver.prototype
		, saveAs = function(blob, name) {
			return new FileSaver(blob, name);
		}
	;
	FS_proto.abort = function() {
		var filesaver = this;
		filesaver.readyState = filesaver.DONE;
		dispatch(filesaver, "abort");
	};
	FS_proto.readyState = FS_proto.INIT = 0;
	FS_proto.WRITING = 1;
	FS_proto.DONE = 2;

	FS_proto.error =
	FS_proto.onwritestart =
	FS_proto.onprogress =
	FS_proto.onwrite =
	FS_proto.onabort =
	FS_proto.onerror =
	FS_proto.onwriteend =
		null;

	return saveAs;
}(
	   typeof self !== "undefined" && self
	|| typeof window !== "undefined" && window
	|| this.content
));
// `self` is undefined in Firefox for Android content script context
// while `this` is nsIContentFrameMessageManager
// with an attribute `content` that corresponds to the window

if (typeof module !== "undefined" && module.exports) {
  module.exports.saveAs = saveAs;
} else if ((typeof define !== "undefined" && define !== null) && (define.amd != null)) {
  define([], function() {
    return saveAs;
  });
}
/* canvas-toBlob.js
 * A canvas.toBlob() implementation.
 * 2013-12-27
 * 
 * By Eli Grey, http://eligrey.com and Devin Samarin, https://github.com/eboyjr
 * License: X11/MIT
 *   See https://github.com/eligrey/canvas-toBlob.js/blob/master/LICENSE.md
 */

/*global self */
/*jslint bitwise: true, regexp: true, confusion: true, es5: true, vars: true, white: true,
  plusplus: true */

/*! @source http://purl.eligrey.com/github/canvas-toBlob.js/blob/master/canvas-toBlob.js */

(function(view) {
"use strict";
var
	  Uint8Array = view.Uint8Array
	, HTMLCanvasElement = view.HTMLCanvasElement
	, canvas_proto = HTMLCanvasElement && HTMLCanvasElement.prototype
	, is_base64_regex = /\s*;\s*base64\s*(?:;|$)/i
	, to_data_url = "toDataURL"
	, base64_ranks
	, decode_base64 = function(base64) {
		var
			  len = base64.length
			, buffer = new Uint8Array(len / 4 * 3 | 0)
			, i = 0
			, outptr = 0
			, last = [0, 0]
			, state = 0
			, save = 0
			, rank
			, code
			, undef
		;
		while (len--) {
			code = base64.charCodeAt(i++);
			rank = base64_ranks[code-43];
			if (rank !== 255 && rank !== undef) {
				last[1] = last[0];
				last[0] = code;
				save = (save << 6) | rank;
				state++;
				if (state === 4) {
					buffer[outptr++] = save >>> 16;
					if (last[1] !== 61 /* padding character */) {
						buffer[outptr++] = save >>> 8;
					}
					if (last[0] !== 61 /* padding character */) {
						buffer[outptr++] = save;
					}
					state = 0;
				}
			}
		}
		// 2/3 chance there's going to be some null bytes at the end, but that
		// doesn't really matter with most image formats.
		// If it somehow matters for you, truncate the buffer up outptr.
		return buffer;
	}
;
if (Uint8Array) {
	base64_ranks = new Uint8Array([
		  62, -1, -1, -1, 63, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1
		, -1, -1,  0, -1, -1, -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9
		, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25
		, -1, -1, -1, -1, -1, -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35
		, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51
	]);
}
if (HTMLCanvasElement && !canvas_proto.toBlob) {
	canvas_proto.toBlob = function(callback, type /*, ...args*/) {
		  if (!type) {
			type = "image/png";
		} if (this.mozGetAsFile) {
			callback(this.mozGetAsFile("canvas", type));
			return;
		} if (this.msToBlob && /^\s*image\/png\s*(?:$|;)/i.test(type)) {
			callback(this.msToBlob());
			return;
		}

		var
			  args = Array.prototype.slice.call(arguments, 1)
			, dataURI = this[to_data_url].apply(this, args)
			, header_end = dataURI.indexOf(",")
			, data = dataURI.substring(header_end + 1)
			, is_base64 = is_base64_regex.test(dataURI.substring(0, header_end))
			, blob
		;
		if (Blob.fake) {
			// no reason to decode a data: URI that's just going to become a data URI again
			blob = new Blob
			if (is_base64) {
				blob.encoding = "base64";
			} else {
				blob.encoding = "URI";
			}
			blob.data = data;
			blob.size = data.length;
		} else if (Uint8Array) {
			if (is_base64) {
				blob = new Blob([decode_base64(data)], {type: type});
			} else {
				blob = new Blob([decodeURIComponent(data)], {type: type});
			}
		}
		callback(blob);
	};

	if (canvas_proto.toDataURLHD) {
		canvas_proto.toBlobHD = function() {
			to_data_url = "toDataURLHD";
			var blob = this.toBlob();
			to_data_url = "toDataURL";
			return blob;
		}
	} else {
		canvas_proto.toBlobHD = canvas_proto.toBlob;
	}
}
}(typeof self !== "undefined" && self || typeof window !== "undefined" && window || this.content || this));
//** Tab Content script v2.0-  Dynamic Drive DHTML code library (http://www.dynamicdrive.com)
//** Updated Oct 7th, 07 to version 2.0. Contains numerous improvements:
//   -Added Auto Mode: Script auto rotates the tabs based on an interval, until a tab is explicitly selected
//   -Ability to expand/contract arbitrary DIVs on the page as the tabbed content is expanded/ contracted
//   -Ability to dynamically select a tab either based on its position within its peers, or its ID attribute (give the target tab one 1st)
//   -Ability to set where the CSS classname "selected" get assigned- either to the target tab's link ("A"), or its parent container
//** Updated Feb 18th, 08 to version 2.1: Adds a "tabinstance.cycleit(dir)" method to cycle forward or backward between tabs dynamically
//** Updated April 8th, 08 to version 2.2: Adds support for expanding a tab using a URL parameter (ie: http://mysite.com/tabcontent.htm?tabinterfaceid=0) 

////NO NEED TO EDIT BELOW////////////////////////

function ddtabcontent(tabinterfaceid){
	this.tabinterfaceid=tabinterfaceid //ID of Tab Menu main container
	this.tabs=document.getElementById(tabinterfaceid).getElementsByTagName("a") //Get all tab links within container
	this.enabletabpersistence=true
	this.hottabspositions=[] //Array to store position of tabs that have a "rel" attr defined, relative to all tab links, within container
	this.currentTabIndex=0 //Index of currently selected hot tab (tab with sub content) within hottabspositions[] array
	this.subcontentids=[] //Array to store ids of the sub contents ("rel" attr values)
	this.revcontentids=[] //Array to store ids of arbitrary contents to expand/contact as well ("rev" attr values)
	this.selectedClassTarget="link" //keyword to indicate which target element to assign "selected" CSS class ("linkparent" or "link")
}

ddtabcontent.getCookie=function(Name){ 
	var re=new RegExp(Name+"=[^;]+", "i"); //construct RE to search for target name/value pair
	if (document.cookie.match(re)) //if cookie found
		return document.cookie.match(re)[0].split("=")[1] //return its value
	return ""
}

ddtabcontent.setCookie=function(name, value){
	document.cookie = name+"="+value+";path=/" //cookie value is domain wide (path=/)
}

ddtabcontent.prototype={

	expandit:function(tabid_or_position){ //PUBLIC function to select a tab either by its ID or position(int) within its peers
		this.cancelautorun() //stop auto cycling of tabs (if running)
		var tabref=""
		try{
			if (typeof tabid_or_position=="string" && document.getElementById(tabid_or_position).getAttribute("rel")) //if specified tab contains "rel" attr
				tabref=document.getElementById(tabid_or_position)
			else if (parseInt(tabid_or_position)!=NaN && this.tabs[tabid_or_position].getAttribute("rel")) //if specified tab contains "rel" attr
				tabref=this.tabs[tabid_or_position]
		}
		catch(err){alert("Invalid Tab ID or position entered!")}
		if (tabref!="") //if a valid tab is found based on function parameter
			this.expandtab(tabref) //expand this tab
	},

	cycleit:function(dir, autorun){ //PUBLIC function to move foward or backwards through each hot tab (tabinstance.cycleit('foward/back') )
		if (dir=="next"){
			var currentTabIndex=(this.currentTabIndex<this.hottabspositions.length-1)? this.currentTabIndex+1 : 0
		}
		else if (dir=="prev"){
			var currentTabIndex=(this.currentTabIndex>0)? this.currentTabIndex-1 : this.hottabspositions.length-1
		}
		if (typeof autorun=="undefined") //if cycleit() is being called by user, versus autorun() function
			this.cancelautorun() //stop auto cycling of tabs (if running)
		this.expandtab(this.tabs[this.hottabspositions[currentTabIndex]])
	},

	setpersist:function(bool){ //PUBLIC function to toggle persistence feature
			this.enabletabpersistence=bool
	},

	setselectedClassTarget:function(objstr){ //PUBLIC function to set which target element to assign "selected" CSS class ("linkparent" or "link")
		this.selectedClassTarget=objstr || "link"
	},

	getselectedClassTarget:function(tabref){ //Returns target element to assign "selected" CSS class to
		return (this.selectedClassTarget==("linkparent".toLowerCase()))? tabref.parentNode : tabref
	},

	urlparamselect:function(tabinterfaceid){
		var result=window.location.search.match(new RegExp(tabinterfaceid+"=(\\d+)", "i")) //check for "?tabinterfaceid=2" in URL
		return (result==null)? null : parseInt(RegExp.$1) //returns null or index, where index (int) is the selected tab's index
	},

	expandtab:function(tabref){
		var subcontentid=tabref.getAttribute("rel") //Get id of subcontent to expand
		//Get "rev" attr as a string of IDs in the format ",john,george,trey,etc," to easily search through
		var associatedrevids=(tabref.getAttribute("rev"))? ","+tabref.getAttribute("rev").replace(/\s+/, "")+"," : ""
		this.expandsubcontent(subcontentid)
		this.expandrevcontent(associatedrevids)
		for (var i=0; i<this.tabs.length; i++){ //Loop through all tabs, and assign only the selected tab the CSS class "selected"
			this.getselectedClassTarget(this.tabs[i]).className=(this.tabs[i].getAttribute("rel")==subcontentid)? "selected" : ""
		}
		if (this.enabletabpersistence) //if persistence enabled, save selected tab position(int) relative to its peers
			ddtabcontent.setCookie(this.tabinterfaceid, tabref.tabposition)
		this.setcurrenttabindex(tabref.tabposition) //remember position of selected tab within hottabspositions[] array
	},

	expandsubcontent:function(subcontentid){
		for (var i=0; i<this.subcontentids.length; i++){
			var subcontent=document.getElementById(this.subcontentids[i]) //cache current subcontent obj (in for loop)
			subcontent.style.display=(subcontent.id==subcontentid)? "block" : "none" //"show" or hide sub content based on matching id attr value
		}
	},

	expandrevcontent:function(associatedrevids){
		var allrevids=this.revcontentids
		for (var i=0; i<allrevids.length; i++){ //Loop through rev attributes for all tabs in this tab interface
			//if any values stored within associatedrevids matches one within allrevids, expand that DIV, otherwise, contract it
			document.getElementById(allrevids[i]).style.display=(associatedrevids.indexOf(","+allrevids[i]+",")!=-1)? "block" : "none"
		}
	},

	setcurrenttabindex:function(tabposition){ //store current position of tab (within hottabspositions[] array)
		for (var i=0; i<this.hottabspositions.length; i++){
			if (tabposition==this.hottabspositions[i]){
				this.currentTabIndex=i
				break
			}
		}
	},

	autorun:function(){ //function to auto cycle through and select tabs based on a set interval
		this.cycleit('next', true)
	},

	cancelautorun:function(){
		if (typeof this.autoruntimer!="undefined")
			clearInterval(this.autoruntimer)
	},

	init:function(automodeperiod){
		var persistedtab=ddtabcontent.getCookie(this.tabinterfaceid) //get position of persisted tab (applicable if persistence is enabled)
		var selectedtab=-1 //Currently selected tab index (-1 meaning none)
		var selectedtabfromurl=this.urlparamselect(this.tabinterfaceid) //returns null or index from: tabcontent.htm?tabinterfaceid=index
		this.automodeperiod=automodeperiod || 0
		for (var i=0; i<this.tabs.length; i++){
			this.tabs[i].tabposition=i //remember position of tab relative to its peers
			if (this.tabs[i].getAttribute("rel")){
				var tabinstance=this
				this.hottabspositions[this.hottabspositions.length]=i //store position of "hot" tab ("rel" attr defined) relative to its peers
				this.subcontentids[this.subcontentids.length]=this.tabs[i].getAttribute("rel") //store id of sub content ("rel" attr value)
				this.tabs[i].onclick=function(){
					tabinstance.expandtab(this)
					tabinstance.cancelautorun() //stop auto cycling of tabs (if running)
					return false
				}
				if (this.tabs[i].getAttribute("rev")){ //if "rev" attr defined, store each value within "rev" as an array element
					this.revcontentids=this.revcontentids.concat(this.tabs[i].getAttribute("rev").split(/\s*,\s*/))
				}
				if (selectedtabfromurl==i || this.enabletabpersistence && selectedtab==-1 && parseInt(persistedtab)==i || !this.enabletabpersistence && selectedtab==-1 && this.getselectedClassTarget(this.tabs[i]).className=="selected"){
					selectedtab=i //Selected tab index, if found
				}
			}
		} //END for loop
		if (selectedtab!=-1) //if a valid default selected tab index is found
			this.expandtab(this.tabs[selectedtab]) //expand selected tab (either from URL parameter, persistent feature, or class="selected" class)
		else //if no valid default selected index found
			this.expandtab(this.tabs[this.hottabspositions[0]]) //Just select first tab that contains a "rel" attr
		if (parseInt(this.automodeperiod)>500 && this.hottabspositions.length>1){
			this.autoruntimer=setInterval(function(){tabinstance.autorun()}, this.automodeperiod)
		}
	} //END int() function

} //END Prototype assignment
// astroem.js: astronomy utilities compiled to js

var Astroem = (function() {

var Module={"noExitRuntime":true};var Module;if(!Module)Module=(typeof Module!=="undefined"?Module:null)||{};var moduleOverrides={};for(var key in Module){if(Module.hasOwnProperty(key)){moduleOverrides[key]=Module[key]}}var ENVIRONMENT_IS_WEB=typeof window==="object";var ENVIRONMENT_IS_WORKER=typeof importScripts==="function";var ENVIRONMENT_IS_NODE=typeof process==="object"&&typeof require==="function"&&!ENVIRONMENT_IS_WEB&&!ENVIRONMENT_IS_WORKER;var ENVIRONMENT_IS_SHELL=!ENVIRONMENT_IS_WEB&&!ENVIRONMENT_IS_NODE&&!ENVIRONMENT_IS_WORKER;if(ENVIRONMENT_IS_NODE){if(!Module["print"])Module["print"]=function print(x){process["stdout"].write(x+"\n")};if(!Module["printErr"])Module["printErr"]=function printErr(x){process["stderr"].write(x+"\n")};var nodeFS=require("fs");var nodePath=require("path");Module["read"]=function read(filename,binary){filename=nodePath["normalize"](filename);var ret=nodeFS["readFileSync"](filename);if(!ret&&filename!=nodePath["resolve"](filename)){filename=path.join(__dirname,"..","src",filename);ret=nodeFS["readFileSync"](filename)}if(ret&&!binary)ret=ret.toString();return ret};Module["readBinary"]=function readBinary(filename){var ret=Module["read"](filename,true);if(!ret.buffer){ret=new Uint8Array(ret)}assert(ret.buffer);return ret};Module["load"]=function load(f){globalEval(read(f))};if(!Module["thisProgram"]){if(process["argv"].length>1){Module["thisProgram"]=process["argv"][1].replace(/\\/g,"/")}else{Module["thisProgram"]="unknown-program"}}Module["arguments"]=process["argv"].slice(2);if(typeof module!=="undefined"){module["exports"]=Module}process["on"]("uncaughtException",(function(ex){if(!(ex instanceof ExitStatus)){throw ex}}));Module["inspect"]=(function(){return"[Emscripten Module object]"})}else if(ENVIRONMENT_IS_SHELL){if(!Module["print"])Module["print"]=print;if(typeof printErr!="undefined")Module["printErr"]=printErr;if(typeof read!="undefined"){Module["read"]=read}else{Module["read"]=function read(){throw"no read() available (jsc?)"}}Module["readBinary"]=function readBinary(f){if(typeof readbuffer==="function"){return new Uint8Array(readbuffer(f))}var data=read(f,"binary");assert(typeof data==="object");return data};if(typeof scriptArgs!="undefined"){Module["arguments"]=scriptArgs}else if(typeof arguments!="undefined"){Module["arguments"]=arguments}}else if(ENVIRONMENT_IS_WEB||ENVIRONMENT_IS_WORKER){Module["read"]=function read(url){var xhr=new XMLHttpRequest;xhr.open("GET",url,false);xhr.send(null);return xhr.responseText};if(typeof arguments!="undefined"){Module["arguments"]=arguments}if(typeof console!=="undefined"){if(!Module["print"])Module["print"]=function print(x){console.log(x)};if(!Module["printErr"])Module["printErr"]=function printErr(x){console.log(x)}}else{var TRY_USE_DUMP=false;if(!Module["print"])Module["print"]=TRY_USE_DUMP&&typeof dump!=="undefined"?(function(x){dump(x)}):(function(x){})}if(ENVIRONMENT_IS_WORKER){Module["load"]=importScripts}if(typeof Module["setWindowTitle"]==="undefined"){Module["setWindowTitle"]=(function(title){document.title=title})}}else{throw"Unknown runtime environment. Where are we?"}function globalEval(x){eval.call(null,x)}if(!Module["load"]&&Module["read"]){Module["load"]=function load(f){globalEval(Module["read"](f))}}if(!Module["print"]){Module["print"]=(function(){})}if(!Module["printErr"]){Module["printErr"]=Module["print"]}if(!Module["arguments"]){Module["arguments"]=[]}if(!Module["thisProgram"]){Module["thisProgram"]="./this.program"}Module.print=Module["print"];Module.printErr=Module["printErr"];Module["preRun"]=[];Module["postRun"]=[];for(var key in moduleOverrides){if(moduleOverrides.hasOwnProperty(key)){Module[key]=moduleOverrides[key]}}var Runtime={setTempRet0:(function(value){tempRet0=value}),getTempRet0:(function(){return tempRet0}),stackSave:(function(){return STACKTOP}),stackRestore:(function(stackTop){STACKTOP=stackTop}),getNativeTypeSize:(function(type){switch(type){case"i1":case"i8":return 1;case"i16":return 2;case"i32":return 4;case"i64":return 8;case"float":return 4;case"double":return 8;default:{if(type[type.length-1]==="*"){return Runtime.QUANTUM_SIZE}else if(type[0]==="i"){var bits=parseInt(type.substr(1));assert(bits%8===0);return bits/8}else{return 0}}}}),getNativeFieldSize:(function(type){return Math.max(Runtime.getNativeTypeSize(type),Runtime.QUANTUM_SIZE)}),STACK_ALIGN:16,prepVararg:(function(ptr,type){if(type==="double"||type==="i64"){if(ptr&7){assert((ptr&7)===4);ptr+=4}}else{assert((ptr&3)===0)}return ptr}),getAlignSize:(function(type,size,vararg){if(!vararg&&(type=="i64"||type=="double"))return 8;if(!type)return Math.min(size,8);return Math.min(size||(type?Runtime.getNativeFieldSize(type):0),Runtime.QUANTUM_SIZE)}),dynCall:(function(sig,ptr,args){if(args&&args.length){if(!args.splice)args=Array.prototype.slice.call(args);args.splice(0,0,ptr);return Module["dynCall_"+sig].apply(null,args)}else{return Module["dynCall_"+sig].call(null,ptr)}}),functionPointers:[],addFunction:(function(func){for(var i=0;i<Runtime.functionPointers.length;i++){if(!Runtime.functionPointers[i]){Runtime.functionPointers[i]=func;return 2*(1+i)}}throw"Finished up all reserved function pointers. Use a higher value for RESERVED_FUNCTION_POINTERS."}),removeFunction:(function(index){Runtime.functionPointers[(index-2)/2]=null}),warnOnce:(function(text){if(!Runtime.warnOnce.shown)Runtime.warnOnce.shown={};if(!Runtime.warnOnce.shown[text]){Runtime.warnOnce.shown[text]=1;Module.printErr(text)}}),funcWrappers:{},getFuncWrapper:(function(func,sig){assert(sig);if(!Runtime.funcWrappers[sig]){Runtime.funcWrappers[sig]={}}var sigCache=Runtime.funcWrappers[sig];if(!sigCache[func]){sigCache[func]=function dynCall_wrapper(){return Runtime.dynCall(sig,func,arguments)}}return sigCache[func]}),getCompilerSetting:(function(name){throw"You must build with -s RETAIN_COMPILER_SETTINGS=1 for Runtime.getCompilerSetting or emscripten_get_compiler_setting to work"}),stackAlloc:(function(size){var ret=STACKTOP;STACKTOP=STACKTOP+size|0;STACKTOP=STACKTOP+15&-16;return ret}),staticAlloc:(function(size){var ret=STATICTOP;STATICTOP=STATICTOP+size|0;STATICTOP=STATICTOP+15&-16;return ret}),dynamicAlloc:(function(size){var ret=DYNAMICTOP;DYNAMICTOP=DYNAMICTOP+size|0;DYNAMICTOP=DYNAMICTOP+15&-16;if(DYNAMICTOP>=TOTAL_MEMORY){var success=enlargeMemory();if(!success){DYNAMICTOP=ret;return 0}}return ret}),alignMemory:(function(size,quantum){var ret=size=Math.ceil(size/(quantum?quantum:16))*(quantum?quantum:16);return ret}),makeBigInt:(function(low,high,unsigned){var ret=unsigned?+(low>>>0)+ +(high>>>0)*+4294967296:+(low>>>0)+ +(high|0)*+4294967296;return ret}),GLOBAL_BASE:8,QUANTUM_SIZE:4,__dummy__:0};Module["Runtime"]=Runtime;var __THREW__=0;var ABORT=false;var EXITSTATUS=0;var undef=0;var tempValue,tempInt,tempBigInt,tempInt2,tempBigInt2,tempPair,tempBigIntI,tempBigIntR,tempBigIntS,tempBigIntP,tempBigIntD,tempDouble,tempFloat;var tempI64,tempI64b;var tempRet0,tempRet1,tempRet2,tempRet3,tempRet4,tempRet5,tempRet6,tempRet7,tempRet8,tempRet9;function assert(condition,text){if(!condition){abort("Assertion failed: "+text)}}var globalScope=this;function getCFunc(ident){var func=Module["_"+ident];if(!func){try{func=eval("_"+ident)}catch(e){}}assert(func,"Cannot call unknown function "+ident+" (perhaps LLVM optimizations or closure removed it?)");return func}var cwrap,ccall;((function(){var JSfuncs={"stackSave":(function(){Runtime.stackSave()}),"stackRestore":(function(){Runtime.stackRestore()}),"arrayToC":(function(arr){var ret=Runtime.stackAlloc(arr.length);writeArrayToMemory(arr,ret);return ret}),"stringToC":(function(str){var ret=0;if(str!==null&&str!==undefined&&str!==0){ret=Runtime.stackAlloc((str.length<<2)+1);writeStringToMemory(str,ret)}return ret})};var toC={"string":JSfuncs["stringToC"],"array":JSfuncs["arrayToC"]};ccall=function ccallFunc(ident,returnType,argTypes,args,opts){var func=getCFunc(ident);var cArgs=[];var stack=0;if(args){for(var i=0;i<args.length;i++){var converter=toC[argTypes[i]];if(converter){if(stack===0)stack=Runtime.stackSave();cArgs[i]=converter(args[i])}else{cArgs[i]=args[i]}}}var ret=func.apply(null,cArgs);if(returnType==="string")ret=Pointer_stringify(ret);if(stack!==0){if(opts&&opts.async){EmterpreterAsync.asyncFinalizers.push((function(){Runtime.stackRestore(stack)}));return}Runtime.stackRestore(stack)}return ret};var sourceRegex=/^function\s*\(([^)]*)\)\s*{\s*([^*]*?)[\s;]*(?:return\s*(.*?)[;\s]*)?}$/;function parseJSFunc(jsfunc){var parsed=jsfunc.toString().match(sourceRegex).slice(1);return{arguments:parsed[0],body:parsed[1],returnValue:parsed[2]}}var JSsource={};for(var fun in JSfuncs){if(JSfuncs.hasOwnProperty(fun)){JSsource[fun]=parseJSFunc(JSfuncs[fun])}}cwrap=function cwrap(ident,returnType,argTypes){argTypes=argTypes||[];var cfunc=getCFunc(ident);var numericArgs=argTypes.every((function(type){return type==="number"}));var numericRet=returnType!=="string";if(numericRet&&numericArgs){return cfunc}var argNames=argTypes.map((function(x,i){return"$"+i}));var funcstr="(function("+argNames.join(",")+") {";var nargs=argTypes.length;if(!numericArgs){funcstr+="var stack = "+JSsource["stackSave"].body+";";for(var i=0;i<nargs;i++){var arg=argNames[i],type=argTypes[i];if(type==="number")continue;var convertCode=JSsource[type+"ToC"];funcstr+="var "+convertCode.arguments+" = "+arg+";";funcstr+=convertCode.body+";";funcstr+=arg+"="+convertCode.returnValue+";"}}var cfuncname=parseJSFunc((function(){return cfunc})).returnValue;funcstr+="var ret = "+cfuncname+"("+argNames.join(",")+");";if(!numericRet){var strgfy=parseJSFunc((function(){return Pointer_stringify})).returnValue;funcstr+="ret = "+strgfy+"(ret);"}if(!numericArgs){funcstr+=JSsource["stackRestore"].body.replace("()","(stack)")+";"}funcstr+="return ret})";return eval(funcstr)}}))();Module["ccall"]=ccall;Module["cwrap"]=cwrap;function setValue(ptr,value,type,noSafe){type=type||"i8";if(type.charAt(type.length-1)==="*")type="i32";switch(type){case"i1":HEAP8[ptr>>0]=value;break;case"i8":HEAP8[ptr>>0]=value;break;case"i16":HEAP16[ptr>>1]=value;break;case"i32":HEAP32[ptr>>2]=value;break;case"i64":tempI64=[value>>>0,(tempDouble=value,+Math_abs(tempDouble)>=+1?tempDouble>+0?(Math_min(+Math_floor(tempDouble/+4294967296),+4294967295)|0)>>>0:~~+Math_ceil((tempDouble- +(~~tempDouble>>>0))/+4294967296)>>>0:0)],HEAP32[ptr>>2]=tempI64[0],HEAP32[ptr+4>>2]=tempI64[1];break;case"float":HEAPF32[ptr>>2]=value;break;case"double":HEAPF64[ptr>>3]=value;break;default:abort("invalid type for setValue: "+type)}}Module["setValue"]=setValue;function getValue(ptr,type,noSafe){type=type||"i8";if(type.charAt(type.length-1)==="*")type="i32";switch(type){case"i1":return HEAP8[ptr>>0];case"i8":return HEAP8[ptr>>0];case"i16":return HEAP16[ptr>>1];case"i32":return HEAP32[ptr>>2];case"i64":return HEAP32[ptr>>2];case"float":return HEAPF32[ptr>>2];case"double":return HEAPF64[ptr>>3];default:abort("invalid type for setValue: "+type)}return null}Module["getValue"]=getValue;var ALLOC_NORMAL=0;var ALLOC_STACK=1;var ALLOC_STATIC=2;var ALLOC_DYNAMIC=3;var ALLOC_NONE=4;Module["ALLOC_NORMAL"]=ALLOC_NORMAL;Module["ALLOC_STACK"]=ALLOC_STACK;Module["ALLOC_STATIC"]=ALLOC_STATIC;Module["ALLOC_DYNAMIC"]=ALLOC_DYNAMIC;Module["ALLOC_NONE"]=ALLOC_NONE;function allocate(slab,types,allocator,ptr){var zeroinit,size;if(typeof slab==="number"){zeroinit=true;size=slab}else{zeroinit=false;size=slab.length}var singleType=typeof types==="string"?types:null;var ret;if(allocator==ALLOC_NONE){ret=ptr}else{ret=[_malloc,Runtime.stackAlloc,Runtime.staticAlloc,Runtime.dynamicAlloc][allocator===undefined?ALLOC_STATIC:allocator](Math.max(size,singleType?1:types.length))}if(zeroinit){var ptr=ret,stop;assert((ret&3)==0);stop=ret+(size&~3);for(;ptr<stop;ptr+=4){HEAP32[ptr>>2]=0}stop=ret+size;while(ptr<stop){HEAP8[ptr++>>0]=0}return ret}if(singleType==="i8"){if(slab.subarray||slab.slice){HEAPU8.set(slab,ret)}else{HEAPU8.set(new Uint8Array(slab),ret)}return ret}var i=0,type,typeSize,previousType;while(i<size){var curr=slab[i];if(typeof curr==="function"){curr=Runtime.getFunctionIndex(curr)}type=singleType||types[i];if(type===0){i++;continue}if(type=="i64")type="i32";setValue(ret+i,curr,type);if(previousType!==type){typeSize=Runtime.getNativeTypeSize(type);previousType=type}i+=typeSize}return ret}Module["allocate"]=allocate;function getMemory(size){if(!staticSealed)return Runtime.staticAlloc(size);if(typeof _sbrk!=="undefined"&&!_sbrk.called||!runtimeInitialized)return Runtime.dynamicAlloc(size);return _malloc(size)}Module["getMemory"]=getMemory;function Pointer_stringify(ptr,length){if(length===0||!ptr)return"";var hasUtf=0;var t;var i=0;while(1){t=HEAPU8[ptr+i>>0];hasUtf|=t;if(t==0&&!length)break;i++;if(length&&i==length)break}if(!length)length=i;var ret="";if(hasUtf<128){var MAX_CHUNK=1024;var curr;while(length>0){curr=String.fromCharCode.apply(String,HEAPU8.subarray(ptr,ptr+Math.min(length,MAX_CHUNK)));ret=ret?ret+curr:curr;ptr+=MAX_CHUNK;length-=MAX_CHUNK}return ret}return Module["UTF8ToString"](ptr)}Module["Pointer_stringify"]=Pointer_stringify;function AsciiToString(ptr){var str="";while(1){var ch=HEAP8[ptr++>>0];if(!ch)return str;str+=String.fromCharCode(ch)}}Module["AsciiToString"]=AsciiToString;function stringToAscii(str,outPtr){return writeAsciiToMemory(str,outPtr,false)}Module["stringToAscii"]=stringToAscii;function UTF8ArrayToString(u8Array,idx){var u0,u1,u2,u3,u4,u5;var str="";while(1){u0=u8Array[idx++];if(!u0)return str;if(!(u0&128)){str+=String.fromCharCode(u0);continue}u1=u8Array[idx++]&63;if((u0&224)==192){str+=String.fromCharCode((u0&31)<<6|u1);continue}u2=u8Array[idx++]&63;if((u0&240)==224){u0=(u0&15)<<12|u1<<6|u2}else{u3=u8Array[idx++]&63;if((u0&248)==240){u0=(u0&7)<<18|u1<<12|u2<<6|u3}else{u4=u8Array[idx++]&63;if((u0&252)==248){u0=(u0&3)<<24|u1<<18|u2<<12|u3<<6|u4}else{u5=u8Array[idx++]&63;u0=(u0&1)<<30|u1<<24|u2<<18|u3<<12|u4<<6|u5}}}if(u0<65536){str+=String.fromCharCode(u0)}else{var ch=u0-65536;str+=String.fromCharCode(55296|ch>>10,56320|ch&1023)}}}Module["UTF8ArrayToString"]=UTF8ArrayToString;function UTF8ToString(ptr){return UTF8ArrayToString(HEAPU8,ptr)}Module["UTF8ToString"]=UTF8ToString;function stringToUTF8Array(str,outU8Array,outIdx,maxBytesToWrite){if(!(maxBytesToWrite>0))return 0;var startIdx=outIdx;var endIdx=outIdx+maxBytesToWrite-1;for(var i=0;i<str.length;++i){var u=str.charCodeAt(i);if(u>=55296&&u<=57343)u=65536+((u&1023)<<10)|str.charCodeAt(++i)&1023;if(u<=127){if(outIdx>=endIdx)break;outU8Array[outIdx++]=u}else if(u<=2047){if(outIdx+1>=endIdx)break;outU8Array[outIdx++]=192|u>>6;outU8Array[outIdx++]=128|u&63}else if(u<=65535){if(outIdx+2>=endIdx)break;outU8Array[outIdx++]=224|u>>12;outU8Array[outIdx++]=128|u>>6&63;outU8Array[outIdx++]=128|u&63}else if(u<=2097151){if(outIdx+3>=endIdx)break;outU8Array[outIdx++]=240|u>>18;outU8Array[outIdx++]=128|u>>12&63;outU8Array[outIdx++]=128|u>>6&63;outU8Array[outIdx++]=128|u&63}else if(u<=67108863){if(outIdx+4>=endIdx)break;outU8Array[outIdx++]=248|u>>24;outU8Array[outIdx++]=128|u>>18&63;outU8Array[outIdx++]=128|u>>12&63;outU8Array[outIdx++]=128|u>>6&63;outU8Array[outIdx++]=128|u&63}else{if(outIdx+5>=endIdx)break;outU8Array[outIdx++]=252|u>>30;outU8Array[outIdx++]=128|u>>24&63;outU8Array[outIdx++]=128|u>>18&63;outU8Array[outIdx++]=128|u>>12&63;outU8Array[outIdx++]=128|u>>6&63;outU8Array[outIdx++]=128|u&63}}outU8Array[outIdx]=0;return outIdx-startIdx}Module["stringToUTF8Array"]=stringToUTF8Array;function stringToUTF8(str,outPtr,maxBytesToWrite){return stringToUTF8Array(str,HEAPU8,outPtr,maxBytesToWrite)}Module["stringToUTF8"]=stringToUTF8;function lengthBytesUTF8(str){var len=0;for(var i=0;i<str.length;++i){var u=str.charCodeAt(i);if(u>=55296&&u<=57343)u=65536+((u&1023)<<10)|str.charCodeAt(++i)&1023;if(u<=127){++len}else if(u<=2047){len+=2}else if(u<=65535){len+=3}else if(u<=2097151){len+=4}else if(u<=67108863){len+=5}else{len+=6}}return len}Module["lengthBytesUTF8"]=lengthBytesUTF8;function UTF16ToString(ptr){var i=0;var str="";while(1){var codeUnit=HEAP16[ptr+i*2>>1];if(codeUnit==0)return str;++i;str+=String.fromCharCode(codeUnit)}}Module["UTF16ToString"]=UTF16ToString;function stringToUTF16(str,outPtr,maxBytesToWrite){if(maxBytesToWrite===undefined){maxBytesToWrite=2147483647}if(maxBytesToWrite<2)return 0;maxBytesToWrite-=2;var startPtr=outPtr;var numCharsToWrite=maxBytesToWrite<str.length*2?maxBytesToWrite/2:str.length;for(var i=0;i<numCharsToWrite;++i){var codeUnit=str.charCodeAt(i);HEAP16[outPtr>>1]=codeUnit;outPtr+=2}HEAP16[outPtr>>1]=0;return outPtr-startPtr}Module["stringToUTF16"]=stringToUTF16;function lengthBytesUTF16(str){return str.length*2}Module["lengthBytesUTF16"]=lengthBytesUTF16;function UTF32ToString(ptr){var i=0;var str="";while(1){var utf32=HEAP32[ptr+i*4>>2];if(utf32==0)return str;++i;if(utf32>=65536){var ch=utf32-65536;str+=String.fromCharCode(55296|ch>>10,56320|ch&1023)}else{str+=String.fromCharCode(utf32)}}}Module["UTF32ToString"]=UTF32ToString;function stringToUTF32(str,outPtr,maxBytesToWrite){if(maxBytesToWrite===undefined){maxBytesToWrite=2147483647}if(maxBytesToWrite<4)return 0;var startPtr=outPtr;var endPtr=startPtr+maxBytesToWrite-4;for(var i=0;i<str.length;++i){var codeUnit=str.charCodeAt(i);if(codeUnit>=55296&&codeUnit<=57343){var trailSurrogate=str.charCodeAt(++i);codeUnit=65536+((codeUnit&1023)<<10)|trailSurrogate&1023}HEAP32[outPtr>>2]=codeUnit;outPtr+=4;if(outPtr+4>endPtr)break}HEAP32[outPtr>>2]=0;return outPtr-startPtr}Module["stringToUTF32"]=stringToUTF32;function lengthBytesUTF32(str){var len=0;for(var i=0;i<str.length;++i){var codeUnit=str.charCodeAt(i);if(codeUnit>=55296&&codeUnit<=57343)++i;len+=4}return len}Module["lengthBytesUTF32"]=lengthBytesUTF32;function demangle(func){var hasLibcxxabi=!!Module["___cxa_demangle"];if(hasLibcxxabi){try{var buf=_malloc(func.length);writeStringToMemory(func.substr(1),buf);var status=_malloc(4);var ret=Module["___cxa_demangle"](buf,0,0,status);if(getValue(status,"i32")===0&&ret){return Pointer_stringify(ret)}}catch(e){}finally{if(buf)_free(buf);if(status)_free(status);if(ret)_free(ret)}}var i=3;var basicTypes={"v":"void","b":"bool","c":"char","s":"short","i":"int","l":"long","f":"float","d":"double","w":"wchar_t","a":"signed char","h":"unsigned char","t":"unsigned short","j":"unsigned int","m":"unsigned long","x":"long long","y":"unsigned long long","z":"..."};var subs=[];var first=true;function dump(x){if(x)Module.print(x);Module.print(func);var pre="";for(var a=0;a<i;a++)pre+=" ";Module.print(pre+"^")}function parseNested(){i++;if(func[i]==="K")i++;var parts=[];while(func[i]!=="E"){if(func[i]==="S"){i++;var next=func.indexOf("_",i);var num=func.substring(i,next)||0;parts.push(subs[num]||"?");i=next+1;continue}if(func[i]==="C"){parts.push(parts[parts.length-1]);i+=2;continue}var size=parseInt(func.substr(i));var pre=size.toString().length;if(!size||!pre){i--;break}var curr=func.substr(i+pre,size);parts.push(curr);subs.push(curr);i+=pre+size}i++;return parts}function parse(rawList,limit,allowVoid){limit=limit||Infinity;var ret="",list=[];function flushList(){return"("+list.join(", ")+")"}var name;if(func[i]==="N"){name=parseNested().join("::");limit--;if(limit===0)return rawList?[name]:name}else{if(func[i]==="K"||first&&func[i]==="L")i++;var size=parseInt(func.substr(i));if(size){var pre=size.toString().length;name=func.substr(i+pre,size);i+=pre+size}}first=false;if(func[i]==="I"){i++;var iList=parse(true);var iRet=parse(true,1,true);ret+=iRet[0]+" "+name+"<"+iList.join(", ")+">"}else{ret=name}paramLoop:while(i<func.length&&limit-->0){var c=func[i++];if(c in basicTypes){list.push(basicTypes[c])}else{switch(c){case"P":list.push(parse(true,1,true)[0]+"*");break;case"R":list.push(parse(true,1,true)[0]+"&");break;case"L":{i++;var end=func.indexOf("E",i);var size=end-i;list.push(func.substr(i,size));i+=size+2;break};case"A":{var size=parseInt(func.substr(i));i+=size.toString().length;if(func[i]!=="_")throw"?";i++;list.push(parse(true,1,true)[0]+" ["+size+"]");break};case"E":break paramLoop;default:ret+="?"+c;break paramLoop}}}if(!allowVoid&&list.length===1&&list[0]==="void")list=[];if(rawList){if(ret){list.push(ret+"?")}return list}else{return ret+flushList()}}var parsed=func;try{if(func=="Object._main"||func=="_main"){return"main()"}if(typeof func==="number")func=Pointer_stringify(func);if(func[0]!=="_")return func;if(func[1]!=="_")return func;if(func[2]!=="Z")return func;switch(func[3]){case"n":return"operator new()";case"d":return"operator delete()"}parsed=parse()}catch(e){parsed+="?"}if(parsed.indexOf("?")>=0&&!hasLibcxxabi){Runtime.warnOnce("warning: a problem occurred in builtin C++ name demangling; build with  -s DEMANGLE_SUPPORT=1  to link in libcxxabi demangling")}return parsed}function demangleAll(text){return text.replace(/__Z[\w\d_]+/g,(function(x){var y=demangle(x);return x===y?x:x+" ["+y+"]"}))}function jsStackTrace(){var err=new Error;if(!err.stack){try{throw new Error(0)}catch(e){err=e}if(!err.stack){return"(no stack trace available)"}}return err.stack.toString()}function stackTrace(){return demangleAll(jsStackTrace())}Module["stackTrace"]=stackTrace;var PAGE_SIZE=4096;function alignMemoryPage(x){if(x%4096>0){x+=4096-x%4096}return x}var HEAP;var HEAP8,HEAPU8,HEAP16,HEAPU16,HEAP32,HEAPU32,HEAPF32,HEAPF64;var STATIC_BASE=0,STATICTOP=0,staticSealed=false;var STACK_BASE=0,STACKTOP=0,STACK_MAX=0;var DYNAMIC_BASE=0,DYNAMICTOP=0;function enlargeMemory(){var OLD_TOTAL_MEMORY=TOTAL_MEMORY;var LIMIT=Math.pow(2,31);if(DYNAMICTOP>=LIMIT)return false;while(TOTAL_MEMORY<=DYNAMICTOP){if(TOTAL_MEMORY<LIMIT/2){TOTAL_MEMORY=alignMemoryPage(2*TOTAL_MEMORY)}else{var last=TOTAL_MEMORY;TOTAL_MEMORY=alignMemoryPage((3*TOTAL_MEMORY+LIMIT)/4);if(TOTAL_MEMORY<=last)return false}}TOTAL_MEMORY=Math.max(TOTAL_MEMORY,16*1024*1024);if(TOTAL_MEMORY>=LIMIT)return false;try{if(ArrayBuffer.transfer){buffer=ArrayBuffer.transfer(buffer,TOTAL_MEMORY)}else{var oldHEAP8=HEAP8;buffer=new ArrayBuffer(TOTAL_MEMORY)}}catch(e){return false}var success=_emscripten_replace_memory(buffer);if(!success)return false;Module["buffer"]=buffer;Module["HEAP8"]=HEAP8=new Int8Array(buffer);Module["HEAP16"]=HEAP16=new Int16Array(buffer);Module["HEAP32"]=HEAP32=new Int32Array(buffer);Module["HEAPU8"]=HEAPU8=new Uint8Array(buffer);Module["HEAPU16"]=HEAPU16=new Uint16Array(buffer);Module["HEAPU32"]=HEAPU32=new Uint32Array(buffer);Module["HEAPF32"]=HEAPF32=new Float32Array(buffer);Module["HEAPF64"]=HEAPF64=new Float64Array(buffer);if(!ArrayBuffer.transfer){HEAP8.set(oldHEAP8)}return true}var byteLength;try{byteLength=Function.prototype.call.bind(Object.getOwnPropertyDescriptor(ArrayBuffer.prototype,"byteLength").get);byteLength(new ArrayBuffer(4))}catch(e){byteLength=(function(buffer){return buffer.byteLength})}var TOTAL_STACK=Module["TOTAL_STACK"]||5242880;var TOTAL_MEMORY=Module["TOTAL_MEMORY"]||67108864;var totalMemory=64*1024;while(totalMemory<TOTAL_MEMORY||totalMemory<2*TOTAL_STACK){if(totalMemory<16*1024*1024){totalMemory*=2}else{totalMemory+=16*1024*1024}}totalMemory=Math.max(totalMemory,16*1024*1024);if(totalMemory!==TOTAL_MEMORY){TOTAL_MEMORY=totalMemory}assert(typeof Int32Array!=="undefined"&&typeof Float64Array!=="undefined"&&!!(new Int32Array(1))["subarray"]&&!!(new Int32Array(1))["set"],"JS engine does not provide full typed array support");var buffer;buffer=new ArrayBuffer(TOTAL_MEMORY);HEAP8=new Int8Array(buffer);HEAP16=new Int16Array(buffer);HEAP32=new Int32Array(buffer);HEAPU8=new Uint8Array(buffer);HEAPU16=new Uint16Array(buffer);HEAPU32=new Uint32Array(buffer);HEAPF32=new Float32Array(buffer);HEAPF64=new Float64Array(buffer);HEAP32[0]=255;assert(HEAPU8[0]===255&&HEAPU8[3]===0,"Typed arrays 2 must be run on a little-endian system");Module["HEAP"]=HEAP;Module["buffer"]=buffer;Module["HEAP8"]=HEAP8;Module["HEAP16"]=HEAP16;Module["HEAP32"]=HEAP32;Module["HEAPU8"]=HEAPU8;Module["HEAPU16"]=HEAPU16;Module["HEAPU32"]=HEAPU32;Module["HEAPF32"]=HEAPF32;Module["HEAPF64"]=HEAPF64;function callRuntimeCallbacks(callbacks){while(callbacks.length>0){var callback=callbacks.shift();if(typeof callback=="function"){callback();continue}var func=callback.func;if(typeof func==="number"){if(callback.arg===undefined){Runtime.dynCall("v",func)}else{Runtime.dynCall("vi",func,[callback.arg])}}else{func(callback.arg===undefined?null:callback.arg)}}}var __ATPRERUN__=[];var __ATINIT__=[];var __ATMAIN__=[];var __ATEXIT__=[];var __ATPOSTRUN__=[];var runtimeInitialized=false;var runtimeExited=false;function preRun(){if(Module["preRun"]){if(typeof Module["preRun"]=="function")Module["preRun"]=[Module["preRun"]];while(Module["preRun"].length){addOnPreRun(Module["preRun"].shift())}}callRuntimeCallbacks(__ATPRERUN__)}function ensureInitRuntime(){if(runtimeInitialized)return;runtimeInitialized=true;callRuntimeCallbacks(__ATINIT__)}function preMain(){callRuntimeCallbacks(__ATMAIN__)}function exitRuntime(){callRuntimeCallbacks(__ATEXIT__);runtimeExited=true}function postRun(){if(Module["postRun"]){if(typeof Module["postRun"]=="function")Module["postRun"]=[Module["postRun"]];while(Module["postRun"].length){addOnPostRun(Module["postRun"].shift())}}callRuntimeCallbacks(__ATPOSTRUN__)}function addOnPreRun(cb){__ATPRERUN__.unshift(cb)}Module["addOnPreRun"]=addOnPreRun;function addOnInit(cb){__ATINIT__.unshift(cb)}Module["addOnInit"]=addOnInit;function addOnPreMain(cb){__ATMAIN__.unshift(cb)}Module["addOnPreMain"]=addOnPreMain;function addOnExit(cb){__ATEXIT__.unshift(cb)}Module["addOnExit"]=addOnExit;function addOnPostRun(cb){__ATPOSTRUN__.unshift(cb)}Module["addOnPostRun"]=addOnPostRun;function intArrayFromString(stringy,dontAddNull,length){var len=length>0?length:lengthBytesUTF8(stringy)+1;var u8array=new Array(len);var numBytesWritten=stringToUTF8Array(stringy,u8array,0,u8array.length);if(dontAddNull)u8array.length=numBytesWritten;return u8array}Module["intArrayFromString"]=intArrayFromString;function intArrayToString(array){var ret=[];for(var i=0;i<array.length;i++){var chr=array[i];if(chr>255){chr&=255}ret.push(String.fromCharCode(chr))}return ret.join("")}Module["intArrayToString"]=intArrayToString;function writeStringToMemory(string,buffer,dontAddNull){var array=intArrayFromString(string,dontAddNull);var i=0;while(i<array.length){var chr=array[i];HEAP8[buffer+i>>0]=chr;i=i+1}}Module["writeStringToMemory"]=writeStringToMemory;function writeArrayToMemory(array,buffer){for(var i=0;i<array.length;i++){HEAP8[buffer++>>0]=array[i]}}Module["writeArrayToMemory"]=writeArrayToMemory;function writeAsciiToMemory(str,buffer,dontAddNull){for(var i=0;i<str.length;++i){HEAP8[buffer++>>0]=str.charCodeAt(i)}if(!dontAddNull)HEAP8[buffer>>0]=0}Module["writeAsciiToMemory"]=writeAsciiToMemory;function unSign(value,bits,ignore){if(value>=0){return value}return bits<=32?2*Math.abs(1<<bits-1)+value:Math.pow(2,bits)+value}function reSign(value,bits,ignore){if(value<=0){return value}var half=bits<=32?Math.abs(1<<bits-1):Math.pow(2,bits-1);if(value>=half&&(bits<=32||value>half)){value=-2*half+value}return value}if(!Math["imul"]||Math["imul"](4294967295,5)!==-5)Math["imul"]=function imul(a,b){var ah=a>>>16;var al=a&65535;var bh=b>>>16;var bl=b&65535;return al*bl+(ah*bl+al*bh<<16)|0};Math.imul=Math["imul"];if(!Math["clz32"])Math["clz32"]=(function(x){x=x>>>0;for(var i=0;i<32;i++){if(x&1<<31-i)return i}return 32});Math.clz32=Math["clz32"];var Math_abs=Math.abs;var Math_cos=Math.cos;var Math_sin=Math.sin;var Math_tan=Math.tan;var Math_acos=Math.acos;var Math_asin=Math.asin;var Math_atan=Math.atan;var Math_atan2=Math.atan2;var Math_exp=Math.exp;var Math_log=Math.log;var Math_sqrt=Math.sqrt;var Math_ceil=Math.ceil;var Math_floor=Math.floor;var Math_pow=Math.pow;var Math_imul=Math.imul;var Math_fround=Math.fround;var Math_min=Math.min;var Math_clz32=Math.clz32;var runDependencies=0;var runDependencyWatcher=null;var dependenciesFulfilled=null;function getUniqueRunDependency(id){return id}function addRunDependency(id){runDependencies++;if(Module["monitorRunDependencies"]){Module["monitorRunDependencies"](runDependencies)}}Module["addRunDependency"]=addRunDependency;function removeRunDependency(id){runDependencies--;if(Module["monitorRunDependencies"]){Module["monitorRunDependencies"](runDependencies)}if(runDependencies==0){if(runDependencyWatcher!==null){clearInterval(runDependencyWatcher);runDependencyWatcher=null}if(dependenciesFulfilled){var callback=dependenciesFulfilled;dependenciesFulfilled=null;callback()}}}Module["removeRunDependency"]=removeRunDependency;Module["preloadedImages"]={};Module["preloadedAudios"]={};var memoryInitializer=null;var ASM_CONSTS=[];STATIC_BASE=8;STATICTOP=STATIC_BASE+795600;__ATINIT__.push();allocate([],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE);allocate([44,66,124,35,100,255,239,63,206,4,147,55,81,230,134,63,201,87,119,237,8,231,115,63,8,49,216,60,72,85,196,190,141,82,239,197,175,26,93,190,79,75,150,130,180,74,73,190,206,4,147,55,81,230,134,191,110,20,118,229,124,255,239,63,190,2,89,23,76,127,252,190,141,82,239,197,175,26,93,62,186,92,192,248,87,85,196,190,213,61,80,36,200,25,210,61,192,143,86,244,8,231,115,191,40,252,144,126,141,121,252,190,190,45,6,62,231,255,239,63,79,75,150,130,180,74,73,62,213,61,80,36,200,25,210,61,56,97,156,141,155,85,196,190,37,201,115,125,31,14,66,191,237,156,102,129,118,135,206,63,237,128,235,138,25,225,219,191,138,238,48,88,55,255,239,63,235,42,147,208,76,230,134,63,26,132,28,8,135,230,115,63,12,31,17,83,34,137,206,191,58,234,232,184,26,217,101,191,7,35,246,9,160,24,137,63,235,42,147,208,76,230,134,191,233,248,162,28,80,255,239,63,8,168,211,46,100,125,252,190,137,181,248,20,0,227,219,63,180,173,102,157,241,125,129,191,174,14,128,184,171,87,97,63,26,132,28,8,135,230,115,191,247,97,247,252,180,122,252,190,137,77,92,117,186,255,239,63,125,150,201,34,100,255,239,63,15,157,207,213,158,230,134,191,92,113,244,103,237,229,115,191,26,77,59,180,100,85,196,62,52,175,17,131,3,27,93,190,173,87,210,121,59,74,73,190,20,1,96,210,158,230,134,63,3,186,248,225,124,255,239,63,250,59,9,220,34,127,252,190,52,175,17,131,3,27,93,62,91,229,82,109,116,85,196,62,232,167,8,110,48,27,210,189,73,225,178,117,237,229,115,63,235,19,108,31,83,119,252,190,10,30,195,64,231,255,239,63,173,87,210,121,59,74,73,62,185,158,187,181,171,23,210,189,73,125,255,4,184,85,196,62,37,201,115,125,31,14,66,191,211,48,124,68,76,137,206,191,34,223,165,212,37,227,219,63,45,6,81,239,144,255,239,63,108,197,229,21,219,230,134,191,214,96,142,220,162,229,115,191,180,174,209,114,160,135,206,63,58,234,232,184,26,217,101,191,180,173,102,157,241,125,129,191,108,197,229,21,219,230,134,63,26,73,6,169,169,255,239,63,26,238,175,96,19,128,252,190,134,170,152,74,63,225,219,191,7,35,246,9,160,24,137,63,174,14,128,184,171,87,97,63,214,96,142,220,162,229,115,63,212,213,62,153,86,117,252,190,150,50,62,6,10,0,240,63],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+10256);allocate([1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,126,132,46,193,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,236,57,250,254,66,46,230,63,2,32,42,250,11,171,252,63,250,44,146,124,167,108,9,64,201,121,68,60,100,38,19,64,202,1,207,58,39,81,26,64,48,204,45,243,225,12,33,64,13,183,252,130,142,53,37,64,207,247,167,33,137,154,41,64,21,77,149,144,117,53,46,64,116,141,58,63,151,128,49,64,67,1,213,22,186,252,51,64,205,50,59,156,90,141,54,64,67,42,22,223,243,48,57,64,71,211,63,166,54,230,59,64,133,26,31,6,255,171,62,64,58,53,47,63,166,192,64,64,82,238,213,242,157,50,66,64,133,57,54,83,129,171,67,64,190,119,237,122,245,42,69,64,169,114,52,100,168,176,70,64,6,79,168,171,79,60,72,64,69,106,133,139,167,205,73,64,20,141,78,3,114,100,75,64,231,101,205,34,118,0,77,64,148,103,124,113,127,161,78,64,238,79,126,182,174,35,80,64,63,2,51,141,241,248,80,64,48,114,145,83,115,208,81,64,229,208,89,139,32,170,82,64,64,90,158,253,230,133,83,64,131,32,148,155,181,99,84,64,74,206,58,99,124,67,85,64,186,150,72,71,44,37,86,64,87,22,225,25,183,8,87,64,88,103,178,121,15,238,87,64,150,61,36,193,40,213,88,64,163,87,82,247,246,189,89,64,17,152,150,194,110,168,90,64,162,43,112,92,133,148,91,64,4,161,156,134,48,130,92,64,237,18,62,129,102,113,93,64,79,186,238,1,30,98,94,64,240,156,166,43,78,84,95,64,160,221,173,67,247,35,96,64,168,30,164,126,123,158,96,64,107,98,98,231,175,25,97,64,89,165,83,200,144,149,97,64,195,110,12,147,26,18,98,64,49,31,235,221,73,143,98,64,53,99,232,97,27,13,99,64,219,29,147,248,139,139,99,64,206,166,51,154,152,10,100,64,25,175,18,92,62,138,100,64,17,135,224,110,122,10,101,64,116,218,57,29,74,139,101,64,71,71,71,202,170,12,102,64,121,121,117,240,153,142,102,64,197,178,67,32,21,17,103,64,89,220,38,255,25,148,103,64,185,111,127,70,166,23,104,64,160,174,160,194,183,155,104,64,99,199,231,81,76,32,105,64,189,164,225,227,97,165,105,64,8,70,126,120,246,42,106,64,38,151,80,31,8,177,106,64,176,215,217,246,148,55,107,64,34,182,223,43,155,190,107,64,246,86,204,248,24,70,108,64,255,144,23,165,12,206,108,64,182,183,184,132,116,86,109,64,113,90,160,247,78,223,109,64,240,107,57,105,154,104,110,64,73,81,241,79,85,242,110,64,131,97,198,44,126,124,111,64,97,52,110,197,137,3,112,64,43,101,139,255,9,73,112,64,99,242,219,19,191,142,112,64,41,177,2,86,168,212,112,64,42,147,248,29,197,26,113,64,54,71,227,199,20,97,113,64,172,225,237,179,150,167,113,64,62,109,35,70,74,238,113,64,213,70,75,230,46,53,114,64,97,41,199,255,67,124,114,64,86,208,114,1,137,195,114,64,86,25,133,93,253,10,115,64,27,143,114,137,160,82,115,64,70,73,209,253,113,154,115,64,247,13,62,54,113,226,115,64,107,163,66,177,157,42,116,64,12,65,61,240,246,114,116,64,102,16,73,119,124,187,116,64,100,175,39,205,45,4,117,64,87,166,43,123,10,77,117,64,236,196,35,13,18,150,117,64,28,90,71,17,68,223,117,64,237,59,35,24,160,40,118,64,97,148,135,180,37,114,118,64,182,105,118,123,212,187,118,64,158,216,18,4,172,5,119,64,190,247,144,231,171,79,119,64,14,92,38,193,211,153,119,64,125,54,251,45,35,228,119,64,104,3,28,205,153,46,120,64,254,196,107,63,55,121,120,64,27,194,150,39,251,195,120,64,95,195,5,42,229,14,121,64,179,200,209,236,244,89,121,64,203,49,184,23,42,165,121,64,93,84,15,84,132,240,121,64,45,123,187,76,3,60,122,64,58,73,36,174,166,135,122,64,192,124,42,38,110,211,122,64,182,13,30,100,89,31,123,64,235,163,180,24,104,107,123,64,10,97,0,246,153,183,123,64,205,250,102,175,238,3,124,64,38,34,153,249,101,80,124,64,12,52,138,138,255,156,124,64,234,48,104,25,187,233,124,64,182,247,147,94,152,54,125,64,18,194,153,19,151,131,125,64,163,222,41,243,182,208,125,64,75,167,17,185,247,29,126,64,200,176,52,34,89,107,126,64,130,49,134,236,218,184,126,64,76,157,2,215,124,6,127,64,39,116,169,161,62,84,127,64,234,65,119,13,32,162,127,64,13,205,95,220,32,240,127,64,93,57,164,104,32,31,128,64,61,87,1,216,63,70,128,64,237,101,163,158,110,109,128,64,57,96,87,159,172,148,128,64,99,150,88,189,249,187,128,64,101,65,78,220,85,227,128,64,41,40,73,224,192,10,129,64,63,86,193,173,58,50,129,64,80,227,147,41,195,89,129,64,194,202,0,57,90,129,129,64,254,210,168,193,255,168,129,64,193,131,139,169,179,208,129,64,254,42,5,215,117,248,129,64,195,239,204,48,70,32,130,64,178,242,242,157,36,72,130,64,141,123,222,5,17,112,130,64,114,51,76,80,11,152,130,64,64,107,76,101,19,192,130,64,222,109,65,45,41,232,130,64,223,221,221,144,76,16,131,64,61,30,35,121,125,56,131,64,192,197,95,207,187,96,131,64,179,28,46,125,7,137,131,64,164,164,114,108,96,177,131,64,207,169,90,135,198,217,131,64,224,221,90,184,57,2,132,64,209,251,45,234,185,42,132,64,143,116,211,7,71,83,132,64,30,36,142,252,224,123,132,64,255,14,227,179,135,164,132,64,148,39,152,25,59,205,132,64,63,27,179,25,251,245,132,64,8,39,120,160,199,30,133,64,136,243,104,154,160,71,133,64,223,119,67,244,133,112,133,64,142,227,0,155,119,153,133,64,232,142,212,123,117,194,133,64,1,242,42,132,127,235,133,64,214,161,168,161,149,20,134,64,139,83,41,194,183,61,134,64,148,229,190,211,229,102,134,64,138,110,176,196,31,144,134,64,152,81,121,131,101,185,134,64,82,88,200,254,182,226,134,64,195,209,126,37,20,12,135,64,151,182,175,230,124,53,135,64,60,210,158,49,241,94,135,64,205,240,191,245,112,136,135,64,179,17,182,34,252,177,135,64,210,158,82,168,146,219,135,64,26,168,148,118,52,5,136,64,116,35,168,125,225,46,136,64,211,48,229,173,153,88,136,64,97,98,207,247,92,130,136,64,157,8,21,76,43,172,136,64,83,130,142,155,4,214,136,64,97,144,61,215,232,255,136,64,19,173,76,240,215,41,137,64,27,103,14,216,209,83,137,64,252,191,252,127,214,125,137,64,212,141,184,217,229,167,137,64,113,224,8,215,255,209,137,64,153,105,218,105,36,252,137,64,108,232,62,132,83,38,138,64,228,151,108,24,141,80,138,64,43,160,189,24,209,122,138,64,245,138,175,119,31,165,138,64,143,186,226,39,120,207,138,64,199,227,25,28,219,249,138,64,108,138,57,71,72,36,139,64,110,128,71,156,191,78,139,64,154,103,106,14,65,121,139,64,184,53,233,144,204,163,139,64,43,187,42,23,98,206,139,64,226,43,181,148,1,249,139,64,153,170,45,253,170,35,140,64,93,214,87,68,94,78,140,64,72,90,21,94,27,121,140,64,77,127,101,62,226,163,140,64,49,192,100,217,178,206,140,64,137,95,76,35,141,249,140,64,185,255,113,16,113,36,141,64,214,60,71,149,94,79,141,64,136,72,89,166,85,122,141,64,184,135,80,56,86,165,141,64,18,50,240,63,96,208,141,64,74,243,21,178,115,251,141,64,43,142,185,131,144,38,142,64,70,129,236,169,182,81,142,64,83,173,217,25,230,124,142,64,42,253,196,200,30,168,142,64,95,15,11,172,96,211,142,64,83,225,32,185,171,254,142,64,227,123,147,229,255,41,143,64,122,161,7,39,93,85,143,64,157,125,57,115,195,128,143,64,239,85,252,191,50,172,143,64,127,60,58,3,171,215,143,64,196,225,121,25,150,1,144,64,56,217,159,34,91,23,144,64,12,94,37,152,36,45,144,64,67,155,43,117,242,66,144,64,209,105,222,180,196,88,144,64,173,45,116,82,155,110,144,64,102,179,45,73,118,132,144,64,88,14,86,148,85,154,144,64,117,119,66,47,57,176,144,64,149,44,82,21,33,198,144,64,87,80,238,65,13,220,144,64,131,202,137,176,253,241,144,64,2,41,161,92,242,7,145,64,69,129,186,65,235,29,145,64,64,82,101,91,232,51,145,64,209,102,58,165,233,73,145,64,174,184,219,26,239,95,145,64,206,83,244,183,248,117,145,64,58,58,56,120,6,140,145,64,95,72,100,87,24,162,145,64,208,25,62,81,46,184,145,64,114,238,147,97,72,206,145,64,24,144,60,132,102,228,145,64,142,56,23,181,136,250,145,64,4,120,11,240,174,16,146,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,128,116,210,26],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+98840);allocate([31,0,0,0,28,0,0,0,31,0,0,0,30,0,0,0,31,0,0,0,30,0,0,0,31,0,0,0,31,0,0,0,30,0,0,0,31,0,0,0,30,0,0,0,31,0,0,0,0,0,0,0,0,0,0,0,26,0,0,0,0,0,0,0,150,48,7,119,44,97,14,238,186,81,9,153,25,196,109,7,143,244,106,112,53,165,99,233,163,149,100,158,50,136,219,14,164,184,220,121,30,233,213,224,136,217,210,151,43,76,182,9,189,124,177,126,7,45,184,231,145,29,191,144,100,16,183,29,242,32,176,106,72,113,185,243,222,65,190,132,125,212,218,26,235,228,221,109,81,181,212,244,199,133,211,131,86,152,108,19,192,168,107,100,122,249,98,253,236,201,101,138,79,92,1,20,217,108,6,99,99,61,15,250,245,13,8,141,200,32,110,59,94,16,105,76,228,65,96,213,114,113,103,162,209,228,3,60,71,212,4,75,253,133,13,210,107,181,10,165,250,168,181,53,108,152,178,66,214,201,187,219,64,249,188,172,227,108,216,50,117,92,223,69,207,13,214,220,89,61,209,171,172,48,217,38,58,0,222,81,128,81,215,200,22,97,208,191,181,244,180,33,35,196,179,86,153,149,186,207,15,165,189,184,158,184,2,40,8,136,5,95,178,217,12,198,36,233,11,177,135,124,111,47,17,76,104,88,171,29,97,193,61,45,102,182,144,65,220,118,6,113,219,1,188,32,210,152,42,16,213,239,137,133,177,113,31,181,182,6,165,228,191,159,51,212,184,232,162,201,7,120,52,249,0,15,142,168,9,150,24,152,14,225,187,13,106,127,45,61,109,8,151,108,100,145,1,92,99,230,244,81,107,107,98,97,108,28,216,48,101,133,78,0,98,242,237,149,6,108,123,165,1,27,193,244,8,130,87,196,15,245,198,217,176,101,80,233,183,18,234,184,190,139,124,136,185,252,223,29,221,98,73,45,218,21,243,124,211,140,101,76,212,251,88,97,178,77,206,81,181,58,116,0,188,163,226,48,187,212,65,165,223,74,215,149,216,61,109,196,209,164,251,244,214,211,106,233,105,67,252,217,110,52,70,136,103,173,208,184,96,218,115,45,4,68,229,29,3,51,95,76,10,170,201,124,13,221,60,113,5,80,170,65,2,39,16,16,11,190,134,32,12,201,37,181,104,87,179,133,111,32,9,212,102,185,159,228,97,206,14,249,222,94,152,201,217,41,34,152,208,176,180,168,215,199,23,61,179,89,129,13,180,46,59,92,189,183,173,108,186,192,32,131,184,237,182,179,191,154,12,226,182,3,154,210,177,116,57,71,213,234,175,119,210,157,21,38,219,4,131,22,220,115,18,11,99,227,132,59,100,148,62,106,109,13,168,90,106,122,11,207,14,228,157,255,9,147,39,174,0,10,177,158,7,125,68,147,15,240,210,163,8,135,104,242,1,30,254,194,6,105,93,87,98,247,203,103,101,128,113,54,108,25,231,6,107,110,118,27,212,254,224,43,211,137,90,122,218,16,204,74,221,103,111,223,185,249,249,239,190,142,67,190,183,23,213,142,176,96,232,163,214,214,126,147,209,161,196,194,216,56,82,242,223,79,241,103,187,209,103,87,188,166,221,6,181,63,75,54,178,72,218,43,13,216,76,27,10,175,246,74,3,54,96,122,4,65,195,239,96,223,85,223,103,168,239,142,110,49,121,190,105,70,140,179,97,203,26,131,102,188,160,210,111,37,54,226,104,82,149,119,12,204,3,71,11,187,185,22,2,34,47,38,5,85,190,59,186,197,40,11,189,178,146,90,180,43,4,106,179,92,167,255,215,194,49,207,208,181,139,158,217,44,29,174,222,91,176,194,100,155,38,242,99,236,156,163,106,117,10,147,109,2,169,6,9,156,63,54,14,235,133,103,7,114,19,87,0,5,130,74,191,149,20,122,184,226,174,43,177,123,56,27,182,12,155,142,210,146,13,190,213,229,183,239,220,124,33,223,219,11,212,210,211,134,66,226,212,241,248,179,221,104,110,131,218,31,205,22,190,129,91,38,185,246,225,119,176,111,119,71,183,24,230,90,8,136,112,106,15,255,202,59,6,102,92,11,1,17,255,158,101,143,105,174,98,248,211,255,107,97,69,207,108,22,120,226,10,160,238,210,13,215,84,131,4,78,194,179,3,57,97,38,103,167,247,22,96,208,77,71,105,73,219,119,110,62,74,106,209,174,220,90,214,217,102,11,223,64,240,59,216,55,83,174,188,169,197,158,187,222,127,207,178,71,233,255,181,48,28,242,189,189,138,194,186,202,48,147,179,83,166,163,180,36,5,54,208,186,147,6,215,205,41,87,222,84,191,103,217,35,46,122,102,179,184,74,97,196,2,27,104,93,148,43,111,42,55,190,11,180,161,142,12,195,27,223,5,90,141,239,2,45,0,0,0,0,65,49,27,25,130,98,54,50,195,83,45,43,4,197,108,100,69,244,119,125,134,167,90,86,199,150,65,79,8,138,217,200,73,187,194,209,138,232,239,250,203,217,244,227,12,79,181,172,77,126,174,181,142,45,131,158,207,28,152,135,81,18,194,74,16,35,217,83,211,112,244,120,146,65,239,97,85,215,174,46,20,230,181,55,215,181,152,28,150,132,131,5,89,152,27,130,24,169,0,155,219,250,45,176,154,203,54,169,93,93,119,230,28,108,108,255,223,63,65,212,158,14,90,205,162,36,132,149,227,21,159,140,32,70,178,167,97,119,169,190,166,225,232,241,231,208,243,232,36,131,222,195,101,178,197,218,170,174,93,93,235,159,70,68,40,204,107,111,105,253,112,118,174,107,49,57,239,90,42,32,44,9,7,11,109,56,28,18,243,54,70,223,178,7,93,198,113,84,112,237,48,101,107,244,247,243,42,187,182,194,49,162,117,145,28,137,52,160,7,144,251,188,159,23,186,141,132,14,121,222,169,37,56,239,178,60,255,121,243,115,190,72,232,106,125,27,197,65,60,42,222,88,5,79,121,240,68,126,98,233,135,45,79,194,198,28,84,219,1,138,21,148,64,187,14,141,131,232,35,166,194,217,56,191,13,197,160,56,76,244,187,33,143,167,150,10,206,150,141,19,9,0,204,92,72,49,215,69,139,98,250,110,202,83,225,119,84,93,187,186,21,108,160,163,214,63,141,136,151,14,150,145,80,152,215,222,17,169,204,199,210,250,225,236,147,203,250,245,92,215,98,114,29,230,121,107,222,181,84,64,159,132,79,89,88,18,14,22,25,35,21,15,218,112,56,36,155,65,35,61,167,107,253,101,230,90,230,124,37,9,203,87,100,56,208,78,163,174,145,1,226,159,138,24,33,204,167,51,96,253,188,42,175,225,36,173,238,208,63,180,45,131,18,159,108,178,9,134,171,36,72,201,234,21,83,208,41,70,126,251,104,119,101,226,246,121,63,47,183,72,36,54,116,27,9,29,53,42,18,4,242,188,83,75,179,141,72,82,112,222,101,121,49,239,126,96,254,243,230,231,191,194,253,254,124,145,208,213,61,160,203,204,250,54,138,131,187,7,145,154,120,84,188,177,57,101,167,168,75,152,131,59,10,169,152,34,201,250,181,9,136,203,174,16,79,93,239,95,14,108,244,70,205,63,217,109,140,14,194,116,67,18,90,243,2,35,65,234,193,112,108,193,128,65,119,216,71,215,54,151,6,230,45,142,197,181,0,165,132,132,27,188,26,138,65,113,91,187,90,104,152,232,119,67,217,217,108,90,30,79,45,21,95,126,54,12,156,45,27,39,221,28,0,62,18,0,152,185,83,49,131,160,144,98,174,139,209,83,181,146,22,197,244,221,87,244,239,196,148,167,194,239,213,150,217,246,233,188,7,174,168,141,28,183,107,222,49,156,42,239,42,133,237,121,107,202,172,72,112,211,111,27,93,248,46,42,70,225,225,54,222,102,160,7,197,127,99,84,232,84,34,101,243,77,229,243,178,2,164,194,169,27,103,145,132,48,38,160,159,41,184,174,197,228,249,159,222,253,58,204,243,214,123,253,232,207,188,107,169,128,253,90,178,153,62,9,159,178,127,56,132,171,176,36,28,44,241,21,7,53,50,70,42,30,115,119,49,7,180,225,112,72,245,208,107,81,54,131,70,122,119,178,93,99,78,215,250,203,15,230,225,210,204,181,204,249,141,132,215,224,74,18,150,175,11,35,141,182,200,112,160,157,137,65,187,132,70,93,35,3,7,108,56,26,196,63,21,49,133,14,14,40,66,152,79,103,3,169,84,126,192,250,121,85,129,203,98,76,31,197,56,129,94,244,35,152,157,167,14,179,220,150,21,170,27,0,84,229,90,49,79,252,153,98,98,215,216,83,121,206,23,79,225,73,86,126,250,80,149,45,215,123,212,28,204,98,19,138,141,45,82,187,150,52,145,232,187,31,208,217,160,6,236,243,126,94,173,194,101,71,110,145,72,108,47,160,83,117,232,54,18,58,169,7,9,35,106,84,36,8,43,101,63,17,228,121,167,150,165,72,188,143,102,27,145,164,39,42,138,189,224,188,203,242,161,141,208,235,98,222,253,192,35,239,230,217,189,225,188,20,252,208,167,13,63,131,138,38,126,178,145,63,185,36,208,112,248,21,203,105,59,70,230,66,122,119,253,91,181,107,101,220,244,90,126,197,55,9,83,238,118,56,72,247,177,174,9,184,240,159,18,161,51,204,63,138,114,253,36,147,0,0,0,0,55,106,194,1,110,212,132,3,89,190,70,2,220,168,9,7,235,194,203,6,178,124,141,4,133,22,79,5,184,81,19,14,143,59,209,15,214,133,151,13,225,239,85,12,100,249,26,9,83,147,216,8,10,45,158,10,61,71,92,11,112,163,38,28,71,201,228,29,30,119,162,31,41,29,96,30,172,11,47,27,155,97,237,26,194,223,171,24,245,181,105,25,200,242,53,18,255,152,247,19,166,38,177,17,145,76,115,16,20,90,60,21,35,48,254,20,122,142,184,22,77,228,122,23,224,70,77,56,215,44,143,57,142,146,201,59,185,248,11,58,60,238,68,63,11,132,134,62,82,58,192,60,101,80,2,61,88,23,94,54,111,125,156,55,54,195,218,53,1,169,24,52,132,191,87,49,179,213,149,48,234,107,211,50,221,1,17,51,144,229,107,36,167,143,169,37,254,49,239,39,201,91,45,38,76,77,98,35,123,39,160,34,34,153,230,32,21,243,36,33,40,180,120,42,31,222,186,43,70,96,252,41,113,10,62,40,244,28,113,45,195,118,179,44,154,200,245,46,173,162,55,47,192,141,154,112,247,231,88,113,174,89,30,115,153,51,220,114,28,37,147,119,43,79,81,118,114,241,23,116,69,155,213,117,120,220,137,126,79,182,75,127,22,8,13,125,33,98,207,124,164,116,128,121,147,30,66,120,202,160,4,122,253,202,198,123,176,46,188,108,135,68,126,109,222,250,56,111,233,144,250,110,108,134,181,107,91,236,119,106,2,82,49,104,53,56,243,105,8,127,175,98,63,21,109,99,102,171,43,97,81,193,233,96,212,215,166,101,227,189,100,100,186,3,34,102,141,105,224,103,32,203,215,72,23,161,21,73,78,31,83,75,121,117,145,74,252,99,222,79,203,9,28,78,146,183,90,76,165,221,152,77,152,154,196,70,175,240,6,71,246,78,64,69,193,36,130,68,68,50,205,65,115,88,15,64,42,230,73,66,29,140,139,67,80,104,241,84,103,2,51,85,62,188,117,87,9,214,183,86,140,192,248,83,187,170,58,82,226,20,124,80,213,126,190,81,232,57,226,90,223,83,32,91,134,237,102,89,177,135,164,88,52,145,235,93,3,251,41,92,90,69,111,94,109,47,173,95,128,27,53,225,183,113,247,224,238,207,177,226,217,165,115,227,92,179,60,230,107,217,254,231,50,103,184,229,5,13,122,228,56,74,38,239,15,32,228,238,86,158,162,236,97,244,96,237,228,226,47,232,211,136,237,233,138,54,171,235,189,92,105,234,240,184,19,253,199,210,209,252,158,108,151,254,169,6,85,255,44,16,26,250,27,122,216,251,66,196,158,249,117,174,92,248,72,233,0,243,127,131,194,242,38,61,132,240,17,87,70,241,148,65,9,244,163,43,203,245,250,149,141,247,205,255,79,246,96,93,120,217,87,55,186,216,14,137,252,218,57,227,62,219,188,245,113,222,139,159,179,223,210,33,245,221,229,75,55,220,216,12,107,215,239,102,169,214,182,216,239,212,129,178,45,213,4,164,98,208,51,206,160,209,106,112,230,211,93,26,36,210,16,254,94,197,39,148,156,196,126,42,218,198,73,64,24,199,204,86,87,194,251,60,149,195,162,130,211,193,149,232,17,192,168,175,77,203,159,197,143,202,198,123,201,200,241,17,11,201,116,7,68,204,67,109,134,205,26,211,192,207,45,185,2,206,64,150,175,145,119,252,109,144,46,66,43,146,25,40,233,147,156,62,166,150,171,84,100,151,242,234,34,149,197,128,224,148,248,199,188,159,207,173,126,158,150,19,56,156,161,121,250,157,36,111,181,152,19,5,119,153,74,187,49,155,125,209,243,154,48,53,137,141,7,95,75,140,94,225,13,142,105,139,207,143,236,157,128,138,219,247,66,139,130,73,4,137,181,35,198,136,136,100,154,131,191,14,88,130,230,176,30,128,209,218,220,129,84,204,147,132,99,166,81,133,58,24,23,135,13,114,213,134,160,208,226,169,151,186,32,168,206,4,102,170,249,110,164,171,124,120,235,174,75,18,41,175,18,172,111,173,37,198,173,172,24,129,241,167,47,235,51,166,118,85,117,164,65,63,183,165,196,41,248,160,243,67,58,161,170,253,124,163,157,151,190,162,208,115,196,181,231,25,6,180,190,167,64,182,137,205,130,183,12,219,205,178,59,177,15,179,98,15,73,177,85,101,139,176,104,34,215,187,95,72,21,186,6,246,83,184,49,156,145,185,180,138,222,188,131,224,28,189,218,94,90,191,237,52,152,190,0,0,0,0,101,103,188,184,139,200,9,170,238,175,181,18,87,151,98,143,50,240,222,55,220,95,107,37,185,56,215,157,239,40,180,197,138,79,8,125,100,224,189,111,1,135,1,215,184,191,214,74,221,216,106,242,51,119,223,224,86,16,99,88,159,87,25,80,250,48,165,232,20,159,16,250,113,248,172,66,200,192,123,223,173,167,199,103,67,8,114,117,38,111,206,205,112,127,173,149,21,24,17,45,251,183,164,63,158,208,24,135,39,232,207,26,66,143,115,162,172,32,198,176,201,71,122,8,62,175,50,160,91,200,142,24,181,103,59,10,208,0,135,178,105,56,80,47,12,95,236,151,226,240,89,133,135,151,229,61,209,135,134,101,180,224,58,221,90,79,143,207,63,40,51,119,134,16,228,234,227,119,88,82,13,216,237,64,104,191,81,248,161,248,43,240,196,159,151,72,42,48,34,90,79,87,158,226,246,111,73,127,147,8,245,199,125,167,64,213,24,192,252,109,78,208,159,53,43,183,35,141,197,24,150,159,160,127,42,39,25,71,253,186,124,32,65,2,146,143,244,16,247,232,72,168,61,88,20,155,88,63,168,35,182,144,29,49,211,247,161,137,106,207,118,20,15,168,202,172,225,7,127,190,132,96,195,6,210,112,160,94,183,23,28,230,89,184,169,244,60,223,21,76,133,231,194,209,224,128,126,105,14,47,203,123,107,72,119,195,162,15,13,203,199,104,177,115,41,199,4,97,76,160,184,217,245,152,111,68,144,255,211,252,126,80,102,238,27,55,218,86,77,39,185,14,40,64,5,182,198,239,176,164,163,136,12,28,26,176,219,129,127,215,103,57,145,120,210,43,244,31,110,147,3,247,38,59,102,144,154,131,136,63,47,145,237,88,147,41,84,96,68,180,49,7,248,12,223,168,77,30,186,207,241,166,236,223,146,254,137,184,46,70,103,23,155,84,2,112,39,236,187,72,240,113,222,47,76,201,48,128,249,219,85,231,69,99,156,160,63,107,249,199,131,211,23,104,54,193,114,15,138,121,203,55,93,228,174,80,225,92,64,255,84,78,37,152,232,246,115,136,139,174,22,239,55,22,248,64,130,4,157,39,62,188,36,31,233,33,65,120,85,153,175,215,224,139,202,176,92,51,59,182,89,237,94,209,229,85,176,126,80,71,213,25,236,255,108,33,59,98,9,70,135,218,231,233,50,200,130,142,142,112,212,158,237,40,177,249,81,144,95,86,228,130,58,49,88,58,131,9,143,167,230,110,51,31,8,193,134,13,109,166,58,181,164,225,64,189,193,134,252,5,47,41,73,23,74,78,245,175,243,118,34,50,150,17,158,138,120,190,43,152,29,217,151,32,75,201,244,120,46,174,72,192,192,1,253,210,165,102,65,106,28,94,150,247,121,57,42,79,151,150,159,93,242,241,35,229,5,25,107,77,96,126,215,245,142,209,98,231,235,182,222,95,82,142,9,194,55,233,181,122,217,70,0,104,188,33,188,208,234,49,223,136,143,86,99,48,97,249,214,34,4,158,106,154,189,166,189,7,216,193,1,191,54,110,180,173,83,9,8,21,154,78,114,29,255,41,206,165,17,134,123,183,116,225,199,15,205,217,16,146,168,190,172,42,70,17,25,56,35,118,165,128,117,102,198,216,16,1,122,96,254,174,207,114,155,201,115,202,34,241,164,87,71,150,24,239,169,57,173,253,204,94,17,69,6,238,77,118,99,137,241,206,141,38,68,220,232,65,248,100,81,121,47,249,52,30,147,65,218,177,38,83,191,214,154,235,233,198,249,179,140,161,69,11,98,14,240,25,7,105,76,161,190,81,155,60,219,54,39,132,53,153,146,150,80,254,46,46,153,185,84,38,252,222,232,158,18,113,93,140,119,22,225,52,206,46,54,169,171,73,138,17,69,230,63,3,32,129,131,187,118,145,224,227,19,246,92,91,253,89,233,73,152,62,85,241,33,6,130,108,68,97,62,212,170,206,139,198,207,169,55,126,56,65,127,214,93,38,195,110,179,137,118,124,214,238,202,196,111,214,29,89,10,177,161,225,228,30,20,243,129,121,168,75,215,105,203,19,178,14,119,171,92,161,194,185,57,198,126,1,128,254,169,156,229,153,21,36,11,54,160,54,110,81,28,142,167,22,102,134,194,113,218,62,44,222,111,44,73,185,211,148,240,129,4,9,149,230,184,177,123,73,13,163,30,46,177,27,72,62,210,67,45,89,110,251,195,246,219,233,166,145,103,81,31,169,176,204,122,206,12,116,148,97,185,102,241,6,5,222,0,0,0,0,119,7,48,150,238,14,97,44,153,9,81,186,7,109,196,25,112,106,244,143,233,99,165,53,158,100,149,163,14,219,136,50,121,220,184,164,224,213,233,30,151,210,217,136,9,182,76,43,126,177,124,189,231,184,45,7,144,191,29,145,29,183,16,100,106,176,32,242,243,185,113,72,132,190,65,222,26,218,212,125,109,221,228,235,244,212,181,81,131,211,133,199,19,108,152,86,100,107,168,192,253,98,249,122,138,101,201,236,20,1,92,79,99,6,108,217,250,15,61,99,141,8,13,245,59,110,32,200,76,105,16,94,213,96,65,228,162,103,113,114,60,3,228,209,75,4,212,71,210,13,133,253,165,10,181,107,53,181,168,250,66,178,152,108,219,187,201,214,172,188,249,64,50,216,108,227,69,223,92,117,220,214,13,207,171,209,61,89,38,217,48,172,81,222,0,58,200,215,81,128,191,208,97,22,33,180,244,181,86,179,196,35,207,186,149,153,184,189,165,15,40,2,184,158,95,5,136,8,198,12,217,178,177,11,233,36,47,111,124,135,88,104,76,17,193,97,29,171,182,102,45,61,118,220,65,144,1,219,113,6,152,210,32,188,239,213,16,42,113,177,133,137,6,182,181,31,159,191,228,165,232,184,212,51,120,7,201,162,15,0,249,52,150,9,168,142,225,14,152,24,127,106,13,187,8,109,61,45,145,100,108,151,230,99,92,1,107,107,81,244,28,108,97,98,133,101,48,216,242,98,0,78,108,6,149,237,27,1,165,123,130,8,244,193,245,15,196,87,101,176,217,198,18,183,233,80,139,190,184,234,252,185,136,124,98,221,29,223,21,218,45,73,140,211,124,243,251,212,76,101,77,178,97,88,58,181,81,206,163,188,0,116,212,187,48,226,74,223,165,65,61,216,149,215,164,209,196,109,211,214,244,251,67,105,233,106,52,110,217,252,173,103,136,70,218,96,184,208,68,4,45,115,51,3,29,229,170,10,76,95,221,13,124,201,80,5,113,60,39,2,65,170,190,11,16,16,201,12,32,134,87,104,181,37,32,111,133,179,185,102,212,9,206,97,228,159,94,222,249,14,41,217,201,152,176,208,152,34,199,215,168,180,89,179,61,23,46,180,13,129,183,189,92,59,192,186,108,173,237,184,131,32,154,191,179,182,3,182,226,12,116,177,210,154,234,213,71,57,157,210,119,175,4,219,38,21,115,220,22,131,227,99,11,18,148,100,59,132,13,109,106,62,122,106,90,168,228,14,207,11,147,9,255,157,10,0,174,39,125,7,158,177,240,15,147,68,135,8,163,210,30,1,242,104,105,6,194,254,247,98,87,93,128,101,103,203,25,108,54,113,110,107,6,231,254,212,27,118,137,211,43,224,16,218,122,90,103,221,74,204,249,185,223,111,142,190,239,249,23,183,190,67,96,176,142,213,214,214,163,232,161,209,147,126,56,216,194,196,79,223,242,82,209,187,103,241,166,188,87,103,63,181,6,221,72,178,54,75,216,13,43,218,175,10,27,76,54,3,74,246,65,4,122,96,223,96,239,195,168,103,223,85,49,110,142,239,70,105,190,121,203,97,179,140,188,102,131,26,37,111,210,160,82,104,226,54,204,12,119,149,187,11,71,3,34,2,22,185,85,5,38,47,197,186,59,190,178,189,11,40,43,180,90,146,92,179,106,4,194,215,255,167,181,208,207,49,44,217,158,139,91,222,174,29,155,100,194,176,236,99,242,38,117,106,163,156,2,109,147,10,156,9,6,169,235,14,54,63,114,7,103,133,5,0,87,19,149,191,74,130,226,184,122,20,123,177,43,174,12,182,27,56,146,210,142,155,229,213,190,13,124,220,239,183,11,219,223,33,134,211,210,212,241,212,226,66,104,221,179,248,31,218,131,110,129,190,22,205,246,185,38,91,111,176,119,225,24,183,71,119,136,8,90,230,255,15,106,112,102,6,59,202,17,1,11,92,143,101,158,255,248,98,174,105,97,107,255,211,22,108,207,69,160,10,226,120,215,13,210,238,78,4,131,84,57,3,179,194,167,103,38,97,208,96,22,247,73,105,71,77,62,110,119,219,174,209,106,74,217,214,90,220,64,223,11,102,55,216,59,240,169,188,174,83,222,187,158,197,71,178,207,127,48,181,255,233,189,189,242,28,202,186,194,138,83,179,147,48,36,180,163,166,186,208,54,5,205,215,6,147,84,222,87,41,35,217,103,191,179,102,122,46,196,97,74,184,93,104,27,2,42,111,43,148,180,11,190,55,195,12,142,161,90,5,223,27,45,2,239,141,0,0,0,0,25,27,49,65,50,54,98,130,43,45,83,195,100,108,197,4,125,119,244,69,86,90,167,134,79,65,150,199,200,217,138,8,209,194,187,73,250,239,232,138,227,244,217,203,172,181,79,12,181,174,126,77,158,131,45,142,135,152,28,207,74,194,18,81,83,217,35,16,120,244,112,211,97,239,65,146,46,174,215,85,55,181,230,20,28,152,181,215,5,131,132,150,130,27,152,89,155,0,169,24,176,45,250,219,169,54,203,154,230,119,93,93,255,108,108,28,212,65,63,223,205,90,14,158,149,132,36,162,140,159,21,227,167,178,70,32,190,169,119,97,241,232,225,166,232,243,208,231,195,222,131,36,218,197,178,101,93,93,174,170,68,70,159,235,111,107,204,40,118,112,253,105,57,49,107,174,32,42,90,239,11,7,9,44,18,28,56,109,223,70,54,243,198,93,7,178,237,112,84,113,244,107,101,48,187,42,243,247,162,49,194,182,137,28,145,117,144,7,160,52,23,159,188,251,14,132,141,186,37,169,222,121,60,178,239,56,115,243,121,255,106,232,72,190,65,197,27,125,88,222,42,60,240,121,79,5,233,98,126,68,194,79,45,135,219,84,28,198,148,21,138,1,141,14,187,64,166,35,232,131,191,56,217,194,56,160,197,13,33,187,244,76,10,150,167,143,19,141,150,206,92,204,0,9,69,215,49,72,110,250,98,139,119,225,83,202,186,187,93,84,163,160,108,21,136,141,63,214,145,150,14,151,222,215,152,80,199,204,169,17,236,225,250,210,245,250,203,147,114,98,215,92,107,121,230,29,64,84,181,222,89,79,132,159,22,14,18,88,15,21,35,25,36,56,112,218,61,35,65,155,101,253,107,167,124,230,90,230,87,203,9,37,78,208,56,100,1,145,174,163,24,138,159,226,51,167,204,33,42,188,253,96,173,36,225,175,180,63,208,238,159,18,131,45,134,9,178,108,201,72,36,171,208,83,21,234,251,126,70,41,226,101,119,104,47,63,121,246,54,36,72,183,29,9,27,116,4,18,42,53,75,83,188,242,82,72,141,179,121,101,222,112,96,126,239,49,231,230,243,254,254,253,194,191,213,208,145,124,204,203,160,61,131,138,54,250,154,145,7,187,177,188,84,120,168,167,101,57,59,131,152,75,34,152,169,10,9,181,250,201,16,174,203,136,95,239,93,79,70,244,108,14,109,217,63,205,116,194,14,140,243,90,18,67,234,65,35,2,193,108,112,193,216,119,65,128,151,54,215,71,142,45,230,6,165,0,181,197,188,27,132,132,113,65,138,26,104,90,187,91,67,119,232,152,90,108,217,217,21,45,79,30,12,54,126,95,39,27,45,156,62,0,28,221,185,152,0,18,160,131,49,83,139,174,98,144,146,181,83,209,221,244,197,22,196,239,244,87,239,194,167,148,246,217,150,213,174,7,188,233,183,28,141,168,156,49,222,107,133,42,239,42,202,107,121,237,211,112,72,172,248,93,27,111,225,70,42,46,102,222,54,225,127,197,7,160,84,232,84,99,77,243,101,34,2,178,243,229,27,169,194,164,48,132,145,103,41,159,160,38,228,197,174,184,253,222,159,249,214,243,204,58,207,232,253,123,128,169,107,188,153,178,90,253,178,159,9,62,171,132,56,127,44,28,36,176,53,7,21,241,30,42,70,50,7,49,119,115,72,112,225,180,81,107,208,245,122,70,131,54,99,93,178,119,203,250,215,78,210,225,230,15,249,204,181,204,224,215,132,141,175,150,18,74,182,141,35,11,157,160,112,200,132,187,65,137,3,35,93,70,26,56,108,7,49,21,63,196,40,14,14,133,103,79,152,66,126,84,169,3,85,121,250,192,76,98,203,129,129,56,197,31,152,35,244,94,179,14,167,157,170,21,150,220,229,84,0,27,252,79,49,90,215,98,98,153,206,121,83,216,73,225,79,23,80,250,126,86,123,215,45,149,98,204,28,212,45,141,138,19,52,150,187,82,31,187,232,145,6,160,217,208,94,126,243,236,71,101,194,173,108,72,145,110,117,83,160,47,58,18,54,232,35,9,7,169,8,36,84,106,17,63,101,43,150,167,121,228,143,188,72,165,164,145,27,102,189,138,42,39,242,203,188,224,235,208,141,161,192,253,222,98,217,230,239,35,20,188,225,189,13,167,208,252,38,138,131,63,63,145,178,126,112,208,36,185,105,203,21,248,66,230,70,59,91,253,119,122,220,101,107,181,197,126,90,244,238,83,9,55,247,72,56,118,184,9,174,177,161,18,159,240,138,63,204,51,147,36,253,114,0,0,0,0,1,194,106,55,3,132,212,110,2,70,190,89,7,9,168,220,6,203,194,235,4,141,124,178,5,79,22,133,14,19,81,184,15,209,59,143,13,151,133,214,12,85,239,225,9,26,249,100,8,216,147,83,10,158,45,10,11,92,71,61,28,38,163,112,29,228,201,71,31,162,119,30,30,96,29,41,27,47,11,172,26,237,97,155,24,171,223,194,25,105,181,245,18,53,242,200,19,247,152,255,17,177,38,166,16,115,76,145,21,60,90,20,20,254,48,35,22,184,142,122,23,122,228,77,56,77,70,224,57,143,44,215,59,201,146,142,58,11,248,185,63,68,238,60,62,134,132,11,60,192,58,82,61,2,80,101,54,94,23,88,55,156,125,111,53,218,195,54,52,24,169,1,49,87,191,132,48,149,213,179,50,211,107,234,51,17,1,221,36,107,229,144,37,169,143,167,39,239,49,254,38,45,91,201,35,98,77,76,34,160,39,123,32,230,153,34,33,36,243,21,42,120,180,40,43,186,222,31,41,252,96,70,40,62,10,113,45,113,28,244,44,179,118,195,46,245,200,154,47,55,162,173,112,154,141,192,113,88,231,247,115,30,89,174,114,220,51,153,119,147,37,28,118,81,79,43,116,23,241,114,117,213,155,69,126,137,220,120,127,75,182,79,125,13,8,22,124,207,98,33,121,128,116,164,120,66,30,147,122,4,160,202,123,198,202,253,108,188,46,176,109,126,68,135,111,56,250,222,110,250,144,233,107,181,134,108,106,119,236,91,104,49,82,2,105,243,56,53,98,175,127,8,99,109,21,63,97,43,171,102,96,233,193,81,101,166,215,212,100,100,189,227,102,34,3,186,103,224,105,141,72,215,203,32,73,21,161,23,75,83,31,78,74,145,117,121,79,222,99,252,78,28,9,203,76,90,183,146,77,152,221,165,70,196,154,152,71,6,240,175,69,64,78,246,68,130,36,193,65,205,50,68,64,15,88,115,66,73,230,42,67,139,140,29,84,241,104,80,85,51,2,103,87,117,188,62,86,183,214,9,83,248,192,140,82,58,170,187,80,124,20,226,81,190,126,213,90,226,57,232,91,32,83,223,89,102,237,134,88,164,135,177,93,235,145,52,92,41,251,3,94,111,69,90,95,173,47,109,225,53,27,128,224,247,113,183,226,177,207,238,227,115,165,217,230,60,179,92,231,254,217,107,229,184,103,50,228,122,13,5,239,38,74,56,238,228,32,15,236,162,158,86,237,96,244,97,232,47,226,228,233,237,136,211,235,171,54,138,234,105,92,189,253,19,184,240,252,209,210,199,254,151,108,158,255,85,6,169,250,26,16,44,251,216,122,27,249,158,196,66,248,92,174,117,243,0,233,72,242,194,131,127,240,132,61,38,241,70,87,17,244,9,65,148,245,203,43,163,247,141,149,250,246,79,255,205,217,120,93,96,216,186,55,87,218,252,137,14,219,62,227,57,222,113,245,188,223,179,159,139,221,245,33,210,220,55,75,229,215,107,12,216,214,169,102,239,212,239,216,182,213,45,178,129,208,98,164,4,209,160,206,51,211,230,112,106,210,36,26,93,197,94,254,16,196,156,148,39,198,218,42,126,199,24,64,73,194,87,86,204,195,149,60,251,193,211,130,162,192,17,232,149,203,77,175,168,202,143,197,159,200,201,123,198,201,11,17,241,204,68,7,116,205,134,109,67,207,192,211,26,206,2,185,45,145,175,150,64,144,109,252,119,146,43,66,46,147,233,40,25,150,166,62,156,151,100,84,171,149,34,234,242,148,224,128,197,159,188,199,248,158,126,173,207,156,56,19,150,157,250,121,161,152,181,111,36,153,119,5,19,155,49,187,74,154,243,209,125,141,137,53,48,140,75,95,7,142,13,225,94,143,207,139,105,138,128,157,236,139,66,247,219,137,4,73,130,136,198,35,181,131,154,100,136,130,88,14,191,128,30,176,230,129,220,218,209,132,147,204,84,133,81,166,99,135,23,24,58,134,213,114,13,169,226,208,160,168,32,186,151,170,102,4,206,171,164,110,249,174,235,120,124,175,41,18,75,173,111,172,18,172,173,198,37,167,241,129,24,166,51,235,47,164,117,85,118,165,183,63,65,160,248,41,196,161,58,67,243,163,124,253,170,162,190,151,157,181,196,115,208,180,6,25,231,182,64,167,190,183,130,205,137,178,205,219,12,179,15,177,59,177,73,15,98,176,139,101,85,187,215,34,104,186,21,72,95,184,83,246,6,185,145,156,49,188,222,138,180,189,28,224,131,191,90,94,218,190,152,52,237,0,0,0,0,184,188,103,101,170,9,200,139,18,181,175,238,143,98,151,87,55,222,240,50,37,107,95,220,157,215,56,185,197,180,40,239,125,8,79,138,111,189,224,100,215,1,135,1,74,214,191,184,242,106,216,221,224,223,119,51,88,99,16,86,80,25,87,159,232,165,48,250,250,16,159,20,66,172,248,113,223,123,192,200,103,199,167,173,117,114,8,67,205,206,111,38,149,173,127,112,45,17,24,21,63,164,183,251,135,24,208,158,26,207,232,39,162,115,143,66,176,198,32,172,8,122,71,201,160,50,175,62,24,142,200,91,10,59,103,181,178,135,0,208,47,80,56,105,151,236,95,12,133,89,240,226,61,229,151,135,101,134,135,209,221,58,224,180,207,143,79,90,119,51,40,63,234,228,16,134,82,88,119,227,64,237,216,13,248,81,191,104,240,43,248,161,72,151,159,196,90,34,48,42,226,158,87,79,127,73,111,246,199,245,8,147,213,64,167,125,109,252,192,24,53,159,208,78,141,35,183,43,159,150,24,197,39,42,127,160,186,253,71,25,2,65,32,124,16,244,143,146,168,72,232,247,155,20,88,61,35,168,63,88,49,29,144,182,137,161,247,211,20,118,207,106,172,202,168,15,190,127,7,225,6,195,96,132,94,160,112,210,230,28,23,183,244,169,184,89,76,21,223,60,209,194,231,133,105,126,128,224,123,203,47,14,195,119,72,107,203,13,15,162,115,177,104,199,97,4,199,41,217,184,160,76,68,111,152,245,252,211,255,144,238,102,80,126,86,218,55,27,14,185,39,77,182,5,64,40,164,176,239,198,28,12,136,163,129,219,176,26,57,103,215,127,43,210,120,145,147,110,31,244,59,38,247,3,131,154,144,102,145,47,63,136,41,147,88,237,180,68,96,84,12,248,7,49,30,77,168,223,166,241,207,186,254,146,223,236,70,46,184,137,84,155,23,103,236,39,112,2,113,240,72,187,201,76,47,222,219,249,128,48,99,69,231,85,107,63,160,156,211,131,199,249,193,54,104,23,121,138,15,114,228,93,55,203,92,225,80,174,78,84,255,64,246,232,152,37,174,139,136,115,22,55,239,22,4,130,64,248,188,62,39,157,33,233,31,36,153,85,120,65,139,224,215,175,51,92,176,202,237,89,182,59,85,229,209,94,71,80,126,176,255,236,25,213,98,59,33,108,218,135,70,9,200,50,233,231,112,142,142,130,40,237,158,212,144,81,249,177,130,228,86,95,58,88,49,58,167,143,9,131,31,51,110,230,13,134,193,8,181,58,166,109,189,64,225,164,5,252,134,193,23,73,41,47,175,245,78,74,50,34,118,243,138,158,17,150,152,43,190,120,32,151,217,29,120,244,201,75,192,72,174,46,210,253,1,192,106,65,102,165,247,150,94,28,79,42,57,121,93,159,150,151,229,35,241,242,77,107,25,5,245,215,126,96,231,98,209,142,95,222,182,235,194,9,142,82,122,181,233,55,104,0,70,217,208,188,33,188,136,223,49,234,48,99,86,143,34,214,249,97,154,106,158,4,7,189,166,189,191,1,193,216,173,180,110,54,21,8,9,83,29,114,78,154,165,206,41,255,183,123,134,17,15,199,225,116,146,16,217,205,42,172,190,168,56,25,17,70,128,165,118,35,216,198,102,117,96,122,1,16,114,207,174,254,202,115,201,155,87,164,241,34,239,24,150,71,253,173,57,169,69,17,94,204,118,77,238,6,206,241,137,99,220,68,38,141,100,248,65,232,249,47,121,81,65,147,30,52,83,38,177,218,235,154,214,191,179,249,198,233,11,69,161,140,25,240,14,98,161,76,105,7,60,155,81,190,132,39,54,219,150,146,153,53,46,46,254,80,38,84,185,153,158,232,222,252,140,93,113,18,52,225,22,119,169,54,46,206,17,138,73,171,3,63,230,69,187,131,129,32,227,224,145,118,91,92,246,19,73,233,89,253,241,85,62,152,108,130,6,33,212,62,97,68,198,139,206,170,126,55,169,207,214,127,65,56,110,195,38,93,124,118,137,179,196,202,238,214,89,29,214,111,225,161,177,10,243,20,30,228,75,168,121,129,19,203,105,215,171,119,14,178,185,194,161,92,1,126,198,57,156,169,254,128,36,21,153,229,54,160,54,11,142,28,81,110,134,102,22,167,62,218,113,194,44,111,222,44,148,211,185,73,9,4,129,240,177,184,230,149,163,13,73,123,27,177,46,30,67,210,62,72,251,110,89,45,233,219,246,195,81,103,145,166,204,176,169,31,116,12,206,122,102,185,97,148,222,5,6,241,0,0,0,0,0,0,0,0,1,0,0,0,4,0,4,0,8,0,4,0,2,0,0,0,4,0,5,0,16,0,8,0,2,0,0,0,4,0,6,0,32,0,32,0,2,0,0,0,4,0,4,0,16,0,16,0,3,0,0,0,8,0,16,0,32,0,32,0,3,0,0,0,8,0,16,0,128,0,128,0,3,0,0,0,8,0,32,0,128,0,0,1,3,0,0,0,32,0,128,0,2,1,0,4,3,0,0,0,32,0,2,1,2,1,0,16,3,0,0,0,186,245,1,0,56,183,1,0,1,1,0,0,30,1,0,0,15,0,0,0,58,250,1,0,32,184,1,0,0,0,0,0,30,0,0,0,15,0,0,0,0,0,0,0,16,185,1,0,0,0,0,0,19,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,10,0,0,0,12,0,0,0,14,0,0,0,16,0,0,0,20,0,0,0,24,0,0,0,28,0,0,0,32,0,0,0,40,0,0,0,48,0,0,0,56,0,0,0,64,0,0,0,80,0,0,0,96,0,0,0,112,0,0,0,128,0,0,0,160,0,0,0,192,0,0,0,224,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,6,0,0,0,6,0,0,0,7,0,0,0,7,0,0,0,8,0,0,0,8,0,0,0,9,0,0,0,9,0,0,0,10,0,0,0,10,0,0,0,11,0,0,0,11,0,0,0,12,0,0,0,12,0,0,0,13,0,0,0,13,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,6,0,0,0,8,0,0,0,12,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,48,0,0,0,64,0,0,0,96,0,0,0,128,0,0,0,192,0,0,0,0,1,0,0,128,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,6,0,0,0,8,0,0,0,12,0,0,0,16,0,0,0,24,0,0,0,32,0,0,0,48,0,0,0,64,0,0,0,96,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,3,0,0,0,7,0,0,0,43,72,9,0,59,72,9,0,201,35,12,0,70,72,9,0,81,72,9,0,94,72,9,0,105,72,9,0,125,72,9,0,138,72,9,0,201,35,12,0,4,0,0,0,0,0,0,0,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,0,0,0,1],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+104e3);allocate([95,199,11,0,137,205,11,0,154,202,11,0,72,51,11,0,16,229,11,0,79,51,11,0,243,228,11,0,86,51,11,0,35,193,11,0,93,51,11,0,143,229,11,0,108,51,11,0,100,51,11,0,123,51,11,0,115,51,11,0,137,51,11,0,130,51,11,0,152,51,11,0,144,51,11,0,166,51,11,0,159,51,11,0,181,51,11,0,173,51,11,0,195,51,11,0,188,51,11,0,210,51,11,0,202,51,11,0,224,51,11,0,217,51,11,0,239,51,11,0,231,51,11,0,128,52,11,0,135,52,11,0,142,52,11,0,149,52,11,0,253,51,11,0,246,51,11,0,11,52,11,0,4,52,11,0,25,52,11,0,18,52,11,0,39,52,11,0,32,52,11,0,77,52,11,0,69,52,11,0,92,52,11,0,84,52,11,0,106,52,11,0,99,52,11,0,55,52,11,0,46,52,11,0,62,52,11,0,169,146,11,0,223,146,11,0,230,146,11,0,239,146,11,0,246,146,11,0,255,146,11,0,6,147,11,0,15,147,11,0,22,147,11,0,30,147,11,0,36,147,11,0,59,147,11,0,66,147,11,0,121,52,11,0,113,52,11,0,44,147,11,0,50,147,11,0,4,148,11,0,119,241,11,0,12,148,11,0,119,241,11,0,26,148,11,0,119,241,11,0,231,147,11,0,119,241,11,0,157,52,11,0,119,241,11,0,164,52,11,0,119,241,11,0,171,52,11,0,119,241,11,0,178,52,11,0,119,241,11,0,185,52,11,0,119,241,11,0,192,52,11,0,119,241,11,0,199,52,11,0,119,241,11,0,206,52,11,0,119,241,11,0,213,52,11,0,119,241,11,0,220,52,11,0,119,241,11,0,227,52,11,0,119,241,11,0,234,52,11,0,119,241,11,0,241,52,11,0,119,241,11,0,248,52,11,0,119,241,11,0,255,52,11,0,119,241,11,0,6,53,11,0,119,241,11,0,13,53,11,0,119,241,11,0,20,53,11,0,119,241,11,0,27,53,11,0,119,241,11,0,34,53,11,0,119,241,11,0,41,53,11,0,119,241,11,0,48,53,11,0,119,241,11,0,55,53,11,0,119,241,11,0,62,53,11,0,119,241,11,0,68,53,11,0,119,241,11,0,75,53,11,0,119,241,11,0,82,53,11,0,119,241,11,0,178,14,12,0,119,241,11,0,131,180,11,0,119,241,11,0,237,147,11,0,119,241,11,0,75,241,11,0,119,241,11,0,93,241,11,0,119,241,11,0,191,241,11,0,189,241,11,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,175,145,11,0,182,145,11,0,189,145,11,0,196,145,11,0,204,145,11,0,211,145,11,0,218,145,11,0,225,145,11,0,233,145,11,0,240,145,11,0,247,145,11,0,254,145,11,0,6,146,11,0,13,146,11,0,20,146,11,0,27,146,11,0,35,146,11,0,42,146,11,0,49,146,11,0,56,146,11,0,64,146,11,0,71,146,11,0,78,146,11,0,85,146,11,0,92,146,11,0,85,146,11,0,99,146,11,0,106,146,11,0,113,146,11,0,106,146,11,0,120,146,11,0,127,146,11,0,134,146,11,0,127,146,11,0,141,146,11,0,148,146,11,0,155,146,11,0,148,146,11,0,162,146,11,0,169,146,11,0,178,146,11,0,185,146,11,0,193,146,11,0,200,146,11,0,208,146,11,0,215,146,11,0,223,146,11,0,230,146,11,0,239,146,11,0,246,146,11,0,255,146,11,0,6,147,11,0,15,147,11,0,22,147,11,0,30,147,11,0,36,147,11,0,44,147,11,0,50,147,11,0,59,147,11,0,66,147,11,0,75,147,11,0,82,147,11,0,91,147,11,0,98,147,11,0,107,147,11,0,114,147,11,0,123,147,11,0,130,147,11,0,139,147,11,0,146,147,11,0,230,146,11,0,189,241,11,0,246,146,11,0,189,241,11,0,6,147,11,0,189,241,11,0,155,147,11,0,189,241,11,0,161,147,11,0,189,241,11,0,170,147,11,0,189,241,11,0,179,147,11,0,189,241,11,0,66,147,11,0,189,241,11,0,188,147,11,0,189,241,11,0,197,147,11,0,189,241,11,0,206,147,11,0,189,241,11,0,215,147,11,0,189,241,11,0,243,187,11,0,189,241,11,0,110,241,11,0,119,241,11,0,136,229,11,0,119,241,11,0,121,241,11,0,119,241,11,0,224,147,11,0,119,241,11,0,60,241,11,0,119,241,11,0,134,241,11,0,119,241,11,0,141,241,11,0,119,241,11,0,231,147,11,0,119,241,11,0,163,241,11,0,119,241,11,0,178,14,12,0,119,241,11,0,131,180,11,0,119,241,11,0,237,147,11,0,119,241,11,0,75,241,11,0,119,241,11,0,93,241,11,0,119,241,11,0,246,147,11,0,119,241,11,0,253,147,11,0,119,241,11,0,175,184,11,0,119,241,11,0,4,148,11,0,119,241,11,0,12,148,11,0,119,241,11,0,19,148,11,0,119,241,11,0,26,148,11,0,119,241,11,0,33,148,11,0,119,241,11,0,40,148,11,0,119,241,11,0,47,148,11,0,119,241,11,0,54,148,11,0,119,241,11,0,61,148,11,0,119,241,11,0,67,148,11,0,119,241,11,0,74,148,11,0,119,241,11,0,80,148,11,0,119,241,11,0,87,148,11,0,119,241,11,0,94,148,11,0,119,241,11,0,101,148,11,0,119,241,11,0,108,148,11,0,119,241,11,0,115,148,11,0,119,241,11,0,122,148,11,0,119,241,11,0,87,148,11,0,119,241,11,0,101,148,11,0,119,241,11,0,129,148,11,0,119,241,11,0,136,148,11,0,119,241,11,0,143,148,11,0,119,241,11,0,150,148,11,0,119,241,11,0,157,148,11,0,119,241,11,0,164,148,11,0,119,241,11,0,171,148,11,0,119,241,11,0,178,148,11,0,119,241,11,0,185,148,11,0,119,241,11,0,192,148,11,0,119,241,11,0,199,148,11,0,119,241,11,0,206,148,11,0,119,241,11,0,213,148,11,0,119,241,11,0,191,241,11,0,189,241,11,0,251,165,11,0,131,180,11,0,178,14,12,0,141,241,11,0,134,241,11,0,216,241,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,71,180,11,0,0,0,0,0,227,180,11,0,1,0,0,0,234,180,11,0,2,0,0,0,110,241,11,0,3,0,0,0,193,241,11,0,4,0,0,0,0,0,0,0,255,255,255,255,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,220,240,11,0,47,241,11,0,54,241,11,0,8,242,11,0,8,242,11,0,8,242,11,0,193,241,11,0,119,241,11,0,136,229,11,0,119,241,11,0,121,241,11,0,119,241,11,0,127,241,11,0,119,241,11,0,60,241,11,0,119,241,11,0,216,241,11,0,119,241,11,0,134,241,11,0,119,241,11,0,75,241,11,0,119,241,11,0,93,241,11,0,119,241,11,0,178,14,12,0,119,241,11,0,248,241,11,0,119,241,11,0,0,242,11,0,119,241,11,0,191,241,11,0,189,241,11,0,84,226,11,0,193,241,11,0,118,238,11,0,110,241,11,0,127,226,11,0,136,229,11,0,163,226,11,0,121,241,11,0,200,241,11,0,127,241,11,0,208,241,11,0,216,241,11,0,223,241,11,0,232,241,11,0,67,241,11,0,60,241,11,0,240,241,11,0,134,241,11,0,84,241,11,0,75,241,11,0,101,241,11,0,93,241,11,0,110,241,11,0,119,241,11,0,136,229,11,0,119,241,11,0,121,241,11,0,119,241,11,0,127,241,11,0,119,241,11,0,60,241,11,0,119,241,11,0,134,241,11,0,119,241,11,0,141,241,11,0,119,241,11,0,149,241,11,0,119,241,11,0,156,241,11,0,119,241,11,0,163,241,11,0,119,241,11,0,41,226,11,0,119,241,11,0,20,227,11,0,119,241,11,0,126,227,11,0,119,241,11,0,169,241,11,0,119,241,11,0,248,227,11,0,119,241,11,0,154,229,11,0,119,241,11,0,176,241,11,0,119,241,11,0,183,241,11,0,119,241,11,0,75,241,11,0,119,241,11,0,93,241,11,0,119,241,11,0,178,14,12,0,189,241,11,0,191,241,11,0,189,241,11,0,0,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,31,0,0,0,63,0,0,0,127,0,0,0,255,0,0,0,255,1,0,0,255,3,0,0,255,7,0,0,255,15,0,0,255,31,0,0,255,63,0,0,255,127,0,0,255,255,0,0,255,255,1,0,255,255,3,0,255,255,7,0,255,255,15,0,255,255,31,0,255,255,63,0,255,255,127,0,255,255,255,0,255,255,255,1,255,255,255,3,255,255,255,7,255,255,255,15,255,255,255,31,255,255,255,63,255,255,255,127,255,255,255,255,0,0,0,0,1,0,0,0,2,0,0,0,2,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,5,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,6,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,7,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,62,0,0,0,0,0,0,0,1,0,0,0,8,0,0,0,2,0,0,0,9,0,0,0,26,0,0,0,27,0,0,0,3,0,0,0,28,0,0,0,10,0,0,0,29,0,0,0,11,0,0,0,30,0,0,0,63,0,0,0,12,0,0,0,6,0,0,0,3,0,0,0,3,0,0,0,4,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,5,0,0,0,3,0,0,0,5,0,0,0,4,0,0,0,5,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,4,0,0,0,0,0,0,0,1,0,0,0,3,0,0,0,7,0,0,0,15,0,0,0,31,0,0,0,63,0,0,0,127,0,0,0,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,1,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,1,0,0,0,13,0,0,0,14,0,0,0,1,0,0,0,15,0,0,0,15,0,0,0,16,0,0,0,16,0,0,0,16,0,0,0,16,0,0,0,16,0,0,0,16,0,0,0,17,0,0,0,17,0,0,0,1,0,0,0,1,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,1,0,0,0,1,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,30,0,0,0,30,0,0,0,31,0,0,0,30,0,0,0,32,0,0,0,33,0,0,0,30,0,0,0,34,0,0,0,35,0,0,0,30,0,0,0,36,0,0,0,37,0,0,0,30,0,0,0,30,0,0,0,38,0,0,0,30,0,0,0,30,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,39,0,0,0,40,0,0,0,1,0,0,0,41,0,0,0,42,0,0,0,23,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,28,0,0,0,47,0,0,0,30,0,0,0,30,0,0,0,48,0,0,0,30,0,0,0,49,0,0,0,50,0,0,0,30,0,0,0,51,0,0,0,52,0,0,0,30,0,0,0,53,0,0,0,54,0,0,0,30,0,0,0,30,0,0,0,38,0,0,0,30,0,0,0,30,0,0,0,1,0,0,0,55,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,2,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,3,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,1,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,4,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,192,3,0,0,192,4,0,0,192,5,0,0,192,6,0,0,192,7,0,0,192,8,0,0,192,9,0,0,192,10,0,0,192,11,0,0,192,12,0,0,192,13,0,0,192,14,0,0,192,15,0,0,192,16,0,0,192,17,0,0,192,18,0,0,192,19,0,0,192,20,0,0,192,21,0,0,192,22,0,0,192,23,0,0,192,24,0,0,192,25,0,0,192,26,0,0,192,27,0,0,192,28,0,0,192,29,0,0,192,30,0,0,192,31,0,0,192,0,0,0,179,1,0,0,195,2,0,0,195,3,0,0,195,4,0,0,195,5,0,0,195,6,0,0,195,7,0,0,195,8,0,0,195,9,0,0,195,10,0,0,195,11,0,0,195,12,0,0,195,13,0,0,211,14,0,0,195,15,0,0,195,0,0,12,187,1,0,12,195,2,0,12,195,3,0,12,195,4,0,12,211,216,232,1,0,72,233,1,0,184,233,1,0,184,233,1,0,0,0,0,0,10,0,0,0,100,0,0,0,232,3,0,0,16,39,0,0,160,134,1,0,64,66,15,0,128,150,152,0,0,225,245,5,199,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,201,35,12,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,3,0,0,0,179,33,12,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,9,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,0,0,0,0,0,0,3,0,0,0,171,29,12,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,3,0,0,0,163,25,12,0,0,4,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,96,7,0,0,0,8,80,0,0,8,16,0,20,8,115,0,18,7,31,0,0,8,112,0,0,8,48,0,0,9,192,0,16,7,10,0,0,8,96,0,0,8,32,0,0,9,160,0,0,8,0,0,0,8,128,0,0,8,64,0,0,9,224,0,16,7,6,0,0,8,88,0,0,8,24,0,0,9,144,0,19,7,59,0,0,8,120,0,0,8,56,0,0,9,208,0,17,7,17,0,0,8,104,0,0,8,40,0,0,9,176,0,0,8,8,0,0,8,136,0,0,8,72,0,0,9,240,0,16,7,4,0,0,8,84,0,0,8,20,0,21,8,227,0,19,7,43,0,0,8,116,0,0,8,52,0,0,9,200,0,17,7,13,0,0,8,100,0,0,8,36,0,0,9,168,0,0,8,4,0,0,8,132,0,0,8,68,0,0,9,232,0,16,7,8,0,0,8,92,0,0,8,28,0,0,9,152,0,20,7,83,0,0,8,124,0,0,8,60,0,0,9,216,0,18,7,23,0,0,8,108,0,0,8,44,0,0,9,184,0,0,8,12,0,0,8,140,0,0,8,76,0,0,9,248,0,16,7,3,0,0,8,82,0,0,8,18,0,21,8,163,0,19,7,35,0,0,8,114,0,0,8,50,0,0,9,196,0,17,7,11,0,0,8,98,0,0,8,34,0,0,9,164,0,0,8,2,0,0,8,130,0,0,8,66,0,0,9,228,0,16,7,7,0,0,8,90,0,0,8,26,0,0,9,148,0,20,7,67,0,0,8,122,0,0,8,58,0,0,9,212,0,18,7,19,0,0,8,106,0,0,8,42,0,0,9,180,0,0,8,10,0,0,8,138,0,0,8,74,0,0,9,244,0,16,7,5,0,0,8,86,0,0,8,22,0,64,8,0,0,19,7,51,0,0,8,118,0,0,8,54,0,0,9,204,0,17,7,15,0,0,8,102,0,0,8,38,0,0,9,172,0,0,8,6,0,0,8,134,0,0,8,70,0,0,9,236,0,16,7,9,0,0,8,94,0,0,8,30,0,0,9,156,0,20,7,99,0,0,8,126,0,0,8,62,0,0,9,220,0,18,7,27,0,0,8,110,0,0,8,46,0,0,9,188,0,0,8,14,0,0,8,142,0,0,8,78,0,0,9,252,0,96,7,0,0,0,8,81,0,0,8,17,0,21,8,131,0,18,7,31,0,0,8,113,0,0,8,49,0,0,9,194,0,16,7,10,0,0,8,97,0,0,8,33,0,0,9,162,0,0,8,1,0,0,8,129,0,0,8,65,0,0,9,226,0,16,7,6,0,0,8,89,0,0,8,25,0,0,9,146,0,19,7,59,0,0,8,121,0,0,8,57,0,0,9,210,0,17,7,17,0,0,8,105,0,0,8,41,0,0,9,178,0,0,8,9,0,0,8,137,0,0,8,73,0,0,9,242,0,16,7,4,0,0,8,85,0,0,8,21,0,16,8,2,1,19,7,43,0,0,8,117,0,0,8,53,0,0,9,202,0,17,7,13,0,0,8,101,0,0,8,37,0,0,9,170,0,0,8,5,0,0,8,133,0,0,8,69,0,0,9,234,0,16,7,8,0,0,8,93,0,0,8,29,0,0,9,154,0,20,7,83,0,0,8,125,0,0,8,61,0,0,9,218,0,18,7,23,0,0,8,109,0,0,8,45,0,0,9,186,0,0,8,13,0,0,8,141,0,0,8,77,0,0,9,250,0,16,7,3,0,0,8,83,0,0,8,19,0,21,8,195,0,19,7,35,0,0,8,115,0,0,8,51,0,0,9,198,0,17,7,11,0,0,8,99,0,0,8,35,0,0,9,166,0,0,8,3,0,0,8,131,0,0,8,67,0,0,9,230,0,16,7,7,0,0,8,91,0,0,8,27,0,0,9,150,0,20,7,67,0,0,8,123,0,0,8,59,0,0,9,214,0,18,7,19,0,0,8,107,0,0,8,43,0,0,9,182,0,0,8,11,0,0,8,139,0,0,8,75,0,0,9,246,0,16,7,5,0,0,8,87,0,0,8,23,0,64,8,0,0,19,7,51,0,0,8,119,0,0,8,55,0,0,9,206,0,17,7,15,0,0,8,103,0,0,8,39,0,0,9,174,0,0,8,7,0,0,8,135,0,0,8,71,0,0,9,238,0,16,7,9,0,0,8,95,0,0,8,31,0,0,9,158,0,20,7,99,0,0,8,127,0,0,8,63,0,0,9,222,0,18,7,27,0,0,8,111,0,0,8,47,0,0,9,190,0,0,8,15,0,0,8,143,0,0,8,79,0,0,9,254,0,96,7,0,0,0,8,80,0,0,8,16,0,20,8,115,0,18,7,31,0,0,8,112,0,0,8,48,0,0,9,193,0,16,7,10,0,0,8,96,0,0,8,32,0,0,9,161,0,0,8,0,0,0,8,128,0,0,8,64,0,0,9,225,0,16,7,6,0,0,8,88,0,0,8,24,0,0,9,145,0,19,7,59,0,0,8,120,0,0,8,56,0,0,9,209,0,17,7,17,0,0,8,104,0,0,8,40,0,0,9,177,0,0,8,8,0,0,8,136,0,0,8,72,0,0,9,241,0,16,7,4,0,0,8,84,0,0,8,20,0,21,8,227,0,19,7,43,0,0,8,116,0,0,8,52,0,0,9,201,0,17,7,13,0,0,8,100,0,0,8,36,0,0,9,169,0,0,8,4,0,0,8,132,0,0,8,68,0,0,9,233,0,16,7,8,0,0,8,92,0,0,8,28,0,0,9,153,0,20,7,83,0,0,8,124,0,0,8,60,0,0,9,217,0,18,7,23,0,0,8,108,0,0,8,44,0,0,9,185,0,0,8,12,0,0,8,140,0,0,8,76,0,0,9,249,0,16,7,3,0,0,8,82,0,0,8,18,0,21,8,163,0,19,7,35,0,0,8,114,0,0,8,50,0,0,9,197,0,17,7,11,0,0,8,98,0,0,8,34,0,0,9,165,0,0,8,2,0,0,8,130,0,0,8,66,0,0,9,229,0,16,7,7,0,0,8,90,0,0,8,26,0,0,9,149,0,20,7,67,0,0,8,122,0,0,8,58,0,0,9,213,0,18,7,19,0,0,8,106,0,0,8,42,0,0,9,181,0,0,8,10,0,0,8,138,0,0,8,74,0,0,9,245,0,16,7,5,0,0,8,86,0,0,8,22,0,64,8,0,0,19,7,51,0,0,8,118,0,0,8,54,0,0,9,205,0,17,7,15,0,0,8,102,0,0,8,38,0,0,9,173,0,0,8,6,0,0,8,134,0,0,8,70,0,0,9,237,0,16,7,9,0,0,8,94,0,0,8,30,0,0,9,157,0,20,7,99,0,0,8,126,0,0,8,62,0,0,9,221,0,18,7,27,0,0,8,110,0,0,8,46,0,0,9,189,0,0,8,14,0,0,8,142,0,0,8,78,0,0,9,253,0,96,7,0,0,0,8,81,0,0,8,17,0,21,8,131,0,18,7,31,0,0,8,113,0,0,8,49,0,0,9,195,0,16,7,10,0,0,8,97,0,0,8,33,0,0,9,163,0,0,8,1,0,0,8,129,0,0,8,65,0,0,9,227,0,16,7,6,0,0,8,89,0,0,8,25,0,0,9,147,0,19,7,59,0,0,8,121,0,0,8,57,0,0,9,211,0,17,7,17,0,0,8,105,0,0,8,41,0,0,9,179,0,0,8,9,0,0,8,137,0,0,8,73,0,0,9,243,0,16,7,4,0,0,8,85,0,0,8,21,0,16,8,2,1,19,7,43,0,0,8,117,0,0,8,53,0,0,9,203,0,17,7,13,0,0,8,101,0,0,8,37,0,0,9,171,0,0,8,5,0,0,8,133,0,0,8,69,0,0,9,235,0,16,7,8,0,0,8,93,0,0,8,29,0,0,9,155,0,20,7,83,0,0,8,125,0,0,8,61,0,0,9,219,0,18,7,23,0,0,8,109,0,0,8,45,0,0,9,187,0,0,8,13,0,0,8,141,0,0,8,77,0,0,9,251,0,16,7,3,0,0,8,83,0,0,8,19,0,21,8,195,0,19,7,35,0,0,8,115,0,0,8,51,0,0,9,199,0,17,7,11,0,0,8,99,0,0,8,35,0,0,9,167,0,0,8,3,0,0,8,131,0,0,8,67,0,0,9,231,0,16,7,7,0,0,8,91,0,0,8,27,0,0,9,151,0,20,7,67,0,0,8,123,0,0,8,59,0,0,9,215,0,18,7,19,0,0,8,107,0,0,8,43,0,0,9,183,0,0,8,11,0,0,8,139,0,0,8,75,0,0,9,247,0,16,7,5,0,0,8,87,0,0,8,23,0,64,8,0,0,19,7,51,0,0,8,119,0,0,8,55,0,0,9,207,0,17,7,15,0,0,8,103,0,0,8,39,0,0,9,175,0,0,8,7,0,0,8,135,0,0,8,71,0,0,9,239,0,16,7,9,0,0,8,95,0,0,8,31,0,0,9,159,0,20,7,99,0,0,8,127,0,0,8,63,0,0,9,223,0,18,7,27,0,0,8,111,0,0,8,47,0,0,9,191,0,0,8,15,0,0,8,143,0,0,8,79,0,0,9,255,0,16,5,1,0,23,5,1,1,19,5,17,0,27,5,1,16,17,5,5,0,25,5,1,4,21,5,65,0,29,5,1,64,16,5,3,0,24,5,1,2,20,5,33,0,28,5,1,32,18,5,9,0,26,5,1,8,22,5,129,0,64,5,0,0,16,5,2,0,23,5,129,1,19,5,25,0,27,5,1,24,17,5,7,0,25,5,1,6,21,5,97,0,29,5,1,96,16,5,4,0,24,5,1,3,20,5,49,0,28,5,1,48,18,5,13,0,26,5,1,12,22,5,193,0,64,5,0,0,16,0,17,0,18,0,0,0,8,0,7,0,9,0,6,0,10,0,5,0,11,0,4,0,12,0,3,0,13,0,2,0,14,0,1,0,15,0,1,0,2,0,3,0,4,0,5,0,7,0,9,0,13,0,17,0,25,0,33,0,49,0,65,0,97,0,129,0,193,0,1,1,129,1,1,2,1,3,1,4,1,6,1,8,1,12,1,16,1,24,1,32,1,48,1,64,1,96,0,0,0,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,13,0,15,0,17,0,19,0,23,0,27,0,31,0,35,0,43,0,51,0,59,0,67,0,83,0,99,0,115,0,131,0,163,0,195,0,227,0,2,1,0,0,0,0,16,0,16,0,16,0,16,0,17,0,17,0,18,0,18,0,19,0,19,0,20,0,20,0,21,0,21,0,22,0,22,0,23,0,23,0,24,0,24,0,25,0,25,0,26,0,26,0,27,0,27,0,28,0,28,0,29,0,29,0,64,0,64,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,16,0,17,0,17,0,17,0,17,0,18,0,18,0,18,0,18,0,19,0,19,0,19,0,19,0,20,0,20,0,20,0,20,0,21,0,21,0,21,0,21,0,16,0,73,0,195,0,12,0,8,0,140,0,8,0,76,0,8,0,204,0,8,0,44,0,8,0,172,0,8,0,108,0,8,0,236,0,8,0,28,0,8,0,156,0,8,0,92,0,8,0,220,0,8,0,60,0,8,0,188,0,8,0,124,0,8,0,252,0,8,0,2,0,8,0,130,0,8,0,66,0,8,0,194,0,8,0,34,0,8,0,162,0,8,0,98,0,8,0,226,0,8,0,18,0,8,0,146,0,8,0,82,0,8,0,210,0,8,0,50,0,8,0,178,0,8,0,114,0,8,0,242,0,8,0,10,0,8,0,138,0,8,0,74,0,8,0,202,0,8,0,42,0,8,0,170,0,8,0,106,0,8,0,234,0,8,0,26,0,8,0,154,0,8,0,90,0,8,0,218,0,8,0,58,0,8,0,186,0,8,0,122,0,8,0,250,0,8,0,6,0,8,0,134,0,8,0,70,0,8,0,198,0,8,0,38,0,8,0,166,0,8,0,102,0,8,0,230,0,8,0,22,0,8,0,150,0,8,0,86,0,8,0,214,0,8,0,54,0,8,0,182,0,8,0,118,0,8,0,246,0,8,0,14,0,8,0,142,0,8,0,78,0,8,0,206,0,8,0,46,0,8,0,174,0,8,0,110,0,8,0,238,0,8,0,30,0,8,0,158,0,8,0,94,0,8,0,222,0,8,0,62,0,8,0,190,0,8,0,126,0,8,0,254,0,8,0,1,0,8,0,129,0,8,0,65,0,8,0,193,0,8,0,33,0,8,0,161,0,8,0,97,0,8,0,225,0,8,0,17,0,8,0,145,0,8,0,81,0,8,0,209,0,8,0,49,0,8,0,177,0,8,0,113,0,8,0,241,0,8,0,9,0,8,0,137,0,8,0,73,0,8,0,201,0,8,0,41,0,8,0,169,0,8,0,105,0,8,0,233,0,8,0,25,0,8,0,153,0,8,0,89,0,8,0,217,0,8,0,57,0,8,0,185,0,8,0,121,0,8,0,249,0,8,0,5,0,8,0,133,0,8,0,69,0,8,0,197,0,8,0,37,0,8,0,165,0,8,0,101,0,8,0,229,0,8,0,21,0,8,0,149,0,8,0,85,0,8,0,213,0,8,0,53,0,8,0,181,0,8,0,117,0,8,0,245,0,8,0,13,0,8,0,141,0,8,0,77,0,8,0,205,0,8,0,45,0,8,0,173,0,8,0,109,0,8,0,237,0,8,0,29,0,8,0,157,0,8,0,93,0,8,0,221,0,8,0,61,0,8,0,189,0,8,0,125,0,8,0,253,0,8,0,19,0,9,0,19,1,9,0,147,0,9,0,147,1,9,0,83,0,9,0,83,1,9,0,211,0,9,0,211,1,9,0,51,0,9,0,51,1,9,0,179,0,9,0,179,1,9,0,115,0,9,0,115,1,9,0,243,0,9,0,243,1,9,0,11,0,9,0,11,1,9,0,139,0,9,0,139,1,9,0,75,0,9,0,75,1,9,0,203,0,9,0,203,1,9,0,43,0,9,0,43,1,9,0,171,0,9,0,171,1,9,0,107,0,9,0,107,1,9,0,235,0,9,0,235,1,9,0,27,0,9,0,27,1,9,0,155,0,9,0,155,1,9,0,91,0,9,0,91,1,9,0,219,0,9,0,219,1,9,0,59,0,9,0,59,1,9,0,187,0,9,0,187,1,9,0,123,0,9,0,123,1,9,0,251,0,9,0,251,1,9,0,7,0,9,0,7,1,9,0,135,0,9,0,135,1,9,0,71,0,9,0,71,1,9,0,199,0,9,0,199,1,9,0,39,0,9,0,39,1,9,0,167,0,9,0,167,1,9,0,103,0,9,0,103,1,9,0,231,0,9,0,231,1,9,0,23,0,9,0,23,1,9,0,151,0,9,0,151,1,9,0,87,0,9,0,87,1,9,0,215,0,9,0,215,1,9,0,55,0,9,0,55,1,9,0,183,0,9,0,183,1,9,0,119,0,9,0,119,1,9,0,247,0,9,0,247,1,9,0,15],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+119092);allocate([9,0,15,1,9,0,143,0,9,0,143,1,9,0,79,0,9,0,79,1,9,0,207,0,9,0,207,1,9,0,47,0,9,0,47,1,9,0,175,0,9,0,175,1,9,0,111,0,9,0,111,1,9,0,239,0,9,0,239,1,9,0,31,0,9,0,31,1,9,0,159,0,9,0,159,1,9,0,95,0,9,0,95,1,9,0,223,0,9,0,223,1,9,0,63,0,9,0,63,1,9,0,191,0,9,0,191,1,9,0,127,0,9,0,127,1,9,0,255,0,9,0,255,1,9,0,0,0,7,0,64,0,7,0,32,0,7,0,96,0,7,0,16,0,7,0,80,0,7,0,48,0,7,0,112,0,7,0,8,0,7,0,72,0,7,0,40,0,7,0,104,0,7,0,24,0,7,0,88,0,7,0,56,0,7,0,120,0,7,0,4,0,7,0,68,0,7,0,36,0,7,0,100,0,7,0,20,0,7,0,84,0,7,0,52,0,7,0,116,0,7,0,3,0,8,0,131,0,8,0,67,0,8,0,195,0,8,0,35,0,8,0,163,0,8,0,99,0,8,0,227,0,8,0,0,0,5,0,16,0,5,0,8,0,5,0,24,0,5,0,4,0,5,0,20,0,5,0,12,0,5,0,28,0,5,0,2,0,5,0,18,0,5,0,10,0,5,0,26,0,5,0,6,0,5,0,22,0,5,0,14,0,5,0,30,0,5,0,1,0,5,0,17,0,5,0,9,0,5,0,25,0,5,0,5,0,5,0,21,0,5,0,13,0,5,0,29,0,5,0,3,0,5,0,19,0,5,0,11,0,5,0,27,0,5,0,7,0,5,0,23,0,5],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+129332);allocate([221,153,0,0,0,0,0,0,27,0,25,0,1,0,24,0,15,0,25,0,25,0,25,0,25,0,25,0,25,0,25,0,7,0,5,0,21,0,25,0,20,0,10,0,10,0,10,0,10,0,6,0,10,0,10,0,10,0,10,0,10,0,14,0,10,0,10,0,10,0,10,0,10,0,10,0,10,0,25,0,1,0,19,0,0,0,9,0,0,0,8,0,0,0,10,0,17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,0,0,0,7,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,5,0,0,0,23,0,18,0,22,0,10,0,10,0,10,0,2,0,10,0,10,0,10,0,4,0,10,0,10,0,10,0,10,0,3,0,10,0,10,0,10,0,10,0,10,0,10,0,10,0,10,0,10,0,10,0,16,0,0,0,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,0,11,0,10,0,20,0,21,0,10,0,10,0,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,15,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,0,0,0,0,0,0,0,0,111,1,112,1,108,1,112,1,90,1,103,1,100,1,99,1,97,1,95,1,32,0,91,1,66,0,103,0,83,1,44,0,82,1,25,0,52,0,60,1,26,0,59,1,34,0,133,0,48,0,61,0,125,0,112,1,0,0,29,0,45,0,60,0,81,0,82,0,93,0,43,1,95,1,112,1,91,1,112,1,88,1,87,1,86,1,112,1,112,1,83,1,58,1,59,1,57,1,38,1,39,1,37,1,112,1,121,0,164,0,51,1,45,1,70,0,117,0,43,0,40,1,20,1,15,1,58,0,86,0,79,0,13,1,152,0,168,0,181,0,112,1,112,1,112,1,151,0,162,0,0,0,180,0,189,0,190,0,191,0,53,1,196,0,199,0,205,0,204,0,211,0,214,0,207,0,223,0,224,0,232,0,238,0,243,0,245,0,222,0,246,0,112,1,55,1,54,1,23,1,26,1,22,1,3,1,6,1,2,1,252,0,30,1,39,1,38,1,37,1,36,1,35,1,34,1,11,1,32,1,2,1,29,1,28,1,22,1,14,1,12,1,3,1,218,0,252,0,8,1,16,1,112,1,251,0,112,1,112,1,4,1,24,1,27,1,236,0,222,0,230,0,193,0,184,0,212,0,208,0,202,0,173,0,156,0,112,1,133,0,126,0,112,1,104,0,98,0,119,0,132,0,80,0,94,0,92,0,112,1,78,0,112,1,67,1,69,1,73,1,77,1,68,0,67,0,81,1,0,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,13,0,20,0,13,0,25,0,23,0,13,0,18,0,18,0,18,0,20,0,23,0,21,0,62,0,32,0,25,0,165,0,164,0,27,0,25,0,13,0,62,0,13,0,32,0,13,0,15,0,15,0,15,0,15,0,27,0,33,0,28,0,15,0,27,0,158,0,21,0,15,0,21,0,15,0,28,0,60,0,33,0,15,0,15,0,15,0,34,0,28,0,66,0,156,0,34,0,155,0,60,0,15,0,37,0,37,0,15,0,66,0,15,0,34,0,15,0,15,0,15,0,16,0,16,0,16,0,16,0,35,0,36,0,68,0,154,0,35,0,36,0,16,0,16,0,68,0,67,0,37,0,36,0,37,0,35,0,56,0,56,0,56,0,56,0,67,0,29,0,29,0,61,0,153,0,152,0,37,0,16,0,16,0,26,0,26,0,26,0,151,0,150,0,61,0,26,0,26,0,26,0,26,0,26,0,26,0,29,0,76,0,148,0,29,0,147,0,29,0,70,0,70,0,70,0,70,0,145,0,26,0,77,0,26,0,26,0,26,0,26,0,26,0,26,0,57,0,57,0,57,0,71,0,71,0,71,0,71,0,77,0,144,0,57,0,57,0,79,0,76,0,71,0,71,0,72,0,79,0,72,0,72,0,72,0,80,0,81,0,82,0,77,0,80,0,81,0,82,0,84,0,57,0,57,0,85,0,84,0,71,0,71,0,85,0,87,0,86,0,143,0,90,0,79,0,86,0,79,0,88,0,142,0,141,0,89,0,140,0,88,0,88,0,89,0,80,0,81,0,82,0,97,0,91,0,92,0,139,0,84,0,91,0,92,0,85,0,87,0,138,0,93,0,137,0,87,0,86,0,93,0,90,0,94,0,88,0,90,0,88,0,94,0,95,0,89,0,96,0,98,0,95,0,136,0,96,0,98,0,130,0,97,0,91,0,92,0,130,0,126,0,108,0,108,0,108,0,133,0,125,0,93,0,124,0,133,0,97,0,108,0,108,0,94,0,127,0,127,0,127,0,123,0,95,0,122,0,96,0,98,0,128,0,128,0,128,0,134,0,130,0,121,0,135,0,134,0,108,0,108,0,135,0,120,0,119,0,133,0,118,0,117,0,116,0,115,0,114,0,113,0,112,0,111,0,110,0,109,0,107,0,106,0,105,0,104,0,103,0,102,0,101,0,100,0,83,0,134,0,69,0,65,0,135,0,160,0,64,0,160,0,160,0,161,0,161,0,162,0,63,0,162,0,162,0,163,0,59,0,163,0,163,0,166,0,58,0,166,0,166,0,54,0,53,0,52,0,51,0,50,0,49,0,48,0,45,0,44,0,43,0,41,0,39,0,38,0,24,0,22,0,19,0,17,0,14,0,12,0,11,0,10,0,9,0,8,0,7,0,5,0,3,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,0,0,159,0,1,0,159,0,159,0,159,0,159,0,159,0,160,0,161,0,162,0,159,0,163,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,159,0,165,0,164,0,164,0,164,0,164,0,164,0,164,0,159,0,159,0,159,0,160,0,159,0,166,0,161,0,162,0,159,0,159,0,163,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,164,0,164,0,165,0,164,0,164,0,164,0,164,0,26,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,164,0,159,0,166,0,166,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,164,0,159,0,159,0,164,0,164,0,164,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,0,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,0,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,4,0,14,0,4,0,15,0,16,0,16,0,16,0,17,0,18,0,19,0,20,0,21,0,22,0,22,0,23,0,24,0,25,0,26,0,22,0,22,0,27,0,28,0,29,0,22,0,22,0,24,0,22,0,22,0,30,0,31,0,32,0,21,0,22,0,33,0,24,0,34,0,22,0,35,0,36,0,37,0,22,0,22,0,24,0,22,0,38,0,49,0,77,0,50,0,81,0,80,0,51,0,73,0,74,0,75,0,78,0,78,0,79,0,115,0,78,0,82,0,78,0,76,0,84,0,78,0,52,0,116,0,53,0,90,0,54,0,56,0,57,0,57,0,57,0,85,0,78,0,86,0,58,0,78,0,157,0,79,0,59,0,78,0,60,0,87,0,111,0,91,0,61,0,62,0,63,0,78,0,78,0,120,0,157,0,92,0,157,0,112,0,64,0,88,0,88,0,65,0,121,0,66,0,93,0,67,0,68,0,69,0,70,0,71,0,71,0,71,0,78,0,78,0,124,0,158,0,94,0,96,0,72,0,72,0,125,0,122,0,88,0,97,0,78,0,95,0,56,0,108,0,108,0,108,0,123,0,88,0,88,0,113,0,157,0,156,0,98,0,72,0,72,0,83,0,83,0,83,0,155,0,154,0,114,0,83,0,83,0,83,0,83,0,83,0,83,0,89,0,129,0,153,0,88,0,152,0,78,0,56,0,57,0,57,0,57,0,146,0,83,0,129,0,78,0,83,0,83,0,83,0,83,0,83,0,57,0,57,0,57,0,70,0,71,0,71,0,71,0,130,0,47,0,72,0,72,0,129,0,78,0,72,0,72,0,127,0,79,0,128,0,128,0,128,0,129,0,129,0,129,0,78,0,74,0,75,0,131,0,129,0,72,0,72,0,129,0,73,0,72,0,72,0,132,0,129,0,129,0,146,0,129,0,79,0,40,0,78,0,129,0,47,0,149,0,129,0,151,0,88,0,88,0,99,0,78,0,78,0,78,0,129,0,129,0,129,0,150,0,78,0,74,0,75,0,78,0,133,0,149,0,129,0,148,0,78,0,78,0,131,0,78,0,129,0,88,0,134,0,78,0,73,0,129,0,78,0,129,0,129,0,132,0,147,0,40,0,99,0,129,0,78,0,78,0,78,0,47,0,99,0,108,0,108,0,108,0,129,0,145,0,78,0,40,0,146,0,135,0,72,0,72,0,78,0,128,0,128,0,128,0,132,0,78,0,73,0,78,0,78,0,128,0,128,0,128,0,129,0,78,0,131,0,129,0,47,0,72,0,72,0,146,0,75,0,74,0,78,0,144,0,99,0,143,0,40,0,132,0,73,0,131,0,75,0,74,0,142,0,141,0,140,0,139,0,138,0,137,0,136,0,101,0,101,0,129,0,78,0,126,0,119,0,78,0,41,0,118,0,41,0,41,0,44,0,44,0,45,0,117,0,45,0,45,0,48,0,110,0,48,0,48,0,100,0,109,0,100,0,100,0,107,0,106,0,105,0,104,0,103,0,102,0,42,0,46,0,159,0,101,0,42,0,39,0,99,0,78,0,78,0,75,0,73,0,55,0,42,0,47,0,46,0,43,0,42,0,40,0,39,0,159,0,3,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,159,0,210,255,41,1,210,255,211,255,210,255,210,255,210,255,210,255,210,255,91,1,142,1,142,1,251,255,0,0,252,255,5,0,19,0,23,0,210,255,210,255,210,255,142,1,142,1,142,1,142,1,142,1,142,1,210,255,142,1,210,255,6,0,17,0,64,4,40,1,188,5,210,5,210,255,210,255,168,1,9,0,94,5,135,0,196,1,168,0,238,5,88,1,75,5,169,5,19,6,246,255,210,255,247,255,142,1,142,1,142,1,142,1,75,5,169,5,250,0,254,255,254,255,10,0,11,0,254,255,10,0,254,255,10,0,107,2,240,0,119,5,144,5,142,1,210,255,142,1,210,255,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,142,1,210,255,142,1,142,1,142,1,142,1,142,1,142,1,142,1,210,255,253,255,253,255,253,255,253,255,253,255,253,255,253,255,253,255,253,255,142,1,210,255,142,1,142,1,142,1,142,1,142,1,142,1,142,1,210,255,142,1,210,255,142,1,210,255,210,255,142,1,210,255,142,1,210,255,210,255,210,255,142,1,142,1,210,255,142,1,210,255,207,4,227,4,247,4,11,5,210,255,210,255,210,255,210,255,75,5,169,5,75,5,169,5,33,5,255,5,255,5,255,5,20,6,20,6,20,6,20,6,55,0,55,0,55,0,216,255,10,0,216,255,216,255,216,2,55,5,144,1,201,0,74,0,111,0,221,255,221,255,216,255,240,2,253,255,253,255,24,0,24,0,24,0,24,0,24,0,24,0,79,0,11,0,11,0,8,3,239,255,239,255,28,0,28,0,28,0,28,0,210,255,224,1,86,1,87,4,151,5,107,4,158,5,252,1,127,4,210,255,210,255,210,255,210,255,142,1,142,1,210,255,142,1,142,1,142,1,142,1,210,255,11,0,20,0,142,1,210,255,142,1,210,255,210,255,142,1,210,255,142,1,210,255,60,0,210,255,142,1,219,5,32,3,219,5,169,5,219,5,169,5,250,0,56,3,80,3,147,4,134,2,24,2,48,0,52,2,142,1,210,255,142,1,210,255,142,1,210,255,142,1,210,255,142,1,210,255,63,0,83,0,210,255,104,3,128,3,152,3,161,2,167,4,39,0,45,0,142,1,210,255,142,1,210,255,142,1,210,255,210,255,142,1,210,255,210,255,176,3,200,3,224,3,80,2,142,1,210,255,142,1,210,255,142,1,210,255,142,1,210,255,248,3,16,4,40,4,187,4,210,255,210,255,210,255,142,1,188,2,210,255,1,0,6,0,42,0,38,0,7,0,50,0,6,0,47,0,9,0,10,0,20,0,20,0,47,0,16,0,31,0,32,0,33,0,34,0,35,0,23,0,21,0,22,0,23,0,1,0,25,0,26,0,20,0,28,0,23,0,20,0,24,0,9,0,10,0,11,0,25,0,26,0,27,0,20,0,29,0,30,0,43,0,24,0,23,0,53,0,53,0,47,0,23,0,38,0,53,0,52,0,28,0,52,0,53,0,54,0,55,0,35,0,47,0,47,0,47,0,35,0,40,0,41,0,53,0,35,0,40,0,41,0,6,0,47,0,20,0,6,0,71,0,47,0,73,0,53,0,75,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,88,0,89,0,6,0,91,0,1,0,53,0,38,0,39,0,96,0,97,0,42,0,53,0,9,0,10,0,27,0,47,0,29,0,30,0,255,255,255,255,108,0,255,255,255,255,255,255,21,0,38,0,91,0,24,0,25,0,26,0,118,0,28,0,40,0,41,0,47,0,123,0,255,255,125,0,255,255,47,0,255,255,255,255,130,0,255,255,132,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,255,255,29,0,30,0,255,255,255,255,255,255,123,0,255,255,255,255,1,0,38,0,255,255,129,0,255,255,255,255,255,255,20,0,9,0,255,255,47,0,255,255,255,255,71,0,255,255,73,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,24,0,255,255,255,255,255,255,28,0,86,0,255,255,255,255,255,255,255,255,91,0,92,0,93,0,94,0,95,0,255,255,255,255,53,0,255,255,255,255,255,255,255,255,25,0,26,0,27,0,255,255,29,0,30,0,255,255,255,255,202,0,203,0,255,255,205,0,206,0,38,0,208,0,255,255,255,255,120,0,212,0,255,255,214,0,255,255,47,0,217,0,255,255,219,0,255,255,255,255,53,0,223,0,22,0,255,255,255,255,255,255,255,255,255,255,207,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,238,0,255,255,240,0,255,255,242,0,255,255,244,0,255,255,246,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,106,0,107,0,255,255,255,255,2,1,255,255,4,1,255,255,6,1,255,255,255,255,9,1,25,0,26,0,27,0,255,255,29,0,30,0,16,1,255,255,18,1,255,255,20,1,255,255,22,1,38,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,31,1,47,0,255,255,255,255,255,255,255,255,255,255,53,0,255,255,205,0,206,0,0,0,1,0,255,255,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,255,255,171,0,172,0,23,0,25,0,26,0,27,0,255,255,29,0,30,0,255,255,255,255,255,255,255,255,255,255,35,0,36,0,38,0,255,255,255,255,255,255,255,255,255,255,43,0,44,0,45,0,47,0,255,255,255,255,50,0,50,0,255,255,52,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,25,0,26,0,27,0,23,0,29,0,30,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,38,0,255,255,35,0,36,0,255,255,255,255,255,255,255,255,255,255,47,0,43,0,44,0,45,0,255,255,255,255,53,0,255,255,53,0,255,255,52,0,53,0,3,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,255,255,255,255,255,255,23,0,22,0,255,255,255,255,25,0,26,0,27,0,255,255,29,0,30,0,255,255,255,255,35,0,36,0,255,255,255,255,255,255,38,0,255,255,255,255,43,0,44,0,45,0,20,0,21,0,255,255,47,0,255,255,255,255,52,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,21,0,255,255,255,255,255,255,255,255,53,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,21,0,255,255,255,255,255,255,255,255,53,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,21,0,255,255,255,255,255,255,255,255,53,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,21,0,255,255,255,255,255,255,255,255,53,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,20,0,21,0,255,255,255,255,255,255,53,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,20,0,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,21,0,255,255,47,0,255,255,255,255,255,255,51,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,20,0,21,0,255,255,255,255,47,0,255,255,255,255,50,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,21,0,42,0,255,255,24,0,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,21,0,42,0,255,255,24,0,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,21,0,42,0,255,255,24,0,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,21,0,42,0,255,255,24,0,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,21,0,22,0,255,255,255,255,47,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,21,0,22,0,255,255,255,255,47,0,255,255,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,21,0,42,0,255,255,255,255,255,255,255,255,47,0,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,255,255,255,255,255,255,255,255,40,0,41,0,255,255,255,255,255,255,255,255,255,255,47,0,255,255,255,255,255,255,255,255,255,255,53,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,255,255,255,255,255,255,255,255,40,0,41,0,255,255,255,255,255,255,255,255,255,255,47,0,255,255,255,255,255,255,255,255,255,255,53,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,25,0,26,0,27,0,53,0,29,0,30,0,255,255,255,255,255,255,255,255,53,0,255,255,255,255,38,0,255,255,255,255,255,255,53,0,255,255,255,255,255,255,255,255,47,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,255,255,255,255,255,255,255,255,40,0,41,0,255,255,255,255,255,255,255,255,255,255,47,0,255,255,255,255,50,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,255,255,28,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,255,255,50,0,42,0,255,255,255,255,255,255,255,255,47,0,29,0,30,0,31,0,32,0,33,0,34,0,35,0,255,255,255,255,255,255,255,255,40,0,41,0,255,255,255,255,255,255,255,255,255,255,47,0,31,0,32,0,33,0,34,0,35,0,36,0,37,0,38,0,39,0,20,0,255,255,42,0,255,255,255,255,255,255,255,255,47,0,255,255,29,0,30,0,31,0,32,0,33,0,34,0,35,0,35,0,36,0,37,0,38,0,39,0,255,255,255,255,42,0,255,255,255,255,255,255,255,255,47,0,32,0,49,0,88,0,96,0,8,0,36,0,51,0,89,0,38,0,42,0,130,0,132,0,97,0,17,0,112,0,113,0,114,0,115,0,116,0,52,0,56,0,59,0,60,0,35,0,63,0,65,0,71,0,67,0,53,0,120,0,72,0,41,0,45,0,48,0,91,0,92,0,93,0,73,0,94,0,95,0,171,0,74,0,54,0,131,0,133,0,89,0,55,0,96,0,50,0,172,0,70,0,134,0,135,0,136,0,137,0,105,0,97,0,97,0,108,0,105,0,106,0,107,0,121,0,116,0,106,0,107,0,236,0,108,0,248,0,0,1,142,0,108,0,144,0,140,0,146,0,147,0,148,0,149,0,150,0,151,0,152,0,153,0,154,0,155,0,156,0,157,0,159,0,160,0,161,0,1,1,162,0,33,0,10,1,86,0,87,0,169,0,170,0,88,0,11,1,39,0,43,0,93,0,89,0,94,0,95,0,0,0,0,0,182,0,0,0,0,0,0,0,57,0,96,0,164,0,61,0,64,0,66,0,190,0,68,0,106,0,107,0,97,0,192,0,0,0,194,0,0,0,108,0,0,0,0,0,196,0,0,0,197,0,183,0,184,0,185,0,186,0,187,0,188,0,189,0,0,0,94,0,95,0,0,0,0,0,0,0,193,0,0,0,0,0,34,0,96,0,0,0,195,0,0,0,0,0,0,0,123,0,40,0,0,0,97,0,0,0,0,0,143,0,0,0,145,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,62,0,0,0,0,0,0,0,69,0,158,0,0,0,0,0,0,0,0,0,163,0,165,0,166,0,167,0,168,0,0,0,0,0,124,0,0,0,0,0,0,0,0,0,91,0,92,0,93,0,0,0,94,0,95,0,0,0,0,0,224,0,225,0,0,0,226,0,228,0,96,0,231,0,0,0,0,0,191,0,232,0,0,0,233,0,0,0,97,0,234,0,0,0,235,0,0,0,0,0,127,0,237,0,207,0,0,0,0,0,0,0,0,0,0,0,230,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,251,0,0,0,252,0,0,0,253,0,0,0,254,0,0,0,255,0,173,0,174,0,175,0,176,0,177,0,178,0,179,0,180,0,181,0,0,0,0,0,12,1,0,0,13,1,0,0,14,1,0,0,0,0,15,1,91,0,92,0,93,0,0,0,94,0,95,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,96,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,32,1,97,0,0,0,0,0,0,0,0,0,0,0,139,0,0,0,227,0,229,0,2,0,3,0,0,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,0,0,210,0,211,0,21,0,91,0,92,0,93,0,0,0,94,0,95,0,0,0,0,0,0,0,0,0,0,0,22,0,23,0,96,0,0,0,0,0,0,0,0,0,0,0,24,0,25,0,26,0,97,0,0,0,0,0,98,0,27,0,0,0,28,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,91,0,92,0,93,0,21,0,94,0,95,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,96,0,0,0,22,0,23,0,0,0,0,0,0,0,0,0,0,0,97,0,24,0,25,0,26,0,0,0,0,0,216,0,0,0,128,0,0,0,28,0,37,0,4,0,5,0,6,0,7,0,8,0,9,0,10,0,11,0,12,0,13,0,14,0,15,0,16,0,17,0,18,0,19,0,20,0,0,0,0,0,0,0,21,0,206,0,0,0,0,0,91,0,92,0,93,0,0,0,94,0,95,0,0,0,0,0,22,0,23,0,0,0,0,0,0,0,96,0,0,0,0,0,24,0,25,0,26,0,118,0,75,0,0,0,97,0,0,0,0,0,28,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,125,0,75,0,0,0,0,0,0,0,119,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,214,0,75,0,0,0,0,0,0,0,126,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,221,0,75,0,0,0,0,0,0,0,215,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,246,0,75,0,0,0,0,0,0,0,222,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,249,0,75,0,0,0,0,0,0,0,247,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,22,1,75,0,0,0,0,0,0,0,250,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,75,0,0,0,0,0,0,0,0,0,23,1,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,75,0,0,0,0,0,0,0,0,0,138,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,75,0,0,0,0,0,0,0,0,0,245,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,75,0,0,0,0,0,0,0,0,0,8,1,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,203,0,75,0,0,0,0,0,0,0,33,1,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,208,0,75,0,0,0,89,0,0,0,0,0,0,0,204,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,212,0,75,0,0,0,89,0,0,0,0,0,0,0,209,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,238,0,75,0,0,0,89,0,0,0,0,0,0,0,213,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,240,0,75,0,0,0,89,0,0,0,0,0,0,0,239,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,242,0,75,0,0,0,89,0,0,0,0,0,0,0,241,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,2,1,75,0,0,0,89,0,0,0,0,0,0,0,243,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,4,1,75,0,0,0,89,0,0,0,0,0,0,0,3,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,6,1,75,0,0,0,89,0,0,0,0,0,0,0,5,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,16,1,75,0,0,0,89,0,0,0,0,0,0,0,7,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,18,1,75,0,0,0,89,0,0,0,0,0,0,0,17,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,20,1,75,0,0,0,89,0,0,0,0,0,0,0,19,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,75,0,0,0,89,0,0,0,0,0,0,0,21,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,75,0,0,0,89,0,0,0,0,0,0,0,28,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,75,0,0,0,89,0,0,0,0,0,0,0,29,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,75,0,0,0,89,0,0,0,0,0,0,0,30,1,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,217,0,75,0,0,0,0,0,89,0,0,0,0,0,90,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,219,0,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,223,0,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,244,0,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,9,1,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,31,1,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,75,0,88,0,0,0,198,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,75,0,88,0,0,0,199,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,75,0,88,0,0,0,200,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,75,0,88,0,0,0,201,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,75,0,202,0,0,0,0,0,89,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,75,0,205,0,0,0,0,0,89,0,0,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,75,0,88,0,0,0,0,0,0,0,0,0,89,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,0,0,0,0,0,0,0,0,106,0,107,0,0,0,0,0,0,0,0,0,0,0,108,0,0,0,0,0,0,0,0,0,0,0,122,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,0,0,0,0,0,0,0,0,106,0,107,0,0,0,0,0,0,0,0,0,0,0,108,0,0,0,0,0,0,0,0,0,0,0,140,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,91,0,92,0,93,0,141,0,94,0,95,0,0,0,0,0,0,0,0,0,218,0,0,0,0,0,96,0,0,0,0,0,0,0,220,0,0,0,0,0,0,0,0,0,97,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,0,0,0,0,0,0,0,0,106,0,107,0,0,0,0,0,0,0,0,0,0,0,108,0,0,0,0,0,109,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,0,0,76,0,77,0,78,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,0,0,117,0,88,0,0,0,0,0,0,0,0,0,89,0,99,0,100,0,101,0,102,0,103,0,104,0,105,0,0,0,0,0,0,0,0,0,106,0,107,0,0,0,0,0,0,0,0,0,0,0,108,0,79,0,80,0,81,0,82,0,83,0,84,0,85,0,86,0,87,0,129,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,0,0,110,0,111,0,112,0,113,0,114,0,115,0,116,0,83,0,84,0,85,0,86,0,87,0,0,0,0,0,88,0,0,0,0,0,0,0,0,0,89,0,210,255,210,255,210,255,210,255,210,255,255,255,90,0,147,0,22,0,123,34,99,114,118,97,108,49,34,58,32,37,46,49,52,103,44,32,34,99,114,118,97,108,50,34,58,32,37,46,49,52,103,44,32,34,99,114,112,105,120,49,34,58,32,37,46,49,52,103,44,32,34,99,114,112,105,120,50,34,58,32,37,46,49,52,103,44,32,34,99,100,101,108,116,49,34,58,32,37,46,49,52,103,44,32,34,99,100,101,108,116,50,34,58,32,37,46,49,52,103,44,32,34,99,114,111,116,34,58,32,37,46,49,52,103,44,32,34,99,116,121,112,101,49,34,58,32,34,37,115,34,44,32,34,99,116,121,112,101,50,34,58,32,34,37,115,34,44,32,32,34,112,116,121,112,101,34,58,32,34,37,115,34,44,32,34,114,97,100,101,99,115,121,115,34,58,32,34,37,115,34,125,0,37,46,51,102,32,37,46,51,102,0,103,97,108,97,99,116,105,99,0,101,99,108,105,112,116,105,99,0,108,105,110,101,97,114,0,100,101,103,114,101,101,115,0,115,101,120,97,103,101,115,105,109,97,108,0,37,115,40,0,37,46,54,102,44,32,37,46,54,102,0,37,115,44,32,37,115,0,116,101,120,116,0,44,37,115,0,44,32,37,115,44,32,37,115,0,44,32,37,46,54,102,44,32,37,46,54,102,0,44,32,37,46,54,102,34,0,44,32,37,46,54,102,39,0,44,32,37,46,54,102,100,0,44,32,37,46,54,102,0,32,123,34,115,105,122,101,34,58,37,46,50,102,44,34,117,110,105,116,115,34,58,34,97,114,99,115,101,99,34,125,0,32,123,34,115,105,122,101,34,58,37,46,54,102,44,34,117,110,105,116,115,34,58,34,97,114,99,109,105,110,34,125,0,32,123,34,115,105,122,101,34,58,37,46,54,102,44,34,117,110,105,116,115,34,58,34,100,101,103,114,101,101,115,34,125,0,37,102,32,37,102],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+326314);allocate([109,84,65,78,72,100,114,0,32,9,0,45,115,0,37,115,115,116,97,116,117,115,95,37,100,46,116,120,116,0,37,115,37,115,0,69,114,114,111,114,58,32,116,97,110,104,100,114,32,102,97,105,108,101,100,59,32,110,111,32,115,116,97,116,117,115,32,102,105,108,101,32,99,114,101,97,116,101,100,0,109,80,114,111,106,101,99,116,80,80,0,69,114,114,111,114,58,32,114,101,112,114,111,106,101,99,116,32,102,97,105,108,101,100,59,32,110,111,32,115,116,97,116,117,115,32,102,105,108,101,32,99,114,101,97,116,101,100,0,37,115,37,115,46,104,100,117,108,105,115,116,0,69,114,114,111,114,58,32,108,105,115,116,104,100,117,32,102,97,105,108,101,100,59,32,110,111,32,108,105,115,116,32,102,105,108,101,32,99,114,101,97,116,101,100,0,69,114,114,111,114,58,32,99,97,110,39,116,32,111,112,101,110,32,100,105,114,101,99,116,111,114,121,32,45,32,37,115,10,0,37,115,47,37,115,0,100,105,114,0,99,100,101,118,0,98,100,101,118,0,102,105,102,111,0,108,105,110,107,0,63,0,102,105,108,101,0,37,115,32,40,100,105,114,41,10,0,37,115,32,40,37,115,41,58,10,9,115,105,122,101,58,32,37,100,10,9,99,114,101,58,32,37,115,9,109,111,100,58,32,37,115,10,0,37,115,32,40,117,110,107,110,111,119,110,41,10,0,102,111,111,0,32,58,44,0,76,84,86,49,0,76,84,86,50,0,76,84,77,49,95,49,0,76,84,77,49,95,50,0,76,84,77,50,95,49,0,76,84,77,50,95,50,0,69,82,82,79,82,58,32,99,97,110,39,116,32,111,112,101,110,32,111,117,116,112,117,116,32,102,105,108,101,58,32,37,115,10,0,34,104,100,117,34,58,32,37,100,0,44,32,34,110,97,109,101,34,58,32,34,37,115,34,0,44,32,34,116,121,112,101,34,58,32,34,105,109,97,103,101,34,44,32,34,110,97,120,105,115,34,58,32,37,100,0,44,32,34,110,97,120,101,115,34,58,32,91,0,44,32,34,98,105,116,112,105,120,34,58,32,37,100,0,44,32,34,116,121,112,101,34,58,32,34,97,115,99,105,105,34,0,44,32,34,116,121,112,101,34,58,32,34,116,97,98,108,101,34,0,44,32,34,114,111,119,115,34,58,32,37,108,100,0,44,32,34,99,111,108,115,34,58,32,91,0,123,34,110,97,109,101,34,58,32,34,37,115,34,44,32,34,116,121,112,101,34,58,32,34,37,115,34,125,0,44,10,0,93,10],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+344632);allocate([122,58,100,58,115,58,98,58,111,58,105,58,104,58,119,58,87,58,116,58,120,58,88,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,68,114,105,122,122,108,101,32,102,97,99,116,111,114,32,115,116,114,105,110,103,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,105,110,116,101,114,112,114,101,116,101,100,32,97,115,32,97,32,114,101,97,108,32,110,117,109,98,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,105,120,101,100,32,119,101,105,103,104,116,32,118,97,108,117,101,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,105,110,116,101,114,112,114,101,116,101,100,32,97,115,32,97,32,114,101,97,108,32,110,117,109,98,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,87,101,105,103,104,116,32,116,104,114,101,115,104,111,108,100,32,115,116,114,105,110,103,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,105,110,116,101,114,112,114,101,116,101,100,32,97,115,32,97,32,114,101,97,108,32,110,117,109,98,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,108,117,120,32,115,99,97,108,101,32,115,116,114,105,110,103,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,105,110,116,101,114,112,114,101,116,101,100,32,97,115,32,97,32,114,101,97,108,32,110,117,109,98,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,66,111,114,100,101,114,32,118,97,108,117,101,32,115,116,114,105,110,103,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,105,110,116,101,114,112,114,101,116,101,100,32,97,115,32,97,110,32,105,110,116,101,103,101,114,32,111,114,32,97,32,115,101,116,32,111,102,32,112,111,108,121,103,111,110,32,118,101,114,116,105,99,101,115,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,66,111,114,100,101,114,32,118,97,108,117,101,32,40,37,100,41,32,109,117,115,116,32,98,101,32,103,114,101,97,116,101,114,32,116,104,97,110,32,111,114,32,101,113,117,97,108,32,116,111,32,122,101,114,111,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,72,68,85,32,118,97,108,117,101,32,40,37,115,41,32,109,117,115,116,32,98,101,32,97,32,110,111,110,45,110,101,103,97,116,105,118,101,32,105,110,116,101,103,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,85,115,97,103,101,58,32,37,115,32,91,45,122,32,102,97,99,116,111,114,93,91,45,100,32,108,101,118,101,108,93,91,45,98,32,98,111,114,100,101,114,93,91,45,115,32,115,116,97,116,117,115,102,105,108,101,93,91,45,111,32,97,108,116,111,117,116,46,104,100,114,93,91,45,105,32,97,108,116,105,110,46,104,100,114,93,91,45,104,32,104,100,117,93,91,45,120,32,115,99,97,108,101,93,91,45,119,32,119,101,105,103,104,116,102,105,108,101,93,91,45,116,32,116,104,114,101,115,104,111,108,100,93,91,45,88,40,101,120,112,97,110,100,41,93,32,105,110,46,102,105,116,115,32,111,117,116,46,102,105,116,115,32,116,101,109,112,108,97,116,101,46,104,100,114,34,93,10,0,46,70,73,84,83,0,46,102,105,116,115,0,46,70,73,84,0,46,102,105,116,0,95,97,114,101,97,46,102,105,116,115,0,10,105,110,112,117,116,95,102,105,108,101,32,32,32,32,61,32,91,37,115,93,10,0,111,117,116,112,117,116,95,102,105,108,101,32,32,32,61,32,91,37,115,93,10,0,97,114,101,97,95,102,105,108,101,32,32,32,32,32,61,32,91,37,115,93,10,0,116,101,109,112,108,97,116,101,95,102,105,108,101,32,61,32,91,37,115,93,10,0,97,108,116,105,110,32,32,32,32,32,32,32,32,32,61,32,91,37,115,93,10,10,0,97,108,116,111,117,116,32,32,32,32,32,32,32,32,61,32,91,37,115,93,10,10,0,10,83,116,97,114,116,105,110,103,32,116,111,32,112,114,111,99,101,115,115,32,112,105,120,101,108,115,32,40,116,105,109,101,32,37,46,48,102,41,10,10,0,105,110,112,117,116,46,110,97,120,101,115,91,48,93,32,32,32,61,32,32,37,108,100,10,0,105,110,112,117,116,46,110,97,120,101,115,91,49,93,32,32,32,61,32,32,37,108,100,10,0,105,110,112,117,116,46,115,121,115,32,32,32,32,32,32,32,32,61,32,32,37,100,10,0,105,110,112,117,116,46,101,112,111,99,104,32,32,32,32,32,32,61,32,32,37,45,103,10,0,105,110,112,117,116,46,99,108,111,99,107,119,105,115,101,32,32,61,32,32,37,100,10,0,105,110,112,117,116,32,112,114,111,106,32,32,32,32,32,32,32,61,32,32,37,115,10,10,0,73,110,112,117,116,32,105,109,97,103,101,32,112,114,111,106,101,99,116,105,111,110,32,40,37,115,41,32,109,117,115,116,32,98,101,32,84,65,78,44,32,83,73,78,44,32,90,69,65,44,32,83,84,71,32,111,114,32,65,82,67,32,102,111,114,32,102,97,115,116,32,114,101,112,114,111,106,101,99,116,105,111,110,0,10,101,120,112,97,110,100,32,111,117,116,112,117,116,32,116,101,109,112,108,97,116,101,32,98,121,32,37,45,103,32,111,110,32,97,108,108,32,115,105,100,101,115,10,10,0,10,111,117,116,112,117,116,46,110,97,120,101,115,91,48,93,32,32,61,32,32,37,108,100,10,0,111,117,116,112,117,116,46,110,97,120,101,115,91,49,93,32,32,61,32,32,37,108,100,10,0,111,117,116,112,117,116,46,115,121,115,32,32,32,32,32,32,32,61,32,32,37,100,10,0,111,117,116,112,117,116,46,101,112,111,99,104,32,32,32,32,32,61,32,32,37,45,103,10,0,111,117,116,112,117,116,46,99,108,111,99,107,119,105,115,101,32,61,32,32,37,100,10,0,111,117,116,112,117,116,32,112,114,111,106,32,32,32,32,32,32,61,32,32,37,115,10,0,79,117,116,112,117,116,32,105,109,97,103,101,32,112,114,111,106,101,99,116,105,111,110,32,40,37,115,41,32,109,117,115,116,32,98,101,32,84,65,78,44,32,83,73,78,44,32,90,69,65,44,32,83,84,71,32,111,114,32,65,82,67,32,102,111,114,32,102,97,115,116,32,114,101,112,114,111,106,101,99,116,105,111,110,0,73,110,112,117,116,32,97,110,100,32,111,117,116,112,117,116,32,109,117,115,116,32,98,101,32,105,110,32,116,104,101,32,115,97,109,101,32,99,111,111,114,100,105,110,97,116,101,32,115,121,115,116,101,109,32,102,111,114,32,102,97,115,116,32,114,101,112,114,111,106,101,99,116,105,111,110,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,67,111,117,108,100,32,110,111,116,32,115,101,116,32,117,112,32,112,108,97,110,101,45,116,111,45,112,108,97,110,101,32,116,114,97,110,115,102,111,114,109,46,32,32,67,104,101,99,107,32,102,111,114,32,99,111,109,112,108,105,97,110,116,32,104,101,97,100,101,114,115,46,34,93,10,0,73,110,105,116,105,97,108,105,122,101,95,84,119,111,80,108,97,110,101,95,66,111,116,104,68,105,115,116,111,114,116,40,41,32,115,117,99,99,101,115,115,102,117,108,0,82,97,110,103,101,58,32,37,45,103,44,37,45,103,32,45,62,32,37,45,103,44,37,45,103,32,40,37,100,41,10,0,32,32,32,32,32,45,62,32,37,45,103,44,37,45,103,32,45,62,37,45,103,44,37,45,103,32,40,37,100,41,10,0,10,79,117,116,112,117,116,32,114,97,110,103,101,58,0,32,111,120,112,105,120,77,105,110,32,61,32,37,45,103,10,0,32,111,120,112,105,120,77,97,120,32,61,32,37,45,103,10,0,32,111,121,112,105,120,77,105,110,32,61,32,37,45,103,10,0,32,111,121,112,105,120,77,97,120,32,61,32,37,45,103,10,0,32,105,115,116,97,114,116,32,32,32,61,32,37,45,100,10,0,32,105,108,101,110,103,116,104,32,32,61,32,37,45,100,10,0,32,106,115,116,97,114,116,32,32,32,61,32,37,45,100,10,0,32,106,108,101,110,103,116,104,32,32,61,32,37,45,100,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,78,111,32,111,118,101,114,108,97,112,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,78,111,116,32,101,110,111,117,103,104,32,109,101,109,111,114,121,32,102,111,114,32,111,117,116,112,117,116,32,100,97,116,97,32,105,109,97,103,101,32,97,114,114,97,121,34,93,10,0,10,37,108,117,32,98,121,116,101,115,32,97,108,108,111,99,97,116,101,100,32,102,111,114,32,105,109,97,103,101,32,112,105,120,101,108,115,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,78,111,116,32,101,110,111,117,103,104,32,109,101,109,111,114,121,32,102,111,114,32,111,117,116,112,117,116,32,97,114,101,97,32,105,109,97,103,101,32,97,114,114,97,121,34,93,10,0,37,108,117,32,98,121,116,101,115,32,97,108,108,111,99,97,116,101,100,32,102,111,114,32,112,105,120,101,108,32,97,114,101,97,115,10,0,13,80,114,111,99,101,115,115,105,110,103,32,105,110,112,117,116,32,114,111,119,32,37,53,100,58,32,98,111,114,100,101,114,32,114,97,110,103,101,32,37,100,32,116,111,32,37,100,32,40,37,100,41,0,13,80,114,111,99,101,115,115,105,110,103,32,105,110,112,117,116,32,114,111,119,32,37,53,100,32,32,0,10,73,110,112,117,116,58,32,108,105,110,101,32,37,100,32,47,32,112,105,120,101,108,32,37,100,44,32,118,97,108,117,101,32,61,32,37,45,103,32,40,119,101,105,103,104,116,58,32,37,45,103,41,10,10,0,10,73,110,112,117,116,58,32,108,105,110,101,32,37,100,32,47,32,112,105,120,101,108,32,37,100,44,32,118,97,108,117,101,32,61,32,37,45,103,10,10,0,32,111,120,112,105,120,77,105,110,32,61,32,37,50,48,46,49,51,101,10,0,32,111,120,112,105,120,77,97,120,32,61,32,37,50,48,46,49,51,101,10,0,32,111,121,112,105,120,77,105,110,32,61,32,37,50,48,46,49,51,101,10,0,32,111,121,112,105,120,77,97,120,32,61,32,37,50,48,46,49,51,101,10,0,79,117,116,112,117,116,32,88,32,114,97,110,103,101,58,32,37,53,100,32,116,111,32,37,53,100,10,0,79,117,116,112,117,116,32,89,32,114,97,110,103,101,58,32,37,53,100,32,116,111,32,37,53,100,10,0,67,111,109,112,97,114,101,32,111,117,116,40,37,100,44,37,100,41,32,116,111,32,105,110,40,37,100,44,37,100,41,32,61,62,32,0,111,118,101,114,108,97,112,65,114,101,97,32,61,32,37,49,50,46,53,101,32,40,37,49,50,46,53,101,32,47,32,37,49,50,46,53,101,41,10,0,10,10,68,111,110,101,32,112,114,111,99,101,115,115,105,110,103,32,112,105,120,101,108,115,32,40,37,46,48,102,32,115,101,99,111,110,100,115,41,10,10,0,68,97,116,97,32,109,105,110,32,61,32,37,45,103,10,0,68,97,116,97,32,109,97,120,32,61,32,37,45,103,10,0,65,114,101,97,32,109,105,110,32,61,32,37,45,103,10,0,65,114,101,97,32,109,97,120,32,61,32,37,45,103,10,10,0,105,32,109,105,110,32,32,32,32,61,32,37,100,10,0,105,32,109,97,120,32,32,32,32,61,32,37,100,10,0,106,32,109,105,110,32,32,32,32,61,32,37,100,10,0,106,32,109,97,120,32,32,32,32,61,32,37,100,10,0,65,108,108,32,112,105,120,101,108,115,32,97,114,101,32,98,108,97,110,107,46,32,67,104,101,99,107,32,102,111,114,32,111,118,101,114,108,97,112,32,111,102,32,111,117,116,112,117,116,32,116,101,109,112,108,97,116,101,32,119,105,116,104,32,105,109,97,103,101,32,102,105,108,101,46,0,10,70,73,84,83,32,100,97,116,97,32,105,109,97,103,101,32,99,114,101,97,116,101,100,32,40,110,111,116,32,121,101,116,32,112,111,112,117,108,97,116,101,100,41,0,70,73,84,83,32,97,114,101,97,32,105,109,97,103,101,32,99,114,101,97,116,101,100,32,40,110,111,116,32,121,101,116,32,112,111,112,117,108,97,116,101,100,41,0,84,101,109,112,108,97,116,101,32,107,101,121,119,111,114,100,115,32,119,114,105,116,116,101,110,32,116,111,32,70,73,84,83,32,100,97,116,97,32,105,109,97,103,101,0,84,101,109,112,108,97,116,101,32,107,101,121,119,111,114,100,115,32,119,114,105,116,116,101,110,32,116,111,32,70,73,84,83,32,97,114,101,97,32,105,109,97,103,101,10,0,84,101,109,112,108,97,116,101,32,107,101,121,119,111,114,100,115,32,66,73,84,80,73,88,44,32,67,82,80,73,88,44,32,97,110,100,32,78,65,88,73,83,32,117,112,100,97,116,101,100,0,68,97,116,97,32,119,114,105,116,116,101,110,32,116,111,32,70,73,84,83,32,100,97,116,97,32,105,109,97,103,101,0,68,97,116,97,32,119,114,105,116,116,101,110,32,116,111,32,70,73,84,83,32,97,114,101,97,32,105,109,97,103,101,10,0,70,73,84,83,32,100,97,116,97,32,105,109,97,103,101,32,102,105,110,97,108,105,122,101,100,0,70,73,84,83,32,97,114,101,97,32,105,109,97,103,101,32,102,105,110,97,108,105,122,101,100,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,79,75,34,44,32,116,105,109,101,61,37,46,48,102,93,10,0,80,111,108,121,103,111,110,32,115,116,114,105,110,103,58,32,91,37,115,93,10,0,80,111,108,121,103,111,110,32,98,111,114,100,101,114,32,32,37,51,100,58,32,37,54,100,32,37,54,100,10,0,73,109,97,103,101,32,102,105,108,101,32,37,115,32,109,105,115,115,105,110,103,32,111,114,32,105,110,118,97,108,105,100,32,70,73,84,83,0,87,101,105,103,104,116,32,102,105,108,101,32,37,115,32,109,105,115,115,105,110,103,32,111,114,32,105,110,118,97,108,105,100,32,70,73,84,83,0,73,110,112,117,116,32,104,101,97,100,101,114,32,116,111,32,119,99,115,105,110,105,116,40,41,32,91,105,110,112,117,116,46,119,99,115,93,58,10,37,115,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,112,117,116,32,119,99,115,105,110,105,116,40,41,32,102,97,105,108,101,100,46,34,93,10,0,73,110,112,117,116,32,112,105,120,101,108,115,32,97,114,101,32,99,108,111,99,107,119,105,115,101,46,0,73,110,112,117,116,32,112,105,120,101,108,115,32,97,114,101,32,99,111,117,110,116,101,114,99,108,111,99,107,119,105,115,101,46,0,114,101,97,100,84,101,109,112,108,97,116,101,40,41,32,102,105,108,101,32,61,32,91,37,115,93,10,0,84,101,109,112,108,97,116,101,32,102,105,108,101,32,91,37,115,93,32,110,111,116,32,102,111,117,110,100,46,0,84,101,109,112,108,97,116,101,32,108,105,110,101,58,32,91,37,115,93,10,0,65,108,116,101,114,110,97,116,101,32,105,110,112,117,116,32,104,101,97,100,101,114,32,116,111,32,119,99,115,105,110,105,116,40,41,32,91,105,110,112,117,116,46,119,99,115,93,58,10,37,115,10,0,65,108,116,101,114,110,97,116,101,32,111,117,116,112,117,116,32,104,101,97,100,101,114,32,116,111,32,119,99,115,105,110,105,116,40,41,32,91,111,117,116,112,117,116,46,119,99,115,93,58,10,37,115,10,0,84,101,109,112,108,97,116,101,32,111,117,116,112,117,116,32,104,101,97,100,101,114,32,116,111,32,119,99,115,105,110,105,116,40,41,32,91,111,117,116,112,117,116,46,119,99,115,93,58,10,37,115,10,0,79,117,116,112,117,116,32,112,105,120,101,108,115,32,97,114,101,32,99,108,111,99,107,119,105,115,101,46,0,79,117,116,112,117,116,32,112,105,120,101,108,115,32,97,114,101,32,99,111,117,110,116,101,114,99,108,111,99,107,119,105,115,101,46,0,66,111,117,110,100,115,58,32,37,45,103,44,37,45,103,32,45,62,32,37,45,103,44,37,45,103,32,40,37,100,41,10,0,107,101,121,119,111,114,100,32,91,37,115,93,32,61,32,118,97,108,117,101,32,91,37,115,93,10,0,78,65,88,73,83,49,32,32,61,32,37,108,100,0,78,65,88,73,83,50,32,32,61,32,37,108,100,0,67,82,80,73,88,49,32,32,61,32,37,49,49,46,54,102,0,67,82,80,73,88,50,32,32,61,32,37,49,49,46,54,102],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+586758);allocate([100,117,105,58,111,58,116,58,115,58,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,45,105,32,40,105,116,101,114,97,116,105,111,110,115,41,32,97,114,103,117,109,101,110,116,32,34,37,115,34,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,45,111,32,40,111,114,100,101,114,41,32,97,114,103,117,109,101,110,116,32,34,37,115,34,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,45,116,32,40,116,111,108,101,114,97,110,99,101,41,32,97,114,103,117,109,101,110,116,32,34,37,115,34,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,67,97,110,110,111,116,32,111,112,101,110,32,115,116,97,116,117,115,32,102,105,108,101,58,32,37,115,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,85,115,97,103,101,58,32,37,115,32,91,45,100,93,91,45,111,32,111,114,100,101,114,93,91,45,105,32,109,97,120,105,116,101,114,93,91,45,116,32,116,111,108,101,114,97,110,99,101,93,91,45,115,32,115,116,97,116,117,115,102,105,108,101,93,32,111,114,105,103,46,104,100,114,32,110,101,119,46,104,100,114,32,40,100,101,102,97,117,108,116,58,32,111,114,100,101,114,32,61,32,53,44,32,109,97,120,105,116,101,114,32,61,32,53,48,44,32,116,111,108,101,114,97,110,99,101,32,61,32,48,46,48,49,41,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,67,97,110,110,111,116,32,111,112,101,110,32,111,117,116,112,117,116,32,116,101,109,112,108,97,116,101,32,102,105,108,101,32,37,115,34,93,10,0,68,69,66,85,71,62,32,67,111,109,109,97,110,100,45,108,105,110,101,32,114,101,97,100,46,0,68,69,66,85,71,62,32,70,87,68,58,32,32,110,97,116,105,118,101,32,112,114,111,106,101,99,116,105,111,110,32,116,111,32,100,105,115,116,111,114,116,101,100,32,84,65,78,32,40,119,99,115,45,62,87,67,83,41,0,68,69,66,85,71,62,32,82,69,86,58,32,32,100,105,115,116,111,114,116,101,100,32,84,65,78,32,116,111,32,110,97,116,105,118,101,32,112,114,111,106,101,99,116,105,111,110,32,40,87,67,83,45,62,119,99,115,41,0,68,69,66,85,71,62,32,68,105,115,116,111,114,116,105,111,110,32,112,97,114,97,109,101,116,101,114,115,32,105,110,105,116,105,97,108,105,122,101,100,46,0,37,49,53,46,49,48,102,0,10,68,69,66,85,71,62,32,70,87,68,58,32,65,32,40,120,45,100,105,114,101,99,116,105,111,110,41,32,100,105,115,116,111,114,116,105,111,110,115,32,91,37,100,93,10,0,68,69,66,85,71,62,32,70,87,68,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,46,53,102,32,37,46,53,102,32,45,62,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,49,48,46,50,102,32,40,37,100,41,10,0,10,68,69,66,85,71,62,32,70,87,68,58,32,66,32,40,121,45,100,105,114,101,99,116,105,111,110,41,32,100,105,115,116,111,114,116,105,111,110,115,32,91,37,100,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,65,108,108,32,112,111,105,110,116,115,32,111,102,102,115,99,97,108,101,32,105,110,32,102,111,114,119,97,114,100,32,116,114,97,110,115,102,111,114,109,34,93,10,0,10,10,70,87,68,58,32,66,101,102,111,114,101,32,103,97,117,115,115,106,40,41,58,0,109,97,116,114,105,120,58,0,37,49,50,46,53,101,32,0,10,118,101,99,116,111,114,58,0,32,32,37,49,50,46,53,101,10,0,10,10,65,102,116,101,114,32,103,97,117,115,115,106,40,41,58,0,97,112,91,37,100,93,91,37,100,93,32,61,32,37,49,50,46,53,101,32,40,37,53,46,49,102,37,37,41,10,0,98,112,91,37,100,93,91,37,100,93,32,61,32,37,49,50,46,53,101,32,40,37,53,46,49,102,37,37,41,10,0,10,68,69,66,85,71,62,32,70,87,68,58,32,109,97,120,105,109,117,109,32,112,111,115,105,116,105,111,110,97,108,32,101,114,114,111,114,32,91,37,100,93,10,0,68,69,66,85,71,62,32,70,87,68,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,46,53,102,32,37,46,53,102,32,45,62,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,40,37,100,41,10,0,102,119,100,109,97,120,120,32,61,32,37,45,103,32,91,37,100,93,10,0,102,119,100,109,97,120,121,32,61,32,37,45,103,32,91,37,100,93,10,0,10,68,69,66,85,71,62,32,82,69,86,58,32,65,32,40,120,45,100,105,114,101,99,116,105,111,110,41,32,100,105,115,116,111,114,116,105,111,110,115,32,91,37,100,93,10,0,68,69,66,85,71,62,32,82,69,86,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,46,53,102,32,37,46,53,102,32,45,62,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,49,48,46,50,102,32,40,37,100,41,10,0,10,68,69,66,85,71,62,32,82,69,86,58,32,66,32,40,121,45,100,105,114,101,99,116,105,111,110,41,32,100,105,115,116,111,114,116,105,111,110,115,32,91,37,100,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,65,108,108,32,112,111,105,110,116,115,32,111,102,102,115,99,97,108,101,32,105,110,32,114,101,118,101,114,115,101,32,116,114,97,110,115,102,111,114,109,34,93,10,0,10,10,82,69,86,58,32,66,101,102,111,114,101,32,103,97,117,115,115,106,40,41,58,0,97,91,37,100,93,91,37,100,93,32,61,32,37,49,50,46,53,101,32,40,37,53,46,49,102,37,37,41,10,0,98,91,37,100,93,91,37,100,93,32,61,32,37,49,50,46,53,101,32,40,37,53,46,49,102,37,37,41,10,0,10,68,69,66,85,71,62,32,82,69,86,58,32,109,97,120,105,109,117,109,32,112,111,115,105,116,105,111,110,97,108,32,101,114,114,111,114,32,91,37,100,93,10,0,68,69,66,85,71,62,32,82,69,86,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,45,62,32,37,46,53,102,32,37,46,53,102,32,45,62,32,37,49,48,46,50,102,32,37,49,48,46,50,102,32,40,37,100,41,10,0,114,101,118,109,97,120,120,32,61,32,37,45,103,32,91,37,100,93,10,0,114,101,118,109,97,120,121,32,61,32,37,45,103,32,91,37,100,93,10,0,10,68,69,66,85,71,62,32,70,87,68,58,32,87,111,114,115,116,32,99,97,115,101,32,112,105,120,101,108,32,101,114,114,111,114,0,102,105,110,97,108,32,102,119,100,109,97,120,120,32,61,32,37,45,103,10,0,102,105,110,97,108,32,102,119,100,109,97,120,121,32,61,32,37,45,103,10,0,10,68,69,66,85,71,62,32,82,69,86,58,32,87,111,114,115,116,32,99,97,115,101,32,112,105,120,101,108,32,101,114,114,111,114,0,102,105,110,97,108,32,114,101,118,109,97,120,120,32,61,32,37,45,103,10,0,102,105,110,97,108,32,114,101,118,109,97,120,121,32,61,32,37,45,103,10,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,79,75,34,44,32,102,119,100,120,101,114,114,61,37,45,103,44,32,102,119,100,121,101,114,114,61,37,45,103,44,32,102,119,100,105,116,101,114,61,37,100,44,32,114,101,118,120,101,114,114,61,37,45,103,44,32,114,101,118,121,101,114,114,61,37,45,103,44,32,114,101,118,105,116,101,114,61,37,100,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,66,97,100,32,116,101,109,112,108,97,116,101,58,32,37,115,34,93,10,0,10,101,120,116,114,97,99,116,67,68,40,41,58,0,99,100,101,108,116,49,32,32,61,32,91,37,115,93,10,0,99,100,101,108,116,50,32,32,61,32,91,37,115,93,10,0,99,114,111,116,97,50,32,32,61,32,91,37,115,93,10,0,99,100,49,49,32,32,32,32,61,32,91,37,115,93,10,0,99,100,49,50,32,32,32,32,61,32,91,37,115,93,10,0,99,100,50,49,32,32,32,32,61,32,91,37,115,93,10,0,99,100,50,50,32,32,32,32,61,32,91,37,115,93,10,0,112,99,49,49,32,32,32,32,61,32,91,37,115,93,10,0,112,99,49,50,32,32,32,32,61,32,91,37,115,93,10,0,112,99,50,49,32,32,32,32,61,32,91,37,115,93,10,0,112,99,50,50,32,32,32,32,61,32,91,37,115,93,10,0,101,112,111,99,104,32,32,32,61,32,91,37,115,93,10,0,101,113,117,105,110,111,120,32,61,32,91,37,115,93,10,0,10,68,69,66,85,71,62,32,79,114,105,103,105,110,97,108,32,72,101,97,100,101,114,58,10,0,37,52,100,58,32,37,115,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,79,117,116,112,117,116,32,119,99,115,105,110,105,116,40,41,32,102,97,105,108,101,100,46,34,93,10,0,68,69,66,85,71,62,32,79,114,105,103,105,110,97,108,32,105,109,97,103,101,32,87,67,83,32,105,110,105,116,105,97,108,105,122,101,100,10,0,68,69,66,85,71,62,32,120,99,111,114,114,101,99,116,105,111,110,32,61,32,37,46,50,102,10,0,68,69,66,85,71,62,32,121,99,111,114,114,101,99,116,105,111,110,32,61,32,37,46,50,102,10,10,0,65,108,108,111,99,97,116,105,111,110,32,102,97,105,108,117,114,101,32,105,110,32,105,118,101,99,116,111,114,40,41,0,83,105,110,103,117,108,97,114,32,77,97,116,114,105,120,45,49,0,83,105,110,103,117,108,97,114,32,77,97,116,114,105,120,45,50,0,10,10,67,111,114,114,101,108,97,116,105,111,110,32,77,97,116,114,105,120,58,0,37,53,46,50,102,32,0,10,77,97,120,105,109,117,109,32,99,111,114,114,101,108,97,116,105,111,110,58,32,37,46,53,102,32,97,116,32,40,37,100,44,37,100,41,10,10,0,83,73,77,80,76,69,32,32,61,32,84,0,66,73,84,80,73,88,32,32,61,32,45,54,52,0,78,65,88,73,83,32,32,32,61,32,50,0,78,65,88,73,83,49,32,32,61,32,37,100,0,78,65,88,73,83,50,32,32,61,32,37,100,0,67,84,89,80,69,49,32,32,61,32,39,82,65,45,45,45,84,65,78,45,83,73,80,39,0,67,84,89,80,69,50,32,32,61,32,39,68,69,67,45,45,84,65,78,45,83,73,80,39,0,67,84,89,80,69,49,32,32,61,32,39,37,115,45,84,65,78,45,83,73,80,39,0,67,84,89,80,69,50,32,32,61,32,39,37,115,45,84,65,78,45,83,73,80,39,0,67,82,86,65,76,49,32,32,61,32,37,49,53,46,49,48,102,0,67,82,86,65,76,50,32,32,61,32,37,49,53,46,49,48,102,0,67,82,80,73,88,49,32,32,61,32,37,49,53,46,49,48,102,0,67,82,80,73,88,50,32,32,61,32,37,49,53,46,49,48,102,0,67,68,69,76,84,49,32,32,61,32,37,46,49,48,102,0,67,68,69,76,84,50,32,32,61,32,37,46,49,48,102,0,67,82,79,84,65,50,32,32,61,32,37,115,0,67,68,49,95,49,32,32,32,61,32,37,115,0,67,68,49,95,50,32,32,32,61,32,37,115,0,67,68,50,95,49,32,32,32,61,32,37,115,0,67,68,50,95,50,32,32,32,61,32,37,115,0,80,67,49,95,49,32,32,32,61,32,37,115,0,80,67,49,95,50,32,32,32,61,32,37,115,0,80,67,50,95,49,32,32,32,61,32,37,115,0,80,67,50,95,50,32,32,32,61,32,37,115,0,69,80,79,67,72,32,32,32,61,32,37,115,0,69,81,85,73,78,79,88,32,61,32,37,115,0,65,95,79,82,68,69,82,32,61,32,37,100,0,65,95,37,100,95,37,100,32,32,32,61,32,37,49,48,46,51,101,0,66,95,79,82,68,69,82,32,61,32,37,100,0,66,95,37,100,95,37,100,32,32,32,61,32,37,49,48,46,51,101,0,65,80,95,79,82,68,69,82,61,32,37,100,0,65,80,95,37,100,95,37,100,32,32,61,32,37,49,48,46,51,101,0,66,80,95,79,82,68,69,82,61,32,37,100,0,66,80,95,37,100,95,37,100,32,32,61,32,37,49,48,46,51,101,0,10,68,69,66,85,71,62,32,68,105,115,116,111,114,116,101,100,32,84,65,78,32,72,101,97,100,101,114,58,10,0,68,69,66,85,71,62,32,68,105,115,116,111,114,116,101,100,32,84,65,78,32,87,67,83,32,105,110,105,116,105,97,108,105,122,101,100,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,32,109,115,103,61,34,37,115,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,68,101,98,117,103,32,108,101,118,101,108,32,115,116,114,105,110,103,32,105,115,32,105,110,118,97,108,105,100,58,32,39,37,115,39,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,68,101,98,117,103,32,108,101,118,101,108,32,118,97,108,117,101,32,99,97,110,110,111,116,32,98,101,32,110,101,103,97,116,105,118,101,34,93,10],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+591855);allocate([119,43,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,67,97,110,110,111,116,32,111,112,101,110,32,111,117,116,112,117,116,32,102,105,108,101,32,37,115,46,34,93,10,0,10,70,73,84,83,32,102,105,108,101,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,73,84,83,32,102,105,108,101,32,40,37,115,41,32,99,97,110,110,111,116,32,98,101,32,117,115,101,100,32,97,115,32,97,32,104,101,97,100,101,114,32,116,101,109,112,108,97,116,101,34,93,10,0,110,99,97,114,100,32,61,32,37,100,10,0,37,45,56,115,61,32,37,50,48,115,0,10,84,101,109,112,108,97,116,101,32,102,105,108,101,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,105,108,101,32,37,115,32,110,111,116,32,102,111,117,110,100,46,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,105,108,101,32,40,37,115,41,32,105,115,32,110,111,116,32,97,32,70,73,84,83,32,105,109,97,103,101,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,70,73,84,83,32,104,101,97,100,101,114,32,108,105,110,101,115,32,99,97,110,110,111,116,32,98,101,32,103,114,101,97,116,101,114,32,116,104,97,110,32,56,48,32,99,104,97,114,97,99,116,101,114,115,46,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,37,115,34,93,10,0,78,111,32,66,73,84,80,73,88,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,78,65,88,73,83,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,78,65,88,73,83,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,78,65,88,73,83,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,84,89,80,69,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,84,89,80,69,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,82,80,73,88,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,82,80,73,88,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,82,86,65,76,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,82,86,65,76,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,68,69,76,84,49,32,107,101,121,119,111,114,100,32,40,111,114,32,105,110,99,111,109,112,108,101,116,101,32,67,68,32,109,97,116,114,105,120,41,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,78,111,32,67,68,69,76,84,50,32,107,101,121,119,111,114,100,32,40,111,114,32,105,110,99,111,109,112,108,101,116,101,32,67,68,32,109,97,116,114,105,120,41,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,0,67,84,89,80,69,49,32,109,117,115,116,32,98,101,32,97,116,32,108,101,97,115,116,32,56,32,99,104,97,114,97,99,116,101,114,115,0,67,84,89,80,69,50,32,109,117,115,116,32,98,101,32,97,116,32,108,101,97,115,116,32,56,32,99,104,97,114,97,99,116,101,114,115,0,73,110,118,97,108,105,100,32,67,84,89,80,69,49,32,111,114,32,67,84,89,80,69,50,32,112,114,111,106,101,99,116,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,0,67,84,89,80,69,49,44,32,67,84,89,80,69,50,32,112,114,111,106,101,99,116,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,32,109,105,115,109,97,116,99,104,0,73,110,118,97,108,105,100,32,67,84,89,80,69,49,32,112,114,111,106,101,99,116,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,0,73,110,118,97,108,105,100,32,67,84,89,80,69,50,32,112,114,111,106,101,99,116,105,111,110,32,105,110,102,111,114,109,97,116,105,111,110,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,37,100,32,69,114,114,111,114,115,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,115,116,97,116,117,115,61,37,100,44,32,109,115,103,61,34,37,115,34,93,10,0,102,105,116,115,67,104,101,99,107,40,41,32,91,37,115,93,32,61,32,91,37,115,93,10,0,83,73,77,80,76,69,32,107,101,121,119,111,114,100,32,109,117,115,116,32,98,101,32,84,32,111,114,32,70,0,66,73,84,80,73,88,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,0,73,110,118,97,108,105,100,32,66,73,84,80,73,88,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,40,109,117,115,116,32,98,101,32,56,44,49,54,44,51,50,44,54,52,44,45,51,50,32,111,114,32,45,54,52,41,0,78,65,88,73,83,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,0,78,65,88,73,83,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,109,117,115,116,32,98,101,32,62,61,32,50,0,78,65,88,73,83,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,0,78,65,88,73,83,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,109,117,115,116,32,98,101,32,62,32,48,0,78,65,88,73,83,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,0,78,65,88,73,83,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,109,117,115,116,32,98,101,32,62,32,48,0,67,82,80,73,88,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,82,80,73,88,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,82,86,65,76,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,82,86,65,76,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,68,69,76,84,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,68,69,76,84,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,82,79,84,65,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,68,49,95,49,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,68,49,95,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,67,68,50,95,50,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,66,83,67,65,76,69,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,66,90,69,82,79,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,66,76,65,78,75,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,69,80,79,67,72,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,69,81,85,73,78,79,88,32,107,101,121,119,111,114,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,32,110,111,116,32,97,32,114,101,97,108,32,110,117,109,98,101,114,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,78,111,32,87,67,83,32,105,110,102,111,114,109,97,116,105,111,110,32,40,111,114,32,110,111,116,32,70,73,84,83,32,104,101,97,100,101,114,41,34,93,10,0,112,114,106,99,111,100,101,32,32,32,32,61,32,37,100,10,0,112,116,121,112,101,32,32,32,32,32,32,61,32,91,37,115,93,10,0,110,120,112,105,120,32,32,32,32,32,32,61,32,37,45,103,10,0,110,121,112,105,120,32,32,32,32,32,32,61,32,37,45,103,10,0,99,49,116,121,112,101,32,32,32,32,32,61,32,91,37,115,93,10,0,99,50,116,121,112,101,32,32,32,32,32,61,32,91,37,115,93,10,0,110,97,120,105,115,32,32,32,32,32,32,61,32,37,100,10,0,99,114,118,97,108,91,37,100,93,32,32,32,61,32,37,45,103,10,0,120,114,101,102,32,32,32,32,32,32,32,61,32,37,45,103,10,0,121,114,101,102,32,32,32,32,32,32,32,61,32,37,45,103,10,0,99,114,112,105,120,91,37,100,93,32,32,32,61,32,37,45,103,10,0,120,114,101,102,112,105,120,32,32,32,32,61,32,37,45,103,10,0,121,114,101,102,112,105,120,32,32,32,32,61,32,37,45,103,10,0,99,100,91,37,100,93,32,32,32,32,32,32,61,32,37,45,103,10,0,99,100,101,108,116,91,37,100,93,32,32,32,61,32,37,45,103,10,0,120,105,110,99,32,32,32,32,32,32,32,61,32,37,45,103,10,0,121,105,110,99,32,32,32,32,32,32,32,61,32,37,45,103,10,0,114,111,116,32,32,32,32,32,32,32,32,61,32,37,45,103,10,0,101,113,117,105,110,111,120,32,32,32,32,61,32,37,45,103,10,0,101,112,111,99,104,32,32,32,32,32,32,61,32,37,45,103,10,0,112,99,91,37,50,100,93,32,32,32,32,32,61,32,37,45,103,10,0,112,114,111,106,112,91,37,50,100,93,32,32,61,32,37,45,103,10,0,108,111,110,103,112,111,108,101,32,32,32,61,32,37,45,103,10,0,108,97,116,112,111,108,101,32,32,32,32,61,32,37,45,103,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,77,117,115,116,32,104,97,118,101,32,97,116,32,108,101,97,115,116,32,116,119,111,32,100,105,109,101,110,115,105,111,110,115,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,112,114,111,106,101,99,116,105,111,110,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,78,65,88,73,83,49,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,78,65,88,73,83,50,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,67,84,89,80,69,49,32,97,110,100,32,67,84,89,80,69,50,32,100,111,110,39,116,32,109,97,116,99,104,34,93,10,0,71,76,79,78,0,71,76,65,84,0,69,76,79,78,0,69,76,65,84,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,67,84,89,80,69,49,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,67,68,32,109,97,116,114,105,120,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,67,68,69,76,84,49,34,93,10,0,91,115,116,114,117,99,116,32,115,116,97,116,61,34,69,82,82,79,82,34,44,32,109,115,103,61,34,73,110,118,97,108,105,100,32,67,68,69,76,84,50,34,93,10,0,69,82,82,79,82,58,32,68,76,67,83,58,32,115,111,109,101,116,104,105,110,103,32,105,115,32,119,114,111,110,103,32,119,105,116,104,32,116,104,101,32,99,100,45,109,97,116,114,105,120,32,102,111,114,32,116,104,101,32,115,101,99,111,110,100,32,102,114,97,109,101,10,0,69,82,82,79,82,58,32,68,76,67,83,58,32,115,111,109,101,116,104,105,110,103,32,105,115,32,119,114,111,110,103,32,119,105,116,104,32,116,104,101,32,99,100,45,109,97,116,114,105,120,32,102,111,114,32,116,104,101,32,102,105,114,115,116,32,102,114,97,109,101,10,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,107,101,121,119,111,114,100,32,91,37,115,93,10,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,67,82,80,73,88,49,10,0,87,67,83,73,78,73,84,78,58,32,87,67,83,32,110,97,109,101,32,37,115,32,110,111,116,32,109,97,116,99,104,101,100,32,105,110,32,70,73,84,83,32,104,101,97,100,101,114,10,0,87,67,83,78,65,77,69,0,87,67,83,65,88,69,83,0,87,67,83,68,73,77,0,87,67,83,73,78,73,84,58,32,78,111,32,87,67,83,65,88,69,83,44,32,78,65,88,73,83,44,32,111,114,32,87,67,83,68,73,77,32,107,101,121,119,111,114,100,0,73,77,65,71,69,87,0,87,67,83,73,78,73,84,58,32,78,111,32,78,65,88,73,83,49,32,111,114,32,73,77,65,71,69,87,32,107,101,121,119,111,114,100,0,73,77,65,71,69,72,0,87,67,83,73,78,73,84,58,32,78,111,32,78,65,88,73,83,50,32,111,114,32,73,77,65,71,69,72,32,107,101,121,119,111,114,100,0,87,67,83,73,78,73,84,58,32,77,105,115,115,105,110,103,32,107,101,121,119,111,114,100,32,37,115,32,97,115,115,117,109,101,100,32,49,10,0,45,84,65,66,0,68,69,84,69,67,84,79,82,0,68,67,45,70,76,65,71,0,87,67,83,68,69,80,0,87,67,83,73,78,73,84,58,32,100,101,112,101,110,100,101,100,32,111,110,32,87,67,83,32,99,111,117,108,100,32,110,111,116,32,98,101,32,115,101,116,0,86,83,79,85,82,67,69,0,90,83,79,85,82,67,69,0,86,69,76,79,67,73,84,89,0,67,84,89,80,69,50,0,67,84,89,80,69,51,0,67,84,89,80,69,52,0,67,85,78,73,84,49,0,117,110,105,116,115,0,67,85,78,73,84,50,0,76,79,78,80,79,76,69,0,76,65,84,80,79,76,69,0,80,82,79,74,82,48,0,80,82,79,74,80,37,100,0,80,86,37,100,95,49,0,80,86,37,100,95,50,0,80,86,37,100,95,51,0,80,86,37,100,95,37,100,0,67,79,49,95,49,0,80,76,65,84,69,0,67,79,49,95,37,100,0,67,79,50,95,49,0,67,79,50,95,37,100,0,83,69,67,80,73,88,0,80,73,88,83,67,65,76,69,0,80,73,88,83,67,65,76,49,0,88,80,73,88,83,73,90,69,0,83,69,67,80,73,88,49,0,83,69,67,80,73,88,50,0,89,80,73,88,83,73,90,69,0,80,73,88,83,67,65,76,50,0,80,67,48,48,49,48,48,49,0,80,67,48,48,37,49,100,48,48,37,49,100,0,80,67,37,49,100,95,37,49,100,0,67,82,79,84,65,49,0,87,67,83,73,78,73,84,58,32,115,101,116,116,105,110,103,32,67,68,69,76,84,32,116,111,32,49,0,68,65,84,69,45,79,66,83,0,68,65,84,69,0,85,84,0,85,84,77,73,68,0,87,67,83,73,78,73,84,67,58,32,78,111,32,105,109,97,103,101,32,115,99,97,108,101,32,102,111,114,32,87,67,83,32,37,99,0,80,76,84,82,65,72,0,80,76,84,82,65,77,0,80,76,84,82,65,83,0,80,76,84,68,69,67,83,78,0,80,76,84,68,69,67,68,0,80,76,84,68,69,67,77,0,80,76,84,68,69,67,83,0,69,81,85,73,78,79,88,0,37,50,46,48,102,58,37,50,46,48,102,58,37,53,46,51,102,32,37,99,37,50,46,48,102,58,37,50,46,48,102,58,37,53,46,51,102,32,37,115,0,80,76,84,83,67,65,76,69,0,88,80,73,88,69,76,83,90,0,89,80,73,88,69,76,83,90,0,67,78,80,73,88,49,0,67,78,80,73,88,50,0,80,80,79,49,0,80,80,79,37,100,0,65,77,68,88,49,0,65,77,68,88,37,100,0,65,77,68,89,49,0,65,77,68,89,37,100,0,82,65,45,45,45,68,83,83,0,68,69,67,45,45,68,83,83,0,87,67,83,73,78,73,84,58,32,78,111,32,82,65,32,119,105,116,104,32,83,69,67,80,73,88,44,32,110,111,32,87,67,83,0,87,67,83,73,78,73,84,32,78,111,32,68,69,67,32,119,105,116,104,32,83,69,67,80,73,88,44,32,110,111,32,87,67,83,0,76,79,78,71,80,79,76,69,0,82,65,45,45,45,84,65,78,0,68,69,67,45,45,84,65,78,0,87,67,83,73,78,73,84,58,32,78,111,32,105,109,97,103,101,32,115,99,97,108,101,0,42,69,114,114,111,114,42,58,32,105,110,99,111,114,114,101,99,116,32,108,105,110,101,97,114,32,99,111,110,118,101,114,115,105,111,110,32,105,110,32,37,115,0,87,65,82,78,73,78,71,58,32,83,105,103,110,105,102,105,99,97,110,116,32,105,110,97,99,99,117,114,97,99,121,32,108,105,107,101,108,121,32,116,111,32,111,99,99,117,114,32,105,110,32,112,114,111,106,101,99,116,105,111,110,0,69,81,85,73,78,79,88,37,99,0,82,65,68,69,67,83,89,83,37,99,0,71,65,76,0,69,67,76,0,71,65,76,65,67,84,73,67,0,69,67,76,73,80,84,73,67,0,83,71,65,76,65,67,84,67,0,72,69,76,73,79,69,67,76,0,65,76,84,65,90,0,76,73,78,69,65,82,0,80,73,88,69,76,0,68,69,84,0,37,45,52,115,37,52,115,0,78,80,79,76,0,78,80,79,76,69,0,83,80,65,45,0,68,69,67,0,73,77,83,89,83,0,105,109,115,121,115,0,37,46,52,102,0,66,49,57,53,48,0,74,50,48,48,48,0,87,67,83,95,67,79,77,77,65,78,68,0,87,67,83,95,67,79,77,77,65,78,68,37,100,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,37,115,9,37,115,0,37,115,32,37,115,0,42,42,42,42,42,42,42,42,42,9,42,42,42,42,42,42,42,42,42,42,0,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,0,42,42,42,42,42,42,42,42,42,42,42,42,42,9,42,42,42,42,42,42,42,42,42,42,42,42,42,0,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,42,0,9,103,97,108,97,99,116,105,99,0,32,103,97,108,97,99,116,105,99,0,9,101,99,108,105,112,116,105,99,0,32,101,99,108,105,112,116,105,99,0,9,108,111,110,103,45,110,112,97,0,32,108,111,110,103,45,110,112,97,0,9,108,111,110,103,45,115,112,97,0,32,108,111,110,103,45,115,112,97,0,42,42,42,42,42,42,42,42,42,42,9,42,42,42,42,42,42,42,42,42,0,50,48,48,48,0,50,48,48,48,46,48,0,73,67,82,83,0,70,75,53,0,49,57,53,48,0,49,57,53,48,46,48,0,70,75,52,0,37,48,50,100,58,37,48,50,100,58,37,48,54,46,51,102,32,37,99,37,48,50,100,58,37,48,50,100,58,37,48,53,46,50,102,0,70,75,52,50,71,65,76,58,32,66,49,57,53,48,32,82,65,44,68,101,99,61,32,37,115,10,0,70,75,52,50,71,65,76,58,32,108,111,110,103,32,61,32,37,46,53,102,32,108,97,116,32,61,32,37,46,53,102,10,0,71,65,76,50,70,75,52,58,32,108,111,110,103,32,61,32,37,46,53,102,32,108,97,116,32,61,32,37,46,53,102,10,0,71,65,76,50,70,75,52,58,32,66,49,57,53,48,32,82,65,44,68,101,99,61,32,37,115,10,0,70,75,53,50,71,65,76,58,32,74,50,48,48,48,32,82,65,44,68,101,99,61,32,37,115,10,0,70,75,53,50,71,65,76,58,32,108,111,110,103,32,61,32,37,46,53,102,32,108,97,116,32,61,32,37,46,53,102,10,0,71,65,76,50,70,75,53,58,32,108,111,110,103,32,61,32,37,46,53,102,32,108,97,116,32,61,32,37,46,53,102,10,0,71,65,76,50,70,75,53,58,32,74,50,48,48,48,32,82,65,44,68,101,99,61,32,37,115,10,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,37,115,95,49,0,37,115,95,37,100,0,37,115,95,48,49,0,37,115,95,48,48,49,0,37,37,46,37,100,102,0,37,48,50,100,58,37,48,50,100,58,37,48,57,46,54,102,0,37,48,50,100,58,37,48,50,100,58,37,48,56,46,53,102,0,37,48,50,100,58,37,48,50,100,58,37,48,55,46,52,102,0,37,48,50,100,58,37,48,50,100,58,37,48,54,46,51,102,0,37,48,50,100,58,37,48,50,100,58,37,48,53,46,50,102,0,37,48,50,100,58,37,48,50,100,58,37,48,52,46,49,102,0,37,48,50,100,58,37,48,50,100,58,37,48,50,100,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,57,46,54,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,56,46,53,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,55,46,52,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,54,46,51,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,53,46,50,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,52,46,49,102,0,37,99,37,48,50,100,58,37,48,50,100,58,37,48,50,100,0,37,37,37,100,46,37,100,102,0,37,37,37,52,100,0,37,37,37,100,100],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+597217);allocate([84,78,88,95,71,83,69,86,65,76,58,32,117,110,107,110,111,119,110,32,115,117,114,102,97,99,101,32,116,121,112,101,10,0,84,78,88,95,71,83,68,69,82,58,32,111,114,100,101,114,32,111,102,32,100,101,114,105,118,97,116,105,118,101,115,32,99,97,110,110,111,116,32,98,101,32,60,32,48,10,0,84,78,88,95,71,83,68,69,82,58,32,117,110,107,110,111,119,110,32,115,117,114,102,97,99,101,32,116,121,112,101,32,37,100,10,0,119,102,95,103,115,114,101,115,116,111,114,101,58,32,105,108,108,101,103,97,108,32,120,32,111,114,100,101,114,32,37,100,10,0,119,102,95,103,115,114,101,115,116,111,114,101,58,32,105,108,108,101,103,97,108,32,121,32,111,114,100,101,114,32,37,100,10,0,119,102,95,103,115,114,101,115,116,111,114,101,58,32,105,108,108,101,103,97,108,32,120,32,114,97,110,103,101,32,37,102,45,37,102,10,0,119,102,95,103,115,114,101,115,116,111,114,101,58,32,105,108,108,101,103,97,108,32,121,32,114,97,110,103,101,32,37,102,45,37,102,10,0,119,102,95,103,115,114,101,115,116,111,114,101,58,32,117,110,107,110,111,119,110,32,115,117,114,102,97,99,101,32,116,121,112,101,32,37,100,10,0,87,65,84,49,0,87,65,84,49,95,48,48,49,61,32,39,119,116,121,112,101,61,122,112,120,32,97,120,116,121,112,101,61,114,97,32,112,114,111,106,112,48,61,48,46,32,112,114,111,106,112,49,61,49,46,32,112,114,111,106,112,50,61,48,46,32,112,114,111,106,112,51,61,51,51,55,46,55,52,32,112,114,111,106,39,87,65,84,50,95,48,48,49,61,32,39,119,116,121,112,101,61,122,112,120,32,97,120,116,121,112,101,61,100,101,99,32,112,114,111,106,112,48,61,48,46,32,112,114,111,106,112,49,61,49,46,32,112,114,111,106,112,50,61,48,46,32,112,114,111,106,112,51,61,51,51,55,46,55,52,32,112,114,111,39,0,87,65,84,50,0,108,111,110,103,112,111,108,101,0,114,111,0,112,114,111,106,112,37,100,0,108,110,103,99,111,114,0,108,97,116,99,111,114,0,10,62,32,37,115,37,115,10,10,0,78,111,116,32,101,110,111,117,103,104,32,109,101,109,111,114,121,32,102,111,114,32,0,112,111,108,121,32,40,49,32,101,108,101,109,101,110,116,115,41,32,33,0,84,104,101,32,100,105,109,101,110,115,105,111,110,97,108,105,116,121,32,111,102,32,116,104,101,32,112,111,108,121,110,111,109,32,40,37,100,41,32,101,120,99,101,101,100,115,32,116,104,101,32,109,97,120,105,109,117,109,10,97,108,108,111,119,101,100,32,111,110,101,32,40,37,100,41,0,42,69,114,114,111,114,42,58,32,0,112,111,108,121,45,62,103,114,111,117,112,32,40,112,111,108,121,45,62,110,100,105,109,32,101,108,101,109,101,110,116,115,41,32,33,0,112,111,108,121,45,62,100,101,103,114,101,101,32,40,112,111,108,121,45,62,110,103,114,111,117,112,32,101,108,101,109,101,110,116,115,41,32,33,0,42,69,114,114,111,114,42,58,32,112,111,108,121,110,111,109,105,97,108,32,71,82,79,85,80,32,111,117,116,32,111,102,32,114,97,110,103,101,0,84,104,101,32,100,101,103,114,101,101,32,111,102,32,116,104,101,32,112,111,108,121,110,111,109,32,40,37,100,41,32,101,120,99,101,101,100,115,32,116,104,101,32,109,97,120,105,109,117,109,10,97,108,108,111,119,101,100,32,111,110,101,32,40,37,100,41,0,112,111,108,121,45,62,98,97,115,105,115,32,40,112,111,108,121,45,62,110,99,111,101,102,102,32,101,108,101,109,101,110,116,115,41,32,33,0,112,111,108,121,45,62,99,111,101,102,102,32,40,112,111,108,121,45,62,110,99,111,101,102,102,32,101,108,101,109,101,110,116,115,41,32,33,0,42,73,110,116,101,114,110,97,108,32,69,114,114,111,114,42,58,32,79,110,101,32,111,102,32,120,32,111,114,32,101,120,116,98,97,115,105,115,32,115,104,111,117,108,100,32,98,101,32,100,105,102,102,101,114,101,110,116,32,102,114,111,109,32,78,85,76,76,10,105,110,32,0,112,111,108,121,95,102,117,110,99,40,41,0,97,108,112,104,97,32,40,109,97,116,115,105,122,101,32,101,108,101,109,101,110,116,115,41,32,33,0,98,101,116,97,32,40,110,99,111,101,102,102,32,101,108,101,109,101,110,116,115,41,32,33,0,118,109,97,116,32,40,110,42,110,32,101,108,101,109,101,110,116,115,41,32,33,0,119,109,97,116,32,40,110,32,101,108,101,109,101,110,116,115,41,32,33,0,112,32,40,110,32,101,108,101,109,101,110,116,115,41,32,33,0,42,69,114,114,111,114,42,58,32,78,111,116,32,101,110,111,117,103,104,32,114,111,119,115,32,102,111,114,32,115,111,108,118,105,110,103,32,116,104,101,32,115,121,115,116,101,109,32,0,105,110,32,115,118,100,102,105,116,40,41,0,114,118,49,32,40,110,32,101,108,101,109,101,110,116,115,41,32,33,0,116,109,112,32,40,110,32,101,108,101,109,101,110,116,115,41,32,33,0,42,69,114,114,111,114,42,58,32,78,111,32,99,111,110,118,101,114,103,101,110,99,101,32,105,110,32,49,48,48,32,83,86,68,32,105,116,101,114,97,116,105,111,110,115,32,0,67,85,66,69,70,65,67,69,0,37,46,51,115,0,82,65,45,45,0,82,65,0,68,69,67,45,45,37,115,0,82,65,45,45,45,37,115,0,76,79,78,0,37,99,76,79,78,0,37,99,76,65,84,0,37,115,45,37,115,0,76,78,0,37,99,37,99,76,78,0,37,99,37,99,76,84,0,76,84,0,71,76,83,0,78,67,80,0,65,90,80,0,83,90,80,0,84,65,78,0,83,84,71,0,83,73,78,0,65,82,67,0,90,80,78,0,90,69,65,0,65,73,82,0,67,89,80,0,67,69,65,0,67,65,82,0,77,69,82,0,67,79,80,0,67,79,69,0,67,79,68,0,67,79,79,0,83,70,76,0,80,65,82,0,77,79,76,0,65,73,84,0,66,79,78,0,80,67,79,0,84,83,67,0,67,83,67,0,81,83,67,0,65,90,80,0,83,90,80,0,84,65,78,0,83,84,71,0,83,73,78,0,65,82,67,0,90,80,78,0,90,69,65,0,65,73,82,0,67,89,80,0,67,69,65,0,67,65,82,0,77,69,82,0,83,70,76,0,80,65,82,0,77,79,76,0,65,73,84,0,67,79,80,0,67,79,69,0,67,79,68,0,67,79,79,0,66,79,78,0,80,67,79,0,84,83,67,0,67,83,67,0,81,83,67,0,65,95,79,82,68,69,82,0,68,73,83,84,73,78,73,84,58,32,77,105,115,115,105,110,103,32,65,95,79,82,68,69,82,32,107,101,121,119,111,114,100,32,102,111,114,32,83,112,105,116,122,101,114,32,100,105,115,116,111,114,116,105,111,110,0,65,95,37,100,95,37,100,0,66,95,79,82,68,69,82,0,68,73,83,84,73,78,73,84,58,32,77,105,115,115,105,110,103,32,66,95,79,82,68,69,82,32,107,101,121,119,111,114,100,32,102,111,114,32,83,112,105,116,122,101,114,32,100,105,115,116,111,114,116,105,111,110,0,66,95,37,100,95,37,100,0,65,80,95,79,82,68,69,82,0,68,73,83,84,73,78,73,84,58,32,77,105,115,115,105,110,103,32,65,80,95,79,82,68,69,82,32,107,101,121,119,111,114,100,32,102,111,114,32,83,112,105,116,122,101,114,32,100,105,115,116,111,114,116,105,111,110,0,65,80,95,37,100,95,37,100,0,66,80,95,79,82,68,69,82,0,68,73,83,84,73,78,73,84,58,32,77,105,115,115,105,110,103,32,66,80,95,79,82,68,69,82,32,107,101,121,119,111,114,100,32,102,111,114,32,83,112,105,116,122,101,114,32,100,105,115,116,111,114,116,105,111,110,0,66,80,95,37,100,95,37,100,0,45,83,73,80,0,58,32,0,105,110,116,101,114,110,97,108,32,101,114,114,111,114,58,32,105,110,102,108,97,116,101,32,115,116,114,101,97,109,32,99,111,114,114,117,112,116,0,99,111,109,112,114,101,115,115,101,100,32,100,97,116,97,32,101,114,114,111,114,0,114,101,113,117,101,115,116,101,100,32,108,101,110,103,116,104,32,100,111,101,115,32,110,111,116,32,102,105,116,32,105,110,32,105,110,116,0,105,110,116,101,114,110,97,108,32,101,114,114,111,114,58,32,100,101,102,108,97,116,101,32,115,116,114,101,97,109,32,99,111,114,114,117,112,116,0,111,117,116,32,111,102,32,109,101,109,111,114,121,0,105,110,99,111,114,114,101,99,116,32,104,101,97,100,101,114,32,99,104,101,99,107,0,117,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,105,111,110,32,109,101,116,104,111,100,0,105,110,118,97,108,105,100,32,119,105,110,100,111,119,32,115,105,122,101,0,117,110,107,110,111,119,110,32,104,101,97,100,101,114,32,102,108,97,103,115,32,115,101,116,0,104,101,97,100,101,114,32,99,114,99,32,109,105,115,109,97,116,99,104,0,105,110,118,97,108,105,100,32,98,108,111,99,107,32,116,121,112,101,0,105,110,118,97,108,105,100,32,115,116,111,114,101,100,32,98,108,111,99,107,32,108,101,110,103,116,104,115,0,116,111,111,32,109,97,110,121,32,108,101,110,103,116,104,32,111,114,32,100,105,115,116,97,110,99,101,32,115,121,109,98,111,108,115,0,105,110,118,97,108,105,100,32,99,111,100,101,32,108,101,110,103,116,104,115,32,115,101,116,0,105,110,118,97,108,105,100,32,98,105,116,32,108,101,110,103,116,104,32,114,101,112,101,97,116,0,105,110,118,97,108,105,100,32,99,111,100,101,32,45,45,32,109,105,115,115,105,110,103,32,101,110,100,45,111,102,45,98,108,111,99,107,0,105,110,118,97,108,105,100,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,115,32,115,101,116,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,115,32,115,101,116,0,105,110,118,97,108,105,100,32,108,105,116,101,114,97,108,47,108,101,110,103,116,104,32,99,111,100,101,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,99,111,100,101,0,105,110,118,97,108,105,100,32,100,105,115,116,97,110,99,101,32,116,111,111,32,102,97,114,32,98,97,99,107,0,105,110,99,111,114,114,101,99,116,32,100,97,116,97,32,99,104,101,99,107,0,105,110,99,111,114,114,101,99,116,32,108,101,110,103,116,104,32,99,104,101,99,107,0,0,1,2,3,4,4,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,0,0,16,17,18,18,19,19,20,20,20,20,21,21,21,21,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,29,0,1,2,3,4,5,6,7,8,8,9,9,10,10,11,11,12,12,12,12,13,13,13,13,14,14,14,14,15,15,15,15,16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,19,19,19,19,19,19,19,19,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,24,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15,110,101,101,100,32,100,105,99,116,105,111,110,97,114,121,0,115,116,114,101,97,109,32,101,110,100,0,102,105,108,101,32,101,114,114,111,114,0,115,116,114,101,97,109,32,101,114,114,111,114,0,100,97,116,97,32,101,114,114,111,114,0,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,98,117,102,102,101,114,32,101,114,114,111,114,0,105,110,99,111,109,112,97,116,105,98,108,101,32,118,101,114,115,105,111,110,0,49,46,50,46,53],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+605048);allocate([117,110,101,120,112,101,99,116,101,100,32,101,110,100,32,111,102,32,102,105,108,101,0,69,82,82,79,82,58,32,105,110,112,117,116,32,46,90,32,102,105,108,101,32,105,115,32,105,110,32,117,110,114,101,99,111,103,110,105,122,101,100,32,99,111,109,112,114,101,115,115,105,111,110,32,102,111,114,109,97,116,46,10,0,119,97,114,110,105,110,103,44,32,117,110,107,110,111,119,110,32,102,108,97,103,115,32,105,110,32,117,110,108,122,119,32,100,101,99,111,109,112,114,101,115,115,105,111,110,0,99,111,109,112,114,101,115,115,101,100,32,119,105,116,104,32,116,111,111,32,109,97,110,121,32,98,105,116,115,59,32,99,97,110,110,111,116,32,104,97,110,100,108,101,32,102,105,108,101],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+641373);allocate([99,111,114,114,117,112,116,32,105,110,112,117,116,46],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+707089);allocate([102,97,105,108,101,100,32,116,111,32,119,114,105,116,101,32,98,117,102,102,101,114,32,116,111,32,117,110,99,111,109,112,114,101,115,115,101,100,32,111,117,116,112,117,116,32,102,105,108,101,32,40,119,114,105,116,101,95,98,117,102,41,0,109,97,108,108,111,99,32,102,97,105,108,101,100,32,119,104,105,108,101,32,117,110,99,111,109,112,114,101,115,115,105,110,103,32,40,119,114,105,116,101,95,98,117,102,41,0,91,99,100,108,95,122,115,99,97,108,101,93,32,37,100,120,37,100,45,37,100,32,32,99,111,110,116,61,37,103,32,111,112,116,115,122,61,37,100,32,108,101,110,61,37,100,10,0,91,99,100,108,95,122,115,99,97,108,101,93,32,122,109,105,110,61,37,103,32,122,109,97,120,61,37,103,32,108,101,102,116,61,37,103,32,109,101,100,105,97,110,61,37,103,10,0,91,99,100,108,95,122,115,99,97,108,101,93,32,109,105,110,112,105,120,61,37,100,32,110,103,114,111,119,61,37,100,32,110,103,111,111,100,112,105,120,61,37,100,10,0,91,99,100,108,95,122,115,99,97,108,101,93,32,122,115,108,111,112,101,61,37,103,32,99,101,110,116,101,114,95,112,105,120,61,37,100,32,122,49,61,37,103,32,122,50,61,37,103,10,0,91,115,97,109,112,108,101,73,109,97,103,101,93,32,111,112,116,95,110,112,105,120,47,108,105,110,101,61,37,100,32,99,111,108,95,115,116,101,112,61,37,100,32,110,47,108,105,110,101,61,37,100,10,0,91,115,97,109,112,108,101,73,109,97,103,101,93,32,110,108,95,105,110,95,115,97,109,112,61,37,100,47,37,100,32,111,112,116,95,110,108,47,115,97,109,112,61,37,100,32,108,115,116,101,112,61,37,100,10,0,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,100,114,105,118,101,114,32,102,111,114,32,112,114,101,45,101,120,105,115,116,105,110,103,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,112,114,101,45,101,120,105,115,116,105,110,103,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,103,101,116,32,116,104,101,32,115,105,122,101,32,111,102,32,116,104,101,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,115,116,114,117,99,116,117,114,101,32,102,111,114,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,102,105,108,101,110,97,109,101,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,104,101,97,100,115,116,97,114,116,32,97,114,114,97,121,58,32,40,102,102,111,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,105,111,98,117,102,102,101,114,32,97,114,114,97,121,58,32,40,102,102,111,109,101,109,41,0,102,102,111,109,101,109,32,99,111,117,108,100,32,110,111,116,32,105,110,116,101,114,112,114,101,116,32,112,114,105,109,97,114,121,32,97,114,114,97,121,32,104,101,97,100,101,114,32,111,102,32,102,105,108,101,58,32,40,102,102,111,109,101,109,41,0,102,102,111,109,101,109,32,99,111,117,108,100,32,110,111,116,32,109,111,118,101,32,116,111,32,116,104,101,32,115,112,101,99,105,102,105,101,100,32,101,120,116,101,110,115,105,111,110,58,0,32,101,120,116,101,110,115,105,111,110,32,110,117,109,98,101,114,32,37,100,32,100,111,101,115,110,39,116,32,101,120,105,115,116,32,111,114,32,99,111,117,108,100,110,39,116,32,98,101,32,111,112,101,110,101,100,46,0,32,101,120,116,101,110,115,105,111,110,32,119,105,116,104,32,69,88,84,78,65,77,69,32,61,32,37,115,44,0,32,32,32,32,32,32,32,32,32,32,32,97,110,100,32,119,105,116,104,32,69,88,84,86,69,82,83,32,61,32,37,100,44,0,32,32,32,32,32,32,32,32,32,32,32,97,110,100,32,119,105,116,104,32,88,84,69,78,83,73,79,78,32,61,32,37,115,44,0,32,100,111,101,115,110,39,116,32,101,120,105,115,116,32,111,114,32,99,111,117,108,100,110,39,116,32,98,101,32,111,112,101,110,101,100,46,0,86,105,116,97,108,32,67,70,73,84,83,73,79,32,112,97,114,97,109,101,116,101,114,115,32,104,101,108,100,32,105,110,32,109,101,109,111,114,121,32,104,97,118,101,32,98,101,101,110,32,99,111,114,114,117,112,116,101,100,33,33,0,70,97,116,97,108,32,99,111,110,100,105,116,105,111,110,32,100,101,116,101,99,116,101,100,32,105,110,32,102,105,116,115,95,114,101,103,105,115,116,101,114,95,100,114,105,118,101,114,46,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,109,101,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,109,101,109,107,101,101,112,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,115,116,100,105,110,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,115,116,100,105,110,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,115,116,100,111,117,116,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,105,114,97,102,109,101,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,114,97,119,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,114,97,119,102,105,108,101,58,47,47,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,99,111,109,112,114,101,115,115,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,99,111,109,112,114,101,115,115,109,101,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,99,111,109,112,114,101,115,115,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,99,111,109,112,114,101,115,115,111,117,116,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,99,111,109,112,114,101,115,115,111,117,116,102,105,108,101,58,47,47,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,114,111,111,116,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,104,116,116,112,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,104,116,116,112,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,104,116,116,112,109,101,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,104,116,116,112,99,111,109,112,114,101,115,115,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,102,116,112,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,102,116,112,102,105,108,101,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,102,116,112,109,101,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,102,116,112,99,111,109,112,114,101,115,115,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,114,101,103,105,115,116,101,114,32,116,104,101,32,115,116,114,101,97,109,58,47,47,32,100,114,105,118,101,114,32,40,105,110,105,116,95,99,102,105,116,115,105,111,41,0,102,97,105,108,101,100,32,116,111,32,99,108,111,115,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,99,108,111,115,41,0,115,112,101,99,105,102,105,101,100,32,101,120,116,101,110,115,105,111,110,32,110,117,109,98,101,114,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,58,0,44,58,59,0,32,44,58,0,105,108,108,101,103,97,108,32,69,88,84,86,69,82,32,118,97,108,117,101,32,105,110,32,105,110,112,117,116,32,85,82,76,58,0,59,0,117,110,107,110,111,119,110,32,116,121,112,101,32,111,102,32,72,68,85,32,105,110,32,105,110,112,117,116,32,85,82,76,58,0,80,0,105,108,108,101,103,97,108,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,111,102,32,105,109,97,103,101,32,105,110,32,116,97,98,108,101,32,99,101,108,108,32,105,110,32,105,110,112,117,116,32,85,82,76,58,0,32,100,105,100,32,110,111,116,32,102,105,110,100,32,97,32,114,111,119,32,101,120,112,114,101,115,115,105,111,110,32,101,110,99,108,111,115,101,100,32,105,110,32,40,32,41,0,78,97,109,101,32,111,102,32,102,105,108,101,32,116,111,32,111,112,101,110,32,105,115,32,98,108,97,110,107,46,32,40,102,102,111,112,101,110,41,0,78,97,109,101,32,111,102,32,102,105,108,101,32,116,111,32,111,112,101,110,32,105,115,32,116,111,111,32,108,111,110,103,46,32,40,102,102,111,112,101,110,41,0,99,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,116,104,101,32,105,110,112,117,116,32,102,105,108,101,110,97,109,101,58,32,40,102,102,111,112,101,110,41,0,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,100,114,105,118,101,114,32,102,111,114,32,116,104,105,115,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,99,104,101,99,107,102,105,108,101,32,102,97,105,108,101,100,32,102,111,114,32,116,104,105,115,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,99,111,117,108,100,32,110,111,116,32,99,104,97,110,103,101,32,100,114,105,118,101,114,32,102,111,114,32,116,104,105,115,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,102,105,110,100,32,111,114,32,111,112,101,110,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,99,97,110,110,111,116,32,111,112,101,110,32,97,110,32,101,120,105,115,116,105,110,103,32,102,105,108,101,32,111,102,32,116,104,105,115,32,116,121,112,101,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,103,101,116,32,116,104,101,32,115,105,122,101,32,111,102,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,115,116,114,117,99,116,117,114,101,32,102,111,114,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,102,105,108,101,110,97,109,101,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,104,101,97,100,115,116,97,114,116,32,97,114,114,97,121,58,32,40,102,102,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,105,111,98,117,102,102,101,114,32,97,114,114,97,121,58,32,40,102,102,111,112,101,110,41,0,102,102,111,112,101,110,32,99,111,117,108,100,32,110,111,116,32,105,110,116,101,114,112,114,101,116,32,112,114,105,109,97,114,121,32,97,114,114,97,121,32,104,101,97,100,101,114,32,111,102,32,102,105,108,101,58,32,0,102,102,111,112,101,110,32,99,111,117,108,100,32,110,111,116,32,109,111,118,101,32,116,111,32,116,104,101,32,115,112,101,99,105,102,105,101,100,32,101,120,116,101,110,115,105,111,110,58,0,103,116,105,0,79,66,83,84,65,66,76,69,0,105,108,108,101,103,97,108,32,114,111,119,110,117,109,32,102,111,114,32,105,109,97,103,101,32,99,101,108,108,58,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,105,109,97,103,101,32,105,110,32,97,32,116,97,98,108,101,32,99,101,108,108,58,0,70,97,105,108,101,100,32,116,111,32,102,105,110,100,32,114,111,119,32,109,97,116,99,104,105,110,103,32,116,104,105,115,32,101,120,112,114,101,115,115,105,111,110,58,0,114,111,119,32,115,116,97,116,105,115,102,121,105,110,103,32,116,104,105,115,32,101,120,112,114,101,115,115,105,111,110,32,100,111,101,115,110,39,116,32,101,120,105,115,116,58,58,0,109,101,109,58,47,47,95,49,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,102,105,108,101,32,102,111,114,32,99,111,112,121,32,111,102,32,105,109,97,103,101,32,105,110,32,116,97,98,108,101,32,99,101,108,108,58,0,70,97,105,108,101,100,32,116,111,32,99,111,112,121,32,116,97,98,108,101,32,99,101,108,108,32,116,111,32,110,101,119,32,112,114,105,109,97,114,121,32,97,114,114,97,121,58,0,101,100,105,116,105,110,103,32,99,111,108,117,109,110,115,32,105,110,32,105,110,112,117,116,32,116,97,98,108,101,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,32,119,104,105,108,101,32,116,114,121,105,110,103,32,116,111,32,112,101,114,102,111,114,109,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,111,112,101,114,97,116,105,111,110,58,0,109,101,109,58,47,47,95,50,0,111,110,45,116,104,101,45,102,108,121,32,115,101,108,101,99,116,105,111,110,32,111,102,32,105,109,97,103,101,32,115,101,99,116,105,111,110,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,32,119,104,105,108,101,32,116,114,121,105,110,103,32,116,111,32,117,115,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,115,101,99,116,105,111,110,32,102,105,108,116,101,114,58,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,115,101,108,101,99,116,101,100,32,99,111,108,117,109,110,115,32,97,114,114,97,121,32,40,102,102,111,112,101,110,41,0,32,119,104,105,108,101,32,116,114,121,105,110,103,32,116,111,32,115,101,108,101,99,116,32,114,111,119,115,32,119,105,116,104,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,116,101,114,58,0,115,101,108,101,99,116,105,111,110,32,111,102,32,114,111,119,115,32,105,110,32,105,110,112,117,116,32,116,97,98,108,101,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,111,110,45,116,104,101,45,102,108,121,32,115,101,108,101,99,116,105,111,110,32,111,102,32,114,111,119,115,32,105,110,32,105,110,112,117,116,32,116,97,98,108,101,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,67,70,73,84,83,73,79,32,117,115,101,100,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,116,101,114,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,99,114,101,97,116,101,32,116,104,105,115,32,116,97,98,108,101,58,0,109,101,109,58,47,47,95,51,0,111,110,45,116,104,101,45,102,108,121,32,104,105,115,116,111,103,114,97,109,109,105,110,103,32,111,102,32,105,110,112,117,116,32,116,97,98,108,101,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,32,119,104,105,108,101,32,116,114,121,105,110,103,32,116,111,32,101,120,101,99,117,116,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,104,105,115,116,111,103,114,97,109,32,115,112,101,99,105,102,105,99,97,116,105,111,110,58,0,67,70,73,84,83,73,79,32,117,115,101,100,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,99,114,101,97,116,101,32,116,104,105,115,32,104,105,115,116,111,103,114,97,109,58,0,109,101,109,58,47,47,95,52,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,102,111,114,32,112,105,120,101,108,32,102,105,108,116,101,114,58,0,112,105,120,101,108,32,102,105,108,116,101,114,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,110,111,116,32,115,112,97,99,101,32,115,101,112,97,114,97,116,101,100,58,0,102,97,105,108,101,100,32,116,111,32,101,120,101,99,117,116,101,32,105,109,97,103,101,32,102,105,108,116,101,114,58,0,112,105,120,101,108,32,102,105,108,116,101,114,105,110,103,32,111,102,32,105,110,112,117,116,32,105,109,97,103,101,32,102,97,105,108,101,100,32,40,102,102,111,112,101,110,41,0,32,119,104,105,108,101,32,116,114,121,105,110,103,32,116,111,32,101,120,101,99,117,116,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,58,0,67,70,73,84,83,73,79,32,117,115,101,100,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,99,114,101,97,116,101,32,116,104,105,115,32,105,109,97,103,101,58,0,99,97,110,110,111,116,32,117,115,101,32,112,105,120,101,108,32,102,105,108,116,101,114,32,111,110,32,110,111,110,45,73,77,65,71,69,32,72,68,85,0,10,69,82,82,79,82,58,32,77,105,115,109,97,116,99,104,32,105,110,32,116,104,101,32,67,70,73,84,83,73,79,95,83,79,78,65,77,69,32,118,97,108,117,101,32,105,110,32,116,104,101,32,102,105,116,115,105,111,46,104,32,105,110,99,108,117,100,101,32,102,105,108,101,0,116,104,97,116,32,119,97,115,32,117,115,101,100,32,116,111,32,98,117,105,108,100,32,116,104,101,32,67,70,73,84,83,73,79,32,108,105,98,114,97,114,121,44,32,97,110,100,32,116,104,101,32,118,97,108,117,101,32,105,110,32,116,104,101,32,105,110,99,108,117,100,101,32,102,105,108,101,0,116,104,97,116,32,119,97,115,32,117,115,101,100,32,119,104,101,110,32,99,111,109,112,105,108,105,110,103,32,116,104,101,32,97,112,112,108,105,99,97,116,105,111,110,32,112,114,111,103,114,97,109,58,0,32,32,32,86,101,114,115,105,111,110,32,117,115,101,100,32,116,111,32,98,117,105,108,100,32,116,104,101,32,67,70,73,84,83,73,79,32,108,105,98,114,97,114,121,32,32,32,61,32,37,100,10,0,32,32,32,86,101,114,115,105,111,110,32,105,110,99,108,117,100,101,100,32,98,121,32,116,104,101,32,97,112,112,108,105,99,97,116,105,111,110,32,112,114,111,103,114,97,109,32,61,32,37,100,10,0,10,70,105,120,32,116,104,105,115,32,98,121,32,114,101,99,111,109,112,105,108,105,110,103,32,97,110,100,32,116,104,101,110,32,114,101,108,105,110,107,105,110,103,32,116,104,105,115,32,97,112,112,108,105,99,97,116,105,111,110,32,112,114,111,103,114,97,109,32,0,119,105,116,104,32,116,104,101,32,67,70,73,84,83,73,79,32,108,105,98,114,97,114,121,46,0,115,116,100,105,110,0,103,115,105,102,116,112,58,0,103,115,105,102,116,112,58,47,47,0,115,104,109,101,109,58,0,102,105,108,101,58,0,78,97,109,101,32,111,102,32,102,105,108,101,32,105,115,32,116,111,111,32,108,111,110,103,46,0,58,91,0,46,105,109,104,0,105,110,112,117,116,32,102,105,108,101,32,85,82,76,32,105,115,32,109,105,115,115,105,110,103,32,99,108,111,115,105,110,103,32,98,114,97,99,107,101,116,32,39,93,39,0,114,97,119,115,116,100,105,110,58,47,47,0,99,111,108,32,0,98,105,110,0,112,105,120,0,99,111,109,112,114,101,115,115,32,0,99,111,109,112,114,101,115,115,93,0,103,116,105,102,105,108,116,101,114,40,0,114,101,103,102,105,108,116,101,114,40,0,91,98,105,110,0,91,66,73,78,0,91,66,105,110,0,91,99,111,108,32,0,91,67,79,76,32,0,91,67,111,108,32,0,108,105,116,101,114,97,108,32,115,116,114,105,110,103,32,105,110,32,105,110,112,117,116,32,102,105,108,101,32,85,82,76,32,105,115,32,109,105,115,115,105,110,103,32,99,108,111,115,105,110,103,32,115,105,110,103,108,101,32,113,117,111,116,101,0,110,101,115,116,101,100,32,98,114,97,99,107,101,116,115,32,105,110,32,105,110,112,117,116,32,102,105,108,101,32,85,82,76,32,105,115,32,109,105,115,115,105,110,103,32,99,108,111,115,105,110,103,32,98,114,97,99,107,101,116,0,91,112,105,120,0,91,80,73,88,0,91,80,105,120,0,91,99,111,109,112,114,101,115,115,0,105,110,112,117,116,32,102,105,108,101,32,85,82,76,32,108,97,99,107,115,32,118,97,108,105,100,32,114,111,119,32,102,105,108,116,101,114,32,101,120,112,114,101,115,115,105,111,110,0,70,73,76,69,58,47,47,0,70,105,108,101,32,110,97,109,101,32,105,115,32,116,111,111,32,108,111,110,103,46,32,40,102,105,116,115,95,97,108,114,101,97,100,121,95,111,112,101,110,41,0,99,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,116,104,101,32,112,114,101,118,105,111,117,115,108,121,32,111,112,101,110,101,100,32,102,105,108,101,110,97,109,101,58,32,40,102,102,111,112,101,110,41,0,99,97,110,110,111,116,32,114,101,111,112,101,110,32,102,105,108,101,32,82,69,65,68,87,82,73,84,69,32,119,104,101,110,32,112,114,101,118,105,111,117,115,108,121,32,111,112,101,110,101,100,32,82,69,65,68,79,78,76,89,0,78,97,109,101,32,111,102,32,102,105,108,101,32,116,111,32,99,114,101,97,116,101,32,105,115,32,98,108,97,110,107,46,32,40,102,102,105,110,105,116,41,0,70,105,108,101,110,97,109,101,32,105,115,32,116,111,111,32,108,111,110,103,46,32,40,102,102,105,110,105,116,41,0,99,111,117,108,100,32,110,111,116,32,112,97,114,115,101,32,116,104,101,32,111,117,116,112,117,116,32,102,105,108,101,110,97,109,101,58,32,40,102,102,105,110,105,116,41,0,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,100,114,105,118,101,114,32,102,111,114,32,116,104,105,115,32,102,105,108,101,58,32,40,102,102,105,110,105,116,41,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,110,101,119,32,102,105,108,101,32,40,97,108,114,101,97,100,121,32,101,120,105,115,116,115,63,41,58,0,99,97,110,110,111,116,32,99,114,101,97,116,101,32,97,32,110,101,119,32,102,105,108,101,32,111,102,32,116,104,105,115,32,116,121,112,101,58,32,40,102,102,105,110,105,116,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,102,105,108,101,110,97,109,101,58,32,40,102,102,105,110,105,116,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,104,101,97,100,115,116,97,114,116,32,97,114,114,97,121,58,32,40,102,102,105,110,105,116,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,105,111,98,117,102,102,101,114,32,97,114,114,97,121,58,32,40,102,102,105,110,105,116,41,0,99,111,108,117,109,110,32,99,111,110,116,97,105,110,105,110,103,32,105,109,97,103,101,32,105,110,32,116,97,98,108,101,32,99,101,108,108,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,58,0,84,104,105,115,32,101,120,116,101,110,115,105,111,110,32,105,115,32,110,111,116,32,97,32,98,105,110,97,114,121,32,116,97,98,108,101,46,0,32,67,97,110,110,111,116,32,111,112,101,110,32,116,104,101,32,105,109,97,103,101,32,105,110,32,97,32,98,105,110,97,114,121,32,116,97,98,108,101,32,99,101,108,108,46,0,69,114,114,111,114,32,103,101,116,116,105,110,103,32,116,104,101,32,100,105,109,101,110,115,105,111,110,115,32,111,102,32,116,104,101,32,105,109,97,103,101,0,69,114,114,111,114,58,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,105,109,97,103,101,32,99,111,108,117,109,110,32,104,97,115,32,105,110,118,97,108,105,100,32,100,97,116,97,116,121,112,101,58,0,67,97,110,110,111,116,32,111,112,101,110,32,97,110,32,105,109,97,103,101,32,105,110,32,97,32,115,105,110,103,108,101,32,114,111,119,32,111,102,32,116,104,105,115,32,99,111,108,117,109,110,46,0,102,97,105,108,101,100,32,116,111,32,119,114,105,116,101,32,114,101,113,117,105,114,101,100,32,112,114,105,109,97,114,121,32,97,114,114,97,121,32,107,101,121,119,111,114,100,115,32,105,110,32,116,104,101,32,111,117,116,112,117,116,32,102,105,108,101,0,72,73,83,84,79,82,89,32,32,84,104,105,115,32,105,109,97,103,101,32,119,97,115,32,99,111,112,105,101,100,32,102,114,111,109,32,114,111,119,32,37,108,100,32,111,102,32,99,111,108,117,109,110,32,39,37,115,39,44,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,100,97,116,97,32,98,117,102,102,101,114,32,102,114,111,109,32,102,105,108,101,58,0,109,101,109,0,99,111,109,112,114,101,115,115,0,104,116,116,112,0,102,116,112,0,103,115,105,102,116,112,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,102,105,108,101,32,102,111,114,32,99,111,112,121,32,40,102,102,101,100,105,116,95,99,111,108,117,109,110,115,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,97,108,108,32,72,68,85,115,32,102,114,111,109,32,105,110,112,117,116,32,102,105,108,101,32,40,102,102,101,100,105,116,95,99,111,108,117,109,110,115,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,116,104,101,32,105,110,112,117,116,32,102,105,108,101,32,40,102,102,101,100,105,116,95,99,111,108,117,109,110,115,41,0,112,97,114,115,105,110,103,32,101,114,114,111,114,32,105,110,32,99,111,108,117,109,110,32,102,105,108,116,101,114,32,101,120,112,114,101,115,115,105,111,110,0,67,111,117,108,100,110,39,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,116,111,32,104,111,108,100,32,116,111,107,101,110,32,115,116,114,105,110,103,32,40,102,105,116,115,95,103,101,116,95,116,111,107,101,110,50,41,46,0,102,97,105,108,101,100,32,116,111,32,100,101,108,101,116,101,32,99,111,108,117,109,110,32,105,110,32,105,110,112,117,116,32,102,105,108,101,58,0,99,111,108,117,109,110,32,111,114,32,107,101,121,119,111,114,100,32,116,111,32,98,101,32,100,101,108,101,116,101,100,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,58,0,40,32,61,0,101,114,114,111,114,58,32,99,111,108,117,109,110,32,111,114,32,107,101,121,119,111,114,100,32,110,97,109,101,32,105,115,32,98,108,97,110,107,58,0,84,104,101,32,107,101,121,119,111,114,100,32,110,97,109,101,58,0,105,115,32,105,110,118,97,108,105,100,32,117,110,108,101,115,115,32,97,32,99,111,108,117,109,110,32,104,97,115,32,98,101,101,110,32,112,114,101,118,105,111,117,115,108,121,0,99,114,101,97,116,101,100,32,111,114,32,101,100,105,116,116,101,100,32,98,121,32,97,32,99,97,108,99,117,108,97,116,111,114,32,99,111,109,109,97,110,100,0,83,121,110,116,97,120,32,101,114,114,111,114,32,105,110,32,99,111,108,117,109,110,115,32,115,112,101,99,105,102,105,101,114,32,105,110,32,105,110,112,117,116,32,85,82,76,58,0,102,97,105,108,101,100,32,116,111,32,114,101,110,97,109,101,32,99,111,108,117,109,110,32,105,110,32,105,110,112,117,116,32,102,105,108,101,0,32,111,108,100,110,97,109,101,32,61,0,32,110,101,119,110,97,109,101,32,61,0,99,111,108,117,109,110,32,111,114,32,107,101,121,119,111,114,100,32,116,111,32,98,101,32,114,101,110,97,109,101,100,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,58,0,40,0,85,110,97,98,108,101,32,116,111,32,99,97,108,99,117,108,97,116,101,32,101,120,112,114,101,115,115,105,111,110,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,102,111,114,32,105,109,97,103,101,32,115,101,99,116,105,111,110,58,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,102,105,108,101,32,102,111,114,32,115,101,108,101,99,116,101,100,32,114,111,119,115,32,102,114,111,109,32,105,110,112,117,116,32,116,97,98,108,101,0,67,79,77,80,82,69,83,83,0,67,111,117,108,100,110,39,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,116,111,32,104,111,108,100,32,65,83,67,73,73,32,102,105,108,101,32,99,111,110,116,101,110,116,115,46,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,65,83,67,73,73,32,102,105,108,101,32,37,115,46,0,73,110,112,117,116,32,105,109,97,103,101,32,101,105,116,104,101,114,32,104,97,100,32,78,65,88,73,83,32,61,32,48,32,40,78,85,76,76,32,105,109,97,103,101,41,32,111,114,32,104,97,115,32,62,32,52,32,100,105,109,101,110,115,105,111,110,115,0,101,114,114,111,114,32,99,111,112,121,105,110,103,32,104,101,97,100,101,114,32,102,114,111,109,32,105,110,112,117,116,32,105,109,97,103,101,32,116,111,32,111,117,116,112,117,116,32,105,109,97,103,101,0,101,114,114,111,114,32,112,97,114,115,105,110,103,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,105,109,97,103,101,32,115,101,99,116,105,111,110,32,115,112,101,99,105,102,105,101,114,58,0,105,109,97,103,101,32,115,101,99,116,105,111,110,32,101,120,99,101,101,100,115,32,100,105,109,101,110,115,105,111,110,115,32,111,102,32,105,110,112,117,116,32,105,109,97,103,101,58,0,67,68,49,95,0,102,105,116,115,95,99,111,112,121,95,105,109,97,103,101,95,115,101,99,116,105,111,110,58,32,110,111,32,109,101,109,111,114,121,32,102,111,114,32,105,109,97,103,101,32,115,101,99,116,105,111,110,0,102,105,116,115,95,99,111,112,121,95,105,109,97,103,101,95,115,101,99,116,105,111,110,58,32,101,114,114,111,114,32,99,111,112,121,105,110,103,32,105,109,97,103,101,32,115,101,99,116,105,111,110,0,32,44,0,115,116,100,111,117,116,0,83,84,68,79,85,84,0,32,32,32,32,48,0,80,67,79,85,78,84,32,32,61,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,48,0,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,100,114,105,118,101,114,32,102,111,114,32,112,114,101,45,101,120,105,115,116,105,110,103,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,105,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,112,114,101,45,101,120,105,115,116,105,110,103,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,105,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,115,116,114,117,99,116,117,114,101,32,102,111,114,32,109,101,109,111,114,121,32,102,105,108,101,58,32,40,102,102,105,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,102,105,108,101,110,97,109,101,58,32,40,102,102,105,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,104,101,97,100,115,116,97,114,116,32,97,114,114,97,121,58,32,40,102,102,105,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,105,111,98,117,102,102,101,114,32,97,114,114,97,121,58,32,40,102,102,105,109,101,109,41,0,83,84,68,73,78,0,102,97,105,108,101,100,32,116,111,32,99,108,111,115,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,100,101,108,116,41,0,102,97,105,108,101,100,32,116,111,32,100,101,108,101,116,101,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,102,105,108,101,58,32,40,102,102,100,101,108,116,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,100,97,116,97,32,98,117,102,102,101,114,32,116,111,32,102,105,108,101,58,0,10,70,73,84,83,73,79,32,115,116,97,116,117,115,32,61,32,37,100,58,32,37,115,10,0,84,83,67,65,76,110,0,84,90,69,82,79,110,0,84,85,78,73,84,110,0,84,78,85,76,76,110,0,68,65,84,65,77,73,78,0,84,68,77,73,78,110,0,68,65,84,65,77,65,88,0,84,68,77,65,88,110,0,67,84,89,80,69,105,0,105,67,84,89,80,110,0,67,84,89,80,69,105,97,0,105,67,84,89,110,97,0,67,85,78,73,84,105,0,105,67,85,78,73,110,0,67,85,78,73,84,105,97,0,105,67,85,78,110,97,0,67,82,86,65,76,105,0,105,67,82,86,76,110,0,67,82,86,65,76,105,97,0,105,67,82,86,110,97,0,67,68,69,76,84,105,0,105,67,68,76,84,110,0,67,68,69,76,84,105,97,0,105,67,68,69,110,97,0,80,67,105,95,106,97,0,105,106,80,67,110,97,0,67,68,105,95,106,97,0,105,106,67,68,110,97,0,80,86,105,95,109,97,0,105,86,110,95,109,97,0,80,83,105,95,109,97,0,105,83,110,95,109,97,0,87,67,83,65,88,69,83,97,0,87,67,65,88,110,97,0,87,67,83,78,110,97,0,67,82,68,69,82,105,97,0,105,67,82,68,110,97,0,67,83,89,69,82,105,97,0,105,67,83,89,110,97,0,67,82,79,84,65,105,0,105,67,82,79,84,110,0,67,78,65,77,69,105,97,0,105,67,78,65,110,97,0,105,67,82,80,88,110,0,67,82,80,73,88,105,0,105,67,82,80,110,97,0,67,82,80,73,88,105,97,0,105,67,84,89,80,109,0,105,67,85,78,73,109,0,105,67,82,86,76,109,0,105,67,68,76,84,109,0,105,67,82,80,88,109,0,105,67,84,89,109,97,0,105,67,85,78,109,97,0,105,67,82,86,109,97,0,105,67,68,69,109,97,0,105,67,82,80,109,97,0,105,106,80,67,109,97,0,105,106,67,68,109,97,0,105,86,109,95,109,97,0,105,83,109,95,109,97,0,105,67,82,68,109,97,0,105,67,83,89,109,97,0,105,67,82,79,84,109,0,87,67,65,88,109,97,0,87,67,83,78,109,97,0,76,79,78,80,109,97,0,76,65,84,80,109,97,0,69,81,85,73,109,97,0,77,74,68,79,66,109,0,77,74,68,65,109,0,82,65,68,69,109,97,0,105,67,78,65,109,97,0,68,65,86,71,109],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+725536);allocate([85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,102,111,114,32,99,111,112,121,32,111,102,32,105,110,112,117,116,32,102,105,108,101,58,0,72,79,77,69,0,72,69,82,65,95,68,65,84,65,95,68,73,82,69,67,84,79,82,89,0,105,110,118,97,108,105,100,32,67,87,68,58,32,100,111,101,115,32,110,111,116,32,109,97,116,99,104,32,114,111,111,116,32,100,97,116,97,32,100,105,114,101,99,116,111,114,121,0,105,110,118,97,108,105,100,32,67,87,68,58,32,110,111,116,32,101,113,117,97,108,32,116,111,32,114,111,111,116,32,100,97,116,97,32,100,105,114,101,99,116,111,114,121,32,43,32,117,115,101,114,110,97,109,101,0,105,110,118,97,108,105,100,32,102,105,108,101,110,97,109,101,58,32,112,97,116,104,32,110,111,116,32,119,105,116,104,105,110,32,117,115,101,114,32,100,105,114,101,99,116,111,114,121,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,99,111,109,112,114,101,115,115,101,100,32,100,105,115,107,32,102,105,108,101,32,40,102,105,108,101,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,117,110,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,32,97,108,114,101,97,100,121,32,101,120,105,115,116,115,58,32,40,102,105,108,101,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,119,43,98,0,99,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,117,110,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,58,32,40,102,105,108,101,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,101,114,114,111,114,32,105,110,32,102,105,108,101,95,99,111,109,112,114,101,115,115,95,111,112,101,110,58,32,102,97,105,108,101,100,32,116,111,32,117,110,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,58,0,32,105,110,116,111,32,110,101,119,32,111,117,116,112,117,116,32,102,105,108,101,58,0,46,122,0,46,122,105,112,0,45,122,0,99,111,109,112,114,101,115,115,109,101,109,58,47,47,0,109,97,108,108,111,99,32,111,102,32,105,110,105,116,105,97,108,32,109,101,109,111,114,121,32,102,97,105,108,101,100,32,40,109,101,109,95,99,114,101,97,116,101,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,109,101,109,95,99,114,101,97,116,101,41,0,45,46,103,122,0,115,116,100,111,117,116,46,103,122,0,83,84,68,79,85,84,46,103,122,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,109,101,109,95,99,114,101,97,116,101,95,99,111,109,112,41,0,70,97,105,108,101,100,32,116,111,32,114,101,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,40,109,101,109,95,116,114,117,110,99,97,116,101,41],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+735569);allocate([85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,116,111,32,99,111,112,121,32,115,116,100,105,110,32,40,115,116,100,105,110,95,111,112,101,110,41,58,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,115,116,100,105,110,32,116,111,32,102,105,108,101,32,40,115,116,100,105,110,95,111,112,101,110,41,0,99,97,110,110,111,116,32,111,112,101,110,32,115,116,100,105,110,32,119,105,116,104,32,87,82,73,84,69,32,97,99,99,101,115,115,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,115,116,100,105,110,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,115,116,100,105,110,32,105,110,116,111,32,109,101,109,111,114,121,32,40,115,116,100,105,110,95,111,112,101,110,41,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,116,104,101,32,115,116,114,105,110,103,32,39,83,73,77,80,76,69,39,32,105,110,32,116,104,101,32,115,116,100,105,110,32,115,116,114,101,97,109,0,99,97,110,110,111,116,32,111,112,101,110,32,99,111,109,112,114,101,115,115,101,100,32,105,110,112,117,116,32,115,116,114,101,97,109,32,119,105,116,104,32,87,82,73,84,69,32,97,99,99,101,115,115,32,40,109,101,109,95,99,111,109,112,114,101,115,115,95,115,116,100,105,110,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,99,111,109,112,114,101,115,115,95,115,116,100,105,110,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,117,110,99,111,109,112,114,101,115,115,32,115,116,100,105,110,32,105,110,116,111,32,109,101,109,111,114,121,32,40,99,111,109,112,114,101,115,115,95,115,116,100,105,110,95,111,112,101,110,41,0,70,97,105,108,101,100,32,116,111,32,114,101,100,117,99,101,32,115,105,122,101,32,111,102,32,97,108,108,111,99,97,116,101,100,32,109,101,109,111,114,121,32,40,99,111,109,112,114,101,115,115,95,115,116,100,105,110,95,111,112,101,110,41,0,67,111,117,108,100,110,39,116,32,102,105,110,100,32,116,104,101,32,115,116,114,105,110,103,32,39,83,73,77,80,76,69,39,32,105,110,32,116,104,101,32,115,116,100,105,110,32,115,116,114,101,97,109,46,0,84,104,105,115,32,100,111,101,115,32,110,111,116,32,108,111,111,107,32,108,105,107,101,32,97,32,70,73,84,83,32,102,105,108,101,46,0,114,101,97,108,108,111,99,32,102,97,105,108,101,100,32,119,104,105,108,101,32,99,111,112,121,105,110,103,32,115,116,100,105,110,32,40,115,116,100,105,110,50,109,101,109,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,109,101,109,111,114,121,32,102,105,108,101,32,116,111,32,115,116,100,111,117,116,32,40,115,116,100,111,117,116,95,99,108,111,115,101,41,0,99,97,110,110,111,116,32,111,112,101,110,32,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,32,119,105,116,104,32,87,82,73,84,69,32,97,99,99,101,115,115,32,40,109,101,109,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,99,111,109,112,114,101,115,115,101,100,32,100,105,115,107,32,102,105,108,101,32,40,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,31,139,0,80,75,0,31,30,0,31,157,0,31,160,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,117,110,99,111,109,112,114,101,115,115,32,102,105,108,101,32,105,110,116,111,32,109,101,109,111,114,121,32,40,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,70,97,105,108,101,100,32,116,111,32,114,101,100,117,99,101,32,115,105,122,101,32,111,102,32,97,108,108,111,99,97,116,101,100,32,109,101,109,111,114,121,32,40,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,101,109,112,116,121,32,109,101,109,111,114,121,32,102,105,108,101,32,40,109,101,109,95,105,114,97,102,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,111,110,118,101,114,116,32,73,82,65,70,32,102,105,108,101,32,105,110,116,111,32,109,101,109,111,114,121,32,40,109,101,109,95,105,114,97,102,95,111,112,101,110,41,0,99,97,110,110,111,116,32,111,112,101,110,32,114,97,119,32,98,105,110,97,114,121,32,102,105,108,101,32,119,105,116,104,32,87,82,73,84,69,32,97,99,99,101,115,115,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,98,105,110,97,114,121,32,102,105,108,101,32,110,97,109,101,32,109,105,115,115,105,110,103,32,39,91,39,32,99,104,97,114,97,99,116,101,114,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,101,114,114,111,114,32,105,110,32,114,97,119,32,98,105,110,97,114,121,32,102,105,108,101,32,100,97,116,97,116,121,112,101,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,111,112,101,110,32,114,97,119,32,32,98,105,110,97,114,121,32,102,105,108,101,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,109,101,109,111,114,121,32,102,105,108,101,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,119,114,105,116,101,32,98,97,115,105,99,32,105,109,97,103,101,32,104,101,97,100,101,114,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,114,97,119,32,102,105,108,101,32,100,97,116,97,32,105,110,116,111,32,109,101,109,111,114,121,32,40,109,101,109,95,114,97,119,102,105,108,101,95,111,112,101,110,41,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,109,101,109,111,114,121,32,102,105,108,101,32,116,111,32,102,105,108,101,32,40,109,101,109,95,99,108,111,115,101,95,99,111,109,112,41,0,114,101,97,108,108,111,99,32,102,117,110,99,116,105,111,110,32,110,111,116,32,100,101,102,105,110,101,100,32,40,109,101,109,95,119,114,105,116,101,41,0,70,97,105,108,101,100,32,116,111,32,114,101,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,40,109,101,109,95,119,114,105,116,101,41,0,67,97,110,39,116,32,111,112,101,110,32,104,116,116,112,58,47,47,32,116,121,112,101,32,102,105,108,101,32,119,105,116,104,32,82,69,65,68,87,82,73,84,69,32,97,99,99,101,115,115,0,32,32,83,112,101,99,105,102,121,32,97,110,32,111,117,116,102,105,108,101,32,102,111,114,32,114,47,119,32,97,99,99,101,115,115,32,40,104,116,116,112,95,111,112,101,110,41,0,84,105,109,101,111,117,116,32,40,104,116,116,112,95,111,112,101,110,41,0,46,90,0,46,103,122,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,104,116,116,112,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,41,58,0,104,116,116,112,32,102,105,108,101,32,110,97,109,101,32,105,115,32,116,111,111,32,108,111,110,103,32,40,104,116,116,112,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,104,116,116,112,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,109,101,109,111,114,121,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,41,0,120,45,103,122,105,112,0,120,45,99,111,109,112,114,101,115,115,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,99,111,109,112,114,101,115,115,101,100,32,109,101,109,111,114,121,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,41,0,67,111,110,116,101,110,116,45,76,101,110,103,116,104,32,110,111,116,32,97,32,109,117,108,116,105,112,108,101,32,111,102,32,50,56,56,48,32,40,104,116,116,112,95,111,112,101,110,41,32,37,100,0,69,114,114,111,114,32,99,111,112,121,105,110,103,32,104,116,116,112,32,102,105,108,101,32,105,110,116,111,32,109,101,109,111,114,121,32,40,104,116,116,112,95,111,112,101,110,41],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+737323);allocate([79,117,116,112,117,116,32,102,105,108,101,32,110,111,116,32,115,101,116,44,32,115,104,111,117,108,100,110,39,116,32,104,97,118,101,32,104,97,112,112,101,110,101,100,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,67,97,110,39,116,32,111,112,101,110,32,99,111,109,112,114,101,115,115,101,100,32,104,116,116,112,58,47,47,32,116,121,112,101,32,102,105,108,101,32,119,105,116,104,32,82,69,65,68,87,82,73,84,69,32,97,99,99,101,115,115,0,32,32,83,112,101,99,105,102,121,32,97,110,32,85,78,67,79,77,80,82,69,83,83,69,68,32,111,117,116,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,104,116,116,112,32,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,100,105,115,107,32,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,58,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,100,105,115,107,32,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,114,101,111,112,101,110,32,100,105,115,107,32,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,109,101,109,111,114,121,32,102,105,108,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,69,114,114,111,114,32,117,110,99,111,109,112,114,101,115,115,105,110,103,32,100,105,115,107,32,102,105,108,101,32,116,111,32,109,101,109,111,114,121,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,67,97,110,32,111,110,108,121,32,104,97,118,101,32,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,115,32,104,101,114,101,32,40,104,116,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,109,101,109,58,0,79,117,116,112,117,116,32,102,105,108,101,32,110,111,116,32,115,101,116,44,32,115,104,111,117,108,100,110,39,116,32,104,97,118,101,32,104,97,112,112,101,110,101,100,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,104,116,116,112,32,102,105,108,101,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,119,0,85,110,97,98,108,101,32,116,111,32,114,101,111,112,101,110,32,116,104,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,69,114,114,111,114,32,117,110,99,111,109,112,114,101,115,115,105,110,103,32,104,116,116,112,32,102,105,108,101,32,116,111,32,100,105,115,107,32,102,105,108,101,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,67,111,110,116,101,110,116,45,76,101,110,103,116,104,32,110,111,116,32,97,32,109,117,108,116,105,112,108,101,32,111,102,32,50,56,56,48,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,32,37,100,0,69,114,114,111,114,32,99,111,112,121,105,110,103,32,104,116,116,112,32,102,105,108,101,32,116,111,32,100,105,115,107,32,102,105,108,101,32,40,104,116,116,112,95,102,105,108,101,95,111,112,101,110,41,0,67,97,110,39,116,32,111,112,101,110,32,102,116,112,58,47,47,32,116,121,112,101,32,102,105,108,101,32,119,105,116,104,32,82,69,65,68,87,82,73,84,69,32,97,99,99,101,115,115,0,83,112,101,99,105,102,121,32,97,110,32,111,117,116,102,105,108,101,32,102,111,114,32,114,47,119,32,97,99,99,101,115,115,32,40,102,116,112,95,111,112,101,110,41,0,102,105,108,101,110,97,109,101,32,116,111,111,32,108,111,110,103,32,40,102,116,112,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,102,116,112,32,102,105,108,101,32,40,102,116,112,95,111,112,101,110,41,0,67,111,117,108,100,32,110,111,116,32,99,114,101,97,116,101,32,109,101,109,111,114,121,32,102,105,108,101,32,116,111,32,112,97,115,115,105,118,101,32,112,111,114,116,32,40,102,116,112,95,111,112,101,110,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,99,111,109,112,114,101,115,115,101,100,32,109,101,109,111,114,121,32,102,105,108,101,32,40,102,116,112,95,111,112,101,110,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,109,101,109,111,114,121,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,41,0,81,85,73,84,10,0,79,117,116,112,117,116,32,102,105,108,101,32,110,111,116,32,115,101,116,44,32,115,104,111,117,108,100,110,39,116,32,104,97,118,101,32,104,97,112,112,101,110,101,100,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,104,116,116,112,32,102,105,108,101,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,114,101,111,112,101,110,32,116,104,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,117,110,99,111,109,112,114,101,115,115,32,116,104,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,102,105,108,101,32,40,102,116,112,95,102,105,108,101,95,111,112,101,110,41,0,67,111,109,112,114,101,115,115,101,100,32,102,105,108,101,115,32,109,117,115,116,32,98,101,32,114,47,111,0,79,117,116,112,117,116,32,102,105,108,101,32,110,111,116,32,115,101,116,44,32,115,104,111,117,108,100,110,39,116,32,104,97,118,101,32,104,97,112,112,101,110,101,100,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,111,112,101,110,32,102,116,112,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,114,101,111,112,101,110,32,100,105,115,107,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,109,101,109,111,114,121,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,99,111,109,112,114,101,115,115,101,100,32,109,101,109,111,114,121,32,102,105,108,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,67,97,110,32,111,110,108,121,32,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,115,32,104,101,114,101,32,40,102,116,112,95,99,111,109,112,114,101,115,115,95,111,112,101,110,41,0,104,116,116,112,109,101,109,58,47,47,0,102,116,112,109,101,109,58,47,47,0,117,112,100,97,116,101,0,114,101,97,100,0,99,114,101,97,116,101,0,85,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,102,105,108,101,0,37,108,100,32,37,108,100,32,0,85,82,76,32,80,97,114,115,101,32,69,114,114,111,114,32,40,114,111,111,116,95,111,112,101,110,41,32,37,115,0,67,111,117,108,100,110,39,116,32,99,111,110,110,101,99,116,32,116,111,32,104,111,115,116,32,40,104,116,116,112,95,111,112,101,110,95,110,101,116,119,111,114,107,41,0,82,79,79,84,85,83,69,82,78,65,77,69,0,85,115,101,114,110,97,109,101,58,32,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,111,110,32,117,115,101,114,110,97,109,101,32,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,111,110,32,117,115,101,114,110,97,109,101,0,69,82,82,79,82,32,111,110,32,82,79,79,84,68,95,85,83,69,82,0,82,79,79,84,80,65,83,83,87,79,82,68,0,80,97,115,115,119,111,114,100,58,32,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,115,101,110,100,105,110,103,32,112,97,115,115,119,111,114,100,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,97,99,107,105,110,103,32,112,97,115,115,119,111,114,100,0,69,82,82,79,82,32,111,110,32,82,79,79,84,68,95,80,65,83,83,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,111,110,32,111,112,101,110,32,0,101,114,114,111,114,32,116,97,108,107,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,121,115,116,101,109,32,111,110,32,111,112,101,110,0,69,82,82,79,82,32,111,110,32,82,79,79,84,68,95,79,80,69,78,0,67,97,110,39,116,32,99,114,101,97,116,101,32,115,111,99,107,101,116,0,104,116,116,112,58,0,108,111,99,97,108,104,111,115,116,0,114,111,111,116,58,0,102,116,112,58,0,102,116,112,32,102,105,108,101,110,97,109,101,32,105,115,32,116,111,111,32,108,111,110,103,32,40,102,116,112,95,111,112,101,110,41,0,85,82,76,32,80,97,114,115,101,32,69,114,114,111,114,32,40,102,116,112,95,111,112,101,110,41,32,37,115,0,97,110,111,110,121,109,111,117,115,0,117,115,101,114,64,104,111,115,116,46,99,111,109,0,102,100,111,112,101,110,32,102,97,105,108,101,100,32,116,111,32,99,111,110,118,101,114,116,32,115,111,99,107,101,116,32,116,111,32,115,116,100,105,111,32,102,105,108,101,32,40,102,116,112,95,111,112,101,110,41,0,50,50,48,32,0,101,114,114,111,114,32,99,111,110,110,101,99,116,105,110,103,32,116,111,32,114,101,109,111,116,101,32,115,101,114,118,101,114,44,32,110,111,32,50,50,48,32,115,101,101,110,32,40,102,116,112,95,111,112,101,110,41,0,85,83,69,82,32,37,115,10,0,51,51,49,32,0,85,83,69,82,32,101,114,114,111,114,32,110,111,32,51,51,49,32,115,101,101,110,32,40,102,116,112,95,111,112,101,110,41,0,80,65,83,83,32,37,115,10,0,50,51,48,32,0,80,65,83,83,32,101,114,114,111,114,44,32,110,111,32,50,51,48,32,115,101,101,110,32,40,102,116,112,95,111,112,101,110,41,0,67,87,68,32,47,10,0,67,87,68,32,37,115,10,0,50,53,48,32,0,67,87,68,32,101,114,114,111,114,44,32,110,111,32,50,53,48,32,115,101,101,110,32,40,102,116,112,95,111,112,101,110,41,0,78,117,108,108,32,102,105,108,101,32,110,97,109,101,32,40,102,116,112,95,111,112,101,110,41,0,50,48,48,32,0,84,89,80,69,32,73,32,101,114,114,111,114,44,32,50,48,48,32,110,111,116,32,115,101,101,110,32,40,102,116,112,95,111,112,101,110,41,0,80,65,83,86,10,0,80,65,83,86,32,101,114,114,111,114,32,40,102,116,112,95,111,112,101,110,41,0,44,41,0,67,111,117,108,100,32,110,111,116,32,99,111,110,110,101,99,116,32,116,111,32,112,97,115,115,105,118,101,32,112,111,114,116,32,40,102,116,112,95,111,112,101,110,41,0,82,69,84,82,32,37,115,10,0,49,53,48,32,0,85,82,76,32,80,97,114,115,101,32,69,114,114,111,114,32,40,104,116,116,112,95,111,112,101,110,41,32,37,115,0,104,116,116,112,95,112,114,111,120,121,0,67,111,117,108,100,110,39,116,32,99,111,110,110,101,99,116,32,116,111,32,104,111,115,116,32,118,105,97,32,112,114,111,120,121,32,115,101,114,118,101,114,32,40,104,116,116,112,95,111,112,101,110,95,110,101,116,119,111,114,107,41,0,102,100,111,112,101,110,32,102,97,105,108,101,100,32,116,111,32,99,111,110,118,101,114,116,32,115,111,99,107,101,116,32,116,111,32,102,105,108,101,32,40,104,116,116,112,95,111,112,101,110,95,110,101,116,119,111,114,107,41,0,71,69,84,32,104,116,116,112,58,47,47,37,115,58,37,45,100,37,115,32,72,84,84,80,47,49,46,48,13,10,0,71,69,84,32,37,115,32,72,84,84,80,47,49,46,48,13,10,0,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,48,49,50,51,52,53,54,55,56,57,43,47,0,65,117,116,104,111,114,105,122,97,116,105,111,110,58,32,66,97,115,105,99,32,37,115,13,10,0,85,115,101,114,45,65,103,101,110,116,58,32,72,69,65,83,65,82,67,47,67,70,73,84,83,73,79,47,37,45,56,46,51,102,13,10,0,72,111,115,116,58,32,37,115,58,37,45,100,13,10,13,10,0,104,116,116,112,32,104,101,97,100,101,114,32,115,104,111,114,116,32,40,104,116,116,112,95,111,112,101,110,95,110,101,116,119,111,114,107,41,32,37,115,0,37,115,32,37,100,0,78,111,116,101,58,32,87,101,98,32,115,101,114,118,101,114,32,114,101,112,108,105,101,100,32,119,105,116,104,32,97,32,116,101,109,112,111,114,97,114,121,32,114,101,100,105,114,101,99,116,32,102,114,111,109,0,78,111,116,101,58,32,87,101,98,32,115,101,114,118,101,114,32,114,101,112,108,105,101,100,32,119,105,116,104,32,97,32,114,101,100,105,114,101,99,116,32,102,114,111,109,0,60,65,32,72,82,69,70,61,34,0,34,0,116,111,32,37,115,10,0,98,117,116,32,119,101,32,119,101,114,101,32,117,110,97,98,108,101,32,116,111,32,102,105,110,100,32,116,104,101,32,114,101,100,105,114,101,99,116,101,100,32,117,114,108,32,105,110,32,116,104,101,32,115,101,114,118,101,114,115,32,114,101,115,112,111,110,115,101,0,67,111,110,116,101,110,116,45,76,101,110,103,116,104,58,0,67,111,110,116,101,110,116,45,69,110,99,111,100,105,110,103,58,0,67,97,110,32,111,110,108,121,32,97,100,100,32,114,111,119,115,32,116,111,32,84,65,66,76,69,32,111,114,32,66,73,78,84,65,66,76,69,32,101,120,116,101,110,115,105,111,110,32,40,102,102,105,114,111,119,41,0,73,110,115,101,114,116,32,112,111,115,105,116,105,111,110,32,103,114,101,97,116,101,114,32,116,104,97,110,32,116,104,101,32,110,117,109,98,101,114,32,111,102,32,114,111,119,115,32,105,110,32,116,104,101,32,116,97,98,108,101,32,40,102,102,105,114,111,119,41,0,73,110,115,101,114,116,32,112,111,115,105,116,105,111,110,32,105,115,32,108,101,115,115,32,116,104,97,110,32,48,32,40,102,102,105,114,111,119,41,0,69,114,114,111,114,32,119,104,105,108,101,32,115,104,105,102,116,105,110,103,32,98,108,111,99,107,32,40,102,102,115,104,102,116,41,0,67,97,110,32,111,110,108,121,32,100,101,108,101,116,101,32,114,111,119,115,32,105,110,32,84,65,66,76,69,32,111,114,32,66,73,78,84,65,66,76,69,32,101,120,116,101,110,115,105,111,110,32,40,102,102,100,114,111,119,41,0,68,101,108,101,116,101,32,112,111,115,105,116,105,111,110,32,103,114,101,97,116,101,114,32,116,104,97,110,32,116,104,101,32,110,117,109,98,101,114,32,111,102,32,114,111,119,115,32,105,110,32,116,104,101,32,116,97,98,108,101,32,40,102,102,100,114,111,119,41,0,68,101,108,101,116,101,32,112,111,115,105,116,105,111,110,32,105,115,32,108,101,115,115,32,116,104,97,110,32,49,32,40,102,102,100,114,111,119,41,0,78,111,46,32,111,102,32,114,111,119,115,32,116,111,32,100,101,108,101,116,101,32,101,120,99,101,101,100,115,32,115,105,122,101,32,111,102,32,116,97,98,108,101,32,40,102,102,100,114,111,119,41,0,67,97,110,32,111,110,108,121,32,97,100,100,32,99,111,108,117,109,110,115,32,116,111,32,84,65,66,76,69,32,111,114,32,66,73,78,84,65,66,76,69,32,101,120,116,101,110,115,105,111,110,32,40,102,102,105,99,111,108,41,0,108,97,98,101,108,32,102,111,114,32,102,105,101,108,100,0,102,111,114,109,97,116,32,111,102,32,102,105,101,108,100,0,98,101,103,105,110,110,105,110,103,32,99,111,108,117,109,110,32,111,102,32,102,105,101,108,100,0,67,97,110,32,111,110,108,121,32,100,101,108,101,116,101,32,99,111,108,117,109,110,32,102,114,111,109,32,84,65,66,76,69,32,111,114,32,66,73,78,84,65,66,76,69,32,101,120,116,101,110,115,105,111,110,32,40,102,102,100,99,111,108,41,0,69,88,84,69,78,68,32,32,0,67,79,77,77,69,78,84,32,32,32,70,73,84,83,32,40,70,108,101,120,105,98,108,101,32,73,109,97,103,101,32,84,114,97,110,115,112,111,114,116,32,83,121,115,116,101,109,41,32,102,111,114,109,97,116,32,105,115,0,80,67,79,85,78,84,32,32,0,71,67,79,85,78,84,32,32,0,73,108,108,101,103,97,108,32,118,97,108,117,101,32,102,111,114,32,84,70,73,69,76,68,83,32,107,101,121,119,111,114,100,58,32,37,100,0,69,120,112,114,101,115,115,105,111,110,32,100,111,101,115,32,110,111,116,32,101,118,97,108,117,97,116,101,32,116,111,32,97,32,108,111,103,105,99,97,108,32,115,99,97,108,97,114,46,0,102,102,105,112,114,115,58,32,117,110,97,98,108,101,32,116,111,32,103,101,116,32,105,109,97,103,101,32,100,105,109,101,110,115,105,111,110,115,0,110,97,120,105,115,61,37,100,44,32,103,80,97,114,115,101,46,116,111,116,97,108,82,111,119,115,61,37,108,100,10,0,66,108,97,110,107,32,101,120,112,114,101,115,115,105,111,110,0,66,97,100,32,114,101,116,117,114,110,32,100,97,116,97,32,116,121,112,101,0,105,110,118,97,108,105,100,32,102,114,101,101,40,103,80,97,114,115,101,46,101,120,112,114,41,32,97,116,32,37,115,58,37,100,10,0,101,118,97,108,95,102,46,99,0,105,110,118,97,108,105,100,32,102,114,101,101,40,103,80,97,114,115,101,46,99,111,108,68,97,116,97,41,32,97,116,32,37,115,58,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,40,40,99,104,97,114,42,42,41,103,80,97,114,115,101,46,118,97,114,68,97,116,97,91,99,111,108,93,46,100,97,116,97,41,91,48,93,41,32,97,116,32,37,115,58,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,103,80,97,114,115,101,46,118,97,114,68,97,116,97,41,32,97,116,32,37,115,58,37,100,10,0,112,97,114,115,101,95,100,97,116,97,40,116,111,116,97,108,61,37,108,100,44,32,111,102,102,115,101,116,61,37,108,100,44,32,102,105,114,115,116,61,37,108,100,44,32,114,111,119,115,61,37,108,100,44,32,99,111,108,115,61,37,100,41,10,0,79,117,116,112,117,116,32,99,111,108,117,109,110,32,102,111,114,32,112,97,114,115,101,114,32,114,101,115,117,108,116,115,32,110,111,116,32,102,111,117,110,100,33,0,105,110,118,97,108,105,100,32,102,114,101,101,40,98,105,116,83,116,114,115,91,48,93,41,32,97,116,32,37,115,58,37,100,10,0,83,101,116,117,112,68,97,116,97,65,114,114,97,121,115,44,32,117,110,104,97,110,100,108,101,100,32,116,121,112,101,32,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,40,40,99,104,97,114,42,42,41,118,97,114,68,97,116,97,45,62,100,97,116,97,41,91,48,93,41,32,97,116,32,37,115,58,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,118,97,114,68,97,116,97,45,62,117,110,100,101,102,41,32,97,116,32,37,115,58,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,114,101,115,117,108,116,45,62,118,97,108,117,101,46,100,97,116,97,46,112,116,114,41,32,97,116,32,37,115,58,37,100,10,0,78,117,109,101,114,105,99,97,108,32,111,118,101,114,102,108,111,119,32,119,104,105,108,101,32,99,111,110,118,101,114,116,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,110,101,99,101,115,115,97,114,121,32,100,97,116,97,116,121,112,101,0,67,97,110,110,111,116,32,99,111,110,118,101,114,116,32,98,105,116,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,100,101,115,105,114,101,100,32,116,121,112,101,46,0,105,110,118,97,108,105,100,32,102,114,101,101,40,114,101,115,117,108,116,45,62,118,97,108,117,101,46,100,97,116,97,46,115,116,114,112,116,114,91,48,93,41,32,97,116,32,37,115,58,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,114,101,115,117,108,116,45,62,118,97,108,117,101,46,100,97,116,97,46,115,116,114,112,116,114,41,32,97,116,32,37,115,58,37,100,10,0,67,97,110,110,111,116,32,99,111,110,118,101,114,116,32,115,116,114,105,110,103,32,101,120,112,114,101,115,115,105,111,110,32,116,111,32,100,101,115,105,114,101,100,32,116,121,112,101,46,0,79,117,116,112,117,116,32,116,97,98,108,101,32,104,97,115,32,100,105,102,102,101,114,101,110,116,32,114,111,119,32,108,101,110,103,116,104,32,102,114,111,109,32,105,110,112,117,116,0,85,110,97,98,108,101,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,114,111,119,32,115,101,108,101,99,116,105,111,110,0,105,110,118,97,108,105,100,32,102,114,101,101,40,73,110,102,111,46,100,97,116,97,80,116,114,41,32,97,116,32,37,115,58,37,100,10,0,67,97,110,110,111,116,32,112,117,116,32,116,97,98,117,108,97,114,32,114,101,115,117,108,116,32,105,110,116,111,32,107,101,121,119,111,114,100,32,40,102,102,99,97,108,99,41,0,72,73,83,84,79,82,89,32,97,110,100,32,67,79,77,77,69,78,84,32,118,97,108,117,101,115,32,109,117,115,116,32,98,101,32,115,116,114,105,110,103,115,32,40,102,102,99,97,108,99,41,0,67,97,110,110,111,116,32,99,114,101,97,116,101,32,76,79,71,73,67,65,76,32,99,111,108,117,109,110,32,105,110,32,65,83,67,73,73,32,116,97,98,108,101,0,68,50,51,46,49,53,0,65,37,108,100,0,37,108,100,37,115,0,78,117,108,108,32,118,97,108,117,101,0,78,117,108,108,32,118,97,108,117,101,32,115,116,114,105,110,103,0,68,69,66,85,71,95,80,73,88,70,73,76,84,69,82,0,117,115,105,110,103,32,100,101,102,97,117,108,116,32,116,97,103,32,39,37,115,39,10,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,99,97,110,110,111,116,32,104,97,118,101,32,115,116,114,105,110,103,32,105,109,97,103,101,0,104,109,109,44,32,105,109,97,103,101,32,102,114,111,109,32,98,105,116,115,63,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,101,120,112,101,99,116,101,100,32,114,101,115,117,108,116,32,100,97,116,97,116,121,112,101,0,76,79,71,73,67,65,76,0,76,79,78,71,0,68,79,85,66,76,69,0,66,73,84,0,85,78,75,78,79,87,78,63,33,0,114,101,115,117,108,116,32,116,121,112,101,32,105,115,32,37,115,32,91,37,100,93,10,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,114,101,97,100,32,105,110,112,117,116,32,105,109,97,103,101,32,112,97,114,97,109,101,116,101,114,115,0,105,110,112,117,116,32,98,105,116,112,105,120,32,37,100,10,0,111,117,116,112,117,116,32,98,105,116,112,105,120,32,37,100,10,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,105,109,97,103,101,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,100,101,116,101,114,109,105,110,101,32,110,117,109,98,101,114,32,111,102,32,107,101,121,99,97,114,100,115,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,114,101,97,100,32,107,101,121,99,97,114,100,32,37,100,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,119,114,105,116,101,32,107,101,121,99,97,114,100,32,39,37,115,39,32,91,37,100,93,10,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,101,120,112,101,99,116,101,100,32,111,117,116,112,117,116,32,98,105,116,112,105,120,32,37,100,10,0,117,110,104,97,110,100,108,101,100,32,112,111,115,105,116,105,118,101,32,111,117,116,112,117,116,32,66,73,84,80,73,88,32,37,100,10,0,117,115,105,110,103,32,98,108,97,110,107,32,37,108,100,10,0,67,82,69,65,84,69,68,0,78,85,76,76,32,112,105,120,101,108,32,118,97,108,117,101,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,119,114,105,116,101,32,66,76,65,78,75,32,107,101,121,119,111,114,100,0,111,117,116,112,117,116,32,104,97,115,32,78,85,76,76,115,0,119,114,111,116,101,32,98,108,97,110,107,32,91,37,100,93,10,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,97,98,108,101,32,116,111,32,114,101,115,101,116,32,105,109,103,110,117,108,108,0,112,105,120,101,108,95,102,105,108,116,101,114,58,32,117,110,101,120,112,101,99,116,101,100,32,99,111,110,115,116,97,110,116,32,114,101,115,117,108,116,32,116,121,112,101,32,91,37,100,93,10,0,117,115,101,32,68,79,85,66,76,69,32,102,111,114,32,37,115,32,119,105,116,104,32,66,83,67,65,76,69,61,37,103,47,66,90,69,82,79,61,37,103,10,0,115,101,116,95,105,109,97,103,101,95,99,111,108,95,116,121,112,101,115,58,32,117,110,114,101,99,111,103,110,105,122,101,100,32,105,109,97,103,101,32,98,105,116,112,105,120,32,91,37,100,93,10,0,108,111,97,100,95,99,111,108,117,109,110,58,32,73,77,65,71,69,95,72,68,85,32,102,82,111,119,61,37,108,100,44,32,110,82,111,119,115,61,37,108,100,32,61,62,32,37,100,10,0,105,110,118,97,108,105,100,32,102,114,101,101,40,40,99,104,97,114,32,42,41,98,121,116,101,115,41,32,97,116,32,37,115,58,37,100,10,0,108,111,97,100,95,99,111,108,117,109,110,58,32,117,110,101,120,112,101,99,116,101,100,32,100,97,116,97,116,121,112,101,32,37,100,0,102,105,110,100,95,99,111,108,117,109,110,40,37,115,41,10,0,102,105,110,100,95,99,111,108,117,109,110,58,32,73,77,65,71,69,95,72,68,85,32,98,117,116,32,110,111,32,80,105,120,101,108,70,105,108,116,101,114,0,102,105,110,100,95,99,111,108,117,109,110,58,32,80,105,120,101,108,70,105,108,116,101,114,32,116,97,103,32,37,115,32,110,111,116,32,102,111,117,110,100,0,84,90,69,82,79,37,100,0,84,83,67,65,76,37,100,0,99,111,108,117,109,110,32,37,100,32,105,115,32,119,105,100,101,114,32,116,104,97,110,32,109,97,120,105,109,117,109,32,37,100,32,99,104,97,114,97,99,116,101,114,115,0,118,97,114,105,97,98,108,101,45,108,101,110,103,116,104,32,97,114,114,97,121,32,99,111,108,117,109,110,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,46,32,116,121,112,101,99,111,100,101,32,61,32,37,100,0,102,102,103,107,101,121,32,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,107,101,121,119,111,114,100,58,32,37,115,0,79,75,32,45,32,110,111,32,101,114,114,111,114,0,110,111,110,45,67,70,73,84,83,73,79,32,112,114,111,103,114,97,109,32,101,114,114,111,114,0,115,97,109,101,32,105,110,112,117,116,32,97,110,100,32,111,117,116,112,117,116,32,102,105,108,101,115,0,97,116,116,101,109,112,116,32,116,111,32,111,112,101,110,32,116,111,111,32,109,97,110,121,32,102,105,108,101,115,0,99,111,117,108,100,32,110,111,116,32,111,112,101,110,32,116,104,101,32,110,97,109,101,100,32,102,105,108,101,0,99,111,117,108,100,110,39,116,32,99,114,101,97,116,101,32,116,104,101,32,110,97,109,101,100,32,102,105,108,101,0,101,114,114,111,114,32,119,114,105,116,105,110,103,32,116,111,32,70,73,84,83,32,102,105,108,101,0,116,114,105,101,100,32,116,111,32,109,111,118,101,32,112,97,115,116,32,101,110,100,32,111,102,32,102,105,108,101,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,102,114,111,109,32,70,73,84,83,32,102,105,108,101,0,99,111,117,108,100,32,110,111,116,32,99,108,111,115,101,32,116,104,101,32,102,105,108,101,0,97,114,114,97,121,32,100,105,109,101,110,115,105,111,110,115,32,116,111,111,32,98,105,103,0,99,97,110,110,111,116,32,119,114,105,116,101,32,116,111,32,114,101,97,100,111,110,108,121,32,102,105,108,101,0,99,111,117,108,100,32,110,111,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,0,105,110,118,97,108,105,100,32,102,105,116,115,102,105,108,101,32,112,111,105,110,116,101,114,0,78,85,76,76,32,105,110,112,117,116,32,112,111,105,110,116,101,114,0,101,114,114,111,114,32,115,101,101,107,105,110,103,32,102,105,108,101,32,112,111,115,105,116,105,111,110,0,105,110,118,97,108,105,100,32,85,82,76,32,112,114,101,102,105,120,0,116,111,111,32,109,97,110,121,32,73,47,79,32,100,114,105,118,101,114,115,0,73,47,79,32,100,114,105,118,101,114,32,105,110,105,116,32,102,97,105,108,101,100,0,110,111,32,73,47,79,32,100,114,105,118,101,114,32,102,111,114,32,116,104,105,115,32,85,82,76,116,121,112,101,0,112,97,114,115,101,32,101,114,114,111,114,32,105,110,32,105,110,112,117,116,32,102,105,108,101,32,85,82,76,0,112,97,114,115,101,32,101,114,114,111,114,32,105,110,32,114,97,110,103,101,32,108,105,115,116,0,98,97,100,32,97,114,103,117,109,101,110,116,32,40,115,104,97,114,101,100,32,109,101,109,32,100,114,118,114,41,0,110,117,108,108,32,112,116,114,32,97,114,103,32,40,115,104,97,114,101,100,32,109,101,109,32,100,114,118,114,41,0,110,111,32,102,114,101,101,32,115,104,97,114,101,100,32,109,101,109,111,114,121,32,104,97,110,100,108,101,115,0,115,104,97,114,101,32,109,101,109,32,100,114,118,114,32,110,111,116,32,105,110,105,116,105,97,108,105,122,101,100,0,73,80,67,32,115,121,115,116,101,109,32,101,114,114,111,114,32,40,115,104,97,114,101,100,32,109,101,109,41,0,110,111,32,109,101,109,111,114,121,32,40,115,104,97,114,101,100,32,109,101,109,32,100,114,118,114,41,0,115,104,97,114,101,32,109,101,109,32,114,101,115,111,117,114,99,101,32,100,101,97,100,108,111,99,107,0,108,111,99,107,32,102,105,108,101,32,111,112,101,110,47,99,114,101,97,116,101,32,102,97,105,108,101,100,0,99,97,110,39,116,32,114,101,115,105,122,101,32,115,104,97,114,101,32,109,101,109,32,98,108,111,99,107,0,104,101,97,100,101,114,32,97,108,114,101,97,100,121,32,104,97,115,32,107,101,121,119,111,114,100,115,0,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,105,110,32,104,101,97,100,101,114,0,107,101,121,119,111,114,100,32,110,117,109,98,101,114,32,111,117,116,32,111,102,32,98,111,117,110,100,115,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,105,115,32,117,110,100,101,102,105,110,101,100,0,115,116,114,105,110,103,32,109,105,115,115,105,110,103,32,99,108,111,115,105,110,103,32,113,117,111,116,101,0,101,114,114,111,114,32,105,110,32,105,110,100,101,120,101,100,32,107,101,121,119,111,114,100,32,110,97,109,101,0,105,108,108,101,103,97,108,32,99,104,97,114,97,99,116,101,114,32,105,110,32,107,101,121,119,111,114,100,0,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,115,32,111,117,116,32,111,102,32,111,114,100,101,114,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,110,111,116,32,112,111,115,105,116,105,118,101,32,105,110,116,0,69,78,68,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,105,108,108,101,103,97,108,32,66,73,84,80,73,88,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,78,65,88,73,83,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,78,65,88,73,83,110,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,80,67,79,85,78,84,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,71,67,79,85,78,84,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,84,70,73,69,76,68,83,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,110,101,103,97,116,105,118,101,32,116,97,98,108,101,32,114,111,119,32,115,105,122,101,0,110,101,103,97,116,105,118,101,32,110,117,109,98,101,114,32,111,102,32,114,111,119,115,0,110,97,109,101,100,32,99,111,108,117,109,110,32,110,111,116,32,102,111,117,110,100,0,105,108,108,101,103,97,108,32,83,73,77,80,76,69,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,102,105,114,115,116,32,107,101,121,119,111,114,100,32,110,111,116,32,83,73,77,80,76,69,0,115,101,99,111,110,100,32,107,101,121,119,111,114,100,32,110,111,116,32,66,73,84,80,73,88,0,116,104,105,114,100,32,107,101,121,119,111,114,100,32,110,111,116,32,78,65,88,73,83,0,109,105,115,115,105,110,103,32,78,65,88,73,83,110,32,107,101,121,119,111,114,100,115,0,102,105,114,115,116,32,107,101,121,119,111,114,100,32,110,111,116,32,88,84,69,78,83,73,79,78,0,67,72,68,85,32,110,111,116,32,97,110,32,65,83,67,73,73,32,116,97,98,108,101,0,67,72,68,85,32,110,111,116,32,97,32,98,105,110,97,114,121,32,116,97,98,108,101,0,80,67,79,85,78,84,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,71,67,79,85,78,84,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,84,70,73,69,76,68,83,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,109,105,115,115,105,110,103,32,84,66,67,79,76,110,32,107,101,121,119,111,114,100,0,109,105,115,115,105,110,103,32,84,70,79,82,77,110,32,107,101,121,119,111,114,100,0,67,72,68,85,32,110,111,116,32,97,110,32,73,77,65,71,69,32,101,120,116,101,110,115,105,111,110,0,105,108,108,101,103,97,108,32,84,66,67,79,76,110,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,67,72,68,85,32,110,111,116,32,97,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,0,99,111,108,117,109,110,32,101,120,99,101,101,100,115,32,119,105,100,116,104,32,111,102,32,116,97,98,108,101,0,109,111,114,101,32,116,104,97,110,32,49,32,109,97,116,99,104,105,110,103,32,99,111,108,46,32,110,97,109,101,0,114,111,119,32,119,105,100,116,104,32,110,111,116,32,61,32,102,105,101,108,100,32,119,105,100,116,104,115,0,117,110,107,110,111,119,110,32,70,73,84,83,32,101,120,116,101,110,115,105,111,110,32,116,121,112,101,0,49,115,116,32,107,101,121,32,110,111,116,32,83,73,77,80,76,69,32,111,114,32,88,84,69,78,83,73,79,78,0,69,78,68,32,107,101,121,119,111,114,100,32,105,115,32,110,111,116,32,98,108,97,110,107,0,72,101,97,100,101,114,32,102,105,108,108,32,97,114,101,97,32,110,111,116,32,98,108,97,110,107,0,68,97,116,97,32,102,105,108,108,32,97,114,101,97,32,105,110,118,97,108,105,100,0,105,108,108,101,103,97,108,32,84,70,79,82,77,32,102,111,114,109,97,116,32,99,111,100,101,0,117,110,107,110,111,119,110,32,84,70,79,82,77,32,100,97,116,97,116,121,112,101,32,99,111,100,101,0,105,108,108,101,103,97,108,32,84,68,73,77,110,32,107,101,121,119,111,114,100,32,118,97,108,117,101,0,105,110,118,97,108,105,100,32,66,73,78,84,65,66,76,69,32,104,101,97,112,32,112,111,105,110,116,101,114,0,117,110,107,110,111,119,110,32,101,114,114,111,114,32,115,116,97,116,117,115,0,105,108,108,101,103,97,108,32,72,68,85,32,110,117,109,98,101,114,0,99,111,108,117,109,110,32,110,117,109,98,101,114,32,60,32,49,32,111,114,32,62,32,116,102,105,101,108,100,115,0,110,101,103,97,116,105,118,101,32,98,121,116,101,32,97,100,100,114,101,115,115,0,110,101,103,97,116,105,118,101,32,110,117,109,98,101,114,32,111,102,32,101,108,101,109,101,110,116,115,0,98,97,100,32,102,105,114,115,116,32,114,111,119,32,110,117,109,98,101,114,0,98,97,100,32,102,105,114,115,116,32,101,108,101,109,101,110,116,32,110,117,109,98,101,114,0,110,111,116,32,97,110,32,65,83,67,73,73,32,40,65,41,32,99,111,108,117,109,110,0,110,111,116,32,97,32,108,111,103,105,99,97,108,32,40,76,41,32,99,111,108,117,109,110,0,98,97,100,32,65,83,67,73,73,32,116,97,98,108,101,32,100,97,116,97,116,121,112,101,0,98,97,100,32,98,105,110,97,114,121,32,116,97,98,108,101,32,100,97,116,97,116,121,112,101,0,110,117,108,108,32,118,97,108,117,101,32,110,111,116,32,100,101,102,105,110,101,100,0,110,111,116,32,97,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,99,111,108,117,109,110,0,105,108,108,101,103,97,108,32,110,117,109,98,101,114,32,111,102,32,100,105,109,101,110,115,105,111,110,115,0,49,115,116,32,112,105,120,101,108,32,110,111,46,32,62,32,108,97,115,116,32,112,105,120,101,108,32,110,111,46,0,66,83,67,65,76,69,32,111,114,32,84,83,67,65,76,110,32,61,32,48,46,0,105,108,108,101,103,97,108,32,97,120,105,115,32,108,101,110,103,116,104,32,60,32,49,0,110,111,116,32,103,114,111,117,112,32,116,97,98,108,101,0,72,68,85,32,97,108,114,101,97,100,121,32,109,101,109,98,101,114,32,111,102,32,103,114,111,117,112,0,103,114,111,117,112,32,109,101,109,98,101,114,32,110,111,116,32,102,111,117,110,100,0,103,114,111,117,112,32,110,111,116,32,102,111,117,110,100,0,98,97,100,32,103,114,111,117,112,32,105,100,0,116,111,111,32,109,97,110,121,32,72,68,85,115,32,116,114,97,99,107,101,100,0,72,68,85,32,97,108,114,101,97,100,32,116,114,97,99,107,101,100,0,98,97,100,32,71,114,111,117,112,105,110,103,32,111,112,116,105,111,110,0,105,100,101,110,116,105,99,97,108,32,112,111,105,110,116,101,114,115,32,40,103,114,111,117,112,115,41,0,109,97,108,108,111,99,32,102,97,105,108,101,100,32,105,110,32,112,97,114,115,101,114,0,102,105,108,101,32,114,101,97,100,32,101,114,114,111,114,32,105,110,32,112,97,114,115,101,114,0,110,117,108,108,32,112,111,105,110,116,101,114,32,97,114,103,32,40,112,97,114,115,101,114,41,0,101,109,112,116,121,32,108,105,110,101,32,40,112,97,114,115,101,114,41,0,99,97,110,110,111,116,32,117,110,114,101,97,100,32,62,32,49,32,108,105,110,101,0,112,97,114,115,101,114,32,116,111,111,32,100,101,101,112,108,121,32,110,101,115,116,101,100,0,102,105,108,101,32,111,112,101,110,32,102,97,105,108,101,100,32,40,112,97,114,115,101,114,41,0,104,105,116,32,69,79,70,32,40,112,97,114,115,101,114,41,0,98,97,100,32,97,114,103,117,109,101,110,116,32,40,112,97,114,115,101,114,41,0,117,110,101,120,112,101,99,116,101,100,32,116,111,107,101,110,32,40,112,97,114,115,101,114,41,0,98,97,100,32,105,110,116,32,116,111,32,115,116,114,105,110,103,32,99,111,110,118,101,114,115,105,111,110,0,98,97,100,32,102,108,111,97,116,32,116,111,32,115,116,114,105,110,103,32,99,111,110,118,101,114,115,105,111,110,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,110,111,116,32,105,110,116,101,103,101,114,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,110,111,116,32,108,111,103,105,99,97,108,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,110,111,116,32,102,108,111,97,116,105,110,103,32,112,116,0,107,101,121,119,111,114,100,32,118,97,108,117,101,32,110,111,116,32,100,111,117,98,108,101,0,98,97,100,32,115,116,114,105,110,103,32,116,111,32,105,110,116,32,99,111,110,118,101,114,115,105,111,110,0,98,97,100,32,115,116,114,105,110,103,32,116,111,32,102,108,111,97,116,32,99,111,110,118,101,114,115,105,111,110,0,98,97,100,32,115,116,114,105,110,103,32,116,111,32,100,111,117,98,108,101,32,99,111,110,118,101,114,116,0,105,108,108,101,103,97,108,32,100,97,116,97,116,121,112,101,32,99,111,100,101,32,118,97,108,117,101,0,105,108,108,101,103,97,108,32,110,111,46,32,111,102,32,100,101,99,105,109,97,108,115,0,100,97,116,97,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,111,118,101,114,102,108,111,119,0,101,114,114,111,114,32,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,0,101,114,114,111,114,32,117,110,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,0,98,97,100,32],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+740633);allocate([100,97,116,101,32,111,114,32,116,105,109,101,32,99,111,110,118,101,114,115,105,111,110,0,115,121,110,116,97,120,32,101,114,114,111,114,32,105,110,32,101,120,112,114,101,115,115,105,111,110,0,101,120,112,114,101,115,115,105,111,110,32,114,101,115,117,108,116,32,119,114,111,110,103,32,116,121,112,101,0,118,101,99,116,111,114,32,114,101,115,117,108,116,32,116,111,111,32,108,97,114,103,101,0,109,105,115,115,105,110,103,32,111,117,116,112,117,116,32,99,111,108,117,109,110,0,98,97,100,32,100,97,116,97,32,105,110,32,112,97,114,115,101,100,32,99,111,108,117,109,110,0,111,117,116,112,117,116,32,101,120,116,101,110,115,105,111,110,32,111,102,32,119,114,111,110,103,32,116,121,112,101,0,87,67,83,32,97,110,103,108,101,32,116,111,111,32,108,97,114,103,101,0,98,97,100,32,87,67,83,32,99,111,111,114,100,105,110,97,116,101,0,101,114,114,111,114,32,105,110,32,87,67,83,32,99,97,108,99,117,108,97,116,105,111,110,0,98,97,100,32,87,67,83,32,112,114,111,106,101,99,116,105,111,110,32,116,121,112,101,0,87,67,83,32,107,101,121,119,111,114,100,115,32,110,111,116,32,102,111,117,110,100],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+750873);allocate([75,101,121,119,111,114,100,32,110,97,109,101,32,99,111,110,116,97,105,110,115,32,101,109,98,101,100,100,101,100,32,115,112,97,99,101,40,115,41,58,32,37,46,56,115,0,67,104,97,114,97,99,116,101,114,32,37,100,32,105,110,32,116,104,105,115,32,107,101,121,119,111,114,100,32,105,115,32,105,108,108,101,103,97,108,58,32,37,46,56,115,0,32,40,84,104,105,115,32,97,32,78,85,76,76,32,40,48,41,32,99,104,97,114,97,99,116,101,114,41,46,0,32,40,84,104,105,115,32,97,110,32,65,83,67,73,73,32,84,65,66,32,40,57,41,32,99,104,97,114,97,99,116,101,114,41,46,0,67,104,97,114,97,99,116,101,114,32,37,100,32,105,110,32,116,104,105,115,32,107,101,121,119,111,114,100,32,105,115,32,105,108,108,101,103,97,108,46,32,72,101,120,32,86,97,108,117,101,32,61,32,37,88,0,32,40,78,85,76,76,32,99,104,97,114,46,41,0,32,40,84,65,66,32,99,104,97,114,46,41,0,32,40,76,105,110,101,32,70,101,101,100,32,99,104,97,114,46,41,0,32,40,86,101,114,116,105,99,97,108,32,84,97,98,41,0,32,40,70,111,114,109,32,70,101,101,100,32,99,104,97,114,46,41,0,32,40,67,97,114,114,105,97,103,101,32,82,101,116,117,114,110,41,0,32,40,69,115,99,97,112,101,32,99,104,97,114,46,41,0,32,40,68,101,108,101,116,101,32,99,104,97,114,46,41,0,73,108,108,101,103,97,108,32,107,101,121,119,111,114,100,32,110,97,109,101,59,32,99,111,110,116,97,105,110,115,32,97,110,32,101,113,117,97,108,115,32,115,105,103,110,32,40,61,41,0,104,105,101,114,97,114,99,104,32,0,61,32,0,73,108,108,101,103,97,108,32,107,101,121,119,111,114,100,32,110,97,109,101,58,0,84,104,101,32,102,111,108,108,111,119,105,110,103,32,107,101,121,119,111,114,100,32,43,32,118,97,108,117,101,32,105,115,32,116,111,111,32,108,111,110,103,32,116,111,32,102,105,116,32,111,110,32,97,32,99,97,114,100,58,0,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,87,97,114,110,105,110,103,58,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,107,101,121,119,111,114,100,32,100,111,101,115,32,110,111,116,32,99,111,110,102,111,114,109,32,116,111,32,116,104,101,32,72,73,69,82,65,82,67,72,32,99,111,110,118,101,110,116,105,111,110,0,70,97,105,108,101,100,32,116,111,32,109,111,118,101,32,116,111,32,72,68,85,32,110,117,109,98,101,114,32,37,100,32,40,102,102,109,97,104,100,41,46,0,69,78,68,32,32,32,32,32,0,84,104,105,115,32,107,101,121,119,111,114,100,32,115,116,114,105,110,103,32,118,97,108,117,101,32,104,97,115,32,110,111,32,99,108,111,115,105,110,103,32,113,117,111,116,101,58,0,84,104,105,115,32,99,111,109,112,108,101,120,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,104,97,115,32,110,111,32,99,108,111,115,105,110,103,32,39,41,39,58,0,32,47,0,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,45,0,32,61,0,39,0,69,114,114,111,114,58,32,65,83,67,73,73,32,116,97,98,108,101,32,84,70,79,82,77,32,99,111,100,101,32,105,115,32,98,108,97,110,107,0,73,108,108,101,103,97,108,32,65,83,67,73,73,32,116,97,98,108,101,32,84,70,79,82,77,110,32,100,97,116,97,116,121,112,101,58,32,39,37,115,39,0,82,97,110,103,101,32,69,114,114,111,114,32,105,110,32,102,102,99,50,105,105,32,99,111,110,118,101,114,116,105,110,103,32,115,116,114,105,110,103,32,116,111,32,108,111,110,103,32,105,110,116,58,32,0,73,108,108,101,103,97,108,32,65,83,67,73,73,32,116,97,98,108,101,32,84,70,79,82,77,110,32,99,111,100,101,58,32,39,37,115,39,0,0,69,114,114,111,114,32,105,110,32,102,102,99,50,114,114,32,99,111,110,118,101,114,116,105,110,103,32,115,116,114,105,110,103,32,116,111,32,102,108,111,97,116,58,32,0,69,114,114,111,114,58,32,98,105,110,97,114,121,32,116,97,98,108,101,32,84,70,79,82,77,32,99,111,100,101,32,105,115,32,98,108,97,110,107,32,40,102,102,98,110,102,109,41,46,0,73,108,108,101,103,97,108,32,98,105,110,97,114,121,32,116,97,98,108,101,32,84,70,79,82,77,110,32,100,97,116,97,116,121,112,101,58,32,39,37,115,39,32,0,102,102,103,99,110,110,32,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,99,111,108,117,109,110,58,32,37,46,52,53,115,0,37,46,48,102,0,67,97,110,110,111,116,32,114,101,97,100,32,102,105,114,115,116,32,107,101,121,119,111,114,100,32,105,110,32,104,101,97,100,101,114,32,40,102,102,114,104,100,117,41,46,0,67,97,110,110,111,116,32,114,101,97,100,32,118,97,108,117,101,32,111,102,32,102,105,114,115,116,32,32,107,101,121,119,111,114,100,32,105,110,32,104,101,97,100,101,114,32,40,102,102,114,104,100,117,41,58,0,65,51,68,84,65,66,76,69,0,51,68,84,65,66,76,69,0,69,120,116,101,110,115,105,111,110,32,100,111,101,115,110,39,116,32,115,116,97,114,116,32,119,105,116,104,32,83,73,77,80,76,69,32,111,114,32,88,84,69,78,83,73,79,78,32,107,101,121,119,111,114,100,46,32,40,102,102,114,104,100,117,41,0,109,97,108,108,111,99,32,102,97,105,108,101,100,32,116,111,32,103,101,116,32,109,101,109,111,114,121,32,102,111,114,32,70,73,84,83,32,97,114,114,97,121,32,100,101,115,99,114,105,112,116,111,114,115,32,40,102,102,112,105,110,105,116,41,0,84,104,105,115,32,115,116,114,105,110,103,32,118,97,108,117,101,32,104,97,115,32,110,111,32,99,108,111,115,105,110,103,32,113,117,111,116,101,32,40,102,102,99,50,115,41,58,0,80,67,79,85,78,84,32,107,101,121,119,111,114,100,32,110,111,116,32,101,113,117,97,108,32,116,111,32,48,32,105,110,32,65,83,67,73,73,32,116,97,98,108,101,32,40,102,102,97,105,110,105,116,41,46,0,32,32,80,67,79,85,78,84,32,61,32,37,108,100,0,109,97,108,108,111,99,32,102,97,105,108,101,100,32,116,111,32,103,101,116,32,109,101,109,111,114,121,32,102,111,114,32,70,73,84,83,32,116,97,98,108,101,32,100,101,115,99,114,105,112,116,111,114,115,32,40,102,102,97,105,110,105,116,41,0,69,78,68,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,105,110,32,65,83,67,73,73,32,116,97,98,108,101,32,104,101,97,100,101,114,32,40,102,102,97,105,110,105,116,41,46,0,82,101,113,117,105,114,101,100,32,37,115,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,40,102,102,97,105,110,105,116,41,46,0,86,97,108,117,101,32,111,102,32,37,115,32,107,101,121,119,111,114,100,32,111,117,116,32,111,102,32,114,97,110,103,101,58,32,37,108,100,32,40,102,102,97,105,110,105,116,41,46,0,67,111,108,117,109,110,32,37,100,32,105,115,32,116,111,111,32,119,105,100,101,32,116,111,32,102,105,116,32,105,110,32,116,97,98,108,101,32,40,102,102,97,105,110,105,116,41,0,32,84,70,79,82,77,32,61,32,37,115,32,97,110,100,32,78,65,88,73,83,49,32,61,32,37,108,100,0,109,97,108,108,111,99,32,102,97,105,108,101,100,32,116,111,32,103,101,116,32,109,101,109,111,114,121,32,102,111,114,32,70,73,84,83,32,116,97,98,108,101,32,100,101,115,99,114,105,112,116,111,114,115,32,40,102,102,98,105,110,105,116,41,0,69,78,68,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,105,110,32,98,105,110,97,114,121,32,116,97,98,108,101,32,104,101,97,100,101,114,32,40,102,102,98,105,110,105,116,41,46,0,82,101,113,117,105,114,101,100,32,37,115,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,40,102,102,98,105,110,105,116,41,46,0,78,65,88,73,83,49,32,61,32,37,108,100,32,105,115,32,110,111,116,32,101,113,117,97,108,32,116,111,32,116,104,101,32,115,117,109,32,111,102,32,99,111,108,117,109,110,32,119,105,100,116,104,115,58,32,37,108,100,0,84,89,80,69,0,70,79,82,77,0,66,67,79,76,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,118,97,108,117,101,32,111,102,32,37,115,32,97,115,32,97,110,32,105,110,116,101,103,101,114,58,32,37,115,0,83,67,65,76,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,118,97,108,117,101,32,111,102,32,37,115,32,97,115,32,97,32,100,111,117,98,108,101,58,32,37,115,0,90,69,82,79,0,82,97,110,103,101,32,69,114,114,111,114,32,105,110,32,102,102,99,50,106,106,32,99,111,110,118,101,114,116,105,110,103,32,115,116,114,105,110,103,32,116,111,32,108,111,110,103,108,111,110,103,32,105,110,116,58,32,0,68,73,77,0,72,69,65,80,0,117,110,107,110,111,119,110,32,98,105,110,97,114,121,32,116,97,98,108,101,32,99,111,108,117,109,110,32,116,121,112,101,58,32,37,115,0,0,69,114,114,111,114,32,105,110,32,102,102,99,50,100,100,32,99,111,110,118,101,114,116,105,110,103,32,115,116,114,105,110,103,32,116,111,32,100,111,117,98,108,101,58,32,0,115,116,114,101,97,109,58,47,47,0,73,109,97,103,101,32,103,114,111,117,112,32,110,117,109,98,101,114,32,105,115,32,108,101,115,115,32,116,104,97,110,32,49,58,32,37,46,48,102,0,83,116,97,114,116,105,110,103,32,114,111,119,32,110,117,109,98,101,114,32,105,115,32,108,101,115,115,32,116,104,97,110,32,49,58,32,37,46,48,102,0,83,116,97,114,116,105,110,103,32,101,108,101,109,101,110,116,32,110,117,109,98,101,114,32,108,101,115,115,32,116,104,97,110,32,49,58,32,37,108,100,0,84,114,105,101,100,32,116,111,32,114,101,97,100,32,111,114,32,119,114,105,116,101,32,108,101,115,115,32,116,104,97,110,32,48,32,101,108,101,109,101,110,116,115,58,32,37,46,48,102,0,32,32,84,104,101,114,101,32,97,114,101,32,37,100,32,99,111,108,117,109,110,115,32,105,110,32,116,104,105,115,32,116,97,98,108,101,46,0,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,65,83,67,73,73,32,115,116,114,105,110,103,32,99,111,108,117,109,110,32,105,115,32,116,111,111,32,119,105,100,101,58,32,37,108,100,59,32,109,97,120,32,115,117,112,112,111,114,116,101,100,32,119,105,100,116,104,32,105,115,32,37,100,0,70,105,114,115,116,32,101,108,101,109,101,110,116,32,116,111,32,119,114,105,116,101,32,105,115,32,116,111,111,32,108,97,114,103,101,58,32,37,108,100,59,32,109,97,120,32,97,108,108,111,119,101,100,32,118,97,108,117,101,32,105,115,32,37,108,100,0,70,97,105,108,101,100,32,116,111,32,97,100,100,32,115,112,97,99,101,32,102,111,114,32,37,46,48,102,32,110,101,119,32,114,111,119,115,32,105,110,32,116,97,98,108,101,46,0,65,116,116,101,109,112,116,101,100,32,116,111,32,114,101,97,100,32,102,114,111,109,32,103,114,111,117,112,32,37,108,100,32,111,102,32,116,104,101,32,72,68,85,44,0,104,111,119,101,118,101,114,32,116,104,101,32,72,68,85,32,111,110,108,121,32,99,111,110,116,97,105,110,115,32,37,108,100,32,103,114,111,117,112,40,115,41,46,0,65,116,116,101,109,112,116,32,116,111,32,114,101,97,100,32,112,97,115,116,32,101,110,100,32,111,102,32,97,114,114,97,121,58,0,32,32,73,109,97,103,101,32,104,97,115,32,32,37,108,100,32,101,108,101,109,101,110,116,115,59,0,32,32,84,114,105,101,100,32,116,111,32,114,101,97,100,32,37,108,100,32,101,108,101,109,101,110,116,115,32,115,116,97,114,116,105,110,103,32,97,116,32,101,108,101,109,101,110,116,32,37,108,100,46,0,65,116,116,101,109,112,116,32,116,111,32,114,101,97,100,32,112,97,115,116,32,101,110,100,32,111,102,32,116,97,98,108,101,58,0,32,32,84,97,98,108,101,32,104,97,115,32,37,46,48,102,32,114,111,119,115,32,119,105,116,104,32,37,46,48,102,32,101,108,101,109,101,110,116,115,32,112,101,114,32,114,111,119,59,0,32,32,84,114,105,101,100,32,116,111,32,114,101,97,100,32,37,46,48,102,32,101,108,101,109,101,110,116,115,32,115,116,97,114,116,105,110,103,32,97,116,32,114,111,119,32,37,46,48,102,44,32,101,108,101,109,101,110,116,32,37,46,48,102,46,0,70,97,105,108,101,100,32,116,111,32,101,120,116,101,110,100,32,116,104,101,32,115,105,122,101,32,111,102,32,116,104,101,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,104,101,97,112,32,98,121,32,37,108,100,32,98,108,111,99,107,115,46,0,65,116,116,101,109,112,116,32,116,111,32,114,101,97,100,32,112,97,115,116,32,101,110,100,32,111,102,32,116,97,98,108,101,0,32,32,84,97,98,108,101,32,104,97,115,32,37,46,48,102,32,114,111,119,115,32,97,110,100,32,116,114,105,101,100,32,116,111,32,114,101,97,100,32,114,111,119,32,37,46,48,102,46,0,83,116,97,114,116,105,110,103,32,101,108,101,109,101,110,116,32,116,111,32,114,101,97,100,32,105,110,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,99,111,108,117,109,110,32,105,115,32,116,111,111,32,108,97,114,103,101,58,32,37,108,100,0,32,32,84,104,105,115,32,114,111,119,32,111,110,108,121,32,99,111,110,116,97,105,110,115,32,37,108,100,32,101,108,101,109,101,110,116,115,0,69,78,68,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,69,114,114,111,114,32,119,104,105,108,101,32,119,114,105,116,105,110,103,32,69,78,68,32,99,97,114,100,32,40,102,102,119,101,110,100,41,46,0,80,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,99,111,108,117,109,110,32,100,101,115,99,114,105,112,116,111,114,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,88,84,69,78,83,73,79,78,61,32,39,73,77,65,71,69,32,32,32,39,32,32,32,32,32,32,32,32,32,32,47,32,73,77,65,71,69,32,101,120,116,101,110,115,105,111,110,0,72,101,97,112,32,105,115,32,116,111,111,32,98,105,103,32,116,111,32,116,101,115,116,32,40,32,62,32,50,42,42,51,49,32,98,121,116,101,115,41,46,32,40,102,102,116,104,101,97,112,41,0,70,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,98,117,102,102,101,114,32,116,111,32,116,101,115,116,32,116,104,101,32,104,101,97,112,0,68,101,115,99,114,105,112,116,111,114,32,105,110,32,114,111,119,32,37,108,100,44,32,99,111,108,117,109,110,32,37,100,32,104,97,115,32,105,110,118,97,108,105,100,32,104,101,97,112,32,97,100,100,114,101,115,115,0,109,101,109,58,47,47,116,101,109,112,104,101,97,112,102,105,108,101,0,70,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,116,101,109,112,111,114,97,114,121,32,102,105,108,101,32,102,111,114,32,116,104,101,32,104,101,97,112,0,70,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,99,111,112,121,32,111,102,32,116,104,101,32,104,101,97,112,0,70,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,98,117,102,102,101,114,32,116,111,32,99,111,112,121,32,116,104,101,32,104,101,97,112,0,69,114,114,111,114,32,100,101,108,101,116,105,110,103,32,70,73,84,83,32,98,108,111,99,107,115,32,40,102,102,100,98,108,107,41,0,69,114,114,111,114,32,119,104,105,108,101,32,99,108,111,115,105,110,103,32,72,68,85,32,110,117,109,98,101,114,32,37,100,32,40,102,102,99,104,100,117,41,46,0,69,114,114,111,114,32,119,104,105,108,101,32,117,112,100,97,116,105,110,103,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,118,101,99,116,111,114,32,84,70,79,82,77,110,32,118,97,108,117,101,115,32,40,102,102,117,112,116,102,41,46,0,40,37,46,48,102,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,68,97,116,97,32,85,110,105,116,32,102,105,108,108,32,98,121,116,101,115,32,40,102,102,112,100,102,108,41,46,0,72,68,85,78,65,77,69,0,73,77,65,71,69,32,32,0,67,77,80,84,89,80,69,0,78,65,77,69,0,86,65,76,0,84,73,76,69,0,66,73,84,80,73,88,32,0,83,67,65,76,69,32,32,0,90,69,82,79,32,32,32,0,66,76,65,78,75,32,32,0,83,73,77,80,76,69,32,0,84,69,78,83,73,79,78,0,69,88,84,69,78,68,32,0,80,67,79,85,78,84,32,0,71,67,79,85,78,84,32,0,81,85,65,78,84,73,90,0,68,73,84,72,69,82,48,0,73,84,80,73,88,32,32,0,76,79,67,75,69,68,32,0,76,65,78,75,32,32,32,0,85,78,73,84,32,32,32,0,79,77,77,69,78,84,0,67,79,77,77,69,78,84,32,32,32,97,110,100,32,65,115,116,114,111,112,104,121,115,105,99,115,39,44,32,118,111,108,117,109,101,32,51,55,54,44,32,112,97,103,101,32,51,0,67,79,77,77,69,78,84,32,32,32,70,73,84,83,32,40,70,108,101,120,105,98,108,101,32,73,109,97,103,101,32,84,114,97,110,115,112,111,114,116,32,83,121,115,116,101,109,0,67,79,77,77,69,78,84,32,32,32,65,115,116,114,111,112,104,121,115,105,99,115,32,83,117,112,112,108,101,109,101,110,116,32,83,101,114,105,101,115,32,118,52,52,47,112,51,0,67,79,77,77,69,78,84,32,32,32,67,111,110,116,97,99,116,32,116,104,101,32,78,65,83,65,32,83,99,105,101,110,99,101,32,79,102,102,105,99,101,32,111,102,32,83,116,0,67,79,77,77,69,78,84,32,32,32,70,73,84,83,32,68,101,102,105,110,105,116,105,111,110,32,100,111,99,117,109,101,110,116,32,35,49,48,48,32,97,110,100,32,111,116,104,0,72,69,67,75,83,85,77,0,79,78,84,73,78,85,69,0,85,78,73,84,0,82,86,65,76,0,82,80,73,88,0,82,79,84,65,0,82,68,69,82,0,83,89,69,82,0,68,69,76,84,0,65,84,65,83,85,77,32,0,65,84,65,77,73,78,32,0,65,84,65,77,65,88,32,0,65,84,69,45,79,66,83,0,88,84,69,78,68,32,32,0,78,68,32,32,32,32,32,0,88,84,78,65,77,69,32,0,88,84,86,69,82,32,32,0,88,84,76,69,86,69,76,0,81,85,73,78,79,88,0,81,85,73,0,80,79,67,72,32,32,32,0,67,79,85,78,84,32,32,0,82,79,85,80,83,32,32,0,68,85,78,65,77,69,32,0,68,85,86,69,82,32,32,0,68,85,76,69,86,69,76,0,73,83,84,79,82,89,0,79,78,80,79,76,69,0,65,84,80,79,76,69,0,79,78,80,0,65,84,80,0,74,68,45,79,66,83,32,0,74,68,79,66,0,65,88,73,83,0,65,68,69,67,83,89,83,0,65,68,69,83,89,83,0,65,68,69,0,73,77,80,76,69,32,32,0,70,73,69,76,68,83,32,0,72,69,65,80,32,32,32,0,68,73,83,80,0,76,77,73,78,0,76,77,65,88,0,68,77,73,78,0,68,77,65,88,0,67,84,89,80,0,67,84,89,0,67,85,78,73,0,67,85,78,0,67,82,86,76,0,67,82,86,0,67,82,80,88,0,67,82,80,0,67,82,79,84,0,67,68,76,84,0,67,68,69,0,67,82,68,0,67,83,89,0,87,67,83,0,67,83,65,88,69,83,0,67,83,78,65,77,69,0,67,65,88,0,67,83,78,0,69,114,114,111,114,32,105,110,32,102,102,99,50,105,32,101,118,97,108,117,97,116,105,110,103,32,115,116,114,105,110,103,32,97,115,32,97,110,32,105,110,116,101,103,101,114,58,32,0,69,114,114,111,114,32,105,110,32,102,102,99,50,106,32,101,118,97,108,117,97,116,105,110,103,32,115,116,114,105,110,103,32,97,115,32,97,32,108,111,110,103,32,105,110,116,101,103,101,114,58,32,0,69,114,114,111,114,32,105,110,32,102,102,99,50,108,32,101,118,97,108,117,97,116,105,110,103,32,115,116,114,105,110,103,32,97,115,32,97,32,108,111,103,105,99,97,108,58,32,0,69,114,114,111,114,32,105,110,32,102,102,99,50,114,32,101,118,97,108,117,97,116,105,110,103,32,115,116,114,105,110,103,32,97,115,32,97,32,102,108,111,97,116,58,32,0,69,114,114,111,114,32,105,110,32,102,102,99,50,100,32,101,118,97,108,117,97,116,105,110,103,32,115,116,114,105,110,103,32,97,115,32,97,32,100,111,117,98,108,101,58,32,0,84,67,84,89,80,110,0,67,84,89,80,69,110,0,84,67,84,89,110,97,0,67,84,89,80,69,110,97,0,84,67,85,78,73,110,0,67,85,78,73,84,110,0,84,67,85,78,110,97,0,67,85,78,73,84,110,97,0,84,67,82,86,76,110,0,67,82,86,65,76,110,0,84,67,82,86,110,97,0,67,82,86,65,76,110,97,0,84,67,68,76,84,110,0,67,68,69,76,84,110,0,84,67,68,69,110,97,0,67,68,69,76,84,110,97,0,84,67,82,80,88,110,0,67,82,80,73,88,110,0,84,67,82,80,110,97,0,67,82,80,73,88,110,97,0,84,67,82,79,84,110,0,67,82,79,84,65,110,0,84,80,110,95,109,97,0,80,67,110,95,109,97,0,84,80,67,110,95,109,0,84,67,110,95,109,97,0,67,68,110,95,109,97,0,84,67,68,110,95,109,0,84,86,110,95,108,97,0,80,86,110,95,108,97,0,84,80,86,110,95,108,0,84,83,110,95,108,97,0,80,83,110,95,108,97,0,84,80,83,110,95,108,0,84,87,67,83,110,97,0,87,67,83,78,65,77,69,97,0,84,67,78,65,110,97,0,67,78,65,77,69,110,97,0,84,67,82,68,110,97,0,67,82,68,69,82,110,97,0,84,67,83,89,110,97,0,67,83,89,69,82,110,97,0,76,79,78,80,110,97,0,76,79,78,80,79,76,69,97,0,76,65,84,80,110,97,0,76,65,84,80,79,76,69,97,0,69,81,85,73,110,97,0,69,81,85,73,78,79,88,97,0,77,74,68,79,66,110,0,77,74,68,45,79,66,83,0,77,74,68,65,110,0,77,74,68,45,65,86,71,0,68,65,86,71,110,0,68,65,84,69,45,65,86,71,0,82,65,68,69,110,97,0,82,65,68,69,83,89,83,97,0,82,70,82,81,110,97,0,82,69,83,84,70,82,81,97,0,82,87,65,86,110,97,0,82,69,83,84,87,65,86,97,0,83,80,69,67,110,97,0,83,80,69,67,83,89,83,97,0,83,79,66,83,110,97,0,83,83,89,83,79,66,83,97,0,83,83,82,67,110,97,0,83,83,89,83,83,82,67,97,0,69,80,79,67,72,0,77,74,68,45,63,63,63,63,0,68,65,84,69,63,63,63,63,0,84,73,77,69,63,63,63,63,0,82,65,68,69,67,83,89,83,0,84,69,76,69,83,67,79,80,0,73,78,83,84,82,85,77,69,0,79,66,83,69,82,86,69,82,0,78,65,88,73,83,105,0,84,68,73,77,35,0,69,88,84,76,69,86,69,76,0,78,65,88,76,69,78,0,65,88,76,69,78,35,0,84,63,63,63,63,35,97,0,84,67,63,63,35,97,0,84,63,63,35,95,35,0,84,87,67,83,35,97,0,76,79,78,80,35,97,0,76,65,84,80,35,97,0,69,81,85,73,35,97,0,77,74,68,79,66,35,0,77,74,68,65,35,0,82,65,68,69,35,97,0,68,65,86,71,35,0,105,67,84,89,80,35,0,105,67,84,89,35,97,0,105,67,85,78,73,35,0,105,67,85,78,35,97,0,105,67,82,86,76,35,0,105,67,68,76,84,35,0,105,67,82,80,88,35,0,105,67,82,86,35,97,0,105,67,68,69,35,97,0,105,67,82,80,35,97,0,105,106,80,67,35,97,0,105,106,67,68,35,97,0,105,86,35,95,35,97,0,105,83,35,95,35,97,0,105,67,82,68,35,97,0,105,67,83,89,35,97,0,105,67,82,79,84,35,0,87,67,65,88,35,97,0,87,67,83,78,35,97,0,105,67,78,65,35,97,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,98,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,98,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,98,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,98,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,100,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,100,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,100,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,100,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,101,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,101,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,101,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,101,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,105,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,105,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,105,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,105,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,106,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,106,41,46,0,102,102,103,115,118,106,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,106,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,107,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,111,117,103,104,32,37,46,48,102,32,111,102,32,108,111,103,105,99,97,108,32,97,114,114,97,121,32,40,102,102,103,99,108,41,46,0,37,49,52,46,54,69,0,78,85,76,76,0,37,50,48,108,108,100,0,37,42,115,0,37,35,50,51,46,49,53,71,0,37,49,49,46,48,102,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,115,41,46,0,67,97,110,110,111,116,32,114,101,97,100,32,98,121,116,101,115,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,119,104,105,99,104,32,104,97,115,32,102,111,114,109,97,116,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,115,98,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,115,98,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,115,98,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,115,98,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,117,107,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,117,107,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,117,105,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,117,105,41,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,117,105,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,117,105,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,67,97,110,110,111,116,32,114,101,97,100,32,110,117,109,98,101,114,115,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,119,104,105,99,104,32,104,97,115,32,102,111,114,109,97,116,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,99,111,108,117,109,110,32,37,100,32,40,102,102,103,99,108,117,106,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,102,114,111,109,32,105,109,97,103,101,32,40,102,102,103,99,108,117,106,41,46,0,78,117,109,101,114,105,99,97,108,32,111,118,101,114,102,108,111,119,32,100,117,114,105,110,103,32,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,119,104,105,108,101,32,114,101,97,100,105,110,103,32,70,73,84,83,32,100,97,116,97,46,0,78,65,88,73,83,32,61,32,37,100,32,105,110,32,99,97,108,108,32,116,111,32,102,102,103,115,118,117,106,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,0,102,102,103,115,118,117,106,58,32,105,108,108,101,103,97,108,32,114,97,110,103,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,97,120,105,115,32,37,108,100,0,67,97,110,110,111,116,32,114,101,97,100,32,110,117,109,98,101,114,32,102,114,111,109,32,65,83,67,73,73,32,116,97,98,108,101,0,67,111,108,117,109,110,32,102,105,101,108,100,32,61,32,37,115,46,0,67,97,110,110,111,116,32,103,101,116,32,107,101,121,119,111,114,100,32,110,117,109,98,101,114,32,37,100,46,32,32,73,116,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,72,73,69,82,65,82,67,72,32,0,107,101,121,119,111,114,100,32,37,115,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,32,99,111,109,112,108,101,120,32,118,97,108,117,101,32,40,102,102,103,107,121,99,41,58,0,41,0,44,0,107,101,121,119,111,114,100,32,37,115,32,100,111,101,115,32,110,111,116,32,104,97,118,101,32,97,32,99,111,109,112,108,101,120,32,118,97,108,117,101,32,40,102,102,103,107,121,109,41,58,0,67,79,78,84,73,78,85,69,32,32,0,68,50,51,52,53,54,55,56,61,32,32,0,78,97,109,101,32,111,102,32,107,101,121,119,111,114,100,32,110,111,46,32,37,100,32,99,111,110,116,97,105,110,115,32,105,108,108,101,103,97,108,32,99,104,97,114,97,99,116,101,114,40,115,41,58,32,37,115,0,32,32,40,84,104,105,115,32,109,97,121,32,105,110,100,105,99,97,116,101,32,97,32,109,105,115,115,105,110,103,32,69,78,68,32,107,101,121,119,111,114,100,41,46,0,73,108,108,101,103,97,108,32,100,105,109,101,110,115,105,111,110,115,32,102,111,114,109,97,116,58,32,37,115,0,111,110,101,32,111,114,32,109,111,114,101,32,100,105,109,101,110,115,105,111,110,32,97,114,101,32,108,101,115,115,32,116,104,97,110,32,48,32,40,102,102,100,116,100,109,41,0,73,108,108,101,103,97,108,32,84,68,73,77,32,107,101,121,119,111,114,100,32,118,97,108,117,101,58,32,37,115,0,111,110,101,32,111,114,32,109,111,114,101,32,84,68,73,77,32,118,97,108,117,101,115,32,97,114,101,32,108,101,115,115,32,116,104,97,110,32,48,32,40,102,102,100,116,100,109,41,0,99,111,108,117,109,110,32,118,101,99,116,111,114,32,108,101,110,103,116,104,44,32,37,46,48,102,44,32,100,111,101,115,32,110,111,116,32,101,113,117,97,108,32,84,68,73,77,110,32,97,114,114,97,121,32,115,105,122,101,44,32,37,46,48,102,0,70,105,114,115,116,32,107,101,121,119,111,114,100,32,111,102,32,116,104,101,32,102,105,108,101,32,105,115,32,110,111,116,32,83,73,77,80,76,69,58,32,37,115,0,66,97,100,32,118,97,108,117,101,32,115,116,114,105,110,103,32,102,111,114,32,88,84,69,78,83,73,79,78,32,107,101,121,119,111,114,100,58,0,73,85,69,73,77,65,71,69,0,70,105,114,115,116,32,107,101,121,119,111,114,100,32,111,102,32,116,104,101,32,101,120,116,101,110,115,105,111,110,32,105,115,32,110,111,116,32,88,84,69,78,83,73,79,78,58,32,37,115,0,84,104,105,115,32,105,115,32,110,111,116,32,97,110,32,73,77,65,71,69,32,101,120,116,101,110,115,105,111,110,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,66,73,84,80,73,88,32,118,97,108,117,101,32,111,102,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,78,65,88,73,83,32,118,97,108,117,101,32,111,102,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,78,65,88,73,83,110,32,118,97,108,117,101,115,32,111,102,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,83,101,99,111,110,100,32,107,101,121,119,111,114,100,32,111,102,32,116,104,101,32,101,120,116,101,110,115,105,111,110,32,105,115,32,110,111,116,32,66,73,84,80,73,88,58,32,37,115,0,86,97,108,117,101,32,111,102,32,66,73,84,80,73,88,32,107,101,121,119,111,114,100,32,105,115,32,110,111,116,32,97,110,32,105,110,116,101,103,101,114,58,32,37,115,0,73,108,108,101,103,97,108,32,118,97,108,117,101,32,102,111,114,32,66,73,84,80,73,88,32,107,101,121,119,111,114,100,58,32,37,115,0,78,65,88,73,83,32,61,32,37,108,100,32,105,115,32,105,108,108,101,103,97,108,0,70,97,105,108,101,100,32,116,111,32,102,105,110,100,32,116,104,101,32,69,78,68,32,107,101,121,119,111,114,100,32,105,110,32,104,101,97,100,101,114,32,40,102,102,103,112,104,100,41,46,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,66,83,67,65,76,69,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,32,100,111,117,98,108,101,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,66,90,69,82,79,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,32,100,111,117,98,108,101,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,66,76,65,78,75,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,110,32,105,110,116,101,103,101,114,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,80,67,79,85,78,84,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,110,32,105,110,116,101,103,101,114,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,71,67,79,85,78,84,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,110,32,105,110,116,101,103,101,114,58,32,37,115,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,69,88,84,69,78,68,32,107,101,121,119,111,114,100,32,118,97,108,117,101,32,97,115,32,97,32,108,111,103,105,99,97,108,58,32,37,115,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,116,104,101,32,114,101,113,117,105,114,101,100,32,112,114,105,109,97,114,121,32,97,114,114,97,121,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,46,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,116,104,101,32,114,101,113,117,105,114,101,100,32,105,109,97,103,101,32,101,120,116,101,110,115,105,111,110,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,46,0,56,0,50,0,49,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,115,32,105,110,32,116,104,101,32,116,97,98,108,101,32,104,101,97,100,101,114,32,40,70,84,71,84,84,66,41,46,0,102,102,103,116,107,110,32,102,111,117,110,100,32,117,110,101,120,112,101,99,116,101,100,32,107,101,121,119,111,114,100,32,111,114,32,118,97,108,117,101,32,102,111,114,32,107,101,121,119,111,114,100,32,110,111,46,32,37,100,46,0,32,69,120,112,101,99,116,101,100,32,112,111,115,105,116,105,118,101,32,105,110,116,101,103,101,114,32,107,101,121,119,111,114,100,32,37,115,44,32,98,117,116,32,105,110,115,116,101,97,100,0,32,102,111,117,110,100,32,107,101,121,119,111,114,100,32,37,115,32,119,105,116,104,32,118,97,108,117,101,32,37,115,0,102,102,103,116,107,110,106,106,32,102,111,117,110,100,32,117,110,101,120,112,101,99,116,101,100,32,107,101,121,119,111,114,100,32,111,114,32,118,97,108,117,101,32,102,111,114,32,107,101,121,119,111,114,100,32,110,111,46,32,37,100,46,0,102,102,116,107,121,110,32,102,111,117,110,100,32,117,110,101,120,112,101,99,116,101,100,32,107,101,121,119,111,114,100,32,111,114,32,118,97,108,117,101,32,102,111,114,32,107,101,121,119,111,114,100,32,110,111,46,32,37,100,46,0,32,69,120,112,101,99,116,101,100,32,107,101,121,119,111,114,100,32,37,115,32,119,105,116,104,32,118,97,108,117,101,32,37,115,44,32,98,117,116,0,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,116,111,32,104,111,108,100,32,97,108,108,32,116,104,101,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,0,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,69,78,68,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,71,82,79,85,80,73,78,71,0,72,68,85,32,99,111,110,116,97,105,110,115,32,97,32,71,114,111,117,112,105,110,103,32,84,97,98,108,101,0,71,114,111,117,112,105,110,103,32,84,97,98,108,101,32,118,101,114,115,46,32,40,116,104,105,115,32,102,105,108,101,41,0,71,82,80,78,65,77,69,0,71,114,111,117,112,105,110,103,32,84,97,98,108,101,32,110,97,109,101,0,77,69,77,66,69,82,95,80,79,83,73,84,73,79,78,0,77,69,77,66,69,82,95,86,69],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+753192);allocate([82,83,73,79,78,0,84,70,79,82,77,37,100,0,84,78,85,76,76,37,100,0,67,111,108,117,109,110,32,78,117,108,108,32,86,97,108,117,101,0,38,0,77,69,77,66,69,82,95,88,84,69,78,83,73,79,78,0,56,65,0,77,69,77,66,69,82,95,78,65,77,69,0,49,74,0,77,69,77,66,69,82,95,85,82,73,95,84,89,80,69,0,51,65,0,77,69,77,66,69,82,95,76,79,67,65,84,73,79,78,0,50,53,54,65,0,73,110,118,97,108,105,100,32,118,97,108,117,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,116,104,101,32,103,114,111,117,112,116,121,112,101,32,112,97,114,97,109,101,116,101,114,32,40,102,102,103,116,100,99,41,0,83,112,101,99,105,102,105,101,100,32,72,68,85,32,105,115,32,110,111,116,32,97,32,71,114,111,117,112,105,110,103,32,84,97,98,108,101,32,40,102,102,103,116,103,99,41,0,83,112,101,99,105,102,105,101,100,32,72,68,85,32,104,97,115,32,109,117,108,116,105,112,117,108,101,32,71,114,111,117,112,32,116,97,98,108,101,32,99,111,108,115,32,100,101,102,105,110,101,100,32,40,102,102,103,116,103,99,41,0,83,112,101,99,105,102,105,101,100,32,72,68,85,32,105,115,32,110,111,116,32,97,32,71,114,111,117,112,105,110,103,32,116,97,98,108,101,32,40,102,102,103,116,110,109,41,0,73,110,118,97,108,105,100,32,118,97,108,117,101,32,102,111,114,32,116,104,101,32,114,109,111,112,116,32,112,97,114,97,109,101,116,101,114,32,115,112,101,99,105,102,105,101,100,32,40,102,102,103,116,114,109,41,0,99,97,110,110,111,116,32,109,111,100,105,102,121,32,103,114,111,117,112,105,110,103,32,116,97,98,108,101,32,40,102,102,103,116,97,109,41,0,71,82,80,73,68,37,100,0,71,82,80,76,67,37,100,0,78,111,32,71,82,80,76,67,37,100,32,102,111,117,110,100,32,102,111,114,32,71,82,80,73,68,37,100,0,73,110,118,97,108,105,100,32,118,97,108,117,101,32,115,112,101,99,105,102,105,101,100,32,102,111,114,32,116,104,101,32,114,109,111,112,116,32,112,97,114,97,109,101,116,101,114,32,40,102,102,103,109,114,109,41,0,67,97,110,110,111,116,32,111,112,101,110,32,116,104,101,32,37,100,116,104,32,103,114,111,117,112,32,116,97,98,108,101,32,40,102,102,103,109,117,108,41,0,84,104,101,32,37,100,116,104,32,103,114,111,117,112,32,99,97,110,110,111,116,32,98,101,32,109,111,100,105,102,105,101,100,32,40,102,102,103,116,97,109,41,0,99,97,110,110,111,116,32,108,111,99,97,116,101,32,109,101,109,98,101,114,39,115,32,101,110,116,114,121,32,105,110,32,103,114,111,117,112,32,116,97,98,108,101,32,40,102,102,103,109,117,108,41,0,67,97,110,110,111,116,32,109,111,100,105,102,121,32,109,101,109,98,101,114,32,72,68,85,44,32,111,112,101,110,101,100,32,82,69,65,68,79,78,76,89,32,40,102,102,103,109,117,108,41,0,80,82,73,77,65,82,89,0,85,82,76,0,67,97,110,110,111,116,32,111,112,101,110,32,109,101,109,98,101,114,32,72,68,85,32,102,105,108,101,32,119,105,116,104,32,85,82,73,32,116,121,112,101,32,37,115,32,40,102,102,103,109,111,112,41,0,109,101,109,98,101,114,32,85,82,76,32,105,115,32,97,98,115,111,108,117,116,101,44,32,116,114,121,32,111,112,101,110,32,82,47,87,32,40,102,102,103,109,111,112,41,0,79,75,44,32,110,111,119,32,116,114,121,32,116,111,32,111,112,101,110,32,114,101,97,100,45,111,110,108,121,32,40,102,102,103,109,111,112,41,0,77,101,109,98,101,114,32,85,82,76,32,105,115,32,111,102,32,116,121,112,101,32,70,73,76,69,32,40,102,102,103,109,111,112,41,0,77,101,109,98,101,114,32,85,82,76,32,115,112,101,99,105,102,105,101,115,32,97,98,115,32,102,105,108,101,32,112,97,116,104,32,40,102,102,103,109,111,112,41,0,84,114,121,32,116,111,32,111,112,101,110,32,109,101,109,98,101,114,32,85,82,76,32,105,110,32,82,47,87,32,109,111,100,101,32,40,102,102,103,109,111,112,41,0,84,114,121,32,116,111,32,111,112,101,110,32,109,101,109,98,101,114,32,102,105,108,101,32,97,115,32,114,101,108,97,116,105,118,101,32,85,82,76,32,40,102,102,103,109,111,112,41,0,110,111,119,32,116,114,121,32,116,111,32,111,112,101,110,32,102,105,108,101,32,97,115,32,82,69,65,68,79,78,76,89,32,40,102,102,103,109,111,112,41,0,67,97,110,110,111,116,32,111,112,101,110,32,109,101,109,98,101,114,32,72,68,85,32,70,73,84,83,32,102,105,108,101,32,40,102,102,103,109,111,112,41,0,67,97,110,110,111,116,32,102,105,110,100,32,115,112,101,99,105,102,105,101,100,32,109,101,109,98,101,114,32,72,68,85,32,40,102,102,103,109,111,112,41,0,102,105,108,101,58,47,47,0,83,112,101,99,105,102,105,101,100,32,72,68,85,32,105,115,32,97,108,114,101,97,100,121,32,97,32,109,101,109,98,101,114,32,111,102,32,116,104,101,32,71,114,111,117,112,105,110,103,32,116,97,98,108,101,32,40,102,102,103,116,97,109,41,0,99,97,110,110,111,116,32,97,100,100,32,71,82,80,73,68,47,76,67,32,107,101,121,119,111,114,100,115,32,116,111,32,109,101,109,98,101,114,32,72,68,85,58,32,40,102,102,103,116,97,109,41,0,0,0,0,0,0,0,0,0,0,0,15,14,0,15,15,12,15,15,15,15,15,15,15,15,15,15,8,0,0,0,0,0,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,0,0,0,0,15,0,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,0,0,0,0,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,0,72,68,85,32,97,108,114,101,97,100,121,32,104,97,115,32,71,82,80,73,68,47,76,67,32,107,101,121,119,111,114,100,115,32,102,111,114,32,103,114,111,117,112,32,116,97,98,108,101,32,40,102,102,103,116,97,109,41,0,69,88,84,86,69,82,32,111,102,32,71,114,111,117,112,32,99,111,110,116,97,105,110,105,110,103,32,116,104,105,115,32,72,68,85,0,85,82,76,32,111,102,32,102,105,108,101,32,99,111,110,116,97,105,110,105,110,103,32,71,114,111,117,112,0,71,82,80,73,68,35,0,71,82,80,73,68,32,105,110,100,101,120,32,37,100,32,108,97,114,103,101,114,32,116,111,116,97,108,32,71,82,80,73,68,32,107,101,121,119,111,114,100,115,32,37,108,100,32,40,102,102,103,116,111,112,41,0,73,110,118,97,108,105,100,32,118,97,108,117,101,32,111,102,32,37,108,100,32,102,111,114,32,71,82,80,73,68,37,100,32,40,102,102,103,116,111,112,41,0,67,97,110,110,111,116,32,102,105,110,100,32,71,82,80,76,67,37,100,32,107,101,121,119,111,114,100,32,40,102,102,103,116,111,112,41,0,84,114,121,32,116,111,32,111,112,101,110,32,103,114,111,117,112,32,116,97,98,108,101,32,102,105,108,101,32,97,115,32,97,98,115,111,108,117,116,101,32,85,82,76,32,40,102,102,103,116,111,112,41,0,79,75,44,32,116,114,121,32,111,112,101,110,32,103,114,111,117,112,32,116,97,98,108,101,32,102,105,108,101,32,97,115,32,82,69,65,68,79,78,76,89,32,40,102,102,103,116,111,112,41,0,111,112,101,110,105,110,103,32,102,105,108,101,32,97,115,32,82,69,65,68,87,82,73,84,69,32,102,97,105,108,101,100,32,40,102,102,103,116,111,112,41,0,79,75,44,32,116,114,121,32,116,111,32,111,112,101,110,32,102,105,108,101,32,97,115,32,82,69,65,68,79,78,76,89,32,40,102,102,103,116,111,112,41,0,67,97,110,110,111,116,32,111,112,101,110,32,111,114,32,102,105,110,100,32,103,114,111,117,112,105,110,103,32,116,97,98,108,101,32,70,73,84,83,32,102,105,108,101,32,40,102,102,103,116,111,112,41,0,58,47,47,0,99,97,110,110,111,116,32,109,97,107,101,32,85,82,76,32,102,114,111,109,32,116,101,109,112,32,77,69,77,58,47,47,32,102,105,108,101,32,40,102,105,116,115,95,103,101,116,95,117,114,108,41,0,109,101,109,107,101,101,112,58,47,47,0,115,104,109,101,109,58,47,47,0,114,111,111,116,58,47,47,0,99,111,109,112,114,101,115,115,102,105,108,101,58,47,47,0,104,116,116,112,102,105,108,101,58,47,47,0,102,116,112,102,105,108,101,58,47,47,0,102,116,112,58,47,47,0,115,116,100,105,110,102,105,108,101,58,47,47,0,115,116,100,105,110,58,47,47,0,99,111,109,112,114,101,115,115,58,47,47,0,104,116,116,112,58,47,47,0,104,116,116,112,99,111,109,112,114,101,115,115,58,47,47,0,102,116,112,99,111,109,112,114,101,115,115,58,47,47,0,99,97,110,110,111,116,32,109,97,107,101,32,118,97,105,108,100,32,85,82,76,32,102,114,111,109,32,115,116,100,105,110,58,47,47,32,40,102,105,116,115,95,103,101,116,95,117,114,108,41,0,115,116,100,111,117,116,58,47,47,0,99,97,110,110,111,116,32,109,97,107,101,32,118,97,105,108,100,32,85,82,76,32,102,114,111,109,32,115,116,100,111,117,116,58,47,47,32,40,102,105,116,115,95,103,101,116,95,117,114,108,41,0,105,114,97,102,109,101,109,58,47,47,0,99,97,110,110,111,116,32,109,97,107,101,32,118,97,105,108,100,32,85,82,76,32,102,114,111,109,32,105,114,97,102,109,101,109,58,47,47,32,40,102,105,116,115,95,103,101,116,95,117,114,108,41,0,77,69,77,58,0,83,72,77,69,77,58,0,114,101,102,32,85,82,76,32,104,97,115,32,97,99,99,101,115,115,32,109,101,109,58,47,47,32,111,114,32,115,104,109,101,109,58,47,47,32,40,102,105,116,115,95,114,101,108,117,114,108,50,117,114,108,41,0,32,32,32,99,97,110,110,111,116,32,99,111,110,115,116,114,117,99,116,32,102,117,108,108,32,85,82,76,32,102,114,111,109,32,97,32,112,97,114,116,105,97,108,32,85,82,76,32,97,110,100,32,0,32,32,32,77,69,77,47,83,72,77,69,77,32,98,97,115,101,32,85,82,76,0,46,46,0,67,97,110,110,111,116,32,109,97,107,101,32,114,101,108,46,32,85,82,76,32,102,114,111,109,32,110,111,110,32,97,98,115,46,32,85,82,76,115,32,40,102,105,116,115,95,117,114,108,50,114,101,108,117,114,108,41,0,78,85,76,76,32,76,111,99,97,116,105,111,110,32,115,116,114,105,110,103,32,103,105,118,101,110,32,61,61,62,32,105,110,103,111,114,101,32,108,111,99,97,116,105,111,110,32,40,102,102,103,109,102,41,0,67,97,110,110,111,116,32,102,105,110,100,32,115,112,101,99,105,102,105,101,100,32,109,101,109,98,101,114,32,72,68,85,32,40,102,102,103,109,102,41,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,72,73,69,82,65,82,67,72,0,92,73,78,67,76,85,68,69,0,67,70,73,84,83,73,79,95,73,78,67,76,85,68,69,95,70,73,76,69,83,0,58,0,84,0,70,0,40,37,108,103,44,37,108,103,41,37,110,0,37,108,103,37,110,0,37,100,37,110,0,69,88,84,86,69,82,0,37,45,56,46,56,115,37,115,0,84,70,79,82,77,37,100,37,99,0,84,84,89,80,69,37,100,37,99,0,78,65,88,73,83,37,100,37,99,0,97,117,116,111,32,97,115,115,105,103,110,101,100,32,98,121,32,116,101,109,112,108,97,116,101,32,112,97,114,115,101,114,0,68,69,70,65,85,76,84,95,71,82,79,85,80,95,37,100,0,92,71,82,79,85,80,0,92,69,78,68,0,98,105,110,110,105,110,103,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,115,121,110,116,97,120,32,101,114,114,111,114,58,0,32,44,41,0,98,105,110,110,105,110,103,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,104,97,115,32,116,111,111,32,109,97,110,121,32,99,111,108,117,109,110,32,110,97,109,101,115,32,111,114,32,105,115,32,109,105,115,115,105,110,103,32,99,108,111,115,105,110,103,32,39,41,39,58,0,105,108,108,101,103,97,108,32,98,105,110,110,105,110,103,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,32,97,110,32,101,113,117,97,108,115,32,115,105,103,110,32,39,61,39,32,109,117,115,116,32,102,111,108,108,111,119,32,116,104,101,32,99,111,108,117,109,110,32,110,97,109,101,115,0,105,108,108,101,103,97,108,32,115,121,110,116,97,120,32,97,102,116,101,114,32,98,105,110,110,105,110,103,32,114,97,110,103,101,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,105,108,108,101,103,97,108,32,115,121,110,116,97,120,32,105,110,32,98,105,110,110,105,110,103,32,114,97,110,103,101,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,105,108,108,101,103,97,108,32,99,104,97,114,97,99,116,101,114,115,32,102,111,108,108,111,119,105,110,103,32,98,105,110,110,105,110,103,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,97,112,112,97,114,101,110,116,108,121,32,103,114,101,97,116,101,114,32,116,104,97,110,32,52,32,104,105,115,116,111,103,114,97,109,32,100,105,109,101,110,115,105,111,110,115,0,105,108,108,101,103,97,108,32,98,105,110,110,105,110,103,32,119,101,105,103,104,116,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,105,108,108,101,103,97,108,32,115,121,110,116,97,120,32,97,102,116,101,114,32,98,105,110,110,105,110,103,32,119,101,105,103,104,116,32,115,112,101,99,105,102,105,99,97,116,105,111,110,32,105,110,32,85,82,76,58,0,32,44,61,58,59,0,32,44,58,59,0,104,105,115,116,111,103,114,97,109,32,104,97,115,32,109,111,114,101,32,116,104,97,110,32,52,32,100,105,109,101,110,115,105,111,110,115,0,102,97,105,108,101,100,32,116,111,32,100,101,116,101,114,109,105,110,101,32,98,105,110,110,105,110,103,32,112,97,114,97,109,101,116,101,114,115,0,107,101,121,119,111,114,100,32,111,114,32,99,111,108,117,109,110,32,102,111,114,32,104,105,115,116,111,103,114,97,109,32,119,101,105,103,104,116,115,32,100,111,101,115,110,39,116,32,101,120,105,115,116,58,32,0,73,108,108,101,103,97,108,32,104,105,115,116,111,103,114,97,109,109,105,110,103,32,119,101,105,103,104,116,105,110,103,32,102,97,99,116,111,114,32,60,61,32,48,46,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,116,101,109,112,32,111,117,116,112,117,116,32,102,105,108,101,32,102,111,114,32,104,105,115,116,111,103,114,97,109,0,102,97,105,108,101,100,32,116,111,32,99,114,101,97,116,101,32,111,117,116,112,117,116,32,104,105,115,116,111,103,114,97,109,32,70,73,84,83,32,105,109,97,103,101,0,102,97,105,108,101,100,32,116,111,32,99,111,112,121,32,112,105,120,101,108,32,108,105,115,116,32,107,101,121,119,111,114,100,115,32,116,111,32,110,101,119,32,104,105,115,116,111,103,114,97,109,32,104,101,97,100,101,114,0,102,97,105,108,101,100,32,116,111,32,99,97,108,99,117,108,97,116,101,32,110,101,119,32,104,105,115,116,111,103,114,97,109,32,118,97,108,117,101,115,0,104,105,115,116,111,103,114,97,109,115,32,119,105,116,104,32,109,111,114,101,32,116,104,97,110,32,52,32,100,105,109,101,110,115,105,111,110,115,32,97,114,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,67,80,82,69,70,0,99,111,108,117,109,110,32,102,111,114,32,104,105,115,116,111,103,114,97,109,32,97,120,105,115,32,100,111,101,115,110,39,116,32,101,120,105,115,116,58,32,0,67,97,110,39,116,32,98,105,110,32,97,32,118,101,99,116,111,114,32,99,111,108,117,109,110,58,32,0,73,110,97,112,112,114,111,112,114,105,97,116,101,32,100,97,116,97,116,121,112,101,59,32,99,97,110,39,116,32,98,105,110,32,116,104,105,115,32,99,111,108,117,109,110,58,32,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,104,105,115,116,111,103,114,97,109,109,105,110,103,32,109,105,110,105,109,117,109,32,107,101,121,119,111,114,100,0,69,114,114,111,114,32,99,97,108,99,117,108,97,116,105,110,103,32,100,97,116,97,109,105,110,32,97,110,100,32,100,97,116,97,109,97,120,32,102,111,114,32,99,111,108,117,109,110,58,32,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,104,105,115,116,111,103,114,97,109,109,105,110,103,32,109,97,120,105,109,117,109,32,107,101,121,119,111,114,100,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,104,105,115,116,111,103,114,97,109,109,105,110,103,32,98,105,110,115,105,122,101,32,107,101,121,119,111,114,100,0,101,114,114,111,114,58,32,104,105,115,116,111,103,114,97,109,32,98,105,110,115,105,122,101,32,61,32,48,0,84,68,66,73,78,0,67,84,89,80,69,0,67,111,111,114,100,105,110,97,116,101,32,84,121,112,101,0,67,85,78,73,84,0,67,111,111,114,100,105,110,97,116,101,32,85,110,105,116,115,0,67,82,80,73,88,0,82,101,102,101,114,101,110,99,101,32,80,105,120,101,108,0,67,82,86,65,76,0,82,101,102,101,114,101,110,99,101,32,86,97,108,117,101,0,67,68,69,76,84,0,80,105,120,101,108,32,115,105,122,101,0,67,68,0,73,82,65,70,50,83,84,82,32,67,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,115,116,114,105,110,103,32,118,97,114,105,97,98,108,101,0,105,109,104,100,114,0,105,109,104,118,50,0,70,105,108,101,32,110,111,116,32,118,97,108,105,100,32,73,82,65,70,32,105,109,97,103,101,32,104,101,97,100,101,114,0,73,82,65,70,71,69,84,67,32,67,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,115,116,114,105,110,103,32,118,97,114,105,97,98,108,101,0,72,68,82,0,69,78,68,0,73,82,65,70,50,70,73,84,83,32,67,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,37,100,45,98,121,116,101,32,70,73,84,83,32,104,101,97,100,101,114,0,85,110,115,117,112,112,111,114,116,101,100,32,73,82,65,70,32,100,97,116,97,32,116,121,112,101,58,32,37,100,0,73,82,65,70,32,46,105,109,104,32,112,105,120,101,108,32,116,121,112,101,0,73,82,65,70,32,46,105,109,104,32,110,97,120,105,115,0,73,82,65,70,32,46,105,109,104,32,105,109,97,103,101,32,110,97,120,105,115,91,49,93,0,73,82,65,70,32,46,105,109,104,32,105,109,97,103,101,32,110,97,120,105,115,91,50,93,0,78,65,88,73,83,51,0,73,82,65,70,32,46,105,109,104,32,105,109,97,103,101,32,110,97,120,105,115,91,51,93,0,78,65,88,73,83,52,0,73,82,65,70,32,46,105,109,104,32,105,109,97,103,101,32,110,97,120,105,115,91,52,93,0,79,66,74,69,67,84,0,73,82,65,70,32,46,105,109,104,32,116,105,116,108,101,0,78,80,65,88,73,83,49,0,73,82,65,70,32,46,105,109,104,32,112,104,121,115,105,99,97,108,32,110,97,120,105,115,91,49,93,0,78,80,65,88,73,83,50,0,73,82,65,70,32,46,105,109,104,32,112,104,121,115,105,99,97,108,32,110,97,120,105,115,91,50,93,0,78,80,65,88,73,83,51,0,73,82,65,70,32,46,105,109,104,32,112,104,121,115,105,99,97,108,32,110,97,120,105,115,91,51,93,0,78,80,65,88,73,83,52,0,73,82,65,70,32,46,105,109,104,32,112,104,121,115,105,99,97,108,32,110,97,120,105,115,91,52,93,0,73,77,72,70,73,76,69,0,73,82,65,70,32,104,101,97,100,101,114,32,102,105,108,101,32,110,97,109,101,0,80,73,88,70,73,76,69,0,73,82,65,70,32,46,112,105,120,32,112,105,120,101,108,32,102,105,108,101,0,80,73,88,79,70,70,0,73,82,65,70,32,46,112,105,120,32,112,105,120,101,108,32,111,102,102,115,101,116,32,40,68,111,32,110,111,116,32,99,104,97,110,103,101,33,41,0,73,77,72,86,69,82,0,73,82,65,70,32,46,105,109,104,32,102,111,114,109,97,116,32,118,101,114,115,105,111,110,32,40,49,32,111,114,32,50,41,0,80,73,88,83,87,65,80,0,73,82,65,70,32,112,105,120,101,108,115,44,32,70,73,84,83,32,98,121,116,101,32,111,114,100,101,114,115,32,100,105,102,102,101,114,32,105,102,32,84,0,79,66,74,69,67,84,32,0,32,32,32,0,114,98,0,73,82,65,70,82,73,77,65,71,69,58,32,67,97,110,110,111,116,32,111,112,101,110,32,73,82,65,70,32,112,105,120,101,108,32,102,105,108,101,58,0,73,82,65,70,82,73,77,65,71,69,58,32,67,97,110,110,111,116,32,97,108,108,111,99,32,109,101,109,111,114,121,32,102,111,114,32,112,105,120,101,108,32,104,101,97,100,101,114,0,73,82,65,70,32,112,105,120,101,108,32,102,105,108,101,58,32,37,100,32,47,32,37,100,32,98,121,116,101,115,32,114,101,97,100,46,0,105,109,112,105,120,0,105,109,112,118,50,0,70,105,108,101,32,110,111,116,32,118,97,108,105,100,32,73,82,65,70,32,112,105,120,101,108,32,102,105,108,101,58,0,73,82,65,70,82,73,77,65,71,69,32,67,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,37,100,45,98,121,116,101,32,105,109,97,103,101,32,98,117,102,102,101,114,0,47,0,45,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,105,114,97,102,102,105,116,115,32,115,97,109,101,95,112,97,116,104,58,32,67,97,110,110,111,116,32,97,108,108,111,99,32,109,101,109,111,114,121,32,102,111,114,32,110,101,119,112,105,120,110,97,109,101,0,72,68,82,36,0,117,110,97,98,108,101,32,116,111,32,111,112,101,110,32,73,82,65,70,32,104,101,97,100,101,114,32,102,105,108,101,58,0,73,82,65,70,82,72,69,65,68,58,32,99,97,110,110,111,116,32,115,101,101,107,32,105,110,32,102,105,108,101,58,0,73,82,65,70,82,72,69,65,68,58,32,99,97,110,110,111,116,32,103,101,116,32,112,111,115,46,32,105,110,32,102,105,108,101,58,0,73,82,65,70,82,72,69,65,68,58,32,99,97,110,110,111,116,32,115,101,101,107,32,116,111,32,98,101,103,105,110,110,105,110,103,32,111,102,32,102,105,108,101,58,0,73,82,65,70,82,72,69,65,68,32,67,97,110,110,111,116,32,97,108,108,111,99,97,116,101,32,37,100,45,98,121,116,101,32,104,101,97,100,101,114,0,73,82,65,70,82,72,69,65,68,32,104,101,97,100,101,114,32,102,105,108,101,58,32,37,100,32,47,32,37,100,32,98,121,116,101,115,32,114,101,97,100,46,0,67,97,110,110,111,116,32,100,101,108,101,116,101,32,107,101,121,119,111,114,100,32,110,117,109,98,101,114,32,37,100,46,32,32,73,116,32,100,111,101,115,32,110,111,116,32,101,120,105,115,116,46,0,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,116,104,101,32,37,115,32,107,101,121,119,111,114,100,32,116,111,32,100,101,108,101,116,101,32,40,102,102,100,107,101,121,41,0,73,108,108,101,103,97,108,32,110,117,109,98,101,114,32,111,102,32,99,111,108,117,109,109,115,32,40,102,102,105,116,101,114,41,0,86,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,97,114,114,97,121,32,110,111,116,32,97,108,108,111,119,101,100,32,102,111,114,32,111,117,116,112,117,116,32,99,111,108,117,109,110,32,110,117,109,98,101,114,32,37,100,32,40,102,102,105,116,101,114,41,0,73,108,108,101,103,97,108,32,100,97,116,97,116,121,112,101,32,102,111,114,32,99,111,108,117,109,110,32,110,117,109,98,101,114,32,37,100,58,32,37,100,32,32,40,102,102,105,116,101,114,41,0,70,105,108,101,32,37,100,32,110,111,116,32,112,111,115,105,116,105,111,110,101,100,32,116,111,32,97,110,32,105,109,97,103,101,32,101,120,116,101,110,115,105,111,110,32,40,102,102,105,116,101,114,41,0,66,85,78,73,84,0,70,105,108,101,32,37,100,32,110,111,116,32,112,111,115,105,116,105,111,110,101,100,32,116,111,32,97,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,32,40,102,102,105,116,101,114,41,0,67,111,108,117,109,110,32,39,37,115,39,32,110,111,116,32,102,111,117,110,100,32,102,111,114,32,99,111,108,117,109,110,32,110,117,109,98,101,114,32,37,100,32,32,40,102,102,105,116,101,114,41,0,67,111,108,117,109,110,32,37,100,32,104,97,115,32,105,108,108,101,103,97,108,32,116,97,98,108,101,32,112,111,115,105,116,105,111,110,32,110,117,109,98,101,114,58,32,37,100,32,32,40,102,102,105,116,101,114,41,0,84,76,77,73,78,0,84,76,77,65,88,0,84,68,73,83,80,0,102,102,105,116,101,114,32,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,110,117,108,108,32,118,97,108,117,101,115,0,84,78,85,76,76,0,102,102,105,116,101,114,32,102,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,97,114,114,97,121,115,0,67,111,108,117,109,110,32,37,100,32,100,97,116,97,116,121,112,101,32,99,117,114,114,101,110,116,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,58,32,37,100,58,32,32,40,102,102,105,116,101,114,41,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,98,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,100,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,101,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,105,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,106,41,46,0,119,114,105,116,105,110,103,32,116,111,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,97,114,114,97,121,32,111,102,32,108,111,103,105,99,97,108,115,32,40,102,102,112,99,108,108,41,46,0,128,64,32,16,8,4,2,1,127,191,223,239,247,251,253,254,83,112,101,99,105,102,105,101,100,32,99,111,108,117,109,110,32,110,117,109,98,101,114,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,58,32,37,100,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,116,111,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,115,116,114,105,110,103,32,99,111,108,117,109,110,32,40,102,102,112,99,108,115,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,115,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,115,98,41,46,0,78,117,108,108,32,118,97,108,117,101,32,115,116,114,105,110,103,32,102,111,114,32,65,83,67,73,73,32,116,97,98,108,101,32,99,111,108,117,109,110,32,105,115,32,110,111,116,32,100,101,102,105,110,101,100,32,40,70,84,80,67,76,85,41,46,0,78,117,108,108,32,118,97,108,117,101,32,102,111,114,32,105,110,116,101,103,101,114,32,116,97,98,108,101,32,99,111,108,117,109,110,32,105,115,32,110,111,116,32,100,101,102,105,110,101,100,32,40,70,84,80,67,76,85,41,46,0,67,97,110,110,111,116,32,119,114,105,116,101,32,110,117,108,108,32,118,97,108,117,101,32,116,111,32,99,111,108,117,109,110,32,37,100,32,119,104,105,99,104,32,104,97,115,32,102,111,114,109,97,116,32,37,115,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,110,117,108,108,32,118,97,108,117,101,115,32,40,102,102,112,99,108,117,41,46,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,117,105,41,46,0,67,97,110,110,111,116,32,119,114,105,116,101,32,110,117,109,98,101,114,115,32,116,111,32,99,111,108,117,109,110,32,37,100,32,119,104,105,99,104,32,104,97,115,32,102,111,114,109,97,116,32,37,115,0,69,114,114,111,114,32,119,114,105,116,105,110,103,32,101,108,101,109,101,110,116,115,32,37,46,48,102,32,116,104,114,117,32,37,46,48,102,32,111,102,32,105,110,112,117,116,32,100,97,116,97,32,97,114,114,97,121,32,40,102,102,112,99,108,117,106,41,46,0,78,117,109,101,114,105,99,97,108,32,111,118,101,114,102,108,111,119,32,100,117,114,105,110,103,32,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,119,104,105,108,101,32,119,114,105,116,105,110,103,32,70,73,84,83,32,100,97,116,97,46,0,102,105,108,101,32,100,111,101,115,32,110,111,116,32,99,111,110,102,111,114,109,32,116,111,32,70,73,84,83,32,115,116,97,110,100,97,114,100,0,73,77,65,71,69,32,101,120,116,101,110,115,105,111,110,0,73,77,65,71,69,0,73,108,108,101,103,97,108,32,118,97,108,117,101,32,102,111,114,32,66,73,84,80,73,88,32,107,101,121,119,111,114,100,58,32,37,100,0,110,117,109,98,101,114,32,111,102,32,98,105,116,115,32,112,101,114,32,100,97,116,97,32,112,105,120,101,108,0,37,108,108,100,0,69,114,114,111,114,32,105,110,32,102,102,105,50,99,32,99,111,110,118,101,114,116,105,110,103,32,105,110,116,101,103,101,114,32,116,111,32,115,116,114,105,110,103,0,73,108,108,101,103,97,108,32,118,97,108,117,101,32,102,111,114,32,78,65,88,73,83,32,107,101,121,119,111,114,100,58,32,37,100,0,110,117,109,98,101,114,32,111,102,32,100,97,116,97,32,97,120,101,115,0,108,101,110,103,116,104,32,111,102,32,100,97,116,97,32,97,120,105,115,32,0,73,108,108,101,103,97,108,32,110,101,103,97,116,105,118,101,32,118,97,108,117,101,32,102,111,114,32,78,65,88,73,83,37,100,32,107,101,121,119,111,114,100,58,32,37,46,48,102,0,37,100,0,70,73,84,83,32,100,97,116,97,115,101,116,32,109,97,121,32,99,111,110,116,97,105,110,32,101,120,116,101,110,115,105,111,110,115,0,112,99,111,117,110,116,32,118,97,108,117,101,32,105,115,32,108,101,115,115,32,116,104,97,110,32,48,0,103,99,111,117,110,116,32,118,97,108,117,101,32,105,115,32,108,101,115,115,32,116,104,97,110,32,49,0,114,97,110,100,111,109,32,103,114,111,117,112,32,114,101,99,111,114,100,115,32,97,114,101,32,112,114,101,115,101,110,116,0,71,82,79,85,80,83,0,110,117,109,98,101,114,32,111,102,32,114,97,110,100,111,109,32,103,114,111,117,112,32,112,97,114,97,109,101,116,101,114,115,0,110,117,109,98,101,114,32,111,102,32,114,97,110,100,111,109,32,103,114,111,117,112,115,0,67,79,77,77,69,78,84,32,32,32,70,73,84,83,32,40,70,108,101,120,105,98,108,101,32,73,109,97,103,101,32,84,114,97,110,115,112,111,114,116,32,83,121,115,116,101,109,41,32,102,111,114,109,97,116,32,105,115,32,100,101,102,105,110,101,100,32,105,110,32,39,65,115,116,114,111,110,111,109,121,0,67,79,77,77,69,78,84,32,32,32,97,110,100,32,65,115,116,114,111,112,104,121,115,105,99,115,39,44,32,118,111,108,117,109,101,32,51,55,54,44,32,112,97,103,101,32,51,53,57,59,32,98,105,98,99,111,100,101,58,32,50,48,48,49,65,38,65,46,46,46,51,55,54,46,46,51,53,57,72,0,105,109,97,103,101,32,101,120,116,101,110,115,105,111,110,115,32,109,117,115,116,32,104,97,118,101,32,112,99,111,117,110,116,32,61,32,48,0,105,109,97,103,101,32,101,120,116,101,110,115,105,111,110,115,32,109,117,115,116,32,104,97,118,101,32,103,99,111,117,110,116,32,61,32,49,0,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,59,32,109,117,115,116,32,61,32,48,0,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,59,32,109,117,115,116,32,61,32,49,0,37,46,42,102,0,69,114,114,111,114,32,105,110,32,102,102,100,50,102,32,99,111,110,118,101,114,116,105,110,103,32,100,111,117,98,108,101,32,116,111,32,115,116,114,105,110,103,0,69,114,114,111,114,32,105,110,32,102,102,100,50,102,58,32,100,111,117,98,108,101,32,118,97,108,117,101,32,105,115,32,97,32,78,97,78,32,111,114,32,73,78,68,69,70,0,66,73,78,84,65,66,76,69,0,98,105,110,97,114,121,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,0,56,45,98,105,116,32,98,121,116,101,115,0,50,45,100,105,109,101,110,115,105,111,110,97,108,32,98,105,110,97,114,121,32,116,97,98,108,101,0,78,65,88,73,83,49,0,119,105,100,116,104,32,111,102,32,116,97,98,108,101,32,105,110,32,98,121,116,101,115,0,110,117,109,98,101,114,32,111,102,32,114,111,119,115,32,105,110,32,116,97,98,108,101,0,115,105,122,101,32,111,102,32,115,112,101,99,105,97,108,32,100,97,116,97,32,97,114,101,97,0,111,110,101,32,100,97,116,97,32,103,114,111,117,112,32,40,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,41,0,110,117,109,98,101,114,32,111,102,32,102,105,101,108,100,115,32,105,110,32,101,97,99,104,32,114,111,119,0,108,97,98,101,108,32,102,111,114,32,102,105,101,108,100,32,37,51,100,0,84,84,89,80,69,0,84,70,79,82,77,0,100,97,116,97,32,102,111,114,109,97,116,32,111,102,32,102,105,101,108,100,0,58,32,65,83,67,73,73,32,67,104,97,114,97,99,116,101,114,0,37,108,100,0,69,82,82,79,82,63,63,32,32,85,83,73,78,71,32,65,83,67,73,73,32,84,65,66,76,69,32,83,89,78,84,65,88,32,66,89,32,77,73,83,84,65,75,69,63,63,0,114,65,119,32,70,79,82,77,65,84,32,69,82,82,79,82,33,32,85,78,73,84,32,87,73,68,84,72,32,119,32,62,32,67,79,76,85,77,78,32,87,73,68,84,72,32,114,0,58,32,66,73,84,0,58,32,66,89,84,69,0,58,32,49,45,98,121,116,101,32,76,79,71,73,67,65,76,0,58,32,50,45,98,121,116,101,32,73,78,84,69,71,69,82,0,58,32,52,45,98,121,116,101,32,73,78,84,69,71,69,82,0,58,32,56,45,98,121,116,101,32,73,78,84,69,71,69,82,0,58,32,52,45,98,121,116,101,32,82,69,65,76,0,58,32,56,45,98,121,116,101,32,68,79,85,66,76,69,0,58,32,67,79,77,80,76,69,88,0,58,32,68,79,85,66,76,69,32,67,79,77,80,76,69,88,0,58,32,118,97,114,105,97,98,108,101,32,108,101,110,103,116,104,32,97,114,114,97,121,0,84,90,69,82,79,0,111,102,102,115,101,116,32,102,111,114,32,115,105,103,110,101,100,32,98,121,116,101,115,0,84,83,67,65,76,0,100,97,116,97,32,97,114,101,32,110,111,116,32,115,99,97,108,101,100,0,111,102,102,115,101,116,32,102,111,114,32,117,110,115,105,103,110,101,100,32,105,110,116,101,103,101,114,115,0,84,85,78,73,84,0,112,104,121,115,105,99,97,108,32,117,110,105,116,32,111,102,32,102,105,101,108,100,0,110,97,109,101,32,111,102,32,116,104,105,115,32,98,105,110,97,114,121,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,0,70,97,105,108,101,100,32,116,111,32,119,114,105,116,101,32,98,105,110,97,114,121,32,116,97,98,108,101,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,32,40,102,102,112,104,98,110,41,0,84,65,66,76,69,0,65,83,67,73,73,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,0,56,45,98,105,116,32,65,83,67,73,73,32,99,104,97,114,97,99,116,101,114,115,0,50,45,100,105,109,101,110,115,105,111,110,97,108,32,65,83,67,73,73,32,116,97,98,108,101,0,119,105,100,116,104,32,111,102,32,116,97,98,108,101,32,105,110,32,99,104,97,114,97,99,116,101,114,115,0,110,111,32,103,114,111,117,112,32,112,97,114,97,109,101,116,101,114,115,32,40,114,101,113,117,105,114,101,100,32,107,101,121,119,111,114,100,41,0,98,101,103,105,110,110,105,110,103,32,99,111,108,117,109,110,32,111,102,32,102,105],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+763432);allocate([101,108,100,32,37,51,100,0,84,66,67,79,76,0,70,111,114,116,114,97,110,45,55,55,32,102,111,114,109,97,116,32,111,102,32,102,105,101,108,100,0,110,97,109,101,32,111,102,32,116,104,105,115,32,65,83,67,73,73,32,116,97,98,108,101,32,101,120,116,101,110,115,105,111,110,0,70,97,105,108,101,100,32,116,111,32,119,114,105,116,101,32,65,83,67,73,73,32,116,97,98,108,101,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,32,40,102,102,112,104,116,98,41,0,102,102,112,107,116,112,32,99,111,117,108,100,32,110,111,116,32,111,112,101,110,32,116,104,101,32,102,111,108,108,111,119,105,110,103,32,116,101,109,112,108,97,116,101,32,102,105,108,101,58,0,61,0,67,79,77,77,69,78,84,32,0,72,73,83,84,79,82,89,32,0,32,32,32,32,32,32,32,32,0,67,79,78,84,73,78,85,69,0,66,97,100,32,107,101,121,119,111,114,100,32,100,97,116,97,116,121,112,101,32,99,111,100,101,58,32,37,100,32,40,102,102,112,107,121,41,0,39,39,0,44,32,0,76,79,78,71,83,84,82,78,0,79,71,73,80,32,49,46,48,0,84,104,101,32,72,69,65,83,65,82,67,32,76,111,110,103,32,83,116,114,105,110,103,32,67,111,110,118,101,110,116,105,111,110,32,109,97,121,32,98,101,32,117,115,101,100,46,0,32,32,84,104,105,115,32,70,73,84,83,32,102,105,108,101,32,109,97,121,32,99,111,110,116,97,105,110,32,108,111,110,103,32,115,116,114,105,110,103,32,107,101,121,119,111,114,100,32,118,97,108,117,101,115,32,116,104,97,116,32,97,114,101,0,32,32,99,111,110,116,105,110,117,101,100,32,111,118,101,114,32,109,117,108,116,105,112,108,101,32,107,101,121,119,111,114,100,115,46,32,32,84,104,101,32,72,69,65,83,65,82,67,32,99,111,110,118,101,110,116,105,111,110,32,117,115,101,115,32,116,104,101,32,38,0,32,32,99,104,97,114,97,99,116,101,114,32,97,116,32,116,104,101,32,101,110,100,32,111,102,32,101,97,99,104,32,115,117,98,115,116,114,105,110,103,32,119,104,105,99,104,32,105,115,32,116,104,101,110,32,99,111,110,116,105,110,117,101,100,0,32,32,111,110,32,116,104,101,32,110,101,120,116,32,107,101,121,119,111,114,100,32,119,104,105,99,104,32,104,97,115,32,116,104,101,32,110,97,109,101,32,67,79,78,84,73,78,85,69,46,0,37,46,42,71,0,69,114,114,111,114,32,105,110,32,102,102,114,50,101,32,99,111,110,118,101,114,116,105,110,103,32,102,108,111,97,116,32,116,111,32,115,116,114,105,110,103,0,37,46,49,69,0,37,46,42,69,0,69,114,114,111,114,32,105,110,32,102,102,114,50,101,58,32,102,108,111,97,116,32,118,97,108,117,101,32,105,115,32,97,32,78,97,78,32,111,114,32,73,78,68,69,70,0,69,114,114,111,114,32,105,110,32,102,102,100,50,102,58,32,32,110,111,46,32,111,102,32,100,101,99,105,109,97,108,32,112,108,97,99,101,115,32,60,32,48,0,69,114,114,111,114,32,105,110,32,102,102,100,50,101,32,99,111,110,118,101,114,116,105,110,103,32,102,108,111,97,116,32,116,111,32,115,116,114,105,110,103,0,69,114,114,111,114,32,105,110,32,102,102,100,50,101,58,32,100,111,117,98,108,101,32,118,97,108,117,101,32,105,115,32,97,32,78,97,78,32,111,114,32,73,78,68,69,70,0,99,111,108,117,109,110,32,110,117,109,98,101,114,32,105,115,32,111,117,116,32,111,102,32,114,97,110,103,101,32,49,32,45,32,57,57,57,32,40,102,102,112,116,100,109,41,0,110,97,120,105,115,32,105,115,32,108,101,115,115,32,116,104,97,110,32,49,32,40,102,102,112,116,100,109,41,0,69,114,114,111,114,58,32,84,104,101,32,84,68,73,77,110,32,107,101,121,119,111,114,100,32,105,115,32,111,110,108,121,32,97,108,108,111,119,101,100,32,105,110,32,66,73,78,84,65,66,76,69,32,101,120,116,101,110,115,105,111,110,115,32,40,102,102,112,116,100,109,41,0,111,110,101,32,111,114,32,109,111,114,101,32,84,68,73,77,32,118,97,108,117,101,115,32,97,114,101,32,108,101,115,115,32,116,104,97,110,32,48,32,40,102,102,112,116,100,109,41,0,99,111,108,117,109,110,32,118,101,99,116,111,114,32,108,101,110,103,116,104,44,32,37,108,100,44,32,100,111,101,115,32,110,111,116,32,101,113,117,97,108,32,84,68,73,77,110,32,97,114,114,97,121,32,115,105,122,101,44,32,37,108,100,0,115,105,122,101,32,111,102,32,116,104,101,32,109,117,108,116,105,100,105,109,101,110,115,105,111,110,97,108,32,97,114,114,97,121,0,84,68,73,77,0,67,111,117,108,100,110,39,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,116,111,32,104,111,108,100,32,82,101,103,105,111,110,32,102,105,108,101,32,99,111,110,116,101,110,116,115,46,0,114,0,67,111,117,108,100,32,110,111,116,32,111,112,101,110,32,82,101,103,105,111,110,32,102,105,108,101,32,37,115,46,0,102,111,114,109,97,116,58,0,70,111,114,109,97,116,32,99,111,100,101,32,101,110,99,111,117,110,116,101,114,101,100,32,97,102,116,101,114,32,114,101,97,100,105,110,103,32,49,32,111,114,32,109,111,114,101,32,115,104,97,112,101,115,46,0,112,105,120,101,108,0,100,101,103,114,101,101,0,104,104,109,109,115,115,0,104,109,115,0,85,110,107,110,111,119,110,32,102,111,114,109,97,116,32,99,111,100,101,32,101,110,99,111,117,110,116,101,114,101,100,32,105,110,32,114,101,103,105,111,110,32,102,105,108,101,46,0,103,108,111,98,0,69,114,114,111,114,32,114,101,97,100,105,110,103,32,82,101,103,105,111,110,32,102,105,108,101,0,105,109,97,103,101,59,0,112,104,121,115,105,99,97,108,59,0,108,105,110,101,97,114,59,0,102,107,52,59,0,102,107,53,59,0,105,99,114,115,59,0,102,107,53,0,102,107,52,0,105,99,114,115,0,105,109,97,103,101,0,112,104,121,115,105,99,97,108,0,103,97,108,97,99,116,105,99,59,0,71,97,108,97,99,116,105,99,32,114,101,103,105,111,110,32,99,111,111,114,100,105,110,97,116,101,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,101,99,108,105,112,116,105,99,59,0,101,99,108,105,112,116,105,99,32,114,101,103,105,111,110,32,99,111,111,114,100,105,110,97,116,101,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,70,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,82,101,103,105,111,110,32,100,97,116,97,0,99,105,114,99,108,101,0,97,110,110,117,108,117,115,0,101,108,108,105,112,115,101,0,101,108,108,105,112,116,97,110,110,117,108,117,115,0,98,111,120,0,114,111,116,98,111,120,0,114,101,99,116,97,110,103,108,101,0,114,111,116,114,101,99,116,97,110,103,108,101,0,100,105,97,109,111,110,100,0,114,111,116,100,105,97,109,111,110,100,0,114,104,111,109,98,117,115,0,114,111,116,114,104,111,109,98,117,115,0,115,101,99,116,111,114,0,112,105,101,0,112,111,105,110,116,0,108,105,110,101,0,112,111,108,121,103,111,110,0,112,97,110,100,97,0,101,112,97,110,100,97,0,98,112,97,110,100,97,0,85,110,114,101,99,111,103,110,105,122,101,100,32,114,101,103,105,111,110,32,102,111,117,110,100,32,105,110,32,114,101,103,105,111,110,32,102,105,108,101,58,0,87,114,111,110,103,32,110,117,109,98,101,114,32,111,102,32,112,97,114,97,109,101,116,101,114,115,32,102,111,117,110,100,32,102,111,114,32,114,101,103,105,111,110,0,67,111,117,108,100,32,110,111,116,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,116,111,32,104,111,108,100,32,112,111,108,121,103,111,110,32,112,97,114,97,109,101,116,101,114,115,0,87,67,83,32,105,110,102,111,114,109,97,116,105,111,110,32,110,101,101,100,101,100,32,116,111,32,99,111,110,118,101,114,116,32,114,101,103,105,111,110,32,99,111,111,114,100,105,110,97,116,101,115,46,0,69,114,114,111,114,32,99,111,110,118,101,114,116,105,110,103,32,114,101,103,105,111,110,32,116,111,32,112,105,120,101,108,32,99,111,111,114,100,105,110,97,116,101,115,46,0,88,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,89,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,83,72,65,80,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,79,84,65,78,71,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,67,79,77,80,79,78,69,78,84,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,69,71,73,79,78,0,67,111,117,108,100,32,110,111,116,32,109,111,118,101,32,116,111,32,82,69,71,73,79,78,32,101,120,116,101,110,115,105,111,110,46,0,67,111,117,108,100,32,110,111,116,32,114,101,97,100,32,78,65,88,73,83,50,32,107,101,121,119,111,114,100,46,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,99,111,108,117,109,110,46,0,70,97,105,108,101,100,32,116,111,32,97,108,108,111,99,97,116,101,32,109,101,109,111,114,121,32,102,111,114,32,82,101,103,105,111,110,32,87,67,83,32,100,97,116,97,0,67,111,117,108,100,32,110,111,116,32,102,105,110,100,32,118,101,99,116,111,114,32,115,105,122,101,32,111,102,32,99,111,108,117,109,110,46,0,32,0,67,111,117,108,100,32,110,111,116,32,114,101,97,100,32,115,104,97,112,101,46,0,80,79,73,78,84,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,67,73,82,67,76,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,69,76,76,73,80,83,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65,78,78,85,76,85,83,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,69,76,76,73,80,84,65,78,78,85,76,85,83,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,66,79,88,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,79,84,66,79,88,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,66,79,88,65,78,78,85,76,85,83,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,69,67,84,65,78,71,76,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,79,84,82,69,67,84,65,78,71,76,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,80,79,76,89,71,79,78,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,80,73,69,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,83,69,67,84,79,82,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,68,73,65,77,79,78,68,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,72,79,77,66,85,83,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,79,84,68,73,65,77,79,78,68,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,82,79,84,82,72,79,77,66,85,83,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,88,32,99,111,108,117,109,110,32,102,111,114,32,112,111,108,121,103,111,110,32,114,101,103,105,111,110,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,89,32,99,111,108,117,109,110,32,102,111,114,32,112,111,108,121,103,111,110,32,114,101,103,105,111,110,0,70,97,105,108,101,100,32,116,111,32,116,114,97,110,115,102,111,114,109,32,99,111,111,114,100,105,110,97,116,101,115,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,82,32,99,111,108,117,109,110,32,102,111,114,32,114,101,103,105,111,110,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,82,79,84,65,78,71,32,99,111,108,117,109,110,32,102,111,114,32,114,101,103,105,111,110,0,70,97,105,108,101,100,32,116,111,32,114,101,97,100,32,67,79,77,80,79,78,69,78,84,32,99,111,108,117,109,110,32,102,111,114,32,114,101,103,105,111,110,0,69,114,114,111,114,32,105,110,32,102,102,103,105,119,99,115,46,32,84,104,105,115,32,72,68,85,32,105,115,32,110,111,116,32,97,110,32,105,109,97,103,101,46,32,67,97,110,39,116,32,114,101,97,100,32,87,67,83,32,107,101,121,119,111,114,100,115,0,101,114,114,111,114,32,99,114,101,97,116,105,110,103,32,115,116,114,105,110,103,32,111,102,32,105,109,97,103,101,32,87,67,83,32,107,101,121,119,111,114,100,115,32,40,102,102,103,105,119,99,115,41,0,67,82,86,65,76,49,0,67,82,86,65,76,50,0,67,82,80,73,88,49,0,67,82,80,73,88,50,0,67,68,69,76,84,49,0,67,68,49,95,49,0,67,68,50,95,49,0,67,68,49,95,50,0,67,68,50,95,50,0,67,68,69,76,84,50,0,67,82,79,84,65,50,0,80,67,49,95,49,0,80,67,50,95,49,0,80,67,49,95,50,0,80,67,50,95,50,0,67,84,89,80,69,49,0,68,69,67,45,0,76,65,84,0,109,101,109,58,47,47,0,102,102,103,116,99,115,32,99,111,117,108,100,32,110,111,116,32,102,105,110,100,32,97,108,108,32,116,104,101,32,99,101,108,101,115,116,105,97,108,32,99,111,111,114,100,105,110,97,116,101,32,107,101,121,119,111,114,100,115,0,102,105,116,115,95,105,110,105,116,95,114,97,110,100,111,109,115,32,103,101,110,101,114,97,116,101,100,32,105,110,99,111,114,114,101,99,116,32,114,97,110,100,111,109,32,110,117,109,98,101,114,32,115,101,113,117,101,110,99,101,0,117,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,105,111,110,32,97,108,103,111,114,105,116,104,109,32,40,102,105,116,115,95,115,101,116,95,99,111,109,112,114,101,115,115,105,111,110,95,116,121,112,101,41,0,105,108,108,101,103,97,108,32,110,117,109,98,101,114,32,111,102,32,116,105,108,101,32,100,105,109,101,110,115,105,111,110,115,32,40,102,105,116,115,95,115,101,116,95,116,105,108,101,95,100,105,109,41,0,105,108,108,101,103,97,108,32,100,105,116,104,101,114,105,110,103,32,118,97,108,117,101,32,40,102,105,116,115,95,115,101,116,95,113,117,97,110,116,105,122,101,95,109,101,116,104,111,100,41,0,76,111,115,115,108,101,115,115,32,99,111,109,112,114,101,115,115,105,111,110,32,111,102,32,102,108,111,97,116,105,110,103,32,112,111,105,110,116,32,105,109,97,103,101,115,32,109,117,115,116,32,117,115,101,32,71,90,73,80,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,72,99,111,109,112,114,101,115,115,32,99,97,110,110,111,116,32,98,101,32,117,115,101,100,32,119,105,116,104,32,49,45,100,105,109,101,110,115,105,111,110,97,108,32,105,109,97,103,101,115,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,72,99,111,109,112,114,101,115,115,32,109,105,110,105,109,117,109,32,105,109,97,103,101,32,100,105,109,101,110,115,105,111,110,32,105,115,32,52,32,112,105,120,101,108,115,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,72,99,111,109,112,114,101,115,115,32,109,105,110,105,109,117,109,32,116,105,108,101,32,100,105,109,101,110,115,105,111,110,32,105,115,32,52,32,112,105,120,101,108,115,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,76,97,115,116,32,116,105,108,101,32,97,108,111,110,103,32,49,115,116,32,100,105,109,101,110,115,105,111,110,32,104,97,115,32,108,101,115,115,32,116,104,97,110,32,52,32,112,105,120,101,108,115,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,76,97,115,116,32,116,105,108,101,32,97,108,111,110,103,32,50,110,100,32,100,105,109,101,110,115,105,111,110,32,104,97,115,32,108,101,115,115,32,116,104,97,110,32,52,32,112,105,120,101,108,115,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,49,68,0,82,73,67,69,95,49,0,71,90,73,80,95,49,0,71,90,73,80,95,50,0,80,76,73,79,95,49,0,72,67,79,77,80,82,69,83,83,95,49,0,78,79,67,79,77,80,82,69,83,83,0,117,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,105,111,110,32,116,121,112,101,32,40,105,109,99,111,109,112,95,105,110,105,116,95,116,97,98,108,101,41,0,90,73,77,65,71,69,0,101,120,116,101,110,115,105,111,110,32,99,111,110,116,97,105,110,115,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,90,83,73,77,80,76,69,0,102,105,108,101,32,100,111,101,115,32,99,111,110,102,111,114,109,32,116,111,32,70,73,84,83,32,115,116,97,110,100,97,114,100,0,90,66,73,84,80,73,88,0,100,97,116,97,32,116,121,112,101,32,111,102,32,111,114,105,103,105,110,97,108,32,105,109,97,103,101,0,90,78,65,88,73,83,0,100,105,109,101,110,115,105,111,110,32,111,102,32,111,114,105,103,105,110,97,108,32,105,109,97,103,101,0,90,78,65,88,73,83,37,100,0,108,101,110,103,116,104,32,111,102,32,111,114,105,103,105,110,97,108,32,105,109,97,103,101,32,97,120,105,115,0,90,84,73,76,69,37,100,0,115,105,122,101,32,111,102,32,116,105,108,101,115,32,116,111,32,98,101,32,99,111,109,112,114,101,115,115,101,100,0,90,81,85,65,78,84,73,90,0,78,79,78,69,0,76,111,115,115,108,101,115,115,32,99,111,109,112,114,101,115,115,105,111,110,32,119,105,116,104,111,117,116,32,113,117,97,110,116,105,122,97,116,105,111,110,0,83,85,66,84,82,65,67,84,73,86,69,95,68,73,84,72,69,82,95,49,0,80,105,120,101,108,32,81,117,97,110,116,105,122,97,116,105,111,110,32,65,108,103,111,114,105,116,104,109,0,90,68,73,84,72,69,82,48,0,100,105,116,104,101,114,105,110,103,32,111,102,102,115,101,116,32,119,104,101,110,32,113,117,97,110,116,105,122,105,110,103,32,102,108,111,97,116,115,0,83,85,66,84,82,65,67,84,73,86,69,95,68,73,84,72,69,82,95,50,0,82,73,67,69,95,79,78,69,0,78,79,95,68,73,84,72,69,82,0,78,111,32,100,105,116,104,101,114,105,110,103,32,100,117,114,105,110,103,32,113,117,97,110,116,105,122,97,116,105,111,110,0,90,67,77,80,84,89,80,69,0,99,111,109,112,114,101,115,115,105,111,110,32,97,108,103,111,114,105,116,104,109,0,90,78,65,77,69,49,0,66,76,79,67,75,83,73,90,69,0,99,111,109,112,114,101,115,115,105,111,110,32,98,108,111,99,107,32,115,105,122,101,0,90,86,65,76,49,0,112,105,120,101,108,115,32,112,101,114,32,98,108,111,99,107,0,90,78,65,77,69,50,0,66,89,84,69,80,73,88,0,98,121,116,101,115,32,112,101,114,32,112,105,120,101,108,32,40,49,44,32,50,44,32,52,44,32,111,114,32,56,41,0,90,86,65,76,50,0,83,67,65,76,69,0,72,67,79,77,80,82,69,83,83,32,115,99,97,108,101,32,102,97,99,116,111,114,0,83,77,79,79,84,72,0,72,67,79,77,80,82,69,83,83,32,115,109,111,111,116,104,32,111,112,116,105,111,110,0,111,102,102,115,101,116,32,100,97,116,97,32,114,97,110,103,101,32,116,111,32,116,104,97,116,32,111,102,32,117,110,115,105,103,110,101,100,32,115,104,111,114,116,0,66,90,69,82,79,0,100,101,102,97,117,108,116,32,115,99,97,108,105,110,103,32,102,97,99,116,111,114,0,66,83,67,65,76,69,0,111,102,102,115,101,116,32,100,97,116,97,32,114,97,110,103,101,32,116,111,32,116,104,97,116,32,111,102,32,115,105,103,110,101,100,32,98,121,116,101,0,111,102,102,115,101,116,32,100,97,116,97,32,114,97,110,103,101,32,116,111,32,116,104,97,116,32,111,102,32,117,110,115,105,103,110,101,100,32,108,111,110,103,0,69,88,84,78,65,77,69,32,61,32,39,67,79,77,80,82,69,83,83,69,68,95,73,77,65,71,69,39,0,66,73,84,80,73,88,0,66,76,65,78,75,0,32,32,32,32,0,90,66,76,65,78,75,0,110,117,108,108,32,118,97,108,117,101,32,105,110,32,116,104,101,32,99,111,109,112,114,101,115,115,101,100,32,105,110,116,101,103,101,114,32,97,114,114,97,121,0,76,111,115,115,108,101,115,115,32,99,111,109,112,114,101,115,115,105,111,110,32,111,102,32,102,108,111,97,116,105,110,103,32,112,111,105,110,116,32,105,109,97,103,101,115,32,109,117,115,116,32,117,115,101,32,71,90,73,80,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,78,79,67,79,77,80,82,69,83,83,73,79,78,32,111,112,116,105,111,110,32,111,110,108,121,32,115,117,112,112,111,114,116,101,100,32,102,111,114,32,105,110,116,42,50,44,32,105,110,116,42,52,44,32,97,110,100,32,102,108,111,97,116,42,52,32,105,109,97,103,101,115,0,85,78,67,79,77,80,82,69,83,83,69,68,95,68,65,84,65,0,73,109,112,108,105,99,105,116,32,100,97,116,97,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,119,104,101,110,32,119,114,105,116,105,110,103,32,116,111,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,115,0,117,110,115,117,112,112,111,114,116,101,100,32,105,109,97,103,101,32,100,97,116,97,116,121,112,101,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,77,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,102,97,105,108,117,114,101,46,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,101,114,114,111,114,32,82,105,99,101,32,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,32,116,105,108,101,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,100,97,116,97,32,111,117,116,32,111,102,32,114,97,110,103,101,32,102,111,114,32,80,76,73,79,32,99,111,109,112,114,101,115,115,105,111,110,32,40,48,32,45,32,50,42,42,50,52,41,0,101,114,114,111,114,32,80,76,73,79,32,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,32,116,105,108,101,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,109,97,108,108,111,99,32,102,97,105,108,101,100,10,0,98,122,105,112,50,32,99,111,109,112,114,101,115,115,105,111,110,32,101,114,114,111,114,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,32,40,102,105,116,115,95,105,110,116,95,116,111,95,108,111,110,103,108,111,110,103,95,105,110,112,108,97,99,101,41,0,71,90,73,80,95,67,79,77,80,82,69,83,83,69,68,95,68,65,84,65,0,49,81,66,0,49,80,66,0,77,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,101,114,114,111,114,46,32,40,105,109,99,111,109,112,95,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,68,97,116,97,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,47,115,99,97,108,105,110,103,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,119,104,101,110,32,119,114,105,116,105,110,103,32,116,111,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,115,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,32,40,102,105,116,115,95,115,104,111,114,116,95,116,111,95,105,110,116,95,105,110,112,108,97,99,101,41,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,32,40,102,105,116,115,95,117,115,104,111,114,116,95,116,111,95,105,110,116,95,105,110,112,108,97,99,101,41,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,32,40,102,105,116,115,95,117,98,121,116,101,95,116,111,95,105,110,116,95,105,110,112,108,97,99,101,41,0,79,117,116,32,111,102,32,109,101,109,111,114,121,46,32,40,102,105,116,115,95,115,98,121,116,101,95,116,111,95,105,110,116,95,105,110,112,108,97,99,101,41,0,85,110,107,110,111,119,110,32,100,105,116,104,101,114,105,110,103,32,109,101,116,104,111,100,46,0,77,97,121,32,110,101,101,100,32,116,111,32,105,110,115,116,97,108,108,32,97,32,110,101,119,101,114,32,118,101,114,115,105,111,110,32,111,102,32,67,70,73,84,83,73,79,46,0,85,110,107,110,111,119,110,32,115,117,98,116,114,97,99,116,105,118,101,32,100,105,116,104,101,114,105,110,103,32,109,101,116,104,111,100,46,0,67,72,68,85,32,105,115,32,110,111,116,32,97,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,40,102,105,116,115,95,119,114,105,116,101,95,99,111,109,112,114,101,115,115,101,100,95,105,109,103,41,0,117,110,115,117,112,112,111,114,116,101,100,32,100,97,116,97,116,121,112,101,32,102,111,114,32,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,0,117,110,115,117,112,112,111,114,116,101,100,32,105,109,97,103,101,32,99,111,109,112,114,101,115,115,105,111,110,32,97,108,103,111,114,105,116,104,109,0,79,117,116,32,111,102,32,109,101,109,111,114,121,32,40,102,105,116,115,95,119,114,105,116,101,95,99,111,109,112,114,101,115,115,95,105,109,103,41,0,101,114,114,111,114,32,97,108,108,111,99,97,116,105,110,103,32,109,101,109,111,114,121,32,102,111,114,32,103,122,105,112,112,101,100,32,116,105,108,101,32,40,105,109,99,111,109,112,95,100,101,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,99,111,109,112,114,101,115,115,101,100,32,98,121,116,101,32,115,116,114,101,97,109,32,102,114,111,109,32,98,105,110,97,114,121,32,116,97,98,108,101,0,105,110,99,111,109,112,97,116,105,98,108,101,32,100,97,116,97,32,116,121,112,101,32,105,110,32,103,122,105,112,112,101,100,32,102,108,111,97,116,105,110,103,45,112,111,105,110,116,32,116,105,108,101,45,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,77,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,102,97,105,108,117,114,101,32,102,111,114,32,116,101,109,112,102,108,111,97,116,46,32,40,105,109,99,111,109,112,95,100,101,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,102,97,105,108,101,100,32,116,111,32,103,117,110,122,105,112,32,116,104,101,32,105,109,97,103,101,32,116,105,108,101,0,105,109,112,108,105,99,105,116,32,100,97,116,97,32,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,102,111,114,32,103,122,105,112,112,101,100,32,105,109,97,103,101,32,116,105,108,101,115,0,105,109,112,108,105,99,105,116,32,100,97,116,97,32,116,121,112,101,32,99,111,110,118,101,114,115,105,111,110,32,105,115,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,105,110,32,116,105,108,101,45,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,115,0,101,114,114,111,114,58,32,117,110,99,111,109,112,114,101,115,115,101,100,32,116,105,108,101,32,104,97,115,32,119,114,111,110,103,32,115,105,122,101,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,115,99,97,108,105,110,103,32,102,97,99,116,111,114,32,97,110,100,32,111,102,102,115,101,116,32,102,111,114,32,99,111,109,112,114,101,115,115,101,100,32,116,105,108,101,0,101,114,114,111,114,32,114,101,97,100,105,110,103,32,110,117,108,108,32,118,97,108,117,101,32,102,111,114,32,99,111,109,112,114,101,115,115,101,100,32,116,105,108,101,0,77,101,109,111,114,121,32,97,108,108,111,99,97,116,105,111,110,32,102,97,105,108,117,114,101,32,102,111,114,32,105,100,97,116,97,46,32,40,105,109,99,111,109,112,95,100,101,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,79,117,116,32,111,102,32,109,101,109,111,114,121,32,102,111,114,32,99,98,117,102,46,32,40,105,109,99,111,109,112,95,100,101,99,111,109,112,114,101,115,115,95,116,105,108,101,41,0,98,122,105,112,50,32,100,101,99,111,109,112,114,101,115,115,105,111,110,32,101,114,114,111,114,0,117,110,107,110,111,119,110,32,99,111,109,112,114,101,115,115,105,111,110,32,97,108,103,111,114,105,116,104,109,0,111,110,108,121,32,49,68,44,32,50,68,44,32,111,114,32,51,68,32,105,109,97,103,101,115,32,97,114,101,32,99,117,114,114,101,110,116,108,121,32,115,117,112,112,111,114,116,101,100,0,67,72,68,85,32,105,115,32,110,111,116,32,97,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,40,102,105,116,115,95,114,101,97,100,95,99,111,109,112,114,101,115,115,101,100,95,105,109,103,41,0,117,110,115,117,112,112,111,114,116,101,100,32,100,97,116,97,116,121,112,101,32,102,111,114,32,117,110,99,111,109,112,114,101,115,115,105,110,103,32,105,109,97,103,101,0,79,117,116,32,111,102,32,109,101,109,111,114,121,32,40,102,105,116,115,95,114,101,97,100,95,99,111,109,112,114,101,115,115,95,105,109,103,41,0,67,72,68,85,32,105,115,32,110,111,116,32,97,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,40,102,105,116,115,95,105,109,103,95,100,101,99,111,109,112,114,101,115,115,41,0,90,84,69,78,83,73,79,78,0,101,114,114,111,114,32,99,114,101,97,116,105,110,103,32,111,117,116,112,117,116,32,100,101,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,72,68,85,0,101,114,114,111,114,32,99,111,112,121,105,110,103,32,104,101,97,100,101,114,32,107,101,121,119,111,114,100,115,32,102,114,111,109,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,101,114,114,111,114,32,99,111,112,121,105,110,103,32,112,114,105,109,97,114,121,32,107,101,121,119,111,114,100,115,32,102,114,111,109,32,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,0,114,101,113,117,105,114,101,100,32,90,67,77,80,84,89,80,69,32,99,111,109,112,114,101,115,115,105,111,110,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,32,105,110,0,32,105,109,99,111,109,112,95,103,101,116,95,99,111,109,112,114,101,115,115,101,100,95,105,109,97,103,101,95,112,97,114,0,66,90,73,80,50,95,49,0,85,110,107,110,111,119,110,32,105,109,97,103,101,32,99,111,109,112,114,101,115,115,105,111,110,32,116,121,112,101,58,0,114,101,113,117,105,114,101,100,32,90,66,73,84,80,73,88,32,99,111,109,112,114,101,115,115,105,111,110,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,114,101,113,117,105,114,101,100,32,90,78,65,88,73,83,32,99,111,109,112,114,101,115,115,105,111,110,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,67,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,104,97,115,32,110,111,32,100,97,116,97,32,40,90,78,65,88,73,83,32,60,32,49,41,0,67,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,32,104,97,115,32,116,111,111,32,109,97,110,121,32,100,105,109,101,110,115,105,111,110,115,0,114,101,113,117,105,114,101,100,32,90,78,65,88,73,83,110,32,99,111,109,112,114,101,115,115,105,111,110,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,110,117,109,98,101,114,32,111,102,32,116,97,98,108,101,32,114,111,119,115,32,33,61,32,116,104,101,32,110,117,109,98,101,114,32,111,102,32,116,105,108,101,115,32,105,110,32,99,111,109,112,114,101,115,115,101,100,32,105,109,97,103,101,0,114,101,113,117,105,114,101,100,32,90,86,65,76,49,32,99,111,109,112,114,101,115,115,105,111,110,32,107,101,121,119,111,114,100,32,110,111,116,32,102,111,117,110,100,0,67,79,77,80,82,69,83,83,69,68,95,68,65,84,65,0,99,111,117,108,100,110,39,116,32,102,105,110,100,32,67,79,77,80,82,69,83,83,69,68,95,68,65,84,65,32,99,111,108,117,109,110,32,40,102,105,116,115,95,103,101,116,95,99,111,109,112,114,101,115,115,101,100,95,105,109,103,95,112,97,114,41,0,90,83,67,65,76,69,0,90,90,69,82,79,0,80,67,79,85,78,84,0,90,80,67,79,85,78,84,0,67,72,69,67,75,83,85,77,0,90,72,69,67,75,83,85,77,0,68,65,84,65,83,85,77,0,90,68,65,84,65,83,85,77,0,88,84,69,78,83,73,79,78,0,45,0,78,65,88,73,83,0,78,65,88,73,83,109,0,71,67,79,85,78,84,0,84,70,73,69,76,68,83,0,84,84,89,80,69,109,0,84,70,79,82,77,109,0,84,72,69,65,80,0,90,84,73,76,69,109,0,90,78,65,77,69,109,0,90,86,65,76,109,0,43,0,42,0,83,73,77,80,76,69,0,90,78,65,88,73,83,109,0,90,69,88,84,69,78,68,0,69,88,84,69,78,68,0,90,66,76,79,67,75,69,68,0,66,76,79,67,75,69,68,0,90,71,67,79,85,78,84,0,72,73,83,84,79,82,89,0,67,79,77,77,69,78,84,0,0,0,102,105,116,115,95,114,99,111,109,112,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,114,105,99,101,95,101,110,99,111,100,101,58,32,101,110,100,32,111,102,32,98,117,102,102,101,114,0,100,101,99,111,109,112,114,101,115,115,105,111,110,32,101,114,114,111,114,58,32,104,105,116,32,101,110,100,32,111,102,32,99,111,109,112,114,101,115,115,101,100,32,98,121,116,101,32,115,116,114,101,97,109,0,100,101,99,111,109,112,114,101,115,115,105,111,110,32,119,97,114,110,105,110,103,58,32,117,110,117,115,101,100,32,98,121,116,101,115,32,97,116,32,101,110,100,32,111,102,32,99,111,109,112,114,101,115,115,101,100,32,98,117,102,102,101,114,0,104,116,114,97,110,115,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,101,110,99,111,100,101,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,101,110,99,111,100,101,58,32,111,117,116,112,117,116,32,98,117,102,102,101,114,32,116,111,111,32,115,109,97,108,108,0,104,116,114,97,110,115,54,52,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,101,110,99,111,100,101,54,52,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,101,110,99,111,100,101,54,52,58,32,111,117,116,112,117,116,32,98,117,102,102,101,114,32,116,111,111,32,115,109,97,108,108,0,113,116,114,101,101,95,101,110,99,111,100,101,54,52,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,113,116,114,101,101,95,101,110,99,111,100,101,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,221,153,98,97,100,32,102,105,108,101,32,102,111,114,109,97,116,0,100,111,100,101,99,111,100,101,58,32,98,97,100,32,98,105,116,32,112,108,97,110,101,32,118,97,108,117,101,115,0,104,105,110,118,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,100,111,100,101,99,111,100,101,54,52,58,32,98,97,100,32,98,105,116,32,112,108,97,110,101,32,118,97,108,117,101,115,0,104,105,110,118,54,52,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,113,116,114,101,101,95,100,101,99,111,100,101,54,52,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,113,116,114,101,101,95,100,101,99,111,100,101,54,52,58,32,98,97,100,32,102,111,114,109,97,116,32,99,111,100,101,0,113,116,114,101,101,95,100,101,99,111,100,101,58,32,105,110,115,117,102,102,105,99,105,101,110,116,32,109,101,109,111,114,121,0,113,116,114,101,101,95,100,101,99,111,100,101,58,32,98,97,100,32,102,111,114,109,97,116,32,99,111,100,101,0,78,117,109,98,101,114,32,111,102,32,98,121,116,101,115,32,116,111,32,119,114,105,116,101,32,105,115,32,103,114,101,97,116,101,114,32,116,104,97,110,32,76,79,78,71,95,77,65,88,32,40,102,102,112,98,121,116,41,46,0,69,114,114,111,114,58,32,116,114,121,105,110,103,32,116,111,32,119,114,105,116,101,32,116,111,32,82,69,65,68,79,78,76,89,32,102,105,108,101,46,0,67,97,110,110,111,116,32,119,114,105,116,101,32,116,111,32,97,32,71,90,73,80,32,111,114,32,67,79,77,80,82,69,83,83,32,99,111,109,112,114,101,115,115,101,100,32,102,105,108,101,46],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+773672);allocate([97,116,116,101,109,112,116,32,116,111,32,114,101,97,100,32,112,97,115,116,32,101,110,100,32,111,102,32,116,97,98,108,101,32,40,102,102,103,116,98,98,41,0,102,102,112,116,98,98,32,102,97,105,108,101,100,32,116,111,32,97,100,100,32,115,112,97,99,101,32,102,111,114,32,37,46,48,102,32,110,101,119,32,114,111,119,115,32,105,110,32,116,97,98,108,101,46,0,111,117,116,32,111,102,32,100,121,110,97,109,105,99,32,109,101,109,111,114,121,32,105,110,32,102,102,101,110,115,117,114,101,95,98,117,102,102,101,114,95,115,116,97,99,107,40,41,0,0,66,105,116,32,115,116,114,105,110,103,32,101,120,99,101,101,100,115,32,109,97,120,105,109,117,109,32,108,101,110,103,116,104,58,32,39,0,46,46,46,39,0,72,101,120,32,115,116,114,105,110,103,32,101,120,99,101,101,100,115,32,109,97,120,105,109,117,109,32,108,101,110,103,116,104,58,32,39,0,48,48,48,48,0,48,48,48,49,0,48,48,49,48,0,48,48,49,49,0,48,49,48,48,0,48,49,48,49,0,48,49,49,48,0,48,49,49,49,0,49,48,48,48,0,49,48,48,49,0,49,48,49,48,0,49,48,49,49,0,49,49,48,48,0,49,49,48,49,0,49,49,49,48,0,49,49,49,49,0,120,120,120,120,0,35,80,73,0,35,69,0,35,68,69,71,0,35,82,79,87,0,35,78,85,76,76,0,35,83,78,85,76,76,0,83,116,114,105,110,103,32,101,120,99,101,101,100,115,32,109,97,120,105,109,117,109,32,108,101,110,103,116,104,58,32,39,0,71,84,73,70,73,76,84,69,82,40,0,82,69,71,70,73,76,84,69,82,40,0,102,97,116,97,108,32,102,108,101,120,32,115,99,97,110,110,101,114,32,105,110,116,101,114,110,97,108,32,101,114,114,111,114,45,45,101,110,100,32,111,102,32,98,117,102,102,101,114,32,109,105,115,115,101,100,0,102,97,116,97,108,32,101,114,114,111,114,32,45,32,115,99,97,110,110,101,114,32,105,110,112,117,116,32,98,117,102,102,101,114,32,111,118,101,114,102,108,111,119,0,114,101,97,100,40,41,32,105,110,32,102,108,101,120,32,115,99,97,110,110,101,114,32,102,97,105,108,101,100,0,111,117,116,32,111,102,32,100,121,110,97,109,105,99,32,109,101,109,111,114,121,32,105,110,32,102,102,95,103,101,116,95,110,101,120,116,95,98,117,102,102,101,114,40,41,0,102,97,116,97,108,32,102,108,101,120,32,115,99,97,110,110,101,114,32,105,110,116,101,114,110,97,108,32,101,114,114,111,114,45,45,110,111,32,97,99,116,105,111,110,32,102,111,117,110,100,0,111,117,116,32,111,102,32,100,121,110,97,109,105,99,32,109,101,109,111,114,121,32,105,110,32,102,102,95,99,114,101,97,116,101,95,98,117,102,102,101,114,40,41,0,85,110,97,98,108,101,32,116,111,32,102,105,110,100,32,100,97,116,97,58,32,0,66,97,100,32,100,97,116,97,116,121,112,101,32,102,111,114,32,100,97,116,97,58,32,0,37,115,10,0,0,2,2,2,2,2,2,2,2,2,50,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,37,41,2,52,53,38,35,20,36,2,39,2,2,2,2,2,2,2,2,2,2,22,2,2,21,2,25,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,47,2,51,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,23,40,24,28,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,26,27,29,30,31,32,33,34,42,43,44,45,46,48,49,2,0,1,0,68,31,32,117,18,0,0,0,0,0,33,69,118,19,35,36,120,0,0,0,0,0,0,4,0,3,0,0,0,0,0,0,9,51,0,0,0,0,0,0,0,0,0,0,0,0,103,0,0,0,0,0,12,10,0,43,44,115,29,66,67,64,65,0,0,0,0,0,17,0,16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5,0,0,0,0,0,0,0,6,0,0,0,0,0,0,0,0,0,0,8,0,0,0,0,0,0,0,7,0,55,0,52,54,0,53,0,96,97,98,0,0,104,0,107,0,0,0,0,45,116,30,121,15,11,13,14,0,81,83,82,79,80,78,77,38,39,37,40,46,41,42,0,0,0,0,91,90,93,92,47,0,0,0,72,71,76,74,73,75,23,22,21,0,85,84,87,89,88,86,122,0,0,0,0,0,0,0,0,34,70,119,20,0,0,59,0,0,0,0,110,29,0,0,24,0,57,99,0,124,0,56,0,105,0,94,0,48,50,49,95,123,0,0,0,0,0,0,0,0,60,0,111,0,25,0,125,0,100,0,0,108,0,0,0,0,0,0,0,0,61,0,112,0,26,58,0,106,109,0,0,0,0,0,62,0,113,0,27,0,101,0,0,0,0,63,114,28,0,0,102,0,2,0,2,1,2,2,2,2,2,2,3,2,3,3,3,2,2,1,1,4,3,3,3,4,6,8,10,12,2,3,1,1,1,4,1,1,3,3,3,3,3,3,2,2,3,3,3,5,5,5,2,3,3,3,3,5,5,9,4,6,8,10,12,2,2,2,2,1,1,4,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,5,5,3,3,3,5,7,11,15,2,3,5,9,3,7,9,4,6,8,10,12,2,3,1,1,4,1,3,3,5,5,7,67,111,117,108,100,110,39,116,32,98,117,105,108,100,32,110,111,100,101,32,115,116,114,117,99,116,117,114,101,58,32,111,117,116,32,111,102,32,109,101,109,111,114,121,63,0,79,102,102,115,101,116,32,97,114,103,117,109,101,110,116,32,109,117,115,116,32,98,101,32,97,32,99,111,110,115,116,97,110,116,32,105,110,116,101,103,101,114,0,67,111,109,98,105,110,101,100,32,98,105,116,32,115,116,114,105,110,103,32,115,105,122,101,32,101,120,99,101,101,100,115,32,50,53,53,32,98,105,116,115,0,73,110,99,111,109,112,97,116,105,98,108,101,32,100,105,109,101,110,115,105,111,110,115,32,105,110,32,39,63,58,39,32,97,114,103,117,109,101,110,116,115,0,73,110,99,111,109,112,97,116,105,98,108,101,32,100,105,109,101,110,115,105,111,110,115,32,105,110,32,39,63,58,39,32,99,111,110,100,105,116,105,111,110,0,82,65,78,68,79,77,40,0,82,65,78,68,79,77,78,40,0,70,117,110,99,116,105,111,110,40,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,85,77,40,0,78,69,76,69,77,40,0,65,67,67,85,77,40,0,70,117,110,99,116,105,111,110,40,98,111,111,108,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,86,65,76,73,68,40,0,70,117,110,99,116,105,111,110,40,115,116,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,77,73,78,40,0,77,65,88,40,0,70,117,110,99,116,105,111,110,40,98,105,116,115,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,65,86,69,82,65,71,69,40,0,83,84,68,68,69,86,40,0,77,69,68,73,65,78,40,0,83,69,81,68,73,70,70,40,0,65,66,83,40,0,83,73,78,40,0,67,79,83,40,0,84,65,78,40,0,65,82,67,83,73,78,40,0,65,83,73,78,40,0,65,82,67,67,79,83,40,0,65,67,79,83,40,0,65,82,67,84,65,78,40,0,65,84,65,78,40,0,83,73,78,72,40,0,67,79,83,72,40,0,84,65,78,72,40,0,69,88,80,40,0,76,79,71,40,0,76,79,71,49,48,40,0,83,81,82,84,40,0,82,79,85,78,68,40,0,70,76,79,79,82,40,0,67,69,73,76,40,0,82,65,78,68,79,77,80,40,0,70,117,110,99,116,105,111,110,40,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,84,82,83,84,82,40,0,68,69,70,78,85,76,76,40,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,68,69,70,78,85,76,76,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,65,82,67,84,65,78,50,40,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,97,114,99,116,97,110,50,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,109,105,110,40,97,44,98,41,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,109,97,120,40,97,44,98,41,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,70,117,110,99,116,105,111,110,40,101,120,112,114,44,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,65,78,71,83,69,80,40,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,65,78,71,83,69,80,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,70,117,110,99,116,105,111,110,40,101,120,112,114,44,101,120,112,114,44,101,120,112,114,44,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,73,83,78,85,76,76,40,0,66,111,111,108,101,97,110,32,70,117,110,99,116,105,111,110,40,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,66,111,111,108,101,97,110,32,70,117,110,99,116,105,111,110,40,101,120,112,114,44,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,78,69,65,82,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,78,69,65,82,40,0,66,111,111,108,101,97,110,32,70,117,110,99,116,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,67,73,82,67,76,69,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,67,73,82,67,76,69,40,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,66,79,88,32,111,114,32,69,76,76,73,80,83,69,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,66,79,88,40,0,69,76,76,73,80,83,69,40,0,83,65,79,32,73,109,97,103,101,32,70,117,110,99,116,105,111,110,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,42,83,84,65,82,84,42,0,42,83,84,79,80,42,0,67,111,109,98,105,110,101,100,32,115,116,114,105,110,103,32,115,105,122,101,32,101,120,99,101,101,100,115,32,50,53,53,32,99,104,97,114,97,99,116,101,114,115,0,67,97,110,110,111,116,32,104,97,118,101,32,97,32,118,101,99,116,111,114,32,115,116,114,105,110,103,32,99,111,108,117,109,110,0,70,117,110,99,116,105,111,110,40,115,116,114,105,110,103,44,115,116,114,105,110,103,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,84,82,77,73,68,40,0,87,104,101,110,32,117,115,105,110,103,32,83,84,82,77,73,68,40,83,44,80,44,78,41,44,32,80,32,97,110,100,32,78,32,109,117,115,116,32,98,101,32,105,110,116,101,103,101,114,115,32,40,97,110,100,32,110,111,116,32,118,101,99,116,111,114,32,99,111,108,117,109,110,115,41,0,83,84,82,77,73,68,40,83,44,80,44,78,41,44,32,78,32,109,117,115,116,32,98,101,32,49,45,50,53,53,0,70,117,110,99,116,105,111,110,40,115,116,114,105,110,103,44,101,120,112,114,44,101,120,112,114,41,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,0,54,55,55,56,56,56,56,56,56,57,57,58,58,58,58,59,60,61,61,61,61,61,61,61,61,61,61,61,61,61,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,59,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,60,62,62,62,62,62,62,62,62,62,255,1,29,30,31,46,47,44,58,115,121,110,116,97,120,32,101,114,114,111,114,0,109,101,109,111,114,121,32,101,120,104,97,117,115,116,101,100,0,88,0,67,111,117,108,100,32,110,111,116,32,98,117,105,108,100,32,88,32,99,111,108,117,109,110,32,102,111,114,32,82,69,71,70,73,76,84,69,82,0,89,0,67,111,117,108,100,32,110,111,116,32,98,117,105,108,100,32,89,32,99,111,108,117,109,110,32,102,111,114,32,82,69,71,70,73,76,84,69,82,0,68,105,109,101,110,115,105,111,110,115,32,111,102,32,82,69,71,70,73,76,84,69,82,32,97,114,103,117,109,101,110,116,115,32,97,114,101,32,110,111,116,32,99,111,109,112,97,116,105,98,108,101,0,67,111,117,108,100,32,110,111,116,32,101,120,116,114,97,99,116,32,118,97,108,105,100,32,112,97,105,114,32,111,102,32,99,111,108,117,109,110,32,110,97,109,101,115,32,102,114,111,109,32,82,69,71,70,73,76,84,69,82,0,67,111,117,108,100,32,110,111,116,32,108,111,99,97,116,101,32,99,111,108,117,109,110,115,32,105,110,100,105,99,97,116,101,100,32,102,111,114,32,87,67,83,32,105,110,102,111,0,70,111,117,110,100,32,109,117,108,116,105,112,108,101,32,88,47,89,32,99,111,108,117,109,110,32,114,101,102,101,114,101,110,99,101,115,32,105,110,32,82,69,71,70,73,76,84,69,82,0,84,73,77,69,0,67,111,117,108,100,32,110,111,116,32,98,117,105,108,100,32,84,73,77,69,32,99,111,108,117,109,110,32,102,111,114,32,71,84,73,70,73,76,84,69,82,0,84,73,77,69,90,69,82,79,0,84,73,77,69,90,69,82,73,0,84,73,77,69,90,69,82,70,0,67,97,110,110,111,116,32,117,115,101,32,112,114,105,109,97,114,121,32,97,114,114,97,121,32,102,111,114,32,71,84,73,32,102,105,108,116,101,114,0,70,105,108,101,32,101,120,116,101,110,115,105,111,110,32,115,112,101,99,105,102,105,101,114,32,108,97,99,107,115,32,99,108,111,115,105,110,103,32,39,93,39,0,69,88,84,78,65,77,69,0,71,84,73,0,71,84,73,32,101,120,116,101,110,115,105,111,110,32,110,111,116,32,102,111,117,110,100,32,105,110,32,116,104,105,115,32,102,105,108,101,0,78,65,88,73,83,50,0,79,117,116,32,111,102,32,114,97,110,103,101,32,97,114,103,117,109,101,110,116,32,116,111,32,97,114,99,115,105,110,0,79,117,116,32,111,102,32,114,97,110,103,101,32,97,114,103,117,109,101,110,116,32,116,111,32,97,114,99,99,111,115,0,79,117,116,32,111,102,32,114,97,110,103,101,32,97,114,103,117,109,101,110,116,32,116,111,32,108,111,103,0,79,117,116,32,111,102,32,114,97,110,103,101,32,97,114,103,117,109,101,110,116,32,116,111,32,108,111,103,49,48,0,79,117,116,32,111,102,32,114,97,110,103,101,32,97,114,103,117,109,101,110,116,32,116,111,32,115,113,114,116,0,83,84,82,77,73,68,40,83,44,80,44,78,41,32,80,32,109,117,115,116,32,98,101,32,48,32,111,114,32,103,114,101,97,116,101,114,0,67,111,117,108,100,32,110,111,116,32,97,108,108,111,99,97,116,101,32,116,101,109,112,111,114,97,114,121,32,109,101,109,111,114,121,32,105,110,32,109,101,100,105,97,110,32,102,117,110,99,116,105,111,110,0,67,97,110,110,111,116,32,105,110,100,101,120,32,97,32,115,99,97,108,97,114,32,118,97,108,117,101,0,67,97,110,110,111,116,32,117,115,101,32,97,110,32,97,114,114,97,121,32,97,115,32,97,110,32,105,110,100,101,120,32,118,97,108,117,101,0,73,110,100,101,120,32,118,97,108,117,101,32,109,117,115,116,32,98,101,32,97,110,32,105,110,116,101,103,101,114,32,116,121,112,101,0,77,117,115,116,32,115,112,101,99,105,102,121,32,106,117,115,116,32,111,110,101,32,111,114,32,97,108,108,32,105,110,100,105,99,101,115,32,102,111,114,32,118,101,99,116,111,114,0,73,110,100,101,120,32,111,117,116,32,111,102,32,114,97,110,103,101,0,78,117,108,108,32,101,110,99,111,117,110,116,101,114,101,100,32,97,115,32,118,101,99,116,111,114,32,105,110,100,101,120,0,65,114,114,97,121,32,115,105,122,101,115,47,100,105,109,115,32,100,111,32,110,111,116,32,109,97,116,99,104,32,102,111,114,32,98,105,110,97,114,121,32,111,112,101,114,97,116,111,114,0,68,105,118,105,100,101,32,98,121,32,90,101,114,111,0,84,33,34,25,13,1,2,3,17,75,28,12,16,4,11,29,18,30,39,104,110,111,112,113,98,32,5,6,15,19,20,21,26,8,22,7,40,36,23,24,9,10,14,27,31,37,35,131,130,125,38,42,43,60,61,62,63,67,71,74,77,88,89,90,91,92,93,94,95,96,97,99,100,101,102,103,105,106,107,108,114,115,116,121,122,123,124,0,73,108,108,101,103,97,108,32,98,121,116,101,32,115,101,113,117,101,110,99,101,0,68,111,109,97,105,110,32,101,114,114,111,114,0,82,101,115,117,108,116,32,110,111,116,32,114,101,112,114,101,115,101,110,116,97,98,108,101,0,78,111,116,32,97,32,116,116,121,0,80,101,114,109,105,115,115,105,111,110,32,100,101,110,105,101,100,0,79,112,101,114,97,116,105,111,110,32,110,111,116,32,112,101,114,109,105,116,116,101,100,0,78,111,32,115,117,99,104,32,102,105,108,101,32,111,114,32,100,105,114,101,99,116,111,114,121,0,78,111,32,115,117,99,104,32,112,114,111,99,101,115,115,0,70,105,108,101,32,101,120,105,115,116,115,0,86,97,108,117,101,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,100,97,116,97,32,116,121,112,101,0,78,111,32,115,112,97,99,101,32,108,101,102,116,32,111,110,32,100,101,118,105,99,101,0,79,117,116,32,111,102,32,109,101,109,111,114,121,0,82,101,115,111,117,114,99,101,32,98,117,115,121,0,73,110,116,101,114,114,117,112,116,101,100,32,115,121,115,116,101,109,32,99,97,108,108,0,82,101,115,111,117,114,99,101,32,116,101,109,112,111,114,97,114,105,108,121,32,117,110,97,118,97,105,108,97,98,108,101,0,73,110,118,97,108,105,100,32,115,101,101,107,0,67,114,111,115,115,45,100,101,118,105,99,101,32,108,105,110,107,0,82,101,97,100,45,111,110,108,121,32,102,105,108,101,32,115,121,115,116,101,109,0,68,105,114,101,99,116,111,114,121,32,110,111,116,32,101,109,112,116,121,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,112,101,101,114,0,79,112,101,114,97,116,105,111,110,32,116,105,109,101,100,32,111,117,116,0,67,111,110,110,101,99,116,105,111,110,32,114,101,102,117,115,101,100,0,72,111,115,116,32,105,115,32,100,111,119,110,0,72,111,115,116,32,105,115,32,117,110,114,101,97,99,104,97,98,108,101,0,65,100,100,114,101,115,115,32,105,110,32,117,115,101,0,66,114,111,107,101,110,32,112,105,112,101,0,73,47,79,32,101,114,114,111,114,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,32,111,114,32,97,100,100,114,101,115,115,0,66,108,111,99,107,32,100,101,118,105,99,101,32,114,101,113,117,105,114,101,100,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,0,78,111,116,32,97,32,100,105,114,101,99,116,111,114,121,0,73,115,32,97,32,100,105,114,101,99,116,111,114,121,0,84,101,120,116,32,102,105,108,101,32,98,117,115,121,0,69,120,101,99,32,102,111,114,109,97,116,32,101,114,114,111,114,0,73,110,118,97,108,105,100,32,97,114,103,117,109,101,110,116,0,65,114,103,117,109,101,110,116,32,108,105,115,116,32,116,111,111,32,108,111,110,103,0,83,121,109,98,111,108,105,99,32,108,105,110,107,32,108,111,111,112,0,70,105,108,101,110,97,109,101,32,116,111,111,32,108,111,110,103,0,84,111,111,32,109,97,110,121,32,111,112,101,110,32,102,105,108,101,115,32,105,110,32,115,121,115,116,101,109,0,78,111,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,115,32,97,118,97,105,108,97,98,108,101,0,66,97,100,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,0,78,111,32,99,104,105,108,100,32,112,114,111,99,101,115,115,0,66,97,100,32,97,100,100,114,101,115,115,0,70,105,108,101,32,116,111,111,32,108,97,114,103,101,0,84,111,111,32,109,97,110,121,32,108,105,110,107,115,0,78,111,32,108,111,99,107,115,32,97,118,97,105,108,97,98,108,101,0,82,101,115,111,117,114,99,101,32,100,101,97,100,108,111,99,107,32,119,111,117,108,100,32,111,99,99,117,114,0,83,116,97,116,101,32,110,111,116,32,114,101,99,111,118,101,114,97,98,108,101,0,80,114,101,118,105,111,117,115,32,111,119,110,101,114,32,100,105,101,100,0,79,112,101,114,97,116,105,111,110,32,99,97,110,99,101,108,101,100,0,70,117,110,99,116,105,111,110,32,110,111,116,32,105,109,112,108,101,109,101,110,116,101,100,0,78,111,32,109,101,115,115,97,103,101,32,111,102,32,100,101,115,105,114,101,100,32,116,121,112,101,0,73,100,101,110,116,105,102,105,101,114,32,114,101,109,111,118,101,100,0,68,101,118,105,99,101,32,110,111,116,32,97,32,115,116,114,101,97,109,0,78,111,32,100,97,116,97,32,97,118,97,105,108,97,98,108,101,0,68,101,118,105,99,101,32,116,105,109,101,111,117,116,0,79,117,116,32,111,102,32,115,116,114,101,97,109,115,32,114,101,115,111,117,114,99,101,115,0,76,105,110,107,32,104,97,115,32,98,101,101,110,32,115,101,118,101,114,101,100,0,80,114,111,116,111,99,111,108,32,101,114,114,111,114,0,66,97,100,32,109,101,115,115,97,103,101,0,70,105,108,101,32,100,101,115,99,114,105,112,116,111,114,32,105,110,32,98,97,100,32,115,116,97,116,101,0,78,111,116,32,97,32,115,111,99,107,101,116,0,68,101,115,116,105,110,97,116,105,111,110,32,97,100,100,114,101,115,115,32,114,101,113,117,105,114,101,100,0,77,101,115,115,97,103,101,32,116,111,111,32,108,97,114,103,101,0,80,114,111,116,111,99,111,108,32,119,114,111,110,103,32,116,121,112,101,32,102,111,114,32,115,111,99,107,101,116,0,80,114,111,116,111,99,111,108,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,80,114,111,116,111,99,111,108,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,111,99,107,101,116,32,116,121,112,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,111,116,32,115,117,112,112,111,114,116,101,100,0,80,114,111,116,111,99,111,108,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,65,100,100,114,101,115,115,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,112,114,111,116,111,99,111,108,0,65,100,100,114,101,115,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,78,101,116,119,111,114,107,32,105,115,32,100,111,119,110,0,78,101,116,119,111,114,107,32,117,110,114,101,97,99,104,97,98,108,101,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,110,101,116,119,111,114,107,0,67,111,110,110,101,99,116,105,111,110,32,97,98,111,114,116,101,100,0,78,111,32,98,117,102,102,101,114,32,115,112,97,99,101,32,97,118,97,105,108,97,98,108,101,0,83,111,99,107,101,116,32,105,115,32,99,111,110,110,101,99,116,101,100,0,83,111,99,107,101,116,32,110,111,116,32,99,111,110,110,101,99,116,101,100,0,67,97,110,110,111,116,32,115,101,110,100,32,97,102,116,101,114,32,115,111,99,107,101,116,32,115,104,117,116,100,111,119,110,0,79,112,101,114,97,116,105,111,110,32,97,108,114,101,97,100,121,32,105,110,32,112,114,111,103,114,101,115,115,0,79,112,101,114,97,116,105,111,110,32,105,110,32,112,114,111,103,114,101,115,115,0,83,116,97,108,101,32,102,105,108,101,32,104,97,110,100,108,101,0,82,101,109,111,116,101,32,73,47,79,32,101,114,114,111,114,0,81,117,111,116,97,32,101,120,99,101,101,100,101,100,0,78,111,32,109,101,100,105,117,109,32,102,111,117,110,100,0,87,114,111,110,103,32,109,101,100,105,117,109,32,116,121,112,101,0,78,111,32,101,114,114,111,114,32,105,110,102,111,114,109,97,116,105,111,110,0,0,105,110,102,105,110,105,116,121,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,4,7,3,6,5,0,58,32,105,108,108,101,103,97,108,32,111,112,116,105,111,110,58,32,0,10,0,58,32,111,112,116,105,111,110,32,114,101,113,117,105,114,101,115,32,97,110,32,97,114,103,117,109,101,110,116,58,32,0,114,119,97],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+786581);allocate([17,0,10,0,17,17,17,0,0,0,0,5,0,0,0,0,0,0,9,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,15,10,17,17,17,3,10,7,0,1,19,9,11,11,0,0,9,6,11,0,0,11,0,6,17,0,0,0,17,17,17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,10,10,17,17,17,0,10,0,0,2,0,9,11,0,0,0,9,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,4,13,0,0,0,0,9,14,0,0,0,0,0,14,0,0,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,0,0,0,0,0,0,0,0,0,0,0,15,0,0,0,0,15,0,0,0,0,9,16,0,0,0,0,0,16,0,0,16,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,0,10,0,0,0,0,9,11,0,0,0,0,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,45,43,32,32,32,48,88,48,120,0,40,110,117,108,108,41,0,45,48,88,43,48,88,32,48,88,45,48,120,43,48,120,32,48,120,0,105,110,102,0,73,78,70,0,110,97,110,0,78,65,78,0,46,0,0],"i8",ALLOC_NONE,Runtime.GLOBAL_BASE+795051);var tempDoublePtr=Runtime.alignMemory(allocate(12,"i8",ALLOC_STATIC),8);assert(tempDoublePtr%8==0);function copyTempFloat(ptr){HEAP8[tempDoublePtr]=HEAP8[ptr];HEAP8[tempDoublePtr+1]=HEAP8[ptr+1];HEAP8[tempDoublePtr+2]=HEAP8[ptr+2];HEAP8[tempDoublePtr+3]=HEAP8[ptr+3]}function copyTempDouble(ptr){HEAP8[tempDoublePtr]=HEAP8[ptr];HEAP8[tempDoublePtr+1]=HEAP8[ptr+1];HEAP8[tempDoublePtr+2]=HEAP8[ptr+2];HEAP8[tempDoublePtr+3]=HEAP8[ptr+3];HEAP8[tempDoublePtr+4]=HEAP8[ptr+4];HEAP8[tempDoublePtr+5]=HEAP8[ptr+5];HEAP8[tempDoublePtr+6]=HEAP8[ptr+6];HEAP8[tempDoublePtr+7]=HEAP8[ptr+7]}Module["_i64Subtract"]=_i64Subtract;var _fabsf=Math_abs;Module["_i64Add"]=_i64Add;var _FtoIHigh=true;var ERRNO_CODES={EPERM:1,ENOENT:2,ESRCH:3,EINTR:4,EIO:5,ENXIO:6,E2BIG:7,ENOEXEC:8,EBADF:9,ECHILD:10,EAGAIN:11,EWOULDBLOCK:11,ENOMEM:12,EACCES:13,EFAULT:14,ENOTBLK:15,EBUSY:16,EEXIST:17,EXDEV:18,ENODEV:19,ENOTDIR:20,EISDIR:21,EINVAL:22,ENFILE:23,EMFILE:24,ENOTTY:25,ETXTBSY:26,EFBIG:27,ENOSPC:28,ESPIPE:29,EROFS:30,EMLINK:31,EPIPE:32,EDOM:33,ERANGE:34,ENOMSG:42,EIDRM:43,ECHRNG:44,EL2NSYNC:45,EL3HLT:46,EL3RST:47,ELNRNG:48,EUNATCH:49,ENOCSI:50,EL2HLT:51,EDEADLK:35,ENOLCK:37,EBADE:52,EBADR:53,EXFULL:54,ENOANO:55,EBADRQC:56,EBADSLT:57,EDEADLOCK:35,EBFONT:59,ENOSTR:60,ENODATA:61,ETIME:62,ENOSR:63,ENONET:64,ENOPKG:65,EREMOTE:66,ENOLINK:67,EADV:68,ESRMNT:69,ECOMM:70,EPROTO:71,EMULTIHOP:72,EDOTDOT:73,EBADMSG:74,ENOTUNIQ:76,EBADFD:77,EREMCHG:78,ELIBACC:79,ELIBBAD:80,ELIBSCN:81,ELIBMAX:82,ELIBEXEC:83,ENOSYS:38,ENOTEMPTY:39,ENAMETOOLONG:36,ELOOP:40,EOPNOTSUPP:95,EPFNOSUPPORT:96,ECONNRESET:104,ENOBUFS:105,EAFNOSUPPORT:97,EPROTOTYPE:91,ENOTSOCK:88,ENOPROTOOPT:92,ESHUTDOWN:108,ECONNREFUSED:111,EADDRINUSE:98,ECONNABORTED:103,ENETUNREACH:101,ENETDOWN:100,ETIMEDOUT:110,EHOSTDOWN:112,EHOSTUNREACH:113,EINPROGRESS:115,EALREADY:114,EDESTADDRREQ:89,EMSGSIZE:90,EPROTONOSUPPORT:93,ESOCKTNOSUPPORT:94,EADDRNOTAVAIL:99,ENETRESET:102,EISCONN:106,ENOTCONN:107,ETOOMANYREFS:109,EUSERS:87,EDQUOT:122,ESTALE:116,ENOTSUP:95,ENOMEDIUM:123,EILSEQ:84,EOVERFLOW:75,ECANCELED:125,ENOTRECOVERABLE:131,EOWNERDEAD:130,ESTRPIPE:86};var ERRNO_MESSAGES={0:"Success",1:"Not super-user",2:"No such file or directory",3:"No such process",4:"Interrupted system call",5:"I/O error",6:"No such device or address",7:"Arg list too long",8:"Exec format error",9:"Bad file number",10:"No children",11:"No more processes",12:"Not enough core",13:"Permission denied",14:"Bad address",15:"Block device required",16:"Mount device busy",17:"File exists",18:"Cross-device link",19:"No such device",20:"Not a directory",21:"Is a directory",22:"Invalid argument",23:"Too many open files in system",24:"Too many open files",25:"Not a typewriter",26:"Text file busy",27:"File too large",28:"No space left on device",29:"Illegal seek",30:"Read only file system",31:"Too many links",32:"Broken pipe",33:"Math arg out of domain of func",34:"Math result not representable",35:"File locking deadlock error",36:"File or path name too long",37:"No record locks available",38:"Function not implemented",39:"Directory not empty",40:"Too many symbolic links",42:"No message of desired type",43:"Identifier removed",44:"Channel number out of range",45:"Level 2 not synchronized",46:"Level 3 halted",47:"Level 3 reset",48:"Link number out of range",49:"Protocol driver not attached",50:"No CSI structure available",51:"Level 2 halted",52:"Invalid exchange",53:"Invalid request descriptor",54:"Exchange full",55:"No anode",56:"Invalid request code",57:"Invalid slot",59:"Bad font file fmt",60:"Device not a stream",61:"No data (for no delay io)",62:"Timer expired",63:"Out of streams resources",64:"Machine is not on the network",65:"Package not installed",66:"The object is remote",67:"The link has been severed",68:"Advertise error",69:"Srmount error",70:"Communication error on send",71:"Protocol error",72:"Multihop attempted",73:"Cross mount point (not really error)",74:"Trying to read unreadable message",75:"Value too large for defined data type",76:"Given log. name not unique",77:"f.d. invalid for this operation",78:"Remote address changed",79:"Can   access a needed shared lib",80:"Accessing a corrupted shared lib",81:".lib section in a.out corrupted",82:"Attempting to link in too many libs",83:"Attempting to exec a shared library",84:"Illegal byte sequence",86:"Streams pipe error",87:"Too many users",88:"Socket operation on non-socket",89:"Destination address required",90:"Message too long",91:"Protocol wrong type for socket",92:"Protocol not available",93:"Unknown protocol",94:"Socket type not supported",95:"Not supported",96:"Protocol family not supported",97:"Address family not supported by protocol family",98:"Address already in use",99:"Address not available",100:"Network interface is not configured",101:"Network is unreachable",102:"Connection reset by network",103:"Connection aborted",104:"Connection reset by peer",105:"No buffer space available",106:"Socket is already connected",107:"Socket is not connected",108:"Can't send after socket shutdown",109:"Too many references",110:"Connection timed out",111:"Connection refused",112:"Host is down",113:"Host is unreachable",114:"Socket already connected",115:"Connection already in progress",116:"Stale file handle",122:"Quota exceeded",123:"No medium (in tape drive)",125:"Operation canceled",130:"Previous owner died",131:"State not recoverable"};function ___setErrNo(value){if(Module["___errno_location"])HEAP32[Module["___errno_location"]()>>2]=value;return value}var PATH={splitPath:(function(filename){var splitPathRe=/^(\/?|)([\s\S]*?)((?:\.{1,2}|[^\/]+?|)(\.[^.\/]*|))(?:[\/]*)$/;return splitPathRe.exec(filename).slice(1)}),normalizeArray:(function(parts,allowAboveRoot){var up=0;for(var i=parts.length-1;i>=0;i--){var last=parts[i];if(last==="."){parts.splice(i,1)}else if(last===".."){parts.splice(i,1);up++}else if(up){parts.splice(i,1);up--}}if(allowAboveRoot){for(;up--;up){parts.unshift("..")}}return parts}),normalize:(function(path){var isAbsolute=path.charAt(0)==="/",trailingSlash=path.substr(-1)==="/";path=PATH.normalizeArray(path.split("/").filter((function(p){return!!p})),!isAbsolute).join("/");if(!path&&!isAbsolute){path="."}if(path&&trailingSlash){path+="/"}return(isAbsolute?"/":"")+path}),dirname:(function(path){var result=PATH.splitPath(path),root=result[0],dir=result[1];if(!root&&!dir){return"."}if(dir){dir=dir.substr(0,dir.length-1)}return root+dir}),basename:(function(path){if(path==="/")return"/";var lastSlash=path.lastIndexOf("/");if(lastSlash===-1)return path;return path.substr(lastSlash+1)}),extname:(function(path){return PATH.splitPath(path)[3]}),join:(function(){var paths=Array.prototype.slice.call(arguments,0);return PATH.normalize(paths.join("/"))}),join2:(function(l,r){return PATH.normalize(l+"/"+r)}),resolve:(function(){var resolvedPath="",resolvedAbsolute=false;for(var i=arguments.length-1;i>=-1&&!resolvedAbsolute;i--){var path=i>=0?arguments[i]:FS.cwd();if(typeof path!=="string"){throw new TypeError("Arguments to path.resolve must be strings")}else if(!path){return""}resolvedPath=path+"/"+resolvedPath;resolvedAbsolute=path.charAt(0)==="/"}resolvedPath=PATH.normalizeArray(resolvedPath.split("/").filter((function(p){return!!p})),!resolvedAbsolute).join("/");return(resolvedAbsolute?"/":"")+resolvedPath||"."}),relative:(function(from,to){from=PATH.resolve(from).substr(1);to=PATH.resolve(to).substr(1);function trim(arr){var start=0;for(;start<arr.length;start++){if(arr[start]!=="")break}var end=arr.length-1;for(;end>=0;end--){if(arr[end]!=="")break}if(start>end)return[];return arr.slice(start,end-start+1)}var fromParts=trim(from.split("/"));var toParts=trim(to.split("/"));var length=Math.min(fromParts.length,toParts.length);var samePartsLength=length;for(var i=0;i<length;i++){if(fromParts[i]!==toParts[i]){samePartsLength=i;break}}var outputParts=[];for(var i=samePartsLength;i<fromParts.length;i++){outputParts.push("..")}outputParts=outputParts.concat(toParts.slice(samePartsLength));return outputParts.join("/")})};var TTY={ttys:[],init:(function(){}),shutdown:(function(){}),register:(function(dev,ops){TTY.ttys[dev]={input:[],output:[],ops:ops};FS.registerDevice(dev,TTY.stream_ops)}),stream_ops:{open:(function(stream){var tty=TTY.ttys[stream.node.rdev];if(!tty){throw new FS.ErrnoError(ERRNO_CODES.ENODEV)}stream.tty=tty;stream.seekable=false}),close:(function(stream){stream.tty.ops.flush(stream.tty)}),flush:(function(stream){stream.tty.ops.flush(stream.tty)}),read:(function(stream,buffer,offset,length,pos){if(!stream.tty||!stream.tty.ops.get_char){throw new FS.ErrnoError(ERRNO_CODES.ENXIO)}var bytesRead=0;for(var i=0;i<length;i++){var result;try{result=stream.tty.ops.get_char(stream.tty)}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EIO)}if(result===undefined&&bytesRead===0){throw new FS.ErrnoError(ERRNO_CODES.EAGAIN)}if(result===null||result===undefined)break;bytesRead++;buffer[offset+i]=result}if(bytesRead){stream.node.timestamp=Date.now()}return bytesRead}),write:(function(stream,buffer,offset,length,pos){if(!stream.tty||!stream.tty.ops.put_char){throw new FS.ErrnoError(ERRNO_CODES.ENXIO)}for(var i=0;i<length;i++){try{stream.tty.ops.put_char(stream.tty,buffer[offset+i])}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EIO)}}if(length){stream.node.timestamp=Date.now()}return i})},default_tty_ops:{get_char:(function(tty){if(!tty.input.length){var result=null;if(ENVIRONMENT_IS_NODE){var BUFSIZE=256;var buf=new Buffer(BUFSIZE);var bytesRead=0;var fd=process.stdin.fd;var usingDevice=false;try{fd=fs.openSync("/dev/stdin","r");usingDevice=true}catch(e){}bytesRead=fs.readSync(fd,buf,0,BUFSIZE,null);if(usingDevice){fs.closeSync(fd)}if(bytesRead>0){result=buf.slice(0,bytesRead).toString("utf-8")}else{result=null}}else if(typeof window!="undefined"&&typeof window.prompt=="function"){result=window.prompt("Input: ");if(result!==null){result+="\n"}}else if(typeof readline=="function"){result=readline();if(result!==null){result+="\n"}}if(!result){return null}tty.input=intArrayFromString(result,true)}return tty.input.shift()}),put_char:(function(tty,val){if(val===null||val===10){Module["print"](UTF8ArrayToString(tty.output,0));tty.output=[]}else{if(val!=0)tty.output.push(val)}}),flush:(function(tty){if(tty.output&&tty.output.length>0){Module["print"](UTF8ArrayToString(tty.output,0));tty.output=[]}})},default_tty1_ops:{put_char:(function(tty,val){if(val===null||val===10){Module["printErr"](UTF8ArrayToString(tty.output,0));tty.output=[]}else{if(val!=0)tty.output.push(val)}}),flush:(function(tty){if(tty.output&&tty.output.length>0){Module["printErr"](UTF8ArrayToString(tty.output,0));tty.output=[]}})}};var MEMFS={ops_table:null,mount:(function(mount){return MEMFS.createNode(null,"/",16384|511,0)}),createNode:(function(parent,name,mode,dev){if(FS.isBlkdev(mode)||FS.isFIFO(mode)){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}if(!MEMFS.ops_table){MEMFS.ops_table={dir:{node:{getattr:MEMFS.node_ops.getattr,setattr:MEMFS.node_ops.setattr,lookup:MEMFS.node_ops.lookup,mknod:MEMFS.node_ops.mknod,rename:MEMFS.node_ops.rename,unlink:MEMFS.node_ops.unlink,rmdir:MEMFS.node_ops.rmdir,readdir:MEMFS.node_ops.readdir,symlink:MEMFS.node_ops.symlink},stream:{llseek:MEMFS.stream_ops.llseek}},file:{node:{getattr:MEMFS.node_ops.getattr,setattr:MEMFS.node_ops.setattr},stream:{llseek:MEMFS.stream_ops.llseek,read:MEMFS.stream_ops.read,write:MEMFS.stream_ops.write,allocate:MEMFS.stream_ops.allocate,mmap:MEMFS.stream_ops.mmap,msync:MEMFS.stream_ops.msync}},link:{node:{getattr:MEMFS.node_ops.getattr,setattr:MEMFS.node_ops.setattr,readlink:MEMFS.node_ops.readlink},stream:{}},chrdev:{node:{getattr:MEMFS.node_ops.getattr,setattr:MEMFS.node_ops.setattr},stream:FS.chrdev_stream_ops}}}var node=FS.createNode(parent,name,mode,dev);if(FS.isDir(node.mode)){node.node_ops=MEMFS.ops_table.dir.node;node.stream_ops=MEMFS.ops_table.dir.stream;node.contents={}}else if(FS.isFile(node.mode)){node.node_ops=MEMFS.ops_table.file.node;node.stream_ops=MEMFS.ops_table.file.stream;node.usedBytes=0;node.contents=null}else if(FS.isLink(node.mode)){node.node_ops=MEMFS.ops_table.link.node;node.stream_ops=MEMFS.ops_table.link.stream}else if(FS.isChrdev(node.mode)){node.node_ops=MEMFS.ops_table.chrdev.node;node.stream_ops=MEMFS.ops_table.chrdev.stream}node.timestamp=Date.now();if(parent){parent.contents[name]=node}return node}),getFileDataAsRegularArray:(function(node){if(node.contents&&node.contents.subarray){var arr=[];for(var i=0;i<node.usedBytes;++i)arr.push(node.contents[i]);return arr}return node.contents}),getFileDataAsTypedArray:(function(node){if(!node.contents)return new Uint8Array;if(node.contents.subarray)return node.contents.subarray(0,node.usedBytes);return new Uint8Array(node.contents)}),expandFileStorage:(function(node,newCapacity){if(node.contents&&node.contents.subarray&&newCapacity>node.contents.length){node.contents=MEMFS.getFileDataAsRegularArray(node);node.usedBytes=node.contents.length}if(!node.contents||node.contents.subarray){var prevCapacity=node.contents?node.contents.buffer.byteLength:0;if(prevCapacity>=newCapacity)return;var CAPACITY_DOUBLING_MAX=1024*1024;newCapacity=Math.max(newCapacity,prevCapacity*(prevCapacity<CAPACITY_DOUBLING_MAX?2:1.125)|0);if(prevCapacity!=0)newCapacity=Math.max(newCapacity,256);var oldContents=node.contents;node.contents=new Uint8Array(newCapacity);if(node.usedBytes>0)node.contents.set(oldContents.subarray(0,node.usedBytes),0);return}if(!node.contents&&newCapacity>0)node.contents=[];while(node.contents.length<newCapacity)node.contents.push(0)}),resizeFileStorage:(function(node,newSize){if(node.usedBytes==newSize)return;if(newSize==0){node.contents=null;node.usedBytes=0;return}if(!node.contents||node.contents.subarray){var oldContents=node.contents;node.contents=new Uint8Array(new ArrayBuffer(newSize));if(oldContents){node.contents.set(oldContents.subarray(0,Math.min(newSize,node.usedBytes)))}node.usedBytes=newSize;return}if(!node.contents)node.contents=[];if(node.contents.length>newSize)node.contents.length=newSize;else while(node.contents.length<newSize)node.contents.push(0);node.usedBytes=newSize}),node_ops:{getattr:(function(node){var attr={};attr.dev=FS.isChrdev(node.mode)?node.id:1;attr.ino=node.id;attr.mode=node.mode;attr.nlink=1;attr.uid=0;attr.gid=0;attr.rdev=node.rdev;if(FS.isDir(node.mode)){attr.size=4096}else if(FS.isFile(node.mode)){attr.size=node.usedBytes}else if(FS.isLink(node.mode)){attr.size=node.link.length}else{attr.size=0}attr.atime=new Date(node.timestamp);attr.mtime=new Date(node.timestamp);attr.ctime=new Date(node.timestamp);attr.blksize=4096;attr.blocks=Math.ceil(attr.size/attr.blksize);return attr}),setattr:(function(node,attr){if(attr.mode!==undefined){node.mode=attr.mode}if(attr.timestamp!==undefined){node.timestamp=attr.timestamp}if(attr.size!==undefined){MEMFS.resizeFileStorage(node,attr.size)}}),lookup:(function(parent,name){throw FS.genericErrors[ERRNO_CODES.ENOENT]}),mknod:(function(parent,name,mode,dev){return MEMFS.createNode(parent,name,mode,dev)}),rename:(function(old_node,new_dir,new_name){if(FS.isDir(old_node.mode)){var new_node;try{new_node=FS.lookupNode(new_dir,new_name)}catch(e){}if(new_node){for(var i in new_node.contents){throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY)}}}delete old_node.parent.contents[old_node.name];old_node.name=new_name;new_dir.contents[new_name]=old_node;old_node.parent=new_dir}),unlink:(function(parent,name){delete parent.contents[name]}),rmdir:(function(parent,name){var node=FS.lookupNode(parent,name);for(var i in node.contents){throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY)}delete parent.contents[name]}),readdir:(function(node){var entries=[".",".."];for(var key in node.contents){if(!node.contents.hasOwnProperty(key)){continue}entries.push(key)}return entries}),symlink:(function(parent,newname,oldpath){var node=MEMFS.createNode(parent,newname,511|40960,0);node.link=oldpath;return node}),readlink:(function(node){if(!FS.isLink(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}return node.link})},stream_ops:{read:(function(stream,buffer,offset,length,position){var contents=stream.node.contents;if(position>=stream.node.usedBytes)return 0;var size=Math.min(stream.node.usedBytes-position,length);assert(size>=0);if(size>8&&contents.subarray){buffer.set(contents.subarray(position,position+size),offset)}else{for(var i=0;i<size;i++)buffer[offset+i]=contents[position+i]}return size}),write:(function(stream,buffer,offset,length,position,canOwn){if(!length)return 0;var node=stream.node;node.timestamp=Date.now();if(buffer.subarray&&(!node.contents||node.contents.subarray)){if(canOwn){node.contents=buffer.subarray(offset,offset+length);node.usedBytes=length;return length}else if(node.usedBytes===0&&position===0){node.contents=new Uint8Array(buffer.subarray(offset,offset+length));node.usedBytes=length;return length}else if(position+length<=node.usedBytes){node.contents.set(buffer.subarray(offset,offset+length),position);return length}}MEMFS.expandFileStorage(node,position+length);if(node.contents.subarray&&buffer.subarray)node.contents.set(buffer.subarray(offset,offset+length),position);else{for(var i=0;i<length;i++){node.contents[position+i]=buffer[offset+i]}}node.usedBytes=Math.max(node.usedBytes,position+length);return length}),llseek:(function(stream,offset,whence){var position=offset;if(whence===1){position+=stream.position}else if(whence===2){if(FS.isFile(stream.node.mode)){position+=stream.node.usedBytes}}if(position<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}return position}),allocate:(function(stream,offset,length){MEMFS.expandFileStorage(stream.node,offset+length);stream.node.usedBytes=Math.max(stream.node.usedBytes,offset+length)}),mmap:(function(stream,buffer,offset,length,position,prot,flags){if(!FS.isFile(stream.node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENODEV)}var ptr;var allocated;var contents=stream.node.contents;if(!(flags&2)&&(contents.buffer===buffer||contents.buffer===buffer.buffer)){allocated=false;ptr=contents.byteOffset}else{if(position>0||position+length<stream.node.usedBytes){if(contents.subarray){contents=contents.subarray(position,position+length)}else{contents=Array.prototype.slice.call(contents,position,position+length)}}allocated=true;ptr=_malloc(length);if(!ptr){throw new FS.ErrnoError(ERRNO_CODES.ENOMEM)}buffer.set(contents,ptr)}return{ptr:ptr,allocated:allocated}}),msync:(function(stream,buffer,offset,length,mmapFlags){if(!FS.isFile(stream.node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENODEV)}if(mmapFlags&2){return 0}var bytesWritten=MEMFS.stream_ops.write(stream,buffer,0,length,offset,false);return 0})}};var IDBFS={dbs:{},indexedDB:(function(){if(typeof indexedDB!=="undefined")return indexedDB;var ret=null;if(typeof window==="object")ret=window.indexedDB||window.mozIndexedDB||window.webkitIndexedDB||window.msIndexedDB;assert(ret,"IDBFS used, but indexedDB not supported");return ret}),DB_VERSION:21,DB_STORE_NAME:"FILE_DATA",mount:(function(mount){return MEMFS.mount.apply(null,arguments)}),syncfs:(function(mount,populate,callback){IDBFS.getLocalSet(mount,(function(err,local){if(err)return callback(err);IDBFS.getRemoteSet(mount,(function(err,remote){if(err)return callback(err);var src=populate?remote:local;var dst=populate?local:remote;IDBFS.reconcile(src,dst,callback)}))}))}),getDB:(function(name,callback){var db=IDBFS.dbs[name];if(db){return callback(null,db)}var req;try{req=IDBFS.indexedDB().open(name,IDBFS.DB_VERSION)}catch(e){return callback(e)}req.onupgradeneeded=(function(e){var db=e.target.result;var transaction=e.target.transaction;var fileStore;if(db.objectStoreNames.contains(IDBFS.DB_STORE_NAME)){fileStore=transaction.objectStore(IDBFS.DB_STORE_NAME)}else{fileStore=db.createObjectStore(IDBFS.DB_STORE_NAME)}if(!fileStore.indexNames.contains("timestamp")){fileStore.createIndex("timestamp","timestamp",{unique:false})}});req.onsuccess=(function(){db=req.result;IDBFS.dbs[name]=db;callback(null,db)});req.onerror=(function(e){callback(this.error);e.preventDefault()})}),getLocalSet:(function(mount,callback){var entries={};function isRealDir(p){return p!=="."&&p!==".."}function toAbsolute(root){return(function(p){return PATH.join2(root,p)})}var check=FS.readdir(mount.mountpoint).filter(isRealDir).map(toAbsolute(mount.mountpoint));while(check.length){var path=check.pop();var stat;try{stat=FS.stat(path)}catch(e){return callback(e)}if(FS.isDir(stat.mode)){check.push.apply(check,FS.readdir(path).filter(isRealDir).map(toAbsolute(path)))}entries[path]={timestamp:stat.mtime}}return callback(null,{type:"local",entries:entries})}),getRemoteSet:(function(mount,callback){var entries={};IDBFS.getDB(mount.mountpoint,(function(err,db){if(err)return callback(err);var transaction=db.transaction([IDBFS.DB_STORE_NAME],"readonly");transaction.onerror=(function(e){callback(this.error);e.preventDefault()});var store=transaction.objectStore(IDBFS.DB_STORE_NAME);var index=store.index("timestamp");index.openKeyCursor().onsuccess=(function(event){var cursor=event.target.result;if(!cursor){return callback(null,{type:"remote",db:db,entries:entries})}entries[cursor.primaryKey]={timestamp:cursor.key};cursor.continue()})}))}),loadLocalEntry:(function(path,callback){var stat,node;try{var lookup=FS.lookupPath(path);node=lookup.node;stat=FS.stat(path)}catch(e){return callback(e)}if(FS.isDir(stat.mode)){return callback(null,{timestamp:stat.mtime,mode:stat.mode})}else if(FS.isFile(stat.mode)){node.contents=MEMFS.getFileDataAsTypedArray(node);return callback(null,{timestamp:stat.mtime,mode:stat.mode,contents:node.contents})}else{return callback(new Error("node type not supported"))}}),storeLocalEntry:(function(path,entry,callback){try{if(FS.isDir(entry.mode)){FS.mkdir(path,entry.mode)}else if(FS.isFile(entry.mode)){FS.writeFile(path,entry.contents,{encoding:"binary",canOwn:true})}else{return callback(new Error("node type not supported"))}FS.chmod(path,entry.mode);FS.utime(path,entry.timestamp,entry.timestamp)}catch(e){return callback(e)}callback(null)}),removeLocalEntry:(function(path,callback){try{var lookup=FS.lookupPath(path);var stat=FS.stat(path);if(FS.isDir(stat.mode)){FS.rmdir(path)}else if(FS.isFile(stat.mode)){FS.unlink(path)}}catch(e){return callback(e)}callback(null)}),loadRemoteEntry:(function(store,path,callback){var req=store.get(path);req.onsuccess=(function(event){callback(null,event.target.result)});req.onerror=(function(e){callback(this.error);e.preventDefault()})}),storeRemoteEntry:(function(store,path,entry,callback){var req=store.put(entry,path);req.onsuccess=(function(){callback(null)});req.onerror=(function(e){callback(this.error);e.preventDefault()})}),removeRemoteEntry:(function(store,path,callback){var req=store.delete(path);req.onsuccess=(function(){callback(null)});req.onerror=(function(e){callback(this.error);e.preventDefault()})}),reconcile:(function(src,dst,callback){var total=0;var create=[];Object.keys(src.entries).forEach((function(key){var e=src.entries[key];var e2=dst.entries[key];if(!e2||e.timestamp>e2.timestamp){create.push(key);total++}}));var remove=[];Object.keys(dst.entries).forEach((function(key){var e=dst.entries[key];var e2=src.entries[key];if(!e2){remove.push(key);total++}}));if(!total){return callback(null)}var errored=false;var completed=0;var db=src.type==="remote"?src.db:dst.db;var transaction=db.transaction([IDBFS.DB_STORE_NAME],"readwrite");var store=transaction.objectStore(IDBFS.DB_STORE_NAME);function done(err){if(err){if(!done.errored){done.errored=true;return callback(err)}return}if(++completed>=total){return callback(null)}}transaction.onerror=(function(e){done(this.error);e.preventDefault()});create.sort().forEach((function(path){if(dst.type==="local"){IDBFS.loadRemoteEntry(store,path,(function(err,entry){if(err)return done(err);IDBFS.storeLocalEntry(path,entry,done)}))}else{IDBFS.loadLocalEntry(path,(function(err,entry){if(err)return done(err);IDBFS.storeRemoteEntry(store,path,entry,done)}))}}));remove.sort().reverse().forEach((function(path){if(dst.type==="local"){IDBFS.removeLocalEntry(path,done)}else{IDBFS.removeRemoteEntry(store,path,done)}}))})};var NODEFS={isWindows:false,staticInit:(function(){NODEFS.isWindows=!!process.platform.match(/^win/)}),mount:(function(mount){assert(ENVIRONMENT_IS_NODE);return NODEFS.createNode(null,"/",NODEFS.getMode(mount.opts.root),0)}),createNode:(function(parent,name,mode,dev){if(!FS.isDir(mode)&&!FS.isFile(mode)&&!FS.isLink(mode)){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var node=FS.createNode(parent,name,mode);node.node_ops=NODEFS.node_ops;node.stream_ops=NODEFS.stream_ops;return node}),getMode:(function(path){var stat;try{stat=fs.lstatSync(path);if(NODEFS.isWindows){stat.mode=stat.mode|(stat.mode&146)>>1}}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}return stat.mode}),realPath:(function(node){var parts=[];while(node.parent!==node){parts.push(node.name);node=node.parent}parts.push(node.mount.opts.root);parts.reverse();return PATH.join.apply(null,parts)}),flagsToPermissionStringMap:{0:"r",1:"r+",2:"r+",64:"r",65:"r+",66:"r+",129:"rx+",193:"rx+",514:"w+",577:"w",578:"w+",705:"wx",706:"wx+",1024:"a",1025:"a",1026:"a+",1089:"a",1090:"a+",1153:"ax",1154:"ax+",1217:"ax",1218:"ax+",4096:"rs",4098:"rs+"},flagsToPermissionString:(function(flags){flags&=~32768;if(flags in NODEFS.flagsToPermissionStringMap){return NODEFS.flagsToPermissionStringMap[flags]}else{throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}}),node_ops:{getattr:(function(node){var path=NODEFS.realPath(node);var stat;try{stat=fs.lstatSync(path)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}if(NODEFS.isWindows&&!stat.blksize){stat.blksize=4096}if(NODEFS.isWindows&&!stat.blocks){stat.blocks=(stat.size+stat.blksize-1)/stat.blksize|0}return{dev:stat.dev,ino:stat.ino,mode:stat.mode,nlink:stat.nlink,uid:stat.uid,gid:stat.gid,rdev:stat.rdev,size:stat.size,atime:stat.atime,mtime:stat.mtime,ctime:stat.ctime,blksize:stat.blksize,blocks:stat.blocks}}),setattr:(function(node,attr){var path=NODEFS.realPath(node);try{if(attr.mode!==undefined){fs.chmodSync(path,attr.mode);node.mode=attr.mode}if(attr.timestamp!==undefined){var date=new Date(attr.timestamp);fs.utimesSync(path,date,date)}if(attr.size!==undefined){fs.truncateSync(path,attr.size)}}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),lookup:(function(parent,name){var path=PATH.join2(NODEFS.realPath(parent),name);var mode=NODEFS.getMode(path);return NODEFS.createNode(parent,name,mode)}),mknod:(function(parent,name,mode,dev){var node=NODEFS.createNode(parent,name,mode,dev);var path=NODEFS.realPath(node);try{if(FS.isDir(node.mode)){fs.mkdirSync(path,node.mode)}else{fs.writeFileSync(path,"",{mode:node.mode})}}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}return node}),rename:(function(oldNode,newDir,newName){var oldPath=NODEFS.realPath(oldNode);var newPath=PATH.join2(NODEFS.realPath(newDir),newName);try{fs.renameSync(oldPath,newPath)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),unlink:(function(parent,name){var path=PATH.join2(NODEFS.realPath(parent),name);try{fs.unlinkSync(path)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),rmdir:(function(parent,name){var path=PATH.join2(NODEFS.realPath(parent),name);try{fs.rmdirSync(path)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),readdir:(function(node){var path=NODEFS.realPath(node);try{return fs.readdirSync(path)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),symlink:(function(parent,newName,oldPath){var newPath=PATH.join2(NODEFS.realPath(parent),newName);try{fs.symlinkSync(oldPath,newPath)}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),readlink:(function(node){var path=NODEFS.realPath(node);try{path=fs.readlinkSync(path);path=NODEJS_PATH.relative(NODEJS_PATH.resolve(node.mount.opts.root),path);return path}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}})},stream_ops:{open:(function(stream){var path=NODEFS.realPath(stream.node);try{if(FS.isFile(stream.node.mode)){stream.nfd=fs.openSync(path,NODEFS.flagsToPermissionString(stream.flags))}}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),close:(function(stream){try{if(FS.isFile(stream.node.mode)&&stream.nfd){fs.closeSync(stream.nfd)}}catch(e){if(!e.code)throw e;throw new FS.ErrnoError(ERRNO_CODES[e.code])}}),read:(function(stream,buffer,offset,length,position){if(length===0)return 0;var nbuffer=new Buffer(length);var res;try{res=fs.readSync(stream.nfd,nbuffer,0,length,position)}catch(e){throw new FS.ErrnoError(ERRNO_CODES[e.code])}if(res>0){for(var i=0;i<res;i++){buffer[offset+i]=nbuffer[i]}}return res}),write:(function(stream,buffer,offset,length,position){var nbuffer=new Buffer(buffer.subarray(offset,offset+length));var res;try{res=fs.writeSync(stream.nfd,nbuffer,0,length,position)}catch(e){throw new FS.ErrnoError(ERRNO_CODES[e.code])}return res}),llseek:(function(stream,offset,whence){var position=offset;if(whence===1){position+=stream.position}else if(whence===2){if(FS.isFile(stream.node.mode)){try{var stat=fs.fstatSync(stream.nfd);position+=stat.size}catch(e){throw new FS.ErrnoError(ERRNO_CODES[e.code])}}}if(position<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}return position})}};var WORKERFS={DIR_MODE:16895,FILE_MODE:33279,reader:null,mount:(function(mount){assert(ENVIRONMENT_IS_WORKER);if(!WORKERFS.reader)WORKERFS.reader=new FileReaderSync;var root=WORKERFS.createNode(null,"/",WORKERFS.DIR_MODE,0);var createdParents={};function ensureParent(path){var parts=path.split("/");var parent=root;for(var i=0;i<parts.length-1;i++){var curr=parts.slice(0,i+1).join("/");if(!createdParents[curr]){createdParents[curr]=WORKERFS.createNode(parent,curr,WORKERFS.DIR_MODE,0)}parent=createdParents[curr]}return parent}function base(path){var parts=path.split("/");return parts[parts.length-1]}Array.prototype.forEach.call(mount.opts["files"]||[],(function(file){WORKERFS.createNode(ensureParent(file.name),base(file.name),WORKERFS.FILE_MODE,0,file,file.lastModifiedDate)}));(mount.opts["blobs"]||[]).forEach((function(obj){WORKERFS.createNode(ensureParent(obj["name"]),base(obj["name"]),WORKERFS.FILE_MODE,0,obj["data"])}));(mount.opts["packages"]||[]).forEach((function(pack){pack["metadata"].files.forEach((function(file){var name=file.filename.substr(1);WORKERFS.createNode(ensureParent(name),base(name),WORKERFS.FILE_MODE,0,pack["blob"].slice(file.start,file.end))}))}));return root}),createNode:(function(parent,name,mode,dev,contents,mtime){var node=FS.createNode(parent,name,mode);node.mode=mode;node.node_ops=WORKERFS.node_ops;node.stream_ops=WORKERFS.stream_ops;node.timestamp=(mtime||new Date).getTime();assert(WORKERFS.FILE_MODE!==WORKERFS.DIR_MODE);if(mode===WORKERFS.FILE_MODE){node.size=contents.size;node.contents=contents}else{node.size=4096;node.contents={}}if(parent){parent.contents[name]=node}return node}),node_ops:{getattr:(function(node){return{dev:1,ino:undefined,mode:node.mode,nlink:1,uid:0,gid:0,rdev:undefined,size:node.size,atime:new Date(node.timestamp),mtime:new Date(node.timestamp),ctime:new Date(node.timestamp),blksize:4096,blocks:Math.ceil(node.size/4096)}}),setattr:(function(node,attr){if(attr.mode!==undefined){node.mode=attr.mode}if(attr.timestamp!==undefined){node.timestamp=attr.timestamp}}),lookup:(function(parent,name){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}),mknod:(function(parent,name,mode,dev){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),rename:(function(oldNode,newDir,newName){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),unlink:(function(parent,name){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),rmdir:(function(parent,name){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),readdir:(function(node){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),symlink:(function(parent,newName,oldPath){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}),readlink:(function(node){throw new FS.ErrnoError(ERRNO_CODES.EPERM)})},stream_ops:{read:(function(stream,buffer,offset,length,position){if(position>=stream.node.size)return 0;var chunk=stream.node.contents.slice(position,position+length);var ab=WORKERFS.reader.readAsArrayBuffer(chunk);buffer.set(new Uint8Array(ab),offset);return chunk.size}),write:(function(stream,buffer,offset,length,position){throw new FS.ErrnoError(ERRNO_CODES.EIO)}),llseek:(function(stream,offset,whence){var position=offset;if(whence===1){position+=stream.position}else if(whence===2){if(FS.isFile(stream.node.mode)){position+=stream.node.size}}if(position<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}return position})}};var _stdin=allocate(1,"i32*",ALLOC_STATIC);var _stdout=allocate(1,"i32*",ALLOC_STATIC);var _stderr=allocate(1,"i32*",ALLOC_STATIC);var FS={root:null,mounts:[],devices:[null],streams:[],nextInode:1,nameTable:null,currentPath:"/",initialized:false,ignorePermissions:true,trackingDelegate:{},tracking:{openFlags:{READ:1,WRITE:2}},ErrnoError:null,genericErrors:{},filesystems:null,handleFSError:(function(e){if(!(e instanceof FS.ErrnoError))throw e+" : "+stackTrace();return ___setErrNo(e.errno)}),lookupPath:(function(path,opts){path=PATH.resolve(FS.cwd(),path);opts=opts||{};if(!path)return{path:"",node:null};var defaults={follow_mount:true,recurse_count:0};for(var key in defaults){if(opts[key]===undefined){opts[key]=defaults[key]}}if(opts.recurse_count>8){throw new FS.ErrnoError(ERRNO_CODES.ELOOP)}var parts=PATH.normalizeArray(path.split("/").filter((function(p){return!!p})),false);var current=FS.root;var current_path="/";for(var i=0;i<parts.length;i++){var islast=i===parts.length-1;if(islast&&opts.parent){break}current=FS.lookupNode(current,parts[i]);current_path=PATH.join2(current_path,parts[i]);if(FS.isMountpoint(current)){if(!islast||islast&&opts.follow_mount){current=current.mounted.root}}if(!islast||opts.follow){var count=0;while(FS.isLink(current.mode)){var link=FS.readlink(current_path);current_path=PATH.resolve(PATH.dirname(current_path),link);var lookup=FS.lookupPath(current_path,{recurse_count:opts.recurse_count});current=lookup.node;if(count++>40){throw new FS.ErrnoError(ERRNO_CODES.ELOOP)}}}}return{path:current_path,node:current}}),getPath:(function(node){var path;while(true){if(FS.isRoot(node)){var mount=node.mount.mountpoint;if(!path)return mount;return mount[mount.length-1]!=="/"?mount+"/"+path:mount+path}path=path?node.name+"/"+path:node.name;node=node.parent}}),hashName:(function(parentid,name){var hash=0;for(var i=0;i<name.length;i++){hash=(hash<<5)-hash+name.charCodeAt(i)|0}return(parentid+hash>>>0)%FS.nameTable.length}),hashAddNode:(function(node){var hash=FS.hashName(node.parent.id,node.name);node.name_next=FS.nameTable[hash];FS.nameTable[hash]=node}),hashRemoveNode:(function(node){var hash=FS.hashName(node.parent.id,node.name);if(FS.nameTable[hash]===node){FS.nameTable[hash]=node.name_next}else{var current=FS.nameTable[hash];while(current){if(current.name_next===node){current.name_next=node.name_next;break}current=current.name_next}}}),lookupNode:(function(parent,name){var err=FS.mayLookup(parent);if(err){throw new FS.ErrnoError(err,parent)}var hash=FS.hashName(parent.id,name);for(var node=FS.nameTable[hash];node;node=node.name_next){var nodeName=node.name;if(node.parent.id===parent.id&&nodeName===name){return node}}return FS.lookup(parent,name)}),createNode:(function(parent,name,mode,rdev){if(!FS.FSNode){FS.FSNode=(function(parent,name,mode,rdev){if(!parent){parent=this}this.parent=parent;this.mount=parent.mount;this.mounted=null;this.id=FS.nextInode++;this.name=name;this.mode=mode;this.node_ops={};this.stream_ops={};this.rdev=rdev});FS.FSNode.prototype={};var readMode=292|73;var writeMode=146;Object.defineProperties(FS.FSNode.prototype,{read:{get:(function(){return(this.mode&readMode)===readMode}),set:(function(val){val?this.mode|=readMode:this.mode&=~readMode})},write:{get:(function(){return(this.mode&writeMode)===writeMode}),set:(function(val){val?this.mode|=writeMode:this.mode&=~writeMode})},isFolder:{get:(function(){return FS.isDir(this.mode)})},isDevice:{get:(function(){return FS.isChrdev(this.mode)})}})}var node=new FS.FSNode(parent,name,mode,rdev);FS.hashAddNode(node);return node}),destroyNode:(function(node){FS.hashRemoveNode(node)}),isRoot:(function(node){return node===node.parent}),isMountpoint:(function(node){return!!node.mounted}),isFile:(function(mode){return(mode&61440)===32768}),isDir:(function(mode){return(mode&61440)===16384}),isLink:(function(mode){return(mode&61440)===40960}),isChrdev:(function(mode){return(mode&61440)===8192}),isBlkdev:(function(mode){return(mode&61440)===24576}),isFIFO:(function(mode){return(mode&61440)===4096}),isSocket:(function(mode){return(mode&49152)===49152}),flagModes:{"r":0,"rs":1052672,"r+":2,"w":577,"wx":705,"xw":705,"w+":578,"wx+":706,"xw+":706,"a":1089,"ax":1217,"xa":1217,"a+":1090,"ax+":1218,"xa+":1218},modeStringToFlags:(function(str){var flags=FS.flagModes[str];if(typeof flags==="undefined"){throw new Error("Unknown file open mode: "+str)}return flags}),flagsToPermissionString:(function(flag){var perms=["r","w","rw"][flag&3];if(flag&512){perms+="w"}return perms}),nodePermissions:(function(node,perms){if(FS.ignorePermissions){return 0}if(perms.indexOf("r")!==-1&&!(node.mode&292)){return ERRNO_CODES.EACCES}else if(perms.indexOf("w")!==-1&&!(node.mode&146)){return ERRNO_CODES.EACCES}else if(perms.indexOf("x")!==-1&&!(node.mode&73)){return ERRNO_CODES.EACCES}return 0}),mayLookup:(function(dir){var err=FS.nodePermissions(dir,"x");if(err)return err;if(!dir.node_ops.lookup)return ERRNO_CODES.EACCES;return 0}),mayCreate:(function(dir,name){try{var node=FS.lookupNode(dir,name);return ERRNO_CODES.EEXIST}catch(e){}return FS.nodePermissions(dir,"wx")}),mayDelete:(function(dir,name,isdir){var node;try{node=FS.lookupNode(dir,name)}catch(e){return e.errno}var err=FS.nodePermissions(dir,"wx");if(err){return err}if(isdir){if(!FS.isDir(node.mode)){return ERRNO_CODES.ENOTDIR}if(FS.isRoot(node)||FS.getPath(node)===FS.cwd()){return ERRNO_CODES.EBUSY}}else{if(FS.isDir(node.mode)){return ERRNO_CODES.EISDIR}}return 0}),mayOpen:(function(node,flags){if(!node){return ERRNO_CODES.ENOENT}if(FS.isLink(node.mode)){return ERRNO_CODES.ELOOP}else if(FS.isDir(node.mode)){if((flags&2097155)!==0||flags&512){return ERRNO_CODES.EISDIR}}return FS.nodePermissions(node,FS.flagsToPermissionString(flags))}),MAX_OPEN_FDS:4096,nextfd:(function(fd_start,fd_end){fd_start=fd_start||0;fd_end=fd_end||FS.MAX_OPEN_FDS;for(var fd=fd_start;fd<=fd_end;fd++){if(!FS.streams[fd]){return fd}}throw new FS.ErrnoError(ERRNO_CODES.EMFILE)}),getStream:(function(fd){return FS.streams[fd]}),createStream:(function(stream,fd_start,fd_end){if(!FS.FSStream){FS.FSStream=(function(){});FS.FSStream.prototype={};Object.defineProperties(FS.FSStream.prototype,{object:{get:(function(){return this.node}),set:(function(val){this.node=val})},isRead:{get:(function(){return(this.flags&2097155)!==1})},isWrite:{get:(function(){return(this.flags&2097155)!==0})},isAppend:{get:(function(){return this.flags&1024})}})}var newStream=new FS.FSStream;for(var p in stream){newStream[p]=stream[p]}stream=newStream;var fd=FS.nextfd(fd_start,fd_end);stream.fd=fd;FS.streams[fd]=stream;return stream}),closeStream:(function(fd){FS.streams[fd]=null}),chrdev_stream_ops:{open:(function(stream){var device=FS.getDevice(stream.node.rdev);stream.stream_ops=device.stream_ops;if(stream.stream_ops.open){stream.stream_ops.open(stream)}}),llseek:(function(){throw new FS.ErrnoError(ERRNO_CODES.ESPIPE)})},major:(function(dev){return dev>>8}),minor:(function(dev){return dev&255}),makedev:(function(ma,mi){return ma<<8|mi}),registerDevice:(function(dev,ops){FS.devices[dev]={stream_ops:ops}}),getDevice:(function(dev){return FS.devices[dev]}),getMounts:(function(mount){var mounts=[];var check=[mount];while(check.length){var m=check.pop();mounts.push(m);check.push.apply(check,m.mounts)}return mounts}),syncfs:(function(populate,callback){if(typeof populate==="function"){callback=populate;populate=false}var mounts=FS.getMounts(FS.root.mount);var completed=0;function done(err){if(err){if(!done.errored){done.errored=true;return callback(err)}return}if(++completed>=mounts.length){callback(null)}}mounts.forEach((function(mount){if(!mount.type.syncfs){return done(null)}mount.type.syncfs(mount,populate,done)}))}),mount:(function(type,opts,mountpoint){var root=mountpoint==="/";var pseudo=!mountpoint;var node;if(root&&FS.root){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}else if(!root&&!pseudo){var lookup=FS.lookupPath(mountpoint,{follow_mount:false});mountpoint=lookup.path;node=lookup.node;if(FS.isMountpoint(node)){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}if(!FS.isDir(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR)}}var mount={type:type,opts:opts,mountpoint:mountpoint,mounts:[]};var mountRoot=type.mount(mount);mountRoot.mount=mount;mount.root=mountRoot;if(root){FS.root=mountRoot}else if(node){node.mounted=mount;if(node.mount){node.mount.mounts.push(mount)}}return mountRoot}),unmount:(function(mountpoint){var lookup=FS.lookupPath(mountpoint,{follow_mount:false});if(!FS.isMountpoint(lookup.node)){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var node=lookup.node;var mount=node.mounted;var mounts=FS.getMounts(mount);Object.keys(FS.nameTable).forEach((function(hash){var current=FS.nameTable[hash];while(current){var next=current.name_next;if(mounts.indexOf(current.mount)!==-1){FS.destroyNode(current)}current=next}}));node.mounted=null;var idx=node.mount.mounts.indexOf(mount);assert(idx!==-1);node.mount.mounts.splice(idx,1)}),lookup:(function(parent,name){return parent.node_ops.lookup(parent,name)}),mknod:(function(path,mode,dev){var lookup=FS.lookupPath(path,{parent:true});var parent=lookup.node;var name=PATH.basename(path);if(!name||name==="."||name===".."){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var err=FS.mayCreate(parent,name);if(err){throw new FS.ErrnoError(err)}if(!parent.node_ops.mknod){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}return parent.node_ops.mknod(parent,name,mode,dev)}),create:(function(path,mode){mode=mode!==undefined?mode:438;mode&=4095;mode|=32768;return FS.mknod(path,mode,0)}),mkdir:(function(path,mode){mode=mode!==undefined?mode:511;mode&=511|512;mode|=16384;return FS.mknod(path,mode,0)}),mkdev:(function(path,mode,dev){if(typeof dev==="undefined"){dev=mode;mode=438}mode|=8192;return FS.mknod(path,mode,dev)}),symlink:(function(oldpath,newpath){if(!PATH.resolve(oldpath)){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}var lookup=FS.lookupPath(newpath,{parent:true});var parent=lookup.node;if(!parent){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}var newname=PATH.basename(newpath);var err=FS.mayCreate(parent,newname);if(err){throw new FS.ErrnoError(err)}if(!parent.node_ops.symlink){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}return parent.node_ops.symlink(parent,newname,oldpath)}),rename:(function(old_path,new_path){var old_dirname=PATH.dirname(old_path);var new_dirname=PATH.dirname(new_path);var old_name=PATH.basename(old_path);var new_name=PATH.basename(new_path);var lookup,old_dir,new_dir;try{lookup=FS.lookupPath(old_path,{parent:true});old_dir=lookup.node;lookup=FS.lookupPath(new_path,{parent:true});new_dir=lookup.node}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}if(!old_dir||!new_dir)throw new FS.ErrnoError(ERRNO_CODES.ENOENT);if(old_dir.mount!==new_dir.mount){throw new FS.ErrnoError(ERRNO_CODES.EXDEV)}var old_node=FS.lookupNode(old_dir,old_name);var relative=PATH.relative(old_path,new_dirname);if(relative.charAt(0)!=="."){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}relative=PATH.relative(new_path,old_dirname);if(relative.charAt(0)!=="."){throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY)}var new_node;try{new_node=FS.lookupNode(new_dir,new_name)}catch(e){}if(old_node===new_node){return}var isdir=FS.isDir(old_node.mode);var err=FS.mayDelete(old_dir,old_name,isdir);if(err){throw new FS.ErrnoError(err)}err=new_node?FS.mayDelete(new_dir,new_name,isdir):FS.mayCreate(new_dir,new_name);if(err){throw new FS.ErrnoError(err)}if(!old_dir.node_ops.rename){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}if(FS.isMountpoint(old_node)||new_node&&FS.isMountpoint(new_node)){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}if(new_dir!==old_dir){err=FS.nodePermissions(old_dir,"w");if(err){throw new FS.ErrnoError(err)}}try{if(FS.trackingDelegate["willMovePath"]){FS.trackingDelegate["willMovePath"](old_path,new_path)}}catch(e){console.log("FS.trackingDelegate['willMovePath']('"+old_path+"', '"+new_path+"') threw an exception: "+e.message)}FS.hashRemoveNode(old_node);try{old_dir.node_ops.rename(old_node,new_dir,new_name)}catch(e){throw e}finally{FS.hashAddNode(old_node)}try{if(FS.trackingDelegate["onMovePath"])FS.trackingDelegate["onMovePath"](old_path,new_path)}catch(e){console.log("FS.trackingDelegate['onMovePath']('"+old_path+"', '"+new_path+"') threw an exception: "+e.message)}}),rmdir:(function(path){var lookup=FS.lookupPath(path,{parent:true});var parent=lookup.node;var name=PATH.basename(path);var node=FS.lookupNode(parent,name);var err=FS.mayDelete(parent,name,true);if(err){throw new FS.ErrnoError(err)}if(!parent.node_ops.rmdir){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}if(FS.isMountpoint(node)){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}try{if(FS.trackingDelegate["willDeletePath"]){FS.trackingDelegate["willDeletePath"](path)}}catch(e){console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: "+e.message)}parent.node_ops.rmdir(parent,name);FS.destroyNode(node);try{if(FS.trackingDelegate["onDeletePath"])FS.trackingDelegate["onDeletePath"](path)}catch(e){console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: "+e.message)}}),readdir:(function(path){var lookup=FS.lookupPath(path,{follow:true});var node=lookup.node;if(!node.node_ops.readdir){throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR)}return node.node_ops.readdir(node)}),unlink:(function(path){var lookup=FS.lookupPath(path,{parent:true});var parent=lookup.node;var name=PATH.basename(path);var node=FS.lookupNode(parent,name);var err=FS.mayDelete(parent,name,false);if(err){if(err===ERRNO_CODES.EISDIR)err=ERRNO_CODES.EPERM;throw new FS.ErrnoError(err)}if(!parent.node_ops.unlink){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}if(FS.isMountpoint(node)){throw new FS.ErrnoError(ERRNO_CODES.EBUSY)}try{if(FS.trackingDelegate["willDeletePath"]){FS.trackingDelegate["willDeletePath"](path)}}catch(e){console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: "+e.message)}parent.node_ops.unlink(parent,name);FS.destroyNode(node);try{if(FS.trackingDelegate["onDeletePath"])FS.trackingDelegate["onDeletePath"](path)}catch(e){console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: "+e.message)}}),readlink:(function(path){var lookup=FS.lookupPath(path);var link=lookup.node;if(!link){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}if(!link.node_ops.readlink){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}return PATH.resolve(FS.getPath(link.parent),link.node_ops.readlink(link))}),stat:(function(path,dontFollow){var lookup=FS.lookupPath(path,{follow:!dontFollow});var node=lookup.node;if(!node){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}if(!node.node_ops.getattr){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}return node.node_ops.getattr(node)}),lstat:(function(path){return FS.stat(path,true)}),chmod:(function(path,mode,dontFollow){var node;if(typeof path==="string"){var lookup=FS.lookupPath(path,{follow:!dontFollow});node=lookup.node}else{node=path}if(!node.node_ops.setattr){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}node.node_ops.setattr(node,{mode:mode&4095|node.mode&~4095,timestamp:Date.now()})}),lchmod:(function(path,mode){FS.chmod(path,mode,true)}),fchmod:(function(fd,mode){var stream=FS.getStream(fd);if(!stream){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}FS.chmod(stream.node,mode)}),chown:(function(path,uid,gid,dontFollow){var node;if(typeof path==="string"){var lookup=FS.lookupPath(path,{follow:!dontFollow});node=lookup.node}else{node=path}if(!node.node_ops.setattr){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}node.node_ops.setattr(node,{timestamp:Date.now()})}),lchown:(function(path,uid,gid){FS.chown(path,uid,gid,true)}),fchown:(function(fd,uid,gid){var stream=FS.getStream(fd);if(!stream){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}FS.chown(stream.node,uid,gid)}),truncate:(function(path,len){if(len<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var node;if(typeof path==="string"){var lookup=FS.lookupPath(path,{follow:true});node=lookup.node}else{node=path}if(!node.node_ops.setattr){throw new FS.ErrnoError(ERRNO_CODES.EPERM)}if(FS.isDir(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.EISDIR)}if(!FS.isFile(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var err=FS.nodePermissions(node,"w");if(err){throw new FS.ErrnoError(err)}node.node_ops.setattr(node,{size:len,timestamp:Date.now()})}),ftruncate:(function(fd,len){var stream=FS.getStream(fd);if(!stream){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}if((stream.flags&2097155)===0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}FS.truncate(stream.node,len)}),utime:(function(path,atime,mtime){var lookup=FS.lookupPath(path,{follow:true});var node=lookup.node;node.node_ops.setattr(node,{timestamp:Math.max(atime,mtime)})}),open:(function(path,flags,mode,fd_start,fd_end){if(path===""){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}flags=typeof flags==="string"?FS.modeStringToFlags(flags):flags;mode=typeof mode==="undefined"?438:mode;if(flags&64){mode=mode&4095|32768}else{mode=0}var node;if(typeof path==="object"){node=path}else{path=PATH.normalize(path);try{var lookup=FS.lookupPath(path,{follow:!(flags&131072)});node=lookup.node}catch(e){}}var created=false;if(flags&64){if(node){if(flags&128){throw new FS.ErrnoError(ERRNO_CODES.EEXIST)}}else{node=FS.mknod(path,mode,0);created=true}}if(!node){throw new FS.ErrnoError(ERRNO_CODES.ENOENT)}if(FS.isChrdev(node.mode)){flags&=~512}if(flags&65536&&!FS.isDir(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR)}if(!created){var err=FS.mayOpen(node,flags);if(err){throw new FS.ErrnoError(err)}}if(flags&512){FS.truncate(node,0)}flags&=~(128|512);var stream=FS.createStream({node:node,path:FS.getPath(node),flags:flags,seekable:true,position:0,stream_ops:node.stream_ops,ungotten:[],error:false},fd_start,fd_end);if(stream.stream_ops.open){stream.stream_ops.open(stream)}if(Module["logReadFiles"]&&!(flags&1)){if(!FS.readFiles)FS.readFiles={};if(!(path in FS.readFiles)){FS.readFiles[path]=1;Module["printErr"]("read file: "+path)}}try{if(FS.trackingDelegate["onOpenFile"]){var trackingFlags=0;if((flags&2097155)!==1){trackingFlags|=FS.tracking.openFlags.READ}if((flags&2097155)!==0){trackingFlags|=FS.tracking.openFlags.WRITE}FS.trackingDelegate["onOpenFile"](path,trackingFlags)}}catch(e){console.log("FS.trackingDelegate['onOpenFile']('"+path+"', flags) threw an exception: "+e.message)}return stream}),close:(function(stream){if(stream.getdents)stream.getdents=null;try{if(stream.stream_ops.close){stream.stream_ops.close(stream)}}catch(e){throw e}finally{FS.closeStream(stream.fd)}}),llseek:(function(stream,offset,whence){if(!stream.seekable||!stream.stream_ops.llseek){throw new FS.ErrnoError(ERRNO_CODES.ESPIPE)}stream.position=stream.stream_ops.llseek(stream,offset,whence);stream.ungotten=[];return stream.position}),read:(function(stream,buffer,offset,length,position){if(length<0||position<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}if((stream.flags&2097155)===1){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}if(FS.isDir(stream.node.mode)){throw new FS.ErrnoError(ERRNO_CODES.EISDIR)}if(!stream.stream_ops.read){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var seeking=true;if(typeof position==="undefined"){position=stream.position;seeking=false}else if(!stream.seekable){throw new FS.ErrnoError(ERRNO_CODES.ESPIPE)}var bytesRead=stream.stream_ops.read(stream,buffer,offset,length,position);if(!seeking)stream.position+=bytesRead;return bytesRead}),write:(function(stream,buffer,offset,length,position,canOwn){if(length<0||position<0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}if((stream.flags&2097155)===0){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}if(FS.isDir(stream.node.mode)){throw new FS.ErrnoError(ERRNO_CODES.EISDIR)}if(!stream.stream_ops.write){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}if(stream.flags&1024){FS.llseek(stream,0,2)}var seeking=true;if(typeof position==="undefined"){position=stream.position;seeking=false}else if(!stream.seekable){throw new FS.ErrnoError(ERRNO_CODES.ESPIPE)}var bytesWritten=stream.stream_ops.write(stream,buffer,offset,length,position,canOwn);if(!seeking)stream.position+=bytesWritten;try{if(stream.path&&FS.trackingDelegate["onWriteToFile"])FS.trackingDelegate["onWriteToFile"](stream.path)}catch(e){console.log("FS.trackingDelegate['onWriteToFile']('"+path+"') threw an exception: "+e.message)}return bytesWritten}),allocate:(function(stream,offset,length){if(offset<0||length<=0){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}if((stream.flags&2097155)===0){throw new FS.ErrnoError(ERRNO_CODES.EBADF)}if(!FS.isFile(stream.node.mode)&&!FS.isDir(node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENODEV)}if(!stream.stream_ops.allocate){throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP)}stream.stream_ops.allocate(stream,offset,length)}),mmap:(function(stream,buffer,offset,length,position,prot,flags){if((stream.flags&2097155)===1){throw new FS.ErrnoError(ERRNO_CODES.EACCES)}if(!stream.stream_ops.mmap){throw new FS.ErrnoError(ERRNO_CODES.ENODEV)}return stream.stream_ops.mmap(stream,buffer,offset,length,position,prot,flags)}),msync:(function(stream,buffer,offset,length,mmapFlags){if(!stream||!stream.stream_ops.msync){return 0}return stream.stream_ops.msync(stream,buffer,offset,length,mmapFlags)}),munmap:(function(stream){return 0}),ioctl:(function(stream,cmd,arg){if(!stream.stream_ops.ioctl){throw new FS.ErrnoError(ERRNO_CODES.ENOTTY)}return stream.stream_ops.ioctl(stream,cmd,arg)}),readFile:(function(path,opts){opts=opts||{};opts.flags=opts.flags||"r";opts.encoding=opts.encoding||"binary";if(opts.encoding!=="utf8"&&opts.encoding!=="binary"){throw new Error('Invalid encoding type "'+opts.encoding+'"')}var ret;var stream=FS.open(path,opts.flags);var stat=FS.stat(path);var length=stat.size;var buf=new Uint8Array(length);FS.read(stream,buf,0,length,0);if(opts.encoding==="utf8"){ret=UTF8ArrayToString(buf,0)}else if(opts.encoding==="binary"){ret=buf}FS.close(stream);return ret}),writeFile:(function(path,data,opts){opts=opts||{};opts.flags=opts.flags||"w";opts.encoding=opts.encoding||"utf8";if(opts.encoding!=="utf8"&&opts.encoding!=="binary"){throw new Error('Invalid encoding type "'+opts.encoding+'"')}var stream=FS.open(path,opts.flags,opts.mode);if(opts.encoding==="utf8"){var buf=new Uint8Array(lengthBytesUTF8(data)+1);var actualNumBytes=stringToUTF8Array(data,buf,0,buf.length);FS.write(stream,buf,0,actualNumBytes,0,opts.canOwn)}else if(opts.encoding==="binary"){FS.write(stream,data,0,data.length,0,opts.canOwn)}FS.close(stream)}),cwd:(function(){return FS.currentPath}),chdir:(function(path){var lookup=FS.lookupPath(path,{follow:true});if(!FS.isDir(lookup.node.mode)){throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR)}var err=FS.nodePermissions(lookup.node,"x");if(err){throw new FS.ErrnoError(err)}FS.currentPath=lookup.path}),createDefaultDirectories:(function(){FS.mkdir("/tmp");FS.mkdir("/home");FS.mkdir("/home/web_user")}),createDefaultDevices:(function(){FS.mkdir("/dev");FS.registerDevice(FS.makedev(1,3),{read:(function(){return 0}),write:(function(stream,buffer,offset,length,pos){return length})});FS.mkdev("/dev/null",FS.makedev(1,3));TTY.register(FS.makedev(5,0),TTY.default_tty_ops);TTY.register(FS.makedev(6,0),TTY.default_tty1_ops);FS.mkdev("/dev/tty",FS.makedev(5,0));FS.mkdev("/dev/tty1",FS.makedev(6,0));var random_device;if(typeof crypto!=="undefined"){var randomBuffer=new Uint8Array(1);random_device=(function(){crypto.getRandomValues(randomBuffer);return randomBuffer[0]})}else if(ENVIRONMENT_IS_NODE){random_device=(function(){return require("crypto").randomBytes(1)[0]})}else{random_device=(function(){return Math.random()*256|0})}FS.createDevice("/dev","random",random_device);FS.createDevice("/dev","urandom",random_device);FS.mkdir("/dev/shm");FS.mkdir("/dev/shm/tmp")}),createSpecialDirectories:(function(){FS.mkdir("/proc");FS.mkdir("/proc/self");FS.mkdir("/proc/self/fd");FS.mount({mount:(function(){var node=FS.createNode("/proc/self","fd",16384|511,73);node.node_ops={lookup:(function(parent,name){var fd=+name;var stream=FS.getStream(fd);if(!stream)throw new FS.ErrnoError(ERRNO_CODES.EBADF);var ret={parent:null,mount:{mountpoint:"fake"},node_ops:{readlink:(function(){return stream.path})}};ret.parent=ret;return ret})};return node})},{},"/proc/self/fd")}),createStandardStreams:(function(){if(Module["stdin"]){FS.createDevice("/dev","stdin",Module["stdin"])}else{FS.symlink("/dev/tty","/dev/stdin")}if(Module["stdout"]){FS.createDevice("/dev","stdout",null,Module["stdout"])}else{FS.symlink("/dev/tty","/dev/stdout")}if(Module["stderr"]){FS.createDevice("/dev","stderr",null,Module["stderr"])}else{FS.symlink("/dev/tty1","/dev/stderr")}var stdin=FS.open("/dev/stdin","r");assert(stdin.fd===0,"invalid handle for stdin ("+stdin.fd+")");var stdout=FS.open("/dev/stdout","w");assert(stdout.fd===1,"invalid handle for stdout ("+stdout.fd+")");var stderr=FS.open("/dev/stderr","w");assert(stderr.fd===2,"invalid handle for stderr ("+stderr.fd+")")}),ensureErrnoError:(function(){if(FS.ErrnoError)return;FS.ErrnoError=function ErrnoError(errno,node){this.node=node;this.setErrno=(function(errno){this.errno=errno;for(var key in ERRNO_CODES){if(ERRNO_CODES[key]===errno){this.code=key;break}}});this.setErrno(errno);this.message=ERRNO_MESSAGES[errno]};FS.ErrnoError.prototype=new Error;FS.ErrnoError.prototype.constructor=FS.ErrnoError;[ERRNO_CODES.ENOENT].forEach((function(code){FS.genericErrors[code]=new FS.ErrnoError(code);FS.genericErrors[code].stack="<generic error, no stack>"}))}),staticInit:(function(){FS.ensureErrnoError();FS.nameTable=new Array(4096);FS.mount(MEMFS,{},"/");FS.createDefaultDirectories();FS.createDefaultDevices();FS.createSpecialDirectories();FS.filesystems={"MEMFS":MEMFS,"IDBFS":IDBFS,"NODEFS":NODEFS,"WORKERFS":WORKERFS}}),init:(function(input,output,error){assert(!FS.init.initialized,"FS.init was previously called. If you want to initialize later with custom parameters, remove any earlier calls (note that one is automatically added to the generated code)");FS.init.initialized=true;FS.ensureErrnoError();Module["stdin"]=input||Module["stdin"];Module["stdout"]=output||Module["stdout"];Module["stderr"]=error||Module["stderr"];FS.createStandardStreams()}),quit:(function(){FS.init.initialized=false;var fflush=Module["_fflush"];if(fflush)fflush(0);for(var i=0;i<FS.streams.length;i++){var stream=FS.streams[i];if(!stream){continue}FS.close(stream)}}),getMode:(function(canRead,canWrite){var mode=0;if(canRead)mode|=292|73;if(canWrite)mode|=146;return mode}),joinPath:(function(parts,forceRelative){var path=PATH.join.apply(null,parts);if(forceRelative&&path[0]=="/")path=path.substr(1);return path}),absolutePath:(function(relative,base){return PATH.resolve(base,relative)}),standardizePath:(function(path){return PATH.normalize(path)}),findObject:(function(path,dontResolveLastLink){var ret=FS.analyzePath(path,dontResolveLastLink);if(ret.exists){return ret.object}else{___setErrNo(ret.error);return null}}),analyzePath:(function(path,dontResolveLastLink){try{var lookup=FS.lookupPath(path,{follow:!dontResolveLastLink});path=lookup.path}catch(e){}var ret={isRoot:false,exists:false,error:0,name:null,path:null,object:null,parentExists:false,parentPath:null,parentObject:null};try{var lookup=FS.lookupPath(path,{parent:true});ret.parentExists=true;ret.parentPath=lookup.path;ret.parentObject=lookup.node;ret.name=PATH.basename(path);lookup=FS.lookupPath(path,{follow:!dontResolveLastLink});ret.exists=true;ret.path=lookup.path;ret.object=lookup.node;ret.name=lookup.node.name;ret.isRoot=lookup.path==="/"}catch(e){ret.error=e.errno}return ret}),createFolder:(function(parent,name,canRead,canWrite){var path=PATH.join2(typeof parent==="string"?parent:FS.getPath(parent),name);var mode=FS.getMode(canRead,canWrite);return FS.mkdir(path,mode)}),createPath:(function(parent,path,canRead,canWrite){parent=typeof parent==="string"?parent:FS.getPath(parent);var parts=path.split("/").reverse();while(parts.length){var part=parts.pop();if(!part)continue;var current=PATH.join2(parent,part);try{FS.mkdir(current)}catch(e){}parent=current}return current}),createFile:(function(parent,name,properties,canRead,canWrite){var path=PATH.join2(typeof parent==="string"?parent:FS.getPath(parent),name);var mode=FS.getMode(canRead,canWrite);return FS.create(path,mode)}),createDataFile:(function(parent,name,data,canRead,canWrite,canOwn){var path=name?PATH.join2(typeof parent==="string"?parent:FS.getPath(parent),name):parent;var mode=FS.getMode(canRead,canWrite);var node=FS.create(path,mode);if(data){if(typeof data==="string"){var arr=new Array(data.length);for(var i=0,len=data.length;i<len;++i)arr[i]=data.charCodeAt(i);data=arr}FS.chmod(node,mode|146);var stream=FS.open(node,"w");FS.write(stream,data,0,data.length,0,canOwn);FS.close(stream);FS.chmod(node,mode)}return node}),createDevice:(function(parent,name,input,output){var path=PATH.join2(typeof parent==="string"?parent:FS.getPath(parent),name);var mode=FS.getMode(!!input,!!output);if(!FS.createDevice.major)FS.createDevice.major=64;var dev=FS.makedev(FS.createDevice.major++,0);FS.registerDevice(dev,{open:(function(stream){stream.seekable=false}),close:(function(stream){if(output&&output.buffer&&output.buffer.length){output(10)}}),read:(function(stream,buffer,offset,length,pos){var bytesRead=0;for(var i=0;i<length;i++){var result;try{result=input()}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EIO)}if(result===undefined&&bytesRead===0){throw new FS.ErrnoError(ERRNO_CODES.EAGAIN)}if(result===null||result===undefined)break;bytesRead++;buffer[offset+i]=result}if(bytesRead){stream.node.timestamp=Date.now()}return bytesRead}),write:(function(stream,buffer,offset,length,pos){for(var i=0;i<length;i++){try{output(buffer[offset+i])}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EIO)}}if(length){stream.node.timestamp=Date.now()}return i})});return FS.mkdev(path,mode,dev)}),createLink:(function(parent,name,target,canRead,canWrite){var path=PATH.join2(typeof parent==="string"?parent:FS.getPath(parent),name);return FS.symlink(target,path)}),forceLoadFile:(function(obj){if(obj.isDevice||obj.isFolder||obj.link||obj.contents)return true;var success=true;if(typeof XMLHttpRequest!=="undefined"){throw new Error("Lazy loading should have been performed (contents set) in createLazyFile, but it was not. Lazy loading only works in web workers. Use --embed-file or --preload-file in emcc on the main thread.")}else if(Module["read"]){try{obj.contents=intArrayFromString(Module["read"](obj.url),true);obj.usedBytes=obj.contents.length}catch(e){success=false}}else{throw new Error("Cannot load without read() or XMLHttpRequest.")}if(!success)___setErrNo(ERRNO_CODES.EIO);return success}),createLazyFile:(function(parent,name,url,canRead,canWrite){function LazyUint8Array(){this.lengthKnown=false;this.chunks=[]}LazyUint8Array.prototype.get=function LazyUint8Array_get(idx){if(idx>this.length-1||idx<0){return undefined}var chunkOffset=idx%this.chunkSize;var chunkNum=idx/this.chunkSize|0;return this.getter(chunkNum)[chunkOffset]};LazyUint8Array.prototype.setDataGetter=function LazyUint8Array_setDataGetter(getter){this.getter=getter};LazyUint8Array.prototype.cacheLength=function LazyUint8Array_cacheLength(){var xhr=new XMLHttpRequest;xhr.open("HEAD",url,false);xhr.send(null);if(!(xhr.status>=200&&xhr.status<300||xhr.status===304))throw new Error("Couldn't load "+url+". Status: "+xhr.status);var datalength=Number(xhr.getResponseHeader("Content-length"));var header;var hasByteServing=(header=xhr.getResponseHeader("Accept-Ranges"))&&header==="bytes";var chunkSize=1024*1024;if(!hasByteServing)chunkSize=datalength;var doXHR=(function(from,to){if(from>to)throw new Error("invalid range ("+from+", "+to+") or no bytes requested!");if(to>datalength-1)throw new Error("only "+datalength+" bytes available! programmer error!");var xhr=new XMLHttpRequest;xhr.open("GET",url,false);if(datalength!==chunkSize)xhr.setRequestHeader("Range","bytes="+from+"-"+to);if(typeof Uint8Array!="undefined")xhr.responseType="arraybuffer";if(xhr.overrideMimeType){xhr.overrideMimeType("text/plain; charset=x-user-defined")}xhr.send(null);if(!(xhr.status>=200&&xhr.status<300||xhr.status===304))throw new Error("Couldn't load "+url+". Status: "+xhr.status);if(xhr.response!==undefined){return new Uint8Array(xhr.response||[])}else{return intArrayFromString(xhr.responseText||"",true)}});var lazyArray=this;lazyArray.setDataGetter((function(chunkNum){var start=chunkNum*chunkSize;var end=(chunkNum+1)*chunkSize-1;end=Math.min(end,datalength-1);if(typeof lazyArray.chunks[chunkNum]==="undefined"){lazyArray.chunks[chunkNum]=doXHR(start,end)}if(typeof lazyArray.chunks[chunkNum]==="undefined")throw new Error("doXHR failed!");return lazyArray.chunks[chunkNum]}));this._length=datalength;this._chunkSize=chunkSize;this.lengthKnown=true};if(typeof XMLHttpRequest!=="undefined"){if(!ENVIRONMENT_IS_WORKER)throw"Cannot do synchronous binary XHRs outside webworkers in modern browsers. Use --embed-file or --preload-file in emcc";var lazyArray=new LazyUint8Array;Object.defineProperty(lazyArray,"length",{get:(function(){if(!this.lengthKnown){this.cacheLength()}return this._length})});Object.defineProperty(lazyArray,"chunkSize",{get:(function(){if(!this.lengthKnown){this.cacheLength()}return this._chunkSize})});var properties={isDevice:false,contents:lazyArray}}else{var properties={isDevice:false,url:url}}var node=FS.createFile(parent,name,properties,canRead,canWrite);if(properties.contents){node.contents=properties.contents}else if(properties.url){node.contents=null;node.url=properties.url}Object.defineProperty(node,"usedBytes",{get:(function(){return this.contents.length})});var stream_ops={};var keys=Object.keys(node.stream_ops);keys.forEach((function(key){var fn=node.stream_ops[key];stream_ops[key]=function forceLoadLazyFile(){if(!FS.forceLoadFile(node)){throw new FS.ErrnoError(ERRNO_CODES.EIO)}return fn.apply(null,arguments)}}));stream_ops.read=function stream_ops_read(stream,buffer,offset,length,position){if(!FS.forceLoadFile(node)){throw new FS.ErrnoError(ERRNO_CODES.EIO)}var contents=stream.node.contents;if(position>=contents.length)return 0;var size=Math.min(contents.length-position,length);assert(size>=0);if(contents.slice){for(var i=0;i<size;i++){buffer[offset+i]=contents[position+i]}}else{for(var i=0;i<size;i++){buffer[offset+i]=contents.get(position+i)}}return size};node.stream_ops=stream_ops;return node}),createPreloadedFile:(function(parent,name,url,canRead,canWrite,onload,onerror,dontCreateFile,canOwn,preFinish){Browser.init();var fullname=name?PATH.resolve(PATH.join2(parent,name)):parent;var dep=getUniqueRunDependency("cp "+fullname);function processData(byteArray){function finish(byteArray){if(preFinish)preFinish();if(!dontCreateFile){FS.createDataFile(parent,name,byteArray,canRead,canWrite,canOwn)}if(onload)onload();removeRunDependency(dep)}var handled=false;Module["preloadPlugins"].forEach((function(plugin){if(handled)return;if(plugin["canHandle"](fullname)){plugin["handle"](byteArray,fullname,finish,(function(){if(onerror)onerror();removeRunDependency(dep)}));handled=true}}));if(!handled)finish(byteArray)}addRunDependency(dep);if(typeof url=="string"){Browser.asyncLoad(url,(function(byteArray){processData(byteArray)}),onerror)}else{processData(url)}}),indexedDB:(function(){return window.indexedDB||window.mozIndexedDB||window.webkitIndexedDB||window.msIndexedDB}),DB_NAME:(function(){return"EM_FS_"+window.location.pathname}),DB_VERSION:20,DB_STORE_NAME:"FILE_DATA",saveFilesToDB:(function(paths,onload,onerror){onload=onload||(function(){});onerror=onerror||(function(){});var indexedDB=FS.indexedDB();try{var openRequest=indexedDB.open(FS.DB_NAME(),FS.DB_VERSION)}catch(e){return onerror(e)}openRequest.onupgradeneeded=function openRequest_onupgradeneeded(){console.log("creating db");var db=openRequest.result;db.createObjectStore(FS.DB_STORE_NAME)};openRequest.onsuccess=function openRequest_onsuccess(){var db=openRequest.result;var transaction=db.transaction([FS.DB_STORE_NAME],"readwrite");var files=transaction.objectStore(FS.DB_STORE_NAME);var ok=0,fail=0,total=paths.length;function finish(){if(fail==0)onload();else onerror()}paths.forEach((function(path){var putRequest=files.put(FS.analyzePath(path).object.contents,path);putRequest.onsuccess=function putRequest_onsuccess(){ok++;if(ok+fail==total)finish()};putRequest.onerror=function putRequest_onerror(){fail++;if(ok+fail==total)finish()}}));transaction.onerror=onerror};openRequest.onerror=onerror}),loadFilesFromDB:(function(paths,onload,onerror){onload=onload||(function(){});onerror=onerror||(function(){});var indexedDB=FS.indexedDB();try{var openRequest=indexedDB.open(FS.DB_NAME(),FS.DB_VERSION)}catch(e){return onerror(e)}openRequest.onupgradeneeded=onerror;openRequest.onsuccess=function openRequest_onsuccess(){var db=openRequest.result;try{var transaction=db.transaction([FS.DB_STORE_NAME],"readonly")}catch(e){onerror(e);return}var files=transaction.objectStore(FS.DB_STORE_NAME);var ok=0,fail=0,total=paths.length;function finish(){if(fail==0)onload();else onerror()}paths.forEach((function(path){var getRequest=files.get(path);getRequest.onsuccess=function getRequest_onsuccess(){if(FS.analyzePath(path).exists){FS.unlink(path)}FS.createDataFile(PATH.dirname(path),PATH.basename(path),getRequest.result,true,true,true);ok++;if(ok+fail==total)finish()};getRequest.onerror=function getRequest_onerror(){fail++;if(ok+fail==total)finish()}}));transaction.onerror=onerror};openRequest.onerror=onerror})};var SYSCALLS={DEFAULT_POLLMASK:5,mappings:{},umask:511,calculateAt:(function(dirfd,path){if(path[0]!=="/"){var dir;if(dirfd===-100){dir=FS.cwd()}else{var dirstream=FS.getStream(dirfd);if(!dirstream)throw new FS.ErrnoError(ERRNO_CODES.EBADF);dir=dirstream.path}path=PATH.join2(dir,path)}return path}),doStat:(function(func,path,buf){try{var stat=func(path)}catch(e){if(e&&e.node&&PATH.normalize(path)!==PATH.normalize(FS.getPath(e.node))){return-ERRNO_CODES.ENOTDIR}throw e}HEAP32[buf>>2]=stat.dev;HEAP32[buf+4>>2]=0;HEAP32[buf+8>>2]=stat.ino;HEAP32[buf+12>>2]=stat.mode;HEAP32[buf+16>>2]=stat.nlink;HEAP32[buf+20>>2]=stat.uid;HEAP32[buf+24>>2]=stat.gid;HEAP32[buf+28>>2]=stat.rdev;HEAP32[buf+32>>2]=0;HEAP32[buf+36>>2]=stat.size;HEAP32[buf+40>>2]=4096;HEAP32[buf+44>>2]=stat.blocks;HEAP32[buf+48>>2]=stat.atime.getTime()/1e3|0;HEAP32[buf+52>>2]=0;HEAP32[buf+56>>2]=stat.mtime.getTime()/1e3|0;HEAP32[buf+60>>2]=0;HEAP32[buf+64>>2]=stat.ctime.getTime()/1e3|0;HEAP32[buf+68>>2]=0;HEAP32[buf+72>>2]=stat.ino;return 0}),doMsync:(function(addr,stream,len,flags){var buffer=new Uint8Array(HEAPU8.subarray(addr,addr+len));FS.msync(stream,buffer,0,len,flags)}),doMkdir:(function(path,mode){path=PATH.normalize(path);if(path[path.length-1]==="/")path=path.substr(0,path.length-1);FS.mkdir(path,mode,0);return 0}),doMknod:(function(path,mode,dev){switch(mode&61440){case 32768:case 8192:case 24576:case 4096:case 49152:break;default:return-ERRNO_CODES.EINVAL}FS.mknod(path,mode,dev);return 0}),doReadlink:(function(path,buf,bufsize){if(bufsize<=0)return-ERRNO_CODES.EINVAL;var ret=FS.readlink(path);ret=ret.slice(0,Math.max(0,bufsize));writeStringToMemory(ret,buf,true);return ret.length}),doAccess:(function(path,amode){if(amode&~7){return-ERRNO_CODES.EINVAL}var node;var lookup=FS.lookupPath(path,{follow:true});node=lookup.node;var perms="";if(amode&4)perms+="r";if(amode&2)perms+="w";if(amode&1)perms+="x";if(perms&&FS.nodePermissions(node,perms)){return-ERRNO_CODES.EACCES}return 0}),doDup:(function(path,flags,suggestFD){var suggest=FS.getStream(suggestFD);if(suggest)FS.close(suggest);return FS.open(path,flags,0,suggestFD,suggestFD).fd}),doReadv:(function(stream,iov,iovcnt,offset){var ret=0;for(var i=0;i<iovcnt;i++){var ptr=HEAP32[iov+i*8>>2];var len=HEAP32[iov+(i*8+4)>>2];var curr=FS.read(stream,HEAP8,ptr,len,offset);if(curr<0)return-1;ret+=curr;if(curr<len)break}return ret}),doWritev:(function(stream,iov,iovcnt,offset){var ret=0;for(var i=0;i<iovcnt;i++){var ptr=HEAP32[iov+i*8>>2];var len=HEAP32[iov+(i*8+4)>>2];var curr=FS.write(stream,HEAP8,ptr,len,offset);if(curr<0)return-1;ret+=curr}return ret}),varargs:0,get:(function(varargs){SYSCALLS.varargs+=4;var ret=HEAP32[SYSCALLS.varargs-4>>2];return ret}),getStr:(function(){var ret=Pointer_stringify(SYSCALLS.get());return ret}),getStreamFromFD:(function(){var stream=FS.getStream(SYSCALLS.get());if(!stream)throw new FS.ErrnoError(ERRNO_CODES.EBADF);return stream}),getSocketFromFD:(function(){var socket=SOCKFS.getSocket(SYSCALLS.get());if(!socket)throw new FS.ErrnoError(ERRNO_CODES.EBADF);return socket}),getSocketAddress:(function(allowNull){var addrp=SYSCALLS.get(),addrlen=SYSCALLS.get();if(allowNull&&addrp===0)return null;var info=__read_sockaddr(addrp,addrlen);if(info.errno)throw new FS.ErrnoError(info.errno);info.addr=DNS.lookup_addr(info.addr)||info.addr;return info}),get64:(function(){var low=SYSCALLS.get(),high=SYSCALLS.get();if(low>=0)assert(high===0);else assert(high===-1);return low}),getZero:(function(){assert(SYSCALLS.get()===0)})};function ___syscall195(which,varargs){SYSCALLS.varargs=varargs;try{var path=SYSCALLS.getStr(),buf=SYSCALLS.get();return SYSCALLS.doStat(FS.stat,path,buf)}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall194(which,varargs){SYSCALLS.varargs=varargs;try{var fd=SYSCALLS.get(),zero=SYSCALLS.getZero(),length=SYSCALLS.get64();FS.ftruncate(fd,length);return 0}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}var _emscripten_check_longjmp=true;var _DtoILow=true;var _emscripten_cleanup_setjmp=true;var _emscripten_postinvoke=true;function _realloc(){throw"bad"}Module["_realloc"]=_realloc;Module["_saveSetjmp"]=_saveSetjmp;function __exit(status){Module["exit"](status)}function _exit(status){__exit(status)}function ___syscall6(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD();FS.close(stream);return 0}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall54(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),op=SYSCALLS.get();switch(op){case 21505:{if(!stream.tty)return-ERRNO_CODES.ENOTTY;return 0};case 21506:{if(!stream.tty)return-ERRNO_CODES.ENOTTY;return 0};case 21519:{if(!stream.tty)return-ERRNO_CODES.ENOTTY;var argp=SYSCALLS.get();HEAP32[argp>>2]=0;return 0};case 21520:{if(!stream.tty)return-ERRNO_CODES.ENOTTY;return-ERRNO_CODES.EINVAL};case 21531:{var argp=SYSCALLS.get();return FS.ioctl(stream,op,argp)};default:abort("bad ioctl syscall "+op)}}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}Module["_bitshift64Ashr"]=_bitshift64Ashr;Module["_bitshift64Lshr"]=_bitshift64Lshr;var _emscripten_prep_setjmp=true;var _BDtoIHigh=true;function _pthread_cleanup_push(routine,arg){__ATEXIT__.push((function(){Runtime.dynCall("vi",routine,[arg])}));_pthread_cleanup_push.level=__ATEXIT__.length}Module["_testSetjmp"]=_testSetjmp;function _longjmp(env,value){asm["setThrew"](env,value||1);throw"longjmp"}function _emscripten_longjmp(env,value){_longjmp(env,value)}var _environ=allocate(1,"i32*",ALLOC_STATIC);var ___environ=_environ;function ___buildEnvironment(env){var MAX_ENV_VALUES=64;var TOTAL_ENV_SIZE=1024;var poolPtr;var envPtr;if(!___buildEnvironment.called){___buildEnvironment.called=true;ENV["USER"]=ENV["LOGNAME"]="web_user";ENV["PATH"]="/";ENV["PWD"]="/";ENV["HOME"]="/home/web_user";ENV["LANG"]="C";ENV["_"]=Module["thisProgram"];poolPtr=allocate(TOTAL_ENV_SIZE,"i8",ALLOC_STATIC);envPtr=allocate(MAX_ENV_VALUES*4,"i8*",ALLOC_STATIC);HEAP32[envPtr>>2]=poolPtr;HEAP32[_environ>>2]=envPtr}else{envPtr=HEAP32[_environ>>2];poolPtr=HEAP32[envPtr>>2]}var strings=[];var totalSize=0;for(var key in env){if(typeof env[key]==="string"){var line=key+"="+env[key];strings.push(line);totalSize+=line.length}}if(totalSize>TOTAL_ENV_SIZE){throw new Error("Environment size exceeded TOTAL_ENV_SIZE!")}var ptrSize=4;for(var i=0;i<strings.length;i++){var line=strings[i];writeAsciiToMemory(line,poolPtr);HEAP32[envPtr+i*ptrSize>>2]=poolPtr;poolPtr+=line.length+1}HEAP32[envPtr+strings.length*ptrSize>>2]=0}var ENV={};function _getenv(name){if(name===0)return 0;name=Pointer_stringify(name);if(!ENV.hasOwnProperty(name))return 0;if(_getenv.ret)_free(_getenv.ret);_getenv.ret=allocate(intArrayFromString(ENV[name]),"i8",ALLOC_NORMAL);return _getenv.ret}function _pthread_cleanup_pop(){assert(_pthread_cleanup_push.level==__ATEXIT__.length,"cannot pop if something else added meanwhile!");__ATEXIT__.pop();_pthread_cleanup_push.level=__ATEXIT__.length}function _getpwnam(){throw"getpwnam: TODO"}function _emscripten_memcpy_big(dest,src,num){HEAPU8.set(HEAPU8.subarray(src,src+num),dest);return dest}Module["_memcpy"]=_memcpy;var _llvm_pow_f64=Math_pow;function _sbrk(bytes){var self=_sbrk;if(!self.called){DYNAMICTOP=alignMemoryPage(DYNAMICTOP);self.called=true;assert(Runtime.dynamicAlloc);self.alloc=Runtime.dynamicAlloc;Runtime.dynamicAlloc=(function(){abort("cannot dynamically allocate, sbrk now has control")})}var ret=DYNAMICTOP;if(bytes!=0){var success=self.alloc(bytes);if(!success)return-1>>>0}return ret}Module["_memmove"]=_memmove;var __sigalrm_handler=0;function _signal(sig,func){if(sig==14){__sigalrm_handler=func}else{}return 0}var _emscripten_preinvoke=true;var _BItoD=true;Module["_llvm_bswap_i32"]=_llvm_bswap_i32;function ___syscall40(which,varargs){SYSCALLS.varargs=varargs;try{var path=SYSCALLS.getStr();FS.rmdir(path);return 0}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function _sysconf(name){switch(name){case 30:return PAGE_SIZE;case 85:return totalMemory/PAGE_SIZE;case 132:case 133:case 12:case 137:case 138:case 15:case 235:case 16:case 17:case 18:case 19:case 20:case 149:case 13:case 10:case 236:case 153:case 9:case 21:case 22:case 159:case 154:case 14:case 77:case 78:case 139:case 80:case 81:case 82:case 68:case 67:case 164:case 11:case 29:case 47:case 48:case 95:case 52:case 51:case 46:return 200809;case 79:return 0;case 27:case 246:case 127:case 128:case 23:case 24:case 160:case 161:case 181:case 182:case 242:case 183:case 184:case 243:case 244:case 245:case 165:case 178:case 179:case 49:case 50:case 168:case 169:case 175:case 170:case 171:case 172:case 97:case 76:case 32:case 173:case 35:return-1;case 176:case 177:case 7:case 155:case 8:case 157:case 125:case 126:case 92:case 93:case 129:case 130:case 131:case 94:case 91:return 1;case 74:case 60:case 69:case 70:case 4:return 1024;case 31:case 42:case 72:return 32;case 87:case 26:case 33:return 2147483647;case 34:case 1:return 47839;case 38:case 36:return 99;case 43:case 37:return 2048;case 0:return 2097152;case 3:return 65536;case 28:return 32768;case 44:return 32767;case 75:return 16384;case 39:return 1e3;case 89:return 700;case 71:return 256;case 40:return 255;case 2:return 100;case 180:return 64;case 25:return 20;case 5:return 16;case 6:return 6;case 73:return 4;case 84:{if(typeof navigator==="object")return navigator["hardwareConcurrency"]||1;return 1}}___setErrNo(ERRNO_CODES.EINVAL);return-1}var _SItoD=true;var _acos=Math_acos;var _SItoF=true;Module["_memset"]=_memset;var _BDtoILow=true;var _DtoIHigh=true;Module["_bitshift64Shl"]=_bitshift64Shl;function _abort(){Module["abort"]()}var _tan=Math_tan;function ___lock(){}function ___unlock(){}var _asin=Math_asin;var _fabs=Math_abs;var _floor=Math_floor;function __inet_pton4_raw(str){var b=str.split(".");for(var i=0;i<4;i++){var tmp=Number(b[i]);if(isNaN(tmp))return null;b[i]=tmp}return(b[0]|b[1]<<8|b[2]<<16|b[3]<<24)>>>0}function _inet_addr(ptr){var addr=__inet_pton4_raw(Pointer_stringify(ptr));if(addr===null){return-1}return addr}var _sqrt=Math_sqrt;function _emscripten_set_main_loop_timing(mode,value){Browser.mainLoop.timingMode=mode;Browser.mainLoop.timingValue=value;if(!Browser.mainLoop.func){return 1}if(mode==0){Browser.mainLoop.scheduler=function Browser_mainLoop_scheduler_setTimeout(){setTimeout(Browser.mainLoop.runner,value)};Browser.mainLoop.method="timeout"}else if(mode==1){Browser.mainLoop.scheduler=function Browser_mainLoop_scheduler_rAF(){Browser.requestAnimationFrame(Browser.mainLoop.runner)};Browser.mainLoop.method="rAF"}else if(mode==2){if(!window["setImmediate"]){var setImmediates=[];var emscriptenMainLoopMessageId="__emcc";function Browser_setImmediate_messageHandler(event){if(event.source===window&&event.data===emscriptenMainLoopMessageId){event.stopPropagation();setImmediates.shift()()}}window.addEventListener("message",Browser_setImmediate_messageHandler,true);window["setImmediate"]=function Browser_emulated_setImmediate(func){setImmediates.push(func);window.postMessage(emscriptenMainLoopMessageId,"*")}}Browser.mainLoop.scheduler=function Browser_mainLoop_scheduler_setImmediate(){window["setImmediate"](Browser.mainLoop.runner)};Browser.mainLoop.method="immediate"}return 0}function _emscripten_set_main_loop(func,fps,simulateInfiniteLoop,arg,noSetTiming){Module["noExitRuntime"]=true;assert(!Browser.mainLoop.func,"emscripten_set_main_loop: there can only be one main loop function at once: call emscripten_cancel_main_loop to cancel the previous one before setting a new one with different parameters.");Browser.mainLoop.func=func;Browser.mainLoop.arg=arg;var thisMainLoopId=Browser.mainLoop.currentlyRunningMainloop;Browser.mainLoop.runner=function Browser_mainLoop_runner(){if(ABORT)return;if(Browser.mainLoop.queue.length>0){var start=Date.now();var blocker=Browser.mainLoop.queue.shift();blocker.func(blocker.arg);if(Browser.mainLoop.remainingBlockers){var remaining=Browser.mainLoop.remainingBlockers;var next=remaining%1==0?remaining-1:Math.floor(remaining);if(blocker.counted){Browser.mainLoop.remainingBlockers=next}else{next=next+.5;Browser.mainLoop.remainingBlockers=(8*remaining+next)/9}}console.log('main loop blocker "'+blocker.name+'" took '+(Date.now()-start)+" ms");Browser.mainLoop.updateStatus();setTimeout(Browser.mainLoop.runner,0);return}if(thisMainLoopId<Browser.mainLoop.currentlyRunningMainloop)return;Browser.mainLoop.currentFrameNumber=Browser.mainLoop.currentFrameNumber+1|0;if(Browser.mainLoop.timingMode==1&&Browser.mainLoop.timingValue>1&&Browser.mainLoop.currentFrameNumber%Browser.mainLoop.timingValue!=0){Browser.mainLoop.scheduler();return}if(Browser.mainLoop.method==="timeout"&&Module.ctx){Module.printErr("Looks like you are rendering without using requestAnimationFrame for the main loop. You should use 0 for the frame rate in emscripten_set_main_loop in order to use requestAnimationFrame, as that can greatly improve your frame rates!");Browser.mainLoop.method=""}Browser.mainLoop.runIter((function(){if(typeof arg!=="undefined"){Runtime.dynCall("vi",func,[arg])}else{Runtime.dynCall("v",func)}}));if(thisMainLoopId<Browser.mainLoop.currentlyRunningMainloop)return;if(typeof SDL==="object"&&SDL.audio&&SDL.audio.queueNewAudioData)SDL.audio.queueNewAudioData();Browser.mainLoop.scheduler()};if(!noSetTiming){if(fps&&fps>0)_emscripten_set_main_loop_timing(0,1e3/fps);else _emscripten_set_main_loop_timing(1,1);Browser.mainLoop.scheduler()}if(simulateInfiniteLoop){throw"SimulateInfiniteLoop"}}var Browser={mainLoop:{scheduler:null,method:"",currentlyRunningMainloop:0,func:null,arg:0,timingMode:0,timingValue:0,currentFrameNumber:0,queue:[],pause:(function(){Browser.mainLoop.scheduler=null;Browser.mainLoop.currentlyRunningMainloop++}),resume:(function(){Browser.mainLoop.currentlyRunningMainloop++;var timingMode=Browser.mainLoop.timingMode;var timingValue=Browser.mainLoop.timingValue;var func=Browser.mainLoop.func;Browser.mainLoop.func=null;_emscripten_set_main_loop(func,0,false,Browser.mainLoop.arg,true);_emscripten_set_main_loop_timing(timingMode,timingValue);Browser.mainLoop.scheduler()}),updateStatus:(function(){if(Module["setStatus"]){var message=Module["statusMessage"]||"Please wait...";var remaining=Browser.mainLoop.remainingBlockers;var expected=Browser.mainLoop.expectedBlockers;if(remaining){if(remaining<expected){Module["setStatus"](message+" ("+(expected-remaining)+"/"+expected+")")}else{Module["setStatus"](message)}}else{Module["setStatus"]("")}}}),runIter:(function(func){if(ABORT)return;if(Module["preMainLoop"]){var preRet=Module["preMainLoop"]();if(preRet===false){return}}try{func()}catch(e){if(e instanceof ExitStatus){return}else{if(e&&typeof e==="object"&&e.stack)Module.printErr("exception thrown: "+[e,e.stack]);throw e}}if(Module["postMainLoop"])Module["postMainLoop"]()})},isFullScreen:false,pointerLock:false,moduleContextCreatedCallbacks:[],workers:[],init:(function(){if(!Module["preloadPlugins"])Module["preloadPlugins"]=[];if(Browser.initted)return;Browser.initted=true;try{new Blob;Browser.hasBlobConstructor=true}catch(e){Browser.hasBlobConstructor=false;console.log("warning: no blob constructor, cannot create blobs with mimetypes")}Browser.BlobBuilder=typeof MozBlobBuilder!="undefined"?MozBlobBuilder:typeof WebKitBlobBuilder!="undefined"?WebKitBlobBuilder:!Browser.hasBlobConstructor?console.log("warning: no BlobBuilder"):null;Browser.URLObject=typeof window!="undefined"?window.URL?window.URL:window.webkitURL:undefined;if(!Module.noImageDecoding&&typeof Browser.URLObject==="undefined"){console.log("warning: Browser does not support creating object URLs. Built-in browser image decoding will not be available.");Module.noImageDecoding=true}var imagePlugin={};imagePlugin["canHandle"]=function imagePlugin_canHandle(name){return!Module.noImageDecoding&&/\.(jpg|jpeg|png|bmp)$/i.test(name)};imagePlugin["handle"]=function imagePlugin_handle(byteArray,name,onload,onerror){var b=null;if(Browser.hasBlobConstructor){try{b=new Blob([byteArray],{type:Browser.getMimetype(name)});if(b.size!==byteArray.length){b=new Blob([(new Uint8Array(byteArray)).buffer],{type:Browser.getMimetype(name)})}}catch(e){Runtime.warnOnce("Blob constructor present but fails: "+e+"; falling back to blob builder")}}if(!b){var bb=new Browser.BlobBuilder;bb.append((new Uint8Array(byteArray)).buffer);b=bb.getBlob()}var url=Browser.URLObject.createObjectURL(b);var img=new Image;img.onload=function img_onload(){assert(img.complete,"Image "+name+" could not be decoded");var canvas=document.createElement("canvas");canvas.width=img.width;canvas.height=img.height;var ctx=canvas.getContext("2d");ctx.drawImage(img,0,0);Module["preloadedImages"][name]=canvas;Browser.URLObject.revokeObjectURL(url);if(onload)onload(byteArray)};img.onerror=function img_onerror(event){console.log("Image "+url+" could not be decoded");if(onerror)onerror()};img.src=url};Module["preloadPlugins"].push(imagePlugin);var audioPlugin={};audioPlugin["canHandle"]=function audioPlugin_canHandle(name){return!Module.noAudioDecoding&&name.substr(-4)in{".ogg":1,".wav":1,".mp3":1}};audioPlugin["handle"]=function audioPlugin_handle(byteArray,name,onload,onerror){var done=false;function finish(audio){if(done)return;done=true;Module["preloadedAudios"][name]=audio;if(onload)onload(byteArray)}function fail(){if(done)return;done=true;Module["preloadedAudios"][name]=new Audio;if(onerror)onerror()}if(Browser.hasBlobConstructor){try{var b=new Blob([byteArray],{type:Browser.getMimetype(name)})}catch(e){return fail()}var url=Browser.URLObject.createObjectURL(b);var audio=new Audio;audio.addEventListener("canplaythrough",(function(){finish(audio)}),false);audio.onerror=function audio_onerror(event){if(done)return;console.log("warning: browser could not fully decode audio "+name+", trying slower base64 approach");function encode64(data){var BASE="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";var PAD="=";var ret="";var leftchar=0;var leftbits=0;for(var i=0;i<data.length;i++){leftchar=leftchar<<8|data[i];leftbits+=8;while(leftbits>=6){var curr=leftchar>>leftbits-6&63;leftbits-=6;ret+=BASE[curr]}}if(leftbits==2){ret+=BASE[(leftchar&3)<<4];ret+=PAD+PAD}else if(leftbits==4){ret+=BASE[(leftchar&15)<<2];ret+=PAD}return ret}audio.src="data:audio/x-"+name.substr(-3)+";base64,"+encode64(byteArray);finish(audio)};audio.src=url;Browser.safeSetTimeout((function(){finish(audio)}),1e4)}else{return fail()}};Module["preloadPlugins"].push(audioPlugin);var canvas=Module["canvas"];function pointerLockChange(){Browser.pointerLock=document["pointerLockElement"]===canvas||document["mozPointerLockElement"]===canvas||document["webkitPointerLockElement"]===canvas||document["msPointerLockElement"]===canvas}if(canvas){canvas.requestPointerLock=canvas["requestPointerLock"]||canvas["mozRequestPointerLock"]||canvas["webkitRequestPointerLock"]||canvas["msRequestPointerLock"]||(function(){});canvas.exitPointerLock=document["exitPointerLock"]||document["mozExitPointerLock"]||document["webkitExitPointerLock"]||document["msExitPointerLock"]||(function(){});canvas.exitPointerLock=canvas.exitPointerLock.bind(document);document.addEventListener("pointerlockchange",pointerLockChange,false);document.addEventListener("mozpointerlockchange",pointerLockChange,false);document.addEventListener("webkitpointerlockchange",pointerLockChange,false);document.addEventListener("mspointerlockchange",pointerLockChange,false);if(Module["elementPointerLock"]){canvas.addEventListener("click",(function(ev){if(!Browser.pointerLock&&canvas.requestPointerLock){canvas.requestPointerLock();ev.preventDefault()}}),false)}}}),createContext:(function(canvas,useWebGL,setInModule,webGLContextAttributes){if(useWebGL&&Module.ctx&&canvas==Module.canvas)return Module.ctx;var ctx;var contextHandle;if(useWebGL){var contextAttributes={antialias:false,alpha:false};if(webGLContextAttributes){for(var attribute in webGLContextAttributes){contextAttributes[attribute]=webGLContextAttributes[attribute]}}contextHandle=GL.createContext(canvas,contextAttributes);if(contextHandle){ctx=GL.getContext(contextHandle).GLctx}canvas.style.backgroundColor="black"}else{ctx=canvas.getContext("2d")}if(!ctx)return null;if(setInModule){if(!useWebGL)assert(typeof GLctx==="undefined","cannot set in module if GLctx is used, but we are a non-GL context that would replace it");Module.ctx=ctx;if(useWebGL)GL.makeContextCurrent(contextHandle);Module.useWebGL=useWebGL;Browser.moduleContextCreatedCallbacks.forEach((function(callback){callback()}));Browser.init()}return ctx}),destroyContext:(function(canvas,useWebGL,setInModule){}),fullScreenHandlersInstalled:false,lockPointer:undefined,resizeCanvas:undefined,requestFullScreen:(function(lockPointer,resizeCanvas,vrDevice){Browser.lockPointer=lockPointer;Browser.resizeCanvas=resizeCanvas;Browser.vrDevice=vrDevice;if(typeof Browser.lockPointer==="undefined")Browser.lockPointer=true;if(typeof Browser.resizeCanvas==="undefined")Browser.resizeCanvas=false;if(typeof Browser.vrDevice==="undefined")Browser.vrDevice=null;var canvas=Module["canvas"];function fullScreenChange(){Browser.isFullScreen=false;var canvasContainer=canvas.parentNode;if((document["webkitFullScreenElement"]||document["webkitFullscreenElement"]||document["mozFullScreenElement"]||document["mozFullscreenElement"]||document["fullScreenElement"]||document["fullscreenElement"]||document["msFullScreenElement"]||document["msFullscreenElement"]||document["webkitCurrentFullScreenElement"])===canvasContainer){canvas.cancelFullScreen=document["cancelFullScreen"]||document["mozCancelFullScreen"]||document["webkitCancelFullScreen"]||document["msExitFullscreen"]||document["exitFullscreen"]||(function(){});canvas.cancelFullScreen=canvas.cancelFullScreen.bind(document);if(Browser.lockPointer)canvas.requestPointerLock();Browser.isFullScreen=true;if(Browser.resizeCanvas)Browser.setFullScreenCanvasSize()}else{canvasContainer.parentNode.insertBefore(canvas,canvasContainer);canvasContainer.parentNode.removeChild(canvasContainer);if(Browser.resizeCanvas)Browser.setWindowedCanvasSize()}if(Module["onFullScreen"])Module["onFullScreen"](Browser.isFullScreen);Browser.updateCanvasDimensions(canvas)}if(!Browser.fullScreenHandlersInstalled){Browser.fullScreenHandlersInstalled=true;document.addEventListener("fullscreenchange",fullScreenChange,false);document.addEventListener("mozfullscreenchange",fullScreenChange,false);document.addEventListener("webkitfullscreenchange",fullScreenChange,false);document.addEventListener("MSFullscreenChange",fullScreenChange,false)}var canvasContainer=document.createElement("div");canvas.parentNode.insertBefore(canvasContainer,canvas);canvasContainer.appendChild(canvas);canvasContainer.requestFullScreen=canvasContainer["requestFullScreen"]||canvasContainer["mozRequestFullScreen"]||canvasContainer["msRequestFullscreen"]||(canvasContainer["webkitRequestFullScreen"]?(function(){canvasContainer["webkitRequestFullScreen"](Element["ALLOW_KEYBOARD_INPUT"])}):null);if(vrDevice){canvasContainer.requestFullScreen({vrDisplay:vrDevice})}else{canvasContainer.requestFullScreen()}}),nextRAF:0,fakeRequestAnimationFrame:(function(func){var now=Date.now();if(Browser.nextRAF===0){Browser.nextRAF=now+1e3/60}else{while(now+2>=Browser.nextRAF){Browser.nextRAF+=1e3/60}}var delay=Math.max(Browser.nextRAF-now,0);setTimeout(func,delay)}),requestAnimationFrame:function requestAnimationFrame(func){if(typeof window==="undefined"){Browser.fakeRequestAnimationFrame(func)}else{if(!window.requestAnimationFrame){window.requestAnimationFrame=window["requestAnimationFrame"]||window["mozRequestAnimationFrame"]||window["webkitRequestAnimationFrame"]||window["msRequestAnimationFrame"]||window["oRequestAnimationFrame"]||Browser.fakeRequestAnimationFrame}window.requestAnimationFrame(func)}},safeCallback:(function(func){return(function(){if(!ABORT)return func.apply(null,arguments)})}),allowAsyncCallbacks:true,queuedAsyncCallbacks:[],pauseAsyncCallbacks:(function(){Browser.allowAsyncCallbacks=false}),resumeAsyncCallbacks:(function(){Browser.allowAsyncCallbacks=true;if(Browser.queuedAsyncCallbacks.length>0){var callbacks=Browser.queuedAsyncCallbacks;Browser.queuedAsyncCallbacks=[];callbacks.forEach((function(func){func()}))}}),safeRequestAnimationFrame:(function(func){return Browser.requestAnimationFrame((function(){if(ABORT)return;if(Browser.allowAsyncCallbacks){func()}else{Browser.queuedAsyncCallbacks.push(func)}}))}),safeSetTimeout:(function(func,timeout){Module["noExitRuntime"]=true;return setTimeout((function(){if(ABORT)return;if(Browser.allowAsyncCallbacks){func()}else{Browser.queuedAsyncCallbacks.push(func)}}),timeout)}),safeSetInterval:(function(func,timeout){Module["noExitRuntime"]=true;return setInterval((function(){if(ABORT)return;if(Browser.allowAsyncCallbacks){func()}}),timeout)}),getMimetype:(function(name){return{"jpg":"image/jpeg","jpeg":"image/jpeg","png":"image/png","bmp":"image/bmp","ogg":"audio/ogg","wav":"audio/wav","mp3":"audio/mpeg"}[name.substr(name.lastIndexOf(".")+1)]}),getUserMedia:(function(func){if(!window.getUserMedia){window.getUserMedia=navigator["getUserMedia"]||navigator["mozGetUserMedia"]}window.getUserMedia(func)}),getMovementX:(function(event){return event["movementX"]||event["mozMovementX"]||event["webkitMovementX"]||0}),getMovementY:(function(event){return event["movementY"]||event["mozMovementY"]||event["webkitMovementY"]||0}),getMouseWheelDelta:(function(event){var delta=0;switch(event.type){case"DOMMouseScroll":delta=event.detail;break;case"mousewheel":delta=event.wheelDelta;break;case"wheel":delta=event["deltaY"];break;default:throw"unrecognized mouse wheel event: "+event.type}return delta}),mouseX:0,mouseY:0,mouseMovementX:0,mouseMovementY:0,touches:{},lastTouches:{},calculateMouseEvent:(function(event){if(Browser.pointerLock){if(event.type!="mousemove"&&"mozMovementX"in event){Browser.mouseMovementX=Browser.mouseMovementY=0}else{Browser.mouseMovementX=Browser.getMovementX(event);Browser.mouseMovementY=Browser.getMovementY(event)}if(typeof SDL!="undefined"){Browser.mouseX=SDL.mouseX+Browser.mouseMovementX;Browser.mouseY=SDL.mouseY+Browser.mouseMovementY}else{Browser.mouseX+=Browser.mouseMovementX;Browser.mouseY+=Browser.mouseMovementY}}else{var rect=Module["canvas"].getBoundingClientRect();var cw=Module["canvas"].width;var ch=Module["canvas"].height;var scrollX=typeof window.scrollX!=="undefined"?window.scrollX:window.pageXOffset;var scrollY=typeof window.scrollY!=="undefined"?window.scrollY:window.pageYOffset;if(event.type==="touchstart"||event.type==="touchend"||event.type==="touchmove"){var touch=event.touch;if(touch===undefined){return}var adjustedX=touch.pageX-(scrollX+rect.left);var adjustedY=touch.pageY-(scrollY+rect.top);adjustedX=adjustedX*(cw/rect.width);adjustedY=adjustedY*(ch/rect.height);var coords={x:adjustedX,y:adjustedY};if(event.type==="touchstart"){Browser.lastTouches[touch.identifier]=coords;Browser.touches[touch.identifier]=coords}else if(event.type==="touchend"||event.type==="touchmove"){var last=Browser.touches[touch.identifier];if(!last)last=coords;Browser.lastTouches[touch.identifier]=last;Browser.touches[touch.identifier]=coords}return}var x=event.pageX-(scrollX+rect.left);var y=event.pageY-(scrollY+rect.top);x=x*(cw/rect.width);y=y*(ch/rect.height);Browser.mouseMovementX=x-Browser.mouseX;Browser.mouseMovementY=y-Browser.mouseY;Browser.mouseX=x;Browser.mouseY=y}}),xhrLoad:(function(url,onload,onerror){var xhr=new XMLHttpRequest;xhr.open("GET",url,true);xhr.responseType="arraybuffer";xhr.onload=function xhr_onload(){if(xhr.status==200||xhr.status==0&&xhr.response){onload(xhr.response)}else{onerror()}};xhr.onerror=onerror;xhr.send(null)}),asyncLoad:(function(url,onload,onerror,noRunDep){Browser.xhrLoad(url,(function(arrayBuffer){assert(arrayBuffer,'Loading data file "'+url+'" failed (no arrayBuffer).');onload(new Uint8Array(arrayBuffer));if(!noRunDep)removeRunDependency("al "+url)}),(function(event){if(onerror){onerror()}else{throw'Loading data file "'+url+'" failed.'}}));if(!noRunDep)addRunDependency("al "+url)}),resizeListeners:[],updateResizeListeners:(function(){var canvas=Module["canvas"];Browser.resizeListeners.forEach((function(listener){listener(canvas.width,canvas.height)}))}),setCanvasSize:(function(width,height,noUpdates){var canvas=Module["canvas"];Browser.updateCanvasDimensions(canvas,width,height);if(!noUpdates)Browser.updateResizeListeners()}),windowedWidth:0,windowedHeight:0,setFullScreenCanvasSize:(function(){if(typeof SDL!="undefined"){var flags=HEAPU32[SDL.screen+Runtime.QUANTUM_SIZE*0>>2];flags=flags|8388608;HEAP32[SDL.screen+Runtime.QUANTUM_SIZE*0>>2]=flags}Browser.updateResizeListeners()}),setWindowedCanvasSize:(function(){if(typeof SDL!="undefined"){var flags=HEAPU32[SDL.screen+Runtime.QUANTUM_SIZE*0>>2];flags=flags&~8388608;HEAP32[SDL.screen+Runtime.QUANTUM_SIZE*0>>2]=flags}Browser.updateResizeListeners()}),updateCanvasDimensions:(function(canvas,wNative,hNative){if(wNative&&hNative){canvas.widthNative=wNative;canvas.heightNative=hNative}else{wNative=canvas.widthNative;hNative=canvas.heightNative}var w=wNative;var h=hNative;if(Module["forcedAspectRatio"]&&Module["forcedAspectRatio"]>0){if(w/h<Module["forcedAspectRatio"]){w=Math.round(h*Module["forcedAspectRatio"])}else{h=Math.round(w/Module["forcedAspectRatio"])}}if((document["webkitFullScreenElement"]||document["webkitFullscreenElement"]||document["mozFullScreenElement"]||document["mozFullscreenElement"]||document["fullScreenElement"]||document["fullscreenElement"]||document["msFullScreenElement"]||document["msFullscreenElement"]||document["webkitCurrentFullScreenElement"])===canvas.parentNode&&typeof screen!="undefined"){var factor=Math.min(screen.width/w,screen.height/h);w=Math.round(w*factor);h=Math.round(h*factor)}if(Browser.resizeCanvas){if(canvas.width!=w)canvas.width=w;if(canvas.height!=h)canvas.height=h;if(typeof canvas.style!="undefined"){canvas.style.removeProperty("width");canvas.style.removeProperty("height")}}else{if(canvas.width!=wNative)canvas.width=wNative;if(canvas.height!=hNative)canvas.height=hNative;if(typeof canvas.style!="undefined"){if(w!=wNative||h!=hNative){canvas.style.setProperty("width",w+"px","important");canvas.style.setProperty("height",h+"px","important")}else{canvas.style.removeProperty("width");canvas.style.removeProperty("height")}}}}),wgetRequests:{},nextWgetRequestHandle:0,getNextWgetRequestHandle:(function(){var handle=Browser.nextWgetRequestHandle;Browser.nextWgetRequestHandle++;return handle})};var _emscripten_get_longjmp_result=true;var ___tm_formatted=allocate(44,"i8",ALLOC_STATIC);var _tzname=allocate(8,"i32*",ALLOC_STATIC);var _daylight=allocate(1,"i32*",ALLOC_STATIC);var _timezone=allocate(1,"i32*",ALLOC_STATIC);function _tzset(){if(_tzset.called)return;_tzset.called=true;HEAP32[_timezone>>2]=-(new Date).getTimezoneOffset()*60;var winter=new Date(2e3,0,1);var summer=new Date(2e3,6,1);HEAP32[_daylight>>2]=Number(winter.getTimezoneOffset()!=summer.getTimezoneOffset());function extractZone(date){var match=date.toTimeString().match(/\(([A-Za-z ]+)\)$/);return match?match[1]:"GMT"}var winterName=extractZone(winter);var summerName=extractZone(summer);var winterNamePtr=allocate(intArrayFromString(winterName),"i8",ALLOC_NORMAL);var summerNamePtr=allocate(intArrayFromString(summerName),"i8",ALLOC_NORMAL);if(summer.getTimezoneOffset()<winter.getTimezoneOffset()){HEAP32[_tzname>>2]=winterNamePtr;HEAP32[_tzname+4>>2]=summerNamePtr}else{HEAP32[_tzname>>2]=summerNamePtr;HEAP32[_tzname+4>>2]=winterNamePtr}}function _mktime(tmPtr){_tzset();var date=new Date(HEAP32[tmPtr+20>>2]+1900,HEAP32[tmPtr+16>>2],HEAP32[tmPtr+12>>2],HEAP32[tmPtr+8>>2],HEAP32[tmPtr+4>>2],HEAP32[tmPtr>>2],0);var dst=HEAP32[tmPtr+32>>2];var guessedOffset=date.getTimezoneOffset();var start=new Date(date.getFullYear(),0,1);var summerOffset=(new Date(2e3,6,1)).getTimezoneOffset();var winterOffset=start.getTimezoneOffset();var dstOffset=Math.min(winterOffset,summerOffset);if(dst<0){HEAP32[tmPtr+32>>2]=Number(winterOffset!=guessedOffset)}else if(dst>0!=(winterOffset!=guessedOffset)){var summerOffset=(new Date(date.getFullYear(),6,1)).getTimezoneOffset();var trueOffset=dst>0?summerOffset:winterOffset;date.setTime(date.getTime()+(trueOffset-guessedOffset)*6e4)}HEAP32[tmPtr+24>>2]=date.getDay();var yday=(date.getTime()-start.getTime())/(1e3*60*60*24)|0;HEAP32[tmPtr+28>>2]=yday;return date.getTime()/1e3|0}function _asctime_r(tmPtr,buf){var date={tm_sec:HEAP32[tmPtr>>2],tm_min:HEAP32[tmPtr+4>>2],tm_hour:HEAP32[tmPtr+8>>2],tm_mday:HEAP32[tmPtr+12>>2],tm_mon:HEAP32[tmPtr+16>>2],tm_year:HEAP32[tmPtr+20>>2],tm_wday:HEAP32[tmPtr+24>>2]};var days=["Sun","Mon","Tue","Wed","Thu","Fri","Sat"];var months=["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"];var s=days[date.tm_wday]+" "+months[date.tm_mon]+(date.tm_mday<10?"  ":" ")+date.tm_mday+(date.tm_hour<10?" 0":" ")+date.tm_hour+(date.tm_min<10?":0":":")+date.tm_min+(date.tm_sec<10?":0":":")+date.tm_sec+" "+(1900+date.tm_year)+"\n";writeStringToMemory(s,buf);return buf}function _asctime(tmPtr){return _asctime_r(tmPtr,___tm_formatted)}var _sin=Math_sin;var _exp=Math_exp;var _atan=Math_atan;function _time(ptr){var ret=Date.now()/1e3|0;if(ptr){HEAP32[ptr>>2]=ret}return ret}function ___syscall10(which,varargs){SYSCALLS.varargs=varargs;try{var path=SYSCALLS.getStr();FS.unlink(path);return 0}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}var ___tm_current=allocate(44,"i8",ALLOC_STATIC);var ___tm_timezone=allocate(intArrayFromString("GMT"),"i8",ALLOC_STATIC);function _gmtime_r(time,tmPtr){var date=new Date(HEAP32[time>>2]*1e3);HEAP32[tmPtr>>2]=date.getUTCSeconds();HEAP32[tmPtr+4>>2]=date.getUTCMinutes();HEAP32[tmPtr+8>>2]=date.getUTCHours();HEAP32[tmPtr+12>>2]=date.getUTCDate();HEAP32[tmPtr+16>>2]=date.getUTCMonth();HEAP32[tmPtr+20>>2]=date.getUTCFullYear()-1900;HEAP32[tmPtr+24>>2]=date.getUTCDay();HEAP32[tmPtr+36>>2]=0;HEAP32[tmPtr+32>>2]=0;var start=Date.UTC(date.getUTCFullYear(),0,1,0,0,0,0);var yday=(date.getTime()-start)/(1e3*60*60*24)|0;HEAP32[tmPtr+28>>2]=yday;HEAP32[tmPtr+40>>2]=___tm_timezone;return tmPtr}function _gmtime(time){return _gmtime_r(time,___tm_current)}var _htons=undefined;Module["_htons"]=_htons;function __inet_pton6_raw(str){var words;var w,offset,z,i;var valid6regx=/^((?=.*::)(?!.*::.+::)(::)?([\dA-F]{1,4}:(:|\b)|){5}|([\dA-F]{1,4}:){6})((([\dA-F]{1,4}((?!\3)::|:\b|$))|(?!\2\3)){2}|(((2[0-4]|1\d|[1-9])?\d|25[0-5])\.?\b){4})$/i;var parts=[];if(!valid6regx.test(str)){return null}if(str==="::"){return[0,0,0,0,0,0,0,0]}if(str.indexOf("::")===0){str=str.replace("::","Z:")}else{str=str.replace("::",":Z:")}if(str.indexOf(".")>0){str=str.replace(new RegExp("[.]","g"),":");words=str.split(":");words[words.length-4]=parseInt(words[words.length-4])+parseInt(words[words.length-3])*256;words[words.length-3]=parseInt(words[words.length-2])+parseInt(words[words.length-1])*256;words=words.slice(0,words.length-2)}else{words=str.split(":")}offset=0;z=0;for(w=0;w<words.length;w++){if(typeof words[w]==="string"){if(words[w]==="Z"){for(z=0;z<8-words.length+1;z++){parts[w+z]=0}offset=z-1}else{parts[w+offset]=_htons(parseInt(words[w],16))}}else{parts[w+offset]=words[w]}}return[parts[1]<<16|parts[0],parts[3]<<16|parts[2],parts[5]<<16|parts[4],parts[7]<<16|parts[6]]}var DNS={address_map:{id:1,addrs:{},names:{}},lookup_name:(function(name){var res=__inet_pton4_raw(name);if(res){return name}res=__inet_pton6_raw(name);if(res){return name}var addr;if(DNS.address_map.addrs[name]){addr=DNS.address_map.addrs[name]}else{var id=DNS.address_map.id++;assert(id<65535,"exceeded max address mappings of 65535");addr="172.29."+(id&255)+"."+(id&65280);DNS.address_map.names[addr]=name;DNS.address_map.addrs[name]=addr}return addr}),lookup_addr:(function(addr){if(DNS.address_map.names[addr]){return DNS.address_map.names[addr]}return null})};function _gethostbyname(name){name=Pointer_stringify(name);var ret=_malloc(20);var nameBuf=_malloc(name.length+1);writeStringToMemory(name,nameBuf);HEAP32[ret>>2]=nameBuf;var aliasesBuf=_malloc(4);HEAP32[aliasesBuf>>2]=0;HEAP32[ret+4>>2]=aliasesBuf;var afinet=2;HEAP32[ret+8>>2]=afinet;HEAP32[ret+12>>2]=4;var addrListBuf=_malloc(12);HEAP32[addrListBuf>>2]=addrListBuf+8;HEAP32[addrListBuf+4>>2]=0;HEAP32[addrListBuf+8>>2]=__inet_pton4_raw(DNS.lookup_name(name));HEAP32[ret+16>>2]=addrListBuf;return ret}function ___syscall183(which,varargs){SYSCALLS.varargs=varargs;try{var buf=SYSCALLS.get(),size=SYSCALLS.get();if(size===0)return-ERRNO_CODES.EINVAL;var cwd=FS.cwd();if(size<cwd.length+1)return-ERRNO_CODES.ERANGE;writeAsciiToMemory(cwd,buf);return buf}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall3(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),buf=SYSCALLS.get(),count=SYSCALLS.get();return FS.read(stream,HEAP8,buf,count)}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}var _ceil=Math_ceil;function ___syscall5(which,varargs){SYSCALLS.varargs=varargs;try{var pathname=SYSCALLS.getStr(),flags=SYSCALLS.get(),mode=SYSCALLS.get();var stream=FS.open(pathname,flags,mode);return stream.fd}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall4(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),buf=SYSCALLS.get(),count=SYSCALLS.get();return FS.write(stream,HEAP8,buf,count)}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}var _log=Math_log;var _emscripten_setjmp=true;var _cos=Math_cos;function _alarm(seconds){setTimeout((function(){if(__sigalrm_handler)Runtime.dynCall("vi",__sigalrm_handler,[0])}),seconds*1e3)}function _clock(){if(_clock.start===undefined)_clock.start=Date.now();return(Date.now()-_clock.start)*(1e6/1e3)|0}var SOCKFS={mount:(function(mount){Module["websocket"]=Module["websocket"]&&"object"===typeof Module["websocket"]?Module["websocket"]:{};Module["websocket"]._callbacks={};Module["websocket"]["on"]=(function(event,callback){if("function"===typeof callback){this._callbacks[event]=callback}return this});Module["websocket"].emit=(function(event,param){if("function"===typeof this._callbacks[event]){this._callbacks[event].call(this,param)}});return FS.createNode(null,"/",16384|511,0)}),createSocket:(function(family,type,protocol){var streaming=type==1;if(protocol){assert(streaming==(protocol==6))}var sock={family:family,type:type,protocol:protocol,server:null,error:null,peers:{},pending:[],recv_queue:[],sock_ops:SOCKFS.websocket_sock_ops};var name=SOCKFS.nextname();var node=FS.createNode(SOCKFS.root,name,49152,0);node.sock=sock;var stream=FS.createStream({path:name,node:node,flags:FS.modeStringToFlags("r+"),seekable:false,stream_ops:SOCKFS.stream_ops});sock.stream=stream;return sock}),getSocket:(function(fd){var stream=FS.getStream(fd);if(!stream||!FS.isSocket(stream.node.mode)){return null}return stream.node.sock}),stream_ops:{poll:(function(stream){var sock=stream.node.sock;return sock.sock_ops.poll(sock)}),ioctl:(function(stream,request,varargs){var sock=stream.node.sock;return sock.sock_ops.ioctl(sock,request,varargs)}),read:(function(stream,buffer,offset,length,position){var sock=stream.node.sock;var msg=sock.sock_ops.recvmsg(sock,length);if(!msg){return 0}buffer.set(msg.buffer,offset);return msg.buffer.length}),write:(function(stream,buffer,offset,length,position){var sock=stream.node.sock;return sock.sock_ops.sendmsg(sock,buffer,offset,length)}),close:(function(stream){var sock=stream.node.sock;sock.sock_ops.close(sock)})},nextname:(function(){if(!SOCKFS.nextname.current){SOCKFS.nextname.current=0}return"socket["+SOCKFS.nextname.current++ +"]"}),websocket_sock_ops:{createPeer:(function(sock,addr,port){var ws;if(typeof addr==="object"){ws=addr;addr=null;port=null}if(ws){if(ws._socket){addr=ws._socket.remoteAddress;port=ws._socket.remotePort}else{var result=/ws[s]?:\/\/([^:]+):(\d+)/.exec(ws.url);if(!result){throw new Error("WebSocket URL must be in the format ws(s)://address:port")}addr=result[1];port=parseInt(result[2],10)}}else{try{var runtimeConfig=Module["websocket"]&&"object"===typeof Module["websocket"];var url="ws:#".replace("#","//");if(runtimeConfig){if("string"===typeof Module["websocket"]["url"]){url=Module["websocket"]["url"]}}if(url==="ws://"||url==="wss://"){var parts=addr.split("/");url=url+parts[0]+":"+port+"/"+parts.slice(1).join("/")}var subProtocols="binary";if(runtimeConfig){if("string"===typeof Module["websocket"]["subprotocol"]){subProtocols=Module["websocket"]["subprotocol"]}}subProtocols=subProtocols.replace(/^ +| +$/g,"").split(/ *, */);var opts=ENVIRONMENT_IS_NODE?{"protocol":subProtocols.toString()}:subProtocols;var WebSocket=ENVIRONMENT_IS_NODE?require("ws"):window["WebSocket"];ws=new WebSocket(url,opts);ws.binaryType="arraybuffer"}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EHOSTUNREACH)}}var peer={addr:addr,port:port,socket:ws,dgram_send_queue:[]};SOCKFS.websocket_sock_ops.addPeer(sock,peer);SOCKFS.websocket_sock_ops.handlePeerEvents(sock,peer);if(sock.type===2&&typeof sock.sport!=="undefined"){peer.dgram_send_queue.push(new Uint8Array([255,255,255,255,"p".charCodeAt(0),"o".charCodeAt(0),"r".charCodeAt(0),"t".charCodeAt(0),(sock.sport&65280)>>8,sock.sport&255]))}return peer}),getPeer:(function(sock,addr,port){return sock.peers[addr+":"+port]}),addPeer:(function(sock,peer){sock.peers[peer.addr+":"+peer.port]=peer}),removePeer:(function(sock,peer){delete sock.peers[peer.addr+":"+peer.port]}),handlePeerEvents:(function(sock,peer){var first=true;var handleOpen=(function(){Module["websocket"].emit("open",sock.stream.fd);try{var queued=peer.dgram_send_queue.shift();while(queued){peer.socket.send(queued);queued=peer.dgram_send_queue.shift()}}catch(e){peer.socket.close()}});function handleMessage(data){assert(typeof data!=="string"&&data.byteLength!==undefined);data=new Uint8Array(data);var wasfirst=first;first=false;if(wasfirst&&data.length===10&&data[0]===255&&data[1]===255&&data[2]===255&&data[3]===255&&data[4]==="p".charCodeAt(0)&&data[5]==="o".charCodeAt(0)&&data[6]==="r".charCodeAt(0)&&data[7]==="t".charCodeAt(0)){var newport=data[8]<<8|data[9];SOCKFS.websocket_sock_ops.removePeer(sock,peer);peer.port=newport;SOCKFS.websocket_sock_ops.addPeer(sock,peer);return}sock.recv_queue.push({addr:peer.addr,port:peer.port,data:data});Module["websocket"].emit("message",sock.stream.fd)}if(ENVIRONMENT_IS_NODE){peer.socket.on("open",handleOpen);peer.socket.on("message",(function(data,flags){if(!flags.binary){return}handleMessage((new Uint8Array(data)).buffer)}));peer.socket.on("close",(function(){Module["websocket"].emit("close",sock.stream.fd)}));peer.socket.on("error",(function(error){sock.error=ERRNO_CODES.ECONNREFUSED;Module["websocket"].emit("error",[sock.stream.fd,sock.error,"ECONNREFUSED: Connection refused"])}))}else{peer.socket.onopen=handleOpen;peer.socket.onclose=(function(){Module["websocket"].emit("close",sock.stream.fd)});peer.socket.onmessage=function peer_socket_onmessage(event){handleMessage(event.data)};peer.socket.onerror=(function(error){sock.error=ERRNO_CODES.ECONNREFUSED;Module["websocket"].emit("error",[sock.stream.fd,sock.error,"ECONNREFUSED: Connection refused"])})}}),poll:(function(sock){if(sock.type===1&&sock.server){return sock.pending.length?64|1:0}var mask=0;var dest=sock.type===1?SOCKFS.websocket_sock_ops.getPeer(sock,sock.daddr,sock.dport):null;if(sock.recv_queue.length||!dest||dest&&dest.socket.readyState===dest.socket.CLOSING||dest&&dest.socket.readyState===dest.socket.CLOSED){mask|=64|1}if(!dest||dest&&dest.socket.readyState===dest.socket.OPEN){mask|=4}if(dest&&dest.socket.readyState===dest.socket.CLOSING||dest&&dest.socket.readyState===dest.socket.CLOSED){mask|=16}return mask}),ioctl:(function(sock,request,arg){switch(request){case 21531:var bytes=0;if(sock.recv_queue.length){bytes=sock.recv_queue[0].data.length}HEAP32[arg>>2]=bytes;return 0;default:return ERRNO_CODES.EINVAL}}),close:(function(sock){if(sock.server){try{sock.server.close()}catch(e){}sock.server=null}var peers=Object.keys(sock.peers);for(var i=0;i<peers.length;i++){var peer=sock.peers[peers[i]];try{peer.socket.close()}catch(e){}SOCKFS.websocket_sock_ops.removePeer(sock,peer)}return 0}),bind:(function(sock,addr,port){if(typeof sock.saddr!=="undefined"||typeof sock.sport!=="undefined"){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}sock.saddr=addr;sock.sport=port;if(sock.type===2){if(sock.server){sock.server.close();sock.server=null}try{sock.sock_ops.listen(sock,0)}catch(e){if(!(e instanceof FS.ErrnoError))throw e;if(e.errno!==ERRNO_CODES.EOPNOTSUPP)throw e}}}),connect:(function(sock,addr,port){if(sock.server){throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP)}if(typeof sock.daddr!=="undefined"&&typeof sock.dport!=="undefined"){var dest=SOCKFS.websocket_sock_ops.getPeer(sock,sock.daddr,sock.dport);if(dest){if(dest.socket.readyState===dest.socket.CONNECTING){throw new FS.ErrnoError(ERRNO_CODES.EALREADY)}else{throw new FS.ErrnoError(ERRNO_CODES.EISCONN)}}}var peer=SOCKFS.websocket_sock_ops.createPeer(sock,addr,port);sock.daddr=peer.addr;sock.dport=peer.port;throw new FS.ErrnoError(ERRNO_CODES.EINPROGRESS)}),listen:(function(sock,backlog){if(!ENVIRONMENT_IS_NODE){throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP)}if(sock.server){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var WebSocketServer=require("ws").Server;var host=sock.saddr;sock.server=new WebSocketServer({host:host,port:sock.sport});Module["websocket"].emit("listen",sock.stream.fd);sock.server.on("connection",(function(ws){if(sock.type===1){var newsock=SOCKFS.createSocket(sock.family,sock.type,sock.protocol);var peer=SOCKFS.websocket_sock_ops.createPeer(newsock,ws);newsock.daddr=peer.addr;newsock.dport=peer.port;sock.pending.push(newsock);Module["websocket"].emit("connection",newsock.stream.fd)}else{SOCKFS.websocket_sock_ops.createPeer(sock,ws);Module["websocket"].emit("connection",sock.stream.fd)}}));sock.server.on("closed",(function(){Module["websocket"].emit("close",sock.stream.fd);sock.server=null}));sock.server.on("error",(function(error){sock.error=ERRNO_CODES.EHOSTUNREACH;Module["websocket"].emit("error",[sock.stream.fd,sock.error,"EHOSTUNREACH: Host is unreachable"])}))}),accept:(function(listensock){if(!listensock.server){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}var newsock=listensock.pending.shift();newsock.stream.flags=listensock.stream.flags;return newsock}),getname:(function(sock,peer){var addr,port;if(peer){if(sock.daddr===undefined||sock.dport===undefined){throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN)}addr=sock.daddr;port=sock.dport}else{addr=sock.saddr||0;port=sock.sport||0}return{addr:addr,port:port}}),sendmsg:(function(sock,buffer,offset,length,addr,port){if(sock.type===2){if(addr===undefined||port===undefined){addr=sock.daddr;port=sock.dport}if(addr===undefined||port===undefined){throw new FS.ErrnoError(ERRNO_CODES.EDESTADDRREQ)}}else{addr=sock.daddr;port=sock.dport}var dest=SOCKFS.websocket_sock_ops.getPeer(sock,addr,port);if(sock.type===1){if(!dest||dest.socket.readyState===dest.socket.CLOSING||dest.socket.readyState===dest.socket.CLOSED){throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN)}else if(dest.socket.readyState===dest.socket.CONNECTING){throw new FS.ErrnoError(ERRNO_CODES.EAGAIN)}}var data;if(buffer instanceof Array||buffer instanceof ArrayBuffer){data=buffer.slice(offset,offset+length)}else{data=buffer.buffer.slice(buffer.byteOffset+offset,buffer.byteOffset+offset+length)}if(sock.type===2){if(!dest||dest.socket.readyState!==dest.socket.OPEN){if(!dest||dest.socket.readyState===dest.socket.CLOSING||dest.socket.readyState===dest.socket.CLOSED){dest=SOCKFS.websocket_sock_ops.createPeer(sock,addr,port)}dest.dgram_send_queue.push(data);return length}}try{dest.socket.send(data);return length}catch(e){throw new FS.ErrnoError(ERRNO_CODES.EINVAL)}}),recvmsg:(function(sock,length){if(sock.type===1&&sock.server){throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN)}var queued=sock.recv_queue.shift();if(!queued){if(sock.type===1){var dest=SOCKFS.websocket_sock_ops.getPeer(sock,sock.daddr,sock.dport);if(!dest){throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN)}else if(dest.socket.readyState===dest.socket.CLOSING||dest.socket.readyState===dest.socket.CLOSED){return null}else{throw new FS.ErrnoError(ERRNO_CODES.EAGAIN)}}else{throw new FS.ErrnoError(ERRNO_CODES.EAGAIN)}}var queuedLength=queued.data.byteLength||queued.data.length;var queuedOffset=queued.data.byteOffset||0;var queuedBuffer=queued.data.buffer||queued.data;var bytesRead=Math.min(length,queuedLength);var res={buffer:new Uint8Array(queuedBuffer,queuedOffset,bytesRead),addr:queued.addr,port:queued.port};if(sock.type===1&&bytesRead<queuedLength){var bytesRemaining=queuedLength-bytesRead;queued.data=new Uint8Array(queuedBuffer,queuedOffset+bytesRead,bytesRemaining);sock.recv_queue.unshift(queued)}return res})}};var Sockets={BUFFER_SIZE:10240,MAX_BUFFER_SIZE:10485760,nextFd:1,fds:{},nextport:1,maxport:65535,peer:null,connections:{},portmap:{},localAddr:4261412874,addrPool:[33554442,50331658,67108874,83886090,100663306,117440522,134217738,150994954,167772170,184549386,201326602,218103818,234881034]};function __inet_ntop4_raw(addr){return(addr&255)+"."+(addr>>8&255)+"."+(addr>>16&255)+"."+(addr>>24&255)}var _ntohs=undefined;Module["_ntohs"]=_ntohs;function __inet_ntop6_raw(ints){var str="";var word=0;var longest=0;var lastzero=0;var zstart=0;var len=0;var i=0;var parts=[ints[0]&65535,ints[0]>>16,ints[1]&65535,ints[1]>>16,ints[2]&65535,ints[2]>>16,ints[3]&65535,ints[3]>>16];var hasipv4=true;var v4part="";for(i=0;i<5;i++){if(parts[i]!==0){hasipv4=false;break}}if(hasipv4){v4part=__inet_ntop4_raw(parts[6]|parts[7]<<16);if(parts[5]===-1){str="::ffff:";str+=v4part;return str}if(parts[5]===0){str="::";if(v4part==="0.0.0.0")v4part="";if(v4part==="0.0.0.1")v4part="1";str+=v4part;return str}}for(word=0;word<8;word++){if(parts[word]===0){if(word-lastzero>1){len=0}lastzero=word;len++}if(len>longest){longest=len;zstart=word-longest+1}}for(word=0;word<8;word++){if(longest>1){if(parts[word]===0&&word>=zstart&&word<zstart+longest){if(word===zstart){str+=":";if(zstart===0)str+=":"}continue}}str+=Number(_ntohs(parts[word]&65535)).toString(16);str+=word<7?":":""}return str}function __read_sockaddr(sa,salen){var family=HEAP16[sa>>1];var port=_ntohs(HEAP16[sa+2>>1]);var addr;switch(family){case 2:if(salen!==16){return{errno:ERRNO_CODES.EINVAL}}addr=HEAP32[sa+4>>2];addr=__inet_ntop4_raw(addr);break;case 10:if(salen!==28){return{errno:ERRNO_CODES.EINVAL}}addr=[HEAP32[sa+8>>2],HEAP32[sa+12>>2],HEAP32[sa+16>>2],HEAP32[sa+20>>2]];addr=__inet_ntop6_raw(addr);break;default:return{errno:ERRNO_CODES.EAFNOSUPPORT}}return{family:family,addr:addr,port:port}}function __write_sockaddr(sa,family,addr,port){switch(family){case 2:addr=__inet_pton4_raw(addr);HEAP16[sa>>1]=family;HEAP32[sa+4>>2]=addr;HEAP16[sa+2>>1]=_htons(port);break;case 10:addr=__inet_pton6_raw(addr);HEAP32[sa>>2]=family;HEAP32[sa+8>>2]=addr[0];HEAP32[sa+12>>2]=addr[1];HEAP32[sa+16>>2]=addr[2];HEAP32[sa+20>>2]=addr[3];HEAP16[sa+2>>1]=_htons(port);HEAP32[sa+4>>2]=0;HEAP32[sa+24>>2]=0;break;default:return{errno:ERRNO_CODES.EAFNOSUPPORT}}return{}}function ___syscall102(which,varargs){SYSCALLS.varargs=varargs;try{var call=SYSCALLS.get(),socketvararg=SYSCALLS.get();SYSCALLS.varargs=socketvararg;switch(call){case 1:{var domain=SYSCALLS.get(),type=SYSCALLS.get(),protocol=SYSCALLS.get();var sock=SOCKFS.createSocket(domain,type,protocol);assert(sock.stream.fd<64);return sock.stream.fd};case 2:{var sock=SYSCALLS.getSocketFromFD(),info=SYSCALLS.getSocketAddress();sock.sock_ops.bind(sock,info.addr,info.port);return 0};case 3:{var sock=SYSCALLS.getSocketFromFD(),info=SYSCALLS.getSocketAddress();sock.sock_ops.connect(sock,info.addr,info.port);return 0};case 4:{var sock=SYSCALLS.getSocketFromFD(),backlog=SYSCALLS.get();sock.sock_ops.listen(sock,backlog);return 0};case 5:{var sock=SYSCALLS.getSocketFromFD(),addr=SYSCALLS.get(),addrlen=SYSCALLS.get();var newsock=sock.sock_ops.accept(sock);if(addr){var res=__write_sockaddr(addr,newsock.family,DNS.lookup_name(newsock.daddr),newsock.dport);assert(!res.errno)}return newsock.stream.fd};case 6:{var sock=SYSCALLS.getSocketFromFD(),addr=SYSCALLS.get(),addrlen=SYSCALLS.get();var res=__write_sockaddr(addr,sock.family,DNS.lookup_name(sock.daddr||"0.0.0.0"),sock.dport);assert(!res.errno);return 0};case 11:{var sock=SYSCALLS.getSocketFromFD(),message=SYSCALLS.get(),length=SYSCALLS.get(),flags=SYSCALLS.get(),dest=SYSCALLS.getSocketAddress(true);if(!dest){return FS.write(sock.stream,HEAP8,message,length)}else{return sock.sock_ops.sendmsg(sock,HEAP8,message,length,dest.addr,dest.port)}};case 12:{var sock=SYSCALLS.getSocketFromFD(),buf=SYSCALLS.get(),len=SYSCALLS.get(),flags=SYSCALLS.get(),addr=SYSCALLS.get(),addrlen=SYSCALLS.get();var msg=sock.sock_ops.recvmsg(sock,len);if(!msg)return 0;if(addr){var res=__write_sockaddr(addr,sock.family,DNS.lookup_name(msg.addr),msg.port);assert(!res.errno)}HEAPU8.set(msg.buffer,buf);return msg.buffer.byteLength};case 14:{return-ERRNO_CODES.ENOPROTOOPT};case 15:{var sock=SYSCALLS.getSocketFromFD(),level=SYSCALLS.get(),optname=SYSCALLS.get(),optval=SYSCALLS.get(),optlen=SYSCALLS.get();if(level===1){if(optname===4){HEAP32[optval>>2]=sock.error;HEAP32[optlen>>2]=4;sock.error=null;return 0}}return-ERRNO_CODES.ENOPROTOOPT};case 16:{var sock=SYSCALLS.getSocketFromFD(),message=SYSCALLS.get(),flags=SYSCALLS.get();var iov=HEAP32[message+8>>2];var num=HEAP32[message+12>>2];var addr,port;var name=HEAP32[message>>2];var namelen=HEAP32[message+4>>2];if(name){var info=__read_sockaddr(name,namelen);if(info.errno)return-info.errno;port=info.port;addr=DNS.lookup_addr(info.addr)||info.addr}var total=0;for(var i=0;i<num;i++){total+=HEAP32[iov+(8*i+4)>>2]}var view=new Uint8Array(total);var offset=0;for(var i=0;i<num;i++){var iovbase=HEAP32[iov+(8*i+0)>>2];var iovlen=HEAP32[iov+(8*i+4)>>2];for(var j=0;j<iovlen;j++){view[offset++]=HEAP8[iovbase+j>>0]}}return sock.sock_ops.sendmsg(sock,view,0,total,addr,port)};case 17:{var sock=SYSCALLS.getSocketFromFD(),message=SYSCALLS.get(),flags=SYSCALLS.get();var iov=HEAP32[message+8>>2];var num=HEAP32[message+12>>2];var total=0;for(var i=0;i<num;i++){total+=HEAP32[iov+(8*i+4)>>2]}var msg=sock.sock_ops.recvmsg(sock,total);if(!msg)return 0;var name=HEAP32[message>>2];if(name){var res=__write_sockaddr(name,sock.family,DNS.lookup_name(msg.addr),msg.port);assert(!res.errno)}var bytesRead=0;var bytesRemaining=msg.buffer.byteLength;for(var i=0;bytesRemaining>0&&i<num;i++){var iovbase=HEAP32[iov+(8*i+0)>>2];var iovlen=HEAP32[iov+(8*i+4)>>2];if(!iovlen){continue}var length=Math.min(iovlen,bytesRemaining);var buf=msg.buffer.subarray(bytesRead,bytesRead+length);HEAPU8.set(buf,iovbase+bytesRead);bytesRead+=length;bytesRemaining-=length}return bytesRead};default:abort("unsupported socketcall syscall "+call)}}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}var _atan2=Math_atan2;function _llvm_trap(){abort("trap!")}var _FtoILow=true;function _pthread_self(){return 0}function ___syscall140(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),offset_high=SYSCALLS.get(),offset_low=SYSCALLS.get(),result=SYSCALLS.get(),whence=SYSCALLS.get();var offset=offset_low;assert(offset_high===0);FS.llseek(stream,offset,whence);HEAP32[result>>2]=stream.position;if(stream.getdents&&offset===0&&whence===0)stream.getdents=null;return 0}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall220(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),dirp=SYSCALLS.get(),count=SYSCALLS.get();if(!stream.getdents){stream.getdents=FS.readdir(stream.path)}var pos=0;while(stream.getdents.length>0&&pos+268<count){var id;var type;var name=stream.getdents.pop();assert(name.length<256);if(name[0]==="."){id=1;type=4}else{var child=FS.lookupNode(stream.node,name);id=child.id;type=FS.isChrdev(child.mode)?2:FS.isDir(child.mode)?4:FS.isLink(child.mode)?10:8}HEAP32[dirp+pos>>2]=id;HEAP32[dirp+pos+4>>2]=stream.position;HEAP16[dirp+pos+8>>1]=268;HEAP8[dirp+pos+10>>0]=type;for(var i=0;i<name.length;i++){HEAP8[dirp+pos+(11+i)>>0]=name.charCodeAt(i)}HEAP8[dirp+pos+(11+i)>>0]=0;pos+=268}return pos}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall146(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),iov=SYSCALLS.get(),iovcnt=SYSCALLS.get();return SYSCALLS.doWritev(stream,iov,iovcnt)}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall221(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),cmd=SYSCALLS.get();switch(cmd){case 0:{var arg=SYSCALLS.get();if(arg<0){return-ERRNO_CODES.EINVAL}var newStream;newStream=FS.open(stream.path,stream.flags,0,arg);return newStream.fd};case 1:case 2:return 0;case 3:return stream.flags;case 4:{var arg=SYSCALLS.get();stream.flags|=arg;return 0};case 12:case 12:{var arg=SYSCALLS.get();var offset=0;HEAP16[arg+offset>>1]=2;return 0};case 13:case 14:case 13:case 14:return 0;case 16:case 8:return-ERRNO_CODES.EINVAL;case 9:___setErrNo(ERRNO_CODES.EINVAL);return-1;default:{return-ERRNO_CODES.EINVAL}}}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}function ___syscall145(which,varargs){SYSCALLS.varargs=varargs;try{var stream=SYSCALLS.getStreamFromFD(),iov=SYSCALLS.get(),iovcnt=SYSCALLS.get();return SYSCALLS.doReadv(stream,iov,iovcnt)}catch(e){if(typeof FS==="undefined"||!(e instanceof FS.ErrnoError))abort(e);return-e.errno}}FS.staticInit();__ATINIT__.unshift((function(){if(!Module["noFSInit"]&&!FS.init.initialized)FS.init()}));__ATMAIN__.push((function(){FS.ignorePermissions=false}));__ATEXIT__.push((function(){FS.quit()}));Module["FS_createFolder"]=FS.createFolder;Module["FS_createPath"]=FS.createPath;Module["FS_createDataFile"]=FS.createDataFile;Module["FS_createPreloadedFile"]=FS.createPreloadedFile;Module["FS_createLazyFile"]=FS.createLazyFile;Module["FS_createLink"]=FS.createLink;Module["FS_createDevice"]=FS.createDevice;Module["FS_unlink"]=FS.unlink;__ATINIT__.unshift((function(){TTY.init()}));__ATEXIT__.push((function(){TTY.shutdown()}));if(ENVIRONMENT_IS_NODE){var fs=require("fs");var NODEJS_PATH=require("path");NODEFS.staticInit()}___buildEnvironment(ENV);Module["requestFullScreen"]=function Module_requestFullScreen(lockPointer,resizeCanvas,vrDevice){Browser.requestFullScreen(lockPointer,resizeCanvas,vrDevice)};Module["requestAnimationFrame"]=function Module_requestAnimationFrame(func){Browser.requestAnimationFrame(func)};Module["setCanvasSize"]=function Module_setCanvasSize(width,height,noUpdates){Browser.setCanvasSize(width,height,noUpdates)};Module["pauseMainLoop"]=function Module_pauseMainLoop(){Browser.mainLoop.pause()};Module["resumeMainLoop"]=function Module_resumeMainLoop(){Browser.mainLoop.resume()};Module["getUserMedia"]=function Module_getUserMedia(){Browser.getUserMedia()};Module["createContext"]=function Module_createContext(canvas,useWebGL,setInModule,webGLContextAttributes){return Browser.createContext(canvas,useWebGL,setInModule,webGLContextAttributes)};__ATINIT__.push((function(){SOCKFS.root=FS.mount(SOCKFS,{},null)}));STACK_BASE=STACKTOP=Runtime.alignMemory(STATICTOP);staticSealed=true;STACK_MAX=STACK_BASE+TOTAL_STACK;DYNAMIC_BASE=DYNAMICTOP=Runtime.alignMemory(STACK_MAX);assert(DYNAMIC_BASE<TOTAL_MEMORY,"TOTAL_MEMORY not big enough for stack");var cttz_i8=allocate([8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0],"i8",ALLOC_DYNAMIC);function invoke_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7){try{return Module["dynCall_iiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_iiii(index,a1,a2,a3){try{return Module["dynCall_iiii"](index,a1,a2,a3)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_iddiii(index,a1,a2,a3,a4,a5){try{return Module["dynCall_iddiii"](index,a1,a2,a3,a4,a5)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_i(index){try{return Module["dynCall_i"](index)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_vi(index,a1){try{Module["dynCall_vi"](index,a1)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_vii(index,a1,a2){try{Module["dynCall_vii"](index,a1,a2)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_ii(index,a1){try{return Module["dynCall_ii"](index,a1)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_viiiiiidii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9){try{Module["dynCall_viiiiiidii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_iiiii(index,a1,a2,a3,a4){try{return Module["dynCall_iiiii"](index,a1,a2,a3,a4)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_iii(index,a1,a2){try{return Module["dynCall_iii"](index,a1,a2)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}function invoke_iiiiii(index,a1,a2,a3,a4,a5){try{return Module["dynCall_iiiiii"](index,a1,a2,a3,a4,a5)}catch(e){if(typeof e!=="number"&&e!=="longjmp")throw e;asm["setThrew"](1,0)}}Module.asmGlobalArg={"Math":Math,"Int8Array":Int8Array,"Int16Array":Int16Array,"Int32Array":Int32Array,"Uint8Array":Uint8Array,"Uint16Array":Uint16Array,"Uint32Array":Uint32Array,"Float32Array":Float32Array,"Float64Array":Float64Array,"NaN":NaN,"Infinity":Infinity,"byteLength":byteLength};Module.asmLibraryArg={"abort":abort,"assert":assert,"invoke_iiiiiiii":invoke_iiiiiiii,"invoke_iiii":invoke_iiii,"invoke_iddiii":invoke_iddiii,"invoke_i":invoke_i,"invoke_vi":invoke_vi,"invoke_vii":invoke_vii,"invoke_ii":invoke_ii,"invoke_viiiiiidii":invoke_viiiiiidii,"invoke_iiiii":invoke_iiiii,"invoke_iii":invoke_iii,"invoke_iiiiii":invoke_iiiiii,"_pthread_cleanup_pop":_pthread_cleanup_pop,"_fabs":_fabs,"__inet_pton4_raw":__inet_pton4_raw,"_exp":_exp,"__inet_ntop6_raw":__inet_ntop6_raw,"___syscall102":___syscall102,"_llvm_pow_f64":_llvm_pow_f64,"_acos":_acos,"___syscall54":___syscall54,"_pthread_cleanup_push":_pthread_cleanup_push,"__inet_ntop4_raw":__inet_ntop4_raw,"___syscall40":___syscall40,"_sin":_sin,"___setErrNo":___setErrNo,"__inet_pton6_raw":__inet_pton6_raw,"__read_sockaddr":__read_sockaddr,"_floor":_floor,"_asin":_asin,"___buildEnvironment":___buildEnvironment,"_longjmp":_longjmp,"_time":_time,"_signal":_signal,"_tzset":_tzset,"_emscripten_set_main_loop_timing":_emscripten_set_main_loop_timing,"_fabsf":_fabsf,"_sbrk":_sbrk,"_atan":_atan,"_atan2":_atan2,"___syscall195":___syscall195,"___syscall194":___syscall194,"__exit":__exit,"_mktime":_mktime,"_clock":_clock,"___syscall6":___syscall6,"_emscripten_memcpy_big":_emscripten_memcpy_big,"___syscall221":___syscall221,"_gmtime":_gmtime,"_cos":_cos,"_pthread_self":_pthread_self,"_asctime_r":_asctime_r,"_getenv":_getenv,"_sqrt":_sqrt,"_log":_log,"___unlock":___unlock,"___lock":___lock,"__write_sockaddr":__write_sockaddr,"_emscripten_set_main_loop":_emscripten_set_main_loop,"___syscall10":___syscall10,"_tan":_tan,"_gmtime_r":_gmtime_r,"_getpwnam":_getpwnam,"___syscall3":___syscall3,"_sysconf":_sysconf,"_gethostbyname":_gethostbyname,"_abort":_abort,"___syscall5":___syscall5,"___syscall183":___syscall183,"_alarm":_alarm,"___syscall220":___syscall220,"_inet_addr":_inet_addr,"_emscripten_longjmp":_emscripten_longjmp,"_ceil":_ceil,"_asctime":_asctime,"___syscall140":___syscall140,"_llvm_trap":_llvm_trap,"_exit":_exit,"___syscall145":___syscall145,"___syscall146":___syscall146,"___syscall4":___syscall4,"STACKTOP":STACKTOP,"STACK_MAX":STACK_MAX,"tempDoublePtr":tempDoublePtr,"ABORT":ABORT,"cttz_i8":cttz_i8};// EMSCRIPTEN_START_ASM
var asm=(function(global,env,buffer) {
"almost asm";var a=global.Int8Array;var b=global.Int16Array;var c=global.Int32Array;var d=global.Uint8Array;var e=global.Uint16Array;var f=global.Uint32Array;var g=global.Float32Array;var h=global.Float64Array;var i=new a(buffer);var j=new b(buffer);var k=new c(buffer);var l=new d(buffer);var m=new e(buffer);var n=new f(buffer);var o=new g(buffer);var p=new h(buffer);var q=global.byteLength;var r=env.STACKTOP|0;var s=env.STACK_MAX|0;var t=env.tempDoublePtr|0;var u=env.ABORT|0;var v=env.cttz_i8|0;var w=0;var x=0;var y=0;var z=0;var A=global.NaN,B=global.Infinity;var C=0,D=0,E=0,F=0,G=0.0,H=0,I=0,J=0,K=0.0;var L=0;var M=0;var N=0;var O=0;var P=0;var Q=0;var R=0;var S=0;var T=0;var U=0;var V=global.Math.floor;var W=global.Math.abs;var X=global.Math.sqrt;var Y=global.Math.pow;var Z=global.Math.cos;var _=global.Math.sin;var $=global.Math.tan;var aa=global.Math.acos;var ba=global.Math.asin;var ca=global.Math.atan;var da=global.Math.atan2;var ea=global.Math.exp;var fa=global.Math.log;var ga=global.Math.ceil;var ha=global.Math.imul;var ia=global.Math.min;var ja=global.Math.clz32;var ka=env.abort;var la=env.assert;var ma=env.invoke_iiiiiiii;var na=env.invoke_iiii;var oa=env.invoke_iddiii;var pa=env.invoke_i;var qa=env.invoke_vi;var ra=env.invoke_vii;var sa=env.invoke_ii;var ta=env.invoke_viiiiiidii;var ua=env.invoke_iiiii;var va=env.invoke_iii;var wa=env.invoke_iiiiii;var xa=env._pthread_cleanup_pop;var ya=env._fabs;var za=env.__inet_pton4_raw;var Aa=env._exp;var Ba=env.__inet_ntop6_raw;var Ca=env.___syscall102;var Da=env._llvm_pow_f64;var Ea=env._acos;var Fa=env.___syscall54;var Ga=env._pthread_cleanup_push;var Ha=env.__inet_ntop4_raw;var Ia=env.___syscall40;var Ja=env._sin;var Ka=env.___setErrNo;var La=env.__inet_pton6_raw;var Ma=env.__read_sockaddr;var Na=env._floor;var Oa=env._asin;var Pa=env.___buildEnvironment;var Qa=env._longjmp;var Ra=env._time;var Sa=env._signal;var Ta=env._tzset;var Ua=env._emscripten_set_main_loop_timing;var Va=env._fabsf;var Wa=env._sbrk;var Xa=env._atan;var Ya=env._atan2;var Za=env.___syscall195;var _a=env.___syscall194;var $a=env.__exit;var ab=env._mktime;var bb=env._clock;var cb=env.___syscall6;var db=env._emscripten_memcpy_big;var eb=env.___syscall221;var fb=env._gmtime;var gb=env._cos;var hb=env._pthread_self;var ib=env._asctime_r;var jb=env._getenv;var kb=env._sqrt;var lb=env._log;var mb=env.___unlock;var nb=env.___lock;var ob=env.__write_sockaddr;var pb=env._emscripten_set_main_loop;var qb=env.___syscall10;var rb=env._tan;var sb=env._gmtime_r;var tb=env._getpwnam;var ub=env.___syscall3;var vb=env._sysconf;var wb=env._gethostbyname;var xb=env._abort;var yb=env.___syscall5;var zb=env.___syscall183;var Ab=env._alarm;var Bb=env.___syscall220;var Cb=env._inet_addr;var Db=env._emscripten_longjmp;var Eb=env._ceil;var Fb=env._asctime;var Gb=env.___syscall140;var Hb=env._llvm_trap;var Ib=env._exit;var Jb=env.___syscall145;var Kb=env.___syscall146;var Lb=env.___syscall4;var Mb=0.0;function Nb(newBuffer){if(q(newBuffer)&16777215||q(newBuffer)<=16777215||q(newBuffer)>2147483648)return false;i=new a(newBuffer);j=new b(newBuffer);k=new c(newBuffer);l=new d(newBuffer);m=new e(newBuffer);n=new f(newBuffer);o=new g(newBuffer);p=new h(newBuffer);buffer=newBuffer;return true}
// EMSCRIPTEN_START_FUNCS
function nj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0;d=r;r=r+2528|0;e=d;f=4;g=cy(40)|0;k[g>>2]=0;h=d+12|0;j=d+2416|0;l=d+1216|0;m=d+16|0;n=d+8|0;o=d+4|0;w=0;p=na(43,739441,741186,4)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){s=oy(k[q>>2]|0,g|0,f|0)|0;if(!s)Qa(q|0,x|0);L=x}else s=-1;do if((s|0)!=1){if(!p){w=0;q=na(30,a|0,0,c|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){u=oy(k[t>>2]|0,g|0,f|0)|0;if(!u)Qa(t|0,x|0);L=x}else u=-1;if((u|0)==1){v=0;y=L;break}z=q;dy(g|0);r=d;return z|0}k[29923]=0;k[29967]=0;k[29968]=0;w=0;q=sa(27,739441)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){A=oy(k[t>>2]|0,g|0,f|0)|0;if(!A)Qa(t|0,x|0);L=x}else A=-1;if((A|0)!=1){if(q){g=ly(119700,1,g|0,f|0)|0;f=L;w=0;t=w;w=0;if((t|0)!=0&(x|0)!=0){B=oy(k[t>>2]|0,g|0,f|0)|0;if(!B)Qa(t|0,x|0);L=x}else B=-1;if((B|0)==1){v=q;y=L;break}v=q;y=0;break}w=0;qa(2,741191);t=w;w=0;if((t|0)!=0&(x|0)!=0){C=oy(k[t>>2]|0,g|0,f|0)|0;if(!C)Qa(t|0,x|0);L=x}else C=-1;if((C|0)!=1){z=104;dy(g|0);r=d;return z|0}else{v=q;y=L}}else{v=q;y=L}}else{v=0;y=L}while(0);a:while(1){b:do if(!y){w=0;va(20,14,3)|0;C=w;w=0;if((C|0)!=0&(x|0)!=0){D=oy(k[C>>2]|0,g|0,f|0)|0;if(!D)Qa(C|0,x|0);L=x}else D=-1;if((D|0)==1){E=v;y=L;v=E;continue a}w=0;sa(26,180)|0;C=w;w=0;if((C|0)!=0&(x|0)!=0){F=oy(k[C>>2]|0,g|0,f|0)|0;if(!F)Qa(C|0,x|0);L=x}else F=-1;if((F|0)==1){E=v;y=L;v=E;continue a}w=0;C=ua(3,a|0,h|0,j|0,n|0)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){G=oy(k[B>>2]|0,g|0,f|0)|0;if(!G)Qa(B|0,x|0);L=x}else G=-1;if((G|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=C;if(C){w=0;sa(26,0)|0;C=w;w=0;if((C|0)!=0&(x|0)!=0){H=oy(k[C>>2]|0,g|0,f|0)|0;if(!H)Qa(C|0,x|0);L=x}else H=-1;if((H|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,741253);C=w;w=0;if((C|0)!=0&(x|0)!=0){I=oy(k[C>>2]|0,g|0,f|0)|0;if(!I)Qa(C|0,x|0);L=x}else I=-1;if((I|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,a|0);C=w;w=0;if((C|0)!=0&(x|0)!=0){J=oy(k[C>>2]|0,g|0,f|0)|0;if(!J)Qa(C|0,x|0);L=x}else J=-1;if((J|0)==1){E=v;y=L;v=E;continue a}break}k[29923]=(k[29923]|0)+1;if((i[739441]|0)==33){if((v|0)>0){C=0;do{B=C;C=C+1|0;i[739441+B>>0]=i[739441+C>>0]|0}while((C|0)!=(v|0))}w=0;C=sa(10,739441)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){K=oy(k[B>>2]|0,g|0,f|0)|0;if(!K)Qa(B|0,x|0);L=x}else K=-1;if((K|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=C}C=k[h>>2]|0;w=0;B=sa(28,C|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){M=oy(k[A>>2]|0,g|0,f|0)|0;if(!M)Qa(A|0,x|0);L=x}else M=-1;if((M|0)==1){E=v;y=L;v=E;continue a}A=B<<24>>24;w=0;va(24,A|0,C|0)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){N=oy(k[B>>2]|0,g|0,f|0)|0;if(!N)Qa(B|0,x|0);L=x}else N=-1;if((N|0)==1){E=v;y=L;v=E;continue a}w=0;B=va(25,j|0,739273)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){O=oy(k[u>>2]|0,g|0,f|0)|0;if(!O)Qa(u|0,x|0);L=x}else O=-1;if((O|0)==1){E=v;y=L;v=E;continue a}do if(B){w=0;u=va(25,j|0,739280)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){P=oy(k[p>>2]|0,g|0,f|0)|0;if(!P)Qa(p|0,x|0);L=x}else P=-1;if((P|0)==1){E=v;y=L;v=E;continue a}if(!((A|0)==31|(u|0)==0)){w=0;u=va(10,739441,c|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Q=oy(k[p>>2]|0,g|0,f|0)|0;if(!Q)Qa(p|0,x|0);L=x}else Q=-1;if((Q|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=u;if(u){w=0;qa(2,741295);u=w;w=0;if((u|0)!=0&(x|0)!=0){R=oy(k[u>>2]|0,g|0,f|0)|0;if(!R)Qa(u|0,x|0);L=x}else R=-1;if((R|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,739441);u=w;w=0;if((u|0)!=0&(x|0)!=0){S=oy(k[u>>2]|0,g|0,f|0)|0;if(!S)Qa(u|0,x|0);L=x}else S=-1;if((S|0)==1){E=v;y=L;v=E;continue a}break b}k[29967]=(k[29967]|0)+1;u=k[n>>2]|0;if((u|0)%2880|0){w=0;k[e>>2]=u;na(42,l|0,741453,e|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){T=oy(k[u>>2]|0,g|0,f|0)|0;if(!T)Qa(u|0,x|0);L=x}else T=-1;if((T|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,l|0);u=w;w=0;if((u|0)!=0&(x|0)!=0){U=oy(k[u>>2]|0,g|0,f|0)|0;if(!U)Qa(u|0,x|0);L=x}else U=-1;if((U|0)==1){E=v;y=L;v=E;continue a}}w=0;sa(26,180)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){V=oy(k[u>>2]|0,g|0,f|0)|0;if(!V)Qa(u|0,x|0);L=x}else V=-1;if((V|0)==1){E=v;y=L;v=E;continue a}while(1){w=0;u=ua(2,m|0,1,1200,C|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){W=oy(k[p>>2]|0,g|0,f|0)|0;if(!W)Qa(p|0,x|0);L=x}else W=-1;if((W|0)==1){E=v;y=L;v=E;continue a}if(!u){X=74;break}w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Y=oy(k[p>>2]|0,g|0,f|0)|0;if(!Y)Qa(p|0,x|0);L=x}else Y=-1;if((Y|0)==1){E=v;y=L;v=E;continue a}w=0;p=na(13,k[c>>2]|0,m|0,u|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){Z=oy(k[u>>2]|0,g|0,f|0)|0;if(!Z)Qa(u|0,x|0);L=x}else Z=-1;if((Z|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=p;if(p)break}if((X|0)==74){X=0;w=0;sa(9,k[c>>2]|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){_=oy(k[p>>2]|0,g|0,f|0)|0;if(!_)Qa(p|0,x|0);L=x}else _=-1;if((_|0)==1){E=v;y=L;v=E;continue a}k[29967]=(k[29967]|0)+-1;break}w=0;qa(2,741511);p=w;w=0;if((p|0)!=0&(x|0)!=0){$=oy(k[p>>2]|0,g|0,f|0)|0;if(!$)Qa(p|0,x|0);L=x}else $=-1;if(($|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,a|0);p=w;w=0;if((p|0)!=0&(x|0)!=0){aa=oy(k[p>>2]|0,g|0,f|0)|0;if(!aa)Qa(p|0,x|0);L=x}else aa=-1;if((aa|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,739441);p=w;w=0;if((p|0)!=0&(x|0)!=0){ba=oy(k[p>>2]|0,g|0,f|0)|0;if(!ba)Qa(p|0,x|0);L=x}else ba=-1;if((ba|0)==1){E=v;y=L;v=E;continue a}break b}else X=33}else X=33;while(0);do if((X|0)==33){X=0;w=0;A=va(10,739441,c|0)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){ca=oy(k[B>>2]|0,g|0,f|0)|0;if(!ca)Qa(B|0,x|0);L=x}else ca=-1;if((ca|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=A;if(A){w=0;qa(2,741295);A=w;w=0;if((A|0)!=0&(x|0)!=0){da=oy(k[A>>2]|0,g|0,f|0)|0;if(!da)Qa(A|0,x|0);L=x}else da=-1;if((da|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,739441);A=w;w=0;if((A|0)!=0&(x|0)!=0){ea=oy(k[A>>2]|0,g|0,f|0)|0;if(!ea)Qa(A|0,x|0);L=x}else ea=-1;if((ea|0)==1){E=v;y=L;v=E;continue a}break b}w=0;sa(9,k[c>>2]|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){fa=oy(k[A>>2]|0,g|0,f|0)|0;if(!fa)Qa(A|0,x|0);L=x}else fa=-1;if((fa|0)==1){E=v;y=L;v=E;continue a}w=0;A=va(7,739441,741341)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){ga=oy(k[B>>2]|0,g|0,f|0)|0;if(!ga)Qa(B|0,x|0);L=x}else ga=-1;if((ga|0)==1){E=v;y=L;v=E;continue a}k[29969]=A;if(!A){w=0;qa(2,741343);A=w;w=0;if((A|0)!=0&(x|0)!=0){ha=oy(k[A>>2]|0,g|0,f|0)|0;if(!ha)Qa(A|0,x|0);L=x}else ha=-1;if((ha|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,739441);A=w;w=0;if((A|0)!=0&(x|0)!=0){ia=oy(k[A>>2]|0,g|0,f|0)|0;if(!ia)Qa(A|0,x|0);L=x}else ia=-1;if((ia|0)==1){E=v;y=L;v=E;continue a}break b}k[29968]=(k[29968]|0)+1;k[o>>2]=0;w=0;sa(26,1800)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){ja=oy(k[A>>2]|0,g|0,f|0)|0;if(!ja)Qa(A|0,x|0);L=x}else ja=-1;if((ja|0)==1){E=v;y=L;v=E;continue a}w=0;A=ua(4,a|0,C|0,k[29969]|0,o|0)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){ka=oy(k[B>>2]|0,g|0,f|0)|0;if(!ka)Qa(B|0,x|0);L=x}else ka=-1;if((ka|0)==1){E=v;y=L;v=E;continue a}k[o>>2]=A;w=0;sa(26,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){la=oy(k[A>>2]|0,g|0,f|0)|0;if(!la)Qa(A|0,x|0);L=x}else la=-1;if((la|0)==1){E=v;y=L;v=E;continue a}if(!(k[o>>2]|0)){w=0;sa(4,k[29969]|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){ma=oy(k[A>>2]|0,g|0,f|0)|0;if(!ma)Qa(A|0,x|0);L=x}else ma=-1;if((ma|0)==1){E=v;y=L;v=E;continue a}k[29968]=(k[29968]|0)+-1;break}w=0;qa(2,741393);A=w;w=0;if((A|0)!=0&(x|0)!=0){oa=oy(k[A>>2]|0,g|0,f|0)|0;if(!oa)Qa(A|0,x|0);L=x}else oa=-1;if((oa|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,a|0);A=w;w=0;if((A|0)!=0&(x|0)!=0){pa=oy(k[A>>2]|0,g|0,f|0)|0;if(!pa)Qa(A|0,x|0);L=x}else pa=-1;if((pa|0)==1){E=v;y=L;v=E;continue a}w=0;qa(2,739441);A=w;w=0;if((A|0)!=0&(x|0)!=0){ra=oy(k[A>>2]|0,g|0,f|0)|0;if(!ra)Qa(A|0,x|0);L=x}else ra=-1;if((ra|0)==1){E=v;y=L;v=E;continue a}break b}while(0);w=0;sa(4,C|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){ta=oy(k[A>>2]|0,g|0,f|0)|0;if(!ta)Qa(A|0,x|0);L=x}else ta=-1;if((ta|0)==1){E=v;y=L;v=E;continue a}k[29923]=(k[29923]|0)+-1;w=0;va(20,14,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){wa=oy(k[A>>2]|0,g|0,f|0)|0;if(!wa)Qa(A|0,x|0);L=x}else wa=-1;if((wa|0)==1){E=v;y=L;v=E;continue a}w=0;sa(26,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){xa=oy(k[A>>2]|0,g|0,f|0)|0;if(!xa)Qa(A|0,x|0);L=x}else xa=-1;if((xa|0)==1){E=v;y=L;v=E;continue a}w=0;ya=na(9,739441,b|0,c|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){za=oy(k[A>>2]|0,g|0,f|0)|0;if(!za)Qa(A|0,x|0);L=x}else za=-1;if((za|0)==1){E=v;y=L;v=E;continue a}else{X=80;break a}}else{w=0;qa(2,739091);A=w;w=0;if((A|0)!=0&(x|0)!=0){Aa=oy(k[A>>2]|0,g|0,f|0)|0;if(!Aa)Qa(A|0,x|0);L=x}else Aa=-1;if((Aa|0)==1){E=v;y=L;v=E;continue a}}while(0);w=0;sa(26,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){Ba=oy(k[A>>2]|0,g|0,f|0)|0;if(!Ba)Qa(A|0,x|0);L=x}else Ba=-1;if((Ba|0)==1){E=v;y=L;v=E;continue}if(k[29923]|0){w=0;sa(4,k[h>>2]|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){Ca=oy(k[A>>2]|0,g|0,f|0)|0;if(!Ca)Qa(A|0,x|0);L=x}else Ca=-1;if((Ca|0)==1){E=v;y=L;v=E;continue}}if(k[29968]|0){w=0;sa(4,k[29969]|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){Da=oy(k[A>>2]|0,g|0,f|0)|0;if(!Da)Qa(A|0,x|0);L=x}else Da=-1;if((Da|0)==1){E=v;y=L;v=E;continue}}if(k[29967]|0){w=0;sa(9,k[c>>2]|0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){Ea=oy(k[A>>2]|0,g|0,f|0)|0;if(!Ea)Qa(A|0,x|0);L=x}else Ea=-1;if((Ea|0)==1){E=v;y=L;v=E;continue}}w=0;va(20,14,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){Fa=oy(k[A>>2]|0,g|0,f|0)|0;if(!Fa)Qa(A|0,x|0);L=x}else Fa=-1;if((Fa|0)==1){E=v;y=L;v=E}else{X=92;break}}if((X|0)==80){z=ya;dy(g|0);r=d;return z|0}else if((X|0)==92){z=104;dy(g|0);r=d;return z|0}return 0}function oj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0;d=r;r=r+2416|0;e=4;f=cy(40)|0;k[f>>2]=0;g=d+8|0;h=d+4|0;j=d;l=d+1216|0;m=d+16|0;k[29924]=0;k[29970]=0;k[29971]=0;if(!b){f=ly(119700,1,f|0,e|0)|0;e=L;w=0;b=w;w=0;if((b|0)!=0&(x|0)!=0){n=oy(k[b>>2]|0,f|0,e|0)|0;if(!n)Qa(b|0,x|0);L=x}else n=-1;if((n|0)==1)o=L;else o=0}else{w=0;qa(2,741565);n=w;w=0;if((n|0)!=0&(x|0)!=0){p=oy(k[n>>2]|0,f|0,e|0)|0;if(!p)Qa(n|0,x|0);L=x}else p=-1;if((p|0)!=1){w=0;qa(2,741615);p=w;w=0;if((p|0)!=0&(x|0)!=0){q=oy(k[p>>2]|0,f|0,e|0)|0;if(!q)Qa(p|0,x|0);L=x}else q=-1;if((q|0)!=1){s=104;dy(f|0);r=d;return s|0}else o=L}else o=L}a:while(1){b:do if(!o){w=0;va(20,14,3)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){t=oy(k[q>>2]|0,f|0,e|0)|0;if(!t)Qa(q|0,x|0);L=x}else t=-1;if((t|0)==1){o=L;continue a}w=0;q=sa(27,a|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){u=oy(k[p>>2]|0,f|0,e|0)|0;if(!u)Qa(p|0,x|0);L=x}else u=-1;if((u|0)==1){o=L;continue a}if(q>>>0>1196){w=0;qa(2,741660);q=w;w=0;if((q|0)!=0&(x|0)!=0){v=oy(k[q>>2]|0,f|0,e|0)|0;if(!v)Qa(q|0,x|0);L=x}else v=-1;if((v|0)==1){o=L;continue a}w=0;qa(2,a|0);q=w;w=0;if((q|0)!=0&(x|0)!=0){y=oy(k[q>>2]|0,f|0,e|0)|0;if(!y)Qa(q|0,x|0);L=x}else y=-1;if((y|0)==1){o=L;continue a}break}w=0;sa(26,180)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){z=oy(k[q>>2]|0,f|0,e|0)|0;if(!z)Qa(q|0,x|0);L=x}else z=-1;if((z|0)==1){o=L;continue a}w=0;va(23,l|0,a|0)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){A=oy(k[q>>2]|0,f|0,e|0)|0;if(!A)Qa(q|0,x|0);L=x}else A=-1;if((A|0)==1){o=L;continue a}w=0;q=va(21,l|0,739111)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){B=oy(k[p>>2]|0,f|0,e|0)|0;if(!B)Qa(p|0,x|0);L=x}else B=-1;if((B|0)==1){o=L;continue a}do if(!q){w=0;p=va(21,l|0,739114)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){C=oy(k[n>>2]|0,f|0,e|0)|0;if(!C)Qa(n|0,x|0);L=x}else C=-1;if((C|0)==1){o=L;continue a}if(!p){w=0;va(23,l|0,a|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){D=oy(k[p>>2]|0,f|0,e|0)|0;if(!D)Qa(p|0,x|0);L=x}else D=-1;if((D|0)==1){o=L;continue a}w=0;p=sa(27,l|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){E=oy(k[n>>2]|0,f|0,e|0)|0;if(!E)Qa(n|0,x|0);L=x}else E=-1;if((E|0)==1){o=L;continue a}n=l+p|0;i[n>>0]=46;i[n+1>>0]=103;i[n+2>>0]=122;i[n+3>>0]=0;w=0;sa(26,180)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){F=oy(k[n>>2]|0,f|0,e|0)|0;if(!F)Qa(n|0,x|0);L=x}else F=-1;if((F|0)==1){o=L;continue a}w=0;n=ua(5,l|0,g|0,h|0,j|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){G=oy(k[p>>2]|0,f|0,e|0)|0;if(!G)Qa(p|0,x|0);L=x}else G=-1;if((G|0)==1){o=L;continue a}if(n){w=0;sa(26,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){H=oy(k[n>>2]|0,f|0,e|0)|0;if(!H)Qa(n|0,x|0);L=x}else H=-1;if((H|0)==1){o=L;continue a}w=0;va(23,l|0,a|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){I=oy(k[n>>2]|0,f|0,e|0)|0;if(!I)Qa(n|0,x|0);L=x}else I=-1;if((I|0)==1){o=L;continue a}w=0;n=sa(27,l|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){J=oy(k[p>>2]|0,f|0,e|0)|0;if(!J)Qa(p|0,x|0);L=x}else J=-1;if((J|0)==1){o=L;continue a}p=l+n|0;i[p>>0]=i[739111]|0;i[p+1>>0]=i[739112]|0;i[p+2>>0]=i[739113]|0;w=0;sa(26,180)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){K=oy(k[p>>2]|0,f|0,e|0)|0;if(!K)Qa(p|0,x|0);L=x}else K=-1;if((K|0)==1){o=L;continue a}w=0;p=ua(5,l|0,g|0,h|0,j|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){M=oy(k[n>>2]|0,f|0,e|0)|0;if(!M)Qa(n|0,x|0);L=x}else M=-1;if((M|0)==1){o=L;continue a}if(!p)break;w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){N=oy(k[p>>2]|0,f|0,e|0)|0;if(!N)Qa(p|0,x|0);L=x}else N=-1;if((N|0)==1){o=L;continue a}w=0;va(23,l|0,a|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){O=oy(k[p>>2]|0,f|0,e|0)|0;if(!O)Qa(p|0,x|0);L=x}else O=-1;if((O|0)==1){o=L;continue a}w=0;sa(26,180)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){P=oy(k[p>>2]|0,f|0,e|0)|0;if(!P)Qa(p|0,x|0);L=x}else P=-1;if((P|0)==1){o=L;continue a}w=0;p=ua(5,l|0,g|0,h|0,j|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Q=oy(k[n>>2]|0,f|0,e|0)|0;if(!Q)Qa(n|0,x|0);L=x}else Q=-1;if((Q|0)==1){o=L;continue a}if(!p)break;w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){R=oy(k[p>>2]|0,f|0,e|0)|0;if(!R)Qa(p|0,x|0);L=x}else R=-1;if((R|0)==1){o=L;continue a}w=0;qa(2,741689);p=w;w=0;if((p|0)!=0&(x|0)!=0){S=oy(k[p>>2]|0,f|0,e|0)|0;if(!S)Qa(p|0,x|0);L=x}else S=-1;if((S|0)==1){o=L;continue a}w=0;qa(2,l|0);p=w;w=0;if((p|0)!=0&(x|0)!=0){T=oy(k[p>>2]|0,f|0,e|0)|0;if(!T)Qa(p|0,x|0);L=x}else T=-1;if((T|0)==1){o=L;continue a}break b}}else U=22}else U=22;while(0);if((U|0)==22){U=0;w=0;sa(26,180)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){V=oy(k[q>>2]|0,f|0,e|0)|0;if(!V)Qa(q|0,x|0);L=x}else V=-1;if((V|0)==1){o=L;continue a}w=0;q=ua(5,a|0,g|0,h|0,j|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){W=oy(k[p>>2]|0,f|0,e|0)|0;if(!W)Qa(p|0,x|0);L=x}else W=-1;if((W|0)==1){o=L;continue a}if(q){w=0;sa(26,0)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){X=oy(k[q>>2]|0,f|0,e|0)|0;if(!X)Qa(q|0,x|0);L=x}else X=-1;if((X|0)==1){o=L;continue a}w=0;qa(2,741689);q=w;w=0;if((q|0)!=0&(x|0)!=0){Y=oy(k[q>>2]|0,f|0,e|0)|0;if(!Y)Qa(q|0,x|0);L=x}else Y=-1;if((Y|0)==1){o=L;continue a}w=0;qa(2,a|0);q=w;w=0;if((q|0)!=0&(x|0)!=0){Z=oy(k[q>>2]|0,f|0,e|0)|0;if(!Z)Qa(q|0,x|0);L=x}else Z=-1;if((Z|0)==1){o=L;continue a}break}}k[29971]=(k[29971]|0)+1;k[29970]=(k[29970]|0)+1;w=0;q=va(12,a|0,c|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){_=oy(k[p>>2]|0,f|0,e|0)|0;if(!_)Qa(p|0,x|0);L=x}else _=-1;if((_|0)==1){o=L;continue a}if(q){w=0;qa(2,741724);q=w;w=0;if((q|0)!=0&(x|0)!=0){$=oy(k[q>>2]|0,f|0,e|0)|0;if(!$)Qa(q|0,x|0);L=x}else $=-1;if(($|0)==1){o=L;continue a}w=0;qa(2,a|0);q=w;w=0;if((q|0)!=0&(x|0)!=0){aa=oy(k[q>>2]|0,f|0,e|0)|0;if(!aa)Qa(q|0,x|0);L=x}else aa=-1;if((aa|0)==1){o=L;continue a}break}k[29924]=(k[29924]|0)+1;q=k[g>>2]|0;w=0;p=sa(28,q|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){ba=oy(k[n>>2]|0,f|0,e|0)|0;if(!ba)Qa(n|0,x|0);L=x}else ba=-1;if((ba|0)==1){o=L;continue a}n=p<<24>>24;w=0;va(24,n|0,q|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){ca=oy(k[p>>2]|0,f|0,e|0)|0;if(!ca)Qa(p|0,x|0);L=x}else ca=-1;if((ca|0)==1){o=L;continue a}w=0;p=va(21,l|0,739114)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){da=oy(k[b>>2]|0,f|0,e|0)|0;if(!da)Qa(b|0,x|0);L=x}else da=-1;if((da|0)==1){o=L;continue a}c:do if(!p){w=0;b=va(21,l|0,739111)|0;ea=w;w=0;if((ea|0)!=0&(x|0)!=0){fa=oy(k[ea>>2]|0,f|0,e|0)|0;if(!fa)Qa(ea|0,x|0);L=x}else fa=-1;if((fa|0)==1){o=L;continue a}if(!((n|0)==31|(b|0)!=0)){w=0;sa(26,180)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ga=oy(k[b>>2]|0,f|0,e|0)|0;if(!ga)Qa(b|0,x|0);L=x}else ga=-1;if((ga|0)==1){o=L;continue a}w=0;b=ua(2,m|0,1,1200,q|0)|0;ea=w;w=0;if((ea|0)!=0&(x|0)!=0){ha=oy(k[ea>>2]|0,f|0,e|0)|0;if(!ha)Qa(ea|0,x|0);L=x}else ha=-1;if((ha|0)==1){o=L;continue a}if(!b)break;else ia=b;while(1){w=0;sa(26,0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ja=oy(k[b>>2]|0,f|0,e|0)|0;if(!ja)Qa(b|0,x|0);L=x}else ja=-1;if((ja|0)==1){o=L;continue a}w=0;b=na(17,k[c>>2]|0,m|0,ia|0)|0;ea=w;w=0;if((ea|0)!=0&(x|0)!=0){ka=oy(k[ea>>2]|0,f|0,e|0)|0;if(!ka)Qa(ea|0,x|0);L=x}else ka=-1;if((ka|0)==1){o=L;continue a}if(b)break;w=0;sa(26,180)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){la=oy(k[b>>2]|0,f|0,e|0)|0;if(!la)Qa(b|0,x|0);L=x}else la=-1;if((la|0)==1){o=L;continue a}w=0;ia=ua(2,m|0,1,1200,q|0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ma=oy(k[b>>2]|0,f|0,e|0)|0;if(!ma)Qa(b|0,x|0);L=x}else ma=-1;if((ma|0)==1){o=L;continue a}if(!ia)break c}w=0;qa(2,741828);b=w;w=0;if((b|0)!=0&(x|0)!=0){oa=oy(k[b>>2]|0,f|0,e|0)|0;if(!oa)Qa(b|0,x|0);L=x}else oa=-1;if((oa|0)==1){o=L;continue a}w=0;qa(2,a|0);b=w;w=0;if((b|0)!=0&(x|0)!=0){pa=oy(k[b>>2]|0,f|0,e|0)|0;if(!pa)Qa(b|0,x|0);L=x}else pa=-1;if((pa|0)==1){o=L;continue a}break b}else U=60}else U=60;while(0);do if((U|0)==60){U=0;w=0;sa(26,1800)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){ra=oy(k[n>>2]|0,f|0,e|0)|0;if(!ra)Qa(n|0,x|0);L=x}else ra=-1;if((ra|0)==1){o=L;continue a}w=0;n=na(41,a|0,q|0,k[c>>2]|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){ta=oy(k[p>>2]|0,f|0,e|0)|0;if(!ta)Qa(p|0,x|0);L=x}else ta=-1;if((ta|0)==1){o=L;continue a}w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){wa=oy(k[p>>2]|0,f|0,e|0)|0;if(!wa)Qa(p|0,x|0);L=x}else wa=-1;if((wa|0)==1){o=L;continue a}if(!n)break;w=0;qa(2,741780);n=w;w=0;if((n|0)!=0&(x|0)!=0){xa=oy(k[n>>2]|0,f|0,e|0)|0;if(!xa)Qa(n|0,x|0);L=x}else xa=-1;if((xa|0)==1){o=L;continue a}w=0;qa(2,a|0);n=w;w=0;if((n|0)!=0&(x|0)!=0){ya=oy(k[n>>2]|0,f|0,e|0)|0;if(!ya)Qa(n|0,x|0);L=x}else ya=-1;if((ya|0)==1){o=L;continue a}break b}while(0);w=0;sa(4,q|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){za=oy(k[n>>2]|0,f|0,e|0)|0;if(!za)Qa(n|0,x|0);L=x}else za=-1;if((za|0)==1){o=L;continue a}k[29971]=(k[29971]|0)+-1;n=k[j>>2]|0;d:do if((n|0)>=0){p=0;while(1){w=0;b=ua(6,n|0,741866+p|0,5-p|0,0)|0;ea=w;w=0;if((ea|0)!=0&(x|0)!=0){Aa=oy(k[ea>>2]|0,f|0,e|0)|0;if(!Aa)Qa(ea|0,x|0);L=x}else Aa=-1;if((Aa|0)==1){o=L;continue a}if((b|0)<1)break d;p=b+p|0;if((p|0)>=5)break}}while(0);w=0;sa(4,k[h>>2]|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ba=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ba)Qa(n|0,x|0);L=x}else Ba=-1;if((Ba|0)==1){o=L;continue a}k[29970]=(k[29970]|0)+-1;w=0;va(20,14,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ca=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ca)Qa(n|0,x|0);L=x}else Ca=-1;if((Ca|0)==1){o=L;continue a}w=0;sa(26,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Da=oy(k[n>>2]|0,f|0,e|0)|0;if(!Da)Qa(n|0,x|0);L=x}else Da=-1;if((Da|0)==1){o=L;continue a}w=0;Ea=na(15,k[c>>2]|0,0,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Fa=oy(k[n>>2]|0,f|0,e|0)|0;if(!Fa)Qa(n|0,x|0);L=x}else Fa=-1;if((Fa|0)==1){o=L;continue a}else{U=88;break a}}else{w=0;qa(2,739091);n=w;w=0;if((n|0)!=0&(x|0)!=0){Ga=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ga)Qa(n|0,x|0);L=x}else Ga=-1;if((Ga|0)==1){o=L;continue a}}while(0);w=0;sa(26,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ha=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ha)Qa(n|0,x|0);L=x}else Ha=-1;if((Ha|0)==1){o=L;continue}if(k[29970]|0){w=0;sa(4,k[h>>2]|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ia=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ia)Qa(n|0,x|0);L=x}else Ia=-1;if((Ia|0)==1){o=L;continue}}if(k[29971]|0){w=0;sa(4,k[g>>2]|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ja=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ja)Qa(n|0,x|0);L=x}else Ja=-1;if((Ja|0)==1){o=L;continue}}if(k[29924]|0){w=0;sa(15,k[c>>2]|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){Ka=oy(k[n>>2]|0,f|0,e|0)|0;if(!Ka)Qa(n|0,x|0);L=x}else Ka=-1;if((Ka|0)==1){o=L;continue}}w=0;va(20,14,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){La=oy(k[n>>2]|0,f|0,e|0)|0;if(!La)Qa(n|0,x|0);L=x}else La=-1;if((La|0)==1)o=L;else{U=100;break}}if((U|0)==88){s=Ea;dy(f|0);r=d;return s|0}else if((U|0)==100){s=104;dy(f|0);r=d;return s|0}return 0}function pj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0;d=r;r=r+1216|0;e=4;f=cy(40)|0;k[f>>2]=0;g=d+12|0;h=d+8|0;j=d+16|0;l=d+4|0;m=d;w=0;n=na(43,739441,741186,4)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){p=oy(k[o>>2]|0,f|0,e|0)|0;if(!p)Qa(o|0,x|0);L=x}else p=-1;do if((p|0)!=1){if(!n){w=0;o=na(34,a|0,0,c|0)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){s=oy(k[q>>2]|0,f|0,e|0)|0;if(!s)Qa(q|0,x|0);L=x}else s=-1;if((s|0)==1){t=0;u=L;break}v=o;dy(f|0);r=d;return v|0}k[29971]=0;k[29970]=0;k[29967]=0;k[29968]=0;w=0;o=sa(27,739441)|0;q=w;w=0;if((q|0)!=0&(x|0)!=0){y=oy(k[q>>2]|0,f|0,e|0)|0;if(!y)Qa(q|0,x|0);L=x}else y=-1;if((y|0)!=1){if(o){f=ly(119700,1,f|0,e|0)|0;e=L;w=0;q=w;w=0;if((q|0)!=0&(x|0)!=0){z=oy(k[q>>2]|0,f|0,e|0)|0;if(!z)Qa(q|0,x|0);L=x}else z=-1;if((z|0)==1){t=o;u=L;break}t=o;u=0;break}w=0;qa(2,741872);q=w;w=0;if((q|0)!=0&(x|0)!=0){A=oy(k[q>>2]|0,f|0,e|0)|0;if(!A)Qa(q|0,x|0);L=x}else A=-1;if((A|0)!=1){v=104;dy(f|0);r=d;return v|0}else{t=o;u=L}}else{t=o;u=L}}else{t=0;u=L}while(0);a:while(1){b:do if(!u){w=0;va(20,14,3)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){B=oy(k[A>>2]|0,f|0,e|0)|0;if(!B)Qa(A|0,x|0);L=x}else B=-1;if((B|0)==1){C=t;u=L;t=C;continue a}w=0;sa(26,180)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){D=oy(k[A>>2]|0,f|0,e|0)|0;if(!D)Qa(A|0,x|0);L=x}else D=-1;if((D|0)==1){C=t;u=L;t=C;continue a}w=0;A=ua(5,a|0,g|0,h|0,l|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){E=oy(k[z>>2]|0,f|0,e|0)|0;if(!E)Qa(z|0,x|0);L=x}else E=-1;if((E|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=A;if(A){w=0;sa(26,0)|0;A=w;w=0;if((A|0)!=0&(x|0)!=0){F=oy(k[A>>2]|0,f|0,e|0)|0;if(!F)Qa(A|0,x|0);L=x}else F=-1;if((F|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,741933);A=w;w=0;if((A|0)!=0&(x|0)!=0){G=oy(k[A>>2]|0,f|0,e|0)|0;if(!G)Qa(A|0,x|0);L=x}else G=-1;if((G|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,a|0);A=w;w=0;if((A|0)!=0&(x|0)!=0){H=oy(k[A>>2]|0,f|0,e|0)|0;if(!H)Qa(A|0,x|0);L=x}else H=-1;if((H|0)==1){C=t;u=L;t=C;continue a}break}k[29971]=(k[29971]|0)+1;k[29970]=(k[29970]|0)+1;if((i[739441]|0)==33){if((t|0)>0){A=0;do{z=A;A=A+1|0;i[739441+z>>0]=i[739441+A>>0]|0}while((A|0)!=(t|0))}w=0;A=sa(10,739441)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){I=oy(k[z>>2]|0,f|0,e|0)|0;if(!I)Qa(z|0,x|0);L=x}else I=-1;if((I|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=A}A=k[g>>2]|0;w=0;z=sa(28,A|0)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){J=oy(k[y>>2]|0,f|0,e|0)|0;if(!J)Qa(y|0,x|0);L=x}else J=-1;if((J|0)==1){C=t;u=L;t=C;continue a}y=z<<24>>24;w=0;va(24,y|0,A|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){K=oy(k[z>>2]|0,f|0,e|0)|0;if(!K)Qa(z|0,x|0);L=x}else K=-1;if((K|0)==1){C=t;u=L;t=C;continue a}w=0;z=va(21,a|0,739114)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){M=oy(k[s>>2]|0,f|0,e|0)|0;if(!M)Qa(s|0,x|0);L=x}else M=-1;if((M|0)==1){C=t;u=L;t=C;continue a}if(!z){w=0;z=va(21,a|0,739111)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){N=oy(k[s>>2]|0,f|0,e|0)|0;if(!N)Qa(s|0,x|0);L=x}else N=-1;if((N|0)==1){C=t;u=L;t=C;continue a}if(!((y|0)==31|(z|0)!=0)){w=0;z=va(10,739441,c|0)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){O=oy(k[y>>2]|0,f|0,e|0)|0;if(!O)Qa(y|0,x|0);L=x}else O=-1;if((O|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=z;if(z){w=0;qa(2,741974);z=w;w=0;if((z|0)!=0&(x|0)!=0){P=oy(k[z>>2]|0,f|0,e|0)|0;if(!P)Qa(z|0,x|0);L=x}else P=-1;if((P|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,739441);z=w;w=0;if((z|0)!=0&(x|0)!=0){Q=oy(k[z>>2]|0,f|0,e|0)|0;if(!Q)Qa(z|0,x|0);L=x}else Q=-1;if((Q|0)==1){C=t;u=L;t=C;continue a}break}k[29967]=(k[29967]|0)+1;w=0;sa(26,180)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){R=oy(k[z>>2]|0,f|0,e|0)|0;if(!R)Qa(z|0,x|0);L=x}else R=-1;if((R|0)==1){C=t;u=L;t=C;continue a}w=0;z=ua(2,j|0,1,1200,A|0)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){S=oy(k[y>>2]|0,f|0,e|0)|0;if(!S)Qa(y|0,x|0);L=x}else S=-1;if((S|0)==1){C=t;u=L;t=C;continue a}c:do if(z){y=z;while(1){w=0;sa(26,0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){T=oy(k[s>>2]|0,f|0,e|0)|0;if(!T)Qa(s|0,x|0);L=x}else T=-1;if((T|0)==1){C=t;u=L;t=C;continue a}w=0;s=na(13,k[c>>2]|0,j|0,y|0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){U=oy(k[n>>2]|0,f|0,e|0)|0;if(!U)Qa(n|0,x|0);L=x}else U=-1;if((U|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=s;if(s)break;w=0;sa(26,180)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){V=oy(k[s>>2]|0,f|0,e|0)|0;if(!V)Qa(s|0,x|0);L=x}else V=-1;if((V|0)==1){C=t;u=L;t=C;continue a}w=0;y=ua(2,j|0,1,1200,A|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){W=oy(k[s>>2]|0,f|0,e|0)|0;if(!W)Qa(s|0,x|0);L=x}else W=-1;if((W|0)==1){C=t;u=L;t=C;continue a}if(!y)break c}w=0;qa(2,742121);y=w;w=0;if((y|0)!=0&(x|0)!=0){X=oy(k[y>>2]|0,f|0,e|0)|0;if(!X)Qa(y|0,x|0);L=x}else X=-1;if((X|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,a|0);y=w;w=0;if((y|0)!=0&(x|0)!=0){Y=oy(k[y>>2]|0,f|0,e|0)|0;if(!Y)Qa(y|0,x|0);L=x}else Y=-1;if((Y|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,739441);y=w;w=0;if((y|0)!=0&(x|0)!=0){Z=oy(k[y>>2]|0,f|0,e|0)|0;if(!Z)Qa(y|0,x|0);L=x}else Z=-1;if((Z|0)==1){C=t;u=L;t=C;continue a}break b}while(0);w=0;sa(9,k[c>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){_=oy(k[z>>2]|0,f|0,e|0)|0;if(!_)Qa(z|0,x|0);L=x}else _=-1;if((_|0)==1){C=t;u=L;t=C;continue a}}else $=33}else $=33;do if(($|0)==33){$=0;w=0;z=va(10,739441,c|0)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){aa=oy(k[y>>2]|0,f|0,e|0)|0;if(!aa)Qa(y|0,x|0);L=x}else aa=-1;if((aa|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=z;if(z){w=0;qa(2,741974);z=w;w=0;if((z|0)!=0&(x|0)!=0){ba=oy(k[z>>2]|0,f|0,e|0)|0;if(!ba)Qa(z|0,x|0);L=x}else ba=-1;if((ba|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,739441);z=w;w=0;if((z|0)!=0&(x|0)!=0){ca=oy(k[z>>2]|0,f|0,e|0)|0;if(!ca)Qa(z|0,x|0);L=x}else ca=-1;if((ca|0)==1){C=t;u=L;t=C;continue a}break b}w=0;sa(9,k[c>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){da=oy(k[z>>2]|0,f|0,e|0)|0;if(!da)Qa(z|0,x|0);L=x}else da=-1;if((da|0)==1){C=t;u=L;t=C;continue a}w=0;z=va(7,739441,741341)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){ea=oy(k[y>>2]|0,f|0,e|0)|0;if(!ea)Qa(y|0,x|0);L=x}else ea=-1;if((ea|0)==1){C=t;u=L;t=C;continue a}k[29969]=z;if(!z){w=0;qa(2,742019);z=w;w=0;if((z|0)!=0&(x|0)!=0){fa=oy(k[z>>2]|0,f|0,e|0)|0;if(!fa)Qa(z|0,x|0);L=x}else fa=-1;if((fa|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,739441);z=w;w=0;if((z|0)!=0&(x|0)!=0){ga=oy(k[z>>2]|0,f|0,e|0)|0;if(!ga)Qa(z|0,x|0);L=x}else ga=-1;if((ga|0)==1){C=t;u=L;t=C;continue a}break b}k[29968]=(k[29968]|0)+1;k[m>>2]=0;w=0;sa(26,1800)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){ha=oy(k[z>>2]|0,f|0,e|0)|0;if(!ha)Qa(z|0,x|0);L=x}else ha=-1;if((ha|0)==1){C=t;u=L;t=C;continue a}w=0;z=ua(4,a|0,A|0,k[29969]|0,m|0)|0;y=w;w=0;if((y|0)!=0&(x|0)!=0){ia=oy(k[y>>2]|0,f|0,e|0)|0;if(!ia)Qa(y|0,x|0);L=x}else ia=-1;if((ia|0)==1){C=t;u=L;t=C;continue a}k[m>>2]=z;w=0;sa(26,0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){ja=oy(k[z>>2]|0,f|0,e|0)|0;if(!ja)Qa(z|0,x|0);L=x}else ja=-1;if((ja|0)==1){C=t;u=L;t=C;continue a}if(!(k[m>>2]|0)){w=0;sa(4,k[29969]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){ka=oy(k[z>>2]|0,f|0,e|0)|0;if(!ka)Qa(z|0,x|0);L=x}else ka=-1;if((ka|0)==1){C=t;u=L;t=C;continue a}k[29968]=(k[29968]|0)+-1;break}w=0;qa(2,742068);z=w;w=0;if((z|0)!=0&(x|0)!=0){la=oy(k[z>>2]|0,f|0,e|0)|0;if(!la)Qa(z|0,x|0);L=x}else la=-1;if((la|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,a|0);z=w;w=0;if((z|0)!=0&(x|0)!=0){ma=oy(k[z>>2]|0,f|0,e|0)|0;if(!ma)Qa(z|0,x|0);L=x}else ma=-1;if((ma|0)==1){C=t;u=L;t=C;continue a}w=0;qa(2,739441);z=w;w=0;if((z|0)!=0&(x|0)!=0){oa=oy(k[z>>2]|0,f|0,e|0)|0;if(!oa)Qa(z|0,x|0);L=x}else oa=-1;if((oa|0)==1){C=t;u=L;t=C;continue a}break b}while(0);w=0;sa(4,A|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){pa=oy(k[z>>2]|0,f|0,e|0)|0;if(!pa)Qa(z|0,x|0);L=x}else pa=-1;if((pa|0)==1){C=t;u=L;t=C;continue a}k[29971]=(k[29971]|0)+-1;z=k[l>>2]|0;d:do if((z|0)>=0){y=0;while(1){w=0;s=ua(6,z|0,741866+y|0,5-y|0,0)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){ra=oy(k[n>>2]|0,f|0,e|0)|0;if(!ra)Qa(n|0,x|0);L=x}else ra=-1;if((ra|0)==1){C=t;u=L;t=C;continue a}if((s|0)<1)break d;y=s+y|0;if((y|0)>=5)break}}while(0);w=0;sa(4,k[h>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){ta=oy(k[z>>2]|0,f|0,e|0)|0;if(!ta)Qa(z|0,x|0);L=x}else ta=-1;if((ta|0)==1){C=t;u=L;t=C;continue a}k[29970]=(k[29970]|0)+-1;w=0;va(20,14,0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){wa=oy(k[z>>2]|0,f|0,e|0)|0;if(!wa)Qa(z|0,x|0);L=x}else wa=-1;if((wa|0)==1){C=t;u=L;t=C;continue a}w=0;sa(26,0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){xa=oy(k[z>>2]|0,f|0,e|0)|0;if(!xa)Qa(z|0,x|0);L=x}else xa=-1;if((xa|0)==1){C=t;u=L;t=C;continue a}w=0;ya=na(9,739441,b|0,c|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){za=oy(k[z>>2]|0,f|0,e|0)|0;if(!za)Qa(z|0,x|0);L=x}else za=-1;if((za|0)==1){C=t;u=L;t=C;continue a}else{$=83;break a}}else{w=0;qa(2,739091);z=w;w=0;if((z|0)!=0&(x|0)!=0){Aa=oy(k[z>>2]|0,f|0,e|0)|0;if(!Aa)Qa(z|0,x|0);L=x}else Aa=-1;if((Aa|0)==1){C=t;u=L;t=C;continue a}}while(0);w=0;sa(26,0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Ba=oy(k[z>>2]|0,f|0,e|0)|0;if(!Ba)Qa(z|0,x|0);L=x}else Ba=-1;if((Ba|0)==1){C=t;u=L;t=C;continue}if(k[29971]|0){w=0;sa(4,k[g>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Ca=oy(k[z>>2]|0,f|0,e|0)|0;if(!Ca)Qa(z|0,x|0);L=x}else Ca=-1;if((Ca|0)==1){C=t;u=L;t=C;continue}}if(k[29970]|0){w=0;sa(4,k[h>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Da=oy(k[z>>2]|0,f|0,e|0)|0;if(!Da)Qa(z|0,x|0);L=x}else Da=-1;if((Da|0)==1){C=t;u=L;t=C;continue}}if(k[29968]|0){w=0;sa(4,k[29969]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Ea=oy(k[z>>2]|0,f|0,e|0)|0;if(!Ea)Qa(z|0,x|0);L=x}else Ea=-1;if((Ea|0)==1){C=t;u=L;t=C;continue}}if(k[29967]|0){w=0;sa(9,k[c>>2]|0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Fa=oy(k[z>>2]|0,f|0,e|0)|0;if(!Fa)Qa(z|0,x|0);L=x}else Fa=-1;if((Fa|0)==1){C=t;u=L;t=C;continue}}w=0;va(20,14,0)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){Ga=oy(k[z>>2]|0,f|0,e|0)|0;if(!Ga)Qa(z|0,x|0);L=x}else Ga=-1;if((Ga|0)==1){C=t;u=L;t=C}else{$=98;break}}if(($|0)==83){v=ya;dy(f|0);r=d;return v|0}else if(($|0)==98){v=104;dy(f|0);r=d;return v|0}return 0}function qj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0,wa=0,xa=0,ya=0,za=0,Aa=0;d=r;r=r+1216|0;e=4;f=cy(40)|0;k[f>>2]=0;g=d+8|0;h=d+4|0;j=d+16|0;l=d;k[29971]=0;k[29970]=0;k[29924]=0;k[29965]=0;k[29964]=0;do if(!b){w=0;m=sa(27,739441)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){o=oy(k[n>>2]|0,f|0,e|0)|0;if(!o)Qa(n|0,x|0);L=x}else o=-1;if((o|0)!=1){if(m){f=ly(119700,1,f|0,e|0)|0;e=L;w=0;n=w;w=0;if((n|0)!=0&(x|0)!=0){p=oy(k[n>>2]|0,f|0,e|0)|0;if(!p)Qa(n|0,x|0);L=x}else p=-1;if((p|0)==1){q=m;s=L;break}q=m;s=0;break}w=0;qa(2,742185);n=w;w=0;if((n|0)!=0&(x|0)!=0){t=oy(k[n>>2]|0,f|0,e|0)|0;if(!t)Qa(n|0,x|0);L=x}else t=-1;if((t|0)!=1){u=104;dy(f|0);r=d;return u|0}else{q=m;s=L}}else{q=m;s=L}}else{w=0;qa(2,742156);m=w;w=0;if((m|0)!=0&(x|0)!=0){v=oy(k[m>>2]|0,f|0,e|0)|0;if(!v)Qa(m|0,x|0);L=x}else v=-1;if((v|0)==1){q=0;s=L}else{u=104;dy(f|0);r=d;return u|0}}while(0);a:while(1){b:do if(!s){w=0;va(20,14,3)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){y=oy(k[v>>2]|0,f|0,e|0)|0;if(!y)Qa(v|0,x|0);L=x}else y=-1;if((y|0)==1){z=q;s=L;q=z;continue a}w=0;sa(26,180)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){A=oy(k[v>>2]|0,f|0,e|0)|0;if(!A)Qa(v|0,x|0);L=x}else A=-1;if((A|0)==1){z=q;s=L;q=z;continue a}w=0;v=ua(5,a|0,g|0,h|0,l|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){B=oy(k[t>>2]|0,f|0,e|0)|0;if(!B)Qa(t|0,x|0);L=x}else B=-1;if((B|0)==1){z=q;s=L;q=z;continue a}if(v){w=0;sa(26,0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){C=oy(k[v>>2]|0,f|0,e|0)|0;if(!C)Qa(v|0,x|0);L=x}else C=-1;if((C|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,742250);v=w;w=0;if((v|0)!=0&(x|0)!=0){D=oy(k[v>>2]|0,f|0,e|0)|0;if(!D)Qa(v|0,x|0);L=x}else D=-1;if((D|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,a|0);v=w;w=0;if((v|0)!=0&(x|0)!=0){E=oy(k[v>>2]|0,f|0,e|0)|0;if(!E)Qa(v|0,x|0);L=x}else E=-1;if((E|0)==1){z=q;s=L;q=z;continue a}break}k[29971]=(k[29971]|0)+1;k[29970]=(k[29970]|0)+1;v=k[g>>2]|0;w=0;t=sa(28,v|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){F=oy(k[p>>2]|0,f|0,e|0)|0;if(!F)Qa(p|0,x|0);L=x}else F=-1;if((F|0)==1){z=q;s=L;q=z;continue a}p=t<<24>>24;w=0;va(24,p|0,v|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){G=oy(k[t>>2]|0,f|0,e|0)|0;if(!G)Qa(t|0,x|0);L=x}else G=-1;if((G|0)==1){z=q;s=L;q=z;continue a}w=0;t=va(21,a|0,739114)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){H=oy(k[o>>2]|0,f|0,e|0)|0;if(!H)Qa(o|0,x|0);L=x}else H=-1;if((H|0)==1){z=q;s=L;q=z;continue a}if(!t){w=0;t=va(21,a|0,739111)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){I=oy(k[o>>2]|0,f|0,e|0)|0;if(!I)Qa(o|0,x|0);L=x}else I=-1;if((I|0)==1){z=q;s=L;q=z;continue a}if(!((p|0)==31|(t|0)!=0)){w=0;qa(2,742534);t=w;w=0;if((t|0)!=0&(x|0)!=0){J=oy(k[t>>2]|0,f|0,e|0)|0;if(!J)Qa(t|0,x|0);L=x}else J=-1;if((J|0)==1){z=q;s=L;q=z;continue a}break}}if((i[739441]|0)==33){if((q|0)>0){t=0;do{p=t;t=t+1|0;i[739441+p>>0]=i[739441+t>>0]|0}while((t|0)!=(q|0))}w=0;sa(10,739441)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){K=oy(k[t>>2]|0,f|0,e|0)|0;if(!K)Qa(t|0,x|0);L=x}else K=-1;if((K|0)==1){z=q;s=L;q=z;continue a}}w=0;t=va(10,739441,c|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){M=oy(k[p>>2]|0,f|0,e|0)|0;if(!M)Qa(p|0,x|0);L=x}else M=-1;if((M|0)==1){z=q;s=L;q=z;continue a}if(t){w=0;qa(2,742294);t=w;w=0;if((t|0)!=0&(x|0)!=0){N=oy(k[t>>2]|0,f|0,e|0)|0;if(!N)Qa(t|0,x|0);L=x}else N=-1;if((N|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,739441);t=w;w=0;if((t|0)!=0&(x|0)!=0){O=oy(k[t>>2]|0,f|0,e|0)|0;if(!O)Qa(t|0,x|0);L=x}else O=-1;if((O|0)==1){z=q;s=L;q=z;continue a}break}k[29964]=(k[29964]|0)+1;w=0;sa(26,180)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){P=oy(k[t>>2]|0,f|0,e|0)|0;if(!P)Qa(t|0,x|0);L=x}else P=-1;if((P|0)==1){z=q;s=L;q=z;continue a}w=0;t=ua(2,j|0,1,1200,v|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Q=oy(k[p>>2]|0,f|0,e|0)|0;if(!Q)Qa(p|0,x|0);L=x}else Q=-1;if((Q|0)==1){z=q;s=L;q=z;continue a}c:do if(t){p=t;while(1){w=0;sa(26,0)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){R=oy(k[o>>2]|0,f|0,e|0)|0;if(!R)Qa(o|0,x|0);L=x}else R=-1;if((R|0)==1){z=q;s=L;q=z;continue a}w=0;o=na(13,k[c>>2]|0,j|0,p|0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){S=oy(k[b>>2]|0,f|0,e|0)|0;if(!S)Qa(b|0,x|0);L=x}else S=-1;if((S|0)==1){z=q;s=L;q=z;continue a}if(o)break;w=0;sa(26,180)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){T=oy(k[o>>2]|0,f|0,e|0)|0;if(!T)Qa(o|0,x|0);L=x}else T=-1;if((T|0)==1){z=q;s=L;q=z;continue a}w=0;p=ua(2,j|0,1,1200,v|0)|0;o=w;w=0;if((o|0)!=0&(x|0)!=0){U=oy(k[o>>2]|0,f|0,e|0)|0;if(!U)Qa(o|0,x|0);L=x}else U=-1;if((U|0)==1){z=q;s=L;q=z;continue a}if(!p)break c}w=0;qa(2,742343);p=w;w=0;if((p|0)!=0&(x|0)!=0){V=oy(k[p>>2]|0,f|0,e|0)|0;if(!V)Qa(p|0,x|0);L=x}else V=-1;if((V|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,a|0);p=w;w=0;if((p|0)!=0&(x|0)!=0){W=oy(k[p>>2]|0,f|0,e|0)|0;if(!W)Qa(p|0,x|0);L=x}else W=-1;if((W|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,739441);p=w;w=0;if((p|0)!=0&(x|0)!=0){X=oy(k[p>>2]|0,f|0,e|0)|0;if(!X)Qa(p|0,x|0);L=x}else X=-1;if((X|0)==1){z=q;s=L;q=z;continue a}break b}while(0);w=0;sa(9,k[c>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){Y=oy(k[t>>2]|0,f|0,e|0)|0;if(!Y)Qa(t|0,x|0);L=x}else Y=-1;if((Y|0)==1){z=q;s=L;q=z;continue a}k[29964]=(k[29964]|0)+-1;w=0;sa(4,v|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){Z=oy(k[t>>2]|0,f|0,e|0)|0;if(!Z)Qa(t|0,x|0);L=x}else Z=-1;if((Z|0)==1){z=q;s=L;q=z;continue a}k[29971]=(k[29971]|0)+-1;t=k[l>>2]|0;d:do if((t|0)>=0){p=0;while(1){w=0;o=ua(6,t|0,741866+p|0,5-p|0,0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){_=oy(k[b>>2]|0,f|0,e|0)|0;if(!_)Qa(b|0,x|0);L=x}else _=-1;if((_|0)==1){z=q;s=L;q=z;continue a}if((o|0)<1)break d;p=o+p|0;if((p|0)>=5)break}}while(0);w=0;sa(4,k[h>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){$=oy(k[t>>2]|0,f|0,e|0)|0;if(!$)Qa(t|0,x|0);L=x}else $=-1;if(($|0)==1){z=q;s=L;q=z;continue a}k[29970]=(k[29970]|0)+-1;w=0;t=va(7,739441,774853)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){aa=oy(k[v>>2]|0,f|0,e|0)|0;if(!aa)Qa(v|0,x|0);L=x}else aa=-1;if((aa|0)==1){z=q;s=L;q=z;continue a}k[29966]=t;if(!t){w=0;qa(2,742381);t=w;w=0;if((t|0)!=0&(x|0)!=0){ba=oy(k[t>>2]|0,f|0,e|0)|0;if(!ba)Qa(t|0,x|0);L=x}else ba=-1;if((ba|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,739441);t=w;w=0;if((t|0)!=0&(x|0)!=0){ca=oy(k[t>>2]|0,f|0,e|0)|0;if(!ca)Qa(t|0,x|0);L=x}else ca=-1;if((ca|0)==1){z=q;s=L;q=z;continue a}else{da=61;break a}}k[29965]=(k[29965]|0)+1;w=0;t=va(12,a|0,c|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){ea=oy(k[v>>2]|0,f|0,e|0)|0;if(!ea)Qa(v|0,x|0);L=x}else ea=-1;if((ea|0)==1){z=q;s=L;q=z;continue a}if(t){w=0;qa(2,742428);t=w;w=0;if((t|0)!=0&(x|0)!=0){fa=oy(k[t>>2]|0,f|0,e|0)|0;if(!fa)Qa(t|0,x|0);L=x}else fa=-1;if((fa|0)==1){z=q;s=L;q=z;continue a}w=0;qa(2,a|0);t=w;w=0;if((t|0)!=0&(x|0)!=0){ga=oy(k[t>>2]|0,f|0,e|0)|0;if(!ga)Qa(t|0,x|0);L=x}else ga=-1;if((ga|0)==1){z=q;s=L;q=z;continue a}break}k[29924]=(k[29924]|0)+1;w=0;t=na(41,a|0,k[29966]|0,k[c>>2]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){ha=oy(k[v>>2]|0,f|0,e|0)|0;if(!ha)Qa(v|0,x|0);L=x}else ha=-1;if((ha|0)==1){z=q;s=L;q=z;continue a}w=0;sa(4,k[29966]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){ia=oy(k[v>>2]|0,f|0,e|0)|0;if(!ia)Qa(v|0,x|0);L=x}else ia=-1;if((ia|0)==1){z=q;s=L;q=z;continue a}k[29965]=(k[29965]|0)+-1;if(t){w=0;qa(2,742477);t=w;w=0;if((t|0)!=0&(x|0)!=0){ja=oy(k[t>>2]|0,f|0,e|0)|0;if(!ja)Qa(t|0,x|0);L=x}else ja=-1;if((ja|0)==1){z=q;s=L;q=z;continue a}break}w=0;va(20,14,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ka=oy(k[t>>2]|0,f|0,e|0)|0;if(!ka)Qa(t|0,x|0);L=x}else ka=-1;if((ka|0)==1){z=q;s=L;q=z;continue a}w=0;sa(26,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){la=oy(k[t>>2]|0,f|0,e|0)|0;if(!la)Qa(t|0,x|0);L=x}else la=-1;if((la|0)==1){z=q;s=L;q=z;continue a}w=0;ma=na(15,k[c>>2]|0,0,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){oa=oy(k[t>>2]|0,f|0,e|0)|0;if(!oa)Qa(t|0,x|0);L=x}else oa=-1;if((oa|0)==1){z=q;s=L;q=z;continue a}else{da=77;break a}}else{w=0;qa(2,739091);t=w;w=0;if((t|0)!=0&(x|0)!=0){pa=oy(k[t>>2]|0,f|0,e|0)|0;if(!pa)Qa(t|0,x|0);L=x}else pa=-1;if((pa|0)==1){z=q;s=L;q=z;continue a}}while(0);w=0;sa(26,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ra=oy(k[t>>2]|0,f|0,e|0)|0;if(!ra)Qa(t|0,x|0);L=x}else ra=-1;if((ra|0)==1){z=q;s=L;q=z;continue}if(k[29971]|0){w=0;sa(4,k[g>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ta=oy(k[t>>2]|0,f|0,e|0)|0;if(!ta)Qa(t|0,x|0);L=x}else ta=-1;if((ta|0)==1){z=q;s=L;q=z;continue}}if(k[29970]|0){w=0;sa(4,k[h>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){wa=oy(k[t>>2]|0,f|0,e|0)|0;if(!wa)Qa(t|0,x|0);L=x}else wa=-1;if((wa|0)==1){z=q;s=L;q=z;continue}}if(k[29965]|0){w=0;sa(4,k[29966]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){xa=oy(k[t>>2]|0,f|0,e|0)|0;if(!xa)Qa(t|0,x|0);L=x}else xa=-1;if((xa|0)==1){z=q;s=L;q=z;continue}}if(k[29924]|0){w=0;sa(15,k[c>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ya=oy(k[t>>2]|0,f|0,e|0)|0;if(!ya)Qa(t|0,x|0);L=x}else ya=-1;if((ya|0)==1){z=q;s=L;q=z;continue}}if(k[29964]|0){w=0;sa(9,k[c>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){za=oy(k[t>>2]|0,f|0,e|0)|0;if(!za)Qa(t|0,x|0);L=x}else za=-1;if((za|0)==1){z=q;s=L;q=z;continue}}w=0;va(20,14,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){Aa=oy(k[t>>2]|0,f|0,e|0)|0;if(!Aa)Qa(t|0,x|0);L=x}else Aa=-1;if((Aa|0)==1){z=q;s=L;q=z}else{da=95;break}}if((da|0)==61){u=104;dy(f|0);r=d;return u|0}else if((da|0)==77){u=ma;dy(f|0);r=d;return u|0}else if((da|0)==95){u=104;dy(f|0);r=d;return u|0}return 0}function rj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;d=r;r=r+2416|0;e=d+1208|0;f=d+4|0;g=d+8|0;h=d;j=a;l=j;i[l>>0]=104;i[l+1>>0]=116;i[l+2>>0]=116;i[l+3>>0]=112;l=j+4|0;i[l>>0]=58;i[l+1>>0]=47;i[l+2>>0]=47;i[l+3>>0]=0;if(!(i[c>>0]|0)){r=d;return 0}if(!(Ax(c,764609,7)|0))ux(739441,c+7|0)|0;else ux(739441,c)|0;if(!(Ax(c,741186,4)|0)){m=a;n=742585;o=m+11|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}if(rx(b,63)|0){m=a;n=765458;o=m+12|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}if(!(Hj(b,f,g,h)|0)){ww(k[f>>2]|0)|0;if((Ex(b,739114)|0)==0?(Ex(b,739111)|0)==0:0){m=a;n=765458;o=m+12|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){m=a;n=765458;o=m+12|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}m=a;n=765530;o=m+16|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}ux(e,b)|0;l=e+(xx(e)|0)|0;i[l>>0]=46;i[l+1>>0]=103;i[l+2>>0]=122;i[l+3>>0]=0;if(!(Hj(e,f,g,h)|0)){ww(k[f>>2]|0)|0;ux(b,e)|0;if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){m=a;n=765458;o=m+12|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}m=a;n=765530;o=m+16|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}ux(e,b)|0;l=e+(xx(e)|0)|0;i[l>>0]=i[739111]|0;i[l+1>>0]=i[739112]|0;i[l+2>>0]=i[739113]|0;if(Hj(e,f,g,h)|0){r=d;return 0}ww(k[f>>2]|0)|0;ux(b,e)|0;if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){m=a;n=765458;o=m+12|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}m=a;n=765530;o=m+16|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));r=d;return 0}function sj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0;d=r;r=r+1216|0;e=d+16|0;f=d+8|0;g=d+4|0;h=d;i[a>>0]=i[765481]|0;i[a+1>>0]=i[765482]|0;i[a+2>>0]=i[765483]|0;i[a+3>>0]=i[765484]|0;i[a+4>>0]=i[765485]|0;i[a+5>>0]=i[765486]|0;i[a+6>>0]=i[765487]|0;if(!(i[c>>0]|0)){r=d;return 0}if(!(Ax(c,764609,7)|0))ux(739441,c+7|0)|0;else ux(739441,c)|0;if(!(Ax(c,741186,4)|0)){j=a;l=742596;m=j+10|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}if(!(Ij(b,f,g,h)|0)){ww(k[f>>2]|0)|0;ww(k[g>>2]|0)|0;if((Ex(b,739114)|0)==0?(Ex(b,739111)|0)==0:0){j=a;l=765470;m=j+11|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){j=a;l=765470;m=j+11|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}j=a;l=765546;m=j+15|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}ux(e,b)|0;n=e+(xx(e)|0)|0;i[n>>0]=46;i[n+1>>0]=103;i[n+2>>0]=122;i[n+3>>0]=0;if(!(Ij(e,f,g,h)|0)){ww(k[f>>2]|0)|0;ww(k[g>>2]|0)|0;ux(b,e)|0;if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){j=a;l=765470;m=j+11|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}j=a;l=765546;m=j+15|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}ux(e,b)|0;n=e+(xx(e)|0)|0;i[n>>0]=i[739111]|0;i[n+1>>0]=i[739112]|0;i[n+2>>0]=i[739113]|0;if(Ij(e,f,g,h)|0){r=d;return 0}ww(k[f>>2]|0)|0;ww(k[g>>2]|0)|0;ux(b,e)|0;if((Ex(c,739114)|0)==0?(Ex(c,739111)|0)==0:0){j=a;l=765470;m=j+11|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}j=a;l=765546;m=j+15|0;do{i[j>>0]=i[l>>0]|0;j=j+1|0;l=l+1|0}while((j|0)<(m|0));r=d;return 0}function tj(){var a=0,b=0;a=0;do{k[82840+(a<<4)>>2]=0;b=82840+(a<<4)+8|0;k[b>>2]=0;k[b+4>>2]=0;a=a+1|0}while((a|0)!=1e3);return 0}function uj(a){a=a|0;return 0}function vj(a){a=a|0;k[a>>2]=0;return 0}function wj(a){a=a|0;k[a>>2]=10;return 0}function xj(){return 0}function yj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0;d=r;r=r+16|0;e=d;k[c>>2]=-1;f=0;while(1){if(!(k[82840+(f<<4)>>2]|0)){g=f;break}f=f+1|0;if((f|0)>=1e3){h=103;i=10;break}}if((i|0)==10){r=d;return h|0}k[c>>2]=g;if((g|0)==-1){h=103;r=d;return h|0}if(!b)j=Jj(a,742613,e)|0;else j=Jj(a,742606,e)|0;if(j){h=j;r=d;return h|0}k[82840+(g<<4)>>2]=k[e>>2];e=82840+(g<<4)+8|0;k[e>>2]=0;k[e+4>>2]=0;h=0;r=d;return h|0}function zj(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0;c=r;r=r+16|0;d=c;k[b>>2]=-1;e=0;while(1){if(!(k[82840+(e<<4)>>2]|0)){f=e;break}e=e+1|0;if((e|0)>=1e3){g=103;h=8;break}}if((h|0)==8){r=c;return g|0}k[b>>2]=f;if((f|0)==-1){g=103;r=c;return g|0}b=Jj(a,742618,d)|0;if(!b){k[82840+(f<<4)>>2]=k[d>>2];d=82840+(f<<4)+8|0;k[d>>2]=0;k[d+4>>2]=0;g=0;r=c;return g|0}else{tk(742625);g=b;r=c;return g|0}return 0}function Aj(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0;c=r;r=r+16|0;d=c+8|0;e=c+4|0;f=c;g=k[82840+(a<<4)>>2]|0;a=Xv(4)|0;k[d>>2]=a;a=Xv(2009)|0;k[d+4>>2]=a;a:do if((g|0)>=0){a=0;do{h=bw(g,d+a|0,8-a|0,0)|0;if((h|0)<1)break a;a=h+a|0}while((a|0)<8)}while(0);Kj(g,f,e)|0;f=Zv(k[e>>2]|0)|0;e=b;k[e>>2]=f;k[e+4>>2]=0;r=c;return 0}function Bj(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=r;r=r+16|0;c=b;d=82840+(a<<4)|0;a=k[d>>2]|0;e=Xv(4)|0;k[c>>2]=e;e=Xv(2008)|0;k[c+4>>2]=e;a:do if((a|0)>=0){e=0;do{f=bw(a,c+e|0,8-e|0,0)|0;if((f|0)<1)break a;e=f+e|0}while((e|0)<8)}while(0);Hx(a)|0;k[d>>2]=0;r=b;return 0}function Cj(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=r;r=r+16|0;c=b;d=k[82840+(a<<4)>>2]|0;a=Xv(4)|0;k[c>>2]=a;a=Xv(2007)|0;k[c+4>>2]=a;if((d|0)<0){r=b;return 0}else e=0;while(1){a=bw(d,c+e|0,8-e|0,0)|0;if((a|0)<1){f=4;break}e=a+e|0;if((e|0)>=8){f=4;break}}if((f|0)==4){r=b;return 0}return 0}function Dj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;d=82840+(a<<4)+8|0;k[d>>2]=b;k[d+4>>2]=c;return 0}function Ej(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;d=r;r=r+128|0;e=d;f=d+16|0;g=d+12|0;h=d+8|0;i=82840+(a<<4)+8|0;k[e>>2]=k[i>>2];k[e+4>>2]=c;Xw(f,742647,e)|0;j=82840+(a<<4)|0;a=k[j>>2]|0;l=xx(f)|0;m=Xv(l+4|0)|0;k[e>>2]=m;m=Xv(2006)|0;k[e+4>>2]=m;a:do if((a|0)>=0){m=0;do{n=bw(a,e+m|0,8-m|0,0)|0;if((n|0)<1){o=n;p=4;break}m=n+m|0}while((m|0)<8);if((p|0)==4?(o|a|0)<0:0){q=(o|0)>-1?-1:o;break}if((l|0)>0){m=0;while(1){n=bw(a,f+m|0,l-m|0,0)|0;if((n|0)<1){q=n;break a}s=n+m|0;if((s|0)<(l|0))m=s;else{q=s;break}}}else q=0}else q=-1;while(0);if((q|0)!=(xx(f)|0)){t=108;r=d;return t|0}k[h>>2]=0;Kj(k[j>>2]|0,g,h)|0;if(k[h>>2]|0){t=108;r=d;return t|0}h=k[j>>2]|0;b:do if((h|0)>=0)if((c|0)>0){j=0;while(1){g=b+j|0;f=c-j|0;q=$v(h,g,f,0)|0;if((q|0)==-1){l=rv()|0;while(1){if((k[l>>2]|0)!=4){u=-1;break b}k[l>>2]=0;a=$v(h,g,f,0)|0;if((a|0)!=-1){v=a;break}}}else v=q;if((v|0)<0){u=v;break b}if(!v){u=j;break b}f=v+j|0;if((f|0)<(c|0))j=f;else{u=f;break}}}else u=0;else u=-1;while(0);if((u|0)!=(c|0)){t=108;r=d;return t|0}u=i;v=ky(k[u>>2]|0,k[u+4>>2]|0,c|0,((c|0)<0)<<31>>31|0)|0;c=i;k[c>>2]=v;k[c+4>>2]=L;t=0;r=d;return t|0}function Fj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;d=r;r=r+128|0;e=d;f=d+16|0;g=d+12|0;h=d+8|0;i=82840+(a<<4)|0;j=k[i>>2]|0;l=82840+(a<<4)+8|0;k[e>>2]=k[l>>2];k[e+4>>2]=c;Xw(f,742647,e)|0;a=xx(f)|0;m=a+1|0;n=Xv(a+5|0)|0;k[e>>2]=n;n=Xv(2005)|0;k[e+4>>2]=n;n=(j|0)<0;a:do if(!n){o=0;do{p=bw(j,e+o|0,8-o|0,0)|0;if((p|0)<1){q=p;s=4;break}o=p+o|0}while((o|0)<8);if((s|0)==4?(q|j|0)<0:0){t=(q|0)>-1?-1:q;break}if((a|0)>-1){o=0;while(1){p=bw(j,f+o|0,m-o|0,0)|0;if((p|0)<1){t=p;break a}u=p+o|0;if((u|0)<(m|0))o=u;else{t=u;break}}}else t=0}else t=-1;while(0);if((t|0)!=(m|0)){v=106;r=d;return v|0}b:do if(!n)if((c|0)>0){m=0;while(1){t=bw(j,b+m|0,c-m|0,0)|0;if((t|0)<1){w=t;break b}f=t+m|0;if((f|0)<(c|0))m=f;else{w=f;break}}}else w=0;else w=-1;while(0);if((w|0)!=(c|0)){v=106;r=d;return v|0}k[g>>2]=0;Kj(k[i>>2]|0,h,g)|0;if(k[g>>2]|0){v=106;r=d;return v|0}g=l;h=ky(k[g>>2]|0,k[g+4>>2]|0,c|0,((c|0)<0)<<31>>31|0)|0;c=l;k[c>>2]=h;k[c+4>>2]=L;v=0;r=d;return v|0}function Gj(a){a=a|0;if((a|0)==14)Qa(119700,14);else Ib(a|0)}function Hj(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0.0,X=0,Y=0;e=r;r=r+10224|0;f=e+1288|0;g=e+1280|0;h=e+1272|0;j=e+1264|0;m=e+1256|0;n=e+1248|0;o=e+1240|0;q=e+1232|0;s=e+1216|0;t=e+1208|0;u=e+1200|0;v=e+1312|0;w=e+1308|0;x=e+9024|0;y=e+7824|0;z=e+7720|0;A=e+6520|0;B=e+5320|0;C=e+5120|0;D=e+3920|0;E=e+2720|0;F=e;G=e+1304|0;H=e+1300|0;I=e+2520|0;J=e+1296|0;K=F;k[K>>2]=1886680168;k[K+4>>2]=3092282;zx(F,a,1192)|0;if(Lj(F,e+5220|0,C,G,E)|0){k[u>>2]=a;Xw(B,743668,u)|0;tk(B);L=104;r=e;return L|0}ux(D,a)|0;a=rx(D,64)|0;if(!a)i[D>>0]=0;else i[a>>0]=0;a=jb(743699)|0;u=(a|0)!=0;do if(u){if(!(Lj(a,e+2620|0,I,J,e+1320|0)|0)){M=Mj(I,k[J>>2]|0)|0;break}k[t>>2]=a;Xw(B,743668,t)|0;tk(B);L=104;r=e;return L|0}else M=Mj(C,k[G>>2]|0)|0;while(0);if((M|0)<0){if(!u){L=104;r=e;return L|0}tk(743710);tk(a);L=104;r=e;return L|0}a=iw(M,774853)|0;k[b>>2]=a;if(!a){tk(743772);Hx(M)|0;L=104;r=e;return L|0}if(u){u=k[G>>2]|0;k[s>>2]=C;k[s+4>>2]=u;k[s+8>>2]=E;Xw(y,743832,s)|0}else{k[q>>2]=E;Xw(y,743863,q)|0}do if(i[D>>0]|0){q=xx(D)|0;a:do if(!q){N=A;O=30}else{E=1200;s=A;u=0-q|0;a=0;t=0;while(1){J=u>>>0>4294967293?0-u|0:3;I=t;K=0;P=a;while(1){Q=l[D+I>>0]|P<<8;R=K+1|0;if((R|0)==(J|0)){S=Q;break}else{I=R+t|0;K=R;P=Q}}P=S<<(((6-((J<<3>>>0)%6|0)|0)>>>0)%6|0);if(E>>>0<4)break a;K=s+2|0;I=s+1|0;i[s>>0]=61;i[s+1>>0]=61;i[s+2>>0]=61;i[s+3>>0]=61;switch(J|0){case 3:{i[s+3>>0]=i[743881+(P&63)>>0]|0;T=P>>>6;O=27;break}case 2:{T=P;O=27;break}case 1:{U=P;O=28;break}default:V=P}if((O|0)==27){O=0;i[K>>0]=i[743881+(T&63)>>0]|0;U=T>>>6;O=28}if((O|0)==28){O=0;i[I>>0]=i[743881+(U&63)>>0]|0;I=U>>>6;i[s>>0]=i[743881+(I&63)>>0]|0;V=I}I=s+4|0;t=t+3|0;if(t>>>0>=q>>>0){N=I;O=30;break a}else{E=E+-4|0;s=I;u=u+3|0;a=V}}}while(0);if((O|0)==30)i[N>>0]=0;k[o>>2]=A;Xw(z,743946,o)|0;q=xx(y)|0;if(((xx(z)|0)+q|0)>>>0>1199){L=104;r=e;return L|0}else{qx(y,z)|0;break}}while(0);W=+pk(H);p[n>>3]=W;Xw(z,743972,n)|0;n=xx(y)|0;if(((xx(z)|0)+n|0)>>>0>1199){L=104;r=e;return L|0}qx(y,z)|0;n=k[G>>2]|0;k[m>>2]=C;k[m+4>>2]=n;Xw(z,744009,m)|0;m=xx(y)|0;if(((xx(z)|0)+m|0)>>>0>1199){L=104;r=e;return L|0}qx(y,z)|0;z=xx(y)|0;b:do if((z|0)>0){m=0;while(1){n=bw(M,y+m|0,z-m|0,0)|0;if((n|0)<1){X=n;break b}C=n+m|0;if((C|0)<(z|0))m=C;else{X=C;break}}}else X=0;while(0);k[v>>2]=X;if(!(Bw(x,1200,k[b>>2]|0)|0)){k[j>>2]=x;Xw(B,744026,j)|0;tk(B);ww(k[b>>2]|0)|0;L=104;r=e;return L|0}k[d>>2]=0;i[c>>0]=0;k[h>>2]=y;k[h+4>>2]=v;Yw(x,744067,h)|0;h=k[v>>2]|0;if((h|0)==200){v=(Bw(x,1200,k[b>>2]|0)|0)==0;if(v|(i[x>>0]|0)==13){L=0;r=e;return L|0}do{v=xx(x)|0;if(v>>>0>3){i[x+(v+-1)>>0]=0;v=x+((xx(x)|0)+-1)|0;i[v>>0]=0}k[f>>2]=y;k[f+4>>2]=w;Yw(x,744067,f)|0;if(!(tx(y,744264)|0))k[d>>2]=k[w>>2];if((tx(y,744280)|0)==0?(v=rx(x,58)|0,(v|0)!=0):0)ux(c,v+2|0)|0;v=(Bw(x,1200,k[b>>2]|0)|0)==0}while(!(v|(i[x>>0]|0)==13));L=0;r=e;return L|0}do if((h+-301|0)>>>0<2){if((h|0)==301)tk(744073);else tk(744129);tk(F);while(1){if(!(Bw(x,1200,k[b>>2]|0)|0)){O=50;break}y=Ex(x,744175)|0;if(y){Y=y;break}}if((O|0)==50){tk(744194);break}ux(F,Fx(Y+16|0,744185)|0)|0;k[g>>2]=F;Xw(B,744187,g)|0;tk(B);ww(k[b>>2]|0)|0;L=Hj(F,b,c,d)|0;r=e;return L|0}while(0);ww(k[b>>2]|0)|0;L=104;r=e;return L|0}function Ij(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;e=r;r=r+7584|0;f=e+1256|0;g=e+1248|0;h=e+1240|0;j=e+1232|0;l=e+1224|0;m=e+1216|0;n=e+1208|0;o=e+1200|0;p=e+6376|0;q=e+1264|0;s=e+5176|0;t=e+3976|0;u=e;v=e+3776|0;w=e+2576|0;x=e+2472|0;y=e+1272|0;z=e+1260|0;if((xx(a)|0)>>>0>1193){tk(743148);A=104;r=e;return A|0};i[y>>0]=i[765481]|0;i[y+1>>0]=i[765482]|0;i[y+2>>0]=i[765483]|0;i[y+3>>0]=i[765484]|0;i[y+4>>0]=i[765485]|0;i[y+5>>0]=i[765486]|0;i[y+6>>0]=i[765487]|0;qx(y,a)|0;if(Lj(y,e+3876|0,v,z,w)|0){k[o>>2]=a;Xw(t,743184,o)|0;tk(t);A=104;r=e;return A|0}k[z>>2]=21;t=Cx(v,64)|0;if(t){i[t>>0]=0;o=t+1|0;t=rx(v,58)|0;if(!t){B=o;C=743224;D=v}else{i[t>>0]=0;B=o;C=t+1|0;D=v}}else{B=v;C=743224;D=743214}v=Mj(B,k[z>>2]|0)|0;k[d>>2]=v;B=iw(v,774853)|0;k[c>>2]=B;if(!B){tk(743238);A=104;r=e;return A|0}v=p+4|0;while(1){if(!(Bw(p,1200,B)|0)){E=14;break}i[v>>0]=0;if(!(tx(p,743295)|0))break;if((i[p>>0]|0)>51){E=14;break}}if((E|0)==14){tk(743300);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}k[n>>2]=D;Xw(u,743358,n)|0;n=k[d>>2]|0;D=xx(u)|0;a:do if((n|0)>-1&(D|0)>0){B=0;do{t=bw(n,u+B|0,D-B|0,0)|0;if((t|0)<1)break a;B=t+B|0}while((B|0)<(D|0))}while(0);D=k[c>>2]|0;while(1){if(!(Bw(p,1200,D)|0)){E=22;break}i[v>>0]=0;if(!(tx(p,743367)|0))break;if((i[p>>0]|0)>51){E=22;break}}if((E|0)==22){tk(743372);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}k[m>>2]=C;Xw(u,743406,m)|0;m=k[d>>2]|0;C=xx(u)|0;b:do if((m|0)>-1&(C|0)>0){D=0;do{n=bw(m,u+D|0,C-D|0,0)|0;if((n|0)<1)break b;D=n+D|0}while((D|0)<(C|0))}while(0);C=k[c>>2]|0;while(1){if(!(Bw(p,1200,C)|0)){E=30;break}i[v>>0]=0;if(!(tx(p,743415)|0))break;if((i[p>>0]|0)>51){E=30;break}}if((E|0)==30){tk(743420);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}C=Cx(w,47)|0;c:do if(!C){i[u>>0]=i[743455]|0;i[u+1>>0]=i[743456]|0;i[u+2>>0]=i[743457]|0;i[u+3>>0]=i[743458]|0;i[u+4>>0]=i[743459]|0;i[u+5>>0]=i[743460]|0;i[u+6>>0]=i[743461]|0;F=w}else{i[C>>0]=0;m=C+1|0;switch(i[w>>0]|0){case 0:{i[u>>0]=i[743455]|0;i[u+1>>0]=i[743456]|0;i[u+2>>0]=i[743457]|0;i[u+3>>0]=i[743458]|0;i[u+4>>0]=i[743459]|0;i[u+5>>0]=i[743460]|0;i[u+6>>0]=i[743461]|0;F=m;break c;break}case 47:{k[l>>2]=w+1;Xw(u,743462,l)|0;F=m;break c;break}default:{k[j>>2]=w;Xw(u,743462,j)|0;F=m;break c}}}while(0);j=k[d>>2]|0;w=xx(u)|0;d:do if((j|0)>-1&(w|0)>0){l=0;do{C=bw(j,u+l|0,w-l|0,0)|0;if((C|0)<1)break d;l=C+l|0}while((l|0)<(w|0))}while(0);w=k[c>>2]|0;while(1){if(!(Bw(p,1200,w)|0)){E=44;break}i[v>>0]=0;if(!(tx(p,743470)|0))break;if((i[p>>0]|0)>51){E=44;break}}if((E|0)==44){tk(743475);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}if(!(i[F>>0]|0)){tk(743509);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}w=u;k[w>>2]=1162893652;k[w+4>>2]=674080;w=k[d>>2]|0;j=xx(u)|0;e:do if((w|0)>-1&(j|0)>0){l=0;do{C=bw(w,u+l|0,j-l|0,0)|0;if((C|0)<1)break e;l=C+l|0}while((l|0)<(j|0))}while(0);j=k[c>>2]|0;while(1){if(!(Bw(p,1200,j)|0)){E=54;break}i[v>>0]=0;if(!(tx(p,743535)|0))break;if((i[p>>0]|0)>51){E=54;break}}if((E|0)==54){tk(743540);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}j=k[d>>2]|0;f:do if((j|0)>=0){w=0;do{l=bw(j,743578+w|0,5-w|0,0)|0;if((l|0)<1)break f;w=l+w|0}while((w|0)<5)}while(0);if(!(Bw(s,1200,k[c>>2]|0)|0)){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}if(((i[s>>0]|0)==50?(i[s+1>>0]|0)==50:0)?(i[s+2>>0]|0)==55:0){j=rx(s,40)|0;if(!j){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}i[j>>0]=0;i[x>>0]=0;s=Fx(j+1|0,743606)|0;if(!s){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}ux(x,s)|0;s=x+(xx(x)|0)|0;i[s>>0]=46;i[s+1>>0]=0;s=Fx(0,743606)|0;if(!s){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}qx(x,s)|0;s=x+(xx(x)|0)|0;i[s>>0]=46;i[s+1>>0]=0;s=Fx(0,743606)|0;if(!s){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}qx(x,s)|0;s=x+(xx(x)|0)|0;i[s>>0]=46;i[s+1>>0]=0;s=Fx(0,743606)|0;if(!s){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}qx(x,s)|0;s=Fx(0,743606)|0;if(!s){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}k[h>>2]=z;Yw(s,772183,h)|0;k[z>>2]=k[z>>2]<<8;h=Fx(0,743606)|0;if(!h){tk(743584);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}k[g>>2]=q;Yw(h,772183,g)|0;g=(k[z>>2]|0)+(k[q>>2]|0)|0;k[z>>2]=g;if(!(i[F>>0]|0)){tk(743509);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}z=iw(Mj(x,g)|0,774853)|0;k[b>>2]=z;if(!z){tk(743609);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}k[f>>2]=F;Xw(u,743654,f)|0;f=k[d>>2]|0;F=xx(u)|0;g:do if((f|0)>-1&(F|0)>0){z=0;do{b=bw(f,u+z|0,F-z|0,0)|0;if((b|0)<1)break g;z=b+z|0}while((z|0)<(F|0))}while(0);F=k[c>>2]|0;while(1){if(!(Bw(p,1200,F)|0)){E=89;break}i[v>>0]=0;if(!(tx(p,743663)|0)){E=87;break}if((i[p>>0]|0)>51){E=89;break}}if((E|0)==87){A=0;r=e;return A|0}else if((E|0)==89){ww(k[c>>2]|0)|0;A=104;r=e;return A|0}}E=k[d>>2]|0;h:do if((E|0)>=0){d=0;do{p=bw(E,741866+d|0,5-d|0,0)|0;if((p|0)<1)break h;d=p+d|0}while((d|0)<5)}while(0);ww(k[c>>2]|0)|0;A=104;r=e;return A|0}function Jj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0;d=r;r=r+5040|0;e=d+1216|0;f=d+1200|0;g=d+1232|0;h=d+3840|0;j=d+2640|0;m=d+2440|0;n=d+1240|0;o=d;p=d+1228|0;q=d+1224|0;s=d+1220|0;t=o;k[t>>2]=1953460082;k[t+4>>2]=3092282;qx(o,a)|0;if(Lj(o,d+2540|0,m,p,n)|0){k[f>>2]=a;Xw(j,742656,f)|0;tk(j);u=104;r=d;return u|0}j=Mj(m,k[p>>2]|0)|0;k[c>>2]=j;if((j|0)<0){tk(742687);u=104;r=d;return u|0}j=jb(742732)|0;if(!j){Rw(742745,d+1208|0)|0;Bw(h,1200,k[31230]|0)|0;p=h+((xx(h)|0)+-1)|0;i[p>>0]=0}else ux(h,j)|0;j=k[c>>2]|0;p=xx(h)|0;m=Xv(p+4|0)|0;k[g>>2]=m;m=Xv(2e3)|0;f=g+4|0;k[f>>2]=m;do if((j|0)<0)v=16;else{m=0;do{a=bw(j,g+m|0,8-m|0,0)|0;if((a|0)<1){w=a;v=11;break}m=a+m|0}while((m|0)<8);if((v|0)==11?(w|j|0)<0:0){v=16;break}if((p|0)>0){m=0;while(1){a=bw(j,h+m|0,p-m|0,0)|0;if((a|0)<1){x=a;break}o=a+m|0;if((o|0)<(p|0))m=o;else{x=o;break}}if((x|0)<0)break}if(!(Kj(k[c>>2]|0,q,s)|0)){tk(742800);u=104;r=d;return u|0}if((k[q>>2]|0)!=2002){tk(742843);tk(h);u=104;r=d;return u|0}m=jb(742863)|0;if(!m){Rw(742876,e)|0;Bw(h,1200,k[31230]|0)|0;o=h+((xx(h)|0)+-1)|0;i[o>>0]=0}else ux(h,m)|0;if(!(i[h>>0]|0))y=0;else{m=0;do{o=h+m|0;i[o>>0]=l[o>>0]^255;m=m+1|0;o=xx(h)|0}while(m>>>0<o>>>0);y=o}m=k[c>>2]|0;o=Xv(y+4|0)|0;k[g>>2]=o;o=Xv(2001)|0;k[f>>2]=o;do if((m|0)<0)v=36;else{o=0;do{a=bw(m,g+o|0,8-o|0,0)|0;if((a|0)<1){z=a;v=31;break}o=a+o|0}while((o|0)<8);if((v|0)==31?(z|m|0)<0:0){v=36;break}if((y|0)>0){o=0;while(1){a=bw(m,h+o|0,y-o|0,0)|0;if((a|0)<1){A=a;break}t=a+o|0;if((t|0)<(y|0))o=t;else{A=t;break}}if((A|0)<0)break}if((Kj(k[c>>2]|0,q,s)|0)<0){tk(742935);u=104;r=d;return u|0}if((k[q>>2]|0)!=2002){tk(742982);tk(h);u=104;r=d;return u|0}ux(h,n)|0;o=h+(xx(h)|0)|0;i[o>>0]=32;i[o+1>>0]=0;qx(h,b)|0;o=k[c>>2]|0;t=xx(h)|0;a=Xv(t+4|0)|0;k[g>>2]=a;a=Xv(2004)|0;k[f>>2]=a;do if((o|0)<0)v=51;else{a=0;do{B=bw(o,g+a|0,8-a|0,0)|0;if((B|0)<1){C=B;v=46;break}a=B+a|0}while((a|0)<8);if((v|0)==46?(C|o|0)<0:0){v=51;break}if((t|0)>0){a=0;while(1){B=bw(o,h+a|0,t-a|0,0)|0;if((B|0)<1){D=B;break}E=B+a|0;if((E|0)<(t|0))a=E;else{D=E;break}}if((D|0)<0)break}if((Kj(k[c>>2]|0,q,s)|0)<0){tk(743042);u=104;r=d;return u|0}if(!((k[q>>2]|0)!=2004&(k[s>>2]|0)!=0)){u=0;r=d;return u|0}tk(743081);tk(h);u=104;r=d;return u|0}while(0);tk(743002);u=104;r=d;return u|0}while(0);tk(742887);u=104;r=d;return u|0}while(0);tk(742756);u=104;r=d;return u|0}function Kj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;d=r;r=r+1216|0;e=d;f=d+8|0;if((a|0)<0){g=-1;r=d;return g|0}else h=0;a:while(1){i=e+h|0;j=4-h|0;l=$v(a,i,j,0)|0;if((l|0)==-1){m=rv()|0;while(1){if((k[m>>2]|0)!=4){g=-1;n=34;break a}k[m>>2]=0;o=$v(a,i,j,0)|0;if((o|0)!=-1){p=o;break}}}else p=l;if((p|0)<0){g=p;n=34;break}if(!p){q=h;n=9;break}j=p+h|0;if((j|0)<4)h=j;else{s=j;break}}if((n|0)==9)if((q|0)<0){g=q;r=d;return g|0}else s=q;else if((n|0)==34){r=d;return g|0}q=(Zv(k[e>>2]|0)|0)+-4|0;k[e>>2]=q;q=0;b:while(1){h=b+q|0;p=4-q|0;j=$v(a,h,p,0)|0;if((j|0)==-1){i=rv()|0;while(1){if((k[i>>2]|0)!=4){g=-1;n=34;break b}k[i>>2]=0;m=$v(a,h,p,0)|0;if((m|0)!=-1){t=m;break}}}else t=j;if((t|0)<0){g=t;n=34;break}if(!t){u=q;n=18;break}p=t+q|0;if((p|0)<4)q=p;else{v=p;break}}if((n|0)==18)if((u|0)<0){g=u;r=d;return g|0}else v=u;else if((n|0)==34){r=d;return g|0}u=v+s|0;s=Zv(k[b>>2]|0)|0;k[b>>2]=s;s=k[e>>2]|0;if((s|0)<=1200)if((s|0)>0){w=s;n=22}else x=v;else{k[e>>2]=1200;w=1200;n=22}if((n|0)==22){n=0;c:while(1){v=f+n|0;s=w-n|0;b=$v(a,v,s,0)|0;if((b|0)==-1){q=rv()|0;while(1){if((k[q>>2]|0)!=4){y=-1;break c}k[q>>2]=0;t=$v(a,v,s,0)|0;if((t|0)!=-1){z=t;break}}}else z=b;if((z|0)<0){y=z;break}if(!z){y=n;break}s=z+n|0;if((s|0)<(w|0))n=s;else{y=s;break}}n=k[e>>2]|0;if((n|0)>4){k[e>>2]=4;A=4}else A=n;py(c|0,f|0,A|0)|0;if((y|0)<0){g=y;r=d;return g|0}else x=y}g=u+x|0;r=d;return g|0}function Lj(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;f=r;r=r+16|0;g=f;h=cy((xx(a)|0)+1|0)|0;ux(h,a)|0;k[d>>2]=80;i[b>>0]=i[743121]|0;i[b+1>>0]=i[743122]|0;i[b+2>>0]=i[743123]|0;i[b+3>>0]=i[743124]|0;i[b+4>>0]=i[743125]|0;i[b+5>>0]=i[743126]|0;a=c;j=743127;l=a+10|0;do{i[a>>0]=i[j>>0]|0;a=a+1|0;j=j+1|0}while((a|0)<(l|0));i[e>>0]=47;i[e+1>>0]=0;j=Ex(h,743121)|0;do if(!j){a=Ex(h,743137)|0;if(a){if((a|0)==(h|0)){m=h+5|0;break}dy(h);n=1;r=f;return n|0}a=Ex(h,743143)|0;if(!a)m=h;else{if((a|0)==(h|0)){i[b>>0]=i[743143]|0;i[b+1>>0]=i[743144]|0;i[b+2>>0]=i[743145]|0;i[b+3>>0]=i[743146]|0;i[b+4>>0]=i[743147]|0;k[d>>2]=21;m=h+4|0;break}dy(h);n=1;r=f;return n|0}}else{if((j|0)==(h|0)){m=h+5|0;break}dy(h);n=1;r=f;return n|0}while(0);if((i[m>>0]|0)==47)o=(i[m+1>>0]|0)==47?m+2|0:m;else o=m;a:do if(!(tx(b,743121)|0)){m=rx(o,64)|0;j=(m|0)==0?o:m+1|0;ux(c,j)|0;m=c;a=j;b:while(1){switch(i[a>>0]|0){case 0:case 58:case 47:{p=m;q=a;break b;break}default:{}}m=m+1|0;a=a+1|0}i[p>>0]=0;a=i[q>>0]|0;if(a<<24>>24==58){m=q+1|0;k[g>>2]=d;Yw(m,772183,g)|0;j=m;while(1){m=i[j>>0]|0;switch(m<<24>>24){case 0:case 47:{s=m;t=j;break a;break}default:{}}j=j+1|0}}else{s=a;t=q}}else{ux(c,o)|0;j=c;m=o;c:while(1){switch(i[m>>0]|0){case 0:case 47:{u=j;v=m;break c;break}default:{}}j=j+1|0;m=m+1|0}i[u>>0]=0;s=i[v>>0]|0;t=v}while(0);if(s<<24>>24)ux(e,t)|0;dy(h);n=0;r=f;return n|0}function Mj(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,l=0,m=0,n=0;c=r;r=r+1232|0;d=c+24|0;e=c+32|0;f=c+8|0;g=c;k[g>>2]=1;ux(e,a)|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j[f>>1]=2;h=Yv(b&65535)|0;j[f+2>>1]=h;do if(a){h=Cb(e|0)|0;k[d>>2]=h;if((h|0)==-1){b=wb(e|0)|0;if(!b){i=rv()|0;k[i>>2]=113;break}else{py(d|0,k[k[b+16>>2]>>2]|0,k[b+12>>2]|0)|0;l=k[d>>2]|0;m=7;break}}else{l=h;m=7}}else{k[d>>2]=0;l=0;m=7}while(0);if((m|0)==7)k[f+4>>2]=l;l=ew(2,1,0)|0;if((l|0)<0){tk(743101);n=-1002;r=c;return n|0}if((Wv(l,f,16)|0)<0){Hx(l)|0;n=-1002;r=c;return n|0}else{dw(l,6,1,g,4)|0;dw(l,1,9,g,4)|0;k[g>>2]=65536;dw(l,1,7,g,4)|0;dw(l,1,8,g,4)|0;n=l;r=c;return n|0}return 0}function Nj(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;g=r;r=r+16|0;h=g;i=k[f>>2]|0;if((i|0)>0){j=i;r=g;return j|0}i=k[a>>2]|0;l=a+4|0;m=k[l>>2]|0;if((i|0)==(k[m+68>>2]|0)){n=m+120|0;if(((k[n>>2]|0)==-1?(k[n+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){j=k[f>>2]|0;r=g;return j|0}}else Fk(a,i+1|0,0,f)|0;i=k[l>>2]|0;if(!(k[i+72>>2]|0)){tk(744298);k[f>>2]=235;j=235;r=g;return j|0}if((e|0)<0){k[f>>2]=306;j=306;r=g;return j|0}if((d|0)==0&(e|0)==0){j=k[f>>2]|0;r=g;return j|0}n=i+952|0;m=k[n>>2]|0;o=k[n+4>>2]|0;n=i+944|0;p=k[n>>2]|0;q=k[n+4>>2]|0;if((q|0)<(c|0)|(q|0)==(c|0)&p>>>0<b>>>0){tk(744356);k[f>>2]=307;j=307;r=g;return j|0}if((c|0)<0){tk(744426);k[f>>2]=307;j=307;r=g;return j|0}n=i+968|0;s=i+976|0;t=ky(k[s>>2]|0,k[s+4>>2]|0,k[n>>2]|0,k[n+4>>2]|0)|0;n=L;s=ky(t|0,n|0,2879,0)|0;u=L;v=xy(s|0,u|0,2880,0)|0;w=jy(s|0,u|0,v|0,L|0)|0;v=jy(w|0,L|0,t|0,n|0)|0;w=L;u=yy(m|0,o|0,d|0,e|0)|0;s=L;if((s|0)>(w|0)|(s|0)==(w|0)&u>>>0>v>>>0){x=ky(u|0,s|0,2879,0)|0;y=jy(x|0,L|0,v|0,w|0)|0;w=wy(y|0,L|0,2880,0)|0;rl(a,w,1,f)|0;z=k[l>>2]|0}else z=i;i=yy(m|0,o|0,b|0,c|0)|0;c=L;b=jy(t|0,n|0,i|0,c|0)|0;n=L;t=z+120|0;z=ky(k[t>>2]|0,k[t+4>>2]|0,i|0,c|0)|0;Oj(a,z,L,b,n,u,s,f)|0;n=(k[l>>2]|0)+968|0;b=n;z=ky(k[b>>2]|0,k[b+4>>2]|0,u|0,s|0)|0;s=L;u=n;k[u>>2]=z;k[u+4>>2]=s;k[h>>2]=0;oq(a,782755,z,s,763480,h)|0;h=ky(p|0,q|0,d|0,e|0)|0;oq(a,790243,h,L,763480,f)|0;h=k[l>>2]|0;l=h+944|0;a=l;q=ky(k[a>>2]|0,k[a+4>>2]|0,d|0,e|0)|0;a=l;k[a>>2]=q;k[a+4>>2]=L;a=h+936|0;h=a;q=ky(k[h>>2]|0,k[h+4>>2]|0,d|0,e|0)|0;e=a;k[e>>2]=q;k[e+4>>2]=L;j=k[f>>2]|0;r=g;return j|0}function Oj(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;i=r;r=r+1e5|0;j=i;l=k[h>>2]|0;if((l|0)>0){m=l;r=i;return m|0}l=(g|0)>0|(g|0)==0&f>>>0>0;n=ky(d|0,e|0,b|0,c|0)|0;o=L;p=l?n:b;q=l?o:c;a:do if(!((d|0)==0&(e|0)==0)){b:do if(l){s=e;t=d;u=p;v=q;while(1){w=(s|0)<0|(s|0)==0&t>>>0<1e5;x=my(0,t|0,32)|0;y=w?x:1e5;x=w?L:0;u=jy(u|0,v|0,y|0,x|0)|0;v=L;au(a,u,v,0,h)|0;fu(a,y,x,j,h)|0;w=ky(u|0,v|0,f|0,g|0)|0;au(a,w,L,1,h)|0;if((cu(a,y,x,j,h)|0)>0)break b;w=t;t=jy(t|0,s|0,y|0,x|0)|0;if((w|0)==(y|0)&(s|0)==(x|0))break a;else s=L}}else{if((g|0)<0){s=e;t=d;v=p;u=q;while(1){x=(s|0)<0|(s|0)==0&t>>>0<1e5;y=my(0,t|0,32)|0;w=x?y:1e5;y=x?L:0;au(a,v,u,0,h)|0;fu(a,w,y,j,h)|0;x=ky(v|0,u|0,f|0,g|0)|0;au(a,x,L,1,h)|0;if((cu(a,w,y,j,h)|0)>0)break b;x=t;t=jy(t|0,s|0,w|0,y|0)|0;z=s;s=L;v=ky(v|0,u|0,w|0,y|0)|0;if((x|0)==(w|0)&(z|0)==(y|0))break a;else u=L}}u=ky(p|0,q|0,f|0,g|0)|0;v=L;s=e;t=d;while(1){y=(s|0)<0|(s|0)==0&t>>>0<1e5;z=my(0,t|0,32)|0;w=y?z:1e5;z=y?L:0;au(a,p,q,0,h)|0;fu(a,w,z,j,h)|0;au(a,u,v,1,h)|0;if((cu(a,w,z,j,h)|0)>0)break b;y=t;t=jy(t|0,s|0,w|0,z|0)|0;if((y|0)==(w|0)&(s|0)==(z|0))break a;else s=L}}while(0);tk(744466);m=k[h>>2]|0;r=i;return m|0}while(0);if((k[(k[a+4>>2]|0)+72>>2]|0)==1)sy(j|0,32,1e5)|0;else sy(j|0,0,1e5)|0;q=(g|0)<0;p=jy(0,0,f|0,g|0)|0;d=L;e=ky(n|0,o|0,f|0,g|0)|0;o=q?p:f;f=q?d:g;au(a,q?e:b,q?L:c,0,h)|0;if(!((o|0)==0&(f|0)==0)){c=f;f=o;while(1){o=(c|0)<0|(c|0)==0&f>>>0<1e5;q=my(0,f|0,32)|0;b=o?q:1e5;q=o?L:0;cu(a,b,q,j,h)|0;o=f;f=jy(f|0,c|0,b|0,q|0)|0;if((o|0)==(b|0)&(c|0)==(q|0))break;else c=L}}m=k[h>>2]|0;r=i;return m|0}function Pj(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;g=r;r=r+96|0;h=g+8|0;i=g;j=g+12|0;l=k[f>>2]|0;if((l|0)>0){m=l;r=g;return m|0}l=k[a>>2]|0;n=a+4|0;o=k[n>>2]|0;if((l|0)==(k[o+68>>2]|0)){p=o+120|0;if(((k[p>>2]|0)==-1?(k[p+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){m=k[f>>2]|0;r=g;return m|0}}else Fk(a,l+1|0,0,f)|0;if(!(k[(k[n>>2]|0)+72>>2]|0)){tk(744502);k[f>>2]=235;m=235;r=g;return m|0}if((e|0)<0){k[f>>2]=306;m=306;r=g;return m|0}if((d|0)==0&(e|0)==0){m=k[f>>2]|0;r=g;return m|0}No(a,772833,i,j,f)|0;j=k[n>>2]|0;l=j+944|0;p=k[l>>2]|0;o=k[l+4>>2]|0;if((o|0)<(c|0)|(o|0)==(c|0)&p>>>0<b>>>0){tk(744563);k[f>>2]=307;m=307;r=g;return m|0}if((c|0)<0|(c|0)==0&b>>>0<1){tk(744633);k[f>>2]=307;m=307;r=g;return m|0}l=ky(b|0,c|0,-1,-1)|0;c=ky(l|0,L|0,d|0,e|0)|0;l=L;if((l|0)>(o|0)|(l|0)==(o|0)&c>>>0>p>>>0){tk(744673);k[f>>2]=307;m=307;r=g;return m|0}b=i;i=k[b>>2]|0;q=k[b+4>>2]|0;b=yy(i|0,q|0,d|0,e|0)|0;s=L;t=j+968|0;u=j+976|0;v=ky(k[u>>2]|0,k[u+4>>2]|0,k[t>>2]|0,k[t+4>>2]|0)|0;t=L;u=yy(i|0,q|0,c|0,l|0)|0;l=L;c=jy(v|0,t|0,u|0,l|0)|0;q=L;i=j+120|0;j=ky(k[i>>2]|0,k[i+4>>2]|0,u|0,l|0)|0;l=L;u=jy(0,0,b|0,s|0)|0;Oj(a,j,l,c,q,u,L,f)|0;u=ky(v|0,t|0,2879,0)|0;t=xy(u|0,L|0,2880,0)|0;u=L;v=ky(b|0,s|0,2879,0)|0;q=jy(v|0,L|0,t|0,u|0)|0;u=wy(q|0,L|0,2880,0)|0;if((u|0)>0)ul(a,u,f)|0;u=(k[n>>2]|0)+968|0;q=u;t=jy(k[q>>2]|0,k[q+4>>2]|0,b|0,s|0)|0;s=u;k[s>>2]=t;k[s+4>>2]=L;k[h>>2]=0;s=my(0,t|0,32)|0;oq(a,782755,s,L,763480,h)|0;h=jy(p|0,o|0,d|0,e|0)|0;oq(a,790243,h,L,763480,f)|0;h=k[n>>2]|0;n=h+944|0;o=n;p=jy(k[o>>2]|0,k[o+4>>2]|0,d|0,e|0)|0;o=n;k[o>>2]=p;k[o+4>>2]=L;o=h+936|0;h=o;p=jy(k[h>>2]|0,k[h+4>>2]|0,d|0,e|0)|0;e=o;k[e>>2]=p;k[e+4>>2]=L;tl(a,f)|0;m=k[f>>2]|0;r=g;return m|0}function Qj(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0;f=r;r=r+16|0;g=f+4|0;h=f;k[g>>2]=c;k[h>>2]=d;Rj(a,b,1,g,h,e)|0;r=f;return k[e>>2]|0}function Rj(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0;g=r;r=r+256|0;h=g+24|0;j=g+20|0;l=g+16|0;m=g;n=g+12|0;o=g+8|0;p=g+176|0;q=g+101|0;s=g+28|0;t=k[f>>2]|0;if((t|0)>0){u=t;r=g;return u|0}t=k[a>>2]|0;v=a+4|0;w=k[v>>2]|0;if((t|0)==(k[w+68>>2]|0)){x=w+120|0;if(((k[x>>2]|0)==-1?(k[x+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){u=k[f>>2]|0;r=g;return u|0}}else Fk(a,t+1|0,0,f)|0;t=k[v>>2]|0;if(!(k[t+72>>2]|0)){tk(744726);k[f>>2]=235;u=235;r=g;return u|0}x=k[t+928>>2]|0;if((b|0)<1){k[f>>2]=302;u=302;r=g;return u|0}t=(x|0)<(b|0);w=x+1|0;y=t?w:b;b=(c|0)>0;if(b){z=0;A=0;B=0;while(1){ux(p,k[e+(B<<2)>>2]|0)|0;Ck(p);a:do if((k[(k[v>>2]|0)+72>>2]|0)==1){Nk(p,h,n,j,f)|0;C=(k[n>>2]|0)+1|0;D=C;E=((C|0)<0)<<31>>31}else{Qk(p,h,o,n,f)|0;C=k[h>>2]|0;if((C|0)<0){F=(rx(p,81)|0)==0;D=F?8:16;E=F?0:0;break}switch(C|0){case 1:{F=((k[o>>2]|0)+7|0)/8|0;D=F;E=((F|0)<0)<<31>>31;break a;break}case 16:{F=k[o>>2]|0;D=F;E=((F|0)<0)<<31>>31;break a;break}default:{F=ha(k[o>>2]|0,(C|0)/10|0)|0;D=F;E=((F|0)<0)<<31>>31;break a}}}while(0);F=ky(D|0,E|0,z|0,A|0)|0;C=L;B=B+1|0;if((B|0)==(c|0)){G=F;H=C;break}else{z=F;A=C}}}else{G=0;H=0}A=k[f>>2]|0;if((A|0)>0){u=A;r=g;return u|0}A=k[v>>2]|0;z=A+952|0;B=k[z>>2]|0;E=k[z+4>>2]|0;z=A+944|0;D=k[z>>2]|0;o=k[z+4>>2]|0;z=A+968|0;C=A+976|0;F=k[C>>2]|0;I=k[C+4>>2]|0;C=ky(F|0,I|0,k[z>>2]|0,k[z+4>>2]|0)|0;z=L;J=ky(C|0,z|0,2879,0)|0;K=L;M=xy(J|0,K|0,2880,0)|0;N=jy(J|0,K|0,M|0,L|0)|0;M=jy(N|0,L|0,C|0,z|0)|0;z=L;C=yy(D|0,o|0,G|0,H|0)|0;N=L;do if((N|0)>(z|0)|(N|0)==(z|0)&C>>>0>M>>>0){K=ky(C|0,N|0,2879,0)|0;J=jy(K|0,L|0,M|0,z|0)|0;K=wy(J|0,L|0,2880,0)|0;if((rl(a,K,1,f)|0)<=0){K=k[v>>2]|0;J=K+976|0;O=k[J+4>>2]|0;P=k[J>>2]|0;Q=K;break}u=k[f>>2]|0;r=g;return u|0}else{O=I;P=F;Q=A}while(0);do if((O|0)>0|(O|0)==0&P>>>0>0){A=Q+120|0;F=Q+968|0;I=ky(k[F>>2]|0,k[F+4>>2]|0,k[A>>2]|0,k[A+4>>2]|0)|0;if((Oj(a,I,L,P,O,C,N,f)|0)<=0){R=k[v>>2]|0;break}u=k[f>>2]|0;r=g;return u|0}else R=Q;while(0);Q=R+968|0;R=Q;O=ky(k[R>>2]|0,k[R+4>>2]|0,C|0,N|0)|0;N=L;C=Q;k[C>>2]=O;k[C+4>>2]=N;k[l>>2]=0;oq(a,782755,O,N,763480,l)|0;if(t){S=B;T=E}else{l=(k[(k[v>>2]|0)+960>>2]|0)+((y+-1|0)*160|0)+72|0;S=k[l>>2]|0;T=k[l+4>>2]|0}Sj(a,B,E,D,o,G,H,S,T,f)|0;if((x|0)>0?(k[(k[v>>2]|0)+72>>2]|0)==1:0){o=0;do{o=o+1|0;Gk(773688,o,q,f)|0;No(a,q,m,s,f)|0;D=m;l=k[D>>2]|0;N=k[D+4>>2]|0;if((N|0)>(T|0)|(N|0)==(T|0)&l>>>0>S>>>0){D=ky(l|0,N|0,G|0,H|0)|0;N=L;l=m;k[l>>2]=D;k[l+4>>2]=N;oq(a,q,D,N,763480,f)|0}}while((o|0)!=(x|0))}o=x+c|0;oq(a,782733,o,((o|0)<0)<<31>>31,763480,f)|0;o=ky(B|0,E|0,G|0,H|0)|0;oq(a,772833,o,L,763480,f)|0;if(!t)Tj(a,y,x,c,f)|0;b:do if(b){x=S;t=T;o=y;H=0;while(1){G=s;E=744787;B=G+16|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));Gk(772997,o,q,f)|0;fs(a,q,k[d+(H<<2)>>2]|0,s,f)|0;G=s;E=744803;B=G+16|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));ux(p,k[e+(H<<2)>>2]|0)|0;Ck(p);Gk(773003,o,q,f)|0;N=k[h>>2]|0;switch(((N|0)>-1?N:0-N|0)|0){case 12:{N=p;while(1)if((i[N>>0]|0)==83){U=N;break}else N=N+1|0;i[U>>0]=66;fs(a,q,p,s,f)|0;Gk(773309,o,q,f)|0;G=s;E=773315;B=G+24|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,-128.0,0,s,f)|0;Gk(773339,o,q,f)|0;G=s;E=773345;B=G+20|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,1.0,0,s,f)|0;break}case 20:{N=p;while(1)if((i[N>>0]|0)==85){V=N;break}else N=N+1|0;i[V>>0]=73;fs(a,q,p,s,f)|0;Gk(773309,o,q,f)|0;G=s;E=773365;B=G+29|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,32768.0,0,s,f)|0;Gk(773339,o,q,f)|0;G=s;E=773345;B=G+20|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,1.0,0,s,f)|0;break}case 40:{N=p;while(1)if((i[N>>0]|0)==86){W=N;break}else N=N+1|0;i[W>>0]=74;fs(a,q,p,s,f)|0;Gk(773309,o,q,f)|0;G=s;E=773365;B=G+29|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,2147483648.0,0,s,f)|0;Gk(773339,o,q,f)|0;G=s;E=773345;B=G+20|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));hs(a,q,1.0,0,s,f)|0;break}default:fs(a,q,p,s,f)|0}if((k[(k[v>>2]|0)+72>>2]|0)==1){N=(o|0)==(w|0);D=ky((N?2:1)|0,(N?0:0)|0,x|0,t|0)|0;N=m;k[N>>2]=D;k[N+4>>2]=L;G=s;E=744819;B=G+26|0;do{i[G>>0]=i[E>>0]|0;G=G+1|0;E=E+1|0}while((G|0)<(B|0));Gk(773688,o,q,f)|0;E=m;gs(a,q,k[E>>2]|0,k[E+4>>2]|0,s,f)|0;Nk(p,h,n,j,f)|0;E=(k[n>>2]|0)+1|0;G=ky(E|0,((E|0)<0)<<31>>31|0,x|0,t|0)|0;X=G;Y=L}else{X=x;Y=t}H=H+1|0;if((H|0)==(c|0))break b;else{x=X;t=Y;o=o+1|0}}}while(0);Vk(a,f)|0;u=k[f>>2]|0;r=g;return u|0}function Sj(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0;l=r;r=r+1e4|0;m=l;n=k[j>>2]|0;if((d|0)==0&(e|0)==0|(n|0)>0){o=n;r=l;return o|0}n=a+4|0;p=k[n>>2]|0;q=(k[p+72>>2]|0)==1?32:0;s=ky(f|0,g|0,b|0,c|0)|0;t=L;if((t|0)<0|(t|0)==0&s>>>0<10001){if((g|0)>0|(g|0)==0&f>>>0>0)sy(m|0,q|0,f|0)|0;u=ky(h|0,i|0,1,0)|0;v=L;w=jy(b|0,c|0,h|0,i|0)|0;x=L;y=m+f|0;ku(a,d,e,u,v,w,x,y,j)|0;z=(k[n>>2]|0)+952|0;k[z>>2]=s;k[z+4>>2]=t;z=ky(w|0,x|0,f|0,g|0)|0;ru(a,d,e,u,v,z,L,m,j)|0;z=(k[n>>2]|0)+952|0;k[z>>2]=b;k[z+4>>2]=c;if((e|0)>0|(e|0)==0&d>>>0>1){z=d;x=e;do{z=ky(z|0,x|0,-1,-1)|0;x=L;ku(a,z,x,u,v,b,c,y,j)|0;w=(k[n>>2]|0)+952|0;k[w>>2]=s;k[w+4>>2]=t;ru(a,z,x,u,v,s,t,m,j)|0;w=(k[n>>2]|0)+952|0;k[w>>2]=b;k[w+4>>2]=c}while((x|0)>0|(x|0)==0&z>>>0>1)}}else{z=ky(b|0,c|0,9999,0)|0;x=L;v=jy(z|0,x|0,h|0,i|0)|0;u=L;y=wy(v|0,u|0,1e4,0)|0;w=L;A=yy(y|0,w|0,1e4,0)|0;B=L;C=ky(h|0,i|0,-9999,-1)|0;D=ky(C|0,L|0,A|0,B|0)|0;B=L;if((u|0)>0|(u|0)==0&v>>>0>9999){v=ky(b|0,c|0,1,0)|0;u=jy(v|0,L|0,D|0,B|0)|0;v=0;A=0;C=D;D=B;B=u;u=L;while(1){ku(a,d,e,C,D,B,u,m,j)|0;E=(k[n>>2]|0)+952|0;k[E>>2]=s;k[E+4>>2]=t;E=ky(C|0,D|0,f|0,g|0)|0;ru(a,d,e,E,L,B,u,m,j)|0;E=k[n>>2]|0;F=E+952|0;k[F>>2]=b;k[F+4>>2]=c;C=ky(C|0,D|0,-1e4,-1)|0;D=L;v=ky(v|0,A|0,1,0)|0;A=L;if(!((A|0)<(w|0)|(A|0)==(w|0)&v>>>0<y>>>0)){G=E;break}else{B=1e4;u=0}}}else G=p;p=wy(z|0,x|0,1e4,0)|0;x=L;if((e|0)>0|(e|0)==0&d>>>0>1?(z=yy(p|0,x|0,1e4,0)|0,u=ky(z|0,L|0,-1e4,-1)|0,z=L,B=ky(h|0,i|0,1,0)|0,y=ky(B|0,L|0,u|0,z|0)|0,B=L,v=jy(b|0,c|0,u|0,z|0)|0,z=L,(c|0)>0|(c|0)==0&b>>>0>0):0){u=d;w=e;do{u=ky(u|0,w|0,-1,-1)|0;w=L;A=y;D=B;C=v;E=z;F=0;H=0;while(1){ku(a,u,w,A,D,C,E,m,j)|0;I=(k[n>>2]|0)+952|0;k[I>>2]=s;k[I+4>>2]=t;I=ky(A|0,D|0,f|0,g|0)|0;ru(a,u,w,I,L,C,E,m,j)|0;I=k[n>>2]|0;J=I+952|0;k[J>>2]=b;k[J+4>>2]=c;A=ky(A|0,D|0,-1e4,-1)|0;D=L;F=ky(F|0,H|0,1,0)|0;H=L;if(!((H|0)<(x|0)|(H|0)==(x|0)&F>>>0<p>>>0)){K=I;break}else{C=1e4;E=0}}}while((w|0)>0|(w|0)==0&u>>>0>1);M=K}else M=G;sy(m|0,q|0,((g|0)<0|(g|0)==0&f>>>0<1e4?f:1e4)|0)|0;q=ky(f|0,g|0,9999,0)|0;G=wy(q|0,L|0,1e4,0)|0;q=L;K=M+952|0;k[K>>2]=s;k[K+4>>2]=t;if(!((e|0)<0|(e|0)==0&d>>>0<1)?(t=ky(h|0,i|0,1,0)|0,i=L,h=yy(G|0,q|0,-1e4,-1)|0,K=L,s=ky(f|0,g|0,1e4,0)|0,u=ky(s|0,L|0,h|0,K|0)|0,K=L,(g|0)>0|(g|0)==0&f>>>0>0):0){f=1;g=0;while(1){h=t;s=i;w=u;p=K;x=0;z=0;while(1){ru(a,f,g,h,s,w,p,m,j)|0;h=ky(w|0,p|0,h|0,s|0)|0;s=L;x=ky(x|0,z|0,1,0)|0;z=L;if(!((z|0)<(q|0)|(z|0)==(q|0)&x>>>0<G>>>0))break;else{w=1e4;p=0}}p=f;f=ky(f|0,g|0,1,0)|0;if(!((g|0)<(e|0)|(g|0)==(e|0)&p>>>0<d>>>0))break;else g=L}N=k[n>>2]|0}else N=M;M=N+952|0;k[M>>2]=b;k[M+4>>2]=c}o=k[j>>2]|0;r=l;return o|0}function Tj(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;f=r;r=r+256|0;g=f+96|0;h=f+88|0;j=f+84|0;l=f;m=f+175|0;n=f+100|0;Co(a,g,f+92|0,e)|0;if((k[g>>2]|0)<9){o=k[e>>2]|0;r=f;return o|0}p=l+1|0;q=(d|0)<1;s=9;while(1){Ro(a,s,l,e)|0;a:do if((i[l>>0]|0)==84){Bx(m,p,4)|0;do if(((((((((((((((Ax(m,755203,4)|0)!=0?(Ax(m,755198,4)|0)!=0:0)?(Ax(m,755193,4)|0)!=0:0)?(Ax(m,755252,4)|0)!=0:0)?(Ax(m,757583,4)|0)!=0:0)?(Ax(m,760184,4)|0)!=0:0)?(Ax(m,755299,4)|0)!=0:0)?(Ax(m,757839,4)|0)!=0:0)?(Ax(m,757844,4)|0)!=0:0)?(Ax(m,757849,4)|0)!=0:0)?(Ax(m,757854,4)|0)!=0:0)?(Ax(m,757859,4)|0)!=0:0)?(Ax(m,757864,4)|0)!=0:0)?(Ax(m,757891,4)|0)!=0:0)?(Ax(m,757882,4)|0)!=0:0){if(!(Ax(m,757905,4)|0)){t=5;break}if(!(Ax(m,757900,4)|0)){t=5;break}if(!(Ax(m,757873,4)|0)){t=5;break}if(!(Ax(l,774793,4)|0))t=4;else{u=s;break a}}else t=5;while(0);i[m>>0]=0;zx(m,l+t|0,8-t|0)|0;k[h>>2]=0;Ok(m,j,h)|0;if((k[h>>2]|0)==0?(v=k[j>>2]|0,!((v|0)<(b|0)|(v|0)>(c|0))):0)if(q&(v|0)==(b|0)){xq(a,s,e)|0;k[g>>2]=(k[g>>2]|0)+-1;u=s+-1|0;break}else{k[j>>2]=v+d;i[m>>0]=0;zx(m,l,t)|0;Gk(m,k[j>>2]|0,n,e)|0;v=l;k[v>>2]=538976288;k[v+4>>2]=538976288;Bx(l,n,xx(n)|0)|0;wq(a,s,l,e)|0;u=s;break}else u=s}else u=s;while(0);if((u|0)<(k[g>>2]|0))s=u+1|0;else break}o=k[e>>2]|0;r=f;return o|0}function Uj(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;l=r;r=r+1e4|0;m=l;n=k[j>>2]|0;if((d|0)==0&(e|0)==0|(n|0)>0){o=n;r=l;return o|0}n=jy(b|0,c|0,f|0,g|0)|0;p=L;if((p|0)<0|(p|0)==0&n>>>0<10001){q=ky(h|0,i|0,1,0)|0;s=L;t=ky(q|0,s|0,f|0,g|0)|0;u=L;if((e|0)>0|(e|0)==0&d>>>0>1){v=a+4|0;w=1;x=0;do{ku(a,w,x,t,u,n,p,m,j)|0;y=(k[v>>2]|0)+952|0;k[y>>2]=n;k[y+4>>2]=p;ru(a,w,x,q,s,n,p,m,j)|0;y=(k[v>>2]|0)+952|0;k[y>>2]=b;k[y+4>>2]=c;w=ky(w|0,x|0,1,0)|0;x=L}while((x|0)<(e|0)|(x|0)==(e|0)&w>>>0<d>>>0)}w=jy(n|0,p|0,h|0,i|0)|0;x=L;if((x|0)>0|(x|0)==0&w>>>0>0){ku(a,d,e,t,u,w,x,m,j)|0;u=a+4|0;t=(k[u>>2]|0)+952|0;k[t>>2]=n;k[t+4>>2]=p;ru(a,d,e,q,s,w,x,m,j)|0;x=(k[u>>2]|0)+952|0;k[x>>2]=b;k[x+4>>2]=c}}else{x=ky(n|0,p|0,9999,0)|0;u=wy(x|0,L|0,1e4,0)|0;x=L;if((e|0)>0|(e|0)==0&d>>>0>1){w=ky(h|0,i|0,1,0)|0;s=L;q=ky(w|0,s|0,f|0,g|0)|0;t=L;v=yy(u|0,x|0,-1e4,-1)|0;y=L;z=ky(n|0,p|0,1e4,0)|0;A=ky(z|0,L|0,v|0,y|0)|0;y=L;v=a+4|0;z=1;B=0;do{C=q;D=t;E=A;F=y;G=w;H=s;I=0;J=0;while(1){ku(a,z,B,C,D,E,F,m,j)|0;K=(k[v>>2]|0)+952|0;k[K>>2]=n;k[K+4>>2]=p;ru(a,z,B,G,H,E,F,m,j)|0;K=(k[v>>2]|0)+952|0;k[K>>2]=b;k[K+4>>2]=c;G=ky(E|0,F|0,G|0,H|0)|0;H=L;C=ky(E|0,F|0,C|0,D|0)|0;D=L;I=ky(I|0,J|0,1,0)|0;J=L;if(!((J|0)<(x|0)|(J|0)==(x|0)&I>>>0<u>>>0))break;else{E=1e4;F=0}}z=ky(z|0,B|0,1,0)|0;B=L}while((B|0)<(e|0)|(B|0)==(e|0)&z>>>0<d>>>0)}z=jy(n|0,p|0,h|0,i|0)|0;B=L;if((B|0)>0|(B|0)==0&z>>>0>0){u=ky(z|0,B|0,9999,0)|0;x=wy(u|0,L|0,1e4,0)|0;u=L;v=ky(h|0,i|0,1,0)|0;i=L;h=ky(z|0,B|0,1e4,0)|0;B=L;z=yy(x|0,u|0,-1e4,-1)|0;s=ky(h|0,B|0,z|0,L|0)|0;z=L;B=ky(v|0,i|0,f|0,g|0)|0;g=a+4|0;f=B;B=L;h=s;s=z;z=v;v=i;i=0;w=0;while(1){ku(a,d,e,f,B,h,s,m,j)|0;y=(k[g>>2]|0)+952|0;k[y>>2]=n;k[y+4>>2]=p;ru(a,d,e,z,v,h,s,m,j)|0;y=(k[g>>2]|0)+952|0;k[y>>2]=b;k[y+4>>2]=c;z=ky(h|0,s|0,z|0,v|0)|0;v=L;f=ky(h|0,s|0,f|0,B|0)|0;B=L;i=ky(i|0,w|0,1,0)|0;w=L;if(!((w|0)<(u|0)|(w|0)==(u|0)&i>>>0<x>>>0))break;else{h=1e4;s=0}}}}o=k[j>>2]|0;r=l;return o|0}function Vj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;d=r;r=r+160|0;e=d+8|0;f=d;g=d+85|0;h=d+12|0;i=k[c>>2]|0;if((i|0)>0){j=i;r=d;return j|0}i=k[a>>2]|0;l=a+4|0;m=k[l>>2]|0;if((i|0)==(k[m+68>>2]|0)){n=m+120|0;if(((k[n>>2]|0)==-1?(k[n+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){j=k[c>>2]|0;r=d;return j|0}}else Fk(a,i+1|0,0,c)|0;i=k[l>>2]|0;n=k[i+72>>2]|0;if(!n){tk(744845);k[c>>2]=235;j=235;r=d;return j|0}if((b|0)>=1?(m=k[i+928>>2]|0,(m|0)>=(b|0)):0){o=k[i+960>>2]|0;p=b+-1|0;q=o+(p*160|0)+72|0;s=k[q>>2]|0;t=k[q+4>>2]|0;do if((n|0)==1){q=k[o+(p*160|0)+152>>2]|0;u=((q|0)<0)<<31>>31;if((m|0)>(b|0)){v=o+(b*160|0)+72|0;w=k[v>>2]|0;x=k[v+4>>2]|0;v=ky(q|0,u|0,s|0,t|0)|0;y=jy(w|0,x|0,v|0,L|0)|0;v=ky((y|0)>0|0,0,q|0,u|0)|0;z=v;A=L;B=s;C=t;break}if((b|0)>1?(v=b+-2|0,y=o+(v*160|0)+72|0,x=jy(s|0,t|0,k[y>>2]|0,k[y+4>>2]|0)|0,y=jy(x|0,L|0,k[o+(v*160|0)+152>>2]|0,0)|0,(y|0)>0):0){y=ky(q|0,u|0,1,0)|0;v=L;x=ky(s|0,t|0,-1,-1)|0;z=y;A=v;B=x;C=L}else{z=q;A=u;B=s;C=t}}else if((m|0)>(b|0)){u=o+(b*160|0)+72|0;q=jy(k[u>>2]|0,k[u+4>>2]|0,s|0,t|0)|0;z=q;A=L;B=s;C=t;break}else{q=i+952|0;u=jy(k[q>>2]|0,k[q+4>>2]|0,s|0,t|0)|0;z=u;A=L;B=s;C=t;break}while(0);t=i+952|0;s=k[t>>2]|0;o=k[t+4>>2]|0;t=i+944|0;m=k[t>>2]|0;p=k[t+4>>2]|0;t=i+968|0;n=i+976|0;i=ky(k[n>>2]|0,k[n+4>>2]|0,k[t>>2]|0,k[t+4>>2]|0)|0;t=L;n=yy(m|0,p|0,z|0,A|0)|0;u=L;q=ky(i|0,t|0,2879,0)|0;t=xy(q|0,L|0,2880,0)|0;q=jy(n|0,u|0,t|0,L|0)|0;t=ky(q|0,L|0,2879,0)|0;q=wy(t|0,L|0,2880,0)|0;Uj(a,s,o,m,p,z,A,B,C,c)|0;p=k[l>>2]|0;m=p+976|0;t=k[m>>2]|0;i=k[m+4>>2]|0;if((i|0)>0|(i|0)==0&t>>>0>0?(m=p+120|0,x=p+968|0,p=ky(k[x>>2]|0,k[x+4>>2]|0,k[m>>2]|0,k[m+4>>2]|0)|0,m=L,x=jy(0,0,n|0,u|0)|0,(Oj(a,p,m,t,i,x,L,c)|0)>0):0){j=k[c>>2]|0;r=d;return j|0}if((q|0)>0)ul(a,q,c)|0;q=(k[l>>2]|0)+968|0;x=q;i=jy(k[x>>2]|0,k[x+4>>2]|0,n|0,u|0)|0;u=q;k[u>>2]=i;k[u+4>>2]=L;k[e>>2]=0;u=my(0,i|0,32)|0;oq(a,782755,u,L,763480,e)|0;e=k[l>>2]|0;if((k[e+72>>2]|0)==1?(k[e+928>>2]|0)>=1:0){u=1;while(1){Gk(773688,u,g,c)|0;No(a,g,f,h,c)|0;i=f;q=k[i>>2]|0;n=k[i+4>>2]|0;if((n|0)>(C|0)|(n|0)==(C|0)&q>>>0>B>>>0){i=jy(q|0,n|0,z|0,A|0)|0;n=L;q=f;k[q>>2]=i;k[q+4>>2]=n;oq(a,g,i,n,763480,c)|0}n=k[l>>2]|0;if((u|0)<(k[n+928>>2]|0))u=u+1|0;else{D=n;break}}}else D=e;e=(k[D+928>>2]|0)+-1|0;oq(a,782733,e,((e|0)<0)<<31>>31,763480,c)|0;e=jy(s|0,o|0,z|0,A|0)|0;oq(a,772833,e,L,763480,c)|0;Tj(a,b,k[(k[l>>2]|0)+928>>2]|0,-1,c)|0;Vk(a,c)|0;j=k[c>>2]|0;r=d;return j|0}k[c>>2]=302;j=302;r=d;return j|0}function Wj(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;if((g|0)>0){h=g;r=e;return h|0}if((a|0)==(b|0)){k[d>>2]=101;h=101;r=e;return h|0}if((Xj(a,b,d)|0)>0){h=k[d>>2]|0;r=e;return h|0}if((c|0)<=0){if((Co(a,0,f,d)|0)>0){h=k[d>>2]|0;r=e;return h|0}g=k[f>>2]|0;if((g|0)>0?(zl(b,g,d)|0,(k[f>>2]|0)>34):0)ol(b,d)|0}else zl(b,c,d)|0;Yj(a,b,d)|0;h=k[d>>2]|0;r=e;return h|0}function Xj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;d=r;r=r+96|0;e=d+8|0;f=d+4|0;g=d;h=d+12|0;j=k[c>>2]|0;if((j|0)>0){l=j;r=d;return l|0}if((a|0)==(b|0)){k[c>>2]=101;l=101;r=d;return l|0}j=k[a>>2]|0;m=a+4|0;if((j|0)!=(k[(k[m>>2]|0)+68>>2]|0))Fk(a,j+1|0,0,c)|0;if((Co(a,e,0,c)|0)>0){l=k[c>>2]|0;r=d;return l|0}j=k[e>>2]|0;n=cy(j*81|0)|0;if(!n){k[c>>2]=113;l=113;r=d;return l|0}if((j|0)>0){j=0;do{o=j;j=j+1|0;Ro(a,j,n+(o*81|0)|0,c)|0}while((j|0)<(k[e>>2]|0))}j=k[a>>2]|0;k[f>>2]=-1;if(!(k[(k[m>>2]|0)+72>>2]|0))Ko(a,782713,f,0,c)|0;a=k[b>>2]|0;m=b+4|0;o=k[m>>2]|0;if((a|0)==(k[o+68>>2]|0)){p=o;q=a}else{Fk(b,a+1|0,0,c)|0;a=k[m>>2]|0;p=a;q=k[a+68>>2]|0}a=p+96|0;m=(k[p+88>>2]|0)+(q<<3)|0;if(!((k[a>>2]|0)==(k[m>>2]|0)?(k[a+4>>2]|0)==(k[m+4>>2]|0):0))Al(b,c)|0;if(!(k[b>>2]|0))if((k[f>>2]|0)<0){Xr(b,8,0,g,c)|0;Al(b,c)|0;s=0}else s=1;else s=0;if((k[c>>2]|0)>0){dy(n);l=k[c>>2]|0;r=d;return l|0}a:do if(!(s|j)){t=h;u=771919;v=t+16|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));fs(b,782702,771935,h,c)|0;if(((k[f>>2]|0)+3|0)>1){g=1;while(1){ds(b,n+(g*81|0)|0,c)|0;if((g|0)<((k[f>>2]|0)+2|0))g=g+1|0;else break}}t=h;u=772318;v=t+34|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));gs(b,782652,0,0,h,c)|0;t=h;u=772352;v=t+24|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));gs(b,782726,1,0,h,c)|0;g=(k[f>>2]|0)+3|0;m=k[e>>2]|0;if((g|0)<(m|0)){a=m;m=g;while(1){g=n+(m*81|0)|0;switch(i[g>>0]|0){case 69:{if(!(Ax(g,744910,8)|0))w=a;else x=31;break}case 67:{if((Ax(g,744919,58)|0)!=0?(Ax(g,757327,47)|0)!=0:0)x=31;else w=a;break}default:x=31}if((x|0)==31){x=0;ds(b,g,c)|0;w=k[e>>2]|0}m=m+1|0;if((m|0)>=(w|0))break a;else a=w}}}else{if(!((j|0)!=0&(s|0)==1)){if((k[e>>2]|0)>0)y=0;else break;while(1){ds(b,n+(y*81|0)|0,c)|0;y=y+1|0;if((y|0)>=(k[e>>2]|0))break a}}t=h;u=778844;v=t+35|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));is(b,782785,1,h,c)|0;if(((k[f>>2]|0)+3|0)>1){a=1;while(1){ds(b,n+(a*81|0)|0,c)|0;if((a|0)<((k[f>>2]|0)+2|0))a=a+1|0;else break}}t=h;u=772186;v=t+36|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));is(b,782808,1,h,c)|0;ds(b,772376,c)|0;ds(b,772457,c)|0;a=(k[f>>2]|0)+3|0;m=k[e>>2]|0;if((a|0)<(m|0)){g=m;m=a;while(1){a=n+(m*81|0)|0;switch(i[a>>0]|0){case 80:{if(!(Ax(a,744978,8)|0))z=g;else x=41;break}case 71:{if(!(Ax(a,744987,8)|0))z=g;else x=41;break}default:x=41}if((x|0)==41){x=0;ds(b,a,c)|0;z=k[e>>2]|0}m=m+1|0;if((m|0)>=(z|0))break a;else g=z}}}while(0);dy(n);l=k[c>>2]|0;r=d;return l|0}function Yj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;d=r;r=r+2912|0;e=d+16|0;f=d+8|0;g=d;h=d+24|0;i=k[c>>2]|0;if((i|0)>0){j=i;r=d;return j|0}if((a|0)==(b|0)){k[c>>2]=101;j=101;r=d;return j|0}cl(a,0,e,f,c)|0;cl(b,0,g,0,c)|0;i=f;f=e;l=k[f>>2]|0;m=k[f+4>>2]|0;f=jy(k[i>>2]|0,k[i+4>>2]|0,l|0,m|0)|0;i=wy(f|0,L|0,2880,0)|0;a:do if((i|0)>0){if((k[a+4>>2]|0)==(k[b+4>>2]|0)){f=l;n=m;o=0;while(1){au(a,f,n,0,c)|0;fu(a,2880,0,h,c)|0;p=g;au(b,k[p>>2]|0,k[p+4>>2]|0,1,c)|0;cu(b,2880,0,h,c)|0;p=e;f=ky(k[p>>2]|0,k[p+4>>2]|0,2880,0)|0;n=L;p=e;k[p>>2]=f;k[p+4>>2]=n;p=g;q=ky(k[p>>2]|0,k[p+4>>2]|0,2880,0)|0;p=g;k[p>>2]=q;k[p+4>>2]=L;o=o+1|0;if((o|0)==(i|0))break a}}au(a,l,m,0,c)|0;o=g;au(b,k[o>>2]|0,k[o+4>>2]|0,1,c)|0;o=0;do{fu(a,2880,0,h,c)|0;cu(b,2880,0,h,c)|0;o=o+1|0}while((o|0)!=(i|0))}while(0);j=k[c>>2]|0;r=d;return j|0}function Zj(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=r;r=r+176|0;o=n;p=n+12|0;q=n+8|0;s=n+4|0;t=n+87|0;u=n+16|0;v=k[m>>2]|0;if((v|0)>0){w=v;r=n;return w|0}i[u>>0]=0;if(h)zx(u,h,70)|0;h=k[a>>2]|0;v=a+4|0;x=k[v>>2]|0;if((h|0)==(k[x+68>>2]|0)){y=x;z=h}else{Fk(a,h+1|0,0,m)|0;h=k[v>>2]|0;y=h;z=k[h+68>>2]|0}h=y+96|0;x=k[y+88>>2]|0;A=x+(z<<3)|0;do if(!((k[h>>2]|0)==(k[A>>2]|0)?(k[h+4>>2]|0)==(k[A+4>>2]|0):0)){if((z|0)==(k[y+80>>2]|0)?(B=x+(z+1<<3)|0,C=k[B+4>>2]|0,D=y+32|0,E=k[D+4>>2]|0,!((C|0)<(E|0)|((C|0)==(E|0)?(k[B>>2]|0)>>>0<(k[D>>2]|0)>>>0:0))):0)break;if((c|0)<0){k[m>>2]=218;w=218;r=n;return w|0}if(d>>>0>999){k[o>>2]=d;Xw(t,744996,o)|0;tk(t);k[m>>2]=216;w=216;r=n;return w|0}D=(d|0)>0;if((g|0)==0|D^1)F=0;else{B=(k[g>>2]|0)==0;E=0;C=0;while(1){if(B)G=C;else G=((i[k[g+(E<<2)>>2]>>0]|0)!=0&1)+C|0;E=E+1|0;if((E|0)==(d|0)){F=G;break}else C=G}}C=((d<<1)+44+F+((i[u>>0]|0)!=0&1)|0)/36|0;if(D){E=0;B=0;H=0;while(1){Qk(k[f+(H<<2)>>2]|0,p,q,s,m)|0;switch(k[p>>2]|0){case 1:{I=((k[q>>2]|0)+7|0)/8|0;break}case 16:{I=k[q>>2]|0;break}default:I=ha(k[s>>2]|0,k[q>>2]|0)|0}J=ky(I|0,((I|0)<0)<<31>>31|0,E|0,B|0)|0;K=L;H=H+1|0;if((H|0)==(d|0)){M=J;N=K;break}else{E=J;B=K}}O=k[v>>2]|0;P=M;Q=N}else{O=y;P=0;Q=0}if((k[O+76>>2]|0)!=1){k[m>>2]=112;w=112;r=n;return w|0}B=yy(P|0,Q|0,b|0,c|0)|0;E=L;H=ky(j|0,l|0,2879,0)|0;D=ky(H|0,L|0,B|0,E|0)|0;E=wy(D|0,L|0,2880,0)|0;Vk(a,m)|0;yl(a,m)|0;D=k[v>>2]|0;B=(k[D+68>>2]|0)+1|0;H=(k[D+88>>2]|0)+(B<<3)|0;K=k[H>>2]|0;J=k[H+4>>2]|0;k[D+72>>2]=2;if((rl(a,E+C|0,1,m)|0)>0){w=k[m>>2]|0;r=n;return w|0}E=k[v>>2]|0;D=E+80|0;H=k[D>>2]|0;R=H+1|0;k[D>>2]=R;D=E+68|0;S=k[D>>2]|0;T=k[E+88>>2]|0;if((H|0)>=(S|0)){H=R;do{R=T+(H<<3)|0;U=k[R+4>>2]|0;V=T+(H+1<<3)|0;k[V>>2]=k[R>>2];k[V+4>>2]=U;H=H+-1|0}while((H|0)>(S|0))}S=T+(B<<3)|0;H=S;k[H>>2]=K;k[H+4>>2]=J;k[D>>2]=B;k[a>>2]=B;H=E+112|0;k[H>>2]=K;k[H+4>>2]=J;H=S;U=k[H+4>>2]|0;V=E+96|0;k[V>>2]=k[H>>2];k[V+4>>2]=U;U=S;S=C*2880|0;V=ky(k[U>>2]|0,k[U+4>>2]|0,S|0,((S|0)<0)<<31>>31|0)|0;S=E+120|0;k[S>>2]=V;k[S+4>>2]=L;k[E+72>>2]=2;as(a,b,c,d,e,f,g,u,j,l,m)|0;Vk(a,m)|0;w=k[m>>2]|0;r=n;return w|0}while(0);$r(a,2,b,c,d,e,f,g,u,m)|0;w=k[m>>2]|0;r=n;return w|0}function _j(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+16|0;e=d+4|0;k[e>>2]=0;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}f=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((f|0)==(k[i+68>>2]|0)){j=f;l=i}else{Fk(a,f+1|0,0,c)|0;f=k[h>>2]|0;j=k[f+68>>2]|0;l=f}if(j){f=k[l+88>>2]|0;i=f+(j+1<<3)|0;m=f+(j<<3)|0;j=jy(k[i>>2]|0,k[i+4>>2]|0,k[m>>2]|0,k[m+4>>2]|0)|0;m=wy(j|0,L|0,2880,0)|0;if((ul(a,m,c)|0)>0){g=k[c>>2]|0;r=d;return g|0}m=k[h>>2]|0;j=k[m+68>>2]|0;i=m+80|0;f=k[i>>2]|0;if((j|0)<(f|0)){n=k[m+88>>2]|0;o=j+1|0;do{j=o;o=o+1|0;p=n+(o<<3)|0;q=k[p+4>>2]|0;s=n+(j<<3)|0;k[s>>2]=k[p>>2];k[s+4>>2]=q}while((j|0)<(f|0));t=n}else t=k[m+88>>2]|0;m=t+(f+1<<3)|0;k[m>>2]=0;k[m+4>>2]=0;k[i>>2]=f+-1;if((dl(a,e,c)|0)>0){k[c>>2]=0;xk();Il(a,(k[(k[h>>2]|0)+68>>2]|0)+-1|0,e,c)|0}}else{f=l+96|0;k[f>>2]=0;k[f+4>>2]=0;f=l+112|0;k[f>>2]=0;k[f+4>>2]=0;Yr(a,1,8,0,d,0,0,1,0,1,c)|0;f=k[h>>2]|0;l=(k[f+88>>2]|0)+((k[f+68>>2]|0)+1<<3)|0;i=ky(k[l>>2]|0,k[l+4>>2]|0,-2880,-1)|0;l=wy(i|0,L|0,2880,0)|0;do if((l|0)>0){if((ul(a,l,c)|0)<=0){u=k[h>>2]|0;break}g=k[c>>2]|0;r=d;return g|0}else u=f;while(0);f=u+120|0;k[f>>2]=-1;k[f+4>>2]=-1;Vk(a,c)|0}if(b)k[b>>2]=k[e>>2];g=k[c>>2]|0;r=d;return g|0}function $j(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;h=r;r=r+48|0;j=h+28|0;l=h+20|0;m=k[g>>2]|0;if(m){n=m;r=h;return n|0}if(ak(a,0,b,5,j,l,h+24|0,h,g)|0){bk();n=k[g>>2]|0;r=h;return n|0}b=k[l>>2]|0;if((b|0)<0){a=0-b|0;k[l>>2]=a;o=a;p=1}else{o=b;p=0}if((k[j>>2]|0)!=14|(o|0)!=1){bk();tk(745034);k[g>>2]=432;n=432;r=h;return n|0}do if(!p){k[j+4>>2]=f;k[j+8>>2]=0;o=j+12|0;k[o>>2]=d;if((Pq(k[30826]|0,k[30827]|0,(c|0)>1?c+-1|0:0,0,1,j,g)|0)!=-1){if(k[g>>2]|0)break}else k[g>>2]=0;k[e>>2]=0;b=k[o>>2]|0;if((b|0)>0){o=0;a=0;while(1){if((i[f+a>>0]|0)==1){l=o+1|0;k[e>>2]=l;q=l}else q=o;a=a+1|0;if((a|0)>=(b|0))break;else o=q}}}else{o=i[(k[30820]|0)+((k[30823]|0)*344|0)+88>>0]|0;k[e>>2]=d;if((d|0)>0)sy(f|0,o|0,d|0)|0}while(0);bk();n=k[g>>2]|0;r=h;return n|0}function ak(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;l=r;r=r+64|0;m=l+8|0;n=l;o=l+24|0;p=l+20|0;q=l+28|0;k[o>>2]=0;s=k[j>>2]|0;if(s){t=s;r=l;return t|0}if(Vk(a,j)|0){t=k[j>>2]|0;r=l;return t|0}k[30810]=a;k[30813]=b;k[30826]=0;k[30827]=0;k[30828]=0;k[30811]=26;k[30812]=1;k[30820]=0;k[30822]=0;k[30821]=0;k[30834]=0;k[30835]=0;Zk(a,123336,j)|0;if(!(k[30834]|0)){Dl(a,9,l+16|0,p,q,j)|0;if(k[j>>2]|0){tk(745084);t=k[j>>2]|0;r=l;return t|0}b=k[p>>2]|0;p=(b|0)>0;s=p&1;k[30832]=s;if(p){p=s;u=0;while(1){v=ha(p,k[q+(u<<2)>>2]|0)|0;u=u+1|0;if((u|0)>=(b|0)){w=v;break}else p=v}k[30832]=w;x=w}else x=s;if(k[29972]|0){k[n>>2]=b;k[n+4>>2]=x;Rw(745123,n)|0}}else if(Ko(a,790243,123328,0,o)|0)k[30832]=0;do if((i[c>>0]|0)==64){if(!(Uh(c+1|0,123268,j)|0)){o=k[30817]|0;y=o;z=xx(o)|0;break}t=k[j>>2]|0;r=l;return t|0}else{o=xx(c)|0;n=cy(o+2|0)|0;k[30817]=n;ux(n,c)|0;y=k[30817]|0;z=o}while(0);c=y+((xx(y+z|0)|0)+z)|0;i[c>>0]=10;i[c+1>>0]=0;k[30818]=0;k[30819]=0;Bu(0);if(Du()|0){k[j>>2]=431;t=431;r=l;return t|0}c=k[30835]|0;k[j>>2]=c;if(c){t=c;r=l;return t|0}if(!(k[30821]|0)){tk(745155);k[j>>2]=431;t=431;r=l;return t|0}if(!(k[30826]|0)){k[29973]=a;k[30827]=119892}a=k[30820]|0;c=k[30823]|0;z=k[a+(c*344|0)+60>>2]|0;k[g>>2]=z;k[f>>2]=k[a+(c*344|0)+56>>2];if((d|0)>0&(z|0)>0){g=0-d|0;d=(z|0)>1?0-z|0:-1;z=0-(d>>>0<g>>>0?g:d)|0;d=0;do{k[h+(d<<2)>>2]=k[a+(c*344|0)+64+(d<<2)>>2];d=d+1|0}while((d|0)!=(z|0))}switch(k[a+(c*344|0)+52>>2]|0){case 258:{k[e>>2]=14;A=14;break}case 259:{k[e>>2]=41;A=41;break}case 260:{k[e>>2]=82;A=82;break}case 262:{k[e>>2]=1;A=1;break}case 261:{k[e>>2]=16;A=16;break}default:{k[e>>2]=0;tk(745172);k[30835]=432;k[j>>2]=432;A=k[e>>2]|0}}k[30833]=A;A=k[30817]|0;if(!A){k[m>>2]=745229;k[m+4>>2]=936;Rw(745193,m)|0}else dy(A);if((k[a+(c*344|0)>>2]|0)==-1e3)k[f>>2]=0-(k[f>>2]|0);t=k[j>>2]|0;r=l;return t|0}function bk(){var a=0,b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;a=r;r=r+32|0;b=a+16|0;c=a+8|0;d=a;if((k[30826]|0)>0){e=k[30827]|0;if(!e){k[d>>2]=745229;k[d+4>>2]=951;Rw(745238,d)|0}else dy(e);e=k[30826]|0;d=k[30828]|0;if((e|0)>0){f=e;e=d;g=0;while(1){if(!(k[e+(g*124|0)+116>>2]|0)){h=f;i=e}else{do if((k[e+(g*124|0)+84>>2]|0)==262){j=k[k[e+(g*124|0)+120>>2]>>2]|0;if(!j){k[c>>2]=745229;k[c+4>>2]=955;Rw(745277,c)|0;break}else{dy(j);break}}while(0);dy(k[(k[30828]|0)+(g*124|0)+116>>2]|0);h=k[30826]|0;i=k[30828]|0}g=g+1|0;if((g|0)>=(h|0)){l=i;break}else{f=h;e=i}}}else l=d;if(!l){k[b>>2]=745229;k[b+4>>2]=958;Rw(745339,b)|0}else dy(l);k[30826]=0}l=k[30821]|0;if((l|0)>0){b=l;do{b=b+-1|0;l=k[30820]|0;switch(k[l+(b*344|0)>>2]|0){case 1032:{d=k[l+((k[l+(b*344|0)+12>>2]|0)*344|0)+88>>2]|0;if(d)dy(d);break}case 1033:{Es(k[l+((k[l+(b*344|0)+12>>2]|0)*344|0)+88>>2]|0);break}default:{}}}while((b|0)!=0);k[30821]=0}b=k[30820]|0;if(!b){k[30820]=0;k[30834]=-1;k[30829]=0;r=a;return}dy(b);k[30820]=0;k[30834]=-1;k[30829]=0;r=a;return}function ck(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0.0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0;h=r;r=r+192|0;m=h+88|0;n=h+80|0;q=h+72|0;s=h+64|0;t=h+56|0;u=h+48|0;v=h+40|0;w=h+32|0;x=h+24|0;y=h;z=h+112|0;A=h+104|0;B=h+100|0;C=h+96|0;D=h+108|0;k[B>>2]=0;if(k[29972]|0){k[y>>2]=a;k[y+4>>2]=b;k[y+8>>2]=c;k[y+12>>2]=d;k[y+16>>2]=e;Rw(745378,y)|0}y=e+-1|0;a:do if((b+1|0)==(c|0)){k[30034]=g;k[g+16>>2]=0;E=g+12|0;F=k[E>>2]|0;do if((F|0)<=0)if((F|0)<0){k[E>>2]=a;break}else{k[E>>2]=d;break}else k[E>>2]=(F|0)>(a|0)?a:F;while(0);k[30035]=c+-1+(k[g+12>>2]|0);F=k[g+4>>2]|0;if(!F){if(!(k[f+(y*244|0)+84>>2]|0)){tk(745443);G=434;r=h;return G|0}k[30036]=k[f+(y*244|0)+88>>2];k[g>>2]=k[f+(y*244|0)+80>>2];k[A>>2]=0;E=98840;k[E>>2]=0;k[E+4>>2]=0;b:do if(!(k[30834]|0)){E=k[(k[30829]|0)+24>>2]|0;if(!E)H=g;else{I=98840;k[I>>2]=E;k[I+4>>2]=((E|0)<0)<<31>>31;H=g}}else{Vo(k[f+(y*244|0)>>2]|0,770576,k[f+(y*244|0)+4>>2]|0,1,98840,C,A)|0;E=k[30034]|0;if((k[A>>2]|0)==403)switch(k[E>>2]|0){case 21:{I=98840;k[I>>2]=-32768;k[I+4>>2]=-1;H=E;break b;break}case 31:{I=98840;k[I>>2]=-2147483648;k[I+4>>2]=-1;H=E;break b;break}case 41:{I=98840;k[I>>2]=-2147483648;k[I+4>>2]=-1;H=E;break b;break}default:{H=E;break b}}else H=E}while(0);J=H;K=f+(y*244|0)+92|0}else{k[30037]=F;E=k[g+8>>2]|0;k[30036]=(E|0)==0?120152:E;J=g;K=(k[30820]|0)+((k[30823]|0)*344|0)+56|0}k[30042]=k[K>>2];switch(k[J>>2]|0){case 11:case 14:case 1:{k[30043]=1;break}case 21:{k[30043]=2;break}case 31:{k[30043]=4;break}case 41:{k[30043]=4;break}case 81:{k[30043]=8;break}case 42:{k[30043]=4;break}case 82:{k[30043]=8;break}case 16:{k[30043]=4;break}default:{}}switch(k[(k[30820]|0)+((k[30823]|0)*344|0)+52>>2]|0){case 258:{k[30044]=1;L=J;break a;break}case 259:{k[30044]=4;L=J;break a;break}case 260:{k[30044]=8;L=J;break a;break}default:{L=J;break a}}}else L=k[30034]|0;while(0);c:do if(!(k[L+4>>2]|0)){k[30037]=(k[f+(y*244|0)+88>>2]|0)+(k[30043]|0);switch(k[L>>2]|0){case 14:{i[k[30036]>>0]=85;break c;break}case 11:{i[k[30036]>>0]=k[24710];break c;break}case 21:{j[k[30036]>>1]=k[24710];break c;break}case 31:{k[k[30036]>>2]=k[24710];break c;break}case 41:{k[k[30036]>>2]=k[24710];break c;break}case 81:{J=98840;K=k[J+4>>2]|0;g=k[30036]|0;k[g>>2]=k[J>>2];k[g+4>>2]=K;break c;break}case 42:{o[k[30036]>>2]=-9.119119840596153e-36;break c;break}case 82:{p[k[30036]>>3]=-9.1191291391491e-36;break c;break}case 16:{i[k[k[30036]>>2]>>0]=1;i[(k[k[30036]>>2]|0)+1>>0]=0;break c;break}default:break c}}while(0);L=(k[30035]|0)-c|0;y=(L|0)<(d|0)?L+1|0:d;k[30830]=c;k[30831]=y;d:do if((e|0)>0){d=y<<2;L=(y|0)>0;K=(y|0)==0;g=(y|0)==1;J=(y|0)==2;H=0;e:while(1){A=k[30828]|0;if((k[f+(H*244|0)+84>>2]|0)!=2){b=k[A+(H*124|0)+88>>2]|0;E=ha(b,y)|0;I=k[A+(H*124|0)+84>>2]|0;switch(I|0){case 262:{M=b+1|0;N=ha(M,y)|0;O=A+(H*124|0)+120|0;P=k[O>>2]|0;do if(P){Q=k[P>>2]|0;if(!Q){k[x>>2]=745229;k[x+4>>2]=1382;Rw(745487,x)|0;break}else{dy(Q);break}}while(0);dy(P);Q=cy(d)|0;if(!Q){R=A;S=O;T=H;U=56;break e}V=cy(N)|0;k[Q>>2]=V;if(!V){W=A;X=O;Y=Q;Z=H;U=69;break e}f:do if(L){_=(b+7|0)/8|0;$=f+(H*244|0)+88|0;if((b|0)<=0){k[Q>>2]=V;i[V>>0]=0;if(g)break;aa=V+M|0;k[Q+4>>2]=aa;i[aa>>0]=0;if(J)break;else ba=2;while(1){aa=(k[Q>>2]|0)+(ha(ba,M)|0)|0;k[Q+(ba<<2)>>2]=aa;i[aa>>0]=0;ba=ba+1|0;if((ba|0)==(y|0))break f}}if((b|0)==1){aa=V;ca=0;while(1){da=aa+(ha(ca,M)|0)|0;ea=Q+(ca<<2)|0;k[ea>>2]=da;fa=(ha(ca,_)|0)+1|0;i[da>>0]=(l[(k[$>>2]|0)+fa>>0]|0)>>>7|48;i[(k[ea>>2]|0)+1>>0]=0;ea=ca+1|0;if((ea|0)==(y|0))break f;aa=k[Q>>2]|0;ca=ea}}else{ga=V;ia=0}while(1){ca=ga+(ha(ia,M)|0)|0;aa=Q+(ia<<2)|0;k[aa>>2]=ca;ea=(ha(ia,_)|0)+1|0;i[ca>>0]=(l[(k[$>>2]|0)+ea>>0]|0)>>>7|48;ca=0;fa=1;da=ea;do{da=((ca|0)==7&1)+da|0;ca=(fa|0)%8|0;i[(k[aa>>2]|0)+fa>>0]=(1<<7-ca&i[(k[$>>2]|0)+da>>0]|0)==0?48:49;fa=fa+1|0}while((fa|0)!=(b|0));i[(k[aa>>2]|0)+b>>0]=0;fa=ia+1|0;if((fa|0)==(y|0))break f;ga=k[Q>>2]|0;ia=fa}}while(0);k[A+(H*124|0)+116>>2]=Q;k[O>>2]=Q;break}case 261:{b=k[f+(H*244|0)+88>>2]|0;M=A+(H*124|0)+116|0;V=k[M>>2]|0;if(V)dy(V);V=cy(y)|0;k[M>>2]=V;if(!V){ja=H;U=76;break e}if(!K){V=y;do{N=k[b>>2]|0;P=i[N>>0]|0;if(!(P<<24>>24))ka=0;else{$=k[b+(V<<2)>>2]|0;_=i[$>>0]|0;if(P<<24>>24>=_<<24>>24)if(P<<24>>24>_<<24>>24)la=1;else la=tx(N,$)|0;else la=-1;ka=(la|0)==0}V=V+-1|0;i[(k[M>>2]|0)+V>>0]=ka&1}while((V|0)!=0)}k[A+(H*124|0)+120>>2]=b+4;break}case 258:{V=k[f+(H*244|0)+88>>2]|0;M=A+(H*124|0)+116|0;Q=k[M>>2]|0;if(Q)dy(Q);Q=cy(E)|0;k[M>>2]=Q;if(!Q){ma=H;U=88;break e}if(E){Q=E;do{O=i[V>>0]|0;if(!(O<<24>>24))na=0;else na=O<<24>>24==(i[V+Q>>0]|0);Q=Q+-1|0;i[(k[M>>2]|0)+Q>>0]=na&1}while((Q|0)!=0)}k[A+(H*124|0)+120>>2]=V+1;break}case 259:{Q=k[f+(H*244|0)+88>>2]|0;M=A+(H*124|0)+116|0;b=k[M>>2]|0;if(b)dy(b);b=cy(E)|0;k[M>>2]=b;if(!b){oa=H;U=97;break e}if(E){b=E;do{O=k[Q>>2]|0;if(!O)pa=0;else pa=(O|0)==(k[Q+(b<<2)>>2]|0);b=b+-1|0;i[(k[M>>2]|0)+b>>0]=pa&1}while((b|0)!=0)}k[A+(H*124|0)+120>>2]=Q+4;break}case 260:{b=k[f+(H*244|0)+88>>2]|0;M=A+(H*124|0)+116|0;V=k[M>>2]|0;if(V)dy(V);V=cy(E)|0;k[M>>2]=V;if(!V){qa=H;U=106;break e}if(E){V=E;do{ra=+p[b>>3];if(ra!=0.0)sa=ra==+p[b+(V<<3)>>3];else sa=0;V=V+-1|0;i[(k[M>>2]|0)+V>>0]=sa&1}while((V|0)!=0)}k[A+(H*124|0)+120>>2]=b+8;break}default:{k[w>>2]=I;Xw(z,745522,w)|0;tk(z)}}if(k[30835]|0){ta=H;break}}H=H+1|0;if((H|0)>=(e|0))break d}if((U|0)==56){k[R+(T*124|0)+116>>2]=0;k[S>>2]=0;k[30835]=113;ta=T}else if((U|0)==69){dy(Y);k[W+(Z*124|0)+116>>2]=0;k[X>>2]=0;k[30835]=113;ta=Z}else if((U|0)==76){k[30835]=113;ta=ja}else if((U|0)==88){k[30835]=113;ta=ma}else if((U|0)==97){k[30835]=113;ta=oa}else if((U|0)==106){k[30835]=113;ta=qa}if(ta){H=ta;do{H=H+-1|0;K=k[30828]|0;do if((k[K+(H*124|0)+84>>2]|0)==262){J=k[k[K+(H*124|0)+120>>2]>>2]|0;if(!J){k[v>>2]=745229;k[v+4>>2]=1488;Rw(745558,v)|0;break}else{dy(J);break}}while(0);I=K+(H*124|0)+116|0;b=k[I>>2]|0;if(!b){k[u>>2]=745229;k[u+4>>2]=1489;Rw(745609,u)|0}else dy(b);k[I>>2]=0}while((H|0)!=0)}}while(0);g:do if(!y)ua=c;else{u=c;v=0;ta=y;while(1){qa=(ta|0)<2500?ta:2500;Eu(u,qa);if(k[30835]|0){ua=u;break g}oa=u+qa|0;ma=ta;ta=ta-qa|0;ja=k[30820]|0;Z=k[30823]|0;X=ja+(Z*344|0)|0;v=(k[X>>2]|0)==-1e3?1:v;W=ja+(Z*344|0)+52|0;h:do switch(k[W>>2]|0){case 260:case 259:case 258:{do if(!v){Y=k[30042]|0;T=ja+(Z*344|0)+56|0;S=k[T>>2]|0;i:do if((Y|0)!=(S|0)){if((S|0)==1){if((qa|0)<=0)break;R=ja+(Z*344|0)+88|0;e=ja+(Z*344|0)+84|0;z=(ma|0)<2500?ma:2500;w=Y;sa=0;while(1){k[C>>2]=0;if((w|0)>0){f=w;pa=0;while(1){na=(k[R>>2]|0)+(ha(k[30044]|0,sa)|0)|0;ka=(ha(f,sa)|0)+pa|0;gk(k[30833]|0,na,(k[e>>2]|0)+sa|0,1,k[k[30034]>>2]|0,k[30036]|0,(k[30037]|0)+(ha(k[30043]|0,ka)|0)|0,B,123340)|0;pa=(k[C>>2]|0)+1|0;k[C>>2]=pa;ka=k[30042]|0;if((pa|0)>=(ka|0)){va=ka;break}else f=ka}}else va=w;sa=sa+1|0;if((sa|0)==(z|0))break i;else w=va}}w=(Y|0)<(S|0)?Y:S;if((qa|0)>0){z=ja+(Z*344|0)+88|0;sa=ja+(Z*344|0)+84|0;e=(ma|0)<2500?ma:2500;R=S;f=Y;pa=0;while(1){ka=ha(R,pa)|0;na=(k[z>>2]|0)+(ha(ka,k[30044]|0)|0)|0;la=ha(f,pa)|0;gk(k[30833]|0,na,(k[sa>>2]|0)+ka|0,w,k[k[30034]>>2]|0,k[30036]|0,(k[30037]|0)+(ha(la,k[30043]|0)|0)|0,B,123340)|0;la=k[30042]|0;if((la|0)>(w|0)){ka=(ha(la,pa)|0)+w|0;na=k[30043]|0;sy((k[30037]|0)+(ha(na,ka)|0)|0,0,ha(na,la-w|0)|0)|0}la=pa+1|0;if((la|0)==(e|0))break i;R=k[T>>2]|0;f=k[30042]|0;pa=la}}}else{pa=ha(Y,qa)|0;gk(k[30833]|0,k[ja+(Z*344|0)+88>>2]|0,k[ja+(Z*344|0)+84>>2]|0,pa,k[k[30034]>>2]|0,k[30036]|0,k[30037]|0,B,123340)|0}while(0);if((k[X>>2]|0)>0){Y=k[ja+(Z*344|0)+88>>2]|0;if(!Y){k[t>>2]=745229;k[t+4>>2]=1210;Rw(745648,t)|0;break}else{dy(Y);break}}}else{i[D>>0]=0;if((qa|0)>0){Y=ja+(Z*344|0)+88|0;T=ja+(Z*344|0)+56|0;S=(ma|0)<2500?ma:2500;aa=k[30042]|0;pa=0;while(1){k[C>>2]=0;if((aa|0)>0){f=aa;R=0;while(1){e=(ha(f,pa)|0)+R|0;gk(k[30833]|0,Y,D,k[T>>2]|0,k[k[30034]>>2]|0,k[30036]|0,(k[30037]|0)+(ha(k[30043]|0,e)|0)|0,B,123340)|0;R=(k[C>>2]|0)+1|0;k[C>>2]=R;e=k[30042]|0;if((R|0)>=(e|0)){wa=e;break}else f=e}}else wa=aa;pa=pa+1|0;if((pa|0)==(S|0))break;else aa=wa}}}while(0);aa=k[30835]|0;if((aa|0)==-11){k[30835]=412;tk(745695);U=213}else xa=aa;break}case 262:{j:do switch(k[k[30034]>>2]|0){case 11:{if((qa|0)>0){aa=ja+(Z*344|0)+56|0;S=(v|0)==0;pa=ja+(Z*344|0)+88|0;T=(ma|0)<2500?ma:2500;Y=-1;f=0;while(1){k[C>>2]=0;if((k[aa>>2]|0)>0)if(S){R=Y;e=0;while(1){if(!(e&7)){w=R+1|0;i[(k[30037]|0)+w>>0]=0;ya=k[C>>2]|0;za=w}else{ya=e;za=R}if((i[(k[(k[pa>>2]|0)+(f<<2)>>2]|0)+ya>>0]|0)==49){w=(k[30037]|0)+za|0;i[w>>0]=l[w>>0]|128>>>(ya&7);Aa=k[C>>2]|0}else Aa=ya;e=Aa+1|0;k[C>>2]=e;if((e|0)>=(k[aa>>2]|0)){Ba=za;break}else R=za}}else{R=Y;e=0;while(1){if(!(e&7)){w=R+1|0;i[(k[30037]|0)+w>>0]=0;Ca=k[C>>2]|0;Da=w}else{Ca=e;Da=R}if((i[pa+Ca>>0]|0)==49){w=(k[30037]|0)+Da|0;i[w>>0]=l[w>>0]|128>>>(Ca&7);Ea=k[C>>2]|0}else Ea=Ca;e=Ea+1|0;k[C>>2]=e;if((e|0)>=(k[aa>>2]|0)){Ba=Da;break}else R=Da}}else Ba=Y;f=f+1|0;if((f|0)==(T|0))break;else Y=Ba}}break}case 14:case 1:{Y=(qa|0)>0;if(!v){if(!Y)break j;T=ja+(Z*344|0)+56|0;f=ja+(Z*344|0)+88|0;aa=(ma|0)<2500?ma:2500;pa=0;do{k[C>>2]=0;S=k[T>>2]|0;if((S|0)>0){R=S;S=0;do{e=(ha(R,pa)|0)+S|0;i[(k[30037]|0)+e>>0]=(i[(k[(k[f>>2]|0)+(pa<<2)>>2]|0)+S>>0]|0)==49&1;S=(k[C>>2]|0)+1|0;k[C>>2]=S;R=k[T>>2]|0}while((S|0)<(R|0))}pa=pa+1|0}while((pa|0)!=(aa|0))}else{if(!Y)break j;aa=ja+(Z*344|0)+56|0;pa=ja+(Z*344|0)+88|0;T=(ma|0)<2500?ma:2500;f=0;do{k[C>>2]=0;R=k[aa>>2]|0;if((R|0)>0){S=R;R=0;do{e=(ha(S,f)|0)+R|0;i[(k[30037]|0)+e>>0]=(i[pa+R>>0]|0)==49&1;R=(k[C>>2]|0)+1|0;k[C>>2]=R;S=k[aa>>2]|0}while((R|0)<(S|0))}f=f+1|0}while((f|0)!=(T|0))}break}case 16:{k[C>>2]=0;T=(qa|0)>0;if(!v){if(!T)break j;f=ja+(Z*344|0)+88|0;aa=0;do{ux(k[(k[30037]|0)+(aa<<2)>>2]|0,k[(k[f>>2]|0)+(aa<<2)>>2]|0)|0;aa=(k[C>>2]|0)+1|0;k[C>>2]=aa}while((aa|0)<(qa|0))}else{if(!T)break j;aa=ja+(Z*344|0)+88|0;f=0;do{ux(k[(k[30037]|0)+(f<<2)>>2]|0,aa)|0;f=(k[C>>2]|0)+1|0;k[C>>2]=f}while((f|0)<(qa|0))}break}default:{tk(745764);k[30835]=432}}while(0);if((k[X>>2]|0)>0){f=ja+(Z*344|0)+88|0;aa=k[k[f>>2]>>2]|0;if(!aa){k[s>>2]=745229;k[s+4>>2]=1270;Rw(745811,s)|0}else dy(aa);aa=k[f>>2]|0;if(!aa){k[q>>2]=745229;k[q+4>>2]=1271;Rw(745864,q)|0;U=213;break h}else{dy(aa);U=213;break h}}else U=213;break}case 261:{k:do if((k[k[30034]>>2]|0)==16){k[C>>2]=0;aa=(qa|0)>0;if(v){if(!aa)break;f=ja+(Z*344|0)+88|0;T=0;while(1){ux(k[(k[30037]|0)+(T<<2)>>2]|0,f)|0;T=(k[C>>2]|0)+1|0;k[C>>2]=T;if((T|0)>=(qa|0))break k}}if(aa){T=ja+(Z*344|0)+84|0;f=ja+(Z*344|0)+88|0;pa=0;do{if(!(i[(k[T>>2]|0)+pa>>0]|0))ux(k[(k[30037]|0)+(pa<<2)>>2]|0,k[(k[f>>2]|0)+(pa<<2)>>2]|0)|0;else{k[B>>2]=1;ux(k[(k[30037]|0)+(pa<<2)>>2]|0,k[k[30036]>>2]|0)|0}pa=(k[C>>2]|0)+1|0;k[C>>2]=pa}while((pa|0)<(qa|0))}}else{tk(745914);k[30835]=432}while(0);if((k[X>>2]|0)>0){pa=ja+(Z*344|0)+88|0;f=k[k[pa>>2]>>2]|0;if(!f){k[n>>2]=745229;k[n+4>>2]=1296;Rw(745811,n)|0}else dy(f);f=k[pa>>2]|0;if(!f){k[m>>2]=745229;k[m+4>>2]=1297;Rw(745864,m)|0;U=213;break h}else{dy(f);U=213;break h}}else U=213;break}default:U=213}while(0);if((U|0)==213){U=0;xa=k[30835]|0}if(xa){ua=oa;break g}X=k[W>>2]|0;do if((X|0)==262){K=k[30037]|0;f=k[30043]|0;if((k[k[30034]>>2]|0)==11){pa=K+(ha(ha(f,qa)|0,((k[ja+(Z*344|0)+56>>2]|0)+7|0)/8|0)|0)|0;k[30037]=pa;break}else{Fa=ha(f,qa)|0;Ga=K;U=222;break}}else{K=k[30037]|0;f=ha(k[30043]|0,qa)|0;if((X|0)==261)k[30037]=K+f;else{Fa=f;Ga=K;U=222}}while(0);if((U|0)==222){U=0;X=Ga+(ha(k[30042]|0,Fa)|0)|0;k[30037]=X}if((ma|0)==(qa|0)){ua=oa;break g}else u=oa}}while(0);Fa=k[30034]|0;do if(!(k[B>>2]|0)){if(!(k[Fa+4>>2]|0)){Ga=k[30036]|0;if((k[Fa>>2]|0)==16){U=k[Ga>>2]|0;xa=j[60076]|0;i[U>>0]=xa;i[U+1>>0]=xa>>8;break}else{py(Ga|0,120152,k[30043]|0)|0;break}}}else k[Fa+16>>2]=1;while(0);if(((k[30834]|0)!=0?!((ua+-1|0)!=(k[30035]|0)|(k[30835]|0)!=0):0)?(k[(k[30034]|0)+12>>2]|0)<(a|0):0){G=-1;r=h;return G|0}G=k[30835]|0;r=h;return G|0}function dk(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0;e=r;r=r+80|0;f=e+16|0;g=e+52|0;h=e+44|0;j=e+8|0;l=e;m=k[d>>2]|0;if(m){n=m;r=e;return n|0}if(ak(a,0,c,5,g,h,e+48|0,e+24|0,d)|0){bk();n=k[d>>2]|0;r=e;return n|0}c=k[h>>2]|0;if((c|0)<0){m=0-c|0;k[h>>2]=m;o=m;p=1}else{o=c;p=0}if((k[g>>2]|0)!=14|(o|0)!=1){bk();tk(745034);k[d>>2]=432;n=432;r=e;return n|0}o=k[a>>2]|0;c=a+4|0;if((o|0)!=(k[(k[c>>2]|0)+68>>2]|0))Fk(a,o+1|0,0,d)|0;if(k[d>>2]|0){bk();n=k[d>>2]|0;r=e;return n|0}o=k[c>>2]|0;m=my(0,k[o+952>>2]|0,32)|0;h=L;q=o+944|0;s=k[q>>2]|0;t=k[q+4>>2]|0;q=o+976|0;o=k[q>>2]|0;u=k[q+4>>2]|0;if((s|0)==0&(t|0)==0){bk();n=k[d>>2]|0;r=e;return n|0}q=k[b>>2]|0;v=b+4|0;w=k[v>>2]|0;if((q|0)==(k[w+68>>2]|0))x=w;else{Fk(b,q+1|0,0,d)|0;x=k[v>>2]|0}if((k[x+120+4>>2]|0)<0)Vk(b,d)|0;if(k[d>>2]|0){bk();n=k[d>>2]|0;r=e;return n|0}x=k[v>>2]|0;q=my(0,k[x+952>>2]|0,32)|0;w=x+944|0;y=k[w>>2]|0;z=k[w+4>>2]|0;w=x+976|0;if((y|0)==0&(z|0)==0){x=w;k[x>>2]=0;k[x+4>>2]=0;A=0;B=0}else{x=w;A=k[x>>2]|0;B=k[x+4>>2]|0}if(!((m|0)==(q|0)&(h|0)==(L|0))){tk(745964);bk();k[d>>2]=436;n=436;r=e;return n|0}h=ky(s|0,t|0,1,0)|0;q=L;x=cy(h)|0;w=g+4|0;k[w>>2]=x;k[g+8>>2]=0;k[g+12>>2]=s;if(!x){tk(746013);bk();k[d>>2]=113;n=113;r=e;return n|0}i[x+s>>0]=0;if(!p){Pq(k[30826]|0,k[30827]|0,0,0,1,g,d)|0;if((t|0)>0|(t|0)==0&s>>>0>0){g=k[w>>2]|0;p=0;C=0;D=0;while(1){E=((i[g+p>>0]|0)!=0&1)+D|0;p=ky(p|0,C|0,1,0)|0;C=L;if(!((C|0)<(t|0)|(C|0)==(t|0)&p>>>0<s>>>0)){F=E;break}else D=E}}else F=0}else{D=i[(k[30820]|0)+((k[30823]|0)*344|0)+88>>0]|0;if((t|0)>0|(t|0)==0&s>>>0>0?(i[x>>0]=D,(t|0)>0|(t|0)==0&s>>>0>1):0){x=1;p=0;do{i[(k[w>>2]|0)+x>>0]=D;x=ky(x|0,p|0,1,0)|0;p=L}while((p|0)<(t|0)|(p|0)==(t|0)&x>>>0<s>>>0)}F=D<<24>>24!=0?s:0}if(!(k[d>>2]|0)){D=cy((m|0)<5e5?5e5:m)|0;if(!D){bk();k[d>>2]=113;n=113;r=e;return n|0}x=5e5/(m|0)|0;p=(x|0)>1?x:1;x=(a|0)==(b|0);if(!x){C=ky(y|0,z|0,1,0)|0;if((C|0)>1){Nj(b,y,z,F,((F|0)<0)<<31>>31,d)|0;G=1;H=C}else{G=1;H=C}}else{C=k[w>>2]|0;g=1;while(1)if(!(i[C+(g+-1)>>0]|0)){G=g;H=g;break}else g=g+1|0}g=((m|0)<0)<<31>>31;C=ha(p,m)|0;E=((C|0)<0)<<31>>31;I=G;G=0;J=H;while(1){do if(!(i[(k[w>>2]|0)+(I+-1)>>0]|0)){K=G;M=J}else{ku(a,I,((I|0)<0)<<31>>31,1,0,m,g,D+(ha(G,m)|0)|0,d)|0;H=G+1|0;if((H|0)!=(p|0)){K=H;M=J;break}ru(b,J,((J|0)<0)<<31>>31,1,0,C,E,D,d)|0;K=0;M=p+J|0}while(0);I=I+1|0;H=((I|0)<0)<<31>>31;if((H|0)>(t|0)|(H|0)==(t|0)&I>>>0>s>>>0|(k[d>>2]|0)!=0){N=K;O=M;break}else{G=K;J=M}}if(!N)P=O;else{M=ha(N,m)|0;ru(b,O,((O|0)<0)<<31>>31,1,0,M,((M|0)<0)<<31>>31,D,d)|0;P=N+O|0}do if(x){O=((P|0)<0)<<31>>31;if((t|0)<(O|0)|(t|0)==(O|0)&s>>>0<P>>>0)break;N=jy(h|0,q|0,P|0,O|0)|0;Pj(a,P,O,N,L,d)|0}else{if(!(((o|0)!=0|(u|0)!=0)&(F|0)!=0))break;N=k[b>>2]|0;O=k[v>>2]|0;if((N|0)==(k[O+68>>2]|0))Q=O;else{Fk(b,N+1|0,0,d)|0;Q=k[v>>2]|0}N=Q+120|0;O=k[N>>2]|0;M=k[N+4>>2]|0;N=Q+968|0;m=k[N>>2]|0;J=k[N+4>>2]|0;N=ky(m|0,J|0,A|0,B|0)|0;K=L;G=ky(N|0,K|0,2879,0)|0;I=L;p=xy(G|0,I|0,2880,0)|0;E=ky(N|0,K|0,p|0,L|0)|0;p=jy(G|0,I|0,E|0,L|0)|0;E=my(0,p|0,32)|0;p=L;if((u|0)>(p|0)|(u|0)==(p|0)&o>>>0>E>>>0){I=ky(o|0,u|0,2879,0)|0;G=jy(I|0,L|0,E|0,p|0)|0;p=wy(G|0,L|0,2880,0)|0;rl(b,p,1,d)|0}p=ky(A|0,B|0,o|0,u|0)|0;jq(b,782652,p,L,0,d)|0;p=k[a>>2]|0;G=k[c>>2]|0;if((p|0)==(k[G+68>>2]|0))R=G;else{Fk(a,p+1|0,0,d)|0;R=k[c>>2]|0}p=ky(O|0,M|0,A|0,B|0)|0;M=ky(p|0,L|0,m|0,J|0)|0;J=L;m=R+968|0;p=R+120|0;O=ky(k[m>>2]|0,k[m+4>>2]|0,k[p>>2]|0,k[p+4>>2]|0)|0;p=u;m=o;G=O;O=L;E=M;M=J;while(1){if(k[d>>2]|0)break;J=(p|0)<0|(p|0)==0&m>>>0<5e5;au(a,G,O,0,d)|0;I=my(0,m|0,32)|0;K=J?I:5e5;I=J?L:0;fu(a,K,I,D,d)|0;au(b,E,M,1,d)|0;cu(b,K,I,D,d)|0;G=ky(K|0,I|0,G|0,O|0)|0;O=L;E=ky(K|0,I|0,E|0,M|0)|0;M=L;J=m;m=jy(m|0,p|0,K|0,I|0)|0;if((J|0)==(K|0)&(p|0)==(I|0))break;else p=L}if((A|0)==0&(B|0)==0)break;p=k[v>>2]|0;if((k[p+928>>2]|0)<1)break;m=ky(F|0,((F|0)<0)<<31>>31|0,y|0,z|0)|0;M=L;E=(F|0)<1;O=p;p=1;while(1){if((k[(k[O+960>>2]|0)+((p+-1|0)*160|0)+80>>2]|0)>-1|E)S=O;else{G=y;I=z;do{G=ky(G|0,I|0,1,0)|0;I=L;pl(b,p,G,I,j,l,d)|0;K=l;J=ky(k[K>>2]|0,k[K+4>>2]|0,A|0,B|0)|0;K=L;N=l;k[N>>2]=J;k[N+4>>2]=K;N=j;ql(b,p,G,I,k[N>>2]|0,k[N+4>>2]|0,J,K,d)|0}while((I|0)<(M|0)|(I|0)==(M|0)&G>>>0<m>>>0);S=k[v>>2]|0}if((p|0)<(k[S+928>>2]|0)){O=S;p=p+1|0}else break}}while(0);dy(D)}D=k[w>>2]|0;if(!D){k[f>>2]=745229;k[f+4>>2]=412;Rw(746057,f)|0}else dy(D);bk();tl(b,d)|0;n=k[d>>2]|0;r=e;return n|0}function ek(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0;g=r;r=r+16|0;h=g+4|0;i=g;k[h>>2]=1;k[i>>2]=2147483647;j=fk(a,b,c,d,e,1,h,i,f)|0;r=g;return j|0}function fk(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;l=r;r=r+224|0;m=l+16|0;n=l+8|0;o=l;q=l+92|0;s=l+88|0;t=l+84|0;u=l+80|0;v=l+60|0;w=l+56|0;x=l+52|0;y=l+48|0;z=l+130|0;A=l+32|0;B=l+121|0;C=l+112|0;D=l+24|0;E=k[j>>2]|0;if(E){F=E;r=l;return F|0}if(ak(a,0,b,5,q,u,s,v,j)|0){bk();F=k[j>>2]|0;r=l;return F|0}b=k[u>>2]|0;if((b|0)<0){E=0-b|0;k[u>>2]=E;G=E;H=1}else{G=b;H=0}k[y>>2]=0;a:do if((Tk(c,0,d,y,j)|0)==219){k[j>>2]=0;b=(H|0)!=0;do if((i[d>>0]|0)!=35){if(!b){k[y>>2]=-1;I=d;break}if((Go(c,d,z,j)|0)==202){k[y>>2]=-1;I=d;break}if(!(k[j>>2]|0)){J=d;K=20}else{bk();F=k[j>>2]|0;r=l;return F|0}}else{if(!b){bk();tk(746094);k[j>>2]=432;F=432;r=l;return F|0}E=d+1|0;if(!(px(E,782840)|0)){if((k[q>>2]|0)==16){J=E;K=20;break}}else{L=(px(E,782848)|0)!=0;if(L|(k[q>>2]|0)==16){J=E;K=20;break}}bk();tk(746142);k[j>>2]=432;F=432;r=l;return F|0}while(0);if((K|0)==20){b=k[y>>2]|0;if((b|0)<0)I=J;else{M=J;N=e;O=b;P=0;break}}k[j>>2]=0;$k(c,y,j)|0;k[y>>2]=(k[y>>2]|0)+1;if((e|0)!=0?(b=i[e>>0]|0,b<<24>>24!=0):0)if(((b<<24>>24)+-48|0)>>>0>9&(k[30834]|0)==2){if(b<<24>>24==66&(k[q>>2]|0)==1){b=(G+7|0)/8|0;k[u>>2]=b;Q=b}else Q=G;k[m>>2]=Q;k[m+4>>2]=e;Xw(A,746250,m)|0;R=A}else R=e;else{b:do if((k[30834]|0)==2){k[o>>2]=G;Xw(A,773048,o)|0;switch(k[q>>2]|0){case 14:{b=A+(xx(A)|0)|0;i[b>>0]=76;i[b+1>>0]=0;break b;break}case 41:{b=A+(xx(A)|0)|0;i[b>>0]=74;i[b+1>>0]=0;break b;break}case 82:{b=A+(xx(A)|0)|0;i[b>>0]=68;i[b+1>>0]=0;break b;break}case 16:{b=A+(xx(A)|0)|0;i[b>>0]=65;i[b+1>>0]=0;break b;break}case 1:{b=A+(xx(A)|0)|0;i[b>>0]=88;i[b+1>>0]=0;break b;break}case 81:{b=A+(xx(A)|0)|0;i[b>>0]=75;i[b+1>>0]=0;break b;break}default:break b}}else switch(k[q>>2]|0){case 14:{bk();tk(746194);k[j>>2]=227;F=227;r=l;return F|0}case 41:{k[A>>2]=3223881;break b;break}case 82:{i[A>>0]=i[746238]|0;i[A+1>>0]=i[746239]|0;i[A+2>>0]=i[746240]|0;i[A+3>>0]=i[746241]|0;i[A+4>>0]=i[746242]|0;i[A+5>>0]=i[746243]|0;i[A+6>>0]=i[746244]|0;break b;break}case 1:case 16:{k[n>>2]=G;Xw(A,746245,n)|0;break b;break}default:break b}while(0);R=A}Qj(c,k[y>>2]|0,I,R,j)|0;b=k[s>>2]|0;if((b|0)>1)ys(c,k[y>>2]|0,b,v,j)|0;Gk(770576,k[y>>2]|0,B,j)|0;if((Go(c,B,z,j)|0)==202){k[j>>2]=0;switch(k[30834]|0){case 2:break;case 1:{fs(c,B,760184,746267,j)|0;Js(c,k[y>>2]|0,760184,j)|0;S=I;T=R;U=1;K=61;break a;break}default:{S=I;T=R;U=0;K=61;break a}}Qk(R,t,w,x,j)|0;b=k[t>>2]|0;switch(b|0){case 11:{V=255;W=0;break}case 21:{V=-32768;W=-1;break}case 31:{V=-2147483648;W=-1;break}case 41:{V=-2147483648;W=-1;break}default:if((b|0)==81){V=0;W=-2147483648}else{S=I;T=R;U=0;K=61;break a}}gs(c,B,V,W,746256,j)|0;Is(c,k[y>>2]|0,V,W,j)|0;S=I;T=R;U=1;K=61}else{S=I;T=R;U=0;K=61}}else{if(k[j>>2]|0){bk();F=k[j>>2]|0;r=l;return F|0}Gk(774793,k[y>>2]|0,C,j)|0;Go(c,C,z,j)|0;b=k[j>>2]|0;switch(b|0){case 0:{ys(c,k[y>>2]|0,k[s>>2]|0,v,j)|0;K=58;break}case 202:{k[j>>2]=0;xk();E=k[s>>2]|0;if((E|0)>1){ys(c,k[y>>2]|0,E,v,j)|0;K=58}else K=58;break}default:X=b}if((K|0)==58)X=k[j>>2]|0;if(!X){S=d;T=e;U=0;K=61}else{bk();F=k[j>>2]|0;r=l;return F|0}}while(0);if((K|0)==61){M=S;N=T;O=k[y>>2]|0;P=U}c:do if((O|0)>0){Ko(a,790243,D,0,j)|0;U=k[30826]|0;if(mk(U,j)|0){bk();F=k[j>>2]|0;r=l;return F|0}Kq((k[30827]|0)+(U*244|0)|0,c,k[y>>2]|0,0,2)|0;k[30826]=(k[30826]|0)+1;do if((f|0)>0){U=q+4|0;T=q+12|0;S=(f|0)==1;e=q+16|0;d=0;X=0;while(1){k[U>>2]=0;v=k[g+(X<<2)>>2]|0;s=(k[h+(X<<2)>>2]|0)-v|0;z=s+1|0;k[T>>2]=z;if((S&(s|0)>8?(k[g>>2]|0)==1:0)?(k[h>>2]|0)==(k[D>>2]|0):0)Y=0;else Y=z;if((Pq(k[30826]|0,k[30827]|0,v+-1|0,Y,1,q,j)|0)!=-1){if(k[j>>2]|0)break}else k[j>>2]=0;v=(k[e>>2]|0)==0?d:1;X=X+1|0;if((X|0)>=(f|0)){Z=v;K=76;break}else d=v}if((K|0)==76){_=(Z|0)!=0;break}bk();F=k[j>>2]|0;r=l;return F|0}else _=0;while(0);if(!((P|0)==0|_))Aq(c,B,j)|0}else{d=k[30820]|0;X=k[30823]|0;switch(k[q>>2]|0){case 82:{lq(c,M,+p[d+(X*344|0)+88>>3],15,N,j)|0;break c;break}case 41:{e=k[d+(X*344|0)+88>>2]|0;jq(c,M,e,((e|0)<0)<<31>>31,N,j)|0;break c;break}case 14:{kq(c,M,i[d+(X*344|0)+88>>0]|0,N,j)|0;break c;break}case 16:case 1:{if(!(px(M,782840)|0)){xs(c,d+(X*344|0)+88|0,j)|0;break c}e=d+(X*344|0)+88|0;if(!(px(M,782848)|0)){rs(c,e,j)|0;break c}else{iq(c,M,e,N,j)|0;break c}break}default:break c}}while(0);bk();F=k[j>>2]|0;r=l;return F|0}function gk(a,b,c,d,e,f,g,h,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;var n=0,q=0,r=0,s=0;switch(e|0){case 14:{switch(a|0){case 11:case 14:{if((d|0)>0){e=0;do{i[g+e>>0]=(i[b+e>>0]|0)!=0&1;e=e+1|0}while((e|0)!=(d|0))}else{n=k[m>>2]|0;return n|0}break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}e=0;do{i[g+e>>0]=(j[b+(e<<1)>>1]|0)!=0&1;e=e+1|0}while((e|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}e=0;do{i[g+e>>0]=(k[b+(e<<2)>>2]|0)!=0&1;e=e+1|0}while((e|0)!=(d|0));break}case 42:{if((d|0)<=0){n=k[m>>2]|0;return n|0}e=0;do{i[g+e>>0]=+o[b+(e<<2)>>2]!=0.0&1;e=e+1|0}while((e|0)!=(d|0));break}case 82:{if((d|0)<=0){n=k[m>>2]|0;return n|0}e=0;do{i[g+e>>0]=+p[b+(e<<3)>>3]!=0.0&1;e=e+1|0}while((e|0)!=(d|0));break}default:k[m>>2]=410}if((d|0)>0)q=0;else{n=k[m>>2]|0;return n|0}do{if(i[c+q>>0]|0){i[g+q>>0]=i[f>>0]|0;k[h>>2]=1}q=q+1|0}while((q|0)!=(d|0));n=k[m>>2]|0;return n|0}case 11:{switch(a|0){case 11:case 14:{if((d|0)>0){q=0;do{i[g+q>>0]=i[b+q>>0]|0;q=q+1|0}while((q|0)!=(d|0))}else{n=k[m>>2]|0;return n|0}break}case 21:{gm(b,d,1.0,0.0,0,0,0,0,0,g,m)|0;break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}q=0;do{do if(!(i[c+q>>0]|0)){e=k[b+(q<<2)>>2]|0;if((e|0)<0){k[m>>2]=-11;i[g+q>>0]=0;break}if((e|0)>255){k[m>>2]=-11;i[g+q>>0]=-1;break}else{i[g+q>>0]=e;break}}else{i[g+q>>0]=i[f>>0]|0;k[h>>2]=1}while(0);q=q+1|0}while((q|0)!=(d|0));n=k[m>>2]|0;return n|0}case 42:{jm(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}case 82:{km(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)>0)r=0;else{n=k[m>>2]|0;return n|0}do{if(i[c+r>>0]|0){i[g+r>>0]=i[f>>0]|0;k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 21:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{j[g+(r<<1)>>1]=l[b+r>>0]|0;r=r+1|0}while((r|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{j[g+(r<<1)>>1]=j[b+(r<<1)>>1]|0;r=r+1|0}while((r|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{do if(!(i[c+r>>0]|0)){q=k[b+(r<<2)>>2]|0;if((q|0)<-32768){k[m>>2]=-11;j[g+(r<<1)>>1]=-32768;break}if((q|0)>32767){k[m>>2]=-11;j[g+(r<<1)>>1]=32767;break}else{j[g+(r<<1)>>1]=q;break}}else{j[g+(r<<1)>>1]=j[f>>1]|0;k[h>>2]=1}while(0);r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 42:{Um(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}case 82:{Vm(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{if(i[c+r>>0]|0){j[g+(r<<1)>>1]=j[f>>1]|0;k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 31:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=l[b+r>>0];r=r+1|0}while((r|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=j[b+(r<<1)>>1];r=r+1|0}while((r|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=k[b+(r<<2)>>2];r=r+1|0}while((r|0)!=(d|0));break}case 42:{Dn(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}case 82:{En(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{if(i[c+r>>0]|0){k[g+(r<<2)>>2]=k[f>>2];k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 41:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=l[b+r>>0];r=r+1|0}while((r|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=j[b+(r<<1)>>1];r=r+1|0}while((r|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=k[b+(r<<2)>>2];r=r+1|0}while((r|0)!=(d|0));break}case 42:{fn(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}case 82:{gn(b,d,1.0,0.0,0,0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{if(i[c+r>>0]|0){k[g+(r<<2)>>2]=k[f>>2];k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 81:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{q=g+(r<<3)|0;k[q>>2]=l[b+r>>0];k[q+4>>2]=0;r=r+1|0}while((r|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{q=j[b+(r<<1)>>1]|0;e=g+(r<<3)|0;k[e>>2]=q;k[e+4>>2]=((q|0)<0)<<31>>31;r=r+1|0}while((r|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{q=k[b+(r<<2)>>2]|0;e=g+(r<<3)|0;k[e>>2]=q;k[e+4>>2]=((q|0)<0)<<31>>31;r=r+1|0}while((r|0)!=(d|0));break}case 42:{sn(b,d,1.0,0.0,0,0,0,0,0,g,m)|0;break}case 82:{tn(b,d,1.0,0.0,0,0,0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{if(i[c+r>>0]|0){q=f;e=k[q+4>>2]|0;s=g+(r<<3)|0;k[s>>2]=k[q>>2];k[s+4>>2]=e;k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 42:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{o[g+(r<<2)>>2]=+(l[b+r>>0]|0);r=r+1|0}while((r|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{o[g+(r<<2)>>2]=+(j[b+(r<<1)>>1]|0);r=r+1|0}while((r|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{o[g+(r<<2)>>2]=+(k[b+(r<<2)>>2]|0);r=r+1|0}while((r|0)!=(d|0));break}case 42:{if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{k[g+(r<<2)>>2]=k[b+(r<<2)>>2];r=r+1|0}while((r|0)!=(d|0));break}case 82:{Km(b,d,1.0,0.0,0,0.0,0,0,g,m)|0;break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}r=0;do{if(i[c+r>>0]|0){k[g+(r<<2)>>2]=k[f>>2];k[h>>2]=1}r=r+1|0}while((r|0)!=(d|0));n=k[m>>2]|0;return n|0}case 82:{switch(a|0){case 11:case 14:{if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{p[g+(a<<3)>>3]=+(l[b+a>>0]|0);a=a+1|0}while((a|0)!=(d|0));break}case 21:{if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{p[g+(a<<3)>>3]=+(j[b+(a<<1)>>1]|0);a=a+1|0}while((a|0)!=(d|0));break}case 41:{if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{p[g+(a<<3)>>3]=+(k[b+(a<<2)>>2]|0);a=a+1|0}while((a|0)!=(d|0));break}case 42:{if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{p[g+(a<<3)>>3]=+o[b+(a<<2)>>2];a=a+1|0}while((a|0)!=(d|0));break}case 82:{if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{p[g+(a<<3)>>3]=+p[b+(a<<3)>>3];a=a+1|0}while((a|0)!=(d|0));break}default:k[m>>2]=410}if((d|0)<=0){n=k[m>>2]|0;return n|0}a=0;do{if(i[c+a>>0]|0){p[g+(a<<3)>>3]=+p[f>>3];k[h>>2]=1}a=a+1|0}while((a|0)!=(d|0));n=k[m>>2]|0;return n|0}default:{k[m>>2]=410;n=k[m>>2]|0;return n|0}}return 0}function hk(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0;e=r;r=r+32|0;f=e+24|0;g=e+20|0;h=k[d>>2]|0;if(h){j=h;r=e;return j|0}if(ak(a,0,b,5,f,g,e+28|0,e,d)|0){bk();j=k[d>>2]|0;r=e;return j|0}b=k[g>>2]|0;if((b|0)<0){h=0-b|0;k[g>>2]=h;l=h;m=1}else{l=b;m=0}if((k[f>>2]|0)!=14|(l|0)!=1){bk();tk(745034);k[d>>2]=432;j=432;r=e;return j|0}k[c>>2]=0;if(!m){if((Pq(k[30826]|0,k[30827]|0,0,0,2,c,d)|0)==-1)k[d>>2]=0}else if((i[(k[30820]|0)+((k[30823]|0)*344|0)+88>>0]|0)!=0?(al(a,g,d)|0,(k[g>>2]|0)!=0):0)k[c>>2]=1;bk();j=k[d>>2]|0;r=e;return j|0}function ik(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0;Eu(c,d);f=k[30835]|0;if(f){h=f;return h|0}e=k[30820]|0;b=k[30823]|0;if((k[e+(b*344|0)>>2]|0)==-1e3){if(!(i[e+(b*344|0)+88>>0]|0)){h=0;return h|0}k[g>>2]=c;h=-1;return h|0}if((d|0)<=0){h=0;return h|0}a=k[e+(b*344|0)+88>>2]|0;j=e+(b*344|0)+84|0;b=0;while(1){if((i[a+b>>0]|0)!=0?(i[(k[j>>2]|0)+b>>0]|0)==0:0){l=b;break}b=b+1|0;if((b|0)>=(d|0)){h=f;m=11;break}}if((m|0)==11)return h|0;k[g>>2]=l+c;h=-1;return h|0}function jk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;c=r;r=r+496|0;d=c+80|0;e=c+72|0;f=c+64|0;g=c+56|0;h=c+48|0;j=c+40|0;l=c+32|0;m=c+24|0;n=c+16|0;o=c+8|0;q=c;s=c+136|0;t=c+132|0;u=c+128|0;v=c+124|0;w=c+104|0;x=c+100|0;y=c+240|0;z=c+96|0;A=c+92|0;B=c+156|0;C=c+88|0;D=c+84|0;k[s>>2]=0;k[s+4>>2]=0;k[s+8>>2]=0;k[s+12>>2]=0;k[s+16>>2]=0;k[x>>2]=789744;E=(jb(746285)|0)!=0;k[29972]=E&1;F=k[b>>2]|0;if(F){G=F;r=c;return G|0}F=a+8|0;H=k[F>>2]|0;if(!(((H|0)!=0?(I=k[H>>2]|0,(I|0)!=0):0)?(i[I>>0]|0)!=0:0))J=5;if((J|0)==5?(k[F>>2]=x,E):0){k[q>>2]=789744;Rw(746301,q)|0}q=k[k[a+12>>2]>>2]|0;E=a+28|0;x=k[E>>2]|0;k[30829]=a;a:do if(!(ak(q,0,k[a+16>>2]|0,5,s,v,t,w,b)|0)){F=k[v>>2]|0;if((F|0)<0)k[v>>2]=0-F;switch(k[s>>2]|0){case 14:{K=746427;J=17;break}case 41:{K=746435;J=17;break}case 82:{K=746440;J=17;break}case 16:{k[b>>2]=-1;tk(746325);J=14;break}case 1:{J=14;break}default:{k[b>>2]=-1;tk(746386);K=746451;J=17}}if((J|0)==14?(k[29972]|0)!=0:0){Tw(746364)|0;K=746447;J=17}if((J|0)==17?(k[29972]|0)!=0:0){F=k[s>>2]|0;k[o>>2]=K;k[o+4>>2]=F;Rw(746461,o)|0}if(!(k[b>>2]|0)){if(Dl(q,5,u,t,w,b)|0){tk(746485);break}if(k[29972]|0){k[n>>2]=k[u>>2];Rw(746537,n)|0}F=k[u>>2]|0;if((k[s>>2]|0)==82&(F|0)!=-64){k[u>>2]=-32;L=-32}else L=F;F=k[a+20>>2]|0;if(!F)M=L;else{k[u>>2]=F;M=F}if(!(k[29972]|0))N=M;else{k[m>>2]=M;Rw(746554,m)|0;N=k[u>>2]|0}if(Xr(x,N,k[t>>2]|0,w,b)|0){tk(746572);break}if(Co(q,z,A,b)|0){tk(746616);break}b:do if((k[z>>2]|0)>=1){F=1;c:while(1){if(Ro(q,F,B,b)|0){O=F;J=37;break}I=Ll(B)|0;do if((I|0)!=10?!((F|0)<12&(I|0)==130):0){if(((I|0)==40|(I|0)==30)&(k[u>>2]|0)<0)break;if(ds(x,B,b)|0){J=42;break c}}while(0);if((F|0)<(k[z>>2]|0))F=F+1|0;else break b}if((J|0)==37){k[l>>2]=O;Xw(y,746669,l)|0;tk(y)}else if((J|0)==42){F=k[b>>2]|0;k[j>>2]=B;k[j+4>>2]=F;Xw(y,746709,j)|0;tk(y)}break a}while(0);F=k[u>>2]|0;switch(F|0){case 8:{k[s>>2]=11;J=52;break}case 16:{k[s>>2]=21;J=52;break}case 32:{k[s>>2]=41;J=52;break}case -32:{k[s>>2]=42;P=0;break}case -64:{k[s>>2]=82;P=0;break}default:{k[h>>2]=F;Xw(y,746758,h)|0;tk(y);k[b>>2]=-1;break a}}if((J|0)==52){F=a+24|0;I=k[F>>2]|0;k[C>>2]=I;if(!I){k[D>>2]=0;d:do if(!(Ko(q,779663,C,0,D)|0))Q=0;else{H=k[u>>2]|0;switch(H|0){case 8:{k[C>>2]=255;Q=1;break d;break}case 16:{k[C>>2]=-32768;Q=1;break d;break}case 32:{k[C>>2]=-2147483648;Q=1;break d;break}default:{k[g>>2]=H;Rw(746801,g)|0;Q=1;break d}}}while(0);H=k[C>>2]|0;k[F>>2]=H;R=H;S=Q}else{R=I;S=0}Hs(x,R,((R|0)<0)<<31>>31,b)|0;if(!(k[29972]|0))P=S;else{k[f>>2]=k[C>>2];Rw(746838,f)|0;P=S}}H=a+32|0;if(i[H>>0]|0){T=a+107|0;U=k[30820]|0;V=k[30823]|0;W=k[s>>2]|0;switch(W|0){case 82:{lq(x,H,+p[U+(V*344|0)+88>>3],15,T,b)|0;break a;break}case 41:{X=k[U+(V*344|0)+88>>2]|0;jq(x,H,X,((X|0)<0)<<31>>31,T,b)|0;break a;break}case 14:{kq(x,H,i[U+(V*344|0)+88>>0]|0,T,b)|0;break a;break}case 16:case 1:{iq(x,H,U+(V*344|0)+88|0,T,b)|0;break a;break}default:{k[d>>2]=W;Xw(y,746997,d)|0;tk(y);break a}}}W=k[30826]|0;if(!(mk(W,b)|0)){k[30826]=(k[30826]|0)+1;T=k[30827]|0;V=T+(W*244|0)|0;U=k[E>>2]|0;k[V>>2]=U;k[T+(W*244|0)+84>>2]=2;nk(U,746855,k[u>>2]|0,(k[30828]|0)+(W*124|0)|0,V)|0;k[s+12>>2]=-1;if((Pq(k[30826]|0,k[30827]|0,0,0,1,s,b)|0)!=-1){if(k[b>>2]|0)break}else k[b>>2]=0;if(!(k[s+16>>2]|0)){if((k[u>>2]|0)<=0)break;if(!(Hs(x,-1234554321,-1,b)|0))break;tk(746959);break}if(P){V=k[a+24>>2]|0;jq(x,779663,V,((V|0)<0)<<31>>31,746863,b)|0;if(k[b>>2]|0)tk(746880);if(!(k[29972]|0))break;Tw(746924)|0;k[e>>2]=k[b>>2];Rw(746941,e)|0}}}}while(0);bk();G=k[b>>2]|0;r=c;return G|0}function kk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0.0,N=0;c=r;r=r+176|0;d=c+56|0;e=c+48|0;f=c+40|0;g=c+32|0;h=c+24|0;j=c+16|0;l=c+80|0;m=c+76|0;n=c+72|0;o=c+68|0;q=c+64|0;s=c+88|0;t=c+8|0;u=c;v=c+60|0;if(k[29972]|0){k[j>>2]=a;Rw(747269,j)|0}if((i[a>>0]|0)==35){w=ok(a+1|0,b)|0;r=c;return w|0}j=k[30810]|0;k[l>>2]=0;x=k[30826]|0;do if(k[30834]|0){do if(!(k[30813]|0)){if(!(Tk(j,0,a,m,l)|0)){y=k[m>>2]|0;break}z=k[l>>2]|0;if((z|0)!=219){k[30835]=z;w=-1;r=c;return w|0}z=ok(a,b)|0;if((z|0)==-1){w=-1;r=c;return w|0}xk();w=z;r=c;return w|0}else{z=k[30816]|0;k[m>>2]=z;y=z}while(0);if(Xk(j,y,n,o,q,l)|0){k[30835]=k[l>>2];w=-1;r=c;return w|0}if(!(mk(x,123340)|0)){z=k[30828]|0;A=k[30827]|0;Kq(A+(x*244|0)|0,j,k[m>>2]|0,0,0)|0;B=z;C=A;D=j;break}else{w=-1;r=c;return w|0}}else{A=k[30829]|0;if(!A){k[30835]=219;tk(747286);w=-1;r=c;return w|0}k[m>>2]=-1;z=k[A>>2]|0;if((z|0)>0){E=k[A+8>>2]|0;F=-1;G=z;z=0;while(1){if(!(px(a,k[E+(z<<2)>>2]|0)|0)){k[m>>2]=z;H=z;I=k[A>>2]|0}else{H=F;I=G}z=z+1|0;if((z|0)>=(I|0)){J=H;break}else{F=H;G=I}}if((J|0)>=0){if(mk(x,123340)|0){w=-1;r=c;return w|0}G=k[30828]|0;F=k[30827]|0;z=F+(x*244|0)|0;A=k[(k[(k[30829]|0)+12>>2]|0)+(J<<2)>>2]|0;Dl(A,5,n,G+(x*124|0)+92|0,G+(x*124|0)+96|0,l)|0;k[G+(x*124|0)+88>>2]=1;if(!(nk(A,a,k[n>>2]|0,G+(x*124|0)|0,z)|0)){k[z>>2]=A;k[F+(x*244|0)+84>>2]=0;B=G;C=F;D=A;break}else{w=-1;r=c;return w|0}}}k[h>>2]=a;Xw(s,747328,h)|0;tk(s);k[30835]=219;w=-1;r=c;return w|0}while(0);Bx(B+(x*124|0)|0,a,80)|0;i[B+(x*124|0)+80>>0]=0;a=k[30834]|0;do if(a){h=k[n>>2]|0;a:do switch(h|0){case 1:{k[B+(x*124|0)+84>>2]=262;k[C+(x*244|0)+80>>2]=11;K=271;break}case 41:case 21:case 11:{k[g>>2]=k[m>>2];Xw(s,747370,g)|0;k[v>>2]=0;if(Io(D,82,s,t,0,v)|0)p[t>>3]=0.0;k[f>>2]=k[m>>2];Xw(s,747378,f)|0;k[v>>2]=0;if(!(Io(D,82,s,u,0,v)|0)){if(+p[u>>3]==1.0)L=38}else{p[u>>3]=1.0;L=38}if((L|0)==38?(M=+p[t>>3],M==0.0|M==32768.0):0){k[B+(x*124|0)+84>>2]=259;k[C+(x*244|0)+80>>2]=41;K=268;break a}k[B+(x*124|0)+84>>2]=260;k[C+(x*244|0)+80>>2]=82;K=268;break}case 82:case 42:case 81:{k[B+(x*124|0)+84>>2]=260;k[C+(x*244|0)+80>>2]=82;K=268;break}case 14:{k[B+(x*124|0)+84>>2]=258;k[C+(x*244|0)+80>>2]=14;K=269;break}case 16:{k[B+(x*124|0)+84>>2]=261;k[C+(x*244|0)+80>>2]=16;J=k[q>>2]|0;if((J|0)>255){k[e>>2]=k[m>>2];k[e+4>>2]=255;Xw(s,747386,e)|0;tk(s);k[30835]=433;w=-1;r=c;return w|0}else{if((a|0)!=1){K=270;break a}k[o>>2]=J;K=270;break a}break}default:{if((h|0)<0){k[d>>2]=h;Xw(s,747432,d)|0;tk(s)}k[30835]=432;w=-1;r=c;return w|0}}while(0);k[B+(x*124|0)+88>>2]=k[o>>2];if(!((k[o>>2]|0)>1&(k[n>>2]|0)!=16)){k[B+(x*124|0)+92>>2]=1;k[B+(x*124|0)+96>>2]=1;N=K;break}if(!(Wo(D,k[m>>2]|0,5,B+(x*124|0)+92|0,B+(x*124|0)+96|0,l)|0))N=K;else{k[30835]=k[l>>2];w=-1;r=c;return w|0}}else N=268;while(0);k[30826]=(k[30826]|0)+1;k[b>>2]=x;w=N;r=c;return w|0}function lk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;f=r;r=r+128|0;g=f+24|0;h=f+16|0;j=f;m=f+40|0;n=f+32|0;o=f+28|0;p=k[30827]|0;q=p+(a*244|0)|0;k[n>>2]=0;a:do if(!(k[30834]|0)){Zl(k[q>>2]|0,k[p+(a*244|0)+80>>2]|0,b,((b|0)<0)<<31>>31,c,((c|0)<0)<<31>>31,d,e,o,n)|0;if(k[29972]|0){s=k[n>>2]|0;k[j>>2]=b;k[j+4>>2]=c;k[j+8>>2]=s;Rw(747145,j)|0}}else{s=p+(a*244|0)+92|0;t=k[s>>2]|0;u=ha(t,c)|0;v=k[p+(a*244|0)+80>>2]|0;switch(v|0){case 11:{w=ha((t+7|0)/8|0,c)|0;t=cy(w)|0;em(k[q>>2]|0,k[p+(a*244|0)+4>>2]|0,b,((b|0)<0)<<31>>31,1,0,w,((w|0)<0)<<31>>31,0,t,o,n)|0;w=k[s>>2]|0;if((c|0)>0){s=(w+7|0)/8|0;x=(w|0)>0;y=0;do{if(x){z=d+(y<<2)|0;A=(ha(s,y)|0)+1|0;B=0;while(1){C=(B|0)%8|0;i[(k[z>>2]|0)+B>>0]=(1<<7-C&(l[t+A>>0]|0)|0)==0?48:49;B=B+1|0;if((B|0)==(w|0)){D=z;E=w;break}else A=((C|0)==7&1)+A|0}}else{D=d+(y<<2)|0;E=0}i[(k[D>>2]|0)+E>>0]=0;y=y+1|0}while((y|0)!=(c|0))}if(!t){k[h>>2]=745229;k[h+4>>2]=2559;Rw(747195,h)|0;break a}else{dy(t);break a}break}case 16:{Kn(k[q>>2]|0,k[p+(a*244|0)+4>>2]|0,b,((b|0)<0)<<31>>31,1,0,c,((c|0)<0)<<31>>31,d,e,o,n)|0;break a;break}case 14:{Gn(k[q>>2]|0,k[p+(a*244|0)+4>>2]|0,b,((b|0)<0)<<31>>31,1,0,u,((u|0)<0)<<31>>31,d,e,o,n)|0;break a;break}case 41:{an(k[q>>2]|0,k[p+(a*244|0)+4>>2]|0,b,((b|0)<0)<<31>>31,1,0,u,((u|0)<0)<<31>>31,d,e,o,n)|0;break a;break}case 82:{rm(k[q>>2]|0,k[p+(a*244|0)+4>>2]|0,b,((b|0)<0)<<31>>31,1,0,u,((u|0)<0)<<31>>31,d,e,o,n)|0;break a;break}default:{k[g>>2]=v;Xw(m,747233,g)|0;tk(m);break a}}}while(0);m=k[n>>2]|0;if(!m){F=0;r=f;return F|0}k[30835]=m;F=-1;r=f;return F|0}function mk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0;if(!((a|0)%25|0)){if(!a){c=cy(6100)|0;k[30827]=c;d=c;e=cy(3100)|0}else{c=a+25|0;f=fy(k[30827]|0,c*244|0)|0;k[30827]=f;f=fy(k[30828]|0,c*124|0)|0;d=k[30827]|0;e=f}k[30828]=e;f=(d|0)==0;if((e|0)==0|f){if(f)g=e;else{dy(d);g=k[30828]|0}if(g)dy(g);k[30827]=0;k[30828]=0;k[b>>2]=113;h=113;return h|0}else i=e}else i=k[30828]|0;k[i+(a*124|0)+120>>2]=0;k[i+(a*124|0)+116>>2]=0;h=0;return h|0}function nk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0.0,o=0,q=0.0,s=0;f=r;r=r+128|0;g=f+40|0;h=f+16|0;i=f+44|0;j=f+8|0;l=f;m=f+48|0;switch(c|0){case 32:case 16:case 8:{k[i>>2]=0;if(Io(a,82,779507,l,0,i)|0)p[l>>3]=0.0;k[i>>2]=0;if(!(Io(a,82,779536,j,0,i)|0)){n=+p[j>>3];if(n==1.0)o=7;else q=n}else{p[j>>3]=1.0;o=7}if((o|0)==7){n=+p[l>>3];if(n==0.0|n==32768.0){k[d+84>>2]=259;k[e+80>>2]=41;s=0;r=f;return s|0}else q=1.0}k[d+84>>2]=260;k[e+80>>2]=82;if(!(k[29972]|0)){s=0;r=f;return s|0}n=+p[l>>3];k[h>>2]=b;p[h+8>>3]=q;p[h+16>>3]=n;Rw(747049,h)|0;s=0;r=f;return s|0}case -64:case -32:case 64:{k[d+84>>2]=260;k[e+80>>2]=82;s=0;r=f;return s|0}default:{k[g>>2]=c;Xw(m,747092,g)|0;tk(m);k[30835]=432;s=432;r=f;return s|0}}return 0}function ok(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0;c=r;r=r+96|0;d=c+8|0;e=c+20|0;f=c+25|0;g=c+24|0;h=c;j=c+16|0;l=c+12|0;k[e>>2]=0;m=k[30810]|0;if(Qo(m,a,f,0,e)|0){n=k[e>>2]|0;if((n|0)==202){k[d>>2]=a;Xw(f,747495,d)|0;tk(f);o=k[e>>2]|0}else o=n;k[30835]=o;q=-1;r=c;return q|0}if(Ml(f,g,e)|0){k[30835]=k[e>>2];q=-1;r=c;return q|0}switch(i[g>>0]|0){case 67:{Jo(m,a,f,0,e)|0;ux(b,f)|0;s=261;break}case 76:{Lo(m,a,j,0,e)|0;i[b>>0]=k[j>>2];s=258;break}case 73:{Ko(m,a,l,0,e)|0;k[b>>2]=k[l>>2];s=259;break}case 70:{Mo(m,a,h,0,e)|0;p[b>>3]=+p[h>>3];s=260;break}default:s=-1}h=k[e>>2]|0;if(!h){q=s;r=c;return q|0}k[30835]=h;q=-1;r=c;return q|0}function pk(a){a=a|0;o[a>>2]=3.390000104904175;return 3.390000104904175}function qk(a,b,c){a=a|0;b=b|0;c=c|0;ux(b,k[(k[a+4>>2]|0)+12>>2]|0)|0;return k[c>>2]|0}function rk(a,b,c){a=a|0;b=b|0;c=c|0;k[b>>2]=k[(k[a+4>>2]|0)+76>>2];return k[c>>2]|0}function sk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;i[b>>0]=0;if(a>>>0>=300){if((a|0)>=600){c=b;d=749645;e=c+21|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}do switch(a|0){case 301:{c=b;d=749666;e=c+19|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 302:{c=b;d=749685;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 304:{c=b;d=749716;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 306:{c=b;d=749738;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 307:{c=b;d=749766;e=c+21|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 308:{c=b;d=749787;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 309:{c=b;d=749812;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 310:{c=b;d=749836;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 311:{c=b;d=749861;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 312:{c=b;d=749886;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 314:{c=b;d=749912;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 317:{c=b;d=749935;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 320:{c=b;d=749964;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 321:{c=b;d=749993;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 322:{c=b;d=750024;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 323:{c=b;d=750046;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 340:{c=b;d=750070;e=c+16|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 341:{c=b;d=750086;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 342:{c=b;d=750114;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 343:{c=b;d=750137;e=c+16|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 344:{c=b;d=750153;e=c+13|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 345:{c=b;d=750166;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 346:{c=b;d=750188;e=c+19|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 347:{c=b;d=750207;e=c+20|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 348:{c=b;d=750227;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 360:{c=b;d=750255;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 361:{c=b;d=750279;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 362:{c=b;d=750305;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 363:{c=b;d=750331;e=c+20|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 364:{c=b;d=750351;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 365:{c=b;d=750374;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 366:{c=b;d=750399;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 367:{c=b;d=750425;e=c+17|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 368:{c=b;d=750442;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 369:{c=b;d=750464;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 401:{c=b;d=750490;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 402:{c=b;d=750519;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 403:{c=b;d=750550;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 404:{c=b;d=750576;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 405:{c=b;d=750602;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 406:{c=b;d=750632;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 407:{c=b;d=750657;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 408:{c=b;d=750686;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 409:{c=b;d=750717;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 410:{c=b;d=750746;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 411:{c=b;d=750774;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 412:{c=b;d=750798;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 413:{c=b;d=750827;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 414:{c=b;d=750851;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 420:{c=b;d=750877;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 431:{c=b;d=750905;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 432:{c=b;d=750932;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 433:{c=b;d=750961;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 434:{c=b;d=750985;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 435:{c=b;d=751007;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 436:{c=b;d=751033;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 501:{c=b;d=751064;e=c+20|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 502:{c=b;d=751084;e=c+19|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 503:{c=b;d=751103;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 504:{c=b;d=751128;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 505:{c=b;d=751152;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}default:{c=b;d=749645;e=c+21|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}}while(0)}do switch(a|0){case 0:{c=b;d=747529;e=c+14|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 1:{c=b;d=747543;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 101:{c=b;d=747569;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 103:{c=b;d=747597;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 104:{c=b;d=747628;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 105:{c=b;d=747658;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 106:{c=b;d=747689;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 107:{c=b;d=747716;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 108:{c=b;d=747747;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 110:{c=b;d=747776;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 111:{c=b;d=747801;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 112:{c=b;d=747826;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 113:{c=b;d=747856;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 114:{c=b;d=747882;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 115:{c=b;d=747907;e=c+19|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 116:{c=b;d=747926;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 121:{c=b;d=747954;e=c+19|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 122:{c=b;d=747973;e=c+21|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 123:{c=b;d=747994;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 124:{c=b;d=748017;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 125:{c=b;d=748048;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 126:{c=b;d=748078;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 151:{c=b;d=748104;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 152:{c=b;d=748135;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 153:{c=b;d=748166;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 154:{c=b;d=748196;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 155:{c=b;d=748227;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 156:{c=b;d=748257;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 157:{c=b;d=748285;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 158:{c=b;d=748313;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 159:{c=b;d=748342;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 201:{c=b;d=748371;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 202:{c=b;d=748399;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 203:{c=b;d=748427;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 204:{c=b;d=748456;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 205:{c=b;d=748483;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 206:{c=b;d=748512;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 207:{c=b;d=748542;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 208:{c=b;d=748571;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 209:{c=b;d=748602;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 210:{c=b;d=748633;e=c+22|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 211:{c=b;d=748655;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 212:{c=b;d=748684;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 213:{c=b;d=748712;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 214:{c=b;d=748741;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 215:{c=b;d=748770;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 216:{c=b;d=748799;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 217:{c=b;d=748829;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 218:{c=b;d=748853;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 219:{c=b;d=748877;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 220:{c=b;d=748900;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 221:{c=b;d=748929;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 222:{c=b;d=748954;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 223:{c=b;d=748980;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 224:{c=b;d=749004;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 225:{c=b;d=749028;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 226:{c=b;d=749055;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 227:{c=b;d=749079;e=c+24|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 228:{c=b;d=749103;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 229:{c=b;d=749128;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 230:{c=b;d=749153;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 231:{c=b;d=749179;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 232:{c=b;d=749202;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 233:{c=b;d=749225;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 234:{c=b;d=749253;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 235:{c=b;d=749282;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 236:{c=b;d=749309;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 237:{c=b;d=749339;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 241:{c=b;d=749370;e=c+29|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 251:{c=b;d=749399;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 252:{c=b;d=749427;e=c+31|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 253:{c=b;d=749458;e=c+25|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 254:{c=b;d=749483;e=c+27|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 255:{c=b;d=749510;e=c+23|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 261:{c=b;d=749533;e=c+26|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 262:{c=b;d=749559;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 263:{c=b;d=749587;e=c+28|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}case 264:{c=b;d=749615;e=c+30|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}default:{c=b;d=749645;e=c+21|0;do{i[c>>0]=i[d>>0]|0;c=c+1|0;d=d+1|0}while((c|0)<(e|0));return}}while(0)}function tk(a){a=a|0;uk(5,a);return}
function _l(a,b,c,d,e,f,g,h,l,m,n,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0;t=r;r=r+16|0;u=t;v=k[s>>2]|0;if((v|0)>0){w=v;r=t;return w|0}a:do switch(b|0){case 1:{Hn(a,c,d,e,f,g,h,l,n,s)|0;break}case 11:{if(!m){bm(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{bm(a,c,d,e,f,g,h,l,1,1,i[m>>0]|0,n,u,q,s)|0;break a}break}case 12:{if(!m){On(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{On(a,c,d,e,f,g,h,l,1,1,i[m>>0]|0,n,u,q,s)|0;break a}break}case 20:{if(!m){ho(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{ho(a,c,d,e,f,g,h,l,1,1,j[m>>1]|0,n,u,q,s)|0;break a}break}case 21:{if(!m){Nm(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{Nm(a,c,d,e,f,g,h,l,1,1,j[m>>1]|0,n,u,q,s)|0;break a}break}case 30:{if(!m){Zn(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{Zn(a,c,d,e,f,g,h,l,1,1,k[m>>2]|0,n,u,q,s)|0;break a}break}case 31:{if(!m){wn(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{wn(a,c,d,e,f,g,h,l,1,1,k[m>>2]|0,n,u,q,s)|0;break a}break}case 40:{if(!m){so(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{so(a,c,d,e,f,g,h,l,1,1,k[m>>2]|0,n,u,q,s)|0;break a}break}case 41:{if(!m){Ym(a,c,d,e,f,g,h,l,1,1,0,n,u,q,s)|0;break a}else{Ym(a,c,d,e,f,g,h,l,1,1,k[m>>2]|0,n,u,q,s)|0;break a}break}case 81:{if(!m){kn(a,c,d,e,f,g,h,l,1,1,0,0,n,u,q,s)|0;break a}else{v=m;kn(a,c,d,e,f,g,h,l,1,1,k[v>>2]|0,k[v+4>>2]|0,n,u,q,s)|0;break a}break}case 42:{if(!m){Bm(a,c,d,e,f,g,h,l,1,1,0.0,n,u,q,s)|0;break a}else{Bm(a,c,d,e,f,g,h,l,1,1,+o[m>>2],n,u,q,s)|0;break a}break}case 82:{if(!m){nm(a,c,d,e,f,g,h,l,1,1,0.0,n,u,q,s)|0;break a}else{nm(a,c,d,e,f,g,h,l,1,1,+p[m>>3],n,u,q,s)|0;break a}break}case 83:{v=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;x=L;y=ty(h|0,l|0,1)|0;z=L;if(!m){Bm(a,c,d,e,v,x,y,z,1,1,0.0,n,u,q,s)|0;break a}else{Bm(a,c,d,e,v,x,y,z,1,1,+o[m>>2],n,u,q,s)|0;break a}break}case 163:{z=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;y=L;x=ty(h|0,l|0,1)|0;v=L;if(!m){nm(a,c,d,e,z,y,x,v,1,1,0.0,n,u,q,s)|0;break a}else{nm(a,c,d,e,z,y,x,v,1,1,+p[m>>3],n,u,q,s)|0;break a}break}case 14:{if(!m){Fn(a,c,d,e,f,g,h,l,1,0,n,u,q,s)|0;break a}else{Fn(a,c,d,e,f,g,h,l,1,i[m>>0]|0,n,u,q,s)|0;break a}break}case 16:{if(!m){i[u>>0]=0;Jn(a,c,d,e,f,g,h,l,1,u,n,u,q,s)|0;break a}else{Jn(a,c,d,e,f,g,h,l,1,m,n,u,q,s)|0;break a}break}default:k[s>>2]=410}while(0);w=k[s>>2]|0;r=t;return w|0}function $l(a,b,c,d,e,f,g,h,i,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0;o=r;r=r+16|0;p=k[n>>2]|0;if((p|0)>0){q=p;r=o;return q|0}do switch(b|0){case 1:{Hn(a,c,d,e,f,g,h,i,j,n)|0;break}case 11:{bm(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 12:{On(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 20:{ho(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 21:{Nm(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 30:{Zn(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 31:{wn(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 40:{so(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 41:{Ym(a,c,d,e,f,g,h,i,1,2,0,j,l,m,n)|0;break}case 81:{kn(a,c,d,e,f,g,h,i,1,2,0,0,j,l,m,n)|0;break}case 42:{Bm(a,c,d,e,f,g,h,i,1,2,0.0,j,l,m,n)|0;break}case 82:{nm(a,c,d,e,f,g,h,i,1,2,0.0,j,l,m,n)|0;break}case 83:{Em(a,c,d,e,f,g,h,i,j,l,m,n)|0;break}case 163:{sm(a,c,d,e,f,g,h,i,j,l,m,n)|0;break}case 14:{Fn(a,c,d,e,f,g,h,i,2,0,j,l,m,n)|0;break}case 16:{Jn(a,c,d,e,f,g,h,i,2,o,j,l,m,n)|0;break}default:k[n>>2]=410}while(0);q=k[n>>2]|0;r=o;return q|0}function am(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0;m=r;r=r+16|0;n=m;if(!(Cl(a,l)|0)){o=(b|0)<1;bm(a,2,o?1:b,o?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,m+1|0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}else{i[n>>0]=g;qt(a,11,c,d,e,f,1,n,h,0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}return 0}function bm(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0.0,X=0.0,Y=0.0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28920|0;D=t+28916|0;E=t+28912|0;F=t+28908|0;G=t+28904|0;H=t+32|0;I=t+24|0;J=t+16|0;K=t+8|0;M=t;N=t+28956|0;O=t+28976|0;P=t+28936|0;Q=t+40|0;R=k[s>>2]|0;if((g|0)==0&(h|0)==0|(R|0)>0){S=R;r=t;return S|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;nl(a,b,c,d,e,f,g,h,j>>31,x,y,N,E,z,A,I,J,F,H,M,B,K,P,s)|0;R=k[z>>2]|0;if((j|0)==1&(R|0)==14){Fn(a,b,c,d,e,f,g,h,l,m,n,o,q,s)|0;S=k[s>>2]|0;r=t;return S|0}f=k[A>>2]|0;if(!(rx(N,65)|0)){T=f;U=((f|0)<0)<<31>>31;V=R}else{if((k[s>>2]|0)==308){k[s>>2]=0;xk()}k[z>>2]=11;k[F>>2]=1;R=k[E>>2]|0;f=H;k[f>>2]=R;k[f+4>>2]=((R|0)<0)<<31>>31;k[E>>2]=1;p[x>>3]=1.0;p[y>>3]=0.0;R=K;k[R>>2]=1234554321;k[R+4>>2]=0;T=28800;U=0;V=11}R=k[s>>2]|0;if((R|0)>0){S=R;r=t;return S|0}R=ha(k[F>>2]|0,j)|0;k[F>>2]=R;if((V|0)==16&(k[B>>2]|0)==1?(Nk(N,C,G,D,s)|0,G=k[D>>2]|0,(G|0)>0):0){D=0;W=1.0;while(1){X=W*10.0;D=D+1|0;if((D|0)>=(G|0)){Y=X;break}else W=X}}else Y=1.0;G=k[z>>2]|0;if(!((l|0)==1&m<<24>>24==0)?(D=K,C=k[D>>2]|0,V=k[D+4>>2]|0,!(((G|0)%10|0|0)==1&((C|0)==1234554321&(V|0)==0))):0){D=ky(C|0,V|0,32768,0)|0;R=L;if(!((G|0)==21&(R>>>0>0|(R|0)==0&D>>>0>65535)))if((G|0)==11&(V>>>0>0|(V|0)==0&C>>>0>255)){Z=0;_=22}else{$=(G|0)==16&(i[P>>0]|0)==1?0:l;_=21}else{aa=U;ba=T;ca=0;da=0}}else{$=0;_=21}if((_|0)==21)if((G|0)==11){Z=$;_=22}else{aa=U;ba=T;ca=0;da=$}if((_|0)==22){$=(h|0)<0|(h|0)==0&g>>>0<2147483647;aa=$?h:0;ba=$?g:2147483647;ca=(Z|0)==0&+p[x>>3]==1.0&+p[y>>3]==0.0;da=Z}Z=(j|0)>-1;$=((j|0)<0)<<31>>31;T=0-j|0;U=((T|0)<0)<<31>>31;G=0;l=0;C=g;g=h;h=0;V=0;a:while(1){D=V;R=C;f=g;A=h;while(1){e=n+A|0;d=o+A|0;c=R;ea=f;while(1){if((c|0)==0&(ea|0)==0){_=56;break a}fa=(ea|0)<(aa|0)|(ea|0)==(aa|0)&c>>>0<ba>>>0;ga=fa?c:ba;fa=my(0,ga|0,32)|0;ga=L;if(Z){ia=H;ja=J;ka=k[ja>>2]|0;la=k[ja+4>>2]|0;ja=ky(k[ia>>2]|0,k[ia+4>>2]|0,-1,-1)|0;ia=jy(ja|0,L|0,ka|0,la|0)|0;ja=wy(ia|0,L|0,j|0,$|0)|0;ia=ky(ja|0,L|0,1,0)|0;ja=L;ma=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<ia>>>0;na=ka;oa=la;pa=ma?fa:ia;qa=ma?ga:ja}else{ja=J;ma=k[ja>>2]|0;ia=k[ja+4>>2]|0;ja=wy(ma|0,ia|0,T|0,U|0)|0;la=ky(ja|0,L|0,1,0)|0;ja=L;ka=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<la>>>0;na=ma;oa=ia;pa=ka?fa:la;qa=ka?ga:ja}ja=I;ga=k[ja>>2]|0;ka=k[ja+4>>2]|0;ja=M;la=yy(k[ja>>2]|0,k[ja+4>>2]|0,G|0,l|0)|0;ja=ky(la|0,L|0,ga|0,ka|0)|0;ka=L;ga=k[F>>2]|0;la=(ga|0)/(j|0)|0;fa=yy(la|0,((la|0)<0)<<31>>31|0,na|0,oa|0)|0;la=ky(ja|0,ka|0,fa|0,L|0)|0;fa=L;switch(k[z>>2]|0){case 11:{lu(a,la,fa,pa,ga,e,s)|0;if(!ca)fm(e,pa,+p[x>>3],+p[y>>3],da,k[K>>2]&255,m,d,q,e,s)|0;break}case 21:{mu(a,la,fa,pa,ga,Q,s)|0;gm(Q,pa,+p[x>>3],+p[y>>3],da,k[K>>2]&65535,m,d,q,e,s)|0;break}case 41:{nu(a,la,fa,pa,ga,Q,s)|0;hm(Q,pa,+p[x>>3],+p[y>>3],da,k[K>>2]|0,m,d,q,e,s)|0;break}case 81:{ou(a,la,fa,pa,ga,Q,s)|0;ka=K;im(Q,pa,+p[x>>3],+p[y>>3],da,k[ka>>2]|0,k[ka+4>>2]|0,m,d,q,e,s)|0;break}case 42:{pu(a,la,fa,pa,ga,Q,s)|0;jm(Q,pa,+p[x>>3],+p[y>>3],da,m,d,q,e,s)|0;break}case 82:{qu(a,la,fa,pa,ga,Q,s)|0;km(Q,pa,+p[x>>3],+p[y>>3],da,m,d,q,e,s)|0;break}case 16:{au(a,la,fa,0,s)|0;fa=k[F>>2]|0;la=k[E>>2]|0;if((fa|0)==(la|0)){ga=ha(fa,pa)|0;fu(a,ga,((ga|0)<0)<<31>>31,Q,s)|0}else gu(a,la,pa,fa-la|0,Q,s)|0;lm(Q,pa,+p[x>>3],+p[y>>3],k[E>>2]|0,Y,da,P,m,d,q,e,s)|0;break}default:{_=42;break a}}if((k[s>>2]|0)>0){ra=pa;sa=A;ta=D;_=46;break a}la=((pa|0)<0)<<31>>31;fa=jy(c|0,ea|0,pa|0,la|0)|0;ga=L;if((c|0)==(pa|0)&(ea|0)==(la|0)){c=fa;ea=ga}else{ua=pa;va=pa;wa=la;xa=fa;ya=ga;break}}ea=ky(va|0,wa|0,A|0,D|0)|0;c=L;e=J;d=ha(ua,j)|0;ga=ky(k[e>>2]|0,k[e+4>>2]|0,d|0,((d|0)<0)<<31>>31|0)|0;d=L;e=J;k[e>>2]=ga;k[e+4>>2]=d;e=H;fa=k[e>>2]|0;la=k[e+4>>2]|0;if(!((d|0)<(la|0)|(d|0)==(la|0)&ga>>>0<fa>>>0)){za=ga;Aa=d;Ba=fa;Ca=la;Da=ea;Ea=c;Fa=xa;Ga=ya;_=52;break}if((d|0)<0){Ha=ga;Ia=d;Ja=fa;Ka=la;La=ea;Ma=c;Na=xa;Oa=ya;_=55;break}else{D=c;R=xa;f=ya;A=ea}}if((_|0)==52){_=0;A=wy(za|0,Aa|0,Ba|0,Ca|0)|0;f=L;R=ky(A|0,f|0,G|0,l|0)|0;D=L;ea=yy(A|0,f|0,Ba|0,Ca|0)|0;f=jy(za|0,Aa|0,ea|0,L|0)|0;Pa=f;Qa=L;Ra=Da;Sa=Ea;Ta=Fa;Ua=Ga;Va=R;Wa=D}else if((_|0)==55){_=0;D=wy(~Ha|0,~Ia|0,Ja|0,Ka|0)|0;R=ky(D|0,L|0,1,0)|0;D=L;f=jy(G|0,l|0,R|0,D|0)|0;ea=L;A=yy(R|0,D|0,Ja|0,Ka|0)|0;D=ky(A|0,L|0,Ha|0,Ia|0)|0;Pa=D;Qa=L;Ra=La;Sa=Ma;Ta=Na;Ua=Oa;Va=f;Wa=ea}ea=J;k[ea>>2]=Pa;k[ea+4>>2]=Qa;G=Va;l=Wa;C=Ta;g=Ua;h=Ra;V=Sa}if((_|0)==42){k[w>>2]=b;k[w+4>>2]=N;Xw(O,760278,w)|0;tk(O);if((k[B>>2]|0)==1){k[s>>2]=311;S=311;r=t;return S|0}else{k[s>>2]=312;S=312;r=t;return S|0}}else if((_|0)==46){Y=+(sa>>>0)+4294967296.0*+(ta|0);W=Y+1.0;X=Y+ +(ra|0);if((k[B>>2]|0)>0){p[v>>3]=W;p[v+8>>3]=X;k[v+16>>2]=b;Xw(O,759004,v)|0}else{p[u>>3]=W;p[u+8>>3]=X;Xw(O,759067,u)|0}tk(O);S=k[s>>2]|0;r=t;return S|0}else if((_|0)==56){_=k[s>>2]|0;if((_|0)!=-11){S=_;r=t;return S|0}tk(761034);k[s>>2]=412;S=412;r=t;return S|0}return 0}function cm(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;bm(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,11,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function dm(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;n=r;r=r+480|0;o=n+232|0;p=n+224|0;q=n+352|0;s=n+316|0;t=n+280|0;u=n+244|0;v=n+144|0;w=n+72|0;x=n;y=n+240|0;z=n+236|0;A=n+389|0;B=n+390|0;C=n+388|0;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(B,759126,p)|0;tk(B);k[m>>2]=320;D=320;r=n;return D|0}if(Cl(a,m)|0){if((c|0)>0){p=0;do{E=k[e+(p<<2)>>2]|0;F=w+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(p<<2)>>2]|0;F=x+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}i[C>>0]=h;mt(a,11,w,x,g,1,C,j,0,l,m)|0;D=k[m>>2]|0;r=n;return D|0}if((Zk(a,y,m)|0)>0){D=k[m>>2]|0;r=n;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(l|0)!=0;if(b)k[l>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;p=v+8|0;E=p;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[o>>2]=xa+1;Xw(B,759171,o)|0;tk(B);k[m>>2]=321;D=321;r=n;return D|0}while(0);B=v+(c<<3)|0;v=B;o=k[u+(c<<2)>>2]|0;xa=yy(o|0,((o|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;o=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(o|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=p;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=o;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((bm(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,j+_b|0,A,z,m)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[l>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[m>>2]|0;r=n;return D|0}while(0);D=k[m>>2]|0;r=n;return D|0}function em(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0;n=r;r=r+16|0;bm(a,b,c,d,e,f,g,h,1,1,i,j,n,l,m)|0;r=n;return k[m>>2]|0}function fm(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0.0,s=0;o=c==1.0&d==0.0;if(!e){if(o){qy(m|0,a|0,b|0)|0;p=k[n>>2]|0;return p|0}if((b|0)>0)q=0;else{p=k[n>>2]|0;return p|0}do{r=+(l[a+q>>0]|0|0)*c+d;do if(!(r<-.49))if(r>255.49){k[n>>2]=-11;i[m+q>>0]=-1;break}else{i[m+q>>0]=~~r;break}else{k[n>>2]=-11;i[m+q>>0]=0}while(0);q=q+1|0}while((q|0)!=(b|0));p=k[n>>2]|0;return p|0}q=(b|0)>0;if(o){if(!q){p=k[n>>2]|0;return p|0}if((e|0)==1){o=0;do{s=i[a+o>>0]|0;if(s<<24>>24==f<<24>>24){k[j>>2]=1;i[m+o>>0]=g}else i[m+o>>0]=s;o=o+1|0}while((o|0)!=(b|0));p=k[n>>2]|0;return p|0}else{o=0;do{s=i[a+o>>0]|0;if(s<<24>>24==f<<24>>24){k[j>>2]=1;i[h+o>>0]=1}else i[m+o>>0]=s;o=o+1|0}while((o|0)!=(b|0));p=k[n>>2]|0;return p|0}}if(!q){p=k[n>>2]|0;return p|0}if((e|0)==1){e=0;do{q=i[a+e>>0]|0;do if(q<<24>>24!=f<<24>>24){r=+(q&255|0)*c+d;if(r<-.49){k[n>>2]=-11;i[m+e>>0]=0;break}if(r>255.49){k[n>>2]=-11;i[m+e>>0]=-1;break}else{i[m+e>>0]=~~r;break}}else{k[j>>2]=1;i[m+e>>0]=g}while(0);e=e+1|0}while((e|0)!=(b|0));p=k[n>>2]|0;return p|0}else{e=0;do{g=i[a+e>>0]|0;do if(g<<24>>24!=f<<24>>24){r=+(g&255|0)*c+d;if(r<-.49){k[n>>2]=-11;i[m+e>>0]=0;break}if(r>255.49){k[n>>2]=-11;i[m+e>>0]=-1;break}else{i[m+e>>0]=~~r;break}}else{k[j>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));p=k[n>>2]|0;return p|0}return 0}function gm(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=j[a+(q<<1)>>1]|0;do if(s<<16>>16>=0)if(s<<16>>16>255){k[n>>2]=-11;i[m+q>>0]=-1;break}else{i[m+q>>0]=s;break}else{k[n>>2]=-11;i[m+q>>0]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)t=0;else{r=k[n>>2]|0;return r|0}do{u=+(j[a+(t<<1)>>1]|0)*c+d;do if(!(u<-.49))if(u>255.49){k[n>>2]=-11;i[m+t>>0]=-1;break}else{i[m+t>>0]=~~u;break}else{k[n>>2]=-11;i[m+t>>0]=0}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;t=0;do{q=j[a+(t<<1)>>1]|0;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){i[m+t>>0]=g;break}else{i[h+t>>0]=1;break}}else{u=+(q<<16>>16)*c+d;if(u<-.49){k[n>>2]=-11;i[m+t>>0]=0;break}if(u>255.49){k[n>>2]=-11;i[m+t>>0]=-1;break}else{i[m+t>>0]=~~u;break}}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){e=0;do{p=j[a+(e<<1)>>1]|0;do if(p<<16>>16!=f<<16>>16){if(p<<16>>16<0){k[n>>2]=-11;i[m+e>>0]=0;break}if(p<<16>>16>255){k[n>>2]=-11;i[m+e>>0]=-1;break}else{i[m+e>>0]=p;break}}else{k[l>>2]=1;i[m+e>>0]=g}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}else{e=0;do{g=j[a+(e<<1)>>1]|0;do if(g<<16>>16!=f<<16>>16){if(g<<16>>16<0){k[n>>2]=-11;i[m+e>>0]=0;break}if(g<<16>>16>255){k[n>>2]=-11;i[m+e>>0]=-1;break}else{i[m+e>>0]=g;break}}else{k[l>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function hm(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0.0;n=c==1.0&d==0.0;o=(b|0)>0;if(!e)if(n){if(o)p=0;else{q=k[m>>2]|0;return q|0}do{r=k[a+(p<<2)>>2]|0;do if((r|0)>=0)if((r|0)>255){k[m>>2]=-11;i[l+p>>0]=-1;break}else{i[l+p>>0]=r;break}else{k[m>>2]=-11;i[l+p>>0]=0}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}else{if(o)s=0;else{q=k[m>>2]|0;return q|0}do{t=+(k[a+(s<<2)>>2]|0)*c+d;do if(!(t<-.49))if(t>255.49){k[m>>2]=-11;i[l+s>>0]=-1;break}else{i[l+s>>0]=~~t;break}else{k[m>>2]=-11;i[l+s>>0]=0}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[m>>2]|0;return q|0}if(n){if(!o){q=k[m>>2]|0;return q|0}n=(e|0)==1;s=0;do{p=k[a+(s<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(n){i[l+s>>0]=g;break}else{i[h+s>>0]=1;break}}else{if((p|0)<0){k[m>>2]=-11;i[l+s>>0]=0;break}if((p|0)>255){k[m>>2]=-11;i[l+s>>0]=-1;break}else{i[l+s>>0]=p;break}}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[m>>2]|0;return q|0}else{if(!o){q=k[m>>2]|0;return q|0}o=(e|0)==1;e=0;do{s=k[a+(e<<2)>>2]|0;do if((s|0)==(f|0)){k[j>>2]=1;if(o){i[l+e>>0]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s|0)*c+d;if(t<-.49){k[m>>2]=-11;i[l+e>>0]=0;break}if(t>255.49){k[m>>2]=-11;i[l+e>>0]=-1;break}else{i[l+e>>0]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}return 0}function im(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=a+(q<<3)|0;t=k[s>>2]|0;u=k[s+4>>2]|0;do if((u|0)>=0)if((u|0)>0|(u|0)==0&t>>>0>255){k[n>>2]=-11;i[m+q>>0]=-1;break}else{i[m+q>>0]=t;break}else{k[n>>2]=-11;i[m+q>>0]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)v=0;else{r=k[n>>2]|0;return r|0}do{q=a+(v<<3)|0;w=(+((k[q>>2]|0)>>>0)+4294967296.0*+(k[q+4>>2]|0))*c+d;do if(!(w<-.49))if(w>255.49){k[n>>2]=-11;i[m+v>>0]=-1;break}else{i[m+v>>0]=~~w;break}else{k[n>>2]=-11;i[m+v>>0]=0}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;v=0;do{q=a+(v<<3)|0;t=k[q>>2]|0;u=k[q+4>>2]|0;do if((t|0)==(f|0)&(u|0)==(g|0)){k[l>>2]=1;if(o){i[m+v>>0]=h;break}else{i[j+v>>0]=1;break}}else{if((u|0)<0){k[n>>2]=-11;i[m+v>>0]=0;break}if((u|0)>0|(u|0)==0&t>>>0>255){k[n>>2]=-11;i[m+v>>0]=-1;break}else{i[m+v>>0]=t;break}}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{v=a+(e<<3)|0;o=k[v>>2]|0;t=k[v+4>>2]|0;do if((o|0)==(f|0)&(t|0)==(g|0)){k[l>>2]=1;if(p){i[m+e>>0]=h;break}else{i[j+e>>0]=1;break}}else{w=(+(o>>>0)+4294967296.0*+(t|0))*c+d;if(w<-.49){k[n>>2]=-11;i[m+e>>0]=0;break}if(w>255.49){k[n>>2]=-11;i[m+e>>0]=-1;break}else{i[m+e>>0]=~~w;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function jm(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-.49))if(s>255.49){k[l>>2]=-11;i[j+p>>0]=-1;break}else{i[j+p>>0]=~~r;break}else{k[l>>2]=-11;i[j+p>>0]=0}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-.49))if(r>255.49){k[l>>2]=-11;i[j+t>>0]=-1;break}else{i[j+t>>0]=~~r;break}else{k[l>>2]=-11;i[j+t>>0]=0}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-.49){k[l>>2]=-11;i[j+u>>0]=0;break a}if(s>255.49){k[l>>2]=-11;i[j+u>>0]=-1;break a}else{i[j+u>>0]=~~r;break a}break}case 1:{k[h>>2]=1;if(p){i[j+u>>0]=f;break a}else{i[g+u>>0]=1;break a}break}default:i[j+u>>0]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;v=d>255.49;u=~~d&255;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-.49){k[l>>2]=-11;i[j+e>>0]=0;break b}if(r>255.49){k[l>>2]=-11;i[j+e>>0]=-1;break b}else{i[j+e>>0]=~~r;break b}break}case 1:{k[h>>2]=1;if(p){i[j+e>>0]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;i[j+e>>0]=0;break b}if(v){k[l>>2]=-11;i[j+e>>0]=-1;break b}else{i[j+e>>0]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function km(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-.49))if(r>255.49){k[l>>2]=-11;i[j+o>>0]=-1;break}else{i[j+o>>0]=~~r;break}else{k[l>>2]=-11;i[j+o>>0]=0}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-.49))if(r>255.49){k[l>>2]=-11;i[j+s>>0]=-1;break}else{i[j+s>>0]=~~r;break}else{k[l>>2]=-11;i[j+s>>0]=0}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-.49){k[l>>2]=-11;i[j+t>>0]=0;break a}if(r>255.49){k[l>>2]=-11;i[j+t>>0]=-1;break a}else{i[j+t>>0]=~~r;break a}break}case 1:{k[h>>2]=1;if(o){i[j+t>>0]=f;break a}else{i[g+t>>0]=1;break a}break}default:i[j+t>>0]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;u=d>255.49;t=~~d&255;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-.49){k[l>>2]=-11;i[j+e>>0]=0;break b}if(r>255.49){k[l>>2]=-11;i[j+e>>0]=-1;break b}else{i[j+e>>0]=~~r;break b}break}case 1:{k[h>>2]=1;if(o){i[j+e>>0]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;i[j+e>>0]=0;break b}if(u){k[l>>2]=-11;i[j+e>>0]=-1;break b}else{i[j+e>>0]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function lm(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0.0,P=0.0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0.0,Z=0,_=0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0,ga=0.0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0;p=r;r=r+96|0;q=p;s=p+4|0;t=xx(h)|0;a:do if((b|0)>0){u=(g|0)==0;v=(g|0)==1;w=a;x=0;b:while(1){y=w+e|0;z=i[y>>0]|0;i[y>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,w,t)|0)==0:0)if(!u){k[m>>2]=1;if(v){i[n+x>>0]=j;A=y;break}else{i[l+x>>0]=1;A=y;break}}else A=y;else{B=w;C=9}while(0);do if((C|0)==9){c:while(1){C=0;D=i[B>>0]|0;switch(D<<24>>24){case 43:case 45:{E=D;F=B;C=11;break c;break}case 32:break;default:{G=D;H=B;I=1;break c}}B=B+1|0;C=9}if((C|0)==11){C=0;D=E<<24>>24==45;J=F;while(1){K=J+1|0;L=i[K>>0]|0;if(L<<24>>24==32)J=K;else{M=L;N=K;break}}G=M;H=N;I=D?-1:1}if((G+-48&255)<10){J=G;K=H;O=0.0;while(1){P=O*10.0+ +(J<<24>>24);L=K;while(1){Q=L+1|0;R=i[Q>>0]|0;if(R<<24>>24==32)L=Q;else{S=R;T=Q;break}}U=P+-48.0;if((S+-48&255)<10){J=S;K=T;O=U}else{V=S;W=T;X=U;break}}}else{V=G;W=H;X=0.0}switch(V<<24>>24){case 44:case 46:{K=W;while(1){J=K+1|0;D=i[J>>0]|0;if(D<<24>>24==32)K=J;else{Z=D;_=J;break}}if((Z+-48&255)<10){K=Z;J=_;O=1.0;U=X;while(1){$=U*10.0+ +(K<<24>>24)+-48.0;D=J;while(1){L=D+1|0;Q=i[L>>0]|0;if(Q<<24>>24==32)D=L;else{aa=Q;ba=L;break}}P=O*10.0;if((aa+-48&255)<10){K=aa;J=ba;O=P;U=$}else{ca=aa;da=ba;ea=1;fa=P;ga=$;break}}}else{ca=Z;da=_;ea=1;fa=1.0;ga=X}break}default:{ca=V;da=W;ea=0;fa=1.0;ga=X}}if((ca&-2)<<24>>24==68){J=da;d:while(1){K=J+1|0;D=i[K>>0]|0;switch(D<<24>>24){case 32:{J=K;break}case 43:case 45:{ia=D;ja=J;C=28;break d;break}default:{ka=D;la=K;ma=1;break d}}}if((C|0)==28){C=0;J=ia<<24>>24==45;K=ja+2|0;while(1){D=i[K>>0]|0;if(D<<24>>24==32)K=K+1|0;else{na=D;oa=K;break}}ka=na;la=oa;ma=J?-1:1}if((ka+-48&255)<10){K=ka;D=la;L=0;while(1){Q=(L*10|0)+-48|0;R=D;while(1){pa=R+1|0;qa=i[pa>>0]|0;if(qa<<24>>24==32)R=pa;else{ra=qa;sa=pa;break}}R=Q+(K<<24>>24)|0;if((ra+-48&255)<10){K=ra;D=sa;L=R}else{ta=ra;ua=sa;va=ma;wa=R;break}}}else{ta=ka;ua=la;va=ma;wa=0}}else{ta=ca;ua=da;va=1;wa=0}if(ta<<24>>24){xa=y;ya=z;za=w;break b}$=+(I|0)*ga/((ea|0)==0?f:fa)*+Y(10.0,+(+(ha(va,wa)|0)))*c+d;if($<-.49){k[o>>2]=-11;i[n+x>>0]=0;A=ua;break}if($>255.49){k[o>>2]=-11;i[n+x>>0]=-1;A=ua;break}else{i[n+x>>0]=~~$;A=ua;break}}while(0);i[y>>0]=z;x=x+1|0;if((x|0)>=(b|0))break a;else w=A}w=s;x=761193;v=w+36|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(v|0));tk(s);k[q>>2]=za;Xw(s,761229,q)|0;tk(s);i[xa>>0]=ya;k[o>>2]=409;Aa=409;r=p;return Aa|0}while(0);Aa=k[o>>2]|0;r=p;return Aa|0}function mm(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=+g;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;nm(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+8|0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}else{p[m>>3]=g;qt(a,82,c,d,e,f,1,m,h,0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}return 0}function nm(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=+m;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28920|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,j>>31,x,y,K,D,z,C,G,H,E,F,J,A,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[C>>2]|0;C=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28924|0,t+28904|0,B,s)|0,e=k[B>>2]|0,(e|0)>0):0){B=0;R=1.0;while(1){S=R*10.0;B=B+1|0;if((B|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;e=k[z>>2]|0;if(!((l|0)==1&m==0.0)?(B=I,d=k[B>>2]|0,c=k[B+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0){B=ky(d|0,c|0,32768,0)|0;P=L;if(!((e|0)==21&(P>>>0>0|(P|0)==0&B>>>0>65535))?!((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)):0){U=(e|0)==16&(i[N>>0]|0)==1?0:l;V=16}else{W=C;X=f;Y=0;Z=0}}else{U=0;V=16}if((V|0)==16)if((e|0)==82){e=(h|0)<0|(h|0)==0&g>>>0<268435455;W=e?h:0;X=e?g:268435455;Y=(U|0)==0&+p[x>>3]==1.0&+p[y>>3]==0.0;Z=U}else{W=C;X=f;Y=0;Z=U}U=(j|0)>-1;f=((j|0)<0)<<31>>31;C=0-j|0;e=((C|0)<0)<<31>>31;l=0;d=0;c=g;g=h;h=0;B=0;a:while(1){P=B;_=c;$=g;aa=h;while(1){ba=n+(aa<<3)|0;ca=o+aa|0;da=_;ea=$;while(1){if((da|0)==0&(ea|0)==0){V=51;break a}fa=(ea|0)<(W|0)|(ea|0)==(W|0)&da>>>0<X>>>0;ga=fa?da:X;fa=my(0,ga|0,32)|0;ga=L;if(U){ia=F;ja=H;ka=k[ja>>2]|0;la=k[ja+4>>2]|0;ja=ky(k[ia>>2]|0,k[ia+4>>2]|0,-1,-1)|0;ia=jy(ja|0,L|0,ka|0,la|0)|0;ja=wy(ia|0,L|0,j|0,f|0)|0;ia=ky(ja|0,L|0,1,0)|0;ja=L;ma=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<ia>>>0;na=ka;oa=la;pa=ma?fa:ia;qa=ma?ga:ja}else{ja=H;ma=k[ja>>2]|0;ia=k[ja+4>>2]|0;ja=wy(ma|0,ia|0,C|0,e|0)|0;la=ky(ja|0,L|0,1,0)|0;ja=L;ka=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<la>>>0;na=ma;oa=ia;pa=ka?fa:la;qa=ka?ga:ja}ja=G;ga=k[ja>>2]|0;ka=k[ja+4>>2]|0;ja=J;la=yy(k[ja>>2]|0,k[ja+4>>2]|0,l|0,d|0)|0;ja=ky(la|0,L|0,ga|0,ka|0)|0;ka=L;ga=k[E>>2]|0;la=(ga|0)/(j|0)|0;fa=yy(la|0,((la|0)<0)<<31>>31|0,na|0,oa|0)|0;la=ky(ja|0,ka|0,fa|0,L|0)|0;fa=L;switch(k[z>>2]|0){case 82:{qu(a,la,fa,pa,ga,ba,s)|0;if(!Y)tm(ba,pa,+p[x>>3],+p[y>>3],Z,m,ca,q,ba,s)|0;break}case 11:{lu(a,la,fa,pa,ga,O,s)|0;um(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]&255,m,ca,q,ba,s)|0;break}case 21:{mu(a,la,fa,pa,ga,O,s)|0;vm(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]&65535,m,ca,q,ba,s)|0;break}case 41:{nu(a,la,fa,pa,ga,O,s)|0;wm(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]|0,m,ca,q,ba,s)|0;break}case 81:{ou(a,la,fa,pa,ga,O,s)|0;ka=I;xm(O,pa,+p[x>>3],+p[y>>3],Z,k[ka>>2]|0,k[ka+4>>2]|0,m,ca,q,ba,s)|0;break}case 42:{pu(a,la,fa,pa,ga,O,s)|0;ym(O,pa,+p[x>>3],+p[y>>3],Z,m,ca,q,ba,s)|0;break}case 16:{au(a,la,fa,0,s)|0;fa=k[E>>2]|0;la=k[D>>2]|0;if((fa|0)==(la|0)){ga=ha(fa,pa)|0;fu(a,ga,((ga|0)<0)<<31>>31,O,s)|0}else gu(a,la,pa,fa-la|0,O,s)|0;zm(O,pa,+p[x>>3],+p[y>>3],k[D>>2]|0,T,Z,N,m,ca,q,ba,s)|0;break}default:{V=37;break a}}if((k[s>>2]|0)>0){ra=pa;sa=aa;ta=P;V=41;break a}la=((pa|0)<0)<<31>>31;fa=jy(da|0,ea|0,pa|0,la|0)|0;ga=L;if((da|0)==(pa|0)&(ea|0)==(la|0)){da=fa;ea=ga}else{ua=pa;va=pa;wa=la;xa=fa;ya=ga;break}}ea=ky(va|0,wa|0,aa|0,P|0)|0;da=L;ba=H;ca=ha(ua,j)|0;ga=ky(k[ba>>2]|0,k[ba+4>>2]|0,ca|0,((ca|0)<0)<<31>>31|0)|0;ca=L;ba=H;k[ba>>2]=ga;k[ba+4>>2]=ca;ba=F;fa=k[ba>>2]|0;la=k[ba+4>>2]|0;if(!((ca|0)<(la|0)|(ca|0)==(la|0)&ga>>>0<fa>>>0)){za=ga;Aa=ca;Ba=fa;Ca=la;Da=ea;Ea=da;Fa=xa;Ga=ya;V=47;break}if((ca|0)<0){Ha=ga;Ia=ca;Ja=fa;Ka=la;La=ea;Ma=da;Na=xa;Oa=ya;V=50;break}else{P=da;_=xa;$=ya;aa=ea}}if((V|0)==47){V=0;aa=wy(za|0,Aa|0,Ba|0,Ca|0)|0;$=my(0,aa|0,32)|0;aa=L;_=ky($|0,aa|0,l|0,d|0)|0;P=L;ea=yy($|0,aa|0,Ba|0,Ca|0)|0;aa=jy(za|0,Aa|0,ea|0,L|0)|0;Pa=aa;Qa=L;Ra=Da;Sa=Ea;Ta=Fa;Ua=Ga;Va=_;Wa=P}else if((V|0)==50){V=0;P=wy(~Ha|0,~Ia|0,Ja|0,Ka|0)|0;_=ky(0,P|0,0,1)|0;P=my(_|0,L|0,32)|0;_=L;aa=jy(l|0,d|0,P|0,_|0)|0;ea=L;$=yy(P|0,_|0,Ja|0,Ka|0)|0;_=ky($|0,L|0,Ha|0,Ia|0)|0;Pa=_;Qa=L;Ra=La;Sa=Ma;Ta=Na;Ua=Oa;Va=aa;Wa=ea}ea=H;k[ea>>2]=Pa;k[ea+4>>2]=Qa;l=Va;d=Wa;c=Ta;g=Ua;h=Ra;B=Sa}if((V|0)==37){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[A>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((V|0)==41){m=+(sa>>>0)+4294967296.0*+(ta|0);T=m+1.0;R=m+ +(ra|0);if((k[A>>2]|0)>0){p[v>>3]=T;p[v+8>>3]=R;k[v+16>>2]=b;Xw(M,759216,v)|0}else{p[u>>3]=T;p[u+8>>3]=R;Xw(M,759279,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((V|0)==51){V=k[s>>2]|0;if((V|0)!=-11){Q=V;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function om(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;nm(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0.0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,82,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function pm(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;m=r;r=r+480|0;n=m+240|0;o=m+232|0;q=m+360|0;s=m+324|0;t=m+288|0;u=m+252|0;v=m+152|0;w=m+80|0;x=m+8|0;y=m+248|0;z=m+244|0;A=m+396|0;B=m+397|0;C=m;if((c+-1|0)>>>0>8){k[o>>2]=c;Xw(B,759338,o)|0;tk(B);k[l>>2]=320;D=320;r=m;return D|0}if(Cl(a,l)|0){if((c|0)>0){o=0;do{E=k[e+(o<<2)>>2]|0;F=w+(o<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(o<<2)>>2]|0;F=x+(o<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;o=o+1|0}while((o|0)!=(c|0))}p[C>>3]=h;mt(a,82,w,x,g,1,C,i,0,j,l)|0;D=k[l>>2]|0;r=m;return D|0}if((Zk(a,y,l)|0)>0){D=k[l>>2]|0;r=m;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;o=v+8|0;E=o;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[n>>2]=xa+1;Xw(B,759383,n)|0;tk(B);k[l>>2]=321;D=321;r=m;return D|0}while(0);B=v+(c<<3)|0;v=B;n=k[u+(c<<2)>>2]|0;xa=yy(n|0,((n|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;n=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(n|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=o;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=n;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((nm(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,i+(_b<<3)|0,A,z,l)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[j>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[l>>2]|0;r=m;return D|0}while(0);D=k[l>>2]|0;r=m;return D|0}function qm(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=+i;j=j|0;l=l|0;m=m|0;var n=0;n=r;r=r+16|0;nm(a,b,c,d,e,f,g,h,1,1,i,j,n,l,m)|0;r=n;return k[m>>2]|0}function rm(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;nm(a,b,c,d,e,f,g,h,1,2,0.0,i,j,l,m)|0;return k[m>>2]|0}function sm(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;o=ty(g|0,h|0,1)|0;p=L;q=ey(o,1)|0;r=ty(e|0,f|0,1)|0;f=ky(r|0,L|0,-1,-1)|0;nm(a,b,c,d,f,L,o,p,1,2,0.0,j,q,m,n)|0;if((h|0)>0|(h|0)==0&g>>>0>0){s=0;t=0;u=0;v=0}else{dy(q);w=k[n>>2]|0;return w|0}do{if((i[q+s>>0]|0)==0?(i[q+(s|1)>>0]|0)==0:0)i[l+t>>0]=0;else i[l+t>>0]=1;s=ky(s|0,u|0,2,0)|0;u=L;t=ky(t|0,v|0,1,0)|0;v=L}while((v|0)<(h|0)|(v|0)==(h|0)&t>>>0<g>>>0);dy(q);w=k[n>>2]|0;return w|0}function tm(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0,s=0,t=0;if(!e){if(c==1.0&d==0.0){qy(j|0,a|0,b<<3|0)|0;n=k[l>>2]|0;return n|0}if((b|0)>0)o=0;else{n=k[l>>2]|0;return n|0}do{p[j+(o<<3)>>3]=+p[a+(o<<3)>>3]*c+d;o=o+1|0}while((o|0)!=(b|0));n=k[l>>2]|0;return n|0}o=a+6|0;q=(b|0)>0;if(c==1.0&d==0.0){if(!q){n=k[l>>2]|0;return n|0}if((e|0)==1){r=0;s=o;while(1){t=(m[s>>1]|0)&32752;switch(((t|0)==32752?1:(t|0)==0?2:0)|0){case 0:{p[j+(r<<3)>>3]=+p[a+(r<<3)>>3];break}case 1:{k[h>>2]=1;p[j+(r<<3)>>3]=f;break}default:p[j+(r<<3)>>3]=0.0}r=r+1|0;if((r|0)==(b|0))break;else s=s+8|0}n=k[l>>2]|0;return n|0}else{s=0;r=o;while(1){t=(m[r>>1]|0)&32752;switch(((t|0)==32752?1:(t|0)==0?2:0)|0){case 0:{p[j+(s<<3)>>3]=+p[a+(s<<3)>>3];break}case 1:{k[h>>2]=1;i[g+s>>0]=1;p[j+(s<<3)>>3]=-9.1191291391491e-36;break}default:p[j+(s<<3)>>3]=0.0}s=s+1|0;if((s|0)==(b|0))break;else r=r+8|0}n=k[l>>2]|0;return n|0}}else{if(!q){n=k[l>>2]|0;return n|0}if((e|0)==1){e=0;q=o;while(1){r=(m[q>>1]|0)&32752;switch(((r|0)==32752?1:(r|0)==0?2:0)|0){case 0:{p[j+(e<<3)>>3]=+p[a+(e<<3)>>3]*c+d;break}case 1:{k[h>>2]=1;p[j+(e<<3)>>3]=f;break}default:p[j+(e<<3)>>3]=d}e=e+1|0;if((e|0)==(b|0))break;else q=q+8|0}n=k[l>>2]|0;return n|0}else{q=0;e=o;while(1){o=(m[e>>1]|0)&32752;switch(((o|0)==32752?1:(o|0)==0?2:0)|0){case 0:{p[j+(q<<3)>>3]=+p[a+(q<<3)>>3]*c+d;break}case 1:{k[h>>2]=1;i[g+q>>0]=1;p[j+(q<<3)>>3]=-9.1191291391491e-36;break}default:p[j+(q<<3)>>3]=d}q=q+1|0;if((q|0)==(b|0))break;else e=e+8|0}n=k[l>>2]|0;return n|0}}return 0}function um(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,q=0,r=0,s=0,t=0,u=0,v=0;o=c==1.0&d==0.0;q=(b|0)>0;a:do if(!e)if(o){if(q)r=0;else break;do{p[m+(r<<3)>>3]=+(l[a+r>>0]|0);r=r+1|0}while((r|0)!=(b|0))}else{if(q)s=0;else break;do{p[m+(s<<3)>>3]=+(l[a+s>>0]|0|0)*c+d;s=s+1|0}while((s|0)!=(b|0))}else{if(!o){if(!q)break;t=(e|0)==1;u=0;while(1){v=i[a+u>>0]|0;do if(v<<24>>24==f<<24>>24){k[j>>2]=1;if(t){p[m+(u<<3)>>3]=g;break}else{i[h+u>>0]=1;break}}else p[m+(u<<3)>>3]=+(v&255|0)*c+d;while(0);u=u+1|0;if((u|0)==(b|0))break a}}if(q)if((e|0)==1){u=0;do{t=i[a+u>>0]|0;if(t<<24>>24==f<<24>>24){k[j>>2]=1;p[m+(u<<3)>>3]=g}else p[m+(u<<3)>>3]=+(t&255);u=u+1|0}while((u|0)!=(b|0))}else{u=0;do{t=i[a+u>>0]|0;if(t<<24>>24==f<<24>>24){k[j>>2]=1;i[h+u>>0]=1}else p[m+(u<<3)>>3]=+(t&255);u=u+1|0}while((u|0)!=(b|0))}}while(0);return k[n>>2]|0}function vm(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,q=0,r=0,s=0,t=0,u=0,v=0;o=c==1.0&d==0.0;q=(b|0)>0;a:do if(!e)if(o){if(q)r=0;else break;do{p[m+(r<<3)>>3]=+(j[a+(r<<1)>>1]|0);r=r+1|0}while((r|0)!=(b|0))}else{if(q)s=0;else break;do{p[m+(s<<3)>>3]=+(j[a+(s<<1)>>1]|0)*c+d;s=s+1|0}while((s|0)!=(b|0))}else{if(!o){if(!q)break;t=(e|0)==1;u=0;while(1){v=j[a+(u<<1)>>1]|0;do if(v<<16>>16==f<<16>>16){k[l>>2]=1;if(t){p[m+(u<<3)>>3]=g;break}else{i[h+u>>0]=1;break}}else p[m+(u<<3)>>3]=+(v<<16>>16)*c+d;while(0);u=u+1|0;if((u|0)==(b|0))break a}}if(q)if((e|0)==1){u=0;do{t=j[a+(u<<1)>>1]|0;if(t<<16>>16==f<<16>>16){k[l>>2]=1;p[m+(u<<3)>>3]=g}else p[m+(u<<3)>>3]=+(t<<16>>16);u=u+1|0}while((u|0)!=(b|0))}else{u=0;do{t=j[a+(u<<1)>>1]|0;if(t<<16>>16==f<<16>>16){k[l>>2]=1;i[h+u>>0]=1}else p[m+(u<<3)>>3]=+(t<<16>>16);u=u+1|0}while((u|0)!=(b|0))}}while(0);return k[n>>2]|0}function wm(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,q=0,r=0,s=0,t=0,u=0;n=c==1.0&d==0.0;o=(b|0)>0;do if(!e)if(n){if(o)q=0;else break;do{p[l+(q<<3)>>3]=+(k[a+(q<<2)>>2]|0);q=q+1|0}while((q|0)!=(b|0))}else{if(o)r=0;else break;do{p[l+(r<<3)>>3]=+(k[a+(r<<2)>>2]|0)*c+d;r=r+1|0}while((r|0)!=(b|0))}else if(n){if(!o)break;s=(e|0)==1;t=0;do{u=k[a+(t<<2)>>2]|0;do if((u|0)==(f|0)){k[j>>2]=1;if(s){p[l+(t<<3)>>3]=g;break}else{i[h+t>>0]=1;break}}else p[l+(t<<3)>>3]=+(u|0);while(0);t=t+1|0}while((t|0)!=(b|0))}else{if(!o)break;t=(e|0)==1;s=0;do{u=k[a+(s<<2)>>2]|0;do if((u|0)==(f|0)){k[j>>2]=1;if(t){p[l+(s<<3)>>3]=g;break}else{i[h+s>>0]=1;break}}else p[l+(s<<3)>>3]=+(u|0)*c+d;while(0);s=s+1|0}while((s|0)!=(b|0))}while(0);return k[m>>2]|0}function xm(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=+h;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,q=0,r=0,s=0,t=0,u=0,v=0;o=c==1.0&d==0.0;q=(b|0)>0;if(!e)if(o){if(q)r=0;else{s=k[n>>2]|0;return s|0}do{t=a+(r<<3)|0;p[m+(r<<3)>>3]=+((k[t>>2]|0)>>>0)+4294967296.0*+(k[t+4>>2]|0);r=r+1|0}while((r|0)!=(b|0));s=k[n>>2]|0;return s|0}else{if(q)u=0;else{s=k[n>>2]|0;return s|0}do{r=a+(u<<3)|0;p[m+(u<<3)>>3]=(+((k[r>>2]|0)>>>0)+4294967296.0*+(k[r+4>>2]|0))*c+d;u=u+1|0}while((u|0)!=(b|0));s=k[n>>2]|0;return s|0}if(o){if(!q){s=k[n>>2]|0;return s|0}o=(e|0)==1;u=0;do{r=a+(u<<3)|0;t=k[r>>2]|0;v=k[r+4>>2]|0;do if((t|0)==(f|0)&(v|0)==(g|0)){k[l>>2]=1;if(o){p[m+(u<<3)>>3]=h;break}else{i[j+u>>0]=1;break}}else p[m+(u<<3)>>3]=+(t>>>0)+4294967296.0*+(v|0);while(0);u=u+1|0}while((u|0)!=(b|0));s=k[n>>2]|0;return s|0}else{if(!q){s=k[n>>2]|0;return s|0}q=(e|0)==1;e=0;do{u=a+(e<<3)|0;o=k[u>>2]|0;v=k[u+4>>2]|0;do if((o|0)==(f|0)&(v|0)==(g|0)){k[l>>2]=1;if(q){p[m+(e<<3)>>3]=h;break}else{i[j+e>>0]=1;break}}else p[m+(e<<3)>>3]=(+(o>>>0)+4294967296.0*+(v|0))*c+d;while(0);e=e+1|0}while((e|0)!=(b|0));s=k[n>>2]|0;return s|0}return 0}function ym(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,q=0,r=0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)q=0;else{r=k[l>>2]|0;return r|0}do{p[j+(q<<3)>>3]=+o[a+(q<<2)>>2];q=q+1|0}while((q|0)!=(b|0));r=k[l>>2]|0;return r|0}else{if(n)s=0;else{r=k[l>>2]|0;return r|0}do{p[j+(s<<3)>>3]=+o[a+(s<<2)>>2]*c+d;s=s+1|0}while((s|0)!=(b|0));r=k[l>>2]|0;return r|0}}s=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){r=k[l>>2]|0;return r|0}q=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32640;a:do switch(((v|0)==32640?1:(v|0)==0?2:0)|0){case 0:{p[j+(t<<3)>>3]=+o[a+(t<<2)>>2];break}case 1:{k[h>>2]=1;if(q){p[j+(t<<3)>>3]=f;break a}else{i[g+t>>0]=1;break a}break}default:p[j+(t<<3)>>3]=0.0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+4|0}r=k[l>>2]|0;return r|0}else{if(!n){r=k[l>>2]|0;return r|0}n=(e|0)==1;e=0;u=s;while(1){s=(m[u>>1]|0)&32640;b:do switch(((s|0)==32640?1:(s|0)==0?2:0)|0){case 0:{p[j+(e<<3)>>3]=+o[a+(e<<2)>>2]*c+d;break}case 1:{k[h>>2]=1;if(n){p[j+(e<<3)>>3]=f;break b}else{i[g+e>>0]=1;break b}break}default:p[j+(e<<3)>>3]=d}while(0);e=e+1|0;if((e|0)==(b|0))break;else u=u+4|0}r=k[l>>2]|0;return r|0}return 0}function zm(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=+j;l=l|0;m=m|0;n=n|0;o=o|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0,S=0,T=0,U=0,V=0.0,W=0,X=0,Z=0.0,_=0,$=0,aa=0.0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0.0,ia=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0;q=r;r=r+96|0;s=q;t=q+4|0;u=xx(h)|0;a:do if((b|0)>0){v=(g|0)==0;w=(g|0)==1;x=a;y=0;while(1){z=x+e|0;A=i[z>>0]|0;i[z>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,x,u)|0)==0:0)if(!v){k[m>>2]=1;if(w){p[n+(y<<3)>>3]=j;B=z;break}else{i[l+y>>0]=1;B=z;break}}else B=z;else{C=x;D=9}while(0);if((D|0)==9){b:while(1){D=0;E=i[C>>0]|0;switch(E<<24>>24){case 43:case 45:{F=E;G=C;D=11;break b;break}case 32:break;default:{H=E;I=C;J=1;break b}}C=C+1|0;D=9}if((D|0)==11){D=0;E=F<<24>>24==45;K=G;while(1){L=K+1|0;M=i[L>>0]|0;if(M<<24>>24==32)K=L;else{N=M;O=L;break}}H=N;I=O;J=E?-1:1}if((H+-48&255)<10){K=H;L=I;P=0.0;while(1){Q=P*10.0+ +(K<<24>>24);M=L;while(1){R=M+1|0;S=i[R>>0]|0;if(S<<24>>24==32)M=R;else{T=S;U=R;break}}V=Q+-48.0;if((T+-48&255)<10){K=T;L=U;P=V}else{W=T;X=U;Z=V;break}}}else{W=H;X=I;Z=0.0}switch(W<<24>>24){case 44:case 46:{L=X;while(1){K=L+1|0;E=i[K>>0]|0;if(E<<24>>24==32)L=K;else{_=E;$=K;break}}if((_+-48&255)<10){L=_;K=$;P=1.0;V=Z;while(1){aa=V*10.0+ +(L<<24>>24)+-48.0;E=K;while(1){M=E+1|0;R=i[M>>0]|0;if(R<<24>>24==32)E=M;else{ba=R;ca=M;break}}Q=P*10.0;if((ba+-48&255)<10){L=ba;K=ca;P=Q;V=aa}else{da=ba;ea=ca;fa=1;ga=Q;ia=aa;break}}}else{da=_;ea=$;fa=1;ga=1.0;ia=Z}break}default:{da=W;ea=X;fa=0;ga=1.0;ia=Z}}if((da&-2)<<24>>24==68){K=ea;c:while(1){L=K+1|0;E=i[L>>0]|0;switch(E<<24>>24){case 32:{K=L;break}case 43:case 45:{ja=E;ka=K;D=28;break c;break}default:{la=E;ma=L;na=1;break c}}}if((D|0)==28){D=0;K=ja<<24>>24==45;L=ka+2|0;while(1){E=i[L>>0]|0;if(E<<24>>24==32)L=L+1|0;else{oa=E;pa=L;break}}la=oa;ma=pa;na=K?-1:1}if((la+-48&255)<10){L=la;E=ma;M=0;while(1){R=(M*10|0)+-48|0;S=E;while(1){qa=S+1|0;ra=i[qa>>0]|0;if(ra<<24>>24==32)S=qa;else{sa=ra;ta=qa;break}}S=R+(L<<24>>24)|0;if((sa+-48&255)<10){L=sa;E=ta;M=S}else{ua=sa;va=ta;wa=na;xa=S;break}}}else{ua=la;va=ma;wa=na;xa=0}}else{ua=da;va=ea;wa=1;xa=0}if(ua<<24>>24){ya=z;za=A;Aa=x;break}aa=+(J|0)*ia/((fa|0)==0?f:ga)*+Y(10.0,+(+(ha(wa,xa)|0)))*c+d;p[n+(y<<3)>>3]=aa;B=va}i[z>>0]=A;y=y+1|0;if((y|0)>=(b|0))break a;else x=B}x=t;y=761193;w=x+36|0;do{i[x>>0]=i[y>>0]|0;x=x+1|0;y=y+1|0}while((x|0)<(w|0));tk(t);k[s>>2]=Aa;Xw(t,761229,s)|0;tk(t);i[ya>>0]=za;k[o>>2]=409;Ba=409;r=q;return Ba|0}while(0);Ba=k[o>>2]|0;r=q;return Ba|0}function Am(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=+g;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,p=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;Bm(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+4|0,i,j)|0;p=k[j>>2]|0;r=l;return p|0}else{o[m>>2]=g;qt(a,42,c,d,e,f,1,m,h,0,i,j)|0;p=k[j>>2]|0;r=l;return p|0}return 0}function Bm(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=+m;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,j>>31,x,y,K,D,z,A,G,H,E,F,J,B,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[A>>2]|0;A=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28920|0,t+28904|0,C,s)|0,e=k[C>>2]|0,(e|0)>0):0){C=0;R=1.0;while(1){S=R*10.0;C=C+1|0;if((C|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;e=k[z>>2]|0;if(!((l|0)==1&m==0.0)?(C=I,d=k[C>>2]|0,c=k[C+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0){C=ky(d|0,c|0,32768,0)|0;P=L;if(!((e|0)==21&(P>>>0>0|(P|0)==0&C>>>0>65535))?!((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)):0){U=(e|0)==16&(i[N>>0]|0)==1?0:l;V=16}else{W=A;X=f;Y=0;Z=0}}else{U=0;V=16}if((V|0)==16)if((e|0)==42){e=(h|0)<0|(h|0)==0&g>>>0<536870911;W=e?h:0;X=e?g:536870911;Y=(U|0)==0&+p[x>>3]==1.0&+p[y>>3]==0.0;Z=U}else{W=A;X=f;Y=0;Z=U}U=(j|0)>-1;f=((j|0)<0)<<31>>31;A=0-j|0;e=((A|0)<0)<<31>>31;l=0;d=0;c=g;g=h;h=0;C=0;a:while(1){P=C;_=c;$=g;aa=h;while(1){ba=n+(aa<<2)|0;ca=o+aa|0;da=_;ea=$;while(1){if((da|0)==0&(ea|0)==0){V=51;break a}fa=(ea|0)<(W|0)|(ea|0)==(W|0)&da>>>0<X>>>0;ga=fa?da:X;fa=my(0,ga|0,32)|0;ga=L;if(U){ia=F;ja=H;ka=k[ja>>2]|0;la=k[ja+4>>2]|0;ja=ky(k[ia>>2]|0,k[ia+4>>2]|0,-1,-1)|0;ia=jy(ja|0,L|0,ka|0,la|0)|0;ja=wy(ia|0,L|0,j|0,f|0)|0;ia=ky(ja|0,L|0,1,0)|0;ja=L;ma=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<ia>>>0;na=ka;oa=la;pa=ma?fa:ia;qa=ma?ga:ja}else{ja=H;ma=k[ja>>2]|0;ia=k[ja+4>>2]|0;ja=wy(ma|0,ia|0,A|0,e|0)|0;la=ky(ja|0,L|0,1,0)|0;ja=L;ka=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<la>>>0;na=ma;oa=ia;pa=ka?fa:la;qa=ka?ga:ja}ja=G;ga=k[ja>>2]|0;ka=k[ja+4>>2]|0;ja=J;la=yy(k[ja>>2]|0,k[ja+4>>2]|0,l|0,d|0)|0;ja=ky(la|0,L|0,ga|0,ka|0)|0;ka=L;ga=k[E>>2]|0;la=(ga|0)/(j|0)|0;fa=yy(la|0,((la|0)<0)<<31>>31|0,na|0,oa|0)|0;la=ky(ja|0,ka|0,fa|0,L|0)|0;fa=L;switch(k[z>>2]|0){case 42:{pu(a,la,fa,pa,ga,ba,s)|0;if(!Y)Fm(ba,pa,+p[x>>3],+p[y>>3],Z,m,ca,q,ba,s)|0;break}case 11:{lu(a,la,fa,pa,ga,O,s)|0;Gm(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]&255,m,ca,q,ba,s)|0;break}case 21:{mu(a,la,fa,pa,ga,O,s)|0;Hm(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]&65535,m,ca,q,ba,s)|0;break}case 41:{nu(a,la,fa,pa,ga,O,s)|0;Im(O,pa,+p[x>>3],+p[y>>3],Z,k[I>>2]|0,m,ca,q,ba,s)|0;break}case 81:{ou(a,la,fa,pa,ga,O,s)|0;ka=I;Jm(O,pa,+p[x>>3],+p[y>>3],Z,k[ka>>2]|0,k[ka+4>>2]|0,m,ca,q,ba,s)|0;break}case 82:{qu(a,la,fa,pa,ga,O,s)|0;Km(O,pa,+p[x>>3],+p[y>>3],Z,m,ca,q,ba,s)|0;break}case 16:{au(a,la,fa,0,s)|0;fa=k[E>>2]|0;la=k[D>>2]|0;if((fa|0)==(la|0)){ga=ha(fa,pa)|0;fu(a,ga,((ga|0)<0)<<31>>31,O,s)|0}else gu(a,la,pa,fa-la|0,O,s)|0;Lm(O,pa,+p[x>>3],+p[y>>3],k[D>>2]|0,T,Z,N,m,ca,q,ba,s)|0;break}default:{V=37;break a}}if((k[s>>2]|0)>0){ra=pa;sa=aa;ta=P;V=41;break a}la=((pa|0)<0)<<31>>31;fa=jy(da|0,ea|0,pa|0,la|0)|0;ga=L;if((da|0)==(pa|0)&(ea|0)==(la|0)){da=fa;ea=ga}else{ua=pa;va=pa;wa=la;xa=fa;ya=ga;break}}ea=ky(va|0,wa|0,aa|0,P|0)|0;da=L;ba=H;ca=ha(ua,j)|0;ga=ky(k[ba>>2]|0,k[ba+4>>2]|0,ca|0,((ca|0)<0)<<31>>31|0)|0;ca=L;ba=H;k[ba>>2]=ga;k[ba+4>>2]=ca;ba=F;fa=k[ba>>2]|0;la=k[ba+4>>2]|0;if(!((ca|0)<(la|0)|(ca|0)==(la|0)&ga>>>0<fa>>>0)){za=ga;Aa=ca;Ba=fa;Ca=la;Da=ea;Ea=da;Fa=xa;Ga=ya;V=47;break}if((ca|0)<0){Ha=ga;Ia=ca;Ja=fa;Ka=la;La=ea;Ma=da;Na=xa;Oa=ya;V=50;break}else{P=da;_=xa;$=ya;aa=ea}}if((V|0)==47){V=0;aa=wy(za|0,Aa|0,Ba|0,Ca|0)|0;$=L;_=ky(aa|0,$|0,l|0,d|0)|0;P=L;ea=yy(aa|0,$|0,Ba|0,Ca|0)|0;$=jy(za|0,Aa|0,ea|0,L|0)|0;Pa=$;Qa=L;Ra=Da;Sa=Ea;Ta=Fa;Ua=Ga;Va=_;Wa=P}else if((V|0)==50){V=0;P=wy(~Ha|0,~Ia|0,Ja|0,Ka|0)|0;_=ky(P|0,L|0,1,0)|0;P=L;$=jy(l|0,d|0,_|0,P|0)|0;ea=L;aa=yy(_|0,P|0,Ja|0,Ka|0)|0;P=ky(aa|0,L|0,Ha|0,Ia|0)|0;Pa=P;Qa=L;Ra=La;Sa=Ma;Ta=Na;Ua=Oa;Va=$;Wa=ea}ea=H;k[ea>>2]=Pa;k[ea+4>>2]=Qa;l=Va;d=Wa;c=Ta;g=Ua;h=Ra;C=Sa}if((V|0)==37){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[B>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((V|0)==41){m=+(sa>>>0)+4294967296.0*+(ta|0);T=m+1.0;R=m+ +(ra|0);if((k[B>>2]|0)>0){p[v>>3]=T;p[v+8>>3]=R;k[v+16>>2]=b;Xw(M,759428,v)|0}else{p[u>>3]=T;p[u+8>>3]=R;Xw(M,759491,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((V|0)==51){V=k[s>>2]|0;if((V|0)!=-11){Q=V;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function Cm(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;Bm(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0.0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,42,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function Dm(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=i|0;j=j|0;l=l|0;var m=0,n=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;m=r;r=r+480|0;n=m+232|0;p=m+224|0;q=m+356|0;s=m+320|0;t=m+284|0;u=m+248|0;v=m+144|0;w=m+72|0;x=m;y=m+244|0;z=m+240|0;A=m+392|0;B=m+393|0;C=m+236|0;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(B,759550,p)|0;tk(B);k[l>>2]=320;D=320;r=m;return D|0}if(Cl(a,l)|0){if((c|0)>0){p=0;do{E=k[e+(p<<2)>>2]|0;F=w+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(p<<2)>>2]|0;F=x+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}o[C>>2]=h;mt(a,42,w,x,g,1,C,i,0,j,l)|0;D=k[l>>2]|0;r=m;return D|0}if((Zk(a,y,l)|0)>0){D=k[l>>2]|0;r=m;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;p=v+8|0;E=p;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[n>>2]=xa+1;Xw(B,759595,n)|0;tk(B);k[l>>2]=321;D=321;r=m;return D|0}while(0);B=v+(c<<3)|0;v=B;n=k[u+(c<<2)>>2]|0;xa=yy(n|0,((n|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;n=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(n|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=p;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=n;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((Bm(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,i+(_b<<2)|0,A,z,l)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[j>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[l>>2]|0;r=m;return D|0}while(0);D=k[l>>2]|0;r=m;return D|0}function Em(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;o=ty(g|0,h|0,1)|0;p=L;q=ey(o,1)|0;r=ty(e|0,f|0,1)|0;f=ky(r|0,L|0,-1,-1)|0;Bm(a,b,c,d,f,L,o,p,1,2,0.0,j,q,m,n)|0;if((h|0)>0|(h|0)==0&g>>>0>0){s=0;t=0;u=0;v=0}else{dy(q);w=k[n>>2]|0;return w|0}do{if((i[q+s>>0]|0)==0?(i[q+(s|1)>>0]|0)==0:0)i[l+t>>0]=0;else i[l+t>>0]=1;s=ky(s|0,u|0,2,0)|0;u=L;t=ky(t|0,v|0,1,0)|0;v=L}while((v|0)<(h|0)|(v|0)==(h|0)&t>>>0<g>>>0);dy(q);w=k[n>>2]|0;return w|0}function Fm(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0,s=0,t=0,u=0.0;if(!e){if(c==1.0&d==0.0){qy(j|0,a|0,b<<2|0)|0;n=k[l>>2]|0;return n|0}if((b|0)>0)p=0;else{n=k[l>>2]|0;return n|0}do{o[j+(p<<2)>>2]=+o[a+(p<<2)>>2]*c+d;p=p+1|0}while((p|0)!=(b|0));n=k[l>>2]|0;return n|0}p=a+2|0;q=(b|0)>0;if(c==1.0&d==0.0){if(!q){n=k[l>>2]|0;return n|0}if((e|0)==1){r=0;s=p;while(1){t=(m[s>>1]|0)&32640;switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{k[j+(r<<2)>>2]=k[a+(r<<2)>>2];break}case 1:{k[h>>2]=1;o[j+(r<<2)>>2]=f;break}default:o[j+(r<<2)>>2]=0.0}r=r+1|0;if((r|0)==(b|0))break;else s=s+4|0}n=k[l>>2]|0;return n|0}else{s=0;r=p;while(1){t=(m[r>>1]|0)&32640;switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{k[j+(s<<2)>>2]=k[a+(s<<2)>>2];break}case 1:{k[h>>2]=1;i[g+s>>0]=1;o[j+(s<<2)>>2]=-9.119119840596153e-36;break}default:o[j+(s<<2)>>2]=0.0}s=s+1|0;if((s|0)==(b|0))break;else r=r+4|0}n=k[l>>2]|0;return n|0}}else{if(!q){n=k[l>>2]|0;return n|0}u=d;if((e|0)==1){e=0;q=p;while(1){r=(m[q>>1]|0)&32640;switch(((r|0)==32640?1:(r|0)==0?2:0)|0){case 0:{o[j+(e<<2)>>2]=+o[a+(e<<2)>>2]*c+d;break}case 1:{k[h>>2]=1;o[j+(e<<2)>>2]=f;break}default:o[j+(e<<2)>>2]=u}e=e+1|0;if((e|0)==(b|0))break;else q=q+4|0}n=k[l>>2]|0;return n|0}else{q=0;e=p;while(1){p=(m[e>>1]|0)&32640;switch(((p|0)==32640?1:(p|0)==0?2:0)|0){case 0:{o[j+(q<<2)>>2]=+o[a+(q<<2)>>2]*c+d;break}case 1:{k[h>>2]=1;i[g+q>>0]=1;o[j+(q<<2)>>2]=-9.119119840596153e-36;break}default:o[j+(q<<2)>>2]=u}q=q+1|0;if((q|0)==(b|0))break;else e=e+4|0}n=k[l>>2]|0;return n|0}}return 0}function Gm(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;j=j|0;m=m|0;n=n|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0;p=c==1.0&d==0.0;q=(b|0)>0;a:do if(!e)if(p){if(q)r=0;else break;do{o[m+(r<<2)>>2]=+(l[a+r>>0]|0);r=r+1|0}while((r|0)!=(b|0))}else{if(q)s=0;else break;do{o[m+(s<<2)>>2]=+(l[a+s>>0]|0)*c+d;s=s+1|0}while((s|0)!=(b|0))}else{if(!p){if(!q)break;t=(e|0)==1;u=0;while(1){v=i[a+u>>0]|0;do if(v<<24>>24==f<<24>>24){k[j>>2]=1;if(t){o[m+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else o[m+(u<<2)>>2]=+(v&255)*c+d;while(0);u=u+1|0;if((u|0)==(b|0))break a}}if(q)if((e|0)==1){u=0;do{t=i[a+u>>0]|0;if(t<<24>>24==f<<24>>24){k[j>>2]=1;o[m+(u<<2)>>2]=g}else o[m+(u<<2)>>2]=+(t&255);u=u+1|0}while((u|0)!=(b|0))}else{u=0;do{t=i[a+u>>0]|0;if(t<<24>>24==f<<24>>24){k[j>>2]=1;i[h+u>>0]=1}else o[m+(u<<2)>>2]=+(t&255);u=u+1|0}while((u|0)!=(b|0))}}while(0);return k[n>>2]|0}function Hm(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;l=l|0;m=m|0;n=n|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0;p=c==1.0&d==0.0;q=(b|0)>0;a:do if(!e)if(p){if(q)r=0;else break;do{o[m+(r<<2)>>2]=+(j[a+(r<<1)>>1]|0);r=r+1|0}while((r|0)!=(b|0))}else{if(q)s=0;else break;do{o[m+(s<<2)>>2]=+(j[a+(s<<1)>>1]|0)*c+d;s=s+1|0}while((s|0)!=(b|0))}else{if(!p){if(!q)break;t=(e|0)==1;u=0;while(1){v=j[a+(u<<1)>>1]|0;do if(v<<16>>16==f<<16>>16){k[l>>2]=1;if(t){o[m+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else o[m+(u<<2)>>2]=+(v<<16>>16)*c+d;while(0);u=u+1|0;if((u|0)==(b|0))break a}}if(q)if((e|0)==1){u=0;do{t=j[a+(u<<1)>>1]|0;if(t<<16>>16==f<<16>>16){k[l>>2]=1;o[m+(u<<2)>>2]=g}else o[m+(u<<2)>>2]=+(t<<16>>16);u=u+1|0}while((u|0)!=(b|0))}else{u=0;do{t=j[a+(u<<1)>>1]|0;if(t<<16>>16==f<<16>>16){k[l>>2]=1;i[h+u>>0]=1}else o[m+(u<<2)>>2]=+(t<<16>>16);u=u+1|0}while((u|0)!=(b|0))}}while(0);return k[n>>2]|0}function Im(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,p=0,q=0,r=0,s=0,t=0,u=0;n=c==1.0&d==0.0;p=(b|0)>0;do if(!e)if(n){if(p)q=0;else break;do{o[l+(q<<2)>>2]=+(k[a+(q<<2)>>2]|0);q=q+1|0}while((q|0)!=(b|0))}else{if(p)r=0;else break;do{o[l+(r<<2)>>2]=+(k[a+(r<<2)>>2]|0)*c+d;r=r+1|0}while((r|0)!=(b|0))}else if(n){if(!p)break;s=(e|0)==1;t=0;do{u=k[a+(t<<2)>>2]|0;do if((u|0)==(f|0)){k[j>>2]=1;if(s){o[l+(t<<2)>>2]=g;break}else{i[h+t>>0]=1;break}}else o[l+(t<<2)>>2]=+(u|0);while(0);t=t+1|0}while((t|0)!=(b|0))}else{if(!p)break;t=(e|0)==1;s=0;do{u=k[a+(s<<2)>>2]|0;do if((u|0)==(f|0)){k[j>>2]=1;if(t){o[l+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else o[l+(s<<2)>>2]=+(u|0)*c+d;while(0);s=s+1|0}while((s|0)!=(b|0))}while(0);return k[m>>2]|0}function Jm(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=+h;j=j|0;l=l|0;m=m|0;n=n|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e)if(p){if(q)r=0;else{s=k[n>>2]|0;return s|0}do{t=a+(r<<3)|0;o[m+(r<<2)>>2]=+((k[t>>2]|0)>>>0)+4294967296.0*+(k[t+4>>2]|0);r=r+1|0}while((r|0)!=(b|0));s=k[n>>2]|0;return s|0}else{if(q)u=0;else{s=k[n>>2]|0;return s|0}do{r=a+(u<<3)|0;o[m+(u<<2)>>2]=(+((k[r>>2]|0)>>>0)+4294967296.0*+(k[r+4>>2]|0))*c+d;u=u+1|0}while((u|0)!=(b|0));s=k[n>>2]|0;return s|0}if(p){if(!q){s=k[n>>2]|0;return s|0}p=(e|0)==1;u=0;do{r=a+(u<<3)|0;t=k[r>>2]|0;v=k[r+4>>2]|0;do if((t|0)==(f|0)&(v|0)==(g|0)){k[l>>2]=1;if(p){o[m+(u<<2)>>2]=h;break}else{i[j+u>>0]=1;break}}else o[m+(u<<2)>>2]=+(t>>>0)+4294967296.0*+(v|0);while(0);u=u+1|0}while((u|0)!=(b|0));s=k[n>>2]|0;return s|0}else{if(!q){s=k[n>>2]|0;return s|0}q=(e|0)==1;e=0;do{u=a+(e<<3)|0;p=k[u>>2]|0;v=k[u+4>>2]|0;do if((p|0)==(f|0)&(v|0)==(g|0)){k[l>>2]=1;if(q){o[m+(e<<2)>>2]=h;break}else{i[j+e>>0]=1;break}}else o[m+(e<<2)>>2]=(+(p>>>0)+4294967296.0*+(v|0))*c+d;while(0);e=e+1|0}while((e|0)!=(b|0));s=k[n>>2]|0;return s|0}return 0}function Km(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)q=0;else{r=k[l>>2]|0;return r|0}do{o[j+(q<<2)>>2]=+p[a+(q<<3)>>3];q=q+1|0}while((q|0)!=(b|0));r=k[l>>2]|0;return r|0}else{if(n)s=0;else{r=k[l>>2]|0;return r|0}do{o[j+(s<<2)>>2]=+p[a+(s<<3)>>3]*c+d;s=s+1|0}while((s|0)!=(b|0));r=k[l>>2]|0;return r|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){r=k[l>>2]|0;return r|0}q=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{o[j+(t<<2)>>2]=+p[a+(t<<3)>>3];break}case 1:{k[h>>2]=1;if(q){o[j+(t<<2)>>2]=f;break a}else{i[g+t>>0]=1;break a}break}default:o[j+(t<<2)>>2]=0.0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}r=k[l>>2]|0;return r|0}else{if(!n){r=k[l>>2]|0;return r|0}w=d;n=(e|0)==1;e=0;u=s;while(1){s=(m[u>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{o[j+(e<<2)>>2]=+p[a+(e<<3)>>3]*c+d;break}case 1:{k[h>>2]=1;if(n){o[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:o[j+(e<<2)>>2]=w}while(0);e=e+1|0;if((e|0)==(b|0))break;else u=u+8|0}r=k[l>>2]|0;return r|0}return 0}function Lm(a,b,c,d,e,f,g,h,j,l,m,n,p){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=+j;l=l|0;m=m|0;n=n|0;p=p|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0,S=0,T=0,U=0,V=0.0,W=0,X=0,Z=0.0,_=0,$=0,aa=0.0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0.0,ia=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0;q=r;r=r+96|0;s=q;t=q+4|0;u=xx(h)|0;a:do if((b|0)>0){v=(g|0)==0;w=(g|0)==1;x=a;y=0;while(1){z=x+e|0;A=i[z>>0]|0;i[z>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,x,u)|0)==0:0)if(!v){k[m>>2]=1;if(w){o[n+(y<<2)>>2]=j;B=z;break}else{i[l+y>>0]=1;B=z;break}}else B=z;else{C=x;D=9}while(0);if((D|0)==9){b:while(1){D=0;E=i[C>>0]|0;switch(E<<24>>24){case 43:case 45:{F=E;G=C;D=11;break b;break}case 32:break;default:{H=E;I=C;J=1;break b}}C=C+1|0;D=9}if((D|0)==11){D=0;E=F<<24>>24==45;K=G;while(1){L=K+1|0;M=i[L>>0]|0;if(M<<24>>24==32)K=L;else{N=M;O=L;break}}H=N;I=O;J=E?-1:1}if((H+-48&255)<10){K=H;L=I;P=0.0;while(1){Q=P*10.0+ +(K<<24>>24);M=L;while(1){R=M+1|0;S=i[R>>0]|0;if(S<<24>>24==32)M=R;else{T=S;U=R;break}}V=Q+-48.0;if((T+-48&255)<10){K=T;L=U;P=V}else{W=T;X=U;Z=V;break}}}else{W=H;X=I;Z=0.0}switch(W<<24>>24){case 44:case 46:{L=X;while(1){K=L+1|0;E=i[K>>0]|0;if(E<<24>>24==32)L=K;else{_=E;$=K;break}}if((_+-48&255)<10){L=_;K=$;P=1.0;V=Z;while(1){aa=V*10.0+ +(L<<24>>24)+-48.0;E=K;while(1){M=E+1|0;R=i[M>>0]|0;if(R<<24>>24==32)E=M;else{ba=R;ca=M;break}}Q=P*10.0;if((ba+-48&255)<10){L=ba;K=ca;P=Q;V=aa}else{da=ba;ea=ca;fa=1;ga=Q;ia=aa;break}}}else{da=_;ea=$;fa=1;ga=1.0;ia=Z}break}default:{da=W;ea=X;fa=0;ga=1.0;ia=Z}}if((da&-2)<<24>>24==68){K=ea;c:while(1){L=K+1|0;E=i[L>>0]|0;switch(E<<24>>24){case 32:{K=L;break}case 43:case 45:{ja=E;ka=K;D=28;break c;break}default:{la=E;ma=L;na=1;break c}}}if((D|0)==28){D=0;K=ja<<24>>24==45;L=ka+2|0;while(1){E=i[L>>0]|0;if(E<<24>>24==32)L=L+1|0;else{oa=E;pa=L;break}}la=oa;ma=pa;na=K?-1:1}if((la+-48&255)<10){L=la;E=ma;M=0;while(1){R=(M*10|0)+-48|0;S=E;while(1){qa=S+1|0;ra=i[qa>>0]|0;if(ra<<24>>24==32)S=qa;else{sa=ra;ta=qa;break}}S=R+(L<<24>>24)|0;if((sa+-48&255)<10){L=sa;E=ta;M=S}else{ua=sa;va=ta;wa=na;xa=S;break}}}else{ua=la;va=ma;wa=na;xa=0}}else{ua=da;va=ea;wa=1;xa=0}if(ua<<24>>24){ya=z;za=A;Aa=x;break}aa=+(J|0)*ia/((fa|0)==0?f:ga)*+Y(10.0,+(+(ha(wa,xa)|0)))*c+d;o[n+(y<<2)>>2]=aa;B=va}i[z>>0]=A;y=y+1|0;if((y|0)>=(b|0))break a;else x=B}x=t;y=761193;w=x+36|0;do{i[x>>0]=i[y>>0]|0;x=x+1|0;y=y+1|0}while((x|0)<(w|0));tk(t);k[s>>2]=Aa;Xw(t,761229,s)|0;tk(t);i[ya>>0]=za;k[p>>2]=409;Ba=409;r=q;return Ba|0}while(0);Ba=k[p>>2]|0;r=q;return Ba|0}function Mm(a,b,c,d,e,f,g,h,i,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;var m=0,n=0,o=0,p=0;m=r;r=r+16|0;n=m;if(!(Cl(a,l)|0)){o=(b|0)<1;Nm(a,2,o?1:b,o?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,m+2|0,i,l)|0;p=k[l>>2]|0;r=m;return p|0}else{j[n>>1]=g;qt(a,21,c,d,e,f,1,n,h,0,i,l)|0;p=k[l>>2]|0;r=m;return p|0}return 0}function Nm(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,j>>31,x,y,K,D,z,A,G,H,E,F,J,B,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[A>>2]|0;A=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28920|0,t+28904|0,C,s)|0,e=k[C>>2]|0,(e|0)>0):0){C=0;R=1.0;while(1){S=R*10.0;C=C+1|0;if((C|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;e=k[z>>2]|0;if(!((l|0)==1&m<<16>>16==0)?(C=I,d=k[C>>2]|0,c=k[C+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0){C=ky(d|0,c|0,32768,0)|0;P=L;if(!((e|0)==21&(P>>>0>0|(P|0)==0&C>>>0>65535)))if((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)){U=A;V=f;W=0;X=0}else{Y=(e|0)==16&(i[N>>0]|0)==1?0:l;Z=16}else{_=0;Z=17}}else{Y=0;Z=16}if((Z|0)==16)if((e|0)==21){_=Y;Z=17}else{U=A;V=f;W=0;X=Y}if((Z|0)==17){Y=(h|0)<0|(h|0)==0&g>>>0<1073741823;U=Y?h:0;V=Y?g:1073741823;W=(_|0)==0&+p[x>>3]==1.0&+p[y>>3]==0.0;X=_}_=(j|0)>-1;Y=((j|0)<0)<<31>>31;f=0-j|0;A=((f|0)<0)<<31>>31;e=0;l=0;d=g;g=h;h=0;c=0;a:while(1){C=c;P=d;$=g;aa=h;while(1){ba=n+(aa<<1)|0;ca=o+aa|0;da=P;ea=$;while(1){if((da|0)==0&(ea|0)==0){Z=51;break a}fa=(ea|0)<(U|0)|(ea|0)==(U|0)&da>>>0<V>>>0;ga=fa?da:V;fa=my(0,ga|0,32)|0;ga=L;if(_){ia=F;ja=H;ka=k[ja>>2]|0;la=k[ja+4>>2]|0;ja=ky(k[ia>>2]|0,k[ia+4>>2]|0,-1,-1)|0;ia=jy(ja|0,L|0,ka|0,la|0)|0;ja=wy(ia|0,L|0,j|0,Y|0)|0;ia=ky(ja|0,L|0,1,0)|0;ja=L;ma=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<ia>>>0;na=ka;oa=la;pa=ma?fa:ia;qa=ma?ga:ja}else{ja=H;ma=k[ja>>2]|0;ia=k[ja+4>>2]|0;ja=wy(ma|0,ia|0,f|0,A|0)|0;la=ky(ja|0,L|0,1,0)|0;ja=L;ka=(ga|0)<(ja|0)|(ga|0)==(ja|0)&fa>>>0<la>>>0;na=ma;oa=ia;pa=ka?fa:la;qa=ka?ga:ja}ja=G;ga=k[ja>>2]|0;ka=k[ja+4>>2]|0;ja=J;la=yy(k[ja>>2]|0,k[ja+4>>2]|0,e|0,l|0)|0;ja=ky(la|0,L|0,ga|0,ka|0)|0;ka=L;ga=k[E>>2]|0;la=(ga|0)/(j|0)|0;fa=yy(la|0,((la|0)<0)<<31>>31|0,na|0,oa|0)|0;la=ky(ja|0,ka|0,fa|0,L|0)|0;fa=L;switch(k[z>>2]|0){case 21:{mu(a,la,fa,pa,ga,ba,s)|0;if(!W)Qm(ba,pa,+p[x>>3],+p[y>>3],X,k[I>>2]&65535,m,ca,q,ba,s)|0;break}case 81:{ou(a,la,fa,pa,ga,O,s)|0;ka=I;Rm(O,pa,+p[x>>3],+p[y>>3],X,k[ka>>2]|0,k[ka+4>>2]|0,m,ca,q,ba,s)|0;break}case 11:{lu(a,la,fa,pa,ga,O,s)|0;Sm(O,pa,+p[x>>3],+p[y>>3],X,k[I>>2]&255,m,ca,q,ba,s)|0;break}case 41:{nu(a,la,fa,pa,ga,O,s)|0;Tm(O,pa,+p[x>>3],+p[y>>3],X,k[I>>2]|0,m,ca,q,ba,s)|0;break}case 42:{pu(a,la,fa,pa,ga,O,s)|0;Um(O,pa,+p[x>>3],+p[y>>3],X,m,ca,q,ba,s)|0;break}case 82:{qu(a,la,fa,pa,ga,O,s)|0;Vm(O,pa,+p[x>>3],+p[y>>3],X,m,ca,q,ba,s)|0;break}case 16:{au(a,la,fa,0,s)|0;fa=k[E>>2]|0;la=k[D>>2]|0;if((fa|0)==(la|0)){ga=ha(fa,pa)|0;fu(a,ga,((ga|0)<0)<<31>>31,O,s)|0}else gu(a,la,pa,fa-la|0,O,s)|0;Wm(O,pa,+p[x>>3],+p[y>>3],k[D>>2]|0,T,X,N,m,ca,q,ba,s)|0;break}default:{Z=37;break a}}if((k[s>>2]|0)>0){ra=pa;sa=aa;ta=C;Z=41;break a}la=((pa|0)<0)<<31>>31;fa=jy(da|0,ea|0,pa|0,la|0)|0;ga=L;if((da|0)==(pa|0)&(ea|0)==(la|0)){da=fa;ea=ga}else{ua=pa;va=pa;wa=la;xa=fa;ya=ga;break}}ea=ky(va|0,wa|0,aa|0,C|0)|0;da=L;ba=H;ca=ha(ua,j)|0;ga=ky(k[ba>>2]|0,k[ba+4>>2]|0,ca|0,((ca|0)<0)<<31>>31|0)|0;ca=L;ba=H;k[ba>>2]=ga;k[ba+4>>2]=ca;ba=F;fa=k[ba>>2]|0;la=k[ba+4>>2]|0;if(!((ca|0)<(la|0)|(ca|0)==(la|0)&ga>>>0<fa>>>0)){za=ga;Aa=ca;Ba=fa;Ca=la;Da=ea;Ea=da;Fa=xa;Ga=ya;Z=47;break}if((ca|0)<0){Ha=ga;Ia=ca;Ja=fa;Ka=la;La=ea;Ma=da;Na=xa;Oa=ya;Z=50;break}else{C=da;P=xa;$=ya;aa=ea}}if((Z|0)==47){Z=0;aa=wy(za|0,Aa|0,Ba|0,Ca|0)|0;$=L;P=ky(aa|0,$|0,e|0,l|0)|0;C=L;ea=yy(aa|0,$|0,Ba|0,Ca|0)|0;$=jy(za|0,Aa|0,ea|0,L|0)|0;Pa=$;Qa=L;Ra=Da;Sa=Ea;Ta=Fa;Ua=Ga;Va=P;Wa=C}else if((Z|0)==50){Z=0;C=wy(~Ha|0,~Ia|0,Ja|0,Ka|0)|0;P=ky(C|0,L|0,1,0)|0;C=L;$=jy(e|0,l|0,P|0,C|0)|0;ea=L;aa=yy(P|0,C|0,Ja|0,Ka|0)|0;C=ky(aa|0,L|0,Ha|0,Ia|0)|0;Pa=C;Qa=L;Ra=La;Sa=Ma;Ta=Na;Ua=Oa;Va=$;Wa=ea}ea=H;k[ea>>2]=Pa;k[ea+4>>2]=Qa;e=Va;l=Wa;d=Ta;g=Ua;h=Ra;c=Sa}if((Z|0)==37){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[B>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((Z|0)==41){T=+(sa>>>0)+4294967296.0*+(ta|0);R=T+1.0;S=T+ +(ra|0);if((k[B>>2]|0)>0){p[v>>3]=R;p[v+8>>3]=S;k[v+16>>2]=b;Xw(M,759640,v)|0}else{p[u>>3]=R;p[u+8>>3]=S;Xw(M,759703,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((Z|0)==51){Z=k[s>>2]|0;if((Z|0)!=-11){Q=Z;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function Om(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;Nm(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,21,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function Pm(a,b,c,d,e,f,g,h,i,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;n=r;r=r+480|0;o=n+232|0;p=n+224|0;q=n+352|0;s=n+316|0;t=n+280|0;u=n+244|0;v=n+144|0;w=n+72|0;x=n;y=n+240|0;z=n+236|0;A=n+390|0;B=n+391|0;C=n+388|0;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(B,759762,p)|0;tk(B);k[m>>2]=320;D=320;r=n;return D|0}if(Cl(a,m)|0){if((c|0)>0){p=0;do{E=k[e+(p<<2)>>2]|0;F=w+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(p<<2)>>2]|0;F=x+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}j[C>>1]=h;mt(a,21,w,x,g,1,C,i,0,l,m)|0;D=k[m>>2]|0;r=n;return D|0}if((Zk(a,y,m)|0)>0){D=k[m>>2]|0;r=n;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(l|0)!=0;if(b)k[l>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;p=v+8|0;E=p;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[o>>2]=xa+1;Xw(B,759807,o)|0;tk(B);k[m>>2]=321;D=321;r=n;return D|0}while(0);B=v+(c<<3)|0;v=B;o=k[u+(c<<2)>>2]|0;xa=yy(o|0,((o|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;o=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(o|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=p;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=o;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((Nm(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,i+(_b<<1)|0,A,z,m)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[l>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[m>>2]|0;r=n;return D|0}while(0);D=k[m>>2]|0;r=n;return D|0}function Qm(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0.0,s=0;o=c==1.0&d==0.0;if(!e){if(o){qy(m|0,a|0,b<<1|0)|0;p=k[n>>2]|0;return p|0}if((b|0)>0)q=0;else{p=k[n>>2]|0;return p|0}do{r=+(j[a+(q<<1)>>1]|0)*c+d;do if(!(r<-32768.49))if(r>32767.49){k[n>>2]=-11;j[m+(q<<1)>>1]=32767;break}else{j[m+(q<<1)>>1]=~~r;break}else{k[n>>2]=-11;j[m+(q<<1)>>1]=-32768}while(0);q=q+1|0}while((q|0)!=(b|0));p=k[n>>2]|0;return p|0}q=(b|0)>0;if(o){if(!q){p=k[n>>2]|0;return p|0}if((e|0)==1){o=0;do{s=j[a+(o<<1)>>1]|0;if(s<<16>>16==f<<16>>16){k[l>>2]=1;j[m+(o<<1)>>1]=g}else j[m+(o<<1)>>1]=s;o=o+1|0}while((o|0)!=(b|0));p=k[n>>2]|0;return p|0}else{o=0;do{s=j[a+(o<<1)>>1]|0;if(s<<16>>16==f<<16>>16){k[l>>2]=1;i[h+o>>0]=1}else j[m+(o<<1)>>1]=s;o=o+1|0}while((o|0)!=(b|0));p=k[n>>2]|0;return p|0}}if(!q){p=k[n>>2]|0;return p|0}if((e|0)==1){e=0;do{q=j[a+(e<<1)>>1]|0;do if(q<<16>>16!=f<<16>>16){r=+(q<<16>>16)*c+d;if(r<-32768.49){k[n>>2]=-11;j[m+(e<<1)>>1]=-32768;break}if(r>32767.49){k[n>>2]=-11;j[m+(e<<1)>>1]=32767;break}else{j[m+(e<<1)>>1]=~~r;break}}else{k[l>>2]=1;j[m+(e<<1)>>1]=g}while(0);e=e+1|0}while((e|0)!=(b|0));p=k[n>>2]|0;return p|0}else{e=0;do{g=j[a+(e<<1)>>1]|0;do if(g<<16>>16!=f<<16>>16){r=+(g<<16>>16)*c+d;if(r<-32768.49){k[n>>2]=-11;j[m+(e<<1)>>1]=-32768;break}if(r>32767.49){k[n>>2]=-11;j[m+(e<<1)>>1]=32767;break}else{j[m+(e<<1)>>1]=~~r;break}}else{k[l>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));p=k[n>>2]|0;return p|0}return 0}function Rm(a,b,c,d,e,f,g,h,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e)if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{t=a+(r<<3)|0;u=k[t>>2]|0;v=k[t+4>>2]|0;do if(!((v|0)<-1|(v|0)==-1&u>>>0<4294934528))if((v|0)>0|(v|0)==0&u>>>0>32767){k[o>>2]=-11;j[n+(r<<1)>>1]=32767;break}else{j[n+(r<<1)>>1]=u;break}else{k[o>>2]=-11;j[n+(r<<1)>>1]=-32768}while(0);r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}else{if(q)w=0;else{s=k[o>>2]|0;return s|0}do{r=a+(w<<3)|0;x=(+((k[r>>2]|0)>>>0)+4294967296.0*+(k[r+4>>2]|0))*c+d;do if(!(x<-32768.49))if(x>32767.49){k[o>>2]=-11;j[n+(w<<1)>>1]=32767;break}else{j[n+(w<<1)>>1]=~~x;break}else{k[o>>2]=-11;j[n+(w<<1)>>1]=-32768}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;w=0;do{r=a+(w<<3)|0;u=k[r>>2]|0;v=k[r+4>>2]|0;do if((u|0)==(f|0)&(v|0)==(g|0)){k[m>>2]=1;if(p){j[n+(w<<1)>>1]=h;break}else{i[l+w>>0]=1;break}}else{x=(+(u>>>0)+4294967296.0*+(v|0))*c+d;if(x<-32768.49){k[o>>2]=-11;j[n+(w<<1)>>1]=-32768;break}if(x>32767.49){k[o>>2]=-11;j[n+(w<<1)>>1]=32767;break}else{j[n+(w<<1)>>1]=~~x;break}}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!q){s=k[o>>2]|0;return s|0}if((e|0)==1){e=0;do{q=a+(e<<3)|0;w=k[q>>2]|0;p=k[q+4>>2]|0;do if(!((w|0)==(f|0)&(p|0)==(g|0))){if((p|0)<-1|(p|0)==-1&w>>>0<4294934528){k[o>>2]=-11;j[n+(e<<1)>>1]=-32768;break}if((p|0)>0|(p|0)==0&w>>>0>32767){k[o>>2]=-11;j[n+(e<<1)>>1]=32767;break}else{j[n+(e<<1)>>1]=w;break}}else{k[m>>2]=1;j[n+(e<<1)>>1]=h}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}else{e=0;do{h=a+(e<<3)|0;w=k[h>>2]|0;p=k[h+4>>2]|0;do if(!((w|0)==(f|0)&(p|0)==(g|0))){if((p|0)<-1|(p|0)==-1&w>>>0<4294934528){k[o>>2]=-11;j[n+(e<<1)>>1]=-32768;break}if((p|0)>0|(p|0)==0&w>>>0>32767){k[o>>2]=-11;j[n+(e<<1)>>1]=32767;break}else{j[n+(e<<1)>>1]=w;break}}else{k[m>>2]=1;i[l+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}return 0}function Sm(a,b,c,d,e,f,g,h,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e){if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{j[n+(r<<1)>>1]=l[a+r>>0]|0;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q)t=0;else{s=k[o>>2]|0;return s|0}do{u=+(l[a+t>>0]|0|0)*c+d;do if(!(u<-32768.49))if(u>32767.49){k[o>>2]=-11;j[n+(t<<1)>>1]=32767;break}else{j[n+(t<<1)>>1]=~~u;break}else{k[o>>2]=-11;j[n+(t<<1)>>1]=-32768}while(0);t=t+1|0}while((t|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;t=0;do{r=i[a+t>>0]|0;do if(r<<24>>24==f<<24>>24){k[m>>2]=1;if(p){j[n+(t<<1)>>1]=g;break}else{i[h+t>>0]=1;break}}else j[n+(t<<1)>>1]=r&255;while(0);t=t+1|0}while((t|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{t=i[a+e>>0]|0;do if(t<<24>>24==f<<24>>24){k[m>>2]=1;if(q){j[n+(e<<1)>>1]=g;break}else{i[h+e>>0]=1;break}}else{u=+(t&255|0)*c+d;if(u<-32768.49){k[o>>2]=-11;j[n+(e<<1)>>1]=-32768;break}if(u>32767.49){k[o>>2]=-11;j[n+(e<<1)>>1]=32767;break}else{j[n+(e<<1)>>1]=~~u;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}function Tm(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=k[a+(q<<2)>>2]|0;do if((s|0)>=-32768)if((s|0)>32767){k[n>>2]=-11;j[m+(q<<1)>>1]=32767;break}else{j[m+(q<<1)>>1]=s;break}else{k[n>>2]=-11;j[m+(q<<1)>>1]=-32768}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)t=0;else{r=k[n>>2]|0;return r|0}do{u=+(k[a+(t<<2)>>2]|0)*c+d;do if(!(u<-32768.49))if(u>32767.49){k[n>>2]=-11;j[m+(t<<1)>>1]=32767;break}else{j[m+(t<<1)>>1]=~~u;break}else{k[n>>2]=-11;j[m+(t<<1)>>1]=-32768}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;t=0;do{q=k[a+(t<<2)>>2]|0;do if((q|0)==(f|0)){k[l>>2]=1;if(o){j[m+(t<<1)>>1]=g;break}else{i[h+t>>0]=1;break}}else{if((q|0)<-32768){k[n>>2]=-11;j[m+(t<<1)>>1]=-32768;break}if((q|0)>32767){k[n>>2]=-11;j[m+(t<<1)>>1]=32767;break}else{j[m+(t<<1)>>1]=q;break}}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{t=k[a+(e<<2)>>2]|0;do if((t|0)==(f|0)){k[l>>2]=1;if(p){j[m+(e<<1)>>1]=g;break}else{i[h+e>>0]=1;break}}else{u=+(t|0)*c+d;if(u<-32768.49){k[n>>2]=-11;j[m+(e<<1)>>1]=-32768;break}if(u>32767.49){k[n>>2]=-11;j[m+(e<<1)>>1]=32767;break}else{j[m+(e<<1)>>1]=~~u;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function Um(a,b,c,d,e,f,g,h,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;var p=0,q=0,r=0,s=0.0,t=0.0,u=0,v=0,w=0,x=0;if(!e){p=(b|0)>0;if(c==1.0&d==0.0){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(q<<2)>>2];t=s;do if(!(t<-32768.49))if(t>32767.49){k[n>>2]=-11;j[l+(q<<1)>>1]=32767;break}else{j[l+(q<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(q<<1)>>1]=-32768}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)u=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(u<<2)>>2]*c+d;do if(!(s<-32768.49))if(s>32767.49){k[n>>2]=-11;j[l+(u<<1)>>1]=32767;break}else{j[l+(u<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(u<<1)>>1]=-32768}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}}u=a+2|0;p=(b|0)>0;if(c==1.0&d==0.0){if(!p){r=k[n>>2]|0;return r|0}q=(e|0)==1;v=0;w=u;while(1){x=(m[w>>1]|0)&32640;a:do switch(((x|0)==32640?1:(x|0)==0?2:0)|0){case 0:{s=+o[a+(v<<2)>>2];t=s;if(t<-32768.49){k[n>>2]=-11;j[l+(v<<1)>>1]=-32768;break a}if(t>32767.49){k[n>>2]=-11;j[l+(v<<1)>>1]=32767;break a}else{j[l+(v<<1)>>1]=~~s;break a}break}case 1:{k[h>>2]=1;if(q){j[l+(v<<1)>>1]=f;break a}else{i[g+v>>0]=1;break a}break}default:j[l+(v<<1)>>1]=0}while(0);v=v+1|0;if((v|0)==(b|0))break;else w=w+4|0}r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=d<-32768.49;w=d>32767.49;v=~~d;q=(e|0)==1;e=0;x=u;while(1){u=(m[x>>1]|0)&32640;b:do switch(((u|0)==32640?1:(u|0)==0?2:0)|0){case 0:{s=+o[a+(e<<2)>>2]*c+d;if(s<-32768.49){k[n>>2]=-11;j[l+(e<<1)>>1]=-32768;break b}if(s>32767.49){k[n>>2]=-11;j[l+(e<<1)>>1]=32767;break b}else{j[l+(e<<1)>>1]=~~s;break b}break}case 1:{k[h>>2]=1;if(q){j[l+(e<<1)>>1]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(p){k[n>>2]=-11;j[l+(e<<1)>>1]=-32768;break b}if(w){k[n>>2]=-11;j[l+(e<<1)>>1]=32767;break b}else{j[l+(e<<1)>>1]=v;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else x=x+4|0}r=k[n>>2]|0;return r|0}function Vm(a,b,c,d,e,f,g,h,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;var o=0,q=0,r=0,s=0.0,t=0,u=0,v=0,w=0;if(!e){o=(b|0)>0;if(c==1.0&d==0.0){if(o)q=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(q<<3)>>3];do if(!(s<-32768.49))if(s>32767.49){k[n>>2]=-11;j[l+(q<<1)>>1]=32767;break}else{j[l+(q<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(q<<1)>>1]=-32768}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(o)t=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(t<<3)>>3]*c+d;do if(!(s<-32768.49))if(s>32767.49){k[n>>2]=-11;j[l+(t<<1)>>1]=32767;break}else{j[l+(t<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(t<<1)>>1]=-32768}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}}t=a+6|0;o=(b|0)>0;if(c==1.0&d==0.0){if(!o){r=k[n>>2]|0;return r|0}q=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32752;a:do switch(((w|0)==32752?1:(w|0)==0?2:0)|0){case 0:{s=+p[a+(u<<3)>>3];if(s<-32768.49){k[n>>2]=-11;j[l+(u<<1)>>1]=-32768;break a}if(s>32767.49){k[n>>2]=-11;j[l+(u<<1)>>1]=32767;break a}else{j[l+(u<<1)>>1]=~~s;break a}break}case 1:{k[h>>2]=1;if(q){j[l+(u<<1)>>1]=f;break a}else{i[g+u>>0]=1;break a}break}default:j[l+(u<<1)>>1]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+8|0}r=k[n>>2]|0;return r|0}if(!o){r=k[n>>2]|0;return r|0}o=d<-32768.49;v=d>32767.49;u=~~d;q=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32752;b:do switch(((t|0)==32752?1:(t|0)==0?2:0)|0){case 0:{s=+p[a+(e<<3)>>3]*c+d;if(s<-32768.49){k[n>>2]=-11;j[l+(e<<1)>>1]=-32768;break b}if(s>32767.49){k[n>>2]=-11;j[l+(e<<1)>>1]=32767;break b}else{j[l+(e<<1)>>1]=~~s;break b}break}case 1:{k[h>>2]=1;if(q){j[l+(e<<1)>>1]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(o){k[n>>2]=-11;j[l+(e<<1)>>1]=-32768;break b}if(v){k[n>>2]=-11;j[l+(e<<1)>>1]=32767;break b}else{j[l+(e<<1)>>1]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+8|0}r=k[n>>2]|0;return r|0}function Wm(a,b,c,d,e,f,g,h,l,m,n,o,p){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0,S=0,T=0,U=0,V=0.0,W=0,X=0,Z=0.0,_=0,$=0,aa=0.0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0.0,ia=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0;q=r;r=r+96|0;s=q;t=q+4|0;u=xx(h)|0;a:do if((b|0)>0){v=(g|0)==0;w=(g|0)==1;x=a;y=0;b:while(1){z=x+e|0;A=i[z>>0]|0;i[z>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,x,u)|0)==0:0)if(!v){k[n>>2]=1;if(w){j[o+(y<<1)>>1]=l;B=z;break}else{i[m+y>>0]=1;B=z;break}}else B=z;else{C=x;D=9}while(0);do if((D|0)==9){c:while(1){D=0;E=i[C>>0]|0;switch(E<<24>>24){case 43:case 45:{F=E;G=C;D=11;break c;break}case 32:break;default:{H=E;I=C;J=1;break c}}C=C+1|0;D=9}if((D|0)==11){D=0;E=F<<24>>24==45;K=G;while(1){L=K+1|0;M=i[L>>0]|0;if(M<<24>>24==32)K=L;else{N=M;O=L;break}}H=N;I=O;J=E?-1:1}if((H+-48&255)<10){K=H;L=I;P=0.0;while(1){Q=P*10.0+ +(K<<24>>24);M=L;while(1){R=M+1|0;S=i[R>>0]|0;if(S<<24>>24==32)M=R;else{T=S;U=R;break}}V=Q+-48.0;if((T+-48&255)<10){K=T;L=U;P=V}else{W=T;X=U;Z=V;break}}}else{W=H;X=I;Z=0.0}switch(W<<24>>24){case 44:case 46:{L=X;while(1){K=L+1|0;E=i[K>>0]|0;if(E<<24>>24==32)L=K;else{_=E;$=K;break}}if((_+-48&255)<10){L=_;K=$;P=1.0;V=Z;while(1){aa=V*10.0+ +(L<<24>>24)+-48.0;E=K;while(1){M=E+1|0;R=i[M>>0]|0;if(R<<24>>24==32)E=M;else{ba=R;ca=M;break}}Q=P*10.0;if((ba+-48&255)<10){L=ba;K=ca;P=Q;V=aa}else{da=ba;ea=ca;fa=1;ga=Q;ia=aa;break}}}else{da=_;ea=$;fa=1;ga=1.0;ia=Z}break}default:{da=W;ea=X;fa=0;ga=1.0;ia=Z}}if((da&-2)<<24>>24==68){K=ea;d:while(1){L=K+1|0;E=i[L>>0]|0;switch(E<<24>>24){case 32:{K=L;break}case 43:case 45:{ja=E;ka=K;D=28;break d;break}default:{la=E;ma=L;na=1;break d}}}if((D|0)==28){D=0;K=ja<<24>>24==45;L=ka+2|0;while(1){E=i[L>>0]|0;if(E<<24>>24==32)L=L+1|0;else{oa=E;pa=L;break}}la=oa;ma=pa;na=K?-1:1}if((la+-48&255)<10){L=la;E=ma;M=0;while(1){R=(M*10|0)+-48|0;S=E;while(1){qa=S+1|0;ra=i[qa>>0]|0;if(ra<<24>>24==32)S=qa;else{sa=ra;ta=qa;break}}S=R+(L<<24>>24)|0;if((sa+-48&255)<10){L=sa;E=ta;M=S}else{ua=sa;va=ta;wa=na;xa=S;break}}}else{ua=la;va=ma;wa=na;xa=0}}else{ua=da;va=ea;wa=1;xa=0}if(ua<<24>>24){ya=z;za=A;Aa=x;break b}aa=+(J|0)*ia/((fa|0)==0?f:ga)*+Y(10.0,+(+(ha(wa,xa)|0)))*c+d;if(aa<-32768.49){k[p>>2]=-11;j[o+(y<<1)>>1]=-32768;B=va;break}if(aa>32767.49){k[p>>2]=-11;j[o+(y<<1)>>1]=32767;B=va;break}else{j[o+(y<<1)>>1]=~~aa;B=va;break}}while(0);i[z>>0]=A;y=y+1|0;if((y|0)>=(b|0))break a;else x=B}x=t;y=761193;w=x+36|0;do{i[x>>0]=i[y>>0]|0;x=x+1|0;y=y+1|0}while((x|0)<(w|0));tk(t);k[s>>2]=Aa;Xw(t,761229,s)|0;tk(t);i[ya>>0]=za;k[p>>2]=409;Ba=409;r=q;return Ba|0}while(0);Ba=k[p>>2]|0;r=q;return Ba|0}function Xm(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;Ym(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+4|0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}else{k[m>>2]=g;qt(a,41,c,d,e,f,1,m,h,0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}return 0}function Ym(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,j>>31,x,y,K,D,z,A,G,H,E,F,J,B,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[A>>2]|0;A=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28920|0,t+28904|0,C,s)|0,e=k[C>>2]|0,(e|0)>0):0){C=0;R=1.0;while(1){S=R*10.0;C=C+1|0;if((C|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;if(((!((l|0)==1&(m|0)==0)?(e=k[z>>2]|0,C=I,d=k[C>>2]|0,c=k[C+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0)?(C=ky(d|0,c|0,32768,0)|0,P=L,!((e|0)==21&(P>>>0>0|(P|0)==0&C>>>0>65535))):0)?!((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)):0)U=(e|0)==16&(i[N>>0]|0)==1?0:l;else U=0;l=(j|0)>-1;e=((j|0)<0)<<31>>31;d=0-j|0;c=((d|0)<0)<<31>>31;C=0;P=0;V=g;g=h;h=0;W=0;a:while(1){X=W;Y=V;Z=g;_=h;while(1){$=o+_|0;aa=n+(_<<2)|0;ba=Y;ca=Z;while(1){if((ba|0)==0&(ca|0)==0){da=48;break a}ea=(ca|0)<(A|0)|(ca|0)==(A|0)&ba>>>0<f>>>0;fa=ea?ba:f;ea=my(0,fa|0,32)|0;fa=L;if(l){ga=F;ia=H;ja=k[ia>>2]|0;ka=k[ia+4>>2]|0;ia=ky(k[ga>>2]|0,k[ga+4>>2]|0,-1,-1)|0;ga=jy(ia|0,L|0,ja|0,ka|0)|0;ia=wy(ga|0,L|0,j|0,e|0)|0;ga=ky(ia|0,L|0,1,0)|0;ia=L;la=(fa|0)<(ia|0)|(fa|0)==(ia|0)&ea>>>0<ga>>>0;ma=ja;na=ka;oa=la?ea:ga;pa=la?fa:ia}else{ia=H;la=k[ia>>2]|0;ga=k[ia+4>>2]|0;ia=wy(la|0,ga|0,d|0,c|0)|0;ka=ky(ia|0,L|0,1,0)|0;ia=L;ja=(fa|0)<(ia|0)|(fa|0)==(ia|0)&ea>>>0<ka>>>0;ma=la;na=ga;oa=ja?ea:ka;pa=ja?fa:ia}ia=G;fa=k[ia>>2]|0;ja=k[ia+4>>2]|0;ia=J;ka=yy(k[ia>>2]|0,k[ia+4>>2]|0,C|0,P|0)|0;ia=ky(ka|0,L|0,fa|0,ja|0)|0;ja=L;fa=k[E>>2]|0;ka=(fa|0)/(j|0)|0;ea=yy(ka|0,((ka|0)<0)<<31>>31|0,ma|0,na|0)|0;ka=ky(ia|0,ja|0,ea|0,L|0)|0;ea=L;switch(k[z>>2]|0){case 41:{nu(a,ka,ea,oa,fa,O,s)|0;bn(O,oa,+p[x>>3],+p[y>>3],U,k[I>>2]|0,m,$,q,aa,s)|0;break}case 81:{ou(a,ka,ea,oa,fa,O,s)|0;ja=I;cn(O,oa,+p[x>>3],+p[y>>3],U,k[ja>>2]|0,k[ja+4>>2]|0,m,$,q,aa,s)|0;break}case 11:{lu(a,ka,ea,oa,fa,O,s)|0;dn(O,oa,+p[x>>3],+p[y>>3],U,k[I>>2]&255,m,$,q,aa,s)|0;break}case 21:{mu(a,ka,ea,oa,fa,O,s)|0;en(O,oa,+p[x>>3],+p[y>>3],U,k[I>>2]&65535,m,$,q,aa,s)|0;break}case 42:{pu(a,ka,ea,oa,fa,O,s)|0;fn(O,oa,+p[x>>3],+p[y>>3],U,m,$,q,aa,s)|0;break}case 82:{qu(a,ka,ea,oa,fa,O,s)|0;gn(O,oa,+p[x>>3],+p[y>>3],U,m,$,q,aa,s)|0;break}case 16:{au(a,ka,ea,0,s)|0;ea=k[E>>2]|0;ka=k[D>>2]|0;if((ea|0)==(ka|0)){fa=ha(ea,oa)|0;fu(a,fa,((fa|0)<0)<<31>>31,O,s)|0}else gu(a,ka,oa,ea-ka|0,O,s)|0;hn(O,oa,+p[x>>3],+p[y>>3],k[D>>2]|0,T,U,N,m,$,q,aa,s)|0;break}default:{da=34;break a}}if((k[s>>2]|0)>0){qa=oa;ra=_;sa=X;da=38;break a}ka=((oa|0)<0)<<31>>31;ea=jy(ba|0,ca|0,oa|0,ka|0)|0;fa=L;if((ba|0)==(oa|0)&(ca|0)==(ka|0)){ba=ea;ca=fa}else{ta=oa;ua=oa;va=ka;wa=ea;xa=fa;break}}ca=ky(ua|0,va|0,_|0,X|0)|0;ba=L;aa=H;$=ha(ta,j)|0;fa=ky(k[aa>>2]|0,k[aa+4>>2]|0,$|0,(($|0)<0)<<31>>31|0)|0;$=L;aa=H;k[aa>>2]=fa;k[aa+4>>2]=$;aa=F;ea=k[aa>>2]|0;ka=k[aa+4>>2]|0;if(!(($|0)<(ka|0)|($|0)==(ka|0)&fa>>>0<ea>>>0)){ya=fa;za=$;Aa=ea;Ba=ka;Ca=ca;Da=ba;Ea=wa;Fa=xa;da=44;break}if(($|0)<0){Ga=fa;Ha=$;Ia=ea;Ja=ka;Ka=ca;La=ba;Ma=wa;Na=xa;da=47;break}else{X=ba;Y=wa;Z=xa;_=ca}}if((da|0)==44){da=0;_=wy(ya|0,za|0,Aa|0,Ba|0)|0;Z=L;Y=ky(_|0,Z|0,C|0,P|0)|0;X=L;ca=yy(_|0,Z|0,Aa|0,Ba|0)|0;Z=jy(ya|0,za|0,ca|0,L|0)|0;Oa=Z;Pa=L;Qa=Ca;Ra=Da;Sa=Ea;Ta=Fa;Ua=Y;Va=X}else if((da|0)==47){da=0;X=wy(~Ga|0,~Ha|0,Ia|0,Ja|0)|0;Y=ky(X|0,L|0,1,0)|0;X=L;Z=jy(C|0,P|0,Y|0,X|0)|0;ca=L;_=yy(Y|0,X|0,Ia|0,Ja|0)|0;X=ky(_|0,L|0,Ga|0,Ha|0)|0;Oa=X;Pa=L;Qa=Ka;Ra=La;Sa=Ma;Ta=Na;Ua=Z;Va=ca}ca=H;k[ca>>2]=Oa;k[ca+4>>2]=Pa;C=Ua;P=Va;V=Sa;g=Ta;h=Qa;W=Ra}if((da|0)==34){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[B>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((da|0)==38){T=+(ra>>>0)+4294967296.0*+(sa|0);R=T+1.0;S=T+ +(qa|0);if((k[B>>2]|0)>0){p[v>>3]=R;p[v+8>>3]=S;k[v+16>>2]=b;Xw(M,759852,v)|0}else{p[u>>3]=R;p[u+8>>3]=S;Xw(M,759915,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((da|0)==48){da=k[s>>2]|0;if((da|0)!=-11){Q=da;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function Zm(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;Ym(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,41,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function _m(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0;m=r;r=r+480|0;n=m+232|0;o=m+224|0;p=m+356|0;q=m+320|0;s=m+284|0;t=m+248|0;u=m+144|0;v=m+72|0;w=m;x=m+244|0;y=m+240|0;z=m+392|0;A=m+393|0;B=m+236|0;if((c+-1|0)>>>0>8){k[o>>2]=c;Xw(A,760019,o)|0;tk(A);k[l>>2]=320;C=320;r=m;return C|0}if(Cl(a,l)|0){if((c|0)>0){o=0;do{D=k[e+(o<<2)>>2]|0;E=v+(o<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;D=k[f+(o<<2)>>2]|0;E=w+(o<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;o=o+1|0}while((o|0)!=(c|0))}k[B>>2]=h;mt(a,41,v,w,g,1,B,i,0,j,l)|0;C=k[l>>2]|0;r=m;return C|0}if((Zk(a,x,l)|0)>0){C=k[l>>2]|0;r=m;return C|0}B=(k[x>>2]|0)==0;if(B){x=(b|0)==0?1:b;F=2;G=1;H=x;I=x}else{F=b;G=k[g+(c<<2)>>2]|0;H=k[f+(c<<2)>>2]|0;I=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[p>>2]=1;k[q>>2]=1;k[s>>2]=1;x=u;k[x>>2]=1;k[x+4>>2]=0;k[t>>2]=1;x=p+4|0;k[x>>2]=1;w=q+4|0;k[w>>2]=1;v=s+4|0;k[v>>2]=1;o=u+8|0;D=o;k[D>>2]=1;k[D+4>>2]=0;D=t+4|0;k[D>>2]=1;E=p+8|0;k[E>>2]=1;J=q+8|0;k[J>>2]=1;K=s+8|0;k[K>>2]=1;M=u+16|0;N=M;k[N>>2]=1;k[N+4>>2]=0;N=t+8|0;k[N>>2]=1;O=p+12|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=u+24|0;S=R;k[S>>2]=1;k[S+4>>2]=0;S=t+12|0;k[S>>2]=1;T=p+16|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=u+32|0;X=W;k[X>>2]=1;k[X+4>>2]=0;X=t+16|0;k[X>>2]=1;Y=p+20|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=u+40|0;aa=$;k[aa>>2]=1;k[aa+4>>2]=0;aa=t+20|0;k[aa>>2]=1;ba=p+24|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=u+48|0;fa=ea;k[fa>>2]=1;k[fa+4>>2]=0;fa=t+24|0;k[fa>>2]=1;ga=p+28|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=u+56|0;la=ka;k[la>>2]=1;k[la+4>>2]=0;la=t+28|0;k[la>>2]=1;ma=p+32|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=u+64|0;qa=pa;k[qa>>2]=1;k[qa+4>>2]=0;qa=t+32|0;k[qa>>2]=1;a:do if((c|0)>0){ra=1;sa=0;ta=0;while(1){ua=k[f+(ta<<2)>>2]|0;va=k[e+(ta<<2)>>2]|0;if((ua|0)<(va|0)){if(!B){wa=ta;break}k[t+(ta<<2)>>2]=-1;xa=-1}else xa=k[t+(ta<<2)>>2]|0;k[p+(ta<<2)>>2]=va;k[q+(ta<<2)>>2]=ua;k[s+(ta<<2)>>2]=k[g+(ta<<2)>>2];ua=k[d+(ta<<2)>>2]|0;va=ra;ra=yy(ua|0,((ua|0)<0)<<31>>31|0,ra|0,sa|0)|0;ua=sa;sa=L;ya=ta;ta=ta+1|0;za=u+(ta<<3)|0;k[za>>2]=ra;k[za+4>>2]=sa;za=yy(xa|0,((xa|0)<0)<<31>>31|0,va|0,ua|0)|0;ua=u+(ya<<3)|0;k[ua>>2]=za;k[ua+4>>2]=L;if((ta|0)>=(c|0))break a}k[n>>2]=wa+1;Xw(A,759974,n)|0;tk(A);k[l>>2]=321;C=321;r=m;return C|0}while(0);A=u+(c<<3)|0;u=A;n=k[t+(c<<2)>>2]|0;wa=yy(n|0,((n|0)<0)<<31>>31|0,k[u>>2]|0,k[u+4>>2]|0)|0;u=A;k[u>>2]=wa;k[u+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Aa=(H-I|0)/(G|0)|0;Ba=G;Ca=I}else{d=k[t>>2]|0;t=ha((k[q>>2]|0)-(k[p>>2]|0)|0,d)|0;Aa=(t|0)/(k[g>>2]|0)|0;Ba=ha(k[s>>2]|0,d)|0;Ca=H}H=Aa+1|0;b:do if((I|0)<=(Ca|0)){Aa=k[qa>>2]|0;d=ha(Aa,k[ma>>2]|0)|0;s=ha(Aa,k[na>>2]|0)|0;g=(d|0)>(s|0);t=k[la>>2]|0;q=ha(t,k[ga>>2]|0)|0;c=ha(t,k[ia>>2]|0)|0;u=(q|0)>(c|0);wa=k[oa>>2]|0;A=k[fa>>2]|0;n=ha(A,k[ba>>2]|0)|0;xa=ha(A,k[ca>>2]|0)|0;B=(n|0)>(xa|0);e=k[ja>>2]|0;f=k[aa>>2]|0;ta=ha(f,k[Y>>2]|0)|0;sa=ha(f,k[Z>>2]|0)|0;ra=(ta|0)>(sa|0);ua=k[da>>2]|0;za=k[X>>2]|0;ya=ha(za,k[T>>2]|0)|0;va=ha(za,k[U>>2]|0)|0;Da=(ya|0)>(va|0);Ea=k[_>>2]|0;Fa=k[S>>2]|0;Ga=ha(Fa,k[O>>2]|0)|0;Ha=ha(Fa,k[P>>2]|0)|0;Ia=(Ga|0)>(Ha|0);Ja=k[V>>2]|0;Ka=k[N>>2]|0;La=ha(Ka,k[E>>2]|0)|0;Ma=ha(Ka,k[J>>2]|0)|0;Na=(La|0)>(Ma|0);Oa=k[Q>>2]|0;Pa=k[D>>2]|0;Qa=ha(Pa,k[x>>2]|0)|0;Ra=ha(Pa,k[w>>2]|0)|0;Sa=(Qa|0)>(Ra|0);Ta=k[K>>2]|0;Ua=k[p>>2]|0;Va=((Ua|0)<0)<<31>>31;Wa=o;Xa=k[Wa>>2]|0;Ya=k[Wa+4>>2]|0;Wa=M;Za=k[Wa>>2]|0;_a=k[Wa+4>>2]|0;Wa=R;$a=k[Wa>>2]|0;ab=k[Wa+4>>2]|0;Wa=W;bb=k[Wa>>2]|0;cb=k[Wa+4>>2]|0;Wa=$;db=k[Wa>>2]|0;eb=k[Wa+4>>2]|0;Wa=ea;fb=k[Wa>>2]|0;gb=k[Wa+4>>2]|0;Wa=ka;hb=k[Wa>>2]|0;ib=k[Wa+4>>2]|0;Wa=pa;jb=k[Wa>>2]|0;kb=k[Wa+4>>2]|0;Wa=((H|0)<0)<<31>>31;lb=k[v>>2]|0;mb=0;nb=I;c:while(1){if(g)ob=mb;else{pb=((nb|0)<0)<<31>>31;qb=mb;rb=d;while(1){if(u)sb=qb;else{tb=rb-Aa|0;ub=yy(tb|0,((tb|0)<0)<<31>>31|0,jb|0,kb|0)|0;tb=L;vb=qb;wb=q;while(1){if(B)xb=vb;else{yb=wb-t|0;zb=yy(yb|0,((yb|0)<0)<<31>>31|0,hb|0,ib|0)|0;yb=L;Ab=vb;Bb=n;while(1){if(ra)Cb=Ab;else{Db=Bb-A|0;Eb=yy(Db|0,((Db|0)<0)<<31>>31|0,fb|0,gb|0)|0;Db=L;Fb=Ab;Gb=ta;while(1){if(Da)Hb=Fb;else{Ib=Gb-f|0;Jb=yy(Ib|0,((Ib|0)<0)<<31>>31|0,db|0,eb|0)|0;Ib=L;Kb=Fb;Lb=ya;while(1){if(Ia)Mb=Kb;else{Nb=Lb-za|0;Ob=yy(Nb|0,((Nb|0)<0)<<31>>31|0,bb|0,cb|0)|0;Nb=L;Pb=Kb;Qb=Ga;while(1){if(Na)Rb=Pb;else{Sb=Qb-Fa|0;Tb=yy(Sb|0,((Sb|0)<0)<<31>>31|0,$a|0,ab|0)|0;Sb=L;Ub=Pb;Vb=La;while(1){if(Sa)Wb=Ub;else{Xb=Vb-Ka|0;Yb=yy(Xb|0,((Xb|0)<0)<<31>>31|0,Za|0,_a|0)|0;Xb=L;Zb=Ub;_b=Qa;while(1){$b=_b-Pa|0;ac=yy(Xa|0,Ya|0,$b|0,(($b|0)<0)<<31>>31|0)|0;$b=ky(ac|0,L|0,Ua|0,Va|0)|0;ac=ky($b|0,L|0,Yb|0,Xb|0)|0;$b=ky(ac|0,L|0,Tb|0,Sb|0)|0;ac=ky($b|0,L|0,Ob|0,Nb|0)|0;$b=ky(ac|0,L|0,Jb|0,Ib|0)|0;ac=ky($b|0,L|0,Eb|0,Db|0)|0;$b=ky(ac|0,L|0,zb|0,yb|0)|0;ac=ky($b|0,L|0,ub|0,tb|0)|0;if((Ym(a,F,nb,pb,ac,L,H,Wa,Ba,1,h,i+(Zb<<2)|0,z,y,l)|0)>0)break c;if(b&(k[y>>2]|0)!=0)k[j>>2]=1;ac=Zb+H|0;_b=lb+_b|0;if((_b|0)>(Ra|0)){Wb=ac;break}else Zb=ac}}Vb=Ta+Vb|0;if((Vb|0)>(Ma|0)){Rb=Wb;break}else Ub=Wb}}Qb=Oa+Qb|0;if((Qb|0)>(Ha|0)){Mb=Rb;break}else Pb=Rb}}Lb=Ja+Lb|0;if((Lb|0)>(va|0)){Hb=Mb;break}else Kb=Mb}}Gb=Ea+Gb|0;if((Gb|0)>(sa|0)){Cb=Hb;break}else Fb=Hb}}Bb=ua+Bb|0;if((Bb|0)>(xa|0)){xb=Cb;break}else Ab=Cb}}wb=e+wb|0;if((wb|0)>(c|0)){sb=xb;break}else vb=xb}}rb=wa+rb|0;if((rb|0)>(s|0)){ob=sb;break}else qb=sb}}nb=nb+G|0;if((nb|0)>(Ca|0))break b;else mb=ob}C=k[l>>2]|0;r=m;return C|0}while(0);C=k[l>>2]|0;r=m;return C|0}function $m(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0;n=r;r=r+16|0;Ym(a,b,c,d,e,f,g,h,1,1,i,j,n,l,m)|0;r=n;return k[m>>2]|0}function an(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;Ym(a,b,c,d,e,f,g,h,1,2,0,i,j,l,m)|0;return k[m>>2]|0}function bn(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0.0;n=c==1.0&d==0.0;o=(b|0)>0;if(!e){if(n){if(o)p=0;else{q=k[m>>2]|0;return q|0}do{k[l+(p<<2)>>2]=k[a+(p<<2)>>2];p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}if(o)r=0;else{q=k[m>>2]|0;return q|0}do{s=+(k[a+(r<<2)>>2]|0)*c+d;do if(!(s<-9223372036854774784.0))if(s>9223372036854774784.0){k[m>>2]=-11;k[l+(r<<2)>>2]=2147483647;break}else{k[l+(r<<2)>>2]=~~s;break}else{k[m>>2]=-11;k[l+(r<<2)>>2]=-2147483648}while(0);r=r+1|0}while((r|0)!=(b|0));q=k[m>>2]|0;return q|0}if(n){if(!o){q=k[m>>2]|0;return q|0}if((e|0)==1){n=0;do{r=k[a+(n<<2)>>2]|0;if((r|0)==(f|0)){k[j>>2]=1;k[l+(n<<2)>>2]=g}else k[l+(n<<2)>>2]=r;n=n+1|0}while((n|0)!=(b|0));q=k[m>>2]|0;return q|0}else{n=0;do{r=k[a+(n<<2)>>2]|0;if((r|0)==(f|0)){k[j>>2]=1;i[h+n>>0]=1}else k[l+(n<<2)>>2]=r;n=n+1|0}while((n|0)!=(b|0));q=k[m>>2]|0;return q|0}}if(!o){q=k[m>>2]|0;return q|0}if((e|0)==1){e=0;do{o=k[a+(e<<2)>>2]|0;do if((o|0)!=(f|0)){s=+(o|0)*c+d;if(s<-9223372036854774784.0){k[m>>2]=-11;k[l+(e<<2)>>2]=-2147483648;break}if(s>9223372036854774784.0){k[m>>2]=-11;k[l+(e<<2)>>2]=2147483647;break}else{k[l+(e<<2)>>2]=~~s;break}}else{k[j>>2]=1;k[l+(e<<2)>>2]=g}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}else{e=0;do{g=k[a+(e<<2)>>2]|0;do if((g|0)!=(f|0)){s=+(g|0)*c+d;if(s<-9223372036854774784.0){k[m>>2]=-11;k[l+(e<<2)>>2]=-2147483648;break}if(s>9223372036854774784.0){k[m>>2]=-11;k[l+(e<<2)>>2]=2147483647;break}else{k[l+(e<<2)>>2]=~~s;break}}else{k[j>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}return 0}function cn(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=a+(q<<3)|0;t=k[s>>2]|0;u=k[s+4>>2]|0;do if(!((u|0)<-1|(u|0)==-1&t>>>0<2147483648))if((u|0)>0|(u|0)==0&t>>>0>2147483647){k[n>>2]=-11;k[m+(q<<2)>>2]=2147483647;break}else{k[m+(q<<2)>>2]=t;break}else{k[n>>2]=-11;k[m+(q<<2)>>2]=-2147483648}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)v=0;else{r=k[n>>2]|0;return r|0}do{q=a+(v<<3)|0;w=(+((k[q>>2]|0)>>>0)+4294967296.0*+(k[q+4>>2]|0))*c+d;do if(!(w<-9223372036854774784.0))if(w>9223372036854774784.0){k[n>>2]=-11;k[m+(v<<2)>>2]=2147483647;break}else{k[m+(v<<2)>>2]=~~w;break}else{k[n>>2]=-11;k[m+(v<<2)>>2]=-2147483648}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;v=0;do{q=a+(v<<3)|0;t=k[q>>2]|0;u=k[q+4>>2]|0;do if((t|0)==(f|0)&(u|0)==(g|0)){k[l>>2]=1;if(o){k[m+(v<<2)>>2]=h;break}else{i[j+v>>0]=1;break}}else{w=(+(t>>>0)+4294967296.0*+(u|0))*c+d;if(w<-9223372036854774784.0){k[n>>2]=-11;k[m+(v<<2)>>2]=-2147483648;break}if(w>9223372036854774784.0){k[n>>2]=-11;k[m+(v<<2)>>2]=2147483647;break}else{k[m+(v<<2)>>2]=~~w;break}}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){e=0;do{p=a+(e<<3)|0;v=k[p>>2]|0;o=k[p+4>>2]|0;do if(!((v|0)==(f|0)&(o|0)==(g|0))){if((o|0)<-1|(o|0)==-1&v>>>0<2147483648){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if((o|0)>0|(o|0)==0&v>>>0>2147483647){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=v;break}}else{k[l>>2]=1;k[m+(e<<2)>>2]=h}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}else{e=0;do{h=a+(e<<3)|0;v=k[h>>2]|0;o=k[h+4>>2]|0;do if(!((v|0)==(f|0)&(o|0)==(g|0))){if((o|0)<-1|(o|0)==-1&v>>>0<2147483648){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if((o|0)>0|(o|0)==0&v>>>0>2147483647){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=v;break}}else{k[l>>2]=1;i[j+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function dn(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=l[a+q>>0];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(l[a+s>>0]|0|0)*c+d;do if(!(t<-9223372036854774784.0))if(t>9223372036854774784.0){k[n>>2]=-11;k[m+(s<<2)>>2]=2147483647;break}else{k[m+(s<<2)>>2]=~~t;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;s=0;do{q=i[a+s>>0]|0;do if(q<<24>>24==f<<24>>24){k[j>>2]=1;if(o){k[m+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else k[m+(s<<2)>>2]=q&255;while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{s=i[a+e>>0]|0;do if(s<<24>>24==f<<24>>24){k[j>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s&255|0)*c+d;if(t<-9223372036854774784.0){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if(t>9223372036854774784.0){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function en(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=j[a+(q<<1)>>1];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(j[a+(s<<1)>>1]|0)*c+d;do if(!(t<-9223372036854774784.0))if(t>9223372036854774784.0){k[n>>2]=-11;k[m+(s<<2)>>2]=2147483647;break}else{k[m+(s<<2)>>2]=~~t;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;s=0;do{q=j[a+(s<<1)>>1]|0;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){k[m+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else k[m+(s<<2)>>2]=q<<16>>16;while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{s=j[a+(e<<1)>>1]|0;do if(s<<16>>16==f<<16>>16){k[l>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s<<16>>16)*c+d;if(t<-9223372036854774784.0){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if(t>9223372036854774784.0){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function fn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-9223372036854774784.0))if(s>9223372036854774784.0){k[l>>2]=-11;k[j+(p<<2)>>2]=2147483647;break}else{k[j+(p<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(p<<2)>>2]=-2147483648}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-9223372036854774784.0))if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(t<<2)>>2]=2147483647;break}else{k[j+(t<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(t<<2)>>2]=-2147483648}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-9223372036854774784.0){k[l>>2]=-11;k[j+(u<<2)>>2]=-2147483648;break a}if(s>9223372036854774784.0){k[l>>2]=-11;k[j+(u<<2)>>2]=2147483647;break a}else{k[j+(u<<2)>>2]=~~r;break a}break}case 1:{k[h>>2]=1;if(p){k[j+(u<<2)>>2]=f;break a}else{i[g+u>>0]=1;break a}break}default:k[j+(u<<2)>>2]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-9223372036854774784.0;v=d>9223372036854774784.0;u=~~d;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-9223372036854774784.0){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=~~r;break b}break}case 1:{k[h>>2]=1;if(p){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(v){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function gn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-9223372036854774784.0))if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(o<<2)>>2]=2147483647;break}else{k[j+(o<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(o<<2)>>2]=-2147483648}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-9223372036854774784.0))if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(s<<2)>>2]=2147483647;break}else{k[j+(s<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-9223372036854774784.0){k[l>>2]=-11;k[j+(t<<2)>>2]=-2147483648;break a}if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(t<<2)>>2]=2147483647;break a}else{k[j+(t<<2)>>2]=~~r;break a}break}case 1:{k[h>>2]=1;if(o){k[j+(t<<2)>>2]=f;break a}else{i[g+t>>0]=1;break a}break}default:k[j+(t<<2)>>2]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-9223372036854774784.0;u=d>9223372036854774784.0;t=~~d;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-9223372036854774784.0){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(r>9223372036854774784.0){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=~~r;break b}break}case 1:{k[h>>2]=1;if(o){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(u){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function hn(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0.0,P=0.0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0.0,Z=0,_=0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0,ga=0.0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0;p=r;r=r+96|0;q=p;s=p+4|0;t=xx(h)|0;a:do if((b|0)>0){u=(g|0)==0;v=(g|0)==1;w=a;x=0;b:while(1){y=w+e|0;z=i[y>>0]|0;i[y>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,w,t)|0)==0:0)if(!u){k[m>>2]=1;if(v){k[n+(x<<2)>>2]=j;A=y;break}else{i[l+x>>0]=1;A=y;break}}else A=y;else{B=w;C=9}while(0);do if((C|0)==9){c:while(1){C=0;D=i[B>>0]|0;switch(D<<24>>24){case 43:case 45:{E=D;F=B;C=11;break c;break}case 32:break;default:{G=D;H=B;I=1;break c}}B=B+1|0;C=9}if((C|0)==11){C=0;D=E<<24>>24==45;J=F;while(1){K=J+1|0;L=i[K>>0]|0;if(L<<24>>24==32)J=K;else{M=L;N=K;break}}G=M;H=N;I=D?-1:1}if((G+-48&255)<10){J=G;K=H;O=0.0;while(1){P=O*10.0+ +(J<<24>>24);L=K;while(1){Q=L+1|0;R=i[Q>>0]|0;if(R<<24>>24==32)L=Q;else{S=R;T=Q;break}}U=P+-48.0;if((S+-48&255)<10){J=S;K=T;O=U}else{V=S;W=T;X=U;break}}}else{V=G;W=H;X=0.0}switch(V<<24>>24){case 44:case 46:{K=W;while(1){J=K+1|0;D=i[J>>0]|0;if(D<<24>>24==32)K=J;else{Z=D;_=J;break}}if((Z+-48&255)<10){K=Z;J=_;O=1.0;U=X;while(1){$=U*10.0+ +(K<<24>>24)+-48.0;D=J;while(1){L=D+1|0;Q=i[L>>0]|0;if(Q<<24>>24==32)D=L;else{aa=Q;ba=L;break}}P=O*10.0;if((aa+-48&255)<10){K=aa;J=ba;O=P;U=$}else{ca=aa;da=ba;ea=1;fa=P;ga=$;break}}}else{ca=Z;da=_;ea=1;fa=1.0;ga=X}break}default:{ca=V;da=W;ea=0;fa=1.0;ga=X}}if((ca&-2)<<24>>24==68){J=da;d:while(1){K=J+1|0;D=i[K>>0]|0;switch(D<<24>>24){case 32:{J=K;break}case 43:case 45:{ia=D;ja=J;C=28;break d;break}default:{ka=D;la=K;ma=1;break d}}}if((C|0)==28){C=0;J=ia<<24>>24==45;K=ja+2|0;while(1){D=i[K>>0]|0;if(D<<24>>24==32)K=K+1|0;else{na=D;oa=K;break}}ka=na;la=oa;ma=J?-1:1}if((ka+-48&255)<10){K=ka;D=la;L=0;while(1){Q=(L*10|0)+-48|0;R=D;while(1){pa=R+1|0;qa=i[pa>>0]|0;if(qa<<24>>24==32)R=pa;else{ra=qa;sa=pa;break}}R=Q+(K<<24>>24)|0;if((ra+-48&255)<10){K=ra;D=sa;L=R}else{ta=ra;ua=sa;va=ma;wa=R;break}}}else{ta=ka;ua=la;va=ma;wa=0}}else{ta=ca;ua=da;va=1;wa=0}if(ta<<24>>24){xa=y;ya=z;za=w;break b}$=+(I|0)*ga/((ea|0)==0?f:fa)*+Y(10.0,+(+(ha(va,wa)|0)))*c+d;if($<-9223372036854774784.0){k[o>>2]=-11;k[n+(x<<2)>>2]=-2147483648;A=ua;break}if($>9223372036854774784.0){k[o>>2]=-11;k[n+(x<<2)>>2]=2147483647;A=ua;break}else{k[n+(x<<2)>>2]=~~$;A=ua;break}}while(0);i[y>>0]=z;x=x+1|0;if((x|0)>=(b|0))break a;else w=A}w=s;x=761193;v=w+36|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(v|0));tk(s);k[q>>2]=za;Xw(s,761229,q)|0;tk(s);i[xa>>0]=ya;k[o>>2]=409;Aa=409;r=p;return Aa|0}while(0);Aa=k[o>>2]|0;r=p;return Aa|0}function jn(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0;m=r;r=r+16|0;n=m;if(!(Cl(a,l)|0)){o=(b|0)<1;kn(a,2,o?1:b,o?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,i,m+8|0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}else{b=n;k[b>>2]=g;k[b+4>>2]=h;qt(a,81,c,d,e,f,1,n,i,0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}return 0}function kn(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s,t){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;t=t|0;var u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0.0,T=0.0,U=0.0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0;u=r;r=r+29072|0;v=u+28888|0;w=u+28864|0;x=u+28856|0;y=u+28848|0;z=u+28840|0;A=u+28932|0;B=u+28928|0;C=u+28924|0;D=u+28916|0;E=u+28912|0;F=u+28908|0;G=u+32|0;H=u+24|0;I=u+16|0;J=u+8|0;K=u;M=u+28956|0;N=u+28976|0;O=u+28936|0;P=u+40|0;Q=k[t>>2]|0;if((g|0)==0&(h|0)==0|(Q|0)>0){R=Q;r=u;return R|0}if(s)k[s>>2]=0;if((l|0)==2)sy(q|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,j>>31,y,z,M,E,A,B,H,I,F,G,K,C,J,O,t)|0)>0){R=k[t>>2]|0;r=u;return R|0}f=k[B>>2]|0;B=((f|0)<0)<<31>>31;e=ha(k[F>>2]|0,j)|0;k[F>>2]=e;if((k[A>>2]|0)==16?(Nk(M,u+28920|0,u+28904|0,D,t)|0,e=k[D>>2]|0,(e|0)>0):0){D=0;S=1.0;while(1){T=S*10.0;D=D+1|0;if((D|0)>=(e|0)){U=T;break}else S=T}}else U=1.0;e=k[A>>2]|0;if(!((l|0)==1&((m|0)==0&(n|0)==0))?(D=J,d=k[D>>2]|0,c=k[D+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0){D=ky(d|0,c|0,32768,0)|0;Q=L;if(!((e|0)==21&(Q>>>0>0|(Q|0)==0&D>>>0>65535))?!((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)):0){V=(e|0)==16&(i[O>>0]|0)==1?0:l;W=16}else{X=B;Y=f;Z=0;_=0}}else{V=0;W=16}if((W|0)==16)if((e|0)==81){e=(h|0)<0|(h|0)==0&g>>>0<268435455;X=e?h:0;Y=e?g:268435455;Z=(V|0)==0&+p[y>>3]==1.0&+p[z>>3]==0.0;_=V}else{X=B;Y=f;Z=0;_=V}V=(j|0)>-1;f=((j|0)<0)<<31>>31;B=0-j|0;e=((B|0)<0)<<31>>31;l=0;d=0;c=g;g=h;h=0;D=0;a:while(1){Q=D;$=c;aa=g;ba=h;while(1){ca=o+(ba<<3)|0;da=q+ba|0;ea=$;fa=aa;while(1){if((ea|0)==0&(fa|0)==0){W=51;break a}ga=(fa|0)<(X|0)|(fa|0)==(X|0)&ea>>>0<Y>>>0;ia=ga?ea:Y;ga=my(0,ia|0,32)|0;ia=L;if(V){ja=G;ka=I;la=k[ka>>2]|0;ma=k[ka+4>>2]|0;ka=ky(k[ja>>2]|0,k[ja+4>>2]|0,-1,-1)|0;ja=jy(ka|0,L|0,la|0,ma|0)|0;ka=wy(ja|0,L|0,j|0,f|0)|0;ja=ky(ka|0,L|0,1,0)|0;ka=L;na=(ia|0)<(ka|0)|(ia|0)==(ka|0)&ga>>>0<ja>>>0;oa=la;pa=ma;qa=na?ga:ja;ra=na?ia:ka}else{ka=I;na=k[ka>>2]|0;ja=k[ka+4>>2]|0;ka=wy(na|0,ja|0,B|0,e|0)|0;ma=ky(ka|0,L|0,1,0)|0;ka=L;la=(ia|0)<(ka|0)|(ia|0)==(ka|0)&ga>>>0<ma>>>0;oa=na;pa=ja;qa=la?ga:ma;ra=la?ia:ka}ka=H;ia=k[ka>>2]|0;la=k[ka+4>>2]|0;ka=K;ma=yy(k[ka>>2]|0,k[ka+4>>2]|0,l|0,d|0)|0;ka=ky(ma|0,L|0,ia|0,la|0)|0;la=L;ia=k[F>>2]|0;ma=(ia|0)/(j|0)|0;ga=yy(ma|0,((ma|0)<0)<<31>>31|0,oa|0,pa|0)|0;ma=ky(ka|0,la|0,ga|0,L|0)|0;ga=L;switch(k[A>>2]|0){case 81:{ou(a,ma,ga,qa,ia,ca,t)|0;if(!Z){la=J;on(ca,qa,+p[y>>3],+p[z>>3],_,k[la>>2]|0,k[la+4>>2]|0,m,n,da,s,ca,t)|0}break}case 41:{nu(a,ma,ga,qa,ia,P,t)|0;pn(P,qa,+p[y>>3],+p[z>>3],_,k[J>>2]|0,m,n,da,s,ca,t)|0;break}case 11:{lu(a,ma,ga,qa,ia,P,t)|0;qn(P,qa,+p[y>>3],+p[z>>3],_,k[J>>2]&255,m,n,da,s,ca,t)|0;break}case 21:{mu(a,ma,ga,qa,ia,P,t)|0;rn(P,qa,+p[y>>3],+p[z>>3],_,k[J>>2]&65535,m,n,da,s,ca,t)|0;break}case 42:{pu(a,ma,ga,qa,ia,P,t)|0;sn(P,qa,+p[y>>3],+p[z>>3],_,m,n,da,s,ca,t)|0;break}case 82:{qu(a,ma,ga,qa,ia,P,t)|0;tn(P,qa,+p[y>>3],+p[z>>3],_,m,n,da,s,ca,t)|0;break}case 16:{au(a,ma,ga,0,t)|0;ga=k[F>>2]|0;ma=k[E>>2]|0;if((ga|0)==(ma|0)){ia=ha(ga,qa)|0;fu(a,ia,((ia|0)<0)<<31>>31,P,t)|0}else gu(a,ma,qa,ga-ma|0,P,t)|0;un(P,qa,+p[y>>3],+p[z>>3],k[E>>2]|0,U,_,O,m,n,da,s,ca,t)|0;break}default:{W=37;break a}}if((k[t>>2]|0)>0){sa=qa;ta=ba;ua=Q;W=41;break a}ma=((qa|0)<0)<<31>>31;ga=jy(ea|0,fa|0,qa|0,ma|0)|0;ia=L;if((ea|0)==(qa|0)&(fa|0)==(ma|0)){ea=ga;fa=ia}else{va=qa;wa=qa;xa=ma;ya=ga;za=ia;break}}fa=ky(wa|0,xa|0,ba|0,Q|0)|0;ea=L;ca=I;da=ha(va,j)|0;ia=ky(k[ca>>2]|0,k[ca+4>>2]|0,da|0,((da|0)<0)<<31>>31|0)|0;da=L;ca=I;k[ca>>2]=ia;k[ca+4>>2]=da;ca=G;ga=k[ca>>2]|0;ma=k[ca+4>>2]|0;if(!((da|0)<(ma|0)|(da|0)==(ma|0)&ia>>>0<ga>>>0)){Aa=ia;Ba=da;Ca=ga;Da=ma;Ea=fa;Fa=ea;Ga=ya;Ha=za;W=47;break}if((da|0)<0){Ia=ia;Ja=da;Ka=ga;La=ma;Ma=fa;Na=ea;Oa=ya;Pa=za;W=50;break}else{Q=ea;$=ya;aa=za;ba=fa}}if((W|0)==47){W=0;ba=wy(Aa|0,Ba|0,Ca|0,Da|0)|0;aa=L;$=ky(ba|0,aa|0,l|0,d|0)|0;Q=L;fa=yy(ba|0,aa|0,Ca|0,Da|0)|0;aa=jy(Aa|0,Ba|0,fa|0,L|0)|0;Qa=aa;Ra=L;Sa=Ea;Ta=Fa;Ua=Ga;Va=Ha;Wa=$;Xa=Q}else if((W|0)==50){W=0;Q=wy(~Ia|0,~Ja|0,Ka|0,La|0)|0;$=ky(Q|0,L|0,1,0)|0;Q=L;aa=jy(l|0,d|0,$|0,Q|0)|0;fa=L;ba=yy($|0,Q|0,Ka|0,La|0)|0;Q=ky(ba|0,L|0,Ia|0,Ja|0)|0;Qa=Q;Ra=L;Sa=Ma;Ta=Na;Ua=Oa;Va=Pa;Wa=aa;Xa=fa}fa=I;k[fa>>2]=Qa;k[fa+4>>2]=Ra;l=Wa;d=Xa;c=Ua;g=Va;h=Sa;D=Ta}if((W|0)==37){k[x>>2]=b;k[x+4>>2]=M;Xw(N,760855,x)|0;tk(N);if((k[C>>2]|0)==1){k[t>>2]=311;R=311;r=u;return R|0}else{k[t>>2]=312;R=312;r=u;return R|0}}else if((W|0)==41){U=+(ta>>>0)+4294967296.0*+(ua|0);S=U+1.0;T=U+ +(sa|0);if((k[C>>2]|0)>0){p[w>>3]=S;p[w+8>>3]=T;k[w+16>>2]=b;Xw(N,759852,w)|0}else{p[v>>3]=S;p[v+8>>3]=T;Xw(N,759915,v)|0}tk(N);R=k[t>>2]|0;r=u;return R|0}else if((W|0)==51){W=k[t>>2]|0;if((W|0)!=-11){R=W;r=u;return R|0}tk(761034);k[t>>2]=412;R=412;r=u;return R|0}return 0}function ln(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;kn(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,81,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function mn(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;n=r;r=r+480|0;o=n+240|0;p=n+232|0;q=n+360|0;s=n+324|0;t=n+288|0;u=n+252|0;v=n+152|0;w=n+80|0;x=n+8|0;y=n+248|0;z=n+244|0;A=n+396|0;B=n+397|0;C=n;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(B,760019,p)|0;tk(B);k[m>>2]=320;D=320;r=n;return D|0}if(Cl(a,m)|0){if((c|0)>0){p=0;do{E=k[e+(p<<2)>>2]|0;F=w+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(p<<2)>>2]|0;F=x+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}p=C;k[p>>2]=h;k[p+4>>2]=i;mt(a,81,w,x,g,1,C,j,0,l,m)|0;D=k[m>>2]|0;r=n;return D|0}if((Zk(a,y,m)|0)>0){D=k[m>>2]|0;r=n;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(l|0)!=0;if(b)k[l>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;p=v+8|0;E=p;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[o>>2]=xa+1;Xw(B,759974,o)|0;tk(B);k[m>>2]=321;D=321;r=n;return D|0}while(0);B=v+(c<<3)|0;v=B;o=k[u+(c<<2)>>2]|0;xa=yy(o|0,((o|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;o=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(o|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=p;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=o;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((kn(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,i,j+(_b<<3)|0,A,z,m)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[l>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[m>>2]|0;r=n;return D|0}while(0);D=k[m>>2]|0;r=n;return D|0}function nn(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;kn(a,b,c,d,e,f,g,h,1,2,0,0,i,j,l,m)|0;return k[m>>2]|0}function on(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e){if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{t=a+(r<<3)|0;u=k[t+4>>2]|0;v=n+(r<<3)|0;k[v>>2]=k[t>>2];k[v+4>>2]=u;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q)w=0;else{s=k[o>>2]|0;return s|0}do{r=a+(w<<3)|0;x=(+((k[r>>2]|0)>>>0)+4294967296.0*+(k[r+4>>2]|0))*c+d;do if(!(x<-9223372036854775808.0))if(x>9223372036854775808.0){k[o>>2]=-11;r=n+(w<<3)|0;k[r>>2]=-1;k[r+4>>2]=2147483647;break}else{r=+W(x)>=1.0?(x>0.0?~~+ia(+V(x/4294967296.0),4294967295.0)>>>0:~~+ga((x-+(~~x>>>0))/4294967296.0)>>>0):0;u=n+(w<<3)|0;k[u>>2]=~~x>>>0;k[u+4>>2]=r;break}else{k[o>>2]=-11;r=n+(w<<3)|0;k[r>>2]=0;k[r+4>>2]=-2147483648}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;w=0;do{r=a+(w<<3)|0;u=k[r>>2]|0;v=k[r+4>>2]|0;do if((u|0)==(f|0)&(v|0)==(g|0)){k[m>>2]=1;if(p){r=n+(w<<3)|0;k[r>>2]=h;k[r+4>>2]=j;break}else{i[l+w>>0]=1;break}}else{r=n+(w<<3)|0;k[r>>2]=u;k[r+4>>2]=v}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{w=a+(e<<3)|0;p=k[w>>2]|0;v=k[w+4>>2]|0;do if((p|0)==(f|0)&(v|0)==(g|0)){k[m>>2]=1;if(q){w=n+(e<<3)|0;k[w>>2]=h;k[w+4>>2]=j;break}else{i[l+e>>0]=1;break}}else{x=(+(p>>>0)+4294967296.0*+(v|0))*c+d;if(x<-9223372036854775808.0){k[o>>2]=-11;w=n+(e<<3)|0;k[w>>2]=0;k[w+4>>2]=-2147483648;break}if(x>9223372036854775808.0){k[o>>2]=-11;w=n+(e<<3)|0;k[w>>2]=-1;k[w+4>>2]=2147483647;break}else{w=+W(x)>=1.0?(x>0.0?~~+ia(+V(x/4294967296.0),4294967295.0)>>>0:~~+ga((x-+(~~x>>>0))/4294967296.0)>>>0):0;u=n+(e<<3)|0;k[u>>2]=~~x>>>0;k[u+4>>2]=w;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}function pn(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=k[a+(q<<2)>>2]|0;t=m+(q<<3)|0;k[t>>2]=s;k[t+4>>2]=((s|0)<0)<<31>>31;q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)u=0;else{r=k[n>>2]|0;return r|0}do{v=+(k[a+(u<<2)>>2]|0)*c+d;do if(!(v<-9223372036854775808.0))if(v>9223372036854775808.0){k[n>>2]=-11;q=m+(u<<3)|0;k[q>>2]=-1;k[q+4>>2]=2147483647;break}else{q=+W(v)>=1.0?(v>0.0?~~+ia(+V(v/4294967296.0),4294967295.0)>>>0:~~+ga((v-+(~~v>>>0))/4294967296.0)>>>0):0;s=m+(u<<3)|0;k[s>>2]=~~v>>>0;k[s+4>>2]=q;break}else{k[n>>2]=-11;q=m+(u<<3)|0;k[q>>2]=0;k[q+4>>2]=-2147483648}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;u=0;do{q=k[a+(u<<2)>>2]|0;do if((q|0)==(f|0)){k[l>>2]=1;if(o){s=m+(u<<3)|0;k[s>>2]=g;k[s+4>>2]=h;break}else{i[j+u>>0]=1;break}}else{s=m+(u<<3)|0;k[s>>2]=q;k[s+4>>2]=((q|0)<0)<<31>>31}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{u=k[a+(e<<2)>>2]|0;do if((u|0)==(f|0)){k[l>>2]=1;if(p){o=m+(e<<3)|0;k[o>>2]=g;k[o+4>>2]=h;break}else{i[j+e>>0]=1;break}}else{v=+(u|0)*c+d;if(v<-9223372036854775808.0){k[n>>2]=-11;o=m+(e<<3)|0;k[o>>2]=0;k[o+4>>2]=-2147483648;break}if(v>9223372036854775808.0){k[n>>2]=-11;o=m+(e<<3)|0;k[o>>2]=-1;k[o+4>>2]=2147483647;break}else{o=+W(v)>=1.0?(v>0.0?~~+ia(+V(v/4294967296.0),4294967295.0)>>>0:~~+ga((v-+(~~v>>>0))/4294967296.0)>>>0):0;q=m+(e<<3)|0;k[q>>2]=~~v>>>0;k[q+4>>2]=o;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function qn(a,b,c,d,e,f,g,h,j,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e){if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{t=n+(r<<3)|0;k[t>>2]=l[a+r>>0];k[t+4>>2]=0;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q)u=0;else{s=k[o>>2]|0;return s|0}do{v=+(l[a+u>>0]|0|0)*c+d;do if(!(v<-9223372036854775808.0))if(v>9223372036854775808.0){k[o>>2]=-11;r=n+(u<<3)|0;k[r>>2]=-1;k[r+4>>2]=2147483647;break}else{r=+W(v)>=1.0?(v>0.0?~~+ia(+V(v/4294967296.0),4294967295.0)>>>0:~~+ga((v-+(~~v>>>0))/4294967296.0)>>>0):0;t=n+(u<<3)|0;k[t>>2]=~~v>>>0;k[t+4>>2]=r;break}else{k[o>>2]=-11;r=n+(u<<3)|0;k[r>>2]=0;k[r+4>>2]=-2147483648}while(0);u=u+1|0}while((u|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;u=0;do{r=i[a+u>>0]|0;do if(r<<24>>24==f<<24>>24){k[m>>2]=1;if(p){t=n+(u<<3)|0;k[t>>2]=g;k[t+4>>2]=h;break}else{i[j+u>>0]=1;break}}else{t=n+(u<<3)|0;k[t>>2]=r&255;k[t+4>>2]=0}while(0);u=u+1|0}while((u|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{u=i[a+e>>0]|0;do if(u<<24>>24==f<<24>>24){k[m>>2]=1;if(q){p=n+(e<<3)|0;k[p>>2]=g;k[p+4>>2]=h;break}else{i[j+e>>0]=1;break}}else{v=+(u&255|0)*c+d;if(v<-9223372036854775808.0){k[o>>2]=-11;p=n+(e<<3)|0;k[p>>2]=0;k[p+4>>2]=-2147483648;break}if(v>9223372036854775808.0){k[o>>2]=-11;p=n+(e<<3)|0;k[p>>2]=-1;k[p+4>>2]=2147483647;break}else{p=+W(v)>=1.0?(v>0.0?~~+ia(+V(v/4294967296.0),4294967295.0)>>>0:~~+ga((v-+(~~v>>>0))/4294967296.0)>>>0):0;r=n+(e<<3)|0;k[r>>2]=~~v>>>0;k[r+4>>2]=p;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}
function rn(a,b,c,d,e,f,g,h,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e){if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{t=j[a+(r<<1)>>1]|0;u=n+(r<<3)|0;k[u>>2]=t;k[u+4>>2]=((t|0)<0)<<31>>31;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q)v=0;else{s=k[o>>2]|0;return s|0}do{w=+(j[a+(v<<1)>>1]|0)*c+d;do if(!(w<-9223372036854775808.0))if(w>9223372036854775808.0){k[o>>2]=-11;r=n+(v<<3)|0;k[r>>2]=-1;k[r+4>>2]=2147483647;break}else{r=+W(w)>=1.0?(w>0.0?~~+ia(+V(w/4294967296.0),4294967295.0)>>>0:~~+ga((w-+(~~w>>>0))/4294967296.0)>>>0):0;t=n+(v<<3)|0;k[t>>2]=~~w>>>0;k[t+4>>2]=r;break}else{k[o>>2]=-11;r=n+(v<<3)|0;k[r>>2]=0;k[r+4>>2]=-2147483648}while(0);v=v+1|0}while((v|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;v=0;do{r=j[a+(v<<1)>>1]|0;do if(r<<16>>16==f<<16>>16){k[m>>2]=1;if(p){t=n+(v<<3)|0;k[t>>2]=g;k[t+4>>2]=h;break}else{i[l+v>>0]=1;break}}else{t=r<<16>>16;u=n+(v<<3)|0;k[u>>2]=t;k[u+4>>2]=((t|0)<0)<<31>>31}while(0);v=v+1|0}while((v|0)!=(b|0));s=k[o>>2]|0;return s|0}if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{v=j[a+(e<<1)>>1]|0;do if(v<<16>>16==f<<16>>16){k[m>>2]=1;if(q){p=n+(e<<3)|0;k[p>>2]=g;k[p+4>>2]=h;break}else{i[l+e>>0]=1;break}}else{w=+(v<<16>>16)*c+d;if(w<-9223372036854775808.0){k[o>>2]=-11;p=n+(e<<3)|0;k[p>>2]=0;k[p+4>>2]=-2147483648;break}if(w>9223372036854775808.0){k[o>>2]=-11;p=n+(e<<3)|0;k[p>>2]=-1;k[p+4>>2]=2147483647;break}else{p=+W(w)>=1.0?(w>0.0?~~+ia(+V(w/4294967296.0),4294967295.0)>>>0:~~+ga((w-+(~~w>>>0))/4294967296.0)>>>0):0;r=n+(e<<3)|0;k[r>>2]=~~w>>>0;k[r+4>>2]=p;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}function sn(a,b,c,d,e,f,g,h,j,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;n=n|0;var p=0,q=0,r=0,s=0.0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;if(!e){p=(b|0)>0;if(c==1.0&d==0.0){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(q<<2)>>2];do if(!(s<-9223372036854775808.0))if(s>9223372036854775808.0){k[n>>2]=-11;t=l+(q<<3)|0;k[t>>2]=-1;k[t+4>>2]=2147483647;break}else{t=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;u=l+(q<<3)|0;k[u>>2]=~~s>>>0;k[u+4>>2]=t;break}else{k[n>>2]=-11;t=l+(q<<3)|0;k[t>>2]=0;k[t+4>>2]=-2147483648}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)v=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(v<<2)>>2]*c+d;do if(!(s<-9223372036854775808.0))if(s>9223372036854775808.0){k[n>>2]=-11;p=l+(v<<3)|0;k[p>>2]=-1;k[p+4>>2]=2147483647;break}else{p=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;q=l+(v<<3)|0;k[q>>2]=~~s>>>0;k[q+4>>2]=p;break}else{k[n>>2]=-11;p=l+(v<<3)|0;k[p>>2]=0;k[p+4>>2]=-2147483648}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}}v=a+2|0;p=(b|0)>0;if(c==1.0&d==0.0){if(!p){r=k[n>>2]|0;return r|0}q=(e|0)==1;t=0;u=v;while(1){w=(m[u>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{s=+o[a+(t<<2)>>2];if(s<-9223372036854775808.0){k[n>>2]=-11;x=l+(t<<3)|0;k[x>>2]=0;k[x+4>>2]=-2147483648;break a}if(s>9223372036854775808.0){k[n>>2]=-11;x=l+(t<<3)|0;k[x>>2]=-1;k[x+4>>2]=2147483647;break a}else{x=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;y=l+(t<<3)|0;k[y>>2]=~~s>>>0;k[y+4>>2]=x;break a}break}case 1:{k[j>>2]=1;if(q){x=l+(t<<3)|0;k[x>>2]=f;k[x+4>>2]=g;break a}else{i[h+t>>0]=1;break a}break}default:{x=l+(t<<3)|0;k[x>>2]=0;k[x+4>>2]=0}}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+4|0}r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=d<-9223372036854775808.0;u=d>9223372036854775808.0;t=~~d>>>0;q=+W(d)>=1.0?(d>0.0?~~+ia(+V(d/4294967296.0),4294967295.0)>>>0:~~+ga((d-+(~~d>>>0))/4294967296.0)>>>0):0;w=(e|0)==1;e=0;x=v;while(1){v=(m[x>>1]|0)&32640;b:do switch(((v|0)==32640?1:(v|0)==0?2:0)|0){case 0:{s=+o[a+(e<<2)>>2]*c+d;if(s<-9223372036854775808.0){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=0;k[y+4>>2]=-2147483648;break b}if(s>9223372036854775808.0){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=-1;k[y+4>>2]=2147483647;break b}else{y=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;z=l+(e<<3)|0;k[z>>2]=~~s>>>0;k[z+4>>2]=y;break b}break}case 1:{k[j>>2]=1;if(w){y=l+(e<<3)|0;k[y>>2]=f;k[y+4>>2]=g;break b}else{i[h+e>>0]=1;break b}break}default:{if(p){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=0;k[y+4>>2]=-2147483648;break b}if(u){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=-1;k[y+4>>2]=2147483647;break b}else{y=l+(e<<3)|0;k[y>>2]=t;k[y+4>>2]=q;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else x=x+4|0}r=k[n>>2]|0;return r|0}function tn(a,b,c,d,e,f,g,h,j,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;n=n|0;var o=0,q=0,r=0,s=0.0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;if(!e){o=(b|0)>0;if(c==1.0&d==0.0){if(o)q=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(q<<3)>>3];do if(!(s<-9223372036854775808.0))if(s>9223372036854775808.0){k[n>>2]=-11;t=l+(q<<3)|0;k[t>>2]=-1;k[t+4>>2]=2147483647;break}else{t=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;u=l+(q<<3)|0;k[u>>2]=~~s>>>0;k[u+4>>2]=t;break}else{k[n>>2]=-11;t=l+(q<<3)|0;k[t>>2]=0;k[t+4>>2]=-2147483648}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(o)v=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(v<<3)>>3]*c+d;do if(!(s<-9223372036854775808.0))if(s>9223372036854775808.0){k[n>>2]=-11;o=l+(v<<3)|0;k[o>>2]=-1;k[o+4>>2]=2147483647;break}else{o=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;q=l+(v<<3)|0;k[q>>2]=~~s>>>0;k[q+4>>2]=o;break}else{k[n>>2]=-11;o=l+(v<<3)|0;k[o>>2]=0;k[o+4>>2]=-2147483648}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}}v=a+6|0;o=(b|0)>0;if(c==1.0&d==0.0){if(!o){r=k[n>>2]|0;return r|0}q=(e|0)==1;t=0;u=v;while(1){w=(m[u>>1]|0)&32752;a:do switch(((w|0)==32752?1:(w|0)==0?2:0)|0){case 0:{s=+p[a+(t<<3)>>3];if(s<-9223372036854775808.0){k[n>>2]=-11;x=l+(t<<3)|0;k[x>>2]=0;k[x+4>>2]=-2147483648;break a}if(s>9223372036854775808.0){k[n>>2]=-11;x=l+(t<<3)|0;k[x>>2]=-1;k[x+4>>2]=2147483647;break a}else{x=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;y=l+(t<<3)|0;k[y>>2]=~~s>>>0;k[y+4>>2]=x;break a}break}case 1:{k[j>>2]=1;if(q){x=l+(t<<3)|0;k[x>>2]=f;k[x+4>>2]=g;break a}else{i[h+t>>0]=1;break a}break}default:{x=l+(t<<3)|0;k[x>>2]=0;k[x+4>>2]=0}}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}r=k[n>>2]|0;return r|0}if(!o){r=k[n>>2]|0;return r|0}o=d<-9223372036854775808.0;u=d>9223372036854775808.0;t=~~d>>>0;q=+W(d)>=1.0?(d>0.0?~~+ia(+V(d/4294967296.0),4294967295.0)>>>0:~~+ga((d-+(~~d>>>0))/4294967296.0)>>>0):0;w=(e|0)==1;e=0;x=v;while(1){v=(m[x>>1]|0)&32752;b:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{s=+p[a+(e<<3)>>3]*c+d;if(s<-9223372036854775808.0){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=0;k[y+4>>2]=-2147483648;break b}if(s>9223372036854775808.0){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=-1;k[y+4>>2]=2147483647;break b}else{y=+W(s)>=1.0?(s>0.0?~~+ia(+V(s/4294967296.0),4294967295.0)>>>0:~~+ga((s-+(~~s>>>0))/4294967296.0)>>>0):0;z=l+(e<<3)|0;k[z>>2]=~~s>>>0;k[z+4>>2]=y;break b}break}case 1:{k[j>>2]=1;if(w){y=l+(e<<3)|0;k[y>>2]=f;k[y+4>>2]=g;break b}else{i[h+e>>0]=1;break b}break}default:{if(o){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=0;k[y+4>>2]=-2147483648;break b}if(u){k[n>>2]=-11;y=l+(e<<3)|0;k[y>>2]=-1;k[y+4>>2]=2147483647;break b}else{y=l+(e<<3)|0;k[y>>2]=t;k[y+4>>2]=q;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else x=x+8|0}r=k[n>>2]|0;return r|0}function un(a,b,c,d,e,f,g,h,j,l,m,n,o,p){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0,S=0,T=0,U=0,X=0.0,Z=0,_=0,$=0.0,aa=0,ba=0,ca=0.0,da=0,ea=0,fa=0,ja=0,ka=0,la=0.0,ma=0.0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0;q=r;r=r+96|0;s=q;t=q+4|0;u=xx(h)|0;a:do if((b|0)>0){v=(g|0)==0;w=(g|0)==1;x=a;y=0;b:while(1){z=x+e|0;A=i[z>>0]|0;i[z>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,x,u)|0)==0:0)if(!v){k[n>>2]=1;if(w){B=o+(y<<3)|0;k[B>>2]=j;k[B+4>>2]=l;C=z;break}else{i[m+y>>0]=1;C=z;break}}else C=z;else{D=x;E=9}while(0);do if((E|0)==9){c:while(1){E=0;B=i[D>>0]|0;switch(B<<24>>24){case 43:case 45:{F=B;G=D;E=11;break c;break}case 32:break;default:{H=B;I=D;J=1;break c}}D=D+1|0;E=9}if((E|0)==11){E=0;B=F<<24>>24==45;K=G;while(1){L=K+1|0;M=i[L>>0]|0;if(M<<24>>24==32)K=L;else{N=M;O=L;break}}H=N;I=O;J=B?-1:1}if((H+-48&255)<10){K=H;L=I;P=0.0;while(1){Q=P*10.0+ +(K<<24>>24);M=L;while(1){R=M+1|0;S=i[R>>0]|0;if(S<<24>>24==32)M=R;else{T=S;U=R;break}}X=Q+-48.0;if((T+-48&255)<10){K=T;L=U;P=X}else{Z=T;_=U;$=X;break}}}else{Z=H;_=I;$=0.0}switch(Z<<24>>24){case 44:case 46:{L=_;while(1){K=L+1|0;B=i[K>>0]|0;if(B<<24>>24==32)L=K;else{aa=B;ba=K;break}}if((aa+-48&255)<10){L=aa;K=ba;P=1.0;X=$;while(1){ca=X*10.0+ +(L<<24>>24)+-48.0;B=K;while(1){M=B+1|0;R=i[M>>0]|0;if(R<<24>>24==32)B=M;else{da=R;ea=M;break}}Q=P*10.0;if((da+-48&255)<10){L=da;K=ea;P=Q;X=ca}else{fa=da;ja=ea;ka=1;la=Q;ma=ca;break}}}else{fa=aa;ja=ba;ka=1;la=1.0;ma=$}break}default:{fa=Z;ja=_;ka=0;la=1.0;ma=$}}if((fa&-2)<<24>>24==68){K=ja;d:while(1){L=K+1|0;B=i[L>>0]|0;switch(B<<24>>24){case 32:{K=L;break}case 43:case 45:{na=B;oa=K;E=28;break d;break}default:{pa=B;qa=L;ra=1;break d}}}if((E|0)==28){E=0;K=na<<24>>24==45;L=oa+2|0;while(1){B=i[L>>0]|0;if(B<<24>>24==32)L=L+1|0;else{sa=B;ta=L;break}}pa=sa;qa=ta;ra=K?-1:1}if((pa+-48&255)<10){L=pa;B=qa;M=0;while(1){R=(M*10|0)+-48|0;S=B;while(1){ua=S+1|0;va=i[ua>>0]|0;if(va<<24>>24==32)S=ua;else{wa=va;xa=ua;break}}S=R+(L<<24>>24)|0;if((wa+-48&255)<10){L=wa;B=xa;M=S}else{ya=wa;za=xa;Aa=ra;Ba=S;break}}}else{ya=pa;za=qa;Aa=ra;Ba=0}}else{ya=fa;za=ja;Aa=1;Ba=0}if(ya<<24>>24){Ca=z;Da=A;Ea=x;break b}ca=+(J|0)*ma/((ka|0)==0?f:la)*+Y(10.0,+(+(ha(Aa,Ba)|0)))*c+d;if(ca<-9223372036854775808.0){k[p>>2]=-11;M=o+(y<<3)|0;k[M>>2]=0;k[M+4>>2]=-2147483648;C=za;break}if(ca>9223372036854775808.0){k[p>>2]=-11;M=o+(y<<3)|0;k[M>>2]=-1;k[M+4>>2]=2147483647;C=za;break}else{M=+W(ca)>=1.0?(ca>0.0?~~+ia(+V(ca/4294967296.0),4294967295.0)>>>0:~~+ga((ca-+(~~ca>>>0))/4294967296.0)>>>0):0;B=o+(y<<3)|0;k[B>>2]=~~ca>>>0;k[B+4>>2]=M;C=za;break}}while(0);i[z>>0]=A;y=y+1|0;if((y|0)>=(b|0))break a;else x=C}x=t;y=761193;w=x+36|0;do{i[x>>0]=i[y>>0]|0;x=x+1|0;y=y+1|0}while((x|0)<(w|0));tk(t);k[s>>2]=Ea;Xw(t,761229,s)|0;tk(t);i[Ca>>0]=Da;k[p>>2]=409;Fa=409;r=q;return Fa|0}while(0);Fa=k[p>>2]|0;r=q;return Fa|0}function vn(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0;l=r;r=r+16|0;m=l;if(Cl(a,j)|0){k[m>>2]=g;qt(a,31,c,d,e,f,1,m,h,0,i,j)|0;n=k[j>>2]|0;r=l;return n|0}if((e|0)==0&(f|0)==0|(k[j>>2]|0)>0){n=k[j>>2]|0;r=l;return n|0}m=(b|0)<1;Ym(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+4|0,i,j)|0;n=k[j>>2]|0;r=l;return n|0}function wn(a,b,c,d,e,f,g,h,i,j,l,m,n,o,p){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,r=0;q=k[p>>2]|0;if((g|0)==0&(h|0)==0|(q|0)>0){r=q;return r|0}Ym(a,b,c,d,e,f,g,h,i,j,l,m,n,o,p)|0;r=k[p>>2]|0;return r|0}function xn(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(Cl(a,j)|0){qt(a,31,c,d,e,f,2,0,g,h,i,j)|0;l=k[j>>2]|0;return l|0}if((e|0)==0&(f|0)==0|(k[j>>2]|0)>0){l=k[j>>2]|0;return l|0}m=(b|0)<1;Ym(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;l=k[j>>2]|0;return l|0}function yn(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0;m=r;r=r+480|0;n=m+232|0;o=m+224|0;p=m+356|0;q=m+320|0;s=m+284|0;t=m+248|0;u=m+144|0;v=m+72|0;w=m;x=m+244|0;y=m+240|0;z=m+392|0;A=m+393|0;B=m+236|0;if((c+-1|0)>>>0>8){k[o>>2]=c;Xw(A,760019,o)|0;tk(A);k[l>>2]=320;C=320;r=m;return C|0}if(Cl(a,l)|0){if((c|0)>0){o=0;do{D=k[e+(o<<2)>>2]|0;E=v+(o<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;D=k[f+(o<<2)>>2]|0;E=w+(o<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;o=o+1|0}while((o|0)!=(c|0))}k[B>>2]=h;mt(a,31,v,w,g,1,B,i,0,j,l)|0;C=k[l>>2]|0;r=m;return C|0}if((Zk(a,x,l)|0)>0){C=k[l>>2]|0;r=m;return C|0}B=(k[x>>2]|0)==0;if(B){x=(b|0)==0?1:b;F=2;G=1;H=x;I=x}else{F=b;G=k[g+(c<<2)>>2]|0;H=k[f+(c<<2)>>2]|0;I=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[p>>2]=1;k[q>>2]=1;k[s>>2]=1;x=u;k[x>>2]=1;k[x+4>>2]=0;k[t>>2]=1;x=p+4|0;k[x>>2]=1;w=q+4|0;k[w>>2]=1;v=s+4|0;k[v>>2]=1;o=u+8|0;D=o;k[D>>2]=1;k[D+4>>2]=0;D=t+4|0;k[D>>2]=1;E=p+8|0;k[E>>2]=1;J=q+8|0;k[J>>2]=1;K=s+8|0;k[K>>2]=1;M=u+16|0;N=M;k[N>>2]=1;k[N+4>>2]=0;N=t+8|0;k[N>>2]=1;O=p+12|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=u+24|0;S=R;k[S>>2]=1;k[S+4>>2]=0;S=t+12|0;k[S>>2]=1;T=p+16|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=u+32|0;X=W;k[X>>2]=1;k[X+4>>2]=0;X=t+16|0;k[X>>2]=1;Y=p+20|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=u+40|0;aa=$;k[aa>>2]=1;k[aa+4>>2]=0;aa=t+20|0;k[aa>>2]=1;ba=p+24|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=u+48|0;fa=ea;k[fa>>2]=1;k[fa+4>>2]=0;fa=t+24|0;k[fa>>2]=1;ga=p+28|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=u+56|0;la=ka;k[la>>2]=1;k[la+4>>2]=0;la=t+28|0;k[la>>2]=1;ma=p+32|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=u+64|0;qa=pa;k[qa>>2]=1;k[qa+4>>2]=0;qa=t+32|0;k[qa>>2]=1;a:do if((c|0)>0){ra=1;sa=0;ta=0;while(1){ua=k[f+(ta<<2)>>2]|0;va=k[e+(ta<<2)>>2]|0;if((ua|0)<(va|0)){if(!B){wa=ta;break}k[t+(ta<<2)>>2]=-1;xa=-1}else xa=k[t+(ta<<2)>>2]|0;k[p+(ta<<2)>>2]=va;k[q+(ta<<2)>>2]=ua;k[s+(ta<<2)>>2]=k[g+(ta<<2)>>2];ua=k[d+(ta<<2)>>2]|0;va=ra;ra=yy(ua|0,((ua|0)<0)<<31>>31|0,ra|0,sa|0)|0;ua=sa;sa=L;ya=ta;ta=ta+1|0;za=u+(ta<<3)|0;k[za>>2]=ra;k[za+4>>2]=sa;za=yy(xa|0,((xa|0)<0)<<31>>31|0,va|0,ua|0)|0;ua=u+(ya<<3)|0;k[ua>>2]=za;k[ua+4>>2]=L;if((ta|0)>=(c|0))break a}k[n>>2]=wa+1;Xw(A,760064,n)|0;tk(A);k[l>>2]=321;C=321;r=m;return C|0}while(0);A=u+(c<<3)|0;u=A;n=k[t+(c<<2)>>2]|0;wa=yy(n|0,((n|0)<0)<<31>>31|0,k[u>>2]|0,k[u+4>>2]|0)|0;u=A;k[u>>2]=wa;k[u+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Aa=(H-I|0)/(G|0)|0;Ba=G;Ca=I}else{d=k[t>>2]|0;t=ha((k[q>>2]|0)-(k[p>>2]|0)|0,d)|0;Aa=(t|0)/(k[g>>2]|0)|0;Ba=ha(k[s>>2]|0,d)|0;Ca=H}H=Aa+1|0;b:do if((I|0)<=(Ca|0)){Aa=k[qa>>2]|0;d=ha(Aa,k[ma>>2]|0)|0;s=ha(Aa,k[na>>2]|0)|0;g=(d|0)>(s|0);t=k[la>>2]|0;q=ha(t,k[ga>>2]|0)|0;c=ha(t,k[ia>>2]|0)|0;u=(q|0)>(c|0);wa=k[oa>>2]|0;A=k[fa>>2]|0;n=ha(A,k[ba>>2]|0)|0;xa=ha(A,k[ca>>2]|0)|0;B=(n|0)>(xa|0);e=k[ja>>2]|0;f=k[aa>>2]|0;ta=ha(f,k[Y>>2]|0)|0;sa=ha(f,k[Z>>2]|0)|0;ra=(ta|0)>(sa|0);ua=k[da>>2]|0;za=k[X>>2]|0;ya=ha(za,k[T>>2]|0)|0;va=ha(za,k[U>>2]|0)|0;Da=(ya|0)>(va|0);Ea=k[_>>2]|0;Fa=k[S>>2]|0;Ga=ha(Fa,k[O>>2]|0)|0;Ha=ha(Fa,k[P>>2]|0)|0;Ia=(Ga|0)>(Ha|0);Ja=k[V>>2]|0;Ka=k[N>>2]|0;La=ha(Ka,k[E>>2]|0)|0;Ma=ha(Ka,k[J>>2]|0)|0;Na=(La|0)>(Ma|0);Oa=k[Q>>2]|0;Pa=k[D>>2]|0;Qa=ha(Pa,k[x>>2]|0)|0;Ra=ha(Pa,k[w>>2]|0)|0;Sa=(Qa|0)>(Ra|0);Ta=k[K>>2]|0;Ua=(H|0)==0;Va=k[v>>2]|0;Wa=((H|0)<0)<<31>>31;Xa=k[p>>2]|0;Ya=((Xa|0)<0)<<31>>31;Za=o;_a=k[Za>>2]|0;$a=k[Za+4>>2]|0;Za=M;ab=k[Za>>2]|0;bb=k[Za+4>>2]|0;Za=R;cb=k[Za>>2]|0;db=k[Za+4>>2]|0;Za=W;eb=k[Za>>2]|0;fb=k[Za+4>>2]|0;Za=$;gb=k[Za>>2]|0;hb=k[Za+4>>2]|0;Za=ea;ib=k[Za>>2]|0;jb=k[Za+4>>2]|0;Za=ka;kb=k[Za>>2]|0;lb=k[Za+4>>2]|0;Za=pa;mb=k[Za>>2]|0;nb=k[Za+4>>2]|0;Za=0;ob=I;c:while(1){if(g)pb=Za;else{qb=((ob|0)<0)<<31>>31;rb=Za;sb=d;while(1){if(u)tb=rb;else{ub=sb-Aa|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,mb|0,nb|0)|0;ub=L;wb=rb;xb=q;while(1){if(B)yb=wb;else{zb=xb-t|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,kb|0,lb|0)|0;zb=L;Bb=wb;Cb=n;while(1){if(ra)Db=Bb;else{Eb=Cb-A|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,ib|0,jb|0)|0;Eb=L;Gb=Bb;Hb=ta;while(1){if(Da)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Jb=L;Lb=Gb;Mb=ya;while(1){if(Ia)Nb=Lb;else{Ob=Mb-za|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,eb|0,fb|0)|0;Ob=L;Qb=Lb;Rb=Ga;while(1){if(Na)Sb=Qb;else{Tb=Rb-Fa|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,cb|0,db|0)|0;Tb=L;Vb=Qb;Wb=La;while(1){if(Sa)Xb=Vb;else{Yb=Wb-Ka|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Yb=L;_b=Vb;$b=Qa;while(1){ac=k[l>>2]|0;if(Ua|(ac|0)>0)bc=ac;else{ac=$b-Pa|0;cc=yy(_a|0,$a|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(cc|0,L|0,Xa|0,Ya|0)|0;cc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(cc|0,L|0,Ub|0,Tb|0)|0;cc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(cc|0,L|0,Kb|0,Jb|0)|0;cc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(cc|0,L|0,Ab|0,zb|0)|0;cc=ky(ac|0,L|0,vb|0,ub|0)|0;Ym(a,F,ob,qb,cc,L,H,Wa,Ba,1,h,i+(_b<<2)|0,z,y,l)|0;bc=k[l>>2]|0}if((bc|0)>0){C=bc;break c}if(b&(k[y>>2]|0)!=0)k[j>>2]=1;cc=_b+H|0;$b=Va+$b|0;if(($b|0)>(Ra|0)){Xb=cc;break}else _b=cc}}Wb=Ta+Wb|0;if((Wb|0)>(Ma|0)){Sb=Xb;break}else Vb=Xb}}Rb=Oa+Rb|0;if((Rb|0)>(Ha|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ja+Mb|0;if((Mb|0)>(va|0)){Ib=Nb;break}else Lb=Nb}}Hb=Ea+Hb|0;if((Hb|0)>(sa|0)){Db=Ib;break}else Gb=Ib}}Cb=ua+Cb|0;if((Cb|0)>(xa|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=wa+sb|0;if((sb|0)>(s|0)){pb=tb;break}else rb=tb}}ob=ob+G|0;if((ob|0)>(Ca|0))break b;else Za=pb}r=m;return C|0}while(0);C=k[l>>2]|0;r=m;return C|0}function zn(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0;n=r;r=r+16|0;o=k[m>>2]|0;if((g|0)==0&(h|0)==0|(o|0)>0){p=o;r=n;return p|0}Ym(a,b,c,d,e,f,g,h,1,1,i,j,n,l,m)|0;p=k[m>>2]|0;r=n;return p|0}function An(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=l[a+q>>0];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(l[a+s>>0]|0|0)*c+d;do if(!(t<-2147483648.49))if(t>2147483647.49){k[n>>2]=-11;k[m+(s<<2)>>2]=2147483647;break}else{k[m+(s<<2)>>2]=~~t;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;s=0;do{q=i[a+s>>0]|0;do if(q<<24>>24==f<<24>>24){k[j>>2]=1;if(o){k[m+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else k[m+(s<<2)>>2]=q&255;while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{s=i[a+e>>0]|0;do if(s<<24>>24==f<<24>>24){k[j>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s&255|0)*c+d;if(t<-2147483648.49){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if(t>2147483647.49){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function Bn(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=j[a+(q<<1)>>1];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(j[a+(s<<1)>>1]|0)*c+d;do if(!(t<-2147483648.49))if(t>2147483647.49){k[n>>2]=-11;k[m+(s<<2)>>2]=2147483647;break}else{k[m+(s<<2)>>2]=~~t;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;s=0;do{q=j[a+(s<<1)>>1]|0;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){k[m+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else k[m+(s<<2)>>2]=q<<16>>16;while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{s=j[a+(e<<1)>>1]|0;do if(s<<16>>16==f<<16>>16){k[l>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s<<16>>16)*c+d;if(t<-2147483648.49){k[n>>2]=-11;k[m+(e<<2)>>2]=-2147483648;break}if(t>2147483647.49){k[n>>2]=-11;k[m+(e<<2)>>2]=2147483647;break}else{k[m+(e<<2)>>2]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function Cn(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0.0;n=c==1.0&d==0.0;o=(b|0)>0;if(!e){if(n){if(o)p=0;else{q=k[m>>2]|0;return q|0}do{k[l+(p<<2)>>2]=k[a+(p<<2)>>2];p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}if(o)r=0;else{q=k[m>>2]|0;return q|0}do{s=+(k[a+(r<<2)>>2]|0)*c+d;do if(!(s<-2147483648.49))if(s>2147483647.49){k[m>>2]=-11;k[l+(r<<2)>>2]=2147483647;break}else{k[l+(r<<2)>>2]=~~s;break}else{k[m>>2]=-11;k[l+(r<<2)>>2]=-2147483648}while(0);r=r+1|0}while((r|0)!=(b|0));q=k[m>>2]|0;return q|0}if(n){if(!o){q=k[m>>2]|0;return q|0}n=(e|0)==1;r=0;do{p=k[a+(r<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(n){k[l+(r<<2)>>2]=g;break}else{i[h+r>>0]=1;break}}else k[l+(r<<2)>>2]=p;while(0);r=r+1|0}while((r|0)!=(b|0));q=k[m>>2]|0;return q|0}if(!o){q=k[m>>2]|0;return q|0}o=(e|0)==1;e=0;do{r=k[a+(e<<2)>>2]|0;do if((r|0)==(f|0)){k[j>>2]=1;if(o){k[l+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{s=+(r|0)*c+d;if(s<-2147483648.49){k[m>>2]=-11;k[l+(e<<2)>>2]=-2147483648;break}if(s>2147483647.49){k[m>>2]=-11;k[l+(e<<2)>>2]=2147483647;break}else{k[l+(e<<2)>>2]=~~s;break}}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}function Dn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-2147483648.49))if(s>2147483647.49){k[l>>2]=-11;k[j+(p<<2)>>2]=2147483647;break}else{k[j+(p<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(p<<2)>>2]=-2147483648}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-2147483648.49))if(r>2147483647.49){k[l>>2]=-11;k[j+(t<<2)>>2]=2147483647;break}else{k[j+(t<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(t<<2)>>2]=-2147483648}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-2147483648.49){k[l>>2]=-11;k[j+(u<<2)>>2]=-2147483648;break a}if(s>2147483647.49){k[l>>2]=-11;k[j+(u<<2)>>2]=2147483647;break a}else{k[j+(u<<2)>>2]=~~r;break a}break}case 1:{k[h>>2]=1;if(p){k[j+(u<<2)>>2]=f;break a}else{i[g+u>>0]=1;break a}break}default:k[j+(u<<2)>>2]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-2147483648.49;v=d>2147483647.49;u=~~d;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-2147483648.49){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(r>2147483647.49){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=~~r;break b}break}case 1:{k[h>>2]=1;if(p){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(v){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function En(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-2147483648.49))if(r>2147483647.49){k[l>>2]=-11;k[j+(o<<2)>>2]=2147483647;break}else{k[j+(o<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(o<<2)>>2]=-2147483648}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-2147483648.49))if(r>2147483647.49){k[l>>2]=-11;k[j+(s<<2)>>2]=2147483647;break}else{k[j+(s<<2)>>2]=~~r;break}else{k[l>>2]=-11;k[j+(s<<2)>>2]=-2147483648}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-2147483648.49){k[l>>2]=-11;k[j+(t<<2)>>2]=-2147483648;break a}if(r>2147483647.49){k[l>>2]=-11;k[j+(t<<2)>>2]=2147483647;break a}else{k[j+(t<<2)>>2]=~~r;break a}break}case 1:{k[h>>2]=1;if(o){k[j+(t<<2)>>2]=f;break a}else{i[g+t>>0]=1;break a}break}default:k[j+(t<<2)>>2]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-2147483648.49;u=d>2147483647.49;t=~~d;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-2147483648.49){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(r>2147483647.49){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=~~r;break b}break}case 1:{k[h>>2]=1;if(o){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=-2147483648;break b}if(u){k[l>>2]=-11;k[j+(e<<2)>>2]=2147483647;break b}else{k[j+(e<<2)>>2]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function Fn(a,b,c,d,e,f,g,h,j,l,m,n,o,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;var s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0;s=r;r=r+29024|0;t=s+56|0;u=s+88|0;v=s+84|0;w=s+72|0;x=s+48|0;y=s+40|0;z=s+32|0;A=s+16|0;B=s+28936|0;C=s+136|0;D=k[q>>2]|0;if((g|0)==0&(h|0)==0|(D|0)>0){E=D;r=s;return E|0}D=(o|0)!=0;if(D)k[o>>2]=0;F=(j|0)==2;if(F)sy(n|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,0,s+8|0,s,s+112|0,s+76|0,u,v,y,z,w,x,A,s+80|0,s+24|0,s+92|0,q)|0)>0){E=k[q>>2]|0;r=s;return E|0}if((k[u>>2]|0)!=14){k[q>>2]=310;E=310;r=s;return E|0}if(!g){E=k[q>>2]|0;r=s;return E|0}u=F&((j|0)!=1|l<<24>>24!=0);j=g;F=0;f=0;e=g;g=h;h=0;d=0;a:while(1){c=e;b=g;G=j;H=F;I=f;while(1){J=k[v>>2]|0;K=(G|0)<(J|0)?G:J;J=((K|0)<0)<<31>>31;M=x;N=z;O=k[N>>2]|0;P=k[N+4>>2]|0;N=jy(k[M>>2]|0,k[M+4>>2]|0,O|0,P|0)|0;M=L;Q=(J|0)<(M|0)|(J|0)==(M|0)&K>>>0<N>>>0?K:N;N=y;K=k[N>>2]|0;M=k[N+4>>2]|0;N=A;J=yy(k[N>>2]|0,k[N+4>>2]|0,h|0,d|0)|0;N=ky(J|0,L|0,K|0,M|0)|0;M=L;K=k[w>>2]|0;J=yy(K|0,((K|0)<0)<<31>>31|0,O|0,P|0)|0;P=ky(N|0,M|0,J|0,L|0)|0;lu(a,P,L,Q,K,C,q)|0;do if((Q|0)>0){K=Q+-1|0;P=C+Q|0;if(D)if(u){J=H;M=I;N=C;while(1){O=i[N>>0]|0;switch(O<<24>>24){case 84:{i[m+J>>0]=1;break}case 70:{i[m+J>>0]=0;break}case 0:{i[m+J>>0]=l;k[o>>2]=1;i[n+J>>0]=1;break}case 1:{i[m+J>>0]=49;break}default:i[m+J>>0]=O}J=ky(J|0,M|0,1,0)|0;N=N+1|0;if((N|0)==(P|0))break;else M=L}M=ky(H|0,I|0,1,0)|0;N=ky(M|0,L|0,K|0,0)|0;R=N;S=L;break}else{N=H;M=I;J=C;while(1){O=i[J>>0]|0;switch(O<<24>>24){case 84:{i[m+N>>0]=1;break}case 70:{i[m+N>>0]=0;break}case 0:{i[m+N>>0]=l;k[o>>2]=1;break}case 1:{i[m+N>>0]=49;break}default:i[m+N>>0]=O}N=ky(N|0,M|0,1,0)|0;J=J+1|0;if((J|0)==(P|0))break;else M=L}M=ky(H|0,I|0,1,0)|0;J=ky(M|0,L|0,K|0,0)|0;R=J;S=L;break}else if(u){J=H;M=I;N=C;while(1){O=i[N>>0]|0;switch(O<<24>>24){case 84:{i[m+J>>0]=1;break}case 70:{i[m+J>>0]=0;break}case 0:{i[m+J>>0]=l;i[n+J>>0]=1;break}case 1:{i[m+J>>0]=49;break}default:i[m+J>>0]=O}J=ky(J|0,M|0,1,0)|0;N=N+1|0;if((N|0)==(P|0))break;else M=L}M=ky(H|0,I|0,1,0)|0;N=ky(M|0,L|0,K|0,0)|0;R=N;S=L;break}else{N=H;M=I;J=C;while(1){O=i[J>>0]|0;switch(O<<24>>24){case 84:{i[m+N>>0]=1;break}case 70:{i[m+N>>0]=0;break}case 0:{i[m+N>>0]=l;break}case 1:{i[m+N>>0]=49;break}default:i[m+N>>0]=O}N=ky(N|0,M|0,1,0)|0;J=J+1|0;if((J|0)==(P|0))break;else M=L}M=ky(H|0,I|0,1,0)|0;P=ky(M|0,L|0,K|0,0)|0;R=P;S=L;break}}else{R=H;S=I}while(0);P=k[q>>2]|0;if((P|0)>0){T=Q;U=R;V=S;W=50;break a}M=((Q|0)<0)<<31>>31;J=c;N=jy(c|0,b|0,Q|0,M|0)|0;O=b;X=L;if(!((J|0)==(Q|0)&(O|0)==(M|0))?(O=z,J=ky(k[O>>2]|0,k[O+4>>2]|0,Q|0,M|0)|0,M=L,O=z,k[O>>2]=J,k[O+4>>2]=M,O=x,(J|0)==(k[O>>2]|0)?(M|0)==(k[O+4>>2]|0):0):0){Y=P;Z=N;_=X;$=R;aa=S;break}if(!N){E=P;W=56;break a}else{c=N;b=X;G=N;H=R;I=S}}I=z;k[I>>2]=0;k[I+4>>2]=0;I=ky(h|0,d|0,1,0)|0;if(!Z){E=Y;W=56;break}else{j=Z;F=$;f=aa;e=Z;g=_;h=I;d=L}}if((W|0)==50){ba=+(U>>>0)+4294967296.0*+(V|0);p[t>>3]=ba+1.0;p[t+8>>3]=+(T|0)+ba;Xw(B,760109,t)|0;tk(B);E=k[q>>2]|0;r=s;return E|0}else if((W|0)==56){r=s;return E|0}return 0}function Gn(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;Fn(a,b,c,d,e,f,g,h,2,0,i,j,l,m)|0;return k[m>>2]|0}function Hn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0;m=r;r=r+16|0;n=m+4|0;o=m;p=m+8|0;q=k[l>>2]|0;if((h|0)<0|(h|0)==0&g>>>0<1|(q|0)>0){s=q;r=m;return s|0}if((d|0)<0|(d|0)==0&c>>>0<1){k[l>>2]=307;s=307;r=m;return s|0}if((f|0)<0|(f|0)==0&e>>>0<1){k[l>>2]=308;s=308;r=m;return s|0}q=k[a>>2]|0;t=a+4|0;u=k[t>>2]|0;if((q|0)==(k[u+68>>2]|0)){v=u+120|0;if(((k[v>>2]|0)==-1?(k[v+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){s=k[l>>2]|0;r=m;return s|0}}else Fk(a,q+1|0,0,l)|0;q=ky(e|0,f|0,7,0)|0;v=wy(q|0,L|0,8,0)|0;q=ky(e|0,f|0,-1,-1)|0;u=L;w=wy(q|0,u|0,8,0)|0;x=ty(w|0,L|0,3)|0;w=jy(q|0,u|0,x|0,L|0)|0;x=ky(c|0,d|0,-1,-1)|0;u=L;q=v+-1|0;y=((q|0)<0)<<31>>31;z=k[t>>2]|0;A=k[z+960>>2]|0;B=b+-1|0;C=k[A+(B*160|0)+80>>2]|0;if((((C|0)>-1?C:0-C|0)|0)>11){k[l>>2]=310;s=310;r=m;return s|0}do if((C|0)>0){D=k[A+(B*160|0)+88>>2]|0;E=(C|0)==1?(D+7|0)/8|0:D;k[o>>2]=E;if((v|0)<=(E|0)){E=z+120|0;D=k[E>>2]|0;F=k[E+4>>2]|0;E=z+952|0;G=yy(k[E>>2]|0,k[E+4>>2]|0,x|0,u|0)|0;E=ky(G|0,L|0,D|0,F|0)|0;F=A+(B*160|0)+72|0;D=ky(E|0,L|0,k[F>>2]|0,k[F+4>>2]|0)|0;H=D;I=L;J=0;break}k[l>>2]=308;s=308;r=m;return s|0}else{vl(a,b,c,d,o,n,l)|0;D=k[o>>2]|0;if((C|0)==-1){F=(D+7|0)/8|0;k[o>>2]=F;K=F}else K=D;D=ky(e|0,f|0,6,0)|0;F=ky(D|0,L|0,g|0,h|0)|0;D=wy(F|0,L|0,8,0)|0;F=L;E=((K|0)<0)<<31>>31;if(!((F|0)>(E|0)|(F|0)==(E|0)&D>>>0>K>>>0)){D=k[t>>2]|0;E=D+120|0;F=k[n>>2]|0;G=ky(F|0,((F|0)<0)<<31>>31|0,k[E>>2]|0,k[E+4>>2]|0)|0;E=D+968|0;D=ky(G|0,L|0,k[E>>2]|0,k[E+4>>2]|0)|0;H=D;I=L;J=1;break}k[l>>2]=308;s=308;r=m;return s|0}while(0);n=ky(H|0,I|0,q|0,y|0)|0;if((au(a,n,L,0,l)|0)>0){s=k[l>>2]|0;r=m;return s|0}n=(J|0)==0;J=A+(B*160|0)+72|0;B=x;x=u;u=q;q=y;y=w;w=0;a:while(1){A=u;I=q;H=y;K=w;while(1){f=H;e=K;while(1){if((fu(a,1,0,p,l)|0)>0){M=28;break a}C=((e|0)<0)<<31>>31;if((f|0)<8&((C|0)<(h|0)|(C|0)==(h|0)&e>>>0<g>>>0)){d=i[p>>0]|0;c=e+1|0;b=((c|0)<0)<<31>>31;z=ky(c|0,b|0,-1,-1)|0;v=L;D=(b|0)<(h|0)|(b|0)==(h|0)&c>>>0<g>>>0;E=jy(z|0,v|0,(D?g:c)|0,(D?h:b)|0)|0;b=L;D=~(7-f);G=b>>>0>4294967295|(b|0)==-1&E>>>0>D>>>0;F=G?E:D;D=jy(z|0,v|0,F|0,(G?b:-1)|0)|0;b=L;G=~F;F=f;v=e;do{i[j+v>>0]=(i[771129+F>>0]&d)<<24>>24!=0&1;F=F+1|0;v=v+1|0;z=((v|0)<0)<<31>>31}while((F|0)<8&((z|0)<(h|0)|(z|0)==(h|0)&v>>>0<g>>>0));N=D;O=b;P=c+G|0}else{N=e;O=C;P=e}if((N|0)==(g|0)&(O|0)==(h|0)){M=32;break a}if(n){Q=P;break}else{f=0;e=P}}A=ky(A|0,I|0,1,0)|0;I=L;e=k[o>>2]|0;if((A|0)==(e|0)&(I|0)==(((e|0)<0)<<31>>31|0)){R=Q;break}else{H=0;K=Q}}K=ky(B|0,x|0,1,0)|0;H=L;I=k[t>>2]|0;A=I+120|0;e=k[A>>2]|0;f=k[A+4>>2]|0;A=I+952|0;I=yy(k[A>>2]|0,k[A+4>>2]|0,K|0,H|0)|0;A=ky(I|0,L|0,e|0,f|0)|0;f=J;e=ky(A|0,L|0,k[f>>2]|0,k[f+4>>2]|0)|0;au(a,e,L,0,l)|0;B=K;x=H;u=0;q=0;y=0;w=R}if((M|0)==28){s=k[l>>2]|0;r=m;return s|0}else if((M|0)==32){s=k[l>>2]|0;r=m;return s|0}return 0}function In(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0;n=r;r=r+16|0;Jn(a,b,c,d,e,f,g,h,1,i,j,n,l,m)|0;r=n;return k[m>>2]|0}function Jn(a,b,c,d,e,f,g,h,l,m,n,q,s,t){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;q=q|0;s=s|0;t=t|0;var u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0.0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0;u=r;r=r+800|0;v=u+112|0;w=u+104|0;x=u+96|0;y=u+88|0;z=u+80|0;A=u+72|0;B=u+64|0;C=u+56|0;D=u+48|0;E=u+40|0;F=u+32|0;G=u+132|0;H=u+128|0;I=u+124|0;J=u+120|0;K=u+715|0;M=u+640|0;N=u+8|0;O=u+216|0;P=u+240|0;Q=u+136|0;R=u;p[R>>3]=1.0;S=k[t>>2]|0;if((g|0)==0&(h|0)==0|(S|0)>0){T=S;r=u;return T|0}S=k[a>>2]|0;U=a+4|0;V=k[U>>2]|0;if((S|0)==(k[V+68>>2]|0)){W=V+120|0;if(((k[W>>2]|0)==-1?(k[W+4>>2]|0)==-1:0)?(Vk(a,t)|0)>0:0){T=k[t>>2]|0;r=u;return T|0}}else Fk(a,S+1|0,0,t)|0;if((b|0)>=1?(k[(k[U>>2]|0)+928>>2]|0)>=(b|0):0){_k(a,b,J,0,0,t)|0;S=k[J>>2]|0;if((S|0)<0){W=0-S|0;k[J>>2]=W;X=W}else X=S;S=k[(k[(k[U>>2]|0)+960>>2]|0)+((b+-1|0)*160|0)+80>>2]|0;U=(S|0)>-1?S:0-S|0;a:do switch(U|0){case 16:{Ln(a,b,c,d,e,f,g,h,l,m,n,q,s,t)|0;break}case 14:{S=cy(g)|0;Fn(a,b,c,d,e,f,g,h,l,i[m>>0]|0,S,q,s,t)|0;if((k[t>>2]|0)<1&((h|0)>0|(h|0)==0&g>>>0>0)){W=0;do{J=i[S+W>>0]|0;do if(J<<24>>24!=1){V=k[n+(W<<2)>>2]|0;if(!(J<<24>>24)){i[V>>0]=70;i[V+1>>0]=0;break}else{i[V>>0]=78;i[V+1>>0]=0;break}}else{V=k[n+(W<<2)>>2]|0;i[V>>0]=84;i[V+1>>0]=0}while(0);W=W+1|0;J=((W|0)<0)<<31>>31}while((J|0)<(h|0)|(J|0)==(h|0)&W>>>0<g>>>0)}dy(S);break}case 83:{W=ty(g|0,h|0,1)|0;J=L;V=ey(W,4)|0;Y=ty(e|0,f|0,1)|0;Z=ky(Y|0,L|0,-1,-1)|0;Bm(a,b,c,d,Z,L,W,J,1,1,-9.119119840596153e-36,V,q,s,t)|0;if((k[t>>2]|0)<1){Mn(a,b,I,t)|0;J=((k[I>>2]|0)+-3|0)/2|0;k[I>>2]=J;Gk(770521,b,M,t)|0;k[H>>2]=0;i[N>>0]=0;if(!(Jo(a,M,O,0,H)|0))Sk(O,N);if(!(i[N>>0]|0)){i[N>>0]=i[760177]|0;i[N+1>>0]=i[760178]|0;i[N+2>>0]=i[760179]|0;i[N+3>>0]=i[760180]|0;i[N+4>>0]=i[760181]|0;i[N+5>>0]=i[760182]|0;i[N+6>>0]=i[760183]|0}if((h|0)>0|(h|0)==0&g>>>0>0){W=(l|0)==2;Z=0;Y=0;while(1){_=n+(Z<<2)|0;$=k[_>>2]|0;i[$>>0]=40;i[$+1>>0]=0;aa=+o[V+(Y<<2)>>2];if(aa==-9.119119840596153e-36){i[P>>0]=i[760184]|0;i[P+1>>0]=i[760185]|0;i[P+2>>0]=i[760186]|0;i[P+3>>0]=i[760187]|0;i[P+4>>0]=i[760188]|0;if(W)i[q+Z>>0]=1}else{p[E>>3]=aa;Xw(P,N,E)|0}zx(k[_>>2]|0,P,J)|0;$=k[_>>2]|0;ba=$+(xx($)|0)|0;i[ba>>0]=44;i[ba+1>>0]=0;aa=+o[V+((Y|1)<<2)>>2];if(aa==-9.119119840596153e-36){i[P>>0]=i[760184]|0;i[P+1>>0]=i[760185]|0;i[P+2>>0]=i[760186]|0;i[P+3>>0]=i[760187]|0;i[P+4>>0]=i[760188]|0;if(W)i[q+Z>>0]=1}else{p[D>>3]=aa;Xw(P,N,D)|0}zx(k[_>>2]|0,P,J)|0;ba=k[_>>2]|0;_=ba+(xx(ba)|0)|0;i[_>>0]=41;i[_+1>>0]=0;Z=Z+1|0;_=((Z|0)<0)<<31>>31;if(!((_|0)<(h|0)|(_|0)==(h|0)&Z>>>0<g>>>0))break;else Y=Y+2|0}}}dy(V);break}case 163:{Y=ty(g|0,h|0,1)|0;Z=L;J=ey(Y,8)|0;W=ty(e|0,f|0,1)|0;S=ky(W|0,L|0,-1,-1)|0;nm(a,b,c,d,S,L,Y,Z,1,1,-9.1191291391491e-36,J,q,s,t)|0;if((k[t>>2]|0)<1){Mn(a,b,I,t)|0;Z=((k[I>>2]|0)+-3|0)/2|0;k[I>>2]=Z;Gk(770521,b,M,t)|0;k[H>>2]=0;i[N>>0]=0;if(!(Jo(a,M,O,0,H)|0))Sk(O,N);if(!(i[N>>0]|0)){Y=N;k[Y>>2]=775107109;k[Y+4>>2]=4535601}if((h|0)>0|(h|0)==0&g>>>0>0){Y=(l|0)==2;S=0;W=0;while(1){_=n+(S<<2)|0;ba=k[_>>2]|0;i[ba>>0]=40;i[ba+1>>0]=0;aa=+p[J+(W<<3)>>3];if(aa==-9.1191291391491e-36){i[P>>0]=i[760184]|0;i[P+1>>0]=i[760185]|0;i[P+2>>0]=i[760186]|0;i[P+3>>0]=i[760187]|0;i[P+4>>0]=i[760188]|0;if(Y)i[q+S>>0]=1}else{p[C>>3]=aa;Xw(P,N,C)|0}zx(k[_>>2]|0,P,Z)|0;ba=k[_>>2]|0;$=ba+(xx(ba)|0)|0;i[$>>0]=44;i[$+1>>0]=0;aa=+p[J+((W|1)<<3)>>3];if(aa==-9.1191291391491e-36){i[P>>0]=i[760184]|0;i[P+1>>0]=i[760185]|0;i[P+2>>0]=i[760186]|0;i[P+3>>0]=i[760187]|0;i[P+4>>0]=i[760188]|0;if(Y)i[q+S>>0]=1}else{p[B>>3]=aa;Xw(P,N,B)|0}zx(k[_>>2]|0,P,Z)|0;$=k[_>>2]|0;_=$+(xx($)|0)|0;i[_>>0]=41;i[_+1>>0]=0;S=S+1|0;_=((S|0)<0)<<31>>31;if(!((_|0)<(h|0)|(_|0)==(h|0)&S>>>0<g>>>0))break;else W=W+2|0}}}dy(J);break}default:{W=(U|0)==81;S=ey(g,8)|0;if(W&(X|0)==81){Z=ey(g,1)|0;k[I>>2]=20;if((nn(a,b,c,d,e,f,g,h,S,Z,s,t)|0)>0){dy(Z);dy(S);T=k[t>>2]|0;r=u;return T|0}if(!m){j[Q>>1]=32;ca=1}else{ux(Q,m)|0;ca=xx(m)|0}if((h|0)>0|(h|0)==0&g>>>0>0){Y=(l|0)==2;if((ca|0)>20){V=0;do{if(i[Z+V>>0]|0){_=n+(V<<2)|0;i[k[_>>2]>>0]=0;zx(k[_>>2]|0,Q,20)|0;if(Y)i[q+V>>0]=1}else{_=S+(V<<3)|0;$=k[_+4>>2]|0;ba=A;k[ba>>2]=k[_>>2];k[ba+4>>2]=$;Xw(P,760189,A)|0;$=n+(V<<2)|0;i[k[$>>2]>>0]=0;zx(k[$>>2]|0,P,20)|0}V=V+1|0;$=((V|0)<0)<<31>>31}while(($|0)<(h|0)|($|0)==(h|0)&V>>>0<g>>>0)}else{V=0;do{if(i[Z+V>>0]|0){J=n+(V<<2)|0;i[k[J>>2]>>0]=0;$=k[J>>2]|0;k[z>>2]=20;k[z+4>>2]=Q;Xw($,760196,z)|0;if(Y)i[q+V>>0]=1}else{$=S+(V<<3)|0;J=k[$+4>>2]|0;ba=y;k[ba>>2]=k[$>>2];k[ba+4>>2]=J;Xw(P,760189,y)|0;J=n+(V<<2)|0;i[k[J>>2]>>0]=0;zx(k[J>>2]|0,P,20)|0}V=V+1|0;J=((V|0)<0)<<31>>31}while((J|0)<(h|0)|(J|0)==(h|0)&V>>>0<g>>>0)}}dy(Z);dy(S);break a}if((nm(a,b,c,d,e,f,g,h,1,l,-9.1191291391491e-36,S,q,s,t)|0)>0){dy(S);T=k[t>>2]|0;r=u;return T|0}Mn(a,b,I,t)|0;Gk(773339,b,M,t)|0;k[H>>2]=0;V=(Mo(a,M,R,0,H)|0)==0;Y=V&+p[R>>3]!=1.0;V=((U|0)>41|Y)&1^1;Gk(770521,b,M,t)|0;k[H>>2]=0;i[N>>0]=0;if(!(Jo(a,M,O,0,H)|0))Sk(O,N);b:do if(!(i[N>>0]|0)){Gk(773003,b,M,t)|0;Jo(a,M,O,0,t)|0;if((U|0)<22&Y){J=N;k[J>>2]=875635493;k[J+4>>2]=4666926;da=V;break}J=(U|0)==41;if(J&Y){ea=N;fa=760200;ga=ea+9|0;do{i[ea>>0]=i[fa>>0]|0;ea=ea+1|0;fa=fa+1|0}while((ea|0)<(ga|0));da=V;break}if(W&Y){ea=N;fa=760200;ga=ea+9|0;do{i[ea>>0]=i[fa>>0]|0;ea=ea+1|0;fa=fa+1|0}while((ea|0)<(ga|0));da=V;break}Zk(a,G,t)|0;if((k[G>>2]|0)==1){Sk(O,N);da=V;break}switch(U|0){case 1:{k[N>>2]=6566949;da=V;break b;break}case 11:{k[N>>2]=6566949;da=V;break b;break}case 21:{k[N>>2]=6567461;da=V;break b;break}default:{if(J){i[N>>0]=i[760209]|0;i[N+1>>0]=i[760210]|0;i[N+2>>0]=i[760211]|0;i[N+3>>0]=i[760212]|0;i[N+4>>0]=i[760213]|0;i[N+5>>0]=i[760214]|0;i[N+6>>0]=i[760215]|0;da=0;break b}switch(U|0){case 42:{ba=N;k[ba>>2]=875635493;k[ba+4>>2]=4666926;da=V;break b;break}case 82:{ea=N;fa=760200;ga=ea+9|0;do{i[ea>>0]=i[fa>>0]|0;ea=ea+1|0;fa=fa+1|0}while((ea|0)<(ga|0));da=V;break b;break}default:{da=V;break b}}}}}else da=V;while(0);if(!m){j[Q>>1]=32;ha=1}else{ux(Q,m)|0;ha=xx(m)|0}if((h|0)>0|(h|0)==0&g>>>0>0){V=(U|0)==1;Y=(da|0)==0;W=k[I>>2]|0;Z=(W|0)<(ha|0);J=0;do{do if(V){ba=~~+p[S+(J<<3)>>3];$=ba&255;_=n+(J<<2)|0;i[k[_>>2]>>0]=(ba&255)>>>7|48;i[(k[_>>2]|0)+1>>0]=($&64)>>>6|48;i[(k[_>>2]|0)+2>>0]=($&32)>>>5|48;i[(k[_>>2]|0)+3>>0]=($&16)>>>4|48;i[(k[_>>2]|0)+4>>0]=($&8)>>>3|48;i[(k[_>>2]|0)+5>>0]=($&4)>>>2|48;i[(k[_>>2]|0)+6>>0]=($&2)>>>1|48;i[(k[_>>2]|0)+7>>0]=$&1|48;i[(k[_>>2]|0)+8>>0]=0}else{switch(l|0){case 1:{if(+p[S+(J<<3)>>3]==-9.1191291391491e-36)ia=107;break}case 2:{if(i[q+J>>0]|0)ia=107;break}default:{}}if((ia|0)==107){ia=0;_=n+(J<<2)|0;i[k[_>>2]>>0]=0;$=k[_>>2]|0;if(Z){zx($,Q,W)|0;break}else{k[x>>2]=W;k[x+4>>2]=Q;Xw($,760196,x)|0;break}}aa=+p[S+(J<<3)>>3];if(Y){p[v>>3]=aa;Xw(P,N,v)|0}else{k[w>>2]=~~aa;Xw(P,N,w)|0}if((xx(P)|0)>(W|0))sy(P|0,42,W|0)|0;$=n+(J<<2)|0;i[k[$>>2]>>0]=0;zx(k[$>>2]|0,P,W)|0}while(0);J=J+1|0;$=((J|0)<0)<<31>>31}while(($|0)<(h|0)|($|0)==(h|0)&J>>>0<g>>>0)}dy(S)}}while(0);T=k[t>>2]|0;r=u;return T|0}k[F>>2]=b;Xw(K,771145,F)|0;tk(K);k[t>>2]=302;T=302;r=u;return T|0}function Kn(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0;n=r;r=r+16|0;Jn(a,b,c,d,e,f,g,h,2,n,i,j,l,m)|0;r=n;return k[m>>2]|0}function Ln(a,b,c,d,e,f,g,h,j,l,m,n,o,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;var s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0.0;s=r;r=r+29024|0;t=s+28864|0;u=s+28856|0;v=s+28896|0;w=s+28892|0;x=s+28888|0;y=s+28884|0;z=s+28880|0;A=s+48|0;B=s+40|0;C=s+32|0;D=s+24|0;E=s+16|0;F=s+8|0;G=s;H=s+28920|0;I=s+28940|0;J=s+28900|0;K=s+56|0;M=k[q>>2]|0;if((g|0)==0&(h|0)==0|(M|0)>0){N=M;r=s;return N|0}M=k[a>>2]|0;O=a+4|0;if((M|0)!=(k[(k[O>>2]|0)+68>>2]|0))Fk(a,M+1|0,0,q)|0;if(o)k[o>>2]=0;if((j|0)==2)sy(n|0,0,g|0)|0;if((b|0)>=1?(M=k[O>>2]|0,(k[M+928>>2]|0)>=(b|0)):0){O=k[(k[M+960>>2]|0)+((b+-1|0)*160|0)+80>>2]|0;k[v>>2]=O;a:do switch(O|0){case -16:{if((nl(a,b,c,d,1,0,1,0,0,F,G,H,y,v,w,B,C,z,A,E,x,D,J,q)|0)<=0){M=k[A>>2]|0;k[y>>2]=M;P=J;Q=M;R=1;S=0;break a}N=k[q>>2]|0;r=s;return N|0}case 16:{if((nl(a,b,c,d,e,f,g,h,0,F,G,H,y,v,w,B,C,z,A,E,x,D,J,q)|0)>0){N=k[q>>2]|0;r=s;return N|0}else{M=k[y>>2]|0;if((M|0)<=2880){P=J;Q=M;R=g;S=h;break a}k[w>>2]=1;k[z>>2]=M;T=A;k[T>>2]=1;k[T+4>>2]=0;P=J;Q=M;R=g;S=h;break a}break}default:{k[q>>2]=309;N=309;r=s;return N|0}}while(0);h=xx(P)|0;g=(h|0)==0?1:h;h=(j|0)==1;J=(l|0)==0;do if(h&J)U=0;else{if(h&(l|0)!=0){if((i[l>>0]|0)==0|(i[P>>0]|0)==1){U=0;break}}else if((i[P>>0]|0)==1){U=0;break}U=(g|0)>(Q|0)?0:j}while(0);if((R|0)==0&(S|0)==0){N=k[q>>2]|0;r=s;return N|0}j=(U|0)==0;U=0;Q=0;D=0;x=0;v=R;R=S;b:while(1){S=D;H=x;G=R;F=v;while(1){f=k[w>>2]|0;e=((f|0)<0)<<31>>31;d=(G|0)<(e|0)|(G|0)==(e|0)&F>>>0<f>>>0;e=d?F:f;f=my(0,e|0,32)|0;e=L;d=A;c=C;O=k[c>>2]|0;M=k[c+4>>2]|0;c=jy(k[d>>2]|0,k[d+4>>2]|0,O|0,M|0)|0;d=L;T=(e|0)<(d|0)|(e|0)==(d|0)&f>>>0<c>>>0;d=T?f:c;c=B;f=k[c>>2]|0;T=k[c+4>>2]|0;c=E;e=yy(k[c>>2]|0,k[c+4>>2]|0,U|0,Q|0)|0;c=ky(e|0,L|0,f|0,T|0)|0;T=L;f=k[z>>2]|0;e=yy(f|0,((f|0)<0)<<31>>31|0,O|0,M|0)|0;M=ky(c|0,T|0,e|0,L|0)|0;au(a,M,L,0,q)|0;M=k[z>>2]|0;e=k[y>>2]|0;if((M|0)==(e|0)){T=ha(d,M)|0;fu(a,T,((T|0)<0)<<31>>31,K,q)|0}else gu(a,e,d,M-e|0,K,q)|0;e=k[y>>2]|0;M=K+((ha(e,d)|0)+-1)|0;T=((d|0)<0)<<31>>31;c=ky(d|0,T|0,S|0,H|0)|0;O=L;f=ky(c|0,O|0,-1,0)|0;V=((f|0)<0)<<31>>31;c:do if(!((V|0)<(H|0)|(V|0)==(H|0)&f>>>0<S>>>0)){if(j){W=e;X=M;Y=f;while(1){Z=W+-1|0;_=(k[m+(Y<<2)>>2]|0)+Z|0;d:do if((W|0)>1){$=_;aa=X;ba=Z;while(1){if((i[aa>>0]|0)!=32){ca=$;da=aa;ea=ba;break d}fa=aa+-1|0;ga=$+-1|0;ia=ba+-1|0;if((ba|0)>1){$=ga;aa=fa;ba=ia}else{ca=ga;da=fa;ea=ia;break}}}else{ca=_;da=X;ea=Z}while(0);i[ca+1>>0]=0;if((ea|0)>-1){Z=1-ea|0;_=ca;ba=da;aa=ea;while(1){i[_>>0]=i[ba>>0]|0;if((aa|0)>0){_=_+-1|0;ba=ba+-1|0;aa=aa+-1|0}else break}ja=da+(Z+-2)|0}else ja=da;aa=Y+-1|0;ba=((aa|0)<0)<<31>>31;if((ba|0)<(H|0)|(ba|0)==(H|0)&aa>>>0<S>>>0)break c;W=k[y>>2]|0;X=ja;Y=aa}}else{ka=e;la=M;ma=f}while(1){Y=m+(ma<<2)|0;X=ka+-1|0;W=(k[Y>>2]|0)+X|0;e:do if((ka|0)>1){aa=W;ba=la;_=X;while(1){if((i[ba>>0]|0)!=32){na=aa;oa=ba;pa=_;break e}$=ba+-1|0;ia=aa+-1|0;fa=_+-1|0;if((_|0)>1){aa=ia;ba=$;_=fa}else{na=ia;oa=$;pa=fa;break}}}else{na=W;oa=la;pa=X}while(0);i[na+1>>0]=0;if((pa|0)>-1){X=1-pa|0;W=na;Z=oa;_=pa;while(1){i[W>>0]=i[Z>>0]|0;if((_|0)>0){W=W+-1|0;Z=Z+-1|0;_=_+-1|0}else break}qa=oa+(X+-2)|0}else qa=oa;_=k[Y>>2]|0;do if(!(Ax(P,_,g)|0)){k[o>>2]=1;if(!h){i[n+ma>>0]=1;break}if(J){i[_>>0]=32;i[_+1>>0]=0;break}else{ux(_,l)|0;break}}while(0);_=ma+-1|0;Y=((_|0)<0)<<31>>31;if((Y|0)<(H|0)|(Y|0)==(H|0)&_>>>0<S>>>0)break c;ka=k[y>>2]|0;la=qa;ma=_}}while(0);f=k[q>>2]|0;if((f|0)>0){ra=S;sa=H;ta=d;break b}M=jy(F|0,G|0,d|0,T|0)|0;e=L;if((F|0)==(d|0)&(G|0)==(T|0)){N=f;ua=63;break b}f=C;V=ky(k[f>>2]|0,k[f+4>>2]|0,d|0,T|0)|0;f=L;_=C;k[_>>2]=V;k[_+4>>2]=f;_=A;if((V|0)==(k[_>>2]|0)?(f|0)==(k[_+4>>2]|0):0){va=M;wa=e;xa=c;ya=O;break}else{S=c;H=O;G=e;F=M}}F=C;k[F>>2]=0;k[F+4>>2]=0;F=ky(U|0,Q|0,1,0)|0;U=F;Q=L;D=xa;x=ya;v=va;R=wa}if((ua|0)==63){r=s;return N|0}za=+(ra>>>0)+4294967296.0*+(sa|0);p[t>>3]=za+1.0;p[t+8>>3]=za+ +(ta|0);Xw(I,760216,t)|0;tk(I);N=k[q>>2]|0;r=s;return N|0}k[u>>2]=b;Xw(I,771145,u)|0;tk(I);k[q>>2]=302;N=302;r=s;return N|0}function Mn(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0;e=r;r=r+208|0;f=e+8|0;g=e+115|0;h=e+40|0;j=e+20|0;l=e+16|0;m=e+12|0;n=e;o=k[d>>2]|0;if((o|0)>0){q=o;r=e;return q|0}o=k[a>>2]|0;s=a+4|0;if((o|0)!=(k[(k[s>>2]|0)+68>>2]|0))Fk(a,o+1|0,0,d)|0;if((b|0)>=1?(o=k[s>>2]|0,(k[o+928>>2]|0)>=(b|0)):0){s=k[(k[o+960>>2]|0)+((b+-1|0)*160|0)+80>>2]|0;o=(s|0)>-1?s:0-s|0;Gk(770521,b,h,d)|0;k[c>>2]=0;k[m>>2]=0;a:do if(!(Jo(a,h,j,0,m)|0)){s=j;b:while(1){t=i[s>>0]|0;switch(t<<24>>24){case 103:case 71:case 100:case 68:case 101:case 69:case 102:case 70:case 122:case 90:case 111:case 79:case 105:case 73:case 97:case 65:{u=t;v=s;break b;break}case 32:break;default:{w=14;break a}}s=s+1|0}if(u<<24>>24!=0&((u<<24>>24)+-48|0)>>>0>9){s=v;while(1){t=s+1|0;x=i[t>>0]|0;if(x<<24>>24!=0&((x<<24>>24)+-48|0)>>>0>9)s=t;else{y=t;break}}}else y=v;s=ex(y)|0;k[c>>2]=s;if((o|0)>82)k[c>>2]=(s<<1)+3;else{z=s;w=15}}else w=14;while(0);if((w|0)==14){z=k[c>>2]|0;w=15}c:do if((w|0)==15?(z|0)==0:0){Gk(773003,b,h,d)|0;Jo(a,h,j,0,d)|0;Gk(773339,b,h,d)|0;k[m>>2]=0;y=(Mo(a,h,n,0,m)|0)==0;v=y&+p[n>>3]!=1.0;if((o|0)<22&v){k[c>>2]=14;break}y=(o|0)==41;if(y&v){k[c>>2]=23;break}u=(o|0)==81;if(u&v){k[c>>2]=23;break}Zk(a,l,d)|0;if((k[l>>2]|0)==1){v=j;while(1){s=i[v>>0]|0;if(s<<24>>24!=0&((s<<24>>24)+-48|0)>>>0>9)v=v+1|0;else{A=v;break}}v=ex(A)|0;k[c>>2]=v;break}switch(o|0){case 1:{k[c>>2]=8;break c;break}case 11:{k[c>>2]=4;break c;break}case 21:{k[c>>2]=6;break c;break}default:{if(y){k[c>>2]=11;break c}if(u){k[c>>2]=20;break c}switch(o|0){case 42:{k[c>>2]=14;break c;break}case 82:{k[c>>2]=23;break c;break}case 83:{k[c>>2]=31;break c;break}case 163:{k[c>>2]=49;break c;break}case 14:{k[c>>2]=1;break c;break}case 16:{v=j;while(1){s=i[v>>0]|0;if(s<<24>>24!=0&((s<<24>>24)+-48|0)>>>0>9)v=v+1|0;else{B=v;break}}v=ex(B)|0;k[c>>2]=(v|0)<1?1:v;break c;break}default:break c}}}}while(0);q=k[d>>2]|0;r=e;return q|0}k[f>>2]=b;Xw(g,771145,f)|0;tk(g);k[d>>2]=302;q=302;r=e;return q|0}function Nn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0;m=r;r=r+16|0;n=m;if(!(Cl(a,l)|0)){o=(b|0)<1;On(a,2,o?1:b,o?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,m+1|0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}else{i[n>>0]=g;qt(a,12,c,d,e,f,1,n,h,0,j,l)|0;p=k[l>>2]|0;r=m;return p|0}return 0}function On(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0.0,V=0.0,W=0.0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28920|0;D=t+28916|0;E=t+28912|0;F=t+28908|0;G=t+28904|0;H=t+32|0;I=t+24|0;J=t+16|0;K=t+8|0;M=t;N=t+28956|0;O=t+28976|0;P=t+28936|0;Q=t+40|0;R=k[s>>2]|0;if((g|0)==0&(h|0)==0|(R|0)>0){S=R;r=t;return S|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;nl(a,b,c,d,e,f,g,h,j>>31,x,y,N,E,z,A,I,J,F,H,M,B,K,P,s)|0;R=k[z>>2]|0;if((j|0)==1&(R|0)==14){Fn(a,b,c,d,e,f,g,h,l,m,n,o,q,s)|0;S=k[s>>2]|0;r=t;return S|0}if(!(rx(N,65)|0))T=R;else{if((k[s>>2]|0)==308){k[s>>2]=0;xk()}k[z>>2]=11;k[F>>2]=1;R=k[E>>2]|0;f=H;k[f>>2]=R;k[f+4>>2]=((R|0)<0)<<31>>31;k[E>>2]=1;p[x>>3]=1.0;p[y>>3]=0.0;R=K;k[R>>2]=1234554321;k[R+4>>2]=0;k[A>>2]=28800;T=11}R=k[s>>2]|0;if((R|0)>0){S=R;r=t;return S|0}R=ha(k[F>>2]|0,j)|0;k[F>>2]=R;if((T|0)==16&(k[B>>2]|0)==1?(Nk(N,C,G,D,s)|0,G=k[D>>2]|0,(G|0)>0):0){D=0;U=1.0;while(1){V=U*10.0;D=D+1|0;if((D|0)>=(G|0)){W=V;break}else U=V}}else W=1.0;if(((!((l|0)==1&m<<24>>24==0)?(G=k[z>>2]|0,D=K,C=k[D>>2]|0,T=k[D+4>>2]|0,!(((G|0)%10|0|0)==1&((C|0)==1234554321&(T|0)==0))):0)?(D=ky(C|0,T|0,32768,0)|0,R=L,!((G|0)==21&(R>>>0>0|(R|0)==0&D>>>0>65535))):0)?!((G|0)==11&(T>>>0>0|(T|0)==0&C>>>0>255)):0)X=(G|0)==16&(i[P>>0]|0)==1?0:l;else X=0;l=(j|0)>-1;G=((j|0)<0)<<31>>31;C=0-j|0;T=((C|0)<0)<<31>>31;D=0;R=0;f=g;g=h;h=0;e=0;a:while(1){d=e;c=f;Y=g;Z=h;while(1){_=n+Z|0;$=o+Z|0;aa=c;ba=Y;while(1){if((aa|0)==0&(ba|0)==0){ca=53;break a}da=k[A>>2]|0;ea=((da|0)<0)<<31>>31;fa=(ba|0)<(ea|0)|(ba|0)==(ea|0)&aa>>>0<da>>>0;ea=fa?aa:da;da=my(0,ea|0,32)|0;ea=L;if(l){fa=H;ga=J;ia=k[ga>>2]|0;ja=k[ga+4>>2]|0;ga=ky(k[fa>>2]|0,k[fa+4>>2]|0,-1,-1)|0;fa=jy(ga|0,L|0,ia|0,ja|0)|0;ga=wy(fa|0,L|0,j|0,G|0)|0;fa=ky(ga|0,L|0,1,0)|0;ga=L;ka=(ea|0)<(ga|0)|(ea|0)==(ga|0)&da>>>0<fa>>>0;la=ia;ma=ja;na=ka?da:fa;oa=ka?ea:ga}else{ga=J;ka=k[ga>>2]|0;fa=k[ga+4>>2]|0;ga=wy(ka|0,fa|0,C|0,T|0)|0;ja=ky(ga|0,L|0,1,0)|0;ga=L;ia=(ea|0)<(ga|0)|(ea|0)==(ga|0)&da>>>0<ja>>>0;la=ka;ma=fa;na=ia?da:ja;oa=ia?ea:ga}ga=I;ea=k[ga>>2]|0;ia=k[ga+4>>2]|0;ga=M;ja=yy(k[ga>>2]|0,k[ga+4>>2]|0,D|0,R|0)|0;ga=ky(ja|0,L|0,ea|0,ia|0)|0;ia=L;ea=k[F>>2]|0;ja=(ea|0)/(j|0)|0;da=yy(ja|0,((ja|0)<0)<<31>>31|0,la|0,ma|0)|0;ja=ky(ga|0,ia|0,da|0,L|0)|0;da=L;switch(k[z>>2]|0){case 11:{lu(a,ja,da,na,ea,_,s)|0;Rn(_,na,+p[x>>3],+p[y>>3],X,k[K>>2]&255,m,$,q,_,s)|0;break}case 21:{mu(a,ja,da,na,ea,Q,s)|0;Sn(Q,na,+p[x>>3],+p[y>>3],X,k[K>>2]&65535,m,$,q,_,s)|0;break}case 41:{nu(a,ja,da,na,ea,Q,s)|0;Tn(Q,na,+p[x>>3],+p[y>>3],X,k[K>>2]|0,m,$,q,_,s)|0;break}case 81:{ou(a,ja,da,na,ea,Q,s)|0;ia=K;Un(Q,na,+p[x>>3],+p[y>>3],X,k[ia>>2]|0,k[ia+4>>2]|0,m,$,q,_,s)|0;break}case 42:{pu(a,ja,da,na,ea,Q,s)|0;Vn(Q,na,+p[x>>3],+p[y>>3],X,m,$,q,_,s)|0;break}case 82:{qu(a,ja,da,na,ea,Q,s)|0;Wn(Q,na,+p[x>>3],+p[y>>3],X,m,$,q,_,s)|0;break}case 16:{au(a,ja,da,0,s)|0;da=k[F>>2]|0;ja=k[E>>2]|0;if((da|0)==(ja|0)){ea=ha(da,na)|0;fu(a,ea,((ea|0)<0)<<31>>31,Q,s)|0}else gu(a,ja,na,da-ja|0,Q,s)|0;Xn(Q,na,+p[x>>3],+p[y>>3],k[E>>2]|0,W,X,P,m,$,q,_,s)|0;break}default:{ca=39;break a}}if((k[s>>2]|0)>0){pa=na;qa=Z;ra=d;ca=43;break a}ja=((na|0)<0)<<31>>31;da=jy(aa|0,ba|0,na|0,ja|0)|0;ea=L;if((aa|0)==(na|0)&(ba|0)==(ja|0)){aa=da;ba=ea}else{sa=na;ta=na;ua=ja;va=da;wa=ea;break}}ba=ky(ta|0,ua|0,Z|0,d|0)|0;aa=L;_=J;$=ha(sa,j)|0;ea=ky(k[_>>2]|0,k[_+4>>2]|0,$|0,(($|0)<0)<<31>>31|0)|0;$=L;_=J;k[_>>2]=ea;k[_+4>>2]=$;_=H;da=k[_>>2]|0;ja=k[_+4>>2]|0;if(!(($|0)<(ja|0)|($|0)==(ja|0)&ea>>>0<da>>>0)){xa=ea;ya=$;za=da;Aa=ja;Ba=ba;Ca=aa;Da=va;Ea=wa;ca=49;break}if(($|0)<0){Fa=ea;Ga=$;Ha=da;Ia=ja;Ja=ba;Ka=aa;La=va;Ma=wa;ca=52;break}else{d=aa;c=va;Y=wa;Z=ba}}if((ca|0)==49){ca=0;Z=wy(xa|0,ya|0,za|0,Aa|0)|0;Y=L;c=ky(Z|0,Y|0,D|0,R|0)|0;d=L;ba=yy(Z|0,Y|0,za|0,Aa|0)|0;Y=jy(xa|0,ya|0,ba|0,L|0)|0;Na=Y;Oa=L;Pa=Ba;Qa=Ca;Ra=Da;Sa=Ea;Ta=c;Ua=d}else if((ca|0)==52){ca=0;d=wy(~Fa|0,~Ga|0,Ha|0,Ia|0)|0;c=ky(d|0,L|0,1,0)|0;d=L;Y=jy(D|0,R|0,c|0,d|0)|0;ba=L;Z=yy(c|0,d|0,Ha|0,Ia|0)|0;d=ky(Z|0,L|0,Fa|0,Ga|0)|0;Na=d;Oa=L;Pa=Ja;Qa=Ka;Ra=La;Sa=Ma;Ta=Y;Ua=ba}ba=J;k[ba>>2]=Na;k[ba+4>>2]=Oa;D=Ta;R=Ua;f=Ra;g=Sa;h=Pa;e=Qa}if((ca|0)==39){k[w>>2]=b;k[w+4>>2]=N;Xw(O,760278,w)|0;tk(O);if((k[B>>2]|0)==1){k[s>>2]=311;S=311;r=t;return S|0}else{k[s>>2]=312;S=312;r=t;return S|0}}else if((ca|0)==43){W=+(qa>>>0)+4294967296.0*+(ra|0);U=W+1.0;V=W+ +(pa|0);if((k[B>>2]|0)>0){p[v>>3]=U;p[v+8>>3]=V;k[v+16>>2]=b;Xw(O,760331,v)|0}else{p[u>>3]=U;p[u+8>>3]=V;Xw(O,760395,u)|0}tk(O);S=k[s>>2]|0;r=t;return S|0}else if((ca|0)==53){ca=k[s>>2]|0;if((ca|0)!=-11){S=ca;r=t;return S|0}tk(761034);k[s>>2]=412;S=412;r=t;return S|0}return 0}function Pn(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;On(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,12,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function Qn(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0;n=r;r=r+480|0;o=n+232|0;p=n+224|0;q=n+352|0;s=n+316|0;t=n+280|0;u=n+244|0;v=n+144|0;w=n+72|0;x=n;y=n+240|0;z=n+236|0;A=n+389|0;B=n+390|0;C=n+388|0;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(B,760455,p)|0;tk(B);k[m>>2]=320;D=320;r=n;return D|0}if(Cl(a,m)|0){if((c|0)>0){p=0;do{E=k[e+(p<<2)>>2]|0;F=w+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;E=k[f+(p<<2)>>2]|0;F=x+(p<<3)|0;k[F>>2]=E;k[F+4>>2]=((E|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}i[C>>0]=h;mt(a,12,w,x,g,1,C,j,0,l,m)|0;D=k[m>>2]|0;r=n;return D|0}if((Zk(a,y,m)|0)>0){D=k[m>>2]|0;r=n;return D|0}C=(k[y>>2]|0)==0;if(C){y=(b|0)==0?1:b;G=2;H=1;I=y;J=y}else{G=b;H=k[g+(c<<2)>>2]|0;I=k[f+(c<<2)>>2]|0;J=k[e+(c<<2)>>2]|0}b=(l|0)!=0;if(b)k[l>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;y=v;k[y>>2]=1;k[y+4>>2]=0;k[u>>2]=1;y=q+4|0;k[y>>2]=1;x=s+4|0;k[x>>2]=1;w=t+4|0;k[w>>2]=1;p=v+8|0;E=p;k[E>>2]=1;k[E+4>>2]=0;E=u+4|0;k[E>>2]=1;F=q+8|0;k[F>>2]=1;K=s+8|0;k[K>>2]=1;M=t+8|0;k[M>>2]=1;N=v+16|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=u+8|0;k[O>>2]=1;P=q+12|0;k[P>>2]=1;Q=s+12|0;k[Q>>2]=1;R=t+12|0;k[R>>2]=1;S=v+24|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=u+12|0;k[T>>2]=1;U=q+16|0;k[U>>2]=1;V=s+16|0;k[V>>2]=1;W=t+16|0;k[W>>2]=1;X=v+32|0;Y=X;k[Y>>2]=1;k[Y+4>>2]=0;Y=u+16|0;k[Y>>2]=1;Z=q+20|0;k[Z>>2]=1;_=s+20|0;k[_>>2]=1;$=t+20|0;k[$>>2]=1;aa=v+40|0;ba=aa;k[ba>>2]=1;k[ba+4>>2]=0;ba=u+20|0;k[ba>>2]=1;ca=q+24|0;k[ca>>2]=1;da=s+24|0;k[da>>2]=1;ea=t+24|0;k[ea>>2]=1;fa=v+48|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;ga=u+24|0;k[ga>>2]=1;ia=q+28|0;k[ia>>2]=1;ja=s+28|0;k[ja>>2]=1;ka=t+28|0;k[ka>>2]=1;la=v+56|0;ma=la;k[ma>>2]=1;k[ma+4>>2]=0;ma=u+28|0;k[ma>>2]=1;na=q+32|0;k[na>>2]=1;oa=s+32|0;k[oa>>2]=1;pa=t+32|0;k[pa>>2]=1;qa=v+64|0;ra=qa;k[ra>>2]=1;k[ra+4>>2]=0;ra=u+32|0;k[ra>>2]=1;a:do if((c|0)>0){sa=1;ta=0;ua=0;while(1){va=k[f+(ua<<2)>>2]|0;wa=k[e+(ua<<2)>>2]|0;if((va|0)<(wa|0)){if(!C){xa=ua;break}k[u+(ua<<2)>>2]=-1;ya=-1}else ya=k[u+(ua<<2)>>2]|0;k[q+(ua<<2)>>2]=wa;k[s+(ua<<2)>>2]=va;k[t+(ua<<2)>>2]=k[g+(ua<<2)>>2];va=k[d+(ua<<2)>>2]|0;wa=sa;sa=yy(va|0,((va|0)<0)<<31>>31|0,sa|0,ta|0)|0;va=ta;ta=L;za=ua;ua=ua+1|0;Aa=v+(ua<<3)|0;k[Aa>>2]=sa;k[Aa+4>>2]=ta;Aa=yy(ya|0,((ya|0)<0)<<31>>31|0,wa|0,va|0)|0;va=v+(za<<3)|0;k[va>>2]=Aa;k[va+4>>2]=L;if((ua|0)>=(c|0))break a}k[o>>2]=xa+1;Xw(B,760501,o)|0;tk(B);k[m>>2]=321;D=321;r=n;return D|0}while(0);B=v+(c<<3)|0;v=B;o=k[u+(c<<2)>>2]|0;xa=yy(o|0,((o|0)<0)<<31>>31|0,k[v>>2]|0,k[v+4>>2]|0)|0;v=B;k[v>>2]=xa;k[v+4>>2]=L;if((c|0)==1?(k[d>>2]|0)==1:0){Ba=(I-J|0)/(H|0)|0;Ca=H;Da=J}else{d=k[u>>2]|0;u=ha((k[s>>2]|0)-(k[q>>2]|0)|0,d)|0;Ba=(u|0)/(k[g>>2]|0)|0;Ca=ha(k[t>>2]|0,d)|0;Da=I}I=Ba+1|0;b:do if((J|0)<=(Da|0)){Ba=k[ra>>2]|0;d=ha(Ba,k[na>>2]|0)|0;t=ha(Ba,k[oa>>2]|0)|0;g=(d|0)>(t|0);u=k[ma>>2]|0;s=ha(u,k[ia>>2]|0)|0;c=ha(u,k[ja>>2]|0)|0;v=(s|0)>(c|0);xa=k[pa>>2]|0;B=k[ga>>2]|0;o=ha(B,k[ca>>2]|0)|0;ya=ha(B,k[da>>2]|0)|0;C=(o|0)>(ya|0);e=k[ka>>2]|0;f=k[ba>>2]|0;ua=ha(f,k[Z>>2]|0)|0;ta=ha(f,k[_>>2]|0)|0;sa=(ua|0)>(ta|0);va=k[ea>>2]|0;Aa=k[Y>>2]|0;za=ha(Aa,k[U>>2]|0)|0;wa=ha(Aa,k[V>>2]|0)|0;Ea=(za|0)>(wa|0);Fa=k[$>>2]|0;Ga=k[T>>2]|0;Ha=ha(Ga,k[P>>2]|0)|0;Ia=ha(Ga,k[Q>>2]|0)|0;Ja=(Ha|0)>(Ia|0);Ka=k[W>>2]|0;La=k[O>>2]|0;Ma=ha(La,k[F>>2]|0)|0;Na=ha(La,k[K>>2]|0)|0;Oa=(Ma|0)>(Na|0);Pa=k[R>>2]|0;Qa=k[E>>2]|0;Ra=ha(Qa,k[y>>2]|0)|0;Sa=ha(Qa,k[x>>2]|0)|0;Ta=(Ra|0)>(Sa|0);Ua=k[M>>2]|0;Va=k[q>>2]|0;Wa=((Va|0)<0)<<31>>31;Xa=p;Ya=k[Xa>>2]|0;Za=k[Xa+4>>2]|0;Xa=N;_a=k[Xa>>2]|0;$a=k[Xa+4>>2]|0;Xa=S;ab=k[Xa>>2]|0;bb=k[Xa+4>>2]|0;Xa=X;cb=k[Xa>>2]|0;db=k[Xa+4>>2]|0;Xa=aa;eb=k[Xa>>2]|0;fb=k[Xa+4>>2]|0;Xa=fa;gb=k[Xa>>2]|0;hb=k[Xa+4>>2]|0;Xa=la;ib=k[Xa>>2]|0;jb=k[Xa+4>>2]|0;Xa=qa;kb=k[Xa>>2]|0;lb=k[Xa+4>>2]|0;Xa=((I|0)<0)<<31>>31;mb=k[w>>2]|0;nb=0;ob=J;c:while(1){if(g)pb=nb;else{qb=((ob|0)<0)<<31>>31;rb=nb;sb=d;while(1){if(v)tb=rb;else{ub=sb-Ba|0;vb=yy(ub|0,((ub|0)<0)<<31>>31|0,kb|0,lb|0)|0;ub=L;wb=rb;xb=s;while(1){if(C)yb=wb;else{zb=xb-u|0;Ab=yy(zb|0,((zb|0)<0)<<31>>31|0,ib|0,jb|0)|0;zb=L;Bb=wb;Cb=o;while(1){if(sa)Db=Bb;else{Eb=Cb-B|0;Fb=yy(Eb|0,((Eb|0)<0)<<31>>31|0,gb|0,hb|0)|0;Eb=L;Gb=Bb;Hb=ua;while(1){if(Ea)Ib=Gb;else{Jb=Hb-f|0;Kb=yy(Jb|0,((Jb|0)<0)<<31>>31|0,eb|0,fb|0)|0;Jb=L;Lb=Gb;Mb=za;while(1){if(Ja)Nb=Lb;else{Ob=Mb-Aa|0;Pb=yy(Ob|0,((Ob|0)<0)<<31>>31|0,cb|0,db|0)|0;Ob=L;Qb=Lb;Rb=Ha;while(1){if(Oa)Sb=Qb;else{Tb=Rb-Ga|0;Ub=yy(Tb|0,((Tb|0)<0)<<31>>31|0,ab|0,bb|0)|0;Tb=L;Vb=Qb;Wb=Ma;while(1){if(Ta)Xb=Vb;else{Yb=Wb-La|0;Zb=yy(Yb|0,((Yb|0)<0)<<31>>31|0,_a|0,$a|0)|0;Yb=L;_b=Vb;$b=Ra;while(1){ac=$b-Qa|0;bc=yy(Ya|0,Za|0,ac|0,((ac|0)<0)<<31>>31|0)|0;ac=ky(bc|0,L|0,Va|0,Wa|0)|0;bc=ky(ac|0,L|0,Zb|0,Yb|0)|0;ac=ky(bc|0,L|0,Ub|0,Tb|0)|0;bc=ky(ac|0,L|0,Pb|0,Ob|0)|0;ac=ky(bc|0,L|0,Kb|0,Jb|0)|0;bc=ky(ac|0,L|0,Fb|0,Eb|0)|0;ac=ky(bc|0,L|0,Ab|0,zb|0)|0;bc=ky(ac|0,L|0,vb|0,ub|0)|0;if((On(a,G,ob,qb,bc,L,I,Xa,Ca,1,h,j+_b|0,A,z,m)|0)>0)break c;if(b&(k[z>>2]|0)!=0)k[l>>2]=1;bc=_b+I|0;$b=mb+$b|0;if(($b|0)>(Sa|0)){Xb=bc;break}else _b=bc}}Wb=Ua+Wb|0;if((Wb|0)>(Na|0)){Sb=Xb;break}else Vb=Xb}}Rb=Pa+Rb|0;if((Rb|0)>(Ia|0)){Nb=Sb;break}else Qb=Sb}}Mb=Ka+Mb|0;if((Mb|0)>(wa|0)){Ib=Nb;break}else Lb=Nb}}Hb=Fa+Hb|0;if((Hb|0)>(ta|0)){Db=Ib;break}else Gb=Ib}}Cb=va+Cb|0;if((Cb|0)>(ya|0)){yb=Db;break}else Bb=Db}}xb=e+xb|0;if((xb|0)>(c|0)){tb=yb;break}else wb=yb}}sb=xa+sb|0;if((sb|0)>(t|0)){pb=tb;break}else rb=tb}}ob=ob+H|0;if((ob|0)>(Da|0))break b;else nb=pb}D=k[m>>2]|0;r=n;return D|0}while(0);D=k[m>>2]|0;r=n;return D|0}function Rn(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;o=c==1.0;p=o&d==-128.0;if(!e){if(p){if((b|0)>0)q=0;else{r=k[n>>2]|0;return r|0}do{i[m+q>>0]=(l[a+q>>0]|0)^128;q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}q=(b|0)>0;if(o&d==0.0){if(q)s=0;else{r=k[n>>2]|0;return r|0}do{t=i[a+s>>0]|0;if(t<<24>>24<0){k[n>>2]=-11;u=127}else u=t;i[m+s>>0]=u;s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(q)v=0;else{r=k[n>>2]|0;return r|0}do{w=+(l[a+v>>0]|0|0)*c+d;do if(!(w<-128.49))if(w>127.49){k[n>>2]=-11;i[m+v>>0]=127;break}else{i[m+v>>0]=~~w;break}else{k[n>>2]=-11;i[m+v>>0]=-128}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p){if((b|0)<=0){r=k[n>>2]|0;return r|0}if((e|0)==1){p=0;do{v=i[a+p>>0]|0;if(v<<24>>24==f<<24>>24){k[j>>2]=1;i[m+p>>0]=g}else i[m+p>>0]=v&255^128;p=p+1|0}while((p|0)!=(b|0));r=k[n>>2]|0;return r|0}else{p=0;do{v=i[a+p>>0]|0;if(v<<24>>24==f<<24>>24){k[j>>2]=1;i[h+p>>0]=1}else i[m+p>>0]=v&255^128;p=p+1|0}while((p|0)!=(b|0));r=k[n>>2]|0;return r|0}}p=(b|0)>0;if(o&d==0.0){if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){o=0;do{v=i[a+o>>0]|0;if(v<<24>>24==f<<24>>24){k[j>>2]=1;i[m+o>>0]=g}else i[m+o>>0]=v;o=o+1|0}while((o|0)!=(b|0));r=k[n>>2]|0;return r|0}else{o=0;do{v=i[a+o>>0]|0;if(v<<24>>24==f<<24>>24){k[j>>2]=1;i[h+o>>0]=1}else i[m+o>>0]=v;o=o+1|0}while((o|0)!=(b|0));r=k[n>>2]|0;return r|0}}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{o=i[a+e>>0]|0;do if(o<<24>>24==f<<24>>24){k[j>>2]=1;if(p){i[m+e>>0]=g;break}else{i[h+e>>0]=1;break}}else{w=+(o&255|0)*c+d;if(w<-128.49){k[n>>2]=-11;i[m+e>>0]=-128;break}if(w>127.49){k[n>>2]=-11;i[m+e>>0]=127;break}else{i[m+e>>0]=~~w;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function Sn(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=j[a+(q<<1)>>1]|0;do if(s<<16>>16>=-128)if(s<<16>>16>127){k[n>>2]=-11;i[m+q>>0]=127;break}else{i[m+q>>0]=s;break}else{k[n>>2]=-11;i[m+q>>0]=-128}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)t=0;else{r=k[n>>2]|0;return r|0}do{u=+(j[a+(t<<1)>>1]|0)*c+d;do if(!(u<-128.49))if(u>127.49){k[n>>2]=-11;i[m+t>>0]=127;break}else{i[m+t>>0]=~~u;break}else{k[n>>2]=-11;i[m+t>>0]=-128}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;t=0;do{q=j[a+(t<<1)>>1]|0;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){i[m+t>>0]=g;break}else{i[h+t>>0]=1;break}}else{u=+(q<<16>>16)*c+d;if(u<-128.49){k[n>>2]=-11;i[m+t>>0]=-128;break}if(u>127.49){k[n>>2]=-11;i[m+t>>0]=127;break}else{i[m+t>>0]=~~u;break}}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){e=0;do{p=j[a+(e<<1)>>1]|0;do if(p<<16>>16!=f<<16>>16){if(p<<16>>16<-128){k[n>>2]=-11;i[m+e>>0]=-128;break}if(p<<16>>16>127){k[n>>2]=-11;i[m+e>>0]=127;break}else{i[m+e>>0]=p;break}}else{k[l>>2]=1;i[m+e>>0]=g}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}else{e=0;do{g=j[a+(e<<1)>>1]|0;do if(g<<16>>16!=f<<16>>16){if(g<<16>>16<-128){k[n>>2]=-11;i[m+e>>0]=-128;break}if(g<<16>>16>127){k[n>>2]=-11;i[m+e>>0]=127;break}else{i[m+e>>0]=g;break}}else{k[l>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function Tn(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0.0;n=c==1.0&d==0.0;o=(b|0)>0;if(!e)if(n){if(o)p=0;else{q=k[m>>2]|0;return q|0}do{r=k[a+(p<<2)>>2]|0;do if((r|0)>=-128)if((r|0)>127){k[m>>2]=-11;i[l+p>>0]=127;break}else{i[l+p>>0]=r;break}else{k[m>>2]=-11;i[l+p>>0]=-128}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}else{if(o)s=0;else{q=k[m>>2]|0;return q|0}do{t=+(k[a+(s<<2)>>2]|0)*c+d;do if(!(t<-128.49))if(t>127.49){k[m>>2]=-11;i[l+s>>0]=127;break}else{i[l+s>>0]=~~t;break}else{k[m>>2]=-11;i[l+s>>0]=-128}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[m>>2]|0;return q|0}if(n){if(!o){q=k[m>>2]|0;return q|0}n=(e|0)==1;s=0;do{p=k[a+(s<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(n){i[l+s>>0]=g;break}else{i[h+s>>0]=1;break}}else{if((p|0)<-128){k[m>>2]=-11;i[l+s>>0]=-128;break}if((p|0)>127){k[m>>2]=-11;i[l+s>>0]=127;break}else{i[l+s>>0]=p;break}}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[m>>2]|0;return q|0}else{if(!o){q=k[m>>2]|0;return q|0}o=(e|0)==1;e=0;do{s=k[a+(e<<2)>>2]|0;do if((s|0)==(f|0)){k[j>>2]=1;if(o){i[l+e>>0]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s|0)*c+d;if(t<-128.49){k[m>>2]=-11;i[l+e>>0]=-128;break}if(t>127.49){k[m>>2]=-11;i[l+e>>0]=127;break}else{i[l+e>>0]=~~t;break}}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}return 0}function Un(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=a+(q<<3)|0;t=k[s>>2]|0;u=k[s+4>>2]|0;do if(!((u|0)<-1|(u|0)==-1&t>>>0<4294967168))if((u|0)>0|(u|0)==0&t>>>0>127){k[n>>2]=-11;i[m+q>>0]=127;break}else{i[m+q>>0]=t;break}else{k[n>>2]=-11;i[m+q>>0]=-128}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)v=0;else{r=k[n>>2]|0;return r|0}do{q=a+(v<<3)|0;w=(+((k[q>>2]|0)>>>0)+4294967296.0*+(k[q+4>>2]|0))*c+d;do if(!(w<-128.49))if(w>127.49){k[n>>2]=-11;i[m+v>>0]=127;break}else{i[m+v>>0]=~~w;break}else{k[n>>2]=-11;i[m+v>>0]=-128}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;v=0;do{q=a+(v<<3)|0;t=k[q>>2]|0;u=k[q+4>>2]|0;do if((t|0)==(f|0)&(u|0)==(g|0)){k[l>>2]=1;if(o){i[m+v>>0]=h;break}else{i[j+v>>0]=1;break}}else{if((u|0)<-1|(u|0)==-1&t>>>0<4294967168){k[n>>2]=-11;i[m+v>>0]=-128;break}if((u|0)>0|(u|0)==0&t>>>0>127){k[n>>2]=-11;i[m+v>>0]=127;break}else{i[m+v>>0]=t;break}}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{v=a+(e<<3)|0;o=k[v>>2]|0;t=k[v+4>>2]|0;do if((o|0)==(f|0)&(t|0)==(g|0)){k[l>>2]=1;if(p){i[m+e>>0]=h;break}else{i[j+e>>0]=1;break}}else{w=(+(o>>>0)+4294967296.0*+(t|0))*c+d;if(w<-128.49){k[n>>2]=-11;i[m+e>>0]=-128;break}if(w>127.49){k[n>>2]=-11;i[m+e>>0]=127;break}else{i[m+e>>0]=~~w;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function Vn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-128.49))if(s>127.49){k[l>>2]=-11;i[j+p>>0]=127;break}else{i[j+p>>0]=~~r;break}else{k[l>>2]=-11;i[j+p>>0]=-128}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-128.49))if(r>127.49){k[l>>2]=-11;i[j+t>>0]=127;break}else{i[j+t>>0]=~~r;break}else{k[l>>2]=-11;i[j+t>>0]=-128}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-128.49){k[l>>2]=-11;i[j+u>>0]=-128;break a}if(s>127.49){k[l>>2]=-11;i[j+u>>0]=127;break a}else{i[j+u>>0]=~~r;break a}break}case 1:{k[h>>2]=1;if(p){i[j+u>>0]=f;break a}else{i[g+u>>0]=1;break a}break}default:i[j+u>>0]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-128.49;v=d>127.49;u=~~d;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-128.49){k[l>>2]=-11;i[j+e>>0]=-128;break b}if(r>127.49){k[l>>2]=-11;i[j+e>>0]=127;break b}else{i[j+e>>0]=~~r;break b}break}case 1:{k[h>>2]=1;if(p){i[j+e>>0]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;i[j+e>>0]=-128;break b}if(v){k[l>>2]=-11;i[j+e>>0]=127;break b}else{i[j+e>>0]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function Wn(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-128.49))if(r>127.49){k[l>>2]=-11;i[j+o>>0]=127;break}else{i[j+o>>0]=~~r;break}else{k[l>>2]=-11;i[j+o>>0]=-128}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-128.49))if(r>127.49){k[l>>2]=-11;i[j+s>>0]=127;break}else{i[j+s>>0]=~~r;break}else{k[l>>2]=-11;i[j+s>>0]=-128}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-128.49){k[l>>2]=-11;i[j+t>>0]=-128;break a}if(r>127.49){k[l>>2]=-11;i[j+t>>0]=127;break a}else{i[j+t>>0]=~~r;break a}break}case 1:{k[h>>2]=1;if(o){i[j+t>>0]=f;break a}else{i[g+t>>0]=1;break a}break}default:i[j+t>>0]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-128.49;u=d>127.49;t=~~d;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-128.49){k[l>>2]=-11;i[j+e>>0]=-128;break b}if(r>127.49){k[l>>2]=-11;i[j+e>>0]=127;break b}else{i[j+e>>0]=~~r;break b}break}case 1:{k[h>>2]=1;if(o){i[j+e>>0]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;i[j+e>>0]=-128;break b}if(u){k[l>>2]=-11;i[j+e>>0]=127;break b}else{i[j+e>>0]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function Xn(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0.0,P=0.0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0.0,Z=0,_=0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0,ga=0.0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0;p=r;r=r+96|0;q=p;s=p+4|0;t=xx(h)|0;a:do if((b|0)>0){u=(g|0)==0;v=(g|0)==1;w=a;x=0;b:while(1){y=w+e|0;z=i[y>>0]|0;i[y>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,w,t)|0)==0:0)if(!u){k[m>>2]=1;if(v){i[n+x>>0]=j;A=y;break}else{i[l+x>>0]=1;A=y;break}}else A=y;else{B=w;C=9}while(0);do if((C|0)==9){c:while(1){C=0;D=i[B>>0]|0;switch(D<<24>>24){case 43:case 45:{E=D;F=B;C=11;break c;break}case 32:break;default:{G=D;H=B;I=1;break c}}B=B+1|0;C=9}if((C|0)==11){C=0;D=E<<24>>24==45;J=F;while(1){K=J+1|0;L=i[K>>0]|0;if(L<<24>>24==32)J=K;else{M=L;N=K;break}}G=M;H=N;I=D?-1:1}if((G+-48&255)<10){J=G;K=H;O=0.0;while(1){P=O*10.0+ +(J<<24>>24);L=K;while(1){Q=L+1|0;R=i[Q>>0]|0;if(R<<24>>24==32)L=Q;else{S=R;T=Q;break}}U=P+-48.0;if((S+-48&255)<10){J=S;K=T;O=U}else{V=S;W=T;X=U;break}}}else{V=G;W=H;X=0.0}switch(V<<24>>24){case 44:case 46:{K=W;while(1){J=K+1|0;D=i[J>>0]|0;if(D<<24>>24==32)K=J;else{Z=D;_=J;break}}if((Z+-48&255)<10){K=Z;J=_;O=1.0;U=X;while(1){$=U*10.0+ +(K<<24>>24)+-48.0;D=J;while(1){L=D+1|0;Q=i[L>>0]|0;if(Q<<24>>24==32)D=L;else{aa=Q;ba=L;break}}P=O*10.0;if((aa+-48&255)<10){K=aa;J=ba;O=P;U=$}else{ca=aa;da=ba;ea=1;fa=P;ga=$;break}}}else{ca=Z;da=_;ea=1;fa=1.0;ga=X}break}default:{ca=V;da=W;ea=0;fa=1.0;ga=X}}if((ca&-2)<<24>>24==68){J=da;d:while(1){K=J+1|0;D=i[K>>0]|0;switch(D<<24>>24){case 32:{J=K;break}case 43:case 45:{ia=D;ja=J;C=28;break d;break}default:{ka=D;la=K;ma=1;break d}}}if((C|0)==28){C=0;J=ia<<24>>24==45;K=ja+2|0;while(1){D=i[K>>0]|0;if(D<<24>>24==32)K=K+1|0;else{na=D;oa=K;break}}ka=na;la=oa;ma=J?-1:1}if((ka+-48&255)<10){K=ka;D=la;L=0;while(1){Q=(L*10|0)+-48|0;R=D;while(1){pa=R+1|0;qa=i[pa>>0]|0;if(qa<<24>>24==32)R=pa;else{ra=qa;sa=pa;break}}R=Q+(K<<24>>24)|0;if((ra+-48&255)<10){K=ra;D=sa;L=R}else{ta=ra;ua=sa;va=ma;wa=R;break}}}else{ta=ka;ua=la;va=ma;wa=0}}else{ta=ca;ua=da;va=1;wa=0}if(ta<<24>>24){xa=y;ya=z;za=w;break b}$=+(I|0)*ga/((ea|0)==0?f:fa)*+Y(10.0,+(+(ha(va,wa)|0)))*c+d;if($<-128.49){k[o>>2]=-11;i[n+x>>0]=-128;A=ua;break}if($>127.49){k[o>>2]=-11;i[n+x>>0]=127;A=ua;break}else{i[n+x>>0]=~~$;A=ua;break}}while(0);i[y>>0]=z;x=x+1|0;if((x|0)>=(b|0))break a;else w=A}w=s;x=761193;v=w+36|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(v|0));tk(s);k[q>>2]=za;Xw(s,761229,q)|0;tk(s);i[xa>>0]=ya;k[o>>2]=409;Aa=409;r=p;return Aa|0}while(0);Aa=k[o>>2]|0;r=p;return Aa|0}function Yn(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0;l=r;r=r+16|0;m=l;if(Cl(a,j)|0){k[m>>2]=g;qt(a,30,c,d,e,f,1,m,h,0,i,j)|0;n=k[j>>2]|0;r=l;return n|0}if((e|0)==0&(f|0)==0|(k[j>>2]|0)>0){n=k[j>>2]|0;r=l;return n|0}m=(b|0)<1;so(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+4|0,i,j)|0;n=k[j>>2]|0;r=l;return n|0}function Zn(a,b,c,d,e,f,g,h,i,j,l,m,n,o,p){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,r=0;q=k[p>>2]|0;if((g|0)==0&(h|0)==0|(q|0)>0){r=q;return r|0}so(a,b,c,d,e,f,g,h,i,j,l,m,n,o,p)|0;r=k[p>>2]|0;return r|0}function _n(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(Cl(a,j)|0){qt(a,30,c,d,e,f,2,0,g,h,i,j)|0;l=k[j>>2]|0;return l|0}if((e|0)==0&(f|0)==0|(k[j>>2]|0)>0){l=k[j>>2]|0;return l|0}m=(b|0)<1;so(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;l=k[j>>2]|0;return l|0}function $n(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0;m=r;r=r+448|0;n=m+232|0;o=m+224|0;p=m+320|0;q=m+284|0;s=m+248|0;t=m+144|0;u=m+72|0;v=m;w=m+244|0;x=m+240|0;y=m+356|0;z=m+357|0;A=m+236|0;if((c+-1|0)>>>0>8){k[o>>2]=c;Xw(z,760547,o)|0;tk(z);k[l>>2]=320;B=320;r=m;return B|0}if(Cl(a,l)|0){if((c|0)>0){o=0;do{C=k[e+(o<<2)>>2]|0;D=u+(o<<3)|0;k[D>>2]=C;k[D+4>>2]=((C|0)<0)<<31>>31;C=k[f+(o<<2)>>2]|0;D=v+(o<<3)|0;k[D>>2]=C;k[D+4>>2]=((C|0)<0)<<31>>31;o=o+1|0}while((o|0)!=(c|0))}k[A>>2]=h;mt(a,30,u,v,g,1,A,i,0,j,l)|0;B=k[l>>2]|0;r=m;return B|0}if((Zk(a,w,l)|0)>0){B=k[l>>2]|0;r=m;return B|0}if(!(k[w>>2]|0)){w=(b|0)==0?1:b;E=2;F=1;G=w;H=w}else{E=b;F=k[g+(c<<2)>>2]|0;G=k[f+(c<<2)>>2]|0;H=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[p>>2]=1;k[q>>2]=1;k[s>>2]=1;w=t;k[w>>2]=1;k[w+4>>2]=0;w=p+4|0;k[w>>2]=1;A=q+4|0;k[A>>2]=1;v=s+4|0;k[v>>2]=1;u=t+8|0;o=u;k[o>>2]=1;k[o+4>>2]=0;o=p+8|0;k[o>>2]=1;C=q+8|0;k[C>>2]=1;D=s+8|0;k[D>>2]=1;I=t+16|0;J=I;k[J>>2]=1;k[J+4>>2]=0;J=p+12|0;k[J>>2]=1;K=q+12|0;k[K>>2]=1;M=s+12|0;k[M>>2]=1;N=t+24|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=p+16|0;k[O>>2]=1;P=q+16|0;k[P>>2]=1;Q=s+16|0;k[Q>>2]=1;R=t+32|0;S=R;k[S>>2]=1;k[S+4>>2]=0;S=p+20|0;k[S>>2]=1;T=q+20|0;k[T>>2]=1;U=s+20|0;k[U>>2]=1;V=t+40|0;W=V;k[W>>2]=1;k[W+4>>2]=0;W=p+24|0;k[W>>2]=1;X=q+24|0;k[X>>2]=1;Y=s+24|0;k[Y>>2]=1;Z=t+48|0;_=Z;k[_>>2]=1;k[_+4>>2]=0;_=p+28|0;k[_>>2]=1;$=q+28|0;k[$>>2]=1;aa=s+28|0;k[aa>>2]=1;ba=t+56|0;ca=ba;k[ca>>2]=1;k[ca+4>>2]=0;ca=p+32|0;k[ca>>2]=1;da=q+32|0;k[da>>2]=1;ea=s+32|0;k[ea>>2]=1;fa=t+64|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;if((c|0)>0){ga=1;ha=0;ia=0;do{ja=k[f+(ia<<2)>>2]|0;ka=k[e+(ia<<2)>>2]|0;if((ja|0)<(ka|0)){la=ia;ma=16;break}k[p+(ia<<2)>>2]=ka;k[q+(ia<<2)>>2]=ja;k[s+(ia<<2)>>2]=k[g+(ia<<2)>>2];ja=k[d+(ia<<2)>>2]|0;ga=yy(ja|0,((ja|0)<0)<<31>>31|0,ga|0,ha|0)|0;ha=L;ia=ia+1|0;ja=t+(ia<<3)|0;k[ja>>2]=ga;k[ja+4>>2]=ha}while((ia|0)<(c|0));if((ma|0)==16){k[n>>2]=la+1;Xw(z,760593,n)|0;tk(z);k[l>>2]=321;B=321;r=m;return B|0}if((c|0)==1?(k[d>>2]|0)==1:0){na=(G-H|0)/(F|0)|0;oa=F;pa=H}else ma=21}else ma=21;if((ma|0)==21){na=((k[q>>2]|0)-(k[p>>2]|0)|0)/(k[g>>2]|0)|0;oa=k[s>>2]|0;pa=G}G=na+1|0;a:do if((H|0)<=(pa|0)){na=k[ca>>2]|0;s=k[da>>2]|0;g=(na|0)>(s|0);q=k[_>>2]|0;ma=k[$>>2]|0;d=(q|0)>(ma|0);c=k[ea>>2]|0;z=k[W>>2]|0;n=k[X>>2]|0;la=(z|0)>(n|0);ia=k[aa>>2]|0;ha=k[S>>2]|0;ga=k[T>>2]|0;t=(ha|0)>(ga|0);e=k[Y>>2]|0;f=k[O>>2]|0;ja=k[P>>2]|0;ka=(f|0)>(ja|0);qa=k[U>>2]|0;ra=k[J>>2]|0;sa=k[K>>2]|0;ta=(ra|0)>(sa|0);ua=k[Q>>2]|0;va=k[o>>2]|0;wa=k[C>>2]|0;xa=(va|0)>(wa|0);ya=k[M>>2]|0;za=k[w>>2]|0;Aa=k[A>>2]|0;Ba=(za|0)>(Aa|0);Ca=k[D>>2]|0;Da=(G|0)==0;Ea=k[v>>2]|0;Fa=((G|0)<0)<<31>>31;Ga=k[p>>2]|0;Ha=((Ga|0)<0)<<31>>31;Ia=u;Ja=k[Ia>>2]|0;Ka=k[Ia+4>>2]|0;Ia=I;La=k[Ia>>2]|0;Ma=k[Ia+4>>2]|0;Ia=N;Na=k[Ia>>2]|0;Oa=k[Ia+4>>2]|0;Ia=R;Pa=k[Ia>>2]|0;Qa=k[Ia+4>>2]|0;Ia=V;Ra=k[Ia>>2]|0;Sa=k[Ia+4>>2]|0;Ia=Z;Ta=k[Ia>>2]|0;Ua=k[Ia+4>>2]|0;Ia=ba;Va=k[Ia>>2]|0;Wa=k[Ia+4>>2]|0;Ia=fa;Xa=k[Ia>>2]|0;Ya=k[Ia+4>>2]|0;Ia=0;Za=H;b:while(1){if(g)_a=Ia;else{$a=((Za|0)<0)<<31>>31;ab=Ia;bb=na;while(1){if(d)cb=ab;else{db=bb+-1|0;eb=yy(Xa|0,Ya|0,db|0,((db|0)<0)<<31>>31|0)|0;db=L;fb=ab;gb=q;while(1){if(la)hb=fb;else{ib=gb+-1|0;jb=yy(Va|0,Wa|0,ib|0,((ib|0)<0)<<31>>31|0)|0;ib=L;kb=fb;lb=z;while(1){if(t)mb=kb;else{nb=lb+-1|0;ob=yy(Ta|0,Ua|0,nb|0,((nb|0)<0)<<31>>31|0)|0;nb=L;pb=kb;qb=ha;while(1){if(ka)rb=pb;else{sb=qb+-1|0;tb=yy(Ra|0,Sa|0,sb|0,((sb|0)<0)<<31>>31|0)|0;sb=L;ub=pb;vb=f;while(1){if(ta)wb=ub;else{xb=vb+-1|0;yb=yy(Pa|0,Qa|0,xb|0,((xb|0)<0)<<31>>31|0)|0;xb=L;zb=ub;Ab=ra;while(1){if(xa)Bb=zb;else{Cb=Ab+-1|0;Db=yy(Na|0,Oa|0,Cb|0,((Cb|0)<0)<<31>>31|0)|0;Cb=L;Eb=zb;Fb=va;while(1){if(Ba)Gb=Eb;else{Hb=Fb+-1|0;Ib=yy(La|0,Ma|0,Hb|0,((Hb|0)<0)<<31>>31|0)|0;Hb=L;Jb=Eb;Kb=za;while(1){Lb=k[l>>2]|0;if(Da|(Lb|0)>0)Mb=Lb;else{Lb=Kb+-1|0;Nb=yy(Ja|0,Ka|0,Lb|0,((Lb|0)<0)<<31>>31|0)|0;Lb=ky(Nb|0,L|0,Ga|0,Ha|0)|0;Nb=ky(Lb|0,L|0,Ib|0,Hb|0)|0;Lb=ky(Nb|0,L|0,Db|0,Cb|0)|0;Nb=ky(Lb|0,L|0,yb|0,xb|0)|0;Lb=ky(Nb|0,L|0,tb|0,sb|0)|0;Nb=ky(Lb|0,L|0,ob|0,nb|0)|0;Lb=ky(Nb|0,L|0,jb|0,ib|0)|0;Nb=ky(Lb|0,L|0,eb|0,db|0)|0;so(a,E,Za,$a,Nb,L,G,Fa,oa,1,h,i+(Jb<<2)|0,y,x,l)|0;Mb=k[l>>2]|0}if((Mb|0)>0){B=Mb;break b}if(b&(k[x>>2]|0)!=0)k[j>>2]=1;Nb=Jb+G|0;Kb=Ea+Kb|0;if((Kb|0)>(Aa|0)){Gb=Nb;break}else Jb=Nb}}Fb=Ca+Fb|0;if((Fb|0)>(wa|0)){Bb=Gb;break}else Eb=Gb}}Ab=ya+Ab|0;if((Ab|0)>(sa|0)){wb=Bb;break}else zb=Bb}}vb=ua+vb|0;if((vb|0)>(ja|0)){rb=wb;break}else ub=wb}}qb=qa+qb|0;if((qb|0)>(ga|0)){mb=rb;break}else pb=rb}}lb=e+lb|0;if((lb|0)>(n|0)){hb=mb;break}else kb=mb}}gb=ia+gb|0;if((gb|0)>(ma|0)){cb=hb;break}else fb=hb}}bb=c+bb|0;if((bb|0)>(s|0)){_a=cb;break}else ab=cb}}Za=Za+F|0;if((Za|0)>(pa|0))break a;else Ia=_a}r=m;return B|0}while(0);B=k[l>>2]|0;r=m;return B|0}function ao(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=l[a+q>>0];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(l[a+s>>0]|0|0)*c+d;do if(!(t<-.49))if(t>4294967295.49){k[n>>2]=-11;k[m+(s<<2)>>2]=-1;break}else{k[m+(s<<2)>>2]=~~t>>>0;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=0}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;s=0;do{q=i[a+s>>0]|0;do if(q<<24>>24==f<<24>>24){k[j>>2]=1;if(o){k[m+(s<<2)>>2]=g;break}else{i[h+s>>0]=1;break}}else k[m+(s<<2)>>2]=q&255;while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{s=i[a+e>>0]|0;do if(s<<24>>24==f<<24>>24){k[j>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(s&255|0)*c+d;if(t<-.49){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if(t>4294967295.49){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=~~t>>>0;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function bo(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=j[a+(q<<1)>>1]|0;if(s<<16>>16<0){k[n>>2]=-11;t=0}else t=s<<16>>16;k[m+(q<<2)>>2]=t;q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)u=0;else{r=k[n>>2]|0;return r|0}do{v=+(j[a+(u<<1)>>1]|0)*c+d;do if(!(v<-.49))if(v>4294967295.49){k[n>>2]=-11;k[m+(u<<2)>>2]=-1;break}else{k[m+(u<<2)>>2]=~~v>>>0;break}else{k[n>>2]=-11;k[m+(u<<2)>>2]=0}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;u=0;do{q=j[a+(u<<1)>>1]|0;t=q<<16>>16;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){k[m+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else if(q<<16>>16<0){k[n>>2]=-11;k[m+(u<<2)>>2]=0;break}else{k[m+(u<<2)>>2]=t;break}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{u=j[a+(e<<1)>>1]|0;do if(u<<16>>16==f<<16>>16){k[l>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{v=+(u<<16>>16)*c+d;if(v<-.49){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if(v>4294967295.49){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=~~v>>>0;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function co(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;n=c==1.0;o=n&d==2147483648.0;if(!e){if(o){if((b|0)>0)p=0;else{q=k[m>>2]|0;return q|0}do{k[l+(p<<2)>>2]=k[a+(p<<2)>>2]^-2147483648;p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}p=(b|0)>0;if(n&d==0.0){if(p)r=0;else{q=k[m>>2]|0;return q|0}do{s=k[a+(r<<2)>>2]|0;if((s|0)<0){k[m>>2]=-11;t=0}else t=s;k[l+(r<<2)>>2]=t;r=r+1|0}while((r|0)!=(b|0));q=k[m>>2]|0;return q|0}if(p)u=0;else{q=k[m>>2]|0;return q|0}do{v=+(k[a+(u<<2)>>2]|0)*c+d;do if(!(v<-.49))if(v>4294967295.49){k[m>>2]=-11;k[l+(u<<2)>>2]=-1;break}else{k[l+(u<<2)>>2]=~~v>>>0;break}else{k[m>>2]=-11;k[l+(u<<2)>>2]=0}while(0);u=u+1|0}while((u|0)!=(b|0));q=k[m>>2]|0;return q|0}if(o){if((b|0)<=0){q=k[m>>2]|0;return q|0}o=(e|0)==1;u=0;do{p=k[a+(u<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(o){k[l+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else k[l+(u<<2)>>2]=p^-2147483648;while(0);u=u+1|0}while((u|0)!=(b|0));q=k[m>>2]|0;return q|0}u=(b|0)>0;if(n&d==0.0){if(!u){q=k[m>>2]|0;return q|0}n=(e|0)==1;o=0;do{p=k[a+(o<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(n){k[l+(o<<2)>>2]=g;break}else{i[h+o>>0]=1;break}}else if((p|0)<0){k[m>>2]=-11;k[l+(o<<2)>>2]=0;break}else{k[l+(o<<2)>>2]=p;break}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[m>>2]|0;return q|0}if(!u){q=k[m>>2]|0;return q|0}u=(e|0)==1;e=0;do{o=k[a+(e<<2)>>2]|0;do if((o|0)==(f|0)){k[j>>2]=1;if(u){k[l+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{v=+(o|0)*c+d;if(v<-.49){k[m>>2]=-11;k[l+(e<<2)>>2]=0;break}if(v>4294967295.49){k[m>>2]=-11;k[l+(e<<2)>>2]=-1;break}else{k[l+(e<<2)>>2]=~~v>>>0;break}}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}function eo(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-.49))if(s>4294967295.49){k[l>>2]=-11;k[j+(p<<2)>>2]=-1;break}else{k[j+(p<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(p<<2)>>2]=0}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-.49))if(r>4294967295.49){k[l>>2]=-11;k[j+(t<<2)>>2]=-1;break}else{k[j+(t<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(t<<2)>>2]=0}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-.49){k[l>>2]=-11;k[j+(u<<2)>>2]=0;break a}if(s>4294967295.49){k[l>>2]=-11;k[j+(u<<2)>>2]=-1;break a}else{k[j+(u<<2)>>2]=~~r>>>0;break a}break}case 1:{k[h>>2]=1;if(p){k[j+(u<<2)>>2]=f;break a}else{i[g+u>>0]=1;break a}break}default:k[j+(u<<2)>>2]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;v=d>4294967295.49;u=~~d>>>0;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-.49){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(r>4294967295.49){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=~~r>>>0;break b}break}case 1:{k[h>>2]=1;if(p){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(v){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function fo(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-.49))if(r>4294967295.49){k[l>>2]=-11;k[j+(o<<2)>>2]=-1;break}else{k[j+(o<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(o<<2)>>2]=0}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-.49))if(r>4294967295.49){k[l>>2]=-11;k[j+(s<<2)>>2]=-1;break}else{k[j+(s<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(s<<2)>>2]=0}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-.49){k[l>>2]=-11;k[j+(t<<2)>>2]=0;break a}if(r>4294967295.49){k[l>>2]=-11;k[j+(t<<2)>>2]=-1;break a}else{k[j+(t<<2)>>2]=~~r>>>0;break a}break}case 1:{k[h>>2]=1;if(o){k[j+(t<<2)>>2]=f;break a}else{i[g+t>>0]=1;break a}break}default:k[j+(t<<2)>>2]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;u=d>4294967295.49;t=~~d>>>0;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-.49){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(r>4294967295.49){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=~~r>>>0;break b}break}case 1:{k[h>>2]=1;if(o){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(u){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function go(a,b,c,d,e,f,g,h,i,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;var m=0,n=0,o=0,p=0;m=r;r=r+16|0;n=m;if(!(Cl(a,l)|0)){o=(b|0)<1;ho(a,2,o?1:b,o?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,m+2|0,i,l)|0;p=k[l>>2]|0;r=m;return p|0}else{j[n>>1]=g;qt(a,20,c,d,e,f,1,n,h,0,i,l)|0;p=k[l>>2]|0;r=m;return p|0}return 0}function ho(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,0,x,y,K,D,z,A,G,H,E,F,J,B,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[A>>2]|0;A=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28920|0,t+28904|0,C,s)|0,e=k[C>>2]|0,(e|0)>0):0){C=0;R=1.0;while(1){S=R*10.0;C=C+1|0;if((C|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;e=k[z>>2]|0;if(!((l|0)==1&m<<16>>16==0)?(C=I,d=k[C>>2]|0,c=k[C+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0){C=ky(d|0,c|0,32768,0)|0;P=L;if(!((e|0)==21&(P>>>0>0|(P|0)==0&C>>>0>65535)))if((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)){U=11;V=0}else{U=e;V=(e|0)==16&(i[N>>0]|0)==1?0:l}else{U=21;V=0}}else{U=e;V=0}e=(U|0)==21;U=(h|0)<0|(h|0)==0&g>>>0<1073741823;l=e?(U?g:1073741823):f;f=e?(U?h:0):A;A=((j|0)<0)<<31>>31;U=0;e=0;d=g;g=h;h=0;c=0;a:while(1){C=c;P=d;W=g;X=h;while(1){Y=n+(X<<1)|0;Z=o+X|0;_=P;$=W;while(1){if((_|0)==0&($|0)==0){aa=42;break a}ba=($|0)<(f|0)|($|0)==(f|0)&_>>>0<l>>>0;ca=ba?_:l;ba=my(0,ca|0,32)|0;ca=L;da=F;ea=H;fa=k[ea>>2]|0;ga=k[ea+4>>2]|0;ea=ky(k[da>>2]|0,k[da+4>>2]|0,-1,-1)|0;da=jy(ea|0,L|0,fa|0,ga|0)|0;ea=wy(da|0,L|0,j|0,A|0)|0;da=ky(ea|0,L|0,1,0)|0;ea=L;ia=(ca|0)<(ea|0)|(ca|0)==(ea|0)&ba>>>0<da>>>0;ea=ia?ba:da;da=G;ba=k[da>>2]|0;ia=k[da+4>>2]|0;da=J;ca=yy(k[da>>2]|0,k[da+4>>2]|0,U|0,e|0)|0;da=ky(ca|0,L|0,ba|0,ia|0)|0;ia=L;ba=k[E>>2]|0;ca=(ba|0)/(j|0)|0;ja=yy(ca|0,((ca|0)<0)<<31>>31|0,fa|0,ga|0)|0;ga=ky(da|0,ia|0,ja|0,L|0)|0;ja=L;switch(k[z>>2]|0){case 21:{mu(a,ga,ja,ea,ba,Y,s)|0;ko(Y,ea,+p[x>>3],+p[y>>3],V,k[I>>2]&65535,m,Z,q,Y,s)|0;break}case 81:{ou(a,ga,ja,ea,ba,O,s)|0;ia=I;lo(O,ea,+p[x>>3],+p[y>>3],V,k[ia>>2]|0,k[ia+4>>2]|0,m,Z,q,Y,s)|0;break}case 11:{lu(a,ga,ja,ea,ba,O,s)|0;mo(O,ea,+p[x>>3],+p[y>>3],V,k[I>>2]&255,m,Z,q,Y,s)|0;break}case 41:{nu(a,ga,ja,ea,ba,O,s)|0;no(O,ea,+p[x>>3],+p[y>>3],V,k[I>>2]|0,m,Z,q,Y,s)|0;break}case 42:{pu(a,ga,ja,ea,ba,O,s)|0;oo(O,ea,+p[x>>3],+p[y>>3],V,m,Z,q,Y,s)|0;break}case 82:{qu(a,ga,ja,ea,ba,O,s)|0;po(O,ea,+p[x>>3],+p[y>>3],V,m,Z,q,Y,s)|0;break}case 16:{au(a,ga,ja,0,s)|0;ja=k[E>>2]|0;ga=k[D>>2]|0;if((ja|0)==(ga|0)){ba=ha(ja,ea)|0;fu(a,ba,((ba|0)<0)<<31>>31,O,s)|0}else gu(a,ga,ea,ja-ga|0,O,s)|0;qo(O,ea,+p[x>>3],+p[y>>3],k[D>>2]|0,T,V,N,m,Z,q,Y,s)|0;break}default:{aa=31;break a}}if((k[s>>2]|0)>0){ka=X;la=C;ma=ea;aa=35;break a}ga=((ea|0)<0)<<31>>31;ja=jy(_|0,$|0,ea|0,ga|0)|0;ba=L;if((_|0)==(ea|0)&($|0)==(ga|0)){_=ja;$=ba}else{na=ea;oa=ga;pa=ja;qa=ba;ra=ea;break}}$=ky(na|0,oa|0,X|0,C|0)|0;_=L;Y=H;Z=ha(ra,j)|0;ea=ky(k[Y>>2]|0,k[Y+4>>2]|0,Z|0,((Z|0)<0)<<31>>31|0)|0;Z=L;Y=H;k[Y>>2]=ea;k[Y+4>>2]=Z;Y=F;ba=k[Y>>2]|0;ja=k[Y+4>>2]|0;if((Z|0)<(ja|0)|(Z|0)==(ja|0)&ea>>>0<ba>>>0){C=_;P=pa;W=qa;X=$}else{sa=ea;ta=Z;ua=ba;va=ja;wa=$;xa=_;ya=pa;za=qa;break}}X=wy(sa|0,ta|0,ua|0,va|0)|0;W=L;P=ky(X|0,W|0,U|0,e|0)|0;C=L;_=yy(X|0,W|0,ua|0,va|0)|0;W=jy(sa|0,ta|0,_|0,L|0)|0;_=H;k[_>>2]=W;k[_+4>>2]=L;U=P;e=C;d=ya;g=za;h=wa;c=xa}if((aa|0)==31){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[B>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((aa|0)==35){T=+(ka>>>0)+4294967296.0*+(la|0);R=T+1.0;S=T+ +(ma|0);if((k[B>>2]|0)>0){p[v>>3]=R;p[v+8>>3]=S;k[v+16>>2]=b;Xw(M,760639,v)|0}else{p[u>>3]=R;p[u+8>>3]=S;Xw(M,760703,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((aa|0)==42){aa=k[s>>2]|0;if((aa|0)!=-11){Q=aa;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function io(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;ho(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,20,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function jo(a,b,c,d,e,f,g,h,i,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0;n=r;r=r+448|0;o=n+232|0;p=n+224|0;q=n+316|0;s=n+280|0;t=n+244|0;u=n+144|0;v=n+72|0;w=n;x=n+240|0;y=n+236|0;z=n+354|0;A=n+355|0;B=n+352|0;if((c+-1|0)>>>0>8){k[p>>2]=c;Xw(A,760763,p)|0;tk(A);k[m>>2]=320;C=320;r=n;return C|0}if(Cl(a,m)|0){if((c|0)>0){p=0;do{D=k[e+(p<<2)>>2]|0;E=v+(p<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;D=k[f+(p<<2)>>2]|0;E=w+(p<<3)|0;k[E>>2]=D;k[E+4>>2]=((D|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(c|0))}j[B>>1]=h;mt(a,20,v,w,g,1,B,i,0,l,m)|0;C=k[m>>2]|0;r=n;return C|0}if((Zk(a,x,m)|0)>0){C=k[m>>2]|0;r=n;return C|0}if(!(k[x>>2]|0)){x=(b|0)==0?1:b;F=2;G=1;H=x;I=x}else{F=b;G=k[g+(c<<2)>>2]|0;H=k[f+(c<<2)>>2]|0;I=k[e+(c<<2)>>2]|0}b=(l|0)!=0;if(b)k[l>>2]=0;k[q>>2]=1;k[s>>2]=1;k[t>>2]=1;x=u;k[x>>2]=1;k[x+4>>2]=0;x=q+4|0;k[x>>2]=1;B=s+4|0;k[B>>2]=1;w=t+4|0;k[w>>2]=1;v=u+8|0;p=v;k[p>>2]=1;k[p+4>>2]=0;p=q+8|0;k[p>>2]=1;D=s+8|0;k[D>>2]=1;E=t+8|0;k[E>>2]=1;J=u+16|0;K=J;k[K>>2]=1;k[K+4>>2]=0;K=q+12|0;k[K>>2]=1;M=s+12|0;k[M>>2]=1;N=t+12|0;k[N>>2]=1;O=u+24|0;P=O;k[P>>2]=1;k[P+4>>2]=0;P=q+16|0;k[P>>2]=1;Q=s+16|0;k[Q>>2]=1;R=t+16|0;k[R>>2]=1;S=u+32|0;T=S;k[T>>2]=1;k[T+4>>2]=0;T=q+20|0;k[T>>2]=1;U=s+20|0;k[U>>2]=1;V=t+20|0;k[V>>2]=1;W=u+40|0;X=W;k[X>>2]=1;k[X+4>>2]=0;X=q+24|0;k[X>>2]=1;Y=s+24|0;k[Y>>2]=1;Z=t+24|0;k[Z>>2]=1;_=u+48|0;$=_;k[$>>2]=1;k[$+4>>2]=0;$=q+28|0;k[$>>2]=1;aa=s+28|0;k[aa>>2]=1;ba=t+28|0;k[ba>>2]=1;ca=u+56|0;da=ca;k[da>>2]=1;k[da+4>>2]=0;da=q+32|0;k[da>>2]=1;ea=s+32|0;k[ea>>2]=1;fa=t+32|0;k[fa>>2]=1;ga=u+64|0;ha=ga;k[ha>>2]=1;k[ha+4>>2]=0;if((c|0)>0){ha=1;ia=0;ja=0;do{ka=k[f+(ja<<2)>>2]|0;la=k[e+(ja<<2)>>2]|0;if((ka|0)<(la|0)){ma=ja;na=16;break}k[q+(ja<<2)>>2]=la;k[s+(ja<<2)>>2]=ka;k[t+(ja<<2)>>2]=k[g+(ja<<2)>>2];ka=k[d+(ja<<2)>>2]|0;ha=yy(ka|0,((ka|0)<0)<<31>>31|0,ha|0,ia|0)|0;ia=L;ja=ja+1|0;ka=u+(ja<<3)|0;k[ka>>2]=ha;k[ka+4>>2]=ia}while((ja|0)<(c|0));if((na|0)==16){k[o>>2]=ma+1;Xw(A,760809,o)|0;tk(A);k[m>>2]=321;C=321;r=n;return C|0}if((c|0)==1?(k[d>>2]|0)==1:0){oa=(H-I|0)/(G|0)|0;pa=G;qa=I}else na=21}else na=21;if((na|0)==21){oa=((k[s>>2]|0)-(k[q>>2]|0)|0)/(k[g>>2]|0)|0;pa=k[t>>2]|0;qa=H}H=oa+1|0;a:do if((I|0)<=(qa|0)){oa=k[da>>2]|0;t=k[ea>>2]|0;g=(oa|0)>(t|0);s=k[$>>2]|0;na=k[aa>>2]|0;d=(s|0)>(na|0);c=k[fa>>2]|0;A=k[X>>2]|0;o=k[Y>>2]|0;ma=(A|0)>(o|0);ja=k[ba>>2]|0;ia=k[T>>2]|0;ha=k[U>>2]|0;u=(ia|0)>(ha|0);e=k[Z>>2]|0;f=k[P>>2]|0;ka=k[Q>>2]|0;la=(f|0)>(ka|0);ra=k[V>>2]|0;sa=k[K>>2]|0;ta=k[M>>2]|0;ua=(sa|0)>(ta|0);va=k[R>>2]|0;wa=k[p>>2]|0;xa=k[D>>2]|0;ya=(wa|0)>(xa|0);za=k[N>>2]|0;Aa=k[x>>2]|0;Ba=k[B>>2]|0;Ca=(Aa|0)>(Ba|0);Da=k[E>>2]|0;Ea=k[q>>2]|0;Fa=((Ea|0)<0)<<31>>31;Ga=v;Ha=k[Ga>>2]|0;Ia=k[Ga+4>>2]|0;Ga=J;Ja=k[Ga>>2]|0;Ka=k[Ga+4>>2]|0;Ga=O;La=k[Ga>>2]|0;Ma=k[Ga+4>>2]|0;Ga=S;Na=k[Ga>>2]|0;Oa=k[Ga+4>>2]|0;Ga=W;Pa=k[Ga>>2]|0;Qa=k[Ga+4>>2]|0;Ga=_;Ra=k[Ga>>2]|0;Sa=k[Ga+4>>2]|0;Ga=ca;Ta=k[Ga>>2]|0;Ua=k[Ga+4>>2]|0;Ga=ga;Va=k[Ga>>2]|0;Wa=k[Ga+4>>2]|0;Ga=((H|0)<0)<<31>>31;Xa=k[w>>2]|0;Ya=0;Za=I;b:while(1){if(g)_a=Ya;else{$a=((Za|0)<0)<<31>>31;ab=Ya;bb=oa;while(1){if(d)cb=ab;else{db=bb+-1|0;eb=yy(Va|0,Wa|0,db|0,((db|0)<0)<<31>>31|0)|0;db=L;fb=ab;gb=s;while(1){if(ma)hb=fb;else{ib=gb+-1|0;jb=yy(Ta|0,Ua|0,ib|0,((ib|0)<0)<<31>>31|0)|0;ib=L;kb=fb;lb=A;while(1){if(u)mb=kb;else{nb=lb+-1|0;ob=yy(Ra|0,Sa|0,nb|0,((nb|0)<0)<<31>>31|0)|0;nb=L;pb=kb;qb=ia;while(1){if(la)rb=pb;else{sb=qb+-1|0;tb=yy(Pa|0,Qa|0,sb|0,((sb|0)<0)<<31>>31|0)|0;sb=L;ub=pb;vb=f;while(1){if(ua)wb=ub;else{xb=vb+-1|0;yb=yy(Na|0,Oa|0,xb|0,((xb|0)<0)<<31>>31|0)|0;xb=L;zb=ub;Ab=sa;while(1){if(ya)Bb=zb;else{Cb=Ab+-1|0;Db=yy(La|0,Ma|0,Cb|0,((Cb|0)<0)<<31>>31|0)|0;Cb=L;Eb=zb;Fb=wa;while(1){if(Ca)Gb=Eb;else{Hb=Fb+-1|0;Ib=yy(Ja|0,Ka|0,Hb|0,((Hb|0)<0)<<31>>31|0)|0;Hb=L;Jb=Eb;Kb=Aa;while(1){Lb=Kb+-1|0;Mb=yy(Ha|0,Ia|0,Lb|0,((Lb|0)<0)<<31>>31|0)|0;Lb=ky(Mb|0,L|0,Ea|0,Fa|0)|0;Mb=ky(Lb|0,L|0,Ib|0,Hb|0)|0;Lb=ky(Mb|0,L|0,Db|0,Cb|0)|0;Mb=ky(Lb|0,L|0,yb|0,xb|0)|0;Lb=ky(Mb|0,L|0,tb|0,sb|0)|0;Mb=ky(Lb|0,L|0,ob|0,nb|0)|0;Lb=ky(Mb|0,L|0,jb|0,ib|0)|0;Mb=ky(Lb|0,L|0,eb|0,db|0)|0;if((ho(a,F,Za,$a,Mb,L,H,Ga,pa,1,h,i+(Jb<<1)|0,z,y,m)|0)>0)break b;if(b&(k[y>>2]|0)!=0)k[l>>2]=1;Mb=Jb+H|0;Kb=Xa+Kb|0;if((Kb|0)>(Ba|0)){Gb=Mb;break}else Jb=Mb}}Fb=Da+Fb|0;if((Fb|0)>(xa|0)){Bb=Gb;break}else Eb=Gb}}Ab=za+Ab|0;if((Ab|0)>(ta|0)){wb=Bb;break}else zb=Bb}}vb=va+vb|0;if((vb|0)>(ka|0)){rb=wb;break}else ub=wb}}qb=ra+qb|0;if((qb|0)>(ha|0)){mb=rb;break}else pb=rb}}lb=e+lb|0;if((lb|0)>(o|0)){hb=mb;break}else kb=mb}}gb=ja+gb|0;if((gb|0)>(na|0)){cb=hb;break}else fb=hb}}bb=c+bb|0;if((bb|0)>(t|0)){_a=cb;break}else ab=cb}}Za=Za+G|0;if((Za|0)>(qa|0))break a;else Ya=_a}C=k[m>>2]|0;r=n;return C|0}while(0);C=k[m>>2]|0;r=n;return C|0}function ko(a,b,c,d,e,f,g,h,l,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0.0;p=c==1.0;q=p&d==32768.0;if(!e){if(q){if((b|0)>0)r=0;else{s=k[o>>2]|0;return s|0}do{j[n+(r<<1)>>1]=m[a+(r<<1)>>1]^32768;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}r=(b|0)>0;if(p&d==0.0){if(r)t=0;else{s=k[o>>2]|0;return s|0}do{u=j[a+(t<<1)>>1]|0;if(u<<16>>16<0){k[o>>2]=-11;v=0}else v=u;j[n+(t<<1)>>1]=v;t=t+1|0}while((t|0)!=(b|0));s=k[o>>2]|0;return s|0}if(r)w=0;else{s=k[o>>2]|0;return s|0}do{x=+(j[a+(w<<1)>>1]|0)*c+d;do if(!(x<-.49))if(x>65535.49){k[o>>2]=-11;j[n+(w<<1)>>1]=-1;break}else{j[n+(w<<1)>>1]=~~x;break}else{k[o>>2]=-11;j[n+(w<<1)>>1]=0}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q){if((b|0)<=0){s=k[o>>2]|0;return s|0}if((e|0)==1){q=0;do{w=j[a+(q<<1)>>1]|0;if(w<<16>>16==f<<16>>16){k[l>>2]=1;j[n+(q<<1)>>1]=g}else j[n+(q<<1)>>1]=w&65535^32768;q=q+1|0}while((q|0)!=(b|0));s=k[o>>2]|0;return s|0}else{q=0;do{w=j[a+(q<<1)>>1]|0;if(w<<16>>16==f<<16>>16){k[l>>2]=1;i[h+q>>0]=1}else j[n+(q<<1)>>1]=w&65535^32768;q=q+1|0}while((q|0)!=(b|0));s=k[o>>2]|0;return s|0}}q=(b|0)>0;if(p&d==0.0){if(!q){s=k[o>>2]|0;return s|0}if((e|0)==1){p=0;do{w=j[a+(p<<1)>>1]|0;do if(w<<16>>16!=f<<16>>16)if(w<<16>>16<0){k[o>>2]=-11;j[n+(p<<1)>>1]=0;break}else{j[n+(p<<1)>>1]=w;break}else{k[l>>2]=1;j[n+(p<<1)>>1]=g}while(0);p=p+1|0}while((p|0)!=(b|0));s=k[o>>2]|0;return s|0}else{p=0;do{w=j[a+(p<<1)>>1]|0;do if(w<<16>>16!=f<<16>>16)if(w<<16>>16<0){k[o>>2]=-11;j[n+(p<<1)>>1]=0;break}else{j[n+(p<<1)>>1]=w;break}else{k[l>>2]=1;i[h+p>>0]=1}while(0);p=p+1|0}while((p|0)!=(b|0));s=k[o>>2]|0;return s|0}}if(!q){s=k[o>>2]|0;return s|0}if((e|0)==1){e=0;do{q=j[a+(e<<1)>>1]|0;do if(q<<16>>16!=f<<16>>16){x=+(q<<16>>16)*c+d;if(x<-.49){k[o>>2]=-11;j[n+(e<<1)>>1]=0;break}if(x>65535.49){k[o>>2]=-11;j[n+(e<<1)>>1]=-1;break}else{j[n+(e<<1)>>1]=~~x;break}}else{k[l>>2]=1;j[n+(e<<1)>>1]=g}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}else{e=0;do{g=j[a+(e<<1)>>1]|0;do if(g<<16>>16!=f<<16>>16){x=+(g<<16>>16)*c+d;if(x<-.49){k[o>>2]=-11;j[n+(e<<1)>>1]=0;break}if(x>65535.49){k[o>>2]=-11;j[n+(e<<1)>>1]=-1;break}else{j[n+(e<<1)>>1]=~~x;break}}else{k[l>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}return 0}function lo(a,b,c,d,e,f,g,h,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e)if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{t=a+(r<<3)|0;u=k[t>>2]|0;v=k[t+4>>2]|0;do if((v|0)>=0)if((v|0)>0|(v|0)==0&u>>>0>65535){k[o>>2]=-11;j[n+(r<<1)>>1]=-1;break}else{j[n+(r<<1)>>1]=u;break}else{k[o>>2]=-11;j[n+(r<<1)>>1]=0}while(0);r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}else{if(q)w=0;else{s=k[o>>2]|0;return s|0}do{r=a+(w<<3)|0;x=(+((k[r>>2]|0)>>>0)+4294967296.0*+(k[r+4>>2]|0))*c+d;do if(!(x<-.49))if(x>65535.49){k[o>>2]=-11;j[n+(w<<1)>>1]=-1;break}else{j[n+(w<<1)>>1]=~~x;break}else{k[o>>2]=-11;j[n+(w<<1)>>1]=0}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}p=(e|0)==1;w=0;do{r=a+(w<<3)|0;u=k[r>>2]|0;v=k[r+4>>2]|0;do if((u|0)==(f|0)&(v|0)==(g|0)){k[m>>2]=1;if(p){j[n+(w<<1)>>1]=h;break}else{i[l+w>>0]=1;break}}else{if((v|0)<0){k[o>>2]=-11;j[n+(w<<1)>>1]=0;break}if((v|0)>0|(v|0)==0&u>>>0>65535){k[o>>2]=-11;j[n+(w<<1)>>1]=-1;break}else{j[n+(w<<1)>>1]=u;break}}while(0);w=w+1|0}while((w|0)!=(b|0));s=k[o>>2]|0;return s|0}else{if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{w=a+(e<<3)|0;p=k[w>>2]|0;u=k[w+4>>2]|0;do if((p|0)==(f|0)&(u|0)==(g|0)){k[m>>2]=1;if(q){j[n+(e<<1)>>1]=h;break}else{i[l+e>>0]=1;break}}else{x=(+(p>>>0)+4294967296.0*+(u|0))*c+d;if(x<-.49){k[o>>2]=-11;j[n+(e<<1)>>1]=0;break}if(x>65535.49){k[o>>2]=-11;j[n+(e<<1)>>1]=-1;break}else{j[n+(e<<1)>>1]=~~x;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}return 0}function mo(a,b,c,d,e,f,g,h,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0.0;p=c==1.0&d==0.0;q=(b|0)>0;if(!e){if(p){if(q)r=0;else{s=k[o>>2]|0;return s|0}do{j[n+(r<<1)>>1]=l[a+r>>0]|0;r=r+1|0}while((r|0)!=(b|0));s=k[o>>2]|0;return s|0}if(q)t=0;else{s=k[o>>2]|0;return s|0}do{u=+(l[a+t>>0]|0|0)*c+d;do if(!(u<-.49))if(u>65535.49){k[o>>2]=-11;j[n+(t<<1)>>1]=-1;break}else{j[n+(t<<1)>>1]=~~u;break}else{k[o>>2]=-11;j[n+(t<<1)>>1]=0}while(0);t=t+1|0}while((t|0)!=(b|0));s=k[o>>2]|0;return s|0}if(p){if(!q){s=k[o>>2]|0;return s|0}if((e|0)==1){p=0;do{t=i[a+p>>0]|0;if(t<<24>>24==f<<24>>24){k[m>>2]=1;j[n+(p<<1)>>1]=g}else j[n+(p<<1)>>1]=t&255;p=p+1|0}while((p|0)!=(b|0));s=k[o>>2]|0;return s|0}else{p=0;do{t=i[a+p>>0]|0;if(t<<24>>24==f<<24>>24){k[m>>2]=1;i[h+p>>0]=1}else j[n+(p<<1)>>1]=t&255;p=p+1|0}while((p|0)!=(b|0));s=k[o>>2]|0;return s|0}}if(!q){s=k[o>>2]|0;return s|0}q=(e|0)==1;e=0;do{p=i[a+e>>0]|0;do if(p<<24>>24==f<<24>>24){k[m>>2]=1;if(q){j[n+(e<<1)>>1]=g;break}else{i[h+e>>0]=1;break}}else{u=+(p&255|0)*c+d;if(u<-.49){k[o>>2]=-11;j[n+(e<<1)>>1]=0;break}if(u>65535.49){k[o>>2]=-11;j[n+(e<<1)>>1]=-1;break}else{j[n+(e<<1)>>1]=~~u;break}}while(0);e=e+1|0}while((e|0)!=(b|0));s=k[o>>2]|0;return s|0}function no(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=k[a+(q<<2)>>2]|0;do if((s|0)>=0)if((s|0)>65535){k[n>>2]=-11;j[m+(q<<1)>>1]=-1;break}else{j[m+(q<<1)>>1]=s;break}else{k[n>>2]=-11;j[m+(q<<1)>>1]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)t=0;else{r=k[n>>2]|0;return r|0}do{u=+(k[a+(t<<2)>>2]|0)*c+d;do if(!(u<-.49))if(u>65535.49){k[n>>2]=-11;j[m+(t<<1)>>1]=-1;break}else{j[m+(t<<1)>>1]=~~u;break}else{k[n>>2]=-11;j[m+(t<<1)>>1]=0}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;t=0;do{q=k[a+(t<<2)>>2]|0;do if((q|0)==(f|0)){k[l>>2]=1;if(o){j[m+(t<<1)>>1]=g;break}else{i[h+t>>0]=1;break}}else{if((q|0)<0){k[n>>2]=-11;j[m+(t<<1)>>1]=0;break}if((q|0)>65535){k[n>>2]=-11;j[m+(t<<1)>>1]=-1;break}else{j[m+(t<<1)>>1]=q;break}}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{t=k[a+(e<<2)>>2]|0;do if((t|0)==(f|0)){k[l>>2]=1;if(p){j[m+(e<<1)>>1]=g;break}else{i[h+e>>0]=1;break}}else{u=+(t|0)*c+d;if(u<-.49){k[n>>2]=-11;j[m+(e<<1)>>1]=0;break}if(u>65535.49){k[n>>2]=-11;j[m+(e<<1)>>1]=-1;break}else{j[m+(e<<1)>>1]=~~u;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function oo(a,b,c,d,e,f,g,h,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;var p=0,q=0,r=0,s=0.0,t=0.0,u=0,v=0,w=0,x=0;if(!e){p=(b|0)>0;if(c==1.0&d==0.0){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(q<<2)>>2];t=s;do if(!(t<-.49))if(t>65535.49){k[n>>2]=-11;j[l+(q<<1)>>1]=-1;break}else{j[l+(q<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(q<<1)>>1]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)u=0;else{r=k[n>>2]|0;return r|0}do{s=+o[a+(u<<2)>>2]*c+d;do if(!(s<-.49))if(s>65535.49){k[n>>2]=-11;j[l+(u<<1)>>1]=-1;break}else{j[l+(u<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(u<<1)>>1]=0}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}}u=a+2|0;p=(b|0)>0;if(c==1.0&d==0.0){if(!p){r=k[n>>2]|0;return r|0}q=(e|0)==1;v=0;w=u;while(1){x=(m[w>>1]|0)&32640;a:do switch(((x|0)==32640?1:(x|0)==0?2:0)|0){case 0:{s=+o[a+(v<<2)>>2];t=s;if(t<-.49){k[n>>2]=-11;j[l+(v<<1)>>1]=0;break a}if(t>65535.49){k[n>>2]=-11;j[l+(v<<1)>>1]=-1;break a}else{j[l+(v<<1)>>1]=~~s;break a}break}case 1:{k[h>>2]=1;if(q){j[l+(v<<1)>>1]=f;break a}else{i[g+v>>0]=1;break a}break}default:j[l+(v<<1)>>1]=0}while(0);v=v+1|0;if((v|0)==(b|0))break;else w=w+4|0}r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=d<-.49;w=d>65535.49;v=~~d&65535;q=(e|0)==1;e=0;x=u;while(1){u=(m[x>>1]|0)&32640;b:do switch(((u|0)==32640?1:(u|0)==0?2:0)|0){case 0:{s=+o[a+(e<<2)>>2]*c+d;if(s<-.49){k[n>>2]=-11;j[l+(e<<1)>>1]=0;break b}if(s>65535.49){k[n>>2]=-11;j[l+(e<<1)>>1]=-1;break b}else{j[l+(e<<1)>>1]=~~s;break b}break}case 1:{k[h>>2]=1;if(q){j[l+(e<<1)>>1]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(p){k[n>>2]=-11;j[l+(e<<1)>>1]=0;break b}if(w){k[n>>2]=-11;j[l+(e<<1)>>1]=-1;break b}else{j[l+(e<<1)>>1]=v;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else x=x+4|0}r=k[n>>2]|0;return r|0}function po(a,b,c,d,e,f,g,h,l,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;var o=0,q=0,r=0,s=0.0,t=0,u=0,v=0,w=0;if(!e){o=(b|0)>0;if(c==1.0&d==0.0){if(o)q=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(q<<3)>>3];do if(!(s<-.49))if(s>65535.49){k[n>>2]=-11;j[l+(q<<1)>>1]=-1;break}else{j[l+(q<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(q<<1)>>1]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(o)t=0;else{r=k[n>>2]|0;return r|0}do{s=+p[a+(t<<3)>>3]*c+d;do if(!(s<-.49))if(s>65535.49){k[n>>2]=-11;j[l+(t<<1)>>1]=-1;break}else{j[l+(t<<1)>>1]=~~s;break}else{k[n>>2]=-11;j[l+(t<<1)>>1]=0}while(0);t=t+1|0}while((t|0)!=(b|0));r=k[n>>2]|0;return r|0}}t=a+6|0;o=(b|0)>0;if(c==1.0&d==0.0){if(!o){r=k[n>>2]|0;return r|0}q=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32752;a:do switch(((w|0)==32752?1:(w|0)==0?2:0)|0){case 0:{s=+p[a+(u<<3)>>3];if(s<-.49){k[n>>2]=-11;j[l+(u<<1)>>1]=0;break a}if(s>65535.49){k[n>>2]=-11;j[l+(u<<1)>>1]=-1;break a}else{j[l+(u<<1)>>1]=~~s;break a}break}case 1:{k[h>>2]=1;if(q){j[l+(u<<1)>>1]=f;break a}else{i[g+u>>0]=1;break a}break}default:j[l+(u<<1)>>1]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+8|0}r=k[n>>2]|0;return r|0}if(!o){r=k[n>>2]|0;return r|0}o=d<-.49;v=d>65535.49;u=~~d&65535;q=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32752;b:do switch(((t|0)==32752?1:(t|0)==0?2:0)|0){case 0:{s=+p[a+(e<<3)>>3]*c+d;if(s<-.49){k[n>>2]=-11;j[l+(e<<1)>>1]=0;break b}if(s>65535.49){k[n>>2]=-11;j[l+(e<<1)>>1]=-1;break b}else{j[l+(e<<1)>>1]=~~s;break b}break}case 1:{k[h>>2]=1;if(q){j[l+(e<<1)>>1]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(o){k[n>>2]=-11;j[l+(e<<1)>>1]=0;break b}if(v){k[n>>2]=-11;j[l+(e<<1)>>1]=-1;break b}else{j[l+(e<<1)>>1]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+8|0}r=k[n>>2]|0;return r|0}function qo(a,b,c,d,e,f,g,h,l,m,n,o,p){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0,S=0,T=0,U=0,V=0.0,W=0,X=0,Z=0.0,_=0,$=0,aa=0.0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0.0,ia=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0;q=r;r=r+96|0;s=q;t=q+4|0;u=xx(h)|0;a:do if((b|0)>0){v=(g|0)==0;w=(g|0)==1;x=a;y=0;b:while(1){z=x+e|0;A=i[z>>0]|0;i[z>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,x,u)|0)==0:0)if(!v){k[n>>2]=1;if(w){j[o+(y<<1)>>1]=l;B=z;break}else{i[m+y>>0]=1;B=z;break}}else B=z;else{C=x;D=9}while(0);do if((D|0)==9){c:while(1){D=0;E=i[C>>0]|0;switch(E<<24>>24){case 43:case 45:{F=E;G=C;D=11;break c;break}case 32:break;default:{H=E;I=C;J=1;break c}}C=C+1|0;D=9}if((D|0)==11){D=0;E=F<<24>>24==45;K=G;while(1){L=K+1|0;M=i[L>>0]|0;if(M<<24>>24==32)K=L;else{N=M;O=L;break}}H=N;I=O;J=E?-1:1}if((H+-48&255)<10){K=H;L=I;P=0.0;while(1){Q=P*10.0+ +(K<<24>>24);M=L;while(1){R=M+1|0;S=i[R>>0]|0;if(S<<24>>24==32)M=R;else{T=S;U=R;break}}V=Q+-48.0;if((T+-48&255)<10){K=T;L=U;P=V}else{W=T;X=U;Z=V;break}}}else{W=H;X=I;Z=0.0}switch(W<<24>>24){case 44:case 46:{L=X;while(1){K=L+1|0;E=i[K>>0]|0;if(E<<24>>24==32)L=K;else{_=E;$=K;break}}if((_+-48&255)<10){L=_;K=$;P=1.0;V=Z;while(1){aa=V*10.0+ +(L<<24>>24)+-48.0;E=K;while(1){M=E+1|0;R=i[M>>0]|0;if(R<<24>>24==32)E=M;else{ba=R;ca=M;break}}Q=P*10.0;if((ba+-48&255)<10){L=ba;K=ca;P=Q;V=aa}else{da=ba;ea=ca;fa=1;ga=Q;ia=aa;break}}}else{da=_;ea=$;fa=1;ga=1.0;ia=Z}break}default:{da=W;ea=X;fa=0;ga=1.0;ia=Z}}if((da&-2)<<24>>24==68){K=ea;d:while(1){L=K+1|0;E=i[L>>0]|0;switch(E<<24>>24){case 32:{K=L;break}case 43:case 45:{ja=E;ka=K;D=28;break d;break}default:{la=E;ma=L;na=1;break d}}}if((D|0)==28){D=0;K=ja<<24>>24==45;L=ka+2|0;while(1){E=i[L>>0]|0;if(E<<24>>24==32)L=L+1|0;else{oa=E;pa=L;break}}la=oa;ma=pa;na=K?-1:1}if((la+-48&255)<10){L=la;E=ma;M=0;while(1){R=(M*10|0)+-48|0;S=E;while(1){qa=S+1|0;ra=i[qa>>0]|0;if(ra<<24>>24==32)S=qa;else{sa=ra;ta=qa;break}}S=R+(L<<24>>24)|0;if((sa+-48&255)<10){L=sa;E=ta;M=S}else{ua=sa;va=ta;wa=na;xa=S;break}}}else{ua=la;va=ma;wa=na;xa=0}}else{ua=da;va=ea;wa=1;xa=0}if(ua<<24>>24){ya=z;za=A;Aa=x;break b}aa=+(J|0)*ia/((fa|0)==0?f:ga)*+Y(10.0,+(+(ha(wa,xa)|0)))*c+d;if(aa<-.49){k[p>>2]=-11;j[o+(y<<1)>>1]=0;B=va;break}if(aa>65535.49){k[p>>2]=-11;j[o+(y<<1)>>1]=-1;B=va;break}else{j[o+(y<<1)>>1]=~~aa;B=va;break}}while(0);i[z>>0]=A;y=y+1|0;if((y|0)>=(b|0))break a;else x=B}x=t;y=761193;w=x+36|0;do{i[x>>0]=i[y>>0]|0;x=x+1|0;y=y+1|0}while((x|0)<(w|0));tk(t);k[s>>2]=Aa;Xw(t,761229,s)|0;tk(t);i[ya>>0]=za;k[p>>2]=409;Ba=409;r=q;return Ba|0}while(0);Ba=k[p>>2]|0;r=q;return Ba|0}function ro(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;so(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,1,1,g,h,l+4|0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}else{k[m>>2]=g;qt(a,40,c,d,e,f,1,m,h,0,i,j)|0;o=k[j>>2]|0;r=l;return o|0}return 0}function so(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;var t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0;t=r;r=r+29072|0;u=t+28888|0;v=t+28864|0;w=t+28856|0;x=t+28848|0;y=t+28840|0;z=t+28932|0;A=t+28928|0;B=t+28924|0;C=t+28916|0;D=t+28912|0;E=t+28908|0;F=t+32|0;G=t+24|0;H=t+16|0;I=t+8|0;J=t;K=t+28956|0;M=t+28976|0;N=t+28936|0;O=t+40|0;P=k[s>>2]|0;if((g|0)==0&(h|0)==0|(P|0)>0){Q=P;r=t;return Q|0}if(q)k[q>>2]=0;if((l|0)==2)sy(o|0,0,g|0)|0;if((nl(a,b,c,d,e,f,g,h,0,x,y,K,D,z,A,G,H,E,F,J,B,I,N,s)|0)>0){Q=k[s>>2]|0;r=t;return Q|0}f=k[A>>2]|0;A=((f|0)<0)<<31>>31;e=ha(k[E>>2]|0,j)|0;k[E>>2]=e;if((k[z>>2]|0)==16?(Nk(K,t+28920|0,t+28904|0,C,s)|0,e=k[C>>2]|0,(e|0)>0):0){C=0;R=1.0;while(1){S=R*10.0;C=C+1|0;if((C|0)>=(e|0)){T=S;break}else R=S}}else T=1.0;if(((!((l|0)==1&(m|0)==0)?(e=k[z>>2]|0,C=I,d=k[C>>2]|0,c=k[C+4>>2]|0,!(((e|0)%10|0|0)==1&((d|0)==1234554321&(c|0)==0))):0)?(C=ky(d|0,c|0,32768,0)|0,P=L,!((e|0)==21&(P>>>0>0|(P|0)==0&C>>>0>65535))):0)?!((e|0)==11&(c>>>0>0|(c|0)==0&d>>>0>255)):0)U=(e|0)==16&(i[N>>0]|0)==1?0:l;else U=0;l=((j|0)<0)<<31>>31;e=0;d=0;c=g;g=h;h=0;C=0;a:while(1){P=C;V=c;W=g;X=h;while(1){Y=o+X|0;Z=n+(X<<2)|0;_=V;$=W;while(1){if((_|0)==0&($|0)==0){aa=42;break a}ba=($|0)<(A|0)|($|0)==(A|0)&_>>>0<f>>>0;ca=ba?_:f;ba=my(0,ca|0,32)|0;ca=L;da=F;ea=H;fa=k[ea>>2]|0;ga=k[ea+4>>2]|0;ea=ky(k[da>>2]|0,k[da+4>>2]|0,-1,-1)|0;da=jy(ea|0,L|0,fa|0,ga|0)|0;ea=wy(da|0,L|0,j|0,l|0)|0;da=ky(ea|0,L|0,1,0)|0;ea=L;ia=(ca|0)<(ea|0)|(ca|0)==(ea|0)&ba>>>0<da>>>0;ea=ia?ba:da;da=G;ba=k[da>>2]|0;ia=k[da+4>>2]|0;da=J;ca=yy(k[da>>2]|0,k[da+4>>2]|0,e|0,d|0)|0;da=ky(ca|0,L|0,ba|0,ia|0)|0;ia=L;ba=k[E>>2]|0;ca=(ba|0)/(j|0)|0;ja=yy(ca|0,((ca|0)<0)<<31>>31|0,fa|0,ga|0)|0;ga=ky(da|0,ia|0,ja|0,L|0)|0;ja=L;switch(k[z>>2]|0){case 41:{nu(a,ga,ja,ea,ba,O,s)|0;vo(O,ea,+p[x>>3],+p[y>>3],U,k[I>>2]|0,m,Y,q,Z,s)|0;break}case 81:{ou(a,ga,ja,ea,ba,O,s)|0;ia=I;wo(O,ea,+p[x>>3],+p[y>>3],U,k[ia>>2]|0,k[ia+4>>2]|0,m,Y,q,Z,s)|0;break}case 11:{lu(a,ga,ja,ea,ba,O,s)|0;xo(O,ea,+p[x>>3],+p[y>>3],U,k[I>>2]&255,m,Y,q,Z,s)|0;break}case 21:{mu(a,ga,ja,ea,ba,O,s)|0;yo(O,ea,+p[x>>3],+p[y>>3],U,k[I>>2]&65535,m,Y,q,Z,s)|0;break}case 42:{pu(a,ga,ja,ea,ba,O,s)|0;zo(O,ea,+p[x>>3],+p[y>>3],U,m,Y,q,Z,s)|0;break}case 82:{qu(a,ga,ja,ea,ba,O,s)|0;Ao(O,ea,+p[x>>3],+p[y>>3],U,m,Y,q,Z,s)|0;break}case 16:{au(a,ga,ja,0,s)|0;ja=k[E>>2]|0;ga=k[D>>2]|0;if((ja|0)==(ga|0)){ba=ha(ja,ea)|0;fu(a,ba,((ba|0)<0)<<31>>31,O,s)|0}else gu(a,ga,ea,ja-ga|0,O,s)|0;Bo(O,ea,+p[x>>3],+p[y>>3],k[D>>2]|0,T,U,N,m,Y,q,Z,s)|0;break}default:{aa=31;break a}}if((k[s>>2]|0)>0){ka=X;la=P;ma=ea;aa=35;break a}ga=((ea|0)<0)<<31>>31;ja=jy(_|0,$|0,ea|0,ga|0)|0;ba=L;if((_|0)==(ea|0)&($|0)==(ga|0)){_=ja;$=ba}else{na=ea;oa=ga;pa=ja;qa=ba;ra=ea;break}}$=ky(na|0,oa|0,X|0,P|0)|0;_=L;Z=H;Y=ha(ra,j)|0;ea=ky(k[Z>>2]|0,k[Z+4>>2]|0,Y|0,((Y|0)<0)<<31>>31|0)|0;Y=L;Z=H;k[Z>>2]=ea;k[Z+4>>2]=Y;Z=F;ba=k[Z>>2]|0;ja=k[Z+4>>2]|0;if((Y|0)<(ja|0)|(Y|0)==(ja|0)&ea>>>0<ba>>>0){P=_;V=pa;W=qa;X=$}else{sa=ea;ta=Y;ua=ba;va=ja;wa=$;xa=_;ya=pa;za=qa;break}}X=wy(sa|0,ta|0,ua|0,va|0)|0;W=L;V=ky(X|0,W|0,e|0,d|0)|0;P=L;_=yy(X|0,W|0,ua|0,va|0)|0;W=jy(sa|0,ta|0,_|0,L|0)|0;_=H;k[_>>2]=W;k[_+4>>2]=L;e=V;d=P;c=ya;g=za;h=wa;C=xa}if((aa|0)==31){k[w>>2]=b;k[w+4>>2]=K;Xw(M,760855,w)|0;tk(M);if((k[B>>2]|0)==1){k[s>>2]=311;Q=311;r=t;return Q|0}else{k[s>>2]=312;Q=312;r=t;return Q|0}}else if((aa|0)==35){T=+(ka>>>0)+4294967296.0*+(la|0);R=T+1.0;S=T+ +(ma|0);if((k[B>>2]|0)>0){p[v>>3]=R;p[v+8>>3]=S;k[v+16>>2]=b;Xw(M,760910,v)|0}else{p[u>>3]=R;p[u+8>>3]=S;Xw(M,760974,u)|0}tk(M);Q=k[s>>2]|0;r=t;return Q|0}else if((aa|0)==42){aa=k[s>>2]|0;if((aa|0)!=-11){Q=aa;r=t;return Q|0}tk(761034);k[s>>2]=412;Q=412;r=t;return Q|0}return 0}function to(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;so(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,1,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{qt(a,40,c,d,e,f,2,0,g,h,i,j)|0;m=k[j>>2]|0;return m|0}return 0}function uo(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0;m=r;r=r+448|0;n=m+232|0;o=m+224|0;p=m+320|0;q=m+284|0;s=m+248|0;t=m+144|0;u=m+72|0;v=m;w=m+244|0;x=m+240|0;y=m+356|0;z=m+357|0;A=m+236|0;if((c+-1|0)>>>0>8){k[o>>2]=c;Xw(z,761101,o)|0;tk(z);k[l>>2]=320;B=320;r=m;return B|0}if(Cl(a,l)|0){if((c|0)>0){o=0;do{C=k[e+(o<<2)>>2]|0;D=u+(o<<3)|0;k[D>>2]=C;k[D+4>>2]=((C|0)<0)<<31>>31;C=k[f+(o<<2)>>2]|0;D=v+(o<<3)|0;k[D>>2]=C;k[D+4>>2]=((C|0)<0)<<31>>31;o=o+1|0}while((o|0)!=(c|0))}k[A>>2]=h;mt(a,40,u,v,g,1,A,i,0,j,l)|0;B=k[l>>2]|0;r=m;return B|0}if((Zk(a,w,l)|0)>0){B=k[l>>2]|0;r=m;return B|0}if(!(k[w>>2]|0)){w=(b|0)==0?1:b;E=2;F=1;G=w;H=w}else{E=b;F=k[g+(c<<2)>>2]|0;G=k[f+(c<<2)>>2]|0;H=k[e+(c<<2)>>2]|0}b=(j|0)!=0;if(b)k[j>>2]=0;k[p>>2]=1;k[q>>2]=1;k[s>>2]=1;w=t;k[w>>2]=1;k[w+4>>2]=0;w=p+4|0;k[w>>2]=1;A=q+4|0;k[A>>2]=1;v=s+4|0;k[v>>2]=1;u=t+8|0;o=u;k[o>>2]=1;k[o+4>>2]=0;o=p+8|0;k[o>>2]=1;C=q+8|0;k[C>>2]=1;D=s+8|0;k[D>>2]=1;I=t+16|0;J=I;k[J>>2]=1;k[J+4>>2]=0;J=p+12|0;k[J>>2]=1;K=q+12|0;k[K>>2]=1;M=s+12|0;k[M>>2]=1;N=t+24|0;O=N;k[O>>2]=1;k[O+4>>2]=0;O=p+16|0;k[O>>2]=1;P=q+16|0;k[P>>2]=1;Q=s+16|0;k[Q>>2]=1;R=t+32|0;S=R;k[S>>2]=1;k[S+4>>2]=0;S=p+20|0;k[S>>2]=1;T=q+20|0;k[T>>2]=1;U=s+20|0;k[U>>2]=1;V=t+40|0;W=V;k[W>>2]=1;k[W+4>>2]=0;W=p+24|0;k[W>>2]=1;X=q+24|0;k[X>>2]=1;Y=s+24|0;k[Y>>2]=1;Z=t+48|0;_=Z;k[_>>2]=1;k[_+4>>2]=0;_=p+28|0;k[_>>2]=1;$=q+28|0;k[$>>2]=1;aa=s+28|0;k[aa>>2]=1;ba=t+56|0;ca=ba;k[ca>>2]=1;k[ca+4>>2]=0;ca=p+32|0;k[ca>>2]=1;da=q+32|0;k[da>>2]=1;ea=s+32|0;k[ea>>2]=1;fa=t+64|0;ga=fa;k[ga>>2]=1;k[ga+4>>2]=0;if((c|0)>0){ga=1;ha=0;ia=0;do{ja=k[f+(ia<<2)>>2]|0;ka=k[e+(ia<<2)>>2]|0;if((ja|0)<(ka|0)){la=ia;ma=16;break}k[p+(ia<<2)>>2]=ka;k[q+(ia<<2)>>2]=ja;k[s+(ia<<2)>>2]=k[g+(ia<<2)>>2];ja=k[d+(ia<<2)>>2]|0;ga=yy(ja|0,((ja|0)<0)<<31>>31|0,ga|0,ha|0)|0;ha=L;ia=ia+1|0;ja=t+(ia<<3)|0;k[ja>>2]=ga;k[ja+4>>2]=ha}while((ia|0)<(c|0));if((ma|0)==16){k[n>>2]=la+1;Xw(z,761147,n)|0;tk(z);k[l>>2]=321;B=321;r=m;return B|0}if((c|0)==1?(k[d>>2]|0)==1:0){na=(G-H|0)/(F|0)|0;oa=F;pa=H}else ma=21}else ma=21;if((ma|0)==21){na=((k[q>>2]|0)-(k[p>>2]|0)|0)/(k[g>>2]|0)|0;oa=k[s>>2]|0;pa=G}G=na+1|0;a:do if((H|0)<=(pa|0)){na=k[ca>>2]|0;s=k[da>>2]|0;g=(na|0)>(s|0);q=k[_>>2]|0;ma=k[$>>2]|0;d=(q|0)>(ma|0);c=k[ea>>2]|0;z=k[W>>2]|0;n=k[X>>2]|0;la=(z|0)>(n|0);ia=k[aa>>2]|0;ha=k[S>>2]|0;ga=k[T>>2]|0;t=(ha|0)>(ga|0);e=k[Y>>2]|0;f=k[O>>2]|0;ja=k[P>>2]|0;ka=(f|0)>(ja|0);qa=k[U>>2]|0;ra=k[J>>2]|0;sa=k[K>>2]|0;ta=(ra|0)>(sa|0);ua=k[Q>>2]|0;va=k[o>>2]|0;wa=k[C>>2]|0;xa=(va|0)>(wa|0);ya=k[M>>2]|0;za=k[w>>2]|0;Aa=k[A>>2]|0;Ba=(za|0)>(Aa|0);Ca=k[D>>2]|0;Da=k[p>>2]|0;Ea=((Da|0)<0)<<31>>31;Fa=u;Ga=k[Fa>>2]|0;Ha=k[Fa+4>>2]|0;Fa=I;Ia=k[Fa>>2]|0;Ja=k[Fa+4>>2]|0;Fa=N;Ka=k[Fa>>2]|0;La=k[Fa+4>>2]|0;Fa=R;Ma=k[Fa>>2]|0;Na=k[Fa+4>>2]|0;Fa=V;Oa=k[Fa>>2]|0;Pa=k[Fa+4>>2]|0;Fa=Z;Qa=k[Fa>>2]|0;Ra=k[Fa+4>>2]|0;Fa=ba;Sa=k[Fa>>2]|0;Ta=k[Fa+4>>2]|0;Fa=fa;Ua=k[Fa>>2]|0;Va=k[Fa+4>>2]|0;Fa=((G|0)<0)<<31>>31;Wa=k[v>>2]|0;Xa=0;Ya=H;b:while(1){if(g)Za=Xa;else{_a=((Ya|0)<0)<<31>>31;$a=Xa;ab=na;while(1){if(d)bb=$a;else{cb=ab+-1|0;db=yy(Ua|0,Va|0,cb|0,((cb|0)<0)<<31>>31|0)|0;cb=L;eb=$a;fb=q;while(1){if(la)gb=eb;else{hb=fb+-1|0;ib=yy(Sa|0,Ta|0,hb|0,((hb|0)<0)<<31>>31|0)|0;hb=L;jb=eb;kb=z;while(1){if(t)lb=jb;else{mb=kb+-1|0;nb=yy(Qa|0,Ra|0,mb|0,((mb|0)<0)<<31>>31|0)|0;mb=L;ob=jb;pb=ha;while(1){if(ka)qb=ob;else{rb=pb+-1|0;sb=yy(Oa|0,Pa|0,rb|0,((rb|0)<0)<<31>>31|0)|0;rb=L;tb=ob;ub=f;while(1){if(ta)vb=tb;else{wb=ub+-1|0;xb=yy(Ma|0,Na|0,wb|0,((wb|0)<0)<<31>>31|0)|0;wb=L;yb=tb;zb=ra;while(1){if(xa)Ab=yb;else{Bb=zb+-1|0;Cb=yy(Ka|0,La|0,Bb|0,((Bb|0)<0)<<31>>31|0)|0;Bb=L;Db=yb;Eb=va;while(1){if(Ba)Fb=Db;else{Gb=Eb+-1|0;Hb=yy(Ia|0,Ja|0,Gb|0,((Gb|0)<0)<<31>>31|0)|0;Gb=L;Ib=Db;Jb=za;while(1){Kb=Jb+-1|0;Lb=yy(Ga|0,Ha|0,Kb|0,((Kb|0)<0)<<31>>31|0)|0;Kb=ky(Lb|0,L|0,Da|0,Ea|0)|0;Lb=ky(Kb|0,L|0,Hb|0,Gb|0)|0;Kb=ky(Lb|0,L|0,Cb|0,Bb|0)|0;Lb=ky(Kb|0,L|0,xb|0,wb|0)|0;Kb=ky(Lb|0,L|0,sb|0,rb|0)|0;Lb=ky(Kb|0,L|0,nb|0,mb|0)|0;Kb=ky(Lb|0,L|0,ib|0,hb|0)|0;Lb=ky(Kb|0,L|0,db|0,cb|0)|0;if((so(a,E,Ya,_a,Lb,L,G,Fa,oa,1,h,i+(Ib<<2)|0,y,x,l)|0)>0)break b;if(b&(k[x>>2]|0)!=0)k[j>>2]=1;Lb=Ib+G|0;Jb=Wa+Jb|0;if((Jb|0)>(Aa|0)){Fb=Lb;break}else Ib=Lb}}Eb=Ca+Eb|0;if((Eb|0)>(wa|0)){Ab=Fb;break}else Db=Fb}}zb=ya+zb|0;if((zb|0)>(sa|0)){vb=Ab;break}else yb=Ab}}ub=ua+ub|0;if((ub|0)>(ja|0)){qb=vb;break}else tb=vb}}pb=qa+pb|0;if((pb|0)>(ga|0)){lb=qb;break}else ob=qb}}kb=e+kb|0;if((kb|0)>(n|0)){gb=lb;break}else jb=lb}}fb=ia+fb|0;if((fb|0)>(ma|0)){bb=gb;break}else eb=gb}}ab=c+ab|0;if((ab|0)>(s|0)){Za=bb;break}else $a=bb}}Ya=Ya+F|0;if((Ya|0)>(pa|0))break a;else Xa=Za}B=k[l>>2]|0;r=m;return B|0}while(0);B=k[l>>2]|0;r=m;return B|0}function vo(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;n=c==1.0;o=n&d==2147483648.0;if(!e){if(o){if((b|0)>0)p=0;else{q=k[m>>2]|0;return q|0}do{k[l+(p<<2)>>2]=k[a+(p<<2)>>2]^-2147483648;p=p+1|0}while((p|0)!=(b|0));q=k[m>>2]|0;return q|0}p=(b|0)>0;if(n&d==0.0){if(p)r=0;else{q=k[m>>2]|0;return q|0}do{s=k[a+(r<<2)>>2]|0;if((s|0)<0){k[m>>2]=-11;t=0}else t=s;k[l+(r<<2)>>2]=t;r=r+1|0}while((r|0)!=(b|0));q=k[m>>2]|0;return q|0}if(p)u=0;else{q=k[m>>2]|0;return q|0}do{v=+(k[a+(u<<2)>>2]|0)*c+d;do if(!(v<-.49))if(v>18446744073709549568.0){k[m>>2]=-11;k[l+(u<<2)>>2]=-1;break}else{k[l+(u<<2)>>2]=~~v>>>0;break}else{k[m>>2]=-11;k[l+(u<<2)>>2]=0}while(0);u=u+1|0}while((u|0)!=(b|0));q=k[m>>2]|0;return q|0}if(o){if((b|0)<=0){q=k[m>>2]|0;return q|0}if((e|0)==1){o=0;do{u=k[a+(o<<2)>>2]|0;if((u|0)==(f|0)){k[j>>2]=1;k[l+(o<<2)>>2]=g}else k[l+(o<<2)>>2]=u^-2147483648;o=o+1|0}while((o|0)!=(b|0));q=k[m>>2]|0;return q|0}else{o=0;do{u=k[a+(o<<2)>>2]|0;if((u|0)==(f|0)){k[j>>2]=1;i[h+o>>0]=1}else k[l+(o<<2)>>2]=u^-2147483648;o=o+1|0}while((o|0)!=(b|0));q=k[m>>2]|0;return q|0}}o=(b|0)>0;if(!(n&d==0.0)){if(!o){q=k[m>>2]|0;return q|0}n=(e|0)==1;u=0;do{p=k[a+(u<<2)>>2]|0;do if((p|0)==(f|0)){k[j>>2]=1;if(n){k[l+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else{v=+(p|0)*c+d;if(v<-.49){k[m>>2]=-11;k[l+(u<<2)>>2]=0;break}if(v>18446744073709549568.0){k[m>>2]=-11;k[l+(u<<2)>>2]=-1;break}else{k[l+(u<<2)>>2]=~~v>>>0;break}}while(0);u=u+1|0}while((u|0)!=(b|0));q=k[m>>2]|0;return q|0}if(!o){q=k[m>>2]|0;return q|0}if((e|0)==1){e=0;do{o=k[a+(e<<2)>>2]|0;do if((o|0)!=(f|0))if((o|0)<0){k[m>>2]=-11;k[l+(e<<2)>>2]=0;break}else{k[l+(e<<2)>>2]=o;break}else{k[j>>2]=1;k[l+(e<<2)>>2]=g}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}else{e=0;do{g=k[a+(e<<2)>>2]|0;do if((g|0)!=(f|0))if((g|0)<0){k[m>>2]=-11;k[l+(e<<2)>>2]=0;break}else{k[l+(e<<2)>>2]=g;break}else{k[j>>2]=1;i[h+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));q=k[m>>2]|0;return q|0}return 0}function wo(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e)if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=a+(q<<3)|0;t=k[s>>2]|0;u=k[s+4>>2]|0;do if((u|0)>=0)if((u|0)>0|(u|0)==0&t>>>0>4294967295){k[n>>2]=-11;k[m+(q<<2)>>2]=-1;break}else{k[m+(q<<2)>>2]=t;break}else{k[n>>2]=-11;k[m+(q<<2)>>2]=0}while(0);q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}else{if(p)v=0;else{r=k[n>>2]|0;return r|0}do{q=a+(v<<3)|0;w=(+((k[q>>2]|0)>>>0)+4294967296.0*+(k[q+4>>2]|0))*c+d;do if(!(w<-.49))if(w>18446744073709549568.0){k[n>>2]=-11;k[m+(v<<2)>>2]=-1;break}else{k[m+(v<<2)>>2]=~~w>>>0;break}else{k[n>>2]=-11;k[m+(v<<2)>>2]=0}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;v=0;do{q=a+(v<<3)|0;t=k[q>>2]|0;u=k[q+4>>2]|0;do if((t|0)==(f|0)&(u|0)==(g|0)){k[l>>2]=1;if(o){k[m+(v<<2)>>2]=h;break}else{i[j+v>>0]=1;break}}else{w=(+(t>>>0)+4294967296.0*+(u|0))*c+d;if(w<-.49){k[n>>2]=-11;k[m+(v<<2)>>2]=0;break}if(w>18446744073709549568.0){k[n>>2]=-11;k[m+(v<<2)>>2]=-1;break}else{k[m+(v<<2)>>2]=~~w>>>0;break}}while(0);v=v+1|0}while((v|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){e=0;do{p=a+(e<<3)|0;v=k[p>>2]|0;o=k[p+4>>2]|0;do if(!((v|0)==(f|0)&(o|0)==(g|0))){if((o|0)<0){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if((o|0)>0|(o|0)==0&v>>>0>4294967295){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=v;break}}else{k[l>>2]=1;k[m+(e<<2)>>2]=h}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}else{e=0;do{h=a+(e<<3)|0;v=k[h>>2]|0;o=k[h+4>>2]|0;do if(!((v|0)==(f|0)&(o|0)==(g|0))){if((o|0)<0){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if((o|0)>0|(o|0)==0&v>>>0>4294967295){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=v;break}}else{k[l>>2]=1;i[j+e>>0]=1}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}return 0}function xo(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{k[m+(q<<2)>>2]=l[a+q>>0];q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)s=0;else{r=k[n>>2]|0;return r|0}do{t=+(l[a+s>>0]|0|0)*c+d;do if(!(t<-.49))if(t>18446744073709549568.0){k[n>>2]=-11;k[m+(s<<2)>>2]=-1;break}else{k[m+(s<<2)>>2]=~~t>>>0;break}else{k[n>>2]=-11;k[m+(s<<2)>>2]=0}while(0);s=s+1|0}while((s|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}if((e|0)==1){o=0;do{s=i[a+o>>0]|0;if(s<<24>>24==f<<24>>24){k[j>>2]=1;k[m+(o<<2)>>2]=g}else k[m+(o<<2)>>2]=s&255;o=o+1|0}while((o|0)!=(b|0));r=k[n>>2]|0;return r|0}else{o=0;do{s=i[a+o>>0]|0;if(s<<24>>24==f<<24>>24){k[j>>2]=1;i[h+o>>0]=1}else k[m+(o<<2)>>2]=s&255;o=o+1|0}while((o|0)!=(b|0));r=k[n>>2]|0;return r|0}}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{o=i[a+e>>0]|0;do if(o<<24>>24==f<<24>>24){k[j>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{t=+(o&255|0)*c+d;if(t<-.49){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if(t>18446744073709549568.0){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=~~t>>>0;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function yo(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0;o=c==1.0&d==0.0;p=(b|0)>0;if(!e){if(o){if(p)q=0;else{r=k[n>>2]|0;return r|0}do{s=j[a+(q<<1)>>1]|0;if(s<<16>>16<0){k[n>>2]=-11;t=0}else t=s<<16>>16;k[m+(q<<2)>>2]=t;q=q+1|0}while((q|0)!=(b|0));r=k[n>>2]|0;return r|0}if(p)u=0;else{r=k[n>>2]|0;return r|0}do{v=+(j[a+(u<<1)>>1]|0)*c+d;do if(!(v<-.49))if(v>18446744073709549568.0){k[n>>2]=-11;k[m+(u<<2)>>2]=-1;break}else{k[m+(u<<2)>>2]=~~v>>>0;break}else{k[n>>2]=-11;k[m+(u<<2)>>2]=0}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(o){if(!p){r=k[n>>2]|0;return r|0}o=(e|0)==1;u=0;do{q=j[a+(u<<1)>>1]|0;t=q<<16>>16;do if(q<<16>>16==f<<16>>16){k[l>>2]=1;if(o){k[m+(u<<2)>>2]=g;break}else{i[h+u>>0]=1;break}}else if(q<<16>>16<0){k[n>>2]=-11;k[m+(u<<2)>>2]=0;break}else{k[m+(u<<2)>>2]=t;break}while(0);u=u+1|0}while((u|0)!=(b|0));r=k[n>>2]|0;return r|0}if(!p){r=k[n>>2]|0;return r|0}p=(e|0)==1;e=0;do{u=j[a+(e<<1)>>1]|0;do if(u<<16>>16==f<<16>>16){k[l>>2]=1;if(p){k[m+(e<<2)>>2]=g;break}else{i[h+e>>0]=1;break}}else{v=+(u<<16>>16)*c+d;if(v<-.49){k[n>>2]=-11;k[m+(e<<2)>>2]=0;break}if(v>18446744073709549568.0){k[n>>2]=-11;k[m+(e<<2)>>2]=-1;break}else{k[m+(e<<2)>>2]=~~v>>>0;break}}while(0);e=e+1|0}while((e|0)!=(b|0));r=k[n>>2]|0;return r|0}function zo(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,p=0,q=0,r=0.0,s=0.0,t=0,u=0,v=0,w=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)p=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(p<<2)>>2];s=r;do if(!(s<-.49))if(s>18446744073709549568.0){k[l>>2]=-11;k[j+(p<<2)>>2]=-1;break}else{k[j+(p<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(p<<2)>>2]=0}while(0);p=p+1|0}while((p|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)t=0;else{q=k[l>>2]|0;return q|0}do{r=+o[a+(t<<2)>>2]*c+d;do if(!(r<-.49))if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(t<<2)>>2]=-1;break}else{k[j+(t<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(t<<2)>>2]=0}while(0);t=t+1|0}while((t|0)!=(b|0));q=k[l>>2]|0;return q|0}}t=a+2|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}p=(e|0)==1;u=0;v=t;while(1){w=(m[v>>1]|0)&32640;a:do switch(((w|0)==32640?1:(w|0)==0?2:0)|0){case 0:{r=+o[a+(u<<2)>>2];s=r;if(s<-.49){k[l>>2]=-11;k[j+(u<<2)>>2]=0;break a}if(s>18446744073709549568.0){k[l>>2]=-11;k[j+(u<<2)>>2]=-1;break a}else{k[j+(u<<2)>>2]=~~r>>>0;break a}break}case 1:{k[h>>2]=1;if(p){k[j+(u<<2)>>2]=f;break a}else{i[g+u>>0]=1;break a}break}default:k[j+(u<<2)>>2]=0}while(0);u=u+1|0;if((u|0)==(b|0))break;else v=v+4|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;v=d>18446744073709549568.0;u=~~d>>>0;p=(e|0)==1;e=0;w=t;while(1){t=(m[w>>1]|0)&32640;b:do switch(((t|0)==32640?1:(t|0)==0?2:0)|0){case 0:{r=+o[a+(e<<2)>>2]*c+d;if(r<-.49){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=~~r>>>0;break b}break}case 1:{k[h>>2]=1;if(p){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(v){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=u;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else w=w+4|0}q=k[l>>2]|0;return q|0}function Ao(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var n=0,o=0,q=0,r=0.0,s=0,t=0,u=0,v=0;if(!e){n=(b|0)>0;if(c==1.0&d==0.0){if(n)o=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(o<<3)>>3];do if(!(r<-.49))if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(o<<2)>>2]=-1;break}else{k[j+(o<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(o<<2)>>2]=0}while(0);o=o+1|0}while((o|0)!=(b|0));q=k[l>>2]|0;return q|0}else{if(n)s=0;else{q=k[l>>2]|0;return q|0}do{r=+p[a+(s<<3)>>3]*c+d;do if(!(r<-.49))if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(s<<2)>>2]=-1;break}else{k[j+(s<<2)>>2]=~~r>>>0;break}else{k[l>>2]=-11;k[j+(s<<2)>>2]=0}while(0);s=s+1|0}while((s|0)!=(b|0));q=k[l>>2]|0;return q|0}}s=a+6|0;n=(b|0)>0;if(c==1.0&d==0.0){if(!n){q=k[l>>2]|0;return q|0}o=(e|0)==1;t=0;u=s;while(1){v=(m[u>>1]|0)&32752;a:do switch(((v|0)==32752?1:(v|0)==0?2:0)|0){case 0:{r=+p[a+(t<<3)>>3];if(r<-.49){k[l>>2]=-11;k[j+(t<<2)>>2]=0;break a}if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(t<<2)>>2]=-1;break a}else{k[j+(t<<2)>>2]=~~r>>>0;break a}break}case 1:{k[h>>2]=1;if(o){k[j+(t<<2)>>2]=f;break a}else{i[g+t>>0]=1;break a}break}default:k[j+(t<<2)>>2]=0}while(0);t=t+1|0;if((t|0)==(b|0))break;else u=u+8|0}q=k[l>>2]|0;return q|0}if(!n){q=k[l>>2]|0;return q|0}n=d<-.49;u=d>18446744073709549568.0;t=~~d>>>0;o=(e|0)==1;e=0;v=s;while(1){s=(m[v>>1]|0)&32752;b:do switch(((s|0)==32752?1:(s|0)==0?2:0)|0){case 0:{r=+p[a+(e<<3)>>3]*c+d;if(r<-.49){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(r>18446744073709549568.0){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=~~r>>>0;break b}break}case 1:{k[h>>2]=1;if(o){k[j+(e<<2)>>2]=f;break b}else{i[g+e>>0]=1;break b}break}default:{if(n){k[l>>2]=-11;k[j+(e<<2)>>2]=0;break b}if(u){k[l>>2]=-11;k[j+(e<<2)>>2]=-1;break b}else{k[j+(e<<2)>>2]=t;break b}}}while(0);e=e+1|0;if((e|0)==(b|0))break;else v=v+8|0}q=k[l>>2]|0;return q|0}function Bo(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=+f;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0.0,P=0.0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0.0,Z=0,_=0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0,ga=0.0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0;p=r;r=r+96|0;q=p;s=p+4|0;t=xx(h)|0;a:do if((b|0)>0){u=(g|0)==0;v=(g|0)==1;w=a;x=0;b:while(1){y=w+e|0;z=i[y>>0]|0;i[y>>0]=0;do if((i[h>>0]|0)!=1?(Ax(h,w,t)|0)==0:0)if(!u){k[m>>2]=1;if(v){k[n+(x<<2)>>2]=j;A=y;break}else{i[l+x>>0]=1;A=y;break}}else A=y;else{B=w;C=9}while(0);do if((C|0)==9){c:while(1){C=0;D=i[B>>0]|0;switch(D<<24>>24){case 43:case 45:{E=D;F=B;C=11;break c;break}case 32:break;default:{G=D;H=B;I=1;break c}}B=B+1|0;C=9}if((C|0)==11){C=0;D=E<<24>>24==45;J=F;while(1){K=J+1|0;L=i[K>>0]|0;if(L<<24>>24==32)J=K;else{M=L;N=K;break}}G=M;H=N;I=D?-1:1}if((G+-48&255)<10){J=G;K=H;O=0.0;while(1){P=O*10.0+ +(J<<24>>24);L=K;while(1){Q=L+1|0;R=i[Q>>0]|0;if(R<<24>>24==32)L=Q;else{S=R;T=Q;break}}U=P+-48.0;if((S+-48&255)<10){J=S;K=T;O=U}else{V=S;W=T;X=U;break}}}else{V=G;W=H;X=0.0}switch(V<<24>>24){case 44:case 46:{K=W;while(1){J=K+1|0;D=i[J>>0]|0;if(D<<24>>24==32)K=J;else{Z=D;_=J;break}}if((Z+-48&255)<10){K=Z;J=_;O=1.0;U=X;while(1){$=U*10.0+ +(K<<24>>24)+-48.0;D=J;while(1){L=D+1|0;Q=i[L>>0]|0;if(Q<<24>>24==32)D=L;else{aa=Q;ba=L;break}}P=O*10.0;if((aa+-48&255)<10){K=aa;J=ba;O=P;U=$}else{ca=aa;da=ba;ea=1;fa=P;ga=$;break}}}else{ca=Z;da=_;ea=1;fa=1.0;ga=X}break}default:{ca=V;da=W;ea=0;fa=1.0;ga=X}}if((ca&-2)<<24>>24==68){J=da;d:while(1){K=J+1|0;D=i[K>>0]|0;switch(D<<24>>24){case 32:{J=K;break}case 43:case 45:{ia=D;ja=J;C=28;break d;break}default:{ka=D;la=K;ma=1;break d}}}if((C|0)==28){C=0;J=ia<<24>>24==45;K=ja+2|0;while(1){D=i[K>>0]|0;if(D<<24>>24==32)K=K+1|0;else{na=D;oa=K;break}}ka=na;la=oa;ma=J?-1:1}if((ka+-48&255)<10){K=ka;D=la;L=0;while(1){Q=(L*10|0)+-48|0;R=D;while(1){pa=R+1|0;qa=i[pa>>0]|0;if(qa<<24>>24==32)R=pa;else{ra=qa;sa=pa;break}}R=Q+(K<<24>>24)|0;if((ra+-48&255)<10){K=ra;D=sa;L=R}else{ta=ra;ua=sa;va=ma;wa=R;break}}}else{ta=ka;ua=la;va=ma;wa=0}}else{ta=ca;ua=da;va=1;wa=0}if(ta<<24>>24){xa=y;ya=z;za=w;break b}$=+(I|0)*ga/((ea|0)==0?f:fa)*+Y(10.0,+(+(ha(va,wa)|0)))*c+d;if($<-.49){k[o>>2]=-11;k[n+(x<<2)>>2]=0;A=ua;break}if($>18446744073709549568.0){k[o>>2]=-11;k[n+(x<<2)>>2]=-1;A=ua;break}else{k[n+(x<<2)>>2]=~~$>>>0;A=ua;break}}while(0);i[y>>0]=z;x=x+1|0;if((x|0)>=(b|0))break a;else w=A}w=s;x=761193;v=w+36|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(v|0));tk(s);k[q>>2]=za;Xw(s,761229,q)|0;tk(s);i[xa>>0]=ya;k[o>>2]=409;Aa=409;r=p;return Aa|0}while(0);Aa=k[o>>2]|0;r=p;return Aa|0}function Co(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}e=k[a>>2]|0;g=a+4|0;if((e|0)!=(k[(k[g>>2]|0)+68>>2]|0))Fk(a,e+1|0,0,d)|0;e=k[g>>2]|0;if(b){g=e+96|0;a=(k[e+88>>2]|0)+(k[e+68>>2]<<3)|0;h=jy(k[g>>2]|0,k[g+4>>2]|0,k[a>>2]|0,k[a+4>>2]|0)|0;a=wy(h|0,L|0,80,0)|0;k[b>>2]=a}a=e+120|0;b=k[a>>2]|0;h=k[a+4>>2]|0;a=(c|0)!=0;if((b|0)==-1&(h|0)==-1){if(a)k[c>>2]=-1}else if(a){a=e+96|0;e=jy(b|0,h|0,k[a>>2]|0,k[a+4>>2]|0)|0;a=wy(e|0,L|0,80,0)|0;e=ky(a|0,L|0,-1,0)|0;k[c>>2]=e}f=k[d>>2]|0;return f|0}function Do(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[a>>2]|0;e=a+4|0;f=k[e>>2]|0;if((d|0)==(k[f+68>>2]|0)){g=d;h=f}else{Fk(a,d+1|0,0,c)|0;d=k[e>>2]|0;g=k[d+68>>2]|0;h=d}d=(k[h+88>>2]|0)+(g<<3)|0;g=(b*80|0)+-80|0;b=ky(k[d>>2]|0,k[d+4>>2]|0,g|0,((g|0)<0)<<31>>31|0)|0;g=h+112|0;k[g>>2]=b;k[g+4>>2]=L;return k[c>>2]|0}function Eo(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+96|0;e=d;f=d+4|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=k[a>>2]|0;j=a+4|0;if((g|0)!=(k[(k[j>>2]|0)+68>>2]|0))Fk(a,g+1|0,0,c)|0;i[b>>0]=0;g=k[j>>2]|0;l=g+112|0;m=k[l>>2]|0;n=k[l+4>>2]|0;l=g+96|0;o=k[l>>2]|0;p=k[l+4>>2]|0;l=g+120|0;q=ky(k[l>>2]|0,k[l+4>>2]|0,-2880,-1)|0;l=L;s=(p|0)>(l|0)|(p|0)==(l|0)&o>>>0>q>>>0;t=s?p:l;l=(k[g+88>>2]|0)+(k[g+68>>2]<<3)|0;g=k[l>>2]|0;p=k[l+4>>2]|0;if((n|0)>(t|0)|(n|0)==(t|0)&m>>>0>(s?o:q)>>>0|((n|0)<(p|0)|(n|0)==(p|0)&m>>>0<g>>>0)){q=jy(m|0,n|0,g|0,p|0)|0;p=wy(q|0,L|0,80,0)|0;q=ky(p|0,L|0,1,0)|0;k[e>>2]=q;Xw(f,761248,e)|0;tk(f);k[c>>2]=203;h=203;r=d;return h|0}au(a,m,n,0,c)|0;i[b+80>>0]=0;if((fu(a,80,0,b,c)|0)<1){a=(k[j>>2]|0)+112|0;j=a;n=ky(k[j>>2]|0,k[j+4>>2]|0,80,0)|0;j=a;k[j>>2]=n;k[j+4>>2]=L;j=79;while(1){if((i[b+j>>0]|0)!=32){u=j;break}n=j+-1|0;if((j|0)>0)j=n;else{u=n;break}}i[b+(u+1)>>0]=0}h=k[c>>2]|0;r=d;return h|0}function Fo(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;h=r;r=r+176|0;j=h+8|0;l=h+4|0;m=h;n=h+87|0;o=h+12|0;i[f>>0]=0;p=k[g>>2]|0;if((p|0)>0){q=p;r=h;return q|0}a:do if((Go(a,782783,n,g)|0)<1){p=(c|0)>0;b:while(1){Ho(n,o,m,g)|0;if(p){s=0;do{Wk(k[b+(s<<2)>>2]|0,o,0,j,l);if(k[j>>2]|0){t=-1;do{t=t+1|0;if((t|0)>=(e|0))break b;Wk(k[d+(t<<2)>>2]|0,o,0,j,l)}while((k[j>>2]|0)==0)}s=s+1|0}while((s|0)<(c|0))}if((Go(a,782783,n,g)|0)>=1)break a}qx(f,n)|0;q=k[g>>2]|0;r=h;return q|0}while(0);q=k[g>>2]|0;r=h;return q|0}function Go(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0;e=r;r=r+176|0;f=e+8|0;g=e+4|0;h=e;j=e+87|0;l=e+12|0;m=k[d>>2]|0;if((m|0)>0){n=m;r=e;return n|0}i[j>>0]=0;m=0;while(1){o=b+m|0;if((i[o>>0]|0)==32)m=m+1|0;else{p=o;break}}zx(j,p,74)|0;p=xx(j)|0;while(1){if((p|0)<=0){q=p;s=7;break}m=p+-1|0;if((i[j+m>>0]|0)==32)p=m;else{t=p;s=8;break}}if((s|0)==7){i[j+q>>0]=0;u=q}else if((s|0)==8){i[j+t>>0]=0;q=0;do{p=j+q|0;m=(nv(i[p>>0]|0)|0)&255;i[p>>0]=m;q=q+1|0}while((q|0)!=(t|0));u=t}if((i[j>>0]|0)==72?(Ax(767038,j,8)|0)==0:0)if((u|0)==8){v=1;w=8;s=18}else{t=j+8|0;a:while(1){q=i[t>>0]|0;switch(q<<24>>24){case 32:{t=t+1|0;break}case 0:{x=0;y=j;break a;break}default:{z=q;A=0;B=j;C=t;s=15;break a}}}if((s|0)==15)while(1){s=0;i[B>>0]=z;t=B+1|0;C=C+1|0;q=A+1|0;z=i[C>>0]|0;if(!(z<<24>>24)){x=q;y=t;break}else{A=q;B=t;s=15}}i[y>>0]=0;D=x;s=17}else{D=u;s=17}if((s|0)==17)if((D|0)<9){v=0;w=D;s=18}else{E=0;F=D;G=D;H=0}do if((s|0)==18){if(!(rx(j,63)|0)){D=rx(j,42)|0;if(!D)if(!(rx(j,35)|0)){E=v;F=w;G=w;H=0;break}else I=0;else I=D}else I=0;E=v;F=w;G=(I|0)==0?w:w+-2|0;H=1}while(0);if((k[d>>2]|0)>0){J=a+4|0;K=0;M=0}else{w=k[a>>2]|0;I=a+4|0;v=k[I>>2]|0;if((w|0)==(k[v+68>>2]|0)){N=v;O=w}else{Fk(a,w+1|0,0,d)|0;w=k[I>>2]|0;N=w;O=k[w+68>>2]|0}w=N+96|0;v=(k[N+88>>2]|0)+(O<<3)|0;O=k[v>>2]|0;D=k[v+4>>2]|0;v=jy(k[w>>2]|0,k[w+4>>2]|0,O|0,D|0)|0;w=wy(v|0,L|0,80,0)|0;v=N+112|0;N=jy(k[v>>2]|0,k[v+4>>2]|0,O|0,D|0)|0;D=wy(N|0,L|0,80,0)|0;N=ky(D|0,L|0,1,0)|0;J=I;K=w+1|0;M=N}N=F+-1|0;w=(N|0)>1?N:1;N=(E|0)==0;E=(H|0)!=0;H=j+w|0;I=l+w|0;D=M+-1|0;O=0;v=K-M|0;b:while(1){c:do if((v|0)>0){if(N)P=0;else{M=0;while(1){Eo(a,c,d)|0;if((i[c>>0]|0)==72?(Ax(767038,c,8)|0)==0:0){s=44;break b}M=M+1|0;if((M|0)>=(v|0))break c}}do{Eo(a,c,d)|0;Ho(c,l,f,d)|0;M=k[f>>2]|0;do if((M|0)>=(G|0)&(E|(M|0)==(F|0))){if((M|0)>0){K=0;do{u=l+K|0;x=i[u>>0]|0;if(x<<24>>24>96){y=(nv(x<<24>>24)|0)&255;i[u>>0]=y}K=K+1|0}while((K|0)<(M|0))}if(E){Wk(j,l,1,g,h);if(!(k[g>>2]|0))break;else{s=45;break b}}if((i[H>>0]|0)!=(i[I>>0]|0))if(!(M|F)){s=47;break b}else break;if((i[j>>0]|0)==(i[l>>0]|0)?(Ax(j,l,w)|0)==0:0){s=46;break b}}while(0);P=P+1|0}while((P|0)<(v|0))}while(0);if(E|(O|0)==1){s=53;break}M=k[a>>2]|0;K=k[J>>2]|0;if((M|0)==(k[K+68>>2]|0)){Q=K;R=M}else{Fk(a,M+1|0,0,d)|0;M=k[J>>2]|0;Q=M;R=k[M+68>>2]|0}M=(k[Q+88>>2]|0)+(R<<3)|0;K=k[M+4>>2]|0;y=Q+112|0;k[y>>2]=k[M>>2];k[y+4>>2]=K;K=O+1|0;if((K|0)<2){O=K;v=D}else{s=53;break}}if((s|0)==44){n=k[d>>2]|0;r=e;return n|0}else if((s|0)==45){n=k[d>>2]|0;r=e;return n|0}else if((s|0)==46){n=k[d>>2]|0;r=e;return n|0}else if((s|0)==47){n=k[d>>2]|0;r=e;return n|0}else if((s|0)==53){k[d>>2]=202;n=202;r=e;return n|0}return 0}
function yh(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;h=r;r=r+32|0;l=h+16|0;m=h;n=(d|0)<(g|0)?d:g;q=(n|0)<1?1:n;n=d+-1|0;s=(n+q|0)/(q|0)|0;t=(s|0)<2?2:s;s=(n+t|0)/(t|0)|0;n=(s|0)<1?1:s;if(!(k[25364]|0))u=1;else{k[m>>2]=q;k[m+4>>2]=t;k[m+8>>2]=n;Rw(725844,m)|0;u=(k[25364]|0)==0}m=(f|0)/(g|0)|0;g=(m|0)<1?1:m;m=(f+-1+n|0)/(n|0)|0;f=(m|0)>(e|0)?e:m;m=(g|0)>(f|0)?g:f;f=(e|0)/(m|0)|0;q=(f|0)<2?2:f;s=(e+-1+q|0)/(q|0)|0;if(!u){k[l>>2]=g;k[l+4>>2]=m;k[l+8>>2]=q;k[l+12>>2]=s;Rw(725898,l)|0}l=ha(s,n)|0;s=cy(l<<2)|0;k[c>>2]=s;c=d<<2;m=cy(c)|0;g=(q+1|0)/2|0;if((g|0)>=(e|0)){v=0;dy(m);r=h;return v|0}u=(n|0)>0;w=(d|0)>0;x=d<<1;y=d<<3;if((b|0)==8){z=g;A=0;B=s;while(1){C=ha(z+-1|0,d)|0;if(w){D=0;do{o[m+(D<<2)>>2]=+(i[a+(D+C)>>0]|0);D=D+1|0}while((D|0)!=(d|0))}if(u){D=0;C=0;while(1){k[B+(D<<2)>>2]=k[m+(C<<2)>>2];D=D+1|0;if((D|0)==(n|0))break;else C=C+t|0}}C=A+n|0;if((C|0)>(l|0)){v=C;E=27;break}z=z+q|0;if((z|0)>=(e|0)){v=C;E=27;break}else{A=C;B=B+(n<<2)|0}}if((E|0)==27){dy(m);r=h;return v|0}}B=(ha(g+-1|0,d)|0)<<2;A=(ha((f|0)>2?f:2,d)|0)<<2;f=d<<2;z=0;C=g;g=0;D=s;while(1){s=a+(B+(ha(A,z)|0))|0;switch(b|0){case -64:{F=a+(ha(y,C+-1|0)|0)|0;if(w){G=0;do{o[m+(G<<2)>>2]=+p[F+(G<<3)>>3];G=G+1|0}while((G|0)!=(d|0))}break}case 16:{G=a+(ha(x,C+-1|0)|0)|0;if(w){F=0;do{o[m+(F<<2)>>2]=+(j[G+(F<<1)>>1]|0);F=F+1|0}while((F|0)!=(d|0))}break}case 32:{F=a+(ha(c,C+-1|0)|0)|0;if(w){G=0;do{o[m+(G<<2)>>2]=+(k[F+(G<<2)>>2]|0);G=G+1|0}while((G|0)!=(d|0))}break}case -32:{if(w)py(m|0,s|0,f|0)|0;break}default:{}}if(u){G=0;F=0;while(1){k[D+(G<<2)>>2]=k[m+(F<<2)>>2];G=G+1|0;if((G|0)==(n|0))break;else F=F+t|0}}F=g+n|0;if((F|0)>(l|0)){v=F;E=27;break}C=C+q|0;if((C|0)>=(e|0)){v=F;E=27;break}else{z=z+1|0;g=F;D=D+(n<<2)|0}}if((E|0)==27){dy(m);r=h;return v|0}return 0}function zh(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=+e;f=f|0;g=g|0;var h=0.0,j=0,l=0,m=0,n=0,p=0,q=0.0,r=0.0,s=0.0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0,C=0,D=0,E=0,F=0,G=0,H=0.0,I=0.0,J=0,K=0.0,L=0.0,M=0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0,V=0,W=0,Y=0.0,Z=0,_=0,$=0.0,aa=0.0,ba=0.0,ca=0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ia=0,ja=0.0,ka=0.0,la=0.0,ma=0.0,na=0.0,oa=0.0,pa=0,qa=0.0,ra=0.0,sa=0.0,ta=0.0,ua=0,va=0.0,wa=0.0,xa=0.0,ya=0,za=0.0;h=e;if((b|0)<1){j=0;return j|0}if((b|0)==1){k[c>>2]=k[a+4>>2];o[d>>2]=0.0;j=1;return j|0}e=2.0/+(b+-1|0);l=b<<2;m=cy(l)|0;n=cy(l)|0;l=ey(b,1)|0;p=0;do{o[n+(p<<2)>>2]=e*+(p|0)+-1.0;p=p+1|0}while((p|0)!=(b|0));q=0.0;r=0.0;s=0.0;t=0;while(1){u=+o[n+(t<<2)>>2];v=+o[a+(t<<2)>>2];w=q+u*u;x=r+u*v;u=s+v;t=t+1|0;if((t|0)==(b|0)){y=w;z=x;A=u;break}else{q=w;r=x;s=u}}s=+(b|0);r=A/s;q=z/y;t=~~(s*.5);p=(t|0)<5?5:t;if((g|0)>0){s=h;t=~f;B=~b;h=0.0;u=A;A=z;z=y;y=q;C=b;D=0;x=r;while(1){E=0;do{o[m+(E<<2)>>2]=+o[a+(E<<2)>>2]-(x+y*+o[n+(E<<2)>>2]);E=E+1|0}while((E|0)!=(b|0));F=0;G=0;H=0.0;I=0.0;while(1){if(!(i[l+F>>0]|0)){w=+o[m+(F<<2)>>2];J=G+1|0;K=H+w;L=I+w*w}else{J=G;K=H;L=I}F=F+1|0;if((F|0)==(b|0)){M=J;N=K;O=L;break}else{G=J;H=K;I=L}}if(M>>>0>=2){E=M+-1|0;w=O/+(E|0)-N*N/+(ha(E,M)|0);if(w<0.0)P=0.0;else P=+X(+w)}else P=-999.0;w=s*P;v=-w;Q=z;R=A;S=u;T=h;E=0;U=t;V=b;while(1){W=~((U|0)>(B|0)?U:B);if((i[l+E>>0]|0)!=1){Y=+o[m+(E<<2)>>2];if(Y<v|Y>w?(Z=E-f|0,_=(Z|0)<0?0:Z,Z=E+f|0,(_|0)<(((Z|0)>(b|0)?b:Z)|0)):0){Y=Q;$=R;aa=T;ba=S;Z=_;_=V;while(1){ca=l+Z|0;do if((i[ca>>0]|0)!=1)if((Z|0)>(E|0)){i[ca>>0]=2;da=Y;ea=$;fa=ba;ga=aa;ia=_;break}else{ja=+o[n+(Z<<2)>>2];ka=+o[a+(Z<<2)>>2];i[ca>>0]=1;da=Y-ja*ja;ea=$-ja*ka;fa=ba-ka;ga=aa-ja;ia=_+-1|0;break}else{da=Y;ea=$;fa=ba;ga=aa;ia=_}while(0);Z=Z+1|0;if((Z|0)==(W|0)){la=da;ma=ea;na=fa;oa=ga;pa=ia;break}else{Y=da;$=ea;aa=ga;ba=fa;_=ia}}}else{la=Q;ma=R;na=S;oa=T;pa=V}}else{la=Q;ma=R;na=S;oa=T;pa=V+-1|0}E=E+1|0;if((E|0)==(b|0)){qa=la;ra=ma;sa=na;ta=oa;ua=pa;break}else{Q=la;R=ma;S=na;T=oa;U=U+-1|0;V=pa}}if((ua|0)>0){T=ta/qa;S=(sa-ra*T)/(+(ua|0)-ta*T);va=(ra-ta*S)/qa;wa=S}else{va=y;wa=x}V=D+1|0;if((ua|0)>=(p|0)&(ua|0)<(C|0)&(V|0)<(g|0)){h=ta;u=sa;A=ra;z=qa;y=va;C=ua;D=V;x=wa}else{xa=va;ya=ua;za=wa;break}}}else{xa=q;ya=b;za=r}o[c>>2]=za-xa;za=e*xa;o[d>>2]=(za>=0.0?za:-za)<.001?e*q:za;dy(m);dy(n);dy(l);j=ya;return j|0}function Ah(a,b){a=a|0;b=b|0;return (+o[a>>2]<=+o[b>>2]?-1:1)|0}function Bh(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0;j=r;r=r+7488|0;l=j+32|0;m=j+24|0;n=j+16|0;o=j+8|0;p=j+52|0;q=j+48|0;s=j+44|0;t=j+40|0;u=j+36|0;v=j+7405|0;w=j;x=j+1164|0;y=j+6380|0;z=j+5355|0;A=j+4330|0;B=j+3305|0;C=j+2280|0;D=j+1255|0;E=j+1184|0;F=j+137|0;G=j+56|0;H=k[h>>2]|0;if((H|0)>0){I=H;r=j;return I|0}k[a>>2]=0;if((k[28269]|0)!=0?(H=Ch()|0,k[h>>2]=H,(H|0)>0):0){I=H;r=j;return I|0}else J=b;while(1)if((i[J>>0]|0)==32)J=J+1|0;else{K=J;break}Jh(K,x,y,z,A,B,C,D,0,0,h)|0;D=x;C=765414;B=D+11|0;do{i[D>>0]=i[C>>0]|0;D=D+1|0;C=C+1|0}while((D|0)<(B|0));C=k[28270]|0;while(1){D=C+-1|0;if((C|0)<=0){L=8;break}if(!(tx(113084+(D*84|0)|0,x)|0)){M=D;break}else C=D}if((L|0)==8){k[h>>2]=124;tk(725954);I=k[h>>2]|0;r=j;return I|0}k[h>>2]=0;L=Ti(d,e,f,g,p)|0;k[h>>2]=L;if((L|0)>0){tk(726015);I=k[h>>2]|0;r=j;return I|0}L=Xb[k[113084+(M*84|0)+64>>2]&31](k[p>>2]|0,w)|0;k[h>>2]=L;if((L|0)>0){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726065);I=k[h>>2]|0;r=j;return I|0}L=ey(1,8)|0;k[a>>2]=L;if(!L){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726114);tk(K);k[h>>2]=113;I=113;r=j;return I|0}g=ey(1,1728)|0;k[L+4>>2]=g;g=(k[a>>2]|0)+4|0;if(!(k[g>>2]|0)){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726114);tk(K);dy(k[a>>2]|0);k[a>>2]=0;k[h>>2]=113;I=113;r=j;return I|0}L=(xx(K)|0)+1|0;f=cy((L|0)>32?L:32)|0;k[(k[g>>2]|0)+12>>2]=f;f=(k[a>>2]|0)+4|0;if(!(k[(k[f>>2]|0)+12>>2]|0)){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726172);tk(K);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[h>>2]=113;I=113;r=j;return I|0}g=ey(1001,8)|0;k[(k[f>>2]|0)+88>>2]=g;g=(k[a>>2]|0)+4|0;if(!(k[(k[g>>2]|0)+88>>2]|0)){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726221);tk(K);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[h>>2]=113;I=113;r=j;return I|0}f=ey(40,2880)|0;k[(k[g>>2]|0)+1244>>2]=f;f=k[(k[a>>2]|0)+4>>2]|0;if(!(k[f+1244>>2]|0)){Ub[k[113084+(M*84|0)+56>>2]&31](k[p>>2]|0)|0;tk(726277);tk(K);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+88>>2]|0);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[h>>2]=113;I=113;r=j;return I|0}else N=0;do{k[f+1568+(N<<2)>>2]=N;k[f+1248+(N<<2)>>2]=-1;N=N+1|0}while((N|0)!=40);k[f+84>>2]=1e3;k[f>>2]=k[p>>2];k[f+4>>2]=M;ux(k[f+12>>2]|0,K)|0;f=w;M=k[f+4>>2]|0;p=k[a>>2]|0;N=k[p+4>>2]|0;g=N+24|0;k[g>>2]=k[f>>2];k[g+4>>2]=M;M=w;w=k[M+4>>2]|0;g=N+32|0;k[g>>2]=k[M>>2];k[g+4>>2]=w;k[N+76>>2]=c;c=N+120|0;k[c>>2]=-1;k[c+4>>2]=-1;k[N+64>>2]=-1;k[N+8>>2]=1;k[N+16>>2]=555;bu(p,0,0,h)|0;p=k[a>>2]|0;N=k[p+4>>2]|0;a:do if((k[h>>2]|0)>0)O=p;else{c=0;while(1){w=115100+(c<<2)|0;c=c+1|0;if(!(k[w>>2]|0)){P=w;break}if((c|0)>=1e3){O=p;break a}}k[P>>2]=N;O=k[a>>2]|0}while(0);if((dl(O,q,h)|0)>0){tk(726332);tk(K);if((k[h>>2]|0)==252)tk(737931);Fh(k[a>>2]|0,h)|0;k[a>>2]=0}i[E>>0]=0;i[F>>0]=0;if(!(i[A>>0]|0)){I=k[h>>2]|0;r=j;return I|0}Gh(A,u,v,t,s,E,F,h)|0;F=k[h>>2]|0;if((F|0)>0){I=F;r=j;return I|0}F=k[u>>2]|0;if(!F){if(i[v>>0]|0)Jl(k[a>>2]|0,k[s>>2]|0,v,k[t>>2]|0,h)|0}else Fk(k[a>>2]|0,F+1|0,q,h)|0;q=k[h>>2]|0;if((q|0)<=0){I=q;r=j;return I|0}tk(726398);if((F|0)>0){k[o>>2]=F;Xw(G,726448,o)|0;tk(G)}else{k[n>>2]=v;Xw(G,726506,n)|0;tk(G);n=k[t>>2]|0;if(n){k[m>>2]=n;Xw(G,726536,m)|0;tk(G)}m=k[s>>2]|0;if((m|0)!=-1){k[l>>2]=k[119100+(m<<2)>>2];Xw(G,726570,l)|0;tk(G)}tk(726605)}I=k[h>>2]|0;r=j;return I|0}function Ch(){var a=0,b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0;if(!(k[28269]|0)){a=0;return a|0}b=k[28270]|0;if((b|0)>=0)if((b|0)<=23){b=li()|0;if(!b){Bx(113084+((k[28270]|0)*84|0)|0,764609,20)|0;c=k[28270]|0;i[113084+(c*84|0)+19>>0]=0;k[113084+(c*84|0)+20>>2]=1;k[113084+(c*84|0)+24>>2]=2;k[113084+(c*84|0)+28>>2]=6;k[113084+(c*84|0)+32>>2]=7;k[113084+(c*84|0)+36>>2]=8;k[113084+(c*84|0)+40>>2]=8;k[113084+(c*84|0)+44>>2]=9;k[113084+(c*84|0)+48>>2]=10;k[113084+(c*84|0)+52>>2]=10;k[113084+(c*84|0)+56>>2]=9;k[113084+(c*84|0)+60>>2]=10;k[113084+(c*84|0)+64>>2]=11;k[113084+(c*84|0)+68>>2]=11;k[113084+(c*84|0)+72>>2]=11;k[113084+(c*84|0)+76>>2]=12;k[113084+(c*84|0)+80>>2]=13;k[28270]=c+1;if((c|0)>=-1)if((c|0)<=22){c=Mi()|0;if(!c){Bx(113084+((k[28270]|0)*84|0)|0,777975,20)|0;d=k[28270]|0;i[113084+(d*84|0)+19>>0]=0;k[113084+(d*84|0)+20>>2]=3;k[113084+(d*84|0)+24>>2]=4;k[113084+(d*84|0)+28>>2]=12;k[113084+(d*84|0)+32>>2]=13;k[113084+(d*84|0)+36>>2]=14;k[113084+(d*84|0)+40>>2]=0;k[113084+(d*84|0)+44>>2]=0;k[113084+(d*84|0)+48>>2]=12;k[113084+(d*84|0)+52>>2]=14;k[113084+(d*84|0)+56>>2]=15;k[113084+(d*84|0)+60>>2]=0;k[113084+(d*84|0)+64>>2]=13;k[113084+(d*84|0)+68>>2]=0;k[113084+(d*84|0)+72>>2]=15;k[113084+(d*84|0)+76>>2]=16;k[113084+(d*84|0)+80>>2]=17;e=d+1|0;k[28270]=e;if((d|0)>=-1){if((d|0)<=22){Bx(113084+(e*84|0)|0,765414,20)|0;e=k[28270]|0;i[113084+(e*84|0)+19>>0]=0;k[113084+(e*84|0)+20>>2]=0;k[113084+(e*84|0)+24>>2]=4;k[113084+(e*84|0)+28>>2]=12;k[113084+(e*84|0)+32>>2]=13;k[113084+(e*84|0)+36>>2]=14;k[113084+(e*84|0)+40>>2]=0;k[113084+(e*84|0)+44>>2]=0;k[113084+(e*84|0)+48>>2]=0;k[113084+(e*84|0)+52>>2]=14;k[113084+(e*84|0)+56>>2]=16;k[113084+(e*84|0)+60>>2]=0;k[113084+(e*84|0)+64>>2]=13;k[113084+(e*84|0)+68>>2]=0;k[113084+(e*84|0)+72>>2]=15;k[113084+(e*84|0)+76>>2]=16;k[113084+(e*84|0)+80>>2]=17;d=e+1|0;k[28270]=d;if((e|0)>=-1){if((e|0)<=22){Bx(113084+(d*84|0)|0,765501,20)|0;d=k[28270]|0;i[113084+(d*84|0)+19>>0]=0;k[113084+(d*84|0)+20>>2]=0;k[113084+(d*84|0)+24>>2]=4;k[113084+(d*84|0)+28>>2]=12;k[113084+(d*84|0)+32>>2]=13;k[113084+(d*84|0)+36>>2]=14;k[113084+(d*84|0)+40>>2]=18;k[113084+(d*84|0)+44>>2]=19;k[113084+(d*84|0)+48>>2]=0;k[113084+(d*84|0)+52>>2]=14;k[113084+(d*84|0)+56>>2]=15;k[113084+(d*84|0)+60>>2]=0;k[113084+(d*84|0)+64>>2]=13;k[113084+(d*84|0)+68>>2]=0;k[113084+(d*84|0)+72>>2]=15;k[113084+(d*84|0)+76>>2]=16;k[113084+(d*84|0)+80>>2]=17;e=d+1|0;k[28270]=e;if((d|0)>=-1){if((d|0)<=22){Bx(113084+(e*84|0)|0,765488,20)|0;e=k[28270]|0;i[113084+(e*84|0)+19>>0]=0;k[113084+(e*84|0)+20>>2]=0;k[113084+(e*84|0)+24>>2]=4;k[113084+(e*84|0)+28>>2]=12;k[113084+(e*84|0)+32>>2]=13;k[113084+(e*84|0)+36>>2]=14;k[113084+(e*84|0)+40>>2]=0;k[113084+(e*84|0)+44>>2]=19;k[113084+(e*84|0)+48>>2]=0;k[113084+(e*84|0)+52>>2]=10;k[113084+(e*84|0)+56>>2]=9;k[113084+(e*84|0)+60>>2]=10;k[113084+(e*84|0)+64>>2]=11;k[113084+(e*84|0)+68>>2]=11;k[113084+(e*84|0)+72>>2]=11;k[113084+(e*84|0)+76>>2]=12;k[113084+(e*84|0)+80>>2]=13;d=e+1|0;k[28270]=d;do if((e|0)<-1){tk(726643);tk(726705)}else{if((e|0)>22)break;Bx(113084+(d*84|0)|0,765612,20)|0;f=k[28270]|0;i[113084+(f*84|0)+19>>0]=0;k[113084+(f*84|0)+20>>2]=0;k[113084+(f*84|0)+24>>2]=4;k[113084+(f*84|0)+28>>2]=12;k[113084+(f*84|0)+32>>2]=13;k[113084+(f*84|0)+36>>2]=14;k[113084+(f*84|0)+40>>2]=0;k[113084+(f*84|0)+44>>2]=0;k[113084+(f*84|0)+48>>2]=12;k[113084+(f*84|0)+52>>2]=14;k[113084+(f*84|0)+56>>2]=17;k[113084+(f*84|0)+60>>2]=0;k[113084+(f*84|0)+64>>2]=13;k[113084+(f*84|0)+68>>2]=0;k[113084+(f*84|0)+72>>2]=15;k[113084+(f*84|0)+76>>2]=16;k[113084+(f*84|0)+80>>2]=17;g=f+1|0;k[28270]=g;do if((f|0)<-1){tk(726643);tk(726705)}else{if((f|0)>22)break;Bx(113084+(g*84|0)|0,765674,20)|0;h=k[28270]|0;i[113084+(h*84|0)+19>>0]=0;k[113084+(h*84|0)+20>>2]=0;k[113084+(h*84|0)+24>>2]=4;k[113084+(h*84|0)+28>>2]=12;k[113084+(h*84|0)+32>>2]=13;k[113084+(h*84|0)+36>>2]=14;k[113084+(h*84|0)+40>>2]=0;k[113084+(h*84|0)+44>>2]=20;k[113084+(h*84|0)+48>>2]=0;k[113084+(h*84|0)+52>>2]=14;k[113084+(h*84|0)+56>>2]=15;k[113084+(h*84|0)+60>>2]=0;k[113084+(h*84|0)+64>>2]=13;k[113084+(h*84|0)+68>>2]=0;k[113084+(h*84|0)+72>>2]=15;k[113084+(h*84|0)+76>>2]=16;k[113084+(h*84|0)+80>>2]=17;j=h+1|0;k[28270]=j;do if((h|0)<-1){tk(726643);tk(726705)}else{if((h|0)>22)break;Bx(113084+(j*84|0)|0,727195,20)|0;l=k[28270]|0;i[113084+(l*84|0)+19>>0]=0;k[113084+(l*84|0)+20>>2]=0;k[113084+(l*84|0)+24>>2]=4;k[113084+(l*84|0)+28>>2]=12;k[113084+(l*84|0)+32>>2]=13;k[113084+(l*84|0)+36>>2]=14;k[113084+(l*84|0)+40>>2]=0;k[113084+(l*84|0)+44>>2]=21;k[113084+(l*84|0)+48>>2]=0;k[113084+(l*84|0)+52>>2]=14;k[113084+(l*84|0)+56>>2]=15;k[113084+(l*84|0)+60>>2]=0;k[113084+(l*84|0)+64>>2]=13;k[113084+(l*84|0)+68>>2]=0;k[113084+(l*84|0)+72>>2]=15;k[113084+(l*84|0)+76>>2]=16;k[113084+(l*84|0)+80>>2]=17;m=l+1|0;k[28270]=m;do if((l|0)<-1){tk(726643);tk(726705)}else{if((l|0)>22)break;Bx(113084+(m*84|0)|0,765510,20)|0;n=k[28270]|0;i[113084+(n*84|0)+19>>0]=0;k[113084+(n*84|0)+20>>2]=0;k[113084+(n*84|0)+24>>2]=4;k[113084+(n*84|0)+28>>2]=12;k[113084+(n*84|0)+32>>2]=13;k[113084+(n*84|0)+36>>2]=14;k[113084+(n*84|0)+40>>2]=0;k[113084+(n*84|0)+44>>2]=22;k[113084+(n*84|0)+48>>2]=0;k[113084+(n*84|0)+52>>2]=14;k[113084+(n*84|0)+56>>2]=15;k[113084+(n*84|0)+60>>2]=0;k[113084+(n*84|0)+64>>2]=13;k[113084+(n*84|0)+68>>2]=0;k[113084+(n*84|0)+72>>2]=15;k[113084+(n*84|0)+76>>2]=16;k[113084+(n*84|0)+80>>2]=17;o=n+1|0;k[28270]=o;do if((n|0)<-1){tk(726643);tk(726705)}else{if((n|0)>22)break;Bx(113084+(o*84|0)|0,736074,20)|0;p=k[28270]|0;i[113084+(p*84|0)+19>>0]=0;k[113084+(p*84|0)+20>>2]=0;k[113084+(p*84|0)+24>>2]=4;k[113084+(p*84|0)+28>>2]=12;k[113084+(p*84|0)+32>>2]=13;k[113084+(p*84|0)+36>>2]=14;k[113084+(p*84|0)+40>>2]=0;k[113084+(p*84|0)+44>>2]=23;k[113084+(p*84|0)+48>>2]=0;k[113084+(p*84|0)+52>>2]=14;k[113084+(p*84|0)+56>>2]=15;k[113084+(p*84|0)+60>>2]=0;k[113084+(p*84|0)+64>>2]=13;k[113084+(p*84|0)+68>>2]=0;k[113084+(p*84|0)+72>>2]=15;k[113084+(p*84|0)+76>>2]=16;k[113084+(p*84|0)+80>>2]=17;q=p+1|0;k[28270]=q;do if((p|0)<-1){tk(726643);tk(726705)}else{if((p|0)>22)break;Bx(113084+(q*84|0)|0,765442,20)|0;r=k[28270]|0;i[113084+(r*84|0)+19>>0]=0;k[113084+(r*84|0)+20>>2]=0;k[113084+(r*84|0)+24>>2]=2;k[113084+(r*84|0)+28>>2]=6;k[113084+(r*84|0)+32>>2]=7;k[113084+(r*84|0)+36>>2]=8;k[113084+(r*84|0)+40>>2]=0;k[113084+(r*84|0)+44>>2]=24;k[113084+(r*84|0)+48>>2]=10;k[113084+(r*84|0)+52>>2]=10;k[113084+(r*84|0)+56>>2]=9;k[113084+(r*84|0)+60>>2]=10;k[113084+(r*84|0)+64>>2]=11;k[113084+(r*84|0)+68>>2]=11;k[113084+(r*84|0)+72>>2]=11;k[113084+(r*84|0)+76>>2]=12;k[113084+(r*84|0)+80>>2]=13;s=r+1|0;k[28270]=s;do if((r|0)<-1){tk(726643);tk(726705)}else{if((r|0)>22)break;Bx(113084+(s*84|0)|0,727448,20)|0;t=k[28270]|0;i[113084+(t*84|0)+19>>0]=0;k[113084+(t*84|0)+20>>2]=0;k[113084+(t*84|0)+24>>2]=4;k[113084+(t*84|0)+28>>2]=12;k[113084+(t*84|0)+32>>2]=13;k[113084+(t*84|0)+36>>2]=14;k[113084+(t*84|0)+40>>2]=0;k[113084+(t*84|0)+44>>2]=0;k[113084+(t*84|0)+48>>2]=14;k[113084+(t*84|0)+52>>2]=14;k[113084+(t*84|0)+56>>2]=18;k[113084+(t*84|0)+60>>2]=10;k[113084+(t*84|0)+64>>2]=13;k[113084+(t*84|0)+68>>2]=0;k[113084+(t*84|0)+72>>2]=15;k[113084+(t*84|0)+76>>2]=16;k[113084+(t*84|0)+80>>2]=17;k[28270]=t+1;do if((t|0)<-1){tk(726643);tk(726705);u=122}else{if((t|0)>22){u=122;break}v=tj()|0;if(v){u=v;break}Bx(113084+((k[28270]|0)*84|0)|0,765434,20)|0;v=k[28270]|0;i[113084+(v*84|0)+19>>0]=0;k[113084+(v*84|0)+20>>2]=5;k[113084+(v*84|0)+24>>2]=6;k[113084+(v*84|0)+28>>2]=19;k[113084+(v*84|0)+32>>2]=20;k[113084+(v*84|0)+36>>2]=21;k[113084+(v*84|0)+40>>2]=0;k[113084+(v*84|0)+44>>2]=25;k[113084+(v*84|0)+48>>2]=15;k[113084+(v*84|0)+52>>2]=0;k[113084+(v*84|0)+56>>2]=22;k[113084+(v*84|0)+60>>2]=0;k[113084+(v*84|0)+64>>2]=16;k[113084+(v*84|0)+68>>2]=23;k[113084+(v*84|0)+72>>2]=26;k[113084+(v*84|0)+76>>2]=27;k[113084+(v*84|0)+80>>2]=28;w=v+1|0;k[28270]=w;do if((v|0)<-1){tk(726643);tk(726705)}else{if((v|0)>22)break;Bx(113084+(w*84|0)|0,765522,20)|0;x=k[28270]|0;i[113084+(x*84|0)+19>>0]=0;k[113084+(x*84|0)+20>>2]=0;k[113084+(x*84|0)+24>>2]=4;k[113084+(x*84|0)+28>>2]=12;k[113084+(x*84|0)+32>>2]=13;k[113084+(x*84|0)+36>>2]=14;k[113084+(x*84|0)+40>>2]=29;k[113084+(x*84|0)+44>>2]=30;k[113084+(x*84|0)+48>>2]=0;k[113084+(x*84|0)+52>>2]=14;k[113084+(x*84|0)+56>>2]=15;k[113084+(x*84|0)+60>>2]=0;k[113084+(x*84|0)+64>>2]=13;k[113084+(x*84|0)+68>>2]=0;k[113084+(x*84|0)+72>>2]=15;k[113084+(x*84|0)+76>>2]=16;k[113084+(x*84|0)+80>>2]=17;y=x+1|0;k[28270]=y;do if((x|0)<-1){tk(726643);tk(726705)}else{if((x|0)>22)break;Bx(113084+(y*84|0)|0,765458,20)|0;z=k[28270]|0;i[113084+(z*84|0)+19>>0]=0;k[113084+(z*84|0)+20>>2]=0;k[113084+(z*84|0)+24>>2]=2;k[113084+(z*84|0)+28>>2]=6;k[113084+(z*84|0)+32>>2]=7;k[113084+(z*84|0)+36>>2]=8;k[113084+(z*84|0)+40>>2]=0;k[113084+(z*84|0)+44>>2]=31;k[113084+(z*84|0)+48>>2]=10;k[113084+(z*84|0)+52>>2]=10;k[113084+(z*84|0)+56>>2]=9;k[113084+(z*84|0)+60>>2]=10;k[113084+(z*84|0)+64>>2]=11;k[113084+(z*84|0)+68>>2]=11;k[113084+(z*84|0)+72>>2]=11;k[113084+(z*84|0)+76>>2]=12;k[113084+(z*84|0)+80>>2]=13;A=z+1|0;k[28270]=A;do if((z|0)<-1){tk(726643);tk(726705)}else{if((z|0)>22)break;Bx(113084+(A*84|0)|0,742585,20)|0;B=k[28270]|0;i[113084+(B*84|0)+19>>0]=0;k[113084+(B*84|0)+20>>2]=0;k[113084+(B*84|0)+24>>2]=4;k[113084+(B*84|0)+28>>2]=12;k[113084+(B*84|0)+32>>2]=13;k[113084+(B*84|0)+36>>2]=14;k[113084+(B*84|0)+40>>2]=29;k[113084+(B*84|0)+44>>2]=31;k[113084+(B*84|0)+48>>2]=0;k[113084+(B*84|0)+52>>2]=14;k[113084+(B*84|0)+56>>2]=15;k[113084+(B*84|0)+60>>2]=0;k[113084+(B*84|0)+64>>2]=13;k[113084+(B*84|0)+68>>2]=0;k[113084+(B*84|0)+72>>2]=15;k[113084+(B*84|0)+76>>2]=16;k[113084+(B*84|0)+80>>2]=17;C=B+1|0;k[28270]=C;do if((B|0)<-1){tk(726643);tk(726705)}else{if((B|0)>22)break;Bx(113084+(C*84|0)|0,765530,20)|0;D=k[28270]|0;i[113084+(D*84|0)+19>>0]=0;k[113084+(D*84|0)+20>>2]=0;k[113084+(D*84|0)+24>>2]=4;k[113084+(D*84|0)+28>>2]=12;k[113084+(D*84|0)+32>>2]=13;k[113084+(D*84|0)+36>>2]=14;k[113084+(D*84|0)+40>>2]=0;k[113084+(D*84|0)+44>>2]=32;k[113084+(D*84|0)+48>>2]=0;k[113084+(D*84|0)+52>>2]=14;k[113084+(D*84|0)+56>>2]=15;k[113084+(D*84|0)+60>>2]=0;k[113084+(D*84|0)+64>>2]=13;k[113084+(D*84|0)+68>>2]=0;k[113084+(D*84|0)+72>>2]=15;k[113084+(D*84|0)+76>>2]=16;k[113084+(D*84|0)+80>>2]=17;E=D+1|0;k[28270]=E;do if((D|0)<-1){tk(726643);tk(726705)}else{if((D|0)>22)break;Bx(113084+(E*84|0)|0,765481,20)|0;F=k[28270]|0;i[113084+(F*84|0)+19>>0]=0;k[113084+(F*84|0)+20>>2]=0;k[113084+(F*84|0)+24>>2]=4;k[113084+(F*84|0)+28>>2]=12;k[113084+(F*84|0)+32>>2]=13;k[113084+(F*84|0)+36>>2]=14;k[113084+(F*84|0)+40>>2]=33;k[113084+(F*84|0)+44>>2]=34;k[113084+(F*84|0)+48>>2]=0;k[113084+(F*84|0)+52>>2]=14;k[113084+(F*84|0)+56>>2]=15;k[113084+(F*84|0)+60>>2]=0;k[113084+(F*84|0)+64>>2]=13;k[113084+(F*84|0)+68>>2]=0;k[113084+(F*84|0)+72>>2]=15;k[113084+(F*84|0)+76>>2]=16;k[113084+(F*84|0)+80>>2]=17;G=F+1|0;k[28270]=G;do if((F|0)<-1){tk(726643);tk(726705)}else{if((F|0)>22)break;Bx(113084+(G*84|0)|0,765470,20)|0;H=k[28270]|0;i[113084+(H*84|0)+19>>0]=0;k[113084+(H*84|0)+20>>2]=0;k[113084+(H*84|0)+24>>2]=2;k[113084+(H*84|0)+28>>2]=6;k[113084+(H*84|0)+32>>2]=7;k[113084+(H*84|0)+36>>2]=8;k[113084+(H*84|0)+40>>2]=0;k[113084+(H*84|0)+44>>2]=35;k[113084+(H*84|0)+48>>2]=10;k[113084+(H*84|0)+52>>2]=10;k[113084+(H*84|0)+56>>2]=9;k[113084+(H*84|0)+60>>2]=10;k[113084+(H*84|0)+64>>2]=11;k[113084+(H*84|0)+68>>2]=11;k[113084+(H*84|0)+72>>2]=11;k[113084+(H*84|0)+76>>2]=12;k[113084+(H*84|0)+80>>2]=13;I=H+1|0;k[28270]=I;do if((H|0)<-1){tk(726643);tk(726705)}else{if((H|0)>22)break;Bx(113084+(I*84|0)|0,742596,20)|0;J=k[28270]|0;i[113084+(J*84|0)+19>>0]=0;k[113084+(J*84|0)+20>>2]=0;k[113084+(J*84|0)+24>>2]=4;k[113084+(J*84|0)+28>>2]=12;k[113084+(J*84|0)+32>>2]=13;k[113084+(J*84|0)+36>>2]=14;k[113084+(J*84|0)+40>>2]=33;k[113084+(J*84|0)+44>>2]=35;k[113084+(J*84|0)+48>>2]=0;k[113084+(J*84|0)+52>>2]=14;k[113084+(J*84|0)+56>>2]=15;k[113084+(J*84|0)+60>>2]=0;k[113084+(J*84|0)+64>>2]=13;k[113084+(J*84|0)+68>>2]=0;k[113084+(J*84|0)+72>>2]=15;k[113084+(J*84|0)+76>>2]=16;k[113084+(J*84|0)+80>>2]=17;K=J+1|0;k[28270]=K;do if((J|0)<-1){tk(726643);tk(726705)}else{if((J|0)>22)break;Bx(113084+(K*84|0)|0,765546,20)|0;L=k[28270]|0;i[113084+(L*84|0)+19>>0]=0;k[113084+(L*84|0)+20>>2]=0;k[113084+(L*84|0)+24>>2]=4;k[113084+(L*84|0)+28>>2]=12;k[113084+(L*84|0)+32>>2]=13;k[113084+(L*84|0)+36>>2]=14;k[113084+(L*84|0)+40>>2]=0;k[113084+(L*84|0)+44>>2]=36;k[113084+(L*84|0)+48>>2]=0;k[113084+(L*84|0)+52>>2]=14;k[113084+(L*84|0)+56>>2]=15;k[113084+(L*84|0)+60>>2]=0;k[113084+(L*84|0)+64>>2]=13;k[113084+(L*84|0)+68>>2]=0;k[113084+(L*84|0)+72>>2]=15;k[113084+(L*84|0)+76>>2]=16;k[113084+(L*84|0)+80>>2]=17;M=L+1|0;k[28270]=M;do if((L|0)<-1){tk(726643);tk(726705)}else{if((L|0)>22)break;Bx(113084+(M*84|0)|0,755455,20)|0;N=k[28270]|0;O=113084+(N*84|0)+44|0;P=113084+(N*84|0)+19|0;Q=P+25|0;do{i[P>>0]=0;P=P+1|0}while((P|0)<(Q|0));k[O>>2]=37;k[113084+(N*84|0)+48>>2]=17;k[113084+(N*84|0)+52>>2]=0;k[113084+(N*84|0)+56>>2]=24;k[113084+(N*84|0)+60>>2]=0;k[113084+(N*84|0)+64>>2]=18;k[113084+(N*84|0)+68>>2]=25;k[113084+(N*84|0)+72>>2]=38;k[113084+(N*84|0)+76>>2]=39;k[113084+(N*84|0)+80>>2]=40;k[28270]=N+1;k[28269]=0;a=0;return a|0}while(0);tk(727970);a=122;return a|0}while(0);tk(727910);a=122;return a|0}while(0);tk(727855);a=122;return a|0}while(0);tk(727799);a=122;return a|0}while(0);tk(727747);a=122;return a|0}while(0);tk(727686);a=122;return a|0}while(0);tk(727630);a=122;return a|0}while(0);tk(727573);a=122;return a|0}while(0);tk(727520);a=122;return a|0}while(0);tk(727467);a=u;return a|0}while(0);tk(727384);a=122;return a|0}while(0);tk(727323);a=122;return a|0}while(0);tk(727263);a=122;return a|0}while(0);tk(727206);a=122;return a|0}while(0);tk(727139);a=122;return a|0}while(0);tk(727083);a=122;return a|0}while(0);tk(727028);a=122;return a|0}}else{tk(726643);tk(726705)}tk(726970);a=122;return a|0}}else{tk(726643);tk(726705)}tk(726916);a=122;return a|0}}else{tk(726643);tk(726705)}tk(726860);a=122;return a|0}else R=c}else R=122;else{tk(726643);tk(726705);R=122}tk(726808);a=R;return a|0}else S=b}else S=122;else{tk(726643);tk(726705);S=122}tk(726755);a=S;return a|0}function Dh(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;return Jh(a,b,c,d,e,f,g,h,0,0,i)|0}function Eh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;c=k[28270]|0;while(1){d=c+-1|0;if((c|0)<=0){e=124;f=5;break}if(!(tx(113084+(d*84|0)|0,a)|0)){g=d;break}else c=d}if((f|0)==5)return e|0;k[b>>2]=g;e=0;return e|0}function Fh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;c=r;r=r+16|0;d=c+4|0;e=c;k[d>>2]=999;k[e>>2]=0;if(!a){k[b>>2]=115;f=115;r=c;return f|0}g=a+4|0;if((k[(k[g>>2]|0)+16>>2]|0)!=555){k[b>>2]=114;f=114;r=c;return f|0}if((k[b>>2]|0)>0)wl(a,d)|0;else wl(a,b)|0;d=(k[g>>2]|0)+8|0;h=(k[d>>2]|0)+-1|0;k[d>>2]=h;if(!h){hu(a,1,b)|0;h=k[g>>2]|0;if((Ub[k[113084+((k[h+4>>2]|0)*84|0)+56>>2]&31](k[h>>2]|0)|0)!=0?(k[b>>2]|0)<1:0){k[b>>2]=110;tk(728025);tk(k[(k[g>>2]|0)+12>>2]|0)}h=k[g>>2]|0;d=0;while(1){i=115100+(d<<2)|0;d=d+1|0;if((k[i>>2]|0)==(h|0)){j=i;l=15;break}if((d|0)>=1e3){m=h;break}}if((l|0)==15){k[j>>2]=0;m=k[g>>2]|0}dy(k[m+1244>>2]|0);dy(k[(k[g>>2]|0)+88>>2]|0);dy(k[(k[g>>2]|0)+12>>2]|0);dy(k[g>>2]|0);dy(a)}else{if((k[b>>2]|0)>0)hu(a,0,e)|0;else hu(a,0,b)|0;dy(a)}f=k[b>>2]|0;r=c;return f|0}function Gh(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;j=r;r=r+80|0;l=j;m=j+8|0;n=j+4|0;k[b>>2]=0;i[c>>0]=0;k[d>>2]=0;k[e>>2]=-1;i[f>>0]=0;i[g>>0]=0;o=k[h>>2]|0;if((o|0)>0){p=o;r=j;return p|0}else q=a;while(1){o=i[q>>0]|0;if(o<<24>>24==32)q=q+1|0;else{s=o;t=q;break}}do if(((s<<24>>24)+-48|0)>>>0<10){q=rv()|0;k[q>>2]=0;o=jx(t,n,10)|0;k[b>>2]=o;u=k[n>>2]|0;a:while(1){switch(i[u>>0]|0){case 59:case 0:{v=7;break a;break}case 32:break;default:{v=8;break a}}w=u+1|0;k[n>>2]=w;u=w}if((v|0)==7)if((k[q>>2]|0)==34)v=8;else{x=o;y=0}if((v|0)==8){k[b>>2]=0;k[q>>2]=0;x=k[b>>2]|0;y=1}if(x>>>0<=99999)if(!y){z=t;break}else{v=12;break}k[b>>2]=0;tk(728070);tk(a);k[h>>2]=125;p=125;r=j;return p|0}else v=12;while(0);b:do if((v|0)==12){b=vx(t,728114)|0;if((b|0)>70){k[h>>2]=125;p=125;r=j;return p|0}zx(c,t,b)|0;c:do if((b|0)>0){y=b;while(1){x=y+-1|0;n=c+x|0;if((i[n>>0]|0)!=32){A=y;break c}i[n>>0]=0;if((y|0)>1)y=x;else{A=x;break}}}else A=b;while(0);b=(Dx(t+A|0,728118)|0)+A|0;q=t+b|0;o=vx(q,767743)|0;if(!o){ux(m,c)|0;Ck(m);if((tx(m,764159)|0)!=0?(tx(m,728193)|0)!=0:0){z=q;break}i[c>>0]=0;z=q;break}k[l>>2]=d;if((Yw(q,772183,l)|0)!=1){tk(728122);tk(a);k[h>>2]=125;p=125;r=j;return p|0}q=b+o|0;o=t+((Dx(t+q|0,728118)|0)+q)|0;if(!(vx(o,728157)|0))z=o;else switch(i[o>>0]|0){case 66:case 98:{k[e>>2]=2;z=o;break b;break}case 65:case 97:case 84:case 116:{k[e>>2]=1;z=o;break b;break}case 73:case 105:{k[e>>2]=0;z=o;break b;break}default:{tk(728159);tk(a);k[h>>2]=125;p=125;r=j;return p|0}}}while(0);e=rx(z,59)|0;do if(e){z=e;while(1){t=z+1|0;if((i[t>>0]|0)==32)z=t;else{B=t;break}}z=rx(B,40)|0;if(!z){tk(728195);tk(728254);tk(a);k[h>>2]=125;p=125;r=j;return p|0}t=z-B|0;if((t|0)>1024){k[h>>2]=125;p=125;r=j;return p|0}zx(f,B,t)|0;t=z;while(1){z=t+1|0;if((i[z>>0]|0)==32)t=z;else{C=z;break}}t=(rx(C,41)|0)-C|0;if((t|0)<=1024){zx(g,C,t)|0;break}k[h>>2]=125;p=125;r=j;return p|0}while(0);p=k[h>>2]|0;r=j;return p|0}function Hh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0;e=r;r=r+13232|0;f=e+168|0;g=e+160|0;h=e+152|0;j=e+144|0;l=e+136|0;m=e+244|0;n=e+240|0;o=e+236|0;q=e+232|0;s=e+228|0;t=e+224|0;u=e+220|0;v=e+128|0;w=e+216|0;x=e+212|0;y=e+208|0;z=e+204|0;A=e+200|0;B=e+196|0;C=e+192|0;D=e+188|0;E=e+104|0;F=e+12198|0;G=e+11173|0;H=e+1076|0;I=e+10148|0;J=e+10077|0;K=e+9052|0;L=e+8981|0;M=e+8910|0;N=e+7885|0;O=e+6860|0;P=e+5835|0;Q=e+4810|0;R=e+3785|0;S=e+2760|0;T=e+1735|0;U=e+1664|0;V=e+1380|0;W=e+1096|0;X=e+792|0;Y=e+72|0;Z=e+40|0;_=e+8|0;$=e;aa=e+184|0;ba=e+180|0;ca=e+176|0;da=e+172|0;ea=e+508|0;fa=e+424|0;k[D>>2]=0;k[ba>>2]=1;ga=k[d>>2]|0;if((ga|0)>0){ha=ga;r=e;return ha|0}switch(ga|0){case -102:{k[d>>2]=0;ia=0;ja=0;ka=1;la=0;break}case -103:{k[d>>2]=0;ia=0;ja=1;ka=0;la=0;break}case -104:{k[d>>2]=0;ia=0;ja=0;ka=0;la=1;break}case -105:{k[d>>2]=0;ia=1;ja=0;ka=0;la=0;break}default:{ia=0;ja=0;ka=0;la=0}}k[a>>2]=0;if((k[28269]|0)!=0?(ga=Ch()|0,k[d>>2]=ga,(ga|0)>0):0){ha=ga;r=e;return ha|0}else ma=b;a:while(1){switch(i[ma>>0]|0){case 0:{na=11;break a;break}case 32:break;default:{oa=ma;break a}}ma=ma+1|0}if((na|0)==11){tk(728301);k[d>>2]=104;ha=104;r=e;return ha|0}do if(!ia)Jh(oa,E,F,G,I,K,O,P,Q,T,d)|0;else{if((xx(oa)|0)>>>0<=1024){ux(F,oa)|0;ma=E;k[ma>>2]=1701603686;k[ma+4>>2]=3092282;i[G>>0]=0;i[I>>0]=0;i[O>>0]=0;i[P>>0]=0;i[K>>0]=0;i[Q>>0]=0;i[T>>0]=0;break}tk(728341);k[d>>2]=104;ha=104;r=e;return ha|0}while(0);if((k[d>>2]|0)>0){tk(728384);tk(oa);ha=k[d>>2]|0;r=e;return ha|0}i[M>>0]=0;i[N>>0]=0;if(i[I>>0]|0){ia=I+((xx(I)|0)+-1)|0;if((i[ia>>0]|0)==35){i[ia>>0]=0;pa=1}else pa=0;Gh(I,z,J,A,C,M,N,d)|0;ia=k[d>>2]|0;if((ia|0)>0){ha=ia;r=e;return ha|0}else qa=pa}else qa=0;i[R>>0]=0;i[S>>0]=0;do if(i[G>>0]|0){if((i[M>>0]|i[O>>0]|i[Q>>0])<<24>>24){ux(R,G)|0;i[G>>0]=0;break}if((i[P>>0]|i[K>>0])<<24>>24){ux(S,G)|0;i[G>>0]=0}}while(0);if((Kh(a,oa,E,F,I,K,O,P,c,u,d)|0)>0){ha=k[d>>2]|0;r=e;return ha|0}do if(!(k[u>>2]|0)){pa=k[28270]|0;while(1){ia=pa+-1|0;if((pa|0)<=0){na=34;break}if(!(tx(113084+(ia*84|0)|0,E)|0)){ra=ia;break}else pa=ia}if((na|0)==34){k[d>>2]=124;tk(728429);tk(E);tk(oa);ha=k[d>>2]|0;r=e;return ha|0}k[d>>2]=0;pa=113084+(ra*84|0)+40|0;do if(k[pa>>2]|0){ux(H,E)|0;ia=Pb[k[pa>>2]&63](E,F,G)|0;k[d>>2]=ia;if(ia){tk(728475);tk(oa);ha=k[d>>2]|0;r=e;return ha|0}if(tx(H,E)|0){ia=k[28270]|0;while(1){ma=ia+-1|0;if((ia|0)<=0)break;if(!(tx(113084+(ma*84|0)|0,E)|0)){sa=ma;na=42;break}else ia=ma}if((na|0)==42){k[d>>2]=0;ta=sa;break}k[d>>2]=124;tk(728516);tk(oa);tk(E);ha=k[d>>2]|0;r=e;return ha|0}else ta=ra}else ta=ra;while(0);pa=k[113084+(ta*84|0)+44>>2]|0;if(!pa){tk(728616);tk(oa);k[d>>2]=104;ha=104;r=e;return ha|0}ia=Pb[pa&63](F,c,B)|0;k[d>>2]=ia;if((ia|0)>0){tk(728564);tk(oa);ha=k[d>>2]|0;r=e;return ha|0}ia=Xb[k[113084+(ta*84|0)+64>>2]&31](k[B>>2]|0,v)|0;k[d>>2]=ia;if((ia|0)>0){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728668);tk(oa);ha=k[d>>2]|0;r=e;return ha|0}ia=ey(1,8)|0;k[a>>2]=ia;if(!ia){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728720);tk(oa);k[d>>2]=113;ha=113;r=e;return ha|0}pa=ey(1,1728)|0;k[ia+4>>2]=pa;pa=(k[a>>2]|0)+4|0;if(!(k[pa>>2]|0)){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728720);tk(oa);dy(k[a>>2]|0);k[a>>2]=0;k[d>>2]=113;ha=113;r=e;return ha|0}ia=(xx(oa)|0)+1|0;ma=cy((ia|0)>32?ia:32)|0;k[(k[pa>>2]|0)+12>>2]=ma;ma=(k[a>>2]|0)+4|0;if(!(k[(k[ma>>2]|0)+12>>2]|0)){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728778);tk(oa);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[d>>2]=113;ha=113;r=e;return ha|0}pa=ey(1001,8)|0;k[(k[ma>>2]|0)+88>>2]=pa;pa=(k[a>>2]|0)+4|0;if(!(k[(k[pa>>2]|0)+88>>2]|0)){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728827);tk(oa);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[d>>2]=113;ha=113;r=e;return ha|0}ma=ey(40,2880)|0;k[(k[pa>>2]|0)+1244>>2]=ma;ma=k[(k[a>>2]|0)+4>>2]|0;if(!(k[ma+1244>>2]|0)){Ub[k[113084+(ta*84|0)+56>>2]&31](k[B>>2]|0)|0;tk(728883);tk(oa);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+88>>2]|0);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[d>>2]=113;ha=113;r=e;return ha|0}else ua=0;do{k[ma+1568+(ua<<2)>>2]=ua;k[ma+1248+(ua<<2)>>2]=-1;ua=ua+1|0}while((ua|0)!=40);k[ma+84>>2]=1e3;k[ma>>2]=k[B>>2];k[ma+4>>2]=ta;ux(k[ma+12>>2]|0,oa)|0;pa=v;ia=k[pa+4>>2]|0;ga=k[a>>2]|0;va=k[ga+4>>2]|0;wa=va+24|0;k[wa>>2]=k[pa>>2];k[wa+4>>2]=ia;ia=v;wa=k[ia+4>>2]|0;pa=va+32|0;k[pa>>2]=k[ia>>2];k[pa+4>>2]=wa;k[va+76>>2]=c;wa=va+120|0;k[wa>>2]=-1;k[wa+4>>2]=-1;k[va+64>>2]=-1;k[va+8>>2]=1;k[va+16>>2]=555;k[va+20>>2]=qa;bu(ga,0,0,d)|0;ga=k[a>>2]|0;va=k[ga+4>>2]|0;b:do if((k[d>>2]|0)>0)xa=ga;else{wa=0;while(1){pa=115100+(wa<<2)|0;wa=wa+1|0;if(!(k[pa>>2]|0)){ya=pa;break}if((wa|0)>=1e3){xa=ga;break b}}k[ya>>2]=va;xa=k[a>>2]|0}while(0);if((dl(xa,s,d)|0)<=0){za=(i[G>>0]|0)!=0&1;break}tk(728938);tk(oa);if((k[d>>2]|0)==252)tk(737931);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}else za=0;while(0);c:do if(!(i[I>>0]|0)){oa=(ja|0)!=0;xa=(la|0)!=0;if((!((ja|ka|la|0)==0?(i[P>>0]|i[M>>0]|i[K>>0]|i[O>>0])<<24>>24==0:0)?(bl(k[a>>2]|0,t)|0,(k[t>>2]|0)==1):0)?(Fl(k[a>>2]|0,ba,d)|0,oa|(k[ba>>2]|0)==0):0){if(!(Bl(k[a>>2]|0,1,s,d)|0))do{ya=k[s>>2]|0;qa=(ya|0)==0;do if(!(oa&qa|xa&(ya|0)!=0)){if(qa){Fl(k[a>>2]|0,ba,d)|0;if((k[ba>>2]|0)>0)break c;else break}k[D>>2]=0;i[L>>0]=0;Io(k[a>>2]|0,16,790194,L,0,D)|0;if(!(Ex(L,790202)|0)){if(Ex(L,729046)|0)break;if(yx(L,729050,8)|0)break c}}while(0)}while((Bl(k[a>>2]|0,1,s,d)|0)==0);if((k[d>>2]|0)==107)k[d>>2]=0;Fk(k[a>>2]|0,1,s,d)|0}}else{xa=k[z>>2]|0;if(!xa){if(i[J>>0]|0)Jl(k[a>>2]|0,k[C>>2]|0,J,k[A>>2]|0,d)|0}else Fk(k[a>>2]|0,xa+1|0,s,d)|0;if((k[d>>2]|0)>0){tk(728996);if((xa|0)>0){k[l>>2]=xa;Xw(fa,726448,l)|0;tk(fa)}else{k[j>>2]=J;Xw(fa,726506,j)|0;tk(fa);xa=k[A>>2]|0;if(xa){k[h>>2]=xa;Xw(fa,726536,h)|0;tk(fa)}xa=k[C>>2]|0;if((xa|0)!=-1){k[g>>2]=k[119100+(xa<<2)>>2];Xw(fa,726570,g)|0;tk(fa)}tk(726605)}Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}}while(0);do if(!(i[M>>0]|0))Aa=za;else{if(((i[N>>0]|0)+-48|0)>>>0<10){k[f>>2]=w;Yw(N,773048,f)|0;if((k[w>>2]|0)<1){tk(729059);tk(N);tk(729090);tk(I);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;k[d>>2]=307;ha=307;r=e;return ha|0}}else{if((hk(k[a>>2]|0,N,w,d)|0)>0){tk(729142);tk(N);tk(729090);tk(I);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}if(!(k[w>>2]|0)){tk(729187);tk(N);tk(729090);tk(I);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;k[d>>2]=307;ha=307;r=e;return ha|0}}if((i[R>>0]|0)!=0&(i[Q>>0]|0)==0)ux(G,R)|0;else{Ba=G;Ca=729235;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0))}if((Lh(q,G,d)|0)>0){tk(729244);tk(G);ha=k[d>>2]|0;r=e;return ha|0}if((Mh(k[a>>2]|0,k[q>>2]|0,M,k[w>>2]|0,d)|0)<=0){Fh(k[a>>2]|0,d)|0;k[a>>2]=k[q>>2];Aa=1;break}tk(729299);tk(I);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}while(0);if(i[P>>0]|0){do if(!Aa){I=Nh(E)|0;if(!I)if((i[S>>0]|0)!=0&(i[G>>0]|0)==0){ux(G,S)|0;Ea=1;break}else{Ba=G;Ca=729235;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0));Ea=1;break}else{Fa=I;na=120}}else{Fa=Aa;na=120}while(0);if((na|0)==120){k[(k[(k[a>>2]|0)+4>>2]|0)+76>>2]=1;i[G>>0]=0;Ea=Fa}if((Oh(a,G,P,d)|0)>0){tk(729347);tk(729394);tk(P);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}else Ga=Ea}else Ga=Aa;do if(!(i[K>>0]|0))Ha=0;else{Zk(k[a>>2]|0,s,d)|0;if(!(k[s>>2]|0)){Aa=(i[G>>0]|0)==0;if(!((i[S>>0]|0)!=0&Aa)){if(Aa){Ba=G;Ca=729444;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0))}}else ux(G,S)|0;if((Ph(a,G,K,d)|0)<=0){Ha=0;break}tk(729453);tk(729507);tk(K);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}if(i[O>>0]|0){al(k[a>>2]|0,x,d)|0;I=k[x>>2]|0;Aa=ey(I,1)|0;if(!Aa){tk(729558);tk(729620);tk(K);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;k[d>>2]=113;ha=113;r=e;return ha|0}if(($j(k[a>>2]|0,K,1,I,y,Aa,d)|0)<=0){Ha=Aa;break}tk(729676);tk(729620);tk(K);dy(Aa);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}do if((Ga|0)==0?(Nh(E)|0)==0:0){Aa=(i[G>>0]|0)==0;if((i[S>>0]|0)!=0&Aa){ux(G,S)|0;break}if(Aa){Ba=G;Ca=729444;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0))}}else na=142;while(0);if((na|0)==142){k[(k[(k[a>>2]|0)+4>>2]|0)+76>>2]=1;i[G>>0]=0}if((Qh(a,G,K,d)|0)<=0){xs(k[a>>2]|0,729785,d)|0;xs(k[a>>2]|0,b,d)|0;Ha=0;break}tk(729725);tk(729620);tk(K);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}while(0);do if(i[O>>0]|0){if((i[R>>0]|0)!=0&(i[Q>>0]|0)==0)ux(G,R)|0;else{Ba=G;Ca=729855;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0))}Qp(O,aa,ca,ea,Y,Z,_,V,W,X,$,U,da,d)|0;Sp(a,G,k[aa>>2]|0,k[ca>>2]|0,ea,Y,Z,_,V,W,X,+p[$>>3],U,k[da>>2]|0,Ha,d)|0;if(Ha)dy(Ha);if((k[d>>2]|0)<=0){xs(k[a>>2]|0,729984,d)|0;xs(k[a>>2]|0,b,d)|0;break}tk(729864);tk(729920);tk(O);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}while(0);do if(i[Q>>0]|0){if(!(i[R>>0]|0)){Ba=G;Ca=730048;Da=Ba+9|0;do{i[Ba>>0]=i[Ca>>0]|0;Ba=Ba+1|0;Ca=Ca+1|0}while((Ba|0)<(Da|0))}else ux(G,R)|0;Zk(k[a>>2]|0,s,d)|0;if(k[s>>2]|0){tk(730331);tk(Q);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;k[d>>2]=233;ha=233;r=e;return ha|0}sy(m|0,0,180)|0;k[n>>2]=789744;k[m>>2]=1;k[m+12>>2]=a;k[m+8>>2]=n;O=m+28|0;d:do if((Lh(O,G,d)|0)>0){tk(730057);tk(G)}else{bl(k[a>>2]|0,o)|0;Ha=Q+3|0;switch(i[Ha>>0]|0){case 66:case 98:{k[m+20>>2]=8;Ia=0;break}case 73:case 105:{k[m+20>>2]=16;Ia=0;break}case 74:case 106:{k[m+20>>2]=32;Ia=0;break}case 82:case 114:{k[m+20>>2]=-32;Ia=0;break}case 68:case 100:{k[m+20>>2]=-64;Ia=0;break}default:Ia=(k[m+20>>2]|0)==0}da=Ia?Ha:Q+4|0;Ha=(i[da>>0]|0)==49;U=Ha?Q+(Ia?4:5)|0:da;da=Ha&1;Ha=k[(k[(k[a>>2]|0)+4>>2]|0)+20>>2]|0;if((i[U>>0]|0)==32)Ja=U;else{tk(730104);tk(U);Ja=U}while(1)if((i[Ja>>0]|0)==32)Ja=Ja+1|0;else{Ka=Ja;break}U=(Ha|da|0)!=0;$=k[o>>2]|0;X=k[a>>2]|0;e:do if(($|0)>1&(U^1)){f:do if(U){Fk(X,1,0,d)|0;if((Wj(k[a>>2]|0,k[O>>2]|0,0,d)|0)>0)break;La=k[o>>2]|0;Ma=k[a>>2]|0;break e}else{W=X;V=1;while(1){Fk(W,V,0,d)|0;V=V+1|0;if((Wj(k[a>>2]|0,k[O>>2]|0,0,d)|0)>0)break f;_=k[o>>2]|0;Z=k[a>>2]|0;if((V|0)>=(_|0)){La=_;Ma=Z;break e}else W=Z}}while(0);Fh(k[O>>2]|0,d)|0;break d}else{La=$;Ma=X}while(0);Fk(Ma,La,0,d)|0;k[m+16>>2]=Ka;if(jk(m,d)|0){tk(730152);tk(Ka);Fh(k[O>>2]|0,d)|0;break}X=k[o>>2]|0;do if(U)Na=X;else{$=X+1|0;if((Fk(k[a>>2]|0,$,0,d)|0)>0){Na=X;break}else Oa=$;while(1){Wj(k[a>>2]|0,k[O>>2]|0,0,d)|0;$=Oa+1|0;if((Fk(k[a>>2]|0,$,0,d)|0)>0){Na=Oa;break}else Oa=$}}while(0);X=k[d>>2]|0;do if((X|0)==107)k[d>>2]=0;else{if((X|0)<=0)break;Fh(k[O>>2]|0,d)|0;break d}while(0);Fh(k[a>>2]|0,d)|0;X=k[O>>2]|0;k[a>>2]=X;U=k[o>>2]|0;if((Na|0)==(U|0))break;Fk(X,U,0,d)|0}while(0);if((k[d>>2]|0)<=0){xs(k[a>>2]|0,730271,d)|0;xs(k[a>>2]|0,b,d)|0;break}tk(730184);tk(730231);tk(Q);Fh(k[a>>2]|0,d)|0;k[a>>2]=0;ha=k[d>>2]|0;r=e;return ha|0}while(0);if(i[T>>0]|0)Rh(k[a>>2]|0,T,d)|0;ha=k[d>>2]|0;r=e;return ha|0}function Ih(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0;f=r;r=r+16|0;g=f+8|0;h=f;if((a|0)==5){Hh(b,c,d,e)|0;i=k[e>>2]|0;r=f;return i|0}else{Tw(730372)|0;Tw(730446)|0;Tw(730524)|0;k[h>>2]=5;Rw(730578,h)|0;k[g>>2]=a;Rw(730631,g)|0;Tw(730684)|0;Tw(730754)|0;k[e>>2]=104;i=104;r=f;return i|0}return 0}function Jh(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=(c|0)!=0;if(n)i[c>>0]=0;p=(b|0)!=0;if(p)i[b>>0]=0;q=(d|0)!=0;if(q)i[d>>0]=0;r=(e|0)!=0;if(r)i[e>>0]=0;s=(g|0)!=0;if(s)i[g>>0]=0;t=(h|0)!=0;if(t)i[h>>0]=0;u=(f|0)!=0;if(u)i[f>>0]=0;v=(j|0)!=0;if(v)i[j>>0]=0;w=(l|0)!=0;if(w)i[l>>0]=0;x=xx(a)|0;if(!x){o=k[m>>2]|0;return o|0}y=x+1|0;z=ey(3,y)|0;if(!z){k[m>>2]=113;o=113;return o|0}A=z+y|0;B=z+(y<<1)|0;a:do if((i[a>>0]|0)==45){y=a+1|0;switch(i[y>>0]|0){case 40:case 91:case 32:case 0:break;default:{C=28;break a}}if(p){D=b+(xx(b)|0)|0;E=765501;F=D+9|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0));G=y;C=63}else H=y}else C=28;while(0);b:do if((C|0)==28){do if(!(yx(a,730780,5)|0)){if(p){D=b+(xx(b)|0)|0;E=765501;F=D+9|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0))}I=a+5|0}else{J=Ex(a,765357)|0;K=rx(a,40)|0;if(!((J|0)==0|(K|0)!=0&K>>>0<J>>>0)){if(p)zx(b,a,3-a+J|0)|0;I=J+3|0;break}if(!(Ax(a,743143,4)|0)){if(p){J=b+(xx(b)|0)|0;i[J>>0]=i[765481]|0;i[J+1>>0]=i[765482]|0;i[J+2>>0]=i[765483]|0;i[J+3>>0]=i[765484]|0;i[J+4>>0]=i[765485]|0;i[J+5>>0]=i[765486]|0;i[J+6>>0]=i[765487]|0}I=a+4|0;break}if(!(Ax(a,730786,7)|0)){if(p){D=b+(xx(b)|0)|0;E=730794;F=D+10|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0))}I=a+7|0;break}if(!(Ax(a,743121,5)|0)){if(p){J=b+(xx(b)|0)|0;K=J;i[K>>0]=104;i[K+1>>0]=116;i[K+2>>0]=116;i[K+3>>0]=112;K=J+4|0;i[K>>0]=58;i[K+1>>0]=47;i[K+2>>0]=47;i[K+3>>0]=0}I=a+5|0;break}if(!(Ax(a,741186,4)|0)){if(p){K=b+(xx(b)|0)|0;i[K>>0]=i[777975]|0;i[K+1>>0]=i[777976]|0;i[K+2>>0]=i[777977]|0;i[K+3>>0]=i[777978]|0;i[K+4>>0]=i[777979]|0;i[K+5>>0]=i[777980]|0;i[K+6>>0]=i[777981]|0}I=a+4|0;break}if(!(Ax(a,730804,6)|0)){if(p){D=b+(xx(b)|0)|0;E=765425;F=D+9|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0))}I=a+6|0;break}if(Ax(a,730811,5)|0){if(!p){H=a;break b}K=b+(xx(b)|0)|0;J=K;i[J>>0]=102;i[J+1>>0]=105;i[J+2>>0]=108;i[J+3>>0]=101;J=K+4|0;i[J>>0]=58;i[J+1>>0]=47;i[J+2>>0]=47;i[J+3>>0]=0;G=a;C=63;break b}if(p){J=b+(xx(b)|0)|0;K=J;i[K>>0]=102;i[K+1>>0]=105;i[K+2>>0]=108;i[K+3>>0]=101;K=J+4|0;i[K>>0]=58;i[K+1>>0]=47;i[K+2>>0]=47;i[K+3>>0]=0}I=a+5|0}while(0);if(p){G=I;C=63}else H=I}while(0);c:do if((C|0)==63)if(!(Ax(b,765522,7)|0)){if((rx(G,40)|0)==0?(rx(G,91)|0)==0:0){H=G;break}I=G+((xx(G)|0)+-1)|0;d:while(1){switch(i[I>>0]|0){case 41:case 93:{H=G;break c;break}case 32:break;default:break d}I=I+-1|0}do if(n){if((xx(G)|0)>>>0<=1024){ux(c,G)|0;break}tk(730817);k[m>>2]=125;o=125;return o|0}while(0);dy(z);o=k[m>>2]|0;return o|0}else H=G;while(0);if((i[H>>0]|0)==91)L=(i[a>>0]|0)==45?H:H+1|0;else{a=Ex(H,730843)|0;L=(a|0)==0?H:a+2|0}a=rx(L,40)|0;G=rx(L,91)|0;do if((a|0)==(G|0)){qx(z,H)|0;M=a}else{if((G|0)!=0?!((a|0)!=0&a>>>0<G>>>0):0){zx(z,H,G-H|0)|0;M=G;break}zx(z,H,a-H|0)|0;L=a+1|0;I=rx(L,41)|0;if(!I){dy(z);k[m>>2]=125;o=125;return o|0}do if(q){y=I-L|0;if((y|0)<=1024){zx(d,L,y)|0;break}dy(z);k[m>>2]=125;o=125;return o|0}while(0);M=rx(I,91)|0}while(0);a=xx(z)|0;e:do if((a|0)>1){H=a;do{H=H+-1|0;G=z+H|0;if((i[G>>0]|0)!=32)break e;i[G>>0]=0}while((H|0)>1)}while(0);f:do if(q?(a=xx(d)|0,(a|0)>1):0){H=a;do{H=H+-1|0;a=d+H|0;if((i[a>>0]|0)!=32)break f;i[a>>0]=0}while((H|0)>1)}while(0);H=Ex(z,730846)|0;if((H|0)!=0?p&(i[H+4>>0]|0)==0:0){D=b;E=765674;F=D+11|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0))}H=xx(z)|0;a=H+-1|0;I=(H|0)>0;g:do if(I){G=a;while(1){L=z+G|0;if((i[L>>0]|0)==43){N=L;O=G;break}if((G|0)>0)G=G+-1|0;else{P=0;break g}}if((O|0)<=0){P=0;break}G=H-O|0;if((G|0)>=7){P=0;break}L=O+1|0;y=z+L|0;h:do if((L|0)<(H|0)){K=L;while(1){if(((i[z+K>>0]|0)+-48|0)>>>0>=10){Q=K;break h}J=K+1|0;if((J|0)<(H|0))K=J;else{Q=J;break}}}else Q=L;while(0);if((Q|0)!=(H|0)){P=0;break}do if(r){if((G|0)<=1024){Bx(e,y,G)|0;break}dy(z);k[m>>2]=125;o=125;return o|0}while(0);i[N>>0]=0;P=1}else P=0;while(0);i:do if(q){if((i[d>>0]|0)!=42|I^1)break;else R=a;while(1){if((R|0)==0?1:(i[z+R>>0]|0)==47){S=R;break}if((R|0)>0)R=R+-1|0;else break i}N=z+(S+1)|0;if((xx(N)|0)>>>0<=1024){ux(d,N)|0;break}dy(z);k[m>>2]=125;o=125;return o|0}while(0);do if(n){if((xx(z)|0)>>>0<=1024){ux(c,z)|0;break}dy(z);k[m>>2]=125;o=125;return o|0}while(0);if(!M){dy(z);o=k[m>>2]|0;return o|0}j:do if(!P){d=M+1|0;S=rx(d,93)|0;if(!S){tk(730851);dy(z);k[m>>2]=125;o=125;return o|0}R=i[d>>0]|0;switch(R<<24>>24){case 85:case 117:case 70:case 102:case 82:case 114:case 68:case 100:case 74:case 106:case 73:case 105:case 66:case 98:{a=i[M+2>>0]|0;switch(a<<24>>24){case 76:case 108:case 66:case 98:{T=i[M+3>>0]|0;U=4;break}default:{T=a;U=3}}k:do if(((T<<24>>24)+-48|0)>>>0<10){if(((T<<24>>24)+-48|0)>>>0<10){a=M+U|0;while(1){I=i[a>>0]|0;if(((I<<24>>24)+-48|0)>>>0<10)a=a+1|0;else{V=I;break}}}else V=T;switch(V<<24>>24){case 93:case 58:case 44:break;default:break k}do if(p)if(!(Ex(b,730780)|0)){D=b;E=727195;F=D+11|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0));break}else{D=b;E=730897;F=D+12|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0));break}while(0);do if(n){a=xx(c)|0;if(((xx(M)|0)+a|0)>>>0>1024){dy(z);k[m>>2]=125;o=125;return o|0}else{qx(c,M)|0;a=rx(c,93)|0;if(!a)break;i[a+1>>0]=0;break}}while(0);if(r){i[e>>0]=48;i[e+1>>0]=0}a=rx(S+1|0,91)|0;do if(u&(a|0)!=0){I=xx(f)|0;q=a+1|0;if(((xx(q)|0)+I|0)>>>0>1024){dy(z);k[m>>2]=125;o=125;return o|0}else{qx(f,q)|0;q=rx(f,93)|0;if(!q)break;i[q>>0]=0;break}}while(0);dy(z);o=k[m>>2]|0;return o|0}while(0);if(R<<24>>24==32)C=136;else{W=R;X=d}break}case 32:{C=136;break}default:{W=R;X=d}}if((C|0)==136){a=M+2|0;while(1){q=i[a>>0]|0;if(q<<24>>24==32)a=a+1|0;else{W=q;X=a;break}}}if(((W<<24>>24)+-48|0)>>>0<10){a=X;do{a=a+1|0;q=i[a>>0]|0}while(((q<<24>>24)+-48|0)>>>0<10);Y=q}else Y=W;switch(Y<<24>>24){case 45:case 42:case 58:{qx(A,M)|0;break j;break}default:{Z=R;_=d}}l:while(1){switch(Z<<24>>24){case 64:{$=Z;aa=_;C=163;break l;break}case 32:break;default:{ba=Z;ca=0;da=_;break l}}a=_+1|0;Z=i[a>>0]|0;_=a}if((C|0)==163){ba=$;ca=1;da=aa}R=(yx(da,730909,4)|0)==0;a=(yx(da,730914,3)|0)==0&1;q=(yx(da,730918,3)|0)==0;I=q&1;if((yx(da,730922,9)|0)!=0?(yx(da,730932,9)|0)!=0:0)ea=0;else ea=1;m:do if(yx(da,730942,10)|0){if(!(yx(da,730953,10)|0))break;N=S;H=d;Q=1-H+N|0;n:do if((Q|0)>0){O=ba;G=0;y=0;L=0;K=da;o:while(1){switch(O<<24>>24){case 46:{fa=1;ga=y;break}case 60:case 62:case 61:{C=173;break}case 124:{if((i[K+1>>0]|0)==124)C=173;else{fa=G;ga=y}break}case 38:{if((i[K+1>>0]|0)==38)C=173;else{fa=G;ga=y}break}default:switch(O<<24>>24){case 93:case 59:case 44:{ha=0;ia=0;ja=G;ka=y;break n;break}case 32:{la=G;ma=y;na=L;oa=K;break o;break}default:{fa=G;ga=y}}}if((C|0)==173){C=0;fa=G;ga=1}J=L+1|0;pa=K+1|0;if((J|0)>=(Q|0)){ha=0;ia=0;ja=fa;ka=ga;break n}O=i[pa>>0]|0;G=fa;y=ga;L=J;K=pa}K=32;L=oa;p:while(1){switch(K<<24>>24){case 93:{ha=0;ia=0;ja=la;ka=ma;break n;break}case 32:break;default:{qa=K;break p}}y=L+1|0;K=i[y>>0]|0;L=y}if(R)break m;if(!((I|a|0)==0|(na|0)>4))break m;switch(qa<<24>>24){case 37:case 47:case 42:case 45:case 43:case 33:case 38:case 124:case 60:case 62:case 61:{ra=1;break}default:ra=0}if(q)break m;else{ha=1;ia=ra;ja=la;ka=ma}}else{ha=0;ia=0;ja=0;ka=0}while(0);if((ca|0)!=0&(ja|0)!=0|(ka|ea|0)!=0|ha&(ia|0)!=0)break;do if(r){Q=N-H|0;if((Q|0)<=1024){zx(e,d,Q)|0;break}dy(z);k[m>>2]=125;o=125;return o|0}while(0);qx(A,S+1|0)|0;break j}while(0);qx(A,M)|0}else qx(A,M)|0;while(0);M=xx(A)|0;q:do if((M|0)>0){e=M;while(1){r=z+(e+x)|0;if((i[r>>0]|0)!=32)break q;i[r>>0]=0;if((e|0)>1)e=e+-1|0;else break}}while(0);if(!(i[A>>0]|0)){dy(z);o=k[m>>2]|0;return o|0}M=Ex(A,730964)|0;do if(!M){e=Ex(A,730969)|0;if(e){sa=e;C=201;break}e=Ex(A,730974)|0;if(e){sa=e;C=201}}else{sa=M;C=201}while(0);r:do if((C|0)==201){M=i[sa+4>>0]|0;switch(M<<24>>24){case 100:case 114:case 106:case 105:case 98:{ta=i[sa+5>>0]|0;break}default:ta=M}switch(ta<<24>>24){case 93:case 32:break;default:break r}do if(s){M=sa+1|0;if((xx(M)|0)>>>0>1024){dy(z);k[m>>2]=125;o=125;return o|0}ux(g,M)|0;M=rx(g,93)|0;if(!M){tk(730851);tk(A);dy(z);k[m>>2]=125;o=125;return o|0}else{i[M>>0]=0;e=M+-1|0;if((i[e>>0]|0)!=32)break;i[e>>0]=0;break}}while(0);ux(B,(rx(sa,93)|0)+1|0)|0;ux(sa,B)|0}while(0);sa=Ex(A,730979)|0;do if(!sa){g=Ex(A,730985)|0;if(g){ua=g;C=215;break}g=Ex(A,730991)|0;if(g){ua=g;C=215}}else{ua=sa;C=215}while(0);do if((C|0)==215){sa=ua+5|0;s:while(1){g=i[sa>>0]|0;switch(g<<24>>24){case 93:{va=sa;C=225;break s;break}case 0:{C=217;break s;break}case 39:{s=rx(sa+1|0,39)|0;if(!s){C=220;break s}wa=i[s>>0]|0;xa=s;break}default:{wa=g;xa=sa}}if(wa<<24>>24==91){g=rx(xa+1|0,93)|0;if(!g){C=223;break}else ya=g}else ya=xa;sa=ya+1|0}if((C|0)==217){tk(730851);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==220){tk(730997);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==223){tk(731062);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==225){sa=va-ua|0;g=sa+-1|0;do if(t)if((g|0)>1024){dy(z);k[m>>2]=125;o=125;return o|0}else{Bx(h,ua+1|0,g)|0;i[h+g>>0]=0;s=sa+-2|0;ta=h+s|0;if((i[ta>>0]|0)==32){za=ta;Aa=s}else break;do{i[za>>0]=0;Aa=Aa+-1|0;za=h+Aa|0}while((i[za>>0]|0)==32)}while(0);ux(B,va+1|0)|0;ux(ua,B)|0;break}}while(0);ua=Ex(A,731123)|0;do if(!ua){va=Ex(A,731128)|0;if(va){Ba=va;C=234;break}va=Ex(A,731133)|0;if(va){Ba=va;C=234}}else{Ba=ua;C=234}while(0);do if((C|0)==234){ua=Ba+4|0;va=i[ua>>0]|0;switch(va<<24>>24){case 68:case 82:case 100:case 114:case 74:case 73:case 66:case 106:case 105:case 98:{za=Ba+5|0;Ca=i[za>>0]|0;Da=6;Ea=za;break}default:{Ca=va;Da=5;Ea=ua}}ua=Ca<<24>>24==49?Ba+Da|0:Ea;if((i[ua>>0]|0)==32){Fa=32;Ga=ua}else break;t:while(1){switch(Fa<<24>>24){case 93:{Ha=Ga;C=246;break t;break}case 0:{C=238;break t;break}case 39:{ua=rx(Ga+1|0,39)|0;if(!ua){C=241;break t}Ia=i[ua>>0]|0;Ja=ua;break}default:{Ia=Fa;Ja=Ga}}if(Ia<<24>>24==91){ua=rx(Ja+1|0,93)|0;if(!ua){C=244;break}else Ka=ua}else Ka=Ja;ua=Ka+1|0;Fa=i[ua>>0]|0;Ga=ua}if((C|0)==238){tk(730851);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==241){tk(730997);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==244){tk(731062);dy(z);k[m>>2]=125;o=125;return o|0}else if((C|0)==246){ua=Ha-Ba|0;va=ua+-1|0;do if(v)if((va|0)>1024){dy(z);k[m>>2]=125;o=125;return o|0}else{Bx(j,Ba+1|0,va)|0;i[j+va>>0]=0;za=ua+-2|0;Aa=j+za|0;if((i[Aa>>0]|0)==32){La=Aa;Ma=za}else break;do{i[La>>0]=0;Ma=Ma+-1|0;La=j+Ma|0}while((i[La>>0]|0)==32)}while(0);ux(B,Ha+1|0)|0;ux(Ba,B)|0;break}}while(0);Ba=Ex(A,731138)|0;u:do if(Ba){switch(i[Ba+9>>0]|0){case 93:case 32:break;default:break u}do if(w){Ha=Ba+1|0;if((xx(Ha)|0)>>>0>1024){dy(z);k[m>>2]=125;o=125;return o|0}ux(l,Ha)|0;Ha=rx(l,93)|0;if(!Ha){tk(730851);tk(A);dy(z);k[m>>2]=125;o=125;return o|0}else{i[Ha>>0]=0;La=Ha+-1|0;if((i[La>>0]|0)!=32)break;i[La>>0]=0;break}}while(0);ux(B,(rx(Ba,93)|0)+1|0)|0;ux(Ba,B)|0}while(0);v:do if(u){B=i[A>>0]|0;if(!(B<<24>>24))break;Ba=z+((xx(A)|0)+x)|0;do if(B<<24>>24==91){if((i[Ba>>0]|0)!=93)break;i[Ba>>0]=0;l=z+(x+2)|0;if((xx(l)|0)>>>0<=1024){ux(f,l)|0;break v}dy(z);k[m>>2]=125;o=125;return o|0}while(0);tk(731148);k[m>>2]=125}while(0);dy(z);o=k[m>>2]|0;return o|0}function Kh(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0;n=r;r=r+9248|0;o=n+8200|0;p=n+8220|0;q=n+7175|0;s=n+6150|0;t=n+5125|0;u=n+4100|0;v=n+3075|0;w=n+2050|0;x=n+1025|0;y=n;k[l>>2]=0;if(!j){z=k[m>>2]|0;r=n;return z|0}do if(!(px(c,731197)|0)){Cp(d,y,m)|0;if((i[y>>0]|0)!=47){yp(w,m)|0;A=w+(xx(w)|0)|0;i[A>>0]=47;i[A+1>>0]=0;A=xx(w)|0;if(((xx(y)|0)+A|0)>>>0<=1024){qx(w,y)|0;zp(w,y,m)|0;break}tk(731205);k[m>>2]=104;z=104;r=n;return z|0}}else ux(y,d)|0;while(0);d=(j|0)==1;j=0;a:while(1){A=k[115100+(j<<2)>>2]|0;do if(A){B=A+12|0;Jh(k[B>>2]|0,o,p,s,q,t,u,v,0,0,m)|0;if((k[m>>2]|0)>0){C=B;D=12;break a}if(!(px(o,731197)|0)){Cp(p,x,m)|0;if((i[x>>0]|0)!=47){yp(w,m)|0;B=w+(xx(w)|0)|0;i[B>>0]=47;i[B+1>>0]=0;qx(w,x)|0;zp(w,x,m)|0}ux(p,x)|0}if((tx(c,o)|0)==0?(tx(y,p)|0)==0:0){if(!(((i[t>>0]|i[f>>0])<<24>>24==0?(i[u>>0]|i[g>>0])<<24>>24==0:0)?(i[v>>0]|i[h>>0])<<24>>24==0:0)){if(tx(f,t)|0)break;if(tx(g,u)|0)break;if(tx(h,v)|0)break;if(tx(e,q)|0)break}if(d?(k[A+76>>2]|0)==0:0){D=28;break a}B=ey(1,8)|0;k[a>>2]=B;if(!B){D=30;break a}k[B+4>>2]=A;k[B>>2]=0;E=(k[B+4>>2]|0)+8|0;k[E>>2]=(k[E>>2]|0)+1;if(i[g>>0]|0)i[e>>0]=0;i[f>>0]=0;i[g>>0]=0;i[h>>0]=0;k[l>>2]=1}}while(0);j=j+1|0;if((j|0)>=1e3){D=35;break}}if((D|0)==12){tk(731248);tk(k[C>>2]|0);z=k[m>>2]|0;r=n;return z|0}else if((D|0)==28){tk(731305);tk(b);k[m>>2]=104;z=104;r=n;return z|0}else if((D|0)==30){tk(728720);tk(b);k[m>>2]=113;z=113;r=n;return z|0}else if((D|0)==35){z=k[m>>2]|0;r=n;return z|0}return 0}function Lh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;d=r;r=r+2160|0;e=d;f=d+1129|0;g=d+104|0;h=d+24|0;j=d+20|0;k[a>>2]=0;l=k[c>>2]|0;if((l|0)>0){m=l;r=d;return m|0}if((l|0)==-106){k[c>>2]=0;n=1}else n=0;if((k[28269]|0)!=0?(l=Ch()|0,k[c>>2]=l,(l|0)>0):0){m=l;r=d;return m|0}else o=b;a:while(1){b=i[o>>0]|0;switch(b<<24>>24){case 0:{p=8;break a;break}case 32:break;default:{q=b;s=o;break a}}o=o+1|0}if((p|0)==8){tk(731366);k[c>>2]=105;m=105;r=d;return m|0}do if(!n){o=q<<24>>24==33;b=o?s+1|0:s;Zh(b,e,f,g,h,c)|0;if((k[c>>2]|0)>0){tk(731439);tk(b);m=k[c>>2]|0;r=d;return m|0}else{t=e;u=o&1;v=b}}else{if((xx(s)|0)>>>0<=1024){ux(f,s)|0;b=e;k[b>>2]=1701603686;k[b+4>>2]=3092282;i[g>>0]=0;i[h>>0]=0;t=e;u=0;v=s;break}tk(731408);k[c>>2]=105;m=105;r=d;return m|0}while(0);s=k[28270]|0;while(1){e=s+-1|0;if((s|0)<=0){p=18;break}if(!(tx(113084+(e*84|0)|0,t)|0)){w=e;break}else s=e}if((p|0)==18){k[c>>2]=124;tk(731485);tk(v);m=k[c>>2]|0;r=d;return m|0}k[c>>2]=0;if((u|0)!=0?(u=k[113084+(w*84|0)+60>>2]|0,(u|0)!=0):0)Ub[u&31](f)|0;u=k[113084+(w*84|0)+48>>2]|0;if(!u){tk(731576);tk(v);k[c>>2]=105;m=105;r=d;return m|0}p=Xb[u&31](f,j)|0;k[c>>2]=p;if(p){tk(731531);tk(v);m=k[c>>2]|0;r=d;return m|0}p=ey(1,8)|0;k[a>>2]=p;if(!p){Ub[k[113084+(w*84|0)+56>>2]&31](k[j>>2]|0)|0;tk(728720);tk(v);k[c>>2]=113;m=113;r=d;return m|0}f=ey(1,1728)|0;k[p+4>>2]=f;f=(k[a>>2]|0)+4|0;if(!(k[f>>2]|0)){Ub[k[113084+(w*84|0)+56>>2]&31](k[j>>2]|0)|0;tk(728720);tk(v);dy(k[a>>2]|0);k[a>>2]=0;k[c>>2]=113;m=113;r=d;return m|0}p=(xx(v)|0)+1|0;u=cy((p|0)>32?p:32)|0;k[(k[f>>2]|0)+12>>2]=u;u=(k[a>>2]|0)+4|0;if(!(k[(k[u>>2]|0)+12>>2]|0)){Ub[k[113084+(w*84|0)+56>>2]&31](k[j>>2]|0)|0;tk(731624);tk(v);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[c>>2]=105;m=105;r=d;return m|0}f=ey(1001,8)|0;k[(k[u>>2]|0)+88>>2]=f;f=(k[a>>2]|0)+4|0;if(!(k[(k[f>>2]|0)+88>>2]|0)){Ub[k[113084+(w*84|0)+56>>2]&31](k[j>>2]|0)|0;tk(731673);tk(v);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[c>>2]=113;m=113;r=d;return m|0}u=ey(40,2880)|0;k[(k[f>>2]|0)+1244>>2]=u;u=k[(k[a>>2]|0)+4>>2]|0;if(!(k[u+1244>>2]|0)){Ub[k[113084+(w*84|0)+56>>2]&31](k[j>>2]|0)|0;tk(731729);tk(v);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+88>>2]|0);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[c>>2]=113;m=113;r=d;return m|0}else x=0;do{k[u+1568+(x<<2)>>2]=x;k[u+1248+(x<<2)>>2]=-1;x=x+1|0}while((x|0)!=40);k[u+84>>2]=1e3;k[u>>2]=k[j>>2];k[u+4>>2]=w;ux(k[u+12>>2]|0,v)|0;v=k[a>>2]|0;u=k[v+4>>2]|0;w=u+24|0;k[w>>2]=0;k[w+4>>2]=0;k[w+8>>2]=0;k[w+12>>2]=0;k[u+76>>2]=1;w=u+120|0;k[w>>2]=-1;k[w+4>>2]=-1;k[u+64>>2]=-1;k[u+8>>2]=1;k[u+16>>2]=555;bu(v,0,1,c)|0;v=k[(k[a>>2]|0)+4>>2]|0;b:do if((k[c>>2]|0)<=0){u=0;while(1){w=115100+(u<<2)|0;u=u+1|0;if(!(k[w>>2]|0)){y=w;break}if((u|0)>=1e3)break b}k[y>>2]=v}while(0);if(i[g>>0]|0)_h(k[a>>2]|0,g,c)|0;if(i[h>>0]|0)Rh(k[a>>2]|0,h,c)|0;m=k[c>>2]|0;r=d;return m|0}function Mh(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;f=r;r=r+30912|0;g=f+128|0;h=f+912|0;j=f+724|0;l=f+720|0;m=f+716|0;n=f+712|0;o=f+148|0;p=f+144|0;q=f+56|0;s=f+48|0;t=f+40|0;u=f+32|0;v=f+24|0;w=f+16|0;x=f+140|0;y=f+136|0;z=f+8|0;A=f;B=f+728|0;C=f+829|0;D=f+748|0;E=f+152|0;F=D;G=F+81|0;do{i[F>>0]=0;F=F+1|0}while((F|0)<(G|0));py(E|0,119112,560)|0;F=k[e>>2]|0;if((F|0)>0){H=F;r=f;return H|0}if((Tk(a,0,c,l,e)|0)>0){tk(731784);tk(c);H=k[e>>2]|0;r=f;return H|0}if((nl(a,k[l>>2]|0,d,((d|0)<0)<<31>>31,1,0,1,0,0,z,A,B,x,m,o,t,u,y,s,v,j,w,h,e)|0)>0){H=k[e>>2]|0;r=f;return H|0}Gk(795593,k[l>>2]|0,D,p)|0;Uk(a,0,D,c,l,p)|0;if((k[j>>2]|0)!=2){tk(731838);tk(731876);k[e>>2]=227;H=227;r=f;return H|0}j=k[m>>2]|0;if((j|0)<0){k[m>>2]=0-j;k[n>>2]=1;j=s;p=k[j+4>>2]|0;D=q;k[D>>2]=k[j>>2];k[D+4>>2]=p}else Yo(a,k[l>>2]|0,9,n,q,e)|0;if((k[e>>2]|0)>0){tk(731923);H=k[e>>2]|0;r=f;return H|0}switch(k[m>>2]|0){case 11:{m=s;I=k[m+4>>2]|0;J=k[m>>2]|0;K=8;break}case 21:{m=s;p=ty(k[m>>2]|0,k[m+4>>2]|0,1)|0;I=L;J=p;K=16;break}case 41:{p=s;m=ty(k[p>>2]|0,k[p+4>>2]|0,2)|0;I=L;J=m;K=32;break}case 42:{m=s;p=ty(k[m>>2]|0,k[m+4>>2]|0,2)|0;I=L;J=p;K=-32;break}case 82:{p=s;m=ty(k[p>>2]|0,k[p+4>>2]|0,3)|0;I=L;J=m;K=-64;break}case 81:{m=s;p=ty(k[m>>2]|0,k[m+4>>2]|0,3)|0;I=L;J=p;K=64;break}case 14:{p=s;I=k[p+4>>2]|0;J=k[p>>2]|0;K=8;break}default:{tk(731965);tk(c);tk(B);tk(732021);k[e>>2]=261;H=261;r=f;return H|0}}if((Zr(b,K,k[n>>2]|0,q,e)|0)>0){tk(732074);H=k[e>>2]|0;r=f;return H|0}Kk(a,b,9,E,70,k[l>>2]|0,0,0,e)|0;k[g>>2]=d;k[g+4>>2]=c;Xw(C,732141,g)|0;hu(a,0,e)|0;g=t;au(a,k[g>>2]|0,k[g+4>>2]|0,1,e)|0;g=(I|0)>0|(I|0)==0&J>>>0>3e4;t=g?3e4:J;C=g?0:I;fu(a,t,C,h,e)|0;ru(b,1,0,1,0,t,C,h,e)|0;a:do if(!((J|0)==(t|0)&(I|0)==(C|0))){g=jy(J|0,I|0,t|0,C|0)|0;c=L;d=ky(t|0,C|0,1,0)|0;l=a+4|0;E=c;c=g;g=d;d=L;while(1){if((k[e>>2]|0)>=1)break a;q=(E|0)>0|(E|0)==0&c>>>0>3e4;n=q?3e4:c;K=q?0:E;q=k[l>>2]|0;B=Pb[k[113084+((k[q+4>>2]|0)*84|0)+76>>2]&63](k[q>>2]|0,h,n)|0;if((B|0)!=107){if((B|0)>0){tk(732201);tk(k[q+12>>2]|0);k[e>>2]=108}}else k[e>>2]=107;ru(b,1,0,g,d,n,K,h,e)|0;q=c;c=jy(c|0,E|0,n|0,K|0)|0;B=E;E=L;g=ky(n|0,K|0,g|0,d|0)|0;if((q|0)==(n|0)&(B|0)==(K|0))break;else d=L}}while(0);Vk(b,e)|0;H=k[e>>2]|0;r=f;return H|0}function Nh(a){a=a|0;var b=0;if(((((Ax(a,732238,3)|0)!=0?(Ax(a,732242,8)|0)!=0:0)?(Ax(a,732251,4)|0)!=0:0)?(Ax(a,732256,3)|0)!=0:0)?(Ax(a,732260,6)|0)!=0:0){i[a>>0]=i[730780]|0;i[a+1>>0]=i[730781]|0;i[a+2>>0]=i[730782]|0;i[a+3>>0]=i[730783]|0;i[a+4>>0]=i[730784]|0;b=(a|0)==0&1}else b=1;return b|0}function Oh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0;e=r;r=r+480|0;f=e+24|0;g=e+20|0;h=e+16|0;j=e+12|0;l=e+8|0;m=e+4|0;n=e+393|0;o=e+322|0;p=e+251|0;q=e+180|0;s=e;t=e+109|0;u=e+28|0;k[h>>2]=-1;k[l>>2]=0;k[m>>2]=0;k[s>>2]=0;if(i[b>>0]|0){if((Lh(f,b,d)|0)>0){tk(732267);v=k[d>>2]|0;r=e;return v|0}bl(k[a>>2]|0,g)|0;b=k[a>>2]|0;w=(k[(k[b+4>>2]|0)+20>>2]|0)==0;x=Fk(b,1,0,d)|0;do if(w){if((x|0)<=0){b=1;do{Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0;b=b+1|0}while((Fk(k[a>>2]|0,b,0,d)|0)<=0)}b=k[d>>2]|0;if((b|0)==107){k[d>>2]=0;break}if((b|0)>0){Fh(k[f>>2]|0,d)|0;tk(732315);v=k[d>>2]|0;r=e;return v|0}}else{Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0;Fk(k[a>>2]|0,k[g>>2]|0,0,d)|0;Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0;if((k[d>>2]|0)<=0){k[g>>2]=2;break}Fh(k[f>>2]|0,d)|0;tk(732315);v=k[d>>2]|0;r=e;return v|0}while(0);Fh(k[a>>2]|0,d)|0;x=k[f>>2]|0;k[a>>2]=x;if((Fk(x,k[g>>2]|0,0,d)|0)>0){tk(732372);v=k[d>>2]|0;r=e;return v|0}}g=c+4|0;a:while(1){switch(i[g>>0]|0){case 64:{y=g;z=19;break a;break}case 32:break;default:{A=g;z=24;break a}}g=g+1|0}if((z|0)==19){if(Uh(y+1|0,s,d)|0){v=k[d>>2]|0;r=e;return v|0}y=k[s>>2]|0;g=y;if((i[g>>0]|0)==32){c=g;while(1){x=c+1|0;if((i[x>>0]|0)==32)c=x;else{B=x;break}}C=B;D=g}else{C=y;D=g}}else if((z|0)==24){C=A;D=0}k[m>>2]=0;$k(k[a>>2]|0,l,m)|0;A=C;if(Th(A)|0){tk(732419);tk(A);if(D)dy(D);k[d>>2]=431;v=431;r=e;return v|0}D=o+1|0;b:do if(!(k[d>>2]|0)){A=C;g=0;y=0;B=0;c:while(1){c=A;if((i[c>>0]|0)==32){x=c;while(1){f=x+1|0;if((i[f>>0]|0)==32)x=f;else{E=f;break}}F=E;G=E}else{F=c;G=A}x=vx(F,728157)|0;if(!x){H=0;I=g;J=y;K=B;break b}f=x+1|0;w=ey(f,1)|0;if(!w){L=g;M=y;N=B;z=35;break}b=G;zx(w,b,x)|0;O=b+x|0;if((x|0)<=0){H=w;I=g;J=y;K=B;break b}A=(i[O>>0]|0)==59?b+f|0:O;i[w+x>>0]=0;x=i[w>>0]|0;d:do switch(x<<24>>24){case 45:case 33:{O=w+1|0;if((Tk(k[a>>2]|0,0,O,h,d)|0)>=1){xk();k[d>>2]=0;if((Aq(k[a>>2]|0,O,d)|0)>0){P=w;Q=g;z=47;break c}else{R=g;S=y;T=B;break d}}if((Vj(k[a>>2]|0,k[h>>2]|0,d)|0)>0){U=w;V=g;z=40;break c}k[l>>2]=(k[l>>2]|0)+-1;k[h>>2]=-1;R=g;S=1;T=B;break}default:{i[o>>0]=0;if(x<<24>>24==32){O=w;while(1){f=O+1|0;if((i[f>>0]|0)==32)O=f;else{W=f;break}}}else W=w;O=vx(W,732613)|0;if(!O){X=w;Y=g;z=55;break c}zx(o,W,O)|0;f=W+O|0;if((i[o>>0]|0)==35?(O=rx(D,35)|0,(O|0)==(o+((xx(o)|0)+-1)|0)):0){if((k[h>>2]|0)<1){Z=w;z=63;break c}i[O>>0]=0;Gk(D,k[h>>2]|0,p,d)|0;O=k[d>>2]|0;if(O){v=O;z=162;break c}ux(D,p)|0}else z=68;e:do if((z|0)==68?(z=0,O=rx(o,35)|0,b=(O|0)==(o+((xx(o)|0)+-1)|0),(k[h>>2]|0)>0&b):0){k[m>>2]=0;Tk(k[a>>2]|0,0,o,j,m)|0;switch(k[m>>2]|0){case 0:case 237:{break e;break}default:{}}xk();ux(t,o)|0;b=t+((xx(t)|0)+-1)|0;i[b>>0]=0;Gk(t,k[h>>2]|0,p,d)|0;if(k[d>>2]|0){_=w;z=71;break c}k[m>>2]=0;if(!(Go(k[a>>2]|0,p,u,m)|0))ux(o,p)|0}while(0);if((i[f>>0]|0)==40){i[p>>0]=0;if((i[f>>0]|0)==32){b=f;while(1){O=b+1|0;if((i[O>>0]|0)==32)b=O;else{$=O;break}}}else $=f;b=vx($,761359)|0;if(!b)aa=$;else{zx(p,$,b)|0;aa=$+b|0}qx(o,p)|0;b=o+(xx(o)|0)|0;i[b>>0]=41;i[b+1>>0]=0;ba=aa+1|0}else ba=f;f:while(1){switch(i[ba>>0]|0){case 61:{ca=ba;break f;break}case 32:break;default:{da=ba;z=84;break f}}ba=ba+1|0}if((z|0)==84){z=0;Tk(k[a>>2]|0,0,o,j,d)|0;f=k[d>>2]|0;g:do if((f|0)==237){b=g;h:while(1){O=k[j>>2]|0;k[h>>2]=O;if(!b)ea=ey(999,4)|0;else ea=b;k[ea+(O+-1<<2)>>2]=1;Tk(k[a>>2]|0,0,o,j,d)|0;O=k[d>>2]|0;switch(O|0){case 237:{b=ea;break}case 219:{fa=ea;break h;break}default:{ga=O;ha=ea;ia=1;z=89;break g}}}k[d>>2]=999;ja=fa;ka=1}else{ga=f;ha=g;ia=B;z=89}while(0);do if((z|0)==89){z=0;if((ga|0)>=1)if((ga|0)==999){ja=ha;ka=ia;break}else{la=da;ma=w;na=ha;z=95;break c}f=k[j>>2]|0;k[h>>2]=f;if(!ha)oa=ey(999,4)|0;else oa=ha;k[oa+(f+-1<<2)>>2]=1;R=oa;S=y;T=1;break d}while(0);k[d>>2]=0;R=ja;S=y;T=ka;break d}f=ca+1|0;if((i[f>>0]|0)==61){b=ca+2|0;while(1)if((i[b>>0]|0)==32)b=b+1|0;else{pa=b;break}i[p>>0]=0;if((i[pa>>0]|0)==32){b=pa;while(1){O=b+1|0;if((i[O>>0]|0)==32)b=O;else{qa=O;break}}}else qa=pa;b=vx(qa,776273)|0;if(b)zx(p,qa,b)|0;if((Tk(k[a>>2]|0,0,p,h,d)|0)>=1){xk();k[d>>2]=0;if((yq(k[a>>2]|0,p,o,d)|0)>0){ra=w;sa=g;z=118;break c}else{R=g;S=y;T=B;break d}}Gk(772997,k[h>>2]|0,n,d)|0;if((mq(k[a>>2]|0,n,o,0,d)|0)>0){ta=w;ua=g;z=109;break c}if(!g)va=ey(999,4)|0;else va=g;k[va+((k[h>>2]|0)+-1<<2)>>2]=1;R=va;S=y;T=1;break d}i[q>>0]=0;i[p>>0]=0;b=i[o>>0]|0;if(b<<24>>24==32){O=o;while(1){wa=O+1|0;xa=i[wa>>0]|0;if(xa<<24>>24==32)O=wa;else{ya=wa;za=xa;break}}}else{ya=o;za=b}O=vx(ya,732921)|0;if(!O){Aa=za;Ba=ya}else{zx(p,ya,O)|0;xa=ya+O|0;Aa=i[xa>>0]|0;Ba=xa}do if(Aa<<24>>24==40){xa=Ba+1|0;i[q>>0]=0;if((i[xa>>0]|0)==32){O=xa;while(1){wa=O+1|0;if((i[wa>>0]|0)==32)O=wa;else{Ca=wa;break}}}else Ca=xa;O=vx(Ca,761359)|0;if(!O)break;zx(q,Ca,O)|0}while(0);b=k[a>>2]|0;if((ek(b,f,b,p,q,d)|0)>0){Da=w;Ea=g;z=133;break c}k[m>>2]=0;Tk(k[a>>2]|0,0,p,j,m)|0;if(k[m>>2]|0){xk();R=g;S=y;T=B;break d}b=k[j>>2]|0;k[h>>2]=b;if(!g)Fa=ey(999,4)|0;else Fa=g;k[Fa+(b+-1<<2)>>2]=1;b=k[l>>2]|0;if((k[h>>2]|0)<=(b|0)){R=Fa;S=y;T=1;break d}k[l>>2]=b+1;R=Fa;S=y;T=1}}while(0);dy(w);if(k[d>>2]|0){H=0;I=R;J=S;K=T;break b}else{g=R;y=S;B=T}}switch(z|0){case 35:{tk(732461);k[d>>2]=113;H=0;I=L;J=M;K=N;break b;break}case 40:{tk(732526);tk(U);if(V)dy(V);B=k[s>>2]|0;if(B)dy(B);dy(U);v=k[d>>2]|0;r=e;return v|0}case 47:{tk(732565);tk(P);if(Q)dy(Q);B=k[s>>2]|0;if(B)dy(B);dy(P);v=k[d>>2]|0;r=e;return v|0}case 55:{tk(732617);tk(X);if(Y)dy(Y);B=k[s>>2]|0;if(B)dy(B);dy(X);k[d>>2]=125;v=125;r=e;return v|0}case 63:{tk(732657);tk(o);tk(732675);tk(732722);B=k[s>>2]|0;if(B)dy(B);dy(Z);k[d>>2]=125;v=125;r=e;return v|0}case 71:{B=k[s>>2]|0;if(B)dy(B);dy(_);v=k[d>>2]|0;r=e;return v|0}case 95:{tk(732765);tk(la);if(na)dy(na);B=k[s>>2]|0;if(B)dy(B);dy(ma);k[d>>2]=125;v=125;r=e;return v|0}case 109:{tk(732813);tk(732851);tk(p);tk(732862);tk(o);if(ua)dy(ua);B=k[s>>2]|0;if(B)dy(B);dy(ta);v=k[d>>2]|0;r=e;return v|0}case 118:{tk(732873);tk(ra);if(sa)dy(sa);B=k[s>>2]|0;if(B)dy(B);dy(ra);v=k[d>>2]|0;r=e;return v|0}case 133:{tk(732923);if(Ea)dy(Ea);B=k[s>>2]|0;if(B)dy(B);dy(Da);v=k[d>>2]|0;r=e;return v|0}case 162:{r=e;return v|0}}}else{H=0;I=0;J=0;K=0}while(0);i:do if(!((J|0)!=0|(K|0)==0)?(Da=k[l>>2]|0,(Da|0)>0):0){Ea=Da;while(1){Da=Ea;Ea=Ea+-1|0;if((k[I+(Ea<<2)>>2]|0)==0?(Vj(k[a>>2]|0,Da,d)|0)>0:0)break;if((Da|0)<=1)break i}tk(732526);tk(H);dy(I);Ea=k[s>>2]|0;if(Ea)dy(Ea);if(H)dy(H);v=k[d>>2]|0;r=e;return v|0}while(0);if(I)dy(I);I=k[s>>2]|0;if(I)dy(I);if(H)dy(H);v=k[d>>2]|0;r=e;return v|0}function Ph(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;e=r;r=r+16|0;f=e+4|0;g=e;if((Lh(f,b,d)|0)>0){tk(732954);tk(b);h=k[d>>2]|0;r=e;return h|0}bl(k[a>>2]|0,g)|0;b=k[a>>2]|0;if(!(k[(k[b+4>>2]|0)+20>>2]|0)){i=k[g>>2]|0;a:do if((i|0)>1){j=b;l=1;while(1){Fk(j,l,0,d)|0;l=l+1|0;if((Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0)>0)break;m=k[g>>2]|0;n=k[a>>2]|0;if((l|0)>=(m|0)){o=m;p=n;break a}else j=n}Fh(k[f>>2]|0,d)|0;h=k[d>>2]|0;r=e;return h|0}else{o=i;p=b}while(0);Fk(p,o,0,d)|0;q=k[a>>2]|0}else q=b;if((Xh(q,k[f>>2]|0,c,d)|0)>0){Fh(k[f>>2]|0,d)|0;h=k[d>>2]|0;r=e;return h|0}c=k[a>>2]|0;q=(k[g>>2]|0)+1|0;do if(!(k[(k[c+4>>2]|0)+20>>2]|0)){if((Fk(c,q,0,d)|0)>0)s=q;else{b=q;while(1){Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0;o=b+1|0;if((Fk(k[a>>2]|0,o,0,d)|0)>0){s=o;break}else b=o}}b=k[d>>2]|0;if((b|0)==107){k[d>>2]=0;t=s;break}if((b|0)>0){Fh(k[f>>2]|0,d)|0;h=k[d>>2]|0;r=e;return h|0}else t=s}else t=q;while(0);Fh(k[a>>2]|0,d)|0;q=k[f>>2]|0;k[a>>2]=q;f=k[g>>2]|0;if((t+-1|0)!=(f|0)){Fk(q,f,0,d)|0;h=k[d>>2]|0;r=e;return h|0}if((Vk(q,d)|0)<=0){h=k[d>>2]|0;r=e;return h|0}Fh(k[a>>2]|0,d)|0;h=k[d>>2]|0;r=e;return h|0}function Qh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;e=r;r=r+16|0;f=e+4|0;g=e;do if(!(i[b>>0]|0)){h=k[a>>2]|0;k[f>>2]=h;j=h;l=j;m=j}else{if((Lh(f,b,d)|0)>0){tk(733002);tk(b);n=k[d>>2]|0;r=e;return n|0}bl(k[a>>2]|0,g)|0;j=k[a>>2]|0;a:do if(!(k[(k[j+4>>2]|0)+20>>2]|0)){h=k[g>>2]|0;if((h|0)>1){Fk(j,1,0,d)|0;if((Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0)<=0){o=2;while(1){p=k[g>>2]|0;if((o|0)>=(p|0)){q=p;break a}Fk(k[a>>2]|0,o,0,d)|0;if((Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0)>0)break;else o=o+1|0}}Fh(k[f>>2]|0,d)|0;n=k[d>>2]|0;r=e;return n|0}else q=h}else{Fk(j,1,0,d)|0;if((Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0)<=0){q=k[g>>2]|0;break}Fh(k[f>>2]|0,d)|0;n=k[d>>2]|0;r=e;return n|0}while(0);Fk(k[a>>2]|0,q,0,d)|0;j=(Xj(k[a>>2]|0,k[f>>2]|0,d)|0)>0;o=k[f>>2]|0;if(j){Fh(o,d)|0;n=k[d>>2]|0;r=e;return n|0}oq(o,790243,0,0,0,d)|0;o=k[f>>2]|0;j=(k[o+4>>2]|0)+936|0;k[j>>2]=0;k[j+4>>2]=0;k[j+8>>2]=0;k[j+12>>2]=0;if((Vk(o,d)|0)<=0){l=k[a>>2]|0;m=k[f>>2]|0;break}Fh(k[f>>2]|0,d)|0;n=k[d>>2]|0;r=e;return n|0}while(0);q=(dk(l,m,c,d)|0)>0;c=(i[b>>0]|0)!=0;if(q){if(!c){n=k[d>>2]|0;r=e;return n|0}Fh(k[f>>2]|0,d)|0;n=k[d>>2]|0;r=e;return n|0}if(!c){n=k[d>>2]|0;r=e;return n|0}c=k[a>>2]|0;do if(!(k[(k[c+4>>2]|0)+20>>2]|0)){q=(k[g>>2]|0)+1|0;if((Fk(c,q,0,d)|0)<=0){b=q;do{Wj(k[a>>2]|0,k[f>>2]|0,0,d)|0;b=b+1|0}while((Fk(k[a>>2]|0,b,0,d)|0)<=0)}b=k[d>>2]|0;if((b|0)==107){k[d>>2]=0;break}if((b|0)>0){Fh(k[f>>2]|0,d)|0;n=k[d>>2]|0;r=e;return n|0}}else k[g>>2]=2;while(0);Fh(k[a>>2]|0,d)|0;c=k[f>>2]|0;k[a>>2]=c;Fk(c,k[g>>2]|0,0,d)|0;n=k[d>>2]|0;r=e;return n|0}function Rh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0.0,M=0,N=0.0,O=0.0,P=0.0,Q=0,R=0.0,S=0,T=0.0,U=0,V=0,W=0.0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0;d=r;r=r+32|0;e=d+24|0;f=d;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;k[f+16>>2]=0;k[f+20>>2]=0;g=b;while(1){k[e>>2]=g;if((i[g>>0]|0)==32)g=g+1|0;else{h=g;break}}if((Ax(h,732242,8)|0)!=0?(Ax(h,733059,8)|0)!=0:0){k[c>>2]=125;j=125;r=d;return j|0}g=h+8|0;a:while(1){k[e>>2]=g;h=i[g>>0]|0;switch(h<<24>>24){case 82:case 114:{l=h;m=g;n=9;break a;break}case 71:case 103:{o=h;p=g;n=11;break a;break}case 80:case 112:{q=h;s=g;n=13;break a;break}case 72:case 104:{t=g;n=15;break a;break}case 32:break;default:{u=h;v=g;w=11;x=0;break a}}g=g+1|0}b:do if((n|0)==9)while(1){n=0;switch(l<<24>>24){case 0:case 59:case 32:{u=l;v=m;w=11;x=0;break b;break}default:{}}g=m+1|0;k[e>>2]=g;l=i[g>>0]|0;m=g;n=9}else if((n|0)==11)while(1){n=0;switch(o<<24>>24){case 0:case 59:case 32:{u=o;v=p;w=21;x=0;break b;break}default:{}}g=p+1|0;k[e>>2]=g;o=i[g>>0]|0;p=g;n=11}else if((n|0)==13)while(1){n=0;switch(q<<24>>24){case 0:case 59:case 32:{u=q;v=s;w=31;x=0;break b;break}default:{}}g=s+1|0;k[e>>2]=g;q=i[g>>0]|0;s=g;n=13}else if((n|0)==15){g=t+1|0;k[e>>2]=g;h=i[g>>0]|0;switch(h<<24>>24){case 83:case 115:{y=1;break}default:y=0}b=h;h=g;while(1){switch(b<<24>>24){case 0:case 59:case 32:{u=b;v=h;w=41;x=y;break b;break}default:{}}g=h+1|0;k[e>>2]=g;b=i[g>>0]|0;h=g}}while(0);if(u<<24>>24==32){y=v;while(1){t=y+1|0;k[e>>2]=t;n=i[t>>0]|0;if(n<<24>>24==32)y=t;else{z=n;A=t;break}}}else{z=u;A=v}if(((z<<24>>24)+-48|0)>>>0<10){v=A;u=0;while(1){y=fx(v)|0;k[f+(u<<2)>>2]=y;u=u+1|0;y=i[v>>0]|0;if(((y<<24>>24)+-48|0)>>>0<10){t=v;while(1){n=t+1|0;k[e>>2]=n;s=i[n>>0]|0;if(((s<<24>>24)+-48|0)>>>0<10)t=n;else{B=n;C=s;break}}}else{B=v;C=y}if(C<<24>>24==44){t=B+1|0;k[e>>2]=t;D=i[t>>0]|0;E=t}else{D=C;E=B}if(D<<24>>24==32){t=E;while(1){s=t+1|0;k[e>>2]=s;n=i[s>>0]|0;if(n<<24>>24==32)t=s;else{F=n;G=s;break}}}else{F=D;G=E}if(!((u|0)<9&((F<<24>>24)+-48|0)>>>0<10)){H=G;I=F;break}else v=G}}else{H=A;I=z}c:do if(I<<24>>24==59){z=H;while(1){A=z+1|0;k[e>>2]=A;G=i[A>>0]|0;if(G<<24>>24==32)z=A;else{J=G;K=A;L=-99.0;M=1;N=0.0;break}}d:while(1){z=J;A=K;O=N;e:while(1){switch(z<<24>>24){case 0:{P=L;Q=M;R=O;break c;break}case 81:case 113:{S=A;T=O;break e;break}case 83:case 115:{U=A;break}default:break d}while(1){G=U+1|0;k[e>>2]=G;if((i[G>>0]|0)==32)U=G;else{V=G;break}}W=+hx(V,e);G=k[e>>2]|0;f:while(1){v=i[G>>0]|0;switch(v<<24>>24){case 44:case 32:break;default:{X=G;Y=v;break f}}v=G+1|0;k[e>>2]=v;G=v}z=Y;A=X;O=W}A=S+1|0;k[e>>2]=A;switch(i[A>>0]|0){case 90:case 122:{z=S+2|0;k[e>>2]=z;Z=z;_=2;break}case 48:{z=S+2|0;k[e>>2]=z;Z=z;_=-1;break}default:{Z=A;_=M}}if((i[Z>>0]|0)==32){A=Z;while(1){z=A+1|0;k[e>>2]=z;if((i[z>>0]|0)==32)A=z;else{$=z;break}}}else $=Z;O=+hx($,e);A=k[e>>2]|0;g:while(1){z=i[A>>0]|0;switch(z<<24>>24){case 44:case 32:break;default:{aa=A;ba=z;break g}}z=A+1|0;k[e>>2]=z;A=z}J=ba;K=aa;L=O;M=_;N=T}k[c>>2]=125;j=125;r=d;return j|0}else{P=-99.0;Q=1;R=0.0}while(0);Ts(a,w,c)|0;Us(a,6,f,c)|0;if((w|0)==41){Xs(a,R,c)|0;Ys(a,x,c)|0}if(P!=-99.0){Vs(a,P,c)|0;Ws(a,Q,c)|0}j=k[c>>2]|0;r=d;return j|0}function Sh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}if(!a){k[c>>2]=115;e=115;return e|0}d=a+4|0;if((k[(k[d>>2]|0)+16>>2]|0)==555){a=ey(1,8)|0;k[b>>2]=a;b=k[d>>2]|0;k[a+4>>2]=b;k[a>>2]=0;a=b+8|0;k[a>>2]=(k[a>>2]|0)+1;e=k[c>>2]|0;return e|0}else{k[c>>2]=114;e=114;return e|0}return 0}function Th(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;b=r;r=r+16|0;c=b;k[c>>2]=a;d=a;a:while(1){a=i[d>>0]|0;switch(a<<24>>24){case 0:{e=0;f=16;break a;break}case 44:{i[d>>0]=59;k[c>>2]=d+1;break}case 40:{k[c>>2]=d+1;if(ii(c)|0){e=1;f=16;break a}break}case 91:{k[c>>2]=d+1;if(ji(c)|0){e=1;f=16;break a}break}case 123:{k[c>>2]=d+1;if(ki(c)|0){e=1;f=16;break a}break}case 34:{g=d+1|0;k[c>>2]=g;h=g;while(1){g=i[h>>0]|0;if(!(g<<24>>24)){e=1;f=16;break a}j=h+1|0;if(g<<24>>24==34){l=j;break}else h=j}k[c>>2]=l;break}default:{h=d+1|0;k[c>>2]=h;if(a<<24>>24==39){j=h;while(1){h=i[j>>0]|0;if(!(h<<24>>24)){e=1;f=16;break a}g=j+1|0;if(h<<24>>24==39){m=g;break}else j=g}k[c>>2]=m}}}d=k[c>>2]|0}if((f|0)==16){r=b;return e|0}return 0}function Uh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;d=r;r=r+272|0;e=d;f=d+8|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=cy(1024)|0;if(!g){tk(733068);k[c>>2]=113;h=113;r=d;return h|0}i[g>>0]=0;j=Dw(a,774853)|0;if(!j){k[e>>2]=a;Xw(f,733122,e)|0;tk(f);dy(g);k[c>>2]=104;h=104;r=d;return h|0}e=f+1|0;a=1024;l=0;m=g;g=0;a:while(1){b:do if(l){if(!(Bw(f,256,j)|0)){n=m;break a}o=xx(f)|0;p=o;q=(o|0)>1}else while(1){if(!(Bw(f,256,j)|0)){n=m;break a}o=xx(f)|0;s=(o|0)>1;if(!(s&(i[f>>0]|0)==47)){p=o;q=s;break b}if((i[e>>0]|0)!=47){p=o;q=1;break}}while(0);c:do if((p|0)>0){o=p+-1|0;s=f+o|0;switch(i[s>>0]|0){case 13:case 10:break;default:{t=0;u=p;break c}}i[s>>0]=0;if(q){s=p+-2|0;v=f+s|0;switch(i[v>>0]|0){case 13:case 10:break;default:{t=1;u=o;break c}}i[v>>0]=0;t=1;u=s}else{t=1;u=o}}else{t=0;u=p}while(0);o=u+g|0;if((o+3|0)>=(a|0)){s=a+256|0;v=fy(m,s)|0;if(!v){w=15;break}else{x=s;y=v}}else{x=a;y=m}ux(y+g|0,f)|0;if(!t){a=x;l=1;m=y;g=o;continue}v=y+o|0;i[v>>0]=32;i[v+1>>0]=0;a=x;l=0;m=y;g=o+1|0}if((w|0)==15){tk(733068);k[c>>2]=113;n=0}ww(j)|0;k[b>>2]=n;h=k[c>>2]|0;r=d;return h|0}function Vh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0;e=r;r=r+80|0;f=e;g=e+4|0;i[c>>0]=0;h=k[a>>2]|0;if((i[h>>0]|0)==32){j=h;while(1){l=j+1|0;k[a>>2]=l;if((i[l>>0]|0)==32)j=l;else{m=l;break}}}else m=h;h=vx(m,b)|0;if(!h){r=e;return h|0}zx(c,m,h)|0;k[a>>2]=(k[a>>2]|0)+h;if(!d){r=e;return h|0}k[d>>2]=1;if(!(rx(c,68)|0))+hx(c,f);else{Bx(g,c,72)|0;i[g+72>>0]=0;c=rx(g,68)|0;k[f>>2]=c;if(c)i[c>>0]=69;+hx(g,f)}switch(i[k[f>>2]>>0]|0){case 32:case 0:break;default:k[d>>2]=0}f=rv()|0;if((k[f>>2]|0)!=34){r=e;return h|0}k[d>>2]=0;r=e;return h|0}function Wh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=Pb[k[113084+((k[a+4>>2]|0)*84|0)+76>>2]&63](k[a>>2]|0,c,b)|0;if((e|0)==107){k[d>>2]=107;f=107;return f|0}if((e|0)>0){tk(732201);tk(k[a+12>>2]|0);k[d>>2]=108;f=108;return f|0}else{f=k[d>>2]|0;return f|0}return 0}function Xh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0.0,N=0.0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;e=r;r=r+400|0;f=e+192|0;g=e+152|0;h=e+148|0;j=e+196|0;l=e+72|0;m=e+68|0;n=e+28|0;o=e+156|0;q=e+112|0;s=e+76|0;t=e+24|0;u=e+313|0;v=e+232|0;w=e+20|0;x=e+16|0;y=e+32|0;z=e+8|0;A=e;B=j;C=121436;D=B+36|0;do{k[B>>2]=k[C>>2];B=B+4|0;C=C+4|0}while((B|0)<(D|0));B=o;C=121436;D=B+36|0;do{k[B>>2]=k[C>>2];B=B+4|0;C=C+4|0}while((B|0)<(D|0));B=q;C=121436;D=B+36|0;do{k[B>>2]=k[C>>2];B=B+4|0;C=C+4|0}while((B|0)<(D|0));B=s;C=121436;D=B+36|0;do{k[B>>2]=k[C>>2];B=B+4|0;C=C+4|0}while((B|0)<(D|0));C=k[d>>2]|0;if((C|0)>0){E=C;r=e;return E|0}El(a,f,d)|0;Fl(a,g,d)|0;if((Gl(a,k[g>>2]|0,j,d)|0)>0){E=k[d>>2]|0;r=e;return E|0}C=k[g>>2]|0;if((C+-1|0)>>>0>3){tk(733152);k[d>>2]=212;E=212;r=e;return E|0}Xr(b,k[f>>2]|0,C,j,d)|0;Co(a,h,0,d)|0;if((k[h>>2]|0)>=4){C=4;while(1){Ro(a,C,v,d)|0;if((Ll(v)|0)>20)ds(b,v,d)|0;if((C|0)<(k[h>>2]|0))C=C+1|0;else break}}if((k[d>>2]|0)>0){tk(733220);E=k[d>>2]|0;r=e;return E|0}k[t>>2]=c;a:do if((k[g>>2]|0)>0){C=u+2|0;h=0;while(1){if((Yh(t,l,m,n,d)|0)>0){F=15;break}v=k[m>>2]|0;if(v){B=k[l>>2]|0;D=k[j+(h<<2)>>2]|0;if(!B){k[l>>2]=D;G=D;H=D;I=v}else{G=B;H=D;I=v}}else{v=k[j+(h<<2)>>2]|0;k[m>>2]=v;G=k[l>>2]|0;H=v;I=v}if((G|0)>(H|0)|(I|0)>(H|0)){F=21;break}v=o+(h<<2)|0;k[v>>2]=G;D=q+(h<<2)|0;k[D>>2]=I;B=k[n>>2]|0;J=s+(h<<2)|0;k[J>>2]=B;K=(((G|0)>(I|0)?G-I|0:I-G|0)+B|0)/(B|0)|0;k[y+(h<<2)>>2]=K;h=h+1|0;Gk(782713,h,u,d)|0;oq(b,u,K,((K|0)<0)<<31>>31,0,d)|0;if(!((k[v>>2]|0)==1?(k[J>>2]|0)==1:0)){L=-1;F=25}if((F|0)==25)while(1){F=0;Gk(768562,h,u,d)|0;K=(L|0)!=-1;if(K){B=xx(u)|0;i[u+B>>0]=L+65;i[u+(B+1)>>0]=0}k[w>>2]=0;do if(!(Io(a,82,u,z,0,w)|0)){B=k[v>>2]|0;if((B|0)>(k[D>>2]|0))M=(+(B|0)-+p[z>>3])/+(k[J>>2]|0);else M=(+p[z>>3]-+(B|0))/+(k[J>>2]|0);N=M+1.0;p[z>>3]=N;rq(b,u,N,15,0,d)|0;if((k[J>>2]|0)==1?(k[v>>2]|0)<=(k[D>>2]|0):0)break;Gk(768606,h,u,d)|0;if(K){B=xx(u)|0;i[u+B>>0]=L+65;i[u+(B+1)>>0]=0}k[w>>2]=0;if(!(Io(a,82,u,A,0,w)|0)){B=k[J>>2]|0;N=+p[A>>3]*+(((k[v>>2]|0)>(k[D>>2]|0)?0-B|0:B)|0);p[A>>3]=N;rq(b,u,N,15,0,d)|0}Gk(733376,h,u,d)|0;if(K){B=xx(u)|0;i[u+B>>0]=L+65;i[u+(B+1)>>0]=0;O=0}else O=0;do{i[C>>0]=O+49;k[w>>2]=0;if(!(Io(a,82,u,A,0,w)|0)){B=k[J>>2]|0;N=+p[A>>3]*+(((k[v>>2]|0)>(k[D>>2]|0)?0-B|0:B)|0);p[A>>3]=N;rq(b,u,N,15,0,d)|0}O=O+1|0}while((O|0)!=9)}while(0);L=L+1|0;if((L|0)==26)break;else F=25}if((h|0)>=(k[g>>2]|0))break a}if((F|0)==15){tk(733274);tk(c);E=k[d>>2]|0;r=e;return E|0}else if((F|0)==21){tk(733327);tk(c);k[d>>2]=212;E=212;r=e;return E|0}}while(0);if((Vk(b,d)|0)>0){E=k[d>>2]|0;r=e;return E|0}Gs(a,1.0,0.0,d)|0;Gs(b,1.0,0.0,d)|0;c=k[y>>2]|0;y=k[f>>2]|0;F=cy(ha((((y|0)>-1?y:0-y|0)|0)/8|0,c)|0)|0;if(!F){tk(733381);k[d>>2]=113;E=113;r=e;return E|0}y=o+4|0;L=k[y>>2]|0;O=q+4|0;u=k[O>>2]|0;A=(L|0)>(u|0);w=s+4|0;z=k[w>>2]|0;G=((A?L-u|0:u-L|0)+z|0)/(z|0)|0;z=o+8|0;u=k[z>>2]|0;I=q+8|0;n=k[I>>2]|0;H=(u|0)>(n|0);l=s+8|0;m=k[l>>2]|0;t=((H?u-n|0:n-u|0)+m|0)/(m|0)|0;m=o+12|0;n=k[m>>2]|0;h=q+12|0;C=k[h>>2]|0;D=(n|0)>(C|0);v=s+12|0;J=k[v>>2]|0;K=((D?n-C|0:C-n|0)+J|0)/(J|0)|0;b:do if((K|0)>0){C=(t|0)>0;B=(G|0)>0;P=((c|0)<0)<<31>>31;Q=ha(G,c)|0;R=J;S=1;T=0;while(1){U=ha(R,T)|0;V=(D?0-U|0:U)+n|0;k[m>>2]=V;k[h>>2]=V;if(C){V=S;U=0;while(1){W=ha(k[l>>2]|0,U)|0;X=(H?0-W|0:W)+u|0;k[z>>2]=X;k[I>>2]=X;if(B){c:do if(A){X=V;W=0;while(1){Y=L-(ha(k[w>>2]|0,W)|0)|0;k[y>>2]=Y;k[O>>2]=Y;switch(k[f>>2]|0){case 8:{dm(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;Qq(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case 16:{Pm(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;dr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case 32:{yn(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;sr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case -32:{Dm(a,1,k[g>>2]|0,j,o,q,s,-9.119119840596153e-36,F,x,d)|0;ar(b,1,X,((X|0)<0)<<31>>31,c,P,F,-9.119119840596153e-36,d)|0;break}case -64:{pm(a,1,k[g>>2]|0,j,o,q,s,-9.1191291391491e-36,F,x,d)|0;Xq(b,1,X,((X|0)<0)<<31>>31,c,P,F,-9.1191291391491e-36,d)|0;break}case 64:{mn(a,1,k[g>>2]|0,j,o,q,s,0,0,F,x,d)|0;nr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}default:{}}W=W+1|0;if((W|0)==(G|0))break c;else X=X+c|0}}else{X=V;W=0;while(1){Y=(ha(k[w>>2]|0,W)|0)+L|0;k[y>>2]=Y;k[O>>2]=Y;switch(k[f>>2]|0){case 8:{dm(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;Qq(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case 16:{Pm(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;dr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case 32:{yn(a,1,k[g>>2]|0,j,o,q,s,0,F,x,d)|0;sr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}case -32:{Dm(a,1,k[g>>2]|0,j,o,q,s,-9.119119840596153e-36,F,x,d)|0;ar(b,1,X,((X|0)<0)<<31>>31,c,P,F,-9.119119840596153e-36,d)|0;break}case -64:{pm(a,1,k[g>>2]|0,j,o,q,s,-9.1191291391491e-36,F,x,d)|0;Xq(b,1,X,((X|0)<0)<<31>>31,c,P,F,-9.1191291391491e-36,d)|0;break}case 64:{mn(a,1,k[g>>2]|0,j,o,q,s,0,0,F,x,d)|0;nr(b,1,X,((X|0)<0)<<31>>31,c,P,F,d)|0;break}default:{}}W=W+1|0;if((W|0)==(G|0))break c;else X=X+c|0}}while(0);Z=Q+V|0}else Z=V;U=U+1|0;if((U|0)==(t|0)){_=Z;break}else V=Z}}else _=S;V=T+1|0;if((V|0)==(K|0))break b;R=k[v>>2]|0;S=_;T=V}}while(0);dy(F);F=k[d>>2]|0;if((F|0)<=0){E=F;r=e;return E|0}tk(733434);E=k[d>>2]|0;r=e;return E|0}function Yh(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;f=r;r=r+80|0;g=f;h=f+4|0;l=k[e>>2]|0;if((l|0)>0){m=l;r=f;return m|0}a:do if(!(Vh(a,728118,h,g)|0)){j[h>>1]=42;n=5}else{l=j[h>>1]|0;switch((l&255)<<24>>24){case 42:{n=5;break a;break}case 45:{if((l&-256)<<16>>16==10752){k[b>>2]=0;k[c>>2]=1;break a}break}default:{}}if((k[g>>2]|0)!=0?(l=k[a>>2]|0,(i[l>>0]|0)==58):0){o=fx(h)|0;k[b>>2]=o;k[a>>2]=l+1;l=(Vh(a,728118,h,g)|0)!=0;if(l&(k[g>>2]|0)!=0){l=fx(h)|0;k[c>>2]=l;break}k[e>>2]=125;m=125;r=f;return m|0}k[e>>2]=125;m=125;r=f;return m|0}while(0);if((n|0)==5){k[b>>2]=1;k[c>>2]=0}n=k[a>>2]|0;do if((i[n>>0]|0)==58){k[a>>2]=n+1;l=(Vh(a,733487,h,g)|0)!=0;if(l&(k[g>>2]|0)!=0){l=fx(h)|0;p=l;q=k[a>>2]|0;break}k[e>>2]=125;m=125;r=f;return m|0}else{p=1;q=n}while(0);k[d>>2]=p;d=i[q>>0]|0;if(d<<24>>24==44){n=q+1|0;k[a>>2]=n;s=i[n>>0]|0;t=n}else{s=d;t=q}if(s<<24>>24==32){s=t;do{s=s+1|0;k[a>>2]=s}while((i[s>>0]|0)==32)}if((k[b>>2]|0)>=0?!((k[c>>2]|0)<0|(p|0)<1):0){m=k[e>>2]|0;r=f;return m|0}k[e>>2]=125;m=125;r=f;return m|0}function Zh(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}g=(b|0)!=0;if(g)i[b>>0]=0;j=(c|0)!=0;if(j)i[c>>0]=0;l=(d|0)!=0;if(l)i[d>>0]=0;m=(e|0)!=0;if(m){i[e>>0]=0;n=a}else n=a;a:while(1){switch(i[n>>0]|0){case 45:{o=n;p=12;break a;break}case 32:break;default:{q=n;p=13;break a}}n=n+1|0}if((p|0)==12)switch(i[o+1>>0]|0){case 32:case 0:{p=15;break}default:{q=o;p=13}}b:do if((p|0)==13)if((tx(q,733490)|0)!=0?(tx(q,733497)|0)!=0:0){o=Ex(q,765357)|0;if(!o)if(g){n=b+(xx(b)|0)|0;a=n;i[a>>0]=102;i[a+1>>0]=105;i[a+2>>0]=108;i[a+3>>0]=101;a=n+4|0;i[a>>0]=58;i[a+1>>0]=47;i[a+2>>0]=47;i[a+3>>0]=0;r=q}else r=q;else{do if(g){a=o-q+3|0;if((a|0)<=19){zx(b,q,a)|0;break}k[f>>2]=125;h=125;return h|0}while(0);r=o+3|0}a=rx(r,40)|0;n=rx(r,91)|0;s=(a|0)==0;do if(j){if(!s){t=a-r|0;if((t|0)<=1024){zx(c,r,t)|0;p=38;break}k[f>>2]=125;h=125;return h|0}if(!n){if((xx(r)|0)>>>0<=1024){ux(c,r)|0;break}k[f>>2]=125;h=125;return h|0}else{t=n-r|0;if((t|0)<=1024){zx(c,r,t)|0;p=45;break}k[f>>2]=125;h=125;return h|0}}else if(s)p=44;else p=38;while(0);do if((p|0)==38){s=a+1|0;o=rx(s,41)|0;if(!o){k[f>>2]=125;h=125;return h|0}if(l){t=o-s|0;if((t|0)<=1024){zx(d,s,t)|0;p=44;break}k[f>>2]=125;h=125;return h|0}else p=44}while(0);if((p|0)==44?(n|0)!=0:0)p=45;do if((p|0)==45){a=n+1|0;t=rx(a,93)|0;if(!t){k[f>>2]=125;h=125;return h|0}if(m){s=t-a|0;if((s|0)<=1024){zx(e,a,s)|0;break}k[f>>2]=125;h=125;return h|0}}while(0);if((g&j?(tx(b,764609)|0)==0:0)?(n=Ex(c,739114)|0,(n|0)!=0):0){switch(i[n+3>>0]|0){case 32:case 0:break;default:break b}u=b;v=727448;w=u+19|0;do{i[u>>0]=i[v>>0]|0;u=u+1|0;v=v+1|0}while((u|0)<(w|0))}}else p=15;while(0);if((p|0)==15?g:0){u=b;v=765612;w=u+10|0;do{i[u>>0]=i[v>>0]|0;u=u+1|0;v=v+1|0}while((u|0)<(w|0))}h=k[f>>2]|0;return h|0}function _h(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;d=r;r=r+112|0;e=d+12|0;f=d+8|0;g=d+4|0;h=d;j=d+16|0;k[f>>2]=0;l=k[c>>2]|0;if((l|0)>0|(b|0)==0){m=l;r=d;return m|0}if(!(i[b>>0]|0)){m=l;r=d;return m|0}Hh(e,b,0,f)|0;if(k[f>>2]|0){uk(2,j);Pp(a,b,c)|0;Fk(a,1,0,c)|0;m=k[c>>2]|0;r=d;return m|0}Fk(k[e>>2]|0,1,0,c)|0;b=k[c>>2]|0;if((b|0)<1){f=j+25|0;do{Co(k[e>>2]|0,g,h,c)|0;l=k[e>>2]|0;if((k[g>>2]|0)<1)n=l;else{o=l;l=1;while(1){Ro(o,l,j,c)|0;if((Ax(j,744978,8)|0)==0?(Ax(f,733504,5)|0)!=0:0){p=j;q=733510;s=p+30|0;do{i[p>>0]=i[q>>0]|0;p=p+1|0;q=q+1|0}while((p|0)<(s|0))}ds(a,j,c)|0;q=k[e>>2]|0;if((l|0)<(k[g>>2]|0)){o=q;l=l+1|0}else{n=q;break}}}Bl(n,1,0,c)|0;Al(a,c)|0;l=k[c>>2]|0}while((l|0)<1);t=l}else t=b;if((t|0)==107)k[c>>2]=0;Fh(k[e>>2]|0,c)|0;Fk(a,1,0,c)|0;m=k[c>>2]|0;r=d;return m|0}function $h(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0;g=r;r=r+16|0;h=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}k[a>>2]=0;if((k[28269]|0)!=0?(j=Ch()|0,k[f>>2]=j,(j|0)>0):0){l=j;r=g;return l|0}j=k[28270]|0;while(1){m=j+-1|0;if((j|0)<=0){n=7;break}if(!(tx(113084+(m*84|0)|0,765414)|0)){o=m;break}else j=m}if((n|0)==7){k[f>>2]=124;tk(733541);l=k[f>>2]|0;r=g;return l|0}k[f>>2]=0;j=Ti(b,c,d,e,h)|0;k[f>>2]=j;if((j|0)>0){tk(733602);l=k[f>>2]|0;r=g;return l|0}j=ey(1,8)|0;k[a>>2]=j;if(!j){Ub[k[113084+(o*84|0)+56>>2]&31](k[h>>2]|0)|0;tk(733652);k[f>>2]=113;l=113;r=g;return l|0}e=ey(1,1728)|0;k[j+4>>2]=e;e=(k[a>>2]|0)+4|0;if(!(k[e>>2]|0)){Ub[k[113084+(o*84|0)+56>>2]&31](k[h>>2]|0)|0;tk(733652);dy(k[a>>2]|0);k[a>>2]=0;k[f>>2]=113;l=113;r=g;return l|0}j=cy(32)|0;k[(k[e>>2]|0)+12>>2]=j;j=(k[a>>2]|0)+4|0;if(!(k[(k[j>>2]|0)+12>>2]|0)){Ub[k[113084+(o*84|0)+56>>2]&31](k[h>>2]|0)|0;tk(733707);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[f>>2]=113;l=113;r=g;return l|0}e=ey(1001,8)|0;k[(k[j>>2]|0)+88>>2]=e;e=(k[a>>2]|0)+4|0;if(!(k[(k[e>>2]|0)+88>>2]|0)){Ub[k[113084+(o*84|0)+56>>2]&31](k[h>>2]|0)|0;tk(733756);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[f>>2]=113;l=113;r=g;return l|0}j=ey(40,2880)|0;k[(k[e>>2]|0)+1244>>2]=j;j=k[(k[a>>2]|0)+4>>2]|0;if(!(k[j+1244>>2]|0)){Ub[k[113084+(o*84|0)+56>>2]&31](k[h>>2]|0)|0;tk(733812);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+88>>2]|0);dy(k[(k[(k[a>>2]|0)+4>>2]|0)+12>>2]|0);dy(k[(k[a>>2]|0)+4>>2]|0);dy(k[a>>2]|0);k[a>>2]=0;k[f>>2]=113;l=113;r=g;return l|0}else p=0;do{k[j+1568+(p<<2)>>2]=p;k[j+1248+(p<<2)>>2]=-1;p=p+1|0}while((p|0)!=40);k[j+84>>2]=1e3;k[j>>2]=k[h>>2];k[j+4>>2]=o;o=k[j+12>>2]|0;j=o;i[j>>0]=109;i[j+1>>0]=101;i[j+2>>0]=109;i[j+3>>0]=102;j=o+4|0;i[j>>0]=105;i[j+1>>0]=108;i[j+2>>0]=101;i[j+3>>0]=0;j=k[c>>2]|0;c=k[a>>2]|0;o=k[c+4>>2]|0;h=o+24|0;k[h>>2]=j;k[h+4>>2]=0;h=o+32|0;k[h>>2]=j;k[h+4>>2]=0;k[o+76>>2]=1;h=o+120|0;k[h>>2]=-1;k[h+4>>2]=-1;k[o+64>>2]=-1;k[o+8>>2]=1;k[o+16>>2]=555;bu(c,0,1,f)|0;c=k[(k[a>>2]|0)+4>>2]|0;a=k[f>>2]|0;if((a|0)>0){l=a;r=g;return l|0}else q=0;while(1){f=115100+(q<<2)|0;q=q+1|0;if(!(k[f>>2]|0)){s=f;break}if((q|0)>=1e3){l=a;n=25;break}}if((n|0)==25){r=g;return l|0}k[s>>2]=c;l=a;r=g;return l|0}function ai(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+1056|0;e=d;f=d+20|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}i[b>>0]=0;i[e>>0]=0;i[f>>0]=0;do if((i[a>>0]|0)==45){g=e+(xx(e)|0)|0;i[g>>0]=45;i[g+1>>0]=0;j=a+1|0}else{if((Ax(a,730780,5)|0)!=0?(Ax(a,733867,5)|0)!=0:0){g=Ex(a,765357)|0;l=rx(a,40)|0;if(!((g|0)==0|(l|0)!=0&l>>>0<g>>>0)){l=g-a+3|0;if((l|0)<=19){zx(e,a,l)|0;j=g+3|0;break}k[c>>2]=125;h=125;r=d;return h|0}if(!(Ax(a,743143,4)|0)){g=e+(xx(e)|0)|0;i[g>>0]=i[765481]|0;i[g+1>>0]=i[765482]|0;i[g+2>>0]=i[765483]|0;i[g+3>>0]=i[765484]|0;i[g+4>>0]=i[765485]|0;i[g+5>>0]=i[765486]|0;i[g+6>>0]=i[765487]|0;j=a+4|0;break}if(!(Ax(a,730786,7)|0)){m=e+(xx(e)|0)|0;n=730794;o=m+10|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));j=a+7|0;break}if(!(Ax(a,743121,5)|0)){g=e+(xx(e)|0)|0;l=g;i[l>>0]=104;i[l+1>>0]=116;i[l+2>>0]=116;i[l+3>>0]=112;l=g+4|0;i[l>>0]=58;i[l+1>>0]=47;i[l+2>>0]=47;i[l+3>>0]=0;j=a+5|0;break}if(!(Ax(a,741186,4)|0)){l=e+(xx(e)|0)|0;i[l>>0]=i[777975]|0;i[l+1>>0]=i[777976]|0;i[l+2>>0]=i[777977]|0;i[l+3>>0]=i[777978]|0;i[l+4>>0]=i[777979]|0;i[l+5>>0]=i[777980]|0;i[l+6>>0]=i[777981]|0;j=a+4|0;break}if(!(Ax(a,730804,6)|0)){m=e+(xx(e)|0)|0;n=765425;o=m+9|0;do{i[m>>0]=i[n>>0]|0;m=m+1|0;n=n+1|0}while((m|0)<(o|0));j=a+6|0;break}else{l=(Ax(a,730811,5)|0)==0;j=l?a+5|0:a;break}}l=e+(xx(e)|0)|0;i[l>>0]=45;i[l+1>>0]=0;j=a+5|0}while(0);a=rx(j,40)|0;n=rx(j,91)|0;do if((a|0)!=(n|0)){if(!n){m=a-j|0;if((m|0)>1024){k[c>>2]=125;h=125;r=d;return h|0}zx(f,j,m)|0;if(rx(a+1|0,41)|0)break;k[c>>2]=125;h=125;r=d;return h|0}if(!((a|0)!=0&a>>>0<n>>>0)){m=n-j|0;if((m|0)<=1024){zx(f,j,m)|0;break}k[c>>2]=125;h=125;r=d;return h|0}m=a-j|0;if((m|0)>1024){k[c>>2]=125;h=125;r=d;return h|0}zx(f,j,m)|0;if(!(rx(a+1|0,41)|0)){k[c>>2]=125;h=125;r=d;return h|0}}else{if((xx(j)|0)>>>0<=1024){qx(f,j)|0;break}k[c>>2]=125;h=125;r=d;return h|0}while(0);j=xx(f)|0;a:do if((j|0)>1){a=j;do{a=a+-1|0;n=f+a|0;if((i[n>>0]|0)!=32)break a;i[n>>0]=0}while((a|0)>1)}while(0);j=xx(f)|0;a=j;while(1){n=a+-1|0;if((a|0)<=0)break;m=f+n|0;if((i[m>>0]|0)==43){p=m;q=a;s=n;t=45;break}else a=n}if((t|0)==45?(q|0)>1&(j-s|0)<5:0){b:do if((q|0)<(j|0)){s=q;while(1){if(((i[f+s>>0]|0)+-48|0)>>>0>=10){u=s;break b}t=s+1|0;if((t|0)<(j|0))s=t;else{u=t;break}}}else u=q;while(0);if((u|0)==(j|0))i[p>>0]=0}p=xx(e)|0;if(((xx(f)|0)+p|0)>>>0>1024){k[c>>2]=125;h=125;r=d;return h|0}else{qx(b,e)|0;qx(b,f)|0;h=k[c>>2]|0;r=d;return h|0}return 0}function bi(a,b,c){a=a|0;b=b|0;c=c|0;ux(b,113084+((k[(k[a+4>>2]|0)+4>>2]|0)*84|0)|0)|0;return k[c>>2]|0}function ci(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;c=r;r=r+16|0;d=c+4|0;e=c;k[d>>2]=999;k[e>>2]=0;if(!a){k[b>>2]=115;f=115;r=c;return f|0}g=a+4|0;if((k[(k[g>>2]|0)+16>>2]|0)!=555){k[b>>2]=114;f=114;r=c;return f|0}if((k[b>>2]|0)>0)wl(a,d)|0;else wl(a,b)|0;hu(a,1,b)|0;d=k[g>>2]|0;if((Ub[k[113084+((k[d+4>>2]|0)*84|0)+56>>2]&31](k[d>>2]|0)|0)!=0?(k[b>>2]|0)<1:0){k[b>>2]=110;tk(733873);tk(k[(k[g>>2]|0)+12>>2]|0)}d=k[g>>2]|0;if(!(k[113084+((k[d+4>>2]|0)*84|0)+60>>2]|0))h=d;else{i=d+12|0;d=cy((xx(k[i>>2]|0)|0)+1|0)|0;if(!d){k[b>>2]=113;f=113;r=c;return f|0}Jh(k[i>>2]|0,0,d,0,0,0,0,0,0,0,e)|0;if((Ub[k[113084+((k[(k[g>>2]|0)+4>>2]|0)*84|0)+60>>2]&31](d)|0)!=0?(tk(733918),tk(k[(k[g>>2]|0)+12>>2]|0),(k[b>>2]|0)==0):0)k[b>>2]=110;dy(d);h=k[g>>2]|0}d=0;while(1){e=115100+(d<<2)|0;d=d+1|0;if((k[e>>2]|0)==(h|0)){j=e;l=21;break}if((d|0)>=1e3){m=h;break}}if((l|0)==21){k[j>>2]=0;m=k[g>>2]|0}dy(k[m+1244>>2]|0);dy(k[(k[g>>2]|0)+88>>2]|0);dy(k[(k[g>>2]|0)+12>>2]|0);dy(k[g>>2]|0);dy(a);f=k[b>>2]|0;r=c;return f|0}function di(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=a+4|0;if(!(k[113084+((k[(k[e>>2]|0)+4>>2]|0)*84|0)+52>>2]|0)){f=k[d>>2]|0;return f|0}else{hu(a,0,d)|0;g=k[e>>2]|0;h=g+24|0;k[h>>2]=b;k[h+4>>2]=c;h=g+56|0;k[h>>2]=b;k[h+4>>2]=c;h=g+32|0;k[h>>2]=b;k[h+4>>2]=c;h=g+48|0;k[h>>2]=b;k[h+4>>2]=c;iu(a,d)|0;a=k[e>>2]|0;e=Pb[k[113084+((k[a+4>>2]|0)*84|0)+52>>2]&63](k[a>>2]|0,b,c)|0;k[d>>2]=e;f=e;return f|0}return 0}function ei(a){a=a|0;var b=0,c=0;b=k[113084+((k[a+4>>2]|0)*84|0)+68>>2]|0;if(!b){c=0;return c|0}c=Ub[b&31](k[a>>2]|0)|0;return c|0}function fi(a,b,c){a=a|0;b=b|0;c=c|0;return Pb[k[113084+((k[a+4>>2]|0)*84|0)+72>>2]&63](k[a>>2]|0,b,c)|0}function gi(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0;if(!(Pb[k[113084+((k[a+4>>2]|0)*84|0)+80>>2]&63](k[a>>2]|0,c,b)|0)){e=k[d>>2]|0;return e|0}else{tk(733964);tk(k[a+12>>2]|0);k[d>>2]=106;e=106;return e|0}return 0}function hi(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;c=r;r=r+128|0;d=c+8|0;e=c;f=c+12|0;g=c+43|0;if(!b){r=c;return}sk(b,f);k[e>>2]=b;k[e+4>>2]=f;Ew(a,733999,e)|0;if(!(wk(g)|0)){r=c;return}do{k[d>>2]=g;Ew(a,787303,d)|0}while((wk(g)|0)!=0);r=c;return}function ii(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;b=r;r=r+16|0;c=b;d=k[a>>2]|0;k[c>>2]=d;e=d;a:while(1){d=i[e>>0]|0;switch(d<<24>>24){case 0:{f=1;break a;break}case 41:{g=e;h=3;break a;break}case 40:{k[c>>2]=e+1;if(ii(c)|0){f=1;break a}break}case 91:{k[c>>2]=e+1;if(ji(c)|0){f=1;break a}break}case 123:{k[c>>2]=e+1;if(ki(c)|0){f=1;break a}break}case 34:{j=e+1|0;k[c>>2]=j;l=j;while(1){j=i[l>>0]|0;if(!(j<<24>>24)){f=1;break a}m=l+1|0;if(j<<24>>24==34){n=m;break}else l=m}k[c>>2]=n;break}default:{l=e+1|0;k[c>>2]=l;if(d<<24>>24==39){m=l;while(1){l=i[m>>0]|0;if(!(l<<24>>24)){f=1;break a}j=m+1|0;if(l<<24>>24==39){o=j;break}else m=j}k[c>>2]=o}}}e=k[c>>2]|0}if((h|0)==3){k[a>>2]=g+1;f=0}r=b;return f|0}function ji(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;b=r;r=r+16|0;c=b;d=k[a>>2]|0;k[c>>2]=d;e=d;a:while(1){d=i[e>>0]|0;switch(d<<24>>24){case 0:{f=1;break a;break}case 93:{g=e;h=3;break a;break}case 40:{k[c>>2]=e+1;if(ii(c)|0){f=1;break a}break}case 91:{k[c>>2]=e+1;if(ji(c)|0){f=1;break a}break}case 123:{k[c>>2]=e+1;if(ki(c)|0){f=1;break a}break}case 34:{j=e+1|0;k[c>>2]=j;l=j;while(1){j=i[l>>0]|0;if(!(j<<24>>24)){f=1;break a}m=l+1|0;if(j<<24>>24==34){n=m;break}else l=m}k[c>>2]=n;break}default:{l=e+1|0;k[c>>2]=l;if(d<<24>>24==39){m=l;while(1){l=i[m>>0]|0;if(!(l<<24>>24)){f=1;break a}j=m+1|0;if(l<<24>>24==39){o=j;break}else m=j}k[c>>2]=o}}}e=k[c>>2]|0}if((h|0)==3){k[a>>2]=g+1;f=0}r=b;return f|0}function ki(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;b=r;r=r+16|0;c=b;d=k[a>>2]|0;k[c>>2]=d;e=d;a:while(1){d=i[e>>0]|0;switch(d<<24>>24){case 0:{f=1;break a;break}case 125:{g=e;h=3;break a;break}case 40:{k[c>>2]=e+1;if(ii(c)|0){f=1;break a}break}case 91:{k[c>>2]=e+1;if(ji(c)|0){f=1;break a}break}case 123:{k[c>>2]=e+1;if(ki(c)|0){f=1;break a}break}case 34:{j=e+1|0;k[c>>2]=j;l=j;while(1){j=i[l>>0]|0;if(!(j<<24>>24)){f=1;break a}m=l+1|0;if(j<<24>>24==34){n=m;break}else l=m}k[c>>2]=n;break}default:{l=e+1|0;k[c>>2]=l;if(d<<24>>24==39){m=l;while(1){l=i[m>>0]|0;if(!(l<<24>>24)){f=1;break a}j=m+1|0;if(l<<24>>24==39){o=j;break}else m=j}k[c>>2]=o}}}e=k[c>>2]|0}if((h|0)==3){k[a>>2]=g+1;f=0}r=b;return f|0}function li(){var a=0;a=0;do{k[10840+(a*24|0)>>2]=0;a=a+1|0}while((a|0)!=1e3);return 0}function mi(a){a=a|0;return 0}function ni(a){a=a|0;k[a>>2]=0;return 0}function oi(a){a=a|0;k[a>>2]=10;return 0}function pi(){return 0}function qi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;d=r;r=r+2896|0;e=d;f=d+8|0;do if(!(i[734552]|0)){k[c>>2]=-1;g=0;while(1){if(!(k[10840+(g*24|0)>>2]|0)){h=g;break}g=g+1|0;if((g|0)>=1e3){j=103;l=23;break}}if((l|0)==23){r=d;return j|0}k[c>>2]=h;if((h|0)==-1){j=103;r=d;return j|0}else{m=ri(a,b,e)|0;break}}else{g=ri(a,0,e)|0;if(g){i[734552]=0;j=g;r=d;return j|0}g=si(734552,c)|0;if(g){tk(735577);tk(734552);i[734552]=0;j=g;r=d;return j|0}g=k[e>>2]|0;n=Hw(f,1,2880,g)|0;a:do if(n){o=n;while(1){p=k[c>>2]|0;q=10840+(p*24|0)+16|0;if((k[q>>2]|0)==1){s=10840+(p*24|0)+8|0;t=s;u=k[t>>2]|0;v=k[t+4>>2]|0;t=10840+(p*24|0)|0;if(Jw(k[t>>2]|0,u,0)|0){w=116;break}x=s;k[x>>2]=u;k[x+4>>2]=v;y=t}else y=10840+(p*24|0)|0;if((Pw(f,1,o,k[y>>2]|0)|0)!=(o|0)){w=106;break}t=10840+(p*24|0)+8|0;p=t;v=ky(k[p>>2]|0,k[p+4>>2]|0,o|0,((o|0)<0)<<31>>31|0)|0;p=t;k[p>>2]=v;k[p+4>>2]=L;k[q>>2]=2;o=Hw(f,1,2880,g)|0;if(!o)break a}i[734552]=0;j=w;r=d;return j|0}while(0);ww(g)|0;n=k[c>>2]|0;o=10840+(n*24|0)|0;if(!(ww(k[o>>2]|0)|0))k[o>>2]=0;k[c>>2]=n;n=ri(734552,b,e)|0;i[734552]=0;m=n}while(0);b=k[c>>2]|0;k[10840+(b*24|0)>>2]=k[e>>2];e=10840+(b*24|0)+8|0;k[e>>2]=0;k[e+4>>2]=0;k[10840+(b*24|0)+16>>2]=0;j=m;r=d;return j|0}function ri(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0;d=r;r=r+1120|0;e=d;f=d+88|0;g=d+8|0;if((b|0)==1)k[e>>2]=6433650;else{i[e>>0]=i[769387]|0;i[e+1>>0]=i[769388]|0;i[e+2>>0]=i[769389]|0}if((i[a>>0]|0)==126){b=a+1|0;h=i[b>>0]|0;do if(h<<24>>24==47){j=jb(735630)|0;if(!j)if((xx(a)|0)>>>0>1023){l=104;r=d;return l|0}else{ux(f,a)|0;break}else{m=xx(j)|0;if(((xx(b)|0)+m|0)>>>0>1023){l=104;r=d;return l|0}else{ux(f,j)|0;qx(f,b)|0;break}}}else{j=h;m=b;n=0;a:while(1){switch(j<<24>>24){case 47:case 0:{o=m;p=n;break a;break}default:{}}i[g+n>>0]=j;q=m+1|0;j=i[q>>0]|0;m=q;n=n+1|0}i[g+p>>0]=0;n=(tb(g|0)|0)+20|0;m=k[n>>2]|0;n=xx(m)|0;if(((xx(o)|0)+n|0)>>>0>1023){l=104;r=d;return l|0}else{ux(f,m)|0;qx(f,o)|0;break}}while(0);s=Dw(f,e)|0}else s=Dw(a,e)|0;k[c>>2]=s;l=(s|0)==0?104:0;r=d;return l|0}function si(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;c=r;r=r+5664|0;d=c+4|0;e=c;f=c+4625|0;g=c+3600|0;h=c+3344|0;j=c+3088|0;l=c+2058|0;m=c+1033|0;n=c+8|0;k[e>>2]=0;o=jb(735635)|0;if(o){if((xx(o)|0)>>>0>200){p=105;r=c;return p|0}ux(h,o)|0;o=rx(h,59)|0;if(!o)i[j>>0]=0;else{i[o>>0]=0;ux(j,o+1|0)|0}yp(f,e)|0;o=xx(f)|0;if((i[f+(o+-1)>>0]|0)!=47){q=f+o|0;i[q>>0]=47;i[q+1>>0]=0}q=xx(h)|0;if(Ax(h,f,q)|0){tk(735655);p=105;r=c;return p|0}Bx(l,f+q|0,50)|0;q=rx(l,47)|0;if(!q){tk(735703);p=105;r=c;return p|0}i[q+1>>0]=0;ux(m,h)|0;qx(m,l)|0;h=xx(m)|0;ux(n,j)|0;qx(n,l)|0;l=xx(n)|0;xp(f,a,g,e)|0;if((Ax(m,g,h)|0)!=0?(Ax(n,g,l)|0)!=0:0){tk(735760);p=105;r=c;return p|0}}k[b>>2]=-1;l=0;while(1){if(!(k[10840+(l*24|0)>>2]|0)){s=l;break}l=l+1|0;if((l|0)>=1e3){p=103;t=23;break}}if((t|0)==23){r=c;return p|0}k[b>>2]=s;if((s|0)==-1){p=103;r=c;return p|0}k[d>>2]=6433655;b=Dw(a,774853)|0;if(b){ww(b)|0;p=105;r=c;return p|0}b=Dw(a,d)|0;if(!b){p=105;r=c;return p|0}k[10840+(s*24|0)>>2]=b;b=10840+(s*24|0)+8|0;k[b>>2]=0;k[b+4>>2]=0;k[10840+(s*24|0)+16>>2]=0;p=0;r=c;return p|0}function ti(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;d=10840+(a*24|0)+16|0;do if((k[d>>2]|0)==1){e=10840+(a*24|0)+8|0;f=e;g=k[f>>2]|0;h=k[f+4>>2]|0;f=10840+(a*24|0)|0;if(!(Jw(k[f>>2]|0,g,0)|0)){i=e;k[i>>2]=g;k[i+4>>2]=h;j=f;break}else{l=116;return l|0}}else j=10840+(a*24|0)|0;while(0);if((Pw(b,1,c,k[j>>2]|0)|0)!=(c|0)){l=106;return l|0}j=10840+(a*24|0)+8|0;a=j;b=ky(k[a>>2]|0,k[a+4>>2]|0,c|0,((c|0)<0)<<31>>31|0)|0;c=j;k[c>>2]=b;k[c+4>>2]=L;k[d>>2]=2;l=0;return l|0}function ui(a){a=a|0;var b=0,c=0;b=10840+(a*24|0)|0;if(ww(k[b>>2]|0)|0){c=110;return c|0}k[b>>2]=0;c=0;return c|0}function vi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=10840+(a*24|0)|0;Ix(Cw(k[d>>2]|0)|0,b)|0;e=10840+(a*24|0)+8|0;if(!(Jw(k[d>>2]|0,b,0)|0)){d=e;k[d>>2]=b;k[d+4>>2]=c}d=e;k[d>>2]=b;k[d+4>>2]=c;k[10840+(a*24|0)+16>>2]=0;return 0}function wi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;if(Jw(k[10840+(a*24|0)>>2]|0,b,0)|0){d=116;return d|0}e=10840+(a*24|0)+8|0;k[e>>2]=b;k[e+4>>2]=c;d=0;return d|0}function xi(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;c=k[10840+(a*24|0)>>2]|0;a=Mw(c)|0;if((a|0)<0){d=116;return d|0}if(Jw(c,0,2)|0){d=116;return d|0}e=Mw(c)|0;if((e|0)<0){d=116;return d|0}if(Jw(c,a,0)|0){d=116;return d|0}a=b;k[a>>2]=e;k[a+4>>2]=((e|0)<0)<<31>>31;d=0;return d|0}function yi(a){a=a|0;Uw(a)|0;return 0}function zi(a){a=a|0;var b=0;b=(zw(k[10840+(a*24|0)>>2]|0)|0)==0;return (b?0:106)|0}function Ai(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;d=10840+(a*24|0)+16|0;do if((k[d>>2]|0)==2){e=10840+(a*24|0)+8|0;f=e;g=k[f>>2]|0;h=k[f+4>>2]|0;f=10840+(a*24|0)|0;if(!(Jw(k[f>>2]|0,g,0)|0)){j=e;k[j>>2]=g;k[j+4>>2]=h;l=f;break}else{m=116;return m|0}}else l=10840+(a*24|0)|0;while(0);f=Hw(b,1,c,k[l>>2]|0)|0;if((f|0)==1){switch(i[b>>0]|0){case 32:case 10:case 0:{m=107;return m|0}default:{}}m=108;return m|0}else{if((f|0)!=(c|0)){m=108;return m|0}f=10840+(a*24|0)+8|0;a=f;b=ky(k[a>>2]|0,k[a+4>>2]|0,c|0,((c|0)<0)<<31>>31|0)|0;c=f;k[c>>2]=b;k[c+4>>2]=L;k[d>>2]=1;m=0;return m|0}return 0}function Bi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;d=r;r=r+16|0;e=d+4|0;f=d;g=ri(a,0,e)|0;k[f>>2]=g;if(g){tk(735809);tk(a);h=g;r=d;return h|0}if((i[734552]|0)!=33){g=Dw(734552,774853)|0;if(!g)j=734552;else{tk(735866);tk(734552);ww(g)|0;i[734552]=0;h=105;r=d;return h|0}}else{Uw(734553)|0;j=734553}g=Dw(j,735921)|0;if(!g){tk(735925);tk(734552);i[734552]=0;h=105;r=d;return h|0}l=k[e>>2]|0;rh(a,l,g,f)|0;ww(l)|0;ww(g)|0;if(!(k[f>>2]|0)){ux(a,j)|0;i[734552]=0;j=qi(a,b,c)|0;k[f>>2]=j;h=j;r=d;return h|0}else{tk(735982);tk(a);tk(736040);tk(734552);i[734552]=0;h=k[f>>2]|0;r=d;return h|0}return 0}function Ci(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0;b=r;r=r+1040|0;c=b;d=b+4|0;e=b+6|0;if(ri(a,0,c)|0){if((xx(a)|0)>>>0>1024){f=0;r=b;return f|0}ux(e,a)|0;g=a+(xx(a)|0)|0;i[g>>0]=46;i[g+1>>0]=103;i[g+2>>0]=122;i[g+3>>0]=0;if((((((ri(a,0,c)|0)!=0?(ux(a,e)|0,g=a+(xx(a)|0)|0,i[g>>0]=i[739111]|0,i[g+1>>0]=i[739112]|0,i[g+2>>0]=i[739113]|0,(ri(a,0,c)|0)!=0):0)?(ux(a,e)|0,g=a+(xx(a)|0)|0,i[g>>0]=i[736063]|0,i[g+1>>0]=i[736064]|0,i[g+2>>0]=i[736065]|0,(ri(a,0,c)|0)!=0):0)?(ux(a,e)|0,g=a+(xx(a)|0)|0,i[g>>0]=i[736066]|0,i[g+1>>0]=i[736067]|0,i[g+2>>0]=i[736068]|0,i[g+3>>0]=i[736069]|0,i[g+4>>0]=i[736070]|0,(ri(a,0,c)|0)!=0):0)?(ux(a,e)|0,g=a+(xx(a)|0)|0,i[g>>0]=i[736071]|0,i[g+1>>0]=i[736072]|0,i[g+2>>0]=i[736073]|0,(ri(a,0,c)|0)!=0):0)?(ux(a,e)|0,g=a+(xx(a)|0)|0,i[g>>0]=45,i[g+1>>0]=103,i[g+2>>0]=122,i[g+3>>0]=0,(ri(a,0,c)|0)!=0):0){ux(a,e)|0;f=0;r=b;return f|0}}e=k[c>>2]|0;c=(Hw(d,1,2,e)|0)==2;ww(e)|0;if(!c){f=0;r=b;return f|0}if(!(lx(d,738185,2)|0)){f=1;r=b;return f|0}if(!(lx(d,738188,2)|0)){f=1;r=b;return f|0}if(!(lx(d,738191,2)|0)){f=1;r=b;return f|0}if(!(lx(d,738194,2)|0)){f=1;r=b;return f|0}f=(lx(d,738197,2)|0)==0&1;r=b;return f|0}function Di(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=(Ci(b)|0)==0;b=(i[c>>0]|0)!=0;if(d){if(!b)return 0;i[734552]=0;zx(734552,c,1024)|0;return 0}if(!b){e=a;f=765510;g=e+12|0;do{i[e>>0]=i[f>>0]|0;e=e+1|0;f=f+1|0}while((e|0)<(g|0));i[734552]=0;return 0}if(!(Ax(c,741186,4)|0)){e=a;f=736074;g=e+15|0;do{i[e>>0]=i[f>>0]|0;e=e+1|0;f=f+1|0}while((e|0)<(g|0));i[734552]=0;return 0}e=a;f=765442;g=e+16|0;do{i[e>>0]=i[f>>0]|0;e=e+1|0;f=f+1|0}while((e|0)<(g|0));if(!(Ax(c,764609,7)|0)){ux(734552,c+7|0)|0;return 0}else{ux(734552,c)|0;return 0}return 0}function Ei(a,b,c){a=a|0;b=b|0;c=c|0;k[c>>2]=1;return 0}function Fi(a,b){a=a|0;b=b|0;k[b>>2]=2;return 0}function Gi(a,b){a=a|0;b=b|0;a=b;k[a>>2]=2147483647;k[a+4>>2]=0;return 0}function Hi(a){a=a|0;return 0}function Ii(a){a=a|0;if((a|0)!=2)return 0;zw(k[31231]|0)|0;return 0}function Ji(a,b,c){a=a|0;b=b|0;c=c|0;return 1}function Ki(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;if((a|0)!=1){d=1;return d|0}a=(Hw(b,1,c,k[31230]|0)|0)==(c|0);d=a?0:107;return d|0}function Li(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;if((a|0)!=2){d=1;return d|0}a=(Pw(b,1,c,k[31231]|0)|0)==(c|0);d=a?0:106;return d|0}function Mi(){var a=0;a=0;do{k[34840+(a*48|0)>>2]=0;k[34840+(a*48|0)+4>>2]=0;a=a+1|0}while((a|0)!=1e3);return 0}function Ni(a){a=a|0;return 0}function Oi(a){a=a|0;k[a>>2]=0;return 0}function Pi(a){a=a|0;k[a>>2]=10;return 0}function Qi(){return 0}function Ri(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0;k[b>>2]=-1;a=0;while(1){c=34840+(a*48|0)|0;if(!(k[c>>2]|0)){d=c;e=a;f=4;break}a=a+1|0;if((a|0)>=1e3){g=103;break}}do if((f|0)==4){k[b>>2]=e;if((e|0)==-1)g=103;else{a=34840+(e*48|0)+4|0;k[d>>2]=a;c=34840+(e*48|0)+12|0;k[34840+(e*48|0)+8>>2]=c;h=cy(2880)|0;k[a>>2]=h;if(!h){tk(736089);g=104;break}k[c>>2]=2880;k[34840+(e*48|0)+16>>2]=2880;c=34840+(e*48|0)+24|0;k[c>>2]=0;k[c+4>>2]=0;k[c+8>>2]=0;k[c+12>>2]=0;k[34840+(e*48|0)+20>>2]=19;i=0;return i|0}}while(0);tk(736137);i=g;return i|0}function Si(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;c=r;r=r+16|0;d=c;do if(((tx(a,736185)|0)!=0?(tx(a,736190)|0)!=0:0)?(tx(a,736200)|0)!=0:0){k[d>>2]=6433655;e=Dw(a,774853)|0;if(!e){f=Dw(a,d)|0;if(!f)g=105;else{h=f;break}r=c;return g|0}else{ww(e)|0;g=105;r=c;return g|0}}else i=4;while(0);if((i|0)==4)h=k[31231]|0;k[b>>2]=-1;d=0;while(1){a=34840+(d*48|0)|0;if(!(k[a>>2]|0)){j=a;l=d;i=11;break}d=d+1|0;if((d|0)>=1e3){m=103;break}}do if((i|0)==11){k[b>>2]=l;if((l|0)==-1)m=103;else{d=34840+(l*48|0)+4|0;k[j>>2]=d;a=34840+(l*48|0)+12|0;k[34840+(l*48|0)+8>>2]=a;e=cy(2880)|0;k[d>>2]=e;if(!e){tk(736089);m=104;break}k[a>>2]=2880;k[34840+(l*48|0)+16>>2]=2880;a=34840+(l*48|0)+24|0;k[a>>2]=0;k[a+4>>2]=0;k[a+8>>2]=0;k[a+12>>2]=0;k[34840+(l*48|0)+20>>2]=19;k[34840+((k[b>>2]|0)*48|0)+40>>2]=h;g=0;r=c;return g|0}}while(0);tk(736210);g=m;r=c;return g|0}function Ti(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0;k[e>>2]=-1;f=0;while(1){if(!(k[34840+(f*48|0)>>2]|0)){g=f;break}f=f+1|0;if((f|0)>=1e3){h=103;i=6;break}}if((i|0)==6)return h|0;k[e>>2]=g;if((g|0)==-1){h=103;return h|0}k[34840+(g*48|0)>>2]=a;k[34840+(g*48|0)+8>>2]=b;k[34840+(g*48|0)+16>>2]=c;c=34840+(g*48|0)+32|0;k[c>>2]=k[b>>2];k[c+4>>2]=0;c=34840+(g*48|0)+24|0;k[c>>2]=0;k[c+4>>2]=0;k[34840+(g*48|0)+20>>2]=d;h=0;return h|0}function Ui(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[34840+(a*48|0)+20>>2]|0;if(d){e=34840+(a*48|0)|0;f=Xb[d&31](k[k[e>>2]>>2]|0,b)|0;if(!f){tk(736263);g=113;return g|0}d=34840+(a*48|0)+8|0;h=k[k[d>>2]>>2]|0;if(0<(c|0)|0==(c|0)&h>>>0<b>>>0)sy(f+h|0,0,b-h|0)|0;k[k[e>>2]>>2]=f;k[k[d>>2]>>2]=b}d=34840+(a*48|0)+24|0;k[d>>2]=b;k[d+4>>2]=c;d=34840+(a*48|0)+32|0;k[d>>2]=b;k[d+4>>2]=c;g=0;return g|0}function Vi(a,b,c){a=a|0;b=b|0;c=c|0;b=(i[c>>0]|0)==0;i[736306]=0;if(b)return 0;zx(736306,c,1024)|0;c=a;a=765488;b=c+13|0;do{i[c>>0]=i[a>>0]|0;c=c+1|0;a=a+1|0}while((c|0)<(b|0));return 0}function Wi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;if(i[736306]|0){d=si(736306,c)|0;if(d){tk(737331);tk(736306);e=d;return e|0}d=Xi(k[c>>2]|0)|0;ui(k[c>>2]|0)|0;if(!d){e=qi(736306,b,c)|0;return e|0}else{tk(737388);tk(736306);e=d;return e|0}}d=k[31230]|0;f=(Aw(d)|0)<<24>>24;Zw(f,d)|0;switch(f|0){case 75:case 31:{e=Yi(a,b,c)|0;return e|0}default:{}}if(b){tk(737430);e=112;return e|0}k[c>>2]=-1;b=0;while(1){a=34840+(b*48|0)|0;if(!(k[a>>2]|0)){g=a;h=b;j=14;break}b=b+1|0;if((b|0)>=1e3){l=103;break}}do if((j|0)==14){k[c>>2]=h;if((h|0)==-1)l=103;else{b=34840+(h*48|0)+4|0;k[g>>2]=b;a=34840+(h*48|0)+12|0;k[34840+(h*48|0)+8>>2]=a;f=cy(2880)|0;k[b>>2]=f;if(!f){tk(736089);l=104;break}k[a>>2]=2880;k[34840+(h*48|0)+16>>2]=2880;a=34840+(h*48|0)+24|0;k[a>>2]=0;k[a+4>>2]=0;k[a+8>>2]=0;k[a+12>>2]=0;k[34840+(h*48|0)+20>>2]=19;a=Zi(k[c>>2]|0)|0;if(!a){e=0;return e|0}tk(737514);dy(k[34840+((k[c>>2]|0)*48|0)+4>>2]|0);e=a;return e|0}}while(0);tk(737466);e=l;return e|0}function Xi(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0;b=r;r=r+1008|0;c=b;d=k[31230]|0;e=Aw(d)|0;a:do if((e|0)!=-1){f=e;g=0;h=0;while(1){if((f|0)==(i[782785+g>>0]|0)){j=g+1|0;if((j|0)==6)break;else l=j}else l=0;h=h+1|0;f=Aw(d)|0;if(!((h|0)<2e3&(f|0)!=-1))break a;else g=l}i[c>>0]=i[782785]|0;i[c+1>>0]=i[782786]|0;i[c+2>>0]=i[782787]|0;i[c+3>>0]=i[782788]|0;i[c+4>>0]=i[782789]|0;i[c+5>>0]=i[782790]|0;g=ti(a,c,(Hw(c+6|0,1,994,d)|0)+6|0)|0;if(g){m=g;r=b;return m|0}while(1){g=Hw(c,1,1e3,d)|0;if(!g){m=0;n=9;break}f=ti(a,c,g)|0;if(f){m=f;n=9;break}}if((n|0)==9){r=b;return m|0}}while(0);tk(737560);m=104;r=b;return m|0}function Yi(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0;d=r;r=r+16|0;e=d+4|0;f=d;if(b){tk(737614);g=112;r=d;return g|0}k[c>>2]=-1;b=0;while(1){h=34840+(b*48|0)|0;if(!(k[h>>2]|0)){i=h;j=b;l=6;break}b=b+1|0;if((b|0)>=1e3){m=103;break}}do if((l|0)==6){k[c>>2]=j;if((j|0)!=-1){b=34840+(j*48|0)+4|0;k[i>>2]=b;h=34840+(j*48|0)+12|0;k[34840+(j*48|0)+8>>2]=h;n=cy(28800)|0;k[b>>2]=n;if(!n){tk(736089);m=104;break}k[h>>2]=28800;k[34840+(j*48|0)+16>>2]=2880;h=34840+(j*48|0)+24|0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;k[34840+(j*48|0)+20>>2]=19;h=k[31230]|0;n=k[c>>2]|0;k[f>>2]=0;b=(Ex(a,739111)|0)==0;o=k[34840+(n*48|0)>>2]|0;p=k[34840+(n*48|0)+8>>2]|0;if(b)ph(a,h,o,p,19,e,f)|0;else uh(a,h,o,p,19,e,f)|0;p=34840+(n*48|0)+24|0;k[p>>2]=0;k[p+4>>2]=0;p=34840+(n*48|0)+32|0;k[p>>2]=k[e>>2];k[p+4>>2]=0;p=k[f>>2]|0;n=k[c>>2]|0;if(p){o=34840+(n*48|0)|0;dy(k[k[o>>2]>>2]|0);k[o>>2]=0;k[34840+(n*48|0)+4>>2]=0;tk(737751);g=p;r=d;return g|0}p=k[34840+(n*48|0)+32>>2]|0;if((k[k[34840+(n*48|0)+8>>2]>>2]|0)>>>0<=(p+256|0)>>>0){g=0;r=d;return g|0}o=fy(k[k[34840+(n*48|0)>>2]>>2]|0,p)|0;if(!o){tk(737812);g=113;r=d;return g|0}else{p=k[c>>2]|0;k[k[34840+(p*48|0)>>2]>>2]=o;k[k[34840+(p*48|0)+8>>2]>>2]=k[34840+(p*48|0)+32>>2];g=0;r=d;return g|0}}else m=103}while(0);tk(737694);g=m;r=d;return g|0}function Zi(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;b=34840+(a*48|0)|0;c=k[k[b>>2]>>2]|0;d=34840+(a*48|0)+8|0;e=k[k[d>>2]>>2]|0;f=k[34840+(a*48|0)+16>>2]|0;g=k[31230]|0;h=Aw(g)|0;a:do if((h|0)!=-1){j=h;l=0;m=0;while(1){if((j|0)==(i[782785+l>>0]|0)){n=l+1|0;if((n|0)==6)break;else o=n}else o=0;m=m+1|0;j=Aw(g)|0;if(!((m|0)<2e3&(j|0)!=-1))break a;else l=o}i[c>>0]=i[782785]|0;i[c+1>>0]=i[782786]|0;i[c+2>>0]=i[782787]|0;i[c+3>>0]=i[782788]|0;i[c+4>>0]=i[782789]|0;i[c+5>>0]=i[782790]|0;l=(Hw(c+6|0,1,e+-6|0,g)|0)+6|0;if(l>>>0<e>>>0){j=34840+(a*48|0)+32|0;k[j>>2]=l;k[j+4>>2]=0;p=0;return p|0}else{q=l;r=0;s=c;t=e}while(1){l=t+f|0;j=fy(s,l)|0;if(!j){u=9;break}m=Hw(j+q|0,1,f,g)|0;n=ky(m|0,0,q|0,r|0)|0;v=L;if(m>>>0<f>>>0){w=l;x=j;y=n;z=v;u=11;break}else{q=n;r=v;s=j;t=l}}if((u|0)==9){tk(737968);p=113;return p|0}else if((u|0)==11){l=34840+(a*48|0)+32|0;k[l>>2]=y;k[l+4>>2]=z;k[k[b>>2]>>2]=x;k[k[d>>2]>>2]=w;p=0;return p|0}}while(0);tk(737876);tk(737931);p=104;return p|0}function _i(a){a=a|0;var b=0,c=0,d=0,e=0;b=34840+(a*48|0)+4|0;c=34840+(a*48|0)+32|0;d=Pw(k[b>>2]|0,1,k[c>>2]|0,k[31231]|0)|0;if((d|0)==(k[c>>2]|0))e=0;else{tk(738015);e=106}dy(k[b>>2]|0);k[34840+(a*48|0)>>2]=0;k[b>>2]=0;return e|0}function $i(a,b,c){a=a|0;b=b|0;c=c|0;return aj(a,0,c)|0}function aj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;d=r;r=r+16|0;e=d+8|0;f=d+4|0;g=d;h=d+12|0;if(b){tk(738067);tk(a);i=112;r=d;return i|0}b=ri(a,0,g)|0;if(b){tk(738133);tk(a);i=b;r=d;return i|0}if((Hw(h,1,2,k[g>>2]|0)|0)!=2){ww(k[g>>2]|0)|0;i=108;r=d;return i|0}do if(lx(h,738185,2)|0){if(!(lx(h,738188,2)|0)){Kw(k[g>>2]|0,22,0)|0;Hw(h,1,4,k[g>>2]|0)|0;j=(l[h+1>>0]|0)<<8|(l[h>>0]|0)|(l[h+2>>0]|0)<<16|(l[h+3>>0]|0)<<24;m=15;break}if(((lx(h,738191,2)|0)!=0?(lx(h,738194,2)|0)!=0:0)?(lx(h,738197,2)|0)!=0:0){ww(k[g>>2]|0)|0;i=1;r=d;return i|0}else{n=1;m=16}}else{Kw(k[g>>2]|0,0,2)|0;Nw(k[g>>2]|0)|0;Kw(k[g>>2]|0,-4,1)|0;Hw(h,1,4,k[g>>2]|0)|0;j=(l[h+1>>0]|0)<<8|(l[h>>0]|0)|(l[h+2>>0]|0)<<16|(l[h+3>>0]|0)<<24;m=15}while(0);if((m|0)==15)if(!j){n=0;m=16}else{o=0;p=j}if((m|0)==16){Kw(k[g>>2]|0,0,2)|0;o=n;p=(Nw(k[g>>2]|0)|0)*3|0}Kw(k[g>>2]|0,0,0)|0;k[c>>2]=-1;n=0;while(1){j=34840+(n*48|0)|0;if(!(k[j>>2]|0)){q=j;s=n;m=20;break}n=n+1|0;if((n|0)>=1e3){t=103;m=25;break}}do if((m|0)==20){k[c>>2]=s;if((s|0)==-1){t=103;m=25}else{n=34840+(s*48|0)+4|0;k[q>>2]=n;j=34840+(s*48|0)+12|0;k[34840+(s*48|0)+8>>2]=j;if((p|0)!=0?(h=cy(p)|0,k[n>>2]=h,(h|0)==0):0){tk(736089);t=104;m=25;break}k[j>>2]=p;k[34840+(s*48|0)+16>>2]=2880;j=34840+(s*48|0)+24|0;k[j>>2]=0;k[j+4>>2]=0;k[j+8>>2]=0;k[j+12>>2]=0;k[34840+(s*48|0)+20>>2]=19}}while(0);a:do if((m|0)==25){b:do if(o){s=(p>>>0)/3|0;k[c>>2]=-1;q=0;while(1){j=34840+(q*48|0)|0;if(!(k[j>>2]|0)){u=j;v=q;break}q=q+1|0;if((q|0)>=1e3){w=103;break b}}k[c>>2]=v;if((v|0)==-1)w=103;else{q=34840+(v*48|0)+4|0;k[u>>2]=q;j=34840+(v*48|0)+12|0;k[34840+(v*48|0)+8>>2]=j;if(p>>>0>=3?(h=cy(s)|0,k[q>>2]=h,(h|0)==0):0){tk(736089);w=104;break}k[j>>2]=s;k[34840+(v*48|0)+16>>2]=2880;j=34840+(v*48|0)+24|0;k[j>>2]=0;k[j+4>>2]=0;k[j+8>>2]=0;k[j+12>>2]=0;k[34840+(v*48|0)+20>>2]=19;break a}}else w=t;while(0);ww(k[g>>2]|0)|0;tk(738200);i=w;r=d;return i|0}while(0);w=k[g>>2]|0;t=k[c>>2]|0;k[f>>2]=0;v=(Ex(a,739111)|0)==0;p=k[34840+(t*48|0)>>2]|0;u=k[34840+(t*48|0)+8>>2]|0;if(v)ph(a,w,p,u,19,e,f)|0;else uh(a,w,p,u,19,e,f)|0;u=34840+(t*48|0)+24|0;k[u>>2]=0;k[u+4>>2]=0;u=34840+(t*48|0)+32|0;k[u>>2]=k[e>>2];k[u+4>>2]=0;u=k[f>>2]|0;ww(k[g>>2]|0)|0;g=k[c>>2]|0;if(u){f=34840+(g*48|0)|0;dy(k[k[f>>2]>>2]|0);k[f>>2]=0;k[34840+(g*48|0)+4>>2]=0;tk(738251);i=u;r=d;return i|0}u=k[34840+(g*48|0)+32>>2]|0;if((k[k[34840+(g*48|0)+8>>2]>>2]|0)>>>0<=(u+256|0)>>>0){i=0;r=d;return i|0}f=fy(k[k[34840+(g*48|0)>>2]>>2]|0,u)|0;if(!f){tk(738305);i=113;r=d;return i|0}else{u=k[c>>2]|0;k[k[34840+(u*48|0)>>2]>>2]=f;k[k[34840+(u*48|0)+8>>2]>>2]=k[34840+(u*48|0)+32>>2];i=0;r=d;return i|0}return 0}function bj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0;d=r;r=r+16|0;e=d+4|0;f=d;k[f>>2]=0;g=(Ex(a,739111)|0)==0;h=k[34840+(c*48|0)>>2]|0;i=k[34840+(c*48|0)+8>>2]|0;if(g)ph(a,b,h,i,19,e,f)|0;else uh(a,b,h,i,19,e,f)|0;i=34840+(c*48|0)+24|0;k[i>>2]=0;k[i+4>>2]=0;i=34840+(c*48|0)+32|0;k[i>>2]=k[e>>2];k[i+4>>2]=0;r=d;return k[f>>2]|0}function cj(a){a=a|0;var b=0;b=34840+(a*48|0)|0;dy(k[k[b>>2]>>2]|0);k[b>>2]=0;k[34840+(a*48|0)+4>>2]=0;return 0}function dj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;b=r;r=r+16|0;d=b+4|0;e=b;k[e>>2]=0;k[c>>2]=-1;f=0;do{g=34840+(f*48|0)|0;if(!(k[g>>2]|0)){h=g;i=f;j=4;break}f=f+1|0}while((f|0)<1e3);if((j|0)==4?(k[c>>2]=i,(i|0)!=-1):0){k[h>>2]=34840+(i*48|0)+4;h=34840+(i*48|0)+12|0;k[34840+(i*48|0)+8>>2]=h;k[h>>2]=0;k[34840+(i*48|0)+16>>2]=2880;h=34840+(i*48|0)+24|0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;k[34840+(i*48|0)+20>>2]=19;k[d>>2]=0;i=k[c>>2]|0;h=$p(a,k[34840+(i*48|0)>>2]|0,k[34840+(i*48|0)+8>>2]|0,e,d)|0;k[d>>2]=h;i=k[c>>2]|0;if(!h){h=34840+(i*48|0)+24|0;k[h>>2]=0;k[h+4>>2]=0;h=34840+(i*48|0)+32|0;k[h>>2]=k[e>>2];k[h+4>>2]=0;l=0;r=b;return l|0}else{h=34840+(i*48|0)|0;dy(k[k[h>>2]>>2]|0);k[h>>2]=0;k[34840+(i*48|0)+4>>2]=0;tk(738414);l=k[d>>2]|0;r=b;return l|0}}k[d>>2]=103;tk(738363);l=103;r=b;return l|0}function ej(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,l=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0;d=r;r=r+1072|0;e=d+40|0;f=d+36|0;g=d+32|0;h=d+12|0;l=d+8|0;n=d+44|0;o=d+4|0;p=d;k[h>>2]=k[29918];k[h+4>>2]=k[29919];k[h+8>>2]=k[29920];k[h+12>>2]=k[29921];k[h+16>>2]=k[29922];k[l>>2]=0;k[o>>2]=0;k[p>>2]=0;if(b){tk(738470);tk(a);q=112;r=d;return q|0}b=rx(a,91)|0;k[o>>2]=b;if(!b){tk(738535);tk(a);q=125;r=d;return q|0}i[n>>0]=0;zx(n,a,b-a|0)|0;s=b;a:while(1){b=s+1|0;k[o>>2]=b;switch(i[b>>0]|0){case 32:{s=b;break}case 66:case 98:{t=s;u=9;break a;break}case 73:case 105:{v=s;u=10;break a;break}case 85:case 117:{w=s;u=11;break a;break}case 74:case 106:{x=s;u=12;break a;break}case 70:case 102:case 82:case 114:{y=s;z=4;A=-32;break a;break}case 68:case 100:{B=s;u=7;break a;break}default:{u=8;break a}}}if((u|0)==7){y=B;z=8;A=-64}else if((u|0)==8){tk(738593);tk(a);q=125;r=d;return q|0}else if((u|0)==9){y=t;z=1;A=8}else if((u|0)==10){y=v;z=2;A=16}else if((u|0)==11){y=w;z=2;A=20}else if((u|0)==12){y=x;z=4;A=32}x=y+2|0;k[o>>2]=x;switch(i[x>>0]|0){case 66:case 98:{w=y+3|0;k[o>>2]=w;C=w;D=0;break}case 76:case 108:{w=y+3|0;k[o>>2]=w;C=w;D=1;break}default:{C=x;D=1}}x=jx(C,p,10)|0;k[h>>2]=x;x=k[p>>2]|0;if(x)if((i[x>>0]|0)==44){w=jx(x+1|0,o,10)|0;k[h+4>>2]=w;w=k[o>>2]|0;if(w)if((i[w>>0]|0)==44){y=jx(w+1|0,p,10)|0;k[h+8>>2]=y;y=k[p>>2]|0;if(y)if((i[y>>0]|0)==44){v=jx(y+1|0,o,10)|0;k[h+12>>2]=v;v=k[o>>2]|0;if(!v)E=4;else E=(i[v>>0]|0)==44?5:4;t=jx(v+1|0,p,10)|0;k[h+16>>2]=t;F=v;G=k[p>>2]|0;H=E}else{F=w;G=y;H=3}else{F=w;G=0;H=3}}else{F=w;G=x;H=2}else{F=0;G=x;H=2}}else{F=C;G=x;H=1}else{F=C;G=0;H=1}C=F>>>0>G>>>0?F:G;k[o>>2]=C;if((i[C>>0]|0)==58)I=jx(C+1|0,0,10)|0;else I=0;C=ha(k[h+4>>2]|0,k[h>>2]|0)|0;o=ha(C,k[h+8>>2]|0)|0;C=ha(o,k[h+12>>2]|0)|0;o=ha(C,k[h+16>>2]|0)|0;C=ha(o,z)|0;z=C+2879|0;G=2880-((z>>>0)%2880|0)+z|0;k[l>>2]=G;z=ri(n,0,e)|0;k[g>>2]=z;if(z){tk(738646);tk(n);q=k[g>>2]|0;r=d;return q|0}k[c>>2]=-1;n=0;while(1){z=34840+(n*48|0)|0;if(!(k[z>>2]|0)){J=z;K=n;u=32;break}n=n+1|0;if((n|0)>=1e3){L=103;break}}do if((u|0)==32){k[c>>2]=K;if((K|0)==-1)L=103;else{n=34840+(K*48|0)+4|0;k[J>>2]=n;z=34840+(K*48|0)+12|0;k[34840+(K*48|0)+8>>2]=z;if((G|0)!=0?(F=cy(G)|0,k[n>>2]=F,(F|0)==0):0){tk(736089);L=104;break}k[z>>2]=G;k[34840+(K*48|0)+16>>2]=2880;z=34840+(K*48|0)+24|0;k[z>>2]=0;k[z+4>>2]=0;k[z+8>>2]=0;k[z+12>>2]=0;k[34840+(K*48|0)+20>>2]=19;k[g>>2]=0;$h(f,k[34840+((k[c>>2]|0)*48|0)>>2]|0,l,0,0,g)|0;Xr(k[f>>2]|0,A,H,h,g)|0;Fh(k[f>>2]|0,g)|0;if((k[g>>2]|0)>0){tk(738745);ww(k[e>>2]|0)|0;z=k[c>>2]|0;F=34840+(z*48|0)|0;dy(k[k[F>>2]>>2]|0);k[F>>2]=0;k[34840+(z*48|0)+4>>2]=0;q=k[g>>2]|0;r=d;return q|0}if((I|0)>0)Kw(k[e>>2]|0,I,0)|0;z=(k[k[34840+((k[c>>2]|0)*48|0)>>2]>>2]|0)+2880|0;if((Hw(z,1,C,k[e>>2]|0)|0)!=(C|0))k[g>>2]=108;ww(k[e>>2]|0)|0;if(k[g>>2]|0){F=k[c>>2]|0;n=34840+(F*48|0)|0;dy(k[k[n>>2]>>2]|0);k[n>>2]=0;k[34840+(F*48|0)+4>>2]=0;tk(738799);q=k[g>>2]|0;r=d;return q|0}do if((A|0)==20){F=(o|0)>0;if((D|0)==1){if(F){M=0;N=z}else{u=53;break}while(1){j[N>>1]=m[N>>1]^32768;M=M+1|0;if((M|0)==(o|0)){u=51;break}else N=N+2|0}}else{if(F){O=0;P=z}else{u=51;break}while(1){j[P>>1]=m[P>>1]^128;O=O+1|0;if((O|0)==(o|0)){u=51;break}else P=P+2|0}}}else u=51;while(0);b:do if((u|0)==51?(D|0)!=0:0)if((A&-8|0)==16)u=53;else switch(A|0){case 32:case -32:{Ls(z,o);break b;break}case -64:{Ms(z,o);break b;break}default:break b}while(0);if((u|0)==53)Ks(z,o);F=k[c>>2]|0;n=34840+(F*48|0)+24|0;k[n>>2]=0;k[n+4>>2]=0;n=34840+(F*48|0)+32|0;k[n>>2]=k[l>>2];k[n+4>>2]=0;q=0;r=d;return q|0}}while(0);k[g>>2]=L;tk(738697);ww(k[e>>2]|0)|0;q=k[g>>2]|0;r=d;return q|0}function fj(a,b){a=a|0;b=b|0;var c=0,d=0;c=34840+(a*48|0)+32|0;a=k[c+4>>2]|0;d=b;k[d>>2]=k[c>>2];k[d+4>>2]=a;return 0}function gj(a){a=a|0;k[34840+(a*48|0)>>2]=0;k[34840+(a*48|0)+4>>2]=0;return 0}function hj(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=r;r=r+16|0;c=b+4|0;k[c>>2]=0;d=34840+(a*48|0)+4|0;e=34840+(a*48|0)+40|0;if(th(k[d>>2]|0,k[34840+(a*48|0)+32>>2]|0,k[e>>2]|0,b,c)|0){tk(738859);k[c>>2]=106}dy(k[d>>2]|0);k[34840+(a*48|0)>>2]=0;k[d>>2]=0;d=k[e>>2]|0;if((d|0)==(k[31231]|0)){f=k[c>>2]|0;r=b;return f|0}ww(d)|0;f=k[c>>2]|0;r=b;return f|0}function ij(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;d=34840+(a*48|0)+32|0;e=k[d+4>>2]|0;if((e|0)<(c|0)|((e|0)==(c|0)?(k[d>>2]|0)>>>0<b>>>0:0)){f=107;return f|0}d=34840+(a*48|0)+24|0;k[d>>2]=b;k[d+4>>2]=c;f=0;return f|0}function jj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;d=34840+(a*48|0)+24|0;e=d;f=k[e>>2]|0;g=((c|0)<0)<<31>>31;h=ky(f|0,k[e+4>>2]|0,c|0,g|0)|0;e=L;i=34840+(a*48|0)+32|0;j=k[i+4>>2]|0;if((e|0)>(j|0)|((e|0)==(j|0)?h>>>0>(k[i>>2]|0)>>>0:0)){l=107;return l|0}py(b|0,(k[k[34840+(a*48|0)>>2]>>2]|0)+f|0,c|0)|0;f=d;a=ky(k[f>>2]|0,k[f+4>>2]|0,c|0,g|0)|0;g=d;k[g>>2]=a;k[g+4>>2]=L;l=0;return l|0}function kj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0;d=34840+(a*48|0)|0;e=34840+(a*48|0)+24|0;f=e;g=k[f>>2]|0;h=k[f+4>>2]|0;f=((c|0)<0)<<31>>31;i=ky(g|0,h|0,c|0,f|0)|0;j=L;l=34840+(a*48|0)+8|0;m=k[k[l>>2]>>2]|0;do if(i>>>0>m>>>0){n=k[34840+(a*48|0)+20>>2]|0;if(!n){tk(738911);o=106;return o|0}p=ky(i|0,j|0,-1,-1)|0;q=wy(p|0,L|0,2880,0)|0;p=ky(q|0,L|0,1,0)|0;q=p*2880|0;p=(k[34840+(a*48|0)+16>>2]|0)+m|0;r=q>>>0>p>>>0?q:p;p=Xb[n&31](k[k[d>>2]>>2]|0,r)|0;if(p){k[k[d>>2]>>2]=p;k[k[l>>2]>>2]=r;r=e;s=d;t=k[r>>2]|0;u=k[r+4>>2]|0;break}tk(738952);o=113;return o|0}else{s=d;t=g;u=h}while(0);py((k[k[s>>2]>>2]|0)+t|0,b|0,c|0)|0;b=e;t=ky(k[b>>2]|0,k[b+4>>2]|0,c|0,f|0)|0;f=L;c=e;k[c>>2]=t;k[c+4>>2]=f;c=34840+(a*48|0)+32|0;a=c;e=k[a>>2]|0;b=k[a+4>>2]|0;a=(b|0)>(f|0)|(b|0)==(f|0)&e>>>0>t>>>0;s=c;k[s>>2]=a?e:t;k[s+4>>2]=a?b:f;o=0;return o|0}function lj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0;d=r;r=r+3728|0;e=d;f=4;g=cy(40)|0;k[g>>2]=0;h=d+8|0;j=d+3616|0;l=d+2416|0;m=d+1216|0;n=d+16|0;o=d+4|0;k[29923]=0;k[29924]=0;if(!b){g=ly(119700,1,g|0,f|0)|0;f=L;w=0;b=w;w=0;if((b|0)!=0&(x|0)!=0){p=oy(k[b>>2]|0,g|0,f|0)|0;if(!p)Qa(b|0,x|0);L=x}else p=-1;if((p|0)==1)q=L;else q=0}else{w=0;qa(2,738992);p=w;w=0;if((p|0)!=0&(x|0)!=0){s=oy(k[p>>2]|0,g|0,f|0)|0;if(!s)Qa(p|0,x|0);L=x}else s=-1;if((s|0)!=1){w=0;qa(2,739043);s=w;w=0;if((s|0)!=0&(x|0)!=0){t=oy(k[s>>2]|0,g|0,f|0)|0;if(!t)Qa(s|0,x|0);L=x}else t=-1;if((t|0)!=1)u=90;else q=L}else q=L}a:while(1){if((u|0)==90){u=0;w=0;sa(26,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){v=oy(k[t>>2]|0,g|0,f|0)|0;if(!v)Qa(t|0,x|0);L=x}else v=-1;if((v|0)==1){q=L;continue}if(k[29923]|0){w=0;sa(4,k[h>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){y=oy(k[t>>2]|0,g|0,f|0)|0;if(!y)Qa(t|0,x|0);L=x}else y=-1;if((y|0)==1){q=L;continue}}if(k[29924]|0){w=0;sa(15,k[c>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){z=oy(k[t>>2]|0,g|0,f|0)|0;if(!z)Qa(t|0,x|0);L=x}else z=-1;if((z|0)==1){q=L;continue}}w=0;va(20,14,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){A=oy(k[t>>2]|0,g|0,f|0)|0;if(!A)Qa(t|0,x|0);L=x}else A=-1;if((A|0)==1)q=L;else{u=98;break}}if(q){w=0;qa(2,739091);t=w;w=0;if((t|0)!=0&(x|0)!=0){B=oy(k[t>>2]|0,g|0,f|0)|0;if(!B)Qa(t|0,x|0);L=x}else B=-1;if((B|0)==1){q=L;continue}u=90;continue}w=0;va(20,14,3)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){C=oy(k[t>>2]|0,g|0,f|0)|0;if(!C)Qa(t|0,x|0);L=x}else C=-1;if((C|0)==1){q=L;continue}w=0;t=va(21,a|0,739111)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){D=oy(k[s>>2]|0,g|0,f|0)|0;if(!D)Qa(s|0,x|0);L=x}else D=-1;if((D|0)==1){q=L;continue}do if(!t){w=0;s=va(21,a|0,739114)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){E=oy(k[p>>2]|0,g|0,f|0)|0;if(!E)Qa(p|0,x|0);L=x}else E=-1;if((E|0)==1){q=L;continue a}if(!s){w=0;s=va(22,a|0,63)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){F=oy(k[p>>2]|0,g|0,f|0)|0;if(!F)Qa(p|0,x|0);L=x}else F=-1;if((F|0)==1){q=L;continue a}if(!s){w=0;s=sa(27,a|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){G=oy(k[p>>2]|0,g|0,f|0)|0;if(!G)Qa(p|0,x|0);L=x}else G=-1;if((G|0)==1){q=L;continue a}if(s>>>0>1195){w=0;qa(2,739156);s=w;w=0;if((s|0)!=0&(x|0)!=0){H=oy(k[s>>2]|0,g|0,f|0)|0;if(!H)Qa(s|0,x|0);L=x}else H=-1;if((H|0)==1){q=L;continue a}w=0;qa(2,a|0);s=w;w=0;if((s|0)!=0&(x|0)!=0){I=oy(k[s>>2]|0,g|0,f|0)|0;if(!I)Qa(s|0,x|0);L=x}else I=-1;if((I|0)==1){q=L;continue a}u=90;continue a}w=0;sa(26,180)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){J=oy(k[s>>2]|0,g|0,f|0)|0;if(!J)Qa(s|0,x|0);L=x}else J=-1;if((J|0)==1){q=L;continue a}w=0;va(23,l|0,a|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){K=oy(k[s>>2]|0,g|0,f|0)|0;if(!K)Qa(s|0,x|0);L=x}else K=-1;if((K|0)==1){q=L;continue a}w=0;s=sa(27,l|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){M=oy(k[p>>2]|0,g|0,f|0)|0;if(!M)Qa(p|0,x|0);L=x}else M=-1;if((M|0)==1){q=L;continue a}p=l+s|0;i[p>>0]=46;i[p+1>>0]=103;i[p+2>>0]=122;i[p+3>>0]=0;w=0;p=ua(3,l|0,h|0,j|0,o|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){N=oy(k[s>>2]|0,g|0,f|0)|0;if(!N)Qa(s|0,x|0);L=x}else N=-1;if((N|0)==1){q=L;continue a}if(p){w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){O=oy(k[p>>2]|0,g|0,f|0)|0;if(!O)Qa(p|0,x|0);L=x}else O=-1;if((O|0)==1){q=L;continue a}w=0;va(23,l|0,a|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){P=oy(k[p>>2]|0,g|0,f|0)|0;if(!P)Qa(p|0,x|0);L=x}else P=-1;if((P|0)==1){q=L;continue a}w=0;p=sa(27,l|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){Q=oy(k[s>>2]|0,g|0,f|0)|0;if(!Q)Qa(s|0,x|0);L=x}else Q=-1;if((Q|0)==1){q=L;continue a}s=l+p|0;i[s>>0]=i[739111]|0;i[s+1>>0]=i[739112]|0;i[s+2>>0]=i[739113]|0;w=0;sa(26,180)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){R=oy(k[s>>2]|0,g|0,f|0)|0;if(!R)Qa(s|0,x|0);L=x}else R=-1;if((R|0)==1){q=L;continue a}w=0;s=ua(3,l|0,h|0,j|0,o|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){S=oy(k[p>>2]|0,g|0,f|0)|0;if(!S)Qa(p|0,x|0);L=x}else S=-1;if((S|0)==1){q=L;continue a}if(!s)break;w=0;sa(26,0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){T=oy(k[s>>2]|0,g|0,f|0)|0;if(!T)Qa(s|0,x|0);L=x}else T=-1;if((T|0)==1){q=L;continue a}w=0;sa(26,180)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){U=oy(k[s>>2]|0,g|0,f|0)|0;if(!U)Qa(s|0,x|0);L=x}else U=-1;if((U|0)==1){q=L;continue a}w=0;s=ua(3,a|0,h|0,j|0,o|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){V=oy(k[p>>2]|0,g|0,f|0)|0;if(!V)Qa(p|0,x|0);L=x}else V=-1;if((V|0)==1){q=L;continue a}if(!s)break;w=0;sa(26,0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){W=oy(k[s>>2]|0,g|0,f|0)|0;if(!W)Qa(s|0,x|0);L=x}else W=-1;if((W|0)==1){q=L;continue a}w=0;qa(2,739195);s=w;w=0;if((s|0)!=0&(x|0)!=0){X=oy(k[s>>2]|0,g|0,f|0)|0;if(!X)Qa(s|0,x|0);L=x}else X=-1;if((X|0)==1){q=L;continue a}w=0;qa(2,a|0);s=w;w=0;if((s|0)!=0&(x|0)!=0){Y=oy(k[s>>2]|0,g|0,f|0)|0;if(!Y)Qa(s|0,x|0);L=x}else Y=-1;if((Y|0)==1){q=L;continue a}u=90;continue a}}else u=17}else u=17}else u=17;while(0);if((u|0)==17){u=0;w=0;sa(26,180)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){Z=oy(k[t>>2]|0,g|0,f|0)|0;if(!Z)Qa(t|0,x|0);L=x}else Z=-1;if((Z|0)==1){q=L;continue}w=0;t=ua(3,a|0,h|0,j|0,o|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){_=oy(k[s>>2]|0,g|0,f|0)|0;if(!_)Qa(s|0,x|0);L=x}else _=-1;if((_|0)==1){q=L;continue}if(t){w=0;sa(26,0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){$=oy(k[t>>2]|0,g|0,f|0)|0;if(!$)Qa(t|0,x|0);L=x}else $=-1;if(($|0)==1){q=L;continue}w=0;qa(2,739118);t=w;w=0;if((t|0)!=0&(x|0)!=0){aa=oy(k[t>>2]|0,g|0,f|0)|0;if(!aa)Qa(t|0,x|0);L=x}else aa=-1;if((aa|0)==1){q=L;continue}w=0;qa(2,a|0);t=w;w=0;if((t|0)!=0&(x|0)!=0){ba=oy(k[t>>2]|0,g|0,f|0)|0;if(!ba)Qa(t|0,x|0);L=x}else ba=-1;if((ba|0)==1){q=L;continue}u=90;continue}}k[29923]=(k[29923]|0)+1;w=0;t=va(12,a|0,c|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){ca=oy(k[s>>2]|0,g|0,f|0)|0;if(!ca)Qa(s|0,x|0);L=x}else ca=-1;if((ca|0)==1){q=L;continue}if(t){w=0;qa(2,739232);t=w;w=0;if((t|0)!=0&(x|0)!=0){da=oy(k[t>>2]|0,g|0,f|0)|0;if(!da)Qa(t|0,x|0);L=x}else da=-1;if((da|0)==1){q=L;continue}u=90;continue}k[29924]=(k[29924]|0)+1;t=k[h>>2]|0;w=0;s=sa(28,t|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){ea=oy(k[p>>2]|0,g|0,f|0)|0;if(!ea)Qa(p|0,x|0);L=x}else ea=-1;if((ea|0)==1){q=L;continue}p=s<<24>>24;w=0;va(24,p|0,t|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){fa=oy(k[s>>2]|0,g|0,f|0)|0;if(!fa)Qa(s|0,x|0);L=x}else fa=-1;if((fa|0)==1){q=L;continue}w=0;s=va(25,j|0,739273)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ga=oy(k[b>>2]|0,g|0,f|0)|0;if(!ga)Qa(b|0,x|0);L=x}else ga=-1;if((ga|0)==1){q=L;continue}b:do if(s){w=0;b=va(25,j|0,739280)|0;ha=w;w=0;if((ha|0)!=0&(x|0)!=0){ia=oy(k[ha>>2]|0,g|0,f|0)|0;if(!ia)Qa(ha|0,x|0);L=x}else ia=-1;if((ia|0)==1){q=L;continue a}if(b){w=0;b=va(21,a|0,739114)|0;ha=w;w=0;if((ha|0)!=0&(x|0)!=0){ja=oy(k[ha>>2]|0,g|0,f|0)|0;if(!ja)Qa(ha|0,x|0);L=x}else ja=-1;if((ja|0)==1){q=L;continue a}if(!b){w=0;b=va(21,a|0,739111)|0;ha=w;w=0;if((ha|0)!=0&(x|0)!=0){ka=oy(k[ha>>2]|0,g|0,f|0)|0;if(!ka)Qa(ha|0,x|0);L=x}else ka=-1;if((ka|0)==1){q=L;continue a}if(!((p|0)==31|(b|0)!=0)){b=k[o>>2]|0;if((b|0)%2880|0){w=0;k[e>>2]=b;na(42,m|0,739340,e|0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){la=oy(k[b>>2]|0,g|0,f|0)|0;if(!la)Qa(b|0,x|0);L=x}else la=-1;if((la|0)==1){q=L;continue a}w=0;qa(2,m|0);b=w;w=0;if((b|0)!=0&(x|0)!=0){ma=oy(k[b>>2]|0,g|0,f|0)|0;if(!ma)Qa(b|0,x|0);L=x}else ma=-1;if((ma|0)==1){q=L;continue a}}w=0;sa(26,180)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){oa=oy(k[b>>2]|0,g|0,f|0)|0;if(!oa)Qa(b|0,x|0);L=x}else oa=-1;if((oa|0)==1){q=L;continue a}w=0;b=ua(2,n|0,1,1200,t|0)|0;ha=w;w=0;if((ha|0)!=0&(x|0)!=0){pa=oy(k[ha>>2]|0,g|0,f|0)|0;if(!pa)Qa(ha|0,x|0);L=x}else pa=-1;if((pa|0)==1){q=L;continue a}if(!b)break;else ra=b;while(1){w=0;sa(26,0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ta=oy(k[b>>2]|0,g|0,f|0)|0;if(!ta)Qa(b|0,x|0);L=x}else ta=-1;if((ta|0)==1){q=L;continue a}w=0;b=na(17,k[c>>2]|0,n|0,ra|0)|0;ha=w;w=0;if((ha|0)!=0&(x|0)!=0){wa=oy(k[ha>>2]|0,g|0,f|0)|0;if(!wa)Qa(ha|0,x|0);L=x}else wa=-1;if((wa|0)==1){q=L;continue a}if(b)break;w=0;sa(26,180)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){xa=oy(k[b>>2]|0,g|0,f|0)|0;if(!xa)Qa(b|0,x|0);L=x}else xa=-1;if((xa|0)==1){q=L;continue a}w=0;ra=ua(2,n|0,1,1200,t|0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){ya=oy(k[b>>2]|0,g|0,f|0)|0;if(!ya)Qa(b|0,x|0);L=x}else ya=-1;if((ya|0)==1){q=L;continue a}if(!ra)break b}w=0;qa(2,739393);b=w;w=0;if((b|0)!=0&(x|0)!=0){za=oy(k[b>>2]|0,g|0,f|0)|0;if(!za)Qa(b|0,x|0);L=x}else za=-1;if((za|0)==1){q=L;continue a}w=0;qa(2,a|0);b=w;w=0;if((b|0)!=0&(x|0)!=0){Aa=oy(k[b>>2]|0,g|0,f|0)|0;if(!Aa)Qa(b|0,x|0);L=x}else Aa=-1;if((Aa|0)==1){q=L;continue a}u=90;continue a}else u=62}else u=62}else u=62}else u=62;while(0);if((u|0)==62){u=0;w=0;sa(26,1800)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Ba=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ba)Qa(p|0,x|0);L=x}else Ba=-1;if((Ba|0)==1){q=L;continue}w=0;p=na(41,a|0,t|0,k[c>>2]|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){Ca=oy(k[s>>2]|0,g|0,f|0)|0;if(!Ca)Qa(s|0,x|0);L=x}else Ca=-1;if((Ca|0)==1){q=L;continue}w=0;sa(26,0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){Da=oy(k[s>>2]|0,g|0,f|0)|0;if(!Da)Qa(s|0,x|0);L=x}else Da=-1;if((Da|0)==1){q=L;continue}if(p){w=0;qa(2,739291);p=w;w=0;if((p|0)!=0&(x|0)!=0){Ea=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ea)Qa(p|0,x|0);L=x}else Ea=-1;if((Ea|0)==1){q=L;continue}w=0;qa(2,a|0);p=w;w=0;if((p|0)!=0&(x|0)!=0){Fa=oy(k[p>>2]|0,g|0,f|0)|0;if(!Fa)Qa(p|0,x|0);L=x}else Fa=-1;if((Fa|0)==1){q=L;continue}u=90;continue}}w=0;sa(4,t|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Ga=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ga)Qa(p|0,x|0);L=x}else Ga=-1;if((Ga|0)==1){q=L;continue}w=0;va(20,14,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Ha=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ha)Qa(p|0,x|0);L=x}else Ha=-1;if((Ha|0)==1){q=L;continue}w=0;sa(26,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Ia=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ia)Qa(p|0,x|0);L=x}else Ia=-1;if((Ia|0)==1){q=L;continue}w=0;Ja=na(15,k[c>>2]|0,0,0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Ka=oy(k[p>>2]|0,g|0,f|0)|0;if(!Ka)Qa(p|0,x|0);L=x}else Ka=-1;if((Ka|0)==1)q=L;else{u=89;break}}if((u|0)==89){La=Ja;dy(g|0);r=d;return La|0}else if((u|0)==98){La=104;dy(g|0);r=d;return La|0}return 0}function mj(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,oa=0,pa=0,ra=0,ta=0;d=r;r=r+1312|0;e=4;f=cy(40)|0;k[f>>2]=0;g=d+4|0;h=d+1208|0;j=d+8|0;l=d;k[29923]=0;k[29964]=0;k[29965]=0;k[29924]=0;w=0;m=sa(27,739441)|0;n=w;w=0;if((n|0)!=0&(x|0)!=0){o=oy(k[n>>2]|0,f|0,e|0)|0;if(!o)Qa(n|0,x|0);L=x}else o=-1;do if((o|0)!=1){if(!m){w=0;qa(2,740641);n=w;w=0;if((n|0)!=0&(x|0)!=0){p=oy(k[n>>2]|0,f|0,e|0)|0;if(!p)Qa(n|0,x|0);L=x}else p=-1;if((p|0)==1){q=L;break}s=73;break}if(!b){f=ly(119700,1,f|0,e|0)|0;e=L;w=0;n=w;w=0;if((n|0)!=0&(x|0)!=0){t=oy(k[n>>2]|0,f|0,e|0)|0;if(!t)Qa(n|0,x|0);L=x}else t=-1;if((t|0)==1){q=L;break}q=0;break}w=0;qa(2,740707);n=w;w=0;if((n|0)!=0&(x|0)!=0){u=oy(k[n>>2]|0,f|0,e|0)|0;if(!u)Qa(n|0,x|0);L=x}else u=-1;if((u|0)!=1){w=0;qa(2,740769);n=w;w=0;if((n|0)!=0&(x|0)!=0){v=oy(k[n>>2]|0,f|0,e|0)|0;if(!v)Qa(n|0,x|0);L=x}else v=-1;if((v|0)!=1)s=73;else q=L}else q=L}else q=L;while(0);a:while(1){if((s|0)==73){s=0;w=0;sa(26,0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){y=oy(k[v>>2]|0,f|0,e|0)|0;if(!y)Qa(v|0,x|0);L=x}else y=-1;if((y|0)==1){q=L;continue}if(k[29923]|0){w=0;sa(4,k[g>>2]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){z=oy(k[v>>2]|0,f|0,e|0)|0;if(!z)Qa(v|0,x|0);L=x}else z=-1;if((z|0)==1){q=L;continue}}if(k[29965]|0){w=0;sa(4,k[29966]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){A=oy(k[v>>2]|0,f|0,e|0)|0;if(!A)Qa(v|0,x|0);L=x}else A=-1;if((A|0)==1){q=L;continue}}if(k[29924]|0){w=0;sa(15,k[c>>2]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){B=oy(k[v>>2]|0,f|0,e|0)|0;if(!B)Qa(v|0,x|0);L=x}else B=-1;if((B|0)==1){q=L;continue}}if(k[29964]|0){w=0;sa(9,k[c>>2]|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){C=oy(k[v>>2]|0,f|0,e|0)|0;if(!C)Qa(v|0,x|0);L=x}else C=-1;if((C|0)==1){q=L;continue}}w=0;va(20,14,0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){D=oy(k[v>>2]|0,f|0,e|0)|0;if(!D)Qa(v|0,x|0);L=x}else D=-1;if((D|0)==1)q=L;else{s=87;break}}if(q){w=0;qa(2,739091);v=w;w=0;if((v|0)!=0&(x|0)!=0){E=oy(k[v>>2]|0,f|0,e|0)|0;if(!E)Qa(v|0,x|0);L=x}else E=-1;if((E|0)==1){q=L;continue}s=73;continue}w=0;va(20,14,3)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){F=oy(k[v>>2]|0,f|0,e|0)|0;if(!F)Qa(v|0,x|0);L=x}else F=-1;if((F|0)==1){q=L;continue}w=0;sa(26,180)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){G=oy(k[v>>2]|0,f|0,e|0)|0;if(!G)Qa(v|0,x|0);L=x}else G=-1;if((G|0)==1){q=L;continue}w=0;v=ua(3,a|0,g|0,h|0,l|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){H=oy(k[u>>2]|0,f|0,e|0)|0;if(!H)Qa(u|0,x|0);L=x}else H=-1;if((H|0)==1){q=L;continue}if(v){w=0;sa(26,0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){I=oy(k[v>>2]|0,f|0,e|0)|0;if(!I)Qa(v|0,x|0);L=x}else I=-1;if((I|0)==1){q=L;continue}w=0;qa(2,740824);v=w;w=0;if((v|0)!=0&(x|0)!=0){J=oy(k[v>>2]|0,f|0,e|0)|0;if(!J)Qa(v|0,x|0);L=x}else J=-1;if((J|0)==1){q=L;continue}w=0;qa(2,a|0);v=w;w=0;if((v|0)!=0&(x|0)!=0){K=oy(k[v>>2]|0,f|0,e|0)|0;if(!K)Qa(v|0,x|0);L=x}else K=-1;if((K|0)==1){q=L;continue}s=73;continue}k[29923]=(k[29923]|0)+1;v=k[g>>2]|0;w=0;u=sa(28,v|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){M=oy(k[t>>2]|0,f|0,e|0)|0;if(!M)Qa(t|0,x|0);L=x}else M=-1;if((M|0)==1){q=L;continue}t=u<<24>>24;w=0;va(24,t|0,v|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){N=oy(k[u>>2]|0,f|0,e|0)|0;if(!N)Qa(u|0,x|0);L=x}else N=-1;if((N|0)==1){q=L;continue}w=0;u=va(25,h|0,739273)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){O=oy(k[b>>2]|0,f|0,e|0)|0;if(!O)Qa(b|0,x|0);L=x}else O=-1;if((O|0)==1){q=L;continue}if(u){w=0;u=va(25,h|0,739280)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){P=oy(k[b>>2]|0,f|0,e|0)|0;if(!P)Qa(b|0,x|0);L=x}else P=-1;if((P|0)==1){q=L;continue}if(!((t|0)==31|(u|0)==0)){w=0;qa(2,741129);u=w;w=0;if((u|0)!=0&(x|0)!=0){Q=oy(k[u>>2]|0,f|0,e|0)|0;if(!Q)Qa(u|0,x|0);L=x}else Q=-1;if((Q|0)==1){q=L;continue}s=73;continue}}if((i[739441]|0)==33){if((m|0)>0){u=0;do{t=u;u=u+1|0;i[739441+t>>0]=i[739441+u>>0]|0}while((u|0)!=(m|0))}w=0;sa(10,739441)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){R=oy(k[u>>2]|0,f|0,e|0)|0;if(!R)Qa(u|0,x|0);L=x}else R=-1;if((R|0)==1){q=L;continue}}w=0;u=va(10,739441,c|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){S=oy(k[t>>2]|0,f|0,e|0)|0;if(!S)Qa(t|0,x|0);L=x}else S=-1;if((S|0)==1){q=L;continue}if(u){w=0;qa(2,740870);u=w;w=0;if((u|0)!=0&(x|0)!=0){T=oy(k[u>>2]|0,f|0,e|0)|0;if(!T)Qa(u|0,x|0);L=x}else T=-1;if((T|0)==1){q=L;continue}w=0;qa(2,739441);u=w;w=0;if((u|0)!=0&(x|0)!=0){U=oy(k[u>>2]|0,f|0,e|0)|0;if(!U)Qa(u|0,x|0);L=x}else U=-1;if((U|0)==1){q=L;continue}s=73;continue}k[29964]=(k[29964]|0)+1;w=0;sa(26,180)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){V=oy(k[u>>2]|0,f|0,e|0)|0;if(!V)Qa(u|0,x|0);L=x}else V=-1;if((V|0)==1){q=L;continue}w=0;u=ua(2,j|0,1,1200,v|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){W=oy(k[t>>2]|0,f|0,e|0)|0;if(!W)Qa(t|0,x|0);L=x}else W=-1;if((W|0)==1){q=L;continue}b:do if(u){t=u;while(1){w=0;sa(26,0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){X=oy(k[b>>2]|0,f|0,e|0)|0;if(!X)Qa(b|0,x|0);L=x}else X=-1;if((X|0)==1){q=L;continue a}w=0;b=na(13,k[c>>2]|0,j|0,t|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){Y=oy(k[p>>2]|0,f|0,e|0)|0;if(!Y)Qa(p|0,x|0);L=x}else Y=-1;if((Y|0)==1){q=L;continue a}if(b)break;w=0;sa(26,180)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){Z=oy(k[b>>2]|0,f|0,e|0)|0;if(!Z)Qa(b|0,x|0);L=x}else Z=-1;if((Z|0)==1){q=L;continue a}w=0;t=ua(2,j|0,1,1200,v|0)|0;b=w;w=0;if((b|0)!=0&(x|0)!=0){_=oy(k[b>>2]|0,f|0,e|0)|0;if(!_)Qa(b|0,x|0);L=x}else _=-1;if((_|0)==1){q=L;continue a}if(!t)break b}w=0;qa(2,740926);t=w;w=0;if((t|0)!=0&(x|0)!=0){$=oy(k[t>>2]|0,f|0,e|0)|0;if(!$)Qa(t|0,x|0);L=x}else $=-1;if(($|0)==1){q=L;continue a}w=0;qa(2,739441);t=w;w=0;if((t|0)!=0&(x|0)!=0){aa=oy(k[t>>2]|0,f|0,e|0)|0;if(!aa)Qa(t|0,x|0);L=x}else aa=-1;if((aa|0)==1){q=L;continue a}s=73;continue a}while(0);w=0;sa(9,k[c>>2]|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){ba=oy(k[u>>2]|0,f|0,e|0)|0;if(!ba)Qa(u|0,x|0);L=x}else ba=-1;if((ba|0)==1){q=L;continue}w=0;sa(4,v|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){ca=oy(k[u>>2]|0,f|0,e|0)|0;if(!ca)Qa(u|0,x|0);L=x}else ca=-1;if((ca|0)==1){q=L;continue}k[29923]=(k[29923]|0)+-1;k[29964]=(k[29964]|0)+-1;w=0;u=va(7,739441,774853)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){da=oy(k[t>>2]|0,f|0,e|0)|0;if(!da)Qa(t|0,x|0);L=x}else da=-1;if((da|0)==1){q=L;continue}k[29966]=u;if(!u){w=0;qa(2,740970);u=w;w=0;if((u|0)!=0&(x|0)!=0){ea=oy(k[u>>2]|0,f|0,e|0)|0;if(!ea)Qa(u|0,x|0);L=x}else ea=-1;if((ea|0)==1){q=L;continue}w=0;qa(2,739441);u=w;w=0;if((u|0)!=0&(x|0)!=0){fa=oy(k[u>>2]|0,f|0,e|0)|0;if(!fa)Qa(u|0,x|0);L=x}else fa=-1;if((fa|0)==1){q=L;continue}s=73;continue}k[29965]=(k[29965]|0)+1;w=0;u=va(12,a|0,c|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ga=oy(k[t>>2]|0,f|0,e|0)|0;if(!ga)Qa(t|0,x|0);L=x}else ga=-1;if((ga|0)==1){q=L;continue}if(u){w=0;qa(2,741018);u=w;w=0;if((u|0)!=0&(x|0)!=0){ha=oy(k[u>>2]|0,f|0,e|0)|0;if(!ha)Qa(u|0,x|0);L=x}else ha=-1;if((ha|0)==1){q=L;continue}s=73;continue}k[29924]=(k[29924]|0)+1;w=0;u=na(41,a|0,k[29966]|0,k[c>>2]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ia=oy(k[t>>2]|0,f|0,e|0)|0;if(!ia)Qa(t|0,x|0);L=x}else ia=-1;if((ia|0)==1){q=L;continue}w=0;sa(4,k[29966]|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){ja=oy(k[t>>2]|0,f|0,e|0)|0;if(!ja)Qa(t|0,x|0);L=x}else ja=-1;if((ja|0)==1){q=L;continue}k[29965]=(k[29965]|0)+-1;if(!u){w=0;va(20,14,0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){ka=oy(k[u>>2]|0,f|0,e|0)|0;if(!ka)Qa(u|0,x|0);L=x}else ka=-1;if((ka|0)==1){q=L;continue}w=0;sa(26,0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){la=oy(k[u>>2]|0,f|0,e|0)|0;if(!la)Qa(u|0,x|0);L=x}else la=-1;if((la|0)==1){q=L;continue}w=0;ma=na(15,k[c>>2]|0,0,0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){oa=oy(k[u>>2]|0,f|0,e|0)|0;if(!oa)Qa(u|0,x|0);L=x}else oa=-1;if((oa|0)==1){q=L;continue}else{s=72;break}}else{w=0;qa(2,741068);u=w;w=0;if((u|0)!=0&(x|0)!=0){pa=oy(k[u>>2]|0,f|0,e|0)|0;if(!pa)Qa(u|0,x|0);L=x}else pa=-1;if((pa|0)==1){q=L;continue}w=0;qa(2,739441);u=w;w=0;if((u|0)!=0&(x|0)!=0){ra=oy(k[u>>2]|0,f|0,e|0)|0;if(!ra)Qa(u|0,x|0);L=x}else ra=-1;if((ra|0)==1){q=L;continue}s=73;continue}}if((s|0)==72){ta=ma;dy(f|0);r=d;return ta|0}else if((s|0)==87){ta=104;dy(f|0);r=d;return ta|0}return 0}
function Ho(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0;i[b>>0]=0;k[c>>2]=0;e=i[a>>0]|0;if(e<<24>>24==72)if(!(Ax(a,761298,9)|0)){f=rx(a,61)|0;if(!f){g=b+(xx(b)|0)|0;h=767038;j=g+9|0;do{i[g>>0]=i[h>>0]|0;g=g+1|0;h=h+1|0}while((g|0)<(j|0));k[c>>2]=8;l=k[d>>2]|0;return l|0}h=a+9|0;while(1)if((i[h>>0]|0)==32)h=h+1|0;else{m=h;break}h=f-m|0;zx(b,m,h)|0;m=h;while(1){if((m|0)<=0){n=m;break}h=m+-1|0;if((i[b+h>>0]|0)==32)m=h;else{n=m;break}}i[b+n>>0]=0;o=n}else{p=72;q=0;r=11}else{p=e;q=0;r=11}do if((r|0)==11){a:while(1){r=0;switch(p<<24>>24){case 0:case 61:case 32:{s=q;break a;break}default:{}}i[b+q>>0]=p;e=q+1|0;if((e|0)>=74){r=15;break}p=i[a+e>>0]|0;q=e;r=11}if((r|0)==15){i[b+74>>0]=0;o=74;break}i[b+s>>0]=0;k[c>>2]=s;l=k[d>>2]|0;return l|0}while(0);k[c>>2]=o;l=k[d>>2]|0;return l|0}function Io(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0.0;g=r;r=r+176|0;h=g+83|0;l=g+12|0;m=g+8|0;n=g;o=k[f>>2]|0;if((o|0)>0){q=o;r=g;return q|0}a:do switch(b|0){case 16:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=5;if((s|0)==5?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;i[d>>0]=0;fl(l,d,f)|0;break}case 11:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=10;if((s|0)==10?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1){o=k[m>>2]|0;if(o>>>0>255){k[f>>2]=412;break a}else{i[d>>0]=o;break a}}break}case 12:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=18;if((s|0)==18?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1){o=k[m>>2]|0;if((o+128|0)>>>0>255){k[f>>2]=412;break a}else{i[d>>0]=o;break a}}break}case 20:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=26;if((s|0)==26?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1){o=k[m>>2]|0;if(o>>>0>65535){k[f>>2]=412;break a}else{j[d>>1]=o;break a}}break}case 21:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=34;if((s|0)==34?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1){o=k[m>>2]|0;if((o+32768|0)>>>0>65535){k[f>>2]=412;break a}else{j[d>>1]=o;break a}}break}case 30:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=42;if((s|0)==42?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1)k[f>>2]=412;break}case 31:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=48;if((s|0)==48?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,m,f)|0;if((k[f>>2]|0)<1)k[d>>2]=k[m>>2];break}case 14:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=54;if((s|0)==54?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Sl(l,d,f)|0;break}case 40:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=59;if((s|0)==59?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ul(l,n,f)|0;if((k[f>>2]|0)<1){t=+p[n>>3];if(t>4294967295.0|t<0.0){k[f>>2]=412;break a}else{k[d>>2]=~~t>>>0;break a}}break}case 41:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=67;if((s|0)==67?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ql(l,d,f)|0;break}case 81:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=72;if((s|0)==72?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Rl(l,d,f)|0;break}case 42:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=77;if((s|0)==77?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Tl(l,d,f)|0;break}case 82:{i[l>>0]=0;if(!((e|0)!=0?(i[e>>0]=0,(k[f>>2]|0)>0):0))s=82;if((s|0)==82?(Go(a,c,h,f)|0)<=0:0)Hk(h,l,e,f)|0;Ul(l,d,f)|0;break}case 83:{Oo(a,c,d,e,f)|0;break}case 163:{Po(a,c,d,e,f)|0;break}default:k[f>>2]=410}while(0);q=k[f>>2]|0;r=g;return q|0}function Jo(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+160|0;g=f+71|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}i[h>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))m=4;if((m|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;i[c>>0]=0;fl(h,c,e)|0;l=k[e>>2]|0;r=f;return l|0}function Ko(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+160|0;g=f+71|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}i[h>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))m=4;if((m|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;Ql(h,c,e)|0;l=k[e>>2]|0;r=f;return l|0}function Lo(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+160|0;g=f+71|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}i[h>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))m=4;if((m|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;Sl(h,c,e)|0;l=k[e>>2]|0;r=f;return l|0}function Mo(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+160|0;g=f+71|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}i[h>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))m=4;if((m|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;Ul(h,c,e)|0;l=k[e>>2]|0;r=f;return l|0}function No(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+160|0;g=f+71|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}i[h>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))m=4;if((m|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;Rl(h,c,e)|0;l=k[e>>2]|0;r=f;return l|0}function Oo(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;f=r;r=r+240|0;g=f;h=f+156|0;j=f+85|0;l=f+4|0;m=k[e>>2]|0;if((m|0)>0){n=m;r=f;return n|0}i[j>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))o=4;if((o|0)==4?(Go(a,b,h,e)|0)<=0:0)Hk(h,j,d,e)|0;if((i[j>>0]|0)==40){i[j>>0]=32;d=j+(vx(j,761359)|0)|0;i[d>>0]=0;d=vx(j,761361)|0;i[j+d>>0]=0;Tl(j,c,e)|0;Tl(j+(d+1)|0,c+4|0,e)|0;n=k[e>>2]|0;r=f;return n|0}else{k[g>>2]=b;Xw(l,761308,g)|0;tk(l);tk(j);k[e>>2]=408;n=408;r=f;return n|0}return 0}function Po(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;f=r;r=r+240|0;g=f;h=f+156|0;j=f+85|0;l=f+4|0;m=k[e>>2]|0;if((m|0)>0){n=m;r=f;return n|0}i[j>>0]=0;if(!((d|0)!=0?(i[d>>0]=0,(k[e>>2]|0)>0):0))o=4;if((o|0)==4?(Go(a,b,h,e)|0)<=0:0)Hk(h,j,d,e)|0;if((i[j>>0]|0)==40){i[j>>0]=32;d=j+(vx(j,761359)|0)|0;i[d>>0]=0;d=vx(j,761361)|0;i[j+d>>0]=0;Ul(j,c,e)|0;Ul(j+(d+1)|0,c+8|0,e)|0;n=k[e>>2]|0;r=f;return n|0}else{k[g>>2]=b;Xw(l,761363,g)|0;tk(l);tk(j);k[e>>2]=409;n=409;r=f;return n|0}return 0}function Qo(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0;f=r;r=r+96|0;g=f;i[c>>0]=0;if(d)i[d>>0]=0;h=k[e>>2]|0;if((h|0)>0){j=h;r=f;return j|0}if((Go(a,b,g,e)|0)>0){j=k[e>>2]|0;r=f;return j|0}else{Hk(g,c,d,e)|0;j=k[e>>2]|0;r=f;return j|0}return 0}function Ro(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}if(!b){e=k[a>>2]|0;g=a+4|0;h=k[g>>2]|0;if((e|0)==(k[h+68>>2]|0)){j=h;l=e}else{Fk(a,e+1|0,0,d)|0;e=k[g>>2]|0;j=e;l=k[e+68>>2]|0}e=(k[j+88>>2]|0)+(l<<3)|0;l=k[e+4>>2]|0;g=j+112|0;k[g>>2]=k[e>>2];k[g+4>>2]=l;if(c)i[c>>0]=0}else if((b|0)>0){l=k[a>>2]|0;g=a+4|0;e=k[g>>2]|0;if((l|0)==(k[e+68>>2]|0)){m=e;n=l}else{Fk(a,l+1|0,0,d)|0;l=k[g>>2]|0;m=l;n=k[l+68>>2]|0}l=(k[m+88>>2]|0)+(n<<3)|0;n=(b*80|0)+-80|0;b=ky(k[l>>2]|0,k[l+4>>2]|0,n|0,((n|0)<0)<<31>>31|0)|0;n=m+112|0;k[n>>2]=b;k[n+4>>2]=L;Eo(a,c,d)|0}f=k[d>>2]|0;return f|0}function So(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;e=r;r=r+160|0;f=e;g=e+75|0;h=e+4|0;j=k[d>>2]|0;if((j|0)>0){l=j;r=e;return l|0}k[f>>2]=0;i[b>>0]=0;if((Eo(a,g,f)|0)>0){l=k[d>>2]|0;r=e;return l|0}if(!(Ax(g,761414,10)|0)){j=g;m=761425;n=j+10|0;do{i[j>>0]=i[m>>0]|0;j=j+1|0;m=m+1|0}while((j|0)<(n|0));Hk(g,h,c,f)|0;fl(h,b,f)|0;if(k[f>>2]|0)i[b>>0]=0}else{b=k[a>>2]|0;f=a+4|0;h=k[f>>2]|0;if((b|0)==(k[h+68>>2]|0))o=h;else{Fk(a,b+1|0,0,d)|0;o=k[f>>2]|0}f=o+112|0;o=f;b=ky(k[o>>2]|0,k[o+4>>2]|0,-80,-1)|0;o=f;k[o>>2]=b;k[o+4>>2]=L}l=k[d>>2]|0;r=e;return l|0}function To(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;f=r;r=r+240|0;g=f+144|0;h=f+73|0;j=f;l=k[e>>2]|0;if((l|0)>0){m=l;r=f;return m|0}k[c>>2]=0;i[h>>0]=0;l=(d|0)==0;if(!(!l?(i[d>>0]=0,(k[e>>2]|0)>0):0))n=4;if((n|0)==4?(Go(a,b,g,e)|0)<=0:0)Hk(g,h,d,e)|0;g=k[e>>2]|0;if((g|0)>0){m=g;r=f;return m|0}if(l)o=0;else o=71-(xx(d)|0)|0;if(!(i[h>>0]|0)){l=cy(1)|0;k[c>>2]=l;i[l>>0]=0;m=g;r=f;return m|0}g=cy((xx(h)|0)+1|0)|0;k[c>>2]=g;fl(h,g,e)|0;g=o;o=1;l=xx(k[c>>2]|0)|0;a:while(1){b=(g|0)<1;n=o;p=l;b:while(1){q=(p|0)==0;s=p+-1|0;t=n;while(1){if(!t)break a;if(q){n=0;p=0;continue b}if((i[(k[c>>2]|0)+s>>0]|0)==38){u=t;break}else t=0}So(a,h,j,e)|0;if(!(i[h>>0]|0)){v=0;w=p}else{i[(k[c>>2]|0)+s>>0]=0;t=xx(h)|0;q=fy(k[c>>2]|0,t+p|0)|0;k[c>>2]=q;qx(q,h)|0;v=u;w=s+t|0}if(b|(i[j>>0]|0)==0){n=v;p=w}else{x=v;y=w;break}}p=d+(xx(d)|0)|0;i[p>>0]=32;i[p+1>>0]=0;zx(d,j,g)|0;g=71-(xx(d)|0)|0;o=x;l=y}m=k[e>>2]|0;r=f;return m|0}function Uo(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+176|0;h=g;j=g+93|0;l=g+12|0;i[c>>0]=0;i[d>>0]=0;if(e)i[e>>0]=0;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}if((Ro(a,b,j,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}Ho(j,c,g+8|0,f)|0;if((Hk(j,d,e,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}if((Bk(c,f)|0)>0?(k[h>>2]=b,k[h+4>>2]=c,Xw(l,761437,h)|0,tk(l),((b|0)%36|0|0)==0):0)tk(761494);n=k[f>>2]|0;r=g;return n|0}function Vo(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0;h=r;r=r+320|0;j=h+4|0;l=h;m=h+241|0;n=h+8|0;o=h+160|0;p=h+89|0;q=h+16|0;s=k[g>>2]|0;if((s|0)>0){t=s;r=h;return t|0}k[f>>2]=0;s=c+-1+d|0;i[m>>0]=0;zx(m,b,74)|0;b=xx(m)|0;if(!b){t=k[g>>2]|0;r=h;return t|0}if((b|0)>0){d=0;do{u=m+d|0;v=(nv(i[u>>0]|0)|0)&255;i[u>>0]=v;d=d+1|0}while((d|0)!=(b|0))}if((k[g>>2]|0)>0){d=a+4|0;v=k[d>>2]|0;w=d;x=a;y=v;z=k[v+68>>2]|0;A=0}else{v=k[a>>2]|0;d=a+4|0;u=k[d>>2]|0;if((v|0)==(k[u+68>>2]|0)){B=u;C=v}else{Fk(a,v+1|0,0,g)|0;v=k[d>>2]|0;B=v;C=k[v+68>>2]|0}v=B+96|0;u=(k[B+88>>2]|0)+(C<<3)|0;D=jy(k[v>>2]|0,k[v+4>>2]|0,k[u>>2]|0,k[u+4>>2]|0)|0;u=wy(D|0,L|0,80,0)|0;w=d;x=a;y=B;z=C;A=u}u=k[x>>2]|0;if((u|0)==(z|0)){E=y;F=z}else{Fk(a,u+1|0,0,g)|0;u=k[w>>2]|0;E=u;F=k[u+68>>2]|0}u=(k[E+88>>2]|0)+(F<<3)|0;F=ky(k[u>>2]|0,k[u+4>>2]|0,160,0)|0;u=E+112|0;k[u>>2]=F;k[u+4>>2]=L;if((A|0)>=3){u=o+b|0;F=b+o|0;E=3;w=0;while(1){if((Eo(a,o,g)|0)>0){G=16;break}if((((Ax(m,o,b)|0)==0?(i[n>>0]=0,z=rx(o,61)|0,(z|0)!=0):0)?(zx(n,u,z-F|0)|0,k[j>>2]=0,(Ok(n,l,j)|0)<1):0)?(z=k[l>>2]|0,!((z|0)>(s|0)|(z|0)<(c|0))):0){Hk(o,p,q,g)|0;Rl(p,e+((k[l>>2]|0)-c<<3)|0,g)|0;z=(k[l>>2]|0)-c|0;if((z|0)>=(k[f>>2]|0))k[f>>2]=z+1;if((k[g>>2]|0)==204){k[g>>2]=0;H=1}else H=w}else H=w;if((E|0)<(A|0)){E=E+1|0;w=H}else{I=H;break}}if((G|0)==16){t=k[g>>2]|0;r=h;return t|0}if(I){I=k[g>>2]|0;if((I|0)>=1){t=I;r=h;return t|0}k[g>>2]=204;t=204;r=h;return t|0}}t=k[g>>2]|0;r=h;return t|0}function Wo(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0;g=r;r=r+304|0;h=g+221|0;j=g+150|0;l=g;m=g+75|0;n=g+4|0;k[l>>2]=0;o=k[f>>2]|0;if((o|0)>0){p=o;r=g;return p|0}Gk(774793,b,m,f)|0;i[j>>0]=0;if((Go(a,m,h,l)|0)<=0)Hk(h,j,0,l)|0;i[n>>0]=0;fl(j,n,l)|0;Xo(a,n,b,c,d,e,f)|0;p=k[f>>2]|0;r=g;return p|0}function Xo(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;h=r;r=r+112|0;j=h+16|0;l=h+8|0;m=h;n=h+24|0;o=h+28|0;p=k[g>>2]|0;if((p|0)>0){q=p;r=h;return q|0}p=(c|0)!=0;do if(p){s=k[a>>2]|0;t=a+4|0;if((s|0)!=(k[(k[t>>2]|0)+68>>2]|0))Fk(a,s+1|0,0,g)|0;if((c|0)>=1?(s=k[t>>2]|0,(k[s+928>>2]|0)>=(c|0)):0){t=k[s+960>>2]|0;s=c+-1|0;if(i[b>>0]|0){u=t+(s*160|0)|0;break}k[e>>2]=1;if((d|0)>0)k[f>>2]=k[t+(s*160|0)+88>>2];q=k[g>>2]|0;r=h;return q|0}k[g>>2]=302;q=302;r=h;return q|0}else u=0;while(0);k[e>>2]=0;c=rx(b,40)|0;k[n>>2]=c;if(!c){k[m>>2]=b;Xw(o,761539,m)|0;k[g>>2]=263;q=263;r=h;return q|0}else{v=c;w=1}while(1){c=v+1|0;k[n>>2]=c;m=jx(c,n,10)|0;c=k[e>>2]|0;if((c|0)<(d|0))k[f+(c<<2)>>2]=m;if((m|0)<0){x=17;break}a=ha(m,w)|0;k[e>>2]=c+1;c=k[n>>2]|0;v=rx(c,44)|0;k[n>>2]=v;if(!v){y=a;z=c;break}else w=a}if((x|0)==17){tk(761569);tk(b);k[g>>2]=263;q=263;r=h;return q|0}x=rx(z,41)|0;k[n>>2]=x;if(!x){k[l>>2]=b;Xw(o,761539,l)|0;k[g>>2]=263;q=263;r=h;return q|0}if((p?(k[u+80>>2]|0)>0:0)?(p=k[u+88>>2]|0,(p|0)!=(y|0)):0){k[j>>2]=p;k[j+4>>2]=y;Xw(o,774694,j)|0;tk(o);tk(b);k[g>>2]=263;q=263;r=h;return q|0}q=k[g>>2]|0;r=h;return q|0}function Yo(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0;g=r;r=r+304|0;h=g+221|0;j=g+150|0;l=g;m=g+75|0;n=g+4|0;k[l>>2]=0;o=k[f>>2]|0;if((o|0)>0){p=o;r=g;return p|0}Gk(774793,b,m,f)|0;i[j>>0]=0;if((Go(a,m,h,l)|0)<=0)Hk(h,j,0,l)|0;i[n>>0]=0;fl(j,n,l)|0;Zo(a,n,b,c,d,e,f)|0;p=k[f>>2]|0;r=g;return p|0}function Zo(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0,z=0,A=0,B=0,C=0;h=r;r=r+128|0;j=h+16|0;l=h+8|0;m=h;n=h+32|0;o=h+36|0;q=k[g>>2]|0;if((q|0)>0){s=q;r=h;return s|0}q=k[a>>2]|0;t=a+4|0;if((q|0)!=(k[(k[t>>2]|0)+68>>2]|0))Fk(a,q+1|0,0,g)|0;if((c|0)>=1?(q=k[t>>2]|0,(k[q+928>>2]|0)>=(c|0)):0){t=k[q+960>>2]|0;q=c+-1|0;if(!(i[b>>0]|0)){k[e>>2]=1;if((d|0)>0){c=t+(q*160|0)+88|0;a=k[c+4>>2]|0;u=f;k[u>>2]=k[c>>2];k[u+4>>2]=a}}else{k[e>>2]=0;a=rx(b,40)|0;k[n>>2]=a;if(!a){k[m>>2]=b;Xw(o,761616,m)|0;k[g>>2]=263;s=263;r=h;return s|0}else{v=a;w=1;x=0}while(1){a=v+1|0;k[n>>2]=a;y=+hx(a,n)+.1;a=~~y>>>0;m=+W(y)>=1.0?(y>0.0?~~+ia(+V(y/4294967296.0),4294967295.0)>>>0:~~+ga((y-+(~~y>>>0))/4294967296.0)>>>0):0;u=k[e>>2]|0;if((u|0)<(d|0)){c=f+(u<<3)|0;k[c>>2]=a;k[c+4>>2]=m}if((m|0)<0){z=15;break}c=yy(a|0,m|0,w|0,x|0)|0;m=L;k[e>>2]=u+1;u=k[n>>2]|0;v=rx(u,44)|0;k[n>>2]=v;if(!v){A=u;B=c;C=m;break}else{w=c;x=m}}if((z|0)==15){tk(761647);tk(b);k[g>>2]=263;s=263;r=h;return s|0}z=rx(A,41)|0;k[n>>2]=z;if(!z){k[l>>2]=b;Xw(o,761616,l)|0;k[g>>2]=263;s=263;r=h;return s|0}if((k[t+(q*160|0)+80>>2]|0)>0?(l=t+(q*160|0)+88|0,q=k[l>>2]|0,t=k[l+4>>2]|0,!((q|0)==(B|0)&(t|0)==(C|0))):0){p[j>>3]=+(q>>>0)+4294967296.0*+(t|0);p[j+8>>3]=+(B>>>0)+4294967296.0*+(C|0);Xw(o,761696,j)|0;tk(o);tk(b);k[g>>2]=263;s=263;r=h;return s|0}}s=k[g>>2]|0;r=h;return s|0}k[g>>2]=302;s=302;r=h;return s|0}function _o(a,b,c,d,e,f,g,h,j,l,m,n,o,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;var s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0;s=r;r=r+656|0;t=s+112|0;u=s+104|0;v=s+96|0;w=s+88|0;x=s+80|0;y=s+72|0;z=s+64|0;A=s+56|0;B=s+48|0;C=s+40|0;D=s+32|0;E=s+24|0;F=s+16|0;G=s+8|0;H=s+124|0;I=s+120|0;J=s+116|0;K=s;L=s+574|0;M=s+499|0;N=s+418|0;O=s+343|0;P=s+272|0;Q=s+199|0;R=s+128|0;S=k[q>>2]|0;if((S|0)>0){T=S;r=s;return T|0}S=k[a>>2]|0;U=a+4|0;if((S|0)!=(k[(k[U>>2]|0)+68>>2]|0))Fk(a,S+1|0,0,q)|0;S=(c|0)!=0;if(S)k[c>>2]=1;Uo(a,1,O,P,Q,q)|0;a:do if(k[(k[U>>2]|0)+68>>2]|0){if(tx(O,782702)|0){k[F>>2]=O;Xw(L,761854,F)|0;tk(L);k[q>>2]=225;T=225;r=s;return T|0}if((fl(P,R,q)|0)>0){tk(761806);tk(P);T=k[q>>2]|0;r=s;return T|0}if((i[P>>0]|0)==39){if(!(tx(R,771935)|0)){V=0;W=30;break}if(!(tx(R,761845)|0)){V=0;W=30;break}}k[E>>2]=P;Xw(L,761905,E)|0;tk(L);if(k[(k[U>>2]|0)+1080>>2]|0){uk(3,L);if((d|0)!=0?(El(a,d,q)|0,(k[q>>2]|0)>0):0){tk(761940);T=k[q>>2]|0;r=s;return T|0}if((e|0)!=0?(Fl(a,e,q)|0,(k[q>>2]|0)>0):0){tk(761987);T=k[q>>2]|0;r=s;return T|0}if((f|0)!=0?(Hl(a,b,f,q)|0,(k[q>>2]|0)>0):0){tk(762033);T=k[q>>2]|0;r=s;return T|0}else{X=9;Y=0}}else{V=1;W=30}}else{if(tx(O,782785)|0){k[G>>2]=O;Xw(L,761762,G)|0;tk(L);k[q>>2]=221;T=221;r=s;return T|0}switch(i[P>>0]|0){case 84:{V=0;W=30;break a;break}case 70:{if(!S){V=0;W=30;break a}k[c>>2]=0;V=0;W=30;break a;break}default:{k[q>>2]=220;T=220;r=s;return T|0}}}while(0);b:do if((W|0)==30){Uo(a,2,O,P,Q,q)|0;if(tx(O,779656)|0){k[D>>2]=O;Xw(L,762081,D)|0;tk(L);k[q>>2]=222;T=222;r=s;return T|0}if((Ok(P,I,q)|0)>0){k[C>>2]=P;Xw(L,762131,C)|0;tk(L);k[q>>2]=211;T=211;r=s;return T|0}c=k[I>>2]|0;switch(c|0){case 8:case 16:case 32:case 64:case -64:case -32:break;default:{k[B>>2]=P;Xw(L,762177,B)|0;tk(L);k[q>>2]=211;T=211;r=s;return T|0}}if(d)k[d>>2]=c;ap(a,3,782713,J,q)|0;c=k[q>>2]|0;if((c|0)==208){k[q>>2]=223;T=223;r=s;return T|0}S=k[J>>2]|0;if((c|0)==209|(S|0)>999){k[A>>2]=S;Xw(L,762214,A)|0;tk(L);k[q>>2]=212;T=212;r=s;return T|0}if(e)k[e>>2]=S;if((S|0)>0){S=(f|0)!=0;c=0;G=4;c:while(1){E=c;c=c+1|0;Gk(782713,c,M,q)|0;bp(a,E+4|0,M,K,q)|0;switch(k[q>>2]|0){case 208:{W=47;break c;break}case 209:{W=48;break c;break}default:{}}if(S&(E|0)<(b|0)){R=K;F=k[R+4>>2]|0;Z=f+(E<<3)|0;k[Z>>2]=k[R>>2];k[Z+4>>2]=F}F=G+1|0;if((c|0)>=(k[J>>2]|0)){X=F;Y=V;break b}else G=F}if((W|0)==47){k[q>>2]=224;T=224;r=s;return T|0}else if((W|0)==48){k[q>>2]=213;T=213;r=s;return T|0}}else{X=4;Y=V}}while(0);V=(l|0)!=0;if(V)p[l>>3]=1.0;J=(m|0)!=0;if(J)p[m>>3]=0.0;f=(g|0)!=0;if(f)k[g>>2]=0;K=(h|0)!=0;if(K)k[h>>2]=1;b=(j|0)!=0;if(b)k[j>>2]=0;M=(n|0)!=0;if(M){e=n;k[e>>2]=1234554321;k[e+4>>2]=0}k[o>>2]=0;e=k[q>>2]|0;A=X;while(1){do if((Ro(a,A,N,q)|0)>0)if((k[q>>2]|0)==203){k[q>>2]=e;_=e;$=1;break}else{tk(762237);aa=0;W=95;break}else{Ho(N,O,H,q)|0;if((Bk(O,q)|0)>0?(k[z>>2]=A,k[z+4>>2]=O,Xw(L,761437,z)|0,tk(L),((A|0)%36|0|0)==0):0)tk(761494);if(V&(tx(O,779536)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((ll(P,l,q)|0)<=0){aa=0;W=95;break}k[q>>2]=e;p[l>>3]=1.0;k[y>>2]=P;Xw(L,762288,y)|0;tk(L);aa=0;W=95;break}if(J&(tx(O,779507)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((ll(P,m,q)|0)<=0){aa=0;W=95;break}k[q>>2]=e;p[m>>3]=0.0;k[x>>2]=P;Xw(L,762339,x)|0;tk(L);aa=0;W=95;break}if(M&(tx(O,779663)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((ml(P,n,q)|0)<=0){aa=0;W=95;break}k[q>>2]=e;X=n;k[X>>2]=1234554321;k[X+4>>2]=0;k[w>>2]=P;Xw(L,762389,w)|0;tk(L);aa=0;W=95;break}if(f&(tx(O,782652)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((Ok(P,g,q)|0)<=0){aa=0;W=95;break}k[v>>2]=P;Xw(L,762441,v)|0;tk(L);aa=0;W=95;break}if(K&(tx(O,782726)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((Ok(P,h,q)|0)<=0){aa=0;W=95;break}k[u>>2]=P;Xw(L,762494,u)|0;tk(L);aa=0;W=95;break}if(b&(tx(O,782808)|0)==0){k[o>>2]=0;Hk(N,P,Q,q)|0;if((Ol(P,j,q)|0)<=0){aa=0;W=95;break}k[q>>2]=e;k[j>>2]=0;k[t>>2]=P;Xw(L,762547,t)|0;tk(L);aa=0;W=95;break}if(!(tx(O,768779)|0)){aa=1;W=95;break}if(!(i[N>>0]|0)){k[o>>2]=(k[o>>2]|0)+1;aa=0;W=95;break}else{k[o>>2]=0;aa=0;W=95;break}}while(0);if((W|0)==95){W=0;_=k[q>>2]|0;$=aa}if((_|0)>0)break;if(!$)A=A+1|0;else{ba=_;W=101;break}}if((W|0)==101){if(!Y){T=ba;r=s;return T|0}k[q>>2]=233;T=233;r=s;return T|0}if(!(k[(k[U>>2]|0)+68>>2]|0))tk(762599);else tk(762658);T=k[q>>2]|0;r=s;return T|0}function $o(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}if((cp(a,2,779656,762719,f)|0)==208){k[f>>2]=222;h=222;return h|0}if((k[f>>2]|0)==209){k[f>>2]=211;h=211;return h|0}if((cp(a,3,782713,762721,f)|0)==208){k[f>>2]=223;h=223;return h|0}if((k[f>>2]|0)==209){k[f>>2]=212;h=212;return h|0}if((bp(a,4,772833,b,f)|0)==208){k[f>>2]=224;h=224;return h|0}if((k[f>>2]|0)==209){h=0;return h|0}if((bp(a,5,790243,c,f)|0)==208){k[f>>2]=224;h=224;return h|0}if((k[f>>2]|0)==209){h=0;return h|0}if((bp(a,6,782652,d,f)|0)==208){k[f>>2]=228;h=228;return h|0}if((k[f>>2]|0)==209){k[f>>2]=214;h=214;return h|0}if((cp(a,7,782726,762723,f)|0)==208){k[f>>2]=229;h=229;return h|0}if((k[f>>2]|0)==209){k[f>>2]=215;h=215;return h|0}if((ap(a,8,782733,e,f)|0)==208){k[f>>2]=230;h=230;return h|0}a=k[f>>2]|0;if((a|0)!=209?(k[e>>2]|0)<=999:0){if((a|0)<=0){h=a;return h|0}tk(762725);h=k[f>>2]|0;return h|0}h=(a|0)==216&1;return h|0}function ap(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;f=r;r=r+336|0;g=f+16|0;h=f+8|0;j=f;l=f+249|0;m=f+178|0;n=f+24|0;o=k[e>>2]|0;if((o|0)>0){p=o;r=f;return p|0}i[l>>0]=0;i[m>>0]=0;do if((Uo(a,b,l,m,f+105|0,e)|0)<1){if(!(tx(l,c)|0)){Ok(m,d,e)|0;if((k[e>>2]|0)<=0?(k[d>>2]|0)>=0:0)break;else q=209}else q=208;k[e>>2]=q;k[j>>2]=b;Xw(n,762787,j)|0;tk(n);k[h>>2]=c;Xw(n,762848,h)|0;tk(n);k[g>>2]=l;k[g+4>>2]=m;Xw(n,762899,g)|0;tk(n)}while(0);p=k[e>>2]|0;r=f;return p|0}function bp(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;f=r;r=r+336|0;g=f+16|0;h=f+8|0;j=f;l=f+249|0;m=f+178|0;n=f+24|0;o=k[e>>2]|0;if((o|0)>0){p=o;r=f;return p|0}i[l>>0]=0;i[m>>0]=0;do if((Uo(a,b,l,m,f+105|0,e)|0)<1){if(!(tx(l,c)|0)){ml(m,d,e)|0;if((k[e>>2]|0)<=0?(k[d+4>>2]|0)>=0:0)break;else q=209}else q=208;k[e>>2]=q;k[j>>2]=b;Xw(n,762931,j)|0;tk(n);k[h>>2]=c;Xw(n,762848,h)|0;tk(n);k[g>>2]=l;k[g+4>>2]=m;Xw(n,762899,g)|0;tk(n)}while(0);p=k[e>>2]|0;r=f;return p|0}function cp(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;f=r;r=r+336|0;g=f+16|0;h=f+8|0;j=f;l=f+249|0;m=f+178|0;n=f+24|0;o=k[e>>2]|0;if((o|0)>0){p=o;r=f;return p|0}i[l>>0]=0;i[m>>0]=0;if((Uo(a,b,l,m,f+105|0,e)|0)<1){if(tx(l,c)|0)k[e>>2]=208;if(tx(d,m)|0)k[e>>2]=209;else q=7}else q=7;if((q|0)==7?(q=k[e>>2]|0,(q|0)<=0):0){p=q;r=f;return p|0}k[j>>2]=b;Xw(n,762994,j)|0;tk(n);k[h>>2]=c;k[h+4>>2]=d;Xw(n,763055,h)|0;tk(n);k[g>>2]=l;k[g+4>>2]=m;Xw(n,762899,g)|0;tk(n);p=k[e>>2]|0;r=f;return p|0}function dp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;d=r;r=r+16|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}f=k[a>>2]|0;h=a+4|0;j=k[h>>2]|0;if((f|0)!=(k[j+68>>2]|0)){Fk(a,f+1|0,0,c)|0;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}else l=k[h>>2]|0}else l=j;j=l+96|0;h=(k[l+88>>2]|0)+(k[l+68>>2]<<3)|0;l=jy(k[j>>2]|0,k[j+4>>2]|0,k[h>>2]|0,k[h+4>>2]|0)|0;h=wy(l|0,L|0,80,0)|0;l=((h|0)/36|0)*2880|0;h=l+2880|0;j=ey(l+2881|0,1)|0;k[b>>2]=j;if(!j){k[c>>2]=113;tk(763095);g=k[c>>2]|0;r=d;return g|0}else{cl(a,e,0,0,c)|0;j=e;au(a,k[j>>2]|0,k[j+4>>2]|0,0,c)|0;fu(a,h,((h|0)<0)<<31>>31,k[b>>2]|0,c)|0;i[(k[b>>2]|0)+h>>0]=0;g=k[c>>2]|0;r=d;return g|0}return 0}function ep(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;h=r;r=r+256|0;j=h+4|0;l=h;m=h+84|0;n=h+8|0;k[f>>2]=0;o=k[g>>2]|0;if((o|0)>0){p=o;r=h;return p|0}q=k[a>>2]|0;s=a+4|0;t=k[s>>2]|0;if((q|0)==(k[t+68>>2]|0)){u=t;v=q;w=o}else{Fk(a,q+1|0,0,g)|0;q=k[s>>2]|0;u=q;v=k[q+68>>2]|0;w=k[g>>2]|0}q=u+96|0;s=(k[u+88>>2]|0)+(v<<3)|0;v=jy(k[q>>2]|0,k[q+4>>2]|0,k[s>>2]|0,k[s+4>>2]|0)|0;s=wy(v|0,L|0,80,0)|0;if((w|0)>0){p=w;r=h;return p|0}w=ey((s*80|0)+81|0,1)|0;k[e>>2]=w;if(!w){k[g>>2]=113;tk(763095);p=k[g>>2]|0;r=h;return p|0}if((s|0)<1)x=w;else{v=(b|0)==0;b=(d|0)>0;q=w;w=1;while(1){Ro(a,w,m,g)|0;y=m+(xx(m)|0)|0;z=763153;A=y+81|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(A|0));i[n>>0]=0;zx(n,m,8)|0;a:do if(v)B=14;else switch(i[n>>0]|0){case 67:{if(!(tx(773862,n)|0)){C=q;break a}else{B=14;break a}break}case 72:{if(!(tx(773871,n)|0)){C=q;break a}else{B=14;break a}break}case 32:{if(!(tx(773880,n)|0)){C=q;break a}else{B=14;break a}break}default:{B=14;break a}}while(0);if((B|0)==14){B=0;b:do if(b){u=0;while(1){Wk(k[c+(u<<2)>>2]|0,n,0,j,l);if(k[j>>2]|0){D=u;break b}o=u+1|0;if((o|0)<(d|0))u=o;else{D=o;break}}}else D=0;while(0);if((D|0)==(d|0)){ux(q,m)|0;k[f>>2]=(k[f>>2]|0)+1;C=q+80|0}else C=q}if((w|0)==(s|0)){x=C;break}else{q=C;w=w+1|0}}}y=x;z=763234;A=y+81|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(A|0));k[f>>2]=(k[f>>2]|0)+1;i[x+80>>0]=0;x=fy(k[e>>2]|0,(k[f>>2]|0)*80|1)|0;k[e>>2]=x;p=k[g>>2]|0;r=h;return p|0}function fp(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0;h=r;r=r+16|0;i=h;j=k[g>>2]|0;if((j|0)>0){l=j;r=h;return l|0}do if(!(Cl(a,g)|0))ep(a,b,c,d,e,f,g)|0;else{if((Lh(i,777975,g)|0)>0){l=k[g>>2]|0;r=h;return l|0}j=(lt(a,k[i>>2]|0,g)|0)>0;m=k[i>>2]|0;if(!j){ep(m,b,c,d,e,f,g)|0;Fh(k[i>>2]|0,g)|0;break}ci(m,g)|0;l=k[g>>2]|0;r=h;return l|0}while(0);l=k[g>>2]|0;r=h;return l|0}function gp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;if(g){h=g;r=e;return h|0}g=Kl(a,f,d)|0;k[d>>2]=g;i=k[f>>2]|0;if(!i)j=g;else{g=Fk(a,i,e+4|0,d)|0;k[d>>2]=g;j=g}if(j)k[d>>2]=0;j=hp(a,b,c,d)|0;k[d>>2]=j;h=j;r=e;return h|0}function hp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;e=r;r=r+464|0;f=e+8|0;g=e;h=e+72|0;j=e+68|0;l=e+64|0;m=e+40|0;n=e+16|0;o=e+358|0;p=e+304|0;q=e+76|0;s=e+229|0;t=e+158|0;u=e+85|0;k[h>>2]=0;k[j>>2]=0;k[l>>2]=0;v=q;w=763315;x=v+9|0;do{i[v>>0]=i[w>>0]|0;v=v+1|0;w=w+1|0}while((v|0)<(x|0));k[m>>2]=o;k[n>>2]=p;k[m+4>>2]=o+17;k[n+4>>2]=p+9;k[m+8>>2]=o+34;k[n+8>>2]=p+18;k[m+12>>2]=o+51;k[n+12>>2]=p+27;k[m+16>>2]=o+68;k[n+16>>2]=p+36;k[m+20>>2]=o+85;k[n+20>>2]=p+45;p=ip(c,0,0,0,0,0,0,m,n,h,d)|0;k[d>>2]=p;p=k[h>>2]|0;h=Zj(a,0,0,p,m,n,0,0,0,0,d)|0;k[d>>2]=h;if(h){y=h;r=e;return y|0}bl(a,j)|0;Qo(a,782733,t,u,d)|0;Cq(a,790194,q,763324,d)|0;Eq(a,767107,0,0,763354,d)|0;if((b|0)!=0?(i[b>>0]|0)!=0:0)Cq(a,763387,b,763395,d)|0;a:do if((p|0)>0){b=k[d>>2]|0;q=0;while(1){if(b){z=1;break a}h=k[m+(q<<2)>>2]|0;if((px(h,763415)|0)!=0?(px(h,763431)|0)!=0:0){A=q+1|0;B=0}else{h=q+1|0;k[g>>2]=h;Xw(s,763446,g)|0;n=Jo(a,s,t,u,d)|0;k[d>>2]=n;k[f>>2]=h;Xw(s,763454,f)|0;n=Eq(a,s,0,0,763462,d)|0;k[d>>2]=n;A=h;B=n}if((A|0)<(p|0)){b=B;q=A}else{z=1;break}}}else z=1;while(0);while(1)if(!(Jl(a,-1,763315,z,d)|0))z=z+1|0;else{C=z;break}if((k[d>>2]|0)==301)k[d>>2]=0;Fk(a,k[j>>2]|0,l,d)|0;oq(a,767107,C,((C|0)<0)<<31>>31,763480,d)|0;y=k[d>>2]|0;r=e;return y|0}function ip(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=r;r=r+144|0;o=n+120|0;p=n+112|0;q=n+100|0;s=n;t=n+83|0;u=n+80|0;v=n+64|0;w=n+56|0;x=n+40|0;y=n+32|0;z=n+16|0;A=n+4|0;B=o;C=763482;D=B+16|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));i[p>>0]=i[763498]|0;i[p+1>>0]=i[763499]|0;i[p+2>>0]=i[763500]|0;B=q;C=763501;D=B+12|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));k[s>>2]=4272691;B=t;C=763431;D=B+15|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));i[u>>0]=i[763513]|0;i[u+1>>0]=i[763514]|0;i[u+2>>0]=i[763515]|0;B=v;C=763415;D=B+16|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));i[w>>0]=i[763513]|0;i[w+1>>0]=i[763514]|0;i[w+2>>0]=i[763515]|0;B=x;C=763516;D=B+16|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));i[y>>0]=i[763532]|0;i[y+1>>0]=i[763533]|0;i[y+2>>0]=i[763534]|0;B=z;C=763535;D=B+16|0;do{i[B>>0]=i[C>>0]|0;B=B+1|0;C=C+1|0}while((B|0)<(D|0));i[A>>0]=i[763551]|0;i[A+1>>0]=i[763552]|0;i[A+2>>0]=i[763553]|0;i[A+3>>0]=i[763554]|0;i[A+4>>0]=i[763555]|0;C=k[m>>2]|0;if(C){E=C;r=n;return E|0}switch(a|0){case 0:{if(!b){ux(k[h>>2]|0,o)|0;ux(k[j>>2]|0,p)|0;F=1}else F=0;if(!c){ux(k[h+(F<<2)>>2]|0,q)|0;ux(k[j+(F<<2)>>2]|0,s)|0;G=F+1|0}else G=F;if(!d){ux(k[h+(G<<2)>>2]|0,t)|0;ux(k[j+(G<<2)>>2]|0,u)|0;H=G+1|0}else H=G;if(!e){ux(k[h+(H<<2)>>2]|0,v)|0;ux(k[j+(H<<2)>>2]|0,w)|0;I=H+1|0}else I=H;if(!f){ux(k[h+(I<<2)>>2]|0,z)|0;ux(k[j+(I<<2)>>2]|0,A)|0;J=I+1|0}else J=I;if(!g){ux(k[h+(J<<2)>>2]|0,x)|0;ux(k[j+(J<<2)>>2]|0,y)|0;K=J+1|0}else K=J;break}case 1:{if(!b){ux(k[h>>2]|0,o)|0;ux(k[j>>2]|0,p)|0;L=1}else L=0;if(!c){ux(k[h+(L<<2)>>2]|0,q)|0;ux(k[j+(L<<2)>>2]|0,s)|0;M=L+1|0}else M=L;if(!d){ux(k[h+(M<<2)>>2]|0,t)|0;ux(k[j+(M<<2)>>2]|0,u)|0;K=M+1|0}else K=M;break}case 2:{if(!e){ux(k[h>>2]|0,v)|0;ux(k[j>>2]|0,w)|0;K=1}else K=0;break}case 3:{if(!b){ux(k[h>>2]|0,o)|0;ux(k[j>>2]|0,p)|0;N=1}else N=0;if(!c){ux(k[h+(N<<2)>>2]|0,q)|0;ux(k[j+(N<<2)>>2]|0,s)|0;O=N+1|0}else O=N;if(!d){ux(k[h+(O<<2)>>2]|0,t)|0;ux(k[j+(O<<2)>>2]|0,u)|0;P=O+1|0}else P=O;if(!e){ux(k[h+(P<<2)>>2]|0,v)|0;ux(k[j+(P<<2)>>2]|0,w)|0;K=P+1|0}else K=P;break}case 11:{if(!b){ux(k[h>>2]|0,o)|0;ux(k[j>>2]|0,p)|0;Q=1}else Q=0;if(!c){ux(k[h+(Q<<2)>>2]|0,q)|0;ux(k[j+(Q<<2)>>2]|0,s)|0;R=Q+1|0}else R=Q;if(!d){ux(k[h+(R<<2)>>2]|0,t)|0;ux(k[j+(R<<2)>>2]|0,u)|0;S=R+1|0}else S=R;if(!f){ux(k[h+(S<<2)>>2]|0,z)|0;ux(k[j+(S<<2)>>2]|0,A)|0;T=S+1|0}else T=S;if(!g){ux(k[h+(T<<2)>>2]|0,x)|0;ux(k[j+(T<<2)>>2]|0,y)|0;K=T+1|0}else K=T;break}case 12:{if(!e){ux(k[h>>2]|0,v)|0;ux(k[j>>2]|0,w)|0;U=1}else U=0;if(!f){ux(k[h+(U<<2)>>2]|0,z)|0;ux(k[j+(U<<2)>>2]|0,A)|0;V=U+1|0}else V=U;if(!g){ux(k[h+(V<<2)>>2]|0,x)|0;ux(k[j+(V<<2)>>2]|0,y)|0;K=V+1|0}else K=V;break}default:{k[m>>2]=347;tk(763556);K=0}}k[l>>2]=K;E=k[m>>2]|0;r=n;return E|0}function jp(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;j=r;r=r+144|0;l=j+73|0;m=k[i>>2]|0;if(m){n=m;r=j;return n|0}m=Jo(a,790194,l,j,i)|0;k[i>>2]=m;if((m|0)==202){k[i>>2]=340;tk(763617);o=k[i>>2]|0}else o=m;a:do if(!o){rp(l);if(px(l,763315)|0){k[i>>2]=340;n=340;r=j;return n|0}m=Tk(a,1,763482,b,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[b>>2]=0;p=k[i>>2]|0}else p=m;if(!p){m=Tk(a,1,763501,c,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[c>>2]=0;q=k[i>>2]|0}else q=m;if(!q){m=Tk(a,1,763431,d,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[d>>2]=0;s=k[i>>2]|0}else s=m;if(!s){m=Tk(a,1,763415,e,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[e>>2]=0;t=k[i>>2]|0}else t=m;if(!t){m=Tk(a,1,763535,f,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[f>>2]=0;u=k[i>>2]|0}else u=m;if(!u){m=Tk(a,1,763516,g,i)|0;k[i>>2]=m;if((m|0)==219){k[i>>2]=0;k[g>>2]=0;v=k[i>>2]|0}else v=m;if(!v){do if((k[b>>2]|0)!=0?(k[c>>2]|0)!=0:0){if(!(k[d>>2]|0))break;do if(k[e>>2]|0){if(!(k[f>>2]|0))break;if(!(k[g>>2]|0))break;k[h>>2]=0;break a}while(0);if(!(k[d>>2]|0))break;do if(k[f>>2]|0){if(!(k[g>>2]|0))break;k[h>>2]=11;break a}while(0);if(!(k[d>>2]|0))break;if(k[e>>2]|0){k[h>>2]=3;break a}if(!(k[d>>2]|0))break;k[h>>2]=1;break a}while(0);if(!(k[e>>2]|0)){k[i>>2]=340;n=340;r=j;return n|0}do if(k[f>>2]|0){if(!(k[g>>2]|0))break;k[h>>2]=12;break a}while(0);k[h>>2]=2}}}}}}}while(0);h=k[i>>2]|0;if((h|0)!=237){n=h;r=j;return n|0}k[i>>2]=340;tk(763664);n=k[i>>2]|0;r=j;return n|0}function kp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0;d=r;r=r+16160|0;e=d+16085|0;f=d+16012|0;g=d+4|0;h=d;i=d+8|0;k[h>>2]=0;j=k[c>>2]|0;if(j){l=j;r=d;return l|0}a:do switch(b|0){case 0:{j=Qo(a,790194,e,f,c)|0;k[c>>2]=j;if((j|0)==202){m=0;n=340}else{rp(e);if(px(e,763315)|0){k[c>>2]=340;tk(763726)}j=Ko(a,790243,h,f,c)|0;m=k[h>>2]|0;n=j}k[c>>2]=n;if((m|0)>0){j=n;o=m;while(1){if(j)break a;j=lp(a,o,1,c)|0;k[c>>2]=j;if((o|0)<=1)break;else o=o+-1|0}}break}case 3:{k[i>>2]=0;o=mp(a,i,0,0)|0;k[c>>2]=o;o=np(a,i,c)|0;k[c>>2]=o;o=k[i>>2]|0;if((o|0)>0){j=0;do{dy(k[i+4+(j<<2)>>2]|0);dy(k[i+8004+(j<<2)>>2]|0);j=j+1|0}while((j|0)<(o|0))}break}default:{k[c>>2]=347;tk(763773)}}while(0);i=op(a,0,c)|0;k[c>>2]=i;i=_j(a,g,c)|0;k[c>>2]=i;l=i;r=d;return l|0}function lp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;e=r;r=r+7616|0;f=e+32|0;g=e+24|0;h=e+16|0;j=e+8|0;l=e;m=e+7543|0;n=e+7470|0;o=e+64|0;p=e+60|0;q=e+56|0;s=e+52|0;t=e+48|0;u=e+44|0;v=e+6445|0;w=e+5420|0;x=e+4395|0;y=e+3370|0;z=e+3295|0;A=e+2270|0;B=e+40|0;C=e+2199|0;D=e+2118|0;E=e+1093|0;F=e+68|0;G=e+36|0;k[o>>2]=0;k[p>>2]=0;k[q>>2]=0;k[s>>2]=0;k[t>>2]=0;k[u>>2]=0;k[G>>2]=0;H=k[d>>2]|0;if(H){I=H;r=e;return I|0}rk(a,p,d)|0;a:do if((k[p>>2]|0)==1){H=pp(a,b,G,d)|0;k[d>>2]=H;H=rk(k[G>>2]|0,p,d)|0;k[d>>2]=H;switch(c|0){case 2:{if((bl(k[G>>2]|0,o)|0)==1){k[d>>2]=301;break a}H=Jo(k[G>>2]|0,790194,C,D,d)|0;k[d>>2]=H;if((H|0)==202){i[C>>0]=0;k[d>>2]=0}rp(C);if(k[d>>2]|0)break a;if(!(px(C,763315)|0)){H=k[G>>2]|0;J=Qo(H,790194,m,n,d)|0;k[d>>2]=J;if((J|0)==202){K=0;L=340}else{rp(m);if(px(m,763315)|0){k[d>>2]=340;tk(763726)}J=Ko(H,790243,s,n,d)|0;K=k[s>>2]|0;L=J}k[d>>2]=L;J=(L|0)==0;if((K|0)>0){H=J;M=K;while(1){if(!H)break a;N=lp(k[G>>2]|0,M,1,d)|0;k[d>>2]=N;O=(N|0)==0;if((M|0)>1){H=O;M=M+-1|0}else{P=O;break}}}else P=J;if(!P)break a}M=op(k[G>>2]|0,0,d)|0;k[d>>2]=M;if(M)break a;Vk(a,d)|0;if(!(k[p>>2]|0))break a;M=_j(k[G>>2]|0,o,d)|0;k[d>>2]=M;break a;break}case 1:{if((k[p>>2]|0)==1){M=Ko(a,767107,t,D,d)|0;k[d>>2]=M;ai(k[(k[(k[G>>2]|0)+4>>2]|0)+12>>2]|0,E,d)|0;M=a+4|0;ai(k[(k[M>>2]|0)+12>>2]|0,F,d)|0;if((k[(k[G>>2]|0)+4>>2]|0)!=(k[M>>2]|0)?(Ax(E,F,1025)|0)!=0:0)k[t>>2]=0-(k[t>>2]|0);M=wp(a,v,w,0,0,0,d)|0;k[d>>2]=M;if(M)break a;M=yp(y,d)|0;k[d>>2]=M;switch(i[v>>0]|0){case 47:case 0:break;default:if(!(up(v)|0)){ux(x,y)|0;M=x+(xx(x)|0)|0;i[M>>0]=47;i[M+1>>0]=0;qx(x,v)|0;zp(x,v,d)|0}}switch(i[w>>0]|0){case 47:case 0:break;default:if(!(up(w)|0)){ux(x,y)|0;M=x+(xx(x)|0)|0;i[M>>0]=47;i[M+1>>0]=0;qx(x,w)|0;zp(x,w,d)|0}}M=sp(k[G>>2]|0,q,d)|0;k[d>>2]=M;M=Ro(k[G>>2]|0,0,D,d)|0;k[d>>2]=M;H=k[q>>2]|0;if((H|0)>=1){b:do if(!M){O=0;N=1;while(1){k[l>>2]=N;Xw(z,763868,l)|0;Q=Ko(k[G>>2]|0,z,u,D,d)|0;k[d>>2]=Q;c:do if(!Q){R=k[u>>2]|0;S=(R|0)==(k[t>>2]|0);if(!((R|0)>0&S))if((R|0)<0&S){k[j>>2]=N;Xw(z,763876,j)|0;S=To(k[G>>2]|0,z,B,D,d)|0;k[d>>2]=S;if(!S){ux(A,k[B>>2]|0)|0;dy(k[B>>2]|0);T=k[d>>2]|0}else T=S;switch(T|0){case 202:{k[h>>2]=N;k[h+4>>2]=N;Xw(D,763884,h)|0;tk(D);k[d>>2]=0;U=O;break c;break}case 0:break;default:{U=O;break c}}rp(A);S=i[A>>0]|0;if(S<<24>>24!=0?!(S<<24>>24==47|(up(A)|0)!=0):0){S=qk(k[G>>2]|0,x,d)|0;k[d>>2]=S;S=Cx(x,47)|0;if(S)i[S>>0]=0;S=x+(xx(x)|0)|0;i[S>>0]=47;i[S+1>>0]=0;qx(x,A)|0;S=zp(x,A,d)|0;k[d>>2]=S}if((tx(A,v)|0)!=0?(tx(A,w)|0)!=0:0){U=O;break}U=N}else U=O;else U=N}else U=O;while(0);if((N|0)>=(H|0)){V=U;break b}if(!(k[d>>2]|U)){O=U;N=N+1|0}else{V=U;break}}}else V=0;while(0);if(V){k[g>>2]=V;Xw(z,763868,g)|0;H=Aq(k[G>>2]|0,z,d)|0;k[d>>2]=H;k[f>>2]=V;Xw(z,763876,f)|0;Aq(k[G>>2]|0,z,d)|0;k[d>>2]=0;H=sp(k[G>>2]|0,q,d)|0;k[d>>2]=H}}}H=Pj(a,b,((b|0)<0)<<31>>31,1,0,d)|0;k[d>>2]=H;break a;break}default:{k[d>>2]=347;tk(763913);break a}}}else{tk(763830);k[d>>2]=350}while(0);b=k[G>>2]|0;if(b)Fh(b,d)|0;I=k[d>>2]|0;r=e;return I|0}function mp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0;e=r;r=r+2064|0;f=e+4|0;g=e;h=e+1033|0;i=e+8|0;k[g>>2]=0;bl(a,f)|0;j=qk(a,h,g)|0;k[g>>2]=j;j=ai(h,i,g)|0;k[g>>2]=j;j=k[b>>2]|0;a:do if((j|0)>0){h=k[f>>2]|0;a=0;while(1){if((k[b+4004+(a<<2)>>2]|0)==(h|0)?(tx(k[b+4+(a<<2)>>2]|0,i)|0)==0:0){l=a;break a}m=a+1|0;if((m|0)<(j|0))a=m;else{l=m;break}}}else l=0;while(0);if((l|0)!=(j|0)){k[g>>2]=346;if(c)k[c>>2]=k[b+12004+(l<<2)>>2];if(!d){n=k[g>>2]|0;r=e;return n|0}ux(d,k[b+8004+(l<<2)>>2]|0)|0;n=k[g>>2]|0;r=e;return n|0}if((j|0)==1e3){k[g>>2]=345;n=k[g>>2]|0;r=e;return n|0}l=cy(1025)|0;k[b+4+(j<<2)>>2]=l;if(!l){k[g>>2]=113;n=k[g>>2]|0;r=e;return n|0}d=cy(1025)|0;c=b+8004+(j<<2)|0;k[c>>2]=d;if(!d){k[g>>2]=113;dy(l);n=k[g>>2]|0;r=e;return n|0}else{d=k[f>>2]|0;k[b+4004+(j<<2)>>2]=d;k[b+12004+(j<<2)>>2]=d;ux(l,i)|0;ux(k[c>>2]|0,i)|0;k[b>>2]=(k[b>>2]|0)+1;n=k[g>>2]|0;r=e;return n|0}return 0}function np(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+304|0;e=d+229|0;f=d+156|0;g=d+8|0;h=d+4|0;j=d+85|0;l=d+12|0;m=d;k[h>>2]=0;k[m>>2]=0;n=k[c>>2]|0;if(n){o=n;r=d;return o|0}n=Qo(a,790194,e,f,c)|0;k[c>>2]=n;if((n|0)==202){p=0;q=340}else{rp(e);if(px(e,763315)|0){k[c>>2]=340;tk(763726)}e=Ko(a,790243,h,f,c)|0;p=k[h>>2]|0;q=e}k[c>>2]=q;if((p|0)>0&(q|0)==0)s=p;else{o=q;r=d;return o|0}while(1){q=pp(a,s,m,c)|0;k[c>>2]=q;a:do switch(q|0){case 342:{k[c>>2]=0;break}case 0:{p=mp(k[m>>2]|0,b,0,0)|0;k[c>>2]=p;switch(p|0){case 346:{k[c>>2]=0;Fh(k[m>>2]|0,c)|0;break a;break}case 0:break;default:break a}p=Jo(k[m>>2]|0,790194,j,l,c)|0;k[c>>2]=p;if((p|0)==202){k[c>>2]=0;i[j>>0]=0}rp(j);if(!(k[c>>2]|0)){if(!(px(j,763315)|0)){p=np(k[m>>2]|0,b,c)|0;k[c>>2]=p}p=(bl(k[m>>2]|0,g)|0)==1;e=k[m>>2]|0;if(p)t=op(e,1,c)|0;else{p=op(e,0,c)|0;k[c>>2]=p;t=_j(k[m>>2]|0,g,c)|0}k[c>>2]=t;Fh(k[m>>2]|0,c)|0}break}default:{}}while(0);if((s|0)<=1)break;q=k[c>>2]|0;if(!q)s=s+-1|0;else{o=q;u=23;break}}if((u|0)==23){r=d;return o|0}o=k[c>>2]|0;r=d;return o|0}function op(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0;d=r;r=r+2416|0;e=d+96|0;f=d+88|0;g=d+80|0;h=d+72|0;j=d+120|0;l=d+116|0;m=d+112|0;n=d+108|0;o=d+104|0;p=d+1376|0;q=d+351|0;s=d;t=d+280|0;u=d+205|0;v=d+124|0;w=d+100|0;k[j>>2]=0;k[m>>2]=0;k[n>>2]=0;k[o>>2]=0;k[w>>2]=0;x=k[c>>2]|0;if(x){y=x;r=d;return y|0}x=Jo(a,782702,s,v,c)|0;k[c>>2]=x;if((x|0)==202){x=s;k[x>>2]=1296650832;k[x+4>>2]=5853761;k[c>>2]=0}rp(s);x=Ko(a,767107,n,v,c)|0;k[c>>2]=x;if((x|0)==202){k[n>>2]=1;k[c>>2]=0}x=Jo(a,790194,t,v,c)|0;k[c>>2]=x;if((x|0)==202){i[t>>0]=0;k[c>>2]=0}rp(t);bl(a,j)|0;x=wp(a,p,q,0,0,0,c)|0;k[c>>2]=x;a:do if(!x){z=sp(a,m,c)|0;k[c>>2]=z;A=k[m>>2]|0;B=(A|0)<1;if(B)C=z;else{if(!z)D=1;else break;while(1){z=tp(a,D,w,c)|0;k[c>>2]=z;do if(!z){rk(k[w>>2]|0,l,c)|0;if((k[l>>2]|0)!=1){k[g>>2]=D;Xw(v,764012,g)|0;tk(v);break}k[o>>2]=0;if(!(i[p>>0]|0))E=k[c>>2]|0;else{F=Bp(k[w>>2]|0,s,t,k[n>>2]|0,k[j>>2]|0,p,o,c)|0;k[c>>2]=F;E=F}if((E|0)!=342|(i[q>>0]|0)==0)G=E;else{k[c>>2]=0;F=Bp(k[w>>2]|0,s,t,k[n>>2]|0,k[j>>2]|0,q,o,c)|0;k[c>>2]=F;G=F}if(!G){F=k[o>>2]|0;H=Pj(k[w>>2]|0,F,((F|0)<0)<<31>>31,1,0,c)|0;k[c>>2]=H;I=H}else I=G;if((I|0)==342)tk(764055);k[c>>2]=0;H=k[w>>2]|0;if(H){Fh(H,c)|0;k[w>>2]=0}}else{k[c>>2]=0;k[h>>2]=D;Xw(v,763970,h)|0;tk(v)}while(0);if((D|0)>=(A|0))break;if(!(k[c>>2]|0))D=D+1|0;else break a}C=k[c>>2]|0}if((b|0)!=0&(C|0)==0){rk(a,l,c)|0;if(!(k[l>>2]|0)){tk(764108);break}if(!B){z=k[c>>2]|0;H=1;while(1){if(z)break a;k[f>>2]=H;Xw(u,763868,f)|0;Aq(a,u,c)|0;k[e>>2]=H;Xw(u,763876,e)|0;Aq(a,u,c)|0;F=k[c>>2]|0;if((F|0)==202){k[c>>2]=0;J=0}else J=F;if((H|0)<(A|0)){z=J;H=H+1|0}else break}}}}while(0);J=k[w>>2]|0;if(J)Fh(J,c)|0;y=k[c>>2]|0;r=d;return y|0}function pp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0;e=r;r=r+6512|0;f=e;g=e+48|0;h=e+44|0;j=e+40|0;l=e+36|0;m=e+32|0;n=e+28|0;o=e+24|0;p=e+20|0;q=e+16|0;s=e+12|0;t=e+8|0;u=e+6426|0;v=e+6355|0;w=e+6284|0;x=e+5259|0;y=e+4234|0;z=e+3209|0;A=e+2184|0;B=e+1159|0;C=e+134|0;D=e+53|0;E=e+52|0;F=e+4|0;k[s>>2]=0;k[t>>2]=0;i[E>>0]=0;G=k[d>>2]|0;if(G){H=G;r=e;return H|0}G=jp(a,g,h,j,l,m,n,o,d)|0;k[d>>2]=G;if(!G){k[F>>2]=u;I=k[g>>2]|0;do if(I){g=In(a,I,b,((b|0)<0)<<31>>31,1,0,1,0,E,F,q,d)|0;k[d>>2]=g;if(!(px(u,764159)|0)){k[p>>2]=0;J=g;break}if(!(px(u,771935)|0)){k[p>>2]=0;J=g;break}if(!(px(u,773513)|0)){k[p>>2]=1;J=g;break}if(!(px(u,772762)|0)){k[p>>2]=2;J=g;break}else{k[p>>2]=-1;J=g;break}}else J=0;while(0);k[F>>2]=v;I=k[h>>2]|0;if(!I)K=J;else{J=In(a,I,b,((b|0)<0)<<31>>31,1,0,1,0,E,F,q,d)|0;k[d>>2]=J;K=J}J=k[j>>2]|0;if(!J)L=K;else{K=$m(a,J,b,((b|0)<0)<<31>>31,1,0,1,0,0,t,q,d)|0;k[d>>2]=K;L=K}K=k[l>>2]|0;if(!K)M=L;else{L=$m(a,K,b,((b|0)<0)<<31>>31,1,0,1,0,0,s,q,d)|0;k[d>>2]=L;M=L}k[F>>2]=z;L=k[m>>2]|0;if(!L)N=M;else{M=In(a,L,b,((b|0)<0)<<31>>31,1,0,1,0,E,F,q,d)|0;k[d>>2]=M;N=M}k[F>>2]=w;M=k[n>>2]|0;if(!M)O=N;else{N=In(a,M,b,((b|0)<0)<<31>>31,1,0,1,0,E,F,q,d)|0;k[d>>2]=N;O=N}a:do if(!O){N=k[o>>2]|0;b:do switch(N|0){case 3:case 1:case 2:{q=Sh(a,c,d)|0;k[d>>2]=q;P=q;Q=55;break}case 0:case 12:case 11:{if(!(i[z>>0]|0)){q=Sh(a,c,d)|0;k[d>>2]=q;P=q;Q=55;break b}if(px(w,764167)|0){k[d>>2]=104;k[f>>2]=w;Xw(D,764171,f)|0;tk(D);Q=65;break a}if(up(z)|0){tk(764225);q=Ih(5,c,z,1,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0;tk(764271);q=Ih(5,c,z,0,d)|0;k[d>>2]=q;P=q;Q=55;break b}tk(764310);if((i[z>>0]|0)==47){tk(764346);q=vp(z,A,d)|0;k[d>>2]=q;tk(764390);q=Ih(5,c,A,1,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0;tk(764271);q=Ih(5,c,A,0,d)|0;k[d>>2]=q;P=q;Q=55;break b}tk(764434);q=wp(a,x,y,0,0,0,d)|0;k[d>>2]=q;q=i[x>>0]|0;if(q<<24>>24){if(!(q<<24>>24==47|(up(x)|0)!=0)){yp(C,d)|0;q=C+(xx(C)|0)|0;i[q>>0]=47;i[q+1>>0]=0;qx(C,x)|0;ux(x,C)|0}q=xp(x,z,A,d)|0;k[d>>2]=q;if(q){Q=65;break a}if(!(up(A)|0)){q=vp(A,B,d)|0;k[d>>2]=q;ux(A,B)|0}q=Ih(5,c,A,1,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0;tk(764483);q=Ih(5,c,A,0,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0}q=i[y>>0]|0;if(q<<24>>24){if(!(q<<24>>24==47|(up(y)|0)!=0)){yp(C,d)|0;q=C+(xx(C)|0)|0;i[q>>0]=47;i[q+1>>0]=0;qx(C,y)|0;ux(y,C)|0}q=xp(y,z,A,d)|0;k[d>>2]=q;if(q){Q=65;break a}if(!(up(A)|0)){q=vp(A,B,d)|0;k[d>>2]=q;ux(A,B)|0}q=Ih(5,c,A,1,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0;tk(764483);q=Ih(5,c,A,0,d)|0;k[d>>2]=q;if(!q)break b;k[d>>2]=0}tk(764525);k[d>>2]=342;Q=65;break a;break}default:{Q=65;break a}}while(0);if((Q|0)==55?(P|0)!=0:0){Q=65;break}switch(N|0){case 12:case 2:{q=Fk(k[c>>2]|0,k[s>>2]|0,p,d)|0;k[d>>2]=q;R=q;break a;break}case 11:case 1:{q=Jl(k[c>>2]|0,k[p>>2]|0,v,k[t>>2]|0,d)|0;k[d>>2]=q;if((q|0)!=301){R=q;break a}k[d>>2]=342;tk(764567);Q=65;break a;break}case 0:case 3:{if((i[u>>0]|0)!=0?(q=k[t>>2]|0,(i[v>>0]|0)!=0&(q|0)>0):0){F=Jl(k[c>>2]|0,k[p>>2]|0,v,q,d)|0;k[d>>2]=F;if((F|0)!=301){R=F;break a}k[d>>2]=342;tk(764567);Q=65;break a}F=Fk(k[c>>2]|0,k[s>>2]|0,p,d)|0;q=(F|0)==107?342:F;k[d>>2]=q;R=q;break a;break}default:{Q=65;break a}}}else Q=65;while(0);if((Q|0)==65)R=k[d>>2]|0;if(!R){H=0;r=e;return H|0}else S=R}else S=G;G=k[c>>2]|0;if(!G){H=S;r=e;return H|0}Fh(G,d)|0;H=k[d>>2]|0;r=e;return H|0}function qp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0;e=r;r=r+10976|0;f=e+104|0;g=e+96|0;h=e+88|0;j=e+80|0;l=e+72|0;m=e+9941|0;n=e+8916|0;o=e+200|0;p=e+196|0;q=e+192|0;s=e+188|0;t=e+184|0;u=e+180|0;v=e+176|0;w=e+172|0;x=e+168|0;y=e+164|0;z=e+160|0;A=e+156|0;B=e+152|0;C=e+148|0;D=e+144|0;E=e+140|0;F=e+136|0;G=e+132|0;H=e+128|0;I=e+124|0;J=e+8845|0;K=e+8774|0;L=e+7749|0;M=e+6724|0;N=e+5699|0;O=e+120|0;P=e;Q=e+5628|0;R=e+116|0;S=e+5557|0;T=e+5486|0;U=e+4461|0;V=e+3436|0;W=e+2411|0;X=e+1386|0;Y=e+361|0;Z=e+112|0;_=e+286|0;$=e+205|0;aa=e+204|0;ba=e+108|0;k[v>>2]=0;k[w>>2]=0;k[x>>2]=0;k[y>>2]=6;k[D>>2]=0;k[E>>2]=0;k[F>>2]=0;k[G>>2]=0;k[H>>2]=0;k[I>>2]=0;k[R>>2]=5001813;i[aa>>0]=0;k[ba>>2]=0;ca=k[d>>2]|0;if(ca){da=ca;r=e;return da|0}rk(a,C,d)|0;a:do if((k[C>>2]|0)==1){if(!b){ca=Sh(a,ba,d)|0;k[d>>2]=ca;ca=Fk(k[ba>>2]|0,c,x,d)|0;k[d>>2]=ca;if(ca){ea=0;fa=132;break}ga=k[ba>>2]|0}else{k[ba>>2]=b;ga=b}ca=Jo(ga,782702,P,$,d)|0;k[d>>2]=ca;if((ca|0)==202){ca=P;k[ca>>2]=1296650832;k[ca+4>>2]=5853761;k[d>>2]=0}rp(P);ca=Ko(k[ba>>2]|0,767107,D,$,d)|0;k[d>>2]=ca;if((ca|0)==202){k[D>>2]=1;k[d>>2]=0}ca=Jo(k[ba>>2]|0,790194,Q,$,d)|0;k[d>>2]=ca;if((ca|0)==202){i[Q>>0]=0;k[d>>2]=0}rp(Q);bl(k[ba>>2]|0,v)|0;ca=wp(k[ba>>2]|0,L,M,J,K,A,d)|0;k[d>>2]=ca;if(!(i[L>>0]|0)){ux(L,M)|0;ux(J,K)|0}ca=wp(a,U,V,S,T,B,d)|0;k[d>>2]=ca;if(!ca){if(!(k[B>>2]|0)){tk(763830);k[d>>2]=349;ha=0;fa=130;break}if((px(S,764609)|0)!=0?(px(J,764609)|0)!=0:0)i[Y>>0]=0;else{ca=yp(Y,d)|0;k[d>>2]=ca;if(!(px(J,764609)|0)){if((i[L>>0]|0)==47)ux(M,L)|0;else{ux(M,Y)|0;ca=M+(xx(M)|0)|0;i[ca>>0]=47;i[ca+1>>0]=0;qx(M,L)|0}ca=zp(M,L,d)|0;k[d>>2]=ca}if(!(px(S,764609)|0)){if((i[U>>0]|0)==47)ux(V,U)|0;else{ux(V,Y)|0;ca=V+(xx(V)|0)|0;i[ca>>0]=47;i[ca+1>>0]=0;qx(V,U)|0}ca=zp(V,U,d)|0;k[d>>2]=ca}if((px(S,764609)|0)==0?(px(J,764609)|0)==0:0){Ap(L,U,V,d)|0;Ap(U,L,M,d)|0;ux(L,M)|0;ux(U,V)|0}}ca=Ko(a,767107,E,$,d)|0;k[d>>2]=ca;ai(k[(k[(k[ba>>2]|0)+4>>2]|0)+12>>2]|0,W,d)|0;ca=a+4|0;ai(k[(k[ca>>2]|0)+12>>2]|0,X,d)|0;if((k[(k[ba>>2]|0)+4>>2]|0)!=(k[ca>>2]|0)?(Ax(W,X,1025)|0)!=0:0)k[E>>2]=0-(k[E>>2]|0);ia=k[d>>2]|0;if(!ia){ja=Qo(a,790194,m,n,d)|0;k[d>>2]=ja;if((ja|0)==202)ka=340;else{rp(m);if(px(m,763315)|0){k[d>>2]=340;tk(763726)}ka=Ko(a,790243,G,n,d)|0}k[d>>2]=ka;la=ka}else la=ia;k[d>>2]=la;ia=Bp(a,P,Q,k[D>>2]|0,k[v>>2]|0,L,F,d)|0;k[d>>2]=ia;b:do switch(ia|0){case 342:{k[d>>2]=0;ja=k[G>>2]|0;ma=Nj(a,ja,((ja|0)<0)<<31>>31,1,0,d)|0;k[d>>2]=ma;k[G>>2]=(k[G>>2]|0)+1;ma=jp(a,o,p,q,s,t,u,w,d)|0;k[d>>2]=ma;k[Z>>2]=P;ma=k[o>>2]|0;if(ma){ja=k[G>>2]|0;Dr(a,ma,ja,((ja|0)<0)<<31>>31,1,0,1,0,Z,d)|0}k[Z>>2]=Q;ja=k[p>>2]|0;do if(ja){ma=k[G>>2]|0;na=((ma|0)<0)<<31>>31;if(!(i[Q>>0]|0)){Rq(a,ja,ma,na,1,0,1,0,aa,d)|0;break}else{Dr(a,ja,ma,na,1,0,1,0,Z,d)|0;break}}while(0);ja=k[q>>2]|0;if(ja){na=k[G>>2]|0;jr(a,ja,na,((na|0)<0)<<31>>31,1,0,1,0,D,d)|0}na=k[s>>2]|0;if(na){ja=k[G>>2]|0;tr(a,na,ja,((ja|0)<0)<<31>>31,1,0,1,0,v,d)|0}k[Z>>2]=L;c:do if(k[t>>2]|0){ai(k[(k[(k[ba>>2]|0)+4>>2]|0)+12>>2]|0,W,d)|0;ai(k[(k[ca>>2]|0)+12>>2]|0,X,d)|0;do if((k[(k[ba>>2]|0)+4>>2]|0)!=(k[ca>>2]|0)){if(!(Ax(W,X,1025)|0))break;ja=k[G>>2]|0;Dr(a,k[t>>2]|0,ja,((ja|0)<0)<<31>>31,1,0,1,0,Z,d)|0;break c}while(0);ja=k[G>>2]|0;Rq(a,k[t>>2]|0,ja,((ja|0)<0)<<31>>31,1,0,1,0,aa,d)|0}while(0);k[Z>>2]=R;if(!(k[u>>2]|0))oa=0;else{ai(k[(k[(k[ba>>2]|0)+4>>2]|0)+12>>2]|0,W,d)|0;ai(k[(k[ca>>2]|0)+12>>2]|0,X,d)|0;do if((k[(k[ba>>2]|0)+4>>2]|0)!=(k[ca>>2]|0)){if(!(Ax(W,X,1025)|0))break;ja=k[G>>2]|0;Dr(a,k[u>>2]|0,ja,((ja|0)<0)<<31>>31,1,0,1,0,Z,d)|0;oa=0;break b}while(0);ja=k[G>>2]|0;Rq(a,k[u>>2]|0,ja,((ja|0)<0)<<31>>31,1,0,1,0,aa,d)|0;oa=0}break}case 0:{tk(764617);oa=341;break}default:{ha=0;fa=130;break a}}while(0);if(!(k[d>>2]|0)){rk(k[ba>>2]|0,C,d)|0;if((k[A>>2]|0)==0|(k[C>>2]|0)!=1){tk(764682);tk(L);ha=oa;fa=130;break}ia=sp(k[ba>>2]|0,H,d)|0;k[d>>2]=ia;k[z>>2]=1;ja=k[H>>2]|0;d:do if((ja|0)>=1?(ia|0)==0:0){na=ja;while(1){k[l>>2]=na;Xw(_,763868,l)|0;ma=Ko(k[ba>>2]|0,_,I,$,d)|0;k[d>>2]=ma;ma=k[I>>2]|0;if((ma|0)==(k[E>>2]|0)){if((ma|0)>=0){fa=73;break}k[j>>2]=na;Xw(_,763876,j)|0;ma=To(b,_,O,$,d)|0;k[d>>2]=ma;if(!ma){ux(N,k[O>>2]|0)|0;dy(k[O>>2]|0)}if(!(i[Y>>0]|0))pa=na;else{do if(!(up(N)|0)){ma=k[d>>2]|0;if((ma|0)<=0){qa=0;ra=0;e:while(1){sa=i[N+qa>>0]|0;ta=qa;f:while(1){switch(sa<<24>>24){case 0:{ua=ra;break e;break}case 47:break;default:{va=sa;wa=ta;fa=81;break f}}xa=ta+1|0;if((i[N+xa>>0]|0)==47){sa=47;ta=xa}else{ya=47;za=xa;break}}if((fa|0)==81){fa=0;ya=va;za=wa+1|0}i[m+ra>>0]=ya;qa=za;ra=ra+1|0}i[m+ua>>0]=0;if(!ma){ra=i[m>>0]|0;if(!(ra<<24>>24))Aa=V;else{qa=ra;ra=m;ta=V;while(1){sa=qa&255;do if((qa&255)>31&qa<<24>>24>-1){if(!(i[764735+(sa+-32)>>0]|0)){fa=88;break}i[ta>>0]=qa;Ba=ta+1|0}else fa=88;while(0);if((fa|0)==88){fa=0;i[ta>>0]=37;i[ta+1>>0]=i[764831+(sa>>>4)>>0]|0;i[ta+2>>0]=i[764831+(sa&15)>>0]|0;Ba=ta+3|0}ra=ra+1|0;qa=i[ra>>0]|0;if(!(qa<<24>>24)){Aa=Ba;break}else ta=Ba}}i[Aa>>0]=0;Ca=k[d>>2]|0}else Ca=ma;k[d>>2]=Ca}if((i[V>>0]|0)==47)break;ux(n,Y)|0;ta=n+(xx(n)|0)|0;i[ta>>0]=47;i[ta+1>>0]=0;qx(n,V)|0;zp(n,N,d)|0}while(0);do if(!(up(U)|0)){ta=k[d>>2]|0;if((ta|0)<=0){qa=0;ra=0;g:while(1){xa=i[U+qa>>0]|0;Da=qa;h:while(1){switch(xa<<24>>24){case 0:{Ea=ra;break g;break}case 47:break;default:{Fa=xa;Ga=Da;fa=99;break h}}Ha=Da+1|0;if((i[U+Ha>>0]|0)==47){xa=47;Da=Ha}else{Ia=47;Ja=Ha;break}}if((fa|0)==99){fa=0;Ia=Fa;Ja=Ga+1|0}i[m+ra>>0]=Ia;qa=Ja;ra=ra+1|0}i[m+Ea>>0]=0;if(!ta){ra=i[m>>0]|0;if(!(ra<<24>>24))Ka=V;else{qa=ra;ra=m;ma=V;while(1){Da=qa&255;do if((qa&255)>31&qa<<24>>24>-1){if(!(i[764735+(Da+-32)>>0]|0)){fa=106;break}i[ma>>0]=qa;La=ma+1|0}else fa=106;while(0);if((fa|0)==106){fa=0;i[ma>>0]=37;i[ma+1>>0]=i[764831+(Da>>>4)>>0]|0;i[ma+2>>0]=i[764831+(Da&15)>>0]|0;La=ma+3|0}ra=ra+1|0;qa=i[ra>>0]|0;if(!(qa<<24>>24)){Ka=La;break}else ma=La}}i[Ka>>0]=0;Ma=k[d>>2]|0}else Ma=ta;k[d>>2]=Ma}if((i[V>>0]|0)==47)break;ux(n,Y)|0;ma=n+(xx(n)|0)|0;i[ma>>0]=47;i[ma+1>>0]=0;qx(n,V)|0;zp(n,V,d)|0}while(0);pa=k[H>>2]|0}Na=pa;Oa=(tx(N,V)|0)==0&1}else{Na=na;Oa=0}ma=k[z>>2]|0;k[z>>2]=ma+1;Pa=(Oa|0)!=0;if(Pa|(ma|0)>=(Na|0)){fa=117;break}if(!(k[d>>2]|0))na=Na;else{Qa=Na;break d}}if((fa|0)==73)k[z>>2]=(k[z>>2]|0)+1;else if((fa|0)==117?!Pa:0){Qa=Na;break}tk(764848);ha=oa;fa=130;break a}else Qa=ja;while(0);if(!Qa){k[z>>2]=0;k[d>>2]=202;i:do if((k[y>>2]|0)>0){ja=202;ia=0;na=1;do{if(!na){Ra=ja;break i}k[d>>2]=0;ja=Go(k[ba>>2]|0,k[121088+(ia<<2)>>2]|0,$,d)|0;k[d>>2]=ja;ia=(k[z>>2]|0)+1|0;k[z>>2]=ia;na=(ja|0)==202}while((ia|0)<(k[y>>2]|0));if(!na)Ra=ja;else fa=124}else fa=124;while(0);if((fa|0)==124){k[d>>2]=0;Co(k[ba>>2]|0,y,z,d)|0;Ro(k[ba>>2]|0,k[y>>2]|0,$,d)|0;Ra=k[d>>2]|0}if(Ra){ha=oa;fa=130;break}}ia=Qa+1|0;k[H>>2]=ia;ai(k[(k[(k[ba>>2]|0)+4>>2]|0)+12>>2]|0,W,d)|0;ai(k[(k[ca>>2]|0)+12>>2]|0,X,d)|0;if((k[(k[ba>>2]|0)+4>>2]|0)!=(k[ca>>2]|0)?(Ax(W,X,1025)|0)!=0:0){k[g>>2]=ia;Xw(_,763868,g)|0;ma=k[E>>2]|0;Eq(k[ba>>2]|0,_,ma,((ma|0)<0)<<31>>31,764907,d)|0;k[f>>2]=ia;Xw(_,763876,f)|0;Dq(k[ba>>2]|0,_,U,764943,d)|0;qs(k[ba>>2]|0,d)|0;ha=oa;fa=130;break}k[h>>2]=ia;Xw(_,763868,h)|0;ia=k[E>>2]|0;Eq(k[ba>>2]|0,_,ia,((ia|0)<0)<<31>>31,764907,d)|0;ha=oa;fa=130}else{ha=oa;fa=130}}else{ha=0;fa=130}}else{tk(763830);k[d>>2]=349;ha=0;fa=130}while(0);if((fa|0)==130)if(!b){ea=ha;fa=132}else{Sa=k[d>>2]|0;Ta=ha}if((fa|0)==132){fa=Fh(k[ba>>2]|0,d)|0;k[d>>2]=fa;Sa=fa;Ta=ea}ea=(Sa|0)==0?Ta:Sa;k[d>>2]=ea;da=ea;r=e;return da|0}function rp(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0;b=xx(a)|0;if((i[a>>0]|0)==39?(i[a+(b+-1)>>0]|0)==39:0){c=b+-2|0;if((c|0)>0){b=0;do{d=b;b=b+1|0;i[a+d>>0]=i[a+b>>0]|0}while((b|0)!=(c|0))}i[a+c>>0]=0}c=(xx(a)|0)+-1|0;a:do if((c|0)>0){b=0;while(1){if((i[a+b>>0]|0)!=32){e=b;break a}d=b+1|0;if((d|0)<(c|0))b=d;else{e=d;break}}}else e=0;while(0);if((e|0)!=(c|0)&(c|0)>-1)f=c;else return;while(1){c=a+f|0;if((i[c>>0]|0)!=32){g=12;break}i[c>>0]=0;if((f|0)>0)f=f+-1|0;else{g=12;break}}if((g|0)==12)return}function sp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;d=r;r=r+352|0;e=d+24|0;f=d+16|0;g=d+8|0;h=d;i=d+36|0;j=d+32|0;l=d+269|0;m=d+194|0;n=d+113|0;o=d+40|0;p=d+28|0;k[j>>2]=764972;q=k[c>>2]|0;if(q){s=q;r=d;return s|0}k[b>>2]=0;q=Ro(a,0,n,c)|0;k[c>>2]=q;t=q;a:while(1){switch(t|0){case 202:{u=6;break a;break}case 0:break;default:{v=t;break a}}q=Fo(a,j,1,0,0,n,c)|0;k[c>>2]=q;if(q){t=q;continue}k[b>>2]=(k[b>>2]|0)+1;t=0}if((u|0)==6){k[c>>2]=0;v=0}if((k[b>>2]|0)<1){s=v;r=d;return s|0}else{w=v;x=1;y=1;z=0}while(1){if(w){s=w;u=17;break}k[h>>2]=y;Xw(l,763868,h)|0;v=Ko(a,l,i,n,c)|0;k[c>>2]=v;if((v|0)!=202){t=x+1|0;if((z|0)>0){j=y-z|0;k[g>>2]=j;Xw(m,763868,g)|0;yq(a,l,m,c)|0;k[f>>2]=y;Xw(l,763876,f)|0;k[e>>2]=j;Xw(m,763876,e)|0;j=To(a,l,p,o,c)|0;k[c>>2]=j;if(!j){Aq(a,l,c)|0;Dq(a,m,k[p>>2]|0,o,c)|0;qs(a,c)|0;dy(k[p>>2]|0);A=k[c>>2]|0}else A=j;if((A|0)==202){k[c>>2]=0;B=0;C=t;D=z}else{B=A;C=t;D=z}}else{B=v;C=t;D=z}}else{k[c>>2]=0;B=0;C=x;D=z+1|0}if((C|0)>(k[b>>2]|0)){s=B;u=17;break}else{w=B;x=C;y=y+1|0;z=D}}if((u|0)==17){r=d;return s|0}return 0}function tp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0;e=r;r=r+4320|0;f=e+32|0;g=e+24|0;h=e+16|0;j=e+8|0;l=e;m=e+56|0;n=e+52|0;o=e+4233|0;p=e+3208|0;q=e+48|0;s=e+2183|0;t=e+1158|0;u=e+133|0;v=e+60|0;w=e+40|0;k[m>>2]=0;k[n>>2]=0;x=k[d>>2]|0;if(x){y=x;r=e;return y|0}k[c>>2]=0;x=sp(a,m,d)|0;k[d>>2]=x;x=k[m>>2]|0;do if((x|0)>=(b|0)){k[j>>2]=b;Xw(o,763868,j)|0;m=Ko(a,o,n,v,d)|0;k[d>>2]=m;if(!m){z=k[n>>2]|0;a:do if((z|0)>0){A=Sh(a,c,d)|0;k[d>>2]=A;B=A;C=29}else{b:do if(!z){k[d>>2]=344;k[h>>2]=0;k[h+4>>2]=b;Xw(v,765035,h)|0;tk(v)}else{k[n>>2]=0-z;k[g>>2]=b;Xw(o,763876,g)|0;A=To(a,o,q,v,d)|0;k[d>>2]=A;if(!A){ux(p,k[q>>2]|0)|0;dy(k[q>>2]|0);D=k[d>>2]|0}else D=A;if((D|0)==202){k[d>>2]=344;k[f>>2]=b;Xw(v,765077,f)|0;tk(v);break}rp(p);if(up(p)|0){tk(765114);A=Ih(5,c,p,1,d)|0;k[d>>2]=A;if(!A)break a;tk(765168);k[d>>2]=0;A=Ih(5,c,p,0,d)|0;k[d>>2]=A;B=A;C=29;break a}A=vp(p,t,d)|0;k[d>>2]=A;A=Ih(5,c,t,1,d)|0;k[d>>2]=A;if(!A)break a;tk(765168);k[d>>2]=0;A=Ih(5,c,t,0,d)|0;k[d>>2]=A;if(!A)break a;k[d>>2]=0;k[w>>2]=t;k[w+4>>2]=u;A=wp(a,t,u,0,0,0,d)|0;k[d>>2]=A;k[c>>2]=0;A=t;E=0;while(1){do if(!(i[A>>0]|0))F=0;else{G=xp(A,p,s,d)|0;k[d>>2]=G;if(G){k[d>>2]=0;F=0;break}if(!(up(s)|0)){G=vp(s,A,d)|0;k[d>>2]=G;ux(s,A)|0}G=Ih(5,c,s,1,d)|0;k[d>>2]=G;if(!G)break b;tk(765219);tk(765261);k[d>>2]=0;G=(Ih(5,c,s,0,d)|0)==0;k[d>>2]=0;F=G&1}while(0);G=E+1|0;if(!((G|0)<2&(F|0)==0))break b;A=k[w+(G<<2)>>2]|0;E=G}}while(0);B=k[d>>2]|0;C=29}while(0);if((C|0)==29?(B|0)!=0:0){H=B;break}z=k[c>>2]|0;if(!z){tk(765303);k[d>>2]=343;H=343;break}else{E=(Jl(z,-1,763315,k[n>>2]|0,d)|0)==0;z=E?0:343;k[d>>2]=z;I=z;C=33;break}}else H=m}else{k[d>>2]=344;k[l>>2]=b;k[l+4>>2]=x;Xw(v,764979,l)|0;tk(v);I=k[d>>2]|0;C=33}while(0);if((C|0)==33)if(!I){y=0;r=e;return y|0}else H=I;I=k[c>>2]|0;if(!I){y=H;r=e;return y|0}Fh(I,d)|0;k[c>>2]=0;y=k[d>>2]|0;r=e;return y|0}function up(a){a=a|0;var b=0,c=0,d=0;b=rx(a,58)|0;if((((((((((b|0)!=0?(c=rx(a,59)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,47)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,63)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,64)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,38)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,61)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,43)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,36)|0,(c|0)==0|c>>>0>b>>>0):0)?(c=rx(a,44)|0,(c|0)==0|c>>>0>b>>>0):0){d=1;return d|0}d=0;return d|0}function vp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;d=r;r=r+1040|0;e=d;f=k[c>>2]|0;if(f){g=f;r=d;return g|0}ux(e,a)|0;f=k[c>>2]|0;if(!f){h=i[a>>0]|0;a:do if(!(h<<24>>24))j=e;else{l=h;m=a;n=e;while(1){o=m+1|0;if(l<<24>>24==37){p=i[o>>0]|0;q=p<<24>>24;if(!(p<<24>>24)){j=n;break a}if((p+-48&255)<10)s=q+-48|0;else s=q+10+((p+-65&255)<6?-65:-97)|0;p=s<<4;i[n>>0]=p;q=i[m+2>>0]|0;if(!(q<<24>>24)){j=n;break a}t=q<<24>>24;if((q+-48&255)<10)u=t+-48|0;else u=t+10+((q+-65&255)<6?-65:-97)|0;i[n>>0]=u+p;v=m+3|0}else{i[n>>0]=l;v=o}o=n+1|0;l=i[v>>0]|0;if(!(l<<24>>24)){j=o;break}else{m=v;n=o}}}while(0);i[j>>0]=0;w=k[c>>2]|0}else w=f;k[c>>2]=w;ux(b,e)|0;g=k[c>>2]|0;r=d;return g|0}function wp(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;h=r;r=r+7184|0;j=h+6157|0;l=h+5132|0;m=h+4107|0;n=h+3082|0;o=h+2057|0;p=h+1032|0;q=h;s=k[g>>2]|0;if(s){t=s;r=h;return t|0}i[q>>0]=0;i[p>>0]=0;i[o>>0]=0;i[n>>0]=0;s=qk(a,n,g)|0;k[g>>2]=s;s=Dh(n,0,l,m,0,o,p,q,g)|0;k[g>>2]=s;s=(i[p>>0]|i[o>>0]|i[q>>0])<<24>>24!=0;u=s<<31>>31;v=bi(a,p,g)|0;k[g>>2]=v;ux(q,p)|0;v=ai(n,o,g)|0;k[g>>2]=v;ux(n,o)|0;do if(!(px(p,764609)|0)){if(!(i[m>>0]|0))i[o>>0]=0;else ux(n,m)|0;v=Ex(n,765357)|0;if(v){ux(l,v+3|0)|0;ux(n,l)|0}v=Ex(o,765357)|0;if(!v){w=1;x=44}else{ux(l,v+3|0)|0;ux(o,l)|0;w=1;x=44}}else{if(!(px(p,777975)|0))if(s){tk(765361);k[g>>2]=125;break}else{i[o>>0]=0;w=1;x=44;break}if(!(px(p,765414)|0)){i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;i[q>>0]=0;i[o>>0]=0;w=1;x=44;break}if(!(px(p,765425)|0)){i[q>>0]=0;i[o>>0]=0;w=1;x=44;break}if(!(px(p,765434)|0)){i[q>>0]=0;i[o>>0]=0;w=1;x=44;break}if(!(px(p,765442)|0)){ux(n,m)|0;ux(o,l)|0;v=p;k[v>>2]=1701603686;k[v+4>>2]=3092282;v=q;k[v>>2]=1701603686;k[v+4>>2]=3092282;w=1;x=44;break}if(!(px(p,765458)|0)){ux(n,m)|0;v=p;k[v>>2]=1701603686;k[v+4>>2]=3092282;v=q;k[v>>2]=1886680168;k[v+4>>2]=3092282;w=1;x=44;break}if(!(px(p,765470)|0)){ux(n,m)|0;v=p;k[v>>2]=1701603686;k[v+4>>2]=3092282;i[q>>0]=i[765481]|0;i[q+1>>0]=i[765482]|0;i[q+2>>0]=i[765483]|0;i[q+3>>0]=i[765484]|0;i[q+4>>0]=i[765485]|0;i[q+5>>0]=i[765486]|0;i[q+6>>0]=i[765487]|0;w=1;x=44;break}if(!(px(p,765488)|0)){ux(n,m)|0;v=p;k[v>>2]=1701603686;k[v+4>>2]=3092282;v=q;a=765501;y=v+9|0;do{i[v>>0]=i[a>>0]|0;v=v+1|0;a=a+1|0}while((v|0)<(y|0));w=1;x=44;break}if(!(px(p,765510)|0)){i[n>>0]=0;ux(o,l)|0;i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;a=q;k[a>>2]=1701603686;k[a+4>>2]=3092282;w=0;x=44;break}if(!(px(p,765522)|0)){i[n>>0]=0;i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;a=q;k[a>>2]=1886680168;k[a+4>>2]=3092282;w=0;x=44;break}if(!(px(p,765530)|0)){i[n>>0]=0;i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;a=q;k[a>>2]=1886680168;k[a+4>>2]=3092282;w=0;x=44;break}if(!(px(p,765481)|0)){i[n>>0]=0;i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;i[q>>0]=i[765481]|0;i[q+1>>0]=i[765482]|0;i[q+2>>0]=i[765483]|0;i[q+3>>0]=i[765484]|0;i[q+4>>0]=i[765485]|0;i[q+5>>0]=i[765486]|0;i[q+6>>0]=i[765487]|0;w=0;x=44;break}if(!(px(p,765546)|0)){i[n>>0]=0;i[p>>0]=i[777975]|0;i[p+1>>0]=i[777976]|0;i[p+2>>0]=i[777977]|0;i[p+3>>0]=i[777978]|0;i[p+4>>0]=i[777979]|0;i[p+5>>0]=i[777980]|0;i[p+6>>0]=i[777981]|0;i[q>>0]=i[765481]|0;i[q+1>>0]=i[765482]|0;i[q+2>>0]=i[765483]|0;i[q+3>>0]=i[765484]|0;i[q+4>>0]=i[765485]|0;i[q+5>>0]=i[765486]|0;i[q+6>>0]=i[765487]|0;w=0;x=44;break}if(!(px(p,765501)|0)){k[g>>2]=125;tk(765561);i[o>>0]=0;i[n>>0]=0;w=u;x=44;break}if(!(px(p,765612)|0)){k[g>>2]=125;tk(765622);i[o>>0]=0;i[n>>0]=0;w=u;x=44;break}if(!(px(p,765674)|0)){k[g>>2]=125;tk(765685);i[o>>0]=0;i[n>>0]=0;w=u;x=44}else{w=u;x=44}}while(0);if((x|0)==44?(k[g>>2]|0)==0:0){do if(b){if(!(i[n>>0]|0)){i[b>>0]=0;break}u=Ex(n,765357)|0;if(u){l=u+3|0;u=l-n|0;Bx(b,n,u)|0;m=k[g>>2]|0;if((m|0)>0){z=m;A=j}else{B=m;C=j;D=b+u|0;E=l;x=51}}else{B=0;C=j;D=b;E=n;x=51}if((x|0)==51){l=0;u=0;a:while(1){m=i[E+l>>0]|0;s=l;b:while(1){switch(m<<24>>24){case 0:{F=u;break a;break}case 47:break;default:{G=m;H=s;x=54;break b}}a=s+1|0;if((i[E+a>>0]|0)==47){m=47;s=a}else{I=47;J=a;break}}if((x|0)==54){x=0;I=G;J=H+1|0}i[j+u>>0]=I;l=J;u=u+1|0}i[j+F>>0]=0;if(!B){u=i[C>>0]|0;if(!(u<<24>>24))K=D;else{l=u;u=C;s=D;while(1){m=l&255;if((l&255)>31&l<<24>>24>-1?(i[764735+(m+-32)>>0]|0)!=0:0){i[s>>0]=l;L=s+1|0}else{i[s>>0]=37;i[s+1>>0]=i[764831+(m>>>4)>>0]|0;i[s+2>>0]=i[764831+(m&15)>>0]|0;L=s+3|0}u=u+1|0;l=i[u>>0]|0;if(!(l<<24>>24)){K=L;break}else s=L}}i[K>>0]=0;M=k[g>>2]|0}else M=B;k[g>>2]=M;z=M;A=C}k[g>>2]=z}while(0);do if(c){if(!(i[o>>0]|0)){i[c>>0]=0;break}z=Ex(o,765357)|0;if(!z){N=0;O=o}else{C=z+3|0;z=C-o|0;Bx(c,o,z)|0;N=z;O=C}C=c+N|0;z=k[g>>2]|0;if((z|0)>0)P=z;else{A=0;M=0;c:while(1){B=i[O+A>>0]|0;K=A;d:while(1){switch(B<<24>>24){case 0:{Q=M;break c;break}case 47:break;default:{R=B;S=K;x=75;break d}}L=K+1|0;if((i[O+L>>0]|0)==47){B=47;K=L}else{T=47;U=L;break}}if((x|0)==75){x=0;T=R;U=S+1|0}i[j+M>>0]=T;A=U;M=M+1|0}i[j+Q>>0]=0;if(!z){M=i[j>>0]|0;if(!(M<<24>>24))V=C;else{A=M;M=j;K=C;while(1){B=A&255;if((A&255)>31&A<<24>>24>-1?(i[764735+(B+-32)>>0]|0)!=0:0){i[K>>0]=A;W=K+1|0}else{i[K>>0]=37;i[K+1>>0]=i[764831+(B>>>4)>>0]|0;i[K+2>>0]=i[764831+(B&15)>>0]|0;W=K+3|0}M=M+1|0;A=i[M>>0]|0;if(!(A<<24>>24)){V=W;break}else K=W}}i[V>>0]=0;X=k[g>>2]|0}else X=z;k[g>>2]=X;P=X}k[g>>2]=P}while(0);if(d)ux(d,p)|0;if(e)ux(e,q)|0;if(f)k[f>>2]=w}t=k[g>>2]|0;r=h;return t|0}function xp(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0;e=r;r=r+1040|0;f=e;g=k[d>>2]|0;if(g){h=g;r=e;return h|0}ux(f,a)|0;if((yx(f,765738,4)|0)!=0?(yx(f,765743,6)|0)!=0:0){if((i[b>>0]|0)==47){i[c>>0]=47;i[c+1>>0]=0;a=(i[b>>0]|0)==47;g=xx(c)|0;if(a){a=0;j=g;while(1){l=c+j|0;i[l>>0]=47;i[l+1>>0]=0;a=a+1|0;l=(i[b+a>>0]|0)==47;m=xx(c)|0;if(!l){n=m;break}else j=m}}else n=g;g=Ex(f,c)|0;if(!g)o=f;else{j=g;do{g=j+n|0;j=Ex(g,c)|0}while((j|0)!=0);o=g}i[c+(n+-1)>>0]=0;n=Ex(o,c)|0;do if(!n){j=Cx(o,47)|0;if(!j){i[f>>0]=0;break}else{i[j>>0]=0;break}}else i[n>>0]=0;while(0);qx(f,b)|0}else{n=Cx(f,47)|0;if(!n)i[f>>0]=0;else i[n+1>>0]=0;qx(f,b)|0}p=zp(f,c,d)|0}else{tk(765750);tk(765806);tk(765859);p=125}k[d>>2]=p;h=p;r=e;return h|0}function yp(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;c=r;r=r+2064|0;d=c+1025|0;e=c;f=k[b>>2]|0;if(f){g=f;r=c;return g|0}Jx(e,1025)|0;f=k[b>>2]|0;if((f|0)>0)h=f;else{j=0;l=0;a:while(1){m=i[e+j>>0]|0;n=j;b:while(1){switch(m<<24>>24){case 0:{o=l;break a;break}case 47:break;default:{p=m;q=n;s=5;break b}}t=n+1|0;if((i[e+t>>0]|0)==47){m=47;n=t}else{u=47;v=t;break}}if((s|0)==5){s=0;u=p;v=q+1|0}i[d+l>>0]=u;j=v;l=l+1|0}i[d+o>>0]=0;if(!f){o=i[d>>0]|0;if(!(o<<24>>24))w=a;else{l=o;o=d;d=a;while(1){a=l&255;if((l&255)>31&l<<24>>24>-1?(i[764735+(a+-32)>>0]|0)!=0:0){i[d>>0]=l;x=d+1|0}else{i[d>>0]=37;i[d+1>>0]=i[764831+(a>>>4)>>0]|0;i[d+2>>0]=i[764831+(a&15)>>0]|0;x=d+3|0}o=o+1|0;l=i[o>>0]|0;if(!(l<<24>>24)){w=x;break}else d=x}}i[w>>0]=0;y=k[b>>2]|0}else y=f;k[b>>2]=y;h=y}g=h;r=c;return g|0}function zp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0;d=k[c>>2]|0;if(d){e=d;return e|0}d=cy(8)|0;f=(d|0)==0;if(!f){k[d>>2]=0;k[d+4>>2]=0}i[b>>0]=0;g=Ex(a,765357)|0;do if(g){h=rx(g+3|0,47)|0;if(!h){ux(b,a)|0;if(f)break;else{j=36;break}}else{l=h-a|0;Bx(b,a,l)|0;i[b+l>>0]=0;m=h;j=7;break}}else{m=a;j=7}while(0);if((j|0)==7){if((i[m>>0]|0)==47){a=b+(xx(b)|0)|0;i[a>>0]=47;i[a+1>>0]=0}a=Fx(m,769609)|0;a:do if(a){g=d+4|0;if(f)while(1)if(!(Fx(0,769609)|0))break a;else{n=0;o=0;p=0;q=0;r=0;s=0;t=a}while(1){do if(!(tx(t,765881)|0)){if(n){if(!o){u=s;v=q;w=p;x=0;y=n;z=0;break}h=k[o+8>>2]|0;l=o+4|0;A=h;k[A+4>>2]=k[l>>2];k[(k[l>>2]|0)+8>>2]=h;dy(o);k[g>>2]=h;h=(k[d>>2]|0)+-1|0;k[d>>2]=h;if(h){u=A;v=h;w=A;x=A;y=h;z=A;break}k[g>>2]=0;u=0;v=0;w=0;x=0;y=0;z=0;break}if((i[m>>0]|0)==47){u=s;v=q;w=p;x=o;y=0;z=r}else{A=cy(12)|0;k[A>>2]=t;if(!p){k[A+4>>2]=A;k[A+8>>2]=A}else{h=p+4|0;l=k[h>>2]|0;k[A+4>>2]=l;k[A+8>>2]=p;k[l+8>>2]=A;k[h>>2]=A}k[g>>2]=A;h=q+1|0;k[d>>2]=h;u=A;v=h;w=A;x=A;y=h;z=A}}else if(!(tx(t,795591)|0)){u=s;v=q;w=p;x=o;y=n;z=r}else{A=cy(12)|0;k[A>>2]=t;if(!r){k[A+4>>2]=A;k[A+8>>2]=A}else{h=r+4|0;l=k[h>>2]|0;k[A+4>>2]=l;k[A+8>>2]=r;k[l+8>>2]=A;k[h>>2]=A}k[g>>2]=A;h=n+1|0;k[d>>2]=h;u=A;v=h;w=A;x=A;y=h;z=A}while(0);t=Fx(0,769609)|0;if(!t){B=u;C=v;break}else{n=y;o=x;p=w;q=v;r=z;s=u}}if(C){g=d+4|0;A=B;h=C;while(1){if(A){l=k[A+4>>2]|0;D=k[l>>2]|0;E=l+4|0;F=k[l+8>>2]|0;k[F+4>>2]=k[E>>2];k[(k[E>>2]|0)+8>>2]=F;dy(l);l=h+-1|0;k[d>>2]=l;if(!l){k[g>>2]=0;G=0;H=0;I=D}else{G=l;H=A;I=D}}else{G=h;H=0;I=0}qx(b,I)|0;D=b+(xx(b)|0)|0;i[D>>0]=47;i[D+1>>0]=0;if(!G)break;else{A=H;h=G}}}}while(0);G=b+((xx(b)|0)+-1)|0;i[G>>0]=0;j=36}if((j|0)==36)dy(d);e=k[c>>2]|0;return e|0}function Ap(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0;e=k[d>>2]|0;if(e){f=e;return f|0}i[c>>0]=0;if((up(a)|0)==0?(i[a>>0]|0)!=47:0)g=6;else g=4;a:do if((g|0)==4){if((up(b)|0)==0?(i[b>>0]|0)!=47:0){g=6;break}e=xx(a)|0;h=xx(b)|0;if((h|0)>0&(e|0)>0){j=0;l=0;while(1){b:do if((j|0)<(h|0)){m=j;while(1){if((i[b+m>>0]|0)!=47){n=m;break b}o=m+1|0;if((o|0)<(h|0))m=o;else{n=o;break}}}else n=j;while(0);c:do if((l|0)<(e|0)){m=l;while(1){if((i[a+m>>0]|0)!=47){p=m;break c}o=m+1|0;if((o|0)<(e|0))m=o;else{p=o;break}}}else p=l;while(0);m=n;while(1)if((m|0)<(h|0)?(i[b+m>>0]|0)!=47:0)m=m+1|0;else{q=m;break}m=p;while(1)if((m|0)<(e|0)?(i[a+m>>0]|0)!=47:0)m=m+1|0;else{r=m;break}if((q|0)!=(r|0)){s=n;t=p;break}if(Ax(b+n|0,a+p|0,q-p|0)|0){s=n;t=p;break}l=q+1|0;j=q+1|0;if(!((j|0)<(h|0)&(l|0)<(e|0)))break a}if((t|0)<(e|0)){l=t;do{if((i[a+l>>0]|0)==47){h=c+(xx(c)|0)|0;i[h>>0]=46;i[h+1>>0]=46;i[h+2>>0]=47;i[h+3>>0]=0}l=l+1|0}while((l|0)!=(e|0))}qx(c,b+s|0)|0}}while(0);if((g|0)==6){k[d>>2]=125;tk(765884)}f=k[d>>2]|0;return f|0}function Bp(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0;j=r;r=r+10384|0;l=j+9351|0;m=j+9278|0;n=j+48|0;o=j+44|0;p=j+40|0;q=j+36|0;s=j+32|0;t=j+28|0;u=j+24|0;v=j+20|0;w=j+16|0;x=j+12|0;y=j+8|0;z=j+6203|0;A=j+5178|0;B=j+4153|0;C=j+3128|0;D=j+2103|0;E=j+1078|0;F=j+53|0;G=j+52|0;H=j;k[u>>2]=0;k[x>>2]=0;k[y>>2]=0;i[G>>0]=0;I=k[h>>2]|0;if(I){J=I;r=j;return J|0}k[g>>2]=0;k[H>>2]=j+8253;k[H+4>>2]=j+7228;I=(f|0)==0;do if(!I){K=i[f>>0]|0;if(!(K<<24>>24)){i[z>>0]=0;break}if(up(f)|0){ux(z,f)|0;break}L=K;K=0;M=0;a:while(1){N=L;O=K;b:while(1){switch(N<<24>>24){case 0:{P=M;break a;break}case 47:break;default:{Q=N;R=O;S=10;break b}}T=O+1|0;U=i[f+T>>0]|0;if(U<<24>>24==47){N=47;O=T}else{V=47;W=T;X=U;break}}if((S|0)==10){S=0;O=R+1|0;V=Q;W=O;X=i[f+O>>0]|0}i[l+M>>0]=V;L=X;K=W;M=M+1|0}i[l+P>>0]=0;M=i[l>>0]|0;if(!(M<<24>>24))Y=z;else{K=M;M=l;L=z;while(1){O=K&255;if((K&255)>31&K<<24>>24>-1?(i[764735+(O+-32)>>0]|0)!=0:0){i[L>>0]=K;Z=L+1|0}else{i[L>>0]=37;i[L+1>>0]=i[764831+(O>>>4)>>0]|0;i[L+2>>0]=i[764831+(O&15)>>0]|0;Z=L+3|0}M=M+1|0;K=i[M>>0]|0;if(!(K<<24>>24)){Y=Z;break}else L=Z}}i[Y>>0]=0;if((i[z>>0]|0)!=47){yp(F,h)|0;L=F+(xx(F)|0)|0;i[L>>0]=47;i[L+1>>0]=0;qx(F,z)|0;zp(F,z,h)|0}}else i[z>>0]=0;while(0);Y=jp(a,n,o,p,q,s,t,v,h)|0;k[h>>2]=Y;if(!Y){t=Qo(a,790194,l,m,h)|0;k[h>>2]=t;if((t|0)==202)_=340;else{rp(l);if(px(l,763315)|0){k[h>>2]=340;tk(763726)}_=Ko(a,790243,x,m,h)|0}k[h>>2]=_;$=_}else $=Y;k[h>>2]=$;c:do if((k[x>>2]|0)<1)S=71;else{$=k[v>>2]|0;Y=1;while(1){if(k[g>>2]|0)break c;if(k[h>>2]|0){S=71;break c}_=k[n>>2]|0;if(!((_|0)!=0?(In(a,_,Y,((Y|0)<0)<<31>>31,1,0,1,0,G,H,w,h)|0,(px(k[H>>2]|0,b)|0)!=0):0))S=34;d:do if((S|0)==34){S=0;_=k[o>>2]|0;if((_|0)!=0?(In(a,_,Y,((Y|0)<0)<<31>>31,1,0,1,0,G,H,w,h)|0,(px(k[H>>2]|0,c)|0)!=0):0)break;_=k[p>>2]|0;if((_|0)!=0?($m(a,_,Y,((Y|0)<0)<<31>>31,1,0,1,0,0,y,w,h)|0,(k[y>>2]|0)!=(d|0)):0)break;_=k[q>>2]|0;e:do if(_){switch($|0){case 2:case 12:break;default:break e}zn(a,_,Y,((Y|0)<0)<<31>>31,1,0,1,0,0,u,w,h)|0;if((k[u>>2]|0)!=(e|0))break d}while(0);if(I){tk(765942);k[g>>2]=Y;break}_=k[s>>2]|0;f:do if(_){In(a,_,Y,((Y|0)<0)<<31>>31,1,0,1,0,G,H,w,h)|0;ux(A,k[H>>2]|0)|0;i[B>>0]=0;m=i[A>>0]|0;if(m<<24>>24){if(!(m<<24>>24==47|(up(A)|0)!=0)){ux(B,A)|0;m=wp(a,D,E,0,0,0,h)|0;k[h>>2]=m;m=i[D>>0]|0;g:do if(!(m<<24>>24))i[A>>0]=0;else{if(!(m<<24>>24==47|(up(D)|0)!=0)){yp(F,h)|0;l=F+(xx(F)|0)|0;i[l>>0]=47;i[l+1>>0]=0;qx(F,D)|0;zp(F,D,h)|0}xp(D,A,C,h)|0;switch(k[h>>2]|0){case 0:{ux(A,C)|0;break g;break}case 125:{k[h>>2]=0;i[A>>0]=0;break g;break}default:break g}}while(0);m=i[E>>0]|0;if(!(m<<24>>24)){i[B>>0]=0;break}if(!(m<<24>>24==47|(up(E)|0)!=0)){yp(F,h)|0;m=F+(xx(F)|0)|0;i[m>>0]=47;i[m+1>>0]=0;qx(F,E)|0;zp(F,E,h)|0}xp(E,B,C,h)|0;switch(k[h>>2]|0){case 0:{ux(B,C)|0;break f;break}case 125:{k[h>>2]=0;i[B>>0]=0;break f;break}default:break f}}}else S=46}else{i[A>>0]=0;S=46}while(0);if((S|0)==46){S=0;_=wp(a,A,B,0,0,0,h)|0;k[h>>2]=_;_=i[A>>0]|0;if(_<<24>>24!=0?!(_<<24>>24==47|(up(A)|0)!=0):0){yp(F,h)|0;_=F+(xx(F)|0)|0;i[_>>0]=47;i[_+1>>0]=0;qx(F,A)|0;zp(F,A,h)|0}_=i[B>>0]|0;if(_<<24>>24!=0?!(_<<24>>24==47|(up(B)|0)!=0):0){yp(F,h)|0;_=F+(xx(F)|0)|0;i[_>>0]=47;i[_+1>>0]=0;qx(F,B)|0;zp(F,B,h)|0}}if((tx(A,z)|0)!=0?(tx(B,z)|0)!=0:0)break;k[g>>2]=Y}while(0);if((Y|0)<(k[x>>2]|0))Y=Y+1|0;else{S=71;break}}}while(0);if(((S|0)==71?(k[g>>2]|0)==0:0)?(k[h>>2]|0)==0:0){k[h>>2]=342;tk(765997)}J=k[h>>2]|0;r=j;return J|0}function Cp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+1040|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}else{h=0;j=0}a:while(1){l=h;b:while(1){m=i[a+l>>0]|0;switch(m<<24>>24){case 0:{n=j;break a;break}case 47:break;default:{o=m;p=l;break b}}q=l+1|0;if((i[a+q>>0]|0)==47)l=q;else{o=m;p=l;break}}i[e+j>>0]=o;h=p+1|0;j=j+1|0}i[e+n>>0]=0;if(!f){n=i[e>>0]|0;if(!(n<<24>>24))s=b;else{j=n;n=e;e=b;while(1){b=j&255;if((j&255)>31&j<<24>>24>-1?(i[764735+(b+-32)>>0]|0)!=0:0){i[e>>0]=j;t=e+1|0}else{i[e>>0]=37;i[e+1>>0]=i[764831+(b>>>4)>>0]|0;i[e+2>>0]=i[764831+(b&15)>>0]|0;t=e+3|0}n=n+1|0;j=i[n>>0]|0;if(!(j<<24>>24)){s=t;break}else e=t}}i[s>>0]=0;u=k[c>>2]|0}else u=f;k[c>>2]=u;g=u;r=d;return g|0}function Dp(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;if((a|0)==0|(b|0)==0){c=368;return c|0}d=k[30309]|0;e=(d|0)==0;f=k[30308]|0;g=(f|0)>0;if(e&g){c=368;return c|0}if((d|0)!=0&(f|0)<1){c=368;return c|0}a:do if(g){h=0;while(1){if(!(tx(a,k[d+(h<<3)>>2]|0)|0)){i=h;break}h=h+1|0;if((h|0)>=(f|0))break a}h=d+(i<<3)+4|0;j=(k[h>>2]|0)+1|0;k[h>>2]=j;k[b>>2]=j;c=0;return c|0}while(0);if(e)l=cy(8)|0;else l=fy(d,(f<<3)+8|0)|0;if(!l){c=360;return c|0}f=cy((xx(a)|0)+1|0)|0;if(!f){dy(l);c=360;return c|0}else{ux(f,a)|0;k[30309]=l;a=k[30308]|0;k[l+(a<<3)>>2]=f;k[l+(a<<3)+4>>2]=1;k[b>>2]=1;k[30308]=(k[30308]|0)+1;c=0;return c|0}return 0}function Ep(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;if(!a){c=368;return c|0}d=k[30309]|0;e=(d|0)==0;f=k[30308]|0;g=(f|0)>0;if(e&g){c=368;return c|0}if((d|0)!=0&(f|0)<1){c=368;return c|0}a:do if(g){h=0;while(1){if(!(tx(a,k[d+(h<<3)>>2]|0)|0)){i=h;break}h=h+1|0;if((h|0)>=(f|0))break a}h=d+(i<<3)+4|0;if((k[h>>2]|0)>=(b|0)){c=0;return c|0}k[h>>2]=b;c=0;return c|0}while(0);if(e)j=cy(8)|0;else j=fy(d,(f<<3)+8|0)|0;if(!j){c=360;return c|0}f=cy((xx(a)|0)+1|0)|0;if(!f){dy(j);c=360;return c|0}else{ux(f,a)|0;k[30309]=j;a=k[30308]|0;k[j+(a<<3)>>2]=f;k[j+(a<<3)+4>>2]=b;k[30308]=a+1;c=0;return c|0}return 0}function Fp(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0;if((a|0)==0|(b|0)==0){c=362;return c|0}d=cy(1)|0;k[b>>2]=d;if(!d){c=360;return c|0}else{e=1;f=0}a:while(1){b:while(1){d=Qw(a)|0;switch(d|0){case -1:{g=e;h=f;j=5;break a;break}case 10:{l=e;m=f;n=0;break a;break}case 13:break;default:{o=d;break b}}}d=f+1|0;p=f+1001|0;q=p-((p|0)%1e3|0)|0;p=k[b>>2]|0;if((q|0)>(e|0)){r=fy(p,q)|0;if(!r){l=e;m=d;n=360;break}k[b>>2]=r;s=r;t=q}else{s=p;t=e}i[s+f>>0]=o;e=t;f=d}if((j|0)==5){f=(yw(a)|0)==0;if(!h){c=367;return c|0}else{l=g;m=h;n=f?0:361}}f=m+1|0;h=k[b>>2]|0;if((f|0)!=(l|0)){l=fy(h,f)|0;if(!l)u=360;else{k[b>>2]=l;i[l+m>>0]=0;j=14}}else{i[h+m>>0]=0;j=14}if((j|0)==14)if(!n){c=0;return c|0}else u=n;dy(k[b>>2]|0);k[b>>2]=0;c=u;return c|0}function Gp(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0;b=k[a>>2]|0;if(!b){c=362;return c|0}d=a+16|0;e=a+8|0;f=a+4|0;g=a+12|0;h=a+20|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;k[f+16>>2]=0;a:do switch(i[b>>0]|0){case 10:case 0:{j=3;break}case 9:case 32:{l=b+1|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+2|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+3|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+4|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+5|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+6|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+7|0;switch(i[l>>0]|0){case 10:case 0:{j=3;break a;break}case 9:case 32:break;default:{m=l;break a}}l=b+8|0;k[d>>2]=l;n=l;b:while(1){switch(i[n>>0]|0){case 10:{o=n;j=6;break b;break}case 0:{p=b;break b;break}default:{}}n=n+1|0}if((j|0)==6){i[o>>0]=0;p=k[a>>2]|0}i[p>>0]=0;k[f>>2]=k[a>>2];k[g>>2]=7;c=0;return c|0}default:m=b}while(0);if((j|0)==3){i[b>>0]=0;b=k[a>>2]|0;k[f>>2]=b;k[d>>2]=b;k[g>>2]=7;c=0;return c|0}k[f>>2]=m;b=m;c:while(1){m=i[b>>0]|0;switch(m<<24>>24){case 10:case 0:{q=b;j=11;break c;break}default:{}}if((yx(767038,b,8)|0)==0?(a=rx(b,61)|0,(a|0)!=0):0){r=1;s=a;break}switch(m<<24>>24){case 9:case 32:{r=0;s=b;break c;break}case 61:{t=b;j=16;break c;break}default:{}}b=b+1|0}if((j|0)==11){i[q>>0]=0;r=0;s=q}else if((j|0)==16){r=1;s=t}if(!(i[s>>0]|0))u=s;else{i[s>>0]=0;u=s+1|0}s=k[f>>2]|0;f=i[s>>0]|0;t=(f+-97&255)<26?(f&255)+224&255:f;d:do if(t<<24>>24<=72){f=s;q=782840;b=72;m=t;while(1){if(b<<24>>24>m<<24>>24){j=23;break}if(!(b<<24>>24))break;q=q+1|0;f=f+1|0;a=i[q>>0]|0;b=(a+-97&255)<26?(a&255)+224&255:a;a=i[f>>0]|0;m=(a+-97&255)<26?(a&255)+224&255:a;if(b<<24>>24<m<<24>>24){j=23;break}}e:do if((j|0)==23){m=t<<24>>24>67;if(m)break d;else{v=s;w=782848;x=67;y=t}do{if(x<<24>>24>y<<24>>24)break;if(!(x<<24>>24))break e;w=w+1|0;v=v+1|0;b=i[w>>0]|0;x=(b+-97&255)<26?(b&255)+224&255:b;b=i[v>>0]|0;y=(b+-97&255)<26?(b&255)+224&255:b}while(x<<24>>24>=y<<24>>24);if(m)break d;else{z=s;A=773889;B=67;C=t}while(1){if(B<<24>>24>C<<24>>24)break d;if(!(B<<24>>24))break e;A=A+1|0;z=z+1|0;b=i[A>>0]|0;B=(b+-97&255)<26?(b&255)+224&255:b;b=i[z>>0]|0;C=(b+-97&255)<26?(b&255)+224&255:b;if(B<<24>>24<C<<24>>24)break d}}while(0);k[d>>2]=u;m=u;f:while(1){switch(i[m>>0]|0){case 10:{D=m;j=33;break f;break}case 0:{break f;break}default:{}}m=m+1|0}if((j|0)==33)i[D>>0]=0;k[g>>2]=7;c=0;return c|0}while(0);g:do if(t<<24>>24>92){E=r;F=u}else{D=s;C=767047;B=92;z=t;while(1){if(B<<24>>24>z<<24>>24){E=r;F=u;break g}if(!(B<<24>>24)){G=u;break}C=C+1|0;D=D+1|0;A=i[C>>0]|0;B=(A+-97&255)<26?(A&255)+224&255:A;A=i[D>>0]|0;z=(A+-97&255)<26?(A&255)+224&255:A;if(B<<24>>24<z<<24>>24){E=r;F=u;break g}}h:while(1){switch(i[G>>0]|0){case 9:case 32:break;default:{H=G;break h}}G=G+1|0}k[e>>2]=H;z=H;i:while(1){switch(i[z>>0]|0){case 10:{I=z;j=44;break i;break}case 0:{break i;break}default:{}}z=z+1|0}if((j|0)==44)i[I>>0]=0;k[g>>2]=0;c=0;return c|0}while(0);j:while(1){I=i[F>>0]|0;switch(I<<24>>24){case 10:case 0:{c=0;j=80;break j;break}case 9:case 32:{J=E;break}default:{if(!((E&1|0)==0&I<<24>>24==61)){K=I;L=F;break j}J=E|1}}E=J;F=F+1|0}if((j|0)==80)return c|0;k:do switch(K<<24>>24){case 47:{F=L+1|0;switch(i[F>>0]|0){case 9:case 32:{M=L+2|0;break}default:M=F}k[d>>2]=M;F=M;l:while(1){switch(i[F>>0]|0){case 10:{N=F;break l;break}case 0:{c=0;j=80;break l;break}default:{}}F=F+1|0}if((j|0)==80)return c|0;i[N>>0]=0;c=0;return c|0}case 39:{F=L+1|0;k[e>>2]=F;k[g>>2]=2;J=L;E=F;I=F;m:while(1){F=i[E>>0]|0;n:do switch(F<<24>>24){case 10:case 0:{O=I;j=60;break m;break}case 39:{H=J+2|0;switch(i[H>>0]|0){case 10:case 0:{P=I;j=62;break m;break}case 32:case 9:{Q=H;R=I;j=63;break m;break}case 39:break;default:{S=39;T=E;break n}}S=39;T=H;break}default:{S=F;T=E}}while(0);i[I>>0]=S;J=T;E=T+1|0;I=I+1|0}if((j|0)==60){i[O>>0]=0;c=0;return c|0}else if((j|0)==62){i[P>>0]=0;c=0;return c|0}else if((j|0)==63){i[R>>0]=0;U=Q;break k}break}default:{k[e>>2]=L;k[g>>2]=0;I=L;o:while(1){switch(i[I>>0]|0){case 10:case 0:{V=I;break o;break}case 9:case 32:{W=I;j=70;break o;break}default:{}}I=I+1|0}if((j|0)==70){i[W>>0]=0;U=W+1|0;break k}i[V>>0]=0;c=0;return c|0}}while(0);p:while(1){switch(i[U>>0]|0){case 47:{X=U;j=73;break p;break}case 10:case 0:{c=0;j=80;break p;break}case 9:case 32:break;default:{j=79;break p}}U=U+1|0}if((j|0)==73){U=X+1|0;switch(i[U>>0]|0){case 9:case 32:{Y=X+2|0;break}default:Y=U}k[d>>2]=Y;d=Y;q:while(1){switch(i[d>>0]|0){case 10:{Z=d;break q;break}case 0:{c=0;j=80;break q;break}default:{}}d=d+1|0}if((j|0)==80)return c|0;i[Z>>0]=0;c=0;return c|0}else if((j|0)==79){k[h>>2]=1;c=0;return c|0}else if((j|0)==80)return c|0;return 0}function Hp(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;b=r;r=r+1e4|0;c=b;if(!a){d=362;r=b;return d|0}if((k[30292]|0)>9){d=365;r=b;return d|0}e=Dw(a,774853)|0;f=k[30292]|0;k[121240+(f<<2)>>2]=e;do if(!e){g=jb(767056)|0;if(g){Bx(c,g,9999)|0;i[c+9999>>0]=0;g=Fx(c,767078)|0;a:do if(g){h=g;while(1){j=xx(a)|0;l=cy(j+2+(xx(h)|0)|0)|0;if(!l){d=360;break}ux(l,h)|0;j=l+(xx(l)|0)|0;i[j>>0]=47;i[j+1>>0]=0;qx(l,a)|0;j=Dw(l,774853)|0;k[121240+(k[30292]<<2)>>2]=j;dy(l);if(k[121240+(k[30292]<<2)>>2]|0)break a;h=Fx(0,767078)|0;if(!h)break a}r=b;return d|0}while(0);g=k[30292]|0;if(k[121240+(g<<2)>>2]|0){m=g;break}}if((i[a>>0]|0)==47|(i[766038]|0)==0){d=366;r=b;return d|0}g=xx(a)|0;h=cy(g+1+(xx(766038)|0)|0)|0;if(!h){d=360;r=b;return d|0}ux(h,766038)|0;qx(h,a)|0;g=Dw(h,774853)|0;k[121240+(k[30292]<<2)>>2]=g;dy(h);h=k[30292]|0;if(!(k[121240+(h<<2)>>2]|0)){d=366;r=b;return d|0}else m=h}else m=f;while(0);k[30292]=m+1;d=0;r=b;return d|0}function Ip(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;b=r;r=r+48|0;c=b+32|0;d=b+24|0;e=b+16|0;f=b;g=b+40|0;h=k[30292]|0;if((h|0)<1){k[30294]=5;j=0;r=b;return j|0}if((h|0)>10){j=365;r=b;return j|0}if(!(k[121240+(h+-1<<2)>>2]|0)){j=362;r=b;return j|0}h=(a|0)==0;a:while(1){a=k[30278]|0;b:while(1){l=k[121240+((k[30292]|0)+-1<<2)>>2]|0;if(a){dy(a);k[30278]=0;k[30279]=0;k[30280]=0;k[30281]=0;k[30282]=0;k[30283]=0;k[30284]=0}if(k[30285]|0){m=10;break a}k[30284]=0;n=Fp(l,121112)|0;switch(n|0){case 367:{m=12;break b;break}case 0:break;default:{j=n;m=74;break a}}if(k[30284]&1){j=0;m=74;break a}n=k[30278]|0;switch(i[n>>0]|0){case 35:{a=n;break}case 0:{m=18;break b;break}default:break b}}if((m|0)==12){m=0;a=(k[30292]|0)+-1|0;k[30292]=a;n=k[121240+(a<<2)>>2]|0;if(!n)o=a;else{ww(n)|0;o=k[30292]|0}k[121240+(o<<2)>>2]=0;if((o|0)<1){m=15;break}else continue}else if((m|0)==18?(m=0,!h):0)continue;n=Gp(121112)|0;if(n){j=n;m=74;break}n=k[30279]|0;if(!n)continue;a=i[n>>0]|0;c:do if(a<<24>>24){l=a;p=n;q=0;while(1){if((l+-97&255)<26)i[p+q>>0]=(l&255)+224;s=q+1|0;if((q|0)==7)break c;t=k[30279]|0;if(s>>>0>=(xx(t)|0)>>>0)break c;l=i[t+s>>0]|0;p=t;q=s}}while(0);n=k[30296]|0;if((n|0)==-1){m=30;break}a=k[30279]|0;q=n;n=0;while(1){if(!(tx(a,k[121180+(n<<3)>>2]|0)|0)){u=q;break}n=n+1|0;q=k[121180+(n<<3)+4>>2]|0;if((q|0)==-1){m=30;break a}}k[30294]=u;if(u)break;q=Hp(k[30280]|0)|0;if(q){j=q;m=74;break}}if((m|0)==10){k[30278]=k[30285];k[30279]=k[30286];k[30280]=k[30287];k[30281]=k[30288];k[30282]=k[30289];k[30283]=k[30290];k[30285]=0;k[30286]=0;k[30287]=0;k[30288]=0;k[30289]=0;k[30290]=0;k[30291]=0;k[30284]=1;j=0;r=b;return j|0}else if((m|0)==15){k[30294]=5;j=0;r=b;return j|0}else if((m|0)==30)k[30294]=-1;else if((m|0)==74){r=b;return j|0}k[24730]=0;u=k[30280]|0;h=k[30281]|0;o=u;d:do if(!u)if((h|0)==7){k[24730]=7;break}else{k[24730]=6;break}else{if((h|0)==2){k[24730]=2;k[24750]=o;break}q=i[u>>0]|0;n=(q+-97&255)<26?(q&255)+224&255:q;e:do if(n<<24>>24<=84){q=u;a=767080;p=84;l=n;while(1){if(p<<24>>24>l<<24>>24){m=41;break}if(!(p<<24>>24))break;a=a+1|0;q=q+1|0;s=i[a>>0]|0;p=(s+-97&255)<26?(s&255)+224&255:s;s=i[q>>0]|0;l=(s+-97&255)<26?(s&255)+224&255:s;if(p<<24>>24<l<<24>>24){m=41;break}}f:do if((m|0)==41){if(n<<24>>24>70)break e;else{v=u;w=767082;x=70;y=n}while(1){if(x<<24>>24>y<<24>>24)break e;if(!(x<<24>>24))break f;w=w+1|0;v=v+1|0;l=i[w>>0]|0;x=(l+-97&255)<26?(l&255)+224&255:l;l=i[v>>0]|0;y=(l+-97&255)<26?(l&255)+224&255:l;if(x<<24>>24<y<<24>>24)break e}}while(0);k[24730]=1;l=i[u>>0]|0;p=(l+-97&255)<26?(l&255)+224&255:l;g:do if(p<<24>>24>84)z=0;else{l=u;q=767080;a=84;s=p;while(1){if(a<<24>>24>s<<24>>24){z=0;break g}if(!(a<<24>>24)){z=1;break g}q=q+1|0;l=l+1|0;t=i[q>>0]|0;a=(t+-97&255)<26?(t&255)+224&255:t;t=i[l>>0]|0;s=(t+-97&255)<26?(t&255)+224&255:t;if(a<<24>>24<s<<24>>24){z=0;break}}}while(0);i[99e3]=z;break d}while(0);k[f>>2]=99e3;k[f+4>>2]=99008;k[f+8>>2]=g;h:do if((Yw(u,767084,f)|0)==2){switch(i[(k[30280]|0)+(k[g>>2]|0)>>0]|0){case 0:case 10:case 9:case 32:break;default:break h}k[24730]=5;break d}while(0);if(!(k[24730]|0)){n=k[30280]|0;i:do if((rx(n,46)|0)!=0?(k[e>>2]=99e3,k[e+4>>2]=g,(Yw(n,767096,e)|0)==1):0){p=k[g>>2]|0;s=(k[30280]|0)+p|0;switch(i[s>>0]|0){case 68:break;case 0:case 10:case 9:case 32:{k[24730]=4;break d;break}default:break i}i[s>>0]=69;s=k[30280]|0;k[d>>2]=99e3;k[d+4>>2]=g;Yw(s,767096,d)|0;s=k[30280]|0;switch(i[s+(k[g>>2]|0)>>0]|0){case 0:case 10:case 9:case 32:{k[24730]=4;break d;break}default:{i[s+p>>0]=68;break i}}}while(0);if(!(k[24730]|0)){n=k[30280]|0;k[c>>2]=99e3;k[c+4>>2]=g;j:do if((Yw(n,767102,c)|0)==1){switch(i[(k[30280]|0)+(k[g>>2]|0)>>0]|0){case 0:case 10:case 9:case 32:break;default:break j}k[24730]=3;break d}while(0);if(!(k[24730]|0)){k[24730]=2;k[24750]=k[30280]}}}}while(0);g=k[30282]|0;if(!g)i[99016]=0;else{Bx(99016,g,80)|0;i[99095]=0}Bx(98924,k[30279]|0,75)|0;i[98998]=0;g=(xx(98924)|0)>>>0>75;j=g?368:0;r=b;return j|0}function Jp(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;a:do if(a){b=a+4|0;if((((((((((tx(782785,b)|0)!=0?(tx(782702,b)|0)!=0:0)?(tx(779656,b)|0)!=0:0)?(tx(782713,b)|0)!=0:0)?(tx(782652,b)|0)!=0:0)?(tx(782726,b)|0)!=0:0)?(tx(782733,b)|0)!=0:0)?(tx(782755,b)|0)!=0:0)?(tx(782808,b)|0)!=0:0)?(tx(767107,b)|0)!=0:0){if(((Ax(782713,b,5)|0)!=0?(Ax(773003,b,5)|0)!=0:0)?(Ax(772997,b,5)|0)!=0:0){c=0;break}if(((i[a+9>>0]|0)+-49&255)>8)c=0;else{b=5;while(1){d=b+1|0;if((d|0)>=8){c=368;break a}e=i[a+4+d>>0]|0;if((e+-48&255)<10)b=d;else{f=e;g=d;break}}switch(f<<24>>24){case 0:{c=368;break a;break}case 32:{h=g;break}default:{c=0;break a}}while(1){h=h+1|0;if((h|0)>=8){c=368;break a}if((i[a+4+h>>0]|0)!=32){c=0;break}}}}else c=368}else c=362;while(0);return c|0}function Kp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;d=r;r=r+240|0;e=d+8|0;f=d;g=d+24|0;h=d+20|0;j=d+32|0;l=d+16|0;if((a|0)==0|(b|0)==0){m=362;r=d;return m|0}k[g>>2]=0;n=k[a>>2]|0;a:do if((n|0)>0){o=a+4|0;if(!(c&1)){p=n;q=0}else{s=0;while(1){t=k[o>>2]|0;u=Jp(t+(s*176|0)|0)|0;k[g>>2]=u;b:do switch(k[t+(s*176|0)>>2]|0){case 1:{k[h>>2]=i[t+(s*176|0)+80>>0];es(b,14,t+(s*176|0)+4|0,h,t+(s*176|0)+96|0,g)|0;v=16;break}case 2:{ps(b,t+(s*176|0)+4|0,k[t+(s*176|0)+80>>2]|0,t+(s*176|0)+96|0,g)|0;v=16;break}case 3:{k[l>>2]=k[t+(s*176|0)+80>>2];es(b,41,t+(s*176|0)+4|0,l,t+(s*176|0)+96|0,g)|0;v=16;break}case 4:{es(b,82,t+(s*176|0)+4|0,t+(s*176|0)+80|0,t+(s*176|0)+96|0,g)|0;v=16;break}case 5:{es(b,163,t+(s*176|0)+4|0,t+(s*176|0)+80|0,t+(s*176|0)+96|0,g)|0;v=16;break}case 6:{ns(b,t+(s*176|0)+4|0,t+(s*176|0)+96|0,g)|0;v=16;break}case 7:{w=t+(s*176|0)+4|0;if(!(tx(782840,w)|0)){xs(b,t+(s*176|0)+96|0,g)|0;v=16;break b}x=t+(s*176|0)+96|0;if(!(tx(782848,w)|0)){rs(b,x,g)|0;v=16;break b}else{k[f>>2]=w;k[f+4>>2]=x;Xw(j,767114,f)|0;ds(b,j,g)|0;v=16;break b}break}default:y=u}while(0);if((v|0)==16){v=0;y=k[g>>2]|0}if(y){m=y;break}s=s+1|0;if((s|0)>=(k[a>>2]|0))break a}r=d;return m|0}while(1){s=k[o>>2]|0;u=Jp(s+(q*176|0)|0)|0;k[g>>2]=u;c:do if(u){k[g>>2]=0;if((u|0)==368){t=s+(q*176|0)+96|0;if(!(i[t>>0]|0))v=35;else{zq(b,s+(q*176|0)+4|0,t,g)|0;v=35}}else z=p}else switch(k[s+(q*176|0)>>2]|0){case 1:{k[h>>2]=i[s+(q*176|0)+80>>0];es(b,14,s+(q*176|0)+4|0,h,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 2:{ps(b,s+(q*176|0)+4|0,k[s+(q*176|0)+80>>2]|0,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 3:{k[l>>2]=k[s+(q*176|0)+80>>2];es(b,41,s+(q*176|0)+4|0,l,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 4:{es(b,82,s+(q*176|0)+4|0,s+(q*176|0)+80|0,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 5:{es(b,163,s+(q*176|0)+4|0,s+(q*176|0)+80|0,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 6:{ns(b,s+(q*176|0)+4|0,s+(q*176|0)+96|0,g)|0;v=35;break c;break}case 7:{t=s+(q*176|0)+4|0;if(!(tx(782840,t)|0)){xs(b,s+(q*176|0)+96|0,g)|0;v=35;break c}x=s+(q*176|0)+96|0;if(!(tx(782848,t)|0)){rs(b,x,g)|0;v=35;break c}else{k[e>>2]=t;k[e+4>>2]=x;Xw(j,767114,e)|0;ds(b,j,g)|0;v=35;break c}break}default:{v=35;break c}}while(0);if((v|0)==35){v=0;s=k[g>>2]|0;if(s){m=s;break}z=k[a>>2]|0}q=q+1|0;if((q|0)>=(z|0))break a;else p=z}r=d;return m|0}while(0);Vk(b,g)|0;m=k[g>>2]|0;r=d;return m|0}function Lp(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;if((a|0)==0|(b|0)==0){c=362;return c|0}d=k[a>>2]|0;if(!d)e=cy(176)|0;else e=fy(k[a+4>>2]|0,(d*176|0)+176|0)|0;if(!e){c=360;return c|0}d=a+4|0;k[d>>2]=e;f=k[a>>2]|0;py(e+(f*176|0)|0,b|0,176)|0;do if((k[b>>2]|0)==2?(g=b+80|0,h=k[g>>2]|0,(h|0)!=0):0){i=cy((xx(h)|0)+1|0)|0;k[e+(f*176|0)+80>>2]=i;i=k[(k[d>>2]|0)+(f*176|0)+80>>2]|0;if(!i){c=360;return c|0}else{ux(i,k[g>>2]|0)|0;j=k[a>>2]|0;break}}else j=f;while(0);k[a>>2]=j+1;c=0;return c|0}function Mp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;d=r;r=r+32|0;e=d+8|0;f=d;g=d+20|0;h=d+16|0;j=d+24|0;if((a|0)==0|(b|0)==0){l=362;r=d;return l|0}if(!(k[b>>2]|0)){l=0;r=d;return l|0}k[g>>2]=0;if((c|0)>=999){l=0;r=d;return l|0}m=b+4|0;n=0;o=c;while(1){o=o+1|0;c=0;p=0;q=795593;while(1){s=(k[m>>2]|0)+(c*176|0)+4|0;k[f>>2]=h;k[f+4>>2]=j;t=(Yw(s,767123,f)|0)==1;s=k[m>>2]|0;if(t)if((k[h>>2]|0)==(o|0)?(k[s+(c*176|0)>>2]|0)==2:0){u=k[s+(c*176|0)+80>>2]|0;v=q}else{u=p;v=q}else{k[e>>2]=h;k[e+4>>2]=j;if((Yw(s+(c*176|0)+4|0,767133,e)|0)==1?(s=k[m>>2]|0,(k[h>>2]|0)==(o|0)?(k[s+(c*176|0)>>2]|0)==2:0):0){u=p;v=k[s+(c*176|0)+80>>2]|0}else{u=p;v=q}}s=(u|0)!=0;if(s?(i[v>>0]|0)!=0:0){w=1;x=n;y=u;z=v;break}if((c|0)<((k[b>>2]|0)+-1|0)){c=c+1|0;p=u;q=v}else{w=s;x=1;y=u;z=v;break}}q=k[g>>2]|0;if(w&(q|0)==0){Qj(a,o,z,y,g)|0;A=k[g>>2]|0}else A=q;if(!((A|x|0)==0&(o|0)<999)){l=A;break}else n=x}r=d;return l|0}function Np(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0;d=r;r=r+4144|0;e=d+8|0;f=d;g=d+4044|0;h=d+4040|0;j=d+40|0;l=d+36|0;m=d+32|0;n=d+4056|0;o=d+4048|0;p=d+44|0;q=d+24|0;s=d+16|0;i[n>>0]=0;t=q+4|0;k[t>>2]=0;k[q>>2]=0;u=Ip(0)|0;k[g>>2]=u;if(u){v=u;r=d;return v|0}u=k[30294]|0;if(!(c&1)){if((u|0)!=3){v=369;r=d;return v|0}}else if((u|0)!=4){v=369;r=d;return v|0}u=Lp(q,98920)|0;k[g>>2]=u;if(!u)w=0;else{v=u;r=d;return v|0}a:while(1){u=Ip(0)|0;k[g>>2]=u;if(u){v=u;x=73;break}switch(k[30294]|0){case 4:{x=8;break a;break}case 1:case 3:case 2:{x=9;break a;break}case 5:{x=21;break a;break}default:{}}u=xx(98924)|0;if((u+-2|0)>>>0<5?(y=u+-1|0,u=98924+y|0,(i[u>>0]|0)==35):0){if(!(i[n>>0]|0)){py(n|0,98924,y|0)|0;i[n+y>>0]=0}if((y|0)==(xx(n)|0))z=((lx(n,98924,y)|0)==0&1)+w|0;else z=w;k[f>>2]=z;Xw(u,772183,f)|0;A=z}else A=w;u=Lp(q,98920)|0;k[g>>2]=u;if(!u)w=A;else{x=64;break}}do if((x|0)==8){k[g>>2]=369;x=64}else if((x|0)==9){if(k[30278]|0)if(!(k[30285]|0)){k[30285]=k[30278];k[30286]=k[30279];k[30287]=k[30280];k[30288]=k[30281];k[30289]=k[30282];k[30290]=k[30283];k[30291]=k[30284];k[30278]=0;k[g>>2]=0;x=21;break}else B=364;else B=363;k[g>>2]=B;x=64}else if((x|0)==73){r=d;return v|0}while(0);b:do if((x|0)==21){sy(p|0,0,3996)|0;c:do if((k[q>>2]|0)>0){B=0;A=0;w=0;z=0;f=0;while(1){n=k[t>>2]|0;u=n+(B*176|0)+4|0;d:do if(!(tx(782702,u)|0))if((k[n+(B*176|0)>>2]|0)==2){y=k[n+(B*176|0)+80>>2]|0;C=y;D=772762;E=0;while(1){F=i[D>>0]|0;G=(F+-97&255)<26?(F&255)+224&255:F;F=i[C>>0]|0;if(G<<24>>24!=((F+-97&255)<26?(F&255)+224&255:F)<<24>>24){H=f;break}if(!(G<<24>>24)){H=3;break}E=E+1|0;if((E|0)>=8){H=3;break}else{C=C+1|0;D=D+1|0}}D=i[y>>0]|0;switch(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24){case 84:{D=i[y+1>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=65){I=A;J=w;K=z;L=H;break d}D=i[y+2>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=66){I=A;J=w;K=z;L=H;break d}D=i[y+3>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=76){I=A;J=w;K=z;L=H;break d}D=i[y+4>>0]|0;I=A;J=w;K=z;L=((D+-97&255)<26?(D&255)+224&255:D)<<24>>24==69?2:H;break d;break}case 73:{D=i[y+1>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=77){I=A;J=w;K=z;L=H;break d}D=i[y+2>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=65){I=A;J=w;K=z;L=H;break d}D=i[y+3>>0]|0;if(((D+-97&255)<26?(D&255)+224&255:D)<<24>>24!=71){I=A;J=w;K=z;L=H;break d}D=i[y+4>>0]|0;I=A;J=w;K=z;L=((D+-97&255)<26?(D&255)+224&255:D)<<24>>24==69?1:H;break d;break}default:{I=A;J=w;K=z;L=H;break d}}}else{I=A;J=w;K=z;L=f}else{if(!(tx(782785,u)|0)){if((k[n+(B*176|0)>>2]|0)!=1){I=A;J=w;K=z;L=f;break}I=A;J=w;K=z;L=(i[n+(B*176|0)+80>>0]|0)==0?f:1;break}if(!(tx(779656,u)|0)){if((k[n+(B*176|0)>>2]|0)!=3){I=A;J=w;K=z;L=f;break}I=k[n+(B*176|0)+80>>2]|0;J=w;K=z;L=f;break}if(!(tx(782713,u)|0)){if((k[n+(B*176|0)>>2]|0)!=3){I=A;J=w;K=z;L=f;break}I=A;J=k[n+(B*176|0)+80>>2]|0;K=z;L=f;break}if(!(tx(790194,u)|0)){if((k[n+(B*176|0)>>2]|0)!=2){I=A;J=w;K=z;L=f;break}I=A;J=w;K=k[n+(B*176|0)+80>>2]|0;L=f;break}k[e>>2]=l;k[e+4>>2]=o;if(((Yw(u,767143,e)|0)==1?(D=k[t>>2]|0,(k[D+(B*176|0)>>2]|0)==3):0)?(C=(k[l>>2]|0)+-1|0,C>>>0<999):0){k[p+(C<<2)>>2]=k[D+(B*176|0)+80>>2];I=A;J=w;K=z;L=f}else{I=A;J=w;K=z;L=f}}while(0);B=B+1|0;if((B|0)>=(k[q>>2]|0)){M=I;N=J;O=K;P=L;break}else{A=I;w=J;z=K;f=L}}switch(P|0){case 1:{if((c&3|0)==2)Xr(a,16,0,0,g)|0;Xr(a,M,N,p,g)|0;if(k[g>>2]|0){x=64;break b}f=Kp(q,a,0)|0;k[g>>2]=f;Q=f;break}case 3:case 2:{$r(a,(P|0)==2?1:2,0,0,0,0,0,0,0,g)|0;if(k[g>>2]|0){x=64;break b}f=Mp(a,q,0)|0;k[g>>2]=f;if(f){x=64;break b}f=Kp(q,a,0)|0;k[g>>2]=f;if(f){x=64;break b}f=k[p+4>>2]|0;if((f|0)>0){Nj(a,0,0,f,((f|0)<0)<<31>>31,g)|0;Q=k[g>>2]|0}else Q=0;break}default:break c}if((O|0)!=0&(Q|0)==0){f=Dp(O,m)|0;k[g>>2]=f;k[s>>2]=k[m>>2];es(a,41,767107,s,767153,g)|0;R=k[g>>2]|0}else R=Q;f=(R|0)==0;if(!((b|0)>0&f))if(f)break b;else{x=64;break b}bl(a,h)|0;Fk(a,b,j,g)|0;qp(a,0,k[h>>2]|0,g)|0;Fk(a,k[h>>2]|0,j,g)|0;f=k[g>>2]|0;if(!f)break b;else v=f;r=d;return v|0}while(0);k[g>>2]=368;x=64}while(0);if((x|0)==64){k[j>>2]=0;_j(a,0,j)|0}j=k[q>>2]|0;a=k[t>>2]|0;if((j|0)>0){x=a;h=j;j=0;while(1){if((k[x+(j*176|0)>>2]|0)==2?(b=k[x+(j*176|0)+80>>2]|0,(b|0)!=0):0){dy(b);k[(k[t>>2]|0)+(j*176|0)+80>>2]=0;S=k[q>>2]|0;T=k[t>>2]|0}else{S=h;T=x}j=j+1|0;if((j|0)>=(S|0)){U=T;break}else{x=T;h=S}}}else U=a;if(U)dy(U);k[t>>2]=0;k[q>>2]=0;v=k[g>>2]|0;r=d;return v|0}function Op(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;d=r;r=r+208|0;e=d+8|0;f=d;g=d+32|0;h=d+28|0;j=d+24|0;l=d+120|0;m=d+40|0;n=d+16|0;i[m>>0]=0;k[30293]=(k[30293]|0)+1;o=n+4|0;k[o>>2]=0;k[n>>2]=0;k[g>>2]=0;p=gp(a,b,0,g)|0;k[g>>2]=p;if(p){q=p;r=d;return q|0}bl(a,h)|0;if((c|0)>0?(Fk(a,c,j,g)|0,qp(a,0,k[h>>2]|0,g)|0,Fk(a,k[h>>2]|0,j,g)|0,c=k[g>>2]|0,(c|0)!=0):0){q=c;r=d;return q|0}c=l+79|0;p=0;b=6;while(1){s=Ip(0)|0;k[g>>2]=s;if(s)break;a:do switch(k[30294]|0){case 5:case 4:{k[g>>2]=369;t=369;u=p;v=b;break}case 2:{k[30293]=(k[30293]|0)+-1;t=0;u=1;v=b;break}case 1:{if((k[24730]|0)==2)Bx(l,k[24750]|0,80)|0;else{s=k[30307]|0;k[30307]=s+1;k[f>>2]=s;Xw(l,767186,f)|0}i[c>>0]=0;s=Op(a,l,k[h>>2]|0)|0;k[g>>2]=s;t=s;u=p;v=b;break}case 3:{if(k[30278]|0)if(!(k[30285]|0)){k[30285]=k[30278];k[30286]=k[30279];k[30287]=k[30280];k[30288]=k[30281];k[30289]=k[30282];k[30290]=k[30283];k[30291]=k[30284];k[30278]=0;k[g>>2]=0;s=Np(a,k[h>>2]|0,0)|0;k[g>>2]=s;t=s;u=p;v=b;break a}else w=364;else w=363;k[g>>2]=w;t=w;u=p;v=b;break}default:{s=xx(98924)|0;if((s+-2|0)>>>0<5?(x=s+-1|0,s=98924+x|0,(i[s>>0]|0)==35):0){if(!(i[m>>0]|0)){py(m|0,98924,x|0)|0;i[m+x>>0]=0}if((x|0)==(xx(m)|0))y=((lx(m,98924,x)|0)==0&1)+b|0;else y=b;k[e>>2]=y;Xw(s,772183,e)|0;z=y}else z=b;s=Lp(n,98920)|0;k[g>>2]=s;t=s;u=p;v=z}}while(0);if(!(t|u)){p=u;b=v}else break}Fk(a,k[h>>2]|0,j,g)|0;if(!(((k[g>>2]|0)==0?(h=Mp(a,n,6)|0,k[g>>2]=h,(h|0)==0):0)?(h=Kp(n,a,0)|0,k[g>>2]=h,(h|0)==0):0)){k[j>>2]=0;kp(a,0,j)|0}j=k[n>>2]|0;a=k[o>>2]|0;if((j|0)>0){h=a;v=j;j=0;while(1){if((k[h+(j*176|0)>>2]|0)==2?(b=k[h+(j*176|0)+80>>2]|0,(b|0)!=0):0){dy(b);k[(k[o>>2]|0)+(j*176|0)+80>>2]=0;A=k[n>>2]|0;B=k[o>>2]|0}else{A=v;B=h}j=j+1|0;if((j|0)>=(A|0)){C=B;break}else{h=B;v=A}}}else C=a;if(C)dy(C);k[o>>2]=0;k[n>>2]=0;q=k[g>>2]|0;r=d;return q|0}function Pp(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0;d=r;r=r+192|0;e=d;f=d+20|0;g=d+16|0;h=d+12|0;j=d+8|0;l=d+104|0;m=d+24|0;n=d+4|0;if(!c){o=362;r=d;return o|0}p=k[c>>2]|0;if(p){o=p;r=d;return o|0}if((a|0)==0|(b|0)==0){k[c>>2]=362;o=362;r=d;return o|0}k[30292]=0;k[30293]=0;k[30307]=1;i[766038]=0;p=k[30309]|0;q=(p|0)==0;s=k[30308]|0;t=(s|0)>0;if(!(q&t)?!((p|0)!=0&(s|0)<1):0){if(!(q&(s|0)==0)){if(t){t=p;s=p;q=0;while(1){u=k[s+(q<<3)>>2]|0;if(!u){v=t;w=s}else{dy(u);k[(k[30309]|0)+(q<<3)>>2]=0;u=k[30309]|0;v=u;w=u}k[w+(q<<3)+4>>2]=0;q=q+1|0;if((q|0)>=(k[30308]|0)){x=v;break}else{t=v;s=w}}}else x=p;dy(x);k[30309]=0;k[30308]=0}bl(a,f)|0;do if((k[f>>2]|0)<2){Fk(a,1,g,c)|0;Co(a,h,j,c)|0;Fk(a,k[f>>2]|0,g,c)|0;x=k[c>>2]|0;if(!x){y=(k[h>>2]|0)<1&1;break}else{o=x;r=d;return o|0}}else{x=2;while(1){k[c>>2]=0;Fk(a,1,g,c)|0;if(k[c>>2]|0){z=24;break}Io(a,16,790194,m,0,c)|0;a:do if(!(k[c>>2]|0)){Io(a,41,767107,n,0,c)|0;p=k[n>>2]|0;switch(k[c>>2]|0){case 204:{k[c>>2]=0;A=1;break}case 0:{A=p;break}default:break a}p=Ep(m,A)|0;k[c>>2]=p}while(0);p=k[f>>2]|0;if((x|0)<(p|0))x=x+1|0;else{B=p;break}}if((z|0)==24)B=k[f>>2]|0;Fk(a,B,g,c)|0;x=k[c>>2]|0;if(!x)y=0;else{o=x;r=d;return o|0}}while(0);g=Hp(b)|0;k[c>>2]=g;if(g){o=g;r=d;return o|0}g=xx(b)|0;while(1){B=g+-1|0;if((B|0)<=-1){C=g;break}if((i[b+B>>0]|0)==47){C=g;break}else g=B}g=(C|0)>999?999:C;if((g|0)>0){py(766038,b|0,g|0)|0;i[766038+g>>0]=0}g=l+79|0;b=0;C=y;while(1){y=Ip(1)|0;if(y){D=y;break}switch(k[30294]|0){case 4:{if(C)if(k[30278]|0)if(!(k[30285]|0)){k[30285]=k[30278];k[30286]=k[30279];k[30287]=k[30280];k[30288]=k[30281];k[30289]=k[30282];k[30290]=k[30283];k[30291]=k[30284];k[30278]=0;E=b;F=0;G=Np(a,0,3)|0}else{E=b;F=C;G=364}else{E=b;F=C;G=363}else{E=b;F=0;G=369}break}case 3:{if(k[30278]|0)if(!(k[30285]|0)){k[30285]=k[30278];k[30286]=k[30279];k[30287]=k[30280];k[30288]=k[30281];k[30289]=k[30282];k[30290]=k[30283];k[30291]=k[30284];k[30278]=0;E=b;F=0;G=Np(a,0,(C|0)!=0?2:0)|0}else{E=b;F=C;G=364}else{E=b;F=C;G=363}break}case 1:{if((k[24730]|0)==2)Bx(l,k[24750]|0,80)|0;else{y=k[30307]|0;k[30307]=y+1;k[e>>2]=y;Xw(l,767186,e)|0}i[g>>0]=0;E=b;F=0;G=Op(a,l,0)|0;break}case 5:{E=1;F=C;G=0;break}default:{E=b;F=C;G=369}}if(!(E|G)){b=E;C=F}else{D=G;break}}G=k[30278]|0;if(G){dy(G);k[30278]=0;k[30279]=0;k[30280]=0;k[30281]=0;k[30282]=0;k[30283]=0;k[30284]=0}G=k[30285]|0;if(G){dy(G);k[30285]=0;k[30286]=0;k[30287]=0;k[30288]=0;k[30289]=0;k[30290]=0;k[30291]=0}G=k[30309]|0;F=(G|0)==0;C=k[30308]|0;E=(C|0)>0;if(!(F&E)?!((G|0)!=0&(C|0)<1|F&(C|0)==0):0){if(E){E=G;C=G;F=0;while(1){b=k[C+(F<<3)>>2]|0;if(!b){H=E;I=C}else{dy(b);k[(k[30309]|0)+(F<<3)>>2]=0;b=k[30309]|0;H=b;I=b}k[I+(F<<3)+4>>2]=0;F=F+1|0;if((F|0)>=(k[30308]|0)){J=H;break}else{E=H;C=I}}}else J=G;dy(J);k[30309]=0;k[30308]=0}k[c>>2]=D;o=D;r=d;return o|0}k[c>>2]=368;o=368;r=d;return o|0}function Qp(a,b,c,d,e,f,g,h,j,l,m,n,o,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;var s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0;s=r;r=r+48|0;t=s+12|0;u=s+16|0;v=s+8|0;w=s;k[v>>2]=0;x=k[q>>2]|0;if((x|0)>0){y=x;r=s;return y|0}k[c>>2]=2;k[b>>2]=31;p[m>>3]=1.0;k[o>>2]=0;i[n>>0]=0;i[d>>0]=0;i[h>>0]=0;i[j>>0]=0;i[l>>0]=0;p[e>>3]=-9.1191291391491e-36;p[f>>3]=-9.1191291391491e-36;p[g>>3]=-9.1191291391491e-36;o=d+71|0;i[o>>0]=0;x=h+71|0;i[x>>0]=0;z=j+71|0;i[z>>0]=0;A=l+71|0;i[A>>0]=0;B=e+8|0;p[B>>3]=-9.1191291391491e-36;C=f+8|0;p[C>>3]=-9.1191291391491e-36;D=g+8|0;p[D>>3]=-9.1191291391491e-36;E=d+142|0;i[E>>0]=0;F=h+142|0;i[F>>0]=0;G=j+142|0;i[G>>0]=0;H=l+142|0;i[H>>0]=0;I=e+16|0;p[I>>3]=-9.1191291391491e-36;J=f+16|0;p[J>>3]=-9.1191291391491e-36;K=g+16|0;p[K>>3]=-9.1191291391491e-36;L=d+213|0;i[L>>0]=0;M=h+213|0;i[M>>0]=0;N=j+213|0;i[N>>0]=0;O=l+213|0;i[O>>0]=0;P=e+24|0;p[P>>3]=-9.1191291391491e-36;Q=f+24|0;p[Q>>3]=-9.1191291391491e-36;R=g+24|0;p[R>>3]=-9.1191291391491e-36;S=a+3|0;k[t>>2]=S;switch(i[S>>0]|0){case 105:{k[b>>2]=21;T=a+4|0;k[t>>2]=T;U=T;V=0;break}case 106:{T=a+4|0;k[t>>2]=T;U=T;V=0;break}case 114:{k[b>>2]=42;T=a+4|0;k[t>>2]=T;U=T;V=0;break}case 100:{k[b>>2]=82;T=a+4|0;k[t>>2]=T;U=T;V=0;break}case 98:{k[b>>2]=11;T=a+4|0;k[t>>2]=T;U=T;V=0;break}default:{U=S;V=1}}switch(i[U>>0]|0){case 0:{y=k[q>>2]|0;r=s;return y|0}case 32:{S=32;T=U;a:while(1){switch(S<<24>>24){case 0:{W=13;break a;break}case 64:{X=T;W=14;break a;break}case 32:break;default:{Y=T;Z=S;break a}}U=T+1|0;k[t>>2]=U;S=i[U>>0]|0;T=U}if((W|0)==13){y=k[q>>2]|0;r=s;return y|0}do if((W|0)==14)if(!(Uh(X+1|0,v,q)|0)){T=k[v>>2]|0;k[t>>2]=T;S=T;T=i[S>>0]|0;if(T<<24>>24==32)_=S;else{Y=S;Z=T;break}while(1){T=_+1|0;k[t>>2]=T;S=i[T>>0]|0;if(S<<24>>24==32)_=T;else{Y=T;Z=S;break}}}else{y=k[q>>2]|0;r=s;return y|0}while(0);b:do if(Z<<24>>24==40){_=Y;X=0;while(1){if((X|0)<4)$=_;else{aa=_;ba=X;break}while(1){S=$+1|0;k[t>>2]=S;if((i[S>>0]|0)==32)$=S;else{ca=S;break}}S=vx(ca,767251)|0;zx(d+(X*71|0)|0,ca,S)|0;T=(k[t>>2]|0)+S|0;while(1){k[t>>2]=T;S=i[T>>0]|0;if(S<<24>>24==32)T=T+1|0;else{da=S;ea=T;break}}T=X+1|0;if(da<<24>>24==41){fa=T;ga=X;ha=ea;W=25;break}else{_=ea;X=T}}if((W|0)==25){k[c>>2]=fa;aa=ha;ba=ga}if((ba|0)==4){tk(767255);tk(a);X=k[v>>2]|0;if(X)dy(X);k[q>>2]=125;y=125;r=s;return y|0}else ia=aa;c:while(1){X=ia+1|0;k[t>>2]=X;switch(i[X>>0]|0){case 32:{ia=X;break}case 0:{W=31;break c;break}case 61:{ja=ia;W=37;break c;break}default:{W=34;break c}}}if((W|0)==31){X=k[v>>2]|0;if(X)dy(X);y=k[q>>2]|0;r=s;return y|0}else if((W|0)==34){tk(767330);tk(767368);tk(a);X=k[v>>2]|0;if(X)dy(X);k[q>>2]=125;y=125;r=s;return y|0}else if((W|0)==37){X=ja+2|0;while(1){k[t>>2]=X;if((i[X>>0]|0)==32)X=X+1|0;else break}Rp(t,u,e,f,g,h,j,l,q)|0;if((k[q>>2]|0)>0){tk(767330);tk(a);X=k[v>>2]|0;if(X)dy(X);y=k[q>>2]|0;r=s;return y|0}if((k[c>>2]|0)>1){X=1;do{p[e+(X<<3)>>3]=+p[e>>3];p[f+(X<<3)>>3]=+p[f>>3];p[g+(X<<3)>>3]=+p[g>>3];ux(h+(X*71|0)|0,h)|0;ux(j+(X*71|0)|0,j)|0;ux(l+(X*71|0)|0,l)|0;X=X+1|0}while((X|0)<(k[c>>2]|0))}X=k[t>>2]|0;d:while(1){switch(i[X>>0]|0){case 59:{ka=X;break b;break}case 0:{W=51;break d;break}case 32:break;default:break d}_=X+1|0;k[t>>2]=_;X=_}if((W|0)==51){y=k[q>>2]|0;r=s;return y|0}tk(767417);tk(a);X=k[v>>2]|0;if(X)dy(X);k[q>>2]=125;y=125;r=s;return y|0}}else{Rp(t,d,e,f,g,h,j,l,q)|0;e:do if((k[q>>2]|0)<=0){X=k[t>>2]|0;f:do switch(i[X>>0]|0){case 59:case 0:{la=X;W=67;break}case 32:{_=32;T=X;g:while(1){switch(_<<24>>24){case 59:case 0:{la=T;W=67;break f;break}case 44:{ma=T;break g;break}case 32:break;default:{W=63;break f}}S=T+1|0;k[t>>2]=S;_=i[S>>0]|0;T=S}k[t>>2]=ma+1;W=63;break}case 44:{k[t>>2]=X+1;W=63;break}default:{}}while(0);if((W|0)==67){k[c>>2]=1;if(i[d>>0]|0){ka=la;break b}if(!(+p[e>>3]==-9.1191291391491e-36)){ka=la;break b}if(!(+p[f>>3]==-9.1191291391491e-36)){ka=la;break b}k[c>>2]=2;p[D>>3]=+p[g>>3];ka=la;break b}h:do if((W|0)==63){Rp(t,o,B,C,D,x,z,A,q)|0;if((k[q>>2]|0)>0)break e;X=k[t>>2]|0;i:do switch(i[X>>0]|0){case 59:case 0:{na=X;oa=2;W=65;break}case 32:{T=32;_=X;j:while(1){switch(T<<24>>24){case 59:case 0:{pa=_;qa=2;W=66;break i;break}case 44:{ra=_;break j;break}case 32:break;default:{W=95;break i}}S=_+1|0;k[t>>2]=S;T=i[S>>0]|0;_=S}k[t>>2]=ra+1;W=95;break}case 44:{k[t>>2]=X+1;W=95;break}default:break h}while(0);k:do if((W|0)==95){Rp(t,E,I,J,K,F,G,H,q)|0;if((k[q>>2]|0)>0)break e;X=k[t>>2]|0;l:do switch(i[X>>0]|0){case 59:case 0:{na=X;oa=3;W=65;break k;break}case 32:{_=32;T=X;m:while(1){switch(_<<24>>24){case 59:case 0:{pa=T;qa=3;W=66;break k;break}case 44:{sa=T;break m;break}case 32:break;default:break l}S=T+1|0;k[t>>2]=S;_=i[S>>0]|0;T=S}k[t>>2]=sa+1;break}case 44:{k[t>>2]=X+1;break}default:break h}while(0);Rp(t,L,P,Q,R,M,N,O,q)|0;if((k[q>>2]|0)>0)break e;X=k[t>>2]|0;n:do switch(i[X>>0]|0){case 59:case 0:{na=X;oa=4;W=65;break k;break}case 32:{T=32;_=X;o:while(1){switch(T<<24>>24){case 59:case 0:{pa=_;qa=4;W=66;break k;break}case 44:{ta=_;break o;break}case 32:break;default:break n}S=_+1|0;k[t>>2]=S;T=i[S>>0]|0;_=S}k[t>>2]=ta+1;break}case 44:{k[t>>2]=X+1;break}default:break h}while(0);tk(767330);tk(767587);tk(a);k[q>>2]=125;y=125;r=s;return y|0}while(0);if((W|0)==65){k[c>>2]=oa;ka=na;break b}else if((W|0)==66){k[c>>2]=qa;ka=pa;break b}}while(0);tk(767528);tk(a);X=k[v>>2]|0;if(X)dy(X);k[q>>2]=125;y=125;r=s;return y|0}while(0);tk(767474);tk(a);X=k[v>>2]|0;if(X)dy(X);y=k[q>>2]|0;r=s;return y|0}while(0);do if((i[ka>>0]|0)==59){pa=ka;p:while(1){qa=pa+1|0;k[t>>2]=qa;switch(i[qa>>0]|0){case 32:{pa=qa;break}case 47:{W=73;break p;break}default:break p}}if((W|0)==73)Hb();Rp(t,n,w,w,m,u,u,u,q)|0;if((k[q>>2]|0)>0){tk(767634);tk(a);pa=k[v>>2]|0;if(pa)dy(pa);y=k[q>>2]|0;r=s;return y|0}else{if(!V)break;if(!(+p[m>>3]!=1.0)?(i[n>>0]|0)==0:0)break;k[b>>2]=42;break}}while(0);b=k[t>>2]|0;q:while(1){switch(i[b>>0]|0){case 0:{break q;break}case 32:break;default:{W=85;break q}}n=b+1|0;k[t>>2]=n;b=n}if((W|0)==85){tk(767679);tk(a);k[q>>2]=125}W=k[v>>2]|0;if(W)dy(W);y=k[q>>2]|0;r=s;return y|0}default:{tk(767215);tk(a);k[q>>2]=125;y=125;r=s;return y|0}}return 0}function Rp(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0.0;l=r;r=r+80|0;m=l;n=l+4|0;o=k[j>>2]|0;if((o|0)>0){q=o;r=l;return q|0}o=Vh(a,767737,n,m)|0;a:do if(!o){switch(i[k[a>>2]>>0]|0){case 59:case 44:case 0:break;default:break a}q=k[j>>2]|0;r=l;return q|0}while(0);do if((k[m>>2]|0)==0?(i[k[a>>2]>>0]|0)!=58:0){if((i[n>>0]|0)==35?(s=n+1|0,((i[s>>0]|0)+-48|0)>>>0<10):0)ux(b,s)|0;else ux(b,n)|0;s=k[a>>2]|0;b:while(1){switch(i[s>>0]|0){case 61:{t=s;u=15;break b;break}case 32:break;default:{u=14;break b}}v=s+1|0;k[a>>2]=v;s=v}if((u|0)==14){q=k[j>>2]|0;r=l;return q|0}else if((u|0)==15){while(1){u=0;t=t+1|0;k[a>>2]=t;if((i[t>>0]|0)!=32)break;else u=15}w=Vh(a,767743,n,m)|0;break}}else w=o;while(0);if((i[k[a>>2]>>0]|0)!=58){if(!(k[m>>2]|0))ux(h,n)|0;else{x=+hx(n,0);p[e>>3]=x}q=k[j>>2]|0;r=l;return q|0}do if(w)if(!(k[m>>2]|0)){ux(f,n)|0;break}else{x=+hx(n,0);p[c>>3]=x;break}while(0);k[a>>2]=(k[a>>2]|0)+1;do if(Vh(a,767743,n,m)|0)if(!(k[m>>2]|0)){ux(g,n)|0;break}else{x=+hx(n,0);p[d>>3]=x;break}while(0);d=k[a>>2]|0;if((i[d>>0]|0)!=58){q=k[j>>2]|0;r=l;return q|0}k[a>>2]=d+1;do if(Vh(a,767743,n,m)|0)if(!(k[m>>2]|0)){ux(h,n)|0;break}else{x=+hx(n,0);p[e>>3]=x;break}while(0);q=k[j>>2]|0;r=l;return q|0}function Sp(a,b,c,d,e,f,g,h,j,l,m,n,p,q,s,t){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=+n;p=p|0;q=q|0;s=s|0;t=t|0;var u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0.0,I=0.0,J=0;u=r;r=r+112|0;v=u+96|0;w=u+80|0;x=u+72|0;y=u+56|0;z=u+40|0;A=u+24|0;B=u+8|0;C=u;D=k[t>>2]|0;if((D|0)>0){E=D;r=u;return E|0}if((d|0)>4){tk(767748);k[t>>2]=320;E=320;r=u;return E|0}D=k[a>>2]|0;F=k[D>>2]|0;if((F|0)!=(k[(k[D+4>>2]|0)+68>>2]|0))Fk(D,F+1|0,0,t)|0;switch(c|0){case 11:{G=8;break}case 21:{G=16;break}case 31:{G=32;break}case 42:{G=-32;break}case 82:{G=-64;break}default:{k[t>>2]=410;E=410;r=u;return E|0}}if((Tp(k[a>>2]|0,d,e,f,g,h,j,l,m,w,y,z,A,B,t)|0)>0){tk(767785);E=k[t>>2]|0;r=u;return E|0}do if(!(i[p>>0]|0)){H=n;o[C>>2]=H;I=H;J=21}else{if(!(Io(k[a>>2]|0,42,p,C,0,t)|0)){I=+o[C>>2];J=21;break}k[t>>2]=0;if((Tk(k[a>>2]|0,0,p,x,t)|0)<=0){o[C>>2]=-9.119119840596153e-36;break}tk(767824);tk(p);E=k[t>>2]|0;r=u;return E|0}while(0);do if((J|0)==21){p=I!=-9.119119840596153e-36;if(I<=0.0&p){tk(767880);k[t>>2]=125;E=125;r=u;return E|0}else{if(!((q|0)!=0&p))break;o[C>>2]=1.0/I;break}}while(0);if((Lh(v,b,t)|0)>0){tk(767925);E=k[t>>2]|0;r=u;return E|0}if((Xr(k[v>>2]|0,G,d,y,t)|0)>0){tk(767973);E=k[t>>2]|0;r=u;return E|0}if((Lk(k[a>>2]|0,k[v>>2]|0,9,d,w,t)|0)>0){tk(768018);E=k[t>>2]|0;r=u;return E|0}Up(k[a>>2]|0,k[v>>2]|0,d,w,t)|0;Vp(k[v>>2]|0,d,z,B,t)|0;if((Wp(k[a>>2]|0,k[v>>2]|0,G,d,y,w,z,A,B,+o[C>>2],k[x>>2]|0,q,s,t)|0)>0){tk(768077);E=k[t>>2]|0;r=u;return E|0}else{Fh(k[a>>2]|0,t)|0;k[a>>2]=k[v>>2];E=k[t>>2]|0;r=u;return E|0}return 0}function Tp(a,b,c,d,e,f,g,h,j,l,m,n,q,s,t){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;q=q|0;s=s|0;t=t|0;var u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0.0,Y=0,Z=0,_=0,$=0.0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0,ga=0,ha=0;u=r;r=r+464|0;v=u+172|0;w=u+91|0;x=u+16|0;y=u+12|0;z=u+8|0;A=u+4|0;B=u;C=k[t>>2]|0;if((C|0)>0){D=C;r=u;return D|0}if((b|0)>4){tk(768118);k[t>>2]=320;D=320;r=u;return D|0}C=k[a>>2]|0;E=a+4|0;if((C|0)!=(k[(k[E>>2]|0)+68>>2]|0))Fk(a,C+1|0,0,t)|0;i[v>>0]=0;C=v+71|0;i[C>>0]=0;F=v+142|0;i[F>>0]=0;G=v+213|0;i[G>>0]=0;k[y>>2]=0;Io(a,16,768175,v,0,y)|0;a:do if(!(k[y>>2]|0)){H=v;b:while(1){switch(i[H>>0]|0){case 0:{break a;break}case 44:{I=H;break b;break}default:{}}H=H+1|0}i[I>>0]=0;H=I;while(1){J=H+1|0;if((i[J>>0]|0)==32)H=J;else{K=J;break}}ux(C,K)|0;H=C;c:while(1){switch(i[H>>0]|0){case 0:{break a;break}case 44:{L=H;break c;break}default:{}}H=H+1|0}i[L>>0]=0;H=L;while(1){J=H+1|0;if((i[J>>0]|0)==32)H=J;else{M=J;break}}ux(F,M)|0;H=F;d:while(1){switch(i[H>>0]|0){case 0:{break a;break}case 44:{N=H;break d;break}default:{}}H=H+1|0}i[N>>0]=0;H=N;while(1){J=H+1|0;if((i[J>>0]|0)==32)H=J;else{O=J;break}}ux(G,O)|0}while(0);e:do if((b|0)>0){O=0;G=0;f:while(1){N=c+(O*71|0)|0;g:do if((i[N>>0]|0)==0?(ux(N,v+(O*71|0)|0)|0,(i[N>>0]|0)==0):0)switch(O|0){case 0:{i[N>>0]=88;i[N+1>>0]=0;break g;break}case 1:{i[N>>0]=89;i[N+1>>0]=0;break g;break}case 2:{i[N>>0]=90;i[N+1>>0]=0;break g;break}case 3:{i[N>>0]=84;i[N+1>>0]=0;break g;break}default:break g}while(0);F=l+(O<<2)|0;if((Tk(a,0,N,F,t)|0)>0){P=N;Q=31;break}M=k[F>>2]|0;if((k[(k[(k[E>>2]|0)+960>>2]|0)+((M+-1|0)*160|0)+88>>2]|0)>1){R=N;Q=33;break}Xk(a,M,z,0,0,t)|0;M=k[z>>2]|0;if((M|0)<0|(M|0)==16){S=N;Q=35;break}o[A>>2]=-9.119119840596153e-36;o[B>>2]=-9.119119840596153e-36;M=g+(O*71|0)|0;L=d+(O<<3)|0;if((i[M>>0]|0)!=0?(Io(a,82,M,L,0,t)|0)!=0:0){T=M;Q=38;break}U=+p[L>>3];if(!(U!=-9.1191291391491e-36)){Gk(770509,k[F>>2]|0,x,t)|0;L=n+(O<<2)|0;if((Io(a,42,x,L,0,t)|0)>0?(k[t>>2]=0,(Yp(a,k[F>>2]|0,L,B,t)|0)>0):0){V=N;Q=43;break}}else o[n+(O<<2)>>2]=U;L=h+(O*71|0)|0;M=e+(O<<3)|0;if((i[L>>0]|0)!=0?(Io(a,82,L,M,0,t)|0)!=0:0){W=L;Q=46;break}U=+p[M>>3];do if(!(U!=-9.1191291391491e-36)){Gk(770515,k[F>>2]|0,x,t)|0;M=q+(O<<2)|0;if((Io(a,42,x,M,0,t)|0)>0){k[t>>2]=0;X=+o[B>>2];if(!(X!=-9.119119840596153e-36))if((Yp(a,k[F>>2]|0,A,M,t)|0)>0){Y=N;Q=53;break f}else{Z=1;break}else{o[M>>2]=X;Z=1;break}}else Z=1}else{o[q+(O<<2)>>2]=U;Z=G}while(0);N=j+(O*71|0)|0;M=f+(O<<3)|0;if((i[N>>0]|0)!=0?(Io(a,82,N,M,0,t)|0)!=0:0){_=N;Q=56;break}U=+p[M>>3];if(U==0.0){Q=58;break}do if(!(U!=-9.1191291391491e-36)){k[y>>2]=0;Gk(768511,k[F>>2]|0,x,y)|0;if((Io(a,82,x,M,0,y)|0)>0){X=(+o[q+(O<<2)>>2]-+o[n+(O<<2)>>2])/10.0;N=s+(O<<2)|0;o[N>>2]=X;if(!(X>1.0))break;o[N>>2]=1.0}}else o[s+(O<<2)>>2]=U;while(0);M=n+(O<<2)|0;U=+o[M>>2];F=q+(O<<2)|0;X=+o[F>>2];if(U>X?(N=s+(O<<2)|0,$=+o[N>>2],$>0.0):0){aa=N;ba=$;Q=67}else{N=s+(O<<2)|0;$=+o[N>>2];if(U<X&$<0.0){aa=N;ba=$;Q=67}else{ca=$;da=U;ea=X}}if((Q|0)==67){Q=0;X=-ba;o[aa>>2]=X;ca=X;da=+o[M>>2];ea=+o[F>>2]}N=~~ca;L=~~da;C=~~ea;do if(+(N|0)==ca&((k[z>>2]|0)<42&+(L|0)==da&+(C|0)==ea)){k[m+(O<<2)>>2]=((C-L|0)/(N|0)|0)+1;if(da<ea){o[M>>2]=da+-.5;o[F>>2]=+o[F>>2]+.5;break}else{o[M>>2]=da+.5;o[F>>2]=+o[F>>2]+-.5;break}}else{X=(ea-da)/ca;if(Z){k[m+(O<<2)>>2]=~~(X+1.0);break}K=~~X;I=m+(O<<2)|0;k[I>>2]=K;X=da+ +(K|0)*ca;if(da<ea){if(!(X<ea))break;k[I>>2]=K+1;break}else{if(!(X>ea))break;k[I>>2]=K+1;break}}while(0);O=O+1|0;if((O|0)>=(b|0))break e;else G=Z}if((Q|0)==31){fa=w;ga=768181;ha=fa+42|0;do{i[fa>>0]=i[ga>>0]|0;fa=fa+1|0;ga=ga+1|0}while((fa|0)<(ha|0));qx(w,P)|0;tk(w);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==33){fa=w;ga=768223;ha=fa+28|0;do{i[fa>>0]=i[ga>>0]|0;fa=fa+1|0;ga=ga+1|0}while((fa|0)<(ha|0));qx(w,R)|0;tk(w);k[t>>2]=410;D=410;r=u;return D|0}else if((Q|0)==35){fa=w;ga=768251;ha=fa+48|0;do{i[fa>>0]=i[ga>>0]|0;fa=fa+1|0;ga=ga+1|0}while((fa|0)<(ha|0));qx(w,S)|0;tk(w);k[t>>2]=410;D=410;r=u;return D|0}else if((Q|0)==38){tk(768299);tk(T);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==43){fa=w;ga=768343;ha=fa+51|0;do{i[fa>>0]=i[ga>>0]|0;fa=fa+1|0;ga=ga+1|0}while((fa|0)<(ha|0));qx(w,V)|0;tk(w);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==46){tk(768394);tk(W);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==53){fa=w;ga=768343;ha=fa+51|0;do{i[fa>>0]=i[ga>>0]|0;fa=fa+1|0;ga=ga+1|0}while((fa|0)<(ha|0));qx(w,Y)|0;tk(w);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==56){tk(768438);tk(_);D=k[t>>2]|0;r=u;return D|0}else if((Q|0)==58){tk(768482);k[t>>2]=322;D=322;r=u;return D|0}}while(0);D=k[t>>2]|0;r=u;return D|0}function Up(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0;f=r;r=r+160|0;g=f+8|0;h=f+83|0;i=f+12|0;j=f;l=k[e>>2]|0;if(!((l|0)<1&(c|0)>0)){m=l;r=f;return m|0}l=0;do{k[g>>2]=0;n=l;l=l+1|0;Gk(768517,l,h,g)|0;Io(b,16,h,i,0,g)|0;if(k[g>>2]|0){k[g>>2]=0;o=d+(n<<2)|0;Gk(772997,k[o>>2]|0,h,g)|0;Io(a,16,h,i,0,g)|0;if(!(k[g>>2]|0)){Gk(768517,l,h,g)|0;es(b,16,h,i,768523,g)|0}k[g>>2]=0;Gk(773394,k[o>>2]|0,h,g)|0;Io(a,16,h,i,0,g)|0;if(!(k[g>>2]|0)){Gk(768539,l,h,g)|0;es(b,16,h,i,768545,g)|0}p[j>>3]=1.0;k[g>>2]=0;Gk(768562,l,h,g)|0;es(b,82,h,j,768568,g)|0;p[j>>3]=1.0;k[g>>2]=0;Gk(768584,l,h,g)|0;es(b,82,h,j,768590,g)|0;k[g>>2]=0;p[j>>3]=1.0;Gk(768606,l,h,g)|0;es(b,82,h,j,768612,g)|0}}while((l|0)!=(c|0));m=k[e>>2]|0;r=f;return m|0}function Vp(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,q=0,s=0,t=0.0,u=0.0,v=0.0,w=0,x=0;f=r;r=r+160|0;g=f+8|0;h=f+83|0;j=f+12|0;l=f;m=k[e>>2]|0;if(!((m|0)<1&(b|0)>0)){n=m;r=f;return n|0}m=0;do{k[g>>2]=0;q=m;m=m+1|0;Gk(768584,m,h,g)|0;Io(a,82,h,l,0,g)|0;s=k[g>>2]|0;t=+p[l>>3];k[g>>2]=0;Gk(768562,m,h,g)|0;Io(a,82,h,l,0,g)|0;if(!(k[g>>2]|0)){u=+p[l>>3];v=(u-+o[c+(q<<2)>>2])/+o[d+(q<<2)>>2]+.5;p[l>>3]=v;rq(a,h,v,-14,0,g)|0;w=u!=1.0?0:(s|0)==0&t==1.0&1}else w=0;k[g>>2]=0;Gk(768606,m,h,g)|0;Io(a,82,h,l,0,g)|0;if(!(k[g>>2]|0)){t=+p[l>>3];s=d+(q<<2)|0;u=t*+o[s>>2];p[l>>3]=u;rq(a,h,u,-14,0,g)|0;if(!((w|0)==0|t!=1.0)){p[l>>3]=1.0;Gk(768562,m,h,g)|0;rq(a,h,+p[l>>3],-14,0,g)|0;Gk(768584,m,h,g)|0;t=+o[c+(q<<2)>>2]+ +o[s>>2]*.5;p[l>>3]=t;rq(a,h,t,-14,0,g)|0}}else{s=d+(q<<2)|0;q=0;do{k[g>>2]=0;q=q+1|0;Gk(768623,q,j,g)|0;x=j+(xx(j)|0)|0;i[x>>0]=95;i[x+1>>0]=0;Gk(j,m,h,g)|0;Io(a,82,h,l,0,g)|0;if(!(k[g>>2]|0)){t=+p[l>>3]*+o[s>>2];p[l>>3]=t;rq(a,h,t,-14,0,g)|0}}while((q|0)!=(b|0))}}while((m|0)!=(b|0));n=k[e>>2]|0;r=f;return n|0}function Wp(a,b,c,d,e,f,g,h,i,j,l,m,n,p){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=+j;l=l|0;m=m|0;n=n|0;p=p|0;var q=0,s=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0.0,E=0.0,F=0.0,G=0.0,H=0,I=0,J=0.0,K=0,L=0,M=0,N=0,O=0;q=r;r=r+432|0;s=q+308|0;u=q+48|0;v=q+32|0;w=q+16|0;x=q;y=q+312|0;z=q+64|0;A=k[p>>2]|0;if((A|0)>0){B=A;r=q;return B|0}if((d|0)>4){tk(767748);k[p>>2]=320;B=320;r=q;return B|0}switch(c|0){case 8:{C=11;break}case 16:{C=21;break}case 32:{C=31;break}case -32:{C=42;break}case -64:{C=82;break}default:{k[p>>2]=410;B=410;r=q;return B|0}}c=k[a>>2]|0;if((c|0)!=(k[(k[a+4>>2]|0)+68>>2]|0))Fk(a,c+1|0,0,p)|0;o[y+104>>2]=j;k[y+100>>2]=l;k[y+96>>2]=m;k[y+4>>2]=a;k[y+28>>2]=C;m=y+8|0;k[m>>2]=d;k[y+108>>2]=n;if((d|0)>0){n=0;do{j=+(k[e+(n<<2)>>2]|0);o[u+(n<<2)>>2]=j;l=k[g+(n<<2)>>2]|0;c=v+(n<<2)|0;k[c>>2]=l;D=+o[h+(n<<2)>>2];E=(k[t>>2]=l,+o[t>>2]);if(E>D?(F=+o[i+(n<<2)>>2],F>0.0):0){G=F;H=19}else H=15;do if((H|0)==15){H=0;l=i+(n<<2)|0;if(E<D){F=+o[l>>2];if(F<0.0){G=F;H=19;break}I=(o[t>>2]=F,k[t>>2]|0)}else I=k[l>>2]|0;k[w+(n<<2)>>2]=I;J=(k[t>>2]=I,+o[t>>2])}while(0);if((H|0)==19){H=0;F=-G;o[w+(n<<2)>>2]=F;J=F}Xk(a,k[f+(n<<2)>>2]|0,s,0,0,p)|0;do if((k[s>>2]|0)<42?+(~~E|0)==E&+(~~D|0)==D&+(~~J|0)==J:0){o[x+(n<<2)>>2]=j+1.0;if(E<D){o[c>>2]=E+-.5;break}else{o[c>>2]=E+.5;break}}else H=26;while(0);if((H|0)==26){H=0;o[x+(n<<2)>>2]=(D-E)/J}n=n+1|0}while((n|0)!=(d|0));K=k[v>>2]|0;L=k[x>>2]|0;M=k[w>>2]|0;N=~~+o[u>>2];O=k[m>>2]|0}else{K=0;L=0;M=0;N=0;O=d}k[y+12>>2]=k[f>>2];k[y+48>>2]=K;k[y+64>>2]=L;k[y+80>>2]=M;k[y+32>>2]=N;if(((O|0)>1?(k[y+16>>2]=k[f+4>>2],k[y+52>>2]=k[v+4>>2],k[y+68>>2]=k[x+4>>2],k[y+84>>2]=k[w+4>>2],k[y+36>>2]=~~+o[u+4>>2],(O|0)>2):0)?(k[y+20>>2]=k[f+8>>2],k[y+56>>2]=k[v+8>>2],k[y+72>>2]=k[x+8>>2],k[y+88>>2]=k[w+8>>2],k[y+40>>2]=~~+o[u+8>>2],(O|0)>3):0){k[y+24>>2]=k[f+12>>2];k[y+60>>2]=k[v+12>>2];k[y+76>>2]=k[x+12>>2];k[y+92>>2]=k[w+12>>2];k[y+44>>2]=~~+o[u+12>>2]}Lq(z,b)|0;Mq(z,C)|0;Nq(z,2)|0;Pq(1,z,0,-1,3,y,p)|0;B=k[p>>2]|0;r=q;return B|0}function Xp(a,b,c,d,e,f,g,h,j,l,m,n,p,q,s,t){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=+n;p=p|0;q=q|0;s=s|0;t=t|0;var u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0.0,H=0.0,I=0;u=r;r=r+112|0;v=u+96|0;w=u+80|0;x=u+72|0;y=u+56|0;z=u+40|0;A=u+24|0;B=u+8|0;C=u;if((k[t>>2]|0)>0){D=0;r=u;return D|0}if((d|0)>4){tk(767748);k[t>>2]=320;D=0;r=u;return D|0}E=k[a>>2]|0;if((E|0)!=(k[(k[a+4>>2]|0)+68>>2]|0))Fk(a,E+1|0,0,t)|0;switch(c|0){case 11:{F=8;break}case 21:{F=16;break}case 31:{F=32;break}case 42:{F=-32;break}case 82:{F=-64;break}default:{k[t>>2]=410;D=0;r=u;return D|0}}if((Tp(a,d,e,f,g,h,j,l,m,w,y,z,A,B,t)|0)>0){tk(767785);D=0;r=u;return D|0}do if(!(i[p>>0]|0)){G=n;o[C>>2]=G;H=G;I=21}else{if(!(Io(a,42,p,C,0,t)|0)){H=+o[C>>2];I=21;break}k[t>>2]=0;if((Tk(a,0,p,x,t)|0)<=0){o[C>>2]=-9.119119840596153e-36;break}tk(767824);tk(p);D=0;r=u;return D|0}while(0);do if((I|0)==21){p=H!=-9.119119840596153e-36;if(H<=0.0&p){tk(767880);k[t>>2]=125;D=0;r=u;return D|0}else{if(!((q|0)!=0&p))break;o[C>>2]=1.0/H;break}}while(0);if((Lh(v,b,t)|0)>0){tk(767925);D=0;r=u;return D|0}if((Xr(k[v>>2]|0,F,d,y,t)|0)>0){tk(767973);D=0;r=u;return D|0}if((Lk(a,k[v>>2]|0,9,d,w,t)|0)>0){tk(768018);D=0;r=u;return D|0}Up(a,k[v>>2]|0,d,w,t)|0;Vp(k[v>>2]|0,d,z,B,t)|0;if((Wp(a,k[v>>2]|0,F,d,y,w,z,A,B,+o[C>>2],k[x>>2]|0,q,s,t)|0)>0){tk(768077);D=0;r=u;return D|0}else{D=k[v>>2]|0;r=u;return D|0}return 0}function Yp(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,p=0,q=0.0,s=0,t=0,u=0.0,v=0.0;f=r;r=r+4016|0;g=f+8|0;h=f+4|0;i=f+16|0;j=f;Io(a,41,790243,h,0,e)|0;o[j>>2]=-9.119119840596153e-36;o[c>>2]=8999999813669171236817415.0e12;o[d>>2]=-8999999813669171236817415.0e12;l=k[h>>2]|0;if(!l){m=k[e>>2]|0;r=f;return m|0}n=l;l=1;while(1){p=(n|0)<100?n:100;_l(a,42,b,l,((l|0)<0)<<31>>31,1,0,p,((p|0)<0)<<31>>31,j,i,g,e)|0;if((p|0)>0){q=+o[j>>2];s=(n|0)<100?n:100;t=0;do{u=+o[i+(t<<2)>>2];if(u!=q){v=+o[c>>2];o[c>>2]=v<u?v:u;v=+o[d>>2];o[d>>2]=v>u?v:u}t=t+1|0}while((t|0)!=(s|0))}s=k[h>>2]|0;n=s-p|0;k[h>>2]=n;if((s|0)==(p|0))break;else l=p+l|0}m=k[e>>2]|0;r=f;return m|0}function Zp(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0;e=r;r=r+1232|0;d=e+4|0;c=e;k[c>>2]=0;switch(k[g+28>>2]|0){case 11:{b=Oq(f)|0;k[g>>2]=b;break}case 21:{b=Oq(f)|0;k[g>>2]=b;break}case 31:{b=Oq(f)|0;k[g>>2]=b;break}case 42:{b=Oq(f)|0;k[g>>2]=b;break}case 82:{b=Oq(f)|0;k[g>>2]=b;break}default:{}}b=g+8|0;f=k[b>>2]|0;if((f|0)>0){a=g+4|0;h=g+12|0;i=0;do{Kq(d+(i*244|0)|0,k[a>>2]|0,k[h+(i<<2)>>2]|0,42,0)|0;i=i+1|0;j=k[b>>2]|0}while((i|0)<(j|0));l=j}else l=f;if(!(+o[g+104>>2]==-9.119119840596153e-36)){m=l;Pq(m,d,0,0,4,g,c)|0;n=k[c>>2]|0;r=e;return n|0}Kq(d+(l*244|0)|0,k[g+4>>2]|0,k[g+100>>2]|0,42,0)|0;m=(k[b>>2]|0)+1|0;Pq(m,d,0,0,4,g,c)|0;n=k[c>>2]|0;r=e;return n|0}function _p(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,n=0,q=0,r=0.0,s=0.0,t=0;if((c|0)==1){c=121280;b=g;g=c+112|0;do{k[c>>2]=k[b>>2];c=c+4|0;b=b+4|0}while((c|0)<(g|0));k[30348]=k[30347];b=Oq(f)|0;k[30349]=b;b=k[30322]|0;if((b|0)>1){c=Oq(f+244|0)|0;k[30350]=c;k[30351]=k[30328];c=k[30322]|0;if((c|0)>2){g=Oq(f+488|0)|0;k[30352]=g;g=ha(k[30329]|0,k[30351]|0)|0;k[30353]=g;g=k[30322]|0;if((g|0)>3){a=Oq(f+732|0)|0;k[30354]=a;a=ha(k[30330]|0,k[30353]|0)|0;k[30355]=a;h=k[30322]|0}else h=g}else h=c}else h=b;if((h|0)<(e|0)){e=Oq(f+(h*244|0)|0)|0;k[30356]=e}}if((d|0)<1)return 0;else n=1;while(1){e=k[30348]|0;if(!((e|0)!=0?(h=(i[e>>0]|0)==0,k[30348]=e+1,h):0))q=11;a:do if(((q|0)==11?(q=0,r=+o[(k[30349]|0)+(n<<2)>>2],!(r==-9.119119840596153e-36)):0)?(s=(r-+o[30332])/+o[30340],h=~~(s+1.0),!(s>+o[30336]|((h|0)<1|(h|0)>(k[30328]|0)))):0){e=k[30322]|0;if((e|0)>1){s=+o[(k[30350]|0)+(n<<2)>>2];if(s==-9.119119840596153e-36)break;r=(s-+o[30333])/+o[30341];f=~~r;if(r<0.0|(f|0)>=(k[30329]|0)|r>+o[30337])break;b=(ha(k[30351]|0,f)|0)+h|0;if((e|0)>2){r=+o[(k[30352]|0)+(n<<2)>>2];if(r==-9.119119840596153e-36)break;s=(r-+o[30334])/+o[30342];f=~~s;if(s<0.0|(f|0)>=(k[30330]|0)|s>+o[30338])break;c=(ha(k[30353]|0,f)|0)+b|0;if((e|0)>3){s=+o[(k[30354]|0)+(n<<2)>>2];if(s==-9.119119840596153e-36)break;r=(s-+o[30335])/+o[30343];e=~~r;if(r<0.0|(e|0)>=(k[30331]|0)|r>+o[30339])break;t=(ha(k[30355]|0,e)|0)+c|0}else t=c}else t=b}else t=h;r=+o[30346];if(r!=-9.119119840596153e-36)switch(k[30327]|0){case 31:{h=(k[30320]|0)+(t<<2)|0;k[h>>2]=(k[h>>2]|0)+~~r;break a;break}case 21:{h=(k[30320]|0)+(t<<1)|0;j[h>>1]=(m[h>>1]|0)+(~~r&65535);break a;break}case 42:{h=(k[30320]|0)+(t<<2)|0;o[h>>2]=r+ +o[h>>2];break a;break}case 82:{h=(k[30320]|0)+(t<<3)|0;p[h>>3]=r+ +p[h>>3];break a;break}case 11:{h=(k[30320]|0)+t|0;i[h>>0]=(l[h>>0]|0)+(~~r&255);break a;break}default:break a}h=k[30327]|0;b=(h|0)==31;if(!(k[30344]|0)){if(b){c=(k[30320]|0)+(t<<2)|0;k[c>>2]=(k[c>>2]|0)+~~+o[(k[30356]|0)+(n<<2)>>2];break}switch(h|0){case 21:{c=(k[30320]|0)+(t<<1)|0;j[c>>1]=(m[c>>1]|0)+(~~+o[(k[30356]|0)+(n<<2)>>2]&65535);break a;break}case 42:{c=(k[30320]|0)+(t<<2)|0;o[c>>2]=+o[(k[30356]|0)+(n<<2)>>2]+ +o[c>>2];break a;break}case 82:{c=(k[30320]|0)+(t<<3)|0;p[c>>3]=+o[(k[30356]|0)+(n<<2)>>2]+ +p[c>>3];break a;break}case 11:{c=(k[30320]|0)+t|0;i[c>>0]=(l[c>>0]|0)+(~~+o[(k[30356]|0)+(n<<2)>>2]&255);break a;break}default:break a}}else{if(b){b=(k[30320]|0)+(t<<2)|0;k[b>>2]=(k[b>>2]|0)+~~(1.0/+o[(k[30356]|0)+(n<<2)>>2]);break}switch(h|0){case 21:{h=(k[30320]|0)+(t<<1)|0;j[h>>1]=(m[h>>1]|0)+(~~(1.0/+o[(k[30356]|0)+(n<<2)>>2])&65535);break a;break}case 42:{h=(k[30320]|0)+(t<<2)|0;o[h>>2]=1.0/+o[(k[30356]|0)+(n<<2)>>2]+ +o[h>>2];break a;break}case 82:{h=(k[30320]|0)+(t<<3)|0;p[h>>3]=1.0/+o[(k[30356]|0)+(n<<2)>>2]+ +p[h>>3];break a;break}case 11:{h=(k[30320]|0)+t|0;i[h>>0]=(l[h>>0]|0)+(~~(1.0/+o[(k[30356]|0)+(n<<2)>>2])&255);break a;break}default:break a}}}while(0);if((n|0)==(d|0))break;else n=n+1|0}return 0}function $p(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0.0,Ra=0.0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0;f=r;r=r+752|0;g=f+192|0;h=f+184|0;m=f+176|0;n=f+168|0;o=f+160|0;p=f+152|0;q=f+144|0;s=f+136|0;t=f+128|0;u=f+120|0;v=f+112|0;w=f+104|0;x=f+32|0;y=f+24|0;z=f+16|0;A=f+8|0;B=f;C=f+208|0;D=f+708|0;E=f+626|0;F=f+545|0;G=f+464|0;H=f+200|0;k[b>>2]=0;k[c>>2]=0;k[d>>2]=0;I=aq(a,H)|0;if(!I){k[e>>2]=104;J=104;r=f;return J|0}K=k[H>>2]|0;i[E>>0]=i[768779]|0;i[E+1>>0]=i[768780]|0;i[E+2>>0]=i[768781]|0;H=E+3|0;L=H+77|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));i[E+80>>0]=0;M=ey(6,1)|0;if(M){N=(i[I>>0]|0)==0&1;i[M>>0]=i[I+N>>0]|0;O=N|2;i[M+1>>0]=i[I+O>>0]|0;i[M+2>>0]=i[I+(O+2)>>0]|0;O=N|6;i[M+3>>0]=i[I+O>>0]|0;i[M+4>>0]=i[I+(O+2)>>0]|0;O=Ax(M,768678,5)|0;dy(M);if(!O){P=1;Q=0;R=24;S=20;T=52;U=88;V=16;W=(K+-2052|0)/162|0;X=10}else X=6}else{tk(768626);X=6}do if((X|0)==6)if(!(Ax(I,768684,5)|0)){P=2;Q=1;R=22;S=18;T=50;U=86;V=10;W=(K+-2046|0)/81|0;X=10;break}else{tk(768690);tk(a);k[e>>2]=104;break}while(0);a:do if((X|0)==10){O=(((W+24|0)/36|0)*2880|0)+14404|0;k[c>>2]=O;M=ey(O,1)|0;if(!M){k[B>>2]=O;Xw(G,768783,B)|0;tk(a);k[e>>2]=104;break}k[b>>2]=M;Bx(M,E,80)|0;j[C>>1]=84;dq(M,782785,C);O=i[I+V>>0]|0;N=O<<24>>24!=0;Y=N&1;k[30357]=Y;if((P|0)==1)Z=Y;else{Y=i[I+14>>0]|0;_=i[I+17>>0]|0;Z=((N?_:Y)&255)<<24|(N?Y:_)&255|l[I+(N?15:16)>>0]<<8|l[I+(N?16:15)>>0]<<16}k[30358]=Z;_=V|1;Y=V+2|0;$=i[I+(V+3)>>0]|0;aa=(N?($&255)<<24:0)|(N?O:$)&255|l[I+(N?_:Y)>>0]<<8|l[I+(N?Y:_)>>0]<<16;switch(aa|0){case 12:case 2:{ba=8;break}case 3:{ba=16;break}case 11:{ba=-16;break}case 5:case 4:{ba=32;break}case 6:{ba=-32;break}case 7:{ba=-64;break}default:{k[A>>2]=aa;Xw(G,768829,A)|0;tk(G);tk(a);k[e>>2]=104;break a}}k[z>>2]=ba;Xw(D,772183,z)|0;dq(M,779656,D);eq(M,779656,768860);aa=(k[30357]|0)==1;_=i[I+S>>0]|0;Y=S|1;N=S+2|0;$=i[I+(S+3)>>0]|0;O=((aa?$:_)&255)<<24|(aa?_:$)&255|l[I+(aa?Y:N)>>0]<<8|l[I+(aa?N:Y)>>0]<<16;k[y>>2]=O;Xw(D,772183,y)|0;dq(M,782713,D);eq(M,782713,768881);Y=(k[30357]|0)==1;N=i[I+R>>0]|0;aa=R|1;$=R+2|0;_=i[I+(R+3)>>0]|0;k[x>>2]=((Y?_:N)&255)<<24|(Y?N:_)&255|l[I+(Y?aa:$)>>0]<<8|l[I+(Y?$:aa)>>0]<<16;Xw(D,772183,x)|0;dq(M,772833,D);eq(M,772833,768897);aa=M+320|0;$=(O|0)>1;if($){Y=(k[30357]|0)==1;_=i[I+(R+4)>>0]|0;N=R+5|0;ca=R+6|0;da=i[I+(R+7)>>0]|0;k[w>>2]=((Y?da:_)&255)<<24|(Y?_:da)&255|l[I+(Y?N:ca)>>0]<<8|l[I+(Y?ca:N)>>0]<<16;Xw(D,772183,w)|0;dq(M,790243,D);eq(M,790243,768922);N=M+400|0;if((O|0)>2){ca=(k[30357]|0)==1;Y=i[I+(R+8)>>0]|0;da=R+9|0;_=R+10|0;ea=i[I+(R+11)>>0]|0;k[v>>2]=((ca?ea:Y)&255)<<24|(ca?Y:ea)&255|l[I+(ca?da:_)>>0]<<8|l[I+(ca?_:da)>>0]<<16;Xw(D,772183,v)|0;dq(M,768947,D);eq(M,768947,768954);if((O|0)>3){O=(k[30357]|0)==1;da=i[I+(R+12)>>0]|0;_=R+13|0;ca=R+14|0;ea=i[I+(R+15)>>0]|0;k[u>>2]=((O?ea:da)&255)<<24|(O?da:ea)&255|l[I+(O?_:ca)>>0]<<8|l[I+(O?ca:_)>>0]<<16;Xw(D,772183,u)|0;dq(M,768979,D);eq(M,768979,768986);fa=1;ga=1;ia=M+560|0}else{fa=0;ga=1;ia=M+480|0}}else{fa=0;ga=0;ia=N}}else{fa=0;ga=0;ia=aa}do if(Q){aa=ey(384,1)|0;if(!aa){tk(768723);ja=0;break}else{N=0;do{_=i[I+(N+638)>>0]|0;i[aa+N>>0]=(_+-1&255)<31?32:_;N=N+1|0}while((N|0)!=383);ja=aa}}else ja=bq(I,732)|0;while(0);aa=xx(ja)|0;if((aa|0)<8){sy(ja+aa|0,32,8-aa|0)|0;i[ja+8>>0]=0}aa=xx(ja)|0;N=(aa|0)>67?67:aa;i[x>>0]=39;aa=x+1|0;Bx(aa,ja,N)|0;i[x+(N+1)>>0]=39;i[x+(N+2)>>0]=0;dq(M,769011,x);eq(M,769011,769018);dy(ja);N=(k[30357]|0)==1;_=i[I+T>>0]|0;ca=T|1;O=T+2|0;ea=i[I+(T+3)>>0]|0;k[t>>2]=((N?ea:_)&255)<<24|(N?_:ea)&255|l[I+(N?ca:O)>>0]<<8|l[I+(N?O:ca)>>0]<<16;Xw(D,772183,t)|0;dq(M,769034,D);eq(M,769034,769042);if($){ca=(k[30357]|0)==1;O=i[I+(T+4)>>0]|0;N=T+5|0;ea=T+6|0;_=i[I+(T+7)>>0]|0;k[s>>2]=((ca?_:O)&255)<<24|(ca?O:_)&255|l[I+(ca?N:ea)>>0]<<8|l[I+(ca?ea:N)>>0]<<16;Xw(D,772183,s)|0;dq(M,769070,D);eq(M,769070,769078);ka=320;la=ia+240|0}else{ka=240;la=ia+160|0}if(ga){N=T|8;ea=(k[30357]|0)==1;ca=i[I+N>>0]|0;_=T|9;O=N+2|0;da=i[I+(N+3)>>0]|0;k[q>>2]=((ea?da:ca)&255)<<24|(ea?ca:da)&255|l[I+(ea?_:O)>>0]<<8|l[I+(ea?O:_)>>0]<<16;Xw(D,772183,q)|0;dq(M,769106,D);eq(M,769106,769114);ma=ia+ka|0}else ma=la;if(fa){_=(k[30357]|0)==1;O=i[I+(T+12)>>0]|0;ea=T+13|0;da=T+14|0;ca=i[I+(T+15)>>0]|0;k[p>>2]=((_?ca:O)&255)<<24|(_?O:ca)&255|l[I+(_?ea:da)>>0]<<8|l[I+(_?da:ea)>>0]<<16;Xw(D,772183,p)|0;dq(M,769142,D);eq(M,769142,769150);na=ma+80|0}else na=ma;ea=xx(a)|0;da=(ea|0)>67?67:ea;i[x>>0]=39;Bx(aa,a,da)|0;i[x+(da+1)>>0]=39;i[x+(da+2)>>0]=0;dq(M,769178,x);eq(M,769178,769186);do if(Q){da=ey(256,1)|0;if(!da){tk(768723);oa=0;break}else{ea=0;do{_=i[I+(ea+126)>>0]|0;i[da+ea>>0]=(_+-1&255)<31?32:_;ea=ea+1|0}while((ea|0)!=255);oa=da}}else oa=bq(I,412)|0;while(0);if((Ax(oa,768775,3)|0)==0?($=cq(oa,a)|0,($|0)!=0):0){dy(oa);pa=$}else pa=oa;if(((rx(pa,47)|0)==0?(rx(pa,36)|0)==0:0)?($=cq(pa,a)|0,($|0)!=0):0){dy(pa);qa=$}else qa=pa;$=rx(qa,33)|0;if(!$){da=xx(qa)|0;ea=(da|0)>67?67:da;i[x>>0]=39;Bx(aa,qa,ea)|0;i[x+(ea+1)>>0]=39;i[x+(ea+2)>>0]=0;dq(M,769208,x)}else{ea=$+1|0;$=xx(ea)|0;da=($|0)>67?67:$;i[x>>0]=39;Bx(aa,ea,da)|0;i[x+(da+1)>>0]=39;i[x+(da+2)>>0]=0;dq(M,769208,x)}dy(qa);eq(M,769208,769216);da=(k[30357]|0)==1;ea=i[I+U>>0]|0;$=U|1;_=U+2|0;ca=i[I+(U+3)>>0]|0;k[o>>2]=((((da?ca:ea)&255)<<24|(da?ea:ca)&255|l[I+(da?$:_)>>0]<<8|l[I+(da?_:$)>>0]<<16)<<1)+-2;Xw(D,772183,o)|0;dq(M,769237,D);eq(M,769237,769244);k[n>>2]=P;Xw(D,772183,n)|0;dq(M,769284,D);eq(M,769284,769291);if(!(k[30358]|0)){j[C>>1]=70;dq(M,769325,C)}else{j[C>>1]=84;dq(M,769325,C)}eq(M,769325,769333);$=na+400|0;i[F+80>>0]=0;b:do if(Q){H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));if((K|0)>2046){_=$;da=2046;ca=0;while(1){ea=i[I+da>>0]|0;switch(ea<<24>>24){case 0:{ra=_;break b;break}case 10:{Bx(_,F,80)|0;O=Ax(F,769375,7)|0;N=_+80|0;H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));sa=(O|0)==0?_:N;ta=0;break}default:{if((ca|0)>80){if(!(Ax(F,769375,7)|0)){ua=_;va=ca}else{Bx(_,F,80)|0;ua=_+80|0;va=9}H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));wa=ua;xa=va}else{wa=_;xa=ca}if(!(ea<<24>>24<33|ea<<24>>24==127))i[F+xa>>0]=ea;sa=wa;ta=xa+1|0}}da=da+1|0;if((da|0)>=(K|0)){ra=sa;break b}else{_=sa;ca=ta}}}else ra=$}else{ca=(k[30357]|0)!=1;H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));_=ca&1;if((K|0)>2052){da=$;N=2052;O=0;while(1){Y=i[I+(N|_)>>0]|0;switch(Y<<24>>24){case 0:{ra=da;break b;break}case 10:{if(!(Ax(F,769375,7)|0))ya=da;else{Bx(da,F,80)|0;ya=da+80|0}H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));za=ya;Aa=0;break}default:{if((O|0)>80){if(!(Ax(F,769375,7)|0)){Ba=da;Ca=O}else{Bx(da,F,80)|0;Ba=da+80|0;Ca=9}H=F;L=H+80|0;do{i[H>>0]=32;H=H+1|0}while((H|0)<(L|0));Da=Ba;Ea=Ca}else{Da=da;Ea=O}if(!(Y<<24>>24<33|Y<<24>>24==127))i[F+Ea>>0]=Y;za=Da;Aa=Ea+1|0}}N=N+2|0;if((N|0)>=(K|0)){ra=za;break b}else{da=za;O=Aa}}}else ra=$}while(0);Bx(ra,E,80)|0;$=0;while(1){aa=M+$|0;if(!(i[aa>>0]|0)){Fa=aa;Ga=$;X=85;break}aa=$+1|0;if((aa|0)<57600)$=aa;else{Ha=aa;X=84;break}}if((X|0)==84){Ia=M+Ha|0;X=87}else if((X|0)==85)if((Ga|0)>0){Ia=Fa;X=87}else{Ja=M;Ka=0}c:do if((X|0)==87){$=Ia;aa=M;O=M;while(1){da=fq(O,768779,$-O|0)|0;if(!da){Ja=aa;Ka=0;break c}N=(da-aa|0)%80|0;_=i[da+3>>0]|0;d:do if((N|0)<=7){e:do if(_<<24>>24>32){switch(_<<24>>24){case 61:case 127:{break e;break}default:{}}La=da+1|0;break d}while(0);ca=da+(0-N)|0;if((N|0)>0){ea=da+1|0;Ma=O;Na=ca;while(1){Oa=(i[Na>>0]|0)==32?Ma:ea;Na=Na+1|0;if(Na>>>0>=da>>>0){Pa=Oa;break}else Ma=Oa}}else Pa=O;if(da>>>0<Pa>>>0)La=Pa;else{Ja=aa;Ka=ca;break c}}else La=da+1|0;while(0);if(La>>>0<Ia>>>0)O=La;else{Ja=aa;Ka=0;break}}}while(0);aa=Ka+80|0;O=k[c>>2]|0;$=M+(O-((O>>>0)%2880|0))|0;k[d>>2]=aa-Ja;i[E>>0]=i[769383]|0;i[E+1>>0]=i[769384]|0;i[E+2>>0]=i[769385]|0;if(aa>>>0<$>>>0){O=aa;do{Bx(O,E,80)|0;O=O+80|0}while(O>>>0<$>>>0)}}while(0);dy(I);I=k[e>>2]|0;if((I|0)>0){J=I;r=f;return J|0}I=(k[d>>2]|0)+-1|0;k[d>>2]=2880-((I>>>0)%2880|0)+I;I=k[b>>2]|0;Ja=gq(I,769208)|0;do if(Ja)if((xx(Ja)|0)<255){ux(C,Ja)|0;break}else{Bx(C,Ja,254)|0;break}while(0);Ja=gq(I,769237)|0;do if(Ja){ux(D,Ja)|0;Qa=+dx(D);Ra=Qa+.001;if(!(Ra>2147483647.0)){if(Qa>=0.0){Sa=~~Ra;break}Ra=Qa+-.001;if(Ra<-2147483648.0)Sa=-2147483648;else Sa=~~Ra}else Sa=2147483647}else Sa=0;while(0);Ja=rx(C,33)|0;if(!Ja)Ta=Dw(C,769387)|0;else Ta=Dw(Ja+1|0,769387)|0;f:do if(Ta){Ja=ey(Sa,1)|0;if(!Ja){tk(769431);tk(C);ww(Ta)|0;k[e>>2]=104;break}Ka=Hw(Ja,1,Sa,Ta)|0;if((Ka|0)<(Sa|0)){k[m>>2]=Ka;k[m+4>>2]=1024;Xw(E,769480,m)|0;tk(E);dy(Ja);ww(Ta)|0;k[e>>2]=104;break}Ka=ey(6,1)|0;if(Ka){La=(i[Ja>>0]|0)==0&1;i[Ka>>0]=i[Ja+La>>0]|0;Ia=La|2;i[Ka+1>>0]=i[Ja+Ia>>0]|0;i[Ka+2>>0]=i[Ja+(Ia+2)>>0]|0;Ia=La|6;i[Ka+3>>0]=i[Ja+Ia>>0]|0;i[Ka+4>>0]=i[Ja+(Ia+2)>>0]|0;Ia=Ax(Ka,769517,5)|0;dy(Ka);if(Ia)X=124}else{tk(768626);X=124}if((X|0)==124?(Ax(Ja,769523,5)|0)!=0:0){tk(769529);tk(C);dy(Ja);ww(Ta)|0;k[e>>2]=104;break}dy(Ja);Ja=gq(I,782713)|0;do if(Ja){ux(D,Ja)|0;Ra=+dx(D);Qa=Ra+.001;if(!(Qa>2147483647.0)){if(Ra>=0.0){Ua=~~Qa;break}Qa=Ra+-.001;if(Qa<-2147483648.0)Ua=-2147483648;else Ua=~~Qa}else Ua=2147483647}else Ua=1;while(0);Ja=gq(I,772833)|0;do if(Ja){ux(D,Ja)|0;Qa=+dx(D);Ra=Qa+.001;if(!(Ra>2147483647.0)){if(Qa>=0.0){Va=~~Ra;break}Ra=Qa+-.001;if(Ra<-2147483648.0)Va=-2147483648;else Va=~~Ra}else Va=2147483647}else Va=1;while(0);Ja=gq(I,769034)|0;do if(Ja){ux(D,Ja)|0;Ra=+dx(D);Qa=Ra+.001;if(!(Qa>2147483647.0)){if(Ra>=0.0){Wa=~~Qa;break}Qa=Ra+-.001;if(Qa<-2147483648.0)Wa=-2147483648;else Wa=~~Qa}else Wa=2147483647}else Wa=1;while(0);do if((Ua|0)>1){Ja=gq(I,790243)|0;do if(Ja){ux(D,Ja)|0;Qa=+dx(D);Ra=Qa+.001;if(Ra>2147483647.0)Xa=2147483647;else{if(Qa>=0.0){Xa=~~Ra;break}Ra=Qa+-.001;if(Ra<-2147483648.0){Xa=-2147483648;break}Xa=~~Ra}}else Xa=1;while(0);Ja=gq(I,769070)|0;if(Ja)ux(D,Ja)|0;if((Ua|0)>2){Ja=gq(I,768947)|0;do if(!Ja)Ya=1;else{ux(D,Ja)|0;Ra=+dx(D);Qa=Ra+.001;if(Qa>2147483647.0){Ya=2147483647;break}if(Ra>=0.0){Ya=~~Qa;break}Qa=Ra+-.001;if(Qa<-2147483648.0){Ya=-2147483648;break}Ya=~~Qa}while(0);if((Ua|0)<=3){Za=Ya;_a=Xa;$a=1;break}Ja=gq(I,768979)|0;do if(!Ja)ab=1;else{ux(D,Ja)|0;Qa=+dx(D);Ra=Qa+.001;if(Ra>2147483647.0){ab=2147483647;break}if(Qa>=0.0){ab=~~Ra;break}Ra=Qa+-.001;if(Ra<-2147483648.0){ab=-2147483648;break}ab=~~Ra}while(0);Za=Ya;_a=Xa;$a=ab}else{Za=1;_a=Xa;$a=1}}else{Za=1;_a=1;$a=1}while(0);Ja=gq(I,779656)|0;do if(Ja){ux(D,Ja)|0;Ra=+dx(D);Qa=Ra+.001;if(!(Qa>2147483647.0)){if(Ra>=0.0){bb=~~Qa;break}Qa=Ra+-.001;if(Qa<-2147483648.0)bb=-2147483648;else bb=~~Qa}else bb=2147483647}else bb=0;while(0);Ja=(((bb|0)<0?0-bb|0:bb)|0)/8|0;Ia=ha(ha(ha(ha(Za,Va)|0,_a)|0,$a)|0,Ja)|0;Ka=Ia+-1+(k[d>>2]|0)|0;La=Ka-((Ka>>>0)%2880|0)+2880|0;if(La>>>0>(k[c>>2]|0)>>>0){Ka=fy(k[b>>2]|0,La)|0;if(!Ka){k[h>>2]=k[d>>2];Xw(E,769561,h)|0;tk(E);tk(C);ww(Ta)|0;k[e>>2]=104;break}else cb=Ka}else cb=I;k[b>>2]=cb;k[c>>2]=La;Ka=k[d>>2]|0;Pa=cb+Ka|0;k[d>>2]=La;if((Wa|0)!=(Va|0)){La=ha(Ja,Wa-Va|0)|0;Fa=ha(Ja,Va)|0;Ja=(Za|0)>1&(_a|0)==1?Za:_a;if((Ja|0)>0){Ga=0;Ha=Pa;ra=0;while(1){Aa=(Hw(Ha,1,Fa,Ta)|0)+ra|0;Kw(Ta,La,1)|0;Ga=Ga+1|0;if((Ga|0)==(Ja|0)){db=Aa;break}else{Ha=Ha+Fa|0;ra=Aa}}}else db=0}else db=Hw(Pa,1,Ia,Ta)|0;ww(Ta)|0;if((db|0)<(Ia|0)){k[g>>2]=db;k[g+4>>2]=Ia;Xw(E,769480,g)|0;tk(E);tk(C);k[e>>2]=104;break}if(k[30358]|0)switch(bb|0){case 16:{if((Ia|0)<2)break f;ra=cb+(Ka+Ia)|0;Fa=Pa;do{Ha=i[Fa>>0]|0;Ja=Fa+1|0;i[Fa>>0]=i[Ja>>0]|0;i[Ja>>0]=Ha;Fa=Fa+2|0}while(Fa>>>0<ra>>>0);break}case 32:{if((Ia|0)<4)break f;ra=cb+(Ka+Ia)|0;Fa=Pa;do{Ha=i[Fa>>0]|0;Ja=Fa+1|0;Ga=i[Ja>>0]|0;La=Fa+2|0;Aa=i[La>>0]|0;za=Fa+3|0;i[Fa>>0]=i[za>>0]|0;i[Ja>>0]=Aa;i[La>>0]=Ga;i[za>>0]=Ha;Fa=Fa+4|0}while(Fa>>>0<ra>>>0);break}case -16:{if((Ia|0)<2)break f;ra=cb+(Ka+Ia)|0;Fa=Pa;do{Ha=i[Fa>>0]|0;za=Fa+1|0;i[Fa>>0]=i[za>>0]|0;i[za>>0]=Ha;Fa=Fa+2|0}while(Fa>>>0<ra>>>0);break}case -32:{if((Ia|0)<4)break f;ra=cb+(Ka+Ia)|0;Fa=Pa;do{Ha=i[Fa>>0]|0;za=Fa+1|0;Ga=i[za>>0]|0;La=Fa+2|0;Aa=i[La>>0]|0;Ja=Fa+3|0;i[Fa>>0]=i[Ja>>0]|0;i[za>>0]=Aa;i[La>>0]=Ga;i[Ja>>0]=Ha;Fa=Fa+4|0}while(Fa>>>0<ra>>>0);break}case -64:{if((Ia|0)<8)break f;ra=cb+(Ka+Ia)|0;Fa=Pa;do{Ha=i[Fa>>0]|0;Ja=Fa+1|0;Ga=i[Ja>>0]|0;La=Fa+2|0;Aa=i[La>>0]|0;za=Fa+3|0;K=i[za>>0]|0;Ea=Fa+4|0;Da=i[Ea>>0]|0;F=Fa+5|0;Ca=i[F>>0]|0;Ba=Fa+6|0;H=i[Ba>>0]|0;L=Fa+7|0;i[Fa>>0]=i[L>>0]|0;i[Ja>>0]=H;i[La>>0]=Ca;i[za>>0]=Da;i[Ea>>0]=K;i[F>>0]=Aa;i[Ba>>0]=Ga;i[L>>0]=Ha;Fa=Fa+8|0}while(Fa>>>0<ra>>>0);break}default:break f}}else{tk(769390);tk(C);k[e>>2]=104}while(0);J=k[e>>2]|0;r=f;return J|0}function aq(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;c=r;r=r+112|0;d=c+8|0;e=c;f=c+16|0;k[b>>2]=0;g=Dw(a,769387)|0;if(!g){tk(769754);tk(a);h=0;r=c;return h|0}if(Kw(g,0,2)|0){tk(769787);tk(a);h=0;r=c;return h|0}i=Nw(g)|0;if((i|0)<0){tk(769819);tk(a);h=0;r=c;return h|0}if(Kw(g,0,0)|0){tk(769855);tk(a);h=0;r=c;return h|0}j=i+5e3|0;l=ey(1,j)|0;if(!l){k[e>>2]=j;Xw(f,769900,e)|0;tk(f);tk(a);h=0;r=c;return h|0}k[b>>2]=j;j=Hw(l,1,i,g)|0;ww(g)|0;if((j|0)>=1024){h=l;r=c;return h|0}k[d>>2]=j;k[d+4>>2]=1024;Xw(f,769941,d)|0;tk(f);tk(a);dy(l);h=0;r=c;return h|0}function bq(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;c=ey(161,1)|0;if(!c){tk(768723);d=0}else{e=0;do{f=i[a+(e+b)>>0]|0;i[c+e>>0]=(f+-1&255)<31?32:f;e=e+1|0}while((e|0)!=160);d=c}c=ey(80,1)|0;if(!c){tk(768626);g=0;dy(d);return g|0}e=0;b=(i[d>>0]|0)==0&1;while(1){i[c+e>>0]=i[d+b>>0]|0;e=e+1|0;if((e|0)==79){g=c;break}else b=b+2|0}dy(d);return g|0}function cq(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0;c=ey(511,1)|0;if(!c){tk(769694);d=0;return d|0}if(!(Ax(a,769749,4)|0)){Bx(c,b,255)|0;e=xx(c)|0;while(1){if((e|0)<=0){f=e;break}g=e+-1|0;if((i[c+g>>0]|0)==47){f=e;break}else e=g}i[c+f>>0]=0;zx(c,a+4|0,255)|0;d=c;return d|0}if((rx(a,47)|0)==0?(rx(a,36)|0)==0:0){Bx(c,b,255)|0;f=xx(c)|0;while(1){if((f|0)<=0){h=f;break}e=f+-1|0;if((i[c+e>>0]|0)==47){h=f;break}else f=e}i[c+h>>0]=0;zx(c,a,255)|0;d=c;return d|0}if(Ax(a,768775,3)|0){d=c;return d|0}Bx(c,b,255)|0;b=xx(c)|0;i[c+(b+-3)>>0]=112;i[c+(b+-2)>>0]=105;i[c+(b+-1)>>0]=120;d=c;return d|0}
function uk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0;switch(a|0){case 1:{if((k[30047]|0)>0){a=0;do{i[k[120192+(a<<2)>>2]>>0]=0;a=a+1|0}while((a|0)<(k[30047]|0))}k[30047]=0;return}case 2:{a=k[30047]|0;if((a|0)>0)c=a;else return;do{a=c+-1|0;k[30047]=a;d=k[120192+(a<<2)>>2]|0;a=i[d>>0]|0;i[d>>0]=0;c=k[30047]|0}while(a<<24>>24!=27&(c|0)>0);return}case 3:{c=k[30047]|0;if((c|0)<=0)return;a=c+-1|0;k[30047]=a;i[k[120192+(a<<2)>>2]>>0]=0;return}case 4:{a=k[30047]|0;while(1){if((a|0)<=0)break;ux(b,k[30048]|0)|0;i[k[30048]>>0]=0;c=k[30047]|0;a=c+-1|0;k[30047]=a;if((c|0)>1){c=0;do{d=c;c=c+1|0;k[120192+(d<<2)>>2]=k[120192+(c<<2)>>2]}while((c|0)<(a|0))}if((i[b>>0]|0)!=27){e=33;break}}if((e|0)==33)return;i[b>>0]=0;return}case 5:{k[30046]=b;if(!(i[b>>0]|0))return;e=b;b=k[30047]|0;do{a:do if((b|0)==25){a=k[30048]|0;k[30073]=a;i[a>>0]=0;a=k[30047]|0;c=a+-1|0;k[30047]=c;if((a|0)>1){a=0;do{d=a;a=a+1|0;k[120192+(d<<2)>>2]=k[120192+(a<<2)>>2]}while((a|0)<(c|0))}a=k[30073]|0;k[120192+(c<<2)>>2]=a;f=a;g=k[30046]|0}else{do if(i[751175]|0)if(i[751256]|0)if(i[751337]|0)if(i[751418]|0)if(i[751499]|0)if(i[751580]|0)if(i[751661]|0)if(i[751742]|0)if(i[751823]|0)if(i[751904]|0)if(i[751985]|0)if(i[752066]|0)if(i[752147]|0)if(!(i[752228]|0))h=752228;else{if(!(i[752309]|0)){h=752309;break}if(!(i[752390]|0)){h=752390;break}if(!(i[752471]|0)){h=752471;break}if(!(i[752552]|0)){h=752552;break}if(!(i[752633]|0)){h=752633;break}if(!(i[752714]|0)){h=752714;break}if(!(i[752795]|0)){h=752795;break}if(!(i[752876]|0)){h=752876;break}if(!(i[752957]|0)){h=752957;break}if(!(i[753038]|0)){h=753038;break}if(!(i[753119]|0)){h=753119;break}f=k[120192+(b<<2)>>2]|0;g=e;break a}else h=752147;else h=752066;else h=751985;else h=751904;else h=751823;else h=751742;else h=751661;else h=751580;else h=751499;else h=751418;else h=751337;else h=751256;else h=751175;while(0);k[120192+(b<<2)>>2]=h;f=h;g=e}while(0);zx(f,g,80)|0;b=(k[30047]|0)+1|0;k[30047]=b;c=k[30046]|0;a=xx(c)|0;e=c+(a>>>0>80?80:a)|0;k[30046]=e}while((i[e>>0]|0)!=0);return}case 6:{e=k[30047]|0;b:do if((e|0)==25){b=k[30048]|0;k[30073]=b;i[b>>0]=0;b=k[30047]|0;g=b+-1|0;k[30047]=g;if((b|0)>1){b=0;do{f=b;b=b+1|0;k[120192+(f<<2)>>2]=k[120192+(b<<2)>>2]}while((b|0)<(g|0))}b=k[30073]|0;k[120192+(g<<2)>>2]=b;j=b}else{do if(i[751175]|0)if(i[751256]|0)if(i[751337]|0)if(i[751418]|0)if(i[751499]|0)if(i[751580]|0)if(i[751661]|0)if(i[751742]|0)if(i[751823]|0)if(i[751904]|0)if(i[751985]|0)if(i[752066]|0)if(i[752147]|0)if(i[752228]|0)if(i[752309]|0)if(!(i[752390]|0))l=752390;else{if(!(i[752471]|0)){l=752471;break}if(!(i[752552]|0)){l=752552;break}if(!(i[752633]|0)){l=752633;break}if(!(i[752714]|0)){l=752714;break}if(!(i[752795]|0)){l=752795;break}if(!(i[752876]|0)){l=752876;break}if(!(i[752957]|0)){l=752957;break}if(!(i[753038]|0)){l=753038;break}if(!(i[753119]|0)){l=753119;break}j=k[120192+(e<<2)>>2]|0;break b}else l=752309;else l=752228;else l=752147;else l=752066;else l=751985;else l=751904;else l=751823;else l=751742;else l=751661;else l=751580;else l=751499;else l=751418;else l=751337;else l=751256;else l=751175;while(0);k[120192+(e<<2)>>2]=l;j=l}while(0);i[j>>0]=27;i[(k[120192+(k[30047]<<2)>>2]|0)+1>>0]=0;k[30047]=(k[30047]|0)+1;return}default:return}}function vk(){uk(6,0);return}function wk(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0;b=k[30047]|0;while(1){if((b|0)<=0)break;ux(a,k[30048]|0)|0;i[k[30048]>>0]=0;c=k[30047]|0;b=c+-1|0;k[30047]=b;if((c|0)>1){c=0;do{d=c;c=c+1|0;k[120192+(d<<2)>>2]=k[120192+(c<<2)>>2]}while((c|0)!=(b|0))}c=i[a>>0]|0;if(c<<24>>24!=27){e=c;f=7;break}}if((f|0)==7){g=e<<24>>24;return g|0}i[a>>0]=0;e=0;g=e<<24>>24;return g|0}function xk(){var a=0;if((k[30047]|0)>0)a=0;else{k[30047]=0;return}do{i[k[120192+(a<<2)>>2]>>0]=0;a=a+1|0}while((a|0)<(k[30047]|0));k[30047]=0;return}function yk(){var a=0,b=0,c=0;a=k[30047]|0;if((a|0)>0)b=a;else return;do{a=b+-1|0;k[30047]=a;c=k[120192+(a<<2)>>2]|0;a=i[c>>0]|0;i[c>>0]=0;b=k[30047]|0}while(a<<24>>24!=27&(b|0)>0);return}function zk(a){a=a|0;var b=0;switch(a|0){case 14:case 11:{b=1;break}case 20:{b=2;break}case 21:{b=2;break}case 40:{b=4;break}case 41:{b=4;break}case 31:{b=4;break}case 30:{b=4;break}case 42:{b=4;break}case 82:{b=8;break}default:b=0}return b|0}function Ak(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;c=r;r=r+112|0;d=c+8|0;e=c;f=c+16|0;g=k[b>>2]|0;if((g|0)>0){h=g;r=c;return h|0}j=xx(a)|0;l=j>>>0>8?8:j;if(!l){h=g;r=c;return h|0}else{m=g;n=0;o=0}a:while(1){g=a+n|0;j=i[g>>0]|0;if(!m)p=j;else p=(nv(j<<24>>24)|0)&255;b:do if(!((p+-65&255)<26|(p+-48&255)<10)){switch(p<<24>>24){case 45:case 95:{q=7;break b;break}default:{}}if((i[g>>0]|0)==32)s=1;else{t=g;u=n;q=12;break a}}else q=7;while(0);if((q|0)==7){q=0;if(!o)s=0;else{q=8;break}}n=n+1|0;g=k[b>>2]|0;if(n>>>0>=l>>>0){h=g;q=18;break}else{m=g;o=s}}if((q|0)==8){if(!(k[b>>2]|0)){k[e>>2]=a;Xw(f,753200,e)|0;uk(5,f)}k[b>>2]=207;h=207;r=c;return h|0}else if((q|0)==12){c:do if(!(k[b>>2]|0)){k[d>>2]=u+1;k[d+4>>2]=a;Xw(f,753246,d)|0;uk(5,f);switch(i[t>>0]|0){case 0:{uk(5,753292);break c;break}case 9:{uk(5,753322);break c;break}default:break c}}while(0);k[b>>2]=207;h=207;r=c;return h|0}else if((q|0)==18){r=c;return h|0}return 0}function Bk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;c=r;r=r+96|0;d=c;e=c+8|0;f=k[b>>2]|0;if((f|0)>0){g=f;r=c;return g|0}h=xx(a)|0;j=8;while(1){if(j>>>0>=h>>>0){g=f;l=15;break}m=a+j|0;n=i[m>>0]|0;o=j+1|0;if(n<<24>>24<32|n<<24>>24==127){p=m;q=n;s=o;break}else j=o}if((l|0)==15){r=c;return g|0}k[d>>2]=s;k[d+4>>2]=q<<24>>24;Xw(e,753358,d)|0;switch(i[p>>0]|0){case 0:{t=e+(xx(e)|0)|0;u=753414;v=t+14|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 9:{t=e+(xx(e)|0)|0;u=753428;v=t+13|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 10:{t=e+(xx(e)|0)|0;u=753441;v=t+19|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 11:{t=e+(xx(e)|0)|0;u=753460;v=t+16|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 12:{t=e+(xx(e)|0)|0;u=753476;v=t+19|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 13:{t=e+(xx(e)|0)|0;u=753495;v=t+19|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 27:{t=e+(xx(e)|0)|0;u=753514;v=t+16|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}case 127:{t=e+(xx(e)|0)|0;u=753530;v=t+16|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));break}default:{}}uk(5,e);Bx(e,a,80)|0;i[e+80>>0]=0;uk(5,e);k[b>>2]=207;g=207;r=c;return g|0}function Ck(a){a=a|0;var b=0,c=0,d=0,e=0;b=xx(a)|0;if(!b)return;else c=0;do{d=a+c|0;e=(nv(i[d>>0]|0)|0)&255;i[d>>0]=e;c=c+1|0}while((c|0)!=(b|0));return}function Dk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;f=r;r=r+160|0;g=f+79|0;h=f+4|0;j=f;k[j>>2]=-1;l=k[e>>2]|0;if((l|0)>0){m=l;r=f;return m|0}i[g>>0]=0;i[h>>0]=0;i[d>>0]=0;l=0;while(1){n=a+l|0;if((i[n>>0]|0)==32)l=l+1|0;else{o=n;break}}zx(g,o,74)|0;o=xx(b)|0;l=xx(g)|0;if(!l)p=0;else{n=g+(l+-1)|0;if((i[n>>0]|0)==32){l=n;while(1){i[l>>0]=0;q=l+-1|0;if((i[q>>0]|0)==32)l=q;else{s=q;break}}}else s=n;p=1-g+s|0}if(rx(g,61)|0){uk(5,753546);uk(5,g);k[e>>2]=207;m=207;r=f;return m|0}if(p>>>0<9?(Ak(g,j)|0)<1:0){qx(d,g)|0;if(p>>>0<8)sy(d+p|0,32,8-p|0)|0;i[d+8>>0]=61;i[d+9>>0]=32;i[d+10>>0]=0;t=10;u=0;v=0}else w=15;a:do if((w|0)==15){s=i[g>>0]|0;do if(s<<24>>24>=72){if(s<<24>>24>72){if(s<<24>>24!=104)break;if(Ax(g,753596,9)|0)break}else if(Ax(g,761298,9)|0)break;qx(d,g)|0;n=p+3|0;l=d+(xx(d)|0)|0;if((n+o|0)>>>0>80){i[l>>0]=i[753606]|0;i[l+1>>0]=i[753607]|0;i[l+2>>0]=i[753608]|0;t=p+2|0;u=0;v=0;break a}else{i[l>>0]=32;i[l+1>>0]=61;i[l+2>>0]=32;i[l+3>>0]=0;t=n;u=0;v=0;break a}}while(0);zx(h,g,74)|0;s=Fx(h,776273)|0;if(s){n=s;s=0;l=0;while(1){k[j>>2]=-1;q=(Ak(n,j)|0)>0;x=q?1:l;n=Fx(0,776273)|0;q=s+1|0;if(!n){y=q;z=x;A=s;break}else{s=q;l=x}}k[j>>2]=-1;if((A|0)>-1){l=d+(xx(d)|0)|0;s=761298;n=l+10|0;do{i[l>>0]=i[s>>0]|0;l=l+1|0;s=s+1|0}while((l|0)<(n|0));qx(d,g)|0;s=p+12|0;l=d+(xx(d)|0)|0;if((s+o|0)>>>0>80){i[l>>0]=i[753606]|0;i[l+1>>0]=i[753607]|0;i[l+2>>0]=i[753608]|0;t=p+11|0;u=y;v=z;break}else{i[l>>0]=32;i[l+1>>0]=61;i[l+2>>0]=32;i[l+3>>0]=0;t=s;u=y;v=z;break}}else{B=y;C=z}}else{k[j>>2]=-1;B=0;C=0}if((Ak(g,j)|0)<1){zx(d,g,74)|0;s=d+(xx(d)|0)|0;i[s>>0]=i[753606]|0;i[s+1>>0]=i[753607]|0;i[s+2>>0]=i[753608]|0;t=p+2|0;u=B;v=C;break}uk(5,753609);uk(5,g);k[e>>2]=207;m=207;r=f;return m|0}while(0);do if(!o){if((t|0)==10?(i[d+8>>0]=32,(c|0)!=0):0)zx(d,c,70)|0}else{if((i[b>>0]|0)==39){if(t>>>0>77){uk(5,753631);uk(5,a);uk(5,b);k[e>>2]=207;m=207;r=f;return m|0}zx(d,b,80-t|0)|0;g=t+o|0;C=g>>>0>80?80:g;if((C|0)==80)i[d+79>>0]=39;if(!c)break;if(C>>>0<30&(i[c>>0]|0)!=0){B=g>>>0<80?~g:-81;sy(d+~B|0,32,B+31|0)|0;i[d+30>>0]=0;D=30}else D=C}else{C=t+o|0;if(C>>>0>80){uk(5,753631);uk(5,a);uk(5,b);k[e>>2]=207;m=207;r=f;return m|0}B=C>>>0<30;if(B)zx(d,753691,30-C|0)|0;zx(d,b,80-t|0)|0;D=B?30:C}if((c|0)!=0&D>>>0<77?(i[c>>0]|0)!=0:0){C=d+(xx(d)|0)|0;i[C>>0]=32;i[C+1>>0]=47;i[C+2>>0]=32;i[C+3>>0]=0;zx(d,c,77-D|0)|0}}while(0);if((u|0)==1|(v|0)==1){uk(5,753712);uk(5,d)}m=k[e>>2]|0;r=f;return m|0}function Ek(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;d=r;r=r+96|0;e=d;f=k[a>>2]|0;g=a+4|0;if((f|0)!=(k[(k[g>>2]|0)+68>>2]|0))Fk(a,f+1|0,0,c)|0;Bx(e,b,80)|0;i[e+80>>0]=0;b=xx(e)|0;if(b){f=0;do{h=e+f|0;j=i[h>>0]|0;if(j<<24>>24<32|j<<24>>24==127)i[h>>0]=32;f=f+1|0}while((f|0)!=(b|0));if(b>>>0<80)l=5}else l=5;if((l|0)==5)sy(e+b|0,32,80-b|0)|0;b=vx(e,773860)|0;l=(b|0)==80?8:b;if(l){b=0;do{f=e+b|0;h=(nv(i[f>>0]|0)|0)&255;i[f>>0]=h;b=b+1|0}while((b|0)!=(l|0))}Ak(e,c)|0;l=(k[g>>2]|0)+112|0;g=ky(k[l>>2]|0,k[l+4>>2]|0,-80,-1)|0;au(a,g,L,0,c)|0;cu(a,80,0,e,c)|0;r=d;return k[c>>2]|0}function Fk(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;e=r;r=r+96|0;f=e;g=e+4|0;h=e+8|0;i=k[d>>2]|0;if((i|0)>0){j=i;r=e;return j|0}if((b|0)<1){k[d>>2]=301;j=301;r=e;return j|0}i=a+4|0;l=k[i>>2]|0;do if((k[l+84>>2]|0)>(b|0))m=l;else{n=fy(k[l+88>>2]|0,(b<<3)+8008|0)|0;if(n){o=k[i>>2]|0;k[o+84>>2]=b+1e3;k[o+88>>2]=n;m=o;break}k[d>>2]=113;j=113;r=e;return j|0}while(0);l=k[m+68>>2]|0;k[a>>2]=l;a:do if((l+1|0)==(b|0)){p=m;q=l}else{o=b+-1|0;n=m;while(1){s=(k[n+80>>2]|0)+1|0;t=(s|0)<(b|0)?s:o;s=(k[n+88>>2]|0)+(t<<3)|0;u=k[s+4>>2]|0;v=n+32|0;w=k[v+4>>2]|0;if(!((u|0)<(w|0)|((u|0)==(w|0)?(k[s>>2]|0)>>>0<(k[v>>2]|0)>>>0:0))){x=18;break}do if((wl(a,d)|0)<1){if((k[d>>2]|0)<=0){v=(k[(k[i>>2]|0)+88>>2]|0)+(t<<3)|0;if((au(a,k[v>>2]|0,k[v+4>>2]|0,0,d)|0)<1?(v=k[i>>2]|0,s=v+68|0,w=k[s>>2]|0,u=v+80|0,y=k[u>>2]|0,z=v+96|0,A=z,B=k[A>>2]|0,C=k[A+4>>2]|0,k[s>>2]=t,k[a>>2]=t,k[u>>2]=(y|0)>(t|0)?y:t,u=v+32|0,v=k[u+4>>2]|0,s=z,k[s>>2]=k[u>>2],k[s+4>>2]=v,(dl(a,c,d)|0)>0):0){v=k[i>>2]|0;k[v+68>>2]=w;k[a>>2]=w;k[v+80>>2]=y;y=v+96|0;k[y>>2]=B;k[y+4>>2]=C}if((k[d>>2]|0)<=0)break}k[g>>2]=0;dl(a,c,g)|0;x=19}else x=19;while(0);if((x|0)==19?(x=0,t=k[d>>2]|0,(t|0)>0):0){D=t;break}t=k[i>>2]|0;C=k[t+68>>2]|0;if((C+1|0)==(b|0)){p=t;q=C;break a}else n=t}if((x|0)==18){k[d>>2]=107;j=107;r=e;return j|0}if((D|0)==107){j=107;r=e;return j|0}k[f>>2]=b;Xw(h,753787,f)|0;uk(5,h);j=k[d>>2]|0;r=e;return j|0}while(0);do if((c|0)!=0?(k[d>>2]|0)<=0:0){h=k[a>>2]|0;if((h|0)==0?(f=p+96|0,(k[f>>2]|0)==0&(k[f+4>>2]|0)==0):0){k[c>>2]=0;break}if((h|0)==(q|0)){f=p+120|0;if(((k[f>>2]|0)==-1?(k[f+4>>2]|0)==-1:0)?(Vk(a,d)|0)>0:0)break}else Fk(a,h+1|0,0,d)|0;h=k[i>>2]|0;f=k[h+72>>2]|0;k[c>>2]=f;k[c>>2]=(k[h+1080>>2]|0)==0?f:0}while(0);j=k[d>>2]|0;r=e;return j|0}function Gk(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0;e=r;r=r+32|0;f=e;g=e+8|0;i[c>>0]=0;h=xx(a)|0;if((b|0)<0|(h|0)==0){k[d>>2]=206;j=206;r=e;return j|0}k[f>>2]=b;Xw(g,772183,f)|0;ux(c,a)|0;a=h;do{a=a+-1|0;h=c+a|0;if((i[h>>0]|0)!=32)break;i[h>>0]=0}while((a|0)!=0);qx(c,g)|0;j=k[d>>2]|0;r=e;return j|0}function Hk(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}i[b>>0]=0;e=(c|0)!=0;if(e)i[c>>0]=0;g=xx(a)|0;h=i[a>>0]|0;do if(h<<24>>24==72?(Ax(a,761298,9)|0)==0:0){j=vx(a,773860)|0;if((j|0)!=(g|0)){l=j+1|0;break}a:do if(e&g>>>0>8?(ux(c,a+8|0)|0,j=g+-9|0,(j|0)>-1):0){m=j;while(1){j=c+m|0;if((i[j>>0]|0)!=32)break a;i[j>>0]=0;if((m|0)>0)m=m+-1|0;else break}}while(0);f=k[d>>2]|0;return f|0}else n=13;while(0);b:do if((n|0)==13){c:do if(g>>>0>=9){do if(h<<24>>24<67){if(h<<24>>24==32?(Ax(a,773880,8)|0)==0:0)n=25}else{if(h<<24>>24<=67){if(!(Ax(a,773862,8)|0)){n=25;break}if(!(Ax(a,773889,8)|0)){n=25;break}else break}if(h<<24>>24<72){if(h<<24>>24!=69)break;if(!(Ax(a,753829,8)|0)){n=25;break}else break}else{if(h<<24>>24>72)break;if(!(Ax(a,773871,8)|0)){n=25;break}else break}}while(0);if((n|0)==25){if(!e)break;ux(c,a+8|0)|0;m=g+-9|0;if((m|0)>-1)o=m;else break;while(1){m=c+o|0;if((i[m>>0]|0)!=32)break c;i[m>>0]=0;if((o|0)>0)o=o+-1|0;else break c}}m=a+8|0;if((i[m>>0]|0)==61?(Ax(m,753606,2)|0)==0:0){l=10;break b}j=vx(a,773860)|0;if((j|0)!=(g|0)){l=j+1|0;break b}d:do if(e?(ux(c,m)|0,j=g+-9|0,(j|0)>-1):0){p=j;while(1){j=c+p|0;if((i[j>>0]|0)!=32)break d;i[j>>0]=0;if((p|0)>0)p=p+-1|0;else break}}while(0);f=k[d>>2]|0;return f|0}while(0);f=k[d>>2]|0;return f|0}while(0);o=(Dx(a+l|0,776273)|0)+l|0;if((o|0)==(g|0)){f=k[d>>2]|0;return f|0}l=a+o|0;e:do switch(i[l>>0]|0){case 47:{q=o+1|0;break}case 39:{i[b>>0]=39;n=o+1|0;f:do if(n>>>0<g>>>0){h=o;m=n;p=1;while(1){j=i[a+m>>0]|0;if(j<<24>>24==39){r=h+2|0;s=a+r|0;t=(i[s>>0]|0)==39;i[b+p>>0]=39;if(!t){u=h;v=m;w=p;break f}x=i[s>>0]|0;y=r;z=p+1|0}else{x=j;y=m;z=p}i[b+z>>0]=x;j=z+1|0;r=y+1|0;if(r>>>0<g>>>0){h=y;m=r;p=j}else{u=y;v=r;w=j;break}}}else{u=o;v=n;w=1}while(0);if((v|0)==(g|0)){n=(w|0)<69?w:69;i[b+n>>0]=39;i[b+(n+1)>>0]=0;uk(5,753838);uk(5,a);q=g;break e}else{i[b+(w+1)>>0]=0;q=u+2|0;break e}break}case 40:{n=vx(l,761359)|0;if((n|0)!=(xx(l)|0)){p=n+1|0;Bx(b,l,p)|0;i[b+p>>0]=0;q=p+o|0;break e}uk(5,753886);uk(5,a);k[d>>2]=205;f=205;return f|0}default:{p=vx(l,753933)|0;Bx(b,l,p)|0;i[b+p>>0]=0;q=p+o|0}}while(0);g:do if(e?(o=(Dx(a+q|0,776273)|0)+q|0,o>>>0<80):0){if((i[a+o>>0]|0)==47){b=o+1|0;A=(i[a+b>>0]|0)==32?o+2|0:b}else A=o;qx(c,a+A|0)|0;o=xx(c)|0;if((o|0)>0){b=o;do{o=b;b=b+-1|0;l=c+b|0;if((i[l>>0]|0)!=32)break g;i[l>>0]=0}while((o|0)>1)}}while(0);f=k[d>>2]|0;return f|0}function Ik(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0.0,H=0.0;e=r;r=r+512|0;f=e+428|0;g=e+8|0;h=e+288|0;l=e+4|0;m=e+148|0;n=e;o=k[d>>2]|0;if((o|0)>0){p=o;r=e;return p|0}i[b>>0]=0;k[c>>2]=0;if((i[a>>0]|0)==32?(Ax(a,773880,8)|0)==0:0){zx(b,a,80)|0;k[c>>2]=1;p=k[d>>2]|0;r=e;return p|0}i[f>>0]=0;i[g>>0]=0;i[h>>0]=0;o=Dx(a,776273)|0;q=a+o|0;if(!(Ax(q,753936,20)|0)){k[d>>2]=207;p=207;r=e;return p|0}if((i[q>>0]|0)==45){k[c>>2]=-1;s=o+1|0;t=Dx(a+s|0,776273)|0;u=t+s|0;s=a+u|0;if((t|0)<8){t=vx(s,753957)|0;if((t|0)>74){k[d>>2]=207;p=207;r=e;return p|0}zx(b,s,t)|0;if((t|0)<9){v=xx(b)|0;if(v){w=0;do{x=b+w|0;y=(nv(i[x>>0]|0)|0)&255;i[x>>0]=y;w=w+1|0}while((w|0)!=(v|0))}k[n>>2]=0;if((Ak(b,n)|0)>0){i[b>>0]=0;zx(b,s,t)|0}}v=t+u|0;z=v;A=a+v|0}else{z=u;A=s}s=a+((Dx(A,776273)|0)+z)|0;switch(i[s>>0]|0){case 61:case 0:{p=k[d>>2]|0;r=e;return p|0}default:{}}k[c>>2]=-2;z=vx(s,776273)|0;if((z|0)>40){k[d>>2]=207;p=207;r=e;return p|0}A=b+(xx(b)|0)|0;u=770037;v=A+41|0;do{i[A>>0]=i[u>>0]|0;A=A+1|0;u=u+1|0}while((A|0)<(v|0));u=b+40|0;Bx(u,s,z+1|0)|0;if((z|0)<9){A=xx(u)|0;if(A){v=0;do{t=b+(v+40)|0;w=(nv(i[t>>0]|0)|0)&255;i[t>>0]=w;v=v+1|0}while((v|0)!=(A|0))}k[n>>2]=0;if((Ak(u,n)|0)>0)Bx(u,s,z)|0}}else{z=vx(q,753957)|0;if((z|0)>74){k[d>>2]=207;p=207;r=e;return p|0}zx(f,q,z)|0;if((z|0)<9){s=xx(f)|0;if(s){u=0;do{A=f+u|0;v=(nv(i[A>>0]|0)|0)&255;i[A>>0]=v;u=u+1|0}while((u|0)!=(s|0))}k[n>>2]=0;if((Ak(f,n)|0)>0){i[f>>0]=0;zx(f,q,z)|0}}q=i[f>>0]|0;a:do if(q<<24>>24==69){if(tx(f,768779)|0){n=z+o|0;B=n;C=a+n|0;break}i[b>>0]=69;i[b+1>>0]=78;i[b+2>>0]=68;i[b+3>>0]=0;k[c>>2]=2;p=k[d>>2]|0;r=e;return p|0}else{n=z+o|0;s=a+n|0;switch(q<<24>>24){case 67:{if(tx(f,782848)|0){B=n;C=s;break a}break}case 72:{if((tx(f,782840)|0)!=0?(tx(f,767038)|0)!=0:0){B=n;C=s;break a}break}default:{B=n;C=s;break a}}k[c>>2]=1;ux(b,f)|0;zx(b,s,73)|0;p=k[d>>2]|0;r=e;return p|0}while(0);c=(Dx(C,753957)|0)+B|0;B=a+c|0;b:do switch(i[B>>0]|0){case 39:{C=B;while(1){q=vx(C+1|0,753960)|0;o=q+2|0;zx(g,C,o)|0;if((i[C+(q+1)>>0]|0)!=39)break;q=C+o|0;if((i[q>>0]|0)==39)C=q;else{D=q;break b}}k[d>>2]=205;p=205;r=e;return p|0}case 0:case 47:{C=g+(xx(g)|0)|0;i[C>>0]=32;i[C+1>>0]=0;D=B;break}default:{C=vx(B,753933)|0;zx(g,B,C)|0;switch(i[B>>0]|0){case 70:case 84:{switch(i[a+(c+1)>>0]|0){case 0:case 47:case 32:{E=C;break}default:F=49}break}default:F=49}c:do if((F|0)==49){G=+hx(g,l);d:do switch(i[k[l>>2]>>0]|0){case 47:case 32:case 0:{H=G;break}default:{ux(m,g)|0;q=rx(m,68)|0;if(q){i[q>>0]=69;H=+hx(m,l);break d}q=rx(m,100)|0;if(q){i[q>>0]=69;H=+hx(m,l);break d}q=rx(m,46)|0;if(!q)H=G;else{i[q>>0]=44;H=+hx(m,l)}}}while(0);switch(i[k[l>>2]>>0]|0){case 47:case 32:case 0:break;default:{j[g>>1]=39;zx(g,B,C)|0;q=g+(xx(g)|0)|0;i[q>>0]=39;i[q+1>>0]=0;if(!(H==0.0)){E=C;break c}E=~~H+C|0;break c}}q=rx(g,101)|0;if(q){i[q>>0]=69;E=C;break}q=rx(g,100)|0;if(!q)E=C;else{i[q>>0]=68;E=C}}while(0);D=a+(E+c)|0}}while(0);c=D+(Dx(D,753933)|0)|0;D=xx(g)|0;E=D+-1|0;if(E>>>0<9&(i[g>>0]|0)==39){i[g+E>>0]=0;zx(g,773880,10-D|0)|0;D=g+((xx(g+9|0)|0)+9)|0;i[D>>0]=39;i[D+1>>0]=0}zx(h,c,70)|0;Dk(f,g,h,b,d)|0}p=k[d>>2]|0;r=e;return p|0}function Jk(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0;p=k[o>>2]|0;if((p|0)>0){q=p;return q|0}if((a|0)==0|(b|0)==0){k[o>>2]=115;q=115;return q|0}i[b>>0]=0;if(!(i[a>>0]|0)){p=a;r=773880;s=p+9|0;do{i[p>>0]=i[r>>0]|0;p=p+1|0;r=r+1|0}while((p|0)<(s|0))}a:do if((d|0)>0){r=(g|0)==0;p=(g|0)!=-1;s=(g|0)!=1;t=0;u=0;v=0;while(1){w=k[c+(v<<3)>>2]|0;x=i[w>>0]|0;if(x<<24>>24==42){y=32;z=0;A=0;B=-1;C=-1;D=1;E=v;break a}b:do if(!((t|0)!=0?x<<24>>24==u<<24>>24:0))if(x<<24>>24){F=x;G=32;H=1;I=0;J=0;K=0;L=0;M=-1;N=-1;O=0;while(1){P=i[a+J>>0]|0;c:do switch(F<<24>>24){case 105:{Q=(P<<24>>24)+-48|0;R=Q>>>0<10;S=R?Q:I;if(R){T=G;U=S;V=J;W=L;X=M;Y=N;Z=1}else{_=G;$=S;aa=L;ba=26}break}case 106:{S=(P<<24>>24)+-48|0;R=S>>>0<10;Q=R?S:L;if(R){T=G;U=I;V=J;W=Q;X=M;Y=N;Z=1}else{_=G;$=I;aa=Q;ba=26}break}default:{Q=F<<24>>24==110;switch(F<<24>>24){case 63:{T=G;U=I;V=J;W=L;X=M;Y=N;Z=1;break c;break}case 35:case 109:case 110:{if(((P<<24>>24)+-48|0)>>>0>=10){ca=G;da=H;ea=I;fa=L;ga=M;ha=N;ia=x;ja=0;break b}d:do if((J|0)<8){R=P;S=J;ka=0;while(1){la=(R<<24>>24)+-48|0;if(la>>>0>=10){ma=S;na=ka;break d}oa=(ka*10|0)+la|0;la=S+1|0;if((la|0)<8){R=i[a+la>>0]|0;S=la;ka=oa}else{ma=la;na=oa;break}}}else{ma=J;na=0}while(0);ka=ma+-1|0;if(!Q){T=G;U=I;V=ka;W=L;X=F<<24>>24==109?na:M;Y=N;Z=1;break c}if((na+-1|0)>>>0>=999){ca=G;da=H;ea=I;fa=L;ga=M;ha=N;ia=x;ja=0;break b}if(r&(na|0)==(e|0)){T=G;U=I;V=ka;W=L;X=M;Y=e;Z=1;break c}if((p|(na|0)>(e|0))&(s|(na|0)<(e|0))){ca=G;da=H;ea=I;fa=L;ga=M;ha=N;ia=x;ja=0;break b}else{T=G;U=I;V=ka;W=L;X=M;Y=na;Z=1;break c}break}case 97:{ka=P<<24>>24==32|(lv(P<<24>>24)|0)!=0;S=ka?P:G;if(ka){T=S;U=I;V=J;W=L;X=M;Y=N;Z=1;break c}else{_=S;$=I;aa=L;ba=26;break c}break}default:if(P<<24>>24==F<<24>>24){T=G;U=I;V=J;W=L;X=M;Y=N;Z=1;break c}else{ca=G;da=H;ea=I;fa=L;ga=M;ha=N;ia=x;ja=0;break b}}}}while(0);if((ba|0)==26){ba=0;if(!O){ca=_;da=H;ea=$;fa=aa;ga=M;ha=N;ia=x;ja=0;break b}else{T=_;U=$;V=J;W=aa;X=M;Y=N;Z=O}}K=K+1|0;P=V+1|0;F=i[w+K>>0]|0;if(!((P|0)<8&F<<24>>24!=0)){pa=P;qa=T;ra=U;sa=V;ta=W;ua=X;va=Y;wa=Z;break}else{G=T;H=0;I=U;J=P;L=W;M=X;N=Y;O=Z}}if((sa|0)>6){y=qa;z=ra;A=ta;B=ua;C=va;D=wa;E=v;break a}if((i[a+pa>>0]|0)==32){y=qa;z=ra;A=ta;B=ua;C=va;D=wa;E=v;break a}else{ca=qa;da=0;ea=ra;fa=ta;ga=ua;ha=va;ia=x;ja=wa}}else{ca=32;da=1;ea=0;fa=0;ga=-1;ha=-1;ia=0;ja=0}else{ca=32;da=t;ea=0;fa=0;ga=-1;ha=-1;ia=u;ja=0}while(0);x=v+1|0;if((x|0)<(d|0)){t=da;u=ia;v=x}else{y=ca;z=ea;A=fa;B=ga;C=ha;D=ja;E=x;break}}}else{y=32;z=0;A=0;B=0;C=0;D=0;E=0}while(0);if(j)k[j>>2]=z;if(l)k[l>>2]=A;if(n)k[n>>2]=C;if(m)k[m>>2]=B;if(h)k[h>>2]=E;h=k[c+(E<<3)+4>>2]|0;if(!D){q=0;return q|0}switch(i[h>>0]|0){case 45:case 0:{q=0;return q|0}default:{}}ux(b,a)|0;a=i[h>>0]|0;if(a<<24>>24==43){q=0;return q|0}D=(B|0)>0;E=z+48&255;z=A+48&255;e:do if((B|0)>-1){A=a;c=0;m=0;n=C;while(1){f:do switch(A<<24>>24){case 0:{xa=c;break e;break}case 105:{i[b+c>>0]=E;ya=c;za=n;break}case 106:{i[b+c>>0]=z;ya=c;za=n;break}case 110:{l=(n|0)==-1?e:n;if((l|0)>0){j=l+f|0;if((j|0)>0){ja=1;while(1){ha=ja*10|0;if(((j|0)/(ha|0)|0|0)>0)ja=ha;else{Aa=ha;break}}if((Aa|0)>9){ja=Aa;ha=c;while(1){ga=ja;ja=(ja|0)/10|0;i[b+ha>>0]=(((j|0)/(ja|0)|0|0)%10|0)+48;fa=ha+1|0;if((ga|0)<=99){Ba=fa;break}else ha=fa}}else Ba=c}else Ba=c;ya=Ba+-1|0;za=j}else{ya=c;za=l}break}case 109:{if(D){ha=1;while(1){ja=ha*10|0;if(((B|0)/(ja|0)|0|0)>0)ha=ja;else{Ca=ja;break}}if((Ca|0)>9){ha=Ca;l=c;while(1){j=ha;ha=(ha|0)/10|0;i[b+l>>0]=(((B|0)/(ha|0)|0|0)%10|0)+48;ja=l+1|0;if((j|0)<=99){Da=ja;break}else l=ja}}else Da=c}else Da=c;ya=Da+-1|0;za=n;break}default:{l=b+c|0;if(A<<24>>24==97){i[l>>0]=y;ya=c;za=n;break f}else{i[l>>0]=A;ya=c;za=n;break f}}}while(0);l=m+1|0;A=i[h+l>>0]|0;c=ya+1|0;m=l;n=za}}else{n=a;m=0;c=0;A=C;while(1){g:do switch(n<<24>>24){case 0:{xa=m;break e;break}case 105:{i[b+m>>0]=E;Ea=m;Fa=A;break}case 106:{i[b+m>>0]=z;Ea=m;Fa=A;break}case 110:{l=(A|0)==-1?e:A;if((l|0)>0){ha=l+f|0;if((ha|0)>0){ja=1;while(1){j=ja*10|0;if(((ha|0)/(j|0)|0|0)>0)ja=j;else{Ga=j;break}}if((Ga|0)>9){ja=Ga;j=m;while(1){fa=ja;ja=(ja|0)/10|0;i[b+j>>0]=(((ha|0)/(ja|0)|0|0)%10|0)+48;ga=j+1|0;if((fa|0)<=99){Ha=ga;break}else j=ga}}else Ha=m}else Ha=m;Ea=Ha+-1|0;Fa=ha}else{Ea=m;Fa=l}break}default:{j=b+m|0;if(n<<24>>24==97){i[j>>0]=y;Ea=m;Fa=A;break g}else{i[j>>0]=n;Ea=m;Fa=A;break g}}}while(0);j=c+1|0;n=i[h+j>>0]|0;m=Ea+1|0;c=j;A=Fa}}while(0);if((xa|0)<8)sy(b+xa|0,32,8-xa|0)|0;q=k[o>>2]|0;return q|0}function Kk(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0;l=r;r=r+192|0;m=l+24|0;n=l+109|0;o=l+16|0;p=l+12|0;q=l+8|0;s=l+4|0;t=l;u=l+28|0;k[o>>2]=0;k[p>>2]=0;k[q>>2]=0;k[s>>2]=0;k[t>>2]=0;v=k[j>>2]|0;if((v|0)>0){w=v;r=l;return w|0}Co(a,m,l+20|0,j)|0;if((k[m>>2]|0)>=(c|0)){v=n+8|0;x=u+8|0;y=c;while(1){i[u>>0]=0;Ro(a,y,n,j)|0;c=xx(n)|0;if((c|0)>8){z=8;do{A=n+z|0;B=i[A>>0]|0;if(B<<24>>24<32|B<<24>>24==127)i[A>>0]=32;z=z+1|0}while((z|0)!=(c|0))}Jk(n,u,d,e,f,g,h,t,o,p,s,q,j)|0;if(i[u>>0]|0)ds(b,u,j)|0;i[v>>0]=0;i[x>>0]=0;if((y|0)<(k[m>>2]|0))y=y+1|0;else break}}w=k[j>>2]|0;r=l;return w|0}function Lk(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;d=r;r=r+992|0;g=d+820|0;h=d+905|0;j=d+824|0;l=d+20|0;m=d+16|0;n=d+12|0;o=d+8|0;p=d+4|0;q=d;s=d+24|0;k[l>>2]=0;py(s|0,120296,792)|0;t=k[f>>2]|0;if((t|0)>0){u=t;r=d;return u|0}Co(a,g,d+816|0,f)|0;if((k[g>>2]|0)>=(c|0)){t=h+8|0;v=j+8|0;w=c;while(1){i[j>>0]=0;Ro(a,w,h,f)|0;Mk(h,j,s,99,0,e,l,m,n,o,p,q,f)|0;if(i[j>>0]|0)ds(b,j,f)|0;i[t>>0]=0;i[v>>0]=0;if((w|0)<(k[g>>2]|0))w=w+1|0;else break}}u=k[f>>2]|0;r=d;return u|0}function Mk(a,b,c,d,e,f,g,h,j,l,m,n,o){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;var p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0;e=k[o>>2]|0;if((e|0)>0){p=e;return p|0}if((a|0)==0|(b|0)==0){k[o>>2]=115;p=115;return p|0}i[b>>0]=0;if(!(i[a>>0]|0)){p=0;return p|0}a:do if((d|0)>0){e=f+4|0;q=f+8|0;r=f+12|0;s=0;t=0;u=0;v=0;w=0;x=0;while(1){y=k[c+(x<<3)>>2]|0;z=i[y>>0]|0;if(z<<24>>24==42){A=32;B=0;C=0;D=t;E=u;F=v;G=1;H=x;break a}b:do if(!((s|0)!=0?z<<24>>24==w<<24>>24:0))if(z<<24>>24){I=z;J=32;K=1;L=0;M=0;N=0;O=0;P=t;Q=u;R=v;S=0;while(1){T=i[a+M>>0]|0;c:do switch(I<<24>>24){case 105:{U=(T<<24>>24)+-48|0;V=U>>>0<10;W=V?U:L;if(V){X=J;Y=W;Z=M;_=O;$=P;aa=Q;ba=R;ca=1}else{da=J;ea=W;fa=M;ga=O;ha=Q;ia=R;ja=S;ka=28}break}case 106:{W=(T<<24>>24)+-48|0;V=W>>>0<10;U=V?W:O;if(V){X=J;Y=L;Z=M;_=U;$=P;aa=Q;ba=R;ca=1}else{da=J;ea=L;fa=M;ga=U;ha=Q;ia=R;ja=S;ka=28}break}default:{U=I<<24>>24==110;switch(I<<24>>24){case 63:{X=J;Y=L;Z=M;_=O;$=P;aa=Q;ba=R;ca=1;break c;break}case 35:case 108:case 109:case 110:{if(((T<<24>>24)+-48|0)>>>0>=10){la=J;ma=K;na=L;oa=O;pa=P;qa=Q;ra=R;sa=z;ta=0;break b}d:do if((M|0)<8){V=T;W=M;ua=0;while(1){va=(V<<24>>24)+-48|0;if(va>>>0>=10){wa=W;xa=ua;break d}ya=va+(ua*10|0)|0;va=W+1|0;if((va|0)<8){V=i[a+va>>0]|0;W=va;ua=ya}else{wa=va;xa=ya;break}}}else{wa=M;xa=0}while(0);ua=wa+-1|0;if((I+-109&255)>=2){X=J;Y=L;Z=ua;_=O;$=I<<24>>24==108?xa:P;aa=Q;ba=R;ca=1;break c}if((xa+-1|0)>>>0>=999){la=J;ma=K;na=L;oa=O;pa=P;qa=Q;ra=R;sa=z;ta=0;break b}do if((xa|0)!=(k[f>>2]|0))if((xa|0)==(k[e>>2]|0)){za=1;Aa=2}else{if((xa|0)==(k[q>>2]|0)){za=1;Aa=3;break}W=(xa|0)==(k[r>>2]|0);za=W&1;Aa=W?4:0}else{za=1;Aa=1}while(0);da=J;ea=L;fa=ua;ga=O;ha=U?Q:Aa;ia=U?Aa:R;ja=za;ka=28;break c;break}case 97:{W=T<<24>>24==32|(lv(T<<24>>24)|0)!=0;V=W?T:J;if(W){X=V;Y=L;Z=M;_=O;$=P;aa=Q;ba=R;ca=1;break c}else{da=V;ea=L;fa=M;ga=O;ha=Q;ia=R;ja=S;ka=28;break c}break}default:if(T<<24>>24==I<<24>>24){X=J;Y=L;Z=M;_=O;$=P;aa=Q;ba=R;ca=1;break c}else{la=J;ma=K;na=L;oa=O;pa=P;qa=Q;ra=R;sa=z;ta=0;break b}}}}while(0);if((ka|0)==28){ka=0;if(!ja){la=da;ma=K;na=ea;oa=ga;pa=P;qa=ha;ra=ia;sa=z;ta=0;break b}else{X=da;Y=ea;Z=fa;_=ga;$=P;aa=ha;ba=ia;ca=ja}}N=N+1|0;T=Z+1|0;I=i[y+N>>0]|0;if(!((T|0)<8&I<<24>>24!=0)){Ba=T;Ca=X;Da=Y;Ea=Z;Fa=_;Ga=$;Ha=aa;Ia=ba;Ja=ca;break}else{J=X;K=0;L=Y;M=T;O=_;P=$;Q=aa;R=ba;S=ca}}if((Ea|0)>6){A=Ca;B=Da;C=Fa;D=Ga;E=Ha;F=Ia;G=Ja;H=x;break a}if((i[a+Ba>>0]|0)==32){A=Ca;B=Da;C=Fa;D=Ga;E=Ha;F=Ia;G=Ja;H=x;break a}else{la=Ca;ma=0;na=Da;oa=Fa;pa=Ga;qa=Ha;ra=Ia;sa=z;ta=Ja}}else{la=32;ma=1;na=0;oa=0;pa=t;qa=u;ra=v;sa=0;ta=0}else{la=32;ma=s;na=0;oa=0;pa=t;qa=u;ra=v;sa=w;ta=0}while(0);z=x+1|0;if((z|0)<(d|0)){s=ma;t=pa;u=qa;v=ra;w=sa;x=z}else{A=la;B=na;C=oa;D=pa;E=qa;F=ra;G=ta;H=z;break}}}else{A=32;B=0;C=0;D=0;E=0;F=0;G=0;H=0}while(0);if(h)k[h>>2]=B;if(j)k[j>>2]=C;if(l)k[l>>2]=F;if(m)k[m>>2]=E;if(n)k[n>>2]=D;if(g)k[g>>2]=H;g=k[c+(H<<3)+4>>2]|0;if(!G){p=0;return p|0}switch(i[g>>0]|0){case 45:case 0:{p=0;return p|0}default:{}}ux(b,a)|0;a=i[g>>0]|0;if(a<<24>>24==43){p=0;return p|0}G=(F|0)>0;H=(E|0)>0;c=(D|0)>-1;n=(D|0)>0;m=B+48&255;B=C+48&255;C=a;a=0;l=0;e:while(1){f:do switch(C<<24>>24){case 0:{Ka=a;break e;break}case 105:{i[b+a>>0]=m;La=a;break}case 106:{i[b+a>>0]=B;La=a;break}default:{if(G&C<<24>>24==110){j=1;while(1){h=j*10|0;if(((F|0)/(h|0)|0|0)>0)j=h;else{Ma=h;break}}if((Ma|0)>9){j=Ma;h=a;while(1){ta=j;j=(j|0)/10|0;i[b+h>>0]=(((F|0)/(j|0)|0|0)%10|0)+48;ra=h+1|0;if((ta|0)<=99){Na=ra;break}else h=ra}}else Na=a;La=Na+-1|0;break f}if(H&C<<24>>24==109){h=1;while(1){j=h*10|0;if(((E|0)/(j|0)|0|0)>0)h=j;else{Oa=j;break}}if((Oa|0)>9){h=Oa;j=a;while(1){ra=h;h=(h|0)/10|0;i[b+j>>0]=(((E|0)/(h|0)|0|0)%10|0)+48;ta=j+1|0;if((ra|0)<=99){Pa=ta;break}else j=ta}}else Pa=a;La=Pa+-1|0;break f}if(!(c&C<<24>>24==108)){j=b+a|0;if(C<<24>>24==97){i[j>>0]=A;La=a;break f}else{i[j>>0]=C;La=a;break f}}if(n){j=1;while(1){h=j*10|0;if(((D|0)/(h|0)|0|0)>0)j=h;else{Qa=h;break}}if((Qa|0)>9){j=Qa;h=a;while(1){ta=j;j=(j|0)/10|0;i[b+h>>0]=(((D|0)/(j|0)|0|0)%10|0)+48;ra=h+1|0;if((ta|0)<=99){Ra=ra;break}else h=ra}}else Ra=a}else Ra=a;La=Ra+-1|0}}while(0);h=l+1|0;C=i[g+h>>0]|0;a=La+1|0;l=h}if((Ka|0)<8)sy(b+Ka|0,32,8-Ka|0)|0;p=k[o>>2]|0;return p|0}function Nk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0.0,J=0,K=0,L=0,M=0,N=0,O=0;f=r;r=r+256|0;g=f+8|0;h=f;j=f+16|0;l=f+172|0;m=f+12|0;n=f+101|0;p=f+20|0;q=k[e>>2]|0;if((q|0)>0){s=q;r=f;return s|0}q=(b|0)!=0;if(q)k[b>>2]=0;t=(c|0)!=0;if(t)k[c>>2]=0;u=(d|0)!=0;if(u){k[d>>2]=0;v=0}else v=0;while(1){w=a+v|0;if((i[w>>0]|0)==32)v=v+1|0;else{x=w;break}}ux(n,x)|0;x=xx(n)|0;if(x){v=0;do{w=n+v|0;y=(nv(i[w>>0]|0)|0)&255;i[w>>0]=y;v=v+1|0}while((v|0)!=(x|0))}switch(i[n>>0]|0){case 0:{uk(5,753962);k[e>>2]=261;s=261;r=f;return s|0}case 65:{z=16;break}case 73:{z=41;break}case 70:{z=42;break}case 69:{z=42;break}case 68:{z=82;break}default:{k[h>>2]=a;Xw(p,754001,h)|0;uk(5,p);k[e>>2]=262;s=262;r=f;return s|0}}if(q)k[b>>2]=z;h=n+1|0;x=(z|0)==41;a:do switch(z|0){case 16:case 41:{if((k[e>>2]|0)>0){A=0;B=z;C=47;break a}v=rv()|0;k[v>>2]=0;y=jx(h,j,10)|0;switch(i[k[j>>2]>>0]|0){case 32:case 0:break;default:k[e>>2]=407}if((k[v>>2]|0)==34){D=l;E=754043;F=D+54|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0));zx(l,h,25)|0;uk(5,l);k[e>>2]=412;k[v>>2]=0}if((k[e>>2]|0)<1)if((y|0)<1){k[e>>2]=261;G=0;H=z;C=48;break a}else{A=y;B=x&(y|0)<5?21:z;C=47;break a}else{A=y;B=z;C=47}break}default:if((Pk(h,m,e)|0)<1){I=+o[m>>2];if(I<=0.0){k[e>>2]=261;G=0;H=z;C=48;break a}w=~~I;J=(w|0)>7&(i[n>>0]|0)==70?82:z;K=(w|0)<10;if((K?i[n+2>>0]|0:i[n+3>>0]|0)<<24>>24==46){L=n+(K?3:4)|0;if((k[e>>2]|0)>0){A=w;B=J;C=47;break a}K=rv()|0;k[K>>2]=0;M=jx(L,j,10)|0;switch(i[k[j>>2]>>0]|0){case 32:case 0:break;default:k[e>>2]=407}if((k[K>>2]|0)==34){D=l;E=754043;F=D+54|0;do{i[D>>0]=i[E>>0]|0;D=D+1|0;E=E+1|0}while((D|0)<(F|0));zx(l,L,25)|0;uk(5,l);k[e>>2]=412;k[K>>2]=0}if((k[e>>2]|0)<1){if(u)k[d>>2]=M;if((M|0)>=(w|0))k[e>>2]=261;A=w;B=(M|0)>6&(i[n>>0]|0)==69?82:J;C=47}else{A=w;B=J;C=47}}else{A=w;B=J;C=47}}else{A=0;B=z;C=47}}while(0);if((C|0)==47)if((k[e>>2]|0)>0){G=A;H=B;C=48}else{N=A;O=B}if((C|0)==48){k[e>>2]=261;k[g>>2]=a;Xw(p,754097,g)|0;uk(5,p);N=G;O=H}if(q)k[b>>2]=O;if(t)k[c>>2]=N;s=k[e>>2]|0;r=f;return s|0}function Ok(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0;d=r;r=r+96|0;e=d;f=d+4|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=rv()|0;k[g>>2]=0;k[b>>2]=0;j=jx(a,e,10)|0;k[b>>2]=j;switch(i[k[e>>2]>>0]|0){case 32:case 0:break;default:k[c>>2]=407}if((k[g>>2]|0)==34){e=f;j=754043;b=e+54|0;do{i[e>>0]=i[j>>0]|0;e=e+1|0;j=j+1|0}while((e|0)<(b|0));zx(f,a,25)|0;uk(5,f);k[c>>2]=412;k[g>>2]=0}h=k[c>>2]|0;r=d;return h|0}function Pk(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,l=0,m=0,n=0,p=0,q=0.0,s=0.0,u=0,v=0,w=0,x=0;d=r;r=r+160|0;e=d;f=d+77|0;g=d+4|0;h=k[c>>2]|0;if((h|0)>0){l=h;r=d;return l|0}h=i[754135]|0;if(!(h<<24>>24)){m=Av()|0;n=i[k[m>>2]>>0]|0;i[754135]=n;p=n}else p=h;h=rv()|0;k[h>>2]=0;o[b>>2]=0.0;if((rx(a,68)|0)!=0|p<<24>>24==44){ux(g,a)|0;p=rx(g,68)|0;k[e>>2]=p;if(p)i[p>>0]=69;if((i[754135]|0)==44?(p=rx(g,46)|0,k[e>>2]=p,(p|0)!=0):0)i[p>>0]=44;q=+hx(g,e)}else q=+hx(a,e);s=q;o[b>>2]=s;switch(i[k[e>>2]>>0]|0){case 32:case 0:{u=(o[t>>2]=s,k[t>>2]|0)>>>16&65535;break}default:{v=f;w=754136;x=v+45|0;do{i[v>>0]=i[w>>0]|0;v=v+1|0;w=w+1|0}while((v|0)<(x|0));zx(f,a,30)|0;uk(5,f);k[c>>2]=408;u=j[b+2>>1]|0}}if((k[h>>2]|0)==34|(u&32640)==32640){v=f;w=754136;x=v+45|0;do{i[v>>0]=i[w>>0]|0;v=v+1|0;w=w+1|0}while((v|0)<(x|0));zx(f,a,30)|0;uk(5,f);o[b>>2]=0.0;k[c>>2]=412;k[h>>2]=0}l=k[c>>2]|0;r=d;return l|0}function Qk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;f=r;r=r+192|0;g=f+16|0;h=f+8|0;j=f;l=f+24|0;m=f+20|0;n=f+109|0;o=f+28|0;p=k[e>>2]|0;if((p|0)>0){q=p;r=f;return q|0}p=(b|0)!=0;if(p)k[b>>2]=0;s=(c|0)!=0;if(s)k[c>>2]=0;t=(d|0)!=0;if(t)k[d>>2]=0;u=xx(a)|0;a:do if(!u)v=0;else{w=0;while(1){if((i[a+w>>0]|0)!=32){v=w;break a}x=w+1|0;if(x>>>0<u>>>0)w=x;else{v=x;break}}}while(0);if((v|0)==(u|0)){uk(5,754181);k[e>>2]=261;q=261;r=f;return q|0}ux(n,a+v|0)|0;v=xx(n)|0;if(!v)y=0;else{u=0;do{w=n+u|0;x=(nv(i[w>>0]|0)|0)&255;i[w>>0]=x;u=u+1|0}while((u|0)!=(v|0));y=0}while(1){v=n+y|0;u=i[v>>0]|0;x=y+1|0;if(((u<<24>>24)+-48|0)>>>0<10)y=x;else{z=v;A=u;B=x;C=y;break}}if(!C){k[m>>2]=1;D=A}else{k[j>>2]=m;Yw(n,773048,j)|0;D=i[z>>0]|0}j=(D&-2)<<24>>24==80;D=j?n+B|0:z;b:do switch(i[D>>0]|0){case 85:{k[l>>2]=2;E=20;break}case 73:{k[l>>2]=2;E=21;break}case 86:{k[l>>2]=4;E=40;break}case 74:{k[l>>2]=4;E=41;break}case 75:{k[l>>2]=8;E=81;break}case 69:{k[l>>2]=4;E=42;break}case 68:{k[l>>2]=8;E=82;break}case 65:{z=n+((j?B:C)+1)|0;switch(i[z>>0]|0){case 0:break;case 40:{F=z;G=29;break}default:{F=D;G=29}}if((G|0)==29?(k[h>>2]=l,(Yw(F+1|0,773048,h)|0)==1):0){if(j){E=16;break b}if((k[l>>2]|0)<=(k[m>>2]|0)){E=16;break b}}k[l>>2]=k[m>>2];E=16;break}case 76:{k[l>>2]=1;E=14;break}case 88:{k[l>>2]=1;E=1;break}case 66:{k[l>>2]=1;E=11;break}case 83:{k[l>>2]=1;E=12;break}case 67:{k[l>>2]=8;E=83;break}case 77:{k[l>>2]=16;E=163;break}default:{k[g>>2]=a;Xw(o,754231,g)|0;uk(5,o);k[e>>2]=262;q=262;r=f;return q|0}}while(0);if(p)k[b>>2]=j?0-E|0:E;if(s)k[c>>2]=k[m>>2];if(t)k[d>>2]=k[l>>2];q=k[e>>2]|0;r=f;return q|0}function Rk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0;i[b>>0]=0;c=0;do{d=a+c|0;e=i[d>>0]|0;f=e<<24>>24==0;c=c+1|0}while(e<<24>>24==32&(f^1));if(f)return;i[b>>0]=37;ux(b+1|0,a+c|0)|0;c=i[d>>0]|0;switch(c<<24>>24){case 65:{a=b+(xx(b)|0)|0;i[a>>0]=115;i[a+1>>0]=0;g=7;break}case 73:{a=b+(xx(b)|0)|0;i[a>>0]=46;i[a+1>>0]=48;i[a+2>>0]=102;i[a+3>>0]=0;g=7;break}default:h=c}if((g|0)==7)h=i[d>>0]|0;if(h<<24>>24==70){g=b+(xx(b)|0)|0;i[g>>0]=102;i[g+1>>0]=0;j=i[d>>0]|0}else j=h;if(j<<24>>24==69){h=b+(xx(b)|0)|0;i[h>>0]=69;i[h+1>>0]=0;k=i[d>>0]|0}else k=j;if(k<<24>>24!=68)return;k=b+(xx(b)|0)|0;i[k>>0]=69;i[k+1>>0]=0;return}function Sk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0;i[b>>0]=0;c=0;do{d=a+c|0;e=i[d>>0]|0;f=e<<24>>24==0;c=c+1|0}while(e<<24>>24==32&(f^1));if(f){i[b>>0]=0;return}if(rx(d,37)|0){i[b>>0]=0;return}i[b>>0]=37;ux(b+1|0,a+c|0)|0;switch(i[d>>0]|0){case 97:case 65:{d=b+(xx(b)|0)|0;i[d>>0]=115;i[d+1>>0]=0;return}case 105:case 73:{d=b+(xx(b)|0)|0;i[d>>0]=100;i[d+1>>0]=0;return}case 111:case 79:{d=b+(xx(b)|0)|0;i[d>>0]=111;i[d+1>>0]=0;return}case 122:case 90:{d=b+(xx(b)|0)|0;i[d>>0]=88;i[d+1>>0]=0;return}case 102:case 70:{d=b+(xx(b)|0)|0;i[d>>0]=102;i[d+1>>0]=0;return}case 101:case 69:{d=b+(xx(b)|0)|0;i[d>>0]=69;i[d+1>>0]=0;return}case 100:case 68:{d=b+(xx(b)|0)|0;i[d>>0]=69;i[d+1>>0]=0;return}case 103:case 71:{d=b+(xx(b)|0)|0;i[d>>0]=71;i[d+1>>0]=0;return}default:{i[b>>0]=0;return}}}function Tk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0;f=r;r=r+80|0;Uk(a,b,c,f,d,e)|0;r=f;return k[e>>2]|0}function Uk(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0;g=r;r=r+192|0;h=g;j=g+12|0;l=g+97|0;m=g+16|0;n=g+8|0;o=g+4|0;p=k[f>>2]|0;do if((p|0)>=1)if((p|0)==237){k[f>>2]=0;q=a+4|0;s=237;break}else{t=p;r=g;return t|0}else{u=a+4|0;k[(k[u>>2]|0)+932>>2]=0;q=u;s=0}while(0);i[d>>0]=0;k[e>>2]=0;p=k[a>>2]|0;u=k[q>>2]|0;if((p|0)==(k[u+68>>2]|0)){v=u+120|0;if(((k[v>>2]|0)==-1?(k[v+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){t=k[f>>2]|0;r=g;return t|0}}else Fk(a,p+1|0,0,f)|0;p=k[q>>2]|0;a=k[p+932>>2]|0;do if((a|0)<(k[p+928>>2]|0)){v=(k[p+960>>2]|0)+(a*160|0)|0;u=0;w=0;x=a;y=0;a:while(1){Wk(c,v,b,n,o);do if(k[n>>2]|0){z=(u|0)!=0;A=(k[o>>2]|0)!=0;if(z&A){B=14;break a}if(!z){if(A){ux(d,v)|0;k[e>>2]=x+1;C=1;D=w;E=y;break}if(!w){ux(d,v)|0;A=x+1|0;k[e>>2]=A;k[(k[q>>2]|0)+932>>2]=A;C=0;D=1;E=1}else{C=0;D=w;E=0}}else{C=u;D=w;E=y}}else{C=u;D=w;E=y}while(0);x=x+1|0;if((x|0)>=(k[(k[q>>2]|0)+928>>2]|0)){F=C;G=D;H=E;break}else{v=v+160|0;u=C;w=D;y=E}}if((B|0)==14){k[(k[q>>2]|0)+932>>2]=k[e>>2];k[f>>2]=237;t=237;r=g;return t|0}if(F){if((s|0)!=237)break;k[f>>2]=237;break}if(G){if((s|0)==237|(H|0)==0)k[f>>2]=237}else B=27}else B=27;while(0);b:do if((B|0)==27){do if(!s){H=rv()|0;k[H>>2]=0;G=jx(c,j,10)|0;switch(i[k[j>>2]>>0]|0){case 32:case 0:{I=0;break}default:I=407}if((k[H>>2]|0)==34){F=l;E=754043;D=F+54|0;do{i[F>>0]=i[E>>0]|0;F=F+1|0;E=E+1|0}while((F|0)<(D|0));zx(l,c,25)|0;uk(5,l);k[H>>2]=0;k[f>>2]=219;break}if(I){k[f>>2]=219;break}E=k[q>>2]|0;if((G|0)>0?(G|0)<=(k[E+928>>2]|0):0){k[e>>2]=G;ux(d,(k[E+960>>2]|0)+((G+-1|0)*160|0)|0)|0;break b}else{k[f>>2]=219;break}}else{k[f>>2]=219;if((s|0)==237)break b}while(0);k[h>>2]=c;Xw(m,754275,h)|0;uk(5,m)}while(0);k[(k[q>>2]|0)+932>>2]=k[e>>2];t=k[f>>2]|0;r=g;return t|0}function Vk(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;c=r;r=r+272|0;d=c+16|0;e=c+28|0;f=c+24|0;g=c+8|0;h=c;i=c+176|0;j=c+103|0;l=c+32|0;k[f>>2]=0;m=k[b>>2]|0;if((m|0)>0){n=m;r=c;return n|0}m=k[a>>2]|0;o=a+4|0;q=k[o>>2]|0;if((m|0)==(k[q+68>>2]|0)){if((k[q+76>>2]|0)==1){s=q+120|0;if(!((k[s>>2]|0)==-1?(k[s+4>>2]|0)==-1:0)){if(k[q+72>>2]|0){Do(a,2,b)|0;s=(No(a,790243,g,j,f)|0)>0;f=k[o>>2]|0;if(s){s=f+944|0;t=k[s>>2]|0;u=k[s+4>>2]|0;s=g;k[s>>2]=t;k[s+4>>2]=u;v=u;w=t}else{t=g;v=k[t+4>>2]|0;w=k[t>>2]|0}t=f+944|0;g=k[t>>2]|0;u=k[t+4>>2]|0;if((u|0)>(v|0)|(u|0)==(v|0)&g>>>0>w>>>0?(t=f+936|0,(k[t>>2]|0)==(w|0)?(k[t+4>>2]|0)==(v|0):0):0){p[d>>3]=+(g>>>0)+4294967296.0*+(u|0);Xw(l,754311,d)|0;Dk(790243,l,j,i,b)|0;Ek(a,i,b)|0;x=k[o>>2]|0}else x=f}else x=q;q=x+976|0;x=k[q+4>>2]|0;if((x|0)>0|(x|0)==0&(k[q>>2]|0)>>>0>0?(Do(a,2,b)|0,No(a,782652,h,j,b)|0,q=(k[o>>2]|0)+976|0,o=k[q>>2]|0,x=k[q+4>>2]|0,q=h,!((o|0)==(k[q>>2]|0)?(x|0)==(k[q+4>>2]|0):0)):0)oq(a,782652,o,x,j,b)|0}if((ol(a,b)|0)<1)dl(a,e,b)|0}}else Fk(a,m+1|0,0,b)|0;n=k[b>>2]|0;r=c;return n|0}function Wk(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0;f=r;r=r+144|0;g=f+71|0;h=f;k[d>>2]=0;k[e>>2]=1;Bx(g,a,71)|0;Bx(h,b,71)|0;i[g+70>>0]=0;i[h+70>>0]=0;b=(xx(g)|0)+-1|0;a:do if((b|0)>-1){a=b;do{j=g+a|0;if((i[j>>0]|0)!=32)break a;i[j>>0]=0;a=a+-1|0}while((a|0)>-1)}while(0);b=(xx(h)|0)+-1|0;b:do if((b|0)>-1){a=b;do{j=h+a|0;if((i[j>>0]|0)!=32)break b;i[j>>0]=0;a=a+-1|0}while((a|0)>-1)}while(0);if(!c){c=xx(g)|0;if(c){b=0;do{a=g+b|0;j=(nv(i[a>>0]|0)|0)&255;i[a>>0]=j;b=b+1|0}while((b|0)!=(c|0))}c=xx(h)|0;if(c){b=0;do{j=h+b|0;a=(nv(i[j>>0]|0)|0)&255;i[j>>0]=a;b=b+1|0}while((b|0)!=(c|0))}}if((i[g>>0]|0)==(i[h>>0]|0)?(tx(g,h)|0)==0:0){k[d>>2]=1;r=f;return}k[e>>2]=0;e=0;c=0;b=0;a=0;j=1;c:while(1){l=c+1|0;m=g+a|0;n=h+l|0;o=e;p=b;while(1){q=i[g+p>>0]|0;s=i[h+o>>0]|0;t=s<<24>>24==0;if(q<<24>>24)if(t){u=q;v=p;w=24;break c}else{x=q;y=s;z=o;A=p}else{if(t){w=20;break c}if(j){w=39;break c}x=i[m>>0]|0;y=i[n>>0]|0;z=l;A=a}if(x<<24>>24==63?1:x<<24>>24==y<<24>>24){o=z+1|0;p=A+1|0;continue}if(!(x<<24>>24==35&((y<<24>>24)+-48|0)>>>0<10)){if(x<<24>>24==42){B=y;C=z;D=A;break}if(j){w=39;break c}else{o=l;p=a;continue}}else{t=z;while(1){s=t+1|0;if(((i[h+s>>0]|0)+-48|0)>>>0<10)t=s;else{E=s;break}}o=E;p=A+1|0;continue}}p=i[g+(D+1)>>0]|0;switch(p<<24>>24){case 32:case 0:{w=34;break c;break}default:{}}if(!(B<<24>>24)){w=39;break}else{F=B;G=C}while(1){if(p<<24>>24==F<<24>>24){H=G;break}G=G+1|0;F=i[h+G>>0]|0;if(!(F<<24>>24)){w=39;break c}}e=H+1|0;c=C;b=D+2|0;a=D;j=0}if((w|0)==20){k[d>>2]=1;r=f;return}else if((w|0)==24){if(u<<24>>24!=42){r=f;return}if(i[g+(v+1)>>0]|0){r=f;return}k[d>>2]=1;r=f;return}else if((w|0)==34){k[d>>2]=1;r=f;return}else if((w|0)==39){r=f;return}}function Xk(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0;g=r;r=r+16|0;h=g+8|0;i=g;Yk(a,b,c,h,i,f)|0;c=k[f>>2]|0;if((c|0)>0){r=g;return c|0}if(d)k[d>>2]=k[h>>2];if(!e){r=g;return c|0}k[e>>2]=k[i>>2];r=g;return c|0}function Yk(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0;g=r;r=r+16|0;h=g+4|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}j=k[a>>2]|0;m=a+4|0;n=k[m>>2]|0;if((j|0)==(k[n+68>>2]|0)){o=n+120|0;if(((k[o>>2]|0)==-1?(k[o+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){l=k[f>>2]|0;r=g;return l|0}}else Fk(a,j+1|0,0,f)|0;if((b|0)>=1?(j=k[m>>2]|0,(k[j+928>>2]|0)>=(b|0)):0){o=k[j+960>>2]|0;n=b+-1|0;b=k[f>>2]|0;if((b|0)>0){l=b;r=g;return l|0}b=k[a>>2]|0;if((b|0)==0?(p=j+96|0,(k[p>>2]|0)==0&(k[p+4>>2]|0)==0):0)q=0;else s=13;do if((s|0)==13){if((b|0)==(k[j+68>>2]|0)){p=j+120|0;if(((k[p>>2]|0)==-1?(k[p+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){q=0;break}}else Fk(a,b+1|0,0,f)|0;p=k[m>>2]|0;q=(k[p+1080>>2]|0)==0?k[p+72>>2]|0:0}while(0);m=k[f>>2]|0;if((m|0)>0){l=m;r=g;return l|0}if((q|0)==1){Nk(o+(n*160|0)+140|0,c,i,h,f)|0;h=k[i>>2]|0;i=e;k[i>>2]=h;k[i+4>>2]=((h|0)<0)<<31>>31;if(d){h=d;k[h>>2]=1;k[h+4>>2]=0}}else{if(c)k[c>>2]=k[o+(n*160|0)+80>>2];if(e){c=k[o+(n*160|0)+152>>2]|0;h=e;k[h>>2]=c;k[h+4>>2]=((c|0)<0)<<31>>31}if(d){c=o+(n*160|0)+88|0;n=k[c+4>>2]|0;o=d;k[o>>2]=k[c>>2];k[o+4>>2]=n}}l=k[f>>2]|0;r=g;return l|0}k[f>>2]=302;l=302;r=g;return l|0}function Zk(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==0?(h=g+96|0,(k[h>>2]|0)==0&(k[h+4>>2]|0)==0):0)i=0;else{if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if(((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){e=k[c>>2]|0;return e|0}}else Fk(a,d+1|0,0,c)|0;d=k[f>>2]|0;f=k[d+72>>2]|0;k[b>>2]=f;i=(k[d+1080>>2]|0)==0?f:0}k[b>>2]=i;e=k[c>>2]|0;return e|0}function _k(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0;g=r;r=r+16|0;h=g+4|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}j=k[a>>2]|0;m=a+4|0;n=k[m>>2]|0;if((j|0)==(k[n+68>>2]|0)){o=n+120|0;if(((k[o>>2]|0)==-1?(k[o+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){l=k[f>>2]|0;r=g;return l|0}}else Fk(a,j+1|0,0,f)|0;if((b|0)>=1?(j=k[m>>2]|0,(k[j+928>>2]|0)>=(b|0)):0){o=k[j+960>>2]|0;n=b+-1|0;b=k[f>>2]|0;if((b|0)>0){l=b;r=g;return l|0}b=k[a>>2]|0;if((b|0)==0?(q=j+96|0,(k[q>>2]|0)==0&(k[q+4>>2]|0)==0):0)s=0;else t=13;do if((t|0)==13){if((b|0)==(k[j+68>>2]|0)){q=j+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0){s=0;break}}else Fk(a,b+1|0,0,f)|0;q=k[m>>2]|0;s=(k[q+1080>>2]|0)==0?k[q+72>>2]|0:0}while(0);m=k[f>>2]|0;if((m|0)>0){l=m;r=g;return l|0}if((s|0)==1){Nk(o+(n*160|0)+140|0,c,i,h,f)|0;if(e){h=k[i>>2]|0;i=e;k[i>>2]=h;k[i+4>>2]=((h|0)<0)<<31>>31}if(d){h=d;k[h>>2]=1;k[h+4>>2]=0}}else{if(c)k[c>>2]=k[o+(n*160|0)+80>>2];if(e){h=k[o+(n*160|0)+152>>2]|0;i=e;k[i>>2]=h;k[i+4>>2]=((h|0)<0)<<31>>31}if(d){h=o+(n*160|0)+88|0;i=k[h+4>>2]|0;e=d;k[e>>2]=k[h>>2];k[e+4>>2]=i}}if(!c){l=k[f>>2]|0;r=g;return l|0}u=+p[o+(n*160|0)+96>>3];v=+p[o+(n*160|0)+104>>3];if(u==1.0&v==0.0){l=k[f>>2]|0;r=g;return l|0}n=k[c>>2]|0;o=(n|0)>-1?n:0-n|0;switch(o|0){case 11:{w=255.0;x=0.0;break}case 21:{w=32767.0;x=-32768.0;break}case 41:{w=2147483647.0;x=-2147483648.0;break}case 81:{w=2147483647.0;x=-2147483648.0;break}default:{l=k[f>>2]|0;r=g;return l|0}}i=!(u>=0.0);y=v+u*x;x=i?v+u*y:y;z=i?y:v+u*w;a:do if(v!=2147483648.0?+(~~u|0)!=u|(v<2147483648.0?+(~~v|0):0.0)!=v:0){switch(o|0){case 11:case 21:{A=42;break a;break}default:{}}A=82}else t=43;while(0);if((t|0)==43)if(!(z==127.0&x==-128.0))if(!(z<=32767.0&x>=-32768.0)){t=x>=0.0;if(!(z<=65535.0&t))if(z<=2147483647.0&x>=-2147483648.0)A=41;else A=z<4294967296.0&t?40:82;else A=20}else A=21;else A=12;k[c>>2]=(n|0)<0?0-A|0:A;l=k[f>>2]|0;r=g;return l|0}k[f>>2]=302;l=302;r=g;return l|0}function $k(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if(((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){e=k[c>>2]|0;return e|0}}else Fk(a,d+1|0,0,c)|0;d=k[f>>2]|0;if(!(k[d+72>>2]|0)){k[c>>2]=235;e=235;return e|0}else{k[b>>2]=k[d+928>>2];e=k[c>>2]|0;return e|0}return 0}function al(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if(((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){e=k[c>>2]|0;return e|0}}else Fk(a,d+1|0,0,c)|0;d=k[f>>2]|0;if(!(k[d+72>>2]|0)){k[c>>2]=235;e=235;return e|0}else{k[b>>2]=k[d+944>>2];e=k[c>>2]|0;return e|0}return 0}function bl(a,b){a=a|0;b=b|0;var c=0;c=(k[a>>2]|0)+1|0;k[b>>2]=c;return c|0}function cl(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}f=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((f|0)==(k[i+68>>2]|0)){j=i+120|0;if(((k[j>>2]|0)==-1?(k[j+4>>2]|0)==-1:0)?(Vk(a,e)|0)>0:0){g=k[e>>2]|0;return g|0}}else if((Fk(a,f+1|0,0,e)|0)>0){g=k[e>>2]|0;return g|0}if(b){f=k[h>>2]|0;a=(k[f+88>>2]|0)+(k[f+68>>2]<<3)|0;f=k[a+4>>2]|0;j=b;k[j>>2]=k[a>>2];k[j+4>>2]=f}if(c){f=(k[h>>2]|0)+120|0;j=k[f+4>>2]|0;a=c;k[a>>2]=k[f>>2];k[a+4>>2]=j}if(d){j=k[h>>2]|0;h=(k[j+88>>2]|0)+((k[j+68>>2]|0)+1<<3)|0;j=k[h+4>>2]|0;a=d;k[a>>2]=k[h>>2];k[a+4>>2]=j}g=k[e>>2]|0;return g|0}function dl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;d=r;r=r+400|0;e=d;f=d+314|0;g=d+239|0;h=d+168|0;j=d+95|0;l=d+24|0;m=d+4|0;n=k[c>>2]|0;if((n|0)>0){o=n;r=d;return o|0}if((Ro(a,1,f,c)|0)>0){uk(5,754316);o=k[c>>2]|0;r=d;return o|0}Bx(g,f,8)|0;i[g+8>>0]=0;n=g+7|0;if((((((((i[n>>0]|0)==32?(i[n>>0]=0,n=g+6|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,n=g+5|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,n=g+4|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,n=g+3|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,n=g+2|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,n=g+1|0,(i[n>>0]|0)==32):0)?(i[n>>0]=0,(i[g>>0]|0)==32):0)i[g>>0]=0;if((Hk(f,h,j,c)|0)>0){uk(5,754362);uk(5,f);o=k[c>>2]|0;r=d;return o|0}a:do if(!(tx(g,782785)|0)){el(a,c)|0;if(b)k[b>>2]=0}else{if(tx(g,782702)|0)switch(i[f>>0]|0){case 10:case 0:{k[c>>2]=107;break a;break}default:{k[c>>2]=252;uk(5,754435);uk(5,f);break a}}if((fl(h,l,c)|0)>0){uk(5,761806);uk(5,h);o=k[c>>2]|0;r=d;return o|0}else p=l;while(1)if((i[p>>0]|0)==32)p=p+1|0;else{q=p;break}if(!(tx(q,773513)|0)){gl(a,c)|0;if(!b)break;k[b>>2]=1;break}if(((tx(q,772762)|0)!=0?(tx(q,754418)|0)!=0:0)?(tx(q,754427)|0)!=0:0){k[e>>2]=0;el(a,e)|0;j=k[e>>2]|0;n=(b|0)!=0;if(n&(j|0)==251){k[b>>2]=-1;break}k[c>>2]=j;if(!n)break;k[b>>2]=0;break}hl(a,c)|0;if(b)k[b>>2]=2}while(0);b=a+4|0;e=k[b>>2]|0;q=(k[e+88>>2]|0)+((k[e+68>>2]|0)+1<<3)|0;p=k[q+4>>2]|0;l=e+32|0;h=k[l+4>>2]|0;f=e+40|0;do if((p|0)<(h|0)|((p|0)==(h|0)?(k[q>>2]|0)>>>0<(k[l>>2]|0)>>>0:0))k[f>>2]=0;else{k[f>>2]=1;bi(a,m,c)|0;if((tx(m,777975)|0)!=0?(tx(m,765414)|0)!=0:0)break;e=k[b>>2]|0;g=(k[e+88>>2]|0)+((k[e+68>>2]|0)+1<<3)|0;di(a,k[g>>2]|0,k[g+4>>2]|0,c)|0}while(0);o=k[c>>2]|0;r=d;return o|0}function el(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0;c=r;r=r+8128|0;d=c+8048|0;e=c+8044|0;f=c+8036|0;g=c+8032|0;h=c+8024|0;i=c+8020|0;j=c+8016|0;l=c+24|0;m=c+16|0;n=c+8|0;o=c;q=c+8052|0;s=k[b>>2]|0;if((s|0)>0){t=s;r=c;return t|0}s=k[a>>2]|0;u=a+4|0;v=k[u>>2]|0;if((s|0)==(k[v+68>>2]|0))w=v;else{Fk(a,s+1|0,0,b)|0;w=k[u>>2]|0}k[w+72>>2]=0;s=w+32|0;v=k[s+4>>2]|0;x=w+96|0;k[x>>2]=k[s>>2];k[x+4>>2]=v;k[d>>2]=0;k[e>>2]=k[b>>2];_o(a,999,c+8040|0,f,g,l,i,j,c+8028|0,n,o,m,h,b)|0;v=k[b>>2]|0;if((v|0)!=233){if((v|0)>0){t=v;r=c;return t|0}}else k[b>>2]=k[e>>2];v=k[u>>2]|0;x=v+112|0;s=k[x>>2]|0;w=k[x+4>>2]|0;x=((k[h>>2]|0)*80|0)+80|0;h=jy(s|0,w|0,x|0,((x|0)<0)<<31>>31|0)|0;x=v+96|0;k[x>>2]=h;k[x+4>>2]=L;x=ky(s|0,w|0,-80,-1)|0;w=L;s=xy(x|0,w|0,2880,0)|0;h=jy(2880,0,s|0,L|0)|0;s=ky(h|0,L|0,x|0,w|0)|0;w=v+120|0;k[w>>2]=s;k[w+4>>2]=L;w=l;if((k[g>>2]|0)>0&((k[w>>2]|0)==0&(k[w+4>>2]|0)==0)?(k[e>>2]=0,Do(a,2,b)|0,(Lo(a,772311,d,q,e)|0)!=0):0)k[d>>2]=0;e=k[f>>2]|0;switch(e|0){case 8:{y=1;z=11;break}case 16:{y=2;z=21;break}case 32:{y=4;z=41;break}case 64:{y=8;z=81;break}case -32:{y=4;z=42;break}default:{f=(e|0)==-64;y=f?8:0;z=f?82:0}}f=k[u>>2]|0;k[f+128>>2]=k[g>>2];e=k[g>>2]|0;if(e){q=(k[d>>2]|0)==0;d=l;a=k[d>>2]|0;w=k[d+4>>2]|0;d=q?a:1;s=q?w:0;q=f+136|0;k[q>>2]=a;k[q+4>>2]=w;if((e|0)>1){w=d;q=s;a=1;while(1){v=l+(a<<3)|0;x=k[v>>2]|0;h=k[v+4>>2]|0;v=yy(x|0,h|0,w|0,q|0)|0;A=L;B=f+136+(a<<3)|0;k[B>>2]=x;k[B+4>>2]=h;a=a+1|0;if((a|0)>=(e|0)){C=v;D=A;break}else{w=v;q=A}}}else{C=d;D=s}}else{C=0;D=0}s=f+120|0;d=k[s>>2]|0;q=k[s+4>>2]|0;s=k[i>>2]|0;w=ky(s|0,((s|0)<0)<<31>>31|0,C|0,D|0)|0;s=yy(w|0,L|0,y|0,0)|0;w=L;e=k[j>>2]|0;j=((e|0)<0)<<31>>31;a=yy(s|0,w|0,e|0,j|0)|0;l=L;A=ky(a|0,l|0,2879,0)|0;v=L;h=xy(A|0,v|0,2880,0)|0;B=jy(d|0,q|0,h|0,L|0)|0;h=ky(B|0,L|0,A|0,v|0)|0;v=(k[f+88>>2]|0)+((k[f+68>>2]|0)+1<<3)|0;k[v>>2]=h;k[v+4>>2]=L;v=f+968|0;k[v>>2]=a;k[v+4>>2]=l;l=f+976|0;k[l>>2]=0;k[l+4>>2]=0;k[f+1080>>2]=0;do if(!(k[g>>2]|0)){l=f+952|0;k[l>>2]=0;k[l+4>>2]=0;k[f+928>>2]=0;if(!(k[f+1220>>2]|0))E=f;else{l=((k[f+1104>>2]|0)+-1|0)/(k[f+1044>>2]|0)|0;if((l|0)>-1){v=f;a=0;while(1){h=k[(k[v+1232>>2]|0)+(a<<2)>>2]|0;if(!h)F=v;else{dy(h);F=k[u>>2]|0}h=k[(k[F+1236>>2]|0)+(a<<2)>>2]|0;if(!h)G=F;else{dy(h);G=k[u>>2]|0}if((a|0)==(l|0)){H=G;break}else{v=G;a=a+1|0}}}else H=f;dy(k[H+1240>>2]|0);dy(k[(k[u>>2]|0)+1228>>2]|0);dy(k[(k[u>>2]|0)+1224>>2]|0);dy(k[(k[u>>2]|0)+1236>>2]|0);dy(k[(k[u>>2]|0)+1232>>2]|0);dy(k[(k[u>>2]|0)+1220>>2]|0);a=k[u>>2]|0;v=a+1220|0;k[v>>2]=0;k[v+4>>2]=0;k[v+8>>2]=0;k[v+12>>2]=0;k[v+16>>2]=0;k[v+20>>2]=0;E=a}a=k[E+960>>2]|0;if(!a)I=E;else{dy(a);I=k[u>>2]|0}k[I+960>>2]=0;a=I+936|0;k[a>>2]=0;k[a+4>>2]=0;k[a+8>>2]=0;k[a+12>>2]=0;J=I}else{a=f+944|0;k[a>>2]=e;k[a+4>>2]=j;a=f+936|0;k[a>>2]=e;k[a+4>>2]=j;a=f+952|0;k[a>>2]=s;k[a+4>>2]=w;k[f+928>>2]=2;if(!(k[f+1220>>2]|0))K=f;else{a=((k[f+1104>>2]|0)+-1|0)/(k[f+1044>>2]|0)|0;if((a|0)>-1){v=f;l=0;while(1){h=k[(k[v+1232>>2]|0)+(l<<2)>>2]|0;if(!h)M=v;else{dy(h);M=k[u>>2]|0}h=k[(k[M+1236>>2]|0)+(l<<2)>>2]|0;if(!h)N=M;else{dy(h);N=k[u>>2]|0}if((l|0)==(a|0)){O=N;break}else{v=N;l=l+1|0}}}else O=f;dy(k[O+1240>>2]|0);dy(k[(k[u>>2]|0)+1228>>2]|0);dy(k[(k[u>>2]|0)+1224>>2]|0);dy(k[(k[u>>2]|0)+1236>>2]|0);dy(k[(k[u>>2]|0)+1232>>2]|0);dy(k[(k[u>>2]|0)+1220>>2]|0);l=k[u>>2]|0;v=l+1220|0;k[v>>2]=0;k[v+4>>2]=0;k[v+8>>2]=0;k[v+12>>2]=0;k[v+16>>2]=0;k[v+20>>2]=0;K=l}l=k[K+960>>2]|0;if(l)dy(l);l=ey(2,160)|0;if(l){v=k[u>>2]|0;k[v+960>>2]=l;a=l+72|0;k[a>>2]=0;k[a+4>>2]=0;k[l+80>>2]=z;k[l+152>>2]=y;a=k[i>>2]|0;h=l+88|0;k[h>>2]=a;k[h+4>>2]=((a|0)<0)<<31>>31;p[l+96>>3]=1.0;p[l+104>>3]=0.0;h=m;A=k[h>>2]|0;B=k[h+4>>2]|0;h=l+112|0;k[h>>2]=A;k[h+4>>2]=B;h=ha(a,y)|0;a=l+232|0;k[a>>2]=h;k[a+4>>2]=((h|0)<0)<<31>>31;k[l+240>>2]=z;k[l+312>>2]=y;h=l+248|0;k[h>>2]=C;k[h+4>>2]=D;p[l+256>>3]=+p[n>>3];p[l+264>>3]=+p[o>>3];h=l+272|0;k[h>>2]=A;k[h+4>>2]=B;J=v;break}uk(5,754501);k[(k[u>>2]|0)+960>>2]=0;k[b>>2]=111;t=111;r=c;return t|0}while(0);u=(k[J+88>>2]|0)+(k[J+68>>2]<<3)|0;o=k[u+4>>2]|0;n=J+112|0;k[n>>2]=k[u>>2];k[n+4>>2]=o;t=k[b>>2]|0;r=c;return t|0}function fl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}switch(i[a>>0]|0){case 39:{d=xx(a)|0;a:do if(d>>>0>1){f=1;g=0;while(1){h=i[a+f>>0]|0;if(h<<24>>24==39){j=f+1|0;if((i[a+j>>0]|0)==39){l=39;m=j}else{n=f;o=g;break a}}else{l=h;m=f}i[b+g>>0]=l;h=m+1|0;j=g+1|0;if(h>>>0<d>>>0){f=h;g=j}else{n=h;o=j;break}}}else{n=1;o=0}while(0);i[b+o>>0]=0;if((n|0)==(d|0)){uk(5,754566);uk(5,a);k[c>>2]=205;e=205;return e|0}b:do if((o|0)>0){d=o;do{n=d;d=d+-1|0;m=b+d|0;if((i[m>>0]|0)!=32)break b;i[m>>0]=0}while((n|0)>1)}while(0);e=k[c>>2]|0;return e|0}case 0:{i[b>>0]=0;k[c>>2]=204;e=204;return e|0}default:{ux(b,a)|0;e=k[c>>2]|0;return e|0}}return 0}function gl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0;c=r;r=r+512|0;d=c+88|0;e=c+80|0;f=c+72|0;g=c+64|0;h=c+56|0;j=c+48|0;l=c+40|0;m=c+32|0;n=c+24|0;o=c+488|0;q=c+96|0;s=c+16|0;t=c+8|0;u=c;v=c+406|0;w=c+335|0;x=c+262|0;y=c+181|0;z=c+100|0;A=k[b>>2]|0;if((A|0)>0){B=A;r=c;return B|0}A=k[a>>2]|0;C=a+4|0;D=k[C>>2]|0;if((A|0)==(k[D+68>>2]|0))E=D;else{Fk(a,A+1|0,0,b)|0;E=k[C>>2]|0}k[E+72>>2]=1;A=E+32|0;D=k[A+4>>2]|0;F=E+96|0;k[F>>2]=k[A>>2];k[F+4>>2]=D;if(($o(a,t,u,s,q,b)|0)>0){B=k[b>>2]|0;r=c;return B|0}D=s;if(!((k[D>>2]|0)==0&(k[D+4>>2]|0)==0)){uk(5,754614);k[n>>2]=k[s>>2];Xw(z,754670,n)|0;uk(5,z);k[b>>2]=214;B=214;r=c;return B|0}z=t;n=k[z+4>>2]|0;s=k[C>>2]|0;D=s+952|0;k[D>>2]=k[z>>2];k[D+4>>2]=n;k[s+928>>2]=k[q>>2];if(!(k[s+1220>>2]|0))G=s;else{n=((k[s+1104>>2]|0)+-1|0)/(k[s+1044>>2]|0)|0;if((n|0)>-1){D=s;z=0;while(1){F=k[(k[D+1232>>2]|0)+(z<<2)>>2]|0;if(!F)H=D;else{dy(F);H=k[C>>2]|0}F=k[(k[H+1236>>2]|0)+(z<<2)>>2]|0;if(!F)I=H;else{dy(F);I=k[C>>2]|0}if((z|0)==(n|0)){J=I;break}else{D=I;z=z+1|0}}}else J=s;dy(k[J+1240>>2]|0);dy(k[(k[C>>2]|0)+1228>>2]|0);dy(k[(k[C>>2]|0)+1224>>2]|0);dy(k[(k[C>>2]|0)+1236>>2]|0);dy(k[(k[C>>2]|0)+1232>>2]|0);dy(k[(k[C>>2]|0)+1220>>2]|0);J=k[C>>2]|0;s=J+1220|0;k[s>>2]=0;k[s+4>>2]=0;k[s+8>>2]=0;k[s+12>>2]=0;k[s+16>>2]=0;k[s+20>>2]=0;G=J}J=k[G+960>>2]|0;if(J)dy(J);J=k[q>>2]|0;if((J|0)>0){G=ey(J,160)|0;if(!G){uk(5,754685);k[(k[C>>2]|0)+960>>2]=0;k[b>>2]=111;B=111;r=c;return B|0}k[(k[C>>2]|0)+960>>2]=G;J=G;G=0;while(1){i[J>>0]=0;p[J+96>>3]=1.0;p[J+104>>3]=0.0;i[J+120>>0]=1;s=J+72|0;k[s>>2]=-1;k[s+4>>2]=-1;k[J+80>>2]=-9999;G=G+1|0;if((G|0)>=(k[q>>2]|0))break;else J=J+160|0}K=k[C>>2]|0}else{J=k[C>>2]|0;k[J+960>>2]=0;K=J}J=u;G=k[J+4>>2]|0;s=K+944|0;k[s>>2]=k[J>>2];k[s+4>>2]=G;G=u;s=k[G+4>>2]|0;J=K+936|0;k[J>>2]=k[G>>2];k[J+4>>2]=s;s=t;J=u;G=yy(k[J>>2]|0,k[J+4>>2]|0,k[s>>2]|0,k[s+4>>2]|0)|0;s=K+968|0;k[s>>2]=G;k[s+4>>2]=L;s=K+976|0;k[s>>2]=0;k[s+4>>2]=0;k[K+1080>>2]=0;K=8;s=0;a:while(1){Uo(a,K,v,w,x,b)|0;G=k[b>>2]|0;switch(G|0){case 107:{M=29;break a;break}case 205:{J=w+(xx(w)|0)|0;i[J>>0]=39;i[J+1>>0]=0;k[b>>2]=0;N=0;break}case 207:{k[b>>2]=0;N=0;break}default:if((G|0)>0){B=G;M=57;break a}else N=G}G=i[v>>0]|0;switch(G<<24>>24){case 84:{il(a,v,w,b)|0;O=i[v>>0]|0;break}case 69:{if(!(tx(v,768779)|0)){P=N;Q=s;M=35;break a}else O=69;break}default:O=G}K=K+1|0;s=(i[w>>0]|O|i[x>>0])<<24>>24==0?s+1|0:0}if((M|0)==29){uk(5,754750);k[b>>2]=210;B=210;r=c;return B|0}else if((M|0)==35){s=k[q>>2]|0;q=k[C>>2]|0;b:do if((s|0)>0){x=q+952|0;O=k[q+960>>2]|0;w=0;while(1){K=O+72|0;N=k[K>>2]|0;a=k[K+4>>2]|0;if((k[O+80>>2]|0)==-9999){R=w;M=38;break}if((N|0)==-1&(a|0)==-1){S=w;M=43;break}K=x;G=k[K>>2]|0;J=k[K+4>>2]|0;if(!((G|0)==0&(J|0)==0)){if(!(((a|0)>-1|(a|0)==-1&N>>>0>4294967295)&((a|0)<(J|0)|(a|0)==(J|0)&N>>>0<G>>>0))){T=N;U=w;M=49;break}K=k[O+152>>2]|0;z=ky(K|0,((K|0)<0)<<31>>31|0,N|0,a|0)|0;a=L;if((a|0)>(J|0)|(a|0)==(J|0)&z>>>0>G>>>0){V=O;W=w;M=54;break}}w=w+1|0;if((w|0)>=(s|0))break b;else O=O+160|0}if((M|0)==38){i[v>>0]=0;if((R|0)<-1)k[b>>2]=206;else{k[m>>2]=R+1;Xw(o,772183,m)|0;i[v>>0]=i[773003]|0;i[v+1>>0]=i[773004]|0;i[v+2>>0]=i[773005]|0;i[v+3>>0]=i[773006]|0;i[v+4>>0]=i[773007]|0;i[v+5>>0]=i[773008]|0;qx(v,o)|0}k[l>>2]=v;Xw(y,754805,l)|0;uk(5,y);k[b>>2]=232;B=232;r=c;return B|0}else if((M|0)==43){i[v>>0]=0;if((S|0)<-1)k[b>>2]=206;else{k[j>>2]=S+1;Xw(o,772183,j)|0;i[v>>0]=i[773688]|0;i[v+1>>0]=i[773689]|0;i[v+2>>0]=i[773690]|0;i[v+3>>0]=i[773691]|0;i[v+4>>0]=i[773692]|0;i[v+5>>0]=i[773693]|0;qx(v,o)|0}k[h>>2]=v;Xw(y,754805,h)|0;uk(5,y);k[b>>2]=231;B=231;r=c;return B|0}else if((M|0)==49){i[v>>0]=0;if((U|0)<-1)k[b>>2]=206;else{k[g>>2]=U+1;Xw(o,772183,g)|0;i[v>>0]=i[773688]|0;i[v+1>>0]=i[773689]|0;i[v+2>>0]=i[773690]|0;i[v+3>>0]=i[773691]|0;i[v+4>>0]=i[773692]|0;i[v+5>>0]=i[773693]|0;qx(v,o)|0}k[f>>2]=v;k[f+4>>2]=T;Xw(y,754846,f)|0;uk(5,y);k[b>>2]=234;B=234;r=c;return B|0}else if((M|0)==54){k[e>>2]=W+1;Xw(y,754895,e)|0;uk(5,y);O=k[(k[C>>2]|0)+952>>2]|0;k[d>>2]=V+140;k[d+4>>2]=O;Xw(y,754943,d)|0;uk(5,y);k[b>>2]=236;B=236;r=c;return B|0}}while(0);b=q+112|0;y=b;d=k[y>>2]|0;V=k[y+4>>2]|0;y=(Q*80|0)+80|0;Q=jy(d|0,V|0,y|0,((y|0)<0)<<31>>31|0)|0;y=q+96|0;k[y>>2]=Q;k[y+4>>2]=L;y=ky(d|0,V|0,-80,-1)|0;V=L;d=xy(y|0,V|0,2880,0)|0;Q=jy(2880,0,d|0,L|0)|0;d=ky(Q|0,L|0,y|0,V|0)|0;V=L;y=q+120|0;k[y>>2]=d;k[y+4>>2]=V;y=t;t=u;u=yy(k[t>>2]|0,k[t+4>>2]|0,k[y>>2]|0,k[y+4>>2]|0)|0;y=ky(u|0,L|0,2879,0)|0;u=L;t=xy(y|0,u|0,2880,0)|0;Q=jy(d|0,V|0,t|0,L|0)|0;t=ky(Q|0,L|0,y|0,u|0)|0;u=k[q+68>>2]|0;y=k[q+88>>2]|0;q=y+(u+1<<3)|0;k[q>>2]=t;k[q+4>>2]=L;q=y+(u<<3)|0;u=k[q+4>>2]|0;y=b;k[y>>2]=k[q>>2];k[y+4>>2]=u;B=P;r=c;return B|0}else if((M|0)==57){r=c;return B|0}return 0}function hl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0;c=r;r=r+384|0;d=c+48|0;e=c+40|0;f=c+32|0;g=c+360|0;h=c+56|0;j=c+24|0;l=c+16|0;m=c+8|0;n=c;o=c+285|0;q=c+214|0;s=c+141|0;t=c+60|0;u=k[b>>2]|0;if((u|0)>0){v=u;r=c;return v|0}u=k[a>>2]|0;w=a+4|0;x=k[w>>2]|0;if((u|0)==(k[x+68>>2]|0))y=x;else{Fk(a,u+1|0,0,b)|0;y=k[w>>2]|0}k[y+72>>2]=2;u=y+32|0;x=k[u+4>>2]|0;z=y+96|0;k[z>>2]=k[u>>2];k[z+4>>2]=x;if(($o(a,l,m,j,h,b)|0)>0){v=k[b>>2]|0;r=c;return v|0}x=l;z=k[x+4>>2]|0;u=k[w>>2]|0;y=u+952|0;k[y>>2]=k[x>>2];k[y+4>>2]=z;k[u+928>>2]=k[h>>2];if(!(k[u+1220>>2]|0))A=u;else{z=((k[u+1104>>2]|0)+-1|0)/(k[u+1044>>2]|0)|0;if((z|0)>-1){y=u;x=0;while(1){B=k[(k[y+1232>>2]|0)+(x<<2)>>2]|0;if(!B)C=y;else{dy(B);C=k[w>>2]|0}B=k[(k[C+1236>>2]|0)+(x<<2)>>2]|0;if(!B)D=C;else{dy(B);D=k[w>>2]|0}if((x|0)==(z|0)){E=D;break}else{y=D;x=x+1|0}}}else E=u;dy(k[E+1240>>2]|0);dy(k[(k[w>>2]|0)+1228>>2]|0);dy(k[(k[w>>2]|0)+1224>>2]|0);dy(k[(k[w>>2]|0)+1236>>2]|0);dy(k[(k[w>>2]|0)+1232>>2]|0);dy(k[(k[w>>2]|0)+1220>>2]|0);E=k[w>>2]|0;u=E+1220|0;k[u>>2]=0;k[u+4>>2]=0;k[u+8>>2]=0;k[u+12>>2]=0;k[u+16>>2]=0;k[u+20>>2]=0;A=E}E=k[A+960>>2]|0;if(E)dy(E);E=k[h>>2]|0;if((E|0)>0){A=ey(E,160)|0;if(!A){uk(5,754972);k[(k[w>>2]|0)+960>>2]=0;k[b>>2]=111;v=111;r=c;return v|0}k[(k[w>>2]|0)+960>>2]=A;E=A;A=0;while(1){i[E>>0]=0;p[E+96>>3]=1.0;p[E+104>>3]=0.0;u=E+112|0;k[u>>2]=1234554321;k[u+4>>2]=0;k[E+80>>2]=-9999;u=E+88|0;k[u>>2]=1;k[u+4>>2]=0;i[E+120>>0]=0;A=A+1|0;if((A|0)>=(k[h>>2]|0))break;else E=E+160|0}F=k[w>>2]|0}else{E=k[w>>2]|0;k[E+960>>2]=0;F=E}E=m;A=k[E+4>>2]|0;u=F+944|0;k[u>>2]=k[E>>2];k[u+4>>2]=A;A=m;u=k[A+4>>2]|0;E=F+936|0;k[E>>2]=k[A>>2];k[E+4>>2]=u;u=l;E=m;m=yy(k[E>>2]|0,k[E+4>>2]|0,k[u>>2]|0,k[u+4>>2]|0)|0;u=F+968|0;k[u>>2]=m;k[u+4>>2]=L;u=j;j=k[u+4>>2]|0;m=F+976|0;k[m>>2]=k[u>>2];k[m+4>>2]=j;k[F+1080>>2]=0;F=8;j=0;a:while(1){Uo(a,F,o,q,s,b)|0;m=k[b>>2]|0;switch(m|0){case 107:{G=27;break a;break}case 205:{u=q+(xx(q)|0)|0;i[u>>0]=39;i[u+1>>0]=0;k[b>>2]=0;break}case 207:{k[b>>2]=0;break}default:if((m|0)>0){v=m;G=48;break a}}switch(i[o>>0]|0){case 84:{il(a,o,q,b)|0;break}case 90:{if((tx(o,778793)|0)==0&(i[q>>0]|0)==84)k[(k[w>>2]|0)+1080>>2]=1;break}case 69:{if(!(tx(o,768779)|0)){H=j;G=35;break a}break}default:{}}F=F+1|0;j=(i[q>>0]|i[o>>0]|i[s>>0])<<24>>24==0?j+1|0:0}if((G|0)==27){uk(5,755037);k[b>>2]=210;v=210;r=c;return v|0}else if((G|0)==35){j=k[h>>2]|0;h=k[w>>2]|0;b:do if((j|0)>0){s=k[h+960>>2]|0;q=0;while(1){F=q;q=q+1|0;if((k[s+80>>2]|0)==-9999){I=q;J=F;break}if((q|0)>=(j|0))break b;else s=s+160|0}i[o>>0]=0;if((J|0)<-1)k[b>>2]=206;else{k[f>>2]=I;Xw(g,772183,f)|0;i[o>>0]=i[773003]|0;i[o+1>>0]=i[773004]|0;i[o+2>>0]=i[773005]|0;i[o+3>>0]=i[773006]|0;i[o+4>>0]=i[773007]|0;i[o+5>>0]=i[773008]|0;qx(o,g)|0}k[e>>2]=o;Xw(t,755093,e)|0;uk(5,t);k[b>>2]=232;v=232;r=c;return v|0}while(0);e=h+112|0;o=k[e>>2]|0;g=k[e+4>>2]|0;e=(H*80|0)+80|0;H=jy(o|0,g|0,e|0,((e|0)<0)<<31>>31|0)|0;e=h+96|0;k[e>>2]=H;k[e+4>>2]=L;e=ky(o|0,g|0,-80,-1)|0;g=L;o=xy(e|0,g|0,2880,0)|0;H=jy(2880,0,o|0,L|0)|0;o=ky(H|0,L|0,e|0,g|0)|0;g=L;e=h+120|0;k[e>>2]=o;k[e+4>>2]=g;e=h+968|0;H=h+976|0;f=k[H>>2]|0;I=k[H+4>>2]|0;H=ky(k[e>>2]|0,k[e+4>>2]|0,2879,0)|0;e=ky(H|0,L|0,f|0,I|0)|0;I=L;f=xy(e|0,I|0,2880,0)|0;H=jy(o|0,g|0,f|0,L|0)|0;f=ky(H|0,L|0,e|0,I|0)|0;I=(k[h+88>>2]|0)+((k[h+68>>2]|0)+1<<3)|0;k[I>>2]=f;k[I+4>>2]=L;jl(a,n,b)|0;I=n;n=k[I>>2]|0;f=l;l=k[f>>2]|0;if(!((n|0)==(l|0)?(k[I+4>>2]|0)==(k[f+4>>2]|0):0)){k[d>>2]=l;k[d+4>>2]=n;Xw(t,755134,d)|0;uk(5,t);k[b>>2]=241}t=k[w>>2]|0;w=(k[t+88>>2]|0)+(k[t+68>>2]<<3)|0;d=k[w+4>>2]|0;n=t+112|0;k[n>>2]=k[w>>2];k[n+4>>2]=d;if((k[t+1080>>2]|0)==1)st(a,b)|0;v=k[b>>2]|0;r=c;return v|0}else if((G|0)==48){r=c;return v|0}return 0}function il(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0;e=r;r=r+320|0;f=e+40|0;g=e+32|0;h=e+24|0;j=e+16|0;l=e+8|0;m=e+72|0;n=e+228|0;o=e+68|0;q=e+64|0;s=e+60|0;t=e+56|0;u=e+52|0;v=e;w=e+157|0;x=e+48|0;y=e+76|0;z=k[d>>2]|0;if((z|0)>0){A=z;r=e;return A|0}k[o>>2]=0;z=k[a>>2]|0;B=a+4|0;if((z|0)!=(k[(k[B>>2]|0)+68>>2]|0))Fk(a,z+1|0,0,d)|0;z=b+1|0;a:do switch(i[z>>0]|0){case 84:{if(!(Ax(z,755193,4)|0)){a=b+5|0;C=rv()|0;k[C>>2]=0;D=jx(a,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[C>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,a,25)|0;uk(5,n);k[o>>2]=412;k[C>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((D|0)>=1?(H=k[B>>2]|0,(D|0)<=(k[H+928>>2]|0)):0){I=k[H+960>>2]|0;if((fl(c,w,o)|0)<=0){ux(I+((D+-1|0)*160|0)|0,w)|0;break a}A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 70:{if(!(Ax(z,755198,4)|0)){I=b+5|0;H=rv()|0;k[H>>2]=0;J=jx(I,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[H>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,I,25)|0;uk(5,n);k[o>>2]=412;k[H>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((J|0)>=1?(D=k[B>>2]|0,(J|0)<=(k[D+928>>2]|0)):0){C=k[D+960>>2]|0;D=J+-1|0;if((fl(c,w,o)|0)>0){A=k[d>>2]|0;r=e;return A|0}Bx(C+(D*160|0)+140|0,w,9)|0;i[C+(D*160|0)+149>>0]=0;if((k[(k[B>>2]|0)+72>>2]|0)==1){if((Nk(w,q,t,s,d)|0)<=0){k[C+(D*160|0)+80>>2]=16;a=C+(D*160|0)+88|0;k[a>>2]=1;k[a+4>>2]=0;k[C+(D*160|0)+152>>2]=k[t>>2];break a}A=k[d>>2]|0;r=e;return A|0}if((Qk(w,q,u,t,d)|0)>0){A=k[d>>2]|0;r=e;return A|0}k[C+(D*160|0)+80>>2]=k[q>>2];a=k[u>>2]|0;K=C+(D*160|0)+88|0;k[K>>2]=a;k[K+4>>2]=((a|0)<0)<<31>>31;if((k[q>>2]|0)!=16){k[C+(D*160|0)+152>>2]=k[t>>2];break a}K=C+(D*160|0)+152|0;D=k[K>>2]|0;if(!((D|0)==0|(D|0)>(a|0)))break a;k[K>>2]=k[t>>2];break a}A=k[d>>2]|0;r=e;return A|0}break}case 66:{if(!(Ax(z,755203,4)|0)){K=b+5|0;a=rv()|0;k[a>>2]=0;D=jx(K,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[a>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,K,25)|0;uk(5,n);k[o>>2]=412;k[a>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((D|0)>=1?(J=k[B>>2]|0,(D|0)<=(k[J+928>>2]|0)):0){H=k[J+960>>2]|0;I=D+-1|0;if((k[J+72>>2]|0)==2){A=k[d>>2]|0;r=e;return A|0}if((k[d>>2]|0)<=0){k[a>>2]=0;J=jx(c,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[d>>2]=407}if((k[a>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,c,25)|0;uk(5,n);k[d>>2]=412;k[a>>2]=0}if((k[d>>2]|0)<=0){D=J+-1|0;K=H+(I*160|0)+72|0;k[K>>2]=D;k[K+4>>2]=((D|0)<0)<<31>>31;break a}}k[l>>2]=b;k[l+4>>2]=c;Xw(y,755208,l)|0;uk(5,y);A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 83:{if(!(Ax(z,755252,4)|0)){D=b+5|0;K=rv()|0;k[K>>2]=0;C=jx(D,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[K>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,D,25)|0;uk(5,n);k[o>>2]=412;k[K>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((C|0)>=1?(I=k[B>>2]|0,(C|0)<=(k[I+928>>2]|0)):0){H=k[I+960>>2]|0;if((ll(c,v,o)|0)<=0){p[H+((C+-1|0)*160|0)+96>>3]=+p[v>>3];break a}k[j>>2]=b;k[j+4>>2]=c;Xw(y,755257,j)|0;uk(5,y);A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 90:{if(!(Ax(z,755299,4)|0)){H=b+5|0;I=rv()|0;k[I>>2]=0;J=jx(H,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[I>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,H,25)|0;uk(5,n);k[o>>2]=412;k[I>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((J|0)>=1?(C=k[B>>2]|0,(J|0)<=(k[C+928>>2]|0)):0){K=k[C+960>>2]|0;if((ll(c,v,o)|0)<=0){p[K+((J+-1|0)*160|0)+104>>3]=+p[v>>3];break a}k[h>>2]=b;k[h+4>>2]=c;Xw(y,755257,h)|0;uk(5,y);A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 78:{if(!(Ax(z,760184,4)|0)){K=b+5|0;C=rv()|0;k[C>>2]=0;D=jx(K,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[C>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,K,25)|0;uk(5,n);k[o>>2]=412;k[C>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((D|0)>=1?(J=k[B>>2]|0,(D|0)<=(k[J+928>>2]|0)):0){I=k[J+960>>2]|0;H=D+-1|0;if((k[J+72>>2]|0)==1){if((fl(c,w,o)|0)<=0){Bx(I+(H*160|0)+120|0,w,17)|0;i[I+(H*160|0)+137>>0]=0;break a}A=k[d>>2]|0;r=e;return A|0}k[C>>2]=0;J=ix(c,m,10)|0;a=L;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[C>>2]|0)==34){E=n;F=755304;G=E+58|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,c,25)|0;uk(5,n);k[o>>2]=412;k[C>>2]=0}if((k[o>>2]|0)<=0){D=I+(H*160|0)+112|0;k[D>>2]=J;k[D+4>>2]=a;break a}k[g>>2]=b;k[g+4>>2]=c;Xw(y,755208,g)|0;uk(5,y);A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 68:{if(!(Ax(z,755362,3)|0)){if((k[(k[B>>2]|0)+72>>2]|0)==1){A=k[d>>2]|0;r=e;return A|0}D=b+4|0;K=rv()|0;k[K>>2]=0;M=jx(D,m,10)|0;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[K>>2]|0)==34){E=n;F=754043;G=E+54|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,D,25)|0;uk(5,n);k[o>>2]=412;k[K>>2]=0}if((k[o>>2]|0)>0){A=k[d>>2]|0;r=e;return A|0}if((M|0)>=1?(a=k[B>>2]|0,(M|0)<=(k[a+928>>2]|0)):0){J=k[a+960>>2]|0;a=M+-1|0;H=k[J+(a*160|0)+80>>2]|0;if((H|0)<16)switch(H|0){case -9999:break;default:N=121}else switch(H|0){case 16:break;default:N=121}if((N|0)==121){A=k[d>>2]|0;r=e;return A|0}H=rx(c,40)|0;k[x>>2]=H;if(!H){A=k[d>>2]|0;r=e;return A|0}I=H+1|0;k[x>>2]=I;H=jx(I,x,10)|0;k[t>>2]=H;I=J+(a*160|0)+88|0;C=k[I>>2]|0;O=k[I+4>>2]|0;I=((H|0)<0)<<31>>31;if(!(((C|0)!=1|(O|0)!=0)&((O|0)<(I|0)|(O|0)==(I|0)&C>>>0<H>>>0))){k[J+(a*160|0)+152>>2]=H;break a}A=k[d>>2]|0;r=e;return A|0}A=k[d>>2]|0;r=e;return A|0}break}case 72:{if(!(Ax(z,755366,4)|0)){if((k[(k[B>>2]|0)+72>>2]|0)==1){A=k[d>>2]|0;r=e;return A|0}H=rv()|0;k[H>>2]=0;a=ix(c,m,10)|0;J=L;switch(i[k[m>>2]>>0]|0){case 32:case 0:break;default:k[o>>2]=407}if((k[H>>2]|0)==34){E=n;F=755304;G=E+58|0;do{i[E>>0]=i[F>>0]|0;E=E+1|0;F=F+1|0}while((E|0)<(G|0));zx(n,c,25)|0;uk(5,n);k[o>>2]=412;k[H>>2]=0}if((k[o>>2]|0)>0){k[f>>2]=b;k[f+4>>2]=c;Xw(y,755208,f)|0;uk(5,y);A=k[d>>2]|0;r=e;return A|0}else{M=(k[B>>2]|0)+968|0;k[M>>2]=a;k[M+4>>2]=J;A=k[d>>2]|0;r=e;return A|0}}break}default:{}}while(0);A=k[d>>2]|0;r=e;return A|0}function jl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;d=r;r=r+96|0;e=d;f=d+4|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=k[a>>2]|0;j=a+4|0;l=k[j>>2]|0;if((g|0)==(k[l+68>>2]|0)){m=l+120|0;if(((k[m>>2]|0)==-1?(k[m+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){h=k[c>>2]|0;r=d;return h|0}}else Fk(a,g+1|0,0,c)|0;g=k[j>>2]|0;j=k[g+928>>2]|0;a=k[g+960>>2]|0;g=b;k[g>>2]=0;k[g+4>>2]=0;a:do if((j|0)>0){g=0;m=0;l=a;n=0;b:while(1){o=l+72|0;k[o>>2]=g;k[o+4>>2]=m;o=k[l+80>>2]|0;c:do switch(o|0){case 16:{p=l+88|0;q=k[p>>2]|0;s=k[p+4>>2]|0;break}case 1:{p=l+88|0;q=wy(ky(k[p>>2]|0,k[p+4>>2]|0,7,0)|0,L|0,8,0)|0;s=L;break}default:{if((o|0)>0){p=l+88|0;t=(o|0)/10|0;u=yy(t|0,((t|0)<0)<<31>>31|0,k[p>>2]|0,k[p+4>>2]|0)|0;q=u;s=L;break c}u=l+140|0;p=u;while(1){t=i[p>>0]|0;if(((t<<24>>24)+-48|0)>>>0<10)p=p+1|0;else{v=t;break}}switch(v<<24>>24){case 80:{p=l+88|0;q=ty(k[p>>2]|0,k[p+4>>2]|0,3)|0;s=L;break c;break}case 81:{p=l+88|0;q=ty(k[p>>2]|0,k[p+4>>2]|0,4)|0;s=L;break c;break}default:{w=u;break b}}}}while(0);o=b;g=ky(k[o>>2]|0,k[o+4>>2]|0,q|0,s|0)|0;m=L;o=b;k[o>>2]=g;k[o+4>>2]=m;n=n+1|0;if((n|0)>=(j|0))break a;else l=l+160|0}k[e>>2]=w;Xw(f,755371,e)|0;uk(5,f);k[c>>2]=261;h=261;r=d;return h|0}while(0);h=k[c>>2]|0;r=d;return h|0}function kl(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0;g=r;r=r+16|0;h=g+8|0;i=g+4|0;j=g;l=k[f>>2]|0;if((l|0)>0){m=l;r=g;return m|0}k[d>>2]=0;if((a|0)<1){m=l;r=g;return m|0}k[e>>2]=1;l=k[d>>2]|0;n=0;while(1){k[e+(n<<2)>>2]=l+1;Nk(k[b+(n<<2)>>2]|0,h,j,i,f)|0;o=(k[j>>2]|0)+c+(k[d>>2]|0)|0;k[d>>2]=o;n=n+1|0;if((n|0)==(a|0)){p=o;break}else l=o}l=k[f>>2]|0;k[d>>2]=p-c;m=l;r=g;return m|0}function ll(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,l=0,m=0,n=0,o=0,q=0.0,s=0,u=0,v=0,w=0;d=r;r=r+160|0;e=d;f=d+77|0;g=d+4|0;h=k[c>>2]|0;if((h|0)>0){l=h;r=d;return l|0}h=i[755408]|0;if(!(h<<24>>24)){m=Av()|0;n=i[k[m>>2]>>0]|0;i[755408]=n;o=n}else o=h;h=rv()|0;k[h>>2]=0;p[b>>3]=0.0;if((rx(a,68)|0)!=0|o<<24>>24==44){ux(g,a)|0;o=rx(g,68)|0;k[e>>2]=o;if(o)i[o>>0]=69;if((i[755408]|0)==44?(o=rx(g,46)|0,k[e>>2]=o,(o|0)!=0):0)i[o>>0]=44;q=+hx(g,e)}else q=+hx(a,e);p[b>>3]=q;g=i[k[e>>2]>>0]|0;p[t>>3]=q;e=ny(k[t>>2]|0,k[t+4>>2]|0,48)|0;switch(g<<24>>24){case 32:case 0:{s=e&65535;break}default:{u=f;v=755409;w=u+46|0;do{i[u>>0]=i[v>>0]|0;u=u+1|0;v=v+1|0}while((u|0)<(w|0));zx(f,a,30)|0;uk(5,f);k[c>>2]=409;s=j[b+6>>1]|0}}if((k[h>>2]|0)==34|(s&32752)==32752){u=f;v=755409;w=u+46|0;do{i[u>>0]=i[v>>0]|0;u=u+1|0;v=v+1|0}while((u|0)<(w|0));zx(f,a,30)|0;uk(5,f);p[b>>3]=0.0;k[c>>2]=412;k[h>>2]=0}l=k[c>>2]|0;r=d;return l|0}function ml(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;d=r;r=r+96|0;e=d;f=d+4|0;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=rv()|0;k[g>>2]=0;j=b;k[j>>2]=0;k[j+4>>2]=0;j=ix(a,e,10)|0;l=b;k[l>>2]=j;k[l+4>>2]=L;switch(i[k[e>>2]>>0]|0){case 32:case 0:break;default:k[c>>2]=407}if((k[g>>2]|0)==34){e=f;l=755304;j=e+58|0;do{i[e>>0]=i[l>>0]|0;e=e+1|0;l=l+1|0}while((e|0)<(j|0));zx(f,a,25)|0;uk(5,f);k[c>>2]=412;k[g>>2]=0}h=k[c>>2]|0;r=d;return h|0}function nl(a,b,c,d,e,f,g,h,j,l,m,n,o,q,s,t,u,v,w,x,y,z,A,B){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;s=s|0;t=t|0;u=u|0;v=v|0;w=w|0;x=x|0;y=y|0;z=z|0;A=A|0;B=B|0;var C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0.0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0;C=r;r=r+304|0;D=C+200|0;E=C+192|0;F=C+176|0;G=C+168|0;H=C+160|0;I=C+136|0;J=C+120|0;K=C+112|0;M=C+104|0;N=C+96|0;O=C+88|0;P=C+80|0;Q=C+72|0;R=C+64|0;S=C+56|0;T=C+48|0;U=C+40|0;V=C+32|0;W=C+24|0;X=C+16|0;Y=C+204|0;Z=C+8|0;_=C;$=C+208|0;k[Y>>2]=0;aa=k[a>>2]|0;ba=a+4|0;ca=k[ba>>2]|0;do if((aa|0)==(k[ca+68>>2]|0)){da=ca+120|0;if((k[da>>2]|0)==-1?(k[da+4>>2]|0)==-1:0){if((Vk(a,B)|0)<=0)break;ea=k[B>>2]|0;r=C;return ea|0}if((j|0)>0){da=k[30045]|0;if((da+-1|0)>>>0>39){Eh(755455,120180)|0;fa=k[ba>>2]|0;ga=k[30045]|0}else{fa=ca;ga=da}if((k[fa+4>>2]|0)==(ga|0)?(da=fa+104|0,ha=k[da>>2]|0,ia=k[da+4>>2]|0,da=fa+96|0,ja=k[da>>2]|0,ka=k[da+4>>2]|0,da=fa+120|0,la=ky(k[da>>2]|0,k[da+4>>2]|0,-2880,-1)|0,da=L,ma=(ka|0)>(da|0)|(ka|0)==(da|0)&ja>>>0>la>>>0,!((ha|0)==((ma?ja:la)|0)&(ia|0)==((ma?ka:da)|0))):0)ol(a,B)|0}}else Fk(a,aa+1|0,0,B)|0;while(0);aa=k[ba>>2]|0;fa=k[aa+72>>2]|0;if((d|0)<0|(d|0)==0&c>>>0<1){na=+(c>>>0)+4294967296.0*+(d|0);if(!fa){p[X>>3]=na;Xw($,755465,X)|0;uk(5,$);k[B>>2]=307;ea=307;r=C;return ea|0}else{p[W>>3]=na;Xw($,755505,W)|0;uk(5,$);k[B>>2]=307;ea=307;r=C;return ea|0}}if(((f|0)<0|(f|0)==0&e>>>0<1)&(fa|0)!=1){k[V>>2]=e;Xw($,755546,V)|0;uk(5,$);k[B>>2]=308;ea=308;r=C;return ea|0}if((h|0)<0){p[U>>3]=+(g>>>0)+4294967296.0*+(h|0);Xw($,755587,U)|0;uk(5,$);k[B>>2]=306;ea=306;r=C;return ea|0}if((b|0)>=1?(k[aa+928>>2]|0)>=(b|0):0){k[y>>2]=fa;fa=aa+952|0;U=k[fa+4>>2]|0;V=x;k[V>>2]=k[fa>>2];k[V+4>>2]=U;U=aa+120|0;V=k[U>>2]|0;fa=k[U+4>>2]|0;U=k[aa+960>>2]|0;aa=b+-1|0;p[l>>3]=+p[U+(aa*160|0)+96>>3];p[m>>3]=+p[U+(aa*160|0)+104>>3];m=U+(aa*160|0)+112|0;l=k[m+4>>2]|0;W=z;k[W>>2]=k[m>>2];k[W+4>>2]=l;l=U+(aa*160|0)+72|0;W=k[l>>2]|0;m=k[l+4>>2]|0;l=U+(aa*160|0)+152|0;k[o>>2]=k[l>>2];k[v>>2]=k[l>>2];l=U+(aa*160|0)+80|0;k[q>>2]=k[l>>2];z=U+(aa*160|0)+88|0;X=k[z+4>>2]|0;ga=w;k[ga>>2]=k[z>>2];k[ga+4>>2]=X;ux(n,U+(aa*160|0)+140|0)|0;ux(A,U+(aa*160|0)+120|0)|0;if((k[y>>2]|0)==1?(i[A>>0]|0)==0:0){aa=A;U=755675;n=aa+18|0;do{i[aa>>0]=i[U>>0]|0;aa=aa+1|0;U=U+1|0}while((aa|0)<(n|0));U=k[o>>2]|0;i[A+((U|0)>17?17:U)>>0]=0}U=(j|0)==-1;A=U?0:j;j=k[q>>2]|0;if((((j|0)>-1?j:0-j|0)|0)==1){aa=j*11|0;k[q>>2]=aa;n=w;X=ky(k[n>>2]|0,k[n+4>>2]|0,7,0)|0;n=wy(X|0,L|0,8,0)|0;X=w;k[X>>2]=n;k[X+4>>2]=L;oa=aa}else oa=j;j=k[y>>2]|0;a:do if((j|0)==2)switch(oa|0){case 16:{aa=w;X=k[o>>2]|0;n=wy(k[aa>>2]|0,k[aa+4>>2]|0,X|0,((X|0)<0)<<31>>31|0)|0;X=w;k[X>>2]=n;k[X+4>>2]=L;break a;break}case -16:{k[v>>2]=1;k[o>>2]=g;break a;break}default:break a}while(0);X=(j|0)==1;j=ky(e|0,f|0,-1,-1)|0;f=u;k[f>>2]=X?0:j;k[f+4>>2]=X?0:L;if((((oa|0)>-1?oa:0-oa|0)|0)>82){X=(((oa|0)>0?1:-1)+oa|0)/2|0;k[q>>2]=X;f=w;j=ty(k[f>>2]|0,k[f+4>>2]|0,1)|0;f=w;k[f>>2]=j;k[f+4>>2]=L;k[o>>2]=(k[o>>2]|0)/2|0;k[v>>2]=(k[v>>2]|0)/2|0;pa=X}else pa=oa;oa=(pa|0)>-1?pa:0-pa|0;b:do switch(oa|0){case 42:{k[s>>2]=7200;break}case 82:{k[s>>2]=3600;break}default:{pa=k[o>>2]|0;if((oa|0)!=16){k[s>>2]=28800/(pa|0)|0;break b}X=28799/(pa|0)|0;k[s>>2]=X;if(!X){k[R>>2]=pa;k[R+4>>2]=28799;Xw($,755693,R)|0;uk(5,$);k[B>>2]=236;ea=236;r=C;return ea|0}}}while(0);R=ky(c|0,d|0,-1,-1)|0;s=x;oa=yy(k[s>>2]|0,k[s+4>>2]|0,R|0,L|0)|0;R=L;s=ky(W|0,m|0,V|0,fa|0)|0;m=ky(s|0,L|0,oa|0,R|0)|0;R=t;k[R>>2]=m;k[R+4>>2]=L;R=(k[y>>2]|0)==0;y=(A|0)!=0;do if(y&R){m=w;oa=k[m>>2]|0;s=k[m+4>>2]|0;m=u;W=ky(k[m>>2]|0,k[m+4>>2]|0,g|0,h|0)|0;m=L;if((s|0)<(m|0)|(s|0)==(m|0)&oa>>>0<W>>>0){oa=w;k[oa>>2]=W;k[oa+4>>2]=m}}else{m=k[q>>2]|0;if((m|0)>0){oa=u;W=k[oa>>2]|0;s=k[oa+4>>2]|0;oa=w;o=k[oa>>2]|0;pa=k[oa+4>>2]|0;if(!((s|0)<(pa|0)|(s|0)==(pa|0)&W>>>0<o>>>0)){oa=ky(W|0,s|0,1,0)|0;k[Q>>2]=oa;k[Q+4>>2]=o;Xw($,755757,Q)|0;uk(5,$);k[B>>2]=308;ea=308;r=C;return ea|0}oa=ky(g|0,h|0,-1,-1)|0;X=ky(oa|0,L|0,W|0,s|0)|0;s=wy(X|0,L|0,o|0,pa|0)|0;pa=ky(s|0,L|0,c|0,d|0)|0;s=L;o=k[ba>>2]|0;X=o+944|0;W=X;oa=k[W>>2]|0;f=k[W+4>>2]|0;W=(s|0)>(f|0)|(s|0)==(f|0)&pa>>>0>oa>>>0;c:do if(y){if(((h|0)>0|(h|0)==0&g>>>0>0)&W){do if(k[o+40>>2]|0){j=o+976|0;n=k[j+4>>2]|0;if((n|0)>0|(n|0)==0&(k[j>>2]|0)>>>0>0)break;j=jy(pa|0,s|0,oa|0,f|0)|0;n=o+952|0;aa=yy(k[n>>2]|0,k[n+4>>2]|0,j|0,L|0)|0;j=o+968|0;n=j;ga=ky(k[n>>2]|0,k[n+4>>2]|0,aa|0,L|0)|0;aa=j;k[aa>>2]=ga;k[aa+4>>2]=L;aa=X;k[aa>>2]=pa;k[aa+4>>2]=s;break c}while(0);aa=jy(pa|0,s|0,oa|0,f|0)|0;ga=L;if((Nj(a,oa,f,aa,ga,B)|0)<=0)break;p[P>>3]=+(aa>>>0)+4294967296.0*+(ga|0);Xw($,755824,P)|0;uk(5,$);ea=k[B>>2]|0;r=C;return ea|0}}else if(W&(U^1)){do if(R)if((f|0)<(d|0)|(f|0)==(d|0)&oa>>>0<c>>>0){k[O>>2]=c;Xw($,755872,O)|0;uk(5,$);k[N>>2]=k[(k[ba>>2]|0)+944>>2];Xw($,755917,N)|0;uk(5,$);break}else{uk(5,755961);k[M>>2]=k[w>>2];Xw($,755996,M)|0;uk(5,$);k[K>>2]=g;k[K+4>>2]=e;Xw($,756023,K)|0;uk(5,$);break}else{uk(5,756077);ga=(k[ba>>2]|0)+944|0;aa=w;na=+((k[aa>>2]|0)>>>0)+4294967296.0*+(k[aa+4>>2]|0);p[J>>3]=+((k[ga>>2]|0)>>>0)+4294967296.0*+(k[ga+4>>2]|0);p[J+8>>3]=na;Xw($,756112,J)|0;uk(5,$);ga=u;aa=ky(k[ga>>2]|0,k[ga+4>>2]|0,1,0)|0;p[I>>3]=+(g>>>0)+4294967296.0*+(h|0);p[I+8>>3]=+(c>>>0)+4294967296.0*+(d|0);p[I+16>>3]=+(aa>>>0)+4294967296.0*+(L|0);Xw($,756162,I)|0;uk(5,$)}while(0);k[B>>2]=307;ea=307;r=C;return ea|0}while(0);oa=w;if(!((A|0)!=2&(((h|0)>0|(h|0)==0&g>>>0>1)&((k[oa>>2]|0)==1&(k[oa+4>>2]|0)==0))))break;oa=x;f=k[oa>>2]|0;W=k[oa+4>>2]|0;if(!((W|0)<0|(W|0)==0&f>>>0<2147483648))break;k[v>>2]=f;f=w;k[f>>2]=g;k[f+4>>2]=h;break}k[q>>2]=0-m;if(!y){f=(k[ba>>2]|0)+944|0;W=k[f+4>>2]|0;if((W|0)<(d|0)|((W|0)==(d|0)?(k[f>>2]|0)>>>0<c>>>0:0)){uk(5,756297);f=(k[ba>>2]|0)+944|0;p[F>>3]=+((k[f>>2]|0)>>>0)+4294967296.0*+(k[f+4>>2]|0);p[F+8>>3]=+(c>>>0)+4294967296.0*+(d|0);Xw($,756331,F)|0;uk(5,$);k[B>>2]=307;ea=307;r=C;return ea|0}pl(a,b,c,d,_,Z,B)|0;f=_;W=k[f>>2]|0;oa=k[f+4>>2]|0;f=w;k[f>>2]=W;k[f+4>>2]=oa;f=k[l>>2]|0;if((f|0)>=-82)if((f|0)==-1){f=ky(W|0,oa|0,7,0)|0;s=wy(f|0,L|0,8,0)|0;f=L;pa=w;k[pa>>2]=s;k[pa+4>>2]=f;qa=f;ra=s}else{qa=oa;ra=W}else{s=ty(W|0,oa|0,1)|0;oa=L;W=w;k[W>>2]=s;k[W+4>>2]=oa;qa=oa;ra=s}s=u;oa=k[s+4>>2]|0;if((oa|0)<(qa|0)|((oa|0)==(qa|0)?(k[s>>2]|0)>>>0<ra>>>0:0)){s=Z;oa=ky(k[s>>2]|0,k[s+4>>2]|0,V|0,fa|0)|0;s=(k[ba>>2]|0)+968|0;W=ky(oa|0,L|0,k[s>>2]|0,k[s+4>>2]|0)|0;s=t;k[s>>2]=W;k[s+4>>2]=L;break}k[E>>2]=e;Xw($,756381,E)|0;uk(5,$);k[D>>2]=k[w>>2];Xw($,756450,D)|0;uk(5,$);k[B>>2]=308;ea=308;r=C;return ea|0}s=u;W=ky(k[s>>2]|0,k[s+4>>2]|0,g|0,h|0)|0;s=w;k[s>>2]=W;k[s+4>>2]=L;s=(k[ba>>2]|0)+944|0;W=k[s+4>>2]|0;if(!((W|0)<(d|0)|((W|0)==(d|0)?(k[s>>2]|0)>>>0<c>>>0:0))?(pl(a,b,c,d,_,Z,Y)|0,(k[Y>>2]|0)==0):0){s=k[l>>2]|0;W=(s|0)<-82;do if(W){oa=_;f=ty(k[oa>>2]|0,k[oa+4>>2]|0,1)|0;oa=L;pa=_;k[pa>>2]=f;k[pa+4>>2]=oa;sa=oa;ta=f}else{f=_;oa=k[f>>2]|0;pa=k[f+4>>2]|0;if((s|0)!=-1){sa=pa;ta=oa;break}f=ky(oa|0,pa|0,7,0)|0;pa=wy(f|0,L|0,8,0)|0;f=L;oa=_;k[oa>>2]=pa;k[oa+4>>2]=f;sa=f;ta=pa}while(0);s=w;m=k[s+4>>2]|0;if(!((sa|0)<(m|0)|((sa|0)==(m|0)?ta>>>0<(k[s>>2]|0)>>>0:0))){s=Z;m=k[s>>2]|0;pa=k[s+4>>2]|0;s=ky(m|0,pa|0,V|0,fa|0)|0;f=(k[ba>>2]|0)+968|0;oa=ky(s|0,L|0,k[f>>2]|0,k[f+4>>2]|0)|0;f=t;k[f>>2]=oa;k[f+4>>2]=L;f=w;oa=k[f>>2]|0;s=k[f+4>>2]|0;if(W){f=wy(oa|0,s|0,2,0)|0;ql(a,b,c,d,f,L,m,pa,B)|0}else ql(a,b,c,d,oa,s,m,pa,B)|0;ea=k[B>>2]|0;r=C;return ea|0}}pa=k[ba>>2]|0;m=pa+944|0;s=k[m>>2]|0;oa=k[m+4>>2]|0;do if((oa|0)<(d|0)|(oa|0)==(d|0)&s>>>0<c>>>0){m=jy(c|0,d|0,s|0,oa|0)|0;f=L;if((Nj(a,s,oa,m,f,B)|0)<=0){ua=k[ba>>2]|0;break}p[H>>3]=+(m>>>0)+4294967296.0*+(f|0);Xw($,755824,H)|0;uk(5,$);ea=k[B>>2]|0;r=C;return ea|0}else ua=pa;while(0);pa=ua+968|0;oa=ky(k[pa>>2]|0,k[pa+4>>2]|0,V|0,fa|0)|0;pa=ua+976|0;s=pa;W=ky(oa|0,L|0,k[s>>2]|0,k[s+4>>2]|0)|0;s=t;k[s>>2]=W;k[s+4>>2]=L;s=w;W=k[s>>2]|0;oa=k[s+4>>2]|0;if((k[l>>2]|0)<-82){s=wy(W|0,oa|0,2,0)|0;f=pa;ql(a,b,c,d,s,L,k[f>>2]|0,k[f+4>>2]|0,B)|0}else{f=pa;ql(a,b,c,d,W,oa,k[f>>2]|0,k[f+4>>2]|0,B)|0}f=k[ba>>2]|0;do if(!(k[f+40>>2]|0)){oa=f+968|0;W=ky(k[oa>>2]|0,k[oa+4>>2]|0,V|0,fa|0)|0;oa=f+976|0;pa=ky(W|0,L|0,k[oa>>2]|0,k[oa+4>>2]|0)|0;oa=L;W=w;s=k[v>>2]|0;m=yy(s|0,((s|0)<0)<<31>>31|0,k[W>>2]|0,k[W+4>>2]|0)|0;W=ky(pa|0,oa|0,m|0,L|0)|0;m=L;oa=(k[f+88>>2]|0)+((k[f+68>>2]|0)+1<<3)|0;pa=k[oa>>2]|0;s=k[oa+4>>2]|0;if(!((m|0)>(s|0)|(m|0)==(s|0)&W>>>0>pa>>>0)){va=f;break}oa=ky(W|0,m|0,-1,-1)|0;m=jy(oa|0,L|0,pa|0,s|0)|0;s=wy(m|0,L|0,2880,0)|0;m=ky(s|0,L|0,1,0)|0;if((rl(a,m,1,B)|0)<=0){va=k[ba>>2]|0;break}k[G>>2]=m;Xw($,756228,G)|0;uk(5,$);ea=k[B>>2]|0;r=C;return ea|0}else va=f;while(0);f=w;m=k[v>>2]|0;s=yy(m|0,((m|0)<0)<<31>>31|0,k[f>>2]|0,k[f+4>>2]|0)|0;f=va+976|0;m=f;pa=ky(k[m>>2]|0,k[m+4>>2]|0,s|0,L|0)|0;s=f;k[s>>2]=pa;k[s+4>>2]=L}while(0);ea=k[B>>2]|0;r=C;return ea|0}k[T>>2]=b;Xw($,771145,T)|0;uk(5,$);k[S>>2]=k[(k[ba>>2]|0)+928>>2];Xw($,755637,S)|0;uk(5,$);k[B>>2]=302;ea=302;r=C;return ea|0}function ol(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0;c=r;r=r+256|0;d=c;e=c+166|0;f=c+85|0;g=c+4|0;h=g;j=h+81|0;do{i[h>>0]=0;h=h+1|0}while((h|0)<(j|0));l=k[b>>2]|0;if((l|0)>0){m=l;r=c;return m|0}l=a+4|0;n=k[l>>2]|0;o=n+96|0;p=k[o>>2]|0;q=k[o+4>>2]|0;o=n+120|0;n=o;s=k[n>>2]|0;t=k[n+4>>2]|0;if((s|0)==-1&(t|0)==-1){n=xy(p|0,q|0,2880,0)|0;u=L;v=ky(p|0,q|0,2880,0)|0;w=jy(v|0,L|0,n|0,u|0)|0;u=L;n=o;k[n>>2]=w;k[n+4>>2]=u;x=w;y=u}else{x=s;y=t}t=jy(x|0,y|0,p|0,q|0)|0;y=wy(t|0,L|0,80,0)|0;h=e;t=770037;j=h+41|0;do{i[h>>0]=i[t>>0]|0;h=h+1|0;t=t+1|0}while((h|0)<(j|0));h=e+(xx(e)|0)|0;t=770037;j=h+41|0;do{i[h>>0]=i[t>>0]|0;h=h+1|0;t=t+1|0}while((h|0)<(j|0));h=f;t=756488;j=h+41|0;do{i[h>>0]=i[t>>0]|0;h=h+1|0;t=t+1|0}while((h|0)<(j|0));h=f+(xx(f)|0)|0;t=770037;j=h+41|0;do{i[h>>0]=i[t>>0]|0;h=h+1|0;t=t+1|0}while((h|0)<(j|0));k[d>>2]=0;au(a,p,q,0,d)|0;t=(y|0)>0;a:do if(t){h=0;while(1){fu(a,80,0,g,d)|0;if(k[d>>2]|0)break a;if((Ax(g,e,80)|0)!=0?(Ax(g,f,80)|0)!=0:0){z=0;A=h;B=10;break a}j=h+1|0;if((j|0)<(y|0))h=j;else{z=0;A=j;B=10;break}}}else{z=(k[d>>2]|0)!=0;A=0;B=10}while(0);if(((B|0)==10?!((A|0)!=(y|0)|z):0)?(z=(k[l>>2]|0)+120|0,A=ky(k[z>>2]|0,k[z+4>>2]|0,-2880,-1)|0,z=L,B=(q|0)>(z|0)|(q|0)==(z|0)&p>>>0>A>>>0,h=B?p:A,A=B?q:z,au(a,h,A,0,d)|0,fu(a,80,0,g,d)|0,z=Ax(g,f,80)|0,(k[d>>2]|z|0)==0):0){z=(k[l>>2]|0)+104|0;k[z>>2]=h;k[z+4>>2]=A;m=k[b>>2]|0;r=c;return m|0}A=(k[l>>2]|0)+96|0;z=k[A>>2]|0;h=k[A+4>>2]|0;au(a,z,h,1,b)|0;if(t){t=0;do{cu(a,80,0,e,b)|0;t=t+1|0}while((t|0)!=(y|0))}y=(k[l>>2]|0)+120|0;t=ky(k[y>>2]|0,k[y+4>>2]|0,-2880,-1)|0;y=L;e=(h|0)>(y|0)|(h|0)==(y|0)&z>>>0>t>>>0;A=e?z:t;t=e?h:y;au(a,A,t,0,b)|0;cu(a,80,0,f,b)|0;f=(k[l>>2]|0)+104|0;k[f>>2]=A;k[f+4>>2]=t;t=k[b>>2]|0;if((t|0)<=0){m=t;r=c;return m|0}uk(5,756529);m=k[b>>2]|0;r=c;return m|0}function pl(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;h=r;r=r+32|0;j=h+16|0;l=h;m=j;k[m>>2]=0;k[m+4>>2]=0;k[l>>2]=0;k[l+4>>2]=0;k[l+8>>2]=0;k[l+12>>2]=0;m=k[g>>2]|0;if((m|0)>0){n=m;r=h;return n|0}m=k[a>>2]|0;o=a+4|0;p=k[o>>2]|0;if((m|0)==(k[p+68>>2]|0)){q=p+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,g)|0)>0:0){n=k[g>>2]|0;r=h;return n|0}}else Fk(a,m+1|0,0,g)|0;m=k[o>>2]|0;o=k[m+960>>2]|0;q=b+-1|0;if((k[o+(q*160|0)+80>>2]|0)>-1){k[g>>2]=317;n=317;r=h;return n|0}b=m+120|0;p=k[b>>2]|0;s=k[b+4>>2]|0;b=m+952|0;m=k[b>>2]|0;t=k[b+4>>2]|0;b=ky(c|0,d|0,-1,-1)|0;d=yy(m|0,t|0,b|0,L|0)|0;b=ky(d|0,L|0,p|0,s|0)|0;s=o+(q*160|0)+72|0;p=ky(b|0,L|0,k[s>>2]|0,k[s+4>>2]|0)|0;s=L;if((i[o+(q*160|0)+140>>0]|0)!=80?(i[o+(q*160|0)+141>>0]|0)!=80:0){if((ou(a,p,s,2,8,l,g)|0)<1){if(e){q=l;o=k[q+4>>2]|0;b=e;k[b>>2]=k[q>>2];k[b+4>>2]=o}if(f){o=l+8|0;l=k[o+4>>2]|0;b=f;k[b>>2]=k[o>>2];k[b+4>>2]=l}}}else u=11;if((u|0)==11?(nu(a,p,s,2,4,j,g)|0)<1:0){if(e){s=e;k[s>>2]=k[j>>2];k[s+4>>2]=0}if(f){s=f;k[s>>2]=k[j+4>>2];k[s+4>>2]=0}}n=k[g>>2]|0;r=h;return n|0}function ql(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;l=r;r=r+32|0;m=l+16|0;n=l;o=k[j>>2]|0;if((o|0)>0){p=o;r=l;return p|0}o=k[a>>2]|0;q=a+4|0;s=k[q>>2]|0;if((o|0)==(k[s+68>>2]|0)){t=s+120|0;if(((k[t>>2]|0)==-1?(k[t+4>>2]|0)==-1:0)?(Vk(a,j)|0)>0:0){p=k[j>>2]|0;r=l;return p|0}}else Fk(a,o+1|0,0,j)|0;o=k[q>>2]|0;q=k[o+960>>2]|0;t=b+-1|0;if((k[q+(t*160|0)+80>>2]|0)>-1)k[j>>2]=317;b=o+120|0;s=k[b>>2]|0;u=k[b+4>>2]|0;b=o+952|0;o=k[b>>2]|0;v=k[b+4>>2]|0;b=ky(c|0,d|0,-1,-1)|0;d=yy(o|0,v|0,b|0,L|0)|0;b=ky(d|0,L|0,s|0,u|0)|0;u=q+(t*160|0)+72|0;s=ky(b|0,L|0,k[u>>2]|0,k[u+4>>2]|0)|0;au(a,s,L,1,j)|0;if((i[q+(t*160|0)+140>>0]|0)!=80?(i[q+(t*160|0)+141>>0]|0)!=80:0){t=n;k[t>>2]=e;k[t+4>>2]=f;t=n+8|0;k[t>>2]=g;k[t+4>>2]=h;vu(a,2,8,n,j)|0}else w=11;do if((w|0)==11){n=h|f;if(!(n>>>0>0|(n|0)==0&(g|e)>>>0>4294967295)){k[m>>2]=e;k[m+4>>2]=g;uu(a,2,4,m,j)|0;break}uk(5,756568);k[j>>2]=412;p=412;r=l;return p|0}while(0);p=k[j>>2]|0;r=l;return p|0}function rl(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;e=r;r=r+5856|0;f=e;g=e+2968|0;h=e+88|0;j=e+4|0;l=k[d>>2]|0;if((b|0)<1|(l|0)>0){m=l;r=e;return m|0}n=(c|0)==0;o=a+4|0;p=k[o>>2]|0;do if(!n){q=(k[p+72>>2]|0)==1?32:0;if((c|0)==-1){s=j;t=756620;u=s+48|0;do{i[s>>0]=i[t>>0]|0;s=s+1|0;t=t+1|0}while((s|0)<(u|0));v=0;w=0;x=q;break}else{t=p+120|0;s=p+968|0;u=k[s>>2]|0;y=k[s+4>>2]|0;s=p+976|0;z=k[s>>2]|0;A=k[s+4>>2]|0;s=ky(k[t>>2]|0,k[t+4>>2]|0,2879,0)|0;t=ky(s|0,L|0,u|0,y|0)|0;y=ky(t|0,L|0,z|0,A|0)|0;A=L;z=xy(y|0,A|0,2880,0)|0;t=jy(y|0,A|0,z|0,L|0)|0;v=t;w=L;x=q;break}}else{t=p+120|0;v=k[t>>2]|0;w=k[t+4>>2]|0;x=32}while(0);sy(h|0,x|0,2880)|0;if((b|0)!=1){x=k[p+68>>2]|0;if((l|0)<1){do{Fk(a,(k[a>>2]|0)+2|0,f,d)|0;p=k[d>>2]|0}while((p|0)<1);B=p}else B=l;if((B|0)==107)k[d>>2]=l;Fk(a,x+1|0,f,d)|0;if((c|0)==-1)wq(a,1,j,d)|0;f=k[o>>2]|0;x=(k[f+88>>2]|0)+((k[f+80>>2]|0)+1<<3)|0;f=k[x>>2]|0;B=k[x+4>>2]|0;x=jy(f|0,B|0,v|0,w|0)|0;p=wy(x|0,L|0,2880,0)|0;a:do if((p|0)>0){x=yy(b|0,((b|0)<0)<<31>>31|0,2880,0)|0;t=L;z=f;A=B;y=0;while(1){z=ky(z|0,A|0,-2880,-1)|0;A=L;if((au(a,z,A,0,d)|0)>0)break;fu(a,2880,0,g,d)|0;u=ky(z|0,A|0,x|0,t|0)|0;au(a,u,L,1,d)|0;cu(a,2880,0,g,d)|0;y=y+1|0;if((y|0)>=(p|0))break a}m=k[d>>2]|0;r=e;return m|0}while(0);au(a,v,w,1,d)|0;if((b|0)>0){p=0;do{cu(a,2880,0,h,d)|0;p=p+1|0}while((p|0)!=(b|0))}}else{if((c|0)==-1)wq(a,1,j,d)|0;au(a,v,w,0,d)|0;fu(a,2880,0,g,d)|0;b:do if((k[d>>2]|0)<1){j=v;c=w;p=g;B=h;while(1){au(a,j,c,0,d)|0;cu(a,2880,0,B,d)|0;f=k[d>>2]|0;if((f|0)>0){m=f;break}f=ky(j|0,c|0,2880,0)|0;y=L;au(a,f,y,0,d)|0;fu(a,2880,0,B,d)|0;if((k[d>>2]|0)<1){t=B;j=f;c=y;B=p;p=t}else{C=f;D=y;E=p;break b}}r=e;return m|0}else{C=v;D=w;E=h}while(0);k[d>>2]=l;au(a,C,D,1,d)|0;cu(a,2880,0,E,d)|0}if(n){n=yy(b|0,((b|0)<0)<<31>>31|0,2880,0)|0;E=k[o>>2]|0;a=E+120|0;D=a;C=ky(k[D>>2]|0,k[D+4>>2]|0,n|0,L|0)|0;n=a;k[n>>2]=C;k[n+4>>2]=L;F=E}else F=k[o>>2]|0;o=k[F+68>>2]|0;E=k[F+80>>2]|0;if((o|0)<=(E|0)){n=yy(b|0,((b|0)<0)<<31>>31|0,2880,0)|0;b=L;C=k[F+88>>2]|0;F=o;do{o=F;F=F+1|0;a=C+(F<<3)|0;D=a;l=ky(k[D>>2]|0,k[D+4>>2]|0,n|0,b|0)|0;D=a;k[D>>2]=l;k[D+4>>2]=L}while((o|0)<(E|0))}m=k[d>>2]|0;r=e;return m|0}function sl(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0;g=r;r=r+128|0;h=g+32|0;j=g+24|0;l=g+16|0;m=g+40|0;n=g+8|0;o=g;p=g+44|0;q=k[f>>2]|0;if((q|0)>0){s=q;r=g;return s|0}q=k[a>>2]|0;t=a+4|0;if((q|0)==(k[(k[t>>2]|0)+68>>2]|0)){if((Vk(a,f)|0)>0){s=k[f>>2]|0;r=g;return s|0}}else Fk(a,q+1|0,0,f)|0;q=(e|0)!=0;if(q)k[e>>2]=1;u=(b|0)!=0;if(u){v=(k[t>>2]|0)+976|0;w=k[v+4>>2]|0;x=b;k[x>>2]=k[v>>2];k[x+4>>2]=w}w=(c|0)!=0;if(w){x=c;k[x>>2]=0;k[x+4>>2]=0}x=(d|0)!=0;if(x){v=d;k[v>>2]=0;k[v+4>>2]=0}v=k[t>>2]|0;if((k[v+72>>2]|0)==2?(y=v+976|0,z=k[y>>2]|0,A=k[y+4>>2]|0,!((z|0)==0&(A|0)==0)):0){if((A|0)>0|(A|0)==0&z>>>0>2147483647){uk(5,756668);k[f>>2]=113;s=113;r=g;return s|0}A=ey(1,z)|0;if(!A){y=p;B=756720;C=y+43|0;do{i[y>>0]=i[B>>0]|0;y=y+1|0;B=B+1|0}while((y|0)<(C|0));uk(5,p);k[f>>2]=113;s=113;r=g;return s|0}a:do if((k[v+928>>2]|0)>=1){B=((z|0)<0)<<31>>31;y=1;while(1){if((k[f>>2]|0)>=1)break a;Yk(a,y,m,j,l,f)|0;C=k[m>>2]|0;if((C|0)<=0){D=(C|0)/-10|0;C=k[t>>2]|0;E=C+944|0;F=k[E+4>>2]|0;if((F|0)<0|(F|0)==0&(k[E>>2]|0)>>>0<1)G=C;else{C=1;E=0;F=1;while(1){pl(a,y,C,E,n,o,f)|0;H=n;I=k[H>>2]|0;if((k[m>>2]|0)==-1){J=ky(I|0,k[H+4>>2]|0,7,0)|0;K=(J|0)/8|0}else K=ha(I,D)|0;I=o;J=k[I>>2]|0;H=k[I+4>>2]|0;do if((H|0)<0)M=34;else{I=ky(J|0,H|0,K|0,((K|0)<0)<<31>>31|0)|0;N=L;if((N|0)>(B|0)|(N|0)==(B|0)&I>>>0>z>>>0){M=34;break}if((K|0)>0)O=0;else break;do{I=ky(J|0,H|0,O|0,0)|0;N=A+I|0;i[N>>0]=(i[N>>0]|0)+1<<24>>24;O=O+1|0}while((O|0)!=(K|0))}while(0);if((M|0)==34){M=0;if(q)k[e>>2]=0;k[h>>2]=F;k[h+4>>2]=y;Xw(p,756763,h)|0;uk(5,p)}H=F+1|0;J=((H|0)<0)<<31>>31;N=k[t>>2]|0;I=N+944|0;P=k[I+4>>2]|0;if((J|0)>(P|0)|((J|0)==(P|0)?H>>>0>(k[I>>2]|0)>>>0:0)){G=N;break}else{C=H;E=J;F=H}}}}else G=k[t>>2]|0;if((y|0)<(k[G+928>>2]|0))y=y+1|0;else break}}while(0);if((z|0)>0){G=0;t=0;p=0;h=0;e=0;while(1){q=i[A+e>>0]|0;M=q<<24>>24==0;K=ky(G|0,t|0,M&1|0,0)|0;O=L;o=ky(p|0,h|0,(M?0:q<<24>>24>1&1)|0,(M?0:0)|0)|0;M=L;e=e+1|0;if((e|0)==(z|0)){Q=K;R=O;S=o;T=M;break}else{G=K;t=O;p=o;h=M}}}else{Q=0;R=0;S=0;T=0}if(u){u=b;k[u>>2]=z;k[u+4>>2]=((z|0)<0)<<31>>31}if(w){w=c;k[w>>2]=Q;k[w+4>>2]=R}if(x){x=d;k[x>>2]=S;k[x+4>>2]=T}dy(A);s=k[f>>2]|0;r=g;return s|0}s=k[f>>2]|0;r=g;return s|0}function tl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;c=r;r=r+240|0;d=c+56|0;e=c+48|0;f=c+40|0;g=c+68|0;h=c+64|0;j=c+60|0;l=c+32|0;m=c+24|0;n=c+16|0;o=c+8|0;p=c+153|0;q=c+72|0;s=c;t=k[b>>2]|0;if((t|0)>0){u=t;r=c;return u|0}sl(a,0,l,m,j,b)|0;if(!(k[j>>2]|0)){k[b>>2]=264;u=264;r=c;return u|0}j=a+4|0;t=k[j>>2]|0;if((((k[t+72>>2]|0)==2?(v=t+976|0,!((k[v>>2]|0)==0&(k[v+4>>2]|0)==0)):0)?(v=l,l=m,!((k[l>>2]|k[v>>2]|0)==0?(k[l+4>>2]|k[v+4>>2]|0)==0:0)):0)?(k[b>>2]|0)<=0:0){if(Lh(g,756821,b)|0){w=q;x=756840;y=w+45|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(y|0));uk(5,q);u=k[b>>2]|0;r=c;return u|0}if(Wj(a,k[g>>2]|0,0,b)|0){w=q;x=756885;y=w+34|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(y|0));uk(5,q);Fh(k[g>>2]|0,b)|0;u=k[b>>2]|0;r=c;return u|0}v=cy(1e4)|0;if(!v){w=q;x=756919;y=w+43|0;do{i[w>>0]=i[x>>0]|0;w=w+1|0;x=x+1|0}while((w|0)<(y|0));uk(5,q);Fh(k[g>>2]|0,b)|0;k[b>>2]=113;u=113;r=c;return u|0}x=k[(k[g>>2]|0)+4>>2]|0;w=x+120|0;y=x+968|0;x=ky(k[y>>2]|0,k[y+4>>2]|0,k[w>>2]|0,k[w+4>>2]|0)|0;w=L;y=k[j>>2]|0;l=y+120|0;m=y+968|0;t=ky(k[m>>2]|0,k[m+4>>2]|0,k[l>>2]|0,k[l+4>>2]|0)|0;l=L;m=y+976|0;z=m;A=k[z>>2]|0;B=k[z+4>>2]|0;z=m;k[z>>2]=0;k[z+4>>2]=0;a:do if((k[y+928>>2]|0)<1)C=v;else{z=v;m=1e4;D=1;b:while(1){if((k[b>>2]|0)>=1){C=z;break a}Yk(k[g>>2]|0,D,h,e,f,b)|0;E=k[h>>2]|0;if((E|0)<=0){F=(E|0)/-10|0;E=k[j>>2]|0;G=E+944|0;H=k[G+4>>2]|0;if((H|0)<0|(H|0)==0&(k[G>>2]|0)>>>0<1){I=E;J=z;K=m}else{E=1;G=0;H=z;M=m;N=1;while(1){pl(k[g>>2]|0,D,E,G,n,o,b)|0;O=n;P=k[O>>2]|0;if((k[h>>2]|0)==-1){Q=ky(P|0,k[O+4>>2]|0,7,0)|0;R=(Q|0)/8|0}else R=ha(P,F)|0;do if((R|0)>(M|0)){P=fy(H,R)|0;if(P){S=P;T=R;break}k[b>>2]=113;S=H;T=M}else{S=H;T=M}while(0);P=k[j>>2]|0;do if(!(k[P+40>>2]|0)){Q=P+976|0;O=k[Q>>2]|0;U=k[Q+4>>2]|0;Q=((R|0)<0)<<31>>31;V=ky(R|0,Q|0,t|0,l|0)|0;W=ky(V|0,L|0,O|0,U|0)|0;U=L;O=(k[P+88>>2]|0)+((k[P+68>>2]|0)+1<<3)|0;V=k[O>>2]|0;X=k[O+4>>2]|0;if(!((U|0)>(X|0)|(U|0)==(X|0)&W>>>0>V>>>0)){Y=R;Z=Q;break}O=ky(W|0,U|0,-1,-1)|0;U=jy(O|0,L|0,V|0,X|0)|0;X=wy(U|0,L|0,2880,0)|0;U=ky(X|0,L|0,1,0)|0;if((rl(a,U,1,b)|0)<=0){Y=R;Z=Q;break}k[d>>2]=U;Xw(q,756228,d)|0;uk(5,q);Y=R;Z=Q}else{Y=R;Z=((R|0)<0)<<31>>31}while(0);P=k[g>>2]|0;Q=o;U=ky(x|0,w|0,k[Q>>2]|0,k[Q+4>>2]|0)|0;au(P,U,L,0,b)|0;fu(k[g>>2]|0,Y,Z,S,b)|0;U=(k[j>>2]|0)+976|0;P=ky(k[U>>2]|0,k[U+4>>2]|0,t|0,l|0)|0;au(a,P,L,1,b)|0;cu(a,Y,Z,S,b)|0;P=n;U=(k[j>>2]|0)+976|0;ql(a,D,E,G,k[P>>2]|0,k[P+4>>2]|0,k[U>>2]|0,k[U+4>>2]|0,b)|0;U=k[j>>2]|0;P=U+976|0;Q=P;X=ky(k[Q>>2]|0,k[Q+4>>2]|0,Y|0,Z|0)|0;Q=P;k[Q>>2]=X;k[Q+4>>2]=L;Q=N+1|0;if((k[b>>2]|0)>0){_=S;break b}X=((Q|0)<0)<<31>>31;P=U+944|0;V=k[P+4>>2]|0;if((X|0)>(V|0)|((X|0)==(V|0)?Q>>>0>(k[P>>2]|0)>>>0:0)){I=U;J=S;K=T;break}else{E=Q;G=X;H=S;M=T;N=Q}}}}else{I=k[j>>2]|0;J=z;K=m}if((D|0)<(k[I+928>>2]|0)){z=J;m=K;D=D+1|0}else{C=J;break a}}dy(_);Fh(k[g>>2]|0,b)|0;u=k[b>>2]|0;r=c;return u|0}while(0);dy(C);Fh(k[g>>2]|0,b)|0;g=k[j>>2]|0;C=(k[g+88>>2]|0)+((k[g+68>>2]|0)+1<<3)|0;_=g+976|0;g=_;J=k[g>>2]|0;K=k[g+4>>2]|0;g=jy(k[C>>2]|0,k[C+4>>2]|0,t|0,l|0)|0;l=jy(g|0,L|0,J|0,K|0)|0;g=wy(l|0,L|0,2880,0)|0;if((g|0)>0){l=_;k[l>>2]=A;k[l+4>>2]=B;ul(a,g,b)|0;g=(k[j>>2]|0)+976|0;k[g>>2]=J;k[g+4>>2]=K}Do(a,2,b)|0;No(a,782652,s,p,b)|0;K=(k[j>>2]|0)+976|0;j=k[K>>2]|0;g=k[K+4>>2]|0;K=s;if(!((j|0)==(k[K>>2]|0)?(g|0)==(k[K+4>>2]|0):0))oq(a,782652,j,g,p,b)|0;Vk(a,b)|0;u=k[b>>2]|0;r=c;return u|0}u=k[b>>2]|0;r=c;return u|0}function ul(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;d=r;r=r+2896|0;e=d+8|0;f=d;g=k[c>>2]|0;if((b|0)<1|(g|0)>0){h=g;r=d;return h|0}k[f>>2]=0;g=a+4|0;i=k[g>>2]|0;j=i+120|0;l=i+968|0;m=k[l>>2]|0;n=k[l+4>>2]|0;l=i+976|0;i=k[l>>2]|0;o=k[l+4>>2]|0;l=ky(k[j>>2]|0,k[j+4>>2]|0,2879,0)|0;j=ky(l|0,L|0,m|0,n|0)|0;n=ky(j|0,L|0,i|0,o|0)|0;o=L;i=xy(n|0,o|0,2880,0)|0;j=jy(n|0,o|0,i|0,L|0)|0;i=L;o=yy(b|0,((b|0)<0)<<31>>31|0,2880,0)|0;n=L;m=jy(j|0,i|0,o|0,n|0)|0;l=L;a:do if(!(au(a,j,i,0,f)|0)){p=m;q=l;s=j;t=i;while(1){if(fu(a,2880,0,e,f)|0){u=p;v=q;break a}au(a,p,q,0,c)|0;cu(a,2880,0,e,c)|0;if((k[c>>2]|0)>0)break;s=ky(s|0,t|0,2880,0)|0;t=L;w=ky(p|0,q|0,2880,0)|0;x=L;if(au(a,s,t,0,f)|0){u=w;v=x;break a}else{p=w;q=x}}uk(5,756962);h=k[c>>2]|0;r=d;return h|0}else{u=m;v=l}while(0);sy(e|0,0,2880)|0;au(a,u,v,0,c)|0;if((b|0)>0){l=0;do{cu(a,2880,0,e,c)|0;l=l+1|0}while((l|0)!=(b|0))}b=ky(u|0,v|0,-1,-1)|0;au(a,b,L,0,c)|0;di(a,u,v,c)|0;v=k[g>>2]|0;g=k[v+68>>2]|0;u=k[v+80>>2]|0;if((g|0)<=(u|0)){a=k[v+88>>2]|0;v=g;do{g=v;v=v+1|0;b=a+(v<<3)|0;l=b;e=jy(k[l>>2]|0,k[l+4>>2]|0,o|0,n|0)|0;l=b;k[l>>2]=e;k[l+4>>2]=L}while((g|0)<(u|0))}h=k[c>>2]|0;r=d;return h|0}function vl(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0;h=r;r=r+16|0;i=h+8|0;j=h;do if((pl(a,b,c,d,i,j,g)|0)<=0){do if(e){l=i;m=k[l>>2]|0;n=k[l+4>>2]|0;if((n|0)>0|(n|0)==0&m>>>0>2147483647){k[g>>2]=412;break}else{k[e>>2]=m;break}}while(0);if(f){m=j;n=k[m>>2]|0;l=k[m+4>>2]|0;if((l|0)>0|(l|0)==0&n>>>0>2147483647){k[g>>2]=412;break}else{k[f>>2]=n;break}}}while(0);r=h;return k[g>>2]|0}function wl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;c=r;r=r+96|0;d=c;e=c+8|0;f=c+4|0;g=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((g|0)==(k[i+68>>2]|0)){if((k[i+76>>2]|0)==1){Eh(755455,f)|0;i=k[h>>2]|0;if((k[i+4>>2]|0)==(k[f>>2]|0))j=i;else{Vk(a,b)|0;j=k[h>>2]|0}i=j+976|0;j=k[i+4>>2]|0;if((j|0)>0|(j|0)==0&(k[i>>2]|0)>>>0>0)xl(a,b)|0;yl(a,b)|0}}else Fk(a,g+1|0,0,b)|0;g=k[h>>2]|0;if((k[g+8>>2]|0)==1?(a=k[g+960>>2]|0,(a|0)!=0):0){dy(a);a=k[h>>2]|0;k[a+960>>2]=0;if(!(k[a+1220>>2]|0))l=a;else{i=((k[a+1104>>2]|0)+-1|0)/(k[a+1044>>2]|0)|0;if((i|0)>-1){j=a;f=0;while(1){m=k[(k[j+1232>>2]|0)+(f<<2)>>2]|0;if(!m)n=j;else{dy(m);n=k[h>>2]|0}m=k[(k[n+1236>>2]|0)+(f<<2)>>2]|0;if(!m)o=n;else{dy(m);o=k[h>>2]|0}if((f|0)==(i|0)){p=o;break}else{j=o;f=f+1|0}}}else p=a;dy(k[p+1240>>2]|0);dy(k[(k[h>>2]|0)+1228>>2]|0);dy(k[(k[h>>2]|0)+1224>>2]|0);dy(k[(k[h>>2]|0)+1236>>2]|0);dy(k[(k[h>>2]|0)+1232>>2]|0);dy(k[(k[h>>2]|0)+1220>>2]|0);p=k[h>>2]|0;h=p+1220|0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;k[h+16>>2]=0;k[h+20>>2]=0;l=p}}else l=g;g=k[b>>2]|0;if((g|0)<1|(g|0)==999){q=g;r=c;return q|0}k[d>>2]=k[l+68>>2];Xw(e,756998,d)|0;uk(5,e);q=k[b>>2]|0;r=c;return q|0}function xl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0;c=r;r=r+560|0;d=c+32|0;e=c+24|0;f=c+544|0;g=c+40|0;h=c+16|0;l=c+8|0;m=c;n=c+466|0;o=c+391|0;q=c+320|0;s=c+44|0;t=c+280|0;u=c+196|0;v=c+115|0;Do(a,2,b)|0;No(a,790243,m,n,b)|0;Ko(a,782733,g,n,b)|0;if((k[g>>2]|0)<1){w=k[b>>2]|0;r=c;return w|0}x=q+1|0;y=1;while(1){i[o>>0]=0;k[e>>2]=y;Xw(f,772183,e)|0;i[o>>0]=i[773003]|0;i[o+1>>0]=i[773004]|0;i[o+2>>0]=i[773005]|0;i[o+3>>0]=i[773006]|0;i[o+4>>0]=i[773007]|0;i[o+5>>0]=i[773008]|0;qx(o,f)|0;if((Jo(a,o,q,n,b)|0)>0)break;z=i[q>>0]|0;if(!(z<<24>>24!=80?!(z<<24>>24==81?1:(i[x>>0]&-2)<<24>>24==80):0)){z=m;A=k[z+4>>2]|0;if((A|0)<0|(A|0)==0&(k[z>>2]|0)>>>0<1){B=0;C=0}else{z=1;A=0;D=0;E=0;while(1){pl(a,y,z,A,h,l,b)|0;F=h;G=k[F>>2]|0;H=k[F+4>>2]|0;F=(H|0)>(D|0)|(H|0)==(D|0)&G>>>0>E>>>0;I=F?G:E;G=F?H:D;H=z;z=ky(z|0,A|0,1,0)|0;F=m;J=k[F+4>>2]|0;if(!((A|0)<(J|0)|((A|0)==(J|0)?H>>>0<(k[F>>2]|0)>>>0:0))){B=I;C=G;break}else{A=L;D=G;E=I}}}j[s>>1]=39;E=rx(q,40)|0;if(E)i[E>>0]=0;qx(s,q)|0;p[d>>3]=+(B>>>0)+4294967296.0*+(C|0);Xw(t,757110,d)|0;qx(s,t)|0;E=xx(s)|0;D=s+E|0;if(E>>>0<9){E=D;while(1){i[E>>0]=32;i[E+1>>0]=0;A=xx(s)|0;z=s+A|0;if(A>>>0<9)E=z;else{K=z;break}}}else K=D;i[K>>0]=39;i[K+1>>0]=0;Dk(o,s,n,u,b)|0;Ek(a,u,b)|0}if((y|0)<(k[g>>2]|0))y=y+1|0;else{M=15;break}}if((M|0)==15){w=k[b>>2]|0;r=c;return w|0}M=v;y=757042;g=M+68|0;do{i[M>>0]=i[y>>0]|0;M=M+1|0;y=y+1|0}while((M|0)<(g|0));uk(5,v);w=k[b>>2]|0;r=c;return w|0}function yl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;c=r;r=r+2896|0;d=c+8|0;e=c;f=k[b>>2]|0;if((f|0)>0){g=f;r=c;return g|0}h=k[a+4>>2]|0;if((k[a>>2]|0)!=(k[h+68>>2]|0)){g=f;r=c;return g|0}j=h+968|0;l=k[j>>2]|0;m=k[j+4>>2]|0;if((l|0)==0&(m|0)==0){g=f;r=c;return g|0}f=h+120|0;j=ky(k[f>>2]|0,k[f+4>>2]|0,l|0,m|0)|0;m=h+976|0;l=ky(j|0,L|0,k[m>>2]|0,k[m+4>>2]|0)|0;m=L;j=ky(l|0,m|0,2879,0)|0;f=L;n=xy(j|0,f|0,2880,0)|0;o=ky(l|0,m|0,n|0,L|0)|0;n=jy(j|0,f|0,o|0,L|0)|0;o=(k[h+72>>2]|0)==1?32:0;k[e>>2]=0;if(!n){h=ky(l|0,m|0,-1,-1)|0;f=L;au(a,h,f,0,e)|0;fu(a,1,0,d,e)|0;if(!(k[e>>2]|0)){g=k[b>>2]|0;r=c;return g|0}else{p=h;q=f;s=1}}else{au(a,l,m,0,e)|0;fu(a,n,((n|0)<0)<<31>>31,d,e)|0;if(!(k[e>>2]|0)){a:do if((n|0)>0){e=o&255;f=0;while(1){if((i[d+f>>0]|0)!=(e|0)){t=f;break a}h=f+1|0;if((h|0)<(n|0))f=h;else{t=h;break}}}else t=0;while(0);if((t|0)==(n|0)){g=k[b>>2]|0;r=c;return g|0}else{p=l;q=m;s=n}}else{p=l;q=m;s=n}}sy(d|0,o|0,s|0)|0;au(a,p,q,1,b)|0;cu(a,s,((s|0)<0)<<31>>31,d,b)|0;d=k[b>>2]|0;if((d|0)<=0){g=d;r=c;return g|0}uk(5,757117);g=k[b>>2]|0;r=c;return g|0}function zl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0;d=k[c>>2]|0;if((b|0)<1|(d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0){Vk(a,c)|0;h=k[f>>2]|0;f=h+96|0;g=b*80|0;b=ky(k[f>>2]|0,k[f+4>>2]|0,g|0,((g|0)<0)<<31>>31|0)|0;g=L;f=xy(b|0,g|0,2880,0)|0;i=jy(b|0,g|0,f|0,L|0)|0;f=ky(i|0,L|0,2880,0)|0;i=L;g=h+120|0;b=g;j=jy(f|0,i|0,k[b>>2]|0,k[b+4>>2]|0)|0;b=g;k[b>>2]=f;k[b+4>>2]=i;i=(k[h+88>>2]|0)+((k[h+68>>2]|0)+1<<3)|0;h=i;b=ky(j|0,L|0,k[h>>2]|0,k[h+4>>2]|0)|0;h=i;k[h>>2]=b;k[h+4>>2]=L}}else Fk(a,d+1|0,0,c)|0;e=k[c>>2]|0;return e|0}function Al(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;c=r;r=r+16|0;d=c;k[d>>2]=0;e=k[b>>2]|0;if((e|0)>0){f=e;r=c;return f|0}e=k[a>>2]|0;g=a+4|0;h=k[g>>2]|0;if((e|0)==(k[h+68>>2]|0)){i=h;j=e}else{Fk(a,e+1|0,0,b)|0;e=k[g>>2]|0;i=e;j=k[e+68>>2]|0}e=i+96|0;h=(k[i+88>>2]|0)+(j<<3)|0;if((k[e>>2]|0)==(k[h>>2]|0)?(k[e+4>>2]|0)==(k[h+4>>2]|0):0){f=k[b>>2]|0;r=c;return f|0}do Fk(a,(k[a>>2]|0)+2|0,0,d)|0;while((k[d>>2]|0)==0);d=k[g>>2]|0;h=k[d+84>>2]|0;do if((k[d+80>>2]|0)==(h|0)){e=fy(k[d+88>>2]|0,(h<<3)+8008|0)|0;if(e){j=k[g>>2]|0;i=j+84|0;k[i>>2]=(k[i>>2]|0)+1e3;k[j+88>>2]=e;break}k[b>>2]=113;f=113;r=c;return f|0}while(0);if((wl(a,b)|0)<1){h=k[g>>2]|0;d=(k[h+88>>2]|0)+((k[h+80>>2]|0)+1<<3)|0;h=k[d>>2]|0;e=k[d+4>>2]|0;au(a,h,e,1,b)|0;d=k[g>>2]|0;g=d+80|0;j=(k[g>>2]|0)+1|0;k[g>>2]=j;k[d+68>>2]=j;k[a>>2]=j;j=d+112|0;k[j>>2]=h;k[j+4>>2]=e;j=d+96|0;k[j>>2]=h;k[j+4>>2]=e;e=d+120|0;k[e>>2]=-1;k[e+4>>2]=-1;k[d+1076>>2]=k[d+1020>>2]}f=k[b>>2]|0;r=c;return f|0}function Bl(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}Fk(a,b+1+(k[a>>2]|0)|0,c,d)|0;f=k[d>>2]|0;return f|0}function Cl(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;if((k[b>>2]|0)>0){c=0;return c|0}d=k[a>>2]|0;e=a+4|0;f=k[e>>2]|0;if((d|0)==(k[f+68>>2]|0)){g=f+120|0;if(((k[g>>2]|0)==-1?(k[g+4>>2]|0)==-1:0)?(Vk(a,b)|0)>0:0){c=k[b>>2]|0;return c|0}}else Fk(a,d+1|0,0,b)|0;c=(k[(k[e>>2]|0)+1080>>2]|0)!=0&1;return c|0}function Dl(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}if(c)El(a,c,f)|0;do if((d|0)!=0?(k[f>>2]|0)<=0:0){c=k[a>>2]|0;g=a+4|0;i=k[g>>2]|0;if((c|0)==(k[i+68>>2]|0)){j=i+120|0;if(((k[j>>2]|0)==-1?(k[j+4>>2]|0)==-1:0)?(Vk(a,f)|0)>0:0)break}else Fk(a,c+1|0,0,f)|0;c=k[g>>2]|0;if(!(k[c+72>>2]|0)){k[d>>2]=k[c+128>>2];break}if(!(k[c+1080>>2]|0)){k[f>>2]=233;break}else{k[d>>2]=k[c+1100>>2];break}}while(0);if(e)Gl(a,b,e,f)|0;h=k[f>>2]|0;return h|0}function El(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if(((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){e=k[c>>2]|0;return e|0}}else Fk(a,d+1|0,0,c)|0;Do(a,1,c)|0;d=k[f>>2]|0;do if(k[d+72>>2]|0)if(!(k[d+1080>>2]|0)){k[c>>2]=233;break}else{Io(a,31,778879,b,0,c)|0;break}else Io(a,31,779656,b,0,c)|0;while(0);e=k[c>>2]|0;return e|0}function Fl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((d|0)==(k[g+68>>2]|0)){h=g+120|0;if(((k[h>>2]|0)==-1?(k[h+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){e=k[c>>2]|0;return e|0}}else Fk(a,d+1|0,0,c)|0;d=k[f>>2]|0;do if(k[d+72>>2]|0)if(!(k[d+1080>>2]|0)){k[c>>2]=233;break}else{k[b>>2]=k[d+1100>>2];break}else k[b>>2]=k[d+128>>2];while(0);e=k[c>>2]|0;return e|0}function Gl(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}e=k[a>>2]|0;g=a+4|0;h=k[g>>2]|0;if((e|0)==(k[h+68>>2]|0)){i=h+120|0;if(((k[i>>2]|0)==-1?(k[i+4>>2]|0)==-1:0)?(Vk(a,d)|0)>0:0){f=k[d>>2]|0;return f|0}}else Fk(a,e+1|0,0,d)|0;e=k[g>>2]|0;do if(!(k[e+72>>2]|0)){g=k[e+128>>2]|0;if((((g|0)<(b|0)?g:b)|0)>0){a=(g|0)<(b|0)?g:b;g=0;do{k[c+(g<<2)>>2]=k[e+136+(g<<3)>>2];g=g+1|0}while((g|0)!=(a|0))}}else{if(!(k[e+1080>>2]|0)){k[d>>2]=233;break}a=k[e+1100>>2]|0;if((((a|0)<(b|0)?a:b)|0)>0){g=(a|0)<(b|0)?a:b;a=0;do{k[c+(a<<2)>>2]=k[e+1104+(a<<2)>>2];a=a+1|0}while((a|0)!=(g|0))}}while(0);f=k[d>>2]|0;return f|0}function Hl(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}e=k[a>>2]|0;g=a+4|0;h=k[g>>2]|0;if((e|0)==(k[h+68>>2]|0)){i=h+120|0;if(((k[i>>2]|0)==-1?(k[i+4>>2]|0)==-1:0)?(Vk(a,d)|0)>0:0){f=k[d>>2]|0;return f|0}}else Fk(a,e+1|0,0,d)|0;e=k[g>>2]|0;do if(!(k[e+72>>2]|0)){g=k[e+128>>2]|0;if((((g|0)<(b|0)?g:b)|0)>0){a=(g|0)<(b|0)?g:b;g=0;do{i=e+136+(g<<3)|0;h=k[i+4>>2]|0;j=c+(g<<3)|0;k[j>>2]=k[i>>2];k[j+4>>2]=h;g=g+1|0}while((g|0)!=(a|0))}}else{if(!(k[e+1080>>2]|0)){k[d>>2]=233;break}a=k[e+1100>>2]|0;if((((a|0)<(b|0)?a:b)|0)>0){g=(a|0)<(b|0)?a:b;a=0;do{h=k[e+1104+(a<<2)>>2]|0;j=c+(a<<3)|0;k[j>>2]=h;k[j+4>>2]=((h|0)<0)<<31>>31;a=a+1|0}while((a|0)!=(g|0))}}while(0);f=k[d>>2]|0;return f|0}function Il(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}e=a+4|0;g=(k[(k[e>>2]|0)+88>>2]|0)+(b<<3)|0;if((au(a,k[g>>2]|0,k[g+4>>2]|0,0,d)|0)<1?(g=k[e>>2]|0,h=g+68|0,i=k[h>>2]|0,j=g+80|0,l=k[j>>2]|0,m=g+96|0,n=m,o=k[n>>2]|0,p=k[n+4>>2]|0,k[h>>2]=b,k[a>>2]=b,k[j>>2]=(l|0)>(b|0)?l:b,b=g+32|0,g=k[b+4>>2]|0,j=m,k[j>>2]=k[b>>2],k[j+4>>2]=g,(dl(a,c,d)|0)>0):0){c=k[e>>2]|0;k[c+68>>2]=i;k[a>>2]=i;k[c+80>>2]=l;l=c+96|0;k[l>>2]=o;k[l+4>>2]=p}f=k[d>>2]|0;return f|0}function Jl(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0;f=r;r=r+96|0;g=f+20|0;h=f+16|0;j=f+12|0;l=f+8|0;m=f+4|0;n=f;o=k[e>>2]|0;if((o|0)>0){p=o;r=f;return p|0}o=(k[a>>2]|0)+1|0;q=a+4|0;if(!(k[(k[q>>2]|0)+20>>2]|0))s=0;else{t=c+((xx(c)|0)+-1)|0;s=(i[t>>0]|0)!=35&1}k[j>>2]=0;a:do if(!(Fk(a,1,h,j)|0)){t=(b|0)==-1;u=(s|0)==0;v=(d|0)==0;w=0;x=0;y=1;while(1){do if((k[e>>2]|0)>0)z=0;else{A=k[a>>2]|0;B=k[q>>2]|0;if((A|0)==(k[B+68>>2]|0)){C=B+120|0;if(((k[C>>2]|0)==-1?(k[C+4>>2]|0)==-1:0)?(Vk(a,e)|0)>0:0){z=k[e>>2]|0;break}}else Fk(a,A+1|0,0,e)|0;z=(k[(k[q>>2]|0)+1080>>2]|0)!=0&1}while(0);if(!t?(A=k[h>>2]|0,!((A|0)==(b|0)|(A|0)==(((z|0)==0?-1:2)|0))):0){D=w;E=x}else{Do(a,2,e)|0;if((Jo(a,790194,g,0,j)|0)<1){if(!u){A=g+((xx(g)|0)+-1)|0;if((i[A>>0]|0)==35){i[A>>0]=0;F=1}else F=0}else F=x;Wk(g,c,0,l,m);G=k[m>>2]|0;H=F}else{G=w;H=x}if(!((k[j>>2]|0)==0&(G|0)!=0)?(k[j>>2]=0,(Jo(a,757162,g,0,j)|0)<1):0){if(!u){A=g+((xx(g)|0)+-1)|0;if((i[A>>0]|0)==35){i[A>>0]=0;I=1}else I=0}else I=H;Wk(g,c,0,l,m);J=k[m>>2]|0;K=I}else{J=G;K=H}if((k[j>>2]|0)==0&(J|0)!=0){if(v){L=K;M=36;break}if((Ko(a,767107,n,0,j)|0)>0){k[n>>2]=1;N=1}else N=k[n>>2]|0;if((N|0)==(d|0)){O=K;M=33;break}else{D=J;E=K}}else{D=J;E=K}}y=y+1|0;k[j>>2]=0;if(Fk(a,y,h,j)|0)break a;else{w=D;x=E}}if((M|0)==33){if(O)k[(k[q>>2]|0)+20>>2]=0;p=k[e>>2]|0;r=f;return p|0}else if((M|0)==36){if(L)k[(k[q>>2]|0)+20>>2]=0;p=k[e>>2]|0;r=f;return p|0}}while(0);Fk(a,o,0,e)|0;k[e>>2]=301;p=301;r=f;return p|0}function Kl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=r;r=r+16|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}f=k[a>>2]|0;h=f+1|0;k[b>>2]=f;f=(k[a+4>>2]|0)+120|0;if((k[f>>2]|0)==-1?(k[f+4>>2]|0)==-1:0){g=k[c>>2]|0;r=d;return g|0}k[e>>2]=0;if((Fk(a,h,0,e)|0)<1){f=h;do{k[b>>2]=f;f=f+1|0}while((Fk(a,f,0,e)|0)<1)}Fk(a,h,0,c)|0;g=k[c>>2]|0;r=d;return g|0}function Ll(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0;b=r;r=r+32|0;c=b;i[c>>0]=0;zx(c,a,8)|0;d=c+(xx(c)|0)|0;e=773880;f=d+9|0;do{i[d>>0]=i[e>>0]|0;d=d+1|0;e=e+1|0}while((d|0)<(f|0));e=xx(c)|0;if(e){d=0;do{f=c+d|0;g=(nv(i[f>>0]|0)|0)&255;i[f>>0]=g;d=d+1|0}while((d|0)!=(e|0))}e=c+1|0;d=c+5|0;g=i[c>>0]|0;a:do switch(g<<24>>24){case 90:{f=i[e>>0]|0;b:do if(f<<24>>24<73){if(f<<24>>24==67)if(!(Ax(e,757178,7)|0)){h=20;break a}else{j=0;k=33;break}if(f<<24>>24>=66)if(f<<24>>24<=66?(Ax(e,757200,7)|0)==0:0){h=20;break a}else k=31;else{j=0;k=33}}else{if(f<<24>>24<=73)if(!(Ax(e,757170,7)|0)){h=20;break a}else{j=0;k=33;break}if(f<<24>>24>=78){do if(f<<24>>24>78)if(f<<24>>24<86){if(f<<24>>24!=84)break;if(Ax(e,757195,4)|0){j=1;k=33;break b}if(((i[d>>0]|0)+-48&255)<10){h=20;break a}else{k=271;break a}}else{if(f<<24>>24>86){k=23;break}if(Ax(e,757191,3)|0){k=23;break}if(((i[c+4>>0]|0)+-48&255)<10){h=20;break a}else{k=271;break a}}else if(!(Ax(e,757186,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=20;break a}else{k=271;break a}else k=23;while(0);if((k|0)==23?f<<24>>24<=78:0){if(Ax(e,782713,5)|0){k=31;break}l=i[c+6>>0]|0;if(l<<24>>24==32|(l+-48&255)<10){h=20;break a}else{k=271;break a}}if(f<<24>>24>=83){if(f<<24>>24<=83)if(!(Ax(e,757208,7)|0)){h=20;break a}else{j=0;k=33;break}if(f<<24>>24==90?(Ax(e,757216,7)|0)==0:0){h=20;break a}else{j=0;k=33}}else{j=0;k=33}}else{j=0;k=33}}while(0);if((k|0)==31)if(f<<24>>24==66){if(!(Ax(e,757224,7)|0)){h=20;break a}if(!(Ax(e,782824,7)|0)){h=20;break a}else k=41}else{j=0;k=33}c:do if((k|0)==33){if(f<<24>>24>=83){if(f<<24>>24<=83?(Ax(e,757232,7)|0)==0:0){h=20;break a}if(j?(Ax(e,757240,7)|0)==0:0){h=20;break a}}switch(f<<24>>24){case 80:{k=42;break c;break}case 71:{k=43;break c;break}case 81:{k=44;break c;break}case 68:{k=45;break c;break}case 69:break;default:{k=271;break a}}if(!(Ax(e,757248,7)|0)){h=20;break a}else k=41}while(0);if((k|0)==41)switch(f<<24>>24){case 80:{k=42;break}case 71:{k=43;break}case 81:{k=44;break}case 68:{k=45;break}default:{k=271;break a}}if((k|0)==42)if(!(Ax(e,757256,7)|0)){h=20;break a}else{k=271;break a}else if((k|0)==43)if(!(Ax(e,757264,7)|0)){h=20;break a}else{k=271;break a}else if((k|0)==44)if(!(Ax(e,757272,7)|0)){h=20;break a}else{k=271;break a}else if((k|0)==45)if(!(Ax(e,757280,7)|0)){h=20;break a}else{k=271;break a}break}case 32:{h=130;break}case 66:{switch(i[e>>0]|0){case 73:{if(!(Ax(e,757288,7)|0)){h=10;break a}else{k=271;break a}break}case 76:{if(!(Ax(e,757296,7)|0)){h=10;break a}if(!(Ax(e,757304,7)|0)){h=40;break a}else{k=271;break a}break}case 83:{if(!(Ax(e,757208,7)|0)){h=30;break a}else{k=271;break a}break}case 90:{if(!(Ax(e,757216,7)|0)){h=30;break a}else{k=271;break a}break}case 85:{if(!(Ax(e,757312,7)|0)){h=70;break a}else{k=271;break a}break}default:{k=271;break a}}break}case 67:{l=i[e>>0]|0;if(l<<24>>24<79){switch(l<<24>>24){case 72:{if(!(Ax(e,757567,7)|0)){h=100;break a}else{k=271;break a}break}case 68:break;default:{k=271;break a}}if(!(Ax(e,757613,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}else if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}}d:do if(l<<24>>24>79){if(l<<24>>24<84){switch(l<<24>>24){case 83:{k=83;break d;break}case 82:break;default:{k=271;break a}}if(!(Ax(e,757588,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(Ax(e,757593,4)|0)break;if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}}if(l<<24>>24<=84){if(Ax(e,755193,4)|0){k=78;break}if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}}if(l<<24>>24<=85?(Ax(e,757583,4)|0)==0:0)if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}else k=78}else{if(Ax(e,757320,6)|0)if(!(Ax(e,757575,7)|0)){h=140;break a}else{k=78;break}if((i[a>>0]|0)==67){if(!(Ax(a,757327,47)|0)){h=10;break a}if(!(Ax(a,757375,47)|0)){h=10;break a}if(!(Ax(a,757423,47)|0)){h=10;break a}if(!(Ax(a,757471,47)|0)){h=10;break a}if(!(Ax(a,757519,47)|0)){h=10;break a}}h=(i[c+7>>0]|0)==32?130:150;break a}while(0);if((k|0)==78)switch(l<<24>>24){case 82:break;case 83:{k=83;break}default:{k=271;break a}}if((k|0)==83){if(Ax(e,757608,4)|0){k=271;break a}if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}}if(!(Ax(e,757598,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if((Ax(e,757603,4)|0)==0?((i[d>>0]|0)+-48&255)<10:0)h=110;else k=271;break}case 68:{if((i[e>>0]|0)==65)if(Ax(e,757618,7)|0)if((Ax(e,757626,7)|0)!=0?(Ax(e,757634,7)|0)!=0:0)if(!(Ax(e,757642,7)|0))h=120;else k=271;else h=60;else h=100;else k=271;break}case 69:{f=i[e>>0]|0;if(f<<24>>24<88)switch(f<<24>>24){case 78:{if(!(Ax(e,757658,7)|0)){h=10;break a}else{k=271;break a}break}case 81:{if(!(Ax(e,757690,6)|0)){h=120;break a}if(Ax(e,757697,3)|0){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=120;break a}else{k=271;break a}break}case 80:{if(!(Ax(e,757701,7)|0)){h=120;break a}else{k=271;break a}break}default:{k=271;break a}}if(f<<24>>24<=88)if(Ax(e,757650,7)|0)if(Ax(e,757666,7)|0){if(!(Ax(e,757674,7)|0)){h=90;break a}if(!(Ax(e,757682,7)|0)){h=90;break a}else{k=271;break a}}else{if((i[a>>0]|0)!=69){h=90;break a}f=(Ax(a,779627,28)|0)==0;h=f?20:90;break a}else h=10;else k=271;break}case 71:{switch(i[e>>0]|0){case 67:{if(!(Ax(e,757709,7)|0)){h=10;break a}else{k=271;break a}break}case 82:{if(!(Ax(e,757717,7)|0)){h=10;break a}else{k=271;break a}break}default:{k=271;break a}}break}case 72:{switch(i[e>>0]|0){case 68:break;case 73:{if(Ax(e,757749,6)|0){k=271;break a}h=(i[c+7>>0]|0)==32?130:150;break a;break}default:{k=271;break a}}if(((Ax(e,757725,7)|0)!=0?(Ax(e,757733,7)|0)!=0:0)?(Ax(e,757741,7)|0)!=0:0)k=271;else h=90;break}case 76:{switch(i[e>>0]|0){case 79:{if(!(Ax(e,757756,6)|0)){h=110;break a}if(Ax(e,757770,3)|0){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 65:{if(!(Ax(e,757763,6)|0)){h=110;break a}if(Ax(e,757774,3)|0){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}default:{k=271;break a}}break}case 77:{if((i[e>>0]|0)==74)if(Ax(e,757778,7)|0)if((Ax(e,757786,4)|0)==0?((i[d>>0]|0)+-48&255)<10:0)h=120;else k=271;else h=120;else k=271;break}case 78:{if(((i[e>>0]|0)==65?(Ax(e,757791,4)|0)==0:0)?(f=i[d>>0]|0,f<<24>>24==32|(f+-48&255)<10):0)h=10;else k=271;break}case 80:{switch(i[e>>0]|0){case 67:{if(!(Ax(e,757709,7)|0)){h=10;break a}if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 83:{if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 86:{if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}default:{k=271;break a}}break}case 82:{if((i[e>>0]|0)==65)if((Ax(e,757796,7)|0)!=0?(Ax(e,757804,6)|0)!=0:0)if((Ax(e,757811,3)|0)==0?((i[c+4>>0]|0)+-48&255)<10:0)h=120;else k=271;else h=120;else k=271;break}case 83:{if((i[e>>0]|0)==73?(Ax(e,757815,7)|0)==0:0)h=10;else k=271;break}case 84:{f=i[e>>0]|0;do if(f<<24>>24<84){if(f<<24>>24==70){if(!(Ax(e,755198,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=10;break a}else{k=271;break a}}else{if(f<<24>>24<66){m=0;k=179;break}if(f<<24>>24<=66){if(Ax(e,755203,4)|0){m=0;k=179;break}if(((i[d>>0]|0)+-48&255)<10){h=10;break a}else{k=271;break a}}if(f<<24>>24<70){if(f<<24>>24!=68){k=188;break}if(!(Ax(e,755362,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=50;break a}else{k=271;break a}if(Ax(e,757839,4)|0){n=1;k=172;break}if(((i[d>>0]|0)+-48&255)<10){h=80;break a}else{k=271;break a}}if(f<<24>>24>70){if(f<<24>>24<72){m=0;k=179;break}if(f<<24>>24<=72)if(!(Ax(e,757831,7)|0)){h=10;break a}else{m=0;k=179;break}if(f<<24>>24<78){m=0;k=179;break}if(f<<24>>24>78)if(f<<24>>24==85){k=168;break}else{n=0;k=172;break}if(Ax(e,760184,4)|0){m=0;k=179;break}if(((i[d>>0]|0)+-48&255)<10){h=40;break a}else{k=271;break a}}}if(!(Ax(e,757823,7)|0)){h=10;break a}else{m=0;k=179}}else{if(f<<24>>24>84)if(f<<24>>24>85){n=0;k=172;break}else{k=168;break}if(!(Ax(e,755193,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=10;break a}else{k=271;break a}else k=219}while(0);if((k|0)==168)if(!(Ax(e,757583,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=70;break a}else{k=271;break a}else{n=0;k=172}if((k|0)==172)if(f<<24>>24>=83){if(f<<24>>24>83){if(f<<24>>24==90?(Ax(e,755299,4)|0)==0:0)if(((i[d>>0]|0)+-48&255)<10){h=30;break a}else{k=271;break a}}else if(!(Ax(e,755252,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=30;break a}else{k=271;break a}if(n)k=186;else k=219}else{m=n;k=179}if((k|0)==179){if(f<<24>>24==76){if(!(Ax(e,757844,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=60;break a}else{k=271;break a}if(!(Ax(e,757849,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=60;break a}else{k=271;break a}}if(m)k=186;else k=188}if((k|0)==186)if(!(Ax(e,757854,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=60;break a}else{k=271;break a}else k=188;e:do if((k|0)==188){if(f<<24>>24>=68){if(f<<24>>24>68){k=219;break}if(Ax(e,757859,4)|0){k=219;break}if(((i[d>>0]|0)+-48&255)<10){h=60;break a}else{k=271;break a}}switch(f<<24>>24){case 80:{k=223;break e;break}case 86:{k=224;break e;break}case 83:{k=225;break e;break}case 67:break;default:{k=271;break a}}if(!(Ax(e,757864,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757869,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757873,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757878,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757882,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757887,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757891,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757896,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757900,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757905,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757910,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757914,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757918,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}else k=219}while(0);f:do if((k|0)==219){switch(f<<24>>24){case 80:{k=223;break f;break}case 86:{k=224;break f;break}case 83:{k=225;break f;break}case 87:break;case 67:{if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}default:{k=271;break a}}if(Ax(e,757922,3)|0){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}}while(0);if((k|0)==223)if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}else if((k|0)==224)if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}else if((k|0)==225)if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 88:{if((i[e>>0]|0)==84?(Ax(e,757240,7)|0)==0:0)h=10;else k=271;break}case 87:{if((i[e>>0]|0)==67)if((Ax(e,757926,6)|0)!=0?(Ax(e,757933,6)|0)!=0:0){if(!(Ax(e,757940,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if((Ax(e,757944,3)|0)==0?((i[c+4>>0]|0)+-48&255)<10:0)h=110;else k=271}else h=110;else k=271;break}default:if((g+-48&255)<10){f=i[e>>0]|0;switch(f<<24>>24){case 67:{if(!(Ax(e,757864,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757869,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757873,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757878,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757882,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757887,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757891,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757896,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757900,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757905,4)|0))if(((i[d>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757910,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(!(Ax(e,757914,3)|0))if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}if(Ax(e,757918,3)|0){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 86:{if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 83:{if(((i[c+2>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}default:{if((f+-48&255)>=10){k=271;break a}switch(i[c+2>>0]|0){case 80:{if((i[c+3>>0]|0)!=67){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}case 67:{if((i[c+3>>0]|0)!=68){k=271;break a}if(((i[c+4>>0]|0)+-48&255)<10){h=110;break a}else{k=271;break a}break}default:{k=271;break a}}}}}else k=271}while(0);if((k|0)==271)h=150;r=b;return h|0}function Ml(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}a:do switch(i[a>>0]|0){case 0:{k[c>>2]=204;e=204;return e|0}case 39:{i[b>>0]=67;break}case 70:case 84:{i[b>>0]=76;break}case 40:{i[b>>0]=88;break}default:{if(rx(a,46)|0){i[b>>0]=70;break a}if((rx(a,69)|0)==0?(rx(a,68)|0)==0:0){i[b>>0]=73;break a}i[b>>0]=70}}while(0);e=k[c>>2]|0;return e|0}function Nl(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;h=r;r=r+96|0;j=h;l=h+4|0;a:do if((k[g>>2]|0)>0)m=11;else switch(i[a>>0]|0){case 0:{k[g>>2]=204;m=11;break a;break}case 39:{i[b>>0]=67;m=22;break a;break}case 70:case 84:{i[b>>0]=76;if((k[g>>2]|0)>0)break a;k[d>>2]=(i[a>>0]|0)==84&1;break a;break}case 40:{i[b>>0]=88;m=22;break a;break}default:{if(rx(a,46)|0){i[b>>0]=70;m=19;break a}if((rx(a,69)|0)==0?(rx(a,68)|0)==0:0){i[b>>0]=73;if((k[g>>2]|0)>0){n=l;o=j;m=18;break a}p=rv()|0;k[p>>2]=0;k[c>>2]=0;q=jx(a,j,10)|0;k[c>>2]=q;switch(i[k[j>>2]>>0]|0){case 32:case 0:break;default:k[g>>2]=407}if((k[p>>2]|0)!=34){n=l;o=j;m=18;break a}q=l;s=754043;t=q+54|0;do{i[q>>0]=i[s>>0]|0;q=q+1|0;s=s+1|0}while((q|0)<(t|0));zx(l,a,25)|0;uk(5,l);k[g>>2]=412;k[p>>2]=0;n=l;o=j;m=18;break a}i[b>>0]=70;m=19;break a}}while(0);b:do if((m|0)==11){switch(i[b>>0]|0){case 70:{m=19;break b;break}case 76:{break b;break}case 73:break;default:{m=22;break b}}n=l;o=j;m=18}while(0);if((m|0)!=18)if((m|0)==19)ll(a,f,g)|0;else if((m|0)==22)fl(a,e,g)|0;r=h;return k[g>>2]|0}function Ol(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}k[b>>2]=(i[a>>0]|0)==84&1;e=k[c>>2]|0;return e|0}function Pl(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;h=r;r=r+96|0;j=h;l=h+4|0;a:do if((k[g>>2]|0)>0)m=11;else switch(i[a>>0]|0){case 0:{k[g>>2]=204;m=11;break a;break}case 39:{i[b>>0]=67;m=22;break a;break}case 70:case 84:{i[b>>0]=76;if((k[g>>2]|0)>0)break a;k[d>>2]=(i[a>>0]|0)==84&1;break a;break}case 40:{i[b>>0]=88;m=22;break a;break}default:{if(rx(a,46)|0){i[b>>0]=70;m=19;break a}if((rx(a,69)|0)==0?(rx(a,68)|0)==0:0){i[b>>0]=73;if((k[g>>2]|0)>0){n=l;o=j;m=18;break a}p=rv()|0;k[p>>2]=0;q=c;k[q>>2]=0;k[q+4>>2]=0;q=ix(a,j,10)|0;s=c;k[s>>2]=q;k[s+4>>2]=L;switch(i[k[j>>2]>>0]|0){case 32:case 0:break;default:k[g>>2]=407}if((k[p>>2]|0)!=34){n=l;o=j;m=18;break a}s=l;q=755304;t=s+58|0;do{i[s>>0]=i[q>>0]|0;s=s+1|0;q=q+1|0}while((s|0)<(t|0));zx(l,a,25)|0;uk(5,l);k[g>>2]=412;k[p>>2]=0;n=l;o=j;m=18;break a}i[b>>0]=70;m=19;break a}}while(0);b:do if((m|0)==11){switch(i[b>>0]|0){case 70:{m=19;break b;break}case 76:{break b;break}case 73:break;default:{m=22;break b}}n=l;o=j;m=18}while(0);if((m|0)!=18)if((m|0)==19)ll(a,f,g)|0;else if((m|0)==22)fl(a,e,g)|0;r=h;return k[g>>2]|0}function Ql(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0.0,o=0;d=r;r=r+176|0;e=d+93|0;f=d+94|0;g=d+12|0;h=d+8|0;j=d;l=k[c>>2]|0;if((l|0)>0){m=l;r=d;return m|0}if(!(i[a>>0]|0)){k[c>>2]=204;m=204;r=d;return m|0}Nl(a,e,b,h,f,j,c)|0;a:do switch(i[e>>0]|0){case 88:{k[c>>2]=403;break}case 67:{if((ll(f,j,c)|0)<1){n=+p[j>>3];if(n>2147483647.0|n<-2147483648.0){k[c>>2]=412;break a}else{k[b>>2]=~~n;o=14;break a}}else o=14;break}case 70:{n=+p[j>>3];if(n>2147483647.0|n<-2147483648.0){k[c>>2]=412;break a}else{k[b>>2]=~~n;o=14;break a}break}case 76:{k[b>>2]=k[h>>2];o=14;break}default:o=14}while(0);if((o|0)==14?(o=k[c>>2]|0,(o|0)<=0):0){m=o;r=d;return m|0}k[b>>2]=0;b=g;o=757948;h=b+49|0;do{i[b>>0]=i[o>>0]|0;b=b+1|0;o=o+1|0}while((b|0)<(h|0));zx(g,a,30)|0;uk(5,g);m=k[c>>2]|0;r=d;return m|0}function Rl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0.0,o=0,q=0;d=r;r=r+176|0;e=d+93|0;f=d+94|0;g=d+12|0;h=d+8|0;j=d;l=k[c>>2]|0;if((l|0)>0){m=l;r=d;return m|0}if(!(i[a>>0]|0)){k[c>>2]=204;m=204;r=d;return m|0}Pl(a,e,b,h,f,j,c)|0;a:do switch(i[e>>0]|0){case 88:{k[c>>2]=403;break}case 67:{if((ll(f,j,c)|0)<1){n=+p[j>>3];if(n>9223372036854775808.0|n<-9223372036854775808.0){k[c>>2]=412;break a}else{l=+W(n)>=1.0?(n>0.0?~~+ia(+V(n/4294967296.0),4294967295.0)>>>0:~~+ga((n-+(~~n>>>0))/4294967296.0)>>>0):0;o=b;k[o>>2]=~~n>>>0;k[o+4>>2]=l;q=14;break a}}else q=14;break}case 70:{n=+p[j>>3];if(n>9223372036854775808.0|n<-9223372036854775808.0){k[c>>2]=412;break a}else{l=+W(n)>=1.0?(n>0.0?~~+ia(+V(n/4294967296.0),4294967295.0)>>>0:~~+ga((n-+(~~n>>>0))/4294967296.0)>>>0):0;o=b;k[o>>2]=~~n>>>0;k[o+4>>2]=l;q=14;break a}break}case 76:{l=k[h>>2]|0;o=b;k[o>>2]=l;k[o+4>>2]=((l|0)<0)<<31>>31;q=14;break}default:q=14}while(0);if((q|0)==14?(q=k[c>>2]|0,(q|0)<=0):0){m=q;r=d;return m|0}q=b;k[q>>2]=0;k[q+4>>2]=0;q=g;b=757997;h=q+53|0;do{i[q>>0]=i[b>>0]|0;q=q+1|0;b=b+1|0}while((q|0)<(h|0));zx(g,a,30)|0;uk(5,g);m=k[c>>2]|0;r=d;return m|0}function Sl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0;d=r;r=r+176|0;e=d+93|0;f=d+12|0;g=d+8|0;h=d;j=k[c>>2]|0;if((j|0)>0){l=j;r=d;return l|0}if(!(i[a>>0]|0)){k[c>>2]=204;l=204;r=d;return l|0}Nl(a,e,g,b,d+94|0,h,c)|0;j=i[e>>0]|0;switch(j<<24>>24){case 88:case 67:{k[c>>2]=404;break}default:if((k[c>>2]|0)<=0){a:do switch(j<<24>>24){case 73:{if(!(k[g>>2]|0)){k[b>>2]=0;break a}else{k[b>>2]=1;break a}break}case 70:{if(+p[h>>3]!=0.0){k[b>>2]=1;break a}else{k[b>>2]=0;break a}break}default:{}}while(0);l=k[c>>2]|0;r=d;return l|0}}k[b>>2]=0;b=f;h=758050;g=b+48|0;do{i[b>>0]=i[h>>0]|0;b=b+1|0;h=h+1|0}while((b|0)<(g|0));zx(f,a,30)|0;uk(5,f);l=k[c>>2]|0;r=d;return l|0}function Tl(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0;d=r;r=r+176|0;e=d+81|0;f=d;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=i[a>>0]|0;switch(g<<24>>24){case 0:{k[c>>2]=204;h=204;r=d;return h|0}case 40:{k[c>>2]=405;break}case 39:{fl(a,e,c)|0;Pk(e,b,c)|0;j=8;break}case 70:case 84:{o[b>>2]=+(g<<24>>24==84|0);j=8;break}default:{Pk(a,b,c)|0;j=8}}if((j|0)==8?(j=k[c>>2]|0,(j|0)<=0):0){h=j;r=d;return h|0}o[b>>2]=0.0;b=f;j=758098;g=b+46|0;do{i[b>>0]=i[j>>0]|0;b=b+1|0;j=j+1|0}while((b|0)<(g|0));zx(f,a,30)|0;uk(5,f);h=k[c>>2]|0;r=d;return h|0}function Ul(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0;d=r;r=r+176|0;e=d+81|0;f=d;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=i[a>>0]|0;switch(g<<24>>24){case 0:{k[c>>2]=204;h=204;r=d;return h|0}case 40:{k[c>>2]=406;break}case 39:{fl(a,e,c)|0;ll(e,b,c)|0;j=8;break}case 70:case 84:{p[b>>3]=+(g<<24>>24==84|0);j=8;break}default:{ll(a,b,c)|0;j=8}}if((j|0)==8?(j=k[c>>2]|0,(j|0)<=0):0){h=j;r=d;return h|0}p[b>>3]=0.0;b=f;j=758144;g=b+47|0;do{i[b>>0]=i[j>>0]|0;b=b+1|0;j=j+1|0}while((b|0)<(g|0));zx(f,a,30)|0;uk(5,f);h=k[c>>2]|0;r=d;return h|0}function Vl(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0;j=r;r=r+800|0;l=j;m=j+792|0;n=k[i>>2]|0;if((d|0)==0&(e|0)==0|(n|0)>0){o=n;r=j;return o|0}Fl(a,m,i)|0;n=k[m>>2]|0;if((n|0)>0){m=0;do{p=k[c+(m<<2)>>2]|0;q=l+(m<<3)|0;k[q>>2]=p;k[q+4>>2]=((p|0)<0)<<31>>31;m=m+1|0}while((m|0)<(n|0))}Wl(a,b,l,d,e,f,g,h,i)|0;o=k[i>>2]|0;r=j;return o|0}function Wl(a,b,c,d,e,f,g,h,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;var m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0;m=r;r=r+192|0;n=m+180|0;q=m+184|0;s=m+72|0;t=m;u=m+144|0;v=t;w=98848;x=v+72|0;do{k[v>>2]=k[w>>2];v=v+4|0;w=w+4|0}while((v|0)<(x|0));v=u;w=121436;x=v+36|0;do{k[v>>2]=k[w>>2];v=v+4|0;w=w+4|0}while((v|0)<(x|0));w=k[l>>2]|0;if((d|0)==0&(e|0)==0|(w|0)>0){y=w;r=m;return y|0}Fl(a,n,l)|0;Hl(a,9,s,l)|0;w=k[n>>2]|0;v=s;x=k[v>>2]|0;z=k[v+4>>2]|0;if((w|0)==0|(x|0)==0&(z|0)==0){k[l>>2]=320;y=320;r=m;return y|0}if((w|0)>0){py(t|0,c|0,((w|0)>1?w<<3:8)|0)|0;v=c;A=ky(k[v>>2]|0,k[v+4>>2]|0,-1,-1)|0;v=L;if((w|0)>1){B=x;x=z;z=1;C=0;D=1;E=A;F=v;while(1){z=yy(B|0,x|0,z|0,C|0)|0;C=L;G=s+(D<<3)|0;B=k[G>>2]|0;x=k[G+4>>2]|0;G=c+(D<<3)|0;H=ky(k[G>>2]|0,k[G+4>>2]|0,-1,-1)|0;G=yy(H|0,L|0,z|0,C|0)|0;H=ky(G|0,L|0,E|0,F|0)|0;G=L;D=D+1|0;if((D|0)>=(w|0)){I=H;J=G;break}else{E=H;F=G}}}else{I=A;J=v}}else{I=0;J=0}v=ky(I|0,J|0,1,0)|0;J=L;if(Cl(a,l)|0){if(((k[n>>2]&-2|0)==2?(n=c,(k[n>>2]|0)==1&(k[n+4>>2]|0)==0):0)?(n=s,I=k[n>>2]|0,A=k[n+4>>2]|0,n=xy(d|0,e|0,I|0,A|0)|0,(n|0)==0&(L|0)==0):0){n=t;k[n>>2]=I;k[n+4>>2]=A;n=c+8|0;F=k[n>>2]|0;E=k[n+4>>2]|0;n=wy(d|0,e|0,I|0,A|0)|0;A=L;I=ky(F|0,E|0,-1,-1)|0;E=ky(I|0,L|0,n|0,A|0)|0;A=L;n=t+8|0;I=n;k[I>>2]=E;k[I+4>>2]=A;I=s+8|0;s=k[I>>2]|0;F=k[I+4>>2]|0;if((A|0)>(F|0)|(A|0)==(F|0)&E>>>0>s>>>0){I=t+16|0;w=I;D=E;E=A;A=k[w>>2]|0;C=k[w+4>>2]|0;while(1){w=jy(D|0,E|0,s|0,F|0)|0;z=L;x=ky(A|0,C|0,1,0)|0;B=L;if((z|0)>(F|0)|(z|0)==(F|0)&w>>>0>s>>>0){D=w;E=z;A=x;C=B}else{K=w;M=z;N=x;O=B;break}}C=n;k[C>>2]=K;k[C+4>>2]=M;M=I;k[M>>2]=N;k[M+4>>2]=O}mt(a,b,c,t,u,1,f,g,0,h,l)|0}else qt(a,b,v,J,d,e,1,f,g,0,h,l)|0;y=k[l>>2]|0;r=m;return y|0}a:do switch(b|0){case 11:{if(!f){bm(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{bm(a,2,1,0,v,J,d,e,1,1,i[f>>0]|0,g,q,h,l)|0;break a}break}case 12:{if(!f){On(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{On(a,2,1,0,v,J,d,e,1,1,i[f>>0]|0,g,q,h,l)|0;break a}break}case 20:{if(!f){ho(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{ho(a,2,1,0,v,J,d,e,1,1,j[f>>1]|0,g,q,h,l)|0;break a}break}case 21:{if(!f){Nm(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{Nm(a,2,1,0,v,J,d,e,1,1,j[f>>1]|0,g,q,h,l)|0;break a}break}case 30:{if(!f){Zn(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{Zn(a,2,1,0,v,J,d,e,1,1,k[f>>2]|0,g,q,h,l)|0;break a}break}case 31:{if(!f){wn(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{wn(a,2,1,0,v,J,d,e,1,1,k[f>>2]|0,g,q,h,l)|0;break a}break}case 40:{if(!f){so(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{so(a,2,1,0,v,J,d,e,1,1,k[f>>2]|0,g,q,h,l)|0;break a}break}case 41:{if(!f){Ym(a,2,1,0,v,J,d,e,1,1,0,g,q,h,l)|0;break a}else{Ym(a,2,1,0,v,J,d,e,1,1,k[f>>2]|0,g,q,h,l)|0;break a}break}case 81:{if(!f){kn(a,2,1,0,v,J,d,e,1,1,0,0,g,q,h,l)|0;break a}else{u=f;kn(a,2,1,0,v,J,d,e,1,1,k[u>>2]|0,k[u+4>>2]|0,g,q,h,l)|0;break a}break}case 42:{if(!f){Bm(a,2,1,0,v,J,d,e,1,1,0.0,g,q,h,l)|0;break a}else{Bm(a,2,1,0,v,J,d,e,1,1,+o[f>>2],g,q,h,l)|0;break a}break}case 82:{if(!f){nm(a,2,1,0,v,J,d,e,1,1,0.0,g,q,h,l)|0;break a}else{nm(a,2,1,0,v,J,d,e,1,1,+p[f>>3],g,q,h,l)|0;break a}break}default:k[l>>2]=410}while(0);y=k[l>>2]|0;r=m;return y|0}function Xl(a,b,c,d,e,f,g,h,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;var m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;m=r;r=r+848|0;n=m;q=m+832|0;s=m+792|0;t=m+796|0;u=k[l>>2]|0;if((u|0)>0){v=u;r=m;return v|0}Fl(a,s,l)|0;Gl(a,9,t,l)|0;if(!(Cl(a,l)|0)){u=k[s>>2]|0;a:do if((u|0)>0){w=1;x=0;y=0;while(1){if((k[e+(y<<2)>>2]|0)!=1){z=w;A=x;B=y;break a}if((k[c+(y<<2)>>2]|0)!=1){z=w;A=x;B=y;break a}C=k[d+(y<<2)>>2]|0;if((C|0)!=(k[t+(y<<2)>>2]|0)){z=w;A=x;B=y;break a}D=yy(C|0,((C|0)<0)<<31>>31|0,w|0,x|0)|0;C=L;E=y+1|0;if((E|0)<(u|0)){w=D;x=C;y=E}else{z=D;A=C;B=E;break}}}else{z=1;A=0;B=0}while(0);if((B|0)==(u|0)){u=k[l>>2]|0;if((z|0)==0&(A|0)==0|(u|0)>0)F=u;else{Fl(a,q,l)|0;u=k[q>>2]|0;if((u|0)>0){q=0;do{B=k[c+(q<<2)>>2]|0;y=n+(q<<3)|0;k[y>>2]=B;k[y+4>>2]=((B|0)<0)<<31>>31;q=q+1|0}while((q|0)!=(u|0))}Wl(a,b,n,z,A,f,g,h,l)|0;F=k[l>>2]|0}v=F;r=m;return v|0}}b:do switch(b|0){case 11:{F=k[s>>2]|0;if(!f){dm(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{dm(a,1,F,t,c,d,e,i[f>>0]|0,g,h,l)|0;break b}break}case 12:{F=k[s>>2]|0;if(!f){Qn(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{Qn(a,1,F,t,c,d,e,i[f>>0]|0,g,h,l)|0;break b}break}case 20:{F=k[s>>2]|0;if(!f){jo(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{jo(a,1,F,t,c,d,e,j[f>>1]|0,g,h,l)|0;break b}break}case 21:{F=k[s>>2]|0;if(!f){Pm(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{Pm(a,1,F,t,c,d,e,j[f>>1]|0,g,h,l)|0;break b}break}case 30:{F=k[s>>2]|0;if(!f){$n(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{$n(a,1,F,t,c,d,e,k[f>>2]|0,g,h,l)|0;break b}break}case 31:{F=k[s>>2]|0;if(!f){yn(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{yn(a,1,F,t,c,d,e,k[f>>2]|0,g,h,l)|0;break b}break}case 40:{F=k[s>>2]|0;if(!f){uo(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{uo(a,1,F,t,c,d,e,k[f>>2]|0,g,h,l)|0;break b}break}case 41:{F=k[s>>2]|0;if(!f){_m(a,1,F,t,c,d,e,0,g,h,l)|0;break b}else{_m(a,1,F,t,c,d,e,k[f>>2]|0,g,h,l)|0;break b}break}case 81:{F=k[s>>2]|0;if(!f){mn(a,1,F,t,c,d,e,0,0,g,h,l)|0;break b}else{A=f;mn(a,1,F,t,c,d,e,k[A>>2]|0,k[A+4>>2]|0,g,h,l)|0;break b}break}case 42:{A=k[s>>2]|0;if(!f){Dm(a,1,A,t,c,d,e,0.0,g,h,l)|0;break b}else{Dm(a,1,A,t,c,d,e,+o[f>>2],g,h,l)|0;break b}break}case 82:{A=k[s>>2]|0;if(!f){pm(a,1,A,t,c,d,e,0.0,g,h,l)|0;break b}else{pm(a,1,A,t,c,d,e,+p[f>>3],g,h,l)|0;break b}break}default:k[l>>2]=410}while(0);v=k[l>>2]|0;r=m;return v|0}function Yl(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0;n=k[m>>2]|0;if((e|0)==0&(f|0)==0|(n|0)>0){q=n;return q|0}a:do switch(b|0){case 11:{if(!g){am(a,1,c,d,e,f,0,h,l,m)|0;break a}else{am(a,1,c,d,e,f,i[g>>0]|0,h,l,m)|0;break a}break}case 12:{if(!g){Nn(a,1,c,d,e,f,0,h,l,m)|0;break a}else{Nn(a,1,c,d,e,f,i[g>>0]|0,h,l,m)|0;break a}break}case 20:{if(!g){go(a,1,c,d,e,f,0,h,l,m)|0;break a}else{go(a,1,c,d,e,f,j[g>>1]|0,h,l,m)|0;break a}break}case 21:{if(!g){Mm(a,1,c,d,e,f,0,h,l,m)|0;break a}else{Mm(a,1,c,d,e,f,j[g>>1]|0,h,l,m)|0;break a}break}case 30:{if(!g){Yn(a,1,c,d,e,f,0,h,l,m)|0;break a}else{Yn(a,1,c,d,e,f,k[g>>2]|0,h,l,m)|0;break a}break}case 31:{if(!g){vn(a,1,c,d,e,f,0,h,l,m)|0;break a}else{vn(a,1,c,d,e,f,k[g>>2]|0,h,l,m)|0;break a}break}case 40:{if(!g){ro(a,1,c,d,e,f,0,h,l,m)|0;break a}else{ro(a,1,c,d,e,f,k[g>>2]|0,h,l,m)|0;break a}break}case 41:{if(!g){Xm(a,1,c,d,e,f,0,h,l,m)|0;break a}else{Xm(a,1,c,d,e,f,k[g>>2]|0,h,l,m)|0;break a}break}case 81:{if(!g){jn(a,1,c,d,e,f,0,0,h,l,m)|0;break a}else{n=g;jn(a,1,c,d,e,f,k[n>>2]|0,k[n+4>>2]|0,h,l,m)|0;break a}break}case 42:{if(!g){Am(a,1,c,d,e,f,0.0,h,l,m)|0;break a}else{Am(a,1,c,d,e,f,+o[g>>2],h,l,m)|0;break a}break}case 82:{if(!g){mm(a,1,c,d,e,f,0.0,h,l,m)|0;break a}else{mm(a,1,c,d,e,f,+p[g>>3],h,l,m)|0;break a}break}default:k[m>>2]=410}while(0);q=k[m>>2]|0;return q|0}function Zl(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;l=k[j>>2]|0;if((e|0)==0&(f|0)==0|(l|0)>0){m=l;return m|0}do switch(b|0){case 11:{cm(a,1,c,d,e,f,g,h,i,j)|0;break}case 12:{Pn(a,1,c,d,e,f,g,h,i,j)|0;break}case 20:{io(a,1,c,d,e,f,g,h,i,j)|0;break}case 21:{Om(a,1,c,d,e,f,g,h,i,j)|0;break}case 30:{_n(a,1,c,d,e,f,g,h,i,j)|0;break}case 31:{xn(a,1,c,d,e,f,g,h,i,j)|0;break}case 40:{to(a,1,c,d,e,f,g,h,i,j)|0;break}case 41:{Zm(a,1,c,d,e,f,g,h,i,j)|0;break}case 81:{ln(a,1,c,d,e,f,g,h,i,j)|0;break}case 42:{Cm(a,1,c,d,e,f,g,h,i,j)|0;break}case 82:{om(a,1,c,d,e,f,g,h,i,j)|0;break}default:k[j>>2]=410}while(0);m=k[j>>2]|0;return m|0}
function ad(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0,A=0,B=0,C=0,D=0,E=0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0,V=0.0,W=0.0,X=0.0,Y=0.0,$=0.0,ba=0.0,ea=0.0,fa=0.0,ga=0,ha=0.0,ia=0.0,ja=0.0,ka=0.0;d=r;r=r+48|0;e=d+24|0;f=d+16|0;g=d+8|0;h=d;i=d+32|0;if(!(k[b+3304>>2]|0)){j=b+56|0;l=b+64|0;m=b+72|0;n=b+80|0;o=b;q=b+8|0}else{j=b+72|0;l=b+80|0;m=b+56|0;n=b+64|0;o=b+8|0;q=b}s=+p[n>>3];u=+p[m>>3];v=+p[l>>3];w=+p[j>>3];x=+p[o>>3];y=+p[q>>3];if(!(k[c+3304>>2]|0)){z=c;A=c+8|0;B=c+56|0;C=c+64|0;D=c+72|0;E=c+80|0}else{z=c+8|0;A=c;B=c+72|0;C=c+80|0;D=c+56|0;E=c+64|0}F=+p[E>>3];G=+p[D>>3];H=+p[C>>3];I=+p[B>>3];J=+p[z>>3];K=+p[A>>3];L=+p[b+48>>3];M=+p[c+48>>3];N=+p[b+32>>3];O=+p[b+40>>3];P=+p[c+32>>3];Q=+p[c+40>>3];R=+p[b+16>>3];S=+p[b+24>>3];T=+p[c+16>>3];U=+p[c+24>>3];k[a+120>>2]=~~+p[b+136>>3];k[a+128>>2]=~~+p[b+144>>3];k[a+124>>2]=~~+p[c+136>>3];k[a+132>>2]=~~+p[c+144>>3];k[b+5968>>2]=0;k[c+5968>>2]=0;if(x==J&y==K){p[a+6680>>3]=w;p[a+6688>>3]=v;p[a+6696>>3]=u;p[a+6704>>3]=s;p[a+6752>>3]=I;p[a+6760>>3]=H;p[a+6768>>3]=G;p[a+6776>>3]=F;p[a+6712>>3]=+p[b+88>>3];p[a+6720>>3]=+p[b+96>>3];p[a+6728>>3]=+p[b+104>>3];p[a+6736>>3]=+p[b+112>>3];p[a+6784>>3]=+p[c+88>>3];p[a+6792>>3]=+p[c+96>>3];p[a+6800>>3]=+p[c+104>>3];p[a+6808>>3]=+p[c+112>>3];V=M*.0174532925199433;W=L*.0174532925199433;X=0.0}else{Bd(c,x,y,g,h,i);L=+p[g>>3];M=+p[h>>3];if(L!=T){Y=M-U;$=L-T;ba=Y;ea=$;fa=+ca(+(Q*Y/P/$))}else{ba=M-U;ea=L-T;fa=M!=U?1.5707963267948966:0.0}M=+da(+(G*ea+F*ba),+(I*ea+H*ba));ba=+Z(+M);ea=+_(+M);M=I*ba+G*ea;p[a+6752>>3]=M;L=H*ba+F*ea;p[a+6760>>3]=L;$=-ea;ea=G*ba+I*$;p[a+6768>>3]=ea;I=F*ba+H*$;p[a+6776>>3]=I;$=1.0/(M*I-L*ea);p[t>>3]=$;h=k[t+4>>2]&2147483647;if(h>>>0>2146435072|(h|0)==2146435072&(k[t>>2]|0)>>>0>0){Pw(600532,72,1,k[31229]|0)|0;ga=1;r=d;return ga|0}p[a+6784>>3]=I*$;p[a+6792>>3]=-($*L);p[a+6800>>3]=-($*ea);p[a+6808>>3]=$*M;Bd(b,J,K,e,f,i);M=+p[e>>3];$=+p[f>>3];if(M!=R){ea=$-S;L=M-R;ha=L;ia=ea;ja=+ca(+(O*ea/N/L))}else{ha=M-R;ia=$-S;ja=$!=S?1.5707963267948966:0.0}$=P*(+p[g>>3]-T)<0.0?fa+3.141592653589793:fa;fa=N*ha>0.0?ja+3.141592653589793:ja;ja=+da(+(u*ha+s*ia),+(w*ha+v*ia))+3.141592653589793;ia=+Z(+ja);ha=+_(+ja);ja=w*ia+u*ha;p[a+6680>>3]=ja;M=v*ia+s*ha;p[a+6688>>3]=M;L=-ha;ha=u*ia+w*L;p[a+6696>>3]=ha;w=s*ia+v*L;p[a+6704>>3]=w;L=1.0/(ja*w-M*ha);p[t>>3]=L;g=k[t+4>>2]&2147483647;if(g>>>0>2146435072|(g|0)==2146435072&(k[t>>2]|0)>>>0>0){Pw(600605,71,1,k[31229]|0)|0;ga=1;r=d;return ga|0}p[a+6712>>3]=w*L;p[a+6720>>3]=-(L*M);p[a+6728>>3]=-(L*ha);p[a+6736>>3]=L*ja;ja=K*.0174532925199433;K=+Z(+ja);L=J*.0174532925199433;J=y*.0174532925199433;y=+Z(+J);ha=x*.0174532925199433;x=K*+Z(+L)*y*+Z(+ha)+ +_(+ha)*(y*(K*+_(+L)))+ +_(+ja)*+_(+J);if(!(x>1.0))if(x<-1.0)ka=-1.0;else ka=x;else ka=1.0;V=$;W=fa;X=+aa(+ka)}p[a+8>>3]=R;p[a+16>>3]=S;p[a+24>>3]=T;p[a+32>>3]=U;U=+Z(+W);p[a+40>>3]=U;U=+_(+W);p[a+48>>3]=U;U=+Z(+V);p[a+56>>3]=U;U=+_(+V);p[a+64>>3]=U;U=+Z(+X);p[a+72>>3]=U;U=+_(+X);p[a+80>>3]=U;p[a+88>>3]=N;p[a+96>>3]=O;p[a+104>>3]=P;p[a+112>>3]=Q;ux(a,b+3467|0)|0;ux(a+4|0,c+3467|0)|0;k[a+6672>>2]=1;k[a+6664>>2]=0;k[a+6668>>2]=0;k[a+6676>>2]=1;k[a+6744>>2]=1;ga=0;r=d;return ga|0}function bd(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=md(b)|0;e=md(c)|0;f=ad(a,d,e)|0;if(f){g=f;return g|0}f=ed(b,a+136|0)|0;k[a+6664>>2]=f;f=ed(c,a+3400|0)|0;k[a+6668>>2]=f;if(e)dy(e);if(!d){g=0;return g|0}dy(d);g=0;return g|0}function cd(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0.0,l=0.0,m=0,n=0,o=0.0,q=0,s=0,t=0.0,u=0,v=0.0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0,E=0.0;f=r;r=r+80|0;g=f;h=k[c+1616>>2]|0;i=k[c+2424>>2]|0;j=a-+p[c+3232>>3];l=b-+p[c+3240>>3];if((h|0)>=0){m=0;while(1){n=h-m|0;o=+p[c+1624+(n*80|0)+(m<<3)>>3];q=g+(m<<3)|0;p[q>>3]=o;if((m|0)>0){s=m;t=o;while(1){u=s;s=s+-1|0;o=l*t+ +p[c+1624+(n*80|0)+(s<<3)>>3];if((u|0)<=1){v=o;break}else t=o}p[q>>3]=v}if((h|0)==(m|0))break;else m=m+1|0}v=+p[g>>3];if((h|0)>0){m=h+1|0;s=h;t=v;while(1){o=j*t+ +p[g+(m-s<<3)>>3];if((s|0)>1){s=s+-1|0;t=o}else{w=g;x=o;break}}}else{w=g;x=v}}else{w=g;x=0.0}p[d>>3]=x;if((i|0)<0){y=+p[w>>3];p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}else D=0;while(1){s=i-D|0;x=+p[c+2432+(s*80|0)+(D<<3)>>3];m=g+(D<<3)|0;p[m>>3]=x;if((D|0)>0){h=D;v=x;while(1){n=h;h=h+-1|0;x=l*v+ +p[c+2432+(s*80|0)+(h<<3)>>3];if((n|0)<=1){E=x;break}else v=x}p[m>>3]=E}if((i|0)==(D|0))break;else D=D+1|0}E=+p[w>>3];if((i|0)<=0){y=E;p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}w=i+1|0;D=i;l=E;while(1){E=j*l+ +p[g+(w-D<<3)>>3];if((D|0)>1){D=D+-1|0;l=E}else{y=E;break}}p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}function dd(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0.0,l=0.0,m=0,n=0,o=0.0,q=0,s=0,t=0.0,u=0,v=0.0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0,E=0.0;f=r;r=r+80|0;g=f;h=k[c>>2]|0;i=k[c+808>>2]|0;j=a-+p[c+3232>>3];l=b-+p[c+3240>>3];if((h|0)>=0){m=0;while(1){n=h-m|0;o=+p[c+8+(n*80|0)+(m<<3)>>3];q=g+(m<<3)|0;p[q>>3]=o;if((m|0)>0){s=m;t=o;while(1){u=s;s=s+-1|0;o=l*t+ +p[c+8+(n*80|0)+(s<<3)>>3];if((u|0)<=1){v=o;break}else t=o}p[q>>3]=v}if((h|0)==(m|0))break;else m=m+1|0}v=+p[g>>3];if((h|0)>0){m=h+1|0;s=h;t=v;while(1){o=j*t+ +p[g+(m-s<<3)>>3];if((s|0)>1){s=s+-1|0;t=o}else{w=g;x=o;break}}}else{w=g;x=v}}else{w=g;x=0.0}p[d>>3]=x;if((i|0)<0){y=+p[w>>3];p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}else D=0;while(1){s=i-D|0;x=+p[c+816+(s*80|0)+(D<<3)>>3];m=g+(D<<3)|0;p[m>>3]=x;if((D|0)>0){h=D;v=x;while(1){n=h;h=h+-1|0;x=l*v+ +p[c+816+(s*80|0)+(h<<3)>>3];if((n|0)<=1){E=x;break}else v=x}p[m>>3]=E}if((i|0)==(D|0))break;else D=D+1|0}E=+p[w>>3];if((i|0)<=0){y=E;p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}w=i+1|0;D=i;l=E;while(1){E=j*l+ +p[g+(w-D<<3)>>3];if((D|0)>1){D=D+-1|0;l=E}else{y=E;break}}p[e>>3]=y;z=+p[d>>3];A=z+a;p[d>>3]=A;B=+p[e>>3];C=B+b;p[e>>3]=C;r=f;return}function ed(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;c=r;r=r+304|0;d=c+144|0;e=c+136|0;f=c+128|0;g=c+120|0;h=c+112|0;j=c+104|0;l=c+96|0;m=c+88|0;n=c+80|0;o=c;p=c+229|0;q=c+156|0;s=c+152|0;k[s>>2]=0;i[o>>0]=i[777959]|0;i[o+1>>0]=i[777960]|0;i[o+2>>0]=i[777961]|0;i[o+3>>0]=i[777962]|0;i[o+4>>0]=i[777963]|0;i[o+5>>0]=i[777964]|0;i[o+6>>0]=i[777965]|0;if(gd(a,p,o)|0){t=k[31229]|0;k[n>>2]=o;Ew(t,600677,n)|0;u=-1;r=c;return u|0}if((xx(p)|0)==8){u=0;r=c;return u|0}Bx(q,p+9|0,4)|0;i[q+4>>0]=0;if(tx(q,606892)|0){u=-1;r=c;return u|0}q=o;k[q>>2]=1380933441;k[q+4>>2]=5391684;if(hd(a,s,o)|0){q=k[31229]|0;k[m>>2]=o;Ew(q,600677,m)|0}m=k[s>>2]|0;k[b>>2]=m;q=(m|0)<0;if(!q){p=(m<<3)+8|0;n=0;while(1){sy(b+8+(n*80|0)|0,0,p|0)|0;if((n|0)==(m|0))break;else n=n+1|0}if(!q){q=m+1|0;n=0;p=q;while(1){if((m|0)>=(n|0)){t=0;do{k[l>>2]=n;k[l+4>>2]=t;Xw(o,606659,l)|0;fd(a,b+8+(n*80|0)+(t<<3)|0,o)|0;t=t+1|0}while((t|0)!=(p|0))}n=n+1|0;if((n|0)==(q|0))break;else p=p+-1|0}}}p=o;k[p>>2]=1380933442;k[p+4>>2]=5391684;if(hd(a,s,o)|0){p=k[31229]|0;k[j>>2]=o;Ew(p,600677,j)|0}j=k[s>>2]|0;k[b+808>>2]=j;p=(j|0)<0;if(!p){q=(j<<3)+8|0;n=0;while(1){sy(b+816+(n*80|0)|0,0,q|0)|0;if((n|0)==(j|0))break;else n=n+1|0}if(!p){p=j+1|0;n=0;q=p;while(1){if((j|0)>=(n|0)){l=0;do{k[h>>2]=n;k[h+4>>2]=l;Xw(o,606732,h)|0;fd(a,b+816+(n*80|0)+(l<<3)|0,o)|0;l=l+1|0}while((l|0)!=(q|0))}n=n+1|0;if((n|0)==(p|0))break;else q=q+-1|0}}}q=o;p=606740;n=q+9|0;do{i[q>>0]=i[p>>0]|0;q=q+1|0;p=p+1|0}while((q|0)<(n|0));if(hd(a,s,o)|0){h=k[31229]|0;k[g>>2]=o;Ew(h,600677,g)|0}g=k[s>>2]|0;k[b+1616>>2]=g;h=(g|0)<0;if(!h){j=(g<<3)+8|0;l=0;while(1){sy(b+1624+(l*80|0)|0,0,j|0)|0;if((l|0)==(g|0))break;else l=l+1|0}if(!h){h=g+1|0;l=0;j=h;while(1){if((g|0)>=(l|0)){m=0;do{i[o>>0]=0;k[f>>2]=l;k[f+4>>2]=m;Xw(o,606807,f)|0;fd(a,b+1624+(l*80|0)+(m<<3)|0,o)|0;m=m+1|0}while((m|0)!=(j|0))}l=l+1|0;if((l|0)==(h|0))break;else j=j+-1|0}}}q=o;p=606816;n=q+9|0;do{i[q>>0]=i[p>>0]|0;q=q+1|0;p=p+1|0}while((q|0)<(n|0));if(hd(a,s,o)|0){p=k[31229]|0;k[e>>2]=o;Ew(p,600677,e)|0}e=k[s>>2]|0;k[b+2424>>2]=e;s=(e|0)<0;if(!s){p=(e<<3)+8|0;q=0;while(1){sy(b+2432+(q*80|0)|0,0,p|0)|0;if((q|0)==(e|0))break;else q=q+1|0}if(!s){s=e+1|0;q=0;p=s;while(1){if((e|0)>=(q|0)){n=0;do{k[d>>2]=q;k[d+4>>2]=n;Xw(o,606883,d)|0;fd(a,b+2432+(q*80|0)+(n<<3)|0,o)|0;n=n+1|0}while((n|0)!=(p|0))}q=q+1|0;if((q|0)==(s|0))break;else p=p+-1|0}}};i[o>>0]=i[777876]|0;i[o+1>>0]=i[777877]|0;i[o+2>>0]=i[777878]|0;i[o+3>>0]=i[777879]|0;i[o+4>>0]=i[777880]|0;i[o+5>>0]=i[777881]|0;i[o+6>>0]=i[777882]|0;if(fd(a,b+3232|0,o)|0){Pw(600705,21,1,k[31229]|0)|0;u=-1;r=c;return u|0};i[o>>0]=i[777883]|0;i[o+1>>0]=i[777884]|0;i[o+2>>0]=i[777885]|0;i[o+3>>0]=i[777886]|0;i[o+4>>0]=i[777887]|0;i[o+5>>0]=i[777888]|0;i[o+6>>0]=i[777889]|0;if(!(fd(a,b+3240|0,o)|0)){u=1;r=c;return u|0}Pw(600705,21,1,k[31229]|0)|0;u=-1;r=c;return u|0}function fd(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,k=0,l=0,m=0,n=0,o=0,q=0,s=0.0;d=r;r=r+96|0;e=d;f=d+16|0;g=d+2|0;j[e>>1]=32;ux(g,c)|0;c=xx(g)|0;h=g+c|0;if((c|0)<8){k=h;l=c;while(1){i[k>>0]=32;i[k+1>>0]=0;l=l+1|0;c=g+(xx(g)|0)|0;if((l|0)==8){m=c;break}else k=c}}else m=h;i[m>>0]=61;i[m+1>>0]=0;m=Ex(a,g)|0;if(!m){n=1;r=d;return n|0}g=rx(m,61)|0;if(!g){n=1;r=d;return n|0}else o=g;while(1){g=o+1|0;if((i[g>>0]|0)==32)o=g;else{q=g;break}}o=vx(q,e)|0;if((o|0)>79){n=1;r=d;return n|0}Bx(f,q,o)|0;i[f+o>>0]=0;s=+dx(f);p[b>>3]=s;n=0;r=d;return n|0}function gd(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,k=0,l=0,m=0,n=0,o=0,p=0;d=r;r=r+96|0;e=d;f=d+16|0;g=d+2|0;j[e>>1]=32;ux(g,c)|0;c=xx(g)|0;h=g+c|0;if((c|0)<8){k=h;l=c;while(1){i[k>>0]=32;i[k+1>>0]=0;l=l+1|0;c=g+(xx(g)|0)|0;if((l|0)==8){m=c;break}else k=c}}else m=h;i[m>>0]=61;i[m+1>>0]=0;m=Ex(a,g)|0;if(!m){n=1;r=d;return n|0}g=rx(m,61)|0;if(!g){n=1;r=d;return n|0}else o=g;while(1){g=o+1|0;if((i[g>>0]|0)==32)o=g;else{p=g;break}}o=vx(p,e)|0;if((o|0)>79){n=1;r=d;return n|0}Bx(f,p,o)|0;i[f+o>>0]=0;ux(b,f)|0;n=0;r=d;return n|0}function hd(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,l=0,m=0,n=0,o=0,p=0,q=0;d=r;r=r+96|0;e=d;f=d+16|0;g=d+2|0;j[e>>1]=32;ux(g,c)|0;c=xx(g)|0;h=g+c|0;if((c|0)<8){l=h;m=c;while(1){i[l>>0]=32;i[l+1>>0]=0;m=m+1|0;c=g+(xx(g)|0)|0;if((m|0)==8){n=c;break}else l=c}}else n=h;i[n>>0]=61;i[n+1>>0]=0;n=Ex(a,g)|0;if(!n){o=1;r=d;return o|0}g=rx(n,61)|0;if(!g){o=1;r=d;return o|0}else p=g;while(1){g=p+1|0;if((i[g>>0]|0)==32)p=g;else{q=g;break}}p=vx(q,e)|0;if((p|0)>79){o=1;r=d;return o|0}Bx(f,q,p)|0;i[f+p>>0]=0;p=ex(f)|0;k[b>>2]=p;o=0;r=d;return o|0}function id(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0.0,j=0,l=0.0,m=0,n=0,o=0.0,p=0.0,q=0.0,s=0.0,t=0.0,u=0.0,v=0;c=r;r=r+16|0;d=c;k[25986]=0;e=(b|0)==0;f=a;while(1)if((i[f>>0]|0)==32)f=f+1|0;else{g=f;break}f=e?d:b;h=+hx(g,f);b=k[f>>2]|0;if(rx(g,46)|0)k[25986]=46;d=i[b>>0]|0;e=d<<24>>24;switch(d<<24>>24){case 32:case 58:case 100:case 104:case 109:{j=6;break}case 39:case 114:case 34:break;default:{l=h;r=c;return +l}}a:do if((j|0)==6){a=b;do if((a-g|0)<5){m=b+1|0;n=i[m>>0]|0;if(((n<<24>>24)+-48|0)>>>0>=10){if(n<<24>>24!=32)break;if(((i[b+2>>0]|0)+-48|0)>>>0>=10)break}k[25986]=e;k[f>>2]=m;n=(i[g>>0]|0)==45;o=n?-h:h;p=n?-1.0:1.0;q=+hx(m,f);b:do if(d<<24>>24!=109){m=k[f>>2]|0;switch(i[m>>0]|0){case 32:case 58:case 109:break;default:{s=o;t=q;u=0.0;break b}}if((m-a|0)<4){n=m+1|0;v=i[n>>0]|0;if(((v<<24>>24)+-48|0)>>>0>=10){if(v<<24>>24!=32){s=o;t=q;u=0.0;break}if(((i[m+2>>0]|0)+-48|0)>>>0>=10){s=o;t=q;u=0.0;break}}k[f>>2]=n;s=o;t=q;u=+hx(n,f)}else{s=o;t=q;u=0.0}}else{s=0.0;t=o;u=q}while(0);l=p*(s+t/60.0+u/3600.0);r=c;return +l}while(0);switch(d<<24>>24){case 39:case 114:case 100:case 34:{break a;break}default:l=h}r=c;return +l}while(0);d=b+1|0;if(i[d>>0]|0){l=h;r=c;return +l}k[25986]=e;k[f>>2]=d;l=h;r=c;return +l}function jd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0.0,Ua=0.0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0.0,cb=0.0,db=0.0,eb=0,fb=0.0,gb=0,hb=0.0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0.0,ub=0.0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0;c=r;r=r+1520|0;d=c+1192|0;e=c+1184|0;f=c+1176|0;g=c+1168|0;h=c+1160|0;j=c+1152|0;l=c+1144|0;m=c+1136|0;n=c+1128|0;o=c+1064|0;q=c+1056|0;s=c+1048|0;t=c+1040|0;u=c+1032|0;v=c+1024|0;w=c+1016|0;x=c+1008|0;y=c+1e3|0;z=c+992|0;A=c+984|0;B=c+976|0;C=c+968|0;D=c+960|0;E=c+952|0;F=c+944|0;G=c+936|0;H=c+928|0;I=c+920|0;J=c+912|0;K=c+904|0;L=c+896|0;M=c+888|0;N=c+880|0;O=c+872|0;P=c+864|0;Q=c+856|0;R=c+848|0;S=c+840|0;T=c+832|0;U=c+824|0;V=c+816|0;W=c+808|0;X=c+1504|0;Y=c+1472|0;$=c+1440|0;aa=c+1328|0;ba=c+1320|0;ca=c+1248|0;da=c+1240|0;ea=c+1232|0;fa=c+152|0;ga=c+144|0;ia=c+136|0;ja=c+128|0;ka=c+120|0;la=c+112|0;ma=c+104|0;na=c+96|0;oa=c+88|0;pa=c+80|0;qa=c+72|0;ra=c+64|0;sa=c+56|0;ta=c+24|0;ua=c+160|0;va=c+1216|0;wa=c+1208|0;xa=c+1204|0;ya=c+1200|0;za=c+1360|0;Aa=c+1256|0;Ba=c+1212|0;Ca=c+16|0;Da=c+8|0;Ea=c;Fa=ey(1,9336)|0;Ga=i[b>>0]|0;b=Ga<<24>>24==32?0:Ga;i[Ba>>0]=b;i[Fa+9324>>0]=b;if(le(a,600777,Ba,63,Aa)|0){b=ey((xx(Aa)|0)+2|0,1)|0;k[Fa+9320>>2]=b;ux(b,Aa)|0}b=Fa+3984|0;k[b>>2]=0;k[Fa+3956>>2]=0;k[Fa+3924>>2]=0;k[Fa+3952>>2]=-1;Ga=Fa+3268|0;k[Ga>>2]=0;Ha=Fa+3272|0;k[Ha>>2]=0;p[ra>>3]=0.0;p[sa>>3]=0.0;Ia=ta+8|0;Ja=ta+16|0;Ka=ta+24|0;k[ta>>2]=0;k[ta+4>>2]=0;k[ta+8>>2]=0;k[ta+12>>2]=0;k[ta+16>>2]=0;k[ta+20>>2]=0;k[ta+24>>2]=0;k[ta+28>>2]=0;p[ua>>3]=0.0;La=Fa+3300|0;k[La>>2]=0;Ma=Fa+48|0;p[Ma>>3]=0.0;k[ya>>2]=0;ae(a,600785,Ba,ya)|0;Na=k[ya>>2]|0;if(!Na){be(a,600785,ya)|0;Oa=k[ya>>2]|0;if(!Oa){be(a,782713,ya)|0;Pa=k[ya>>2]|0;if(!Pa){be(a,600793,ya)|0;Qa=k[ya>>2]|0}else Qa=Pa}else Qa=Oa}else Qa=Na;if((Qa|0)<1){xd(600800);qd(Fa);Ra=0;r=c;return Ra|0}if((Qa|0)>2){k[ya>>2]=2;Sa=2}else Sa=Qa;Qa=Fa+3316|0;k[Qa>>2]=Sa;Na=Fa+3320|0;k[Na>>2]=Sa;k[Fa+3960>>2]=Sa;Sa=Fa+136|0;p[Sa>>3]=0.0;ie(a,772833,Sa)|0;Ta=+p[Sa>>3];if(Ta<1.0){ie(a,600846,Sa)|0;Ua=+p[Sa>>3]}else Ua=Ta;if(Ua<1.0){xd(600853);qd(Fa);Ra=0;r=c;return Ra|0}Oa=Fa+144|0;p[Oa>>3]=0.0;ie(a,790243,Oa)|0;if(+p[Oa>>3]<1.0)ie(a,600890,Oa)|0;Pa=k[ya>>2]|0;if((Pa|0)>1)if(+p[Oa>>3]<1.0){xd(600897);qd(Fa);Ra=0;r=c;return Ra|0}else Va=19;else if((Pa|0)>0)Va=19;else Wa=0;if((Va|0)==19){Pa=k[31229]|0;Xa=0;Ya=0;while(1){i[va>>0]=i[782713]|0;i[va+1>>0]=i[782714]|0;i[va+2>>0]=i[782715]|0;i[va+3>>0]=i[782716]|0;i[va+4>>0]=i[782717]|0;i[va+5>>0]=i[782718]|0;Za=Xa;Xa=Xa+1|0;k[W>>2]=Xa;Xw(za,772183,W)|0;qx(va,za)|0;a:do if(!(be(a,va,xa)|0)){switch(Za|0){case 0:{Ua=+p[Sa>>3];if(Ua>1.0){k[xa>>2]=~~Ua;break a}break}case 1:{Ua=+p[Oa>>3];if(Ua>1.0){k[xa>>2]=~~Ua;break a}break}default:{}}k[V>>2]=va;Ew(Pa,600934,V)|0}while(0);i[va>>0]=i[768517]|0;i[va+1>>0]=i[768518]|0;i[va+2>>0]=i[768519]|0;i[va+3>>0]=i[768520]|0;i[va+4>>0]=i[768521]|0;i[va+5>>0]=i[768522]|0;qx(va,za)|0;if((me(a,va,16,za)|0)!=0?(ne(za,600973)|0)!=0:0)k[xa>>2]=0;Za=((k[xa>>2]|0)>1&1)+Ya|0;if((Xa|0)>=(k[ya>>2]|0)){Wa=Za;break}else Ya=Za}}k[ya>>2]=Wa;k[Na>>2]=Wa;k[Qa>>2]=Wa;me(a,758734,16,Fa+3336|0)|0;be(a,600978,Fa+3332|0)|0;Wa=Jd()|0;Na=Fa+3324|0;k[Na>>2]=Wa;Wa=Fa+9328|0;k[Wa>>2]=0;be(a,600987,Wa)|0;Wa=Fa+832|0;sy(Wa|0,0,648)|0;sy(ua|0,0,648)|0;Ya=k[ya>>2]|0;Xa=(Ya|0)>0;if(Xa){V=0;do{Pa=Wa+((ha(Ya,V)|0)+V<<3)|0;p[Pa>>3]=1.0;V=V+1|0}while((V|0)<(Ya|0));if(Xa){V=0;do{Pa=ua+((ha(Ya,V)|0)+V<<3)|0;p[Pa>>3]=1.0;V=V+1|0}while((V|0)<(Ya|0));V=Fa+760|0;_a=V;$a=_a+72|0;do{k[_a>>2]=0;_a=_a+4|0}while((_a|0)<($a|0));if(Xa){Xa=0;do{p[V+(Xa<<3)>>3]=1.0;Xa=Xa+1|0}while((Xa|0)<(Ya|0));ab=V}else ab=V}else Va=37}else Va=37;if((Va|0)==37){V=Fa+760|0;_a=V;$a=_a+72|0;do{k[_a>>2]=0;_a=_a+4|0}while((_a|0)<($a|0));ab=V}do if(!(le(a,600995,Ba,63,Aa)|0))k[Fa+9312>>2]=0;else{V=kd(a,Aa)|0;i[X>>0]=V;if(V<<24>>24!=95){V=jd(a,X)|0;k[Fa+9312>>2]=V;if(V){k[V+9316>>2]=Fa;break}}else{V=k[31229]|0;k[U>>2]=Aa;Ew(V,600727,U)|0;k[Fa+9312>>2]=0}xd(601002);qd(Fa);Ra=0;r=c;return Ra|0}while(0);U=Fa+3224|0;Aa=Fa+3232|0;k[U>>2]=0;k[U+4>>2]=0;k[U+8>>2]=0;k[U+12>>2]=0;do if(!(he(a,601044,Ba,U)|0)){if(he(a,601052,Ba,Aa)|0){p[U>>3]=+p[Aa>>3]*299792.5;break}if(ie(a,601060,U)|0)p[Aa>>3]=+p[U>>3]/299792.5}else p[Aa>>3]=+p[U>>3]/299792.5;while(0);U=Fa+4096|0;_a=U;$a=_a+80|0;do{k[_a>>2]=0;_a=_a+4|0}while((_a|0)<($a|0));do if(!(le(a,777959,Ba,16,Y)|0)){Aa=i[Ba>>0]|0;switch(Aa<<24>>24){case 32:case 0:break;default:{k[q>>2]=Aa<<24>>24;Xw(za,601360,q)|0;xd(za);qd(Fa);Ra=0;r=c;return Ra|0}}if($d(a,601396)|0){k[Fa+3260>>2]=29;Aa=$d(a,601396)|0;ie(Aa,601396,fa)|0;ie(Aa,601403,ga)|0;ie(Aa,601410,ia)|0;p[Fa+152>>3]=(+p[fa>>3]+ +p[ga>>3]/60.0+ +p[ia>>3]/3600.0)*15.0*3.141592653589793/180.0;i[ea>>0]=43;me(Aa,601417,1,ea)|0;Ua=(i[ea>>0]|0)==45?-1.0:1.0;ie(Aa,601426,ja)|0;ie(Aa,601434,ka)|0;ie(Aa,601442,la)|0;p[Fa+160>>3]=Ua*(+p[ja>>3]+ +p[ka>>3]/60.0+ +p[la>>3]/3600.0)*3.141592653589793/180.0;Aa=Fa+120|0;ie(a,601450,Aa)|0;be(a,601450,wa)|0;V=Fa+3764|0;Ya=(k[wa>>2]|0)==1950?3427142:3492678;i[V>>0]=Ya;i[V+1>>0]=Ya>>8;i[V+2>>0]=Ya>>16;i[V+3>>0]=Ya>>24;Ya=Fa+128|0;p[Ya>>3]=+p[Aa>>3];ie(a,758683,Ya)|0;Ua=+p[ga>>3];Ta=+p[ia>>3];Ya=i[ea>>0]|0;bb=+p[ja>>3];cb=+p[ka>>3];db=+p[la>>3];p[o>>3]=+p[fa>>3];p[o+8>>3]=Ua;p[o+16>>3]=Ta;k[o+24>>2]=Ya;p[o+32>>3]=bb;p[o+40>>3]=cb;p[o+48>>3]=db;k[o+56>>2]=V;Xw(Fa+3892|0,601458,o)|0;ie(a,601499,Fa+168|0)|0;ie(a,601508,Fa+192|0)|0;ie(a,601517,Fa+200|0)|0;ie(a,601526,Fa+176|0)|0;ie(a,601533,Fa+184|0)|0;V=$d(a,601540)|0;Ya=Fa+208|0;k[n>>2]=1;Xw(va,601545,n)|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;k[m>>2]=2;Xw(va,601545,m)|0;Ya=Fa+216|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;k[l>>2]=3;Xw(va,601545,l)|0;Ya=Fa+224|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;k[j>>2]=4;Xw(va,601545,j)|0;Ya=Fa+232|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;k[h>>2]=5;Xw(va,601545,h)|0;Ya=Fa+240|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;k[g>>2]=6;Xw(va,601545,g)|0;Ya=Fa+248|0;p[Ya>>3]=0.0;ie(V,va,Ya)|0;Ya=$d(a,601551)|0;V=Fa+256|0;Aa=0;do{Xa=Aa;Aa=Aa+1|0;k[f>>2]=Aa;Xw(va,601557,f)|0;Pa=V+(Xa<<3)|0;p[Pa>>3]=0.0;ie(Ya,va,Pa)|0}while((Aa|0)!=20);Aa=$d(a,601564)|0;Ya=Fa+416|0;V=0;do{Pa=V;V=V+1|0;k[e>>2]=V;Xw(va,601570,e)|0;Xa=Ya+(Pa<<3)|0;p[Xa>>3]=0.0;ie(Aa,va,Xa)|0}while((V|0)!=20);V=Fa+3312|0;k[V>>2]=1;Aa=Fa+3449|0;i[Aa>>0]=i[606317]|0;i[Aa+1>>0]=i[606318]|0;i[Aa+2>>0]=i[606319]|0;Aa=Fa+3458|0;i[Aa>>0]=68;i[Aa+1>>0]=69;i[Aa+2>>0]=67;i[Aa+3>>0]=0;Aa=Fa+3467|0;i[Aa>>0]=68;i[Aa+1>>0]=83;i[Aa+2>>0]=83;i[Aa+3>>0]=0;k[Fa+3292>>2]=0;k[Fa+3288>>2]=3;_a=Fa+3368|0;Aa=601577;$a=_a+9|0;do{i[_a>>0]=i[Aa>>0]|0;_a=_a+1|0;Aa=Aa+1|0}while((_a|0)<($a|0));_a=Fa+3377|0;Aa=601586;$a=_a+9|0;do{i[_a>>0]=i[Aa>>0]|0;_a=_a+1|0;Aa=Aa+1|0}while((_a|0)<($a|0));db=+p[Sa>>3]*.5;Aa=Fa+616|0;p[Aa>>3]=db;cb=+p[Oa>>3]*.5;Ya=Fa+624|0;p[Ya>>3]=cb;p[Fa+16>>3]=db;p[Fa+24>>3]=cb;Me(db,cb,Fa,na,pa)|0;cb=+p[na>>3];p[Fa+688>>3]=cb;db=+p[pa>>3];p[Fa+696>>3]=db;p[Fa>>3]=cb;p[Fa+8>>3]=db;Me(+p[Aa>>3],+p[Ya>>3]+1.0,Fa,oa,qa)|0;db=+p[qa>>3]-+p[pa>>3];p[Fa+40>>3]=db;p[Fa+32>>3]=-db;pd(Fa);k[V>>2]=1;zd(Fa);db=+p[Ma>>3]*3.141592653589793/180.0;p[Da>>3]=db;cb=+p[Aa>>3]+ +Z(+db);Me(cb,+p[Ya>>3]+ +_(+db),Fa,oa,qa)|0;db=-+Dd(+p[na>>3],+p[pa>>3],+p[oa>>3],+p[qa>>3]);p[ab>>3]=db;db=+p[Da>>3];cb=+p[Aa>>3]+ +_(+db);Me(cb,+p[Ya>>3]+ +Z(+db),Fa,oa,qa)|0;db=+Dd(+p[na>>3],+p[pa>>3],+p[oa>>3],+p[qa>>3]);p[Fa+768>>3]=db;td(Fa,+p[ab>>3],db,+p[Ma>>3]);eb=Aa;break}if((((($d(a,601202)|0)==0?($d(a,601209)|0)==0:0)?($d(a,601218)|0)==0:0)?($d(a,601227)|0)==0:0)?($d(a,601236)|0)==0:0){xd(601692);qd(Fa);Ra=0;r=c;return Ra|0}p[ma>>3]=0.0;ie(a,601202,ma)|0;db=+p[ma>>3];if(db==0.0){ie(a,601209,ma)|0;fb=+p[ma>>3]}else fb=db;do if(fb==0.0){ie(a,601236,ma)|0;db=+p[ma>>3];if(db!=0.0){p[ra>>3]=-db/3600.0;ie(a,601244,ma)|0;p[sa>>3]=+p[ma>>3]/3600.0;break}ie(a,601227,ma)|0;db=+p[ma>>3];if(db!=0.0){p[ra>>3]=-db/3600.0;ie(a,601252,ma)|0;p[sa>>3]=+p[ma>>3]/3600.0;break}else{ie(a,601218,ma)|0;p[ra>>3]=-+p[ma>>3]/3600.0;ie(a,601261,ma)|0;p[sa>>3]=+p[ma>>3]/3600.0;break}}else{db=fb/3600.0;p[sa>>3]=db;p[ra>>3]=-db}while(0);p[Da>>3]=0.0;ie(a,601302,Da)|0;if(+p[Ma>>3]==0.0)ie(a,777928,Da)|0;td(Fa,+p[ra>>3],+p[sa>>3],+p[Da>>3]);V=Fa+616|0;p[V>>3]=+p[Sa>>3]*.5+.5;Aa=Fa+624|0;p[Aa>>3]=+p[Oa>>3]*.5+.5;if($d(a,777876)|0){ie(a,777876,V)|0;ie(a,777883,Aa)|0}p[Fa+16>>3]=+p[V>>3];p[Fa+24>>3]=+p[Aa>>3];Aa=Fa+688|0;p[Aa>>3]=-999.0;if(!(ee(a,606317,Aa)|0)){xd(601595);qd(Fa);Ra=0;r=c;return Ra|0}Ya=Fa+696|0;p[Ya>>3]=-999.0;if(!(fe(a,601933,Ya)|0)){xd(601630);qd(Fa);Ra=0;r=c;return Ra|0}db=+p[Aa>>3];p[Fa>>3]=db;cb=+p[Ya>>3];p[Fa+8>>3]=cb;Ya=Fa+3304|0;k[Ya>>2]=0;p[Fa+3992>>3]=db;p[Fa+4e3>>3]=cb;Aa=Fa+4008|0;p[Aa>>3]=999.0;if(!(ie(a,601110,Aa)|0))ie(a,601665,Aa)|0;Aa=Fa+4016|0;p[Aa>>3]=999.0;ie(a,601118,Aa)|0;do if(!(ie(a,758550,Ca)|0)){Aa=Fa+128|0;if(!(je(a,601337,Aa)|0)){if(je(a,601346,Aa)|0)break;if(ie(a,758683,Aa)|0)break;p[Aa>>3]=+p[Fa+120>>3];break}me(a,601337,32,aa)|0;if(rx(aa,84)|0)break;if(ie(a,601351,Ea)|0){p[Aa>>3]=+p[Aa>>3]+ +p[Ea>>3]/8765.812770744;break}if(!(ie(a,601354,Ea)|0))break;p[Aa>>3]=+p[Aa>>3]+ +p[Ea>>3]/8765.812770744}else p[Fa+128>>3]=(+p[Ca>>3]+-15019.81352)/365.242198781+1900.0;while(0);sd(Fa,601674,601683)|0;k[Ya>>2]=0;i[X>>0]=0;od(a,Fa,X);pd(Fa);k[Fa+3292>>2]=0;k[Fa+3288>>2]=3;k[Fa+3312>>2]=1;eb=V}else{ux($,Y)|0;Aa=(le(a,601069,Ba,16,$)|0)==0;ux(Fa+3368|0,Y)|0;ux(Fa+3377|0,$)|0;if(!(ne($,777971)|0)){Xa=(ne($,601933)|0)==0;gb=Xa?1:2}else gb=2;Xa=Fa+3386|0;i[Xa>>0]=0;le(a,601076,Ba,9,Xa)|0;Xa=Fa+3395|0;i[Xa>>0]=0;le(a,601083,Ba,9,Xa)|0;if(sd(Fa,Y,$)|0){qd(Fa);Ra=0;r=c;return Ra|0}Xa=Fa+3260|0;do if(!(k[Xa>>2]|0)){Pa=Fa+3476|0;if((le(a,601090,Ba,16,Pa)|0)==0?(ve(a,605354,601097,16,Pa)|0)==0:0)i[Pa>>0]=0;if(!(tx(Pa,774950)|0))k[Xa>>2]=-1;if(!Aa){W=Fa+3508|0;do if(!(le(a,601103,Ba,16,W)|0)){if(ve(a,605520,601097,16,W)|0)break;i[W>>0]=0}while(0);if(tx(Pa,774950)|0)break;k[Xa>>2]=-1}}while(0);Aa=Fa+616|0;p[Aa>>3]=1.0;he(a,777876,Ba,Aa)|0;V=Fa+624|0;p[V>>3]=1.0;he(a,777883,Ba,V)|0;p[Fa+16>>3]=+p[Aa>>3];p[Fa+24>>3]=+p[V>>3];Ya=Fa+688|0;p[Ya>>3]=0.0;he(a,777862,Ba,Ya)|0;W=Fa+696|0;p[W>>3]=0.0;he(a,777869,Ba,W)|0;switch(k[Fa+3884>>2]|0){case 7:{p[W>>3]=90.0-+p[W>>3];Va=74;break}case 8:{cb=+p[W>>3]+-90.0;p[W>>3]=cb;hb=cb;break}default:Va=74}if((Va|0)==74)hb=+p[W>>3];cb=+p[Ya>>3];p[Fa>>3]=cb;p[Fa+8>>3]=hb;Ya=Fa+3992|0;if(!(k[Fa+3304>>2]|0)){p[Ya>>3]=cb;p[Fa+4e3>>3]=hb}else{p[Ya>>3]=hb;p[Fa+4e3>>3]=cb}Ya=Fa+3176|0;p[Ya>>3]=999.0;he(a,601110,Ba,Ya)|0;p[Fa+4008>>3]=+p[Ya>>3];Ya=Fa+3184|0;p[Ya>>3]=999.0;he(a,601118,Ba,Ya)|0;p[Fa+4016>>3]=+p[Ya>>3];k[Fa+3964>>2]=Aa;k[Fa+3972>>2]=ab;k[Fa+3968>>2]=Wa;Ya=Fa+4088|0;p[Ya>>3]=0.0;he(a,601126,Ba,Ya)|0;k[T>>2]=0;Xw(va,601133,T)|0;he(a,va,Ba,U)|0;k[S>>2]=1;Xw(va,601133,S)|0;Ya=Fa+4104|0;he(a,va,Ba,Ya)|0;k[R>>2]=2;Xw(va,601133,R)|0;W=Fa+4112|0;he(a,va,Ba,W)|0;k[Q>>2]=3;Xw(va,601133,Q)|0;Za=Fa+4120|0;he(a,va,Ba,Za)|0;k[P>>2]=4;Xw(va,601133,P)|0;ib=Fa+4128|0;he(a,va,Ba,ib)|0;k[O>>2]=5;Xw(va,601133,O)|0;jb=Fa+4136|0;he(a,va,Ba,jb)|0;k[N>>2]=6;Xw(va,601133,N)|0;kb=Fa+4144|0;he(a,va,Ba,kb)|0;k[M>>2]=7;Xw(va,601133,M)|0;lb=Fa+4152|0;he(a,va,Ba,lb)|0;k[L>>2]=8;Xw(va,601133,L)|0;mb=Fa+4160|0;he(a,va,Ba,mb)|0;k[K>>2]=9;Xw(va,601133,K)|0;nb=Fa+4168|0;he(a,va,Ba,nb)|0;k[J>>2]=gb;Xw(ba,601141,J)|0;k[I>>2]=gb;Xw(ca,601148,I)|0;k[H>>2]=gb;Xw(da,601155,H)|0;ob=k[Xa>>2]|0;switch(ob|0){case 17:case 15:case 16:case 14:case 4:case 1:{he(a,ba,Ba,Fa+4104|0)|0;he(a,ca,Ba,Fa+4112|0)|0;Va=97;break}case 2:{he(a,ba,Ba,Fa+4104|0)|0;he(a,ca,Ba,Fa+4112|0)|0;pb=Fa+4120|0;if(+p[pb>>3]==0.0)p[pb>>3]=90.0;he(a,da,Ba,pb)|0;Va=97;break}case 13:{pb=Fa+4104|0;if(+p[pb>>3]==0.0)p[pb>>3]=1.0;he(a,ba,Ba,pb)|0;Va=97;break}case 10:{pb=Fa+4104|0;if(+p[pb>>3]==0.0)p[pb>>3]=1.0;he(a,ba,Ba,pb)|0;pb=Fa+4112|0;if(+p[pb>>3]==0.0)p[pb>>3]=1.0;he(a,ca,Ba,pb)|0;Va=97;break}case 9:{pb=Fa+4104|0;if(+p[pb>>3]==0.0)p[pb>>3]=90.0;he(a,ba,Ba,pb)|0;Va=97;break}case 18:{he(a,ba,Ba,Fa+4104|0)|0;Va=97;break}case 7:{k[G>>2]=gb;k[G+4>>2]=0;Xw(va,601162,G)|0;he(a,va,Ba,U)|0;k[F>>2]=gb;k[F+4>>2]=1;Xw(va,601162,F)|0;he(a,va,Ba,Ya)|0;k[E>>2]=gb;k[E+4>>2]=2;Xw(va,601162,E)|0;he(a,va,Ba,W)|0;k[D>>2]=gb;k[D+4>>2]=3;Xw(va,601162,D)|0;he(a,va,Ba,Za)|0;k[C>>2]=gb;k[C+4>>2]=4;Xw(va,601162,C)|0;he(a,va,Ba,ib)|0;k[B>>2]=gb;k[B+4>>2]=5;Xw(va,601162,B)|0;he(a,va,Ba,jb)|0;k[A>>2]=gb;k[A+4>>2]=6;Xw(va,601162,A)|0;he(a,va,Ba,kb)|0;k[z>>2]=gb;k[z+4>>2]=7;Xw(va,601162,z)|0;he(a,va,Ba,lb)|0;k[y>>2]=gb;k[y+4>>2]=8;Xw(va,601162,y)|0;he(a,va,Ba,mb)|0;k[x>>2]=gb;k[x+4>>2]=9;Xw(va,601162,x)|0;he(a,va,Ba,nb)|0;Va=97;break}default:qb=ob}if((Va|0)==97)qb=k[Xa>>2]|0;do if((qb|0)==31)if(!(Ce(a,Fa)|0)){rb=k[Xa>>2]|0;Va=102;break}else{i[Fa+3374>>0]=65;i[Fa+3375>>0]=78;i[Fa+3383>>0]=65;i[Fa+3384>>0]=78;k[Xa>>2]=3;break}else{rb=qb;Va=102}while(0);b:do if((Va|0)==102){do if((rb|0)==32)if(!(Je(a,Fa)|0)){sb=k[Xa>>2]|0;break}else{i[Fa+3375>>0]=78;i[Fa+3384>>0]=78;k[Xa>>2]=7;break b}else sb=rb;while(0);if((sb|0)!=33)break;i[Fa+3374>>0]=65;i[Fa+3375>>0]=78;i[Fa+3383>>0]=65;i[Fa+3384>>0]=78;k[Xa>>2]=3}while(0);if((k[Na>>2]|0)>0)od(a,Fa,Ba);pd(Fa);vg(Fa,a);k[Ga>>2]=0;k[Ha>>2]=0;Pa=he(a,777897,Ba,ta)|0;ob=he(a,777909,Ba,Ia)|0;nb=he(a,777903,Ba,Ja)|0;mb=he(a,777915,Ba,Ka)|0;do if((k[Na>>2]|0)==2)Va=121;else{lb=$d(a,601170)|0;if(!lb){Va=121;break}k[Xa>>2]=30;kb=Fa+3467|0;i[kb>>0]=i[601176]|0;i[kb+1>>0]=i[601177]|0;i[kb+2>>0]=i[601178]|0;i[kb+3>>0]=i[601179]|0;i[kb+4>>0]=i[601180]|0;i[kb+5>>0]=i[601181]|0;kb=Fa+256|0;jb=0;do{ib=jb;jb=jb+1|0;k[w>>2]=jb;Xw(va,601182,w)|0;Za=kb+(ib<<3)|0;p[Za>>3]=0.0;if(ie(lb,va,Za)|0)k[Ga>>2]=jb}while((jb|0)!=20);jb=$d(a,601189)|0;lb=Fa+416|0;kb=0;do{Za=kb;kb=kb+1|0;k[v>>2]=kb;Xw(va,601195,v)|0;ib=lb+(Za<<3)|0;p[ib>>3]=0.0;if(ie(jb,va,ib)|0)k[Ha>>2]=kb}while((kb|0)!=20);Ae(+p[Aa>>3],+p[V>>3],Fa,na,pa)|0;Ae(+p[Aa>>3],+p[V>>3]+1.0,Fa,oa,qa)|0;cb=+p[qa>>3]-+p[pa>>3];kb=Fa+40|0;p[kb>>3]=cb;jb=Fa+32|0;p[jb>>3]=-cb;k[Fa+3312>>2]=1;zd(Fa);p[Da>>3]=+p[Ma>>3]*3.141592653589793/180.0;Ae(+p[Aa>>3],+p[V>>3],Fa,na,pa)|0;cb=+p[Da>>3];db=+p[Aa>>3]+ +Z(+cb);Ae(db,+p[V>>3]+ +_(+cb),Fa,oa,qa)|0;cb=-+Dd(+p[na>>3],+p[pa>>3],+p[oa>>3],+p[qa>>3]);p[ab>>3]=cb;p[jb>>3]=cb;cb=+p[Da>>3];db=+p[Aa>>3]+ +_(+cb);Ae(db,+p[V>>3]+ +Z(+cb),Fa,oa,qa)|0;cb=+Dd(+p[na>>3],+p[pa>>3],+p[oa>>3],+p[qa>>3]);p[Fa+768>>3]=cb;p[kb>>3]=cb;kb=Fa+56|0;p[kb>>3]=+p[ta>>3];p[Fa+64>>3]=+p[Ia>>3];p[Fa+72>>3]=+p[Ja>>3];p[Fa+80>>3]=+p[Ka>>3];_e(2,kb,Fa+88|0)|0}while(0);c:do if((Va|0)==121){if(ob|Pa|nb|mb){k[La>>2]=1;wd(Fa,ta);break}if(!(he(a,777890,Ba,ra)|0)){p[Fa+32>>3]=1.0;p[Fa+40>>3]=1.0;p[ab>>3]=1.0;p[Fa+768>>3]=1.0;p[Ma>>3]=0.0;k[La>>2]=0;xd(601309);break}he(a,777921,Ba,sa)|0;if(!(+p[ra>>3]==0.0)?(cb=+p[sa>>3],!(+p[Oa>>3]>1.0&cb==0.0)):0)tb=cb;else Va=126;d:do if((Va|0)==126){do if(!($d(a,601202)|0)){if($d(a,601209)|0)break;if($d(a,601218)|0)break;if($d(a,601227)|0)break;if($d(a,601236)|0)break;tb=+p[sa>>3];break d}while(0);p[ma>>3]=0.0;ie(a,601202,ma)|0;cb=+p[ma>>3];if(cb==0.0){ie(a,601209,ma)|0;ub=+p[ma>>3]}else ub=cb;if(!(ub==0.0)){if(+p[ra>>3]==0.0)p[ra>>3]=-ub/3600.0;cb=+p[sa>>3];if(!(cb==0.0)){tb=cb;break}cb=ub/3600.0;p[sa>>3]=cb;tb=cb;break}ie(a,601236,ma)|0;cb=+p[ma>>3];if(cb!=0.0){if(+p[ra>>3]==0.0)p[ra>>3]=-cb/3600.0;cb=+p[sa>>3];if(!(cb==0.0)){tb=cb;break}ie(a,601244,ma)|0;cb=+p[ma>>3]/3600.0;p[sa>>3]=cb;tb=cb;break}ie(a,601227,ma)|0;cb=+p[ma>>3];if(cb!=0.0){if(+p[ra>>3]==0.0)p[ra>>3]=-cb/3600.0;cb=+p[sa>>3];if(!(cb==0.0)){tb=cb;break}ie(a,601252,ma)|0;cb=+p[ma>>3]/3600.0;p[sa>>3]=cb;tb=cb;break}else{ie(a,601218,ma)|0;cb=+p[ma>>3];if(cb!=0.0&+p[ra>>3]==0.0)p[ra>>3]=-cb/3600.0;cb=+p[sa>>3];if(!(cb==0.0)){tb=cb;break}ie(a,601261,ma)|0;cb=+p[ma>>3]/3600.0;p[sa>>3]=cb;tb=cb;break}}while(0);do if(tb==0.0){if(!(+p[Oa>>3]>1.0))break;p[sa>>3]=-+p[ra>>3]}while(0);p[Fa+776>>3]=1.0;p[Fa+784>>3]=1.0;sy(ua|0,0,648)|0;sy(Wa|0,0,648)|0;V=k[ya>>2]|0;if((V|0)>0){kb=0;do{jb=ua+((ha(V,kb)|0)+kb<<3)|0;p[jb>>3]=1.0;kb=kb+1|0}while((kb|0)<(V|0))}do if(!(i[Ba>>0]|0)){if(!(ie(a,601270,ua)|0))break;V=k[ya>>2]|0;if((V|0)>0){kb=V;V=0;jb=0;while(1){k[xa>>2]=0;lb=V;V=V+1|0;if((kb|0)>0){ib=0;Za=jb;while(1){W=ua+(Za<<3)|0;p[W>>3]=(lb|0)==(ib|0)?1.0:0.0;k[u>>2]=V;k[u+4>>2]=ib+1;Xw(va,601279,u)|0;Ya=Za+1|0;ie(a,va,W)|0;ib=(k[xa>>2]|0)+1|0;k[xa>>2]=ib;W=k[ya>>2]|0;if((ib|0)>=(W|0)){vb=W;wb=Ya;break}else Za=Ya}}else{vb=kb;wb=jb}if((V|0)>=(vb|0))break;else{kb=vb;jb=wb}}}yd(Fa,+p[ra>>3],+p[sa>>3],ua);break c}while(0);if(!(he(a,777935,Ba,ua)|0)){p[Da>>3]=0.0;if((gb|0)==2)he(a,777928,Ba,Da)|0;else he(a,601302,Ba,Da)|0;td(Fa,+p[ra>>3],+p[sa>>3],+p[Da>>3]);break}jb=k[ya>>2]|0;if((jb|0)>0){kb=jb;jb=0;V=0;while(1){k[xa>>2]=0;Za=jb;jb=jb+1|0;if((kb|0)>0){ib=0;lb=V;while(1){Ya=ua+(lb<<3)|0;p[Ya>>3]=(Za|0)==(ib|0)?1.0:0.0;k[t>>2]=jb;k[t+4>>2]=ib+1;Xw(va,601292,t)|0;W=lb+1|0;he(a,va,Ba,Ya)|0;ib=(k[xa>>2]|0)+1|0;k[xa>>2]=ib;Ya=k[ya>>2]|0;if((ib|0)>=(Ya|0)){xb=Ya;yb=W;break}else lb=W}}else{xb=kb;yb=V}if((jb|0)>=(xb|0))break;else{kb=xb;V=yb}}}yd(Fa,+p[ra>>3],+p[sa>>3],ua)}while(0);do if((k[Xa>>2]|0)==3){if((k[Qa>>2]|0)!=2)break;mb=Fa+3168|0;nb=k[mb>>2]|0;if(nb){Qe(nb);k[mb>>2]=0}mb=Fa+3172|0;nb=k[mb>>2]|0;if(nb){Qe(nb);k[mb>>2]=0}k[Fa+1560>>2]=0;mb=Fa+1568|0;nb=Fa+4344|0;Pa=0;do{p[mb+(Pa<<3)>>3]=0.0;p[nb+(Pa<<3)>>3]=0.0;Pa=Pa+1|0}while((Pa|0)!=200);k[xa>>2]=0;Pa=0;ob=0;while(1){k[s>>2]=1;k[s+4>>2]=ob;Xw(va,601162,s)|0;if(!(he(a,va,Ba,mb+(k[xa>>2]<<3)|0)|0)){V=k[xa>>2]|0;p[mb+(V<<3)>>3]=0.0;zb=V;Ab=Pa}else{zb=k[xa>>2]|0;Ab=Pa+1|0}ob=zb+1|0;k[xa>>2]=ob;if((ob|0)>=100){Bb=Ab;break}else Pa=Ab}k[xa>>2]=0;Pa=Bb;ob=0;while(1){k[d>>2]=2;k[d+4>>2]=ob;Xw(va,601162,d)|0;if(!(he(a,va,Ba,mb+((k[xa>>2]|0)+100<<3)|0)|0)){V=k[xa>>2]|0;p[mb+(V+100<<3)>>3]=0.0;Cb=V;Db=Pa}else{Cb=k[xa>>2]|0;Db=Pa+1|0}ob=Cb+1|0;k[xa>>2]=ob;if((ob|0)>=100){Eb=Db;break}else Pa=Db}if((Eb|0)<=0)break;if((k[Fa+5968>>2]|0)==1)break;Pa=Fa+3948|0;ob=Fa+3944|0;V=100;kb=0;while(1){cb=+p[mb+(((k[Pa>>2]|0)*100|0)+V<<3)>>3];p[nb+(V<<3)>>3]=cb;db=+p[mb+(((k[ob>>2]|0)*100|0)+V<<3)>>3];p[nb+(V+100<<3)>>3]=db;if(!kb)Fb=db!=0.0|cb!=0.0?V+1|0:0;else Fb=kb;if((V|0)>0){V=V+-1|0;kb=Fb}else break}nd(Fa);k[b>>2]=0}while(0);Xa=Fa+3467|0;if(!((Ax(Xa,601892,6)|0)!=0?(Ax(Xa,601899,5)|0)!=0:0)){k[Fa+3292>>2]=-1;k[Fa+3288>>2]=5}do if(!(ie(a,758550,Ca)|0)){Xa=Fa+128|0;if(!(je(a,601337,Xa)|0)){if(je(a,601346,Xa)|0)break;if(ie(a,758683,Xa)|0)break;p[Xa>>3]=+p[Fa+120>>3];break}me(a,601337,32,aa)|0;if(rx(aa,84)|0)break;if(ie(a,601351,Ea)|0){p[Xa>>3]=+p[Xa>>3]+ +p[Ea>>3]/8765.812770744;break}if(!(ie(a,601354,Ea)|0))break;p[Xa>>3]=+p[Xa>>3]+ +p[Ea>>3]/8765.812770744}else p[Fa+128>>3]=(+p[Ca>>3]+-15019.81352)/365.242198781+1900.0;while(0);k[Fa+3312>>2]=1;eb=Aa}while(0);k[Fa+3964>>2]=eb;k[Fa+3972>>2]=ab;k[Fa+3968>>2]=Wa;k[Fa+3284>>2]=1;k[Fa+3296>>2]=0;k[Fa+3328>>2]=0;vd(Fa);Ra=Fa;r=c;return Ra|0}function kd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;c=r;r=r+96|0;d=c;e=c+16|0;if(!b){f=0;r=c;return f|0}g=xx(b)|0;h=ey(1,g+1|0)|0;if((g|0)>0){j=0;do{l=i[b+j>>0]|0;if((l+-97&255)<26)m=(l&255)+224&255;else m=l;i[h+j>>0]=m;j=j+1|0}while((j|0)!=(g|0))}i[h+g>>0]=0;if((xx(h)|0)==1){f=i[h>>0]|0;r=c;return f|0}g=d;k[g>>2]=1314079575;k[g+4>>2]=4541761;i[d+8>>0]=0;g=d+7|0;j=95;m=0;while(1){b=(m|0)>0?m+64&255:0;i[g>>0]=b;if(!(me(a,d,72,e)|0))n=j;else{l=xx(e)|0;o=ey(1,l+1|0)|0;if((l|0)>0){p=0;do{q=i[e+p>>0]|0;if((q+-97&255)<26)s=(q&255)+224&255;else s=q;i[o+p>>0]=s;p=p+1|0}while((p|0)!=(l|0))}i[o+l>>0]=0;p=(tx(o,h)|0)==0;dy(o);n=p?b:j}m=m+1|0;if((m|0)==27){t=n;break}else j=n}dy(h);f=t;r=c;return f|0}function ld(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;i[d>>0]=0;_d(a,b)|0;b=jd(a,d)|0;r=c;return b|0}function md(a){a=a|0;var b=0,c=0,d=0;b=r;r=r+16|0;c=b;i[c>>0]=0;d=jd(a,c)|0;r=b;return d|0}function nd(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;b=r;r=r+192|0;c=b+104|0;d=b+96|0;e=b+88|0;f=b+72|0;g=b+56|0;h=b+40|0;i=b+32|0;j=b+108|0;l=b+112|0;m=b+24|0;n=b+16|0;o=b+8|0;q=b;s=i;k[s>>2]=1;k[s+4>>2]=1;s=k[a+3944>>2]|0;t=k[a+3948>>2]|0;if((k[a+3316>>2]|0)!=2){r=b;return}u=a+3928|0;if(tx(u,606498)|0){r=b;return}if(+p[a+1568+((s*100|1)<<3)>>3]==0.0?+p[a+1568+((t*100|1)<<3)>>3]==0.0:0){r=b;return}v=a+9312|0;w=k[v>>2]|0;if(!w){p[m>>3]=0.0;p[n>>3]=0.0;p[o>>3]=+p[a+136>>3];p[q>>3]=+p[a+144>>3]}else{Ad(w,0.0,0.0,m,n);Ad(k[v>>2]|0,+p[a+136>>3],+p[a+144>>3],o,q)}if(!s){x=+p[m>>3];y=+p[n>>3];z=y;A=+p[q>>3]-y;B=x;C=+p[o>>3]-x}else{x=+p[n>>3];y=+p[m>>3];z=y;A=+p[o>>3]-y;B=x;C=+p[q>>3]-x}x=A/11.0;A=C/11.0;q=ey(288,8)|0;o=ey(144,8)|0;m=ey(144,8)|0;C=z+.5;n=h+(t<<3)|0;p[n>>3]=C;v=g+(t<<3)|0;p[v>>3]=C;C=B+.5;w=h+(s<<3)|0;p[w>>3]=C;D=g+(s<<3)|0;p[D>>3]=C;E=a+3956|0;F=f+(s<<3)|0;s=f+(t<<3)|0;t=a+4064|0;B=C;G=11;H=m;I=o;J=q;while(1){p[D>>3]=B;K=H+96|0;L=11;M=H;N=I;O=J;while(1){if(af(g,E,f)|0){k[e>>2]=u;Xw(l,601716,e)|0;xd(l)}p[N>>3]=+p[F>>3];p[M>>3]=+p[s>>3];tf(t,+p[F>>3],+p[s>>3],O,O+8|0)|0;p[D>>3]=A+ +p[D>>3];if(!L)break;else{L=L+-1|0;M=M+8|0;N=N+8|0;O=O+16|0}}p[v>>3]=x+ +p[v>>3];if(!G)break;B=+p[w>>3];G=G+-1|0;H=K;I=I+96|0;J=J+192|0}af(h,E,f)|0;p[F>>3]=+p[F>>3]+2.777777777777778e-04;$e(f,E,g)|0;B=+p[D>>3]-+p[w>>3];x=+p[v>>3]-+p[n>>3];A=+X(+(B*B+x*x))*3600.0;if(!(A!=0.0)){k[d>>2]=u;Xw(l,601716,d)|0;xd(l)}x=.04/A;k[j>>2]=1;d=0;F=1;a:while(1){if((F|0)>1)Qe(d);J=Pe(i,2,j,1)|0;Se(J,q,o,0,144,0);I=143;H=o;G=q;while(1){A=+Re(J,G);if(+W(+(A-+p[H>>3]))>x)break;if(!I){P=J;Q=27;break a}else{I=I+-1|0;H=H+8|0;G=G+16|0}}F=(k[j>>2]|0)+1|0;k[j>>2]=F;if((F|0)>=10){R=J;Q=28;break}else d=J}if((Q|0)==27){k[j>>2]=(k[j>>2]|0)+1;S=P}else if((Q|0)==28){xd(601759);k[a+1560>>2]=1;S=R}k[a+3168>>2]=S;k[a+5944>>2]=S;af(h,E,f)|0;p[s>>3]=+p[s>>3]+2.777777777777778e-04;$e(f,E,g)|0;x=+p[D>>3]-+p[w>>3];A=+p[v>>3]-+p[n>>3];B=+X(+(x*x+A*A))*3600.0;if(!(B!=0.0)){k[c>>2]=u;Xw(l,601716,c)|0;xd(l)}A=.04/B;k[j>>2]=1;l=S;S=1;b:while(1){if((S|0)>1)Qe(l);c=Pe(i,2,j,1)|0;Se(c,q,m,0,144,0);u=143;n=m;v=q;while(1){B=+Re(c,v);if(+W(+(B-+p[n>>3]))>A)break;if(!u){T=c;Q=38;break b}else{u=u+-1|0;n=n+8|0;v=v+16|0}}S=(k[j>>2]|0)+1|0;k[j>>2]=S;if((S|0)>=10){U=c;Q=39;break}else l=c}if((Q|0)==38){k[j>>2]=(k[j>>2]|0)+1;V=T}else if((Q|0)==39){xd(601759);k[a+1560>>2]=1;V=U}k[a+3172>>2]=V;k[a+5948>>2]=V;dy(q);dy(o);dy(m);r=b;return}function od(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0.0,x=0,y=0,z=0,A=0,B=0,C=0.0,D=0,E=0,F=0;d=r;r=r+160|0;e=d+32|0;f=d+24|0;g=d+72|0;h=d+40|0;j=d+128|0;l=d+112|0;m=d+8|0;n=d+80|0;o=d;k[g>>2]=0;i[h>>0]=0;i[j>>0]=0;q=i[c>>0]|0;if(!(q<<24>>24)){s=m;k[s>>2]=1230328133;k[s+4>>2]=5787470;t=l;u=758716;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0))}else{k[f>>2]=q<<24>>24;Xw(m,601821,f)|0;k[e>>2]=i[c>>0];Xw(l,601831,e)|0}if((me(a,m,31,j)|0)==0?(me(a,601450,31,j)|0)!=0:0){e=m;k[e>>2]=1230328133;k[e+4>>2]=5787470}if((me(a,l,31,h)|0)==0?(me(a,758716,31,h)|0)!=0:0){t=l;u=758716;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0))}a:do switch(i[j>>0]|0){case 74:{e=j+1|0;w=+dx(e);p[b+120>>3]=w;c=ex(e)|0;k[g>>2]=c;k[h>>2]=3492678;x=c;y=0;z=31;break}case 66:{c=j+1|0;w=+dx(c);p[b+120>>3]=w;e=~~+dx(c);k[g>>2]=e;k[h>>2]=3427142;x=e;y=0;z=31;break}default:{do if(!(be(a,m,g)|0)){if(be(a,758683,g)|0)if(!(k[g>>2]|0)){k[g>>2]=1950;p[b+120>>3]=1950.0;A=0;break}else{ie(a,758683,b+120|0)|0;A=1;break}if(i[h>>0]|0){if(!(Ax(h,602314,3)|0)){p[b+120>>3]=1950.0;k[g>>2]=1950;B=0;break a}if(!(Ax(h,602293,4)|0)){p[b+120>>3]=2.0e3;k[g>>2]=2e3;B=0;break a}if(!(Ax(h,602298,3)|0)){p[b+120>>3]=2.0e3;k[g>>2]=2e3;B=0;break a}if(!(Ax(h,601842,3)|0)){p[b+120>>3]=2.0e3;k[g>>2]=2e3;B=0;break a}if(!(Ax(h,601846,3)|0)){p[b+120>>3]=2.0e3;k[g>>2]=2e3;B=0;break a}else A=0}else A=0}else{ie(a,m,b+120|0)|0;A=1}while(0);x=k[g>>2]|0;y=A;z=31}}while(0);do if((z|0)==31)if(!x){p[b+120>>3]=2.0e3;k[g>>2]=2e3;A=b+3449|0;if((Ax(A,606317,2)|0)!=0?(Ax(A,601933,3)|0)!=0:0){B=y;break}k[h>>2]=3492678;B=y}else B=y;while(0);y=b+128|0;do if(!(je(a,601337,y)|0))if((je(a,601346,y)|0)==0?(ie(a,758683,y)|0)==0:0){w=+p[b+120>>3];p[y>>3]=w;C=w}else z=44;else{me(a,601337,32,n)|0;if(!(rx(n,84)|0)){if(ie(a,601351,o)|0){w=+p[y>>3]+ +p[o>>3]/8765.812770744;p[y>>3]=w;C=w;break}if(ie(a,601354,o)|0){w=+p[y>>3]+ +p[o>>3]/8765.812770744;p[y>>3]=w;C=w}else z=44}else z=44}while(0);if((z|0)==44)C=+p[y>>3];if(C==0.0)p[y>>3]=+p[b+120>>3];do if((i[h>>0]|0)==0?(me(a,l,31,h)|0,(i[h>>0]|0)==0):0){if((k[b+3884>>2]|0)!=7){y=b+3764|0;if((k[g>>2]|0)>1980){i[y>>0]=70;i[y+1>>0]=75;i[y+2>>0]=53;i[y+3>>0]=0;break}else{i[y>>0]=70;i[y+1>>0]=75;i[y+2>>0]=52;i[y+3>>0]=0;break}}}else z=49;while(0);do if((z|0)==49?(l=b+3764|0,ux(l,h)|0,(B|0)==0):0){if(!(Ax(l,602314,3)|0)){p[b+120>>3]=1950.0;break}if(!(Ax(l,602298,3)|0)){p[b+120>>3]=2.0e3;break}if(!(Ax(l,602293,4)|0)){p[b+120>>3]=2.0e3;break}a=Ax(l,601842,3)|0;if(!(k[g>>2]|a))p[b+120>>3]=2.0e3}while(0);switch(i[b+3449>>0]|0){case 71:{g=b+3764|0;t=g;u=601850;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}case 69:{g=b+3764|0;t=g;u=601859;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}case 83:{g=b+3764|0;t=g;u=601868;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}case 72:{g=b+3764|0;t=g;u=601877;v=t+9|0;do{i[t>>0]=i[u>>0]|0;t=t+1|0;u=u+1|0}while((t|0)<(v|0));D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}case 65:{g=b+3764|0;i[g>>0]=i[601886]|0;i[g+1>>0]=i[601887]|0;i[g+2>>0]=i[601888]|0;i[g+3>>0]=i[601889]|0;i[g+4>>0]=i[601890]|0;i[g+5>>0]=i[601891]|0;D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}case 76:{g=b+3764|0;i[g>>0]=i[601892]|0;i[g+1>>0]=i[601893]|0;i[g+2>>0]=i[601894]|0;i[g+3>>0]=i[601895]|0;i[g+4>>0]=i[601896]|0;i[g+5>>0]=i[601897]|0;i[g+6>>0]=i[601898]|0;D=g;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}default:{D=b+3764|0;E=Md(D)|0;F=b+3884|0;k[F>>2]=E;r=d;return}}}function pd(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0;b=a+3764|0;c=k[a+3260>>2]|0;if(i[b>>0]|0)if(!c){d=0;e=3}else f=c;else{d=c;e=3}if((e|0)==3){i[b>>0]=i[601892]|0;i[b+1>>0]=i[601893]|0;i[b+2>>0]=i[601894]|0;i[b+3>>0]=i[601895]|0;i[b+4>>0]=i[601896]|0;i[b+5>>0]=i[601897]|0;i[b+6>>0]=i[601898]|0;f=d}if((f|0)==-1){i[b>>0]=i[601899]|0;i[b+1>>0]=i[601900]|0;i[b+2>>0]=i[601901]|0;i[b+3>>0]=i[601902]|0;i[b+4>>0]=i[601903]|0;i[b+5>>0]=i[601904]|0}f=Md(b)|0;k[a+3884>>2]=f;do if((f|0)!=2){d=a+3796|0;if((f|0)==1){i[d>>0]=70;i[d+1>>0]=75;i[d+2>>0]=53;i[d+3>>0]=0;g=d;break}else{ux(d,b)|0;g=d;break}}else{d=a+3796|0;i[d>>0]=70;i[d+1>>0]=75;i[d+2>>0]=52;i[d+3>>0]=0;g=d}while(0);f=Md(g)|0;k[a+3888>>2]=f;f=a+120|0;p[a+3872>>3]=+p[f>>3];g=a+3828|0;ux(g,b)|0;b=Md(g)|0;k[a+3880>>2]=b;p[a+3864>>3]=+p[f>>3];return}function qd(a){a=a|0;var b=0,c=0;if(!a)return;if(!(k[a+3312>>2]|0)){dy(a);return}b=a+9312|0;c=k[b>>2]|0;if(c){qd(c);k[b>>2]=0}rd(a);b=k[a+9320>>2]|0;if(b)dy(b);b=k[a+3980>>2]|0;if(b)dy(b);b=k[a+3976>>2]|0;if(b)dy(b);b=k[a+3168>>2]|0;if(b)Qe(b);b=k[a+3172>>2]|0;if(b)Qe(b);dy(a);return}function rd(a){a=a|0;var b=0;b=k[25987]|0;if(b){dy(b);k[25987]=0}b=k[25988]|0;if(b){dy(b);k[25988]=0}b=k[25989]|0;if(b){dy(b);k[25989]=0}b=k[25990]|0;if(b){dy(b);k[25990]=0}b=k[25991]|0;if(b){dy(b);k[25991]=0}b=k[25992]|0;if(b){dy(b);k[25992]=0}b=k[25993]|0;if(b){dy(b);k[25993]=0}b=k[25994]|0;if(b){dy(b);k[25994]=0}b=k[25995]|0;if(b){dy(b);k[25995]=0}b=k[25996]|0;if(b){dy(b);k[25996]=0}if(!a)return;if(!(k[a+3312>>2]|0))return;b=k[a+9208>>2]|0;if(b)dy(b);b=k[a+9212>>2]|0;if(b)dy(b);b=k[a+9216>>2]|0;if(b)dy(b);b=k[a+9220>>2]|0;if(b)dy(b);b=k[a+9224>>2]|0;if(b)dy(b);b=k[a+9228>>2]|0;if(b)dy(b);b=k[a+9232>>2]|0;if(b)dy(b);b=k[a+9236>>2]|0;if(b)dy(b);b=k[a+9240>>2]|0;if(b)dy(b);b=k[a+9244>>2]|0;if(!b)return;dy(b);return}function sd(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;d=r;r=r+160|0;e=d+8|0;f=d;g=d+16|0;k[g>>2]=5130572;k[g+4>>2]=5265985;k[g+8>>2]=5266003;k[g+12>>2]=5128532;k[g+16>>2]=5130579;k[g+20>>2]=4674643;k[g+24>>2]=4411969;k[g+28>>2]=5132378;k[g+32>>2]=4277594;k[g+36>>2]=5392705;k[g+40>>2]=5265731;k[g+44>>2]=5390659;k[g+48>>2]=5391693;k[g+52>>2]=4277571;k[g+56>>2]=5263171;k[g+60>>2]=4476739;k[g+64>>2]=4542275;k[g+68>>2]=5197635;k[g+72>>2]=5132098;k[g+76>>2]=5194576;k[g+80>>2]=4998739;k[g+84>>2]=5390672;k[g+88>>2]=5523777;k[g+92>>2]=5001037;k[g+96>>2]=4412227;k[g+100>>2]=4412241;k[g+104>>2]=4412244;k[g+108>>2]=5260110;k[g+112>>2]=5459015;k[g+116>>2]=5460804;k[g+120>>2]=5524560;k[g+124>>2]=5787220;k[g+128>>2]=5787738;k[g+132>>2]=5656660;if(!(Ax(b,746435,4)|0)){i[b>>0]=88;i[b+1>>0]=76;i[b+2>>0]=79;i[b+3>>0]=78}h=a+3368|0;ux(h,b)|0;j=a+3449|0;ux(j,b)|0;l=a+3467|0;ux(l,b)|0;a:do if(!(Ax(b,601892,6)|0))k[a+3260>>2]=0;else{if(!(Ax(b,601899,6)|0)){k[a+3260>>2]=-1;break}if(ne(b,601905)|0){k[a+3260>>2]=-1;break}m=i[b>>0]|0;switch(m<<24>>24){case 65:case 68:case 82:break;default:if((i[b+1>>0]|0)!=76){k[a+3260>>2]=0;r=d;return 0}}i[j>>0]=m;i[a+3450>>0]=i[b+1>>0]|0;m=i[b+2>>0]|0;n=a+3451|0;do if(m<<24>>24!=45){i[n>>0]=m;o=i[b+3>>0]|0;p=a+3452|0;if(o<<24>>24==45){i[p>>0]=0;q=4;break}else{i[p>>0]=o;i[a+3453>>0]=0;q=4;break}}else{i[n>>0]=0;q=3}while(0);n=((i[b+q>>0]|0)==45&1)+q|0;m=n+((i[b+n>>0]|0)==45&1)|0;n=m+((i[b+m>>0]|0)==45&1)|0;m=n+((i[b+n>>0]|0)==45&1)|0;i[l>>0]=i[b+m>>0]|0;i[a+3468>>0]=i[b+(m+1)>>0]|0;i[a+3469>>0]=i[b+(m+2)>>0]|0;i[a+3470>>0]=0;k[f>>2]=j;k[f+4>>2]=l;Xw(h,601909,f)|0;if((i[h>>0]|0)==32)i[h>>0]=45;m=a+3369|0;if((i[m>>0]|0)==32)i[m>>0]=45;m=a+3370|0;if((i[m>>0]|0)==32)i[m>>0]=45;m=a+3371|0;if((i[m>>0]|0)==32)i[m>>0]=45;m=a+3372|0;if((i[m>>0]|0)==32)i[m>>0]=45;m=a+3373|0;if((i[m>>0]|0)==32)i[m>>0]=45;m=a+3374|0;if((i[m>>0]|0)==32)i[m>>0]=45;n=a+3375|0;if((i[n>>0]|0)==32)i[n>>0]=45;o=a+3260|0;k[o>>2]=0;p=0;s=1;while(1){if(!(Ax(l,g+(s<<2)|0,3)|0)){k[o>>2]=s;t=s}else t=p;s=s+1|0;if((s|0)==34){u=t;break}else p=t}p=a+3324|0;switch(k[p>>2]|0){case 0:{k[p>>2]=3;break a;break}case 1:{k[p>>2]=2;break}case 2:break;default:break a}switch(u|0){case 31:{i[m>>0]=65;i[n>>0]=78;k[o>>2]=3;break a;break}case 32:{i[m>>0]=80;i[n>>0]=78;k[o>>2]=7;break a;break}default:break a}}while(0);do if(Ax(c,601917,4)|0)if(!(Ax(c,601928,4)|0)){i[c>>0]=i[b>>0]|0;u=c+1|0;i[u>>0]=i[777971]|0;i[u+1>>0]=i[777972]|0;i[u+2>>0]=i[777973]|0;k[a+3264>>2]=-90;u=a+3764|0;i[u>>0]=83;i[u+1>>0]=80;i[u+2>>0]=65;i[u+3>>0]=0;k[a+3884>>2]=8;break}else{k[a+3264>>2]=0;break}else{i[c>>0]=i[b>>0]|0;u=c+1|0;i[u>>0]=i[777971]|0;i[u+1>>0]=i[777972]|0;i[u+2>>0]=i[777973]|0;k[a+3264>>2]=90;u=a+3764|0;i[u>>0]=i[601922]|0;i[u+1>>0]=i[601923]|0;i[u+2>>0]=i[601924]|0;i[u+3>>0]=i[601925]|0;i[u+4>>0]=i[601926]|0;i[u+5>>0]=i[601927]|0;k[a+3884>>2]=7}while(0);u=a+3377|0;ux(u,c)|0;t=a+3458|0;ux(t,c)|0;b:do if(Ax(c,601892,6)|0){if(!(Ax(c,601899,6)|0)){k[a+3260>>2]=-1;break}g=i[c>>0]|0;switch(g<<24>>24){case 65:case 68:case 82:break;default:if((i[c+1>>0]|0)!=76){k[a+3260>>2]=0;break b}}i[t>>0]=g;g=c+1|0;i[a+3459>>0]=i[g>>0]|0;h=i[c+2>>0]|0;f=a+3460|0;do if(h<<24>>24!=45){i[f>>0]=h;j=i[c+3>>0]|0;q=a+3461|0;if(j<<24>>24==45){i[q>>0]=0;v=4;break}else{i[q>>0]=j;i[a+3462>>0]=0;v=4;break}}else{i[f>>0]=0;v=3}while(0);f=((i[c+v>>0]|0)==45&1)+v|0;h=f+((i[c+f>>0]|0)==45&1)|0;f=h+((i[c+h>>0]|0)==45&1)|0;h=f+((i[c+f>>0]|0)==45&1)|0;i[l>>0]=i[c+h>>0]|0;i[a+3468>>0]=i[c+(h+1)>>0]|0;i[a+3469>>0]=i[c+(h+2)>>0]|0;i[a+3470>>0]=0;if((Ax(b,601933,3)|0)!=0?(Ax(b+1|0,777971,3)|0)!=0:0)k[a+3304>>2]=0;else k[a+3304>>2]=1;if((i[g>>0]|0)!=76?(i[c>>0]|0)!=65:0){k[a+3292>>2]=0;k[a+3288>>2]=3}else{k[a+3292>>2]=1;k[a+3288>>2]=5}k[e>>2]=t;k[e+4>>2]=l;Xw(u,601909,e)|0;if((i[u>>0]|0)==32)i[u>>0]=45;h=a+3378|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3379|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3380|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3381|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3382|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3383|0;if((i[h>>0]|0)==32)i[h>>0]=45;h=a+3384|0;if((i[h>>0]|0)==32)i[h>>0]=45}else k[a+3260>>2]=0;while(0);yg(a,b);r=d;return 0}function td(a,b,c,d){a=a|0;b=+b;c=+c;d=+d;var e=0,f=0,g=0,h=0.0,i=0,j=0,l=0,m=0,n=0.0,o=0.0,q=0.0;e=k[a+3316>>2]|0;f=a+760|0;p[f>>3]=b;g=a+768|0;h=c!=0.0?c:b;p[g>>3]=h;p[a+32>>3]=b;p[a+40>>3]=h;if((e|0)>0){i=(e|0)<2?e:2;e=0;j=a+832|0;while(1){l=0;m=j;while(1){p[m>>3]=(e|0)==(l|0)?1.0:0.0;l=l+1|0;if((l|0)==(i|0))break;else m=m+8|0}e=e+1|0;if((e|0)==(i|0))break;else j=j+(i<<3)|0}n=+p[f>>3];o=+p[g>>3]}else{n=b;o=h}k[a+3300>>2]=0;i=a+48|0;h=d<0.0?d+360.0:d;d=!(h>=360.0)?h:h+-360.0;p[i>>3]=d;h=d*3.141592653589793/180.0;d=+Z(+h);q=+_(+(b*c>0.0?-h:h));j=a+56|0;p[j>>3]=d*n;h=q*+W(+o);p[a+64>>3]=n<0.0?-h:h;h=q*+W(+n);p[a+72>>3]=o<0.0?h:-h;p[a+80>>3]=d*o;_e(2,j,a+88|0)|0;Kd(a);o=+p[f>>3];f=o<0.0;if(!(k[a+3304>>2]|0)){if(f?+p[g>>3]>0.0:0){k[a+3256>>2]=0;d=+p[i>>3];p[a+3200>>3]=d;h=d+90.0;p[a+3208>>3]=h>180.0?h+-360.0:h;h=d+180.0;j=a+3216|0;p[j>>3]=h;if(!(h>180.0))return;p[j>>3]=h+-360.0;return}do if(o>0.0){if(!(+p[g>>3]<0.0)){if(!(+p[g>>3]>0.0))break;k[a+3256>>2]=1;h=+p[i>>3];p[a+3200>>3]=-h;d=90.0-h;p[a+3208>>3]=d>180.0?d+-360.0:d;p[a+3216>>3]=h;return}k[a+3256>>2]=0;h=+p[i>>3]+180.0;d=h>180.0?h+-360.0:h;p[a+3200>>3]=d;h=d+90.0;p[a+3208>>3]=h>180.0?h+-360.0:h;h=d+180.0;j=a+3216|0;p[j>>3]=h;if(!(h>180.0))return;p[j>>3]=h+-360.0;return}while(0);if(!f)return;if(!(+p[g>>3]<0.0))return;k[a+3256>>2]=1;h=+p[i>>3];d=h+180.0;n=d>180.0?d+-360.0:d;p[a+3200>>3]=n;d=n+90.0;p[a+3208>>3]=d>180.0?d+-360.0:d;d=h+90.0;j=a+3216|0;p[j>>3]=d;if(!(d>180.0))return;p[j>>3]=d+-360.0;return}if(f?+p[g>>3]>0.0:0){k[a+3256>>2]=1;d=+p[i>>3];h=d+-90.0;j=h<-180.0;n=h+360.0;p[a+3200>>3]=j?n:h;p[a+3208>>3]=d;e=a+3216|0;p[e>>3]=h;if(!j)return;p[e>>3]=n;return}if(o>0.0){if(+p[g>>3]<0.0){k[a+3256>>2]=1;o=+p[i>>3];n=o+90.0;p[a+3200>>3]=n>180.0?n+-360.0:n;p[a+3208>>3]=o;n=o+-90.0;e=a+3216|0;p[e>>3]=n;if(!(n<-180.0))return;p[e>>3]=n+360.0;return}if(+p[g>>3]>0.0){k[a+3256>>2]=0;n=+p[i>>3]+90.0;e=n>180.0;o=n+-360.0;h=e?o:n;p[a+3200>>3]=h;p[a+3208>>3]=h;j=a+3216|0;p[j>>3]=n;if(!e)return;p[j>>3]=o;return}}if(!f)return;if(!(+p[g>>3]<0.0))return;k[a+3256>>2]=0;o=+p[i>>3];n=o+-90.0;h=n<-180.0?n+360.0:n;p[a+3200>>3]=h;p[a+3208>>3]=h;h=o+90.0;i=a+3216|0;p[i>>3]=h;if(!(h>180.0))return;p[i>>3]=h+-360.0;return}function ud(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0.0,l=0,m=0;c=r;r=r+16|0;d=c+8|0;e=c;if(!a){r=c;return}f=a+3312|0;if(!(k[f>>2]|0)){r=c;return}do if((((b|0)!=0?(i[b>>0]|0)!=0:0)?(tx(b,601937)|0)!=0:0)?(tx(b,601943)|0)!=0:0){g=Md(b)|0;if((g|0)<0){r=c;return}h=k[a+3884>>2]|0;if((g|0)!=(h|0)&(h+-5|0)>>>0<2){r=c;return}else{ux(a+3796|0,b)|0;j=+Nd(b);p[a+3872>>3]=j;l=g;break}}else m=7;while(0);a:do if((m|0)==7){b=k[a+3884>>2]|0;g=a+3796|0;ux(g,a+3764|0)|0;j=+p[a+120>>3];p[a+3872>>3]=j;switch(b|0){case 2:{if(!(j!=1950.0)){i[g>>0]=i[601954]|0;i[g+1>>0]=i[601955]|0;i[g+2>>0]=i[601956]|0;i[g+3>>0]=i[601957]|0;i[g+4>>0]=i[601958]|0;i[g+5>>0]=i[601959]|0;l=2;break a}i[g>>0]=66;p[e>>3]=j;Xw(a+3797|0,601949,e)|0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)==48)i[h>>0]=0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)==48)i[h>>0]=0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)!=48){l=2;break a}i[h>>0]=0;l=2;break a;break}case 1:{if(!(j!=2.0e3)){i[g>>0]=i[601960]|0;i[g+1>>0]=i[601961]|0;i[g+2>>0]=i[601962]|0;i[g+3>>0]=i[601963]|0;i[g+4>>0]=i[601964]|0;i[g+5>>0]=i[601965]|0;l=1;break a}i[g>>0]=74;p[d>>3]=j;Xw(a+3797|0,601949,d)|0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)==48)i[h>>0]=0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)==48)i[h>>0]=0;h=(xx(g)|0)+-1+(a+3796)|0;if((i[h>>0]|0)!=48){l=1;break a}i[h>>0]=0;l=1;break a;break}default:{l=b;break a}}}while(0);k[a+3888>>2]=l;if(!(k[f>>2]|0)){r=c;return}switch(l|0){case 9:case 4:case 3:{k[a+3292>>2]=1;k[a+3288>>2]=5;r=c;return}case 5:{k[a+3292>>2]=1;k[a+3288>>2]=5;r=c;return}case 8:case 7:{k[a+3292>>2]=1;k[a+3288>>2]=5;r=c;return}default:{k[a+3292>>2]=0;k[a+3288>>2]=3;r=c;return}}}function vd(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;b=r;r=r+32|0;c=b;d=b+8|0;if(!a){r=b;return}e=a+3312|0;if(!(k[e>>2]|0)){r=b;return}f=a+9212|0;g=a+9216|0;h=a+9220|0;j=a+9224|0;l=a+9228|0;m=0;a:while(1){if(!m){n=d;o=601966;p=n+12|0;do{i[n>>0]=i[o>>0]|0;n=n+1|0;o=o+1|0}while((n|0)<(p|0))}else{k[c>>2]=m;Xw(d,601978,c)|0}o=k[103948+(m<<2)>>2]|0;b:do if(o){if((k[e>>2]|0)!=0?(n=xx(o)|0,(n|0)>0):0){p=a+9208+(m<<2)|0;q=k[p>>2]|0;if(q)dy(q);q=ey(n+2|0,1)|0;k[p>>2]=q;if(q){s=q;q=0;while(1){t=i[o+q>>0]|0;i[s+q>>0]=t<<24>>24==95?32:t;q=q+1|0;t=k[p>>2]|0;if((q|0)==(n|0)){u=t;break}else s=t}i[u+n>>0]=0}}}else{s=jb(d|0)|0;if(s){if(!(k[e>>2]|0))break;q=xx(s)|0;if((q|0)<=0)break;p=a+9208+(m<<2)|0;t=k[p>>2]|0;if(t)dy(t);t=ey(q+2|0,1)|0;k[p>>2]=t;if(!t)break;else{v=t;w=0}while(1){t=i[s+w>>0]|0;i[v+w>>0]=t<<24>>24==95?32:t;w=w+1|0;t=k[p>>2]|0;if((w|0)==(q|0)){x=t;break}else v=t}i[x+q>>0]=0;break}switch(m|0){case 1:{if(k[e>>2]|0){p=k[f>>2]|0;if(p)dy(p);p=ey(13,1)|0;k[f>>2]=p;if(p){i[p>>0]=115;i[p+1>>0]=117;i[p+2>>0]=97;i[p+3>>0]=50;i[p+4>>0]=32;i[p+5>>0]=45;i[p+6>>0]=97;i[p+7>>0]=104;i[p+8>>0]=32;i[p+9>>0]=37;i[p+10>>0]=115;i[p+11>>0]=0}}break}case 2:{if(k[e>>2]|0){p=k[g>>2]|0;if(p)dy(p);p=ey(13,1)|0;k[g>>2]=p;if(p){i[p>>0]=115;i[p+1>>0]=103;i[p+2>>0]=115;i[p+3>>0]=99;i[p+4>>0]=32;i[p+5>>0]=45;i[p+6>>0]=97;i[p+7>>0]=104;i[p+8>>0]=32;i[p+9>>0]=37;i[p+10>>0]=115;i[p+11>>0]=0}}break}case 3:{if(k[e>>2]|0){p=k[h>>2]|0;if(p)dy(p);p=ey(13,1)|0;k[h>>2]=p;if(p){i[p>>0]=115;i[p+1>>0]=116;i[p+2>>0]=121;i[p+3>>0]=50;i[p+4>>0]=32;i[p+5>>0]=45;i[p+6>>0]=97;i[p+7>>0]=104;i[p+8>>0]=32;i[p+9>>0]=37;i[p+10>>0]=115;i[p+11>>0]=0}}break}case 4:{if(k[e>>2]|0){p=k[j>>2]|0;if(p)dy(p);p=ey(13,1)|0;k[j>>2]=p;if(p){i[p>>0]=115;i[p+1>>0]=112;i[p+2>>0]=112;i[p+3>>0]=109;i[p+4>>0]=32;i[p+5>>0]=45;i[p+6>>0]=97;i[p+7>>0]=104;i[p+8>>0]=32;i[p+9>>0]=37;i[p+10>>0]=115;i[p+11>>0]=0}}break}case 5:{if(k[e>>2]|0){p=k[l>>2]|0;if(p)dy(p);p=ey(13,1)|0;k[l>>2]=p;if(p){i[p>>0]=115;i[p+1>>0]=115;i[p+2>>0]=97;i[p+3>>0]=111;i[p+4>>0]=32;i[p+5>>0]=45;i[p+6>>0]=97;i[p+7>>0]=104;i[p+8>>0]=32;i[p+9>>0]=37;i[p+10>>0]=115;i[p+11>>0]=0}}break}default:{k[a+9208+(m<<2)>>2]=0;break b}}m=m+1|0;continue a}while(0);m=m+1|0;if((m|0)==10)break}r=b;return}function wd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,i=0.0;if(!b)return;k[a+3300>>2]=1;c=a+56|0;p[c>>3]=+p[b>>3];d=b+8|0;p[a+64>>3]=+p[d>>3];e=b+16|0;p[a+72>>3]=+p[e>>3];f=b+24|0;p[a+80>>3]=+p[f>>3];_e(2,c,a+88|0)|0;g=+p[b>>3];h=+p[e>>3];i=+X(+(g*g+h*h));b=a+32|0;p[b>>3]=i;i=+p[d>>3];h=+p[f>>3];g=+X(+(i*i+h*h));f=a+40|0;p[f>>3]=g;if(k[a+3304>>2]|0){g=+p[d>>3];p[d>>3]=-+p[e>>3];p[e>>3]=-g}Kd(a);k[a+3312>>2]=1;zd(a);p[a+760>>3]=+p[b>>3];p[a+768>>3]=+p[f>>3];return}function xd(a){a=a|0;ux(601992,a)|0;return}function yd(a,b,c,d){a=a|0;b=+b;c=+c;d=d|0;var e=0,f=0,g=0,h=0,i=0.0,j=0,l=0,m=0,n=0,o=0,q=0.0,r=0,s=0,t=0;if(!d)return;e=a+3316|0;f=k[e>>2]|0;if((f+-1|0)>>>0>8){g=k[a+3320>>2]|0;k[e>>2]=g;h=g}else h=f;f=a+760|0;p[f>>3]=b;g=a+768|0;i=c!=0.0?c:b;p[g>>3]=i;p[a+32>>3]=b;p[a+40>>3]=i;if((h|0)>0){e=0;j=d;l=a+832|0;while(1){m=0;n=j;j=j+(h<<3)|0;o=l;while(1){p[o>>3]=+p[n>>3];m=m+1|0;if((m|0)==(h|0))break;else{n=n+8|0;o=o+8|0}}e=e+1|0;if((e|0)==(h|0))break;else l=l+(h<<3)|0}i=+p[f>>3];f=a+56|0;p[f>>3]=+p[d>>3]*i;if((h|0)>1){p[a+64>>3]=+p[d+8>>3]*i;i=+p[g>>3];p[a+72>>3]=+p[d+(h<<3)>>3]*i;q=+p[d+(h+1<<3)>>3]*i;r=f}else{s=f;t=12}}else{f=a+56|0;p[f>>3]=+p[d>>3]*b;s=f;t=12}if((t|0)==12){t=a+64|0;k[t>>2]=0;k[t+4>>2]=0;k[t+8>>2]=0;k[t+12>>2]=0;q=1.0;r=s}p[a+80>>3]=q;_e(2,r,a+88|0)|0;k[a+3300>>2]=1;Ze(a+3956|0)|0;k[a+3312>>2]=1;zd(a);return}function zd(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0.0,m=0,n=0.0,o=0,q=0.0,s=0.0,t=0.0,u=0,v=0.0,w=0.0;b=r;r=r+64|0;c=b+48|0;d=b+40|0;e=b+32|0;f=b+24|0;g=b+16|0;h=b+8|0;i=b;if(!(+p[a+136>>3]<1.5)?!(+p[a+144>>3]<1.5):0){if((k[a+3884>>2]|0)==6){r=b;return}j=a+32|0;l=+W(+(+p[j>>3]));p[j>>3]=l;m=a+40|0;l=+W(+(+p[m>>3]));p[m>>3]=l;l=+p[a+16>>3];n=+p[a+24>>3];Ad(a,l,n,d,e);o=a+3304|0;q=+p[d>>3];if(!(k[o>>2]|0)){s=+p[e>>3];d=a+3828|0;Hd(a,q+ +p[j>>3],s,d,g,i,c);Hd(a,q,s+ +p[m>>3],d,f,h,c)}else{s=+p[e>>3];e=a+3828|0;Hd(a,q+ +p[m>>3],s,e,g,i,c);Hd(a,q,s+ +p[j>>3],e,f,h,c)}s=+da(+(+p[h>>3]-n),+(+p[f>>3]-l))*180.0/3.141592653589793;q=s<-90.0?s+360.0:s;p[a+3208>>3]=q;s=+da(+(+p[i>>3]-n),+(+p[g>>3]-l))*180.0/3.141592653589793;l=s<-90.0?s+360.0:s;p[a+3216>>3]=l;s=q+(q<-90.0?270.0:-90.0);p[a+3200>>3]=s;g=(k[o>>2]|0)==0;if(!g){n=s+90.0;o=a+48|0;p[o>>3]=n;if(n<0.0){t=n+360.0;p[o>>3]=t;u=o;v=t}else{u=o;v=n}}else{o=a+48|0;p[o>>3]=s;u=o;v=s}if(v<0.0){s=v+360.0;p[u>>3]=s;w=s}else w=v;if(w>=360.0)p[u>>3]=w+-360.0;w=l-q;v=q-l;u=v>80.0&v<100.0|w<280.0&w>260.0?1:w<-80.0&w>-100.0&1;k[a+3256>>2]=u;o=(u|0)!=0;if(g){if(o){r=b;return}p[j>>3]=-+p[j>>3];r=b;return}else{if(!o){r=b;return}p[m>>3]=-+p[m>>3];r=b;return}}w=+p[a+48>>3];p[a+3200>>3]=w;p[a+3208>>3]=w+90.0;p[a+3216>>3]=w+180.0;r=b;return}function Ad(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0.0,x=0.0,y=0.0,z=0.0;f=r;r=r+144|0;g=f+112|0;h=f+80|0;i=f+48|0;j=f+40|0;l=f+32|0;m=f+24|0;n=f+16|0;o=f+8|0;q=f;if(!a){r=f;return}if(!(k[a+3312>>2]|0)){r=f;return}p[a+576>>3]=b;p[a+584>>3]=c;p[a+592>>3]=+p[1282];s=a+3308|0;k[s>>2]=0;t=k[a+9312>>2]|0;if(!t)xg(a,b,c,m,n);else Ad(t,b,c,m,n);t=a+3260|0;u=k[t>>2]|0;a:do switch(u|0){case 29:{if(!(Me(+p[m>>3],+p[n>>3],a,o,q)|0))v=24;else{k[s>>2]=1;w=0.0;x=0.0}break}case 30:{if(!(Ae(+p[m>>3],+p[n>>3],a,o,q)|0))v=24;else{k[s>>2]=1;w=0.0;x=0.0}break}case 31:{if(!(De(+p[m>>3],+p[n>>3],a,o,q)|0))v=24;else{k[s>>2]=1;w=0.0;x=0.0}break}case 32:{if(!(Ke(+p[m>>3],+p[n>>3],a,o,q)|0))v=24;else{k[s>>2]=1;w=0.0;x=0.0}break}default:{c=+p[m>>3];b=+p[n>>3];if((u|0)<1?1:(k[a+3324>>2]|0)==2){if(!(ye(c,b,a,o,q)|0)){v=24;break a}k[s>>2]=1;w=0.0;x=0.0;break a}p[o>>3]=0.0;p[q>>3]=0.0;p[i>>3]=c;p[i+8>>3]=b;if((u+-24|0)>>>0<3)y=+((k[25998]|0)+1|0);else y=+p[1282];p[i+16>>3]=y;p[i+24>>3]=1.0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;k[h+16>>2]=0;k[h+20>>2]=0;k[h+24>>2]=0;k[h+28>>2]=0;if(!(Ye(a+3368|0,a+3924|0,i,a+3956|0,h,a+4064|0,j,l,a+688|0,a+3984|0,g)|0)){p[o>>3]=+p[g+(k[a+3944>>2]<<3)>>3];p[q>>3]=+p[g+(k[a+3948>>2]<<3)>>3];v=24;break a}else{k[s>>2]=1;w=0.0;x=0.0;break a}}}while(0);if((v|0)==24)if(!(k[s>>2]|0)){if((k[t>>2]|0)>0)Ld(k[a+3884>>2]|0,k[a+3888>>2]|0,+p[a+120>>3],+p[a+3872>>3],o,q,+p[a+128>>3]);switch(k[a+3264>>2]|0){case 90:{y=90.0-+p[q>>3];p[q>>3]=y;z=y;break}case -90:{y=+p[q>>3]+-90.0;p[q>>3]=y;z=y;break}default:z=+p[q>>3]}y=+p[o>>3];p[a+600>>3]=y;p[a+608>>3]=z;w=z;x=y}else{w=0.0;x=0.0}p[d>>3]=x;p[e>>3]=w;e=k[a+3888>>2]|0;if((e|0)<1){r=f;return}switch(e|0){case 6:case 10:{r=f;return}default:{}}w=+p[d>>3];if(w<0.0){p[d>>3]=w+360.0;r=f;return}if(!(w>360.0)){r=f;return}p[d>>3]=w+-360.0;r=f;return}function Bd(a,b,c,d,e,f){a=a|0;b=+b;c=+c;d=d|0;e=e|0;f=f|0;Hd(a,b,c,a+3828|0,d,e,f);return}function Cd(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0;f=r;r=r+128|0;g=f+56|0;h=f+48|0;j=f+40|0;l=f+32|0;m=f+24|0;n=f+16|0;o=f+8|0;q=f;s=f+96|0;t=f+64|0;if((a|0)!=0?(k[a+3312>>2]|0)!=0:0){Ad(a,b,c,o,q);if(k[a+3308>>2]|0){u=d;v=u;i[v>>0]=79;i[v+1>>0]=102;i[v+2>>0]=102;i[v+3>>0]=32;v=u+4|0;i[v>>0]=109;i[v+1>>0]=97;i[v+2>>0]=112;i[v+3>>0]=0;w=1;r=f;return w|0}a:do switch(k[a+3292>>2]|0){case 1:{v=a+3288|0;u=k[v>>2]|0;x=(u<<1)+9|0;if((x|0)>=(e|0))if(!(k[a+3296>>2]|0)){Bx(d,602105,e)|0;y=0;break a}else{Bx(d,602084,e)|0;y=0;break a}se(s,32,+p[o>>3],u);se(t,32,+p[q>>3],k[v>>2]|0);if(!(k[a+3296>>2]|0)){k[m>>2]=s;k[m+4>>2]=t;Xw(d,602078,m)|0}else{k[n>>2]=s;k[n+4>>2]=t;Xw(d,602072,n)|0}y=e-x|0;break}case 0:{x=a+3288|0;v=k[x>>2]|0;u=(v<<1)+18|0;if((u|0)>=(e|0))if(!(k[a+3296>>2]|0)){Bx(d,602153,e)|0;y=0;break a}else{Bx(d,602125,e)|0;y=0;break a}c=+p[o>>3];if(((k[a+3888>>2]|0)+-1|0)>>>0<2){qe(s,32,c,v);re(t,32,+p[q>>3],(k[x>>2]|0)+-1|0);z=t}else{re(s,32,c,v);re(t,32,+p[q>>3],k[x>>2]|0);z=t}if(!(k[a+3296>>2]|0)){k[j>>2]=s;k[j+4>>2]=z;Xw(d,602078,j)|0}else{k[l>>2]=s;k[l+4>>2]=z;Xw(d,602072,l)|0}y=e-u|0;break}default:y=e}while(0);switch(k[a+3888>>2]|0){case 3:{if((y|0)<=9){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}l=(k[a+3296>>2]|0)==0;z=d+(xx(d)|0)|0;if(l){A=z;B=602190;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}else{A=z;B=602180;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}break}case 4:{if((y|0)<=9){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}z=(k[a+3296>>2]|0)==0;l=d+(xx(d)|0)|0;if(z){A=l;B=602210;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}else{A=l;B=602200;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}break}case 9:{if((y|0)<=9){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}l=(k[a+3296>>2]|0)==0;z=d+(xx(d)|0)|0;if(l){l=z;j=l;i[j>>0]=32;i[j+1>>0]=112;i[j+2>>0]=108;i[j+3>>0]=97;j=l+4|0;i[j>>0]=110;i[j+1>>0]=101;i[j+2>>0]=116;i[j+3>>0]=0;w=1;r=f;return w|0}else{j=z;z=j;i[z>>0]=9;i[z+1>>0]=112;i[z+2>>0]=108;i[z+3>>0]=97;z=j+4|0;i[z>>0]=110;i[z+1>>0]=101;i[z+2>>0]=116;i[z+3>>0]=0;w=1;r=f;return w|0}break}case 5:{if((y|0)<=7){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}z=(k[a+3296>>2]|0)==0;j=d+(xx(d)|0)|0;if(z){z=j;l=z;i[l>>0]=32;i[l+1>>0]=97;i[l+2>>0]=108;i[l+3>>0]=116;l=z+4|0;i[l>>0]=45;i[l+1>>0]=97;i[l+2>>0]=122;i[l+3>>0]=0;w=1;r=f;return w|0}else{l=j;j=l;i[j>>0]=9;i[j+1>>0]=97;i[j+2>>0]=108;i[j+3>>0]=116;j=l+4|0;i[j>>0]=45;i[j+1>>0]=97;i[j+2>>0]=122;i[j+3>>0]=0;w=1;r=f;return w|0}break}case 7:{if((y|0)<=7){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}j=(k[a+3296>>2]|0)==0;l=d+(xx(d)|0)|0;if(j){A=l;B=602230;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}else{A=l;B=602220;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}break}case 8:{if((y|0)<=7){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}l=(k[a+3296>>2]|0)==0;j=d+(xx(d)|0)|0;if(l){A=j;B=602250;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}else{A=j;B=602240;C=A+10|0;do{i[A>>0]=i[B>>0]|0;A=A+1|0;B=B+1|0}while((A|0)<(C|0));w=1;r=f;return w|0}break}case 1:case 2:{B=a+3796|0;if((y|0)<=((xx(B)|0)+1|0)){w=1;r=f;return w|0}if(!(k[a+3284>>2]|0)){w=1;r=f;return w|0}A=(k[a+3296>>2]|0)==0;C=d+(xx(d)|0)|0;j=A?32:9;i[C>>0]=j;i[C+1>>0]=j>>8;qx(d,B)|0;w=1;r=f;return w|0}default:{B=a+3288|0;te(s,+p[o>>3],0,k[B>>2]|0);te(t,+p[q>>3],0,k[B>>2]|0);B=xx(s)|0;q=B+1+(xx(t)|0)|0;o=a+3476|0;j=a+3508|0;C=(xx(o)|0)+2+(xx(j)|0)|0;A=a+3884|0;if(((k[A>>2]|0)==6?(k[a+3328>>2]|0)==1:0)?(l=C+q|0,(y|0)>(l|0)):0){if(i[o>>0]|0){z=s+B|0;i[z>>0]=32;i[z+1>>0]=0;qx(s,o)|0}if(!(i[j>>0]|0))D=l;else{z=t+(xx(t)|0)|0;i[z>>0]=32;i[z+1>>0]=0;qx(t,j)|0;D=l}}else D=q;q=(k[a+3296>>2]|0)!=0;do if((y|0)>(D|0))if(q){k[h>>2]=s;k[h+4>>2]=t;Xw(d,602072,h)|0;break}else{k[g>>2]=s;k[g+4>>2]=t;Xw(d,602078,g)|0;break}else if(q){Bx(d,602260,y)|0;break}else{Bx(d,602105,y)|0;break}while(0);if((k[A>>2]|0)!=6){w=1;r=f;return w|0}q=a+3328|0;a=k[q>>2]|0;do if((y|0)>(D+7|0)&(a|0)!=1){g=d+(xx(d)|0)|0;t=g;i[t>>0]=32;i[t+1>>0]=108;i[t+2>>0]=105;i[t+3>>0]=110;t=g+4|0;i[t>>0]=101;i[t+1>>0]=97;i[t+2>>0]=114;i[t+3>>0]=0;if((k[A>>2]|0)==6){E=k[q>>2]|0;break}else{w=1;r=f;return w|0}}else E=a;while(0);if((E|0)!=2){w=1;r=f;return w|0}if((y|0)<=(C+7+D|0)){w=1;r=f;return w|0}if(i[o>>0]|0){D=d+(xx(d)|0)|0;i[D>>0]=32;i[D+1>>0]=0;qx(d,o)|0}if(!(i[j>>0]|0)){w=1;r=f;return w|0}o=d+(xx(d)|0)|0;i[o>>0]=32;i[o+1>>0]=0;qx(d,j)|0;w=1;r=f;return w|0}}}if((e|0)<=0){w=0;r=f;return w|0}i[d>>0]=0;w=0;r=f;return w|0}function Dd(a,b,c,d){a=+a;b=+b;c=+c;d=+d;var e=0,f=0,g=0;e=r;r=r+48|0;f=e+24|0;g=e;Zd(a,b,1.0,f);Zd(c,d,1.0,g);d=+p[f>>3]-+p[g>>3];c=+p[f+8>>3]-+p[g+8>>3];b=+p[f+16>>3]-+p[g+16>>3];a=(d*d+0.0+c*c+b*b)*.25;b=a>1.0?1.0:a;a=+da(+(+X(+b)),+(+X(+(1.0-b))))*2.0*180.0/3.141592653589793;r=e;return +a}function Ed(a){a=a|0;var b=0;if(!a)b=0;else b=(k[a+3312>>2]|0)!=0;return (b?a+3796|0:0)|0}function Fd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;if(!a){c=0;return c|0}if(!(k[a+3312>>2]|0)){c=0;return c|0}d=a+3292|0;e=k[d>>2]|0;k[d>>2]=b;if((b|0)==1&(e|0)==0?(d=a+3288|0,(k[d>>2]|0)==3):0)k[d>>2]=6;if(!((b|0)==0&(e|0)==1)){c=e;return c|0}e=a+3288|0;if((k[e>>2]|0)!=5){c=1;return c|0}k[e>>2]=3;c=1;return c|0}function Gd(a){a=a|0;var b=0;if(!a)b=0;else b=(k[a+3312>>2]|0)!=0;return (b?a+3764|0:0)|0}function Hd(a,b,c,d,e,f,g){a=a|0;b=+b;c=+c;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0.0,o=0,q=0,s=0.0,t=0.0,u=0.0,v=0;h=r;r=r+32|0;i=h+24|0;j=h+16|0;l=h+8|0;m=h;if(!a){r=h;return}if(!(k[a+3312>>2]|0)){r=h;return}k[g>>2]=0;p[i>>3]=b;p[j>>3]=c;switch(k[a+3264>>2]|0){case 90:{p[j>>3]=90.0-c;break}case -90:{p[j>>3]=c+-90.0;break}default:{}}if(!d){n=+p[a+120>>3];o=k[a+3884>>2]|0}else{q=Md(d)|0;n=+Nd(d);o=q}p[a+592>>3]=1.0;a:do if((o|0)>0){switch(o|0){case 6:case 10:{break a;break}default:{}}Ld(o,k[a+3884>>2]|0,n,+p[a+120>>3],i,j,+p[a+128>>3])}while(0);o=k[a+3260>>2]|0;b:do switch(o|0){case 29:{if(Ne(+p[i>>3],+p[j>>3],a,l,m)|0)k[g>>2]=1;break}case 30:{if(Be(+p[i>>3],+p[j>>3],a,l,m)|0)k[g>>2]=1;break}case 31:{if(Ee(+p[i>>3],+p[j>>3],a,l,m)|0)k[g>>2]=1;break}case 32:{if(Le(+p[i>>3],+p[j>>3],a,l,m)|0)k[g>>2]=1;break}default:{n=+p[i>>3];s=+p[j>>3];if((o|0)<1?1:(k[a+3324>>2]|0)==2){if(!(ze(n,s,a,l,m)|0))break b;k[g>>2]=1;break b}else{if(!(Id(n,s,a,l,m)|0))break b;k[g>>2]=1;break b}}}while(0);o=k[a+9312>>2]|0;s=+p[l>>3];n=+p[m>>3];do if(!o){wg(a,s,n,e,f);m=k[g>>2]|0;if(!m){t=+p[e>>3];if(!(t<.5)?(u=+p[f>>3],!(u<.5)):0){if(!(t>+p[a+136>>3]+.5)?!(u>+p[a+144>>3]+.5):0){v=0;break}k[g>>2]=2;v=2;break}k[g>>2]=2;v=2}else v=m}else{Hd(o,s,n,0,e,f,g);v=k[g>>2]|0}while(0);k[a+3308>>2]=v;p[a+600>>3]=b;p[a+608>>3]=c;p[a+576>>3]=+p[e>>3];p[a+584>>3]=+p[f>>3];r=h;return}function Id(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0;f=r;r=r+112|0;g=f+80|0;h=f+48|0;i=f+16|0;j=f+8|0;l=f;p[d>>3]=0.0;p[e>>3]=0.0;m=c+3924|0;if((k[m>>2]|0)!=137){n=c+3368|0;if(!(We(k[c+3960>>2]|0,n,m)|0))o=n;else{q=1;r=f;return q|0}}else o=c+3368|0;k[g>>2]=0;k[g+4>>2]=0;k[g+8>>2]=0;k[g+12>>2]=0;k[g+16>>2]=0;k[g+20>>2]=0;k[g+24>>2]=0;k[g+28>>2]=0;p[g+(k[c+3944>>2]<<3)>>3]=a;p[g+(k[c+3948>>2]<<3)>>3]=b;n=i+16|0;k[i>>2]=0;k[i+4>>2]=0;k[i+8>>2]=0;k[i+12>>2]=0;p[n>>3]=1.0;p[i+24>>3]=1.0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;p[h+16>>3]=1.0;p[h+24>>3]=1.0;s=Xe(o,m,g,c+688|0,c+3984|0,j,l,c+4064|0,h,c+3956|0,i)|0;if(s){q=s;r=f;return q|0}p[d>>3]=+p[i>>3];p[e>>3]=+p[i+8>>3];b=+p[n>>3];if(((k[c+3260>>2]|0)+-24|0)>>>0<3){p[c+592>>3]=b+-1.0;q=0;r=f;return q|0}else{p[c+592>>3]=b;q=0;r=f;return q|0}return 0}function Jd(){return k[25997]|0}function Kd(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;b=a+3316|0;c=k[b>>2]|0;d=(c|0)>2?2:c;if((d+-1|0)>>>0>8){c=k[a+3320>>2]|0;k[b>>2]=c;e=c}else e=d;d=ha(e<<3,e)|0;c=a+3976|0;b=k[c>>2]|0;if(!b){f=cy(d)|0;k[c>>2]=f;if(!f)return;else g=f}else g=b;b=a+3980|0;if((k[b>>2]|0)==0?(f=cy(d)|0,k[b>>2]=f,(f|0)==0):0)return;f=a+3956|0;k[f>>2]=137;switch(e|0){case 2:{p[g>>3]=+p[a+56>>3];p[g+8>>3]=+p[a+64>>3];p[g+16>>3]=+p[a+72>>3];p[g+24>>3]=+p[a+80>>3];h=g;break}case 3:{i=g;j=i+72|0;do{k[i>>2]=0;i=i+4|0}while((i|0)<(j|0));d=k[c>>2]|0;p[d>>3]=+p[a+56>>3];p[d+8>>3]=+p[a+64>>3];p[d+24>>3]=+p[a+72>>3];p[d+32>>3]=+p[a+80>>3];p[d+64>>3]=1.0;h=d;break}case 4:{i=g;j=i+128|0;do{k[i>>2]=0;i=i+4|0}while((i|0)<(j|0));i=k[c>>2]|0;p[i>>3]=+p[a+56>>3];p[i+8>>3]=+p[a+64>>3];p[i+32>>3]=+p[a+72>>3];p[i+40>>3]=+p[a+80>>3];p[i+80>>3]=1.0;p[i+120>>3]=1.0;h=i;break}default:h=g}_e(e,h,k[b>>2]|0)|0;k[a+3964>>2]=a+616;k[a+3972>>2]=a+760;k[a+3968>>2]=a+832;k[f>>2]=137;return}function Ld(a,b,c,d,e,f,g){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=+g;var h=0,i=0,j=0,k=0,l=0,m=0.0,n=0,o=0,q=0,s=0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0;h=r;r=r+96|0;i=h+24|0;j=h+16|0;k=h+8|0;l=h;m=c==0.0?((a|0)==2?1950.0:2.0e3):c;n=(b|0)==2;o=(b|0)==11;q=(a|0)==1&o&m==2.0e3;c=(a|0)==11&o|q?m:d==0.0?(n?1950.0:2.0e3):d;o=q?b:a;a=(b|0)==1;q=a&(o|0)==11&c==2.0e3;d=q?c:m;s=q?b:o;if((s|0)==(b|0)&d==c){r=h;return}o=d!=c;if(o){if((s|0)==2&d!=1950.0)Wd(d,1950.0,e,f);if((s|0)==1&d!=2.0e3)Xd(d,2.0e3,e,f)}a:do if(n)switch(s|0){case 1:{if(g>0.0){p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Od(e,f,k,l,i,j);d=g+-1950.0;p[e>>3]=+p[e>>3]+d*+p[k>>3];p[f>>3]=+p[f>>3]+d*+p[l>>3];break a}else{p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Od(e,f,k,l,i,j);break a}break}case 3:{Rd(e,f);break a;break}case 4:{if(g>0.0){Vd(e,f,g);p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Od(e,f,k,l,i,j);d=g+-1950.0;p[e>>3]=+p[e>>3]+d*+p[k>>3];p[f>>3]=+p[f>>3]+d*+p[l>>3];break a}else{Vd(e,f,1950.0);p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Od(e,f,k,l,i,j);p[e>>3]=+p[e>>3]+ +p[k>>3]*0.0;p[f>>3]=+p[f>>3]+ +p[l>>3]*0.0;break a}break}default:break a}else{if(a)switch(s|0){case 2:{if(g>0.0){p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Pd(e,f,k,l,i,j);d=g+-2.0e3;p[e>>3]=+p[e>>3]+d*+p[k>>3];p[f>>3]=+p[f>>3]+d*+p[l>>3];break a}else{p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Pd(e,f,k,l,i,j);break a}break}case 3:{Td(e,f);break a;break}case 4:{if(g>0.0){Vd(e,f,g);break a}else{d=+p[e>>3]*3.141592653589793/180.0;m=+p[f>>3]*3.141592653589793/180.0;t=+Z(+m);u=+Z(+d)*t;v=t*+_(+d);d=+_(+m);Yd(1,.40909280422232386,0.0,0.0,i);m=u*+p[i>>3]+0.0+v*+p[i+24>>3]+d*+p[i+48>>3];t=u*+p[i+8>>3]+0.0+v*+p[i+32>>3]+d*+p[i+56>>3];w=+da(+t,+m);x=w<0.0?w+6.283185307179586:w;w=+da(+(u*+p[i+16>>3]+0.0+v*+p[i+40>>3]+d*+p[i+64>>3]),+(+X(+(m*m+t*t))));p[e>>3]=(x>6.283185307179586?x+-6.283185307179586:x)*180.0/3.141592653589793;p[f>>3]=w*180.0/3.141592653589793;break a}break}default:break a}switch(b|0){case 3:{switch(s|0){case 2:{Qd(e,f);break a;break}case 1:{Sd(e,f);break a;break}case 4:{if(g>0.0)Vd(e,f,g);else{w=+p[e>>3]*3.141592653589793/180.0;x=+p[f>>3]*3.141592653589793/180.0;t=+Z(+x);m=+Z(+w)*t;d=t*+_(+w);w=+_(+x);Yd(1,.40909280422232386,0.0,0.0,i);x=m*+p[i>>3]+0.0+d*+p[i+24>>3]+w*+p[i+48>>3];t=m*+p[i+8>>3]+0.0+d*+p[i+32>>3]+w*+p[i+56>>3];v=+da(+t,+x);u=v<0.0?v+6.283185307179586:v;v=+da(+(m*+p[i+16>>3]+0.0+d*+p[i+40>>3]+w*+p[i+64>>3]),+(+X(+(x*x+t*t))));p[e>>3]=(u>6.283185307179586?u+-6.283185307179586:u)*180.0/3.141592653589793;p[f>>3]=v*180.0/3.141592653589793}Sd(e,f);break a;break}default:break a}break}case 4:break;default:break a}switch(s|0){case 2:{if(g>0.0){p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Pd(e,f,k,l,i,j);v=g+-2.0e3;p[e>>3]=+p[e>>3]+v*+p[k>>3];p[f>>3]=+p[f>>3]+v*+p[l>>3];Ud(e,f,g);break a}else{p[k>>3]=0.0;p[l>>3]=0.0;p[i>>3]=0.0;p[j>>3]=0.0;Pd(e,f,k,l,i,j);p[e>>3]=+p[e>>3]-+p[k>>3]*50.0;p[f>>3]=+p[f>>3]-+p[l>>3]*50.0;Ud(e,f,1950.0);break a}break}case 1:{if(g>0.0){Ud(e,f,g);break a}else{v=+p[e>>3]*3.141592653589793/180.0;u=+p[f>>3]*3.141592653589793/180.0;t=+Z(+u);x=+Z(+v)*t;w=t*+_(+v);v=+_(+u);Yd(1,.40909280422232386,0.0,0.0,i);u=x*+p[i>>3]+0.0+w*+p[i+8>>3]+v*+p[i+16>>3];t=x*+p[i+24>>3]+0.0+w*+p[i+32>>3]+v*+p[i+40>>3];d=+da(+t,+u);m=d<0.0?d+6.283185307179586:d;d=+da(+(x*+p[i+48>>3]+0.0+w*+p[i+56>>3]+v*+p[i+64>>3]),+(+X(+(u*u+t*t))));p[e>>3]=(m>6.283185307179586?m+-6.283185307179586:m)*180.0/3.141592653589793;p[f>>3]=d*180.0/3.141592653589793;break a}break}case 3:{Td(e,f);if(g>0.0){Ud(e,f,g);break a}else{d=+p[e>>3]*3.141592653589793/180.0;m=+p[f>>3]*3.141592653589793/180.0;t=+Z(+m);u=+Z(+d)*t;v=t*+_(+d);d=+_(+m);Yd(1,.40909280422232386,0.0,0.0,i);m=u*+p[i>>3]+0.0+v*+p[i+8>>3]+d*+p[i+16>>3];t=u*+p[i+24>>3]+0.0+v*+p[i+32>>3]+d*+p[i+40>>3];w=+da(+t,+m);x=w<0.0?w+6.283185307179586:w;w=+da(+(u*+p[i+48>>3]+0.0+v*+p[i+56>>3]+d*+p[i+64>>3]),+(+X(+(m*m+t*t))));p[e>>3]=(x>6.283185307179586?x+-6.283185307179586:x)*180.0/3.141592653589793;p[f>>3]=w*180.0/3.141592653589793;break a}break}default:break a}}while(0);if(o){if(n&c!=1950.0)Wd(1950.0,c,e,f);if(a&c!=2.0e3)Xd(2.0e3,c,e,f)}c=+p[f>>3];do if(!(c>90.0))if(c<-90.0){p[f>>3]=-180.0-c;g=+p[e>>3]+180.0;p[e>>3]=g;y=g;break}else{y=+p[e>>3];break}else{p[f>>3]=180.0-c;g=+p[e>>3]+180.0;p[e>>3]=g;y=g}while(0);if(y>360.0){p[e>>3]=y+-360.0;r=h;return}if(!(y<0.0)){r=h;return}p[e>>3]=y+360.0;r=h;return}function Md(a){a=a|0;var b=0,c=0,d=0.0;b=i[a>>0]|0;a:do switch(b<<24>>24){case 106:case 74:{c=1;break}default:if((((((tx(a,602281)|0)!=0?(tx(a,602286)|0)!=0:0)?(tx(a,602293)|0)!=0:0)?(tx(a,775102)|0)!=0:0)?(Ax(a,602298,3)|0)!=0:0)?(Ax(a,775094,3)|0)!=0:0){switch(b<<24>>24){case 98:case 66:{c=2;break a;break}default:{}}if((((tx(a,602302)|0)!=0?(tx(a,602307)|0)!=0:0)?(Ax(a,602314,3)|0)!=0:0)?(Ax(a,775098,3)|0)!=0:0)switch(b<<24>>24){case 105:case 73:{c=11;break a;break}case 103:case 71:{c=3;break a;break}case 101:case 69:{c=4;break a;break}case 97:case 65:{c=5;break a;break}case 110:case 78:{c=7;break a;break}case 108:case 76:{c=6;break a;break}default:{if(!(yx(a,774950,5)|0)){c=10;break a}switch(b<<24>>24){case 112:case 80:{c=9;break a;break}default:{}}if(!(de(a)|0)){c=-1;break a}d=+dx(a);if(d>1980.0){c=1;break a}c=d>1900.0?2:-1;break a}}else c=2}else c=1}while(0);return c|0}function Nd(a){a=a|0;var b=0,c=0.0;b=i[a>>0]|0;switch(b<<24>>24){case 98:case 66:case 106:case 74:{c=+dx(a+1|0);break}default:if((Ax(a,602314,3)|0)!=0?(Ax(a,775098,3)|0)!=0:0)if((((Ax(a,602298,3)|0)!=0?(Ax(a,775094,3)|0)!=0:0)?(Ax(a,602293,4)|0)!=0:0)?(Ax(a,775102,4)|0)!=0:0)if((b+-49&255)<2)c=+dx(a);else c=0.0;else c=2.0e3;else c=1950.0}return +c}function Od(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0.0,j=0.0,k=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0.0,u=0.0,v=0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0;g=r;r=r+48|0;h=g;i=+p[a>>3]*3.141592653589793/180.0;j=+p[b>>3]*3.141592653589793/180.0;k=+p[c>>3]*36.0e4;l=+p[d>>3]*36.0e4;m=+_(+i);n=+Z(+i);i=+_(+j);o=+Z(+j);j=n*o;q=m*o;if(k!=0.0|l!=0.0){s=l*o;t=k*j-l*(m*i);u=-(k*q)-l*(n*i)}else{s=0.0;t=0.0;u=0.0}v=0;do{p[h+(v<<3)>>3]=+p[10264+(v*48|0)>>3]*j+0.0+ +p[10264+(v*48|0)+8>>3]*q+ +p[10264+(v*48|0)+16>>3]*i+ +p[10264+(v*48|0)+24>>3]*u+ +p[10264+(v*48|0)+32>>3]*t+ +p[10264+(v*48|0)+40>>3]*s;v=v+1|0}while((v|0)!=6);s=+p[h>>3];t=+p[h+8>>3];u=+p[h+16>>3];i=+X(+(s*s+t*t+u*u));q=t*-3.1919e-07-s*1.62557e-06-u*1.3843e-07;j=s*q;n=s-i*1.62557e-06-j;m=t*q;o=t-i*3.1919e-07-m;w=u*q;q=u-i*1.3843e-07-w;i=+X(+(q*q+(n*n+o*o)));o=s*.001245-t*.00158-u*.000659;n=s-i*1.62557e-06-j;j=t-i*3.1919e-07-m;m=u-i*1.3843e-07-w;w=+p[h+24>>3]+i*.001245-o*n;u=+p[h+32>>3]-i*.00158-o*j;t=+p[h+40>>3]-i*.000659-o*m;o=n*n+j*j;s=+X(+o);if(!(n==0.0&j==0.0)){q=+da(+j,+n);if(q<0.0)x=q+6.283185307179586;else x=q}else x=0.0;q=+da(+m,+s);if(s>1.0e-30){y=(t*o-m*(n*w+j*u))/(s*(m*m+o));z=(n*u-j*w)/o}else{y=l;z=k}k=+p[e>>3];if(!(k>1.0e-30)){A=x*180.0;B=A/3.141592653589793;p[a>>3]=B;C=q*180.0;D=C/3.141592653589793;p[b>>3]=D;E=z/36.0e4;p[c>>3]=E;F=y/36.0e4;p[d>>3]=F;r=g;return}p[f>>3]=(m*t+(n*w+j*u))/(i*(k*21.095));p[e>>3]=+p[e>>3]/i;A=x*180.0;B=A/3.141592653589793;p[a>>3]=B;C=q*180.0;D=C/3.141592653589793;p[b>>3]=D;E=z/36.0e4;p[c>>3]=E;F=y/36.0e4;p[d>>3]=F;r=g;return}function Pd(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0.0,j=0.0,k=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0;g=r;r=r+48|0;h=g;i=+p[a>>3]*3.141592653589793/180.0;j=+p[b>>3]*3.141592653589793/180.0;k=+p[c>>3]*36.0e4;l=+p[d>>3]*36.0e4;m=+_(+i);n=+Z(+i);i=+_(+j);o=+Z(+j);j=n*o;q=m*o;s=+p[f>>3];t=+p[e>>3];u=s*21.095*t;if(!(k!=0.0|l!=0.0)?!(s!=0.0&t!=0.0):0){v=-.001245;w=.00158;x=.000659}else{v=-(k*q)-l*(n*i)+u*j+-.001245;w=k*j-l*(m*i)+u*q+.00158;x=l*o+u*i+.000659}u=q*-3.1919e-07-j*1.62557e-06-i*1.3843e-07;o=u*j+(j+1.62557e-06);m=u*q+(q+3.1919e-07);n=u*i+(i+1.3843e-07);u=j*.001245-q*.00158-i*.000659;s=v+u*j;j=w+u*q;q=x+u*i;y=0;do{p[h+(y<<3)>>3]=+p[10552+(y*48|0)>>3]*o+0.0+ +p[10552+(y*48|0)+8>>3]*m+ +p[10552+(y*48|0)+16>>3]*n+ +p[10552+(y*48|0)+24>>3]*s+ +p[10552+(y*48|0)+32>>3]*j+ +p[10552+(y*48|0)+40>>3]*q;y=y+1|0}while((y|0)!=6);q=+p[h>>3];j=+p[h+8>>3];s=+p[h+16>>3];n=+p[h+24>>3];m=+p[h+32>>3];o=+p[h+40>>3];i=q*q+j*j;u=+X(+i);x=i+s*s;w=+X(+x);v=q*n+j*m;z=v+s*o;if(!(q==0.0&j==0.0)){A=+da(+j,+q);if(A<0.0)B=A+6.283185307179586;else B=A}else B=0.0;A=+da(+s,+u);if(u>1.0e-30){C=(i*o-s*v)/(x*u);D=(q*m-j*n)/i}else{C=l;D=k}if(!(t>1.0e-30)){E=B*180.0;F=E/3.141592653589793;p[a>>3]=F;G=A*180.0;H=G/3.141592653589793;p[b>>3]=H;I=D/36.0e4;p[c>>3]=I;J=C/36.0e4;p[d>>3]=J;r=g;return}p[f>>3]=z/(w*t*21.095);p[e>>3]=+p[e>>3]/w;E=B*180.0;F=E/3.141592653589793;p[a>>3]=F;G=A*180.0;H=G/3.141592653589793;p[b>>3]=H;I=D/36.0e4;p[c>>3]=I;J=C/36.0e4;p[d>>3]=J;r=g;return}function Qd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0,u=0,v=0,w=0;c=r;r=r+64|0;d=c+48|0;e=c+40|0;f=c;g=+p[a>>3];h=+p[b>>3];j=g*3.141592653589793/180.0;l=h*3.141592653589793/180.0;m=+Z(+l);n=+Z(+j)*m;o=m*+_(+j);j=+_(+l);l=n*-.066988739415+o*-.872755765852+j*-.483538914632;m=n*.492728466075+o*-.45034695802+j*.744584633283;q=+da(+m,+l);s=q<0.0?q+6.283185307179586:q;q=(s>6.283185307179586?s+-6.283185307179586:s)*180.0/3.141592653589793;s=+da(+(n*-.867600811151+o*-.188374601723+j*.460199784784),+(+X(+(l*l+m*m))))*180.0/3.141592653589793;p[a>>3]=q;p[b>>3]=s;if(!(k[25999]|0)){r=c;return}m=g/15.0;b=~~m;g=(m-+(b|0))*60.0;a=~~g;t=h<0.0;m=t?-h:h;u=~~m;h=(m-+(u|0))*60.0;v=~~h;w=cy(32)|0;k[f>>2]=b;k[f+4>>2]=a;p[f+8>>3]=(g-+(a|0))*60.0;k[f+16>>2]=t?45:43;k[f+20>>2]=u;k[f+24>>2]=v;p[f+32>>3]=(h-+(v|0))*60.0;Xw(w,602318,f)|0;f=w+6|0;if((i[f>>0]|0)==32)i[f>>0]=48;f=w+20|0;if((i[f>>0]|0)==32)i[f>>0]=48;f=k[31229]|0;k[e>>2]=w;Ew(f,602354,e)|0;p[d>>3]=q;p[d+8>>3]=s;Ew(f,602381,d)|0;dy(w);r=c;return}function Rd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0,u=0,v=0,w=0;c=r;r=r+64|0;d=c+56|0;e=c+16|0;f=c;g=+p[a>>3];h=+p[b>>3];j=g*3.141592653589793/180.0;l=h*3.141592653589793/180.0;m=+Z(+l);n=+Z(+j)*m;o=m*+_(+j);j=+_(+l);l=n*-.066988739415+o*.492728466075+j*-.867600811151;m=n*-.872755765852+o*-.45034695802+j*-.188374601723;q=+da(+m,+l);s=q<0.0?q+6.283185307179586:q;q=(s>6.283185307179586?s+-6.283185307179586:s)*180.0/3.141592653589793;s=+da(+(n*-.483538914632+o*.744584633283+j*.460199784784),+(+X(+(l*l+m*m))))*180.0/3.141592653589793;p[a>>3]=q;p[b>>3]=s;if(!(k[25999]|0)){r=c;return}b=k[31229]|0;p[f>>3]=g;p[f+8>>3]=h;Ew(b,602414,f)|0;h=q/15.0;f=~~h;q=(h-+(f|0))*60.0;a=~~q;t=s<0.0;h=t?-s:s;u=~~h;s=(h-+(u|0))*60.0;v=~~s;w=cy(32)|0;k[e>>2]=f;k[e+4>>2]=a;p[e+8>>3]=(q-+(a|0))*60.0;k[e+16>>2]=t?45:43;k[e+20>>2]=u;k[e+24>>2]=v;p[e+32>>3]=(s-+(v|0))*60.0;Xw(w,602318,e)|0;e=w+6|0;if((i[e>>0]|0)==32)i[e>>0]=48;e=w+20|0;if((i[e>>0]|0)==32)i[e>>0]=48;k[d>>2]=w;Ew(b,602447,d)|0;dy(w);r=c;return}function Sd(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0,u=0,v=0,w=0;c=r;r=r+64|0;d=c+48|0;e=c+40|0;f=c;g=+p[a>>3];h=+p[b>>3];j=g*3.141592653589793/180.0;l=h*3.141592653589793/180.0;m=+Z(+l);n=+Z(+j)*m;o=m*+_(+j);j=+_(+l);l=n*-.054875539726+o*-.87343710801+j*-.483834985808;m=n*.494109453312+o*-.444829589425+j*.74698225181;q=+da(+m,+l);s=q<0.0?q+6.283185307179586:q;q=(s>6.283185307179586?s+-6.283185307179586:s)*180.0/3.141592653589793;s=+da(+(n*-.867666135858+o*-.198076386122+j*.455983795705),+(+X(+(l*l+m*m))))*180.0/3.141592653589793;p[a>>3]=q;p[b>>3]=s;if(!(k[25999]|0)){r=c;return}m=g/15.0;b=~~m;g=(m-+(b|0))*60.0;a=~~g;t=h<0.0;m=t?-h:h;u=~~m;h=(m-+(u|0))*60.0;v=~~h;w=cy(32)|0;k[f>>2]=b;k[f+4>>2]=a;p[f+8>>3]=(g-+(a|0))*60.0;k[f+16>>2]=t?45:43;k[f+20>>2]=u;k[f+24>>2]=v;p[f+32>>3]=(h-+(v|0))*60.0;Xw(w,602318,f)|0;f=w+6|0;if((i[f>>0]|0)==32)i[f>>0]=48;f=w+20|0;if((i[f>>0]|0)==32)i[f>>0]=48;f=k[31229]|0;k[e>>2]=w;Ew(f,602474,e)|0;p[d>>3]=q;p[d+8>>3]=s;Ew(f,602501,d)|0;dy(w);r=c;return}function Td(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0,u=0,v=0,w=0;c=r;r=r+64|0;d=c+56|0;e=c+16|0;f=c;g=+p[a>>3];h=+p[b>>3];j=g*3.141592653589793/180.0;l=h*3.141592653589793/180.0;m=+Z(+l);n=+Z(+j)*m;o=m*+_(+j);j=+_(+l);l=n*-.054875539726+o*.494109453312+j*-.867666135858;m=n*-.87343710801+o*-.444829589425+j*-.198076386122;q=+da(+m,+l);s=q<0.0?q+6.283185307179586:q;q=(s>6.283185307179586?s+-6.283185307179586:s)*180.0/3.141592653589793;s=+da(+(n*-.483834985808+o*.74698225181+j*.455983795705),+(+X(+(l*l+m*m))))*180.0/3.141592653589793;p[a>>3]=q;p[b>>3]=s;if(!(k[25999]|0)){r=c;return}b=k[31229]|0;p[f>>3]=g;p[f+8>>3]=h;Ew(b,602534,f)|0;h=q/15.0;f=~~h;q=(h-+(f|0))*60.0;a=~~q;t=s<0.0;h=t?-s:s;u=~~h;s=(h-+(u|0))*60.0;v=~~s;w=cy(32)|0;k[e>>2]=f;k[e+4>>2]=a;p[e+8>>3]=(q-+(a|0))*60.0;k[e+16>>2]=t?45:43;k[e+20>>2]=u;k[e+24>>2]=v;p[e+32>>3]=(s-+(v|0))*60.0;Xw(w,602318,e)|0;e=w+6|0;if((i[e>>0]|0)==32)i[e>>0]=48;e=w+20|0;if((i[e>>0]|0)==32)i[e>>0]=48;k[d>>2]=w;Ew(b,602567,d)|0;dy(w);r=c;return}function Ud(a,b,c){a=a|0;b=b|0;c=+c;var d=0,e=0,f=0.0,g=0.0,h=0.0,i=0.0,j=0.0,k=0.0;d=r;r=r+80|0;e=d;if(c!=2.0e3)Xd(2.0e3,c,a,b);f=+p[a>>3]*3.141592653589793/180.0;g=+p[b>>3]*3.141592653589793/180.0;h=+Z(+g);i=+Z(+f)*h;j=h*+_(+f);f=+_(+g);g=(c+-2.0e3)*.01;Yd(1,(g*(g*(g*.001813+-.00059)+-46.815)+84381.448)*4.8481368110953e-06,0.0,0.0,e);g=+p[e>>3]*i+0.0+ +p[e+8>>3]*j+ +p[e+16>>3]*f;c=+p[e+24>>3]*i+0.0+ +p[e+32>>3]*j+ +p[e+40>>3]*f;h=+da(+c,+g);k=h<0.0?h+6.283185307179586:h;h=+da(+(+p[e+48>>3]*i+0.0+ +p[e+56>>3]*j+ +p[e+64>>3]*f),+(+X(+(g*g+c*c))));p[a>>3]=(k>6.283185307179586?k+-6.283185307179586:k)*180.0/3.141592653589793;p[b>>3]=h*180.0/3.141592653589793;r=d;return}function Vd(a,b,c){a=a|0;b=b|0;c=+c;var d=0,e=0,f=0.0,g=0.0,h=0.0,i=0.0,j=0.0,k=0.0,l=0.0;d=r;r=r+80|0;e=d;f=+p[a>>3]*3.141592653589793/180.0;g=+p[b>>3]*3.141592653589793/180.0;h=+Z(+g);i=+Z(+f)*h;j=h*+_(+f);f=+_(+g);g=(c+-2.0e3)*.01;Yd(1,(g*(g*(g*.001813+-.00059)+-46.815)+84381.448)*4.8481368110953e-06,0.0,0.0,e);g=+p[e>>3]*i+0.0+ +p[e+24>>3]*j+ +p[e+48>>3]*f;h=+p[e+8>>3]*i+0.0+ +p[e+32>>3]*j+ +p[e+56>>3]*f;k=+da(+h,+g);l=k<0.0?k+6.283185307179586:k;k=+da(+(+p[e+16>>3]*i+0.0+ +p[e+40>>3]*j+ +p[e+64>>3]*f),+(+X(+(g*g+h*h))));p[a>>3]=(l>6.283185307179586?l+-6.283185307179586:l)*180.0/3.141592653589793;p[b>>3]=k*180.0/3.141592653589793;if(!(c!=2.0e3)){r=d;return}Xd(c,2.0e3,a,b);r=d;return}function Wd(a,b,c,d){a=+a;b=+b;c=c|0;d=d|0;var e=0,f=0,g=0.0,h=0.0,i=0.0,j=0.0,k=0.0;e=r;r=r+80|0;f=e;g=+p[c>>3]*3.141592653589793/180.0;h=+p[d>>3]*3.141592653589793/180.0;i=(a+-1850.0)/100.0;j=(b-a)/100.0;a=j*4.8481368110953e-06;b=i*(i*.000059+1.3972)+2303.5548;k=i*.000365;Yd(323,-(a*(b+j*(.30242-i*.000269+j*.017996))),a*(i*(-.85294-k)+2005.1125+j*(-.42647-k-j*.041802)),-(a*(b+j*(i*.000387+1.09478+j*.018324))),f);j=+Z(+h);i=+Z(+g)*j;b=j*+_(+g);g=+_(+h);h=+p[f>>3]*i+0.0+ +p[f+8>>3]*b+ +p[f+16>>3]*g;j=+p[f+24>>3]*i+0.0+ +p[f+32>>3]*b+ +p[f+40>>3]*g;a=+da(+j,+h);k=a<0.0?a+6.283185307179586:a;a=+da(+(+p[f+48>>3]*i+0.0+ +p[f+56>>3]*b+ +p[f+64>>3]*g),+(+X(+(h*h+j*j))));p[c>>3]=(k>6.283185307179586?k+-6.283185307179586:k)*180.0/3.141592653589793;p[d>>3]=a*180.0/3.141592653589793;r=e;return}function Xd(a,b,c,d){a=+a;b=+b;c=c|0;d=d|0;var e=0,f=0,g=0.0,h=0.0,i=0.0,j=0.0,k=0.0;e=r;r=r+80|0;f=e;g=+p[c>>3]*3.141592653589793/180.0;h=+p[d>>3]*3.141592653589793/180.0;i=(a+-2.0e3)/100.0;j=(b-a)/100.0;a=j*4.8481368110953e-06;b=i*(1.39656-i*.000139)+2306.2181;k=i*.000217;Yd(323,-(a*(b+j*(.30188-i*.000344+j*.017998))),a*(i*(-.8533-k)+2004.3109+j*(-.42665-k-j*.041833)),-(a*(b+j*(i*.000066+1.09468+j*.018203))),f);j=+Z(+h);i=+Z(+g)*j;b=j*+_(+g);g=+_(+h);h=i*+p[f>>3]+0.0+b*+p[f+8>>3]+g*+p[f+16>>3];j=i*+p[f+24>>3]+0.0+b*+p[f+32>>3]+g*+p[f+40>>3];a=+da(+j,+h);k=a<0.0?a+6.283185307179586:a;a=+da(+(i*+p[f+48>>3]+0.0+b*+p[f+56>>3]+g*+p[f+64>>3]),+(+X(+(h*h+j*j))));p[c>>3]=(k>6.283185307179586?k+-6.283185307179586:k)*180.0/3.141592653589793;p[d>>3]=a*180.0/3.141592653589793;r=e;return}function Yd(a,b,c,d,e){a=a|0;b=+b;c=+c;d=+d;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0.0,M=0.0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0,T=0.0,U=0.0,V=0.0,W=0.0,X=0.0,Y=0.0,$=0.0,aa=0.0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ia=0.0,ja=0.0,ka=0.0,la=0.0,ma=0,na=0.0,oa=0.0,pa=0.0,qa=0.0,ra=0.0,sa=0.0,ta=0.0,ua=0.0,va=0.0,wa=0.0,xa=0.0,ya=0.0,za=0.0,Aa=0.0,Ba=0.0,Ca=0.0,Da=0.0,Ea=0.0;f=r;r=r+192|0;g=f+72|0;h=f+96|0;i=f;j=f+168|0;l=e+8|0;p[e>>3]=1.0;p[l>>3]=0.0;p[e+16>>3]=0.0;m=e+24|0;p[m>>3]=0.0;p[e+32>>3]=1.0;p[e+40>>3]=0.0;p[e+48>>3]=0.0;p[e+56>>3]=0.0;p[e+64>>3]=1.0;n=(a|0)/100|0;k[j>>2]=n;o=(a|0)>99;q=ha(n,-100)|0;n=o&1;s=(o?q:0)+a|0;a=(s|0)/10|0;k[j+(n<<2)>>2]=a;if((s|0)>9){q=s+(ha(a,-10)|0)|0;t=q;u=o?2:1}else{t=s;u=n}k[j+(u<<2)>>2]=t;n=((t|0)>0&1)+u|0;p[g>>3]=b;p[g+8>>3]=c;p[g+16>>3]=d;if((n|0)<=0){r=f;return}u=i+8|0;t=i+16|0;s=i+24|0;o=i+32|0;q=i+40|0;a=i+48|0;v=i+56|0;w=i+64|0;x=e+48|0;y=e+32|0;z=e+56|0;A=e+16|0;B=e+40|0;C=e+64|0;D=h+8|0;E=h+16|0;F=h+24|0;G=h+32|0;H=h+40|0;I=h+48|0;J=h+56|0;K=h+64|0;d=+p[e>>3];c=+p[m>>3];L=+p[x>>3];M=+p[y>>3];N=+p[z>>3];O=+p[A>>3];P=+p[B>>3];Q=b;b=+p[l>>3];R=1.0;S=0;while(1){T=+_(+Q);U=+Z(+Q);switch(k[j+(S<<2)>>2]|0){case 1:{V=1.0;W=0.0;X=0.0;Y=0.0;$=U;aa=T;ba=0.0;ca=-T;da=U;break}case 2:{V=U;W=0.0;X=-T;Y=0.0;$=1.0;aa=0.0;ba=T;ca=0.0;da=U;break}default:{V=U;W=T;X=0.0;Y=-T;$=U;aa=0.0;ba=0.0;ca=0.0;da=1.0}}U=V*d+0.0+W*c+X*L;T=V*b+0.0+W*M+X*N;ea=V*O+0.0+W*P+X*R;fa=Y*d+0.0+$*c+aa*L;ga=Y*b+0.0+$*M+aa*N;ia=Y*O+0.0+$*P+aa*R;ja=ba*d+0.0+ca*c+da*L;ka=ba*b+0.0+ca*M+da*N;la=ba*O+0.0+ca*P+da*R;ma=S+1|0;if((ma|0)==(n|0)){na=Y;oa=$;pa=aa;qa=ba;ra=ca;sa=da;ta=U;ua=T;va=ea;wa=fa;xa=ga;ya=ia;za=ja;Aa=ka;Ba=la;Ca=V;Da=W;Ea=X;break}d=U;c=fa;L=ja;M=ga;N=ka;O=ea;P=ia;Q=+p[g+(ma<<3)>>3];b=T;R=la;S=ma}p[i>>3]=Ca;p[u>>3]=Da;p[t>>3]=Ea;p[s>>3]=na;p[o>>3]=oa;p[q>>3]=pa;p[a>>3]=qa;p[v>>3]=ra;p[w>>3]=sa;p[h>>3]=ta;p[D>>3]=ua;p[E>>3]=va;p[F>>3]=wa;p[G>>3]=xa;p[H>>3]=ya;p[I>>3]=za;p[J>>3]=Aa;p[K>>3]=Ba;p[e>>3]=ta;p[m>>3]=wa;p[x>>3]=za;p[l>>3]=ua;p[y>>3]=xa;p[z>>3]=Aa;p[A>>3]=va;p[B>>3]=ya;p[C>>3]=Ba;r=f;return}function Zd(a,b,c,d){a=+a;b=+b;c=+c;d=d|0;var e=0.0,f=0.0;e=a*3.141592653589793/180.0;a=b*3.141592653589793/180.0;b=+Z(+a);f=+Z(+e)*c*b;p[d>>3]=f;f=b*(+_(+e)*c);p[d+8>>3]=f;f=+_(+a)*c;p[d+16>>3]=f;return}function _d(a,b){a=a|0;b=b|0;var c=0;if((b|0)>0)c=b;else{k[26e3]=0;c=($d(a,768779)|0)+80-a|0}k[26e3]=c;return c|0}function $d(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0;c=k[26e3]|0;d=(c|0)==0?256e3:c;if((d|0)>0)e=0;else{f=0;return f|0}while(1){if((i[a+e>>0]|0)<1){g=e;break}c=e+1|0;if((c|0)<(d|0))e=c;else{g=c;break}}e=a+g|0;d=e;if((g|0)<=0){f=0;return f|0}g=a;c=a;a:while(1){a=oe(c,b,d-c|0)|0;if(!a){f=0;h=17;break}j=(a-g|0)%80|0;l=a+(xx(b)|0)|0;m=i[l>>0]|0;b:do if((j|0)<=7){c:do if(m<<24>>24>32){switch(m<<24>>24){case 61:case 127:{break c;break}default:{}}n=a+1|0;break b}while(0);l=a+(0-j)|0;if((j|0)>0){o=a+1|0;p=c;q=l;while(1){r=(i[q>>0]|0)==32?p:o;q=q+1|0;if(q>>>0>=a>>>0){s=r;break}else p=r}}else s=c;if(a>>>0<s>>>0)n=s;else{f=l;h=17;break a}}else n=a+1|0;while(0);if(n>>>0<e>>>0)c=n;else{f=0;h=17;break}}if((h|0)==17)return f|0;return 0}function ae(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;if((i[c>>0]|0)<64){g=be(a,b,d)|0;r=e;return g|0}else{ux(f,b)|0;h=xx(b)|0;i[f+h>>0]=i[c>>0]|0;i[f+(h+1)>>0]=0;g=be(a,f,d)|0;r=e;return g|0}return 0}function be(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0.0,g=0.0;d=ce(a,b)|0;if(!d){e=0;return e|0}b=(i[d>>0]|0)==35?d+1|0:d;if((xx(b)|0)>81){Bx(602594,b,81)|0;i[602675]=0}else ux(602594,b)|0;if((de(602594)|0)==2){b=rx(602594,68)|0;if(b)i[b>>0]=101;b=rx(602594,100)|0;if(b)i[b>>0]=101;b=rx(602594,69)|0;if(b)i[b>>0]=101}f=+dx(602594);g=f+.001;if(g>2147483647.0){k[c>>2]=2147483647;e=1;return e|0}if(f>=0.0){k[c>>2]=~~g;e=1;return e|0}g=f+-.001;if(g<-2147483648.0){k[c>>2]=-2147483648;e=1;return e|0}else{k[c>>2]=~~g;e=1;return e|0}return 0}function ce(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0;c=r;r=r+208|0;d=c+196|0;e=c+112|0;f=c+8|0;g=c+6|0;h=c+4|0;j=c+2|0;l=c;m=c+114|0;n=c+12|0;i[e>>0]=39;i[e+1>>0]=0;i[f>>0]=34;i[f+1>>0]=0;i[g>>0]=91;i[g+1>>0]=0;i[l>>0]=44;i[l+1>>0]=0;i[h>>0]=93;i[h+1>>0]=0;i[j>>0]=47;i[j+1>>0]=0;Bx(m,b,80)|0;b=xx(m)|0;o=xx(g)|0;a:do if(o){if((b|0)!=0?(p=o+-1|0,q=i[g+p>>0]|0,s=b+1-o|0,t=m+s|0,(s|0)>0):0){s=(o|0)==1;u=(o|0)<3;v=m;do{b:do if((i[v>>0]|0)==91){if(s){w=v;x=23;break a}if((i[v+p>>0]|0)==q<<24>>24){if(u){w=v;x=23;break a}else y=1;while(1){if((i[v+y>>0]|0)!=(i[g+y>>0]|0))break b;y=y+1|0;if((y|0)>=(o|0)){w=v;x=23;break a}}}}while(0);v=v+1|0}while(v>>>0<t>>>0)}t=xx(m)|0;v=xx(l)|0;if(v)if((t|0)!=0?(u=v+-1|0,q=i[l+u>>0]|0,p=t+1-v|0,t=m+p|0,(p|0)>0):0){p=(v|0)==1;s=(v|0)<3;z=m;while(1){c:do if((i[z>>0]|0)==44){if(p){w=z;x=23;break a}if((i[z+u>>0]|0)==q<<24>>24){if(s){w=z;x=23;break a}else A=1;while(1){if((i[z+A>>0]|0)!=(i[l+A>>0]|0))break c;A=A+1|0;if((A|0)>=(v|0)){w=z;x=23;break a}}}}while(0);z=z+1|0;if(z>>>0>=t>>>0){B=0;break}}}else B=0;else{w=m;x=23}}else{w=m;x=23}while(0);if((x|0)==23){i[w>>0]=0;B=w+1|0}w=$d(a,m)|0;if(!w){C=0;r=c;return C|0}m=n;a=m+100|0;do{i[m>>0]=0;m=m+1|0}while((m|0)<(a|0));Bx(n,w,80)|0;w=xx(n)|0;m=xx(e)|0;d:do if(m)if((w|0)!=0?(a=i[e>>0]|0,A=m+-1|0,l=i[e+A>>0]|0,o=w+1-m|0,y=n+o|0,(o|0)>0):0){o=(m|0)==1;g=(m|0)<3;b=n;while(1){e:do if((i[b>>0]|0)==a<<24>>24){if(o){D=b;break d}if((i[b+A>>0]|0)==l<<24>>24){if(g){D=b;break d}else E=1;while(1){if((i[b+E>>0]|0)!=(i[e+E>>0]|0))break e;E=E+1|0;if((E|0)>=(m|0)){D=b;break d}}}}while(0);b=b+1|0;if(b>>>0>=y>>>0){D=0;break}}}else D=0;else D=n;while(0);m=xx(n)|0;E=xx(j)|0;f:do if(E)if((m|0)!=0?(w=i[j>>0]|0,y=E+-1|0,b=i[j+y>>0]|0,g=m+1-E|0,l=n+g|0,(g|0)>0):0){g=(E|0)==1;A=(E|0)<3;o=n;while(1){g:do if((i[o>>0]|0)==w<<24>>24){if(g){F=o;break f}if((i[o+y>>0]|0)==b<<24>>24){if(A){F=o;break f}else G=1;while(1){if((i[o+G>>0]|0)!=(i[j+G>>0]|0))break g;G=G+1|0;if((G|0)>=(E|0)){F=o;break f}}}}while(0);o=o+1|0;if(o>>>0>=l>>>0){F=0;break}}}else F=0;else F=n;while(0);h:do if(!D){E=xx(n)|0;G=xx(f)|0;i:do if(!G)H=n;else{if(!E){x=116;break h}j=i[f>>0]|0;m=G+-1|0;l=i[f+m>>0]|0;o=E+1-G|0;A=n+o|0;if((o|0)<=0){x=116;break h}o=(G|0)==1;b=(G|0)<3;y=n;while(1){j:do if((i[y>>0]|0)==j<<24>>24){if(o){H=y;break i}if((i[y+m>>0]|0)==l<<24>>24){if(b){H=y;break i}else I=1;while(1){if((i[y+I>>0]|0)!=(i[f+I>>0]|0))break j;I=I+1|0;if((I|0)>=(G|0)){H=y;break i}}}}while(0);y=y+1|0;if(y>>>0>=A>>>0){x=116;break h}}}while(0);if(H>>>0<F>>>0){G=H+1|0;E=xx(G)|0;A=xx(f)|0;if(!A){J=H;K=G;x=115;break}if((E|0)!=0?(y=i[f>>0]|0,b=i[f+(A+-1)>>0]|0,l=E+1-A|0,E=H+(l+1)|0,(l|0)>0):0){l=(A|0)==1;m=(A|0)<3;o=H;j=G;while(1){k:do if((i[j>>0]|0)==y<<24>>24){if(l){J=H;K=j;x=115;break h}if((i[o+A>>0]|0)==b<<24>>24){if(m){J=H;K=j;x=115;break h}else L=1;while(1){G=L;L=L+1|0;if((i[o+L>>0]|0)!=(i[f+G>>0]|0))break k;if((L|0)>=(A|0)){J=H;K=j;x=115;break h}}}}while(0);G=j+1|0;if(G>>>0<E>>>0){g=j;j=G;o=g}else{M=F;break}}}else M=F;while(1){o=M+-1|0;if((i[o>>0]|0)==32)M=o;else{J=H;K=M;x=115;break h}}}if(!F){o=H+1|0;j=xx(o)|0;E=xx(f)|0;if(!E){J=H;K=o;x=115}else{if((j|0)!=0?(A=i[f>>0]|0,m=i[f+(E+-1)>>0]|0,b=j+1-E|0,j=H+(b+1)|0,(b|0)>0):0){b=(E|0)==1;l=(E|0)<3;y=H;g=o;while(1){l:do if((i[g>>0]|0)==A<<24>>24){if(b){J=H;K=g;x=115;break h}if((i[y+E>>0]|0)==m<<24>>24){if(l){J=H;K=g;x=115;break h}else N=1;while(1){o=N;N=N+1|0;if((i[y+N>>0]|0)!=(i[f+o>>0]|0))break l;if((N|0)>=(E|0)){J=H;K=g;x=115;break h}}}}while(0);o=g+1|0;if(o>>>0<j>>>0){G=g;g=o;y=G}else break}}y=n+79|0;while(1)if((i[y>>0]|0)==32)y=y+-1|0;else{O=y;break}J=H;K=O+1|0;x=115}}else x=116}else{if(D>>>0<F>>>0){y=D+1|0;g=xx(y)|0;j=xx(e)|0;if(!j){J=D;K=y;x=115;break}if((g|0)!=0?(E=i[e>>0]|0,l=i[e+(j+-1)>>0]|0,m=g+1-j|0,g=D+(m+1)|0,(m|0)>0):0){m=(j|0)==1;b=(j|0)<3;A=D;G=y;while(1){m:do if((i[G>>0]|0)==E<<24>>24){if(m){J=D;K=G;x=115;break h}if((i[A+j>>0]|0)==l<<24>>24){if(b){J=D;K=G;x=115;break h}else P=1;while(1){y=P;P=P+1|0;if((i[A+P>>0]|0)!=(i[e+y>>0]|0))break m;if((P|0)>=(j|0)){J=D;K=G;x=115;break h}}}}while(0);y=G+1|0;if(y>>>0<g>>>0){o=G;G=y;A=o}else{Q=F;break}}}else Q=F;while(1){A=Q+-1|0;if((i[A>>0]|0)==32)Q=A;else{J=D;K=Q;x=115;break h}}}if(!F){A=D+1|0;G=xx(A)|0;g=xx(e)|0;if(!g){J=D;K=A;x=115}else{if((G|0)!=0?(j=i[e>>0]|0,b=i[e+(g+-1)>>0]|0,l=G+1-g|0,G=D+(l+1)|0,(l|0)>0):0){l=(g|0)==1;m=(g|0)<3;E=D;o=A;while(1){n:do if((i[o>>0]|0)==j<<24>>24){if(l){J=D;K=o;x=115;break h}if((i[E+g>>0]|0)==b<<24>>24){if(m){J=D;K=o;x=115;break h}else R=1;while(1){A=R;R=R+1|0;if((i[E+R>>0]|0)!=(i[e+A>>0]|0))break n;if((R|0)>=(g|0)){J=D;K=o;x=115;break h}}}}while(0);A=o+1|0;if(A>>>0<G>>>0){y=o;o=A;E=y}else break}}E=n+79|0;while(1)if((i[E>>0]|0)==32)E=E+-1|0;else{S=E;break}J=D;K=S+1|0;x=115}}else x=116}while(0);if((x|0)==115){T=J+1|0;U=K}else if((x|0)==116){K=xx(n)|0;o:do if((K|0)!=0?(J=n+K|0,(K|0)>0):0){S=n;while(1){if((i[S>>0]|0)==61){V=S;break o}S=S+1|0;if(S>>>0>=J>>>0){V=0;break}}}else V=0;while(0);K=(V|0)==0?n+9|0:V+1|0;V=xx(n)|0;p:do if((V|0)!=0?(J=n+V|0,(V|0)>0):0){S=n;while(1){if((i[S>>0]|0)==47){W=S;break p}S=S+1|0;if(S>>>0>=J>>>0){W=0;break}}}else W=0;while(0);T=K;U=(W|0)==0?n+79|0:W}if(!(k[26001]|0)){W=T;while(1)if(W>>>0<U>>>0&(i[W>>0]|0)==32)W=W+1|0;else{X=W;break}}else X=T;i[U>>0]=0;q:do if(!(k[26001]|0)){T=U;while(1){W=T+-1|0;switch(i[W>>0]|0){case 13:case 32:break;default:break q}if(W>>>0<=X>>>0)break q;i[W>>0]=0;T=W}}while(0);U=(tx(X,769611)|0)==0;T=U?X+1|0:X;ux(602676,T)|0;if(!B){C=602676;r=c;return C|0}X=xx(B)|0;U=xx(h)|0;r:do if(U){if((X|0)!=0?(W=i[h>>0]|0,n=U+-1|0,K=i[h+n>>0]|0,V=X+1-U|0,J=B+V|0,(V|0)>0):0){V=(U|0)==1;S=(U|0)<3;D=B;s:while(1){t:do if((i[D>>0]|0)==W<<24>>24){if(V){Y=D;break s}if((i[D+n>>0]|0)==K<<24>>24){if(S){Z=D;x=144;break r}else _=1;while(1){if((i[D+_>>0]|0)!=(i[h+_>>0]|0))break t;_=_+1|0;if((_|0)>=(U|0)){Z=D;x=144;break r}}}}while(0);D=D+1|0;if(D>>>0>=J>>>0)break r}if(Y){Z=Y;x=144}}}else{Z=B;x=144}while(0);if((x|0)==144)i[Z>>0]=0;if(!(de(B)|0)){Z=xx(B)|0;if((Z|0)>0){x=0;do{Y=B+x|0;U=i[Y>>0]|0;if((U+-65&255)<26)i[Y>>0]=(U&255)+32;x=x+1|0}while((x|0)!=(Z|0))}Z=xe(602676,B)|0;if(!Z){C=0;r=c;return C|0}ux(602676,Z)|0;C=602676;r=c;return C|0}Z=ex(B)|0;i[d>>0]=32;i[d+1>>0]=0;if((Z|0)>0){B=1;x=T;while(1){U=Fx(x,d)|0;if((B|0)==(Z|0)){$=U;break}else{B=B+1|0;x=0}}if(!$){C=0;r=c;return C|0}ux(602676,$)|0;C=602676;r=c;return C|0}if((Z|0)>=0){C=602676;r=c;return C|0}$=0-Z|0;u:do if((Z|0)>=-1)if(!T){C=0;r=c;return C|0}else aa=T;else{x=1;B=T;while(1){d=rx(B,32)|0;if(!d){C=0;break}U=d+1|0;x=x+1|0;if((x|0)>=($|0)){aa=U;break u}else B=U}r=c;return C|0}while(0);ux(602676,aa)|0;C=602676;r=c;return C|0}function de(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0;if(!a){b=0;return b|0}c=i[a>>0]|0;switch(c<<24>>24){case 101:case 69:case 100:case 68:{b=0;return b|0}default:{}}d=xx(a)|0;while(1){e=d+-1|0;if((i[a+e>>0]|0)==32)d=e;else{f=d;break}}if((f|0)<=0){b=0;return b|0}a:do if(c<<24>>24!=10){d=c;e=0;g=1;h=0;j=0;b:while(1){c:do if((j|0)==0&d<<24>>24==32){k=e;l=g;m=0}else{d:do if((d+-48&255)>=10){switch(d<<24>>24){case 101:case 100:case 69:case 68:case 58:case 46:{n=13;break d;break}case 45:case 43:break;default:{b=0;n=22;break b}}switch(i[a+(h+1)>>0]|0){case 43:case 45:{b=0;n=22;break b;break}default:{}}if((h|0)>0)switch(i[a+(h+-1)>>0]|0){case 32:case 58:case 101:case 69:case 100:case 68:{o=e;p=j;break}default:{b=0;n=22;break b}}else{o=e;p=j}}else n=13;while(0);do if((n|0)==13){n=0;if((d+-47&255)<11){o=e;p=j+1|0;break}else{o=(d<<24>>24==58&1)+e|0;p=j;break}}while(0);switch(d<<24>>24){case 101:case 100:case 46:break;default:{k=o;l=g;m=p;break c}}k=o;l=2;m=p}while(0);h=h+1|0;if((h|0)>=(f|0)){q=k;r=l;s=m;break a}d=i[a+h>>0]|0;if(d<<24>>24==10){q=k;r=l;s=m;break a}else{e=k;g=l;j=m}}if((n|0)==22)return b|0}else{q=0;r=1;s=0}while(0);if((s|0)<=0){b=0;return b|0}b=(q|0)==0?r:3;return b|0}function ee(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0.0,g=0,h=0;d=ce(a,b)|0;if(!d){e=0;return e|0}f=+ge(d);b=xx(d)|0;a:do if((b|0)!=0?(a=d+b|0,(b|0)>0):0){g=d;while(1){if((i[g>>0]|0)==58){h=g;break a}g=g+1|0;if(g>>>0>=a>>>0){h=0;break}}}else h=0;while(0);p[c>>3]=(h|0)==0?f:f*15.0;e=1;return e|0}function fe(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0.0;d=ce(a,b)|0;if(!d){e=0;return e|0}f=+ge(d);p[c>>3]=f;e=1;return e|0}function ge(a){a=a|0;var b=0.0,c=0,d=0,e=0,f=0,g=0,h=0.0,j=0,k=0,l=0,m=0.0,n=0,o=0,p=0.0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0,z=0.0,A=0.0;if(!a){b=0.0;return +b}c=i[a>>0]|0;if(!(c<<24>>24)){b=0.0;return +b}else{d=c;e=a}a:while(1){switch(d<<24>>24){case 45:{f=e;g=5;break a;break}case 43:{h=1.0;j=e;g=6;break a;break}case 32:break;default:{k=e;g=7;break a}}a=e+1|0;d=i[a>>0]|0;e=a}if((g|0)==5){h=-1.0;j=f;g=6}else if((g|0)==7){f=xx(k)|0;if(!k){l=1;m=1.0;n=0}else{o=f;p=1.0;q=k;g=8}}if((g|0)==6){k=j+1|0;o=xx(k)|0;p=h;q=k;g=8}b:do if((g|0)==8)if((o|0)!=0?(k=q+o|0,(o|0)>0):0){j=q;while(1){if((i[j>>0]|0)==44){r=j;break}j=j+1|0;if(j>>>0>=k>>>0){l=0;m=p;n=q;break b}}if(r){i[r>>0]=32;l=0;m=p;n=q}else{l=0;m=p;n=q}}else{l=0;m=p;n=q}while(0);q=xx(n)|0;while(1){r=q+-1|0;if((i[n+r>>0]|0)==32)q=r;else{s=q;break}}q=xx(n)|0;c:do if(!l){do if((q|0)!=0?(r=n+q|0,(q|0)>0):0){o=n;do{if((i[o>>0]|0)==58){t=o;g=21;break}o=o+1|0}while(o>>>0<r>>>0);if((g|0)==21?(t|0)!=0:0){u=t;break}if(!s)break c;else g=24}else g=23;while(0);if((g|0)==23)if(!s)break;else g=24;if((g|0)==24){r=n+s|0;if((s|0)>0)v=n;else break;while(1){if((i[v>>0]|0)==32){w=v;break}v=v+1|0;if(v>>>0>=r>>>0)break c}if(!w)break;else u=w}i[u>>0]=0;p=+(ex(n)|0);i[u>>0]=58;r=u+1|0;o=xx(r)|0;d:do if((o|0)!=0?(k=u+(o+1)|0,(o|0)>0):0){j=r;while(1){if((i[j>>0]|0)==58){x=j;g=36;break d}j=j+1|0;if(j>>>0>=k>>>0){g=32;break}}}else g=32;while(0);e:do if((g|0)==32){o=xx(r)|0;if((o|0)!=0?(k=u+(o+1)|0,(o|0)>0):0){o=r;do{if((i[o>>0]|0)==32){x=o;g=36;break e}o=o+1|0}while(o>>>0<k>>>0)}k=xx(r)|0;f:do if((k|0)!=0?(o=u+(k+1)|0,(k|0)>0):0){j=r;while(1){if((i[j>>0]|0)==46)break;j=j+1|0;if(j>>>0>=o>>>0){y=0.0;break f}}y=+dx(r)}else y=0.0;while(0);if(!(i[r>>0]|0)){z=y;A=0.0}else{z=+(ex(r)|0);A=0.0}}while(0);if((g|0)==36){i[x>>0]=0;h=+(ex(r)|0);i[x>>0]=58;z=h;A=+dx(x+1|0)}b=m*(p+z/60.0+A/3600.0);return +b}while(0);if((de(n)|0)!=2){b=m*+(ex(n)|0);return +b}x=rx(n,68)|0;if(x)i[x>>0]=101;x=rx(n,100)|0;if(x)i[x>>0]=101;x=rx(n,69)|0;if(x)i[x>>0]=101;b=m*+dx(n);return +b}function he(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;if((i[c>>0]|0)<64){g=ie(a,b,d)|0;r=e;return g|0}else{ux(f,b)|0;h=xx(b)|0;i[f+h>>0]=i[c>>0]|0;i[f+(h+1)>>0]=0;g=ie(a,f,d)|0;r=e;return g|0}return 0}function ie(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0.0;d=ce(a,b)|0;if(!d){e=0;return e|0}b=(i[d>>0]|0)==35?d+1|0:d;if((xx(b)|0)>81){Bx(602594,b,81)|0;i[602675]=0}else ux(602594,b)|0;if((de(602594)|0)==2){b=rx(602594,68)|0;if(b)i[b>>0]=101;b=rx(602594,100)|0;if(b)i[b>>0]=101;b=rx(602594,69)|0;if(b)i[b>>0]=101}f=+dx(602594);p[c>>3]=f;e=1;return e|0}function je(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0.0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0.0;d=ce(a,b)|0;if(!d){e=0;return e|0}b=rx(d,47)|0;a=rx(d,45)|0;if(b>>>0>d>>>0){i[b>>0]=0;f=~~+dx(d);i[b>>0]=47;g=b+1|0;b=rx(g,47)|0;if(!b)h=rx(g,45)|0;else h=b;if(h>>>0<=d>>>0){e=0;return e|0}i[h>>0]=0;b=~~+dx(g);i[h>>0]=47;g=~~+dx(h+1|0);h=(f|0)>31;j=h?f:g;l=h?g:f;if(j>>>0<50)m=j+2e3|0;else m=(j|0)<100?j+1900|0:j;j=(m&3|0)==0&(((m|0)%100|0|0)!=0|((m|0)%400|0|0)==0);k[26003]=j?29:28;f=b+-1|0;g=k[104008+(f<<2)>>2]|0;n=j?366.0:365.0;j=((l|0)>(g|0)?g:(l|0)<1?1:l)+-1|0;if((b|0)>1){b=0;l=j;while(1){g=(k[104008+(b<<2)>>2]|0)+l|0;b=b+1|0;if((b|0)==(f|0)){o=g;break}else l=g}}else o=j;p[c>>3]=+(m|0)+ +(o|0)/n;e=1;return e|0}if(a>>>0<=d>>>0){e=0;return e|0}i[a>>0]=0;o=~~+dx(d);i[a>>0]=45;m=a+1|0;a=rx(m,45)|0;do if(a>>>0>d>>>0){i[a>>0]=0;j=~~+dx(m);i[a>>0]=45;l=a+1|0;f=rx(l,84)|0;if(f>>>0>d>>>0){i[f>>0]=0;b=~~+dx(l);i[f>>0]=84;q=b;r=j;s=f;break}else{q=~~+dx(l);r=j;s=f;break}}else{q=1;r=1;s=0}while(0);a=(o|0)<32;m=a?q+1900|0:o;f=a?o:q;q=(((m|0)%100|0|0)!=0|((m|0)%400|0|0)==0)&(m&3|0)==0;k[26003]=q?29:28;o=r+-1|0;a=k[104008+(o<<2)>>2]|0;n=q?366.0:365.0;q=((f|0)>(a|0)?a:(f|0)<1?1:f)+-1|0;if((r|0)>1){r=0;f=q;while(1){a=(k[104008+(r<<2)>>2]|0)+f|0;r=r+1|0;if((r|0)==(o|0)){t=a;break}else f=a}}else t=q;p[c>>3]=+(m|0)+ +(t|0)/n;if(s>>>0<=d>>>0){e=1;return e|0}t=s+1|0;s=rx(t,58)|0;do if(s>>>0>d>>>0){i[s>>0]=0;m=~~+dx(t);i[s>>0]=58;q=s+1|0;f=rx(q,58)|0;if(f>>>0>d>>>0){i[f>>0]=0;o=~~+dx(q);i[f>>0]=58;u=m;v=o;w=+dx(f+1|0);break}else{u=m;v=~~+dx(q);w=0.0;break}}else{u=0;v=0;w=0.0}while(0);p[c>>3]=+p[c>>3]+(w+(+(u|0)*3600.0+ +(v|0)*60.0))/86400.0/n;e=1;return e|0}function ke(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;e=r;r=r+64|0;f=e+32|0;g=e+24|0;h=e+16|0;j=e+8|0;l=e+40|0;m=e;k[j>>2]=b;Xw(l,602756,j)|0;do if(!($d(a,l)|0)){k[h>>2]=b;Xw(l,602767,h)|0;if($d(a,l)|0){j=m;k[j>>2]=627012389;k[j+4>>2]=6566448;break}k[g>>2]=b;Xw(l,602773,g)|0;if($d(a,l)|0){j=m;k[j>>2]=627012389;k[j+4>>2]=6566704;break}if(!($d(a,l)|0)){n=0;r=e;return n|0}else{j=m;k[j>>2]=627012389;k[j+4>>2]=6566704;break}}else{i[m>>0]=i[602761]|0;i[m+1>>0]=i[602762]|0;i[m+2>>0]=i[602763]|0;i[m+3>>0]=i[602764]|0;i[m+4>>0]=i[602765]|0;i[m+5>>0]=i[602766]|0}while(0);k[26001]=1;g=1;h=c;c=d;while(1){k[f>>2]=b;k[f+4>>2]=g;Xw(l,m,f)|0;j=ce(a,l)|0;if(!j){o=g;break}p=xx(j)|0;if((h|0)<=(p|0)){q=j;s=g;t=h;u=c;v=13;break}ux(c,j)|0;j=g+1|0;if((j|0)<500){g=j;h=h-p|0;c=c+p|0}else{o=j;break}}do if((v|0)==13)if((t|0)>1){Bx(u,q,t+-1|0)|0;i[u+t>>0]=0;o=s;break}else{i[d>>0]=i[q>>0]|0;o=s;break}while(0);k[26001]=0;n=(o|0)>1&1;r=e;return n|0}function le(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0;f=r;r=r+16|0;g=f;do if((i[c>>0]|0)<64){h=ce(a,b)|0;if(h){if((xx(h)|0)<(d|0)){ux(e,h)|0;j=1;break}if((d|0)>1){Bx(e,h,d+-1|0)|0;j=1;break}else{i[e>>0]=i[h>>0]|0;j=1;break}}else j=0}else{ux(g,b)|0;h=xx(b)|0;i[g+h>>0]=i[c>>0]|0;i[g+(h+1)>>0]=0;h=ce(a,g)|0;if(h){if((xx(h)|0)<(d|0)){ux(e,h)|0;j=1;break}if((d|0)>1){Bx(e,h,d+-1|0)|0;j=1;break}else{i[e>>0]=i[h>>0]|0;j=1;break}}else j=0}while(0);r=f;return j|0}function me(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=ce(a,b)|0;if(!e){f=0;return f|0}if((xx(e)|0)<(c|0)){ux(d,e)|0;f=1;return f|0}if((c|0)>1){Bx(d,e,c+-1|0)|0;f=1;return f|0}else{i[d>>0]=i[e>>0]|0;f=1;return f|0}return 0}function ne(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0;c=xx(a)|0;if((a|0)==0|(b|0)==0){d=0;return d|0}e=xx(b)|0;if(!e){d=a;return d|0}if(!c){d=0;return d|0}f=i[b>>0]|0;g=e+-1|0;h=i[b+g>>0]|0;j=c+1-e|0;c=a+j|0;if((j|0)<=0){d=0;return d|0}j=(e|0)==1;k=(e|0)<3;l=a;a:while(1){b:do if((i[l>>0]|0)==f<<24>>24){if(j){d=l;m=13;break a}if((i[l+g>>0]|0)==h<<24>>24){if(k){d=l;m=13;break a}else n=1;while(1){if((i[l+n>>0]|0)!=(i[b+n>>0]|0))break b;n=n+1|0;if((n|0)>=(e|0)){d=l;m=13;break a}}}}while(0);l=l+1|0;if(l>>>0>=c>>>0){d=0;m=13;break}}if((m|0)==13)return d|0;return 0}function oe(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0;if((a|0)==0|(b|0)==0){d=0;return d|0}e=xx(b)|0;if(!e){d=a;return d|0}if(!c){d=0;return d|0}do if((e|0)<3){f=i[b>>0]|0;g=f<<24>>24;if((f+-97&255)>=26)if((f+-65&255)<26)h=g+32&255;else h=f;else h=g+224&255;if((e|0)>1){g=i[b+1>>0]|0;j=g<<24>>24;if((g+-97&255)<26){k=f;l=g;m=h;n=j+224&255;o=0;break}if((g+-65&255)<26){k=f;l=g;m=h;n=j+32&255;o=0}else{k=f;l=g;m=h;n=g;o=0}}else{k=f;l=32;m=h;n=32;o=0}}else{f=ey(e,1)|0;g=0;do{j=i[b+g>>0]|0;p=j&255;do if((j+-97&255)>=26)if((j+-65&255)<26){i[f+g>>0]=p+32;break}else{i[f+g>>0]=j;break}else i[f+g>>0]=p+224;while(0);g=g+1|0}while((g|0)!=(e|0));g=e+-1|0;k=i[b>>0]|0;l=i[b+g>>0]|0;m=i[f>>0]|0;n=i[f+g>>0]|0;o=f}while(0);h=c+1-e|0;c=a+h|0;a:do if((h|0)>0){g=e+-1|0;p=(e|0)>1;switch(e|0){case 1:{j=a;while(1){q=i[j>>0]|0;if(q<<24>>24==k<<24>>24?1:q<<24>>24==m<<24>>24){r=j;break}j=j+1|0;if(j>>>0>=c>>>0)break a}if(!o){d=r;return d|0}dy(o);d=r;return d|0}case 2:{j=a;while(1){f=i[j>>0]|0;if((f<<24>>24==k<<24>>24?1:f<<24>>24==m<<24>>24)?(f=i[j+g>>0]|0,f<<24>>24==l<<24>>24?1:f<<24>>24==n<<24>>24):0){s=j;break}j=j+1|0;if(j>>>0>=c>>>0)break a}if(!o){d=s;return d|0}dy(o);d=s;return d|0}default:{j=a;b:while(1){f=i[j>>0]|0;c:do if((f<<24>>24==k<<24>>24?1:f<<24>>24==m<<24>>24)?(q=i[j+g>>0]|0,q<<24>>24==l<<24>>24?1:q<<24>>24==n<<24>>24):0){if(p)t=1;else{u=j;break b}while(1){q=i[j+t>>0]|0;if(q<<24>>24!=(i[b+t>>0]|0)?q<<24>>24!=(i[o+t>>0]|0):0)break c;t=t+1|0;if((t|0)>=(e|0)){u=j;break b}}}while(0);j=j+1|0;if(j>>>0>=c>>>0)break a}if(!o){d=u;return d|0}dy(o);d=u;return d|0}}}while(0);if(!o){d=0;return d|0}dy(o);d=0;return d|0}function pe(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0;if((a|0)==0|(b|0)==0){d=0;return d|0}e=xx(b)|0;if(!e){d=a;return d|0}if(!c){d=0;return d|0}f=i[b>>0]|0;g=e+-1|0;h=i[b+g>>0]|0;j=c+1-e|0;c=a+j|0;if((j|0)<=0){d=0;return d|0}j=(e|0)==1;k=(e|0)<3;l=a;a:while(1){b:do if((i[l>>0]|0)==f<<24>>24){if(j){d=l;m=13;break a}if((i[l+g>>0]|0)==h<<24>>24){if(k){d=l;m=13;break a}else n=1;while(1){if((i[l+n>>0]|0)!=(i[b+n>>0]|0))break b;n=n+1|0;if((n|0)>=(e|0)){d=l;m=13;break a}}}}while(0);l=l+1|0;if(l>>>0>=c>>>0){d=0;m=13;break}}if((m|0)==13)return d|0;return 0}function qe(a,b,c,d){a=a|0;b=b|0;c=+c;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0.0,t=0,u=0,v=0,w=0,x=0;e=r;r=r+176|0;f=e+96|0;g=e+80|0;h=e+64|0;j=e+48|0;l=e+32|0;m=e+16|0;n=e;o=e+112|0;q=c<0.0;s=(q?-1.0:1.0)*+Gv(q?-c:c,360.0);c=(s<0.0?s+360.0:s)/15.0;q=~~c;s=(c-+(q|0))*60.0;t=~~s;c=(s-+(t|0))*60.0;do if((d|0)<=5){if((d|0)>4){u=c>59.99999;v=(u&1)+t|0;w=(v|0)>59;k[m>>2]=((w&1)+q|0)%24|0;k[m+4>>2]=w?0:v;p[m+8>>3]=u?0.0:c;Xw(o,602804,m)|0;break}if((d|0)>3){u=c>59.9999;v=(u&1)+t|0;w=(v|0)>59;k[l>>2]=((w&1)+q|0)%24|0;k[l+4>>2]=w?0:v;p[l+8>>3]=u?0.0:c;Xw(o,602821,l)|0;break}if((d|0)>2){u=c>59.999;v=(u&1)+t|0;w=(v|0)>59;k[j>>2]=((w&1)+q|0)%24|0;k[j+4>>2]=w?0:v;p[j+8>>3]=u?0.0:c;Xw(o,602838,j)|0;break}if((d|0)>1){u=c>59.99;v=(u&1)+t|0;w=(v|0)>59;k[h>>2]=((w&1)+q|0)%24|0;k[h+4>>2]=w?0:v;p[h+8>>3]=u?0.0:c;Xw(o,602855,h)|0;break}if((d|0)>0){u=c>59.9;v=(u&1)+t|0;w=(v|0)>59;k[g>>2]=((w&1)+q|0)%24|0;k[g+4>>2]=w?0:v;p[g+8>>3]=u?0.0:c;Xw(o,602872,g)|0;break}else{u=~~(c+.5);v=(u|0)>59;w=(v&1)+t|0;x=(w|0)>59;k[f>>2]=((x&1)+q|0)%24|0;k[f+4>>2]=x?0:w;k[f+8>>2]=v?0:u;Xw(o,602889,f)|0;break}}else{u=c>59.999999;v=(u&1)+t|0;w=(v|0)>59;k[n>>2]=((w&1)+q|0)%24|0;k[n+4>>2]=w?0:v;p[n+8>>3]=u?0.0:c;Xw(o,602787,n)|0}while(0);n=b+-1|0;if((xx(o)|0)<(n|0)){ux(a,o)|0;r=e;return}else{Bx(a,o,n)|0;i[a+n>>0]=0;r=e;return}}function re(a,b,c,d){a=a|0;b=b|0;c=+c;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0.0,t=0,u=0,v=0,w=0,x=0,y=0;e=r;r=r+224|0;f=e+144|0;g=e+120|0;h=e+96|0;j=e+72|0;l=e+48|0;m=e+24|0;n=e;o=e+160|0;q=c<0.0;s=(q?-1.0:1.0)*+Gv(q?-c:c,360.0);c=!(s<=-180.0)?s:s+360.0;q=c<0.0;t=q?45:43;s=q?-c:c;q=~~s;c=(s-+(q|0))*60.0;u=~~c;s=(c-+(u|0))*60.0;do if((d|0)<=5){if((d|0)>4){v=s>59.99999;w=(v&1)+u|0;x=(w|0)>59;k[m>>2]=t&255;k[m+4>>2]=(x&1)+q;k[m+8>>2]=x?0:w;p[m+16>>3]=v?0.0:s;Xw(o,602923,m)|0;break}if((d|0)>3){v=s>59.9999;w=(v&1)+u|0;x=(w|0)>59;k[l>>2]=t&255;k[l+4>>2]=(x&1)+q;k[l+8>>2]=x?0:w;p[l+16>>3]=v?0.0:s;Xw(o,602942,l)|0;break}if((d|0)>2){v=s>59.999;w=(v&1)+u|0;x=(w|0)>59;k[j>>2]=t&255;k[j+4>>2]=(x&1)+q;k[j+8>>2]=x?0:w;p[j+16>>3]=v?0.0:s;Xw(o,602961,j)|0;break}if((d|0)>1){v=s>59.99;w=(v&1)+u|0;x=(w|0)>59;k[h>>2]=t&255;k[h+4>>2]=(x&1)+q;k[h+8>>2]=x?0:w;p[h+16>>3]=v?0.0:s;Xw(o,602980,h)|0;break}if((d|0)>0){v=s>59.9;w=(v&1)+u|0;x=(w|0)>59;k[g>>2]=t&255;k[g+4>>2]=(x&1)+q;k[g+8>>2]=x?0:w;p[g+16>>3]=v?0.0:s;Xw(o,602999,g)|0;break}else{v=~~(s+.5);w=(v|0)>59;x=(w&1)+u|0;y=(x|0)>59;k[f>>2]=t&255;k[f+4>>2]=(y&1)+q;k[f+8>>2]=y?0:x;k[f+12>>2]=w?0:v;Xw(o,603018,f)|0;break}}else{v=s>59.999999;w=(v&1)+u|0;x=(w|0)>59;k[n>>2]=t&255;k[n+4>>2]=(x&1)+q;k[n+8>>2]=x?0:w;p[n+16>>3]=v?0.0:s;Xw(o,602904,n)|0}while(0);n=b+-1|0;if((xx(o)|0)<(n|0)){ux(a,o)|0;r=e;return}else{Bx(a,o,n)|0;i[a+n>>0]=0;r=e;return}}function se(a,b,c,d){a=a|0;b=b|0;c=+c;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0.0;e=r;r=r+112|0;f=e+24|0;g=e+16|0;h=e+8|0;j=e;l=e+32|0;m=e+40|0;n=c<0.0;o=(n?-1.0:1.0)*+Gv(n?-c:c,360.0);c=!(o<=-180.0)?o:o+360.0;n=d+4|0;if((d|0)>0){k[j>>2]=n;k[j+4>>2]=d;Xw(l,603035,j)|0;p[h>>3]=c;Xw(m,l,h)|0}else{k[g>>2]=n;Xw(l,603044,g)|0;k[f>>2]=~~c;Xw(m,l,f)|0}f=b+-1|0;if((xx(m)|0)<(f|0)){ux(a,m)|0;r=e;return}else{Bx(a,m,f)|0;i[a+f>>0]=0;r=e;return}}function te(a,b,c,d){a=a|0;b=+b;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0;e=r;r=r+64|0;f=e+48|0;g=e+40|0;h=e+32|0;i=e+24|0;j=e+16|0;l=e+8|0;m=e;n=e+56|0;o=(d|0)>0;if((c|0)>0)if(o){k[m>>2]=c;k[m+4>>2]=d;Xw(n,603035,m)|0;p[l>>3]=b;Xw(a,n,l)|0;r=e;return}else{k[j>>2]=c;Xw(n,603050,j)|0;k[i>>2]=~~b;Xw(a,n,i)|0;r=e;return}else if(o){k[h>>2]=d;Xw(n,602780,h)|0;p[g>>3]=b;Xw(a,n,g)|0;r=e;return}else{k[f>>2]=~~b;Xw(a,772183,f)|0;r=e;return}}function ue(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0.0;d=r;r=r+32|0;e=d;f=xe(a,b)|0;if(!f){g=0;r=d;return g|0}ux(e,f)|0;h=+dx(e);p[c>>3]=h;g=1;r=d;return g|0}function ve(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0;f=cy(2e3)|0;if(!(ke(a,b,2e3,f)|0)){dy(f);g=0;return g|0}b=xe(f,c)|0;do if(b){if((xx(b)|0)<(d|0)){ux(e,b)|0;h=1;break}if((d|0)>1){Bx(e,b,d+-1|0)|0;h=1;break}else{i[e>>0]=i[b>>0]|0;h=1;break}}else h=0;while(0);dy(f);g=h;return g|0}function we(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=xe(a,b)|0;if(!e){f=0;return f|0}if((xx(e)|0)<(c|0)){ux(d,e)|0;f=1;return f|0}if((c|0)>1){Bx(d,e,c+-1|0)|0;f=1;return f|0}else{i[d>>0]=i[e>>0]|0;f=1;return f|0}return 0}function xe(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;c=r;r=r+2032|0;d=c+24|0;e=c+18|0;f=c+16|0;g=c;h=c+32|0;i[e>>0]=91;i[e+1>>0]=0;i[f>>0]=93;i[f+1>>0]=0;ux(g,b)|0;b=ne(g,e)|0;e=(b|0)!=0;if(e){i[b>>0]=0;j=0}else j=0;while(1){if(!(i[a+j>>0]|0)){k=j;l=6;break}m=j+1|0;if((m|0)<57600)j=m;else{n=m;l=5;break}}if((l|0)==5){o=n;p=xx(g)|0}else if((l|0)==6){n=xx(g)|0;if((k|0)>0){o=k;p=n}else{q=0;r=c;return q|0}}n=a+o|0;o=n;k=a;a:while(1){j=pe(k,g,o-k|0)|0;if(!j){q=0;l=33;break}m=i[j+p>>0]|0;s=i[j+-1>>0]|0;if(m<<24>>24>32)switch(m<<24>>24){case 61:case 127:{l=11;break}default:{}}else l=11;if((l|0)==11){l=0;if((j|0)==(a|0)){t=a;l=14;break}switch(s<<24>>24){case 9:case 32:{t=j;l=14;break a;break}default:{}}}k=j+1|0;if(k>>>0>=n>>>0){q=0;l=33;break}}if((l|0)==14){n=t+p|0;b:while(1){switch(i[n>>0]|0){case 61:case 32:break;default:{u=n;break b}}n=n+1|0}if(!u){q=0;r=c;return q|0}sy(h|0,0,2e3)|0;n=i[u>>0]|0;c:do if(n<<24>>24==34){p=u+1|0;t=i[p>>0]|0;if(t<<24>>24){k=t;t=0;a=p;while(1){if(!((t|0)<2e3&k<<24>>24!=34))break c;i[h+t>>0]=k;a=a+1|0;k=i[a>>0]|0;if(!(k<<24>>24))break;else t=t+1|0}}}else{t=n;k=0;a=u;while(1){switch(t<<24>>24){case 9:case 32:{break c;break}default:{}}if(!((k|0)<2e3&t<<24>>24>0))break c;p=a+1|0;i[h+k>>0]=t;t=i[p>>0]|0;k=k+1|0;a=p}}while(0);if(!e){ux(603056,h)|0;q=603056;r=c;return q|0}e=b+1|0;b=ne(e,f)|0;if(!b){q=603056;r=c;return q|0}i[b>>0]=0;b=ex(e)|0;if((b|0)<=0){q=603056;r=c;return q|0}i[d>>0]=32;i[d+1>>0]=44;i[d+2>>0]=0;e=Fx(h,d)|0;if((b|0)>1){h=1;do{f=Fx(0,d)|0;h=h+1|0}while((h|0)!=(b|0));v=f}else v=e;if(!v){q=603056;r=c;return q|0}ux(603056,v)|0;q=603056;r=c;return q|0}else if((l|0)==33){r=c;return q|0}return 0}function ye(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0,r=0.0,s=0.0,t=0.0,u=0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0;f=k[c+3260>>2]|0;g=+p[c>>3];h=+p[c+8>>3];i=+p[c+32>>3];j=+p[c+40>>3];l=+p[c+48>>3]*3.141592653589793/180.0;m=+Z(+l);n=+_(+l);o=a-+p[c+16>>3];a=b-+p[c+24>>3];do if(!(k[c+3300>>2]|0))if(i==0.0|j==0.0){p[d>>3]=0.0;p[e>>3]=0.0;q=2;return q|0}else{b=o*i;r=a*j;if(!(l!=0.0)){s=b;t=r;break}s=b*m-r*n;t=r*m+b*n;break}else{s=o*+p[c+56>>3]+a*+p[c+64>>3];t=o*+p[c+72>>3]+a*+p[c+80>>3]}while(0);u=(k[c+3304>>2]|0)==0;a=u?t:s;o=u?s:t;p[d>>3]=g+o;p[e>>3]=h+a;if((f|0)<1){q=0;return q|0}t=g*3.141592653589793/180.0;g=h*3.141592653589793/180.0;s=u?g:t;l=u?t:g;t=o*3.141592653589793/180.0;o=a*3.141592653589793/180.0;a=t*t;b=a+o*o;r=+Z(+s);v=+_(+s);a:do switch(f|0){case 11:{w=s+o;x=l+t;break}case 4:{if(b>1.0){q=1;return q|0}y=+X(+(1.0-b));z=o*r+v*y;if(z>1.0|z<-1.0){q=1;return q|0}A=r*y-o*v;if(t==0.0&A==0.0){q=1;return q|0}else{w=+ba(+z);x=l+ +da(+t,+A);break a}break}case 32:case 33:case 31:case 3:{if(b>1.0){q=1;return q|0}A=r-o*v;if(A==0.0){q=1;return q|0}else{z=l+ +da(+t,+A);w=+ca(+((o*r+v)*+Z(+(z-l))/A));x=z;break a}break}case 6:{if(b>=9.869604401089369){q=1;return q|0}z=+X(+b);A=+Z(+z);if(z!=0.0)B=+_(+z)/z;else B=1.0;z=v*A+o*r*B;if(z>1.0|z<-1.0){q=1;return q|0}y=A-v*z;A=r*(t*B);if(A==0.0&y==0.0){q=1;return q|0}else{w=+ba(+z);x=l+ +da(+A,+y);break a}break}case 27:{y=r-o*v;if(y==0.0){q=1;return q|0}A=l+ +da(+t,+y);z=+Z(+(A-l));if(z==0.0){q=1;return q|0}C=y/z;if(C>1.0|C<-1.0){q=1;return q|0}z=+aa(+C);if(s<0.0){w=-z;x=A}else{w=z;x=A}break}case 20:case 28:{A=s+o;if(+W(+A)>1.5707963267948974){q=1;return q|0}z=+Z(+A);if(+W(+t)>z*6.28318530717959*.5){q=1;return q|0}if(z>1.0e-05){w=A;x=l+t/z}else{w=A;x=l}break}case 12:{A=j*m+i*n;z=A==0.0?1.0:A;A=(h*.5+45.0)*3.141592653589793/180.0;C=+fa(+(+$(+A)));y=z*3.141592653589793/180.0/(+fa(+(+$(+(z*.5*.01745329252+A))))-C);A=+Z(+g);z=l+t/(!(A<=0.0)?A:1.0);if(+W(+(z-l))>6.28318530717959){q=1;return q|0}else{w=+ca(+(+ea(+(y!=0.0?(o+C*y)/y:0.0))))*2.0+-1.5707963267948974;x=z;break a}break}case 22:{z=j*m+i*n;y=z==0.0?.017453292519943295:z*3.141592653589793/180.0;z=g+y;C=+_(+g);A=+Z(+g);D=+X(+((A+1.0)*.5));E=+_(+z)/+X(+((+Z(+z)+1.0)*.5))-C/D;z=y/(E==0.0?1.0:E);E=i*m-j*n;y=E==0.0?.017453292519943295:E*3.141592653589793/180.0;E=y*.5;F=A*2.0*+_(+E);G=y*+X(+((A*+Z(+E)+1.0)*.5))/(F==0.0?1.0:F);if(!(t==0.0&o==0.0)){F=o+C*z/D;D=F/z;C=4.0-a/(G*(G*4.0))-D*D;if(C>4.0|C<2.0){q=1;return q|0}D=+X(+C)*.5;C=F*D/z;if(+W(+C)>1.0){q=1;return q|0}z=+ba(+C);C=+Z(+z);if(+W(+C)<1.0e-05){q=1;return q|0}F=t*D/(G*2.0*C);if(+W(+F)>1.0){q=1;return q|0}else{w=z;x=l+ +ba(+F)*2.0;break a}}else{w=s;x=l}break}case 5:{F=(4.0-b)/(b+4.0);if(+W(+F)>1.0){q=1;return q|0}z=F+1.0;C=v*F+o*r*z*.5;if(+W(+C)>1.0){q=1;return q|0}F=+ba(+C);C=+Z(+F);if(+W(+C)<1.0e-05){q=1;return q|0}G=t*z/(C*2.0);if(+W(+G)>1.0){q=1;return q|0}z=+ba(+G);G=+_(+F);D=+Z(+z);E=v*G+1.0+r*C*D;if(+W(+E)<1.0e-05){q=1;return q|0}else{u=+W(+((r*G-v*C*D)*2.0/E-o))>1.0e-05;w=F;x=l+(u?3.141592653589795-z:z);break a}break}case 16:{z=1.0/+$(+s)-o;w=+ba(+(1.0/(v*2.0)*(v*v*(1.0-(a+z*z))+1.0)));x=l-+da(+t,+(s<0.0?-z:z))/v;break}default:{w=0.0;x=0.0}}while(0);v=x-l>3.141592653589795?x+-6.28318530717959:x;x=v-l<-3.141592653589795?v+6.28318530717959:v;p[d>>3]=(x<0.0?x+6.28318530717959:x)*180.0/3.141592653589793;p[e>>3]=w*180.0/3.141592653589793;q=0;return q|0}function ze(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0,r=0,s=0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0,V=0.0;f=+p[c>>3];g=+p[c+8>>3];h=+p[c+16>>3];i=+p[c+24>>3];j=+p[c+32>>3];l=+p[c+40>>3];m=+p[c+48>>3]*3.141592653589793/180.0;n=+Z(+m);o=+_(+m);q=k[c+3260>>2]|0;r=(q|0)>0;if(r){s=(k[c+3304>>2]|0)==0;t=f*3.141592653589793/180.0;u=g*3.141592653589793/180.0;v=s?u:t;w=s?t:u;u=a-(s?f:g);t=h*j;if(t>180.0|t<-180.0){t=u>360.0?a+-360.0:a;if(u<0.0)x=t+360.0;else x=t}else{t=u>180.0?a+-360.0:a;if(u<-180.0)x=t+360.0;else x=t}t=x*3.141592653589793/180.0;u=b*3.141592653589793/180.0;y=+Z(+u);z=+_(+u);A=t-w;B=x;C=y;D=u;E=v;F=y*+_(+A);G=t;H=w;I=z;J=z*+_(+v)+y*+Z(+v)*+Z(+A)}else{B=a;C=0.0;D=0.0;E=0.0;F=0.0;G=0.0;H=0.0;I=0.0;J=0.0}a:do switch(q|0){case 11:{K=G-H;L=D-E;break}case 4:{if(J<0.0){M=1;return M|0}else{K=F;L=I*+Z(+E)-C*+_(+E)*+Z(+(G-H));break a}break}case 3:case 32:case 33:case 31:{if(!(J<=0.0)){a=+_(+E);A=+Z(+E);v=+Z(+(G-H));y=I*a+C*A*v;K=F/y;L=(I*A-C*a*v)/y;break a}else{M=1;return M|0}break}case 6:{y=+_(+E);v=+Z(+E);a=+Z(+(G-H));A=I*y+C*v*a;z=A<-1.0?-1.0:A;A=+aa(+(z>1.0?1.0:z));if(A!=0.0)N=A/+_(+A);else N=1.0;K=F*N;L=(I*v-C*y*a)*N;break}case 27:{if(E==0.0){M=1;return M|0}else{K=F;L=(+Z(+E)-C*+Z(+(G-H)))/+_(+E);break a}break}case 20:case 28:{if(+W(+D)>1.5707963267948974){M=1;return M|0}if(+W(+E)>1.5707963267948974){M=1;return M|0}else{K=C*(G-H);L=D-E;break a}break}case 12:{a=l*n+j*o;y=a==0.0?1.0:a;a=(g*.5+45.0)*3.141592653589793/180.0;v=+fa(+(+$(+a)));A=y*3.141592653589793/180.0/(+fa(+(+$(+(y*.5*.01745329252+a))))-v);a=+Z(+(g*3.141592653589793/180.0));y=+$(+(D*.5+.7853981633974487));if(y<1.0e-05){M=2;return M|0}else{K=(G-H)*(!(a<=0.0)?a:1.0);L=A*+fa(+y)-v*A;break a}break}case 22:{A=(G-H)*.5;if(+W(+A)>1.5707963267948974){M=1;return M|0}v=l*n+j*o;y=v==0.0?.017453292519943295:v*3.141592653589793/180.0;v=g*3.141592653589793/180.0;a=v+y;z=+_(+v);w=+Z(+v);v=+X(+((w+1.0)*.5));t=+_(+a)/+X(+((+Z(+a)+1.0)*.5))-z/v;a=y/(t==0.0?1.0:t);t=j*n-l*o;y=t==0.0?.017453292519943295:t*3.141592653589793/180.0;t=y*.5;u=w*2.0*+_(+t);x=+Z(+D);O=+X(+((x*+Z(+A)+1.0)*.5));if(+W(+O)<1.0e-05){M=3;return M|0}else{K=+_(+A)*(x*(y*+X(+((w*+Z(+t)+1.0)*.5))/(u==0.0?1.0:u)*2.0))/O;L=a*+_(+D)/O-z*a/v;break a}break}case 5:{if(+W(+D)>1.5707963267948974){M=1;return M|0}v=+_(+E);a=+Z(+E);z=+Z(+(G-H));O=I*v+1.0+C*a*z;if(+W(+O)<1.0e-05){M=1;return M|0}else{u=2.0/O;K=F*u;L=(I*a-C*v*z)*u;break a}break}case 16:{u=+$(+E);z=1.0/u;v=u*-2.0;a=u*u;O=u/3.0;u=v*v;t=(H-G)*+_(+E);w=D-E;y=z*(w*(v*.5+w*(a*.5-u*.125+w*(O*.5+(a*(v*-.25)+v*u*.0625)+w*(a*.1875*u+(a*(a*-.125)-v*.25*O)-u*u*.0390625))))+1.0);u=t*t;if(!(k[c+3300>>2]|0))P=j/+W(+j);else{O=+p[c+56>>3];P=O/+W(+O)}K=(1.0-u/6.0)*(t*y)*P;L=z-(1.0-u*.5)*y;break}default:{K=F;L=0.0}}while(0);if(r){Q=K*180.0/3.141592653589793;R=L*180.0/3.141592653589793}else{Q=B-f;R=b-g}r=(k[c+3304>>2]|0)==0;g=r?R:Q;b=r?Q:R;if(!(k[c+3300>>2]|0)){if(m!=0.0){S=n*b+o*g;T=n*g-o*b}else{S=b;T=g}o=j!=0.0?S/j:S;if(l!=0.0){U=o;V=T/l}else{U=o;V=T}}else{U=b*+p[c+88>>3]+g*+p[c+96>>3];V=b*+p[c+104>>3]+g*+p[c+112>>3]}g=h+U;p[d>>3]=g;do if((q|0)==11){U=+p[c+136>>3];if(g>U){h=g-360.0/j;if(!(h>0.0))break;p[d>>3]=h;break}if(g<0.0?(h=g+360.0/j,h<=U):0)p[d>>3]=h}while(0);p[e>>3]=i+V;M=0;return M|0}function Ae(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0;f=k[c+3268>>2]|0;g=k[c+3272>>2]|0;h=a-+p[c+616>>3];a=b-+p[c+624>>3];b=h*h;i=a*a;j=h*b;l=a*i;m=b+i;n=+p[c+256>>3]+h*+p[c+264>>3]+a*+p[c+272>>3]+b*+p[c+280>>3]+i*+p[c+288>>3]+a*(h*+p[c+296>>3]);if((f|0)>6){o=n+j*+p[c+304>>3]+l*+p[c+312>>3];if((f|0)>8)q=m*+p[c+336>>3]+(o+a*(b*+p[c+320>>3])+i*(h*+p[c+328>>3]))+m*(h*+p[c+344>>3])+m*(a*+p[c+352>>3]);else q=o}else q=n;n=+p[c+416>>3]+h*+p[c+424>>3]+a*+p[c+432>>3]+b*+p[c+440>>3]+i*+p[c+448>>3]+a*(h*+p[c+456>>3]);if((g|0)>6){o=n+j*+p[c+464>>3]+l*+p[c+472>>3];if((g|0)>8)r=m*+p[c+496>>3]+(o+a*(b*+p[c+480>>3])+h*(i*+p[c+488>>3]))+m*(h*+p[c+504>>3])+m*(a*+p[c+512>>3]);else r=o}else r=n;n=r*3.141592653589793/180.0;r=+p[c+696>>3]*3.141592653589793/180.0;o=+$(+r);a=1.0-n*o;m=+da(+(q*3.141592653589793/180.0/+Z(+r)),+a);r=+p[c+688>>3]*3.141592653589793/180.0+m;p[d>>3]=(r<0.0?r+6.28318530717959:r)*180.0/3.141592653589793;r=+ca(+(+Z(+m)/(a/(n+o))))*180.0/3.141592653589793;p[e>>3]=r;return 0}function Be(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0,u=0,v=0,w=0,x=0,y=0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0,U=0.0,V=0.0,X=0.0,Y=0.0,_=0.0,aa=0.0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ha=0.0,ia=0.0,ja=0.0,ka=0.0,la=0.0,ma=0.0,na=0.0,oa=0.0,pa=0.0,qa=0.0,ra=0.0,sa=0.0,ta=0.0,ua=0.0,va=0.0,wa=0.0,xa=0.0,ya=0.0,za=0.0,Aa=0;f=k[c+3268>>2]|0;g=k[c+3272>>2]|0;h=+$(+(b*3.141592653589793/180.0));b=+p[c+696>>3]*3.141592653589793/180.0;i=+$(+b);j=a*3.141592653589793/180.0-+p[c+688>>3]*3.141592653589793/180.0;a=+Z(+j);l=(1.0-i*a/h)/(i+a/h);h=+Z(+b)*+$(+j)*(1.0-i*l)*180.0/3.141592653589793;i=l*180.0/3.141592653589793;l=h*+p[c+88>>3]+i*+p[c+96>>3];j=h*+p[c+104>>3]+i*+p[c+112>>3];b=+p[c+256>>3];a=+p[c+264>>3];m=+p[c+272>>3];n=+p[c+280>>3];o=+p[c+288>>3];q=+p[c+296>>3];r=n*2.0;s=o*2.0;t=(f|0)>8;u=c+320|0;v=c+328|0;w=c+336|0;x=c+344|0;y=c+352|0;z=+p[c+416>>3];A=+p[c+424>>3];B=+p[c+432>>3];C=+p[c+440>>3];D=+p[c+448>>3];E=+p[c+456>>3];F=C*2.0;G=D*2.0;H=(g|0)>6;I=c+464|0;J=c+472|0;K=(g|0)>8;g=c+480|0;L=c+488|0;M=c+496|0;N=c+504|0;O=c+512|0;a:do if((f|0)>6){P=+p[c+304>>3];Q=+p[c+312>>3];R=P*3.0;S=Q*3.0;T=0;U=l;V=j;while(1){X=V*U;Y=U*U;_=V*V;aa=U*Y;ba=V*_;ca=V*Y;da=_*U;ea=_+Y;fa=b+U*a+V*m+Y*n+_*o+X*q+aa*P+ba*Q;ga=a+U*r+V*q+Y*R;ha=U*q+(m+V*s)+_*S;if(t){ia=+p[u>>3];ja=+p[v>>3];ka=+p[w>>3];la=+p[x>>3];ma=+p[y>>3];na=ka*2.0;oa=fa+ca*ia+da*ja+ea*ka+ea*(U*la)+ea*(V*ma);pa=(_+Y*3.0)*la+(_*ja+(ga+X*(ia*2.0))+U*na)+X*(ma*2.0);qa=(_*3.0+Y)*ma+(ha+Y*ia+X*(ja*2.0)+V*na+X*(la*2.0))}else{oa=fa;pa=ga;qa=ha}ha=z+U*A+V*B+Y*C+_*D+X*E;ga=A+U*F+V*E;fa=U*E+(B+V*G);if(H){la=+p[I>>3];na=+p[J>>3];ja=ha+aa*la+ba*na;ba=ga+Y*(la*3.0);la=fa+_*(na*3.0);if(K){na=+p[g>>3];aa=+p[L>>3];ia=+p[M>>3];ma=+p[N>>3];ka=+p[O>>3];ra=ia*2.0;sa=ja+ca*na+da*aa+ea*ia+ea*(U*ma)+ea*(V*ka);ta=(_+Y*3.0)*ma+(_*aa+(ba+X*(na*2.0))+U*ra)+X*(ka*2.0);ua=(_*3.0+Y)*ka+(la+Y*na+X*(aa*2.0)+V*ra+X*(ma*2.0))}else{sa=ja;ta=ba;ua=la}}else{sa=ha;ta=ga;ua=fa}fa=oa-h;ga=sa-i;ha=pa*ua-qa*ta;la=(qa*ga-fa*ua)/ha;ba=(fa*ta-pa*ga)/ha;ha=U+la;ga=V+ba;if(+W(+la)<5.0e-07?+W(+ba)<5.0e-07:0){va=ha;wa=ga;break a}T=T+1|0;if((T|0)>=50){va=ha;wa=ga;break}else{U=ha;V=ga}}}else{T=0;V=l;U=j;while(1){S=U*V;R=V*V;Q=U*U;P=U*R;ga=Q*V;ha=Q+R;ba=b+V*a+U*m+R*n+Q*o+S*q;la=a+V*r+U*q;fa=V*q+(m+U*s);ja=z+V*A+U*B+R*C+Q*D+S*E;ma=A+V*F+U*E;X=V*E+(B+U*G);if(H){ra=+p[I>>3];aa=+p[J>>3];na=ja+V*R*ra+U*Q*aa;Y=ma+R*(ra*3.0);ra=X+Q*(aa*3.0);if(K){aa=+p[g>>3];ka=+p[L>>3];_=+p[M>>3];ea=+p[N>>3];ia=+p[O>>3];da=_*2.0;xa=na+P*aa+ga*ka+ha*_+ha*(V*ea)+ha*(U*ia);ya=(Q+R*3.0)*ea+(Q*ka+(Y+S*(aa*2.0))+V*da)+S*(ia*2.0);za=(Q*3.0+R)*ia+(ra+R*aa+S*(ka*2.0)+U*da+S*(ea*2.0))}else{xa=na;ya=Y;za=ra}}else{xa=ja;ya=ma;za=X}X=ba-h;ba=xa-i;ma=la*za-fa*ya;ja=(fa*ba-X*za)/ma;fa=(X*ya-la*ba)/ma;ma=V+ja;ba=U+fa;if(+W(+ja)<5.0e-07?+W(+fa)<5.0e-07:0){va=ma;wa=ba;break a}T=T+1|0;if((T|0)>=50){va=ma;wa=ba;break}else{V=ma;U=ba}}}while(0);p[d>>3]=va+ +p[c+616>>3];va=wa+ +p[c+624>>3];p[e>>3]=va;wa=+p[d>>3];if(wa<.5){Aa=-1;return Aa|0}if(va<.5?1:wa>+p[c+136>>3]+.5){Aa=-1;return Aa|0}if(va>+p[c+144>>3]+.5){Aa=-1;return Aa|0}Aa=0;return Aa|0}function Ce(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0;c=cy(2e3)|0;d=cy(2e3)|0;ke(a,605354,2e3,c)|0;ke(a,605520,2e3,d)|0;a=cy(2e3)|0;e=cy(2e3)|0;f=b+3176|0;if((+p[f>>3]>360.0?(ue(c,605525,f)|0)==0:0)?(ue(d,605525,f)|0)==0:0)p[f>>3]=180.0;f=b+3192|0;if((ue(c,605534,f)|0)==0?(ue(d,605534,f)|0)==0:0)p[f>>3]=57.29577951308232;do if(!(we(c,605545,2e3,a)|0))if(!(we(d,605545,2e3,a)|0)){k[b+5960>>2]=0;break}else{f=Fe(a)|0;k[b+5960>>2]=f;break}else{f=Fe(a)|0;k[b+5960>>2]=f}while(0);do if(!(we(d,605552,2e3,e)|0))if(!(we(c,605552,2e3,e)|0)){f=b+5964|0;k[f>>2]=0;g=f;break}else{f=Fe(e)|0;h=b+5964|0;k[h>>2]=f;g=h;break}else{h=Fe(e)|0;f=b+5964|0;k[f>>2]=h;g=f}while(0);zd(b);dy(c);dy(d);dy(a);dy(e);if((k[g>>2]|0)==0?(k[b+5960>>2]|0)==0:0){i=1;return i|0}i=0;return i|0}function De(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0,s=0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0;f=a-+p[c+616>>3];a=b-+p[c+624>>3];if(!(k[c+3300>>2]|0)){b=+p[c+760>>3];if(b==0.0){g=2;h=0.0;i=0.0;p[d>>3]=i;p[e>>3]=h;return g|0}j=+p[c+768>>3];if(j==0.0){g=2;h=0.0;i=0.0;p[d>>3]=i;p[e>>3]=h;return g|0}l=f*b;b=a*j;j=+p[c+48>>3];if(j!=0.0){m=j*3.141592653589793/180.0;j=+Z(+m);n=+_(+m);o=l*j-b*n;q=b*j+l*n}else{o=l;q=b}}else{o=f*+p[c+56>>3]+a*+p[c+64>>3];q=f*+p[c+72>>3]+a*+p[c+80>>3]}r=(k[c+3304>>2]|0)==0&1;a=(90.0-+p[c+688+(r<<3)>>3])*3.141592653589793/180.0;f=+Z(+a);b=+_(+a);l=+p[c+3176>>3]*3.141592653589793/180.0;s=k[c+5960>>2]|0;if(!s)t=o;else t=o+ +Ge(s,o,q);s=k[c+5964>>2]|0;if(!s)u=q;else u=q+ +Ge(s,o,q);q=+X(+(t*t+u*u));if(q==0.0)v=0.0;else v=+da(+t,+-u);u=+da(+(+p[c+3192>>3]),+q);q=+Z(+u);t=+_(+u);o=v-l;l=+Z(+o);v=+_(+o);n=f*q;j=b*t-n*l;if(+W(+j)<1.0e-05)w=n*(1.0-l)-+Z(+(a+u));else w=j;j=q*v;if(j!=-0.0|w!=0.0)x=+da(+-j,+w);else x=o+3.141592653589793;v=+p[c+688+((r^1)<<3)>>3];n=v+x*180.0/3.141592653589793;if(!(v>=0.0))if(n>0.0)y=n+-360.0;else y=n;else if(n<0.0)y=n+360.0;else y=n;if(!(y>360.0))if(y<-360.0)z=y+360.0;else z=y;else z=y+-360.0;if(+Gv(o,3.141592653589793)==0.0){o=(u+a*l)*180.0/3.141592653589793;a=o>90.0?180.0-o:o;if(!(a<-90.0)){g=0;h=a;i=z;p[d>>3]=i;p[e>>3]=h;return g|0}g=0;h=-180.0-a;i=z;p[d>>3]=i;p[e>>3]=h;return g|0}a=f*t+b*q*l;if(!(+W(+a)>.99)){g=0;h=+ba(+a)*180.0/3.141592653589793;i=z;p[d>>3]=i;p[e>>3]=h;return g|0}l=+aa(+(+X(+(j*j+w*w))));if(!(a>=0.0)){g=0;h=l*-180.0/3.141592653589793;i=z;p[d>>3]=i;p[e>>3]=h;return g|0}else{g=0;h=l*180.0/3.141592653589793;i=z;p[d>>3]=i;p[e>>3]=h;return g|0}return 0}function Ee(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0,B=0,C=0,D=0.0,E=0.0,F=0.0,G=0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0,N=0.0,O=0,P=0.0,Q=0.0,R=0.0,S=0,T=0.0,U=0.0,V=0.0;f=c+3304|0;g=(k[f>>2]|0)==0;h=g&1;i=(a-+p[c+688+((h^1)<<3)>>3])*3.141592653589793/180.0;a=b*3.141592653589793/180.0;b=+Z(+i);j=+_(+i);l=+Z(+a);m=+_(+a);n=(90.0-+p[c+688+(h<<3)>>3])*3.141592653589793/180.0;o=+Z(+n);q=+_(+n);r=+p[c+3176>>3];s=l*o;t=m*q-b*s;if(+W(+t)<1.0e-05)u=(1.0-b)*s-+Z(+(a+n));else u=t;t=j*l;if(t!=-0.0|u!=0.0)v=+da(+-t,+u);else v=i+-3.141592653589793;j=(r==999.0?3.141592653589793:r*3.141592653589793/180.0)+v;if(!(j>3.141592653589793))if(j<-3.141592653589793)w=j+6.283185307179586;else w=j;else w=j+-6.283185307179586;do if(+Gv(i,3.141592653589793)==0.0){j=a+b*n;v=j>1.5707963267948966?3.141592653589793-j:j;if(v<-1.5707963267948966)x=-3.141592653589793-v;else x=v}else{v=m*o+b*(l*q);if(!(+W(+v)>.99)){x=+ba(+v);break}j=+aa(+(+X(+(t*t+u*u))));if(!(v>=0.0))x=-j;else x=j}while(0);u=+_(+x);do if(u==0.0){y=0.0;z=0.0}else{t=+p[c+3192>>3]*+Z(+x)/u;h=c+5960|0;A=k[h>>2]|0;B=c+5964|0;if((A|0)==0?(k[B>>2]|0)==0:0){q=t*+_(+w);l=-(t*+Z(+w));y=g?q:l;z=g?l:q;break}q=t*+_(+w);l=t*+Z(+w);C=A;A=0;t=q;b=-l;while(1){if(!C){D=t;E=1.0;F=0.0}else{o=t+ +Ge(C,t,b);m=+He(k[h>>2]|0,t,b,1,0)+1.0;D=o;E=m;F=+He(k[h>>2]|0,t,b,0,1)}m=D-q;G=k[B>>2]|0;if(!G){H=b;I=0.0;J=1.0}else{o=b+ +Ge(G,t,b);n=+He(k[B>>2]|0,t,b,1,0);H=o;I=n;J=+He(k[B>>2]|0,t,b,0,1)+1.0}n=l+H;o=E*J-F*I;if(o==0.0){K=t;L=b;break}a=(F*n-m*J)/o;i=(m*I-E*n)/o;o=t+a;j=b+i;G=+W(+a)>+W(+i);v=G?a:i;i=+W(+v);G=+W(+m)>+W(+n);a=G?m:n;G=i>+W(+a);M=A+1|0;if(!(!(+W(+(G?v:a))<2.8e-08)&(M|0)<500)){K=o;L=j;break}C=k[h>>2]|0;A=M;t=o;b=j}A=(k[f>>2]|0)==0;y=A?K:L;z=A?L:K}while(0);if(k[c+3300>>2]|0){p[d>>3]=y*+p[c+88>>3]+z*+p[c+96>>3];p[e>>3]=y*+p[c+104>>3]+z*+p[c+112>>3];N=+p[d>>3];O=c+16|0;P=+p[O>>3];Q=N+P;p[d>>3]=Q;R=+p[e>>3];S=c+24|0;T=+p[S>>3];U=R+T;p[e>>3]=U;return 0}K=+p[c+48>>3];if(K!=0.0){L=K*3.141592653589793/180.0;K=+Z(+L);E=+_(+L);p[d>>3]=y*K+z*E;V=z*K-y*E}else{p[d>>3]=y;V=z}p[e>>3]=V;V=+p[c+32>>3];if(V!=0.0)p[d>>3]=+p[d>>3]/V;V=+p[c+40>>3];if(!(V!=0.0)){N=+p[d>>3];O=c+16|0;P=+p[O>>3];Q=N+P;p[d>>3]=Q;R=+p[e>>3];S=c+24|0;T=+p[S>>3];U=R+T;p[e>>3]=U;return 0}p[e>>3]=+p[e>>3]/V;N=+p[d>>3];O=c+16|0;P=+p[O>>3];Q=N+P;p[d>>3]=Q;R=+p[e>>3];S=c+24|0;T=+p[S>>3];U=R+T;p[e>>3]=U;return 0}function Fe(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0.0,o=0,q=0,s=0,t=0,u=0,v=0.0,w=0,x=0,y=0;b=r;r=r+16|0;c=b;if(!(i[a+1>>0]|0)){d=0;r=b;return d|0}e=cy(160)|0;k[c>>2]=a;f=a;g=a;a=e;e=0;h=20;a:while(1){j=i[g>>0]|0;while(1){if(!(j<<24>>24)){l=a;m=e;break a}n=+hx(f,c);o=k[c>>2]|0;q=i[o>>0]|0;if(q<<24>>24==46){s=o+1|0;k[c>>2]=s;t=i[s>>0]|0;u=s}else{t=q;u=o}if(!(t<<24>>24))j=0;else{v=n;w=u;break}}j=e+1|0;if((j|0)<(h|0)){x=a;y=h}else{o=h+20|0;x=fy(a,o<<3)|0;y=o}p[x+(e<<3)>>3]=v;o=w;while(1)if((i[o>>0]|0)==32)o=o+1|0;else{f=o;g=w;a=x;e=j;h=y;continue a}}y=Ie(l)|0;dy(l);d=(m|0)==0?0:y;r=b;return d|0}function Ge(a,b,c){a=a|0;b=+b;c=+c;var d=0,e=0.0,f=0.0,g=0,h=0.0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0.0,A=0,B=0,C=0,D=0,E=0,F=0.0,G=0.0,H=0,I=0.0,J=0.0,K=0,L=0.0,M=0.0,N=0;switch(k[a+32>>2]|0){case 1:{d=k[a+36>>2]|0;e=+p[a+8>>3];f=+p[a>>3];g=k[a+56>>2]|0;p[g>>3]=1.0;if(((d|0)!=1?(h=(e+b)*f,p[g+8>>3]=h,(d|0)>2):0)?(f=h*2.0,e=f*h+-1.0,p[g+16>>3]=e,(d|0)!=3):0){h=e;i=3;j=2;while(1){h=f*h-+p[g+(j+-1<<3)>>3];p[g+(i<<3)>>3]=h;l=i+1|0;if((l|0)==(d|0))break;else{m=i;i=l;j=m}}}j=k[a+40>>2]|0;h=+p[a+24>>3];f=+p[a+16>>3];i=k[a+60>>2]|0;p[i>>3]=1.0;if((j|0)!=1){e=(h+c)*f;p[i+8>>3]=e;if((j|0)>2){f=e*2.0;h=f*e+-1.0;p[i+16>>3]=h;if((j|0)==3){n=g;o=i;q=3;r=d;s=25}else{e=h;m=3;l=2;while(1){e=f*e-+p[i+(l+-1<<3)>>3];p[i+(m<<3)>>3]=e;t=m+1|0;if((t|0)==(j|0)){u=d;v=j;w=g;x=i;s=26;break}else{y=m;m=t;l=y}}}}else{u=d;v=j;w=g;x=i;s=26}}else{n=g;o=i;q=1;r=d;s=25}break}case 2:{d=k[a+36>>2]|0;e=+p[a+8>>3];f=+p[a>>3];i=k[a+56>>2]|0;p[i>>3]=1.0;if(((d|0)!=1?(h=(e+b)*f,p[i+8>>3]=h,(d|0)>2):0)?(f=(h*(h*3.0)+-1.0)*.5,p[i+16>>3]=f,(d|0)!=3):0){g=3;e=f;j=2;while(1){f=+(g|0);e=(e*(h*(f*2.0+-1.0))-+p[i+(j+-1<<3)>>3]*(f+-1.0))/f;p[i+(g<<3)>>3]=e;l=g+1|0;if((l|0)==(d|0))break;else{m=g;g=l;j=m}}}j=k[a+40>>2]|0;e=+p[a+24>>3];h=+p[a+16>>3];g=k[a+60>>2]|0;p[g>>3]=1.0;if((j|0)!=1){f=(e+c)*h;p[g+8>>3]=f;if((j|0)>2){h=(f*(f*3.0)+-1.0)*.5;p[g+16>>3]=h;if((j|0)==3){n=i;o=g;q=3;r=d;s=25}else{m=3;e=h;l=2;while(1){h=+(m|0);e=(e*(f*(h*2.0+-1.0))-+p[g+(l+-1<<3)>>3]*(h+-1.0))/h;p[g+(m<<3)>>3]=e;y=m+1|0;if((y|0)==(j|0)){u=d;v=j;w=i;x=g;s=26;break}else{t=m;m=y;l=t}}}}else{u=d;v=j;w=i;x=g;s=26}}else{n=i;o=g;q=1;r=d;s=25}break}case 3:{d=k[a+36>>2]|0;g=k[a+56>>2]|0;p[g>>3]=1.0;if((d|0)!=1?(p[g+8>>3]=b,(d|0)>2):0){e=b;i=2;do{e=e*b;p[g+(i<<3)>>3]=e;i=i+1|0}while((i|0)!=(d|0))}i=k[a+40>>2]|0;j=k[a+60>>2]|0;p[j>>3]=1.0;if((i|0)!=1){p[j+8>>3]=c;if((i|0)>2){e=c;l=2;do{e=e*c;p[j+(l<<3)>>3]=e;l=l+1|0}while((l|0)!=(i|0));u=d;v=i;w=g;x=j;s=26}else{u=d;v=i;w=g;x=j;s=26}}else{n=g;o=j;q=1;r=d;s=25}break}default:{Pw(605056,33,1,k[31229]|0)|0;z=0.0;return +z}}if((s|0)==25){A=(r|0)>(q|0)?r:q;B=r;C=o;D=q;E=n}else if((s|0)==26)if((v|0)>0){A=(u|0)>(v|0)?u:v;B=u;C=x;D=v;E=w}else{z=0.0;return +z}w=a+52|0;switch(k[a+44>>2]|0){case 0:{a=0;v=0;e=0.0;x=B;while(1){if((x|0)>0){u=k[w>>2]|0;c=0.0;s=v;n=0;while(1){b=c+ +p[u+(s<<3)>>3]*+p[E+(n<<3)>>3];n=n+1|0;if((n|0)==(x|0)){F=b;break}else{c=b;s=s+1|0}}G=F;H=x+v|0}else{G=0.0;H=v}c=e+G*+p[C+(a<<3)>>3];a=a+1|0;if((a|0)>=(D|0)){z=c;break}else{v=H;e=c;x=1}}return +z}case 2:{x=0;H=0;e=0.0;v=B;while(1){if((v|0)>0){a=k[w>>2]|0;G=0.0;s=H;n=0;while(1){F=G+ +p[a+(s<<3)>>3]*+p[E+(n<<3)>>3];n=n+1|0;if((n|0)==(v|0)){I=F;break}else{G=F;s=s+1|0}}J=I;K=v+H|0}else{J=0.0;K=H}G=e+J*+p[C+(x<<3)>>3];v=(((x+1+B|0)>(A|0))<<31>>31)+v|0;x=x+1|0;if((x|0)>=(D|0)){z=G;break}else{H=K;e=G}}return +z}default:{K=(B|0)>0;H=0;x=0;e=0.0;while(1){if(K){v=k[w>>2]|0;J=0.0;A=x;s=0;while(1){I=J+ +p[v+(A<<3)>>3]*+p[E+(s<<3)>>3];s=s+1|0;if((s|0)==(B|0)){L=I;break}else{J=I;A=A+1|0}}M=L;N=B+x|0}else{M=0.0;N=x}J=e+M*+p[C+(H<<3)>>3];H=H+1|0;if((H|0)>=(D|0)){z=J;break}else{x=N;e=J}}return +z}}return +(0.0)}
function dq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0;d=r;r=r+160|0;e=d+52|0;f=d;g=xx(b)|0;h=xx(c)|0;do if((g|0)==7){if((Ax(b,782848,7)|0)!=0?(Ax(b,782840,7)|0)!=0:0){j=0;break}else k=0;while(1){l=a+k|0;if(!(i[l>>0]|0)){m=l;n=k;o=7;break}l=k+1|0;if((l|0)<57600)k=l;else{p=l;o=6;break}}if((o|0)==6){q=a+p|0;o=8}else if((o|0)==7)if((n|0)>0){q=m;o=8}else s=0;a:do if((o|0)==8){l=q;t=a;u=a;while(1){v=fq(u,768779,l-u|0)|0;if(!v){s=0;break a}w=(v-t|0)%80|0;x=i[v+3>>0]|0;b:do if((w|0)<=7){c:do if(x<<24>>24>32){switch(x<<24>>24){case 61:case 127:{break c;break}default:{}}y=v+1|0;break b}while(0);z=v+(0-w)|0;if((w|0)>0){A=v+1|0;B=u;C=z;while(1){D=(i[C>>0]|0)==32?B:A;C=C+1|0;if(C>>>0>=v>>>0){E=D;break}else B=D}}else E=u;if(v>>>0<E>>>0)y=E;else{s=z;break a}}else y=v+1|0;while(0);if(y>>>0<q>>>0)u=y;else{s=0;break}}}while(0);u=s+80|0;Bx(u,s,80)|0;Bx(s,b,7)|0;t=s+8|0;sy(s+7|0,32,(t>>>0>u>>>0?t:u)+(-7-s)|0)|0;Bx(s+9|0,c,h)|0;r=d;return}else j=0;while(0);while(1){s=a+j|0;if(!(i[s>>0]|0)){F=s;G=j;o=24;break}s=j+1|0;if((s|0)<57600)j=s;else{H=s;o=23;break}}if((o|0)==23){I=a+H|0;o=25}else if((o|0)==24)if((G|0)>0){I=F;o=25}else{J=0;o=38}d:do if((o|0)==25){F=I;G=a;H=a;e:while(1){j=fq(H,b,F-H|0)|0;if(!j){J=0;o=38;break d}s=(j-G|0)%80|0;y=j+(xx(b)|0)|0;q=i[y>>0]|0;f:do if((s|0)<=7){g:do if(q<<24>>24>32){switch(q<<24>>24){case 61:case 127:{break g;break}default:{}}K=j+1|0;break f}while(0);y=j+(0-s)|0;if((s|0)>0){E=j+1|0;m=H;n=y;while(1){p=(i[n>>0]|0)==32?m:E;n=n+1|0;if(n>>>0>=j>>>0){L=p;break}else m=p}}else L=H;if(j>>>0<L>>>0)K=L;else{M=j;N=s;O=y;break e}}else K=j+1|0;while(0);if(K>>>0<I>>>0)H=K;else{J=0;o=38;break d}}if(!O){J=0;o=38}else{Bx(e,O,80)|0;i[e+80>>0]=0;H=M+(80-N)|0;G=rx(e,39)|0;if(!G)P=e;else P=rx(G+1|0,39)|0;G=rx(P,47)|0;if(!G){i[f>>0]=0;Q=0;R=O;S=H;break}F=e-G|0;Bx(f,G+1|0,F+80|0)|0;h:do if((F+79|0)>0){G=f+(F+78)|0;while(1){if((i[G>>0]|0)!=32)break h;i[G>>0]=0;if(G>>>0>f>>>0)G=G+-1|0;else break}}while(0);Q=xx(f)|0;R=O;S=H}}while(0);if((o|0)==38){while(1){o=0;if(!(i[a+J>>0]|0)){T=J;o=40;break}O=J+1|0;if((O|0)<57600){J=O;o=38}else{U=O;o=41;break}}if((o|0)==40)if((T|0)>0){U=T;o=41}else{V=0;o=56}i:do if((o|0)==41){T=a+U|0;J=T;O=a;e=a;j:while(1){P=fq(e,768779,J-e|0)|0;if(!P){V=0;o=56;break i}N=(P-O|0)%80|0;M=i[P+3>>0]|0;k:do if((N|0)<=7){l:do if(M<<24>>24>32){switch(M<<24>>24){case 61:case 127:{break l;break}default:{}}W=P+1|0;break k}while(0);y=P+(0-N)|0;if((N|0)>0){K=P+1|0;I=e;L=y;while(1){F=(i[L>>0]|0)==32?I:K;L=L+1|0;if(L>>>0>=P>>>0){X=F;break}else I=F}}else X=e;if(P>>>0<X>>>0)W=X;else{Y=y;break j}}else W=P+1|0;while(0);if(W>>>0<T>>>0)e=W;else{V=0;o=56;break i}}if(!((Y|0)==0|(Y|0)==(a|0))){e=Y;while(1){T=e+-80|0;if(!(Ax(T,773880,8)|0))e=T;else{Z=e;break}}if(!(Z>>>0>=Y>>>0|(Z|0)==0)){_=Z;$=Z+80|0}else{V=0;o=56}}else{V=0;o=56}}while(0);if((o|0)==56){while(1){o=0;Z=a+V|0;if(!(i[Z>>0]|0)){aa=Z;ba=V;o=59;break}Z=V+1|0;if((Z|0)<57600){V=Z;o=56}else{ca=Z;o=58;break}}if((o|0)==58){da=a+ca|0;o=60}else if((o|0)==59)if((ba|0)>0){da=aa;o=60}else ea=0;m:do if((o|0)==60){aa=da;ba=a;ca=a;while(1){V=fq(ca,768779,aa-ca|0)|0;if(!V){ea=0;break m}Z=(V-ba|0)%80|0;Y=i[V+3>>0]|0;n:do if((Z|0)<=7){o:do if(Y<<24>>24>32){switch(Y<<24>>24){case 61:case 127:{break o;break}default:{}}fa=V+1|0;break n}while(0);y=V+(0-Z)|0;if((Z|0)>0){W=V+1|0;X=ca;U=y;while(1){e=(i[U>>0]|0)==32?X:W;U=U+1|0;if(U>>>0>=V>>>0){ga=e;break}else X=e}}else ga=ca;if(V>>>0<ga>>>0)fa=ga;else{ea=y;break m}}else fa=V+1|0;while(0);if(fa>>>0<da>>>0)ca=fa;else{ea=0;break}}}while(0);fa=ea+80|0;Bx(fa,ea,80)|0;_=ea;$=fa}i[f>>0]=0;Q=0;R=_;S=$}$=R;if(R>>>0<S>>>0)sy(R|0,32,S+(0-$)|0)|0;Bx(R,b,g)|0;i[R+8>>0]=61;i[R+9>>0]=32;if((i[c>>0]|0)==39){Bx(R+10|0,c,h)|0;g=h+12|0;ha=(g|0)>31?g:30}else{Bx(R+(30-h)|0,c,h)|0;ha=30}if((Q|0)<=0){r=d;return}h=(Q+2+ha|0)>80?78-ha|0:Q;i[R+(ha+2)>>0]=47;c=ha+3|0;Bx(R+c|0,f,h)|0;if((R+(h+c)|0)>>>0>=S>>>0){r=d;return}c=-3-Q-ha|0;ha=(c|0)>-81?c:-81;sy(R+(0-ha)|0,32,S+(ha-$)|0)|0;r=d;return}function eq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0;d=r;r=r+112|0;e=d;do if((xx(b)|0)==7){if((Ax(b,782848,7)|0)!=0?(Ax(b,782840,7)|0)!=0:0){f=0;g=21;break}else h=0;while(1){j=a+h|0;if(!(i[j>>0]|0)){k=j;l=h;g=7;break}j=h+1|0;if((j|0)<57600)h=j;else{m=j;g=6;break}}if((g|0)==6){n=a+m|0;g=8}else if((g|0)==7)if((l|0)>0){n=k;g=8}else o=0;a:do if((g|0)==8){j=n;p=a;q=a;while(1){s=fq(q,768779,j-q|0)|0;if(!s){o=0;break a}t=(s-p|0)%80|0;u=i[s+3>>0]|0;b:do if((t|0)<=7){c:do if(u<<24>>24>32){switch(u<<24>>24){case 61:case 127:{break c;break}default:{}}v=s+1|0;break b}while(0);w=s+(0-t)|0;if((t|0)>0){x=s+1|0;y=q;z=w;while(1){A=(i[z>>0]|0)==32?y:x;z=z+1|0;if(z>>>0>=s>>>0){B=A;break}else y=A}}else B=q;if(s>>>0<B>>>0)v=B;else{o=w;break a}}else v=s+1|0;while(0);if(v>>>0<n>>>0)q=v;else{o=0;break}}}while(0);q=o+80|0;Bx(q,o,80)|0;p=o;j=p+80|0;do{i[p>>0]=32;p=p+1|0}while((p|0)<(j|0));Bx(o,b,7)|0;C=0;D=q}else{f=0;g=21}while(0);do if((g|0)==21){while(1){g=0;o=a+f|0;if(!(i[o>>0]|0)){E=o;F=f;g=24;break}o=f+1|0;if((o|0)<57600){f=o;g=21}else{G=o;g=23;break}}if((g|0)==23)H=a+G|0;else if((g|0)==24)if((F|0)>0)H=E;else{r=d;return}q=H;o=a;v=a;d:while(1){n=fq(v,b,q-v|0)|0;if(!n){g=44;break}B=(n-o|0)%80|0;k=n+(xx(b)|0)|0;l=i[k>>0]|0;e:do if((B|0)<=7){f:do if(l<<24>>24>32){switch(l<<24>>24){case 61:case 127:{break f;break}default:{}}I=n+1|0;break e}while(0);k=n+(0-B)|0;if((B|0)>0){m=n+1|0;h=v;p=k;while(1){j=(i[p>>0]|0)==32?h:m;p=p+1|0;if(p>>>0>=n>>>0){J=j;break}else h=j}}else J=v;if(n>>>0<J>>>0)I=J;else{K=n;L=B;M=k;g=37;break d}}else I=n+1|0;while(0);if(I>>>0<H>>>0)v=I;else{g=44;break}}if((g|0)==37){v=K+(80-L)|0;if(!M){r=d;return}Bx(e,M,80)|0;o=rx(e,39)|0;if((o|0)!=0?(q=rx(o+1|0,39)|0,(q|0)!=0):0){o=q-e|0;N=(o|0)<31?31:o+2|0}else N=31;o=K+(N-L)|0;i[o>>0]=47;i[o+1>>0]=32;C=o;D=v;break}else if((g|0)==44){r=d;return}}while(0);g=xx(c)|0;if((g|0)<=0){r=d;return}L=C+2|0;Bx(L,c,(C+(g+2)|0)>>>0>D>>>0?D-L|0:g)|0;r=d;return}function fq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0;if((a|0)==0|(b|0)==0){d=0;return d|0}e=xx(b)|0;if(!e){d=a;return d|0}if(!c){d=0;return d|0}f=i[b>>0]|0;g=e+-1|0;h=i[b+g>>0]|0;j=c+1-e|0;c=a+j|0;if((j|0)<=0){d=0;return d|0}switch(e|0){case 1:{j=a;while(1){if((i[j>>0]|0)==f<<24>>24){d=j;k=20;break}j=j+1|0;if(j>>>0>=c>>>0){d=0;k=20;break}}if((k|0)==20)return d|0;break}case 2:{j=a;while(1){if((i[j>>0]|0)==f<<24>>24?(i[j+g>>0]|0)==h<<24>>24:0){d=j;k=20;break}j=j+1|0;if(j>>>0>=c>>>0){d=0;k=20;break}}if((k|0)==20)return d|0;break}default:{if((e|0)<=1){j=a;while(1){if((i[j>>0]|0)==f<<24>>24?(i[j+g>>0]|0)==h<<24>>24:0){d=j;k=20;break}j=j+1|0;if(j>>>0>=c>>>0){d=0;k=20;break}}if((k|0)==20)return d|0}else l=a;a:while(1){b:do if((i[l>>0]|0)==f<<24>>24?(i[l+g>>0]|0)==h<<24>>24:0){a=1;while(1){if((i[l+a>>0]|0)!=(i[b+a>>0]|0))break b;a=a+1|0;if((a|0)>=(e|0)){d=l;k=20;break a}}}while(0);l=l+1|0;if(l>>>0>=c>>>0){d=0;k=20;break}}if((k|0)==20)return d|0}}return 0}function gq(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;c=r;r=r+208|0;d=c+196|0;e=c+112|0;f=c+8|0;g=c+6|0;h=c+4|0;j=c+2|0;k=c;l=c+114|0;m=c+12|0;i[e>>0]=39;i[e+1>>0]=0;i[f>>0]=34;i[f+1>>0]=0;i[g>>0]=91;i[g+1>>0]=0;i[k>>0]=44;i[k+1>>0]=0;i[h>>0]=93;i[h+1>>0]=0;i[j>>0]=47;i[j+1>>0]=0;Bx(l,b,80)|0;b=xx(l)|0;n=fq(l,g,b)|0;if(!n){g=fq(l,k,b)|0;if(!g)o=0;else{p=g;q=3}}else{p=n;q=3}if((q|0)==3){i[p>>0]=0;o=p+1|0}p=0;while(1){n=a+p|0;if(!(i[n>>0]|0)){s=n;t=p;q=8;break}n=p+1|0;if((n|0)<57600)p=n;else{u=n;q=7;break}}if((q|0)==7)v=a+u|0;else if((q|0)==8)if((t|0)>0)v=s;else{w=0;r=c;return w|0}s=v;t=a;u=a;a:while(1){a=fq(u,l,s-u|0)|0;if(!a){w=0;q=45;break}p=(a-t|0)%80|0;n=a+(xx(l)|0)|0;g=i[n>>0]|0;b:do if((p|0)<=7){c:do if(g<<24>>24>32){switch(g<<24>>24){case 61:case 127:{break c;break}default:{}}x=a+1|0;break b}while(0);n=a+(0-p)|0;if((p|0)>0){b=a+1|0;k=u;y=n;while(1){z=(i[y>>0]|0)==32?k:b;y=y+1|0;if(y>>>0>=a>>>0){A=z;break}else k=z}}else A=u;if(a>>>0<A>>>0)x=A;else{B=n;q=21;break a}}else x=a+1|0;while(0);if(x>>>0<v>>>0)u=x;else{w=0;q=45;break}}if((q|0)==21){if(!B){w=0;r=c;return w|0}x=m;u=x+100|0;do{i[x>>0]=0;x=x+1|0}while((x|0)<(u|0));Bx(m,B,80)|0;B=xx(m)|0;x=fq(m,e,B)|0;u=fq(m,j,B)|0;do if(!x){j=fq(m,f,B)|0;if(j){if(j>>>0<u>>>0){v=j+1|0;C=j;D=fq(v,f,xx(v)|0)|0;break}if(!u){v=j+1|0;C=j;D=fq(v,f,xx(v)|0)|0}else q=32}else q=32}else{if(x>>>0<u>>>0){v=x+1|0;C=x;D=fq(v,e,xx(v)|0)|0;break}if(!u){v=x+1|0;C=x;D=fq(v,e,xx(v)|0)|0}else q=32}while(0);if((q|0)==32){e=fq(m,773860,B)|0;x=fq(m,769609,B)|0;C=e;D=(x|0)==0?m+79|0:x}x=C;while(1){C=x+1|0;if(C>>>0<D>>>0&(i[C>>0]|0)==32)x=C;else{E=C;F=x;break}}i[D>>0]=0;x=D+-1|0;if((i[x>>0]|0)==32&x>>>0>E>>>0){D=x;do{i[D>>0]=0;D=D+-1|0}while((i[D>>0]|0)==32&D>>>0>E>>>0)}D=(tx(E,769611)|0)==0;x=D?F+2|0:E;ux(769614,x)|0;if(!o){w=769614;r=c;return w|0}E=fq(o,h,xx(o)|0)|0;if(E)i[E>>0]=0;E=ex(o)|0;if((E|0)<=0){w=769614;r=c;return w|0}i[d>>0]=32;i[d+1>>0]=0;o=1;h=x;while(1){x=Fx(h,d)|0;if((o|0)==(E|0)){G=x;break}else{o=o+1|0;h=0}}if(!G){w=0;r=c;return w|0}ux(769614,G)|0;w=769614;r=c;return w|0}else if((q|0)==45){r=c;return w|0}return 0}function hq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,n=0,q=0,r=0.0,s=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}do switch(b|0){case 16:{if((mq(a,c,d,e,f)|0)==202){k[f>>2]=g;fs(a,c,d,e,f)|0}break}case 11:{n=l[d>>0]|0;if((oq(a,c,n,0,e,f)|0)==202){k[f>>2]=g;gs(a,c,n,0,e,f)|0}break}case 12:{n=i[d>>0]|0;q=((n|0)<0)<<31>>31;if((oq(a,c,n,q,e,f)|0)==202){k[f>>2]=g;gs(a,c,n,q,e,f)|0}break}case 20:{q=m[d>>1]|0;if((oq(a,c,q,0,e,f)|0)==202){k[f>>2]=g;gs(a,c,q,0,e,f)|0}break}case 21:{q=j[d>>1]|0;n=((q|0)<0)<<31>>31;if((oq(a,c,q,n,e,f)|0)==202){k[f>>2]=g;gs(a,c,q,n,e,f)|0}break}case 31:{n=k[d>>2]|0;q=((n|0)<0)<<31>>31;if((oq(a,c,n,q,e,f)|0)==202){k[f>>2]=g;gs(a,c,n,q,e,f)|0}break}case 30:{r=+((k[d>>2]|0)>>>0);if((qq(a,c,r,0,e,f)|0)==202){k[f>>2]=g;hs(a,c,r,0,e,f)|0}break}case 14:{q=k[d>>2]|0;if((nq(a,c,q,e,f)|0)==202){k[f>>2]=g;is(a,c,q,e,f)|0}break}case 40:{r=+((k[d>>2]|0)>>>0);if((qq(a,c,r,0,e,f)|0)==202){k[f>>2]=g;hs(a,c,r,0,e,f)|0}break}case 41:{q=k[d>>2]|0;n=((q|0)<0)<<31>>31;if((oq(a,c,q,n,e,f)|0)==202){k[f>>2]=g;gs(a,c,q,n,e,f)|0}break}case 81:{n=d;q=k[n>>2]|0;s=k[n+4>>2]|0;if((oq(a,c,q,s,e,f)|0)==202){k[f>>2]=g;gs(a,c,q,s,e,f)|0}break}case 42:{r=+o[d>>2];if((pq(a,c,r,-7,e,f)|0)==202){k[f>>2]=g;js(a,c,r,-7,e,f)|0}break}case 82:{r=+p[d>>3];if((rq(a,c,r,-15,e,f)|0)==202){k[f>>2]=g;ks(a,c,r,-15,e,f)|0}break}case 83:{if((sq(a,c,d,-7,e,f)|0)==202){k[f>>2]=g;ls(a,c,d,-7,e,f)|0}break}case 163:{if((tq(a,c,d,-15,e,f)|0)==202){k[f>>2]=g;ms(a,c,d,-15,e,f)|0}break}default:k[f>>2]=410}while(0);h=k[f>>2]|0;return h|0}function iq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}if((mq(a,b,c,d,e)|0)==202){k[e>>2]=f;fs(a,b,c,d,e)|0}g=k[e>>2]|0;return g|0}function jq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}if((oq(a,b,c,d,e,f)|0)==202){k[f>>2]=g;gs(a,b,c,d,e,f)|0}h=k[f>>2]|0;return h|0}function kq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}if((nq(a,b,c,d,e)|0)==202){k[e>>2]=f;is(a,b,c,d,e)|0}g=k[e>>2]|0;return g|0}function lq(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}if((rq(a,b,c,d,e,f)|0)==202){k[f>>2]=g;ks(a,b,c,d,e,f)|0}h=k[f>>2]|0;return h|0}function mq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;f=r;r=r+384|0;g=f+298|0;h=f+227|0;j=f+154|0;l=f+73|0;m=f;n=k[e>>2]|0;if((n|0)>0){o=n;r=f;return o|0}if((Qo(a,b,g,j,e)|0)>0){o=k[e>>2]|0;r=f;return o|0}os(c,h,e)|0;if((d|0)!=0?(i[d>>0]|0)!=38:0)Dk(b,h,d,l,e)|0;else Dk(b,h,j,l,e)|0;Ek(a,l,e)|0;l=k[a+4>>2]|0;j=l+112|0;b=(k[l+88>>2]|0)+(k[l+68>>2]<<3)|0;l=jy(k[j>>2]|0,k[j+4>>2]|0,k[b>>2]|0,k[b+4>>2]|0)|0;b=wy(l|0,L|0,80,0)|0;l=ky(b|0,L|0,1,0)|0;b=k[e>>2]|0;if((b|0)>0){o=b;r=f;return o|0}vk();fl(g,h,e)|0;if((k[e>>2]|0)==204){yk();k[e>>2]=0;o=0;r=f;return o|0}g=xx(h)|0;a:do if(g){b=g;do{if((i[h+(b+-1)>>0]|0)!=38)break a;So(a,h,m,e)|0;if(!(i[h>>0]|0))break a;xq(a,l,e)|0;b=xx(h)|0}while((b|0)!=0)}while(0);o=k[e>>2]|0;r=f;return o|0}function nq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0;f=r;r=r+240|0;g=f+154|0;h=f+81|0;j=f;l=k[e>>2]|0;if((l|0)>0){m=l;r=f;return m|0}if((Qo(a,b,g,h,e)|0)>0){m=k[e>>2]|0;r=f;return m|0}ss(c,g,e)|0;if((d|0)!=0?(i[d>>0]|0)!=38:0)Dk(b,g,d,j,e)|0;else Dk(b,g,h,j,e)|0;Ek(a,j,e)|0;m=k[e>>2]|0;r=f;return m|0}function oq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+240|0;h=g+154|0;j=g+81|0;l=g;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}if((Qo(a,b,h,j,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}ts(c,d,h,f)|0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,l,f)|0;else Dk(b,h,j,l,f)|0;Ek(a,l,f)|0;n=k[f>>2]|0;r=g;return n|0}function pq(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+240|0;h=g+154|0;j=g+81|0;l=g;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}if((Qo(a,b,h,j,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}us(c,d,h,f)|0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,l,f)|0;else Dk(b,h,j,l,f)|0;Ek(a,l,f)|0;n=k[f>>2]|0;r=g;return n|0}function qq(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+240|0;h=g+154|0;j=g+81|0;l=g;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}if((Qo(a,b,h,j,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}vs(c,d,h,f)|0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,l,f)|0;else Dk(b,h,j,l,f)|0;Ek(a,l,f)|0;n=k[f>>2]|0;r=g;return n|0}function rq(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+240|0;h=g+154|0;j=g+81|0;l=g;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}if((Qo(a,b,h,j,f)|0)>0){n=k[f>>2]|0;r=g;return n|0}ws(c,d,h,f)|0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,l,f)|0;else Dk(b,h,j,l,f)|0;Ek(a,l,f)|0;n=k[f>>2]|0;r=g;return n|0}function sq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,m=0,n=0,p=0,q=0;g=r;r=r+304|0;h=g;l=g+225|0;m=g+152|0;n=g+71|0;p=k[f>>2]|0;if((p|0)>0){q=p;r=g;return q|0}if((Qo(a,b,h,m,f)|0)>0){q=k[f>>2]|0;r=g;return q|0}j[h>>1]=40;us(+o[c>>2],d,l,f)|0;qx(h,l)|0;p=h+(xx(h)|0)|0;i[p>>0]=i[773939]|0;i[p+1>>0]=i[773940]|0;i[p+2>>0]=i[773941]|0;us(+o[c+4>>2],d,l,f)|0;qx(h,l)|0;l=h+(xx(h)|0)|0;i[l>>0]=41;i[l+1>>0]=0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,n,f)|0;else Dk(b,h,m,n,f)|0;Ek(a,n,f)|0;q=k[f>>2]|0;r=g;return q|0}function tq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,m=0,n=0,o=0,q=0;g=r;r=r+304|0;h=g;l=g+225|0;m=g+152|0;n=g+71|0;o=k[f>>2]|0;if((o|0)>0){q=o;r=g;return q|0}if((Qo(a,b,h,m,f)|0)>0){q=k[f>>2]|0;r=g;return q|0}j[h>>1]=40;ws(+p[c>>3],d,l,f)|0;qx(h,l)|0;o=h+(xx(h)|0)|0;i[o>>0]=i[773939]|0;i[o+1>>0]=i[773940]|0;i[o+2>>0]=i[773941]|0;ws(+p[c+8>>3],d,l,f)|0;qx(h,l)|0;l=h+(xx(h)|0)|0;i[l>>0]=41;i[l+1>>0]=0;if((e|0)!=0?(i[e>>0]|0)!=38:0)Dk(b,h,e,n,f)|0;else Dk(b,h,m,n,f)|0;Ek(a,n,f)|0;q=k[f>>2]|0;r=g;return q|0}function uq(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}if((vq(a,b,c,d)|0)==202){k[d>>2]=e;ds(a,c,d)|0}f=k[d>>2]|0;return f|0}function vq(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0;e=r;r=r+400|0;f=e+316|0;g=e+235|0;h=e+73|0;j=e;l=k[d>>2]|0;if((l|0)>0){m=l;r=e;return m|0}if((Go(a,b,f,d)|0)>0){m=k[d>>2]|0;r=e;return m|0}Ek(a,c,d)|0;c=k[a+4>>2]|0;b=c+112|0;l=(k[c+88>>2]|0)+(k[c+68>>2]<<3)|0;c=jy(k[b>>2]|0,k[b+4>>2]|0,k[l>>2]|0,k[l+4>>2]|0)|0;l=wy(c|0,L|0,80,0)|0;c=l+1|0;Hk(f,g,e+154|0,d)|0;f=k[d>>2]|0;if((f|0)>0){m=f;r=e;return m|0}vk();fl(g,h,d)|0;if((k[d>>2]|0)==204){yk();k[d>>2]=0;m=0;r=e;return m|0}g=xx(h)|0;a:do if(g){f=g;do{if((i[h+(f+-1)>>0]|0)!=38)break a;So(a,h,j,d)|0;if(!(i[h>>0]|0))break a;xq(a,c,d)|0;f=xx(h)|0}while((f|0)!=0)}while(0);m=k[d>>2]|0;r=e;return m|0}function wq(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=k[d>>2]|0;if((e|0)>0){f=e;return f|0}Do(a,b+1|0,d)|0;Ek(a,c,d)|0;f=k[d>>2]|0;return f|0}function xq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+256|0;e=d;f=d+166|0;g=d+85|0;h=d+4|0;j=k[c>>2]|0;if((j|0)>0){l=j;r=d;return l|0}j=k[a>>2]|0;m=a+4|0;if((j|0)!=(k[(k[m>>2]|0)+68>>2]|0))Fk(a,j+1|0,0,c)|0;if((b|0)>=1?(j=((b|0)<0)<<31>>31,n=k[m>>2]|0,o=n+96|0,p=k[o>>2]|0,q=k[o+4>>2]|0,o=(k[n+88>>2]|0)+(k[n+68>>2]<<3)|0,s=k[o>>2]|0,t=k[o+4>>2]|0,o=wy(s|0,t|0,80,0)|0,u=jy(p|0,q|0,o|0,L|0)|0,o=L,!((j|0)>(o|0)|(j|0)==(o|0)&b>>>0>u>>>0)):0){u=(b*80|0)+-80|0;o=ky(s|0,t|0,u|0,((u|0)<0)<<31>>31|0)|0;u=L;t=n+112|0;k[t>>2]=o;k[t+4>>2]=u;t=jy(p|0,q|0,o|0,u|0)|0;u=wy(t|0,L|0,80,0)|0;if((u|0)<1){k[e>>2]=b;Xw(h,769984,e)|0;tk(h);k[c>>2]=203;l=203;r=d;return l|0}h=g;e=770037;b=h+41|0;do{i[h>>0]=i[e>>0]|0;h=h+1|0;e=e+1|0}while((h|0)<(b|0));h=g+(xx(g)|0)|0;e=770037;b=h+41|0;do{i[h>>0]=i[e>>0]|0;h=h+1|0;e=e+1|0}while((h|0)<(b|0));e=p;p=q;q=0;h=f;f=g;while(1){e=ky(e|0,p|0,-80,-1)|0;p=L;au(a,e,p,0,c)|0;fu(a,80,0,h,c)|0;au(a,e,p,0,c)|0;cu(a,80,0,f,c)|0;q=q+1|0;if((q|0)==(u|0))break;else{g=f;f=h;h=g}}h=(k[m>>2]|0)+96|0;m=h;f=ky(k[m>>2]|0,k[m+4>>2]|0,-80,-1)|0;m=h;k[m>>2]=f;k[m+4>>2]=L;l=k[c>>2]|0;r=d;return l|0}k[c>>2]=203;l=203;r=d;return l|0}function yq(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0;e=r;r=r+240|0;f=e+152|0;g=e+81|0;h=e;i=k[d>>2]|0;if((i|0)>0){j=i;r=e;return j|0}if((Qo(a,b,g,f,d)|0)>0){j=k[d>>2]|0;r=e;return j|0}else{Dk(c,g,f,h,d)|0;Ek(a,h,d)|0;j=k[d>>2]|0;r=e;return j|0}return 0}function zq(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=r;r=r+240|0;f=e+81|0;g=e;h=k[d>>2]|0;if((h|0)>0){i=h;r=e;return i|0}if((Qo(a,b,f,e+152|0,d)|0)>0){i=k[d>>2]|0;r=e;return i|0}else{Dk(b,f,c,g,d)|0;Ek(a,g,d)|0;i=k[d>>2]|0;r=e;return i|0}return 0}function Aq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;d=r;r=r+384|0;e=d;f=d+302|0;g=d+158|0;h=d+77|0;j=d+4|0;l=k[c>>2]|0;if((l|0)>0){m=l;r=d;return m|0}if((Qo(a,b,f,d+229|0,c)|0)>0){k[e>>2]=b;Xw(h,770078,e)|0;tk(h);m=k[c>>2]|0;r=d;return m|0}h=k[a+4>>2]|0;e=h+112|0;b=(k[h+88>>2]|0)+(k[h+68>>2]<<3)|0;h=jy(k[e>>2]|0,k[e+4>>2]|0,k[b>>2]|0,k[b+4>>2]|0)|0;b=wy(h|0,L|0,80,0)|0;xq(a,b,c)|0;h=k[c>>2]|0;if((h|0)>0){m=h;r=d;return m|0}vk();fl(f,g,c)|0;if((k[c>>2]|0)==204){yk();k[c>>2]=0;m=0;r=d;return m|0}f=xx(g)|0;a:do if(f){h=f;do{if((i[g+(h+-1)>>0]|0)!=38)break a;So(a,g,j,c)|0;if(!(i[g>>0]|0))break a;xq(a,b,c)|0;h=xx(g)|0}while((h|0)!=0)}while(0);m=k[c>>2]|0;r=d;return m|0}function Bq(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;d=r;r=r+176|0;e=d+81|0;f=d;g=k[c>>2]|0;if((g|0)>0){h=g;r=d;return h|0}g=k[a>>2]|0;j=a+4|0;l=k[j>>2]|0;if((g|0)==(k[l+68>>2]|0))m=l;else{Fk(a,g+1|0,0,c)|0;m=k[j>>2]|0}g=m+120|0;l=m+96|0;n=k[l>>2]|0;o=k[l+4>>2]|0;l=jy(k[g>>2]|0,k[g+4>>2]|0,n|0,o|0)|0;do if((l|0)==80&(L|0)==0){if((rl(a,1,0,c)|0)<=0){g=k[j>>2]|0;p=g+96|0;q=g;s=k[p>>2]|0;t=k[p+4>>2]|0;break}h=k[c>>2]|0;r=d;return h|0}else{q=m;s=n;t=o}while(0);o=q+112|0;q=jy(s|0,t|0,k[o>>2]|0,k[o+4>>2]|0)|0;o=wy(q|0,L|0,80,0)|0;Bx(f,b,80)|0;i[f+80>>0]=0;b=xx(f)|0;if((b|0)>0){q=0;do{t=f+q|0;s=i[t>>0]|0;if(s<<24>>24<32|s<<24>>24==127)i[t>>0]=32;q=q+1|0}while((q|0)!=(b|0));if((b|0)<80)u=10}else u=10;if((u|0)==10)sy(f+b|0,32,80-b|0)|0;b=vx(f,773860)|0;u=(b|0)==80?8:b;if((u|0)>0){b=0;do{q=f+b|0;t=(nv(i[q>>0]|0)|0)&255;i[q>>0]=t;b=b+1|0}while((b|0)!=(u|0))}Ak(f,c)|0;u=(k[j>>2]|0)+112|0;b=k[u>>2]|0;t=k[u+4>>2]|0;au(a,b,t,0,c)|0;if((o|0)>0){u=b;b=t;t=0;q=e;e=f;while(1){fu(a,80,0,q,c)|0;au(a,u,b,0,c)|0;cu(a,80,0,e,c)|0;u=ky(u|0,b|0,80,0)|0;t=t+1|0;if((t|0)==(o|0)){v=q;break}else{s=e;b=L;e=q;q=s}}}else v=f;cu(a,80,0,v,c)|0;v=k[j>>2]|0;j=v+96|0;a=j;f=ky(k[a>>2]|0,k[a+4>>2]|0,80,0)|0;a=j;k[a>>2]=f;k[a+4>>2]=L;a=v+112|0;v=a;f=ky(k[v>>2]|0,k[v+4>>2]|0,80,0)|0;v=a;k[v>>2]=f;k[v+4>>2]=L;h=k[c>>2]|0;r=d;return h|0}function Cq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0;f=r;r=r+160|0;g=f+81|0;h=f;i=k[e>>2]|0;if((i|0)>0){j=i;r=f;return j|0}os(c,g,e)|0;Dk(b,g,d,h,e)|0;Bq(a,h,e)|0;j=k[e>>2]|0;r=f;return j|0}function Dq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;f=r;r=r+320|0;g=f+237|0;h=f+156|0;j=f+75|0;l=f+4|0;m=f;k[m>>2]=-1;n=k[e>>2]|0;if((n|0)>0){o=n;r=f;return o|0}n=xx(c)|0;p=rx(c,39)|0;if(!p)q=0;else{s=p;p=0;while(1){t=p+1|0;s=rx(s+1|0,39)|0;if(!s){q=t;break}else p=t}}Bx(j,b,80)|0;i[j+80>>0]=0;p=j;while(1)if((i[p>>0]|0)==32)p=p+1|0;else{u=p;break}p=xx(u)|0;if((p|0)<9?(Ak(u,m)|0)<1:0)v=68-q|0;else v=75-q-p|0;if((n|0)>0){p=68-q|0;q=h+8|0;m=0;u=v;v=0;j=n;while(1){Bx(l,c+v|0,u)|0;i[l+u>>0]=0;os(l,g,e)|0;do if((j|0)>(u|0)){n=xx(g)|0;s=u+-1|0;t=g+(n+-2)|0;if((i[t>>0]|0)==39){i[g+(n+-3)>>0]=38;i[g+(n+-1)>>0]=0;w=s;break}else{i[t>>0]=38;w=s;break}}else w=u;while(0);if(!m)Dk(b,g,d,h,e)|0;else{Dk(773889,g,d,h,e)|0;i[q>>0]=32;i[q+1>>0]=32}Bq(a,h,e)|0;s=j-w|0;if((s|0)>0){m=1;u=p;v=w+v|0;j=s}else break}}o=k[e>>2]|0;r=f;return o|0}function Eq(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0;g=r;r=r+160|0;h=g+81|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}ts(c,d,h,f)|0;Dk(b,h,e,i,f)|0;Bq(a,i,f)|0;l=k[f>>2]|0;r=g;return l|0}function Fq(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;h=r;r=r+80|0;i=h+72|0;j=h;l=k[g>>2]|0;if((l|0)>0){m=l;r=h;return m|0}Fl(a,i,g)|0;Hl(a,9,j,g)|0;l=k[i>>2]|0;if((l|0)>0){i=1;n=0;o=0;p=0;q=0;while(1){s=(k[c+(q<<2)>>2]|0)+-1|0;t=yy(s|0,((s|0)<0)<<31>>31|0,i|0,n|0)|0;s=ky(t|0,L|0,o|0,p|0)|0;t=L;u=j+(q<<3)|0;i=yy(k[u>>2]|0,k[u+4>>2]|0,i|0,n|0)|0;q=q+1|0;if((q|0)>=(l|0)){v=s;w=t;break}else{n=L;o=s;p=t}}p=ky(v|0,w|0,1,0)|0;x=p;y=L}else{x=1;y=0}do switch(b|0){case 11:{Qq(a,1,x,y,d,e,f,g)|0;break}case 12:{Fr(a,1,x,y,d,e,f,g)|0;break}case 20:{Mr(a,1,x,y,d,e,f,g)|0;break}case 21:{dr(a,1,x,y,d,e,f,g)|0;break}case 30:{wr(a,1,x,y,d,e,f,g)|0;break}case 31:{sr(a,1,x,y,d,e,f,g)|0;break}case 40:{Sr(a,1,x,y,d,e,f,g)|0;break}case 41:{ir(a,1,x,y,d,e,f,g)|0;break}case 81:{nr(a,1,x,y,d,e,f,g)|0;break}case 42:{_q(a,1,x,y,d,e,f,g)|0;break}case 82:{Vq(a,1,x,y,d,e,f,g)|0;break}default:k[g>>2]=410}while(0);m=k[g>>2]|0;r=h;return m|0}function Gq(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0;i=k[h>>2]|0;if((i|0)>0){j=i;return j|0}do switch(b|0){case 11:{Qq(a,1,c,d,e,f,g,h)|0;break}case 12:{Fr(a,1,c,d,e,f,g,h)|0;break}case 20:{Mr(a,1,c,d,e,f,g,h)|0;break}case 21:{dr(a,1,c,d,e,f,g,h)|0;break}case 30:{wr(a,1,c,d,e,f,g,h)|0;break}case 31:{sr(a,1,c,d,e,f,g,h)|0;break}case 40:{Sr(a,1,c,d,e,f,g,h)|0;break}case 41:{ir(a,1,c,d,e,f,g,h)|0;break}case 81:{nr(a,1,c,d,e,f,g,h)|0;break}case 42:{_q(a,1,c,d,e,f,g,h)|0;break}case 82:{Vq(a,1,c,d,e,f,g,h)|0;break}default:k[h>>2]=410}while(0);j=k[h>>2]|0;return j|0}function Hq(a,b,c,d,e,f,g,h,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;var m=0,n=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}if(!h){Gq(a,b,c,d,e,f,g,l)|0;n=k[l>>2]|0;return n|0}do switch(b|0){case 11:{Sq(a,1,c,d,e,f,g,i[h>>0]|0,l)|0;break}case 12:{Hr(a,1,c,d,e,f,g,i[h>>0]|0,l)|0;break}case 20:{Or(a,1,c,d,e,f,g,j[h>>1]|0,l)|0;break}case 21:{fr(a,1,c,d,e,f,g,j[h>>1]|0,l)|0;break}case 30:{yr(a,1,c,d,e,f,g,k[h>>2]|0,l)|0;break}case 31:{ur(a,1,c,d,e,f,g,k[h>>2]|0,l)|0;break}case 40:{Ur(a,1,c,d,e,f,g,k[h>>2]|0,l)|0;break}case 41:{kr(a,1,c,d,e,f,g,k[h>>2]|0,l)|0;break}case 81:{m=h;pr(a,1,c,d,e,f,g,k[m>>2]|0,k[m+4>>2]|0,l)|0;break}case 42:{ar(a,1,c,d,e,f,g,+o[h>>2],l)|0;break}case 82:{Xq(a,1,c,d,e,f,g,+p[h>>3],l)|0;break}default:k[l>>2]=410}while(0);n=k[l>>2]|0;return n|0}function Iq(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}do switch(b|0){case 1:{Cr(a,c,d,e,f,h,j,l)|0;break}case 11:{Rq(a,c,d,e,f,g,h,i,j,l)|0;break}case 12:{Gr(a,c,d,e,f,g,h,i,j,l)|0;break}case 20:{Nr(a,c,d,e,f,g,h,i,j,l)|0;break}case 21:{er(a,c,d,e,f,g,h,i,j,l)|0;break}case 30:{xr(a,c,d,e,f,g,h,i,j,l)|0;break}case 31:{tr(a,c,d,e,f,g,h,i,j,l)|0;break}case 40:{Tr(a,c,d,e,f,g,h,i,j,l)|0;break}case 41:{jr(a,c,d,e,f,g,h,i,j,l)|0;break}case 81:{or(a,c,d,e,f,g,h,i,j,l)|0;break}case 42:{$q(a,c,d,e,f,g,h,i,j,l)|0;break}case 82:{Wq(a,c,d,e,f,g,h,i,j,l)|0;break}case 83:{m=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;$q(a,c,d,e,m,L,ty(h|0,i|0,1)|0,L,j,l)|0;break}case 163:{m=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;Wq(a,c,d,e,m,L,ty(h|0,i|0,1)|0,L,j,l)|0;break}case 14:{Ar(a,c,d,e,f,g,h,i,j,l)|0;break}case 16:{Dr(a,c,d,e,f,g,h,i,j,l)|0;break}default:k[l>>2]=410}while(0);n=k[l>>2]|0;return n|0}function Jq(a,b,c,d,e,f,g,h,l,m,n,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;q=q|0;var r=0,s=0;r=k[q>>2]|0;if((r|0)>0){s=r;return s|0}if(!n){Iq(a,b,c,d,e,f,g,h,l,m,q)|0;s=k[q>>2]|0;return s|0}do switch(b|0){case 11:{Tq(a,c,d,e,f,g,h,l,m,i[n>>0]|0,q)|0;break}case 12:{Ir(a,c,d,e,f,g,h,l,m,i[n>>0]|0,q)|0;break}case 20:{Pr(a,c,d,e,f,g,h,l,m,j[n>>1]|0,q)|0;break}case 21:{gr(a,c,d,e,f,g,h,l,m,j[n>>1]|0,q)|0;break}case 30:{zr(a,c,d,e,f,g,h,l,m,k[n>>2]|0,q)|0;break}case 31:{vr(a,c,d,e,f,g,h,l,m,k[n>>2]|0,q)|0;break}case 40:{Vr(a,c,d,e,f,g,h,l,m,k[n>>2]|0,q)|0;break}case 41:{lr(a,c,d,e,f,g,h,l,m,k[n>>2]|0,q)|0;break}case 81:{r=n;qr(a,c,d,e,f,g,h,l,m,k[r>>2]|0,k[r+4>>2]|0,q)|0;break}case 42:{br(a,c,d,e,f,g,h,l,m,+o[n>>2],q)|0;break}case 82:{Yq(a,c,d,e,f,g,h,l,m,+p[n>>3],q)|0;break}case 83:{r=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;br(a,c,d,e,r,L,ty(h|0,l|0,1)|0,L,m,+o[n>>2],q)|0;break}case 163:{r=ky(ty(f|0,g|0,1)|0,L|0,-1,-1)|0;Yq(a,c,d,e,r,L,ty(h|0,l|0,1)|0,L,m,+p[n>>3],q)|0;break}case 14:{Br(a,c,d,e,f,g,h,l,m,i[n>>0]|0,q)|0;break}case 16:{Er(a,c,d,e,f,g,h,l,m,n,q)|0;break}default:k[q>>2]=410}while(0);s=k[q>>2]|0;return s|0}function Kq(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;k[a>>2]=b;k[a+4>>2]=c;k[a+80>>2]=d;k[a+84>>2]=e;return 0}function Lq(a,b){a=a|0;b=b|0;k[a>>2]=b;return 0}function Mq(a,b){a=a|0;b=b|0;k[a+80>>2]=b;return 0}function Nq(a,b){a=a|0;b=b|0;k[a+84>>2]=b;return 0}function Oq(a){a=a|0;return k[a+88>>2]|0}function Pq(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0;h=r;r=r+400|0;l=h+64|0;m=h+56|0;n=h+48|0;q=h+40|0;s=h+32|0;t=h+24|0;u=h+16|0;v=h+8|0;w=h+164|0;x=h+124|0;y=h+120|0;z=h+116|0;A=h+112|0;B=h+108|0;C=h+104|0;D=h+100|0;E=h+96|0;F=h+92|0;G=h+88|0;H=h+84|0;I=h+80|0;J=h+76|0;K=h+128|0;L=h+72|0;M=h;N=h+314|0;O=h+239|0;P=h+168|0;Q=K;R=121436;S=Q+36|0;do{k[Q>>2]=k[R>>2];Q=Q+4|0;R=R+4|0}while((Q|0)<(S|0));p[M>>3]=0.0;R=k[g>>2]|0;if((R|0)>0){T=R;r=h;return T|0}if(a>>>0>999){tk(770127);k[g>>2]=302;T=302;r=h;return T|0}Zk(k[b>>2]|0,A,g)|0;R=(a|0)>0;a:do if(R){Q=0;b:while(1){S=k[b+(Q*244|0)+80>>2]|0;U=((S|0)<0?(k[b+(Q*244|0)+84>>2]|0)!=2:0)?0-S|0:S;switch(U|0){case 0:case 11:case 12:case 14:case 16:case 20:case 21:case 31:case 40:case 41:case 42:case 81:case 82:case 83:case 163:break;default:{V=U;W=S;X=Q;Y=6;break b}}S=b+(Q*244|0)+96|0;k[S>>2]=0;U=b+(Q*244|0)+100|0;k[U>>2]=0;Z=b+(Q*244|0)+104|0;i[Z>>0]=0;_=b+(Q*244|0)+174|0;i[_>>0]=0;$=b+(Q*244|0)|0;Zk(k[$>>2]|0,B,g)|0;aa=(k[B>>2]|0)==0;if(!(k[A>>2]|0)){if(!aa){ba=Q;Y=12;break}k[b+(Q*244|0)+4>>2]=0;ca=b+(Q*244|0)+8|0;i[ca>>0]=i[771935]|0;i[ca+1>>0]=i[771936]|0;i[ca+2>>0]=i[771937]|0;i[ca+3>>0]=i[771938]|0;i[ca+4>>0]=i[771939]|0;i[ca+5>>0]=i[771940]|0;k[w>>2]=0;Jo(k[$>>2]|0,770339,Z,0,w)|0}else{if(aa){da=Q;Y=15;break}aa=b+(Q*244|0)+4|0;ca=k[aa>>2]|0;if((ca|0)<1){ea=b+(Q*244|0)+8|0;if(Tk(k[$>>2]|0,0,ea,aa,g)|0){fa=ea;ga=Q;Y=18;break}ea=k[aa>>2]|0;if((ea|0)<1){ia=ea;ja=Q;Y=21;break}else ka=ea}else ka=ca;if((ka|0)>(k[(k[(k[$>>2]|0)+4>>2]|0)+928>>2]|0)){ia=ka;ja=Q;Y=21;break}k[w>>2]=0;Gk(770509,ka,O,w)|0;Ko(k[$>>2]|0,O,S,0,w)|0;k[w>>2]=0;Gk(770515,k[aa>>2]|0,O,w)|0;Ko(k[$>>2]|0,O,U,0,w)|0;k[w>>2]=0;Gk(772997,k[aa>>2]|0,O,w)|0;U=b+(Q*244|0)+8|0;Jo(k[$>>2]|0,O,U,0,w)|0;if(k[w>>2]|0)i[U>>0]=0;k[w>>2]=0;Gk(773394,k[aa>>2]|0,O,w)|0;Jo(k[$>>2]|0,O,Z,0,w)|0;k[w>>2]=0;Gk(770521,k[aa>>2]|0,O,w)|0;Jo(k[$>>2]|0,O,_,0,w)|0}Q=Q+1|0;if((Q|0)>=(a|0))break a}if((Y|0)==6){Q=X+1|0;if((V|0)<0){k[v>>2]=Q;Xw(N,770162,v)|0}else{k[u>>2]=Q;k[u+4>>2]=W;Xw(N,770233,u)|0}tk(N);k[g>>2]=410;T=410;r=h;return T|0}else if((Y|0)==12){k[t>>2]=ba+1;Xw(N,770285,t)|0;k[g>>2]=233;T=233;r=h;return T|0}else if((Y|0)==15){k[s>>2]=da+1;Xw(N,770345,s)|0;k[g>>2]=235;T=235;r=h;return T|0}else if((Y|0)==18){k[q>>2]=fa;k[q+4>>2]=ga+1;Xw(N,770398,q)|0;tk(N);T=k[g>>2]|0;r=h;return T|0}else if((Y|0)==21){k[n>>2]=ja+1;k[n+4>>2]=ia;Xw(N,770451,n)|0;tk(N);k[g>>2]=302;T=302;r=h;return T|0}}while(0);n=(c|0)>0?c:0;c=k[b>>2]|0;if(!(k[A>>2]|0)){Fl(c,x,g)|0;Gl(k[b>>2]|0,9,K,g)|0;k[w>>2]=0;Ko(k[b>>2]|0,772311,L,0,w)|0;ia=k[x>>2]|0;ja=k[K>>2]|0;if((k[w>>2]|0)==0&(k[L>>2]|0)!=0&(ia|0)>1&(ja|0)==0){Ko(k[b>>2]|0,782726,D,0,g)|0;la=k[x>>2]|0}else{k[D>>2]=ja;la=ia}if((la|0)>1){ia=k[D>>2]|0;ja=1;while(1){x=ha(ia,k[K+(ja<<2)>>2]|0)|0;ja=ja+1|0;if((ja|0)>=(la|0)){ma=x;break}else ia=x}k[D>>2]=ma}na=n+1|0;oa=1}else{Ko(c,790243,D,0,g)|0;na=1;oa=n+1|0}c=(k[D>>2]|0)-n|0;ma=(c|0)>0?c:0;k[D>>2]=ma;do if(d)if((d|0)<0){k[E>>2]=ma;break}else{k[E>>2]=(ma|0)>(d|0)?d:ma;break}else{ju(k[b>>2]|0,E,g)|0;if((a|0)>1){c=1;ia=1;while(1){la=k[b+(c*244|0)>>2]|0;ja=0;while(1){if((k[b+(ja*244|0)>>2]|0)==(la|0)){pa=ja;break}K=ja+1|0;if((K|0)<(c|0))ja=K;else{pa=K;break}}if((pa|0)==(c|0)){ju(k[b+(c*244|0)>>2]|0,F,g)|0;ja=k[E>>2]|0;la=k[F>>2]|0;k[E>>2]=(ja|0)<(la|0)?ja:la;qa=ia+1|0}else qa=ia;c=c+1|0;if((c|0)==(a|0)){ra=qa;break}else ia=qa}}else ra=1;ia=(k[E>>2]|0)/(ra|0)|0;k[E>>2]=(ia|0)>1?ia:1}while(0);ra=ey(a,16)|0;if(!ra){tk(770527);k[g>>2]=113;T=113;r=h;return T|0}c:do if(R){qa=0;d:while(1){F=b+(qa*244|0)|0;pa=k[F>>2]|0;e:do if(k[A>>2]|0){if((Xk(pa,k[b+(qa*244|0)+4>>2]|0,z,G,I,g)|0)>0)break c;ma=k[z>>2]|0;if((ma|0)<0?(k[E>>2]=1,(k[b+(qa*244|0)+84>>2]|0)==2):0){sa=qa;Y=63;break d}else ta=ma}else{El(pa,y,g)|0;switch(k[y>>2]|0){case 8:{k[z>>2]=11;ta=11;break e;break}case 16:{k[z>>2]=21;ta=21;break e;break}case 32:{k[z>>2]=41;ta=41;break e;break}case -32:{k[z>>2]=42;ta=42;break e;break}case -64:{k[z>>2]=82;ta=82;break e;break}case 64:{k[z>>2]=81;ta=81;break e;break}default:{ta=k[z>>2]|0;break e}}}while(0);if((((ta|0)>-1?ta:0-ta|0)|0)==1){pa=ta*11|0;k[z>>2]=pa;k[G>>2]=((k[G>>2]|0)+7|0)/8|0;ua=pa}else ua=ta;pa=b+(qa*244|0)+80|0;ma=k[pa>>2]|0;if(!ma){d=(ua|0)>-1?ua:0-ua|0;k[pa>>2]=d;va=d}else va=ma;f:do if((k[A>>2]|0)==0|(va|0)==16){ma=k[E>>2]|0;k[b+(qa*244|0)+92>>2]=1;switch(((ua|0)>-1?ua:0-ua|0)|0){case 41:case 21:case 11:break;default:{wa=ma;break f}}k[w>>2]=0;Ko(k[F>>2]|0,779663,J,0,w)|0;if(!(k[w>>2]|0))wa=ma;else{k[J>>2]=0;wa=ma}}else{if((ua|0)<0){k[G>>2]=1;if((k[D>>2]|0)>0){ma=b+(qa*244|0)+4|0;d=0;do{ia=d+oa|0;vl(k[F>>2]|0,k[ma>>2]|0,ia,((ia|0)<0)<<31>>31,H,0,g)|0;ia=k[G>>2]|0;c=k[H>>2]|0;la=(ia|0)>(c|0)?ia:c;k[G>>2]=la;d=d+1|0}while((d|0)<(k[D>>2]|0));xa=la;ya=xa;za=k[z>>2]|0}else{ya=1;za=ua}}else{ya=k[G>>2]|0;za=ua}d=ha(ya,k[E>>2]|0)|0;k[b+(qa*244|0)+92>>2]=ya;switch(((za|0)>-1?za:0-za|0)|0){case 41:case 21:case 11:break;default:{wa=d;break f}}k[w>>2]=0;ma=(k[A>>2]|0)==1;Gk(770576,k[b+(qa*244|0)+4>>2]|0,O,w)|0;la=k[F>>2]|0;if(!ma){Ko(la,O,J,0,w)|0;if(k[w>>2]|0){k[J>>2]=0;wa=d;break}if(k[J>>2]|0){wa=d;break}k[J>>2]=-2147483648;wa=d;break}Jo(la,O,P,0,w)|0;if(!(k[w>>2]|0))Aa=P;else{k[J>>2]=0;wa=d;break}g:while(1){switch(i[Aa>>0]|0){case 0:{Y=84;break g;break}case 32:break;default:break g}Aa=Aa+1|0}if((Y|0)==84){Y=0;k[J>>2]=-2147483648;wa=d;break}Ok(P,J,w)|0;if(!(k[w>>2]|0))wa=d;else{k[J>>2]=-2147483648;wa=d}}while(0);la=k[pa>>2]|0;h:do switch(la|0){case 11:{ma=ey(wa+1|0,1)|0;c=b+(qa*244|0)+88|0;k[c>>2]=ma;k[ra+(qa<<4)>>2]=1;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{ma=k[J>>2]|0;ia=(ma|0)<255?ma:255;ma=(ia|0)>0?ia:0;k[J>>2]=ma;i[ra+(qa<<4)+8>>0]=ma;Ba=c;break h;break}default:{i[ra+(qa<<4)+8>>0]=-1;Ba=c;break h}}break}case 12:{c=ey(wa+1|0,1)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=c;k[ra+(qa<<4)>>2]=1;c=k[z>>2]|0;switch(((c|0)>-1?c:0-c|0)|0){case 41:case 21:case 11:{c=k[J>>2]|0;ia=(c|0)<127?c:127;c=(ia|0)>-128?ia:-128;k[J>>2]=c;i[ra+(qa<<4)+8>>0]=c;Ba=ma;break h;break}default:{i[ra+(qa<<4)+8>>0]=-128;Ba=ma;break h}}break}case 21:{ma=ey(wa+1|0,2)|0;c=b+(qa*244|0)+88|0;k[c>>2]=ma;k[ra+(qa<<4)>>2]=2;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{ma=k[J>>2]|0;ia=(ma|0)<32767?ma:32767;ma=(ia|0)>-32768?ia:-32768;k[J>>2]=ma;j[ra+(qa<<4)+8>>1]=ma;Ba=c;break h;break}default:{j[ra+(qa<<4)+8>>1]=-32768;Ba=c;break h}}break}case 20:{c=ey(wa+1|0,2)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=c;k[ra+(qa<<4)>>2]=2;c=k[z>>2]|0;switch(((c|0)>-1?c:0-c|0)|0){case 41:case 21:case 11:{c=k[J>>2]|0;ia=(c|0)<65535?c:65535;c=(ia|0)>0?ia:0;k[J>>2]=c;j[ra+(qa<<4)+8>>1]=c;Ba=ma;break h;break}default:{j[ra+(qa<<4)+8>>1]=-1;Ba=ma;break h}}break}case 31:{ma=ey(4,wa+1|0)|0;c=b+(qa*244|0)+88|0;k[c>>2]=ma;k[ra+(qa<<4)>>2]=4;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{k[ra+(qa<<4)+8>>2]=k[J>>2];Ba=c;break h;break}default:{k[ra+(qa<<4)+8>>2]=-2147483648;Ba=c;break h}}break}case 30:{c=ey(wa+1|0,4)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=c;k[ra+(qa<<4)>>2]=4;c=k[z>>2]|0;switch(((c|0)>-1?c:0-c|0)|0){case 41:case 21:case 11:{c=k[J>>2]|0;ia=(c|0)>0?c:0;k[J>>2]=ia;k[ra+(qa<<4)+8>>2]=ia;Ba=ma;break h;break}default:{k[ra+(qa<<4)+8>>2]=-1;Ba=ma;break h}}break}case 41:{ma=ey(wa+1|0,4)|0;ia=b+(qa*244|0)+88|0;k[ia>>2]=ma;k[ra+(qa<<4)>>2]=4;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{k[ra+(qa<<4)+8>>2]=k[J>>2];Ba=ia;break h;break}default:{k[ra+(qa<<4)+8>>2]=-2147483648;Ba=ia;break h}}break}case 40:{ia=ey(wa+1|0,4)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=ia;k[ra+(qa<<4)>>2]=4;ia=k[z>>2]|0;switch(((ia|0)>-1?ia:0-ia|0)|0){case 41:case 21:case 11:break;default:{k[ra+(qa<<4)+8>>2]=2147483647;Ba=ma;break h}}ia=k[J>>2]|0;c=ra+(qa<<4)+8|0;if((ia|0)<0){k[c>>2]=2147483647;Ba=ma;break h}else{k[c>>2]=ia;Ba=ma;break h}break}case 42:{ma=ey(wa+1|0,4)|0;ia=b+(qa*244|0)+88|0;k[ia>>2]=ma;k[ra+(qa<<4)>>2]=4;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{o[ra+(qa<<4)+8>>2]=+(k[J>>2]|0);Ba=ia;break h;break}default:{o[ra+(qa<<4)+8>>2]=-9.119119840596153e-36;Ba=ia;break h}}break}case 83:{ia=ey(wa<<1|1,4)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=ia;k[ra+(qa<<4)>>2]=4;o[ra+(qa<<4)+8>>2]=-9.119119840596153e-36;Ba=ma;break}case 82:{ma=ey(wa+1|0,8)|0;ia=b+(qa*244|0)+88|0;k[ia>>2]=ma;k[ra+(qa<<4)>>2]=8;ma=k[z>>2]|0;switch(((ma|0)>-1?ma:0-ma|0)|0){case 41:case 21:case 11:{p[ra+(qa<<4)+8>>3]=+(k[J>>2]|0);Ba=ia;break h;break}default:{p[ra+(qa<<4)+8>>3]=-9.1191291391491e-36;Ba=ia;break h}}break}case 163:{ia=ey(wa<<1|1,8)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=ia;k[ra+(qa<<4)>>2]=8;p[ra+(qa<<4)+8>>3]=-9.1191291391491e-36;Ba=ma;break}case 16:{if((k[A>>2]|0)==1){ma=k[I>>2]|0;k[G>>2]=ma;Ca=ma}else Ca=k[G>>2]|0;ma=wa+1|0;ia=ey(ma,4)|0;c=b+(qa*244|0)+88|0;k[c>>2]=ia;ja=Ca+1|0;k[ra+(qa<<4)>>2]=ja;if(ia){K=ey(ja,1)|0;x=ra+(qa<<4)+8|0;k[x>>2]=K;i[K+1>>0]=1;K=ey(ha(ja,ma)|0,1)|0;k[ia>>2]=K;if(!K){Y=135;break d}if((wa|0)>=1){L=K;K=1;do{L=L+ja|0;k[ia+(K<<2)>>2]=L;K=K+1|0}while((K|0)!=(ma|0))}k[w>>2]=0;Gk(770576,k[b+(qa*244|0)+4>>2]|0,O,w)|0;Jo(k[F>>2]|0,O,P,0,w)|0;if(!(k[w>>2]|0)){zx(k[x>>2]|0,P,k[G>>2]|0)|0;Ba=c}else Ba=c}else Ba=c;break}case 14:{ma=ey(wa+1|0,1)|0;K=b+(qa*244|0)+88|0;k[K>>2]=ma;k[ra+(qa<<4)>>2]=1;i[ra+(qa<<4)+8>>0]=2;Ba=K;break}case 81:{K=ey(wa+1|0,8)|0;ma=b+(qa*244|0)+88|0;k[ma>>2]=K;k[ra+(qa<<4)>>2]=8;K=k[z>>2]|0;switch(((K|0)>-1?K:0-K|0)|0){case 81:case 41:case 21:case 11:{K=k[J>>2]|0;L=ra+(qa<<4)+8|0;k[L>>2]=K;k[L+4>>2]=((K|0)<0)<<31>>31;Ba=ma;break h;break}default:{K=ra+(qa<<4)+8|0;k[K>>2]=0;k[K+4>>2]=-2147483648;Ba=ma;break h}}break}default:{Da=la;Ea=qa;Y=140;break d}}while(0);if(!(k[Ba>>2]|0)){Y=142;break}qa=qa+1|0;if((qa|0)>=(a|0)){Y=144;break c}}if((Y|0)==63){k[m>>2]=sa+1;Xw(N,770162,m)|0;tk(N);k[g>>2]=410;T=410;r=h;return T|0}else if((Y|0)==135){tk(770582);k[g>>2]=113;break}else if((Y|0)==140){k[l>>2]=Ea+1;k[l+4>>2]=Da;Xw(N,770622,l)|0;tk(N);k[g>>2]=410;break}else if((Y|0)==142){tk(770582);k[g>>2]=113;break}}else Y=144;while(0);i:do if((Y|0)==144?(N=k[D>>2]|0,(N|0)!=0):0){l=na;Da=oa;Ea=N;while(1){N=((Da|0)<0)<<31>>31;m=l;sa=Ea;while(1){Ba=k[E>>2]|0;J=(sa|0)<(Ba|0)?sa:Ba;j:do if(R){Ba=((m|0)<0)<<31>>31;wa=0;do{do if((k[b+(wa*244|0)+84>>2]|0)!=2){P=b+(wa*244|0)+80|0;O=k[P>>2]|0;Ca=b+(wa*244|0)+88|0;Aa=k[Ca>>2]|0;if((O|0)==16){Fa=Aa+4|0;Ga=k[ra+(wa<<4)+8>>2]|0}else{Fa=Aa+(k[ra+(wa<<4)>>2]|0)|0;Ga=ra+(wa<<4)+8|0}Aa=b+(wa*244|0)|0;za=k[Aa>>2]|0;if(!(k[A>>2]|0)){ya=ha(k[b+(wa*244|0)+92>>2]|0,J)|0;if((Yl(za,O,m,Ba,ya,((ya|0)<0)<<31>>31,Ga,Fa,C,g)|0)>0)break j}else{ya=b+(wa*244|0)+4|0;if((Xk(za,k[ya>>2]|0,z,G,I,g)|0)>0)break i;if((k[z>>2]|0)<0){za=b+(wa*244|0)+92|0;vl(k[Aa>>2]|0,k[ya>>2]|0,Da,N,za,0,g)|0;Ha=za}else Ha=b+(wa*244|0)+92|0;za=ha(k[Ha>>2]|0,J)|0;if((_l(k[Aa>>2]|0,k[P>>2]|0,k[ya>>2]|0,Da,N,m,Ba,za,((za|0)<0)<<31>>31,Ga,Fa,C,g)|0)>0)break j}za=(k[P>>2]|0)==16;P=k[Ca>>2]|0;Ca=k[ra+(wa<<4)>>2]|0;if(!(k[C>>2]|0))if(za){sy(k[P>>2]|0,0,Ca|0)|0;break}else{sy(P|0,0,Ca|0)|0;break}else if(za){py(k[P>>2]|0,k[ra+(wa<<4)+8>>2]|0,Ca|0)|0;break}else{py(P|0,Ga|0,Ca|0)|0;break}}while(0);wa=wa+1|0}while((wa|0)<(a|0))}while(0);if((k[g>>2]|0)>0)break i;c=k[D>>2]|0;if(!(k[A>>2]|0))Ia=Ob[e&7](c,n,m,J,a,b,f)|0;else Ia=Ob[e&7](c,n,Da,J,a,b,f)|0;k[g>>2]=Ia;if((Ia+1|0)>>>0>1)break i;k[w>>2]=0;k:do if(R){c=((m|0)<0)<<31>>31;x=0;do{do if(k[b+(x*244|0)+84>>2]|0){wa=b+(x*244|0)+80|0;Ba=k[wa>>2]|0;Ca=k[b+(x*244|0)+88>>2]|0;if((Ba|0)==16){Ja=Ca+4|0;Ka=2;La=k[Ca>>2]|0}else{P=k[ra+(x<<4)>>2]|0;Ja=Ca+P|0;Ka=P;La=Ca}Ca=(lx(La,M,Ka)|0)==0;P=(k[A>>2]|0)==0;za=b+(x*244|0)|0;ya=k[za>>2]|0;if(Ca){if(P){Ca=ha(k[b+(x*244|0)+92>>2]|0,J)|0;if((Gq(ya,Ba,m,c,Ca,((Ca|0)<0)<<31>>31,Ja,w)|0)>0)break k;else break}Ca=b+(x*244|0)+4|0;if((Xk(ya,k[Ca>>2]|0,z,G,I,g)|0)>0)break i;if((k[z>>2]|0)<0){Aa=b+(x*244|0)+92|0;vl(k[za>>2]|0,k[Ca>>2]|0,Da,N,Aa,0,g)|0;Ma=Aa}else Ma=b+(x*244|0)+92|0;Aa=ha(k[Ma>>2]|0,J)|0;if((Iq(k[za>>2]|0,k[wa>>2]|0,k[Ca>>2]|0,Da,N,m,c,Aa,((Aa|0)<0)<<31>>31,Ja,w)|0)>0)break k;else break}else{if(P){P=ha(k[b+(x*244|0)+92>>2]|0,J)|0;if((Hq(ya,Ba,m,c,P,((P|0)<0)<<31>>31,Ja,La,w)|0)>0)break k;else break}P=b+(x*244|0)+4|0;if((Xk(ya,k[P>>2]|0,z,G,I,g)|0)>0)break i;if((k[z>>2]|0)<0){ya=b+(x*244|0)+92|0;vl(k[za>>2]|0,k[P>>2]|0,Da,N,ya,0,g)|0;Na=ya}else Na=b+(x*244|0)+92|0;ya=ha(k[Na>>2]|0,J)|0;if((Jq(k[za>>2]|0,k[wa>>2]|0,k[P>>2]|0,Da,N,m,c,ya,((ya|0)<0)<<31>>31,Ja,La,w)|0)>0)break k;else break}}while(0);x=x+1|0}while((x|0)<(a|0))}while(0);if(k[g>>2]|0)break i;x=k[w>>2]|0;k[g>>2]=x;if(x)break i;x=sa;sa=sa-J|0;if(k[A>>2]|0){Oa=sa;Pa=J;Qa=m;Ra=x;break}if((x|0)==(J|0))break i;else m=J+m|0}if((Ra|0)==(Pa|0))break;else{l=Qa;Da=Pa+Da|0;Ea=Oa}}}while(0);if(R){R=0;do{Oa=b+(R*244|0)+88|0;if((k[b+(R*244|0)+80>>2]|0)==16?(Pa=k[Oa>>2]|0,(Pa|0)!=0):0){dy(k[Pa>>2]|0);dy(k[ra+(R<<4)+8>>2]|0)}Pa=k[Oa>>2]|0;if(Pa)dy(Pa);R=R+1|0}while((R|0)!=(a|0))}dy(ra);T=k[g>>2]|0;r=h;return T|0}function Qq(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;Rq(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,11,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function Rq(a,b,c,d,e,f,g,h,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;n=n|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0.0,ea=0.0,fa=0.0,ja=0.0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0;q=r;r=r+29056|0;s=q+28864|0;t=q+28856|0;u=q+28896|0;v=q+28892|0;w=q+28888|0;x=q+28884|0;y=q+28880|0;z=q+48|0;A=q+40|0;B=q+32|0;C=q+24|0;D=q+8|0;E=q;F=q+28940|0;G=q+28920|0;H=q+28960|0;I=q+56|0;J=k[n>>2]|0;if((J|0)>0){K=J;r=q;return K|0}if((nl(a,b,c,d,e,f,g,h,1,D,E,F,x,u,v,A,B,y,z,C,w,q+16|0,q+28900|0,n)|0)>0){K=k[n>>2]|0;r=q;return K|0}f=k[v>>2]|0;v=k[u>>2]|0;if((v|0)==16){Rk(F,G);M=k[u>>2]|0}else M=v;v=+p[D>>3]==1.0&+p[E>>3]==0.0&(M|0)==11;M=(h|0)<0|(h|0)==0&g>>>0<2147483647;e=v?(M?g:2147483647):f;d=v?(M?h:0):((f|0)<0)<<31>>31;f=G+1|0;M=g;g=h;h=0;c=0;J=0;N=0;a:while(1){O=c;P=h;Q=M;R=g;while(1){S=m+P|0;T=Q;U=R;while(1){if((T|0)==0&(U|0)==0){X=90;break a}Y=(U|0)<(d|0)|(U|0)==(d|0)&T>>>0<e>>>0;Z=Y?T:e;Y=my(0,Z|0,32)|0;Z=L;_=z;$=B;aa=k[$>>2]|0;ba=k[$+4>>2]|0;$=jy(k[_>>2]|0,k[_+4>>2]|0,aa|0,ba|0)|0;_=L;ca=(Z|0)<(_|0)|(Z|0)==(_|0)&Y>>>0<$>>>0;_=ca?Y:$;$=A;Y=k[$>>2]|0;ca=k[$+4>>2]|0;$=C;Z=yy(k[$>>2]|0,k[$+4>>2]|0,J|0,N|0)|0;$=ky(Z|0,L|0,Y|0,ca|0)|0;ca=L;Y=k[y>>2]|0;Z=yy(Y|0,((Y|0)<0)<<31>>31|0,aa|0,ba|0)|0;ba=ky($|0,ca|0,Z|0,L|0)|0;au(a,ba,L,1,n)|0;b:do switch(k[u>>2]|0){case 11:{if(v){su(a,_,k[y>>2]|0,S,n)|0;break b}da=+p[D>>3];ea=+p[E>>3];if(!(da==1.0&ea==0.0)){if((_|0)>0){ba=0;do{fa=(+(l[m+(ba+P)>>0]|0)-ea)/da;do if(!(fa<-.49))if(fa>255.49){k[n>>2]=-11;i[I+ba>>0]=-1;break}else{i[I+ba>>0]=~~(fa+.5);break}else{k[n>>2]=-11;i[I+ba>>0]=0}while(0);ba=ba+1|0}while((ba|0)!=(_|0))}}else py(I|0,S|0,_|0)|0;su(a,_,k[y>>2]|0,I,n)|0;break}case 81:{da=+p[D>>3];ea=+p[E>>3];ba=(_|0)>0;if(da==1.0&ea==0.0){if(ba){Z=0;do{ca=I+(Z<<3)|0;k[ca>>2]=l[m+(Z+P)>>0];k[ca+4>>2]=0;Z=Z+1|0}while((Z|0)!=(_|0))}}else if(ba){Z=0;do{fa=(+(l[m+(Z+P)>>0]|0)-ea)/da;do if(!(fa<-9223372036854775808.0)){if(fa>9223372036854775808.0){k[n>>2]=-11;ca=I+(Z<<3)|0;k[ca>>2]=-1;k[ca+4>>2]=2147483647;break}if(!(fa>=0.0)){ja=fa+-.5;ca=+W(ja)>=1.0?(ja>0.0?~~+ia(+V(ja/4294967296.0),4294967295.0)>>>0:~~+ga((ja-+(~~ja>>>0))/4294967296.0)>>>0):0;$=I+(Z<<3)|0;k[$>>2]=~~ja>>>0;k[$+4>>2]=ca;break}else{ja=fa+.5;ca=+W(ja)>=1.0?(ja>0.0?~~+ia(+V(ja/4294967296.0),4294967295.0)>>>0:~~+ga((ja-+(~~ja>>>0))/4294967296.0)>>>0):0;$=I+(Z<<3)|0;k[$>>2]=~~ja>>>0;k[$+4>>2]=ca;break}}else{k[n>>2]=-11;ca=I+(Z<<3)|0;k[ca>>2]=0;k[ca+4>>2]=-2147483648}while(0);Z=Z+1|0}while((Z|0)!=(_|0))}vu(a,_,k[y>>2]|0,I,n)|0;break}case 21:{da=+p[D>>3];ea=+p[E>>3];Z=(_|0)>0;if(da==1.0&ea==0.0){if(Z){ba=0;do{j[I+(ba<<1)>>1]=l[m+(ba+P)>>0]|0;ba=ba+1|0}while((ba|0)!=(_|0))}}else if(Z){ba=0;do{fa=(+(l[m+(ba+P)>>0]|0)-ea)/da;do if(!(fa<-32768.49)){if(fa>32767.49){k[n>>2]=-11;j[I+(ba<<1)>>1]=32767;break}if(!(fa>=0.0)){j[I+(ba<<1)>>1]=~~(fa+-.5);break}else{j[I+(ba<<1)>>1]=~~(fa+.5);break}}else{k[n>>2]=-11;j[I+(ba<<1)>>1]=-32768}while(0);ba=ba+1|0}while((ba|0)!=(_|0))}tu(a,_,k[y>>2]|0,I,n)|0;break}case 41:{da=+p[D>>3];ea=+p[E>>3];ba=(_|0)>0;if(da==1.0&ea==0.0){if(ba){Z=0;do{k[I+(Z<<2)>>2]=l[m+(Z+P)>>0];Z=Z+1|0}while((Z|0)!=(_|0))}}else if(ba){Z=0;do{fa=(+(l[m+(Z+P)>>0]|0)-ea)/da;do if(!(fa<-2147483648.49)){if(fa>2147483647.49){k[n>>2]=-11;k[I+(Z<<2)>>2]=2147483647;break}if(!(fa>=0.0)){k[I+(Z<<2)>>2]=~~(fa+-.5);break}else{k[I+(Z<<2)>>2]=~~(fa+.5);break}}else{k[n>>2]=-11;k[I+(Z<<2)>>2]=-2147483648}while(0);Z=Z+1|0}while((Z|0)!=(_|0))}uu(a,_,k[y>>2]|0,I,n)|0;break}case 42:{da=+p[D>>3];ea=+p[E>>3];Z=(_|0)>0;if(da==1.0&ea==0.0){if(Z){ba=0;do{o[I+(ba<<2)>>2]=+(l[m+(ba+P)>>0]|0);ba=ba+1|0}while((ba|0)!=(_|0))}}else if(Z){ba=0;do{o[I+(ba<<2)>>2]=(+(l[m+(ba+P)>>0]|0)-ea)/da;ba=ba+1|0}while((ba|0)!=(_|0))}wu(a,_,k[y>>2]|0,I,n)|0;break}case 82:{da=+p[D>>3];ea=+p[E>>3];ba=(_|0)>0;if(da==1.0&ea==0.0){if(ba){Z=0;do{p[I+(Z<<3)>>3]=+(l[m+(Z+P)>>0]|0);Z=Z+1|0}while((Z|0)!=(_|0))}}else if(ba){Z=0;do{p[I+(Z<<3)>>3]=(+(l[m+(Z+P)>>0]|0)-ea)/da;Z=Z+1|0}while((Z|0)!=(_|0))}xu(a,_,k[y>>2]|0,I,n)|0;break}case 16:{if(rx(F,65)|0){Z=k[y>>2]|0;ba=k[x>>2]|0;if((Z|0)==(ba|0)){cu(a,_,((_|0)<0)<<31>>31,S,n)|0;break b}else{eu(a,ba,(_|0)/(ba|0)|0,Z-ba|0,S,n)|0;break b}}if((i[f>>0]|0)==115){X=82;break a}Uq(S,_,+p[D>>3],+p[E>>3],G,k[x>>2]|0,I,n)|0;ba=k[y>>2]|0;Z=k[x>>2]|0;if((ba|0)==(Z|0)){ca=ha(ba,_)|0;cu(a,ca,((ca|0)<0)<<31>>31,I,n)|0;break b}else{eu(a,Z,_,ba-Z|0,I,n)|0;break b}break}default:{X=82;break a}}while(0);if((k[n>>2]|0)>0){ka=P;la=O;ma=_;X=86;break a}Z=((_|0)<0)<<31>>31;ba=jy(T|0,U|0,_|0,Z|0)|0;ca=L;if((T|0)==(_|0)&(U|0)==(Z|0)){T=ba;U=ca}else{na=_;oa=Z;pa=ba;qa=ca;break}}U=ky(na|0,oa|0,P|0,O|0)|0;T=L;S=B;ca=ky(k[S>>2]|0,k[S+4>>2]|0,na|0,oa|0)|0;S=L;ba=B;k[ba>>2]=ca;k[ba+4>>2]=S;ba=z;if((ca|0)==(k[ba>>2]|0)?(S|0)==(k[ba+4>>2]|0):0){ra=U;sa=T;ta=pa;ua=qa;break}else{O=T;P=U;Q=pa;R=qa}}R=B;k[R>>2]=0;k[R+4>>2]=0;R=ky(J|0,N|0,1,0)|0;M=ta;g=ua;h=ra;c=sa;J=R;N=L}if((X|0)==82){k[t>>2]=b;k[t+4>>2]=F;Xw(H,771690,t)|0;tk(H);if((k[w>>2]|0)==1){k[n>>2]=311;K=311;r=q;return K|0}else{k[n>>2]=312;K=312;r=q;return K|0}}else if((X|0)==86){w=ky(ka|0,la|0,1,0)|0;da=+(w>>>0)+4294967296.0*+(L|0);w=ky(ma|0,((ma|0)<0)<<31>>31|0,ka|0,la|0)|0;p[s>>3]=da;p[s+8>>3]=+(w>>>0)+4294967296.0*+(L|0);Xw(H,770680,s)|0;tk(H);K=k[n>>2]|0;r=q;return K|0}else if((X|0)==90){X=k[n>>2]|0;if((X|0)!=-11){K=X;r=q;return K|0}tk(771813);k[n>>2]=412;K=412;r=q;return K|0}return 0}function Sq(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;Tq(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,j)|0;o=k[j>>2]|0;r=l;return o|0}else{i[m>>0]=h;jt(a,11,c,d,e,f,1,g,m,j)|0;o=k[j>>2]|0;r=l;return o|0}return 0}function Tq(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[n+(p*160|0)+80>>2]|0;q=(r|0)>0;do if(!q){s=ky(e|0,f|0,-1,-1)|0;t=ky(s|0,L|0,g|0,h|0)|0;s=L;if((r|0)<0?(Rq(a,b,c,d,e,f,g,h,j,m)|0)>0:0){u=k[m>>2]|0;if((u|0)==412){k[m>>2]=0;v=t;w=s;break}else{o=u;return o|0}}else{v=t;w=s}}else{s=n+(p*160|0)+88|0;v=k[s>>2]|0;w=k[s+4>>2]|0}while(0);p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,w|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){o=k[m>>2]|0;return o|0}a:do if(q){e=0;d=0;c=0;n=0;r=0;s=0;t=0;while(1){if((i[j+e>>0]|0)==l<<24>>24){if(!((r|0)==0&(s|0)==0)?(u=jy(e|0,n|0,r|0,s|0)|0,x=ky(u|0,L|0,p|0,f|0)|0,y=L,z=ky(x|0,y|0,-1,-1)|0,A=wy(z|0,L|0,v|0,w|0)|0,z=L,B=yy(A|0,z|0,v|0,w|0)|0,C=jy(x|0,y|0,B|0,L|0)|0,B=L,y=ky(A|0,z|0,1,0)|0,(Rq(a,b,y,L,C,B,r,s,j+u|0,m)|0)>0):0){u=k[m>>2]|0;if((u|0)!=412){o=u;break}k[m>>2]=0;D=1}else D=t;u=ky(d|0,c|0,1,0)|0;E=0;F=0;G=u;H=L;I=D}else{if(!((d|0)==0&(c|0)==0)?(u=jy(e|0,n|0,d|0,c|0)|0,B=ky(u|0,L|0,p|0,f|0)|0,u=L,C=ky(B|0,u|0,-1,-1)|0,y=wy(C|0,L|0,v|0,w|0)|0,C=L,z=ky(y|0,C|0,1,0)|0,A=L,x=yy(y|0,C|0,v|0,w|0)|0,C=jy(B|0,u|0,x|0,L|0)|0,(Kr(a,b,z,A,C,L,d,c,m)|0)>0):0){J=29;break a}C=ky(r|0,s|0,1,0)|0;E=C;F=L;G=0;H=0;I=t}C=ky(e|0,n|0,1,0)|0;A=L;if((A|0)<(h|0)|(A|0)==(h|0)&C>>>0<g>>>0){e=C;d=G;c=H;n=A;r=E;s=F;t=I}else{K=E;M=F;N=C;O=A;P=G;Q=H;R=I;break a}}return o|0}else{t=0;s=0;r=0;n=0;c=0;d=0;while(1){if((i[j+t>>0]|0)==l<<24>>24){e=ky(s|0,r|0,1,0)|0;S=0;T=0;U=e;V=L}else{if(!((s|0)==0&(r|0)==0)?(e=jy(t|0,n|0,s|0,r|0)|0,A=ky(e|0,L|0,p|0,f|0)|0,e=L,C=ky(A|0,e|0,-1,-1)|0,z=wy(C|0,L|0,v|0,w|0)|0,C=L,x=ky(z|0,C|0,1,0)|0,u=L,B=yy(z|0,C|0,v|0,w|0)|0,C=jy(A|0,e|0,B|0,L|0)|0,(Kr(a,b,x,u,C,L,s,r,m)|0)>0):0){J=29;break a}C=ky(c|0,d|0,1,0)|0;S=C;T=L;U=0;V=0}C=ky(t|0,n|0,1,0)|0;u=L;if((u|0)<(h|0)|(u|0)==(h|0)&C>>>0<g>>>0){t=C;s=U;r=V;n=u;c=S;d=T}else{K=S;M=T;N=C;O=u;P=U;Q=V;R=0;break}}}while(0);if((J|0)==29){o=k[m>>2]|0;return o|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){J=jy(N|0,O|0,P|0,Q|0)|0;V=ky(J|0,L|0,p|0,f|0)|0;J=L;U=ky(V|0,J|0,-1,-1)|0;T=wy(U|0,L|0,v|0,w|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,v|0,w|0)|0;U=jy(V|0,J|0,h|0,L|0)|0;Kr(a,b,S,g,U,L,P,Q,m)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,p|0,f|0)|0;f=L;p=ky(O|0,f|0,-1,-1)|0;N=wy(p|0,L|0,v|0,w|0)|0;p=L;if(q){q=yy(N|0,p|0,v|0,w|0)|0;w=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,p|0,1,0)|0;Rq(a,b,f,L,w,q,K,M,j+Q|0,m)|0}}Q=k[m>>2]|0;if(!((R|0)!=0&(Q|0)<1)){o=Q;return o|0}k[m>>2]=412;o=412;return o|0}function Uq(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0;j=r;r=r+16|0;m=j+8|0;n=j;o=(b|0)>0;if(c==1.0&d==0.0){if(o){q=g;s=0;do{p[n>>3]=+(l[a+s>>0]|0);Xw(q,e,n)|0;q=q+f|0;if(i[q>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}}else if(o){o=g;s=0;do{p[m>>3]=(+(l[a+s>>0]|0)-d)/c;Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}b=rx(g,44)|0;if(!b){t=k[h>>2]|0;r=j;return t|0}else u=b;do{i[u>>0]=46;u=rx(u,44)|0}while((u|0)!=0);t=k[h>>2]|0;r=j;return t|0}function Vq(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;Wq(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,82,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function Wq(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}f=k[u>>2]|0;u=((f|0)<0)<<31>>31;if((k[t>>2]|0)==16)Rk(E,F);e=F+1|0;d=g;g=h;h=0;c=0;I=0;K=0;a:while(1){M=h;N=c;O=d;P=g;while(1){Q=l+(M<<3)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=104;break a}U=(S|0)<(u|0)|(S|0)==(u|0)&R>>>0<f>>>0;X=U?R:f;U=my(0,X|0,32)|0;X=L;Y=y;Z=A;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=jy(k[Y>>2]|0,k[Y+4>>2]|0,_|0,$|0)|0;Y=L;aa=(X|0)<(Y|0)|(X|0)==(Y|0)&U>>>0<Z>>>0;Y=aa?U:Z;Z=z;U=k[Z>>2]|0;aa=k[Z+4>>2]|0;Z=B;X=yy(k[Z>>2]|0,k[Z+4>>2]|0,I|0,K|0)|0;Z=ky(X|0,L|0,U|0,aa|0)|0;aa=L;U=k[x>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,_|0,$|0)|0;$=ky(Z|0,aa|0,X|0,L|0)|0;au(a,$,L,1,m)|0;b:do switch(k[t>>2]|0){case 82:{ba=+p[C>>3];ca=+p[D>>3];if(!(ba==1.0&ca==0.0)){if((Y|0)>0){$=0;do{p[H+($<<3)>>3]=(+p[l+($+M<<3)>>3]-ca)/ba;$=$+1|0}while(($|0)!=(Y|0))}}else py(H|0,Q|0,Y<<3|0)|0;xu(a,Y,k[x>>2]|0,H,m)|0;break}case 81:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{da=+p[l+(X+M<<3)>>3];do if(!(da<-9223372036854775808.0))if(da>9223372036854775808.0){k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}else{aa=+W(da)>=1.0?(da>0.0?~~+ia(+V(da/4294967296.0),4294967295.0)>>>0:~~+ga((da-+(~~da>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~da>>>0;k[Z+4>>2]=aa;break}else{k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+p[l+(X+M<<3)>>3]-ca)/ba;do if(!(da<-9223372036854775808.0)){if(da>9223372036854775808.0){k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}if(!(da>=0.0)){ea=da+-.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}else{ea=da+.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}}else{k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}vu(a,Y,k[x>>2]|0,H,m)|0;break}case 11:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{da=+p[l+($+M<<3)>>3];do if(!(da<-.49))if(da>255.49){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=~~da;break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+p[l+($+M<<3)>>3]-ca)/ba;do if(!(da<-.49))if(da>255.49){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=~~(da+.5);break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}su(a,Y,k[x>>2]|0,H,m)|0;break}case 21:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{da=+p[l+(X+M<<3)>>3];do if(!(da<-32768.49))if(da>32767.49){k[m>>2]=-11;j[H+(X<<1)>>1]=32767;break}else{j[H+(X<<1)>>1]=~~da;break}else{k[m>>2]=-11;j[H+(X<<1)>>1]=-32768}while(0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+p[l+(X+M<<3)>>3]-ca)/ba;do if(!(da<-32768.49)){if(da>32767.49){k[m>>2]=-11;j[H+(X<<1)>>1]=32767;break}if(!(da>=0.0)){j[H+(X<<1)>>1]=~~(da+-.5);break}else{j[H+(X<<1)>>1]=~~(da+.5);break}}else{k[m>>2]=-11;j[H+(X<<1)>>1]=-32768}while(0);X=X+1|0}while((X|0)!=(Y|0))}tu(a,Y,k[x>>2]|0,H,m)|0;break}case 41:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{da=+p[l+($+M<<3)>>3];do if(!(da<-2147483648.49))if(da>2147483647.49){k[m>>2]=-11;k[H+($<<2)>>2]=2147483647;break}else{k[H+($<<2)>>2]=~~da;break}else{k[m>>2]=-11;k[H+($<<2)>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+p[l+($+M<<3)>>3]-ca)/ba;do if(!(da<-2147483648.49)){if(da>2147483647.49){k[m>>2]=-11;k[H+($<<2)>>2]=2147483647;break}if(!(da>=0.0)){k[H+($<<2)>>2]=~~(da+-.5);break}else{k[H+($<<2)>>2]=~~(da+.5);break}}else{k[m>>2]=-11;k[H+($<<2)>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(Y|0))}uu(a,Y,k[x>>2]|0,H,m)|0;break}case 42:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{o[H+(X<<2)>>2]=+p[l+(X+M<<3)>>3];X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{o[H+(X<<2)>>2]=(+p[l+(X+M<<3)>>3]-ca)/ba;X=X+1|0}while((X|0)!=(Y|0))}wu(a,Y,k[x>>2]|0,H,m)|0;break}case 16:{if((i[e>>0]|0)==115){T=96;break a}Zq(Q,Y,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;X=k[x>>2]|0;$=k[w>>2]|0;if((X|0)==($|0)){aa=ha(X,Y)|0;cu(a,aa,((aa|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,$,Y,X-$|0,H,m)|0;break b}break}default:{T=96;break a}}while(0);if((k[m>>2]|0)>0){fa=M;ja=N;ka=Y;T=100;break a}$=((Y|0)<0)<<31>>31;X=jy(R|0,S|0,Y|0,$|0)|0;aa=L;if((R|0)==(Y|0)&(S|0)==($|0)){R=X;S=aa}else{la=Y;ma=$;na=X;oa=aa;break}}S=ky(la|0,ma|0,M|0,N|0)|0;R=L;Q=A;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,la|0,ma|0)|0;Q=L;X=A;k[X>>2]=aa;k[X+4>>2]=Q;X=y;if((aa|0)==(k[X>>2]|0)?(Q|0)==(k[X+4>>2]|0):0){pa=S;qa=R;ra=na;sa=oa;break}else{M=S;N=R;O=na;P=oa}}P=A;k[P>>2]=0;k[P+4>>2]=0;P=ky(I|0,K|0,1,0)|0;d=ra;g=sa;h=pa;c=qa;I=P;K=L}if((T|0)==96){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((T|0)==100){v=ky(fa|0,ja|0,1,0)|0;ba=+(v>>>0)+4294967296.0*+(L|0);v=ky(ka|0,((ka|0)<0)<<31>>31|0,fa|0,ja|0)|0;p[q>>3]=ba;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,770748,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((T|0)==104){T=k[m>>2]|0;if((T|0)!=-11){J=T;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function Xq(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;Yq(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{p[l>>3]=h;jt(a,82,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function Yq(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=+j;l=l|0;var m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;o=a+4|0;q=k[o>>2]|0;if((m|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[o>>2]|0)+960>>2]|0;o=b+-1|0;r=k[m+(o*160|0)+80>>2]|0;q=(r|0)>0;if(q){s=m+(o*160|0)+88|0;t=k[s>>2]|0;u=k[s+4>>2]|0}else{s=ky(e|0,f|0,-1,-1)|0;o=ky(s|0,L|0,g|0,h|0)|0;t=o;u=L}o=ty(t|0,u|0,(((r|0)>-1?r:0-r|0)|0)>82|0)|0;u=L;do if((r|0)<0?(Wq(a,b,c,d,e,f,g,h,i,l)|0)>0:0){t=k[l>>2]|0;if((t|0)==412){k[l>>2]=0;break}else{n=t;return n|0}}while(0);r=ky(c|0,d|0,-1,-1)|0;d=yy(o|0,u|0,r|0,L|0)|0;r=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(q){e=0;d=0;c=0;t=0;s=0;m=0;v=0;while(1){if(+p[i+(d<<3)>>3]!=j){if(!((m|0)==0&(e|0)==0)?(w=jy(d|0,s|0,m|0,e|0)|0,x=ky(w|0,L|0,r|0,f|0)|0,w=L,y=ky(x|0,w|0,-1,-1)|0,z=wy(y|0,L|0,o|0,u|0)|0,y=L,A=ky(z|0,y|0,1,0)|0,B=L,C=yy(z|0,y|0,o|0,u|0)|0,y=jy(x|0,w|0,C|0,L|0)|0,(Lr(a,b,A,B,y,L,m,e,l)|0)>0):0){D=30;break a}y=ky(c|0,t|0,1,0)|0;E=y;F=L;G=0;H=0;I=v}else{if(!((c|0)==0&(t|0)==0)?(y=jy(d|0,s|0,c|0,t|0)|0,B=ky(y|0,L|0,r|0,f|0)|0,A=L,C=ky(B|0,A|0,-1,-1)|0,w=wy(C|0,L|0,o|0,u|0)|0,C=L,x=yy(w|0,C|0,o|0,u|0)|0,z=jy(B|0,A|0,x|0,L|0)|0,x=L,A=ky(w|0,C|0,1,0)|0,(Wq(a,b,A,L,z,x,c,t,i+(y<<3)|0,l)|0)>0):0){y=k[l>>2]|0;if((y|0)!=412){n=y;break}k[l>>2]=0;J=1}else J=v;y=ky(m|0,e|0,1,0)|0;E=0;F=0;G=y;H=L;I=J}y=ky(d|0,s|0,1,0)|0;x=L;if((x|0)<(h|0)|(x|0)==(h|0)&y>>>0<g>>>0){e=H;d=y;c=E;t=F;s=x;m=G;v=I}else{K=E;M=F;N=y;O=x;P=G;Q=H;R=I;break a}}return n|0}else{v=0;m=0;s=0;t=0;c=0;d=0;while(1){if(+p[i+(v<<3)>>3]!=j){if(!((m|0)==0&(s|0)==0)?(e=jy(v|0,t|0,m|0,s|0)|0,x=ky(e|0,L|0,r|0,f|0)|0,e=L,y=ky(x|0,e|0,-1,-1)|0,z=wy(y|0,L|0,o|0,u|0)|0,y=L,A=ky(z|0,y|0,1,0)|0,C=L,w=yy(z|0,y|0,o|0,u|0)|0,y=jy(x|0,e|0,w|0,L|0)|0,(Lr(a,b,A,C,y,L,m,s,l)|0)>0):0){D=30;break a}y=ky(c|0,d|0,1,0)|0;S=y;T=L;U=0;V=0}else{y=ky(m|0,s|0,1,0)|0;S=0;T=0;U=y;V=L}y=ky(v|0,t|0,1,0)|0;C=L;if((C|0)<(h|0)|(C|0)==(h|0)&y>>>0<g>>>0){v=y;m=U;s=V;t=C;c=S;d=T}else{K=S;M=T;N=y;O=C;P=U;Q=V;R=0;break}}}while(0);if((D|0)==30){n=k[l>>2]|0;return n|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){D=jy(N|0,O|0,P|0,Q|0)|0;V=ky(D|0,L|0,r|0,f|0)|0;D=L;U=ky(V|0,D|0,-1,-1)|0;T=wy(U|0,L|0,o|0,u|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,o|0,u|0)|0;U=jy(V|0,D|0,h|0,L|0)|0;Lr(a,b,S,g,U,L,P,Q,l)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,r|0,f|0)|0;f=L;r=ky(O|0,f|0,-1,-1)|0;N=wy(r|0,L|0,o|0,u|0)|0;r=L;if(q){q=yy(N|0,r|0,o|0,u|0)|0;u=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,r|0,1,0)|0;Wq(a,b,f,L,u,q,K,M,i+(Q<<3)|0,l)|0}}Q=k[l>>2]|0;if(!((R|0)!=0&(Q|0)<1)){n=Q;return n|0}k[l>>2]=412;n=412;return n|0}function Zq(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){o=g;q=0;do{p[m>>3]=+p[a+(q<<3)>>3];Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}}else if(n){n=g;q=0;do{p[l>>3]=(+p[a+(q<<3)>>3]-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}b=rx(g,44)|0;if(!b){s=k[h>>2]|0;r=j;return s|0}else t=b;do{i[t>>0]=46;t=rx(t,44)|0}while((t|0)!=0);s=k[h>>2]|0;r=j;return s|0}function _q(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;$q(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,42,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function $q(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}f=k[u>>2]|0;u=((f|0)<0)<<31>>31;if((k[t>>2]|0)==16)Rk(E,F);e=F+1|0;d=g;g=h;h=0;c=0;I=0;K=0;a:while(1){M=h;N=c;O=d;P=g;while(1){Q=l+(M<<2)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=104;break a}U=(S|0)<(u|0)|(S|0)==(u|0)&R>>>0<f>>>0;X=U?R:f;U=my(0,X|0,32)|0;X=L;Y=y;Z=A;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=jy(k[Y>>2]|0,k[Y+4>>2]|0,_|0,$|0)|0;Y=L;aa=(X|0)<(Y|0)|(X|0)==(Y|0)&U>>>0<Z>>>0;Y=aa?U:Z;Z=z;U=k[Z>>2]|0;aa=k[Z+4>>2]|0;Z=B;X=yy(k[Z>>2]|0,k[Z+4>>2]|0,I|0,K|0)|0;Z=ky(X|0,L|0,U|0,aa|0)|0;aa=L;U=k[x>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,_|0,$|0)|0;$=ky(Z|0,aa|0,X|0,L|0)|0;au(a,$,L,1,m)|0;b:do switch(k[t>>2]|0){case 42:{ba=+p[C>>3];ca=+p[D>>3];if(!(ba==1.0&ca==0.0)){if((Y|0)>0){$=0;do{o[H+($<<2)>>2]=(+o[l+($+M<<2)>>2]-ca)/ba;$=$+1|0}while(($|0)!=(Y|0))}}else py(H|0,Q|0,Y<<2|0)|0;wu(a,Y,k[x>>2]|0,H,m)|0;break}case 81:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{da=+o[l+(X+M<<2)>>2];do if(!(da<-9223372036854775808.0))if(da>9223372036854775808.0){k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}else{aa=~~da;Z=H+(X<<3)|0;k[Z>>2]=aa;k[Z+4>>2]=((aa|0)<0)<<31>>31;break}else{k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+o[l+(X+M<<2)>>2]-ca)/ba;do if(!(da<-9223372036854775808.0)){if(da>2147483647.49){k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}if(!(da>=0.0)){ea=da+-.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}else{ea=da+.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}}else{k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}vu(a,Y,k[x>>2]|0,H,m)|0;break}case 11:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{da=+o[l+($+M<<2)>>2];ea=da;do if(!(ea<-.49))if(ea>255.49){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=~~da;break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+o[l+($+M<<2)>>2]-ca)/ba;do if(!(da<-.49))if(da>255.49){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=~~(da+.5);break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}su(a,Y,k[x>>2]|0,H,m)|0;break}case 21:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{da=+o[l+(X+M<<2)>>2];ea=da;do if(!(ea<-32768.49))if(ea>32767.49){k[m>>2]=-11;j[H+(X<<1)>>1]=32767;break}else{j[H+(X<<1)>>1]=~~da;break}else{k[m>>2]=-11;j[H+(X<<1)>>1]=-32768}while(0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+o[l+(X+M<<2)>>2]-ca)/ba;do if(!(da<-32768.49)){if(da>32767.49){k[m>>2]=-11;j[H+(X<<1)>>1]=32767;break}if(!(da>=0.0)){j[H+(X<<1)>>1]=~~(da+-.5);break}else{j[H+(X<<1)>>1]=~~(da+.5);break}}else{k[m>>2]=-11;j[H+(X<<1)>>1]=-32768}while(0);X=X+1|0}while((X|0)!=(Y|0))}tu(a,Y,k[x>>2]|0,H,m)|0;break}case 41:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{da=+o[l+($+M<<2)>>2];ea=da;do if(!(ea<-2147483648.49))if(ea>2147483647.49){k[m>>2]=-11;k[H+($<<2)>>2]=2147483647;break}else{k[H+($<<2)>>2]=~~da;break}else{k[m>>2]=-11;k[H+($<<2)>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+o[l+($+M<<2)>>2]-ca)/ba;do if(!(da<-2147483648.49)){if(da>2147483647.49){k[m>>2]=-11;k[H+($<<2)>>2]=2147483647;break}if(!(da>=0.0)){k[H+($<<2)>>2]=~~(da+-.5);break}else{k[H+($<<2)>>2]=~~(da+.5);break}}else{k[m>>2]=-11;k[H+($<<2)>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(Y|0))}uu(a,Y,k[x>>2]|0,H,m)|0;break}case 82:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{p[H+(X<<3)>>3]=+o[l+(X+M<<2)>>2];X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{p[H+(X<<3)>>3]=(+o[l+(X+M<<2)>>2]-ca)/ba;X=X+1|0}while((X|0)!=(Y|0))}xu(a,Y,k[x>>2]|0,H,m)|0;break}case 16:{if((i[e>>0]|0)==115){T=96;break a}cr(Q,Y,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;X=k[x>>2]|0;$=k[w>>2]|0;if((X|0)==($|0)){aa=ha(X,Y)|0;cu(a,aa,((aa|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,$,Y,X-$|0,H,m)|0;break b}break}default:{T=96;break a}}while(0);if((k[m>>2]|0)>0){fa=M;ja=N;ka=Y;T=100;break a}$=((Y|0)<0)<<31>>31;X=jy(R|0,S|0,Y|0,$|0)|0;aa=L;if((R|0)==(Y|0)&(S|0)==($|0)){R=X;S=aa}else{la=Y;ma=$;na=X;oa=aa;break}}S=ky(la|0,ma|0,M|0,N|0)|0;R=L;Q=A;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,la|0,ma|0)|0;Q=L;X=A;k[X>>2]=aa;k[X+4>>2]=Q;X=y;if((aa|0)==(k[X>>2]|0)?(Q|0)==(k[X+4>>2]|0):0){pa=S;qa=R;ra=na;sa=oa;break}else{M=S;N=R;O=na;P=oa}}P=A;k[P>>2]=0;k[P+4>>2]=0;P=ky(I|0,K|0,1,0)|0;d=ra;g=sa;h=pa;c=qa;I=P;K=L}if((T|0)==96){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((T|0)==100){v=ky(fa|0,ja|0,1,0)|0;ba=+(v>>>0)+4294967296.0*+(L|0);v=ky(ka|0,((ka|0)<0)<<31>>31|0,fa|0,ja|0)|0;p[q>>3]=ba;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,770816,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((T|0)==104){T=k[m>>2]|0;if((T|0)!=-11){J=T;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function ar(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;br(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{o[l>>2]=h;jt(a,42,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function br(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=+j;l=l|0;var m=0,n=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((m|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[m+(p*160|0)+80>>2]|0;q=(r|0)>0;if(q){s=m+(p*160|0)+88|0;t=k[s>>2]|0;u=k[s+4>>2]|0}else{s=ky(e|0,f|0,-1,-1)|0;p=ky(s|0,L|0,g|0,h|0)|0;t=p;u=L}p=ty(t|0,u|0,(((r|0)>-1?r:0-r|0)|0)>82|0)|0;u=L;do if((r|0)<0?($q(a,b,c,d,e,f,g,h,i,l)|0)>0:0){t=k[l>>2]|0;if((t|0)==412){k[l>>2]=0;break}else{n=t;return n|0}}while(0);r=ky(c|0,d|0,-1,-1)|0;d=yy(p|0,u|0,r|0,L|0)|0;r=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(q){e=0;d=0;c=0;t=0;s=0;m=0;v=0;while(1){if(+o[i+(d<<2)>>2]!=j){if(!((m|0)==0&(e|0)==0)?(w=jy(d|0,s|0,m|0,e|0)|0,x=ky(w|0,L|0,r|0,f|0)|0,w=L,y=ky(x|0,w|0,-1,-1)|0,z=wy(y|0,L|0,p|0,u|0)|0,y=L,A=ky(z|0,y|0,1,0)|0,B=L,C=yy(z|0,y|0,p|0,u|0)|0,y=jy(x|0,w|0,C|0,L|0)|0,(Lr(a,b,A,B,y,L,m,e,l)|0)>0):0){D=30;break a}y=ky(c|0,t|0,1,0)|0;E=y;F=L;G=0;H=0;I=v}else{if(!((c|0)==0&(t|0)==0)?(y=jy(d|0,s|0,c|0,t|0)|0,B=ky(y|0,L|0,r|0,f|0)|0,A=L,C=ky(B|0,A|0,-1,-1)|0,w=wy(C|0,L|0,p|0,u|0)|0,C=L,x=yy(w|0,C|0,p|0,u|0)|0,z=jy(B|0,A|0,x|0,L|0)|0,x=L,A=ky(w|0,C|0,1,0)|0,($q(a,b,A,L,z,x,c,t,i+(y<<2)|0,l)|0)>0):0){y=k[l>>2]|0;if((y|0)!=412){n=y;break}k[l>>2]=0;J=1}else J=v;y=ky(m|0,e|0,1,0)|0;E=0;F=0;G=y;H=L;I=J}y=ky(d|0,s|0,1,0)|0;x=L;if((x|0)<(h|0)|(x|0)==(h|0)&y>>>0<g>>>0){e=H;d=y;c=E;t=F;s=x;m=G;v=I}else{K=E;M=F;N=y;O=x;P=G;Q=H;R=I;break a}}return n|0}else{v=0;m=0;s=0;t=0;c=0;d=0;while(1){if(+o[i+(v<<2)>>2]!=j){if(!((m|0)==0&(s|0)==0)?(e=jy(v|0,t|0,m|0,s|0)|0,x=ky(e|0,L|0,r|0,f|0)|0,e=L,y=ky(x|0,e|0,-1,-1)|0,z=wy(y|0,L|0,p|0,u|0)|0,y=L,A=ky(z|0,y|0,1,0)|0,C=L,w=yy(z|0,y|0,p|0,u|0)|0,y=jy(x|0,e|0,w|0,L|0)|0,(Lr(a,b,A,C,y,L,m,s,l)|0)>0):0){D=30;break a}y=ky(c|0,d|0,1,0)|0;S=y;T=L;U=0;V=0}else{y=ky(m|0,s|0,1,0)|0;S=0;T=0;U=y;V=L}y=ky(v|0,t|0,1,0)|0;C=L;if((C|0)<(h|0)|(C|0)==(h|0)&y>>>0<g>>>0){v=y;m=U;s=V;t=C;c=S;d=T}else{K=S;M=T;N=y;O=C;P=U;Q=V;R=0;break}}}while(0);if((D|0)==30){n=k[l>>2]|0;return n|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){D=jy(N|0,O|0,P|0,Q|0)|0;V=ky(D|0,L|0,r|0,f|0)|0;D=L;U=ky(V|0,D|0,-1,-1)|0;T=wy(U|0,L|0,p|0,u|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,p|0,u|0)|0;U=jy(V|0,D|0,h|0,L|0)|0;Lr(a,b,S,g,U,L,P,Q,l)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,r|0,f|0)|0;f=L;r=ky(O|0,f|0,-1,-1)|0;N=wy(r|0,L|0,p|0,u|0)|0;r=L;if(q){q=yy(N|0,r|0,p|0,u|0)|0;u=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,r|0,1,0)|0;$q(a,b,f,L,u,q,K,M,i+(Q<<2)|0,l)|0}}Q=k[l>>2]|0;if(!((R|0)!=0&(Q|0)<1)){n=Q;return n|0}k[l>>2]=412;n=412;return n|0}function cr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){q=g;s=0;do{p[m>>3]=+o[a+(s<<2)>>2];Xw(q,e,m)|0;q=q+f|0;if(i[q>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}}else if(n){n=g;s=0;do{p[l>>3]=(+o[a+(s<<2)>>2]-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}b=rx(g,44)|0;if(!b){t=k[h>>2]|0;r=j;return t|0}else u=b;do{i[u>>0]=46;u=rx(u,44)|0}while((u|0)!=0);t=k[h>>2]|0;r=j;return t|0}function dr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;er(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,21,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function er(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}f=k[u>>2]|0;u=((f|0)<0)<<31>>31;if((k[t>>2]|0)==16)Rk(E,F);e=F+1|0;d=g;g=h;h=0;c=0;I=0;K=0;a:while(1){M=h;N=c;O=d;P=g;while(1){Q=l+(M<<1)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=89;break a}U=(S|0)<(u|0)|(S|0)==(u|0)&R>>>0<f>>>0;X=U?R:f;U=my(0,X|0,32)|0;X=L;Y=y;Z=A;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=jy(k[Y>>2]|0,k[Y+4>>2]|0,_|0,$|0)|0;Y=L;aa=(X|0)<(Y|0)|(X|0)==(Y|0)&U>>>0<Z>>>0;Y=aa?U:Z;Z=z;U=k[Z>>2]|0;aa=k[Z+4>>2]|0;Z=B;X=yy(k[Z>>2]|0,k[Z+4>>2]|0,I|0,K|0)|0;Z=ky(X|0,L|0,U|0,aa|0)|0;aa=L;U=k[x>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,_|0,$|0)|0;$=ky(Z|0,aa|0,X|0,L|0)|0;au(a,$,L,1,m)|0;b:do switch(k[t>>2]|0){case 21:{ba=+p[C>>3];ca=+p[D>>3];if(!(ba==1.0&ca==0.0)){if((Y|0)>0){$=0;do{da=(+(j[l+($+M<<1)>>1]|0)-ca)/ba;do if(!(da<-32768.49)){if(da>32767.49){k[m>>2]=-11;j[H+($<<1)>>1]=32767;break}if(!(da>=0.0)){j[H+($<<1)>>1]=~~(da+-.5);break}else{j[H+($<<1)>>1]=~~(da+.5);break}}else{k[m>>2]=-11;j[H+($<<1)>>1]=-32768}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else py(H|0,Q|0,Y<<1|0)|0;tu(a,Y,k[x>>2]|0,H,m)|0;break}case 81:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{aa=j[l+(X+M<<1)>>1]|0;Z=H+(X<<3)|0;k[Z>>2]=aa;k[Z+4>>2]=((aa|0)<0)<<31>>31;X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+(j[l+(X+M<<1)>>1]|0)-ca)/ba;do if(!(da<-9223372036854775808.0)){if(da>9223372036854775808.0){k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}if(!(da>=0.0)){ea=da+-.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}else{ea=da+.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=H+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}}else{k[m>>2]=-11;aa=H+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}vu(a,Y,k[x>>2]|0,H,m)|0;break}case 11:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{aa=j[l+($+M<<1)>>1]|0;do if(aa<<16>>16>=0)if(aa<<16>>16>255){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=aa;break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+(j[l+($+M<<1)>>1]|0)-ca)/ba;do if(!(da<-.49))if(da>255.49){k[m>>2]=-11;i[H+$>>0]=-1;break}else{i[H+$>>0]=~~(da+.5);break}else{k[m>>2]=-11;i[H+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}su(a,Y,k[x>>2]|0,H,m)|0;break}case 41:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{k[H+(X<<2)>>2]=j[l+(X+M<<1)>>1];X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+(j[l+(X+M<<1)>>1]|0)-ca)/ba;do if(!(da<-2147483648.49)){if(da>2147483647.49){k[m>>2]=-11;k[H+(X<<2)>>2]=2147483647;break}if(!(da>=0.0)){k[H+(X<<2)>>2]=~~(da+-.5);break}else{k[H+(X<<2)>>2]=~~(da+.5);break}}else{k[m>>2]=-11;k[H+(X<<2)>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}uu(a,Y,k[x>>2]|0,H,m)|0;break}case 42:{ba=+p[C>>3];ca=+p[D>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{o[H+($<<2)>>2]=+(j[l+($+M<<1)>>1]|0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{o[H+($<<2)>>2]=(+(j[l+($+M<<1)>>1]|0)-ca)/ba;$=$+1|0}while(($|0)!=(Y|0))}wu(a,Y,k[x>>2]|0,H,m)|0;break}case 82:{ba=+p[C>>3];ca=+p[D>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{p[H+(X<<3)>>3]=+(j[l+(X+M<<1)>>1]|0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{p[H+(X<<3)>>3]=(+(j[l+(X+M<<1)>>1]|0)-ca)/ba;X=X+1|0}while((X|0)!=(Y|0))}xu(a,Y,k[x>>2]|0,H,m)|0;break}case 16:{if((i[e>>0]|0)==115){T=81;break a}hr(Q,Y,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;X=k[x>>2]|0;$=k[w>>2]|0;if((X|0)==($|0)){aa=ha(X,Y)|0;cu(a,aa,((aa|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,$,Y,X-$|0,H,m)|0;break b}break}default:{T=81;break a}}while(0);if((k[m>>2]|0)>0){fa=M;ja=N;ka=Y;T=85;break a}$=((Y|0)<0)<<31>>31;X=jy(R|0,S|0,Y|0,$|0)|0;aa=L;if((R|0)==(Y|0)&(S|0)==($|0)){R=X;S=aa}else{la=Y;ma=$;na=X;oa=aa;break}}S=ky(la|0,ma|0,M|0,N|0)|0;R=L;Q=A;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,la|0,ma|0)|0;Q=L;X=A;k[X>>2]=aa;k[X+4>>2]=Q;X=y;if((aa|0)==(k[X>>2]|0)?(Q|0)==(k[X+4>>2]|0):0){pa=S;qa=R;ra=na;sa=oa;break}else{M=S;N=R;O=na;P=oa}}P=A;k[P>>2]=0;k[P+4>>2]=0;P=ky(I|0,K|0,1,0)|0;d=ra;g=sa;h=pa;c=qa;I=P;K=L}if((T|0)==81){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((T|0)==85){v=ky(fa|0,ja|0,1,0)|0;ba=+(v>>>0)+4294967296.0*+(L|0);v=ky(ka|0,((ka|0)<0)<<31>>31|0,fa|0,ja|0)|0;p[q>>3]=ba;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,770884,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((T|0)==89){T=k[m>>2]|0;if((T|0)!=-11){J=T;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function fr(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,i)|0)){n=(b|0)<1;gr(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;o=k[i>>2]|0;r=l;return o|0}else{j[m>>1]=h;jt(a,21,c,d,e,f,1,g,m,i)|0;o=k[i>>2]|0;r=l;return o|0}return 0}function gr(a,b,c,d,e,f,g,h,i,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[n+(p*160|0)+80>>2]|0;q=(r|0)>0;do if(!q){s=ky(e|0,f|0,-1,-1)|0;t=ky(s|0,L|0,g|0,h|0)|0;s=L;if((r|0)<0?(er(a,b,c,d,e,f,g,h,i,m)|0)>0:0){u=k[m>>2]|0;if((u|0)==412){k[m>>2]=0;v=t;w=s;break}else{o=u;return o|0}}else{v=t;w=s}}else{s=n+(p*160|0)+88|0;v=k[s>>2]|0;w=k[s+4>>2]|0}while(0);p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,w|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){o=k[m>>2]|0;return o|0}a:do if(q){e=0;d=0;c=0;n=0;r=0;s=0;t=0;while(1){if((j[i+(e<<1)>>1]|0)==l<<16>>16){if(!((r|0)==0&(s|0)==0)?(u=jy(e|0,n|0,r|0,s|0)|0,x=ky(u|0,L|0,p|0,f|0)|0,y=L,z=ky(x|0,y|0,-1,-1)|0,A=wy(z|0,L|0,v|0,w|0)|0,z=L,B=yy(A|0,z|0,v|0,w|0)|0,C=jy(x|0,y|0,B|0,L|0)|0,B=L,y=ky(A|0,z|0,1,0)|0,(er(a,b,y,L,C,B,r,s,i+(u<<1)|0,m)|0)>0):0){u=k[m>>2]|0;if((u|0)!=412){o=u;break}k[m>>2]=0;D=1}else D=t;u=ky(d|0,c|0,1,0)|0;E=0;F=0;G=u;H=L;I=D}else{if(!((d|0)==0&(c|0)==0)?(u=jy(e|0,n|0,d|0,c|0)|0,B=ky(u|0,L|0,p|0,f|0)|0,u=L,C=ky(B|0,u|0,-1,-1)|0,y=wy(C|0,L|0,v|0,w|0)|0,C=L,z=ky(y|0,C|0,1,0)|0,A=L,x=yy(y|0,C|0,v|0,w|0)|0,C=jy(B|0,u|0,x|0,L|0)|0,(Kr(a,b,z,A,C,L,d,c,m)|0)>0):0){J=29;break a}C=ky(r|0,s|0,1,0)|0;E=C;F=L;G=0;H=0;I=t}C=ky(e|0,n|0,1,0)|0;A=L;if((A|0)<(h|0)|(A|0)==(h|0)&C>>>0<g>>>0){e=C;d=G;c=H;n=A;r=E;s=F;t=I}else{K=E;M=F;N=C;O=A;P=G;Q=H;R=I;break a}}return o|0}else{t=0;s=0;r=0;n=0;c=0;d=0;while(1){if((j[i+(t<<1)>>1]|0)==l<<16>>16){e=ky(s|0,r|0,1,0)|0;S=0;T=0;U=e;V=L}else{if(!((s|0)==0&(r|0)==0)?(e=jy(t|0,n|0,s|0,r|0)|0,A=ky(e|0,L|0,p|0,f|0)|0,e=L,C=ky(A|0,e|0,-1,-1)|0,z=wy(C|0,L|0,v|0,w|0)|0,C=L,x=ky(z|0,C|0,1,0)|0,u=L,B=yy(z|0,C|0,v|0,w|0)|0,C=jy(A|0,e|0,B|0,L|0)|0,(Kr(a,b,x,u,C,L,s,r,m)|0)>0):0){J=29;break a}C=ky(c|0,d|0,1,0)|0;S=C;T=L;U=0;V=0}C=ky(t|0,n|0,1,0)|0;u=L;if((u|0)<(h|0)|(u|0)==(h|0)&C>>>0<g>>>0){t=C;s=U;r=V;n=u;c=S;d=T}else{K=S;M=T;N=C;O=u;P=U;Q=V;R=0;break}}}while(0);if((J|0)==29){o=k[m>>2]|0;return o|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){J=jy(N|0,O|0,P|0,Q|0)|0;V=ky(J|0,L|0,p|0,f|0)|0;J=L;U=ky(V|0,J|0,-1,-1)|0;T=wy(U|0,L|0,v|0,w|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,v|0,w|0)|0;U=jy(V|0,J|0,h|0,L|0)|0;Kr(a,b,S,g,U,L,P,Q,m)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,p|0,f|0)|0;f=L;p=ky(O|0,f|0,-1,-1)|0;N=wy(p|0,L|0,v|0,w|0)|0;p=L;if(q){q=yy(N|0,p|0,v|0,w|0)|0;w=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,p|0,1,0)|0;er(a,b,f,L,w,q,K,M,i+(Q<<1)|0,m)|0}}Q=k[m>>2]|0;if(!((R|0)!=0&(Q|0)<1)){o=Q;return o|0}k[m>>2]=412;o=412;return o|0}function hr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0;l=r;r=r+16|0;m=l+8|0;n=l;o=(b|0)>0;if(c==1.0&d==0.0){if(o){q=g;s=0;do{p[n>>3]=+(j[a+(s<<1)>>1]|0);Xw(q,e,n)|0;q=q+f|0;if(i[q>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}}else if(o){o=g;s=0;do{p[m>>3]=(+(j[a+(s<<1)>>1]|0)-d)/c;Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}b=rx(g,44)|0;if(!b){t=k[h>>2]|0;r=l;return t|0}else u=b;do{i[u>>0]=46;u=rx(u,44)|0}while((u|0)!=0);t=k[h>>2]|0;r=l;return t|0}function ir(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;jr(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,41,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function jr(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0.0,ea=0.0,fa=0,ja=0,ka=0.0,la=0.0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}f=k[u>>2]|0;u=((f|0)<0)<<31>>31;if((k[t>>2]|0)==16)Rk(E,F);e=F+1|0;d=g;g=h;h=0;c=0;I=0;K=0;a:while(1){M=h;N=c;O=d;P=g;while(1){Q=l+(M<<2)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=95;break a}U=(S|0)<(u|0)|(S|0)==(u|0)&R>>>0<f>>>0;X=U?R:f;U=my(0,X|0,32)|0;X=L;Y=y;Z=k[Y>>2]|0;_=k[Y+4>>2]|0;Y=A;$=k[Y>>2]|0;aa=k[Y+4>>2]|0;Y=jy(Z|0,_|0,$|0,aa|0)|0;ba=L;ca=(X|0)<(ba|0)|(X|0)==(ba|0)&U>>>0<Y>>>0;ba=ca?U:Y;Y=z;U=k[Y>>2]|0;ca=k[Y+4>>2]|0;Y=B;X=yy(k[Y>>2]|0,k[Y+4>>2]|0,I|0,K|0)|0;Y=ky(X|0,L|0,U|0,ca|0)|0;ca=L;U=k[x>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,$|0,aa|0)|0;U=ky(Y|0,ca|0,X|0,L|0)|0;au(a,U,L,1,m)|0;b:do switch(k[t>>2]|0){case 41:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){X=ky($|0,aa|0,-1,-1)|0;ca=jy(X|0,L|0,Z|0,_|0)|0;X=L;Y=(u|0)>(S|0)|(u|0)==(S|0)&f>>>0>R>>>0;fa=Y?R:f;Y=my(0,fa|0,32)|0;fa=~Y;Y=~L;ja=(X|0)>(Y|0)|(X|0)==(Y|0)&ca>>>0>fa>>>0;Y=ja?ca:fa;py(H|0,Q|0,(Y^1073741823)<<2|0)|0}}else if(U){U=0;do{ka=(+(k[l+(U+M<<2)>>2]|0)-ea)/da;do if(!(ka<-2147483648.49)){if(ka>2147483647.49){k[m>>2]=-11;k[H+(U<<2)>>2]=2147483647;break}if(!(ka>=0.0)){k[H+(U<<2)>>2]=~~(ka+-.5);break}else{k[H+(U<<2)>>2]=~~(ka+.5);break}}else{k[m>>2]=-11;k[H+(U<<2)>>2]=-2147483648}while(0);U=U+1|0}while((U|0)!=(ba|0))}uu(a,ba,k[x>>2]|0,H,m)|0;break}case 81:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{fa=k[l+(Y+M<<2)>>2]|0;ca=H+(Y<<3)|0;k[ca>>2]=fa;k[ca+4>>2]=((fa|0)<0)<<31>>31;Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{ka=(+(k[l+(Y+M<<2)>>2]|0)-ea)/da;do if(!(ka<-9223372036854775808.0)){if(ka>9223372036854775808.0){k[m>>2]=-11;fa=H+(Y<<3)|0;k[fa>>2]=-1;k[fa+4>>2]=2147483647;break}if(!(ka>=0.0)){la=ka+-.5;fa=+W(la)>=1.0?(la>0.0?~~+ia(+V(la/4294967296.0),4294967295.0)>>>0:~~+ga((la-+(~~la>>>0))/4294967296.0)>>>0):0;ca=H+(Y<<3)|0;k[ca>>2]=~~la>>>0;k[ca+4>>2]=fa;break}else{la=ka+.5;fa=+W(la)>=1.0?(la>0.0?~~+ia(+V(la/4294967296.0),4294967295.0)>>>0:~~+ga((la-+(~~la>>>0))/4294967296.0)>>>0):0;ca=H+(Y<<3)|0;k[ca>>2]=~~la>>>0;k[ca+4>>2]=fa;break}}else{k[m>>2]=-11;fa=H+(Y<<3)|0;k[fa>>2]=0;k[fa+4>>2]=-2147483648}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}vu(a,ba,k[x>>2]|0,H,m)|0;break}case 11:{da=+p[C>>3];ea=+p[D>>3];Y=(ba|0)>0;if(da==1.0&ea==0.0){if(Y){U=0;do{fa=k[l+(U+M<<2)>>2]|0;do if((fa|0)>=0)if((fa|0)>255){k[m>>2]=-11;i[H+U>>0]=-1;break}else{i[H+U>>0]=fa;break}else{k[m>>2]=-11;i[H+U>>0]=0}while(0);U=U+1|0}while((U|0)!=(ba|0))}}else if(Y){U=0;do{ka=(+(k[l+(U+M<<2)>>2]|0)-ea)/da;do if(!(ka<-.49))if(ka>255.49){k[m>>2]=-11;i[H+U>>0]=-1;break}else{i[H+U>>0]=~~(ka+.5);break}else{k[m>>2]=-11;i[H+U>>0]=0}while(0);U=U+1|0}while((U|0)!=(ba|0))}su(a,ba,k[x>>2]|0,H,m)|0;break}case 21:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{fa=k[l+(Y+M<<2)>>2]|0;do if((fa|0)>=-32768)if((fa|0)>32767){k[m>>2]=-11;j[H+(Y<<1)>>1]=32767;break}else{j[H+(Y<<1)>>1]=fa;break}else{k[m>>2]=-11;j[H+(Y<<1)>>1]=-32768}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{ka=(+(k[l+(Y+M<<2)>>2]|0)-ea)/da;do if(!(ka<-32768.49)){if(ka>32767.49){k[m>>2]=-11;j[H+(Y<<1)>>1]=32767;break}if(!(ka>=0.0)){j[H+(Y<<1)>>1]=~~(ka+-.5);break}else{j[H+(Y<<1)>>1]=~~(ka+.5);break}}else{k[m>>2]=-11;j[H+(Y<<1)>>1]=-32768}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}tu(a,ba,k[x>>2]|0,H,m)|0;break}case 42:{da=+p[C>>3];ea=+p[D>>3];Y=(ba|0)>0;if(da==1.0&ea==0.0){if(Y){U=0;do{o[H+(U<<2)>>2]=+(k[l+(U+M<<2)>>2]|0);U=U+1|0}while((U|0)!=(ba|0))}}else if(Y){U=0;do{o[H+(U<<2)>>2]=(+(k[l+(U+M<<2)>>2]|0)-ea)/da;U=U+1|0}while((U|0)!=(ba|0))}wu(a,ba,k[x>>2]|0,H,m)|0;break}case 82:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{p[H+(Y<<3)>>3]=+(k[l+(Y+M<<2)>>2]|0);Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{p[H+(Y<<3)>>3]=(+(k[l+(Y+M<<2)>>2]|0)-ea)/da;Y=Y+1|0}while((Y|0)!=(ba|0))}xu(a,ba,k[x>>2]|0,H,m)|0;break}case 16:{if((i[e>>0]|0)==115){T=87;break a}mr(Q,ba,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;Y=k[x>>2]|0;U=k[w>>2]|0;if((Y|0)==(U|0)){fa=ha(Y,ba)|0;cu(a,fa,((fa|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,U,ba,Y-U|0,H,m)|0;break b}break}default:{T=87;break a}}while(0);if((k[m>>2]|0)>0){ma=M;na=N;oa=ba;T=91;break a}_=((ba|0)<0)<<31>>31;Z=jy(R|0,S|0,ba|0,_|0)|0;aa=L;if((R|0)==(ba|0)&(S|0)==(_|0)){R=Z;S=aa}else{pa=ba;qa=_;ra=Z;sa=aa;break}}S=ky(pa|0,qa|0,M|0,N|0)|0;R=L;Q=A;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,pa|0,qa|0)|0;Q=L;Z=A;k[Z>>2]=aa;k[Z+4>>2]=Q;Z=y;if((aa|0)==(k[Z>>2]|0)?(Q|0)==(k[Z+4>>2]|0):0){ta=S;ua=R;va=ra;wa=sa;break}else{M=S;N=R;O=ra;P=sa}}P=A;k[P>>2]=0;k[P+4>>2]=0;P=ky(I|0,K|0,1,0)|0;d=va;g=wa;h=ta;c=ua;I=P;K=L}if((T|0)==87){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((T|0)==91){v=ky(ma|0,na|0,1,0)|0;da=+(v>>>0)+4294967296.0*+(L|0);v=ky(oa|0,((oa|0)<0)<<31>>31|0,ma|0,na|0)|0;p[q>>3]=da;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,770952,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((T|0)==95){T=k[m>>2]|0;if((T|0)!=-11){J=T;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function kr(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;lr(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{k[l>>2]=h;jt(a,41,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function lr(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;o=a+4|0;p=k[o>>2]|0;if((m|0)==(k[p+68>>2]|0)){q=p+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[o>>2]|0)+960>>2]|0;o=b+-1|0;q=k[m+(o*160|0)+80>>2]|0;p=(q|0)>0;do if(!p){r=ky(e|0,f|0,-1,-1)|0;s=ky(r|0,L|0,g|0,h|0)|0;r=L;if((q|0)<0?(jr(a,b,c,d,e,f,g,h,i,l)|0)>0:0){t=k[l>>2]|0;if((t|0)==412){k[l>>2]=0;u=s;v=r;break}else{n=t;return n|0}}else{u=s;v=r}}else{r=m+(o*160|0)+88|0;u=k[r>>2]|0;v=k[r+4>>2]|0}while(0);o=ky(c|0,d|0,-1,-1)|0;d=yy(u|0,v|0,o|0,L|0)|0;o=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(p){e=0;d=0;c=0;m=0;q=0;r=0;s=0;while(1){if((k[i+(e<<2)>>2]|0)==(j|0)){if(!((q|0)==0&(r|0)==0)?(t=jy(e|0,m|0,q|0,r|0)|0,w=ky(t|0,L|0,o|0,f|0)|0,x=L,y=ky(w|0,x|0,-1,-1)|0,z=wy(y|0,L|0,u|0,v|0)|0,y=L,A=yy(z|0,y|0,u|0,v|0)|0,B=jy(w|0,x|0,A|0,L|0)|0,A=L,x=ky(z|0,y|0,1,0)|0,(jr(a,b,x,L,B,A,q,r,i+(t<<2)|0,l)|0)>0):0){t=k[l>>2]|0;if((t|0)!=412){n=t;break}k[l>>2]=0;C=1}else C=s;t=ky(d|0,c|0,1,0)|0;D=0;E=0;F=t;G=L;H=C}else{if(!((d|0)==0&(c|0)==0)?(t=jy(e|0,m|0,d|0,c|0)|0,A=ky(t|0,L|0,o|0,f|0)|0,t=L,B=ky(A|0,t|0,-1,-1)|0,x=wy(B|0,L|0,u|0,v|0)|0,B=L,y=ky(x|0,B|0,1,0)|0,z=L,w=yy(x|0,B|0,u|0,v|0)|0,B=jy(A|0,t|0,w|0,L|0)|0,(Kr(a,b,y,z,B,L,d,c,l)|0)>0):0){I=29;break a}B=ky(q|0,r|0,1,0)|0;D=B;E=L;F=0;G=0;H=s}B=ky(e|0,m|0,1,0)|0;z=L;if((z|0)<(h|0)|(z|0)==(h|0)&B>>>0<g>>>0){e=B;d=F;c=G;m=z;q=D;r=E;s=H}else{J=D;K=E;M=B;N=z;O=F;P=G;Q=H;break a}}return n|0}else{s=0;r=0;q=0;m=0;c=0;d=0;while(1){if((k[i+(s<<2)>>2]|0)==(j|0)){e=ky(r|0,q|0,1,0)|0;R=0;S=0;T=e;U=L}else{if(!((r|0)==0&(q|0)==0)?(e=jy(s|0,m|0,r|0,q|0)|0,z=ky(e|0,L|0,o|0,f|0)|0,e=L,B=ky(z|0,e|0,-1,-1)|0,y=wy(B|0,L|0,u|0,v|0)|0,B=L,w=ky(y|0,B|0,1,0)|0,t=L,A=yy(y|0,B|0,u|0,v|0)|0,B=jy(z|0,e|0,A|0,L|0)|0,(Kr(a,b,w,t,B,L,r,q,l)|0)>0):0){I=29;break a}B=ky(c|0,d|0,1,0)|0;R=B;S=L;T=0;U=0}B=ky(s|0,m|0,1,0)|0;t=L;if((t|0)<(h|0)|(t|0)==(h|0)&B>>>0<g>>>0){s=B;r=T;q=U;m=t;c=R;d=S}else{J=R;K=S;M=B;N=t;O=T;P=U;Q=0;break}}}while(0);if((I|0)==29){n=k[l>>2]|0;return n|0}if((J|0)==0&(K|0)==0){if(!((O|0)==0&(P|0)==0)){I=jy(M|0,N|0,O|0,P|0)|0;U=ky(I|0,L|0,o|0,f|0)|0;I=L;T=ky(U|0,I|0,-1,-1)|0;S=wy(T|0,L|0,u|0,v|0)|0;T=L;R=ky(S|0,T|0,1,0)|0;g=L;h=yy(S|0,T|0,u|0,v|0)|0;T=jy(U|0,I|0,h|0,L|0)|0;Kr(a,b,R,g,T,L,O,P,l)|0}}else{P=jy(M|0,N|0,J|0,K|0)|0;N=ky(P|0,L|0,o|0,f|0)|0;f=L;o=ky(N|0,f|0,-1,-1)|0;M=wy(o|0,L|0,u|0,v|0)|0;o=L;if(p){p=yy(M|0,o|0,u|0,v|0)|0;v=jy(N|0,f|0,p|0,L|0)|0;p=L;f=ky(M|0,o|0,1,0)|0;jr(a,b,f,L,v,p,J,K,i+(P<<2)|0,l)|0}}P=k[l>>2]|0;if(!((Q|0)!=0&(P|0)<1)){n=P;return n|0}k[l>>2]=412;n=412;return n|0}function mr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){o=g;q=0;do{p[m>>3]=+(k[a+(q<<2)>>2]|0);Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}}else if(n){n=g;q=0;do{p[l>>3]=(+(k[a+(q<<2)>>2]|0)-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}b=rx(g,44)|0;if(!b){s=k[h>>2]|0;r=j;return s|0}else t=b;do{i[t>>0]=46;t=rx(t,44)|0}while((t|0)!=0);s=k[h>>2]|0;r=j;return s|0}function nr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0;if(!(Cl(a,h)|0)){i=(b|0)<1;or(a,2,i?1:b,i?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;j=k[h>>2]|0;return j|0}else{tk(771020);k[h>>2]=413;j=413;return j|0}return 0}function or(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0.0,ea=0.0,fa=0,ja=0,ka=0.0,la=0.0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}f=k[u>>2]|0;u=((f|0)<0)<<31>>31;if((k[t>>2]|0)==16)Rk(E,F);e=F+1|0;d=g;g=h;h=0;c=0;I=0;K=0;a:while(1){M=h;N=c;O=d;P=g;while(1){Q=l+(M<<3)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=100;break a}U=(S|0)<(u|0)|(S|0)==(u|0)&R>>>0<f>>>0;X=U?R:f;U=my(0,X|0,32)|0;X=L;Y=y;Z=k[Y>>2]|0;_=k[Y+4>>2]|0;Y=A;$=k[Y>>2]|0;aa=k[Y+4>>2]|0;Y=jy(Z|0,_|0,$|0,aa|0)|0;ba=L;ca=(X|0)<(ba|0)|(X|0)==(ba|0)&U>>>0<Y>>>0;ba=ca?U:Y;Y=z;U=k[Y>>2]|0;ca=k[Y+4>>2]|0;Y=B;X=yy(k[Y>>2]|0,k[Y+4>>2]|0,I|0,K|0)|0;Y=ky(X|0,L|0,U|0,ca|0)|0;ca=L;U=k[x>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,$|0,aa|0)|0;U=ky(Y|0,ca|0,X|0,L|0)|0;au(a,U,L,1,m)|0;b:do switch(k[t>>2]|0){case 81:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){X=ky($|0,aa|0,-1,-1)|0;ca=jy(X|0,L|0,Z|0,_|0)|0;X=L;Y=(u|0)>(S|0)|(u|0)==(S|0)&f>>>0>R>>>0;fa=Y?R:f;Y=my(0,fa|0,32)|0;fa=~Y;Y=~L;ja=(X|0)>(Y|0)|(X|0)==(Y|0)&ca>>>0>fa>>>0;Y=ja?ca:fa;py(H|0,Q|0,(Y^536870911)<<3|0)|0}}else if(U){U=0;do{Y=l+(U+M<<3)|0;ka=(+((k[Y>>2]|0)>>>0)+4294967296.0*+(k[Y+4>>2]|0)-ea)/da;do if(!(ka<-9223372036854775808.0)){if(ka>9223372036854775808.0){k[m>>2]=-11;Y=H+(U<<3)|0;k[Y>>2]=-1;k[Y+4>>2]=2147483647;break}if(!(ka>=0.0)){la=ka+-.5;Y=+W(la)>=1.0?(la>0.0?~~+ia(+V(la/4294967296.0),4294967295.0)>>>0:~~+ga((la-+(~~la>>>0))/4294967296.0)>>>0):0;fa=H+(U<<3)|0;k[fa>>2]=~~la>>>0;k[fa+4>>2]=Y;break}else{la=ka+.5;Y=+W(la)>=1.0?(la>0.0?~~+ia(+V(la/4294967296.0),4294967295.0)>>>0:~~+ga((la-+(~~la>>>0))/4294967296.0)>>>0):0;fa=H+(U<<3)|0;k[fa>>2]=~~la>>>0;k[fa+4>>2]=Y;break}}else{k[m>>2]=-11;Y=H+(U<<3)|0;k[Y>>2]=0;k[Y+4>>2]=-2147483648}while(0);U=U+1|0}while((U|0)!=(ba|0))}vu(a,ba,k[x>>2]|0,H,m)|0;break}case 41:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{fa=l+(Y+M<<3)|0;ca=k[fa>>2]|0;ja=k[fa+4>>2]|0;do if(!((ja|0)<-1|(ja|0)==-1&ca>>>0<2147483648))if((ja|0)>0|(ja|0)==0&ca>>>0>2147483647){k[m>>2]=-11;k[H+(Y<<2)>>2]=2147483647;break}else{k[H+(Y<<2)>>2]=ca;break}else{k[m>>2]=-11;k[H+(Y<<2)>>2]=-2147483648}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{ca=l+(Y+M<<3)|0;ka=(+((k[ca>>2]|0)>>>0)+4294967296.0*+(k[ca+4>>2]|0)-ea)/da;do if(!(ka<-2147483648.49)){if(ka>2147483647.49){k[m>>2]=-11;k[H+(Y<<2)>>2]=2147483647;break}if(!(ka>=0.0)){k[H+(Y<<2)>>2]=~~(ka+-.5);break}else{k[H+(Y<<2)>>2]=~~(ka+.5);break}}else{k[m>>2]=-11;k[H+(Y<<2)>>2]=-2147483648}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}uu(a,ba,k[x>>2]|0,H,m)|0;break}case 11:{da=+p[C>>3];ea=+p[D>>3];Y=(ba|0)>0;if(da==1.0&ea==0.0){if(Y){U=0;do{ca=l+(U+M<<3)|0;ja=k[ca>>2]|0;fa=k[ca+4>>2]|0;do if((fa|0)>=0)if((fa|0)>0|(fa|0)==0&ja>>>0>255){k[m>>2]=-11;i[H+U>>0]=-1;break}else{i[H+U>>0]=ja;break}else{k[m>>2]=-11;i[H+U>>0]=0}while(0);U=U+1|0}while((U|0)!=(ba|0))}}else if(Y){U=0;do{ja=l+(U+M<<3)|0;ka=(+((k[ja>>2]|0)>>>0)+4294967296.0*+(k[ja+4>>2]|0)-ea)/da;do if(!(ka<-.49))if(ka>255.49){k[m>>2]=-11;i[H+U>>0]=-1;break}else{i[H+U>>0]=~~(ka+.5);break}else{k[m>>2]=-11;i[H+U>>0]=0}while(0);U=U+1|0}while((U|0)!=(ba|0))}su(a,ba,k[x>>2]|0,H,m)|0;break}case 21:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{ja=l+(Y+M<<3)|0;fa=k[ja>>2]|0;ca=k[ja+4>>2]|0;do if(!((ca|0)<-1|(ca|0)==-1&fa>>>0<4294934528))if((ca|0)>0|(ca|0)==0&fa>>>0>32767){k[m>>2]=-11;j[H+(Y<<1)>>1]=32767;break}else{j[H+(Y<<1)>>1]=fa;break}else{k[m>>2]=-11;j[H+(Y<<1)>>1]=-32768}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{fa=l+(Y+M<<3)|0;ka=(+((k[fa>>2]|0)>>>0)+4294967296.0*+(k[fa+4>>2]|0)-ea)/da;do if(!(ka<-32768.49)){if(ka>32767.49){k[m>>2]=-11;j[H+(Y<<1)>>1]=32767;break}if(!(ka>=0.0)){j[H+(Y<<1)>>1]=~~(ka+-.5);break}else{j[H+(Y<<1)>>1]=~~(ka+.5);break}}else{k[m>>2]=-11;j[H+(Y<<1)>>1]=-32768}while(0);Y=Y+1|0}while((Y|0)!=(ba|0))}tu(a,ba,k[x>>2]|0,H,m)|0;break}case 42:{da=+p[C>>3];ea=+p[D>>3];Y=(ba|0)>0;if(da==1.0&ea==0.0){if(Y){U=0;do{fa=l+(U+M<<3)|0;o[H+(U<<2)>>2]=+((k[fa>>2]|0)>>>0)+4294967296.0*+(k[fa+4>>2]|0);U=U+1|0}while((U|0)!=(ba|0))}}else if(Y){U=0;do{fa=l+(U+M<<3)|0;o[H+(U<<2)>>2]=(+((k[fa>>2]|0)>>>0)+4294967296.0*+(k[fa+4>>2]|0)-ea)/da;U=U+1|0}while((U|0)!=(ba|0))}wu(a,ba,k[x>>2]|0,H,m)|0;break}case 82:{da=+p[C>>3];ea=+p[D>>3];U=(ba|0)>0;if(da==1.0&ea==0.0){if(U){Y=0;do{fa=l+(Y+M<<3)|0;p[H+(Y<<3)>>3]=+((k[fa>>2]|0)>>>0)+4294967296.0*+(k[fa+4>>2]|0);Y=Y+1|0}while((Y|0)!=(ba|0))}}else if(U){Y=0;do{fa=l+(Y+M<<3)|0;p[H+(Y<<3)>>3]=(+((k[fa>>2]|0)>>>0)+4294967296.0*+(k[fa+4>>2]|0)-ea)/da;Y=Y+1|0}while((Y|0)!=(ba|0))}xu(a,ba,k[x>>2]|0,H,m)|0;break}case 16:{if((i[e>>0]|0)==115){T=92;break a}rr(Q,ba,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;Y=k[x>>2]|0;U=k[w>>2]|0;if((Y|0)==(U|0)){fa=ha(Y,ba)|0;cu(a,fa,((fa|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,U,ba,Y-U|0,H,m)|0;break b}break}default:{T=92;break a}}while(0);if((k[m>>2]|0)>0){ma=M;na=N;oa=ba;T=96;break a}_=((ba|0)<0)<<31>>31;Z=jy(R|0,S|0,ba|0,_|0)|0;aa=L;if((R|0)==(ba|0)&(S|0)==(_|0)){R=Z;S=aa}else{pa=ba;qa=_;ra=Z;sa=aa;break}}S=ky(pa|0,qa|0,M|0,N|0)|0;R=L;Q=A;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,pa|0,qa|0)|0;Q=L;Z=A;k[Z>>2]=aa;k[Z+4>>2]=Q;Z=y;if((aa|0)==(k[Z>>2]|0)?(Q|0)==(k[Z+4>>2]|0):0){ta=S;ua=R;va=ra;wa=sa;break}else{M=S;N=R;O=ra;P=sa}}P=A;k[P>>2]=0;k[P+4>>2]=0;P=ky(I|0,K|0,1,0)|0;d=va;g=wa;h=ta;c=ua;I=P;K=L}if((T|0)==92){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((T|0)==96){v=ky(ma|0,na|0,1,0)|0;da=+(v>>>0)+4294967296.0*+(L|0);v=ky(oa|0,((oa|0)<0)<<31>>31|0,ma|0,na|0)|0;p[q>>3]=da;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,770952,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((T|0)==100){T=k[m>>2]|0;if((T|0)!=-11){J=T;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function pr(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;if(!(Cl(a,j)|0)){l=(b|0)<1;qr(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i,j)|0;m=k[j>>2]|0;return m|0}else{tk(771020);k[j>>2]=413;m=413;return m|0}return 0}function qr(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[n+(p*160|0)+80>>2]|0;q=(r|0)>0;do if(!q){s=ky(e|0,f|0,-1,-1)|0;t=ky(s|0,L|0,g|0,h|0)|0;s=L;if((r|0)<0?(or(a,b,c,d,e,f,g,h,i,m)|0)>0:0){u=k[m>>2]|0;if((u|0)==412){k[m>>2]=0;v=t;w=s;break}else{o=u;return o|0}}else{v=t;w=s}}else{s=n+(p*160|0)+88|0;v=k[s>>2]|0;w=k[s+4>>2]|0}while(0);p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,w|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;if((h|0)>0|(h|0)==0&g>>>0>0){x=0;y=0;z=0;A=0;B=0;C=0;D=0}else{o=k[m>>2]|0;return o|0}while(1){e=i+(z<<3)|0;if((k[e>>2]|0)==(j|0)?(k[e+4>>2]|0)==(l|0):0){if((!((x|0)==0&(y|0)==0)?(e=jy(z|0,C|0,x|0,y|0)|0,d=ky(e|0,L|0,p|0,f|0)|0,c=L,n=ky(d|0,c|0,-1,-1)|0,r=wy(n|0,L|0,v|0,w|0)|0,n=L,q):0)?(s=yy(r|0,n|0,v|0,w|0)|0,t=jy(d|0,c|0,s|0,L|0)|0,s=L,c=ky(r|0,n|0,1,0)|0,(or(a,b,c,L,t,s,x,y,i+(e<<3)|0,m)|0)>0):0){e=k[m>>2]|0;if((e|0)!=412){o=e;E=34;break}k[m>>2]=0;F=1}else F=D;e=ky(A|0,B|0,1,0)|0;G=0;H=0;I=e;J=L;K=F}else{if(!((A|0)==0&(B|0)==0)?(e=jy(z|0,C|0,A|0,B|0)|0,s=ky(e|0,L|0,p|0,f|0)|0,e=L,t=ky(s|0,e|0,-1,-1)|0,c=wy(t|0,L|0,v|0,w|0)|0,t=L,n=ky(c|0,t|0,1,0)|0,r=L,d=yy(c|0,t|0,v|0,w|0)|0,t=jy(s|0,e|0,d|0,L|0)|0,(Kr(a,b,n,r,t,L,A,B,m)|0)>0):0){E=18;break}t=ky(x|0,y|0,1,0)|0;G=t;H=L;I=0;J=0;K=D}t=ky(z|0,C|0,1,0)|0;r=L;if((r|0)<(h|0)|(r|0)==(h|0)&t>>>0<g>>>0){x=G;y=H;z=t;A=I;B=J;C=r;D=K}else{M=G;N=H;O=t;P=r;Q=I;R=J;S=K;E=27;break}}if((E|0)==18){o=k[m>>2]|0;return o|0}else if((E|0)==27){if((M|0)==0&(N|0)==0){if(!((Q|0)==0&(R|0)==0)){K=jy(O|0,P|0,Q|0,R|0)|0;J=ky(K|0,L|0,p|0,f|0)|0;K=L;I=ky(J|0,K|0,-1,-1)|0;H=wy(I|0,L|0,v|0,w|0)|0;I=L;G=ky(H|0,I|0,1,0)|0;D=L;C=yy(H|0,I|0,v|0,w|0)|0;I=jy(J|0,K|0,C|0,L|0)|0;Kr(a,b,G,D,I,L,Q,R,m)|0}}else{R=jy(O|0,P|0,M|0,N|0)|0;P=ky(R|0,L|0,p|0,f|0)|0;f=L;p=ky(P|0,f|0,-1,-1)|0;O=wy(p|0,L|0,v|0,w|0)|0;p=L;if(q){q=yy(O|0,p|0,v|0,w|0)|0;w=jy(P|0,f|0,q|0,L|0)|0;q=L;f=ky(O|0,p|0,1,0)|0;or(a,b,f,L,w,q,M,N,i+(R<<3)|0,m)|0}}R=k[m>>2]|0;if(!((S|0)!=0&(R|0)<1)){o=R;return o|0}k[m>>2]=412;o=412;return o|0}else if((E|0)==34)return o|0;return 0}function rr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){o=g;q=0;do{s=a+(q<<3)|0;p[m>>3]=+((k[s>>2]|0)>>>0)+4294967296.0*+(k[s+4>>2]|0);Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}}else if(n){n=g;q=0;do{o=a+(q<<3)|0;p[l>>3]=(+((k[o>>2]|0)>>>0)+4294967296.0*+(k[o+4>>2]|0)-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}b=rx(g,44)|0;if(!b){t=k[h>>2]|0;r=j;return t|0}else u=b;do{i[u>>0]=46;u=rx(u,44)|0}while((u|0)!=0);t=k[h>>2]|0;r=j;return t|0}function sr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;j=i;if(!(Cl(a,h)|0)){if((k[h>>2]|0)<=0){l=(b|0)<1;jr(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0}}else jt(a,31,c,d,e,f,0,g,j,h)|0;r=i;return k[h>>2]|0}function tr(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;l=k[j>>2]|0;if((l|0)>0){m=l;return m|0}jr(a,b,c,d,e,f,g,h,i,j)|0;m=k[j>>2]|0;return m|0}function ur(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;vr(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{k[l>>2]=h;jt(a,31,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function vr(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;o=a+4|0;p=k[o>>2]|0;if((m|0)==(k[p+68>>2]|0)){q=p+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[o>>2]|0)+960>>2]|0;o=b+-1|0;q=k[m+(o*160|0)+80>>2]|0;p=(q|0)>0;do if(!p){r=ky(e|0,f|0,-1,-1)|0;s=ky(r|0,L|0,g|0,h|0)|0;r=L;if((q|0)<0){t=k[l>>2]|0;if((t|0)<=0){jr(a,b,c,d,e,f,g,h,i,l)|0;u=k[l>>2]|0;if((u|0)>0)v=u;else{w=s;x=r;break}}else v=t;if((v|0)==412){k[l>>2]=0;w=s;x=r;break}else{n=v;return n|0}}else{w=s;x=r}}else{r=m+(o*160|0)+88|0;w=k[r>>2]|0;x=k[r+4>>2]|0}while(0);o=ky(c|0,d|0,-1,-1)|0;d=yy(w|0,x|0,o|0,L|0)|0;o=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(p){e=0;d=0;c=0;m=0;v=0;q=0;r=0;b:while(1){if((k[i+(e<<2)>>2]|0)==(j|0)){do if((v|0)==0&(q|0)==0)y=r;else{s=jy(e|0,m|0,v|0,q|0)|0;t=ky(s|0,L|0,o|0,f|0)|0;u=L;z=ky(t|0,u|0,-1,-1)|0;A=wy(z|0,L|0,w|0,x|0)|0;z=L;B=k[l>>2]|0;if((B|0)<=0){C=ky(A|0,z|0,1,0)|0;D=L;E=yy(A|0,z|0,w|0,x|0)|0;z=jy(t|0,u|0,E|0,L|0)|0;jr(a,b,C,D,z,L,v,q,i+(s<<2)|0,l)|0;s=k[l>>2]|0;if((s|0)>0)F=s;else{y=r;break}}else F=B;if((F|0)!=412){n=F;break b}k[l>>2]=0;y=1}while(0);B=ky(d|0,c|0,1,0)|0;G=0;H=0;I=B;J=L;K=y}else{if(!((d|0)==0&(c|0)==0)?(B=jy(e|0,m|0,d|0,c|0)|0,s=ky(B|0,L|0,o|0,f|0)|0,B=L,z=ky(s|0,B|0,-1,-1)|0,D=wy(z|0,L|0,w|0,x|0)|0,z=L,C=ky(D|0,z|0,1,0)|0,E=L,u=yy(D|0,z|0,w|0,x|0)|0,z=jy(s|0,B|0,u|0,L|0)|0,(Kr(a,b,C,E,z,L,d,c,l)|0)>0):0){M=31;break a}z=ky(v|0,q|0,1,0)|0;G=z;H=L;I=0;J=0;K=r}z=ky(e|0,m|0,1,0)|0;E=L;if((E|0)<(h|0)|(E|0)==(h|0)&z>>>0<g>>>0){e=z;d=I;c=J;m=E;v=G;q=H;r=K}else{N=G;O=H;P=z;Q=E;R=I;S=J;T=K;break a}}return n|0}else{r=0;q=0;v=0;m=0;c=0;d=0;while(1){if((k[i+(r<<2)>>2]|0)==(j|0)){e=ky(q|0,v|0,1,0)|0;U=0;V=0;W=e;X=L}else{if(!((q|0)==0&(v|0)==0)?(e=jy(r|0,m|0,q|0,v|0)|0,E=ky(e|0,L|0,o|0,f|0)|0,e=L,z=ky(E|0,e|0,-1,-1)|0,C=wy(z|0,L|0,w|0,x|0)|0,z=L,u=ky(C|0,z|0,1,0)|0,B=L,s=yy(C|0,z|0,w|0,x|0)|0,z=jy(E|0,e|0,s|0,L|0)|0,(Kr(a,b,u,B,z,L,q,v,l)|0)>0):0){M=31;break a}z=ky(c|0,d|0,1,0)|0;U=z;V=L;W=0;X=0}z=ky(r|0,m|0,1,0)|0;B=L;if((B|0)<(h|0)|(B|0)==(h|0)&z>>>0<g>>>0){r=z;q=W;v=X;m=B;c=U;d=V}else{N=U;O=V;P=z;Q=B;R=W;S=X;T=0;break}}}while(0);if((M|0)==31){n=k[l>>2]|0;return n|0}if((N|0)==0&(O|0)==0){if(!((R|0)==0&(S|0)==0)){M=jy(P|0,Q|0,R|0,S|0)|0;X=ky(M|0,L|0,o|0,f|0)|0;M=L;W=ky(X|0,M|0,-1,-1)|0;V=wy(W|0,L|0,w|0,x|0)|0;W=L;U=ky(V|0,W|0,1,0)|0;g=L;h=yy(V|0,W|0,w|0,x|0)|0;W=jy(X|0,M|0,h|0,L|0)|0;Kr(a,b,U,g,W,L,R,S,l)|0}}else{S=jy(P|0,Q|0,N|0,O|0)|0;Q=ky(S|0,L|0,o|0,f|0)|0;f=L;o=ky(Q|0,f|0,-1,-1)|0;P=wy(o|0,L|0,w|0,x|0)|0;o=L;if(p?(k[l>>2]|0)<=0:0){p=ky(P|0,o|0,1,0)|0;R=L;W=yy(P|0,o|0,w|0,x|0)|0;x=jy(Q|0,f|0,W|0,L|0)|0;jr(a,b,p,R,x,L,N,O,i+(S<<2)|0,l)|0}}S=k[l>>2]|0;if(!((T|0)!=0&(S|0)<1)){n=S;return n|0}k[l>>2]=412;n=412;return n|0}function wr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;j=i;if(!(Cl(a,h)|0)){if((k[h>>2]|0)<=0){l=(b|0)<1;Tr(a,2,l?1:b,l?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0}}else jt(a,30,c,d,e,f,0,g,j,h)|0;r=i;return k[h>>2]|0}function xr(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0;l=k[j>>2]|0;if((l|0)>0){m=l;return m|0}Tr(a,b,c,d,e,f,g,h,i,j)|0;m=k[j>>2]|0;return m|0}function yr(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;zr(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{k[l>>2]=h;jt(a,30,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function zr(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;o=a+4|0;p=k[o>>2]|0;if((m|0)==(k[p+68>>2]|0)){q=p+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[o>>2]|0)+960>>2]|0;o=b+-1|0;q=k[m+(o*160|0)+80>>2]|0;p=(q|0)>0;do if(!p){r=ky(e|0,f|0,-1,-1)|0;s=ky(r|0,L|0,g|0,h|0)|0;r=L;if((q|0)<0){t=k[l>>2]|0;if((t|0)<=0){Tr(a,b,c,d,e,f,g,h,i,l)|0;u=k[l>>2]|0;if((u|0)>0)v=u;else{w=s;x=r;break}}else v=t;if((v|0)==412){k[l>>2]=0;w=s;x=r;break}else{n=v;return n|0}}else{w=s;x=r}}else{r=m+(o*160|0)+88|0;w=k[r>>2]|0;x=k[r+4>>2]|0}while(0);o=ky(c|0,d|0,-1,-1)|0;d=yy(w|0,x|0,o|0,L|0)|0;o=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(p){e=0;d=0;c=0;m=0;v=0;q=0;r=0;b:while(1){if((k[i+(e<<2)>>2]|0)==(j|0)){do if((v|0)==0&(q|0)==0)y=r;else{s=jy(e|0,m|0,v|0,q|0)|0;t=ky(s|0,L|0,o|0,f|0)|0;u=L;z=ky(t|0,u|0,-1,-1)|0;A=wy(z|0,L|0,w|0,x|0)|0;z=L;B=k[l>>2]|0;if((B|0)<=0){C=ky(A|0,z|0,1,0)|0;D=L;E=yy(A|0,z|0,w|0,x|0)|0;z=jy(t|0,u|0,E|0,L|0)|0;Tr(a,b,C,D,z,L,v,q,i+(s<<2)|0,l)|0;s=k[l>>2]|0;if((s|0)>0)F=s;else{y=r;break}}else F=B;if((F|0)!=412){n=F;break b}k[l>>2]=0;y=1}while(0);B=ky(d|0,c|0,1,0)|0;G=0;H=0;I=B;J=L;K=y}else{if(!((d|0)==0&(c|0)==0)?(B=jy(e|0,m|0,d|0,c|0)|0,s=ky(B|0,L|0,o|0,f|0)|0,B=L,z=ky(s|0,B|0,-1,-1)|0,D=wy(z|0,L|0,w|0,x|0)|0,z=L,C=ky(D|0,z|0,1,0)|0,E=L,u=yy(D|0,z|0,w|0,x|0)|0,z=jy(s|0,B|0,u|0,L|0)|0,(Kr(a,b,C,E,z,L,d,c,l)|0)>0):0){M=31;break a}z=ky(v|0,q|0,1,0)|0;G=z;H=L;I=0;J=0;K=r}z=ky(e|0,m|0,1,0)|0;E=L;if((E|0)<(h|0)|(E|0)==(h|0)&z>>>0<g>>>0){e=z;d=I;c=J;m=E;v=G;q=H;r=K}else{N=G;O=H;P=z;Q=E;R=I;S=J;T=K;break a}}return n|0}else{r=0;q=0;v=0;m=0;c=0;d=0;while(1){if((k[i+(r<<2)>>2]|0)==(j|0)){e=ky(q|0,v|0,1,0)|0;U=0;V=0;W=e;X=L}else{if(!((q|0)==0&(v|0)==0)?(e=jy(r|0,m|0,q|0,v|0)|0,E=ky(e|0,L|0,o|0,f|0)|0,e=L,z=ky(E|0,e|0,-1,-1)|0,C=wy(z|0,L|0,w|0,x|0)|0,z=L,u=ky(C|0,z|0,1,0)|0,B=L,s=yy(C|0,z|0,w|0,x|0)|0,z=jy(E|0,e|0,s|0,L|0)|0,(Kr(a,b,u,B,z,L,q,v,l)|0)>0):0){M=31;break a}z=ky(c|0,d|0,1,0)|0;U=z;V=L;W=0;X=0}z=ky(r|0,m|0,1,0)|0;B=L;if((B|0)<(h|0)|(B|0)==(h|0)&z>>>0<g>>>0){r=z;q=W;v=X;m=B;c=U;d=V}else{N=U;O=V;P=z;Q=B;R=W;S=X;T=0;break}}}while(0);if((M|0)==31){n=k[l>>2]|0;return n|0}if((N|0)==0&(O|0)==0){if(!((R|0)==0&(S|0)==0)){M=jy(P|0,Q|0,R|0,S|0)|0;X=ky(M|0,L|0,o|0,f|0)|0;M=L;W=ky(X|0,M|0,-1,-1)|0;V=wy(W|0,L|0,w|0,x|0)|0;W=L;U=ky(V|0,W|0,1,0)|0;g=L;h=yy(V|0,W|0,w|0,x|0)|0;W=jy(X|0,M|0,h|0,L|0)|0;Kr(a,b,U,g,W,L,R,S,l)|0}}else{S=jy(P|0,Q|0,N|0,O|0)|0;Q=ky(S|0,L|0,o|0,f|0)|0;f=L;o=ky(Q|0,f|0,-1,-1)|0;P=wy(o|0,L|0,w|0,x|0)|0;o=L;if(p?(k[l>>2]|0)<=0:0){p=ky(P|0,o|0,1,0)|0;R=L;W=yy(P|0,o|0,w|0,x|0)|0;x=jy(Q|0,f|0,W|0,L|0)|0;Tr(a,b,p,R,x,L,N,O,i+(S<<2)|0,l)|0}}S=k[l>>2]|0;if(!((T|0)!=0&(S|0)<1)){n=S;return n|0}k[l>>2]=412;n=412;return n|0}function Ar(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0;m=r;r=r+224|0;n=m+56|0;o=m+80|0;q=m+64|0;s=m+48|0;t=m+40|0;u=m+32|0;v=m+24|0;w=m+105|0;x=m+104|0;y=m+128|0;i[w>>0]=84;i[x>>0]=70;z=k[l>>2]|0;if((z|0)>0){A=z;r=m;return A|0}if((nl(a,b,c,d,e,f,g,h,1,m+8|0,m,m+108|0,m+68|0,o,m+76|0,t,u,q,s,v,m+72|0,m+16|0,m+84|0,l)|0)>0){A=k[l>>2]|0;r=m;return A|0}if((k[o>>2]|0)==14){B=0;C=0;D=0;E=g;F=h;G=0}else{k[l>>2]=310;A=310;r=m;return A|0}a:while(1){h=E;g=F;o=G;f=B;while(1){e=j+o|0;d=h;c=g;while(1){if((d|0)==0&(c|0)==0){H=17;break a}b=t;z=k[b>>2]|0;I=k[b+4>>2]|0;b=v;J=yy(k[b>>2]|0,k[b+4>>2]|0,C|0,D|0)|0;b=ky(J|0,L|0,z|0,I|0)|0;I=L;z=u;J=k[q>>2]|0;K=yy(J|0,((J|0)<0)<<31>>31|0,k[z>>2]|0,k[z+4>>2]|0)|0;z=ky(b|0,I|0,K|0,L|0)|0;au(a,z,L,1,l)|0;if(!(i[e>>0]|0))cu(a,1,0,x,l)|0;else cu(a,1,0,w,l)|0;if((k[l>>2]|0)>0){M=o;N=f;H=11;break a}z=ky(d|0,c|0,-1,-1)|0;K=L;if((z|0)==0&(K|0)==0){d=0;c=0}else{O=z;P=K;break}}c=ky(o|0,f|0,1,0)|0;d=L;e=u;K=ky(k[e>>2]|0,k[e+4>>2]|0,1,0)|0;e=L;z=u;k[z>>2]=K;k[z+4>>2]=e;z=s;if((K|0)==(k[z>>2]|0)?(e|0)==(k[z+4>>2]|0):0){Q=c;R=d;S=O;T=P;break}else{h=O;g=P;o=c;f=d}}f=u;k[f>>2]=0;k[f+4>>2]=0;f=ky(C|0,D|0,1,0)|0;B=R;C=f;D=L;E=S;F=T;G=Q}if((H|0)==11){Q=ky(M|0,N|0,1,0)|0;p[n>>3]=+(Q>>>0)+4294967296.0*+(L|0);Xw(y,771065,n)|0;tk(y);A=k[l>>2]|0;r=m;return A|0}else if((H|0)==17){A=k[l>>2]|0;r=m;return A|0}return 0}function Br(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;if((k[n+(p*160|0)+80>>2]|0)>0){r=n+(p*160|0)+88|0;s=k[r>>2]|0;t=k[r+4>>2]|0}else{r=ky(e|0,f|0,-1,-1)|0;p=ky(r|0,L|0,g|0,h|0)|0;s=p;t=L}if((Ar(a,b,c,d,e,f,g,h,j,m)|0)>0){o=k[m>>2]|0;return o|0}p=ky(c|0,d|0,-1,-1)|0;d=yy(s|0,t|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;do if((h|0)>0|(h|0)==0&g>>>0>0){e=0;d=0;c=0;r=0;n=0;q=0;while(1){if((i[j+e>>0]|0)==l<<24>>24){u=ky(d|0,c|0,1,0)|0;v=0;w=0;x=u;y=L}else{if(!((d|0)==0&(c|0)==0)?(u=jy(e|0,r|0,d|0,c|0)|0,z=ky(u|0,L|0,p|0,f|0)|0,u=L,A=ky(z|0,u|0,-1,-1)|0,B=wy(A|0,L|0,s|0,t|0)|0,A=L,C=ky(B|0,A|0,1,0)|0,D=L,E=yy(B|0,A|0,s|0,t|0)|0,A=jy(z|0,u|0,E|0,L|0)|0,(Kr(a,b,C,D,A,L,d,c,m)|0)>0):0){F=16;break}A=ky(n|0,q|0,1,0)|0;v=A;w=L;x=0;y=0}A=ky(e|0,r|0,1,0)|0;D=L;if((D|0)<(h|0)|(D|0)==(h|0)&A>>>0<g>>>0){e=A;d=x;c=y;r=D;n=v;q=w}else{G=v;H=w;I=x;J=y;K=A;M=D;F=20;break}}if((F|0)==16){o=k[m>>2]|0;return o|0}else if((F|0)==20){if((G|0)!=0|(H|0)!=0|(I|0)==0&(J|0)==0)break;q=jy(K|0,M|0,I|0,J|0)|0;n=ky(q|0,L|0,p|0,f|0)|0;q=L;r=ky(n|0,q|0,-1,-1)|0;c=wy(r|0,L|0,s|0,t|0)|0;r=L;d=ky(c|0,r|0,1,0)|0;e=L;D=yy(c|0,r|0,s|0,t|0)|0;r=jy(n|0,q|0,D|0,L|0)|0;Kr(a,b,d,e,r,L,I,J,m)|0;break}}while(0);o=k[m>>2]|0;return o|0}function Cr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;j=r;r=r+112|0;l=j+56|0;m=j+48|0;n=j+40|0;o=j+32|0;p=j+24|0;q=j+80|0;s=j+76|0;t=j+72|0;u=j+68|0;v=j+64|0;w=j+16|0;x=j+100|0;y=j+88|0;z=j+84|0;A=j+8|0;B=j;C=k[h>>2]|0;if((f|0)<1|(C|0)>0){D=C;r=j;return D|0}if((d|0)<0|(d|0)==0&c>>>0<1){k[h>>2]=307;D=307;r=j;return D|0}if((e|0)<1){k[h>>2]=308;D=308;r=j;return D|0}C=k[a>>2]|0;E=a+4|0;F=k[E>>2]|0;if((C|0)==(k[F+68>>2]|0)){G=F+120|0;if(((k[G>>2]|0)==-1?(k[G+4>>2]|0)==-1:0)?(Vk(a,h)|0)>0:0){D=k[h>>2]|0;r=j;return D|0}}else Fk(a,C+1|0,0,h)|0;C=(e+7|0)/8|0;G=f+e|0;F=1-C+((G+6|0)/8|0)|0;H=((C|0)<0)<<31>>31;if((nl(a,b,c,d,C,H,F,((F|0)<0)<<31>>31,1,w,w,x,q,t,u,l,o,s,m,n,v,p,y,h)|0)>0){D=k[h>>2]|0;r=j;return D|0}y=e+-1|0;e=y-(((y|0)/8|0)<<3)|0;y=ky(c|0,d|0,-1,-1)|0;p=L;v=C+-1|0;n=((v|0)<0)<<31>>31;s=k[E>>2]|0;o=k[s+960>>2]|0;u=b+-1|0;q=k[o+(u*160|0)+80>>2]|0;k[t>>2]=q;if((((q|0)>-1?q:0-q|0)|0)>11){k[h>>2]=310;D=310;r=j;return D|0}do if((q|0)>0){t=o+(u*160|0)+88|0;x=k[t>>2]|0;w=k[t+4>>2]|0;t=(q|0)==1;F=ky(x|0,w|0,7,0)|0;I=wy(F|0,L|0,8,0)|0;F=t?I:x;x=t?L:w;w=m;k[w>>2]=F;k[w+4>>2]=x;if(!((H|0)>(x|0)|(H|0)==(x|0)&C>>>0>F>>>0)){F=s+120|0;x=k[F>>2]|0;w=k[F+4>>2]|0;F=s+952|0;t=yy(k[F>>2]|0,k[F+4>>2]|0,y|0,p|0)|0;F=ky(t|0,L|0,x|0,w|0)|0;w=o+(u*160|0)+72|0;x=w;t=ky(F|0,L|0,k[x>>2]|0,k[x+4>>2]|0)|0;J=w;K=t;M=L;N=1;break}k[h>>2]=308;D=308;r=j;return D|0}else{t=G+-1|0;w=m;k[w>>2]=t;k[w+4>>2]=((t|0)<0)<<31>>31;if((q|0)==-1){pl(a,b,c,d,B,A,h)|0;t=my(0,k[m>>2]|0,32)|0;w=A;ql(a,b,c,d,t,L,k[w>>2]|0,k[w+4>>2]|0,h)|0}w=l;J=o+(u*160|0)+72|0;K=k[w>>2]|0;M=k[w+4>>2]|0;N=0}while(0);u=ky(K|0,M|0,v|0,n|0)|0;M=L;K=l;k[K>>2]=u;k[K+4>>2]=M;au(a,u,M,1,h)|0;M=y;y=p;p=v;v=n;n=e;e=0;a:while(1){u=p;K=v;o=n;d=e;while(1){if(!N){O=o;P=d;Q=35;break a}if((fu(a,1,0,z,h)|0)==107){k[h>>2]=0;i[z>>0]=0}c=l;au(a,k[c>>2]|0,k[c+4>>2]|0,1,h)|0;if((d|0)<(f|0)&(o|0)<8){c=d-f|0;b=o+-8|0;A=c>>>0>b>>>0?c:b;b=o-A|0;c=o;B=d;q=i[z>>0]|0;while(1){if(!(i[g+B>>0]|0))R=i[771137+c>>0]&q;else R=i[771129+c>>0]|q;c=c+1|0;if((c|0)==(b|0)){S=R;break}else{B=B+1|0;q=R}}i[z>>0]=S;T=d-A|0}else T=d;cu(a,1,0,z,h)|0;if((T|0)==(f|0))break a;q=l;B=ky(k[q>>2]|0,k[q+4>>2]|0,1,0)|0;q=l;k[q>>2]=B;k[q+4>>2]=L;u=ky(u|0,K|0,1,0)|0;K=L;q=m;if((u|0)==(k[q>>2]|0)?(K|0)==(k[q+4>>2]|0):0){U=T;break}else{o=0;d=T}}d=ky(M|0,y|0,1,0)|0;o=L;K=k[E>>2]|0;u=K+120|0;q=k[u>>2]|0;B=k[u+4>>2]|0;u=K+952|0;K=yy(k[u>>2]|0,k[u+4>>2]|0,d|0,o|0)|0;u=ky(K|0,L|0,q|0,B|0)|0;B=J;q=ky(u|0,L|0,k[B>>2]|0,k[B+4>>2]|0)|0;B=L;u=l;k[u>>2]=q;k[u+4>>2]=B;au(a,q,B,1,h)|0;M=d;y=o;p=0;v=0;n=0;e=U}b:do if((Q|0)==35)while(1){Q=0;if((fu(a,1,0,z,h)|0)==107){k[h>>2]=0;i[z>>0]=0}U=l;au(a,k[U>>2]|0,k[U+4>>2]|0,1,h)|0;if((P|0)<(f|0)&(O|0)<8){U=P-f|0;e=O+-8|0;n=U>>>0>e>>>0?U:e;e=O-n|0;U=O;v=P;p=i[z>>0]|0;while(1){if(!(i[g+v>>0]|0))V=i[771137+U>>0]&p;else V=i[771129+U>>0]|p;U=U+1|0;if((U|0)==(e|0)){W=V;break}else{v=v+1|0;p=V}}i[z>>0]=W;X=P-n|0}else X=P;cu(a,1,0,z,h)|0;if((X|0)==(f|0))break b;p=l;v=ky(k[p>>2]|0,k[p+4>>2]|0,1,0)|0;p=l;k[p>>2]=v;k[p+4>>2]=L;O=0;P=X;Q=35}while(0);D=k[h>>2]|0;r=j;return D|0}function Dr(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0.0;m=r;r=r+29024|0;n=m+28864|0;o=m+28856|0;q=m+28896|0;s=m+28892|0;t=m+28888|0;u=m+28884|0;v=m+28880|0;w=m+48|0;x=m+40|0;y=m+32|0;z=m+24|0;A=m+16|0;B=m+8|0;C=m;D=m+28920|0;E=m+28940|0;F=m+28900|0;G=m+56|0;H=k[l>>2]|0;if((H|0)>0){I=H;r=m;return I|0}H=k[a>>2]|0;J=a+4|0;K=k[J>>2]|0;if((H|0)==(k[K+68>>2]|0)){M=K+120|0;if(((k[M>>2]|0)==-1?(k[M+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){I=k[l>>2]|0;r=m;return I|0}}else Fk(a,H+1|0,0,l)|0;if((b|0)>=1?(H=k[J>>2]|0,(k[H+928>>2]|0)>=(b|0)):0){J=k[(k[H+960>>2]|0)+((b+-1|0)*160|0)+80>>2]|0;k[q>>2]=J;switch(J|0){case -16:{J=k[j>>2]|0;if(!(i[J>>0]|0)){N=1;O=0}else{H=xx(J)|0;N=H;O=((H|0)<0)<<31>>31}if((nl(a,b,c,d,1,0,N,O,1,B,C,D,u,q,s,x,y,v,w,z,t,A,F,l)|0)>0){I=k[l>>2]|0;r=m;return I|0}H=x;au(a,k[H>>2]|0,k[H+4>>2]|0,1,l)|0;cu(a,N,O,k[j>>2]|0,l)|0;O=k[l>>2]|0;if((O|0)<=0){I=O;r=m;return I|0}O=E;N=771189;H=O+57|0;do{i[O>>0]=i[N>>0]|0;O=O+1|0;N=N+1|0}while((O|0)<(H|0));tk(E);I=k[l>>2]|0;r=m;return I|0}case 16:{if((nl(a,b,c,d,e,f,g,h,1,B,C,D,u,q,s,x,y,v,w,z,t,A,F,l)|0)>0){I=k[l>>2]|0;r=m;return I|0}F=k[u>>2]|0;if((F|0)>2880){k[s>>2]=1;k[v>>2]=F;F=w;k[F>>2]=1;k[F+4>>2]=0}if((g|0)==0&(h|0)==0){I=k[l>>2]|0;r=m;return I|0}else{P=0;Q=0;R=0;S=0;T=g;U=h}a:while(1){h=R;g=S;F=U;A=T;while(1){t=k[s>>2]|0;q=((t|0)<0)<<31>>31;D=(F|0)<(q|0)|(F|0)==(q|0)&A>>>0<t>>>0;C=D?A:t;D=my(0,C|0,32)|0;C=L;B=w;f=k[B>>2]|0;e=k[B+4>>2]|0;B=y;d=k[B>>2]|0;c=k[B+4>>2]|0;B=jy(f|0,e|0,d|0,c|0)|0;N=L;O=(C|0)<(N|0)|(C|0)==(N|0)&D>>>0<B>>>0;N=O?D:B;B=x;D=k[B>>2]|0;O=k[B+4>>2]|0;B=z;C=yy(k[B>>2]|0,k[B+4>>2]|0,P|0,Q|0)|0;B=ky(C|0,L|0,D|0,O|0)|0;O=L;D=k[v>>2]|0;C=yy(D|0,((D|0)<0)<<31>>31|0,d|0,c|0)|0;D=ky(B|0,O|0,C|0,L|0)|0;au(a,D,L,1,l)|0;if((N|0)>0){D=ky(d|0,c|0,-1,-1)|0;c=jy(D|0,L|0,f|0,e|0)|0;e=L;f=(q|0)>(F|0)|(q|0)==(F|0)&t>>>0>A>>>0;q=f?A:t;t=my(0,q|0,32)|0;q=~t;t=~L;f=(e|0)>(t|0)|(e|0)==(t|0)&c>>>0>q>>>0;t=f?c:q;q=~t;t=q+-1|0;c=k[u>>2]|0;f=h;e=g;D=G;d=0;while(1){b:do if((c|0)>0){C=c;O=k[j+(f<<2)>>2]|0;B=D;H=0;while(1){J=i[O>>0]|0;if(!(J<<24>>24)){V=C;W=B;X=H;break b}i[B>>0]=J;J=B+1|0;M=H+1|0;K=k[u>>2]|0;if((M|0)<(K|0)){C=K;O=O+1|0;B=J;H=M}else{V=K;W=J;X=M;break}}}else{V=c;W=D;X=0}while(0);if((X|0)<(V|0)){H=W;B=X;while(1){i[H>>0]=32;O=H+1|0;B=B+1|0;C=k[u>>2]|0;if((B|0)>=(C|0)){Y=C;Z=O;break}else H=O}}else{Y=V;Z=W}f=ky(f|0,e|0,1,0)|0;d=d+1|0;if((d|0)==(q|0)){_=Y;break}else{c=Y;e=L;D=Z}}D=ky(h|0,g|0,1,0)|0;e=ky(D|0,L|0,t|0,0)|0;$=_;aa=e;ba=L}else{$=k[u>>2]|0;aa=h;ba=g}e=k[v>>2]|0;if((e|0)==($|0)){D=ha($,N)|0;cu(a,D,((D|0)<0)<<31>>31,G,l)|0}else eu(a,$,N,e-$|0,G,l)|0;e=k[l>>2]|0;if((e|0)>0){ca=aa;da=ba;ea=N;break a}D=((N|0)<0)<<31>>31;c=jy(A|0,F|0,N|0,D|0)|0;q=L;if((A|0)==(N|0)&(F|0)==(D|0)){I=e;fa=44;break a}e=y;d=ky(k[e>>2]|0,k[e+4>>2]|0,N|0,D|0)|0;D=L;e=y;k[e>>2]=d;k[e+4>>2]=D;e=w;if((d|0)==(k[e>>2]|0)?(D|0)==(k[e+4>>2]|0):0){ga=c;ia=q;ja=aa;ka=ba;break}else{h=aa;g=ba;F=q;A=c}}A=y;k[A>>2]=0;k[A+4>>2]=0;A=ky(P|0,Q|0,1,0)|0;P=A;Q=L;R=ja;S=ka;T=ga;U=ia}if((fa|0)==44){r=m;return I|0}fa=ky(ca|0,da|0,1,0)|0;la=+(fa>>>0)+4294967296.0*+(L|0);fa=ky(ca|0,da|0,ea|0,((ea|0)<0)<<31>>31|0)|0;p[n>>3]=la;p[n+8>>3]=+(fa>>>0)+4294967296.0*+(L|0);Xw(E,771246,n)|0;tk(E);I=k[l>>2]|0;r=m;return I|0}default:{k[l>>2]=309;I=309;r=m;return I|0}}}k[o>>2]=b;Xw(E,771145,o)|0;tk(E);k[l>>2]=302;I=302;r=m;return I|0}function Er(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0;m=r;r=r+16|0;n=m+4|0;o=m;p=k[l>>2]|0;if((p|0)>0){q=p;r=m;return q|0}p=k[a>>2]|0;s=a+4|0;t=k[s>>2]|0;if((p|0)==(k[t+68>>2]|0)){u=t+120|0;if(((k[u>>2]|0)==-1?(k[u+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){q=k[l>>2]|0;r=m;return q|0}}else Fk(a,p+1|0,0,l)|0;Xk(a,b,0,n,o,l)|0;p=k[n>>2]|0;if((k[(k[s>>2]|0)+72>>2]|0)==2){s=(p|0)/(k[o>>2]|0)|0;k[n>>2]=s;v=s}else v=p;p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,((v|0)<0)<<31>>31|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;do if((h|0)>0|(h|0)==0&g>>>0>0){e=0;d=0;v=0;c=0;s=0;o=0;while(1){if(!(tx(j,k[i+(e<<2)>>2]|0)|0)){if(!((s|0)==0&(o|0)==0)?(u=jy(e|0,c|0,s|0,o|0)|0,t=ky(u|0,L|0,p|0,f|0)|0,w=L,x=ky(t|0,w|0,-1,-1)|0,y=k[n>>2]|0,z=((y|0)<0)<<31>>31,A=wy(x|0,L|0,y|0,z|0)|0,x=L,B=ky(A|0,x|0,1,0)|0,C=L,D=yy(A|0,x|0,y|0,z|0)|0,z=jy(t|0,w|0,D|0,L|0)|0,(Dr(a,b,B,C,z,L,s,o,i+(u<<2)|0,l)|0)>0):0){E=17;break}u=ky(d|0,v|0,1,0)|0;F=0;G=0;H=u;I=L}else{if(!((d|0)==0&(v|0)==0)?(u=jy(e|0,c|0,d|0,v|0)|0,z=ky(u|0,L|0,p|0,f|0)|0,u=L,C=ky(z|0,u|0,-1,-1)|0,B=k[n>>2]|0,D=((B|0)<0)<<31>>31,w=wy(C|0,L|0,B|0,D|0)|0,C=L,t=ky(w|0,C|0,1,0)|0,y=L,x=yy(w|0,C|0,B|0,D|0)|0,D=jy(z|0,u|0,x|0,L|0)|0,(Kr(a,b,t,y,D,L,d,v,l)|0)>0):0){E=13;break}D=ky(s|0,o|0,1,0)|0;F=D;G=L;H=0;I=0}D=ky(e|0,c|0,1,0)|0;y=L;if((y|0)<(h|0)|(y|0)==(h|0)&D>>>0<g>>>0){e=D;d=H;v=I;c=y;s=F;o=G}else{J=F;K=G;M=D;N=y;O=H;P=I;E=20;break}}if((E|0)==13){q=k[l>>2]|0;r=m;return q|0}else if((E|0)==17){q=k[l>>2]|0;r=m;return q|0}else if((E|0)==20){if(!((J|0)==0&(K|0)==0)){o=jy(M|0,N|0,J|0,K|0)|0;s=ky(o|0,L|0,p|0,f|0)|0;c=L;v=ky(s|0,c|0,-1,-1)|0;d=k[n>>2]|0;e=((d|0)<0)<<31>>31;y=wy(v|0,L|0,d|0,e|0)|0;v=L;D=ky(y|0,v|0,1,0)|0;t=L;x=yy(y|0,v|0,d|0,e|0)|0;e=jy(s|0,c|0,x|0,L|0)|0;Dr(a,b,D,t,e,L,J,K,i+(o<<2)|0,l)|0;break}if((O|0)==0&(P|0)==0)break;o=jy(M|0,N|0,O|0,P|0)|0;e=ky(o|0,L|0,p|0,f|0)|0;o=L;t=ky(e|0,o|0,-1,-1)|0;D=k[n>>2]|0;x=((D|0)<0)<<31>>31;c=wy(t|0,L|0,D|0,x|0)|0;t=L;s=ky(c|0,t|0,1,0)|0;d=L;v=yy(c|0,t|0,D|0,x|0)|0;x=jy(e|0,o|0,v|0,L|0)|0;Kr(a,b,s,d,x,L,O,P,l)|0;break}}while(0);q=k[l>>2]|0;r=m;return q|0}function Fr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;Gr(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,12,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function Gr(a,b,c,d,e,f,g,h,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;n=n|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0,ea=0,fa=0.0,ja=0.0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0;q=r;r=r+29056|0;s=q+28864|0;t=q+28856|0;u=q+28896|0;v=q+28892|0;w=q+28888|0;x=q+28884|0;y=q+28880|0;z=q+48|0;A=q+40|0;B=q+32|0;C=q+24|0;D=q+8|0;E=q;F=q+28940|0;G=q+28920|0;H=q+28960|0;I=q+56|0;J=k[n>>2]|0;if((J|0)>0){K=J;r=q;return K|0}if((nl(a,b,c,d,e,f,g,h,1,D,E,F,x,u,v,A,B,y,z,C,w,q+16|0,q+28900|0,n)|0)>0){K=k[n>>2]|0;r=q;return K|0}if((k[u>>2]|0)==16)Rk(F,G);f=G+1|0;e=g;g=h;h=0;d=0;c=0;J=0;a:while(1){M=h;N=d;O=e;P=g;while(1){Q=m+M|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=94;break a}U=k[v>>2]|0;X=((U|0)<0)<<31>>31;Y=(S|0)<(X|0)|(S|0)==(X|0)&R>>>0<U>>>0;X=Y?R:U;U=my(0,X|0,32)|0;X=L;Y=z;Z=B;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=jy(k[Y>>2]|0,k[Y+4>>2]|0,_|0,$|0)|0;Y=L;aa=(X|0)<(Y|0)|(X|0)==(Y|0)&U>>>0<Z>>>0;Y=aa?U:Z;Z=A;U=k[Z>>2]|0;aa=k[Z+4>>2]|0;Z=C;X=yy(k[Z>>2]|0,k[Z+4>>2]|0,c|0,J|0)|0;Z=ky(X|0,L|0,U|0,aa|0)|0;aa=L;U=k[y>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,_|0,$|0)|0;$=ky(Z|0,aa|0,X|0,L|0)|0;au(a,$,L,1,n)|0;b:do switch(k[u>>2]|0){case 11:{ba=+p[D>>3];ca=+p[E>>3];$=ba==1.0;c:do if($&ca==-128.0){if((Y|0)>0){X=0;do{i[I+X>>0]=l[m+(X+M)>>0]^128;X=X+1|0}while((X|0)!=(Y|0))}}else{X=(Y|0)>0;if($&ca==0.0){if(X)da=0;else break;while(1){aa=i[m+(da+M)>>0]|0;if(aa<<24>>24<0){k[n>>2]=-11;ea=0}else ea=aa;i[I+da>>0]=ea;da=da+1|0;if((da|0)==(Y|0))break c}}if(X){aa=0;do{fa=(+(i[m+(aa+M)>>0]|0)-ca)/ba;do if(!(fa<-.49))if(fa>255.49){k[n>>2]=-11;i[I+aa>>0]=-1;break}else{i[I+aa>>0]=~~(fa+.5);break}else{k[n>>2]=-11;i[I+aa>>0]=0}while(0);aa=aa+1|0}while((aa|0)!=(Y|0))}}while(0);su(a,Y,k[y>>2]|0,I,n)|0;break}case 81:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){aa=0;do{X=i[m+(aa+M)>>0]|0;Z=I+(aa<<3)|0;k[Z>>2]=X;k[Z+4>>2]=((X|0)<0)<<31>>31;aa=aa+1|0}while((aa|0)!=(Y|0))}}else if($){aa=0;do{fa=(+(i[m+(aa+M)>>0]|0)-ca)/ba;do if(!(fa<-9223372036854775808.0)){if(fa>9223372036854775808.0){k[n>>2]=-11;X=I+(aa<<3)|0;k[X>>2]=-1;k[X+4>>2]=2147483647;break}if(!(fa>=0.0)){ja=fa+-.5;X=+W(ja)>=1.0?(ja>0.0?~~+ia(+V(ja/4294967296.0),4294967295.0)>>>0:~~+ga((ja-+(~~ja>>>0))/4294967296.0)>>>0):0;Z=I+(aa<<3)|0;k[Z>>2]=~~ja>>>0;k[Z+4>>2]=X;break}else{ja=fa+.5;X=+W(ja)>=1.0?(ja>0.0?~~+ia(+V(ja/4294967296.0),4294967295.0)>>>0:~~+ga((ja-+(~~ja>>>0))/4294967296.0)>>>0):0;Z=I+(aa<<3)|0;k[Z>>2]=~~ja>>>0;k[Z+4>>2]=X;break}}else{k[n>>2]=-11;X=I+(aa<<3)|0;k[X>>2]=0;k[X+4>>2]=-2147483648}while(0);aa=aa+1|0}while((aa|0)!=(Y|0))}vu(a,Y,k[y>>2]|0,I,n)|0;break}case 21:{ba=+p[D>>3];ca=+p[E>>3];aa=(Y|0)>0;if(ba==1.0&ca==0.0){if(aa){$=0;do{j[I+($<<1)>>1]=i[m+($+M)>>0]|0;$=$+1|0}while(($|0)!=(Y|0))}}else if(aa){$=0;do{fa=(+(i[m+($+M)>>0]|0)-ca)/ba;do if(!(fa<-32768.49)){if(fa>32767.49){k[n>>2]=-11;j[I+($<<1)>>1]=32767;break}if(!(fa>=0.0)){j[I+($<<1)>>1]=~~(fa+-.5);break}else{j[I+($<<1)>>1]=~~(fa+.5);break}}else{k[n>>2]=-11;j[I+($<<1)>>1]=-32768}while(0);$=$+1|0}while(($|0)!=(Y|0))}tu(a,Y,k[y>>2]|0,I,n)|0;break}case 41:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){aa=0;do{k[I+(aa<<2)>>2]=i[m+(aa+M)>>0];aa=aa+1|0}while((aa|0)!=(Y|0))}}else if($){aa=0;do{fa=(+(i[m+(aa+M)>>0]|0)-ca)/ba;do if(!(fa<-2147483648.49)){if(fa>2147483647.49){k[n>>2]=-11;k[I+(aa<<2)>>2]=2147483647;break}if(!(fa>=0.0)){k[I+(aa<<2)>>2]=~~(fa+-.5);break}else{k[I+(aa<<2)>>2]=~~(fa+.5);break}}else{k[n>>2]=-11;k[I+(aa<<2)>>2]=-2147483648}while(0);aa=aa+1|0}while((aa|0)!=(Y|0))}uu(a,Y,k[y>>2]|0,I,n)|0;break}case 42:{ba=+p[D>>3];ca=+p[E>>3];aa=(Y|0)>0;if(ba==1.0&ca==0.0){if(aa){$=0;do{o[I+($<<2)>>2]=+(i[m+($+M)>>0]|0);$=$+1|0}while(($|0)!=(Y|0))}}else if(aa){$=0;do{o[I+($<<2)>>2]=(+(i[m+($+M)>>0]|0)-ca)/ba;$=$+1|0}while(($|0)!=(Y|0))}wu(a,Y,k[y>>2]|0,I,n)|0;break}case 82:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){aa=0;do{p[I+(aa<<3)>>3]=+(i[m+(aa+M)>>0]|0);aa=aa+1|0}while((aa|0)!=(Y|0))}}else if($){aa=0;do{p[I+(aa<<3)>>3]=(+(i[m+(aa+M)>>0]|0)-ca)/ba;aa=aa+1|0}while((aa|0)!=(Y|0))}xu(a,Y,k[y>>2]|0,I,n)|0;break}case 16:{if(rx(F,65)|0){aa=k[y>>2]|0;$=k[x>>2]|0;if((aa|0)==($|0)){cu(a,Y,((Y|0)<0)<<31>>31,Q,n)|0;break b}else{eu(a,$,(Y|0)/($|0)|0,aa-$|0,Q,n)|0;break b}}if((i[f>>0]|0)==115){T=86;break a}Jr(Q,Y,+p[D>>3],+p[E>>3],G,k[x>>2]|0,I,n)|0;$=k[y>>2]|0;aa=k[x>>2]|0;if(($|0)==(aa|0)){X=ha($,Y)|0;cu(a,X,((X|0)<0)<<31>>31,I,n)|0;break b}else{eu(a,aa,Y,$-aa|0,I,n)|0;break b}break}default:{T=86;break a}}while(0);if((k[n>>2]|0)>0){ka=M;la=N;ma=Y;T=90;break a}aa=((Y|0)<0)<<31>>31;$=jy(R|0,S|0,Y|0,aa|0)|0;X=L;if((R|0)==(Y|0)&(S|0)==(aa|0)){R=$;S=X}else{na=Y;oa=aa;pa=$;qa=X;break}}S=ky(na|0,oa|0,M|0,N|0)|0;R=L;Q=B;X=ky(k[Q>>2]|0,k[Q+4>>2]|0,na|0,oa|0)|0;Q=L;$=B;k[$>>2]=X;k[$+4>>2]=Q;$=z;if((X|0)==(k[$>>2]|0)?(Q|0)==(k[$+4>>2]|0):0){ra=S;sa=R;ta=pa;ua=qa;break}else{M=S;N=R;O=pa;P=qa}}P=B;k[P>>2]=0;k[P+4>>2]=0;P=ky(c|0,J|0,1,0)|0;e=ta;g=ua;h=ra;d=sa;c=P;J=L}if((T|0)==86){k[t>>2]=b;k[t+4>>2]=F;Xw(H,771690,t)|0;tk(H);if((k[w>>2]|0)==1){k[n>>2]=311;K=311;r=q;return K|0}else{k[n>>2]=312;K=312;r=q;return K|0}}else if((T|0)==90){w=ky(ka|0,la|0,1,0)|0;ba=+(w>>>0)+4294967296.0*+(L|0);w=ky(ma|0,((ma|0)<0)<<31>>31|0,ka|0,la|0)|0;p[s>>3]=ba;p[s+8>>3]=+(w>>>0)+4294967296.0*+(L|0);Xw(H,771314,s)|0;tk(H);K=k[n>>2]|0;r=q;return K|0}else if((T|0)==94){T=k[n>>2]|0;if((T|0)!=-11){K=T;r=q;return K|0}tk(771813);k[n>>2]=412;K=412;r=q;return K|0}return 0}function Hr(a,b,c,d,e,f,g,h,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,j)|0)){n=(b|0)<1;Ir(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,j)|0;o=k[j>>2]|0;r=l;return o|0}else{i[m>>0]=h;jt(a,12,c,d,e,f,1,g,m,j)|0;o=k[j>>2]|0;r=l;return o|0}return 0}function Ir(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[n+(p*160|0)+80>>2]|0;q=(r|0)>0;do if(!q){s=ky(e|0,f|0,-1,-1)|0;t=ky(s|0,L|0,g|0,h|0)|0;s=L;if((r|0)<0?(Gr(a,b,c,d,e,f,g,h,j,m)|0)>0:0){u=k[m>>2]|0;if((u|0)==412){k[m>>2]=0;v=t;w=s;break}else{o=u;return o|0}}else{v=t;w=s}}else{s=n+(p*160|0)+88|0;v=k[s>>2]|0;w=k[s+4>>2]|0}while(0);p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,w|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){o=k[m>>2]|0;return o|0}a:do if(q){e=0;d=0;c=0;n=0;r=0;s=0;t=0;while(1){if((i[j+e>>0]|0)==l<<24>>24){if(!((r|0)==0&(s|0)==0)?(u=jy(e|0,n|0,r|0,s|0)|0,x=ky(u|0,L|0,p|0,f|0)|0,y=L,z=ky(x|0,y|0,-1,-1)|0,A=wy(z|0,L|0,v|0,w|0)|0,z=L,B=yy(A|0,z|0,v|0,w|0)|0,C=jy(x|0,y|0,B|0,L|0)|0,B=L,y=ky(A|0,z|0,1,0)|0,(Gr(a,b,y,L,C,B,r,s,j+u|0,m)|0)>0):0){u=k[m>>2]|0;if((u|0)!=412){o=u;break}k[m>>2]=0;D=1}else D=t;u=ky(d|0,c|0,1,0)|0;E=0;F=0;G=u;H=L;I=D}else{if(!((d|0)==0&(c|0)==0)?(u=jy(e|0,n|0,d|0,c|0)|0,B=ky(u|0,L|0,p|0,f|0)|0,u=L,C=ky(B|0,u|0,-1,-1)|0,y=wy(C|0,L|0,v|0,w|0)|0,C=L,z=ky(y|0,C|0,1,0)|0,A=L,x=yy(y|0,C|0,v|0,w|0)|0,C=jy(B|0,u|0,x|0,L|0)|0,(Kr(a,b,z,A,C,L,d,c,m)|0)>0):0){J=29;break a}C=ky(r|0,s|0,1,0)|0;E=C;F=L;G=0;H=0;I=t}C=ky(e|0,n|0,1,0)|0;A=L;if((A|0)<(h|0)|(A|0)==(h|0)&C>>>0<g>>>0){e=C;d=G;c=H;n=A;r=E;s=F;t=I}else{K=E;M=F;N=C;O=A;P=G;Q=H;R=I;break a}}return o|0}else{t=0;s=0;r=0;n=0;c=0;d=0;while(1){if((i[j+t>>0]|0)==l<<24>>24){e=ky(s|0,r|0,1,0)|0;S=0;T=0;U=e;V=L}else{if(!((s|0)==0&(r|0)==0)?(e=jy(t|0,n|0,s|0,r|0)|0,A=ky(e|0,L|0,p|0,f|0)|0,e=L,C=ky(A|0,e|0,-1,-1)|0,z=wy(C|0,L|0,v|0,w|0)|0,C=L,x=ky(z|0,C|0,1,0)|0,u=L,B=yy(z|0,C|0,v|0,w|0)|0,C=jy(A|0,e|0,B|0,L|0)|0,(Kr(a,b,x,u,C,L,s,r,m)|0)>0):0){J=29;break a}C=ky(c|0,d|0,1,0)|0;S=C;T=L;U=0;V=0}C=ky(t|0,n|0,1,0)|0;u=L;if((u|0)<(h|0)|(u|0)==(h|0)&C>>>0<g>>>0){t=C;s=U;r=V;n=u;c=S;d=T}else{K=S;M=T;N=C;O=u;P=U;Q=V;R=0;break}}}while(0);if((J|0)==29){o=k[m>>2]|0;return o|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){J=jy(N|0,O|0,P|0,Q|0)|0;V=ky(J|0,L|0,p|0,f|0)|0;J=L;U=ky(V|0,J|0,-1,-1)|0;T=wy(U|0,L|0,v|0,w|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,v|0,w|0)|0;U=jy(V|0,J|0,h|0,L|0)|0;Kr(a,b,S,g,U,L,P,Q,m)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,p|0,f|0)|0;f=L;p=ky(O|0,f|0,-1,-1)|0;N=wy(p|0,L|0,v|0,w|0)|0;p=L;if(q){q=yy(N|0,p|0,v|0,w|0)|0;w=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,p|0,1,0)|0;Gr(a,b,f,L,w,q,K,M,j+Q|0,m)|0}}Q=k[m>>2]|0;if(!((R|0)!=0&(Q|0)<1)){o=Q;return o|0}k[m>>2]=412;o=412;return o|0}function Jr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){o=g;q=0;do{p[m>>3]=+(i[a+q>>0]|0);Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}}else if(n){n=g;q=0;do{p[l>>3]=(+(i[a+q>>0]|0)-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}b=rx(g,44)|0;if(!b){s=k[h>>2]|0;r=j;return s|0}else t=b;do{i[t>>0]=46;t=rx(t,44)|0}while((t|0)!=0);s=k[h>>2]|0;r=j;return s|0}
function Kr(a,b,c,d,e,f,g,h,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0.0;m=r;r=r+256|0;n=m+80|0;o=m+72|0;q=m+116|0;s=m+108|0;t=m+120|0;u=m+104|0;v=m+100|0;w=m+96|0;x=m+64|0;y=m+56|0;z=m+48|0;A=m+40|0;B=m+32|0;C=m+24|0;D=m+165|0;E=m+164|0;F=m+144|0;G=m+166|0;H=m+124|0;I=m;i[E>>0]=0;J=I;k[J>>2]=-1;k[J+4>>2]=-1;J=k[l>>2]|0;if((J|0)>0){K=J;r=m;return K|0}Xk(a,b,q,0,0,l)|0;J=k[q>>2]|0;if((((J|0)>-1?J:0-J|0)|0)>82){M=ty(e|0,f|0,1)|0;N=ky(M|0,L|0,-1,-1)|0;M=L;O=ty(g|0,h|0,1)|0;P=N;Q=M;R=O;S=L}else{P=e;Q=f;R=g;S=h}if((nl(a,b,c,d,P,Q,R,S,(J>>31)+2&-2,m+16|0,m+8|0,F,v,q,m+112|0,A,B,w,z,C,s,x,H,l)|0)>0){K=k[l>>2]|0;r=m;return K|0}J=k[q>>2]|0;a:do switch(J|0){case 16:{if((i[H>>0]|0)==1){tk(771383);k[l>>2]=314;K=314;r=m;return K|0}Q=k[v>>2]|0;P=(Q|0)<20?20:Q;Q=cy(P)|0;if(Q){sy(Q|0,32,P|0)|0;P=xx(H)|0;Bx(Q,H,((k[s>>2]|0)==2&1)+P|0)|0;T=Q;break a}k[l>>2]=113;K=113;r=m;return K|0}case 11:case 21:case 41:case 81:{Q=x;P=k[Q>>2]|0;d=k[Q+4>>2]|0;if((P|0)==1234554321&(d|0)==0){tk(771449);k[l>>2]=314;K=314;r=m;return K|0}switch(J|0){case 11:{i[D>>0]=P;T=0;break a;break}case 21:{j[t>>1]=P;Ks(t,1);T=0;break a;break}case 41:{k[u>>2]=P;Ls(u,1);T=0;break a;break}default:{Q=y;k[Q>>2]=P;k[Q+4>>2]=d;Ms(y,1);T=0;break a}}break}default:T=0}while(0);J=R;R=S;S=0;x=0;s=0;H=0;b:while(1){d=S;Q=x;P=J;c=R;while(1){h=P;g=c;while(1){if((h|0)==0&(g|0)==0){U=47;break b}f=z;e=B;O=k[e>>2]|0;M=k[e+4>>2]|0;e=jy(k[f>>2]|0,k[f+4>>2]|0,O|0,M|0)|0;f=L;N=(g|0)<(f|0)|(g|0)==(f|0)&h>>>0<e>>>0;V=N?h:e;e=N?g:f;f=A;N=k[f>>2]|0;W=k[f+4>>2]|0;f=C;X=yy(k[f>>2]|0,k[f+4>>2]|0,s|0,H|0)|0;f=ky(X|0,L|0,N|0,W|0)|0;W=L;N=k[w>>2]|0;X=yy(N|0,((N|0)<0)<<31>>31|0,O|0,M|0)|0;M=ky(f|0,W|0,X|0,L|0)|0;au(a,M,L,1,l)|0;switch(k[q>>2]|0){case 11:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,1,0,D,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 21:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,2,0,t,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 41:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,4,0,u,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 81:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,8,0,y,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 42:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,4,0,I,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 82:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,8,0,I,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 14:{if((e|0)>0|(e|0)==0&V>>>0>0){M=0;X=0;do{cu(a,1,0,E,l)|0;M=ky(M|0,X|0,1,0)|0;X=L}while((X|0)<(e|0)|(X|0)==(e|0)&M>>>0<V>>>0)}break}case 16:{M=k[v>>2]|0;cu(a,M,((M|0)<0)<<31>>31,T,l)|0;break}default:{U=39;break b}}if((k[l>>2]|0)>0){Y=d;Z=Q;_=V;$=e;U=41;break b}M=jy(h|0,g|0,V|0,e|0)|0;X=L;if((h|0)==(V|0)&(g|0)==(e|0)){h=M;g=X}else{aa=V;ba=e;ca=M;da=X;break}}g=ky(aa|0,ba|0,d|0,Q|0)|0;h=L;X=B;M=ky(k[X>>2]|0,k[X+4>>2]|0,aa|0,ba|0)|0;X=L;W=B;k[W>>2]=M;k[W+4>>2]=X;W=z;if((M|0)==(k[W>>2]|0)?(X|0)==(k[W+4>>2]|0):0){ea=g;fa=h;ga=ca;ha=da;break}else{d=g;Q=h;P=ca;c=da}}c=B;k[c>>2]=0;k[c+4>>2]=0;c=ky(s|0,H|0,1,0)|0;J=ga;R=ha;S=ea;x=fa;s=c;H=L}if((U|0)==39){k[o>>2]=b;k[o+4>>2]=F;Xw(G,771510,o)|0;tk(G);K=k[l>>2]|0;r=m;return K|0}else if((U|0)==41){o=ky(Y|0,Z|0,1,0)|0;ia=+(o>>>0)+4294967296.0*+(L|0);o=ky(_|0,$|0,Y|0,Z|0)|0;p[n>>3]=ia;p[n+8>>3]=+(o>>>0)+4294967296.0*+(L|0);Xw(G,771567,n)|0;tk(G);if(T)dy(T);K=k[l>>2]|0;r=m;return K|0}else if((U|0)==47){if(T)dy(T);K=k[l>>2]|0;r=m;return K|0}return 0}function Lr(a,b,c,d,e,f,g,h,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0.0;m=r;r=r+256|0;n=m+80|0;o=m+72|0;q=m+116|0;s=m+108|0;t=m+120|0;u=m+104|0;v=m+100|0;w=m+96|0;x=m+64|0;y=m+56|0;z=m+48|0;A=m+40|0;B=m+32|0;C=m+24|0;D=m+165|0;E=m+164|0;F=m+144|0;G=m+166|0;H=m+124|0;I=m;i[E>>0]=0;J=I;k[J>>2]=-1;k[J+4>>2]=-1;J=k[l>>2]|0;if((J|0)>0){K=J;r=m;return K|0}Xk(a,b,q,0,0,l)|0;if((nl(a,b,c,d,e,f,g,h,(k[q>>2]>>31)+2&-2,m+16|0,m+8|0,F,v,q,m+112|0,A,B,w,z,C,s,x,H,l)|0)>0){K=k[l>>2]|0;r=m;return K|0}f=k[q>>2]|0;a:do switch(f|0){case 16:{if((i[H>>0]|0)==1){tk(771383);k[l>>2]=314;K=314;r=m;return K|0}e=k[v>>2]|0;d=(e|0)<20?20:e;e=cy(d)|0;if(e){sy(e|0,32,d|0)|0;d=xx(H)|0;Bx(e,H,((k[s>>2]|0)==2&1)+d|0)|0;M=e;break a}k[l>>2]=113;K=113;r=m;return K|0}case 11:case 21:case 41:case 81:{e=x;d=k[e>>2]|0;c=k[e+4>>2]|0;if((d|0)==1234554321&(c|0)==0){tk(771449);k[l>>2]=314;K=314;r=m;return K|0}switch(f|0){case 11:{i[D>>0]=d;M=0;break a;break}case 21:{j[t>>1]=d;Ks(t,1);M=0;break a;break}case 41:{k[u>>2]=d;Ls(u,1);M=0;break a;break}default:{e=y;k[e>>2]=d;k[e+4>>2]=c;Ls(y,2);M=0;break a}}break}default:M=0}while(0);f=g;g=h;h=0;x=0;s=0;H=0;b:while(1){c=h;e=x;d=f;J=g;while(1){N=d;O=J;while(1){if((N|0)==0&(O|0)==0){P=45;break b}Q=z;R=B;S=k[R>>2]|0;T=k[R+4>>2]|0;R=jy(k[Q>>2]|0,k[Q+4>>2]|0,S|0,T|0)|0;Q=L;U=(O|0)<(Q|0)|(O|0)==(Q|0)&N>>>0<R>>>0;V=U?N:R;R=U?O:Q;Q=A;U=k[Q>>2]|0;W=k[Q+4>>2]|0;Q=C;X=yy(k[Q>>2]|0,k[Q+4>>2]|0,s|0,H|0)|0;Q=ky(X|0,L|0,U|0,W|0)|0;W=L;U=k[w>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,S|0,T|0)|0;T=ky(Q|0,W|0,X|0,L|0)|0;au(a,T,L,1,l)|0;switch(k[q>>2]|0){case 11:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,1,0,D,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 21:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,2,0,t,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 41:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,4,0,u,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 81:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,8,0,y,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 42:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,4,0,I,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 82:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,8,0,I,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 14:{if((R|0)>0|(R|0)==0&V>>>0>0){T=0;X=0;do{cu(a,1,0,E,l)|0;T=ky(T|0,X|0,1,0)|0;X=L}while((X|0)<(R|0)|(X|0)==(R|0)&T>>>0<V>>>0)}break}case 16:{T=k[v>>2]|0;cu(a,T,((T|0)<0)<<31>>31,M,l)|0;break}default:{P=37;break b}}if((k[l>>2]|0)>0){Y=c;Z=e;_=V;$=R;P=39;break b}T=jy(N|0,O|0,V|0,R|0)|0;X=L;if((N|0)==(V|0)&(O|0)==(R|0)){N=T;O=X}else{aa=V;ba=R;ca=T;da=X;break}}O=ky(aa|0,ba|0,c|0,e|0)|0;N=L;X=B;T=ky(k[X>>2]|0,k[X+4>>2]|0,aa|0,ba|0)|0;X=L;W=B;k[W>>2]=T;k[W+4>>2]=X;W=z;if((T|0)==(k[W>>2]|0)?(X|0)==(k[W+4>>2]|0):0){ea=O;fa=N;ga=ca;ha=da;break}else{c=O;e=N;d=ca;J=da}}J=B;k[J>>2]=0;k[J+4>>2]=0;J=ky(s|0,H|0,1,0)|0;f=ga;g=ha;h=ea;x=fa;s=J;H=L}if((P|0)==37){k[o>>2]=b;k[o+4>>2]=F;Xw(G,771510,o)|0;tk(G);K=k[l>>2]|0;r=m;return K|0}else if((P|0)==39){o=ky(Y|0,Z|0,1,0)|0;ia=+(o>>>0)+4294967296.0*+(L|0);o=ky(_|0,$|0,Y|0,Z|0)|0;p[n>>3]=ia;p[n+8>>3]=+(o>>>0)+4294967296.0*+(L|0);Xw(G,771567,n)|0;tk(G);if(M)dy(M);K=k[l>>2]|0;r=m;return K|0}else if((P|0)==45){if(M)dy(M);K=k[l>>2]|0;r=m;return K|0}return 0}function Mr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;Nr(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,20,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function Nr(a,b,c,d,e,f,g,h,l,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;n=n|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0;q=r;r=r+29056|0;s=q+28864|0;t=q+28856|0;u=q+28896|0;v=q+28892|0;w=q+28888|0;x=q+28884|0;y=q+28880|0;z=q+48|0;A=q+40|0;B=q+32|0;C=q+24|0;D=q+8|0;E=q;F=q+28940|0;G=q+28920|0;H=q+28960|0;I=q+56|0;J=k[n>>2]|0;if((J|0)>0){K=J;r=q;return K|0}if((nl(a,b,c,d,e,f,g,h,1,D,E,F,x,u,v,A,B,y,z,C,w,q+16|0,q+28900|0,n)|0)>0){K=k[n>>2]|0;r=q;return K|0}if((k[u>>2]|0)==16)Rk(F,G);f=G+1|0;e=g;g=h;h=0;d=0;c=0;J=0;a:while(1){M=h;N=d;O=e;P=g;while(1){Q=l+(M<<1)|0;R=O;S=P;while(1){if((R|0)==0&(S|0)==0){T=76;break a}U=k[v>>2]|0;X=((U|0)<0)<<31>>31;Y=(S|0)<(X|0)|(S|0)==(X|0)&R>>>0<U>>>0;X=Y?R:U;U=my(0,X|0,32)|0;X=L;Y=z;Z=B;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=jy(k[Y>>2]|0,k[Y+4>>2]|0,_|0,$|0)|0;Y=L;aa=(X|0)<(Y|0)|(X|0)==(Y|0)&U>>>0<Z>>>0;Y=aa?U:Z;Z=A;U=k[Z>>2]|0;aa=k[Z+4>>2]|0;Z=C;X=yy(k[Z>>2]|0,k[Z+4>>2]|0,c|0,J|0)|0;Z=ky(X|0,L|0,U|0,aa|0)|0;aa=L;U=k[y>>2]|0;X=yy(U|0,((U|0)<0)<<31>>31|0,_|0,$|0)|0;$=ky(Z|0,aa|0,X|0,L|0)|0;au(a,$,L,1,n)|0;b:do switch(k[u>>2]|0){case 21:{Qr(Q,Y,+p[D>>3],+p[E>>3],I,n)|0;tu(a,Y,k[y>>2]|0,I,n)|0;break}case 81:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{aa=I+(X<<3)|0;k[aa>>2]=m[l+(X+M<<1)>>1];k[aa+4>>2]=0;X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+(m[l+(X+M<<1)>>1]|0)-ca)/ba;do if(!(da<-9223372036854775808.0)){if(da>9223372036854775808.0){k[n>>2]=-11;aa=I+(X<<3)|0;k[aa>>2]=-1;k[aa+4>>2]=2147483647;break}if(!(da>=0.0)){ea=da+-.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=I+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}else{ea=da+.5;aa=+W(ea)>=1.0?(ea>0.0?~~+ia(+V(ea/4294967296.0),4294967295.0)>>>0:~~+ga((ea-+(~~ea>>>0))/4294967296.0)>>>0):0;Z=I+(X<<3)|0;k[Z>>2]=~~ea>>>0;k[Z+4>>2]=aa;break}}else{k[n>>2]=-11;aa=I+(X<<3)|0;k[aa>>2]=0;k[aa+4>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}vu(a,Y,k[y>>2]|0,I,n)|0;break}case 11:{ba=+p[D>>3];ca=+p[E>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{aa=j[l+($+M<<1)>>1]|0;if((aa&65535)>255){k[n>>2]=-11;fa=-1}else fa=aa&255;i[I+$>>0]=fa;$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{da=(+(m[l+($+M<<1)>>1]|0)-ca)/ba;do if(!(da<-.49))if(da>255.49){k[n>>2]=-11;i[I+$>>0]=-1;break}else{i[I+$>>0]=~~(da+.5);break}else{k[n>>2]=-11;i[I+$>>0]=0}while(0);$=$+1|0}while(($|0)!=(Y|0))}su(a,Y,k[y>>2]|0,I,n)|0;break}case 41:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{k[I+(X<<2)>>2]=m[l+(X+M<<1)>>1];X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{da=(+(m[l+(X+M<<1)>>1]|0)-ca)/ba;do if(!(da<-2147483648.49)){if(da>2147483647.49){k[n>>2]=-11;k[I+(X<<2)>>2]=2147483647;break}if(!(da>=0.0)){k[I+(X<<2)>>2]=~~(da+-.5);break}else{k[I+(X<<2)>>2]=~~(da+.5);break}}else{k[n>>2]=-11;k[I+(X<<2)>>2]=-2147483648}while(0);X=X+1|0}while((X|0)!=(Y|0))}uu(a,Y,k[y>>2]|0,I,n)|0;break}case 42:{ba=+p[D>>3];ca=+p[E>>3];X=(Y|0)>0;if(ba==1.0&ca==0.0){if(X){$=0;do{o[I+($<<2)>>2]=+(m[l+($+M<<1)>>1]|0);$=$+1|0}while(($|0)!=(Y|0))}}else if(X){$=0;do{o[I+($<<2)>>2]=(+(m[l+($+M<<1)>>1]|0)-ca)/ba;$=$+1|0}while(($|0)!=(Y|0))}wu(a,Y,k[y>>2]|0,I,n)|0;break}case 82:{ba=+p[D>>3];ca=+p[E>>3];$=(Y|0)>0;if(ba==1.0&ca==0.0){if($){X=0;do{p[I+(X<<3)>>3]=+(m[l+(X+M<<1)>>1]|0);X=X+1|0}while((X|0)!=(Y|0))}}else if($){X=0;do{p[I+(X<<3)>>3]=(+(m[l+(X+M<<1)>>1]|0)-ca)/ba;X=X+1|0}while((X|0)!=(Y|0))}xu(a,Y,k[y>>2]|0,I,n)|0;break}case 16:{if((i[f>>0]|0)==115){T=68;break a}Rr(Q,Y,+p[D>>3],+p[E>>3],G,k[x>>2]|0,I,n)|0;X=k[y>>2]|0;$=k[x>>2]|0;if((X|0)==($|0)){aa=ha(X,Y)|0;cu(a,aa,((aa|0)<0)<<31>>31,I,n)|0;break b}else{eu(a,$,Y,X-$|0,I,n)|0;break b}break}default:{T=68;break a}}while(0);if((k[n>>2]|0)>0){ja=M;ka=N;la=Y;T=72;break a}$=((Y|0)<0)<<31>>31;X=jy(R|0,S|0,Y|0,$|0)|0;aa=L;if((R|0)==(Y|0)&(S|0)==($|0)){R=X;S=aa}else{ma=Y;na=$;oa=X;pa=aa;break}}S=ky(ma|0,na|0,M|0,N|0)|0;R=L;Q=B;aa=ky(k[Q>>2]|0,k[Q+4>>2]|0,ma|0,na|0)|0;Q=L;X=B;k[X>>2]=aa;k[X+4>>2]=Q;X=z;if((aa|0)==(k[X>>2]|0)?(Q|0)==(k[X+4>>2]|0):0){qa=S;ra=R;sa=oa;ta=pa;break}else{M=S;N=R;O=oa;P=pa}}P=B;k[P>>2]=0;k[P+4>>2]=0;P=ky(c|0,J|0,1,0)|0;e=sa;g=ta;h=qa;d=ra;c=P;J=L}if((T|0)==68){k[t>>2]=b;k[t+4>>2]=F;Xw(H,771690,t)|0;tk(H);if((k[w>>2]|0)==1){k[n>>2]=311;K=311;r=q;return K|0}else{k[n>>2]=312;K=312;r=q;return K|0}}else if((T|0)==72){w=ky(ja|0,ka|0,1,0)|0;ba=+(w>>>0)+4294967296.0*+(L|0);w=ky(la|0,((la|0)<0)<<31>>31|0,ja|0,ka|0)|0;p[s>>3]=ba;p[s+8>>3]=+(w>>>0)+4294967296.0*+(L|0);Xw(H,771621,s)|0;tk(H);K=k[n>>2]|0;r=q;return K|0}else if((T|0)==76){T=k[n>>2]|0;if((T|0)!=-11){K=T;r=q;return K|0}tk(771813);k[n>>2]=412;K=412;r=q;return K|0}return 0}function Or(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var l=0,m=0,n=0,o=0;l=r;r=r+16|0;m=l;if(!(Cl(a,i)|0)){n=(b|0)<1;Pr(a,2,n?1:b,n?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;o=k[i>>2]|0;r=l;return o|0}else{j[m>>1]=h;jt(a,20,c,d,e,f,1,g,m,i)|0;o=k[i>>2]|0;r=l;return o|0}return 0}function Pr(a,b,c,d,e,f,g,h,i,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;n=k[m>>2]|0;if((n|0)>0){o=n;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){r=q+120|0;if(((k[r>>2]|0)==-1?(k[r+4>>2]|0)==-1:0)?(Vk(a,m)|0)>0:0){o=k[m>>2]|0;return o|0}}else Fk(a,n+1|0,0,m)|0;n=k[(k[p>>2]|0)+960>>2]|0;p=b+-1|0;r=k[n+(p*160|0)+80>>2]|0;q=(r|0)>0;do if(!q){s=ky(e|0,f|0,-1,-1)|0;t=ky(s|0,L|0,g|0,h|0)|0;s=L;if((r|0)<0?(Nr(a,b,c,d,e,f,g,h,i,m)|0)>0:0){u=k[m>>2]|0;if((u|0)==412){k[m>>2]=0;v=t;w=s;break}else{o=u;return o|0}}else{v=t;w=s}}else{s=n+(p*160|0)+88|0;v=k[s>>2]|0;w=k[s+4>>2]|0}while(0);p=ky(c|0,d|0,-1,-1)|0;d=yy(v|0,w|0,p|0,L|0)|0;p=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){o=k[m>>2]|0;return o|0}a:do if(q){e=0;d=0;c=0;n=0;r=0;s=0;t=0;while(1){if((j[i+(e<<1)>>1]|0)==l<<16>>16){if(!((r|0)==0&(s|0)==0)?(u=jy(e|0,n|0,r|0,s|0)|0,x=ky(u|0,L|0,p|0,f|0)|0,y=L,z=ky(x|0,y|0,-1,-1)|0,A=wy(z|0,L|0,v|0,w|0)|0,z=L,B=yy(A|0,z|0,v|0,w|0)|0,C=jy(x|0,y|0,B|0,L|0)|0,B=L,y=ky(A|0,z|0,1,0)|0,(Nr(a,b,y,L,C,B,r,s,i+(u<<1)|0,m)|0)>0):0){u=k[m>>2]|0;if((u|0)!=412){o=u;break}k[m>>2]=0;D=1}else D=t;u=ky(d|0,c|0,1,0)|0;E=0;F=0;G=u;H=L;I=D}else{if(!((d|0)==0&(c|0)==0)?(u=jy(e|0,n|0,d|0,c|0)|0,B=ky(u|0,L|0,p|0,f|0)|0,u=L,C=ky(B|0,u|0,-1,-1)|0,y=wy(C|0,L|0,v|0,w|0)|0,C=L,z=ky(y|0,C|0,1,0)|0,A=L,x=yy(y|0,C|0,v|0,w|0)|0,C=jy(B|0,u|0,x|0,L|0)|0,(Kr(a,b,z,A,C,L,d,c,m)|0)>0):0){J=29;break a}C=ky(r|0,s|0,1,0)|0;E=C;F=L;G=0;H=0;I=t}C=ky(e|0,n|0,1,0)|0;A=L;if((A|0)<(h|0)|(A|0)==(h|0)&C>>>0<g>>>0){e=C;d=G;c=H;n=A;r=E;s=F;t=I}else{K=E;M=F;N=C;O=A;P=G;Q=H;R=I;break a}}return o|0}else{t=0;s=0;r=0;n=0;c=0;d=0;while(1){if((j[i+(t<<1)>>1]|0)==l<<16>>16){e=ky(s|0,r|0,1,0)|0;S=0;T=0;U=e;V=L}else{if(!((s|0)==0&(r|0)==0)?(e=jy(t|0,n|0,s|0,r|0)|0,A=ky(e|0,L|0,p|0,f|0)|0,e=L,C=ky(A|0,e|0,-1,-1)|0,z=wy(C|0,L|0,v|0,w|0)|0,C=L,x=ky(z|0,C|0,1,0)|0,u=L,B=yy(z|0,C|0,v|0,w|0)|0,C=jy(A|0,e|0,B|0,L|0)|0,(Kr(a,b,x,u,C,L,s,r,m)|0)>0):0){J=29;break a}C=ky(c|0,d|0,1,0)|0;S=C;T=L;U=0;V=0}C=ky(t|0,n|0,1,0)|0;u=L;if((u|0)<(h|0)|(u|0)==(h|0)&C>>>0<g>>>0){t=C;s=U;r=V;n=u;c=S;d=T}else{K=S;M=T;N=C;O=u;P=U;Q=V;R=0;break}}}while(0);if((J|0)==29){o=k[m>>2]|0;return o|0}if((K|0)==0&(M|0)==0){if(!((P|0)==0&(Q|0)==0)){J=jy(N|0,O|0,P|0,Q|0)|0;V=ky(J|0,L|0,p|0,f|0)|0;J=L;U=ky(V|0,J|0,-1,-1)|0;T=wy(U|0,L|0,v|0,w|0)|0;U=L;S=ky(T|0,U|0,1,0)|0;g=L;h=yy(T|0,U|0,v|0,w|0)|0;U=jy(V|0,J|0,h|0,L|0)|0;Kr(a,b,S,g,U,L,P,Q,m)|0}}else{Q=jy(N|0,O|0,K|0,M|0)|0;O=ky(Q|0,L|0,p|0,f|0)|0;f=L;p=ky(O|0,f|0,-1,-1)|0;N=wy(p|0,L|0,v|0,w|0)|0;p=L;if(q){q=yy(N|0,p|0,v|0,w|0)|0;w=jy(O|0,f|0,q|0,L|0)|0;q=L;f=ky(N|0,p|0,1,0)|0;Nr(a,b,f,L,w,q,K,M,i+(Q<<1)|0,m)|0}}Q=k[m>>2]|0;if(!((R|0)!=0&(Q|0)<1)){o=Q;return o|0}k[m>>2]=412;o=412;return o|0}function Qr(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;var g=0,h=0,i=0,l=0,n=0,o=0,p=0.0;g=c==1.0;if(g&d==32768.0){if((b|0)>0)h=0;else{i=k[f>>2]|0;return i|0}do{j[e+(h<<1)>>1]=(m[a+(h<<1)>>1]|0)^32768;h=h+1|0}while((h|0)!=(b|0));i=k[f>>2]|0;return i|0}h=(b|0)>0;if(g&d==0.0){if(h)l=0;else{i=k[f>>2]|0;return i|0}do{g=j[a+(l<<1)>>1]|0;if(g<<16>>16<0){k[f>>2]=-11;n=32767}else n=g;j[e+(l<<1)>>1]=n;l=l+1|0}while((l|0)!=(b|0));i=k[f>>2]|0;return i|0}if(h)o=0;else{i=k[f>>2]|0;return i|0}do{p=(+(m[a+(o<<1)>>1]|0)-d)/c;do if(!(p<-32768.49)){if(p>32767.49){k[f>>2]=-11;j[e+(o<<1)>>1]=32767;break}if(!(p>=0.0)){j[e+(o<<1)>>1]=~~(p+-.5);break}else{j[e+(o<<1)>>1]=~~(p+.5);break}}else{k[f>>2]=-11;j[e+(o<<1)>>1]=-32768}while(0);o=o+1|0}while((o|0)!=(b|0));i=k[f>>2]|0;return i|0}function Rr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,n=0,o=0,q=0,s=0,t=0,u=0;j=r;r=r+16|0;l=j+8|0;n=j;o=(b|0)>0;if(c==1.0&d==0.0){if(o){q=g;s=0;do{p[n>>3]=+(m[a+(s<<1)>>1]|0);Xw(q,e,n)|0;q=q+f|0;if(i[q>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}}else if(o){o=g;s=0;do{p[l>>3]=(+(m[a+(s<<1)>>1]|0)-d)/c;Xw(o,e,l)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;s=s+1|0}while((s|0)!=(b|0))}b=rx(g,44)|0;if(!b){t=k[h>>2]|0;r=j;return t|0}else u=b;do{i[u>>0]=46;u=rx(u,44)|0}while((u|0)!=0);t=k[h>>2]|0;r=j;return t|0}function Sr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0;i=r;r=r+16|0;if(!(Cl(a,h)|0)){j=(b|0)<1;Tr(a,2,j?1:b,j?0:((b|0)<0)<<31>>31,c,d,e,f,g,h)|0;l=k[h>>2]|0;r=i;return l|0}else{jt(a,40,c,d,e,f,0,g,i,h)|0;l=k[h>>2]|0;r=i;return l|0}return 0}function Tr(a,b,c,d,e,f,g,h,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;var n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,X=0,Y=0,Z=0,_=0,$=0,aa=0.0,ba=0.0,ca=0,da=0,ea=0.0,fa=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0;n=r;r=r+29056|0;q=n+28864|0;s=n+28856|0;t=n+28896|0;u=n+28892|0;v=n+28888|0;w=n+28884|0;x=n+28880|0;y=n+48|0;z=n+40|0;A=n+32|0;B=n+24|0;C=n+8|0;D=n;E=n+28940|0;F=n+28920|0;G=n+28960|0;H=n+56|0;I=k[m>>2]|0;if((I|0)>0){J=I;r=n;return J|0}if((nl(a,b,c,d,e,f,g,h,1,C,D,E,w,t,u,z,A,x,y,B,v,n+16|0,n+28900|0,m)|0)>0){J=k[m>>2]|0;r=n;return J|0}if((k[t>>2]|0)==16)Rk(E,F);f=F+1|0;e=g;g=h;h=0;d=0;c=0;I=0;a:while(1){K=h;M=d;N=e;O=g;while(1){P=l+(K<<2)|0;Q=N;R=O;while(1){if((Q|0)==0&(R|0)==0){S=96;break a}T=k[u>>2]|0;U=((T|0)<0)<<31>>31;X=(R|0)<(U|0)|(R|0)==(U|0)&Q>>>0<T>>>0;U=X?Q:T;T=my(0,U|0,32)|0;U=L;X=y;Y=A;Z=k[Y>>2]|0;_=k[Y+4>>2]|0;Y=jy(k[X>>2]|0,k[X+4>>2]|0,Z|0,_|0)|0;X=L;$=(U|0)<(X|0)|(U|0)==(X|0)&T>>>0<Y>>>0;X=$?T:Y;Y=z;T=k[Y>>2]|0;$=k[Y+4>>2]|0;Y=B;U=yy(k[Y>>2]|0,k[Y+4>>2]|0,c|0,I|0)|0;Y=ky(U|0,L|0,T|0,$|0)|0;$=L;T=k[x>>2]|0;U=yy(T|0,((T|0)<0)<<31>>31|0,Z|0,_|0)|0;_=ky(Y|0,$|0,U|0,L|0)|0;au(a,_,L,1,m)|0;b:do switch(k[t>>2]|0){case 41:{aa=+p[C>>3];ba=+p[D>>3];_=aa==1.0;c:do if(_&ba==2147483648.0){if((X|0)>0){U=0;do{k[H+(U<<2)>>2]=k[l+(U+K<<2)>>2]^-2147483648;U=U+1|0}while((U|0)!=(X|0))}}else{U=(X|0)>0;if(_&ba==0.0){if(U)ca=0;else break;while(1){$=k[l+(ca+K<<2)>>2]|0;if(($|0)<0){k[m>>2]=-11;da=2147483647}else da=$;k[H+(ca<<2)>>2]=da;ca=ca+1|0;if((ca|0)==(X|0))break c}}if(U){$=0;do{ea=(+((k[l+($+K<<2)>>2]|0)>>>0)-ba)/aa;do if(!(ea<-2147483648.49)){if(ea>2147483647.49){k[m>>2]=-11;k[H+($<<2)>>2]=2147483647;break}if(!(ea>=0.0)){k[H+($<<2)>>2]=~~(ea+-.5);break}else{k[H+($<<2)>>2]=~~(ea+.5);break}}else{k[m>>2]=-11;k[H+($<<2)>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(X|0))}}while(0);uu(a,X,k[x>>2]|0,H,m)|0;break}case 81:{aa=+p[C>>3];ba=+p[D>>3];_=(X|0)>0;if(aa==1.0&ba==0.0){if(_){$=0;do{U=H+($<<3)|0;k[U>>2]=k[l+($+K<<2)>>2];k[U+4>>2]=0;$=$+1|0}while(($|0)!=(X|0))}}else if(_){$=0;do{ea=(+((k[l+($+K<<2)>>2]|0)>>>0)-ba)/aa;do if(!(ea<-9223372036854775808.0)){if(ea>9223372036854775808.0){k[m>>2]=-11;U=H+($<<3)|0;k[U>>2]=-1;k[U+4>>2]=2147483647;break}if(!(ea>=0.0)){fa=ea+-.5;U=+W(fa)>=1.0?(fa>0.0?~~+ia(+V(fa/4294967296.0),4294967295.0)>>>0:~~+ga((fa-+(~~fa>>>0))/4294967296.0)>>>0):0;Y=H+($<<3)|0;k[Y>>2]=~~fa>>>0;k[Y+4>>2]=U;break}else{fa=ea+.5;U=+W(fa)>=1.0?(fa>0.0?~~+ia(+V(fa/4294967296.0),4294967295.0)>>>0:~~+ga((fa-+(~~fa>>>0))/4294967296.0)>>>0):0;Y=H+($<<3)|0;k[Y>>2]=~~fa>>>0;k[Y+4>>2]=U;break}}else{k[m>>2]=-11;U=H+($<<3)|0;k[U>>2]=0;k[U+4>>2]=-2147483648}while(0);$=$+1|0}while(($|0)!=(X|0))}vu(a,X,k[x>>2]|0,H,m)|0;break}case 11:{aa=+p[C>>3];ba=+p[D>>3];$=(X|0)>0;if(aa==1.0&ba==0.0){if($){_=0;do{U=k[l+(_+K<<2)>>2]|0;if(U>>>0>255){k[m>>2]=-11;ja=-1}else ja=U&255;i[H+_>>0]=ja;_=_+1|0}while((_|0)!=(X|0))}}else if($){_=0;do{ea=(+((k[l+(_+K<<2)>>2]|0)>>>0)-ba)/aa;do if(!(ea<-.49))if(ea>255.49){k[m>>2]=-11;i[H+_>>0]=-1;break}else{i[H+_>>0]=~~(ea+.5);break}else{k[m>>2]=-11;i[H+_>>0]=0}while(0);_=_+1|0}while((_|0)!=(X|0))}su(a,X,k[x>>2]|0,H,m)|0;break}case 21:{aa=+p[C>>3];ba=+p[D>>3];_=(X|0)>0;if(aa==1.0&ba==0.0){if(_){$=0;do{U=k[l+($+K<<2)>>2]|0;if(U>>>0>32767){k[m>>2]=-11;ka=32767}else ka=U&65535;j[H+($<<1)>>1]=ka;$=$+1|0}while(($|0)!=(X|0))}}else if(_){$=0;do{ea=(+((k[l+($+K<<2)>>2]|0)>>>0)-ba)/aa;do if(!(ea<-32768.49)){if(ea>32767.49){k[m>>2]=-11;j[H+($<<1)>>1]=32767;break}if(!(ea>=0.0)){j[H+($<<1)>>1]=~~(ea+-.5);break}else{j[H+($<<1)>>1]=~~(ea+.5);break}}else{k[m>>2]=-11;j[H+($<<1)>>1]=-32768}while(0);$=$+1|0}while(($|0)!=(X|0))}tu(a,X,k[x>>2]|0,H,m)|0;break}case 42:{aa=+p[C>>3];ba=+p[D>>3];$=(X|0)>0;if(aa==1.0&ba==0.0){if($){_=0;do{o[H+(_<<2)>>2]=+((k[l+(_+K<<2)>>2]|0)>>>0);_=_+1|0}while((_|0)!=(X|0))}}else if($){_=0;do{o[H+(_<<2)>>2]=(+((k[l+(_+K<<2)>>2]|0)>>>0)-ba)/aa;_=_+1|0}while((_|0)!=(X|0))}wu(a,X,k[x>>2]|0,H,m)|0;break}case 82:{aa=+p[C>>3];ba=+p[D>>3];_=(X|0)>0;if(aa==1.0&ba==0.0){if(_){$=0;do{p[H+($<<3)>>3]=+((k[l+($+K<<2)>>2]|0)>>>0);$=$+1|0}while(($|0)!=(X|0))}}else if(_){$=0;do{p[H+($<<3)>>3]=(+((k[l+($+K<<2)>>2]|0)>>>0)-ba)/aa;$=$+1|0}while(($|0)!=(X|0))}xu(a,X,k[x>>2]|0,H,m)|0;break}case 16:{if((i[f>>0]|0)==115){S=88;break a}Wr(P,X,+p[C>>3],+p[D>>3],F,k[w>>2]|0,H,m)|0;$=k[x>>2]|0;_=k[w>>2]|0;if(($|0)==(_|0)){U=ha($,X)|0;cu(a,U,((U|0)<0)<<31>>31,H,m)|0;break b}else{eu(a,_,X,$-_|0,H,m)|0;break b}break}default:{S=88;break a}}while(0);if((k[m>>2]|0)>0){la=K;ma=M;na=X;S=92;break a}_=((X|0)<0)<<31>>31;$=jy(Q|0,R|0,X|0,_|0)|0;U=L;if((Q|0)==(X|0)&(R|0)==(_|0)){Q=$;R=U}else{oa=X;pa=_;qa=$;ra=U;break}}R=ky(oa|0,pa|0,K|0,M|0)|0;Q=L;P=A;U=ky(k[P>>2]|0,k[P+4>>2]|0,oa|0,pa|0)|0;P=L;$=A;k[$>>2]=U;k[$+4>>2]=P;$=y;if((U|0)==(k[$>>2]|0)?(P|0)==(k[$+4>>2]|0):0){sa=R;ta=Q;ua=qa;va=ra;break}else{K=R;M=Q;N=qa;O=ra}}O=A;k[O>>2]=0;k[O+4>>2]=0;O=ky(c|0,I|0,1,0)|0;e=ua;g=va;h=sa;d=ta;c=O;I=L}if((S|0)==88){k[s>>2]=b;k[s+4>>2]=E;Xw(G,771690,s)|0;tk(G);if((k[v>>2]|0)==1){k[m>>2]=311;J=311;r=n;return J|0}else{k[m>>2]=312;J=312;r=n;return J|0}}else if((S|0)==92){v=ky(la|0,ma|0,1,0)|0;aa=+(v>>>0)+4294967296.0*+(L|0);v=ky(na|0,((na|0)<0)<<31>>31|0,la|0,ma|0)|0;p[q>>3]=aa;p[q+8>>3]=+(v>>>0)+4294967296.0*+(L|0);Xw(G,771744,q)|0;tk(G);J=k[m>>2]|0;r=n;return J|0}else if((S|0)==96){S=k[m>>2]|0;if((S|0)!=-11){J=S;r=n;return J|0}tk(771813);k[m>>2]=412;J=412;r=n;return J|0}return 0}function Ur(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0;j=r;r=r+16|0;l=j;if(!(Cl(a,i)|0)){m=(b|0)<1;Vr(a,2,m?1:b,m?0:((b|0)<0)<<31>>31,c,d,e,f,g,h,i)|0;n=k[i>>2]|0;r=j;return n|0}else{k[l>>2]=h;jt(a,40,c,d,e,f,1,g,l,i)|0;n=k[i>>2]|0;r=j;return n|0}return 0}function Vr(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0;m=k[l>>2]|0;if((m|0)>0){n=m;return n|0}m=k[a>>2]|0;o=a+4|0;p=k[o>>2]|0;if((m|0)==(k[p+68>>2]|0)){q=p+120|0;if(((k[q>>2]|0)==-1?(k[q+4>>2]|0)==-1:0)?(Vk(a,l)|0)>0:0){n=k[l>>2]|0;return n|0}}else Fk(a,m+1|0,0,l)|0;m=k[(k[o>>2]|0)+960>>2]|0;o=b+-1|0;q=k[m+(o*160|0)+80>>2]|0;p=(q|0)>0;do if(!p){r=ky(e|0,f|0,-1,-1)|0;s=ky(r|0,L|0,g|0,h|0)|0;r=L;if((q|0)<0?(Tr(a,b,c,d,e,f,g,h,i,l)|0)>0:0){t=k[l>>2]|0;if((t|0)==412){k[l>>2]=0;u=s;v=r;break}else{n=t;return n|0}}else{u=s;v=r}}else{r=m+(o*160|0)+88|0;u=k[r>>2]|0;v=k[r+4>>2]|0}while(0);o=ky(c|0,d|0,-1,-1)|0;d=yy(u|0,v|0,o|0,L|0)|0;o=ky(d|0,L|0,e|0,f|0)|0;f=L;if(!((h|0)>0|(h|0)==0&g>>>0>0)){n=k[l>>2]|0;return n|0}a:do if(p){e=0;d=0;c=0;m=0;q=0;r=0;s=0;while(1){if((k[i+(e<<2)>>2]|0)==(j|0)){if(!((q|0)==0&(r|0)==0)?(t=jy(e|0,m|0,q|0,r|0)|0,w=ky(t|0,L|0,o|0,f|0)|0,x=L,y=ky(w|0,x|0,-1,-1)|0,z=wy(y|0,L|0,u|0,v|0)|0,y=L,A=yy(z|0,y|0,u|0,v|0)|0,B=jy(w|0,x|0,A|0,L|0)|0,A=L,x=ky(z|0,y|0,1,0)|0,(Tr(a,b,x,L,B,A,q,r,i+(t<<2)|0,l)|0)>0):0){t=k[l>>2]|0;if((t|0)!=412){n=t;break}k[l>>2]=0;C=1}else C=s;t=ky(d|0,c|0,1,0)|0;D=0;E=0;F=t;G=L;H=C}else{if(!((d|0)==0&(c|0)==0)?(t=jy(e|0,m|0,d|0,c|0)|0,A=ky(t|0,L|0,o|0,f|0)|0,t=L,B=ky(A|0,t|0,-1,-1)|0,x=wy(B|0,L|0,u|0,v|0)|0,B=L,y=ky(x|0,B|0,1,0)|0,z=L,w=yy(x|0,B|0,u|0,v|0)|0,B=jy(A|0,t|0,w|0,L|0)|0,(Kr(a,b,y,z,B,L,d,c,l)|0)>0):0){I=29;break a}B=ky(q|0,r|0,1,0)|0;D=B;E=L;F=0;G=0;H=s}B=ky(e|0,m|0,1,0)|0;z=L;if((z|0)<(h|0)|(z|0)==(h|0)&B>>>0<g>>>0){e=B;d=F;c=G;m=z;q=D;r=E;s=H}else{J=D;K=E;M=B;N=z;O=F;P=G;Q=H;break a}}return n|0}else{s=0;r=0;q=0;m=0;c=0;d=0;while(1){if((k[i+(s<<2)>>2]|0)==(j|0)){e=ky(r|0,q|0,1,0)|0;R=0;S=0;T=e;U=L}else{if(!((r|0)==0&(q|0)==0)?(e=jy(s|0,m|0,r|0,q|0)|0,z=ky(e|0,L|0,o|0,f|0)|0,e=L,B=ky(z|0,e|0,-1,-1)|0,y=wy(B|0,L|0,u|0,v|0)|0,B=L,w=ky(y|0,B|0,1,0)|0,t=L,A=yy(y|0,B|0,u|0,v|0)|0,B=jy(z|0,e|0,A|0,L|0)|0,(Kr(a,b,w,t,B,L,r,q,l)|0)>0):0){I=29;break a}B=ky(c|0,d|0,1,0)|0;R=B;S=L;T=0;U=0}B=ky(s|0,m|0,1,0)|0;t=L;if((t|0)<(h|0)|(t|0)==(h|0)&B>>>0<g>>>0){s=B;r=T;q=U;m=t;c=R;d=S}else{J=R;K=S;M=B;N=t;O=T;P=U;Q=0;break}}}while(0);if((I|0)==29){n=k[l>>2]|0;return n|0}if((J|0)==0&(K|0)==0){if(!((O|0)==0&(P|0)==0)){I=jy(M|0,N|0,O|0,P|0)|0;U=ky(I|0,L|0,o|0,f|0)|0;I=L;T=ky(U|0,I|0,-1,-1)|0;S=wy(T|0,L|0,u|0,v|0)|0;T=L;R=ky(S|0,T|0,1,0)|0;g=L;h=yy(S|0,T|0,u|0,v|0)|0;T=jy(U|0,I|0,h|0,L|0)|0;Kr(a,b,R,g,T,L,O,P,l)|0}}else{P=jy(M|0,N|0,J|0,K|0)|0;N=ky(P|0,L|0,o|0,f|0)|0;f=L;o=ky(N|0,f|0,-1,-1)|0;M=wy(o|0,L|0,u|0,v|0)|0;o=L;if(p){p=yy(M|0,o|0,u|0,v|0)|0;v=jy(N|0,f|0,p|0,L|0)|0;p=L;f=ky(M|0,o|0,1,0)|0;Tr(a,b,f,L,v,p,J,K,i+(P<<2)|0,l)|0}}P=k[l>>2]|0;if(!((Q|0)!=0&(P|0)<1)){n=P;return n|0}k[l>>2]=412;n=412;return n|0}function Wr(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=+c;d=+d;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0;j=r;r=r+16|0;l=j+8|0;m=j;n=(b|0)>0;if(c==1.0&d==0.0){if(n){o=g;q=0;do{p[m>>3]=+((k[a+(q<<2)>>2]|0)>>>0);Xw(o,e,m)|0;o=o+f|0;if(i[o>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}}else if(n){n=g;q=0;do{p[l>>3]=(+((k[a+(q<<2)>>2]|0)>>>0)-d)/c;Xw(n,e,l)|0;n=n+f|0;if(i[n>>0]|0)k[h>>2]=-11;q=q+1|0}while((q|0)!=(b|0))}b=rx(g,44)|0;if(!b){s=k[h>>2]|0;r=j;return s|0}else t=b;do{i[t>>0]=46;t=rx(t,44)|0}while((t|0)!=0);s=k[h>>2]|0;r=j;return s|0}function Xr(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0;f=r;r=r+160|0;g=f;h=k[e>>2]|0;if((h|0)>0){i=h;r=f;return i|0}h=k[a>>2]|0;j=a+4|0;l=k[j>>2]|0;if((h|0)==(k[l+68>>2]|0)){m=l;n=h}else{Fk(a,h+1|0,0,e)|0;h=k[j>>2]|0;m=h;n=k[h+68>>2]|0}h=m+96|0;j=(k[m+88>>2]|0)+(n<<3)|0;if(!((k[h>>2]|0)==(k[j>>2]|0)?(k[h+4>>2]|0)==(k[j+4>>2]|0):0))Al(a,e)|0;if((c|0)>0){j=(0-c|0)>>>0>4294967276?c:20;h=0;do{n=k[d+(h<<2)>>2]|0;m=g+(h<<3)|0;k[m>>2]=n;k[m+4>>2]=((n|0)<0)<<31>>31;h=h+1|0}while((h|0)!=(j|0))}_r(a,1,b,c,g,0,0,1,0,1,e)|0;i=k[e>>2]|0;r=f;return i|0}function Yr(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0;m=r;r=r+160|0;n=m;if((d|0)>0){o=(0-d|0)>>>0>4294967276?d:20;p=0;do{q=k[e+(p<<2)>>2]|0;s=n+(p<<3)|0;k[s>>2]=q;k[s+4>>2]=((q|0)<0)<<31>>31;p=p+1|0}while((p|0)!=(o|0))}_r(a,b,c,d,n,f,g,h,i,j,l)|0;r=m;return k[l>>2]|0}function Zr(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}f=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((f|0)==(k[i+68>>2]|0)){j=i;l=f}else{Fk(a,f+1|0,0,e)|0;f=k[h>>2]|0;j=f;l=k[f+68>>2]|0}f=j+96|0;h=(k[j+88>>2]|0)+(l<<3)|0;if(!((k[f>>2]|0)==(k[h>>2]|0)?(k[f+4>>2]|0)==(k[h+4>>2]|0):0))Al(a,e)|0;_r(a,1,b,c,d,0,0,1,0,1,e)|0;g=k[e>>2]|0;return g|0}function _r(a,b,c,d,e,f,g,h,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;l=l|0;m=m|0;n=n|0;var o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0;o=r;r=r+656|0;q=o+176|0;s=o+160|0;t=o+144|0;u=o+128|0;v=o+112|0;w=o+96|0;x=o+88|0;y=o+80|0;z=o+72|0;A=o+64|0;B=o+56|0;C=o+48|0;D=o+32|0;E=o+24|0;F=o+16|0;G=o+8|0;H=o;I=o+272|0;J=o+572|0;K=o+192|0;L=o+497|0;M=o+424|0;N=o+343|0;O=k[n>>2]|0;if((O|0)>0){P=O;r=o;return P|0}O=k[a>>2]|0;Q=a+4|0;R=k[Q>>2]|0;if((O|0)==(k[R+68>>2]|0)){S=R;T=O}else{Fk(a,O+1|0,0,n)|0;O=k[Q>>2]|0;S=O;T=k[O+68>>2]|0}O=S+96|0;R=(k[S+88>>2]|0)+(T<<3)|0;if(!((k[O>>2]|0)==(k[R>>2]|0)?(k[O+4>>2]|0)==(k[R+4>>2]|0):0)){k[n>>2]=201;P=201;r=o;return P|0}if((d|0)!=0?(k[S+984>>2]|0)!=0:0){if((d|0)>0){S=0;do{k[K+(S<<2)>>2]=k[e+(S<<3)>>2];S=S+1|0}while((S|0)!=(d|0))}Zs(a,c,d,K,1,n)|0;P=k[n>>2]|0;r=o;return P|0}if(!T){T=(b|0)==0;if(T){U=M;V=771880;W=U+39|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0))}else{U=M;V=778844;W=U+35|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0))}if((k[n>>2]|0)<=0){j[I>>1]=T?70:84;Dk(782785,I,M,J,n)|0;ds(a,J,n)|0}}else{U=M;V=771919;W=U+16|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));fs(a,782702,771935,M,n)|0}T=(c|0)==20;a:do if(!T)if((c|0)==40)X=32;else{b=(c|0)==10?8:c;switch(b|0){case 8:case 16:case 32:case 64:case -64:case -32:{X=b;break a;break}default:{}}k[H>>2]=c;Xw(N,771941,H)|0;tk(N);k[n>>2]=211;P=211;r=o;return P|0}else X=16;while(0);U=M;V=771978;W=U+30|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));H=k[n>>2]|0;if((H|0)>0){P=H;r=o;return P|0}i[I>>0]=0;H=G;k[H>>2]=X;k[H+4>>2]=((X|0)<0)<<31>>31;if((Xw(I,772008,G)|0)<0){tk(772013);k[n>>2]=401}Dk(779656,I,M,J,n)|0;ds(a,J,n)|0;G=k[n>>2]|0;if((G|0)>0){P=G;r=o;return P|0}if(d>>>0>999){k[F>>2]=d;Xw(N,772057,F)|0;tk(N);k[n>>2]=212;P=212;r=o;return P|0}U=M;V=772093;W=U+20|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));i[I>>0]=0;F=E;k[F>>2]=d;k[F+4>>2]=((d|0)<0)<<31>>31;if((Xw(I,772008,E)|0)<0){tk(772013);k[n>>2]=401}Dk(782713,I,M,J,n)|0;ds(a,J,n)|0;U=M;V=772113;W=U+21|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));b:do if((d|0)>0){E=M+20|0;F=0;while(1){G=e+(F<<3)|0;X=G;H=k[X+4>>2]|0;F=F+1|0;if((H|0)<0){Y=F;Z=k[X>>2]|0;_=H;break}k[C>>2]=F;Xw(E,772183,C)|0;Gk(782713,F,L,n)|0;H=G;G=k[H>>2]|0;X=k[H+4>>2]|0;if((k[n>>2]|0)<=0){i[I>>0]=0;H=B;k[H>>2]=G;k[H+4>>2]=X;if((Xw(I,772008,B)|0)<0){tk(772013);k[n>>2]=401}Dk(L,I,M,J,n)|0;ds(a,J,n)|0}if((F|0)>=(d|0))break b}k[D>>2]=Y;p[D+8>>3]=+(Z>>>0)+4294967296.0*+(_|0);Xw(N,772134,D)|0;tk(N);k[n>>2]=213;P=213;r=o;return P|0}while(0);do if(k[(k[Q>>2]|0)+68>>2]|0){if(!((f|0)==0&(g|0)==0)){tk(772537);k[n>>2]=214;break}if(!((h|0)==1&(l|0)==0)){tk(772575);k[n>>2]=215;break}U=M;V=772613;W=U+27|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));N=k[n>>2]|0;if((N|0)>0)$=N;else{i[I>>0]=0;N=y;k[N>>2]=0;k[N+4>>2]=0;if((Xw(I,772008,y)|0)<0){tk(772013);k[n>>2]=401}Dk(782652,I,M,J,n)|0;ds(a,J,n)|0;$=k[n>>2]|0}U=M;V=772640;W=U+27|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if(($|0)<=0){i[I>>0]=0;N=x;k[N>>2]=1;k[N+4>>2]=0;if((Xw(I,772008,x)|0)<0){tk(772013);k[n>>2]=401}Dk(782726,I,M,J,n)|0;ds(a,J,n)|0}}else{if(m){U=M;V=772186;W=U+36|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((k[n>>2]|0)<=0){j[I>>1]=84;Dk(782808,I,M,J,n)|0;ds(a,J,n)|0}}if((g|0)<0){tk(772222);k[n>>2]=214;P=214;r=o;return P|0}if((l|0)<0|(l|0)==0&h>>>0<1){tk(772250);k[n>>2]=215;P=215;r=o;return P|0}if((g|0)>0|(g|0)==0&f>>>0>0|((l|0)>0|(l|0)==0&h>>>0>1)){U=M;V=772278;W=U+33|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));N=k[n>>2]|0;if((N|0)>0)aa=N;else{j[I>>1]=84;Dk(772311,I,M,J,n)|0;ds(a,J,n)|0;aa=k[n>>2]|0}U=M;V=772318;W=U+34|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((aa|0)>0)ba=aa;else{i[I>>0]=0;N=A;k[N>>2]=f;k[N+4>>2]=g;if((Xw(I,772008,A)|0)<0){tk(772013);k[n>>2]=401}Dk(782652,I,M,J,n)|0;ds(a,J,n)|0;ba=k[n>>2]|0}U=M;V=772352;W=U+24|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((ba|0)<=0){i[I>>0]=0;N=z;k[N>>2]=h;k[N+4>>2]=l;if((Xw(I,772008,z)|0)<0){tk(772013);k[n>>2]=401}Dk(782726,I,M,J,n)|0;ds(a,J,n)|0}}ds(a,772376,n)|0;ds(a,772457,n)|0}while(0);c:do if(T){U=M;V=779463;W=U+44|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));z=k[n>>2]|0;if((z|0)>0)ca=z;else{i[I>>0]=0;k[w>>2]=0;p[w+8>>3]=32768.0;if((Xw(I,772667,w)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779507,I,M,J,n)|0;ds(a,J,n)|0;ca=k[n>>2]|0}U=M;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((ca|0)<=0){i[I>>0]=0;k[v>>2]=0;p[v+8>>3]=1.0;if((Xw(I,772667,v)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779536,I,M,J,n)|0;ds(a,J,n)|0}}else switch(c|0){case 40:{U=M;V=779584;W=U+43|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));z=k[n>>2]|0;if((z|0)>0)da=z;else{i[I>>0]=0;k[u>>2]=0;p[u+8>>3]=2147483648.0;if((Xw(I,772667,u)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779507,I,M,J,n)|0;ds(a,J,n)|0;da=k[n>>2]|0}U=M;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((da|0)<=0){i[I>>0]=0;k[t>>2]=0;p[t+8>>3]=1.0;if((Xw(I,772667,t)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779536,I,M,J,n)|0;ds(a,J,n)|0}break c;break}case 10:{U=M;V=779543;W=U+41|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));z=k[n>>2]|0;if((z|0)>0)ea=z;else{i[I>>0]=0;k[s>>2]=0;p[s+8>>3]=-128.0;if((Xw(I,772667,s)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779507,I,M,J,n)|0;ds(a,J,n)|0;ea=k[n>>2]|0}U=M;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));if((ea|0)<=0){i[I>>0]=0;k[q>>2]=0;p[q+8>>3]=1.0;if((Xw(I,772667,q)|0)<0){tk(772672);k[n>>2]=402}z=rx(I,44)|0;if(z)i[z>>0]=46;if(rx(I,78)|0){tk(772715);k[n>>2]=402}Dk(779536,I,M,J,n)|0;ds(a,J,n)|0}break c;break}default:break c}while(0);P=k[n>>2]|0;r=o;return P|0}function $r(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;l=r;r=r+160|0;m=l;n=k[j>>2]|0;if((n|0)>0){o=n;r=l;return o|0}n=k[a>>2]|0;p=a+4|0;q=k[p>>2]|0;if((n|0)==(k[q+68>>2]|0)){s=q;t=n}else{Fk(a,n+1|0,0,j)|0;n=k[p>>2]|0;s=n;t=k[n+68>>2]|0}n=s+96|0;q=(k[s+88>>2]|0)+(t<<3)|0;if((k[n>>2]|0)==(k[q>>2]|0)?(k[n+4>>2]|0)==(k[q+4>>2]|0):0){u=t;v=s}else{Al(a,j)|0;s=k[p>>2]|0;u=k[s+68>>2]|0;v=s}if(!u){if((k[j>>2]|0)<=0){u=k[a>>2]|0;if(!u){w=v;x=0}else{Fk(a,u+1|0,0,j)|0;u=k[p>>2]|0;w=u;x=k[u+68>>2]|0}u=w+96|0;p=(k[w+88>>2]|0)+(x<<3)|0;if(!((k[u>>2]|0)==(k[p>>2]|0)?(k[u+4>>2]|0)==(k[p+4>>2]|0):0))Al(a,j)|0;_r(a,1,16,0,m,0,0,1,0,1,j)|0}Al(a,j)|0}switch(b|0){case 2:{as(a,c,d,e,f,g,h,i,0,0,j)|0;break}case 1:{bs(a,0,0,c,d,e,f,0,g,h,i,j)|0;break}default:k[j>>2]=235}o=k[j>>2]|0;r=l;return o|0}function as(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0;j=r;r=r+592|0;n=j+152|0;o=j+136|0;q=j+120|0;s=j+104|0;t=j+88|0;u=j+72|0;v=j+64|0;w=j+56|0;x=j+48|0;y=j+40|0;z=j+32|0;A=j+24|0;B=j+16|0;C=j+8|0;D=j;E=j+510|0;F=j+429|0;G=j+176|0;H=j+172|0;I=j+168|0;J=j+180|0;K=j+354|0;M=j+281|0;N=j+210|0;O=k[m>>2]|0;if((O|0)>0){P=O;r=j;return P|0}O=k[a>>2]|0;Q=a+4|0;R=k[Q>>2]|0;if((O|0)==(k[R+68>>2]|0)){S=R;T=O}else{Fk(a,O+1|0,0,m)|0;O=k[Q>>2]|0;S=O;T=k[O+68>>2]|0}O=S+96|0;Q=(k[S+88>>2]|0)+(T<<3)|0;if(!((k[O>>2]|0)==(k[Q>>2]|0)?(k[O+4>>2]|0)==(k[Q+4>>2]|0):0)){k[m>>2]=201;P=201;r=j;return P|0}if((c|0)<0){k[m>>2]=218;P=218;r=j;return P|0}if((l|0)<0){k[m>>2]=214;P=214;r=j;return P|0}if(d>>>0>999){k[m>>2]=216;P=216;r=j;return P|0}i[N>>0]=0;if(h)zx(N,h,70)|0;fs(a,782702,772762,772771,m)|0;h=k[m>>2]|0;if((h|0)>0)U=h;else{i[E>>0]=0;h=D;k[h>>2]=8;k[h+4>>2]=0;if((Xw(E,772008,D)|0)<0){tk(772013);k[m>>2]=401}Dk(779656,E,772794,F,m)|0;ds(a,F,m)|0;U=k[m>>2]|0}if((U|0)<=0){i[E>>0]=0;U=C;k[U>>2]=2;k[U+4>>2]=0;if((Xw(E,772008,C)|0)<0){tk(772013);k[m>>2]=401}Dk(782713,E,772806,F,m)|0;ds(a,F,m)|0}do if((d|0)>0){C=0;U=0;D=0;while(1){h=f+(D<<2)|0;Qk(k[h>>2]|0,G,H,I,m)|0;l=k[G>>2]|0;a:do switch(l|0){case 16:{Q=k[H>>2]|0;V=Q;W=((Q|0)<0)<<31>>31;break}case 1:{Q=((k[H>>2]|0)+7|0)/8|0;V=Q;W=((Q|0)<0)<<31>>31;break}default:{if((l|0)>0){Q=ha((l|0)/10|0,k[H>>2]|0)|0;V=Q;W=((Q|0)<0)<<31>>31;break a}Q=k[h>>2]|0;if((i[Q>>0]|0)==80){V=8;W=0}else{O=(i[Q+1>>0]|0)==80;V=O?8:16;W=O?0:0}}}while(0);h=ky(V|0,W|0,C|0,U|0)|0;l=L;O=k[m>>2]|0;D=D+1|0;if((O|0)>0){X=O;Y=33;break}if((D|0)>=(d|0)){Z=h;_=l;Y=24;break}else{C=h;U=l}}if((Y|0)==24){$=Z;aa=_;Y=35;break}else if((Y|0)==33){ba=X;break}}else{U=k[m>>2]|0;if((U|0)>0)ba=U;else{$=0;aa=0;Y=35}}while(0);if((Y|0)==35){i[E>>0]=0;Y=B;k[Y>>2]=$;k[Y+4>>2]=aa;if((Xw(E,772008,B)|0)<0){tk(772013);k[m>>2]=401}Dk(772833,E,772840,F,m)|0;ds(a,F,m)|0;ba=k[m>>2]|0}if((ba|0)>0)ca=ba;else{i[E>>0]=0;ba=A;k[ba>>2]=b;k[ba+4>>2]=c;if((Xw(E,772008,A)|0)<0){tk(772013);k[m>>2]=401}Dk(790243,E,772864,F,m)|0;ds(a,F,m)|0;ca=k[m>>2]|0}if((ca|0)>0)da=ca;else{i[E>>0]=0;ca=z;k[ca>>2]=0;k[ca+4>>2]=0;if((Xw(E,772008,z)|0)<0){tk(772013);k[m>>2]=401}Dk(782652,E,772888,F,m)|0;ds(a,F,m)|0;da=k[m>>2]|0}if((da|0)>0)ea=da;else{i[E>>0]=0;da=y;k[da>>2]=1;k[da+4>>2]=0;if((Xw(E,772008,y)|0)<0){tk(772013);k[m>>2]=401}Dk(782726,E,772914,F,m)|0;ds(a,F,m)|0;ea=k[m>>2]|0}if((ea|0)<=0){i[E>>0]=0;ea=x;k[ea>>2]=d;k[ea+4>>2]=((d|0)<0)<<31>>31;if((Xw(E,772008,x)|0)<0){tk(772013);k[m>>2]=401}Dk(782733,E,772948,F,m)|0;ds(a,F,m)|0}x=(g|0)==0;ea=0;do{if((ea|0)>=(d|0))break;y=e+(ea<<2)|0;da=ea;ea=ea+1|0;if(i[k[y>>2]>>0]|0){k[w>>2]=ea;Xw(M,772977,w)|0;Gk(772997,ea,K,m)|0;fs(a,K,k[y>>2]|0,M,m)|0}ux(J,k[f+(da<<2)>>2]|0)|0;Ck(J);Gk(773003,ea,K,m)|0;y=M;z=773009;ca=y+21|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));Qk(J,G,H,I,m)|0;A=k[G>>2]|0;b:do switch(A|0){case 16:{y=M+(xx(M)|0)|0;z=773030;ca=y+18|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));c=(rx(J,65)|0)+1|0;k[v>>2]=I;if((Yw(c,773048,v)|0)==1?(c=k[H>>2]|0,(k[I>>2]|0)>(c|0)):0)if((c|0)==1){y=M;z=773052;ca=y+47|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break b}else{y=M;z=773099;ca=y+48|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break b}break}case 1:{c=M+(xx(M)|0)|0;i[c>>0]=i[773147]|0;i[c+1>>0]=i[773148]|0;i[c+2>>0]=i[773149]|0;i[c+3>>0]=i[773150]|0;i[c+4>>0]=i[773151]|0;i[c+5>>0]=i[773152]|0;break}case 11:{c=M+(xx(M)|0)|0;i[c>>0]=i[773153]|0;i[c+1>>0]=i[773154]|0;i[c+2>>0]=i[773155]|0;i[c+3>>0]=i[773156]|0;i[c+4>>0]=i[773157]|0;i[c+5>>0]=i[773158]|0;i[c+6>>0]=i[773159]|0;break}case 14:{y=M+(xx(M)|0)|0;z=773160;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 21:{y=M+(xx(M)|0)|0;z=773177;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 20:{y=M+(xx(M)|0)|0;z=773177;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 41:{y=M+(xx(M)|0)|0;z=773194;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 81:{y=M+(xx(M)|0)|0;z=773211;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 40:{y=M+(xx(M)|0)|0;z=773194;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 42:{y=M+(xx(M)|0)|0;z=773228;ca=y+14|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 82:{y=M+(xx(M)|0)|0;z=773242;ca=y+16|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 83:{y=M+(xx(M)|0)|0;z=773258;ca=y+10|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}case 163:{y=M+(xx(M)|0)|0;z=773268;ca=y+17|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));break}default:if((A|0)<0){y=M+(xx(M)|0)|0;z=773285;ca=y+24|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0))}}while(0);A=k[G>>2]|0;switch(((A|0)>-1?A:0-A|0)|0){case 12:{A=J;while(1)if((i[A>>0]|0)==83){fa=A;break}else A=A+1|0;i[fa>>0]=66;fs(a,K,J,M,m)|0;Gk(773309,ea,K,m)|0;y=M;z=773315;ca=y+24|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[u>>2]=0;p[u+8>>3]=-128.0;if((Xw(E,772667,u)|0)<0){tk(772672);k[m>>2]=402}A=rx(E,44)|0;if(A)i[A>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}Gk(773339,ea,K,m)|0;y=M;z=773345;ca=y+20|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[t>>2]=0;p[t+8>>3]=1.0;if((Xw(E,772667,t)|0)<0){tk(772672);k[m>>2]=402}A=rx(E,44)|0;if(A)i[A>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}break}case 20:{A=J;while(1)if((i[A>>0]|0)==85){ga=A;break}else A=A+1|0;i[ga>>0]=73;fs(a,K,J,M,m)|0;Gk(773309,ea,K,m)|0;y=M;z=773365;ca=y+29|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[s>>2]=0;p[s+8>>3]=32768.0;if((Xw(E,772667,s)|0)<0){tk(772672);k[m>>2]=402}A=rx(E,44)|0;if(A)i[A>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}Gk(773339,ea,K,m)|0;y=M;z=773345;ca=y+20|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[q>>2]=0;p[q+8>>3]=1.0;if((Xw(E,772667,q)|0)<0){tk(772672);k[m>>2]=402}A=rx(E,44)|0;if(A)i[A>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}break}case 40:{A=J;while(1)if((i[A>>0]|0)==86){ia=A;break}else A=A+1|0;i[ia>>0]=74;fs(a,K,J,M,m)|0;Gk(773309,ea,K,m)|0;y=M;z=773365;ca=y+29|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[o>>2]=0;p[o+8>>3]=2147483648.0;if((Xw(E,772667,o)|0)<0){tk(772672);k[m>>2]=402}A=rx(E,44)|0;if(A)i[A>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}Gk(773339,ea,K,m)|0;y=M;z=773345;ca=y+20|0;do{i[y>>0]=i[z>>0]|0;y=y+1|0;z=z+1|0}while((y|0)<(ca|0));if((k[m>>2]|0)<=0){i[E>>0]=0;k[n>>2]=0;p[n+8>>3]=1.0;if((Xw(E,772667,n)|0)<0){tk(772672);k[m>>2]=402}z=rx(E,44)|0;if(z)i[z>>0]=46;if(rx(E,78)|0){tk(772715);k[m>>2]=402}Dk(K,E,M,F,m)|0;ds(a,F,m)|0}break}default:fs(a,K,J,M,m)|0}do if(!x?(z=g+(da<<2)|0,y=k[z>>2]|0,(y|0)!=0):0){if(!(i[y>>0]|0))break;Gk(773394,ea,K,m)|0;fs(a,K,k[z>>2]|0,773400,m)|0}while(0)}while((k[m>>2]|0)<=0);if(i[N>>0]|0)fs(a,790194,N,773423,m)|0;N=k[m>>2]|0;if((N|0)<=0){P=N;r=j;return P|0}tk(773459);P=k[m>>2]|0;r=j;return P|0}function bs(a,b,c,d,e,f,g,h,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;o=r;r=r+512|0;p=o+96|0;q=o+88|0;s=o+80|0;t=o+72|0;u=o+64|0;v=o+56|0;w=o+48|0;x=o+40|0;y=o+32|0;z=o+24|0;A=o+16|0;B=o+8|0;C=o;D=o+439|0;E=o+358|0;F=o+104|0;G=o+180|0;H=o+283|0;I=o+210|0;J=o+108|0;K=k[a>>2]|0;L=a+4|0;if((K|0)!=(k[(k[L>>2]|0)+68>>2]|0))Fk(a,K+1|0,0,n)|0;K=k[n>>2]|0;if((K|0)>0){M=K;r=o;return M|0}K=k[L>>2]|0;L=K+96|0;N=(k[K+88>>2]|0)+(k[K+68>>2]<<3)|0;if(!((k[L>>2]|0)==(k[N>>2]|0)?(k[L+4>>2]|0)==(k[N+4>>2]|0):0)){k[n>>2]=201;M=201;r=o;return M|0}if((c|0)<0){k[n>>2]=217;M=217;r=o;return M|0}if((e|0)<0){k[n>>2]=218;M=218;r=o;return M|0}if(f>>>0>999){k[n>>2]=216;M=216;r=o;return M|0}i[J>>0]=0;if(m)zx(J,m,70)|0;k[F>>2]=b;if(((h|0)!=0?(k[h>>2]|0)!=0:0)?!((b|0)==0&(c|0)==0&(f|0)!=0):0){O=h;P=0}else{h=ey((f|0)<5?5:f,4)|0;if(!h){O=h;P=0}else{kl(f,j,1,F,h,n)|0;O=h;P=1}}fs(a,782702,773513,773519,n)|0;h=k[n>>2]|0;if((h|0)>0)Q=h;else{i[D>>0]=0;h=C;k[h>>2]=8;k[h+4>>2]=0;if((Xw(D,772008,C)|0)<0){tk(772013);k[n>>2]=401}Dk(779656,D,773541,E,n)|0;ds(a,E,n)|0;Q=k[n>>2]|0}if((Q|0)>0)R=Q;else{i[D>>0]=0;Q=B;k[Q>>2]=2;k[Q+4>>2]=0;if((Xw(D,772008,B)|0)<0){tk(772013);k[n>>2]=401}Dk(782713,D,773564,E,n)|0;ds(a,E,n)|0;R=k[n>>2]|0}B=k[F>>2]|0;if((R|0)>0)S=R;else{i[D>>0]=0;R=A;k[R>>2]=B;k[R+4>>2]=((B|0)<0)<<31>>31;if((Xw(D,772008,A)|0)<0){tk(772013);k[n>>2]=401}Dk(772833,D,773590,E,n)|0;ds(a,E,n)|0;S=k[n>>2]|0}if((S|0)>0)T=S;else{i[D>>0]=0;S=z;k[S>>2]=d;k[S+4>>2]=e;if((Xw(D,772008,z)|0)<0){tk(772013);k[n>>2]=401}Dk(790243,D,772864,E,n)|0;ds(a,E,n)|0;T=k[n>>2]|0}if((T|0)>0)U=T;else{i[D>>0]=0;T=y;k[T>>2]=0;k[T+4>>2]=0;if((Xw(D,772008,y)|0)<0){tk(772013);k[n>>2]=401}Dk(782652,D,773619,E,n)|0;ds(a,E,n)|0;U=k[n>>2]|0}if((U|0)>0)V=U;else{i[D>>0]=0;U=x;k[U>>2]=1;k[U+4>>2]=0;if((Xw(D,772008,x)|0)<0){tk(772013);k[n>>2]=401}Dk(782726,D,772914,E,n)|0;ds(a,E,n)|0;V=k[n>>2]|0}if((V|0)<=0){i[D>>0]=0;V=w;k[V>>2]=f;k[V+4>>2]=((f|0)<0)<<31>>31;if((Xw(D,772008,w)|0)<0){tk(772013);k[n>>2]=401}Dk(782733,D,772948,E,n)|0;ds(a,E,n)|0}a:do if(!l){w=0;do{if((w|0)>=(f|0))break a;V=g+(w<<2)|0;if(i[k[V>>2]>>0]|0){x=w+1|0;k[v>>2]=x;Xw(I,772977,v)|0;Gk(772997,x,H,n)|0;fs(a,H,k[V>>2]|0,I,n)|0}V=O+(w<<2)|0;x=k[V>>2]|0;if((x|0)<1?1:(x|0)>(k[F>>2]|0))k[n>>2]=234;x=w;w=w+1|0;k[u>>2]=w;Xw(I,773658,u)|0;Gk(773688,w,H,n)|0;U=k[V>>2]|0;if((k[n>>2]|0)<=0){i[D>>0]=0;V=t;k[V>>2]=U;k[V+4>>2]=((U|0)<0)<<31>>31;if((Xw(D,772008,t)|0)<0){tk(772013);k[n>>2]=401}Dk(H,D,I,E,n)|0;ds(a,E,n)|0}ux(G,k[j+(x<<2)>>2]|0)|0;Ck(G);Gk(773003,w,H,n)|0;fs(a,H,G,773694,n)|0}while((k[n>>2]|0)<=0)}else{w=0;do{if((w|0)>=(f|0))break a;x=g+(w<<2)|0;if(i[k[x>>2]>>0]|0){U=w+1|0;k[s>>2]=U;Xw(I,772977,s)|0;Gk(772997,U,H,n)|0;fs(a,H,k[x>>2]|0,I,n)|0}x=O+(w<<2)|0;U=k[x>>2]|0;if((U|0)<1?1:(U|0)>(k[F>>2]|0))k[n>>2]=234;U=w;w=w+1|0;k[q>>2]=w;Xw(I,773658,q)|0;Gk(773688,w,H,n)|0;V=k[x>>2]|0;if((k[n>>2]|0)<=0){i[D>>0]=0;x=p;k[x>>2]=V;k[x+4>>2]=((V|0)<0)<<31>>31;if((Xw(D,772008,p)|0)<0){tk(772013);k[n>>2]=401}Dk(H,D,I,E,n)|0;ds(a,E,n)|0}ux(G,k[j+(U<<2)>>2]|0)|0;Ck(G);Gk(773003,w,H,n)|0;fs(a,H,G,773694,n)|0;V=l+(U<<2)|0;U=k[V>>2]|0;do if(U){if(!(i[U>>0]|0))break;Gk(773394,w,H,n)|0;fs(a,H,k[V>>2]|0,773400,n)|0}while(0)}while((k[n>>2]|0)<=0)}while(0);if(i[J>>0]|0)fs(a,790194,J,773721,n)|0;if((k[n>>2]|0)>0)tk(773756);if(P)dy(O);M=k[n>>2]|0;r=o;return M|0}function cs(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;d=r;r=r+400|0;e=d+315|0;f=d+154|0;g=d+79|0;h=d+4|0;j=d;l=k[c>>2]|0;if((l|0)>0){m=l;r=d;return m|0}l=Dw(b,774853)|0;if(!l){tk(773809);tk(b);k[c>>2]=104;m=104;r=d;return m|0}b=f+160|0;a:do if(Bw(f,160,l)|0){n=g+8|0;o=e+40|0;p=h+8|0;while(1){i[b>>0]=0;q=f+((xx(f)|0)+-1)|0;i[q>>0]=0;if((Ik(f,e,j,c)|0)>0)break a;Bx(g,e,8)|0;i[n>>0]=0;switch(k[j>>2]|0){case -2:{Bx(h,o,8)|0;i[p>>0]=0;yq(a,g,h,c)|0;break}case -1:{Aq(a,g,c)|0;break}case 0:{uq(a,g,e,c)|0;break}case 1:{ds(a,e,c)|0;break}default:break a}if(!(Bw(f,160,l)|0))break a}}while(0);ww(l)|0;m=k[c>>2]|0;r=d;return m|0}function ds(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0;d=r;r=r+96|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}f=k[a>>2]|0;h=a+4|0;j=k[h>>2]|0;if((f|0)==(k[j+68>>2]|0))l=j;else{Fk(a,f+1|0,0,c)|0;l=k[h>>2]|0}f=l+120|0;j=l+96|0;l=jy(k[f>>2]|0,k[f+4>>2]|0,k[j>>2]|0,k[j+4>>2]|0)|0;if((l|0)==80&(L|0)==0?(rl(a,1,0,c)|0)>0:0){g=k[c>>2]|0;r=d;return g|0}Bx(e,b,80)|0;i[e+80>>0]=0;b=xx(e)|0;if(b){l=0;do{j=e+l|0;f=i[j>>0]|0;if(f<<24>>24<32|f<<24>>24==127)i[j>>0]=32;l=l+1|0}while((l|0)!=(b|0));if(b>>>0<80)m=9}else m=9;if((m|0)==9)sy(e+b|0,32,80-b|0)|0;b=vx(e,773860)|0;l=(b|0)==80?8:b;if((((yx(773862,e,8)|0)!=0?(yx(773871,e,8)|0)!=0:0)?(yx(773880,e,8)|0)!=0:0)?(yx(773889,e,8)|0)!=0:0){if(l){n=l;m=18}}else{n=8;m=18}if((m|0)==18){m=0;do{l=e+m|0;b=(nv(i[l>>0]|0)|0)&255;i[l>>0]=b;m=m+1|0}while((m|0)!=(n|0))}Ak(e,c)|0;n=(k[h>>2]|0)+96|0;au(a,k[n>>2]|0,k[n+4>>2]|0,1,c)|0;cu(a,80,0,e,c)|0;e=k[c>>2]|0;if((e|0)>=1){g=e;r=d;return g|0}c=(k[h>>2]|0)+96|0;h=c;a=ky(k[h>>2]|0,k[h+4>>2]|0,80,0)|0;h=c;k[h>>2]=a;k[h+4>>2]=L;g=e;r=d;return g|0}function es(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0.0,G=0;g=r;r=r+336|0;h=g+88|0;n=g+80|0;q=g+72|0;s=g+56|0;t=g+48|0;u=g+32|0;v=g+24|0;w=g+16|0;x=g+8|0;y=g;z=g+92|0;A=g+244|0;B=g+163|0;C=k[f>>2]|0;if((C|0)>0){D=C;r=g;return D|0}do switch(b|0){case 16:{fs(a,c,d,e,f)|0;break}case 11:{C=l[d>>0]|0;i[z>>0]=0;E=y;k[E>>2]=C;k[E+4>>2]=0;if((Xw(z,772008,y)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 12:{E=i[d>>0]|0;i[z>>0]=0;C=x;k[C>>2]=E;k[C+4>>2]=((E|0)<0)<<31>>31;if((Xw(z,772008,x)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 20:{E=m[d>>1]|0;i[z>>0]=0;C=w;k[C>>2]=E;k[C+4>>2]=0;if((Xw(z,772008,w)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 21:{C=j[d>>1]|0;i[z>>0]=0;E=v;k[E>>2]=C;k[E+4>>2]=((C|0)<0)<<31>>31;if((Xw(z,772008,v)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 30:{F=+((k[d>>2]|0)>>>0);i[z>>0]=0;k[u>>2]=0;p[u+8>>3]=F;if((Xw(z,772667,u)|0)<0){tk(772672);k[f>>2]=402}C=rx(z,44)|0;if(C)i[C>>0]=46;if(rx(z,78)|0){tk(772715);k[f>>2]=402}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 31:{C=k[d>>2]|0;i[z>>0]=0;E=t;k[E>>2]=C;k[E+4>>2]=((C|0)<0)<<31>>31;if((Xw(z,772008,t)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 14:{j[z>>1]=(k[d>>2]|0)==0?70:84;Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 40:{F=+((k[d>>2]|0)>>>0);i[z>>0]=0;k[s>>2]=0;p[s+8>>3]=F;if((Xw(z,772667,s)|0)<0){tk(772672);k[f>>2]=402}C=rx(z,44)|0;if(C)i[C>>0]=46;if(rx(z,78)|0){tk(772715);k[f>>2]=402}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 41:{C=k[d>>2]|0;i[z>>0]=0;E=q;k[E>>2]=C;k[E+4>>2]=((C|0)<0)<<31>>31;if((Xw(z,772008,q)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 81:{C=d;E=k[C>>2]|0;G=k[C+4>>2]|0;i[z>>0]=0;C=n;k[C>>2]=E;k[C+4>>2]=G;if((Xw(z,772008,n)|0)<0){tk(772013);k[f>>2]=401}Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 42:{us(+o[d>>2],-7,z,f)|0;Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 82:{ws(+p[d>>3],-15,z,f)|0;Dk(c,z,e,A,f)|0;ds(a,A,f)|0;break}case 83:{ls(a,c,d,-7,e,f)|0;break}case 163:{ms(a,c,d,-15,e,f)|0;break}default:{k[h>>2]=b;Xw(B,773898,h)|0;tk(B);k[f>>2]=410}}while(0);D=k[f>>2]|0;r=g;return D|0}function fs(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;f=r;r=r+160|0;g=f+81|0;h=f;j=k[e>>2]|0;if((j|0)>0){l=j;r=f;return l|0}do if(!c){i[g>>0]=i[773936]|0;i[g+1>>0]=i[773937]|0;i[g+2>>0]=i[773938]|0}else{i[g>>0]=39;j=xx(c)|0;m=j>>>0>68?68:j;if(m){j=0;n=1;while(1){o=i[c+j>>0]|0;i[g+n>>0]=o;if(o<<24>>24==39){o=n+1|0;i[g+o>>0]=39;p=o}else p=n;j=j+1|0;o=p+1|0;if(!(j>>>0<m>>>0&o>>>0<69)){q=o;break}else n=o}if(q>>>0>=9)if((q|0)==70){i[g+69>>0]=0;break}else s=q;else{t=q;u=6}}else{t=1;u=6}if((u|0)==6){sy(g+t|0,32,9-t|0)|0;s=9}i[g+s>>0]=39;i[g+(s+1)>>0]=0}while(0);Dk(b,g,d,h,e)|0;ds(a,h,e)|0;l=k[e>>2]|0;r=f;return l|0}function gs(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0;g=r;r=r+160|0;h=g;j=g+89|0;l=g+8|0;m=k[f>>2]|0;if((m|0)>0){n=m;r=g;return n|0}i[j>>0]=0;m=h;k[m>>2]=c;k[m+4>>2]=d;if((Xw(j,772008,h)|0)<0){tk(772013);k[f>>2]=401}Dk(b,j,e,l,f)|0;ds(a,l,f)|0;n=k[f>>2]|0;r=g;return n|0}function hs(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0;g=r;r=r+160|0;h=g+81|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}vs(c,d,h,f)|0;Dk(b,h,e,i,f)|0;ds(a,i,f)|0;l=k[f>>2]|0;r=g;return l|0}function is(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,l=0;f=r;r=r+160|0;g=f;h=f+71|0;i=k[e>>2]|0;if((i|0)>0){l=i;r=f;return l|0}j[g>>1]=(c|0)==0?70:84;Dk(b,g,d,h,e)|0;ds(a,h,e)|0;l=k[e>>2]|0;r=f;return l|0}function js(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0;g=r;r=r+160|0;h=g+81|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}us(c,d,h,f)|0;Dk(b,h,e,i,f)|0;ds(a,i,f)|0;l=k[f>>2]|0;r=g;return l|0}function ks(a,b,c,d,e,f){a=a|0;b=b|0;c=+c;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0;g=r;r=r+160|0;h=g+81|0;i=g;j=k[f>>2]|0;if((j|0)>0){l=j;r=g;return l|0}ws(c,d,h,f)|0;Dk(b,h,e,i,f)|0;ds(a,i,f)|0;l=k[f>>2]|0;r=g;return l|0}function ls(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,m=0,n=0,p=0;g=r;r=r+224|0;h=g;l=g+152|0;m=g+71|0;n=k[f>>2]|0;if((n|0)>0){p=n;r=g;return p|0}j[h>>1]=40;us(+o[c>>2],d,l,f)|0;qx(h,l)|0;n=h+(xx(h)|0)|0;i[n>>0]=i[773939]|0;i[n+1>>0]=i[773940]|0;i[n+2>>0]=i[773941]|0;us(+o[c+4>>2],d,l,f)|0;qx(h,l)|0;l=h+(xx(h)|0)|0;i[l>>0]=41;i[l+1>>0]=0;Dk(b,h,e,m,f)|0;ds(a,m,f)|0;p=k[f>>2]|0;r=g;return p|0}function ms(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,m=0,n=0,o=0;g=r;r=r+224|0;h=g;l=g+152|0;m=g+71|0;n=k[f>>2]|0;if((n|0)>0){o=n;r=g;return o|0}j[h>>1]=40;ws(+p[c>>3],d,l,f)|0;qx(h,l)|0;n=h+(xx(h)|0)|0;i[n>>0]=i[773939]|0;i[n+1>>0]=i[773940]|0;i[n+2>>0]=i[773941]|0;ws(+p[c+8>>3],d,l,f)|0;qx(h,l)|0;l=h+(xx(h)|0)|0;i[l>>0]=41;i[l+1>>0]=0;Dk(b,h,e,m,f)|0;ds(a,m,f)|0;o=k[f>>2]|0;r=g;return o|0}function ns(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=r;r=r+160|0;f=e;g=e+71|0;h=k[d>>2]|0;if((h|0)>0){i=h;r=e;return i|0}j[f>>1]=32;Dk(b,f,c,g,d)|0;ds(a,g,d)|0;i=k[d>>2]|0;r=e;return i|0}function os(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}if(!a){i[b>>0]=i[773936]|0;i[b+1>>0]=i[773937]|0;i[b+2>>0]=i[773938]|0;e=k[c>>2]|0;return e|0}i[b>>0]=39;d=xx(a)|0;f=d>>>0>68?68:d;if(f){d=0;g=1;while(1){h=a+d|0;i[b+g>>0]=i[h>>0]|0;if((i[h>>0]|0)==39){h=g+1|0;i[b+h>>0]=39;j=h}else j=g;d=d+1|0;h=j+1|0;if(!(d>>>0<f>>>0&h>>>0<69)){l=h;break}else g=h}if(l>>>0>=9)if((l|0)==70)i[b+69>>0]=0;else{m=l;n=12}else{o=l;n=6}}else{o=1;n=6}if((n|0)==6){sy(b+o|0,32,9-o|0)|0;m=9;n=12}if((n|0)==12){i[b+m>>0]=39;i[b+(m+1)>>0]=0}e=k[c>>2]|0;return e|0}function ps(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;f=r;r=r+336|0;g=f+247|0;h=f+166|0;j=f+85|0;l=f+4|0;m=f;k[m>>2]=-1;n=k[e>>2]|0;if((n|0)>0){o=n;r=f;return o|0}n=xx(c)|0;p=n>>>0>1?n:1;if(!d)q=3;else q=(xx(d)|0)+3|0;i[l>>0]=0;zx(l,c,68)|0;n=rx(l,39)|0;if(!n)s=0;else{t=n;n=0;while(1){u=n+1|0;t=rx(t+1|0,39)|0;if(!t){s=u;break}else n=u}}Bx(j,b,80)|0;i[j+80>>0]=0;n=j;while(1)if((i[n>>0]|0)==32)n=n+1|0;else{v=n;break}n=xx(v)|0;if((n|0)<9?(Ak(v,m)|0)<1:0)w=68-s|0;else w=75-s-n|0;if((p|0)>0){n=h+8|0;m=g+69|0;v=0;j=w;w=0;t=0;u=s;s=p;while(1){i[l>>0]=0;zx(l,c+w|0,j)|0;do if((k[e>>2]|0)<=0){i[g>>0]=39;p=xx(l)|0;x=p>>>0>68?68:p;if(x){p=0;y=1;while(1){z=i[l+p>>0]|0;i[g+y>>0]=z;if(z<<24>>24==39){z=y+1|0;i[g+z>>0]=39;A=z}else A=y;p=p+1|0;z=A+1|0;if(!(p>>>0<x>>>0&z>>>0<69)){B=z;break}else y=z}if(B>>>0>=9)if((B|0)==70){i[m>>0]=0;break}else C=B;else{D=B;E=17}}else{D=1;E=17}if((E|0)==17){E=0;sy(g+D|0,32,9-D|0)|0;C=9}i[g+C>>0]=39;i[g+(C+1)>>0]=0}while(0);do if((s|0)>(j|0)){y=xx(g)|0;x=j+-1|0;p=g+(y+-2)|0;if((i[p>>0]|0)==39){i[g+(y+-3)>>0]=38;i[g+(y+-1)>>0]=0;F=x;break}else{i[p>>0]=38;F=x;break}}else F=j;while(0);if(!v)Dk(b,g,d,h,e)|0;else{if(!t)Dk(773889,g,d,h,e)|0;else Dk(773889,g,0,h,e)|0;i[n>>0]=32;i[n+1>>0]=32}ds(a,h,e)|0;s=s-F|0;w=F+w|0;x=(s|0)>0;if(x){i[l>>0]=0;zx(l,c+w|0,68)|0;p=rx(l,39)|0;if(!p)G=0;else{y=p;p=0;while(1){z=p+1|0;y=rx(y+1|0,39)|0;if(!y){G=z;break}else p=z}}H=68-G|0;I=G}else{H=F;I=u}p=I+s|0;if((q+p|0)>68){y=(I|0)<18&(p|0)>18;J=y?18-I|0:H;K=y&1}else{J=H;K=0}if(!x)break;else{v=1;j=J;t=K;u=I}}}o=k[e>>2]|0;r=f;return o|0}function qs(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0;c=r;r=r+240|0;d=c+148|0;e=c;f=k[b>>2]|0;if((f|0)>0){g=f;r=c;return g|0}k[e>>2]=0;if(!(Jo(a,773942,c+77|0,c+4|0,e)|0)){g=k[b>>2]|0;r=c;return g|0}fs(a,773942,773951,773960,b)|0;e=k[b>>2]|0;if((e|0)<=0){f=d;h=773862;j=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));f=d+(xx(d)|0)|0;h=774008;j=f+65|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));ds(a,d,b)|0;l=k[b>>2]|0;if((l|0)<=0){f=d;h=773862;j=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));f=d+(xx(d)|0)|0;h=774073;j=f+71|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));ds(a,d,b)|0;m=k[b>>2]|0;if((m|0)<=0){f=d;h=773862;j=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));f=d+(xx(d)|0)|0;h=774144;j=f+65|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));ds(a,d,b)|0;n=k[b>>2]|0;if((n|0)>0)o=n;else{f=d;h=773862;j=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));f=d+(xx(d)|0)|0;h=774209;j=f+51|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(j|0));ds(a,d,b)|0;o=k[b>>2]|0}}else{p=m;q=9}}else{s=l;q=7}}else{s=e;q=7}if((q|0)==7){p=s;q=9}if((q|0)==9)o=p;g=o;r=c;return g|0}function rs(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;d=r;r=r+96|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}h=xx(b)|0;if((h|0)>0){j=0;l=h}else{g=f;r=d;return g|0}while(1){f=e;h=773862;m=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(m|0));zx(e,b+j|0,72)|0;ds(a,e,c)|0;if((l|0)>72){j=j+72|0;l=l+-72|0}else break}g=k[c>>2]|0;r=d;return g|0}function ss(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=k[c>>2]|0;if((d|0)>0){e=d;return e|0}d=(a|0)==0?70:84;i[b>>0]=d;i[b+1>>0]=d>>8;e=k[c>>2]|0;return e|0}function ts(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;if((g|0)>0){h=g;r=e;return h|0}i[c>>0]=0;g=f;k[g>>2]=a;k[g+4>>2]=b;if((Xw(c,772008,f)|0)<0){tk(772013);k[d>>2]=401;h=401;r=e;return h|0}else{h=k[d>>2]|0;r=e;return h|0}return 0}function us(a,b,c,d){a=+a;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0.0,n=0;e=r;r=r+48|0;f=e+24|0;g=e+16|0;h=e;j=k[d>>2]|0;if((j|0)>0){l=j;r=e;return l|0}i[c>>0]=0;do if((b|0)<0){m=a;k[h>>2]=0-b;p[h+8>>3]=m;if((Xw(c,774260,h)|0)<0){tk(774265);k[d>>2]=402;break}if((rx(c,46)|0)==0?(rx(c,69)|0)!=0:0){p[g>>3]=m;if((Xw(c,774307,g)|0)<0){tk(774265);k[d>>2]=402;l=402;r=e;return l|0}else{l=k[d>>2]|0;r=e;return l|0}}else n=12}else{k[f>>2]=b;p[f+8>>3]=a;if((Xw(c,774312,f)|0)<0){tk(774265);k[d>>2]=402}else n=12}while(0);do if((n|0)==12?(k[d>>2]|0)<1:0){f=rx(c,44)|0;if(f)i[f>>0]=46;if(rx(c,78)|0){tk(774317);k[d>>2]=402;break}if((rx(c,46)|0)==0?(rx(c,69)|0)==0:0){f=c+(xx(c)|0)|0;i[f>>0]=46;i[f+1>>0]=0}}while(0);l=k[d>>2]|0;r=e;return l|0}function vs(a,b,c,d){a=+a;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;if((g|0)>0){h=g;r=e;return h|0}i[c>>0]=0;if((b|0)<0){tk(774363);k[d>>2]=411;h=411;r=e;return h|0}k[f>>2]=b;p[f+8>>3]=a;if((Xw(c,772667,f)|0)<0){tk(772672);k[d>>2]=402}f=rx(c,44)|0;if(f)i[f>>0]=46;if(!(rx(c,78)|0)){h=k[d>>2]|0;r=e;return h|0}else{tk(772715);k[d>>2]=402;h=402;r=e;return h|0}return 0}function ws(a,b,c,d){a=+a;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0;e=r;r=r+48|0;f=e+24|0;g=e+16|0;h=e;j=k[d>>2]|0;if((j|0)>0){l=j;r=e;return l|0}i[c>>0]=0;do if((b|0)<0){k[h>>2]=0-b;p[h+8>>3]=a;if((Xw(c,774260,h)|0)<0){tk(774406);k[d>>2]=402;break}if((rx(c,46)|0)==0?(rx(c,69)|0)!=0:0){p[g>>3]=a;if((Xw(c,774307,g)|0)<0){tk(774406);k[d>>2]=402;l=402;r=e;return l|0}else{l=k[d>>2]|0;r=e;return l|0}}else m=12}else{k[f>>2]=b;p[f+8>>3]=a;if((Xw(c,774312,f)|0)<0){tk(774406);k[d>>2]=402}else m=12}while(0);do if((m|0)==12?(k[d>>2]|0)<1:0){f=rx(c,44)|0;if(f)i[f>>0]=46;if(rx(c,78)|0){tk(774448);k[d>>2]=402;break}if((rx(c,46)|0)==0?(rx(c,69)|0)==0:0){f=c+(xx(c)|0)|0;i[f>>0]=46;i[f+1>>0]=0}}while(0);l=k[d>>2]|0;r=e;return l|0}function xs(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;d=r;r=r+96|0;e=d;f=k[c>>2]|0;if((f|0)>0){g=f;r=d;return g|0}h=xx(b)|0;if((h|0)>0){j=0;l=h}else{g=f;r=d;return g|0}while(1){f=e;h=773871;m=f+9|0;do{i[f>>0]=i[h>>0]|0;f=f+1|0;h=h+1|0}while((f|0)<(m|0));zx(e,b+j|0,72)|0;ds(a,e,c)|0;if((l|0)>72){j=j+72|0;l=l+-72|0}else break}g=k[c>>2]|0;r=d;return g|0}function ys(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;f=r;r=r+416|0;g=f+8|0;h=f;l=f+329|0;m=f+20|0;n=f+256|0;o=f+176|0;p=f+91|0;q=f+16|0;s=k[e>>2]|0;if((s|0)>0){t=s;r=f;return t|0}s=b+-1|0;if(s>>>0>998){tk(774495);k[e>>2]=302;t=302;r=f;return t|0}if((c|0)<1){tk(774542);k[e>>2]=320;t=320;r=f;return t|0}u=k[a>>2]|0;v=a+4|0;w=k[v>>2]|0;if((u|0)==(k[w+68>>2]|0)){x=w+120|0;if(((k[x>>2]|0)==-1?(k[x+4>>2]|0)==-1:0)?(Vk(a,e)|0)>0:0){t=k[e>>2]|0;r=f;return t|0}}else Fk(a,u+1|0,0,e)|0;u=k[v>>2]|0;if((k[u+72>>2]|0)!=2){tk(774572);k[e>>2]=227;t=227;r=f;return t|0}j[m>>1]=40;do if((c|0)>0){x=0;w=1;while(1){if((x|0)>0){y=m+(xx(m)|0)|0;i[y>>0]=44;i[y+1>>0]=0}y=d+(x<<2)|0;z=k[y>>2]|0;if((z|0)<0)break;k[h>>2]=z;Xw(o,773048,h)|0;qx(m,o)|0;z=ha(k[y>>2]|0,w)|0;x=x+1|0;if((x|0)>=(c|0)){A=z;B=19;break}else w=z}if((B|0)==19){C=k[v>>2]|0;D=A;break}tk(774645);k[e>>2]=263;t=263;r=f;return t|0}else{C=u;D=1}while(0);u=(k[C+960>>2]|0)+(s*160|0)+88|0;if((k[u>>2]|0)!=(D|0)?(Gk(773003,b,l,e)|0,Jo(a,l,o,0,e)|0,Qk(o,0,q,0,e)|0,!((k[e>>2]|0)<1?(k[q>>2]|0)==(D|0):0)):0){k[g>>2]=k[u>>2];k[g+4>>2]=D;Xw(p,774694,g)|0;tk(p);k[e>>2]=263;t=263;r=f;return t|0}p=m+(xx(m)|0)|0;i[p>>0]=41;i[p+1>>0]=0;p=n;g=774758;D=p+35|0;do{i[p>>0]=i[g>>0]|0;p=p+1|0;g=g+1|0}while((p|0)<(D|0));Gk(774793,b,l,e)|0;fs(a,l,m,n,e)|0;t=k[e>>2]|0;r=f;return t|0}function zs(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=r;r=r+16|0;f=e+4|0;g=e;k[g>>2]=0;h=k[d>>2]|0;if(h){i=h;r=e;return i|0}vk();if(!(Hh(f,a,0,g)|0))Bs(k[f>>2]|0,b,c,d)|0;else{yk();As(a,b,c,d)|0}i=k[d>>2]|0;r=e;return i|0}function As(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0.0,H=0.0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0.0,ma=0.0,na=0.0,oa=0.0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,gc=0,hc=0,ic=0,jc=0,kc=0,lc=0.0,mc=0.0,nc=0,oc=0,pc=0,qc=0,rc=0.0,sc=0.0,tc=0.0,uc=0.0,vc=0.0,wc=0.0,xc=0.0,yc=0.0,zc=0,Ac=0.0,Bc=0.0,Cc=0.0,Dc=0.0,Ec=0,Fc=0,Gc=0.0,Hc=0,Ic=0,Jc=0,Kc=0.0,Lc=0.0,Mc=0,Nc=0,Oc=0,Pc=0,Qc=0,Rc=0.0,Sc=0.0,Tc=0,Uc=0,Vc=0,Wc=0,Xc=0,Yc=0,Zc=0,_c=0,$c=0,ad=0,bd=0;e=r;r=r+32|0;f=e+16|0;g=e+20|0;h=e+8|0;j=e;l=k[d>>2]|0;if(l){m=l;r=e;return m|0}l=cy(80)|0;if(!l){tk(774798);k[d>>2]=113;m=113;r=e;return m|0}k[l>>2]=0;n=l+4|0;k[n>>2]=0;o=(b|0)==0;if(!o?(k[b>>2]|0)!=0:0){q=l+8|0;s=b;t=q+72|0;do{k[q>>2]=k[s>>2];q=q+4|0;s=s+4|0}while((q|0)<(t|0))}else k[l+8>>2]=0;s=cy(512)|0;if(!s){dy(l);tk(774798);k[d>>2]=113;m=113;r=e;return m|0}u=Dw(a,774853)|0;if(!u){k[f>>2]=a;Xw(s,774855,f)|0;tk(s);dy(s);dy(l);k[d>>2]=104;m=104;r=e;return m|0}k[d>>2]=104;a:do if(!(Bw(s,512,u)|0)){v=s;w=226}else{f=b+8|0;a=b+16|0;x=b+24|0;y=b+32|0;z=b+40|0;A=b+48|0;B=b+56|0;C=b+64|0;D=512;E=0;F=s;G=0.0;H=0.0;b:while(1){I=D;J=E;K=F;c:while(1){L=I;M=K;while(1){k[d>>2]=0;N=xx(M)|0;d:do if((N|0)==(L+-1|0)){O=L;P=M;Q=N;while(1){if((i[P+(Q+-1)>>0]|0)==10){R=O;S=P;break d}T=O<<1;U=fy(P,T)|0;if(!U){V=P;w=20;break b}W=U+Q|0;Bw(W,O+1|0,u)|0;Q=(xx(W)|0)+Q|0;if((Q|0)!=(T+-1|0)){R=T;S=U;break}else{O=T;P=U}}}else{R=L;S=M}while(0);N=i[S>>0]|0;if(N<<24>>24==35){P=S;while(1){O=P+1|0;if(!(kv(i[O>>0]|0)|0)){Y=P;Z=O;break}else P=O}if(!(yx(Z,774886,7)|0)){_=R;$=S;aa=Y;break}}else if(yx(S,775022,4)|0){ba=N;ca=R;da=J;ea=S;break c}if(!(Bw(S,R,u)|0)){v=S;w=226;break a}else{L=R;M=S}}if(k[l>>2]|0){fa=$;w=26;break b}M=aa+8|0;while(1)if(!(kv(i[M>>0]|0)|0)){ga=M;break}else M=M+1|0;if(yx(ga,774950,5)|0)if(yx(ga,774956,6)|0)if((yx(ga,774963,6)|0)!=0?(yx(ga,774970,3)|0)!=0:0){ha=$;w=34;break b}else ia=2;else ia=1;else ia=0;if(!(Bw($,_,u)|0)){v=$;w=226;break a}else{I=_;J=ia;K=$}}e:do if(!(ba<<24>>24)){ja=da;ka=ea;la=G;ma=H}else{K=da;J=ea;na=G;oa=H;while(1){I=K;M=J;f:while(1){L=I;P=M;O=1;Q=0;g:while(1){U=P;pa=O;T=Q;h:while(1){W=U;qa=T;i:while(1){ra=(qa|0)==0;sa=(qa|0)==0;ta=W;j:while(1){ua=(k[d>>2]|0)==0;k:while(1){if(!ua){va=ea;w=59;break b}wa=i[ta>>0]|0;if(!(wa<<24>>24)){va=ea;w=59;break b}switch(wa<<24>>24|0){case 58:{xa=ta;ya=pa;za=qa;w=53;break h;break}case 100:{Aa=ta;Ba=pa;Ca=qa;w=54;break h;break}case 44:{Da=ta;Ea=qa;w=55;break i;break}case 41:{Fa=ta;break j;break}case 40:{break k;break}case 10:case 35:break;default:{Ga=ta;Ha=pa;Ia=qa;break i}}i[ta>>0]=0;ua=(k[d>>2]|0)==0;if(ra){w=57;break g}}i[ta>>0]=0;wa=ta+1|0;if(sa){W=wa;qa=wa;continue i}k[d>>2]=1;ta=wa}i[Fa>>0]=0;sa=Fa+1|0;if(qa){Ja=sa;Ka=L;La=pa;Ma=qa;w=50;break g}k[d>>2]=1;W=sa;qa=0}if((w|0)==55){w=0;Ga=Da;Ha=pa+1|0;Ia=Ea}U=Ga+1|0;pa=Ha;T=Ia}if((w|0)==53){w=0;L=(za|0)==0?L:2;P=xa+1|0;O=ya;Q=za;continue}else if((w|0)==54){w=0;L=(Ca|0)==0?L:1;P=Aa+1|0;O=Ba;Q=Ca;continue}}if((w|0)==50){w=0;if(!(k[d>>2]|0)){Na=Ka;Oa=Ja;Pa=La;Qa=Ma}else{va=ea;w=59;break b}}else if((w|0)==57){w=0;if(ua){Na=L;Oa=ta;Pa=pa;Qa=qa}else{va=ea;w=59;break b}}Q=M;while(1)if(!(kv(i[Q>>0]|0)|0)){Ra=Q;break}else Q=Q+1|0;do if((Qa|0)!=0|(i[Ra>>0]|0)!=0){if(!(yx(Ra,775053,6)|0)){Sa=Oa;Ta=Pa;Ua=Ra;Va=Qa;w=64;break f}if(!(yx(Ra,775060,9)|0)){Wa=Oa;Xa=Pa;Ya=Ra;Za=Qa;w=66;break f}if(!(yx(Ra,775070,7)|0)){_a=Oa;$a=Pa;ab=Ra;bb=Qa;w=68;break f}if(!(yx(Ra,775078,4)|0)){cb=Oa;db=Pa;eb=Ra;fb=Qa;w=70;break f}if(!(yx(Ra,775083,4)|0)){gb=Oa;hb=Pa;ib=Ra;jb=Qa;w=72;break f}if(!(yx(Ra,775088,5)|0)){kb=Oa;lb=Pa;mb=Ra;nb=Qa;w=74;break f}if(!(yx(Ra,775094,3)|0)){ob=1;break}if(!(yx(Ra,775098,3)|0)){ob=1;break}if(!(yx(Ra,775102,4)|0)){ob=1;break}if(!(yx(Ra,775107,5)|0)){ob=0;break}if(!(yx(Ra,775113,8)|0))ob=0;else{pb=Na;qb=Oa;rb=Pa;sb=Ra;tb=Qa;w=80;break f}}else ob=Na;while(0);if(!(i[Oa>>0]|0)){ja=ob;ka=ea;la=na;ma=oa;break e}else{I=ob;M=Oa}}if((w|0)==64){w=0;ub=0;vb=Sa;wb=Ta;xb=Ua+6|0;yb=Va}else if((w|0)==66){w=0;ub=0;vb=Wa;wb=Xa;xb=Ya+9|0;yb=Za}else if((w|0)==68){w=0;ub=0;vb=_a;wb=$a;xb=ab+7|0;yb=bb}else if((w|0)==70){w=0;ub=1;vb=cb;wb=db;xb=eb+4|0;yb=fb}else if((w|0)==72){w=0;ub=1;vb=gb;wb=hb;xb=ib+4|0;yb=jb}else if((w|0)==74){w=0;ub=1;vb=kb;wb=lb;xb=mb+5|0;yb=nb}else if((w|0)==80){w=0;if(!(yx(sb,775122,9)|0)){zb=ea;Ab=sb;w=81;break b}if(!(yx(sb,775174,9)|0)){Bb=ea;Cb=sb;w=83;break b}else{ub=pb;vb=qb;wb=rb;xb=sb;yb=tb}}M=k[l>>2]|0;I=k[n>>2]|0;if(!((M|0)%10|0)){if(!I)Db=cy(1680)|0;else Db=fy(I,(M*168|0)+1680|0)|0;if(!Db){Eb=ea;w=90;break b}k[n>>2]=Db;Fb=k[l>>2]|0;Gb=Db}else{Fb=M;Gb=I}k[l>>2]=Fb+1;I=Gb+(Fb*168|0)|0;i[I>>0]=1;M=Gb+(Fb*168|0)+4|0;k[M>>2]=0;q=Gb+48+(Fb*21<<3)|0;t=q+64|0;do{k[q>>2]=0;q=q+4|0}while((q|0)<(t|0));N=Gb+(Fb*168|0)+136|0;k[N>>2]=0;k[N+4>>2]=0;k[N+8>>2]=0;k[N+12>>2]=0;k[N+16>>2]=0;k[N+20>>2]=0;k[N+24>>2]=0;k[N+28>>2]=0;N=xb;while(1){Q=N+1|0;if(!(kv(i[N>>0]|0)|0)){Hb=Q;Ib=N;break}else N=Q}N=Gb+(Fb*168|0)+48|0;switch(i[Ib>>0]|0){case 43:{Jb=Hb;break}case 45:{i[I>>0]=0;Jb=Hb;break}default:Jb=Ib}while(1)if(!(kv(i[Jb>>0]|0)|0)){Kb=Jb;break}else Jb=Jb+1|0;if(!(i[Kb>>0]|0)){Lb=ea;w=98;break b}Q=(xx(Kb)|0)+-1|0;L=Kb+Q|0;if(kv(i[L>>0]|0)|0){O=L;L=Q;do{i[O>>0]=0;L=L+-1|0;O=Kb+L|0}while((kv(i[O>>0]|0)|0)!=0)}l:do if(!(px(Kb,775268)|0)){k[M>>2]=2;if((wb|0)==3){Mb=2;Nb=2;Ob=3}else{Pb=ea;Qb=Kb;w=103;break b}}else{if(!(px(Kb,775275)|0)){k[M>>2]=3;if((wb|0)==4){Mb=3;Nb=2;Ob=4;break}else{Rb=ea;Sb=Kb;w=106;break b}}if(!(px(Kb,775283)|0)){if((wb+-4|0)>>>0>4){Tb=ea;Ub=Kb;w=109;break b}if((wb|0)<6){k[M>>2]=4;p[Gb+(Fb*168|0)+80>>3]=0.0;Mb=4;Nb=2;Ob=wb;break}else{k[M>>2]=5;O=Gb+(Fb*168|0)+96|0;k[O>>2]=0;k[O+4>>2]=0;k[O+8>>2]=0;k[O+12>>2]=0;Mb=5;Nb=2;Ob=wb;break}}if(!(px(Kb,775291)|0)){k[M>>2]=5;switch(wb|0){case 6:case 8:break;default:k[d>>2]=431}O=Gb+(Fb*168|0)+96|0;k[O>>2]=0;k[O+4>>2]=0;k[O+8>>2]=0;k[O+12>>2]=0;Mb=5;Nb=2;Ob=wb;break}do if(px(Kb,775305)|0){if(!(px(Kb,775309)|0))break;do if(px(Kb,775316)|0){if(!(px(Kb,775326)|0))break;do if(px(Kb,775339)|0){if(!(px(Kb,775347)|0))break;if(!(px(Kb,775358)|0))break;if(!(px(Kb,775366)|0))break;do if(px(Kb,775377)|0){if(!(px(Kb,775384)|0))break;if(!(px(Kb,775388)|0)){k[M>>2]=0;if((wb|0)==2){Mb=0;Nb=2;Ob=2;break l}else{Vb=ea;Wb=Kb;w=142;break b}}if(!(px(Kb,775394)|0)){k[M>>2]=1;if((wb|0)==4){Mb=1;Nb=4;Ob=4;break l}else{Xb=ea;Yb=Kb;w=145;break b}}if(!(px(Kb,775399)|0)){k[M>>2]=11;if((wb|0)>5&(wb&1|0)==0){Mb=11;Nb=wb;Ob=wb;break l}else{Zb=ea;_b=Kb;w=148;break b}}if(!(px(Kb,775407)|0)){k[M>>2]=12;if((wb|0)==8){Mb=12;Nb=2;Ob=8;break l}else{$b=ea;ac=Kb;w=151;break b}}if(!(px(Kb,775413)|0)){k[M>>2]=13;if((wb&-2|0)!=10)k[d>>2]=431;p[Gb+(Fb*168|0)+128>>3]=0.0;Mb=13;Nb=2;Ob=wb;break l}if(px(Kb,775420)|0){bc=ea;cc=Kb;w=160;break b}k[M>>2]=14;if((wb&-2|0)!=10)k[d>>2]=431;p[Gb+(Fb*168|0)+128>>3]=0.0;Mb=14;Nb=2;Ob=wb;break l}while(0);k[M>>2]=10;if((wb|0)==4){Mb=10;Nb=2;Ob=4;break l}else{dc=ea;ec=Kb;w=139;break b}}while(0);k[M>>2]=9;if((wb&-2|0)!=4)k[d>>2]=431;p[Gb+(Fb*168|0)+80>>3]=0.0;Mb=9;Nb=2;Ob=wb;break l}while(0);k[M>>2]=8;if((wb&-2|0)!=4)k[d>>2]=431;p[Gb+(Fb*168|0)+80>>3]=0.0;Mb=8;Nb=4;Ob=wb;break l}while(0);if((wb+-4|0)>>>0>4){fc=ea;gc=Kb;w=120;break b}if((wb|0)<6){k[M>>2]=6;p[Gb+(Fb*168|0)+80>>3]=0.0;Mb=6;Nb=2;Ob=wb;break}else{k[M>>2]=7;O=Gb+(Fb*168|0)+96|0;k[O>>2]=0;k[O+4>>2]=0;k[O+8>>2]=0;k[O+12>>2]=0;Mb=7;Nb=2;Ob=wb;break}}while(0);if(k[d>>2]|0){hc=ea;ic=Kb;break b}if((Mb|0)==11){O=cy(Ob<<3)|0;k[N+4>>2]=O;if(!O){jc=ea;w=165;break b}k[N>>2]=Ob;kc=O}else kc=N;if((Nb|0)>0){O=ub;L=0;Q=yb;lc=na;mc=oa;while(1){P=Q;while(1){T=P+1|0;if((i[P>>0]|0)==44){nc=T;oc=P;break}else P=T}i[oc>>0]=0;P=nc;m:while(1){T=P+1|0;switch(i[P>>0]|0){case 0:case 44:{pc=T;qc=P;break m;break}default:P=T}}i[qc>>0]=0;if(!(rx(Q,58)|0)){rc=+dx(Q);sc=+dx(nc);P=(L|0)==0;tc=P?rc:lc;uc=P?sc:mc;if(!O){vc=tc;wc=rc;xc=uc;yc=sc;zc=0}else{Ac=tc;Bc=rc;Cc=uc;Dc=sc;Ec=O;w=200}}else{P=jx(Q,g,10)|0;T=k[g>>2]|0;do if(!T){Fc=0;Gc=0.0}else{if((i[T>>0]|0)!=58){Fc=0;Gc=0.0;break}U=jx(T+1|0,g,10)|0;W=k[g>>2]|0;if(!W){Fc=U;Gc=0.0;break}if((i[W>>0]|0)!=58){Fc=U;Gc=0.0;break}Fc=U;Gc=+dx(W+1|0)}while(0);sc=+(P|0)+ +(Fc|0)/60.0+Gc/3600.0;T=oc;W=nc;while(1)if(!(kv(i[W>>0]|0)|0)){Hc=T;Ic=W;break}else{U=W;W=W+1|0;T=U}uc=sc*15.0;T=(i[Ic>>0]|0)==45;W=jx(T?Hc+2|0:Ic,g,10)|0;P=k[g>>2]|0;do if(!P){Jc=0;Kc=0.0}else{if((i[P>>0]|0)!=58){Jc=0;Kc=0.0;break}U=jx(P+1|0,g,10)|0;sa=k[g>>2]|0;if(!sa){Jc=U;Kc=0.0;break}if((i[sa>>0]|0)!=58){Jc=U;Kc=0.0;break}Jc=U;Kc=+dx(sa+1|0)}while(0);if(T)Lc=+(0-W|0)-+(Jc|0)/60.0-Kc/3600.0;else Lc=+(W|0)+ +(Jc|0)/60.0+Kc/3600.0;P=(L|0)==0;Ac=P?uc:lc;Bc=uc;Cc=P?Lc:mc;Dc=Lc;Ec=2;w=200}if((w|0)==200){w=0;if(o){Mc=ea;w=202;break b}if(!(k[b>>2]|0)){Mc=ea;w=202;break b}if(Rs(Bc,Dc,+p[f>>3],+p[a>>3],+p[x>>3],+p[y>>3],+p[z>>3],+p[A>>3],+p[B>>3],C,h,j,d)|0){Nc=ea;w=204;break b}vc=Ac;wc=+p[h>>3];xc=Cc;yc=+p[j>>3];zc=Ec}p[kc+(L<<3)>>3]=wc;p[kc+((L|1)<<3)>>3]=yc;P=L+2|0;if((P|0)<(Nb|0)){O=zc;L=P;Q=pc;lc=vc;mc=xc}else{Oc=zc;Pc=P;Qc=pc;Rc=vc;Sc=xc;break}}}else{Oc=ub;Pc=0;Qc=yb;Rc=na;Sc=oa}if((Pc|0)<(Ob|0)){Q=kc+8|0;if(Sc<0.0){L=Pc;O=Qc;while(1){N=O;n:while(1){P=N+1|0;switch(i[N>>0]|0){case 0:case 44:{Tc=P;Uc=N;break n;break}default:N=P}}i[Uc>>0]=0;uc=+hx(O,g);N=kc+(L<<3)|0;p[N>>3]=uc;W=k[g>>2]|0;o:do if(W){T=i[W>>0]|0;switch(T<<24>>24){case 100:case 39:case 34:break;default:break o}if(Rs(Rc,Sc+uc/(T<<24>>24==39?60.0:T<<24>>24==34?3600.0:1.0),+p[f>>3],+p[a>>3],+p[x>>3],+p[y>>3],+p[z>>3],+p[A>>3],+p[B>>3],C,h,j,d)|0){Vc=ea;w=212;break b}mc=+p[h>>3]-+p[kc>>3];lc=+p[j>>3]-+p[Q>>3];sc=+X(+(mc*mc+lc*lc));p[N>>3]=sc}while(0);N=L+1|0;if((N|0)<(Ob|0)){L=N;O=Tc}else break}}else{O=Pc;L=Qc;while(1){N=L;p:while(1){W=N+1|0;switch(i[N>>0]|0){case 0:case 44:{Wc=W;Xc=N;break p;break}default:N=W}}i[Xc>>0]=0;uc=+hx(L,g);N=kc+(O<<3)|0;p[N>>3]=uc;W=k[g>>2]|0;q:do if(W){T=i[W>>0]|0;switch(T<<24>>24){case 100:case 39:case 34:break;default:break q}if(Rs(Rc,Sc-uc/(T<<24>>24==39?60.0:T<<24>>24==34?3600.0:1.0),+p[f>>3],+p[a>>3],+p[x>>3],+p[y>>3],+p[z>>3],+p[A>>3],+p[B>>3],C,h,j,d)|0){Vc=ea;w=212;break b}sc=+p[h>>3]-+p[kc>>3];lc=+p[j>>3]-+p[Q>>3];mc=+X(+(sc*sc+lc*lc));p[N>>3]=mc}while(0);N=O+1|0;if((N|0)<(Ob|0)){O=N;L=Wc}else break}}}L=k[M>>2]|0;if((L|0)==5){if((Ob|0)==7)w=218}else if((Ob|0)==7&(L|0)==7)w=218;if((w|0)==218){w=0;p[kc+56>>3]=+p[kc+48>>3]}r:do if(Oc)switch(L|0){case 12:case 10:{O=kc+16|0;p[O>>3]=+p[B>>3]+ +p[O>>3];O=kc+24|0;p[O>>3]=+p[B>>3]+ +p[O>>3];break r;break}case 4:case 9:case 8:case 6:{O=kc+32|0;p[O>>3]=+p[B>>3]+ +p[O>>3];break r;break}case 5:case 7:{O=kc+48|0;p[O>>3]=+p[B>>3]+ +p[O>>3];O=kc+56|0;p[O>>3]=+p[B>>3]+ +p[O>>3];break r;break}case 14:case 13:{O=kc+16|0;p[O>>3]=+p[B>>3]+ +p[O>>3];O=kc+24|0;p[O>>3]=+p[B>>3]+ +p[O>>3];O=kc+80|0;p[O>>3]=+p[B>>3]+ +p[O>>3];break r;break}default:break r}while(0);Cs(I);if(!(i[vb>>0]|0)){ja=Oc;ka=ea;la=Rc;ma=Sc;break e}else{K=Oc;J=vb;na=Rc;oa=Sc}}}while(0);if(!(Bw(ea,ca,u)|0)){v=ka;w=226;break a}else{D=ca;E=ja;F=ka;G=la;H=ma}}switch(w|0){case 20:{tk(774798);k[d>>2]=113;Yc=V;w=228;break a;break}case 26:{tk(774894);k[d>>2]=431;Yc=fa;w=228;break a;break}case 34:{tk(774974);k[d>>2]=431;Yc=ha;w=228;break a;break}case 59:{tk(775027);k[d>>2]=431;Yc=va;w=228;break a;break}case 81:{tk(775132);tk(Ab);k[d>>2]=431;Yc=zb;w=228;break a;break}case 83:{tk(775184);tk(Cb);k[d>>2]=431;Yc=Bb;w=228;break a;break}case 90:{tk(775226);k[d>>2]=113;Yc=Eb;w=228;break a;break}case 98:{tk(775027);k[d>>2]=431;Yc=Lb;w=228;break a;break}case 103:{k[d>>2]=431;hc=Pb;ic=Qb;break}case 106:{k[d>>2]=431;hc=Rb;ic=Sb;break}case 109:{k[d>>2]=431;hc=Tb;ic=Ub;break}case 120:{k[d>>2]=431;hc=fc;ic=gc;break}case 139:{k[d>>2]=431;hc=dc;ic=ec;break}case 142:{k[d>>2]=431;hc=Vb;ic=Wb;break}case 145:{k[d>>2]=431;hc=Xb;ic=Yb;break}case 148:{k[d>>2]=431;hc=Zb;ic=_b;break}case 151:{k[d>>2]=431;hc=$b;ic=ac;break}case 160:{tk(775427);tk(cc);k[d>>2]=431;Yc=bc;w=228;break a;break}case 165:{tk(775513);k[d>>2]=113;Yc=jc;w=228;break a;break}case 202:{tk(775566);k[d>>2]=505;Yc=Mc;w=228;break a;break}case 204:{tk(775620);Zc=Nc;w=227;break a;break}case 212:{tk(775620);Zc=Vc;w=227;break a;break}}tk(775469);tk(ic);Zc=hc;w=227}while(0);if((w|0)==226){Ds(l);Zc=v;w=227}if((w|0)==227)if(!(k[d>>2]|0)){k[c>>2]=l;_c=Zc}else{Yc=Zc;w=228}if((w|0)==228){w=k[l>>2]|0;Zc=k[n>>2]|0;if((w|0)>0){c=Zc;v=w;w=0;while(1){if((k[c+(w*168|0)+4>>2]|0)==11){dy(k[c+(w*168|0)+48+4>>2]|0);$c=k[l>>2]|0;ad=k[n>>2]|0}else{$c=v;ad=c}w=w+1|0;if((w|0)>=($c|0)){bd=ad;break}else{c=ad;v=$c}}}else bd=Zc;if(bd)dy(bd);dy(l);_c=Yc}ww(u)|0;dy(_c);m=k[d>>2]|0;r=e;return m|0}function Bs(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0.0,fa=0.0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0.0,wa=0.0,xa=0,ya=0,za=0,Aa=0.0,Ba=0.0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0.0,Ha=0.0,Ia=0,Ja=0,Ka=0,La=0.0,Ma=0.0,Na=0.0,Oa=0,Pa=0,Qa=0.0,Ra=0,Sa=0,Ta=0.0,Ua=0.0,Va=0,Wa=0.0,Xa=0.0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0;e=r;r=r+608|0;f=e+80|0;g=e+72|0;h=e+68|0;j=e+64|0;l=e+40|0;m=e+24|0;n=e+16|0;o=e+8|0;q=e;s=e+32|0;t=e+530|0;u=e+104|0;py(u|0,775666,426)|0;v=k[d>>2]|0;if(v){w=v;r=e;return w|0}v=cy(80)|0;if(!v){tk(774798);k[d>>2]=113;w=113;r=e;return w|0}k[v>>2]=0;x=v+4|0;k[x>>2]=0;if((b|0)!=0?(k[b>>2]|0)!=0:0){y=v+8|0;z=b;A=y+72|0;do{k[y>>2]=k[z>>2];y=y+4|0;z=z+4|0}while((y|0)<(A|0))}else k[v+8>>2]=0;k[j>>2]=0;Tk(a,0,u,f,j)|0;z=u+71|0;B=f+4|0;Tk(a,0,z,B,j)|0;C=u+142|0;D=f+8|0;Tk(a,0,C,D,j)|0;E=u+213|0;F=f+12|0;Tk(a,0,E,F,j)|0;G=u+284|0;H=f+16|0;Tk(a,0,G,H,j)|0;if((k[j>>2]|0)!=0?(Jl(a,2,776092,1,d)|0)!=0:0){tk(776099);I=92}else I=11;a:do if((I|0)==11){if(Io(a,31,790243,v,t,d)|0){tk(776135);I=92;break}j=cy((k[v>>2]|0)*168|0)|0;k[x>>2]=j;if(!j){tk(775226);k[d>>2]=113;I=93;break}if(((((Tk(a,0,u,f,d)|0)==0?(Tk(a,0,z,B,d)|0)==0:0)?(Tk(a,0,C,D,d)|0)==0:0)?(Tk(a,0,E,F,d)|0)==0:0)?(Tk(a,0,G,H,d)|0)==0:0){j=f+20|0;J=(Tk(a,0,u+355|0,j,d)|0)==0;b:do if(k[v+8>>2]|0){K=cy(72)|0;if(!K){tk(776189);k[d>>2]=113;I=93;break a}k[K>>2]=1;L=K+8|0;M=K+16|0;N=K+24|0;O=K+32|0;P=K+40|0;Q=K+48|0;R=K+56|0;S=K+64|0;if(Ps(a,k[f>>2]|0,k[B>>2]|0,L,M,N,O,P,Q,R,S,d)|0){k[K>>2]=0;k[d>>2]=0;T=0;U=K;break}if((k[K>>2]|0)!=0?(k[b>>2]|0)!=0:0){do if(!(+W(+(+p[L>>3]-+p[b+8>>3]))>1.0e-06)){if(+W(+(+p[M>>3]-+p[b+16>>3]))>1.0e-06)break;if(+W(+(+p[N>>3]-+p[b+24>>3]))>1.0e-06)break;if(+W(+(+p[O>>3]-+p[b+32>>3]))>1.0e-06)break;if(+W(+(+p[P>>3]-+p[b+40>>3]))>1.0e-06)break;if(+W(+(+p[Q>>3]-+p[b+48>>3]))>1.0e-06)break;if(+W(+(+p[R>>3]-+p[b+56>>3]))>1.0e-06)break;if(tx(S,b+64|0)|0){T=0;U=K;break b}}while(0);T=1;U=K}else{T=0;U=K}}else{T=0;U=0}while(0);do if(((((Wo(a,k[f>>2]|0,1,g,l,d)|0)==0?(Wo(a,k[B>>2]|0,1,g,l+4|0,d)|0)==0:0)?(Wo(a,k[D>>2]|0,1,g,l+8|0,d)|0)==0:0)?(Wo(a,k[F>>2]|0,1,g,l+12|0,d)|0)==0:0)?(Wo(a,k[H>>2]|0,1,g,l+16|0,d)|0)==0:0){if(Wo(a,k[j>>2]|0,1,g,l+20|0,d)|0)break;S=cy(72)|0;k[s>>2]=S;if((k[v>>2]|0)<1){I=92;break a}S=(T|0)!=0;R=U+8|0;Q=U+16|0;P=U+24|0;O=U+32|0;N=U+40|0;M=U+48|0;L=U+56|0;V=U+64|0;Y=b+8|0;Z=b+16|0;_=b+24|0;$=b+32|0;aa=b+40|0;ba=b+48|0;ca=b+56|0;da=b+64|0;ea=0.0;fa=0.0;ga=1;ha=0;c:while(1){ia=ga+-1|0;ja=k[x>>2]|0;y=ja+(ha*168|0)+48|0;A=y+64|0;do{k[y>>2]=0;y=y+4|0}while((y|0)<(A|0));ka=ja+(ia*168|0)|0;la=ja+(ia*168|0)+48|0;ma=ja+(ia*168|0)+136|0;k[ma>>2]=0;k[ma+4>>2]=0;k[ma+8>>2]=0;k[ma+12>>2]=0;k[ma+16>>2]=0;k[ma+20>>2]=0;k[ma+24>>2]=0;k[ma+28>>2]=0;ma=((ga|0)<0)<<31>>31;if(In(a,k[D>>2]|0,ga,ma,1,0,1,0,776273,s,h,d)|0){I=38;break}i[ka>>0]=1;na=k[s>>2]|0;if((i[na>>0]|0)==33){i[ka>>0]=0;oa=na+1|0}else oa=na;na=ja+(ia*168|0)+4|0;if(!(tx(oa,776297)|0))k[na>>2]=0;if(!(tx(oa,776368)|0))k[na>>2]=2;if(!(tx(oa,776439)|0))k[na>>2]=4;if(!(tx(oa,776510)|0))k[na>>2]=3;if(!(tx(oa,776581)|0))k[na>>2]=5;if(!(tx(oa,776652)|0))k[na>>2]=6;if(!(tx(oa,776723)|0))k[na>>2]=6;if(!(tx(oa,776794)|0))k[na>>2]=7;do if(!(tx(oa,776865)|0)){k[na>>2]=8;pa=8;I=46}else{qa=k[na>>2]|0;if((qa|0)!=11){pa=qa;I=46;break}qa=k[l>>2]<<1;ra=ey(qa,8)|0;k[la+4>>2]=ra;if(!ra){I=45;break c}k[la>>2]=qa;qa=k[na>>2]|0;if((qa|0)!=11){sa=qa;ta=ra;ua=1;break}sa=11;ta=ra;ua=(k[la>>2]|0)/2|0}while(0);if((I|0)==46){I=0;sa=pa;ta=la;ua=1}ra=(sa|0)==8?2:ua;va=ea;wa=fa;qa=ta;xa=0;d:while(1){ya=qa;za=xa;while(1){if((za|0)>=(ra|0)){Aa=va;Ba=wa;Ca=ya;Da=ra;break d}Ea=za+1|0;Fa=((Ea|0)<0)<<31>>31;if(qm(a,k[f>>2]|0,ga,ma,Ea,Fa,1,0,-9.1191291391491e-36,ya,h,d)|0){I=53;break c}if(+p[ya>>3]==-9.1191291391491e-36){Ga=va;Ha=wa;Ia=ya;Ja=za;I=55;break d}Ka=ya+8|0;if(qm(a,k[B>>2]|0,ga,ma,Ea,Fa,1,0,-9.1191291391491e-36,Ka,h,d)|0){I=57;break c}La=+p[Ka>>3];if(La==-9.1191291391491e-36){Ma=va;Na=wa;Oa=ya;Pa=za;I=59;break d}Ka=ya+16|0;Qa=+p[ya>>3];if(!za){va=Qa;wa=La;qa=Ka;xa=Ea;continue d}if(wa==La&va==Qa){Ra=Ea;Sa=Ka;Ta=va;Ua=wa;I=62;break d}else{ya=Ka;za=Ea}}}if((I|0)==55){I=0;k[la>>2]=Ja<<1;Aa=Ga;Ba=Ha;Ca=Ia;Da=Ja}else if((I|0)==59){I=0;k[la>>2]=Pa<<1;Aa=Ma;Ba=Na;Ca=Oa;Da=Pa}else if((I|0)==62){I=0;k[la>>2]=Ra<<1;Aa=Ta;Ba=Ua;Ca=Sa;Da=Ra}do if(S){xa=Da<<1;qa=0-xa|0;wa=+p[Ca+(qa<<3)>>3];va=+p[Ca+((qa|1)<<3)>>3];if((Da|0)>0)Va=0;else{Wa=wa;Xa=va;break}while(1){qa=Va<<1;ra=Ca+(qa-xa<<3)|0;za=Ca+((qa|1)-xa<<3)|0;Qs(+p[ra>>3],+p[za>>3],+p[R>>3],+p[Q>>3],+p[P>>3],+p[O>>3],+p[N>>3],+p[M>>3],+p[L>>3],V,o,q,d)|0;Rs(+p[o>>3],+p[q>>3],+p[Y>>3],+p[Z>>3],+p[_>>3],+p[$>>3],+p[aa>>3],+p[ba>>3],+p[ca>>3],da,ra,za,d)|0;Va=Va+1|0;if(k[d>>2]|0){I=67;break c}if((Va|0)>=(Da|0)){Wa=wa;Xa=va;break}}}else{Wa=Aa;Xa=Ba}while(0);xa=k[na>>2]|0;switch(xa|0){case 2:{Ya=1;I=71;break}case 10:case 3:case 4:case 9:case 6:{Ya=2;I=71;break}case 5:case 7:{Ya=4;I=71;break}default:{Za=xa;_a=Ca}}if((I|0)==71){I=0;if(qm(a,k[F>>2]|0,ga,ma,1,0,Ya,0,0.0,Ca,h,d)|0){I=72;break}if(S){xa=ja+(ia*168|0)+56|0;za=Ca;ra=0;while(1){va=Xa+ +p[za>>3];p[n>>3]=va;p[m>>3]=Wa;Qs(Wa,va,+p[R>>3],+p[Q>>3],+p[P>>3],+p[O>>3],+p[N>>3],+p[M>>3],+p[L>>3],V,o,q,d)|0;Rs(+p[o>>3],+p[q>>3],+p[Y>>3],+p[Z>>3],+p[_>>3],+p[$>>3],+p[aa>>3],+p[ba>>3],+p[ca>>3],da,m,n,d)|0;if(k[d>>2]|0){I=76;break c}va=+p[m>>3]-+p[la>>3];wa=+p[n>>3]-+p[xa>>3];Qa=+X(+(va*va+wa*wa));qa=za+8|0;p[za>>3]=Qa;ra=ra+1|0;if((ra|0)>=(Ya|0)){$a=qa;break}else za=qa}}else $a=Ca+(Ya<<3)|0;Za=k[na>>2]|0;_a=$a}switch(Za|0){case 4:case 9:case 6:{ab=1;I=82;break}case 10:case 5:case 7:{ab=2;I=82;break}default:{}}do if((I|0)==82){I=0;if(qm(a,k[H>>2]|0,ga,ma,1,0,ab,0,0.0,_a,h,d)|0){I=83;break c}if(!S)break;Qa=+p[ca>>3]-+p[L>>3];za=_a;ra=0;while(1){p[za>>3]=Qa+ +p[za>>3];ra=ra+1|0;if((ra|0)>=(ab|0))break;else za=za+8|0}}while(0);if(J){if(_l(a,31,k[j>>2]|0,ga,ma,1,0,1,0,0,ja+(ia*168|0)+8|0,h,d)|0){I=89;break}}else k[ja+(ia*168|0)+8>>2]=1;Cs(ka);if((ga|0)<(k[v>>2]|0)){ea=Wa;fa=Xa;ga=ga+1|0;ha=ha+1|0}else{I=92;break a}}if((I|0)==38){tk(776275);I=92;break a}else if((I|0)==45){tk(775513);k[d>>2]=113;I=93;break a}else if((I|0)==53){tk(777504);I=92;break a}else if((I|0)==57){tk(777547);I=92;break a}else if((I|0)==67){tk(777590);I=92;break a}else if((I|0)==72){tk(777622);I=92;break a}else if((I|0)==76){tk(777590);I=92;break a}else if((I|0)==83){tk(777657);I=92;break a}else if((I|0)==89){tk(777697);I=92;break a}}while(0);tk(776235);I=92;break}tk(776166);I=92}while(0);if((I|0)==92)if(!(k[d>>2]|0))k[c>>2]=v;else I=93;if((I|0)==93){I=k[v>>2]|0;c=k[x>>2]|0;if((I|0)>0){h=c;ab=I;I=0;while(1){if((k[h+(I*168|0)+4>>2]|0)==11){dy(k[h+(I*168|0)+48+4>>2]|0);bb=k[v>>2]|0;cb=k[x>>2]|0}else{bb=ab;cb=h}I=I+1|0;if((I|0)>=(bb|0)){db=cb;break}else{h=cb;ab=bb}}}else db=c;if(db)dy(db);dy(v)}Fh(a,d)|0;w=k[d>>2]|0;r=e;return w|0}function Cs(a){a=a|0;var b=0,c=0,d=0.0,e=0,f=0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0,r=0,s=0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0;b=k[a+4>>2]|0;c=a+48|0;a:do if((b|0)!=11){do switch(b|0){case 2:{d=+p[a+64>>3];p[a+152>>3]=d*d;e=52;break}case 3:{d=+p[a+64>>3];p[a+152>>3]=d*d;d=+p[a+72>>3];p[a+160>>3]=d*d;e=53;break}case 10:{f=a+64|0;d=+p[f>>3];if(d>180.0){g=d;while(1){h=g+-360.0;if(h>180.0)g=h;else{i=h;break}}p[f>>3]=i;j=i}else j=d;if(j<=-180.0){g=j;while(1){h=g+360.0;if(!(h<=-180.0)){l=h;break}else g=h}p[f>>3]=l}m=a+72|0;g=+p[m>>3];if(g>180.0){d=g;while(1){h=d+-360.0;if(h>180.0)d=h;else{n=h;break}}p[m>>3]=n;o=n}else o=g;if(!(o<=-180.0))e=51;else{d=o;while(1){h=d+360.0;if(!(h<=-180.0)){q=h;break}else d=h}p[m>>3]=q;e=51}break}case 4:{f=a+80|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+136>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+144>>3]=d;e=54;break}case 5:{f=a+96|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+152>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+160>>3]=d;f=a+104|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+136>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+144>>3]=d;e=55;break}case 6:{f=a+80|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+136>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+144>>3]=d;e=56;break}case 7:{f=a+96|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+152>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+160>>3]=d;f=a+104|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+136>>3]=d;d=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+144>>3]=d;e=57;break}case 8:{f=a+80|0;d=+_(+(+p[f>>3]/180.0*3.141592653589793));p[a+136>>3]=d;g=+Z(+(+p[f>>3]/180.0*3.141592653589793));p[a+144>>3]=g;f=a+64|0;h=(+p[f>>3]-+p[c>>3])*.5;r=a+72|0;s=a+56|0;t=(+p[r>>3]-+p[s>>3])*.5;u=+W(+(g*h+d*t));p[a+152>>3]=u;u=+W(+(t*g-h*d));p[a+160>>3]=u;p[a+88>>3]=(+p[f>>3]+ +p[c>>3])*.5;p[a+96>>3]=(+p[r>>3]+ +p[s>>3])*.5;e=67;break}case 9:{s=a+80|0;u=+_(+(+p[s>>3]/180.0*3.141592653589793));p[a+136>>3]=u;u=+Z(+(+p[s>>3]/180.0*3.141592653589793));p[a+144>>3]=u;e=58;break}case 1:{u=+p[a+64>>3]-+p[c>>3];d=+p[a+72>>3]-+p[a+56>>3];h=+X(+(u*u+d*d));s=h!=0.0;p[a+136>>3]=s?d/h:0.0;p[a+144>>3]=s?u/h:1.0;p[a+152>>3]=h+.5;e=78;break}case 12:{s=a+64|0;h=+p[s>>3];if(h>180.0){u=h;while(1){d=u+-360.0;if(d>180.0)u=d;else{v=d;break}}p[s>>3]=v;w=v}else w=h;if(w<=-180.0){u=w;while(1){d=u+360.0;if(!(d<=-180.0)){x=d;break}else u=d}p[s>>3]=x}m=a+72|0;u=+p[m>>3];if(u>180.0){h=u;while(1){d=h+-360.0;if(d>180.0)h=d;else{y=d;break}}p[m>>3]=y;z=y}else z=u;if(z<=-180.0){h=z;while(1){d=h+360.0;if(!(d<=-180.0)){A=d;break}else h=d}p[m>>3]=A}h=+p[a+88>>3];p[a+152>>3]=h*h;h=+p[a+96>>3];p[a+160>>3]=h*h;e=51;break}case 14:case 13:{s=a+64|0;h=+p[s>>3];if(h>180.0){u=h;while(1){d=u+-360.0;if(d>180.0)u=d;else{B=d;break}}p[s>>3]=B;C=B}else C=h;if(C<=-180.0){u=C;while(1){d=u+360.0;if(!(d<=-180.0)){D=d;break}else u=d}p[s>>3]=D}m=a+72|0;u=+p[m>>3];if(u>180.0){h=u;while(1){d=h+-360.0;if(d>180.0)h=d;else{E=d;break}}p[m>>3]=E;F=E}else F=u;if(F<=-180.0){h=F;while(1){d=h+360.0;if(!(d<=-180.0)){G=d;break}else h=d}p[m>>3]=G}s=a+128|0;h=+_(+(+p[s>>3]/180.0*3.141592653589793));p[a+136>>3]=h;h=+Z(+(+p[s>>3]/180.0*3.141592653589793));p[a+144>>3]=h;e=51;break}case 0:{H=1.0;e=65;break}default:return}while(0);b:do if((e|0)==51)switch(b|0){case 11:{I=c;break a;break}case 2:{e=52;break b;break}case 3:{e=53;break b;break}case 4:{e=54;break b;break}case 5:{e=55;break b;break}case 6:{e=56;break b;break}case 7:{e=57;break b;break}case 9:{e=58;break b;break}case 0:{H=1.0;e=65;break b;break}case 8:{e=67;break b;break}case 1:{e=78;break b;break}case 10:{e=81;break b;break}case 12:{J=+p[a+96>>3];e=64;break b;break}case 13:{h=+p[a+104>>3];u=+p[a+112>>3];J=h>u?h:u;e=64;break b;break}case 14:{u=+p[a+104>>3]*+p[a+112>>3];J=+X(+(u+u))*.5;e=64;break b;break}default:return}while(0);do if((e|0)==52){J=+p[a+64>>3];e=64}else if((e|0)==53){J=+p[a+72>>3];e=64}else if((e|0)==54){u=+p[a+64>>3];h=+p[a+72>>3];J=u>h?u:h;e=64}else if((e|0)==55){h=+p[a+80>>3];u=+p[a+88>>3];J=h>u?h:u;e=64}else if((e|0)==56){u=+p[a+64>>3];h=+p[a+72>>3];J=+X(+(u*u+h*h))*.5;e=64}else if((e|0)==57){h=+p[a+80>>3]*+p[a+88>>3];J=+X(+(h+h))*.5;e=64}else if((e|0)==58){h=+p[a+64>>3];u=+p[a+72>>3];if(h>u){J=h*.5;e=64;break}else{J=u*.5;e=64;break}}while(0);c:do if((e|0)==64)if(J>0.0){H=J;e=65}else{switch(b|0){case 11:{I=c;break a;break}case 8:{e=67;break c;break}case 1:{e=78;break c;break}case 10:{e=81;break c;break}default:{}}return}while(0);if((e|0)==65){p[a+16>>3]=+p[c>>3]-H;p[a+24>>3]=H+ +p[c>>3];s=a+56|0;p[a+32>>3]=+p[s>>3]-H;p[a+40>>3]=H+ +p[s>>3];return}else if((e|0)==67){s=a+88|0;u=+p[s>>3];h=u-+p[c>>3];r=a+96|0;d=+p[r>>3]-+p[a+56>>3];g=+X(+(h*h+d*d));p[a+16>>3]=u-g;p[a+24>>3]=g+ +p[s>>3];p[a+32>>3]=+p[r>>3]-g;p[a+40>>3]=g+ +p[r>>3];return}else if((e|0)==78){g=+p[c>>3];r=a+64|0;u=+p[r>>3];s=g>u;p[a+16>>3]=s?u:g;p[a+24>>3]=s?+p[c>>3]:+p[r>>3];r=a+56|0;g=+p[r>>3];s=a+72|0;u=+p[s>>3];f=a+32|0;if(g>u){p[f>>3]=u;p[a+40>>3]=+p[r>>3];return}else{p[f>>3]=g;p[a+40>>3]=+p[s>>3];return}}else if((e|0)==81){p[a+16>>3]=1.0;p[a+24>>3]=-1.0;p[a+32>>3]=1.0;p[a+40>>3]=-1.0;return}}else I=k[c+4>>2]|0;while(0);H=+p[I>>3];e=a+16|0;p[e>>3]=H;J=+p[I>>3];b=a+24|0;p[b>>3]=J;s=I+8|0;G=+p[s>>3];f=a+32|0;p[f>>3]=G;F=+p[s>>3];s=a+40|0;p[s>>3]=F;a=k[c>>2]|0;if((a|0)>2){K=H;L=J;M=G;N=F;O=2}else return;while(1){c=I+(O<<3)|0;F=+p[c>>3];if(K>F){p[e>>3]=F;P=+p[c>>3];Q=F}else{P=F;Q=K}if(L<P){p[b>>3]=P;R=P}else R=L;c=I+((O|1)<<3)|0;F=+p[c>>3];if(M>F){p[f>>3]=F;S=+p[c>>3];T=F}else{S=F;T=M}if(N<S){p[s>>3]=S;U=S}else U=N;O=O+2|0;if((O|0)>=(a|0))break;else{K=Q;L=R;M=T;N=U}}return}function Ds(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0;if((k[a>>2]|0)<=0)return;b=a+4|0;c=0;while(1){d=k[b>>2]|0;a:do if(!(i[d+(c*168|0)>>0]|0)){e=c+-1|0;while(1){if((e|0)<=0){f=e;g=9;break}h=e+-1|0;if(!(i[d+(e*168|0)>>0]|0))e=h;else{j=h;break}}if((g|0)==9){g=0;j=f+-1|0}if((j|0)>-1){e=d;h=c;l=j;while(1){if(!(i[e+(l*168|0)>>0]|0))m=h;else{n=fy(e,((k[a>>2]|0)*168|0)+168|0)|0;k[b>>2]=n;o=k[a>>2]|0;k[a>>2]=o+1;p=l+1|0;if((o|0)>(p|0)){q=n;r=o;while(1){o=r;r=r+-1|0;py(q+(o*168|0)|0,q+(r*168|0)|0,168)|0;o=k[b>>2]|0;if((r|0)<=(p|0)){s=o;break}else q=o}}else s=n;q=h+1|0;py(s+(p*168|0)|0,s+(q*168|0)|0,168)|0;m=q}if((l|0)<=0){t=m;break a}e=k[b>>2]|0;h=m;l=l+-1|0}}else t=c}else t=c;while(0);c=t+1|0;d=k[a>>2]|0;if((c|0)>=(d|0)){u=d;break}}if((u|0)<=0)return;c=k[a+4>>2]|0;a=0;t=0;do{t=((i[c+(a*168|0)>>0]|0)!=0&1)+t|0;k[c+(a*168|0)+8>>2]=t;a=a+1|0}while((a|0)<(u|0));return}function Es(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0;b=k[a>>2]|0;c=a+4|0;d=k[c>>2]|0;if((b|0)>0){e=b;b=d;f=0;while(1){if((k[b+(f*168|0)+4>>2]|0)==11){dy(k[b+(f*168|0)+48+4>>2]|0);g=k[a>>2]|0;h=k[c>>2]|0}else{g=e;h=b}f=f+1|0;if((f|0)>=(g|0)){i=h;break}else{e=g;b=h}}}else i=d;if(!i){dy(a);return}dy(i);dy(a);return}function Fs(a,b,c){a=+a;b=+b;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0,F=0.0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0.0;d=k[c+4>>2]|0;e=k[c>>2]|0;if((e|0)<=0){f=0;g=0;h=f|g;j=(h|0)!=0;l=j&1;return l|0}c=d;m=0;n=k[d+8>>2]|0;d=0;o=0;while(1){q=k[c+8>>2]|0;if((d|0)!=0&(q|0)==(n|0)){r=i[c>>0]|0;s=m;t=n;u=o}else{v=i[c>>0]|0;r=v;s=v<<24>>24==0&1;t=q;u=(m|o|0)!=0&1}q=r<<24>>24==0;if(!s)if(q)w=0;else x=9;else if(q)x=9;else w=s;if((x|0)==9){x=0;a:do switch(k[c+4>>2]|0){case 6:{y=a-+p[c+48>>3];z=b-+p[c+56>>3];A=+p[c+144>>3];B=+p[c+136>>3];C=y*A+z*B;D=z*A-y*B;B=+p[c+64>>3]*.5;y=+p[c+72>>3]*.5;E=(D>y|(C<-B|C>B|D<-y))&1^1;break}case 7:{y=a-+p[c+48>>3];D=b-+p[c+56>>3];B=+p[c+144>>3];C=+p[c+136>>3];A=y*B+D*C;z=-y;F=D*B+C*z;C=+p[c+80>>3]*.5;B=+p[c+88>>3]*.5;if(F>B|(A<-C|A>C|F<-B))E=0;else{B=+p[c+160>>3];F=+p[c+152>>3];C=y*B+D*F;y=D*B+F*z;z=+p[c+64>>3]*.5;F=+p[c+72>>3]*.5;E=(!(y<=F)|(!(C>=-z)|!(C<=z)|!(y>=-F)))&1}break}case 8:{F=a-+p[c+88>>3];y=b-+p[c+96>>3];z=+p[c+144>>3];C=+p[c+136>>3];B=F*z+y*C;D=y*z-F*C;C=+p[c+152>>3];F=+p[c+160>>3];E=(D>F|(D<-F|(B<-C|B>C)))&1^1;break}case 9:{C=a-+p[c+48>>3];B=b-+p[c+56>>3];F=+p[c+144>>3];D=+p[c+136>>3];if(+W(+((C*F+B*D)/(+p[c+64>>3]*.5)))+ +W(+((B*F-C*D)/(+p[c+72>>3]*.5)))>1.0)E=0;else E=1;break}case 2:{D=a-+p[c+48>>3];C=b-+p[c+56>>3];if(D*D+C*C>+p[c+152>>3])E=0;else E=1;break}case 3:{C=a-+p[c+48>>3];D=b-+p[c+56>>3];F=C*C+D*D;if(!(F<+p[c+152>>3])?!(F>+p[c+160>>3]):0){E=1;break a}E=0;break}case 10:{F=a-+p[c+48>>3];D=b-+p[c+56>>3];if(F!=0.0|D!=0.0){C=+da(+D,+F)*180.0/3.141592653589793;F=+p[c+64>>3];D=+p[c+72>>3];v=C<F;G=C>D;if(!(F<=D)){E=v&G&1^1;break a}else{E=(v|G)&1^1;break a}}else E=1;break}case 4:{D=a-+p[c+48>>3];F=b-+p[c+56>>3];C=+p[c+144>>3];B=+p[c+136>>3];z=(D*C+F*B)/+p[c+64>>3];y=(F*C-D*B)/+p[c+72>>3];if(z*z+y*y>1.0)E=0;else E=1;break}case 5:{y=a-+p[c+48>>3];z=b-+p[c+56>>3];B=+p[c+144>>3];D=+p[c+136>>3];C=-y;F=(y*B+z*D)/+p[c+80>>3];A=(z*B+D*C)/+p[c+88>>3];if(!(F*F+A*A>1.0)?(A=+p[c+160>>3],F=+p[c+152>>3],D=(y*A+z*F)/+p[c+64>>3],y=(z*A+F*C)/+p[c+72>>3],!(D*D+y*y<1.0)):0)E=1;else E=0;break}case 1:{y=a-+p[c+48>>3];D=b-+p[c+56>>3];C=+p[c+144>>3];F=+p[c+136>>3];A=y*C+D*F;z=D*C-y*F;if(!(A<-.5|(z<-.5|z>=.5))?!(A>=+p[c+152>>3]):0){E=1;break a}E=0;break}case 0:{A=a-+p[c+48>>3];z=b-+p[c+56>>3];E=(z>=.5|(A<-.5|A>=.5|z<-.5))&1^1;break}case 11:{if((((!(+p[c+16>>3]>a)?!(+p[c+24>>3]<a):0)?!(+p[c+32>>3]>b):0)?!(+p[c+40>>3]<b):0)?(G=c+48|0,v=k[G>>2]|0,H=k[G+4>>2]|0,G=v+-1|0,(v|0)>0):0){I=0;J=0;z=+p[H+(v+-2<<3)>>3];A=+p[H+(G<<3)>>3];while(1){F=z;z=+p[H+(J<<3)>>3];K=J|1;y=A;A=+p[H+(K<<3)>>3];do if((!(y<b)|!(A<=b)?!(y>b)|!(A>=b):0)?!(F<a)|!(z<=a):0){if(F<=a|z<a){C=b-y;D=A-y;if(+W(+D)<1.0e-10)if(+W(+C)<1.0e-10){E=1;break a}else{L=I;break}B=F+C*((z-F)/D)-a;if(B<-1.0e-10){L=I;break}if(B<1.0e-10){E=1;break a}}if(y!=b){L=1-I|0;break}else M=K;do{M=(M|0)>1?M+-2|0:G;B=+p[H+(M<<3)>>3]}while(B==b);N=B;if((A-b)*(b-N)>0.0)L=1-I|0;else L=I}else L=I;while(0);J=J+2|0;if((J|0)>=(v|0)){E=L;break}else I=L}}else E=0;break}case 12:{A=a-+p[c+48>>3];z=b-+p[c+56>>3];y=A*A+z*z;if(!(y<+p[c+152>>3])?!(y>+p[c+160>>3]):0)if(A!=0.0|z!=0.0){y=+da(+z,+A)*180.0/3.141592653589793;A=+p[c+64>>3];z=+p[c+72>>3];I=y<A;v=y>z;if(!(A<=z)){E=I&v&1^1;break a}else{E=(I|v)&1^1;break a}}else E=1;else E=0;break}case 13:{z=a-+p[c+48>>3];A=b-+p[c+56>>3];y=+p[c+144>>3];F=+p[c+136>>3];B=z*y+A*F;D=A*y-z*F;F=B/+p[c+104>>3];z=D/+p[c+112>>3];if(!(F*F+z*z>1.0)?(z=B/+p[c+88>>3],F=D/+p[c+96>>3],!(z*z+F*F<1.0)):0)if(B!=0.0|D!=0.0){F=+da(+D,+B)*180.0/3.141592653589793;B=+p[c+64>>3];D=+p[c+72>>3];v=F<B;I=F>D;if(!(B<=D)){E=v&I&1^1;break a}else{E=(v|I)&1^1;break a}}else E=1;else E=0;break}case 14:{D=a-+p[c+48>>3];B=b-+p[c+56>>3];F=+p[c+144>>3];z=+p[c+136>>3];y=D*F+B*z;A=B*F-D*z;z=+p[c+104>>3]*.5;D=+p[c+112>>3]*.5;if(!(A>D|(y<-z|y>z|A<-D))?(D=+p[c+88>>3]*.5,z=+p[c+96>>3]*.5,!(A<=z)|(!(y>=-D)|!(y<=D)|!(A>=-z))):0)if(y!=0.0|A!=0.0){z=+da(+A,+y)*180.0/3.141592653589793;y=+p[c+64>>3];A=+p[c+72>>3];I=z<y;v=z>A;if(!(y<=A)){E=I&v&1^1;break a}else{E=(I|v)&1^1;break a}}else E=1;else E=0;break}default:E=1}while(0);if(q)w=(E|0)==0&1;else w=E}d=d+1|0;if((d|0)>=(e|0)){f=w;g=u;break}else{c=c+168|0;m=w;n=t;o=u}}h=f|g;j=(h|0)!=0;l=j&1;return l|0}function Gs(a,b,c,d){a=a|0;b=+b;c=+c;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;do if((g|0)<=0){if(b==0.0){k[d>>2]=322;h=322;break}if((Zk(a,f,d)|0)>0){h=k[d>>2]|0;break}if(k[f>>2]|0){k[d>>2]=233;h=233;break}i=(Cl(a,d)|0)==0;j=k[a+4>>2]|0;if(i){i=k[j+960>>2]|0;p[i+256>>3]=b;p[i+264>>3]=c;h=k[d>>2]|0;break}else{p[j+1176>>3]=b;p[j+1184>>3]=c;h=k[d>>2]|0;break}}else h=g;while(0);r=e;return h|0}function Hs(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;do if((g|0)<=0){if((Zk(a,f,d)|0)>0){h=k[d>>2]|0;break}if(k[f>>2]|0){k[d>>2]=233;h=233;break}if(!(Cl(a,d)|0)){i=(k[(k[a+4>>2]|0)+960>>2]|0)+272|0;k[i>>2]=b;k[i+4>>2]=c;h=k[d>>2]|0;break}else{h=k[d>>2]|0;break}}else h=g;while(0);r=e;return h|0}function Is(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0;f=r;r=r+16|0;g=f;h=k[e>>2]|0;if((h|0)>0){i=h;r=f;return i|0}if((Zk(a,g,e)|0)>0){i=k[e>>2]|0;r=f;return i|0}if((k[g>>2]|0)==2){g=(k[(k[a+4>>2]|0)+960>>2]|0)+((b+-1|0)*160|0)+112|0;k[g>>2]=c;k[g+4>>2]=d;i=k[e>>2]|0;r=f;return i|0}else{k[e>>2]=227;i=227;r=f;return i|0}return 0}function Js(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;e=r;r=r+16|0;f=e;g=k[d>>2]|0;if((g|0)>0){h=g;r=e;return h|0}if((Zk(a,f,d)|0)>0){h=k[d>>2]|0;r=e;return h|0}if((k[f>>2]|0)==1){f=(k[(k[a+4>>2]|0)+960>>2]|0)+((b+-1|0)*160|0)+120|0;i[f>>0]=0;zx(f,c,19)|0;h=k[d>>2]|0;r=e;return h|0}else{k[d>>2]=226;h=226;r=e;return h|0}return 0}function Ks(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;if((b|0)>0)c=0;else return;do{d=a+(c<<1)|0;e=m[d>>1]|0;j[d>>1]=e>>>8|e<<8;c=c+1|0}while((c|0)!=(b|0));return}function Ls(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;if((b|0)>0)c=0;else return;do{d=a+(c<<2)|0;e=i[d>>0]|0;f=d+3|0;i[d>>0]=i[f>>0]|0;i[f>>0]=e;e=d+1|0;f=i[e>>0]|0;g=d+2|0;i[e>>0]=i[g>>0]|0;i[g>>0]=f;c=c+1|0}while((c|0)!=(b|0));return}function Ms(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0;c=b<<3;if((b|0)>0)d=0;else return;do{b=a+d|0;e=i[b>>0]|0;f=a+(d|7)|0;i[b>>0]=i[f>>0]|0;i[f>>0]=e;e=a+(d|1)|0;f=i[e>>0]|0;b=a+(d|6)|0;i[e>>0]=i[b>>0]|0;i[b>>0]=f;f=a+(d|2)|0;b=i[f>>0]|0;e=a+(d|5)|0;i[f>>0]=i[e>>0]|0;i[e>>0]=b;b=a+(d|3)|0;e=i[b>>0]|0;f=a+(d|4)|0;i[b>>0]=i[f>>0]|0;i[f>>0]=e;d=d+8|0}while((d|0)<(c|0));return}function Ns(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=r;r=r+16|0;e=d;f=k[c>>2]|0;do if((f|0)<=0){Zk(a,e,c)|0;if(k[e>>2]|0){tk(777740);k[c>>2]=233;g=233;break}if((dp(a,b,c)|0)>0){tk(777808);g=k[c>>2]|0;break}else{g=k[c>>2]|0;break}}else g=f;while(0);r=d;return g|0}function Os(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0.0,F=0.0,G=0.0,H=0.0,I=0,J=0,K=0,L=0.0,M=0.0;m=r;r=r+144|0;n=m+64|0;o=m+68|0;q=m+56|0;s=m+48|0;t=m+40|0;u=m+32|0;v=m+24|0;w=m+16|0;x=m+8|0;y=m;k[n>>2]=0;p[q>>3]=0.0;p[s>>3]=0.0;p[t>>3]=0.0;p[u>>3]=0.0;p[v>>3]=1.0;p[w>>3]=0.0;p[x>>3]=1.0;p[y>>3]=0.0;z=k[l>>2]|0;if((z|0)>0){A=z;r=m;return A|0}k[n>>2]=0;if(Mo(a,777862,b,0,n)|0)p[b>>3]=0.0;k[n>>2]=0;if(Mo(a,777869,c,0,n)|0)p[c>>3]=0.0;k[n>>2]=0;if(Mo(a,777876,d,0,n)|0)p[d>>3]=0.0;k[n>>2]=0;if(Mo(a,777883,e,0,n)|0)p[e>>3]=0.0;k[n>>2]=0;do if(!(Mo(a,777890,f,0,n)|0)){if(Mo(a,777921,g,0,n)|0)p[g>>3]=1.0;k[n>>2]=0;if(Mo(a,777928,h,0,n)|0){p[h>>3]=0.0;k[n>>2]=0;if(!(Mo(a,777935,v,0,n)|0))B=1;else{k[n>>2]=0;B=0}if(!(Mo(a,777941,w,0,n)|0))C=1;else{k[n>>2]=0;C=B}if(!(Mo(a,777947,y,0,n)|0))D=1;else{k[n>>2]=0;D=C}if((Mo(a,777953,x,0,n)|0)!=0?(k[n>>2]=0,(D|0)==0):0)break;E=+da(+(+p[w>>3]),+(+p[v>>3]));F=+da(+-+p[y>>3],+(+p[x>>3]));G=E<F?E:F;H=E>F?E:F;F=H-G>1.5707963267948966?G+3.141592653589793:G;if(+W(+(F-H))>.0002)k[l>>2]=506;p[h>>3]=(H+F)*.5*180.0/3.141592653589793}}else{k[n>>2]=0;if(!(Mo(a,777897,q,0,n)|0))I=1;else{k[n>>2]=0;I=0}if(!(Mo(a,777903,s,0,n)|0))J=1;else{k[n>>2]=0;J=I}if(!(Mo(a,777909,u,0,n)|0))K=1;else{k[n>>2]=0;K=J}if((Mo(a,777915,t,0,n)|0)!=0?(k[n>>2]=0,(K|0)==0):0){p[f>>3]=1.0;k[n>>2]=0;if(Mo(a,777921,g,0,n)|0)p[g>>3]=1.0;k[n>>2]=0;if(!(Mo(a,777928,h,0,n)|0))break;p[h>>3]=0.0;break}F=+p[q>>3];H=+da(+(+p[s>>3]),+F);G=+p[t>>3];E=+da(+-+p[u>>3],+G);L=H<E?H:E;M=H>E?H:E;E=M-L>1.5707963267948966?L+3.141592653589793:L;if(+W(+(E-M))>.0002)k[l>>2]=506;L=(M+E)*.5;E=+Z(+L);p[f>>3]=F/E;p[g>>3]=G/E;p[h>>3]=L*180.0/3.141592653589793;if(+p[g>>3]<0.0){p[f>>3]=-+p[f>>3];p[g>>3]=-+p[g>>3];p[h>>3]=+p[h>>3]+-180.0}}while(0);k[n>>2]=0;do if(!(Jo(a,777959,o,0,n)|0)){Bx(j,o+4|0,4)|0;i[j+4>>0]=0;if((Ax(o,777966,4)|0)!=0?(Ax(o+1|0,777971,3)|0)!=0:0)break;p[h>>3]=90.0-+p[h>>3];p[g>>3]=-+p[g>>3];L=+p[b>>3];p[b>>3]=+p[c>>3];p[c>>3]=L}else i[j>>0]=0;while(0);A=k[l>>2]|0;r=m;return A|0}function Ps(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0;n=r;r=r+32|0;o=n+16|0;p=n+8|0;q=n;s=k[m>>2]|0;if((s|0)>0){t=s;r=n;return t|0}k[o>>2]=b;k[o+4>>2]=c;k[p>>2]=10;k[p+4>>2]=10;Lh(q,777975,m)|0;Xr(k[q>>2]|0,32,2,p,m)|0;Lk(a,k[q>>2]|0,9,2,o,m)|0;Up(a,k[q>>2]|0,2,o,m)|0;o=k[m>>2]|0;if((o|0)>0){t=o;r=n;return t|0}Os(k[q>>2]|0,d,e,f,g,h,i,j,l,m)|0;if((k[m>>2]|0)>0){tk(777982);k[m>>2]=505;t=505;r=n;return t|0}else{ci(k[q>>2]|0,m)|0;t=k[m>>2]|0;r=n;return t|0}return 0}function Qs(a,b,c,d,e,f,g,h,j,l,m,n,o){a=+a;b=+b;c=+c;d=+d;e=+e;f=+f;g=+g;h=+h;j=+j;l=l|0;m=m|0;n=n|0;o=o|0;var q=0,r=0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0,F=0.0;q=k[o>>2]|0;if((q|0)>0){r=q;return r|0}s=(a-e)*g;e=(b-f)*h;f=j*.01745329252;b=+Z(+f);a=+_(+f);if(j!=0.0){t=s*b-e*a;u=e*b+s*a}else{t=s;u=e}e=c*.01745329252;c=d*.01745329252;s=t*.01745329252;t=u*.01745329252;u=s*s;j=t*t+u;f=+Z(+c);v=+_(+c);if((i[l>>0]|0)!=45){k[o>>2]=504;r=504;return r|0}a:do switch(i[l+1>>0]|0){case 67:{if((i[l+2>>0]|0)==65?(i[l+3>>0]|0)==82:0){w=c+t;x=e+s;break a}k[o>>2]=504;r=504;return r|0}case 84:{if((i[l+2>>0]|0)==65?(i[l+3>>0]|0)==78:0){y=+Z(+e);z=+_(+e);A=f*y-s*z-v*(t*y);B=s*y+f*z-v*(t*z);w=+ca(+((v+t*f)/+X(+(A*A+B*B))));x=+da(+B,+A);break a}k[o>>2]=504;r=504;return r|0}case 83:{switch(i[l+2>>0]|0){case 73:{if((i[l+3>>0]|0)==78){if(j>1.0){k[o>>2]=501;r=501;return r|0}A=+X(+(1.0-j));B=t*f+v*A;if(B>1.0|B<-1.0){k[o>>2]=501;r=501;return r|0}z=f*A-t*v;if(!(s==0.0&z==0.0)){w=+ba(+B);x=e+ +da(+s,+z);break a}k[o>>2]=501;r=501;return r|0}break}case 84:{if((i[l+3>>0]|0)==71){z=(4.0-j)/(j+4.0);if(+W(+z)>1.0){k[o>>2]=501;r=501;return r|0}B=z+1.0;A=v*z+t*f*B*.5;if(+W(+A)>1.0){k[o>>2]=501;r=501;return r|0}z=+ba(+A);A=+Z(+z);if(+W(+A)<1.0e-05){k[o>>2]=501;r=501;return r|0}y=s*B/(A*2.0);if(+W(+y)>1.0){k[o>>2]=501;r=501;return r|0}B=+ba(+y);y=+_(+z);C=+Z(+B);D=v*y+1.0+f*A*C;if(!(+W(+D)<1.0e-05)){E=+W(+((f*y-v*A*C)*2.0/D-t))>1.0e-05;w=z;x=e+(E?3.141592653589795-B:B);break a}k[o>>2]=501;r=501;return r|0}break}default:{}}k[o>>2]=504;r=504;return r|0}case 65:{switch(i[l+2>>0]|0){case 82:{if((i[l+3>>0]|0)==67){if(j>=9.869604401089369){k[o>>2]=501;r=501;return r|0}B=+X(+j);z=+Z(+B);if(B!=0.0)F=+_(+B)/B;else F=1.0;B=v*z+t*f*F;if(B>1.0|B<-1.0){k[o>>2]=501;r=501;return r|0}D=z-v*B;z=f*(s*F);if(!(z==0.0&D==0.0)){w=+ba(+B);x=e+ +da(+z,+D);break a}k[o>>2]=501;r=501;return r|0}break}case 73:{if((i[l+3>>0]|0)==84){D=b*h+a*g;z=D==0.0?.01745329252:D*.01745329252;D=c+z;B=+X(+((f+1.0)*.5));C=+_(+D)/+X(+((+Z(+D)+1.0)*.5))-v/B;D=z/(C==0.0?1.0:C);C=b*g-a*h;z=C==0.0?.01745329252:C*.01745329252;C=z*.5;A=f*2.0*+_(+C);y=z*+X(+((f*+Z(+C)+1.0)*.5))/(A==0.0?1.0:A);if(!(t!=0.0|s!=0.0)){w=c;x=e;break a}A=t+v*D/B;B=A/D;C=4.0-u/(y*(y*4.0))-B*B;if(C>4.0|C<2.0){k[o>>2]=501;r=501;return r|0}B=+X(+C)*.5;C=A*B/D;if(+W(+C)>1.0){k[o>>2]=501;r=501;return r|0}D=+ba(+C);C=+Z(+D);if(+W(+C)<1.0e-05){k[o>>2]=501;r=501;return r|0}A=s*B/(y*2.0*C);if(!(+W(+A)>1.0)){w=D;x=e+ +ba(+A)*2.0;break a}k[o>>2]=501;r=501;return r|0}break}default:{}}k[o>>2]=504;r=504;return r|0}case 78:{if((i[l+2>>0]|0)==67?(i[l+3>>0]|0)==80:0){A=f-t*v;if(A==0.0){k[o>>2]=501;r=501;return r|0}D=e+ +da(+s,+A);C=+Z(+(D-e));if(C==0.0){k[o>>2]=501;r=501;return r|0}y=A/C;if(y>1.0|y<-1.0){k[o>>2]=501;r=501;return r|0}else{C=+aa(+y);if(!(c<0.0)){w=C;x=D;break a}w=-C;x=D;break a}}k[o>>2]=504;r=504;return r|0}case 71:{if((i[l+2>>0]|0)==76?(i[l+3>>0]|0)==83:0){D=c+t;if(+W(+D)>1.5707963267948974){k[o>>2]=501;r=501;return r|0}C=+Z(+D);if(+W(+s)>C*6.28318530717959*.5){k[o>>2]=501;r=501;return r|0}else{if(!(C>1.0e-05)){w=D;x=e;break a}w=D;x=e+s/C;break a}}k[o>>2]=504;r=504;return r|0}case 77:{if((i[l+2>>0]|0)==69?(i[l+3>>0]|0)==82:0){C=b*h+a*g;D=C==0.0?1.0:C;C=(d*.5+45.0)*.01745329252;y=+fa(+(+$(+C)));A=D*.01745329252/(+fa(+(+$(+(C+D*.5*.01745329252))))-y);D=e+s/(!(f<=0.0)?f:1.0);if(!(+W(+(D-e))>6.28318530717959)){w=+ca(+(+ea(+(A!=0.0?(t+y*A)/A:0.0))))*2.0+-1.5707963267948974;x=D;break a}k[o>>2]=501;r=501;return r|0}k[o>>2]=504;r=504;return r|0}default:{k[o>>2]=504;r=504;return r|0}}while(0);t=x-e>3.141592653589795?x+-6.28318530717959:x;x=t-e<-3.141592653589795?t+6.28318530717959:t;p[m>>3]=(x<0.0?x+6.28318530717959:x)/.01745329252;p[n>>3]=w/.01745329252;r=q;return r|0}function Rs(a,b,c,d,e,f,g,h,j,l,m,n,o){a=+a;b=+b;c=+c;d=+d;e=+e;f=+f;g=+g;h=+h;j=+j;l=l|0;m=m|0;n=n|0;o=o|0;var q=0,r=0.0,s=0.0,t=0.0,u=0.0,v=0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0.0;if((i[l>>0]|0)!=45){k[o>>2]=504;q=504;return q|0}r=a-c;s=r>180.0?a+-360.0:a;a=r<-180.0?s+360.0:s;s=a-c;r=b-d;t=j*.01745329252;j=+Z(+t);u=+_(+t);if(g==0.0|h==0.0){p[m>>3]=0.0;p[n>>3]=0.0;k[o>>2]=502;q=502;return q|0}p[m>>3]=(s*j+r*u)/g+e;p[n>>3]=(r*j-s*u)/h+f;v=i[l+1>>0]|0;if(v<<24>>24==67){if((i[l+2>>0]|0)==65?(i[l+3>>0]|0)==82:0){q=k[o>>2]|0;return q|0}k[o>>2]=504;q=504;return q|0}s=c*.01745329252;c=d*.01745329252;r=a*.01745329252;a=b*.01745329252;b=+Z(+a);t=+_(+a);w=+Z(+c);x=+_(+c);y=r-s;z=b*+_(+y);A=t*x;B=+Z(+y);C=b*w*B;D=A+C;a:do switch(v<<24>>24){case 84:{if((i[l+2>>0]|0)==65?(i[l+3>>0]|0)==78:0){if(D<=0.0){k[o>>2]=501;q=501;return q|0}if(w<.001){E=b*B/A;F=(w*(E*E+1.0)-E)/x}else F=(t/D-x)/w;E=+_(+s);if(+W(+E)<.3){G=(x*(F*E)+(b*+_(+r)/D-w*E))/+Z(+s);H=F;break a}else{I=+Z(+s);G=(b*+Z(+r)/D-w*I+x*(F*I))/-E;H=F;break a}}k[o>>2]=504;q=504;return q|0}case 83:{switch(i[l+2>>0]|0){case 73:{if((i[l+3>>0]|0)==78){if(!(D<0.0)){G=z;H=t*w-b*x*B;break a}k[o>>2]=501;q=501;return q|0}break}case 84:{if((i[l+3>>0]|0)==71){if(+W(+a)>1.5707963267948974){k[o>>2]=501;q=501;return q|0}E=A+1.0+C;if(!(+W(+E)<1.0e-05)){I=2.0/E;G=z*I;H=(t*w-b*x*B)*I;break a}k[o>>2]=501;q=501;return q|0}break}default:{}}k[o>>2]=504;q=504;return q|0}case 65:{switch(i[l+2>>0]|0){case 82:{if((i[l+3>>0]|0)==67){I=D<-1.0?-1.0:D;E=+aa(+(I>1.0?1.0:I));if(E!=0.0)J=E/+_(+E);else J=1.0;G=z*J;H=(t*w-b*x*B)*J;break a}break}case 73:{if((i[l+3>>0]|0)==84){E=y*.5;if(+W(+E)>1.5707963267948974){k[o>>2]=501;q=501;return q|0}I=j*h+u*g;K=I==0.0?.01745329252:I*.01745329252;I=c+K;L=+X(+((w+1.0)*.5));M=+_(+I)/+X(+((+Z(+I)+1.0)*.5))-x/L;I=K/(M==0.0?1.0:M);M=j*g-u*h;K=M==0.0?.01745329252:M*.01745329252;M=K*.5;N=w*2.0*+_(+M);O=+X(+((b*+Z(+E)+1.0)*.5));if(!(+W(+O)<1.0e-05)){G=+_(+E)*(b*(K*+X(+((w*+Z(+M)+1.0)*.5))/(N==0.0?1.0:N)*2.0))/O;H=t*I/O-x*I/L;break a}k[o>>2]=503;q=503;return q|0}break}default:{}}k[o>>2]=504;q=504;return q|0}case 78:{if((i[l+2>>0]|0)==67?(i[l+3>>0]|0)==80:0){if(!(c==0.0)){G=z;H=(w-b*B)/x;break a}k[o>>2]=501;q=501;return q|0}k[o>>2]=504;q=504;return q|0}case 71:{if((i[l+2>>0]|0)==76?(i[l+3>>0]|0)==83:0){if(+W(+a)>1.5707963267948974){k[o>>2]=501;q=501;return q|0}if(!(+W(+c)>1.5707963267948974)){G=y*b;H=a-c;break a}k[o>>2]=501;q=501;return q|0}k[o>>2]=504;q=504;return q|0}case 77:{if((i[l+2>>0]|0)==69?(i[l+3>>0]|0)==82:0){L=j*h+u*g;I=L==0.0?1.0:L;L=(d*.5+45.0)*.01745329252;O=+fa(+(+$(+L)));N=I*.01745329252/(+fa(+(+$(+(L+I*.5*.01745329252))))-O);I=+$(+(a*.5+.7853981633974487));if(!(I<1.0e-05)){G=y*(!(w<=0.0)?w:1.0);H=N*+fa(+I)-O*N;break a}k[o>>2]=502;q=502;return q|0}k[o>>2]=504;q=504;return q|0}default:{k[o>>2]=504;q=504;return q|0}}while(0);w=G/.01745329252;G=H/.01745329252;p[m>>3]=(j*w+u*G)/g+e;p[n>>3]=(j*G-u*w)/h+f;q=k[o>>2]|0;return q|0}function Ss(){var a=0,b=0,c=0,d=0.0,e=0.0,f=0.0,g=0.0;if(k[30368]|0){a=0;return a|0}b=ey(1e4,4)|0;k[30368]=b;if(!b){a=113;return a|0}else{c=0;d=1.0}while(1){e=d*16807.0;f=e-+(~~(e/2147483647.0)|0)*2147483647.0;o[b+(c<<2)>>2]=f/2147483647.0;c=c+1|0;if((c|0)==1e4){g=f;break}else d=f}if((~~g|0)==1043618065){a=0;return a|0}tk(778042);a=1;return a|0}function Ts(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;switch(b|0){case 0:case 11:case 21:case 22:case 31:case 41:case 51:case -1:{k[(k[a+4>>2]|0)+984>>2]=b;d=k[c>>2]|0;return d|0}default:{tk(778103);k[c>>2]=413;d=413;return d|0}}return 0}function Us(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0;if(b>>>0<=6){if((b|0)>0){e=k[a+4>>2]|0;a=0;do{k[e+988+(a<<2)>>2]=k[c+(a<<2)>>2];a=a+1|0}while((a|0)!=(b|0))}}else{k[d>>2]=320;tk(778161)}return k[d>>2]|0}function Vs(a,b,c){a=a|0;b=+b;c=c|0;o[(k[a+4>>2]|0)+1012>>2]=b==0.0?9999.0:b;return k[c>>2]|0}function Ws(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;if((b+1|0)>>>0>3){tk(778215);k[c>>2]=413;d=413;return d|0}else{k[(k[a+4>>2]|0)+1016>>2]=(b|0)==0?1:b;d=k[c>>2]|0;return d|0}return 0}function Xs(a,b,c){a=a|0;b=+b;c=c|0;o[(k[a+4>>2]|0)+1032>>2]=b;return k[c>>2]|0}function Ys(a,b,c){a=a|0;b=b|0;c=c|0;k[(k[a+4>>2]|0)+1036>>2]=b;return k[c>>2]|0}function Zs(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0.0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0;g=r;r=r+256|0;h=g+24|0;j=g+16|0;l=g+177|0;m=g;n=g+84|0;p=g+72|0;q=g+68|0;s=g+100|0;t=g+96|0;u=g+56|0;v=g+104|0;w=g+32|0;k[n>>2]=k[30369];k[n+4>>2]=k[30370];k[n+8>>2]=k[30371];k[u>>2]=k[30372];k[u+4>>2]=k[30373];k[u+8>>2]=k[30374];x=k[f>>2]|0;if((x|0)>0){y=x;r=g;return y|0}x=a+4|0;z=k[x>>2]|0;A=z+1012|0;B=+o[A>>2];C=z+984|0;D=k[C>>2]|0;if(B==9999.0)if((D+-21|0)>>>0<2)E=D;else{tk(778266);k[f>>2]=413;y=413;r=g;return y|0}else if(!D){k[C>>2]=11;E=11}else E=D;do if((b|0)<0)if(B!=9999.0){D=z+1016|0;F=k[D>>2]|0;if(!F){k[D>>2]=1;G=1}else G=F;if(B==0.0)if((G|0)==-1){o[A>>2]=16.0;H=16;break}else{o[A>>2]=4.0;H=16;break}else H=16}else H=16;else if((b|0)==20){I=1;J=16}else H=16;while(0);a:do if((H|0)==16)switch(b|0){case 40:{I=0;J=32;break a;break}case 10:{I=0;J=8;break a;break}default:{I=0;J=b;break a}}while(0);G=z+988|0;k[w>>2]=k[G>>2];k[w+4>>2]=k[G+4>>2];k[w+8>>2]=k[G+8>>2];k[w+12>>2]=k[G+12>>2];k[w+16>>2]=k[G+16>>2];k[w+20>>2]=k[G+20>>2];do if((E|0)==41){if((c|0)<2){tk(778346);k[f>>2]=413;y=413;r=g;return y|0}G=k[d>>2]|0;if((G|0)>=4?(F=k[d+4>>2]|0,(F|0)>=4):0){D=k[w>>2]|0;do if((D|0)<1){K=w+4|0;L=k[K>>2]|0;if((L|0)==-1){k[w>>2]=G;k[K>>2]=F;if((c|0)>2)M=2;else{N=G;break}do{k[w+(M<<2)>>2]=1;M=M+1|0}while((M|0)!=(c|0));N=k[w>>2]|0;break}if(L>>>0<2){k[w>>2]=G;if((F|0)<31){k[K>>2]=F;N=G;break}O=(F|0)%16|0;if((O|0)==0|(O|0)>3){k[K>>2]=16;N=G;break}O=(F|0)%24|0;if((O|0)==0|(O|0)>3){k[K>>2]=24;N=G;break}O=(F|0)%20|0;if((O|0)==0|(O|0)>3){k[K>>2]=20;N=G;break}O=(F|0)%30|0;if((O|0)==0|(O|0)>3){k[K>>2]=30;N=G;break}O=(F|0)%28|0;if((O|0)==0|(O|0)>3){k[K>>2]=28;N=G;break}O=(F|0)%26|0;if((O|0)==0|(O|0)>3){k[K>>2]=26;N=G;break}O=(F|0)%22|0;if((O|0)==0|(O|0)>3){k[K>>2]=22;N=G;break}O=(F|0)%18|0;if((O|0)==0|(O|0)>3){k[K>>2]=18;N=G;break}O=(F|0)%14|0;if((O|0)==0|(O|0)>3){k[K>>2]=14;N=G;break}else{k[K>>2]=17;N=G;break}}else H=54}else if((D|0)>=4?(k[w+4>>2]|0)>=4:0)N=D;else H=54;while(0);if((H|0)==54){tk(778483);k[f>>2]=413;y=413;r=g;return y|0}if((((G|0)%(N|0)|0)+-1|0)>>>0<3?(D=N+1|0,k[w>>2]=D,(((G|0)%(D|0)|0)+-1|0)>>>0<3):0){tk(778548);k[f>>2]=413;y=413;r=g;return y|0}D=w+4|0;O=k[D>>2]|0;if((((F|0)%(O|0)|0)+-1|0)>>>0>=3)break;P=O+1|0;k[D>>2]=P;if((((F|0)%(P|0)|0)+-1|0)>>>0>=3)break;tk(778621);k[f>>2]=413;y=413;r=g;return y|0}tk(778417);k[f>>2]=413;y=413;r=g;return y|0}while(0);N=(c|0)>0;if(N){M=0;do{E=w+(M<<2)|0;P=k[E>>2]|0;do if(!M){if((P|0)<1)k[E>>2]=k[d>>2]}else{if((P|0)<0){k[E>>2]=k[d+(M<<2)>>2];break}if(!P)k[E>>2]=1}while(0);M=M+1|0}while((M|0)!=(c|0))}M=z+1028|0;k[q>>2]=(k[M>>2]|0)==0?4345905:4346161;i[s>>0]=i[778694]|0;i[s+1>>0]=i[778695]|0;i[s+2>>0]=i[778696]|0;i[t>>0]=i[778694]|0;i[t+1>>0]=i[778695]|0;i[t+2>>0]=i[778696]|0;k[p>>2]=q;k[p+4>>2]=s;k[p+8>>2]=t;if(N){t=0;s=1;while(1){z=ha((((k[d+(t<<2)>>2]|0)+-1|0)/(k[w+(t<<2)>>2]|0)|0)+1|0,s)|0;t=t+1|0;if((t|0)==(c|0)){Q=z;break}else s=z}R=Q;S=((Q|0)<0)<<31>>31}else{R=1;S=0}Q=(J|0)<0;if(Q?+o[A>>2]!=9999.0:0)T=3;else T=1;b:do switch(k[C>>2]|0){case 11:{i[m>>0]=i[778697]|0;i[m+1>>0]=i[778698]|0;i[m+2>>0]=i[778699]|0;i[m+3>>0]=i[778700]|0;i[m+4>>0]=i[778701]|0;i[m+5>>0]=i[778702]|0;i[m+6>>0]=i[778703]|0;break}case 21:{i[m>>0]=i[778704]|0;i[m+1>>0]=i[778705]|0;i[m+2>>0]=i[778706]|0;i[m+3>>0]=i[778707]|0;i[m+4>>0]=i[778708]|0;i[m+5>>0]=i[778709]|0;i[m+6>>0]=i[778710]|0;break}case 22:{i[m>>0]=i[778711]|0;i[m+1>>0]=i[778712]|0;i[m+2>>0]=i[778713]|0;i[m+3>>0]=i[778714]|0;i[m+4>>0]=i[778715]|0;i[m+5>>0]=i[778716]|0;i[m+6>>0]=i[778717]|0;break}case 51:{A=m;k[A>>2]=1346984514;k[A+4>>2]=3235634;break}case 31:{i[m>>0]=i[778718]|0;i[m+1>>0]=i[778719]|0;i[m+2>>0]=i[778720]|0;i[m+3>>0]=i[778721]|0;i[m+4>>0]=i[778722]|0;i[m+5>>0]=i[778723]|0;i[m+6>>0]=i[778724]|0;if(!(k[M>>2]|0)){k[q>>2]=4804657;break b}else{k[q>>2]=4804913;break b}break}case 41:{U=m;V=778725;W=U+12|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));break}case -1:{U=m;V=778737;W=U+11|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));break}default:{tk(778748);k[f>>2]=413;y=413;r=g;return y|0}}while(0);$r(a,2,R,S,T,n,p,u,0,f)|0;is(a,778793,1,778800,f)|0;if(e){is(a,778836,1,778844,f)|0;gs(a,778879,J,((J|0)<0)<<31>>31,778887,f)|0;gs(a,778915,c,((c|0)<0)<<31>>31,778922,f)|0;if(N){e=0;do{u=e;e=e+1|0;k[j>>2]=e;Xw(l,778950,j)|0;p=k[d+(u<<2)>>2]|0;gs(a,l,p,((p|0)<0)<<31>>31,778959,f)|0}while((e|0)!=(c|0));H=89}}else H=89;if((H|0)==89?N:0){N=0;do{H=N;N=N+1|0;k[h>>2]=N;Xw(l,778989,h)|0;e=k[w+(H<<2)>>2]|0;gs(a,l,e,((e|0)<0)<<31>>31,778997,f)|0}while((N|0)!=(c|0))}c:do if(Q){c=k[x>>2]|0;if(+o[c+1012>>2]==9999.0){fs(a,779028,779037,779042,f)|0;break}N=c+1016|0;switch(k[N>>2]|0){case 0:{k[N>>2]=1;break}case 1:break;case 2:{fs(a,779028,779183,779105,f)|0;es(a,31,779134,(k[x>>2]|0)+1020|0,779143,f)|0;if(tx(m,778697)|0)break c;U=m;V=779204;W=U+9|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));break c;break}case -1:{fs(a,779028,779213,779223,f)|0;break c;break}default:break c}fs(a,779028,779084,779105,f)|0;es(a,31,779134,(k[x>>2]|0)+1020|0,779143,f)|0}while(0);fs(a,779256,m,779265,f)|0;d:do switch(k[(k[x>>2]|0)+984>>2]|0){case 11:{fs(a,779287,779294,779304,f)|0;gs(a,779327,32,0,779333,f)|0;fs(a,779350,779357,779365,f)|0;switch(J|0){case 8:{gs(a,779397,1,0,779365,f)|0;break d;break}case 16:{gs(a,779397,2,0,779365,f)|0;break d;break}default:{gs(a,779397,4,0,779365,f)|0;break d}}break}case 41:{fs(a,779287,779403,779409,f)|0;js(a,779327,+o[(k[x>>2]|0)+1032>>2],7,779409,f)|0;fs(a,779350,779432,779439,f)|0;m=k[(k[x>>2]|0)+1036>>2]|0;gs(a,779397,m,((m|0)<0)<<31>>31,779439,f)|0;break}default:{}}while(0);e:do if(I){U=v;V=779463;W=U+44|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779507,32768.0,0,v,f)|0;U=v;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779536,1.0,0,v,f)|0}else switch(b|0){case 10:{U=v;V=779543;W=U+41|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779507,-128.0,0,v,f)|0;U=v;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779536,1.0,0,v,f)|0;break e;break}case 40:{U=v;V=779584;W=U+43|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779507,2147483648.0,0,v,f)|0;U=v;V=779513;W=U+23|0;do{i[U>>0]=i[V>>0]|0;U=U+1|0;V=V+1|0}while((U|0)<(W|0));hs(a,779536,1.0,0,v,f)|0;break e;break}default:break e}while(0);y=k[f>>2]|0;r=g;return y|0}function _s(a,b,c,d,e,f,g,h,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;var m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0.0,P=0.0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0,ca=0.0,da=0.0,ea=0.0;m=r;r=r+80|0;n=m+64|0;q=m+60|0;s=m+56|0;t=m+52|0;u=m+48|0;v=m+44|0;w=m+32|0;x=m+24|0;y=m+16|0;z=m+8|0;A=m;B=m+40|0;k[q>>2]=1;k[v>>2]=0;p[z>>3]=1.0;C=A;k[C>>2]=0;k[C+4>>2]=0;C=k[l>>2]|0;if((C|0)>0){D=C;r=m;return D|0}C=a+4|0;E=k[C>>2]|0;if(+o[E+1068>>2]==9999.0?((k[E+1040>>2]|0)+-21|0)>>>0>=2:0){tk(779724);k[l>>2]=413;D=413;r=m;return D|0}F=k[E+1220>>2]|0;if((F|0)!=0?(G=(b+-1|0)%((((k[E+1104>>2]|0)+-1|0)/(k[E+1044>>2]|0)|0)+1|0)|0,(k[F+(G<<2)>>2]|0)==(b|0)):0){F=k[(k[E+1232>>2]|0)+(G<<2)>>2]|0;if(!F)H=E;else{dy(F);H=k[C>>2]|0}F=k[(k[H+1236>>2]|0)+(G<<2)>>2]|0;if(!F)I=H;else{dy(F);I=k[C>>2]|0}k[(k[I+1232>>2]|0)+(G<<2)>>2]=0;k[(k[(k[C>>2]|0)+1236>>2]|0)+(G<<2)>>2]=0;I=k[C>>2]|0;k[(k[I+1220>>2]|0)+(G<<2)>>2]=0;k[(k[I+1224>>2]|0)+(G<<2)>>2]=0;k[(k[I+1228>>2]|0)+(G<<2)>>2]=0;k[(k[I+1240>>2]|0)+(G<<2)>>2]=0;J=I}else J=E;if((k[J+1040>>2]|0)==-1){a:do if((k[J+1140>>2]|0)<1){switch(c|0){case 21:{k[n>>2]=4804657;break}case 31:{k[n>>2]=4870193;break}case 42:{k[n>>2]=4542769;break}default:{tk(779807);k[l>>2]=413;K=413;break a}}Qj(a,999,779880,n,l)|0;M=k[C>>2]|0;N=20}else{M=J;N=20}while(0);if((N|0)==20){Tk(a,0,779880,M+1140|0,l)|0;Iq(a,c,k[(k[C>>2]|0)+1140>>2]|0,b,((b|0)<0)<<31>>31,1,0,e,((e|0)<0)<<31>>31,d,l)|0;K=k[l>>2]|0}D=K;r=m;return D|0}K=k[J+1096>>2]|0;M=k[J+1200>>2]|0;N=((K|0)>0?(k[J+1156>>2]|0)!=-1:0)?0:h;O=+p[J+1176>>3];P=+p[J+1184>>3];b:do switch(c|0){case 21:{$s(a,d,e,N,j,M,K,O,P,+p[J+1192>>3],s,l)|0;break}case 20:{at(a,d,e,N,j,M,K,O,P,s,l)|0;break}case 11:{bt(a,d,e,N,j,M,K,O,P,s,l)|0;break}case 12:{ct(a,d,e,N,j,M,K,O,P,s,l)|0;break}case 31:{if((K|0)!=32|O!=1.0|P!=0.0){tk(779898);k[l>>2]=413;D=413;r=m;return D|0}k[s>>2]=4;if((N|0)==1?(h=k[j>>2]|0,(e|0)>0&(h|0)!=(M|0)):0){n=e;do{E=n;n=n+-1|0;I=d+(n<<2)|0;if((k[I>>2]|0)==(h|0))k[I>>2]=M}while((E|0)>1)}break}case 30:{if((K|0)!=32|O!=1.0|P!=2147483648.0){tk(779898);k[l>>2]=413;D=413;r=m;return D|0}k[s>>2]=4;if((N|0)==1){h=k[j>>2]|0;if((e|0)>0)Q=e;else break b;do{n=Q;Q=Q+-1|0;E=d+(Q<<2)|0;I=k[E>>2]|0;k[E>>2]=(I|0)==(h|0)?M:I^-2147483648}while((n|0)>1)}else{if((e|0)>0)R=e;else break b;do{h=R;R=R+-1|0;n=d+(R<<2)|0;k[n>>2]=k[n>>2]^-2147483648}while((h|0)>1)}break}case 42:{dt(a,b,d,e,f,g,N,j,M,K,O,P,s,q,z,A,l)|0;break}case 82:{et(a,b,d,e,f,g,N,j,M,K,O,P,s,q,z,A,l)|0;break}default:{tk(779978);k[l>>2]=410;D=410;r=m;return D|0}}while(0);R=k[l>>2]|0;if((R|0)>0){D=R;r=m;return D|0}R=k[C>>2]|0;if(!(k[q>>2]|0)){if((k[R+1144>>2]|0)<1){if(!(k[R+1028>>2]|0))Qj(a,999,780328,780353,l)|0;else Qj(a,999,780328,780349,l)|0;if((k[l>>2]|0)<1)Tk(a,0,780328,(k[C>>2]|0)+1144|0,l)|0}if((c|0)==42){q=e<<2;Q=~~(+(q>>>0)*1.1)>>>0;k[t>>2]=Q;J=ey(Q,1)|0;k[u>>2]=J;if(!J){tk(780357);k[l>>2]=113;D=113;r=m;return D|0}if((N|0)==1?(P=+o[j>>2],(e|0)>0):0){J=0;do{if(+o[d+(J<<2)>>2]==P)k[d+(J<<2)>>2]=-1;J=J+1|0}while((J|0)!=(e|0))}Ls(d,e);sh(d,q,u,t,19,v,l)|0;S=u}else{q=e<<3;J=~~(+(q>>>0)*1.1)>>>0;k[t>>2]=J;Q=ey(J,1)|0;k[u>>2]=Q;if(!Q){tk(780357);k[l>>2]=113;D=113;r=m;return D|0}if((N|0)==1?(P=+p[j>>3],(e|0)>0):0){j=0;do{if(+p[d+(j<<3)>>3]==P){Q=d+(j<<3)|0;k[Q>>2]=-1;k[Q+4>>2]=-1}j=j+1|0}while((j|0)!=(e|0))}Ms(d,e);sh(d,q,u,t,19,v,l)|0;S=u}Rq(a,k[(k[C>>2]|0)+1144>>2]|0,b,((b|0)<0)<<31>>31,1,0,k[v>>2]|0,0,k[S>>2]|0,l)|0;dy(k[S>>2]|0)}else{S=k[R+1132>>2]|0;k[t>>2]=S;q=ey(S,1)|0;k[u>>2]=q;if(!q){tk(780028);k[l>>2]=113;D=113;r=m;return D|0}c:do switch(k[R+1040>>2]|0){case 11:{switch(k[s>>2]|0){case 2:{T=Dt(d,e,q,S,k[R+1204>>2]|0)|0;break}case 1:{T=Et(d,e,q,S,k[R+1204>>2]|0)|0;break}default:T=Ct(d,e,q,S,k[R+1204>>2]|0)|0}if((T|0)>=0){Rq(a,k[(k[C>>2]|0)+1136>>2]|0,b,((b|0)<0)<<31>>31,1,0,T,((T|0)<0)<<31>>31,k[u>>2]|0,l)|0;break c}dy(k[u>>2]|0);tk(780078);k[l>>2]=413;D=413;r=m;return D|0}case 31:{d:do if((e|0)>0){j=0;while(1){if((k[d+(j<<2)>>2]|0)>>>0>16777215)break;j=j+1|0;if((j|0)>=(e|0))break d}tk(780135);k[l>>2]=413;D=413;r=m;return D|0}while(0);j=It(d,1,q,e)|0;if((j|0)>=0){er(a,k[(k[C>>2]|0)+1136>>2]|0,b,((b|0)<0)<<31>>31,1,0,j,((j|0)<0)<<31>>31,k[u>>2]|0,l)|0;break c}dy(k[u>>2]|0);tk(780186);k[l>>2]=413;D=413;r=m;return D|0}case 22:case 21:{j=+o[R+1068>>2]==9999.0;e:do if((c|0)==42&j){Ls(d,e);do if((k[(k[C>>2]|0)+1040>>2]|0)==22){Q=((e|0)<0)<<31>>31;J=e<<2;h=cy(J)|0;if(!h){tk(780243);U=J;break}if((e|0)>0){n=e<<1;I=e*3|0;E=0;G=0;F=h;H=d;while(1){i[F>>0]=i[H>>0]|0;i[F+e>>0]=i[H+1>>0]|0;i[F+n>>0]=i[H+2>>0]|0;i[F+I>>0]=i[H+3>>0]|0;E=ky(E|0,G|0,1,0)|0;G=L;if(!((G|0)<(Q|0)|(G|0)==(Q|0)&E>>>0<e>>>0))break;else{F=F+1|0;H=H+4|0}}}py(d|0,h|0,J|0)|0;dy(h);U=J}else U=e<<2;while(0);sh(d,U,u,t,19,v,l)|0}else{if((c|0)==82&j){Ms(d,e);if((k[(k[C>>2]|0)+1040>>2]|0)==22){H=((e|0)<0)<<31>>31;F=e<<3;E=ey(1,F)|0;if((e|0)>0){Q=0;G=0;I=d;while(1){i[E+Q>>0]=i[I>>0]|0;n=Q+e|0;i[E+n>>0]=i[I+1>>0]|0;V=n+e|0;i[E+V>>0]=i[I+2>>0]|0;n=V+e|0;i[E+n>>0]=i[I+3>>0]|0;V=n+e|0;i[E+V>>0]=i[I+4>>0]|0;n=V+e|0;i[E+n>>0]=i[I+5>>0]|0;V=n+e|0;i[E+V>>0]=i[I+6>>0]|0;i[E+(V+e)>>0]=i[I+7>>0]|0;Q=ky(Q|0,G|0,1,0)|0;G=L;if(!((G|0)<(H|0)|(G|0)==(H|0)&Q>>>0<e>>>0))break;else I=I+8|0}}py(d|0,E|0,F|0)|0;dy(E);W=F}else W=e<<3;sh(d,W,u,t,19,v,l)|0;break}switch(k[s>>2]|0){case 2:{Ks(d,e);if((k[(k[C>>2]|0)+1040>>2]|0)==22){I=((e|0)<0)<<31>>31;Q=e<<1;H=cy(Q)|0;if((e|0)>0){G=0;V=0;n=H;X=d;while(1){i[n>>0]=i[X>>0]|0;i[n+e>>0]=i[X+1>>0]|0;G=ky(G|0,V|0,1,0)|0;V=L;if(!((V|0)<(I|0)|(V|0)==(I|0)&G>>>0<e>>>0))break;else{n=n+1|0;X=X+2|0}}}py(d|0,H|0,Q|0)|0;dy(H);Y=Q}else Y=e<<1;sh(d,Y,u,t,19,v,l)|0;break e;break}case 4:{Ls(d,e);break}case 1:{sh(d,e,u,t,19,v,l)|0;break e;break}default:{}}do if((k[(k[C>>2]|0)+1040>>2]|0)==22){X=((e|0)<0)<<31>>31;n=e<<2;G=cy(n)|0;if(!G){tk(780243);Z=n;break}if((e|0)>0){I=e<<1;V=e*3|0;F=0;E=0;_=G;$=d;while(1){i[_>>0]=i[$>>0]|0;i[_+e>>0]=i[$+1>>0]|0;i[_+I>>0]=i[$+2>>0]|0;i[_+V>>0]=i[$+3>>0]|0;F=ky(F|0,E|0,1,0)|0;E=L;if(!((E|0)<(X|0)|(E|0)==(X|0)&F>>>0<e>>>0))break;else{_=_+1|0;$=$+4|0}}}py(d|0,G|0,n|0)|0;dy(G);Z=n}else Z=e<<2;while(0);sh(d,Z,u,t,19,v,l)|0}while(0);Rq(a,k[(k[C>>2]|0)+1136>>2]|0,b,((b|0)<0)<<31>>31,1,0,k[v>>2]|0,0,k[u>>2]|0,l)|0;break}case 51:{switch(k[s>>2]|0){case 2:{Ks(d,e);break}case 4:{Ls(d,e);break}default:{}}tk(780258);k[l>>2]=413;D=413;r=m;return D|0}case 41:{P=+o[R+1212>>2];O=P;if(!(P>0.0))if(P<0.0){aa=S;ba=-P}else{aa=S;ba=P}else{wt(d,f,g,N,M,0,0,0,0,0,0,w,x,y,l)|0;P=+p[w>>3];ca=+p[x>>3];if(P!=0.0&P<ca){p[x>>3]=P;da=P}else da=ca;ca=+p[y>>3];if(ca!=0.0&ca<da){p[x>>3]=ca;ea=ca}else ea=da;aa=k[t>>2]|0;ba=O*ea}j=~~(ba+.5);k[B>>2]=aa;switch(K|0){case 8:case 16:{Kt(d,f,g,j,k[u>>2]|0,B,l)|0;break}default:{do if((k[l>>2]|0)<=0){Q=(e|0)<1e4?e:1e4;H=cy(Q<<3)|0;if(!H){tk(780282);k[l>>2]=113;break}f:do if((Q|0)>0){$=e-Q|0;_=Q;while(1){if((_|0)>0){F=0;do{X=k[d+(F+$<<2)>>2]|0;E=H+(F<<3)|0;k[E>>2]=X;k[E+4>>2]=((X|0)<0)<<31>>31;F=F+1|0}while((F|0)!=(_|0))}py(d+($<<3)|0,H|0,_<<3|0)|0;if(!$)break f;F=($|0)>1e4;_=F?_:$;if((_|0)<=0)break;else $=F?$+-1e4|0:0}}while(0);dy(H)}while(0);Lt(d,f,g,j,k[u>>2]|0,B,l)|0}}Q=k[B>>2]|0;Rq(a,k[(k[C>>2]|0)+1136>>2]|0,b,((b|0)<0)<<31>>31,1,0,Q,((Q|0)<0)<<31>>31,k[u>>2]|0,l)|0;break}default:{}}while(0);B=k[(k[C>>2]|0)+1148>>2]|0;if((B|0)>0){g=((b|0)<0)<<31>>31;Wq(a,B,b,g,1,0,1,0,z,l)|0;Wq(a,k[(k[C>>2]|0)+1152>>2]|0,b,g,1,0,1,0,A,l)|0}dy(k[u>>2]|0)}D=k[l>>2]|0;r=m;return D|0}function $s(a,b,c,d,e,f,g,h,i,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=+i;l=+l;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;if((g|0)!=16|h!=1.0|i!=0.0){tk(780405);k[n>>2]=413;o=413;return o|0}g=k[(k[a+4>>2]|0)+1040>>2]|0;a:do switch(g|0){case 51:case 22:case 21:case 11:{k[m>>2]=2;if((d|0)==1?(a=j[e>>1]|0,(a<<16>>16|0)!=(f|0)&(c|0)>0):0){p=f&65535;q=c;do{r=q;q=q+-1|0;s=b+(q<<1)|0;if((j[s>>1]|0)==a<<16>>16)j[s>>1]=p}while((r|0)>1)}break}default:{k[m>>2]=4;if((g|0)==41){if((d|0)==1){p=j[e>>1]|0;if((c|0)>0)t=c;else break a;while(1){a=t;t=t+-1|0;q=j[b+(t<<1)>>1]|0;k[b+(t<<2)>>2]=q<<16>>16==p<<16>>16?f:q<<16>>16;if((a|0)<=1)break a}}if((k[n>>2]|0)>0)break a;p=(c|0)<1e4?c:1e4;a=cy(p<<2)|0;if(!a){tk(780484);k[n>>2]=113;break a}b:do if((p|0)>0){q=c-p|0;r=p;while(1){if((r|0)>0){s=0;do{k[a+(s<<2)>>2]=j[b+(s+q<<1)>>1];s=s+1|0}while((s|0)!=(r|0))}py(b+(q<<2)|0,a|0,r<<2|0)|0;if(!q)break b;s=(q|0)>1e4;r=s?r:q;if((r|0)<=0)break;else q=s?q+-1e4|0:0}}while(0);dy(a);break a}p=(d|0)==1;if(l==32768.0){if(p){q=j[e>>1]|0;if((c|0)>0)u=c;else break a;while(1){r=u;u=u+-1|0;s=j[b+(u<<1)>>1]|0;k[b+(u<<2)>>2]=s<<16>>16==q<<16>>16?f:(s<<16>>16)+32768|0;if((r|0)<=1)break a}}if((k[n>>2]|0)>0)break a;q=(c|0)<1e4?c:1e4;a=cy(q<<2)|0;if(!a){tk(780484);k[n>>2]=113;break a}c:do if((q|0)>0){r=c-q|0;s=q;while(1){if((s|0)>0){v=0;do{k[a+(v<<2)>>2]=j[b+(v+r<<1)>>1];v=v+1|0}while((v|0)!=(s|0))}py(b+(r<<2)|0,a|0,s<<2|0)|0;if(!r)break c;v=(r|0)>1e4;s=v?s:r;if((s|0)<=0)break;else r=v?r+-1e4|0:0}}while(0);dy(a);break a}else{if(p){q=j[e>>1]|0;if((c|0)>0)w=c;else break a;while(1){r=w;w=w+-1|0;s=j[b+(w<<1)>>1]|0;k[b+(w<<2)>>2]=s<<16>>16==q<<16>>16?f:s<<16>>16;if((r|0)<=1)break a}}if((k[n>>2]|0)>0)break a;q=(c|0)<1e4?c:1e4;p=cy(q<<2)|0;if(!p){tk(780484);k[n>>2]=113;break a}d:do if((q|0)>0){a=c-q|0;r=q;while(1){if((r|0)>0){s=0;do{k[p+(s<<2)>>2]=j[b+(s+a<<1)>>1];s=s+1|0}while((s|0)!=(r|0))}py(b+(a<<2)|0,p|0,r<<2|0)|0;if(!a)break d;s=(a|0)>1e4;r=s?r:a;if((r|0)<=0)break;else a=s?a+-1e4|0:0}}while(0);dy(p);break a}}}while(0);o=k[n>>2]|0;return o|0}function at(a,b,c,d,e,f,g,h,i,l,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=+i;l=l|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0;if((g|0)!=16|h!=1.0|i!=32768.0){tk(779898);k[n>>2]=413;o=413;return o|0}a:do switch(k[(k[a+4>>2]|0)+1040>>2]|0){case 51:case 22:case 21:case 11:{k[l>>2]=2;if((d|0)==1){g=j[e>>1]|0;if((c|0)>0)p=c;else break a;do{q=p;p=p+-1|0;r=b+(p<<1)|0;s=j[r>>1]|0;j[r>>1]=s<<16>>16==g<<16>>16?f:s&65535^32768}while((q|0)>1)}else{if((c|0)>0)t=c;else break a;do{g=t;t=t+-1|0;q=b+(t<<1)|0;j[q>>1]=(m[q>>1]|0)^32768}while((g|0)>1)}break}default:{k[l>>2]=4;if((d|0)==1){g=j[e>>1]|0;if((c|0)>0)u=c;else break a;while(1){q=u;u=u+-1|0;s=j[b+(u<<1)>>1]|0;k[b+(u<<2)>>2]=s<<16>>16==g<<16>>16?f:(s&65535)+-32768|0;if((q|0)<=1)break a}}if((k[n>>2]|0)<=0){g=(c|0)<1e4?c:1e4;q=cy(g<<2)|0;if(!q){tk(780527);k[n>>2]=113;break a}b:do if((g|0)>0){s=c-g|0;r=g;while(1){if((r|0)>0){v=0;do{k[q+(v<<2)>>2]=m[b+(v+s<<1)>>1];v=v+1|0}while((v|0)!=(r|0))}py(b+(s<<2)|0,q|0,r<<2|0)|0;if(!s)break b;v=(s|0)>1e4;r=v?r:s;if((r|0)<=0)break;else s=v?s+-1e4|0:0}}while(0);dy(q)}}}while(0);o=k[n>>2]|0;return o|0}function bt(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;j=+j;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0;if((g|0)!=8|h!=1.0|j!=0.0){tk(779898);k[n>>2]=413;o=413;return o|0}a:do switch(k[(k[a+4>>2]|0)+1040>>2]|0){case 51:case 22:case 21:case 11:{k[m>>2]=1;if((d|0)==1?(g=i[e>>0]|0,(g&255|0)!=(f|0)&(c|0)>0):0){p=f&255;q=c;do{r=q;q=q+-1|0;s=b+q|0;if((i[s>>0]|0)==g<<24>>24)i[s>>0]=p}while((r|0)>1)}break}default:{k[m>>2]=4;if((d|0)==1){p=i[e>>0]|0;if((c|0)>0)t=c;else break a;while(1){g=t;t=t+-1|0;q=i[b+t>>0]|0;k[b+(t<<2)>>2]=q<<24>>24==p<<24>>24?f:q&255;if((g|0)<=1)break a}}if((k[n>>2]|0)<=0){p=(c|0)<1e4?c:1e4;g=cy(p<<2)|0;if(!g){tk(780571);k[n>>2]=113;break a}b:do if((p|0)>0){q=c-p|0;r=p;while(1){if((r|0)>0){s=0;do{k[g+(s<<2)>>2]=l[b+(s+q)>>0];s=s+1|0}while((s|0)!=(r|0))}py(b+(q<<2)|0,g|0,r<<2|0)|0;if(!q)break b;s=(q|0)>1e4;r=s?r:q;if((r|0)<=0)break;else q=s?q+-1e4|0:0}}while(0);dy(g)}}}while(0);o=k[n>>2]|0;return o|0}function ct(a,b,c,d,e,f,g,h,j,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;j=+j;m=m|0;n=n|0;var o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0;if((g|0)!=8|h!=1.0|j!=-128.0){tk(779898);k[n>>2]=413;o=413;return o|0}a:do switch(k[(k[a+4>>2]|0)+1040>>2]|0){case 51:case 22:case 21:case 11:{k[m>>2]=1;if((d|0)==1){g=i[e>>0]|0;if((c|0)>0)p=c;else break a;do{q=p;p=p+-1|0;r=b+p|0;s=i[r>>0]|0;i[r>>0]=s<<24>>24==g<<24>>24?f:s&255^128}while((q|0)>1)}else{if((c|0)>0)t=c;else break a;do{g=t;t=t+-1|0;q=b+t|0;i[q>>0]=l[q>>0]^128}while((g|0)>1)}break}default:{k[m>>2]=4;if((d|0)==1){g=i[e>>0]|0;if((c|0)>0)u=c;else break a;while(1){q=u;u=u+-1|0;s=i[b+u>>0]|0;k[b+(u<<2)>>2]=s<<24>>24==g<<24>>24?f:(s<<24>>24)+128|0;if((q|0)<=1)break a}}if((k[n>>2]|0)<=0){g=(c|0)<1e4?c:1e4;q=cy(g<<2)|0;if(!q){tk(780614);k[n>>2]=113;break a}b:do if((g|0)>0){s=c-g|0;r=g;while(1){if((r|0)>0){v=0;do{k[q+(v<<2)>>2]=(i[b+(v+s)>>0]|0)+128;v=v+1|0}while((v|0)!=(r|0))}py(b+(s<<2)|0,q|0,r<<2|0)|0;if(!s)break b;v=(s|0)>1e4;r=v?r:s;if((r|0)<=0)break;else s=v?s+-1e4|0:0}}while(0);dy(q)}}}while(0);o=k[n>>2]|0;return o|0}
function Zb(a){a=a|0;var b=0;b=r;r=r+a|0;r=r+15&-16;return b|0}function _b(){return r|0}function $b(a){a=a|0;r=a}function ac(a,b){a=a|0;b=b|0;r=a;s=b}function bc(a,b){a=a|0;b=b|0;if(!w){w=a;x=b}}function cc(a){a=a|0;i[t>>0]=i[a>>0];i[t+1>>0]=i[a+1>>0];i[t+2>>0]=i[a+2>>0];i[t+3>>0]=i[a+3>>0]}function dc(a){a=a|0;i[t>>0]=i[a>>0];i[t+1>>0]=i[a+1>>0];i[t+2>>0]=i[a+2>>0];i[t+3>>0]=i[a+3>>0];i[t+4>>0]=i[a+4>>0];i[t+5>>0]=i[a+5>>0];i[t+6>>0]=i[a+6>>0];i[t+7>>0]=i[a+7>>0]}function ec(a){a=a|0;L=a}function fc(){return L|0}function gc(a){a=a|0;Qa(101460,((a|0)==0?1:a)|0)}function hc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0;if((b|0)>0)c=ld(a,b)|0;else c=md(a)|0;a=(c|0)==0;if(!a)ud(c,Gd(c)|0);b=k[25404]|0;if(!b){k[25404]=10;d=cy(82e3)|0;k[25405]=d;if(!d){e=-4;return e|0}else f=10}else f=b;b=k[25406]|0;a:do if((b|0)<(f|0))g=b;else{d=f;h=k[25405]|0;while(1){j=d+10|0;k[25404]=j;h=fy(h,j*8200|0)|0;k[25405]=h;if(!h){e=-3;break}d=k[25404]|0;j=k[25406]|0;if((j|0)<(d|0)){g=j;break a}}return e|0}while(0);if(a){e=-1;return e|0}k[(k[25405]|0)+(g*8200|0)>>2]=c;c=k[25405]|0;k[c+(g*8200|0)+4>>2]=0;i[c+((k[25406]|0)*8200|0)+8>>0]=0;k[25406]=(k[25406]|0)+1;e=g;return e|0}function ic(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0,s=0,t=0,u=0,v=0.0;b=r;r=r+80|0;c=b;d=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;a=k[d>>2]|0;if(!a){e=0;f=0;g=0.0;h=0.0;i=0.0;j=0.0;l=0.0;m=0.0;n=0.0;o=0;q=0}else{s=(k[a+3304>>2]|0)==0;t=a+768|0;u=a+760|0;v=+p[a+48>>3];e=a+3449|0;f=a+3458|0;g=+p[(s?u:t)>>3];h=+p[(s?t:u)>>3];i=(k[a+3256>>2]|0)==0?v:-v;j=+p[a+616>>3];l=+p[a+624>>3];m=+p[a+688>>3];n=+p[a+696>>3];o=a+3467|0;q=a+3764|0}a=d+8|0;p[c>>3]=m;p[c+8>>3]=n;p[c+16>>3]=j;p[c+24>>3]=l;p[c+32>>3]=g;p[c+40>>3]=h;p[c+48>>3]=i;k[c+56>>2]=e;k[c+60>>2]=f;k[c+64>>2]=o;k[c+68>>2]=q;Ww(a,8191,336022,c)|0;r=b;return a|0}function jc(a,b,c){a=a|0;b=+b;c=+c;var d=0,e=0,f=0;d=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;a=k[d>>2]|0;if(!a){e=0;return e|0}f=d+8|0;i[f>>0]=0;Cd(a,b+-1.0,c+-1.0,f,8192)|0;e=f;return e|0}function kc(a,b,c){a=a|0;b=+b;c=+c;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0;d=r;r=r+8240|0;e=d+16|0;f=d+40|0;g=d+8|0;h=d;j=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;a=k[j>>2]|0;if(!a){l=0;r=d;return l|0}Bd(a,b,c,g,h,d+32|0);c=+p[h>>3]+1.0;p[e>>3]=+p[g>>3]+1.0;p[e+8>>3]=c;Ww(f,8191,336206,e)|0;e=j+8|0;j=i[f>>0]|0;a:do if(j<<24>>24){g=f;h=j;while(1){if(!(kv(h<<24>>24)|0)){m=g;break}g=g+1|0;h=i[g>>0]|0;if(!(h<<24>>24))break a}h=i[m>>0]|0;if(h<<24>>24){g=e;a=m;n=h;while(1){a=a+1|0;h=g+1|0;i[g>>0]=n;n=i[a>>0]|0;if(!(n<<24>>24)){o=h;break}else g=h}i[o>>0]=0;if((o|0)==(e|0)){l=e;r=d;return l|0}g=o;n=o-e|0;while(1){g=g+-1|0;if(!(kv(i[g>>0]|0)|0)){l=e;q=12;break}i[g>>0]=0;n=n+-1|0;if(!n){l=e;q=12;break}}if((q|0)==12){r=d;return l|0}}}while(0);i[e>>0]=0;l=e;r=d;return l|0}function lc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;c=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;a=k[c>>2]|0;if(!a){d=0;return d|0}e=c+8|0;i[e>>0]=0;do if((b|0)!=0?(i[b>>0]|0)!=0:0){if(((px(b,336216)|0)!=0?(px(b,336225)|0)!=0:0)?(px(b,336234)|0)!=0:0){if(!(+Nd(b)>0.0))break;f=k[c>>2]|0}else f=a;ud(f,b)}while(0);Bx(e,Ed(k[c>>2]|0)|0,8192)|0;if(!(px(e,336216)|0)){g=e;h=336216;j=g+9|0;do{i[g>>0]=i[h>>0]|0;g=g+1|0;h=h+1|0}while((g|0)<(j|0));d=e;return d|0}if(!(px(e,336225)|0)){g=e;h=336225;j=g+9|0;do{i[g>>0]=i[h>>0]|0;g=g+1|0;h=h+1|0}while((g|0)<(j|0));d=e;return d|0}h=i[e>>0]|0;if(!(h<<24>>24)){d=e;return d|0}else{l=e;m=h}do{if(jv(m<<24>>24)|0){h=(nv(i[l>>0]|0)|0)&255;i[l>>0]=h}l=l+1|0;m=i[l>>0]|0}while(m<<24>>24!=0);d=e;return d|0}function mc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;c=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;a=k[c>>2]|0;if(!a){d=0;return d|0}e=c+8|0;i[e>>0]=0;do if((b|0)!=0?(i[b>>0]|0)!=0:0)if(!(px(b,336241)|0)){Fd(a,1)|0;k[c+4>>2]=1;break}else{Fd(a,0)|0;k[c+4>>2]=0;break}while(0);switch(k[c+4>>2]|0){case 1:{Bx(e,336241,8191)|0;break}case 0:{Bx(e,336249,8191)|0;break}default:{}}c=i[e>>0]|0;if(!(c<<24>>24)){d=e;return d|0}else{f=e;g=c}do{if(lv(g<<24>>24)|0){c=(mv(i[f>>0]|0)|0)&255;i[f>>0]=c}f=f+1|0;g=i[f>>0]|0}while(g<<24>>24!=0);d=e;return d|0}function nc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0.0,R=0,S=0,T=0.0,U=0.0,V=0.0,W=0.0,X=0.0,Y=0.0,Z=0.0,_=0.0,$=0.0,aa=0.0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0.0;c=r;r=r+24832|0;d=c+232|0;e=c+224|0;f=c+216|0;g=c+208|0;h=c+200|0;j=c+192|0;l=c+184|0;m=c+176|0;n=c+168|0;o=c+152|0;q=c+144|0;s=c+128|0;t=c+112|0;u=c+96|0;v=c+88|0;w=c+80|0;x=c+64|0;y=c+56|0;z=c+40|0;A=c+32|0;B=c+16632|0;C=c+8440|0;D=c+248|0;E=c+240|0;F=c+236|0;G=c+24|0;H=c+16|0;I=c+8|0;J=c;K=(a|0)>0&(k[25404]|0)>(a|0)?(k[25405]|0)+(a*8200|0)|0:0;k[E>>2]=0;k[F>>2]=0;if(!(k[K>>2]|0)){L=0;r=c;return L|0}M=lc(a,0)|0;if(((tx(M,336216)|0)!=0?(tx(M,336225)|0)!=0:0)?(tx(M,336234)|0)!=0:0)N=0;else N=1;M=K+8|0;i[M>>0]=0;a=wx(b)|0;b=Fx(a,728157)|0;if(b){O=K+4|0;P=(N|0)==0;Q=0.0;N=b;while(1){b=rx(N,32)|0;if(!b){k[E>>2]=795593;R=795593;S=0}else{k[E>>2]=b+1;i[b>>0]=0;R=k[E>>2]|0;S=N}T=+hx(R,F);if(T!=0.0?(U=+hx(k[F>>2]|0,E),U!=0.0):0){Ad(k[K>>2]|0,T+-1.0,U+-1.0,G,H);b=(S|0)!=0;if(b){k[A>>2]=S;Ww(B,8192,336261,A)|0;zx(M,B,8191)|0}switch(k[O>>2]|0){case 1:{U=+p[H>>3];p[z>>3]=+p[G>>3];p[z+8>>3]=U;Ww(B,8192,336265,z)|0;zx(M,B,8191)|0;break}case 0:{U=+p[G>>3];if(P)qe(C,8191,U,3);else re(C,8191,U,3);re(D,8191,+p[H>>3],3);k[y>>2]=C;k[y+4>>2]=D;Ww(B,8192,336276,y)|0;zx(M,B,8191)|0;break}default:{U=+p[H>>3];p[x>>3]=+p[G>>3];p[x+8>>3]=U;Ww(B,8192,336265,x)|0;zx(M,B,8191)|0}}a:do if(tx(S,336283)|0){if((tx(S,775399)|0)!=0?(tx(S,775394)|0)!=0:0){U=+hx(k[E>>2]|0,F);if(U!=0.0){V=U;W=Q}else{X=U;Y=Q;break}while(1){U=+hx(k[F>>2]|0,E);if(!(U!=0.0)){X=V;Y=W;break a}Z=+hx(k[E>>2]|0,F);if(!(Z!=0.0)){X=V;Y=W;break a}_=+hx(k[F>>2]|0,E);if(!(_!=0.0)){X=V;Y=W;break a}Ad(k[K>>2]|0,V+-1.0,U+-1.0,G,H);Ad(k[K>>2]|0,Z+-1.0,_+-1.0,I,J);_=+Dd(+p[G>>3],+p[H>>3],+p[I>>3],+p[J>>3])*3600.0;do if(!(_<=60.0))if(!(_<=3600.0)){p[l>>3]=_/3600.0;Ww(B,8192,336330,l)|0;zx(M,B,8191)|0;break}else{p[m>>3]=_/60.0;Ww(B,8192,336322,m)|0;zx(M,B,8191)|0;break}else{p[n>>3]=_;Ww(B,8192,336314,n)|0;zx(M,B,8191)|0}while(0);Z=+hx(k[E>>2]|0,F);if(Z!=0.0){V=Z;W=_}else{X=Z;Y=_;break a}}}Z=+hx(k[E>>2]|0,F);b:do if(Z!=0.0)if(P){U=Z;while(1){$=+hx(k[F>>2]|0,E);if(!($!=0.0)){aa=U;break b}Ad(k[K>>2]|0,U+-1.0,$+-1.0,G,H);switch(k[O>>2]|0){case 1:{$=+p[H>>3];p[u>>3]=+p[G>>3];p[u+8>>3]=$;Ww(B,8192,336301,u)|0;zx(M,B,8191)|0;break}case 0:{qe(C,8191,+p[G>>3],3);re(D,8191,+p[H>>3],3);k[v>>2]=C;k[v+4>>2]=D;Ww(B,8192,336292,v)|0;zx(M,B,8191)|0;break}default:{$=+p[H>>3];p[t>>3]=+p[G>>3];p[t+8>>3]=$;Ww(B,8192,336301,t)|0;zx(M,B,8191)|0}}$=+hx(k[E>>2]|0,F);if($!=0.0)U=$;else{aa=$;break b}}}else{U=Z;while(1){_=+hx(k[F>>2]|0,E);if(!(_!=0.0)){aa=U;break b}Ad(k[K>>2]|0,U+-1.0,_+-1.0,G,H);switch(k[O>>2]|0){case 1:{_=+p[H>>3];p[s>>3]=+p[G>>3];p[s+8>>3]=_;Ww(B,8192,336301,s)|0;zx(M,B,8191)|0;break}case 0:{re(C,8191,+p[G>>3],3);re(D,8191,+p[H>>3],3);k[q>>2]=C;k[q+4>>2]=D;Ww(B,8192,336292,q)|0;zx(M,B,8191)|0;break}default:{_=+p[H>>3];p[o>>3]=+p[G>>3];p[o+8>>3]=_;Ww(B,8192,336301,o)|0;zx(M,B,8191)|0}}_=+hx(k[E>>2]|0,F);if(_!=0.0)U=_;else{aa=_;break b}}}else aa=Z;while(0);if(!(tx(S,775394)|0)){k[E>>2]=R;Z=+hx(R,F);if((Z!=0.0?(U=+hx(k[F>>2]|0,E),U!=0.0):0)?(_=+hx(k[E>>2]|0,F),_!=0.0):0){$=_;_=Z;ba=U;U=Q;while(1){ca=ba;ba=+hx(k[F>>2]|0,E);if(!(ba!=0.0)){X=_;Y=U;break a}Ad(k[K>>2]|0,_+-1.0,ca+-1.0,G,H);Ad(k[K>>2]|0,$+-1.0,ba+-1.0,I,J);ca=U+ +Dd(+p[G>>3],+p[H>>3],+p[I>>3],+p[J>>3])*3600.0;da=+hx(k[E>>2]|0,F);if(!(da!=0.0)){X=$;Y=ca;break}else{ea=$;$=da;U=ca;_=ea}}}else{X=Z;Y=Q}}else{X=aa;Y=Q}}else{k[w>>2]=k[E>>2];Ww(B,8192,336288,w)|0;zx(M,B,8191)|0;X=T;Y=Q}while(0);if(!((tx(S,775305)|0)!=0?(tx(S,775283)|0)!=0:0)){if(X<0.0){T=X;while(1){_=T+6.283185307179586;if(_<0.0)T=_;else{fa=_;break}}}else fa=X;p[j>>3]=fa*57.29577951308232;Ww(B,8192,336338,j)|0;zx(M,B,8191)|0}if(b){Ww(B,8192,761359,h)|0;zx(M,B,8191)|0}do if(!(tx(S,775394)|0)){if(Y<=60.0){p[g>>3]=Y;Ww(B,8192,336345,g)|0;zx(M,B,8191)|0;break}if(!(Y<=3600.0)){p[e>>3]=Y/3600.0;Ww(B,8192,336409,e)|0;zx(M,B,8191)|0;break}else{p[f>>3]=Y/60.0;Ww(B,8192,336377,f)|0;zx(M,B,8191)|0;break}}while(0);Ww(B,8192,728157,d)|0;zx(M,B,8191)|0;ga=Y}else ga=Q;N=Fx(0,728157)|0;if(!N)break;else Q=ga}}if(!a){L=M;r=c;return L|0}dy(a);L=M;r=c;return L|0}function oc(a){a=a|0;return +(+id(a,0))}function pc(){return k[25986]|0}function qc(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=+e;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0.0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;h=r;r=r+8224|0;j=h;l=4;m=cy(40)|0;k[m>>2]=0;n=h+20|0;q=h+16|0;s=h+24|0;i[s>>0]=0;m=ly(101460,1,m|0,l|0)|0;l=L;w=0;t=w;w=0;if((t|0)!=0&(x|0)!=0){u=oy(k[t>>2]|0,m|0,l|0)|0;if(!u)Qa(t|0,x|0);L=x}else u=-1;if((u|0)==1)v=L;else v=0;a:while(1){if(v){y=12;break}w=0;ta(1,a|0,b|0,c|0,d|0,n|0,q|0,+e,f|0,g|0);u=w;w=0;if((u|0)!=0&(x|0)!=0){z=oy(k[u>>2]|0,m|0,l|0)|0;if(!z)Qa(u|0,x|0);L=x}else z=-1;if((z|0)==1){v=L;continue}A=+o[q>>2];w=0;p[j>>3]=+o[n>>2];p[j+8>>3]=A;ua(1,s|0,8191,336442,j|0)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){B=oy(k[u>>2]|0,m|0,l|0)|0;if(!B)Qa(u|0,x|0);L=x}else B=-1;if((B|0)==1){v=L;continue}u=i[s>>0]|0;if(!(u<<24>>24)){y=12;break}else{C=s;D=u}while(1){w=0;u=sa(2,D<<24>>24|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){E=oy(k[t>>2]|0,m|0,l|0)|0;if(!E)Qa(t|0,x|0);L=x}else E=-1;if((E|0)==1){v=L;continue a}if(!u)break;u=C+1|0;D=i[u>>0]|0;if(!(D<<24>>24)){y=12;break a}else C=u}u=i[C>>0]|0;if(!(u<<24>>24)){y=12;break}else{F=336448;G=C;H=u}do{G=G+1|0;u=F;F=F+1|0;i[u>>0]=H;H=i[G>>0]|0}while(H<<24>>24!=0);i[F>>0]=0;if((F|0)==336448){y=18;break}u=F;t=F-336448|0;while(1){u=u+-1|0;w=0;I=sa(2,i[u>>0]|0)|0;J=w;w=0;if((J|0)!=0&(x|0)!=0){K=oy(k[J>>2]|0,m|0,l|0)|0;if(!K)Qa(J|0,x|0);L=x}else K=-1;if((K|0)==1){v=L;continue a}if(!I){y=18;break a}i[u>>0]=0;t=t+-1|0;if(!t){y=18;break a}}}if((y|0)==12){i[336448]=0;dy(m|0);r=h;return 336448}else if((y|0)==18){dy(m|0);r=h;return 336448}return 0}function rc(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0;d=r;r=r+139296|0;e=d+16|0;f=d+8|0;g=d;h=4;j=cy(40)|0;k[j>>2]=0;l=d+24|0;m=d+57368|0;n=d+49176|0;o=d+40984|0;p=d+32792|0;k[l>>2]=344640;a:do if((c|0)!=0?(i[c>>0]|0)!=0:0){w=0;q=sa(3,c|0)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){t=oy(k[s>>2]|0,j|0,h|0)|0;if(!t)Qa(s|0,x|0);L=x}else t=-1;if((t|0)!=1){w=0;s=va(5,q|0,344648)|0;u=w;w=0;if((u|0)!=0&(x|0)!=0){v=oy(k[u>>2]|0,j|0,h|0)|0;if(!v)Qa(u|0,x|0);L=x}else v=-1;if((v|0)!=1){if(!s)y=1;else{u=1;z=0;A=s;while(1){s=m+(z<<13)|0;w=0;na(6,s|0,A|0,8191)|0;B=w;w=0;if((B|0)!=0&(x|0)!=0){C=oy(k[B>>2]|0,j|0,h|0)|0;if(!C)Qa(B|0,x|0);L=x}else C=-1;if((C|0)==1){D=L;E=0;break a}z=z+1|0;B=u+1|0;k[l+(u<<2)>>2]=s;w=0;A=va(5,0,344648)|0;s=w;w=0;if((s|0)!=0&(x|0)!=0){F=oy(k[s>>2]|0,j|0,h|0)|0;if(!F)Qa(s|0,x|0);L=x}else F=-1;if((F|0)==1){D=L;E=0;break a}if(!((z|0)<10&(A|0)!=0)){y=B;break}else u=B}}if(q){w=0;qa(1,q|0);u=w;w=0;if((u|0)!=0&(x|0)!=0){G=oy(k[u>>2]|0,j|0,h|0)|0;if(!G)Qa(u|0,x|0);L=x}else G=-1;if((G|0)==1){D=L;E=0}else{H=y;I=12}}else{H=y;I=12}}else{D=L;E=0}}else{D=L;E=0}}else{H=1;I=12}while(0);if((I|0)==12){y=H+1|0;k[l+(H<<2)>>2]=344651;G=k[25407]|0;k[25407]=G+1;w=0;k[g>>2]=769609;k[g+4>>2]=G;ua(1,n|0,8191,344654,g|0)|0;g=w;w=0;if((g|0)!=0&(x|0)!=0){J=oy(k[g>>2]|0,j|0,h|0)|0;if(!J)Qa(g|0,x|0);L=x}else J=-1;if((J|0)!=1){J=H+2|0;k[l+(y<<2)>>2]=n;w=0;k[f>>2]=769609;k[f+4>>2]=a;ua(1,o|0,8191,344670,f|0)|0;f=w;w=0;if((f|0)!=0&(x|0)!=0){K=oy(k[f>>2]|0,j|0,h|0)|0;if(!K)Qa(f|0,x|0);L=x}else K=-1;if((K|0)!=1){K=H+3|0;k[l+(J<<2)>>2]=o;w=0;k[e>>2]=769609;k[e+4>>2]=b;ua(1,p|0,8191,344670,e|0)|0;e=w;w=0;if((e|0)!=0&(x|0)!=0){M=oy(k[e>>2]|0,j|0,h|0)|0;if(!M)Qa(e|0,x|0);L=x}else M=-1;if((M|0)!=1){k[l+(K<<2)>>2]=p;j=ly(101460,1,j|0,h|0)|0;h=L;w=0;p=w;w=0;if((p|0)!=0&(x|0)!=0){N=oy(k[p>>2]|0,j|0,h|0)|0;if(!N)Qa(p|0,x|0);L=x}else N=-1;if((N|0)!=1){D=0;E=H}else{D=L;E=H}}else{D=L;E=H}}else{D=L;E=H}}else{D=L;E=H}}while(1){if(!D){w=0;va(6,E+4|0,l|0)|0;H=w;w=0;if((H|0)!=0&(x|0)!=0){O=oy(k[H>>2]|0,j|0,h|0)|0;if(!O)Qa(H|0,x|0);L=x}else O=-1;P=E;D=L;E=P;continue}w=0;H=va(7,n|0,774853)|0;N=w;w=0;if((N|0)!=0&(x|0)!=0){Q=oy(k[N>>2]|0,j|0,h|0)|0;if(!Q)Qa(N|0,x|0);L=x}else Q=-1;if((Q|0)==1){P=E;D=L;E=P;continue}if(!H){R=344675;I=26;break}w=0;N=ua(2,336448,1,8191,H|0)|0;p=w;w=0;if((p|0)!=0&(x|0)!=0){S=oy(k[p>>2]|0,j|0,h|0)|0;if(!S)Qa(p|0,x|0);L=x}else S=-1;if((S|0)==1){P=E;D=L;E=P;continue}w=0;sa(4,H|0)|0;H=w;w=0;if((H|0)!=0&(x|0)!=0){T=oy(k[H>>2]|0,j|0,h|0)|0;if(!T)Qa(H|0,x|0);L=x}else T=-1;if((T|0)==1){P=E;D=L;E=P;continue}i[336448+N>>0]=0;if((N|0)<=0){R=344675;I=26;break}w=0;sa(5,n|0)|0;N=w;w=0;if((N|0)!=0&(x|0)!=0){U=oy(k[N>>2]|0,j|0,h|0)|0;if(!U)Qa(N|0,x|0);L=x}else U=-1;if((U|0)==1){P=E;D=L;E=P}else{I=25;break}}if((I|0)==25){R=336448;dy(j|0);r=d;return R|0}else if((I|0)==26){dy(j|0);r=d;return R|0}return 0}function sc(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0;e=r;r=r+147488|0;f=e+24|0;g=e+16|0;h=e+8|0;j=e;l=4;m=cy(40)|0;k[m>>2]=0;n=e+32|0;o=e+65568|0;p=e+57376|0;q=e+49184|0;s=e+40992|0;t=e+32800|0;k[n>>2]=344720;a:do if((d|0)!=0?(i[d>>0]|0)!=0:0){w=0;u=sa(3,d|0)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){y=oy(k[v>>2]|0,m|0,l|0)|0;if(!y)Qa(v|0,x|0);L=x}else y=-1;if((y|0)!=1){w=0;v=va(5,u|0,344648)|0;z=w;w=0;if((z|0)!=0&(x|0)!=0){A=oy(k[z>>2]|0,m|0,l|0)|0;if(!A)Qa(z|0,x|0);L=x}else A=-1;if((A|0)!=1){if(!v)B=1;else{z=1;C=0;D=v;while(1){v=o+(C<<13)|0;w=0;na(6,v|0,D|0,8191)|0;E=w;w=0;if((E|0)!=0&(x|0)!=0){F=oy(k[E>>2]|0,m|0,l|0)|0;if(!F)Qa(E|0,x|0);L=x}else F=-1;if((F|0)==1){G=L;H=0;break a}C=C+1|0;E=z+1|0;k[n+(z<<2)>>2]=v;w=0;D=va(5,0,344648)|0;v=w;w=0;if((v|0)!=0&(x|0)!=0){I=oy(k[v>>2]|0,m|0,l|0)|0;if(!I)Qa(v|0,x|0);L=x}else I=-1;if((I|0)==1){G=L;H=0;break a}if(!((C|0)<10&(D|0)!=0)){B=E;break}else z=E}}if(u){w=0;qa(1,u|0);z=w;w=0;if((z|0)!=0&(x|0)!=0){J=oy(k[z>>2]|0,m|0,l|0)|0;if(!J)Qa(z|0,x|0);L=x}else J=-1;if((J|0)==1){G=L;H=0}else{K=B;M=12}}else{K=B;M=12}}else{G=L;H=0}}else{G=L;H=0}}else{K=1;M=12}while(0);if((M|0)==12){B=K+1|0;k[n+(K<<2)>>2]=344651;J=k[25407]|0;k[25407]=J+1;w=0;k[j>>2]=769609;k[j+4>>2]=J;ua(1,p|0,8191,344654,j|0)|0;j=w;w=0;if((j|0)!=0&(x|0)!=0){N=oy(k[j>>2]|0,m|0,l|0)|0;if(!N)Qa(j|0,x|0);L=x}else N=-1;if((N|0)!=1){N=K+2|0;k[n+(B<<2)>>2]=p;w=0;k[h>>2]=769609;k[h+4>>2]=a;ua(1,q|0,8191,344670,h|0)|0;h=w;w=0;if((h|0)!=0&(x|0)!=0){O=oy(k[h>>2]|0,m|0,l|0)|0;if(!O)Qa(h|0,x|0);L=x}else O=-1;if((O|0)!=1){O=K+3|0;k[n+(N<<2)>>2]=q;w=0;k[g>>2]=769609;k[g+4>>2]=b;ua(1,s|0,8191,344670,g|0)|0;g=w;w=0;if((g|0)!=0&(x|0)!=0){P=oy(k[g>>2]|0,m|0,l|0)|0;if(!P)Qa(g|0,x|0);L=x}else P=-1;if((P|0)!=1){P=K+4|0;k[n+(O<<2)>>2]=s;w=0;k[f>>2]=769609;k[f+4>>2]=c;ua(1,t|0,8191,344670,f|0)|0;f=w;w=0;if((f|0)!=0&(x|0)!=0){Q=oy(k[f>>2]|0,m|0,l|0)|0;if(!Q)Qa(f|0,x|0);L=x}else Q=-1;if((Q|0)!=1){k[n+(P<<2)>>2]=t;m=ly(101460,1,m|0,l|0)|0;l=L;w=0;t=w;w=0;if((t|0)!=0&(x|0)!=0){R=oy(k[t>>2]|0,m|0,l|0)|0;if(!R)Qa(t|0,x|0);L=x}else R=-1;if((R|0)!=1){G=0;H=K}else{G=L;H=K}}else{G=L;H=K}}else{G=L;H=K}}else{G=L;H=K}}else{G=L;H=K}}while(1){if(!G){w=0;va(8,H+5|0,n|0)|0;K=w;w=0;if((K|0)!=0&(x|0)!=0){S=oy(k[K>>2]|0,m|0,l|0)|0;if(!S)Qa(K|0,x|0);L=x}else S=-1;T=H;G=L;H=T;continue}w=0;K=va(7,p|0,774853)|0;R=w;w=0;if((R|0)!=0&(x|0)!=0){U=oy(k[R>>2]|0,m|0,l|0)|0;if(!U)Qa(R|0,x|0);L=x}else U=-1;if((U|0)==1){T=H;G=L;H=T;continue}if(!K){V=344731;M=27;break}w=0;R=ua(2,336448,1,8191,K|0)|0;t=w;w=0;if((t|0)!=0&(x|0)!=0){W=oy(k[t>>2]|0,m|0,l|0)|0;if(!W)Qa(t|0,x|0);L=x}else W=-1;if((W|0)==1){T=H;G=L;H=T;continue}w=0;sa(4,K|0)|0;K=w;w=0;if((K|0)!=0&(x|0)!=0){X=oy(k[K>>2]|0,m|0,l|0)|0;if(!X)Qa(K|0,x|0);L=x}else X=-1;if((X|0)==1){T=H;G=L;H=T;continue}i[336448+R>>0]=0;if((R|0)<=-1){V=344731;M=27;break}w=0;sa(5,p|0)|0;R=w;w=0;if((R|0)!=0&(x|0)!=0){Y=oy(k[R>>2]|0,m|0,l|0)|0;if(!Y)Qa(R|0,x|0);L=x}else Y=-1;if((Y|0)==1){T=H;G=L;H=T}else{M=26;break}}if((M|0)==26){V=336448;dy(m|0);r=e;return V|0}else if((M|0)==27){dy(m|0);r=e;return V|0}return 0}function tc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0;b=r;r=r+16400|0;c=b+8|0;d=b;e=b+8208|0;f=b+16|0;k[d>>2]=769609;k[d+4>>2]=a;Ww(e,8191,344670,d)|0;k[c>>2]=769609;k[c+4>>2]=a;Ww(f,8191,344779,c)|0;Ec(e,f)|0;e=Dw(f,774853)|0;if(!e){g=344792;r=b;return g|0}c=Hw(336448,1,8191,e)|0;ww(e)|0;i[336448+c>>0]=0;if((c|0)<=-1){g=344792;r=b;return g|0}Nx(f)|0;g=336448;r=b;return g|0}function uc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;c=r;r=r+8208|0;d=c;e=c+8|0;k[d>>2]=769609;k[d+4>>2]=a;Ww(e,8191,344670,d)|0;i[336448]=0;d=Dw(e,774853)|0;if(!d){r=c;return 336448}e=b+-1|0;b=Hw(336448,1,e>>>0>8190?8190:e,d)|0;ww(d)|0;i[336448+b>>0]=0;r=c;return 336448}function vc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;b=r;r=r+8336|0;c=b+56|0;d=b+32|0;e=b+24|0;f=b+16|0;g=b+8|0;h=b;j=b+136|0;l=b+60|0;if((a|0)!=0?(i[a>>0]|0)!=0:0)m=a;else m=769609;a=pv(m)|0;if(!a){n=k[31231]|0;k[h>>2]=m;Ew(n,344836,h)|0;zw(n)|0;o=0;r=b;return o|0}n=qv(a)|0;a:do if(!n)p=0;else{h=k[31231]|0;q=l+64|0;s=l+56|0;t=l+12|0;u=l+36|0;v=n;w=0;while(1){x=v;while(1){y=x+11|0;if((tx(y,795591)|0)!=0?(tx(y,765881)|0)!=0:0){z=m+((xx(m)|0)+-1)|0;if((i[z>>0]|0)==47){k[g>>2]=m;k[g+4>>2]=y;Ww(j,8192,344670,g)|0}else{k[f>>2]=m;k[f+4>>2]=y;Ww(j,8192,344870,f)|0}if((hw(j,l)|0)>-1){A=y;break}k[c>>2]=y;Ew(h,344954,c)|0;zw(h)|0}x=qv(a)|0;if(!x){p=w;break a}}x=Fb(fb(q|0)|0)|0;y=Fb(fb(s|0)|0)|0;z=k[t>>2]&61440;b:do if((z|0)<24576){if((z|0)<8192){switch(z|0){case 4096:break;default:{B=20;break b}}C=344890;break}if((z|0)<16384){switch(z|0){case 8192:break;default:{B=20;break b}}C=344880;break}else{switch(z|0){case 16384:break;default:{B=20;break b}}C=344876;break}}else{if((z|0)<32768){switch(z|0){case 24576:break;default:{B=20;break b}}C=344885;break}if((z|0)<40960)switch(z|0){case 32768:{C=344902;break b;break}default:{B=20;break b}}switch(z|0){case 40960:break;default:{B=20;break b}}C=344895}while(0);if((B|0)==20){B=0;C=344900}if(!(tx(C,344876)|0)){k[e>>2]=A;Ew(h,344907,e)|0}else{z=k[u>>2]|0;k[d>>2]=A;k[d+4>>2]=C;k[d+8>>2]=z;k[d+12>>2]=x;k[d+16>>2]=y;Ew(h,344917,d)|0}zw(h)|0;z=w+1|0;v=qv(a)|0;if(!v){p=z;break}else w=z}}while(0);ov(a)|0;o=p;r=b;return o|0}function wc(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0;e=r;r=r+16|0;f=e+8|0;g=e+4|0;h=e;bl(a,f)|0;k[d>>2]=0;Fl(a,g,d)|0;k[d>>2]=0;a:do if((k[f>>2]|0)==1&(k[g>>2]|0)==0){if(b){i=wx(b)|0;j=Fx(i,776273)|0;b:do if(j){l=j;while(1){Jl(a,-1,l,0,d)|0;if(!(k[d>>2]|0))break;k[d>>2]=0;l=Fx(0,776273)|0;if(!l)break b}dy(i);break a}while(0);dy(i)}Fk(a,2,h,d)|0}while(0);Zk(a,c,d)|0;r=e;return a|0}function xc(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0;e=r;r=r+16|0;f=e;Ih(5,f,a,0,d)|0;if(k[d>>2]|0){g=0;r=e;return g|0}g=wc(k[f>>2]|0,b,c,d)|0;r=e;return g|0}function yc(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0;f=r;r=r+16|0;g=f;Bh(g,344968,0,a,b,0,0,e)|0;if(k[e>>2]|0){h=0;r=f;return h|0}h=wc(k[g>>2]|0,c,d,e)|0;r=f;return h|0}function zc(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0;j=r;r=r+208|0;l=j+108|0;m=j+104|0;n=j+88|0;o=j+72|0;q=j+56|0;s=j+40|0;t=j+16|0;u=j+8|0;v=j+112|0;w=j;x=j+24|0;k[m>>2]=0;p[t>>3]=1.0;p[u>>3]=0.0;y=w;k[y>>2]=0;k[y+4>>2]=1;k[x>>2]=0;k[x+4>>2]=0;k[x+8>>2]=0;k[x+12>>2]=0;k[n>>2]=0;k[n+4>>2]=0;k[n+8>>2]=0;k[n+12>>2]=0;k[o>>2]=1;k[q>>2]=1;k[s>>2]=1;k[o+4>>2]=1;k[q+4>>2]=1;k[s+4>>2]=1;k[o+8>>2]=1;k[q+8>>2]=1;k[s+8>>2]=1;k[o+12>>2]=1;k[q+12>>2]=1;k[s+12>>2]=1;Fl(a,l,h)|0;y=k[l>>2]|0;Gl(a,(y|0)<4?y:4,n,h)|0;El(a,g,h)|0;if((k[l>>2]|0)<2){k[h>>2]=320;z=0;r=j;return z|0}if((d|0)!=0?(i[d>>0]|0)!=0:0){y=wx(d)|0;d=Fx(y,344972)|0;do if(!d){A=0;B=1}else{C=0;D=0;E=d;while(1){if(!(tx(E,782783)|0))if((C|0)<2){k[w+(C<<2)>>2]=D;F=C+1|0}else F=C;else{G=ex(E)|0;k[x+(D<<2)>>2]=G;if((G|0)<1)break;if((G|0)>(k[n+(D<<2)>>2]|0))break;else F=C}E=Fx(0,344972)|0;D=D+1|0;if(!((E|0)!=0&(D|0)<4)){H=13;break}else C=F}if((H|0)==13){A=k[w>>2]|0;B=k[w+4>>2]|0;break}k[h>>2]=116;z=0;r=j;return z|0}while(0);dy(y);I=A;J=B}else{I=0;J=1}if(((b|0)!=0?(B=k[b>>2]|0,(B|0)!=0):0)?(A=k[b+4>>2]|0,(A|0)!=0):0){b=k[n+(I<<2)>>2]|0;y=(B|0)>(b|0)?b:B;B=k[n+(J<<2)>>2]|0;w=(A|0)>(B|0)?B:A;if(!c){A=(y|0)/2|0;B=(w|0)/2|0;K=A;L=B;M=A;N=B}else{K=(y|0)/2|0;L=(w|0)/2|0;M=~~+p[c>>3];N=~~+p[c+8>>3]}c=o+(I<<2)|0;k[c>>2]=M-((y+1|0)/2|0);y=o+(J<<2)|0;k[y>>2]=N-((w+1|0)/2|0);w=q+(I<<2)|0;k[w>>2]=K+M;M=q+(J<<2)|0;k[M>>2]=L+N;O=w;P=y;Q=M;R=c;S=b}else{b=o+(I<<2)|0;k[b>>2]=1;c=o+(J<<2)|0;k[c>>2]=1;M=k[n+(I<<2)>>2]|0;y=q+(I<<2)|0;k[y>>2]=M;I=q+(J<<2)|0;k[I>>2]=k[n+(J<<2)>>2];O=y;P=c;Q=I;R=b;S=M}M=k[R>>2]|0;b=(M|0)<1?1:M;k[R>>2]=(b|0)>(S|0)?S:b;b=k[O>>2]|0;M=(b|0)<1?1:b;k[O>>2]=(M|0)>(S|0)?S:M;M=k[P>>2]|0;b=(M|0)<1?1:M;M=(b|0)>(S|0)?S:b;k[P>>2]=M;b=k[Q>>2]|0;I=(b|0)<1?1:b;b=(I|0)>(S|0)?S:I;k[Q>>2]=b;I=k[l>>2]|0;if((I|0)>0){l=(I|0)<4?I:4;I=0;do{S=k[x+(I<<2)>>2]|0;if(S){c=(S|0)<1?1:S;S=k[n+(I<<2)>>2]|0;y=(c|0)>(S|0)?S:c;k[o+(I<<2)>>2]=y;k[q+(I<<2)>>2]=y}I=I+1|0}while((I|0)<(l|0));T=k[Q>>2]|0;U=k[P>>2]|0}else{T=b;U=M}k[e>>2]=(k[O>>2]|0)+1-(k[R>>2]|0);R=T-U+1|0;k[f>>2]=R;f=ha(R,k[e>>2]|0)|0;if((f|0)<2){k[h>>2]=323;z=0;r=j;return z|0}Io(a,82,779536,t,v,m)|0;if((k[m>>2]|0)!=204)Io(a,82,779507,u,v,m)|0;m=+p[t>>3]!=1.0|+p[u>>3]!=0.0;a:do switch(k[g>>2]|0){case 8:{if(m){k[g>>2]=-32;V=f<<2;W=42}else{V=f;W=11}break}case 16:{if(m){k[g>>2]=-32;V=f<<2;W=42;break a}else{V=f<<1;W=21;break a}break}case -16:{if(m){k[g>>2]=-32;V=f<<2;W=42;break a}else{V=f<<1;W=20;break a}break}case 32:{if(m){k[g>>2]=-32;V=f<<2;W=42;break a}else{V=f<<2;W=31;break a}break}case 64:{if(m){k[g>>2]=-32;V=f<<2;W=42;break a}else{V=f<<3;W=81;break a}break}case -32:{V=f<<2;W=42;break}case -64:{V=f<<3;W=82;break}default:{z=0;r=j;return z|0}}while(0);if((V|0)>(k[25408]|0)){k[h>>2]=113;z=0;r=j;return z|0}f=cy(V)|0;if(!f){k[h>>2]=113;z=0;r=j;return z|0}else{Xl(a,W,o,q,s,0,f,0,h)|0;z=f;r=j;return z|0}return 0}function Ac(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0.0,I=0,J=0,K=0.0,L=0.0,M=0.0,N=0.0,O=0,P=0,Q=0.0,R=0.0,S=0,T=0,U=0.0,V=0.0;h=r;r=r+1488|0;j=h+192|0;l=h+188|0;m=h+184|0;n=h+96|0;q=h+64|0;s=h+32|0;t=h;u=h+1403|0;v=h+1332|0;w=h+1048|0;x=h+764|0;y=h+480|0;z=h+196|0;A=h+168|0;B=h+152|0;C=h+136|0;D=h+120|0;E=h+104|0;F=(f|0)==0?1:f;i[v>>0]=0;if(((c|0)!=0?(f=k[c>>2]|0,(f|0)!=0):0)?(G=c+4|0,(k[G>>2]|0)!=0):0){ux(w,f)|0;ux(w+71|0,k[G>>2]|0)|0}else{i[w>>0]=0;i[w+71>>0]=0}H=+(F|0);p[q>>3]=-9.1191291391491e-36;p[s>>3]=-9.1191291391491e-36;p[t>>3]=H;i[x>>0]=0;i[y>>0]=0;i[z>>0]=0;G=q+8|0;p[G>>3]=-9.1191291391491e-36;f=s+8|0;p[f>>3]=-9.1191291391491e-36;p[t+8>>3]=H;i[x+71>>0]=0;i[y+71>>0]=0;i[z+71>>0]=0;p[q+16>>3]=-9.1191291391491e-36;p[s+16>>3]=-9.1191291391491e-36;p[t+16>>3]=H;i[x+142>>0]=0;i[y+142>>0]=0;i[z+142>>0]=0;p[q+24>>3]=-9.1191291391491e-36;p[s+24>>3]=-9.1191291391491e-36;p[t+24>>3]=H;i[x+213>>0]=0;i[y+213>>0]=0;i[z+213>>0]=0;al(a,l,g)|0;c=k[l>>2]|0;l=cy(c+1|0)|0;if((b|0)!=0?(i[b>>0]|0)!=0:0){$j(a,b,0,c,m,l,g)|0;if((k[g>>2]|0)>0){I=0;r=h;return I|0}}else J=7;if((J|0)==7?(c|0)>-1:0)sy(l|0,1,((c|0)>0?c+1|0:1)|0)|0;Tp(a,2,w,q,s,t,x,y,z,A,B,C,D,E,g)|0;if((k[g>>2]|0)>0){I=0;r=h;return I|0}K=+((~~(+o[D>>2]+ +o[C>>2])|0)/2|0|0);L=+((~~(+o[D+4>>2]+ +o[C+4>>2])|0)/2|0|0);C=k[B>>2]|0;D=(d|0)!=0;if((D?(E=k[d>>2]|0,(E|0)!=0):0)?(A=k[d+4>>2]|0,(A|0)!=0):0){if(((e|0)!=0?(M=+p[e>>3],M!=0.0):0)?(N=+p[e+8>>3],N!=0.0):0){O=E;P=A;Q=M;R=N}else{O=(C|0)<(E|0)?C:E;P=(k[B+4>>2]|0)<(A|0)?C:A;Q=K;R=L}A=ha(O,F)|0;O=ha(P,F)|0;p[q>>3]=+(~~(Q-+((A+1|0)/2|0|0))|0);p[G>>3]=+(~~(R-+((O+1|0)/2|0|0))|0);p[s>>3]=+(~~(Q+ +((A|0)/2|0|0))|0);p[f>>3]=+(~~(R+ +((O|0)/2|0|0))|0);S=A;T=O;U=Q;V=R}else{S=C;T=C;U=K;V=L}C=Xp(a,777975,31,2,w,q,s,t,x,y,z,1.0,v,0,l,g)|0;if((k[g>>2]|0)>0){I=0;r=h;return I|0}p[n>>3]=0.0;i[u>>0]=0;k[j>>2]=0;Io(a,82,344976,n,u,j)|0;L=+(F|0);p[n>>3]=(+((S|0)/2|0|0)-U)/L;k[j>>2]=0;hq(C,82,344976,n,u,j)|0;p[n>>3]=0.0;i[u>>0]=0;k[j>>2]=0;Io(a,82,344981,n,u,j)|0;p[n>>3]=(+((T|0)/2|0|0)-V)/L;k[j>>2]=0;hq(C,82,344981,n,u,j)|0;L=1.0/H;p[n>>3]=L;i[u>>0]=0;k[j>>2]=0;Io(a,82,344986,n,u,j)|0;k[j>>2]=0;hq(C,82,344986,n,u,j)|0;p[n>>3]=0.0;i[u>>0]=0;k[j>>2]=0;Io(a,82,344993,n,u,j)|0;k[j>>2]=0;hq(C,82,344993,n,u,j)|0;p[n>>3]=0.0;i[u>>0]=0;k[j>>2]=0;Io(a,82,345e3,n,u,j)|0;k[j>>2]=0;hq(C,82,345e3,n,u,j)|0;p[n>>3]=L;i[u>>0]=0;k[j>>2]=0;Io(a,82,345007,n,u,j)|0;k[j>>2]=0;hq(C,82,345007,n,u,j)|0;if(D){k[d>>2]=(S|0)/(F|0)|0;k[d+4>>2]=(T|0)/(F|0)|0}if(e){p[e>>3]=U;p[e+8>>3]=V}if(!l){I=C;r=h;return I|0}dy(l);I=C;r=h;return I|0}function Bc(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;fp(a,1,0,0,b,c,d)|0;return}function Cc(a,b){a=a|0;b=b|0;Fh(a,b)|0;return}function Dc(a){a=a|0;var b=0;b=k[25408]|0;if(!a)return b|0;k[25408]=a;return b|0}function Ec(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;c=r;r=r+368|0;d=c+56|0;e=c+48|0;f=c+40|0;g=c+32|0;h=c+24|0;j=c+16|0;l=c+8|0;m=c;n=c+136|0;o=c+282|0;p=c+211|0;q=c+140|0;s=c+88|0;t=c+84|0;u=c+80|0;v=c+76|0;w=c+72|0;x=c+68|0;y=c+96|0;z=c+64|0;k[s>>2]=0;A=k[31231]|0;if(!(Ih(5,n,a,0,s)|0)){if((b|0)!=0?(i[b>>0]|0)!=0:0){a=Dw(b,741341)|0;if(!a){B=k[31229]|0;k[m>>2]=b;Ew(B,345014,m)|0;C=-1;r=c;return C|0}else D=a}else D=A;Fk(k[n>>2]|0,1,u,s)|0;bl(k[n>>2]|0,t)|0;k[t>>2]=(k[t>>2]|0)+-1;Fw(91,D)|0;a:do if(!(k[s>>2]|0)){while(1){Fw(123,D)|0;Zk(k[n>>2]|0,u,s)|0;k[l>>2]=k[t>>2];Ew(D,345049,l)|0;Io(k[n>>2]|0,16,790194,p,0,s)|0;if(!(k[s>>2]|0)){k[j>>2]=p;Ew(D,345059,j)|0}k[s>>2]=0;a=k[n>>2]|0;if(!(k[u>>2]|0)){Dl(a,10,v,w,y,s)|0;k[h>>2]=k[w>>2];Ew(D,345074,h)|0;Pw(345105,12,1,D)|0;if((k[w>>2]|0)>0){m=0;do{if(m)Pw(773939,2,1,D)|0;k[g>>2]=k[y+(m<<2)>>2];Ew(D,773048,g)|0;m=m+1|0}while((m|0)<(k[w>>2]|0))}Fw(93,D)|0;k[f>>2]=k[v>>2];Ew(D,345118,f)|0}else{al(a,z,s)|0;$k(k[n>>2]|0,x,s)|0;if((k[u>>2]|0)==1)Pw(345133,17,1,D)|0;else Pw(345151,17,1,D)|0;k[e>>2]=k[z>>2];Ew(D,345169,e)|0;Pw(345183,11,1,D)|0;if((k[x>>2]|0)>=1){m=1;while(1){Gk(772997,m,o,s)|0;Io(k[n>>2]|0,16,o,p,0,s)|0;Gk(773003,m,o,s)|0;Io(k[n>>2]|0,16,o,q,0,s)|0;k[d>>2]=p;k[d+4>>2]=q;Ew(D,345195,d)|0;if((m|0)==(k[x>>2]|0))E=m;else{Pw(773939,2,1,D)|0;E=k[x>>2]|0}if((m|0)<(E|0))m=m+1|0;else break}}Fw(93,D)|0}Fw(125,D)|0;Bl(k[n>>2]|0,1,0,s)|0;if(k[s>>2]|0)break;Pw(345224,2,1,D)|0;m=k[s>>2]|0;k[t>>2]=(k[t>>2]|0)+1;if(m)break a}k[t>>2]=(k[t>>2]|0)+1}while(0);Pw(345227,2,1,D)|0;if((k[s>>2]|0)==107)k[s>>2]=0;Fh(k[n>>2]|0,s)|0;F=D}else F=A;D=k[s>>2]|0;if(D)hi(k[31229]|0,D);if((F|0)==(A|0))zw(A)|0;else ww(F)|0;C=k[s>>2]|0;r=c;return C|0}function Fc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,u=0,v=0,w=0,x=0,y=0,z=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,W=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0.0,ab=0,bb=0.0,cb=0.0,db=0.0,eb=0.0,fb=0.0,gb=0.0,hb=0.0,ib=0.0,jb=0.0,kb=0,lb=0,mb=0,nb=0,ob=0.0,pb=0,qb=0.0,rb=0.0,sb=0.0,tb=0.0,ub=0,vb=0.0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0.0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0.0,Sb=0.0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0.0,$b=0.0,ac=0,bc=0.0,cc=0.0,dc=0,ec=0,fc=0,hc=0.0,ic=0.0,jc=0,kc=0,lc=0,mc=0,nc=0,oc=0,pc=0.0,qc=0.0,rc=0.0,sc=0,tc=0,uc=0,vc=0,wc=0,xc=0.0,yc=0.0,zc=0.0,Ac=0.0,Bc=0.0,Cc=0.0,Dc=0.0,Ec=0.0,Fc=0.0,Mc=0.0,Nc=0.0,Oc=0.0,Pc=0.0,Qc=0.0,Rc=0.0,Sc=0.0,Uc=0.0,Xc=0.0,Yc=0.0,Zc=0.0,$c=0.0,ad=0.0,cd=0.0,dd=0.0,ed=0.0,fd=0.0,gd=0.0,hd=0.0,id=0.0,jd=0.0,kd=0.0,ld=0.0,md=0,nd=0,od=0,pd=0,qd=0,rd=0.0,sd=0.0,td=0.0,ud=0.0,vd=0,wd=0,xd=0,yd=0,zd=0,Cd=0.0,Dd=0.0,Ed=0.0,Fd=0.0,Gd=0,Hd=0,Id=0,Jd=0,Kd=0,Ld=0,Md=0,Nd=0,Od=0.0,Pd=0,Qd=0;c=r;r=r+1600|0;d=c+1032|0;e=c+1024|0;f=c+1016|0;g=c+1008|0;h=c+1e3|0;j=c+992|0;l=c+984|0;m=c+976|0;n=c+968|0;o=c+960|0;q=c+936|0;s=c+920|0;u=c+912|0;v=c+904|0;w=c+896|0;x=c+888|0;y=c+880|0;z=c+872|0;B=c+856|0;C=c+832|0;D=c+824|0;E=c+808|0;F=c+800|0;G=c+792|0;H=c+784|0;I=c+776|0;J=c+768|0;K=c+760|0;L=c+752|0;M=c+744|0;N=c+736|0;O=c+728|0;P=c+688|0;Q=c+648|0;R=c+608|0;S=c+568|0;T=c+528|0;U=c+488|0;W=c+448|0;Y=c+408|0;Z=c+400|0;_=c+392|0;$=c+384|0;aa=c+376|0;ba=c+368|0;ca=c+360|0;da=c+352|0;ea=c+344|0;fa=c+336|0;ga=c+328|0;ia=c+320|0;ja=c+312|0;ka=c+304|0;la=c+296|0;ma=c+288|0;na=c+280|0;oa=c+272|0;pa=c+264|0;qa=c+256|0;ra=c+248|0;sa=c+240|0;ta=c+232|0;ua=c+224|0;va=c+216|0;wa=c+208|0;xa=c+200|0;ya=c+192|0;za=c+184|0;Aa=c+176|0;Ba=c+168|0;Ca=c+160|0;Da=c+152|0;Ea=c+1072|0;Fa=c+1056|0;Ga=c+120|0;Ha=c+88|0;Ia=c+80|0;Ja=c+72|0;Ka=c+64|0;La=c+56|0;Ma=c+48|0;Na=c+40|0;Oa=c+32|0;Pa=c+24|0;Qa=c+16|0;Sa=c+8|0;Ta=c+1048|0;Ua=c+1044|0;Va=c+1336|0;Wa=c+1080|0;Xa=c+1040|0;Ya=c;k[Ua>>2]=0;p[Ya>>3]=A;Ra(103728)|0;k[25933]=k[25932];k[25409]=0;k[31173]=0;k[25410]=0;k[25411]=0;i[345230]=0;i[345486]=0;c=k[31231]|0;k[25412]=c;k[31172]=1;Za=c;c=0;_a=0;$a=1.0;ab=0;bb=1.0;cb=1.0;db=0.0;eb=1.0;a:while(1)do switch(Pv(a,b,586766)|0){case 88:{fb=eb;gb=db;hb=cb;ib=bb;jb=$a;kb=_a;lb=c;ab=1;eb=fb;db=gb;cb=hb;bb=ib;$a=jb;_a=kb;c=lb;continue a;break}case -1:{mb=c;nb=_a;ob=$a;pb=ab;qb=bb;rb=cb;sb=db;tb=eb;ub=25;break a;break}case 122:{vb=+hx(k[31177]|0,Xa);wb=k[31177]|0;if((k[Xa>>2]|0)>>>0<(wb+(xx(wb)|0)|0)>>>0){xb=wb;ub=4;break a}else{fb=eb;gb=db;hb=cb;ib=bb;yb=ab;kb=_a;lb=c;$a=vb;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;_a=kb;c=lb;continue a}break}case 100:{wb=Vc(k[31177]|0)|0;k[25409]=wb;fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a;break}case 115:{wb=Dw(k[31177]|0,597225)|0;k[25412]=wb;if(!wb){ub=7;break a}else{fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a}break}case 105:{ux(345486,k[31177]|0)|0;fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a;break}case 111:{ux(345230,k[31177]|0)|0;fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a;break}case 119:{k[25411]=1;ux(345742,k[31177]|0)|0;fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a;break}case 87:{vb=+hx(k[31177]|0,Xa);wb=k[31177]|0;if((k[Xa>>2]|0)>>>0<(wb+(xx(wb)|0)|0)>>>0){zb=wb;ub=12;break a}else{gb=db;hb=cb;yb=ab;jb=$a;kb=_a;lb=c;bb=vb;eb=vb;db=gb;cb=hb;ab=yb;$a=jb;_a=kb;c=lb;continue a}break}case 116:{vb=+hx(k[31177]|0,Xa);wb=k[31177]|0;if((k[Xa>>2]|0)>>>0<(wb+(xx(wb)|0)|0)>>>0){Ab=wb;ub=14;break a}else{fb=eb;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;db=vb;eb=fb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a}break}case 120:{vb=+hx(k[31177]|0,Xa);wb=k[31177]|0;if((k[Xa>>2]|0)>>>0<(wb+(xx(wb)|0)|0)>>>0){Bb=wb;ub=16;break a}else{fb=eb;gb=db;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;cb=vb;eb=fb;db=gb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a}break}case 98:{wb=jx(k[31177]|0,Xa,10)|0;Cb=k[31177]|0;if((k[Xa>>2]|0)>>>0<(Cb+(xx(Cb)|0)|0)>>>0)if((Gc(Cb)|0)<4){ub=19;break a}else{fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;c=0;_a=1;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;continue a}else if((wb|0)<0){Db=wb;ub=21;break a}else{fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;c=wb;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;continue a}break}case 104:{wb=jx(k[31177]|0,Xa,10)|0;k[25410]=wb;Cb=k[31177]|0;if((wb|0)<0|(k[Xa>>2]|0)>>>0<(Cb+(xx(Cb)|0)|0)>>>0){Eb=Cb;ub=23;break a}else{fb=eb;gb=db;hb=cb;ib=bb;yb=ab;jb=$a;kb=_a;lb=c;eb=fb;db=gb;cb=hb;bb=ib;ab=yb;$a=jb;_a=kb;c=lb;continue a}break}default:{ub=24;break a}}while(0);switch(ub|0){case 4:{lb=k[25412]|0;k[Da>>2]=xb;Ew(lb,586790,Da)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 7:{k[Ca>>2]=k[31177];Ew(0,592096,Ca)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 12:{Ca=k[25412]|0;k[Ba>>2]=zb;Ew(Ca,586886,Ba)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 14:{Ba=k[25412]|0;k[Aa>>2]=Ab;Ew(Ba,586979,Aa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 16:{Aa=k[25412]|0;k[za>>2]=Bb;Ew(Aa,587077,za)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 19:{za=k[25412]|0;k[ya>>2]=k[31177];Ew(za,587169,ya)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 21:{ya=k[25412]|0;k[xa>>2]=Db;Ew(ya,587289,xa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 23:{xa=k[25412]|0;k[wa>>2]=Eb;Ew(xa,587375,wa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 24:{wa=k[25412]|0;k[va>>2]=k[b>>2];Ew(wa,587451,va)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1);break}case 25:{va=k[31172]|0;if((a-va|0)<3){a=k[25412]|0;k[ua>>2]=k[b>>2];Ew(a,587451,ua)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}ux(345998,k[b+(va<<2)>>2]|0)|0;ux(346254,k[b+((k[31172]|0)+1<<2)>>2]|0)|0;ux(Va,k[b+((k[31172]|0)+2<<2)>>2]|0)|0;Wc(345998,0,k[25410]|0)|0;Wc(Va,1,0)|0;if(i[345486]|0)Wc(345486,1,0)|0;if(i[345230]|0)Wc(345230,1,0)|0;b=xx(346254)|0;do if(b>>>0>5){va=346254+(b+-5)|0;if(!(Ax(va,587657,5)|0)){i[va>>0]=0;break}va=346254+(b+-5)|0;if(!(Ax(va,587663,5)|0))i[va>>0]=0;else ub=37}else if(b>>>0>4)ub=37;while(0);do if((ub|0)==37){va=346254+(b+-4)|0;if(!(Ax(va,587669,4)|0)){i[va>>0]=0;break}va=346254+(b+-4)|0;if(!(Ax(va,587674,4)|0))i[va>>0]=0}while(0);ux(346510,346254)|0;b=346254+(xx(346254)|0)|0;i[b>>0]=i[587663]|0;i[b+1>>0]=i[587664]|0;i[b+2>>0]=i[587665]|0;i[b+3>>0]=i[587666]|0;i[b+4>>0]=i[587667]|0;i[b+5>>0]=i[587668]|0;b=346510+(xx(346510)|0)|0;va=587679;ua=b+11|0;do{i[b>>0]=i[va>>0]|0;b=b+1|0;va=va+1|0}while((b|0)<(ua|0));if((k[25409]|0)>0){k[ta>>2]=345998;Rw(587690,ta)|0;k[sa>>2]=346254;Rw(587713,sa)|0;k[ra>>2]=346510;Rw(587735,ra)|0;k[qa>>2]=Va;Rw(587757,qa)|0;if(i[345486]|0){k[pa>>2]=345486;Rw(587779,pa)|0}if(i[345230]|0){k[oa>>2]=345230;Rw(587802,oa)|0}zw(Za)|0;if((k[25409]|0)>0){Ra(103728)|0;p[na>>3]=+((k[25932]|0)-(k[25933]|0)|0);Rw(587825,na)|0;zw(Za)|0}}Hc(345998,345742)|0;if(i[345486]|0)Ic(345486,1)|0;if((k[25409]|0)>0){k[ma>>2]=k[3];Rw(587867,ma)|0;k[la>>2]=k[4];Rw(587892,la)|0;k[ka>>2]=k[6];Rw(587917,ka)|0;p[ja>>3]=+p[4];Rw(587941,ja)|0;k[ia>>2]=k[10];Rw(587966,ia)|0;k[ga>>2]=(k[5]|0)+3467;Rw(587990,ga)|0;zw(Za)|0}ga=(k[5]|0)+3467|0;if(((((tx(ga,606498)|0)!=0?(tx(ga,606506)|0)!=0:0)?(tx(ga,606518)|0)!=0:0)?(tx(ga,606502)|0)!=0:0)?(tx(ga,606510)|0)!=0:0){k[fa>>2]=ga;Xw(Wa,588015,fa)|0;Jc(Wa)}p[6]=0.0;if(!pb)Fb=0.0;else{pb=k[3]|0;fa=ha(pb,pb)|0;pb=k[4]|0;jb=+(~~+X(+(+((ha(pb,pb)|0)+fa|0)))|0);p[6]=jb;Fb=jb}if((k[25409]|0)>0){p[ea>>3]=Fb;Rw(588099,ea)|0;zw(Za)|0}Ic(Va,0)|0;if(i[345230]|0)Ic(345230,2)|0;if((k[25409]|0)>0){k[da>>2]=k[15];Rw(588145,da)|0;k[ca>>2]=k[16];Rw(588171,ca)|0;k[ba>>2]=k[18];Rw(588196,ba)|0;p[aa>>3]=+p[10];Rw(588220,aa)|0;k[$>>2]=k[22];Rw(588245,$)|0;k[_>>2]=(k[17]|0)+3467;Rw(588269,_)|0;zw(Za)|0}_=(k[17]|0)+3467|0;if(((((tx(_,606498)|0)!=0?(tx(_,606506)|0)!=0:0)?(tx(_,606518)|0)!=0:0)?(tx(_,606502)|0)!=0:0)?(tx(_,606510)|0)!=0:0){k[Z>>2]=_;Xw(Wa,588293,Z)|0;Jc(Wa)}if((k[6]|0)!=(k[18]|0))Jc(588378);Wa=(i[345230]|0)!=0;do if(!(i[345486]|0)){Z=k[25413]|0;if(Wa){_=bd(96,Z,426766)|0;k[Ua>>2]=_;Gb=_;break}else{_=bd(96,Z,506766)|0;k[Ua>>2]=_;Gb=_;break}}else if(Wa){_=bd(96,346766,426766)|0;k[Ua>>2]=_;Gb=_;break}else{_=bd(96,346766,506766)|0;k[Ua>>2]=_;Gb=_;break}while(0);if(Gb){Pw(588455,102,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}if((k[25409]|0)>1){Tw(588558)|0;zw(Za)|0}Gb=k[3]|0;Wa=(Gb*24|0)+24|0;_=cy(Wa)|0;k[25414]=_;_=cy(Wa)|0;k[25415]=_;_=cy(Wa)|0;k[25416]=_;_=cy(Wa)|0;k[25417]=_;_=Gb<<3;Wa=cy(_)|0;if(!(k[25411]|0))Hb=0;else Hb=cy(_)|0;p[Oa>>3]=1.0e8;p[Qa>>3]=-1.0e8;p[Pa>>3]=1.0e8;p[Sa>>3]=-1.0e8;if((k[4]|0)>-1){_=0;while(1){Fb=+(_|0)+.5;_c(.5,Fb,Ia,Ja,96)|0;jb=+p[Ia>>3];do if(jb<-.5)Ib=1;else{$a=+p[Ja>>3];if($a<-.5|jb>+(k[55]|0)+1.5){Ib=1;break}Ib=$a>+(k[57]|0)+1.5}while(0);if((k[25409]|0)>2){$a=+p[Ja>>3];p[Y>>3]=.5;p[Y+8>>3]=Fb;p[Y+16>>3]=jb;p[Y+24>>3]=$a;k[Y+32>>2]=Ib&1;Rw(588603,Y)|0;zw(Za)|0;Ad(k[5]|0,.5,Fb,Ma,Na);Bd(k[17]|0,+p[Ma>>3],+p[Na>>3],Ka,La,Ta);$a=+p[Na>>3];ib=+p[Ka>>3];bb=+p[La>>3];Z=k[Ta>>2]|0;p[W>>3]=+p[Ma>>3];p[W+8>>3]=$a;p[W+16>>3]=ib;p[W+24>>3]=bb;k[W+32>>2]=Z;Rw(588635,W)|0}do if(!Ib){bb=+p[Ia>>3];if(bb<+p[Oa>>3])p[Oa>>3]=bb;if(bb>+p[Qa>>3])p[Qa>>3]=bb;bb=+p[Ja>>3];if(bb<+p[Pa>>3])p[Pa>>3]=bb;if(!(bb>+p[Sa>>3]))break;p[Sa>>3]=bb}while(0);_c(+(k[3]|0)+.5,Fb,Ia,Ja,96)|0;jb=+p[Ia>>3];do if(jb<-.5)Jb=1;else{bb=+p[Ja>>3];if(bb<-.5|jb>+(k[55]|0)+1.5){Jb=1;break}Jb=bb>+(k[57]|0)+1.5}while(0);if((k[25409]|0)>2){bb=+p[Ja>>3];p[U>>3]=+(k[3]|0)+.5;p[U+8>>3]=Fb;p[U+16>>3]=jb;p[U+24>>3]=bb;k[U+32>>2]=Jb&1;Rw(588603,U)|0;zw(Za)|0;Ad(k[5]|0,+(k[3]|0)+.5,Fb,Ma,Na);Bd(k[17]|0,+p[Ma>>3],+p[Na>>3],Ka,La,Ta);bb=+p[Na>>3];ib=+p[Ka>>3];$a=+p[La>>3];Z=k[Ta>>2]|0;p[T>>3]=+p[Ma>>3];p[T+8>>3]=bb;p[T+16>>3]=ib;p[T+24>>3]=$a;k[T+32>>2]=Z;Rw(588635,T)|0}do if(!Jb){$a=+p[Ia>>3];if($a<+p[Oa>>3])p[Oa>>3]=$a;if($a>+p[Qa>>3])p[Qa>>3]=$a;$a=+p[Ja>>3];if($a<+p[Pa>>3])p[Pa>>3]=$a;if(!($a>+p[Sa>>3]))break;p[Sa>>3]=$a}while(0);if((_|0)<(k[4]|0))_=_+1|0;else break}Kb=k[3]|0}else Kb=Gb;if((Kb|0)>-1){Kb=0;while(1){Fb=+(Kb|0)+.5;_c(Fb,.5,Ia,Ja,96)|0;jb=+p[Ia>>3];do if(jb<-.5)Lb=1;else{$a=+p[Ja>>3];if($a<-.5|jb>+(k[55]|0)+1.5){Lb=1;break}Lb=$a>+(k[57]|0)+1.5}while(0);if((k[25409]|0)>2){$a=+p[Ja>>3];p[S>>3]=Fb;p[S+8>>3]=.5;p[S+16>>3]=jb;p[S+24>>3]=$a;k[S+32>>2]=Lb&1;Rw(588603,S)|0;zw(Za)|0;Ad(k[5]|0,Fb,.5,Ma,Na);Bd(k[17]|0,+p[Ma>>3],+p[Na>>3],Ka,La,Ta);$a=+p[Na>>3];ib=+p[Ka>>3];bb=+p[La>>3];Gb=k[Ta>>2]|0;p[R>>3]=+p[Ma>>3];p[R+8>>3]=$a;p[R+16>>3]=ib;p[R+24>>3]=bb;k[R+32>>2]=Gb;Rw(588635,R)|0}do if(!Lb){bb=+p[Ia>>3];if(bb<+p[Oa>>3])p[Oa>>3]=bb;if(bb>+p[Qa>>3])p[Qa>>3]=bb;bb=+p[Ja>>3];if(bb<+p[Pa>>3])p[Pa>>3]=bb;if(!(bb>+p[Sa>>3]))break;p[Sa>>3]=bb}while(0);_c(Fb,+(k[4]|0)+.5,Ia,Ja,96)|0;jb=+p[Ia>>3];do if(jb<-.5)Mb=1;else{bb=+p[Ja>>3];if(bb<-.5|jb>+(k[55]|0)+1.5){Mb=1;break}Mb=bb>+(k[57]|0)+1.5}while(0);if((k[25409]|0)>2){bb=+(k[4]|0)+.5;ib=+p[Ja>>3];p[Q>>3]=Fb;p[Q+8>>3]=bb;p[Q+16>>3]=jb;p[Q+24>>3]=ib;k[Q+32>>2]=Mb&1;Rw(588603,Q)|0;zw(Za)|0;Ad(k[5]|0,Fb,+(k[4]|0)+.5,Ma,Na);Bd(k[17]|0,+p[Ma>>3],+p[Na>>3],Ka,La,Ta);ib=+p[Na>>3];bb=+p[Ka>>3];$a=+p[La>>3];Gb=k[Ta>>2]|0;p[P>>3]=+p[Ma>>3];p[P+8>>3]=ib;p[P+16>>3]=bb;p[P+24>>3]=$a;k[P+32>>2]=Gb;Rw(588635,P)|0}do if(!Mb){$a=+p[Ia>>3];if($a<+p[Oa>>3])p[Oa>>3]=$a;if($a>+p[Qa>>3])p[Qa>>3]=$a;$a=+p[Ja>>3];if($a<+p[Pa>>3])p[Pa>>3]=$a;if(!($a>+p[Sa>>3]))break;p[Sa>>3]=$a}while(0);if((Kb|0)<(k[3]|0))Kb=Kb+1|0;else break}}Kb=k[16]|0;if((Kb|0)>-1){Mb=0;while(1){p[Ia>>3]=.5;Fb=+(Mb|0)+.5;p[Ja>>3]=Fb;Kc(.5,Fb,Oa,Qa,Pa,Sa);Fb=+(k[15]|0)+.5;p[Ia>>3]=Fb;Kc(Fb,+p[Ja>>3],Oa,Qa,Pa,Sa);P=k[16]|0;if((Mb|0)<(P|0))Mb=Mb+1|0;else{Nb=P;break}}}else Nb=Kb;Kb=k[15]|0;if((Kb|0)>-1){Mb=0;while(1){Fb=+(Mb|0)+.5;p[Ia>>3]=Fb;p[Ja>>3]=.5;Kc(Fb,.5,Oa,Qa,Pa,Sa);Fb=+(k[16]|0)+.5;p[Ja>>3]=Fb;Kc(+p[Ia>>3],Fb,Oa,Qa,Pa,Sa);P=k[15]|0;if((Mb|0)<(P|0))Mb=Mb+1|0;else{Ob=P;break}}Pb=Ob;Qb=k[16]|0}else{Pb=Kb;Qb=Nb}Fb=+p[Oa>>3];Nb=~~(Fb+-1.0);Kb=(Nb|0)<0?0:Nb;jb=+p[Qa>>3];Nb=~~(jb-Fb+2.0);Ob=(Nb|0)>(Pb|0)?Pb:Nb;$a=+p[Pa>>3];Mb=~~($a+-1.0);Ia=(Mb|0)<0?0:Mb;Mb=~~(+p[Sa>>3]-$a+2.0);Ja=(Mb|0)>(Qb|0)?Qb:Mb;if((k[25409]|0)>1){Tw(588667)|0;p[O>>3]=+p[Oa>>3];Rw(588682,O)|0;p[N>>3]=+p[Qa>>3];Rw(588699,N)|0;p[M>>3]=+p[Pa>>3];Rw(588716,M)|0;p[L>>3]=+p[Sa>>3];Rw(588733,L)|0;k[K>>2]=Kb;Rw(588750,K)|0;k[J>>2]=Ob;Rw(588767,J)|0;k[I>>2]=Ia;Rw(588784,I)|0;k[H>>2]=Ja;Rw(588801,H)|0;zw(Za)|0;Rb=+p[Oa>>3];Sb=+p[Qa>>3]}else{Rb=Fb;Sb=jb}if(Rb>Sb){Tb=k[25412]|0;Pw(588818,40,1,Tb)|0;Ub=k[25412]|0;zw(Ub)|0;Vb=k[25412]|0;ww(Vb)|0;gc(1)}if(+p[Pa>>3]>+p[Sa>>3]){Tb=k[25412]|0;Pw(588818,40,1,Tb)|0;Ub=k[25412]|0;zw(Ub)|0;Vb=k[25412]|0;ww(Vb)|0;gc(1)}Vb=Ja<<2;Ub=cy(Vb)|0;if(!Ub){Pw(588859,75,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}Tb=(Ja|0)>0;b:do if(Tb){H=Ob<<3;I=0;while(1){J=cy(H)|0;k[Ub+(I<<2)>>2]=J;I=I+1|0;if(!J)break;if((I|0)>=(Ja|0))break b}Pw(588859,75,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}while(0);if((k[25409]|0)>0){I=ha(Ob<<3,Ja)|0;k[G>>2]=I;Rw(588935,G)|0;zw(Za)|0}if(Tb){G=(Ob|0)>0;I=(Nb|0)>(Pb|0)?Pb:Nb;H=(Qb|0)>(Mb|0)?Mb:Qb;J=0;do{if(G){K=k[Ub+(J<<2)>>2]|0;L=0;do{p[K+(L<<3)>>3]=A;L=L+1|0}while((L|0)!=(I|0))}J=J+1|0}while((J|0)!=(H|0))}H=cy(Vb)|0;if(!H){Pw(588974,75,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}c:do if(Tb){Vb=Ob<<3;J=(Ob|0)>0;I=-8-(((Nb|0)>(Pb|0)?Pb:Nb)<<3^-8)|0;G=0;while(1){L=cy(Vb)|0;k[H+(G<<2)>>2]=L;if(!L)break;if(J)sy(L|0,0,I|0)|0;G=G+1|0;if((G|0)>=(Ja|0))break c}Pw(588974,75,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}while(0);if((k[25409]|0)>0){G=ha(Ob<<3,Ja)|0;k[F>>2]=G;Rw(589050,F)|0;zw(Za)|0}k[Fa>>2]=1;F=Fa+4|0;k[F>>2]=mb+1;k[Fa+8>>2]=1;k[Fa+12>>2]=1;d:do if((mb|0)<((k[4]|0)-mb|0)){G=k[3]|0;I=(nb|0)==1;J=((G|0)<0)<<31>>31;Vb=ob!=1.0;L=ob==1.0;Sb=ob*.5;K=Ga+8|0;M=Ha+8|0;N=Ga+16|0;O=Ha+16|0;P=Ga+24|0;Ma=Ha+24|0;Ta=G;La=0;Ka=mb;Rb=0.0;jb=tb;e:while(1){Na=Ta-mb|0;do if(I){Q=Ta+-1|0;Fb=+(Ka|0);Lb=k[25419]|0;do if((Lb|0)<1){Wb=Q;Xb=0;Yb=0}else{R=0;S=1;Gb=k[25420]|0;_=k[25421]|0;$a=0.0;bb=+(Q|0)+1.0;while(1){Jb=(S|0)==(Lb|0)?0:S;T=Gb;Gb=k[101680+(Jb<<3)>>2]|0;U=_;_=k[101680+(Jb<<3)+4>>2]|0;ib=+(U|0);hb=+(_|0);do if(Fb>+((ib<hb?U:_)|0)){if(!(Fb<+((ib>hb?U:_)|0))){Zb=R;_b=$a;$b=bb;break}cb=+(T|0)+(Fb-ib)*+(Gb-T|0)/+(_-U|0);Zb=1;_b=$a>cb?$a:cb;$b=bb<cb?bb:cb}else{Zb=R;_b=$a;$b=bb}while(0);if((S|0)==(Lb|0)){ac=Zb;bc=_b;cc=$b;break}else{R=Zb;S=S+1|0;$a=_b;bb=$b}}if(!ac){Wb=Q;Xb=0;Yb=0;break}S=~~cc;R=~~(bc+.5);Wb=(R|0)>(Q|0)?Q:R;Xb=ac;Yb=(S|0)<0?0:S}while(0);if((k[25409]|0)>1){k[E>>2]=Ka;k[E+4>>2]=Yb;k[E+8>>2]=Wb;k[E+12>>2]=Xb;Rw(589087,E)|0;zw(Za)|0}if(Xb){dc=Wb;ec=Yb;ub=201;break}k[F>>2]=(k[F>>2]|0)+1;fc=La;hc=Rb;ic=jb}else{if((k[25409]|0)!=2){dc=Na;ec=mb;ub=201;break}k[D>>2]=Ka;Rw(589141,D)|0;zw(Za)|0;dc=Na;ec=mb;ub=201}while(0);do if((ub|0)==201){ub=0;if(Vl(k[2]|0,82,Fa,G,J,Ya,Wa,Ea,Ua)|0){ub=202;break e}if((k[25411]|0)!=0?(Vl(k[1728]|0,82,Fa,G,J,Ya,Hb,Ea,Ua)|0)!=0:0){ub=205;break e}k[F>>2]=(k[F>>2]|0)+1;f:do if(Vb|(La|0)==0){if((k[3]|0)<=-1){jc=La;break}Fb=+(Ka|0)+.5;bb=+(Ka+1|0)-Sb;if(L)kc=0;else{Na=0;while(1){$a=+(Na+1|0);Q=k[25414]|0;Lb=_c($a-Sb,bb,Q+(Na*24|0)|0,Q+(Na*24|0)+8|0,96)|0;k[(k[25414]|0)+(Na*24|0)+16>>2]=Lb;Lb=k[25415]|0;Q=_c(Sb+$a,bb,Lb+(Na*24|0)|0,Lb+(Na*24|0)+8|0,96)|0;k[(k[25415]|0)+(Na*24|0)+16>>2]=Q;Q=k[3]|0;if((Na|0)<(Q|0))Na=Na+1|0;else{lc=Q;mc=La;ub=214;break f}}}while(1){Na=k[25414]|0;Q=_c(+(kc|0)+.5,Fb,Na+(kc*24|0)|0,Na+(kc*24|0)+8|0,96)|0;Na=k[25414]|0;k[Na+(kc*24|0)+16>>2]=Q;if((kc|0)>0){Lb=k[25415]|0;S=kc+-1|0;p[Lb+(S*24|0)>>3]=+p[Na+(kc*24|0)>>3];p[Lb+(S*24|0)+8>>3]=+p[Na+(kc*24|0)+8>>3];k[Lb+(S*24|0)+16>>2]=Q}Q=k[3]|0;if((kc|0)<(Q|0))kc=kc+1|0;else{lc=Q;mc=1;ub=214;break}}}else{Q=k[25414]|0;k[25414]=k[25416];k[25416]=Q;Q=k[25415]|0;k[25418]=Q;k[25415]=k[25417];k[25417]=Q;lc=k[3]|0;mc=La;ub=214}while(0);g:do if((ub|0)==214){ub=0;if((lc|0)<=-1){jc=mc;break}Fb=+(Ka|0)+1.5;bb=Sb+ +(Ka+1|0);if(L)nc=0;else{Q=0;while(1){$a=+(Q+1|0);S=k[25416]|0;Lb=_c($a-Sb,bb,S+(Q*24|0)|0,S+(Q*24|0)+8|0,96)|0;k[(k[25416]|0)+(Q*24|0)+16>>2]=Lb;Lb=k[25417]|0;S=_c(Sb+$a,bb,Lb+(Q*24|0)|0,Lb+(Q*24|0)+8|0,96)|0;k[(k[25417]|0)+(Q*24|0)+16>>2]=S;if((Q|0)<(k[3]|0))Q=Q+1|0;else{jc=mc;break g}}}while(1){Q=k[25416]|0;S=_c(+(nc|0)+.5,Fb,Q+(nc*24|0)|0,Q+(nc*24|0)+8|0,96)|0;Q=k[25416]|0;k[Q+(nc*24|0)+16>>2]=S;if((nc|0)>0){Lb=k[25417]|0;Na=nc+-1|0;p[Lb+(Na*24|0)>>3]=+p[Q+(nc*24|0)>>3];p[Lb+(Na*24|0)+8>>3]=+p[Q+(nc*24|0)+8>>3];k[Lb+(Na*24|0)+16>>2]=S}if((nc|0)<(k[3]|0))nc=nc+1|0;else{jc=mc;break}}}while(0);if((ec|0)<(dc|0)){oc=ec;pc=Rb;qc=jb}else{fc=jc;hc=Rb;ic=jb;break}while(1){Fb=+p[Wa+(oc<<3)>>3];S=(k[25411]|0)==0;if(S)rc=qc;else{bb=+p[Hb+(oc<<3)>>3];rc=qb*(bb<sb?0.0:bb)}p[t>>3]=Fb;Na=k[t+4>>2]|0;Lb=Na&2147483647;Q=Na&2146435072;do if((Lb>>>0<2146435072|(Lb|0)==2146435072&(k[t>>2]|0)>>>0<1)&(Q>>>0<2146435072|(Q|0)==2146435072&0<0)){bb=rb*Fb;if((k[25409]|0)>2){if(S){k[B>>2]=Ka;k[B+4>>2]=oc;p[B+8>>3]=bb;Rw(589225,B)|0}else{k[C>>2]=Ka;k[C+4>>2]=oc;p[C+8>>3]=bb;p[C+16>>3]=rc;Rw(589169,C)|0}zw(Za)|0}p[Oa>>3]=1.0e8;p[Qa>>3]=-1.0e8;p[Pa>>3]=1.0e8;p[Sa>>3]=-1.0e8;if(!(k[10]|0)){Na=k[25415]|0;p[Ga>>3]=+p[Na+(oc*24|0)>>3];p[Ha>>3]=+p[Na+(oc*24|0)+8>>3];R=k[25414]|0;p[K>>3]=+p[R+(oc*24|0)>>3];p[M>>3]=+p[R+(oc*24|0)+8>>3];_=k[25416]|0;p[N>>3]=+p[_+(oc*24|0)>>3];p[O>>3]=+p[_+(oc*24|0)+8>>3];Gb=k[25417]|0;sc=Gb;tc=R;uc=Na;vc=_;wc=Gb}else{Gb=k[25417]|0;p[Ga>>3]=+p[Gb+(oc*24|0)>>3];p[Ha>>3]=+p[Gb+(oc*24|0)+8>>3];_=k[25416]|0;p[K>>3]=+p[_+(oc*24|0)>>3];p[M>>3]=+p[_+(oc*24|0)+8>>3];Na=k[25414]|0;p[N>>3]=+p[Na+(oc*24|0)>>3];p[O>>3]=+p[Na+(oc*24|0)+8>>3];R=k[25415]|0;sc=R;tc=Na;uc=R;vc=_;wc=Gb}p[P>>3]=+p[sc+(oc*24|0)>>3];p[Ma>>3]=+p[sc+(oc*24|0)+8>>3];Gb=tc+(oc*24|0)|0;$a=+p[Gb>>3];if($a<1.0e8){p[Oa>>3]=$a;xc=+p[Gb>>3];yc=$a}else{xc=$a;yc=1.0e8}if(xc>-1.0e8){p[Qa>>3]=xc;zc=xc}else zc=-1.0e8;Gb=tc+(oc*24|0)+8|0;$a=+p[Gb>>3];if($a<1.0e8){p[Pa>>3]=$a;Ac=+p[Gb>>3];Bc=$a}else{Ac=$a;Bc=1.0e8}if(Ac>-1.0e8){p[Sa>>3]=Ac;Cc=Ac}else Cc=-1.0e8;Gb=uc+(oc*24|0)|0;$a=+p[Gb>>3];if($a<yc){p[Oa>>3]=$a;Dc=+p[Gb>>3];Ec=$a}else{Dc=$a;Ec=yc}if(Dc>zc){p[Qa>>3]=Dc;Fc=Dc}else Fc=zc;Gb=uc+(oc*24|0)+8|0;$a=+p[Gb>>3];if($a<Bc){p[Pa>>3]=$a;Mc=+p[Gb>>3];Nc=$a}else{Mc=$a;Nc=Bc}if(Mc>Cc){p[Sa>>3]=Mc;Oc=Mc}else Oc=Cc;Gb=vc+(oc*24|0)|0;$a=+p[Gb>>3];if($a<Ec){p[Oa>>3]=$a;Pc=+p[Gb>>3];Qc=$a}else{Pc=$a;Qc=Ec}if(Pc>Fc){p[Qa>>3]=Pc;Rc=Pc}else Rc=Fc;Gb=vc+(oc*24|0)+8|0;$a=+p[Gb>>3];if($a<Nc){p[Pa>>3]=$a;Sc=+p[Gb>>3];Uc=$a}else{Sc=$a;Uc=Nc}if(Sc>Oc){p[Sa>>3]=Sc;Xc=Sc}else Xc=Oc;Gb=wc+(oc*24|0)|0;$a=+p[Gb>>3];if($a<Qc){p[Oa>>3]=$a;Yc=+p[Gb>>3];Zc=$a}else{Yc=$a;Zc=Qc}if(Yc>Rc){p[Qa>>3]=Yc;$c=Yc}else $c=Rc;Gb=wc+(oc*24|0)+8|0;$a=+p[Gb>>3];if($a<Uc){p[Pa>>3]=$a;ad=+p[Gb>>3];cd=$a}else{ad=$a;cd=Uc}if(ad>Xc){p[Sa>>3]=ad;dd=ad}else dd=Xc;if(k[uc+(oc*24|0)+16>>2]|k[tc+(oc*24|0)+16>>2]|k[vc+(oc*24|0)+16>>2]|k[wc+(oc*24|0)+16>>2]){ed=pc;break}Gb=~~+V(+(Zc+-.5));_=~~(+V(+($c+-.5))+1.0);R=~~+V(+(cd+-.5));Na=~~(+V(+(dd+-.5))+1.0);if((k[25409]|0)>2){Sw(10)|0;p[z>>3]=+p[Oa>>3];Rw(589267,z)|0;p[y>>3]=+p[Qa>>3];Rw(589288,y)|0;p[x>>3]=+p[Pa>>3];Rw(589309,x)|0;p[w>>3]=+p[Sa>>3];Rw(589330,w)|0;Sw(10)|0;k[v>>2]=Gb;k[v+4>>2]=_;Rw(589351,v)|0;k[u>>2]=R;k[u+4>>2]=Na;Rw(589379,u)|0;Sw(10)|0}if((R|0)>=(Na|0)){ed=pc;break}U=rc>0.0;T=(Gb|0)>=(_|0);Jb=R;$a=pc;while(1){R=Jb-Ia|0;if((R|0)>=(Ja|0)|(R|0)<0|T)fd=$a;else{ib=+(Jb|0);hb=ib+1.5;cb=ib+.5;Ib=Ub+(R<<2)|0;W=H+(R<<2)|0;R=Gb;ib=$a;while(1){Y=R-Kb|0;do if((Y|0)>-1&(Y|0)<(Ob|0)){gb=+(R|0);if(U)gd=+Tc(Ga,Ha,gb+.5,gb+1.5,cb,hb,+p[869]);else gd=ib;Z=(k[Ib>>2]|0)+(Y<<3)|0;gb=+p[Z>>3];p[t>>3]=gb;$=k[t+4>>2]|0;aa=$&2147483647;ba=$&2146435072;db=rc*(bb*gd);p[Z>>3]=(aa>>>0<2146435072|(aa|0)==2146435072&(k[t>>2]|0)>>>0<1)&(ba>>>0<2146435072|(ba|0)==2146435072&0<0)?gb+db:db;ba=(k[W>>2]|0)+(Y<<3)|0;p[ba>>3]=rc*gd+ +p[ba>>3];if((k[25409]|0)<=2){hd=gd;break}k[s>>2]=Jb;k[s+4>>2]=R;k[s+8>>2]=Ka;k[s+12>>2]=oc;Rw(589407,s)|0;db=+p[Z>>3];gb=+p[ba>>3];p[q>>3]=gd;p[q+8>>3]=db;p[q+16>>3]=gb;Rw(589443,q)|0;zw(Za)|0;hd=gd}else hd=ib;while(0);R=R+1|0;if((R|0)==(_|0)){fd=hd;break}else ib=hd}}Jb=Jb+1|0;if((Jb|0)==(Na|0)){ed=fd;break}else $a=fd}}else ed=pc;while(0);oc=oc+1|0;if((oc|0)==(dc|0)){fc=jc;hc=ed;ic=rc;break}else{pc=ed;qc=rc}}}while(0);S=Ka+1|0;if((S|0)>=((k[4]|0)-mb|0))break d;Ta=k[3]|0;La=fc;Ka=S;Rb=hc;jb=ic}if((ub|0)==202)Lc(k[Ua>>2]|0);else if((ub|0)==205)Lc(k[Ua>>2]|0)}while(0);if((k[25409]|0)>0){Ra(103728)|0;p[o>>3]=+((k[25932]|0)-(k[25933]|0)|0);Rw(589483,o)|0;zw(Za)|0}if(Fh(k[2]|0,Ua)|0)Lc(k[Ua>>2]|0);if(Tb&(Ob|0)>0){Ob=(Nb|0)>(Pb|0)?Pb:Nb;Nb=(Qb|0)>(Mb|0)?Mb:Qb;ic=0.0;hc=0.0;rc=0.0;qc=0.0;Qb=0;Mb=0;Pb=99999;Tb=0;o=0;fc=99999;while(1){mb=k[H+(Tb<<2)>>2]|0;jc=Ub+(Tb<<2)|0;ed=ic;pc=hc;fd=rc;hd=qc;dc=Qb;oc=0;q=Mb;s=Pb;Ha=o;Ga=fc;while(1){Ja=mb+(oc<<3)|0;gd=+p[Ja>>3];if(gd>0.0){u=(k[jc>>2]|0)+(oc<<3)|0;dd=+p[u>>3]/gd;p[u>>3]=dd;u=(dc|0)==0;gd=+p[Ja>>3];cd=u?dd:hd;$c=u?dd:fd;Zc=u?gd:pc;Xc=u?gd:ed;id=gd>Xc?gd:Xc;jd=gd<Zc?gd:Zc;kd=dd>$c?dd:$c;ld=dd<cd?dd:cd;md=u?1:dc;nd=(oc|0)>(q|0)?oc:q;od=(oc|0)<(s|0)?oc:s;pd=(Tb|0)>(Ha|0)?Tb:Ha;qd=(Tb|0)<(Ga|0)?Tb:Ga}else{p[(k[jc>>2]|0)+(oc<<3)>>3]=+p[Ya>>3];p[Ja>>3]=0.0;id=ed;jd=pc;kd=fd;ld=hd;md=dc;nd=q;od=s;pd=Ha;qd=Ga}oc=oc+1|0;if((oc|0)==(Ob|0)){rd=id;sd=jd;td=kd;ud=ld;vd=md;wd=nd;xd=od;yd=pd;zd=qd;break}else{ed=id;pc=jd;fd=kd;hd=ld;dc=md;q=nd;s=od;Ha=pd;Ga=qd}}Ga=Tb+1|0;if((Ga|0)==(Nb|0)){Cd=rd;Dd=sd;Ed=td;Fd=ud;Gd=wd;Hd=xd;Id=yd;Jd=zd;break}else{ic=rd;hc=sd;rc=td;qc=ud;Qb=vd;Mb=wd;Pb=xd;Tb=Ga;o=yd;fc=zd}}}else{Cd=0.0;Dd=0.0;Ed=0.0;Fd=0.0;Gd=0;Hd=99999;Id=0;Jd=99999}zd=Hd+Kb|0;fc=Gd+Kb|0;Kb=Jd+Ia|0;Jd=Id+Ia|0;if((k[25409]|0)>0){p[n>>3]=Fd;Rw(589525,n)|0;p[m>>3]=Ed;Rw(589541,m)|0;p[l>>3]=Dd;Rw(589557,l)|0;p[j>>3]=Cd;Rw(589573,j)|0;k[h>>2]=zd;Rw(589590,h)|0;k[g>>2]=fc;Rw(589605,g)|0;k[f>>2]=Kb;Rw(589620,f)|0;k[e>>2]=Jd;Rw(589635,e)|0}if((Jd|0)<(Kb|0)|(fc|0)<(zd|0))Jc(589650);Uw(346254)|0;Uw(346510)|0;if(Lh(56,346254,Ua)|0)Lc(k[Ua>>2]|0);if(Lh(6960,346510,Ua)|0)Lc(k[Ua>>2]|0);if(Xr(k[14]|0,-64,2,60,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)>0){Tw(589726)|0;zw(Za)|0}if(Xr(k[1740]|0,-64,2,6964,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)>0){Tw(589771)|0;zw(Za)|0}if(cs(k[14]|0,Va,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)>0){Tw(589815)|0;zw(Za)|0}if(cs(k[1740]|0,Va,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)>0){Tw(589860)|0;zw(Za)|0}if(jq(k[14]|0,779656,-64,-1,0,Ua)|0)Lc(k[Ua>>2]|0);if(jq(k[1740]|0,779656,-64,-1,0,Ua)|0)Lc(k[Ua>>2]|0);if(jq(k[14]|0,782713,2,0,0,Ua)|0)Lc(k[Ua>>2]|0);Va=fc+1-zd|0;fc=((Va|0)<0)<<31>>31;if(jq(k[14]|0,772833,Va,fc,0,Ua)|0)Lc(k[Ua>>2]|0);e=1-Kb+Jd|0;f=((e|0)<0)<<31>>31;if(jq(k[14]|0,790243,e,f,0,Ua)|0)Lc(k[Ua>>2]|0);Cd=+(zd|0);if(lq(k[14]|0,777876,+p[875]-Cd,-14,0,Ua)|0)Lc(k[Ua>>2]|0);Dd=+(Kb|0);if(lq(k[14]|0,777883,+p[876]-Dd,-14,0,Ua)|0)Lc(k[Ua>>2]|0);if(jq(k[1740]|0,782713,2,0,0,Ua)|0)Lc(k[Ua>>2]|0);if(jq(k[1740]|0,772833,Va,fc,0,Ua)|0)Lc(k[Ua>>2]|0);if(jq(k[1740]|0,790243,e,f,0,Ua)|0)Lc(k[Ua>>2]|0);if(lq(k[1740]|0,777876,+p[875]-Cd,-14,0,Ua)|0)Lc(k[Ua>>2]|0);if(lq(k[1740]|0,777883,+p[876]-Dd,-14,0,Ua)|0)Lc(k[Ua>>2]|0);if(k[25409]|0){Tw(589906)|0;zw(Za)|0}k[Fa>>2]=1;k[F>>2]=1;f=Kb;while(1){if(Fq(k[14]|0,82,Fa,Va,fc,(k[Ub+(f-Ia<<2)>>2]|0)+(Hd<<3)|0,Ua)|0){ub=345;break}k[F>>2]=(k[F>>2]|0)+1;if((f|0)<(Jd|0))f=f+1|0;else break}if((ub|0)==345)Lc(k[Ua>>2]|0);dy(k[Ub>>2]|0);if((k[25409]|0)>0){Tw(589957)|0;zw(Za)|0}k[Fa>>2]=1;k[F>>2]=1;Ub=Kb;while(1){if(Fq(k[1740]|0,82,Fa,Va,fc,(k[H+(Ub-Ia<<2)>>2]|0)+(Hd<<3)|0,Ua)|0){ub=351;break}k[F>>2]=(k[F>>2]|0)+1;if((Ub|0)<(Jd|0))Ub=Ub+1|0;else break}if((ub|0)==351)Lc(k[Ua>>2]|0);dy(k[H>>2]|0);if((k[25409]|0)>0){Tw(589989)|0;zw(Za)|0}if(Fh(k[14]|0,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)>0){Tw(590022)|0;zw(Za)|0}if(Fh(k[1740]|0,Ua)|0)Lc(k[Ua>>2]|0);if((k[25409]|0)<=0){Ra(103728)|0;Kd=k[25412]|0;Ld=k[25932]|0;Md=k[25933]|0;Nd=Ld-Md|0;Od=+(Nd|0);p[d>>3]=Od;Ew(Kd,590075,d)|0;zw(Za)|0;Pd=k[25412]|0;zw(Pd)|0;Qd=k[25412]|0;ww(Qd)|0;gc(0)}Tw(590048)|0;zw(Za)|0;Ra(103728)|0;Kd=k[25412]|0;Ld=k[25932]|0;Md=k[25933]|0;Nd=Ld-Md|0;Od=+(Nd|0);p[d>>3]=Od;Ew(Kd,590075,d)|0;zw(Za)|0;Pd=k[25412]|0;zw(Pd)|0;Qd=k[25412]|0;ww(Qd)|0;gc(0);break}}return 0}function Gc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;b=r;r=r+8224|0;c=b+8|0;d=b;e=b+24|0;k[25419]=0;ux(e,a)|0;if((k[25409]|0)>2){k[d>>2]=e;Rw(590106,d)|0;zw(k[31231]|0)|0}d=e+(xx(e)|0)|0;a=e;while(1){f=a>>>0<d>>>0;if(f&(i[a>>0]|0)==32)a=a+1|0;else break}if(!f){g=k[25419]|0;r=b;return g|0}f=k[31231]|0;e=a;h=a;while(1){a=e;a:while(1){switch(i[a>>0]|0){case 44:case 32:{j=a;break a;break}default:{}}if(a>>>0>=d>>>0){j=a;break}a=a+1|0}i[j>>0]=0;a=ex(h)|0;l=k[25419]|0;k[101680+(l<<3)>>2]=a;a=j;while(1){m=a+1|0;n=i[m>>0]|0;o=m>>>0<d>>>0;if(n<<24>>24==32&o)a=m;else break}if(o){p=n;q=a;s=m}else{g=l;t=21;break}b:while(1){switch(p<<24>>24){case 44:case 32:{u=q;v=s;break b;break}default:{}}if(s>>>0>=d>>>0){u=q;v=s;break}w=s+1|0;x=s;p=i[w>>0]|0;s=w;q=x}i[v>>0]=0;l=ex(m)|0;a=k[25419]|0;k[101680+(a<<3)+4>>2]=l;if(!(k[25409]|0))y=a;else{x=k[101680+(a<<3)>>2]|0;k[c>>2]=a;k[c+4>>2]=x;k[c+8>>2]=l;Rw(590128,c)|0;zw(f)|0;y=k[25419]|0}l=y+1|0;k[25419]=l;h=u+2|0;if(h>>>0>=d>>>0){g=l;t=21;break}else e=v}if((t|0)==21){r=b;return g|0}return 0}function Hc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0.0,l=0.0,m=0,n=0,o=0.0,q=0;c=r;r=r+304|0;d=c+16|0;e=c+8|0;f=c;g=c+32|0;h=c+40|0;i=c+24|0;k[g>>2]=0;if(Ih(5,8,a,0,g)|0){k[f>>2]=a;Xw(h,590158,f)|0;Jc(h)}f=k[25410]|0;if((f|0)>0?(Fk(k[2]|0,f+1|0,0,g)|0)!=0:0)Lc(k[g>>2]|0);if(Ns(k[2]|0,101652,g)|0)Lc(k[g>>2]|0);if(k[25411]|0){if(Ih(5,6912,b,0,g)|0){k[e>>2]=b;Xw(h,590196,e)|0;Jc(h)}h=k[25410]|0;if((h|0)>0?(Fk(k[1728]|0,h+1|0,0,g)|0)!=0:0)Lc(k[g>>2]|0)}if((k[25409]|0)>2){k[d>>2]=k[25413];Rw(590235,d)|0;zw(k[31231]|0)|0}d=md(k[25413]|0)|0;k[5]=d;if(!d){Pw(590278,53,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}h=d+136|0;j=+p[h>>3]+ +p[6]*2.0;p[h>>3]=j;h=d+144|0;l=+p[6]*2.0+ +p[h>>3];p[h>>3]=l;h=d+16|0;p[h>>3]=+p[6]+ +p[h>>3];h=d+24|0;p[h>>3]=+p[6]+ +p[h>>3];k[3]=~~j;k[4]=~~l;if(!(Gl(k[2]|0,2,i,g)|0)){k[3]=k[i>>2];k[4]=k[i+4>>2]}else k[g>>2]=0;k[10]=0;g=k[5]|0;l=+p[g+32>>3];if(l<0.0?+p[g+40>>3]<0.0:0)m=25;else if(l>0.0?+p[g+40>>3]>0.0:0)m=25;else n=0;if((m|0)==25){k[10]=1;n=1}do if((k[25409]|0)>2)if(!n){Tw(590360)|0;break}else{Tw(590332)|0;break}while(0);n=k[5]|0;switch(k[n+3884>>2]|0){case 1:{if(!(+p[n+120>>3]==1950.0)){o=2.0e3;q=0;k[6]=q;p[4]=o;r=c;return 0}o=1950.0;q=0;k[6]=q;p[4]=o;r=c;return 0}case 2:{if(!(+p[n+120>>3]==2.0e3)){o=1950.0;q=1;k[6]=q;p[4]=o;r=c;return 0}o=2.0e3;q=1;k[6]=q;p[4]=o;r=c;return 0}case 4:{if(!(+p[n+120>>3]==1950.0)){o=2.0e3;q=2;k[6]=q;p[4]=o;r=c;return 0}o=1950.0;q=3;k[6]=q;p[4]=o;r=c;return 0}case 3:{o=2.0e3;q=4;k[6]=q;p[4]=o;r=c;return 0}default:{o=2.0e3;q=0;k[6]=q;p[4]=o;r=c;return 0}}return 0}function Ic(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0.0,s=0,t=0,u=0,v=0.0,w=0,x=0;c=r;r=r+80304|0;d=c+40|0;e=c+32|0;f=c+24|0;g=c+16|0;h=c+8|0;j=c;l=c+80048|0;m=c+48|0;if((k[25409]|0)>2){k[j>>2]=a;Rw(590395,j)|0;zw(k[31231]|0)|0}j=Dw(a,774853)|0;if(!j){k[h>>2]=a;Xw(l,590423,h)|0;Jc(l)}i[m>>0]=0;h=k[31231]|0;a=0;do{if(!(Bw(l,256,j)|0))break;n=l+((xx(l)|0)+-1)|0;if((i[n>>0]|0)==10)i[n>>0]=0;n=l+((xx(l)|0)+-1)|0;if((i[n>>0]|0)==13)i[n>>0]=0;if((k[25409]|0)>2){k[g>>2]=l;Rw(590453,g)|0;zw(h)|0}Mc(l,b)|0;n=xx(m)|0;o=xx(l)|0;if(!((o|0)>0?(py(m+n|0,l|0,o|0)|0,(o|0)>=80):0))sy(m+(o+n)|0,32,80-o|0)|0;i[m+(n+80)>>0]=0;a=a+1|0}while((a|0)<1e3);a=(k[25409]|0)>2;if((b|0)==1){if(a){k[f>>2]=m;Rw(590474,f)|0;zw(h)|0}ux(346766,m)|0;f=md(m)|0;k[5]=f;if(!f){Pw(594122,54,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}switch(k[f+3884>>2]|0){case 1:{if(+p[f+120>>3]==1950.0){q=1950.0;s=0}else{q=2.0e3;s=0}break}case 2:{if(+p[f+120>>3]==2.0e3){q=2.0e3;s=1}else{q=1950.0;s=1}break}case 3:{q=2.0e3;s=4;break}case 4:{if(+p[f+120>>3]==1950.0){q=1950.0;s=3}else{q=2.0e3;s=2}break}default:{q=2.0e3;s=0}}k[6]=s;p[4]=q;k[10]=0;q=+p[f+32>>3];if(q<0.0?+p[f+40>>3]<0.0:0)t=34;else if(q>0.0?+p[f+40>>3]>0.0:0)t=34;else u=1;if((t|0)==34){k[10]=1;u=0}k[10]=u&1;if((k[25409]|0)<=2){r=c;return 0}if(u){Tw(590332)|0;r=c;return 0}else{Tw(590360)|0;r=c;return 0}}u=(b|0)==2;do if(a)if(u){k[e>>2]=m;Rw(590527,e)|0;zw(h)|0;t=44;break}else{k[d>>2]=m;Rw(590582,d)|0;zw(h)|0;t=45;break}else if(u)t=44;else t=45;while(0);if((t|0)==44)ux(426766,m)|0;else if((t|0)==45)ux(506766,m)|0;u=md(m)|0;k[17]=u;if(!u){Pw(594122,54,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}k[1743]=u;m=u+32|0;h=u+40|0;q=+W(+(+p[m>>3]*+p[h>>3]))*.017453292519943295*.017453292519943295;p[869]=q;switch(k[u+3884>>2]|0){case 1:{if(+p[u+120>>3]==1950.0){v=1950.0;w=0}else{v=2.0e3;w=0}break}case 2:{if(+p[u+120>>3]==2.0e3){v=2.0e3;w=1}else{v=1950.0;w=1}break}case 3:{v=2.0e3;w=4;break}case 4:{if(+p[u+120>>3]==1950.0){v=1950.0;w=3}else{v=2.0e3;w=2}break}default:{v=2.0e3;w=0}}k[18]=w;p[10]=v;k[1744]=w;p[873]=v;k[22]=0;v=+p[m>>3];if(v<0.0?+p[h>>3]<0.0:0)t=60;else if(v>0.0?+p[h>>3]>0.0:0)t=60;else x=0;if((t|0)==60){k[22]=1;x=1}if((k[25409]|0)<=2){r=c;return 0}if(!x){Tw(590665)|0;r=c;return 0}else{Tw(590636)|0;r=c;return 0}return 0}function Jc(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;b=k[25412]|0;k[c>>2]=a;Ew(b,597611,c)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}function Kc(a,b,c,d,e,f){a=+a;b=+b;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0.0,m=0.0,n=0;g=r;r=r+64|0;h=g+16|0;i=g+8|0;j=g;$c(a,b,i,j,96)|0;l=+p[i>>3];if(!(l<-.5)?(m=+p[j>>3],!(m<-.5|l>+(k[54]|0)+1.5)):0)n=m>+(k[56]|0)+1.5;else n=1;if((k[25409]|0)>2){m=+p[j>>3];p[h>>3]=a;p[h+8>>3]=b;p[h+16>>3]=l;p[h+24>>3]=m;k[h+32>>2]=n&1;Rw(590701,h)|0;zw(k[31231]|0)|0}if(n){r=g;return}if(+p[c>>3]>a)p[c>>3]=a;if(+p[d>>3]<a)p[d>>3]=a;if(+p[e>>3]>b)p[e>>3]=b;if(!(+p[f>>3]<b)){r=g;return}p[f>>3]=b;r=g;return}function Lc(a){a=a|0;var b=0,c=0,d=0;b=r;r=r+48|0;c=b;d=b+8|0;sk(a,d);b=k[25412]|0;k[c>>2]=a;k[c+4>>2]=d;Ew(b,598376,c)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}function Mc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0.0,w=0.0;c=r;r=r+304|0;d=c+32|0;e=c+24|0;f=c+16|0;g=c+8|0;h=c;j=c+40|0;ux(j,a)|0;l=j+(xx(j)|0)|0;m=j;while(1){j=i[m>>0]|0;if(m>>>0<l>>>0&j<<24>>24==32)m=m+1|0;else{n=j;o=m;break}}m=n;n=o;a:while(1){switch(m<<24>>24){case 61:case 32:{q=m;s=n;break a;break}default:{}}if(n>>>0>=l>>>0){q=m;s=n;break}j=n+1|0;m=i[j>>0]|0;n=j}n=q;q=s;b:while(1){switch(n<<24>>24){case 39:case 32:case 61:break;default:{t=q;break b}}if(q>>>0>=l>>>0){t=q;break}m=q+1|0;n=i[m>>0]|0;q=m}i[s>>0]=0;s=(i[t>>0]|0)==39?t+1|0:t;c:while(1){switch(i[s>>0]|0){case 39:case 32:{u=s;break c;break}default:{}}if(s>>>0>=l>>>0){u=s;break}s=s+1|0}i[u>>0]=0;if((k[25409]|0)>1){k[h>>2]=o;k[h+4>>2]=t;Rw(590734,h)|0;zw(k[31231]|0)|0}if(b&-3){r=c;return 0}if(!(tx(o,772833)|0)){v=+(ex(t)|0);w=+p[6]*2.0;b=~~(v+w);k[15]=b;h=~~(+(ex(t)|0)+w);k[1741]=h;k[g>>2]=b;Xw(a,590761,g)|0}if(!(tx(o,790243)|0)){w=+(ex(t)|0);v=+p[6]*2.0;g=~~(w+v);k[16]=g;b=~~(+(ex(t)|0)+v);k[1742]=b;k[f>>2]=g;Xw(a,590775,f)|0}if(!(tx(o,777876)|0)){v=+dx(t);w=v+ +p[6];p[875]=w;p[e>>3]=w;Xw(a,590789,e)|0}if(tx(o,777883)|0){r=c;return 0}w=+dx(t);v=w+ +p[6];p[876]=v;p[d>>3]=v;Xw(a,590806,d)|0;r=c;return 0}function Nc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,X=0,$=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0.0,ra=0,sa=0.0,ta=0,ua=0,va=0.0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0.0,Ea=0,Fa=0,Ga=0,Ha=0.0,Ia=0.0,Ja=0.0,Ka=0.0,La=0.0,Ma=0.0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0.0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0.0,Za=0,_a=0,$a=0.0,ab=0.0,bb=0.0,cb=0.0,db=0.0,eb=0,fb=0.0,gb=0.0,hb=0.0,ib=0.0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0.0,sb=0,tb=0,ub=0,vb=0,wb=0.0,xb=0,yb=0,zb=0.0,Ab=0.0,Bb=0,Cb=0.0,Db=0.0,Eb=0.0,Fb=0.0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0.0,Lb=0.0,Mb=0,Nb=0.0,Ob=0.0,Pb=0.0,Qb=0.0,Rb=0.0,Sb=0.0,Tb=0,Ub=0.0,Vb=0.0,Wb=0.0,Xb=0.0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,hc=0,ic=0,jc=0,kc=0,lc=0;c=r;r=r+1504|0;d=c+928|0;e=c+920|0;f=c+912|0;g=c+856|0;h=c+848|0;i=c+840|0;j=c+784|0;l=c+768|0;m=c+752|0;n=c+696|0;o=c+688|0;q=c+664|0;s=c+640|0;t=c+632|0;u=c+624|0;v=c+616|0;w=c+608|0;x=c+544|0;y=c+536|0;z=c+472|0;A=c+464|0;B=c+448|0;C=c+432|0;D=c+376|0;E=c+368|0;F=c+344|0;G=c+320|0;H=c+312|0;I=c+304|0;J=c+296|0;K=c+288|0;L=c+224|0;M=c+216|0;N=c+152|0;O=c+144|0;P=c+136|0;Q=c+128|0;R=c+120|0;S=c+112|0;T=c+104|0;U=c+96|0;V=c+88|0;X=c+80|0;$=c+72|0;ba=c+64|0;ca=c+1240|0;da=c+984|0;ea=c+980|0;fa=c+56|0;ga=c+48|0;ia=c+40|0;ja=c+32|0;ka=c+24|0;la=c+16|0;ma=c+8|0;na=c;oa=c+976|0;p[877]=.017453292519943295;k[31173]=0;k[25409]=0;k[25934]=5;c=k[31231]|0;k[25412]=c;k[31172]=1;pa=c;c=50;qa=.01;ra=0;a:while(1)switch(Pv(a,b,591863)|0){case 117:{sa=qa;ta=c;ra=1;qa=sa;c=ta;continue a;break}case -1:{ua=c;va=qa;wa=ra;xa=18;break a;break}case 100:{k[25409]=2;ya=ra;sa=qa;ta=c;ra=ya;qa=sa;c=ta;continue a;break}case 105:{za=jx(k[31177]|0,oa,0)|0;Aa=k[31177]|0;if((k[oa>>2]|0)>>>0<(Aa+(xx(Aa)|0)|0)>>>0){Ba=Aa;xa=5;break a}ya=ra;sa=qa;c=(za|0)<1?1:za;ra=ya;qa=sa;continue a;break}case 111:{za=jx(k[31177]|0,oa,0)|0;k[25934]=za;Aa=k[31177]|0;if((k[oa>>2]|0)>>>0<(Aa+(xx(Aa)|0)|0)>>>0){Ca=Aa;xa=8;break a}if((za|0)>=0){ya=ra;sa=qa;ta=c;ra=ya;qa=sa;c=ta;continue a}k[25934]=0;ya=ra;sa=qa;ta=c;ra=ya;qa=sa;c=ta;continue a;break}case 116:{Da=+hx(k[31177]|0,oa);za=k[31177]|0;if((k[oa>>2]|0)>>>0<(za+(xx(za)|0)|0)>>>0){Ea=za;xa=12;break a}if(!(Da<0.0)){ya=ra;ta=c;qa=Da;ra=ya;c=ta;continue a}ya=ra;ta=c;qa=0.0;ra=ya;c=ta;continue a;break}case 115:{za=Dw(k[31177]|0,597225)|0;k[25412]=za;if(!za){xa=16;break a}else{ya=ra;sa=qa;ta=c;ra=ya;qa=sa;c=ta;continue a}break}default:{xa=17;break a}}if((xa|0)==5){ta=k[25412]|0;k[ba>>2]=Ba;Ew(ta,591874,ba)|0;zw(pa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else if((xa|0)==8){ba=k[25412]|0;k[$>>2]=Ca;Ew(ba,591949,$)|0;zw(pa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else if((xa|0)==12){$=k[25412]|0;k[X>>2]=Ea;Ew($,592019,X)|0;zw(pa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else if((xa|0)==16){k[V>>2]=k[31177];Ew(0,592096,V)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else if((xa|0)==17){V=k[25412]|0;k[U>>2]=k[b>>2];Ew(V,592154,U)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else if((xa|0)==18){U=k[31172]|0;if((a-U|0)<2){a=k[25412]|0;k[T>>2]=k[b>>2];Ew(a,592154,T)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}ux(ca,k[b+(U<<2)>>2]|0)|0;ux(da,k[b+((k[31172]|0)+1<<2)>>2]|0)|0;Wc(ca,1,0)|0;Oc(ca)|0;b=Dw(da,597225)|0;k[25935]=b;if(!b){b=k[25412]|0;k[S>>2]=da;Ew(b,592320,S)|0;zw(pa)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}if(k[25409]|0){Tw(592386)|0;Tw(592412)|0;Tw(592471)|0;zw(pa)|0}S=k[25934]|0;b=(S<<2)+4|0;da=cy(b)|0;k[25936]=da;U=cy(b)|0;k[25937]=U;U=(S|0)>0;T=S+1|0;b:do if(U){a=T<<3;V=(S<<3)+8|0;X=da;$=0;while(1){Ea=cy(a)|0;k[X+($<<2)>>2]=Ea;Ea=cy(V)|0;k[(k[25937]|0)+($<<2)>>2]=Ea;Ea=k[(k[25936]|0)+($<<2)>>2]|0;ba=k[(k[25937]|0)+($<<2)>>2]|0;Ca=0;do{p[Ea+(Ca<<3)>>3]=0.0;p[ba+(Ca<<3)>>3]=0.0;Ca=Ca+1|0}while((Ca|0)<(S|0));Ca=$+1|0;if((Ca|0)>=(S|0))break b;X=k[25936]|0;$=Ca}}while(0);da=cy(T<<2)|0;k[25938]=da;T=cy(b)|0;k[25939]=T;c:do if(U){T=(S<<3)+8|0;b=da;$=0;while(1){X=cy(T)|0;k[b+($<<2)>>2]=X;X=cy(T)|0;k[(k[25939]|0)+($<<2)>>2]=X;X=k[(k[25938]|0)+($<<2)>>2]|0;V=k[(k[25939]|0)+($<<2)>>2]|0;a=0;do{p[X+(a<<3)>>3]=0.0;p[V+(a<<3)>>3]=0.0;a=a+1|0}while((a|0)<(S|0));a=$+1|0;if((a|0)>=(S|0))break c;b=k[25938]|0;$=a}}while(0);if(!(k[25409]|0))Fa=S;else{Tw(592530)|0;zw(pa)|0;Fa=k[25934]|0}pa=ha(Fa,Fa)|0;S=pa<<1;da=S|1;U=da<<2;$=cy(U)|0;b=cy(U)|0;U=(pa|0)==0;if(!U){T=da<<3;da=0;do{a=cy(16)|0;k[$+(da<<2)>>2]=a;a=cy(T)|0;k[b+(da<<2)>>2]=a;da=da+1|0}while((da|0)<(S|0))}Pc(ca)|0;ca=k[25953]|0;if((k[ca+3260>>2]|0)==29){k[25940]=1;p[R>>3]=+p[ca+760>>3];Xw(590823,592572,R)|0;k[25941]=1;p[Q>>3]=+p[(k[25953]|0)+768>>3];Xw(590903,592572,Q)|0;k[25940]=1;p[P>>3]=+p[(k[25953]|0)+48>>3];Xw(590983,592572,P)|0;Ga=k[25953]|0}else Ga=ca;sa=(+p[Ga+136>>3]+1.0)*.5;p[fa>>3]=sa;qa=(+p[Ga+144>>3]+1.0)*.5;p[ga>>3]=qa;Ad(Ga,sa,qa,ka,la);qa=+dx(590823);Ad(k[25953]|0,+p[fa>>3]+1.0,+p[ga>>3],ma,na);sa=+p[877];Da=+p[ka>>3]*sa;Ha=+p[la>>3]*sa;Ia=+Z(+Ha);Ja=+p[ma>>3]*sa;Ka=+p[na>>3]*sa;La=+Z(+Ka);Ma=+aa(+(+_(+Ha)*+_(+Ka)+(+Z(+Da)*Ia*(+Z(+Ja)*La)+Ia*+_(+Da)*(La*+_(+Ja)))))/sa;p[878]=qa<0.0?-Ma:Ma;Ma=+dx(590903);Ad(k[25953]|0,+p[fa>>3],+p[ga>>3]+1.0,ma,na);qa=+p[877];sa=+p[ka>>3]*qa;Ja=+p[la>>3]*qa;La=+Z(+Ja);Da=+p[ma>>3]*qa;Ia=+p[na>>3]*qa;Ka=+Z(+Ia);Ha=+aa(+(+_(+Ja)*+_(+Ia)+(+Z(+sa)*La*(+Z(+Da)*Ka)+La*+_(+sa)*(Ka*+_(+Da)))))/qa;p[879]=Ma<0.0?-Ha:Ha;na=k[25953]|0;ma=~~+p[na+136>>3];Ha=+p[na+144>>3];na=~~(+((Ha<+(ma|0)?~~Ha:ma)|0)*.125);ma=~~(+(na|0)/20.0);la=(na|0)<1?1:na;na=(ma|0)<1?1:ma;Rc()|0;ma=(ua|0)>0;d:do if(ma){ka=k[31231]|0;Ga=pa+-1|0;ca=b+(Ga<<2)|0;P=(wa|0)==0;Q=(pa<<1|0)>1?pa<<4:8;R=0;while(1){if(!U){da=0;do{p[k[$+(da<<2)>>2]>>3]=0.0;sy(k[b+(da<<2)>>2]|0,0,Q|0)|0;da=da+1|0}while((da|0)<(S|0))}if((k[25409]|0)>1){k[O>>2]=R;Rw(592580,O)|0;zw(ka)|0}da=k[25953]|0;if(+p[da+136>>3]+1.0>0.0){T=da;Ha=0.0;a=0;while(1){if(+p[T+144>>3]+1.0>0.0){Ma=Ha+-.5;qa=0.0;V=T;X=0;while(1){Da=qa+-.5;Ad(V,Ma,Da,ia,ja);k[ea>>2]=0;Bd(k[25954]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);Ka=+p[fa>>3];sa=Ma-Ka;if((k[25409]|0)>1){La=+p[ia>>3];Ia=+p[ja>>3];Ja=+p[ga>>3];Ca=k[ea>>2]|0;p[N>>3]=Ma;p[N+8>>3]=Da;p[N+16>>3]=La;p[N+24>>3]=Ia;p[N+32>>3]=Ka;p[N+40>>3]=Ja;p[N+48>>3]=sa;k[N+56>>2]=Ca;Rw(592627,N)|0;zw(ka)|0}do if(!(P&(k[ea>>2]|0)!=0)){Ca=k[25934]|0;if((Ca|0)>0)Na=0;else break;do{ba=ha(Na,Fa)|0;Ja=+Y(+Ma,+(+(Na|0)));Ea=0;do{ta=Ea+ba|0;Ka=sa*(Ja*+Y(+Da,+(+(Ea|0))));Ba=k[$+(ta<<2)>>2]|0;p[Ba>>3]=+p[Ba>>3]+Ka;Ba=k[b+(ta<<2)>>2]|0;ta=0;do{c=ha(ta,Fa)|0;Ka=+Y(+Ma,+(+(ta+Na|0)));ya=0;do{Ia=Ka*+Y(+Da,+(+(ya+Ea|0)));ra=Ba+(ya+c<<3)|0;p[ra>>3]=Ia+ +p[ra>>3];ya=ya+1|0}while((ya|0)!=(Ca|0));ta=ta+1|0}while((ta|0)!=(Ca|0));Ea=Ea+1|0}while((Ea|0)!=(Ca|0));Na=Na+1|0}while((Na|0)<(Ca|0))}while(0);X=X+la|0;qa=+(X|0);Ca=k[25953]|0;if(!(qa<+p[Ca+144>>3]+1.0)){Oa=Ca;break}else V=Ca}}else Oa=T;a=a+la|0;Ha=+(a|0);if(!(Ha<+p[Oa+136>>3]+1.0)){Pa=Oa;break}else T=Oa}}else Pa=da;if((k[25409]|0)>1){k[M>>2]=R;Rw(592698,M)|0;zw(ka)|0;Qa=k[25953]|0}else Qa=Pa;if(+p[Qa+136>>3]+1.0>0.0){T=Qa;Ha=0.0;a=0;while(1){if(+p[T+144>>3]+1.0>0.0){qa=Ha+-.5;Ma=0.0;V=T;X=0;while(1){Da=Ma+-.5;Ad(V,qa,Da,ia,ja);k[ea>>2]=0;Bd(k[25954]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);sa=+p[ga>>3];Ja=Da-sa;if((k[25409]|0)>1){Ka=+p[ia>>3];Ia=+p[ja>>3];La=+p[fa>>3];Ca=k[ea>>2]|0;p[L>>3]=qa;p[L+8>>3]=Da;p[L+16>>3]=Ka;p[L+24>>3]=Ia;p[L+32>>3]=La;p[L+40>>3]=sa;p[L+48>>3]=Ja;k[L+56>>2]=Ca;Rw(592627,L)|0;zw(ka)|0}do if(!(P&(k[ea>>2]|0)!=0)){Ca=k[25934]|0;if((Ca|0)>0)Ra=0;else break;do{sa=+(Ra|0);Ea=(ha(Ra,Fa)|0)+pa|0;ba=0;do{La=+Y(+(+p[fa>>3]),+sa);Ia=Ja*(La*+Y(+(+p[ga>>3]),+(+(ba|0))));ta=Ea+ba|0;Ba=k[$+(ta<<2)>>2]|0;p[Ba>>3]=+p[Ba>>3]+Ia;Ba=k[b+(ta<<2)>>2]|0;ta=0;do{Ia=+(ta+Ra|0);ya=(ha(ta,Fa)|0)+pa|0;c=0;do{La=+Y(+(+p[fa>>3]),+Ia);Ka=La*+Y(+(+p[ga>>3]),+(+(c+ba|0)));ra=Ba+(ya+c<<3)|0;p[ra>>3]=Ka+ +p[ra>>3];c=c+1|0}while((c|0)!=(Ca|0));ta=ta+1|0}while((ta|0)!=(Ca|0));ba=ba+1|0}while((ba|0)!=(Ca|0));Ra=Ra+1|0}while((Ra|0)<(Ca|0))}while(0);X=X+la|0;Ma=+(X|0);Ca=k[25953]|0;if(!(Ma<+p[Ca+144>>3]+1.0)){Sa=Ca;break}else V=Ca}}else Sa=T;a=a+la|0;Ha=+(a|0);if(!(Ha<+p[Sa+136>>3]+1.0))break;else T=Sa}}if(+p[(k[ca>>2]|0)+(Ga<<3)>>3]==0.0){xa=85;break}if(k[25409]|0){Tw(592816)|0;if(!U){T=0;do{Tw(592840)|0;a=k[b+(T<<2)>>2]|0;da=0;do{p[K>>3]=+p[a+(da<<3)>>3];Rw(592848,K)|0;da=da+1|0}while((da|0)<(pa|0));Tw(592856)|0;p[J>>3]=+p[k[$+(T<<2)>>2]>>3];Rw(592865,J)|0;T=T+1|0}while((T|0)<(pa|0))}Sw(10)|0}Qc(b,S,$,1);if(k[25409]|0){Tw(592875)|0;if(!U){T=0;do{Tw(592840)|0;da=k[b+(T<<2)>>2]|0;a=0;do{p[I>>3]=+p[da+(a<<3)>>3];Rw(592848,I)|0;a=a+1|0}while((a|0)<(pa|0));Tw(592856)|0;p[H>>3]=+p[k[$+(T<<2)>>2]>>3];Rw(592865,H)|0;T=T+1|0}while((T|0)<(pa|0))}Sw(10)|0}T=k[25934]|0;if((T|0)>0){a=T;da=0;while(1){if((a|0)>0){V=ha(da,Fa)|0;X=a;Ca=0;while(1){ba=k[$+(Ca+V<<2)>>2]|0;Ea=(k[(k[25937]|0)+(da<<2)>>2]|0)+(Ca<<3)|0;Ha=+p[ba>>3]*.5+ +p[Ea>>3];p[Ea>>3]=Ha;if(Ha!=0.0)Ta=+p[ba>>3]/Ha*100.0;else Ta=-999.0;if(!(k[25409]|0))Ua=X;else{k[G>>2]=da;k[G+4>>2]=Ca;p[G+8>>3]=Ha;p[G+16>>3]=Ta;Rw(592893,G)|0;zw(ka)|0;Ua=k[25934]|0}Ca=Ca+1|0;if((Ca|0)>=(Ua|0)){Va=Ua;break}else X=Ua}}else Va=a;da=da+1|0;if((da|0)>=(Va|0)){Wa=Va;break}else a=Va}}else Wa=T;if(!(k[25409]|0))Xa=Wa;else{Sw(10)|0;Xa=k[25934]|0}if((Xa|0)>0){a=Xa;da=0;while(1){if((a|0)>0){X=(ha(da,Fa)|0)+pa|0;Ca=a;V=0;while(1){ba=k[$+(X+V<<2)>>2]|0;Ea=(k[(k[25939]|0)+(da<<2)>>2]|0)+(V<<3)|0;Ha=+p[ba>>3]*.5+ +p[Ea>>3];p[Ea>>3]=Ha;if(Ha!=0.0)Ya=+p[ba>>3]/Ha*100.0;else Ya=-999.0;if(!(k[25409]|0))Za=Ca;else{k[F>>2]=da;k[F+4>>2]=V;p[F+8>>3]=Ha;p[F+16>>3]=Ya;Rw(592924,F)|0;zw(ka)|0;Za=k[25934]|0}V=V+1|0;if((V|0)>=(Za|0)){_a=Za;break}else Ca=Za}}else _a=a;da=da+1|0;if((da|0)>=(_a|0))break;else a=_a}}Rc()|0;if((k[25409]|0)>1){k[E>>2]=R;Rw(592955,E)|0;zw(ka)|0}a=k[25953]|0;if(+p[a+136>>3]+1.0>0.0){da=a;Ha=0.0;Ma=0.0;qa=0.0;a=0;while(1){if(+p[da+144>>3]+1.0>0.0){Ja=Ha+-.5;sa=0.0;T=da;Ia=Ma;Ka=qa;Ca=0;while(1){La=sa+-.5;Ad(T,Ja,La,ia,ja);k[ea>>2]=0;Bd(k[25954]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);if((k[25409]|0)>1){Da=+p[ia>>3];$a=+p[ja>>3];ab=+p[fa>>3];bb=+p[ga>>3];V=k[ea>>2]|0;p[D>>3]=Ja;p[D+8>>3]=La;p[D+16>>3]=Da;p[D+24>>3]=$a;p[D+32>>3]=ab;p[D+40>>3]=bb;k[D+48>>2]=V;Rw(592999,D)|0;zw(ka)|0}do if(P&(k[ea>>2]|0)!=0){cb=Ia;db=Ka}else{bb=+W(+(Ja-+p[fa>>3]));ab=bb>Ia?bb:Ia;bb=+W(+(La-+p[ga>>3]));if(!(bb>Ka)){cb=ab;db=Ka;break}cb=ab;db=bb}while(0);Ca=Ca+la|0;sa=+(Ca|0);V=k[25953]|0;if(!(sa<+p[V+144>>3]+1.0)){eb=V;fb=cb;gb=db;break}else{T=V;Ia=cb;Ka=db}}}else{eb=da;fb=Ma;gb=qa}a=a+la|0;Ha=+(a|0);if(!(Ha<+p[eb+136>>3]+1.0)){hb=fb;ib=gb;break}else{da=eb;Ma=fb;qa=gb}}}else{hb=0.0;ib=0.0}if(k[25409]|0){p[C>>3]=hb;k[C+8>>2]=R;Rw(593060,C)|0;p[B>>3]=ib;k[B+8>>2]=R;Rw(593080,B)|0;zw(ka)|0}if(hb<va&ib<va){jb=1;kb=R;break}da=R+1|0;if((da|0)<(ua|0))R=da;else{jb=0;kb=da;break}}if((xa|0)==85){Pw(592745,70,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}Rc()|0;if(ma){R=k[31231]|0;ka=pa+-1|0;P=b+(ka<<2)|0;Ga=(wa|0)==0;ca=(pa<<1|0)>1?pa<<4:8;Q=0;while(1){if(!U){da=0;do{p[k[$+(da<<2)>>2]>>3]=0.0;sy(k[b+(da<<2)>>2]|0,0,ca|0)|0;da=da+1|0}while((da|0)<(S|0))}if((k[25409]|0)>1){k[A>>2]=Q;Rw(593100,A)|0;zw(R)|0}da=k[25953]|0;if(+p[da+136>>3]+1.0>0.0){a=da;qa=0.0;T=0;while(1){if(+p[a+144>>3]+1.0>0.0){Ma=qa+-.5;Ha=0.0;Ca=0;do{Ka=Ha+-.5;Ad(k[25954]|0,Ma,Ka,ia,ja);k[ea>>2]=0;Bd(k[25953]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);Ia=+p[fa>>3]-+p[880];p[fa>>3]=Ia;sa=+p[ga>>3]-+p[881];p[ga>>3]=sa;k[ea>>2]=0;Ja=Ma-Ia;if(!((k[25409]|0)>1?(La=+p[ia>>3],bb=+p[ja>>3],p[z>>3]=Ma,p[z+8>>3]=Ka,p[z+16>>3]=La,p[z+24>>3]=bb,p[z+32>>3]=Ia,p[z+40>>3]=sa,p[z+48>>3]=Ja,k[z+56>>2]=0,Rw(593147,z)|0,zw(R)|0,Ga&(k[ea>>2]|0)!=0):0))xa=145;do if((xa|0)==145){xa=0;V=k[25934]|0;if((V|0)>0)lb=0;else break;do{X=ha(lb,Fa)|0;sa=+Y(+Ma,+(+(lb|0)));ba=0;do{Ea=ba+X|0;Ia=Ja*(sa*+Y(+Ka,+(+(ba|0))));ta=k[$+(Ea<<2)>>2]|0;p[ta>>3]=+p[ta>>3]+Ia;ta=k[b+(Ea<<2)>>2]|0;Ea=0;do{Ba=ha(Ea,Fa)|0;Ia=+Y(+Ma,+(+(Ea+lb|0)));c=0;do{bb=Ia*+Y(+Ka,+(+(c+ba|0)));ya=ta+(c+Ba<<3)|0;p[ya>>3]=bb+ +p[ya>>3];c=c+1|0}while((c|0)!=(V|0));Ea=Ea+1|0}while((Ea|0)!=(V|0));ba=ba+1|0}while((ba|0)!=(V|0));lb=lb+1|0}while((lb|0)<(V|0))}while(0);Ca=Ca+la|0;Ha=+(Ca|0);V=k[25953]|0}while(Ha<+p[V+144>>3]+1.0);mb=V}else mb=a;T=T+la|0;qa=+(T|0);if(!(qa<+p[mb+136>>3]+1.0)){nb=mb;break}else a=mb}}else nb=da;if((k[25409]|0)>1){k[y>>2]=Q;Rw(593218,y)|0;zw(R)|0;ob=k[25953]|0}else ob=nb;if(+p[ob+136>>3]+1.0>0.0){a=ob;qa=0.0;T=0;while(1){if(+p[a+144>>3]+1.0>0.0){Ha=qa+-.5;Ma=0.0;Ca=0;do{Ka=Ma+-.5;Ad(k[25954]|0,Ha,Ka,ia,ja);k[ea>>2]=0;Bd(k[25953]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);Ja=+p[fa>>3]-+p[880];p[fa>>3]=Ja;sa=+p[ga>>3]-+p[881];p[ga>>3]=sa;k[ea>>2]=0;Ia=Ka-sa;if(!((k[25409]|0)>1?(bb=+p[ia>>3],La=+p[ja>>3],p[x>>3]=Ha,p[x+8>>3]=Ka,p[x+16>>3]=bb,p[x+24>>3]=La,p[x+32>>3]=Ja,p[x+40>>3]=sa,p[x+48>>3]=Ia,k[x+56>>2]=0,Rw(593147,x)|0,zw(R)|0,Ga&(k[ea>>2]|0)!=0):0))xa=162;do if((xa|0)==162){xa=0;V=k[25934]|0;if((V|0)>0)pb=0;else break;do{sa=+(pb|0);ba=(ha(pb,Fa)|0)+pa|0;X=0;do{Ja=+Y(+(+p[fa>>3]),+sa);La=Ia*(Ja*+Y(+(+p[ga>>3]),+(+(X|0))));Ea=ba+X|0;ta=k[$+(Ea<<2)>>2]|0;p[ta>>3]=+p[ta>>3]+La;ta=k[b+(Ea<<2)>>2]|0;Ea=0;do{La=+(Ea+pb|0);c=(ha(Ea,Fa)|0)+pa|0;Ba=0;do{Ja=+Y(+(+p[fa>>3]),+La);bb=Ja*+Y(+(+p[ga>>3]),+(+(Ba+X|0)));ya=ta+(c+Ba<<3)|0;p[ya>>3]=bb+ +p[ya>>3];Ba=Ba+1|0}while((Ba|0)!=(V|0));Ea=Ea+1|0}while((Ea|0)!=(V|0));X=X+1|0}while((X|0)!=(V|0));pb=pb+1|0}while((pb|0)<(V|0))}while(0);Ca=Ca+la|0;Ma=+(Ca|0);V=k[25953]|0}while(Ma<+p[V+144>>3]+1.0);qb=V}else qb=a;T=T+la|0;qa=+(T|0);if(!(qa<+p[qb+136>>3]+1.0))break;else a=qb}}if(+p[(k[P>>2]|0)+(ka<<3)>>3]==0.0)break;if(k[25409]|0){Tw(593336)|0;if(!U){a=0;do{Tw(592840)|0;T=k[b+(a<<2)>>2]|0;da=0;do{p[w>>3]=+p[T+(da<<3)>>3];Rw(592848,w)|0;da=da+1|0}while((da|0)<(pa|0));Tw(592856)|0;p[v>>3]=+p[k[$+(a<<2)>>2]>>3];Rw(592865,v)|0;a=a+1|0}while((a|0)<(pa|0))}Sw(10)|0}Qc(b,S,$,1);if(k[25409]|0){Tw(592875)|0;if(!U){a=0;do{Tw(592840)|0;da=k[b+(a<<2)>>2]|0;T=0;do{p[u>>3]=+p[da+(T<<3)>>3];Rw(592848,u)|0;T=T+1|0}while((T|0)<(pa|0));Tw(592856)|0;p[t>>3]=+p[k[$+(a<<2)>>2]>>3];Rw(592865,t)|0;a=a+1|0}while((a|0)<(pa|0))}Sw(10)|0}a=k[25934]|0;if((a|0)>0){T=a;da=0;while(1){if((T|0)>0){Ca=ha(da,Fa)|0;V=T;X=0;while(1){ba=k[$+(X+Ca<<2)>>2]|0;Ea=(k[(k[25936]|0)+(da<<2)>>2]|0)+(X<<3)|0;qa=+p[ba>>3]*.5+ +p[Ea>>3];p[Ea>>3]=qa;if(qa!=0.0)rb=+p[ba>>3]/qa*100.0;else rb=-999.0;if(!(k[25409]|0))sb=V;else{k[s>>2]=da;k[s+4>>2]=X;p[s+8>>3]=qa;p[s+16>>3]=rb;Rw(593360,s)|0;zw(R)|0;sb=k[25934]|0}X=X+1|0;if((X|0)>=(sb|0)){tb=sb;break}else V=sb}}else tb=T;da=da+1|0;if((da|0)>=(tb|0)){ub=tb;break}else T=tb}}else ub=a;if(!(k[25409]|0))vb=ub;else{Sw(10)|0;vb=k[25934]|0}if((vb|0)>0){T=vb;da=0;while(1){if((T|0)>0){V=(ha(da,Fa)|0)+pa|0;X=T;Ca=0;while(1){ba=k[$+(V+Ca<<2)>>2]|0;Ea=(k[(k[25938]|0)+(da<<2)>>2]|0)+(Ca<<3)|0;qa=+p[ba>>3]*.5+ +p[Ea>>3];p[Ea>>3]=qa;if(qa!=0.0)wb=+p[ba>>3]/qa*100.0;else wb=-999.0;if(!(k[25409]|0))xb=X;else{k[q>>2]=da;k[q+4>>2]=Ca;p[q+8>>3]=qa;p[q+16>>3]=wb;Rw(593390,q)|0;zw(R)|0;xb=k[25934]|0}Ca=Ca+1|0;if((Ca|0)>=(xb|0)){yb=xb;break}else X=xb}}else yb=T;da=da+1|0;if((da|0)>=(yb|0))break;else T=yb}}Rc()|0;if((k[25409]|0)>1){k[o>>2]=Q;Rw(593420,o)|0;zw(R)|0}T=k[25953]|0;if(+p[T+136>>3]+1.0>0.0){da=T;qa=0.0;T=0;Ma=0.0;Ha=0.0;while(1){if(+p[da+144>>3]+1.0>0.0){Ia=qa+-.5;sa=0.0;a=0;La=Ma;bb=Ha;while(1){Ja=sa+-.5;Ad(k[25954]|0,Ia,Ja,ia,ja);k[ea>>2]=0;Bd(k[25953]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);Ka=+p[fa>>3]-+p[880];p[fa>>3]=Ka;ab=+p[ga>>3]-+p[881];p[ga>>3]=ab;k[ea>>2]=0;if((k[25409]|0)>1?($a=+p[ia>>3],Da=+p[ja>>3],p[n>>3]=Ia,p[n+8>>3]=Ja,p[n+16>>3]=$a,p[n+24>>3]=Da,p[n+32>>3]=Ka,p[n+40>>3]=ab,k[n+48>>2]=0,Rw(593464,n)|0,zw(R)|0,Ga&(k[ea>>2]|0)!=0):0){zb=La;Ab=bb}else xa=213;do if((xa|0)==213){xa=0;ab=+W(+(Ia-+p[fa>>3]));Ka=ab>La?ab:La;ab=+W(+(Ja-+p[ga>>3]));if(!(ab>bb)){zb=Ka;Ab=bb;break}zb=Ka;Ab=ab}while(0);a=a+la|0;sa=+(a|0);X=k[25953]|0;if(!(sa<+p[X+144>>3]+1.0)){Bb=X;Cb=zb;Db=Ab;break}else{La=zb;bb=Ab}}}else{Bb=da;Cb=Ma;Db=Ha}T=T+la|0;qa=+(T|0);if(!(qa<+p[Bb+136>>3]+1.0)){Eb=Cb;Fb=Db;break}else{da=Bb;Ma=Cb;Ha=Db}}}else{Eb=0.0;Fb=0.0}if(k[25409]|0){p[m>>3]=Eb;k[m+8>>2]=Q;Rw(593525,m)|0;p[l>>3]=Fb;k[l+8>>2]=Q;Rw(593545,l)|0;zw(R)|0}if(Eb<va&Fb<va){Gb=1;Hb=jb;Ib=kb;Jb=Q;break d}da=Q+1|0;if((da|0)<(ua|0))Q=da;else{Gb=0;Hb=jb;Ib=kb;Jb=da;break d}}Pw(593265,70,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}else{Gb=0;Hb=jb;Ib=kb;Jb=0}}else{Rc()|0;Gb=0;Hb=0;Ib=0;Jb=0}while(0);if((k[25409]|0)!=0?(Sw(10)|0,(k[25409]|0)>1):0){Tw(593565)|0;zw(k[31231]|0)|0}kb=k[25953]|0;if(+p[kb+136>>3]+1.0>0.0){jb=k[31231]|0;l=(wa|0)==0;m=kb;va=0.0;Fb=0.0;Eb=0.0;kb=0;while(1){if(+p[m+144>>3]+1.0>0.0){Db=va+-.5;Cb=0.0;Bb=m;Ab=Fb;zb=Eb;la=0;while(1){wb=Cb+-.5;Ad(Bb,Db,wb,ia,ja);k[ea>>2]=0;Bd(k[25954]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);if((k[25409]|0)>1){rb=+p[ia>>3];ib=+p[ja>>3];hb=+p[fa>>3];gb=+p[ga>>3];n=k[ea>>2]|0;p[j>>3]=Db;p[j+8>>3]=wb;p[j+16>>3]=rb;p[j+24>>3]=ib;p[j+32>>3]=hb;p[j+40>>3]=gb;k[j+48>>2]=n;Rw(592999,j)|0;zw(jb)|0}do if(l&(k[ea>>2]|0)!=0){Kb=Ab;Lb=zb}else{gb=+W(+(Db-+p[fa>>3]));hb=gb>Ab?gb:Ab;gb=+W(+(wb-+p[ga>>3]));if(!(gb>zb)){Kb=hb;Lb=zb;break}Kb=hb;Lb=gb}while(0);la=la+na|0;Cb=+(la|0);n=k[25953]|0;if(!(Cb<+p[n+144>>3]+1.0)){Mb=n;Nb=Kb;Ob=Lb;break}else{Bb=n;Ab=Kb;zb=Lb}}}else{Mb=m;Nb=Fb;Ob=Eb}kb=kb+na|0;va=+(kb|0);if(!(va<+p[Mb+136>>3]+1.0)){Pb=Nb;Qb=Ob;break}else{m=Mb;Fb=Nb;Eb=Ob}}}else{Pb=0.0;Qb=0.0}if((k[25409]|0)!=0?(Sw(10)|0,p[i>>3]=Pb,Rw(593601,i)|0,p[h>>3]=Qb,Rw(593622,h)|0,h=k[31231]|0,zw(h)|0,(k[25409]|0)>1):0){Tw(593643)|0;zw(h)|0}h=k[25953]|0;if(+p[h+136>>3]+1.0>0.0){i=k[31231]|0;Mb=(wa|0)==0;wa=h;Ob=0.0;h=0;Eb=0.0;Nb=0.0;while(1){if(+p[wa+144>>3]+1.0>0.0){Fb=Ob+-.5;va=0.0;m=0;Lb=Eb;Kb=Nb;while(1){zb=va+-.5;Ad(k[25954]|0,Fb,zb,ia,ja);k[ea>>2]=0;Bd(k[25953]|0,+p[ia>>3],+p[ja>>3],fa,ga,ea);Ab=+p[fa>>3]-+p[880];p[fa>>3]=Ab;Cb=+p[ga>>3]-+p[881];p[ga>>3]=Cb;k[ea>>2]=0;if((k[25409]|0)>1?(Db=+p[ia>>3],wb=+p[ja>>3],p[g>>3]=Fb,p[g+8>>3]=zb,p[g+16>>3]=Db,p[g+24>>3]=wb,p[g+32>>3]=Ab,p[g+40>>3]=Cb,k[g+48>>2]=0,Rw(593464,g)|0,zw(i)|0,Mb&(k[ea>>2]|0)!=0):0){Rb=Lb;Sb=Kb}else xa=244;do if((xa|0)==244){xa=0;Cb=+W(+(Fb-+p[fa>>3]));Ab=Cb>Lb?Cb:Lb;Cb=+W(+(zb-+p[ga>>3]));if(!(Cb>Kb)){Rb=Ab;Sb=Kb;break}Rb=Ab;Sb=Cb}while(0);m=m+na|0;va=+(m|0);kb=k[25953]|0;if(!(va<+p[kb+144>>3]+1.0)){Tb=kb;Ub=Rb;Vb=Sb;break}else{Lb=Rb;Kb=Sb}}}else{Tb=wa;Ub=Eb;Vb=Nb}h=h+na|0;Ob=+(h|0);if(!(Ob<+p[Tb+136>>3]+1.0)){Wb=Ub;Xb=Vb;break}else{wa=Tb;Eb=Ub;Nb=Vb}}}else{Wb=0.0;Xb=0.0}if(!(k[25409]|0)){Yb=k[31231]|0;Zb=k[25935]|0;ww(Zb)|0;_b=ua+-1|0;$b=Hb?Ib:_b;ac=Gb?Jb:_b;bc=k[25412]|0;cc=$b+1|0;dc=ac+1|0;p[d>>3]=Pb;ec=d+8|0;p[ec>>3]=Qb;fc=d+16|0;k[fc>>2]=cc;hc=d+24|0;p[hc>>3]=Wb;ic=d+32|0;p[ic>>3]=Xb;jc=d+40|0;k[jc>>2]=dc;Ew(bc,593722,d)|0;zw(Yb)|0;kc=k[25412]|0;zw(kc)|0;lc=k[25412]|0;ww(lc)|0;gc(0)}else{p[f>>3]=Wb;Rw(593679,f)|0;p[e>>3]=Xb;Rw(593700,e)|0;e=k[31231]|0;zw(e)|0;Yb=e;Zb=k[25935]|0;ww(Zb)|0;_b=ua+-1|0;$b=Hb?Ib:_b;ac=Gb?Jb:_b;bc=k[25412]|0;cc=$b+1|0;dc=ac+1|0;p[d>>3]=Pb;ec=d+8|0;p[ec>>3]=Qb;fc=d+16|0;k[fc>>2]=cc;hc=d+24|0;p[hc>>3]=Wb;ic=d+32|0;p[ic>>3]=Xb;jc=d+40|0;k[jc>>2]=dc;Ew(bc,593722,d)|0;zw(Yb)|0;kc=k[25412]|0;zw(kc)|0;lc=k[25412]|0;ww(lc)|0;gc(0)}}return 0}function Oc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;b=r;r=r+368|0;c=b+104|0;d=b+96|0;e=b+88|0;f=b+80|0;g=b+72|0;h=b+64|0;j=b+56|0;l=b+48|0;m=b+40|0;n=b+32|0;o=b+24|0;p=b+16|0;q=b+8|0;s=b;t=b+112|0;u=Dw(a,774853)|0;if(!u){v=k[25412]|0;k[s>>2]=a;Ew(v,593818,s)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}k[25940]=0;k[25941]=0;k[25942]=0;k[25943]=0;k[25944]=0;k[25945]=0;k[25946]=0;k[25947]=0;k[25948]=0;k[25949]=0;k[25950]=0;k[25951]=0;k[25952]=0;i[590823]=0;i[590903]=0;i[590983]=0;i[591063]=0;i[591143]=0;i[591223]=0;i[591303]=0;i[591383]=0;i[591463]=0;i[591543]=0;i[591623]=0;i[591703]=0;i[591783]=0;if(Bw(t,256,u)|0)do{s=t+(xx(t)|0)|0;v=t;while(1){a=i[v>>0]|0;if(v>>>0<s>>>0&a<<24>>24==32)v=v+1|0;else{w=a;x=v;break}}v=w;a=x;a:while(1){switch(v<<24>>24){case 61:case 32:{y=v;z=a;break a;break}default:{}}if(a>>>0>=s>>>0){y=v;z=a;break}A=a+1|0;v=i[A>>0]|0;a=A}a=y;v=z;b:while(1){switch(a<<24>>24){case 39:case 32:case 61:break;default:{B=v;break b}}if(v>>>0>=s>>>0){B=v;break}A=v+1|0;a=i[A>>0]|0;v=A}i[z>>0]=0;v=(i[B>>0]|0)==39?B+1|0:B;c:while(1){switch(i[v>>0]|0){case 10:case 13:case 39:case 32:{C=v;break c;break}default:{}}if(v>>>0>=s>>>0){C=v;break}v=v+1|0}i[C>>0]=0;do if(tx(x,777890)|0){if(!(tx(x,777921)|0)){k[25941]=1;ux(590903,B)|0;break}if(!(tx(x,777928)|0)){k[25942]=1;ux(590983,B)|0;break}if(!(tx(x,777897)|0)){k[25943]=1;ux(591063,B)|0;break}if(!(tx(x,777909)|0)){k[25944]=1;ux(591143,B)|0;break}if(!(tx(x,777903)|0)){k[25945]=1;ux(591223,B)|0;break}if(!(tx(x,777915)|0)){k[25946]=1;ux(591303,B)|0;break}if(!(tx(x,777935)|0)){k[25947]=1;ux(591383,B)|0;break}if(!(tx(x,777947)|0)){k[25948]=1;ux(591463,B)|0;break}if(!(tx(x,777941)|0)){k[25949]=1;ux(591543,B)|0;break}if(!(tx(x,777953)|0)){k[25950]=1;ux(591623,B)|0;break}if(!(tx(x,758683)|0)){k[25951]=1;ux(591703,B)|0;break}if(!(tx(x,601450)|0)){k[25952]=1;ux(591783,B)|0}}else{k[25940]=1;ux(590823,B)|0}while(0)}while((Bw(t,256,u)|0)!=0);if(!(k[25409]|0)){r=b;return 0}Tw(593865)|0;if(k[25940]|0){k[q>>2]=590823;Rw(593879,q)|0}if(k[25941]|0){k[p>>2]=590903;Rw(593895,p)|0}if(k[25942]|0){k[o>>2]=590983;Rw(593911,o)|0}if(k[25943]|0){k[n>>2]=591063;Rw(593927,n)|0}if(k[25944]|0){k[m>>2]=591143;Rw(593943,m)|0}if(k[25945]|0){k[l>>2]=591223;Rw(593959,l)|0}if(k[25946]|0){k[j>>2]=591303;Rw(593975,j)|0}if(k[25947]|0){k[h>>2]=591383;Rw(593991,h)|0}if(k[25948]|0){k[g>>2]=591463;Rw(594007,g)|0}if(k[25949]|0){k[f>>2]=591543;Rw(594023,f)|0}if(k[25950]|0){k[e>>2]=591623;Rw(594039,e)|0}if(k[25951]|0){k[d>>2]=591703;Rw(594055,d)|0}if(k[25952]|0){k[c>>2]=591783;Rw(594071,c)|0}Sw(10)|0;r=b;return 0}function Pc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0;b=r;r=r+80416|0;c=b+56|0;d=b+48|0;e=b+40|0;f=b+32|0;g=b+80328|0;h=b+80072|0;j=b+72|0;l=b+24|0;m=b+16|0;n=b+8|0;o=b;q=b+64|0;s=Dw(a,774853)|0;if(!s){t=k[25412]|0;k[f>>2]=a;Ew(t,593818,f)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}i[j>>0]=0;f=0;do{if(!(Bw(h,256,s)|0))break;t=h+((xx(h)|0)+-1)|0;if((i[t>>0]|0)==10)i[t>>0]=0;t=h+((xx(h)|0)+-1)|0;if((i[t>>0]|0)==13)i[t>>0]=0;t=xx(j)|0;a=xx(h)|0;if(!((a|0)>0?(py(j+t|0,h|0,a|0)|0,(a|0)>=80):0))sy(j+(a+t)|0,32,80-a|0)|0;i[j+(t+80)>>0]=0;f=f+1|0}while((f|0)<1e3);if(k[25409]|0){Tw(594087)|0;f=k[31231]|0;zw(f)|0;h=xx(j)|0;t=g+80|0;a=0;do{u=g;v=u+80|0;do{i[u>>0]=0;u=u+1|0}while((u|0)<(v|0));u=a*80|0;v=0;while(1){w=v+u|0;if((w|0)>(h|0)){x=1;break}i[g+v>>0]=i[j+w>>0]|0;v=v+1|0;if((v|0)>=80){x=0;break}}i[t>>0]=0;v=0;u=80;a:while(1){w=g+u|0;switch(v<<24>>24){case 0:case 32:break;default:break a}i[w>>0]=0;w=u+-1|0;if((u|0)<=0)break;v=i[g+w>>0]|0;u=w}a=a+1|0;if(i[g>>0]|0){k[e>>2]=a;k[e+4>>2]=g;Rw(594113,e)|0}}while(!x);Sw(10)|0;zw(f)|0}f=md(j)|0;k[25953]=f;if(!f){Pw(594122,54,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}Zc(f,0)|0;if(k[25409]|0){Tw(594177)|0;zw(k[31231]|0)|0}k[q>>2]=0;Ad(k[25953]|0,.5,.5,n,o);Bd(k[25953]|0,+p[n>>3],+p[o>>3],l,m,q);y=+p[l>>3]+-.5;p[880]=y;p[881]=+p[m>>3]+-.5;if(!(k[25409]|0)){ww(s)|0;r=b;return 0}p[d>>3]=y;Rw(594216,d)|0;p[c>>3]=+p[881];Rw(594243,c)|0;zw(k[31231]|0)|0;ww(s)|0;r=b;return 0}function Qc(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0.0,t=0,u=0,v=0,w=0.0,x=0,y=0,z=0,A=0.0,B=0,C=0,D=0,E=0,F=0.0,G=0.0,H=0,I=0,J=0,K=0,L=0,M=0,N=0.0,O=0,P=0,Q=0.0,R=0,S=0,T=0.0,U=0,V=0;e=r;r=r+32|0;f=e+8|0;g=e;h=b<<2;i=cy(h)|0;if(!i)Sc(594271);j=cy(h)|0;if(!j)Sc(594271);l=cy(h)|0;if(!l)Sc(594271);h=(b|0)>0;do if(h){sy(l|0,0,b<<2|0)|0;m=(d|0)>0;n=0;o=0;q=0;a:while(1){s=0.0;t=o;u=q;v=0;while(1){if((k[l+(v<<2)>>2]|0)==1){w=s;x=t;y=u}else{z=a+(v<<2)|0;A=s;B=t;C=u;D=0;while(1){E=k[l+(D<<2)>>2]|0;if(!E){F=+W(+(+p[(k[z>>2]|0)+(D<<3)>>3]));if(!(F>=A)){G=A;H=B;I=C}else{G=F;H=D;I=v}}else if((E|0)>1){J=22;break a}else{G=A;H=B;I=C}D=D+1|0;if((D|0)>=(b|0)){w=G;x=H;y=I;break}else{A=G;B=H;C=I}}}v=v+1|0;if((v|0)>=(b|0)){K=x;L=y;break}else{s=w;t=x;u=y}}k[l+(K<<2)>>2]=(k[l+(K<<2)>>2]|0)+1;if((L|0)!=(K|0)){u=k[a+(L<<2)>>2]|0;t=k[a+(K<<2)>>2]|0;v=0;do{C=u+(v<<3)|0;s=+p[C>>3];B=t+(v<<3)|0;p[C>>3]=+p[B>>3];p[B>>3]=s;v=v+1|0}while((v|0)!=(b|0));if(m){v=k[c+(L<<2)>>2]|0;t=k[c+(K<<2)>>2]|0;u=0;do{B=v+(u<<3)|0;s=+p[B>>3];C=t+(u<<3)|0;p[B>>3]=+p[C>>3];p[C>>3]=s;u=u+1|0}while((u|0)!=(d|0))}}k[j+(n<<2)>>2]=L;k[i+(n<<2)>>2]=K;u=k[a+(K<<2)>>2]|0;t=u+(K<<3)|0;s=+p[t>>3];if(s==0.0){J=30;break}A=1.0/s;p[t>>3]=1.0;t=0;do{v=u+(t<<3)|0;p[v>>3]=A*+p[v>>3];t=t+1|0}while((t|0)!=(b|0));if(m){t=k[c+(K<<2)>>2]|0;v=0;do{C=t+(v<<3)|0;p[C>>3]=A*+p[C>>3];v=v+1|0}while((v|0)!=(d|0))}v=c+(K<<2)|0;t=0;do{if((t|0)!=(K|0)){C=k[a+(t<<2)>>2]|0;B=C+(K<<3)|0;A=+p[B>>3];p[B>>3]=0.0;B=0;do{D=C+(B<<3)|0;p[D>>3]=+p[D>>3]-A*+p[u+(B<<3)>>3];B=B+1|0}while((B|0)!=(b|0));if(m){B=k[v>>2]|0;C=k[c+(t<<2)>>2]|0;D=0;do{z=C+(D<<3)|0;p[z>>3]=+p[z>>3]-A*+p[B+(D<<3)>>3];D=D+1|0}while((D|0)!=(d|0))}}t=t+1|0}while((t|0)!=(b|0));t=n+1|0;if((t|0)<(b|0)){n=t;o=K;q=L}else{J=18;break}}if((J|0)==18){if(h)M=b;else break;do{q=M;M=M+-1|0;o=k[j+(M<<2)>>2]|0;n=k[i+(M<<2)>>2]|0;if((o|0)!=(n|0)){m=0;do{t=k[a+(m<<2)>>2]|0;v=t+(o<<3)|0;A=+p[v>>3];u=t+(n<<3)|0;p[v>>3]=+p[u>>3];p[u>>3]=A;m=m+1|0}while((m|0)!=(b|0))}}while((q|0)>1)}else if((J|0)==22)Sc(594303);else if((J|0)==30)Sc(594321)}while(0);if(!(k[25409]|0)){dy(l);dy(j);dy(i);r=e;return}Tw(594339)|0;if(h){h=0;w=0.0;J=-1;M=-1;while(1){L=a+(h<<2)|0;K=0;G=w;d=J;c=M;while(1){y=k[a+(K<<2)>>2]|0;A=+p[y+(h<<3)>>3]/+X(+(+W(+(+p[y+(K<<3)>>3]*+p[(k[L>>2]|0)+(h<<3)>>3]))));p[g>>3]=A;Rw(594361,g)|0;if((K|0)!=(h|0)?(s=+W(+A),s>G):0){N=s;O=K;P=h}else{N=G;O=d;P=c}K=K+1|0;if((K|0)==(b|0)){Q=N;R=O;S=P;break}else{G=N;d=O;c=P}}Sw(10)|0;h=h+1|0;if((h|0)==(b|0)){T=Q;U=R;V=S;break}else{w=Q;J=R;M=S}}}else{T=0.0;U=-1;V=-1}p[f>>3]=T;k[f+8>>2]=U;k[f+12>>2]=V;Rw(594368,f)|0;dy(l);dy(j);dy(i);r=e;return}function Rc(){var a=0,b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0.0,wa=0,xa=0.0,ya=0,za=0.0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0;a=r;r=r+33680|0;b=a+640|0;c=a+552|0;d=a+544|0;e=a+528|0;f=a+520|0;g=a+512|0;h=a+504|0;j=a+488|0;l=a+480|0;m=a+472|0;n=a+464|0;o=a+448|0;q=a+440|0;s=a+432|0;t=a+424|0;u=a+408|0;v=a+400|0;w=a+392|0;x=a+384|0;y=a+376|0;z=a+368|0;A=a+360|0;B=a+352|0;C=a+344|0;D=a+336|0;E=a+328|0;F=a+320|0;G=a+312|0;H=a+304|0;I=a+296|0;J=a+288|0;K=a+280|0;L=a+272|0;M=a+264|0;N=a+256|0;O=a+248|0;P=a+240|0;Q=a+232|0;R=a+224|0;S=a+216|0;T=a+208|0;U=a+200|0;V=a+192|0;W=a+184|0;X=a+176|0;Y=a+168|0;Z=a+160|0;_=a+152|0;$=a+144|0;aa=a+136|0;ba=a+128|0;ca=a+120|0;da=a+112|0;ea=a+104|0;fa=a+96|0;ga=a+88|0;ha=a+80|0;ia=a+72|0;ja=a+64|0;ka=a+56|0;la=a+48|0;ma=a+40|0;na=a+32|0;oa=a+24|0;pa=a+16|0;qa=a+904|0;ra=a+648|0;sa=a+8|0;ta=a;Vw(k[25935]|0);ua=k[25953]|0;va=+p[ua+136>>3];wa=~~va;xa=+p[ua+144>>3];ya=~~xa;za=(va+1.0)*.5;va=(xa+1.0)*.5;Ad(ua,za,va,sa,ta);i[qa>>0]=0;ua=ra;Aa=594408;Ba=ua+12|0;do{i[ua>>0]=i[Aa>>0]|0;ua=ua+1|0;Aa=Aa+1|0}while((ua|0)<(Ba|0));Ca=xx(qa)|0;Da=xx(ra)|0;if(!((Da|0)>0?(py(qa+Ca|0,ra|0,Da|0)|0,(Da|0)>=80):0))sy(qa+(Da+Ca)|0,32,80-Da|0)|0;i[qa+(Ca+80)>>0]=0;Ca=k[25935]|0;k[pa>>2]=ra;Ew(Ca,787303,pa)|0;ua=ra;Aa=594420;Ba=ua+14|0;do{i[ua>>0]=i[Aa>>0]|0;ua=ua+1|0;Aa=Aa+1|0}while((ua|0)<(Ba|0));pa=xx(qa)|0;Ca=xx(ra)|0;if(!((Ca|0)>0?(py(qa+pa|0,ra|0,Ca|0)|0,(Ca|0)>=80):0))sy(qa+(Ca+pa)|0,32,80-Ca|0)|0;i[qa+(pa+80)>>0]=0;pa=k[25935]|0;k[oa>>2]=ra;Ew(pa,787303,oa)|0;ua=ra;Aa=594434;Ba=ua+12|0;do{i[ua>>0]=i[Aa>>0]|0;ua=ua+1|0;Aa=Aa+1|0}while((ua|0)<(Ba|0));oa=xx(qa)|0;pa=xx(ra)|0;if(!((pa|0)>0?(py(qa+oa|0,ra|0,pa|0)|0,(pa|0)>=80):0))sy(qa+(pa+oa)|0,32,80-pa|0)|0;i[qa+(oa+80)>>0]=0;oa=k[25935]|0;k[na>>2]=ra;Ew(oa,787303,na)|0;k[ma>>2]=wa;Xw(ra,594446,ma)|0;ma=xx(qa)|0;wa=xx(ra)|0;if(!((wa|0)>0?(py(qa+ma|0,ra|0,wa|0)|0,(wa|0)>=80):0))sy(qa+(wa+ma)|0,32,80-wa|0)|0;i[qa+(ma+80)>>0]=0;ma=k[25935]|0;k[la>>2]=ra;Ew(ma,787303,la)|0;k[ka>>2]=ya;Xw(ra,594459,ka)|0;ka=xx(qa)|0;ya=xx(ra)|0;if(!((ya|0)>0?(py(qa+ka|0,ra|0,ya|0)|0,(ya|0)>=80):0))sy(qa+(ya+ka)|0,32,80-ya|0)|0;i[qa+(ka+80)>>0]=0;ka=k[25935]|0;k[ja>>2]=ra;Ew(ka,787303,ja)|0;ja=(k[25953]|0)+3449|0;if(!(Ax(ja,606317,2)|0)){ua=ra;Aa=594472;Ba=ua+25|0;do{i[ua>>0]=i[Aa>>0]|0;ua=ua+1|0;Aa=Aa+1|0}while((ua|0)<(Ba|0));ka=xx(qa)|0;ya=xx(ra)|0;if(!((ya|0)>0?(py(qa+ka|0,ra|0,ya|0)|0,(ya|0)>=80):0))sy(qa+(ya+ka)|0,32,80-ya|0)|0;i[qa+(ka+80)>>0]=0;ka=k[25935]|0;k[ia>>2]=ra;Ew(ka,787303,ia)|0;ua=ra;Aa=594497;Ba=ua+25|0;do{i[ua>>0]=i[Aa>>0]|0;ua=ua+1|0;Aa=Aa+1|0}while((ua|0)<(Ba|0));Aa=xx(qa)|0;ia=xx(ra)|0;if(!((ia|0)>0?(py(qa+Aa|0,ra|0,ia|0)|0,(ia|0)>=80):0))sy(qa+(ia+Aa)|0,32,80-ia|0)|0;i[qa+(Aa+80)>>0]=0;Aa=k[25935]|0;k[ha>>2]=ra;Ew(Aa,787303,ha)|0}else{k[ga>>2]=ja;Xw(ra,594522,ga)|0;ga=xx(qa)|0;ja=xx(ra)|0;if(!((ja|0)>0?(py(qa+ga|0,ra|0,ja|0)|0,(ja|0)>=80):0))sy(qa+(ja+ga)|0,32,80-ja|0)|0;i[qa+(ga+80)>>0]=0;ga=k[25935]|0;k[fa>>2]=ra;Ew(ga,787303,fa)|0;k[ea>>2]=(k[25953]|0)+3458;Xw(ra,594545,ea)|0;ea=xx(qa)|0;fa=xx(ra)|0;if(!((fa|0)>0?(py(qa+ea|0,ra|0,fa|0)|0,(fa|0)>=80):0))sy(qa+(fa+ea)|0,32,80-fa|0)|0;i[qa+(ea+80)>>0]=0;ea=k[25935]|0;k[da>>2]=ra;Ew(ea,787303,da)|0}p[ca>>3]=+p[sa>>3];Xw(ra,594568,ca)|0;ca=xx(qa)|0;sa=xx(ra)|0;if(!((sa|0)>0?(py(qa+ca|0,ra|0,sa|0)|0,(sa|0)>=80):0))sy(qa+(sa+ca)|0,32,80-sa|0)|0;i[qa+(ca+80)>>0]=0;ca=k[25935]|0;k[ba>>2]=ra;Ew(ca,787303,ba)|0;p[aa>>3]=+p[ta>>3];Xw(ra,594586,aa)|0;aa=xx(qa)|0;ta=xx(ra)|0;if(!((ta|0)>0?(py(qa+aa|0,ra|0,ta|0)|0,(ta|0)>=80):0))sy(qa+(ta+aa)|0,32,80-ta|0)|0;i[qa+(aa+80)>>0]=0;aa=k[25935]|0;k[$>>2]=ra;Ew(aa,787303,$)|0;p[_>>3]=za;Xw(ra,594604,_)|0;_=xx(qa)|0;$=xx(ra)|0;if(!(($|0)>0?(py(qa+_|0,ra|0,$|0)|0,($|0)>=80):0))sy(qa+($+_)|0,32,80-$|0)|0;i[qa+(_+80)>>0]=0;_=k[25935]|0;k[Z>>2]=ra;Ew(_,787303,Z)|0;p[Y>>3]=va;Xw(ra,594622,Y)|0;Y=xx(qa)|0;Z=xx(ra)|0;if(!((Z|0)>0?(py(qa+Y|0,ra|0,Z|0)|0,(Z|0)>=80):0))sy(qa+(Z+Y)|0,32,80-Z|0)|0;i[qa+(Y+80)>>0]=0;Y=k[25935]|0;k[X>>2]=ra;Ew(Y,787303,X)|0;if(k[25940]|0){p[W>>3]=+p[878];Xw(ra,594640,W)|0;W=xx(qa)|0;X=xx(ra)|0;if(!((X|0)>0?(py(qa+W|0,ra|0,X|0)|0,(X|0)>=80):0))sy(qa+(X+W)|0,32,80-X|0)|0;i[qa+(W+80)>>0]=0;W=k[25935]|0;k[V>>2]=ra;Ew(W,787303,V)|0}if(k[25941]|0){p[U>>3]=+p[879];Xw(ra,594656,U)|0;U=xx(qa)|0;V=xx(ra)|0;if(!((V|0)>0?(py(qa+U|0,ra|0,V|0)|0,(V|0)>=80):0))sy(qa+(V+U)|0,32,80-V|0)|0;i[qa+(U+80)>>0]=0;U=k[25935]|0;k[T>>2]=ra;Ew(U,787303,T)|0}if(k[25942]|0){k[S>>2]=590983;Xw(ra,594672,S)|0;S=xx(qa)|0;T=xx(ra)|0;if(!((T|0)>0?(py(qa+S|0,ra|0,T|0)|0,(T|0)>=80):0))sy(qa+(T+S)|0,32,80-T|0)|0;i[qa+(S+80)>>0]=0;S=k[25935]|0;k[R>>2]=ra;Ew(S,787303,R)|0}if(k[25943]|0){k[Q>>2]=591063;Xw(ra,594685,Q)|0;Q=xx(qa)|0;R=xx(ra)|0;if(!((R|0)>0?(py(qa+Q|0,ra|0,R|0)|0,(R|0)>=80):0))sy(qa+(R+Q)|0,32,80-R|0)|0;i[qa+(Q+80)>>0]=0;Q=k[25935]|0;k[P>>2]=ra;Ew(Q,787303,P)|0}if(k[25944]|0){k[O>>2]=591143;Xw(ra,594698,O)|0;O=xx(qa)|0;P=xx(ra)|0;if(!((P|0)>0?(py(qa+O|0,ra|0,P|0)|0,(P|0)>=80):0))sy(qa+(P+O)|0,32,80-P|0)|0;i[qa+(O+80)>>0]=0;O=k[25935]|0;k[N>>2]=ra;Ew(O,787303,N)|0}if(k[25945]|0){k[M>>2]=591223;Xw(ra,594711,M)|0;M=xx(qa)|0;N=xx(ra)|0;if(!((N|0)>0?(py(qa+M|0,ra|0,N|0)|0,(N|0)>=80):0))sy(qa+(N+M)|0,32,80-N|0)|0;i[qa+(M+80)>>0]=0;M=k[25935]|0;k[L>>2]=ra;Ew(M,787303,L)|0}if(k[25946]|0){k[K>>2]=591303;Xw(ra,594724,K)|0;K=xx(qa)|0;L=xx(ra)|0;if(!((L|0)>0?(py(qa+K|0,ra|0,L|0)|0,(L|0)>=80):0))sy(qa+(L+K)|0,32,80-L|0)|0;i[qa+(K+80)>>0]=0;K=k[25935]|0;k[J>>2]=ra;Ew(K,787303,J)|0}if(k[25947]|0){k[I>>2]=591383;Xw(ra,594737,I)|0;I=xx(qa)|0;J=xx(ra)|0;if(!((J|0)>0?(py(qa+I|0,ra|0,J|0)|0,(J|0)>=80):0))sy(qa+(J+I)|0,32,80-J|0)|0;i[qa+(I+80)>>0]=0;I=k[25935]|0;k[H>>2]=ra;Ew(I,787303,H)|0}if(k[25948]|0){k[G>>2]=591463;Xw(ra,594750,G)|0;G=xx(qa)|0;H=xx(ra)|0;if(!((H|0)>0?(py(qa+G|0,ra|0,H|0)|0,(H|0)>=80):0))sy(qa+(H+G)|0,32,80-H|0)|0;i[qa+(G+80)>>0]=0;G=k[25935]|0;k[F>>2]=ra;Ew(G,787303,F)|0}if(k[25949]|0){k[E>>2]=591543;Xw(ra,594763,E)|0;E=xx(qa)|0;F=xx(ra)|0;if(!((F|0)>0?(py(qa+E|0,ra|0,F|0)|0,(F|0)>=80):0))sy(qa+(F+E)|0,32,80-F|0)|0;i[qa+(E+80)>>0]=0;E=k[25935]|0;k[D>>2]=ra;Ew(E,787303,D)|0}if(k[25950]|0){k[C>>2]=591623;Xw(ra,594776,C)|0;C=xx(qa)|0;D=xx(ra)|0;if(!((D|0)>0?(py(qa+C|0,ra|0,D|0)|0,(D|0)>=80):0))sy(qa+(D+C)|0,32,80-D|0)|0;i[qa+(C+80)>>0]=0;C=k[25935]|0;k[B>>2]=ra;Ew(C,787303,B)|0}if(k[25951]|0){k[A>>2]=591703;Xw(ra,594789,A)|0;A=xx(qa)|0;B=xx(ra)|0;if(!((B|0)>0?(py(qa+A|0,ra|0,B|0)|0,(B|0)>=80):0))sy(qa+(B+A)|0,32,80-B|0)|0;i[qa+(A+80)>>0]=0;A=k[25935]|0;k[z>>2]=ra;Ew(A,787303,z)|0}if(k[25952]|0){k[y>>2]=591783;Xw(ra,594802,y)|0;y=xx(qa)|0;z=xx(ra)|0;if(!((z|0)>0?(py(qa+y|0,ra|0,z|0)|0,(z|0)>=80):0))sy(qa+(z+y)|0,32,80-z|0)|0;i[qa+(y+80)>>0]=0;y=k[25935]|0;k[x>>2]=ra;Ew(y,787303,x)|0}k[w>>2]=(k[25934]|0)+-1;Xw(ra,594815,w)|0;w=xx(qa)|0;x=xx(ra)|0;if(!((x|0)>0?(py(qa+w|0,ra|0,x|0)|0,(x|0)>=80):0))sy(qa+(x+w)|0,32,80-x|0)|0;i[qa+(w+80)>>0]=0;w=k[25935]|0;k[v>>2]=ra;Ew(w,787303,v)|0;v=k[25934]|0;if((v|0)>0){w=v;x=0;while(1){if((w|0)>0){y=0;do{va=+p[(k[(k[25936]|0)+(x<<2)>>2]|0)+(y<<3)>>3];k[u>>2]=x;k[u+4>>2]=y;p[u+8>>3]=va;Xw(ra,594828,u)|0;z=xx(qa)|0;A=xx(ra)|0;if(!((A|0)>0?(py(qa+z|0,ra|0,A|0)|0,(A|0)>=80):0))sy(qa+(A+z)|0,32,80-A|0)|0;i[qa+(z+80)>>0]=0;z=k[25935]|0;k[t>>2]=ra;Ew(z,787303,t)|0;y=y+1|0;z=k[25934]|0}while((y|0)<(z|0));Ea=z}else Ea=w;x=x+1|0;if((x|0)>=(Ea|0)){Fa=Ea;break}else w=Ea}}else Fa=v;k[s>>2]=Fa+-1;Xw(ra,594847,s)|0;s=xx(qa)|0;Fa=xx(ra)|0;if(!((Fa|0)>0?(py(qa+s|0,ra|0,Fa|0)|0,(Fa|0)>=80):0))sy(qa+(Fa+s)|0,32,80-Fa|0)|0;i[qa+(s+80)>>0]=0;s=k[25935]|0;k[q>>2]=ra;Ew(s,787303,q)|0;q=k[25934]|0;if((q|0)>0){s=q;Fa=0;while(1){if((s|0)>0){v=0;do{va=+p[(k[(k[25938]|0)+(Fa<<2)>>2]|0)+(v<<3)>>3];k[o>>2]=Fa;k[o+4>>2]=v;p[o+8>>3]=va;Xw(ra,594860,o)|0;Ea=xx(qa)|0;w=xx(ra)|0;if(!((w|0)>0?(py(qa+Ea|0,ra|0,w|0)|0,(w|0)>=80):0))sy(qa+(w+Ea)|0,32,80-w|0)|0;i[qa+(Ea+80)>>0]=0;Ea=k[25935]|0;k[n>>2]=ra;Ew(Ea,787303,n)|0;v=v+1|0;Ea=k[25934]|0}while((v|0)<(Ea|0));Ga=Ea}else Ga=s;Fa=Fa+1|0;if((Fa|0)>=(Ga|0)){Ha=Ga;break}else s=Ga}}else Ha=q;k[m>>2]=Ha+-1;Xw(ra,594879,m)|0;m=xx(qa)|0;Ha=xx(ra)|0;if(!((Ha|0)>0?(py(qa+m|0,ra|0,Ha|0)|0,(Ha|0)>=80):0))sy(qa+(Ha+m)|0,32,80-Ha|0)|0;i[qa+(m+80)>>0]=0;m=k[25935]|0;k[l>>2]=ra;Ew(m,787303,l)|0;l=k[25934]|0;if((l|0)>0){m=l;Ha=0;while(1){if((m|0)>0){q=0;do{va=+p[(k[(k[25937]|0)+(Ha<<2)>>2]|0)+(q<<3)>>3];k[j>>2]=Ha;k[j+4>>2]=q;p[j+8>>3]=va;Xw(ra,594892,j)|0;Ga=xx(qa)|0;s=xx(ra)|0;if(!((s|0)>0?(py(qa+Ga|0,ra|0,s|0)|0,(s|0)>=80):0))sy(qa+(s+Ga)|0,32,80-s|0)|0;i[qa+(Ga+80)>>0]=0;Ga=k[25935]|0;k[h>>2]=ra;Ew(Ga,787303,h)|0;q=q+1|0;Ga=k[25934]|0}while((q|0)<(Ga|0));Ia=Ga}else Ia=m;Ha=Ha+1|0;if((Ha|0)>=(Ia|0)){Ja=Ia;break}else m=Ia}}else Ja=l;k[g>>2]=Ja+-1;Xw(ra,594911,g)|0;g=xx(qa)|0;Ja=xx(ra)|0;if(!((Ja|0)>0?(py(qa+g|0,ra|0,Ja|0)|0,(Ja|0)>=80):0))sy(qa+(Ja+g)|0,32,80-Ja|0)|0;i[qa+(g+80)>>0]=0;g=k[25935]|0;k[f>>2]=ra;Ew(g,787303,f)|0;f=k[25934]|0;if((f|0)>0){g=f;f=0;while(1){if((g|0)>0){Ja=0;do{va=+p[(k[(k[25939]|0)+(f<<2)>>2]|0)+(Ja<<3)>>3];k[e>>2]=f;k[e+4>>2]=Ja;p[e+8>>3]=va;Xw(ra,594924,e)|0;l=xx(qa)|0;Ia=xx(ra)|0;if(!((Ia|0)>0?(py(qa+l|0,ra|0,Ia|0)|0,(Ia|0)>=80):0))sy(qa+(Ia+l)|0,32,80-Ia|0)|0;i[qa+(l+80)>>0]=0;l=k[25935]|0;k[d>>2]=ra;Ew(l,787303,d)|0;Ja=Ja+1|0;l=k[25934]|0}while((Ja|0)<(l|0));Ka=l}else Ka=g;f=f+1|0;if((f|0)>=(Ka|0))break;else g=Ka}}k[ra>>2]=4476485;Ka=xx(qa)|0;g=xx(ra)|0;if(!((g|0)>0?(py(qa+Ka|0,ra|0,g|0)|0,(g|0)>=80):0))sy(qa+(g+Ka)|0,32,80-g|0)|0;i[qa+(Ka+80)>>0]=0;Ka=k[25935]|0;k[c>>2]=ra;Ew(Ka,787303,c)|0;if(k[25409]|0){Tw(594943)|0;Ka=xx(qa)|0;ra=c+80|0;g=0;do{ua=c;Ba=ua+80|0;do{i[ua>>0]=0;ua=ua+1|0}while((ua|0)<(Ba|0));f=g*80|0;d=0;while(1){e=d+f|0;if((e|0)>(Ka|0)){La=1;break}i[c+d>>0]=i[qa+e>>0]|0;d=d+1|0;if((d|0)>=80){La=0;break}}i[ra>>0]=0;d=0;f=80;a:while(1){e=c+f|0;switch(d<<24>>24){case 0:case 32:break;default:break a}i[e>>0]=0;e=f+-1|0;if((f|0)<=0)break;d=i[c+e>>0]|0;f=e}g=g+1|0;if(i[c>>0]|0){k[b>>2]=g;k[b+4>>2]=c;Rw(594113,b)|0}}while(!La);Sw(10)|0;zw(k[31231]|0)|0}La=md(qa)|0;k[25954]=La;if(!La){Pw(594122,54,1,k[25412]|0)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}if(!(k[25409]|0)){Ma=k[25935]|0;zw(Ma)|0;r=a;return 0}Tw(594974)|0;zw(k[31231]|0)|0;Ma=k[25935]|0;zw(Ma)|0;r=a;return 0}function Sc(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;b=k[25412]|0;k[c>>2]=a;Ew(b,595012,c)|0;zw(k[25412]|0)|0;ww(k[25412]|0)|0;gc(1)}function Tc(a,b,c,d,e,f,g){a=a|0;b=b|0;c=+c;d=+d;e=+e;f=+f;g=+g;var h=0,i=0,j=0,k=0,l=0,m=0,n=0,o=0.0,q=0,s=0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0,z=0,A=0;h=r;r=r+1664|0;i=h+864|0;j=h+64|0;k=h+32|0;l=h;m=Uc(4,a,b,7056,7856,c,1)|0;if((((m|0)>0?(n=Uc(m,7056,7856,8656,9456,d,0)|0,(n|0)>0):0)?(m=Uc(n,9456,8656,7856,7056,e,1)|0,(m|0)>0):0)?(n=Uc(m,7856,7056,j,i,f,0)|0,(n|0)>=3):0){o=0.0;m=0;while(1){q=m;m=m+1|0;s=(m|0)==(n|0);t=s?0:m;u=o+(+p[i+(q<<3)>>3]*+p[j+(t<<3)>>3]-+p[i+(t<<3)>>3]*+p[j+(q<<3)>>3]);if(s){v=u;break}else o=u}w=+W(+v)*.5*g;r=h;return +w}p[k>>3]=c;p[l>>3]=e;p[k+8>>3]=d;p[l+8>>3]=e;p[k+16>>3]=d;p[l+16>>3]=f;p[k+24>>3]=c;p[l+24>>3]=f;f=+p[a>>3];c=+p[b>>3];b=0;a=0;a:while(1){j=a;while(1){i=j;j=j+1|0;m=(j|0)%4|0;d=+p[l+(i<<3)>>3];if(d<=c?(e=+p[l+(m<<3)>>3],e>c):0){x=e;y=12}else y=9;if(((y|0)==9?(y=0,d>c):0)?(e=+p[l+(m<<3)>>3],e<=c):0){x=e;y=12}if((y|0)==12?(y=0,e=+p[k+(i<<3)>>3],e+(c-d)/(x-d)*(+p[k+(m<<3)>>3]-e)>f):0){z=j;break}if((j|0)>=4){A=b;break a}}j=b+1|0;if((z|0)<4){b=j;a=z}else{A=j;break}}w=(A&1|0)==0?0.0:g;r=h;return +w}function Uc(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=+f;g=g|0;var h=0,i=0,j=0.0,k=0,l=0,m=0,n=0,o=0,q=0,r=0.0,s=0.0,t=0,u=0.0,v=0,w=0,x=0,y=0,z=0.0,A=0,B=0.0,C=0,D=0,E=0.0;h=a+-1|0;i=b+(h<<3)|0;j=+p[i>>3];k=(g|0)==0;g=(k?j<=f:j>=f)&1;if((a|0)<=0){l=0;return l|0}m=c+(h<<3)|0;if(k){k=0;h=g;n=0;while(1){o=b+(k<<3)|0;j=+p[o>>3];q=j<=f;do if(!h)if(q){if(!k){r=+p[m>>3];s=+p[i>>3];t=c;u=r+(+p[c>>3]-r)*(f-s)/(j-s)}else{v=k+-1|0;s=+p[c+(v<<3)>>3];w=c+(k<<3)|0;r=+p[b+(v<<3)>>3];t=w;u=s+(+p[w>>3]-s)*(f-r)/(j-r)}p[d+(n<<3)>>3]=f;p[e+(n<<3)>>3]=u;w=n+1|0;p[d+(w<<3)>>3]=+p[o>>3];p[e+(w<<3)>>3]=+p[t>>3];x=1;y=n+2|0}else{x=0;y=n}else{if(q){p[d+(n<<3)>>3]=j;p[e+(n<<3)>>3]=+p[c+(k<<3)>>3];x=h;y=n+1|0;break}if(!k){r=+p[m>>3];s=+p[i>>3];z=r+(+p[c>>3]-r)*(f-s)/(+p[b>>3]-s)}else{w=k+-1|0;s=+p[c+(w<<3)>>3];r=+p[b+(w<<3)>>3];z=s+(+p[c+(k<<3)>>3]-s)*(f-r)/(j-r)}p[d+(n<<3)>>3]=f;p[e+(n<<3)>>3]=z;x=0;y=n+1|0}while(0);k=k+1|0;if((k|0)==(a|0)){l=y;break}else{h=x;n=y}}return l|0}else{y=0;n=g;g=0;while(1){x=b+(y<<3)|0;z=+p[x>>3];h=z>=f;do if(!n)if(h){if(!y){u=+p[m>>3];j=+p[i>>3];A=c;B=u+(+p[c>>3]-u)*(f-j)/(z-j)}else{k=y+-1|0;j=+p[c+(k<<3)>>3];t=c+(y<<3)|0;u=+p[b+(k<<3)>>3];A=t;B=j+(+p[t>>3]-j)*(f-u)/(z-u)}p[d+(g<<3)>>3]=f;p[e+(g<<3)>>3]=B;t=g+1|0;p[d+(t<<3)>>3]=+p[x>>3];p[e+(t<<3)>>3]=+p[A>>3];C=1;D=g+2|0}else{C=0;D=g}else{if(h){p[d+(g<<3)>>3]=z;p[e+(g<<3)>>3]=+p[c+(y<<3)>>3];C=n;D=g+1|0;break}if(!y){u=+p[m>>3];j=+p[i>>3];E=u+(+p[c>>3]-u)*(f-j)/(+p[b>>3]-j)}else{t=y+-1|0;j=+p[c+(t<<3)>>3];u=+p[b+(t<<3)>>3];E=j+(+p[c+(y<<3)>>3]-j)*(f-u)/(z-u)}p[d+(g<<3)>>3]=f;p[e+(g<<3)>>3]=E;C=0;D=g+1|0}while(0);y=y+1|0;if((y|0)==(a|0)){l=D;break}else{n=C;g=D}}return l|0}return 0}function Vc(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=r;r=r+16|0;c=b;d=b+4|0;e=jx(a,d,0)|0;f=(k[d>>2]|0)-a|0;if((f|0)<(xx(a)|0)){f=k[25412]|0;k[c>>2]=a;Ew(f,595044,c)|0;gc(1)}if((e|0)<0){Pw(595110,66,1,k[25412]|0)|0;gc(1)}else{r=b;return e|0}return 0}function Wc(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0;d=r;r=r+2736|0;e=d+344|0;f=d+336|0;g=d+328|0;h=d+320|0;j=d+312|0;l=d+304|0;m=d+296|0;n=d+288|0;o=d+280|0;p=d+272|0;q=d+264|0;s=d+256|0;t=d+248|0;u=d+240|0;v=d+232|0;w=d+224|0;x=d+216|0;y=d+208|0;z=d+200|0;A=d+192|0;B=d+184|0;C=d+176|0;D=d+168|0;E=d+160|0;F=d+152|0;G=d+144|0;H=d+136|0;I=d+128|0;J=d+120|0;K=d+112|0;L=d+104|0;M=d+96|0;N=d+88|0;O=d+80|0;P=d+72|0;Q=d+64|0;R=d+56|0;S=d+48|0;T=d+40|0;U=d+32|0;V=d+24|0;W=d+16|0;X=d+8|0;Y=d;Z=d+360|0;_=d+356|0;$=d+352|0;aa=d+2656|0;ba=d+2576|0;ca=d+2496|0;da=d+2416|0;ea=d+1392|0;fa=d+368|0;ga=d+348|0;k[$>>2]=0;ha=k[25982]|0;if(!ha){ia=cy(8e4)|0;k[25982]=ia;k[25983]=8e4;ja=ia}else ja=ha;k[25958]=0;k[25959]=0;k[25960]=0;k[25961]=0;k[25962]=0;k[25963]=0;k[25964]=0;k[25965]=0;k[25966]=0;k[25967]=0;k[25968]=0;k[25969]=0;k[25970]=0;k[25971]=0;k[25972]=0;k[25973]=0;k[25974]=0;k[25975]=0;k[25976]=0;k[25977]=0;k[25978]=0;k[25979]=0;k[25980]=0;k[25957]=0;ha=k[25955]|0;do if(ha){ia=Dw(ha,597225)|0;k[25935]=ia;if(!ia){ia=k[25412]|0;k[Y>>2]=k[25955];Ew(ia,597228,Y)|0;zw(k[25412]|0)|0;gc(1)}else{ka=k[25982]|0;break}}else ka=ja;while(0);i[ka>>0]=0;ka=(Ih(5,ga,a,0,$)|0)==0;ja=(k[25956]|0)!=0;if(ka){if(ja){Tw(597286)|0;zw(k[31231]|0)|0}if((b|0)==1){ka=k[25412]|0;k[X>>2]=a;Ew(ka,597297,X)|0;zw(k[25412]|0)|0;gc(1)}if((c|0)>0?(Fk(k[ga>>2]|0,c+1|0,0,$)|0)!=0:0)Xc(k[$>>2]|0)|0;if(Co(k[ga>>2]|0,Z,_,$)|0)Xc(k[$>>2]|0)|0;_=k[Z>>2]|0;if((_|0)>1e3){c=fy(k[25982]|0,(_*80|0)+1024|0)|0;k[25982]=c}if(k[25956]|0){k[W>>2]=k[Z>>2];Rw(597378,W)|0;zw(k[31231]|0)|0}a:do if((k[Z>>2]|0)>=1){W=ba+1|0;c=1;while(1){if(Uo(k[ga>>2]|0,c,aa,ba,ca,$)|0)break;if((i[ba>>0]|0)==39){ux(da,W)|0;_=da+((xx(da)|0)+-1)|0;if((i[_>>0]|0)==39)i[_>>0]=0}else ux(da,ba)|0;Yc(aa,da)|0;k[V>>2]=aa;k[V+4>>2]=ba;Xw(ea,597390,V)|0;if(Ax(ea,782848,7)|0){_=k[25982]|0;X=xx(_)|0;ka=xx(ea)|0;if((ka|0)>0){Y=0;do{i[_+(Y+X)>>0]=i[ea+Y>>0]|0;Y=Y+1|0}while((Y|0)!=(ka|0));if((ka|0)<80)la=33}else la=33;if((la|0)==33){la=0;sy(_+(ka+X)|0,32,80-ka|0)|0}i[_+(X+80)>>0]=0}if((c|0)<(k[Z>>2]|0))c=c+1|0;else break a}Xc(k[$>>2]|0)|0}while(0);Z=k[25982]|0;V=xx(Z)|0;i[Z+V>>0]=69;i[Z+(V+1)>>0]=78;i[Z+(V+2)>>0]=68;ba=Z+(V+3)|0;aa=ba+77|0;do{i[ba>>0]=32;ba=ba+1|0}while((ba|0)<(aa|0));i[Z+(V+80)>>0]=0;if(Fh(k[ga>>2]|0,$)|0)Xc(k[$>>2]|0)|0}else{if(ja){Tw(597401)|0;zw(k[31231]|0)|0}ja=Dw(a,774853)|0;$=(ja|0)==0;if(!b)if($){b=k[25412]|0;k[U>>2]=a;Ew(b,597416,U)|0;zw(k[25412]|0)|0;gc(1)}else{ww(ja)|0;U=k[25412]|0;k[T>>2]=a;Ew(U,597465,T)|0;zw(k[25412]|0)|0;gc(1)}if($){$=k[25412]|0;k[S>>2]=a;Ew($,597416,S)|0;zw(k[25412]|0)|0;gc(1)}b:do if(Bw(ea,1024,ja)|0){while(1){S=ea+((xx(ea)|0)+-1)|0;if((i[S>>0]|0)==10)i[S>>0]=0;S=ea+((xx(ea)|0)+-1)|0;if((i[S>>0]|0)==13)i[S>>0]=0;ux(fa,ea)|0;if((xx(ea)|0)>80)break;S=fa+(xx(fa)|0)|0;$=fa;while(1){a=i[$>>0]|0;if($>>>0<S>>>0&a<<24>>24==32)$=$+1|0;else{ma=a;na=$;break}}$=ma;X=na;c:while(1){switch($<<24>>24){case 61:case 32:{oa=$;pa=X;break c;break}default:{}}if(X>>>0>=S>>>0){oa=$;pa=X;break}_=X+1|0;$=i[_>>0]|0;X=_}X=oa;$=pa;d:while(1){switch(X<<24>>24){case 39:case 32:case 61:break;default:{qa=$;break d}}if($>>>0>=S>>>0){qa=$;break}_=$+1|0;X=i[_>>0]|0;$=_}i[pa>>0]=0;$=(i[qa>>0]|0)==39?qa+1|0:qa;e:while(1){switch(i[$>>0]|0){case 39:case 32:{ra=$;break e;break}default:{}}if($>>>0>=S>>>0){ra=$;break}$=$+1|0}i[ra>>0]=0;Yc(na,qa)|0;$=k[25982]|0;S=xx($)|0;X=xx(ea)|0;if((X|0)>0){_=0;do{i[$+(_+S)>>0]=i[ea+_>>0]|0;_=_+1|0}while((_|0)!=(X|0));if((X|0)<80)la=70}else la=70;if((la|0)==70){la=0;sy($+(X+S)|0,32,80-X|0)|0}i[$+(S+80)>>0]=0;_=k[25982]|0;ka=(xx(_)|0)+160|0;a=k[25983]|0;if((ka|0)>(a|0)){ka=a+8e4|0;k[25983]=ka;a=fy(_,ka)|0;k[25982]=a}if(!(Bw(ea,1024,ja)|0))break b}Pw(597525,85,1,k[25412]|0)|0;zw(k[25412]|0)|0;gc(1)}while(0);ww(ja)|0}do if(!(k[25960]|0))if(!(k[25955]|0)){ja=k[25412]|0;k[R>>2]=597644;Ew(ja,597611,R)|0;zw(k[25412]|0)|0;gc(1)}else{ja=k[25935]|0;k[Q>>2]=597644;Ew(ja,787303,Q)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25961]|0))if(!(k[25955]|0)){Q=k[25412]|0;k[P>>2]=597677;Ew(Q,597611,P)|0;zw(k[25412]|0)|0;gc(1)}else{Q=k[25935]|0;k[O>>2]=597677;Ew(Q,787303,O)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25962]|0))if(!(k[25955]|0)){O=k[25412]|0;k[N>>2]=597709;Ew(O,597611,N)|0;zw(k[25412]|0)|0;gc(1)}else{O=k[25935]|0;k[M>>2]=597709;Ew(O,787303,M)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25963]|0))if(!(k[25955]|0)){M=k[25412]|0;k[L>>2]=597742;Ew(M,597611,L)|0;zw(k[25412]|0)|0;gc(1)}else{M=k[25935]|0;k[K>>2]=597742;Ew(M,787303,K)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);if(k[25958]|0){dy(k[25982]|0);k[25983]=0;k[25982]=0;r=d;return 0}do if(!(k[25964]|0))if(!(k[25955]|0)){K=k[25412]|0;k[J>>2]=597775;Ew(K,597611,J)|0;zw(k[25412]|0)|0;gc(1)}else{K=k[25935]|0;k[I>>2]=597775;Ew(K,787303,I)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25965]|0))if(!(k[25955]|0)){I=k[25412]|0;k[H>>2]=597808;Ew(I,597611,H)|0;zw(k[25412]|0)|0;gc(1)}else{I=k[25935]|0;k[G>>2]=597808;Ew(I,787303,G)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25966]|0))if(!(k[25955]|0)){G=k[25412]|0;k[F>>2]=597841;Ew(G,597611,F)|0;zw(k[25412]|0)|0;gc(1)}else{G=k[25935]|0;k[E>>2]=597841;Ew(G,787303,E)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25967]|0))if(!(k[25955]|0)){E=k[25412]|0;k[D>>2]=597874;Ew(E,597611,D)|0;zw(k[25412]|0)|0;gc(1)}else{E=k[25935]|0;k[C>>2]=597874;Ew(E,787303,C)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25974]|0))if(!(k[25955]|0)){C=k[25412]|0;k[B>>2]=597907;Ew(C,597611,B)|0;zw(k[25412]|0)|0;gc(1)}else{C=k[25935]|0;k[A>>2]=597907;Ew(C,787303,A)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25975]|0))if(!(k[25955]|0)){A=k[25412]|0;k[z>>2]=597940;Ew(A,597611,z)|0;zw(k[25412]|0)|0;gc(1)}else{A=k[25935]|0;k[y>>2]=597940;Ew(A,787303,y)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(!(k[25971]|k[25970]|k[25972]|k[25973])){if(!(k[25968]|0))if(!(k[25955]|0)){y=k[25412]|0;k[x>>2]=597973;Ew(y,597611,x)|0;zw(k[25412]|0)|0;gc(1)}else{y=k[25935]|0;k[w>>2]=597973;Ew(y,787303,w)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}if(!(k[25969]|0))if(!(k[25955]|0)){y=k[25412]|0;k[v>>2]=598032;Ew(y,597611,v)|0;zw(k[25412]|0)|0;gc(1)}else{y=k[25935]|0;k[u>>2]=598032;Ew(y,787303,u)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}}while(0);do if((xx(595177)|0)>>>0<8)if(!(k[25955]|0)){u=k[25412]|0;k[t>>2]=598091;Ew(u,597611,t)|0;zw(k[25412]|0)|0;gc(1)}else{u=k[25935]|0;k[s>>2]=598091;Ew(u,787303,s)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if((xx(596201)|0)>>>0<8)if(!(k[25955]|0)){s=k[25412]|0;k[q>>2]=598128;Ew(s,597611,q)|0;zw(k[25412]|0)|0;gc(1)}else{s=k[25935]|0;k[p>>2]=598128;Ew(s,787303,p)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;sa=595177;break}else sa=595177;while(0);f:while(1){p=i[sa>>0]|0;switch(p<<24>>24){case 45:{ta=sa;la=137;break f;break}case 0:{ua=sa;va=p;break f;break}default:{}}sa=sa+1|0}if((la|0)==137)while(1){la=0;sa=ta+1|0;p=i[sa>>0]|0;if(p<<24>>24==45){ta=sa;la=137}else{ua=sa;va=p;break}}ta=596201;g:while(1){p=i[ta>>0]|0;switch(p<<24>>24){case 45:{wa=ta;la=140;break g;break}case 0:{xa=p;ya=ta;break g;break}default:{}}ta=ta+1|0}if((la|0)==140)while(1){la=0;ta=wa+1|0;p=i[ta>>0]|0;if(p<<24>>24==45){wa=ta;la=140}else{xa=p;ya=ta;break}}do if(xa<<24>>24==0|va<<24>>24==0)if(!(k[25955]|0)){la=k[25412]|0;k[o>>2]=598165;Ew(la,597611,o)|0;zw(k[25412]|0)|0;gc(1)}else{la=k[25935]|0;k[n>>2]=598165;Ew(la,787303,n)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(tx(ua,ya)|0)if(!(k[25955]|0)){n=k[25412]|0;k[m>>2]=598213;Ew(n,597611,m)|0;zw(k[25412]|0)|0;gc(1)}else{n=k[25935]|0;k[l>>2]=598213;Ew(n,787303,l)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);do if(k[25981]|0){do if((xx(ua)|0)!=3)if(!(k[25955]|0)){l=k[25412]|0;k[j>>2]=598260;Ew(l,597611,j)|0;zw(k[25412]|0)|0;gc(1)}else{l=k[25935]|0;k[h>>2]=598260;Ew(l,787303,h)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);if((xx(ya)|0)==3)break;if(!(k[25955]|0)){l=k[25412]|0;k[g>>2]=598298;Ew(l,597611,g)|0;zw(k[25412]|0)|0;gc(1)}else{l=k[25935]|0;k[f>>2]=598298;Ew(l,787303,f)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}}while(0);f=md(k[25982]|0)|0;k[25984]=f;Zc(f,0)|0;f=k[25957]|0;if((f|0)>0){g=k[25412]|0;k[e>>2]=f;Ew(g,598336,e)|0;zw(k[25412]|0)|0;gc(1)}else{r=d;return 0}return 0}function Xc(a){a=a|0;var b=0,c=0,d=0;b=r;r=r+48|0;c=b;d=b+8|0;sk(a,d);b=k[25412]|0;k[c>>2]=a;k[c+4>>2]=d;Ew(b,598376,c)|0;gc(1);return 0}function Yc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0.0;c=r;r=r+416|0;d=c+400|0;e=c+392|0;f=c+384|0;g=c+376|0;h=c+368|0;i=c+360|0;j=c+352|0;l=c+344|0;m=c+336|0;n=c+328|0;o=c+320|0;p=c+312|0;q=c+304|0;s=c+296|0;t=c+288|0;u=c+280|0;v=c+272|0;w=c+264|0;x=c+256|0;y=c+248|0;z=c+240|0;A=c+232|0;B=c+224|0;C=c+216|0;D=c+208|0;E=c+200|0;F=c+192|0;G=c+184|0;H=c+176|0;I=c+168|0;J=c+160|0;K=c+152|0;L=c+144|0;M=c+136|0;N=c+128|0;O=c+120|0;P=c+112|0;Q=c+104|0;R=c+96|0;S=c+88|0;T=c+80|0;U=c+72|0;V=c+64|0;W=c+56|0;X=c+48|0;Y=c+40|0;Z=c+32|0;_=c+24|0;$=c+16|0;aa=c+8|0;ba=c;ca=c+404|0;if(k[25956]|0){k[ba>>2]=a;k[ba+4>>2]=b;Rw(598420,ba)|0;zw(k[31231]|0)|0}if(!(tx(a,782785)|0)){k[25959]=1;if(!(tx(b,767080)|0)){r=c;return 0}if(!(tx(b,767082)|0)){r=c;return 0}if(!(k[25955]|0)){ba=k[25412]|0;k[aa>>2]=598445;Ew(ba,597611,aa)|0;zw(k[25412]|0)|0;gc(1)}aa=k[25935]|0;k[$>>2]=598445;Ew(aa,787303,$)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,779656)|0)){k[25960]=1;$=jx(b,ca,0)|0;aa=k[ca>>2]|0;do if(aa>>>0<(b+(xx(b)|0)|0)>>>0)if(!(k[25955]|0)){ba=k[25412]|0;k[_>>2]=598475;Ew(ba,597611,_)|0;zw(k[25412]|0)|0;gc(1)}else{ba=k[25935]|0;k[Z>>2]=598475;Ew(ba,787303,Z)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);switch($|0){case 8:case 16:case 32:case 64:case -64:case -32:{r=c;return 0}default:{}}if(!(k[25955]|0)){$=k[25412]|0;k[Y>>2]=598520;Ew($,597611,Y)|0;zw(k[25412]|0)|0;gc(1)}Y=k[25935]|0;k[X>>2]=598520;Ew(Y,787303,X)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,782713)|0)){k[25961]=1;X=jx(b,ca,0)|0;Y=k[ca>>2]|0;do if(Y>>>0<(b+(xx(b)|0)|0)>>>0)if(!(k[25955]|0)){$=k[25412]|0;k[W>>2]=598582;Ew($,597611,W)|0;zw(k[25412]|0)|0;gc(1)}else{$=k[25935]|0;k[V>>2]=598582;Ew($,787303,V)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);if((X|0)>=2){r=c;return 0}if(!(k[25955]|0)){X=k[25412]|0;k[U>>2]=598626;Ew(X,597611,U)|0;zw(k[25412]|0)|0;gc(1)}U=k[25935]|0;k[T>>2]=598626;Ew(U,787303,T)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,772833)|0)){k[25962]=1;T=jx(b,ca,0)|0;U=k[ca>>2]|0;do if(U>>>0<(b+(xx(b)|0)|0)>>>0)if(!(k[25955]|0)){X=k[25412]|0;k[S>>2]=598668;Ew(X,597611,S)|0;zw(k[25412]|0)|0;gc(1)}else{X=k[25935]|0;k[R>>2]=598668;Ew(X,787303,R)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);if((T|0)>=0){r=c;return 0}if(!(k[25955]|0)){T=k[25412]|0;k[Q>>2]=598713;Ew(T,597611,Q)|0;zw(k[25412]|0)|0;gc(1)}Q=k[25935]|0;k[P>>2]=598713;Ew(Q,787303,P)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,790243)|0)){k[25963]=1;P=jx(b,ca,0)|0;Q=k[ca>>2]|0;do if(Q>>>0<(b+(xx(b)|0)|0)>>>0)if(!(k[25955]|0)){T=k[25412]|0;k[O>>2]=598755;Ew(T,597611,O)|0;zw(k[25412]|0)|0;gc(1)}else{T=k[25935]|0;k[N>>2]=598755;Ew(T,787303,N)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;break}while(0);if((P|0)>=0){r=c;return 0}if(!(k[25955]|0)){P=k[25412]|0;k[M>>2]=598800;Ew(P,597611,M)|0;zw(k[25412]|0)|0;gc(1)}M=k[25935]|0;k[L>>2]=598800;Ew(M,787303,L)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,601396)|0)){k[25958]=1;r=c;return 0}if(!(tx(a,777959)|0)){k[25964]=1;ux(595177,b)|0;r=c;return 0}if(!(tx(a,601069)|0)){k[25965]=1;ux(596201,b)|0;r=c;return 0}if(!(tx(a,777876)|0)){k[25966]=1;+hx(b,ca);L=k[ca>>2]|0;if(L>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){L=k[25412]|0;k[K>>2]=598842;Ew(L,597611,K)|0;zw(k[25412]|0)|0;gc(1)}K=k[25935]|0;k[J>>2]=598842;Ew(K,787303,J)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777883)|0)){k[25967]=1;+hx(b,ca);J=k[ca>>2]|0;if(J>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){J=k[25412]|0;k[I>>2]=598890;Ew(J,597611,I)|0;zw(k[25412]|0)|0;gc(1)}I=k[25935]|0;k[H>>2]=598890;Ew(I,787303,H)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777862)|0)){k[25974]=1;+hx(b,ca);H=k[ca>>2]|0;if(H>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){H=k[25412]|0;k[G>>2]=598938;Ew(H,597611,G)|0;zw(k[25412]|0)|0;gc(1)}G=k[25935]|0;k[F>>2]=598938;Ew(G,787303,F)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777869)|0)){k[25975]=1;+hx(b,ca);F=k[ca>>2]|0;if(F>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){F=k[25412]|0;k[E>>2]=598986;Ew(F,597611,E)|0;zw(k[25412]|0)|0;gc(1)}E=k[25935]|0;k[D>>2]=598986;Ew(E,787303,D)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777890)|0)){k[25968]=1;+hx(b,ca);D=k[ca>>2]|0;if(D>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){D=k[25412]|0;k[C>>2]=599034;Ew(D,597611,C)|0;zw(k[25412]|0)|0;gc(1)}C=k[25935]|0;k[B>>2]=599034;Ew(C,787303,B)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777921)|0)){k[25969]=1;+hx(b,ca);B=k[ca>>2]|0;if(B>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){B=k[25412]|0;k[A>>2]=599082;Ew(B,597611,A)|0;zw(k[25412]|0)|0;gc(1)}A=k[25935]|0;k[z>>2]=599082;Ew(A,787303,z)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777928)|0)){+hx(b,ca);z=k[ca>>2]|0;if(z>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){z=k[25412]|0;k[y>>2]=599130;Ew(z,597611,y)|0;zw(k[25412]|0)|0;gc(1)}y=k[25935]|0;k[x>>2]=599130;Ew(y,787303,x)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777897)|0)){k[25970]=1;+hx(b,ca);x=k[ca>>2]|0;if(x>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){x=k[25412]|0;k[w>>2]=599178;Ew(x,597611,w)|0;zw(k[25412]|0)|0;gc(1)}w=k[25935]|0;k[v>>2]=599178;Ew(w,787303,v)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777909)|0)){k[25971]=1;+hx(b,ca);v=k[ca>>2]|0;if(v>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){v=k[25412]|0;k[u>>2]=599225;Ew(v,597611,u)|0;zw(k[25412]|0)|0;gc(1)}u=k[25935]|0;k[t>>2]=599225;Ew(u,787303,t)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777903)|0)){k[25972]=1;+hx(b,ca);t=k[ca>>2]|0;if(t>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){t=k[25412]|0;k[s>>2]=599225;Ew(t,597611,s)|0;zw(k[25412]|0)|0;gc(1)}s=k[25935]|0;k[q>>2]=599225;Ew(s,787303,q)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,777915)|0)){k[25973]=1;+hx(b,ca);q=k[ca>>2]|0;if(q>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){q=k[25412]|0;k[p>>2]=599272;Ew(q,597611,p)|0;zw(k[25412]|0)|0;gc(1)}p=k[25935]|0;k[o>>2]=599272;Ew(p,787303,o)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,779536)|0)){+hx(b,ca);o=k[ca>>2]|0;if(o>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){o=k[25412]|0;k[n>>2]=599319;Ew(o,597611,n)|0;zw(k[25412]|0)|0;gc(1)}n=k[25935]|0;k[m>>2]=599319;Ew(n,787303,m)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,779507)|0)){+hx(b,ca);m=k[ca>>2]|0;if(m>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){m=k[25412]|0;k[l>>2]=599367;Ew(m,597611,l)|0;zw(k[25412]|0)|0;gc(1)}l=k[25935]|0;k[j>>2]=599367;Ew(l,787303,j)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,779663)|0)){+hx(b,ca);j=k[ca>>2]|0;if(j>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){j=k[25412]|0;k[i>>2]=599414;Ew(j,597611,i)|0;zw(k[25412]|0)|0;gc(1)}i=k[25935]|0;k[h>>2]=599414;Ew(i,787303,h)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(!(tx(a,758683)|0)){+hx(b,ca);h=k[ca>>2]|0;if(h>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(k[25955]|0)){h=k[25412]|0;k[g>>2]=599461;Ew(h,597611,g)|0;zw(k[25412]|0)|0;gc(1)}g=k[25935]|0;k[f>>2]=599461;Ew(g,787303,f)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}if(tx(a,601450)|0){r=c;return 0}da=+hx(b,ca);a=k[ca>>2]|0;if(a>>>0>=(b+(xx(b)|0)|0)>>>0){r=c;return 0}if(!(da<1900.0|da>2050.0)){r=c;return 0}if(!(k[25955]|0)){b=k[25412]|0;k[e>>2]=599508;Ew(b,597611,e)|0;zw(k[25412]|0)|0;gc(1)}e=k[25935]|0;k[d>>2]=599508;Ew(e,787303,d)|0;zw(k[25935]|0)|0;k[25957]=(k[25957]|0)+1;r=c;return 0}function Zc(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0;c=r;r=r+672|0;d=c+664|0;e=c+656|0;f=c+640|0;g=c+624|0;h=c+608|0;i=c+592|0;j=c+576|0;l=c+560|0;m=c+544|0;n=c+528|0;o=c+512|0;q=c+496|0;s=c+480|0;t=c+464|0;u=c+448|0;v=c+432|0;w=c+416|0;x=c+400|0;y=c+384|0;z=c+368|0;A=c+352|0;B=c+336|0;C=c+320|0;D=c+304|0;E=c+288|0;F=c+272|0;G=c+256|0;H=c+240|0;I=c+232|0;J=c+224|0;K=c+216|0;L=c+208|0;M=c+200|0;N=c+184|0;O=c+168|0;P=c+152|0;Q=c+136|0;R=c+120|0;S=c+112|0;T=c+104|0;U=c+88|0;V=c+80|0;W=c+72|0;X=c+56|0;Y=c+48|0;Z=c+40|0;_=c+32|0;$=c+24|0;aa=c+16|0;ba=c+8|0;ca=c;if(!a)if(!b){Pw(599557,69,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if(!(k[25985]|0))ea=a+3316|0;else{k[ca>>2]=k[a+3260>>2];Rw(599627,ca)|0;k[ba>>2]=a+3467;Rw(599644,ba)|0;p[aa>>3]=+p[a+136>>3];Rw(599663,aa)|0;p[$>>3]=+p[a+144>>3];Rw(599681,$)|0;k[_>>2]=a+3449;Rw(599699,_)|0;k[Z>>2]=a+3458;Rw(599718,Z)|0;Z=a+3316|0;k[Y>>2]=k[Z>>2];Rw(599737,Y)|0;if((k[Z>>2]|0)>0){Y=0;do{fa=+p[a+688+(Y<<3)>>3];k[X>>2]=Y;p[X+8>>3]=fa;Rw(599754,X)|0;Y=Y+1|0}while((Y|0)<(k[Z>>2]|0))}p[W>>3]=+p[a>>3];Rw(599773,W)|0;p[V>>3]=+p[a+8>>3];Rw(599791,V)|0;if((k[Z>>2]|0)>0){V=0;do{fa=+p[a+616+(V<<3)>>3];k[U>>2]=V;p[U+8>>3]=fa;Rw(599809,U)|0;V=V+1|0}while((V|0)<(k[Z>>2]|0))}p[T>>3]=+p[a+16>>3];Rw(599828,T)|0;p[S>>3]=+p[a+24>>3];Rw(599846,S)|0;if(!(k[a+3300>>2]|0)){if((k[Z>>2]|0)>0){S=0;do{fa=+p[a+760+(S<<3)>>3];k[N>>2]=S;p[N+8>>3]=fa;Rw(599883,N)|0;S=S+1|0}while((S|0)<(k[Z>>2]|0))}}else{fa=+p[a+56>>3];k[R>>2]=0;p[R+8>>3]=fa;Rw(599864,R)|0;fa=+p[a+64>>3];k[Q>>2]=1;p[Q+8>>3]=fa;Rw(599864,Q)|0;fa=+p[a+72>>3];k[P>>2]=2;p[P+8>>3]=fa;Rw(599864,P)|0;fa=+p[a+80>>3];k[O>>2]=3;p[O+8>>3]=fa;Rw(599864,O)|0}p[M>>3]=+p[a+32>>3];Rw(599902,M)|0;p[L>>3]=+p[a+40>>3];Rw(599920,L)|0;p[K>>3]=+p[a+48>>3];Rw(599938,K)|0;p[J>>3]=+p[a+120>>3];Rw(599956,J)|0;p[I>>3]=+p[a+128>>3];Rw(599974,I)|0;fa=+p[a+832>>3];k[H>>2]=0;p[H+8>>3]=fa;Rw(599992,H)|0;fa=+p[a+840>>3];k[G>>2]=1;p[G+8>>3]=fa;Rw(599992,G)|0;fa=+p[a+848>>3];k[F>>2]=2;p[F+8>>3]=fa;Rw(599992,F)|0;fa=+p[a+856>>3];k[E>>2]=3;p[E+8>>3]=fa;Rw(599992,E)|0;fa=+p[a+864>>3];k[D>>2]=4;p[D+8>>3]=fa;Rw(599992,D)|0;fa=+p[a+872>>3];k[C>>2]=5;p[C+8>>3]=fa;Rw(599992,C)|0;fa=+p[a+880>>3];k[B>>2]=6;p[B+8>>3]=fa;Rw(599992,B)|0;fa=+p[a+888>>3];k[A>>2]=7;p[A+8>>3]=fa;Rw(599992,A)|0;fa=+p[a+896>>3];k[z>>2]=8;p[z+8>>3]=fa;Rw(599992,z)|0;fa=+p[a+904>>3];k[y>>2]=9;p[y+8>>3]=fa;Rw(599992,y)|0;fa=+p[a+912>>3];k[x>>2]=10;p[x+8>>3]=fa;Rw(599992,x)|0;fa=+p[a+920>>3];k[w>>2]=11;p[w+8>>3]=fa;Rw(599992,w)|0;fa=+p[a+928>>3];k[v>>2]=12;p[v+8>>3]=fa;Rw(599992,v)|0;fa=+p[a+936>>3];k[u>>2]=13;p[u+8>>3]=fa;Rw(599992,u)|0;fa=+p[a+944>>3];k[t>>2]=14;p[t+8>>3]=fa;Rw(599992,t)|0;fa=+p[a+952>>3];k[s>>2]=15;p[s+8>>3]=fa;Rw(599992,s)|0;fa=+p[a+1480>>3];k[q>>2]=0;p[q+8>>3]=fa;Rw(600011,q)|0;fa=+p[a+1488>>3];k[o>>2]=1;p[o+8>>3]=fa;Rw(600011,o)|0;fa=+p[a+1496>>3];k[n>>2]=2;p[n+8>>3]=fa;Rw(600011,n)|0;fa=+p[a+1504>>3];k[m>>2]=3;p[m+8>>3]=fa;Rw(600011,m)|0;fa=+p[a+1512>>3];k[l>>2]=4;p[l+8>>3]=fa;Rw(600011,l)|0;fa=+p[a+1520>>3];k[j>>2]=5;p[j+8>>3]=fa;Rw(600011,j)|0;fa=+p[a+1528>>3];k[i>>2]=6;p[i+8>>3]=fa;Rw(600011,i)|0;fa=+p[a+1536>>3];k[h>>2]=7;p[h+8>>3]=fa;Rw(600011,h)|0;fa=+p[a+1544>>3];k[g>>2]=8;p[g+8>>3]=fa;Rw(600011,g)|0;fa=+p[a+1552>>3];k[f>>2]=9;p[f+8>>3]=fa;Rw(600011,f)|0;p[e>>3]=+p[a+3176>>3];Rw(600030,e)|0;p[d>>3]=+p[a+3184>>3];Rw(600048,d)|0;ea=Z}if((k[ea>>2]|0)<2)if(!b){Pw(600066,63,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if((k[a+3260>>2]|0)<1)if(!b){Pw(600130,48,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if(+p[a+136>>3]<=0.0)if(!b){Pw(600179,44,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if(+p[a+144>>3]<=0.0)if(!b){Pw(600224,44,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}ea=a+3449|0;do if(!(tx(ea,606317)|0)){if(tx(a+3458|0,601933)|0)if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}else{if(!(tx(ea,601933)|0)){if(!(tx(a+3458|0,606317)|0))break;if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}if(!(tx(ea,600329)|0)){if(!(tx(a+3458|0,600334)|0))break;if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}if(!(tx(ea,600334)|0)){if(!(tx(a+3458|0,600329)|0))break;if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}if(!(tx(ea,600339)|0)){if(!(tx(a+3458|0,600344)|0))break;if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}if(tx(ea,600344)|0)if(!b){Pw(600349,44,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if(tx(a+3458|0,600339)|0)if(!b){Pw(600269,59,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}while(0);if(!(k[a+3300>>2]|0)){if(+p[a+32>>3]==0.0)if(!b){Pw(600442,44,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}if(!(+p[a+40>>3]==0.0)){da=0;r=c;return da|0}if(!b){Pw(600487,44,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}else{if(!(+p[a+56>>3]==0.0?+p[a+64>>3]==0.0:0)){if(!(+p[a+72>>3]==0.0)){da=0;r=c;return da|0}if(!(+p[a+80>>3]==0.0)){da=0;r=c;return da|0}}if(!b){Pw(600394,47,1,k[25412]|0)|0;gc(1)}else{da=1;r=c;return da|0}}return 0}function _c(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0,x=0.0,y=0.0,z=0.0,A=0.0;f=r;r=r+3264|0;g=f;p[c>>3]=0.0;p[d>>3]=0.0;if((k[e+6672>>2]|0)!=1){h=-1;r=f;return h|0}if((k[e+6664>>2]|0)>0){py(g|0,e+136|0,3264)|0;dd(a,b,g,c,d);i=+p[c>>3];j=+p[d>>3]}else{i=a;j=b}b=+p[e+8>>3];a=+p[e+16>>3];if(!(k[e+6676>>2]|0)){l=+p[e+88>>3]*(i-b);m=+p[e+40>>3];n=+p[e+96>>3]*(j-a);o=+p[e+48>>3];q=m*l+o*n;s=m*n-l*o}else{o=i-b;b=j-a;q=+p[e+6680>>3]*o+ +p[e+6688>>3]*b;s=o*+p[e+6696>>3]+b*+p[e+6704>>3]}b=q*.0174532925199433;do if(tx(e,606498)|0){if(!(tx(e,606506)|0)){q=b*b+s*s*.0174532925199433*.0174532925199433;if(q>1.0){h=2;r=f;return h|0}if(q<1.0)t=+X(+(1.0-q));else t=0.0;u=b/t;v=s/t;break}if(!(tx(e,606518)|0)){q=(b*b+s*s*.0174532925199433*.0174532925199433)*.25;if(q>.5){h=2;r=f;return h|0}else{o=+X(+(1.0-q))*.5/(1.0-q*2.0);u=b*o;v=s*o;break}}if(!(tx(e,606502)|0)){o=(b*b+s*s*.0174532925199433*.0174532925199433)*.25;if(!(o>=1.0)){q=1.0-o;u=b*q;v=s*q;break}else{h=2;r=f;return h|0}}if(tx(e,606510)|0){if(!(+p[e+72>>3]-b*+p[e+80>>3]<=0.0)){u=b;v=s;break}else h=2;r=f;return h|0}q=s*s+b*b/3.0461741978670873e-04;if(!(q<=0.0)){o=+X(+q);q=+$(+o)/o;u=b*q;v=s*q;break}else{h=2;r=f;return h|0}}else{u=b;v=s}while(0);s=+p[e+72>>3];b=+p[e+80>>3];t=s-u*b;q=v/t;v=(b+u*s)/t/.0174532925199433;w=e+4|0;do if(tx(w,606498)|0){if(!(tx(w,606506)|0)){t=1.0/+X(+((q*q+v*v)*.0174532925199433*.0174532925199433+1.0));x=v*t;y=q*t;break}if(!(tx(w,606518)|0)){t=(q*q+v*v)*.0174532925199433*.0174532925199433;if(!(t>1.0e-23)){x=v;y=q;break}s=+X(+((1.0-1.0/+X(+(t+1.0)))*2.0))/+X(+t);x=v*s;y=q*s;break}if(!(tx(w,606502)|0)){s=1.0/(+X(+((q*q+v*v)*.0174532925199433*.0174532925199433+1.0))+1.0);x=v*s;y=q*s;break}if(!(tx(w,606510)|0)){s=(q*q+v*v)*.0174532925199433*.0174532925199433;if(!(s<=0.0)){t=+X(+s);s=+ca(+t)/t;x=v*s;y=q*s;break}else{h=2;r=f;return h|0}}else{x=v;y=q}}else{x=v;y=q}while(0);if(!(k[e+6744>>2]|0)){q=+p[e+56>>3];v=+p[e+64>>3];z=(x*q-y*v)/+p[e+104>>3];A=(y*q+x*v)/+p[e+112>>3]}else{z=x*+p[e+6784>>3]+y*+p[e+6792>>3];A=x*+p[e+6800>>3]+y*+p[e+6808>>3]}y=z+ +p[e+24>>3];z=A+ +p[e+32>>3];if((k[e+6668>>2]|0)>0){py(g|0,e+3400|0,3264)|0;cd(y,z,g,c,d)}else{p[c>>3]=y;p[d>>3]=z}z=+p[c>>3];if(z<.5){h=1;r=f;return h|0}if(z>+(k[e+124>>2]|0)+.5){h=1;r=f;return h|0}z=+p[d>>3];if(z<.5){h=1;r=f;return h|0}if(z>+(k[e+132>>2]|0)+.5){h=1;r=f;return h|0}h=0;r=f;return h|0}function $c(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,s=0.0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0;f=r;r=r+3264|0;g=f;p[c>>3]=0.0;p[d>>3]=0.0;if((k[e+6672>>2]|0)!=1){h=-1;r=f;return h|0}if((k[e+6668>>2]|0)>0){py(g|0,e+3400|0,3264)|0;dd(a,b,g,c,d);i=+p[c>>3];j=+p[d>>3]}else{i=a;j=b}b=+p[e+24>>3];a=+p[e+32>>3];if(!(k[e+6744>>2]|0)){l=+p[e+104>>3]*(i-b);m=+p[e+56>>3];n=+p[e+112>>3]*(j-a);o=+p[e+64>>3];q=m*l+o*n;s=m*n-l*o}else{o=i-b;b=j-a;q=+p[e+6752>>3]*o+ +p[e+6760>>3]*b;s=o*+p[e+6768>>3]+b*+p[e+6776>>3]}b=q*.0174532925199433;t=e+4|0;do if(tx(t,606498)|0){if(!(tx(t,606506)|0)){q=b*b+s*s*.0174532925199433*.0174532925199433;if(q>1.0){h=2;r=f;return h|0}if(q<1.0)u=+X(+(1.0-q));else u=0.0;v=b/u;w=s/u;break}if(!(tx(t,606518)|0)){q=(b*b+s*s*.0174532925199433*.0174532925199433)*.25;if(q>.5){h=2;r=f;return h|0}else{o=+X(+(1.0-q))*.5/(1.0-q*2.0);v=b*o;w=s*o;break}}if(!(tx(t,606502)|0)){o=(b*b+s*s*.0174532925199433*.0174532925199433)*.25;if(!(o>=1.0)){q=1.0-o;v=b*q;w=s*q;break}else{h=2;r=f;return h|0}}if(!(tx(t,606510)|0)){q=s*s+b*b/3.0461741978670873e-04;if(!(q<=0.0)){o=+X(+q);q=+$(+o)/o;v=b*q;w=s*q;break}else{h=2;r=f;return h|0}}else{v=b;w=s}}else{v=b;w=s}while(0);s=+p[e+72>>3];b=+p[e+80>>3];u=s+v*b;if(u<=0.0){h=2;r=f;return h|0}q=w/u;w=(v*s-b)/u/.0174532925199433;do if(tx(e,606498)|0){if(!(tx(e,606506)|0)){u=1.0/+X(+((q*q+w*w)*.0174532925199433*.0174532925199433+1.0));x=w*u;y=q*u;break}if(!(tx(e,606518)|0)){u=(q*q+w*w)*.0174532925199433*.0174532925199433;if(!(u>1.0e-23)){x=w;y=q;break}b=+X(+((1.0-1.0/+X(+(u+1.0)))*2.0))/+X(+u);x=w*b;y=q*b;break}if(!(tx(e,606502)|0)){b=1.0/(+X(+((q*q+w*w)*.0174532925199433*.0174532925199433+1.0))+1.0);x=w*b;y=q*b;break}if(!(tx(e,606510)|0)){b=(q*q+w*w)*.0174532925199433*.0174532925199433;if(!(b<=0.0)){u=+X(+b);b=+ca(+u)/u;x=w*b;y=q*b;break}else{h=2;r=f;return h|0}}else{x=w;y=q}}else{x=w;y=q}while(0);if(!(k[e+6676>>2]|0)){q=+p[e+40>>3];w=+p[e+48>>3];z=(x*q-y*w)/+p[e+88>>3];A=(y*q+x*w)/+p[e+96>>3]}else{z=x*+p[e+6712>>3]+y*+p[e+6720>>3];A=x*+p[e+6728>>3]+y*+p[e+6736>>3]}y=z+ +p[e+8>>3];z=A+ +p[e+16>>3];if((k[e+6664>>2]|0)>0){py(g|0,e+136|0,3264)|0;cd(y,z,g,c,d)}else{p[c>>3]=y;p[d>>3]=z}z=+p[c>>3];if(z<.5){h=1;r=f;return h|0}if(z>+(k[e+120>>2]|0)+.5){h=1;r=f;return h|0}z=+p[d>>3];if(z<.5){h=1;r=f;return h|0}if(z>+(k[e+128>>2]|0)+.5){h=1;r=f;return h|0}h=0;r=f;return h|0}
function Eg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0;if(!a){c=-2;return c|0}d=a+28|0;e=k[d>>2]|0;if(b>>>0>5|(e|0)==0){c=-2;return c|0}f=a+12|0;do if(k[f>>2]|0){if((k[a>>2]|0)==0?(k[a+4>>2]|0)!=0:0)break;g=e+4|0;h=k[g>>2]|0;m=(b|0)!=4;if(!(m&(h|0)==666)){n=a+16|0;if(!(k[n>>2]|0)){k[a+24>>2]=k[28254];c=-5;return c|0}k[e>>2]=a;o=e+40|0;p=k[o>>2]|0;k[o>>2]=b;do if((h|0)==42){if((k[e+24>>2]|0)!=2){q=(k[e+48>>2]<<12)+-30720|0;if((k[e+136>>2]|0)<=1?(r=k[e+132>>2]|0,(r|0)>=2):0)if((r|0)<6)s=64;else s=(r|0)==6?128:192;else s=0;r=s|q;q=e+108|0;t=(k[q>>2]|0)==0?r:r|32;k[g>>2]=113;r=e+20|0;u=k[r>>2]|0;k[r>>2]=u+1;v=e+8|0;i[(k[v>>2]|0)+u>>0]=t>>>8;u=k[r>>2]|0;k[r>>2]=u+1;i[(k[v>>2]|0)+u>>0]=(t|((t>>>0)%31|0))^31;t=a+48|0;if(k[q>>2]|0){q=k[t>>2]|0;u=k[r>>2]|0;k[r>>2]=u+1;i[(k[v>>2]|0)+u>>0]=q>>>24;u=k[r>>2]|0;k[r>>2]=u+1;i[(k[v>>2]|0)+u>>0]=q>>>16;q=k[t>>2]|0;u=k[r>>2]|0;k[r>>2]=u+1;i[(k[v>>2]|0)+u>>0]=q>>>8;u=k[r>>2]|0;k[r>>2]=u+1;i[(k[v>>2]|0)+u>>0]=q}q=zg(0,0,0)|0;k[t>>2]=q;w=k[g>>2]|0;x=30;break}q=Ag(0,0,0)|0;t=a+48|0;k[t>>2]=q;q=e+20|0;u=k[q>>2]|0;k[q>>2]=u+1;v=e+8|0;i[(k[v>>2]|0)+u>>0]=31;u=k[q>>2]|0;k[q>>2]=u+1;i[(k[v>>2]|0)+u>>0]=-117;u=k[q>>2]|0;k[q>>2]=u+1;i[(k[v>>2]|0)+u>>0]=8;u=e+28|0;r=k[u>>2]|0;if(!r){y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=0;y=k[e+132>>2]|0;if((y|0)==9)z=2;else z=((y|0)<2?1:(k[e+136>>2]|0)>1)?4:0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=z;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=3;k[g>>2]=113;break}y=(((k[r+44>>2]|0)!=0?2:0)|(k[r>>2]|0)!=0|((k[r+16>>2]|0)==0?0:4)|((k[r+28>>2]|0)==0?0:8)|((k[r+36>>2]|0)==0?0:16))&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=k[(k[u>>2]|0)+4>>2]&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=(k[(k[u>>2]|0)+4>>2]|0)>>>8&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=(k[(k[u>>2]|0)+4>>2]|0)>>>16&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=(k[(k[u>>2]|0)+4>>2]|0)>>>24&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=k[e+132>>2]|0;if((y|0)==9)A=2;else A=((y|0)<2?1:(k[e+136>>2]|0)>1)?4:0;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=A;y=k[(k[u>>2]|0)+12>>2]&255;r=k[q>>2]|0;k[q>>2]=r+1;i[(k[v>>2]|0)+r>>0]=y;y=k[u>>2]|0;if(!(k[y+16>>2]|0))B=y;else{r=k[y+20>>2]&255;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=r;r=(k[(k[u>>2]|0)+20>>2]|0)>>>8&255;y=k[q>>2]|0;k[q>>2]=y+1;i[(k[v>>2]|0)+y>>0]=r;B=k[u>>2]|0}if(k[B+44>>2]|0){r=Ag(k[t>>2]|0,k[v>>2]|0,k[q>>2]|0)|0;k[t>>2]=r}k[e+32>>2]=0;k[g>>2]=69;C=u;x=32}else{w=h;x=30}while(0);if((x|0)==30)if((w|0)==69){C=e+28|0;x=32}else{D=w;x=53}do if((x|0)==32){h=k[C>>2]|0;if(!(k[h+16>>2]|0)){k[g>>2]=73;E=C;F=h;x=55;break}u=e+20|0;r=k[u>>2]|0;t=e+32|0;q=k[t>>2]|0;a:do if(q>>>0<(k[h+20>>2]&65535)>>>0){v=e+12|0;y=a+48|0;G=e+8|0;H=a+20|0;I=q;J=r;K=h;L=r;while(1){if((J|0)==(k[v>>2]|0)){if(J>>>0>L>>>0&(k[K+44>>2]|0)!=0){M=Ag(k[y>>2]|0,(k[G>>2]|0)+L|0,J-L|0)|0;k[y>>2]=M}M=k[d>>2]|0;N=k[M+20>>2]|0;O=k[n>>2]|0;P=N>>>0>O>>>0?O:N;if((P|0)!=0?(py(k[f>>2]|0,k[M+16>>2]|0,P|0)|0,k[f>>2]=(k[f>>2]|0)+P,M=k[d>>2]|0,N=M+16|0,k[N>>2]=(k[N>>2]|0)+P,k[H>>2]=(k[H>>2]|0)+P,k[n>>2]=(k[n>>2]|0)-P,O=M+20|0,Q=k[O>>2]|0,k[O>>2]=Q-P,(Q|0)==(P|0)):0)k[N>>2]=k[M+8>>2];M=k[u>>2]|0;if((M|0)==(k[v>>2]|0)){R=M;break}S=k[C>>2]|0;T=k[t>>2]|0;U=M;V=M}else{S=K;T=I;U=J;V=L}M=i[(k[S+16>>2]|0)+T>>0]|0;k[u>>2]=U+1;i[(k[G>>2]|0)+U>>0]=M;M=(k[t>>2]|0)+1|0;k[t>>2]=M;N=k[C>>2]|0;if(M>>>0>=(k[N+20>>2]&65535)>>>0){W=N;X=V;break a}I=M;J=k[u>>2]|0;K=N;L=V}W=k[C>>2]|0;X=R}else{W=h;X=r}while(0);if((k[W+44>>2]|0)!=0?(r=k[u>>2]|0,r>>>0>X>>>0):0){h=a+48|0;q=Ag(k[h>>2]|0,(k[e+8>>2]|0)+X|0,r-X|0)|0;k[h>>2]=q;Y=k[C>>2]|0}else Y=W;if((k[t>>2]|0)==(k[Y+20>>2]|0)){k[t>>2]=0;k[g>>2]=73;E=C;F=Y;x=55;break}else{D=k[g>>2]|0;x=53;break}}while(0);if((x|0)==53)if((D|0)==73){q=e+28|0;E=q;F=k[q>>2]|0;x=55}else{Z=D;x=73}do if((x|0)==55){if(!(k[F+28>>2]|0)){k[g>>2]=91;_=E;x=75;break}q=e+20|0;h=k[q>>2]|0;r=e+12|0;L=a+48|0;K=e+8|0;J=a+20|0;I=e+32|0;G=h;v=h;while(1){if((G|0)==(k[r>>2]|0)){if(G>>>0>v>>>0?(k[(k[E>>2]|0)+44>>2]|0)!=0:0){h=Ag(k[L>>2]|0,(k[K>>2]|0)+v|0,G-v|0)|0;k[L>>2]=h}h=k[d>>2]|0;H=k[h+20>>2]|0;y=k[n>>2]|0;N=H>>>0>y>>>0?y:H;if((N|0)!=0?(py(k[f>>2]|0,k[h+16>>2]|0,N|0)|0,k[f>>2]=(k[f>>2]|0)+N,h=k[d>>2]|0,H=h+16|0,k[H>>2]=(k[H>>2]|0)+N,k[J>>2]=(k[J>>2]|0)+N,k[n>>2]=(k[n>>2]|0)-N,y=h+20|0,M=k[y>>2]|0,k[y>>2]=M-N,(M|0)==(N|0)):0)k[H>>2]=k[h+8>>2];h=k[q>>2]|0;if((h|0)==(k[r>>2]|0)){$=h;aa=1;break}else{ba=h;ca=h}}else{ba=G;ca=v}h=k[I>>2]|0;k[I>>2]=h+1;H=i[(k[(k[E>>2]|0)+28>>2]|0)+h>>0]|0;k[q>>2]=ba+1;i[(k[K>>2]|0)+ba>>0]=H;if(!(H<<24>>24)){$=ca;aa=H&255;break}G=k[q>>2]|0;v=ca}if((k[(k[E>>2]|0)+44>>2]|0)!=0?(v=k[q>>2]|0,v>>>0>$>>>0):0){G=Ag(k[L>>2]|0,(k[K>>2]|0)+$|0,v-$|0)|0;k[L>>2]=G}if(!aa){k[I>>2]=0;k[g>>2]=91;_=E;x=75;break}else{Z=k[g>>2]|0;x=73;break}}while(0);if((x|0)==73)if((Z|0)==91){_=e+28|0;x=75}else{da=Z;x=93}do if((x|0)==75){if(!(k[(k[_>>2]|0)+36>>2]|0)){k[g>>2]=103;ea=_;x=95;break}G=e+20|0;v=k[G>>2]|0;r=e+12|0;J=a+48|0;t=e+8|0;u=a+20|0;H=e+32|0;h=v;N=v;while(1){if((h|0)==(k[r>>2]|0)){if(h>>>0>N>>>0?(k[(k[_>>2]|0)+44>>2]|0)!=0:0){v=Ag(k[J>>2]|0,(k[t>>2]|0)+N|0,h-N|0)|0;k[J>>2]=v}v=k[d>>2]|0;M=k[v+20>>2]|0;y=k[n>>2]|0;P=M>>>0>y>>>0?y:M;if((P|0)!=0?(py(k[f>>2]|0,k[v+16>>2]|0,P|0)|0,k[f>>2]=(k[f>>2]|0)+P,v=k[d>>2]|0,M=v+16|0,k[M>>2]=(k[M>>2]|0)+P,k[u>>2]=(k[u>>2]|0)+P,k[n>>2]=(k[n>>2]|0)-P,y=v+20|0,Q=k[y>>2]|0,k[y>>2]=Q-P,(Q|0)==(P|0)):0)k[M>>2]=k[v+8>>2];v=k[G>>2]|0;if((v|0)==(k[r>>2]|0)){fa=v;ga=1;break}else{ha=v;ia=v}}else{ha=h;ia=N}v=k[H>>2]|0;k[H>>2]=v+1;M=i[(k[(k[_>>2]|0)+36>>2]|0)+v>>0]|0;k[G>>2]=ha+1;i[(k[t>>2]|0)+ha>>0]=M;if(!(M<<24>>24)){fa=ia;ga=M&255;break}h=k[G>>2]|0;N=ia}if((k[(k[_>>2]|0)+44>>2]|0)!=0?(N=k[G>>2]|0,N>>>0>fa>>>0):0){h=Ag(k[J>>2]|0,(k[t>>2]|0)+fa|0,N-fa|0)|0;k[J>>2]=h}if(!ga){k[g>>2]=103;ea=_;x=95;break}else{da=k[g>>2]|0;x=93;break}}while(0);if((x|0)==93?(da|0)==103:0){ea=e+28|0;x=95}do if((x|0)==95){if(!(k[(k[ea>>2]|0)+44>>2]|0)){k[g>>2]=113;break}h=e+20|0;N=e+12|0;if((((k[h>>2]|0)+2|0)>>>0>(k[N>>2]|0)>>>0?(H=k[d>>2]|0,r=k[H+20>>2]|0,u=k[n>>2]|0,I=r>>>0>u>>>0?u:r,(I|0)!=0):0)?(py(k[f>>2]|0,k[H+16>>2]|0,I|0)|0,k[f>>2]=(k[f>>2]|0)+I,H=k[d>>2]|0,r=H+16|0,k[r>>2]=(k[r>>2]|0)+I,u=a+20|0,k[u>>2]=(k[u>>2]|0)+I,k[n>>2]=(k[n>>2]|0)-I,u=H+20|0,L=k[u>>2]|0,k[u>>2]=L-I,(L|0)==(I|0)):0)k[r>>2]=k[H+8>>2];H=k[h>>2]|0;if((H+2|0)>>>0<=(k[N>>2]|0)>>>0){N=a+48|0;r=k[N>>2]&255;k[h>>2]=H+1;I=e+8|0;i[(k[I>>2]|0)+H>>0]=r;r=(k[N>>2]|0)>>>8&255;H=k[h>>2]|0;k[h>>2]=H+1;i[(k[I>>2]|0)+H>>0]=r;r=Ag(0,0,0)|0;k[N>>2]=r;k[g>>2]=113}}while(0);r=e+20|0;if(!(k[r>>2]|0)){if(m&(p|0)>=(b|0)&(k[a+4>>2]|0)==0){k[a+24>>2]=k[28254];c=-5;return c|0}}else{N=k[d>>2]|0;H=k[N+20>>2]|0;I=k[n>>2]|0;h=H>>>0>I>>>0?I:H;if(h){py(k[f>>2]|0,k[N+16>>2]|0,h|0)|0;k[f>>2]=(k[f>>2]|0)+h;N=k[d>>2]|0;H=N+16|0;k[H>>2]=(k[H>>2]|0)+h;L=a+20|0;k[L>>2]=(k[L>>2]|0)+h;L=(k[n>>2]|0)-h|0;k[n>>2]=L;u=N+20|0;K=k[u>>2]|0;k[u>>2]=K-h;if((K|0)==(h|0)){k[H>>2]=k[N+8>>2];ja=L}else ja=L}else ja=I;if(!ja){k[o>>2]=-1;c=0;return c|0}}I=(k[g>>2]|0)==666;L=(k[a+4>>2]|0)==0;if(I)if(L)x=115;else{k[a+24>>2]=k[28254];c=-5;return c|0}else if(L)x=115;else x=116;if((x|0)==115?!((k[e+116>>2]|0)==0&((b|0)==0|I)):0)x=116;do if((x|0)==116){b:do switch(k[e+136>>2]|0){case 2:{I=e+116|0;L=e+96|0;N=e+108|0;H=e+56|0;h=e+5792|0;K=e+5796|0;u=e+5784|0;q=e+5788|0;M=e+92|0;while(1){if((k[I>>2]|0)==0?(Fg(e),(k[I>>2]|0)==0):0)break;k[L>>2]=0;v=i[(k[H>>2]|0)+(k[N>>2]|0)>>0]|0;P=k[h>>2]|0;j[(k[K>>2]|0)+(P<<1)>>1]=0;k[h>>2]=P+1;i[(k[u>>2]|0)+P>>0]=v;P=e+148+((v&255)<<2)|0;j[P>>1]=(j[P>>1]|0)+1<<16>>16;P=(k[h>>2]|0)==((k[q>>2]|0)+-1|0);k[I>>2]=(k[I>>2]|0)+-1;v=(k[N>>2]|0)+1|0;k[N>>2]=v;if(!P)continue;P=k[M>>2]|0;if((P|0)>-1)ka=(k[H>>2]|0)+P|0;else ka=0;ih(e,ka,v-P|0,0);k[M>>2]=k[N>>2];P=k[e>>2]|0;v=P+28|0;Q=k[v>>2]|0;y=k[Q+20>>2]|0;O=P+16|0;la=k[O>>2]|0;ma=y>>>0>la>>>0?la:y;if((ma|0)!=0?(y=P+12|0,py(k[y>>2]|0,k[Q+16>>2]|0,ma|0)|0,k[y>>2]=(k[y>>2]|0)+ma,y=k[v>>2]|0,v=y+16|0,k[v>>2]=(k[v>>2]|0)+ma,Q=P+20|0,k[Q>>2]=(k[Q>>2]|0)+ma,k[O>>2]=(k[O>>2]|0)-ma,O=y+20|0,Q=k[O>>2]|0,k[O>>2]=Q-ma,(Q|0)==(ma|0)):0)k[v>>2]=k[y+8>>2];if(!(k[(k[e>>2]|0)+16>>2]|0))break b}if(b){I=k[M>>2]|0;if((I|0)>-1)na=(k[H>>2]|0)+I|0;else na=0;q=(b|0)==4;ih(e,na,(k[N>>2]|0)-I|0,q&1);k[M>>2]=k[N>>2];I=k[e>>2]|0;h=I+28|0;u=k[h>>2]|0;K=k[u+20>>2]|0;L=I+16|0;y=k[L>>2]|0;v=K>>>0>y>>>0?y:K;if((v|0)!=0?(K=I+12|0,py(k[K>>2]|0,k[u+16>>2]|0,v|0)|0,k[K>>2]=(k[K>>2]|0)+v,K=k[h>>2]|0,h=K+16|0,k[h>>2]=(k[h>>2]|0)+v,u=I+20|0,k[u>>2]=(k[u>>2]|0)+v,k[L>>2]=(k[L>>2]|0)-v,L=K+20|0,u=k[L>>2]|0,k[L>>2]=u-v,(u|0)==(v|0)):0)k[h>>2]=k[K+8>>2];if(!(k[(k[e>>2]|0)+16>>2]|0)){oa=q?2:0;x=175;break b}else{oa=q?3:1;x=175;break b}}break}case 3:{q=e+116|0;K=(b|0)==0;h=e+96|0;v=e+108|0;u=e+5792|0;L=e+5796|0;I=e+5784|0;y=e+2440+(l[607512]<<2)|0;ma=e+5788|0;Q=e+56|0;O=e+92|0;while(1){P=k[q>>2]|0;if(P>>>0<258){Fg(e);la=k[q>>2]|0;if(K&la>>>0<258)break b;if(!la)break;k[h>>2]=0;if(la>>>0>2){pa=la;x=143}else{qa=k[v>>2]|0;x=158}}else{k[h>>2]=0;pa=P;x=143}if((x|0)==143){x=0;P=k[v>>2]|0;if(P){la=k[Q>>2]|0;ra=i[la+(P+-1)>>0]|0;if((ra<<24>>24==(i[la+P>>0]|0)?ra<<24>>24==(i[la+(P+1)>>0]|0):0)?(sa=la+(P+2)|0,ra<<24>>24==(i[sa>>0]|0)):0){ta=la+(P+258)|0;la=sa;while(1){sa=la+1|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+2|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+3|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+4|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+5|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+6|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+7|0;if(ra<<24>>24!=(i[sa>>0]|0)){ua=sa;break}sa=la+8|0;if(sa>>>0<ta>>>0?ra<<24>>24==(i[sa>>0]|0):0)la=sa;else{ua=sa;break}}la=ua-ta+258|0;ra=la>>>0>pa>>>0?pa:la;k[h>>2]=ra;if(ra>>>0>2){la=ra+253|0;ra=k[u>>2]|0;j[(k[L>>2]|0)+(ra<<1)>>1]=1;k[u>>2]=ra+1;i[(k[I>>2]|0)+ra>>0]=la;ra=e+148+((l[608024+(la&255)>>0]|256)+1<<2)|0;j[ra>>1]=(j[ra>>1]|0)+1<<16>>16;j[y>>1]=(j[y>>1]|0)+1<<16>>16;ra=(k[u>>2]|0)==((k[ma>>2]|0)+-1|0)&1;la=k[h>>2]|0;k[q>>2]=(k[q>>2]|0)-la;sa=(k[v>>2]|0)+la|0;k[v>>2]=sa;k[h>>2]=0;va=sa;wa=ra}else{qa=P;x=158}}else{qa=P;x=158}}else{qa=0;x=158}}if((x|0)==158){x=0;ra=i[(k[Q>>2]|0)+qa>>0]|0;sa=k[u>>2]|0;j[(k[L>>2]|0)+(sa<<1)>>1]=0;k[u>>2]=sa+1;i[(k[I>>2]|0)+sa>>0]=ra;sa=e+148+((ra&255)<<2)|0;j[sa>>1]=(j[sa>>1]|0)+1<<16>>16;sa=(k[u>>2]|0)==((k[ma>>2]|0)+-1|0)&1;k[q>>2]=(k[q>>2]|0)+-1;ra=(k[v>>2]|0)+1|0;k[v>>2]=ra;va=ra;wa=sa}if(!wa)continue;sa=k[O>>2]|0;if((sa|0)>-1)xa=(k[Q>>2]|0)+sa|0;else xa=0;ih(e,xa,va-sa|0,0);k[O>>2]=k[v>>2];sa=k[e>>2]|0;ra=sa+28|0;la=k[ra>>2]|0;ya=k[la+20>>2]|0;za=sa+16|0;Aa=k[za>>2]|0;Ba=ya>>>0>Aa>>>0?Aa:ya;if((Ba|0)!=0?(ya=sa+12|0,py(k[ya>>2]|0,k[la+16>>2]|0,Ba|0)|0,k[ya>>2]=(k[ya>>2]|0)+Ba,ya=k[ra>>2]|0,ra=ya+16|0,k[ra>>2]=(k[ra>>2]|0)+Ba,la=sa+20|0,k[la>>2]=(k[la>>2]|0)+Ba,k[za>>2]=(k[za>>2]|0)-Ba,za=ya+20|0,la=k[za>>2]|0,k[za>>2]=la-Ba,(la|0)==(Ba|0)):0)k[ra>>2]=k[ya+8>>2];if(!(k[(k[e>>2]|0)+16>>2]|0))break b}q=k[O>>2]|0;if((q|0)>-1)Ca=(k[Q>>2]|0)+q|0;else Ca=0;ma=(b|0)==4;ih(e,Ca,(k[v>>2]|0)-q|0,ma&1);k[O>>2]=k[v>>2];q=k[e>>2]|0;u=q+28|0;I=k[u>>2]|0;L=k[I+20>>2]|0;h=q+16|0;y=k[h>>2]|0;K=L>>>0>y>>>0?y:L;if((K|0)!=0?(L=q+12|0,py(k[L>>2]|0,k[I+16>>2]|0,K|0)|0,k[L>>2]=(k[L>>2]|0)+K,L=k[u>>2]|0,u=L+16|0,k[u>>2]=(k[u>>2]|0)+K,I=q+20|0,k[I>>2]=(k[I>>2]|0)+K,k[h>>2]=(k[h>>2]|0)-K,h=L+20|0,I=k[h>>2]|0,k[h>>2]=I-K,(I|0)==(K|0)):0)k[u>>2]=k[L+8>>2];if(!(k[(k[e>>2]|0)+16>>2]|0)){oa=ma?2:0;x=175;break b}else{oa=ma?3:1;x=175;break b}break}default:{oa=Xb[k[112260+((k[e+132>>2]|0)*12|0)+8>>2]&31](e,b)|0;x=175}}while(0);if((x|0)==175){if((oa&-2|0)==2)k[g>>2]=666;if(oa&-3){if((oa|0)!=1)break;switch(b|0){case 1:{hh(e);break}case 5:break;default:{gh(e,0,0,0);if((b|0)==3?(J=k[e+76>>2]|0,t=k[e+68>>2]|0,j[t+(J+-1<<1)>>1]=0,sy(t|0,0,(J<<1)+-2|0)|0,(k[e+116>>2]|0)==0):0){k[e+108>>2]=0;k[e+92>>2]=0}}}J=k[d>>2]|0;t=k[J+20>>2]|0;G=k[n>>2]|0;ma=t>>>0>G>>>0?G:t;if(ma){py(k[f>>2]|0,k[J+16>>2]|0,ma|0)|0;k[f>>2]=(k[f>>2]|0)+ma;J=k[d>>2]|0;t=J+16|0;k[t>>2]=(k[t>>2]|0)+ma;L=a+20|0;k[L>>2]=(k[L>>2]|0)+ma;L=(k[n>>2]|0)-ma|0;k[n>>2]=L;u=J+20|0;K=k[u>>2]|0;k[u>>2]=K-ma;if((K|0)==(ma|0)){k[t>>2]=k[J+8>>2];Da=L}else Da=L}else Da=G;if(Da)break;k[o>>2]=-1;c=0;return c|0}}if(k[n>>2]|0){c=0;return c|0}k[o>>2]=-1;c=0;return c|0}while(0);if(m){c=0;return c|0}o=e+24|0;g=k[o>>2]|0;if((g|0)<1){c=1;return c|0}p=a+48|0;G=k[p>>2]|0;if((g|0)==2){g=k[r>>2]|0;k[r>>2]=g+1;L=e+8|0;i[(k[L>>2]|0)+g>>0]=G;g=(k[p>>2]|0)>>>8&255;J=k[r>>2]|0;k[r>>2]=J+1;i[(k[L>>2]|0)+J>>0]=g;g=(k[p>>2]|0)>>>16&255;J=k[r>>2]|0;k[r>>2]=J+1;i[(k[L>>2]|0)+J>>0]=g;g=(k[p>>2]|0)>>>24&255;J=k[r>>2]|0;k[r>>2]=J+1;i[(k[L>>2]|0)+J>>0]=g;g=a+8|0;J=k[g>>2]&255;t=k[r>>2]|0;k[r>>2]=t+1;i[(k[L>>2]|0)+t>>0]=J;J=(k[g>>2]|0)>>>8&255;t=k[r>>2]|0;k[r>>2]=t+1;i[(k[L>>2]|0)+t>>0]=J;J=(k[g>>2]|0)>>>16&255;t=k[r>>2]|0;k[r>>2]=t+1;i[(k[L>>2]|0)+t>>0]=J;J=(k[g>>2]|0)>>>24&255;g=k[r>>2]|0;k[r>>2]=g+1;i[(k[L>>2]|0)+g>>0]=J}else{J=k[r>>2]|0;k[r>>2]=J+1;g=e+8|0;i[(k[g>>2]|0)+J>>0]=G>>>24;J=k[r>>2]|0;k[r>>2]=J+1;i[(k[g>>2]|0)+J>>0]=G>>>16;G=k[p>>2]|0;p=k[r>>2]|0;k[r>>2]=p+1;i[(k[g>>2]|0)+p>>0]=G>>>8;p=k[r>>2]|0;k[r>>2]=p+1;i[(k[g>>2]|0)+p>>0]=G}G=k[d>>2]|0;p=k[G+20>>2]|0;g=k[n>>2]|0;J=p>>>0>g>>>0?g:p;if((J|0)!=0?(py(k[f>>2]|0,k[G+16>>2]|0,J|0)|0,k[f>>2]=(k[f>>2]|0)+J,G=k[d>>2]|0,p=G+16|0,k[p>>2]=(k[p>>2]|0)+J,g=a+20|0,k[g>>2]=(k[g>>2]|0)+J,k[n>>2]=(k[n>>2]|0)-J,g=G+20|0,L=k[g>>2]|0,k[g>>2]=L-J,(L|0)==(J|0)):0)k[p>>2]=k[G+8>>2];G=k[o>>2]|0;if((G|0)>0)k[o>>2]=0-G;c=(k[r>>2]|0)==0&1;return c|0}}while(0);k[a+24>>2]=k[28251];c=-2;return c|0}function Fg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;b=a+44|0;c=k[b>>2]|0;d=a+60|0;e=a+116|0;f=a+108|0;g=c+-262|0;h=a+56|0;i=a+72|0;n=a+88|0;o=a+84|0;p=a+112|0;q=a+92|0;r=a+76|0;s=a+68|0;t=a+64|0;u=k[e>>2]|0;v=c;while(1){w=k[f>>2]|0;x=(k[d>>2]|0)-u-w|0;if(w>>>0<(g+v|0)>>>0){y=w;z=x}else{w=k[h>>2]|0;py(w|0,w+c|0,c|0)|0;k[p>>2]=(k[p>>2]|0)-c;w=(k[f>>2]|0)-c|0;k[f>>2]=w;k[q>>2]=(k[q>>2]|0)-c;A=k[r>>2]|0;B=A;C=(k[s>>2]|0)+(A<<1)|0;do{C=C+-2|0;A=m[C>>1]|0;j[C>>1]=A>>>0<c>>>0?0:A-c&65535;B=B+-1|0}while((B|0)!=0);B=c;C=(k[t>>2]|0)+(c<<1)|0;do{C=C+-2|0;A=m[C>>1]|0;j[C>>1]=A>>>0<c>>>0?0:A-c&65535;B=B+-1|0}while((B|0)!=0);y=w;z=x+c|0}B=k[a>>2]|0;C=B+4|0;A=k[C>>2]|0;if(!A){D=24;break}E=k[e>>2]|0;F=(k[h>>2]|0)+(E+y)|0;G=A>>>0>z>>>0?z:A;if(!G){H=0;I=E}else{k[C>>2]=A-G;switch(k[(k[B+28>>2]|0)+24>>2]|0){case 1:{A=B+48|0;C=zg(k[A>>2]|0,k[B>>2]|0,G)|0;k[A>>2]=C;J=B;break}case 2:{C=B+48|0;A=Ag(k[C>>2]|0,k[B>>2]|0,G)|0;k[C>>2]=A;J=B;break}default:J=B}py(F|0,k[J>>2]|0,G|0)|0;k[J>>2]=(k[J>>2]|0)+G;F=B+8|0;k[F>>2]=(k[F>>2]|0)+G;H=G;I=k[e>>2]|0}G=I+H|0;k[e>>2]=G;if(G>>>0>2?(F=k[f>>2]|0,B=k[h>>2]|0,A=l[B+F>>0]|0,k[i>>2]=A,k[i>>2]=((l[B+(F+1)>>0]|0)^A<<k[n>>2])&k[o>>2],G>>>0>=262):0){K=G;break}if(!(k[(k[a>>2]|0)+4>>2]|0)){K=G;break}u=G;v=k[b>>2]|0}if((D|0)==24)return;D=a+5824|0;a=k[D>>2]|0;b=k[d>>2]|0;if(b>>>0<=a>>>0)return;d=K+(k[f>>2]|0)|0;if(a>>>0<d>>>0){f=b-d|0;K=f>>>0>258?258:f;sy((k[h>>2]|0)+d|0,0,K|0)|0;k[D>>2]=K+d;return}K=d+258|0;if(K>>>0<=a>>>0)return;d=K-a|0;K=b-a|0;b=d>>>0>K>>>0?K:d;sy((k[h>>2]|0)+a|0,0,b|0)|0;k[D>>2]=(k[D>>2]|0)+b;return}function Gg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0;c=(k[a+12>>2]|0)+-5|0;d=c>>>0<65535?c:65535;c=a+116|0;e=a+108|0;f=a+92|0;g=a+44|0;h=a+56|0;while(1){i=k[c>>2]|0;if(i>>>0<2){Fg(a);j=k[c>>2]|0;if(!(j|b)){l=0;m=28;break}if(!j){m=20;break}else n=j}else n=i;i=(k[e>>2]|0)+n|0;k[e>>2]=i;k[c>>2]=0;j=k[f>>2]|0;o=j+d|0;if((i|0)!=0&i>>>0<o>>>0){p=i;q=j}else{k[c>>2]=i-o;k[e>>2]=o;if((j|0)>-1)r=(k[h>>2]|0)+j|0;else r=0;ih(a,r,d,0);k[f>>2]=k[e>>2];j=k[a>>2]|0;o=j+28|0;i=k[o>>2]|0;s=k[i+20>>2]|0;t=j+16|0;u=k[t>>2]|0;v=s>>>0>u>>>0?u:s;if((v|0)!=0?(s=j+12|0,py(k[s>>2]|0,k[i+16>>2]|0,v|0)|0,k[s>>2]=(k[s>>2]|0)+v,s=k[o>>2]|0,o=s+16|0,k[o>>2]=(k[o>>2]|0)+v,i=j+20|0,k[i>>2]=(k[i>>2]|0)+v,k[t>>2]=(k[t>>2]|0)-v,t=s+20|0,i=k[t>>2]|0,k[t>>2]=i-v,(i|0)==(v|0)):0)k[o>>2]=k[s+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){l=0;m=28;break}p=k[e>>2]|0;q=k[f>>2]|0}s=p-q|0;if(s>>>0<((k[g>>2]|0)+-262|0)>>>0)continue;if((q|0)>-1)w=(k[h>>2]|0)+q|0;else w=0;ih(a,w,s,0);k[f>>2]=k[e>>2];s=k[a>>2]|0;o=s+28|0;v=k[o>>2]|0;i=k[v+20>>2]|0;t=s+16|0;j=k[t>>2]|0;u=i>>>0>j>>>0?j:i;if((u|0)!=0?(i=s+12|0,py(k[i>>2]|0,k[v+16>>2]|0,u|0)|0,k[i>>2]=(k[i>>2]|0)+u,i=k[o>>2]|0,o=i+16|0,k[o>>2]=(k[o>>2]|0)+u,v=s+20|0,k[v>>2]=(k[v>>2]|0)+u,k[t>>2]=(k[t>>2]|0)-u,t=i+20|0,v=k[t>>2]|0,k[t>>2]=v-u,(v|0)==(u|0)):0)k[o>>2]=k[i+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){l=0;m=28;break}}if((m|0)==20){w=k[f>>2]|0;if((w|0)>-1)x=(k[h>>2]|0)+w|0;else x=0;h=(b|0)==4;ih(a,x,(k[e>>2]|0)-w|0,h&1);k[f>>2]=k[e>>2];e=k[a>>2]|0;f=e+28|0;w=k[f>>2]|0;x=k[w+20>>2]|0;b=e+16|0;q=k[b>>2]|0;g=x>>>0>q>>>0?q:x;if((g|0)!=0?(x=e+12|0,py(k[x>>2]|0,k[w+16>>2]|0,g|0)|0,k[x>>2]=(k[x>>2]|0)+g,x=k[f>>2]|0,f=x+16|0,k[f>>2]=(k[f>>2]|0)+g,w=e+20|0,k[w>>2]=(k[w>>2]|0)+g,k[b>>2]=(k[b>>2]|0)-g,b=x+20|0,w=k[b>>2]|0,k[b>>2]=w-g,(w|0)==(g|0)):0)k[f>>2]=k[x+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){l=h?2:0;return l|0}else{l=h?3:1;return l|0}}else if((m|0)==28)return l|0;return 0}function Hg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0;c=a+116|0;d=(b|0)==0;e=a+72|0;f=a+88|0;g=a+108|0;h=a+56|0;m=a+84|0;n=a+68|0;o=a+52|0;p=a+64|0;q=a+44|0;r=a+96|0;s=a+112|0;t=a+5792|0;u=a+5796|0;v=a+5784|0;w=a+5788|0;x=a+128|0;y=a+92|0;while(1){if((k[c>>2]|0)>>>0<262){Fg(a);z=k[c>>2]|0;if(d&z>>>0<262){A=0;B=32;break}if(!z){B=24;break}if(z>>>0<=2)B=9;else B=6}else B=6;if((B|0)==6){B=0;z=k[g>>2]|0;C=((l[(k[h>>2]|0)+(z+2)>>0]|0)^k[e>>2]<<k[f>>2])&k[m>>2];k[e>>2]=C;D=(k[n>>2]|0)+(C<<1)|0;C=j[D>>1]|0;j[(k[p>>2]|0)+((k[o>>2]&z)<<1)>>1]=C;E=C&65535;j[D>>1]=z;if(C<<16>>16!=0?(z-E|0)>>>0<=((k[q>>2]|0)+-262|0)>>>0:0){z=Jg(a,E)|0;k[r>>2]=z;F=z}else B=9}if((B|0)==9){B=0;F=k[r>>2]|0}do if(F>>>0>2){z=F+253|0;E=(k[g>>2]|0)-(k[s>>2]|0)|0;C=k[t>>2]|0;j[(k[u>>2]|0)+(C<<1)>>1]=E;k[t>>2]=C+1;i[(k[v>>2]|0)+C>>0]=z;C=a+148+((l[608024+(z&255)>>0]|0|256)+1<<2)|0;j[C>>1]=(j[C>>1]|0)+1<<16>>16;C=E+65535&65535;E=a+2440+((l[607512+(C>>>0<256?C:(C>>>7)+256|0)>>0]|0)<<2)|0;j[E>>1]=(j[E>>1]|0)+1<<16>>16;E=(k[t>>2]|0)==((k[w>>2]|0)+-1|0)&1;C=k[r>>2]|0;z=(k[c>>2]|0)-C|0;k[c>>2]=z;if(!(z>>>0>2?C>>>0<=(k[x>>2]|0)>>>0:0)){z=(k[g>>2]|0)+C|0;k[g>>2]=z;k[r>>2]=0;D=k[h>>2]|0;G=l[D+z>>0]|0;k[e>>2]=G;k[e>>2]=((l[D+(z+1)>>0]|0)^G<<k[f>>2])&k[m>>2];H=z;I=E;break}z=C+-1|0;k[r>>2]=z;C=k[f>>2]|0;G=k[h>>2]|0;D=k[m>>2]|0;J=k[n>>2]|0;K=k[o>>2]|0;L=k[p>>2]|0;M=z;z=k[g>>2]|0;N=k[e>>2]|0;while(1){O=z+1|0;k[g>>2]=O;N=((l[G+(z+3)>>0]|0)^N<<C)&D;k[e>>2]=N;P=J+(N<<1)|0;j[L+((K&O)<<1)>>1]=j[P>>1]|0;j[P>>1]=O;M=M+-1|0;k[r>>2]=M;if(!M){Q=z;break}else z=O}z=Q+2|0;k[g>>2]=z;H=z;I=E}else{z=i[(k[h>>2]|0)+(k[g>>2]|0)>>0]|0;M=k[t>>2]|0;j[(k[u>>2]|0)+(M<<1)>>1]=0;k[t>>2]=M+1;i[(k[v>>2]|0)+M>>0]=z;M=a+148+((z&255)<<2)|0;j[M>>1]=(j[M>>1]|0)+1<<16>>16;M=(k[t>>2]|0)==((k[w>>2]|0)+-1|0)&1;k[c>>2]=(k[c>>2]|0)+-1;z=(k[g>>2]|0)+1|0;k[g>>2]=z;H=z;I=M}while(0);if(!I)continue;M=k[y>>2]|0;if((M|0)>-1)R=(k[h>>2]|0)+M|0;else R=0;ih(a,R,H-M|0,0);k[y>>2]=k[g>>2];M=k[a>>2]|0;z=M+28|0;K=k[z>>2]|0;L=k[K+20>>2]|0;N=M+16|0;J=k[N>>2]|0;D=L>>>0>J>>>0?J:L;if((D|0)!=0?(L=M+12|0,py(k[L>>2]|0,k[K+16>>2]|0,D|0)|0,k[L>>2]=(k[L>>2]|0)+D,L=k[z>>2]|0,z=L+16|0,k[z>>2]=(k[z>>2]|0)+D,K=M+20|0,k[K>>2]=(k[K>>2]|0)+D,k[N>>2]=(k[N>>2]|0)-D,N=L+20|0,K=k[N>>2]|0,k[N>>2]=K-D,(K|0)==(D|0)):0)k[z>>2]=k[L+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){A=0;B=32;break}}if((B|0)==24){H=k[y>>2]|0;if((H|0)>-1)S=(k[h>>2]|0)+H|0;else S=0;h=(b|0)==4;ih(a,S,(k[g>>2]|0)-H|0,h&1);k[y>>2]=k[g>>2];g=k[a>>2]|0;y=g+28|0;H=k[y>>2]|0;S=k[H+20>>2]|0;b=g+16|0;R=k[b>>2]|0;I=S>>>0>R>>>0?R:S;if((I|0)!=0?(S=g+12|0,py(k[S>>2]|0,k[H+16>>2]|0,I|0)|0,k[S>>2]=(k[S>>2]|0)+I,S=k[y>>2]|0,y=S+16|0,k[y>>2]=(k[y>>2]|0)+I,H=g+20|0,k[H>>2]=(k[H>>2]|0)+I,k[b>>2]=(k[b>>2]|0)-I,b=S+20|0,H=k[b>>2]|0,k[b>>2]=H-I,(H|0)==(I|0)):0)k[y>>2]=k[S+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){A=h?2:0;return A|0}else{A=h?3:1;return A|0}}else if((B|0)==32)return A|0;return 0}function Ig(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;c=a+116|0;d=(b|0)==0;e=a+72|0;f=a+88|0;g=a+108|0;h=a+56|0;m=a+84|0;n=a+68|0;o=a+52|0;p=a+64|0;q=a+96|0;r=a+120|0;s=a+112|0;t=a+100|0;u=a+5792|0;v=a+5796|0;w=a+5784|0;x=a+5788|0;y=a+104|0;z=a+92|0;A=a+128|0;B=a+44|0;C=a+136|0;a:while(1){D=k[c>>2]|0;while(1){if(D>>>0<262){Fg(a);E=k[c>>2]|0;if(d&E>>>0<262){F=0;G=48;break a}if(!E){G=38;break a}if(E>>>0<=2){k[r>>2]=k[q>>2];k[t>>2]=k[s>>2];k[q>>2]=2;H=2;G=16}else G=8}else G=8;do if((G|0)==8){G=0;E=k[g>>2]|0;I=((l[(k[h>>2]|0)+(E+2)>>0]|0)^k[e>>2]<<k[f>>2])&k[m>>2];k[e>>2]=I;J=(k[n>>2]|0)+(I<<1)|0;I=j[J>>1]|0;j[(k[p>>2]|0)+((k[o>>2]&E)<<1)>>1]=I;K=I&65535;j[J>>1]=E;E=k[q>>2]|0;k[r>>2]=E;k[t>>2]=k[s>>2];k[q>>2]=2;if(I<<16>>16)if(E>>>0<(k[A>>2]|0)>>>0)if(((k[g>>2]|0)-K|0)>>>0<=((k[B>>2]|0)+-262|0)>>>0){I=Jg(a,K)|0;k[q>>2]=I;if(I>>>0<6){if((k[C>>2]|0)!=1){if((I|0)!=3){H=I;G=16;break}if(((k[g>>2]|0)-(k[s>>2]|0)|0)>>>0<=4096){H=3;G=16;break}}k[q>>2]=2;H=2;G=16}else{H=I;G=16}}else{H=2;G=16}else{L=E;M=2}else{H=2;G=16}}while(0);if((G|0)==16){G=0;L=k[r>>2]|0;M=H}if(!(L>>>0<3|M>>>0>L>>>0)){N=L;break}if(!(k[y>>2]|0)){k[y>>2]=1;k[g>>2]=(k[g>>2]|0)+1;E=(k[c>>2]|0)+-1|0;k[c>>2]=E;D=E;continue}E=i[(k[h>>2]|0)+((k[g>>2]|0)+-1)>>0]|0;I=k[u>>2]|0;j[(k[v>>2]|0)+(I<<1)>>1]=0;k[u>>2]=I+1;i[(k[w>>2]|0)+I>>0]=E;I=a+148+((E&255)<<2)|0;j[I>>1]=(j[I>>1]|0)+1<<16>>16;if((k[u>>2]|0)==((k[x>>2]|0)+-1|0)){I=k[z>>2]|0;if((I|0)>-1)O=(k[h>>2]|0)+I|0;else O=0;ih(a,O,(k[g>>2]|0)-I|0,0);k[z>>2]=k[g>>2];I=k[a>>2]|0;E=I+28|0;K=k[E>>2]|0;J=k[K+20>>2]|0;P=I+16|0;Q=k[P>>2]|0;R=J>>>0>Q>>>0?Q:J;if((R|0)!=0?(J=I+12|0,py(k[J>>2]|0,k[K+16>>2]|0,R|0)|0,k[J>>2]=(k[J>>2]|0)+R,J=k[E>>2]|0,E=J+16|0,k[E>>2]=(k[E>>2]|0)+R,K=I+20|0,k[K>>2]=(k[K>>2]|0)+R,k[P>>2]=(k[P>>2]|0)-R,P=J+20|0,K=k[P>>2]|0,k[P>>2]=K-R,(K|0)==(R|0)):0)k[E>>2]=k[J+8>>2]}k[g>>2]=(k[g>>2]|0)+1;D=(k[c>>2]|0)+-1|0;k[c>>2]=D;if(!(k[(k[a>>2]|0)+16>>2]|0)){F=0;G=48;break a}}D=k[g>>2]|0;J=D+-3+(k[c>>2]|0)|0;E=N+253|0;R=D+65535-(k[t>>2]|0)|0;D=k[u>>2]|0;j[(k[v>>2]|0)+(D<<1)>>1]=R;k[u>>2]=D+1;i[(k[w>>2]|0)+D>>0]=E;D=a+148+((l[608024+(E&255)>>0]|0|256)+1<<2)|0;j[D>>1]=(j[D>>1]|0)+1<<16>>16;D=R+65535&65535;R=a+2440+((l[607512+(D>>>0<256?D:(D>>>7)+256|0)>>0]|0)<<2)|0;j[R>>1]=(j[R>>1]|0)+1<<16>>16;R=k[u>>2]|0;D=(k[x>>2]|0)+-1|0;E=k[r>>2]|0;k[c>>2]=1-E+(k[c>>2]|0);K=E+-2|0;k[r>>2]=K;E=k[g>>2]|0;P=K;while(1){K=E+1|0;k[g>>2]=K;if(K>>>0<=J>>>0){I=((l[(k[h>>2]|0)+(E+3)>>0]|0)^k[e>>2]<<k[f>>2])&k[m>>2];k[e>>2]=I;Q=(k[n>>2]|0)+(I<<1)|0;j[(k[p>>2]|0)+((k[o>>2]&K)<<1)>>1]=j[Q>>1]|0;j[Q>>1]=K}P=P+-1|0;k[r>>2]=P;if(!P){S=E;break}else E=K}k[y>>2]=0;k[q>>2]=2;E=S+2|0;k[g>>2]=E;if((R|0)!=(D|0))continue;P=k[z>>2]|0;if((P|0)>-1)T=(k[h>>2]|0)+P|0;else T=0;ih(a,T,E-P|0,0);k[z>>2]=k[g>>2];P=k[a>>2]|0;E=P+28|0;J=k[E>>2]|0;K=k[J+20>>2]|0;Q=P+16|0;I=k[Q>>2]|0;U=K>>>0>I>>>0?I:K;if((U|0)!=0?(K=P+12|0,py(k[K>>2]|0,k[J+16>>2]|0,U|0)|0,k[K>>2]=(k[K>>2]|0)+U,K=k[E>>2]|0,E=K+16|0,k[E>>2]=(k[E>>2]|0)+U,J=P+20|0,k[J>>2]=(k[J>>2]|0)+U,k[Q>>2]=(k[Q>>2]|0)-U,Q=K+20|0,J=k[Q>>2]|0,k[Q>>2]=J-U,(J|0)==(U|0)):0)k[E>>2]=k[K+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){F=0;G=48;break}}if((G|0)==38){if(k[y>>2]|0){T=i[(k[h>>2]|0)+((k[g>>2]|0)+-1)>>0]|0;S=k[u>>2]|0;j[(k[v>>2]|0)+(S<<1)>>1]=0;k[u>>2]=S+1;i[(k[w>>2]|0)+S>>0]=T;S=a+148+((T&255)<<2)|0;j[S>>1]=(j[S>>1]|0)+1<<16>>16;k[y>>2]=0}y=k[z>>2]|0;if((y|0)>-1)V=(k[h>>2]|0)+y|0;else V=0;h=(b|0)==4;ih(a,V,(k[g>>2]|0)-y|0,h&1);k[z>>2]=k[g>>2];g=k[a>>2]|0;z=g+28|0;y=k[z>>2]|0;V=k[y+20>>2]|0;b=g+16|0;S=k[b>>2]|0;T=V>>>0>S>>>0?S:V;if((T|0)!=0?(V=g+12|0,py(k[V>>2]|0,k[y+16>>2]|0,T|0)|0,k[V>>2]=(k[V>>2]|0)+T,V=k[z>>2]|0,z=V+16|0,k[z>>2]=(k[z>>2]|0)+T,y=g+20|0,k[y>>2]=(k[y>>2]|0)+T,k[b>>2]=(k[b>>2]|0)-T,b=V+20|0,y=k[b>>2]|0,k[b>>2]=y-T,(y|0)==(T|0)):0)k[z>>2]=k[V+8>>2];if(!(k[(k[a>>2]|0)+16>>2]|0)){F=h?2:0;return F|0}else{F=h?3:1;return F|0}}else if((G|0)==48)return F|0;return 0}function Jg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0;c=k[a+124>>2]|0;d=k[a+56>>2]|0;e=k[a+108>>2]|0;f=d+e|0;g=k[a+120>>2]|0;h=k[a+144>>2]|0;j=(k[a+44>>2]|0)+-262|0;l=e>>>0>j>>>0?e-j|0:0;j=k[a+64>>2]|0;n=k[a+52>>2]|0;o=d+(e+258)|0;p=k[a+116>>2]|0;q=h>>>0>p>>>0?p:h;h=a+112|0;r=d+(e+1)|0;s=d+(e+2)|0;t=o;u=e+257|0;v=b;b=g;w=g>>>0<(k[a+140>>2]|0)>>>0?c:c>>>2;c=i[d+(g+e)>>0]|0;a=i[d+(e+-1+g)>>0]|0;while(1){g=d+v|0;if((((i[d+(v+b)>>0]|0)==c<<24>>24?(i[d+(b+-1+v)>>0]|0)==a<<24>>24:0)?(i[g>>0]|0)==(i[f>>0]|0):0)?(i[d+(v+1)>>0]|0)==(i[r>>0]|0):0){g=d+(v+2)|0;x=s;while(1){y=x+1|0;if((i[y>>0]|0)!=(i[g+1>>0]|0)){z=y;break}y=x+2|0;if((i[y>>0]|0)!=(i[g+2>>0]|0)){z=y;break}y=x+3|0;if((i[y>>0]|0)!=(i[g+3>>0]|0)){z=y;break}y=x+4|0;if((i[y>>0]|0)!=(i[g+4>>0]|0)){z=y;break}y=x+5|0;if((i[y>>0]|0)!=(i[g+5>>0]|0)){z=y;break}y=x+6|0;if((i[y>>0]|0)!=(i[g+6>>0]|0)){z=y;break}y=x+7|0;if((i[y>>0]|0)!=(i[g+7>>0]|0)){z=y;break}y=x+8|0;g=g+8|0;if(!(y>>>0<o>>>0?(i[y>>0]|0)==(i[g>>0]|0):0)){z=y;break}else x=y}x=z-t|0;g=x+258|0;if((g|0)>(b|0)){k[h>>2]=v;if((g|0)>=(q|0)){A=g;B=20;break}C=g;D=i[d+(g+e)>>0]|0;E=i[d+(u+x)>>0]|0}else{C=b;D=c;E=a}}else{C=b;D=c;E=a}v=m[j+((v&n)<<1)>>1]|0;if(v>>>0<=l>>>0){A=C;B=20;break}w=w+-1|0;if(!w){A=C;B=20;break}else{b=C;c=D;a=E}}if((B|0)==20)return (A>>>0>p>>>0?p:A)|0;return 0}function Kg(a){a=a|0;var b=0;do if(a)if((k[a>>2]|0)==7247){b=Rg(a)|0;break}else{b=Wg(a)|0;break}else b=-2;while(0);return b|0}function Lg(a,b){a=a|0;b=b|0;return Pg(a,-1,b)|0}function Mg(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;if(!a){d=-1;return d|0}e=k[a>>2]|0;a:do if((e|0)<31153){switch(e|0){case 7247:{break a;break}default:d=-1}return d|0}else{switch(e|0){case 31153:{break a;break}default:d=-1}return d|0}while(0);f=a+76|0;if(c>>>0>1|(k[f>>2]|0)!=0){d=-1;return d|0}if(c){c=a+72|0;if(!(k[c>>2]|0)){g=b;h=c}else{g=(k[a+68>>2]|0)+b|0;h=c}}else{g=b-(k[a+12>>2]|0)|0;h=a+72|0}k[h>>2]=0;b=(e|0)==7247;if((b?(k[a+52>>2]|0)==1:0)?(c=a+12|0,((k[c>>2]|0)+g|0)>=(k[a+48>>2]|0)):0){i=a+36|0;if((Lx(k[a+4>>2]|0,g-(k[i>>2]|0)|0,1)|0)==-1){d=-1;return d|0}k[i>>2]=0;k[a+40>>2]=0;k[h>>2]=0;i=a+80|0;j=k[i>>2]|0;if(j){if((k[f>>2]|0)!=-4)dy(j);k[i>>2]=0}k[f>>2]=0;k[a+88>>2]=0;i=(k[c>>2]|0)+g|0;k[c>>2]=i;d=i;return d|0}if((g|0)<0){if(!b){d=-1;return d|0}b=a+12|0;i=(k[b>>2]|0)+g|0;if((i|0)<0){d=-1;return d|0}if((Lx(k[a+4>>2]|0,k[a+44>>2]|0,0)|0)==-1){d=-1;return d|0}c=k[a>>2]|0;if((c|0)==7247){k[a+36>>2]=0;k[a+40>>2]=0;k[a+52>>2]=0;k[a+56>>2]=1}k[h>>2]=0;j=a+80|0;l=k[j>>2]|0;if(!l)m=c;else{if((k[f>>2]|0)==-4)n=c;else{dy(l);n=k[a>>2]|0}k[j>>2]=0;m=n}k[f>>2]=0;k[b>>2]=0;k[a+88>>2]=0;o=i;p=m}else{o=g;p=e}if((p|0)==7247){p=a+36|0;e=k[p>>2]|0;g=(e|0)<0|(e|0)>(o|0)?o:e;k[p>>2]=e-g;e=a+32|0;k[e>>2]=(k[e>>2]|0)+g;e=a+12|0;k[e>>2]=g+(k[e>>2]|0);q=o-g|0}else q=o;if(q){k[h>>2]=1;k[a+68>>2]=q}d=(k[a+12>>2]|0)+q|0;return d|0}function Ng(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;d=a+80|0;e=k[d>>2]|0;f=a+76|0;if(e){if((k[f>>2]|0)!=-4)dy(e);k[d>>2]=0}k[f>>2]=b;if(!c)return;if((b|0)==-4){k[d>>2]=c;return}b=k[a+8>>2]|0;a=xx(b)|0;e=cy(a+3+(xx(c)|0)|0)|0;k[d>>2]=e;if(!e){k[f>>2]=-4;k[d>>2]=607037;return}else{ux(e,b)|0;b=k[d>>2]|0;e=b+(xx(b)|0)|0;i[e>>0]=i[606897]|0;i[e+1>>0]=i[606898]|0;i[e+2>>0]=i[606899]|0;qx(k[d>>2]|0,c)|0;return}}function Og(a,b,c){a=a|0;b=b|0;c=c|0;return Mg(a,b,c)|0}function Pg(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;d=r;r=r+16|0;e=d;f=cy(140)|0;if(!f){g=0;r=d;return g|0}k[f+16>>2]=0;k[f+20>>2]=8192;h=f+80|0;k[h>>2]=0;k[f>>2]=0;j=f+60|0;k[j>>2]=-1;l=f+64|0;k[l>>2]=0;m=i[c>>0]|0;if(m<<24>>24){n=c;c=m;m=0;a:while(1){o=c<<24>>24;b:do if((c+-48&255)<10){k[j>>2]=o+-48;p=m}else switch(o|0){case 43:{q=9;break a;break}case 114:{k[f>>2]=7247;p=7247;break b;break}case 119:{k[f>>2]=31153;p=31153;break b;break}case 97:{k[f>>2]=1;p=1;break b;break}case 70:{k[l>>2]=4;p=m;break b;break}case 102:{k[l>>2]=1;p=m;break b;break}case 104:{k[l>>2]=2;p=m;break b;break}case 82:{k[l>>2]=3;p=m;break b;break}default:{p=m;break b}}while(0);n=n+1|0;c=i[n>>0]|0;if(!(c<<24>>24)){s=p;break}else m=p}if((q|0)==9){dy(f);g=0;r=d;return g|0}if(s){s=cy((xx(a)|0)+1|0)|0;q=f+8|0;k[q>>2]=s;if(!s){dy(f);g=0;r=d;return g|0}ux(s,a)|0;if((b|0)==-1){s=k[f>>2]|0;k[e>>2]=438;p=uv(a,(s|0)==7247?32768:(s|0)==31153?33345:33857,e)|0;k[f+4>>2]=p;if((p|0)==-1){dy(k[q>>2]|0);dy(f);g=0;r=d;return g|0}else t=p}else{k[f+4>>2]=b;t=b}b=k[f>>2]|0;c:do if((b|0)>=7247){switch(b|0){case 7247:break;default:break c}p=Lx(t,0,1)|0;k[f+44>>2]=(p|0)==-1?0:p;if((k[f>>2]|0)==7247){k[f+36>>2]=0;k[f+40>>2]=0;k[f+52>>2]=0;k[f+56>>2]=1}}else{switch(b|0){case 1:break;default:break c}k[f>>2]=31153}while(0);k[f+72>>2]=0;b=k[h>>2]|0;t=f+76|0;if(b){if((k[t>>2]|0)!=-4)dy(b);k[h>>2]=0}k[t>>2]=0;k[f+12>>2]=0;k[f+88>>2]=0;g=f;r=d;return g|0}}dy(f);g=0;r=d;return g|0}function Qg(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0;if(!a){d=-1;return d|0}if((k[a>>2]|0)!=7247){d=-1;return d|0}if(k[a+76>>2]|0){d=-1;return d|0}if((c|0)<0){Ng(a,-5,606961);d=-1;return d|0}if(!c){d=0;return d|0}e=a+72|0;a:do if(k[e>>2]|0){k[e>>2]=0;f=k[a+68>>2]|0;g=a+88|0;h=a+36|0;i=a+40|0;j=a+32|0;l=a+12|0;if(f){m=a+52|0;n=a+28|0;o=a+16|0;p=a+4|0;q=a+100|0;r=a+96|0;s=f;b:while(1){c:while(1){f=k[h>>2]|0;while(1){if(f){t=f;break c}if((k[i>>2]|0)!=0?(k[g>>2]|0)==0:0){u=h;v=i;w=j;x=g;y=l;z=j;break a}A=k[m>>2]|0;if(A){B=A;break}if((Sg(a)|0)==-1){d=-1;C=59;break b}f=k[h>>2]|0;if(!f){C=21;break}}if((C|0)==21){C=0;B=k[m>>2]|0}switch(B|0){case 1:break;case 2:{k[q>>2]=k[o>>2]<<1;k[r>>2]=k[n>>2];if((Tg(a)|0)==-1){d=-1;C=59;break b}else continue c;break}default:continue c}f=k[n>>2]|0;A=k[o>>2]<<1;k[h>>2]=0;D=0;do{E=Mx(k[p>>2]|0,f+D|0,A-D|0)|0;if((E|0)<1){F=E;C=26;break}D=(k[h>>2]|0)+E|0;k[h>>2]=D}while(D>>>0<A>>>0);if((C|0)==26){C=0;if((F|0)<0){C=28;break b}k[i>>2]=1}k[j>>2]=k[n>>2]}A=(t|0)<0|(t|0)>(s|0)?s:t;k[h>>2]=t-A;k[j>>2]=(k[j>>2]|0)+A;k[l>>2]=(k[l>>2]|0)+A;if((s|0)==(A|0)){u=h;v=i;w=j;x=g;y=l;z=j;break a}else s=s-A|0}if((C|0)==28){s=rv()|0;Ng(a,-1,sv(k[s>>2]|0)|0);d=-1;return d|0}else if((C|0)==59)return d|0}else{u=h;v=i;w=j;x=g;y=l;z=j}}else{s=a+32|0;u=a+36|0;v=a+40|0;w=s;x=a+88|0;y=a+12|0;z=s}while(0);t=a+52|0;F=a+28|0;B=a+16|0;e=a+4|0;s=a+100|0;n=a+96|0;p=b;b=c;c=0;d:while(1){o=k[u>>2]|0;e:do if(!o){if((k[v>>2]|0)!=0?(k[x>>2]|0)==0:0){d=c;C=59;break d}r=k[t>>2]|0;if(r)if(b>>>0<k[B>>2]<<1>>>0)G=r;else{if((r|0)==1)H=0;else{k[s>>2]=b;k[n>>2]=p;if((Tg(a)|0)==-1){d=-1;C=59;break d}r=k[u>>2]|0;k[u>>2]=0;I=r;C=57;break}while(1){r=Mx(k[e>>2]|0,p+H|0,b-H|0)|0;if((r|0)<1){J=H;K=r;break}q=r+H|0;if(q>>>0<b>>>0)H=q;else{I=q;C=57;break e}}if((K|0)<0){C=54;break d}k[v>>2]=1;I=J;C=57;break}else{if((Sg(a)|0)==-1){d=-1;C=59;break d}if(k[u>>2]|0){L=p;M=b;N=c;break}G=k[t>>2]|0}switch(G|0){case 1:break;case 2:{k[s>>2]=k[B>>2]<<1;k[n>>2]=k[F>>2];if((Tg(a)|0)==-1){d=-1;C=59;break d}else{L=p;M=b;N=c;break e}break}default:{L=p;M=b;N=c;break e}}q=k[F>>2]|0;r=k[B>>2]<<1;k[u>>2]=0;m=0;do{A=Mx(k[e>>2]|0,q+m|0,r-m|0)|0;if((A|0)<1){O=A;C=44;break}m=(k[u>>2]|0)+A|0;k[u>>2]=m}while(m>>>0<r>>>0);if((C|0)==44){C=0;if((O|0)<0){C=46;break d}k[v>>2]=1}k[w>>2]=k[F>>2];L=p;M=b;N=c}else{r=o>>>0>b>>>0?b:o;py(p|0,k[z>>2]|0,r|0)|0;k[z>>2]=(k[z>>2]|0)+r;k[u>>2]=(k[u>>2]|0)-r;I=r;C=57}while(0);if((C|0)==57){C=0;k[y>>2]=(k[y>>2]|0)+I;L=p+I|0;M=b-I|0;N=I+c|0}if(!M){d=N;C=59;break}else{p=L;b=M;c=N}}if((C|0)==46){N=rv()|0;Ng(a,-1,sv(k[N>>2]|0)|0);d=-1;return d|0}else if((C|0)==54){N=rv()|0;Ng(a,-1,sv(k[N>>2]|0)|0);d=-1;return d|0}else if((C|0)==59)return d|0;return 0}function Rg(a){a=a|0;var b=0,c=0;if(!a){b=-2;return b|0}if((k[a>>2]|0)!=7247){b=-2;return b|0}if(k[a+16>>2]|0){ch(a+84|0)|0;dy(k[a+28>>2]|0);dy(k[a+24>>2]|0)}Ng(a,0,0);dy(k[a+8>>2]|0);c=Hx(k[a+4>>2]|0)|0;dy(a);b=((c|0)!=0)<<31>>31;return b|0}function Sg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0;b=a+84|0;c=a+16|0;do if(!(k[c>>2]|0)){d=k[a+20>>2]|0;e=cy(d)|0;f=a+24|0;k[f>>2]=e;g=cy(d<<1)|0;h=a+28|0;k[h>>2]=g;j=(g|0)==0;if(!(j|(e|0)==0)){k[c>>2]=d;k[a+116>>2]=0;k[a+120>>2]=0;k[a+124>>2]=0;d=a+88|0;k[d>>2]=0;k[b>>2]=0;if(!(ah(b,-15,608415,56)|0)){m=d;break}dy(k[h>>2]|0);dy(k[f>>2]|0);k[c>>2]=0;Ng(a,-4,607037);n=-1;return n|0}if(j)o=e;else{dy(g);o=k[f>>2]|0}if(o)dy(o);Ng(a,-4,607037);n=-1;return n|0}else m=a+88|0;while(0);o=k[m>>2]|0;if(!o){if(k[a+76>>2]|0){n=-1;return n|0}f=a+40|0;if(k[f>>2]|0){n=0;return n|0}g=a+24|0;e=k[g>>2]|0;j=k[c>>2]|0;k[m>>2]=0;h=a+4|0;d=0;while(1){p=Mx(k[h>>2]|0,e+d|0,j-d|0)|0;if((p|0)<1){q=p;r=17;break}s=(k[m>>2]|0)+p|0;k[m>>2]=s;if(s>>>0<j>>>0)d=s;else{t=s;break}}do if((r|0)==17){if((q|0)>=0){k[f>>2]=1;t=k[m>>2]|0;break}d=rv()|0;Ng(a,-1,sv(k[d>>2]|0)|0);n=-1;return n|0}while(0);k[b>>2]=k[g>>2];if(!t){n=0;return n|0}else u=t}else u=o;o=k[b>>2]|0;if((i[o>>0]|0)==31){t=u+-1|0;k[m>>2]=t;g=o+1|0;k[b>>2]=g;if(!t){if(k[a+76>>2]|0){n=-1;return n|0}o=a+40|0;if(!(k[o>>2]|0)){f=a+24|0;q=k[f>>2]|0;d=k[c>>2]|0;k[m>>2]=0;j=a+4|0;e=0;while(1){h=Mx(k[j>>2]|0,q+e|0,d-e|0)|0;if((h|0)<1){v=h;r=29;break}s=(k[m>>2]|0)+h|0;k[m>>2]=s;if(s>>>0<d>>>0)e=s;else{w=s;break}}do if((r|0)==29){if((v|0)>=0){k[o>>2]=1;w=k[m>>2]|0;break}e=rv()|0;Ng(a,-1,sv(k[e>>2]|0)|0);n=-1;return n|0}while(0);o=k[f>>2]|0;k[b>>2]=o;if(w){x=o;y=w;r=33}}}else{x=g;y=t;r=33}if((r|0)==33?(i[x>>0]|0)==-117:0){t=y+-1|0;k[m>>2]=t;y=x+1|0;k[b>>2]=y;a:do if(!t){if((k[a+76>>2]|0)==0?(x=a+40|0,(k[x>>2]|0)==0):0){g=a+24|0;w=k[g>>2]|0;o=k[c>>2]|0;k[m>>2]=0;f=a+4|0;v=0;while(1){e=Mx(k[f>>2]|0,w+v|0,o-v|0)|0;if((e|0)<1){z=e;r=40;break}d=(k[m>>2]|0)+e|0;k[m>>2]=d;if(d>>>0<o>>>0)v=d;else{A=d;break}}do if((r|0)==40)if((z|0)<0){v=rv()|0;Ng(a,-1,sv(k[v>>2]|0)|0);break a}else{k[x>>2]=1;A=k[m>>2]|0;break}while(0);x=k[g>>2]|0;k[b>>2]=x;if(A){B=A;C=x;r=44}}}else{B=t;C=y;r=44}while(0);if((r|0)==44?(y=B+-1|0,k[m>>2]=y,B=C+1|0,k[b>>2]=B,(i[C>>0]|0)==8):0){b:do if(!y){if((k[a+76>>2]|0)==0?(C=a+40|0,(k[C>>2]|0)==0):0){t=a+24|0;A=k[t>>2]|0;z=k[c>>2]|0;k[m>>2]=0;x=a+4|0;v=0;while(1){o=Mx(k[x>>2]|0,A+v|0,z-v|0)|0;if((o|0)<1){D=o;r=52;break}w=(k[m>>2]|0)+o|0;k[m>>2]=w;if(w>>>0<z>>>0)v=w;else{E=w;break}}do if((r|0)==52)if((D|0)<0){v=rv()|0;Ng(a,-1,sv(k[v>>2]|0)|0);break b}else{k[C>>2]=1;E=k[m>>2]|0;break}while(0);C=k[t>>2]|0;k[b>>2]=C;if(E){F=E;G=C;r=56}}}else{F=y;G=B;r=56}while(0);if((r|0)==56?(B=F+-1|0,k[m>>2]=B,F=G+1|0,k[b>>2]=F,y=l[G>>0]|0,(y&224|0)==0):0){c:do if(!B){G=a+76|0;if(!(k[G>>2]|0)){E=a+40|0;if(!(k[E>>2]|0)){D=a+24|0;C=k[D>>2]|0;v=k[c>>2]|0;k[m>>2]=0;z=a+4|0;A=0;while(1){x=Mx(k[z>>2]|0,C+A|0,v-A|0)|0;if((x|0)<1){H=x;r=64;break}g=(k[m>>2]|0)+x|0;k[m>>2]=g;if(g>>>0<v>>>0)A=g;else{I=g;break}}do if((r|0)==64)if((H|0)<0){A=rv()|0;Ng(a,-1,sv(k[A>>2]|0)|0);J=k[m>>2]|0;r=69;break c}else{k[E>>2]=1;I=k[m>>2]|0;break}while(0);E=k[D>>2]|0;k[b>>2]=E;if(I){K=I;L=E;r=68}else{M=G;r=71}}else{M=G;r=71}}else r=70}else{K=B;L=F;r=68}while(0);if((r|0)==68){F=K+-1|0;k[m>>2]=F;k[b>>2]=L+1;J=F;r=69}if((r|0)==69)if(!J)r=70;else{N=J;r=80}if((r|0)==70){M=a+76|0;r=71}d:do if((r|0)==71)if(!(k[M>>2]|0)){J=a+40|0;if(!(k[J>>2]|0)){F=a+24|0;L=k[F>>2]|0;K=k[c>>2]|0;k[m>>2]=0;B=a+4|0;I=0;while(1){H=Mx(k[B>>2]|0,L+I|0,K-I|0)|0;if((H|0)<1){O=H;r=76;break}E=(k[m>>2]|0)+H|0;k[m>>2]=E;if(E>>>0<K>>>0)I=E;else{P=E;break}}do if((r|0)==76)if((O|0)<0){I=rv()|0;Ng(a,-1,sv(k[I>>2]|0)|0);Q=k[m>>2]|0;r=81;break d}else{k[J>>2]=1;P=k[m>>2]|0;break}while(0);k[b>>2]=k[F>>2];if(P){N=P;r=80}else{R=M;r=83}}else{R=M;r=83}}else r=82;while(0);if((r|0)==80){M=N+-1|0;k[m>>2]=M;k[b>>2]=(k[b>>2]|0)+1;Q=M;r=81}if((r|0)==81)if(!Q)r=82;else{S=Q;r=92}if((r|0)==82){R=a+76|0;r=83}e:do if((r|0)==83)if(!(k[R>>2]|0)){Q=a+40|0;if(!(k[Q>>2]|0)){M=a+24|0;N=k[M>>2]|0;P=k[c>>2]|0;k[m>>2]=0;O=a+4|0;J=0;while(1){I=Mx(k[O>>2]|0,N+J|0,P-J|0)|0;if((I|0)<1){T=I;r=88;break}K=(k[m>>2]|0)+I|0;k[m>>2]=K;if(K>>>0<P>>>0)J=K;else{U=K;break}}do if((r|0)==88)if((T|0)<0){J=rv()|0;Ng(a,-1,sv(k[J>>2]|0)|0);V=k[m>>2]|0;r=93;break e}else{k[Q>>2]=1;U=k[m>>2]|0;break}while(0);k[b>>2]=k[M>>2];if(U){S=U;r=92}else{W=R;r=95}}else{W=R;r=95}}else r=94;while(0);if((r|0)==92){R=S+-1|0;k[m>>2]=R;k[b>>2]=(k[b>>2]|0)+1;V=R;r=93}if((r|0)==93)if(!V)r=94;else{X=V;r=104}if((r|0)==94){W=a+76|0;r=95}f:do if((r|0)==95)if(!(k[W>>2]|0)){V=a+40|0;if(!(k[V>>2]|0)){R=a+24|0;S=k[R>>2]|0;U=k[c>>2]|0;k[m>>2]=0;T=a+4|0;Q=0;while(1){J=Mx(k[T>>2]|0,S+Q|0,U-Q|0)|0;if((J|0)<1){Y=J;r=100;break}P=(k[m>>2]|0)+J|0;k[m>>2]=P;if(P>>>0<U>>>0)Q=P;else{Z=P;break}}do if((r|0)==100)if((Y|0)<0){Q=rv()|0;Ng(a,-1,sv(k[Q>>2]|0)|0);_=k[m>>2]|0;r=105;break f}else{k[V>>2]=1;Z=k[m>>2]|0;break}while(0);k[b>>2]=k[R>>2];if(Z){X=Z;r=104}else{$=W;r=107}}else{$=W;r=107}}else r=106;while(0);if((r|0)==104){W=X+-1|0;k[m>>2]=W;k[b>>2]=(k[b>>2]|0)+1;_=W;r=105}if((r|0)==105)if(!_)r=106;else{aa=_;r=116}if((r|0)==106){$=a+76|0;r=107}g:do if((r|0)==107)if(!(k[$>>2]|0)){_=a+40|0;if(!(k[_>>2]|0)){W=a+24|0;X=k[W>>2]|0;Z=k[c>>2]|0;k[m>>2]=0;Y=a+4|0;V=0;while(1){Q=Mx(k[Y>>2]|0,X+V|0,Z-V|0)|0;if((Q|0)<1){ba=Q;r=112;break}U=(k[m>>2]|0)+Q|0;k[m>>2]=U;if(U>>>0<Z>>>0)V=U;else{ca=U;break}}do if((r|0)==112)if((ba|0)<0){V=rv()|0;Ng(a,-1,sv(k[V>>2]|0)|0);da=k[m>>2]|0;r=117;break g}else{k[_>>2]=1;ca=k[m>>2]|0;break}while(0);k[b>>2]=k[W>>2];if(ca){aa=ca;r=116}else{ea=$;r=119}}else{ea=$;r=119}}else r=118;while(0);if((r|0)==116){$=aa+-1|0;k[m>>2]=$;k[b>>2]=(k[b>>2]|0)+1;da=$;r=117}if((r|0)==117)if(!da)r=118;else{fa=da;r=128}if((r|0)==118){ea=a+76|0;r=119}h:do if(((r|0)==119?(k[ea>>2]|0)==0:0)?(da=a+40|0,(k[da>>2]|0)==0):0){$=a+24|0;aa=k[$>>2]|0;ca=k[c>>2]|0;k[m>>2]=0;ba=a+4|0;_=0;while(1){V=Mx(k[ba>>2]|0,aa+_|0,ca-_|0)|0;if((V|0)<1){ga=V;r=124;break}Z=(k[m>>2]|0)+V|0;k[m>>2]=Z;if(Z>>>0<ca>>>0)_=Z;else{ha=Z;break}}do if((r|0)==124)if((ga|0)<0){_=rv()|0;Ng(a,-1,sv(k[_>>2]|0)|0);break h}else{k[da>>2]=1;ha=k[m>>2]|0;break}while(0);k[b>>2]=k[$>>2];if(ha){fa=ha;r=128}}while(0);if((r|0)==128){k[m>>2]=fa+-1;k[b>>2]=(k[b>>2]|0)+1}i:do if(y&4){fa=k[m>>2]|0;j:do if(!fa){ha=a+76|0;if(!(k[ha>>2]|0)){ga=a+40|0;if(!(k[ga>>2]|0)){ea=a+24|0;da=k[ea>>2]|0;_=k[c>>2]|0;k[m>>2]=0;ca=a+4|0;aa=0;while(1){ba=Mx(k[ca>>2]|0,da+aa|0,_-aa|0)|0;if((ba|0)<1){ia=ba;r=136;break}W=(k[m>>2]|0)+ba|0;k[m>>2]=W;if(W>>>0<_>>>0)aa=W;else{ja=W;break}}do if((r|0)==136)if((ia|0)<0){aa=rv()|0;Ng(a,-1,sv(k[aa>>2]|0)|0);ka=-1;la=k[m>>2]|0;r=141;break j}else{k[ga>>2]=1;ja=k[m>>2]|0;break}while(0);k[b>>2]=k[ea>>2];if(ja){ma=ja;r=140}else{na=ha;oa=-1;r=143}}else{na=ha;oa=-1;r=143}}else{pa=-1;r=142}}else{ma=fa;r=140}while(0);if((r|0)==140){fa=ma+-1|0;k[m>>2]=fa;$=k[b>>2]|0;k[b>>2]=$+1;ka=l[$>>0]|0;la=fa;r=141}if((r|0)==141)if(!la){pa=ka;r=142}else{qa=la;ra=ka;r=152}if((r|0)==142){na=a+76|0;oa=pa;r=143}k:do if((r|0)==143)if((k[na>>2]|0)==0?(fa=a+40|0,(k[fa>>2]|0)==0):0){$=a+24|0;ga=k[$>>2]|0;aa=k[c>>2]|0;k[m>>2]=0;_=a+4|0;da=0;while(1){ca=Mx(k[_>>2]|0,ga+da|0,aa-da|0)|0;if((ca|0)<1){sa=ca;r=148;break}W=(k[m>>2]|0)+ca|0;k[m>>2]=W;if(W>>>0<aa>>>0)da=W;else{ta=W;break}}do if((r|0)==148)if((sa|0)<0){da=rv()|0;Ng(a,-1,sv(k[da>>2]|0)|0);ua=-256;va=oa;break k}else{k[fa>>2]=1;ta=k[m>>2]|0;break}while(0);k[b>>2]=k[$>>2];if(ta){qa=ta;ra=oa;r=152}else{ua=-256;va=oa}}else{ua=-256;va=oa}while(0);if((r|0)==152){k[m>>2]=qa+-1;fa=k[b>>2]|0;k[b>>2]=fa+1;ua=l[fa>>0]<<8;va=ra}fa=ua+va|0;if(fa){da=a+76|0;aa=a+40|0;ga=a+24|0;_=a+4|0;ha=fa;fa=k[m>>2]|0;while(1){ha=ha+-1|0;if(!fa){if(k[da>>2]|0)break i;if(k[aa>>2]|0)break i;ea=k[ga>>2]|0;W=k[c>>2]|0;k[m>>2]=0;ca=0;while(1){ba=Mx(k[_>>2]|0,ea+ca|0,W-ca|0)|0;if((ba|0)<1){wa=ba;r=161;break}Z=(k[m>>2]|0)+ba|0;k[m>>2]=Z;if(Z>>>0<W>>>0)ca=Z;else{xa=Z;break}}if((r|0)==161){r=0;if((wa|0)<0)break;k[aa>>2]=1;xa=k[m>>2]|0}k[b>>2]=k[ga>>2];if(!xa)break i;else ya=xa}else ya=fa;fa=ya+-1|0;k[m>>2]=fa;k[b>>2]=(k[b>>2]|0)+1;if(!ha)break i}ha=rv()|0;Ng(a,-1,sv(k[ha>>2]|0)|0)}}while(0);l:do if(y&8){ya=a+76|0;xa=a+40|0;wa=a+24|0;va=a+4|0;ua=k[m>>2]|0;while(1){if(!ua){if(k[ya>>2]|0)break l;if(k[xa>>2]|0)break l;ra=k[wa>>2]|0;qa=k[c>>2]|0;k[m>>2]=0;oa=0;while(1){ta=Mx(k[va>>2]|0,ra+oa|0,qa-oa|0)|0;if((ta|0)<1){za=ta;r=174;break}sa=(k[m>>2]|0)+ta|0;k[m>>2]=sa;if(sa>>>0<qa>>>0)oa=sa;else{Aa=sa;break}}if((r|0)==174){r=0;if((za|0)<0)break;k[xa>>2]=1;Aa=k[m>>2]|0}k[b>>2]=k[wa>>2];if(!Aa)break l;else Ba=Aa}else Ba=ua;ua=Ba+-1|0;k[m>>2]=ua;oa=k[b>>2]|0;k[b>>2]=oa+1;if(!(i[oa>>0]|0))break l}ua=rv()|0;Ng(a,-1,sv(k[ua>>2]|0)|0)}while(0);m:do if(y&16){Ba=a+76|0;Aa=a+40|0;za=a+24|0;ua=a+4|0;wa=k[m>>2]|0;while(1){if(!wa){if(k[Ba>>2]|0)break m;if(k[Aa>>2]|0)break m;xa=k[za>>2]|0;va=k[c>>2]|0;k[m>>2]=0;ya=0;while(1){oa=Mx(k[ua>>2]|0,xa+ya|0,va-ya|0)|0;if((oa|0)<1){Ca=oa;r=187;break}qa=(k[m>>2]|0)+oa|0;k[m>>2]=qa;if(qa>>>0<va>>>0)ya=qa;else{Da=qa;break}}if((r|0)==187){r=0;if((Ca|0)<0)break;k[Aa>>2]=1;Da=k[m>>2]|0}k[b>>2]=k[za>>2];if(!Da)break m;else Ea=Da}else Ea=wa;wa=Ea+-1|0;k[m>>2]=wa;ya=k[b>>2]|0;k[b>>2]=ya+1;if(!(i[ya>>0]|0))break m}wa=rv()|0;Ng(a,-1,sv(k[wa>>2]|0)|0)}while(0);n:do if(y&2){Ea=k[m>>2]|0;o:do if(!Ea){Da=a+76|0;if(k[Da>>2]|0){r=205;break}Ca=a+40|0;if(k[Ca>>2]|0){Fa=Da;r=206;break}wa=a+24|0;za=k[wa>>2]|0;Aa=k[c>>2]|0;k[m>>2]=0;ua=a+4|0;Ba=0;while(1){ya=Mx(k[ua>>2]|0,za+Ba|0,Aa-Ba|0)|0;if((ya|0)<1){Ga=ya;r=199;break}va=(k[m>>2]|0)+ya|0;k[m>>2]=va;if(va>>>0<Aa>>>0)Ba=va;else{Ha=va;break}}do if((r|0)==199)if((Ga|0)<0){Ba=rv()|0;Ng(a,-1,sv(k[Ba>>2]|0)|0);Ia=k[m>>2]|0;r=204;break o}else{k[Ca>>2]=1;Ha=k[m>>2]|0;break}while(0);k[b>>2]=k[wa>>2];if(!Ha){Fa=Da;r=206}else{Ja=Ha;r=203}}else{Ja=Ea;r=203}while(0);if((r|0)==203){Ea=Ja+-1|0;k[m>>2]=Ea;k[b>>2]=(k[b>>2]|0)+1;Ia=Ea;r=204}if((r|0)==204)if(!Ia)r=205;else Ka=Ia;if((r|0)==205){Fa=a+76|0;r=206}if((r|0)==206){if(k[Fa>>2]|0)break;Ea=a+40|0;if(k[Ea>>2]|0)break;Ca=a+24|0;Ba=k[Ca>>2]|0;Aa=k[c>>2]|0;k[m>>2]=0;za=a+4|0;ua=0;while(1){va=Mx(k[za>>2]|0,Ba+ua|0,Aa-ua|0)|0;if((va|0)<1){La=va;r=211;break}ya=(k[m>>2]|0)+va|0;k[m>>2]=ya;if(ya>>>0<Aa>>>0)ua=ya;else{Ma=ya;break}}do if((r|0)==211)if((La|0)<0){ua=rv()|0;Ng(a,-1,sv(k[ua>>2]|0)|0);break n}else{k[Ea>>2]=1;Ma=k[m>>2]|0;break}while(0);k[b>>2]=k[Ca>>2];if(!Ma)break;else Ka=Ma}k[m>>2]=Ka+-1;k[b>>2]=(k[b>>2]|0)+1}while(0);_g(b)|0;Ka=Ag(0,0,0)|0;k[a+132>>2]=Ka;k[a+52>>2]=2;k[a+56>>2]=0;n=0;return n|0}Ng(a,-3,607121);n=-1;return n|0}Ng(a,-3,607074);n=-1;return n|0}Ka=a+28|0;i[k[Ka>>2]>>0]=31;k[a+36>>2]=1;Na=Ka;Oa=k[m>>2]|0}else{Na=a+28|0;Oa=u}k[a+48>>2]=k[a+12>>2];u=k[Na>>2]|0;k[a+32>>2]=u;if(Oa){Na=a+36|0;py(u+(k[Na>>2]|0)|0,k[b>>2]|0,Oa|0)|0;k[Na>>2]=(k[Na>>2]|0)+(k[m>>2]|0);k[m>>2]=0}k[a+52>>2]=1;k[a+56>>2]=1;n=0;return n|0}function Tg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;b=r;r=r+16|0;c=b+4|0;d=b;e=a+84|0;f=a+100|0;g=k[f>>2]|0;h=a+88|0;i=a+76|0;j=a+40|0;l=a+24|0;m=a+16|0;n=a+4|0;a:while(1){if(!(k[h>>2]|0)){if(k[i>>2]|0){o=-1;p=27;break}if(k[j>>2]|0){p=12;break}q=k[l>>2]|0;s=k[m>>2]|0;k[h>>2]=0;t=0;while(1){u=Mx(k[n>>2]|0,q+t|0,s-t|0)|0;if((u|0)<1){v=u;p=8;break}w=(k[h>>2]|0)+u|0;k[h>>2]=w;if(w>>>0<s>>>0)t=w;else{x=w;break}}if((p|0)==8){p=0;if((v|0)<0){p=10;break}k[j>>2]=1;x=k[h>>2]|0}k[e>>2]=k[l>>2];if(!x){p=12;break}}t=bh(e,0)|0;switch(t|0){case 2:case -2:{p=14;break a;break}case -4:{p=15;break a;break}case -3:{p=16;break a;break}default:{}}s=k[f>>2]|0;if(!((t|0)!=1&(s|0)!=0)){y=t;z=s;p=18;break}}if((p|0)==10){f=rv()|0;Ng(a,-1,sv(k[f>>2]|0)|0);o=-1;r=b;return o|0}else if((p|0)==12){Ng(a,-3,641381);o=-1;r=b;return o|0}else if((p|0)==14){Ng(a,-2,606900);o=-1;r=b;return o|0}else if((p|0)==15){Ng(a,-4,607037);o=-1;r=b;return o|0}else if((p|0)==16){f=k[a+108>>2]|0;Ng(a,-3,(f|0)==0?606939:f);o=-1;r=b;return o|0}else if((p|0)==18){f=g-z|0;k[a+36>>2]=f;z=(k[a+96>>2]|0)+(0-f)|0;k[a+32>>2]=z;g=a+132|0;e=Ag(k[g>>2]|0,z,f)|0;k[g>>2]=e;if((y|0)!=1){o=0;r=b;return o|0}if((Ug(a,c)|0)!=-1?(Ug(a,d)|0)!=-1:0){if((k[c>>2]|0)!=(k[g>>2]|0)){Ng(a,-3,607468);o=-1;r=b;return o|0}if((k[d>>2]|0)==(k[a+104>>2]|0)){k[a+52>>2]=0;o=0;r=b;return o|0}else{Ng(a,-3,607489);o=-1;r=b;return o|0}}Ng(a,-3,641381);o=-1;r=b;return o|0}else if((p|0)==27){r=b;return o|0}return 0}function Ug(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;c=a+84|0;d=a+88|0;e=k[d>>2]|0;a:do if(!e){f=a+76|0;if(!(k[f>>2]|0)){g=a+40|0;if(!(k[g>>2]|0)){h=a+24|0;i=k[h>>2]|0;j=k[a+16>>2]|0;k[d>>2]=0;m=a+4|0;n=0;while(1){o=Mx(k[m>>2]|0,i+n|0,j-n|0)|0;if((o|0)<1){p=o;q=7;break}r=(k[d>>2]|0)+o|0;k[d>>2]=r;if(r>>>0<j>>>0)n=r;else{s=r;break}}do if((q|0)==7)if((p|0)<0){n=rv()|0;Ng(a,-1,sv(k[n>>2]|0)|0);t=-1;u=k[d>>2]|0;q=12;break a}else{k[g>>2]=1;s=k[d>>2]|0;break}while(0);k[c>>2]=k[h>>2];if(s){v=s;q=11}else{w=f;x=-1;q=14}}else{w=f;x=-1;q=14}}else{y=-1;q=13}}else{v=e;q=11}while(0);if((q|0)==11){e=v+-1|0;k[d>>2]=e;v=k[c>>2]|0;k[c>>2]=v+1;t=l[v>>0]|0;u=e;q=12}if((q|0)==12)if(!u){y=t;q=13}else{z=t;A=u;q=25}if((q|0)==13){w=a+76|0;x=y;q=14}b:do if((q|0)==14){if(k[w>>2]|0){B=x+-256|0;q=27;break}y=a+40|0;if(!(k[y>>2]|0)){u=a+24|0;t=k[u>>2]|0;e=k[a+16>>2]|0;k[d>>2]=0;v=a+4|0;s=0;while(1){p=Mx(k[v>>2]|0,t+s|0,e-s|0)|0;if((p|0)<1){C=p;q=20;break}g=(k[d>>2]|0)+p|0;k[d>>2]=g;if(g>>>0<e>>>0)s=g;else{D=g;break}}do if((q|0)==20)if((C|0)<0){s=rv()|0;Ng(a,-1,sv(k[s>>2]|0)|0);E=x;F=-256;G=k[d>>2]|0;q=26;break b}else{k[y>>2]=1;D=k[d>>2]|0;break}while(0);k[c>>2]=k[u>>2];if(D){z=x;A=D;q=25;break}}H=w;I=x+-256|0;q=28}while(0);if((q|0)==25){x=A+-1|0;k[d>>2]=x;A=k[c>>2]|0;k[c>>2]=A+1;E=z;F=(l[A>>0]|0)<<8;G=x;q=26}if((q|0)==26){x=F+E|0;if(!G){B=x;q=27}else{J=G;K=x;q=39}}if((q|0)==27){H=a+76|0;I=B;q=28}c:do if((q|0)==28){if(k[H>>2]|0){L=I+-65536|0;q=41;break}B=a+40|0;if(!(k[B>>2]|0)){x=a+24|0;G=k[x>>2]|0;E=k[a+16>>2]|0;k[d>>2]=0;F=a+4|0;A=0;while(1){z=Mx(k[F>>2]|0,G+A|0,E-A|0)|0;if((z|0)<1){M=z;q=34;break}w=(k[d>>2]|0)+z|0;k[d>>2]=w;if(w>>>0<E>>>0)A=w;else{N=w;break}}do if((q|0)==34)if((M|0)<0){A=rv()|0;Ng(a,-1,sv(k[A>>2]|0)|0);O=I;P=-65536;Q=k[d>>2]|0;q=40;break c}else{k[B>>2]=1;N=k[d>>2]|0;break}while(0);k[c>>2]=k[x>>2];if(N){J=N;K=I;q=39;break}}R=H;S=I+-65536|0;q=42}while(0);if((q|0)==39){I=J+-1|0;k[d>>2]=I;J=k[c>>2]|0;k[c>>2]=J+1;O=K;P=(l[J>>0]|0)<<16;Q=I;q=40}if((q|0)==40){I=P+O|0;if(!Q){L=I;q=41}else{T=Q;U=I}}if((q|0)==41){R=a+76|0;S=L;q=42}if((q|0)==42){if(k[R>>2]|0){V=-1;return V|0}R=a+40|0;if(k[R>>2]|0){V=-1;return V|0}L=a+24|0;I=k[L>>2]|0;Q=k[a+16>>2]|0;k[d>>2]=0;O=a+4|0;P=0;while(1){J=Mx(k[O>>2]|0,I+P|0,Q-P|0)|0;if((J|0)<1){W=J;q=47;break}K=(k[d>>2]|0)+J|0;k[d>>2]=K;if(K>>>0<Q>>>0)P=K;else{X=K;break}}do if((q|0)==47){if((W|0)>=0){k[R>>2]=1;X=k[d>>2]|0;break}P=rv()|0;Ng(a,-1,sv(k[P>>2]|0)|0);V=-1;return V|0}while(0);k[c>>2]=k[L>>2];if(!X){V=-1;return V|0}else{T=X;U=S}}k[d>>2]=T+-1;T=k[c>>2]|0;k[c>>2]=T+1;k[b>>2]=((l[T>>0]|0)<<24)+U;V=0;return V|0}function Vg(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0;if(!a){d=0;return d|0}e=a+84|0;if((k[a>>2]|0)!=31153){d=0;return d|0}if(k[a+76>>2]|0){d=0;return d|0}if((c|0)<0){Ng(a,-5,606961);d=0;return d|0}if(!c){d=0;return d|0}f=a+16|0;if((k[f>>2]|0)==0?(Xg(a)|0)==-1:0){d=0;return d|0}g=a+72|0;a:do if(k[g>>2]|0){k[g>>2]=0;h=k[a+68>>2]|0;i=a+88|0;if((k[i>>2]|0)!=0?(Yg(a,0)|0)==-1:0){d=0;return d|0}if(h){j=a+24|0;l=a+12|0;m=h;h=0;while(1){n=k[f>>2]|0;o=(n|0)<0|(n|0)>(m|0)?m:n;if(!h)sy(k[j>>2]|0,0,o|0)|0;k[i>>2]=o;k[e>>2]=k[j>>2];k[l>>2]=(k[l>>2]|0)+o;if((Yg(a,0)|0)==-1){d=0;break}if((m|0)==(o|0))break a;else{m=m-o|0;h=1}}return d|0}}while(0);g=a+88|0;b:do if((k[f>>2]|0)>>>0<=c>>>0){if((k[g>>2]|0)!=0?(Yg(a,0)|0)==-1:0){d=0;return d|0}k[g>>2]=c;k[e>>2]=b;h=a+12|0;k[h>>2]=(k[h>>2]|0)+c;if((Yg(a,0)|0)==-1){d=0;return d|0}}else{h=a+24|0;m=a+12|0;l=b;j=c;while(1){i=k[g>>2]|0;if(!i){o=k[h>>2]|0;k[e>>2]=o;p=o}else p=k[e>>2]|0;o=(k[f>>2]|0)-i|0;n=o>>>0>j>>>0?j:o;py(p+i|0,l|0,n|0)|0;k[g>>2]=(k[g>>2]|0)+n;k[m>>2]=(k[m>>2]|0)+n;if((j|0)==(n|0))break b;if((Yg(a,0)|0)==-1){d=0;break}else{l=l+n|0;j=j-n|0}}return d|0}while(0);d=c;return d|0}function Wg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0;if(!a){b=-2;return b|0}if((k[a>>2]|0)!=31153){b=-2;return b|0}c=a+72|0;a:do if(k[c>>2]|0){k[c>>2]=0;d=k[a+68>>2]|0;e=a+84|0;f=a+88|0;if((k[f>>2]|0)!=0?(Yg(a,0)|0)==-1:0){g=e;h=-1;break}if(!d){g=e;h=0}else{i=a+16|0;j=a+24|0;l=a+12|0;m=d;d=0;while(1){n=k[i>>2]|0;o=(n|0)<0|(n|0)>(m|0)?m:n;if(!d)sy(k[j>>2]|0,0,o|0)|0;k[f>>2]=o;k[e>>2]=k[j>>2];k[l>>2]=(k[l>>2]|0)+o;if((Yg(a,0)|0)==-1){g=e;h=-1;break a}if((m|0)==(o|0)){g=e;h=0;break}else{m=m-o|0;d=1}}}}else{g=a+84|0;h=0}while(0);c=(Yg(a,4)|0)+h|0;Cg(g)|0;dy(k[a+28>>2]|0);dy(k[a+24>>2]|0);Ng(a,0,0);dy(k[a+8>>2]|0);g=Hx(k[a+4>>2]|0)|0;dy(a);b=((c|0)!=(0-g|0))<<31>>31;return b|0}function Xg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;b=a+20|0;c=k[b>>2]|0;d=cy(c)|0;e=a+24|0;k[e>>2]=d;f=cy(c)|0;c=a+28|0;k[c>>2]=f;g=(f|0)==0;if(!(g|(d|0)==0)){k[a+116>>2]=0;k[a+120>>2]=0;k[a+124>>2]=0;if(!(Bg(a+84|0,k[a+60>>2]|0,8,31,8,k[a+64>>2]|0,608415,56)|0)){h=k[b>>2]|0;k[a+16>>2]=h;k[a+100>>2]=h;h=k[c>>2]|0;k[a+96>>2]=h;k[a+32>>2]=h;i=0;return i|0}else{dy(k[e>>2]|0);Ng(a,-4,607037);i=-1;return i|0}}if(g)j=d;else{dy(f);j=k[e>>2]|0}if(j)dy(j);Ng(a,-4,607037);i=-1;return i|0}function Yg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;c=a+84|0;d=a+16|0;if((k[d>>2]|0)==0?(Xg(a)|0)==-1:0){e=-1;return e|0}f=a+100|0;g=a+96|0;h=a+32|0;i=a+28|0;j=a+4|0;l=(b|0)==4;m=k[f>>2]|0;a:do if(!b){n=m;while(1){if(!n){o=k[g>>2]|0;p=k[h>>2]|0;q=o-p|0;if((o|0)!=(p|0)){o=Ox(k[j>>2]|0,p,q)|0;if(!((o|0)>-1&(o|0)==(q|0))){r=26;break a}q=k[f>>2]|0;if(q){s=k[g>>2]|0;t=q}else r=10}else r=10;if((r|0)==10){r=0;q=k[d>>2]|0;k[f>>2]=q;o=k[i>>2]|0;k[g>>2]=o;s=o;t=q}k[h>>2]=s;u=t}else u=n;if((Eg(c,0)|0)==-2){r=32;break a}n=k[f>>2]|0;if((u|0)==(n|0)){r=34;break}}}else{if(l){v=m;w=0}else{n=m;while(1){q=k[g>>2]|0;o=k[h>>2]|0;p=q-o|0;if((q|0)==(o|0))x=n;else{q=Ox(k[j>>2]|0,o,p)|0;if(!((q|0)>-1&(q|0)==(p|0))){r=26;break a}x=k[f>>2]|0}if(!x){p=k[d>>2]|0;k[f>>2]=p;q=k[i>>2]|0;k[g>>2]=q;y=q;z=p}else{y=k[g>>2]|0;z=x}k[h>>2]=y;if((Eg(c,b)|0)==-2){r=32;break a}n=k[f>>2]|0;if((z|0)==(n|0)){r=34;break a}}}while(1){if((v|0)==0|(w|0)==1){n=k[g>>2]|0;p=k[h>>2]|0;q=n-p|0;if((n|0)==(p|0))A=v;else{n=Ox(k[j>>2]|0,p,q)|0;if(!((n|0)>-1&(n|0)==(q|0))){r=26;break a}A=k[f>>2]|0}if(!A){q=k[d>>2]|0;k[f>>2]=q;n=k[i>>2]|0;k[g>>2]=n;B=n;C=q}else{B=k[g>>2]|0;C=A}k[h>>2]=B;D=C}else D=v;w=Eg(c,4)|0;if((w|0)==-2){r=32;break a}v=k[f>>2]|0;if((D|0)==(v|0)){r=34;break}}}while(0);if((r|0)==26){v=rv()|0;Ng(a,-1,sv(k[v>>2]|0)|0);e=-1;return e|0}else if((r|0)==32){Ng(a,-2,606998);e=-1;return e|0}else if((r|0)==34){if(!l){e=0;return e|0}Dg(c)|0;e=0;return e|0}return 0}function Zg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0;c=k[a+28>>2]|0;d=k[a>>2]|0;e=a+4|0;f=d+((k[e>>2]|0)+-6)|0;g=a+12|0;h=k[g>>2]|0;m=a+16|0;n=k[m>>2]|0;o=h+(n+-258)|0;p=k[c+44>>2]|0;q=k[c+48>>2]|0;r=k[c+52>>2]|0;s=c+56|0;t=c+60|0;u=k[c+76>>2]|0;v=k[c+80>>2]|0;w=(1<<k[c+84>>2])+-1|0;x=(1<<k[c+88>>2])+-1|0;y=h+(n+~b)|0;b=c+7104|0;n=r+-1|0;z=(q|0)==0;A=(k[c+40>>2]|0)+-1|0;B=A+q|0;C=q+-1|0;D=y+-1|0;E=y-q|0;F=y+-2|0;G=k[t>>2]|0;H=k[s>>2]|0;I=d+-1|0;d=h+-1|0;a:while(1){if(G>>>0<15){h=I+2|0;J=G+16|0;K=((l[I+1>>0]|0)<<G)+H+((l[h>>0]|0)<<G+8)|0;L=h}else{J=G;K=H;L=I}h=K&w;M=i[u+(h<<2)>>0]|0;N=j[u+(h<<2)+2>>1]|0;O=l[u+(h<<2)+1>>0]|0;h=K>>>O;P=J-O|0;do if(M<<24>>24){O=h;Q=P;R=M&255;S=N;while(1){if(R&16){T=R;U=Q;V=O;W=S;break}if(R&64){X=R;Y=Q;Z=O;_=L;$=d;aa=57;break a}ba=(O&(1<<R)+-1)+(S&65535)|0;ca=i[u+(ba<<2)>>0]|0;da=j[u+(ba<<2)+2>>1]|0;ea=l[u+(ba<<2)+1>>0]|0;ba=O>>>ea;fa=Q-ea|0;if(!(ca<<24>>24)){ga=ba;ha=fa;ia=da;aa=6;break}else{O=ba;Q=fa;R=ca&255;S=da}}if((aa|0)==6){aa=0;ja=ga;ka=ha;la=ia&255;aa=7;break}S=W&65535;R=T&15;if(!R){ma=U;na=V;oa=L;pa=S}else{if(U>>>0<R>>>0){Q=L+1|0;qa=U+8|0;ra=((l[Q>>0]|0)<<U)+V|0;sa=Q}else{qa=U;ra=V;sa=L}ma=qa-R|0;na=ra>>>R;oa=sa;pa=(ra&(1<<R)+-1)+S|0}if(ma>>>0<15){S=oa+2|0;ta=ma+16|0;ua=((l[oa+1>>0]|0)<<ma)+na+((l[S>>0]|0)<<ma+8)|0;va=S}else{ta=ma;ua=na;va=oa}S=ua&x;R=j[v+(S<<2)+2>>1]|0;Q=l[v+(S<<2)+1>>0]|0;O=ua>>>Q;da=ta-Q|0;Q=l[v+(S<<2)>>0]|0;if(!(Q&16)){S=Q;ca=O;fa=da;ba=R;while(1){if(S&64){wa=fa;xa=ca;ya=va;za=d;aa=54;break a}ea=(ca&(1<<S)+-1)+(ba&65535)|0;Aa=j[v+(ea<<2)+2>>1]|0;Ba=l[v+(ea<<2)+1>>0]|0;Ca=ca>>>Ba;Da=fa-Ba|0;Ba=l[v+(ea<<2)>>0]|0;if(!(Ba&16)){S=Ba;ca=Ca;fa=Da;ba=Aa}else{Ea=Ca;Fa=Da;Ga=Ba;Ha=Aa;break}}}else{Ea=O;Fa=da;Ga=Q;Ha=R}ba=Ha&65535;fa=Ga&15;if(Fa>>>0<fa>>>0){ca=va+1|0;S=((l[ca>>0]|0)<<Fa)+Ea|0;Aa=Fa+8|0;if(Aa>>>0<fa>>>0){Ba=va+2|0;Ia=Fa+16|0;Ja=((l[Ba>>0]|0)<<Aa)+S|0;Ka=Ba}else{Ia=Aa;Ja=S;Ka=ca}}else{Ia=Fa;Ja=Ea;Ka=va}ca=(Ja&(1<<fa)+-1)+ba|0;ba=Ja>>>fa;S=Ia-fa|0;fa=d;Aa=fa-y|0;if(ca>>>0<=Aa>>>0){Ba=2-pa|0;Da=pa+(Ba>>>0>4294967293?Ba:-3)|0;Ba=Da-((Da>>>0)%3|0)|0;Da=Ba-ca|0;Ca=d+(0-ca)|0;ea=pa;La=d;do{i[La+1>>0]=i[Ca+1>>0]|0;i[La+2>>0]=i[Ca+2>>0]|0;Ca=Ca+3|0;La=La+3|0;i[La>>0]=i[Ca>>0]|0;ea=ea+-3|0}while(ea>>>0>2);ea=pa+-3|0;if((ea|0)==(Ba|0)){Ma=S;Na=ba;Oa=Ka;Pa=d+(Ba+3)|0;break}Ca=d+(Ba+4)|0;i[Ca>>0]=i[d+(Da+4)>>0]|0;if((ea-Ba|0)>>>0<=1){Ma=S;Na=ba;Oa=Ka;Pa=Ca;break}Ca=d+(Ba+5)|0;i[Ca>>0]=i[d+(Da+5)>>0]|0;Ma=S;Na=ba;Oa=Ka;Pa=Ca;break}Ca=ca-Aa|0;if(Ca>>>0>p>>>0?(k[b>>2]|0)!=0:0){Qa=ba;Ra=S;Sa=Ka;Ta=d;aa=22;break a}do if(z){ea=r+(A-Ca)|0;if(pa>>>0>Ca>>>0){La=pa-Ca|0;R=ca-fa|0;Q=ea;da=Ca;O=d;do{Q=Q+1|0;O=O+1|0;i[O>>0]=i[Q>>0]|0;da=da+-1|0}while((da|0)!=0);Ua=d+(D+R+(1-ca))|0;Va=La;Wa=d+(y+R)|0}else{Ua=ea;Va=pa;Wa=d}}else{if(Ca>>>0<=q>>>0){da=r+(C-Ca)|0;if(pa>>>0<=Ca>>>0){Ua=da;Va=pa;Wa=d;break}Q=pa-Ca|0;O=ca-fa|0;Xa=da;da=Ca;Ya=d;do{Xa=Xa+1|0;Ya=Ya+1|0;i[Ya>>0]=i[Xa>>0]|0;da=da+-1|0}while((da|0)!=0);Ua=d+(D+O+(1-ca))|0;Va=Q;Wa=d+(y+O)|0;break}da=r+(B-Ca)|0;Xa=Ca-q|0;if(pa>>>0>Xa>>>0){Ya=pa-Xa|0;ea=ca-fa|0;R=da;La=Xa;Xa=d;do{R=R+1|0;Xa=Xa+1|0;i[Xa>>0]=i[R>>0]|0;La=La+-1|0}while((La|0)!=0);La=d+(E+ea)|0;if(Ya>>>0>q>>>0){R=Ya-q|0;Xa=n;O=q;Q=La;do{Xa=Xa+1|0;Q=Q+1|0;i[Q>>0]=i[Xa>>0]|0;O=O+-1|0}while((O|0)!=0);Ua=d+(F+ea+(2-ca))|0;Va=R;Wa=d+(y+ea)|0}else{Ua=n;Va=Ya;Wa=La}}else{Ua=da;Va=pa;Wa=d}}while(0);if(Va>>>0>2){ca=Va+-3|0;fa=(ca>>>0)%3|0;Ca=ca-fa|0;ca=Ua;Aa=Va;Da=Wa;do{i[Da+1>>0]=i[ca+1>>0]|0;i[Da+2>>0]=i[ca+2>>0]|0;ca=ca+3|0;Da=Da+3|0;i[Da>>0]=i[ca>>0]|0;Aa=Aa+-3|0}while(Aa>>>0>2);Za=Ua+(Ca+3)|0;_a=fa;$a=Wa+(Ca+3)|0}else{Za=Ua;_a=Va;$a=Wa}if(_a){Aa=$a+1|0;i[Aa>>0]=i[Za+1>>0]|0;if(_a>>>0>1){ca=$a+2|0;i[ca>>0]=i[Za+2>>0]|0;Ma=S;Na=ba;Oa=Ka;Pa=ca}else{Ma=S;Na=ba;Oa=Ka;Pa=Aa}}else{Ma=S;Na=ba;Oa=Ka;Pa=$a}}else{ja=h;ka=P;la=N&255;aa=7}while(0);if((aa|0)==7){aa=0;N=d+1|0;i[N>>0]=la;Ma=ka;Na=ja;Oa=L;Pa=N}if(Pa>>>0<o>>>0&Oa>>>0<f>>>0){G=Ma;H=Na;I=Oa;d=Pa}else{ab=Ma;bb=Na;cb=Oa;db=Pa;break}}do if((aa|0)==22){k[a+24>>2]=607438;k[c>>2]=29;ab=Ra;bb=Qa;cb=Sa;db=Ta}else if((aa|0)==54){k[a+24>>2]=607416;k[c>>2]=29;ab=wa;bb=xa;cb=ya;db=za}else if((aa|0)==57)if(!(X&32)){k[a+24>>2]=607388;k[c>>2]=29;ab=Y;bb=Z;cb=_;db=$;break}else{k[c>>2]=11;ab=Y;bb=Z;cb=_;db=$;break}while(0);$=ab>>>3;_=ab-($<<3)|0;k[a>>2]=cb+(1-$);k[g>>2]=db+1;k[e>>2]=f+5-(cb+(0-$));k[m>>2]=o+257-db;k[s>>2]=(1<<_)+-1&bb;k[t>>2]=_;return}function _g(a){a=a|0;var b=0,c=0;if(!a){b=-2;return b|0}c=k[a+28>>2]|0;if(!c){b=-2;return b|0}k[c+28>>2]=0;k[a+20>>2]=0;k[a+8>>2]=0;k[a+24>>2]=0;k[a+48>>2]=1;k[c>>2]=0;k[c+4>>2]=0;k[c+12>>2]=0;k[c+20>>2]=32768;k[c+32>>2]=0;k[c+40>>2]=0;k[c+44>>2]=0;k[c+48>>2]=0;k[c+56>>2]=0;k[c+60>>2]=0;a=c+1328|0;k[c+108>>2]=a;k[c+80>>2]=a;k[c+76>>2]=a;k[c+7104>>2]=1;k[c+7108>>2]=-1;b=0;return b|0}function $g(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;if(!a){c=-2;return c|0}d=a+28|0;e=k[d>>2]|0;if(!e){c=-2;return c|0}if((b|0)<0){f=0-b|0;g=0}else{f=(b|0)<48?b&15:b;g=(b>>4)+1|0}if(!((f|0)==0|(f&-8|0)==8)){c=-2;return c|0}b=e+52|0;h=k[b>>2]|0;i=e+36|0;if((h|0)!=0?(k[i>>2]|0)!=(f|0):0){Tb[k[a+36>>2]&1](k[a+40>>2]|0,h);k[b>>2]=0;b=k[d>>2]|0;k[e+8>>2]=g;k[i>>2]=f;if(!b){c=-2;return c|0}else j=b}else{k[e+8>>2]=g;k[i>>2]=f;j=e}k[j+28>>2]=0;k[a+20>>2]=0;k[a+8>>2]=0;k[a+24>>2]=0;k[a+48>>2]=1;k[j>>2]=0;k[j+4>>2]=0;k[j+12>>2]=0;k[j+20>>2]=32768;k[j+32>>2]=0;k[j+40>>2]=0;k[j+44>>2]=0;k[j+48>>2]=0;k[j+56>>2]=0;k[j+60>>2]=0;a=j+1328|0;k[j+108>>2]=a;k[j+80>>2]=a;k[j+76>>2]=a;k[j+7104>>2]=1;k[j+7108>>2]=-1;c=0;return c|0}function ah(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0;if(!c){e=-6;return e|0}if((d|0)!=56|(i[c>>0]|0)!=49){e=-6;return e|0}if(!a){e=-2;return e|0}k[a+24>>2]=0;c=a+32|0;d=k[c>>2]|0;if(!d){k[c>>2]=7;k[a+40>>2]=0;f=7}else f=d;d=a+36|0;if(!(k[d>>2]|0))k[d>>2]=1;c=a+40|0;g=Pb[f&63](k[c>>2]|0,1,7116)|0;if(!g){e=-4;return e|0}f=a+28|0;k[f>>2]=g;k[g+52>>2]=0;h=$g(a,b)|0;if(!h){e=0;return e|0}Tb[k[d>>2]&1](k[c>>2]|0,g);k[f>>2]=0;e=h;return e|0}function bh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,gc=0,hc=0,ic=0,jc=0,kc=0,lc=0,mc=0,nc=0,oc=0,pc=0,qc=0,rc=0,sc=0,tc=0,uc=0,vc=0,wc=0,xc=0,yc=0,zc=0,Ac=0,Bc=0,Cc=0,Dc=0,Ec=0,Fc=0,Gc=0,Hc=0,Ic=0,Jc=0,Kc=0,Lc=0,Mc=0,Nc=0,Oc=0,Pc=0,Qc=0,Rc=0,Sc=0,Tc=0,Uc=0,Vc=0,Wc=0,Xc=0,Yc=0,Zc=0,_c=0,$c=0,ad=0,bd=0,cd=0,dd=0,ed=0,fd=0,gd=0,hd=0,id=0,jd=0,kd=0,ld=0,md=0,nd=0,od=0,pd=0,qd=0,rd=0,sd=0,td=0,ud=0,vd=0,wd=0,xd=0,yd=0,zd=0,Ad=0,Bd=0,Cd=0,Dd=0,Ed=0,Fd=0,Gd=0,Hd=0,Id=0,Jd=0,Kd=0,Ld=0,Md=0,Nd=0,Od=0,Pd=0,Qd=0,Rd=0,Sd=0,Td=0,Ud=0,Vd=0,Wd=0,Xd=0,Yd=0,Zd=0,_d=0,$d=0,ae=0,be=0,ce=0,de=0,ee=0,fe=0,ge=0,he=0,ie=0,je=0,ke=0,le=0,me=0,ne=0,oe=0,pe=0,qe=0,re=0,se=0,te=0,ue=0,ve=0,we=0,xe=0,ye=0,ze=0,Ae=0,Be=0,Ce=0,De=0,Ee=0,Fe=0,Ge=0,He=0,Ie=0,Je=0,Ke=0,Le=0,Me=0,Ne=0,Oe=0,Pe=0,Qe=0,Re=0,Se=0,Te=0,Ue=0,Ve=0,We=0,Xe=0,Ye=0,Ze=0,_e=0,$e=0,af=0,bf=0,cf=0,df=0,ef=0,ff=0,gf=0,hf=0,jf=0,kf=0,lf=0,mf=0,nf=0,of=0,pf=0,qf=0,rf=0,sf=0,tf=0,uf=0,vf=0,wf=0,xf=0,yf=0,zf=0,Af=0,Bf=0,Cf=0,Df=0,Ef=0,Ff=0,Gf=0,Hf=0,If=0,Jf=0,Kf=0,Lf=0,Mf=0,Nf=0,Of=0,Pf=0,Qf=0,Rf=0,Sf=0,Tf=0,Uf=0,Vf=0,Wf=0,Xf=0,Yf=0,Zf=0,_f=0,$f=0,ag=0,bg=0,cg=0,dg=0,eg=0,fg=0,gg=0,hg=0,ig=0,jg=0,kg=0,lg=0,mg=0,ng=0,og=0,pg=0,qg=0,rg=0,sg=0,tg=0,ug=0,vg=0,wg=0,xg=0,yg=0,Bg=0,Cg=0,Dg=0,Eg=0,Fg=0,Gg=0,Hg=0,Ig=0,Jg=0,Kg=0,Lg=0,Mg=0,Ng=0,Og=0,Pg=0,Qg=0,Rg=0,Sg=0,Tg=0,Ug=0,Vg=0,Wg=0,Xg=0,Yg=0,_g=0,$g=0,ah=0,bh=0,ch=0,fh=0,gh=0;c=r;r=r+16|0;d=c;if(!a){e=-2;r=c;return e|0}f=k[a+28>>2]|0;if(!f){e=-2;r=c;return e|0}g=a+12|0;h=k[g>>2]|0;if(!h){e=-2;r=c;return e|0}n=k[a>>2]|0;if((n|0)==0?(k[a+4>>2]|0)!=0:0){e=-2;r=c;return e|0}o=k[f>>2]|0;if((o|0)==11){k[f>>2]=12;p=12}else p=o;o=a+16|0;q=k[o>>2]|0;s=a+4|0;t=k[s>>2]|0;u=f+56|0;v=f+60|0;w=f+8|0;x=f+24|0;y=d+1|0;z=f+16|0;A=f+32|0;B=a+24|0;C=f+36|0;D=f+20|0;E=a+48|0;F=f+64|0;G=f+12|0;H=(b+-5|0)>>>0<2;I=f+4|0;J=f+76|0;K=f+84|0;L=f+80|0;M=f+88|0;N=(b|0)==6;O=f+7108|0;P=f+72|0;Q=f+7112|0;R=f+68|0;S=f+44|0;T=f+7104|0;U=f+48|0;V=f+52|0;W=f+40|0;X=a+20|0;Y=f+28|0;Z=d+2|0;_=d+3|0;$=f+104|0;aa=f+96|0;ba=f+100|0;ca=f+624|0;da=f+1328|0;ea=f+108|0;fa=f+112|0;ga=f+752|0;ha=f+92|0;ia=p;p=k[v>>2]|0;ja=t;ka=k[u>>2]|0;la=q;ma=n;n=q;q=h;h=0;a:while(1){b:do switch(ia|0){case 28:{na=la;oa=p;pa=ja;qa=ka;ra=ma;sa=n;ta=q;ua=1;break a;break}case 29:{va=p;wa=ja;xa=ka;ya=la;za=ma;Aa=n;Ba=q;Ca=284;break a;break}case 30:{e=-4;Ca=297;break a;break}case 0:{Da=k[w>>2]|0;if(!Da){k[f>>2]=12;Ea=p;Fa=ja;Ga=ka;Ha=la;Ia=ma;Ja=n;Ka=q;La=h;break b}if(p>>>0<16){Ma=p;Na=ja;Oa=ka;Pa=ma;while(1){if(!Na){na=la;oa=Ma;pa=0;qa=Oa;ra=Pa;sa=n;ta=q;ua=h;break a}Qa=Na+-1|0;Ra=Pa+1|0;Sa=(l[Pa>>0]<<Ma)+Oa|0;Ta=Ma+8|0;if(Ta>>>0<16){Ma=Ta;Na=Qa;Oa=Sa;Pa=Ra}else{Ua=Ta;Va=Qa;Wa=Sa;Xa=Ra;break}}}else{Ua=p;Va=ja;Wa=ka;Xa=ma}if((Wa|0)==35615&(Da&2|0)!=0){Pa=Ag(0,0,0)|0;k[x>>2]=Pa;i[d>>0]=31;i[y>>0]=-117;Pa=Ag(k[x>>2]|0,d,2)|0;k[x>>2]=Pa;k[f>>2]=1;Ea=0;Fa=Va;Ga=0;Ha=la;Ia=Xa;Ja=n;Ka=q;La=h;break b}k[z>>2]=0;Pa=k[A>>2]|0;if(Pa)k[Pa+48>>2]=-1;if((Da&1|0)!=0?((((Wa<<8&65280)+(Wa>>>8)|0)>>>0)%31|0|0)==0:0){if((Wa&15|0)!=8){k[B>>2]=607074;k[f>>2]=29;Ea=Ua;Fa=Va;Ga=Wa;Ha=la;Ia=Xa;Ja=n;Ka=q;La=h;break b}Pa=Wa>>>4;Oa=Ua+-4|0;Na=(Pa&15)+8|0;Ma=k[C>>2]|0;if(Ma){if(Na>>>0>Ma>>>0){k[B>>2]=607101;k[f>>2]=29;Ea=Oa;Fa=Va;Ga=Pa;Ha=la;Ia=Xa;Ja=n;Ka=q;La=h;break b}}else k[C>>2]=Na;k[D>>2]=1<<Na;Na=zg(0,0,0)|0;k[x>>2]=Na;k[E>>2]=Na;k[f>>2]=Wa>>>12&2^11;Ea=0;Fa=Va;Ga=0;Ha=la;Ia=Xa;Ja=n;Ka=q;La=h;break b}k[B>>2]=607051;k[f>>2]=29;Ea=Ua;Fa=Va;Ga=Wa;Ha=la;Ia=Xa;Ja=n;Ka=q;La=h;break}case 1:{if(p>>>0<16){Na=p;Pa=ja;Oa=ka;Ma=ma;while(1){if(!Pa){na=la;oa=Na;pa=0;qa=Oa;ra=Ma;sa=n;ta=q;ua=h;break a}Ra=Pa+-1|0;Sa=Ma+1|0;Qa=(l[Ma>>0]<<Na)+Oa|0;Ta=Na+8|0;if(Ta>>>0<16){Na=Ta;Pa=Ra;Oa=Qa;Ma=Sa}else{Ya=Qa;Za=Ta;_a=Ra;$a=Sa;break}}}else{Ya=ka;Za=p;_a=ja;$a=ma}k[z>>2]=Ya;if((Ya&255|0)!=8){k[B>>2]=607074;k[f>>2]=29;Ea=Za;Fa=_a;Ga=Ya;Ha=la;Ia=$a;Ja=n;Ka=q;La=h;break b}if(Ya&57344){k[B>>2]=607121;k[f>>2]=29;Ea=Za;Fa=_a;Ga=Ya;Ha=la;Ia=$a;Ja=n;Ka=q;La=h;break b}Ma=k[A>>2]|0;if(Ma)k[Ma>>2]=Ya>>>8&1;if(Ya&512){i[d>>0]=Ya;i[y>>0]=Ya>>>8;Ma=Ag(k[x>>2]|0,d,2)|0;k[x>>2]=Ma}k[f>>2]=2;ab=0;bb=_a;cb=0;db=$a;Ca=47;break}case 2:{if(p>>>0<32){ab=p;bb=ja;cb=ka;db=ma;Ca=47}else{eb=ja;fb=ka;gb=ma;Ca=49}break}case 3:{if(p>>>0<16){hb=p;ib=ja;jb=ka;kb=ma;Ca=55}else{lb=ja;mb=ka;nb=ma;Ca=57}break}case 4:{ob=p;pb=ja;qb=ka;rb=ma;Ca=62;break}case 5:{sb=p;tb=ja;ub=ka;vb=ma;Ca=73;break}case 6:{wb=p;xb=ja;yb=ka;zb=ma;Ca=83;break}case 7:{Ab=p;Bb=ja;Cb=ka;Db=ma;Ca=96;break}case 8:{Eb=p;Fb=ja;Gb=ka;Hb=ma;Ca=109;break}case 9:{if(p>>>0<32){Ma=p;Oa=ja;Pa=ka;Na=ma;while(1){if(!Oa){na=la;oa=Ma;pa=0;qa=Pa;ra=Na;sa=n;ta=q;ua=h;break a}Da=Oa+-1|0;Sa=Na+1|0;Ra=(l[Na>>0]<<Ma)+Pa|0;Ma=Ma+8|0;if(Ma>>>0>=32){Ib=Da;Jb=Ra;Kb=Sa;break}else{Oa=Da;Pa=Ra;Na=Sa}}}else{Ib=ja;Jb=ka;Kb=ma}Na=ry(Jb|0)|0;k[x>>2]=Na;k[E>>2]=Na;k[f>>2]=10;Lb=0;Mb=Ib;Nb=0;Ob=Kb;Ca=121;break}case 10:{Lb=p;Mb=ja;Nb=ka;Ob=ma;Ca=121;break}case 11:{Pb=p;Qb=ja;Rb=ka;Sb=ma;Ca=124;break}case 12:{Tb=p;Ub=ja;Vb=ka;Wb=ma;Ca=125;break}case 13:{Na=p&7;Pa=ka>>>Na;Oa=p-Na|0;if(Oa>>>0<32){Na=Oa;Ma=ja;Sa=Pa;Ra=ma;while(1){if(!Ma){na=la;oa=Na;pa=0;qa=Sa;ra=Ra;sa=n;ta=q;ua=h;break a}Da=Ma+-1|0;Ta=Ra+1|0;Qa=(l[Ra>>0]<<Na)+Sa|0;Xb=Na+8|0;if(Xb>>>0<32){Na=Xb;Ma=Da;Sa=Qa;Ra=Ta}else{Yb=Xb;Zb=Da;_b=Qa;$b=Ta;break}}}else{Yb=Oa;Zb=ja;_b=Pa;$b=ma}Ra=_b&65535;if((Ra|0)==(_b>>>16^65535|0)){k[F>>2]=Ra;k[f>>2]=14;if(N){na=la;oa=0;pa=Zb;qa=0;ra=$b;sa=n;ta=q;ua=h;break a}else{ac=0;bc=Zb;cc=0;dc=$b;Ca=143;break b}}else{k[B>>2]=607185;k[f>>2]=29;Ea=Yb;Fa=Zb;Ga=_b;Ha=la;Ia=$b;Ja=n;Ka=q;La=h;break b}break}case 14:{ac=p;bc=ja;cc=ka;dc=ma;Ca=143;break}case 15:{ec=p;fc=ja;gc=ka;hc=ma;Ca=144;break}case 16:{if(p>>>0<14){Ra=p;Sa=ja;Ma=ka;Na=ma;while(1){if(!Sa){na=la;oa=Ra;pa=0;qa=Ma;ra=Na;sa=n;ta=q;ua=h;break a}Ta=Sa+-1|0;Qa=Na+1|0;Da=(l[Na>>0]<<Ra)+Ma|0;Xb=Ra+8|0;if(Xb>>>0<14){Ra=Xb;Sa=Ta;Ma=Da;Na=Qa}else{ic=Xb;jc=Ta;kc=Da;lc=Qa;break}}}else{ic=p;jc=ja;kc=ka;lc=ma}Na=(kc&31)+257|0;k[aa>>2]=Na;Ma=(kc>>>5&31)+1|0;k[ba>>2]=Ma;Sa=(kc>>>10&15)+4|0;k[ha>>2]=Sa;Ra=kc>>>14;Pa=ic+-14|0;if(Na>>>0>286|Ma>>>0>30){k[B>>2]=607214;k[f>>2]=29;Ea=Pa;Fa=jc;Ga=Ra;Ha=la;Ia=lc;Ja=n;Ka=q;La=h;break b}else{k[$>>2]=0;k[f>>2]=17;mc=Sa;nc=0;oc=Pa;pc=jc;qc=Ra;rc=lc;Ca=154;break b}break}case 17:{Ra=k[$>>2]|0;Pa=k[ha>>2]|0;if(Ra>>>0<Pa>>>0){mc=Pa;nc=Ra;oc=p;pc=ja;qc=ka;rc=ma;Ca=154}else{sc=Ra;tc=p;uc=ja;vc=ka;wc=ma;Ca=155}break}case 18:{xc=k[$>>2]|0;yc=p;zc=ja;Ac=ka;Bc=ma;Cc=h;Ca=165;break}case 19:{Dc=p;Ec=ja;Fc=ka;Gc=ma;Hc=h;Ca=202;break}case 20:{Ic=p;Jc=ja;Kc=ka;Lc=ma;Mc=h;Ca=203;break}case 21:{Nc=k[P>>2]|0;Oc=p;Pc=ja;Qc=ka;Rc=ma;Sc=h;Ca=221;break}case 22:{Tc=p;Uc=ja;Vc=ka;Wc=ma;Xc=h;Ca=228;break}case 23:{Yc=k[P>>2]|0;Zc=p;_c=ja;$c=ka;ad=ma;bd=h;Ca=240;break}case 24:{cd=p;dd=ja;ed=ka;fd=ma;gd=h;Ca=246;break}case 25:{if(!la){na=0;oa=p;pa=ja;qa=ka;ra=ma;sa=n;ta=q;ua=h;break a}i[q>>0]=k[F>>2];k[f>>2]=20;Ea=p;Fa=ja;Ga=ka;Ha=la+-1|0;Ia=ma;Ja=n;Ka=q+1|0;La=h;break}case 26:{if(k[w>>2]|0){if(p>>>0<32){Ra=p;Pa=ja;Sa=ka;Ma=ma;while(1){if(!Pa){na=la;oa=Ra;pa=0;qa=Sa;ra=Ma;sa=n;ta=q;ua=h;break a}Na=Pa+-1|0;Oa=Ma+1|0;Qa=(l[Ma>>0]<<Ra)+Sa|0;Da=Ra+8|0;if(Da>>>0<32){Ra=Da;Pa=Na;Sa=Qa;Ma=Oa}else{hd=Da;id=Na;jd=Qa;kd=Oa;break}}}else{hd=p;id=ja;jd=ka;kd=ma}Ma=n-la|0;k[X>>2]=(k[X>>2]|0)+Ma;k[Y>>2]=(k[Y>>2]|0)+Ma;if((n|0)==(la|0))ld=k[x>>2]|0;else{Sa=k[x>>2]|0;Pa=q+(0-Ma)|0;if(!(k[z>>2]|0))md=zg(Sa,Pa,Ma)|0;else md=Ag(Sa,Pa,Ma)|0;k[x>>2]=md;k[E>>2]=md;ld=md}Ma=(k[z>>2]|0)==0;Pa=ry(jd|0)|0;if(((Ma?Pa:jd)|0)==(ld|0)){nd=0;od=id;pd=0;qd=kd;rd=la}else{k[B>>2]=607468;k[f>>2]=29;Ea=hd;Fa=id;Ga=jd;Ha=la;Ia=kd;Ja=la;Ka=q;La=h;break b}}else{nd=p;od=ja;pd=ka;qd=ma;rd=n}k[f>>2]=27;sd=nd;td=od;ud=pd;vd=qd;wd=rd;Ca=276;break}case 27:{sd=p;td=ja;ud=ka;vd=ma;wd=n;Ca=276;break}default:{Ca=296;break a}}while(0);if((Ca|0)==47)while(1){Ca=0;if(!bb){na=la;oa=ab;pa=0;qa=cb;ra=db;sa=n;ta=q;ua=h;break a}Pa=bb+-1|0;Ma=db+1|0;Sa=(l[db>>0]<<ab)+cb|0;ab=ab+8|0;if(ab>>>0>=32){eb=Pa;fb=Sa;gb=Ma;Ca=49;break}else{bb=Pa;cb=Sa;db=Ma;Ca=47}}else if((Ca|0)==121){Ca=0;if(!(k[G>>2]|0)){xd=Lb;yd=Mb;zd=Nb;Ad=la;Bd=Ob;Cd=q;Ca=122;break}Ma=zg(0,0,0)|0;k[x>>2]=Ma;k[E>>2]=Ma;k[f>>2]=11;Pb=Lb;Qb=Mb;Rb=Nb;Sb=Ob;Ca=124}else if((Ca|0)==143){Ca=0;k[f>>2]=15;ec=ac;fc=bc;gc=cc;hc=dc;Ca=144}else if((Ca|0)==154){Ca=0;Ma=nc;Sa=oc;Pa=pc;Ra=qc;Oa=rc;while(1){if(Sa>>>0<3){Qa=Sa;Na=Pa;Da=Ra;Ta=Oa;while(1){if(!Na){na=la;oa=Qa;pa=0;qa=Da;ra=Ta;sa=n;ta=q;ua=h;break a}Xb=Na+-1|0;Dd=Ta+1|0;Ed=(l[Ta>>0]<<Qa)+Da|0;Fd=Qa+8|0;if(Fd>>>0<3){Qa=Fd;Na=Xb;Da=Ed;Ta=Dd}else{Gd=Fd;Hd=Xb;Id=Ed;Jd=Dd;break}}}else{Gd=Sa;Hd=Pa;Id=Ra;Jd=Oa}Ta=Ma+1|0;k[$>>2]=Ta;j[f+112+(m[128152+(Ma<<1)>>1]<<1)>>1]=Id&7;Da=Id>>>3;Na=Gd+-3|0;if(Ta>>>0<mc>>>0){Ma=Ta;Sa=Na;Pa=Hd;Ra=Da;Oa=Jd}else{sc=Ta;tc=Na;uc=Hd;vc=Da;wc=Jd;Ca=155;break}}}else if((Ca|0)==276){Ca=0;if(!(k[w>>2]|0)){Kd=sd;Ld=td;Md=ud;Nd=la;Od=vd;Pd=wd;Qd=q;Ca=283;break}if(!(k[z>>2]|0)){Kd=sd;Ld=td;Md=ud;Nd=la;Od=vd;Pd=wd;Qd=q;Ca=283;break}if(sd>>>0<32){Oa=sd;Ra=td;Pa=ud;Sa=vd;while(1){if(!Ra){na=la;oa=Oa;pa=0;qa=Pa;ra=Sa;sa=wd;ta=q;ua=h;break a}Ma=Ra+-1|0;Da=Sa+1|0;Na=(l[Sa>>0]<<Oa)+Pa|0;Ta=Oa+8|0;if(Ta>>>0<32){Oa=Ta;Ra=Ma;Pa=Na;Sa=Da}else{Rd=Ta;Sd=Ma;Td=Na;Ud=Da;break}}}else{Rd=sd;Sd=td;Td=ud;Ud=vd}if((Td|0)==(k[Y>>2]|0)){Kd=0;Ld=Sd;Md=0;Nd=la;Od=Ud;Pd=wd;Qd=q;Ca=283;break}k[B>>2]=607489;k[f>>2]=29;Ea=Rd;Fa=Sd;Ga=Td;Ha=la;Ia=Ud;Ja=wd;Ka=q;La=h}do if((Ca|0)==49){Ca=0;Sa=k[A>>2]|0;if(Sa)k[Sa+4>>2]=fb;if(k[z>>2]&512){i[d>>0]=fb;i[y>>0]=fb>>>8;i[Z>>0]=fb>>>16;i[_>>0]=fb>>>24;Sa=Ag(k[x>>2]|0,d,4)|0;k[x>>2]=Sa}k[f>>2]=3;hb=0;ib=eb;jb=0;kb=gb;Ca=55}else if((Ca|0)==124){Ca=0;if(H){na=la;oa=Pb;pa=Qb;qa=Rb;ra=Sb;sa=n;ta=q;ua=h;break a}else{Tb=Pb;Ub=Qb;Vb=Rb;Wb=Sb;Ca=125}}else if((Ca|0)==144){Ca=0;Sa=k[F>>2]|0;if(!Sa){k[f>>2]=11;Ea=ec;Fa=fc;Ga=gc;Ha=la;Ia=hc;Ja=n;Ka=q;La=h;break}Pa=Sa>>>0>fc>>>0?fc:Sa;Sa=Pa>>>0>la>>>0?la:Pa;if(!Sa){na=la;oa=ec;pa=fc;qa=gc;ra=hc;sa=n;ta=q;ua=h;break a}py(q|0,hc|0,Sa|0)|0;k[F>>2]=(k[F>>2]|0)-Sa;Ea=ec;Fa=fc-Sa|0;Ga=gc;Ha=la-Sa|0;Ia=hc+Sa|0;Ja=n;Ka=q+Sa|0;La=h}else if((Ca|0)==155){Ca=0;if(sc>>>0<19){Sa=sc;do{j[f+112+(m[128152+(Sa<<1)>>1]<<1)>>1]=0;Sa=Sa+1|0}while((Sa|0)!=19);k[$>>2]=19}k[ea>>2]=da;k[J>>2]=da;k[K>>2]=7;Sa=eh(0,fa,19,ea,K,ga)|0;if(!Sa){k[$>>2]=0;k[f>>2]=18;xc=0;yc=tc;zc=uc;Ac=vc;Bc=wc;Cc=0;Ca=165;break}else{k[B>>2]=607250;k[f>>2]=29;Ea=tc;Fa=uc;Ga=vc;Ha=la;Ia=wc;Ja=n;Ka=q;La=Sa;break}}while(0);c:do if((Ca|0)==55)while(1){Ca=0;if(!ib){na=la;oa=hb;pa=0;qa=jb;ra=kb;sa=n;ta=q;ua=h;break a}Sa=ib+-1|0;Pa=kb+1|0;Ra=(l[kb>>0]<<hb)+jb|0;hb=hb+8|0;if(hb>>>0>=16){lb=Sa;mb=Ra;nb=Pa;Ca=57;break}else{ib=Sa;jb=Ra;kb=Pa;Ca=55}}else if((Ca|0)==125){Ca=0;if(k[I>>2]|0){Pa=Tb&7;k[f>>2]=26;Ea=Tb-Pa|0;Fa=Ub;Ga=Vb>>>Pa;Ha=la;Ia=Wb;Ja=n;Ka=q;La=h;break}if(Tb>>>0<3){Pa=Tb;Ra=Ub;Sa=Vb;Oa=Wb;while(1){if(!Ra){na=la;oa=Pa;pa=0;qa=Sa;ra=Oa;sa=n;ta=q;ua=h;break a}Da=Ra+-1|0;Na=Oa+1|0;Ma=(l[Oa>>0]<<Pa)+Sa|0;Ta=Pa+8|0;if(Ta>>>0<3){Pa=Ta;Ra=Da;Sa=Ma;Oa=Na}else{Vd=Ta;Wd=Da;Xd=Ma;Yd=Na;break}}}else{Vd=Tb;Wd=Ub;Xd=Vb;Yd=Wb}k[I>>2]=Xd&1;switch(Xd>>>1&3|0){case 0:{k[f>>2]=13;break}case 1:{k[J>>2]=125976;k[K>>2]=9;k[L>>2]=128024;k[M>>2]=5;k[f>>2]=19;if(N){Zd=Vd;_d=Wd;$d=Xd;ae=la;be=Yd;ce=n;de=q;ee=h;Ca=133;break a}break}case 2:{k[f>>2]=16;break}case 3:{k[B>>2]=607166;k[f>>2]=29;break}default:{}}Ea=Vd+-3|0;Fa=Wd;Ga=Xd>>>3;Ha=la;Ia=Yd;Ja=n;Ka=q;La=h}else if((Ca|0)==165){Ca=0;Oa=k[aa>>2]|0;Sa=k[ba>>2]|0;do if(xc>>>0<(Sa+Oa|0)>>>0){Ra=xc;Pa=Sa;Na=Oa;Ma=yc;Da=zc;Ta=Ac;Qa=Bc;d:while(1){Dd=(1<<k[K>>2])+-1|0;Ed=Dd&Ta;Xb=k[J>>2]|0;Fd=l[Xb+(Ed<<2)+1>>0]|0;if(Fd>>>0>Ma>>>0){fe=Ma;ge=Da;he=Ta;ie=Qa;while(1){if(!ge){na=la;oa=fe;pa=0;qa=he;ra=ie;sa=n;ta=q;ua=Cc;break a}je=ge+-1|0;ke=ie+1|0;le=(l[ie>>0]<<fe)+he|0;me=fe+8|0;ne=Dd&le;oe=l[Xb+(ne<<2)+1>>0]|0;if(oe>>>0>me>>>0){fe=me;ge=je;he=le;ie=ke}else{pe=oe;qe=ne;re=me;se=je;te=le;ue=ke;break}}}else{pe=Fd;qe=Ed;re=Ma;se=Da;te=Ta;ue=Qa}ie=j[Xb+(qe<<2)+2>>1]|0;if((ie&65535)>=16){switch(ie<<16>>16){case 16:{he=pe+2|0;if(re>>>0<he>>>0){ge=re;fe=se;Dd=te;ke=ue;while(1){if(!fe){na=la;oa=ge;pa=0;qa=Dd;ra=ke;sa=n;ta=q;ua=Cc;break a}le=fe+-1|0;je=ke+1|0;me=(l[ke>>0]<<ge)+Dd|0;ne=ge+8|0;if(ne>>>0<he>>>0){ge=ne;fe=le;Dd=me;ke=je}else{ve=ne;we=le;xe=me;ye=je;break}}}else{ve=re;we=se;xe=te;ye=ue}ke=xe>>>pe;Dd=ve-pe|0;if(!Ra){ze=ke;Ae=Dd;Be=we;Ce=ye;Ca=182;break d}De=Dd+-2|0;Ee=(ke&3)+3|0;Fe=we;Ge=ke>>>2;He=m[f+112+(Ra+-1<<1)>>1]|0;Ie=ye;break}case 17:{ke=pe+3|0;if(re>>>0<ke>>>0){Dd=re;fe=se;ge=te;he=ue;while(1){if(!fe){na=la;oa=Dd;pa=0;qa=ge;ra=he;sa=n;ta=q;ua=Cc;break a}Xb=fe+-1|0;Ed=he+1|0;Fd=(l[he>>0]<<Dd)+ge|0;je=Dd+8|0;if(je>>>0<ke>>>0){Dd=je;fe=Xb;ge=Fd;he=Ed}else{Je=je;Ke=Xb;Le=Fd;Me=Ed;break}}}else{Je=re;Ke=se;Le=te;Me=ue}he=Le>>>pe;De=-3-pe+Je|0;Ee=(he&7)+3|0;Fe=Ke;Ge=he>>>3;He=0;Ie=Me;break}default:{he=pe+7|0;if(re>>>0<he>>>0){ge=re;fe=se;Dd=te;ke=ue;while(1){if(!fe){na=la;oa=ge;pa=0;qa=Dd;ra=ke;sa=n;ta=q;ua=Cc;break a}Ed=fe+-1|0;Fd=ke+1|0;Xb=(l[ke>>0]<<ge)+Dd|0;je=ge+8|0;if(je>>>0<he>>>0){ge=je;fe=Ed;Dd=Xb;ke=Fd}else{Ne=je;Oe=Ed;Pe=Xb;Qe=Fd;break}}}else{Ne=re;Oe=se;Pe=te;Qe=ue}ke=Pe>>>pe;De=-7-pe+Ne|0;Ee=(ke&127)+11|0;Fe=Oe;Ge=ke>>>7;He=0;Ie=Qe}}if((Ra+Ee|0)>>>0>(Pa+Na|0)>>>0){Re=De;Se=Fe;Te=Ge;Ue=Ie;Ca=192;break}ke=He&65535;Dd=Ee+-1|0;k[$>>2]=Ra+1;j[f+112+(Ra<<1)>>1]=ke;if(!Dd){Ve=De;We=Fe;Xe=Ge;Ye=Ie}else{fe=Dd;do{Dd=k[$>>2]|0;fe=fe+-1|0;k[$>>2]=Dd+1;j[f+112+(Dd<<1)>>1]=ke}while((fe|0)!=0);Ve=De;We=Fe;Xe=Ge;Ye=Ie}}else{if(re>>>0<pe>>>0){fe=re;ke=se;Dd=te;ge=ue;while(1){if(!ke){na=la;oa=fe;pa=0;qa=Dd;ra=ge;sa=n;ta=q;ua=Cc;break a}he=ke+-1|0;Fd=ge+1|0;Xb=(l[ge>>0]<<fe)+Dd|0;Ed=fe+8|0;if(Ed>>>0<pe>>>0){fe=Ed;ke=he;Dd=Xb;ge=Fd}else{Ze=Ed;_e=he;$e=Xb;af=Fd;break}}}else{Ze=re;_e=se;$e=te;af=ue}k[$>>2]=Ra+1;j[f+112+(Ra<<1)>>1]=ie;Ve=Ze-pe|0;We=_e;Xe=$e>>>pe;Ye=af}Ra=k[$>>2]|0;ge=k[aa>>2]|0;Pa=k[ba>>2]|0;if(Ra>>>0>=(Pa+ge|0)>>>0){bf=ge;cf=Ve;df=We;ef=Xe;ff=Ye;Ca=194;break}else{Na=ge;Ma=Ve;Da=We;Ta=Xe;Qa=Ye}}if((Ca|0)==182){Ca=0;k[B>>2]=607275;k[f>>2]=29;Ea=Ae;Fa=Be;Ga=ze;Ha=la;Ia=Ce;Ja=n;Ka=q;La=Cc;break c}else if((Ca|0)==192){Ca=0;k[B>>2]=607275;k[f>>2]=29;Ea=Re;Fa=Se;Ga=Te;Ha=la;Ia=Ue;Ja=n;Ka=q;La=Cc;break c}else if((Ca|0)==194){Ca=0;if((k[f>>2]|0)==29){Ea=cf;Fa=df;Ga=ef;Ha=la;Ia=ff;Ja=n;Ka=q;La=Cc;break c}else{gf=bf;hf=cf;jf=df;kf=ef;lf=ff;break}}}else{gf=Oa;hf=yc;jf=zc;kf=Ac;lf=Bc}while(0);if(!(j[ca>>1]|0)){k[B>>2]=607301;k[f>>2]=29;Ea=hf;Fa=jf;Ga=kf;Ha=la;Ia=lf;Ja=n;Ka=q;La=Cc;break}k[ea>>2]=da;k[J>>2]=da;k[K>>2]=9;Oa=eh(1,fa,gf,ea,K,ga)|0;if(Oa){k[B>>2]=607338;k[f>>2]=29;Ea=hf;Fa=jf;Ga=kf;Ha=la;Ia=lf;Ja=n;Ka=q;La=Oa;break}k[L>>2]=k[ea>>2];k[M>>2]=6;Oa=eh(2,f+112+(k[aa>>2]<<1)|0,k[ba>>2]|0,ea,M,ga)|0;if(!Oa){k[f>>2]=19;if(N){na=la;oa=hf;pa=jf;qa=kf;ra=lf;sa=n;ta=q;ua=0;break a}else{Dc=hf;Ec=jf;Fc=kf;Gc=lf;Hc=0;Ca=202;break}}else{k[B>>2]=607366;k[f>>2]=29;Ea=hf;Fa=jf;Ga=kf;Ha=la;Ia=lf;Ja=n;Ka=q;La=Oa;break}}while(0);if((Ca|0)==57){Ca=0;Oa=k[A>>2]|0;if(Oa){k[Oa+8>>2]=mb&255;k[Oa+12>>2]=mb>>>8}if(k[z>>2]&512){i[d>>0]=mb;i[y>>0]=mb>>>8;Oa=Ag(k[x>>2]|0,d,2)|0;k[x>>2]=Oa}k[f>>2]=4;ob=0;pb=lb;qb=0;rb=nb;Ca=62}else if((Ca|0)==202){Ca=0;k[f>>2]=20;Ic=Dc;Jc=Ec;Kc=Fc;Lc=Gc;Mc=Hc;Ca=203}do if((Ca|0)==62){Ca=0;Oa=k[z>>2]|0;if(!(Oa&1024)){Sa=k[A>>2]|0;if(!Sa){mf=ob;nf=pb;of=qb;pf=rb}else{k[Sa+16>>2]=0;mf=ob;nf=pb;of=qb;pf=rb}}else{if(ob>>>0<16){Sa=ob;Qa=pb;Ta=qb;Da=rb;while(1){if(!Qa){na=la;oa=Sa;pa=0;qa=Ta;ra=Da;sa=n;ta=q;ua=h;break a}Ma=Qa+-1|0;Na=Da+1|0;Pa=(l[Da>>0]<<Sa)+Ta|0;Sa=Sa+8|0;if(Sa>>>0>=16){qf=Ma;rf=Pa;sf=Na;break}else{Qa=Ma;Ta=Pa;Da=Na}}}else{qf=pb;rf=qb;sf=rb}k[F>>2]=rf;Da=k[A>>2]|0;if(Da)k[Da+20>>2]=rf;if(!(Oa&512)){mf=0;nf=qf;of=0;pf=sf}else{i[d>>0]=rf;i[y>>0]=rf>>>8;Da=Ag(k[x>>2]|0,d,2)|0;k[x>>2]=Da;mf=0;nf=qf;of=0;pf=sf}}k[f>>2]=5;sb=mf;tb=nf;ub=of;vb=pf;Ca=73}else if((Ca|0)==203){Ca=0;if(la>>>0>257&Jc>>>0>5){k[g>>2]=q;k[o>>2]=la;k[a>>2]=Lc;k[s>>2]=Jc;k[u>>2]=Kc;k[v>>2]=Ic;Zg(a,n);Da=k[g>>2]|0;Ta=k[o>>2]|0;Qa=k[a>>2]|0;Sa=k[s>>2]|0;Na=k[u>>2]|0;Pa=k[v>>2]|0;if((k[f>>2]|0)!=11){Ea=Pa;Fa=Sa;Ga=Na;Ha=Ta;Ia=Qa;Ja=n;Ka=Da;La=Mc;break}k[O>>2]=-1;Ea=Pa;Fa=Sa;Ga=Na;Ha=Ta;Ia=Qa;Ja=n;Ka=Da;La=Mc;break}k[O>>2]=0;Da=(1<<k[K>>2])+-1|0;Qa=Da&Kc;Ta=k[J>>2]|0;Na=i[Ta+(Qa<<2)+1>>0]|0;Sa=Na&255;if(Sa>>>0>Ic>>>0){Pa=Ic;Ma=Jc;Ra=Kc;ge=Lc;while(1){if(!Ma){na=la;oa=Pa;pa=0;qa=Ra;ra=ge;sa=n;ta=q;ua=Mc;break a}Dd=Ma+-1|0;ke=ge+1|0;fe=(l[ge>>0]<<Pa)+Ra|0;Fd=Pa+8|0;Xb=Da&fe;he=i[Ta+(Xb<<2)+1>>0]|0;Ed=he&255;if(Ed>>>0>Fd>>>0){Pa=Fd;Ma=Dd;Ra=fe;ge=ke}else{tf=he;uf=Ed;vf=Xb;wf=Fd;xf=Dd;yf=fe;zf=ke;break}}}else{tf=Na;uf=Sa;vf=Qa;wf=Ic;xf=Jc;yf=Kc;zf=Lc}ge=i[Ta+(vf<<2)>>0]|0;Ra=j[Ta+(vf<<2)+2>>1]|0;Ma=ge&255;if(ge<<24>>24!=0&(Ma&240|0)==0){Pa=Ra&65535;Da=(1<<uf+Ma)+-1|0;Ma=((yf&Da)>>>uf)+Pa|0;Oa=i[Ta+(Ma<<2)+1>>0]|0;if(((Oa&255)+uf|0)>>>0>wf>>>0){ke=wf;fe=xf;Dd=yf;Fd=zf;while(1){if(!fe){na=la;oa=ke;pa=0;qa=Dd;ra=Fd;sa=n;ta=q;ua=Mc;break a}Xb=fe+-1|0;Ed=Fd+1|0;he=(l[Fd>>0]<<ke)+Dd|0;je=ke+8|0;me=((he&Da)>>>uf)+Pa|0;le=i[Ta+(me<<2)+1>>0]|0;if(((le&255)+uf|0)>>>0>je>>>0){ke=je;fe=Xb;Dd=he;Fd=Ed}else{Af=me;Bf=le;Cf=je;Df=Xb;Ef=he;Ff=Ed;break}}}else{Af=Ma;Bf=Oa;Cf=wf;Df=xf;Ef=yf;Ff=zf}Fd=j[Ta+(Af<<2)+2>>1]|0;Dd=i[Ta+(Af<<2)>>0]|0;k[O>>2]=uf;Gf=uf;Hf=Cf-uf|0;If=Df;Jf=Dd;Kf=Bf;Lf=Fd;Mf=Ef>>>uf;Nf=Ff}else{Gf=0;Hf=wf;If=xf;Jf=ge;Kf=tf;Lf=Ra;Mf=yf;Nf=zf}Fd=Kf&255;Dd=Mf>>>Fd;fe=Hf-Fd|0;k[O>>2]=Gf+Fd;k[F>>2]=Lf&65535;Fd=Jf&255;if(!(Jf<<24>>24)){k[f>>2]=25;Ea=fe;Fa=If;Ga=Dd;Ha=la;Ia=Nf;Ja=n;Ka=q;La=Mc;break}if(Fd&32){k[O>>2]=-1;k[f>>2]=11;Ea=fe;Fa=If;Ga=Dd;Ha=la;Ia=Nf;Ja=n;Ka=q;La=Mc;break}if(!(Fd&64)){ke=Fd&15;k[P>>2]=ke;k[f>>2]=21;Nc=ke;Oc=fe;Pc=If;Qc=Dd;Rc=Nf;Sc=Mc;Ca=221;break}else{k[B>>2]=607388;k[f>>2]=29;Ea=fe;Fa=If;Ga=Dd;Ha=la;Ia=Nf;Ja=n;Ka=q;La=Mc;break}}while(0);if((Ca|0)==73){Ca=0;Dd=k[z>>2]|0;if(Dd&1024){fe=k[F>>2]|0;ke=fe>>>0>tb>>>0?tb:fe;if(!ke){Of=fe;Pf=tb;Qf=vb}else{Fd=k[A>>2]|0;if((Fd|0)!=0?(Pa=k[Fd+16>>2]|0,(Pa|0)!=0):0){Da=(k[Fd+20>>2]|0)-fe|0;fe=k[Fd+24>>2]|0;py(Pa+Da|0,vb|0,((Da+ke|0)>>>0>fe>>>0?fe-Da|0:ke)|0)|0;Rf=k[z>>2]|0}else Rf=Dd;if(Rf&512){Dd=Ag(k[x>>2]|0,vb,ke)|0;k[x>>2]=Dd}Dd=(k[F>>2]|0)-ke|0;k[F>>2]=Dd;Of=Dd;Pf=tb-ke|0;Qf=vb+ke|0}if(!Of){Sf=Pf;Tf=Qf}else{na=la;oa=sb;pa=Pf;qa=ub;ra=Qf;sa=n;ta=q;ua=h;break}}else{Sf=tb;Tf=vb}k[F>>2]=0;k[f>>2]=6;wb=sb;xb=Sf;yb=ub;zb=Tf;Ca=83}else if((Ca|0)==221){Ca=0;if(!Nc){Uf=k[F>>2]|0;Vf=Oc;Wf=Pc;Xf=Qc;Yf=Rc}else{if(Oc>>>0<Nc>>>0){ke=Oc;Dd=Pc;Da=Qc;fe=Rc;while(1){if(!Dd){na=la;oa=ke;pa=0;qa=Da;ra=fe;sa=n;ta=q;ua=Sc;break a}Pa=Dd+-1|0;Fd=fe+1|0;Qa=(l[fe>>0]<<ke)+Da|0;Sa=ke+8|0;if(Sa>>>0<Nc>>>0){ke=Sa;Dd=Pa;Da=Qa;fe=Fd}else{Zf=Sa;_f=Pa;$f=Qa;ag=Fd;break}}}else{Zf=Oc;_f=Pc;$f=Qc;ag=Rc}fe=(k[F>>2]|0)+((1<<Nc)+-1&$f)|0;k[F>>2]=fe;k[O>>2]=(k[O>>2]|0)+Nc;Uf=fe;Vf=Zf-Nc|0;Wf=_f;Xf=$f>>>Nc;Yf=ag}k[Q>>2]=Uf;k[f>>2]=22;Tc=Vf;Uc=Wf;Vc=Xf;Wc=Yf;Xc=Sc;Ca=228}do if((Ca|0)==83){Ca=0;if(!(k[z>>2]&2048)){fe=k[A>>2]|0;if(!fe){bg=xb;cg=zb}else{k[fe+28>>2]=0;bg=xb;cg=zb}}else{if(!xb){na=la;oa=wb;pa=0;qa=yb;ra=zb;sa=n;ta=q;ua=h;break a}else dg=0;while(1){fe=dg+1|0;Da=i[zb+dg>>0]|0;Dd=k[A>>2]|0;if(((Dd|0)!=0?(ke=k[Dd+28>>2]|0,(ke|0)!=0):0)?(Fd=k[F>>2]|0,Fd>>>0<(k[Dd+32>>2]|0)>>>0):0){k[F>>2]=Fd+1;i[ke+Fd>>0]=Da}Fd=Da<<24>>24!=0;if(Fd&xb>>>0>fe>>>0)dg=fe;else{eg=fe;fg=Fd;break}}if(k[z>>2]&512){Ra=Ag(k[x>>2]|0,zb,eg)|0;k[x>>2]=Ra}Ra=xb-eg|0;ge=zb+eg|0;if(fg){na=la;oa=wb;pa=Ra;qa=yb;ra=ge;sa=n;ta=q;ua=h;break a}else{bg=Ra;cg=ge}}k[F>>2]=0;k[f>>2]=7;Ab=wb;Bb=bg;Cb=yb;Db=cg;Ca=96}else if((Ca|0)==228){Ca=0;ge=(1<<k[M>>2])+-1|0;Ra=ge&Vc;Ta=k[L>>2]|0;Oa=i[Ta+(Ra<<2)+1>>0]|0;Ma=Oa&255;if(Ma>>>0>Tc>>>0){Fd=Tc;fe=Uc;Da=Vc;ke=Wc;while(1){if(!fe){na=la;oa=Fd;pa=0;qa=Da;ra=ke;sa=n;ta=q;ua=Xc;break a}Dd=fe+-1|0;Qa=ke+1|0;Pa=(l[ke>>0]<<Fd)+Da|0;Sa=Fd+8|0;Na=ge&Pa;Ed=i[Ta+(Na<<2)+1>>0]|0;he=Ed&255;if(he>>>0>Sa>>>0){Fd=Sa;fe=Dd;Da=Pa;ke=Qa}else{gg=Ed;hg=he;ig=Na;jg=Sa;kg=Dd;lg=Pa;mg=Qa;break}}}else{gg=Oa;hg=Ma;ig=Ra;jg=Tc;kg=Uc;lg=Vc;mg=Wc}ke=i[Ta+(ig<<2)>>0]|0;Da=j[Ta+(ig<<2)+2>>1]|0;fe=ke&255;if(!(fe&240)){Fd=Da&65535;ge=(1<<hg+fe)+-1|0;fe=((lg&ge)>>>hg)+Fd|0;Qa=i[Ta+(fe<<2)+1>>0]|0;if(((Qa&255)+hg|0)>>>0>jg>>>0){Pa=jg;Dd=kg;Sa=lg;Na=mg;while(1){if(!Dd){na=la;oa=Pa;pa=0;qa=Sa;ra=Na;sa=n;ta=q;ua=Xc;break a}he=Dd+-1|0;Ed=Na+1|0;Xb=(l[Na>>0]<<Pa)+Sa|0;je=Pa+8|0;le=((Xb&ge)>>>hg)+Fd|0;me=i[Ta+(le<<2)+1>>0]|0;if(((me&255)+hg|0)>>>0>je>>>0){Pa=je;Dd=he;Sa=Xb;Na=Ed}else{ng=le;og=me;pg=je;qg=he;rg=Xb;sg=Ed;break}}}else{ng=fe;og=Qa;pg=jg;qg=kg;rg=lg;sg=mg}Na=j[Ta+(ng<<2)+2>>1]|0;Sa=i[Ta+(ng<<2)>>0]|0;Dd=(k[O>>2]|0)+hg|0;k[O>>2]=Dd;tg=Dd;ug=pg-hg|0;vg=qg;wg=Sa;xg=og;yg=Na;Bg=rg>>>hg;Cg=sg}else{tg=k[O>>2]|0;ug=jg;vg=kg;wg=ke;xg=gg;yg=Da;Bg=lg;Cg=mg}Na=xg&255;Sa=Bg>>>Na;Dd=ug-Na|0;k[O>>2]=tg+Na;Na=wg&255;if(!(Na&64)){k[R>>2]=yg&65535;Pa=Na&15;k[P>>2]=Pa;k[f>>2]=23;Yc=Pa;Zc=Dd;_c=vg;$c=Sa;ad=Cg;bd=Xc;Ca=240;break}else{k[B>>2]=607416;k[f>>2]=29;Ea=Dd;Fa=vg;Ga=Sa;Ha=la;Ia=Cg;Ja=n;Ka=q;La=Xc;break}}while(0);if((Ca|0)==96){Ca=0;if(!(k[z>>2]&4096)){Sa=k[A>>2]|0;if(!Sa){Dg=Bb;Eg=Db}else{k[Sa+36>>2]=0;Dg=Bb;Eg=Db}}else{if(!Bb){na=la;oa=Ab;pa=0;qa=Cb;ra=Db;sa=n;ta=q;ua=h;break}else Fg=0;while(1){Sa=Fg+1|0;Dd=i[Db+Fg>>0]|0;Pa=k[A>>2]|0;if(((Pa|0)!=0?(Na=k[Pa+36>>2]|0,(Na|0)!=0):0)?(Fd=k[F>>2]|0,Fd>>>0<(k[Pa+40>>2]|0)>>>0):0){k[F>>2]=Fd+1;i[Na+Fd>>0]=Dd}Fd=Dd<<24>>24!=0;if(Fd&Bb>>>0>Sa>>>0)Fg=Sa;else{Gg=Sa;Hg=Fd;break}}if(k[z>>2]&512){Fd=Ag(k[x>>2]|0,Db,Gg)|0;k[x>>2]=Fd}Fd=Bb-Gg|0;Sa=Db+Gg|0;if(Hg){na=la;oa=Ab;pa=Fd;qa=Cb;ra=Sa;sa=n;ta=q;ua=h;break}else{Dg=Fd;Eg=Sa}}k[f>>2]=8;Eb=Ab;Fb=Dg;Gb=Cb;Hb=Eg;Ca=109}else if((Ca|0)==240){Ca=0;if(!Yc){Ig=Zc;Jg=_c;Kg=$c;Lg=ad}else{if(Zc>>>0<Yc>>>0){Sa=Zc;Fd=_c;Dd=$c;Na=ad;while(1){if(!Fd){na=la;oa=Sa;pa=0;qa=Dd;ra=Na;sa=n;ta=q;ua=bd;break a}Pa=Fd+-1|0;ge=Na+1|0;Ra=(l[Na>>0]<<Sa)+Dd|0;Ma=Sa+8|0;if(Ma>>>0<Yc>>>0){Sa=Ma;Fd=Pa;Dd=Ra;Na=ge}else{Mg=Ma;Ng=Pa;Og=Ra;Pg=ge;break}}}else{Mg=Zc;Ng=_c;Og=$c;Pg=ad}k[R>>2]=(k[R>>2]|0)+((1<<Yc)+-1&Og);k[O>>2]=(k[O>>2]|0)+Yc;Ig=Mg-Yc|0;Jg=Ng;Kg=Og>>>Yc;Lg=Pg}k[f>>2]=24;cd=Ig;dd=Jg;ed=Kg;fd=Lg;gd=bd;Ca=246}do if((Ca|0)==109){Ca=0;Na=k[z>>2]|0;if(Na&512){if(Eb>>>0<16){Dd=Eb;Fd=Fb;Sa=Gb;ge=Hb;while(1){if(!Fd){na=la;oa=Dd;pa=0;qa=Sa;ra=ge;sa=n;ta=q;ua=h;break a}Ra=Fd+-1|0;Pa=ge+1|0;Ma=(l[ge>>0]<<Dd)+Sa|0;Oa=Dd+8|0;if(Oa>>>0<16){Dd=Oa;Fd=Ra;Sa=Ma;ge=Pa}else{Qg=Oa;Rg=Ra;Sg=Ma;Tg=Pa;break}}}else{Qg=Eb;Rg=Fb;Sg=Gb;Tg=Hb}if((Sg|0)==(k[x>>2]&65535|0)){Ug=0;Vg=Rg;Wg=0;Xg=Tg}else{k[B>>2]=607146;k[f>>2]=29;Ea=Qg;Fa=Rg;Ga=Sg;Ha=la;Ia=Tg;Ja=n;Ka=q;La=h;break}}else{Ug=Eb;Vg=Fb;Wg=Gb;Xg=Hb}ge=k[A>>2]|0;if(ge){k[ge+44>>2]=Na>>>9&1;k[ge+48>>2]=1}ge=Ag(0,0,0)|0;k[x>>2]=ge;k[E>>2]=ge;k[f>>2]=11;Ea=Ug;Fa=Vg;Ga=Wg;Ha=la;Ia=Xg;Ja=n;Ka=q;La=h}else if((Ca|0)==246){Ca=0;if(!la){na=0;oa=cd;pa=dd;qa=ed;ra=fd;sa=n;ta=q;ua=gd;break a}ge=n-la|0;Sa=k[R>>2]|0;if(Sa>>>0>ge>>>0){Fd=Sa-ge|0;if(Fd>>>0>(k[S>>2]|0)>>>0?(k[T>>2]|0)!=0:0){k[B>>2]=607438;k[f>>2]=29;Ea=cd;Fa=dd;Ga=ed;Ha=la;Ia=fd;Ja=n;Ka=q;La=gd;break}ge=k[U>>2]|0;if(Fd>>>0>ge>>>0){Dd=Fd-ge|0;Yg=Dd;_g=(k[V>>2]|0)+((k[W>>2]|0)-Dd)|0}else{Yg=Fd;_g=(k[V>>2]|0)+(ge-Fd)|0}Fd=k[F>>2]|0;$g=Fd;ah=Yg>>>0>Fd>>>0?Fd:Yg;bh=_g}else{Fd=k[F>>2]|0;$g=Fd;ah=Fd;bh=q+(0-Sa)|0}Sa=ah>>>0>la>>>0?la:ah;k[F>>2]=$g-Sa;Fd=ah>>>0>la>>>0?la:ah;ge=Sa;Dd=bh;Da=q;while(1){i[Da>>0]=i[Dd>>0]|0;ge=ge+-1|0;if(!ge)break;else{Dd=Dd+1|0;Da=Da+1|0}}Da=la-Sa|0;Dd=q+Fd|0;if(!(k[F>>2]|0)){k[f>>2]=20;Ea=cd;Fa=dd;Ga=ed;Ha=Da;Ia=fd;Ja=n;Ka=Dd;La=gd}else{Ea=cd;Fa=dd;Ga=ed;Ha=Da;Ia=fd;Ja=n;Ka=Dd;La=gd}}while(0);ia=k[f>>2]|0;p=Ea;ja=Fa;ka=Ga;la=Ha;ma=Ia;n=Ja;q=Ka;h=La}if((Ca|0)==122){k[g>>2]=Cd;k[o>>2]=Ad;k[a>>2]=Bd;k[s>>2]=yd;k[u>>2]=zd;k[v>>2]=xd;e=2;r=c;return e|0}else if((Ca|0)==133){na=ae;oa=Zd+-3|0;pa=_d;qa=$d>>>3;ra=be;sa=ce;ta=de;ua=ee}else if((Ca|0)==283){k[f>>2]=28;na=Nd;oa=Kd;pa=Ld;qa=Md;ra=Od;sa=Pd;ta=Qd;ua=1}else if((Ca|0)==284){na=ya;oa=va;pa=wa;qa=xa;ra=za;sa=Aa;ta=Ba;ua=-3}else if((Ca|0)==296){e=-2;r=c;return e|0}else if((Ca|0)==297){r=c;return e|0}k[g>>2]=ta;k[o>>2]=na;k[a>>2]=ra;k[s>>2]=pa;k[u>>2]=qa;k[v>>2]=oa;if((k[W>>2]|0)==0?(k[f>>2]|0)>>>0>25|(sa|0)==(na|0):0){ch=pa;fh=na}else Ca=287;do if((Ca|0)==287){if(!(dh(a,sa)|0)){ch=k[s>>2]|0;fh=k[o>>2]|0;break}k[f>>2]=30;e=-4;r=c;return e|0}while(0);o=sa-fh|0;s=a+8|0;k[s>>2]=t-ch+(k[s>>2]|0);k[X>>2]=(k[X>>2]|0)+o;k[Y>>2]=(k[Y>>2]|0)+o;if((sa|0)!=(fh|0)&(k[w>>2]|0)!=0){w=k[x>>2]|0;Y=(k[g>>2]|0)+(0-o)|0;if(!(k[z>>2]|0))gh=zg(w,Y,o)|0;else gh=Ag(w,Y,o)|0;k[x>>2]=gh;k[E>>2]=gh}gh=k[f>>2]|0;k[a+44>>2]=((k[I>>2]|0)!=0?64:0)+(k[v>>2]|0)+((gh|0)==11?128:0)+((gh|0)==19|(gh|0)==14?256:0);e=(ua|0)==0&((b|0)==4|(t|0)==(ch|0)&(sa|0)==(fh|0))?-5:ua;r=c;return e|0}function ch(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;if(!a){b=-2;return b|0}c=a+28|0;d=k[c>>2]|0;if(!d){b=-2;return b|0}e=a+36|0;f=k[e>>2]|0;if(!f){b=-2;return b|0}g=k[d+52>>2]|0;h=a+40|0;if(!g){i=f;j=d}else{Tb[f&1](k[h>>2]|0,g);i=k[e>>2]|0;j=k[c>>2]|0}Tb[i&1](k[h>>2]|0,j);k[c>>2]=0;b=0;return b|0}function dh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;c=k[a+28>>2]|0;d=c+52|0;e=k[d>>2]|0;if(!e){f=Pb[k[a+32>>2]&63](k[a+40>>2]|0,1<<k[c+36>>2],1)|0;k[d>>2]=f;if(!f){g=1;return g|0}else h=f}else h=e;e=c+40|0;f=k[e>>2]|0;if(!f){i=1<<k[c+36>>2];k[e>>2]=i;k[c+48>>2]=0;k[c+44>>2]=0;j=i}else j=f;f=b-(k[a+16>>2]|0)|0;if(f>>>0>=j>>>0){py(h|0,(k[a+12>>2]|0)+(0-j)|0,j|0)|0;k[c+48>>2]=0;k[c+44>>2]=k[e>>2];g=0;return g|0}b=c+48|0;i=k[b>>2]|0;l=j-i|0;j=l>>>0>f>>>0?f:l;l=a+12|0;py(h+i|0,(k[l>>2]|0)+(0-f)|0,j|0)|0;i=f-j|0;if((f|0)!=(j|0)){py(k[d>>2]|0,(k[l>>2]|0)+(0-i)|0,i|0)|0;k[b>>2]=i;k[c+44>>2]=k[e>>2];g=0;return g|0}i=(k[b>>2]|0)+f|0;l=k[e>>2]|0;k[b>>2]=(i|0)==(l|0)?0:i;i=c+44|0;c=k[i>>2]|0;if(c>>>0>=l>>>0){g=0;return g|0}k[i>>2]=c+f;g=0;return g|0}function eh(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,l=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0;g=r;r=r+64|0;h=g+32|0;l=g;n=h;o=n+32|0;do{j[n>>1]=0;n=n+2|0}while((n|0)<(o|0));n=(c|0)==0;if(!n){o=0;do{p=h+(m[b+(o<<1)>>1]<<1)|0;j[p>>1]=(j[p>>1]|0)+1<<16>>16;o=o+1|0}while((o|0)!=(c|0));o=j[h+30>>1]|0;p=k[e>>2]|0;if(!(o<<16>>16)){q=p;s=5}else{t=p;u=o;v=15;s=6}}else{q=k[e>>2]|0;s=5}do if((s|0)==5)if(!(j[h+28>>1]|0))if(!(j[h+26>>1]|0))if(!(j[h+24>>1]|0))if(!(j[h+22>>1]|0))if(!(j[h+20>>1]|0))if(!(j[h+18>>1]|0))if(!(j[h+16>>1]|0))if(!(j[h+14>>1]|0))if(!(j[h+12>>1]|0))if(!(j[h+10>>1]|0))if(!(j[h+8>>1]|0))if(!(j[h+6>>1]|0))if(!(j[h+4>>1]|0)){if(j[h+2>>1]|0){w=0;x=q>>>0>1?1:q;y=1;z=1;break}o=k[d>>2]|0;k[d>>2]=o+4;i[o>>0]=64;i[o+1>>0]=1;j[o+2>>1]=0;o=k[d>>2]|0;k[d>>2]=o+4;i[o>>0]=64;i[o+1>>0]=1;j[o+2>>1]=0;k[e>>2]=1;A=0;r=g;return A|0}else{t=q;u=0;v=2;s=6}else{t=q;u=0;v=3;s=6}else{t=q;u=0;v=4;s=6}else{t=q;u=0;v=5;s=6}else{t=q;u=0;v=6;s=6}else{t=q;u=0;v=7;s=6}else{t=q;u=0;v=8;s=6}else{t=q;u=0;v=9;s=6}else{t=q;u=0;v=10;s=6}else{t=q;u=0;v=11;s=6}else{t=q;u=0;v=12;s=6}else{t=q;u=0;v=13;s=6}else{t=q;u=0;v=14;s=6}while(0);a:do if((s|0)==6){q=t>>>0>v>>>0?v:t;o=1;while(1){if(j[h+(o<<1)>>1]|0){w=u;x=q;y=v;z=o;break a}p=o+1|0;if(p>>>0<v>>>0)o=p;else{w=u;x=q;y=v;z=p;break}}}while(0);v=x>>>0<z>>>0?z:x;x=j[h+2>>1]|0;u=2-(x&65535)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+4>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+6>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+8>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+10>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+12>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+14>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+16>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+18>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+20>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+22>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+24>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(m[h+26>>1]|0)|0;if((u|0)<0){A=-1;r=g;return A|0}t=(u<<1)-(m[h+28>>1]|0)|0;if((t|0)<0){A=-1;r=g;return A|0}u=(t<<1)-(w&65535)|0;if((u|0)<0){A=-1;r=g;return A|0}if((u|0)>0?(a|0)==0|(y|0)!=1:0){A=-1;r=g;return A|0}j[l+2>>1]=0;j[l+4>>1]=x;u=(m[h+4>>1]|0)+(x&65535)|0;j[l+6>>1]=u;x=(m[h+6>>1]|0)+(u&65535)|0;j[l+8>>1]=x;u=(m[h+8>>1]|0)+(x&65535)|0;j[l+10>>1]=u;x=(m[h+10>>1]|0)+(u&65535)|0;j[l+12>>1]=x;u=(m[h+12>>1]|0)+(x&65535)|0;j[l+14>>1]=u;x=(m[h+14>>1]|0)+(u&65535)|0;j[l+16>>1]=x;u=(m[h+16>>1]|0)+(x&65535)|0;j[l+18>>1]=u;x=(m[h+18>>1]|0)+(u&65535)|0;j[l+20>>1]=x;u=(m[h+20>>1]|0)+(x&65535)|0;j[l+22>>1]=u;x=(m[h+22>>1]|0)+(u&65535)|0;j[l+24>>1]=x;u=(m[h+24>>1]|0)+(x&65535)|0;j[l+26>>1]=u;x=(m[h+26>>1]|0)+(u&65535)|0;j[l+28>>1]=x;j[l+30>>1]=(m[h+28>>1]|0)+(x&65535);if(!n){n=0;do{x=j[b+(n<<1)>>1]|0;if(x<<16>>16){u=l+((x&65535)<<1)|0;x=j[u>>1]|0;j[u>>1]=x+1<<16>>16;j[f+((x&65535)<<1)>>1]=n}n=n+1|0}while((n|0)!=(c|0))}switch(a|0){case 0:{B=0;C=0;D=f;E=19;F=f;break}case 1:{if(v>>>0>9){A=1;r=g;return A|0}else{B=0;C=1;D=127740;E=256;F=127866}break}default:{c=(a|0)==2;if(c&v>>>0>9){A=1;r=g;return A|0}else{B=c;C=0;D=128190;E=-1;F=128316}}}c=1<<v;a=c+-1|0;n=v&255;l=v;x=0;u=0;w=z;z=-1;t=k[d>>2]|0;q=0;o=c;b:while(1){c=1<<l;p=u;G=w;H=q;while(1){I=G-x|0;J=I&255;K=j[f+(H<<1)>>1]|0;L=K&65535;do if((L|0)<(E|0)){M=0;N=K}else{if((L|0)<=(E|0)){M=96;N=0;break}M=j[F+(L<<1)>>1]&255;N=j[D+(L<<1)>>1]|0}while(0);L=1<<I;K=p>>>x;O=c;do{P=O;O=O-L|0;Q=O+K|0;i[t+(Q<<2)>>0]=M;i[t+(Q<<2)+1>>0]=J;j[t+(Q<<2)+2>>1]=N}while((P|0)!=(L|0));L=1<<G+-1;while(1)if(!(L&p)){R=L;break}else L=L>>>1;if(!R)S=0;else S=(R+-1&p)+R|0;L=H+1|0;K=h+(G<<1)|0;O=(j[K>>1]|0)+-1<<16>>16;j[K>>1]=O;if(!(O<<16>>16)){if((G|0)==(y|0)){T=J;U=x;V=S;W=z;X=t;Y=o;break b}Z=m[b+(m[f+(L<<1)>>1]<<1)>>1]|0}else Z=G;if(Z>>>0<=v>>>0){p=S;G=Z;H=L;continue}O=S&a;if((O|0)==(z|0)){p=S;G=Z;H=L}else{_=L;$=O;aa=S;ba=Z;break}}H=(x|0)==0?v:x;G=t+(c<<2)|0;p=ba-H|0;c:do if(ba>>>0<y>>>0){O=ba;L=p;K=1<<p;while(1){I=K-(m[h+(O<<1)>>1]|0)|0;if((I|0)<1){ca=L;break c}P=L+1|0;O=P+H|0;if(O>>>0>=y>>>0){ca=P;break}else{L=P;K=I<<1}}}else ca=p;while(0);p=(1<<ca)+o|0;if(C&p>>>0>851|B&p>>>0>591){A=1;s=49;break}c=k[d>>2]|0;i[c+($<<2)>>0]=ca;i[c+($<<2)+1>>0]=n;j[c+($<<2)+2>>1]=(G-c|0)>>>2;l=ca;x=H;u=aa;w=ba;z=$;t=G;q=_;o=p}if((s|0)==49){r=g;return A|0}d:do if(V){s=U;o=T;_=V;q=y;t=X;while(1){if((s|0)==0|(_&a|0)==(W|0)){da=s;ea=o;fa=q;ga=t}else{da=0;ea=n;fa=v;ga=k[d>>2]|0}$=_>>>da;i[ga+($<<2)>>0]=64;i[ga+($<<2)+1>>0]=ea;j[ga+($<<2)+2>>1]=0;$=1<<fa+-1;while(1)if(!($&_)){ha=$;break}else $=$>>>1;if(!ha)break d;_=(ha+-1&_)+ha|0;if(!_)break;else{s=da;o=ea;q=fa;t=ga}}}while(0);k[d>>2]=(k[d>>2]|0)+(Y<<2);k[e>>2]=v;A=0;r=g;return A|0}function fh(a){a=a|0;k[a+2840>>2]=a+148;k[a+2848>>2]=112380;k[a+2852>>2]=a+2440;k[a+2860>>2]=112400;k[a+2864>>2]=a+2684;k[a+2872>>2]=112420;j[a+5816>>1]=0;k[a+5820>>2]=0;k[a+5812>>2]=8;jh(a);return}function gh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,l=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0;e=a+5820|0;f=k[e>>2]|0;g=d&65535;d=a+5816|0;h=m[d>>1]|0|g<<f;j[d>>1]=h;if((f|0)>13){l=a+20|0;n=k[l>>2]|0;k[l>>2]=n+1;o=a+8|0;i[(k[o>>2]|0)+n>>0]=h;n=(m[d>>1]|0)>>>8&255;p=k[l>>2]|0;k[l>>2]=p+1;i[(k[o>>2]|0)+p>>0]=n;n=k[e>>2]|0;p=g>>>(16-n|0);j[d>>1]=p;q=p;r=n+-13|0}else{q=h;r=f+3|0}f=q&255;k[e>>2]=r;do if((r|0)<=8){q=a+20|0;if((r|0)>0){h=k[q>>2]|0;k[q>>2]=h+1;n=a+8|0;i[(k[n>>2]|0)+h>>0]=f;s=q;t=n;break}else{s=q;t=a+8|0;break}}else{q=a+20|0;n=k[q>>2]|0;k[q>>2]=n+1;h=a+8|0;i[(k[h>>2]|0)+n>>0]=f;n=(m[d>>1]|0)>>>8&255;p=k[q>>2]|0;k[q>>2]=p+1;i[(k[h>>2]|0)+p>>0]=n;s=q;t=h}while(0);j[d>>1]=0;k[e>>2]=0;k[a+5812>>2]=8;a=k[s>>2]|0;k[s>>2]=a+1;i[(k[t>>2]|0)+a>>0]=c;a=k[s>>2]|0;k[s>>2]=a+1;i[(k[t>>2]|0)+a>>0]=c>>>8;a=c&65535^65535;e=k[s>>2]|0;k[s>>2]=e+1;i[(k[t>>2]|0)+e>>0]=a;e=k[s>>2]|0;k[s>>2]=e+1;i[(k[t>>2]|0)+e>>0]=a>>>8;if(!c)return;else{u=c;v=b}while(1){u=u+-1|0;b=i[v>>0]|0;c=k[s>>2]|0;k[s>>2]=c+1;i[(k[t>>2]|0)+c>>0]=b;if(!u)break;else v=v+1|0}return}function hh(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,l=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;b=a+5820|0;c=k[b>>2]|0;d=a+5816|0;e=m[d>>1]|0|2<<c;f=e&65535;j[d>>1]=f;if((c|0)>13){g=a+20|0;h=k[g>>2]|0;k[g>>2]=h+1;l=a+8|0;i[(k[l>>2]|0)+h>>0]=e;e=(m[d>>1]|0)>>>8&255;h=k[g>>2]|0;k[g>>2]=h+1;i[(k[l>>2]|0)+h>>0]=e;e=k[b>>2]|0;h=2>>>(16-e|0)&65535;j[d>>1]=h;n=h;o=e+-13|0}else{n=f;o=c+3|0}k[b>>2]=o;if((o|0)>9){c=a+20|0;f=k[c>>2]|0;k[c>>2]=f+1;e=a+8|0;i[(k[e>>2]|0)+f>>0]=n;f=(m[d>>1]|0)>>>8&255;h=k[c>>2]|0;k[c>>2]=h+1;i[(k[e>>2]|0)+h>>0]=f;j[d>>1]=0;p=(k[b>>2]|0)+-9|0;q=0}else{p=o+7|0;q=n}k[b>>2]=p;if((p|0)!=16)if((p|0)>7){n=a+20|0;o=k[n>>2]|0;k[n>>2]=o+1;i[(k[a+8>>2]|0)+o>>0]=q;o=(m[d>>1]|0)>>>8;j[d>>1]=o;n=(k[b>>2]|0)+-8|0;k[b>>2]=n;r=n;s=o}else{r=p;s=q}else{p=a+20|0;o=k[p>>2]|0;k[p>>2]=o+1;n=a+8|0;i[(k[n>>2]|0)+o>>0]=q;q=(m[d>>1]|0)>>>8&255;o=k[p>>2]|0;k[p>>2]=o+1;i[(k[n>>2]|0)+o>>0]=q;j[d>>1]=0;k[b>>2]=0;r=0;s=0}q=a+5812|0;if((11-r+(k[q>>2]|0)|0)>=9){k[q>>2]=7;return}o=s&65535|2<<r;j[d>>1]=o;if((r|0)>13){s=a+20|0;n=k[s>>2]|0;k[s>>2]=n+1;p=a+8|0;i[(k[p>>2]|0)+n>>0]=o;n=(m[d>>1]|0)>>>8&255;f=k[s>>2]|0;k[s>>2]=f+1;i[(k[p>>2]|0)+f>>0]=n;n=k[b>>2]|0;f=2>>>(16-n|0);j[d>>1]=f;t=f;u=n+-13|0}else{t=o;u=r+3|0}r=t&255;k[b>>2]=u;if((u|0)>9){t=a+20|0;o=k[t>>2]|0;k[t>>2]=o+1;n=a+8|0;i[(k[n>>2]|0)+o>>0]=r;o=(m[d>>1]|0)>>>8&255;f=k[t>>2]|0;k[t>>2]=f+1;i[(k[n>>2]|0)+f>>0]=o;j[d>>1]=0;v=0;w=(k[b>>2]|0)+-9|0}else{v=r;w=u+7|0}k[b>>2]=w;if((w|0)==16){u=a+20|0;r=k[u>>2]|0;k[u>>2]=r+1;o=a+8|0;i[(k[o>>2]|0)+r>>0]=v;r=(m[d>>1]|0)>>>8&255;f=k[u>>2]|0;k[u>>2]=f+1;i[(k[o>>2]|0)+f>>0]=r;j[d>>1]=0;k[b>>2]=0;k[q>>2]=7;return}if((w|0)<=7){k[q>>2]=7;return}w=a+20|0;r=k[w>>2]|0;k[w>>2]=r+1;i[(k[a+8>>2]|0)+r>>0]=v;j[d>>1]=(m[d>>1]|0)>>>8;k[b>>2]=(k[b>>2]|0)+-8;k[q>>2]=7;return}function ih(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;if((k[a+132>>2]|0)>0){e=(k[a>>2]|0)+44|0;if((k[e>>2]|0)==2){f=-201342849;g=0;while(1){if((f&1|0)!=0?(j[a+148+(g<<2)>>1]|0)!=0:0){h=0;break}g=g+1|0;if((g|0)>=32){n=6;break}else f=f>>>1}a:do if((n|0)==6)if(((j[a+184>>1]|0)==0?(j[a+188>>1]|0)==0:0)?(j[a+200>>1]|0)==0:0){f=32;while(1){if(j[a+148+(f<<2)>>1]|0){h=1;break a}f=f+1|0;if((f|0)>=256){h=0;break}}}else h=1;while(0);k[e>>2]=h}kh(a,a+2840|0);kh(a,a+2852|0);h=k[a+2844>>2]|0;e=j[a+150>>1]|0;n=e<<16>>16==0;j[a+148+(h+1<<2)+2>>1]=-1;f=a+2752|0;g=a+2756|0;o=a+2748|0;p=n?138:7;q=n?3:4;n=0;r=e&65535;e=-1;b:while(1){s=0;t=n;while(1){if((t|0)>(h|0))break b;u=t+1|0;v=j[a+148+(u<<2)+2>>1]|0;w=v&65535;x=s+1|0;y=(r|0)==(w|0);if((x|0)<(p|0)&y){s=x;t=u}else{z=u;A=v;B=w;C=x;D=y;break}}do if((C|0)>=(q|0))if(!r)if((C|0)<11){j[f>>1]=(j[f>>1]|0)+1<<16>>16;break}else{j[g>>1]=(j[g>>1]|0)+1<<16>>16;break}else{if((r|0)!=(e|0)){t=a+2684+(r<<2)|0;j[t>>1]=(j[t>>1]|0)+1<<16>>16}j[o>>1]=(j[o>>1]|0)+1<<16>>16;break}else{t=a+2684+(r<<2)|0;j[t>>1]=(m[t>>1]|0)+C}while(0);t=A<<16>>16==0;s=r;p=t?138:D?6:7;q=t|D?3:4;n=z;r=B;e=s}e=k[a+2856>>2]|0;B=j[a+2442>>1]|0;r=B<<16>>16==0;j[a+2440+(e+1<<2)+2>>1]=-1;z=r?138:7;n=r?3:4;r=0;D=B&65535;B=-1;c:while(1){q=0;p=r;while(1){if((p|0)>(e|0))break c;A=p+1|0;C=j[a+2440+(A<<2)+2>>1]|0;h=C&65535;s=q+1|0;t=(D|0)==(h|0);if((s|0)<(z|0)&t){q=s;p=A}else{E=A;F=C;G=h;H=s;I=t;break}}do if((H|0)>=(n|0))if(!D)if((H|0)<11){j[f>>1]=(j[f>>1]|0)+1<<16>>16;break}else{j[g>>1]=(j[g>>1]|0)+1<<16>>16;break}else{if((D|0)!=(B|0)){p=a+2684+(D<<2)|0;j[p>>1]=(j[p>>1]|0)+1<<16>>16}j[o>>1]=(j[o>>1]|0)+1<<16>>16;break}else{p=a+2684+(D<<2)|0;j[p>>1]=(m[p>>1]|0)+H}while(0);p=F<<16>>16==0;q=D;z=p?138:I?6:7;n=p|I?3:4;r=E;D=G;B=q}kh(a,a+2864|0);if(!(j[a+2746>>1]|0))if(!(j[a+2690>>1]|0))if(!(j[a+2742>>1]|0))if(!(j[a+2694>>1]|0))if(!(j[a+2738>>1]|0))if(!(j[a+2698>>1]|0))if(!(j[a+2734>>1]|0))if(!(j[a+2702>>1]|0))if(!(j[a+2730>>1]|0))if(!(j[a+2706>>1]|0))if(!(j[a+2726>>1]|0))if(!(j[a+2710>>1]|0))if(!(j[a+2722>>1]|0))if(!(j[a+2714>>1]|0))if(!(j[a+2718>>1]|0))J=(j[a+2686>>1]|0)==0?2:3;else J=4;else J=5;else J=6;else J=7;else J=8;else J=9;else J=10;else J=11;else J=12;else J=13;else J=14;else J=15;else J=16;else J=17;else J=18;B=a+5800|0;G=(J*3|0)+17+(k[B>>2]|0)|0;k[B>>2]=G;B=(G+10|0)>>>3;G=((k[a+5804>>2]|0)+10|0)>>>3;K=J;L=G>>>0>B>>>0?B:G;M=G}else{G=c+5|0;K=0;L=G;M=G}do if((b|0)!=0&(c+4|0)>>>0<=L>>>0)gh(a,b,c,d);else{G=a+5820|0;B=k[G>>2]|0;J=(B|0)>13;if((M|0)==(L|0)?1:(k[a+136>>2]|0)==4){D=d+2&65535;E=a+5816|0;r=m[E>>1]|D<<B;j[E>>1]=r;if(J){I=a+20|0;n=k[I>>2]|0;k[I>>2]=n+1;z=a+8|0;i[(k[z>>2]|0)+n>>0]=r;r=(m[E>>1]|0)>>>8&255;n=k[I>>2]|0;k[I>>2]=n+1;i[(k[z>>2]|0)+n>>0]=r;r=k[G>>2]|0;j[E>>1]=D>>>(16-r|0);N=r+-13|0}else N=B+3|0;k[G>>2]=N;lh(a,128442,129594);break}r=d+4&65535;D=a+5816|0;E=m[D>>1]|r<<B;j[D>>1]=E;if(J){J=a+20|0;n=k[J>>2]|0;k[J>>2]=n+1;z=a+8|0;i[(k[z>>2]|0)+n>>0]=E;n=(m[D>>1]|0)>>>8&255;I=k[J>>2]|0;k[J>>2]=I+1;i[(k[z>>2]|0)+I>>0]=n;n=k[G>>2]|0;I=r>>>(16-n|0);j[D>>1]=I;O=I;P=n+-13|0}else{O=E;P=B+3|0}k[G>>2]=P;B=k[a+2844>>2]|0;E=k[a+2856>>2]|0;n=B+65280&65535;I=O&65535|n<<P;j[D>>1]=I;if((P|0)>11){r=a+20|0;z=k[r>>2]|0;k[r>>2]=z+1;J=a+8|0;i[(k[J>>2]|0)+z>>0]=I;z=(m[D>>1]|0)>>>8&255;F=k[r>>2]|0;k[r>>2]=F+1;i[(k[J>>2]|0)+F>>0]=z;z=k[G>>2]|0;F=n>>>(16-z|0);j[D>>1]=F;Q=z+-11|0;R=F}else{Q=P+5|0;R=I}k[G>>2]=Q;I=E&65535;F=I<<Q|R&65535;j[D>>1]=F;if((Q|0)>11){z=a+20|0;n=k[z>>2]|0;k[z>>2]=n+1;J=a+8|0;i[(k[J>>2]|0)+n>>0]=F;n=(m[D>>1]|0)>>>8&255;r=k[z>>2]|0;k[z>>2]=r+1;i[(k[J>>2]|0)+r>>0]=n;n=k[G>>2]|0;r=I>>>(16-n|0);j[D>>1]=r;S=n+-11|0;T=r}else{S=Q+5|0;T=F}k[G>>2]=S;F=K+65533&65535;r=F<<S|T&65535;j[D>>1]=r;if((S|0)>12){n=a+20|0;I=k[n>>2]|0;k[n>>2]=I+1;J=a+8|0;i[(k[J>>2]|0)+I>>0]=r;I=(m[D>>1]|0)>>>8&255;z=k[n>>2]|0;k[n>>2]=z+1;i[(k[J>>2]|0)+z>>0]=I;I=k[G>>2]|0;z=F>>>(16-I|0);j[D>>1]=z;U=J;V=n;W=z;X=I+-12|0}else{U=a+8|0;V=a+20|0;W=r;X=S+4|0}k[G>>2]=X;r=X;I=W;z=0;while(1){n=m[a+2684+(l[608280+z>>0]<<2)+2>>1]|0;J=n<<r|I&65535;j[D>>1]=J;if((r|0)>13){F=k[V>>2]|0;k[V>>2]=F+1;i[(k[U>>2]|0)+F>>0]=J;F=(m[D>>1]|0)>>>8&255;H=k[V>>2]|0;k[V>>2]=H+1;i[(k[U>>2]|0)+H>>0]=F;F=k[G>>2]|0;H=n>>>(16-F|0);j[D>>1]=H;Y=H;Z=F+-13|0}else{Y=J;Z=r+3|0}k[G>>2]=Z;if((z|0)==(K|0))break;else{r=Z;I=Y;z=z+1|0}}z=a+148|0;mh(a,z,B);I=a+2440|0;mh(a,I,E);lh(a,z,I)}while(0);jh(a);if(!d)return;d=a+5820|0;Y=k[d>>2]|0;if((Y|0)<=8){Z=a+5816|0;if((Y|0)>0){Y=j[Z>>1]&255;K=a+20|0;U=k[K>>2]|0;k[K>>2]=U+1;i[(k[a+8>>2]|0)+U>>0]=Y;_=Z}else _=Z}else{Z=a+5816|0;Y=j[Z>>1]&255;U=a+20|0;K=k[U>>2]|0;k[U>>2]=K+1;V=a+8|0;i[(k[V>>2]|0)+K>>0]=Y;Y=(m[Z>>1]|0)>>>8&255;K=k[U>>2]|0;k[U>>2]=K+1;i[(k[V>>2]|0)+K>>0]=Y;_=Z}j[_>>1]=0;k[d>>2]=0;return}function jh(a){a=a|0;var b=0;b=0;do{j[a+148+(b<<2)>>1]=0;b=b+1|0}while((b|0)!=286);j[a+2440>>1]=0;j[a+2444>>1]=0;j[a+2448>>1]=0;j[a+2452>>1]=0;j[a+2456>>1]=0;j[a+2460>>1]=0;j[a+2464>>1]=0;j[a+2468>>1]=0;j[a+2472>>1]=0;j[a+2476>>1]=0;j[a+2480>>1]=0;j[a+2484>>1]=0;j[a+2488>>1]=0;j[a+2492>>1]=0;j[a+2496>>1]=0;j[a+2500>>1]=0;j[a+2504>>1]=0;j[a+2508>>1]=0;j[a+2512>>1]=0;j[a+2516>>1]=0;j[a+2520>>1]=0;j[a+2524>>1]=0;j[a+2528>>1]=0;j[a+2532>>1]=0;j[a+2536>>1]=0;j[a+2540>>1]=0;j[a+2544>>1]=0;j[a+2548>>1]=0;j[a+2552>>1]=0;j[a+2556>>1]=0;j[a+2684>>1]=0;j[a+2688>>1]=0;j[a+2692>>1]=0;j[a+2696>>1]=0;j[a+2700>>1]=0;j[a+2704>>1]=0;j[a+2708>>1]=0;j[a+2712>>1]=0;j[a+2716>>1]=0;j[a+2720>>1]=0;j[a+2724>>1]=0;j[a+2728>>1]=0;j[a+2732>>1]=0;j[a+2736>>1]=0;j[a+2740>>1]=0;j[a+2744>>1]=0;j[a+2748>>1]=0;j[a+2752>>1]=0;j[a+2756>>1]=0;j[a+1172>>1]=1;k[a+5804>>2]=0;k[a+5800>>2]=0;k[a+5808>>2]=0;k[a+5792>>2]=0;return}function kh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0;c=r;r=r+32|0;d=c;e=k[b>>2]|0;f=b+8|0;g=k[f>>2]|0;h=k[g>>2]|0;n=k[g+12>>2]|0;g=a+5200|0;k[g>>2]=0;o=a+5204|0;k[o>>2]=573;if((n|0)>0){p=-1;q=0;while(1){if(!(j[e+(q<<2)>>1]|0)){j[e+(q<<2)+2>>1]=0;s=p}else{t=(k[g>>2]|0)+1|0;k[g>>2]=t;k[a+2908+(t<<2)>>2]=q;i[a+5208+q>>0]=0;s=q}q=q+1|0;if((q|0)==(n|0)){u=s;break}else p=s}s=k[g>>2]|0;if((s|0)<2){v=s;w=u;x=3}else{y=s;z=u}}else{v=0;w=-1;x=3}if((x|0)==3){x=a+5800|0;u=a+5804|0;if(!h){s=v;p=w;while(1){q=(p|0)<2;t=p+1|0;A=q?t:p;B=q?t:0;t=s+1|0;k[g>>2]=t;k[a+2908+(t<<2)>>2]=B;j[e+(B<<2)>>1]=1;i[a+5208+B>>0]=0;k[x>>2]=(k[x>>2]|0)+-1;B=k[g>>2]|0;if((B|0)<2){s=B;p=A}else{y=B;z=A;break}}}else{p=v;v=w;while(1){w=(v|0)<2;s=v+1|0;A=w?s:v;B=w?s:0;s=p+1|0;k[g>>2]=s;k[a+2908+(s<<2)>>2]=B;j[e+(B<<2)>>1]=1;i[a+5208+B>>0]=0;k[x>>2]=(k[x>>2]|0)+-1;k[u>>2]=(k[u>>2]|0)-(m[h+(B<<2)+2>>1]|0);B=k[g>>2]|0;if((B|0)<2){p=B;v=A}else{y=B;z=A;break}}}}v=b+4|0;k[v>>2]=z;p=y;h=(y|0)/2|0;while(1){y=k[a+2908+(h<<2)>>2]|0;u=a+5208+y|0;x=h<<1;a:do if((x|0)>(p|0))C=h;else{A=e+(y<<2)|0;B=h;s=p;w=x;while(1){do if((w|0)<(s|0)){t=w|1;q=k[a+2908+(t<<2)>>2]|0;D=j[e+(q<<2)>>1]|0;E=k[a+2908+(w<<2)>>2]|0;F=j[e+(E<<2)>>1]|0;if((D&65535)>=(F&65535)){if(D<<16>>16!=F<<16>>16){G=w;break}if((l[a+5208+q>>0]|0)>(l[a+5208+E>>0]|0)){G=w;break}}G=t}else G=w;while(0);t=j[A>>1]|0;E=k[a+2908+(G<<2)>>2]|0;q=j[e+(E<<2)>>1]|0;if((t&65535)<(q&65535)){C=B;break a}if(t<<16>>16==q<<16>>16?(l[u>>0]|0)<=(l[a+5208+E>>0]|0):0){C=B;break a}k[a+2908+(B<<2)>>2]=E;w=G<<1;s=k[g>>2]|0;if((w|0)>(s|0)){C=G;break}else B=G}}while(0);k[a+2908+(C<<2)>>2]=y;if((h|0)<=1)break;p=k[g>>2]|0;h=h+-1|0}h=a+2912|0;p=k[g>>2]|0;C=n;do{n=k[h>>2]|0;G=p+-1|0;k[g>>2]=G;u=k[a+2908+(p<<2)>>2]|0;k[h>>2]=u;x=a+5208+u|0;b:do if((p|0)<3)H=1;else{B=e+(u<<2)|0;s=1;w=G;A=2;while(1){do if((A|0)<(w|0)){E=A|1;q=k[a+2908+(E<<2)>>2]|0;t=j[e+(q<<2)>>1]|0;F=k[a+2908+(A<<2)>>2]|0;D=j[e+(F<<2)>>1]|0;if((t&65535)>=(D&65535)){if(t<<16>>16!=D<<16>>16){I=A;break}if((l[a+5208+q>>0]|0)>(l[a+5208+F>>0]|0)){I=A;break}}I=E}else I=A;while(0);E=j[B>>1]|0;F=k[a+2908+(I<<2)>>2]|0;q=j[e+(F<<2)>>1]|0;if((E&65535)<(q&65535)){H=s;break b}if(E<<16>>16==q<<16>>16?(l[x>>0]|0)<=(l[a+5208+F>>0]|0):0){H=s;break b}k[a+2908+(s<<2)>>2]=F;A=I<<1;w=k[g>>2]|0;if((A|0)>(w|0)){H=I;break}else s=I}}while(0);k[a+2908+(H<<2)>>2]=u;x=k[h>>2]|0;G=(k[o>>2]|0)+-1|0;k[o>>2]=G;k[a+2908+(G<<2)>>2]=n;G=(k[o>>2]|0)+-1|0;k[o>>2]=G;k[a+2908+(G<<2)>>2]=x;G=e+(C<<2)|0;j[G>>1]=(m[e+(x<<2)>>1]|0)+(m[e+(n<<2)>>1]|0);y=i[a+5208+n>>0]|0;s=i[a+5208+x>>0]|0;w=a+5208+C|0;i[w>>0]=(((y&255)<(s&255)?s:y)&255)+1;y=C&65535;j[e+(x<<2)+2>>1]=y;j[e+(n<<2)+2>>1]=y;k[h>>2]=C;y=k[g>>2]|0;c:do if((y|0)<2)J=1;else{x=1;s=y;A=2;while(1){do if((A|0)<(s|0)){B=A|1;F=k[a+2908+(B<<2)>>2]|0;q=j[e+(F<<2)>>1]|0;E=k[a+2908+(A<<2)>>2]|0;D=j[e+(E<<2)>>1]|0;if((q&65535)>=(D&65535)){if(q<<16>>16!=D<<16>>16){K=A;break}if((l[a+5208+F>>0]|0)>(l[a+5208+E>>0]|0)){K=A;break}}K=B}else K=A;while(0);B=j[G>>1]|0;E=k[a+2908+(K<<2)>>2]|0;F=j[e+(E<<2)>>1]|0;if((B&65535)<(F&65535)){J=x;break c}if(B<<16>>16==F<<16>>16?(l[w>>0]|0)<=(l[a+5208+E>>0]|0):0){J=x;break c}k[a+2908+(x<<2)>>2]=E;A=K<<1;s=k[g>>2]|0;if((A|0)>(s|0)){J=K;break}else x=K}}while(0);k[a+2908+(J<<2)>>2]=C;C=C+1|0;p=k[g>>2]|0}while((p|0)>1);p=k[h>>2]|0;h=(k[o>>2]|0)+-1|0;k[o>>2]=h;k[a+2908+(h<<2)>>2]=p;p=k[b>>2]|0;b=k[v>>2]|0;v=k[f>>2]|0;f=k[v>>2]|0;h=k[v+4>>2]|0;g=k[v+8>>2]|0;C=k[v+16>>2]|0;v=a+2876|0;J=v+32|0;do{j[v>>1]=0;v=v+2|0}while((v|0)<(J|0));v=k[o>>2]|0;j[p+(k[a+2908+(v<<2)>>2]<<2)+2>>1]=0;o=v+1|0;d:do if((o|0)<573){v=a+5800|0;J=a+5804|0;if(!f){K=o;H=0;while(1){I=k[a+2908+(K<<2)>>2]|0;w=p+(I<<2)+2|0;G=m[p+(m[w>>1]<<2)+2>>1]|0;y=(G|0)<(C|0);n=y?G+1|0:C;G=(y&1^1)+H|0;j[w>>1]=n;if((I|0)<=(b|0)){w=a+2876+(n<<1)|0;j[w>>1]=(j[w>>1]|0)+1<<16>>16;if((I|0)<(g|0))L=0;else L=k[h+(I-g<<2)>>2]|0;w=ha(m[p+(I<<2)>>1]|0,L+n|0)|0;k[v>>2]=w+(k[v>>2]|0)}K=K+1|0;if((K|0)==573){M=G;break}else H=G}}else{H=o;K=0;while(1){G=k[a+2908+(H<<2)>>2]|0;w=p+(G<<2)+2|0;n=m[p+(m[w>>1]<<2)+2>>1]|0;I=(n|0)<(C|0);y=I?n+1|0:C;n=(I&1^1)+K|0;j[w>>1]=y;if((G|0)<=(b|0)){w=a+2876+(y<<1)|0;j[w>>1]=(j[w>>1]|0)+1<<16>>16;if((G|0)<(g|0))N=0;else N=k[h+(G-g<<2)>>2]|0;w=m[p+(G<<2)>>1]|0;I=ha(w,N+y|0)|0;k[v>>2]=I+(k[v>>2]|0);I=ha((m[f+(G<<2)+2>>1]|0)+N|0,w)|0;k[J>>2]=I+(k[J>>2]|0)}H=H+1|0;if((H|0)==573){M=n;break}else K=n}}if(M){K=a+2876+(C<<1)|0;H=M;while(1){J=C;while(1){n=J+-1|0;I=a+2876+(n<<1)|0;w=j[I>>1]|0;if(!(w<<16>>16))J=n;else{O=I;P=w;Q=J;break}}j[O>>1]=P+-1<<16>>16;J=a+2876+(Q<<1)|0;j[J>>1]=(m[J>>1]|0)+2;J=(j[K>>1]|0)+-1<<16>>16;j[K>>1]=J;if((H|0)>2)H=H+-2|0;else{R=J;break}}if(C){H=R;K=C;J=573;while(1){w=K&65535;if(!(H<<16>>16))S=J;else{I=J;n=H&65535;while(1){G=I;while(1){y=G+-1|0;u=k[a+2908+(y<<2)>>2]|0;if((u|0)>(b|0))G=y;else{T=y;U=u;break}}G=p+(U<<2)+2|0;u=m[G>>1]|0;if((K|0)!=(u|0)){y=ha(m[p+(U<<2)>>1]|0,K-u|0)|0;k[v>>2]=y+(k[v>>2]|0);j[G>>1]=w}n=n+-1|0;if(!n){S=T;break}else I=T}}I=K+-1|0;if(!I)break d;H=j[a+2876+(I<<1)>>1]|0;K=I;J=S}}}}while(0);S=1;T=0;do{T=(m[a+2876+(S+-1<<1)>>1]|0)+(T&65534)<<1;j[d+(S<<1)>>1]=T;S=S+1|0}while((S|0)!=16);if((z|0)<0){r=c;return}else V=0;while(1){S=j[e+(V<<2)+2>>1]|0;T=S&65535;if(S<<16>>16){S=d+(T<<1)|0;a=j[S>>1]|0;j[S>>1]=a+1<<16>>16;S=T;T=a&65535;a=0;while(1){U=a|T&1;if((S|0)>1){S=S+-1|0;T=T>>>1;a=U<<1}else{W=U;break}}j[e+(V<<2)>>1]=W}if((V|0)==(z|0))break;else V=V+1|0}r=c;return}function lh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0;d=a+5792|0;if(!(k[d>>2]|0)){e=a+5820|0;f=a+5816|0;g=f;h=e;n=k[e>>2]|0;o=j[f>>1]|0}else{f=a+5796|0;e=a+5784|0;p=a+5820|0;q=a+5816|0;r=a+20|0;s=a+8|0;t=0;do{u=j[(k[f>>2]|0)+(t<<1)>>1]|0;v=u&65535;w=l[(k[e>>2]|0)+t>>0]|0;t=t+1|0;do if(!(u<<16>>16)){x=m[b+(w<<2)+2>>1]|0;y=k[p>>2]|0;z=m[b+(w<<2)>>1]|0;A=m[q>>1]|0|z<<y;B=A&65535;j[q>>1]=B;if((y|0)>(16-x|0)){C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=A;A=(m[q>>1]|0)>>>8&255;C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=A;A=k[p>>2]|0;C=z>>>(16-A|0)&65535;j[q>>1]=C;z=x+-16+A|0;k[p>>2]=z;D=C;E=z;break}else{z=y+x|0;k[p>>2]=z;D=B;E=z;break}}else{z=l[608024+w>>0]|0;B=(z|256)+1|0;x=m[b+(B<<2)+2>>1]|0;y=k[p>>2]|0;C=m[b+(B<<2)>>1]|0;B=m[q>>1]|0|C<<y;A=B&65535;j[q>>1]=A;if((y|0)>(16-x|0)){F=k[r>>2]|0;k[r>>2]=F+1;i[(k[s>>2]|0)+F>>0]=B;B=(m[q>>1]|0)>>>8&255;F=k[r>>2]|0;k[r>>2]=F+1;i[(k[s>>2]|0)+F>>0]=B;B=k[p>>2]|0;F=C>>>(16-B|0)&65535;j[q>>1]=F;G=F;H=x+-16+B|0}else{G=A;H=y+x|0}k[p>>2]=H;x=k[112440+(z<<2)>>2]|0;do if((z+-8|0)>>>0<20){y=w-(k[112556+(z<<2)>>2]|0)&65535;A=y<<H|G&65535;B=A&65535;j[q>>1]=B;if((H|0)>(16-x|0)){F=k[r>>2]|0;k[r>>2]=F+1;i[(k[s>>2]|0)+F>>0]=A;A=(m[q>>1]|0)>>>8&255;F=k[r>>2]|0;k[r>>2]=F+1;i[(k[s>>2]|0)+F>>0]=A;A=k[p>>2]|0;F=y>>>(16-A|0)&65535;j[q>>1]=F;y=x+-16+A|0;k[p>>2]=y;I=y;J=F;break}else{F=H+x|0;k[p>>2]=F;I=F;J=B;break}}else{I=H;J=G}while(0);x=v+-1|0;z=l[607512+(x>>>0<256?x:(x>>>7)+256|0)>>0]|0;B=m[c+(z<<2)+2>>1]|0;F=m[c+(z<<2)>>1]|0;y=J&65535|F<<I;A=y&65535;j[q>>1]=A;if((I|0)>(16-B|0)){C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=y;y=(m[q>>1]|0)>>>8&255;C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=y;y=k[p>>2]|0;C=F>>>(16-y|0)&65535;j[q>>1]=C;K=B+-16+y|0;L=C}else{K=I+B|0;L=A}k[p>>2]=K;A=k[112672+(z<<2)>>2]|0;if((z+-4|0)>>>0<26){B=x-(k[112792+(z<<2)>>2]|0)&65535;z=B<<K|L&65535;x=z&65535;j[q>>1]=x;if((K|0)>(16-A|0)){C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=z;z=(m[q>>1]|0)>>>8&255;C=k[r>>2]|0;k[r>>2]=C+1;i[(k[s>>2]|0)+C>>0]=z;z=k[p>>2]|0;C=B>>>(16-z|0)&65535;j[q>>1]=C;B=A+-16+z|0;k[p>>2]=B;D=C;E=B;break}else{B=K+A|0;k[p>>2]=B;D=x;E=B;break}}else{D=L;E=K}}while(0)}while(t>>>0<(k[d>>2]|0)>>>0);g=q;h=p;n=E;o=D}D=b+1026|0;E=m[D>>1]|0;p=m[b+1024>>1]|0;b=o&65535|p<<n;j[g>>1]=b;if((n|0)>(16-E|0)){o=a+20|0;q=k[o>>2]|0;k[o>>2]=q+1;d=a+8|0;i[(k[d>>2]|0)+q>>0]=b;b=(m[g>>1]|0)>>>8&255;q=k[o>>2]|0;k[o>>2]=q+1;i[(k[d>>2]|0)+q>>0]=b;b=k[h>>2]|0;j[g>>1]=p>>>(16-b|0);M=E+-16+b|0;k[h>>2]=M;N=j[D>>1]|0;O=N&65535;P=a+5812|0;k[P>>2]=O;return}else{M=n+E|0;k[h>>2]=M;N=j[D>>1]|0;O=N&65535;P=a+5812|0;k[P>>2]=O;return}}function mh(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,l=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0;d=j[b+2>>1]|0;e=d<<16>>16==0;f=a+2754|0;g=a+5820|0;h=a+2752|0;l=a+5816|0;n=a+20|0;o=a+8|0;p=a+2758|0;q=a+2756|0;r=a+2750|0;s=a+2748|0;t=e?138:7;u=e?3:4;e=0;v=d&65535;d=-1;a:while(1){w=0;x=e;while(1){if((x|0)>(c|0))break a;y=x+1|0;z=j[b+(y<<2)+2>>1]|0;A=z&65535;B=w+1|0;C=(v|0)==(A|0);if((B|0)<(t|0)&C){w=B;x=y}else{D=y;E=z;F=A;G=B;H=C;I=w;break}}do if((G|0)>=(u|0)){if(v){if((v|0)==(d|0)){J=j[l>>1]|0;K=k[g>>2]|0;L=G}else{w=m[a+2684+(v<<2)+2>>1]|0;x=k[g>>2]|0;C=m[a+2684+(v<<2)>>1]|0;B=m[l>>1]|0|C<<x;A=B&65535;j[l>>1]=A;if((x|0)>(16-w|0)){z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=B;B=(m[l>>1]|0)>>>8&255;z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=B;B=k[g>>2]|0;z=C>>>(16-B|0)&65535;j[l>>1]=z;M=z;N=w+-16+B|0}else{M=A;N=x+w|0}k[g>>2]=N;J=M;K=N;L=I}w=m[r>>1]|0;x=m[s>>1]|0;A=J&65535|x<<K;j[l>>1]=A;if((K|0)>(16-w|0)){B=k[n>>2]|0;k[n>>2]=B+1;i[(k[o>>2]|0)+B>>0]=A;B=(m[l>>1]|0)>>>8&255;z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=B;B=k[g>>2]|0;z=x>>>(16-B|0);j[l>>1]=z;O=w+-16+B|0;P=z}else{O=K+w|0;P=A}k[g>>2]=O;A=L+65533&65535;w=P&65535|A<<O;j[l>>1]=w;if((O|0)>14){z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=w;w=(m[l>>1]|0)>>>8&255;z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=w;w=k[g>>2]|0;j[l>>1]=A>>>(16-w|0);k[g>>2]=w+-14;break}else{k[g>>2]=O+2;break}}if((G|0)<11){w=m[f>>1]|0;A=k[g>>2]|0;z=m[h>>1]|0;B=m[l>>1]|0|z<<A;j[l>>1]=B;if((A|0)>(16-w|0)){x=k[n>>2]|0;k[n>>2]=x+1;i[(k[o>>2]|0)+x>>0]=B;x=(m[l>>1]|0)>>>8&255;C=k[n>>2]|0;k[n>>2]=C+1;i[(k[o>>2]|0)+C>>0]=x;x=k[g>>2]|0;C=z>>>(16-x|0);j[l>>1]=C;Q=w+-16+x|0;R=C}else{Q=A+w|0;R=B}k[g>>2]=Q;B=I+65534&65535;w=R&65535|B<<Q;j[l>>1]=w;if((Q|0)>13){A=k[n>>2]|0;k[n>>2]=A+1;i[(k[o>>2]|0)+A>>0]=w;w=(m[l>>1]|0)>>>8&255;A=k[n>>2]|0;k[n>>2]=A+1;i[(k[o>>2]|0)+A>>0]=w;w=k[g>>2]|0;j[l>>1]=B>>>(16-w|0);k[g>>2]=w+-13;break}else{k[g>>2]=Q+3;break}}else{w=m[p>>1]|0;B=k[g>>2]|0;A=m[q>>1]|0;C=m[l>>1]|0|A<<B;j[l>>1]=C;if((B|0)>(16-w|0)){x=k[n>>2]|0;k[n>>2]=x+1;i[(k[o>>2]|0)+x>>0]=C;x=(m[l>>1]|0)>>>8&255;z=k[n>>2]|0;k[n>>2]=z+1;i[(k[o>>2]|0)+z>>0]=x;x=k[g>>2]|0;z=A>>>(16-x|0);j[l>>1]=z;S=w+-16+x|0;T=z}else{S=B+w|0;T=C}k[g>>2]=S;C=I+65526&65535;w=T&65535|C<<S;j[l>>1]=w;if((S|0)>9){B=k[n>>2]|0;k[n>>2]=B+1;i[(k[o>>2]|0)+B>>0]=w;w=(m[l>>1]|0)>>>8&255;B=k[n>>2]|0;k[n>>2]=B+1;i[(k[o>>2]|0)+B>>0]=w;w=k[g>>2]|0;j[l>>1]=C>>>(16-w|0);k[g>>2]=w+-9;break}else{k[g>>2]=S+7;break}}}else{w=a+2684+(v<<2)+2|0;C=a+2684+(v<<2)|0;B=k[g>>2]|0;z=j[l>>1]|0;x=G;while(1){A=m[w>>1]|0;y=m[C>>1]|0;U=z&65535|y<<B;V=U&65535;j[l>>1]=V;if((B|0)>(16-A|0)){W=k[n>>2]|0;k[n>>2]=W+1;i[(k[o>>2]|0)+W>>0]=U;U=(m[l>>1]|0)>>>8&255;W=k[n>>2]|0;k[n>>2]=W+1;i[(k[o>>2]|0)+W>>0]=U;U=k[g>>2]|0;W=y>>>(16-U|0)&65535;j[l>>1]=W;X=W;Y=A+-16+U|0}else{X=V;Y=B+A|0}k[g>>2]=Y;x=x+-1|0;if(!x)break;else{B=Y;z=X}}}while(0);z=E<<16>>16==0;B=v;t=z?138:H?6:7;u=z|H?3:4;e=D;v=F;d=B}return}function nh(a,b,c){a=a|0;b=b|0;c=c|0;return cy(ha(c,b)|0)|0}function oh(a,b){a=a|0;b=b|0;dy(b);return}function ph(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;a=r;r=r+64|0;h=a;i=k[g>>2]|0;if((i|0)>0){j=i;r=a;return j|0}i=cy(115200)|0;if(!i){k[g>>2]=113;j=113;r=a;return j|0}k[h+32>>2]=0;k[h+36>>2]=0;k[h+40>>2]=0;l=h+12|0;k[l>>2]=k[c>>2];m=h+16|0;k[m>>2]=k[d>>2];if(ah(h,31,608415,56)|0){k[g>>2]=414;j=414;r=a;return j|0}n=Hw(i,1,115200,b)|0;a:do if(!(yw(b)|0)){o=h+4|0;p=(e|0)==0;q=h+20|0;s=n;b:while(1){if(!s){t=22;break}k[h>>2]=i;k[o>>2]=s;c:do if(p){switch(bh(h,0)|0){case 1:{break c;break}case 0:break;default:{t=19;break b}}if(k[o>>2]|0){t=18;break b}}else while(1){switch(bh(h,0)|0){case 1:{break c;break}case 0:break;default:{t=19;break b}}if(!(k[o>>2]|0))break c;u=Xb[e&31](k[c>>2]|0,(k[d>>2]|0)+28800|0)|0;k[c>>2]=u;if(!u){t=16;break b}k[m>>2]=28800;v=k[d>>2]|0;k[l>>2]=u+v;k[d>>2]=v+28800}while(0);if(xw(b)|0){t=22;break}v=k[q>>2]|0;k[l>>2]=(k[c>>2]|0)+v;k[m>>2]=(k[d>>2]|0)-v;s=Hw(i,1,115200,b)|0;if(yw(b)|0)break a}if((t|0)==16){ch(h)|0;dy(i);k[g>>2]=414;j=414;r=a;return j|0}else if((t|0)==18){ch(h)|0;dy(i);k[g>>2]=414;j=414;r=a;return j|0}else if((t|0)==19){ch(h)|0;dy(i);k[g>>2]=414;j=414;r=a;return j|0}else if((t|0)==22){k[f>>2]=k[q>>2];dy(i);if(!(ch(h)|0)){j=k[g>>2]|0;r=a;return j|0}else{k[g>>2]=414;j=414;r=a;return j|0}}}while(0);ch(h)|0;dy(i);k[g>>2]=414;j=414;r=a;return j|0}function qh(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0;h=r;r=r+64|0;i=h;j=k[g>>2]|0;if((j|0)>0){l=j;r=h;return l|0}k[i+32>>2]=0;k[i+36>>2]=0;k[i+40>>2]=0;if(ah(i,31,608415,56)|0){k[g>>2]=414;l=414;r=h;return l|0}k[i>>2]=a;k[i+4>>2]=b;b=i+12|0;k[b>>2]=k[c>>2];a=i+16|0;k[a>>2]=k[d>>2];a:do if(!e){switch(bh(i,0)|0){case 1:{break a;break}case 0:break;default:{m=11;break a}}ch(i)|0;k[g>>2]=414;l=414;r=h;return l|0}else{while(1){switch(bh(i,0)|0){case 1:{break a;break}case 0:break;default:{m=11;break a}}j=Xb[e&31](k[c>>2]|0,(k[d>>2]|0)+28800|0)|0;k[c>>2]=j;if(!j)break;k[a>>2]=28800;n=k[d>>2]|0;k[b>>2]=j+n;k[d>>2]=n+28800}ch(i)|0;k[g>>2]=414;l=414;r=h;return l|0}while(0);if((m|0)==11){ch(i)|0;k[g>>2]=414;l=414;r=h;return l|0}if(f)k[f>>2]=k[i+20>>2];if(!(ch(i)|0)){l=k[g>>2]|0;r=h;return l|0}else{k[g>>2]=414;l=414;r=h;return l|0}return 0}function rh(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;a=r;r=r+64|0;e=a;f=k[d>>2]|0;if((f|0)>0){g=f;r=a;return g|0}f=cy(115200)|0;if(!f){k[d>>2]=113;g=113;r=a;return g|0}h=cy(115200)|0;if(!h){k[d>>2]=113;g=113;r=a;return g|0}k[e+32>>2]=0;k[e+36>>2]=0;k[e+40>>2]=0;i=e+12|0;k[i>>2]=h;j=e+16|0;k[j>>2]=115200;if(ah(e,31,608415,56)|0){k[d>>2]=414;g=414;r=a;return g|0}l=e+4|0;m=0;a:while(1){n=Hw(f,1,115200,b)|0;if(yw(b)|0){o=10;break}if(!n){p=m;o=20;break}k[e>>2]=f;k[l>>2]=n;n=m;b:while(1){switch(bh(e,0)|0){case 1:{q=n;break b;break}case 0:break;default:{o=18;break a}}if(!(k[l>>2]|0)){q=n;break}if((Pw(h,1,115200,c)|0)!=115200){o=16;break a}k[i>>2]=h;k[j>>2]=115200;n=n+115200|0}if(!(xw(b)|0))m=q;else{p=q;o=20;break}}if((o|0)==10){ch(e)|0;dy(f);dy(h);k[d>>2]=414;g=414;r=a;return g|0}else if((o|0)==16){ch(e)|0;dy(f);dy(h);k[d>>2]=414;g=414;r=a;return g|0}else if((o|0)==18){ch(e)|0;dy(f);dy(h);k[d>>2]=414;g=414;r=a;return g|0}else if((o|0)==20){o=e+20|0;q=k[o>>2]|0;if(q>>>0>p>>>0?(m=Pw(h,1,q-p|0,c)|0,(m|0)!=((k[o>>2]|0)-p|0)):0){ch(e)|0;dy(f);dy(h);k[d>>2]=414;g=414;r=a;return g|0}dy(f);dy(h);if(!(ch(e)|0)){g=k[d>>2]|0;r=a;return g|0}else{k[d>>2]=414;g=414;r=a;return g|0}}return 0}function sh(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0;h=r;r=r+64|0;i=h;j=k[g>>2]|0;if((j|0)>0){l=j;r=h;return l|0}k[i+32>>2]=0;k[i+36>>2]=0;k[i+40>>2]=0;if(Bg(i,1,8,31,8,0,608415,56)|0){k[g>>2]=413;l=413;r=h;return l|0}k[i>>2]=a;k[i+4>>2]=b;b=i+12|0;k[b>>2]=k[c>>2];a=i+16|0;k[a>>2]=k[d>>2];a:do if(!e){switch(Eg(i,4)|0){case 1:{break a;break}case 0:break;default:{m=11;break a}}Cg(i)|0;k[g>>2]=413;l=413;r=h;return l|0}else{while(1){switch(Eg(i,4)|0){case 1:{break a;break}case 0:break;default:{m=11;break a}}j=Xb[e&31](k[c>>2]|0,(k[d>>2]|0)+28800|0)|0;k[c>>2]=j;if(!j)break;k[a>>2]=28800;n=k[d>>2]|0;k[b>>2]=j+n;k[d>>2]=n+28800}Cg(i)|0;k[g>>2]=413;l=413;r=h;return l|0}while(0);if((m|0)==11){Cg(i)|0;k[g>>2]=413;l=413;r=h;return l|0}if(f)k[f>>2]=k[i+20>>2];if(!(Cg(i)|0)){l=k[g>>2]|0;r=h;return l|0}else{k[g>>2]=413;l=413;r=h;return l|0}return 0}function th(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0;f=r;r=r+64|0;g=f;h=k[e>>2]|0;if((h|0)>0){i=h;r=f;return i|0}h=cy(115200)|0;if(!h){k[e>>2]=113;i=113;r=f;return i|0}k[g+32>>2]=0;k[g+36>>2]=0;k[g+40>>2]=0;if(Bg(g,1,8,31,8,0,608415,56)|0){k[e>>2]=413;i=413;r=f;return i|0}k[g>>2]=a;k[g+4>>2]=b;b=g+12|0;k[b>>2]=h;a=g+16|0;j=0;a:while(1){k[a>>2]=115200;switch(Eg(g,4)|0){case 1:{l=j;m=12;break a;break}case 0:break;default:{m=11;break a}}if((Pw(h,1,115200,c)|0)!=115200){m=9;break}k[b>>2]=h;j=j+115200|0}if((m|0)==9){Cg(g)|0;dy(h);k[e>>2]=413;i=413;r=f;return i|0}else if((m|0)==11){Cg(g)|0;dy(h);k[e>>2]=413;i=413;r=f;return i|0}else if((m|0)==12){m=g+20|0;j=k[m>>2]|0;if(j>>>0>l>>>0?(b=Pw(h,1,j-l|0,c)|0,(b|0)!=((k[m>>2]|0)-l|0)):0){Cg(g)|0;dy(h);k[e>>2]=413;i=413;r=f;return i|0}dy(h);if(d)k[d>>2]=k[m>>2];if(!(Cg(g)|0)){i=k[e>>2]|0;r=f;return i|0}else{k[e>>2]=413;i=413;r=f;return i|0}}return 0}function uh(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;h=r;r=r+16|0;j=h;m=k[g>>2]|0;if((m|0)>0){n=m;r=h;return n|0}i[608421]=0;zx(608421,a,127)|0;k[28260]=b;k[28261]=c;k[28262]=d;k[28258]=e;k[28263]=0;k[28264]=0;k[28265]=0;k[28266]=0;e=0;d=b;a:while(1){b=Hw(608549+e|0,1,32768-e|0,d)|0;switch(b|0){case 0:case -1:{o=6;break a;break}default:{}}c=(k[28266]|0)+b|0;k[28266]=c;if(c>>>0>=32768){p=c;o=8;break}e=c;d=k[28260]|0}if((o|0)==6){d=k[28266]|0;if(!d){tk(608421);tk(641381);q=1;s=k[28263]|0;t=k[28266]|0}else{p=d;o=8}}if((o|0)==8){k[28265]=(k[28265]|0)+p;k[28263]=1;q=i[608549]|0;s=1;t=p}i[j>>0]=q;do if(s>>>0<t>>>0){k[28263]=s+1;u=l[608549+s>>0]|0}else{k[28266]=0;q=0;b:while(1){p=Hw(608549+q|0,1,32768-q|0,k[28260]|0)|0;switch(p|0){case 0:case -1:{o=14;break b;break}default:{}}d=(k[28266]|0)+p|0;k[28266]=d;if(d>>>0<32768)q=d;else{v=d;break}}if((o|0)==14){q=k[28266]|0;if(!q){tk(608421);tk(641381);u=1;break}else v=q}k[28265]=(k[28265]|0)+v;k[28263]=1;u=l[608549]|0}while(0);i[j+1>>0]=u;if(lx(j,738194,2)|0){tk(608421);tk(641404);n=-1;r=h;return n|0}k[28257]=4;if(vh(k[28260]|0,k[28267]|0)|0)k[g>>2]=414;if(f)k[f>>2]=k[28264];n=k[g>>2]|0;r=h;return n|0}function vh(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0;k[28267]=b;b=k[28263]|0;do if(b>>>0<(k[28266]|0)>>>0){k[28263]=b+1;c=608549+b|0;d=9}else{k[28266]=0;e=0;a:while(1){f=Hw(608549+e|0,1,32768-e|0,k[28260]|0)|0;switch(f|0){case 0:case -1:{d=6;break a;break}default:{}}g=(k[28266]|0)+f|0;k[28266]=g;if(g>>>0<32768)e=g;else{h=g;break}}if((d|0)==6){e=k[28266]|0;if(!e){tk(608421);tk(641381);k[28259]=0;k[28268]=1;m=2;break}else h=e}k[28265]=(k[28265]|0)+h;k[28263]=1;c=608549;d=9}while(0);if((d|0)==9){h=l[c>>0]|0;k[28268]=h;k[28259]=h&128;if(!(h&96))n=h;else{tk(608421);tk(641465);n=k[28268]|0}h=n&31;k[28268]=h;if(h>>>0>16){tk(608421);tk(641511);o=1;return o|0}else m=1<<h}h=k[28266]|0;n=k[28263]|0;c=(k[28259]|0)!=0;sy(129714,0,256)|0;b=255;while(1){i[641561+b>>0]=b;if((b|0)>0)b=b+-1|0;else break}b=511;e=0;g=c?257:256;c=511;f=9;p=-1;q=0;r=n<<3;n=h;b:while(1){h=1-f|0;s=e;t=g;u=p;v=q;w=r;x=n;c:while(1){y=w>>3;z=(k[28266]|0)-y|0;if((z|0)>0){A=0;do{i[608549+A>>0]=i[608549+(A+y)>>0]|0;A=A+1|0}while((A|0)!=(z|0))}k[28266]=z;if(z>>>0<64){A=Hw(608549+z|0,1,32768,a)|0;if((A|0)==-1){d=20;break b}y=(k[28266]|0)+A|0;k[28266]=y;k[28265]=(k[28265]|0)+A;B=y;C=A}else{B=z;C=x}A=(C|0)!=0;if(A)D=B-((B>>>0)%(f>>>0)|0)<<3;else D=h+(B<<3)|0;d:do if((D|0)>0){y=s;E=t;F=u;G=v;H=0;while(1){I=(E|0)<(m|0);if((E|0)>(c|0)){J=y;K=E;L=F;M=G;N=H;O=C;d=27;break c}else{P=y;Q=F;R=G;S=H}while(1){T=S>>3;U=((l[608549+(T+1)>>0]|0)<<8|(l[608549+T>>0]|0)|(l[608549+(T+2)>>0]|0)<<16)>>>(S&7)&b;T=S+f|0;if((Q|0)!=-1){if((U|0)==256&(k[28259]|0)!=0){V=T;W=P;X=Q;Y=R;Z=C;d=35;break c}if((U|0)<(E|0)){_=U;$=326320}else{if((U|0)>(E|0)){aa=R;d=38;break b}i[326320+-1>>0]=P;_=Q;$=326320+-1|0}ba=i[641561+_>>0]|0;if(_>>>0>255){ca=ba;da=_;ea=$;while(1){fa=ea+-1|0;i[fa>>0]=ca;ga=j[129714+(da<<1)>>1]|0;da=ga&65535;ha=i[641561+da>>0]|0;if((ga&65535)<=255){ia=ha;ja=fa;break}else{ca=ha;ea=fa}}}else{ia=ba;ja=$}ea=ia&255;ca=ja+-1|0;i[ca>>0]=ia;da=326320-ca|0;fa=da+R|0;if((fa|0)>16383){ha=da;ga=R;ka=ca;while(1){la=16384-ga|0;ma=(ha|0)>(la|0)?la:ha;if((ma|0)>0){py(707112+ga|0,ka|0,ma|0)|0;na=ma+ga|0}else na=ga;if((na|0)>16383){wh(na);k[28264]=(k[28264]|0)+na;oa=0}else oa=na;ka=ka+ma|0;ha=326320-ka|0;if((ha|0)<=0){pa=oa;break}else ga=oa}}else{py(707112+R|0,ca|0,da|0)|0;pa=fa}if(I){qa=ia;ra=U;sa=T;ta=ea;ua=Q;va=pa;break}else{wa=ea;xa=pa}}else{if((U|0)>255){d=31;break b}i[707112+R>>0]=U;wa=U;xa=R+1|0}if((D|0)>(T|0)){P=wa;Q=U;R=xa;S=T}else{ya=wa;za=E;Aa=U;Ba=xa;Ca=T;break d}}j[129714+(E<<1)>>1]=ua;i[641561+E>>0]=qa;I=E+1|0;if((D|0)>(sa|0)){y=ta;E=I;F=ra;G=va;H=sa}else{ya=ta;za=I;Aa=ra;Ba=va;Ca=sa;break}}}else{ya=s;za=t;Aa=u;Ba=v;Ca=0}while(0);if(A){s=ya;t=za;u=Aa;v=Ba;w=Ca;x=C}else{Da=Ba;d=54;break b}}if((d|0)==27){d=0;x=f<<3;w=N+-1+x|0;v=f+1|0;u=(1<<v)+-1|0;b=u;e=J;g=K;c=(v|0)==(k[28268]|0)?m:u;f=v;p=L;q=M;r=w-((w|0)%(x|0)|0)|0;n=O;continue}else if((d|0)==35){d=0;sy(129714,0,256)|0;x=f<<3;w=V+-1+x|0;b=511;e=W;g=256;c=511;f=9;p=X;q=Y;r=w-((w|0)%(x|0)|0)|0;n=Z;continue}}if((d|0)==20){tk(608421);tk(641381);o=1;return o|0}else if((d|0)==31){tk(608421);tk(707097);o=1;return o|0}else if((d|0)==38){if((aa|0)>0){wh(aa);k[28264]=(k[28264]|0)+aa}tk(608421);tk(707097);o=1;return o|0}else if((d|0)==54){if((Da|0)<=0){o=0;return o|0}wh(Da);k[28264]=(k[28264]|0)+Da;o=0;return o|0}return 0}function wh(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;b=k[28258]|0;if(!b){if((Pw(707112,1,a,k[28267]|0)|0)==(a|0))return;tk(608421);tk(725544);return}c=k[28264]|0;d=c+a|0;e=k[k[28261]>>2]|0;if(d>>>0>(k[k[28262]>>2]|0)>>>0){f=Xb[b&31](e,d)|0;k[k[28261]>>2]=f;f=k[28264]|0;k[k[28262]>>2]=f+a;d=k[k[28261]>>2]|0;if(!d){tk(608421);tk(725607);return}else{g=d;h=f}}else{g=e;h=c}py(g+h|0,707112,a|0)|0;return}function xh(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=+g;h=h|0;i=i|0;var j=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0;j=r;r=r+128|0;l=j+80|0;m=j+64|0;n=j+32|0;q=j;s=j+116|0;t=j+112|0;u=g;k[t>>2]=0;if(k[25364]|0){k[q>>2]=b;k[q+4>>2]=c;k[q+8>>2]=d;p[q+16>>3]=u;k[q+24>>2]=h;k[q+28>>2]=i;Rw(725653,q)|0}q=yh(a,d,t,b,c,h,i)|0;i=k[t>>2]|0;gx(i,q,4,9);g=+o[i>>2];v=+o[i+(q+-1<<2)>>2];t=q+1|0;h=(t|0)<2?1:(t|0)/2|0;t=h+-1|0;c=i+(t<<2)|0;w=+o[c>>2];if(((q|0)%2|0|0)!=1&(q|0)>(h|0))x=(w+ +o[i+(h<<2)>>2])*.5;else x=w;w=+(q|0);b=~~(w*.5);d=(b|0)<5?5:b;b=~~(w*.01+.5);a=(b|0)<1?1:b;b=zh(i,q,j+120|0,s,2.5,a,5)|0;if((b|0)<(d|0)){o[e>>2]=g;y=v}else{w=+o[s>>2];if(u>0.0){z=w/u;o[s>>2]=z;A=z}else A=w;w=x-+(t|0)*A;o[e>>2]=g>w?g:w;w=x+ +(q-h|0)*A;y=v<w?v:w}o[f>>2]=y;if(!(k[25364]|0)){dy(i);r=j;return}y=+o[c>>2];p[n>>3]=g;p[n+8>>3]=v;p[n+16>>3]=y;p[n+24>>3]=x;Rw(725701,n)|0;k[m>>2]=d;k[m+4>>2]=a;k[m+8>>2]=b;Rw(725749,m)|0;x=+o[e>>2];y=+o[f>>2];p[l>>3]=+o[s>>2];k[l+8>>2]=h;p[l+16>>3]=x;p[l+24>>3]=y;Rw(725794,l)|0;dy(i);r=j;return}
function He(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0.0,P=0,Q=0,R=0,S=0,T=0.0,U=0.0,V=0,W=0,X=0,Z=0.0,_=0.0,$=0.0,aa=0,ba=0.0,ca=0.0;f=r;r=r+16|0;g=f;if(!a){h=0.0;r=f;return +h}i=e|d;if((i|0)<0){Pw(605090,46,1,k[31229]|0)|0;h=0.0;r=f;return +h}if(!i){h=+Ge(a,b,c);r=f;return +h}i=cy(64)|0;j=a+36|0;l=k[j>>2]|0;m=l+-1|0;n=(m|0)>(d|0)?d:m;m=a+40|0;d=k[m>>2]|0;o=d+-1|0;q=(o|0)>(e|0)?e:o;o=k[a+32>>2]|0;s=i+32|0;k[s>>2]=o;if((o+-1|0)>>>0>=3){t=k[31229]|0;k[g>>2]=o;Ew(t,605137,g)|0;h=0.0;r=f;return +h}g=k[a+44>>2]|0;t=i+44|0;k[t>>2]=g;a:do switch(g|0){case 0:{o=(n|0)>0;u=(q|0)>0;if(o&u){v=i+36|0;k[v>>2]=1;w=i+40|0;k[w>>2]=1;x=i+48|0;k[x>>2]=1;y=x;z=v;A=w;B=1;C=1;D=1;break a}if(o){o=l-n|0;w=(o|0)<1?1:o;o=i+36|0;k[o>>2]=w;v=i+40|0;k[v>>2]=1;x=i+48|0;k[x>>2]=w;y=x;z=o;A=v;B=w;C=w;D=1;break a}if(u){u=i+36|0;k[u>>2]=1;w=d-q|0;v=(w|0)<1?1:w;w=i+40|0;k[w>>2]=v;o=i+48|0;k[o>>2]=v;y=o;z=u;A=w;B=v;C=1;D=v;break a}else{y=i+48|0;z=i+36|0;A=i+40|0;B=0;C=0;D=0;break a}break}case 2:{v=((l|0)>(d|0)?l:d)-q-n|0;w=l-n|0;u=(v|0)<(w|0)?v:w;w=(u|0)<1?1:u;u=i+36|0;k[u>>2]=w;o=d-q|0;x=(v|0)<(o|0)?v:o;o=(x|0)<1?1:x;x=i+40|0;k[x>>2]=o;v=(w|0)<(o|0)?w:o;E=(ha(o,w)|0)-((ha(v+-1|0,v)|0)/2|0)|0;v=i+48|0;k[v>>2]=E;y=v;z=u;A=x;B=E;C=w;D=o;break}default:{o=l-n|0;w=(o|0)<1?1:o;o=i+36|0;k[o>>2]=w;E=d-q|0;x=(E|0)<1?1:E;E=i+40|0;k[E>>2]=x;u=ha(x,w)|0;v=i+48|0;k[v>>2]=u;y=v;z=o;A=E;B=u;C=w;D=x}}while(0);p[i>>3]=+p[a>>3];p[i+8>>3]=+p[a+8>>3];l=i+16|0;p[l>>3]=+p[a+16>>3];p[i+24>>3]=+p[a+24>>3];g=cy(B<<3)|0;B=i+52|0;k[B>>2]=g;g=cy(C<<3)|0;C=i+56|0;k[C>>2]=g;g=cy(D<<3)|0;D=i+60|0;k[D>>2]=g;g=a+48|0;x=k[g>>2]|0;w=x<<3;u=k[26014]|0;if((w|0)>(u|0)){if((u|0)>0)F=fy(k[26015]|0,w)|0;else F=cy(w)|0;k[26015]=F;k[26014]=w;G=F;H=k[g>>2]|0}else{G=k[26015]|0;H=x}if((H|0)>0){x=k[a+52>>2]|0;a=0;do{p[G+(a<<3)>>3]=+p[x+(a<<3)>>3];a=a+1|0}while((a|0)!=(H|0))}b:do switch(k[t>>2]|0){case 1:{a=k[m>>2]|0;x=a+-1|0;if((a|0)>(q|0)){a=k[j>>2]|0;g=G+((ha(x,a)|0)<<3)|0;F=k[z>>2]|0;w=1-q|0;u=n+1|0;E=(a|0)<(u|0);o=(F|0)>0;v=0-a|0;I=1-n|0;J=0-F|0;K=(F|0)>0;L=x;x=g;g=(k[B>>2]|0)+((ha((k[A>>2]|0)+-1|0,F)|0)<<3)|0;while(1){M=w+L|0;if((L|0)>=(M|0)){N=L;while(1){if(K){O=+(N|0);P=0;do{Q=x+(P+n<<3)|0;p[Q>>3]=O*+p[Q>>3];P=P+1|0}while((P|0)<(F|0))}if((N|0)>(M|0))N=N+-1|0;else break}}if(!E){N=a;do{M=I+N|0;P=N;N=N+-1|0;if((P|0)>=(M|0)){Q=x+(N<<3)|0;O=+p[Q>>3];R=P;while(1){S=R;R=R+-1|0;T=+(R|0)*O;if((S|0)<=(M|0)){U=T;break}else O=T}p[Q>>3]=U}}while((P|0)>(u|0))}if(o){N=0;do{p[g+(N<<3)>>3]=+p[x+(N+n<<3)>>3];N=N+1|0}while((N|0)<(F|0))}if((L|0)>(q|0)){L=L+-1|0;x=x+(v<<3)|0;g=g+(J<<3)|0}else break}}break}case 2:{J=k[j>>2]|0;g=k[m>>2]|0;v=((J|0)>(g|0)?J:g)+1|0;x=k[z>>2]|0;L=k[A>>2]|0;F=q+1|0;if((g|0)>=(F|0)){o=F+((x|0)>(L|0)?x:L)|0;L=1-q|0;u=n+1|0;I=1-n|0;a=g;g=G+(H<<3)|0;E=(k[B>>2]|0)+(k[y>>2]<<3)|0;while(1){K=v-a|0;w=(K|0)<(J|0)?K:J;K=(w|0)<0?0:w;w=o-a|0;N=(w|0)<(x|0)?w:x;w=(N|0)<0?0:N;N=g;g=g+(0-K<<3)|0;M=L+a|0;if((a|0)>(M|0)){R=(w|0)>0;S=n-K|0;V=a;do{V=V+-1|0;if(R){O=+(V|0);W=0;do{X=N+(S+W<<3)|0;p[X>>3]=O*+p[X>>3];W=W+1|0}while((W|0)<(w|0))}}while((V|0)>(M|0))}if((K|0)>=(u|0)){M=K;do{V=I+M|0;S=M;M=M+-1|0;if((S|0)>=(V|0)){R=N+(M-K<<3)|0;O=+p[R>>3];W=S;while(1){P=W;W=W+-1|0;T=+(W|0)*O;if((P|0)<=(V|0)){Z=T;break}else O=T}p[R>>3]=Z}}while((S|0)>(u|0))}if((w|0)>0){M=n-K|0;V=0;do{p[E+(V-w<<3)>>3]=+p[N+(M+V<<3)>>3];V=V+1|0}while((V|0)<(w|0))}E=E+(0-w<<3)|0;if((a|0)<=(F|0))break;else a=a+-1|0}}break}default:{a=(n|0)>0;F=(q|0)>0;if(a&F){p[k[B>>2]>>3]=0.0;break b}if(a){a=k[j>>2]|0;E=n+1|0;if((a|0)<(E|0))break b;u=1-n|0;I=a;a=(k[B>>2]|0)+((k[y>>2]|0)+-1<<3)|0;while(1){L=u+I|0;g=I;I=I+-1|0;x=G+(I<<3)|0;O=+p[x>>3];if((g|0)<(L|0))_=O;else{T=O;o=g;while(1){J=o;o=o+-1|0;O=T*+(o|0);if((J|0)<=(L|0)){$=O;break}else T=O}p[x>>3]=$;_=$}p[a>>3]=_;if((g|0)<=(E|0))break b;else a=a+-8|0}}if(F){a=H+-1|0;E=k[B>>2]|0;I=k[m>>2]|0;u=q+1|0;if((I|0)<(u|0))aa=a;else{L=1-q|0;o=d+-1|0;w=((o|0)<(e|0)?o:e)+1|0;o=H+-3-(I+~((I|0)>(w|0)?w:I))|0;w=I;I=G+(a<<3)|0;while(1){a=L+w|0;if((w|0)>=(a|0)){T=+p[I>>3];J=w;while(1){v=J;J=J+-1|0;O=+(J|0)*T;if((v|0)<=(a|0)){ba=O;break}else T=O}p[I>>3]=ba}if((w|0)>(u|0)){w=w+-1|0;I=I+-8|0}else{aa=o;break}}}o=k[y>>2]|0;if((o|0)>0){I=0;do{w=I;I=I+1|0;p[E+(w<<3)>>3]=+p[G+(aa+I<<3)>>3]}while((I|0)<(o|0))}}}}while(0);ba=+Ge(i,b,c);if((k[s>>2]|0)==3)ca=ba;else{c=+Y(+(+p[i>>3]),+(+(n|0)));ca=ba*(c*+Y(+(+p[l>>3]),+(+(q|0))))}q=k[C>>2]|0;if(q)dy(q);q=k[D>>2]|0;if(q)dy(q);q=k[B>>2]|0;if(q)dy(q);dy(i);h=ca;r=f;return +h}function Ie(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0.0,m=0.0,n=0.0,o=0.0,q=0;b=r;r=r+64|0;c=b+48|0;d=b+32|0;e=b+16|0;f=b+8|0;g=b;h=~~(+p[a+8>>3]+.5);if((h|0)<1){i=k[31229]|0;k[g>>2]=h;Ew(i,605173,g)|0;j=0;r=b;return j|0}g=~~(+p[a+16>>3]+.5);if((g|0)<1){i=k[31229]|0;k[f>>2]=g;Ew(i,605207,f)|0;j=0;r=b;return j|0}l=+p[a+32>>3];m=+p[a+40>>3];if(m<=l){f=k[31229]|0;p[e>>3]=l;p[e+8>>3]=m;Ew(f,605241,e)|0;j=0;r=b;return j|0}n=+p[a+48>>3];o=+p[a+56>>3];if(o<=n){e=k[31229]|0;p[d>>3]=n;p[d+8>>3]=o;Ew(e,605278,d)|0;j=0;r=b;return j|0}d=~~(+p[a>>3]+.5);if((d+-1|0)>>>0>=3){e=k[31229]|0;k[c>>2]=d;Ew(e,605315,c)|0;j=0;r=b;return j|0}c=cy(64)|0;k[c+36>>2]=h;p[c>>3]=2.0/(m-l);p[c+8>>3]=(l+m)*-.5;k[c+40>>2]=g;p[c+16>>3]=2.0/(o-n);p[c+24>>3]=(n+o)*-.5;e=~~+p[a+24>>3];k[c+44>>2]=e;switch(e|0){case 0:{e=h+-1+g|0;k[c+48>>2]=e;q=e;break}case 2:{e=(h|0)<(g|0)?h:g;f=(ha(g,h)|0)-((ha(e+-1|0,e)|0)/2|0)|0;k[c+48>>2]=f;q=f;break}case 1:{f=ha(g,h)|0;k[c+48>>2]=f;q=f;break}default:q=0}k[c+32>>2]=d;d=cy(q<<3)|0;k[c+52>>2]=d;if((q|0)>0){f=0;do{p[d+(f<<3)>>3]=+p[a+(f+8<<3)>>3];f=f+1|0}while((f|0)<(q|0))}q=cy(h<<3)|0;k[c+56>>2]=q;q=cy(g<<3)|0;k[c+60>>2]=q;j=c;r=b;return j|0}function Je(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0;c=r;r=r+96|0;d=c+72|0;e=c+64|0;f=c+56|0;g=c+48|0;h=c+40|0;i=c+32|0;j=c+24|0;l=c+16|0;m=c+8|0;n=c;o=c+80|0;q=cy(2e3)|0;s=cy(2e3)|0;if(!(ke(a,605354,2e3,q)|0)){t=cy((xx(a)|0)+200|0)|0;py(t|0,605359,161)|0;qx(t,a)|0;ke(t,605354,2e3,q)|0;ke(t,605520,2e3,s)|0;dy(t)}ke(a,605520,2e3,s)|0;a=cy(2e3)|0;t=cy(2e3)|0;u=b+3176|0;if((+p[u>>3]>360.0?(ue(q,605525,u)|0)==0:0)?(ue(s,605525,u)|0)==0:0)p[u>>3]=180.0;u=b+3192|0;if((ue(q,605534,u)|0)==0?(ue(s,605534,u)|0)==0:0)p[u>>3]=57.29577951308232;k[n>>2]=0;Xw(o,605537,n)|0;n=b+4096|0;if(!(ue(q,o,n)|0))p[n>>3]=0.0;k[m>>2]=1;Xw(o,605537,m)|0;m=b+4104|0;if(!(ue(q,o,m)|0))p[m>>3]=0.0;k[l>>2]=2;Xw(o,605537,l)|0;l=b+4112|0;if(!(ue(q,o,l)|0))p[l>>3]=0.0;k[j>>2]=3;Xw(o,605537,j)|0;j=b+4120|0;if(!(ue(q,o,j)|0))p[j>>3]=0.0;k[i>>2]=4;Xw(o,605537,i)|0;i=b+4128|0;if(!(ue(q,o,i)|0))p[i>>3]=0.0;k[h>>2]=5;Xw(o,605537,h)|0;h=b+4136|0;if(!(ue(q,o,h)|0))p[h>>3]=0.0;k[g>>2]=6;Xw(o,605537,g)|0;g=b+4144|0;if(!(ue(q,o,g)|0))p[g>>3]=0.0;k[f>>2]=7;Xw(o,605537,f)|0;f=b+4152|0;if(!(ue(q,o,f)|0))p[f>>3]=0.0;k[e>>2]=8;Xw(o,605537,e)|0;e=b+4160|0;if(!(ue(q,o,e)|0))p[e>>3]=0.0;k[d>>2]=9;Xw(o,605537,d)|0;d=b+4168|0;if(!(ue(q,o,d)|0))p[d>>3]=0.0;do if(!(we(q,605545,2e3,a)|0))if(!(we(s,605545,2e3,a)|0)){k[b+5960>>2]=0;break}else{o=Fe(a)|0;k[b+5960>>2]=o;break}else{o=Fe(a)|0;k[b+5960>>2]=o}while(0);do if(!(we(s,605552,2e3,t)|0))if(!(we(q,605552,2e3,t)|0)){k[b+5964>>2]=0;break}else{o=Fe(t)|0;k[b+5964>>2]=o;break}else{o=Fe(t)|0;k[b+5964>>2]=o}while(0);a:do if(+p[d>>3]==0.0)if(+p[e>>3]==0.0)if(+p[f>>3]==0.0)if(+p[g>>3]==0.0){if(!(+p[h>>3]==0.0)){v=5;w=25;break}if(!(+p[i>>3]==0.0)){v=4;w=25;break}if(!(+p[j>>3]==0.0)){v=3;w=25;break}do if(+p[l>>3]==0.0){if(!(+p[m>>3]==0.0)){x=1;break}if(!(+p[n>>3]==0.0)){x=0;break}k[b+3276>>2]=-1;break a}else x=2;while(0);k[b+3276>>2]=x}else{v=6;w=25}else{v=7;w=25}else{v=8;w=25}else{v=9;w=25}while(0);if((w|0)==25){k[b+3276>>2]=v;y=+p[m>>3];m=1;z=0.0;while(1){A=+(m|0)*3.141592653589793/180.0;B=0.0;w=v;while(1){C=A*B+ +(w|0)*+p[b+4096+(w<<3)>>3];if((w|0)>1){B=C;w=w+-1|0}else{D=C;break}}m=m+1|0;if(D<=0.0){E=D;F=A;G=y;H=z;break}if((m|0)>=181){E=D;F=A;G=D;H=A;break}else{y=D;z=A}}b:do if(!(E<=0.0))I=3.141592653589793;else{z=G;D=E;m=1;y=H;B=F;while(1){C=y-z*(B-y)/(D-z);J=0.0;w=v;while(1){K=C*J+ +(w|0)*+p[b+4096+(w<<3)>>3];if((w|0)>1){J=K;w=w+-1|0}else{L=K;break}}if(+W(+L)<1.0e-13){I=C;break b}w=L<0.0;m=m+1|0;if((m|0)>=11){I=C;break}else{z=w?z:L;D=w?L:D;y=w?y:C;B=w?C:B}}}while(0);m=v;L=0.0;while(1){F=I*L+ +p[b+4096+(m<<3)>>3];if((m|0)>0){m=m+-1|0;L=F}else{M=F;break}}p[b+3240>>3]=I;p[b+3248>>3]=M}zd(b);dy(q);dy(s);dy(a);dy(t);if((k[b+5964>>2]|0)==0?(k[b+5960>>2]|0)==0:0){N=1;r=c;return N|0}N=0;r=c;return N|0}function Ke(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0,o=0,q=0,r=0,s=0.0,t=0.0,u=0.0,v=0.0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0;f=a-+p[c+616>>3];a=b-+p[c+624>>3];do if(!(k[c+3300>>2]|0)){b=+p[c+760>>3];if(!(b==0.0)?(g=+p[c+768>>3],!(g==0.0)):0){h=f*b;b=a*g;g=+p[c+48>>3];if(!(g!=0.0)){i=h;j=b;break}l=g*3.141592653589793/180.0;g=+Z(+l);m=+_(+l);i=h*g-b*m;j=b*g+h*m;break}p[d>>3]=0.0;p[e>>3]=0.0;n=2;return n|0}else{i=f*+p[c+56>>3]+a*+p[c+64>>3];j=f*+p[c+72>>3]+a*+p[c+80>>3]}while(0);o=(k[c+3304>>2]|0)==0&1;q=o^1;a=(90.0-+p[c+688+(o<<3)>>3])*3.141592653589793/180.0;f=+Z(+a);m=+_(+a);h=+p[c+3176>>3]*3.141592653589793/180.0;o=k[c+3276>>2]|0;r=k[c+5960>>2]|0;if(!r)s=i;else s=i+ +Ge(r,i,j);r=k[c+5964>>2]|0;if(!r)t=j;else t=j+ +Ge(r,i,j);j=+X(+(s*s+t*t));i=j/+p[c+3192>>3];if((o|0)<1){p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}a:do switch(o|0){case 1:{u=(i-+p[c+4096>>3])/+p[c+4104>>3];break}case 2:{j=+p[c+4112>>3];g=+p[c+4104>>3];b=g*g-j*4.0*(+p[c+4096>>3]-i);if(b<0.0){p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}l=+X(+b);b=j*2.0;j=(l-g)/b;v=(-g-l)/b;b=j<v?j:v;l=b<-1.0e-13?(j>v?j:v):b;if(l<0.0){if(!(l<-1.0e-13)){u=0.0;break a}p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}if(l>3.141592653589793)if(l>3.141592653589893){p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}else u=3.141592653589793;else u=l;break}default:{l=+p[c+4096>>3];b=+p[c+3240>>3];v=+p[c+3248>>3];if(i<l){if(!(i<l+-1.0e-13)){u=0.0;break a}p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}if(i>v){if(!(i>v+1.0e-13)){u=b;break a}p[d>>3]=0.0;p[e>>3]=0.0;n=1;return n|0}else{w=0;x=l;y=v;z=0.0;A=b}while(1){b=(y-i)/(y-x);if(!(b<.1))if(b>.9)B=.9;else B=b;else B=.1;b=A-(A-z)*B;r=o;v=0.0;while(1){l=b*v+ +p[c+4096+(r<<3)>>3];if((r|0)>0){r=r+-1|0;v=l}else{C=l;break}}if(C<i)if(i-C<1.0e-13){u=b;break a}else{D=C;E=y;F=b;G=A}else if(C-i<1.0e-13){u=b;break a}else{D=x;E=C;F=z;G=b}w=w+1|0;if(!(!(+W(+(G-F))<1.0e-13)&(w|0)<100)){u=b;break}else{x=D;y=E;z=F;A=G}}}}while(0);if(i==0.0)H=0.0;else H=+da(+s,+-t);t=1.5707963267948966-u;u=+Z(+t);s=+_(+t);i=H-h;h=+Z(+i);H=+_(+i);G=f*u;A=m*s-G*h;if(+W(+A)<1.0e-05)I=G*(1.0-h)-+Z(+(a+t));else I=A;A=u*H;if(A!=-0.0|I!=0.0)J=+da(+-A,+I);else J=i+3.141592653589793;H=+p[c+688+(q<<3)>>3];G=H+J*180.0/3.141592653589793;if(!(H>=0.0))if(G>0.0)K=G+-360.0;else K=G;else if(G<0.0)K=G+360.0;else K=G;if(!(K>360.0))if(K<-360.0)L=K+360.0;else L=K;else L=K+-360.0;do if(+Gv(i,3.141592653589793)==0.0){K=(t+a*h)*180.0/3.141592653589793;G=K>90.0?180.0-K:K;if(G<-90.0)M=-180.0-G;else M=G}else{G=f*s+m*u*h;if(!(+W(+G)>.99)){M=+ba(+G)*180.0/3.141592653589793;break}K=+aa(+(+X(+(A*A+I*I))));if(!(G>=0.0)){M=K*-180.0/3.141592653589793;break}else{M=K*180.0/3.141592653589793;break}}while(0);p[d>>3]=L;p[e>>3]=M;n=0;return n|0}function Le(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0,z=0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0,M=0.0,N=0,O=0.0,P=0.0,Q=0.0,R=0,S=0.0,T=0.0,U=0.0;f=c+3304|0;g=(k[f>>2]|0)==0;h=g&1;i=(a-+p[c+688+((h^1)<<3)>>3])*3.141592653589793/180.0;a=b*3.141592653589793/180.0;b=+Z(+i);j=+_(+i);l=+Z(+a);m=+_(+a);n=(90.0-+p[c+688+(h<<3)>>3])*3.141592653589793/180.0;o=+Z(+n);q=+_(+n);r=+p[c+3176>>3];s=l*o;t=m*q-b*s;if(+W(+t)<1.0e-05)u=(1.0-b)*s-+Z(+(a+n));else u=t;t=j*l;if(t!=-0.0|u!=0.0)v=+da(+-t,+u);else v=i+-3.141592653589793;j=(r==999.0?3.141592653589793:r*3.141592653589793/180.0)+v;if(!(j>3.141592653589793))if(j<-3.141592653589793)w=j+6.283185307179586;else w=j;else w=j+-6.283185307179586;do if(+Gv(i,3.141592653589793)==0.0){j=a+b*n;v=j>1.5707963267948966?3.141592653589793-j:j;if(v<-1.5707963267948966)x=-3.141592653589793-v;else x=v}else{v=m*o+b*(l*q);if(!(+W(+v)>.99)){x=+ba(+v);break}j=+aa(+(+X(+(t*t+u*u))));if(!(v>=0.0))x=-j;else x=j}while(0);u=1.5707963267948966-x;x=(u*(u*(u*(u*(u*(u*(u*(u*(u*(u*0.0+ +p[c+4168>>3])+ +p[c+4160>>3])+ +p[c+4152>>3])+ +p[c+4144>>3])+ +p[c+4136>>3])+ +p[c+4128>>3])+ +p[c+4120>>3])+ +p[c+4112>>3])+ +p[c+4104>>3])+ +p[c+4096>>3])*+p[c+3192>>3];h=c+5960|0;y=k[h>>2]|0;z=c+5964|0;if((y|0)==0?(k[z>>2]|0)==0:0){u=x*+_(+w);t=-(x*+Z(+w));A=g?u:t;B=g?t:u}else{u=x*+_(+w);t=x*+Z(+w);g=y;y=0;w=u;x=-t;while(1){if(!g){C=w;D=1.0;E=0.0}else{q=w+ +Ge(g,w,x);l=+He(k[h>>2]|0,w,x,1,0)+1.0;C=q;D=l;E=+He(k[h>>2]|0,w,x,0,1)}l=C-u;F=k[z>>2]|0;if(!F){G=x;H=0.0;I=1.0}else{q=x+ +Ge(F,w,x);b=+He(k[z>>2]|0,w,x,1,0);G=q;H=b;I=+He(k[z>>2]|0,w,x,0,1)+1.0}b=t+G;q=D*I-E*H;if(q==0.0){J=w;K=x;break}o=(E*b-l*I)/q;m=(l*H-D*b)/q;q=w+o;n=x+m;F=+W(+o)>+W(+m);a=F?o:m;m=+W(+a);F=+W(+l)>+W(+b);o=F?l:b;F=m>+W(+o);L=y+1|0;if(!(!(+W(+(F?a:o))<2.8e-08)&(L|0)<500)){J=q;K=n;break}g=k[h>>2]|0;y=L;w=q;x=n}y=(k[f>>2]|0)==0;A=y?J:K;B=y?K:J}if(k[c+3300>>2]|0){p[d>>3]=A*+p[c+88>>3]+B*+p[c+96>>3];p[e>>3]=A*+p[c+104>>3]+B*+p[c+112>>3];M=+p[d>>3];N=c+16|0;O=+p[N>>3];P=M+O;p[d>>3]=P;Q=+p[e>>3];R=c+24|0;S=+p[R>>3];T=Q+S;p[e>>3]=T;return 0}J=+p[c+48>>3];if(J!=0.0){K=J*3.141592653589793/180.0;J=+Z(+K);x=+_(+K);p[d>>3]=A*J+B*x;U=B*J-A*x}else{p[d>>3]=A;U=B}p[e>>3]=U;U=+p[c+32>>3];if(U!=0.0)p[d>>3]=+p[d>>3]/U;U=+p[c+40>>3];if(!(U!=0.0)){M=+p[d>>3];N=c+16|0;O=+p[N>>3];P=M+O;p[d>>3]=P;Q=+p[e>>3];R=c+24|0;S=+p[R>>3];T=Q+S;p[e>>3]=T;return 0}p[e>>3]=+p[e>>3]/U;M=+p[d>>3];N=c+16|0;O=+p[N>>3];P=M+O;p[d>>3]=P;Q=+p[e>>3];R=c+24|0;S=+p[R>>3];T=Q+S;p[e>>3]=T;return 0}function Me(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0.0,i=0.0,j=0.0,k=0.0,l=0.0,m=0.0,n=0.0,o=0.0;f=(+p[c+224>>3]-(+p[c+176>>3]+a+-1.0+.5)*+p[c+192>>3])/1.0e3;a=((+p[c+184>>3]+b+-1.0+.5)*+p[c+200>>3]-+p[c+248>>3])/1.0e3;b=f*f;g=a*a;h=f*b;i=a*g;j=b+g;k=(h*+p[c+496>>3]+(i*+p[c+472>>3]+(j*+p[c+464>>3]+(b*+p[c+456>>3]+(+p[c+432>>3]+(a*+p[c+416>>3]+f*+p[c+424>>3])+g*+p[c+440>>3]+a*(f*+p[c+448>>3]))))+f*(g*+p[c+480>>3])+b*(a*+p[c+488>>3]))+j*(a*+p[c+504>>3])+j*(j*(a*+p[c+512>>3])))/206264.8062470964;l=+p[c+160>>3];m=+$(+l);n=1.0-k*m;o=+da(+((i*+p[c+336>>3]+(h*+p[c+312>>3]+(j*+p[c+304>>3]+(g*+p[c+296>>3]+(+p[c+272>>3]+(f*+p[c+256>>3]+a*+p[c+264>>3])+b*+p[c+280>>3]+a*(f*+p[c+288>>3]))))+a*(b*+p[c+320>>3])+g*(f*+p[c+328>>3]))+j*(f*+p[c+344>>3])+j*(j*(f*+p[c+352>>3])))/206264.8062470964/+Z(+l)),+n);l=o+ +p[c+152>>3];p[d>>3]=(l<0.0?l+6.28318530717959:l)/.01745329252;l=+ca(+(+Z(+o)*((k+m)/n)))/.01745329252;p[e>>3]=l;return 0}function Ne(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0,i=0.0,j=0.0,k=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0,V=0.0,X=0.0,Y=0.0,$=0.0,aa=0.0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ha=0.0,ia=0.0,ja=0.0,ka=0.0,la=0.0,ma=0.0,na=0.0,oa=0.0,pa=0.0,qa=0.0,ra=0.0,sa=0.0,ta=0.0,ua=0.0;p[d>>3]=0.0;p[e>>3]=0.0;f=b*3.141592653589793/180.0;b=+_(+f);g=+Z(+f);h=c+160|0;f=+p[h>>3];if(f==0.0){i=+p[c+8>>3]*3.141592653589793/180.0;p[h>>3]=i;j=i}else j=f;f=+_(+j);i=+Z(+j);h=c+152|0;j=+p[h>>3];if(j==0.0){k=+p[c+8>>3]*3.141592653589793/180.0;p[h>>3]=k;l=k}else l=j;j=a*3.141592653589793/180.0-l;l=+Z(+j);a=b*f+g*i*l;if(a==0.0){m=1;return m|0}k=g*+_(+j)*206264.8062470964/a;j=(b*i-g*f*l)*206264.8062470964/a;a=+p[c+168>>3];if(a==0.0){m=1;return m|0}l=+p[c+256>>3];f=+p[c+264>>3];g=+p[c+272>>3];i=+p[c+280>>3];b=+p[c+288>>3];n=+p[c+296>>3];o=+p[c+304>>3];q=+p[c+312>>3];r=+p[c+320>>3];s=+p[c+328>>3];t=+p[c+336>>3];u=+p[c+344>>3];v=+p[c+352>>3];w=i*2.0;x=o*2.0;y=q*3.0;z=r*2.0;A=n*2.0;B=s*2.0;C=t*3.0;D=u*2.0;E=v*4.0;F=+p[c+416>>3];G=+p[c+424>>3];H=+p[c+432>>3];I=+p[c+440>>3];J=+p[c+448>>3];K=+p[c+456>>3];L=+p[c+464>>3];M=+p[c+472>>3];N=+p[c+480>>3];O=+p[c+488>>3];P=+p[c+496>>3];Q=+p[c+504>>3];R=+p[c+512>>3];S=K*2.0;T=L*2.0;U=O*2.0;V=P*3.0;X=Q*2.0;Y=R*4.0;$=I*2.0;aa=M*3.0;ba=N*2.0;h=0;ca=k/a;da=j/a;while(1){a=da*ca;ea=ca*ca;fa=da*da;ga=da*ea;ha=fa*ca;ia=fa+ea;ja=ia*ia;ka=ca*ea;la=da*fa;ma=ea*ea;na=fa*fa;oa=fa*(ea*6.0);pa=fa*s+(da*b+(l+ca*w)+ca*x+ea*y+a*z)+(fa+ea*3.0)*u+(na+(ma*5.0+oa))*v;qa=f+ca*b+da*A+da*x+ea*r+a*B+fa*C+a*D+ia*(a*E);ra=G+da*J+ca*S+ca*T+fa*N+a*U+ea*V+a*X+ia*(a*Y);sa=ea*O+(ca*J+(F+da*$)+da*T+fa*aa+a*ba)+(fa*3.0+ea)*Q+(ma+(na*5.0+oa))*R;oa=g+(ca*l+da*f)+ea*i+a*b+fa*n+ia*o+ka*q+ga*r+ha*s+la*t+ia*(ca*u)+ja*(ca*v)-k;na=H+(da*F+ca*G)+fa*I+a*J+ea*K+ia*L+la*M+ha*N+ga*O+ka*P+ia*(da*Q)+ja*(da*R)-j;ja=pa*sa-qa*ra;ia=(qa*na-oa*sa)/ja;sa=(oa*ra-pa*na)/ja;ja=ca+ia;na=da+sa;if(+W(+ia)<5.0e-07?+W(+sa)<5.0e-07:0){ta=ja;ua=na;break}h=h+1|0;if((h|0)>=50){ta=ja;ua=na;break}else{ca=ja;da=na}}da=+p[c+192>>3];if(da==0.0){m=1;return m|0}ca=+p[c+200>>3];if(ca==0.0){m=1;return m|0}j=(ua*1.0e3+ +p[c+248>>3])/ca;p[d>>3]=(+p[c+224>>3]-ta*1.0e3)/da-+p[c+176>>3]+1.0+-.5;da=j-+p[c+184>>3]+1.0+-.5;p[e>>3]=da;j=+p[d>>3];if(j<.5){m=-1;return m|0}if(da<.5?1:j>+p[c+136>>3]+.5){m=-1;return m|0}if(da>+p[c+144>>3]+.5){m=-1;return m|0}m=0;return m|0}function Oe(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;c=k[31229]|0;k[d>>2]=a;k[d+4>>2]=b;Ew(c,605559,d)|0;gc(-1)}function Pe(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;e=r;r=r+544|0;f=e+8|0;g=e;h=e+32|0;i=e+16|0;j=ey(1,28)|0;if(!j)Oe(605569,605592);k[j+16>>2]=b;if((b|0)>4){k[g>>2]=b;k[g+4>>2]=4;Xw(h,605612,g)|0;Oe(605688,h)}if(b){g=cy(b<<2)|0;k[j+12>>2]=g;if(!g)Oe(605569,605698);else{l=a;a=b;m=g;while(1){a=a+-1|0;k[m>>2]=(k[l>>2]|0)+-1;if(!a){n=g;break}else{l=l+4|0;m=m+4|0}}}}else n=0;k[j+24>>2]=d;a:do if(d){m=d<<2;l=cy(m)|0;k[j+20>>2]=l;if(!l)Oe(605569,605734);sy(i|0,0,m|0)|0;b:do if((b|0)>0){m=0;while(1){g=k[n+(m<<2)>>2]|0;if((g|0)>=(d|0))break;a=i+(g<<2)|0;k[a>>2]=(k[a>>2]|0)+1;m=m+1|0;if((m|0)>=(b|0))break b}Oe(605773,795593)}while(0);m=j+8|0;k[m>>2]=1;if((d|0)>0){a=c;g=0;while(1){o=k[a>>2]|0;a=a+4|0;k[l+(g<<2)>>2]=o;if((o|0)>10){p=g;break}q=k[i+(g<<2)>>2]|0;if(!o){s=1;t=1}else{u=o;o=1;v=1;while(1){w=ha(v,u+q|0)|0;x=ha(o,u)|0;u=u+-1|0;if(!u){s=x;t=w;break}else{o=x;v=w}}}v=ha(k[m>>2]|0,(t|0)/(s|0)|0)|0;k[m>>2]=v;g=g+1|0;if((g|0)>=(d|0)){y=v;break a}}k[f>>2]=k[l+(p<<2)>>2];k[f+4>>2]=10;Xw(h,605812,f)|0;Oe(605688,h)}else y=1}else{k[j+8>>2]=1;y=1}while(0);h=cy(y<<3)|0;k[j>>2]=h;if(!h)Oe(605569,605880);h=ey(y,8)|0;k[j+4>>2]=h;if(!h)Oe(605569,605918);else{r=e;return j|0}return 0}function Qe(a){a=a|0;if(!a)return;dy(k[a+4>>2]|0);dy(k[a>>2]|0);dy(k[a+20>>2]|0);dy(k[a+12>>2]|0);dy(a);return}function Re(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0.0,s=0.0,t=0.0,u=0,v=0.0,w=0,x=0,y=0.0,z=0.0,A=0.0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0;c=r;r=r+80|0;d=c;e=c+60|0;f=c+40|0;g=k[a+16>>2]|0;h=k[a>>2]|0;i=k[a+4>>2]|0;j=k[a+12>>2]|0;l=k[a+20>>2]|0;if(g){m=g+-1|0;if(m){n=m;m=e;o=d;do{o=o+8|0;p[o>>3]=1.0;m=m+4|0;k[m>>2]=0;n=n+-1|0}while((n|0)!=0)}n=k[a+24>>2]|0;if(n){m=l;l=n;n=f;while(1){l=l+-1|0;k[n>>2]=k[m>>2];if(!l)break;else{m=m+4|0;n=n+4|0}}}n=f+(k[j>>2]<<2)|0;m=k[n>>2]|0;if(m)k[n>>2]=m+-1}m=i+8|0;q=+p[i>>3];i=h+8|0;p[h>>3]=1.0;k[e>>2]=1;s=+p[b>>3];p[d>>3]=s;h=k[a+8>>2]|0;a=h+-1|0;if(!a){t=q;r=c;return +t}if((g|0)>0){u=a;v=s;w=i;x=m;y=q}else{p[i>>3]=s;z=q+s*+p[m>>3];a=h+-2|0;if(!a){t=z;r=c;return +t}else{A=z;B=a;C=i;D=m}while(1){D=D+8|0;C=C+8|0;z=+p[d>>3];p[C>>3]=z;s=A+z*+p[D>>3];B=B+-1|0;if(!B){t=s;break}else A=s}r=c;return +t}while(1){B=w+8|0;p[w>>3]=v;D=x+8|0;A=y+v*+p[x>>3];C=0;m=e;i=j;a=b;h=d;while(1){n=f+(k[i>>2]<<2)|0;l=k[n>>2]|0;k[n>>2]=l+-1;n=k[m>>2]|0;if(l){E=n;F=C;G=m;H=a;I=h;J=14;break}k[f+(k[i>>2]<<2)>>2]=n;k[m>>2]=0;p[h>>3]=1.0;C=C+1|0;if((C|0)>=(g|0))break;else{m=m+4|0;i=i+4|0;a=a+8|0;h=h+8|0}}if((J|0)==14?(J=0,k[G>>2]=E+1,s=+p[H>>3]*+p[I>>3],p[I>>3]=s,(F|0)!=0):0){h=F;a=I;do{a=a+-8|0;h=h+-1|0;p[a>>3]=s}while((h|0)!=0)}h=u+-1|0;if(!h){t=A;break}u=h;v=+p[d>>3];w=B;x=D;y=A}r=c;return +t}function Se(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0.0,C=0.0,D=0,E=0.0,F=0,G=0,H=0;g=r;r=r+32|0;h=g;i=(b|0)!=0;j=(f|0)!=0;if(!(i|j))Oe(605956,606029);l=k[a+8>>2]|0;m=k[a+16>>2]|0;n=ha(l,l)|0;o=k[a>>2]|0;q=ey(n,8)|0;if(!q)Oe(605569,606041);n=ey(l,8)|0;if(!n)Oe(605569,606068);if(e){s=(l|0)==0;t=(m|0)>0;u=s|j^1;j=d;d=c;c=e;e=f;f=b;while(1){c=c+-1|0;if(i){if(t){b=0;v=f;while(1){p[h+(b<<3)>>3]=+p[v>>3];b=b+1|0;if((b|0)==(m|0))break;else v=v+8|0}w=f+(m<<3)|0}else w=f;+Re(a,h);if(u){x=e;y=w}else{v=o;b=e;z=l;while(1){z=z+-1|0;p[b>>3]=+p[v>>3];if(!z)break;else{v=v+8|0;b=b+8|0}}x=e+(l<<3)|0;y=w}}else if(s){x=e;y=f}else{b=o;v=e;z=l;while(1){z=z+-1|0;p[b>>3]=+p[v>>3];if(!z)break;else{b=b+8|0;v=v+8|0}}x=e+(l<<3)|0;y=f}if(!j){A=0;B=1.0}else{A=j+8|0;B=+p[j>>3]}C=+p[d>>3];d=d+8|0;if(!s){v=l;b=q;z=o;D=n;while(1){v=v+-1|0;E=B*+p[z>>3];z=z+8|0;p[D>>3]=+p[D>>3]+C*E;D=D+8|0;F=b;G=o;H=l;while(1){H=H+-1|0;p[F>>3]=+p[F>>3]+E*+p[G>>3];if(!H)break;else{F=F+8|0;G=G+8|0}}if(!v)break;else b=b+(l<<3)|0}}if(!c)break;else{j=A;e=x;f=y}}}Te(q,n,l);dy(q);if(!l){dy(n);r=g;return}q=n;y=k[a+4>>2]|0;a=l;while(1){a=a+-1|0;p[y>>3]=+p[q>>3];if(!a)break;else{q=q+8|0;y=y+8|0}}dy(n);r=g;return}function Te(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;if(!(Ue(a,b,c)|0))return;d=c<<3;e=cy(ha(d,c)|0)|0;if(!e)Oe(605569,606093);f=cy(d)|0;if(!f)Oe(605569,606115);Ve(a,b,c,c,e,f);dy(e);dy(f);return}function Ue(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,k=0,l=0.0,m=0,n=0.0,o=0.0,q=0,r=0,s=0.0,t=0.0;d=cy(c<<3)|0;if(!d)Oe(605569,606135);e=(c|0)>0;if(e){f=0;a:do{g=ha(f,c)|0;h=d+(f<<3)|0;if((f|0)>0){i=f;do{j=ha(i,c)|0;k=f;l=+p[a+(i+g<<3)>>3];while(1){m=k;k=k+-1|0;n=l-+p[a+(k+g<<3)>>3]*+p[a+(k+j<<3)>>3];if((m|0)<=1){o=n;break}else l=n}if((f|0)==(i|0)){if(o<=0.0){q=15;break a}l=+X(+o);p[h>>3]=l}else{j=a+((ha(i,c)|0)+f<<3)|0;p[j>>3]=o/+p[h>>3]}i=i+1|0}while((i|0)<(c|0))}else{i=f;do{l=+p[a+(i+g<<3)>>3];if((f|0)==(i|0)){if(l<=0.0){q=15;break a}n=+X(+l);p[h>>3]=n}else{j=a+((ha(i,c)|0)+f<<3)|0;p[j>>3]=l/+p[h>>3]}i=i+1|0}while((i|0)<(c|0))}f=f+1|0}while((f|0)<(c|0));if((q|0)==15){dy(d);r=-1;return r|0}if(e){q=0;do{f=b+(q<<3)|0;o=+p[f>>3];if((q|0)>0){i=ha(q,c)|0;h=q;l=o;while(1){g=h;h=h+-1|0;n=l-+p[a+(h+i<<3)>>3]*+p[b+(h<<3)>>3];if((g|0)<=1){s=n;break}else l=n}}else s=o;p[f>>3]=s/+p[d+(q<<3)>>3];q=q+1|0}while((q|0)!=(c|0));if(e){e=c;do{q=e;e=e+-1|0;h=b+(e<<3)|0;s=+p[h>>3];if((q|0)<(c|0)){i=q;l=s;while(1){g=a+((ha(i,c)|0)+e<<3)|0;n=l-+p[g>>3]*+p[b+(i<<3)>>3];i=i+1|0;if((i|0)==(c|0)){t=n;break}else l=n}}else t=s;p[h>>3]=t/+p[d+(e<<3)>>3]}while((q|0)>1)}}}dy(d);r=0;return r|0}function Ve(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,k=0.0,l=0.0,m=0,n=0.0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0,E=0,F=0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0,O=0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0,V=0,Y=0,Z=0,_=0.0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0.0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0.0,sa=0.0,ta=0,ua=0.0,va=0.0,wa=0.0,xa=0.0,ya=0.0,za=0.0,Aa=0.0,Ba=0.0,Ca=0.0,Da=0.0,Ea=0.0,Fa=0.0,Ga=0,Ha=0.0,Ia=0,Ja=0.0,Ka=0.0,La=0,Ma=0.0,Na=0,Oa=0,Pa=0.0;if((c|0)<(d|0))Oe(606152,606201);g=d<<3;h=cy(g)|0;if(!h)Oe(605569,606213);i=cy(g)|0;if(!i)Oe(605569,606232);g=(d|0)>0;a:do if(g){j=c+1|0;k=0.0;l=0.0;m=0;n=0.0;while(1){o=m;m=m+1|0;q=d-m|0;r=h+(o<<3)|0;p[r>>3]=l*n;s=c-o|0;if((s|0)>0?(t=a+((ha(o,j)|0)<<3)|0,u=(o|0)==(c|0),!u):0){v=t;w=s;x=0.0;while(1){w=w+-1|0;y=x+ +W(+(+p[v>>3]));if(!w){z=y;break}else{v=v+8|0;x=y}}if(z!=0.0){if(u)A=0.0;else{v=t;w=s;x=0.0;while(1){w=w+-1|0;y=+p[v>>3]/z;p[v>>3]=y;B=x+y*y;if(!w){A=B;break}else{v=v+8|0;x=B}}}x=+p[t>>3];B=+W(+(+X(+A)));y=x>=0.0?B:-B;B=-y;C=x*B-A;p[t>>3]=x+y;if((m|0)!=(d|0)){v=q;w=a+((ha(m,c)|0)+o<<3)|0;while(1){v=v+-1|0;if(!u){D=t;E=w;F=s;y=0.0;while(1){F=F+-1|0;x=y+ +p[E>>3]*+p[D>>3];if(!F){G=x;break}else{D=D+8|0;E=E+8|0;y=x}}y=G/C;if(!u){E=t;D=w;F=s;while(1){F=F+-1|0;p[D>>3]=+p[D>>3]+y*+p[E>>3];if(!F)break;else{E=E+8|0;D=D+8|0}}}}if(!v)break;else w=w+(c<<3)|0}}if(u){H=B;I=z}else{w=t;v=s;while(1){v=v+-1|0;p[w>>3]=z*+p[w>>3];if(!v){H=B;I=z;break}else w=w+8|0}}}else{H=0.0;I=z}}else{H=0.0;I=0.0}B=H*I;w=f+(o<<3)|0;p[w>>3]=B;if(!((o|0)>=(c|0)|(m|0)==(d|0))){v=ha(m,c)|0;s=v+o|0;t=a+(s<<3)|0;u=t;D=q;C=0.0;while(1){D=D+-1|0;y=C+ +W(+(+p[u>>3]));if(!D){J=y;break}else{u=u+(c<<3)|0;C=y}}if(J!=0.0){u=t;D=q;C=0.0;while(1){D=D+-1|0;y=+p[u>>3]/J;p[u>>3]=y;x=C+y*y;if(!D){K=x;break}else{u=u+(c<<3)|0;C=x}}C=+p[t>>3];x=+W(+(+X(+K)));y=C>=0.0?x:-x;x=-y;L=C*x-K;M=C+y;p[t>>3]=M;u=h+(m<<3)|0;D=q+-1|0;p[u>>3]=M/L;if(D){o=a+(s+c<<3)|0;E=D;D=u;while(1){D=D+8|0;E=E+-1|0;p[D>>3]=+p[o>>3]/L;if(!E)break;else o=o+(c<<3)|0}}if((m|0)==(c|0)){N=t;O=q}else{o=c-m|0;E=a+(v+m<<3)|0;while(1){o=o+-1|0;D=t;s=E;F=q;L=0.0;while(1){F=F+-1|0;M=L+ +p[s>>3]*+p[D>>3];if(!F){P=M;break}else{D=D+(c<<3)|0;s=s+(c<<3)|0;L=M}}s=E;D=q;F=u;while(1){D=D+-1|0;p[s>>3]=+p[s>>3]+P*+p[F>>3];if(!D)break;else{s=s+(c<<3)|0;F=F+8|0}}if(!o){N=t;O=q;break}else E=E+8|0}}while(1){O=O+-1|0;p[N>>3]=J*+p[N>>3];if(!O)break;else N=N+(c<<3)|0}Q=+p[w>>3];R=x;S=J}else{Q=B;R=0.0;S=J}}else{Q=B;R=0.0;S=0.0}L=+W(+Q);M=L+ +W(+(+p[r>>3]));L=k>M?k:M;if((m|0)==(d|0)){T=L;U=R;break}else{k=L;l=R;n=S}}m=d+-1|0;if(g){n=U;j=m;E=0;q=d;t=0;while(1){do if((j|0)<(m|0)){if(n!=0.0){o=j+(ha(q,c)|0)|0;u=a+(o<<3)|0;v=e+((ha(j,d)|0)+q<<3)|0;F=e+((ha(q,d)|0)+q<<3)|0;l=+p[u>>3];k=n*l;s=(t|0)==0;if(s)break;D=t+-1|0;p[v>>3]=l/k;if(D){V=a+(o+c<<3)|0;o=D;D=v;while(1){D=D+8|0;o=o+-1|0;p[D>>3]=+p[V>>3]/k;if(!o)break;else V=V+(c<<3)|0}}if(s)break;else{Y=t;Z=F}while(1){Y=Y+-1|0;V=u;o=t;k=0.0;D=Z;while(1){o=o+-1|0;l=k+ +p[V>>3]*+p[D>>3];if(!o){_=l;break}else{V=V+(c<<3)|0;k=l;D=D+8|0}}D=t;V=v;o=Z;while(1){D=D+-1|0;p[o>>3]=+p[o>>3]+_*+p[V>>3];if(!D)break;else{V=V+8|0;o=o+8|0}}if(!Y)break;else Z=Z+(d<<3)|0}}if(t){v=t;u=e+(j+(ha(q,d)|0)<<3)|0;F=e+((ha(j,d)|0)+q<<3)|0;while(1){v=v+-1|0;p[F>>3]=0.0;p[u>>3]=0.0;if(!v)break;else{u=u+(d<<3)|0;F=F+8|0}}}}while(0);r=e+((ha(j,d)|0)+j<<3)|0;p[r>>3]=1.0;t=d-j|0;E=E+1|0;if((E|0)==(d|0))break;else{r=j;n=+p[h+(j<<3)>>3];j=j+-1|0;q=r}}if(g){q=(d|0)<(c|0)?d:c;j=~q;E=a+((ha(q+-1|0,c)|0)+-2-j<<3)|0;q=-8-(c<<3)|0;t=(c+j<<3)+16|0;j=d;m=0;while(1){r=E+(ha(q,m)|0)|0;w=t+(m<<3)|0;F=j;j=j+-1|0;u=d-F|0;v=c-j|0;n=+p[f+(j<<3)>>3];s=ha(j,c)|0;o=s+j|0;V=a+(o<<3)|0;D=a+(o+c<<3)|0;o=(F|0)==(d|0);if(!o){$=D;aa=u;while(1){aa=aa+-1|0;p[$>>3]=0.0;if(!aa)break;else $=$+(c<<3)|0}}if(n!=0.0){B=1.0/n;if(!o){$=v+-1|0;aa=($|0)==0;ba=(j|0)==(c|0);ca=v+-1|0;da=(ca|0)==0;ea=a+(s+F<<3)|0;fa=u;ga=D;while(1){fa=fa+-1|0;if(aa)ia=0.0;else{ja=$;ka=V;la=ga;x=0.0;while(1){ka=ka+8|0;la=la+8|0;k=x+ +p[ka>>3]*+p[la>>3];ja=ja+-1|0;if(!ja){ia=k;break}else x=k}}x=+p[V>>3];k=B*(ia/x);if(!ba?(p[ga>>3]=+p[ga>>3]+k*x,!da):0){ja=ea;la=ca;ka=ga;while(1){ka=ka+8|0;la=la+-1|0;p[ka>>3]=+p[ka>>3]+k*+p[ja>>3];if(!la)break;else ja=ja+8|0}}if(!fa)break;else ga=ga+(c<<3)|0}}if((j|0)!=(c|0)){ga=V;fa=v;while(1){fa=fa+-1|0;p[ga>>3]=B*+p[ga>>3];if(!fa)break;else ga=ga+8|0}}}else if((j|0)!=(c|0))sy(r|0,0,w|0)|0;p[V>>3]=+p[V>>3]+1.0;if((F|0)<=1)break;else m=m+1|0}if(g){m=(c|0)==0;j=d;t=0;b:while(1){q=j;j=j+-1|0;E=f+(j<<3)|0;ga=q+-2|0;fa=f+(ga<<3)|0;v=h+(ga<<3)|0;ca=h+(j<<3)|0;ea=0;da=t;while(1){ba=j;$=da;while(1){if((ba|0)<=-1){ma=ba;na=$;oa=85;break}aa=ba+-1|0;if(T+ +W(+(+p[h+(ba<<3)>>3]))==T){pa=ba;qa=aa;break}if(T+ +W(+(+p[f+(aa<<3)>>3]))==T){ma=ba;na=aa;oa=85;break}else{ba=aa;$=aa}}c:do if((oa|0)==85){oa=0;$=a+((ha(na,c)|0)<<3)|0;if((ma|0)>(j|0)){pa=ma;qa=na}else{ba=a+((ha(ma,c)|0)<<3)|0;aa=ma;B=1.0;while(1){n=B*+p[h+(aa<<3)>>3];k=+W(+n);if(T+k==T){pa=ma;qa=na;break c}D=f+(aa<<3)|0;x=+p[D>>3];l=+W(+x);if(!(k>l))if(x!=0.0){L=k/l;ra=l*+X(+(L*L+1.0))}else ra=0.0;else{L=l/k;ra=k*+X(+(L*L+1.0))}p[D>>3]=ra;L=1.0/ra;k=x*L;B=-(n*L);if(!m){D=$;u=ba;s=c;while(1){s=s+-1|0;L=+p[u>>3];n=+p[D>>3];p[D>>3]=L*B+k*n;p[u>>3]=k*L-n*B;if(!s)break;else{D=D+8|0;u=u+8|0}}}if((aa|0)>=(j|0)){pa=ma;qa=na;break}else{ba=ba+(c<<3)|0;aa=aa+1|0}}}}while(0);B=+p[E>>3];if((pa|0)==(j|0)){sa=B;ta=qa;oa=96;break}if((ea|0)==99)break b;k=+p[f+(pa<<3)>>3];n=+p[fa>>3];L=+p[v>>3];x=+p[ca>>3];l=((n-B)*(B+n)+(L-x)*(L+x))/(n*(x*2.0));L=+W(+l);if(L>1.0){M=1.0/L;ua=L*+X(+(M*M+1.0))}else ua=+X(+(L*L+1.0));L=+W(+ua);M=((k-B)*(B+k)+x*(n/(l+(l>=0.0?L:-L))-x))/k;if((pa|0)>(ga|0)){va=M;wa=k}else{aa=a+((ha(pa,c)|0)<<3)|0;x=1.0;L=M;ba=pa;M=1.0;$=e+((ha(pa,d)|0)<<3)|0;l=k;while(1){u=ba;ba=ba+1|0;k=+p[h+(ba<<3)>>3];n=+p[f+(ba<<3)>>3];B=M*k;y=x*k;k=+W(+L);C=+W(+B);if(!(k>C))if(B!=0.0){xa=k/C;ya=C*+X(+(xa*xa+1.0))}else ya=0.0;else{xa=C/k;ya=k*+X(+(xa*xa+1.0))}p[h+(u<<3)>>3]=ya;xa=L/ya;k=B/ya;B=l*xa+y*k;C=y*xa-l*k;y=n*k;za=n*xa;D=$;$=$+(d<<3)|0;s=d;o=$;ja=D;while(1){s=s+-1|0;n=+p[o>>3];Aa=+p[ja>>3];p[ja>>3]=k*n+xa*Aa;p[o>>3]=xa*n-k*Aa;if(!s)break;else{o=o+8|0;ja=ja+8|0}}Aa=+W(+B);n=+W(+y);do if(!(Aa>n))if(y!=0.0){Ba=Aa/n;Ca=n*+X(+(Ba*Ba+1.0));oa=119;break}else{p[f+(u<<3)>>3]=0.0;Da=xa;Ea=k;break}else{Ba=n/Aa;Ca=Aa*+X(+(Ba*Ba+1.0));oa=119}while(0);if((oa|0)==119){oa=0;p[f+(u<<3)>>3]=Ca;if(Ca!=0.0){Aa=1.0/Ca;Da=B*Aa;Ea=y*Aa}else{Da=xa;Ea=k}}Aa=C*Da+za*Ea;n=za*Da-C*Ea;ja=aa;aa=aa+(c<<3)|0;if(!m){o=aa;s=ja;ja=c;while(1){ja=ja+-1|0;Ba=+p[o>>3];Fa=+p[s>>3];p[s>>3]=Ea*Ba+Da*Fa;p[o>>3]=Da*Ba-Ea*Fa;if(!ja)break;else{o=o+8|0;s=s+8|0}}}if((u|0)>=(ga|0)){va=Aa;wa=n;break}else{x=Da;L=Aa;M=Ea;l=n}}}p[h+(pa<<3)>>3]=0.0;p[ca>>3]=va;p[E>>3]=wa;ea=ea+1|0;if((ea|0)>=100){Ga=ga;break}else da=ga}if((oa|0)==96){oa=0;if(sa<0.0){p[E>>3]=-sa;ga=d;da=e+((ha(j,d)|0)<<3)|0;while(1){ga=ga+-1|0;p[da>>3]=-+p[da>>3];if(!ga){Ga=ta;break}else da=da+8|0}}else Ga=ta}if((q|0)<=1)break a;else t=Ga}Oe(606251,606201)}}}}while(0);Ga=(d|0)==0;if(!Ga){ta=d;oa=f;sa=0.0;while(1){ta=ta+-1|0;wa=+p[oa>>3];va=wa>sa?wa:sa;if(!ta){Ha=va;break}else{oa=oa+8|0;sa=va}}sa=Ha*1.0e-11;if(!Ga){oa=d;ta=f;while(1){oa=oa+-1|0;if(+p[ta>>3]<sa)p[ta>>3]=0.0;if(!oa)break;else ta=ta+8|0}if(!Ga){Ga=(c|0)==0;ta=d;oa=a;a=i;pa=f;while(1){ta=ta+-1|0;sa=+p[pa>>3];if(sa!=0.0){if(Ga){Ia=oa;Ja=0.0}else{f=oa;qa=b;na=c;Ha=0.0;while(1){na=na+-1|0;va=Ha+ +p[f>>3]*+p[qa>>3];if(!na){Ka=va;break}else{f=f+8|0;qa=qa+8|0;Ha=va}}Ia=oa+(c<<3)|0;Ja=Ka}La=Ia;Ma=Ja/sa}else{La=oa+(c<<3)|0;Ma=0.0}p[a>>3]=Ma;if(!ta)break;else{oa=La;a=a+8|0;pa=pa+8|0}}}}}if(g){Na=0;Oa=e}else{dy(i);dy(h);return}while(1){e=d;Ma=0.0;g=i;pa=Oa;while(1){e=e+-1|0;Ja=Ma+ +p[pa>>3]*+p[g>>3];if(!e){Pa=Ja;break}else{Ma=Ja;g=g+8|0;pa=pa+(d<<3)|0}}p[b+(Na<<3)>>3]=Pa;Na=Na+1|0;if((Na|0)==(d|0))break;else Oa=Oa+8|0}dy(i);dy(h);return}function We(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0;d=r;r=r+144|0;e=d+120|0;f=d+112|0;g=d+104|0;h=d+96|0;j=d+88|0;l=d+80|0;m=d+72|0;n=d+64|0;o=d+56|0;p=d+48|0;q=d+40|0;s=d+32|0;t=d+24|0;u=d+16|0;v=d+8|0;w=d;x=d+128|0;y=w;k[y>>2]=5260110;k[y+4>>2]=5459015;y=c+4|0;i[y>>0]=0;i[x>>0]=0;z=c+20|0;k[z>>2]=-1;A=c+24|0;k[A>>2]=-1;B=c+28|0;k[B>>2]=-1;do if((a|0)>0){C=c+8|0;D=c+13|0;E=w+4|0;F=0;G=0;a:while(1){H=b+(F*9|0)|0;do if((i[b+(F*9|0)+4>>0]|0)!=45)if(!(tx(H,606298)|0)){if((k[B>>2]|0)!=-1){I=1;J=34;break a}k[B>>2]=F;K=G}else K=G;else{L=b+(F*9|0)+5|0;M=k[26016]|0;b:do if((M|0)>0){N=0;while(1){if(!(Ax(L,606386+(N<<2)|0,3)|0)){O=N;break b}P=N+1|0;if((P|0)<(M|0))N=P;else{O=P;break}}}else O=0;while(0);if(((O|0)==(M|0)?(Ax(L,w,3)|0)!=0:0)?(Ax(L,E,3)|0)!=0:0){K=G;break}if(i[y>>0]|0){if((G|0)==0|(Ax(H,x,8)|0)!=0){I=1;J=34;break a}k[G>>2]=F;i[x>>0]=0;K=G;break}k[v>>2]=L;Xw(y,606307,v)|0;if(!(Ax(H,606312,4)|0)){k[z>>2]=F;i[C>>0]=i[606317]|0;i[C+1>>0]=i[606318]|0;i[C+2>>0]=i[606319]|0;i[D>>0]=68;i[D+1>>0]=69;i[D+2>>0]=67;i[D+3>>0]=0;k[u>>2]=y;Xw(x,606320,u)|0;K=A;break}if(!(Ax(H,777966,4)|0)){k[A>>2]=F;i[C>>0]=i[606317]|0;i[C+1>>0]=i[606318]|0;i[C+2>>0]=i[606319]|0;i[D>>0]=68;i[D+1>>0]=69;i[D+2>>0]=67;i[D+3>>0]=0;k[t>>2]=y;Xw(x,606328,t)|0;K=z;break}N=b+(F*9|0)+1|0;if(!(Ax(N,606336,3)|0)){k[z>>2]=F;k[s>>2]=i[H>>0];Xw(C,606340,s)|0;k[q>>2]=i[H>>0];Xw(D,606346,q)|0;k[p>>2]=D;k[p+4>>2]=y;Xw(x,606352,p)|0;K=A;break}if(!(Ax(N,777971,3)|0)){k[A>>2]=F;k[o>>2]=i[H>>0];Xw(C,606340,o)|0;k[n>>2]=i[H>>0];Xw(D,606346,n)|0;k[m>>2]=C;k[m+4>>2]=y;Xw(x,606352,m)|0;K=z;break}P=b+(F*9|0)+2|0;if(!(Ax(P,606358,2)|0)){k[z>>2]=F;Q=i[N>>0]|0;k[l>>2]=i[H>>0];k[l+4>>2]=Q;Xw(C,606361,l)|0;Q=i[N>>0]|0;k[j>>2]=i[H>>0];k[j+4>>2]=Q;Xw(D,606368,j)|0;k[h>>2]=D;k[h+4>>2]=y;Xw(x,606352,h)|0;K=A;break}if(Ax(P,606375,2)|0){I=1;J=34;break a}k[A>>2]=F;P=i[N>>0]|0;k[g>>2]=i[H>>0];k[g+4>>2]=P;Xw(C,606361,g)|0;P=i[N>>0]|0;k[f>>2]=i[H>>0];k[f+4>>2]=P;Xw(D,606368,f)|0;k[e>>2]=C;k[e+4>>2]=y;Xw(x,606352,e)|0;K=z}while(0);F=F+1|0;if((F|0)>=(a|0)){J=29;break}else G=K}if((J|0)==29){if(!(i[x>>0]|0))break;else I=1;r=d;return I|0}else if((J|0)==34){r=d;return I|0}}while(0);if(!(Ax(y,606378,3)|0)){i[y>>0]=83;i[y+1>>0]=70;i[y+2>>0]=76;i[y+3>>0]=0;R=137}else R=(i[y>>0]|0)==0?999:137;k[c>>2]=R;I=0;r=d;return I|0}function Xe(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,q=0,r=0,s=0.0,t=0.0,u=0,v=0.0;n=l+4|0;if((k[b>>2]|0)!=137?(We(k[n>>2]|0,a,b)|0)!=0:0){o=1;return o|0}a=k[n>>2]|0;if((a|0)>0){n=k[b+20>>2]|0;q=b+24|0;r=0;do{if((r|0)!=(n|0)?(r|0)!=(k[q>>2]|0):0)p[j+(r<<3)>>3]=+p[c+(r<<3)>>3]-+p[d+(r<<3)>>3];r=r+1|0}while((r|0)<(a|0))}do if((k[b>>2]|0)!=999){a=b+4|0;do if(!(tx(a,606382)|0)){r=e+16|0;if(+p[r>>3]==0.0){o=2;return o|0}else{i[a>>0]=83;i[a+1>>0]=73;i[a+2>>0]=78;i[a+3>>0]=0;p[h+40>>3]=0.0;s=+og(+p[r>>3]);t=s/+pg(+p[r>>3]);p[h+48>>3]=t;r=h+4|0;k[r>>2]=k[r>>2]>>31;break}}while(0);r=b+20|0;d=k[r>>2]|0;q=b+24|0;n=k[q>>2]|0;u=cf(a,+p[c+(d<<3)>>3],+p[c+(n<<3)>>3],e,f,g,h,j+(d<<3)|0,j+(n<<3)|0)|0;if(u){o=u;return o|0}u=k[b+28>>2]|0;if((u|0)!=-1){t=+p[h+24>>3];s=t==0.0?90.0:t*3.141592653589793*.5;n=j+(k[q>>2]<<3)|0;t=+p[n>>3];if(t<s*-.5){p[n>>3]=t+s;p[j+(u<<3)>>3]=5.0;break}v=s*.5;if(t>v){p[n>>3]=t-s;p[j+(u<<3)>>3]=0.0;break}n=j+(k[r>>2]<<3)|0;t=+p[n>>3];if(t>s*2.5){p[n>>3]=t-s*3.0;p[j+(u<<3)>>3]=4.0;break}if(t>s*1.5){p[n>>3]=t-s*2.0;p[j+(u<<3)>>3]=3.0;break}if(t>v){p[n>>3]=t-s;p[j+(u<<3)>>3]=2.0;break}else{p[j+(u<<3)>>3]=1.0;break}}}while(0);h=($e(j,l,m)|0)==0;o=h?0:4;return o|0}function Ye(a,b,c,d,e,f,g,h,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,q=0.0,r=0.0;if((k[b>>2]|0)!=137?(We(k[d+4>>2]|0,a,b)|0)!=0:0){n=1;return n|0}if(af(c,d,e)|0){n=4;return n|0}c=k[d+4>>2]|0;if((c|0)>0){d=k[b+20>>2]|0;a=b+24|0;o=0;do{if((o|0)!=(d|0)?(o|0)!=(k[a>>2]|0):0)p[m+(o<<3)>>3]=+p[e+(o<<3)>>3]+ +p[j+(o<<3)>>3];o=o+1|0}while((o|0)<(c|0))}if((k[b>>2]|0)!=999){c=k[b+28>>2]|0;a:do if((c|0)!=-1){q=+p[e+(c<<3)>>3];o=~~(q+.5);if(+W(+(q-+(o|0)))>1.0e-10){n=3;return n|0}q=+p[f+24>>3];r=q==0.0?90.0:q*3.141592653589793*.5;switch(o|0){case 1:{break a;break}case 0:{o=e+(k[b+24>>2]<<3)|0;p[o>>3]=r+ +p[o>>3];break a;break}case 2:{o=e+(k[b+20>>2]<<3)|0;p[o>>3]=r+ +p[o>>3];break a;break}case 3:{o=e+(k[b+20>>2]<<3)|0;p[o>>3]=r*2.0+ +p[o>>3];break a;break}case 4:{o=e+(k[b+20>>2]<<3)|0;p[o>>3]=r*3.0+ +p[o>>3];break a;break}case 5:{o=e+(k[b+24>>2]<<3)|0;p[o>>3]=+p[o>>3]-r;break a;break}default:{n=3;return n|0}}}while(0);c=b+4|0;do if(!(tx(c,606382)|0)){o=l+16|0;if(+p[o>>3]==0.0){n=2;return n|0}else{i[c>>0]=83;i[c+1>>0]=73;i[c+2>>0]=78;i[c+3>>0]=0;p[f+40>>3]=0.0;r=+og(+p[o>>3]);q=r/+pg(+p[o>>3]);p[f+48>>3]=q;o=f+4|0;k[o>>2]=k[o>>2]>>31;break}}while(0);o=k[b+20>>2]|0;j=k[b+24>>2]|0;b=df(c,+p[e+(o<<3)>>3],+p[e+(j<<3)>>3],f,g,h,l,m+(o<<3)|0,m+(j<<3)|0)|0;if(b){n=b;return n|0}}n=0;return n|0}function Ze(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;b=k[a+4>>2]|0;c=ha(b<<3,b)|0;d=cy(c)|0;k[a+20>>2]=d;if(!d){e=1;return e|0}f=cy(c)|0;k[a+24>>2]=f;if(!f){dy(d);e=1;return e|0}if((b|0)>0){c=k[a+16>>2]|0;g=k[a+12>>2]|0;h=0;i=0;while(1){j=c+(h<<3)|0;l=i;m=0;while(1){p[d+(l<<3)>>3]=+p[j>>3]*+p[g+(l<<3)>>3];m=m+1|0;if((m|0)==(b|0))break;else l=l+1|0}h=h+1|0;if((h|0)==(b|0))break;else i=b+i|0}}if(_e(b,d,f)|0){e=2;return e|0}k[a>>2]=137;e=0;return e|0}function _e(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0.0,q=0,r=0,s=0.0,t=0.0,u=0.0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;d=a<<2;e=cy(d)|0;if(!e){f=1;return f|0}g=cy(d)|0;if(!g){dy(e);f=1;return f|0}d=a<<3;h=cy(d)|0;if(!h){dy(e);dy(g);f=1;return f|0}i=cy(ha(d,a)|0)|0;if(!i){dy(e);dy(g);dy(h);f=1;return f|0}d=(a|0)>0;if(d){j=a<<3;l=0;m=0;while(1){k[e+(l<<2)>>2]=l;n=h+(l<<3)|0;p[n>>3]=0.0;py(i+(m<<3)|0,b+(m<<3)|0,j|0)|0;o=0.0;q=m;r=0;while(1){s=+W(+(+p[b+(q<<3)>>3]));if(s>o){p[n>>3]=s;t=s}else t=o;r=r+1|0;if((r|0)==(a|0)){u=t;break}else{o=t;q=q+1|0}}l=l+1|0;if(u==0.0){v=17;break}if((l|0)>=(a|0))break;else m=m+a|0}if((v|0)==17){dy(e);dy(g);dy(h);dy(i);f=2;return f|0}if(d){v=0;do{m=ha(v,a)|0;l=i+(m+v<<3)|0;b=h+(v<<3)|0;j=v;v=v+1|0;q=(v|0)<(a|0);if(q){u=+W(+(+p[l>>3]));t=u/+p[b>>3];r=v;n=j;while(1){w=i+((ha(r,a)|0)+j<<3)|0;u=+W(+(+p[w>>3]));o=u/+p[h+(r<<3)>>3];w=o>t;x=w?r:n;r=r+1|0;if((r|0)==(a|0)){y=x;break}else{t=w?o:t;n=x}}if((y|0)>(j|0)){n=0;r=m;x=ha(y,a)|0;while(1){w=i+(x<<3)|0;t=+p[w>>3];z=i+(r<<3)|0;p[w>>3]=+p[z>>3];p[z>>3]=t;n=n+1|0;if((n|0)==(a|0))break;else{r=r+1|0;x=x+1|0}}x=h+(y<<3)|0;t=+p[x>>3];p[x>>3]=+p[b>>3];p[b>>3]=t;x=e+(y<<2)|0;r=k[x>>2]|0;n=e+(j<<2)|0;k[x>>2]=k[n>>2];k[n>>2]=r}if(q){r=i+(v+m<<3)|0;n=j+2|0;x=(n|0)==(a|0);z=v;do{w=ha(z,a)|0;A=i+(w+j<<3)|0;t=+p[A>>3];if(t!=0.0?(o=t/+p[l>>3],p[A>>3]=o,B=i+(v+w<<3)|0,p[B>>3]=+p[B>>3]-o*+p[r>>3],!x):0){B=n;do{C=i+(B+w<<3)|0;p[C>>3]=+p[C>>3]-+p[A>>3]*+p[i+(B+m<<3)>>3];B=B+1|0}while((B|0)!=(a|0))}z=z+1|0}while((z|0)!=(a|0))}}}while((v|0)!=(a|0));if(d){v=0;do{k[g+(k[e+(v<<2)>>2]<<2)>>2]=v;v=v+1|0}while((v|0)!=(a|0));if(d){v=a<<3;y=0;z=0;while(1){sy(c+(z<<3)|0,0,v|0)|0;y=y+1|0;if((y|0)==(a|0))break;else z=z+a|0}if(d){d=0;do{z=k[g+(d<<2)>>2]|0;y=c+((ha(z,a)|0)+d<<3)|0;p[y>>3]=1.0;y=z+1|0;if((y|0)<(a|0)){v=y;do{if((z|0)<(v|0)){y=ha(v,a)|0;m=c+(y+d<<3)|0;o=+p[m>>3];n=z;do{x=c+((ha(n,a)|0)+d<<3)|0;o=o-+p[i+(n+y<<3)>>3]*+p[x>>3];p[m>>3]=o;n=n+1|0}while((n|0)!=(v|0))}v=v+1|0}while((v|0)!=(a|0));D=a}else D=a;do{v=D;D=D+-1|0;z=ha(D,a)|0;n=c+(z+d<<3)|0;if((v|0)<(a|0)){o=+p[n>>3];m=v;do{y=c+((ha(m,a)|0)+d<<3)|0;o=o-+p[i+(m+z<<3)>>3]*+p[y>>3];p[n>>3]=o;m=m+1|0}while((m|0)!=(a|0))}p[n>>3]=+p[n>>3]/+p[i+(z+D<<3)>>3]}while((v|0)>1);d=d+1|0}while((d|0)!=(a|0))}}}}}dy(e);dy(g);dy(h);dy(i);f=0;return f|0}function $e(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0.0,m=0,n=0;d=k[b+4>>2]|0;if((k[b>>2]|0)!=137?(Ze(b)|0)!=0:0){e=1;return e|0}f=(d|0)>0;if(!f){e=0;return e|0}g=k[b+24>>2]|0;h=0;i=0;while(1){j=c+(h<<3)|0;p[j>>3]=0.0;l=0.0;m=i;n=0;while(1){l=l+ +p[g+(m<<3)>>3]*+p[a+(n<<3)>>3];p[j>>3]=l;n=n+1|0;if((n|0)==(d|0))break;else m=m+1|0}h=h+1|0;if((h|0)==(d|0))break;else i=d+i|0}if(!f){e=0;return e|0}f=k[b+8>>2]|0;b=0;do{i=c+(b<<3)|0;p[i>>3]=+p[f+(b<<3)>>3]+ +p[i>>3];b=b+1|0}while((b|0)!=(d|0));e=0;return e|0}function af(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0.0,i=0,j=0,l=0;d=k[b+4>>2]|0;if((k[b>>2]|0)!=137?(Ze(b)|0)!=0:0){e=1;return e|0}if((d|0)<=0){e=0;return e|0}sy(c|0,0,d<<3|0)|0;f=k[b+8>>2]|0;g=k[b+20>>2]|0;b=0;do{h=+p[a+(b<<3)>>3]-+p[f+(b<<3)>>3];i=0;j=b;while(1){l=c+(i<<3)|0;p[l>>3]=+p[l>>3]+h*+p[g+(j<<3)>>3];i=i+1|0;if((i|0)==(d|0))break;else j=j+d|0}b=b+1|0}while((b|0)!=(d|0));e=0;return e|0}function bf(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0,z=0,A=0,B=0.0,C=0,D=0.0;if(ef(a,c)|0){d=1;return d|0}a=b+8|0;e=b+24|0;f=+p[e>>3]==999.0;g=c+16|0;h=+p[g>>3];do if(!(h==90.0)){c=b+16|0;i=+p[c>>3];if(f)p[e>>3]=i<h?180.0:0.0;j=+og(i);i=+pg(+p[c>>3]);l=+og(+p[e>>3]);m=+pg(+p[e>>3]);n=+og(+p[g>>3]);o=+pg(+p[g>>3]);q=l*n;l=+X(+(q*q+o*o));do if(l==0.0)if(i!=0.0){d=1;return d|0}else{r=+p[b+32>>3];break}else{s=i/l;if(+W(+s)>1.0){d=1;return d|0}t=+ug(o,q);u=+rg(s);s=t+u;if(!(s>180.0))if(s<-180.0)v=s+360.0;else v=s;else v=s+-360.0;s=t-u;if(!(s>180.0))if(s<-180.0)w=s+360.0;else w=s;else w=s+-360.0;c=b+32|0;s=+p[c>>3];if(+W(+(s-v))<+W(+(s-w))){x=+W(+v)<90.0000000001;y=x?v:w}else{x=+W(+w)<90.0000000001;y=x?w:v}p[c>>3]=y;r=y}while(0);c=b+40|0;x=b+48|0;p[x>>3]=90.0-r;q=j*+og(r);do if(+W(+q)<1.0e-10){if(+W(+j)<1.0e-10){l=+p[a>>3];p[c>>3]=l;p[x>>3]=90.0-+p[g>>3];z=a;A=c;B=l;break}if(r>0.0){l=+p[a>>3]+ +p[e>>3]+-180.0;p[c>>3]=l;p[x>>3]=0.0;z=a;A=c;B=l;break}if(r<0.0){l=+p[a>>3]-+p[e>>3];p[c>>3]=l;p[x>>3]=180.0;z=a;A=c;B=l;break}else{z=a;A=c;B=+p[c>>3];break}}else{l=(o-i*+pg(r))/q;s=m*n/j;if(l==0.0&s==0.0){d=1;return d|0}else{u=+p[a>>3];t=u-+ug(s,l);p[c>>3]=t;z=a;A=c;B=t;break}}while(0);if(!(+p[z>>3]>=0.0)){if(!(B>0.0)){C=x;D=r;break}p[A>>3]=B+-360.0;C=x;D=r;break}else{if(!(B<0.0)){C=x;D=r;break}p[A>>3]=B+360.0;C=x;D=r;break}}else{if(f)p[e>>3]=180.0;j=+p[b+16>>3];p[b+32>>3]=j;p[b+40>>3]=+p[a>>3];c=b+48|0;p[c>>3]=90.0-j;C=c;D=j}while(0);p[b+56>>3]=+p[e>>3];r=+og(+p[C>>3]);p[b+64>>3]=r;r=+pg(+p[C>>3]);p[b+72>>3]=r;k[b>>2]=137;b=+W(+D)>90.0000000001;d=b?2:0;return d|0}function cf(a,b,c,d,e,f,g,h,i){a=a|0;b=+b;c=+c;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0;if((k[d>>2]|0)!=137?(bf(a,d,g)|0)!=0:0){j=1;return j|0}mg(b,c,d+40|0,e,f)|0;d=Qb[k[g+1888>>2]&63](+p[e>>3],+p[f>>3],g,h,i)|0;if(!d){j=0;return j|0}j=(d|0)==1?2:3;return j|0}function df(a,b,c,d,e,f,g,h,i){a=a|0;b=+b;c=+c;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0;if((k[g>>2]|0)!=137?(bf(a,g,d)|0)!=0:0){j=1;return j|0}a=Qb[k[d+1892>>2]&63](b,c,d,e,f)|0;if(!a){ng(+p[e>>3],+p[f>>3],g+40|0,h,i)|0;j=0;return j|0}else{j=(a|0)==1?2:3;return j|0}return 0}function ef(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0.0,P=0.0;if(!(tx(a,606490)|0)){ff(b)|0;c=0;return c|0}if(!(tx(a,606494)|0)){gf(b)|0;c=0;return c|0}if(!(tx(a,606498)|0)){i[b>>0]=84;i[b+1>>0]=65;i[b+2>>0]=78;i[b+3>>0]=0;d=b+4|0;k[d>>2]=(k[d>>2]>>31&-206)+103;p[b+8>>3]=0.0;p[b+16>>3]=90.0;d=b+24|0;if(+p[d>>3]==0.0)p[d>>3]=57.29577951308232;k[b+1888>>2]=1;k[b+1892>>2]=2;d=99;while(1){if(!(+p[b+280+(d<<3)>>3]==0.0)){e=d;break}if(!(+p[b+280+(d+100<<3)>>3]==0.0)){e=d;break}f=d+-1|0;if((d|0)>0)d=f;else{e=f;break}}k[b+276>>2]=(e|0)<0?0:e;c=0;return c|0}if(!(tx(a,606502)|0)){i[b>>0]=83;i[b+1>>0]=84;i[b+2>>0]=71;i[b+3>>0]=0;k[b+4>>2]=104;p[b+8>>3]=0.0;p[b+16>>3]=90.0;e=b+24|0;g=+p[e>>3];if(g==0.0){p[e>>3]=57.29577951308232;h=.008726646259971648;j=114.59155902616465}else{l=g*2.0;h=1.0/l;j=l}p[b+112>>3]=j;p[b+120>>3]=h;k[b+1888>>2]=3;k[b+1892>>2]=4;c=0;return c|0}if(!(tx(a,606506)|0)){i[b>>0]=83;i[b+1>>0]=73;i[b+2>>0]=78;i[b+3>>0]=0;e=b+4|0;k[e>>2]=(k[e>>2]>>31&-210)+105;p[b+8>>3]=0.0;p[b+16>>3]=90.0;e=b+24|0;h=+p[e>>3];if(h==0.0){p[e>>3]=57.29577951308232;m=57.29577951308232}else m=h;p[b+112>>3]=1.0/m;m=+p[b+40>>3];h=+p[b+48>>3];j=m*m+h*h;p[b+120>>3]=j;p[b+128>>3]=j+1.0;p[b+136>>3]=j+-1.0;k[b+1888>>2]=5;k[b+1892>>2]=6;c=0;return c|0}if(!(tx(a,606510)|0)){i[b>>0]=65;i[b+1>>0]=82;i[b+2>>0]=67;i[b+3>>0]=0;k[b+4>>2]=106;p[b+8>>3]=0.0;p[b+16>>3]=90.0;e=b+24|0;j=+p[e>>3];if(j==0.0){p[e>>3]=57.29577951308232;n=1.0;o=1.0}else{h=j*3.141592653589793/180.0;n=1.0/h;o=h}p[b+112>>3]=o;p[b+120>>3]=n;k[b+1888>>2]=7;k[b+1892>>2]=8;c=0;return c|0}if(!(tx(a,606514)|0)){hf(b)|0;c=0;return c|0}if(!(tx(a,606518)|0)){i[b>>0]=90;i[b+1>>0]=69;i[b+2>>0]=65;i[b+3>>0]=0;k[b+4>>2]=108;p[b+8>>3]=0.0;p[b+16>>3]=90.0;e=b+24|0;n=+p[e>>3];if(n==0.0){p[e>>3]=57.29577951308232;q=.008726646259971648;r=114.59155902616465}else{o=n*2.0;q=1.0/o;r=o}p[b+112>>3]=r;p[b+120>>3]=q;k[b+1888>>2]=9;k[b+1892>>2]=10;c=0;return c|0}if(!(tx(a,606522)|0)){i[b>>0]=65;i[b+1>>0]=73;i[b+2>>0]=82;i[b+3>>0]=0;k[b+4>>2]=109;p[b+8>>3]=0.0;p[b+16>>3]=90.0;e=b+24|0;q=+p[e>>3];if(q==0.0){p[e>>3]=57.29577951308232;s=57.29577951308232}else s=q;q=s*2.0;e=b+112|0;p[e>>3]=q;s=+p[b+40>>3];do if(!(s==90.0))if(s>-90.0){r=+og((90.0-s)*.5);o=r*r;n=+fa(+r)*o/(1.0-o);p[b+120>>3]=n;o=.5-n;p[b+128>>3]=o;t=o;u=+p[e>>3];break}else{c=0;return c|0}else{p[b+120>>3]=-.5;p[b+128>>3]=1.0;t=1.0;u=q}while(0);p[b+136>>3]=t*u;p[b+144>>3]=.0001;p[b+152>>3]=t*.0001;p[b+160>>3]=57.29577951308232/t;k[b+1888>>2]=11;k[b+1892>>2]=12;c=0;return c|0}if(!(tx(a,606526)|0)){i[b>>0]=67;i[b+1>>0]=89;i[b+2>>0]=80;i[b+3>>0]=0;k[b+4>>2]=201;e=b+8|0;d=b+24|0;k[e>>2]=0;k[e+4>>2]=0;k[e+8>>2]=0;k[e+12>>2]=0;t=+p[d>>3];do if(t==0.0){p[d>>3]=57.29577951308232;u=+p[b+48>>3];p[b+112>>3]=u;if(u==0.0){c=0;return c|0}p[b+120>>3]=1.0/u;q=(u+ +p[b+40>>3])*57.29577951308232;p[b+128>>3]=q;if(q==0.0){c=0;return c|0}else{p[b+136>>3]=1.0/q;break}}else{q=+p[b+48>>3];u=t*q*3.141592653589793/180.0;p[b+112>>3]=u;if(u==0.0){c=0;return c|0}p[b+120>>3]=1.0/u;u=t*(q+ +p[b+40>>3]);p[b+128>>3]=u;if(u==0.0){c=0;return c|0}else{p[b+136>>3]=1.0/u;break}}while(0);k[b+1888>>2]=13;k[b+1892>>2]=14;c=0;return c|0}if(!(tx(a,606530)|0)){i[b>>0]=67;i[b+1>>0]=69;i[b+2>>0]=65;i[b+3>>0]=0;k[b+4>>2]=202;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;t=+p[e>>3];do if(t==0.0){p[e>>3]=57.29577951308232;p[b+112>>3]=1.0;p[b+120>>3]=1.0;u=+p[b+40>>3];if(u<=0.0|u>1.0){c=0;return c|0}else{p[b+128>>3]=57.29577951308232/u;p[b+136>>3]=u/57.29577951308232;break}}else{p[b+112>>3]=t*3.141592653589793/180.0;p[b+120>>3]=57.29577951308232/t;u=+p[b+40>>3];if(u<=0.0|u>1.0){c=0;return c|0}else{p[b+128>>3]=t/u;p[b+136>>3]=u/t;break}}while(0);k[b+1888>>2]=15;k[b+1892>>2]=16;c=0;return c|0}if(!(tx(a,606534)|0)){i[b>>0]=67;i[b+1>>0]=65;i[b+2>>0]=82;i[b+3>>0]=0;k[b+4>>2]=203;e=b+8|0;d=b+24|0;k[e>>2]=0;k[e+4>>2]=0;k[e+8>>2]=0;k[e+12>>2]=0;t=+p[d>>3];if(t==0.0){p[d>>3]=57.29577951308232;v=1.0;w=1.0}else{u=t*3.141592653589793/180.0;v=1.0/u;w=u}p[b+112>>3]=w;p[b+120>>3]=v;k[b+1888>>2]=17;k[b+1892>>2]=18;c=0;return c|0}if(!(tx(a,606538)|0)){i[b>>0]=77;i[b+1>>0]=69;i[b+2>>0]=82;i[b+3>>0]=0;k[b+4>>2]=204;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;v=+p[e>>3];if(v==0.0){p[e>>3]=57.29577951308232;x=1.0;y=1.0}else{w=v*3.141592653589793/180.0;x=1.0/w;y=w}p[b+112>>3]=y;p[b+120>>3]=x;k[b+1888>>2]=19;k[b+1892>>2]=20;c=0;return c|0}if(!(tx(a,606542)|0)){i[b>>0]=83;i[b+1>>0]=70;i[b+2>>0]=76;i[b+3>>0]=0;k[b+4>>2]=301;e=b+8|0;d=b+24|0;k[e>>2]=0;k[e+4>>2]=0;k[e+8>>2]=0;k[e+12>>2]=0;x=+p[d>>3];if(x==0.0){p[d>>3]=57.29577951308232;z=1.0;A=1.0}else{y=x*3.141592653589793/180.0;z=1.0/y;A=y}p[b+112>>3]=A;p[b+120>>3]=z;k[b+1888>>2]=21;k[b+1892>>2]=22;c=0;return c|0}if(!(tx(a,606546)|0)){i[b>>0]=80;i[b+1>>0]=65;i[b+2>>0]=82;i[b+3>>0]=0;k[b+4>>2]=302;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;z=+p[e>>3];if(z==0.0){p[e>>3]=57.29577951308232;B=.005555555555555556;C=180.0;D=1.0;E=1.0}else{A=z*3.141592653589793;z=A/180.0;B=1.0/A;C=A;D=1.0/z;E=z}p[b+112>>3]=E;p[b+120>>3]=D;p[b+128>>3]=C;p[b+136>>3]=B;k[b+1888>>2]=23;k[b+1892>>2]=24;c=0;return c|0}if(!(tx(a,606550)|0)){i[b>>0]=77;i[b+1>>0]=79;i[b+2>>0]=76;i[b+3>>0]=0;k[b+4>>2]=303;e=b+8|0;d=b+24|0;k[e>>2]=0;k[e+4>>2]=0;k[e+8>>2]=0;k[e+12>>2]=0;B=+p[d>>3];if(B==0.0){p[d>>3]=57.29577951308232;F=57.29577951308232}else F=B;B=F*1.4142135623730951;p[b+112>>3]=B;p[b+120>>3]=B/90.0;p[b+128>>3]=1.0/B;p[b+136>>3]=90.0/F;p[b+144>>3]=.6366197723675814;k[b+1888>>2]=25;k[b+1892>>2]=26;c=0;return c|0}if(!(tx(a,606554)|0)){i[b>>0]=65;i[b+1>>0]=73;i[b+2>>0]=84;i[b+3>>0]=0;k[b+4>>2]=401;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;F=+p[e>>3];if(F==0.0){p[e>>3]=57.29577951308232;G=57.29577951308232}else G=F;F=G*2.0;B=G*F;p[b+112>>3]=B;G=1.0/(B*2.0);p[b+120>>3]=G;p[b+128>>3]=G*.25;p[b+136>>3]=1.0/F;k[b+1888>>2]=27;k[b+1892>>2]=28;c=0;return c|0}if(!(tx(a,606558)|0)){i[b>>0]=67;i[b+1>>0]=79;i[b+2>>0]=80;i[b+3>>0]=0;e=b+4|0;k[e>>2]=(k[e>>2]>>31&-1002)+501;p[b+8>>3]=0.0;e=b+40|0;F=+p[e>>3];p[b+16>>3]=F;d=b+24|0;if(+p[d>>3]==0.0)p[d>>3]=57.29577951308232;G=+pg(F);p[b+112>>3]=G;if(G==0.0){c=0;return c|0}p[b+120>>3]=1.0/G;G=+p[d>>3];F=G*+og(+p[b+48>>3]);d=b+136|0;p[d>>3]=F;if(F==0.0){c=0;return c|0}p[b+144>>3]=1.0/F;F=1.0/+qg(+p[e>>3]);p[b+152>>3]=F;p[b+128>>3]=F*+p[d>>3];k[b+1888>>2]=29;k[b+1892>>2]=30;c=0;return c|0}if(!(tx(a,606562)|0)){jf(b)|0;c=0;return c|0}if(!(tx(a,606566)|0)){kf(b)|0;c=0;return c|0}if(!(tx(a,606570)|0)){lf(b)|0;c=0;return c|0}if(!(tx(a,606574)|0)){mf(b)|0;c=0;return c|0}if(!(tx(a,606578)|0)){i[b>>0]=80;i[b+1>>0]=67;i[b+2>>0]=79;i[b+3>>0]=0;k[b+4>>2]=602;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;F=+p[e>>3];if(F==0.0){p[e>>3]=57.29577951308232;H=114.59155902616465;I=1.0;J=1.0}else{G=F*3.141592653589793/180.0;H=F*2.0;I=1.0/G;J=G}p[b+112>>3]=J;p[b+120>>3]=I;p[b+128>>3]=H;k[b+1888>>2]=31;k[b+1892>>2]=32;c=0;return c|0}if(!(tx(a,606582)|0)){i[b>>0]=84;i[b+1>>0]=83;i[b+2>>0]=67;i[b+3>>0]=0;k[b+4>>2]=701;e=b+8|0;d=b+24|0;k[e>>2]=0;k[e+4>>2]=0;k[e+8>>2]=0;k[e+12>>2]=0;H=+p[d>>3];if(H==0.0){p[d>>3]=57.29577951308232;K=.022222222222222223;L=45.0}else{I=H*3.141592653589793*.25;K=1.0/I;L=I}p[b+112>>3]=L;p[b+120>>3]=K;k[b+1888>>2]=33;k[b+1892>>2]=34;c=0;return c|0}if(!(tx(a,606586)|0)){i[b>>0]=67;i[b+1>>0]=83;i[b+2>>0]=67;i[b+3>>0]=0;k[b+4>>2]=702;d=b+8|0;e=b+24|0;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;K=+p[e>>3];if(K==0.0){p[e>>3]=57.29577951308232;M=.022222222222222223;N=45.0}else{L=K*3.141592653589793*.25;M=1.0/L;N=L}p[b+112>>3]=N;p[b+120>>3]=M;k[b+1888>>2]=35;k[b+1892>>2]=36;c=0;return c|0}if(tx(a,606590)|0){c=1;return c|0}i[b>>0]=81;i[b+1>>0]=83;i[b+2>>0]=67;i[b+3>>0]=0;k[b+4>>2]=703;a=b+8|0;e=b+24|0;k[a>>2]=0;k[a+4>>2]=0;k[a+8>>2]=0;k[a+12>>2]=0;M=+p[e>>3];if(M==0.0){p[e>>3]=57.29577951308232;O=.022222222222222223;P=45.0}else{N=M*3.141592653589793*.25;O=1.0/N;P=N}p[b+112>>3]=P;p[b+120>>3]=O;k[b+1888>>2]=37;k[b+1892>>2]=38;c=0;return c|0}function ff(a){a=a|0;var b=0,c=0.0,d=0.0,e=0,f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0;i[a>>0]=65;i[a+1>>0]=90;i[a+2>>0]=80;i[a+3>>0]=0;b=a+4|0;k[b>>2]=(k[b>>2]>>31&-202)+101;p[a+8>>3]=0.0;p[a+16>>3]=90.0;b=a+24|0;c=+p[b>>3];if(c==0.0){p[b>>3]=57.29577951308232;d=57.29577951308232}else d=c;b=a+40|0;c=d*(+p[b>>3]+1.0);p[a+112>>3]=c;if(c==0.0){e=1;return e|0}f=a+48|0;c=+og(+p[f>>3]);g=a+136|0;p[g>>3]=c;if(c==0.0){e=1;return e|0}p[a+128>>3]=1.0/c;c=+pg(+p[f>>3]);p[a+144>>3]=c;d=+p[g>>3];p[a+120>>3]=c/d;c=+p[b>>3];if(+W(+c)>1.0){h=+sg(-1.0/c);j=h;l=+p[b>>3];m=+p[g>>3]}else{j=-90.0;l=c;m=d}p[a+152>>3]=j;j=l*m;p[a+160>>3]=j;g=+W(+j)<1.0;p[a+168>>3]=g?1.0:0.0;k[a+1888>>2]=39;k[a+1892>>2]=40;e=0;return e|0}function gf(a){a=a|0;var b=0,c=0.0,d=0.0,e=0,f=0,g=0,h=0,j=0,l=0,m=0.0;i[a>>0]=83;i[a+1>>0]=90;i[a+2>>0]=80;i[a+3>>0]=0;b=a+4|0;k[b>>2]=(k[b>>2]>>31&-204)+102;p[a+8>>3]=0.0;p[a+16>>3]=90.0;b=a+24|0;c=+p[b>>3];if(c==0.0){p[b>>3]=57.29577951308232;d=57.29577951308232}else d=c;p[a+112>>3]=1.0/d;e=a+40|0;d=+p[e>>3];f=a+56|0;c=d*+pg(+p[f>>3])+1.0;g=a+136|0;p[g>>3]=c;if(c==0.0){h=1;return h|0}c=+p[e>>3];d=c*+og(+p[f>>3]);j=a+48|0;c=-(d*+pg(+p[j>>3]));l=a+120|0;p[l>>3]=c;c=+p[e>>3];d=c*+og(+p[f>>3]);c=d*+og(+p[j>>3]);p[a+128>>3]=c;d=+p[b>>3];p[a+144>>3]=d*+p[l>>3];p[a+152>>3]=d*c;c=+p[g>>3];p[a+160>>3]=d*c;d=c+-1.0;p[a+168>>3]=c*d+-1.0;if(+W(+d)<1.0)m=+sg(1.0-c);else m=-90.0;p[a+176>>3]=m;k[a+1888>>2]=41;k[a+1892>>2]=42;h=0;return h|0}function hf(a){a=a|0;var b=0,c=0,d=0,e=0,f=0.0,g=0.0,h=0.0,j=0.0,l=0,m=0,n=0.0,o=0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0;i[a>>0]=90;i[a+1>>0]=80;i[a+2>>0]=78;i[a+3>>0]=0;b=a+4|0;k[b>>2]=(k[b>>2]>>31&-214)+107;c=a+8|0;p[c>>3]=0.0;d=a+16|0;p[d>>3]=90.0;e=a+24|0;f=+p[e>>3];if(f==0.0){p[e>>3]=57.29577951308232;g=57.29577951308232}else g=f;if(+p[a+104>>3]==0.0)if(+p[a+96>>3]==0.0)if(+p[a+88>>3]==0.0)if(+p[a+80>>3]==0.0)if(+p[a+72>>3]==0.0)if(+p[a+64>>3]==0.0)if(+p[a+56>>3]==0.0)if(+p[a+48>>3]==0.0)if(+p[a+40>>3]==0.0)if(+p[a+32>>3]==0.0){i[a>>0]=65;i[a+1>>0]=82;i[a+2>>0]=67;i[a+3>>0]=0;k[b>>2]=106;p[c>>3]=0.0;p[d>>3]=90.0;if(g==0.0){p[e>>3]=57.29577951308232;h=1.0;j=1.0}else{f=g*3.141592653589793/180.0;h=1.0/f;j=f}p[a+112>>3]=j;p[a+120>>3]=h;k[a+1888>>2]=7;k[a+1892>>2]=8;l=0;return l|0}else m=0;else m=1;else m=2;else m=3;else m=4;else m=5;else m=6;else m=7;else m=8;else m=9;k[a+272>>2]=m;k[a+1888>>2]=43;k[a+1892>>2]=44;if(m>>>0<=2){l=0;return l|0}h=+p[a+40>>3];if(!(h<=0.0)){n=h;o=0;q=0.0}else{l=1;return l|0}while(1){h=+(o|0)*3.141592653589793/180.0;j=0.0;e=m;while(1){f=h*j+ +(e|0)*+p[a+32+(e<<3)>>3];if((e|0)>1){j=f;e=e+-1|0}else{r=f;break}}if(r<=0.0){s=r;t=h;u=n;v=o;w=q;break}e=o+1|0;if((e|0)<180){n=r;o=e;q=h}else{s=r;t=h;u=r;v=e;w=h;break}}a:do if((v|0)==180)x=3.141592653589793;else{r=u;q=s;o=1;n=w;j=t;while(1){f=n-r*(j-n)/(q-r);g=0.0;e=m;while(1){y=f*g+ +(e|0)*+p[a+32+(e<<3)>>3];if((e|0)>1){g=y;e=e+-1|0}else{z=y;break}}if(+W(+z)<1.0e-13){x=f;break a}e=z<0.0;o=o+1|0;if((o|0)>=11){x=f;break}else{r=e?r:z;q=e?z:q;n=e?n:f;j=e?f:j}}}while(0);v=m;z=0.0;while(1){t=x*z+ +p[a+32+(v<<3)>>3];if((v|0)>0){v=v+-1|0;z=t}else{A=t;break}}p[a+112>>3]=x;p[a+120>>3]=A;l=0;return l|0}function jf(a){a=a|0;var b=0,c=0.0,d=0,e=0.0,f=0.0,g=0.0,h=0,j=0,l=0;i[a>>0]=67;i[a+1>>0]=79;i[a+2>>0]=69;i[a+3>>0]=0;k[a+4>>2]=502;p[a+8>>3]=0.0;b=a+40|0;c=+p[b>>3];p[a+16>>3]=c;d=a+24|0;if(+p[d>>3]==0.0)p[d>>3]=57.29577951308232;e=+p[a+48>>3];f=c-e;g=c+e;e=+pg(f);c=(e+ +pg(g))*.5;h=a+112|0;p[h>>3]=c;if(c==0.0){j=1;return j|0}p[a+120>>3]=1.0/c;l=a+136|0;p[l>>3]=+p[d>>3]/c;c=+pg(f);f=c*+pg(g)+1.0;p[a+144>>3]=f;g=+p[h>>3]*2.0;p[a+152>>3]=g;c=+p[l>>3];p[a+160>>3]=c*c*f;p[a+168>>3]=1.0/(+p[d>>3]*2.0*c);e=c*+X(+(f+g));p[a+176>>3]=e;e=c*+X(+(f-g*+pg(+p[b>>3])));p[a+128>>3]=e;k[a+1888>>2]=45;k[a+1892>>2]=46;j=0;return j|0}function kf(a){a=a|0;var b=0,c=0.0,d=0,e=0.0,f=0.0,g=0,h=0,j=0.0,l=0;i[a>>0]=67;i[a+1>>0]=79;i[a+2>>0]=68;i[a+3>>0]=0;k[a+4>>2]=503;p[a+8>>3]=0.0;b=a+40|0;c=+p[b>>3];p[a+16>>3]=c;d=a+24|0;e=+p[d>>3];if(e==0.0){p[d>>3]=57.29577951308232;f=57.29577951308232}else f=e;g=a+48|0;h=+p[g>>3]==0.0;e=f*+pg(c);if(h)j=e*3.141592653589793/180.0;else{c=e*+pg(+p[g>>3]);j=c/+p[g>>3]}h=a+112|0;p[h>>3]=j;if(j==0.0){l=1;return l|0}p[a+120>>3]=1.0/j;j=+p[d>>3];c=j*+og(+p[g>>3]);j=c*+og(+p[b>>3]);c=j/+p[h>>3];p[a+128>>3]=c;p[a+136>>3]=c+ +p[b>>3];k[a+1888>>2]=47;k[a+1892>>2]=48;l=0;return l|0}function lf(a){a=a|0;var b=0,c=0.0,d=0,e=0.0,f=0.0,g=0.0,h=0.0,j=0,l=0;i[a>>0]=67;i[a+1>>0]=79;i[a+2>>0]=79;i[a+3>>0]=0;k[a+4>>2]=504;p[a+8>>3]=0.0;b=a+40|0;c=+p[b>>3];p[a+16>>3]=c;d=a+24|0;if(+p[d>>3]==0.0)p[d>>3]=57.29577951308232;e=+p[a+48>>3];f=c-e;g=c+e;e=+qg((90.0-f)*.5);c=+og(f);if(f==g)h=+pg(f);else{f=+qg((90.0-g)*.5);h=+fa(+(+og(g)/c))/+fa(+(f/e))}j=a+112|0;p[j>>3]=h;if(h==0.0){l=1;return l|0}p[a+120>>3]=1.0/h;f=+p[d>>3]*(c/h)/+Y(+e,+h);d=a+136|0;p[d>>3]=f;if(f==0.0){l=1;return l|0}h=+qg((90.0-+p[b>>3])*.5);e=f*+Y(+h,+(+p[j>>3]));p[a+128>>3]=e;p[a+144>>3]=1.0/+p[d>>3];k[a+1888>>2]=49;k[a+1892>>2]=50;l=0;return l|0}function mf(a){a=a|0;var b=0,c=0,d=0.0,e=0.0,f=0.0,g=0.0,h=0,j=0,l=0;i[a>>0]=66;i[a+1>>0]=79;i[a+2>>0]=78;i[a+3>>0]=0;k[a+4>>2]=601;b=a+8|0;c=a+24|0;k[b>>2]=0;k[b+4>>2]=0;k[b+8>>2]=0;k[b+12>>2]=0;d=+p[c>>3];if(d==0.0){p[c>>3]=57.29577951308232;p[a+120>>3]=1.0;c=a+40|0;e=+og(+p[c>>3])*57.29577951308232;f=e/+pg(+p[c>>3]);g=f+ +p[c>>3];h=a+128|0;p[h>>3]=g;j=a+1888|0;k[j>>2]=51;l=a+1892|0;k[l>>2]=52;return 0}else{p[a+120>>3]=d*3.141592653589793/180.0;c=a+40|0;f=+og(+p[c>>3]);e=f/+pg(+p[c>>3]);g=d*(e+ +p[c>>3]*3.141592653589793/180.0);h=a+128|0;p[h>>3]=g;j=a+1888|0;k[j>>2]=51;l=a+1892|0;k[l>>2]=52;return 0}return 0}function nf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0;f=c+4|0;g=k[f>>2]|0;if((((g|0)>-1?g:0-g|0)|0)!=101?(ff(c)|0)!=0:0){h=1;return h|0}i=+og(a);j=+og(b);l=i*+p[c+120>>3];g=c+40|0;m=+p[g>>3];n=j*l+(m+ +pg(b));if(n==0.0){h=2;return h|0}m=j*+p[c+112>>3]/n;n=m*+pg(a);p[d>>3]=n;p[e>>3]=-(i*m*+p[c+128>>3]);if((k[f>>2]|0)>0){if(+p[c+152>>3]>b){h=2;return h|0}if((+p[c+168>>3]>0.0?(m=+p[g>>3]/+X(+(l*l+1.0)),+W(+m)<=1.0):0)?(i=+tg(-l),l=+sg(m),m=i-l,n=i+l+180.0,l=m>90.0?m+-360.0:m,m=n>90.0?n+-360.0:n,(l>m?l:m)>b):0){h=2;return h|0}}h=0;return h|0}function of(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0;f=k[c+4>>2]|0;if((((f|0)>-1?f:0-f|0)|0)!=101?(ff(c)|0)!=0:0){g=1;return g|0}h=+p[c+136>>3]*b;i=+X(+(a*a+h*h));if(i==0.0){p[d>>3]=0.0;j=90.0}else{l=+ug(a,-h);p[d>>3]=l;l=i/(+p[c+112>>3]+ +p[c+144>>3]*b);b=l*+p[c+40>>3]/+X(+(l*l+1.0));i=+ug(1.0,l);if(+W(+b)>1.0){l=b<0.0?-90.0:90.0;if(+W(+l)>1.0000000000001){g=2;return g|0}else m=l}else m=+sg(b);b=i-m;l=i+m+180.0;m=b>90.0?b+-360.0:b;b=l>90.0?l+-360.0:l;j=m>b?m:b}p[e>>3]=j;g=0;return g|0}function pf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0,l=0.0,m=0.0,n=0;f=c+4|0;g=k[f>>2]|0;if((((g|0)>-1?g:0-g|0)|0)!=102?(gf(c)|0)!=0:0){h=1;return h|0}i=+og(a);j=+pg(a);a=+og(b);l=1.0-+pg(b);g=c+136|0;m=+p[g>>3]-l;if(m==0.0){h=2;return h|0}n=c+160|0;p[d>>3]=(j*(a*+p[n>>3])-l*+p[c+144>>3])/m;p[e>>3]=-(i*(a*+p[n>>3])+l*+p[c+152>>3])/m;if((k[f>>2]|0)>0){if(+p[c+176>>3]>b){h=2;return h|0}if((+W(+(+p[c+40>>3]))>1.0?(m=j*+p[c+120>>3]-i*+p[c+128>>3],i=1.0/+X(+(+p[c+168>>3]+m*m)),+W(+i)<=1.0):0)?(j=+ug(m,+p[g>>3]+-1.0),m=+sg(i),i=j-m,l=j+m+180.0,m=i>90.0?i+-360.0:i,i=l>90.0?l+-360.0:l,(m>i?m:i)>b):0){h=2;return h|0}}h=0;return h|0}function qf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0;f=k[c+4>>2]|0;if((((f|0)>-1?f:0-f|0)|0)!=102?(gf(c)|0)!=0:0){g=1;return g|0}h=+p[c+112>>3];i=h*a;a=h*b;b=i*i+a*a;h=+p[c+136>>3];j=(i-+p[c+120>>3])/h;l=(a-+p[c+128>>3])/h;h=i*j+a*l;do if(!(b<1.0e-10)){m=j*j+l*l;n=m+1.0;o=h-m;q=o*o-n*(m+(b-h-h)+-1.0);if(q<0.0){g=2;return g|0}m=+X(+q);q=(m-o)/n;r=(-o-m)/n;n=q>r?q:r;if(n>1.0)if(n+-1.0<1.0e-13)s=1.0;else s=q<r?q:r;else s=n;n=s<-1.0&s+1.0>-1.0e-13?-1.0:s;if(n>1.0|n<-1.0){g=2;return g|0}else{r=+sg(n);p[e>>3]=r;t=1.0-n;break}}else{n=90.0-+X(+(b/(h+1.0)))*57.29577951308232;p[e>>3]=n;t=b*.5}while(0);b=+ug(i-j*t,-(a-l*t));p[d>>3]=b;g=0;return g|0}function rf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0.0,o=0,q=0.0,s=0.0,t=0.0,u=0.0;f=r;r=r+16|0;g=f;h=c+4|0;j=k[h>>2]|0;if((((j|0)>-1?j:0-j|0)|0)!=103){i[c>>0]=84;i[c+1>>0]=65;i[c+2>>0]=78;i[c+3>>0]=0;k[h>>2]=(j>>31&-206)+103;p[c+8>>3]=0.0;p[c+16>>3]=90.0;j=c+24|0;if(+p[j>>3]==0.0)p[j>>3]=57.29577951308232;k[c+1888>>2]=1;k[c+1892>>2]=2;j=99;while(1){if(!(+p[c+280+(j<<3)>>3]==0.0)){l=j;break}if(!(+p[c+280+(j+100<<3)>>3]==0.0)){l=j;break}m=j+-1|0;if((j|0)>0)j=m;else{l=m;break}}k[c+276>>2]=(l|0)<0?0:l}n=+pg(b);if(n<=0.0){o=2;r=f;return o|0}q=+p[c+24>>3];s=q*+og(b)/n;b=s*+pg(a);p[g>>3]=b;q=-(s*+og(a));l=g+8|0;p[l>>3]=q;j=k[c+1880>>2]|0;if(!j)t=b;else t=+Re(j,g);p[d>>3]=t;d=k[c+1884>>2]|0;if(!d)u=+p[l>>3];else u=+Re(d,g);p[e>>3]=u;o=n<0.0&(k[h>>2]|0)>0?2:0;r=f;return o|0}function sf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0.0,q=0.0,s=0.0,t=0,u=0.0,v=0.0;f=r;r=r+16|0;g=f+8|0;h=f;j=c+4|0;l=k[j>>2]|0;if((((l|0)>-1?l:0-l|0)|0)==103)m=k[c+276>>2]|0;else{i[c>>0]=84;i[c+1>>0]=65;i[c+2>>0]=78;i[c+3>>0]=0;k[j>>2]=(l>>31&-206)+103;p[c+8>>3]=0.0;p[c+16>>3]=90.0;l=c+24|0;if(+p[l>>3]==0.0)p[l>>3]=57.29577951308232;k[c+1888>>2]=1;k[c+1892>>2]=2;l=99;while(1){if(!(+p[c+280+(l<<3)>>3]==0.0)){n=l;break}if(!(+p[c+280+(l+100<<3)>>3]==0.0)){n=l;break}j=l+-1|0;if((l|0)>0)l=j;else{n=j;break}}l=(n|0)<0?0:n;k[c+276>>2]=l;m=l}if(!m){p[g>>3]=a;p[h>>3]=b;o=a;q=b}else{tf(c,a,b,g,h)|0;o=+p[g>>3];q=+p[h>>3]}b=+X(+(o*o+q*q));if(b==0.0){s=0.0;p[d>>3]=s;t=c+24|0;u=+p[t>>3];v=+ug(u,b);p[e>>3]=v;r=f;return 0}s=+ug(o,-q);p[d>>3]=s;t=c+24|0;u=+p[t>>3];v=+ug(u,b);p[e>>3]=v;r=f;return 0}function tf(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0,M=0.0,N=0.0,O=0.0,P=0.0,Q=0.0,R=0.0,S=0.0,T=0.0,U=0.0,V=0.0,W=0.0,Y=0.0,Z=0.0,_=0.0,$=0.0,aa=0.0,ba=0.0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ha=0.0,ia=0.0,ja=0.0;f=k[a+276>>2]|0;g=+p[a+1080>>3]+ +p[a+1088>>3]*b;h=+p[a+280>>3]+ +p[a+288>>3]*c;do if((f|0)!=1){i=g+ +p[a+1096>>3]*c;j=h+ +p[a+296>>3]*b;if((f|0)!=2){l=b*b;m=c*c;n=+X(+(l+m));o=i+n*+p[a+1104>>3];q=j+n*+p[a+304>>3];if((f|0)!=3){r=o+l*+p[a+1112>>3];s=q+m*+p[a+312>>3];if((f|0)!=4){t=b*c;u=r+t*+p[a+1120>>3];v=s+t*+p[a+320>>3];if((f|0)!=5){t=u+m*+p[a+1128>>3];w=v+l*+p[a+328>>3];if((f|0)!=6){x=l*b;y=t+x*+p[a+1136>>3];z=m*c;A=w+z*+p[a+336>>3];if((f|0)!=7){B=y+l*+p[a+1144>>3]*c;C=A+m*+p[a+344>>3]*b;if((f|0)!=8){D=B+m*(+p[a+1152>>3]*b);E=C+l*(+p[a+352>>3]*c);if((f|0)!=9){F=D+z*+p[a+1160>>3];G=E+x*+p[a+360>>3];if((f|0)!=10){H=n*(n*n);I=F+H*+p[a+1168>>3];J=G+H*+p[a+368>>3];if((f|0)!=11){K=l*l;L=I+K*+p[a+1176>>3];M=m*m;N=J+M*+p[a+376>>3];if((f|0)!=12){O=L+x*+p[a+1184>>3]*c;P=N+z*+p[a+384>>3]*b;if((f|0)!=13){Q=O+m*(l*+p[a+1192>>3]);R=P+m*(l*+p[a+392>>3]);if((f|0)!=14){S=Q+z*(+p[a+1200>>3]*b);T=R+x*(+p[a+400>>3]*c);if((f|0)!=15){U=S+M*+p[a+1208>>3];V=T+K*+p[a+408>>3];if((f|0)!=16){W=K*b;Y=U+W*+p[a+1216>>3];Z=M*c;_=V+Z*+p[a+416>>3];if((f|0)!=17){$=Y+K*+p[a+1224>>3]*c;aa=_+M*+p[a+424>>3]*b;if((f|0)!=18){ba=$+m*(x*+p[a+1232>>3]);ca=aa+l*(z*+p[a+432>>3]);if((f|0)==19){da=ba;ea=ca}else{fa=ba+z*(l*+p[a+1240>>3]);ba=ca+x*(m*+p[a+440>>3]);if((f|0)==20){da=fa;ea=ba;break}ca=fa+M*(+p[a+1248>>3]*b);fa=ba+K*(+p[a+448>>3]*c);if((f|0)==21){da=ca;ea=fa;break}ba=ca+Z*+p[a+1256>>3];ca=fa+W*+p[a+456>>3];if((f|0)==22){da=ba;ea=ca;break}fa=n*(n*H);H=ba+fa*+p[a+1264>>3];ba=ca+fa*+p[a+464>>3];if((f|0)==23){da=H;ea=ba;break}ca=W*b;ga=H+ca*+p[a+1272>>3];H=Z*c;ha=ba+H*+p[a+472>>3];if((f|0)==24){da=ga;ea=ha;break}ba=ga+W*+p[a+1280>>3]*c;ga=ha+Z*+p[a+480>>3]*b;if((f|0)==25){da=ba;ea=ga;break}ha=ba+m*(K*+p[a+1288>>3]);ba=ga+l*(M*+p[a+488>>3]);if((f|0)==26){da=ha;ea=ba;break}ga=ha+z*(x*+p[a+1296>>3]);ha=ba+x*(z*+p[a+496>>3]);if((f|0)==27){da=ga;ea=ha;break}ba=ga+M*(l*+p[a+1304>>3]);ga=ha+K*(m*+p[a+504>>3]);if((f|0)==28){da=ba;ea=ga;break}ha=ba+Z*(+p[a+1312>>3]*b);ba=ga+W*(+p[a+512>>3]*c);if((f|0)==29){da=ha;ea=ba;break}ga=ha+H*+p[a+1320>>3];ha=ba+ca*+p[a+520>>3];if((f|0)==30){da=ga;ea=ha;break}ba=ca*b;ia=ga+ba*+p[a+1328>>3];ga=H*c;ja=ha+ga*+p[a+528>>3];if((f|0)==31){da=ia;ea=ja;break}ha=ia+ca*+p[a+1336>>3]*c;ia=ja+H*+p[a+536>>3]*b;if((f|0)==32){da=ha;ea=ia;break}ja=ha+m*(W*+p[a+1344>>3]);ha=ia+l*(Z*+p[a+544>>3]);if((f|0)==33){da=ja;ea=ha;break}ia=ja+z*(K*+p[a+1352>>3]);ja=ha+x*(M*+p[a+552>>3]);if((f|0)==34){da=ia;ea=ja;break}ha=ia+M*(x*+p[a+1360>>3]);x=ja+K*(z*+p[a+560>>3]);if((f|0)==35){da=ha;ea=x;break}z=ha+Z*(l*+p[a+1368>>3]);l=x+W*(m*+p[a+568>>3]);if((f|0)==36){da=z;ea=l;break}m=z+H*(+p[a+1376>>3]*b);H=l+ca*(+p[a+576>>3]*c);if((f|0)==37){da=m;ea=H;break}ca=m+ga*+p[a+1384>>3];ga=H+ba*+p[a+584>>3];if((f|0)==38){da=ca;ea=ga;break}ba=n*(n*fa);da=ca+ba*+p[a+1392>>3];ea=ga+ba*+p[a+592>>3]}}else{da=$;ea=aa}}else{da=Y;ea=_}}else{da=U;ea=V}}else{da=S;ea=T}}else{da=Q;ea=R}}else{da=O;ea=P}}else{da=L;ea=N}}else{da=I;ea=J}}else{da=F;ea=G}}else{da=D;ea=E}}else{da=B;ea=C}}else{da=y;ea=A}}else{da=t;ea=w}}else{da=u;ea=v}}else{da=r;ea=s}}else{da=o;ea=q}}else{da=i;ea=j}}else{da=g;ea=h}while(0);p[d>>3]=da;p[e>>3]=ea;return 0}function uf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0;f=c+4|0;if((k[f>>2]|0)!=104){i[c>>0]=83;i[c+1>>0]=84;i[c+2>>0]=71;i[c+3>>0]=0;k[f>>2]=104;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;g=+p[f>>3];if(g==0.0){p[f>>3]=57.29577951308232;h=.008726646259971648;j=114.59155902616465}else{l=g*2.0;h=1.0/l;j=l}p[c+112>>3]=j;p[c+120>>3]=h;k[c+1888>>2]=3;k[c+1892>>2]=4}h=+pg(b)+1.0;if(h==0.0){m=2;return m|0}j=+p[c+112>>3];l=j*+og(b)/h;h=l*+pg(a);p[d>>3]=h;h=-(l*+og(a));p[e>>3]=h;m=0;return m|0}function vf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0;f=c+4|0;if((k[f>>2]|0)!=104){i[c>>0]=83;i[c+1>>0]=84;i[c+2>>0]=71;i[c+3>>0]=0;k[f>>2]=104;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;g=+p[f>>3];if(g==0.0){p[f>>3]=57.29577951308232;h=.008726646259971648;j=114.59155902616465}else{l=g*2.0;h=1.0/l;j=l}p[c+112>>3]=j;p[c+120>>3]=h;k[c+1888>>2]=3;k[c+1892>>2]=4}h=+X(+(a*a+b*b));if(h==0.0)m=0.0;else m=+ug(a,-b);p[d>>3]=m;m=90.0-+tg(h*+p[c+120>>3])*2.0;p[e>>3]=m;return 0}function wf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0;f=c+4|0;g=k[f>>2]|0;if((((g|0)>-1?g:0-g|0)|0)!=105){i[c>>0]=83;i[c+1>>0]=73;i[c+2>>0]=78;i[c+3>>0]=0;k[f>>2]=(g>>31&-210)+105;p[c+8>>3]=0.0;p[c+16>>3]=90.0;g=c+24|0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=57.29577951308232}else j=h;p[c+112>>3]=1.0/j;j=+p[c+40>>3];h=+p[c+48>>3];l=j*j+h*h;p[c+120>>3]=l;p[c+128>>3]=l+1.0;p[c+136>>3]=l+-1.0;k[c+1888>>2]=5;k[c+1892>>2]=6}l=(90.0-+W(+b))*3.141592653589793/180.0;if(l<1.0e-05){h=l*l*.5;m=l;n=b>0.0?h:2.0-h}else{h=1.0-+pg(b);m=+og(b);n=h}h=+og(a);l=+pg(a);g=c+24|0;o=c+40|0;p[d>>3]=+p[g>>3]*(m*l+n*+p[o>>3]);d=c+48|0;p[e>>3]=-(+p[g>>3]*(m*h-n*+p[d>>3]));do if((k[f>>2]|0)>0)if(+p[c+120>>3]==0.0){if(b<0.0)q=2;else break;return q|0}else{if(b<-+tg(l*+p[o>>3]-h*+p[d>>3]))q=2;else break;return q|0}while(0);q=0;return q|0}function xf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0,s=0.0,t=0.0,u=0.0;f=c+4|0;g=k[f>>2]|0;if((((g|0)>-1?g:0-g|0)|0)==105){h=+p[c+112>>3];j=+p[c+120>>3]}else{i[c>>0]=83;i[c+1>>0]=73;i[c+2>>0]=78;i[c+3>>0]=0;k[f>>2]=(g>>31&-210)+105;p[c+8>>3]=0.0;p[c+16>>3]=90.0;g=c+24|0;l=+p[g>>3];if(l==0.0){p[g>>3]=57.29577951308232;m=57.29577951308232}else m=l;l=1.0/m;p[c+112>>3]=l;m=+p[c+40>>3];n=+p[c+48>>3];o=m*m+n*n;p[c+120>>3]=o;p[c+128>>3]=o+1.0;p[c+136>>3]=o+-1.0;k[c+1888>>2]=5;k[c+1892>>2]=6;h=l;j=o}o=h*a;a=h*b;b=o*o+a*a;if(j==0.0){if(b!=0.0)q=+ug(o,-a);else q=0.0;p[d>>3]=q;if(b<.5){q=+rg(+X(+b));p[e>>3]=q;r=0;return r|0}if(!(b<=1.0)){r=2;return r|0}q=+sg(+X(+(1.0-b)));p[e>>3]=q;r=0;return r|0}g=c+40|0;f=c+48|0;q=o*+p[g>>3]+a*+p[f>>3];do if(!(b<1.0e-10)){h=+p[c+128>>3];l=q-j;n=l*l-h*(b-q-q+ +p[c+136>>3]);if(n<0.0){r=2;return r|0}m=+X(+n);n=(m-l)/h;s=(-l-m)/h;h=n>s?n:s;if(h>1.0)if(h+-1.0<1.0e-13)t=1.0;else t=n<s?n:s;else t=h;h=t<-1.0&t+1.0>-1.0e-13?-1.0:t;if(h>1.0|h<-1.0){r=2;return r|0}else{s=+sg(h);p[e>>3]=s;u=1.0-h;break}}else{h=90.0-+X(+(b/(q+1.0)))*57.29577951308232;p[e>>3]=h;u=b*.5}while(0);b=u*+p[f>>3]-a;a=o-u*+p[g>>3];if(b==0.0&a==0.0){p[d>>3]=0.0;r=0;return r|0}else{u=+ug(a,b);p[d>>3]=u;r=0;return r|0}return 0}function yf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0;f=c+4|0;if((k[f>>2]|0)==106){g=+p[c+112>>3];h=90.0-b;j=h*g;l=+pg(a);m=j*l;p[d>>3]=m;n=+og(a);o=j*n;q=-o;p[e>>3]=q;return 0}i[c>>0]=65;i[c+1>>0]=82;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=106;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;r=+p[f>>3];if(r==0.0){p[f>>3]=57.29577951308232;s=1.0;t=1.0}else{u=r*3.141592653589793/180.0;s=1.0/u;t=u}p[c+112>>3]=t;p[c+120>>3]=s;k[c+1888>>2]=7;k[c+1892>>2]=8;g=t;h=90.0-b;j=h*g;l=+pg(a);m=j*l;p[d>>3]=m;n=+og(a);o=j*n;q=-o;p[e>>3]=q;return 0}function zf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0,o=0.0,q=0.0,r=0.0;f=c+4|0;if((k[f>>2]|0)!=106){i[c>>0]=65;i[c+1>>0]=82;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=106;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;g=+p[f>>3];if(g==0.0){p[f>>3]=57.29577951308232;h=1.0;j=1.0}else{l=g*3.141592653589793/180.0;h=1.0/l;j=l}p[c+112>>3]=j;p[c+120>>3]=h;k[c+1888>>2]=7;k[c+1892>>2]=8}h=+X(+(a*a+b*b));if(h==0.0){m=0.0;p[d>>3]=m;n=c+120|0;o=+p[n>>3];q=h*o;r=90.0-q;p[e>>3]=r;return 0}m=+ug(a,-b);p[d>>3]=m;n=c+120|0;o=+p[n>>3];q=h*o;r=90.0-q;p[e>>3]=r;return 0}function Af(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0.0,j=0.0;f=c+4|0;g=k[f>>2]|0;if((((g|0)>-1?g:0-g|0)|0)!=107?(hf(c)|0)!=0:0){h=1;return h|0}i=(90.0-b)*3.141592653589793/180.0;b=(i*(i*(i*(i*(i*(i*(i*(i*(i*(i*0.0+ +p[c+104>>3])+ +p[c+96>>3])+ +p[c+88>>3])+ +p[c+80>>3])+ +p[c+72>>3])+ +p[c+64>>3])+ +p[c+56>>3])+ +p[c+48>>3])+ +p[c+40>>3])+ +p[c+32>>3])*+p[c+24>>3];j=b*+pg(a);p[d>>3]=j;j=-(b*+og(a));p[e>>3]=j;if((k[f>>2]|0)>0?i>+p[c+112>>3]:0){h=2;return h|0}h=0;return h|0}function Bf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0;f=k[c+4>>2]|0;if((((f|0)>-1?f:0-f|0)|0)!=107?(hf(c)|0)!=0:0){g=1;return g|0}f=k[c+272>>2]|0;h=+X(+(a*a+b*b));i=h/+p[c+24>>3];if((f|0)<1){g=1;return g|0}a:do switch(f|0){case 1:{j=(i-+p[c+32>>3])/+p[c+40>>3];break}case 2:{h=+p[c+48>>3];l=+p[c+40>>3];m=l*l-h*4.0*(+p[c+32>>3]-i);if(m<0.0){g=2;return g|0}n=+X(+m);m=h*2.0;h=(n-l)/m;o=(-l-n)/m;m=h<o?h:o;n=m<-1.0e-13?(h>o?h:o):m;if(n<0.0){if(n<-1.0e-13)g=2;else{j=0.0;break a}return g|0}if(n>3.141592653589793)if(n>3.141592653589893){g=2;return g|0}else j=3.141592653589793;else j=n;break}default:{n=+p[c+32>>3];m=+p[c+112>>3];o=+p[c+120>>3];if(i<n){if(i<n+-1.0e-13)g=2;else{j=0.0;break a}return g|0}if(i>o){if(i>o+1.0e-13)g=2;else{j=m;break a}return g|0}else{q=0;r=n;s=o;t=0.0;u=m}while(1){m=(s-i)/(s-r);if(!(m<.1))if(m>.9)v=.9;else v=m;else v=.1;m=u-(u-t)*v;w=f;o=0.0;while(1){n=m*o+ +p[c+32+(w<<3)>>3];if((w|0)>0){w=w+-1|0;o=n}else{x=n;break}}if(x<i)if(i-x<1.0e-13){j=m;break a}else{y=x;z=s;A=m;B=u}else if(x-i<1.0e-13){j=m;break a}else{y=r;z=x;A=t;B=m}q=q+1|0;if(!(!(+W(+(B-A))<1.0e-13)&(q|0)<100)){j=m;break}else{r=y;s=z;t=A;u=B}}}}while(0);if(i==0.0)C=0.0;else C=+ug(a,-b);p[d>>3]=C;p[e>>3]=90.0-j*180.0/3.141592653589793;g=0;return g|0}function Cf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0;f=c+4|0;if((k[f>>2]|0)==108)g=+p[c+112>>3];else{i[c>>0]=90;i[c+1>>0]=69;i[c+2>>0]=65;i[c+3>>0]=0;k[f>>2]=108;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;h=+p[f>>3];if(h==0.0){p[f>>3]=57.29577951308232;j=.008726646259971648;l=114.59155902616465}else{m=h*2.0;j=1.0/m;l=m}p[c+112>>3]=l;p[c+120>>3]=j;k[c+1888>>2]=9;k[c+1892>>2]=10;g=l}l=g*+pg((90.0-b)*.5);b=l*+pg(a);p[d>>3]=b;b=-(l*+og(a));p[e>>3]=b;return 0}function Df(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0;f=c+4|0;if((k[f>>2]|0)!=108){i[c>>0]=90;i[c+1>>0]=69;i[c+2>>0]=65;i[c+3>>0]=0;k[f>>2]=108;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;g=+p[f>>3];if(g==0.0){p[f>>3]=57.29577951308232;h=.008726646259971648;j=114.59155902616465}else{l=g*2.0;h=1.0/l;j=l}p[c+112>>3]=j;p[c+120>>3]=h;k[c+1888>>2]=9;k[c+1892>>2]=10}h=+X(+(a*a+b*b));if(h==0.0)m=0.0;else m=+ug(a,-b);p[d>>3]=m;m=h*+p[c+120>>3];if(+W(+m)>1.0)if(+W(+(h-+p[c+112>>3]))<1.0e-12)n=-90.0;else{o=2;return o|0}else n=90.0-+sg(m)*2.0;p[e>>3]=n;o=0;return o|0}function Ef(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0,r=0.0;f=c+4|0;if((k[f>>2]|0)!=109){i[c>>0]=65;i[c+1>>0]=73;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=109;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;g=+p[f>>3];if(g==0.0){p[f>>3]=57.29577951308232;h=57.29577951308232}else h=g;g=h*2.0;f=c+112|0;p[f>>3]=g;h=+p[c+40>>3];do if(!(h==90.0))if(h>-90.0){j=+og((90.0-h)*.5);l=j*j;m=+fa(+j)*l/(1.0-l);p[c+120>>3]=m;l=.5-m;p[c+128>>3]=l;n=l;o=+p[f>>3];break}else{q=1;return q|0}else{p[c+120>>3]=-.5;p[c+128>>3]=1.0;n=1.0;o=g}while(0);p[c+136>>3]=n*o;p[c+144>>3]=.0001;p[c+152>>3]=n*.0001;p[c+160>>3]=57.29577951308232/n;k[c+1888>>2]=11;k[c+1892>>2]=12}do if(!(b==90.0)){if(!(b>-90.0)){q=2;return q|0}n=90.0-b;o=n*.017453292519943295*.5;if(o<+p[c+144>>3]){r=o*+p[c+136>>3];break}else{o=+og(n*.5);n=+X(+(1.0-o*o))/o;g=+fa(+o)/n;r=-(+p[c+112>>3]*(g+n*+p[c+120>>3]));break}}else r=0.0;while(0);b=r*+pg(a);p[d>>3]=b;b=-(r*+og(a));p[e>>3]=b;q=0;return q|0}function Ff(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0,s=0.0,t=0.0,u=0,v=0.0,w=0.0,x=0.0,y=0.0,z=0,A=0.0,B=0.0,C=0,D=0,E=0.0,F=0.0,G=0,H=0.0,I=0.0,J=0.0,K=0.0,L=0.0;f=c+4|0;if((k[f>>2]|0)==109)g=+p[c+112>>3];else{i[c>>0]=65;i[c+1>>0]=73;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=109;p[c+8>>3]=0.0;p[c+16>>3]=90.0;f=c+24|0;h=+p[f>>3];if(h==0.0){p[f>>3]=57.29577951308232;j=57.29577951308232}else j=h;h=j*2.0;f=c+112|0;p[f>>3]=h;j=+p[c+40>>3];do if(!(j==90.0))if(j>-90.0){l=+og((90.0-j)*.5);m=l*l;n=+fa(+l)*m/(1.0-m);p[c+120>>3]=n;m=.5-n;p[c+128>>3]=m;o=m;q=+p[f>>3];break}else{r=1;return r|0}else{p[c+120>>3]=-.5;p[c+128>>3]=1.0;o=1.0;q=h}while(0);p[c+136>>3]=o*q;p[c+144>>3]=.0001;p[c+152>>3]=o*.0001;p[c+160>>3]=57.29577951308232/o;k[c+1888>>2]=11;k[c+1892>>2]=12;g=q}q=+X(+(a*a+b*b))/g;if(q==0.0){s=0.0;t=0.0}else{do if(!(q<+p[c+152>>3])){g=+p[c+120>>3];f=0;o=0.0;h=1.0;while(1){j=h*.5;m=+X(+(1.0-j*j))/j;n=-(+fa(+j)/m+m*g);if(q<=n){u=0;v=o;w=n;x=h;y=j;break}z=f+1|0;if((z|0)<30){f=z;o=n;h=j}else{A=j;B=n;C=z;D=16;break}}if((D|0)==16)if((C|0)==30){r=2;return r|0}else{u=0;v=B;w=B;x=A;y=A}while(1){h=(w-q)/(w-v);if(!(h<.1))if(h>.9)E=.9;else E=h;else E=.1;h=y-(y-x)*E;o=+X(+(1.0-h*h))/h;n=+fa(+h)/o+o*g;o=-n;if(q>o)if(q+n<1.0e-12){F=h;G=u;break}else{H=o;I=w;J=h;K=y}else if(o-q<1.0e-12){F=h;G=u;break}else{H=v;I=o;J=x;K=h}f=u+1|0;if((f|0)<100){u=f;v=H;w=I;x=J;y=K}else{F=h;G=f;break}}if((G|0)==100){r=2;return r|0}else{L=+rg(F);break}}else L=q*+p[c+160>>3];while(0);s=+ug(a,-b);t=L}p[d>>3]=s;p[e>>3]=90.0-t*2.0;r=0;return r|0}function Gf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0;f=c+4|0;if((k[f>>2]|0)==201)g=+p[c+40>>3];else{i[c>>0]=67;i[c+1>>0]=89;i[c+2>>0]=80;i[c+3>>0]=0;k[f>>2]=201;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];do if(j==0.0){p[h>>3]=57.29577951308232;l=+p[c+48>>3];p[c+112>>3]=l;if(l==0.0){m=1;return m|0}p[c+120>>3]=1.0/l;n=+p[c+40>>3];o=(l+n)*57.29577951308232;p[c+128>>3]=o;if(o==0.0){m=1;return m|0}else{p[c+136>>3]=1.0/o;q=n;break}}else{n=+p[c+48>>3];o=j*n*3.141592653589793/180.0;p[c+112>>3]=o;if(o==0.0){m=1;return m|0}p[c+120>>3]=1.0/o;o=+p[c+40>>3];l=j*(n+o);p[c+128>>3]=l;if(l==0.0){m=1;return m|0}else{p[c+136>>3]=1.0/l;q=o;break}}while(0);k[c+1888>>2]=13;k[c+1892>>2]=14;g=q}q=g+ +og(b);if(q==0.0){m=2;return m|0}p[d>>3]=+p[c+112>>3]*a;a=+p[c+128>>3];g=a*+pg(b)/q;p[e>>3]=g;m=0;return m|0}function Hf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0;f=c+4|0;if((k[f>>2]|0)==201)g=+p[c+120>>3];else{i[c>>0]=67;i[c+1>>0]=89;i[c+2>>0]=80;i[c+3>>0]=0;k[f>>2]=201;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];do if(j==0.0){p[h>>3]=57.29577951308232;l=+p[c+48>>3];p[c+112>>3]=l;if(l==0.0){m=1;return m|0}n=1.0/l;p[c+120>>3]=n;o=(l+ +p[c+40>>3])*57.29577951308232;p[c+128>>3]=o;if(o==0.0){m=1;return m|0}else{p[c+136>>3]=1.0/o;q=n;break}}else{n=+p[c+48>>3];o=j*n*3.141592653589793/180.0;p[c+112>>3]=o;if(o==0.0){m=1;return m|0}l=1.0/o;p[c+120>>3]=l;o=j*(n+ +p[c+40>>3]);p[c+128>>3]=o;if(o==0.0){m=1;return m|0}else{p[c+136>>3]=1.0/o;q=l;break}}while(0);k[c+1888>>2]=13;k[c+1892>>2]=14;g=q}p[d>>3]=g*a;a=+p[c+136>>3]*b;b=+ug(a,1.0);g=b+ +sg(a*+p[c+40>>3]/+X(+(a*a+1.0)));p[e>>3]=g;m=0;return m|0}function If(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0,n=0.0,o=0.0;f=c+4|0;if((k[f>>2]|0)==202)g=+p[c+112>>3];else{i[c>>0]=67;i[c+1>>0]=69;i[c+2>>0]=65;i[c+3>>0]=0;k[f>>2]=202;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];do if(j==0.0){p[h>>3]=57.29577951308232;p[c+112>>3]=1.0;p[c+120>>3]=1.0;l=+p[c+40>>3];if(l<=0.0|l>1.0){m=1;return m|0}else{p[c+128>>3]=57.29577951308232/l;p[c+136>>3]=l/57.29577951308232;n=1.0;break}}else{l=j*3.141592653589793/180.0;p[c+112>>3]=l;p[c+120>>3]=57.29577951308232/j;o=+p[c+40>>3];if(o<=0.0|o>1.0){m=1;return m|0}else{p[c+128>>3]=j/o;p[c+136>>3]=o/j;n=l;break}}while(0);k[c+1888>>2]=15;k[c+1892>>2]=16;g=n}p[d>>3]=g*a;a=+p[c+128>>3];g=a*+pg(b);p[e>>3]=g;m=0;return m|0}function Jf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0;f=c+4|0;if((k[f>>2]|0)==202)g=+p[c+136>>3];else{i[c>>0]=67;i[c+1>>0]=69;i[c+2>>0]=65;i[c+3>>0]=0;k[f>>2]=202;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];do if(j==0.0){p[h>>3]=57.29577951308232;p[c+112>>3]=1.0;p[c+120>>3]=1.0;l=+p[c+40>>3];if(l<=0.0|l>1.0){m=1;return m|0}else{p[c+128>>3]=57.29577951308232/l;n=l/57.29577951308232;p[c+136>>3]=n;o=n;break}}else{p[c+112>>3]=j*3.141592653589793/180.0;p[c+120>>3]=57.29577951308232/j;n=+p[c+40>>3];if(n<=0.0|n>1.0){m=1;return m|0}else{p[c+128>>3]=j/n;l=n/j;p[c+136>>3]=l;o=l;break}}while(0);k[c+1888>>2]=15;k[c+1892>>2]=16;g=o}o=g*b;b=+W(+o);do if(b>1.0)if(b>1.0000000000001){m=2;return m|0}else{q=o<0.0?-1.0:1.0;break}else q=o;while(0);p[d>>3]=+p[c+120>>3]*a;a=+sg(q);p[e>>3]=a;m=0;return m|0}function Kf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0;f=c+4|0;if((k[f>>2]|0)==203){g=c+112|0;h=g;j=+p[g>>3];l=j*a;p[d>>3]=l;m=+p[h>>3];n=m*b;p[e>>3]=n;return 0}i[c>>0]=67;i[c+1>>0]=65;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=203;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;o=+p[g>>3];if(o==0.0){p[g>>3]=57.29577951308232;q=1.0;r=1.0}else{s=o*3.141592653589793/180.0;q=1.0/s;r=s}g=c+112|0;p[g>>3]=r;p[c+120>>3]=q;k[c+1888>>2]=17;k[c+1892>>2]=18;h=g;j=r;l=j*a;p[d>>3]=l;m=+p[h>>3];n=m*b;p[e>>3]=n;return 0}function Lf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0;f=c+4|0;if((k[f>>2]|0)==203){g=c+120|0;h=g;j=+p[g>>3];l=j*a;p[d>>3]=l;m=+p[h>>3];n=m*b;p[e>>3]=n;return 0}i[c>>0]=67;i[c+1>>0]=65;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=203;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;o=+p[g>>3];if(o==0.0){p[g>>3]=57.29577951308232;q=1.0;r=1.0}else{s=o*3.141592653589793/180.0;q=1.0/s;r=s}p[c+112>>3]=r;g=c+120|0;p[g>>3]=q;k[c+1888>>2]=17;k[c+1892>>2]=18;h=g;j=q;l=j*a;p[d>>3]=l;m=+p[h>>3];n=m*b;p[e>>3]=n;return 0}function Mf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0;f=c+4|0;if((k[f>>2]|0)!=204){i[c>>0]=77;i[c+1>>0]=69;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=204;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=1.0;l=1.0}else{m=h*3.141592653589793/180.0;j=1.0/m;l=m}p[c+112>>3]=l;p[c+120>>3]=j;k[c+1888>>2]=19;k[c+1892>>2]=20}if(b<=-90.0|b>=90.0){n=2;return n|0}p[d>>3]=+p[c+112>>3]*a;a=+p[c+24>>3];j=a*+fa(+(+qg((b+90.0)*.5)));p[e>>3]=j;n=0;return n|0}function Nf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0,l=0.0,m=0.0,n=0.0,o=0.0;f=c+4|0;if((k[f>>2]|0)==204){g=c+24|0;h=+p[c+120>>3]}else{i[c>>0]=77;i[c+1>>0]=69;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=204;f=c+8|0;j=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;l=+p[j>>3];if(l==0.0){p[j>>3]=57.29577951308232;m=1.0;n=1.0}else{o=l*3.141592653589793/180.0;m=1.0/o;n=o}p[c+112>>3]=n;p[c+120>>3]=m;k[c+1888>>2]=19;k[c+1892>>2]=20;g=j;h=m}p[d>>3]=h*a;a=+tg(+ea(+(b/+p[g>>3])))*2.0+-90.0;p[e>>3]=a;return 0}function Of(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0.0,u=0.0;f=c+4|0;if((k[f>>2]|0)==301){g=c+112|0;h=g;j=+p[g>>3];l=j*a;m=+og(b);n=l*m;p[d>>3]=n;o=+p[h>>3];q=o*b;p[e>>3]=q;return 0}i[c>>0]=83;i[c+1>>0]=70;i[c+2>>0]=76;i[c+3>>0]=0;k[f>>2]=301;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;r=+p[g>>3];if(r==0.0){p[g>>3]=57.29577951308232;s=1.0;t=1.0}else{u=r*3.141592653589793/180.0;s=1.0/u;t=u}g=c+112|0;p[g>>3]=t;p[c+120>>3]=s;k[c+1888>>2]=21;k[c+1892>>2]=22;h=g;j=t;l=j*a;m=+og(b);n=l*m;p[d>>3]=n;o=+p[h>>3];q=o*b;p[e>>3]=q;return 0}function Pf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0;f=c+4|0;if((k[f>>2]|0)==301)g=+p[c+24>>3];else{i[c>>0]=83;i[c+1>>0]=70;i[c+2>>0]=76;i[c+3>>0]=0;k[f>>2]=301;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=1.0;m=1.0;n=57.29577951308232}else{o=j*3.141592653589793/180.0;l=1.0/o;m=o;n=j}p[c+112>>3]=m;p[c+120>>3]=l;k[c+1888>>2]=21;k[c+1892>>2]=22;g=n}n=+Z(+(b/g));h=c+120|0;if(n==0.0){q=0.0;p[d>>3]=q;r=+p[h>>3];s=r*b;p[e>>3]=s;return 0}q=+p[h>>3]*a/n;p[d>>3]=q;r=+p[h>>3];s=r*b;p[e>>3]=s;return 0}function Qf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0;f=c+4|0;if((k[f>>2]|0)==302){g=c+128|0;h=c+112|0}else{i[c>>0]=80;i[c+1>>0]=65;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=302;f=c+8|0;j=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;l=+p[j>>3];if(l==0.0){p[j>>3]=57.29577951308232;m=.005555555555555556;n=180.0;o=1.0;q=1.0}else{r=l*3.141592653589793;l=r/180.0;m=1.0/r;n=r;o=1.0/l;q=l}j=c+112|0;p[j>>3]=q;p[c+120>>3]=o;f=c+128|0;p[f>>3]=n;p[c+136>>3]=m;k[c+1888>>2]=23;k[c+1892>>2]=24;g=f;h=j}m=+pg(b/3.0);p[d>>3]=+p[h>>3]*a*(1.0-m*(m*4.0));p[e>>3]=m*+p[g>>3];return 0}function Rf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0,s=0.0;f=c+4|0;if((k[f>>2]|0)==302)g=+p[c+136>>3];else{i[c>>0]=80;i[c+1>>0]=65;i[c+2>>0]=82;i[c+3>>0]=0;k[f>>2]=302;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=.005555555555555556;m=180.0;n=1.0;o=1.0}else{q=j*3.141592653589793;j=q/180.0;l=1.0/q;m=q;n=1.0/j;o=j}p[c+112>>3]=o;p[c+120>>3]=n;p[c+128>>3]=m;p[c+136>>3]=l;k[c+1888>>2]=23;k[c+1892>>2]=24;g=l}l=g*b;if(l>1.0|l<-1.0){r=2;return r|0}b=1.0-l*(l*4.0);if(b==0.0)if(a==0.0)s=0.0;else{r=2;return r|0}else s=+p[c+120>>3]*a/b;p[d>>3]=s;s=+sg(l)*3.0;p[e>>3]=s;r=0;return r|0}function Sf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0;f=c+4|0;if((k[f>>2]|0)!=303){i[c>>0]=77;i[c+1>>0]=79;i[c+2>>0]=76;i[c+3>>0]=0;k[f>>2]=303;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=57.29577951308232}else j=h;h=j*1.4142135623730951;p[c+112>>3]=h;p[c+120>>3]=h/90.0;p[c+128>>3]=1.0/h;p[c+136>>3]=90.0/j;p[c+144>>3]=.6366197723675814;k[c+1888>>2]=25;k[c+1892>>2]=26}if(+W(+b)==90.0){p[d>>3]=0.0;j=+W(+(+p[c+112>>3]));p[e>>3]=b<0.0?-j:j;return 0}if(b==0.0){p[d>>3]=+p[c+120>>3]*a;p[e>>3]=0.0;return 0}j=+pg(b)*3.141592653589793;g=0;b=j;h=-3.141592653589793;l=3.141592653589793;while(1){m=b-j+ +_(+b);if(m<0.0)if(m>-1.0e-13){n=b;break}else{o=b;q=l}else if(m<1.0e-13){n=b;break}else{o=h;q=b}m=(o+q)*.5;g=g+1|0;if((g|0)>=100){n=m;break}else{b=m;h=o;l=q}}q=n*.5;n=+p[c+120>>3]*a*+Z(+q);p[d>>3]=n;n=+p[c+112>>3]*+_(+q);p[e>>3]=n;return 0}function Tf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0,n=0.0,o=0.0,q=0.0,r=0.0;f=c+4|0;if((k[f>>2]|0)==303)g=+p[c+24>>3];else{i[c>>0]=77;i[c+1>>0]=79;i[c+2>>0]=76;i[c+3>>0]=0;k[f>>2]=303;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=57.29577951308232}else l=j;j=l*1.4142135623730951;p[c+112>>3]=j;p[c+120>>3]=j/90.0;p[c+128>>3]=1.0/j;p[c+136>>3]=90.0/l;p[c+144>>3]=.6366197723675814;k[c+1888>>2]=25;k[c+1892>>2]=26;g=l}l=b/g;g=2.0-l*l;if(g<=1.0e-12){if(g<-1.0e-12){m=2;return m|0}if(+W(+a)>1.0e-12){m=2;return m|0}else{n=0.0;o=0.0}}else{j=+X(+g);n=j;o=+p[c+136>>3]*a/j}p[d>>3]=o;o=+p[c+128>>3]*b;b=+W(+o);do if(b>1.0)if(b>1.000000000001){m=2;return m|0}else{q=(o<0.0?-1.0:1.0)+l*n/3.141592653589793;break}else{j=+ba(+o);q=j*+p[c+144>>3]+l*n/3.141592653589793}while(0);n=+W(+q);do if(n>1.0)if(n>1.000000000001){m=2;return m|0}else{r=q<0.0?-1.0:1.0;break}else r=q;while(0);q=+sg(r);p[e>>3]=q;m=0;return m|0}function Uf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0.0,m=0.0;f=c+4|0;if((k[f>>2]|0)==401)g=c+112|0;else{i[c>>0]=65;i[c+1>>0]=73;i[c+2>>0]=84;i[c+3>>0]=0;k[f>>2]=401;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=57.29577951308232}else l=j;j=l*2.0;m=l*j;h=c+112|0;p[h>>3]=m;l=1.0/(m*2.0);p[c+120>>3]=l;p[c+128>>3]=l*.25;p[c+136>>3]=1.0/j;k[c+1888>>2]=27;k[c+1892>>2]=28;g=h}j=+og(b);l=+p[g>>3];m=a*.5;a=+X(+(l/(j*+og(m)+1.0)));l=+pg(m)*(j*(a*2.0));p[d>>3]=l;l=a*+pg(b);p[e>>3]=l;return 0}function Vf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,j=0,l=0.0,m=0.0,n=0.0,o=0,q=0.0,r=0.0,s=0.0;f=c+4|0;if((k[f>>2]|0)==401){g=+p[c+128>>3];h=+p[c+120>>3]}else{i[c>>0]=65;i[c+1>>0]=73;i[c+2>>0]=84;i[c+3>>0]=0;k[f>>2]=401;f=c+8|0;j=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;l=+p[j>>3];if(l==0.0){p[j>>3]=57.29577951308232;m=57.29577951308232}else m=l;l=m*2.0;n=m*l;p[c+112>>3]=n;m=1.0/(n*2.0);p[c+120>>3]=m;n=m*.25;p[c+128>>3]=n;p[c+136>>3]=1.0/l;k[c+1888>>2]=27;k[c+1892>>2]=28;g=n;h=m}m=1.0-a*a*g-b*b*h;if(m<0.0)if(m<-1.0e-13){o=2;return o|0}else q=0.0;else q=m;m=+X(+q);q=m*b/+p[c+24>>3];b=+W(+q);do if(b>1.0)if(b>1.0000000000001){o=2;return o|0}else{r=q<0.0?-1.0:1.0;break}else r=q;while(0);q=m*(m*2.0)+-1.0;b=m*a*+p[c+136>>3];if(q==0.0&b==0.0)s=0.0;else s=+ug(b,q)*2.0;p[d>>3]=s;s=+sg(r);p[e>>3]=s;o=0;return o|0}function Wf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0,m=0.0,n=0,o=0,q=0.0,r=0.0;f=c+4|0;g=k[f>>2]|0;do if((((g|0)>-1?g:0-g|0)|0)!=501){i[c>>0]=67;i[c+1>>0]=79;i[c+2>>0]=80;i[c+3>>0]=0;k[f>>2]=(g>>31&-1002)+501;p[c+8>>3]=0.0;h=c+40|0;j=+p[h>>3];p[c+16>>3]=j;l=c+24|0;if(+p[l>>3]==0.0)p[l>>3]=57.29577951308232;m=+pg(j);p[c+112>>3]=m;if(m==0.0){n=1;return n|0}p[c+120>>3]=1.0/m;m=+p[l>>3];j=m*+og(+p[c+48>>3]);l=c+136|0;p[l>>3]=j;if(j==0.0){n=1;return n|0}else{p[c+144>>3]=1.0/j;j=1.0/+qg(+p[h>>3]);p[c+152>>3]=j;p[c+128>>3]=j*+p[l>>3];k[c+1888>>2]=29;k[c+1892>>2]=30;o=h;break}}else o=c+40|0;while(0);j=b-+p[o>>3];b=+og(j);if(b==0.0){n=2;return n|0}o=c+112|0;m=+p[o>>3]*a;g=c+128|0;a=+p[g>>3];q=+p[c+136>>3];r=a-q*+pg(j)/b;b=r*+pg(m);p[d>>3]=b;b=+p[g>>3];j=b-r*+og(m);p[e>>3]=j;if((k[f>>2]|0)>0?r*+p[o>>3]<0.0:0){n=2;return n|0}n=0;return n|0}function Xf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0,m=0.0,n=0,o=0,q=0.0,r=0.0;f=c+4|0;g=k[f>>2]|0;do if((((g|0)>-1?g:0-g|0)|0)!=501){i[c>>0]=67;i[c+1>>0]=79;i[c+2>>0]=80;i[c+3>>0]=0;k[f>>2]=(g>>31&-1002)+501;p[c+8>>3]=0.0;h=c+40|0;j=+p[h>>3];p[c+16>>3]=j;l=c+24|0;if(+p[l>>3]==0.0)p[l>>3]=57.29577951308232;m=+pg(j);p[c+112>>3]=m;if(m==0.0){n=1;return n|0}p[c+120>>3]=1.0/m;m=+p[l>>3];j=m*+og(+p[c+48>>3]);l=c+136|0;p[l>>3]=j;if(j==0.0){n=1;return n|0}else{p[c+144>>3]=1.0/j;j=1.0/+qg(+p[h>>3]);p[c+152>>3]=j;m=j*+p[l>>3];p[c+128>>3]=m;k[c+1888>>2]=29;k[c+1892>>2]=30;o=h;q=m;break}}else{o=c+40|0;q=+p[c+128>>3]}while(0);m=q-b;b=+X(+(a*a+m*m));q=+p[o>>3]<0.0?-b:b;if(q==0.0)r=0.0;else r=+ug(a/q,m/q);p[d>>3]=r*+p[c+120>>3];r=+p[o>>3];m=r+ +tg(+p[c+152>>3]-q*+p[c+144>>3]);p[e>>3]=m;n=0;return n|0}function Yf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0,j=0.0;if((k[c+4>>2]|0)!=502?(jf(c)|0)!=0:0){f=1;return f|0}g=+p[c+112>>3]*a;if(b==-90.0)h=+p[c+176>>3];else{a=+p[c+136>>3];i=+p[c+144>>3];j=+p[c+152>>3];h=a*+X(+(i-j*+pg(b)))}b=h*+pg(g);p[d>>3]=b;b=+p[c+128>>3];j=b-h*+og(g);p[e>>3]=j;f=0;return f|0}function Zf(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0;if((k[c+4>>2]|0)!=502?(jf(c)|0)!=0:0){f=1;return f|0}g=+p[c+128>>3]-b;b=+X(+(a*a+g*g));h=+p[c+40>>3]<0.0?-b:b;if(h==0.0)i=0.0;else i=+ug(a/h,g/h);p[d>>3]=i*+p[c+120>>3];if(+W(+(h-+p[c+176>>3]))<1.0e-12){p[e>>3]=-90.0;f=0;return f|0}i=(+p[c+160>>3]-h*h)*+p[c+168>>3];if(!(+W(+i)>1.0)){h=+sg(i);p[e>>3]=h;f=0;return f|0}if(+W(+(i+-1.0))<1.0e-12){p[e>>3]=90.0;f=0;return f|0}if(!(+W(+(i+1.0))<1.0e-12)){f=2;return f|0}p[e>>3]=-90.0;f=0;return f|0}function _f(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0;if((k[c+4>>2]|0)!=503?(kf(c)|0)!=0:0){f=1;return f|0}g=+p[c+112>>3]*a;a=+p[c+136>>3]-b;b=a*+pg(g);p[d>>3]=b;b=+p[c+128>>3];h=b-a*+og(g);p[e>>3]=h;f=0;return f|0}function $f(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0;if((k[c+4>>2]|0)!=503?(kf(c)|0)!=0:0){f=1;return f|0}g=+p[c+128>>3]-b;b=+X(+(a*a+g*g));h=+p[c+40>>3]<0.0?-b:b;if(h==0.0)i=0.0;else i=+ug(a/h,g/h);p[d>>3]=i*+p[c+120>>3];p[e>>3]=+p[c+136>>3]-h;f=0;return f|0}function ag(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0.0,j=0.0;if((k[c+4>>2]|0)!=504?(lf(c)|0)!=0:0){f=1;return f|0}g=c+112|0;h=+p[g>>3];i=h*a;if(b==-90.0)if(h<0.0)j=0.0;else{f=2;return f|0}else{h=+p[c+136>>3];a=+qg((90.0-b)*.5);j=h*+Y(+a,+(+p[g>>3]))}a=j*+pg(i);p[d>>3]=a;a=+p[c+128>>3];h=a-j*+og(i);p[e>>3]=h;f=0;return f|0}function bg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0.0,i=0.0,j=0;if((k[c+4>>2]|0)!=504?(lf(c)|0)!=0:0){f=1;return f|0}g=+p[c+128>>3]-b;b=+X(+(a*a+g*g));h=+p[c+40>>3]<0.0?-b:b;if(h==0.0){p[d>>3]=+p[c+120>>3]*0.0;if(+p[c+112>>3]<0.0)i=-90.0;else{f=2;return f|0}}else{b=+ug(a/h,g/h);j=c+120|0;p[d>>3]=b*+p[j>>3];i=90.0-+tg(+Y(+(h*+p[c+144>>3]),+(+p[j>>3])))*2.0}p[e>>3]=i;f=0;return f|0}function cg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0.0,r=0.0,s=0.0;f=c+4|0;g=k[f>>2]|0;if(!(+p[c+40>>3]==0.0)){if((g|0)!=601)mf(c)|0;h=c+128|0;j=+p[h>>3]-+p[c+120>>3]*b;l=+p[c+24>>3]*a;m=l*+og(b)/j;l=j*+pg(m);p[d>>3]=l;l=+p[h>>3];n=l-j*+og(m);p[e>>3]=n;return 0}if((g|0)==301){g=c+112|0;o=g;q=+p[g>>3]}else{i[c>>0]=83;i[c+1>>0]=70;i[c+2>>0]=76;i[c+3>>0]=0;k[f>>2]=301;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;m=+p[g>>3];if(m==0.0){p[g>>3]=57.29577951308232;r=1.0;s=1.0}else{j=m*3.141592653589793/180.0;r=1.0/j;s=j}g=c+112|0;p[g>>3]=s;p[c+120>>3]=r;k[c+1888>>2]=21;k[c+1892>>2]=22;o=g;q=s}s=q*a*+og(b);p[d>>3]=s;n=+p[o>>3]*b;p[e>>3]=n;return 0}function dg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0,t=0,u=0.0,v=0.0,w=0.0,x=0.0,y=0.0;f=c+40|0;g=+p[f>>3];h=c+4|0;j=k[h>>2]|0;if(!(g==0.0)){if((j|0)==601)l=g;else{mf(c)|0;l=+p[f>>3]}f=c+128|0;g=+p[f>>3];m=g-b;n=+X(+(a*a+m*m));o=l<0.0?-n:n;if(o==0.0){q=g;r=0.0}else{g=+ug(a/o,m/o);q=+p[f>>3];r=g}g=(q-o)/+p[c+120>>3];p[e>>3]=g;q=+og(g);if(q==0.0)s=0.0;else s=r*(o/+p[c+24>>3])/q;p[d>>3]=s;return 0}if((j|0)==301){t=c+120|0;u=+p[c+24>>3]}else{i[c>>0]=83;i[c+1>>0]=70;i[c+2>>0]=76;i[c+3>>0]=0;k[h>>2]=301;h=c+8|0;j=c+24|0;k[h>>2]=0;k[h+4>>2]=0;k[h+8>>2]=0;k[h+12>>2]=0;s=+p[j>>3];if(s==0.0){p[j>>3]=57.29577951308232;v=1.0;w=1.0;x=57.29577951308232}else{q=s*3.141592653589793/180.0;v=1.0/q;w=q;x=s}p[c+112>>3]=w;j=c+120|0;p[j>>3]=v;k[c+1888>>2]=21;k[c+1892>>2]=22;t=j;u=x}x=+Z(+(b/u));if(x==0.0)y=0.0;else y=+p[t>>3]*a/x;p[d>>3]=y;p[e>>3]=+p[t>>3]*b;return 0}function eg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0;f=c+4|0;if((k[f>>2]|0)!=602){i[c>>0]=80;i[c+1>>0]=67;i[c+2>>0]=79;i[c+3>>0]=0;k[f>>2]=602;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=114.59155902616465;l=1.0;m=1.0}else{n=h*3.141592653589793/180.0;j=h*2.0;l=1.0/n;m=n}p[c+112>>3]=m;p[c+120>>3]=l;p[c+128>>3]=j;k[c+1888>>2]=31;k[c+1892>>2]=32}j=+og(b);l=+pg(b);m=l*a;if(l==0.0){p[d>>3]=+p[c+112>>3]*a;o=0.0;p[e>>3]=o;return 0}else{a=j/l;g=c+24|0;l=a*+p[g>>3];j=l*+pg(m);p[d>>3]=j;j=+p[g>>3];o=j*(b*3.141592653589793/180.0+a*(1.0-+og(m)));p[e>>3]=o;return 0}return 0}function fg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0.0,q=0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0;f=c+4|0;if((k[f>>2]|0)==602)g=+p[c+120>>3];else{i[c>>0]=80;i[c+1>>0]=67;i[c+2>>0]=79;i[c+3>>0]=0;k[f>>2]=602;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=114.59155902616465;m=1.0;n=1.0}else{o=j*3.141592653589793/180.0;l=j*2.0;m=1.0/o;n=o}p[c+112>>3]=n;p[c+120>>3]=m;p[c+128>>3]=l;k[c+1888>>2]=31;k[c+1892>>2]=32;g=m}m=+W(+(g*b));if(m<1.0e-12){p[d>>3]=g*a;p[e>>3]=0.0;return 0}if(+W(+(m+-90.0))<1.0e-12){p[d>>3]=0.0;p[e>>3]=b<0.0?-90.0:90.0;return 0}m=b>0.0?90.0:-90.0;g=a*a;h=c+112|0;l=b-m*+p[h>>3];f=c+128|0;n=-999.0;o=g+l*l;q=0;l=0.0;j=m;while(1){if(n<-100.0)r=(l+j)*.5;else{m=o/(o-n);if(!(m<.1))if(m>.9)s=.9;else s=m;else s=.1;r=j-(j-l)*s}p[e>>3]=r;m=b-r*+p[h>>3];t=+qg(r);u=g+m*(m-+p[f>>3]/t);if(+W(+u)<1.0e-12){v=m;w=t;break}if(+W(+(j-l))<1.0e-12){v=m;w=t;break}x=u>0.0;y=+p[e>>3];q=q+1|0;if((q|0)>=64){v=m;w=t;break}else{n=x?n:u;o=x?u:o;l=x?l:y;j=x?y:j}}j=+p[c+24>>3]-w*v;v=w*a;if(v==0.0&j==0.0){p[d>>3]=0.0;return 0}else{a=+ug(v,j);j=a/+pg(+p[e>>3]);p[d>>3]=j;return 0}return 0}function gg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0,o=0,q=0.0,r=0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0,z=0.0;f=c+4|0;if((k[f>>2]|0)!=701){i[c>>0]=84;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=701;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=.022222222222222223;l=45.0}else{m=h*3.141592653589793*.25;j=1.0/m;l=m}p[c+112>>3]=l;p[c+120>>3]=j;k[c+1888>>2]=33;k[c+1892>>2]=34}j=+og(b);l=j*+og(a);m=j*+pg(a);a=+pg(b);g=l>a;b=g?l:a;f=m>b;j=f?m:b;b=-l;n=j<b;h=n?b:j;j=-m;o=h<j;q=o?j:h;h=-a;r=q<h;s=r?h:q;switch((r?5:o?4:n?3:f?2:g&1)|0){case 0:{t=0.0;u=m/s;v=2.0;w=b/s;break}case 1:{t=0.0;u=m/s;v=0.0;w=a/s;break}case 2:{t=2.0;u=b/s;v=0.0;w=a/s;break}case 3:{t=4.0;u=j/s;v=0.0;w=a/s;break}case 4:{t=6.0;u=l/s;v=0.0;w=a/s;break}case 5:{t=0.0;u=m/s;v=-2.0;w=l/s;break}default:{t=0.0;u=0.0;v=0.0;w=0.0}}s=+W(+u);do if(s>1.0)if(s>1.000000000001){x=2;return x|0}else{y=u<0.0?-1.0:1.0;break}else y=u;while(0);u=+W(+w);do if(u>1.0)if(u>1.000000000001){x=2;return x|0}else{z=w<0.0?-1.0:1.0;break}else z=w;while(0);g=c+112|0;p[d>>3]=(t+y)*+p[g>>3];p[e>>3]=(v+z)*+p[g>>3];x=0;return x|0}function hg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0.0,r=0.0,s=0.0,t=0.0;f=c+4|0;if((k[f>>2]|0)==701)g=+p[c+120>>3];else{i[c>>0]=84;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=701;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=.022222222222222223;m=45.0}else{n=j*3.141592653589793*.25;l=1.0/n;m=n}p[c+112>>3]=m;p[c+120>>3]=l;k[c+1888>>2]=33;k[c+1892>>2]=34;g=l}l=g*a;a=g*b;b=+W(+l);if(!(b<=1.0)){if(b>7.0){o=2;return o|0}if(+W(+a)>1.0){o=2;return o|0}}else if(+W(+a)>3.0){o=2;return o|0}b=l<-1.0?l+8.0:l;do if(!(b>5.0)){if(b>3.0){l=b+-4.0;g=-1.0/+X(+(a*a+(l*l+1.0)));q=g;r=l*g;s=-(a*g);break}if(b>1.0){g=b+-2.0;l=1.0/+X(+(a*a+(g*g+1.0)));q=-(g*l);r=l;s=a*l;break}if(a>1.0){l=a+-2.0;g=1.0/+X(+(l*l+(b*b+1.0)));q=-(l*g);r=b*g;s=g;break}if(a<-1.0){g=a+2.0;l=-1.0/+X(+(g*g+(b*b+1.0)));m=-l;q=g*m;r=b*m;s=l;break}else{l=1.0/+X(+(a*a+(b*b+1.0)));q=l;r=b*l;s=a*l;break}}else{l=b+-6.0;m=-1.0/+X(+(a*a+(l*l+1.0)));g=-m;q=l*g;r=m;s=a*g}while(0);if(q==0.0&r==0.0)t=0.0;else t=+ug(r,q);p[d>>3]=t;t=+sg(s);p[e>>3]=t;o=0;return o|0}function ig(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0,o=0,q=0.0,r=0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0,z=0.0;f=c+4|0;if((k[f>>2]|0)!=702){i[c>>0]=67;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=702;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=.022222222222222223;l=45.0}else{m=h*3.141592653589793*.25;j=1.0/m;l=m}p[c+112>>3]=l;p[c+120>>3]=j;k[c+1888>>2]=35;k[c+1892>>2]=36}j=+og(b);l=j*+og(a);m=j*+pg(a);a=+pg(b);g=l>a;b=g?l:a;f=m>b;j=f?m:b;b=-l;n=j<b;h=n?b:j;j=-m;o=h<j;q=o?j:h;h=-a;r=q<h;s=r?h:q;switch((r?5:o?4:n?3:f?2:g&1)|0){case 0:{t=b;u=0.0;v=m;w=2.0;break}case 1:{t=a;u=0.0;v=m;w=0.0;break}case 2:{t=a;u=2.0;v=b;w=0.0;break}case 3:{t=a;u=4.0;v=j;w=0.0;break}case 4:{t=a;u=6.0;v=l;w=0.0;break}default:{t=r?l:0.0;u=0.0;v=r?m:0.0;w=r?-2.0:0.0}}m=v/s;v=t/s;s=m*m;t=v*v;l=1.0-s;a=1.0-t;j=+W(+(m*v));b=s>1.0e-16?s*s:0.0;q=t>1.0e-16?t*t:0.0;h=j>1.0e-16?s*t*.15384112298488617:0.0;j=m*(s+l*(s*(-.15959623456001282-l*(.07591962069272995-s*.021776249632239342))+(t*(s*.004869491793215275-l*.1316167116165161+a*(q*.10695946961641312+(s*.08097013086080551+.14118963479995728-t*.2815285325050354+h-b*.1782512068748474)))+1.374848484992981)));m=v*(t+a*(t*(-.15959623456001282-a*(.07591962069272995-t*.021776249632239342))+(s*(t*.004869491793215275-a*.1316167116165161+l*(b*.10695946961641312+(t*.08097013086080551+.14118963479995728-s*.2815285325050354+h-q*.1782512068748474)))+1.374848484992981)));q=+W(+j);do if(q>1.0)if(q>1.0000001000000012){x=2;return x|0}else{y=j<0.0?-1.0:1.0;break}else y=j;while(0);j=+W(+m);do if(j>1.0)if(j>1.0000001000000012){x=2;return x|0}else{z=m<0.0?-1.0:1.0;break}else z=m;while(0);r=c+112|0;p[d>>3]=(u+y)*+p[r>>3];p[e>>3]=+p[r>>3]*(w+z);x=0;return x|0}function jg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0;f=c+4|0;if((k[f>>2]|0)==702)g=+p[c+120>>3];else{i[c>>0]=67;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=702;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=.022222222222222223;m=45.0}else{n=j*3.141592653589793*.25;l=1.0/n;m=n}p[c+112>>3]=m;p[c+120>>3]=l;k[c+1888>>2]=35;k[c+1892>>2]=36;g=l}l=g*a;a=g*b;b=+W(+l);if(!(b<=1.0)){if(b>7.0){o=2;return o|0}if(+W(+a)>1.0){o=2;return o|0}}else if(+W(+a)>3.0){o=2;return o|0}b=l<-1.0?l+8.0:l;do if(!(b>5.0)){if(b>3.0){q=3;r=b+-4.0;s=a;break}if(b>1.0){q=2;r=b+-2.0;s=a;break}if(a>1.0){q=0;r=b;s=a+-2.0;break}if(a<-1.0){q=5;r=b;s=a+2.0}else{q=1;r=b;s=a}}else{q=4;r=b+-6.0;s=a}while(0);a=r*r;b=s*s;l=r;r=l+l*(1.0-a)*(a*(a*(a*(a*(a*(a*.025843750685453415+.25795793533325195)+-.6293006539344788)+.5485238432884216)+-.22797055542469025)+-.07629968971014023)+-.2729269564151764+b*(a*(a*(a*(a*(1.715475082397461-a*.5302233695983887)+-1.7411445379257202)+.48051509261131287)+-.014715650118887424)+-.028194520622491837+b*(a*(a*(a*(.9893810153007507-a*.8318046927452087)+.30803316831588745)+-.5680093765258789)+.27058160305023193+b*(a*(a*(a*.08693841099739075+-.9367857575416565)+1.5088008642196655)+-.6044155955314636+b*(a*(a*.3388744592666626+-1.4160192012786865)+.934120774269104+b*(b*.14381584525108337+(a*.5203223824501038+-.6391530632972717)))))));l=s;s=l+l*(1.0-b)*(b*(b*(b*(b*(b*(b*.025843750685453415+.25795793533325195)+-.6293006539344788)+.5485238432884216)+-.22797055542469025)+-.07629968971014023)+-.2729269564151764+a*(b*(b*(b*(b*(1.715475082397461-b*.5302233695983887)+-1.7411445379257202)+.48051509261131287)+-.014715650118887424)+-.028194520622491837+a*(b*(b*(b*(.9893810153007507-b*.8318046927452087)+.30803316831588745)+-.5680093765258789)+.27058160305023193+a*(b*(b*(b*.08693841099739075+-.9367857575416565)+1.5088008642196655)+-.6044155955314636+a*(b*(b*.3388744592666626+-1.4160192012786865)+.934120774269104+a*(b*.5203223824501038+-.6391530632972717+a*.14381584525108337))))));switch(q|0){case 0:{a=1.0/+X(+(s*s+r*r+1.0));t=-s*a;u=r*a;v=a;break}case 1:{a=1.0/+X(+(s*s+r*r+1.0));t=a;u=r*a;v=s*a;break}case 2:{a=1.0/+X(+(s*s+r*r+1.0));t=-r*a;u=a;v=s*a;break}case 3:{a=-1.0/+X(+(s*s+r*r+1.0));t=a;u=r*a;v=-s*a;break}case 4:{a=-1.0/+X(+(s*s+r*r+1.0));t=-r*a;u=a;v=-s*a;break}case 5:{a=-1.0/+X(+(s*s+r*r+1.0));t=-s*a;u=-r*a;v=a;break}default:{t=0.0;u=0.0;v=0.0}}if(u==0.0&t==0.0)w=0.0;else w=+ug(u,t);p[d>>3]=w;w=+sg(v);p[e>>3]=w;o=0;return o|0}function kg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0.0,j=0.0,l=0.0,m=0.0,n=0,o=0.0,q=0,r=0.0,s=0,t=0.0,u=0,v=0.0,w=0.0,x=0.0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0;f=c+4|0;if((k[f>>2]|0)!=703){i[c>>0]=81;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=703;f=c+8|0;g=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;h=+p[g>>3];if(h==0.0){p[g>>3]=57.29577951308232;j=.022222222222222223;l=45.0}else{m=h*3.141592653589793*.25;j=1.0/m;l=m}p[c+112>>3]=l;p[c+120>>3]=j;k[c+1888>>2]=37;k[c+1892>>2]=38}if(+W(+b)==90.0){p[d>>3]=0.0;j=+W(+(+p[c+112>>3]*2.0));p[e>>3]=b<0.0?-j:j;n=0;return n|0}j=+og(b);l=j*+og(a);m=j*+pg(a);j=+pg(b);g=l>j;h=g?l:j;f=m>h;o=f?m:h;h=-l;q=o<h;r=q?h:o;o=-m;s=r<o;t=s?o:r;r=-j;u=t<r;v=1.0-(u?r:t);switch((u?5:s?4:q?3:f?2:g&1)|0){case 0:{if(v<1.0e-08){t=(90.0-b)*3.141592653589793/180.0;w=h;x=t*t*.5;y=0.0;z=m;A=2.0}else{w=h;x=v;y=0.0;z=m;A=2.0}break}case 1:{if(v<1.0e-08){t=b*3.141592653589793/180.0;r=+Gv(a,360.0);B=r<-180.0?r+360.0:r;r=(B>180.0?B+-360.0:B)*.017453292519943295;w=j;x=(t*t+r*r)*.5;y=0.0;z=m;A=0.0}else{w=j;x=v;y=0.0;z=m;A=0.0}break}case 2:{if(v<1.0e-08){r=b*3.141592653589793/180.0;t=+Gv(a,360.0);B=(90.0-(t<-180.0?t+360.0:t))*3.141592653589793/180.0;w=j;x=(r*r+B*B)*.5;y=2.0;z=h;A=0.0}else{w=j;x=v;y=2.0;z=h;A=0.0}break}case 3:{if(v<1.0e-08){h=b*3.141592653589793/180.0;B=+Gv(a,360.0);r=(180.0-(B<0.0?B+360.0:B))*3.141592653589793/180.0;w=j;x=(h*h+r*r)*.5;y=4.0;z=o;A=0.0}else{w=j;x=v;y=4.0;z=o;A=0.0}break}case 4:{if(v<1.0e-08){o=b*3.141592653589793/180.0;r=+Gv(a,360.0);a=r>180.0?r+-360.0:r;r=(a+90.0)*3.141592653589793/180.0*a;w=j;x=(o*o+r*r)*.5;y=6.0;z=l;A=0.0}else{w=j;x=v;y=6.0;z=l;A=0.0}break}case 5:{if(v<1.0e-08){j=(b+90.0)*3.141592653589793/180.0;w=l;x=j*j*.5;y=0.0;z=m;A=-2.0}else{w=l;x=v;y=0.0;z=m;A=-2.0}break}default:{w=0.0;x=v;y=0.0;z=0.0;A=0.0}}do if(!(z==0.0&w==0.0)){v=+W(+w);if(v<=-z){m=w/z;l=m*m+1.0;j=-+X(+(x/(1.0-1.0/+X(+(l+1.0)))));b=+tg(m);C=j;D=j/15.0*(b-+sg(m/+X(+(l+l))));break}if(z>=v){v=w/z;l=v*v+1.0;m=+X(+(x/(1.0-1.0/+X(+(l+1.0)))));b=+tg(v);C=m;D=m/15.0*(b-+sg(v/+X(+(l+l))));break}l=+W(+z);if(l<-w){v=z/w;b=v*v+1.0;m=-+X(+(x/(1.0-1.0/+X(+(b+1.0)))));j=+tg(v);C=m/15.0*(j-+sg(v/+X(+(b+b))));D=m;break}if(w>l){l=z/w;m=l*l+1.0;b=+X(+(x/(1.0-1.0/+X(+(m+1.0)))));v=+tg(l);C=b/15.0*(v-+sg(l/+X(+(m+m))));D=b}else{C=0.0;D=0.0}}else{C=0.0;D=0.0}while(0);x=+W(+C);do if(x>1.0)if(x>1.000000000001){n=2;return n|0}else{E=C<0.0?-1.0:1.0;break}else E=C;while(0);C=+W(+D);do if(C>1.0)if(C>1.000000000001){n=2;return n|0}else{F=D<0.0?-1.0:1.0;break}else F=D;while(0);g=c+112|0;p[d>>3]=(y+E)*+p[g>>3];p[e>>3]=(A+F)*+p[g>>3];n=0;return n|0}function lg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0,g=0.0,h=0,j=0.0,l=0.0,m=0.0,n=0.0,o=0,q=0,r=0.0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0,z=0.0,A=0.0,B=0.0,C=0.0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0;f=c+4|0;if((k[f>>2]|0)==703)g=+p[c+120>>3];else{i[c>>0]=81;i[c+1>>0]=83;i[c+2>>0]=67;i[c+3>>0]=0;k[f>>2]=703;f=c+8|0;h=c+24|0;k[f>>2]=0;k[f+4>>2]=0;k[f+8>>2]=0;k[f+12>>2]=0;j=+p[h>>3];if(j==0.0){p[h>>3]=57.29577951308232;l=.022222222222222223;m=45.0}else{n=j*3.141592653589793*.25;l=1.0/n;m=n}p[c+112>>3]=m;p[c+120>>3]=l;k[c+1888>>2]=37;k[c+1892>>2]=38;g=l}l=g*a;a=g*b;b=+W(+l);if(!(b<=1.0)){if(b>7.0){o=2;return o|0}if(+W(+a)>1.0){o=2;return o|0}}else if(+W(+a)>3.0){o=2;return o|0}b=l<-1.0?l+8.0:l;do if(!(b>5.0)){if(b>3.0){q=3;r=b+-4.0;s=a;break}if(b>1.0){q=2;r=b+-2.0;s=a;break}if(a>1.0){q=0;r=b;s=a+-2.0;break}if(a<-1.0){q=5;r=b;s=a+2.0}else{q=1;r=b;s=a}}else{q=4;r=b+-6.0;s=a}while(0);c=+W(+r)>+W(+s);if(c)if(r==0.0){t=0.0;u=1.0;v=0.0;w=1.0;x=28}else{a=s*15.0/r;b=+pg(a);l=b/(+og(a)+-.7071067811865475);a=l*l+1.0;y=l;z=r*r*(1.0-1.0/+X(+(a+1.0)));A=a;x=26}else if(s==0.0){t=0.0;u=1.0;v=0.0;w=1.0;x=28}else{a=r*15.0/s;l=+pg(a);b=l/(+og(a)+-.7071067811865475);a=b*b+1.0;y=b;z=s*s*(1.0-1.0/+X(+(a+1.0)));A=a;x=26}if((x|0)==26){a=1.0-z;if(a<-1.0)if(a<-1.000000000001){o=2;return o|0}else{B=y;C=-1.0;D=0.0}else{t=y;u=a;v=z;w=A;x=28}}if((x|0)==28){B=t;C=u;D=+X(+(v*(2.0-v)/w))}a:do switch(q|0){case 0:{if(c){w=r<0.0?-D:D;E=-(B*w);F=w;G=C;break a}else{w=s>0.0?-D:D;E=w;F=-(B*w);G=C;break a}break}case 1:{if(c){w=r<0.0?-D:D;E=C;F=w;G=B*w;break a}else{w=s<0.0?-D:D;E=C;F=B*w;G=w;break a}break}case 2:{if(c){w=r>0.0?-D:D;E=w;F=C;G=-(B*w);break a}else{w=s<0.0?-D:D;E=-(B*w);F=C;G=w;break a}break}case 3:{w=-C;if(c){v=r>0.0?-D:D;E=w;F=v;G=-(B*v);break a}else{v=s<0.0?-D:D;E=w;F=-(B*v);G=v;break a}break}case 4:{v=-C;if(c){w=r<0.0?-D:D;E=w;F=v;G=B*w;break a}else{w=s<0.0?-D:D;E=B*w;F=v;G=w;break a}break}case 5:{w=-C;if(c){v=r<0.0?-D:D;E=B*v;F=v;G=w;break a}else{v=s<0.0?-D:D;E=v;F=B*v;G=w;break a}break}default:{E=0.0;F=0.0;G=0.0}}while(0);if(E==0.0&F==0.0)H=0.0;else H=+ug(F,E);p[d>>3]=H;H=+sg(G);p[e>>3]=H;o=0;return o|0}function mg(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0.0,i=0.0,j=0,k=0,l=0.0,m=0.0,n=0.0,o=0.0;f=+og(b);g=+pg(b);h=a-+p[c>>3];a=+og(h);i=+pg(h);j=c+32|0;k=c+24|0;l=g*+p[j>>3]-a*(f*+p[k>>3]);if(+W(+l)<1.0e-05){m=+og(+p[c+8>>3]+b);n=(1.0-a)*(f*+p[k>>3])-m}else n=l;l=f*i;if(l!=-0.0|n!=0.0)o=+ug(-l,n);else o=h+-180.0;i=o+ +p[c+16>>3];p[d>>3]=i;if(!(i>180.0)){if(i<-180.0)p[d>>3]=i+360.0}else p[d>>3]=i+-360.0;if(+Gv(h,180.0)==0.0){h=a*+p[c+8>>3]+b;b=h>90.0?180.0-h:h;p[e>>3]=b;if(!(b<-90.0))return 0;p[e>>3]=-180.0-b;return 0}b=g*+p[k>>3]+a*(f*+p[j>>3]);if(!(+W(+b)>.99)){f=+sg(b);p[e>>3]=f;return 0}f=+rg(+X(+(l*l+n*n)));if(b<0.0){p[e>>3]=-f;return 0}else{p[e>>3]=f;return 0}return 0}function ng(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;var f=0.0,g=0.0,h=0.0,i=0.0,j=0,k=0,l=0.0,m=0.0,n=0.0,o=0.0,q=0.0;f=+og(b);g=+pg(b);h=a-+p[c+16>>3];a=+og(h);i=+pg(h);j=c+32|0;k=c+24|0;l=g*+p[j>>3]-a*(f*+p[k>>3]);if(+W(+l)<1.0e-05){m=+og(+p[c+8>>3]+b);n=(1.0-a)*(f*+p[k>>3])-m}else n=l;l=f*i;if(l!=-0.0|n!=0.0)o=+ug(-l,n);else o=h+180.0;i=o+ +p[c>>3];p[d>>3]=i;if(!(+p[c>>3]>=0.0))if(i>0.0){o=i+-360.0;p[d>>3]=o;q=o}else q=i;else if(i<0.0){o=i+360.0;p[d>>3]=o;q=o}else q=i;if(!(q>360.0)){if(q<-360.0)p[d>>3]=q+360.0}else p[d>>3]=q+-360.0;if(+Gv(h,180.0)==0.0){h=a*+p[c+8>>3]+b;b=h>90.0?180.0-h:h;p[e>>3]=b;if(!(b<-90.0))return 0;p[e>>3]=-180.0-b;return 0}b=g*+p[k>>3]+a*(f*+p[j>>3]);if(!(+W(+b)>.99)){f=+sg(b);p[e>>3]=f;return 0}f=+rg(+X(+(l*l+n*n)));if(b<0.0){p[e>>3]=-f;return 0}else{p[e>>3]=f;return 0}return 0}function og(a){a=+a;var b=0.0,c=0.0,d=0.0;b=+Gv(a,360.0);c=+W(+b);if(!(b==0.0))if(!(c==90.0))if(!(c==180.0))if(c==270.0)d=0.0;else d=+Z(+(a*.017453292519943295));else d=-1.0;else d=0.0;else d=1.0;return +d}function pg(a){a=+a;var b=0.0,c=0.0;b=+Gv(a+-90.0,360.0);if(!(b==0.0))if(!(b==90.0))if(!(b==180.0))if(b==270.0)c=0.0;else c=+_(+(a*.017453292519943295));else c=-1.0;else c=0.0;else c=1.0;return +c}function qg(a){a=+a;var b=0.0,c=0.0;b=+Gv(a,360.0);if(!(b==0.0)?!(+W(+b)==180.0):0)if(!(b==45.0|b==225.0))if(b==-135.0|b==-315.0)c=-1.0;else c=+$(+(a*.017453292519943295));else c=1.0;else c=0.0;return +c}function rg(a){a=+a;var b=0.0,c=0;if(!(a>=1.0))if(!(a==0.0))if(a<=-1.0&a+1.0>-1.0e-10)b=180.0;else c=5;else b=90.0;else if(a+-1.0<1.0e-10)b=0.0;else c=5;if((c|0)==5)b=+aa(+a)*57.29577951308232;return +b}function sg(a){a=+a;var b=0.0,c=0;if(!(a<=-1.0))if(!(a==0.0))if(a>=1.0&a+-1.0<1.0e-10)b=90.0;else c=5;else b=0.0;else if(a+1.0>-1.0e-10)b=-90.0;else c=5;if((c|0)==5)b=+ba(+a)*57.29577951308232;return +b}function tg(a){a=+a;var b=0.0;if(!(a==-1.0))if(!(a==0.0))if(a==1.0)b=45.0;else b=+ca(+a)*57.29577951308232;else b=0.0;else b=-45.0;return +b}function ug(a,b){a=+a;b=+b;var c=0.0,d=0;if(a==0.0)if(!(b>=0.0))if(b<0.0)c=180.0;else d=7;else c=0.0;else if(b==0.0)if(!(a>0.0))if(a<0.0)c=-90.0;else d=7;else c=90.0;else d=7;if((d|0)==7)c=+da(+a,+b)*57.29577951308232;return +c}function vg(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0;c=r;r=r+48|0;d=c+24|0;e=c+16|0;f=c+8|0;g=c;h=c+32|0;if((k[a+5968>>2]|0)!=1){r=c;return}i=a+3324|0;if((k[i>>2]|0)==2){k[i>>2]=3;k[a+5976>>2]=0;k[a+6784>>2]=0;k[a+7592>>2]=0;k[a+8400>>2]=0;r=c;return}i=a+5976|0;if(be(b,606594,i)|0){j=k[i>>2]|0;i=(j|0)<0;if(!i){l=(j<<3)+8|0;m=0;while(1){sy(a+5984+(m*80|0)|0,0,l|0)|0;if((m|0)==(j|0))break;else m=m+1|0}if(!i){i=j+1|0;m=0;l=i;while(1){if((j|0)>=(m|0)){n=0;do{k[g>>2]=m;k[g+4>>2]=n;Xw(h,606659,g)|0;ie(b,h,a+5984+(m*80|0)+(n<<3)|0)|0;n=n+1|0}while((n|0)!=(l|0))}m=m+1|0;if((m|0)==(i|0))break;else l=l+-1|0}}}}else xd(606602);l=a+6784|0;if(be(b,606667,l)|0){i=k[l>>2]|0;l=(i|0)<0;if(!l){m=(i<<3)+8|0;g=0;while(1){sy(a+6792+(g*80|0)|0,0,m|0)|0;if((g|0)==(i|0))break;else g=g+1|0}if(!l){l=i+1|0;g=0;m=l;while(1){if((i|0)>=(g|0)){j=0;do{k[f>>2]=g;k[f+4>>2]=j;Xw(h,606732,f)|0;ie(b,h,a+6792+(g*80|0)+(j<<3)|0)|0;j=j+1|0}while((j|0)!=(m|0))}g=g+1|0;if((g|0)==(l|0))break;else m=m+-1|0}}}}else xd(606675);m=a+7592|0;if(be(b,606740,m)|0){l=k[m>>2]|0;m=(l|0)<0;if(!m){g=(l<<3)+8|0;f=0;while(1){sy(a+7600+(f*80|0)|0,0,g|0)|0;if((f|0)==(l|0))break;else f=f+1|0}if(!m){m=l+1|0;f=0;g=m;while(1){if((l|0)>=(f|0)){i=0;do{k[e>>2]=f;k[e+4>>2]=i;Xw(h,606807,e)|0;ie(b,h,a+7600+(f*80|0)+(i<<3)|0)|0;i=i+1|0}while((i|0)!=(g|0))}f=f+1|0;if((f|0)==(m|0))break;else g=g+-1|0}}}}else xd(606749);g=a+8400|0;if(!(be(b,606816,g)|0)){xd(606825);r=c;return}m=k[g>>2]|0;g=(m|0)<0;if(g){r=c;return}f=(m<<3)+8|0;e=0;while(1){sy(a+8408+(e*80|0)|0,0,f|0)|0;if((e|0)==(m|0))break;else e=e+1|0}if(g){r=c;return}g=m+1|0;e=0;f=g;while(1){if((m|0)>=(e|0)){l=0;do{k[d>>2]=e;k[d+4>>2]=l;Xw(h,606883,d)|0;ie(b,h,a+8408+(e*80|0)+(l<<3)|0)|0;l=l+1|0}while((l|0)!=(f|0))}e=e+1|0;if((e|0)==(g|0))break;else f=f+-1|0}r=c;return}function wg(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0,j=0,l=0.0,m=0.0,n=0,o=0,q=0.0,s=0,t=0,u=0.0,v=0,w=0.0,x=0,y=0.0,z=0.0,A=0.0;f=r;r=r+80|0;g=f;if((k[a+5968>>2]|0)!=1){p[d>>3]=b;h=c;p[e>>3]=h;r=f;return}i=k[a+7592>>2]|0;j=k[a+8400>>2]|0;l=b-+p[a+16>>3];m=c-+p[a+24>>3];if((i|0)>=0){n=0;while(1){o=i-n|0;q=+p[a+7600+(o*80|0)+(n<<3)>>3];s=g+(n<<3)|0;p[s>>3]=q;if((n|0)>0){t=n;u=q;while(1){v=t;t=t+-1|0;q=m*u+ +p[a+7600+(o*80|0)+(t<<3)>>3];if((v|0)<=1){w=q;break}else u=q}p[s>>3]=w}if((i|0)==(n|0))break;else n=n+1|0}w=+p[g>>3];if((i|0)>0){n=i+1|0;t=i;u=w;while(1){q=l*u+ +p[g+(n-t<<3)>>3];if((t|0)>1){t=t+-1|0;u=q}else{x=g;y=q;break}}}else{x=g;y=w}}else{x=g;y=0.0}p[d>>3]=y;if((j|0)>=0){t=0;while(1){n=j-t|0;y=+p[a+8408+(n*80|0)+(t<<3)>>3];i=g+(t<<3)|0;p[i>>3]=y;if((t|0)>0){o=t;w=y;while(1){v=o;o=o+-1|0;y=m*w+ +p[a+8408+(n*80|0)+(o<<3)>>3];if((v|0)<=1){z=y;break}else w=y}p[i>>3]=z}if((j|0)==(t|0))break;else t=t+1|0}z=+p[x>>3];if((j|0)>0){t=j+1|0;a=j;m=z;while(1){w=l*m+ +p[g+(t-a<<3)>>3];if((a|0)>1){a=a+-1|0;m=w}else{A=w;break}}}else A=z}else A=+p[x>>3];p[e>>3]=A;p[d>>3]=+p[d>>3]+b;h=+p[e>>3]+c;p[e>>3]=h;r=f;return}function xg(a,b,c,d,e){a=a|0;b=+b;c=+c;d=d|0;e=e|0;var f=0,g=0,h=0.0,i=0,j=0,l=0.0,m=0.0,n=0,o=0,q=0.0,s=0,t=0,u=0.0,v=0,w=0.0,x=0,y=0.0,z=0.0,A=0.0;f=r;r=r+80|0;g=f;if((k[a+5968>>2]|0)!=1){p[d>>3]=b;h=c;p[e>>3]=h;r=f;return}i=k[a+5976>>2]|0;j=k[a+6784>>2]|0;l=b-+p[a+16>>3];m=c-+p[a+24>>3];if((i|0)>=0){n=0;while(1){o=i-n|0;q=+p[a+5984+(o*80|0)+(n<<3)>>3];s=g+(n<<3)|0;p[s>>3]=q;if((n|0)>0){t=n;u=q;while(1){v=t;t=t+-1|0;q=m*u+ +p[a+5984+(o*80|0)+(t<<3)>>3];if((v|0)<=1){w=q;break}else u=q}p[s>>3]=w}if((i|0)==(n|0))break;else n=n+1|0}w=+p[g>>3];if((i|0)>0){n=i+1|0;t=i;u=w;while(1){q=l*u+ +p[g+(n-t<<3)>>3];if((t|0)>1){t=t+-1|0;u=q}else{x=g;y=q;break}}}else{x=g;y=w}}else{x=g;y=0.0}p[d>>3]=y;if((j|0)>=0){t=0;while(1){n=j-t|0;y=+p[a+6792+(n*80|0)+(t<<3)>>3];i=g+(t<<3)|0;p[i>>3]=y;if((t|0)>0){o=t;w=y;while(1){v=o;o=o+-1|0;y=m*w+ +p[a+6792+(n*80|0)+(o<<3)>>3];if((v|0)<=1){z=y;break}else w=y}p[i>>3]=z}if((j|0)==(t|0))break;else t=t+1|0}z=+p[x>>3];if((j|0)>0){t=j+1|0;a=j;m=z;while(1){w=l*m+ +p[g+(t-a<<3)>>3];if((a|0)>1){a=a+-1|0;m=w}else{A=w;break}}}else A=z}else A=+p[x>>3];p[e>>3]=A;p[d>>3]=+p[d>>3]+b;h=+p[e>>3]+c;p[e>>3]=h;r=f;return}function yg(a,b){a=a|0;b=b|0;var c=0;if((xx(b)|0)<9){k[a+5968>>2]=0;return}c=a+5968|0;if(!(Ax(b+8|0,606892,4)|0)){k[c>>2]=1;return}else{k[c>>2]=0;return}}function zg(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,k=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0;d=a>>>16;e=a&65535;if((c|0)==1){a=(l[b>>0]|0)+e|0;f=a>>>0>65520?a+-65521|0:a;a=f+d|0;g=(a>>>0>65520?a+15|0:a)<<16|f;return g|0}if(!b){g=1;return g|0}if(c>>>0<16){if(!c){h=e;i=d}else{f=b;a=c;j=e;k=d;while(1){a=a+-1|0;m=(l[f>>0]|0)+j|0;n=m+k|0;if(!a){h=m;i=n;break}else{f=f+1|0;j=m;k=n}}}g=((i>>>0)%65521|0)<<16|(h>>>0>65520?h+-65521|0:h);return g|0}if(c>>>0>5551){h=c+-5552|0;i=(h>>>0)%5552|0;k=h-i+5552|0;h=b;j=c;f=e;a=d;while(1){j=j+-5552|0;n=h;m=f;o=347;p=a;while(1){q=(l[n>>0]|0)+m|0;r=q+(l[n+1>>0]|0)|0;s=r+(l[n+2>>0]|0)|0;t=s+(l[n+3>>0]|0)|0;u=t+(l[n+4>>0]|0)|0;v=u+(l[n+5>>0]|0)|0;w=v+(l[n+6>>0]|0)|0;x=w+(l[n+7>>0]|0)|0;y=x+(l[n+8>>0]|0)|0;z=y+(l[n+9>>0]|0)|0;A=z+(l[n+10>>0]|0)|0;B=A+(l[n+11>>0]|0)|0;C=B+(l[n+12>>0]|0)|0;D=C+(l[n+13>>0]|0)|0;E=D+(l[n+14>>0]|0)|0;F=E+(l[n+15>>0]|0)|0;G=q+p+r+s+t+u+v+w+x+y+z+A+B+C+D+E+F|0;o=o+-1|0;if(!o){H=F;I=G;break}else{n=n+16|0;m=F;p=G}}p=(H>>>0)%65521|0;m=(I>>>0)%65521|0;if(j>>>0<=5551){J=p;K=m;break}else{h=h+5552|0;f=p;a=m}}a=b+k|0;if(i)if(i>>>0>15){L=i;M=a;N=J;O=K;P=15}else{Q=i;R=a;S=J;T=K;P=18}else{U=J;V=K}}else{L=c;M=b;N=e;O=d;P=15}if((P|0)==15){d=L+-16|0;e=d&-16;b=M+(e+16)|0;c=L;L=M;M=N;N=O;while(1){c=c+-16|0;O=(l[L>>0]|0)+M|0;K=O+(l[L+1>>0]|0)|0;J=K+(l[L+2>>0]|0)|0;a=J+(l[L+3>>0]|0)|0;i=a+(l[L+4>>0]|0)|0;k=i+(l[L+5>>0]|0)|0;f=k+(l[L+6>>0]|0)|0;h=f+(l[L+7>>0]|0)|0;j=h+(l[L+8>>0]|0)|0;I=j+(l[L+9>>0]|0)|0;H=I+(l[L+10>>0]|0)|0;m=H+(l[L+11>>0]|0)|0;p=m+(l[L+12>>0]|0)|0;n=p+(l[L+13>>0]|0)|0;o=n+(l[L+14>>0]|0)|0;G=o+(l[L+15>>0]|0)|0;F=O+N+K+J+a+i+k+f+h+j+I+H+m+p+n+o+G|0;if(c>>>0<=15){W=G;X=F;break}else{L=L+16|0;M=G;N=F}}if((d|0)==(e|0)){Y=W;Z=X;P=19}else{Q=d-e|0;R=b;S=W;T=X;P=18}}if((P|0)==18)while(1){P=0;Q=Q+-1|0;X=(l[R>>0]|0)+S|0;W=X+T|0;if(!Q){Y=X;Z=W;P=19;break}else{R=R+1|0;S=X;T=W;P=18}}if((P|0)==19){U=(Y>>>0)%65521|0;V=(Z>>>0)%65521|0}g=V<<16|U;return g|0}function Ag(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;if(!b){d=0;return d|0}e=~a;a:do if(c){a=b;f=c;g=e;while(1){if(!(a&3)){h=a;i=f;j=g;break}m=k[104068+(((l[a>>0]|0)^g&255)<<2)>>2]^g>>>8;f=f+-1|0;if(!f){n=m;break a}else{a=a+1|0;g=m}}if(i>>>0>31){g=i+-32|0;a=g&-32;f=a+32|0;m=i;o=h;p=j;while(1){q=k[o>>2]^p;r=k[106116+((q>>>8&255)<<2)>>2]^k[107140+((q&255)<<2)>>2]^k[105092+((q>>>16&255)<<2)>>2]^k[104068+(q>>>24<<2)>>2]^k[o+4>>2];q=k[106116+((r>>>8&255)<<2)>>2]^k[107140+((r&255)<<2)>>2]^k[105092+((r>>>16&255)<<2)>>2]^k[104068+(r>>>24<<2)>>2]^k[o+8>>2];r=k[106116+((q>>>8&255)<<2)>>2]^k[107140+((q&255)<<2)>>2]^k[105092+((q>>>16&255)<<2)>>2]^k[104068+(q>>>24<<2)>>2]^k[o+12>>2];q=k[106116+((r>>>8&255)<<2)>>2]^k[107140+((r&255)<<2)>>2]^k[105092+((r>>>16&255)<<2)>>2]^k[104068+(r>>>24<<2)>>2]^k[o+16>>2];r=k[106116+((q>>>8&255)<<2)>>2]^k[107140+((q&255)<<2)>>2]^k[105092+((q>>>16&255)<<2)>>2]^k[104068+(q>>>24<<2)>>2]^k[o+20>>2];q=k[106116+((r>>>8&255)<<2)>>2]^k[107140+((r&255)<<2)>>2]^k[105092+((r>>>16&255)<<2)>>2]^k[104068+(r>>>24<<2)>>2]^k[o+24>>2];r=k[106116+((q>>>8&255)<<2)>>2]^k[107140+((q&255)<<2)>>2]^k[105092+((q>>>16&255)<<2)>>2]^k[104068+(q>>>24<<2)>>2]^k[o+28>>2];q=k[106116+((r>>>8&255)<<2)>>2]^k[107140+((r&255)<<2)>>2]^k[105092+((r>>>16&255)<<2)>>2]^k[104068+(r>>>24<<2)>>2];m=m+-32|0;if(m>>>0<=31){s=q;break}else{o=o+32|0;p=q}}t=g-a|0;u=h+f|0;v=s}else{t=i;u=h;v=j}if(t>>>0>3){p=t+-4|0;o=p>>>2;m=o<<2;q=t;r=u;w=v;while(1){x=k[r>>2]^w;y=k[106116+((x>>>8&255)<<2)>>2]^k[107140+((x&255)<<2)>>2]^k[105092+((x>>>16&255)<<2)>>2]^k[104068+(x>>>24<<2)>>2];q=q+-4|0;if(q>>>0<=3){z=y;break}else{r=r+4|0;w=y}}A=p-m|0;B=u+(o+1<<2)|0;C=z}else{A=t;B=u;C=v}if(!A)n=C;else{w=B;r=A;q=C;while(1){f=k[104068+(((l[w>>0]|0)^q&255)<<2)>>2]^q>>>8;r=r+-1|0;if(!r){n=f;break}else{w=w+1|0;q=f}}}}else n=e;while(0);d=~n;return d|0}function Bg(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0;if(!g){j=-6;return j|0}if((h|0)!=56|(i[g>>0]|0)!=49){j=-6;return j|0}if(!a){j=-2;return j|0}g=a+24|0;k[g>>2]=0;h=a+32|0;l=k[h>>2]|0;if(!l){k[h>>2]=7;k[a+40>>2]=0;m=7}else m=l;l=a+36|0;if(!(k[l>>2]|0))k[l>>2]=1;n=(b|0)==-1?6:b;if((d|0)<0){o=0-d|0;p=0}else{b=(d|0)>15;o=b?d+-16|0:d;p=b?2:1}if(f>>>0>4|(n>>>0>9|((c|0)!=8|(e+-1|0)>>>0>8|(o&-8|0)!=8))){j=-2;return j|0}c=(o|0)==8?9:o;o=a+40|0;b=Pb[m&63](k[o>>2]|0,1,5828)|0;if(!b){j=-4;return j|0}m=a+28|0;k[m>>2]=b;k[b>>2]=a;k[b+24>>2]=p;k[b+28>>2]=0;k[b+48>>2]=c;p=1<<c;c=b+44|0;k[c>>2]=p;k[b+52>>2]=p+-1;d=e+7|0;k[b+80>>2]=d;q=1<<d;d=b+76|0;k[d>>2]=q;k[b+84>>2]=q+-1;k[b+88>>2]=((e+9|0)>>>0)/3|0;q=Pb[k[h>>2]&63](k[o>>2]|0,p,2)|0;p=b+56|0;k[p>>2]=q;q=Pb[k[h>>2]&63](k[o>>2]|0,k[c>>2]|0,2)|0;r=b+64|0;k[r>>2]=q;sy(q|0,0,k[c>>2]<<1|0)|0;c=Pb[k[h>>2]&63](k[o>>2]|0,k[d>>2]|0,2)|0;d=b+68|0;k[d>>2]=c;k[b+5824>>2]=0;c=1<<e+6;e=b+5788|0;k[e>>2]=c;q=Pb[k[h>>2]&63](k[o>>2]|0,c,4)|0;k[b+8>>2]=q;c=k[e>>2]|0;k[b+12>>2]=c<<2;if(((k[p>>2]|0)!=0?(k[r>>2]|0)!=0:0)?!((k[d>>2]|0)==0|(q|0)==0):0){k[b+5796>>2]=q+(c>>>1<<1);k[b+5784>>2]=q+(c*3|0);k[b+132>>2]=n;k[b+136>>2]=f;i[b+36>>0]=8;j=Dg(a)|0;return j|0}k[b+4>>2]=666;k[g>>2]=k[28253];g=k[m>>2]|0;if(!g){j=-4;return j|0}switch(k[g+4>>2]|0){case 42:case 69:case 73:case 91:case 103:case 113:case 666:break;default:{j=-4;return j|0}}b=k[g+8>>2]|0;if(!b)s=g;else{Tb[k[l>>2]&1](k[o>>2]|0,b);s=k[m>>2]|0}b=k[s+68>>2]|0;if(!b)t=s;else{Tb[k[l>>2]&1](k[o>>2]|0,b);t=k[m>>2]|0}b=k[t+64>>2]|0;if(!b)u=t;else{Tb[k[l>>2]&1](k[o>>2]|0,b);u=k[m>>2]|0}b=k[u+56>>2]|0;if(!b)v=u;else{Tb[k[l>>2]&1](k[o>>2]|0,b);v=k[m>>2]|0}Tb[k[l>>2]&1](k[o>>2]|0,v);k[m>>2]=0;j=-4;return j|0}function Cg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;if(!a){b=-2;return b|0}c=a+28|0;d=k[c>>2]|0;if(!d){b=-2;return b|0}e=k[d+4>>2]|0;switch(e|0){case 42:case 69:case 73:case 91:case 103:case 113:case 666:break;default:{b=-2;return b|0}}f=k[d+8>>2]|0;if(!f)g=d;else{Tb[k[a+36>>2]&1](k[a+40>>2]|0,f);g=k[c>>2]|0}f=k[g+68>>2]|0;if(!f)h=g;else{Tb[k[a+36>>2]&1](k[a+40>>2]|0,f);h=k[c>>2]|0}f=k[h+64>>2]|0;if(!f)i=h;else{Tb[k[a+36>>2]&1](k[a+40>>2]|0,f);i=k[c>>2]|0}f=k[i+56>>2]|0;if(!f){j=a+40|0;l=a+36|0;m=i}else{i=a+36|0;h=a+40|0;Tb[k[i>>2]&1](k[h>>2]|0,f);j=h;l=i;m=k[c>>2]|0}Tb[k[l>>2]&1](k[j>>2]|0,m);k[c>>2]=0;b=(e|0)==113?-3:0;return b|0}function Dg(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;if(!a){b=-2;return b|0}c=k[a+28>>2]|0;if(!c){b=-2;return b|0}if(!(k[a+32>>2]|0)){b=-2;return b|0}if(!(k[a+36>>2]|0)){b=-2;return b|0}k[a+20>>2]=0;k[a+8>>2]=0;k[a+24>>2]=0;k[a+44>>2]=2;k[c+20>>2]=0;k[c+16>>2]=k[c+8>>2];d=c+24|0;e=k[d>>2]|0;if((e|0)<0){f=0-e|0;k[d>>2]=f;g=f}else g=e;k[c+4>>2]=(g|0)!=0?42:113;if((g|0)==2)h=Ag(0,0,0)|0;else h=zg(0,0,0)|0;k[a+48>>2]=h;k[c+40>>2]=0;fh(c);k[c+60>>2]=k[c+44>>2]<<1;h=k[c+76>>2]|0;a=k[c+68>>2]|0;j[a+(h+-1<<1)>>1]=0;sy(a|0,0,(h<<1)+-2|0)|0;h=k[c+132>>2]|0;k[c+128>>2]=m[112260+(h*12|0)+2>>1];k[c+140>>2]=m[112260+(h*12|0)>>1];k[c+144>>2]=m[112260+(h*12|0)+4>>1];k[c+124>>2]=m[112260+(h*12|0)+6>>1];k[c+108>>2]=0;k[c+92>>2]=0;k[c+116>>2]=0;k[c+120>>2]=2;k[c+96>>2]=2;k[c+112>>2]=0;k[c+104>>2]=0;k[c+72>>2]=0;b=0;return b|0}
function yu(){var a=0,b=0,c=0,d=0,e=0,f=0,g=0,h=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0.0,eb=0,fb=0,gb=0;a=r;r=r+624|0;b=a+360|0;c=a+104|0;d=a;if(!(k[30836]|0)){k[30836]=1;if(!(k[30837]|0))k[30837]=1;e=k[30806]|0;if(!e){f=k[31230]|0;k[30806]=f;g=f}else g=e;if(!(k[30807]|0))k[30807]=k[31231];e=k[30838]|0;do if(!e){f=cy(4)|0;k[30838]=f;if(!f)Cu(786687);else{k[f>>2]=0;k[30840]=1;k[30839]=0;h=g;m=17;break}}else{f=k[30839]|0;n=k[e+(f<<2)>>2]|0;if(!n){o=k[30840]|0;if(f>>>0>=(o+-1|0)>>>0){q=o+8|0;o=fy(e,q<<2)|0;k[30838]=o;if(!o)Cu(786687);else{s=o+(k[30840]<<2)|0;k[s>>2]=0;k[s+4>>2]=0;k[s+8>>2]=0;k[s+12>>2]=0;k[s+16>>2]=0;k[s+20>>2]=0;k[s+24>>2]=0;k[s+28>>2]=0;k[30840]=q;h=k[30806]|0;m=17;break}}else{h=g;m=17}}else{t=e;u=f;v=n}}while(0);if((m|0)==17){e=zu(h,16384)|0;h=k[30839]|0;k[(k[30838]|0)+(h<<2)>>2]=e;e=k[30838]|0;t=e;u=h;v=k[e+(h<<2)>>2]|0}k[30841]=k[v+16>>2];h=k[v+8>>2]|0;k[30842]=h;k[30808]=h;k[30806]=k[k[t+(u<<2)>>2]>>2];i[786736]=i[h>>0]|0}a:while(1){h=k[30842]|0;i[h>>0]=i[786736]|0;u=h;t=h;h=k[30837]|0;b:while(1){v=t;e=h;while(1){g=k[123372+(l[v>>0]<<2)>>2]|0;if(!(j[326324+(e<<1)>>1]|0)){w=g;x=e}else{k[31099]=e;k[31100]=v;w=g;x=e}c:while(1){g=w&255;n=x;while(1){f=(j[326644+(n<<1)>>1]|0)+g|0;if((j[326978+(f<<1)>>1]|0)==(n|0)){y=f;break c}f=j[327826+(n<<1)>>1]|0;q=f<<16>>16;if(f<<16>>16>159){z=q;break}else n=q}w=k[124404+(g<<2)>>2]|0;x=z}n=j[328160+(y<<1)>>1]|0;q=v+1|0;if((j[326644+(n<<1)>>1]|0)==368){A=u;B=q;C=n;break}else{v=q;e=n}}d:while(1){e=A;v=B;n=C;e:while(1){q=j[326324+(n<<1)>>1]|0;if(!(q<<16>>16)){D=j[326324+(k[31099]<<1)>>1]|0;E=k[31100]|0}else{D=q<<16>>16;E=v}k[30808]=A;q=E;k[30809]=q-e;i[786736]=i[E>>0]|0;i[E>>0]=0;k[30842]=E;f=D;f:while(1){switch(f|0){case 1:{continue a;break}case 2:{m=34;break a;break}case 3:{m=37;break a;break}case 4:{m=54;break a;break}case 5:{m=79;break a;break}case 6:{m=80;break a;break}case 7:{m=83;break a;break}case 8:{m=84;break a;break}case 9:{m=96;break a;break}case 10:{m=100;break a;break}case 11:{F=0;m=103;break a;break}case 12:{m=205;break a;break}case 13:{m=206;break a;break}case 14:{m=207;break a;break}case 15:{m=208;break a;break}case 17:{m=115;break a;break}case 18:{m=116;break a;break}case 19:{m=117;break a;break}case 20:{m=118;break a;break}case 21:{m=119;break a;break}case 22:{m=120;break a;break}case 23:{m=121;break a;break}case 24:{m=122;break a;break}case 25:{m=123;break a;break}case 28:{m=125;break a;break}case 16:{G=275;m=209;break a;break}case 26:{break b;break}case 0:{break f;break}case 27:break;default:{m=204;break a}}s=k[30808]|0;i[E>>0]=i[786736]|0;o=k[30839]|0;H=k[30838]|0;I=k[H+(o<<2)>>2]|0;if(!(k[I+44>>2]|0)){k[30841]=k[I+16>>2];k[I>>2]=k[30806];J=k[30838]|0;k[(k[J+(o<<2)>>2]|0)+44>>2]=1;K=k[30839]|0;L=k[J+(K<<2)>>2]|0;M=J;N=K}else{L=I;M=H;N=o}o=k[30842]|0;H=k[30841]|0;I=k[L+4>>2]|0;K=k[30808]|0;J=K;if(o>>>0<=(I+H|0)>>>0){O=s;P=K;Q=q;break e}if(o>>>0>(I+(H+1)|0)>>>0){m=147;break a}H=o-J|0;if(!(k[L+40>>2]|0)){R=(H|0)==1?1:2;S=J;T=K;U=M;V=N}else{J=H+-1|0;if((H|0)>1){o=I;I=0;W=K;while(1){i[o>>0]=i[W>>0]|0;I=I+1|0;if((I|0)>=(J|0))break;else{o=o+1|0;W=W+1|0}}W=k[30838]|0;X=k[W+(k[30839]<<2)>>2]|0;Y=W}else{X=L;Y=M}if((k[X+44>>2]|0)==2){k[30841]=0;k[X+16>>2]=0;Z=Y}else{W=(k[X+12>>2]|0)-J+-1|0;if((W|0)<1){o=Y;I=X;K=k[30842]|0;while(1){_=(o|0)==0?0:I;$=_+4|0;aa=k[$>>2]|0;if(!(k[_+20>>2]|0)){ba=$;m=158;break a}ca=_+12|0;_=k[ca>>2]|0;da=_<<1;ea=(da|0)<1?(_>>>3)+_|0:da;k[ca>>2]=ea;ca=fy(aa,ea+2|0)|0;k[$>>2]=ca;if(!ca){m=160;break a}K=ca+(K-aa)|0;k[30842]=K;o=k[30838]|0;aa=k[o+(k[30839]<<2)>>2]|0;ca=(k[aa+12>>2]|0)-J+-1|0;if((ca|0)>=1){fa=aa;ga=ca;break}else I=aa}}else{fa=X;ga=W}I=(ga|0)>8192?8192:ga;o=k[fa+4>>2]|0;if(!(k[30819]|0)){K=k[30818]|0;aa=k[30817]|0;ca=0;while(1){k[30818]=K+1;ha=ca+1|0;i[o+(ca+J)>>0]=i[aa+K>>0]|0;K=k[30818]|0;aa=k[30817]|0;ia=(i[aa+K>>0]|0)==0;if((ha|0)>=(I|0)){m=167;break}if(ia){ja=ha;ka=ca;m=168;break}else ca=ha}if((m|0)==167){m=0;if(ia){ja=ha;ka=ca;m=168}else{la=ha;ma=ca}}if((m|0)==168){m=0;k[30819]=1;la=ja;ma=ka}i[o+(ma+H)>>0]=0;k[30841]=la;if((ma|0)<-1){m=170;break a}else na=la}else{i[o+J>>0]=0;k[30841]=0;na=0}I=k[30838]|0;k[(k[I+(k[30839]<<2)>>2]|0)+16>>2]=na;Z=I}do if(!(k[30841]|0))if(!J){Bu(k[30806]|0);oa=k[30838]|0;pa=1;break}else{k[(k[Z+(k[30839]<<2)>>2]|0)+44>>2]=2;oa=Z;pa=2;break}else{oa=Z;pa=0}while(0);o=k[30841]|0;H=o+J|0;ca=k[oa+(k[30839]<<2)>>2]|0;if(H>>>0>(k[ca+12>>2]|0)>>>0){I=fy(k[ca+4>>2]|0,H+(o>>1)|0)|0;H=k[30839]|0;k[(k[(k[30838]|0)+(H<<2)>>2]|0)+4>>2]=I;I=k[(k[(k[30838]|0)+(H<<2)>>2]|0)+4>>2]|0;if(!I){m=180;break a}qa=k[30841]|0;ra=I}else{qa=o;ra=k[ca+4>>2]|0}ca=qa+J|0;k[30841]=ca;i[ra+ca>>0]=0;i[(k[(k[(k[30838]|0)+(k[30839]<<2)>>2]|0)+4>>2]|0)+((k[30841]|0)+1)>>0]=0;ca=k[30839]|0;o=k[30838]|0;I=k[(k[o+(ca<<2)>>2]|0)+4>>2]|0;k[30808]=I;R=pa;S=I;T=I;U=o;V=ca}switch(R|0){case 0:{sa=s;ta=T;ua=q;break d;break}case 2:{va=T;wa=U;xa=V;m=194;break e;break}case 1:break;default:continue a}k[30842]=S;f=(((k[30837]|0)+-1|0)/2|0)+28|0}i[E>>0]=i[786736]|0;v=k[31100]|0;n=k[31099]|0}if((m|0)==194){m=0;n=(k[(k[wa+(xa<<2)>>2]|0)+4>>2]|0)+(k[30841]|0)|0;k[30842]=n;v=k[30837]|0;if(va>>>0<n>>>0){ya=va;za=v}else{A=va;B=n;C=v;continue}while(1){v=i[ya>>0]|0;if(!(v<<24>>24))Aa=1;else Aa=k[123372+((v&255)<<2)>>2]|0;if(!(j[326324+(za<<1)>>1]|0)){Ba=Aa;Ca=za}else{k[31099]=za;k[31100]=ya;Ba=Aa;Ca=za}g:while(1){v=Ba&255;e=Ca;while(1){f=(j[326644+(e<<1)>>1]|0)+v|0;if((j[326978+(f<<1)>>1]|0)==(e|0)){Da=f;break g}f=j[327826+(e<<1)>>1]|0;q=f<<16>>16;if(f<<16>>16>159){Ea=q;break}else e=q}Ba=k[124404+(v<<2)>>2]|0;Ca=Ea}e=j[328160+(Da<<1)>>1]|0;s=ya+1|0;if((s|0)==(n|0)){A=va;B=n;C=e;continue d}else{ya=s;za=e}}}n=Q-O|0;B=P+(n+-1)|0;k[30842]=B;e=k[30837]|0;s=B;if((n|0)>1){J=P+(n+-1)|0;q=P;f=e;while(1){g=i[q>>0]|0;if(!(g<<24>>24))Fa=1;else Fa=k[123372+((g&255)<<2)>>2]|0;if(!(j[326324+(f<<1)>>1]|0)){Ga=Fa;Ha=f}else{k[31099]=f;k[31100]=q;Ga=Fa;Ha=f}h:while(1){g=Ga&255;ca=Ha;while(1){o=(j[326644+(ca<<1)>>1]|0)+g|0;if((j[326978+(o<<1)>>1]|0)==(ca|0)){Ia=o;break h}o=j[327826+(ca<<1)>>1]|0;I=o<<16>>16;if(o<<16>>16>159){Ja=I;break}else ca=I}Ga=k[124404+(g<<2)>>2]|0;Ha=Ja}ca=j[328160+(Ia<<1)>>1]|0;q=q+1|0;if((q|0)==(J|0)){Ka=ca;break}else f=ca}}else Ka=e;if(j[326324+(Ka<<1)>>1]|0){k[31099]=Ka;k[31100]=s}f=(j[326644+(Ka<<1)>>1]|0)+1|0;if((j[326978+(f<<1)>>1]|0)==(Ka|0))La=f;else{f=Ka;do{J=j[327826+(f<<1)>>1]|0;f=J<<16>>16;q=(j[326644+(f<<1)>>1]|0)+1|0}while((j[326978+(q<<1)>>1]|0)!=J<<16>>16);La=q}f=j[328160+(La<<1)>>1]|0;if(!(f<<16>>16==159|(La|0)==0)){Ma=P;Na=n;Oa=f;m=145;break}else{A=P;C=Ka}}if((m|0)==145){m=0;f=Ma+Na|0;k[30842]=f;u=Ma;t=f;h=Oa<<16>>16;continue}f=ua-sa|0;s=ta+(f+-1)|0;k[30842]=s;e=k[30837]|0;if((f|0)>1){Pa=ta;Qa=e}else{u=ta;t=s;h=e;continue}while(1){e=i[Pa>>0]|0;if(!(e<<24>>24))Ra=1;else Ra=k[123372+((e&255)<<2)>>2]|0;if(!(j[326324+(Qa<<1)>>1]|0)){Sa=Ra;Ta=Qa}else{k[31099]=Qa;k[31100]=Pa;Sa=Ra;Ta=Qa}i:while(1){e=Sa&255;f=Ta;while(1){q=(j[326644+(f<<1)>>1]|0)+e|0;if((j[326978+(q<<1)>>1]|0)==(f|0)){Ua=q;break i}q=j[327826+(f<<1)>>1]|0;J=q<<16>>16;if(q<<16>>16>159){Va=J;break}else f=J}Sa=k[124404+(e<<2)>>2]|0;Ta=Va}n=j[328160+(Ua<<1)>>1]|0;f=Pa+1|0;if(f>>>0<s>>>0){Pa=f;Qa=n}else{u=ta;t=s;h=n;continue b}}}Pw(k[30808]|0,k[30809]|0,1,k[30807]|0)|0}switch(m|0){case 34:{ta=k[30808]|0;Qa=xx(ta)|0;while(1){Pa=Qa+-1|0;if((i[ta+Qa>>0]|0)==32)Qa=Pa;else{Wa=Pa;break}}Bx(99104,ta+1|0,Wa)|0;i[99104+Wa>>0]=0;G=262;r=a;return G|0}case 37:{Wa=k[30808]|0;ta=xx(Wa)|0;if((ta|0)>255){k[30835]=431;Xa=d;Ya=786737;Za=Xa+37|0;do{i[Xa>>0]=i[Ya>>0]|0;Xa=Xa+1|0;Ya=Ya+1|0}while((Xa|0)<(Za|0));zx(d,Wa,20)|0;Qa=d+(xx(d)|0)|0;i[Qa>>0]=i[786774]|0;i[Qa+1>>0]=i[786775]|0;i[Qa+2>>0]=i[786776]|0;i[Qa+3>>0]=i[786777]|0;i[Qa+4>>0]=i[786778]|0;tk(d);_a=0}else{Qa=ta;while(1){ta=Qa+-1|0;if((i[Wa+Qa>>0]|0)==32)Qa=ta;else{$a=ta;break}}Bx(b,Wa+1|0,$a)|0;_a=$a}i[b+_a>>0]=0;i[c>>0]=0;_a=i[b>>0]|0;j:do if(_a<<24>>24){$a=_a;Wa=0;while(1){switch($a<<24>>24|0){case 48:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=48;i[Qa+1>>0]=48;i[Qa+2>>0]=48;i[Qa+3>>0]=0;break}case 49:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=48;i[Qa+1>>0]=48;i[Qa+2>>0]=49;i[Qa+3>>0]=0;break}case 50:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=48;i[Qa+1>>0]=49;i[Qa+2>>0]=48;i[Qa+3>>0]=0;break}case 51:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=48;i[Qa+1>>0]=49;i[Qa+2>>0]=49;i[Qa+3>>0]=0;break}case 52:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=49;i[Qa+1>>0]=48;i[Qa+2>>0]=48;i[Qa+3>>0]=0;break}case 53:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=49;i[Qa+1>>0]=48;i[Qa+2>>0]=49;i[Qa+3>>0]=0;break}case 54:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=49;i[Qa+1>>0]=49;i[Qa+2>>0]=48;i[Qa+3>>0]=0;break}case 55:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=49;i[Qa+1>>0]=49;i[Qa+2>>0]=49;i[Qa+3>>0]=0;break}case 88:case 120:{Qa=c+(xx(c)|0)|0;i[Qa>>0]=120;i[Qa+1>>0]=120;i[Qa+2>>0]=120;i[Qa+3>>0]=0;break}default:{}}Wa=Wa+1|0;$a=i[b+Wa>>0]|0;if(!($a<<24>>24))break j}}while(0);ux(99104,c)|0;G=262;r=a;return G|0}case 54:{_a=k[30808]|0;$a=xx(_a)|0;if(($a|0)>255){k[30835]=431;Xa=d;Ya=786779;Za=Xa+37|0;do{i[Xa>>0]=i[Ya>>0]|0;Xa=Xa+1|0;Ya=Ya+1|0}while((Xa|0)<(Za|0));zx(d,_a,20)|0;Wa=d+(xx(d)|0)|0;i[Wa>>0]=i[786774]|0;i[Wa+1>>0]=i[786775]|0;i[Wa+2>>0]=i[786776]|0;i[Wa+3>>0]=i[786777]|0;i[Wa+4>>0]=i[786778]|0;tk(d);ab=0}else{d=$a;while(1){$a=d+-1|0;if((i[_a+d>>0]|0)==32)d=$a;else{bb=$a;break}}Bx(b,_a+1|0,bb)|0;ab=bb}i[b+ab>>0]=0;i[c>>0]=0;ab=i[b>>0]|0;k:do if(ab<<24>>24){bb=ab;_a=0;while(1){do switch(bb<<24>>24|0){case 48:{d=c+(xx(c)|0)|0;i[d>>0]=i[786816]|0;i[d+1>>0]=i[786817]|0;i[d+2>>0]=i[786818]|0;i[d+3>>0]=i[786819]|0;i[d+4>>0]=i[786820]|0;break}case 49:{d=c+(xx(c)|0)|0;i[d>>0]=i[786821]|0;i[d+1>>0]=i[786822]|0;i[d+2>>0]=i[786823]|0;i[d+3>>0]=i[786824]|0;i[d+4>>0]=i[786825]|0;break}case 50:{d=c+(xx(c)|0)|0;i[d>>0]=i[786826]|0;i[d+1>>0]=i[786827]|0;i[d+2>>0]=i[786828]|0;i[d+3>>0]=i[786829]|0;i[d+4>>0]=i[786830]|0;break}case 51:{d=c+(xx(c)|0)|0;i[d>>0]=i[786831]|0;i[d+1>>0]=i[786832]|0;i[d+2>>0]=i[786833]|0;i[d+3>>0]=i[786834]|0;i[d+4>>0]=i[786835]|0;break}case 52:{d=c+(xx(c)|0)|0;i[d>>0]=i[786836]|0;i[d+1>>0]=i[786837]|0;i[d+2>>0]=i[786838]|0;i[d+3>>0]=i[786839]|0;i[d+4>>0]=i[786840]|0;break}case 53:{d=c+(xx(c)|0)|0;i[d>>0]=i[786841]|0;i[d+1>>0]=i[786842]|0;i[d+2>>0]=i[786843]|0;i[d+3>>0]=i[786844]|0;i[d+4>>0]=i[786845]|0;break}case 54:{d=c+(xx(c)|0)|0;i[d>>0]=i[786846]|0;i[d+1>>0]=i[786847]|0;i[d+2>>0]=i[786848]|0;i[d+3>>0]=i[786849]|0;i[d+4>>0]=i[786850]|0;break}case 55:{d=c+(xx(c)|0)|0;i[d>>0]=i[786851]|0;i[d+1>>0]=i[786852]|0;i[d+2>>0]=i[786853]|0;i[d+3>>0]=i[786854]|0;i[d+4>>0]=i[786855]|0;break}case 56:{d=c+(xx(c)|0)|0;i[d>>0]=i[786856]|0;i[d+1>>0]=i[786857]|0;i[d+2>>0]=i[786858]|0;i[d+3>>0]=i[786859]|0;i[d+4>>0]=i[786860]|0;break}case 57:{d=c+(xx(c)|0)|0;i[d>>0]=i[786861]|0;i[d+1>>0]=i[786862]|0;i[d+2>>0]=i[786863]|0;i[d+3>>0]=i[786864]|0;i[d+4>>0]=i[786865]|0;break}case 65:case 97:{d=c+(xx(c)|0)|0;i[d>>0]=i[786866]|0;i[d+1>>0]=i[786867]|0;i[d+2>>0]=i[786868]|0;i[d+3>>0]=i[786869]|0;i[d+4>>0]=i[786870]|0;break}case 66:case 98:{d=c+(xx(c)|0)|0;i[d>>0]=i[786871]|0;i[d+1>>0]=i[786872]|0;i[d+2>>0]=i[786873]|0;i[d+3>>0]=i[786874]|0;i[d+4>>0]=i[786875]|0;break}case 67:case 99:{d=c+(xx(c)|0)|0;i[d>>0]=i[786876]|0;i[d+1>>0]=i[786877]|0;i[d+2>>0]=i[786878]|0;i[d+3>>0]=i[786879]|0;i[d+4>>0]=i[786880]|0;break}case 68:case 100:{d=c+(xx(c)|0)|0;i[d>>0]=i[786881]|0;i[d+1>>0]=i[786882]|0;i[d+2>>0]=i[786883]|0;i[d+3>>0]=i[786884]|0;i[d+4>>0]=i[786885]|0;break}case 69:case 101:{d=c+(xx(c)|0)|0;i[d>>0]=i[786886]|0;i[d+1>>0]=i[786887]|0;i[d+2>>0]=i[786888]|0;i[d+3>>0]=i[786889]|0;i[d+4>>0]=i[786890]|0;break}case 70:case 102:{d=c+(xx(c)|0)|0;i[d>>0]=i[786891]|0;i[d+1>>0]=i[786892]|0;i[d+2>>0]=i[786893]|0;i[d+3>>0]=i[786894]|0;i[d+4>>0]=i[786895]|0;break}case 88:case 120:{d=c+(xx(c)|0)|0;i[d>>0]=i[786896]|0;i[d+1>>0]=i[786897]|0;i[d+2>>0]=i[786898]|0;i[d+3>>0]=i[786899]|0;i[d+4>>0]=i[786900]|0;break}default:{}}while(0);_a=_a+1|0;bb=i[b+_a>>0]|0;if(!(bb<<24>>24))break k}}while(0);ux(99104,c)|0;G=262;r=a;return G|0}case 79:{c=fx(k[30808]|0)|0;k[24776]=c;G=259;r=a;return G|0}case 80:{switch(i[k[30808]>>0]|0){case 84:case 116:{cb=1;break}default:cb=0}i[99104]=cb;G=258;r=a;return G|0}case 83:{db=+dx(k[30808]|0);p[12388]=db;G=260;r=a;return G|0}case 84:{cb=k[30808]|0;if(!(px(cb,786901)|0)){p[12388]=3.141592653589793;G=260;r=a;return G|0}if(!(px(cb,786905)|0)){p[12388]=2.718281828459045;G=260;r=a;return G|0}if(!(px(cb,786908)|0)){p[12388]=.017453292519943295;G=260;r=a;return G|0}if(!(px(cb,786913)|0)){G=272;r=a;return G|0}if(!(px(cb,786918)|0)){G=273;r=a;return G|0}if(!(px(cb,786924)|0)){G=274;r=a;return G|0}if((i[cb+1>>0]|0)==36){c=xx(cb)|0;i[99104]=35;Bx(99105,cb+2|0,c+-3|0)|0;i[99104+(c+-2)>>0]=0;k[30808]=99104;eb=99104}else eb=cb;G=Xb[k[123244>>2]&31](eb,99104)|0;r=a;return G|0}case 96:{eb=k[30808]|0;cb=(xx(eb)|0)+-2|0;if((cb|0)>255){k[30835]=431;Xa=b;Ya=786931;Za=Xa+33|0;do{i[Xa>>0]=i[Ya>>0]|0;Xa=Xa+1|0;Ya=Ya+1|0}while((Xa|0)<(Za|0));zx(b,eb+1|0,20)|0;Ya=b+(xx(b)|0)|0;i[Ya>>0]=i[786774]|0;i[Ya+1>>0]=i[786775]|0;i[Ya+2>>0]=i[786776]|0;i[Ya+3>>0]=i[786777]|0;i[Ya+4>>0]=i[786778]|0;tk(b);fb=0}else{Bx(99104,eb+1|0,cb)|0;fb=cb}i[99104+fb>>0]=0;G=261;r=a;return G|0}case 100:{fb=k[30808]|0;if((i[fb>>0]|0)==36){cb=(xx(fb)|0)+-2|0;Bx(99104,fb+1|0,cb)|0;i[99104+cb>>0]=0;k[30808]=99104;gb=99104}else gb=fb;G=Au(gb,99104)|0;r=a;return G|0}case 103:{while(1){m=0;gb=(nv(i[(k[30808]|0)+F>>0]|0)|0)&255;i[99104+F>>0]=gb;if(!(gb<<24>>24))break;else{F=F+1|0;m=103}}switch(i[99104]|0){case 66:{if(!(tx(99104,789244)|0)){G=264;r=a;return G|0}else m=110;break}case 67:{if(!(tx(99104,789178)|0)){G=264;r=a;return G|0}else m=110;break}case 69:{if(!(tx(99104,789249)|0)){G=264;r=a;return G|0}else m=110;break}case 78:{if(!(tx(99104,789091)|0)){G=264;r=a;return G|0}else m=110;break}case 73:{if(!(tx(99104,788956)|0)){G=264;r=a;return G|0}else m=113;break}case 71:{if(!(tx(99104,786964)|0)){G=266;r=a;return G|0}else m=113;break}case 82:{if(!(tx(99104,786975)|0)){G=267;r=a;return G|0}else m=113;break}case 83:{F=(tx(99104,788588)|0)==0;G=F?265:263;r=a;return G|0}default:{G=263;r=a;return G|0}}if((m|0)==110){G=263;r=a;return G|0}else if((m|0)==113){G=263;r=a;return G|0}break}case 115:{G=276;r=a;return G|0}case 116:{G=278;r=a;return G|0}case 117:{G=277;r=a;return G|0}case 118:{G=282;r=a;return G|0}case 119:{G=281;r=a;return G|0}case 120:{G=279;r=a;return G|0}case 121:{G=280;r=a;return G|0}case 122:{G=10;r=a;return G|0}case 123:{G=i[k[30808]>>0]|0;r=a;return G|0}case 125:{G=0;r=a;return G|0}case 147:{Cu(786986);break}case 158:{k[ba>>2]=0;Cu(787042);break}case 160:{Cu(787042);break}case 170:{Cu(787086);break}case 180:{Cu(787116);break}case 204:{Cu(787162);break}case 205:{G=286;r=a;return G|0}case 206:{G=285;r=a;return G|0}case 207:{G=283;r=a;return G|0}case 208:{G=284;r=a;return G|0}case 209:{r=a;return G|0}}return 0}function zu(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0;c=cy(48)|0;if(!c)Cu(787213);k[c+12>>2]=b;d=cy(b+2|0)|0;k[c+4>>2]=d;if(!d)Cu(787213);k[c+20>>2]=1;b=rv()|0;e=k[b>>2]|0;k[c+16>>2]=0;i[d>>0]=0;i[d+1>>0]=0;k[c+8>>2]=d;k[c+28>>2]=1;k[c+44>>2]=0;d=k[30838]|0;f=(d|0)==0;if(f)g=0;else g=k[d+(k[30839]<<2)>>2]|0;if((g|0)==(c|0)){g=d+(k[30839]<<2)|0;h=k[g>>2]|0;k[30841]=k[h+16>>2];j=k[h+8>>2]|0;k[30842]=j;k[30808]=j;k[30806]=k[k[g>>2]>>2];i[786736]=i[j>>0]|0}k[c>>2]=a;k[c+40>>2]=1;if(f)l=0;else l=k[d+(k[30839]<<2)>>2]|0;if((l|0)!=(c|0)){k[c+32>>2]=1;k[c+36>>2]=0}if(!a){m=0;n=c+24|0;k[n>>2]=m;k[b>>2]=e;return c|0}m=(Kx(Cw(a)|0)|0)>0&1;n=c+24|0;k[n>>2]=m;k[b>>2]=e;return c|0}function Au(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;c=r;r=r+112|0;d=c;e=k[30826]|0;a:do if((e|0)>0){f=k[30828]|0;g=0;while(1){if(!(yx(f+(g*124|0)|0,a,80)|0)){h=g;break}g=g+1|0;if((g|0)>=(e|0))break a}if((h|0)>=0){switch(k[f+(h*124|0)+84>>2]|0){case 260:case 259:{j=268;break}case 258:{j=269;break}case 261:{j=270;break}case 262:{j=271;break}default:{k[30835]=431;l=d;m=787279;n=l+24|0;do{i[l>>0]=i[m>>0]|0;l=l+1|0;m=m+1|0}while((l|0)<(n|0));zx(d,a,80)|0;tk(d);j=-1}}k[b>>2]=h;o=j;r=c;return o|0}}while(0);j=k[30811]|0;if(!j){k[30835]=431;l=d;m=787257;n=l+22|0;do{i[l>>0]=i[m>>0]|0;l=l+1|0;m=m+1|0}while((l|0)<(n|0));zx(d,a,80)|0;tk(d);o=-1;r=c;return o|0}else{o=Xb[j&31](a,b)|0;r=c;return o|0}return 0}function Bu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;b=k[30838]|0;do if(!b){c=cy(4)|0;k[30838]=c;if(!c)Cu(786687);else{k[c>>2]=0;k[30840]=1;k[30839]=0;d=10;break}}else{c=k[30839]|0;if(!(k[b+(c<<2)>>2]|0)){e=k[30840]|0;if(c>>>0>=(e+-1|0)>>>0){f=e+8|0;e=fy(b,f<<2)|0;k[30838]=e;if(!e)Cu(786687);else{g=e+(k[30840]<<2)|0;k[g>>2]=0;k[g+4>>2]=0;k[g+8>>2]=0;k[g+12>>2]=0;k[g+16>>2]=0;k[g+20>>2]=0;k[g+24>>2]=0;k[g+28>>2]=0;k[30840]=f;d=10;break}}else d=10}else{h=b;j=c;d=12}}while(0);if((d|0)==10){b=zu(k[30806]|0,16384)|0;c=k[30839]|0;k[(k[30838]|0)+(c<<2)>>2]=b;b=k[30838]|0;if(!b){f=rv()|0;l=0;m=k[f>>2]|0;n=f}else{h=b;j=c;d=12}}if((d|0)==12){d=k[h+(j<<2)>>2]|0;j=rv()|0;h=k[j>>2]|0;if(d){k[d+16>>2]=0;c=d+4|0;i[k[c>>2]>>0]=0;i[(k[c>>2]|0)+1>>0]=0;k[d+8>>2]=k[c>>2];k[d+28>>2]=1;k[d+44>>2]=0;c=k[30838]|0;if(!c)o=0;else o=k[c+(k[30839]<<2)>>2]|0;if((o|0)==(d|0)){o=c+(k[30839]<<2)|0;c=k[o>>2]|0;k[30841]=k[c+16>>2];b=k[c+8>>2]|0;k[30842]=b;k[30808]=b;k[30806]=k[k[o>>2]>>2];i[786736]=i[b>>0]|0;l=d;m=h;n=j}else{l=d;m=h;n=j}}else{l=0;m=h;n=j}}k[l>>2]=a;k[l+40>>2]=1;j=k[30838]|0;if(!j)p=0;else p=k[j+(k[30839]<<2)>>2]|0;if((p|0)!=(l|0)){k[l+32>>2]=1;k[l+36>>2]=0}if(!a){q=0;r=j;s=l+24|0;k[s>>2]=q;k[n>>2]=m;t=k[30839]|0;u=r+(t<<2)|0;v=k[u>>2]|0;w=v+16|0;x=k[w>>2]|0;k[30841]=x;y=v+8|0;z=k[y>>2]|0;k[30842]=z;k[30808]=z;A=k[u>>2]|0;B=k[A>>2]|0;k[30806]=B;C=z;D=i[C>>0]|0;i[786736]=D;return}j=(Kx(Cw(a)|0)|0)>0&1;q=j;r=k[30838]|0;s=l+24|0;k[s>>2]=q;k[n>>2]=m;t=k[30839]|0;u=r+(t<<2)|0;v=k[u>>2]|0;w=v+16|0;x=k[w>>2]|0;k[30841]=x;y=v+8|0;z=k[y>>2]|0;k[30842]=z;k[30808]=z;A=k[u>>2]|0;B=k[A>>2]|0;k[30806]=B;C=z;D=i[C>>0]|0;i[786736]=D;return}function Cu(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;b=k[31229]|0;k[c>>2]=a;Ew(b,787303,c)|0;Ib(2)}function Du(){var a=0,b=0,c=0,d=0,e=0,f=0,g=0,h=0,n=0,o=0,q=0,s=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,gc=0,hc=0,ic=0,jc=0,kc=0,lc=0,mc=0,nc=0,oc=0,pc=0,qc=0,rc=0,sc=0,tc=0,uc=0,vc=0,wc=0,xc=0,yc=0,zc=0,Ac=0,Bc=0,Cc=0,Dc=0,Ec=0,Fc=0,Gc=0,Hc=0,Ic=0,Jc=0,Kc=0,Lc=0,Mc=0,Nc=0,Oc=0,Pc=0,Qc=0,Rc=0,Sc=0,Tc=0,Uc=0,Vc=0,Wc=0,Xc=0,Yc=0,Zc=0,_c=0,$c=0,ad=0,bd=0,cd=0,dd=0,ed=0,fd=0,gd=0,hd=0,id=0,jd=0,kd=0,ld=0,md=0,nd=0,od=0,pd=0,qd=0,rd=0,sd=0,td=0,ud=0,vd=0,wd=0,xd=0,yd=0,zd=0,Ad=0,Bd=0,Cd=0,Dd=0,Ed=0,Fd=0,Gd=0,Hd=0,Id=0,Jd=0,Kd=0,Ld=0,Md=0,Nd=0,Od=0,Pd=0,Qd=0,Rd=0,Sd=0,Td=0,Ud=0,Vd=0,Wd=0,Xd=0,Yd=0,Zd=0,_d=0,$d=0,ae=0,be=0,ce=0,de=0,ee=0,fe=0,ge=0,he=0,ie=0,je=0,ke=0,le=0,me=0,ne=0,oe=0,pe=0,qe=0,re=0,se=0,te=0,ue=0,ve=0,we=0,xe=0,ye=0,ze=0,Ae=0,Be=0,Ce=0,De=0,Ee=0,Fe=0,Ge=0,He=0,Ie=0,Je=0,Ke=0,Le=0,Me=0,Ne=0,Oe=0,Pe=0,Qe=0,Re=0,Se=0,Te=0,Ue=0,Ve=0,We=0,Xe=0,Ye=0,Ze=0,_e=0,$e=0,af=0,bf=0,cf=0,df=0,ef=0,ff=0,gf=0,hf=0,jf=0,kf=0,lf=0,mf=0,nf=0,of=0,pf=0,qf=0,rf=0,sf=0,tf=0,uf=0,vf=0,wf=0,xf=0,yf=0,zf=0,Af=0,Bf=0,Cf=0,Df=0,Ef=0,Ff=0,Gf=0,Hf=0,If=0,Jf=0,Kf=0,Lf=0,Mf=0,Nf=0,Of=0,Pf=0,Qf=0,Rf=0,Sf=0,Tf=0,Uf=0,Vf=0,Wf=0,Xf=0,Yf=0,Zf=0,_f=0,$f=0,ag=0,bg=0,cg=0,dg=0,eg=0,fg=0,gg=0;a=r;r=r+26144|0;b=a+26056|0;c=a+25856|0;d=a;e=a+25600|0;k[31157]=0;k[31158]=-2;f=b+79|0;g=0;h=c;n=c;o=100;q=0;s=d;u=d;a:while(1){j[n>>1]=q;if((h+(o+-1<<1)|0)>>>0<=n>>>0){d=n-h>>1;v=d+1|0;if(o>>>0>9999){w=h;x=1309;break}y=o<<1;z=y>>>0>1e4?1e4:y;y=cy((z*258|0)+255|0)|0;if(!y){w=h;x=1309;break}py(y|0,h|0,v<<1|0)|0;A=y+(((z<<1)+255|0)>>>8<<8)|0;py(A|0,s|0,v<<8|0)|0;if((h|0)!=(c|0))dy(h);if((z+-1|0)>(d|0)){B=y;C=y+(d<<1)|0;D=z;E=A;F=A+(d<<8)|0}else{G=1;H=y;break}}else{B=h;C=n;D=o;E=s;F=u}if((q|0)==2){I=0;J=B;x=1312;break}y=j[329008+(q<<1)>>1]|0;d=y<<16>>16;if(y<<16>>16!=-46){y=k[31158]|0;if((y|0)==-2){A=yu()|0;k[31158]=A;K=A}else K=y;if((K|0)>=1)if(K>>>0<290)L=l[787307+K>>0]|0;else L=2;else{k[31158]=0;L=0}y=L+d|0;if(y>>>0<=1603?(j[329588+(y<<1)>>1]|0)==(L|0):0){d=j[332796+(y<<1)>>1]|0;if(!(d<<16>>16))x=1294;else{k[31158]=-2;y=F+256|0;py(y|0,99104,256)|0;M=(g|0)==0?0:g+-1|0;N=C;O=d&65535;P=y}}else x=21}else x=21;do if((x|0)==21){x=0;y=i[787597+q>>0]|0;d=y&255;if(!(y<<24>>24))x=1294;else{y=l[787887+d>>0]|0;A=F+(1-y<<8)|0;z=k[A>>2]|0;py(e|0,A+4|0,252)|0;b:do switch(d|0){case 125:{v=F+-1536|0;if((i[v>>0]|0)==83?(tx(v,789423)|0)==0:0){v=k[F+-768>>2]|0;Q=k[30820]|0;if((((k[Q+(v*344|0)+52>>2]|0)==259?(k[Q+(v*344|0)+56>>2]|0)==1:0)?(R=k[F+-256>>2]|0,(k[Q+(R*344|0)+52>>2]|0)==259):0)?(k[Q+(R*344|0)+56>>2]|0)==1:0){if((k[Q+(R*344|0)>>2]|0)==-1e3)S=Q+(R*344|0)+88|0;else S=Q+((k[F+-1280>>2]|0)*344|0)+56|0;Q=k[S>>2]|0;if((Q+-1|0)>>>0<=254){T=Mu(0,1044,3,k[F+-1280>>2]|0,v,R,0,0,0,0,Q)|0;if((T|0)<0){x=1301;break b}else{U=g;V=T;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,789507,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,789431,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,789538,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 5:{T=k[F+-256>>2]|0;if((T|0)>=0){k[30823]=T;U=g;V=z;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788013,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 6:{T=k[F+-256>>2]|0;if((T|0)>=0){k[30823]=T;U=g;V=z;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788013,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 7:{T=k[F+-256>>2]|0;if((T|0)>=0){k[30823]=T;U=g;V=z;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788013,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 8:{T=k[F+-256>>2]|0;if((T|0)>=0){k[30823]=T;U=g;V=z;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788013,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 9:{U=0;V=z;break}case 10:{T=k[F>>2]|0;Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){R=k[30820]|0;if(!R){k[30822]=100;W=cy(34400)|0}else{k[30822]=Q<<1;W=fy(R,Q*688|0)|0}if(!W){k[30835]=113;x=1301;break b}else{k[30820]=W;X=k[30821]|0;break}}else X=Q;while(0);k[30821]=X+1;if((X|0)>-1){Q=k[30820]|0;k[Q+(X*344|0)+52>>2]=k[Q+(T*344|0)+52>>2];k[Q+(X*344|0)+8>>2]=1;k[Q+(X*344|0)+12>>2]=T;k[Q+(X*344|0)>>2]=123;k[Q+(X*344|0)+4>>2]=4;U=g;V=X}else x=1301;break}case 11:{Q=F+-512|0;R=k[Q>>2]|0;v=k[30820]|0;Y=v+(R*344|0)+8|0;Z=k[Y>>2]|0;if((Z|0)>9){_=v+(R*344|0)+52|0;$=v;aa=Z;ba=0;ca=0;while(1){da=v+(R*344|0)+12+(ba<<2)|0;ea=k[da>>2]|0;fa=k[_>>2]|0;if((k[$+(ea*344|0)+52>>2]|0)==(fa|0)){ga=$;ha=ea;ia=aa}else{ja=Iu(fa,0,ea)|0;k[da>>2]=ja;if((ja|0)<0){x=57;break}ga=k[30820]|0;ha=ja;ia=k[Y>>2]|0}ja=(k[ga+(ha*344|0)+56>>2]|0)+ca|0;ba=ba+1|0;if((ba|0)>=(ia|0)){ka=ga;la=ja;break}else{$=ga;aa=ia;ca=ja}}if((x|0)==57){x=0;k[Q>>2]=-1;x=1301;break b}k[v+(R*344|0)+60>>2]=1;k[v+(R*344|0)+56>>2]=la;k[v+(R*344|0)+64>>2]=la;k[Q>>2]=R;if((R|0)<0){x=1301;break b}ca=k[30821]|0;do if((ca|0)==(k[30822]|0)){if(!ka){k[30822]=100;ma=cy(34400)|0}else{k[30822]=ca<<1;ma=fy(ka,ca*688|0)|0}if(!ma){k[30835]=113;x=1301;break b}else{k[30820]=ma;na=k[30821]|0;oa=ma;break}}else{na=ca;oa=ka}while(0);k[30821]=na+1;if((na|0)<=-1){x=1301;break b}k[oa+(na*344|0)+52>>2]=k[oa+(R*344|0)+52>>2];k[oa+(na*344|0)+8>>2]=1;k[oa+(na*344|0)+12>>2]=R;k[oa+(na*344|0)>>2]=123;k[oa+(na*344|0)+4>>2]=4;pa=oa;qa=k[oa+(na*344|0)+8>>2]|0;ra=na}else{pa=v;qa=Z;ra=R}ca=k[F>>2]|0;k[pa+(ra*344|0)+8>>2]=qa+1;k[pa+(ra*344|0)+12+(qa<<2)>>2]=ca;U=g;V=ra;break}case 12:{ca=k[F>>2]|0;Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){aa=k[30820]|0;if(!aa){k[30822]=100;sa=cy(34400)|0}else{k[30822]=Q<<1;sa=fy(aa,Q*688|0)|0}if(!sa){k[30835]=113;x=1301;break b}else{k[30820]=sa;ta=k[30821]|0;break}}else ta=Q;while(0);k[30821]=ta+1;if((ta|0)>-1){Q=k[30820]|0;k[Q+(ta*344|0)+52>>2]=k[Q+(ca*344|0)+52>>2];k[Q+(ta*344|0)+8>>2]=1;k[Q+(ta*344|0)+12>>2]=ca;k[Q+(ta*344|0)>>2]=123;k[Q+(ta*344|0)+4>>2]=4;U=g;V=ta}else x=1301;break}case 13:{Q=F+-512|0;R=k[Q>>2]|0;Z=k[30820]|0;v=Z+(R*344|0)+52|0;aa=k[Z+((k[F>>2]|0)*344|0)+52>>2]|0;if((k[v>>2]|0)<(aa|0)){k[v>>2]=aa;ua=k[Q>>2]|0}else ua=R;R=Z+(ua*344|0)+8|0;aa=k[R>>2]|0;if((aa|0)>9){v=Z+(ua*344|0)+52|0;$=Z;ba=aa;Y=0;_=0;while(1){T=Z+(ua*344|0)+12+(Y<<2)|0;ja=k[T>>2]|0;da=k[v>>2]|0;if((k[$+(ja*344|0)+52>>2]|0)==(da|0)){va=$;wa=ja;xa=ba}else{ea=Iu(da,0,ja)|0;k[T>>2]=ea;if((ea|0)<0){x=86;break}va=k[30820]|0;wa=ea;xa=k[R>>2]|0}ea=(k[va+(wa*344|0)+56>>2]|0)+_|0;Y=Y+1|0;if((Y|0)>=(xa|0)){ya=va;za=ea;break}else{$=va;ba=xa;_=ea}}if((x|0)==86){x=0;k[Q>>2]=-1;x=1301;break b}k[Z+(ua*344|0)+60>>2]=1;k[Z+(ua*344|0)+56>>2]=za;k[Z+(ua*344|0)+64>>2]=za;k[Q>>2]=ua;if((ua|0)<0){x=1301;break b}_=k[30821]|0;do if((_|0)==(k[30822]|0)){if(!ya){k[30822]=100;Aa=cy(34400)|0}else{k[30822]=_<<1;Aa=fy(ya,_*688|0)|0}if(!Aa){k[30835]=113;x=1301;break b}else{k[30820]=Aa;Ba=k[30821]|0;Ca=Aa;break}}else{Ba=_;Ca=ya}while(0);k[30821]=Ba+1;if((Ba|0)<=-1){x=1301;break b}k[Ca+(Ba*344|0)+52>>2]=k[Ca+(ua*344|0)+52>>2];k[Ca+(Ba*344|0)+8>>2]=1;k[Ca+(Ba*344|0)+12>>2]=ua;k[Ca+(Ba*344|0)>>2]=123;k[Ca+(Ba*344|0)+4>>2]=4;Da=Ca;Ea=k[Ca+(Ba*344|0)+8>>2]|0;Fa=Ba}else{Da=Z;Ea=aa;Fa=ua}_=k[F>>2]|0;k[Da+(Fa*344|0)+8>>2]=Ea+1;k[Da+(Fa*344|0)+12+(Ea<<2)>>2]=_;U=g;V=Fa;break}case 14:{_=F+-512|0;Q=k[_>>2]|0;ba=k[30820]|0;$=ba+(Q*344|0)+8|0;Y=k[$>>2]|0;if((Y|0)>9){R=ba+(Q*344|0)+52|0;v=ba;ca=Y;ea=0;T=0;while(1){ja=ba+(Q*344|0)+12+(ea<<2)|0;da=k[ja>>2]|0;fa=k[R>>2]|0;if((k[v+(da*344|0)+52>>2]|0)==(fa|0)){Ga=v;Ha=da;Ia=ca}else{Ja=Iu(fa,0,da)|0;k[ja>>2]=Ja;if((Ja|0)<0){x=104;break}Ga=k[30820]|0;Ha=Ja;Ia=k[$>>2]|0}Ja=(k[Ga+(Ha*344|0)+56>>2]|0)+T|0;ea=ea+1|0;if((ea|0)>=(Ia|0)){Ka=Ga;La=Ja;break}else{v=Ga;ca=Ia;T=Ja}}if((x|0)==104){x=0;k[_>>2]=-1;x=1301;break b}k[ba+(Q*344|0)+60>>2]=1;k[ba+(Q*344|0)+56>>2]=La;k[ba+(Q*344|0)+64>>2]=La;k[_>>2]=Q;if((Q|0)<0){x=1301;break b}T=k[30821]|0;do if((T|0)==(k[30822]|0)){if(!Ka){k[30822]=100;Ma=cy(34400)|0}else{k[30822]=T<<1;Ma=fy(Ka,T*688|0)|0}if(!Ma){k[30835]=113;x=1301;break b}else{k[30820]=Ma;Na=k[30821]|0;Oa=Ma;break}}else{Na=T;Oa=Ka}while(0);k[30821]=Na+1;if((Na|0)<=-1){x=1301;break b}k[Oa+(Na*344|0)+52>>2]=k[Oa+(Q*344|0)+52>>2];k[Oa+(Na*344|0)+8>>2]=1;k[Oa+(Na*344|0)+12>>2]=Q;k[Oa+(Na*344|0)>>2]=123;k[Oa+(Na*344|0)+4>>2]=4;Pa=Oa;Qa=k[Oa+(Na*344|0)+8>>2]|0;Ra=Na}else{Pa=ba;Qa=Y;Ra=Q}T=k[F>>2]|0;k[Pa+(Ra*344|0)+8>>2]=Qa+1;k[Pa+(Ra*344|0)+12+(Qa<<2)>>2]=T;U=g;V=Ra;break}case 15:{T=k[30820]|0;_=F+-512|0;k[T+((k[_>>2]|0)*344|0)+52>>2]=k[T+((k[F>>2]|0)*344|0)+52>>2];ca=k[_>>2]|0;v=T+(ca*344|0)+8|0;ea=k[v>>2]|0;if((ea|0)>9){$=T+(ca*344|0)+52|0;R=T;aa=ea;Z=0;Ja=0;while(1){ja=T+(ca*344|0)+12+(Z<<2)|0;da=k[ja>>2]|0;fa=k[$>>2]|0;if((k[R+(da*344|0)+52>>2]|0)==(fa|0)){Sa=R;Ta=da;Ua=aa}else{Va=Iu(fa,0,da)|0;k[ja>>2]=Va;if((Va|0)<0){x=122;break}Sa=k[30820]|0;Ta=Va;Ua=k[v>>2]|0}Va=(k[Sa+(Ta*344|0)+56>>2]|0)+Ja|0;Z=Z+1|0;if((Z|0)>=(Ua|0)){Wa=Sa;Xa=Va;break}else{R=Sa;aa=Ua;Ja=Va}}if((x|0)==122){x=0;k[_>>2]=-1;x=1301;break b}k[T+(ca*344|0)+60>>2]=1;k[T+(ca*344|0)+56>>2]=Xa;k[T+(ca*344|0)+64>>2]=Xa;k[_>>2]=ca;if((ca|0)<0){x=1301;break b}Ja=k[30821]|0;do if((Ja|0)==(k[30822]|0)){if(!Wa){k[30822]=100;Ya=cy(34400)|0}else{k[30822]=Ja<<1;Ya=fy(Wa,Ja*688|0)|0}if(!Ya){k[30835]=113;x=1301;break b}else{k[30820]=Ya;Za=k[30821]|0;_a=Ya;break}}else{Za=Ja;_a=Wa}while(0);k[30821]=Za+1;if((Za|0)<=-1){x=1301;break b}k[_a+(Za*344|0)+52>>2]=k[_a+(ca*344|0)+52>>2];k[_a+(Za*344|0)+8>>2]=1;k[_a+(Za*344|0)+12>>2]=ca;k[_a+(Za*344|0)>>2]=123;k[_a+(Za*344|0)+4>>2]=4;$a=_a;ab=k[_a+(Za*344|0)+8>>2]|0;bb=Za}else{$a=T;ab=ea;bb=ca}Ja=k[F>>2]|0;k[$a+(bb*344|0)+8>>2]=ab+1;k[$a+(bb*344|0)+12+(ab<<2)>>2]=Ja;U=g;V=bb;break}case 16:{Ja=k[F+-256>>2]|0;_=k[30820]|0;aa=_+(Ja*344|0)+8|0;R=k[aa>>2]|0;if((R|0)>0){Z=_+(Ja*344|0)+52|0;v=_;$=R;R=0;Q=0;while(1){Y=_+(Ja*344|0)+12+(R<<2)|0;ba=k[Y>>2]|0;Va=k[Z>>2]|0;if((k[v+(ba*344|0)+52>>2]|0)==(Va|0)){cb=v;db=ba;eb=$}else{ja=Iu(Va,0,ba)|0;k[Y>>2]=ja;if((ja|0)<0){x=1301;break b}cb=k[30820]|0;db=ja;eb=k[aa>>2]|0}ja=(k[cb+(db*344|0)+56>>2]|0)+Q|0;R=R+1|0;if((R|0)>=(eb|0)){fb=ja;break}else{v=cb;$=eb;Q=ja}}}else fb=0;k[_+(Ja*344|0)+60>>2]=1;k[_+(Ja*344|0)+56>>2]=fb;k[_+(Ja*344|0)+64>>2]=fb;if((Ja|0)<0)x=1301;else{U=g;V=Ja}break}case 17:{Q=k[F+-256>>2]|0;$=k[30820]|0;v=$+(Q*344|0)+8|0;R=k[v>>2]|0;if((R|0)>0){aa=$+(Q*344|0)+52|0;Z=$;ca=R;R=0;ea=0;while(1){T=$+(Q*344|0)+12+(R<<2)|0;ja=k[T>>2]|0;Y=k[aa>>2]|0;if((k[Z+(ja*344|0)+52>>2]|0)==(Y|0)){gb=Z;hb=ja;ib=ca}else{ba=Iu(Y,0,ja)|0;k[T>>2]=ba;if((ba|0)<0){x=1301;break b}gb=k[30820]|0;hb=ba;ib=k[v>>2]|0}ba=(k[gb+(hb*344|0)+56>>2]|0)+ea|0;R=R+1|0;if((R|0)>=(ib|0)){jb=ba;break}else{Z=gb;ca=ib;ea=ba}}}else jb=0;k[$+(Q*344|0)+60>>2]=1;k[$+(Q*344|0)+56>>2]=jb;k[$+(Q*344|0)+64>>2]=jb;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 18:{ea=(xx(F)|0)+1|0;ca=k[30821]|0;do if((ca|0)==(k[30822]|0)){Z=k[30820]|0;if(!Z){k[30822]=100;kb=cy(34400)|0}else{k[30822]=ca<<1;kb=fy(Z,ca*688|0)|0}if(!kb){k[30835]=113;x=1301;break b}else{k[30820]=kb;lb=k[30821]|0;break}}else lb=ca;while(0);k[30821]=lb+1;if((lb|0)>-1){ca=k[30820]|0;k[ca+(lb*344|0)>>2]=-1e3;k[ca+(lb*344|0)+4>>2]=0;k[ca+(lb*344|0)+8>>2]=0;k[ca+(lb*344|0)+52>>2]=262;py(ca+(lb*344|0)+88|0,F|0,ea|0)|0;k[ca+(lb*344|0)+84>>2]=0;k[ca+(lb*344|0)+56>>2]=1;k[ca+(lb*344|0)+60>>2]=1;k[ca+(lb*344|0)+64>>2]=1;ca=xx(F)|0;k[(k[30820]|0)+(lb*344|0)+56>>2]=ca;U=g;V=lb}else x=1301;break}case 19:{ca=k[F>>2]|0;Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){$=k[30820]|0;if(!$){k[30822]=100;mb=cy(34400)|0}else{k[30822]=Q<<1;mb=fy($,Q*688|0)|0}if(!mb){k[30835]=113;x=1301;break b}else{k[30820]=mb;nb=k[30821]|0;break}}else nb=Q;while(0);k[30821]=nb+1;if((nb|0)>-1){Q=k[30820]|0;k[Q+(nb*344|0)>>2]=0-ca;k[Q+(nb*344|0)+4>>2]=0;k[Q+(nb*344|0)+8>>2]=0;ea=k[30828]|0;k[Q+(nb*344|0)+52>>2]=k[ea+(ca*124|0)+84>>2];k[Q+(nb*344|0)+56>>2]=k[ea+(ca*124|0)+88>>2];$=k[ea+(ca*124|0)+92>>2]|0;k[Q+(nb*344|0)+60>>2]=$;if(($|0)>0){Z=0;do{k[Q+(nb*344|0)+64+(Z<<2)>>2]=k[ea+(ca*124|0)+96+(Z<<2)>>2];Z=Z+1|0}while((Z|0)!=($|0));if((nb|0)<0)x=1301;else{U=g;V=nb}}else{U=g;V=nb}}else x=1301;break}case 20:{$=k[F+-256>>2]|0;Z=k[30820]|0;if((k[Z+($*344|0)+52>>2]|0)==259?(k[Z+($*344|0)>>2]|0)==-1e3:0){ca=k[F+-768>>2]|0;ea=k[30821]|0;do if((ea|0)==(k[30822]|0)){k[30822]=ea<<1;Q=fy(Z,ea*688|0)|0;if(!Q){k[30835]=113;x=1301;break b}else{k[30820]=Q;ob=k[30821]|0;pb=Q;break}}else{ob=ea;pb=Z}while(0);Z=ob+1|0;k[30821]=Z;if((ob|0)<=-1){x=1301;break b}k[pb+(ob*344|0)>>2]=0-ca;k[pb+(ob*344|0)+4>>2]=0;k[pb+(ob*344|0)+8>>2]=0;ea=k[30828]|0;k[pb+(ob*344|0)+52>>2]=k[ea+(ca*124|0)+84>>2];k[pb+(ob*344|0)+56>>2]=k[ea+(ca*124|0)+88>>2];Q=k[ea+(ca*124|0)+92>>2]|0;k[pb+(ob*344|0)+60>>2]=Q;if((Q|0)>0){R=0;do{k[pb+(ob*344|0)+64+(R<<2)>>2]=k[ea+(ca*124|0)+96+(R<<2)>>2];R=R+1|0}while((R|0)!=(Q|0));if((ob|0)<0){x=1301;break b}}do if((Z|0)==(k[30822]|0)){if(!pb){k[30822]=100;qb=cy(34400)|0}else{k[30822]=Z<<1;qb=fy(pb,Z*688|0)|0}if(!qb){k[30835]=113;x=1301;break b}else{k[30820]=qb;rb=k[30821]|0;sb=qb;break}}else{rb=Z;sb=pb}while(0);k[30821]=rb+1;if((rb|0)<=-1){x=1301;break b}k[sb+(rb*344|0)>>2]=123;k[sb+(rb*344|0)+4>>2]=5;k[sb+(rb*344|0)+8>>2]=2;k[sb+(rb*344|0)+12>>2]=ob;k[sb+(rb*344|0)+16>>2]=$;Z=k[30828]|0;k[sb+(rb*344|0)+52>>2]=k[Z+(ca*124|0)+84>>2];k[sb+(rb*344|0)+56>>2]=k[Z+(ca*124|0)+88>>2];Q=k[Z+(ca*124|0)+92>>2]|0;k[sb+(rb*344|0)+60>>2]=Q;if((Q|0)>0)tb=0;else{U=g;V=rb;break b}do{k[sb+(rb*344|0)+64+(tb<<2)>>2]=k[Z+(ca*124|0)+96+(tb<<2)>>2];tb=tb+1|0}while((tb|0)!=(Q|0));if((rb|0)<0){x=1301;break b}else{U=g;V=rb;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788059,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 21:{Q=F+-512|0;ca=Ku(262,k[Q>>2]|0,38,k[F>>2]|0)|0;if((ca|0)<0)x=1301;else{Z=k[30820]|0;$=k[Z+((k[Q>>2]|0)*344|0)+56>>2]|0;Q=k[Z+((k[F>>2]|0)*344|0)+56>>2]|0;k[Z+(ca*344|0)+56>>2]=($|0)>(Q|0)?$:Q;U=g;V=ca}break}case 22:{ca=F+-512|0;Q=Ku(262,k[ca>>2]|0,124,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{$=k[30820]|0;Z=k[$+((k[ca>>2]|0)*344|0)+56>>2]|0;ca=k[$+((k[F>>2]|0)*344|0)+56>>2]|0;k[$+(Q*344|0)+56>>2]=(Z|0)>(ca|0)?Z:ca;U=g;V=Q}break}case 23:{Q=F+-512|0;ca=k[Q>>2]|0;Z=k[30820]|0;$=k[F>>2]|0;if(((k[Z+($*344|0)+56>>2]|0)+(k[Z+(ca*344|0)+56>>2]|0)|0)<=255){Z=Ku(262,ca,43,$)|0;if((Z|0)<0){x=1301;break b}$=k[30820]|0;k[$+(Z*344|0)+56>>2]=(k[$+((k[F>>2]|0)*344|0)+56>>2]|0)+(k[$+((k[Q>>2]|0)*344|0)+56>>2]|0);U=g;V=Z;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788102,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 24:{Z=Lu(k[F+-768>>2]|0,1,k[F+-256>>2]|0,0,0,0,0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 25:{Z=Lu(k[F+-1280>>2]|0,2,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0,0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 26:{Z=Lu(k[F+-1792>>2]|0,3,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 27:{Z=Lu(k[F+-2304>>2]|0,4,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 28:{Z=Lu(k[F+-2816>>2]|0,5,k[F+-2304>>2]|0,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 29:{Z=Iu(262,284,k[F>>2]|0)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 30:{U=g;V=k[F+-256>>2]|0;break}case 31:{Z=k[30821]|0;do if((Z|0)==(k[30822]|0)){Q=k[30820]|0;if(!Q){k[30822]=100;ub=cy(34400)|0}else{k[30822]=Z<<1;ub=fy(Q,Z*688|0)|0}if(!ub){k[30835]=113;x=1301;break b}else{k[30820]=ub;vb=k[30821]|0;break}}else vb=Z;while(0);k[30821]=vb+1;if((vb|0)>-1){Z=k[30820]|0;k[Z+(vb*344|0)>>2]=-1e3;k[Z+(vb*344|0)+4>>2]=0;k[Z+(vb*344|0)+8>>2]=0;k[Z+(vb*344|0)+52>>2]=259;Q=Z+(vb*344|0)+88|0;$=l[F>>0]|l[F+1>>0]<<8|l[F+2>>0]<<16|l[F+3>>0]<<24;i[Q>>0]=$;i[Q+1>>0]=$>>8;i[Q+2>>0]=$>>16;i[Q+3>>0]=$>>24;k[Z+(vb*344|0)+84>>2]=0;k[Z+(vb*344|0)+56>>2]=1;k[Z+(vb*344|0)+60>>2]=1;k[Z+(vb*344|0)+64>>2]=1;U=g;V=vb}else x=1301;break}case 32:{Z=k[30821]|0;do if((Z|0)==(k[30822]|0)){$=k[30820]|0;if(!$){k[30822]=100;wb=cy(34400)|0}else{k[30822]=Z<<1;wb=fy($,Z*688|0)|0}if(!wb){k[30835]=113;x=1301;break b}else{k[30820]=wb;xb=k[30821]|0;break}}else xb=Z;while(0);k[30821]=xb+1;if((xb|0)>-1){Z=k[30820]|0;k[Z+(xb*344|0)>>2]=-1e3;k[Z+(xb*344|0)+4>>2]=0;k[Z+(xb*344|0)+8>>2]=0;k[Z+(xb*344|0)+52>>2]=260;$=F;Q=$;ca=l[Q>>0]|l[Q+1>>0]<<8|l[Q+2>>0]<<16|l[Q+3>>0]<<24;Q=$+4|0;$=l[Q>>0]|l[Q+1>>0]<<8|l[Q+2>>0]<<16|l[Q+3>>0]<<24;Q=Z+(xb*344|0)+88|0;R=Q;i[R>>0]=ca;i[R+1>>0]=ca>>8;i[R+2>>0]=ca>>16;i[R+3>>0]=ca>>24;ca=Q+4|0;i[ca>>0]=$;i[ca+1>>0]=$>>8;i[ca+2>>0]=$>>16;i[ca+3>>0]=$>>24;k[Z+(xb*344|0)+84>>2]=0;k[Z+(xb*344|0)+56>>2]=1;k[Z+(xb*344|0)+60>>2]=1;k[Z+(xb*344|0)+64>>2]=1;U=g;V=xb}else x=1301;break}case 33:{Z=k[F>>2]|0;$=k[30821]|0;do if(($|0)==(k[30822]|0)){ca=k[30820]|0;if(!ca){k[30822]=100;yb=cy(34400)|0}else{k[30822]=$<<1;yb=fy(ca,$*688|0)|0}if(!yb){k[30835]=113;x=1301;break b}else{k[30820]=yb;zb=k[30821]|0;break}}else zb=$;while(0);k[30821]=zb+1;if((zb|0)>-1){$=k[30820]|0;k[$+(zb*344|0)>>2]=0-Z;k[$+(zb*344|0)+4>>2]=0;k[$+(zb*344|0)+8>>2]=0;ca=k[30828]|0;k[$+(zb*344|0)+52>>2]=k[ca+(Z*124|0)+84>>2];k[$+(zb*344|0)+56>>2]=k[ca+(Z*124|0)+88>>2];Q=k[ca+(Z*124|0)+92>>2]|0;k[$+(zb*344|0)+60>>2]=Q;if((Q|0)>0){R=0;do{k[$+(zb*344|0)+64+(R<<2)>>2]=k[ca+(Z*124|0)+96+(R<<2)>>2];R=R+1|0}while((R|0)!=(Q|0));if((zb|0)<0)x=1301;else{U=g;V=zb}}else{U=g;V=zb}}else x=1301;break}case 34:{Q=k[F+-256>>2]|0;R=k[30820]|0;if((k[R+(Q*344|0)+52>>2]|0)==259?(k[R+(Q*344|0)>>2]|0)==-1e3:0){Z=k[F+-768>>2]|0;ca=k[30821]|0;do if((ca|0)==(k[30822]|0)){k[30822]=ca<<1;$=fy(R,ca*688|0)|0;if(!$){k[30835]=113;x=1301;break b}else{k[30820]=$;Ab=k[30821]|0;Bb=$;break}}else{Ab=ca;Bb=R}while(0);R=Ab+1|0;k[30821]=R;if((Ab|0)<=-1){x=1301;break b}k[Bb+(Ab*344|0)>>2]=0-Z;k[Bb+(Ab*344|0)+4>>2]=0;k[Bb+(Ab*344|0)+8>>2]=0;ca=k[30828]|0;k[Bb+(Ab*344|0)+52>>2]=k[ca+(Z*124|0)+84>>2];k[Bb+(Ab*344|0)+56>>2]=k[ca+(Z*124|0)+88>>2];$=k[ca+(Z*124|0)+92>>2]|0;k[Bb+(Ab*344|0)+60>>2]=$;if(($|0)>0){ea=0;do{k[Bb+(Ab*344|0)+64+(ea<<2)>>2]=k[ca+(Z*124|0)+96+(ea<<2)>>2];ea=ea+1|0}while((ea|0)!=($|0));if((Ab|0)<0){x=1301;break b}}do if((R|0)==(k[30822]|0)){if(!Bb){k[30822]=100;Cb=cy(34400)|0}else{k[30822]=R<<1;Cb=fy(Bb,R*688|0)|0}if(!Cb){k[30835]=113;x=1301;break b}else{k[30820]=Cb;Db=k[30821]|0;Eb=Cb;break}}else{Db=R;Eb=Bb}while(0);k[30821]=Db+1;if((Db|0)<=-1){x=1301;break b}k[Eb+(Db*344|0)>>2]=123;k[Eb+(Db*344|0)+4>>2]=5;k[Eb+(Db*344|0)+8>>2]=2;k[Eb+(Db*344|0)+12>>2]=Ab;k[Eb+(Db*344|0)+16>>2]=Q;R=k[30828]|0;k[Eb+(Db*344|0)+52>>2]=k[R+(Z*124|0)+84>>2];k[Eb+(Db*344|0)+56>>2]=k[R+(Z*124|0)+88>>2];$=k[R+(Z*124|0)+92>>2]|0;k[Eb+(Db*344|0)+60>>2]=$;if(($|0)>0)Fb=0;else{U=g;V=Db;break b}do{k[Eb+(Db*344|0)+64+(Fb<<2)>>2]=k[R+(Z*124|0)+96+(Fb<<2)>>2];Fb=Fb+1|0}while((Fb|0)!=($|0));if((Db|0)<0){x=1301;break b}else{U=g;V=Db;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788059,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 35:{U=g;V=Mu(259,1035,0,0,0,0,0,0,0,0,0)|0;break}case 36:{U=g;V=Mu(259,1036,0,0,0,0,0,0,0,0,0)|0;break}case 37:{$=F+-512|0;Z=k[$>>2]|0;R=k[30820]|0;Q=k[R+(Z*344|0)+52>>2]|0;ea=k[F>>2]|0;ca=k[R+(ea*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){R=Iu(ca,0,Z)|0;k[$>>2]=R;Gb=R;Hb=k[F>>2]|0}else{Gb=Z;Hb=ea}else{Z=Iu(Q,0,ea)|0;k[F>>2]=Z;Gb=k[$>>2]|0;Hb=Z}Z=Ku(k[(k[30820]|0)+(Gb*344|0)+52>>2]|0,Gb,37,Hb)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 38:{Z=F+-512|0;$=k[Z>>2]|0;ea=k[30820]|0;Q=k[ea+($*344|0)+52>>2]|0;R=k[F>>2]|0;ca=k[ea+(R*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){ea=Iu(ca,0,$)|0;k[Z>>2]=ea;Ib=ea;Jb=k[F>>2]|0}else{Ib=$;Jb=R}else{$=Iu(Q,0,R)|0;k[F>>2]=$;Ib=k[Z>>2]|0;Jb=$}$=Ku(k[(k[30820]|0)+(Ib*344|0)+52>>2]|0,Ib,43,Jb)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 39:{$=F+-512|0;Z=k[$>>2]|0;R=k[30820]|0;Q=k[R+(Z*344|0)+52>>2]|0;ea=k[F>>2]|0;ca=k[R+(ea*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){R=Iu(ca,0,Z)|0;k[$>>2]=R;Kb=R;Lb=k[F>>2]|0}else{Kb=Z;Lb=ea}else{Z=Iu(Q,0,ea)|0;k[F>>2]=Z;Kb=k[$>>2]|0;Lb=Z}Z=Ku(k[(k[30820]|0)+(Kb*344|0)+52>>2]|0,Kb,45,Lb)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 40:{Z=F+-512|0;$=k[Z>>2]|0;ea=k[30820]|0;Q=k[ea+($*344|0)+52>>2]|0;R=k[F>>2]|0;ca=k[ea+(R*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){ea=Iu(ca,0,$)|0;k[Z>>2]=ea;Mb=ea;Nb=k[F>>2]|0}else{Mb=$;Nb=R}else{$=Iu(Q,0,R)|0;k[F>>2]=$;Mb=k[Z>>2]|0;Nb=$}$=Ku(k[(k[30820]|0)+(Mb*344|0)+52>>2]|0,Mb,42,Nb)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 41:{$=F+-512|0;Z=k[$>>2]|0;R=k[30820]|0;Q=k[R+(Z*344|0)+52>>2]|0;ea=k[F>>2]|0;ca=k[R+(ea*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){R=Iu(ca,0,Z)|0;k[$>>2]=R;Ob=R;Pb=k[F>>2]|0}else{Ob=Z;Pb=ea}else{Z=Iu(Q,0,ea)|0;k[F>>2]=Z;Ob=k[$>>2]|0;Pb=Z}Z=Ku(k[(k[30820]|0)+(Ob*344|0)+52>>2]|0,Ob,47,Pb)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 42:{Z=F+-512|0;$=k[Z>>2]|0;ea=k[30820]|0;Q=k[ea+($*344|0)+52>>2]|0;R=k[F>>2]|0;ca=k[ea+(R*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){ea=Iu(ca,0,$)|0;k[Z>>2]=ea;Qb=ea;Rb=k[F>>2]|0}else{Qb=$;Rb=R}else{$=Iu(Q,0,R)|0;k[F>>2]=$;Qb=k[Z>>2]|0;Rb=$}$=Ku(k[(k[30820]|0)+(Qb*344|0)+52>>2]|0,Qb,283,Rb)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 43:{U=g;V=k[F>>2]|0;break}case 44:{$=k[F>>2]|0;Z=Iu(k[(k[30820]|0)+($*344|0)+52>>2]|0,287,$)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 45:{U=g;V=k[F+-256>>2]|0;break}case 46:{Z=F+-512|0;$=Iu(k[(k[30820]|0)+((k[Z>>2]|0)*344|0)+52>>2]|0,0,k[F>>2]|0)|0;k[F>>2]=$;R=k[Z>>2]|0;Z=Ku(k[(k[30820]|0)+(R*344|0)+52>>2]|0,R,42,$)|0;if((Z|0)<0)x=1301;else{U=g;V=Z}break}case 47:{Z=F+-512|0;$=Iu(k[(k[30820]|0)+((k[F>>2]|0)*344|0)+52>>2]|0,0,k[Z>>2]|0)|0;k[Z>>2]=$;Z=k[F>>2]|0;R=Ku(k[(k[30820]|0)+(Z*344|0)+52>>2]|0,$,42,Z)|0;if((R|0)<0)x=1301;else{U=g;V=R}break}case 48:{R=F+-512|0;Z=k[R>>2]|0;$=k[30820]|0;Q=k[$+(Z*344|0)+52>>2]|0;ea=k[F>>2]|0;ca=k[$+(ea*344|0)+52>>2]|0;if((Q|0)<=(ca|0))if((Q|0)<(ca|0)){$=Iu(ca,0,Z)|0;k[R>>2]=$;Sb=k[F>>2]|0;Tb=$}else{Sb=ea;Tb=Z}else{Z=Iu(Q,0,ea)|0;k[F>>2]=Z;Sb=Z;Tb=k[R>>2]|0}do if((Sb|Tb|0)>=0){Z=k[30820]|0;ea=k[Z+(Tb*344|0)+56>>2]|0;if((ea|0)!=1?(Q=k[Z+(Sb*344|0)+56>>2]|0,(Q|0)!=1):0){if(!((ea|0)==(Q|0)?(k[Z+(Tb*344|0)+52>>2]|0)==(k[Z+(Sb*344|0)+52>>2]|0):0))break;Q=k[Z+(Tb*344|0)+60>>2]|0;if((Q|0)!=(k[Z+(Sb*344|0)+60>>2]|0))break;if((Q|0)>0){ea=0;$=1;while(1){ca=(k[Z+(Tb*344|0)+64+(ea<<2)>>2]|0)==(k[Z+(Sb*344|0)+64+(ea<<2)>>2]|0)?$:0;ea=ea+1|0;if((ea|0)==(Q|0)){Ub=ca;break}else $=ca}if(!Ub)break}}$=F+-1024|0;Q=Mu(0,1034,3,Tb,Sb,k[$>>2]|0,0,0,0,0,0)|0;if((Q|0)<0){x=1301;break b}ea=k[30820]|0;Z=k[F>>2]|0;ca=k[ea+(Z*344|0)+56>>2]|0;if(!((Z|Q|0)<0?1:(k[ea+((k[R>>2]|0)*344|0)+56>>2]|0)>=(ca|0))?(k[ea+(Q*344|0)+56>>2]=ca,ca=ea+(Z*344|0)+60|0,k[ea+(Q*344|0)+60>>2]=k[ca>>2],(k[ca>>2]|0)>0):0){v=0;do{k[ea+(Q*344|0)+64+(v<<2)>>2]=k[ea+(Z*344|0)+64+(v<<2)>>2];v=v+1|0}while((v|0)<(k[ca>>2]|0))}k[ea+((k[$>>2]|0)*344|0)+52>>2]=k[ea+((k[R>>2]|0)*344|0)+52>>2];ca=k[$>>2]|0;c:do if((ca|Q|0)>=0){v=k[ea+(ca*344|0)+56>>2]|0;Z=ea+(Q*344|0)+56|0;aa=k[Z>>2]|0;do if((v|0)!=1)if((aa|0)!=1){if(!((v|0)==(aa|0)?(k[ea+(ca*344|0)+52>>2]|0)==(k[ea+(Q*344|0)+52>>2]|0):0))break c;Ja=k[ea+(ca*344|0)+60>>2]|0;if((Ja|0)!=(k[ea+(Q*344|0)+60>>2]|0))break c;if((Ja|0)>0){Vb=0;Wb=1}else{Xb=v;break}while(1){_=(k[ea+(ca*344|0)+64+(Vb<<2)>>2]|0)==(k[ea+(Q*344|0)+64+(Vb<<2)>>2]|0)?Wb:0;Vb=Vb+1|0;if((Vb|0)==(Ja|0)){Yb=_;break}else Wb=_}if(!Yb)break c;else Xb=v}else Xb=1;else Xb=aa;while(0);k[ea+(ca*344|0)+52>>2]=258;aa=k[$>>2]|0;v=k[ea+(aa*344|0)+56>>2]|0;if((Xb|0)>=(v|0)|(aa|Q|0)<0){U=g;V=Q;break b}k[Z>>2]=v;v=ea+(aa*344|0)+60|0;k[ea+(Q*344|0)+60>>2]=k[v>>2];if((k[v>>2]|0)>0)Zb=0;else{U=g;V=Q;break b}while(1){k[ea+(Q*344|0)+64+(Zb<<2)>>2]=k[ea+(aa*344|0)+64+(Zb<<2)>>2];Zb=Zb+1|0;if((Zb|0)>=(k[v>>2]|0)){U=g;V=Q;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788186,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788144,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 49:{R=F+-512|0;Q=k[R>>2]|0;ea=k[30820]|0;$=k[ea+(Q*344|0)+52>>2]|0;ca=k[F>>2]|0;v=k[ea+(ca*344|0)+52>>2]|0;if(($|0)<=(v|0))if(($|0)<(v|0)){ea=Iu(v,0,Q)|0;k[R>>2]=ea;_b=k[F>>2]|0;$b=ea}else{_b=ca;$b=Q}else{Q=Iu($,0,ca)|0;k[F>>2]=Q;_b=Q;$b=k[R>>2]|0}do if((_b|$b|0)>=0){Q=k[30820]|0;ca=k[Q+($b*344|0)+56>>2]|0;if((ca|0)!=1?($=k[Q+(_b*344|0)+56>>2]|0,($|0)!=1):0){if(!((ca|0)==($|0)?(k[Q+($b*344|0)+52>>2]|0)==(k[Q+(_b*344|0)+52>>2]|0):0))break;$=k[Q+($b*344|0)+60>>2]|0;if(($|0)!=(k[Q+(_b*344|0)+60>>2]|0))break;if(($|0)>0){ca=0;ea=1;while(1){v=(k[Q+($b*344|0)+64+(ca<<2)>>2]|0)==(k[Q+(_b*344|0)+64+(ca<<2)>>2]|0)?ea:0;ca=ca+1|0;if((ca|0)==($|0)){ac=v;break}else ea=v}if(!ac)break}}ea=F+-1024|0;$=Mu(0,1034,3,$b,_b,k[ea>>2]|0,0,0,0,0,0)|0;if(($|0)<0){x=1301;break b}ca=k[30820]|0;Q=k[F>>2]|0;v=k[ca+(Q*344|0)+56>>2]|0;if(!((Q|$|0)<0?1:(k[ca+((k[R>>2]|0)*344|0)+56>>2]|0)>=(v|0))?(k[ca+($*344|0)+56>>2]=v,v=ca+(Q*344|0)+60|0,k[ca+($*344|0)+60>>2]=k[v>>2],(k[v>>2]|0)>0):0){aa=0;do{k[ca+($*344|0)+64+(aa<<2)>>2]=k[ca+(Q*344|0)+64+(aa<<2)>>2];aa=aa+1|0}while((aa|0)<(k[v>>2]|0))}k[ca+((k[ea>>2]|0)*344|0)+52>>2]=k[ca+((k[R>>2]|0)*344|0)+52>>2];v=k[ea>>2]|0;d:do if((v|$|0)>=0){aa=k[ca+(v*344|0)+56>>2]|0;Q=ca+($*344|0)+56|0;Z=k[Q>>2]|0;do if((aa|0)!=1)if((Z|0)!=1){if(!((aa|0)==(Z|0)?(k[ca+(v*344|0)+52>>2]|0)==(k[ca+($*344|0)+52>>2]|0):0))break d;Ja=k[ca+(v*344|0)+60>>2]|0;if((Ja|0)!=(k[ca+($*344|0)+60>>2]|0))break d;if((Ja|0)>0){bc=0;cc=1}else{dc=aa;break}while(1){_=(k[ca+(v*344|0)+64+(bc<<2)>>2]|0)==(k[ca+($*344|0)+64+(bc<<2)>>2]|0)?cc:0;bc=bc+1|0;if((bc|0)==(Ja|0)){ec=_;break}else cc=_}if(!ec)break d;else dc=aa}else dc=1;else dc=Z;while(0);k[ca+(v*344|0)+52>>2]=258;Z=k[ea>>2]|0;aa=k[ca+(Z*344|0)+56>>2]|0;if((dc|0)>=(aa|0)|(Z|$|0)<0){U=g;V=$;break b}k[Q>>2]=aa;aa=ca+(Z*344|0)+60|0;k[ca+($*344|0)+60>>2]=k[aa>>2];if((k[aa>>2]|0)>0)fc=0;else{U=g;V=$;break b}while(1){k[ca+($*344|0)+64+(fc<<2)>>2]=k[ca+(Z*344|0)+64+(fc<<2)>>2];fc=fc+1|0;if((fc|0)>=(k[aa>>2]|0)){U=g;V=$;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788186,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788144,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 50:{R=F+-512|0;$=k[R>>2]|0;ca=k[30820]|0;ea=k[ca+($*344|0)+52>>2]|0;v=k[F>>2]|0;aa=k[ca+(v*344|0)+52>>2]|0;if((ea|0)<=(aa|0))if((ea|0)<(aa|0)){ca=Iu(aa,0,$)|0;k[R>>2]=ca;gc=k[F>>2]|0;hc=ca}else{gc=v;hc=$}else{$=Iu(ea,0,v)|0;k[F>>2]=$;gc=$;hc=k[R>>2]|0}do if((gc|hc|0)>=0){$=k[30820]|0;v=k[$+(hc*344|0)+56>>2]|0;if((v|0)!=1?(ea=k[$+(gc*344|0)+56>>2]|0,(ea|0)!=1):0){if(!((v|0)==(ea|0)?(k[$+(hc*344|0)+52>>2]|0)==(k[$+(gc*344|0)+52>>2]|0):0))break;ea=k[$+(hc*344|0)+60>>2]|0;if((ea|0)!=(k[$+(gc*344|0)+60>>2]|0))break;if((ea|0)>0){v=0;ca=1;while(1){aa=(k[$+(hc*344|0)+64+(v<<2)>>2]|0)==(k[$+(gc*344|0)+64+(v<<2)>>2]|0)?ca:0;v=v+1|0;if((v|0)==(ea|0)){ic=aa;break}else ca=aa}if(!ic)break}}ca=F+-1024|0;ea=Mu(0,1034,3,hc,gc,k[ca>>2]|0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}v=k[30820]|0;$=k[F>>2]|0;aa=k[v+($*344|0)+56>>2]|0;if(!(($|ea|0)<0?1:(k[v+((k[R>>2]|0)*344|0)+56>>2]|0)>=(aa|0))?(k[v+(ea*344|0)+56>>2]=aa,aa=v+($*344|0)+60|0,k[v+(ea*344|0)+60>>2]=k[aa>>2],(k[aa>>2]|0)>0):0){Z=0;do{k[v+(ea*344|0)+64+(Z<<2)>>2]=k[v+($*344|0)+64+(Z<<2)>>2];Z=Z+1|0}while((Z|0)<(k[aa>>2]|0))}k[v+((k[ca>>2]|0)*344|0)+52>>2]=k[v+((k[R>>2]|0)*344|0)+52>>2];aa=k[ca>>2]|0;e:do if((aa|ea|0)>=0){Z=k[v+(aa*344|0)+56>>2]|0;$=v+(ea*344|0)+56|0;Q=k[$>>2]|0;do if((Z|0)!=1)if((Q|0)!=1){if(!((Z|0)==(Q|0)?(k[v+(aa*344|0)+52>>2]|0)==(k[v+(ea*344|0)+52>>2]|0):0))break e;Ja=k[v+(aa*344|0)+60>>2]|0;if((Ja|0)!=(k[v+(ea*344|0)+60>>2]|0))break e;if((Ja|0)>0){jc=0;kc=1}else{lc=Z;break}while(1){_=(k[v+(aa*344|0)+64+(jc<<2)>>2]|0)==(k[v+(ea*344|0)+64+(jc<<2)>>2]|0)?kc:0;jc=jc+1|0;if((jc|0)==(Ja|0)){mc=_;break}else kc=_}if(!mc)break e;else lc=Z}else lc=1;else lc=Q;while(0);k[v+(aa*344|0)+52>>2]=258;Q=k[ca>>2]|0;Z=k[v+(Q*344|0)+56>>2]|0;if((lc|0)>=(Z|0)|(Q|ea|0)<0){U=g;V=ea;break b}k[$>>2]=Z;Z=v+(Q*344|0)+60|0;k[v+(ea*344|0)+60>>2]=k[Z>>2];if((k[Z>>2]|0)>0)nc=0;else{U=g;V=ea;break b}while(1){k[v+(ea*344|0)+64+(nc<<2)>>2]=k[v+(Q*344|0)+64+(nc<<2)>>2];nc=nc+1|0;if((nc|0)>=(k[Z>>2]|0)){U=g;V=ea;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788186,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788144,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 51:{R=F+-256|0;do if((i[R>>0]|0)==82){if(!(tx(R,788228)|0))oc=Mu(260,1001,0,0,0,0,0,0,0,0,0)|0;else{if(tx(R,788236)|0)break;oc=Mu(260,1042,0,0,0,0,0,0,0,0,0)|0}if((oc|0)<0){x=1301;break b}else{U=g;V=oc;break b}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788245,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 52:{R=F+-512|0;ea=i[R>>0]|0;f:do if(ea<<24>>24<83){switch(ea<<24>>24){case 78:{if(tx(R,788275)|0)break f;v=k[30820]|0;ca=v+((k[F+-256>>2]|0)*344|0)+56|0;aa=k[30821]|0;do if((aa|0)==(k[30822]|0)){if(!v){k[30822]=100;pc=cy(34400)|0}else{k[30822]=aa<<1;pc=fy(v,aa*688|0)|0}if(!pc){k[30835]=113;x=1301;break b}else{k[30820]=pc;qc=k[30821]|0;rc=pc;break}}else{qc=aa;rc=v}while(0);k[30821]=qc+1;if((qc|0)<=-1){x=1301;break b}k[rc+(qc*344|0)>>2]=-1e3;k[rc+(qc*344|0)+4>>2]=0;k[rc+(qc*344|0)+8>>2]=0;k[rc+(qc*344|0)+52>>2]=259;v=rc+(qc*344|0)+88|0;aa=l[ca>>0]|l[ca+1>>0]<<8|l[ca+2>>0]<<16|l[ca+3>>0]<<24;i[v>>0]=aa;i[v+1>>0]=aa>>8;i[v+2>>0]=aa>>16;i[v+3>>0]=aa>>24;k[rc+(qc*344|0)+84>>2]=0;k[rc+(qc*344|0)+56>>2]=1;k[rc+(qc*344|0)+60>>2]=1;k[rc+(qc*344|0)+64>>2]=1;U=g;V=qc;break b;break}case 65:break;default:break f}if(!(tx(R,788282)|0)){aa=k[F+-256>>2]|0;v=k[30821]|0;do if((v|0)==(k[30822]|0)){Z=k[30820]|0;if(!Z){k[30822]=100;sc=cy(34400)|0}else{k[30822]=v<<1;sc=fy(Z,v*688|0)|0}if(!sc){k[30835]=113;tc=-1;break}else{k[30820]=sc;uc=k[30821]|0;x=432;break}}else{uc=v;x=432}while(0);if((x|0)==432){x=0;k[30821]=uc+1;if((uc|0)>-1){v=k[30820]|0;k[v+(uc*344|0)>>2]=-1e3;k[v+(uc*344|0)+4>>2]=0;k[v+(uc*344|0)+8>>2]=0;k[v+(uc*344|0)+52>>2]=259;ca=v+(uc*344|0)+88|0;i[ca>>0]=0;i[ca+1>>0]=0;i[ca+2>>0]=0;i[ca+3>>0]=0;k[v+(uc*344|0)+84>>2]=0;k[v+(uc*344|0)+56>>2]=1;k[v+(uc*344|0)+60>>2]=1;k[v+(uc*344|0)+64>>2]=1;tc=uc}else tc=uc}vc=Ku(259,aa,289,tc)|0;x=438}}else if(ea<<24>>24<=83?(tx(R,788270)|0)==0:0){vc=Mu(259,1002,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;x=438}while(0);if((x|0)==438){x=0;if((vc|0)<0){x=1301;break b}else{U=g;V=vc;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788289,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 53:{R=F+-512|0;do if((i[R>>0]|0)==78){if(tx(R,788275)|0){if(tx(R,788318)|0)break;ea=Mu(259,1040,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}else{U=g;V=ea;break b}}ea=k[30820]|0;v=ea+((k[F+-256>>2]|0)*344|0)+56|0;ca=k[30821]|0;do if((ca|0)==(k[30822]|0)){if(!ea){k[30822]=100;wc=cy(34400)|0}else{k[30822]=ca<<1;wc=fy(ea,ca*688|0)|0}if(!wc){k[30835]=113;x=1301;break b}else{k[30820]=wc;xc=k[30821]|0;yc=wc;break}}else{xc=ca;yc=ea}while(0);k[30821]=xc+1;if((xc|0)<=-1){x=1301;break b}k[yc+(xc*344|0)>>2]=-1e3;k[yc+(xc*344|0)+4>>2]=0;k[yc+(xc*344|0)+8>>2]=0;k[yc+(xc*344|0)+52>>2]=259;ea=yc+(xc*344|0)+88|0;ca=l[v>>0]|l[v+1>>0]<<8|l[v+2>>0]<<16|l[v+3>>0]<<24;i[ea>>0]=ca;i[ea+1>>0]=ca>>8;i[ea+2>>0]=ca>>16;i[ea+3>>0]=ca>>24;k[yc+(xc*344|0)+84>>2]=0;k[yc+(xc*344|0)+56>>2]=1;k[yc+(xc*344|0)+60>>2]=1;k[yc+(xc*344|0)+64>>2]=1;U=g;V=xc;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788326,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 54:{R=F+-512|0;ca=i[R>>0]|0;g:do if(ca<<24>>24<78){switch(ca<<24>>24){case 77:{if(!(tx(R,788354)|0)){ea=k[F+-256>>2]|0;aa=Mu(k[(k[30820]|0)+(ea*344|0)+52>>2]|0,1022,1,ea,0,0,0,0,0,0,0)|0;k[(k[30820]|0)+(aa*344|0)+56>>2]=1;zc=aa;x=499;break g}if(tx(R,788359)|0)break g;aa=k[F+-256>>2]|0;ea=Mu(k[(k[30820]|0)+(aa*344|0)+52>>2]|0,1024,1,aa,0,0,0,0,0,0,0)|0;k[(k[30820]|0)+(ea*344|0)+56>>2]=1;zc=ea;x=499;break g;break}case 65:break;default:break g}if(!(tx(R,788282)|0)){ea=k[F+-256>>2]|0;aa=k[30821]|0;do if((aa|0)==(k[30822]|0)){Z=k[30820]|0;if(!Z){k[30822]=100;Ac=cy(34400)|0}else{k[30822]=aa<<1;Ac=fy(Z,aa*688|0)|0}if(!Ac){k[30835]=113;Bc=-1;break}else{k[30820]=Ac;Cc=k[30821]|0;x=491;break}}else{Cc=aa;x=491}while(0);if((x|0)==491){x=0;k[30821]=Cc+1;if((Cc|0)>-1){aa=k[30820]|0;k[aa+(Cc*344|0)>>2]=-1e3;k[aa+(Cc*344|0)+4>>2]=0;k[aa+(Cc*344|0)+8>>2]=0;k[aa+(Cc*344|0)+52>>2]=259;v=aa+(Cc*344|0)+88|0;i[v>>0]=0;i[v+1>>0]=0;i[v+2>>0]=0;i[v+3>>0]=0;k[aa+(Cc*344|0)+84>>2]=0;k[aa+(Cc*344|0)+56>>2]=1;k[aa+(Cc*344|0)+60>>2]=1;k[aa+(Cc*344|0)+64>>2]=1;Bc=Cc}else Bc=Cc}zc=Ku(259,ea,289,Bc)|0;x=499}}else{if(ca<<24>>24>78){if(ca<<24>>24!=83)break;if(tx(R,788270)|0)break;zc=Mu(259,1002,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;x=499;break}if(!(tx(R,788275)|0)){aa=k[30820]|0;v=aa+((k[F+-256>>2]|0)*344|0)+56|0;Z=k[30821]|0;do if((Z|0)==(k[30822]|0)){if(!aa){k[30822]=100;Dc=cy(34400)|0}else{k[30822]=Z<<1;Dc=fy(aa,Z*688|0)|0}if(!Dc){k[30835]=113;x=1301;break b}else{k[30820]=Dc;Ec=k[30821]|0;Fc=Dc;break}}else{Ec=Z;Fc=aa}while(0);k[30821]=Ec+1;if((Ec|0)<=-1){x=1301;break b}k[Fc+(Ec*344|0)>>2]=-1e3;k[Fc+(Ec*344|0)+4>>2]=0;k[Fc+(Ec*344|0)+8>>2]=0;k[Fc+(Ec*344|0)+52>>2]=259;aa=Fc+(Ec*344|0)+88|0;Z=l[v>>0]|l[v+1>>0]<<8|l[v+2>>0]<<16|l[v+3>>0]<<24;i[aa>>0]=Z;i[aa+1>>0]=Z>>8;i[aa+2>>0]=Z>>16;i[aa+3>>0]=Z>>24;k[Fc+(Ec*344|0)+84>>2]=0;k[Fc+(Ec*344|0)+56>>2]=1;k[Fc+(Ec*344|0)+60>>2]=1;k[Fc+(Ec*344|0)+64>>2]=1;U=g;V=Ec;break b}if(!(tx(R,788318)|0)){Z=k[30820]|0;aa=Z+((k[F+-256>>2]|0)*344|0)+56|0;ea=k[30821]|0;do if((ea|0)==(k[30822]|0)){if(!Z){k[30822]=100;Gc=cy(34400)|0}else{k[30822]=ea<<1;Gc=fy(Z,ea*688|0)|0}if(!Gc){k[30835]=113;x=1301;break b}else{k[30820]=Gc;Hc=k[30821]|0;Ic=Gc;break}}else{Hc=ea;Ic=Z}while(0);k[30821]=Hc+1;if((Hc|0)<=-1){x=1301;break b}k[Ic+(Hc*344|0)>>2]=-1e3;k[Ic+(Hc*344|0)+4>>2]=0;k[Ic+(Hc*344|0)+8>>2]=0;k[Ic+(Hc*344|0)+52>>2]=259;Z=Ic+(Hc*344|0)+88|0;ea=l[aa>>0]|l[aa+1>>0]<<8|l[aa+2>>0]<<16|l[aa+3>>0]<<24;i[Z>>0]=ea;i[Z+1>>0]=ea>>8;i[Z+2>>0]=ea>>16;i[Z+3>>0]=ea>>24;k[Ic+(Hc*344|0)+84>>2]=0;k[Ic+(Hc*344|0)+56>>2]=1;k[Ic+(Hc*344|0)+60>>2]=1;k[Ic+(Hc*344|0)+64>>2]=1;U=g;V=Hc;break b}}while(0);if((x|0)==499){x=0;if((zc|0)<0){x=1301;break b}else{U=g;V=zc;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788364,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 55:{R=F+-512|0;ca=i[R>>0]|0;h:do if(ca<<24>>24<83){i:do switch(ca<<24>>24){case 65:{if(!(tx(R,788393)|0)){Jc=Mu(260,1038,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;break h}if((tx(R,788282)|0)==0?(ea=k[F+-256>>2]|0,Z=k[30820]|0,(k[Z+(ea*344|0)+52>>2]|0)==259):0){v=k[30821]|0;do if((v|0)==(k[30822]|0)){if(!Z){k[30822]=100;Kc=cy(34400)|0}else{k[30822]=v<<1;Kc=fy(Z,v*688|0)|0}if(!Kc){k[30835]=113;Lc=-1;break}else{k[30820]=Kc;Mc=k[30821]|0;Nc=Kc;x=534;break}}else{Mc=v;Nc=Z;x=534}while(0);if((x|0)==534){x=0;k[30821]=Mc+1;if((Mc|0)>-1){k[Nc+(Mc*344|0)>>2]=-1e3;k[Nc+(Mc*344|0)+4>>2]=0;k[Nc+(Mc*344|0)+8>>2]=0;k[Nc+(Mc*344|0)+52>>2]=259;Z=Nc+(Mc*344|0)+88|0;i[Z>>0]=0;i[Z+1>>0]=0;i[Z+2>>0]=0;i[Z+3>>0]=0;k[Nc+(Mc*344|0)+84>>2]=0;k[Nc+(Mc*344|0)+56>>2]=1;k[Nc+(Mc*344|0)+60>>2]=1;k[Nc+(Mc*344|0)+64>>2]=1;Lc=Mc}else Lc=Mc}Jc=Ku(259,ea,289,Lc)|0;break h}else Oc=1;break}case 77:{if(tx(R,788410)|0){x=581;break h}Z=k[F+-256>>2]|0;Jc=Mu(k[(k[30820]|0)+(Z*344|0)+52>>2]|0,1037,1,Z,0,0,0,0,0,0,0)|0;break h;break}default:{if(ca<<24>>24<78)if(ca<<24>>24<65){x=589;break h}else{Oc=0;break i}if(ca<<24>>24>78){x=585;break h}if(tx(R,788275)|0){if(tx(R,788318)|0){x=585;break h}Jc=Mu(259,1040,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;break h}Z=k[30820]|0;v=Z+((k[F+-256>>2]|0)*344|0)+56|0;Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){if(!Z){k[30822]=100;Pc=cy(34400)|0}else{k[30822]=Q<<1;Pc=fy(Z,Q*688|0)|0}if(!Pc){k[30835]=113;x=1301;break b}else{k[30820]=Pc;Qc=k[30821]|0;Rc=Pc;break}}else{Qc=Q;Rc=Z}while(0);k[30821]=Qc+1;if((Qc|0)<=-1){x=1301;break b}k[Rc+(Qc*344|0)>>2]=-1e3;k[Rc+(Qc*344|0)+4>>2]=0;k[Rc+(Qc*344|0)+8>>2]=0;k[Rc+(Qc*344|0)+52>>2]=259;Z=Rc+(Qc*344|0)+88|0;Q=l[v>>0]|l[v+1>>0]<<8|l[v+2>>0]<<16|l[v+3>>0]<<24;i[Z>>0]=Q;i[Z+1>>0]=Q>>8;i[Z+2>>0]=Q>>16;i[Z+3>>0]=Q>>24;k[Rc+(Qc*344|0)+84>>2]=0;k[Rc+(Qc*344|0)+56>>2]=1;k[Rc+(Qc*344|0)+60>>2]=1;k[Rc+(Qc*344|0)+64>>2]=1;U=g;V=Qc;break b}}while(0);if((ca<<24>>24<=65?(tx(R,788282)|0)==0:0)?(aa=k[F+-256>>2]|0,Q=k[30820]|0,(k[Q+(aa*344|0)+52>>2]|0)==260):0){Z=k[30821]|0;do if((Z|0)==(k[30822]|0)){if(!Q){k[30822]=100;Sc=cy(34400)|0}else{k[30822]=Z<<1;Sc=fy(Q,Z*688|0)|0}if(!Sc){k[30835]=113;Tc=-1;break}else{k[30820]=Sc;Uc=k[30821]|0;Vc=Sc;x=548;break}}else{Uc=Z;Vc=Q;x=548}while(0);if((x|0)==548){x=0;k[30821]=Uc+1;if((Uc|0)>-1){k[Vc+(Uc*344|0)>>2]=-1e3;k[Vc+(Uc*344|0)+4>>2]=0;k[Vc+(Uc*344|0)+8>>2]=0;k[Vc+(Uc*344|0)+52>>2]=260;Q=Vc+(Uc*344|0)+88|0;p[t>>3]=0.0;i[Q>>0]=i[t>>0];i[Q+1>>0]=i[t+1>>0];i[Q+2>>0]=i[t+2>>0];i[Q+3>>0]=i[t+3>>0];i[Q+4>>0]=i[t+4>>0];i[Q+5>>0]=i[t+5>>0];i[Q+6>>0]=i[t+6>>0];i[Q+7>>0]=i[t+7>>0];k[Vc+(Uc*344|0)+84>>2]=0;k[Vc+(Uc*344|0)+56>>2]=1;k[Vc+(Uc*344|0)+60>>2]=1;k[Vc+(Uc*344|0)+64>>2]=1;Tc=Uc}else Tc=Uc}Jc=Ku(260,aa,289,Tc)|0;break}if(ca<<24>>24==83){Wc=Oc;x=552}else{Xc=Oc;x=576}}else if(ca<<24>>24<=83){if(!(tx(R,788270)|0)){Q=k[F+-256>>2]|0;Jc=Mu(k[(k[30820]|0)+(Q*344|0)+52>>2]|0,1002,1,Q,0,0,0,0,0,0,0)|0;break}if(!(tx(R,788402)|0))Jc=Mu(260,1039,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;else{Wc=0;x=552}}else x=589;while(0);do if((x|0)==552){x=0;if((tx(R,788418)|0)==0?(Q=k[F+-256>>2]|0,Z=k[30820]|0,(k[Z+(Q*344|0)+52>>2]|0)==259):0){ea=k[30821]|0;do if((ea|0)==(k[30822]|0)){if(!Z){k[30822]=100;Yc=cy(34400)|0}else{k[30822]=ea<<1;Yc=fy(Z,ea*688|0)|0}if(!Yc){k[30835]=113;Zc=-1;break}else{k[30820]=Yc;_c=k[30821]|0;$c=Yc;x=561;break}}else{_c=ea;$c=Z;x=561}while(0);if((x|0)==561){x=0;k[30821]=_c+1;if((_c|0)>-1){k[$c+(_c*344|0)>>2]=-1e3;k[$c+(_c*344|0)+4>>2]=0;k[$c+(_c*344|0)+8>>2]=0;k[$c+(_c*344|0)+52>>2]=259;Z=$c+(_c*344|0)+88|0;i[Z>>0]=0;i[Z+1>>0]=0;i[Z+2>>0]=0;i[Z+3>>0]=0;k[$c+(_c*344|0)+84>>2]=0;k[$c+(_c*344|0)+56>>2]=1;k[$c+(_c*344|0)+60>>2]=1;k[$c+(_c*344|0)+64>>2]=1;Zc=_c}else Zc=_c}Jc=Ku(259,Q,288,Zc)|0;break}if((tx(R,788418)|0)==0?(Z=k[F+-256>>2]|0,ea=k[30820]|0,(k[ea+(Z*344|0)+52>>2]|0)==260):0){aa=k[30821]|0;do if((aa|0)==(k[30822]|0)){if(!ea){k[30822]=100;ad=cy(34400)|0}else{k[30822]=aa<<1;ad=fy(ea,aa*688|0)|0}if(!ad){k[30835]=113;bd=-1;break}else{k[30820]=ad;cd=k[30821]|0;dd=ad;x=573;break}}else{cd=aa;dd=ea;x=573}while(0);if((x|0)==573){x=0;k[30821]=cd+1;if((cd|0)>-1){k[dd+(cd*344|0)>>2]=-1e3;k[dd+(cd*344|0)+4>>2]=0;k[dd+(cd*344|0)+8>>2]=0;k[dd+(cd*344|0)+52>>2]=260;ea=dd+(cd*344|0)+88|0;p[t>>3]=0.0;i[ea>>0]=i[t>>0];i[ea+1>>0]=i[t+1>>0];i[ea+2>>0]=i[t+2>>0];i[ea+3>>0]=i[t+3>>0];i[ea+4>>0]=i[t+4>>0];i[ea+5>>0]=i[t+5>>0];i[ea+6>>0]=i[t+6>>0];i[ea+7>>0]=i[t+7>>0];k[dd+(cd*344|0)+84>>2]=0;k[dd+(cd*344|0)+56>>2]=1;k[dd+(cd*344|0)+60>>2]=1;k[dd+(cd*344|0)+64>>2]=1;bd=cd}else bd=cd}Jc=Ku(260,Z,288,bd)|0}else{Xc=Wc;x=576}}while(0);do if((x|0)==576){x=0;if(Xc?(tx(R,788427)|0)==0:0){Jc=Mu(0,1017,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;break}if(ca<<24>>24>=77)if(ca<<24>>24>77)x=585;else x=581;else x=589}while(0);do if((x|0)==581){x=0;if(!(tx(R,788354)|0)){ea=k[F+-256>>2]|0;Jc=Mu(k[(k[30820]|0)+(ea*344|0)+52>>2]|0,1022,1,ea,0,0,0,0,0,0,0)|0;break}if(!(tx(R,788359)|0)){ea=k[F+-256>>2]|0;Jc=Mu(k[(k[30820]|0)+(ea*344|0)+52>>2]|0,1024,1,ea,0,0,0,0,0,0,0)|0}else x=589}else if((x|0)==585){x=0;if(ca<<24>>24==82){if(!(tx(R,788228)|0)){ea=Mu(0,1001,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}k[(k[30820]|0)+(ea*344|0)+52>>2]=260;U=g;V=ea;break b}ea=(tx(R,788236)|0)==0;aa=F+-256|0;Q=k[aa>>2]|0;if(ea){ea=Mu(0,1042,1,Q,0,0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}k[(k[30820]|0)+(ea*344|0)+52>>2]=260;U=g;V=ea;break b}else{ed=Q;fd=aa;x=593}}else x=589}while(0);if((x|0)==589){x=0;aa=F+-256|0;ed=k[aa>>2]|0;fd=aa;x=593}j:do if((x|0)==593){x=0;if((k[(k[30820]|0)+(ed*344|0)+52>>2]|0)==260){gd=ca;hd=ed}else{aa=Iu(260,0,ed)|0;k[fd>>2]=aa;gd=i[R>>0]|0;hd=aa}aa=gd<<24>>24<83;k:do if(!aa){if(gd<<24>>24>83){if(gd<<24>>24>84){id=0;jd=0;x=620;break}if(tx(R,788442)|0){id=0;jd=0;x=620;break}Jc=Mu(0,1006,1,hd,0,0,0,0,0,0,0)|0;break j}if(!(tx(R,788432)|0)){Jc=Mu(0,1004,1,hd,0,0,0,0,0,0,0)|0;break j}if(!(tx(R,788489)|0)){Jc=Mu(0,1010,1,hd,0,0,0,0,0,0,0)|0;break j}else{id=0;jd=1;x=620}}else{switch(gd<<24>>24){case 67:{if(!(tx(R,788437)|0)){Jc=Mu(0,1005,1,hd,0,0,0,0,0,0,0)|0;break j}if(tx(R,788495)|0){id=1;jd=1;x=620;break k}Jc=Mu(0,1011,1,hd,0,0,0,0,0,0,0)|0;break j;break}case 65:{if(tx(R,788447)|0)x=608;break}default:{if(gd<<24>>24<65)break k;if(gd<<24>>24>65){kd=0;x=624;break k}else x=608}}if((x|0)==608?(x=0,(tx(R,788455)|0)!=0):0){if((tx(R,788461)|0)!=0?(tx(R,788469)|0)!=0:0){if((tx(R,788475)|0)!=0?(tx(R,788483)|0)!=0:0){id=0;jd=1;x=620;break}Jc=Mu(0,1009,1,hd,0,0,0,0,0,0,0)|0;break j}Jc=Mu(0,1008,1,hd,0,0,0,0,0,0,0)|0;break j}Jc=Mu(0,1007,1,hd,0,0,0,0,0,0,0)|0;break j}while(0);if((x|0)==620){x=0;if(!jd)if(gd<<24>>24<=84)if(!(tx(R,788501)|0)){Jc=Mu(0,1012,1,hd,0,0,0,0,0,0,0)|0;break}else{kd=id;x=624}else{ld=id;x=634}else{kd=id;x=624}}do if((x|0)==624){x=0;if(gd<<24>>24==69)if(tx(R,788507)|0)if(kd){x=646;break}else break;else{Jc=Mu(0,1013,1,hd,0,0,0,0,0,0,0)|0;break j}if(gd<<24>>24<76){if(gd<<24>>24==70?(tx(R,788537)|0)==0:0){Jc=Mu(0,1020,1,hd,0,0,0,0,0,0,0)|0;break j}if(kd){x=646;break}else{x=648;break}}if(gd<<24>>24<=76){if(!(tx(R,788512)|0)){Jc=Mu(0,1014,1,hd,0,0,0,0,0,0,0)|0;break j}if(!(tx(R,788517)|0)){Jc=Mu(0,1015,1,hd,0,0,0,0,0,0,0)|0;break j}else{ld=kd;x=634}}else{ld=kd;x=634}}while(0);if((x|0)==634){x=0;if(aa){if(gd<<24>>24==82?(tx(R,788530)|0)==0:0){Jc=Mu(0,1021,1,hd,0,0,0,0,0,0,0)|0;break}}else if(gd<<24>>24<=83?(tx(R,788524)|0)==0:0){Jc=Mu(0,1016,1,hd,0,0,0,0,0,0,0)|0;break}if(ld)x=646;else x=648}if((x|0)==646){x=0;if(!(tx(R,788544)|0)){Jc=Mu(0,1019,1,hd,0,0,0,0,0,0,0)|0;break}else x=648}if(((x|0)==648?(x=0,gd<<24>>24==82):0)?(tx(R,788550)|0)==0:0){Z=Mu(0,1043,1,hd,0,0,0,0,0,0,0)|0;k[(k[30820]|0)+(Z*344|0)+52>>2]=259;Jc=Z;break}if(!(k[30835]|0))k[30835]=431;Bx(b,788559,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if((Jc|0)<0)x=1301;else{U=g;V=Jc}break}case 56:{R=F+-1024|0;if((i[R>>0]|0)==83?(tx(R,788588)|0)==0:0){R=Mu(259,1045,2,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0,0,0,0,0)|0;if((R|0)<0)x=1301;else{U=g;V=R}}else{U=g;V=z}break}case 57:{R=F+-1024|0;ca=i[R>>0]|0;do if(ca<<24>>24<68){if(ca<<24>>24==65?(tx(R,788656)|0)==0:0){Z=F+-768|0;Q=k[Z>>2]|0;ea=k[30820]|0;if((k[ea+(Q*344|0)+52>>2]|0)==260){md=ea;nd=Q}else{ea=Iu(260,0,Q)|0;k[Z>>2]=ea;md=k[30820]|0;nd=ea}ea=F+-256|0;Q=k[ea>>2]|0;if((k[md+(Q*344|0)+52>>2]|0)==260){od=Q;pd=nd}else{$=Iu(260,0,Q)|0;k[ea>>2]=$;od=$;pd=k[Z>>2]|0}l:do if((od|pd|0)>=0){$=k[30820]|0;Q=k[$+(pd*344|0)+56>>2]|0;do if((Q|0)!=1?(Ja=k[$+(od*344|0)+56>>2]|0,(Ja|0)!=1):0){if(!((Q|0)==(Ja|0)?(k[$+(pd*344|0)+52>>2]|0)==(k[$+(od*344|0)+52>>2]|0):0))break l;Ja=k[$+(pd*344|0)+60>>2]|0;if((Ja|0)!=(k[$+(od*344|0)+60>>2]|0))break l;if((Ja|0)>0){qd=0;rd=1}else break;while(1){_=(k[$+(pd*344|0)+64+(qd<<2)>>2]|0)==(k[$+(od*344|0)+64+(qd<<2)>>2]|0)?rd:0;qd=qd+1|0;if((qd|0)==(Ja|0)){sd=_;break}else rd=_}if(!sd)break l}while(0);$=Mu(0,1018,2,pd,od,0,0,0,0,0,0)|0;if(($|0)<0){x=1301;break b}Q=k[30820]|0;v=k[ea>>2]|0;Ja=k[Q+(v*344|0)+56>>2]|0;if((v|$|0)<0?1:(k[Q+((k[Z>>2]|0)*344|0)+56>>2]|0)>=(Ja|0)){U=g;V=$;break b}k[Q+($*344|0)+56>>2]=Ja;Ja=Q+(v*344|0)+60|0;k[Q+($*344|0)+60>>2]=k[Ja>>2];if((k[Ja>>2]|0)>0)td=0;else{U=g;V=$;break b}while(1){k[Q+($*344|0)+64+(td<<2)>>2]=k[Q+(v*344|0)+64+(td<<2)>>2];td=td+1|0;if((td|0)>=(k[Ja>>2]|0)){U=g;V=$;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788665,80)|0;i[f>>0]=0;tk(b);x=1301;break b}}else{if(ca<<24>>24<=68){if(tx(R,788596)|0)break;Z=F+-768|0;ea=k[Z>>2]|0;aa=k[30820]|0;$=k[aa+(ea*344|0)+56>>2]|0;Ja=F+-256|0;v=k[Ja>>2]|0;Q=k[aa+(v*344|0)+56>>2]|0;m:do if(!(($|0)<(Q|0)|(v|ea|0)<0)){_=k[aa+(ea*344|0)+52>>2]|0;ba=k[aa+(v*344|0)+52>>2]|0;do if(($|0)==1|(Q|0)==1){if((_|0)>(ba|0)){T=Iu(_,0,v)|0;k[Ja>>2]=T;ud=k[Z>>2]|0;vd=T;break}if((_|0)<(ba|0)){T=Iu(ba,0,ea)|0;k[Z>>2]=T;ud=T;vd=k[Ja>>2]|0}else{ud=ea;vd=v}}else{if(!(($|0)==(Q|0)&(_|0)==(ba|0)))break m;T=k[aa+(ea*344|0)+60>>2]|0;if((T|0)!=(k[aa+(v*344|0)+60>>2]|0))break m;if((T|0)>0){ja=0;Y=1;while(1){Va=(k[aa+(ea*344|0)+64+(ja<<2)>>2]|0)==(k[aa+(v*344|0)+64+(ja<<2)>>2]|0)?Y:0;ja=ja+1|0;if((ja|0)==(T|0)){wd=Va;break}else Y=Va}if(!wd)break m;else{ud=ea;vd=v}}else{ud=ea;vd=v}}while(0);ba=Mu(0,1031,2,ud,vd,0,0,0,0,0,0)|0;if((ba|0)<0){x=1301;break b}else{U=g;V=ba;break b}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788605,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(ca<<24>>24==77){if(!(tx(R,788354)|0)){v=F+-768|0;ea=k[v>>2]|0;aa=k[30820]|0;Q=k[aa+(ea*344|0)+52>>2]|0;$=F+-256|0;Ja=k[$>>2]|0;Z=k[aa+(Ja*344|0)+52>>2]|0;if((Q|0)<=(Z|0))if((Q|0)<(Z|0)){aa=Iu(Z,0,ea)|0;k[v>>2]=aa;xd=k[$>>2]|0;yd=aa}else{xd=Ja;yd=ea}else{ea=Iu(Q,0,Ja)|0;k[$>>2]=ea;xd=ea;yd=k[v>>2]|0}n:do if((xd|yd|0)>=0){ea=k[30820]|0;Ja=k[ea+(yd*344|0)+56>>2]|0;do if((Ja|0)!=1?(Q=k[ea+(xd*344|0)+56>>2]|0,(Q|0)!=1):0){if(!((Ja|0)==(Q|0)?(k[ea+(yd*344|0)+52>>2]|0)==(k[ea+(xd*344|0)+52>>2]|0):0))break n;Q=k[ea+(yd*344|0)+60>>2]|0;if((Q|0)!=(k[ea+(xd*344|0)+60>>2]|0))break n;if((Q|0)>0){zd=0;Ad=1}else break;while(1){aa=(k[ea+(yd*344|0)+64+(zd<<2)>>2]|0)==(k[ea+(xd*344|0)+64+(zd<<2)>>2]|0)?Ad:0;zd=zd+1|0;if((zd|0)==(Q|0)){Bd=aa;break}else Ad=aa}if(!Bd)break n}while(0);ea=Mu(0,1023,2,yd,xd,0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}Ja=k[30820]|0;Q=k[$>>2]|0;aa=k[Ja+(Q*344|0)+56>>2]|0;if((Q|ea|0)<0?1:(k[Ja+((k[v>>2]|0)*344|0)+56>>2]|0)>=(aa|0)){U=g;V=ea;break b}k[Ja+(ea*344|0)+56>>2]=aa;aa=Ja+(Q*344|0)+60|0;k[Ja+(ea*344|0)+60>>2]=k[aa>>2];if((k[aa>>2]|0)>0)Cd=0;else{U=g;V=ea;break b}while(1){k[Ja+(ea*344|0)+64+(Cd<<2)>>2]=k[Ja+(Q*344|0)+64+(Cd<<2)>>2];Cd=Cd+1|0;if((Cd|0)>=(k[aa>>2]|0)){U=g;V=ea;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788716,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(!(tx(R,788359)|0)){v=F+-768|0;$=k[v>>2]|0;ea=k[30820]|0;aa=k[ea+($*344|0)+52>>2]|0;Q=F+-256|0;Ja=k[Q>>2]|0;Z=k[ea+(Ja*344|0)+52>>2]|0;if((aa|0)<=(Z|0))if((aa|0)<(Z|0)){ea=Iu(Z,0,$)|0;k[v>>2]=ea;Dd=k[Q>>2]|0;Ed=ea}else{Dd=Ja;Ed=$}else{$=Iu(aa,0,Ja)|0;k[Q>>2]=$;Dd=$;Ed=k[v>>2]|0}o:do if((Dd|Ed|0)>=0){$=k[30820]|0;Ja=k[$+(Ed*344|0)+56>>2]|0;do if((Ja|0)!=1?(aa=k[$+(Dd*344|0)+56>>2]|0,(aa|0)!=1):0){if(!((Ja|0)==(aa|0)?(k[$+(Ed*344|0)+52>>2]|0)==(k[$+(Dd*344|0)+52>>2]|0):0))break o;aa=k[$+(Ed*344|0)+60>>2]|0;if((aa|0)!=(k[$+(Dd*344|0)+60>>2]|0))break o;if((aa|0)>0){Fd=0;Gd=1}else break;while(1){ea=(k[$+(Ed*344|0)+64+(Fd<<2)>>2]|0)==(k[$+(Dd*344|0)+64+(Fd<<2)>>2]|0)?Gd:0;Fd=Fd+1|0;if((Fd|0)==(aa|0)){Hd=ea;break}else Gd=ea}if(!Hd)break o}while(0);$=Mu(0,1025,2,Ed,Dd,0,0,0,0,0,0)|0;if(($|0)<0){x=1301;break b}Ja=k[30820]|0;aa=k[Q>>2]|0;ea=k[Ja+(aa*344|0)+56>>2]|0;if((aa|$|0)<0?1:(k[Ja+((k[v>>2]|0)*344|0)+56>>2]|0)>=(ea|0)){U=g;V=$;break b}k[Ja+($*344|0)+56>>2]=ea;ea=Ja+(aa*344|0)+60|0;k[Ja+($*344|0)+60>>2]=k[ea>>2];if((k[ea>>2]|0)>0)Id=0;else{U=g;V=$;break b}while(1){k[Ja+($*344|0)+64+(Id<<2)>>2]=k[Ja+(aa*344|0)+64+(Id<<2)>>2];Id=Id+1|0;if((Id|0)>=(k[ea>>2]|0)){U=g;V=$;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788768,80)|0;i[f>>0]=0;tk(b);x=1301;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788820,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 58:{R=F+-2048|0;if((i[R>>0]|0)==65?(tx(R,788854)|0)==0:0){R=F+-1792|0;ca=k[R>>2]|0;v=k[30820]|0;if((k[v+(ca*344|0)+52>>2]|0)==260)Jd=v;else{v=Iu(260,0,ca)|0;k[R>>2]=v;Jd=k[30820]|0}v=F+-1280|0;ca=k[v>>2]|0;if((k[Jd+(ca*344|0)+52>>2]|0)==260)Kd=Jd;else{Q=Iu(260,0,ca)|0;k[v>>2]=Q;Kd=k[30820]|0}Q=F+-768|0;ca=k[Q>>2]|0;if((k[Kd+(ca*344|0)+52>>2]|0)==260)Ld=Kd;else{$=Iu(260,0,ca)|0;k[Q>>2]=$;Ld=k[30820]|0}$=F+-256|0;ca=k[$>>2]|0;if((k[Ld+(ca*344|0)+52>>2]|0)==260)Md=ca;else{ea=Iu(260,0,ca)|0;k[$>>2]=ea;Md=ea}ea=k[R>>2]|0;ca=k[v>>2]|0;p:do if((ca|ea|0)>=0){aa=k[30820]|0;Ja=k[aa+(ea*344|0)+56>>2]|0;do if((Ja|0)!=1?(Z=k[aa+(ca*344|0)+56>>2]|0,(Z|0)!=1):0){if(!((Ja|0)==(Z|0)?(k[aa+(ea*344|0)+52>>2]|0)==(k[aa+(ca*344|0)+52>>2]|0):0))break p;Z=k[aa+(ea*344|0)+60>>2]|0;if((Z|0)!=(k[aa+(ca*344|0)+60>>2]|0))break p;if((Z|0)>0){Nd=0;Od=1}else break;while(1){ba=(k[aa+(ea*344|0)+64+(Nd<<2)>>2]|0)==(k[aa+(ca*344|0)+64+(Nd<<2)>>2]|0)?Od:0;Nd=Nd+1|0;if((Nd|0)==(Z|0)){Pd=ba;break}else Od=ba}if(!Pd)break p}while(0);Ja=k[Q>>2]|0;if((Ja|ca|0)>=0){Z=k[aa+(ca*344|0)+56>>2]|0;do if((Z|0)!=1){ba=k[aa+(Ja*344|0)+56>>2]|0;if((ba|0)==1){x=767;break}if(!((Z|0)==(ba|0)?(k[aa+(ca*344|0)+52>>2]|0)==(k[aa+(Ja*344|0)+52>>2]|0):0))break p;ba=k[aa+(ca*344|0)+60>>2]|0;if((ba|0)!=(k[aa+(Ja*344|0)+60>>2]|0))break p;if((ba|0)>0){Qd=0;Rd=1}else{x=767;break}while(1){_=(k[aa+(ca*344|0)+64+(Qd<<2)>>2]|0)==(k[aa+(Ja*344|0)+64+(Qd<<2)>>2]|0)?Rd:0;Qd=Qd+1|0;if((Qd|0)==(ba|0)){Sd=_;break}else Rd=_}if((Sd|0)==0|(Md|Ja|0)<0)break p}else x=767;while(0);if((x|0)==767?(x=0,(Md|Ja|0)<0):0)break;Z=k[aa+(Ja*344|0)+56>>2]|0;do if((Z|0)!=1){ba=k[aa+(Md*344|0)+56>>2]|0;if((ba|0)==1)break;if(!((Z|0)==(ba|0)?(k[aa+(Ja*344|0)+52>>2]|0)==(k[aa+(Md*344|0)+52>>2]|0):0))break p;ba=k[aa+(Ja*344|0)+60>>2]|0;if((ba|0)!=(k[aa+(Md*344|0)+60>>2]|0))break p;if((ba|0)>0){Td=0;Ud=1}else break;while(1){_=(k[aa+(Ja*344|0)+64+(Td<<2)>>2]|0)==(k[aa+(Md*344|0)+64+(Td<<2)>>2]|0)?Ud:0;Td=Td+1|0;if((Td|0)==(ba|0)){Vd=_;break}else Ud=_}if(!Vd)break p}while(0);aa=Mu(0,1041,4,ea,ca,Ja,Md,0,0,0,0)|0;if((aa|0)<0){x=1301;break b}Z=k[30820]|0;ba=k[v>>2]|0;_=k[Z+(ba*344|0)+56>>2]|0;do if(!((ba|aa|0)<0?1:(k[Z+((k[R>>2]|0)*344|0)+56>>2]|0)>=(_|0))){k[Z+(aa*344|0)+56>>2]=_;Y=Z+(ba*344|0)+60|0;k[Z+(aa*344|0)+60>>2]=k[Y>>2];if((k[Y>>2]|0)>0)Wd=0;else break;do{k[Z+(aa*344|0)+64+(Wd<<2)>>2]=k[Z+(ba*344|0)+64+(Wd<<2)>>2];Wd=Wd+1|0}while((Wd|0)<(k[Y>>2]|0))}while(0);ba=k[Q>>2]|0;_=k[Z+(ba*344|0)+56>>2]|0;do if(!((ba|aa|0)<0?1:(k[Z+((k[v>>2]|0)*344|0)+56>>2]|0)>=(_|0))){k[Z+(aa*344|0)+56>>2]=_;Ja=Z+(ba*344|0)+60|0;k[Z+(aa*344|0)+60>>2]=k[Ja>>2];if((k[Ja>>2]|0)>0)Xd=0;else break;do{k[Z+(aa*344|0)+64+(Xd<<2)>>2]=k[Z+(ba*344|0)+64+(Xd<<2)>>2];Xd=Xd+1|0}while((Xd|0)<(k[Ja>>2]|0))}while(0);ba=k[$>>2]|0;_=k[Z+(ba*344|0)+56>>2]|0;if((ba|aa|0)<0?1:(k[Z+((k[Q>>2]|0)*344|0)+56>>2]|0)>=(_|0)){U=g;V=aa;break b}k[Z+(aa*344|0)+56>>2]=_;_=Z+(ba*344|0)+60|0;k[Z+(aa*344|0)+60>>2]=k[_>>2];if((k[_>>2]|0)>0)Yd=0;else{U=g;V=aa;break b}while(1){k[Z+(aa*344|0)+64+(Yd<<2)>>2]=k[Z+(ba*344|0)+64+(Yd<<2)>>2];Yd=Yd+1|0;if((Yd|0)>=(k[_>>2]|0)){U=g;V=aa;break b}}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788862,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788912,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 59:{Q=Lu(k[F+-768>>2]|0,1,k[F+-256>>2]|0,0,0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 60:{Q=Lu(k[F+-1280>>2]|0,2,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 61:{Q=Lu(k[F+-1792>>2]|0,3,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 62:{Q=Lu(k[F+-2304>>2]|0,4,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 63:{Q=Lu(k[F+-2816>>2]|0,5,k[F+-2304>>2]|0,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 64:{Q=Iu(259,286,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 65:{Q=Iu(259,286,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 66:{Q=Iu(260,285,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 67:{Q=Iu(260,285,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 68:{Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){$=k[30820]|0;if(!$){k[30822]=100;Zd=cy(34400)|0}else{k[30822]=Q<<1;Zd=fy($,Q*688|0)|0}if(!Zd){k[30835]=113;x=1301;break b}else{k[30820]=Zd;_d=k[30821]|0;break}}else _d=Q;while(0);k[30821]=_d+1;if((_d|0)>-1){Q=k[30820]|0;k[Q+(_d*344|0)>>2]=-1e3;k[Q+(_d*344|0)+4>>2]=0;k[Q+(_d*344|0)+8>>2]=0;k[Q+(_d*344|0)+52>>2]=258;i[Q+(_d*344|0)+88>>0]=i[F>>0]|0;k[Q+(_d*344|0)+84>>2]=0;k[Q+(_d*344|0)+56>>2]=1;k[Q+(_d*344|0)+60>>2]=1;k[Q+(_d*344|0)+64>>2]=1;U=g;V=_d}else x=1301;break}case 69:{Q=k[F>>2]|0;$=k[30821]|0;do if(($|0)==(k[30822]|0)){v=k[30820]|0;if(!v){k[30822]=100;$d=cy(34400)|0}else{k[30822]=$<<1;$d=fy(v,$*688|0)|0}if(!$d){k[30835]=113;x=1301;break b}else{k[30820]=$d;ae=k[30821]|0;break}}else ae=$;while(0);k[30821]=ae+1;if((ae|0)>-1){$=k[30820]|0;k[$+(ae*344|0)>>2]=0-Q;k[$+(ae*344|0)+4>>2]=0;k[$+(ae*344|0)+8>>2]=0;v=k[30828]|0;k[$+(ae*344|0)+52>>2]=k[v+(Q*124|0)+84>>2];k[$+(ae*344|0)+56>>2]=k[v+(Q*124|0)+88>>2];R=k[v+(Q*124|0)+92>>2]|0;k[$+(ae*344|0)+60>>2]=R;if((R|0)>0){ca=0;do{k[$+(ae*344|0)+64+(ca<<2)>>2]=k[v+(Q*124|0)+96+(ca<<2)>>2];ca=ca+1|0}while((ca|0)!=(R|0));if((ae|0)<0)x=1301;else{U=g;V=ae}}else{U=g;V=ae}}else x=1301;break}case 70:{R=k[F+-256>>2]|0;ca=k[30820]|0;if((k[ca+(R*344|0)+52>>2]|0)==259?(k[ca+(R*344|0)>>2]|0)==-1e3:0){Q=k[F+-768>>2]|0;v=k[30821]|0;do if((v|0)==(k[30822]|0)){k[30822]=v<<1;$=fy(ca,v*688|0)|0;if(!$){k[30835]=113;x=1301;break b}else{k[30820]=$;be=k[30821]|0;ce=$;break}}else{be=v;ce=ca}while(0);ca=be+1|0;k[30821]=ca;if((be|0)<=-1){x=1301;break b}k[ce+(be*344|0)>>2]=0-Q;k[ce+(be*344|0)+4>>2]=0;k[ce+(be*344|0)+8>>2]=0;v=k[30828]|0;k[ce+(be*344|0)+52>>2]=k[v+(Q*124|0)+84>>2];k[ce+(be*344|0)+56>>2]=k[v+(Q*124|0)+88>>2];$=k[v+(Q*124|0)+92>>2]|0;k[ce+(be*344|0)+60>>2]=$;if(($|0)>0){ea=0;do{k[ce+(be*344|0)+64+(ea<<2)>>2]=k[v+(Q*124|0)+96+(ea<<2)>>2];ea=ea+1|0}while((ea|0)!=($|0));if((be|0)<0){x=1301;break b}}do if((ca|0)==(k[30822]|0)){if(!ce){k[30822]=100;de=cy(34400)|0}else{k[30822]=ca<<1;de=fy(ce,ca*688|0)|0}if(!de){k[30835]=113;x=1301;break b}else{k[30820]=de;ee=k[30821]|0;fe=de;break}}else{ee=ca;fe=ce}while(0);k[30821]=ee+1;if((ee|0)<=-1){x=1301;break b}k[fe+(ee*344|0)>>2]=123;k[fe+(ee*344|0)+4>>2]=5;k[fe+(ee*344|0)+8>>2]=2;k[fe+(ee*344|0)+12>>2]=be;k[fe+(ee*344|0)+16>>2]=R;ca=k[30828]|0;k[fe+(ee*344|0)+52>>2]=k[ca+(Q*124|0)+84>>2];k[fe+(ee*344|0)+56>>2]=k[ca+(Q*124|0)+88>>2];$=k[ca+(Q*124|0)+92>>2]|0;k[fe+(ee*344|0)+60>>2]=$;if(($|0)>0)ge=0;else{U=g;V=ee;break b}do{k[fe+(ee*344|0)+64+(ge<<2)>>2]=k[ca+(Q*124|0)+96+(ge<<2)>>2];ge=ge+1|0}while((ge|0)!=($|0));if((ee|0)<0){x=1301;break b}else{U=g;V=ee;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788059,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 71:{$=Ku(258,k[F+-512>>2]|0,278,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 72:{$=Ku(258,k[F+-512>>2]|0,277,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 73:{$=Ku(258,k[F+-512>>2]|0,281,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 74:{$=Ku(258,k[F+-512>>2]|0,280,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 75:{$=Ku(258,k[F+-512>>2]|0,282,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 76:{$=Ku(258,k[F+-512>>2]|0,279,k[F>>2]|0)|0;if(($|0)<0)x=1301;else{k[(k[30820]|0)+($*344|0)+56>>2]=1;U=g;V=$}break}case 77:{$=F+-512|0;Q=k[$>>2]|0;ca=k[30820]|0;R=k[ca+(Q*344|0)+52>>2]|0;ea=k[F>>2]|0;v=k[ca+(ea*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ca=Iu(v,0,Q)|0;k[$>>2]=ca;he=ca;ie=k[F>>2]|0}else{he=Q;ie=ea}else{Q=Iu(R,0,ea)|0;k[F>>2]=Q;he=k[$>>2]|0;ie=Q}Q=Ku(258,he,282,ie)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 78:{Q=F+-512|0;$=k[Q>>2]|0;ea=k[30820]|0;R=k[ea+($*344|0)+52>>2]|0;ca=k[F>>2]|0;v=k[ea+(ca*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ea=Iu(v,0,$)|0;k[Q>>2]=ea;je=ea;ke=k[F>>2]|0}else{je=$;ke=ca}else{$=Iu(R,0,ca)|0;k[F>>2]=$;je=k[Q>>2]|0;ke=$}$=Ku(258,je,281,ke)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 79:{$=F+-512|0;Q=k[$>>2]|0;ca=k[30820]|0;R=k[ca+(Q*344|0)+52>>2]|0;ea=k[F>>2]|0;v=k[ca+(ea*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ca=Iu(v,0,Q)|0;k[$>>2]=ca;le=ca;me=k[F>>2]|0}else{le=Q;me=ea}else{Q=Iu(R,0,ea)|0;k[F>>2]=Q;le=k[$>>2]|0;me=Q}Q=Ku(258,le,279,me)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 80:{Q=F+-512|0;$=k[Q>>2]|0;ea=k[30820]|0;R=k[ea+($*344|0)+52>>2]|0;ca=k[F>>2]|0;v=k[ea+(ca*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ea=Iu(v,0,$)|0;k[Q>>2]=ea;ne=ea;oe=k[F>>2]|0}else{ne=$;oe=ca}else{$=Iu(R,0,ca)|0;k[F>>2]=$;ne=k[Q>>2]|0;oe=$}$=Ku(258,ne,280,oe)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 81:{$=F+-512|0;Q=k[$>>2]|0;ca=k[30820]|0;R=k[ca+(Q*344|0)+52>>2]|0;ea=k[F>>2]|0;v=k[ca+(ea*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ca=Iu(v,0,Q)|0;k[$>>2]=ca;pe=ca;qe=k[F>>2]|0}else{pe=Q;qe=ea}else{Q=Iu(R,0,ea)|0;k[F>>2]=Q;pe=k[$>>2]|0;qe=Q}Q=Ku(258,pe,126,qe)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 82:{Q=F+-512|0;$=k[Q>>2]|0;ea=k[30820]|0;R=k[ea+($*344|0)+52>>2]|0;ca=k[F>>2]|0;v=k[ea+(ca*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ea=Iu(v,0,$)|0;k[Q>>2]=ea;re=ea;se=k[F>>2]|0}else{re=$;se=ca}else{$=Iu(R,0,ca)|0;k[F>>2]=$;re=k[Q>>2]|0;se=$}$=Ku(258,re,278,se)|0;if(($|0)<0)x=1301;else{U=g;V=$}break}case 83:{$=F+-512|0;Q=k[$>>2]|0;ca=k[30820]|0;R=k[ca+(Q*344|0)+52>>2]|0;ea=k[F>>2]|0;v=k[ca+(ea*344|0)+52>>2]|0;if((R|0)<=(v|0))if((R|0)<(v|0)){ca=Iu(v,0,Q)|0;k[$>>2]=ca;te=ca;ue=k[F>>2]|0}else{te=Q;ue=ea}else{Q=Iu(R,0,ea)|0;k[F>>2]=Q;te=k[$>>2]|0;ue=Q}Q=Ku(258,te,277,ue)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 84:{Q=Ku(258,k[F+-512>>2]|0,278,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 85:{Q=Ku(258,k[F+-512>>2]|0,277,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 86:{Q=Ku(258,k[F+-512>>2]|0,282,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 87:{Q=Ku(258,k[F+-512>>2]|0,279,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 88:{Q=Ku(258,k[F+-512>>2]|0,281,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 89:{Q=Ku(258,k[F+-512>>2]|0,280,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{k[(k[30820]|0)+(Q*344|0)+56>>2]=1;U=g;V=Q}break}case 90:{Q=Ku(258,k[F+-512>>2]|0,276,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 91:{Q=Ku(258,k[F+-512>>2]|0,275,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 92:{Q=Ku(258,k[F+-512>>2]|0,278,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 93:{Q=Ku(258,k[F+-512>>2]|0,277,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 94:{Q=F+-1024|0;$=k[Q>>2]|0;ea=k[30820]|0;R=k[ea+($*344|0)+52>>2]|0;ca=F+-512|0;v=k[ca>>2]|0;aa=k[ea+(v*344|0)+52>>2]|0;if((R|0)<=(aa|0))if((R|0)<(aa|0)){ea=Iu(aa,0,$)|0;k[Q>>2]=ea;ve=ea}else ve=$;else{$=Iu(R,0,v)|0;k[ca>>2]=$;ve=k[Q>>2]|0}$=k[30820]|0;v=k[$+(ve*344|0)+52>>2]|0;R=k[F>>2]|0;ea=k[$+(R*344|0)+52>>2]|0;if((v|0)<=(ea|0))if((v|0)<(ea|0)){$=Iu(ea,0,ve)|0;k[Q>>2]=$;we=k[F>>2]|0}else we=R;else{$=Iu(v,0,R)|0;k[F>>2]=$;we=$}$=k[ca>>2]|0;R=k[30820]|0;v=k[R+($*344|0)+52>>2]|0;ea=k[R+(we*344|0)+52>>2]|0;if((v|0)<=(ea|0))if((v|0)<(ea|0)){R=Iu(ea,0,$)|0;k[ca>>2]=R;xe=R}else xe=$;else{$=Iu(v,0,we)|0;k[F>>2]=$;xe=k[ca>>2]|0}$=Ku(258,xe,280,k[Q>>2]|0)|0;k[ca>>2]=$;$=Ku(258,k[Q>>2]|0,280,k[F>>2]|0)|0;k[F>>2]=$;Q=Ku(258,k[ca>>2]|0,276,$)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 95:{Q=F+-512|0;$=k[Q>>2]|0;ca=k[F>>2]|0;do if((ca|$|0)>=0){v=k[30820]|0;R=k[v+($*344|0)+56>>2]|0;if((R|0)!=1?(ea=k[v+(ca*344|0)+56>>2]|0,(ea|0)!=1):0){if(!((R|0)==(ea|0)?(k[v+($*344|0)+52>>2]|0)==(k[v+(ca*344|0)+52>>2]|0):0))break;ea=k[v+($*344|0)+60>>2]|0;if((ea|0)!=(k[v+(ca*344|0)+60>>2]|0))break;if((ea|0)>0){R=0;aa=1;while(1){_=(k[v+($*344|0)+64+(R<<2)>>2]|0)==(k[v+(ca*344|0)+64+(R<<2)>>2]|0)?aa:0;R=R+1|0;if((R|0)==(ea|0)){ye=_;break}else aa=_}if(!ye)break}}aa=F+-1024|0;ea=Mu(0,1034,3,$,ca,k[aa>>2]|0,0,0,0,0,0)|0;if((ea|0)<0){x=1301;break b}R=k[30820]|0;v=k[F>>2]|0;_=k[R+(v*344|0)+56>>2]|0;if(!((v|ea|0)<0?1:(k[R+((k[Q>>2]|0)*344|0)+56>>2]|0)>=(_|0))?(k[R+(ea*344|0)+56>>2]=_,_=R+(v*344|0)+60|0,k[R+(ea*344|0)+60>>2]=k[_>>2],(k[_>>2]|0)>0):0){ba=0;do{k[R+(ea*344|0)+64+(ba<<2)>>2]=k[R+(v*344|0)+64+(ba<<2)>>2];ba=ba+1|0}while((ba|0)<(k[_>>2]|0))}_=k[aa>>2]|0;do if((_|ea|0)>=0){ba=k[R+(_*344|0)+56>>2]|0;v=R+(ea*344|0)+56|0;Z=k[v>>2]|0;if((ba|0)!=1)if((Z|0)!=1){if(!((ba|0)==(Z|0)?(k[R+(_*344|0)+52>>2]|0)==(k[R+(ea*344|0)+52>>2]|0):0))break;Ja=k[R+(_*344|0)+60>>2]|0;if((Ja|0)!=(k[R+(ea*344|0)+60>>2]|0))break;if((Ja|0)>0){ze=0;Ae=1}else{U=g;V=ea;break b}while(1){Y=(k[R+(_*344|0)+64+(ze<<2)>>2]|0)==(k[R+(ea*344|0)+64+(ze<<2)>>2]|0)?Ae:0;ze=ze+1|0;if((ze|0)==(Ja|0)){Be=Y;break}else Ae=Y}if(!Be)break;else{U=g;V=ea;break b}}else Ce=1;else Ce=Z;if((Ce|0)>=(ba|0)){U=g;V=ea;break b}k[v>>2]=ba;Ja=R+(_*344|0)+60|0;k[R+(ea*344|0)+60>>2]=k[Ja>>2];if((k[Ja>>2]|0)>0)De=0;else{U=g;V=ea;break b}while(1){k[R+(ea*344|0)+64+(De<<2)>>2]=k[R+(_*344|0)+64+(De<<2)>>2];De=De+1|0;if((De|0)>=(k[Ja>>2]|0)){U=g;V=ea;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788186,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788144,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 96:{Q=F+-512|0;if((i[Q>>0]|0)==73?(tx(Q,788956)|0)==0:0){Q=Mu(0,1030,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;if((Q|0)<0){x=1301;break b}k[(k[30820]|0)+(Q*344|0)+52>>2]=258;U=g;V=Q;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788964,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 97:{Q=F+-512|0;if((i[Q>>0]|0)==73?(tx(Q,788956)|0)==0:0){Q=Mu(0,1030,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;if((Q|0)<0){x=1301;break b}k[(k[30820]|0)+(Q*344|0)+52>>2]=258;U=g;V=Q;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,788964,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 98:{Q=F+-512|0;if((i[Q>>0]|0)==73?(tx(Q,788956)|0)==0:0){Q=Mu(258,1030,1,k[F+-256>>2]|0,0,0,0,0,0,0,0)|0;if((Q|0)<0){x=1301;break b}else{U=g;V=Q;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788964,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 99:{Q=F+-1024|0;if((i[Q>>0]|0)==68?(tx(Q,788596)|0)==0:0){Q=k[F+-768>>2]|0;ca=k[30820]|0;$=k[ca+(Q*344|0)+56>>2]|0;ea=k[F+-256>>2]|0;_=k[ca+(ea*344|0)+56>>2]|0;do if(!(($|0)<(_|0)|(ea|Q|0)<0)){if(!(($|0)==1|(_|0)==1)){if(!(($|0)==(_|0)?(k[ca+(Q*344|0)+52>>2]|0)==(k[ca+(ea*344|0)+52>>2]|0):0))break;R=k[ca+(Q*344|0)+60>>2]|0;if((R|0)!=(k[ca+(ea*344|0)+60>>2]|0))break;if((R|0)>0){aa=0;Ja=1;while(1){ba=(k[ca+(Q*344|0)+64+(aa<<2)>>2]|0)==(k[ca+(ea*344|0)+64+(aa<<2)>>2]|0)?Ja:0;aa=aa+1|0;if((aa|0)==(R|0)){Ee=ba;break}else Ja=ba}if(!Ee)break}}Ja=Mu(0,1031,2,Q,ea,0,0,0,0,0,0)|0;if((Ja|0)<0){x=1301;break b}else{U=g;V=Ja;break b}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,788605,80)|0;i[f>>0]=0;tk(b);x=1301;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,789001,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 100:{ea=F+-1280|0;Q=k[ea>>2]|0;ca=k[30820]|0;if((k[ca+(Q*344|0)+52>>2]|0)==260)Fe=ca;else{ca=Iu(260,0,Q)|0;k[ea>>2]=ca;Fe=k[30820]|0}ca=F+-768|0;Q=k[ca>>2]|0;if((k[Fe+(Q*344|0)+52>>2]|0)==260){Ge=Fe;He=Q}else{_=Iu(260,0,Q)|0;k[ca>>2]=_;Ge=k[30820]|0;He=_}_=F+-256|0;Q=k[_>>2]|0;if((k[Ge+(Q*344|0)+52>>2]|0)==260){Ie=He;Je=Q}else{$=Iu(260,0,Q)|0;k[_>>2]=$;Ie=k[ca>>2]|0;Je=$}$=k[ea>>2]|0;do if((Ie|$|0)>=0){Q=k[30820]|0;Ja=k[Q+($*344|0)+56>>2]|0;if((Ja|0)!=1?(R=k[Q+(Ie*344|0)+56>>2]|0,(R|0)!=1):0){if(!((Ja|0)==(R|0)?(k[Q+($*344|0)+52>>2]|0)==(k[Q+(Ie*344|0)+52>>2]|0):0))break;R=k[Q+($*344|0)+60>>2]|0;if((R|0)!=(k[Q+(Ie*344|0)+60>>2]|0))break;if((R|0)>0){Ja=0;aa=1;while(1){ba=(k[Q+($*344|0)+64+(Ja<<2)>>2]|0)==(k[Q+(Ie*344|0)+64+(Ja<<2)>>2]|0)?aa:0;Ja=Ja+1|0;if((Ja|0)==(R|0)){Ke=ba;break}else aa=ba}if((Ke|0)==0|(Je|Ie|0)<0)break}else x=999}else x=999;if((x|0)==999?(x=0,(Je|Ie|0)<0):0)break;aa=k[Q+(Ie*344|0)+56>>2]|0;if((aa|0)!=1?(R=k[Q+(Je*344|0)+56>>2]|0,(R|0)!=1):0){if(!((aa|0)==(R|0)?(k[Q+(Ie*344|0)+52>>2]|0)==(k[Q+(Je*344|0)+52>>2]|0):0))break;R=k[Q+(Ie*344|0)+60>>2]|0;if((R|0)!=(k[Q+(Je*344|0)+60>>2]|0))break;if((R|0)>0){aa=0;Ja=1;while(1){ba=(k[Q+(Ie*344|0)+64+(aa<<2)>>2]|0)==(k[Q+(Je*344|0)+64+(aa<<2)>>2]|0)?Ja:0;aa=aa+1|0;if((aa|0)==(R|0)){Le=ba;break}else Ja=ba}if(!Le)break}}Ja=F+-1536|0;if((i[Ja>>0]|0)==78?(tx(Ja,789091)|0)==0:0){Ja=Mu(258,1026,3,$,Ie,Je,0,0,0,0,0)|0;if((Ja|0)<0){x=1301;break b}R=k[30820]|0;aa=R+(Ja*344|0)+56|0;Q=k[ea>>2]|0;ba=k[R+(Q*344|0)+56>>2]|0;do if(!((Q|Ja|0)<0?1:(k[aa>>2]|0)>=(ba|0))){k[aa>>2]=ba;v=R+(Q*344|0)+60|0;k[R+(Ja*344|0)+60>>2]=k[v>>2];if((k[v>>2]|0)>0)Me=0;else break;do{k[R+(Ja*344|0)+64+(Me<<2)>>2]=k[R+(Q*344|0)+64+(Me<<2)>>2];Me=Me+1|0}while((Me|0)<(k[v>>2]|0))}while(0);Q=k[ca>>2]|0;ba=k[R+(Q*344|0)+56>>2]|0;do if(!((Q|Ja|0)<0?1:(k[R+((k[ea>>2]|0)*344|0)+56>>2]|0)>=(ba|0))){k[aa>>2]=ba;v=R+(Q*344|0)+60|0;k[R+(Ja*344|0)+60>>2]=k[v>>2];if((k[v>>2]|0)>0)Ne=0;else break;do{k[R+(Ja*344|0)+64+(Ne<<2)>>2]=k[R+(Q*344|0)+64+(Ne<<2)>>2];Ne=Ne+1|0}while((Ne|0)<(k[v>>2]|0))}while(0);Q=k[_>>2]|0;ba=k[R+(Q*344|0)+56>>2]|0;if((Q|Ja|0)<0?1:(k[R+((k[ca>>2]|0)*344|0)+56>>2]|0)>=(ba|0)){U=g;V=Ja;break b}k[aa>>2]=ba;ba=R+(Q*344|0)+60|0;k[R+(Ja*344|0)+60>>2]=k[ba>>2];if((k[ba>>2]|0)>0)Oe=0;else{U=g;V=Ja;break b}while(1){k[R+(Ja*344|0)+64+(Oe<<2)>>2]=k[R+(Q*344|0)+64+(Oe<<2)>>2];Oe=Oe+1|0;if((Oe|0)>=(k[ba>>2]|0)){U=g;V=Ja;break b}}}if(!(k[30835]|0))k[30835]=431;Bx(b,789097,80)|0;i[f>>0]=0;tk(b);x=1301;break b}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,789043,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 101:{ca=F+-2304|0;_=k[ca>>2]|0;ea=k[30820]|0;if((k[ea+(_*344|0)+52>>2]|0)==260)Pe=ea;else{ea=Iu(260,0,_)|0;k[ca>>2]=ea;Pe=k[30820]|0}ea=F+-1792|0;_=k[ea>>2]|0;if((k[Pe+(_*344|0)+52>>2]|0)==260)Qe=Pe;else{$=Iu(260,0,_)|0;k[ea>>2]=$;Qe=k[30820]|0}$=F+-1280|0;_=k[$>>2]|0;if((k[Qe+(_*344|0)+52>>2]|0)==260)Re=Qe;else{Ja=Iu(260,0,_)|0;k[$>>2]=Ja;Re=k[30820]|0}Ja=F+-768|0;_=k[Ja>>2]|0;if((k[Re+(_*344|0)+52>>2]|0)==260)Se=Re;else{ba=Iu(260,0,_)|0;k[Ja>>2]=ba;Se=k[30820]|0}ba=F+-256|0;_=k[ba>>2]|0;if((k[Se+(_*344|0)+52>>2]|0)==260)Te=_;else{Q=Iu(260,0,_)|0;k[ba>>2]=Q;Te=Q}Q=k[ca>>2]|0;_=k[ea>>2]|0;q:do if((_|Q|0)>=0){R=k[30820]|0;aa=k[R+(Q*344|0)+56>>2]|0;if((aa|0)!=1?(v=k[R+(_*344|0)+56>>2]|0,(v|0)!=1):0){if(!((aa|0)==(v|0)?(k[R+(Q*344|0)+52>>2]|0)==(k[R+(_*344|0)+52>>2]|0):0))break;v=k[R+(Q*344|0)+60>>2]|0;if((v|0)!=(k[R+(_*344|0)+60>>2]|0))break;if((v|0)>0){aa=0;Z=1;while(1){Y=(k[R+(Q*344|0)+64+(aa<<2)>>2]|0)==(k[R+(_*344|0)+64+(aa<<2)>>2]|0)?Z:0;aa=aa+1|0;if((aa|0)==(v|0)){Ue=Y;break}else Z=Y}if(!Ue)break}}Z=k[$>>2]|0;if((Z|_|0)>=0){v=k[R+(_*344|0)+56>>2]|0;do if((v|0)!=1?(aa=k[R+(Z*344|0)+56>>2]|0,(aa|0)!=1):0){if(!((v|0)==(aa|0)?(k[R+(_*344|0)+52>>2]|0)==(k[R+(Z*344|0)+52>>2]|0):0))break q;aa=k[R+(_*344|0)+60>>2]|0;if((aa|0)!=(k[R+(Z*344|0)+60>>2]|0))break q;if((aa|0)>0){Ve=0;We=1}else break;while(1){Y=(k[R+(_*344|0)+64+(Ve<<2)>>2]|0)==(k[R+(Z*344|0)+64+(Ve<<2)>>2]|0)?We:0;Ve=Ve+1|0;if((Ve|0)==(aa|0)){Xe=Y;break}else We=Y}if(!Xe)break q}while(0);v=k[Ja>>2]|0;if((v|Z|0)>=0){aa=k[R+(Z*344|0)+56>>2]|0;do if((aa|0)!=1){Y=k[R+(v*344|0)+56>>2]|0;if((Y|0)==1){x=1059;break}if(!((aa|0)==(Y|0)?(k[R+(Z*344|0)+52>>2]|0)==(k[R+(v*344|0)+52>>2]|0):0))break q;Y=k[R+(Z*344|0)+60>>2]|0;if((Y|0)!=(k[R+(v*344|0)+60>>2]|0))break q;if((Y|0)>0){Ye=0;Ze=1}else{x=1059;break}while(1){T=(k[R+(Z*344|0)+64+(Ye<<2)>>2]|0)==(k[R+(v*344|0)+64+(Ye<<2)>>2]|0)?Ze:0;Ye=Ye+1|0;if((Ye|0)==(Y|0)){_e=T;break}else Ze=T}if((_e|0)==0|(Te|v|0)<0)break q}else x=1059;while(0);if((x|0)==1059?(x=0,(Te|v|0)<0):0)break;aa=k[R+(v*344|0)+56>>2]|0;do if((aa|0)!=1){Y=k[R+(Te*344|0)+56>>2]|0;if((Y|0)==1)break;if(!((aa|0)==(Y|0)?(k[R+(v*344|0)+52>>2]|0)==(k[R+(Te*344|0)+52>>2]|0):0))break q;Y=k[R+(v*344|0)+60>>2]|0;if((Y|0)!=(k[R+(Te*344|0)+60>>2]|0))break q;if((Y|0)>0){$e=0;af=1}else break;while(1){T=(k[R+(v*344|0)+64+($e<<2)>>2]|0)==(k[R+(Te*344|0)+64+($e<<2)>>2]|0)?af:0;$e=$e+1|0;if(($e|0)==(Y|0)){bf=T;break}else af=T}if(!bf)break q}while(0);R=F+-2560|0;do if((i[R>>0]|0)==67){if(tx(R,789178)|0)break;aa=Mu(258,1027,5,Q,_,Z,v,Te,0,0,0)|0;if((aa|0)<0){x=1301;break b}Y=k[30820]|0;T=Y+(aa*344|0)+56|0;ja=k[ca>>2]|0;Va=k[Y+(ja*344|0)+56>>2]|0;do if(!((ja|aa|0)<0?1:(k[T>>2]|0)>=(Va|0))){k[T>>2]=Va;da=Y+(ja*344|0)+60|0;k[Y+(aa*344|0)+60>>2]=k[da>>2];if((k[da>>2]|0)>0)cf=0;else break;do{k[Y+(aa*344|0)+64+(cf<<2)>>2]=k[Y+(ja*344|0)+64+(cf<<2)>>2];cf=cf+1|0}while((cf|0)<(k[da>>2]|0))}while(0);ja=k[ea>>2]|0;Va=k[Y+(ja*344|0)+56>>2]|0;do if(!((ja|aa|0)<0?1:(k[Y+((k[ca>>2]|0)*344|0)+56>>2]|0)>=(Va|0))){k[T>>2]=Va;da=Y+(ja*344|0)+60|0;k[Y+(aa*344|0)+60>>2]=k[da>>2];if((k[da>>2]|0)>0)df=0;else break;do{k[Y+(aa*344|0)+64+(df<<2)>>2]=k[Y+(ja*344|0)+64+(df<<2)>>2];df=df+1|0}while((df|0)<(k[da>>2]|0))}while(0);ja=k[$>>2]|0;Va=k[Y+(ja*344|0)+56>>2]|0;do if(!((ja|aa|0)<0?1:(k[Y+((k[ea>>2]|0)*344|0)+56>>2]|0)>=(Va|0))){k[T>>2]=Va;da=Y+(ja*344|0)+60|0;k[Y+(aa*344|0)+60>>2]=k[da>>2];if((k[da>>2]|0)>0)ef=0;else break;do{k[Y+(aa*344|0)+64+(ef<<2)>>2]=k[Y+(ja*344|0)+64+(ef<<2)>>2];ef=ef+1|0}while((ef|0)<(k[da>>2]|0))}while(0);ja=k[Ja>>2]|0;Va=k[Y+(ja*344|0)+56>>2]|0;do if(!((ja|aa|0)<0?1:(k[Y+((k[$>>2]|0)*344|0)+56>>2]|0)>=(Va|0))){k[T>>2]=Va;da=Y+(ja*344|0)+60|0;k[Y+(aa*344|0)+60>>2]=k[da>>2];if((k[da>>2]|0)>0)ff=0;else break;do{k[Y+(aa*344|0)+64+(ff<<2)>>2]=k[Y+(ja*344|0)+64+(ff<<2)>>2];ff=ff+1|0}while((ff|0)<(k[da>>2]|0))}while(0);ja=k[ba>>2]|0;Va=k[Y+(ja*344|0)+56>>2]|0;if((ja|aa|0)<0?1:(k[Y+((k[Ja>>2]|0)*344|0)+56>>2]|0)>=(Va|0)){U=g;V=aa;break b}k[T>>2]=Va;Va=Y+(ja*344|0)+60|0;k[Y+(aa*344|0)+60>>2]=k[Va>>2];if((k[Va>>2]|0)>0)gf=0;else{U=g;V=aa;break b}while(1){k[Y+(aa*344|0)+64+(gf<<2)>>2]=k[Y+(ja*344|0)+64+(gf<<2)>>2];gf=gf+1|0;if((gf|0)>=(k[Va>>2]|0)){U=g;V=aa;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,789097,80)|0;i[f>>0]=0;tk(b);x=1301;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,789128,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 102:{Ja=F+-3328|0;ba=k[Ja>>2]|0;$=k[30820]|0;if((k[$+(ba*344|0)+52>>2]|0)==260)hf=$;else{$=Iu(260,0,ba)|0;k[Ja>>2]=$;hf=k[30820]|0}$=F+-2816|0;ba=k[$>>2]|0;if((k[hf+(ba*344|0)+52>>2]|0)==260)jf=hf;else{ea=Iu(260,0,ba)|0;k[$>>2]=ea;jf=k[30820]|0}ea=F+-2304|0;ba=k[ea>>2]|0;if((k[jf+(ba*344|0)+52>>2]|0)==260)kf=jf;else{ca=Iu(260,0,ba)|0;k[ea>>2]=ca;kf=k[30820]|0}ca=F+-1792|0;ba=k[ca>>2]|0;if((k[kf+(ba*344|0)+52>>2]|0)==260)lf=kf;else{_=Iu(260,0,ba)|0;k[ca>>2]=_;lf=k[30820]|0}_=F+-1280|0;ba=k[_>>2]|0;if((k[lf+(ba*344|0)+52>>2]|0)==260)mf=lf;else{Q=Iu(260,0,ba)|0;k[_>>2]=Q;mf=k[30820]|0}Q=F+-768|0;ba=k[Q>>2]|0;if((k[mf+(ba*344|0)+52>>2]|0)==260)nf=mf;else{v=Iu(260,0,ba)|0;k[Q>>2]=v;nf=k[30820]|0}v=F+-256|0;ba=k[v>>2]|0;if((k[nf+(ba*344|0)+52>>2]|0)==260)of=ba;else{Z=Iu(260,0,ba)|0;k[v>>2]=Z;of=Z}Z=k[Ja>>2]|0;ba=k[$>>2]|0;r:do if((ba|Z|0)>=0){R=k[30820]|0;aa=k[R+(Z*344|0)+56>>2]|0;do if((aa|0)!=1?(Va=k[R+(ba*344|0)+56>>2]|0,(Va|0)!=1):0){if(!((aa|0)==(Va|0)?(k[R+(Z*344|0)+52>>2]|0)==(k[R+(ba*344|0)+52>>2]|0):0))break r;Va=k[R+(Z*344|0)+60>>2]|0;if((Va|0)!=(k[R+(ba*344|0)+60>>2]|0))break r;if((Va|0)>0){pf=0;qf=1}else break;while(1){ja=(k[R+(Z*344|0)+64+(pf<<2)>>2]|0)==(k[R+(ba*344|0)+64+(pf<<2)>>2]|0)?qf:0;pf=pf+1|0;if((pf|0)==(Va|0)){rf=ja;break}else qf=ja}if(!rf)break r}while(0);aa=k[ea>>2]|0;if((aa|ba|0)>=0){Va=k[R+(ba*344|0)+56>>2]|0;do if((Va|0)!=1){ja=k[R+(aa*344|0)+56>>2]|0;if((ja|0)==1)break;if(!((Va|0)==(ja|0)?(k[R+(ba*344|0)+52>>2]|0)==(k[R+(aa*344|0)+52>>2]|0):0))break r;ja=k[R+(ba*344|0)+60>>2]|0;if((ja|0)!=(k[R+(aa*344|0)+60>>2]|0))break r;if((ja|0)>0){sf=0;tf=1}else break;while(1){Y=(k[R+(ba*344|0)+64+(sf<<2)>>2]|0)==(k[R+(aa*344|0)+64+(sf<<2)>>2]|0)?tf:0;sf=sf+1|0;if((sf|0)==(ja|0)){uf=Y;break}else tf=Y}if(!uf)break r}while(0);Va=k[ca>>2]|0;if((Va|aa|0)>=0){ja=k[R+(aa*344|0)+56>>2]|0;do if((ja|0)!=1){Y=k[R+(Va*344|0)+56>>2]|0;if((Y|0)==1)break;if(!((ja|0)==(Y|0)?(k[R+(aa*344|0)+52>>2]|0)==(k[R+(Va*344|0)+52>>2]|0):0))break r;Y=k[R+(aa*344|0)+60>>2]|0;if((Y|0)!=(k[R+(Va*344|0)+60>>2]|0))break r;if((Y|0)>0){vf=0;wf=1}else break;while(1){T=(k[R+(aa*344|0)+64+(vf<<2)>>2]|0)==(k[R+(Va*344|0)+64+(vf<<2)>>2]|0)?wf:0;vf=vf+1|0;if((vf|0)==(Y|0)){xf=T;break}else wf=T}if(!xf)break r}while(0);ja=k[_>>2]|0;if((ja|Va|0)<0)break;Y=k[R+(Va*344|0)+56>>2]|0;do if((Y|0)!=1){T=k[R+(ja*344|0)+56>>2]|0;if((T|0)==1)break;if(!((Y|0)==(T|0)?(k[R+(Va*344|0)+52>>2]|0)==(k[R+(ja*344|0)+52>>2]|0):0))break r;T=k[R+(Va*344|0)+60>>2]|0;if((T|0)!=(k[R+(ja*344|0)+60>>2]|0))break r;if((T|0)>0){yf=0;zf=1}else break;while(1){da=(k[R+(Va*344|0)+64+(yf<<2)>>2]|0)==(k[R+(ja*344|0)+64+(yf<<2)>>2]|0)?zf:0;yf=yf+1|0;if((yf|0)==(T|0)){Af=da;break}else zf=da}if(!Af)break r}while(0);Y=k[Q>>2]|0;if((Y|ja|0)<0)break;T=k[R+(ja*344|0)+56>>2]|0;do if((T|0)!=1){da=k[R+(Y*344|0)+56>>2]|0;if((da|0)==1){x=1145;break}if(!((T|0)==(da|0)?(k[R+(ja*344|0)+52>>2]|0)==(k[R+(Y*344|0)+52>>2]|0):0))break r;da=k[R+(ja*344|0)+60>>2]|0;if((da|0)!=(k[R+(Y*344|0)+60>>2]|0))break r;if((da|0)>0){Bf=0;Cf=1}else{x=1145;break}while(1){fa=(k[R+(ja*344|0)+64+(Bf<<2)>>2]|0)==(k[R+(Y*344|0)+64+(Bf<<2)>>2]|0)?Cf:0;Bf=Bf+1|0;if((Bf|0)==(da|0)){Df=fa;break}else Cf=fa}if((Df|0)==0|(of|Y|0)<0)break r}else x=1145;while(0);if((x|0)==1145?(x=0,(of|Y|0)<0):0)break;T=k[R+(Y*344|0)+56>>2]|0;do if((T|0)!=1){da=k[R+(of*344|0)+56>>2]|0;if((da|0)==1)break;if(!((T|0)==(da|0)?(k[R+(Y*344|0)+52>>2]|0)==(k[R+(of*344|0)+52>>2]|0):0))break r;da=k[R+(Y*344|0)+60>>2]|0;if((da|0)!=(k[R+(of*344|0)+60>>2]|0))break r;if((da|0)>0){Ef=0;Ff=1}else break;while(1){fa=(k[R+(Y*344|0)+64+(Ef<<2)>>2]|0)==(k[R+(of*344|0)+64+(Ef<<2)>>2]|0)?Ff:0;Ef=Ef+1|0;if((Ef|0)==(da|0)){Gf=fa;break}else Ff=fa}if(!Gf)break r}while(0);R=F+-3584|0;T=i[R>>0]|0;do if(T<<24>>24>=66){if(T<<24>>24>66){if(T<<24>>24!=69)break;if(tx(R,789249)|0)break;Hf=Mu(258,1029,7,Z,ba,aa,Va,ja,Y,of,0)|0}else{if(tx(R,789244)|0)break;Hf=Mu(258,1028,7,Z,ba,aa,Va,ja,Y,of,0)|0}if((Hf|0)<0){x=1301;break b}da=k[30820]|0;fa=da+(Hf*344|0)+56|0;If=k[Ja>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[fa>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Lf=0;else break;do{k[da+(Hf*344|0)+64+(Lf<<2)>>2]=k[da+(If*344|0)+64+(Lf<<2)>>2];Lf=Lf+1|0}while((Lf|0)<(k[Kf>>2]|0))}while(0);If=k[$>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[da+((k[Ja>>2]|0)*344|0)+56>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Mf=0;else break;do{k[da+(Hf*344|0)+64+(Mf<<2)>>2]=k[da+(If*344|0)+64+(Mf<<2)>>2];Mf=Mf+1|0}while((Mf|0)<(k[Kf>>2]|0))}while(0);If=k[ea>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[da+((k[$>>2]|0)*344|0)+56>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Nf=0;else break;do{k[da+(Hf*344|0)+64+(Nf<<2)>>2]=k[da+(If*344|0)+64+(Nf<<2)>>2];Nf=Nf+1|0}while((Nf|0)<(k[Kf>>2]|0))}while(0);If=k[ca>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[da+((k[ea>>2]|0)*344|0)+56>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Of=0;else break;do{k[da+(Hf*344|0)+64+(Of<<2)>>2]=k[da+(If*344|0)+64+(Of<<2)>>2];Of=Of+1|0}while((Of|0)<(k[Kf>>2]|0))}while(0);If=k[_>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[da+((k[ca>>2]|0)*344|0)+56>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Pf=0;else break;do{k[da+(Hf*344|0)+64+(Pf<<2)>>2]=k[da+(If*344|0)+64+(Pf<<2)>>2];Pf=Pf+1|0}while((Pf|0)<(k[Kf>>2]|0))}while(0);If=k[Q>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;do if(!((If|Hf|0)<0?1:(k[da+((k[_>>2]|0)*344|0)+56>>2]|0)>=(Jf|0))){k[fa>>2]=Jf;Kf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Kf>>2];if((k[Kf>>2]|0)>0)Qf=0;else break;do{k[da+(Hf*344|0)+64+(Qf<<2)>>2]=k[da+(If*344|0)+64+(Qf<<2)>>2];Qf=Qf+1|0}while((Qf|0)<(k[Kf>>2]|0))}while(0);If=k[v>>2]|0;Jf=k[da+(If*344|0)+56>>2]|0;if((If|Hf|0)<0?1:(k[da+((k[Q>>2]|0)*344|0)+56>>2]|0)>=(Jf|0)){U=g;V=Hf;break b}k[fa>>2]=Jf;Jf=da+(If*344|0)+60|0;k[da+(Hf*344|0)+60>>2]=k[Jf>>2];if((k[Jf>>2]|0)>0)Rf=0;else{U=g;V=Hf;break b}while(1){k[da+(Hf*344|0)+64+(Rf<<2)>>2]=k[da+(If*344|0)+64+(Rf<<2)>>2];Rf=Rf+1|0;if((Rf|0)>=(k[Jf>>2]|0)){U=g;V=Hf;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,789258,80)|0;i[f>>0]=0;tk(b);x=1301;break b}}}while(0);if(!(k[30835]|0))k[30835]=431;Bx(b,789186,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 103:{Q=Nu(795593,-99,789291,789299)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 104:{Q=Nu(F+-256|0,-99,789291,789299)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 105:{Q=Nu(F+-768|0,k[F+-256>>2]|0,789291,789299)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 106:{Q=Nu(F+-1792|0,k[F+-1280>>2]|0,F+-768|0,F+-256|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 107:{Q=Ou(F+-256|0,-99,-99,795593)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 108:{Q=Ou(F+-1280|0,k[F+-768>>2]|0,k[F+-256>>2]|0,795593)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 109:{Q=Ou(F+-1792|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,F+-256|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 110:{Q=Lu(k[F+-768>>2]|0,1,k[F+-256>>2]|0,0,0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 111:{Q=Lu(k[F+-1280>>2]|0,2,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 112:{Q=Lu(k[F+-1792>>2]|0,3,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 113:{Q=Lu(k[F+-2304>>2]|0,4,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0,0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 114:{Q=Lu(k[F+-2816>>2]|0,5,k[F+-2304>>2]|0,k[F+-1792>>2]|0,k[F+-1280>>2]|0,k[F+-768>>2]|0,k[F+-256>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 115:{Q=Iu(258,284,k[F>>2]|0)|0;if((Q|0)<0)x=1301;else{U=g;V=Q}break}case 116:{U=g;V=k[F+-256>>2]|0;break}case 117:{Q=(xx(F)|0)+1|0;v=k[30821]|0;do if((v|0)==(k[30822]|0)){_=k[30820]|0;if(!_){k[30822]=100;Sf=cy(34400)|0}else{k[30822]=v<<1;Sf=fy(_,v*688|0)|0}if(!Sf){k[30835]=113;x=1301;break b}else{k[30820]=Sf;Tf=k[30821]|0;break}}else Tf=v;while(0);k[30821]=Tf+1;if((Tf|0)>-1){v=k[30820]|0;k[v+(Tf*344|0)>>2]=-1e3;k[v+(Tf*344|0)+4>>2]=0;k[v+(Tf*344|0)+8>>2]=0;k[v+(Tf*344|0)+52>>2]=261;py(v+(Tf*344|0)+88|0,F|0,Q|0)|0;k[v+(Tf*344|0)+84>>2]=0;k[v+(Tf*344|0)+56>>2]=1;k[v+(Tf*344|0)+60>>2]=1;k[v+(Tf*344|0)+64>>2]=1;v=xx(F)|0;k[(k[30820]|0)+(Tf*344|0)+56>>2]=v;U=g;V=Tf}else x=1301;break}case 118:{v=k[F>>2]|0;_=k[30821]|0;do if((_|0)==(k[30822]|0)){ca=k[30820]|0;if(!ca){k[30822]=100;Uf=cy(34400)|0}else{k[30822]=_<<1;Uf=fy(ca,_*688|0)|0}if(!Uf){k[30835]=113;x=1301;break b}else{k[30820]=Uf;Vf=k[30821]|0;break}}else Vf=_;while(0);k[30821]=Vf+1;if((Vf|0)>-1){_=k[30820]|0;k[_+(Vf*344|0)>>2]=0-v;k[_+(Vf*344|0)+4>>2]=0;k[_+(Vf*344|0)+8>>2]=0;Q=k[30828]|0;k[_+(Vf*344|0)+52>>2]=k[Q+(v*124|0)+84>>2];k[_+(Vf*344|0)+56>>2]=k[Q+(v*124|0)+88>>2];ca=k[Q+(v*124|0)+92>>2]|0;k[_+(Vf*344|0)+60>>2]=ca;if((ca|0)>0){ea=0;do{k[_+(Vf*344|0)+64+(ea<<2)>>2]=k[Q+(v*124|0)+96+(ea<<2)>>2];ea=ea+1|0}while((ea|0)!=(ca|0));if((Vf|0)<0)x=1301;else{U=g;V=Vf}}else{U=g;V=Vf}}else x=1301;break}case 119:{ca=k[F+-256>>2]|0;ea=k[30820]|0;if((k[ea+(ca*344|0)+52>>2]|0)==259?(k[ea+(ca*344|0)>>2]|0)==-1e3:0){v=k[F+-768>>2]|0;Q=k[30821]|0;do if((Q|0)==(k[30822]|0)){k[30822]=Q<<1;_=fy(ea,Q*688|0)|0;if(!_){k[30835]=113;x=1301;break b}else{k[30820]=_;Wf=k[30821]|0;Xf=_;break}}else{Wf=Q;Xf=ea}while(0);ea=Wf+1|0;k[30821]=ea;if((Wf|0)<=-1){x=1301;break b}k[Xf+(Wf*344|0)>>2]=0-v;k[Xf+(Wf*344|0)+4>>2]=0;k[Xf+(Wf*344|0)+8>>2]=0;Q=k[30828]|0;k[Xf+(Wf*344|0)+52>>2]=k[Q+(v*124|0)+84>>2];k[Xf+(Wf*344|0)+56>>2]=k[Q+(v*124|0)+88>>2];_=k[Q+(v*124|0)+92>>2]|0;k[Xf+(Wf*344|0)+60>>2]=_;if((_|0)>0){$=0;do{k[Xf+(Wf*344|0)+64+($<<2)>>2]=k[Q+(v*124|0)+96+($<<2)>>2];$=$+1|0}while(($|0)!=(_|0));if((Wf|0)<0){x=1301;break b}}do if((ea|0)==(k[30822]|0)){if(!Xf){k[30822]=100;Yf=cy(34400)|0}else{k[30822]=ea<<1;Yf=fy(Xf,ea*688|0)|0}if(!Yf){k[30835]=113;x=1301;break b}else{k[30820]=Yf;Zf=k[30821]|0;_f=Yf;break}}else{Zf=ea;_f=Xf}while(0);k[30821]=Zf+1;if((Zf|0)<=-1){x=1301;break b}k[_f+(Zf*344|0)>>2]=123;k[_f+(Zf*344|0)+4>>2]=5;k[_f+(Zf*344|0)+8>>2]=2;k[_f+(Zf*344|0)+12>>2]=Wf;k[_f+(Zf*344|0)+16>>2]=ca;ea=k[30828]|0;k[_f+(Zf*344|0)+52>>2]=k[ea+(v*124|0)+84>>2];k[_f+(Zf*344|0)+56>>2]=k[ea+(v*124|0)+88>>2];_=k[ea+(v*124|0)+92>>2]|0;k[_f+(Zf*344|0)+60>>2]=_;if((_|0)>0)$f=0;else{U=g;V=Zf;break b}do{k[_f+(Zf*344|0)+64+($f<<2)>>2]=k[ea+(v*124|0)+96+($f<<2)>>2];$f=$f+1|0}while(($f|0)!=(_|0));if((Zf|0)<0){x=1301;break b}else{U=g;V=Zf;break b}}if(!(k[30835]|0))k[30835]=431;Bx(b,788059,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 120:{U=g;V=Mu(261,1036,0,0,0,0,0,0,0,0,0)|0;break}case 121:{U=g;V=k[F+-256>>2]|0;break}case 122:{_=F+-512|0;v=k[_>>2]|0;ea=k[30820]|0;ca=k[F>>2]|0;if(((k[ea+(ca*344|0)+56>>2]|0)+(k[ea+(v*344|0)+56>>2]|0)|0)<=255){ea=Ku(261,v,43,ca)|0;if((ea|0)<0){x=1301;break b}ca=k[30820]|0;k[ca+(ea*344|0)+56>>2]=(k[ca+((k[F>>2]|0)*344|0)+56>>2]|0)+(k[ca+((k[_>>2]|0)*344|0)+56>>2]|0);U=g;V=ea;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,789306,80)|0;i[f>>0]=0;tk(b);x=1301;break}case 123:{ea=k[F+-1024>>2]|0;_=k[30820]|0;if((k[_+(ea*344|0)+56>>2]|0)!=1){if(!(k[30835]|0))k[30835]=431;Bx(b,789350,80)|0;i[f>>0]=0;tk(b);x=1301;break b}ca=F+-512|0;v=k[ca>>2]|0;$=k[_+(v*344|0)+56>>2]|0;Q=k[F>>2]|0;Ja=k[_+(Q*344|0)+56>>2]|0;_=Mu(0,1034,3,v,Q,ea,0,0,0,0,(Ja|0)>($|0)?Ja:$)|0;if((_|0)>=0){$=k[30820]|0;Ja=k[F>>2]|0;ea=k[$+(Ja*344|0)+56>>2]|0;if(!((Ja|_|0)<0?1:(k[$+((k[ca>>2]|0)*344|0)+56>>2]|0)>=(ea|0))?(k[$+(_*344|0)+56>>2]=ea,ea=$+(Ja*344|0)+60|0,k[$+(_*344|0)+60>>2]=k[ea>>2],(k[ea>>2]|0)>0):0){ca=0;do{k[$+(_*344|0)+64+(ca<<2)>>2]=k[$+(Ja*344|0)+64+(ca<<2)>>2];ca=ca+1|0}while((ca|0)<(k[ea>>2]|0));U=g;V=_}else{U=g;V=_}}else x=1301;break}case 124:{ea=F+-1024|0;if((i[ea>>0]|0)==68?(tx(ea,788596)|0)==0:0){ea=F+-768|0;ca=k[ea>>2]|0;Ja=k[30820]|0;$=k[Ja+(ca*344|0)+56>>2]|0;Q=F+-256|0;v=k[Q>>2]|0;ba=k[Ja+(v*344|0)+56>>2]|0;Ja=Mu(0,1031,2,ca,v,0,0,0,0,0,(ba|0)>($|0)?ba:$)|0;if((Ja|0)<0){x=1301;break b}$=k[30820]|0;ba=k[$+((k[Q>>2]|0)*344|0)+56>>2]|0;if((ba|0)<=(k[$+((k[ea>>2]|0)*344|0)+56>>2]|0)){U=g;V=Ja;break b}k[$+(Ja*344|0)+56>>2]=ba;U=g;V=Ja;break b}if(!(k[30835]|0))k[30835]=431;Bx(b,789385,80)|0;i[f>>0]=0;tk(b);x=1301;break}default:{U=g;V=z}}while(0);if((x|0)==1301){x=0;z=0-y|0;Ja=C+(z<<1)|0;ag=Ja;bg=j[Ja>>1]|0;cg=F+(z<<8)|0;x=1302;break}z=0-y|0;Ja=C+(z<<1)|0;k[A>>2]=V;py(F+(z<<8)+260|0,e|0,252)|0;z=(l[789579+d>>0]|0)+-54|0;ba=j[Ja>>1]|0;$=(ba<<16>>16)+(j[336004+(z<<1)>>1]|0)|0;if($>>>0<1604?(j[329588+($<<1)>>1]|0)==ba<<16>>16:0){M=U;N=Ja;O=m[332796+($<<1)>>1]|0;P=A;break}M=U;N=Ja;O=i[789705+z>>0]|0;P=A}}while(0);s:do if((x|0)==1294){x=0;switch(g|0){case 0:{k[31157]=(k[31157]|0)+1;if(!(k[30835]|0))k[30835]=431;Bx(b,789714,80)|0;i[f>>0]=0;tk(b);ag=C;bg=q;cg=F;x=1302;break s;break}case 3:{z=k[31158]|0;if((z|0)<1)if(!z){I=1;J=B;x=1312;break a}else{ag=C;bg=q;cg=F;x=1302;break s}else{k[31158]=-2;ag=C;bg=q;cg=F;x=1302;break s}break}default:{ag=C;bg=q;cg=F;x=1302;break s}}}while(0);if((x|0)==1302){while(1){x=0;z=j[329008+(bg<<1)>>1]|0;if(((z<<16>>16!=-46?(Ja=(z<<16>>16)+1|0,z<<16>>16>-2&(Ja|0)<1604):0)?(j[329588+(Ja<<1)>>1]|0)==1:0)?(z=j[332796+(Ja<<1)>>1]|0,z<<16>>16!=0):0){dg=z;eg=ag;fg=cg;break}if((ag|0)==(B|0)){I=1;J=B;x=1312;break a}z=ag+-2|0;ag=z;bg=j[z>>1]|0;cg=cg+-256|0;x=1302}z=fg+256|0;py(z|0,99104,256)|0;M=3;N=eg;O=dg&65535;P=z}g=M;h=B;n=N+2|0;o=D;q=O;s=E;u=P}if((x|0)==1309){if(!(k[30835]|0))k[30835]=431;Bx(b,789727,80)|0;i[f>>0]=0;tk(b);I=2;J=w;x=1312}if((x|0)==1312)if((J|0)==(c|0)){gg=I;r=a;return gg|0}else{G=I;H=J}dy(H);gg=G;r=a;return gg|0}function Eu(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0;if(!(k[31159]|0)){ev(Ra(0)|0);k[31159]=1}k[30824]=a;k[30825]=b;b=a-(k[30830]|0)|0;a=k[30821]|0;if((a|0)>0)c=0;else{d=k[30823]|0;Pu(d);return}do{e=k[30820]|0;f=k[e+(c*344|0)>>2]|0;a:do if(!((f|0)>0|(f|0)==-1e3)){g=0-f|0;h=k[30828]|0;i=ha(k[h+(g*124|0)+88>>2]|0,b)|0;k[e+(c*344|0)+84>>2]=(k[h+(g*124|0)+116>>2]|0)+i;switch(k[e+(c*344|0)+52>>2]|0){case 262:{k[e+(c*344|0)+88>>2]=(k[h+(g*124|0)+120>>2]|0)+(b<<2);k[(k[30820]|0)+(c*344|0)+84>>2]=0;break a;break}case 261:{k[e+(c*344|0)+88>>2]=(k[h+(g*124|0)+120>>2]|0)+(b<<2);k[(k[30820]|0)+(c*344|0)+84>>2]=(k[(k[30828]|0)+(g*124|0)+116>>2]|0)+b;break a;break}case 258:{k[e+(c*344|0)+88>>2]=(k[h+(g*124|0)+120>>2]|0)+i;break a;break}case 259:{k[e+(c*344|0)+88>>2]=(k[h+(g*124|0)+120>>2]|0)+(i<<2);break a;break}case 260:{k[e+(c*344|0)+88>>2]=(k[h+(g*124|0)+120>>2]|0)+(i<<3);break a;break}default:break a}}while(0);c=c+1|0}while((c|0)<(a|0));d=k[30823]|0;Pu(d);return}function Fu(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0;c=b+-1|0;d=(c|0)/2|0;if((b|0)>1){e=c;f=0}else{g=a+(d<<2)|0;h=k[g>>2]|0;return h|0}while(1){c=f+1|0;if((e|0)==(c|0)){i=e;j=f;break}b=a+(((f+e|0)/2|0)<<2)|0;l=k[b>>2]|0;m=a+(e<<2)|0;n=k[m>>2]|0;if((l|0)>(n|0)){k[b>>2]=n;k[m>>2]=l;o=l}else o=n;n=a+(f<<2)|0;l=k[n>>2]|0;if((l|0)>(o|0)){k[n>>2]=o;k[m>>2]=l;p=k[n>>2]|0}else p=l;l=k[b>>2]|0;if((l|0)>(p|0)){k[b>>2]=p;k[n>>2]=l;q=k[b>>2]|0}else q=l;l=a+(c<<2)|0;k[b>>2]=k[l>>2];k[l>>2]=q;l=e;b=c;while(1){c=k[n>>2]|0;m=b;while(1){r=m+1|0;s=a+(r<<2)|0;t=k[s>>2]|0;if((c|0)>(t|0))m=r;else{u=r;v=s;w=t;break}}m=l;while(1){t=m+-1|0;s=a+(t<<2)|0;r=k[s>>2]|0;if((r|0)>(c|0))m=t;else{x=t;y=s;z=r;A=m;break}}if((A|0)<=(u|0)){B=u;C=x;D=y;E=z;F=c;G=A;break}k[v>>2]=z;k[y>>2]=w;l=x;b=u}k[n>>2]=E;k[D>>2]=F;f=(C|0)>(d|0)?f:B;e=(G|0)>(d|0)?G+-2|0:e;if((e|0)<=(f|0)){H=19;break}}if((H|0)==19){g=a+(d<<2)|0;h=k[g>>2]|0;return h|0}H=a+(j<<2)|0;j=k[H>>2]|0;f=a+(i<<2)|0;i=k[f>>2]|0;if((j|0)<=(i|0)){g=a+(d<<2)|0;h=k[g>>2]|0;return h|0}k[H>>2]=i;k[f>>2]=j;g=a+(d<<2)|0;h=k[g>>2]|0;return h|0}function Gu(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0.0,i=0,j=0,k=0.0,l=0,m=0.0,n=0.0,o=0,q=0.0,r=0.0,s=0,t=0,u=0,v=0,w=0.0,x=0,y=0,z=0.0,A=0,B=0,C=0,D=0,E=0.0,F=0.0,G=0,H=0;c=b+-1|0;d=(c|0)/2|0;if((b|0)>1){e=c;f=0}else{g=a+(d<<3)|0;h=+p[g>>3];return +h}while(1){c=f+1|0;if((e|0)==(c|0)){i=e;j=f;break}b=a+(((f+e|0)/2|0)<<3)|0;k=+p[b>>3];l=a+(e<<3)|0;m=+p[l>>3];if(k>m){p[b>>3]=m;p[l>>3]=k;n=k}else n=m;o=a+(f<<3)|0;m=+p[o>>3];if(m>n){p[o>>3]=n;p[l>>3]=m;q=+p[o>>3]}else q=m;m=+p[b>>3];if(m>q){p[b>>3]=q;p[o>>3]=m;r=+p[b>>3]}else r=m;l=a+(c<<3)|0;p[b>>3]=+p[l>>3];p[l>>3]=r;l=e;b=c;while(1){m=+p[o>>3];c=b;while(1){s=c+1|0;t=a+(s<<3)|0;k=+p[t>>3];if(m>k)c=s;else{u=s;v=t;w=k;break}}c=l;while(1){t=c+-1|0;s=a+(t<<3)|0;k=+p[s>>3];if(k>m)c=t;else{x=t;y=s;z=k;A=c;break}}if((A|0)<=(u|0)){B=u;C=x;D=y;E=z;F=m;G=A;break}p[v>>3]=z;p[y>>3]=w;l=x;b=u}p[o>>3]=E;p[D>>3]=F;f=(C|0)>(d|0)?f:B;e=(G|0)>(d|0)?G+-2|0:e;if((e|0)<=(f|0)){H=19;break}}if((H|0)==19){g=a+(d<<3)|0;h=+p[g>>3];return +h}H=a+(j<<3)|0;F=+p[H>>3];j=a+(i<<3)|0;E=+p[j>>3];if(!(F>E)){g=a+(d<<3)|0;h=+p[g>>3];return +h}p[H>>3]=E;p[j>>3]=F;g=a+(d<<3)|0;h=+p[g>>3];return +h}function Hu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0;Su(a);b=a+8|0;c=k[b>>2]|0;if(!(k[30835]|0)){if((c|0)<=0)return;d=a+56|0;e=a+84|0;f=a+52|0;g=a+88|0;h=0;j=0;while(1){l=k[30820]|0;m=k[a+12+(h<<2)>>2]|0;n=k[30825]|0;a:do if((k[l+(m*344|0)>>2]|0)!=-1e3){o=l+(m*344|0)+56|0;if(!n)q=o;else{r=k[o>>2]|0;s=l+(m*344|0)+84|0;t=l+(m*344|0)+88|0;u=n;v=r;w=ha(r,n)|0;while(1){r=u+-1|0;x=(ha(k[d>>2]|0,r)|0)+j|0;if(!v)y=w;else{z=v;A=w;do{z=z+-1|0;A=A+-1|0;B=x+z|0;i[(k[e>>2]|0)+B>>0]=i[(k[s>>2]|0)+A>>0]|0;switch(k[f>>2]|0){case 258:{i[(k[g>>2]|0)+B>>0]=i[(k[t>>2]|0)+A>>0]|0;break}case 259:{k[(k[g>>2]|0)+(B<<2)>>2]=k[(k[t>>2]|0)+(A<<2)>>2];break}case 260:{p[(k[g>>2]|0)+(B<<3)>>3]=+p[(k[t>>2]|0)+(A<<3)>>3];break}default:{}}}while((z|0)!=0);y=w-v|0}if(!r){q=o;break a}u=r;v=k[o>>2]|0;w=y}}}else{w=k[d>>2]|0;o=(ha(w,n)|0)+j-w|0;b:do if((o|0)>-1){w=l+(m*344|0)+88|0;v=l+(m*344|0)+88|0;u=o;while(1){i[(k[e>>2]|0)+u>>0]=0;switch(k[f>>2]|0){case 258:{i[(k[g>>2]|0)+u>>0]=i[w>>0]|0;break}case 259:{k[(k[g>>2]|0)+(u<<2)>>2]=k[w>>2];break}case 260:{p[(k[g>>2]|0)+(u<<3)>>3]=+p[v>>3];break}default:{}}u=u-(k[d>>2]|0)|0;if((u|0)<=-1)break b}}while(0);q=l+(m*344|0)+56|0}while(0);h=h+1|0;m=k[b>>2]|0;if((h|0)>=(m|0)){C=m;break}else j=(k[q>>2]|0)+j|0}}else C=c;if((C|0)>0){D=C;E=0}else return;while(1){C=k[a+12+(E<<2)>>2]|0;c=k[30820]|0;if((k[c+(C*344|0)>>2]|0)>0){dy(k[c+(C*344|0)+88>>2]|0);F=k[b>>2]|0}else F=D;E=E+1|0;if((E|0)>=(F|0))break;else D=F}return}function Iu(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0;if((c|0)<0){d=-1;return d|0}e=k[30820]|0;f=(b|0)==0?a:b;switch(f|0){case 260:case 285:{if((k[e+(c*344|0)+52>>2]|0)==260){d=c;return d|0}break}case 259:case 286:{if((k[e+(c*344|0)+52>>2]|0)==259){d=c;return d|0}break}case 258:{if((k[e+(c*344|0)+52>>2]|0)==258){d=c;return d|0}break}default:{}}b=k[30821]|0;do if((b|0)==(k[30822]|0)){if(!e){k[30822]=100;g=cy(34400)|0}else{k[30822]=b<<1;g=fy(e,b*688|0)|0}if(g){k[30820]=g;h=k[30821]|0;i=g;break}k[30835]=113;d=-1;return d|0}else{h=b;i=e}while(0);k[30821]=h+1;if((h|0)<=-1){d=h;return d|0}e=i+(h*344|0)|0;k[e>>2]=f;f=i+(h*344|0)+4|0;k[f>>2]=6;k[i+(h*344|0)+8>>2]=1;k[i+(h*344|0)+12>>2]=c;k[i+(h*344|0)+52>>2]=a;k[i+(h*344|0)+56>>2]=k[i+(c*344|0)+56>>2];a=i+(c*344|0)+60|0;k[i+(h*344|0)+60>>2]=k[a>>2];b=k[a>>2]|0;if((b|0)>0){a=0;do{k[i+(h*344|0)+64+(a<<2)>>2]=k[i+(c*344|0)+64+(a<<2)>>2];a=a+1|0}while((a|0)<(b|0))}if((k[i+(c*344|0)>>2]|0)!=-1e3){d=h;return d|0}Sb[k[f>>2]&31](e);d=h;return d|0}function Ju(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0;b=k[30820]|0;c=k[a+12>>2]|0;d=b+(c*344|0)|0;e=k[b+((k[a+16>>2]|0)*344|0)+88>>2]|0;Su(a);f=(k[30824]|0)+e|0;g=a+52|0;h=k[g>>2]|0;if((h+-261|0)>>>0<2)j=1;else j=k[a+56>>2]|0;l=k[30830]|0;a:do if((l|0)>(f|0)){m=l-f|0;n=k[30825]|0;o=(m|0)>(n|0)?n:m;m=n-o|0;if((o|0)>0&(f|0)<1){q=a+56|0;r=a+88|0;s=(j|0)==0;t=a+84|0;u=f+-1|0;v=u-((f|0)>0?f:0)|0;w=u-l|0;u=~n;n=(w|0)>(u|0)?w:u;u=n+-1-((n|0)>-2?n:-2)|0;w=v>>>0>u>>>0?v:u;u=w+-1|0;v=f-w|0;w=h;x=f;y=0;while(1){if((w|0)!=262)if(s)z=y;else{A=j;B=y;while(1){A=A+-1|0;i[(k[t>>2]|0)+B>>0]=1;if(!A)break;else B=B+1|0}z=j+y|0}else{B=k[q>>2]|0;i[(k[(k[r>>2]|0)+(y<<2)>>2]|0)+B>>0]=0;if(B){A=B;do{A=A+-1|0;i[(k[(k[r>>2]|0)+(y<<2)>>2]|0)+A>>0]=48}while((A|0)!=0)}z=y+1|0}A=x+1|0;if((A|0)==(v|0)){C=z;break}w=k[g>>2]|0;x=A;y=z}D=v;E=m;F=u-n|0;G=C;H=27}else{D=f;E=m;F=o;G=0;H=27}}else{y=k[30825]|0;x=y+f|0;w=(k[30831]|0)+l|0;if((x|0)>(w|0)){r=x-w|0;x=(r|0)>(y|0);q=x?y:r;r=x?f:w;w=y-q|0;x=ha(w,j)|0;if((q|0)>0?(q+r|0)>(k[30832]|0):0){t=a+56|0;s=a+88|0;A=(j|0)==0;B=a+84|0;I=h;J=ha(y,j)|0;K=q;while(1){if((I|0)==262){L=k[t>>2]|0;M=J+-1|0;i[(k[(k[s>>2]|0)+(M<<2)>>2]|0)+L>>0]=0;if(!L)N=M;else{O=L;do{O=O+-1|0;i[(k[(k[s>>2]|0)+(M<<2)>>2]|0)+O>>0]=48}while((O|0)!=0);N=M}}else if(A)N=J;else{M=J;O=j;do{O=O+-1|0;M=M+-1|0;i[(k[B>>2]|0)+M>>0]=1}while((O|0)!=0);N=J-j|0}O=K+-1|0;if(!((K|0)>1?(O+r|0)>(k[30832]|0):0)){D=r;E=w;F=O;G=x;H=27;break a}I=k[g>>2]|0;J=N;K=O}}else{D=r;E=w;F=q;G=x;H=27}}else P=y}while(0);b:do if((H|0)==27)if((F|0)>0)switch(k[g>>2]|0){case 261:case 262:{Yb[k[123248>>2]&1](0-(k[d>>2]|0)|0,D,F,(k[a+88>>2]|0)+(G<<2)|0,(k[a+84>>2]|0)+G|0)|0;P=E;break b;break}case 258:{Yb[k[123248>>2]&1](0-(k[d>>2]|0)|0,D,F,(k[a+88>>2]|0)+G|0,(k[a+84>>2]|0)+G|0)|0;P=E;break b;break}case 259:{Yb[k[123248>>2]&1](0-(k[d>>2]|0)|0,D,F,(k[a+88>>2]|0)+(G<<2)|0,(k[a+84>>2]|0)+G|0)|0;P=E;break b;break}case 260:{Yb[k[123248>>2]&1](0-(k[d>>2]|0)|0,D,F,(k[a+88>>2]|0)+(G<<3)|0,(k[a+84>>2]|0)+G|0)|0;P=E;break b;break}default:{P=E;break b}}else P=E;while(0);if((P|0)<1)return;E=ha(j,e)|0;if(k[30835]|0)return;G=a+88|0;F=b+(c*344|0)+88|0;D=b+(c*344|0)+84|0;c=a+84|0;if(!j)return;a=P;b=ha((e|0)>0?P:k[30825]|0,j)|0;while(1){P=a+-1|0;e=j;d=b;while(1){e=e+-1|0;N=d+-1|0;h=N+E|0;c:do if((k[g>>2]|0)==262)H=43;else{i[(k[c>>2]|0)+N>>0]=i[(k[D>>2]|0)+h>>0]|0;switch(k[g>>2]|0){case 262:{H=43;break c;break}case 261:{ux(k[(k[G>>2]|0)+(N<<2)>>2]|0,k[(k[F>>2]|0)+(h<<2)>>2]|0)|0;break c;break}case 258:{i[(k[G>>2]|0)+N>>0]=i[(k[F>>2]|0)+h>>0]|0;break c;break}case 259:{k[(k[G>>2]|0)+(N<<2)>>2]=k[(k[F>>2]|0)+(h<<2)>>2];break c;break}case 260:{p[(k[G>>2]|0)+(N<<3)>>3]=+p[(k[F>>2]|0)+(h<<3)>>3];break c;break}default:break c}}while(0);if((H|0)==43){H=0;ux(k[(k[G>>2]|0)+(N<<2)>>2]|0,k[(k[F>>2]|0)+(h<<2)>>2]|0)|0}f=k[30835]|0;if(!((e|0)!=0&(f|0)==0)){Q=N;R=f;break}else d=N}if((P|0)!=0&(R|0)==0){a=P;b=Q}else break}return}function Ku(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;e=r;r=r+80|0;f=e;if((d|b|0)<0){g=-1;r=e;return g|0}h=k[30821]|0;do if((h|0)==(k[30822]|0)){j=k[30820]|0;if(!j){k[30822]=100;l=cy(34400)|0}else{k[30822]=h<<1;l=fy(j,h*688|0)|0}if(l){k[30820]=l;m=k[30821]|0;break}k[30835]=113;g=-1;r=e;return g|0}else m=h;while(0);k[30821]=m+1;if((m|0)<=-1){g=m;r=e;return g|0}h=k[30820]|0;l=h+(m*344|0)|0;k[l>>2]=c;k[h+(m*344|0)+8>>2]=2;k[h+(m*344|0)+12>>2]=b;k[h+(m*344|0)+16>>2]=d;k[h+(m*344|0)+52>>2]=a;if((k[h+(b*344|0)>>2]|0)==-1e3)n=(k[h+(d*344|0)>>2]|0)==-1e3;else n=0;a=k[h+(b*344|0)+52>>2]|0;j=k[h+(b*344|0)+56>>2]|0;do if((a+-261|0)>>>0>=2)if((j|0)!=1){o=k[h+(d*344|0)+56>>2]|0;if((o|0)==1)p=j;else{if(((j|0)==(o|0)?(a|0)==(k[h+(d*344|0)+52>>2]|0):0)?(o=k[h+(b*344|0)+60>>2]|0,(o|0)==(k[h+(d*344|0)+60>>2]|0)):0){if((o|0)>0){q=0;s=1}else{p=j;break}while(1){t=(k[h+(b*344|0)+64+(q<<2)>>2]|0)==(k[h+(d*344|0)+64+(q<<2)>>2]|0)?s:0;q=q+1|0;if((q|0)==(o|0)){u=t;break}else s=t}if(u){p=j;break}}o=k[30821]|0;if(o)k[30821]=o+-1;if(!(k[30835]|0))k[30835]=431;Bx(f,790698,80)|0;i[f+79>>0]=0;tk(f);g=-1;r=e;return g|0}}else p=1;else p=j;while(0);j=(p|0)==1?d:b;b=h+(m*344|0)+56|0;k[b>>2]=k[h+(j*344|0)+56>>2];d=h+(j*344|0)+60|0;p=h+(m*344|0)+60|0;k[p>>2]=k[d>>2];if((k[d>>2]|0)>0){f=0;do{k[h+(m*344|0)+64+(f<<2)>>2]=k[h+(j*344|0)+64+(f<<2)>>2];f=f+1|0}while((f|0)<(k[d>>2]|0))}d=h+(j*344|0)+52|0;if((c|0)==289?(k[d>>2]|0)==262:0){k[b>>2]=1;k[p>>2]=1;k[h+(m*344|0)+64>>2]=1}switch(k[d>>2]|0){case 262:{k[h+(m*344|0)+4>>2]=7;break}case 261:{k[h+(m*344|0)+4>>2]=8;break}case 258:{k[h+(m*344|0)+4>>2]=9;break}case 259:{k[h+(m*344|0)+4>>2]=10;break}case 260:{k[h+(m*344|0)+4>>2]=11;break}default:{}}if(!n){g=m;r=e;return g|0}Sb[k[h+(m*344|0)+4>>2]&31](l);g=m;r=e;return g|0}function Lu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0;h=r;r=r+112|0;j=h+24|0;l=h;if((c|a|d|e|f|g|0)<0){m=-1;r=h;return m|0}n=k[30820]|0;if((k[n+(a*344|0)>>2]|0)!=-1e3?(k[n+(a*344|0)+56>>2]|0)!=1:0){o=k[30821]|0;do if((o|0)==(k[30822]|0)){k[30822]=o<<1;p=fy(n,o*688|0)|0;if(p){k[30820]=p;q=k[30821]|0;s=p;break}k[30835]=113;m=-1;r=h;return m|0}else{q=o;s=n}while(0);k[30821]=q+1;if((q|0)<=-1){m=q;r=h;return m|0}n=s+(q*344|0)|0;k[s+(q*344|0)+8>>2]=b+1;k[s+(q*344|0)+12>>2]=a;k[s+(q*344|0)+16>>2]=c;o=s+(c*344|0)|0;k[l>>2]=o;k[s+(q*344|0)+20>>2]=d;k[l+4>>2]=s+(d*344|0);k[s+(q*344|0)+24>>2]=e;k[l+8>>2]=s+(e*344|0);k[s+(q*344|0)+28>>2]=f;k[l+12>>2]=s+(f*344|0);k[s+(q*344|0)+32>>2]=g;k[l+16>>2]=s+(g*344|0);g=(k[s+(a*344|0)>>2]|0)==-1e3&1;f=(b|0)>0;a:do if(f){e=g;d=0;while(1){if(!e)t=0;else t=(k[k[l+(d<<2)>>2]>>2]|0)==-1e3;p=t&1;d=d+1|0;if((d|0)==(b|0)){u=p;break}else e=p}if(f){b:do if((k[s+(c*344|0)+56>>2]|0)<=1){e=o;d=0;while(1){d=d+1|0;if((k[e+52>>2]|0)!=259)break;if((d|0)>=(b|0)){v=u;break a}e=k[l+(d<<2)>>2]|0;if((k[e+56>>2]|0)>1)break b}e=k[30821]|0;if(e)k[30821]=e+-1;if(!(k[30835]|0))k[30835]=431;Bx(j,790562,80)|0;i[j+79>>0]=0;tk(j);m=-1;r=h;return m|0}while(0);e=k[30821]|0;if(e)k[30821]=e+-1;if(!(k[30835]|0))k[30835]=431;Bx(j,790524,80)|0;i[j+79>>0]=0;tk(j);m=-1;r=h;return m|0}else v=u}else v=g;while(0);k[n>>2]=91;g=s+(q*344|0)+4|0;k[g>>2]=12;k[s+(q*344|0)+52>>2]=k[s+(a*344|0)+52>>2];u=k[s+(a*344|0)+60>>2]|0;do if((u|0)==(b|0)){k[s+(q*344|0)+56>>2]=1;k[s+(q*344|0)+60>>2]=1;k[s+(q*344|0)+64>>2]=1}else{if((b|0)==1){l=s+(q*344|0)+60|0;k[l>>2]=u+-1;if((u|0)>1){o=1;c=0;while(1){f=k[s+(a*344|0)+64+(c<<2)>>2]|0;k[s+(q*344|0)+64+(c<<2)>>2]=f;t=ha(f,o)|0;c=c+1|0;if((c|0)>=(k[l>>2]|0)){w=t;break}else o=t}}else w=1;k[s+(q*344|0)+56>>2]=w;break}o=k[30821]|0;if(o)k[30821]=o+-1;if(!(k[30835]|0))k[30835]=431;Bx(j,790598,80)|0;i[j+79>>0]=0;tk(j);m=-1;r=h;return m|0}while(0);if(!v){m=q;r=h;return m|0}Sb[k[g>>2]&31](n);m=q;r=h;return m|0}if(!(k[30835]|0))k[30835]=431;Bx(j,790496,80)|0;i[j+79>>0]=0;tk(j);m=-1;r=h;return m|0}function Mu(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,r=0,s=0;if((e|d|f|g|h|i|j|0)<0){m=-1;return m|0}n=k[30821]|0;do if((n|0)==(k[30822]|0)){o=k[30820]|0;if(!o){k[30822]=100;p=cy(34400)|0}else{k[30822]=n<<1;p=fy(o,n*688|0)|0}if(p){k[30820]=p;q=k[30821]|0;break}k[30835]=113;m=-1;return m|0}else q=n;while(0);k[30821]=q+1;if((q|0)<=-1){m=q;return m|0}n=k[30820]|0;p=n+(q*344|0)|0;k[p>>2]=b;o=n+(q*344|0)+4|0;k[o>>2]=13;k[n+(q*344|0)+8>>2]=c;k[n+(q*344|0)+12>>2]=d;k[n+(q*344|0)+16>>2]=e;k[n+(q*344|0)+20>>2]=f;k[n+(q*344|0)+24>>2]=g;k[n+(q*344|0)+28>>2]=h;k[n+(q*344|0)+32>>2]=i;k[n+(q*344|0)+36>>2]=j;j=(b|0)==1043?0:c;if(!c)r=j;else{b=c;c=j;while(1){b=b+-1|0;if(!c)s=0;else s=(k[n+((k[n+(q*344|0)+12+(b<<2)>>2]|0)*344|0)>>2]|0)==-1e3;j=s&1;if(!b){r=j;break}else c=j}}if(!a){k[n+(q*344|0)+52>>2]=k[n+(d*344|0)+52>>2];k[n+(q*344|0)+56>>2]=k[n+(d*344|0)+56>>2];c=n+(d*344|0)+60|0;k[n+(q*344|0)+60>>2]=k[c>>2];if((k[c>>2]|0)>0){b=0;do{k[n+(q*344|0)+64+(b<<2)>>2]=k[n+(d*344|0)+64+(b<<2)>>2];b=b+1|0}while((b|0)<(k[c>>2]|0))}}else{k[n+(q*344|0)+52>>2]=a;k[n+(q*344|0)+56>>2]=1;k[n+(q*344|0)+60>>2]=1;k[n+(q*344|0)+64>>2]=1}if((l|0)>0)k[n+(q*344|0)+56>>2]=l;if(!r){m=q;return m|0}Sb[k[o>>2]&31](p);m=q;return m|0}
function dt(a,b,c,d,e,f,g,h,i,j,m,n,p,q,s,t,u){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;m=+m;n=+n;p=p|0;q=q|0;s=s|0;t=t|0;u=u|0;var v=0,w=0,x=0,y=0,z=0.0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0;v=r;r=r+16|0;w=v+4|0;x=v;k[w>>2]=0;k[x>>2]=0;if(n!=0.0|(m!=1.0|(j|0)!=-32&((j|0)!=32&(j|0)!=-64))){tk(779898);k[u>>2]=413;y=413;r=v;return y|0}k[p>>2]=4;p=a+4|0;j=k[p>>2]|0;do if((k[j+1148>>2]|0)>0){if((g|0)==1)z=+o[h>>2];else z=-9.119119840596153e-36;switch(k[j+1072>>2]|0){case 2:case 1:{A=7;break}case -1:{B=j;C=-1;D=0;break}default:{tk(780657);tk(780683);k[u>>2]=413;y=413;r=v;return y|0}}if((A|0)==7){E=k[j+1020>>2]|0;if(!E){if(!(k[j+1076>>2]|0)){F=Ra(0)|0;G=((bb()|0)/1e4|0)+F|0;F=k[p>>2]|0;H=F+1076|0;k[H>>2]=((G+(k[F+68>>2]|0)|0)%1e4|0)+1;hq(a,31,779134,H,0,u)|0}}else if((E|0)<0?(E=j+1076|0,(k[E>>2]|0)<0):0){H=d<<2;if((d|0)>0){F=0;G=0;while(1){I=(l[c+G>>0]|0)+F|0;G=G+1|0;if((G|0)>=(H|0)){J=I;break}else F=I}K=((J>>>0)%1e4|0)+1|0}else K=1;k[E>>2]=K;hq(a,31,779134,E,0,u)|0}F=k[p>>2]|0;B=F;C=k[F+1072>>2]|0;D=b+-1+(k[F+1076>>2]|0)|0}F=ut(D,c,e,f,g,z,+o[B+1068>>2],C,c,s,t,w,x)|0;k[q>>2]=F;if((F|0)>1){k[u>>2]=F;y=F;r=v;return y|0}}else{if(+o[j+1068>>2]!=9999.0){ft(c,d,c,g,+o[h>>2],i,u)|0;break}if((g|0)==1?(m=+o[h>>2],(d|0)>0):0){F=0;do{if(+o[c+(F<<2)>>2]==m)k[c+(F<<2)>>2]=-1;F=F+1|0}while((F|0)!=(d|0))}}while(0);y=k[u>>2]|0;r=v;return y|0}function et(a,b,c,d,e,f,g,h,i,j,m,n,q,s,t,u,v){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;m=+m;n=+n;q=q|0;s=s|0;t=t|0;u=u|0;v=v|0;var w=0,x=0,y=0,z=0,A=0.0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0;w=r;r=r+16|0;x=w+4|0;y=w;k[x>>2]=0;k[y>>2]=0;if(n!=0.0|(m!=1.0|(j|0)!=-32&((j|0)!=32&(j|0)!=-64))){tk(779898);k[v>>2]=413;z=413;r=w;return z|0}k[q>>2]=4;q=a+4|0;j=k[q>>2]|0;a:do if((k[j+1148>>2]|0)>0){if((g|0)==1)A=+p[h>>3];else A=-9.1191291391491e-36;switch(k[j+1072>>2]|0){case 2:case 1:{B=7;break}case -1:{C=j;D=-1;E=0;break}default:{tk(780731);tk(780683);k[v>>2]=413;z=413;r=w;return z|0}}if((B|0)==7){F=k[j+1020>>2]|0;if(!F){if(!(k[j+1076>>2]|0)){G=Ra(0)|0;H=((bb()|0)/1e4|0)+G|0;G=k[q>>2]|0;I=G+1076|0;k[I>>2]=((H+(k[G+68>>2]|0)|0)%1e4|0)+1;hq(a,31,779134,I,0,v)|0}}else if((F|0)<0?(F=j+1076|0,(k[F>>2]|0)<0):0){I=d<<3;if((d|0)>0){G=0;H=0;while(1){J=(l[c+H>>0]|0)+G|0;H=H+1|0;if((H|0)>=(I|0)){K=J;break}else G=J}L=((K>>>0)%1e4|0)+1|0}else L=1;k[F>>2]=L;hq(a,31,779134,F,0,v)|0}G=k[q>>2]|0;C=G;D=k[G+1072>>2]|0;E=b+-1+(k[G+1076>>2]|0)|0}G=vt(E,c,e,f,g,A,+o[C+1068>>2],D,c,t,u,x,y)|0;k[s>>2]=G;if((G|0)>1){k[v>>2]=G;z=G;r=w;return z|0}}else{if(!(+o[j+1068>>2]!=9999.0)){if((g|0)!=1)break;m=+p[h>>3];if((d|0)>0)M=0;else break;while(1){if(+p[c+(M<<3)>>3]==m){G=c+(M<<3)|0;k[G>>2]=-1;k[G+4>>2]=-1}M=M+1|0;if((M|0)==(d|0))break a}}m=+p[h>>3];F=(d|0)>0;if((g|0)!=1){if(F)N=0;else break;while(1){n=+p[c+(N<<3)>>3];do if(!(n<-2147483648.49)){if(n>2147483647.49){k[v>>2]=-11;k[c+(N<<2)>>2]=2147483647;break}if(!(n>=0.0)){k[c+(N<<2)>>2]=~~(n+-.5);break}else{k[c+(N<<2)>>2]=~~(n+.5);break}}else{k[v>>2]=-11;k[c+(N<<2)>>2]=-2147483648}while(0);N=N+1|0;if((N|0)==(d|0))break a}}if(F){G=0;do{n=+p[c+(G<<3)>>3];do if(!(n==m)){if(n<-2147483648.49){k[v>>2]=-11;k[c+(G<<2)>>2]=-2147483648;break}if(n>2147483647.49){k[v>>2]=-11;k[c+(G<<2)>>2]=2147483647;break}if(!(n>=0.0)){k[c+(G<<2)>>2]=~~(n+-.5);break}else{k[c+(G<<2)>>2]=~~(n+.5);break}}else k[c+(G<<2)>>2]=i;while(0);G=G+1|0}while((G|0)!=(d|0))}}while(0);z=k[v>>2]|0;r=w;return z|0}function ft(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=+e;f=f|0;g=g|0;var h=0,i=0,j=0,l=0.0,m=0.0,n=0;h=(b|0)>0;if((d|0)!=1){if(h)i=0;else{j=k[g>>2]|0;return j|0}do{l=+o[a+(i<<2)>>2];m=l;do if(!(m<-2147483648.49)){if(m>2147483647.49){k[g>>2]=-11;k[c+(i<<2)>>2]=2147483647;break}if(!(l>=0.0)){k[c+(i<<2)>>2]=~~(m+-.5);break}else{k[c+(i<<2)>>2]=~~(m+.5);break}}else{k[g>>2]=-11;k[c+(i<<2)>>2]=-2147483648}while(0);i=i+1|0}while((i|0)!=(b|0));j=k[g>>2]|0;return j|0}if(h)n=0;else{j=k[g>>2]|0;return j|0}do{m=+o[a+(n<<2)>>2];do if(!(m==e)){l=m;if(l<-2147483648.49){k[g>>2]=-11;k[c+(n<<2)>>2]=-2147483648;break}if(l>2147483647.49){k[g>>2]=-11;k[c+(n<<2)>>2]=2147483647;break}if(!(m>=0.0)){k[c+(n<<2)>>2]=~~(l+-.5);break}else{k[c+(n<<2)>>2]=~~(l+.5);break}}else k[c+(n<<2)>>2]=f;while(0);n=n+1|0}while((n|0)!=(b|0));j=k[g>>2]|0;return j|0}function gt(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;var i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0;i=r;r=r+320|0;j=i+200|0;l=i+176|0;m=i+152|0;n=i+128|0;o=i+104|0;p=i+80|0;q=i+56|0;s=i+32|0;t=i+8|0;u=i+4|0;v=i;w=i+224|0;x=k[h>>2]|0;if((x|0)>0){y=x;r=i;return y|0}if(!(Cl(a,h)|0)){tk(780769);k[h>>2]=413;y=413;r=i;return y|0}x=k[a>>2]|0;z=a+4|0;A=k[z>>2]|0;if((x|0)==(k[A+68>>2]|0)){B=A+120|0;if(((k[B>>2]|0)==-1?(k[B+4>>2]|0)==-1:0)?(Vk(a,h)|0)>0:0){y=k[h>>2]|0;r=i;return y|0}}else Fk(a,x+1|0,0,h)|0;x=b&-2;a:do if((x|0)!=20)if((x|0)!=30)if((b+-11|0)>>>0>=2)if((x|0)==40)C=4;else{switch(b|0){case 42:{C=4;break a;break}case 82:{C=8;break a;break}default:{}}tk(780828);k[h>>2]=410;y=410;r=i;return y|0}else C=1;else C=4;else C=2;while(0);x=k[z>>2]|0;b:do switch(k[x+1040>>2]|0){case 41:{switch(k[x+1096>>2]|0){case 16:case 8:break;default:{D=8;break b}}D=C>>>0>4?C:4;break}case 31:{D=C>>>0>4?C:4;break}case 51:case 22:case 21:case 11:{switch(k[x+1096>>2]|0){case 8:{D=C>>>0>1?C:1;break b;break}case 16:{D=C>>>0>2?C:2;break b;break}default:{D=C>>>0>4?C:4;break b}}break}default:{tk(780871);k[h>>2]=410;y=410;r=i;return y|0}}while(0);B=ey(k[x+1128>>2]|0,D)|0;if(!B){tk(780911);k[h>>2]=113;y=113;r=i;return y|0}k[j>>2]=1;k[l>>2]=1;k[m>>2]=1;k[n>>2]=1;k[q>>2]=1;D=j+4|0;k[D>>2]=1;A=l+4|0;k[A>>2]=1;E=m+4|0;k[E>>2]=1;F=n+4|0;k[F>>2]=1;G=q+4|0;k[G>>2]=1;H=j+8|0;k[H>>2]=1;I=l+8|0;k[I>>2]=1;J=m+8|0;k[J>>2]=1;K=n+8|0;k[K>>2]=1;L=q+8|0;k[L>>2]=1;M=j+12|0;k[M>>2]=1;N=l+12|0;k[N>>2]=1;O=m+12|0;k[O>>2]=1;P=n+12|0;k[P>>2]=1;Q=q+12|0;k[Q>>2]=1;R=j+16|0;k[R>>2]=1;S=l+16|0;k[S>>2]=1;T=m+16|0;k[T>>2]=1;U=n+16|0;k[U>>2]=1;V=q+16|0;k[V>>2]=1;W=j+20|0;k[W>>2]=1;X=l+20|0;k[X>>2]=1;Y=m+20|0;k[Y>>2]=1;Z=n+20|0;k[Z>>2]=1;_=q+20|0;k[_>>2]=1;$=k[x+1100>>2]|0;do if(($|0)>0){aa=0;ba=1;while(1){ca=k[c+(aa<<2)>>2]|0;k[s+(aa<<2)>>2]=ca;da=k[d+(aa<<2)>>2]|0;k[t+(aa<<2)>>2]=da;ea=k[x+1104+(aa<<2)>>2]|0;k[j+(aa<<2)>>2]=ea;if((ca|0)<1)break;fa=k[x+1044+(aa<<2)>>2]|0;k[l+(aa<<2)>>2]=fa;ga=((ea+-1|0)/(fa|0)|0)+1|0;k[m+(aa<<2)>>2]=((ca+-1|0)/(fa|0)|0)+1;ca=((da+-1|0)/(fa|0)|0)+1|0;k[n+(aa<<2)>>2]=(ca|0)<(ga|0)?ca:ga;k[q+(aa<<2)>>2]=ba;ba=ha(ga,ba)|0;aa=aa+1|0;if((aa|0)>=($|0)){ia=31;break}}if((ia|0)==31){aa=k[Y>>2]|0;ba=k[Z>>2]|0;if((aa|0)>(ba|0))break;else{ja=ba;ka=aa;ia=32;break}}dy(B);k[h>>2]=321;y=321;r=i;return y|0}else{ja=1;ka=1;ia=32}while(0);if((ia|0)==32){ia=k[X>>2]|0;X=o+20|0;Z=k[W>>2]|0;W=p+20|0;Y=k[_>>2]|0;_=k[T>>2]|0;T=k[U>>2]|0;U=(_|0)>(T|0);q=k[S>>2]|0;S=o+16|0;x=k[R>>2]|0;R=p+16|0;d=k[V>>2]|0;V=k[O>>2]|0;O=k[P>>2]|0;P=(V|0)>(O|0);c=k[N>>2]|0;N=o+12|0;aa=k[M>>2]|0;M=p+12|0;ba=k[Q>>2]|0;Q=k[J>>2]|0;J=k[K>>2]|0;K=(Q|0)>(J|0);ga=k[I>>2]|0;I=o+8|0;ca=k[H>>2]|0;H=p+8|0;fa=k[L>>2]|0;L=k[E>>2]|0;E=k[F>>2]|0;F=(L|0)>(E|0);da=k[A>>2]|0;A=o+4|0;ea=k[D>>2]|0;D=p+4|0;la=k[G>>2]|0;G=k[m>>2]|0;m=k[n>>2]|0;n=(G|0)>(m|0);ma=k[l>>2]|0;l=k[j>>2]|0;j=ka;while(1){ka=j+-1|0;na=ha(ia,ka)|0;k[X>>2]=na+1;oa=ha(ia,j)|0;pa=(oa|0)<(Z|0)?oa:Z;k[W>>2]=pa;oa=pa-na|0;na=ha(Y,ka)|0;if(!U){ka=_;while(1){pa=ka+-1|0;qa=ha(q,pa)|0;k[S>>2]=qa+1;ra=ha(q,ka)|0;sa=(ra|0)<(x|0)?ra:x;k[R>>2]=sa;ra=ha(sa-qa|0,oa)|0;qa=(ha(d,pa)|0)+na|0;if(!P){pa=V;while(1){sa=pa+-1|0;ta=ha(c,sa)|0;k[N>>2]=ta+1;ua=ha(c,pa)|0;va=(ua|0)<(aa|0)?ua:aa;k[M>>2]=va;ua=ha(ra,va-ta|0)|0;ta=qa+(ha(ba,sa)|0)|0;if(!K){sa=Q;while(1){va=sa+-1|0;wa=ha(ga,va)|0;k[I>>2]=wa+1;xa=ha(ga,sa)|0;ya=(xa|0)<(ca|0)?xa:ca;k[H>>2]=ya;xa=ha(ua,ya-wa|0)|0;wa=ta+(ha(fa,va)|0)|0;if(!F){va=L;while(1){ya=va+-1|0;za=ha(da,ya)|0;Aa=za+1|0;k[A>>2]=Aa;Ba=ha(da,va)|0;Ca=(Ba|0)<(ea|0)?Ba:ea;k[D>>2]=Ca;Ba=ha(xa,Ca-za|0)|0;za=wa+(ha(la,ya)|0)|0;if(!n){ya=Ca+1-Aa|0;Aa=G;while(1){Ca=ha(ma,Aa+-1|0)|0;Da=Ca+1|0;k[o>>2]=Da;Ea=ha(ma,Aa)|0;Fa=(Ea|0)<(l|0)?Ea:l;k[p>>2]=Fa;Ea=ha(Ba,Fa-Ca|0)|0;Ca=za+Aa|0;ht(a,Ca,Ea,b,e,g,B,0,u,h)|0;if((k[h>>2]|0)==415){sy(B|0,0,ha(Ea,C)|0)|0;k[h>>2]=0}it(B,C,$,o,p,0,f,s,t,0,h)|0;_s(a,Ca,b,B,Ea,Fa+1-Da|0,ya,e,g,h)|0;if((Aa|0)<(m|0))Aa=Aa+1|0;else break}}if((va|0)<(E|0))va=va+1|0;else break}}if((sa|0)<(J|0))sa=sa+1|0;else break}}if((pa|0)<(O|0))pa=pa+1|0;else break}}if((ka|0)<(T|0))ka=ka+1|0;else break}}if((j|0)<(ja|0))j=j+1|0;else break}}dy(B);if(((e|0)!=0?(k[(k[z>>2]|0)+1096>>2]|0)<0:0)?(k[v>>2]=0,Go(a,779674,w,v)|0,(k[v>>2]|0)!=0):0){Go(a,779256,w,h)|0;Eq(a,779674,-2147483647,-1,779681,h)|0;a=k[z>>2]|0;k[a+1200>>2]=-2147483647;k[a+1156>>2]=-1}y=k[h>>2]|0;r=i;return y|0}function ht(a,b,c,d,e,f,g,h,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;m=m|0;n=n|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0.0,V=0.0,W=0.0,X=0.0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0.0,ga=0.0,ia=0.0,ja=0.0,ka=0,la=0.0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0.0,sa=0.0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0.0,Da=0.0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0.0,Qa=0.0,Ra=0,Sa=0.0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0.0,Za=0.0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0.0,ib=0.0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0;q=r;r=r+80|0;s=q+72|0;t=q+68|0;u=q+64|0;v=q+60|0;w=q+56|0;x=q+78|0;y=q+76|0;z=q+52|0;A=q+32|0;B=q+24|0;C=q+16|0;D=q+48|0;E=q+44|0;F=q+40|0;G=q+8|0;H=q;k[s>>2]=g;k[t>>2]=0;i[x>>0]=0;j[y>>1]=0;k[z>>2]=0;p[C>>3]=0.0;g=G;k[g>>2]=0;k[g+4>>2]=0;g=H;k[g>>2]=0;k[g+4>>2]=0;g=k[n>>2]|0;if((g|0)>0){I=g;r=q;return I|0}g=a+4|0;J=k[g>>2]|0;K=J+1220|0;M=k[K>>2]|0;do if(!M){N=k[J+1104>>2]|0;O=k[J+1044>>2]|0;P=((N+-1|0)/(O|0)|0)+1|0;if((N|0)==(O|0)?(k[J+1048>>2]|0)==1:0){Q=0;break}O=ey(P,4)|0;k[K>>2]=O;O=ey(P,4)|0;k[J+1232>>2]=O;O=ey(P,4)|0;k[J+1236>>2]=O;O=ey(P,4)|0;k[J+1224>>2]=O;O=ey(P,4)|0;k[J+1228>>2]=O;O=ey(P,4)|0;k[J+1240>>2]=O;O=k[K>>2]|0;if(!O)Q=0;else{R=O;S=6}}else{R=M;S=6}while(0);if((S|0)==6){M=(b+-1|0)%((((k[J+1104>>2]|0)+-1|0)/(k[J+1044>>2]|0)|0)+1|0)|0;if((k[R+(M<<2)>>2]|0)==(b|0)?(k[(k[J+1228>>2]|0)+(M<<2)>>2]|0)==(d|0):0){py(k[s>>2]|0,k[(k[J+1232>>2]|0)+(M<<2)>>2]|0,k[(k[J+1224>>2]|0)+(M<<2)>>2]|0)|0;if((e|0)==2)py(h|0,k[(k[(k[g>>2]|0)+1236>>2]|0)+(M<<2)>>2]|0,c|0)|0;k[m>>2]=k[(k[(k[g>>2]|0)+1240>>2]|0)+(M<<2)>>2];I=k[n>>2]|0;r=q;return I|0}else Q=M}M=((b|0)<0)<<31>>31;pl(a,k[J+1136>>2]|0,b,M,G,H,n)|0;if((k[n>>2]|0)==107){k[n>>2]=415;I=415;r=q;return I|0}J=G;if((k[J>>2]|0)==0&(k[J+4>>2]|0)==0){J=k[g>>2]|0;R=k[J+1140>>2]|0;do if((R|0)>0){pl(a,R,b,M,G,H,n)|0;K=G;O=k[K>>2]|0;P=H;if((k[P>>2]|O|0)==0?(k[P+4>>2]|k[K+4>>2]|0)==0:0){k[n>>2]=415;I=415;r=q;return I|0}K=k[(k[g>>2]|0)+1140>>2]|0;P=my(0,O|0,32)|0;O=L;N=k[s>>2]|0;if((e|0)<2){_l(a,d,K,b,M,1,0,P,O,f,N,m,n)|0;break}else{$l(a,d,K,b,M,1,0,P,O,N,h,m,n)|0;break}}else{N=k[J+1144>>2]|0;if((N|0)<=0){k[n>>2]=415;break}pl(a,N,b,M,G,H,n)|0;N=G;O=k[N>>2]|0;P=H;if((k[P>>2]|O|0)==0?(k[P+4>>2]|k[N+4>>2]|0)==0:0){k[n>>2]=415;I=415;r=q;return I|0}N=cy(O)|0;if(!N){tk(780951);k[n>>2]=113;I=113;r=q;return I|0}P=k[(k[g>>2]|0)+1144>>2]|0;K=my(0,O|0,32)|0;if((_l(a,11,P,b,M,1,0,K,L,x,N,0,n)|0)>0){tk(781017);dy(N);I=k[n>>2]|0;r=q;return I|0}K=(k[g>>2]|0)+1096|0;switch(k[K>>2]|0){case -32:{T=c<<2;break}case -64:{T=c<<3;break}default:{tk(781072);dy(N);k[n>>2]=414;I=414;r=q;return I|0}}k[u>>2]=T;P=(d|0)==82;if(P?(k[K>>2]|0)==-32:0){K=cy(T)|0;k[z>>2]=K;if(!K){tk(781143);dy(N);k[n>>2]=113;I=113;r=q;return I|0}if(qh(N,k[G>>2]|0,z,u,0,v,n)|0){tk(781209);dy(k[z>>2]|0);dy(N);I=k[n>>2]|0;r=q;return I|0}}else S=37;if((S|0)==37?(qh(N,k[G>>2]|0,s,u,0,v,n)|0)!=0:0){tk(781209);dy(N);I=k[n>>2]|0;r=q;return I|0}dy(N);N=k[v>>2]|0;if((N|0)==(c<<2|0)){K=k[z>>2]|0;if(!K)Ls(k[s>>2]|0,c);else Ls(K,c);if((d|0)==42){if(!f)U=0.0;else U=+o[f>>2];K=k[s>>2]|0;Fm(K,c,1.0,0.0,e,U,h,m,K,n)|0;break}if(!P){tk(781241);k[n>>2]=414;I=414;r=q;return I|0}if(!f)V=0.0;else V=+p[f>>3];ym(k[z>>2]|0,c,1.0,0.0,e,V,h,m,k[s>>2]|0,n)|0;dy(k[z>>2]|0);break}else{if((N|0)!=(c<<3|0)){tk(781385);k[n>>2]=414;I=414;r=q;return I|0}Ms(k[s>>2]|0,c);if((d|0)==42){if(!f)W=0.0;else W=+o[f>>2];N=k[s>>2]|0;Km(N,c,1.0,0.0,e,W,h,m,N,n)|0;break}if(!P){tk(781312);k[n>>2]=414;I=414;r=q;return I|0}if(!f)X=0.0;else X=+p[f>>3];P=k[s>>2]|0;tm(P,c,1.0,0.0,e,X,h,m,P,n)|0;break}}while(0);I=k[n>>2]|0;r=q;return I|0}if((e|0)==2&(c|0)>0)sy(h|0,0,c|0)|0;if(m)k[m>>2]=0;z=k[g>>2]|0;X=+p[z+1192>>3];T=k[z+1148>>2]|0;a:do switch(T|0){case 0:{p[A>>3]=1.0;p[B>>3]=0.0;Y=z;break}case -1:{p[A>>3]=+p[z+1160>>3];p[B>>3]=+p[z+1168>>3];Y=z;break}default:{qm(a,T,b,M,1,0,1,0,0.0,A,0,n)|0;qm(a,k[(k[g>>2]|0)+1152>>2]|0,b,M,1,0,1,0,0.0,B,0,n)|0;if((k[n>>2]|0)>0){tk(781425);I=k[n>>2]|0;r=q;return I|0}H=k[g>>2]|0;switch(k[H+1096>>2]|0){case -64:case -32:break;default:{Y=H;break a}}J=H+1176|0;W=+p[J>>3];R=H+1184|0;if(!(W!=1.0)?!(+p[R>>3]!=0.0):0){Y=H;break a}p[A>>3]=+p[A>>3]*W;p[B>>3]=+p[B>>3]*+p[J>>3]+ +p[R>>3];Y=H}}while(0);if(+p[A>>3]==1.0&+p[B>>3]==0.0){p[A>>3]=+p[Y+1176>>3];p[B>>3]=+p[Y+1184>>3]}T=k[Y+1156>>2]|0;b:do switch(T|0){case 0:{Z=0;_=Y;break}case -1:{k[w>>2]=k[Y+1200>>2];Z=e;_=Y;break}default:{zn(a,T,b,M,1,0,1,0,0,w,0,n)|0;if((k[n>>2]|0)<=0){Z=e;_=k[g>>2]|0;break b}tk(781485);I=k[n>>2]|0;r=q;return I|0}}while(0);e=_+1040|0;T=k[e>>2]|0;c:do if((T|0)!=41){Y=(T|0)==11;if(Y){do if((k[_+1096>>2]|0)==8)if((k[_+1208>>2]|0)!=1)if(Y)break;else{S=105;break c}else{k[u>>2]=c;$=c;break c}while(0);if((k[_+1096>>2]|0)!=16){S=105;break}if((k[_+1208>>2]|0)!=2){S=105;break}Y=c<<1;k[u>>2]=Y;$=Y;break}switch(T|0){case 51:case 22:case 21:break;default:{S=105;break c}}if((k[_+1096>>2]|0)==8){k[u>>2]=c;$=c;break}switch(T|0){case 51:case 22:case 21:break;default:{S=105;break c}}if((k[_+1096>>2]|0)==16){Y=c<<1;k[u>>2]=Y;$=Y;break}switch(T|0){case 51:case 22:case 21:break;default:{S=105;break c}}if((k[_+1096>>2]|0)==-64){Y=c<<3;k[u>>2]=Y;$=Y}else S=105}else{switch(k[_+1096>>2]|0){case 16:case 8:{S=105;break c;break}default:{}}Y=c<<3;k[u>>2]=Y;$=Y}while(0);if((S|0)==105){S=c<<2;k[u>>2]=S;$=S}S=cy($)|0;k[t>>2]=S;if(!S){tk(781530);k[n>>2]=113;I=113;r=q;return I|0}S=k[G>>2]|0;if((k[e>>2]|0)==31)aa=cy(S<<1)|0;else aa=cy(S)|0;if(!aa){tk(781592);dy(k[t>>2]|0);k[n>>2]=113;I=113;r=q;return I|0}$=(k[e>>2]|0)==31;e=k[_+1136>>2]|0;_=my(0,S|0,32)|0;S=L;if($)_l(a,21,e,b,M,1,0,_,S,y,aa,0,n)|0;else _l(a,11,e,b,M,1,0,_,S,x,aa,0,n)|0;if((k[n>>2]|0)>0){tk(781017);dy(aa);dy(k[t>>2]|0);I=k[n>>2]|0;r=q;return I|0}x=k[g>>2]|0;d:do switch(k[x+1040>>2]|0){case 11:{S=k[x+1204>>2]|0;_=k[x+1208>>2]|0;if((_|0)==1){M=Ht(aa,k[G>>2]|0,k[t>>2]|0,c,S)|0;k[n>>2]=M;ba=11;break d}M=k[G>>2]|0;e=k[t>>2]|0;if((_|0)==2){_=Gt(aa,M,e,c,S)|0;k[n>>2]=_;ba=21;break d}else{_=Ft(aa,M,e,c,S)|0;k[n>>2]=_;ba=31;break d}break}case 41:{_=k[x+1216>>2]|0;switch(k[x+1096>>2]|0){case 16:case 8:{ca=Tt(aa,_,k[t>>2]|0,D,E,F,n)|0;break}default:ca=Ut(aa,_,k[t>>2]|0,D,E,F,n)|0}k[n>>2]=ca;ba=31;break}case 31:{Jt(aa,1,k[t>>2]|0,c)|0;ba=31;break}case 22:case 21:{qh(aa,k[G>>2]|0,t,u,19,v,n)|0;_=k[v>>2]|0;S=c<<1;if((_|0)==(S|0)){if((k[(k[g>>2]|0)+1040>>2]|0)==22){e=k[t>>2]|0;M=((c|0)<0)<<31>>31;a=cy(S)|0;y=S+-1|0;if((c|0)>0){$=0-c|0;T=0;Y=0;z=a+y|0;H=e+y|0;while(1){i[z>>0]=i[H>>0]|0;i[z+-1>>0]=i[H+$>>0]|0;T=ky(T|0,Y|0,1,0)|0;Y=L;if(!((Y|0)<(M|0)|(Y|0)==(M|0)&T>>>0<c>>>0))break;else{z=z+-2|0;H=H+-1|0}}}py(e|0,a|0,S|0)|0;dy(a)}Ks(k[t>>2]|0,c);ba=21;break d}H=c<<2;if((_|0)==(H|0)){if((k[(k[g>>2]|0)+1040>>2]|0)==22){z=k[t>>2]|0;T=((c|0)<0)<<31>>31;M=cy(H)|0;Y=H+-1|0;if((c|0)>0){$=0-c|0;y=0-S|0;R=ha(c,-3)|0;J=0;P=0;N=M+Y|0;K=z+Y|0;while(1){i[N>>0]=i[K>>0]|0;i[N+-1>>0]=i[K+$>>0]|0;i[N+-2>>0]=i[K+y>>0]|0;i[N+-3>>0]=i[K+R>>0]|0;J=ky(J|0,P|0,1,0)|0;P=L;if(!((P|0)<(T|0)|(P|0)==(T|0)&J>>>0<c>>>0))break;else{N=N+-4|0;K=K+-1|0}}}py(z|0,M|0,H|0)|0;dy(M)}Ls(k[t>>2]|0,c);ba=31;break d}if((_|0)==(c<<3|0)){if((k[(k[g>>2]|0)+1040>>2]|0)==22)tt(k[t>>2]|0,c,((c|0)<0)<<31>>31);Ms(k[t>>2]|0,c);ba=82;break d}else{if((_|0)==(c|0)){ba=11;break d}tk(781385);dy(k[t>>2]|0);k[n>>2]=414;I=414;r=q;return I|0}break}case 51:{tk(781641);dy(k[t>>2]|0);dy(aa);k[n>>2]=414;I=414;r=q;return I|0}default:{tk(781667);dy(k[t>>2]|0);k[n>>2]=414;I=414;r=q;return I|0}}while(0);dy(aa);if(k[n>>2]|0){dy(k[t>>2]|0);I=k[n>>2]|0;r=q;return I|0}aa=(f|0)==0?C:f;e:do switch(d|0){case 21:{f=k[g>>2]|0;C=(ba|0)==31;if(+o[f+1068>>2]==9999.0){v=k[t>>2]|0;if(C){Um(v,c,+p[A>>3],+p[B>>3],Z,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}else{Vm(v,c,+p[A>>3],+p[B>>3],Z,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}}if(!C)switch(ba|0){case 21:{Qm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e;break}case 11:{Sm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e;break}default:{da=2;break e}}C=k[t>>2]|0;W=+p[A>>3];V=+p[B>>3];if(X==32768.0?(k[f+1040>>2]|0)==31:0){Tm(C,c,W,V+-32768.0,Z,k[w>>2]|0,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}Tm(C,c,W,V,Z,k[w>>2]|0,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;if((k[(k[g>>2]|0)+1040>>2]|0)==41?(k[n>>2]|0)==412:0){k[n>>2]=0;da=2}else da=2;break}case 31:{C=k[g>>2]|0;f=(ba|0)==31;if(+o[C+1068>>2]==9999.0){v=k[t>>2]|0;if(f){Dn(v,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{En(v,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}if(f){f=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];if(X==32768.0?(k[C+1040>>2]|0)==31:0){Cn(f,c,V,W+-32768.0,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{Cn(f,c,V,W,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}else switch(ba|0){case 21:{Bn(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}case 11:{An(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}default:{da=4;break e}}break}case 41:{f=k[g>>2]|0;C=(ba|0)==31;if(+o[f+1068>>2]==9999.0){v=k[t>>2]|0;if(C){fn(v,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{gn(v,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}if(C){C=k[t>>2]|0;W=+p[A>>3];V=+p[B>>3];if(X==32768.0?(k[f+1040>>2]|0)==31:0){bn(C,c,W,V+-32768.0,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{bn(C,c,W,V,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}else switch(ba|0){case 21:{en(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}case 11:{dn(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}default:{da=4;break e}}break}case 42:{V=+o[aa>>2];C=k[g>>2]|0;if(+o[C+1068>>2]==9999.0){f=k[t>>2]|0;if((ba|0)==31){Fm(f,c,+p[A>>3],+p[B>>3],Z,V,h,m,k[s>>2]|0,n)|0;da=4;break e}else{Km(f,c,+p[A>>3],+p[B>>3],Z,V,h,m,k[s>>2]|0,n)|0;da=4;break e}}f=k[C+1072>>2]|0;if((f+-1|0)>>>0>=2)switch(ba|0){case 31:{v=k[t>>2]|0;W=+p[A>>3];U=+p[B>>3];if(X==32768.0?(k[C+1040>>2]|0)==31:0){Im(v,c,W,U+-32768.0,Z,k[w>>2]|0,V,h,m,k[s>>2]|0,n)|0;da=4;break e}else{Im(v,c,W,U,Z,k[w>>2]|0,V,h,m,k[s>>2]|0,n)|0;da=4;break e}break}case 21:{Hm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,V,h,m,k[s>>2]|0,n)|0;da=4;break e;break}case 11:{Gm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,V,h,m,k[s>>2]|0,n)|0;da=4;break e;break}default:{da=4;break e}}switch(ba|0){case 31:{v=k[C+1076>>2]|0;u=k[t>>2]|0;U=+p[A>>3];W=+p[B>>3];G=k[w>>2]|0;ca=k[s>>2]|0;F=k[30368]|0;do if(!F){E=ey(1e4,4)|0;k[30368]=E;if(!E){da=4;break e}else{ea=0;fa=1.0}while(1){ga=fa*16807.0;ia=ga-+(~~(ga/2147483647.0)|0)*2147483647.0;o[E+(ea<<2)>>2]=ia/2147483647.0;ea=ea+1|0;if((ea|0)==1e4){ja=ia;break}else fa=ia}if((~~ja|0)==1043618065){ka=E;break}tk(778042);da=4;break e}else ka=F;while(0);F=(b+-2+v|0)%1e4|0;_=~~(+o[ka+(F<<2)>>2]*500.0);M=(c|0)>0;if(!Z){if(!M){da=4;break e}H=(f|0)==2;z=0;D=F;x=_;while(1){K=k[u+(z<<2)>>2]|0;if(H&(K|0)==-2147483646)la=0.0;else la=W+U*(+(K|0)-+o[ka+(x<<2)>>2]+.5);o[ca+(z<<2)>>2]=la;K=x+1|0;if((K|0)==1e4){N=D+1|0;J=(N|0)==1e4?0:N;ma=J;na=~~(+o[ka+(J<<2)>>2]*500.0)}else{ma=D;na=K}z=z+1|0;if((z|0)==(c|0)){da=4;break e}else{D=ma;x=na}}}if(!M){da=4;break e}x=(Z|0)==1;D=(f|0)==2;z=0;H=F;v=_;while(1){K=k[u+(z<<2)>>2]|0;do if((K|0)==(G|0)){k[m>>2]=1;if(x){o[ca+(z<<2)>>2]=V;break}else{i[h+z>>0]=1;break}}else if(D&(K|0)==-2147483646){o[ca+(z<<2)>>2]=0.0;break}else{o[ca+(z<<2)>>2]=W+U*(+(K|0)-+o[(k[30368]|0)+(v<<2)>>2]+.5);break}while(0);K=v+1|0;if((K|0)==1e4){E=H+1|0;J=(E|0)==1e4?0:E;oa=J;pa=~~(+o[(k[30368]|0)+(J<<2)>>2]*500.0)}else{oa=H;pa=K}z=z+1|0;if((z|0)==(c|0)){da=4;break}else{H=oa;v=pa}}break}case 21:{v=k[C+1076>>2]|0;H=k[t>>2]|0;U=+p[A>>3];W=+p[B>>3];z=k[w>>2]&65535;ca=k[s>>2]|0;D=k[30368]|0;do if(!D){x=ey(1e4,4)|0;k[30368]=x;if(!x){da=4;break e}else{qa=0;ra=1.0}while(1){ia=ra*16807.0;ga=ia-+(~~(ia/2147483647.0)|0)*2147483647.0;o[x+(qa<<2)>>2]=ga/2147483647.0;qa=qa+1|0;if((qa|0)==1e4){sa=ga;break}else ra=ga}if((~~sa|0)==1043618065){ta=x;break}tk(778042);da=4;break e}else ta=D;while(0);D=(b+-2+v|0)%1e4|0;G=~~(+o[ta+(D<<2)>>2]*500.0);u=(c|0)>0;if(!Z){if(u){ua=0;va=D;wa=G}else{da=4;break e}while(1){o[ca+(ua<<2)>>2]=W+U*(+(j[H+(ua<<1)>>1]|0)-+o[ta+(wa<<2)>>2]+.5);_=wa+1|0;if((_|0)==1e4){F=va+1|0;f=(F|0)==1e4?0:F;xa=f;ya=~~(+o[ta+(f<<2)>>2]*500.0)}else{xa=va;ya=_}ua=ua+1|0;if((ua|0)==(c|0)){da=4;break e}else{va=xa;wa=ya}}}if(!u){da=4;break e}v=(Z|0)==1;_=0;f=D;F=G;while(1){M=j[H+(_<<1)>>1]|0;do if(M<<16>>16==z<<16>>16){k[m>>2]=1;if(v){o[ca+(_<<2)>>2]=V;break}else{i[h+_>>0]=1;break}}else o[ca+(_<<2)>>2]=W+U*(+(M<<16>>16)-+o[(k[30368]|0)+(F<<2)>>2]+.5);while(0);M=F+1|0;if((M|0)==1e4){x=f+1|0;K=(x|0)==1e4?0:x;za=K;Aa=~~(+o[(k[30368]|0)+(K<<2)>>2]*500.0)}else{za=f;Aa=M}_=_+1|0;if((_|0)==(c|0)){da=4;break}else{f=za;F=Aa}}break}case 11:{F=k[C+1076>>2]|0;f=k[t>>2]|0;U=+p[A>>3];W=+p[B>>3];_=k[w>>2]&255;ca=k[s>>2]|0;v=k[30368]|0;do if(!v){z=ey(1e4,4)|0;k[30368]=z;if(!z){da=4;break e}else{Ba=0;Ca=1.0}while(1){ga=Ca*16807.0;ia=ga-+(~~(ga/2147483647.0)|0)*2147483647.0;o[z+(Ba<<2)>>2]=ia/2147483647.0;Ba=Ba+1|0;if((Ba|0)==1e4){Da=ia;break}else Ca=ia}if((~~Da|0)==1043618065){Ea=z;break}tk(778042);da=4;break e}else Ea=v;while(0);v=(b+-2+F|0)%1e4|0;C=~~(+o[Ea+(v<<2)>>2]*500.0);H=(c|0)>0;if(!Z){if(H){Fa=Ea;Ga=0;Ha=v;Ia=C}else{da=4;break e}while(1){o[ca+(Ga<<2)>>2]=W+U*(+(l[f+Ga>>0]|0)-+o[Fa+(Ia<<2)>>2]+.5);G=Ia+1|0;if((G|0)==1e4){D=Ha+1|0;u=(D|0)==1e4?0:D;Ja=Ea;Ka=u;La=~~(+o[Ea+(u<<2)>>2]*500.0)}else{Ja=Fa;Ka=Ha;La=G}Ga=Ga+1|0;if((Ga|0)==(c|0)){da=4;break e}else{Fa=Ja;Ha=Ka;Ia=La}}}if(!H){da=4;break e}F=(Z|0)==1;G=0;u=v;D=C;while(1){M=i[f+G>>0]|0;do if(M<<24>>24==_<<24>>24){k[m>>2]=1;if(F){o[ca+(G<<2)>>2]=V;break}else{i[h+G>>0]=1;break}}else o[ca+(G<<2)>>2]=W+U*(+(M&255)-+o[(k[30368]|0)+(D<<2)>>2]+.5);while(0);M=D+1|0;if((M|0)==1e4){z=u+1|0;K=(z|0)==1e4?0:z;Ma=K;Na=~~(+o[(k[30368]|0)+(K<<2)>>2]*500.0)}else{Ma=u;Na=M}G=G+1|0;if((G|0)==(c|0)){da=4;break}else{u=Ma;D=Na}}break}default:{da=4;break e}}break}case 82:{U=+p[aa>>3];D=k[g>>2]|0;if(+o[D+1068>>2]==9999.0){u=k[t>>2]|0;if((ba|0)==31){ym(u,c,+p[A>>3],+p[B>>3],Z,U,h,m,k[s>>2]|0,n)|0;da=8;break e}else{tm(u,c,+p[A>>3],+p[B>>3],Z,U,h,m,k[s>>2]|0,n)|0;da=8;break e}}u=k[D+1072>>2]|0;if((u+-1|0)>>>0>=2)switch(ba|0){case 31:{G=k[t>>2]|0;W=+p[A>>3];V=+p[B>>3];if(X==32768.0?(k[D+1040>>2]|0)==31:0){wm(G,c,W,V+-32768.0,Z,k[w>>2]|0,U,h,m,k[s>>2]|0,n)|0;da=8;break e}else{wm(G,c,W,V,Z,k[w>>2]|0,U,h,m,k[s>>2]|0,n)|0;da=8;break e}break}case 21:{vm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,U,h,m,k[s>>2]|0,n)|0;da=8;break e;break}case 11:{um(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,U,h,m,k[s>>2]|0,n)|0;da=8;break e;break}default:{da=8;break e}}switch(ba|0){case 31:{G=k[D+1076>>2]|0;ca=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];F=k[w>>2]|0;_=k[s>>2]|0;f=k[30368]|0;do if(!f){C=ey(1e4,4)|0;k[30368]=C;if(!C){da=8;break e}else{Oa=0;Pa=1.0}while(1){ia=Pa*16807.0;ga=ia-+(~~(ia/2147483647.0)|0)*2147483647.0;o[C+(Oa<<2)>>2]=ga/2147483647.0;Oa=Oa+1|0;if((Oa|0)==1e4){Qa=ga;break}else Pa=ga}if((~~Qa|0)==1043618065){Ra=C;break}tk(778042);da=8;break e}else Ra=f;while(0);f=(b+-2+G|0)%1e4|0;v=~~(+o[Ra+(f<<2)>>2]*500.0);H=(c|0)>0;if(!Z){if(!H){da=8;break e}M=(u|0)==2;K=0;z=f;x=v;while(1){J=k[ca+(K<<2)>>2]|0;if(M&(J|0)==-2147483646)Sa=0.0;else Sa=W+V*(+(J|0)-+o[Ra+(x<<2)>>2]+.5);p[_+(K<<3)>>3]=Sa;J=x+1|0;if((J|0)==1e4){E=z+1|0;N=(E|0)==1e4?0:E;Ta=N;Ua=~~(+o[Ra+(N<<2)>>2]*500.0)}else{Ta=z;Ua=J}K=K+1|0;if((K|0)==(c|0)){da=8;break e}else{z=Ta;x=Ua}}}if(!H){da=8;break e}x=(Z|0)==1;z=(u|0)==2;K=0;M=f;G=v;while(1){J=k[ca+(K<<2)>>2]|0;do if((J|0)==(F|0)){k[m>>2]=1;if(x){p[_+(K<<3)>>3]=U;break}else{i[h+K>>0]=1;break}}else if(z&(J|0)==-2147483646){p[_+(K<<3)>>3]=0.0;break}else{p[_+(K<<3)>>3]=W+V*(+(J|0)-+o[(k[30368]|0)+(G<<2)>>2]+.5);break}while(0);J=G+1|0;if((J|0)==1e4){C=M+1|0;N=(C|0)==1e4?0:C;Va=N;Wa=~~(+o[(k[30368]|0)+(N<<2)>>2]*500.0)}else{Va=M;Wa=J}K=K+1|0;if((K|0)==(c|0)){da=8;break}else{M=Va;G=Wa}}break}case 21:{G=k[D+1076>>2]|0;M=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];K=k[w>>2]&65535;_=k[s>>2]|0;z=k[30368]|0;do if(!z){x=ey(1e4,4)|0;k[30368]=x;if(!x){da=8;break e}else{Xa=0;Ya=1.0}while(1){ga=Ya*16807.0;ia=ga-+(~~(ga/2147483647.0)|0)*2147483647.0;o[x+(Xa<<2)>>2]=ia/2147483647.0;Xa=Xa+1|0;if((Xa|0)==1e4){Za=ia;break}else Ya=ia}if((~~Za|0)==1043618065){_a=x;break}tk(778042);da=8;break e}else _a=z;while(0);z=(b+-2+G|0)%1e4|0;F=~~(+o[_a+(z<<2)>>2]*500.0);ca=(c|0)>0;if(!Z){if(ca){$a=0;ab=z;bb=F}else{da=8;break e}while(1){p[_+($a<<3)>>3]=W+V*(+(j[M+($a<<1)>>1]|0)-+o[_a+(bb<<2)>>2]+.5);v=bb+1|0;if((v|0)==1e4){f=ab+1|0;u=(f|0)==1e4?0:f;cb=u;db=~~(+o[_a+(u<<2)>>2]*500.0)}else{cb=ab;db=v}$a=$a+1|0;if(($a|0)==(c|0)){da=8;break e}else{ab=cb;bb=db}}}if(!ca){da=8;break e}G=(Z|0)==1;v=0;u=z;f=F;while(1){H=j[M+(v<<1)>>1]|0;do if(H<<16>>16==K<<16>>16){k[m>>2]=1;if(G){p[_+(v<<3)>>3]=U;break}else{i[h+v>>0]=1;break}}else p[_+(v<<3)>>3]=W+V*(+(H<<16>>16)-+o[(k[30368]|0)+(f<<2)>>2]+.5);while(0);H=f+1|0;if((H|0)==1e4){x=u+1|0;J=(x|0)==1e4?0:x;eb=J;fb=~~(+o[(k[30368]|0)+(J<<2)>>2]*500.0)}else{eb=u;fb=H}v=v+1|0;if((v|0)==(c|0)){da=8;break}else{u=eb;f=fb}}break}case 11:{f=k[D+1076>>2]|0;u=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];v=k[w>>2]&255;_=k[s>>2]|0;G=k[30368]|0;do if(!G){K=ey(1e4,4)|0;k[30368]=K;if(!K){da=8;break e}else{gb=0;hb=1.0}while(1){ia=hb*16807.0;ga=ia-+(~~(ia/2147483647.0)|0)*2147483647.0;o[K+(gb<<2)>>2]=ga/2147483647.0;gb=gb+1|0;if((gb|0)==1e4){ib=ga;break}else hb=ga}if((~~ib|0)==1043618065){jb=K;break}tk(778042);da=8;break e}else jb=G;while(0);G=(b+-2+f|0)%1e4|0;D=~~(+o[jb+(G<<2)>>2]*500.0);M=(c|0)>0;if(!Z){if(M){kb=jb;lb=0;mb=G;nb=D}else{da=8;break e}while(1){p[_+(lb<<3)>>3]=W+V*(+(l[u+lb>>0]|0)-+o[kb+(nb<<2)>>2]+.5);F=nb+1|0;if((F|0)==1e4){z=mb+1|0;ca=(z|0)==1e4?0:z;ob=jb;pb=ca;qb=~~(+o[jb+(ca<<2)>>2]*500.0)}else{ob=kb;pb=mb;qb=F}lb=lb+1|0;if((lb|0)==(c|0)){da=8;break e}else{kb=ob;mb=pb;nb=qb}}}if(!M){da=8;break e}f=(Z|0)==1;F=0;ca=G;z=D;while(1){H=i[u+F>>0]|0;do if(H<<24>>24==v<<24>>24){k[m>>2]=1;if(f){p[_+(F<<3)>>3]=U;break}else{i[h+F>>0]=1;break}}else p[_+(F<<3)>>3]=W+V*(+(H&255)-+o[(k[30368]|0)+(z<<2)>>2]+.5);while(0);H=z+1|0;if((H|0)==1e4){K=ca+1|0;J=(K|0)==1e4?0:K;rb=J;sb=~~(+o[(k[30368]|0)+(J<<2)>>2]*500.0)}else{rb=ca;sb=H}F=F+1|0;if((F|0)==(c|0)){da=8;break}else{ca=rb;z=sb}}break}default:{da=8;break e}}break}case 11:{switch(ba|0){case 31:{hm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]|0,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}case 21:{gm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}case 11:{fm(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}default:{da=1;break e}}break}case 12:{switch(ba|0){case 31:{Tn(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]|0,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}case 21:{Sn(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}case 11:{Rn(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,i[aa>>0]|0,h,m,k[s>>2]|0,n)|0;da=1;break e;break}default:{da=1;break e}}break}case 20:{z=k[g>>2]|0;ca=(ba|0)==31;if(+o[z+1068>>2]==9999.0){F=k[t>>2]|0;if(ca){oo(F,c,+p[A>>3],+p[B>>3],Z,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}else{po(F,c,+p[A>>3],+p[B>>3],Z,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}}if(ca){ca=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];if(X==32768.0?(k[z+1040>>2]|0)==31:0){no(ca,c,V,W+-32768.0,Z,k[w>>2]|0,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}else{no(ca,c,V,W,Z,k[w>>2]|0,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e}}else switch(ba|0){case 21:{ko(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e;break}case 11:{mo(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,j[aa>>1]|0,h,m,k[s>>2]|0,n)|0;da=2;break e;break}default:{da=2;break e}}break}case 30:{ca=k[g>>2]|0;z=(ba|0)==31;if(+o[ca+1068>>2]==9999.0){F=k[t>>2]|0;if(z){eo(F,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{fo(F,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}if(z){z=k[t>>2]|0;W=+p[A>>3];V=+p[B>>3];if(X==32768.0?(k[ca+1040>>2]|0)==31:0){co(z,c,W,V+-32768.0,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{co(z,c,W,V,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}else switch(ba|0){case 21:{bo(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}case 11:{ao(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}default:{da=4;break e}}break}case 40:{z=k[g>>2]|0;ca=(ba|0)==31;if(+o[z+1068>>2]==9999.0){F=k[t>>2]|0;if(ca){zo(F,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{Ao(F,c,+p[A>>3],+p[B>>3],Z,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}if(ca){ca=k[t>>2]|0;V=+p[A>>3];W=+p[B>>3];if(X==32768.0?(k[z+1040>>2]|0)==31:0){vo(ca,c,V,W+-32768.0,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}else{vo(ca,c,V,W,Z,k[w>>2]|0,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e}}else switch(ba|0){case 21:{yo(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&65535,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}case 11:{xo(k[t>>2]|0,c,+p[A>>3],+p[B>>3],Z,k[w>>2]&255,k[aa>>2]|0,h,m,k[s>>2]|0,n)|0;da=4;break e;break}default:{da=4;break e}}break}default:{k[n>>2]=410;da=0}}while(0);dy(k[t>>2]|0);t=k[g>>2]|0;do if(k[t+1220>>2]|0){if((k[t+1104>>2]|0)==(k[t+1044>>2]|0)?(k[t+1048>>2]|0)==1:0)break;aa=ha(da,c)|0;if((aa|0)==(k[(k[t+1224>>2]|0)+(Q<<2)>>2]|0)?(k[(k[t+1228>>2]|0)+(Q<<2)>>2]|0)==(d|0):0)tb=t;else{w=k[(k[t+1232>>2]|0)+(Q<<2)>>2]|0;if(!w)ub=t;else{dy(w);ub=k[g>>2]|0}w=k[ub+1236>>2]|0;B=k[w+(Q<<2)>>2]|0;if(!B)vb=w;else{dy(B);vb=k[(k[g>>2]|0)+1236>>2]|0}k[vb+(Q<<2)>>2]=0;B=k[g>>2]|0;k[(k[B+1220>>2]|0)+(Q<<2)>>2]=0;k[(k[B+1224>>2]|0)+(Q<<2)>>2]=0;k[(k[B+1228>>2]|0)+(Q<<2)>>2]=0;w=cy(aa)|0;k[(k[B+1232>>2]|0)+(Q<<2)>>2]=w;w=k[g>>2]|0;if(!(k[(k[w+1232>>2]|0)+(Q<<2)>>2]|0)){I=k[n>>2]|0;r=q;return I|0}do if((Z|0)==2){B=cy(c)|0;k[(k[w+1236>>2]|0)+(Q<<2)>>2]=B;B=k[g>>2]|0;if(k[(k[B+1236>>2]|0)+(Q<<2)>>2]|0){wb=B;break}I=k[n>>2]|0;r=q;return I|0}else wb=w;while(0);k[(k[wb+1224>>2]|0)+(Q<<2)>>2]=aa;k[(k[wb+1228>>2]|0)+(Q<<2)>>2]=d;tb=wb}py(k[(k[tb+1232>>2]|0)+(Q<<2)>>2]|0,k[s>>2]|0,aa|0)|0;if((Z|0)==2){w=(k[g>>2]|0)+1236|0;B=k[w>>2]|0;if(!B){A=cy(c)|0;k[(k[w>>2]|0)+(Q<<2)>>2]=A;xb=k[(k[g>>2]|0)+1236>>2]|0}else xb=B;py(k[xb+(Q<<2)>>2]|0,h|0,c|0)|0}B=k[g>>2]|0;k[(k[B+1220>>2]|0)+(Q<<2)>>2]=b;k[(k[B+1240>>2]|0)+(Q<<2)>>2]=k[m>>2]}while(0);I=k[n>>2]|0;r=q;return I|0}function it(a,b,c,d,e,f,g,h,i,j,l){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;var m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0;j=r;r=r+144|0;f=j+120|0;m=j+96|0;n=j+72|0;o=j+48|0;p=j+24|0;q=j;s=k[l>>2]|0;if((s|0)>0){t=s;r=j;return t|0};k[n>>2]=0;k[n+4>>2]=0;k[n+8>>2]=0;k[n+12>>2]=0;k[n+16>>2]=0;k[n+20>>2]=0;k[o>>2]=0;k[o+4>>2]=0;k[o+8>>2]=0;k[o+12>>2]=0;k[o+16>>2]=0;k[o+20>>2]=0;k[p>>2]=0;k[p+4>>2]=0;k[p+8>>2]=0;k[p+12>>2]=0;k[p+16>>2]=0;k[p+20>>2]=0;k[q>>2]=1;k[f>>2]=1;k[m>>2]=1;k[q+4>>2]=1;u=f+4|0;k[u>>2]=1;v=m+4|0;k[v>>2]=1;k[q+8>>2]=1;w=f+8|0;k[w>>2]=1;x=m+8|0;k[x>>2]=1;k[q+12>>2]=1;y=f+12|0;k[y>>2]=1;z=m+12|0;k[z>>2]=1;k[q+16>>2]=1;k[f+16>>2]=1;k[m+16>>2]=1;k[q+20>>2]=1;k[f+20>>2]=1;k[m+20>>2]=1;a:do if((c|0)>0){A=0;b:while(1){B=k[e+(A<<2)>>2]|0;C=k[h+(A<<2)>>2]|0;if((B|0)<(C|0)){t=s;D=38;break}E=k[d+(A<<2)>>2]|0;F=k[i+(A<<2)>>2]|0;if((E|0)>(F|0)){t=s;D=38;break}G=k[q+(A<<2)>>2]|0;H=(G|0)>-1?G:0-G|0;G=(F-C|0)/(H|0)|0;F=G+1|0;I=f+(A<<2)|0;k[I>>2]=F;if((G|0)<0){D=6;break}J=B-E|0;K=J+1|0;L=m+(A<<2)|0;k[L>>2]=K;if((J|0)<0){D=8;break}J=(A|0)>0;if(J){M=ha(k[m+(A+-1<<2)>>2]|0,K)|0;k[L>>2]=M}M=B+-1|0;B=E+-1|0;while(1){K=B;B=B+1|0;N=B-C|0;if(!((N|0)%(H|0)|0)){O=N;P=K;break}if((K|0)>=(M|0)){t=s;D=38;break b}}B=M;while(1){K=B+1-C|0;if(!((K|0)%(H|0)|0)){Q=K;break}B=B+-1|0;if((P|0)>(B|0)){t=s;D=38;break b}}B=(O|0)/(H|0)|0;k[n+(A<<2)>>2]=(B|0)>0?B:0;B=(Q|0)/(H|0)|0;k[o+(A<<2)>>2]=(B|0)<(G|0)?B:G;B=C-E|0;M=(B|0)>0?B:0;B=p+(A<<2)|0;k[B>>2]=M;K=M;while(1){if(!((K+E-C|0)%(H|0)|0))break;K=K+1|0;k[B>>2]=K;if((K|0)>=(k[L>>2]|0)){t=s;D=38;break b}}if(J){L=ha(F,k[f+(A+-1<<2)>>2]|0)|0;k[I>>2]=L}A=A+1|0;if((A|0)>=(c|0))break a}if((D|0)==6){k[l>>2]=323;t=323;r=j;return t|0}else if((D|0)==8){k[l>>2]=323;t=323;r=j;return t|0}else if((D|0)==38){r=j;return t|0}}while(0);D=(k[o>>2]|0)+1-(k[n>>2]|0)|0;c=ha(D,b)|0;Q=k[o+16>>2]|0;O=k[n+16>>2]|0;if((Q|0)<(O|0)){t=s;r=j;return t|0}s=k[p+16>>2]|0;P=k[y>>2]|0;y=k[z>>2]|0;z=k[o+12>>2]|0;q=k[n+12>>2]|0;i=(z|0)<(q|0);d=Q-O|0;Q=k[w>>2]|0;w=k[p+12>>2]|0;h=k[x>>2]|0;x=k[o+8>>2]|0;e=k[n+8>>2]|0;A=(x|0)<(e|0);L=z-q|0;z=k[u>>2]|0;u=k[p+8>>2]|0;K=k[v>>2]|0;v=k[o+4>>2]|0;B=k[n+4>>2]|0;H=(v|0)<(B|0);C=x-e|0;x=k[p>>2]|0;E=k[p+4>>2]|0;p=k[m>>2]|0;m=k[f>>2]|0;f=k[n>>2]|0;n=k[o>>2]|0;o=(f|0)>(n|0);G=v-B|0;v=0;M=0;while(1){N=ha(O+v|0,P)|0;R=ha(s+M|0,y)|0;if(!i){S=0;T=0;while(1){U=(ha(q+S|0,Q)|0)+N|0;if(!A){V=(ha(w+T|0,h)|0)+R|0;W=0;X=0;while(1){Y=U+(ha(e+W|0,z)|0)|0;if(!H){Z=V+(ha(u+X|0,K)|0)+x|0;_=0;$=0;while(1){if(!o){aa=Y+(ha(B+_|0,m)|0)+f|0;ba=f;ca=Z+(ha(E+$|0,p)|0)|0;while(1){py(a+(ha(ca,b)|0)|0,g+(ha(aa,b)|0)|0,c|0)|0;ba=ba+D|0;if((ba|0)>(n|0))break;else{aa=D+aa|0;ca=D+ca|0}}}if((_|0)<(G|0)){_=_+1|0;$=$+1|0}else break}}if((W|0)<(C|0)){W=W+1|0;X=X+1|0}else break}}if((S|0)<(L|0)){S=S+1|0;T=T+1|0}else break}}if((v|0)<(d|0)){v=v+1|0;M=M+1|0}else break}t=k[l>>2]|0;r=j;return t|0}function jt(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;var l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;l=r;r=r+144|0;m=l+128|0;n=l+104|0;o=l+96|0;p=l;q=l+72|0;s=l+48|0;t=k[j>>2]|0;if((t|0)>0){u=t;r=l;return u|0}t=zk(b)|0;k[q>>2]=0;k[q+4>>2]=0;k[q+8>>2]=0;k[q+12>>2]=0;k[q+16>>2]=0;k[q+20>>2]=0;k[s>>2]=0;k[s+4>>2]=0;k[s+8>>2]=0;k[s+12>>2]=0;k[s+16>>2]=0;k[s+20>>2]=0;k[n>>2]=1;v=n+4|0;k[v>>2]=1;w=n+8|0;k[w>>2]=1;x=n+12|0;k[x>>2]=1;y=n+16|0;k[y>>2]=1;k[n+20>>2]=1;Fl(a,m,j)|0;Gl(a,6,n,j)|0;z=p;k[z>>2]=1;k[z+4>>2]=0;z=k[n>>2]|0;A=((z|0)<0)<<31>>31;B=p+8|0;k[B>>2]=z;k[B+4>>2]=A;B=k[v>>2]|0;v=yy(B|0,((B|0)<0)<<31>>31|0,z|0,A|0)|0;A=L;C=p+16|0;k[C>>2]=v;k[C+4>>2]=A;C=k[w>>2]|0;w=yy(C|0,((C|0)<0)<<31>>31|0,v|0,A|0)|0;A=L;v=p+24|0;k[v>>2]=w;k[v+4>>2]=A;v=k[x>>2]|0;x=yy(v|0,((v|0)<0)<<31>>31|0,w|0,A|0)|0;A=L;w=p+32|0;k[w>>2]=x;k[w+4>>2]=A;w=k[y>>2]|0;y=yy(w|0,((w|0)<0)<<31>>31|0,x|0,A|0)|0;A=p+40|0;k[A>>2]=y;k[A+4>>2]=L;A=ky(c|0,d|0,-1,-1)|0;d=L;c=k[m>>2]|0;if((c|0)>0){m=ky(e|0,f|0,-1,-1)|0;f=ky(m|0,L|0,A|0,d|0)|0;m=A;A=d;d=f;f=L;e=c;while(1){y=e;e=e+-1|0;x=p+(e<<3)|0;w=k[x>>2]|0;v=k[x+4>>2]|0;x=wy(m|0,A|0,w|0,v|0)|0;k[q+(e<<2)>>2]=x;C=wy(d|0,f|0,w|0,v|0)|0;k[s+(e<<2)>>2]=C;D=yy(x|0,((x|0)<0)<<31>>31|0,w|0,v|0)|0;m=jy(m|0,A|0,D|0,L|0)|0;A=L;D=yy(C|0,((C|0)<0)<<31>>31|0,w|0,v|0)|0;d=jy(d|0,f|0,D|0,L|0)|0;if((y|0)<=1)break;else f=L}}switch(c|0){case 1:{k[q>>2]=(k[q>>2]|0)+1;k[s>>2]=(k[s>>2]|0)+1;gt(a,b,q,s,g,h,i,j)|0;u=k[j>>2]|0;r=l;return u|0}case 2:{kt(a,b,t,0,q,s,n,g,h,i,o,j)|0;break}case 3:{if((((k[q>>2]|0)==0?(c=q+4|0,(k[c>>2]|0)==0):0)?(f=k[s>>2]|0,(f|0)==(z+-1|0)):0)?(d=s+4|0,A=k[d>>2]|0,(A|0)==(B+-1|0)):0){k[q>>2]=1;k[s>>2]=f+1;k[c>>2]=1;k[d>>2]=A+1;A=q+8|0;k[A>>2]=(k[A>>2]|0)+1;A=s+8|0;k[A>>2]=(k[A>>2]|0)+1;A=q+12|0;k[A>>2]=(k[A>>2]|0)+1;A=s+12|0;k[A>>2]=(k[A>>2]|0)+1;A=q+16|0;k[A>>2]=(k[A>>2]|0)+1;A=s+16|0;k[A>>2]=(k[A>>2]|0)+1;A=q+20|0;k[A>>2]=(k[A>>2]|0)+1;A=s+20|0;k[A>>2]=(k[A>>2]|0)+1;gt(a,b,q,s,g,h,i,j)|0;u=k[j>>2]|0;r=l;return u|0}A=k[s>>2]|0;d=s+4|0;c=k[d>>2]|0;f=k[q+8>>2]|0;m=k[s+8>>2]|0;if((f|0)<(m|0)){k[s>>2]=z+-1;k[d>>2]=B+-1}if((f|0)<=(m|0)){B=q+4|0;z=h;h=f;while(1){if((h|0)==(m|0)){k[s>>2]=A;k[d>>2]=c}kt(a,b,t,h,q,s,n,g,z,i,o,j)|0;k[q>>2]=0;k[B>>2]=0;z=z+(ha(k[o>>2]|0,t)|0)|0;if((h|0)>=(m|0))break;else h=h+1|0}}break}default:{tk(781697);k[j>>2]=413;u=413;r=l;return u|0}}u=k[j>>2]|0;r=l;return u|0}function kt(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;n=r;r=r+48|0;o=n+24|0;p=n;k[l>>2]=0;q=d+1|0;k[o+8>>2]=q;k[p+8>>2]=q;q=k[e>>2]|0;do if(q){d=q+1|0;k[o>>2]=d;s=e+4|0;t=k[s>>2]|0;u=t+1|0;v=o+4|0;k[v>>2]=u;k[p+4>>2]=u;u=f+4|0;if((k[u>>2]|0)==(t|0))w=(k[f>>2]|0)+1|0;else w=k[g>>2]|0;k[p>>2]=w;gt(a,b,o,p,h,i,j,m)|0;k[l>>2]=(k[l>>2]|0)+1+w-d;t=k[s>>2]|0;if((k[u>>2]|0)==(t|0)){x=k[m>>2]|0;r=n;return x|0}else{k[e>>2]=0;y=t+1|0;k[s>>2]=y;z=p;A=u;B=v;C=o;D=y;E=i+(ha(w+1-d|0,c)|0)|0;break}}else{z=p;A=f+4|0;B=o+4|0;C=o;D=k[e+4>>2]|0;E=i}while(0);k[C>>2]=1;i=D+1|0;k[B>>2]=i;D=k[g>>2]|0;k[z>>2]=D;e=k[A>>2]|0;o=e+(((k[f>>2]|0)+1|0)==(D|0)&1)|0;D=p+4|0;k[D>>2]=o;do if((o|0)>=(i|0)){gt(a,b,C,z,h,E,j,m)|0;p=o+1-i|0;w=(ha(p,k[g>>2]|0)|0)+(k[l>>2]|0)|0;k[l>>2]=w;w=k[A>>2]|0;if((w+1|0)==(o|0)){x=k[m>>2]|0;r=n;return x|0}else{F=w;G=E+(ha(ha(k[g>>2]|0,c)|0,p)|0)|0;break}}else{F=e;G=E}while(0);E=F+1|0;if((o|0)==(E|0)){x=k[m>>2]|0;r=n;return x|0}o=k[f>>2]|0;k[z>>2]=o+1;k[D>>2]=E;k[B>>2]=E;gt(a,b,C,z,h,G,j,m)|0;k[l>>2]=(k[l>>2]|0)+1+o;x=k[m>>2]|0;r=n;return x|0}function lt(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+240|0;e=d+48|0;f=d+44|0;g=d+40|0;h=d+36|0;i=d+32|0;j=d+152|0;l=d+28|0;m=d+24|0;n=d;o=k[c>>2]|0;if((o|0)>0){p=o;r=d;return p|0}if((o|0)==-1){k[c>>2]=0;q=1}else q=0;if(!(Cl(a,c)|0)){tk(781889);k[c>>2]=414;p=414;r=d;return p|0}bl(a,g)|0;bl(b,f)|0;Co(b,h,0,c)|0;k[i>>2]=0;o=(Go(a,778836,j,i)|0)==0;if(o&(k[f>>2]|0)==1)if(k[h>>2]|0){Dl(b,6,m,l,n,c)|0;if(!(k[l>>2]|0)){l=(k[g>>2]|0)==2&1;g=k[h>>2]|0;if((g|0)>0){m=g;while(1){xq(b,m,c)|0;if((m|0)>1)m=m+-1|0;else{s=l;t=0;break}}}else{s=l;t=0}}else u=11}else{s=0;t=0}else u=11;do if((u|0)==11){k[i>>2]=0;if(Go(a,781942,j,i)|0){l=k[a+4>>2]|0;if((Xr(b,k[l+1096>>2]|0,k[l+1100>>2]|0,l+1104|0,c)|0)<=0){s=0;t=1;break}tk(781951);p=k[c>>2]|0;r=d;return p|0}if(q){l=k[a+4>>2]|0;if((Xr(b,k[l+1096>>2]|0,k[l+1100>>2]|0,l+1104|0,c)|0)<=0){s=0;t=1;break}tk(781951);p=k[c>>2]|0;r=d;return p|0}if(k[h>>2]|0){Al(b,c)|0;s=0;t=0;break}Xr(b,8,0,n,c)|0;if((Al(b,c)|0)>0){tk(781951);p=k[c>>2]|0;r=d;return p|0}else{s=0;t=0}}while(0);if((k[c>>2]|0)>0){tk(781951);p=k[c>>2]|0;r=d;return p|0}if((nt(a,b,t,c)|0)>0)tk(781996);if(s){Fk(a,1,0,c)|0;Vk(b,c)|0;s=e;t=121500;n=s+104|0;do{k[s>>2]=k[t>>2];s=s+4|0;t=t+4|0}while((s|0)<(n|0));if(!((k[c>>2]|0)<=0?(Kk(a,b,1,e,13,0,0,0,c)|0,(k[c>>2]|0)<=0):0))tk(782048);Fk(a,2,0,c)|0}p=k[c>>2]|0;r=d;return p|0}function mt(a,b,c,d,e,f,g,h,i,n,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;n=n|0;q=q|0;var s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0.0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0;s=r;r=r+256|0;t=s+224|0;u=s+200|0;v=s+176|0;w=s+152|0;x=s+128|0;y=s+104|0;z=s+80|0;A=s+56|0;B=s+32|0;C=s+8|0;D=s;E=k[q>>2]|0;if((E|0)>0){F=E;r=s;return F|0}if(!(Cl(a,q)|0)){tk(781747);k[q>>2]=414;F=414;r=s;return F|0}a:do switch(b|0){case 21:{E=k[a+4>>2]|0;G=cy(k[E+1128>>2]<<1)|0;if(!g){H=E;I=G;J=2;K=0.0}else{H=E;I=G;J=2;K=+(j[g>>1]|0)}break}case 31:{G=k[a+4>>2]|0;E=cy(k[G+1128>>2]<<2)|0;if(!g){H=G;I=E;J=4;K=0.0}else{H=G;I=E;J=4;K=+(k[g>>2]|0)}break}case 41:{E=k[a+4>>2]|0;G=cy(k[E+1128>>2]<<2)|0;if(!g){H=E;I=G;J=4;K=0.0}else{H=E;I=G;J=4;K=+(k[g>>2]|0)}break}case 42:{G=k[a+4>>2]|0;E=cy(k[G+1128>>2]<<2)|0;if(!g){H=G;I=E;J=4;K=0.0}else{H=G;I=E;J=4;K=+o[g>>2]}break}case 82:{E=k[a+4>>2]|0;G=cy(k[E+1128>>2]<<3)|0;if(!g){H=E;I=G;J=8;K=0.0}else{H=E;I=G;J=8;K=+p[g>>3]}break}case 20:{G=k[a+4>>2]|0;E=cy(k[G+1128>>2]<<1)|0;if(!g){H=G;I=E;J=2;K=0.0}else{H=G;I=E;J=2;K=+(m[g>>1]|0)}break}case 30:{E=k[a+4>>2]|0;G=cy(k[E+1128>>2]<<2)|0;if(!g){H=E;I=G;J=4;K=0.0}else{H=E;I=G;J=4;K=+((k[g>>2]|0)>>>0)}break}case 40:{G=k[a+4>>2]|0;E=cy(k[G+1128>>2]<<2)|0;if(!g){H=G;I=E;J=4;K=0.0}else{H=G;I=E;J=4;K=+((k[g>>2]|0)>>>0)}break}default:if((b+-11|0)>>>0<2){E=k[a+4>>2]|0;G=cy(k[E+1128>>2]|0)|0;if(!g){H=E;I=G;J=1;K=0.0;break a}H=E;I=G;J=1;K=+(l[g>>0]|0);break a}else{tk(781805);k[q>>2]=410;F=410;r=s;return F|0}}while(0);G=(f|0)==1&K==0.0?0:f;if(!I){tk(781850);k[q>>2]=113;F=113;r=s;return F|0}f=(G|0)==2;if(f){E=ey(k[H+1128>>2]|0,1)|0;if(!E){tk(781850);dy(I);k[q>>2]=113;F=113;r=s;return F|0}else L=E}else L=0;k[t>>2]=1;k[u>>2]=1;k[v>>2]=1;k[w>>2]=1;k[z>>2]=1;E=t+4|0;k[E>>2]=1;M=u+4|0;k[M>>2]=1;N=v+4|0;k[N>>2]=1;O=w+4|0;k[O>>2]=1;P=z+4|0;k[P>>2]=1;Q=t+8|0;k[Q>>2]=1;R=u+8|0;k[R>>2]=1;S=v+8|0;k[S>>2]=1;T=w+8|0;k[T>>2]=1;U=z+8|0;k[U>>2]=1;V=t+12|0;k[V>>2]=1;W=u+12|0;k[W>>2]=1;X=v+12|0;k[X>>2]=1;Y=w+12|0;k[Y>>2]=1;Z=z+12|0;k[Z>>2]=1;_=t+16|0;k[_>>2]=1;$=u+16|0;k[$>>2]=1;aa=v+16|0;k[aa>>2]=1;ba=w+16|0;k[ba>>2]=1;ca=z+16|0;k[ca>>2]=1;da=t+20|0;k[da>>2]=1;ea=u+20|0;k[ea>>2]=1;fa=v+20|0;k[fa>>2]=1;ga=w+20|0;k[ga>>2]=1;ia=z+20|0;k[ia>>2]=1;ja=k[H+1100>>2]|0;b:do if((ja|0)>0){ka=0;la=1;while(1){ma=c+(ka<<3)|0;na=k[ma>>2]|0;oa=k[ma+4>>2]|0;ma=d+(ka<<3)|0;pa=k[ma>>2]|0;qa=k[ma+4>>2]|0;ma=k[e+(ka<<2)>>2]|0;if((oa|0)>(qa|0)|(oa|0)==(qa|0)&na>>>0>pa>>>0){ra=0-ma|0;sa=pa;ta=na}else{ra=ma;sa=na;ta=pa}k[A+(ka<<2)>>2]=sa;k[B+(ka<<2)>>2]=ta;k[C+(ka<<2)>>2]=ra;pa=k[H+1104+(ka<<2)>>2]|0;k[t+(ka<<2)>>2]=pa;if((sa|0)<1)break;na=k[H+1044+(ka<<2)>>2]|0;k[u+(ka<<2)>>2]=na;ma=((pa+-1|0)/(na|0)|0)+1|0;k[v+(ka<<2)>>2]=((sa+-1|0)/(na|0)|0)+1;pa=((ta+-1|0)/(na|0)|0)+1|0;k[w+(ka<<2)>>2]=(pa|0)<(ma|0)?pa:ma;k[z+(ka<<2)>>2]=la;la=ha(ma,la)|0;ka=ka+1|0;if((ka|0)>=(ja|0))break b}if(f)dy(L);dy(I);k[q>>2]=321;F=321;r=s;return F|0}while(0);z=(n|0)!=0;if(z)k[n>>2]=0;ta=k[fa>>2]|0;fa=k[ga>>2]|0;if((ta|0)<=(fa|0)){ga=k[ea>>2]|0;ea=x+20|0;sa=k[da>>2]|0;da=y+20|0;H=k[ia>>2]|0;ia=k[aa>>2]|0;aa=k[ba>>2]|0;ba=(ia|0)>(aa|0);ra=k[$>>2]|0;$=x+16|0;e=k[_>>2]|0;_=y+16|0;d=k[ca>>2]|0;ca=k[X>>2]|0;X=k[Y>>2]|0;Y=(ca|0)>(X|0);c=k[W>>2]|0;W=x+12|0;ka=k[V>>2]|0;V=y+12|0;la=k[Z>>2]|0;Z=k[S>>2]|0;S=k[T>>2]|0;T=(Z|0)>(S|0);ma=k[R>>2]|0;R=x+8|0;pa=k[Q>>2]|0;Q=y+8|0;na=k[U>>2]|0;U=k[N>>2]|0;N=k[O>>2]|0;O=(U|0)>(N|0);qa=k[M>>2]|0;M=x+4|0;oa=k[E>>2]|0;E=y+4|0;ua=k[P>>2]|0;P=k[v>>2]|0;v=k[w>>2]|0;w=(P|0)>(v|0);va=k[u>>2]|0;u=k[t>>2]|0;t=ta;while(1){ta=t+-1|0;wa=ha(ga,ta)|0;k[ea>>2]=wa+1;xa=ha(ga,t)|0;ya=(xa|0)<(sa|0)?xa:sa;k[da>>2]=ya;xa=ya-wa|0;wa=ha(H,ta)|0;if(!ba){ta=ia;while(1){ya=ta+-1|0;za=ha(ra,ya)|0;k[$>>2]=za+1;Aa=ha(ra,ta)|0;Ba=(Aa|0)<(e|0)?Aa:e;k[_>>2]=Ba;Aa=ha(Ba-za|0,xa)|0;za=(ha(d,ya)|0)+wa|0;if(!Y){ya=ca;while(1){Ba=ya+-1|0;Ca=ha(c,Ba)|0;k[W>>2]=Ca+1;Da=ha(c,ya)|0;Ea=(Da|0)<(ka|0)?Da:ka;k[V>>2]=Ea;Da=ha(Aa,Ea-Ca|0)|0;Ca=za+(ha(la,Ba)|0)|0;if(!T){Ba=Z;while(1){Ea=Ba+-1|0;Fa=ha(ma,Ea)|0;k[R>>2]=Fa+1;Ga=ha(ma,Ba)|0;Ha=(Ga|0)<(pa|0)?Ga:pa;k[Q>>2]=Ha;Ga=ha(Da,Ha-Fa|0)|0;Fa=Ca+(ha(na,Ea)|0)|0;if(!O){Ea=U;while(1){Ha=Ea+-1|0;Ia=ha(qa,Ha)|0;k[M>>2]=Ia+1;Ja=ha(qa,Ea)|0;Ka=(Ja|0)<(oa|0)?Ja:oa;k[E>>2]=Ka;Ja=ha(Ga,Ka-Ia|0)|0;Ia=Fa+(ha(ua,Ha)|0)|0;if(!w){Ha=P;while(1){Ka=ha(va,Ha+-1|0)|0;k[x>>2]=Ka+1;La=ha(va,Ha)|0;Ma=(La|0)<(u|0)?La:u;k[y>>2]=Ma;if(ot(ja,x,y,A,B,C,q)|0){ht(a,Ia+Ha|0,ha(Ja,Ma-Ka|0)|0,b,G,g,I,L,D,q)|0;if(z&(k[D>>2]|0)!=0)k[n>>2]=1;pt(I,J,ja,x,y,L,h,A,B,C,G,i,q)|0}if((Ha|0)<(v|0))Ha=Ha+1|0;else break}}if((Ea|0)<(N|0))Ea=Ea+1|0;else break}}if((Ba|0)<(S|0))Ba=Ba+1|0;else break}}if((ya|0)<(X|0))ya=ya+1|0;else break}}if((ta|0)<(aa|0))ta=ta+1|0;else break}}if((t|0)<(fa|0))t=t+1|0;else break}}if(f)dy(L);dy(I);F=k[q>>2]|0;r=s;return F|0}function nt(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,l=0,m=0,n=0,o=0,p=0;e=r;r=r+432|0;f=e+338|0;g=e+16|0;h=e+336|0;i=e+8|0;l=e+4|0;m=e;j[h>>1]=45;k[i>>2]=0;n=k[d>>2]|0;if((n|0)>0){o=n;r=e;return o|0}n=(c|0)==0;c=0;do{k[g+(c<<3)>>2]=k[121604+(c<<3)>>2];if(n)k[g+(c<<3)+4>>2]=k[121604+(c<<3)+4>>2];else k[g+(c<<3)+4>>2]=h;c=c+1|0}while((c|0)!=11);p=0;do{c=p+11|0;k[g+(c<<3)>>2]=k[121692+(p<<3)>>2];k[g+(c<<3)+4>>2]=k[121692+(p<<3)+4>>2];p=p+1|0}while((p|0)!=22);Go(a,790194,f,i)|0;if((k[i>>2]|0)==0?(Ax(f,779627,28)|0)==0:0)k[g+252>>2]=h;Kk(a,b,1,g,33,0,0,0,d)|0;Co(a,l,m,d)|0;l=k[m>>2]|0;k[m>>2]=(l|0)/36|0;if((l|0)>35){l=0;do{a=0;do{ds(b,779669,d)|0;a=a+1|0}while((a|0)!=36);l=l+1|0}while((l|0)<(k[m>>2]|0))}o=k[d>>2]|0;r=e;return o|0}function ot(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;h=r;r=r+48|0;i=h+24|0;j=h;l=k[g>>2]|0;if((l|0)>0){m=l;r=h;return m|0}if((a|0)>0)n=0;else{m=1;r=h;return m|0}a:while(1){l=k[c+(n<<2)>>2]|0;o=k[d+(n<<2)>>2]|0;if((l|0)<(o|0)){m=0;p=23;break}q=k[b+(n<<2)>>2]|0;s=k[e+(n<<2)>>2]|0;if((q|0)>(s|0)){m=0;p=23;break}t=k[f+(n<<2)>>2]|0;u=(t|0)>-1?t:0-t|0;t=(s-o|0)/(u|0)|0;s=t+1|0;v=i+(n<<2)|0;k[v>>2]=s;if((t|0)<0){p=6;break}t=l-q|0;w=t+1|0;x=j+(n<<2)|0;k[x>>2]=w;if((t|0)<0){p=8;break}t=(n|0)>0;if(t){y=ha(k[j+(n+-1<<2)>>2]|0,w)|0;k[x>>2]=y}y=l+-1|0;l=q+-1|0;while(1){w=l;l=l+1|0;if(!((l-o|0)%(u|0)|0)){z=w;break}if((w|0)>=(y|0)){m=0;p=23;break a}}l=y;while(1){if(!((l+1-o|0)%(u|0)|0))break;l=l+-1|0;if((z|0)>(l|0)){m=0;p=23;break a}}l=o-q|0;y=(l|0)>0?l:0;while(1){if(!((y+q-o|0)%(u|0)|0))break;y=y+1|0;if((y|0)>=(k[x>>2]|0)){m=0;p=23;break a}}if(t){x=ha(s,k[i+(n+-1<<2)>>2]|0)|0;k[v>>2]=x}n=n+1|0;if((n|0)>=(a|0)){m=1;p=23;break}}if((p|0)==6){k[g>>2]=323;m=0;r=h;return m|0}else if((p|0)==8){k[g>>2]=323;m=0;r=h;return m|0}else if((p|0)==23){r=h;return m|0}return 0}function pt(a,b,c,d,e,f,g,h,i,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;var o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0;o=r;r=r+144|0;p=o+120|0;q=o+96|0;s=o+72|0;t=o+48|0;u=o+24|0;v=o;w=k[n>>2]|0;if((w|0)>0){x=w;r=o;return x|0};k[s>>2]=0;k[s+4>>2]=0;k[s+8>>2]=0;k[s+12>>2]=0;k[s+16>>2]=0;k[s+20>>2]=0;k[t>>2]=0;k[t+4>>2]=0;k[t+8>>2]=0;k[t+12>>2]=0;k[t+16>>2]=0;k[t+20>>2]=0;k[u>>2]=0;k[u+4>>2]=0;k[u+8>>2]=0;k[u+12>>2]=0;k[u+16>>2]=0;k[u+20>>2]=0;k[v>>2]=1;k[p>>2]=1;k[q>>2]=1;y=v+4|0;k[y>>2]=1;z=p+4|0;k[z>>2]=1;A=q+4|0;k[A>>2]=1;B=v+8|0;k[B>>2]=1;C=p+8|0;k[C>>2]=1;D=q+8|0;k[D>>2]=1;E=v+12|0;k[E>>2]=1;F=p+12|0;k[F>>2]=1;G=q+12|0;k[G>>2]=1;H=v+16|0;k[H>>2]=1;I=p+16|0;k[I>>2]=1;k[q+16>>2]=1;k[v+20>>2]=1;k[p+20>>2]=1;k[q+20>>2]=1;do if((c|0)>0){J=0;a:while(1){K=k[e+(J<<2)>>2]|0;L=k[h+(J<<2)>>2]|0;if((K|0)<(L|0)){x=w;M=66;break}N=k[d+(J<<2)>>2]|0;O=k[i+(J<<2)>>2]|0;if((N|0)>(O|0)){x=w;M=66;break}P=k[j+(J<<2)>>2]|0;k[v+(J<<2)>>2]=P;Q=(P|0)>-1?P:0-P|0;P=(O-L|0)/(Q|0)|0;O=P+1|0;R=p+(J<<2)|0;k[R>>2]=O;if((P|0)<0){M=6;break}S=K-N|0;T=S+1|0;U=q+(J<<2)|0;k[U>>2]=T;if((S|0)<0){M=8;break}S=(J|0)>0;if(S){V=ha(k[q+(J+-1<<2)>>2]|0,T)|0;k[U>>2]=V}V=K+-1|0;K=N+-1|0;while(1){T=K;K=K+1|0;W=K-L|0;if(!((W|0)%(Q|0)|0)){X=W;Y=T;break}if((T|0)>=(V|0)){x=w;M=66;break a}}K=V;while(1){T=K+1-L|0;if(!((T|0)%(Q|0)|0)){Z=T;break}K=K+-1|0;if((Y|0)>(K|0)){x=w;M=66;break a}}K=(X|0)/(Q|0)|0;k[s+(J<<2)>>2]=(K|0)>0?K:0;K=(Z|0)/(Q|0)|0;k[t+(J<<2)>>2]=(K|0)<(P|0)?K:P;K=L-N|0;V=(K|0)>0?K:0;K=u+(J<<2)|0;k[K>>2]=V;T=V;while(1){if(!((T+N-L|0)%(Q|0)|0))break;T=T+1|0;k[K>>2]=T;if((T|0)>=(k[U>>2]|0)){x=w;M=66;break a}}if(S){U=ha(O,k[p+(J+-1<<2)>>2]|0)|0;k[R>>2]=U}J=J+1|0;if((J|0)>=(c|0)){M=23;break}}if((M|0)==6){k[n>>2]=323;x=323;r=o;return x|0}else if((M|0)==8){k[n>>2]=323;x=323;r=o;return x|0}else if((M|0)==23){J=k[v>>2]|0;if((J|0)==1){M=24;break}else{_=J;$=1;break}}else if((M|0)==66){r=o;return x|0}}else M=24;while(0);if((M|0)==24){_=1;$=(k[t>>2]|0)+1-(k[s>>2]|0)|0}M=ha($,b)|0;v=k[t+16>>2]|0;Z=k[s+16>>2]|0;if((v|0)<(Z|0)){x=w;r=o;return x|0}w=d+16|0;X=(c|0)>4;Y=k[H>>2]|0;H=(Y|0)>0;j=k[F>>2]|0;F=k[u+16>>2]|0;i=k[G>>2]|0;G=k[t+12>>2]|0;e=k[s+12>>2]|0;J=(G|0)<(e|0);U=v-Z|0;v=d+12|0;T=(c|0)>3;K=k[E>>2]|0;E=(K|0)>0;Q=k[C>>2]|0;C=k[u+12>>2]|0;L=k[D>>2]|0;D=k[t+8>>2]|0;N=k[s+8>>2]|0;P=(D|0)<(N|0);V=G-e|0;G=d+8|0;W=(c|0)>2;aa=k[B>>2]|0;B=(aa|0)>0;ba=k[z>>2]|0;z=k[u+8>>2]|0;ca=k[A>>2]|0;A=k[t+4>>2]|0;da=k[s+4>>2]|0;ea=(A|0)<(da|0);fa=D-N|0;D=d+4|0;d=(c|0)>1;c=k[u>>2]|0;ga=k[u+4>>2]|0;u=k[q>>2]|0;q=k[y>>2]|0;y=(q|0)>0;ia=k[p>>2]|0;p=(_|0)>0;ja=k[s>>2]|0;s=k[t>>2]|0;t=(ja|0)>(s|0);ka=A-da|0;A=(l|0)==2;l=ha((_|0)>-1?_:0-_|0,$)|0;_=p?$:0-$|0;la=h+4|0;ma=(q|0)>-1?q:0-q|0;q=h+8|0;na=(aa|0)>-1?aa:0-aa|0;aa=h+12|0;oa=(K|0)>-1?K:0-K|0;K=k[I>>2]|0;I=h+16|0;h=(Y|0)>-1?Y:0-Y|0;Y=p?ja:ia+-1-ja|0;p=0;pa=0;while(1){if(X){qa=k[w>>2]|0;ra=k[I>>2]|0;sa=pa;while(1)if(!((qa+sa+F-ra|0)%(h|0)|0)){ta=sa;break}else sa=sa+1|0}else ta=pa;if(H)ua=ha(Z+p|0,j)|0;else ua=K-(ha(p+1+Z|0,j)|0)|0;sa=ha(F+ta|0,i)|0;if(!J){ra=0;qa=0;while(1){if(T){va=k[v>>2]|0;wa=k[aa>>2]|0;xa=qa;while(1)if(!((va+xa+C-wa|0)%(oa|0)|0)){ya=xa;break}else xa=xa+1|0}else ya=qa;if(E)za=ha(e+ra|0,Q)|0;else za=j-(ha(ra+1+e|0,Q)|0)|0;xa=za+ua|0;if(!P){wa=(ha(C+ya|0,L)|0)+sa|0;va=0;R=0;while(1){if(W){O=k[G>>2]|0;S=k[q>>2]|0;Aa=R;while(1)if(!((O+Aa+z-S|0)%(na|0)|0)){Ba=Aa;break}else Aa=Aa+1|0}else Ba=R;if(B)Ca=ha(N+va|0,ba)|0;else Ca=Q-(ha(va+1+N|0,ba)|0)|0;Aa=xa+Ca|0;if(!ea){S=wa+(ha(z+Ba|0,ca)|0)+c|0;O=0;Da=0;while(1){if(d){Ea=k[D>>2]|0;Fa=k[la>>2]|0;Ga=Da;while(1)if(!((Ea+Ga+ga-Fa|0)%(ma|0)|0)){Ha=Ga;break}else Ga=Ga+1|0}else Ha=Da;Ga=S+(ha(ga+Ha|0,u)|0)|0;if(y)Ia=ha(da+O|0,ia)|0;else Ia=ba-(ha(O+1+da|0,ia)|0)|0;if(!t){Fa=Aa+Ia+Y|0;Ea=ja;Ja=Ga;while(1){if(A)py(m+Fa|0,f+Ja|0,$|0)|0;py(g+(ha(Fa,b)|0)|0,a+(ha(Ja,b)|0)|0,M|0)|0;Ea=Ea+$|0;if((Ea|0)>(s|0))break;else{Fa=_+Fa|0;Ja=l+Ja|0}}}if((O|0)<(ka|0)){O=O+1|0;Da=Ha+1|0}else break}}if((va|0)<(fa|0)){va=va+1|0;R=Ba+1|0}else break}}if((ra|0)<(V|0)){ra=ra+1|0;qa=ya+1|0}else break}}if((p|0)<(U|0)){p=p+1|0;pa=ta+1|0}else break}x=k[n>>2]|0;r=o;return x|0}function qt(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0;n=r;r=r+208|0;o=n+204|0;p=n+200|0;q=n+176|0;s=n+168|0;t=n+144|0;u=n+96|0;v=n+48|0;w=n;x=k[m>>2]|0;if((x|0)>0){y=x;r=n;return y|0}x=zk(b)|0;z=v;A=z+48|0;do{k[z>>2]=0;z=z+4|0}while((z|0)<(A|0));z=w;A=z+48|0;do{k[z>>2]=0;z=z+4|0}while((z|0)<(A|0));k[q>>2]=1;k[t>>2]=1;z=q+4|0;k[z>>2]=1;k[t+4>>2]=1;A=q+8|0;k[A>>2]=1;k[t+8>>2]=1;B=q+12|0;k[B>>2]=1;k[t+12>>2]=1;C=q+16|0;k[C>>2]=1;k[t+16>>2]=1;k[q+20>>2]=1;k[t+20>>2]=1;Fl(a,o,m)|0;Gl(a,6,q,m)|0;D=u;k[D>>2]=1;k[D+4>>2]=0;D=k[q>>2]|0;E=((D|0)<0)<<31>>31;F=u+8|0;k[F>>2]=D;k[F+4>>2]=E;F=k[z>>2]|0;z=yy(F|0,((F|0)<0)<<31>>31|0,D|0,E|0)|0;E=L;G=u+16|0;k[G>>2]=z;k[G+4>>2]=E;G=k[A>>2]|0;A=yy(G|0,((G|0)<0)<<31>>31|0,z|0,E|0)|0;E=L;z=u+24|0;k[z>>2]=A;k[z+4>>2]=E;z=k[B>>2]|0;B=yy(z|0,((z|0)<0)<<31>>31|0,A|0,E|0)|0;E=L;A=u+32|0;k[A>>2]=B;k[A+4>>2]=E;A=k[C>>2]|0;C=yy(A|0,((A|0)<0)<<31>>31|0,B|0,E|0)|0;E=u+40|0;k[E>>2]=C;k[E+4>>2]=L;E=ky(c|0,d|0,-1,-1)|0;d=L;c=k[o>>2]|0;if((c|0)>0){o=ky(e|0,f|0,-1,-1)|0;f=ky(o|0,L|0,E|0,d|0)|0;o=E;E=d;d=f;f=L;e=c;while(1){C=e;e=e+-1|0;B=u+(e<<3)|0;A=k[B>>2]|0;z=k[B+4>>2]|0;B=wy(o|0,E|0,A|0,z|0)|0;G=L;H=v+(e<<3)|0;k[H>>2]=B;k[H+4>>2]=G;H=wy(d|0,f|0,A|0,z|0)|0;I=L;J=w+(e<<3)|0;k[J>>2]=H;k[J+4>>2]=I;J=yy(A|0,z|0,B|0,G|0)|0;o=jy(o|0,E|0,J|0,L|0)|0;E=L;J=yy(A|0,z|0,H|0,I|0)|0;d=jy(d|0,f|0,J|0,L|0)|0;if((C|0)<=1)break;else f=L}}switch(c|0){case 1:{c=v;f=ky(k[c>>2]|0,k[c+4>>2]|0,1,0)|0;c=v;k[c>>2]=f;k[c+4>>2]=L;c=w;f=ky(k[c>>2]|0,k[c+4>>2]|0,1,0)|0;c=w;k[c>>2]=f;k[c+4>>2]=L;mt(a,b,v,w,t,g,h,i,j,l,m)|0;y=k[m>>2]|0;r=n;return y|0}case 2:{rt(a,b,x,0,v,w,t,q,g,h,i,j,l,s,m)|0;break}case 3:{c=v;if((((k[c>>2]|0)==0&(k[c+4>>2]|0)==0?(c=v+8|0,f=c,(k[f>>2]|0)==0&(k[f+4>>2]|0)==0):0)?(f=w,d=k[f>>2]|0,E=k[f+4>>2]|0,f=D+-1|0,(d|0)==(f|0)&(E|0)==(((f|0)<0)<<31>>31|0)):0)?(f=w+8|0,o=f,e=k[o>>2]|0,u=k[o+4>>2]|0,o=F+-1|0,(e|0)==(o|0)&(u|0)==(((o|0)<0)<<31>>31|0)):0){o=v;k[o>>2]=1;k[o+4>>2]=0;o=ky(d|0,E|0,1,0)|0;E=w;k[E>>2]=o;k[E+4>>2]=L;E=c;k[E>>2]=1;k[E+4>>2]=0;E=ky(e|0,u|0,1,0)|0;u=f;k[u>>2]=E;k[u+4>>2]=L;u=v+16|0;E=u;f=ky(k[E>>2]|0,k[E+4>>2]|0,1,0)|0;E=u;k[E>>2]=f;k[E+4>>2]=L;E=w+16|0;f=E;u=ky(k[f>>2]|0,k[f+4>>2]|0,1,0)|0;f=E;k[f>>2]=u;k[f+4>>2]=L;f=v+24|0;u=f;E=ky(k[u>>2]|0,k[u+4>>2]|0,1,0)|0;u=f;k[u>>2]=E;k[u+4>>2]=L;u=w+24|0;E=u;f=ky(k[E>>2]|0,k[E+4>>2]|0,1,0)|0;E=u;k[E>>2]=f;k[E+4>>2]=L;E=v+32|0;f=E;u=ky(k[f>>2]|0,k[f+4>>2]|0,1,0)|0;f=E;k[f>>2]=u;k[f+4>>2]=L;f=w+32|0;u=f;E=ky(k[u>>2]|0,k[u+4>>2]|0,1,0)|0;u=f;k[u>>2]=E;k[u+4>>2]=L;u=v+40|0;E=u;f=ky(k[E>>2]|0,k[E+4>>2]|0,1,0)|0;E=u;k[E>>2]=f;k[E+4>>2]=L;E=w+40|0;f=E;u=ky(k[f>>2]|0,k[f+4>>2]|0,1,0)|0;f=E;k[f>>2]=u;k[f+4>>2]=L;mt(a,b,v,w,t,g,h,i,j,l,m)|0;y=k[m>>2]|0;r=n;return y|0}f=(l|0)!=0;if(f)k[l>>2]=0;u=w;E=k[u>>2]|0;e=k[u+4>>2]|0;u=w+8|0;c=u;o=k[c>>2]|0;d=k[c+4>>2]|0;c=v+16|0;C=k[c>>2]|0;J=k[c+4>>2]|0;c=w+16|0;I=c;H=k[I>>2]|0;z=k[I+4>>2]|0;if((J|0)<(z|0)|(J|0)==(z|0)&C>>>0<H>>>0){J=D+-1|0;D=w;k[D>>2]=J;k[D+4>>2]=((J|0)<0)<<31>>31;J=F+-1|0;F=u;k[F>>2]=J;k[F+4>>2]=((J|0)<0)<<31>>31}J=((C|0)<0)<<31>>31;if(!((J|0)>(z|0)|(J|0)==(z|0)&C>>>0>H>>>0)){F=(g|0)==2;D=C;I=H;H=J;J=z;z=i;i=C;C=j;while(1){if((D|0)==(I|0)&(H|0)==(J|0)){j=w;k[j>>2]=E;k[j+4>>2]=e;j=u;k[j>>2]=o;k[j+4>>2]=d}rt(a,b,x,i,v,w,t,q,g,h,z,C,p,s,m)|0;if(f&(k[p>>2]|0)!=0)k[l>>2]=1;k[v>>2]=0;k[v+4>>2]=0;k[v+8>>2]=0;k[v+12>>2]=0;j=k[s>>2]|0;A=z+(ha(j,x)|0)|0;G=i+1|0;B=((G|0)<0)<<31>>31;K=c;M=k[K>>2]|0;N=k[K+4>>2]|0;if((B|0)>(N|0)|(B|0)==(N|0)&G>>>0>M>>>0)break;else{D=G;I=M;H=B;J=N;z=A;i=G;C=F&(C|0)!=0?C+j|0:C}}}break}default:{tk(781697);k[m>>2]=414;y=414;r=n;return y|0}}y=k[m>>2]|0;r=n;return y|0}function rt(a,b,c,d,e,f,g,h,i,j,l,m,n,o,p){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;p=p|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0;q=r;r=r+112|0;s=q+48|0;t=q;u=q+96|0;v=(n|0)!=0;if(v)k[n>>2]=0;k[o>>2]=0;w=d+1|0;d=((w|0)<0)<<31>>31;x=s+16|0;k[x>>2]=w;k[x+4>>2]=d;x=t+16|0;k[x>>2]=w;k[x+4>>2]=d;d=e;x=k[d>>2]|0;w=k[d+4>>2]|0;if(!((x|0)==0&(w|0)==0)){d=ky(x|0,w|0,1,0)|0;w=L;x=s;k[x>>2]=d;k[x+4>>2]=w;x=e+8|0;y=x;z=k[y>>2]|0;A=k[y+4>>2]|0;y=ky(z|0,A|0,1,0)|0;B=L;C=s+8|0;D=C;k[D>>2]=y;k[D+4>>2]=B;D=t+8|0;k[D>>2]=y;k[D+4>>2]=B;B=f+8|0;D=B;if((k[D>>2]|0)==(z|0)?(k[D+4>>2]|0)==(A|0):0){A=f;D=ky(k[A>>2]|0,k[A+4>>2]|0,1,0)|0;E=D;F=L}else{D=k[h>>2]|0;E=D;F=((D|0)<0)<<31>>31}D=t;k[D>>2]=E;k[D+4>>2]=F;mt(a,b,s,t,g,i,j,l,m,u,p)|0;D=k[o>>2]|0;A=ky(E|0,F|0,1,0)|0;z=jy(A|0,L|0,d|0,w|0)|0;k[o>>2]=z+D;if(v&(k[u>>2]|0)!=0)k[n>>2]=1;D=B;z=x;A=k[z>>2]|0;y=k[z+4>>2]|0;if((k[D>>2]|0)==(A|0)?(k[D+4>>2]|0)==(y|0):0){G=k[p>>2]|0;r=q;return G|0}D=e;k[D>>2]=0;k[D+4>>2]=0;D=ky(A|0,y|0,1,0)|0;y=L;A=x;k[A>>2]=D;k[A+4>>2]=y;A=jy(E|0,F|0,d|0,w|0)|0;w=ky(A|0,L|0,1,0)|0;A=yy(w|0,L|0,c|0,((c|0)<0)<<31>>31|0)|0;d=l+A|0;if((i|0)==2&(m|0)!=0){H=C;I=t;J=B;K=s;M=D;N=y;O=d;P=m+w|0}else{H=C;I=t;J=B;K=s;M=D;N=y;O=d;P=m}}else{d=e+8|0;H=s+8|0;I=t;J=f+8|0;K=s;M=k[d>>2]|0;N=k[d+4>>2]|0;O=l;P=m}m=K;k[m>>2]=1;k[m+4>>2]=0;m=ky(M|0,N|0,1,0)|0;N=L;M=H;k[M>>2]=m;k[M+4>>2]=N;M=k[h>>2]|0;l=((M|0)<0)<<31>>31;d=I;k[d>>2]=M;k[d+4>>2]=l;d=f;s=ky(k[d>>2]|0,k[d+4>>2]|0,1,0)|0;d=J;e=k[d>>2]|0;y=k[d+4>>2]|0;d=ky(e|0,y|0,(s|0)==(M|0)&(L|0)==(l|0)&1|0,0)|0;l=L;M=t+8|0;t=M;k[t>>2]=d;k[t+4>>2]=l;if(!((l|0)<(N|0)|(l|0)==(N|0)&d>>>0<m>>>0)){mt(a,b,K,I,g,i,j,O,P,u,p)|0;t=k[o>>2]|0;s=ky(d|0,l|0,1,0)|0;D=jy(s|0,L|0,m|0,N|0)|0;N=L;m=k[h>>2]|0;s=yy(D|0,N|0,m|0,((m|0)<0)<<31>>31|0)|0;k[o>>2]=s+t;if(v&(k[u>>2]|0)!=0)k[n>>2]=1;t=J;J=k[t>>2]|0;s=k[t+4>>2]|0;t=ky(J|0,s|0,1,0)|0;if((t|0)==(d|0)&(L|0)==(l|0)){G=k[p>>2]|0;r=q;return G|0}t=k[h>>2]|0;h=yy(D|0,N|0,t|0,((t|0)<0)<<31>>31|0)|0;t=yy(h|0,L|0,c|0,((c|0)<0)<<31>>31|0)|0;c=O+t|0;if((i|0)==2&(P|0)!=0){Q=J;R=s;S=c;T=P+h|0}else{Q=J;R=s;S=c;T=P}}else{Q=e;R=y;S=O;T=P}P=ky(Q|0,R|0,1,0)|0;R=L;if((d|0)==(P|0)&(l|0)==(R|0)){G=k[p>>2]|0;r=q;return G|0}l=f;f=ky(k[l>>2]|0,k[l+4>>2]|0,1,0)|0;l=I;k[l>>2]=f;k[l+4>>2]=L;l=M;k[l>>2]=P;k[l+4>>2]=R;l=H;k[l>>2]=P;k[l+4>>2]=R;mt(a,b,K,I,g,i,j,S,T,u,p)|0;if(v&(k[u>>2]|0)!=0)k[n>>2]=1;n=k[o>>2]|0;u=I;I=K;K=k[I>>2]|0;v=k[I+4>>2]|0;I=ky(k[u>>2]|0,k[u+4>>2]|0,1,0)|0;u=jy(I|0,L|0,K|0,v|0)|0;k[o>>2]=u+n;G=k[p>>2]|0;r=q;return G|0}function st(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0.0;c=r;r=r+176|0;d=c+8|0;e=c;f=c+91|0;g=c+20|0;h=c+16|0;j=c+12|0;l=k[b>>2]|0;if((l|0)>0){m=l;r=c;return m|0}if((Io(a,16,779256,g,0,b)|0)>0){tk(782100);tk(782151);m=k[b>>2]|0;r=c;return m|0}l=a+4|0;i[(k[l>>2]|0)+1084>>0]=0;zx((k[l>>2]|0)+1084|0,g,11)|0;n=i[g>>0]|0;a:do switch(n<<24>>24){case 82:{if((tx(g,778697)|0)!=0?(tx(g,779204)|0)!=0:0){q=16;break a}k[(k[l>>2]|0)+1040>>2]=11;break}case 72:{if(!(tx(g,778725)|0))k[(k[l>>2]|0)+1040>>2]=41;else q=16;break}case 71:{if(!(tx(g,778704)|0)){k[(k[l>>2]|0)+1040>>2]=21;break a}if(!(tx(g,778711)|0))k[(k[l>>2]|0)+1040>>2]=22;else q=21;break}case 66:{if(!(tx(g,782184)|0))k[(k[l>>2]|0)+1040>>2]=51;else q=21;break}case 80:{q=17;break}case 78:{q=19;break}default:q=21}while(0);if((q|0)==16)switch(n<<24>>24){case 80:{q=17;break}case 78:{q=19;break}default:q=21}if((q|0)==17)if(!(tx(g,778718)|0))k[(k[l>>2]|0)+1040>>2]=31;else q=21;else if((q|0)==19)if(!(tx(g,778737)|0))k[(k[l>>2]|0)+1040>>2]=-1;else q=21;if((q|0)==21){tk(782192);tk(g);k[b>>2]=414;m=414;r=c;return m|0}k[h>>2]=0;b:do if((Io(a,16,779028,g,0,h)|0)>0)k[(k[l>>2]|0)+1072>>2]=0;else{switch(i[g>>0]|0){case 78:{if(!(tx(g,779037)|0)){o[(k[l>>2]|0)+1068>>2]=9999.0;break b}if(!(tx(g,779213)|0)){k[(k[l>>2]|0)+1072>>2]=-1;break b}break}case 83:{if(!(tx(g,779084)|0)){k[(k[l>>2]|0)+1072>>2]=1;break b}if(!(tx(g,779183)|0)){k[(k[l>>2]|0)+1072>>2]=2;break b}break}default:{}}k[(k[l>>2]|0)+1072>>2]=0}while(0);k[h>>2]=0;if((Io(a,31,779134,j,0,h)|0)>0){g=k[l>>2]|0;k[g+1076>>2]=1;s=g}else{g=k[l>>2]|0;k[g+1076>>2]=k[j>>2];s=g}if((Io(a,31,778879,s+1096|0,0,b)|0)>0){tk(782224);m=k[b>>2]|0;r=c;return m|0}if((Io(a,31,778915,(k[l>>2]|0)+1100|0,0,b)|0)>0){tk(782271);m=k[b>>2]|0;r=c;return m|0}s=k[(k[l>>2]|0)+1100>>2]|0;if((s|0)<1){tk(782317);k[b>>2]=212;m=212;r=c;return m|0}if((s|0)>6){tk(782359);k[b>>2]=212;m=212;r=c;return m|0}else{t=1;u=0;v=1}while(1){s=u;u=u+1|0;k[e>>2]=u;Xw(f,778950,e)|0;Io(a,41,f,(k[l>>2]|0)+1104+(s<<2)|0,0,b)|0;if((k[b>>2]|0)>0){q=46;break}k[d>>2]=u;Xw(f,778989,d)|0;g=k[l>>2]|0;if(!s){k[g+1044>>2]=k[g+1104>>2];w=g+1044|0}else{j=g+1044+(s<<2)|0;k[j>>2]=1;w=j}k[h>>2]=0;Io(a,41,f,w,0,h)|0;j=k[l>>2]|0;g=k[j+1044+(s<<2)>>2]|0;n=ha((((k[j+1104+(s<<2)>>2]|0)+-1|0)/(g|0)|0)+1|0,t)|0;s=ha(g,v)|0;if((u|0)>=(k[j+1100>>2]|0)){x=j;y=n;z=s;break}else{t=n;v=s}}if((q|0)==46){tk(782400);m=k[b>>2]|0;r=c;return m|0}q=x+944|0;if(!((y|0)==(k[q>>2]|0)?(((y|0)<0)<<31>>31|0)==(k[q+4>>2]|0):0)){tk(782447);k[b>>2]=414;m=414;r=c;return m|0}c:do switch(k[x+1040>>2]|0){case 11:{if((Io(a,31,779327,x+1204|0,0,b)|0)>0){tk(782511);m=k[b>>2]|0;r=c;return m|0}k[h>>2]=0;q=(Io(a,31,779397,(k[l>>2]|0)+1208|0,0,h)|0)>0;y=k[l>>2]|0;if(q)k[y+1208>>2]=4;q=y+1204|0;if((k[q>>2]|0)<16?(v=y+1208|0,t=k[v>>2]|0,(t|0)>8):0){k[h>>2]=t;k[v>>2]=k[q>>2];k[q>>2]=k[h>>2];A=y}else A=y;break}case 41:{if((Io(a,42,779327,x+1212|0,0,b)|0)<=0){k[h>>2]=0;Io(a,31,779397,(k[l>>2]|0)+1216|0,0,h)|0;A=k[l>>2]|0;break c}tk(782511);m=k[b>>2]|0;r=c;return m|0}default:A=x}while(0);k[A+1128>>2]=z;x=k[A+1040>>2]|0;y=k[A+1096>>2]|0;q=k[A+1204>>2]|0;d:do if((x|0)==11)if((y|0)==16){B=(z<<1)+6+((z|0)/(q|0)|0)|0;break}else{B=(z<<2)+6+((z|0)/(q|0)|0)|0;break}else{if((x+-21|0)>>>0<2)switch(y|0){case 8:{B=z;break d;break}case 16:{B=z<<1;break d;break}default:{B=z<<2;break d}}switch(x|0){case 51:{B=~~(+(z|0)*1.01*+(y|0)*.125+601.0);break d;break}case 41:{C=+(z|0);switch(y|0){case 8:case 16:{B=~~(C*2.2+26.0);break d;break}default:{B=~~(C*4.4+26.0);break d}}break}default:{B=z<<2;break d}}}while(0);k[A+1132>>2]=B;if((Tk(a,0,782556,A+1136|0,b)|0)>0){tk(782572);k[b>>2]=414;m=414;r=c;return m|0}vk();k[h>>2]=0;Tk(a,0,779880,(k[l>>2]|0)+1140|0,h)|0;k[h>>2]=0;Tk(a,0,780328,(k[l>>2]|0)+1144|0,h)|0;k[h>>2]=0;if((Tk(a,0,782639,(k[l>>2]|0)+1148|0,h)|0)>0?(k[h>>2]=0,(Io(a,82,782639,(k[l>>2]|0)+1160|0,0,h)|0)<1):0)k[(k[l>>2]|0)+1148>>2]=-1;k[h>>2]=0;if((Tk(a,0,782646,(k[l>>2]|0)+1152|0,h)|0)>0?(k[h>>2]=0,(Io(a,82,782646,(k[l>>2]|0)+1168|0,0,h)|0)<1):0)k[(k[l>>2]|0)+1152>>2]=-1;k[h>>2]=0;do if((Tk(a,0,779674,(k[l>>2]|0)+1156|0,h)|0)>0){k[h>>2]=0;if((Io(a,31,779674,(k[l>>2]|0)+1200|0,0,h)|0)<1){k[(k[l>>2]|0)+1156>>2]=-1;break}k[h>>2]=0;if((Io(a,31,779663,(k[l>>2]|0)+1200|0,0,h)|0)<1)k[(k[l>>2]|0)+1156>>2]=-1}while(0);k[h>>2]=0;A=(Io(a,82,779536,(k[l>>2]|0)+1176|0,0,h)|0)>0;B=k[l>>2]|0;if(A)p[B+1176>>3]=1.0;k[h>>2]=0;A=(Io(a,82,779507,B+1184|0,0,h)|0)>0;h=k[l>>2]|0;l=h+1184|0;if(A){k[l>>2]=0;k[l+4>>2]=0;k[l+8>>2]=0;k[l+12>>2]=0}else p[h+1192>>3]=+p[l>>3];C=+o[h+1012>>2];if(C!=0.0)o[h+1068>>2]=C;yk();m=k[b>>2]|0;r=c;return m|0}function tt(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0;d=ty(b|0,c|0,3)|0;e=cy(d)|0;f=d+-1|0;if(!((c|0)>0|(c|0)==0&b>>>0>0)){py(a|0,e|0,d|0)|0;dy(e);return}g=0-b|0;h=ty(b|0,c|0,1)|0;j=0-h|0;h=yy(b|0,c|0,3,0)|0;k=0-h|0;h=ty(b|0,c|0,2)|0;l=0-h|0;h=yy(b|0,c|0,5,0)|0;m=0-h|0;h=yy(b|0,c|0,6,0)|0;n=0-h|0;h=yy(b|0,c|0,7,0)|0;o=0-h|0;h=0;p=0;q=e+f|0;r=a+f|0;while(1){i[q>>0]=i[r>>0]|0;i[q+-1>>0]=i[r+g>>0]|0;i[q+-2>>0]=i[r+j>>0]|0;i[q+-3>>0]=i[r+k>>0]|0;i[q+-4>>0]=i[r+l>>0]|0;i[q+-5>>0]=i[r+m>>0]|0;i[q+-6>>0]=i[r+n>>0]|0;i[q+-7>>0]=i[r+o>>0]|0;h=ky(h|0,p|0,1,0)|0;p=L;if(!((p|0)<(c|0)|(p|0)==(c|0)&h>>>0<b>>>0))break;else{q=q+-8|0;r=r+-1|0}}py(a|0,e|0,d|0)|0;dy(e);return}function ut(a,b,c,d,e,f,g,h,i,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=+f;g=+g;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;var q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0.0,C=0,D=0.0,E=0.0,F=0.0,G=0.0,H=0.0,I=0.0,J=0,K=0.0,L=0,M=0,N=0.0,O=0.0,P=0.0,Q=0,R=0.0,S=0.0,T=0,U=0,X=0.0,Y=0.0,Z=0,_=0.0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ja=0,ka=0,la=0,ma=0.0,na=0.0,oa=0,pa=0.0,qa=0.0,ra=0,sa=0,ta=0.0,ua=0.0,va=0.0,wa=0.0,xa=0,ya=0,za=0.0,Aa=0.0,Ba=0.0,Ca=0.0,Da=0.0,Ea=0.0,Fa=0.0,Ga=0.0,Ha=0,Ia=0.0,Ja=0.0,Ka=0,La=0,Ma=0,Na=0.0,Oa=0.0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0.0,Va=0.0,Wa=0,Xa=0.0,Ya=0.0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0;q=r;r=r+48|0;s=q+36|0;t=q+32|0;u=q+16|0;v=q+8|0;w=q;x=q+28|0;y=q+24|0;k[t>>2]=0;o[x>>2]=0.0;o[y>>2]=0.0;z=ha(d,c)|0;if((z|0)<2){p[j>>3]=1.0;p[l>>3]=0.0;A=0;r=q;return A|0}B=g;do if(g>=0.0){xt(b,c,d,e,f,t,x,y,u,v,w,s);C=k[t>>2]|0;if((e|0)!=0&(C|0)==0){o[x>>2]=0.0;o[y>>2]=1.0;D=1.0}else{E=+p[v>>3];F=+p[u>>3];G=F!=0.0&F<E?F:E;E=+p[w>>3];D=E!=0.0&E<G?E:G}G=g==0.0?D*.25:D/B;if(G==0.0){A=0;r=q;return A|0}else{H=+o[y>>2];I=+o[x>>2];J=C;K=G;break}}else{G=-g;C=(c|0)<5;L=C?1:d;M=ha(C?d:1,c)|0;if((M|0)<5){a:do if((M|0)>0){if(!e){C=0;E=-3402823466385288598117041.0e14;F=3402823466385288598117041.0e14;while(1){N=+o[b+(C<<2)>>2];O=N<F?N:F;P=N>E?N:E;C=C+1|0;if((C|0)==(M|0)){Q=M;R=P;S=O;break a}else{E=P;F=O}}}else{T=0;U=0;X=-3402823466385288598117041.0e14;Y=3402823466385288598117041.0e14}while(1){F=+o[b+(T<<2)>>2];if(F==f){Z=U;_=X;$=Y}else{Z=U+1|0;_=F>X?F:X;$=F<Y?F:Y}T=T+1|0;if((T|0)==(M|0)){Q=Z;R=_;S=$;break}else{U=Z;X=_;Y=$}}}else{Q=0;R=-3402823466385288598117041.0e14;S=3402823466385288598117041.0e14}while(0);o[x>>2]=S;o[y>>2]=R;k[t>>2]=Q;H=R;I=S;J=Q;K=G;break}if((L|0)>0){C=(e|0)!=0;aa=M+-1|0;ba=0;ca=0;F=-3402823466385288598117041.0e14;E=3402823466385288598117041.0e14;while(1){da=ha(ba,M)|0;b:do if(C){ea=0;while(1){if(!(+o[b+(ea+da<<2)>>2]==f)){fa=ea;break b}ja=ea+1|0;if((ja|0)<(M|0))ea=ja;else{fa=ja;break}}}else fa=0;while(0);if((fa|0)!=(M|0)){O=+o[b+(fa+da<<2)>>2];P=O<E?O:E;N=O>F?O:F;ea=fa+1|0;c:do if(C&(ea|0)<(M|0)){ja=ea;while(1){if(!(+o[b+(ja+da<<2)>>2]==f)){ka=ja;break c}la=ja+1|0;if((la|0)<(M|0))ja=la;else{ka=la;break}}}else ka=ea;while(0);if((ka|0)!=(M|0)){O=+o[b+(ka+da<<2)>>2];ma=O<P?O:P;na=O>N?O:N;ea=ka+1|0;d:do if(C&(ea|0)<(M|0)){ja=ea;while(1){if(!(+o[b+(ja+da<<2)>>2]==f)){oa=ja;break d}la=ja+1|0;if((la|0)<(M|0))ja=la;else{oa=la;break}}}else oa=ea;while(0);if((oa|0)!=(M|0)){O=+o[b+(oa+da<<2)>>2];pa=O<ma?O:ma;qa=O>na?O:na;ea=oa+1|0;e:do if(C&(ea|0)<(M|0)){ja=ea;while(1){if(!(+o[b+(ja+da<<2)>>2]==f)){ra=ja;break e}la=ja+1|0;if((la|0)<(M|0))ja=la;else{ra=la;break}}}else ra=ea;while(0);if((ra|0)==(M|0)){sa=ca;ta=qa;ua=pa}else{O=+o[b+(ra+da<<2)>>2];va=O<pa?O:pa;wa=O>qa?O:qa;ea=ra+1|0;f:do if((ea|0)<(M|0)){if(C){xa=ea;ya=ca;za=wa;Aa=va}else{ja=aa+ca|0;la=ea;O=wa;Ba=va;while(1){Ca=+o[b+(la+da<<2)>>2];Da=Ca<Ba?Ca:Ba;Ea=Ca>O?Ca:O;la=la+1|0;if((la|0)==(M|0)){Fa=Ea;Ga=Da;break}else{O=Ea;Ba=Da}}Ha=ja-ra|0;Ia=Fa;Ja=Ga;break}while(1){g:do if((xa|0)<(M|0)){la=xa;while(1){if(!(+o[b+(la+da<<2)>>2]==f)){Ka=la;break g}La=la+1|0;if((La|0)<(M|0))la=La;else{Ka=La;break}}}else Ka=xa;while(0);if((Ka|0)==(M|0)){Ha=ya;Ia=za;Ja=Aa;break f}Ba=+o[b+(Ka+da<<2)>>2];O=Ba<Aa?Ba:Aa;Da=Ba>za?Ba:za;la=ya+1|0;xa=Ka+1|0;if((xa|0)>=(M|0)){Ha=la;Ia=Da;Ja=O;break}else{ya=la;za=Da;Aa=O}}}else{Ha=ca;Ia=wa;Ja=va}while(0);sa=Ha+4|0;ta=Ia;ua=Ja}}else{sa=ca;ta=na;ua=ma}}else{sa=ca;ta=N;ua=P}}else{sa=ca;ta=F;ua=E}ba=ba+1|0;if((ba|0)==(L|0)){Ma=sa;Na=ta;Oa=ua;break}else{ca=sa;F=ta;E=ua}}}else{Ma=0;Na=-3402823466385288598117041.0e14;Oa=3402823466385288598117041.0e14}k[t>>2]=Ma;o[x>>2]=Oa;o[y>>2]=Na;H=Na;I=Oa;J=Ma;K=G}while(0);if((H-I)/K>4294967284.0){A=0;r=q;return A|0}Ma=(a|0)>0;if(Ma){sa=k[30368]|0;do if(!sa)if(!(Ss()|0)){Pa=k[t>>2]|0;Qa=k[30368]|0;break}else{A=113;r=q;return A|0}else{Pa=J;Qa=sa}while(0);sa=(a+-1|0)%1e4|0;Ra=Pa;Sa=sa;Ta=~~(+o[Qa+(sa<<2)>>2]*500.0)}else{Ra=J;Sa=0;Ta=0}h:do if((Ra|0)==(z|0)){J=(h|0)==2;do if(!J){I=+o[y>>2];H=+o[x>>2];if((I-H)/K<2147483637.0){Oa=H/K+.5;Ua=H;Va=K*(+(~~Oa>>>0>>>0)+4294967296.0*+((+W(Oa)>=1.0?(Oa>0.0?~~+ia(+V(Oa/4294967296.0),4294967295.0)>>>0:~~+ga((Oa-+(~~Oa>>>0))/4294967296.0)>>>0):0)|0));break}else{Ua=H;Va=(I+H)*.5;break}}else{H=+o[x>>2];Ua=H;Va=K*2147483637.0+H}while(0);sa=(z|0)>0;if(!Ma){if(sa)Wa=0;else{Xa=Ua;Ya=Va;break}while(1){G=(+o[b+(Wa<<2)>>2]-Va)/K;k[i+(Wa<<2)>>2]=~~(G+(G>=0.0?.5:-.5));Wa=Wa+1|0;if((Wa|0)==(z|0)){Xa=Ua;Ya=Va;break h}}}if(sa){Qa=k[30368]|0;if(J){Za=0;_a=Sa;$a=Ta}else{Pa=0;a=Sa;t=Ta;while(1){G=(+o[b+(Pa<<2)>>2]-Va)/K+ +o[Qa+(t<<2)>>2]+-.5;k[i+(Pa<<2)>>2]=~~(G+(G>=0.0?.5:-.5));Ha=t+1|0;if((Ha|0)==1e4){ya=a+1|0;xa=(ya|0)==1e4?0:ya;ab=xa;bb=~~(+o[Qa+(xa<<2)>>2]*500.0)}else{ab=a;bb=Ha}Pa=Pa+1|0;if((Pa|0)==(z|0)){Xa=Ua;Ya=Va;break h}else{a=ab;t=bb}}}while(1){G=+o[b+(Za<<2)>>2];if(G==0.0)k[i+(Za<<2)>>2]=-2147483646;else{H=(G-Va)/K+ +o[Qa+($a<<2)>>2]+-.5;k[i+(Za<<2)>>2]=~~(H+(H>=0.0?.5:-.5))}t=$a+1|0;if((t|0)==1e4){a=_a+1|0;Pa=(a|0)==1e4?0:a;cb=Pa;db=~~(+o[Qa+(Pa<<2)>>2]*500.0)}else{cb=_a;db=t}Za=Za+1|0;if((Za|0)==(z|0)){Xa=Ua;Ya=Va;break}else{_a=cb;$a=db}}}else{Xa=Ua;Ya=Va}}else{H=+o[x>>2];G=K*2147483637.0+H;Qa=(z|0)>0;if(!Ma){if(Qa)eb=0;else{Xa=H;Ya=G;break}while(1){I=+o[b+(eb<<2)>>2];if(I!=f){Oa=(I-G)/K;fb=~~(Oa+(Oa>=0.0?.5:-.5))}else fb=-2147483647;k[i+(eb<<2)>>2]=fb;eb=eb+1|0;if((eb|0)==(z|0)){Xa=H;Ya=G;break h}}}if(Qa){t=(h|0)==2;Pa=k[30368]|0;a=0;J=Sa;sa=Ta;while(1){Oa=+o[b+(a<<2)>>2];do if(Oa!=f)if(t&Oa==0.0){k[i+(a<<2)>>2]=-2147483646;break}else{I=(Oa-G)/K+ +o[Pa+(sa<<2)>>2]+-.5;k[i+(a<<2)>>2]=~~(I+(I>=0.0?.5:-.5));break}else k[i+(a<<2)>>2]=-2147483647;while(0);Ha=sa+1|0;if((Ha|0)==1e4){xa=J+1|0;ya=(xa|0)==1e4?0:xa;gb=ya;hb=~~(+o[Pa+(ya<<2)>>2]*500.0)}else{gb=J;hb=Ha}a=a+1|0;if((a|0)==(z|0)){Xa=H;Ya=G;break}else{J=gb;sa=hb}}}else{Xa=H;Ya=G}}while(0);f=(Xa-Ya)/K;k[m>>2]=~~(f+(f>=0.0?.5:-.5));f=(+o[y>>2]-Ya)/K;k[n>>2]=~~(f+(f>=0.0?.5:-.5));p[j>>3]=K;p[l>>3]=Ya;A=1;r=q;return A|0}function vt(a,b,c,d,e,f,g,h,i,j,l,m,n){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=+f;g=+g;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;var q=0,r=0,s=0.0,t=0.0,u=0,v=0,w=0,x=0.0,y=0.0,z=0,A=0.0,B=0,C=0,D=0.0,E=0.0,F=0.0,G=0,H=0.0,I=0.0,J=0,K=0,L=0,M=0,N=0.0,O=0.0,P=0,Q=0,R=0.0,S=0.0,T=0,U=0.0,X=0.0,Y=0,Z=0,_=0.0,$=0.0,aa=0.0,ba=0.0,ca=0.0,da=0,ea=0,fa=0,ja=0,ka=0.0,la=0.0,ma=0.0,na=0.0,oa=0.0,pa=0,qa=0.0,ra=0.0,sa=0,ta=0.0,ua=0.0,va=0.0,wa=0.0,xa=0.0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0.0,La=0.0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0.0,Ya=0.0,Za=0.0,_a=0.0,$a=0.0,ab=0.0,bb=0.0,cb=0.0,db=0.0,eb=0.0,fb=0.0,gb=0.0,hb=0.0,ib=0.0,jb=0.0,kb=0.0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0.0,rb=0.0,sb=0.0,tb=0.0,ub=0.0,vb=0,wb=0.0,xb=0.0,yb=0,zb=0,Ab=0,Bb=0.0,Cb=0.0,Db=0.0,Eb=0.0,Fb=0.0,Gb=0.0,Hb=0.0,Ib=0.0,Jb=0.0,Kb=0.0,Lb=0.0,Mb=0.0,Nb=0.0,Ob=0.0,Pb=0.0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0.0,Xb=0.0,Yb=0.0,Zb=0.0,_b=0,$b=0.0,ac=0.0,bc=0.0,cc=0.0,dc=0.0,ec=0.0,fc=0.0,gc=0.0,hc=0,ic=0,jc=0,kc=0.0,lc=0,mc=0.0,nc=0,oc=0,pc=0,qc=0,rc=0,sc=0;q=ha(d,c)|0;if((q|0)<2){p[j>>3]=1.0;p[l>>3]=0.0;r=0;return r|0}s=g;a:do if(!(g>=0.0)){t=-g;u=(c|0)<5;v=u?1:d;w=ha(u?d:1,c)|0;if((w|0)<5){if((w|0)<=0){x=-1797693134862315708145274.0e284;y=1797693134862315708145274.0e284;z=0;A=t;break}u=(e|0)!=0;B=0;C=0;D=-1797693134862315708145274.0e284;E=1797693134862315708145274.0e284;while(1){F=+p[b+(B<<3)>>3];if(u&F==f){G=C;H=D;I=E}else{G=C+1|0;H=F>D?F:D;I=F<E?F:E}B=B+1|0;if((B|0)==(w|0)){x=H;y=I;z=G;A=t;break a}else{C=G;D=H;E=I}}}if((v|0)>0){C=(e|0)!=0;B=0;u=0;E=-1797693134862315708145274.0e284;D=1797693134862315708145274.0e284;while(1){J=ha(B,w)|0;b:do if(C){K=0;while(1){if(!(+p[b+(K+J<<3)>>3]==f)){L=K;break b}M=K+1|0;if((M|0)<(w|0))K=M;else{L=M;break}}}else L=0;while(0);if((L|0)!=(w|0)){F=+p[b+(L+J<<3)>>3];N=F<D?F:D;O=F>E?F:E;K=L+1|0;c:do if(C&(K|0)<(w|0)){M=K;while(1){if(!(+p[b+(M+J<<3)>>3]==f)){P=M;break c}Q=M+1|0;if((Q|0)<(w|0))M=Q;else{P=Q;break}}}else P=K;while(0);if((P|0)!=(w|0)){F=+p[b+(P+J<<3)>>3];R=F<N?F:N;S=F>O?F:O;K=P+1|0;d:do if(C&(K|0)<(w|0)){M=K;while(1){if(!(+p[b+(M+J<<3)>>3]==f)){T=M;break d}Q=M+1|0;if((Q|0)<(w|0))M=Q;else{T=Q;break}}}else T=K;while(0);if((T|0)!=(w|0)){F=+p[b+(T+J<<3)>>3];U=F<R?F:R;X=F>S?F:S;K=T+1|0;e:do if(C&(K|0)<(w|0)){M=K;while(1){if(!(+p[b+(M+J<<3)>>3]==f)){Y=M;break e}Q=M+1|0;if((Q|0)<(w|0))M=Q;else{Y=Q;break}}}else Y=K;while(0);if((Y|0)==(w|0)){Z=u;_=X;$=U}else{F=+p[b+(Y+J<<3)>>3];aa=F<U?F:U;ba=F>X?F:X;K=Y+1|0;f:do if((K|0)<(w|0)){M=K;Q=u;F=ba;ca=aa;while(1){g:do if(C&(M|0)<(w|0)){da=M;while(1){if(!(+p[b+(da+J<<3)>>3]==f)){ea=da;break g}fa=da+1|0;if((fa|0)<(w|0))da=fa;else{ea=fa;break}}}else ea=M;while(0);if((ea|0)==(w|0)){ja=Q;ka=F;la=ca;break f}ma=+p[b+(ea+J<<3)>>3];na=ma<ca?ma:ca;oa=ma>F?ma:F;da=Q+1|0;M=ea+1|0;if((M|0)>=(w|0)){ja=da;ka=oa;la=na;break}else{Q=da;F=oa;ca=na}}}else{ja=u;ka=ba;la=aa}while(0);Z=ja+4|0;_=ka;$=la}}else{Z=u;_=S;$=R}}else{Z=u;_=O;$=N}}else{Z=u;_=E;$=D}B=B+1|0;if((B|0)==(v|0)){x=_;y=$;z=Z;A=t;break}else{u=Z;E=_;D=$}}}else{x=-1797693134862315708145274.0e284;y=1797693134862315708145274.0e284;z=0;A=t}}else{u=(c|0)<9;v=u?1:d;B=ha(u?d:1,c)|0;do if((B|0)<9)if((B|0)>0){u=(e|0)!=0;w=0;C=0;D=-1797693134862315708145274.0e284;E=1797693134862315708145274.0e284;while(1){aa=+p[b+(w<<3)>>3];if(u&aa==f){pa=C;qa=D;ra=E}else{pa=C+1|0;qa=aa>D?aa:D;ra=aa<E?aa:E}w=w+1|0;if((w|0)==(B|0)){sa=pa;ta=0.0;ua=0.0;va=0.0;wa=ra;xa=qa;break}else{C=pa;D=qa;E=ra}}}else{sa=0;ta=0.0;ua=0.0;va=0.0;wa=1797693134862315708145274.0e284;xa=-1797693134862315708145274.0e284}else{C=ey(B,8)|0;if(!C){sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0}else{w=ey(B,8)|0;if(!w){dy(C);sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0;break}u=ey(B,8)|0;if(!u){dy(C);dy(w);sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0;break}J=ey(v,8)|0;if(!J){dy(C);dy(w);dy(u);sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0;break}K=ey(v,8)|0;if(!K){dy(C);dy(w);dy(u);dy(J);sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0;break}Q=ey(v,8)|0;if(!Q){dy(C);dy(w);dy(u);dy(J);dy(K);sa=0;ta=0.0;ua=0.0;va=0.0;wa=0.0;xa=0.0;break}h:do if((v|0)>0){M=(e|0)!=0;da=0;fa=0;ya=0;za=0;E=-1797693134862315708145274.0e284;D=1797693134862315708145274.0e284;while(1){Aa=ha(da,B)|0;i:do if(M){Ba=0;while(1){if(!(+p[b+(Ba+Aa<<3)>>3]==f)){Ca=Ba;break i}Da=Ba+1|0;if((Da|0)<(B|0))Ba=Da;else{Ca=Da;break}}}else Ca=0;while(0);j:do if((Ca|0)!=(B|0)){N=+p[b+(Ca+Aa<<3)>>3];Ba=fa+1|0;O=N<D?N:D;R=N>E?N:E;Da=Ca+1|0;k:do if(M&(Da|0)<(B|0)){Ea=Da;while(1){if(!(+p[b+(Ea+Aa<<3)>>3]==f)){Fa=Ea;break k}Ga=Ea+1|0;if((Ga|0)<(B|0))Ea=Ga;else{Fa=Ga;break}}}else Fa=Da;while(0);if((Fa|0)==(B|0)){Ha=Ba;Ia=ya;Ja=za;Ka=R;La=O}else{S=+p[b+(Fa+Aa<<3)>>3];Da=fa+2|0;aa=S<O?S:O;ba=S>R?S:R;Ea=Fa+1|0;l:do if(M&(Ea|0)<(B|0)){Ga=Ea;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Ma=Ga;break l}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Ma=Na;break}}}else Ma=Ea;while(0);if((Ma|0)==(B|0)){Ha=Da;Ia=ya;Ja=za;Ka=ba;La=aa;break}R=+p[b+(Ma+Aa<<3)>>3];Ea=fa+3|0;O=R<aa?R:aa;X=R>ba?R:ba;Ba=Ma+1|0;m:do if(M&(Ba|0)<(B|0)){Ga=Ba;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Oa=Ga;break m}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Oa=Na;break}}}else Oa=Ba;while(0);if((Oa|0)==(B|0)){Ha=Ea;Ia=ya;Ja=za;Ka=X;La=O;break}ba=+p[b+(Oa+Aa<<3)>>3];Ba=fa+4|0;aa=ba<O?ba:O;U=ba>X?ba:X;Da=Oa+1|0;n:do if(M&(Da|0)<(B|0)){Ga=Da;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Pa=Ga;break n}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Pa=Na;break}}}else Pa=Da;while(0);if((Pa|0)==(B|0)){Ha=Ba;Ia=ya;Ja=za;Ka=U;La=aa;break}X=+p[b+(Pa+Aa<<3)>>3];Da=fa+5|0;O=X<aa?X:aa;ca=X>U?X:U;Ea=Pa+1|0;o:do if(M&(Ea|0)<(B|0)){Ga=Ea;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Qa=Ga;break o}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Qa=Na;break}}}else Qa=Ea;while(0);if((Qa|0)==(B|0)){Ha=Da;Ia=ya;Ja=za;Ka=ca;La=O;break}U=+p[b+(Qa+Aa<<3)>>3];Ea=fa+6|0;aa=U<O?U:O;F=U>ca?U:ca;Ba=Qa+1|0;p:do if(M&(Ba|0)<(B|0)){Ga=Ba;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Ra=Ga;break p}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Ra=Na;break}}}else Ra=Ba;while(0);if((Ra|0)==(B|0)){Ha=Ea;Ia=ya;Ja=za;Ka=F;La=aa;break}ca=+p[b+(Ra+Aa<<3)>>3];Ba=fa+7|0;O=ca<aa?ca:aa;na=ca>F?ca:F;Da=Ra+1|0;q:do if(M&(Da|0)<(B|0)){Ga=Da;while(1){if(!(+p[b+(Ga+Aa<<3)>>3]==f)){Sa=Ga;break q}Na=Ga+1|0;if((Na|0)<(B|0))Ga=Na;else{Sa=Na;break}}}else Sa=Da;while(0);if((Sa|0)==(B|0)){Ha=Ba;Ia=ya;Ja=za;Ka=na;La=O;break}F=+p[b+(Sa+Aa<<3)>>3];Da=fa+8|0;aa=F<O?F:O;oa=F>na?F:na;Ea=Sa+1|0;if((Ea|0)<(B|0)){Ta=Ea;Ua=Da;Va=0;Wa=0;Xa=N;Ya=S;Za=R;_a=ba;$a=X;ab=U;bb=ca;cb=F;db=oa;eb=aa}else{Ha=Da;Ia=ya;Ja=za;Ka=oa;La=aa;break}r:while(1){Da=Ta;Ea=Va;Ga=Wa;aa=Xa;oa=Ya;F=Za;ma=_a;fb=$a;gb=ab;hb=bb;ib=cb;jb=db;kb=eb;while(1){s:do if(M&(Da|0)<(B|0)){Na=Da;while(1){if(!(+p[b+(Na+Aa<<3)>>3]==f)){lb=Na;break s}mb=Na+1|0;if((mb|0)<(B|0))Na=mb;else{lb=mb;break}}}else lb=Da;while(0);if((lb|0)==(B|0)){nb=Ua;ob=Ea;pb=Ga;qb=jb;rb=kb;break r}sb=+p[b+(lb+Aa<<3)>>3];tb=sb<kb?sb:kb;ub=sb>jb?sb:jb;Na=fb==gb;mb=gb==hb;if(mb&Na)vb=Ga;else{wb=+W(+(fb-hb));p[C+(Ga<<3)>>3]=wb;vb=Ga+1|0}if(mb&(Na&(ma==fb&F==ma))){xb=sb;yb=lb;zb=Ea;Ab=vb;Bb=oa;Cb=F;Db=ma;Eb=fb;Fb=gb;Gb=hb;Hb=ib;Ib=ub;Jb=tb;break}wb=+W(+(fb*2.0-F-hb));p[w+(Ea<<3)>>3]=wb;wb=+W(+(aa+(fb*6.0-F*4.0-hb*4.0)+sb));p[u+(Ea<<3)>>3]=wb;Na=Ea+1|0;Da=lb+1|0;if((Da|0)>=(B|0)){nb=Ua;ob=Na;pb=vb;qb=ub;rb=tb;break r}else{wb=ib;Kb=hb;Lb=gb;Mb=fb;Nb=ma;Ob=F;Pb=oa;Ea=Na;Ga=vb;ib=sb;jb=ub;kb=tb;hb=wb;gb=Kb;fb=Lb;ma=Mb;F=Nb;oa=Ob;aa=Pb}}Ga=Ua+1|0;Ea=yb+1|0;if((Ea|0)<(B|0)){Ta=Ea;Ua=Ga;Va=zb;Wa=Ab;Xa=Bb;Ya=Cb;Za=Db;_a=Eb;$a=Fb;ab=Gb;bb=Hb;cb=xb;db=Ib;eb=Jb}else{nb=Ga;ob=zb;pb=Ab;qb=Ib;rb=Jb;break}}Ba=ob+nb|0;switch(ob|0){case 0:{Ha=Ba;Ia=ya;Ja=za;Ka=qb;La=rb;break j;break}case 1:{if((pb|0)==1){p[J+(za<<3)>>3]=+p[C>>3];Qb=za+1|0}else Qb=za;p[K+(ya<<3)>>3]=+p[w>>3];p[Q+(ya<<3)>>3]=+p[u>>3];Rb=Qb;break}default:{if((pb|0)>1){ca=+yt(C,ob);p[J+(za<<3)>>3]=ca;Sb=za+1|0}else Sb=za;ca=+yt(w,ob);p[K+(ya<<3)>>3]=ca;ca=+yt(u,ob);p[Q+(ya<<3)>>3]=ca;Rb=Sb}}Ha=Ba;Ia=ya+1|0;Ja=Rb;Ka=qb;La=rb}}else{Ha=fa;Ia=ya;Ja=za;Ka=E;La=D}while(0);da=da+1|0;if((da|0)==(v|0)){Tb=Ha;Ub=Ia;Vb=Ja;Wb=Ka;Xb=La;break}else{fa=Ha;ya=Ia;za=Ja;E=Ka;D=La}}switch(Ub|0){case 0:{Yb=0.0;Zb=0.0;break}case 1:{Yb=+p[K>>3];Zb=+p[Q>>3];break}default:{gx(K,Ub,8,27);gx(Q,Ub,8,27);za=(Ub+-1|0)/2|0;ya=(Ub|0)/2|0;Yb=(+p[K+(za<<3)>>3]+ +p[K+(ya<<3)>>3])*.5;Zb=(+p[Q+(za<<3)>>3]+ +p[Q+(ya<<3)>>3])*.5}}switch(Vb|0){case 0:{_b=Tb;$b=Wb;ac=Xb;bc=0.0;cc=Yb;dc=Zb;break h;break}case 1:{_b=Tb;$b=Wb;ac=Xb;bc=+p[J>>3];cc=Yb;dc=Zb;break h;break}default:{gx(J,Vb,8,27);_b=Tb;$b=Wb;ac=Xb;bc=(+p[J+(((Vb+-1|0)/2|0)<<3)>>3]+ +p[J+(((Vb|0)/2|0)<<3)>>3])*.5;cc=Yb;dc=Zb;break h}}}else{_b=0;$b=-1797693134862315708145274.0e284;ac=1797693134862315708145274.0e284;bc=0.0;cc=0.0;dc=0.0}while(0);dy(Q);dy(K);dy(J);dy(u);dy(w);dy(C);sa=_b;ta=bc*1.0483579;ua=cc*.6052697;va=dc*.1772048;wa=ac;xa=$b}}while(0);if((e|0)!=0&(sa|0)==0){ec=0.0;fc=1.0;gc=1.0}else{t=ta!=0.0&ta<ua?ta:ua;ec=wa;fc=xa;gc=va!=0.0&va<t?va:t}t=g==0.0?gc*.25:gc/s;if(t==0.0){r=0;return r|0}else{x=fc;y=ec;z=sa;A=t}}while(0);ec=(x-y)/A;if(ec>4294967284.0){r=0;return r|0}sa=(a|0)>0;if(sa){e=k[30368]|0;do if(!e)if(!(Ss()|0)){hc=k[30368]|0;break}else{r=113;return r|0}else hc=e;while(0);e=(a+-1|0)%1e4|0;ic=e;jc=~~(+o[hc+(e<<2)>>2]*500.0)}else{ic=0;jc=0}t:do if((z|0)==(q|0)){e=(h|0)==2;do if(!e)if(ec<2147483637.0){fc=y/A+.5;kc=A*(+(~~fc>>>0>>>0)+4294967296.0*+((+W(fc)>=1.0?(fc>0.0?~~+ia(+V(fc/4294967296.0),4294967295.0)>>>0:~~+ga((fc-+(~~fc>>>0))/4294967296.0)>>>0):0)|0));break}else{kc=(y+x)*.5;break}else kc=y+A*2147483637.0;while(0);hc=(q|0)>0;if(!sa){if(hc)lc=0;else{mc=kc;break}while(1){fc=(+p[b+(lc<<3)>>3]-kc)/A;k[i+(lc<<2)>>2]=~~(fc+(fc>=0.0?.5:-.5));lc=lc+1|0;if((lc|0)==(q|0)){mc=kc;break t}}}if(hc){a=k[30368]|0;_b=0;Vb=ic;Tb=jc;while(1){fc=+p[b+(_b<<3)>>3];if(e&fc==0.0)k[i+(_b<<2)>>2]=-2147483646;else{s=(fc-kc)/A+ +o[a+(Tb<<2)>>2]+-.5;k[i+(_b<<2)>>2]=~~(s+(s>=0.0?.5:-.5))}Ub=Tb+1|0;if((Ub|0)==1e4){Ja=Vb+1|0;Ia=(Ja|0)==1e4?0:Ja;nc=Ia;oc=~~(+o[a+(Ia<<2)>>2]*500.0)}else{nc=Vb;oc=Ub}_b=_b+1|0;if((_b|0)==(q|0)){mc=kc;break}else{Vb=nc;Tb=oc}}}else mc=kc}else{s=y+A*2147483637.0;Tb=(q|0)>0;if(!sa){if(Tb)pc=0;else{mc=s;break}while(1){fc=+p[b+(pc<<3)>>3];if(fc!=f){gc=(fc-s)/A;qc=~~(gc+(gc>=0.0?.5:-.5))}else qc=-2147483647;k[i+(pc<<2)>>2]=qc;pc=pc+1|0;if((pc|0)==(q|0)){mc=s;break t}}}if(Tb){Vb=(h|0)==2;_b=k[30368]|0;a=0;e=ic;hc=jc;while(1){gc=+p[b+(a<<3)>>3];do if(gc!=f)if(Vb&gc==0.0){k[i+(a<<2)>>2]=-2147483646;break}else{fc=(gc-s)/A+ +o[_b+(hc<<2)>>2]+-.5;k[i+(a<<2)>>2]=~~(fc+(fc>=0.0?.5:-.5));break}else k[i+(a<<2)>>2]=-2147483647;while(0);C=hc+1|0;if((C|0)==1e4){w=e+1|0;u=(w|0)==1e4?0:w;rc=u;sc=~~(+o[_b+(u<<2)>>2]*500.0)}else{rc=e;sc=C}a=a+1|0;if((a|0)==(q|0)){mc=s;break}else{e=rc;hc=sc}}}else mc=s}while(0);f=(y-mc)/A;k[m>>2]=~~(f+(f>=0.0?.5:-.5));f=(x-mc)/A;k[n>>2]=~~(f+(f>=0.0?.5:-.5));p[j>>3]=A;p[l>>3]=mc;r=1;return r|0}function wt(a,b,c,d,e,f,g,h,i,j,l,m,n,o,q){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;n=n|0;o=o|0;q=q|0;var r=0,s=0,t=0,u=0,v=0,w=0.0,x=0.0,y=0,z=0.0,A=0.0,B=0.0,C=0,D=0.0,E=0.0,F=0,G=0,H=0,I=0.0,J=0.0,K=0,M=0.0,N=0.0,O=0,P=0.0,Q=0,R=0,S=0,T=0.0,U=0.0,V=0,Y=0,Z=0.0,_=0.0,$=0.0,aa=0.0,ba=0,ca=0,da=0.0,ea=0.0,fa=0.0,ga=0.0,ia=0.0,ja=0,ka=0,la=0,ma=0.0,na=0.0,oa=0,pa=0.0,qa=0.0,ra=0,sa=0,ta=0.0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0.0,Ca=0.0,Da=0.0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0.0,_b=0.0,$b=0,ac=0,bc=0,cc=0.0,dc=0.0,ec=0.0;r=(i|0)!=0;s=(j|0)!=0;if(r|s){t=ha(c,b)|0;u=(t|0)>0;if(!d)if(u){v=0;w=0.0;x=0.0;y=a;while(1){z=+(k[y>>2]|0);A=w+z;B=x+z*z;v=v+1|0;if((v|0)==(t|0)){C=t;D=A;E=B;F=9;break}else{w=A;x=B;y=y+4|0}}}else F=12;else if(u){u=0;y=0;x=0.0;w=0.0;v=a;while(1){G=k[v>>2]|0;if((G|0)==(e|0)){H=y;I=x;J=w}else{B=+(G|0);H=y+1|0;I=x+B;J=w+B*B}u=u+1|0;if((u|0)==(t|0)){C=H;D=I;E=J;F=9;break}else{y=H;x=I;w=J;v=v+4|0}}}else F=12;do if((F|0)==9)if((C|0)<=1)if((C|0)==1){K=1;M=D;N=0.0;break}else{F=12;break}else{J=+(C|0);w=D/J;K=C;M=w;N=+X(+(E/J-w*w));break}while(0);if((F|0)==12){K=0;M=0.0;N=0.0}if(f)k[f>>2]=K;if(r)p[i>>3]=M;if(s){p[j>>3]=N;O=K}else O=K}else O=0;if(l){do if((b|0)<3)P=0.0;else{K=ey(b,4)|0;if(!K){k[q>>2]=113;P=0.0;break}j=ey(c,8)|0;if(!j){dy(K);k[q>>2]=113;P=0.0;break}a:do if((c|0)>0){s=(d|0)!=0;i=0;r=0;while(1){C=ha(i,b)|0;b:do if(s){v=0;while(1){if((k[a+(v+C<<2)>>2]|0)!=(e|0)){Q=v;break b}H=v+1|0;if((H|0)<(b|0))v=H;else{Q=H;break}}}else Q=0;while(0);if((Q|0)!=(b|0)?(v=Q+1|0,(v|0)<(b|0)):0){H=v;v=0;y=k[a+(Q+C<<2)>>2]|0;while(1){c:do if(s&(H|0)<(b|0)){t=H;while(1){if((k[a+(t+C<<2)>>2]|0)!=(e|0)){R=t;break c}u=t+1|0;if((u|0)<(b|0))t=u;else{R=u;break}}}else R=H;while(0);if((R|0)==(b|0)){S=v;break}t=y;y=k[a+(R+C<<2)>>2]|0;k[K+(v<<2)>>2]=t-y;t=v+1|0;H=R+1|0;if((H|0)>=(b|0)){S=t;break}else v=t}if((S|0)>=2){v=0;N=0.0;M=0.0;H=K;while(1){E=+(k[H>>2]|0);D=N+E;w=M+E*E;v=v+1|0;if((v|0)==(S|0)){T=D;U=w;break}else{N=D;M=w;H=H+4|0}}do if((S|0)>1){M=+(S|0);N=T/M;w=+X(+(U/M-N*N));if(w>0.0){M=w*5.0;H=0;v=0;while(1){y=k[K+(H<<2)>>2]|0;if(+W(+(+(y|0)-N))<M){if((v|0)<(H|0))k[K+(v<<2)>>2]=y;V=v+1|0}else V=v;H=H+1|0;if((H|0)==(S|0)){Y=V;break}else v=V}if((Y|0)!=(S|0)){v=(Y|0)>0;do if(v){H=0;M=0.0;N=0.0;y=K;while(1){D=+(k[y>>2]|0);E=M+D;J=N+D*D;H=H+1|0;if((H|0)==(Y|0)){Z=E;_=J;break}else{M=E;N=J;y=y+4|0}}if((Y|0)<=1)if((Y|0)==1){$=0.0;aa=Z;break}else{F=54;break}else{N=+(Y|0);M=Z/N;$=+X(+(_/N-M*M));aa=M;break}}else F=54;while(0);if((F|0)==54){F=0;$=0.0;aa=0.0}if(v){M=$*5.0;y=0;H=0;while(1){C=k[K+(y<<2)>>2]|0;if(+W(+(+(C|0)-aa))<M){if((H|0)<(y|0))k[K+(H<<2)>>2]=C;ba=H+1|0}else ba=H;y=y+1|0;if((y|0)==(Y|0)){ca=ba;break}else H=ba}}else ca=0;if((ca|0)==(Y|0)){da=$;break}H=(ca|0)>0;do if(H){y=0;M=0.0;N=0.0;v=K;while(1){J=+(k[v>>2]|0);E=M+J;D=N+J*J;y=y+1|0;if((y|0)==(ca|0)){ea=E;fa=D;break}else{M=E;N=D;v=v+4|0}}if((ca|0)<=1)if((ca|0)==1){ga=0.0;ia=ea;break}else{F=73;break}else{N=+(ca|0);M=ea/N;ga=+X(+(fa/N-M*M));ia=M;break}}else F=73;while(0);if((F|0)==73){F=0;ga=0.0;ia=0.0}if(H){M=ga*5.0;v=0;y=0;while(1){C=k[K+(v<<2)>>2]|0;if(+W(+(+(C|0)-ia))<M){if((y|0)<(v|0))k[K+(y<<2)>>2]=C;ja=y+1|0}else ja=y;v=v+1|0;if((v|0)==(ca|0)){ka=ja;break}else y=ja}}else ka=0;if((ka|0)==(ca|0)){da=ga;break}if((ka|0)>0){la=0;ma=0.0;na=0.0;oa=K}else{da=0.0;break}while(1){M=+(k[oa>>2]|0);N=ma+M;D=na+M*M;la=la+1|0;if((la|0)==(ka|0)){pa=N;qa=D;break}else{ma=N;na=D;oa=oa+4|0}}if((ka|0)<=1){da=0.0;break}D=+(ka|0);N=pa/D;da=+X(+(qa/D-N*N))}else da=w}else da=w}else da=0.0;while(0);p[j+(r<<3)>>3]=da;ra=r+1|0}else ra=r}else ra=r;i=i+1|0;if((i|0)==(c|0)){sa=ra;break}else r=ra}switch(sa|0){case 0:{ta=0.0;break a;break}case 1:{ta=+p[j>>3];break a;break}default:{gx(j,sa,8,27);ta=(+p[j+(((sa+-1|0)/2|0)<<3)>>3]+ +p[j+(((sa|0)/2|0)<<3)>>3])*.5;break a}}}else ta=0.0;while(0);dy(j);dy(K);P=ta*.70710678}while(0);p[l>>3]=P}l=(g|0)!=0;sa=(h|0)!=0;ra=(n|0)!=0;if(!(l|sa|ra)){ua=k[q>>2]|0;return ua|0}ka=(b|0)<9;oa=ka?1:c;la=ha(ka?c:1,b)|0;do if((la|0)<9)if((la|0)>0){b=(d|0)!=0;c=0;ka=0;ca=-2147483648;ja=2147483647;while(1){F=k[a+(c<<2)>>2]|0;if(b&(F|0)==(e|0)){va=ka;wa=ca;xa=ja}else{va=ka+1|0;wa=(F|0)>(ca|0)?F:ca;xa=(F|0)<(ja|0)?F:ja}c=c+1|0;if((c|0)==(la|0)){ya=va;za=xa;Aa=wa;Ba=0.0;Ca=0.0;Da=0.0;break}else{ka=va;ca=wa;ja=xa}}}else{ya=0;za=2147483647;Aa=-2147483648;Ba=0.0;Ca=0.0;Da=0.0}else{ja=ey(la,8)|0;if(!ja){k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}ca=ey(la,8)|0;if(!ca){dy(ja);k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}ka=ey(la,8)|0;if(!ka){dy(ja);dy(ca);k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}c=ey(oa,8)|0;if(!c){dy(ja);dy(ca);dy(ka);k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}b=ey(oa,8)|0;if(!b){dy(ja);dy(ca);dy(ka);dy(c);k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}K=ey(oa,8)|0;if(!K){dy(ja);dy(ca);dy(ka);dy(c);dy(b);k[q>>2]=113;ya=O;za=0;Aa=0;Ba=0.0;Ca=0.0;Da=0.0;break}d:do if((oa|0)>0){j=(d|0)!=0;F=0;Y=0;ba=0;S=0;V=-2147483648;R=2147483647;while(1){Q=ha(F,la)|0;e:do if(j){r=0;while(1){if((k[a+(r+Q<<2)>>2]|0)!=(e|0)){Ea=r;break e}i=r+1|0;if((i|0)<(la|0))r=i;else{Ea=i;break}}}else Ea=0;while(0);f:do if((Ea|0)!=(la|0)){r=k[a+(Ea+Q<<2)>>2]|0;i=Y+1|0;s=(r|0)<(R|0)?r:R;y=(r|0)>(V|0)?r:V;v=Ea+1|0;g:do if(j&(v|0)<(la|0)){H=v;while(1){if((k[a+(H+Q<<2)>>2]|0)!=(e|0)){Fa=H;break g}C=H+1|0;if((C|0)<(la|0))H=C;else{Fa=C;break}}}else Fa=v;while(0);if((Fa|0)==(la|0)){Ga=i;Ha=ba;Ia=S;Ja=y;Ka=s}else{v=k[a+(Fa+Q<<2)>>2]|0;H=Y+2|0;C=(v|0)<(s|0)?v:s;t=(v|0)>(y|0)?v:y;u=Fa+1|0;h:do if(j&(u|0)<(la|0)){G=u;while(1){if((k[a+(G+Q<<2)>>2]|0)!=(e|0)){La=G;break h}Ma=G+1|0;if((Ma|0)<(la|0))G=Ma;else{La=Ma;break}}}else La=u;while(0);if((La|0)==(la|0)){Ga=H;Ha=ba;Ia=S;Ja=t;Ka=C;break}u=k[a+(La+Q<<2)>>2]|0;y=Y+3|0;s=(u|0)<(C|0)?u:C;i=(u|0)>(t|0)?u:t;G=La+1|0;i:do if(j&(G|0)<(la|0)){Ma=G;while(1){if((k[a+(Ma+Q<<2)>>2]|0)!=(e|0)){Na=Ma;break i}Oa=Ma+1|0;if((Oa|0)<(la|0))Ma=Oa;else{Na=Oa;break}}}else Na=G;while(0);if((Na|0)==(la|0)){Ga=y;Ha=ba;Ia=S;Ja=i;Ka=s;break}G=k[a+(Na+Q<<2)>>2]|0;t=Y+4|0;C=(G|0)<(s|0)?G:s;H=(G|0)>(i|0)?G:i;Ma=Na+1|0;j:do if(j&(Ma|0)<(la|0)){Oa=Ma;while(1){if((k[a+(Oa+Q<<2)>>2]|0)!=(e|0)){Pa=Oa;break j}Qa=Oa+1|0;if((Qa|0)<(la|0))Oa=Qa;else{Pa=Qa;break}}}else Pa=Ma;while(0);if((Pa|0)==(la|0)){Ga=t;Ha=ba;Ia=S;Ja=H;Ka=C;break}Ma=k[a+(Pa+Q<<2)>>2]|0;i=Y+5|0;s=(Ma|0)<(C|0)?Ma:C;y=(Ma|0)>(H|0)?Ma:H;Oa=Pa+1|0;k:do if(j&(Oa|0)<(la|0)){Qa=Oa;while(1){if((k[a+(Qa+Q<<2)>>2]|0)!=(e|0)){Ra=Qa;break k}Sa=Qa+1|0;if((Sa|0)<(la|0))Qa=Sa;else{Ra=Sa;break}}}else Ra=Oa;while(0);if((Ra|0)==(la|0)){Ga=i;Ha=ba;Ia=S;Ja=y;Ka=s;break}Oa=k[a+(Ra+Q<<2)>>2]|0;H=Y+6|0;C=(Oa|0)<(s|0)?Oa:s;t=(Oa|0)>(y|0)?Oa:y;Qa=Ra+1|0;l:do if(j&(Qa|0)<(la|0)){Sa=Qa;while(1){if((k[a+(Sa+Q<<2)>>2]|0)!=(e|0)){Ta=Sa;break l}Ua=Sa+1|0;if((Ua|0)<(la|0))Sa=Ua;else{Ta=Ua;break}}}else Ta=Qa;while(0);if((Ta|0)==(la|0)){Ga=H;Ha=ba;Ia=S;Ja=t;Ka=C;break}Qa=k[a+(Ta+Q<<2)>>2]|0;y=Y+7|0;s=(Qa|0)<(C|0)?Qa:C;i=(Qa|0)>(t|0)?Qa:t;Sa=Ta+1|0;m:do if(j&(Sa|0)<(la|0)){Ua=Sa;while(1){if((k[a+(Ua+Q<<2)>>2]|0)!=(e|0)){Va=Ua;break m}Wa=Ua+1|0;if((Wa|0)<(la|0))Ua=Wa;else{Va=Wa;break}}}else Va=Sa;while(0);if((Va|0)==(la|0)){Ga=y;Ha=ba;Ia=S;Ja=i;Ka=s;break}Sa=k[a+(Va+Q<<2)>>2]|0;t=Y+8|0;C=(Sa|0)<(s|0)?Sa:s;H=(Sa|0)>(i|0)?Sa:i;Ua=Va+1|0;if((Ua|0)<(la|0)){Xa=Ua;Ya=t;Za=0;_a=0;$a=r;ab=v;bb=u;cb=G;db=Ma;eb=Oa;fb=Qa;gb=Sa;hb=H;ib=C}else{Ga=t;Ha=ba;Ia=S;Ja=H;Ka=C;break}n:while(1){C=Xa;H=Za;t=_a;Sa=$a;Ua=ab;Wa=bb;jb=cb;kb=db;lb=eb;mb=fb;nb=gb;ob=hb;pb=ib;while(1){o:do if(j&(C|0)<(la|0)){qb=C;while(1){if((k[a+(qb+Q<<2)>>2]|0)!=(e|0)){rb=qb;break o}sb=qb+1|0;if((sb|0)<(la|0))qb=sb;else{rb=sb;break}}}else rb=C;while(0);if((rb|0)==(la|0)){tb=Ya;ub=H;vb=t;wb=ob;xb=pb;break n}qb=k[a+(rb+Q<<2)>>2]|0;sb=(qb|0)<(pb|0)?qb:pb;yb=(qb|0)>(ob|0)?qb:ob;zb=(kb|0)==(lb|0);Ab=(lb|0)==(mb|0);if(Ab&zb)Bb=t;else{Cb=jy(kb|0,((kb|0)<0)<<31>>31|0,mb|0,((mb|0)<0)<<31>>31|0)|0;Db=L;Eb=(Db|0)<0;Fb=jy(0,0,Cb|0,Db|0)|0;Gb=ja+(t<<3)|0;k[Gb>>2]=Eb?Fb:Cb;k[Gb+4>>2]=Eb?L:Db;Bb=t+1|0}if(Ab&(zb&((jb|0)==(kb|0)&(Wa|0)==(jb|0)))){Hb=qb;Ib=yb;Jb=sb;Kb=rb;Lb=H;Mb=Bb;Nb=Ua;Ob=mb;Pb=nb;break}zb=((kb|0)<0)<<31>>31;Ab=ty(kb|0,zb|0,1)|0;Db=((Wa|0)<0)<<31>>31;Eb=jy(Ab|0,L|0,Wa|0,Db|0)|0;Ab=((mb|0)<0)<<31>>31;Gb=jy(Eb|0,L|0,mb|0,Ab|0)|0;Eb=L;Cb=(Eb|0)<0;Fb=jy(0,0,Gb|0,Eb|0)|0;Qb=ca+(H<<3)|0;k[Qb>>2]=Cb?Fb:Gb;k[Qb+4>>2]=Cb?L:Eb;Eb=yy(kb|0,zb|0,6,0)|0;zb=L;Cb=ty(Wa|0,Db|0,2)|0;Db=L;Qb=ty(mb|0,Ab|0,2)|0;Ab=jy(Eb|0,zb|0,Qb|0,L|0)|0;Qb=jy(Ab|0,L|0,Cb|0,Db|0)|0;Db=ky(Qb|0,L|0,Sa|0,((Sa|0)<0)<<31>>31|0)|0;Qb=ky(Db|0,L|0,qb|0,((qb|0)<0)<<31>>31|0)|0;Db=L;Cb=(Db|0)<0;Ab=jy(0,0,Qb|0,Db|0)|0;zb=ka+(H<<3)|0;k[zb>>2]=Cb?Ab:Qb;k[zb+4>>2]=Cb?L:Db;Db=H+1|0;C=rb+1|0;if((C|0)>=(la|0)){tb=Ya;ub=Db;vb=Bb;wb=yb;xb=sb;break n}else{Cb=nb;zb=mb;Qb=lb;Ab=kb;Eb=jb;Gb=Wa;Fb=Ua;H=Db;t=Bb;nb=qb;ob=yb;pb=sb;mb=Cb;lb=zb;kb=Qb;jb=Ab;Wa=Eb;Ua=Gb;Sa=Fb}}Sa=Ya+1|0;Ua=Kb+1|0;if((Ua|0)<(la|0)){Xa=Ua;Ya=Sa;Za=Lb;_a=Mb;$a=Nb;ab=Ob;bb=Ob;cb=Ob;db=Ob;eb=Ob;fb=Pb;gb=Hb;hb=Ib;ib=Jb}else{tb=Sa;ub=Lb;vb=Mb;wb=Ib;xb=Jb;break}}Qa=ub+tb|0;switch(ub|0){case 0:{Ga=Qa;Ha=ba;Ia=S;Ja=wb;Ka=xb;break f;break}case 1:{if((vb|0)==1){Oa=ja;p[c+(S<<3)>>3]=+((k[Oa>>2]|0)>>>0)+4294967296.0*+(k[Oa+4>>2]|0);Rb=S+1|0}else Rb=S;Oa=ca;p[b+(ba<<3)>>3]=+((k[Oa>>2]|0)>>>0)+4294967296.0*+(k[Oa+4>>2]|0);Oa=ka;p[K+(ba<<3)>>3]=+((k[Oa>>2]|0)>>>0)+4294967296.0*+(k[Oa+4>>2]|0);Sb=Rb;break}default:{if((vb|0)>1){Oa=At(ja,ub)|0;p[c+(S<<3)>>3]=+(Oa>>>0)+4294967296.0*+(L|0);Tb=S+1|0}else Tb=S;Oa=At(ca,ub)|0;p[b+(ba<<3)>>3]=+(Oa>>>0)+4294967296.0*+(L|0);Oa=At(ka,ub)|0;p[K+(ba<<3)>>3]=+(Oa>>>0)+4294967296.0*+(L|0);Sb=Tb}}Ga=Qa;Ha=ba+1|0;Ia=Sb;Ja=wb;Ka=xb}}else{Ga=Y;Ha=ba;Ia=S;Ja=V;Ka=R}while(0);F=F+1|0;if((F|0)==(oa|0)){Ub=Ga;Vb=Ha;Wb=Ia;Xb=Ja;Yb=Ka;break}else{Y=Ga;ba=Ha;S=Ia;V=Ja;R=Ka}}switch(Vb|0){case 0:{Zb=0.0;_b=0.0;break}case 1:{Zb=+p[b>>3];_b=+p[K>>3];break}default:{gx(b,Vb,8,27);gx(K,Vb,8,27);R=(Vb+-1|0)/2|0;V=(Vb|0)/2|0;Zb=(+p[b+(R<<3)>>3]+ +p[b+(V<<3)>>3])*.5;_b=(+p[K+(R<<3)>>3]+ +p[K+(V<<3)>>3])*.5}}switch(Wb|0){case 0:{$b=Ub;ac=Xb;bc=Yb;cc=0.0;dc=Zb;ec=_b;break d;break}case 1:{$b=Ub;ac=Xb;bc=Yb;cc=+p[c>>3];dc=Zb;ec=_b;break d;break}default:{gx(c,Wb,8,27);$b=Ub;ac=Xb;bc=Yb;cc=(+p[c+(((Wb+-1|0)/2|0)<<3)>>3]+ +p[c+(((Wb|0)/2|0)<<3)>>3])*.5;dc=Zb;ec=_b;break d}}}else{$b=0;ac=-2147483648;bc=2147483647;cc=0.0;dc=0.0;ec=0.0}while(0);dy(K);dy(b);dy(c);dy(ka);dy(ca);dy(ja);ya=$b;za=bc;Aa=ac;Ba=cc*1.0483579;Ca=dc*.6052697;Da=ec*.1772048}while(0);if(f)k[f>>2]=ya;if(l)k[g>>2]=za;if(sa)k[h>>2]=Aa;if(m)p[m>>3]=Ba;if(ra)p[n>>3]=Ca;if(!o){ua=k[q>>2]|0;return ua|0}p[o>>3]=Da;ua=k[q>>2]|0;return ua|0}function xt(a,b,c,d,e,f,g,h,i,j,l,m){a=a|0;b=b|0;c=c|0;d=d|0;e=+e;f=f|0;g=g|0;h=h|0;i=i|0;j=j|0;l=l|0;m=m|0;var n=0,q=0,r=0,s=0.0,t=0.0,u=0.0,v=0.0,w=0.0,x=0,y=0.0,z=0.0,A=0,B=0,C=0.0,D=0.0,E=0,F=0.0,G=0.0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0.0,S=0.0,T=0,U=0,V=0,X=0.0,Y=0.0,Z=0,_=0,$=0,aa=0,ba=0,ca=0.0,da=0.0,ea=0.0,fa=0.0,ga=0,ia=0.0,ja=0.0,ka=0,la=0.0,ma=0.0,na=0,oa=0.0,pa=0.0,qa=0,ra=0.0,sa=0.0,ta=0,ua=0.0,va=0.0,wa=0,xa=0,ya=0,za=0,Aa=0.0,Ba=0.0,Ca=0.0,Da=0.0,Ea=0.0,Fa=0.0,Ga=0.0,Ha=0.0,Ia=0.0,Ja=0.0,Ka=0.0,La=0.0,Ma=0.0,Na=0.0,Oa=0.0,Pa=0.0,Qa=0.0,Ra=0.0,Sa=0.0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0.0,_a=0.0,$a=0.0,ab=0.0,bb=0.0,cb=0.0,db=0,eb=0.0,fb=0,gb=0,hb=0,ib=0.0,jb=0.0,kb=0.0,lb=0.0,mb=0.0,nb=0.0,ob=0.0,pb=0.0,qb=0.0,rb=0.0,sb=0.0,tb=0.0,ub=0.0,vb=0.0,wb=0.0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0.0,Eb=0.0,Fb=0.0,Gb=0.0,Hb=0,Ib=0.0,Jb=0.0,Kb=0.0,Lb=0.0,Mb=0.0;n=(b|0)<9;q=n?1:c;r=ha(n?c:1,b)|0;if((r|0)<9){a:do if((r|0)>0){if(!d){b=0;c=0;s=-3402823466385288598117041.0e14;t=3402823466385288598117041.0e14;while(1){u=+o[a+(b<<2)>>2];v=u<t?u:t;w=u>s?u:s;n=c+1|0;b=b+1|0;if((b|0)==(r|0)){x=n;y=w;z=v;break a}else{c=n;s=w;t=v}}}else{A=0;B=0;C=-3402823466385288598117041.0e14;D=3402823466385288598117041.0e14}while(1){t=+o[a+(A<<2)>>2];if(t==e){E=B;F=C;G=D}else{E=B+1|0;F=t>C?t:C;G=t<D?t:D}A=A+1|0;if((A|0)==(r|0)){x=E;y=F;z=G;break}else{B=E;C=F;D=G}}}else{x=0;y=-3402823466385288598117041.0e14;z=3402823466385288598117041.0e14}while(0);if(g)o[g>>2]=z;if(h)o[h>>2]=y;if(f)k[f>>2]=x;if(i)p[i>>3]=0.0;if(j)p[j>>3]=0.0;if(!l)return;p[l>>3]=0.0;return}x=(g|0)!=0;E=(h|0)!=0;B=x|E;A=ey(r,4)|0;if(!A){k[m>>2]=113;return}c=ey(r,4)|0;if(!c){dy(A);k[m>>2]=113;return}b=ey(r,4)|0;if(!b){dy(A);dy(c);k[m>>2]=113;return}n=ey(q,8)|0;if(!n){dy(A);dy(c);dy(b);k[m>>2]=113;return}H=ey(q,8)|0;if(!H){dy(A);dy(c);dy(b);dy(n);k[m>>2]=113;return}I=ey(q,8)|0;if(!I){dy(A);dy(c);dy(b);dy(n);dy(H);k[m>>2]=113;return}b:do if((q|0)>0){m=(d|0)!=0;J=0;K=0;L=0;M=0;y=-3402823466385288598117041.0e14;z=3402823466385288598117041.0e14;while(1){N=ha(J,r)|0;c:do if(m){O=0;while(1){if(!(+o[a+(O+N<<2)>>2]==e)){P=O;break c}Q=O+1|0;if((Q|0)<(r|0))O=Q;else{P=Q;break}}}else P=0;while(0);d:do if((P|0)!=(r|0)){G=+o[a+(P+N<<2)>>2];O=K+1|0;if(B){D=G<z?G:z;if(G>y){R=G;S=D}else{R=y;S=D}}else{R=y;S=z}Q=P+1|0;e:do if(m&(Q|0)<(r|0)){T=Q;while(1){if(!(+o[a+(T+N<<2)>>2]==e)){U=T;break e}V=T+1|0;if((V|0)<(r|0))T=V;else{U=V;break}}}else U=Q;while(0);if((U|0)!=(r|0)){D=+o[a+(U+N<<2)>>2];Q=K+2|0;if(B){F=D<S?D:S;if(D>R){X=D;Y=F}else{X=R;Y=F}}else{X=R;Y=S}T=U+1|0;f:do if(m&(T|0)<(r|0)){V=T;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){Z=V;break f}_=V+1|0;if((_|0)<(r|0))V=_;else{Z=_;break}}}else Z=T;while(0);if((Z|0)==(r|0)){$=Q;aa=L;ba=M;ca=X;da=Y}else{F=+o[a+(Z+N<<2)>>2];T=K+3|0;do if(B){C=F<Y?F:Y;if(!(F>X)){ea=X;fa=C;break}ea=F;fa=C}else{ea=X;fa=Y}while(0);Q=Z+1|0;g:do if(m&(Q|0)<(r|0)){V=Q;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){ga=V;break g}_=V+1|0;if((_|0)<(r|0))V=_;else{ga=_;break}}}else ga=Q;while(0);if((ga|0)==(r|0)){$=T;aa=L;ba=M;ca=ea;da=fa;break}C=+o[a+(ga+N<<2)>>2];Q=K+4|0;do if(B){t=C<fa?C:fa;if(!(C>ea)){ia=ea;ja=t;break}ia=C;ja=t}else{ia=ea;ja=fa}while(0);T=ga+1|0;h:do if(m&(T|0)<(r|0)){V=T;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){ka=V;break h}_=V+1|0;if((_|0)<(r|0))V=_;else{ka=_;break}}}else ka=T;while(0);if((ka|0)==(r|0)){$=Q;aa=L;ba=M;ca=ia;da=ja;break}t=+o[a+(ka+N<<2)>>2];T=K+5|0;do if(B){s=t<ja?t:ja;if(!(t>ia)){la=ia;ma=s;break}la=t;ma=s}else{la=ia;ma=ja}while(0);Q=ka+1|0;i:do if(m&(Q|0)<(r|0)){V=Q;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){na=V;break i}_=V+1|0;if((_|0)<(r|0))V=_;else{na=_;break}}}else na=Q;while(0);if((na|0)==(r|0)){$=T;aa=L;ba=M;ca=la;da=ma;break}s=+o[a+(na+N<<2)>>2];Q=K+6|0;do if(B){v=s<ma?s:ma;if(!(s>la)){oa=la;pa=v;break}oa=s;pa=v}else{oa=la;pa=ma}while(0);T=na+1|0;j:do if(m&(T|0)<(r|0)){V=T;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){qa=V;break j}_=V+1|0;if((_|0)<(r|0))V=_;else{qa=_;break}}}else qa=T;while(0);if((qa|0)==(r|0)){$=Q;aa=L;ba=M;ca=oa;da=pa;break}v=+o[a+(qa+N<<2)>>2];T=K+7|0;do if(B){w=v<pa?v:pa;if(!(v>oa)){ra=oa;sa=w;break}ra=v;sa=w}else{ra=oa;sa=pa}while(0);Q=qa+1|0;k:do if(m&(Q|0)<(r|0)){V=Q;while(1){if(!(+o[a+(V+N<<2)>>2]==e)){ta=V;break k}_=V+1|0;if((_|0)<(r|0))V=_;else{ta=_;break}}}else ta=Q;while(0);if((ta|0)==(r|0)){$=T;aa=L;ba=M;ca=ra;da=sa;break}w=+o[a+(ta+N<<2)>>2];Q=K+8|0;do if(B){u=w<sa?w:sa;if(!(w>ra)){ua=ra;va=u;break}ua=w;va=u}else{ua=ra;va=sa}while(0);T=ta+1|0;if((T|0)<(r|0)){wa=T;xa=Q;ya=0;za=0;Aa=G;Ba=D;Ca=F;Da=C;Ea=t;Fa=s;Ga=v;Ha=w;Ia=ua;Ja=va}else{$=Q;aa=L;ba=M;ca=ua;da=va;break}l:while(1){T=wa;V=ya;_=za;u=Aa;Ka=Ba;La=Ca;Ma=Da;Na=Ea;Oa=Fa;Pa=Ga;Qa=Ha;Ra=Ia;Sa=Ja;while(1){m:do if(m&(T|0)<(r|0)){Ta=T;while(1){if(!(+o[a+(Ta+N<<2)>>2]==e)){Ua=Ta;break m}Va=Ta+1|0;if((Va|0)<(r|0))Ta=Va;else{Ua=Va;break}}}else Ua=T;while(0);if((Ua|0)==(r|0)){Wa=xa;Xa=V;Ya=_;Za=Ra;_a=Sa;break l}$a=+o[a+(Ua+N<<2)>>2];do if(B){ab=$a<Sa?$a:Sa;if(!($a>Ra)){bb=Ra;cb=ab;break}bb=$a;cb=ab}else{bb=Ra;cb=Sa}while(0);Ta=Na==Oa;Va=Oa==Pa;if(Ta&Va)db=_;else{ab=+W(+(Na-Pa));o[A+(_<<2)>>2]=ab;db=_+1|0}if(Ta&(La==Ma&Ma==Na)&Va){eb=$a;fb=Ua;gb=V;hb=db;ib=Ka;jb=La;kb=Ma;lb=Na;mb=Oa;nb=Pa;ob=Qa;pb=bb;qb=cb;break}ab=+W(+(Na*2.0-La-Pa));o[c+(V<<2)>>2]=ab;ab=+W(+(u+(Na*6.0-La*4.0-Pa*4.0)+$a));o[b+(V<<2)>>2]=ab;Va=V+1|0;T=Ua+1|0;if((T|0)>=(r|0)){Wa=xa;Xa=Va;Ya=db;Za=bb;_a=cb;break l}else{ab=Qa;rb=Pa;sb=Oa;tb=Na;ub=Ma;vb=La;wb=Ka;V=Va;_=db;Qa=$a;Ra=bb;Sa=cb;Pa=ab;Oa=rb;Na=sb;Ma=tb;La=ub;Ka=vb;u=wb}}_=xa+1|0;V=fb+1|0;if((V|0)<(r|0)){wa=V;xa=_;ya=gb;za=hb;Aa=ib;Ba=jb;Ca=kb;Da=lb;Ea=mb;Fa=nb;Ga=ob;Ha=eb;Ia=pb;Ja=qb}else{Wa=_;Xa=gb;Ya=hb;Za=pb;_a=qb;break}}Q=Xa+Wa|0;switch(Xa|0){case 0:{$=Q;aa=L;ba=M;ca=Za;da=_a;break d;break}case 1:{if((Ya|0)==1){p[n+(M<<3)>>3]=+o[A>>2];xb=M+1|0}else xb=M;p[H+(L<<3)>>3]=+o[c>>2];p[I+(L<<3)>>3]=+o[b>>2];yb=xb;break}default:{if((Ya|0)>1){w=+Bt(A,Xa);p[n+(M<<3)>>3]=w;zb=M+1|0}else zb=M;w=+Bt(c,Xa);p[H+(L<<3)>>3]=w;w=+Bt(b,Xa);p[I+(L<<3)>>3]=w;yb=zb}}$=Q;aa=L+1|0;ba=yb;ca=Za;da=_a}}else{$=O;aa=L;ba=M;ca=R;da=S}}else{$=K;aa=L;ba=M;ca=y;da=z}while(0);J=J+1|0;if((J|0)==(q|0)){Ab=$;Bb=aa;Cb=ba;Db=ca;Eb=da;break}else{K=$;L=aa;M=ba;y=ca;z=da}}switch(Bb|0){case 0:{Fb=0.0;Gb=0.0;break}case 1:{Fb=+p[H>>3];Gb=+p[I>>3];break}default:{gx(H,Bb,8,27);gx(I,Bb,8,27);M=(Bb+-1|0)/2|0;L=(Bb|0)/2|0;Fb=(+p[H+(M<<3)>>3]+ +p[H+(L<<3)>>3])*.5;Gb=(+p[I+(M<<3)>>3]+ +p[I+(L<<3)>>3])*.5}}switch(Cb|0){case 0:{Hb=Ab;Ib=Db;Jb=Eb;Kb=0.0;Lb=Fb;Mb=Gb;break b;break}case 1:{Hb=Ab;Ib=Db;Jb=Eb;Kb=+p[n>>3];Lb=Fb;Mb=Gb;break b;break}default:{gx(n,Cb,8,27);Hb=Ab;Ib=Db;Jb=Eb;Kb=(+p[n+(((Cb+-1|0)/2|0)<<3)>>3]+ +p[n+(((Cb|0)/2|0)<<3)>>3])*.5;Lb=Fb;Mb=Gb;break b}}}else{Hb=0;Ib=-3402823466385288598117041.0e14;Jb=3402823466385288598117041.0e14;Kb=0.0;Lb=0.0;Mb=0.0}while(0);if(f)k[f>>2]=Hb;if(x)o[g>>2]=Jb;if(E)o[h>>2]=Ib;if(i)p[i>>3]=Kb*1.0483579;if(j)p[j>>3]=Lb*.6052697;if(l)p[l>>3]=Mb*.1772048;dy(I);dy(H);dy(n);dy(b);dy(c);dy(A);return}function yt(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0.0,i=0,j=0,k=0.0,l=0,m=0.0,n=0.0,o=0,q=0.0,r=0.0,s=0,t=0,u=0,v=0,w=0.0,x=0,y=0,z=0.0,A=0,B=0,C=0,D=0,E=0.0,F=0.0,G=0,H=0;c=b+-1|0;d=(c|0)/2|0;if((b|0)>1){e=c;f=0}else{g=a+(d<<3)|0;h=+p[g>>3];return +h}while(1){c=f+1|0;if((e|0)==(c|0)){i=e;j=f;break}b=a+(((f+e|0)/2|0)<<3)|0;k=+p[b>>3];l=a+(e<<3)|0;m=+p[l>>3];if(k>m){p[b>>3]=m;p[l>>3]=k;n=k}else n=m;o=a+(f<<3)|0;m=+p[o>>3];if(m>n){p[o>>3]=n;p[l>>3]=m;q=+p[o>>3]}else q=m;m=+p[b>>3];if(m>q){p[b>>3]=q;p[o>>3]=m;r=+p[b>>3]}else r=m;l=a+(c<<3)|0;p[b>>3]=+p[l>>3];p[l>>3]=r;l=e;b=c;while(1){m=+p[o>>3];c=b;while(1){s=c+1|0;t=a+(s<<3)|0;k=+p[t>>3];if(m>k)c=s;else{u=s;v=t;w=k;break}}c=l;while(1){t=c+-1|0;s=a+(t<<3)|0;k=+p[s>>3];if(k>m)c=t;else{x=t;y=s;z=k;A=c;break}}if((A|0)<=(u|0)){B=u;C=x;D=y;E=z;F=m;G=A;break}p[v>>3]=z;p[y>>3]=w;l=x;b=u}p[o>>3]=E;p[D>>3]=F;f=(C|0)>(d|0)?f:B;e=(G|0)>(d|0)?G+-2|0:e;if((e|0)<=(f|0)){H=19;break}}if((H|0)==19){g=a+(d<<3)|0;h=+p[g>>3];return +h}H=a+(j<<3)|0;F=+p[H>>3];j=a+(i<<3)|0;E=+p[j>>3];if(!(F>E)){g=a+(d<<3)|0;h=+p[g>>3];return +h}p[H>>3]=E;p[j>>3]=F;g=a+(d<<3)|0;h=+p[g>>3];return +h}function zt(a,b){a=a|0;b=b|0;var c=0.0,d=0.0;c=+p[a>>3];d=+p[b>>3];return (c<d?-1:c>d&1)|0}function At(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;c=b+-1|0;d=(c|0)/2|0;if((b|0)>1){e=c;f=0}else{g=a+(d<<3)|0;h=g;i=h;j=k[i>>2]|0;l=h+4|0;m=l;n=k[m>>2]|0;L=n;return j|0}while(1){c=f+1|0;if((e|0)==(c|0)){o=e;p=f;break}b=a+(((f+e|0)/2|0)<<3)|0;q=b;r=k[q>>2]|0;s=k[q+4>>2]|0;q=a+(e<<3)|0;t=q;u=k[t>>2]|0;v=k[t+4>>2]|0;if((s|0)>(v|0)|(s|0)==(v|0)&r>>>0>u>>>0){t=b;k[t>>2]=u;k[t+4>>2]=v;t=q;k[t>>2]=r;k[t+4>>2]=s;w=s;x=r}else{w=v;x=u}u=a+(f<<3)|0;v=u;r=k[v>>2]|0;s=k[v+4>>2]|0;if((s|0)>(w|0)|(s|0)==(w|0)&r>>>0>x>>>0){v=u;k[v>>2]=x;k[v+4>>2]=w;v=q;k[v>>2]=r;k[v+4>>2]=s;v=u;y=k[v+4>>2]|0;z=k[v>>2]|0}else{y=s;z=r}r=b;s=k[r>>2]|0;v=k[r+4>>2]|0;if((v|0)>(y|0)|(v|0)==(y|0)&s>>>0>z>>>0){r=b;k[r>>2]=z;k[r+4>>2]=y;r=u;k[r>>2]=s;k[r+4>>2]=v;r=b;A=k[r>>2]|0;B=k[r+4>>2]|0}else{A=s;B=v}v=a+(c<<3)|0;s=v;r=k[s+4>>2]|0;q=b;k[q>>2]=k[s>>2];k[q+4>>2]=r;r=v;k[r>>2]=A;k[r+4>>2]=B;r=e;v=c;while(1){c=u;q=k[c>>2]|0;s=k[c+4>>2]|0;c=v;while(1){b=c+1|0;t=a+(b<<3)|0;C=t;D=k[C>>2]|0;E=k[C+4>>2]|0;if((s|0)>(E|0)|(s|0)==(E|0)&q>>>0>D>>>0)c=b;else{F=b;G=t;H=D;I=E;break}}c=r;while(1){E=c+-1|0;D=a+(E<<3)|0;t=D;b=k[t>>2]|0;C=k[t+4>>2]|0;if((C|0)>(s|0)|(C|0)==(s|0)&b>>>0>q>>>0)c=E;else{J=E;K=D;M=b;N=C;O=c;break}}if((O|0)<=(F|0)){P=F;Q=J;R=K;S=M;T=N;U=q;V=s;W=O;break}c=G;k[c>>2]=M;k[c+4>>2]=N;c=K;k[c>>2]=H;k[c+4>>2]=I;r=J;v=F}v=u;k[v>>2]=S;k[v+4>>2]=T;v=R;k[v>>2]=U;k[v+4>>2]=V;f=(Q|0)>(d|0)?f:P;e=(W|0)>(d|0)?W+-2|0:e;if((e|0)<=(f|0)){X=19;break}}if((X|0)==19){g=a+(d<<3)|0;h=g;i=h;j=k[i>>2]|0;l=h+4|0;m=l;n=k[m>>2]|0;L=n;return j|0}X=a+(p<<3)|0;p=X;f=k[p>>2]|0;e=k[p+4>>2]|0;p=a+(o<<3)|0;o=p;W=k[o>>2]|0;P=k[o+4>>2]|0;if(!((e|0)>(P|0)|(e|0)==(P|0)&f>>>0>W>>>0)){g=a+(d<<3)|0;h=g;i=h;j=k[i>>2]|0;l=h+4|0;m=l;n=k[m>>2]|0;L=n;return j|0}o=X;k[o>>2]=W;k[o+4>>2]=P;P=p;k[P>>2]=f;k[P+4>>2]=e;g=a+(d<<3)|0;h=g;i=h;j=k[i>>2]|0;l=h+4|0;m=l;n=k[m>>2]|0;L=n;return j|0}function Bt(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0.0,i=0,j=0,l=0.0,m=0,n=0.0,p=0.0,q=0,r=0.0,s=0,u=0,v=0,w=0,x=0,y=0,z=0.0,A=0,B=0,C=0.0,D=0,E=0,F=0,G=0,H=0.0,I=0.0,J=0,K=0;c=b+-1|0;d=(c|0)/2|0;if((b|0)>1){e=c;f=0}else{g=a+(d<<2)|0;h=+o[g>>2];return +h}while(1){c=f+1|0;if((e|0)==(c|0)){i=e;j=f;break}b=a+(((f+e|0)/2|0)<<2)|0;l=+o[b>>2];m=a+(e<<2)|0;n=+o[m>>2];if(l>n){o[b>>2]=n;o[m>>2]=l;p=l}else p=n;q=a+(f<<2)|0;n=+o[q>>2];if(n>p){o[q>>2]=p;o[m>>2]=n;r=+o[q>>2]}else r=n;n=+o[b>>2];if(n>r){o[b>>2]=r;o[q>>2]=n;s=b;u=k[b>>2]|0}else{s=b;u=(o[t>>2]=n,k[t>>2]|0)}b=a+(c<<2)|0;k[s>>2]=k[b>>2];k[b>>2]=u;b=e;m=c;while(1){n=+o[q>>2];c=m;while(1){v=c+1|0;w=a+(v<<2)|0;l=+o[w>>2];if(n>l)c=v;else{x=v;y=w;z=l;break}}c=b;while(1){w=c+-1|0;v=a+(w<<2)|0;l=+o[v>>2];if(l>n)c=w;else{A=w;B=v;C=l;D=c;break}}if((D|0)<=(x|0)){E=x;F=A;G=B;H=C;I=n;J=D;break}o[y>>2]=C;o[B>>2]=z;b=A;m=x}o[q>>2]=H;o[G>>2]=I;f=(F|0)>(d|0)?f:E;e=(J|0)>(d|0)?J+-2|0:e;if((e|0)<=(f|0)){K=20;break}}if((K|0)==20){g=a+(d<<2)|0;h=+o[g>>2];return +h}K=a+(j<<2)|0;I=+o[K>>2];j=a+(i<<2)|0;H=+o[j>>2];if(!(I>H)){g=a+(d<<2)|0;h=+o[g>>2];return +h}o[K>>2]=H;o[j>>2]=I;g=a+(d<<2)|0;h=+o[g>>2];return +h}function Ct(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0,w=0,x=0,y=0,z=0.0,A=0,B=0.0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0;f=c+d|0;d=cy(e<<2)|0;if(!d){tk(782858);g=-1;return g|0}h=k[a>>2]|0;j=h>>>24;i[c>>0]=j;i[c+1>>0]=h>>>16;i[c+2>>0]=h>>>8;i[c+3>>0]=h;l=c+4|0;do if((b|0)>0){m=j<<24|h&16777215;n=l;o=0;p=k[a>>2]|0;q=8;r=e;a:while(1){s=b-o|0;r=(s|0)<(e|0)?s:r;s=(r|0)>0;if(s){t=0;u=p;v=0.0;while(1){w=k[a+(t+o<<2)>>2]|0;x=w-u|0;y=x<<1^x>>31;k[d+(t<<2)>>2]=y;z=v+ +(y>>>0);t=t+1|0;if((t|0)==(r|0)){A=w;B=z;break}else{u=w;v=z}}}else{A=p;B=0.0}v=(B-+((r|0)/2|0|0)+-1.0)/+(r|0);u=0;t=v<0.0?0:~~v>>>0;while(1){t=t>>>1;w=u+1|0;if(!t){C=w;D=u;break}else u=w}do if((D|0)>24){if((q+5|0)>32){u=5-q|0;t=k[121868+(q<<2)>>2]&26>>>u|m<<q;i[n>>0]=t;E=u;F=n+1|0;G=t;H=8}else{E=5;F=n;G=m;H=q}t=k[121868+(E<<2)>>2]&26|G<<E;u=H-E|0;if((u|0)<1){w=E-H|0;y=w&-8;x=w>>>3;w=F;I=u;while(1){i[w>>0]=t>>0-I;I=I+8|0;if((I|0)>=1)break;else w=w+1|0}J=F+(x+1)|0;K=8-E+H+y|0}else{J=F;K=u}if(s){w=t;I=J;L=0;M=K;while(1){N=k[d+(L<<2)>>2]|0;if((M|0)>0){O=32-M|0;P=k[121868+(M<<2)>>2]&N>>O|w<<M;i[I>>0]=P;Q=O;R=I+1|0;S=P;T=8}else{Q=32;R=I;S=w;T=M}P=k[121868+(Q<<2)>>2]&N|S<<Q;N=T-Q|0;if((N|0)<1){O=Q-T|0;U=O&-8;V=O>>>3;O=R;W=N;while(1){i[O>>0]=P>>0-W;W=W+8|0;if((W|0)>=1)break;else O=O+1|0}X=R+(V+1)|0;Y=8-Q+T+U|0}else{X=R;Y=N}L=L+1|0;if((L|0)==(r|0)){Z=X;_=P;$=Y;break}else{w=P;I=X;M=Y}}}else{Z=J;_=t;$=K}}else{M=(q+5|0)>32;if(B==0.0&(D|0)==0){if(M){I=m<<q;i[n>>0]=I;aa=5-q|0;ba=n+1|0;ca=I;da=8}else{aa=5;ba=n;ca=m;da=q}I=ca<<aa;w=da-aa|0;if((w|0)>=1){Z=ba;_=I;$=w;break}L=aa-da|0;u=L&-8;y=L>>>3;L=ba;x=w;while(1){i[L>>0]=I>>0-x;x=x+8|0;if((x|0)>=1)break;else L=L+1|0}Z=ba+(y+1)|0;_=I;$=8-aa+da+u|0;break}if(M){L=5-q|0;x=k[121868+(q<<2)>>2]&C>>L|m<<q;i[n>>0]=x;ea=L;fa=n+1|0;ga=x;ha=8}else{ea=5;fa=n;ga=m;ha=q}x=k[121868+(ea<<2)>>2]&C|ga<<ea;L=ha-ea|0;if((L|0)<1){t=ea-ha|0;w=t&-8;O=t>>>3;t=fa;W=L;while(1){i[t>>0]=x>>0-W;W=W+8|0;if((W|0)>=1)break;else t=t+1|0}ia=fa+(O+1)|0;ja=8-ea+ha+w|0}else{ia=fa;ja=L}t=(1<<D)+-1|0;b:do if(s){if((D|0)<=0){W=ia;M=0;u=x;I=ja;while(1){y=k[d+(M<<2)>>2]>>D;ka=y+1|0;if((I|0)<(ka|0)){i[W>>0]=u<<I;la=y-I|0;ma=W+1|0;if((la|0)>7){na=y+-8-I|0;y=na>>>3;sy(ma|0,0,y+1|0)|0;oa=W+(y+2)|0;pa=na-(y<<3)|0}else{oa=ma;pa=la}qa=oa;ra=1;sa=7-pa|0}else{qa=W;ra=u<<ka|1;sa=I-ka|0}M=M+1|0;if((M|0)==(r|0)){ta=qa;ua=ra;va=sa;break b}else{W=qa;u=ra;I=sa}}}I=8-D|0;u=ia;W=0;M=x;P=ja;while(1){N=k[d+(W<<2)>>2]|0;U=N>>D;V=U+1|0;if((P|0)<(V|0)){i[u>>0]=M<<P;ka=U-P|0;la=u+1|0;if((ka|0)>7){ma=U+-8-P|0;U=ma>>>3;sy(la|0,0,U+1|0)|0;wa=u+(U+2)|0;xa=ma-(U<<3)|0}else{wa=la;xa=ka}ya=wa;za=1;Aa=7-xa|0}else{ya=u;za=M<<V|1;Aa=P-V|0}V=za<<D|N&t;N=Aa-D|0;if((N|0)<1){ka=(D-Aa|0)>>>3;la=ka<<3;U=ya;ma=N;while(1){i[U>>0]=V>>0-ma;ma=ma+8|0;if((ma|0)>=1)break;else U=U+1|0}Ba=ya+(ka+1)|0;Ca=I+Aa+la|0}else{Ba=ya;Ca=N}W=W+1|0;if((W|0)==(r|0)){ta=Ba;ua=V;va=Ca;break}else{u=Ba;M=V;P=Ca}}}else{ta=ia;ua=x;va=ja}while(0);if(ta>>>0>f>>>0){Da=57;break a}else{Z=ta;_=ua;$=va}}while(0);o=o+e|0;if((o|0)>=(b|0)){Ea=Z;Fa=_;Ga=$;Da=59;break}else{m=_;n=Z;p=A;q=$}}if((Da|0)==57){tk(782890);dy(d);g=-1;return g|0}else if((Da|0)==59){if((Ga|0)>=8){Ha=Ea;break}i[Ea>>0]=Fa<<Ga;Ha=Ea+1|0;break}}else Ha=l;while(0);dy(d);g=Ha-c|0;return g|0}function Dt(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0.0,w=0,x=0,y=0,z=0.0,A=0,B=0.0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0;f=c+d|0;d=cy(e<<2)|0;if(!d){tk(782858);g=-1;return g|0}h=j[a>>1]|0;i[c>>0]=(h&65535)>>>8;i[c+1>>0]=h;l=c+2|0;do if((b|0)>0){m=h&65535;n=l;o=0;p=j[a>>1]|0;q=8;r=e;a:while(1){s=b-o|0;r=(s|0)<(e|0)?s:r;s=(r|0)>0;if(s){t=0;u=p;v=0.0;while(1){w=j[a+(t+o<<1)>>1]|0;x=w-u<<16;y=x>>15^x>>31;k[d+(t<<2)>>2]=y;z=v+ +(y>>>0);t=t+1|0;if((t|0)==(r|0)){A=w;B=z;break}else{u=w;v=z}}}else{A=p;B=0.0}v=(B-+((r|0)/2|0|0)+-1.0)/+(r|0);u=0;t=v<0.0?0:~~v&65535;while(1){t=(t&65535)>>>1;w=u+1|0;if(!(t<<16>>16)){C=w;D=u;break}else u=w}do if((D|0)>13){if((q+4|0)>32){u=4-q|0;t=k[121868+(q<<2)>>2]&15>>>u|m<<q;i[n>>0]=t;E=u;F=n+1|0;G=t;H=8}else{E=4;F=n;G=m;H=q}t=k[121868+(E<<2)>>2]&15|G<<E;u=H-E|0;if((u|0)<1){w=E-H|0;y=w&-8;x=w>>>3;w=F;I=u;while(1){i[w>>0]=t>>0-I;I=I+8|0;if((I|0)>=1)break;else w=w+1|0}J=F+(x+1)|0;K=8-E+H+y|0}else{J=F;K=u}if(s){w=t;I=J;L=0;M=K;while(1){N=k[d+(L<<2)>>2]|0;if((M+16|0)>32){O=16-M|0;P=k[121868+(M<<2)>>2]&N>>O|w<<M;i[I>>0]=P;Q=O;R=I+1|0;S=P;T=8}else{Q=16;R=I;S=w;T=M}P=k[121868+(Q<<2)>>2]&N|S<<Q;N=T-Q|0;if((N|0)<1){O=Q-T|0;U=O&-8;V=O>>>3;O=R;W=N;while(1){i[O>>0]=P>>0-W;W=W+8|0;if((W|0)>=1)break;else O=O+1|0}X=R+(V+1)|0;Y=8-Q+T+U|0}else{X=R;Y=N}L=L+1|0;if((L|0)==(r|0)){Z=X;_=P;$=Y;break}else{w=P;I=X;M=Y}}}else{Z=J;_=t;$=K}}else{M=(q+4|0)>32;if(B==0.0&(D|0)==0){if(M){I=m<<q;i[n>>0]=I;aa=4-q|0;ba=n+1|0;ca=I;da=8}else{aa=4;ba=n;ca=m;da=q}I=ca<<aa;w=da-aa|0;if((w|0)>=1){Z=ba;_=I;$=w;break}L=aa-da|0;u=L&-8;y=L>>>3;L=ba;x=w;while(1){i[L>>0]=I>>0-x;x=x+8|0;if((x|0)>=1)break;else L=L+1|0}Z=ba+(y+1)|0;_=I;$=8-aa+da+u|0;break}if(M){L=4-q|0;x=k[121868+(q<<2)>>2]&C>>L|m<<q;i[n>>0]=x;ea=L;fa=n+1|0;ga=x;ha=8}else{ea=4;fa=n;ga=m;ha=q}x=k[121868+(ea<<2)>>2]&C|ga<<ea;L=ha-ea|0;if((L|0)<1){t=ea-ha|0;w=t&-8;O=t>>>3;t=fa;W=L;while(1){i[t>>0]=x>>0-W;W=W+8|0;if((W|0)>=1)break;else t=t+1|0}ia=fa+(O+1)|0;ja=8-ea+ha+w|0}else{ia=fa;ja=L}t=(1<<D)+-1|0;b:do if(s){if((D|0)<=0){W=ia;M=0;u=x;I=ja;while(1){y=k[d+(M<<2)>>2]>>D;ka=y+1|0;if((I|0)<(ka|0)){i[W>>0]=u<<I;la=y-I|0;ma=W+1|0;if((la|0)>7){na=y+-8-I|0;y=na>>>3;sy(ma|0,0,y+1|0)|0;oa=W+(y+2)|0;pa=na-(y<<3)|0}else{oa=ma;pa=la}qa=oa;ra=1;sa=7-pa|0}else{qa=W;ra=u<<ka|1;sa=I-ka|0}M=M+1|0;if((M|0)==(r|0)){ta=qa;ua=ra;va=sa;break b}else{W=qa;u=ra;I=sa}}}I=8-D|0;u=ia;W=0;M=x;P=ja;while(1){N=k[d+(W<<2)>>2]|0;U=N>>D;V=U+1|0;if((P|0)<(V|0)){i[u>>0]=M<<P;ka=U-P|0;la=u+1|0;if((ka|0)>7){ma=U+-8-P|0;U=ma>>>3;sy(la|0,0,U+1|0)|0;wa=u+(U+2)|0;xa=ma-(U<<3)|0}else{wa=la;xa=ka}ya=wa;za=1;Aa=7-xa|0}else{ya=u;za=M<<V|1;Aa=P-V|0}V=za<<D|N&t;N=Aa-D|0;if((N|0)<1){ka=(D-Aa|0)>>>3;la=ka<<3;U=ya;ma=N;while(1){i[U>>0]=V>>0-ma;ma=ma+8|0;if((ma|0)>=1)break;else U=U+1|0}Ba=ya+(ka+1)|0;Ca=I+Aa+la|0}else{Ba=ya;Ca=N}W=W+1|0;if((W|0)==(r|0)){ta=Ba;ua=V;va=Ca;break}else{u=Ba;M=V;P=Ca}}}else{ta=ia;ua=x;va=ja}while(0);if(ta>>>0>f>>>0){Da=57;break a}else{Z=ta;_=ua;$=va}}while(0);o=o+e|0;if((o|0)>=(b|0)){Ea=Z;Fa=_;Ga=$;Da=59;break}else{m=_;n=Z;p=A;q=$}}if((Da|0)==57){tk(782890);dy(d);g=-1;return g|0}else if((Da|0)==59){if((Ga|0)>=8){Ha=Ea;break}i[Ea>>0]=Fa<<Ga;Ha=Ea+1|0;break}}else Ha=l;while(0);dy(d);g=Ha-c|0;return g|0}function Et(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0.0,v=0,w=0,x=0,y=0.0,z=0,A=0.0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0;f=c+d|0;d=cy(e<<2)|0;if(!d){tk(782858);g=-1;return g|0}h=i[a>>0]|0;i[c>>0]=h;j=c+1|0;do if((b|0)>0){l=h&255;m=j;n=0;o=i[a>>0]|0;p=8;q=e;a:while(1){r=b-n|0;q=(r|0)<(e|0)?r:q;r=(q|0)>0;if(r){s=0;t=o;u=0.0;while(1){v=i[a+(s+n)>>0]|0;w=v-t<<24;x=w>>23^w>>31;k[d+(s<<2)>>2]=x;y=u+ +(x>>>0);s=s+1|0;if((s|0)==(q|0)){z=v;A=y;break}else{t=v;u=y}}}else{z=o;A=0.0}u=(A-+((q|0)/2|0|0)+-1.0)/+(q|0);t=0;s=u<0.0?0:~~u&255;while(1){s=(s&255)>>>1;v=t+1|0;if(!(s<<24>>24)){B=v;C=t;break}else t=v}do if((C|0)>5){if((p+3|0)>32){t=3-p|0;s=k[121868+(p<<2)>>2]&7>>>t|l<<p;i[m>>0]=s;D=t;E=m+1|0;F=s;G=8}else{D=3;E=m;F=l;G=p}s=k[121868+(D<<2)>>2]&7|F<<D;t=G-D|0;if((t|0)<1){v=D-G|0;x=v&-8;w=v>>>3;v=E;H=t;while(1){i[v>>0]=s>>0-H;H=H+8|0;if((H|0)>=1)break;else v=v+1|0}I=E+(w+1)|0;J=8-D+G+x|0}else{I=E;J=t}if(r){v=s;H=I;K=0;L=J;while(1){M=k[d+(K<<2)>>2]|0;if((L+8|0)>32){N=8-L|0;O=k[121868+(L<<2)>>2]&M>>N|v<<L;i[H>>0]=O;P=N;Q=H+1|0;R=O;S=8}else{P=8;Q=H;R=v;S=L}O=k[121868+(P<<2)>>2]&M|R<<P;M=S-P|0;if((M|0)<1){N=P-S|0;T=N&-8;U=N>>>3;N=Q;V=M;while(1){i[N>>0]=O>>0-V;V=V+8|0;if((V|0)>=1)break;else N=N+1|0}W=Q+(U+1)|0;X=8-P+S+T|0}else{W=Q;X=M}K=K+1|0;if((K|0)==(q|0)){Y=W;Z=O;_=X;break}else{v=O;H=W;L=X}}}else{Y=I;Z=s;_=J}}else{L=(p+3|0)>32;if(A==0.0&(C|0)==0){if(L){H=l<<p;i[m>>0]=H;$=3-p|0;aa=m+1|0;ba=H;ca=8}else{$=3;aa=m;ba=l;ca=p}H=ba<<$;v=ca-$|0;if((v|0)>=1){Y=aa;Z=H;_=v;break}K=$-ca|0;t=K&-8;x=K>>>3;K=aa;w=v;while(1){i[K>>0]=H>>0-w;w=w+8|0;if((w|0)>=1)break;else K=K+1|0}Y=aa+(x+1)|0;Z=H;_=8-$+ca+t|0;break}if(L){K=3-p|0;w=k[121868+(p<<2)>>2]&B>>K|l<<p;i[m>>0]=w;da=K;ea=m+1|0;fa=w;ga=8}else{da=3;ea=m;fa=l;ga=p}w=k[121868+(da<<2)>>2]&B|fa<<da;K=ga-da|0;if((K|0)<1){s=da-ga|0;v=s&-8;N=s>>>3;s=ea;V=K;while(1){i[s>>0]=w>>0-V;V=V+8|0;if((V|0)>=1)break;else s=s+1|0}ha=ea+(N+1)|0;ia=8-da+ga+v|0}else{ha=ea;ia=K}s=(1<<C)+-1|0;if(r){V=(C|0)>0;L=8-C|0;t=ha;H=0;x=w;ja=ia;while(1){ka=k[d+(H<<2)>>2]|0;la=ka>>C;ma=la+1|0;if((ja|0)<(ma|0)){i[t>>0]=x<<ja;na=la-ja|0;oa=t+1|0;if((na|0)>7){pa=la+-8-ja|0;la=pa>>>3;sy(oa|0,0,la+1|0)|0;qa=t+(la+2)|0;ra=pa-(la<<3)|0}else{qa=oa;ra=na}sa=qa;ta=1;ua=7-ra|0}else{sa=t;ta=x<<ma|1;ua=ja-ma|0}if(V){ma=ta<<C|ka&s;ka=ua-C|0;if((ka|0)<1){na=(C-ua|0)>>>3;oa=na<<3;la=sa;pa=ka;while(1){i[la>>0]=ma>>0-pa;pa=pa+8|0;if((pa|0)>=1)break;else la=la+1|0}va=sa+(na+1)|0;wa=ma;xa=L+ua+oa|0}else{va=sa;wa=ma;xa=ka}}else{va=sa;wa=ta;xa=ua}H=H+1|0;if((H|0)==(q|0)){ya=va;za=wa;Aa=xa;break}else{t=va;x=wa;ja=xa}}}else{ya=ha;za=w;Aa=ia}if(ya>>>0>f>>>0){Ba=51;break a}else{Y=ya;Z=za;_=Aa}}while(0);n=n+e|0;if((n|0)>=(b|0)){Ca=Y;Da=Z;Ea=_;Ba=53;break}else{l=Z;m=Y;o=z;p=_}}if((Ba|0)==51){tk(782890);dy(d);g=-1;return g|0}else if((Ba|0)==53){if((Ea|0)>=8){Fa=Ca;break}i[Ca>>0]=Da<<Ea;Fa=Ca+1|0;break}}else Fa=j;while(0);dy(d);g=Fa-c|0;return g|0}function Ft(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0;f=a+b|0;b=a+5|0;g=l[a+4>>0]|0;h=0;j=(l[a+1>>0]|0)<<16|(l[a>>0]|0)<<24|(l[a+2>>0]|0)<<8|(l[a+3>>0]|0);a=8;while(1){if((h|0)>=(d|0)){m=b;break}n=a+-5|0;if((a|0)<5){o=(((n|0)>-8?a+7|0:4)-a|0)>>>3;p=o<<3;q=b;r=g;s=n;while(1){t=l[q>>0]|0|r<<8;if((s|0)<-8){q=q+1|0;r=t;s=s+8|0}else{u=t;break}}v=b+(o+1)|0;w=u;x=a+3+p|0}else{v=b;w=g;x=n}s=w>>>x;r=s+-1|0;q=(1<<x)+-1|0;t=q&w;y=h+e|0;z=(y|0)>(d|0)?d:y;a:do if((r|0)<0)if((h|0)<(z|0)){A=(y|0)>(d|0)?d:y;B=h;do{k[c+(B<<2)>>2]=j;B=B+1|0}while((B|0)!=(A|0));C=v;D=t;E=A;F=j;G=x}else{C=v;D=t;E=h;F=j;G=x}else{A=(h|0)<(z|0);if((r|0)==25){if(!A){C=v;D=t;E=h;F=j;G=x;break}B=32-x|0;H=B+-8|0;I=(H|0)>-1;J=(x|0)>0;K=24-x|0;L=K>>>3;M=K-(L<<3)|0;K=(y|0)>(d|0)?d:y;N=L+1|0;L=v;O=t;P=h;Q=j;while(1){R=O<<B;if(I){S=L;T=R;U=H;while(1){V=(l[S>>0]|0)<<U|T;U=U+-8|0;if((U|0)<=-1){W=V;break}else{S=S+1|0;T=V}}X=L+N|0;Y=W;Z=M}else{X=L;Y=R;Z=B}if(J){T=l[X>>0]|0;_=X+1|0;$=T&q;aa=T>>>(8-Z|0)|Y}else{_=X;$=0;aa=Y}T=(aa<<31>>31^aa>>>1)+Q|0;k[c+(P<<2)>>2]=T;P=P+1|0;if((P|0)==(K|0)){C=_;D=$;E=K;F=T;G=x;break a}else{L=_;O=$;Q=T}}}if(A){Q=8-s|0;O=s+7|0;L=(y|0)>(d|0)?d:y;K=v;P=t;J=h;B=j;M=x;while(1){if(!P){N=K;H=M;while(1){I=H+8|0;T=N+1|0;S=i[N>>0]|0;if(!(S<<24>>24)){N=T;H=I}else{ba=I;ca=T;da=S;break}}ea=ca;fa=da&255;ga=ba}else{ea=K;fa=P;ga=M}H=k[122e3+(fa<<2)>>2]|0;N=ga-H|0;R=ga+-1-N|0;S=1<<R^fa;T=R-r|0;if((T|0)<0){R=H-s|0;I=(O+((R|0)>-8?R:-8)-H|0)>>>3;R=I<<3;U=ea;V=S;ha=T;while(1){ia=l[U>>0]|0|V<<8;if((ha|0)<-8){U=U+1|0;V=ia;ha=ha+8|0}else{ja=ia;break}}ka=ea+(I+1)|0;la=ja;ma=Q+H+R|0}else{ka=ea;la=S;ma=T}ha=la>>>ma|N<<r;V=(1<<ma)+-1&la;U=(ha<<31>>31^ha>>>1)+B|0;k[c+(J<<2)>>2]=U;J=J+1|0;if((J|0)==(L|0)){C=ka;D=V;E=L;F=U;G=ma;break}else{K=ka;P=V;B=U;M=ma}}}else{C=v;D=t;E=h;F=j;G=x}}while(0);if(C>>>0>f>>>0){na=31;break}else{b=C;g=D;h=E;j=F;a=G}}if((na|0)==31){tk(782917);oa=1;return oa|0}if(m>>>0>=f>>>0){oa=0;return oa|0}tk(782972);oa=0;return oa|0}
function wv(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0;a:do if(b>>>0>36){f=rv()|0;k[f>>2]=22;g=0;h=0}else{f=a+4|0;j=a+100|0;do{m=k[f>>2]|0;if(m>>>0<(k[j>>2]|0)>>>0){k[f>>2]=m+1;n=l[m>>0]|0}else n=yv(a)|0}while((kv(n)|0)!=0);o=n;b:do switch(o|0){case 43:case 45:{m=((o|0)==45)<<31>>31;p=k[f>>2]|0;if(p>>>0<(k[j>>2]|0)>>>0){k[f>>2]=p+1;q=l[p>>0]|0;r=m;break b}else{q=yv(a)|0;r=m;break b}break}default:{q=o;r=0}}while(0);m=(b|0)==0;do if((b&-17|0)==0&(q|0)==48){p=k[f>>2]|0;if(p>>>0<(k[j>>2]|0)>>>0){k[f>>2]=p+1;s=l[p>>0]|0}else s=yv(a)|0;if((s|32|0)!=120)if(m){t=8;u=s;v=46;break}else{w=b;x=s;v=32;break}p=k[f>>2]|0;if(p>>>0<(k[j>>2]|0)>>>0){k[f>>2]=p+1;y=l[p>>0]|0}else y=yv(a)|0;if((l[792664+(y+1)>>0]|0)>15){p=(k[j>>2]|0)==0;if(!p)k[f>>2]=(k[f>>2]|0)+-1;if(!c){xv(a,0);g=0;h=0;break a}if(p){g=0;h=0;break a}k[f>>2]=(k[f>>2]|0)+-1;g=0;h=0;break a}else{t=16;u=y;v=46}}else{p=m?10:b;if((l[792664+(q+1)>>0]|0)>>>0<p>>>0){w=p;x=q;v=32}else{if(k[j>>2]|0)k[f>>2]=(k[f>>2]|0)+-1;xv(a,0);p=rv()|0;k[p>>2]=22;g=0;h=0;break a}}while(0);if((v|0)==32)if((w|0)==10){m=x+-48|0;if(m>>>0<10){p=m;m=0;while(1){z=(m*10|0)+p|0;A=k[f>>2]|0;if(A>>>0<(k[j>>2]|0)>>>0){k[f>>2]=A+1;B=l[A>>0]|0}else B=yv(a)|0;p=B+-48|0;if(!(p>>>0<10&z>>>0<429496729)){C=z;D=B;break}else m=z}E=C;F=0;G=D}else{E=0;F=0;G=x}m=G+-48|0;if(m>>>0<10){p=E;z=F;A=m;m=G;while(1){H=yy(p|0,z|0,10,0)|0;I=L;J=((A|0)<0)<<31>>31;K=~J;if(I>>>0>K>>>0|(I|0)==(K|0)&H>>>0>~A>>>0){M=A;N=p;O=z;P=m;break}K=ky(H|0,I|0,A|0,J|0)|0;J=L;I=k[f>>2]|0;if(I>>>0<(k[j>>2]|0)>>>0){k[f>>2]=I+1;Q=l[I>>0]|0}else Q=yv(a)|0;I=Q+-48|0;if(I>>>0<10&(J>>>0<429496729|(J|0)==429496729&K>>>0<2576980378)){p=K;z=J;A=I;m=Q}else{M=I;N=K;O=J;P=Q;break}}if(M>>>0>9){R=O;S=N;T=r}else{U=10;V=N;W=O;X=P;v=72}}else{R=F;S=E;T=r}}else{t=w;u=x;v=46}c:do if((v|0)==46){if(!(t+-1&t)){m=i[792921+((t*23|0)>>>5&7)>>0]|0;A=i[792664+(u+1)>>0]|0;z=A&255;if(z>>>0<t>>>0){p=z;z=0;while(1){J=p|z<<m;K=k[f>>2]|0;if(K>>>0<(k[j>>2]|0)>>>0){k[f>>2]=K+1;Y=l[K>>0]|0}else Y=yv(a)|0;K=i[792664+(Y+1)>>0]|0;p=K&255;if(!(J>>>0<134217728&p>>>0<t>>>0)){Z=J;_=K;$=Y;break}else z=J}aa=_;ba=0;ca=Z;da=$}else{aa=A;ba=0;ca=0;da=u}z=ny(-1,-1,m|0)|0;p=L;if((aa&255)>>>0>=t>>>0|(ba>>>0>p>>>0|(ba|0)==(p|0)&ca>>>0>z>>>0)){U=t;V=ca;W=ba;X=da;v=72;break}else{ea=ca;fa=ba;ga=aa}while(1){J=ty(ea|0,fa|0,m|0)|0;K=L;I=ga&255|J;J=k[f>>2]|0;if(J>>>0<(k[j>>2]|0)>>>0){k[f>>2]=J+1;ia=l[J>>0]|0}else ia=yv(a)|0;ga=i[792664+(ia+1)>>0]|0;if((ga&255)>>>0>=t>>>0|(K>>>0>p>>>0|(K|0)==(p|0)&I>>>0>z>>>0)){U=t;V=I;W=K;X=ia;v=72;break c}else{ea=I;fa=K}}}z=i[792664+(u+1)>>0]|0;p=z&255;if(p>>>0<t>>>0){m=p;p=0;while(1){A=m+(ha(p,t)|0)|0;K=k[f>>2]|0;if(K>>>0<(k[j>>2]|0)>>>0){k[f>>2]=K+1;ja=l[K>>0]|0}else ja=yv(a)|0;K=i[792664+(ja+1)>>0]|0;m=K&255;if(!(A>>>0<119304647&m>>>0<t>>>0)){ka=A;la=K;ma=ja;break}else p=A}na=la;oa=ka;pa=0;qa=ma}else{na=z;oa=0;pa=0;qa=u}if((na&255)>>>0<t>>>0){p=zy(-1,-1,t|0,0)|0;m=L;A=pa;K=oa;I=na;J=qa;while(1){if(A>>>0>m>>>0|(A|0)==(m|0)&K>>>0>p>>>0){U=t;V=K;W=A;X=J;v=72;break c}H=yy(K|0,A|0,t|0,0)|0;ra=L;sa=I&255;if(ra>>>0>4294967295|(ra|0)==-1&H>>>0>~sa>>>0){U=t;V=K;W=A;X=J;v=72;break c}ta=ky(sa|0,0,H|0,ra|0)|0;ra=L;H=k[f>>2]|0;if(H>>>0<(k[j>>2]|0)>>>0){k[f>>2]=H+1;ua=l[H>>0]|0}else ua=yv(a)|0;I=i[792664+(ua+1)>>0]|0;if((I&255)>>>0>=t>>>0){U=t;V=ta;W=ra;X=ua;v=72;break}else{A=ra;K=ta;J=ua}}}else{U=t;V=oa;W=pa;X=qa;v=72}}while(0);if((v|0)==72)if((l[792664+(X+1)>>0]|0)>>>0<U>>>0){do{J=k[f>>2]|0;if(J>>>0<(k[j>>2]|0)>>>0){k[f>>2]=J+1;va=l[J>>0]|0}else va=yv(a)|0}while((l[792664+(va+1)>>0]|0)>>>0<U>>>0);J=rv()|0;k[J>>2]=34;R=e;S=d;T=(d&1|0)==0&0==0?r:0}else{R=W;S=V;T=r}if(k[j>>2]|0)k[f>>2]=(k[f>>2]|0)+-1;if(!(R>>>0<e>>>0|(R|0)==(e|0)&S>>>0<d>>>0)){if(!((d&1|0)!=0|0!=0|(T|0)!=0)){J=rv()|0;k[J>>2]=34;J=ky(d|0,e|0,-1,-1)|0;g=L;h=J;break}if(R>>>0>e>>>0|(R|0)==(e|0)&S>>>0>d>>>0){J=rv()|0;k[J>>2]=34;g=e;h=d;break}}J=((T|0)<0)<<31>>31;K=jy(S^T|0,R^J|0,T|0,J|0)|0;g=L;h=K}while(0);L=g;return h|0}function xv(a,b){a=a|0;b=b|0;var c=0,d=0,e=0;k[a+104>>2]=b;c=k[a+4>>2]|0;d=k[a+8>>2]|0;e=d-c|0;k[a+108>>2]=e;if((b|0)!=0&(e|0)>(b|0))k[a+100>>2]=c+b;else k[a+100>>2]=d;return}function yv(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0;b=a+104|0;c=k[b>>2]|0;if((c|0)!=0?(k[a+108>>2]|0)>=(c|0):0)d=4;else{c=vw(a)|0;if((c|0)>=0){e=k[b>>2]|0;b=a+8|0;if(e){f=k[b>>2]|0;g=k[a+4>>2]|0;h=f;j=e-(k[a+108>>2]|0)+-1|0;if((h-g|0)>(j|0)){k[a+100>>2]=g+j;m=f}else{n=h;o=f;d=9}}else{f=k[b>>2]|0;n=f;o=f;d=9}if((d|0)==9){k[a+100>>2]=n;m=o}o=k[a+4>>2]|0;if(m){n=a+108|0;k[n>>2]=m+1-o+(k[n>>2]|0)}n=o+-1|0;if((l[n>>0]|0|0)==(c|0))p=c;else{i[n>>0]=c;p=c}}else d=4}if((d|0)==4){k[a+100>>2]=0;p=-1}return p|0}function zv(a){a=a|0;var b=0,c=0;if(a>>>0>4294963200){b=rv()|0;k[b>>2]=0-a;c=-1}else c=a;return c|0}function Av(){return 124968}function Bv(a){a=+a;return +(+ea(+(a+-1416.0996898839683))*2247116418577894884661631.0e283*2247116418577894884661631.0e283)}function Cv(a,b){a=+a;b=+b;var c=0,d=0,e=0;p[t>>3]=a;c=k[t>>2]|0;d=k[t+4>>2]|0;p[t>>3]=b;e=k[t+4>>2]&-2147483648|d&2147483647;k[t>>2]=c;k[t+4>>2]=e;return +(+p[t>>3])}function Dv(a,b){a=+a;b=+b;return +(+Cv(a,b))}function Ev(a){a=+a;var b=0,c=0.0,d=0.0;p[t>>3]=a;b=k[t+4>>2]&2147483647;k[t>>2]=k[t>>2];k[t+4>>2]=b;a=+p[t>>3];do if(b>>>0<1072049730)if(b>>>0<1045430272)c=1.0;else{d=+Fv(a);c=d*d/((d+1.0)*2.0)+1.0}else if(b>>>0<1082535490){d=+ea(+a);c=(d+1.0/d)*.5;break}else{c=+Bv(a);break}while(0);return +c}function Fv(a){a=+a;var b=0,c=0,d=0,e=0,f=0,g=0.0,h=0,i=0.0,j=0.0,l=0,m=0.0,n=0,o=0.0,q=0.0,r=0.0;p[t>>3]=a;b=k[t>>2]|0;c=k[t+4>>2]|0;d=c&2147483647;e=ny(b|0,c|0,63)|0;do if(d>>>0>1078159481){f=c&2147483647;if(!(f>>>0>2146435072|(f|0)==2146435072&b>>>0>0))if(!e)if(a>709.782712893384)g=a*8988465674311579538646525.0e283;else h=11;else g=-1.0;else g=a}else{if(d>>>0<=1071001154)if(d>>>0<1016070144){g=a;break}else{i=a;j=0.0;l=0;h=14;break}if(d>>>0<1072734898)if(!e){m=a+-.6931471803691238;n=1;o=1.9082149292705877e-10;h=12;break}else{m=a+.6931471803691238;n=-1;o=-1.9082149292705877e-10;h=12;break}else h=11}while(0);if((h|0)==11){d=~~(a*1.4426950408889634+((e|0)!=0?-.5:.5));q=+(d|0);m=a-q*.6931471803691238;n=d;o=q*1.9082149292705877e-10;h=12}if((h|0)==12){q=m-o;i=q;j=m-q-o;l=n;h=14}a:do if((h|0)==14){o=i*.5;q=i*o;m=q*(q*(q*(q*(4.008217827329362e-06-q*2.0109921818362437e-07)+-7.93650757867488e-05)+1.5873015872548146e-03)+-.03333333333333313)+1.0;a=3.0-o*m;o=q*((m-a)/(6.0-i*a));if(!l){g=i-(i*o-q);break}a=i*(o-j)-j-q;switch(l|0){case -1:{g=(i-a)*.5+-.5;break a;break}case 1:{if(i<-.25){g=(a-(i+.5))*-2.0;break a}else{g=(i-a)*2.0+1.0;break a}break}default:{n=ty(l+1023|0,0,52)|0;d=L;k[t>>2]=n;k[t+4>>2]=d;q=+p[t>>3];if(l>>>0>56){o=i-a+1.0;g=((l|0)==1024?o*2.0*8988465674311579538646525.0e283:q*o)+-1.0;break a}d=ty(1023-l|0,0,52)|0;n=L;if((l|0)<20){k[t>>2]=d;k[t+4>>2]=n;r=1.0-+p[t>>3]+(i-a)}else{k[t>>2]=d;k[t+4>>2]=n;r=i-(+p[t>>3]+a)+1.0}g=q*r;break a}}}while(0);return +g}function Gv(a,b){a=+a;b=+b;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0.0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0.0;p[t>>3]=a;c=k[t>>2]|0;d=k[t+4>>2]|0;p[t>>3]=b;e=k[t>>2]|0;f=k[t+4>>2]|0;g=ny(c|0,d|0,52)|0;h=g&2047;g=ny(e|0,f|0,52)|0;i=g&2047;g=d&-2147483648;j=ty(e|0,f|0,1)|0;l=L;a:do if(!((j|0)==0&(l|0)==0)?(m=f&2147483647,!(m>>>0>2146435072|(m|0)==2146435072&e>>>0>0|(h|0)==2047)):0){m=ty(c|0,d|0,1)|0;n=L;if(!(n>>>0>l>>>0|(n|0)==(l|0)&m>>>0>j>>>0))return +((m|0)==(j|0)&(n|0)==(l|0)?a*0.0:a);if(!h){n=ty(c|0,d|0,12)|0;m=L;if((m|0)>-1|(m|0)==-1&n>>>0>4294967295){o=n;n=m;m=0;while(1){q=m+-1|0;o=ty(o|0,n|0,1)|0;n=L;if(!((n|0)>-1|(n|0)==-1&o>>>0>4294967295)){r=q;break}else m=q}}else r=0;m=ty(c|0,d|0,1-r|0)|0;s=m;u=L;v=r}else{s=c;u=d&1048575|1048576;v=h}if(!i){m=ty(e|0,f|0,12)|0;o=L;if((o|0)>-1|(o|0)==-1&m>>>0>4294967295){n=m;m=o;o=0;while(1){q=o+-1|0;n=ty(n|0,m|0,1)|0;m=L;if(!((m|0)>-1|(m|0)==-1&n>>>0>4294967295)){w=q;break}else o=q}}else w=0;o=ty(e|0,f|0,1-w|0)|0;x=o;y=L;z=w}else{x=e;y=f&1048575|1048576;z=i}o=jy(s|0,u|0,x|0,y|0)|0;n=L;m=(n|0)>-1|(n|0)==-1&o>>>0>4294967295;b:do if((v|0)>(z|0)){q=m;A=o;B=n;C=s;D=u;E=v;while(1){if(q)if((C|0)==(x|0)&(D|0)==(y|0))break;else{F=A;G=B}else{F=C;G=D}H=ty(F|0,G|0,1)|0;I=L;J=E+-1|0;K=jy(H|0,I|0,x|0,y|0)|0;M=L;N=(M|0)>-1|(M|0)==-1&K>>>0>4294967295;if((J|0)>(z|0)){q=N;A=K;B=M;C=H;D=I;E=J}else{O=N;P=H;Q=I;R=K;S=M;T=J;break b}}U=a*0.0;break a}else{O=m;P=s;Q=u;R=o;S=n;T=v}while(0);if(O)if((P|0)==(x|0)&(Q|0)==(y|0)){U=a*0.0;break}else{V=S;W=R}else{V=Q;W=P}if(V>>>0<1048576|(V|0)==1048576&W>>>0<0){n=W;o=V;m=T;while(1){E=ty(n|0,o|0,1)|0;D=L;C=m+-1|0;if(D>>>0<1048576|(D|0)==1048576&E>>>0<0){n=E;o=D;m=C}else{X=E;Y=D;Z=C;break}}}else{X=W;Y=V;Z=T}if((Z|0)>0){m=ky(X|0,Y|0,0,-1048576)|0;o=L;n=ty(Z|0,0,52)|0;_=o|L;$=m|n}else{n=ny(X|0,Y|0,1-Z|0)|0;_=L;$=n}k[t>>2]=$;k[t+4>>2]=_|g;U=+p[t>>3]}else aa=3;while(0);if((aa|0)==3){ba=a*b;U=ba/ba}return +U}function Hv(a,b){a=+a;b=+b;return +(+Gv(a,b))}function Iv(a,b){a=+a;b=b|0;var c=0,d=0,e=0,f=0,g=0.0,h=0.0,i=0,j=0.0;p[t>>3]=a;c=k[t>>2]|0;d=k[t+4>>2]|0;e=ny(c|0,d|0,52)|0;f=e&2047;switch(f|0){case 0:{if(a!=0.0){g=+Iv(a*18446744073709551616.0,b);h=g;i=(k[b>>2]|0)+-64|0}else{h=a;i=0}k[b>>2]=i;j=h;break}case 2047:{j=a;break}default:{k[b>>2]=f+-1022;k[t>>2]=c;k[t+4>>2]=d&-2146435073|1071644672;j=+p[t>>3]}}return +j}function Jv(a,b){a=+a;b=b|0;return +(+Iv(a,b))}function Kv(a){a=+a;var b=0,c=0,d=0,e=0.0,f=0,g=0,h=0,i=0,j=0,l=0,m=0.0,n=0.0,o=0.0,q=0.0,r=0.0,s=0.0;p[t>>3]=a;b=k[t>>2]|0;c=k[t+4>>2]|0;d=(c|0)<0;do if(d|c>>>0<1048576){if((b|0)==0&(c&2147483647|0)==0){e=-1.0/(a*a);break}if(d){e=(a-a)/0.0;break}else{p[t>>3]=a*18014398509481984.0;f=k[t+4>>2]|0;g=k[t>>2]|0;h=f;i=f;j=-1077;l=9;break}}else if(c>>>0<=2146435071)if((b|0)==0&0==0&(c|0)==1072693248)e=0.0;else{g=b;h=c;i=c;j=-1023;l=9}else e=a;while(0);if((l|0)==9){l=i+614242|0;k[t>>2]=g;k[t+4>>2]=(l&1048575)+1072079006;a=+p[t>>3]+-1.0;m=a*(a*.5);n=a/(a+2.0);o=n*n;q=o*o;p[t>>3]=a-m;g=k[t+4>>2]|0;k[t>>2]=0;k[t+4>>2]=g;r=+p[t>>3];s=n*(m+(q*(q*(q*.15313837699209373+.22222198432149784)+.3999999999940942)+o*(q*(q*(q*.14798198605116586+.1818357216161805)+.2857142874366239)+.6666666666666735)))+(a-r-m);m=r*.4342944818781689;a=+(j+(l>>>20)|0);q=a*.30102999566361177;o=q+m;e=o+(m+(q-o)+(s*.4342944818781689+(a*3.694239077158931e-13+(r+s)*2.5082946711645275e-11)))}return +e}function Lv(a,b){a=+a;b=b|0;var c=0.0,d=0,e=0,f=0,g=0.0;if((b|0)>1023){c=a*8988465674311579538646525.0e283;d=b+-1023|0;if((d|0)>1023){e=b+-2046|0;f=(e|0)>1023?1023:e;g=c*8988465674311579538646525.0e283}else{f=d;g=c}}else if((b|0)<-1022){c=a*2.2250738585072014e-308;d=b+1022|0;if((d|0)<-1022){e=b+2044|0;f=(e|0)<-1022?-1022:e;g=c*2.2250738585072014e-308}else{f=d;g=c}}else{f=b;g=a}b=ty(f+1023|0,0,52)|0;f=L;k[t>>2]=b;k[t+4>>2]=f;return +(g*+p[t>>3])}function Mv(a,b){a=+a;b=b|0;return +(+Lv(a,b))}function Nv(a){a=+a;var b=0,c=0.0,d=0,e=0.0,f=0.0,g=0.0;p[t>>3]=a;b=k[t+4>>2]|0;c=(b|0)<0?-.5:.5;d=b&2147483647;k[t>>2]=k[t>>2];k[t+4>>2]=d;e=+p[t>>3];do if(d>>>0<1082535490){f=+Fv(e);if(d>>>0>=1072693248){g=c*(f+f/(f+1.0));break}if(d>>>0<1045430272)g=a;else g=c*(f*2.0-f*f/(f+1.0))}else g=c*2.0*+Bv(e);while(0);return +g}function Ov(a){a=+a;var b=0,c=0,d=0.0,e=0.0;p[t>>3]=a;b=k[t+4>>2]|0;c=b&2147483647;k[t>>2]=k[t>>2];k[t+4>>2]=c;a=+p[t>>3];do if(c>>>0>1071748074)if(c>>>0>1077149696){d=1.0-0.0/a;break}else{d=1.0-2.0/(+Fv(a*2.0)+2.0);break}else{if(c>>>0>1070618798){e=+Fv(a*2.0);d=e/(e+2.0);break}if(c>>>0>1048575){e=+Fv(a*-2.0);d=-e/(e+2.0)}else d=a}while(0);return +((b|0)<0?-d:d)}function Pv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0;d=r;r=r+16|0;e=d+4|0;f=d;g=k[31172]|0;if((g|0)==0|(k[31174]|0)!=0){k[31174]=0;k[31175]=0;k[31172]=1;h=1}else h=g;a:do if(((h|0)<(a|0)?(g=k[b+(h<<2)>>2]|0,(g|0)!=0):0)?(i[g>>0]|0)==45:0){switch(i[g+1>>0]|0){case 0:{j=-1;break a;break}case 45:{if(!(i[g+2>>0]|0)){k[31172]=h+1;j=-1;break a}break}default:{}}l=k[31175]|0;if(!l){k[31175]=1;m=1}else m=l;l=Tv(e,g+m|0,4)|0;if((l|0)<0){k[e>>2]=65533;n=65533;o=1}else{n=k[e>>2]|0;o=l}l=k[31172]|0;g=k[b+(l<<2)>>2]|0;p=k[31175]|0;q=g+p|0;k[31176]=n;s=p+o|0;k[31175]=s;if(!(i[g+s>>0]|0)){k[31172]=l+1;k[31175]=0}l=Tv(f,c,4)|0;b:do if(!l)t=0;else{s=l;g=0;while(1){if((k[f>>2]|0)==(k[e>>2]|0)){t=g;break b}p=((s|0)<1?1:s)+g|0;s=Tv(f,c+p|0,4)|0;if(!s){t=p;break}else g=p}}while(0);l=k[f>>2]|0;if((l|0)!=(k[e>>2]|0)){if(!((i[c>>0]|0)!=58&(k[31173]|0)!=0)){j=63;break}g=k[b>>2]|0;Ox(2,g,xx(g)|0)|0;Ox(2,792930,18)|0;Ox(2,q,o)|0;Ox(2,792949,1)|0;j=63;break}if((i[c+(t+1)>>0]|0)==58){g=k[31172]|0;if((g|0)<(a|0)){k[31172]=g+1;k[31177]=(k[b+(g<<2)>>2]|0)+(k[31175]|0);k[31175]=0;j=l;break}if((i[c>>0]|0)!=58)if(!(k[31173]|0))j=63;else{g=k[b>>2]|0;Ox(2,g,xx(g)|0)|0;Ox(2,792951,31)|0;Ox(2,q,o)|0;Ox(2,792949,1)|0;j=63}else j=58}else j=l}else j=-1;while(0);r=d;return j|0}function Qv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=r;r=r+32|0;e=d;f=d+16|0;k[f>>2]=c;c=(k[f>>2]|0)+(4-1)&~(4-1);g=k[c>>2]|0;k[f>>2]=c+4;k[e>>2]=a;k[e+4>>2]=b;k[e+8>>2]=g;g=zv(Fa(54,e|0)|0)|0;r=d;return g|0}function Rv(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;e=r;r=r+16|0;f=e;g=(d|0)==0?125024:d;d=k[g>>2]|0;a:do if(!b)if(!d)h=0;else j=15;else{l=(a|0)==0?f:a;if(!c)h=-2;else{if(!d){m=i[b>>0]|0;n=m&255;if(m<<24>>24>-1){k[l>>2]=n;h=m<<24>>24!=0&1;break}m=n+-194|0;if(m>>>0>50){j=15;break}n=k[124712+(m<<2)>>2]|0;m=c+-1|0;if(!m)o=n;else{p=m;q=n;s=b+1|0;j=9}}else{p=c;q=d;s=b;j=9}b:do if((j|0)==9){n=i[s>>0]|0;m=(n&255)>>>3;if((m+-16|m+(q>>26))>>>0>7){j=15;break a}else{t=p;u=n;v=q;w=s}while(1){w=w+1|0;v=(u&255)+-128|v<<6;t=t+-1|0;if((v|0)>=0){x=v;y=t;break}if(!t){o=v;break b}u=i[w>>0]|0;if((u&-64)<<24>>24!=-128){j=15;break a}}k[g>>2]=0;k[l>>2]=x;h=c-y|0;break a}while(0);k[g>>2]=o;h=-2}}while(0);if((j|0)==15){k[g>>2]=0;g=rv()|0;k[g>>2]=84;h=-1}r=e;return h|0}function Sv(a){a=a|0;var b=0;if(!a)b=1;else b=(k[a>>2]|0)==0;return b&1|0}function Tv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,m=0;d=r;r=r+16|0;e=d;a:do if(!b)f=0;else{do if(c){g=(a|0)==0?e:a;h=i[b>>0]|0;j=h&255;if(h<<24>>24>-1){k[g>>2]=j;f=h<<24>>24!=0&1;break a}h=j+-194|0;if(h>>>0<=50){j=b+1|0;m=k[124712+(h<<2)>>2]|0;if(c>>>0<4?(m&-2147483648>>>((c*6|0)+-6|0)|0)!=0:0)break;h=l[j>>0]|0;j=h>>>3;if((j+-16|j+(m>>26))>>>0<=7){j=h+-128|m<<6;if((j|0)>=0){k[g>>2]=j;f=2;break a}m=l[b+2>>0]|0;if((m&192|0)==128){h=m+-128|j<<6;if((h|0)>=0){k[g>>2]=h;f=3;break a}j=l[b+3>>0]|0;if((j&192|0)==128){k[g>>2]=j+-128|h<<6;f=4;break a}}}}}while(0);h=rv()|0;k[h>>2]=84;f=-1}while(0);r=d;return f|0}function Uv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;do if(a){if(b>>>0<128){i[a>>0]=b;d=1;break}if(b>>>0<2048){i[a>>0]=b>>>6|192;i[a+1>>0]=b&63|128;d=2;break}if(b>>>0<55296|(b&-8192|0)==57344){i[a>>0]=b>>>12|224;i[a+1>>0]=b>>>6&63|128;i[a+2>>0]=b&63|128;d=3;break}if((b+-65536|0)>>>0<1048576){i[a>>0]=b>>>18|240;i[a+1>>0]=b>>>12&63|128;i[a+2>>0]=b>>>6&63|128;i[a+3>>0]=b&63|128;d=4;break}else{c=rv()|0;k[c>>2]=84;d=-1;break}}else d=1;while(0);return d|0}function Vv(a,b){a=a|0;b=b|0;var c=0;if(!a)c=0;else c=Uv(a,b,0)|0;return c|0}function Wv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;d=r;r=r+32|0;e=d;f=d+8|0;k[f>>2]=a;k[f+4>>2]=b;k[f+8>>2]=c;k[f+12>>2]=0;k[f+16>>2]=0;k[f+20>>2]=0;k[e>>2]=3;k[e+4>>2]=f;f=zv(Ca(102,e|0)|0)|0;r=d;return f|0}function Xv(a){a=a|0;return ry(a|0)|0}function Yv(a){a=a|0;var b=0;b=a&65535;return (b<<8|b>>>8)&65535|0}function Zv(a){a=a|0;return ry(a|0)|0}function _v(a){a=a|0;var b=0;b=a&65535;return (b<<8|b>>>8)&65535|0}function $v(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;return aw(a,b,c,d,0,0)|0}function aw(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0;g=r;r=r+32|0;h=g;i=g+8|0;k[i>>2]=a;k[i+4>>2]=b;k[i+8>>2]=c;k[i+12>>2]=d;k[i+16>>2]=e;k[i+20>>2]=f;k[h>>2]=12;k[h+4>>2]=i;i=zv(Ca(102,h|0)|0)|0;r=g;return i|0}function bw(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;return cw(a,b,c,d,0,0)|0}function cw(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0;g=r;r=r+32|0;h=g;i=g+8|0;k[i>>2]=a;k[i+4>>2]=b;k[i+8>>2]=c;k[i+12>>2]=d;k[i+16>>2]=e;k[i+20>>2]=f;k[h>>2]=11;k[h+4>>2]=i;i=zv(Ca(102,h|0)|0)|0;r=g;return i|0}function dw(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0;f=r;r=r+32|0;g=f;h=f+8|0;k[h>>2]=a;k[h+4>>2]=b;k[h+8>>2]=c;k[h+12>>2]=d;k[h+16>>2]=e;k[h+20>>2]=0;k[g>>2]=14;k[g+4>>2]=h;h=zv(Ca(102,g|0)|0)|0;r=f;return h|0}function ew(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0;d=r;r=r+96|0;e=d+32|0;f=d+24|0;g=d+16|0;h=d+8|0;i=d;j=d+64|0;l=d+40|0;k[j>>2]=a;k[j+4>>2]=b;k[j+8>>2]=c;k[j+12>>2]=0;k[j+16>>2]=0;k[j+20>>2]=0;k[i>>2]=1;k[i+4>>2]=j;j=zv(Ca(102,i|0)|0)|0;a:do if((j|0)<0){i=rv()|0;switch(k[i>>2]|0){case 93:case 22:break;default:{m=j;break a}}if(b&526336){k[l>>2]=a;k[l+4>>2]=b&-526337;k[l+8>>2]=c;k[l+12>>2]=0;k[l+16>>2]=0;k[l+20>>2]=0;k[h>>2]=1;k[h+4>>2]=l;i=zv(Ca(102,h|0)|0)|0;if((i|0)>=0){if(b&524288){k[g>>2]=1;tv(i,2,g)|0}if(b&2048){n=tv(i,3,f)|0|2048;k[e>>2]=n;tv(i,4,e)|0;m=i}else m=i}else m=i}else m=j}else m=j;while(0);r=d;return m|0}function fw(a){a=a|0;var b=0;b=101448;k[b>>2]=a+-1;k[b+4>>2]=0;return}function gw(){var a=0,b=0,c=0;a=101448;b=yy(k[a>>2]|0,k[a+4>>2]|0,1284865837,1481765933)|0;a=ky(b|0,L|0,1,0)|0;b=L;c=101448;k[c>>2]=a;k[c+4>>2]=b;c=ny(a|0,b|0,33)|0;return c|0}function hw(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;k[d>>2]=a;k[d+4>>2]=b;b=zv(Za(195,d|0)|0)|0;r=c;return b|0}function iw(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;c=r;r=r+112|0;d=c+40|0;e=c+24|0;f=c+16|0;g=c;h=c+52|0;j=i[b>>0]|0;if(kx(792983,j<<24>>24,4)|0){l=cy(1144)|0;if(!l)m=0;else{n=l;o=n+112|0;do{k[n>>2]=0;n=n+4|0}while((n|0)<(o|0));if(!(rx(b,43)|0))k[l>>2]=j<<24>>24==114?8:4;if(!(rx(b,101)|0))p=j;else{k[g>>2]=a;k[g+4>>2]=2;k[g+8>>2]=1;eb(221,g|0)|0;p=i[b>>0]|0}if(p<<24>>24==97){k[f>>2]=a;k[f+4>>2]=3;p=eb(221,f|0)|0;if(!(p&1024)){k[e>>2]=a;k[e+4>>2]=4;k[e+8>>2]=p|1024;eb(221,e|0)|0}e=k[l>>2]|128;k[l>>2]=e;q=e}else q=k[l>>2]|0;k[l+60>>2]=a;k[l+44>>2]=l+120;k[l+48>>2]=1024;e=l+75|0;i[e>>0]=-1;if((q&8|0)==0?(k[d>>2]=a,k[d+4>>2]=21505,k[d+8>>2]=h,(Fa(54,d|0)|0)==0):0)i[e>>0]=10;k[l+32>>2]=4;k[l+36>>2]=2;k[l+40>>2]=3;k[l+12>>2]=1;if(!(k[31162]|0))k[l+76>>2]=-1;nb(124672);e=k[31167]|0;k[l+56>>2]=e;if(e)k[e+52>>2]=l;k[31167]=l;mb(124672);m=l}}else{l=rv()|0;k[l>>2]=22;m=0}r=c;return m|0}function jw(a){a=a|0;var b=0,c=0,d=0,e=0;b=(rx(a,43)|0)==0;c=i[a>>0]|0;d=b?c<<24>>24!=114&1:2;b=(rx(a,120)|0)==0;e=b?d:d|128;d=(rx(a,101)|0)==0;a=d?e:e|524288;e=c<<24>>24==114?a:a|64;a=c<<24>>24==119?e|512:e;return (c<<24>>24==97?a|1024:a)|0}function kw(a){a=a|0;return 0}function lw(a){a=a|0;return}function mw(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0;c=r;r=r+16|0;d=c;e=b&255;i[d>>0]=e;f=a+16|0;g=k[f>>2]|0;if(!g)if(!(uw(a)|0)){h=k[f>>2]|0;j=4}else m=-1;else{h=g;j=4}do if((j|0)==4){g=a+20|0;f=k[g>>2]|0;if(f>>>0<h>>>0?(n=b&255,(n|0)!=(i[a+75>>0]|0)):0){k[g>>2]=f+1;i[f>>0]=e;m=n;break}if((Pb[k[a+36>>2]&63](a,d,1)|0)==1)m=l[d>>0]|0;else m=-1}while(0);r=c;return m|0}function nw(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;k[c>>2]=k[a+60>>2];a=zv(cb(6,c|0)|0)|0;r=b;return a|0}function ow(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0;d=r;r=r+48|0;e=d+16|0;f=d;g=d+32|0;k[g>>2]=b;h=g+4|0;j=a+48|0;l=k[j>>2]|0;k[h>>2]=c-((l|0)!=0&1);m=a+44|0;k[g+8>>2]=k[m>>2];k[g+12>>2]=l;if(!(k[31161]|0)){k[e>>2]=k[a+60>>2];k[e+4>>2]=g;k[e+8>>2]=2;n=zv(Jb(145,e|0)|0)|0}else{Ga(16,a|0);k[f>>2]=k[a+60>>2];k[f+4>>2]=g;k[f+8>>2]=2;g=zv(Jb(145,f|0)|0)|0;xa(0);n=g}if((n|0)>=1){g=k[h>>2]|0;if(n>>>0>g>>>0){h=k[m>>2]|0;m=a+4|0;k[m>>2]=h;f=h;k[a+8>>2]=f+(n-g);if(!(k[j>>2]|0))o=c;else{k[m>>2]=f+1;i[b+(c+-1)>>0]=i[f>>0]|0;o=c}}else o=n}else{k[a>>2]=k[a>>2]|n&48^16;k[a+8>>2]=0;k[a+4>>2]=0;o=n}r=d;return o|0}function pw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=r;r=r+32|0;e=d;f=d+20|0;k[e>>2]=k[a+60>>2];k[e+4>>2]=0;k[e+8>>2]=b;k[e+12>>2]=f;k[e+16>>2]=c;if((zv(Gb(140,e|0)|0)|0)<0){k[f>>2]=-1;g=-1}else g=k[f>>2]|0;r=d;return g|0}function qw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0;d=r;r=r+48|0;e=d+16|0;f=d;g=d+32|0;h=a+28|0;i=k[h>>2]|0;k[g>>2]=i;j=a+20|0;l=(k[j>>2]|0)-i|0;k[g+4>>2]=l;k[g+8>>2]=b;k[g+12>>2]=c;b=a+60|0;i=a+44|0;m=g;g=2;n=l+c|0;while(1){if(!(k[31161]|0)){k[e>>2]=k[b>>2];k[e+4>>2]=m;k[e+8>>2]=g;o=zv(Kb(146,e|0)|0)|0}else{Ga(17,a|0);k[f>>2]=k[b>>2];k[f+4>>2]=m;k[f+8>>2]=g;l=zv(Kb(146,f|0)|0)|0;xa(0);o=l}if((n|0)==(o|0)){p=6;break}if((o|0)<0){q=m;s=g;p=8;break}l=n-o|0;t=k[m+4>>2]|0;if(o>>>0<=t>>>0)if((g|0)==2){k[h>>2]=(k[h>>2]|0)+o;u=t;v=o;w=m;x=2}else{u=t;v=o;w=m;x=g}else{y=k[i>>2]|0;k[h>>2]=y;k[j>>2]=y;u=k[m+12>>2]|0;v=o-t|0;w=m+8|0;x=g+-1|0}k[w>>2]=(k[w>>2]|0)+v;k[w+4>>2]=u-v;m=w;g=x;n=l}if((p|0)==6){n=k[i>>2]|0;k[a+16>>2]=n+(k[a+48>>2]|0);i=n;k[h>>2]=i;k[j>>2]=i;z=c}else if((p|0)==8){k[a+16>>2]=0;k[h>>2]=0;k[j>>2]=0;k[a>>2]=k[a>>2]|32;if((s|0)==2)z=0;else z=c-(k[q+4>>2]|0)|0}r=d;return z|0}function rw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+80|0;e=d;k[a+36>>2]=2;if((k[a>>2]&64|0)==0?(k[e>>2]=k[a+60>>2],k[e+4>>2]=21505,k[e+8>>2]=d+12,(Fa(54,e|0)|0)!=0):0)i[a+75>>0]=-1;e=qw(a,b,c)|0;r=d;return e|0}function sw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0;d=a+84|0;e=k[d>>2]|0;f=c+256|0;g=kx(e,0,f)|0;h=(g|0)==0?f:g-e|0;g=h>>>0<c>>>0?h:c;py(b|0,e|0,g|0)|0;k[a+4>>2]=e+g;b=e+h|0;k[a+8>>2]=b;k[d>>2]=b;return g|0}function tw(a){a=a|0;var b=0,c=0,d=0;b=a+74|0;c=i[b>>0]|0;i[b>>0]=c+255|c;c=a+20|0;b=a+44|0;if((k[c>>2]|0)>>>0>(k[b>>2]|0)>>>0)Pb[k[a+36>>2]&63](a,0,0)|0;k[a+16>>2]=0;k[a+28>>2]=0;k[c>>2]=0;c=k[a>>2]|0;if(c&20)if(!(c&4))d=-1;else{k[a>>2]=c|32;d=-1}else{c=k[b>>2]|0;k[a+8>>2]=c;k[a+4>>2]=c;d=0}return d|0}function uw(a){a=a|0;var b=0,c=0,d=0;b=a+74|0;c=i[b>>0]|0;i[b>>0]=c+255|c;c=k[a>>2]|0;if(!(c&8)){k[a+8>>2]=0;k[a+4>>2]=0;b=k[a+44>>2]|0;k[a+28>>2]=b;k[a+20>>2]=b;k[a+16>>2]=b+(k[a+48>>2]|0);d=0}else{k[a>>2]=c|32;d=-1}return d|0}function vw(a){a=a|0;var b=0,c=0,d=0;b=r;r=r+16|0;c=b;if((k[a+8>>2]|0)==0?(tw(a)|0)!=0:0)d=-1;else if((Pb[k[a+32>>2]&63](a,c,1)|0)==1)d=l[c>>0]|0;else d=-1;r=b;return d|0}function ww(a){a=a|0;var b=0,c=0,d=0,e=0;b=(k[a>>2]&1|0)!=0;if(!b){nb(124672);c=k[a+52>>2]|0;d=a+56|0;if(c)k[c+56>>2]=k[d>>2];e=k[d>>2]|0;if(e)k[e+52>>2]=c;if((k[31167]|0)==(a|0))k[31167]=e;mb(124672)}e=zw(a)|0;c=Ub[k[a+12>>2]&31](a)|0|e;e=k[a+92>>2]|0;if(e)dy(e);if(!b)dy(a);return c|0}function xw(a){a=a|0;var b=0,c=0,d=0;if((k[a+76>>2]|0)>-1){b=(kw(a)|0)==0;c=(k[a>>2]|0)>>>4&1;if(b)d=c;else d=c}else d=(k[a>>2]|0)>>>4&1;return d|0}function yw(a){a=a|0;var b=0,c=0,d=0;if((k[a+76>>2]|0)>-1){b=(kw(a)|0)==0;c=(k[a>>2]|0)>>>5&1;if(b)d=c;else d=c}else d=(k[a>>2]|0)>>>5&1;return d|0}function zw(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;do if(a){if((k[a+76>>2]|0)<=-1){b=Sx(a)|0;break}c=(kw(a)|0)==0;d=Sx(a)|0;if(c)b=d;else{lw(a);b=d}}else{if(!(k[31232]|0))e=0;else e=zw(k[31232]|0)|0;nb(124672);d=k[31167]|0;if(!d)f=e;else{c=d;d=e;while(1){if((k[c+76>>2]|0)>-1)g=kw(c)|0;else g=0;if((k[c+20>>2]|0)>>>0>(k[c+28>>2]|0)>>>0)h=Sx(c)|0|d;else h=d;if(g)lw(c);c=k[c+56>>2]|0;if(!c){f=h;break}else d=h}}mb(124672);b=f}while(0);return b|0}function Aw(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;if((k[a+76>>2]|0)>=0?(kw(a)|0)!=0:0){b=a+4|0;c=k[b>>2]|0;if(c>>>0<(k[a+8>>2]|0)>>>0){k[b>>2]=c+1;d=l[c>>0]|0}else d=vw(a)|0;e=d}else f=3;do if((f|0)==3){d=a+4|0;c=k[d>>2]|0;if(c>>>0<(k[a+8>>2]|0)>>>0){k[d>>2]=c+1;e=l[c>>0]|0;break}else{e=vw(a)|0;break}}while(0);return e|0}function Bw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;if((k[c+76>>2]|0)>-1)d=kw(c)|0;else d=0;e=b+-1|0;if((b|0)<2){b=c+74|0;f=i[b>>0]|0;i[b>>0]=f+255|f;if(d)lw(c);if(!e){i[a>>0]=0;g=a}else g=0}else{a:do if(e){f=c+4|0;b=c+8|0;h=e;j=a;while(1){m=k[f>>2]|0;n=m;o=(k[b>>2]|0)-n|0;p=kx(m,10,o)|0;q=(p|0)==0;r=q?o:1-n+p|0;p=r>>>0<h>>>0;n=p?r:h;py(j|0,m|0,n|0)|0;m=k[f>>2]|0;r=m+n|0;k[f>>2]=r;o=j+n|0;s=h-n|0;if(!(q&p)){t=o;u=18;break a}if(r>>>0<(k[b>>2]|0)>>>0){p=n+1|0;k[f>>2]=m+p;v=p;w=l[r>>0]|0}else{r=vw(c)|0;if((r|0)<0){x=o;break}v=n+1|0;w=r}h=s+-1|0;s=j+v|0;i[o>>0]=w;if(!((h|0)!=0&(w&255|0)!=10)){t=s;u=18;break a}else j=s}if((x|0)!=(a|0)?(k[c>>2]&16|0)!=0:0){t=x;u=18}else y=0}else{t=a;u=18}while(0);if((u|0)==18)if(!a)y=0;else{i[t>>0]=0;y=a}if(!d)g=y;else{lw(c);g=y}}return g|0}function Cw(a){a=a|0;(k[a+76>>2]|0)>-1?(kw(a)|0)!=0:0;return k[a+60>>2]|0}function Dw(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;c=r;r=r+32|0;d=c+16|0;e=c;if(kx(792983,i[b>>0]|0,4)|0){f=jw(b)|0|32768;k[e>>2]=a;k[e+4>>2]=f;k[e+8>>2]=438;f=zv(yb(5,e|0)|0)|0;if((f|0)>=0){e=iw(f,b)|0;if(!e){k[d>>2]=f;cb(6,d|0)|0;g=0}else g=e}else g=0}else{e=rv()|0;k[e>>2]=22;g=0}r=c;return g|0}function Ew(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+16|0;e=d;k[e>>2]=c;c=_w(a,b,e)|0;r=d;return c|0}function Fw(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;if((k[b+76>>2]|0)>=0?(kw(b)|0)!=0:0){if((i[b+75>>0]|0)!=(a|0)?(c=b+20|0,d=k[c>>2]|0,d>>>0<(k[b+16>>2]|0)>>>0):0){k[c>>2]=d+1;i[d>>0]=a;e=a&255}else e=mw(b,a)|0;lw(b);f=e}else g=3;do if((g|0)==3){if((i[b+75>>0]|0)!=(a|0)?(e=b+20|0,d=k[e>>2]|0,d>>>0<(k[b+16>>2]|0)>>>0):0){k[e>>2]=d+1;i[d>>0]=a;f=a&255;break}f=mw(b,a)|0}while(0);return f|0}function Gw(a,b){a=a|0;b=b|0;return (Pw(a,xx(a)|0,1,b)|0)+-1|0}function Hw(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;e=ha(c,b)|0;if((k[d+76>>2]|0)>-1)f=kw(d)|0;else f=0;g=d+74|0;h=i[g>>0]|0;i[g>>0]=h+255|h;h=d+4|0;g=k[h>>2]|0;j=(k[d+8>>2]|0)-g|0;if((j|0)>0){l=j>>>0<e>>>0?j:e;py(a|0,g|0,l|0)|0;k[h>>2]=g+l;m=a+l|0;n=e-l|0}else{m=a;n=e}a:do if(!n)o=13;else{a=d+32|0;l=m;g=n;while(1){if(tw(d)|0){p=g;break}h=Pb[k[a>>2]&63](d,l,g)|0;if((h+1|0)>>>0<2){p=g;break}if((g|0)==(h|0)){o=13;break a}else{l=l+h|0;g=g-h|0}}if(f)lw(d);q=((e-p|0)>>>0)/(b>>>0)|0}while(0);if((o|0)==13)if(!f)q=c;else{lw(d);q=c}return q|0}function Iw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;if((c|0)==1)d=b-(k[a+8>>2]|0)+(k[a+4>>2]|0)|0;else d=b;b=a+20|0;e=a+28|0;if((k[b>>2]|0)>>>0>(k[e>>2]|0)>>>0?(Pb[k[a+36>>2]&63](a,0,0)|0,(k[b>>2]|0)==0):0)f=-1;else{k[a+16>>2]=0;k[e>>2]=0;k[b>>2]=0;if((Pb[k[a+40>>2]&63](a,d,c)|0)<0)f=-1;else{k[a+8>>2]=0;k[a+4>>2]=0;k[a>>2]=k[a>>2]&-17;f=0}}return f|0}function Jw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;if((k[a+76>>2]|0)>-1){d=(kw(a)|0)==0;e=Iw(a,b,c)|0;if(d)f=e;else{lw(a);f=e}}else f=Iw(a,b,c)|0;return f|0}function Kw(a,b,c){a=a|0;b=b|0;c=c|0;return Jw(a,b,c)|0}function Lw(a){a=a|0;var b=0,c=0,d=0;if(!(k[a>>2]&128))b=1;else b=(k[a+20>>2]|0)>>>0>(k[a+28>>2]|0)>>>0?2:1;c=Pb[k[a+40>>2]&63](a,0,b)|0;if((c|0)<0)d=c;else d=c-(k[a+8>>2]|0)+(k[a+4>>2]|0)+(k[a+20>>2]|0)-(k[a+28>>2]|0)|0;return d|0}function Mw(a){a=a|0;var b=0,c=0,d=0;if((k[a+76>>2]|0)>-1){b=(kw(a)|0)==0;c=Lw(a)|0;if(b)d=c;else d=c}else d=Lw(a)|0;return d|0}function Nw(a){a=a|0;return Mw(a)|0}function Ow(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0;d=c+16|0;e=k[d>>2]|0;if(!e)if(!(uw(c)|0)){f=k[d>>2]|0;g=4}else h=0;else{f=e;g=4}a:do if((g|0)==4){e=c+20|0;d=k[e>>2]|0;if((f-d|0)>>>0<b>>>0){h=Pb[k[c+36>>2]&63](c,a,b)|0;break}b:do if((i[c+75>>0]|0)>-1){j=b;while(1){if(!j){l=b;m=a;n=d;o=0;break b}p=j+-1|0;if((i[a+p>>0]|0)==10){q=j;break}else j=p}if((Pb[k[c+36>>2]&63](c,a,q)|0)>>>0<q>>>0){h=q;break a}l=b-q|0;m=a+q|0;n=k[e>>2]|0;o=q}else{l=b;m=a;n=d;o=0}while(0);py(n|0,m|0,l|0)|0;k[e>>2]=(k[e>>2]|0)+l;h=o+l|0}while(0);return h|0}function Pw(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=ha(c,b)|0;if((k[d+76>>2]|0)>-1){f=(kw(d)|0)==0;g=Ow(a,e,d)|0;if(f)h=g;else{lw(d);h=g}}else h=Ow(a,e,d)|0;if((h|0)==(e|0))i=c;else i=(h>>>0)/(b>>>0)|0;return i|0}function Qw(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;if((k[a+76>>2]|0)>=0?(kw(a)|0)!=0:0){b=a+4|0;c=k[b>>2]|0;if(c>>>0<(k[a+8>>2]|0)>>>0){k[b>>2]=c+1;d=l[c>>0]|0}else d=vw(a)|0;e=d}else f=3;do if((f|0)==3){d=a+4|0;c=k[d>>2]|0;if(c>>>0<(k[a+8>>2]|0)>>>0){k[d>>2]=c+1;e=l[c>>0]|0;break}else{e=vw(a)|0;break}}while(0);return e|0}function Rw(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;k[d>>2]=b;b=_w(k[31231]|0,a,d)|0;r=c;return b|0}function Sw(a){a=a|0;return Fw(a,k[31231]|0)|0}function Tw(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=k[31231]|0;if((k[b+76>>2]|0)>-1)c=kw(b)|0;else c=0;do if((Gw(a,b)|0)<0)d=1;else{if((i[b+75>>0]|0)!=10?(e=b+20|0,f=k[e>>2]|0,f>>>0<(k[b+16>>2]|0)>>>0):0){k[e>>2]=f+1;i[f>>0]=10;d=0;break}d=(mw(b,10)|0)<0}while(0);if(c)lw(b);return d<<31>>31|0}function Uw(a){a=a|0;var b=0,c=0,d=0,e=0,f=0;b=r;r=r+16|0;c=b+8|0;d=b;k[d>>2]=a;e=zv(qb(10,d|0)|0)|0;if(e){d=rv()|0;if((k[d>>2]|0)==21){k[c>>2]=a;f=zv(Ia(40,c|0)|0)|0}else f=e}else f=0;r=b;return f|0}function Vw(a){a=a|0;var b=0;if((k[a+76>>2]|0)>-1){b=(kw(a)|0)==0;Iw(a,0,0)|0;k[a>>2]=k[a>>2]&-33;if(!b)lw(a)}else{Iw(a,0,0)|0;k[a>>2]=k[a>>2]&-33}return}function Ww(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=r;r=r+16|0;f=e;k[f>>2]=d;d=ax(a,b,c,f)|0;r=e;return d|0}function Xw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+16|0;e=d;k[e>>2]=c;c=bx(a,b,e)|0;r=d;return c|0}function Yw(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+16|0;e=d;k[e>>2]=c;c=cx(a,b,e)|0;r=d;return c|0}function Zw(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0;do if((a|0)!=-1){if((k[b+76>>2]|0)>-1)c=kw(b)|0;else c=0;if(!((k[b+8>>2]|0)==0?(tw(b)|0)!=0:0))d=6;if((d|0)==6?(e=b+4|0,f=k[e>>2]|0,f>>>0>((k[b+44>>2]|0)+-8|0)>>>0):0){g=f+-1|0;k[e>>2]=g;i[g>>0]=a;k[b>>2]=k[b>>2]&-17;if(!c){h=a;break}lw(b);h=a;break}if(c){lw(b);h=-1}else h=-1}else h=-1;while(0);return h|0}function _w(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+224|0;e=d+80|0;f=d+96|0;g=d;h=d+136|0;j=f;l=j+40|0;do{k[j>>2]=0;j=j+4|0}while((j|0)<(l|0));k[e>>2]=k[c>>2];if((Tx(0,b,e,g,f)|0)<0)m=-1;else{if((k[a+76>>2]|0)>-1)n=kw(a)|0;else n=0;c=k[a>>2]|0;j=c&32;if((i[a+74>>0]|0)<1)k[a>>2]=c&-33;c=a+48|0;if(!(k[c>>2]|0)){l=a+44|0;o=k[l>>2]|0;k[l>>2]=h;p=a+28|0;k[p>>2]=h;q=a+20|0;k[q>>2]=h;k[c>>2]=80;s=a+16|0;k[s>>2]=h+80;h=Tx(a,b,e,g,f)|0;if(!o)t=h;else{Pb[k[a+36>>2]&63](a,0,0)|0;u=(k[q>>2]|0)==0?-1:h;k[l>>2]=o;k[c>>2]=0;k[s>>2]=0;k[p>>2]=0;k[q>>2]=0;t=u}}else t=Tx(a,b,e,g,f)|0;f=k[a>>2]|0;k[a>>2]=f|j;if(n)lw(a);m=(f&32|0)==0?t:-1}r=d;return m|0}function $w(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,m=0,n=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0.0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0;d=r;r=r+304|0;e=d+16|0;f=d+8|0;g=d+33|0;h=d;m=d+32|0;if((k[a+76>>2]|0)>-1)n=kw(a)|0;else n=0;q=i[b>>0]|0;a:do if(q<<24>>24){s=a+4|0;t=a+100|0;u=a+108|0;v=a+8|0;w=g+10|0;x=g+33|0;y=f+4|0;z=g+46|0;A=g+94|0;B=q;C=0;D=b;E=0;F=0;G=0;b:while(1){c:do if(!(kv(B&255)|0)){H=(i[D>>0]|0)==37;d:do if(H){I=D+1|0;J=i[I>>0]|0;e:do switch(J<<24>>24){case 37:{break d;break}case 42:{K=0;M=D+2|0;break}default:{N=(J&255)+-48|0;if(N>>>0<10?(i[D+2>>0]|0)==36:0){k[e>>2]=k[c>>2];O=N;while(1){N=(k[e>>2]|0)+(4-1)&~(4-1);P=k[N>>2]|0;k[e>>2]=N+4;if(O>>>0>1)O=O+-1|0;else{Q=P;break}}K=Q;M=D+3|0;break e}O=(k[c>>2]|0)+(4-1)&~(4-1);P=k[O>>2]|0;k[c>>2]=O+4;K=P;M=I}}while(0);I=i[M>>0]|0;J=I&255;if((J+-48|0)>>>0<10){P=J;J=M;O=0;while(1){N=(O*10|0)+-48+P|0;R=J+1|0;S=i[R>>0]|0;P=S&255;if((P+-48|0)>>>0>=10){T=S;U=R;V=N;break}else{J=R;O=N}}}else{T=I;U=M;V=0}if(T<<24>>24==109){O=U+1|0;W=i[O>>0]|0;X=(K|0)!=0&1;Y=O;Z=0;_=0}else{W=T;X=0;Y=U;Z=F;_=G}O=Y+1|0;switch(W&255|0){case 104:{J=(i[O>>0]|0)==104;$=J?Y+2|0:O;aa=J?-2:-1;break}case 108:{J=(i[O>>0]|0)==108;$=J?Y+2|0:O;aa=J?3:1;break}case 106:{$=O;aa=3;break}case 116:case 122:{$=O;aa=1;break}case 76:{$=O;aa=2;break}case 110:case 112:case 67:case 83:case 91:case 99:case 115:case 88:case 71:case 70:case 69:case 65:case 103:case 102:case 101:case 97:case 120:case 117:case 111:case 105:case 100:{$=Y;aa=0;break}default:{ba=X;ca=C;da=Z;ea=_;fa=152;break b}}O=l[$>>0]|0;J=(O&47|0)==3;P=J?O|32:O;O=J?1:aa;switch(P|0){case 99:{ga=E;ha=(V|0)<1?1:V;break}case 91:{ga=E;ha=V;break}case 110:{if(!K){ia=C;ja=$;ka=E;la=Z;ma=_;break c}switch(O|0){case -2:{i[K>>0]=E;ia=C;ja=$;ka=E;la=Z;ma=_;break c;break}case -1:{j[K>>1]=E;ia=C;ja=$;ka=E;la=Z;ma=_;break c;break}case 0:{k[K>>2]=E;ia=C;ja=$;ka=E;la=Z;ma=_;break c;break}case 1:{k[K>>2]=E;ia=C;ja=$;ka=E;la=Z;ma=_;break c;break}case 3:{J=K;k[J>>2]=E;k[J+4>>2]=((E|0)<0)<<31>>31;ia=C;ja=$;ka=E;la=Z;ma=_;break c;break}default:{ia=C;ja=$;ka=E;la=Z;ma=_;break c}}break}default:{xv(a,0);do{J=k[s>>2]|0;if(J>>>0<(k[t>>2]|0)>>>0){k[s>>2]=J+1;na=l[J>>0]|0}else na=yv(a)|0}while((kv(na)|0)!=0);I=k[s>>2]|0;if(!(k[t>>2]|0))oa=I;else{J=I+-1|0;k[s>>2]=J;oa=J}ga=(k[u>>2]|0)+E+oa-(k[v>>2]|0)|0;ha=V}}xv(a,ha);J=k[s>>2]|0;I=k[t>>2]|0;if(J>>>0<I>>>0){k[s>>2]=J+1;pa=I}else{if((yv(a)|0)<0){ba=X;ca=C;da=Z;ea=_;fa=152;break b}pa=k[t>>2]|0}if(pa)k[s>>2]=(k[s>>2]|0)+-1;f:do switch(P|0){case 91:case 99:case 115:{I=(P|0)==99;g:do if((P&239|0)==99){sy(g|0,-1,257)|0;i[g>>0]=0;if((P|0)==115){i[x>>0]=0;i[w>>0]=0;i[w+1>>0]=0;i[w+2>>0]=0;i[w+3>>0]=0;i[w+4>>0]=0;qa=$}else qa=$}else{J=$+1|0;N=(i[J>>0]|0)==94;R=N&1;S=N?J:$;ra=N?$+2|0:J;sy(g|0,N&1|0,257)|0;i[g>>0]=0;switch(i[ra>>0]|0){case 45:{N=(R^1)&255;i[z>>0]=N;sa=N;ta=S+2|0;break}case 93:{N=(R^1)&255;i[A>>0]=N;sa=N;ta=S+2|0;break}default:{sa=(R^1)&255;ta=ra}}ra=ta;while(1){R=i[ra>>0]|0;h:do switch(R<<24>>24){case 0:{ba=X;ca=C;da=Z;ea=_;fa=152;break b;break}case 93:{qa=ra;break g;break}case 45:{S=ra+1|0;N=i[S>>0]|0;switch(N<<24>>24){case 93:case 0:{ua=45;va=ra;break h;break}default:{}}J=i[ra+-1>>0]|0;if((J&255)<(N&255)){wa=J&255;do{wa=wa+1|0;i[g+wa>>0]=sa;J=i[S>>0]|0}while((wa|0)<(J&255|0));ua=J;va=S}else{ua=N;va=S}break}default:{ua=R;va=ra}}while(0);i[g+((ua&255)+1)>>0]=sa;ra=va+1|0}}while(0);ra=I?ha+1|0:31;R=(O|0)==1;wa=(X|0)!=0;i:do if(R){if(wa){J=cy(ra<<2)|0;if(!J){ba=X;ca=C;da=0;ea=J;fa=152;break b}else xa=J}else xa=K;k[f>>2]=0;k[y>>2]=0;J=0;ya=ra;za=xa;j:while(1){Aa=(za|0)==0;Ba=J;while(1){k:while(1){Ca=k[s>>2]|0;if(Ca>>>0<(k[t>>2]|0)>>>0){k[s>>2]=Ca+1;Da=l[Ca>>0]|0}else Da=yv(a)|0;if(!(i[g+(Da+1)>>0]|0)){Ea=Ba;Fa=za;break j}i[m>>0]=Da;switch(Rv(h,m,1,f)|0){case -1:{ba=X;ca=C;da=0;ea=za;fa=152;break b;break}case -2:break;default:break k}}if(Aa)Ga=Ba;else{k[za+(Ba<<2)>>2]=k[h>>2];Ga=Ba+1|0}if(wa&(Ga|0)==(ya|0))break;else Ba=Ga}Ba=ya<<1|1;Aa=fy(za,Ba<<2)|0;if(!Aa){ba=X;ca=C;da=0;ea=za;fa=152;break b}S=ya;ya=Ba;za=Aa;J=S}if(!(Sv(f)|0)){ba=X;ca=C;da=0;ea=Fa;fa=152;break b}else{Ha=Ea;Ia=0;Ja=Fa}}else{if(wa){J=cy(ra)|0;if(!J){ba=X;ca=C;da=0;ea=0;fa=152;break b}else{Ka=0;La=ra;Ma=J}while(1){J=Ka;do{za=k[s>>2]|0;if(za>>>0<(k[t>>2]|0)>>>0){k[s>>2]=za+1;Na=l[za>>0]|0}else Na=yv(a)|0;if(!(i[g+(Na+1)>>0]|0)){Ha=J;Ia=Ma;Ja=0;break i}i[Ma+J>>0]=Na;J=J+1|0}while((J|0)!=(La|0));J=La<<1|1;za=fy(Ma,J)|0;if(!za){ba=X;ca=C;da=Ma;ea=0;fa=152;break b}else{ya=La;La=J;Ma=za;Ka=ya}}}if(!K){ya=pa;while(1){za=k[s>>2]|0;if(za>>>0<ya>>>0){k[s>>2]=za+1;Oa=l[za>>0]|0}else Oa=yv(a)|0;if(!(i[g+(Oa+1)>>0]|0)){Ha=0;Ia=0;Ja=0;break i}ya=k[t>>2]|0}}else{ya=pa;za=0;while(1){J=k[s>>2]|0;if(J>>>0<ya>>>0){k[s>>2]=J+1;Pa=l[J>>0]|0}else Pa=yv(a)|0;if(!(i[g+(Pa+1)>>0]|0)){Ha=za;Ia=K;Ja=0;break i}i[K+za>>0]=Pa;ya=k[t>>2]|0;za=za+1|0}}}while(0);ra=k[s>>2]|0;if(!(k[t>>2]|0))Qa=ra;else{za=ra+-1|0;k[s>>2]=za;Qa=za}za=Qa-(k[v>>2]|0)+(k[u>>2]|0)|0;if(!za){Ra=X;Sa=C;Ta=Ia;Ua=Ja;break b}if(!((za|0)==(ha|0)|I^1)){Ra=X;Sa=C;Ta=Ia;Ua=Ja;break b}do if(wa)if(R){k[K>>2]=Ja;break}else{k[K>>2]=Ia;break}while(0);if(I){Va=qa;Wa=Ia;Xa=Ja}else{if(Ja)k[Ja+(Ha<<2)>>2]=0;if(!Ia){Va=qa;Wa=0;Xa=Ja;break f}i[Ia+Ha>>0]=0;Va=qa;Wa=Ia;Xa=Ja}break}case 120:case 88:case 112:{Ya=16;fa=134;break}case 111:{Ya=8;fa=134;break}case 117:case 100:{Ya=10;fa=134;break}case 105:{Ya=0;fa=134;break}case 71:case 103:case 70:case 102:case 69:case 101:case 65:case 97:{Za=+vv(a,O,0);if((k[u>>2]|0)==((k[v>>2]|0)-(k[s>>2]|0)|0)){Ra=X;Sa=C;Ta=Z;Ua=_;break b}if(!K){Va=$;Wa=Z;Xa=_}else switch(O|0){case 0:{o[K>>2]=Za;Va=$;Wa=Z;Xa=_;break f;break}case 1:{p[K>>3]=Za;Va=$;Wa=Z;Xa=_;break f;break}case 2:{p[K>>3]=Za;Va=$;Wa=Z;Xa=_;break f;break}default:{Va=$;Wa=Z;Xa=_;break f}}break}default:{Va=$;Wa=Z;Xa=_}}while(0);l:do if((fa|0)==134){fa=0;R=wv(a,Ya,0,-1,-1)|0;if((k[u>>2]|0)==((k[v>>2]|0)-(k[s>>2]|0)|0)){Ra=X;Sa=C;Ta=Z;Ua=_;break b}if((K|0)!=0&(P|0)==112){k[K>>2]=R;Va=$;Wa=Z;Xa=_;break}if(!K){Va=$;Wa=Z;Xa=_}else switch(O|0){case -2:{i[K>>0]=R;Va=$;Wa=Z;Xa=_;break l;break}case -1:{j[K>>1]=R;Va=$;Wa=Z;Xa=_;break l;break}case 0:{k[K>>2]=R;Va=$;Wa=Z;Xa=_;break l;break}case 1:{k[K>>2]=R;Va=$;Wa=Z;Xa=_;break l;break}case 3:{wa=K;k[wa>>2]=R;k[wa+4>>2]=L;Va=$;Wa=Z;Xa=_;break l;break}default:{Va=$;Wa=Z;Xa=_;break l}}}while(0);ia=((K|0)!=0&1)+C|0;ja=Va;ka=(k[u>>2]|0)+ga+(k[s>>2]|0)-(k[v>>2]|0)|0;la=Wa;ma=Xa;break c}while(0);O=D+(H&1)|0;xv(a,0);P=k[s>>2]|0;if(P>>>0<(k[t>>2]|0)>>>0){k[s>>2]=P+1;_a=l[P>>0]|0}else _a=yv(a)|0;if((_a|0)!=(l[O>>0]|0)){$a=_a;ab=C;bb=F;cb=G;fa=21;break b}ia=C;ja=O;ka=E+1|0;la=F;ma=G}else{O=D;while(1){P=O+1|0;if(!(kv(l[P>>0]|0)|0)){db=O;break}else O=P}xv(a,0);do{O=k[s>>2]|0;if(O>>>0<(k[t>>2]|0)>>>0){k[s>>2]=O+1;eb=l[O>>0]|0}else eb=yv(a)|0}while((kv(eb)|0)!=0);O=k[s>>2]|0;if(!(k[t>>2]|0))fb=O;else{H=O+-1|0;k[s>>2]=H;fb=H}ia=C;ja=db;ka=(k[u>>2]|0)+E+fb-(k[v>>2]|0)|0;la=F;ma=G}while(0);D=ja+1|0;B=i[D>>0]|0;if(!(B<<24>>24)){gb=ia;break a}else{C=ia;E=ka;F=la;G=ma}}if((fa|0)==21){if(k[t>>2]|0)k[s>>2]=(k[s>>2]|0)+-1;if((ab|0)!=0|($a|0)>-1){gb=ab;break}else{hb=0;ib=bb;jb=cb;fa=153}}else if((fa|0)==152)if(!ca){hb=ba;ib=da;jb=ea;fa=153}else{Ra=ba;Sa=ca;Ta=da;Ua=ea}if((fa|0)==153){Ra=hb;Sa=-1;Ta=ib;Ua=jb}if(!Ra)gb=Sa;else{dy(Ta);dy(Ua);gb=Sa}}else gb=0;while(0);if(n)lw(a);r=d;return gb|0}function ax(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0;e=r;r=r+128|0;f=e+112|0;g=e;h=g;j=125028;l=h+112|0;do{k[h>>2]=k[j>>2];h=h+4|0;j=j+4|0}while((h|0)<(l|0));if((b+-1|0)>>>0>2147483646)if(!b){m=f;n=1;o=4}else{f=rv()|0;k[f>>2]=75;p=-1}else{m=a;n=b;o=4}if((o|0)==4){o=-2-m|0;b=n>>>0>o>>>0?o:n;k[g+48>>2]=b;n=g+20|0;k[n>>2]=m;k[g+44>>2]=m;o=m+b|0;m=g+16|0;k[m>>2]=o;k[g+28>>2]=o;o=_w(g,c,d)|0;if(!b)p=o;else{b=k[n>>2]|0;i[b+(((b|0)==(k[m>>2]|0))<<31>>31)>>0]=0;p=o}}r=e;return p|0}function bx(a,b,c){a=a|0;b=b|0;c=c|0;return ax(a,2147483647,b,c)|0}function cx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=r;r=r+112|0;e=d;f=e;g=f+112|0;do{k[f>>2]=0;f=f+4|0}while((f|0)<(g|0));k[e+32>>2]=44;k[e+44>>2]=a;k[e+76>>2]=-1;k[e+84>>2]=a;a=$w(e,b,c)|0;r=d;return a|0}function dx(a){a=a|0;return +(+hx(a,0))}function ex(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0;b=a;while(1){a=b+1|0;if(!(kv(i[b>>0]|0)|0)){c=b;d=a;break}else b=a}b=i[c>>0]|0;switch(b<<24>>24|0){case 45:{e=1;f=5;break}case 43:{e=0;f=5;break}default:{g=c;h=b;j=0}}if((f|0)==5){g=d;h=i[d>>0]|0;j=e}e=(h<<24>>24)+-48|0;if(e>>>0<10){h=g;g=e;e=0;while(1){h=h+1|0;d=(e*10|0)-g|0;g=(i[h>>0]|0)+-48|0;if(g>>>0>=10){k=d;break}else e=d}}else k=0;return ((j|0)!=0?k:0-k|0)|0}function fx(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0;b=a;while(1){a=b+1|0;if(!(kv(i[b>>0]|0)|0)){c=b;d=a;break}else b=a}b=i[c>>0]|0;switch(b<<24>>24|0){case 45:{e=1;f=5;break}case 43:{e=0;f=5;break}default:{g=c;h=b;j=0}}if((f|0)==5){g=d;h=i[d>>0]|0;j=e}e=(h<<24>>24)+-48|0;if(e>>>0<10){h=g;g=e;e=0;while(1){h=h+1|0;d=(e*10|0)-g|0;g=(i[h>>0]|0)+-48|0;if(g>>>0>=10){k=d;break}else e=d}}else k=0;return ((j|0)!=0?k:0-k|0)|0}function gx(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0;e=r;r=r+208|0;f=e+8|0;g=e;h=ha(c,b)|0;b=g;k[b>>2]=1;k[b+4>>2]=0;if(h){b=h-c|0;k[f+4>>2]=c;k[f>>2]=c;i=c;j=c;l=2;while(1){m=j+c+i|0;k[f+(l<<2)>>2]=m;if(m>>>0<h>>>0){n=i;i=m;l=l+1|0;j=n}else break}j=0-c|0;l=a+b|0;i=g+4|0;if((b|0)>0){b=l;h=1;n=a;m=1;while(1){do if((h&3|0)==3){Vx(n,c,d,m,f);o=k[i>>2]|0;p=o<<30|(k[g>>2]|0)>>>2;k[g>>2]=p;k[i>>2]=o>>>2;q=p;s=m+2|0}else{p=m+-1|0;if((k[f+(p<<2)>>2]|0)>>>0<(b-n|0)>>>0)Vx(n,c,d,m,f);else Wx(n,c,d,g,m,0,f);if((m|0)==1){o=k[g>>2]|0;k[i>>2]=o>>>31|k[i>>2]<<1;t=o<<1;k[g>>2]=t;q=t;s=0;break}if(p>>>0>31){t=k[g>>2]|0;k[i>>2]=t;k[g>>2]=0;u=m+-33|0;v=t;w=0}else{u=p;v=k[i>>2]|0;w=k[g>>2]|0}k[i>>2]=w>>>(32-u|0)|v<<u;p=w<<u;k[g>>2]=p;q=p;s=1}while(0);h=q|1;k[g>>2]=h;p=n+c|0;if(p>>>0>=l>>>0){x=p;y=s;break}else{n=p;m=s}}}else{x=a;y=1}Wx(x,c,d,g,y,0,f);a=g+4|0;s=k[g>>2]|0;if(!((y|0)==1&(s|0)==1&(k[a>>2]|0)==0)){m=s;s=x;x=y;while(1){if((x|0)<2){y=m+-1|0;do if(y){if(!(y&1)){n=y;l=0;while(1){h=l+1|0;n=n>>>1;if(n&1){z=h;break}else l=h}if(z)A=z;else B=24}else B=24;if((B|0)==24){B=0;l=k[i>>2]|0;if(!l){C=64;B=30;break}if(!(l&1)){D=l;E=0}else{F=0;G=m;H=l;I=0;break}while(1){n=E+1|0;D=D>>>1;if(D&1){J=n;K=E;break}else E=n}if(!J){F=0;G=m;H=l;I=0;break}else A=K+33|0}if(A>>>0>31){C=A;B=30}else{F=A;G=m;H=k[i>>2]|0;I=A}}else{C=32;B=30}while(0);if((B|0)==30){B=0;y=k[i>>2]|0;k[g>>2]=y;k[i>>2]=0;F=C+-32|0;G=y;H=0;I=C}k[g>>2]=H<<32-F|G>>>F;k[i>>2]=H>>>F;L=s+j|0;M=I+x|0}else{y=m>>>30;n=y|k[i>>2]<<2;h=x+-2|0;k[g>>2]=(m<<1&2147483646|y<<31)^3;k[i>>2]=n>>>1;Wx(s+(0-((k[f+(h<<2)>>2]|0)+c))|0,c,d,g,x+-1|0,1,f);n=k[g>>2]|0;k[i>>2]=n>>>31|k[i>>2]<<1;k[g>>2]=n<<1|1;n=s+j|0;Wx(n,c,d,g,h,1,f);L=n;M=h}m=k[g>>2]|0;if((M|0)==1&(m|0)==1&(k[a>>2]|0)==0)break;else{s=L;x=M}}}}r=e;return}function hx(a,b){a=a|0;b=b|0;return +(+Xx(a,b,1))}function ix(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;d=Yx(a,b,c,0,-2147483648)|0;return d|0}function jx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;d=Yx(a,b,c,-2147483648,0)|0;return d|0}function kx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;d=b&255;e=(c|0)!=0;a:do if(e&(a&3|0)!=0){f=b&255;g=c;h=a;while(1){if((i[h>>0]|0)==f<<24>>24){j=g;l=h;m=6;break a}n=h+1|0;o=g+-1|0;p=(o|0)!=0;if(p&(n&3|0)!=0){g=o;h=n}else{q=o;r=p;s=n;m=5;break}}}else{q=c;r=e;s=a;m=5}while(0);if((m|0)==5)if(r){j=q;l=s;m=6}else{t=0;u=s}b:do if((m|0)==6){s=b&255;if((i[l>>0]|0)==s<<24>>24){t=j;u=l}else{q=ha(d,16843009)|0;c:do if(j>>>0>3){r=j;a=l;while(1){e=k[a>>2]^q;if((e&-2139062144^-2139062144)&e+-16843009){v=r;w=a;break}e=a+4|0;c=r+-4|0;if(c>>>0>3){r=c;a=e}else{x=c;y=e;m=11;break c}}z=v;A=w}else{x=j;y=l;m=11}while(0);if((m|0)==11)if(!x){t=0;u=y;break}else{z=x;A=y}while(1){if((i[A>>0]|0)==s<<24>>24){t=z;u=A;break b}q=A+1|0;z=z+-1|0;if(!z){t=0;u=q;break}else A=q}}}while(0);return ((t|0)!=0?u:0)|0}function lx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0;a:do if(!c)d=0;else{e=c;f=a;g=b;while(1){h=i[f>>0]|0;j=i[g>>0]|0;if(h<<24>>24!=j<<24>>24){k=h;l=j;break}e=e+-1|0;if(!e){d=0;break a}else{f=f+1|0;g=g+1|0}}d=(k&255)-(l&255)|0}while(0);return d|0}function mx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=b&255;b=c;while(1){if(!b){e=0;break}b=b+-1|0;c=a+b|0;if((i[c>>0]|0)==d<<24>>24){e=c;break}}return e|0}function nx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0;c=b;a:do if(!((c^a)&3)){if(!(c&3)){d=b;e=a}else{f=a;g=b;while(1){h=i[g>>0]|0;i[f>>0]=h;if(!(h<<24>>24)){j=f;break a}h=g+1|0;l=f+1|0;if(!(h&3)){d=h;e=l;break}else{f=l;g=h}}}g=k[d>>2]|0;if(!((g&-2139062144^-2139062144)&g+-16843009)){f=g;g=e;h=d;while(1){l=h+4|0;m=g+4|0;k[g>>2]=f;f=k[l>>2]|0;if((f&-2139062144^-2139062144)&f+-16843009){n=m;o=l;break}else{g=m;h=l}}}else{n=e;o=d}p=o;q=n;r=8}else{p=b;q=a;r=8}while(0);if((r|0)==8){r=i[p>>0]|0;i[q>>0]=r;if(!(r<<24>>24))j=q;else{r=q;q=p;while(1){q=q+1|0;p=r+1|0;a=i[q>>0]|0;i[p>>0]=a;if(!(a<<24>>24)){j=p;break}else r=p}}}return j|0}function ox(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0;d=b;do if(!((d^a)&3)){e=(c|0)!=0;a:do if(e&(d&3|0)!=0){f=c;g=b;h=a;while(1){j=i[g>>0]|0;i[h>>0]=j;if(!(j<<24>>24)){l=f;m=g;n=h;break a}j=f+-1|0;o=g+1|0;p=h+1|0;q=(j|0)!=0;if(q&(o&3|0)!=0){f=j;g=o;h=p}else{r=j;s=o;t=p;u=q;v=5;break}}}else{r=c;s=b;t=a;u=e;v=5}while(0);if((v|0)==5)if(u){l=r;m=s;n=t}else{w=t;x=0;break}if(!(i[m>>0]|0)){w=n;x=l}else{b:do if(l>>>0>3){e=l;h=n;g=m;while(1){f=k[g>>2]|0;if((f&-2139062144^-2139062144)&f+-16843009){y=e;z=h;A=g;break b}k[h>>2]=f;f=e+-4|0;q=g+4|0;p=h+4|0;if(f>>>0>3){e=f;h=p;g=q}else{y=f;z=p;A=q;break}}}else{y=l;z=n;A=m}while(0);B=A;C=z;D=y;v=11}}else{B=b;C=a;D=c;v=11}while(0);c:do if((v|0)==11)if(!D){w=C;x=0}else{c=B;a=C;b=D;while(1){y=i[c>>0]|0;i[a>>0]=y;if(!(y<<24>>24)){w=a;x=b;break c}b=b+-1|0;y=a+1|0;if(!b){w=y;x=0;break}else{c=c+1|0;a=y}}}while(0);sy(w|0,0,x|0)|0;return w|0}function px(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,m=0,n=0,o=0;c=i[a>>0]|0;a:do if(!(c<<24>>24)){d=0;e=b}else{f=c;g=c&255;h=a;j=b;while(1){k=i[j>>0]|0;if(!(k<<24>>24)){d=f;e=j;break a}if(f<<24>>24!=k<<24>>24?(k=mv(g)|0,(k|0)!=(mv(l[j>>0]|0)|0)):0){m=h;n=j;break}h=h+1|0;k=j+1|0;o=i[h>>0]|0;if(!(o<<24>>24)){d=0;e=k;break a}else{f=o;g=o&255;j=k}}d=i[m>>0]|0;e=n}while(0);n=mv(d&255)|0;return n-(mv(l[e>>0]|0)|0)|0}function qx(a,b){a=a|0;b=b|0;ux(a+(xx(a)|0)|0,b)|0;return a|0}function rx(a,b){a=a|0;b=b|0;var c=0;c=sx(a,b)|0;return ((i[c>>0]|0)==(b&255)<<24>>24?c:0)|0}function sx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;c=b&255;a:do if(!c)d=a+(xx(a)|0)|0;else{if(!(a&3))e=a;else{f=b&255;g=a;while(1){h=i[g>>0]|0;if(h<<24>>24==0?1:h<<24>>24==f<<24>>24){d=g;break a}h=g+1|0;if(!(h&3)){e=h;break}else g=h}}g=ha(c,16843009)|0;f=k[e>>2]|0;b:do if(!((f&-2139062144^-2139062144)&f+-16843009)){h=f;j=e;while(1){l=h^g;if((l&-2139062144^-2139062144)&l+-16843009){m=j;break b}l=j+4|0;h=k[l>>2]|0;if((h&-2139062144^-2139062144)&h+-16843009){m=l;break}else j=l}}else m=e;while(0);g=b&255;f=m;while(1){j=i[f>>0]|0;if(j<<24>>24==0?1:j<<24>>24==g<<24>>24){d=f;break}else f=f+1|0}}while(0);return d|0}function tx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0;c=i[a>>0]|0;d=i[b>>0]|0;if(c<<24>>24==0?1:c<<24>>24!=d<<24>>24){e=c;f=d}else{d=a;a=b;do{d=d+1|0;a=a+1|0;b=i[d>>0]|0;c=i[a>>0]|0}while(!(b<<24>>24==0?1:b<<24>>24!=c<<24>>24));e=b;f=c}return (e&255)-(f&255)|0}function ux(a,b){a=a|0;b=b|0;nx(a,b)|0;return a|0}function vx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0;c=r;r=r+32|0;d=c;e=i[b>>0]|0;if(e<<24>>24!=0?(i[b+1>>0]|0)!=0:0){k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;k[d+16>>2]=0;k[d+20>>2]=0;k[d+24>>2]=0;k[d+28>>2]=0;f=b;b=e;do{g=d+(((b&255)>>>5&255)<<2)|0;k[g>>2]=k[g>>2]|1<<(b&31);f=f+1|0;b=i[f>>0]|0}while(b<<24>>24!=0);b=i[a>>0]|0;a:do if(!(b<<24>>24))h=a;else{f=a;g=b;while(1){if(k[d+(((g&255)>>>5&255)<<2)>>2]&1<<(g&31)){h=f;break a}j=f+1|0;g=i[j>>0]|0;if(!(g<<24>>24)){h=j;break}else f=j}}while(0);l=h-a|0}else l=(sx(a,e<<24>>24)|0)-a|0;r=c;return l|0}function wx(a){a=a|0;var b=0,c=0,d=0;b=(xx(a)|0)+1|0;c=cy(b)|0;if(!c)d=0;else{py(c|0,a|0,b|0)|0;d=c}return d|0}function xx(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0;b=a;a:do if(!(b&3)){c=a;d=4}else{e=a;f=b;while(1){if(!(i[e>>0]|0)){g=f;break a}h=e+1|0;f=h;if(!(f&3)){c=h;d=4;break}else e=h}}while(0);if((d|0)==4){d=c;while(1){c=k[d>>2]|0;if(!((c&-2139062144^-2139062144)&c+-16843009))d=d+4|0;else{j=c;l=d;break}}if(!((j&255)<<24>>24))m=l;else{j=l;while(1){l=j+1|0;if(!(i[l>>0]|0)){m=l;break}else j=l}}g=m}return g-b|0}function yx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,m=0,n=0,o=0,p=0,q=0,r=0;if(!c)d=0;else{e=i[a>>0]|0;a:do if(!(e<<24>>24)){f=0;g=b}else{h=c;j=e;k=e&255;m=a;n=b;while(1){h=h+-1|0;o=i[n>>0]|0;if(!((h|0)!=0&o<<24>>24!=0)){f=j;g=n;break a}if(j<<24>>24!=o<<24>>24?(o=mv(k)|0,(o|0)!=(mv(l[n>>0]|0)|0)):0){p=m;q=n;break}m=m+1|0;o=n+1|0;r=i[m>>0]|0;if(!(r<<24>>24)){f=0;g=o;break a}else{j=r;k=r&255;n=o}}f=i[p>>0]|0;g=q}while(0);q=mv(f&255)|0;d=q-(mv(l[g>>0]|0)|0)|0}return d|0}function zx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0;d=a+(xx(a)|0)|0;a:do if(!c)e=d;else{f=c;g=b;h=d;while(1){j=i[g>>0]|0;if(!(j<<24>>24)){e=h;break a}f=f+-1|0;k=h+1|0;i[h>>0]=j;if(!f){e=k;break}else{g=g+1|0;h=k}}}while(0);i[e>>0]=0;return a|0}function Ax(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,m=0,n=0;if(!c)d=0;else{e=i[a>>0]|0;a:do if(!(e<<24>>24)){f=0;g=b}else{h=c;j=e;k=a;m=b;while(1){h=h+-1|0;n=i[m>>0]|0;if(!(j<<24>>24==n<<24>>24&((h|0)!=0&n<<24>>24!=0))){f=j;g=m;break a}k=k+1|0;n=m+1|0;j=i[k>>0]|0;if(!(j<<24>>24)){f=0;g=n;break}else m=n}}while(0);d=(f&255)-(l[g>>0]|0)|0}return d|0}function Bx(a,b,c){a=a|0;b=b|0;c=c|0;ox(a,b,c)|0;return a|0}function Cx(a,b){a=a|0;b=b|0;return mx(a,b,(xx(a)|0)+1|0)|0}function Dx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0;c=r;r=r+32|0;d=c;k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;k[d+16>>2]=0;k[d+20>>2]=0;k[d+24>>2]=0;k[d+28>>2]=0;e=i[b>>0]|0;do if(!(e<<24>>24))f=0;else{if(!(i[b+1>>0]|0)){g=a;while(1)if((i[g>>0]|0)==e<<24>>24)g=g+1|0;else{h=g;break}f=h-a|0;break}else{j=b;l=e}do{g=d+(((l&255)>>>5&255)<<2)|0;k[g>>2]=k[g>>2]|1<<(l&31);j=j+1|0;l=i[j>>0]|0}while(l<<24>>24!=0);g=i[a>>0]|0;a:do if(!(g<<24>>24))m=a;else{n=a;o=g;while(1){if(!(k[d+(((o&255)>>>5&255)<<2)>>2]&1<<(o&31))){m=n;break a}p=n+1|0;o=i[p>>0]|0;if(!(o<<24>>24)){m=p;break}else n=p}}while(0);f=m-a|0}while(0);r=c;return f|0}function Ex(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0;c=r;r=r+1056|0;d=c+1024|0;e=c;f=i[b>>0]|0;do if(f<<24>>24){g=rx(a,f<<24>>24)|0;if(g){h=i[b+1>>0]|0;if(h<<24>>24){j=g+1|0;m=i[j>>0]|0;if(m<<24>>24){n=i[b+2>>0]|0;if(!(n<<24>>24)){o=h&255|(f&255)<<8;p=j;j=m;q=g;s=l[g>>0]<<8|m&255;while(1){t=s&65535;if((t|0)==(o|0)){u=q;v=j;break}w=p+1|0;x=i[w>>0]|0;if(!(x<<24>>24)){u=p;v=0;break}else{y=p;p=w;j=x;s=x&255|t<<8;q=y}}z=v<<24>>24!=0?u:0;break}q=g+2|0;s=i[q>>0]|0;if(s<<24>>24){j=i[b+3>>0]|0;if(!(j<<24>>24)){p=(h&255)<<16|(f&255)<<24|(n&255)<<8;o=(s&255)<<8|(m&255)<<16|l[g>>0]<<24;if((o|0)==(p|0)){A=q;B=s}else{y=q;q=o;while(1){o=y+1|0;t=i[o>>0]|0;q=(t&255|q)<<8;if(t<<24>>24==0|(q|0)==(p|0)){A=o;B=t;break}else y=o}}z=B<<24>>24!=0?A+-2|0:0;break}y=g+3|0;p=i[y>>0]|0;if(p<<24>>24){if(!(i[b+4>>0]|0)){q=(h&255)<<16|(f&255)<<24|(n&255)<<8|j&255;o=(s&255)<<8|(m&255)<<16|p&255|l[g>>0]<<24;if((o|0)==(q|0)){C=y;D=p}else{p=y;y=o;while(1){o=p+1|0;t=i[o>>0]|0;y=t&255|y<<8;if(t<<24>>24==0|(y|0)==(q|0)){C=o;D=t;break}else p=o}}z=D<<24>>24!=0?C+-3|0:0;break};k[d>>2]=0;k[d+4>>2]=0;k[d+8>>2]=0;k[d+12>>2]=0;k[d+16>>2]=0;k[d+20>>2]=0;k[d+24>>2]=0;k[d+28>>2]=0;p=f;q=0;while(1){if(!(i[g+q>>0]|0)){E=0;break}y=d+(((p&255)>>>5&255)<<2)|0;k[y>>2]=k[y>>2]|1<<(p&31);y=q+1|0;k[e+((p&255)<<2)>>2]=y;p=i[b+y>>0]|0;if(!(p<<24>>24)){F=y;G=q;H=23;break}else q=y}a:do if((H|0)==23){b:do if(F>>>0>1){q=1;p=-1;y=0;c:while(1){m=q;s=y;j=1;while(1){n=m;h=s;d:while(1){o=n;t=1;while(1){x=i[b+(t+p)>>0]|0;w=i[b+o>>0]|0;if(x<<24>>24!=w<<24>>24){I=o;J=x;K=w;L=h;break d}if((t|0)==(j|0))break;t=t+1|0;o=t+h|0;if(o>>>0>=F>>>0){M=p;N=j;break c}}h=h+j|0;n=h+1|0;if(n>>>0>=F>>>0){M=p;N=j;break c}}n=I-p|0;if((J&255)<=(K&255)){O=L;break}h=I+1|0;if(h>>>0<F>>>0){m=h;s=I;j=n}else{M=p;N=n;break c}}q=O+2|0;if(q>>>0>=F>>>0){M=O;N=1;break}else{p=O;y=O+1|0}}y=1;p=-1;q=0;while(1){j=y;s=q;m=1;while(1){n=j;h=s;e:while(1){o=n;t=1;while(1){w=i[b+(t+p)>>0]|0;x=i[b+o>>0]|0;if(w<<24>>24!=x<<24>>24){P=o;Q=w;R=x;S=h;break e}if((t|0)==(m|0))break;t=t+1|0;o=t+h|0;if(o>>>0>=F>>>0){T=M;U=p;V=N;W=m;break b}}h=h+m|0;n=h+1|0;if(n>>>0>=F>>>0){T=M;U=p;V=N;W=m;break b}}n=P-p|0;if((Q&255)>=(R&255)){X=S;break}h=P+1|0;if(h>>>0<F>>>0){j=h;s=P;m=n}else{T=M;U=p;V=N;W=n;break b}}y=X+2|0;if(y>>>0>=F>>>0){T=M;U=X;V=N;W=1;break}else{p=X;q=X+1|0}}}else{T=-1;U=-1;V=1;W=1}while(0);q=(U+1|0)>>>0>(T+1|0)>>>0;p=q?W:V;y=q?U:T;q=y+1|0;if(!(lx(b,b+p|0,q)|0)){Y=F-p|0;Z=p}else{p=F-y+-1|0;Y=0;Z=(y>>>0>p>>>0?y:p)+1|0}p=F|63;m=(Y|0)!=0;s=F-Z|0;j=g;n=0;h=g;f:while(1){o=j;do if((h-o|0)>>>0<F>>>0){t=kx(h,0,p)|0;if(t)if((t-o|0)>>>0<F>>>0){E=0;break a}else{_=t;break}else{_=h+p|0;break}}else _=h;while(0);o=i[j+G>>0]|0;if(!(1<<(o&31)&k[d+(((o&255)>>>5&255)<<2)>>2])){j=j+F|0;n=0;h=_;continue}t=k[e+((o&255)<<2)>>2]|0;o=F-t|0;if((F|0)!=(t|0)){j=j+(m&(n|0)!=0&o>>>0<Z>>>0?s:o)|0;n=0;h=_;continue}o=q>>>0>n>>>0?q:n;t=i[b+o>>0]|0;g:do if(!(t<<24>>24))$=q;else{x=t;w=o;while(1){if(x<<24>>24!=(i[j+w>>0]|0)){aa=w;break}w=w+1|0;x=i[b+w>>0]|0;if(!(x<<24>>24)){$=q;break g}}j=j+(aa-y)|0;n=0;h=_;continue f}while(0);do{if($>>>0<=n>>>0){E=j;break a}$=$+-1|0}while((i[b+$>>0]|0)==(i[j+$>>0]|0));j=j+Z|0;n=Y;h=_}}while(0);z=E}else z=0}else z=0}else z=0}else z=g}else z=0}else z=a;while(0);r=c;return z|0}function Fx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0;if(!a){c=k[31285]|0;if(!c)d=0;else{e=c;f=3}}else{e=a;f=3}do if((f|0)==3){a=Dx(e,b)|0;c=e+a|0;if(!(i[c>>0]|0)){k[31285]=0;d=0;break}g=(vx(c,b)|0)+a|0;a=e+g|0;k[31285]=a;if(!(i[a>>0]|0)){k[31285]=0;d=c;break}else{k[31285]=e+(g+1);i[a>>0]=0;d=c;break}}while(0);return d|0}function Gx(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;k[d>>2]=b;b=((Qv(a,21505,d)|0)!=0)<<31>>31;r=c;return b|0}function Hx(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;k[c>>2]=a;a=cb(6,c|0)|0;c=zv((a|0)==-4?-115:a)|0;r=b;return c|0}function Ix(a,b){a=a|0;b=b|0;var c=0,d=0;c=r;r=r+16|0;d=c;k[d>>2]=a;k[d+4>>2]=0;k[d+8>>2]=b;k[d+12>>2]=((b|0)<0)<<31>>31;b=zv(_a(194,d|0)|0)|0;r=c;return b|0}function Jx(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0;c=r;r=r+4112|0;d=c;e=c+8|0;if(a)if(!b){f=rv()|0;k[f>>2]=22;g=0}else{h=b;i=a;j=4}else{h=4096;i=e;j=4}if((j|0)==4){k[d>>2]=i;k[d+4>>2]=h;if((zv(zb(183,d|0)|0)|0)>=0)if((i|0)==(e|0))g=wx(e)|0;else g=i;else g=0}r=c;return g|0}function Kx(a){a=a|0;var b=0,c=0;b=r;r=r+64|0;c=(Gx(a,b)|0)==0&1;r=b;return c|0}function Lx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;d=r;r=r+32|0;e=d;f=d+20|0;k[e>>2]=a;k[e+4>>2]=0;k[e+8>>2]=b;k[e+12>>2]=f;k[e+16>>2]=c;c=(zv(Gb(140,e|0)|0)|0)!=0;r=d;return (c?-1:k[f>>2]|0)|0}function Mx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+16|0;e=d;k[e>>2]=a;k[e+4>>2]=b;k[e+8>>2]=c;c=zv(ub(3,e|0)|0)|0;r=d;return c|0}function Nx(a){a=a|0;var b=0,c=0;b=r;r=r+16|0;c=b;k[c>>2]=a;a=zv(qb(10,c|0)|0)|0;r=b;return a|0}function Ox(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=r;r=r+16|0;e=d;k[e>>2]=a;k[e+4>>2]=b;k[e+8>>2]=c;c=zv(Lb(4,e|0)|0)|0;r=d;return c|0}function Px(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0;c=a+4|0;d=k[c>>2]|0;e=a+100|0;if(d>>>0<(k[e>>2]|0)>>>0){k[c>>2]=d+1;f=l[d>>0]|0}else f=yv(a)|0;switch(f|0){case 43:case 45:{d=(f|0)==45&1;g=k[c>>2]|0;if(g>>>0<(k[e>>2]|0)>>>0){k[c>>2]=g+1;h=l[g>>0]|0}else h=yv(a)|0;if((b|0)!=0&(h+-48|0)>>>0>9?(k[e>>2]|0)!=0:0){k[c>>2]=(k[c>>2]|0)+-1;i=h;j=d}else{i=h;j=d}break}default:{i=f;j=0}}if((i+-48|0)>>>0>9)if(!(k[e>>2]|0)){m=-2147483648;n=0}else{k[c>>2]=(k[c>>2]|0)+-1;m=-2147483648;n=0}else{f=i;i=0;while(1){d=f+-48+(i*10|0)|0;h=k[c>>2]|0;if(h>>>0<(k[e>>2]|0)>>>0){k[c>>2]=h+1;o=l[h>>0]|0}else o=yv(a)|0;if((o+-48|0)>>>0<10&(d|0)<214748364){f=o;i=d}else{p=d;q=o;break}}o=((p|0)<0)<<31>>31;if((q+-48|0)>>>0<10){i=p;f=o;d=q;while(1){h=yy(i|0,f|0,10,0)|0;b=L;g=ky(d|0,((d|0)<0)<<31>>31|0,-48,-1)|0;r=ky(g|0,L|0,h|0,b|0)|0;b=L;h=k[c>>2]|0;if(h>>>0<(k[e>>2]|0)>>>0){k[c>>2]=h+1;s=l[h>>0]|0}else s=yv(a)|0;if((s+-48|0)>>>0<10&((b|0)<21474836|(b|0)==21474836&r>>>0<2061584302)){i=r;f=b;d=s}else{t=r;u=b;v=s;break}}}else{t=p;u=o;v=q}if((v+-48|0)>>>0<10)do{v=k[c>>2]|0;if(v>>>0<(k[e>>2]|0)>>>0){k[c>>2]=v+1;w=l[v>>0]|0}else w=yv(a)|0}while((w+-48|0)>>>0<10);if(k[e>>2]|0)k[c>>2]=(k[c>>2]|0)+-1;c=(j|0)!=0;j=jy(0,0,t|0,u|0)|0;m=c?L:u;n=c?j:t}L=m;return n|0}function Qx(a){a=a|0;if(!(k[a+68>>2]|0))lw(a);return}function Rx(a){a=a|0;if(!(k[a+68>>2]|0))lw(a);return}function Sx(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;b=a+20|0;c=a+28|0;if((k[b>>2]|0)>>>0>(k[c>>2]|0)>>>0?(Pb[k[a+36>>2]&63](a,0,0)|0,(k[b>>2]|0)==0):0)d=-1;else{e=a+4|0;f=k[e>>2]|0;g=a+8|0;h=k[g>>2]|0;if(f>>>0<h>>>0)Pb[k[a+40>>2]&63](a,f-h|0,1)|0;k[a+16>>2]=0;k[c>>2]=0;k[b>>2]=0;k[g>>2]=0;k[e>>2]=0;d=0}return d|0}function Tx(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,m=0,n=0,o=0,q=0,s=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0.0,bb=0.0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0.0,ib=0.0,jb=0.0,kb=0.0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0.0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0.0,Wb=0.0,Xb=0.0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,gc=0,hc=0,ic=0,jc=0,kc=0,lc=0,mc=0,nc=0,oc=0,pc=0,qc=0,rc=0,sc=0,tc=0,uc=0,vc=0,wc=0,xc=0,yc=0,zc=0,Ac=0,Bc=0,Cc=0,Dc=0,Ec=0;f=r;r=r+624|0;g=f+24|0;h=f+16|0;m=f+588|0;n=f+576|0;o=f;q=f+536|0;s=f+8|0;u=f+528|0;v=(a|0)!=0;w=q+40|0;x=w;y=q+39|0;q=s+4|0;z=n+12|0;A=n+11|0;n=m;B=z;C=B-n|0;D=-2-n|0;E=B+2|0;F=g+288|0;G=m+9|0;H=G;I=m+8|0;J=0;K=b;b=0;M=0;a:while(1){do if((J|0)>-1)if((b|0)>(2147483647-J|0)){N=rv()|0;k[N>>2]=75;O=-1;break}else{O=b+J|0;break}else O=J;while(0);N=i[K>>0]|0;if(!(N<<24>>24)){P=O;Q=M;R=245;break}else{S=N;T=K}b:while(1){switch(S<<24>>24){case 37:{U=T;V=T;R=9;break b;break}case 0:{W=T;X=T;break b;break}default:{}}N=T+1|0;S=i[N>>0]|0;T=N}c:do if((R|0)==9)while(1){R=0;if((i[U+1>>0]|0)!=37){W=U;X=V;break c}N=V+1|0;Y=U+2|0;if((i[Y>>0]|0)==37){U=Y;V=N;R=9}else{W=Y;X=N;break}}while(0);N=X-K|0;if(v?(k[a>>2]&32|0)==0:0)Ow(K,N,a)|0;if((X|0)!=(K|0)){J=O;K=W;b=N;continue}Y=W+1|0;Z=i[Y>>0]|0;_=(Z<<24>>24)+-48|0;if(_>>>0<10){$=(i[W+2>>0]|0)==36;aa=$?W+3|0:Y;ba=i[aa>>0]|0;ca=$?_:-1;da=$?1:M;ea=aa}else{ba=Z;ca=-1;da=M;ea=Y}Y=ba<<24>>24;d:do if((Y&-32|0)==32){Z=Y;aa=ba;$=0;_=ea;while(1){if(!(1<<Z+-32&75913)){fa=aa;ga=$;ia=_;break d}ja=1<<(aa<<24>>24)+-32|$;ka=_+1|0;la=i[ka>>0]|0;Z=la<<24>>24;if((Z&-32|0)!=32){fa=la;ga=ja;ia=ka;break}else{aa=la;$=ja;_=ka}}}else{fa=ba;ga=0;ia=ea}while(0);do if(fa<<24>>24==42){Y=ia+1|0;_=(i[Y>>0]|0)+-48|0;if(_>>>0<10?(i[ia+2>>0]|0)==36:0){k[e+(_<<2)>>2]=10;ma=1;na=ia+3|0;oa=k[d+((i[Y>>0]|0)+-48<<3)>>2]|0}else{if(da){pa=-1;break a}if(!v){qa=ga;ra=Y;sa=0;ta=0;break}_=(k[c>>2]|0)+(4-1)&~(4-1);$=k[_>>2]|0;k[c>>2]=_+4;ma=0;na=Y;oa=$}if((oa|0)<0){qa=ga|8192;ra=na;sa=ma;ta=0-oa|0}else{qa=ga;ra=na;sa=ma;ta=oa}}else{$=(fa<<24>>24)+-48|0;if($>>>0<10){Y=ia;_=0;aa=$;while(1){$=(_*10|0)+aa|0;Z=Y+1|0;aa=(i[Z>>0]|0)+-48|0;if(aa>>>0>=10){ua=$;va=Z;break}else{Y=Z;_=$}}if((ua|0)<0){pa=-1;break a}else{qa=ga;ra=va;sa=da;ta=ua}}else{qa=ga;ra=ia;sa=da;ta=0}}while(0);e:do if((i[ra>>0]|0)==46){_=ra+1|0;Y=i[_>>0]|0;if(Y<<24>>24!=42){aa=(Y<<24>>24)+-48|0;if(aa>>>0<10){wa=_;xa=0;ya=aa}else{za=_;Aa=0;break}while(1){_=(xa*10|0)+ya|0;aa=wa+1|0;ya=(i[aa>>0]|0)+-48|0;if(ya>>>0>=10){za=aa;Aa=_;break e}else{wa=aa;xa=_}}}_=ra+2|0;aa=(i[_>>0]|0)+-48|0;if(aa>>>0<10?(i[ra+3>>0]|0)==36:0){k[e+(aa<<2)>>2]=10;za=ra+4|0;Aa=k[d+((i[_>>0]|0)+-48<<3)>>2]|0;break}if(sa){pa=-1;break a}if(v){aa=(k[c>>2]|0)+(4-1)&~(4-1);Y=k[aa>>2]|0;k[c>>2]=aa+4;za=_;Aa=Y}else{za=_;Aa=0}}else{za=ra;Aa=-1}while(0);_=za;Y=0;while(1){aa=(i[_>>0]|0)+-65|0;if(aa>>>0>57){pa=-1;break a}$=_+1|0;Z=i[795059+(Y*58|0)+aa>>0]|0;aa=Z&255;if((aa+-1|0)>>>0<8){_=$;Y=aa}else{Ba=$;Ca=Z;Da=aa;Ea=_;Fa=Y;break}}if(!(Ca<<24>>24)){pa=-1;break}Y=(ca|0)>-1;do if(Ca<<24>>24==19)if(Y){pa=-1;break a}else R=52;else{if(Y){k[e+(ca<<2)>>2]=Da;_=d+(ca<<3)|0;aa=k[_+4>>2]|0;Z=o;k[Z>>2]=k[_>>2];k[Z+4>>2]=aa;R=52;break}if(!v){pa=0;break a}$x(o,Da,c)}while(0);if((R|0)==52?(R=0,!v):0){J=O;K=Ba;b=N;M=sa;continue}Y=i[Ea>>0]|0;aa=(Fa|0)!=0&(Y&15|0)==3?Y&-33:Y;Y=qa&-65537;Z=(qa&8192|0)==0?qa:Y;f:do switch(aa|0){case 110:{switch(Fa|0){case 0:{k[k[o>>2]>>2]=O;J=O;K=Ba;b=N;M=sa;continue a;break}case 1:{k[k[o>>2]>>2]=O;J=O;K=Ba;b=N;M=sa;continue a;break}case 2:{_=k[o>>2]|0;k[_>>2]=O;k[_+4>>2]=((O|0)<0)<<31>>31;J=O;K=Ba;b=N;M=sa;continue a;break}case 3:{j[k[o>>2]>>1]=O;J=O;K=Ba;b=N;M=sa;continue a;break}case 4:{i[k[o>>2]>>0]=O;J=O;K=Ba;b=N;M=sa;continue a;break}case 6:{k[k[o>>2]>>2]=O;J=O;K=Ba;b=N;M=sa;continue a;break}case 7:{_=k[o>>2]|0;k[_>>2]=O;k[_+4>>2]=((O|0)<0)<<31>>31;J=O;K=Ba;b=N;M=sa;continue a;break}default:{J=O;K=Ba;b=N;M=sa;continue a}}break}case 112:{Ga=Z|8;Ha=Aa>>>0>8?Aa:8;Ia=120;R=64;break}case 88:case 120:{Ga=Z;Ha=Aa;Ia=aa;R=64;break}case 111:{_=o;$=k[_>>2]|0;ka=k[_+4>>2]|0;if(($|0)==0&(ka|0)==0)Ja=w;else{_=w;ja=$;$=ka;while(1){ka=_+-1|0;i[ka>>0]=ja&7|48;ja=ny(ja|0,$|0,3)|0;$=L;if((ja|0)==0&($|0)==0){Ja=ka;break}else _=ka}}if(!(Z&8)){Ka=Ja;La=Z;Ma=Aa;Na=0;Oa=795539;R=77}else{_=x-Ja+1|0;Ka=Ja;La=Z;Ma=(Aa|0)<(_|0)?_:Aa;Na=0;Oa=795539;R=77}break}case 105:case 100:{_=o;$=k[_>>2]|0;ja=k[_+4>>2]|0;if((ja|0)<0){_=jy(0,0,$|0,ja|0)|0;ka=L;la=o;k[la>>2]=_;k[la+4>>2]=ka;Pa=_;Qa=ka;Ra=1;Sa=795539;R=76;break f}if(!(Z&2048)){ka=Z&1;Pa=$;Qa=ja;Ra=ka;Sa=(ka|0)==0?795539:795541;R=76}else{Pa=$;Qa=ja;Ra=1;Sa=795540;R=76}break}case 117:{ja=o;Pa=k[ja>>2]|0;Qa=k[ja+4>>2]|0;Ra=0;Sa=795539;R=76;break}case 99:{i[y>>0]=k[o>>2];Ta=y;Ua=Y;Va=1;Wa=0;Xa=795539;Ya=w;break}case 109:{ja=rv()|0;Za=sv(k[ja>>2]|0)|0;R=82;break}case 115:{ja=k[o>>2]|0;Za=(ja|0)!=0?ja:795549;R=82;break}case 67:{k[s>>2]=k[o>>2];k[q>>2]=0;k[o>>2]=s;_a=-1;R=86;break}case 83:{if(!Aa){by(a,32,ta,0,Z);$a=0;R=98}else{_a=Aa;R=86}break}case 65:case 71:case 70:case 69:case 97:case 103:case 102:case 101:{ab=+p[o>>3];k[h>>2]=0;p[t>>3]=ab;if((k[t+4>>2]|0)>=0)if(!(Z&2048)){ja=Z&1;bb=ab;cb=ja;db=(ja|0)==0?795557:795562}else{bb=ab;cb=1;db=795559}else{bb=-ab;cb=1;db=795556}p[t>>3]=bb;ja=k[t+4>>2]&2146435072;do if(ja>>>0<2146435072|(ja|0)==2146435072&0<0){ab=+Jv(bb,h)*2.0;$=ab!=0.0;if($)k[h>>2]=(k[h>>2]|0)+-1;ka=aa|32;if((ka|0)==97){_=aa&32;la=(_|0)==0?db:db+9|0;eb=cb|2;fb=12-Aa|0;do if(!(Aa>>>0>11|(fb|0)==0)){gb=fb;hb=8.0;while(1){gb=gb+-1|0;ib=hb*16.0;if(!gb){jb=ib;break}else hb=ib}if((i[la>>0]|0)==45){kb=-(jb+(-ab-jb));break}else{kb=ab+jb-jb;break}}else kb=ab;while(0);fb=k[h>>2]|0;gb=(fb|0)<0?0-fb|0:fb;lb=ay(gb,((gb|0)<0)<<31>>31,z)|0;if((lb|0)==(z|0)){i[A>>0]=48;mb=A}else mb=lb;i[mb+-1>>0]=(fb>>31&2)+43;fb=mb+-2|0;i[fb>>0]=aa+15;lb=(Aa|0)<1;gb=(Z&8|0)==0;hb=kb;nb=m;while(1){ob=~~hb;pb=nb+1|0;i[nb>>0]=l[795523+ob>>0]|_;hb=(hb-+(ob|0))*16.0;do if((pb-n|0)==1){if(gb&(lb&hb==0.0)){qb=pb;break}i[pb>>0]=46;qb=nb+2|0}else qb=pb;while(0);if(!(hb!=0.0)){rb=qb;break}else nb=qb}nb=rb;lb=(Aa|0)!=0&(D+nb|0)<(Aa|0)?E+Aa-fb|0:C-fb+nb|0;gb=lb+eb|0;by(a,32,ta,gb,Z);if(!(k[a>>2]&32))Ow(la,eb,a)|0;by(a,48,ta,gb,Z^65536);_=nb-n|0;if(!(k[a>>2]&32))Ow(m,_,a)|0;nb=B-fb|0;by(a,48,lb-(_+nb)|0,0,0);if(!(k[a>>2]&32))Ow(fb,nb,a)|0;by(a,32,ta,gb,Z^8192);sb=(gb|0)<(ta|0)?ta:gb;break}gb=(Aa|0)<0?6:Aa;if($){nb=(k[h>>2]|0)+-28|0;k[h>>2]=nb;tb=ab*268435456.0;ub=nb}else{tb=ab;ub=k[h>>2]|0}nb=(ub|0)<0?g:F;_=nb;hb=tb;lb=nb;while(1){pb=~~hb>>>0;k[lb>>2]=pb;ob=lb+4|0;hb=(hb-+(pb>>>0))*1.0e9;if(!(hb!=0.0)){vb=ob;break}else lb=ob}lb=k[h>>2]|0;if((lb|0)>0){$=lb;fb=nb;eb=vb;while(1){la=($|0)>29?29:$;ob=eb+-4|0;do if(ob>>>0<fb>>>0)wb=fb;else{pb=0;xb=ob;while(1){yb=ty(k[xb>>2]|0,0,la|0)|0;zb=ky(yb|0,L|0,pb|0,0)|0;yb=L;Ab=Ay(zb|0,yb|0,1e9,0)|0;k[xb>>2]=Ab;Ab=zy(zb|0,yb|0,1e9,0)|0;xb=xb+-4|0;if(xb>>>0<fb>>>0){Bb=Ab;break}else pb=Ab}if(!Bb){wb=fb;break}pb=fb+-4|0;k[pb>>2]=Bb;wb=pb}while(0);ob=eb;while(1){if(ob>>>0<=wb>>>0){Cb=ob;break}pb=ob+-4|0;if(!(k[pb>>2]|0))ob=pb;else{Cb=ob;break}}ob=(k[h>>2]|0)-la|0;k[h>>2]=ob;if((ob|0)>0){$=ob;fb=wb;eb=Cb}else{Db=ob;Eb=wb;Fb=Cb;break}}}else{Db=lb;Eb=nb;Fb=vb}if((Db|0)<0){eb=((gb+25|0)/9|0)+1|0;fb=(ka|0)==102;$=Db;ob=Eb;pb=Fb;while(1){xb=0-$|0;Ab=(xb|0)>9?9:xb;do if(ob>>>0<pb>>>0){xb=(1<<Ab)+-1|0;yb=1e9>>>Ab;zb=0;Gb=ob;while(1){Hb=k[Gb>>2]|0;k[Gb>>2]=(Hb>>>Ab)+zb;Ib=ha(Hb&xb,yb)|0;Gb=Gb+4|0;if(Gb>>>0>=pb>>>0){Jb=Ib;break}else zb=Ib}zb=(k[ob>>2]|0)==0?ob+4|0:ob;if(!Jb){Kb=zb;Lb=pb;break}k[pb>>2]=Jb;Kb=zb;Lb=pb+4|0}else{Kb=(k[ob>>2]|0)==0?ob+4|0:ob;Lb=pb}while(0);la=fb?nb:Kb;zb=(Lb-la>>2|0)>(eb|0)?la+(eb<<2)|0:Lb;$=(k[h>>2]|0)+Ab|0;k[h>>2]=$;if(($|0)>=0){Mb=Kb;Nb=zb;break}else{ob=Kb;pb=zb}}}else{Mb=Eb;Nb=Fb}do if(Mb>>>0<Nb>>>0){pb=(_-Mb>>2)*9|0;ob=k[Mb>>2]|0;if(ob>>>0<10){Ob=pb;break}else{Pb=pb;Qb=10}while(1){Qb=Qb*10|0;pb=Pb+1|0;if(ob>>>0<Qb>>>0){Ob=pb;break}else Pb=pb}}else Ob=0;while(0);ob=(ka|0)==103;Ab=(gb|0)!=0;pb=gb-((ka|0)!=102?Ob:0)+((Ab&ob)<<31>>31)|0;if((pb|0)<(((Nb-_>>2)*9|0)+-9|0)){$=pb+9216|0;pb=($|0)/9|0;eb=nb+(pb+-1023<<2)|0;fb=(($|0)%9|0)+1|0;if((fb|0)<9){$=10;lb=fb;while(1){fb=$*10|0;lb=lb+1|0;if((lb|0)==9){Rb=fb;break}else $=fb}}else Rb=10;$=k[eb>>2]|0;lb=($>>>0)%(Rb>>>0)|0;if((lb|0)==0?(nb+(pb+-1022<<2)|0)==(Nb|0):0){Sb=Mb;Tb=eb;Ub=Ob}else R=163;do if((R|0)==163){R=0;hb=((($>>>0)/(Rb>>>0)|0)&1|0)==0?9007199254740992.0:9007199254740994.0;ka=(Rb|0)/2|0;do if(lb>>>0<ka>>>0)Vb=.5;else{if((lb|0)==(ka|0)?(nb+(pb+-1022<<2)|0)==(Nb|0):0){Vb=1.0;break}Vb=1.5}while(0);do if(!cb){Wb=hb;Xb=Vb}else{if((i[db>>0]|0)!=45){Wb=hb;Xb=Vb;break}Wb=-hb;Xb=-Vb}while(0);ka=$-lb|0;k[eb>>2]=ka;if(!(Wb+Xb!=Wb)){Sb=Mb;Tb=eb;Ub=Ob;break}fb=ka+Rb|0;k[eb>>2]=fb;if(fb>>>0>999999999){fb=Mb;ka=eb;while(1){zb=ka+-4|0;k[ka>>2]=0;if(zb>>>0<fb>>>0){la=fb+-4|0;k[la>>2]=0;Yb=la}else Yb=fb;la=(k[zb>>2]|0)+1|0;k[zb>>2]=la;if(la>>>0>999999999){fb=Yb;ka=zb}else{Zb=Yb;_b=zb;break}}}else{Zb=Mb;_b=eb}ka=(_-Zb>>2)*9|0;fb=k[Zb>>2]|0;if(fb>>>0<10){Sb=Zb;Tb=_b;Ub=ka;break}else{$b=ka;ac=10}while(1){ac=ac*10|0;ka=$b+1|0;if(fb>>>0<ac>>>0){Sb=Zb;Tb=_b;Ub=ka;break}else $b=ka}}while(0);eb=Tb+4|0;bc=Sb;cc=Ub;dc=Nb>>>0>eb>>>0?eb:Nb}else{bc=Mb;cc=Ob;dc=Nb}eb=0-cc|0;lb=dc;while(1){if(lb>>>0<=bc>>>0){ec=0;fc=lb;break}$=lb+-4|0;if(!(k[$>>2]|0))lb=$;else{ec=1;fc=lb;break}}do if(ob){lb=(Ab&1^1)+gb|0;if((lb|0)>(cc|0)&(cc|0)>-5){gc=aa+-1|0;hc=lb+-1-cc|0}else{gc=aa+-2|0;hc=lb+-1|0}lb=Z&8;if(lb){ic=gc;jc=hc;kc=lb;break}do if(ec){lb=k[fc+-4>>2]|0;if(!lb){lc=9;break}if(!((lb>>>0)%10|0)){mc=10;nc=0}else{lc=0;break}while(1){mc=mc*10|0;$=nc+1|0;if((lb>>>0)%(mc>>>0)|0){lc=$;break}else nc=$}}else lc=9;while(0);lb=((fc-_>>2)*9|0)+-9|0;if((gc|32|0)==102){$=lb-lc|0;pb=($|0)<0?0:$;ic=gc;jc=(hc|0)<(pb|0)?hc:pb;kc=0;break}else{pb=lb+cc-lc|0;lb=(pb|0)<0?0:pb;ic=gc;jc=(hc|0)<(lb|0)?hc:lb;kc=0;break}}else{ic=aa;jc=gb;kc=Z&8}while(0);gb=jc|kc;_=(gb|0)!=0&1;Ab=(ic|32|0)==102;if(Ab){oc=(cc|0)>0?cc:0;pc=0}else{ob=(cc|0)<0?eb:cc;lb=ay(ob,((ob|0)<0)<<31>>31,z)|0;if((B-lb|0)<2){ob=lb;while(1){pb=ob+-1|0;i[pb>>0]=48;if((B-pb|0)<2)ob=pb;else{qc=pb;break}}}else qc=lb;i[qc+-1>>0]=(cc>>31&2)+43;ob=qc+-2|0;i[ob>>0]=ic;oc=B-ob|0;pc=ob}ob=cb+1+jc+_+oc|0;by(a,32,ta,ob,Z);if(!(k[a>>2]&32))Ow(db,cb,a)|0;by(a,48,ta,ob,Z^65536);do if(Ab){eb=bc>>>0>nb>>>0?nb:bc;pb=eb;while(1){$=ay(k[pb>>2]|0,0,G)|0;do if((pb|0)==(eb|0)){if(($|0)!=(G|0)){rc=$;break}i[I>>0]=48;rc=I}else{if($>>>0>m>>>0)sc=$;else{rc=$;break}while(1){fb=sc+-1|0;i[fb>>0]=48;if(fb>>>0>m>>>0)sc=fb;else{rc=fb;break}}}while(0);if(!(k[a>>2]&32))Ow(rc,H-rc|0,a)|0;$=pb+4|0;if($>>>0>nb>>>0){tc=$;break}else pb=$}do if(gb){if(k[a>>2]&32)break;Ow(795591,1,a)|0}while(0);if((jc|0)>0&tc>>>0<fc>>>0){pb=jc;eb=tc;while(1){$=ay(k[eb>>2]|0,0,G)|0;if($>>>0>m>>>0){fb=$;while(1){ka=fb+-1|0;i[ka>>0]=48;if(ka>>>0>m>>>0)fb=ka;else{uc=ka;break}}}else uc=$;if(!(k[a>>2]&32))Ow(uc,(pb|0)>9?9:pb,a)|0;eb=eb+4|0;fb=pb+-9|0;if(!((pb|0)>9&eb>>>0<fc>>>0)){vc=fb;break}else pb=fb}}else vc=jc;by(a,48,vc+9|0,9,0)}else{pb=ec?fc:bc+4|0;if((jc|0)>-1){eb=(kc|0)==0;fb=jc;ka=bc;while(1){zb=ay(k[ka>>2]|0,0,G)|0;if((zb|0)==(G|0)){i[I>>0]=48;wc=I}else wc=zb;do if((ka|0)==(bc|0)){zb=wc+1|0;if(!(k[a>>2]&32))Ow(wc,1,a)|0;if(eb&(fb|0)<1){xc=zb;break}if(k[a>>2]&32){xc=zb;break}Ow(795591,1,a)|0;xc=zb}else{if(wc>>>0>m>>>0)yc=wc;else{xc=wc;break}while(1){zb=yc+-1|0;i[zb>>0]=48;if(zb>>>0>m>>>0)yc=zb;else{xc=zb;break}}}while(0);$=H-xc|0;if(!(k[a>>2]&32))Ow(xc,(fb|0)>($|0)?$:fb,a)|0;zb=fb-$|0;ka=ka+4|0;if(!(ka>>>0<pb>>>0&(zb|0)>-1)){zc=zb;break}else fb=zb}}else zc=jc;by(a,48,zc+18|0,18,0);if(k[a>>2]&32)break;Ow(pc,B-pc|0,a)|0}while(0);by(a,32,ta,ob,Z^8192);sb=(ob|0)<(ta|0)?ta:ob}else{gb=(aa&32|0)!=0;nb=bb!=bb|0.0!=0.0;Ab=nb?0:cb;_=Ab+3|0;by(a,32,ta,_,Y);lb=k[a>>2]|0;if(!(lb&32)){Ow(db,Ab,a)|0;Ac=k[a>>2]|0}else Ac=lb;if(!(Ac&32))Ow(nb?(gb?795583:795587):gb?795575:795579,3,a)|0;by(a,32,ta,_,Z^8192);sb=(_|0)<(ta|0)?ta:_}while(0);J=O;K=Ba;b=sb;M=sa;continue a;break}default:{Ta=K;Ua=Z;Va=Aa;Wa=0;Xa=795539;Ya=w}}while(0);g:do if((R|0)==64){R=0;aa=o;N=k[aa>>2]|0;ja=k[aa+4>>2]|0;aa=Ia&32;if(!((N|0)==0&(ja|0)==0)){_=w;gb=N;N=ja;while(1){ja=_+-1|0;i[ja>>0]=l[795523+(gb&15)>>0]|aa;gb=ny(gb|0,N|0,4)|0;N=L;if((gb|0)==0&(N|0)==0){Bc=ja;break}else _=ja}_=o;if((Ga&8|0)==0|(k[_>>2]|0)==0&(k[_+4>>2]|0)==0){Ka=Bc;La=Ga;Ma=Ha;Na=0;Oa=795539;R=77}else{Ka=Bc;La=Ga;Ma=Ha;Na=2;Oa=795539+(Ia>>4)|0;R=77}}else{Ka=w;La=Ga;Ma=Ha;Na=0;Oa=795539;R=77}}else if((R|0)==76){R=0;Ka=ay(Pa,Qa,w)|0;La=Z;Ma=Aa;Na=Ra;Oa=Sa;R=77}else if((R|0)==82){R=0;_=kx(Za,0,Aa)|0;N=(_|0)==0;Ta=Za;Ua=Y;Va=N?Aa:_-Za|0;Wa=0;Xa=795539;Ya=N?Za+Aa|0:_}else if((R|0)==86){R=0;_=0;N=0;gb=k[o>>2]|0;while(1){aa=k[gb>>2]|0;if(!aa){Cc=_;Dc=N;break}ja=Vv(u,aa)|0;if((ja|0)<0|ja>>>0>(_a-_|0)>>>0){Cc=_;Dc=ja;break}aa=ja+_|0;if(_a>>>0>aa>>>0){_=aa;N=ja;gb=gb+4|0}else{Cc=aa;Dc=ja;break}}if((Dc|0)<0){pa=-1;break a}by(a,32,ta,Cc,Z);if(!Cc){$a=0;R=98}else{gb=0;N=k[o>>2]|0;while(1){_=k[N>>2]|0;if(!_){$a=Cc;R=98;break g}ja=Vv(u,_)|0;gb=ja+gb|0;if((gb|0)>(Cc|0)){$a=Cc;R=98;break g}if(!(k[a>>2]&32))Ow(u,ja,a)|0;if(gb>>>0>=Cc>>>0){$a=Cc;R=98;break}else N=N+4|0}}}while(0);if((R|0)==98){R=0;by(a,32,ta,$a,Z^8192);J=O;K=Ba;b=(ta|0)>($a|0)?ta:$a;M=sa;continue}if((R|0)==77){R=0;Y=(Ma|0)>-1?La&-65537:La;N=o;gb=(k[N>>2]|0)!=0|(k[N+4>>2]|0)!=0;if((Ma|0)!=0|gb){N=(gb&1^1)+(x-Ka)|0;Ta=Ka;Ua=Y;Va=(Ma|0)>(N|0)?Ma:N;Wa=Na;Xa=Oa;Ya=w}else{Ta=w;Ua=Y;Va=0;Wa=Na;Xa=Oa;Ya=w}}Y=Ya-Ta|0;N=(Va|0)<(Y|0)?Y:Va;gb=Wa+N|0;ja=(ta|0)<(gb|0)?gb:ta;by(a,32,ja,gb,Ua);if(!(k[a>>2]&32))Ow(Xa,Wa,a)|0;by(a,48,ja,gb,Ua^65536);by(a,48,N,Y,0);if(!(k[a>>2]&32))Ow(Ta,Y,a)|0;by(a,32,ja,gb,Ua^8192);J=O;K=Ba;b=ja;M=sa}h:do if((R|0)==245)if(!a)if(Q){sa=1;while(1){M=k[e+(sa<<2)>>2]|0;if(!M){Ec=sa;break}$x(d+(sa<<3)|0,M,c);sa=sa+1|0;if((sa|0)>=10){pa=1;break h}}if((Ec|0)<10){sa=Ec;while(1){if(k[e+(sa<<2)>>2]|0){pa=-1;break h}sa=sa+1|0;if((sa|0)>=10){pa=1;break}}}else pa=1}else pa=0;else pa=P;while(0);r=f;return pa|0}function Ux(a,b,c){a=a|0;b=b|0;c=c|0;return sw(a,b,c)|0}function Vx(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0;f=r;r=r+240|0;g=f;k[g>>2]=a;a:do if((d|0)>1){h=0-b|0;i=d;j=a;l=a;m=1;while(1){n=j+h|0;o=i+-2|0;p=j+(0-((k[e+(o<<2)>>2]|0)+b))|0;if((Xb[c&31](l,p)|0)>-1?(Xb[c&31](l,n)|0)>-1:0){q=m;break a}s=m+1|0;t=g+(m<<2)|0;if((Xb[c&31](p,n)|0)>-1){k[t>>2]=p;u=p;v=i+-1|0}else{k[t>>2]=n;u=n;v=o}if((v|0)<=1){q=s;break a}i=v;j=u;l=k[g>>2]|0;m=s}}else q=1;while(0);Zx(b,g,q);r=f;return}function Wx(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0;h=r;r=r+240|0;i=h;j=k[d>>2]|0;l=k[d+4>>2]|0;k[i>>2]=a;d=0-b|0;a:do if((l|0)!=0|(j|0)!=1?(m=a+(0-(k[g+(e<<2)>>2]|0))|0,(Xb[c&31](m,a)|0)>=1):0){n=a;o=e;p=(f|0)==0;q=m;m=j;s=l;t=1;while(1){if(p&(o|0)>1){u=k[g+(o+-2<<2)>>2]|0;if((Xb[c&31](n+d|0,q)|0)>-1){v=n;w=o;x=t;y=20;break a}if((Xb[c&31](n+(0-(u+b))|0,q)|0)>-1){v=n;w=o;x=t;y=20;break a}}u=t+1|0;k[i+(t<<2)>>2]=q;z=m+-1|0;do if(z){if(!(z&1)){A=z;B=0;while(1){C=B+1|0;A=A>>>1;if(A&1){D=C;break}else B=C}if(D)E=D;else y=11}else y=11;if((y|0)==11){y=0;if(!s){F=64;y=16;break}if(!(s&1)){G=s;H=0}else{I=0;J=m;K=s;L=0;break}while(1){B=H+1|0;G=G>>>1;if(G&1){M=B;N=H;break}else H=B}if(!M){I=0;J=m;K=s;L=0;break}else E=N+33|0}if(E>>>0>31){F=E;y=16}else{I=E;J=m;K=s;L=E}}else{F=32;y=16}while(0);if((y|0)==16){y=0;I=F+-32|0;J=s;K=0;L=F}m=K<<32-I|J>>>I;s=K>>>I;z=L+o|0;if(!((s|0)!=0|(m|0)!=1)){v=q;w=z;x=u;y=20;break a}B=q+(0-(k[g+(z<<2)>>2]|0))|0;if((Xb[c&31](B,k[i>>2]|0)|0)<1){O=q;P=z;Q=0;R=u;y=19;break}else{A=q;o=z;p=1;q=B;t=u;n=A}}}else{O=a;P=e;Q=f;R=1;y=19}while(0);if((y|0)==19?(Q|0)==0:0){v=O;w=P;x=R;y=20}if((y|0)==20){Zx(b,i,x);Vx(v,b,c,w,g)}r=h;return}function Xx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0.0;d=r;r=r+112|0;e=d;f=e;g=f+112|0;do{k[f>>2]=0;f=f+4|0}while((f|0)<(g|0));f=e+4|0;k[f>>2]=a;g=e+8|0;k[g>>2]=-1;k[e+44>>2]=a;k[e+76>>2]=-1;xv(e,0);h=+vv(e,c,1);c=(k[f>>2]|0)-(k[g>>2]|0)+(k[e+108>>2]|0)|0;if(b)k[b>>2]=(c|0)!=0?a+c|0:a;r=d;return +h}function Yx(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0;f=r;r=r+112|0;g=f;k[g>>2]=0;h=g+4|0;k[h>>2]=a;k[g+44>>2]=a;i=g+8|0;k[i>>2]=(a|0)<0?-1:a+2147483647|0;k[g+76>>2]=-1;xv(g,0);j=wv(g,c,1,d,e)|0;if(b)k[b>>2]=a+((k[h>>2]|0)+(k[g+108>>2]|0)-(k[i>>2]|0));r=f;return j|0}function Zx(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;d=r;r=r+256|0;e=d;a:do if((c|0)>=2?(f=b+(c<<2)|0,k[f>>2]=e,(a|0)!=0):0){g=a;h=e;while(1){i=g>>>0>256?256:g;py(h|0,k[b>>2]|0,i|0)|0;j=0;do{l=b+(j<<2)|0;j=j+1|0;py(k[l>>2]|0,k[b+(j<<2)>>2]|0,i|0)|0;k[l>>2]=(k[l>>2]|0)+i}while((j|0)!=(c|0));if((g|0)==(i|0))break a;g=g-i|0;h=k[f>>2]|0}}while(0);r=d;return}function _x(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0;d=a+20|0;e=k[d>>2]|0;f=(k[a+16>>2]|0)-e|0;a=f>>>0>c>>>0?c:f;py(e|0,b|0,a|0)|0;k[d>>2]=(k[d>>2]|0)+a;return c|0}function $x(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0.0;a:do if(b>>>0<=20)do switch(b|0){case 9:{d=(k[c>>2]|0)+(4-1)&~(4-1);e=k[d>>2]|0;k[c>>2]=d+4;k[a>>2]=e;break a;break}case 10:{e=(k[c>>2]|0)+(4-1)&~(4-1);d=k[e>>2]|0;k[c>>2]=e+4;e=a;k[e>>2]=d;k[e+4>>2]=((d|0)<0)<<31>>31;break a;break}case 11:{d=(k[c>>2]|0)+(4-1)&~(4-1);e=k[d>>2]|0;k[c>>2]=d+4;d=a;k[d>>2]=e;k[d+4>>2]=0;break a;break}case 12:{d=(k[c>>2]|0)+(8-1)&~(8-1);e=d;f=k[e>>2]|0;g=k[e+4>>2]|0;k[c>>2]=d+8;d=a;k[d>>2]=f;k[d+4>>2]=g;break a;break}case 13:{g=(k[c>>2]|0)+(4-1)&~(4-1);d=k[g>>2]|0;k[c>>2]=g+4;g=(d&65535)<<16>>16;d=a;k[d>>2]=g;k[d+4>>2]=((g|0)<0)<<31>>31;break a;break}case 14:{g=(k[c>>2]|0)+(4-1)&~(4-1);d=k[g>>2]|0;k[c>>2]=g+4;g=a;k[g>>2]=d&65535;k[g+4>>2]=0;break a;break}case 15:{g=(k[c>>2]|0)+(4-1)&~(4-1);d=k[g>>2]|0;k[c>>2]=g+4;g=(d&255)<<24>>24;d=a;k[d>>2]=g;k[d+4>>2]=((g|0)<0)<<31>>31;break a;break}case 16:{g=(k[c>>2]|0)+(4-1)&~(4-1);d=k[g>>2]|0;k[c>>2]=g+4;g=a;k[g>>2]=d&255;k[g+4>>2]=0;break a;break}case 17:{g=(k[c>>2]|0)+(8-1)&~(8-1);h=+p[g>>3];k[c>>2]=g+8;p[a>>3]=h;break a;break}case 18:{g=(k[c>>2]|0)+(8-1)&~(8-1);h=+p[g>>3];k[c>>2]=g+8;p[a>>3]=h;break a;break}default:break a}while(0);while(0);return}function ay(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0;if(b>>>0>0|(b|0)==0&a>>>0>4294967295){d=c;e=a;f=b;while(1){b=Ay(e|0,f|0,10,0)|0;g=d+-1|0;i[g>>0]=b|48;b=zy(e|0,f|0,10,0)|0;if(f>>>0>9|(f|0)==9&e>>>0>4294967295){d=g;e=b;f=L}else{h=g;j=b;break}}k=h;l=j}else{k=c;l=a}if(!l)m=k;else{a=k;k=l;while(1){l=a+-1|0;i[l>>0]=(k>>>0)%10|0|48;if(k>>>0<10){m=l;break}else{a=l;k=(k>>>0)/10|0}}}return m|0}function by(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0;f=r;r=r+256|0;g=f;do if((c|0)>(d|0)&(e&73728|0)==0){h=c-d|0;sy(g|0,b|0,(h>>>0>256?256:h)|0)|0;i=k[a>>2]|0;j=(i&32|0)==0;if(h>>>0>255){l=c-d|0;m=h;n=i;i=j;while(1){if(i){Ow(g,256,a)|0;o=k[a>>2]|0}else o=n;m=m+-256|0;i=(o&32|0)==0;if(m>>>0<=255)break;else n=o}if(i)p=l&255;else break}else if(j)p=h;else break;Ow(g,p,a)|0}while(0);r=f;return}function cy(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Sa=0,Ta=0;do if(a>>>0<245){b=a>>>0<11?16:a+11&-8;c=b>>>3;d=k[31370]|0;e=d>>>c;if(e&3){f=(e&1^1)+c|0;g=f<<1;h=125520+(g<<2)|0;i=125520+(g+2<<2)|0;g=k[i>>2]|0;j=g+8|0;l=k[j>>2]|0;do if((h|0)!=(l|0)){if(l>>>0<(k[31374]|0)>>>0)xb();m=l+12|0;if((k[m>>2]|0)==(g|0)){k[m>>2]=h;k[i>>2]=l;break}else xb()}else k[31370]=d&~(1<<f);while(0);l=f<<3;k[g+4>>2]=l|3;i=g+(l|4)|0;k[i>>2]=k[i>>2]|1;n=j;return n|0}i=k[31372]|0;if(b>>>0>i>>>0){if(e){l=2<<c;h=e<<c&(l|0-l);l=(h&0-h)+-1|0;h=l>>>12&16;m=l>>>h;l=m>>>5&8;o=m>>>l;m=o>>>2&4;p=o>>>m;o=p>>>1&2;q=p>>>o;p=q>>>1&1;r=(l|h|m|o|p)+(q>>>p)|0;p=r<<1;q=125520+(p<<2)|0;o=125520+(p+2<<2)|0;p=k[o>>2]|0;m=p+8|0;h=k[m>>2]|0;do if((q|0)!=(h|0)){if(h>>>0<(k[31374]|0)>>>0)xb();l=h+12|0;if((k[l>>2]|0)==(p|0)){k[l>>2]=q;k[o>>2]=h;s=k[31372]|0;break}else xb()}else{k[31370]=d&~(1<<r);s=i}while(0);i=r<<3;d=i-b|0;k[p+4>>2]=b|3;h=p+b|0;k[p+(b|4)>>2]=d|1;k[p+i>>2]=d;if(s){i=k[31375]|0;o=s>>>3;q=o<<1;c=125520+(q<<2)|0;e=k[31370]|0;j=1<<o;if(e&j){o=125520+(q+2<<2)|0;g=k[o>>2]|0;if(g>>>0<(k[31374]|0)>>>0)xb();else{t=o;u=g}}else{k[31370]=e|j;t=125520+(q+2<<2)|0;u=c}k[t>>2]=i;k[u+12>>2]=i;k[i+8>>2]=u;k[i+12>>2]=c}k[31372]=d;k[31375]=h;n=m;return n|0}h=k[31371]|0;if(h){d=(h&0-h)+-1|0;h=d>>>12&16;c=d>>>h;d=c>>>5&8;i=c>>>d;c=i>>>2&4;q=i>>>c;i=q>>>1&2;j=q>>>i;q=j>>>1&1;e=k[125784+((d|h|c|i|q)+(j>>>q)<<2)>>2]|0;q=(k[e+4>>2]&-8)-b|0;j=e;i=e;while(1){e=k[j+16>>2]|0;if(!e){c=k[j+20>>2]|0;if(!c){v=q;w=i;break}else x=c}else x=e;e=(k[x+4>>2]&-8)-b|0;c=e>>>0<q>>>0;q=c?e:q;j=x;i=c?x:i}i=k[31374]|0;if(w>>>0<i>>>0)xb();j=w+b|0;if(w>>>0>=j>>>0)xb();q=k[w+24>>2]|0;m=k[w+12>>2]|0;do if((m|0)==(w|0)){p=w+20|0;r=k[p>>2]|0;if(!r){c=w+16|0;e=k[c>>2]|0;if(!e){y=0;break}else{z=e;A=c}}else{z=r;A=p}while(1){p=z+20|0;r=k[p>>2]|0;if(r){z=r;A=p;continue}p=z+16|0;r=k[p>>2]|0;if(!r){B=z;C=A;break}else{z=r;A=p}}if(C>>>0<i>>>0)xb();else{k[C>>2]=0;y=B;break}}else{p=k[w+8>>2]|0;if(p>>>0<i>>>0)xb();r=p+12|0;if((k[r>>2]|0)!=(w|0))xb();c=m+8|0;if((k[c>>2]|0)==(w|0)){k[r>>2]=m;k[c>>2]=p;y=m;break}else xb()}while(0);do if(q){m=k[w+28>>2]|0;i=125784+(m<<2)|0;if((w|0)==(k[i>>2]|0)){k[i>>2]=y;if(!y){k[31371]=k[31371]&~(1<<m);break}}else{if(q>>>0<(k[31374]|0)>>>0)xb();m=q+16|0;if((k[m>>2]|0)==(w|0))k[m>>2]=y;else k[q+20>>2]=y;if(!y)break}m=k[31374]|0;if(y>>>0<m>>>0)xb();k[y+24>>2]=q;i=k[w+16>>2]|0;do if(i)if(i>>>0<m>>>0)xb();else{k[y+16>>2]=i;k[i+24>>2]=y;break}while(0);i=k[w+20>>2]|0;if(i)if(i>>>0<(k[31374]|0)>>>0)xb();else{k[y+20>>2]=i;k[i+24>>2]=y;break}}while(0);if(v>>>0<16){q=v+b|0;k[w+4>>2]=q|3;i=w+(q+4)|0;k[i>>2]=k[i>>2]|1}else{k[w+4>>2]=b|3;k[w+(b|4)>>2]=v|1;k[w+(v+b)>>2]=v;i=k[31372]|0;if(i){q=k[31375]|0;m=i>>>3;i=m<<1;p=125520+(i<<2)|0;c=k[31370]|0;r=1<<m;if(c&r){m=125520+(i+2<<2)|0;e=k[m>>2]|0;if(e>>>0<(k[31374]|0)>>>0)xb();else{D=m;E=e}}else{k[31370]=c|r;D=125520+(i+2<<2)|0;E=p}k[D>>2]=q;k[E+12>>2]=q;k[q+8>>2]=E;k[q+12>>2]=p}k[31372]=v;k[31375]=j}n=w+8|0;return n|0}else F=b}else F=b}else if(a>>>0<=4294967231){p=a+11|0;q=p&-8;i=k[31371]|0;if(i){r=0-q|0;c=p>>>8;if(c)if(q>>>0>16777215)G=31;else{p=(c+1048320|0)>>>16&8;e=c<<p;c=(e+520192|0)>>>16&4;m=e<<c;e=(m+245760|0)>>>16&2;h=14-(c|p|e)+(m<<e>>>15)|0;G=q>>>(h+7|0)&1|h<<1}else G=0;h=k[125784+(G<<2)>>2]|0;a:do if(!h){H=r;I=0;J=0;K=86}else{e=r;m=0;p=q<<((G|0)==31?0:25-(G>>>1)|0);c=h;d=0;while(1){g=k[c+4>>2]&-8;o=g-q|0;if(o>>>0<e>>>0)if((g|0)==(q|0)){L=o;M=c;N=c;K=90;break a}else{O=o;P=c}else{O=e;P=d}o=k[c+20>>2]|0;c=k[c+16+(p>>>31<<2)>>2]|0;g=(o|0)==0|(o|0)==(c|0)?m:o;if(!c){H=O;I=g;J=P;K=86;break}else{e=O;m=g;p=p<<1;d=P}}}while(0);if((K|0)==86){if((I|0)==0&(J|0)==0){h=2<<G;r=i&(h|0-h);if(!r){F=q;break}h=(r&0-r)+-1|0;r=h>>>12&16;b=h>>>r;h=b>>>5&8;j=b>>>h;b=j>>>2&4;d=j>>>b;j=d>>>1&2;p=d>>>j;d=p>>>1&1;Q=k[125784+((h|r|b|j|d)+(p>>>d)<<2)>>2]|0;R=0}else{Q=I;R=J}if(!Q){S=H;T=R}else{L=H;M=Q;N=R;K=90}}if((K|0)==90)while(1){K=0;d=(k[M+4>>2]&-8)-q|0;p=d>>>0<L>>>0;j=p?d:L;d=p?M:N;p=k[M+16>>2]|0;if(p){L=j;M=p;N=d;K=90;continue}M=k[M+20>>2]|0;if(!M){S=j;T=d;break}else{L=j;N=d;K=90}}if((T|0)!=0?S>>>0<((k[31372]|0)-q|0)>>>0:0){i=k[31374]|0;if(T>>>0<i>>>0)xb();d=T+q|0;if(T>>>0>=d>>>0)xb();j=k[T+24>>2]|0;p=k[T+12>>2]|0;do if((p|0)==(T|0)){b=T+20|0;r=k[b>>2]|0;if(!r){h=T+16|0;m=k[h>>2]|0;if(!m){U=0;break}else{V=m;W=h}}else{V=r;W=b}while(1){b=V+20|0;r=k[b>>2]|0;if(r){V=r;W=b;continue}b=V+16|0;r=k[b>>2]|0;if(!r){X=V;Y=W;break}else{V=r;W=b}}if(Y>>>0<i>>>0)xb();else{k[Y>>2]=0;U=X;break}}else{b=k[T+8>>2]|0;if(b>>>0<i>>>0)xb();r=b+12|0;if((k[r>>2]|0)!=(T|0))xb();h=p+8|0;if((k[h>>2]|0)==(T|0)){k[r>>2]=p;k[h>>2]=b;U=p;break}else xb()}while(0);do if(j){p=k[T+28>>2]|0;i=125784+(p<<2)|0;if((T|0)==(k[i>>2]|0)){k[i>>2]=U;if(!U){k[31371]=k[31371]&~(1<<p);break}}else{if(j>>>0<(k[31374]|0)>>>0)xb();p=j+16|0;if((k[p>>2]|0)==(T|0))k[p>>2]=U;else k[j+20>>2]=U;if(!U)break}p=k[31374]|0;if(U>>>0<p>>>0)xb();k[U+24>>2]=j;i=k[T+16>>2]|0;do if(i)if(i>>>0<p>>>0)xb();else{k[U+16>>2]=i;k[i+24>>2]=U;break}while(0);i=k[T+20>>2]|0;if(i)if(i>>>0<(k[31374]|0)>>>0)xb();else{k[U+20>>2]=i;k[i+24>>2]=U;break}}while(0);b:do if(S>>>0>=16){k[T+4>>2]=q|3;k[T+(q|4)>>2]=S|1;k[T+(S+q)>>2]=S;j=S>>>3;if(S>>>0<256){i=j<<1;p=125520+(i<<2)|0;b=k[31370]|0;h=1<<j;if(b&h){j=125520+(i+2<<2)|0;r=k[j>>2]|0;if(r>>>0<(k[31374]|0)>>>0)xb();else{Z=j;_=r}}else{k[31370]=b|h;Z=125520+(i+2<<2)|0;_=p}k[Z>>2]=d;k[_+12>>2]=d;k[T+(q+8)>>2]=_;k[T+(q+12)>>2]=p;break}p=S>>>8;if(p)if(S>>>0>16777215)$=31;else{i=(p+1048320|0)>>>16&8;h=p<<i;p=(h+520192|0)>>>16&4;b=h<<p;h=(b+245760|0)>>>16&2;r=14-(p|i|h)+(b<<h>>>15)|0;$=S>>>(r+7|0)&1|r<<1}else $=0;r=125784+($<<2)|0;k[T+(q+28)>>2]=$;k[T+(q+20)>>2]=0;k[T+(q+16)>>2]=0;h=k[31371]|0;b=1<<$;if(!(h&b)){k[31371]=h|b;k[r>>2]=d;k[T+(q+24)>>2]=r;k[T+(q+12)>>2]=d;k[T+(q+8)>>2]=d;break}b=k[r>>2]|0;c:do if((k[b+4>>2]&-8|0)!=(S|0)){r=S<<(($|0)==31?0:25-($>>>1)|0);h=b;while(1){i=h+16+(r>>>31<<2)|0;p=k[i>>2]|0;if(!p){aa=i;ba=h;break}if((k[p+4>>2]&-8|0)==(S|0)){ca=p;break c}else{r=r<<1;h=p}}if(aa>>>0<(k[31374]|0)>>>0)xb();else{k[aa>>2]=d;k[T+(q+24)>>2]=ba;k[T+(q+12)>>2]=d;k[T+(q+8)>>2]=d;break b}}else ca=b;while(0);b=ca+8|0;h=k[b>>2]|0;r=k[31374]|0;if(h>>>0>=r>>>0&ca>>>0>=r>>>0){k[h+12>>2]=d;k[b>>2]=d;k[T+(q+8)>>2]=h;k[T+(q+12)>>2]=ca;k[T+(q+24)>>2]=0;break}else xb()}else{h=S+q|0;k[T+4>>2]=h|3;b=T+(h+4)|0;k[b>>2]=k[b>>2]|1}while(0);n=T+8|0;return n|0}else F=q}else F=q}else F=-1;while(0);T=k[31372]|0;if(T>>>0>=F>>>0){S=T-F|0;ca=k[31375]|0;if(S>>>0>15){k[31375]=ca+F;k[31372]=S;k[ca+(F+4)>>2]=S|1;k[ca+T>>2]=S;k[ca+4>>2]=F|3}else{k[31372]=0;k[31375]=0;k[ca+4>>2]=T|3;S=ca+(T+4)|0;k[S>>2]=k[S>>2]|1}n=ca+8|0;return n|0}ca=k[31373]|0;if(ca>>>0>F>>>0){S=ca-F|0;k[31373]=S;ca=k[31376]|0;k[31376]=ca+F;k[ca+(F+4)>>2]=S|1;k[ca+4>>2]=F|3;n=ca+8|0;return n|0}do if(!(k[31488]|0)){ca=vb(30)|0;if(!(ca+-1&ca)){k[31490]=ca;k[31489]=ca;k[31491]=-1;k[31492]=-1;k[31493]=0;k[31481]=0;ca=(Ra(0)|0)&-16^1431655768;k[31488]=ca;break}else xb()}while(0);ca=F+48|0;S=k[31490]|0;T=F+47|0;ba=S+T|0;aa=0-S|0;S=ba&aa;if(S>>>0<=F>>>0){n=0;return n|0}$=k[31480]|0;if(($|0)!=0?(_=k[31478]|0,Z=_+S|0,Z>>>0<=_>>>0|Z>>>0>$>>>0):0){n=0;return n|0}d:do if(!(k[31481]&4)){$=k[31376]|0;e:do if($){Z=125928;while(1){_=k[Z>>2]|0;if(_>>>0<=$>>>0?(U=Z+4|0,(_+(k[U>>2]|0)|0)>>>0>$>>>0):0){da=Z;ea=U;break}Z=k[Z+8>>2]|0;if(!Z){K=174;break e}}Z=ba-(k[31373]|0)&aa;if(Z>>>0<2147483647){U=Wa(Z|0)|0;_=(U|0)==((k[da>>2]|0)+(k[ea>>2]|0)|0);X=_?Z:0;if(_)if((U|0)==(-1|0))fa=X;else{ga=U;ha=X;K=194;break d}else{ia=U;ja=Z;ka=X;K=184}}else fa=0}else K=174;while(0);do if((K|0)==174){$=Wa(0)|0;if(($|0)!=(-1|0)){q=$;X=k[31489]|0;Z=X+-1|0;if(!(Z&q))la=S;else la=S-q+(Z+q&0-X)|0;X=k[31478]|0;q=X+la|0;if(la>>>0>F>>>0&la>>>0<2147483647){Z=k[31480]|0;if((Z|0)!=0?q>>>0<=X>>>0|q>>>0>Z>>>0:0){fa=0;break}Z=Wa(la|0)|0;q=(Z|0)==($|0);X=q?la:0;if(q){ga=$;ha=X;K=194;break d}else{ia=Z;ja=la;ka=X;K=184}}else fa=0}else fa=0}while(0);f:do if((K|0)==184){X=0-ja|0;do if(ca>>>0>ja>>>0&(ja>>>0<2147483647&(ia|0)!=(-1|0))?(Z=k[31490]|0,$=T-ja+Z&0-Z,$>>>0<2147483647):0)if((Wa($|0)|0)==(-1|0)){Wa(X|0)|0;fa=ka;break f}else{ma=$+ja|0;break}else ma=ja;while(0);if((ia|0)==(-1|0))fa=ka;else{ga=ia;ha=ma;K=194;break d}}while(0);k[31481]=k[31481]|4;na=fa;K=191}else{na=0;K=191}while(0);if((((K|0)==191?S>>>0<2147483647:0)?(fa=Wa(S|0)|0,S=Wa(0)|0,fa>>>0<S>>>0&((fa|0)!=(-1|0)&(S|0)!=(-1|0))):0)?(ma=S-fa|0,S=ma>>>0>(F+40|0)>>>0,S):0){ga=fa;ha=S?ma:na;K=194}if((K|0)==194){na=(k[31478]|0)+ha|0;k[31478]=na;if(na>>>0>(k[31479]|0)>>>0)k[31479]=na;na=k[31376]|0;g:do if(na){ma=125928;do{S=k[ma>>2]|0;fa=ma+4|0;ia=k[fa>>2]|0;if((ga|0)==(S+ia|0)){oa=S;pa=fa;qa=ia;ra=ma;K=204;break}ma=k[ma+8>>2]|0}while((ma|0)!=0);if(((K|0)==204?(k[ra+12>>2]&8|0)==0:0)?na>>>0<ga>>>0&na>>>0>=oa>>>0:0){k[pa>>2]=qa+ha;ma=(k[31373]|0)+ha|0;ia=na+8|0;fa=(ia&7|0)==0?0:0-ia&7;ia=ma-fa|0;k[31376]=na+fa;k[31373]=ia;k[na+(fa+4)>>2]=ia|1;k[na+(ma+4)>>2]=40;k[31377]=k[31492];break}ma=k[31374]|0;if(ga>>>0<ma>>>0){k[31374]=ga;sa=ga}else sa=ma;ma=ga+ha|0;ia=125928;while(1){if((k[ia>>2]|0)==(ma|0)){ta=ia;ua=ia;K=212;break}ia=k[ia+8>>2]|0;if(!ia){va=125928;break}}if((K|0)==212)if(!(k[ua+12>>2]&8)){k[ta>>2]=ga;ia=ua+4|0;k[ia>>2]=(k[ia>>2]|0)+ha;ia=ga+8|0;ma=(ia&7|0)==0?0:0-ia&7;ia=ga+(ha+8)|0;fa=(ia&7|0)==0?0:0-ia&7;ia=ga+(fa+ha)|0;S=ma+F|0;ka=ga+S|0;ja=ia-(ga+ma)-F|0;k[ga+(ma+4)>>2]=F|3;h:do if((ia|0)!=(na|0)){if((ia|0)==(k[31375]|0)){T=(k[31372]|0)+ja|0;k[31372]=T;k[31375]=ka;k[ga+(S+4)>>2]=T|1;k[ga+(T+S)>>2]=T;break}T=ha+4|0;ca=k[ga+(T+fa)>>2]|0;if((ca&3|0)==1){la=ca&-8;ea=ca>>>3;i:do if(ca>>>0>=256){da=k[ga+((fa|24)+ha)>>2]|0;aa=k[ga+(ha+12+fa)>>2]|0;do if((aa|0)==(ia|0)){ba=fa|16;X=ga+(T+ba)|0;$=k[X>>2]|0;if(!$){Z=ga+(ba+ha)|0;ba=k[Z>>2]|0;if(!ba){wa=0;break}else{xa=ba;ya=Z}}else{xa=$;ya=X}while(1){X=xa+20|0;$=k[X>>2]|0;if($){xa=$;ya=X;continue}X=xa+16|0;$=k[X>>2]|0;if(!$){za=xa;Aa=ya;break}else{xa=$;ya=X}}if(Aa>>>0<sa>>>0)xb();else{k[Aa>>2]=0;wa=za;break}}else{X=k[ga+((fa|8)+ha)>>2]|0;if(X>>>0<sa>>>0)xb();$=X+12|0;if((k[$>>2]|0)!=(ia|0))xb();Z=aa+8|0;if((k[Z>>2]|0)==(ia|0)){k[$>>2]=aa;k[Z>>2]=X;wa=aa;break}else xb()}while(0);if(!da)break;aa=k[ga+(ha+28+fa)>>2]|0;X=125784+(aa<<2)|0;do if((ia|0)!=(k[X>>2]|0)){if(da>>>0<(k[31374]|0)>>>0)xb();Z=da+16|0;if((k[Z>>2]|0)==(ia|0))k[Z>>2]=wa;else k[da+20>>2]=wa;if(!wa)break i}else{k[X>>2]=wa;if(wa)break;k[31371]=k[31371]&~(1<<aa);break i}while(0);aa=k[31374]|0;if(wa>>>0<aa>>>0)xb();k[wa+24>>2]=da;X=fa|16;Z=k[ga+(X+ha)>>2]|0;do if(Z)if(Z>>>0<aa>>>0)xb();else{k[wa+16>>2]=Z;k[Z+24>>2]=wa;break}while(0);Z=k[ga+(T+X)>>2]|0;if(!Z)break;if(Z>>>0<(k[31374]|0)>>>0)xb();else{k[wa+20>>2]=Z;k[Z+24>>2]=wa;break}}else{Z=k[ga+((fa|8)+ha)>>2]|0;aa=k[ga+(ha+12+fa)>>2]|0;da=125520+(ea<<1<<2)|0;do if((Z|0)!=(da|0)){if(Z>>>0<sa>>>0)xb();if((k[Z+12>>2]|0)==(ia|0))break;xb()}while(0);if((aa|0)==(Z|0)){k[31370]=k[31370]&~(1<<ea);break}do if((aa|0)==(da|0))Ba=aa+8|0;else{if(aa>>>0<sa>>>0)xb();X=aa+8|0;if((k[X>>2]|0)==(ia|0)){Ba=X;break}xb()}while(0);k[Z+12>>2]=aa;k[Ba>>2]=Z}while(0);Ca=ga+((la|fa)+ha)|0;Da=la+ja|0}else{Ca=ia;Da=ja}ea=Ca+4|0;k[ea>>2]=k[ea>>2]&-2;k[ga+(S+4)>>2]=Da|1;k[ga+(Da+S)>>2]=Da;ea=Da>>>3;if(Da>>>0<256){T=ea<<1;ca=125520+(T<<2)|0;da=k[31370]|0;X=1<<ea;do if(!(da&X)){k[31370]=da|X;Ea=125520+(T+2<<2)|0;Fa=ca}else{ea=125520+(T+2<<2)|0;$=k[ea>>2]|0;if($>>>0>=(k[31374]|0)>>>0){Ea=ea;Fa=$;break}xb()}while(0);k[Ea>>2]=ka;k[Fa+12>>2]=ka;k[ga+(S+8)>>2]=Fa;k[ga+(S+12)>>2]=ca;break}T=Da>>>8;do if(!T)Ga=0;else{if(Da>>>0>16777215){Ga=31;break}X=(T+1048320|0)>>>16&8;da=T<<X;la=(da+520192|0)>>>16&4;$=da<<la;da=($+245760|0)>>>16&2;ea=14-(la|X|da)+($<<da>>>15)|0;Ga=Da>>>(ea+7|0)&1|ea<<1}while(0);T=125784+(Ga<<2)|0;k[ga+(S+28)>>2]=Ga;k[ga+(S+20)>>2]=0;k[ga+(S+16)>>2]=0;ca=k[31371]|0;ea=1<<Ga;if(!(ca&ea)){k[31371]=ca|ea;k[T>>2]=ka;k[ga+(S+24)>>2]=T;k[ga+(S+12)>>2]=ka;k[ga+(S+8)>>2]=ka;break}ea=k[T>>2]|0;j:do if((k[ea+4>>2]&-8|0)!=(Da|0)){T=Da<<((Ga|0)==31?0:25-(Ga>>>1)|0);ca=ea;while(1){da=ca+16+(T>>>31<<2)|0;$=k[da>>2]|0;if(!$){Ha=da;Ia=ca;break}if((k[$+4>>2]&-8|0)==(Da|0)){Ja=$;break j}else{T=T<<1;ca=$}}if(Ha>>>0<(k[31374]|0)>>>0)xb();else{k[Ha>>2]=ka;k[ga+(S+24)>>2]=Ia;k[ga+(S+12)>>2]=ka;k[ga+(S+8)>>2]=ka;break h}}else Ja=ea;while(0);ea=Ja+8|0;ca=k[ea>>2]|0;T=k[31374]|0;if(ca>>>0>=T>>>0&Ja>>>0>=T>>>0){k[ca+12>>2]=ka;k[ea>>2]=ka;k[ga+(S+8)>>2]=ca;k[ga+(S+12)>>2]=Ja;k[ga+(S+24)>>2]=0;break}else xb()}else{ca=(k[31373]|0)+ja|0;k[31373]=ca;k[31376]=ka;k[ga+(S+4)>>2]=ca|1}while(0);n=ga+(ma|8)|0;return n|0}else va=125928;while(1){S=k[va>>2]|0;if(S>>>0<=na>>>0?(ka=k[va+4>>2]|0,ja=S+ka|0,ja>>>0>na>>>0):0){Ka=S;La=ka;Ma=ja;break}va=k[va+8>>2]|0}ma=Ka+(La+-39)|0;ja=Ka+(La+-47+((ma&7|0)==0?0:0-ma&7))|0;ma=na+16|0;ka=ja>>>0<ma>>>0?na:ja;ja=ka+8|0;S=ga+8|0;ia=(S&7|0)==0?0:0-S&7;S=ha+-40-ia|0;k[31376]=ga+ia;k[31373]=S;k[ga+(ia+4)>>2]=S|1;k[ga+(ha+-36)>>2]=40;k[31377]=k[31492];S=ka+4|0;k[S>>2]=27;k[ja>>2]=k[31482];k[ja+4>>2]=k[31483];k[ja+8>>2]=k[31484];k[ja+12>>2]=k[31485];k[31482]=ga;k[31483]=ha;k[31485]=0;k[31484]=ja;ja=ka+28|0;k[ja>>2]=7;if((ka+32|0)>>>0<Ma>>>0){ia=ja;do{ja=ia;ia=ia+4|0;k[ia>>2]=7}while((ja+8|0)>>>0<Ma>>>0)}if((ka|0)!=(na|0)){ia=ka-na|0;k[S>>2]=k[S>>2]&-2;k[na+4>>2]=ia|1;k[ka>>2]=ia;ja=ia>>>3;if(ia>>>0<256){fa=ja<<1;ca=125520+(fa<<2)|0;ea=k[31370]|0;T=1<<ja;if(ea&T){ja=125520+(fa+2<<2)|0;Z=k[ja>>2]|0;if(Z>>>0<(k[31374]|0)>>>0)xb();else{Na=ja;Oa=Z}}else{k[31370]=ea|T;Na=125520+(fa+2<<2)|0;Oa=ca}k[Na>>2]=na;k[Oa+12>>2]=na;k[na+8>>2]=Oa;k[na+12>>2]=ca;break}ca=ia>>>8;if(ca)if(ia>>>0>16777215)Pa=31;else{fa=(ca+1048320|0)>>>16&8;T=ca<<fa;ca=(T+520192|0)>>>16&4;ea=T<<ca;T=(ea+245760|0)>>>16&2;Z=14-(ca|fa|T)+(ea<<T>>>15)|0;Pa=ia>>>(Z+7|0)&1|Z<<1}else Pa=0;Z=125784+(Pa<<2)|0;k[na+28>>2]=Pa;k[na+20>>2]=0;k[ma>>2]=0;T=k[31371]|0;ea=1<<Pa;if(!(T&ea)){k[31371]=T|ea;k[Z>>2]=na;k[na+24>>2]=Z;k[na+12>>2]=na;k[na+8>>2]=na;break}ea=k[Z>>2]|0;k:do if((k[ea+4>>2]&-8|0)!=(ia|0)){Z=ia<<((Pa|0)==31?0:25-(Pa>>>1)|0);T=ea;while(1){fa=T+16+(Z>>>31<<2)|0;ca=k[fa>>2]|0;if(!ca){Qa=fa;Sa=T;break}if((k[ca+4>>2]&-8|0)==(ia|0)){Ta=ca;break k}else{Z=Z<<1;T=ca}}if(Qa>>>0<(k[31374]|0)>>>0)xb();else{k[Qa>>2]=na;k[na+24>>2]=Sa;k[na+12>>2]=na;k[na+8>>2]=na;break g}}else Ta=ea;while(0);ea=Ta+8|0;ia=k[ea>>2]|0;ma=k[31374]|0;if(ia>>>0>=ma>>>0&Ta>>>0>=ma>>>0){k[ia+12>>2]=na;k[ea>>2]=na;k[na+8>>2]=ia;k[na+12>>2]=Ta;k[na+24>>2]=0;break}else xb()}}else{ia=k[31374]|0;if((ia|0)==0|ga>>>0<ia>>>0)k[31374]=ga;k[31482]=ga;k[31483]=ha;k[31485]=0;k[31379]=k[31488];k[31378]=-1;ia=0;do{ea=ia<<1;ma=125520+(ea<<2)|0;k[125520+(ea+3<<2)>>2]=ma;k[125520+(ea+2<<2)>>2]=ma;ia=ia+1|0}while((ia|0)!=32);ia=ga+8|0;ma=(ia&7|0)==0?0:0-ia&7;ia=ha+-40-ma|0;k[31376]=ga+ma;k[31373]=ia;k[ga+(ma+4)>>2]=ia|1;k[ga+(ha+-36)>>2]=40;k[31377]=k[31492]}while(0);ha=k[31373]|0;if(ha>>>0>F>>>0){ga=ha-F|0;k[31373]=ga;ha=k[31376]|0;k[31376]=ha+F;k[ha+(F+4)>>2]=ga|1;k[ha+4>>2]=F|3;n=ha+8|0;return n|0}}ha=rv()|0;k[ha>>2]=12;n=0;return n|0}function dy(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0;if(!a)return;b=a+-8|0;c=k[31374]|0;if(b>>>0<c>>>0)xb();d=k[a+-4>>2]|0;e=d&3;if((e|0)==1)xb();f=d&-8;g=a+(f+-8)|0;do if(!(d&1)){h=k[b>>2]|0;if(!e)return;i=-8-h|0;j=a+i|0;l=h+f|0;if(j>>>0<c>>>0)xb();if((j|0)==(k[31375]|0)){m=a+(f+-4)|0;n=k[m>>2]|0;if((n&3|0)!=3){o=j;p=l;break}k[31372]=l;k[m>>2]=n&-2;k[a+(i+4)>>2]=l|1;k[g>>2]=l;return}n=h>>>3;if(h>>>0<256){h=k[a+(i+8)>>2]|0;m=k[a+(i+12)>>2]|0;q=125520+(n<<1<<2)|0;if((h|0)!=(q|0)){if(h>>>0<c>>>0)xb();if((k[h+12>>2]|0)!=(j|0))xb()}if((m|0)==(h|0)){k[31370]=k[31370]&~(1<<n);o=j;p=l;break}if((m|0)!=(q|0)){if(m>>>0<c>>>0)xb();q=m+8|0;if((k[q>>2]|0)==(j|0))r=q;else xb()}else r=m+8|0;k[h+12>>2]=m;k[r>>2]=h;o=j;p=l;break}h=k[a+(i+24)>>2]|0;m=k[a+(i+12)>>2]|0;do if((m|0)==(j|0)){q=a+(i+20)|0;n=k[q>>2]|0;if(!n){s=a+(i+16)|0;t=k[s>>2]|0;if(!t){u=0;break}else{v=t;w=s}}else{v=n;w=q}while(1){q=v+20|0;n=k[q>>2]|0;if(n){v=n;w=q;continue}q=v+16|0;n=k[q>>2]|0;if(!n){x=v;y=w;break}else{v=n;w=q}}if(y>>>0<c>>>0)xb();else{k[y>>2]=0;u=x;break}}else{q=k[a+(i+8)>>2]|0;if(q>>>0<c>>>0)xb();n=q+12|0;if((k[n>>2]|0)!=(j|0))xb();s=m+8|0;if((k[s>>2]|0)==(j|0)){k[n>>2]=m;k[s>>2]=q;u=m;break}else xb()}while(0);if(h){m=k[a+(i+28)>>2]|0;q=125784+(m<<2)|0;if((j|0)==(k[q>>2]|0)){k[q>>2]=u;if(!u){k[31371]=k[31371]&~(1<<m);o=j;p=l;break}}else{if(h>>>0<(k[31374]|0)>>>0)xb();m=h+16|0;if((k[m>>2]|0)==(j|0))k[m>>2]=u;else k[h+20>>2]=u;if(!u){o=j;p=l;break}}m=k[31374]|0;if(u>>>0<m>>>0)xb();k[u+24>>2]=h;q=k[a+(i+16)>>2]|0;do if(q)if(q>>>0<m>>>0)xb();else{k[u+16>>2]=q;k[q+24>>2]=u;break}while(0);q=k[a+(i+20)>>2]|0;if(q)if(q>>>0<(k[31374]|0)>>>0)xb();else{k[u+20>>2]=q;k[q+24>>2]=u;o=j;p=l;break}else{o=j;p=l}}else{o=j;p=l}}else{o=b;p=f}while(0);if(o>>>0>=g>>>0)xb();b=a+(f+-4)|0;u=k[b>>2]|0;if(!(u&1))xb();if(!(u&2)){if((g|0)==(k[31376]|0)){c=(k[31373]|0)+p|0;k[31373]=c;k[31376]=o;k[o+4>>2]=c|1;if((o|0)!=(k[31375]|0))return;k[31375]=0;k[31372]=0;return}if((g|0)==(k[31375]|0)){c=(k[31372]|0)+p|0;k[31372]=c;k[31375]=o;k[o+4>>2]=c|1;k[o+c>>2]=c;return}c=(u&-8)+p|0;x=u>>>3;do if(u>>>0>=256){y=k[a+(f+16)>>2]|0;w=k[a+(f|4)>>2]|0;do if((w|0)==(g|0)){v=a+(f+12)|0;r=k[v>>2]|0;if(!r){e=a+(f+8)|0;d=k[e>>2]|0;if(!d){z=0;break}else{A=d;B=e}}else{A=r;B=v}while(1){v=A+20|0;r=k[v>>2]|0;if(r){A=r;B=v;continue}v=A+16|0;r=k[v>>2]|0;if(!r){C=A;D=B;break}else{A=r;B=v}}if(D>>>0<(k[31374]|0)>>>0)xb();else{k[D>>2]=0;z=C;break}}else{v=k[a+f>>2]|0;if(v>>>0<(k[31374]|0)>>>0)xb();r=v+12|0;if((k[r>>2]|0)!=(g|0))xb();e=w+8|0;if((k[e>>2]|0)==(g|0)){k[r>>2]=w;k[e>>2]=v;z=w;break}else xb()}while(0);if(y){w=k[a+(f+20)>>2]|0;l=125784+(w<<2)|0;if((g|0)==(k[l>>2]|0)){k[l>>2]=z;if(!z){k[31371]=k[31371]&~(1<<w);break}}else{if(y>>>0<(k[31374]|0)>>>0)xb();w=y+16|0;if((k[w>>2]|0)==(g|0))k[w>>2]=z;else k[y+20>>2]=z;if(!z)break}w=k[31374]|0;if(z>>>0<w>>>0)xb();k[z+24>>2]=y;l=k[a+(f+8)>>2]|0;do if(l)if(l>>>0<w>>>0)xb();else{k[z+16>>2]=l;k[l+24>>2]=z;break}while(0);l=k[a+(f+12)>>2]|0;if(l)if(l>>>0<(k[31374]|0)>>>0)xb();else{k[z+20>>2]=l;k[l+24>>2]=z;break}}}else{l=k[a+f>>2]|0;w=k[a+(f|4)>>2]|0;y=125520+(x<<1<<2)|0;if((l|0)!=(y|0)){if(l>>>0<(k[31374]|0)>>>0)xb();if((k[l+12>>2]|0)!=(g|0))xb()}if((w|0)==(l|0)){k[31370]=k[31370]&~(1<<x);break}if((w|0)!=(y|0)){if(w>>>0<(k[31374]|0)>>>0)xb();y=w+8|0;if((k[y>>2]|0)==(g|0))E=y;else xb()}else E=w+8|0;k[l+12>>2]=w;k[E>>2]=l}while(0);k[o+4>>2]=c|1;k[o+c>>2]=c;if((o|0)==(k[31375]|0)){k[31372]=c;return}else F=c}else{k[b>>2]=u&-2;k[o+4>>2]=p|1;k[o+p>>2]=p;F=p}p=F>>>3;if(F>>>0<256){u=p<<1;b=125520+(u<<2)|0;c=k[31370]|0;E=1<<p;if(c&E){p=125520+(u+2<<2)|0;g=k[p>>2]|0;if(g>>>0<(k[31374]|0)>>>0)xb();else{G=p;H=g}}else{k[31370]=c|E;G=125520+(u+2<<2)|0;H=b}k[G>>2]=o;k[H+12>>2]=o;k[o+8>>2]=H;k[o+12>>2]=b;return}b=F>>>8;if(b)if(F>>>0>16777215)I=31;else{H=(b+1048320|0)>>>16&8;G=b<<H;b=(G+520192|0)>>>16&4;u=G<<b;G=(u+245760|0)>>>16&2;E=14-(b|H|G)+(u<<G>>>15)|0;I=F>>>(E+7|0)&1|E<<1}else I=0;E=125784+(I<<2)|0;k[o+28>>2]=I;k[o+20>>2]=0;k[o+16>>2]=0;G=k[31371]|0;u=1<<I;a:do if(G&u){H=k[E>>2]|0;b:do if((k[H+4>>2]&-8|0)!=(F|0)){b=F<<((I|0)==31?0:25-(I>>>1)|0);c=H;while(1){g=c+16+(b>>>31<<2)|0;p=k[g>>2]|0;if(!p){J=g;K=c;break}if((k[p+4>>2]&-8|0)==(F|0)){L=p;break b}else{b=b<<1;c=p}}if(J>>>0<(k[31374]|0)>>>0)xb();else{k[J>>2]=o;k[o+24>>2]=K;k[o+12>>2]=o;k[o+8>>2]=o;break a}}else L=H;while(0);H=L+8|0;c=k[H>>2]|0;b=k[31374]|0;if(c>>>0>=b>>>0&L>>>0>=b>>>0){k[c+12>>2]=o;k[H>>2]=o;k[o+8>>2]=c;k[o+12>>2]=L;k[o+24>>2]=0;break}else xb()}else{k[31371]=G|u;k[E>>2]=o;k[o+24>>2]=E;k[o+12>>2]=o;k[o+8>>2]=o}while(0);o=(k[31378]|0)+-1|0;k[31378]=o;if(!o)M=125936;else return;while(1){o=k[M>>2]|0;if(!o)break;else M=o+8|0}k[31378]=-1;return}function ey(a,b){a=a|0;b=b|0;var c=0,d=0;if(a){c=ha(b,a)|0;if((b|a)>>>0>65535)d=((c>>>0)/(a>>>0)|0|0)==(b|0)?c:-1;else d=c}else d=0;c=cy(d)|0;if(!c)return c|0;if(!(k[c+-4>>2]&3))return c|0;sy(c|0,0,d|0)|0;return c|0}function fy(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0;if(!a){c=cy(b)|0;return c|0}if(b>>>0>4294967231){d=rv()|0;k[d>>2]=12;c=0;return c|0}d=gy(a+-8|0,b>>>0<11?16:b+11&-8)|0;if(d){c=d+8|0;return c|0}d=cy(b)|0;if(!d){c=0;return c|0}e=k[a+-4>>2]|0;f=(e&-8)-((e&3|0)==0?8:4)|0;py(d|0,a|0,(f>>>0<b>>>0?f:b)|0)|0;dy(a);c=d;return c|0}function gy(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;c=a+4|0;d=k[c>>2]|0;e=d&-8;f=a+e|0;g=k[31374]|0;h=d&3;if(!((h|0)!=1&a>>>0>=g>>>0&a>>>0<f>>>0))xb();i=a+(e|4)|0;j=k[i>>2]|0;if(!(j&1))xb();if(!h){if(b>>>0<256){l=0;return l|0}if(e>>>0>=(b+4|0)>>>0?(e-b|0)>>>0<=k[31490]<<1>>>0:0){l=a;return l|0}l=0;return l|0}if(e>>>0>=b>>>0){h=e-b|0;if(h>>>0<=15){l=a;return l|0}k[c>>2]=d&1|b|2;k[a+(b+4)>>2]=h|3;k[i>>2]=k[i>>2]|1;hy(a+b|0,h);l=a;return l|0}if((f|0)==(k[31376]|0)){h=(k[31373]|0)+e|0;if(h>>>0<=b>>>0){l=0;return l|0}i=h-b|0;k[c>>2]=d&1|b|2;k[a+(b+4)>>2]=i|1;k[31376]=a+b;k[31373]=i;l=a;return l|0}if((f|0)==(k[31375]|0)){i=(k[31372]|0)+e|0;if(i>>>0<b>>>0){l=0;return l|0}h=i-b|0;if(h>>>0>15){k[c>>2]=d&1|b|2;k[a+(b+4)>>2]=h|1;k[a+i>>2]=h;m=a+(i+4)|0;k[m>>2]=k[m>>2]&-2;n=a+b|0;o=h}else{k[c>>2]=d&1|i|2;h=a+(i+4)|0;k[h>>2]=k[h>>2]|1;n=0;o=0}k[31372]=o;k[31375]=n;l=a;return l|0}if(j&2){l=0;return l|0}n=(j&-8)+e|0;if(n>>>0<b>>>0){l=0;return l|0}o=n-b|0;h=j>>>3;do if(j>>>0>=256){i=k[a+(e+24)>>2]|0;m=k[a+(e+12)>>2]|0;do if((m|0)==(f|0)){p=a+(e+20)|0;q=k[p>>2]|0;if(!q){r=a+(e+16)|0;s=k[r>>2]|0;if(!s){t=0;break}else{u=s;v=r}}else{u=q;v=p}while(1){p=u+20|0;q=k[p>>2]|0;if(q){u=q;v=p;continue}p=u+16|0;q=k[p>>2]|0;if(!q){w=u;x=v;break}else{u=q;v=p}}if(x>>>0<g>>>0)xb();else{k[x>>2]=0;t=w;break}}else{p=k[a+(e+8)>>2]|0;if(p>>>0<g>>>0)xb();q=p+12|0;if((k[q>>2]|0)!=(f|0))xb();r=m+8|0;if((k[r>>2]|0)==(f|0)){k[q>>2]=m;k[r>>2]=p;t=m;break}else xb()}while(0);if(i){m=k[a+(e+28)>>2]|0;p=125784+(m<<2)|0;if((f|0)==(k[p>>2]|0)){k[p>>2]=t;if(!t){k[31371]=k[31371]&~(1<<m);break}}else{if(i>>>0<(k[31374]|0)>>>0)xb();m=i+16|0;if((k[m>>2]|0)==(f|0))k[m>>2]=t;else k[i+20>>2]=t;if(!t)break}m=k[31374]|0;if(t>>>0<m>>>0)xb();k[t+24>>2]=i;p=k[a+(e+16)>>2]|0;do if(p)if(p>>>0<m>>>0)xb();else{k[t+16>>2]=p;k[p+24>>2]=t;break}while(0);p=k[a+(e+20)>>2]|0;if(p)if(p>>>0<(k[31374]|0)>>>0)xb();else{k[t+20>>2]=p;k[p+24>>2]=t;break}}}else{p=k[a+(e+8)>>2]|0;m=k[a+(e+12)>>2]|0;i=125520+(h<<1<<2)|0;if((p|0)!=(i|0)){if(p>>>0<g>>>0)xb();if((k[p+12>>2]|0)!=(f|0))xb()}if((m|0)==(p|0)){k[31370]=k[31370]&~(1<<h);break}if((m|0)!=(i|0)){if(m>>>0<g>>>0)xb();i=m+8|0;if((k[i>>2]|0)==(f|0))y=i;else xb()}else y=m+8|0;k[p+12>>2]=m;k[y>>2]=p}while(0);if(o>>>0<16){k[c>>2]=n|d&1|2;y=a+(n|4)|0;k[y>>2]=k[y>>2]|1;l=a;return l|0}else{k[c>>2]=d&1|b|2;k[a+(b+4)>>2]=o|3;d=a+(n|4)|0;k[d>>2]=k[d>>2]|1;hy(a+b|0,o);l=a;return l|0}return 0}function hy(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0;c=a+b|0;d=k[a+4>>2]|0;do if(!(d&1)){e=k[a>>2]|0;if(!(d&3))return;f=a+(0-e)|0;g=e+b|0;h=k[31374]|0;if(f>>>0<h>>>0)xb();if((f|0)==(k[31375]|0)){i=a+(b+4)|0;j=k[i>>2]|0;if((j&3|0)!=3){l=f;m=g;break}k[31372]=g;k[i>>2]=j&-2;k[a+(4-e)>>2]=g|1;k[c>>2]=g;return}j=e>>>3;if(e>>>0<256){i=k[a+(8-e)>>2]|0;n=k[a+(12-e)>>2]|0;o=125520+(j<<1<<2)|0;if((i|0)!=(o|0)){if(i>>>0<h>>>0)xb();if((k[i+12>>2]|0)!=(f|0))xb()}if((n|0)==(i|0)){k[31370]=k[31370]&~(1<<j);l=f;m=g;break}if((n|0)!=(o|0)){if(n>>>0<h>>>0)xb();o=n+8|0;if((k[o>>2]|0)==(f|0))p=o;else xb()}else p=n+8|0;k[i+12>>2]=n;k[p>>2]=i;l=f;m=g;break}i=k[a+(24-e)>>2]|0;n=k[a+(12-e)>>2]|0;do if((n|0)==(f|0)){o=16-e|0;j=a+(o+4)|0;q=k[j>>2]|0;if(!q){r=a+o|0;o=k[r>>2]|0;if(!o){s=0;break}else{t=o;u=r}}else{t=q;u=j}while(1){j=t+20|0;q=k[j>>2]|0;if(q){t=q;u=j;continue}j=t+16|0;q=k[j>>2]|0;if(!q){v=t;w=u;break}else{t=q;u=j}}if(w>>>0<h>>>0)xb();else{k[w>>2]=0;s=v;break}}else{j=k[a+(8-e)>>2]|0;if(j>>>0<h>>>0)xb();q=j+12|0;if((k[q>>2]|0)!=(f|0))xb();r=n+8|0;if((k[r>>2]|0)==(f|0)){k[q>>2]=n;k[r>>2]=j;s=n;break}else xb()}while(0);if(i){n=k[a+(28-e)>>2]|0;h=125784+(n<<2)|0;if((f|0)==(k[h>>2]|0)){k[h>>2]=s;if(!s){k[31371]=k[31371]&~(1<<n);l=f;m=g;break}}else{if(i>>>0<(k[31374]|0)>>>0)xb();n=i+16|0;if((k[n>>2]|0)==(f|0))k[n>>2]=s;else k[i+20>>2]=s;if(!s){l=f;m=g;break}}n=k[31374]|0;if(s>>>0<n>>>0)xb();k[s+24>>2]=i;h=16-e|0;j=k[a+h>>2]|0;do if(j)if(j>>>0<n>>>0)xb();else{k[s+16>>2]=j;k[j+24>>2]=s;break}while(0);j=k[a+(h+4)>>2]|0;if(j)if(j>>>0<(k[31374]|0)>>>0)xb();else{k[s+20>>2]=j;k[j+24>>2]=s;l=f;m=g;break}else{l=f;m=g}}else{l=f;m=g}}else{l=a;m=b}while(0);s=k[31374]|0;if(c>>>0<s>>>0)xb();v=a+(b+4)|0;w=k[v>>2]|0;if(!(w&2)){if((c|0)==(k[31376]|0)){u=(k[31373]|0)+m|0;k[31373]=u;k[31376]=l;k[l+4>>2]=u|1;if((l|0)!=(k[31375]|0))return;k[31375]=0;k[31372]=0;return}if((c|0)==(k[31375]|0)){u=(k[31372]|0)+m|0;k[31372]=u;k[31375]=l;k[l+4>>2]=u|1;k[l+u>>2]=u;return}u=(w&-8)+m|0;t=w>>>3;do if(w>>>0>=256){p=k[a+(b+24)>>2]|0;d=k[a+(b+12)>>2]|0;do if((d|0)==(c|0)){j=a+(b+20)|0;n=k[j>>2]|0;if(!n){e=a+(b+16)|0;i=k[e>>2]|0;if(!i){x=0;break}else{y=i;z=e}}else{y=n;z=j}while(1){j=y+20|0;n=k[j>>2]|0;if(n){y=n;z=j;continue}j=y+16|0;n=k[j>>2]|0;if(!n){A=y;B=z;break}else{y=n;z=j}}if(B>>>0<s>>>0)xb();else{k[B>>2]=0;x=A;break}}else{j=k[a+(b+8)>>2]|0;if(j>>>0<s>>>0)xb();n=j+12|0;if((k[n>>2]|0)!=(c|0))xb();e=d+8|0;if((k[e>>2]|0)==(c|0)){k[n>>2]=d;k[e>>2]=j;x=d;break}else xb()}while(0);if(p){d=k[a+(b+28)>>2]|0;g=125784+(d<<2)|0;if((c|0)==(k[g>>2]|0)){k[g>>2]=x;if(!x){k[31371]=k[31371]&~(1<<d);break}}else{if(p>>>0<(k[31374]|0)>>>0)xb();d=p+16|0;if((k[d>>2]|0)==(c|0))k[d>>2]=x;else k[p+20>>2]=x;if(!x)break}d=k[31374]|0;if(x>>>0<d>>>0)xb();k[x+24>>2]=p;g=k[a+(b+16)>>2]|0;do if(g)if(g>>>0<d>>>0)xb();else{k[x+16>>2]=g;k[g+24>>2]=x;break}while(0);g=k[a+(b+20)>>2]|0;if(g)if(g>>>0<(k[31374]|0)>>>0)xb();else{k[x+20>>2]=g;k[g+24>>2]=x;break}}}else{g=k[a+(b+8)>>2]|0;d=k[a+(b+12)>>2]|0;p=125520+(t<<1<<2)|0;if((g|0)!=(p|0)){if(g>>>0<s>>>0)xb();if((k[g+12>>2]|0)!=(c|0))xb()}if((d|0)==(g|0)){k[31370]=k[31370]&~(1<<t);break}if((d|0)!=(p|0)){if(d>>>0<s>>>0)xb();p=d+8|0;if((k[p>>2]|0)==(c|0))C=p;else xb()}else C=d+8|0;k[g+12>>2]=d;k[C>>2]=g}while(0);k[l+4>>2]=u|1;k[l+u>>2]=u;if((l|0)==(k[31375]|0)){k[31372]=u;return}else D=u}else{k[v>>2]=w&-2;k[l+4>>2]=m|1;k[l+m>>2]=m;D=m}m=D>>>3;if(D>>>0<256){w=m<<1;v=125520+(w<<2)|0;u=k[31370]|0;C=1<<m;if(u&C){m=125520+(w+2<<2)|0;c=k[m>>2]|0;if(c>>>0<(k[31374]|0)>>>0)xb();else{E=m;F=c}}else{k[31370]=u|C;E=125520+(w+2<<2)|0;F=v}k[E>>2]=l;k[F+12>>2]=l;k[l+8>>2]=F;k[l+12>>2]=v;return}v=D>>>8;if(v)if(D>>>0>16777215)G=31;else{F=(v+1048320|0)>>>16&8;E=v<<F;v=(E+520192|0)>>>16&4;w=E<<v;E=(w+245760|0)>>>16&2;C=14-(v|F|E)+(w<<E>>>15)|0;G=D>>>(C+7|0)&1|C<<1}else G=0;C=125784+(G<<2)|0;k[l+28>>2]=G;k[l+20>>2]=0;k[l+16>>2]=0;E=k[31371]|0;w=1<<G;if(!(E&w)){k[31371]=E|w;k[C>>2]=l;k[l+24>>2]=C;k[l+12>>2]=l;k[l+8>>2]=l;return}w=k[C>>2]|0;a:do if((k[w+4>>2]&-8|0)==(D|0))H=w;else{C=D<<((G|0)==31?0:25-(G>>>1)|0);E=w;while(1){F=E+16+(C>>>31<<2)|0;v=k[F>>2]|0;if(!v){I=F;J=E;break}if((k[v+4>>2]&-8|0)==(D|0)){H=v;break a}else{C=C<<1;E=v}}if(I>>>0<(k[31374]|0)>>>0)xb();k[I>>2]=l;k[l+24>>2]=J;k[l+12>>2]=l;k[l+8>>2]=l;return}while(0);J=H+8|0;I=k[J>>2]|0;D=k[31374]|0;if(!(I>>>0>=D>>>0&H>>>0>=D>>>0))xb();k[I+12>>2]=l;k[J>>2]=l;k[l+8>>2]=I;k[l+12>>2]=H;k[l+24>>2]=0;return}function iy(){}function jy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0;e=b-d>>>0;e=b-d-(c>>>0>a>>>0|0)>>>0;return (L=e,a-c>>>0|0)|0}function ky(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0;e=a+c>>>0;return (L=b+d+(e>>>0<a>>>0|0)>>>0,e|0)|0}function ly(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0;y=y+1|0;k[a>>2]=y;while((e|0)<(d|0)){if(!(k[c+(e<<3)>>2]|0)){k[c+(e<<3)>>2]=y;k[c+((e<<3)+4)>>2]=b;k[c+((e<<3)+8)>>2]=0;L=d;return c|0}e=e+1|0}d=d*2|0;c=fy(c|0,8*(d+1|0)|0)|0;c=ly(a|0,b|0,c|0,d|0)|0;L=d;return c|0}function my(a,b,c){a=a|0;b=b|0;c=c|0;if((c|0)<32){L=b>>c;return a>>>c|(b&(1<<c)-1)<<32-c}L=(b|0)<0?-1:0;return b>>c-32|0}function ny(a,b,c){a=a|0;b=b|0;c=c|0;if((c|0)<32){L=b>>>c;return a>>>c|(b&(1<<c)-1)<<32-c}L=0;return b>>>c-32|0}function oy(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;while((d|0)<(c|0)){e=k[b+(d<<3)>>2]|0;if(!e)break;if((e|0)==(a|0))return k[b+((d<<3)+4)>>2]|0;d=d+1|0}return 0}function py(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;if((c|0)>=4096)return db(a|0,b|0,c|0)|0;d=a|0;if((a&3)==(b&3)){while(a&3){if(!c)return d|0;i[a>>0]=i[b>>0]|0;a=a+1|0;b=b+1|0;c=c-1|0}while((c|0)>=4){k[a>>2]=k[b>>2];a=a+4|0;b=b+4|0;c=c-4|0}}while((c|0)>0){i[a>>0]=i[b>>0]|0;a=a+1|0;b=b+1|0;c=c-1|0}return d|0}function qy(a,b,c){a=a|0;b=b|0;c=c|0;var d=0;if((b|0)<(a|0)&(a|0)<(b+c|0)){d=a;b=b+c|0;a=a+c|0;while((c|0)>0){a=a-1|0;b=b-1|0;c=c-1|0;i[a>>0]=i[b>>0]|0}a=d}else py(a,b,c)|0;return a|0}function ry(a){a=a|0;return (a&255)<<24|(a>>8&255)<<16|(a>>16&255)<<8|a>>>24|0}function sy(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=a+c|0;if((c|0)>=20){b=b&255;e=a&3;f=b|b<<8|b<<16|b<<24;g=d&~3;if(e){e=a+4-e|0;while((a|0)<(e|0)){i[a>>0]=b;a=a+1|0}}while((a|0)<(g|0)){k[a>>2]=f;a=a+4|0}}while((a|0)<(d|0)){i[a>>0]=b;a=a+1|0}return a-c|0}function ty(a,b,c){a=a|0;b=b|0;c=c|0;if((c|0)<32){L=b<<c|(a&(1<<c)-1<<32-c)>>>32-c;return a<<c}L=a<<c-32;return 0}function uy(a){a=a|0;var b=0;b=i[v+(a&255)>>0]|0;if((b|0)<8)return b|0;b=i[v+(a>>8&255)>>0]|0;if((b|0)<8)return b+8|0;b=i[v+(a>>16&255)>>0]|0;if((b|0)<8)return b+16|0;return (i[v+(a>>>24)>>0]|0)+24|0}function vy(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0;c=a&65535;d=b&65535;e=ha(d,c)|0;f=a>>>16;a=(e>>>16)+(ha(d,f)|0)|0;d=b>>>16;b=ha(d,c)|0;return (L=(a>>>16)+(ha(d,f)|0)+(((a&65535)+b|0)>>>16)|0,a+b<<16|e&65535|0)|0}function wy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0;e=b>>31|((b|0)<0?-1:0)<<1;f=((b|0)<0?-1:0)>>31|((b|0)<0?-1:0)<<1;g=d>>31|((d|0)<0?-1:0)<<1;h=((d|0)<0?-1:0)>>31|((d|0)<0?-1:0)<<1;i=jy(e^a,f^b,e,f)|0;b=L;a=g^e;e=h^f;return jy((By(i,b,jy(g^c,h^d,g,h)|0,L,0)|0)^a,L^e,a,e)|0}function xy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0;e=r;r=r+16|0;f=e|0;g=b>>31|((b|0)<0?-1:0)<<1;h=((b|0)<0?-1:0)>>31|((b|0)<0?-1:0)<<1;i=d>>31|((d|0)<0?-1:0)<<1;j=((d|0)<0?-1:0)>>31|((d|0)<0?-1:0)<<1;l=jy(g^a,h^b,g,h)|0;b=L;By(l,b,jy(i^c,j^d,i,j)|0,L,f)|0;j=jy(k[f>>2]^g,k[f+4>>2]^h,g,h)|0;h=L;r=e;return (L=h,j)|0}function yy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=a;a=c;c=vy(e,a)|0;f=L;return (L=(ha(b,a)|0)+(ha(d,e)|0)+f|f&0,c|0|0)|0}function zy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;return By(a,b,c,d,0)|0}function Ay(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0;e=r;r=r+16|0;f=e|0;By(a,b,c,d,f)|0;r=e;return (L=k[f+4>>2]|0,k[f>>2]|0)|0}function By(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;f=a;g=b;h=g;i=c;j=d;l=j;if(!h){m=(e|0)!=0;if(!l){if(m){k[e>>2]=(f>>>0)%(i>>>0);k[e+4>>2]=0}n=0;o=(f>>>0)/(i>>>0)>>>0;return (L=n,o)|0}else{if(!m){n=0;o=0;return (L=n,o)|0}k[e>>2]=a|0;k[e+4>>2]=b&0;n=0;o=0;return (L=n,o)|0}}m=(l|0)==0;do if(i){if(!m){p=(ja(l|0)|0)-(ja(h|0)|0)|0;if(p>>>0<=31){q=p+1|0;r=31-p|0;s=p-31>>31;t=q;u=f>>>(q>>>0)&s|h<<r;v=h>>>(q>>>0)&s;w=0;x=f<<r;break}if(!e){n=0;o=0;return (L=n,o)|0}k[e>>2]=a|0;k[e+4>>2]=g|b&0;n=0;o=0;return (L=n,o)|0}r=i-1|0;if(r&i){s=(ja(i|0)|0)+33-(ja(h|0)|0)|0;q=64-s|0;p=32-s|0;y=p>>31;z=s-32|0;A=z>>31;t=s;u=p-1>>31&h>>>(z>>>0)|(h<<p|f>>>(s>>>0))&A;v=A&h>>>(s>>>0);w=f<<q&y;x=(h<<q|f>>>(z>>>0))&y|f<<p&s-33>>31;break}if(e){k[e>>2]=r&f;k[e+4>>2]=0}if((i|0)==1){n=g|b&0;o=a|0|0;return (L=n,o)|0}else{r=uy(i|0)|0;n=h>>>(r>>>0)|0;o=h<<32-r|f>>>(r>>>0)|0;return (L=n,o)|0}}else{if(m){if(e){k[e>>2]=(h>>>0)%(i>>>0);k[e+4>>2]=0}n=0;o=(h>>>0)/(i>>>0)>>>0;return (L=n,o)|0}if(!f){if(e){k[e>>2]=0;k[e+4>>2]=(h>>>0)%(l>>>0)}n=0;o=(h>>>0)/(l>>>0)>>>0;return (L=n,o)|0}r=l-1|0;if(!(r&l)){if(e){k[e>>2]=a|0;k[e+4>>2]=r&h|b&0}n=0;o=h>>>((uy(l|0)|0)>>>0);return (L=n,o)|0}r=(ja(l|0)|0)-(ja(h|0)|0)|0;if(r>>>0<=30){s=r+1|0;p=31-r|0;t=s;u=h<<p|f>>>(s>>>0);v=h>>>(s>>>0);w=0;x=f<<p;break}if(!e){n=0;o=0;return (L=n,o)|0}k[e>>2]=a|0;k[e+4>>2]=g|b&0;n=0;o=0;return (L=n,o)|0}while(0);if(!t){B=x;C=w;D=v;E=u;F=0;G=0}else{b=c|0|0;c=j|d&0;d=ky(b|0,c|0,-1,-1)|0;j=L;g=x;x=w;w=v;v=u;u=t;t=0;do{a=g;g=x>>>31|g<<1;x=t|x<<1;f=v<<1|a>>>31|0;a=v>>>31|w<<1|0;jy(d,j,f,a)|0;h=L;l=h>>31|((h|0)<0?-1:0)<<1;t=l&1;v=jy(f,a,l&b,(((h|0)<0?-1:0)>>31|((h|0)<0?-1:0)<<1)&c)|0;w=L;u=u-1|0}while((u|0)!=0);B=g;C=x;D=w;E=v;F=0;G=t}t=C;C=0;if(e){k[e>>2]=E;k[e+4>>2]=D}n=(t|0)>>>31|(B|C)<<1|(C<<1|t>>>31)&0|F;o=(t<<1|0>>>31)&-2|G;return (L=n,o)|0}function Cy(a,b,c,d,e,f,g,h){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;return Ob[a&7](b|0,c|0,d|0,e|0,f|0,g|0,h|0)|0}function Dy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;return Pb[a&63](b|0,c|0,d|0)|0}function Ey(a,b,c,d,e,f){a=a|0;b=+b;c=+c;d=d|0;e=e|0;f=f|0;return Qb[a&63](+b,+c,d|0,e|0,f|0)|0}function Fy(a){a=a|0;return Rb[a&7]()|0}function Gy(a,b){a=a|0;b=b|0;Sb[a&31](b|0)}function Hy(a,b,c){a=a|0;b=b|0;c=c|0;Tb[a&1](b|0,c|0)}function Iy(a,b){a=a|0;b=b|0;return Ub[a&31](b|0)|0}function Jy(a,b,c,d,e,f,g,h,i,j){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=+h;i=i|0;j=j|0;Vb[a&1](b|0,c|0,d|0,e|0,f|0,g|0,+h,i|0,j|0)}function Ky(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;return Wb[a&7](b|0,c|0,d|0,e|0)|0}function Ly(a,b,c){a=a|0;b=b|0;c=c|0;return Xb[a&31](b|0,c|0)|0}function My(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;return Yb[a&1](b|0,c|0,d|0,e|0,f|0)|0}function Ny(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;ka(0);return 0}function Oy(a,b,c){a=a|0;b=b|0;c=c|0;ka(1);return 0}function Py(a,b,c,d,e){a=+a;b=+b;c=c|0;d=d|0;e=e|0;ka(2);return 0}function Qy(){ka(3);return 0}function Ry(a){a=a|0;ka(4)}function Sy(a,b){a=a|0;b=b|0;ka(5)}function Ty(a){a=a|0;ka(6);return 0}function Uy(a){a=a|0;return Ab(a|0)|0}function Vy(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=+g;h=h|0;i=i|0;ka(7)}function Wy(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;ka(8);return 0}function Xy(a,b){a=a|0;b=b|0;ka(9);return 0}function Yy(a,b){a=a|0;b=b|0;return Sa(a|0,b|0)|0}function Zy(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;ka(10);return 0}
function Nu(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,X=0,Y=0.0,Z=0.0;e=r;r=r+560|0;f=e+472|0;g=e+328|0;h=e+324|0;j=e+320|0;l=e+316|0;m=e+312|0;n=e+308|0;o=e+304|0;q=e+300|0;s=e+296|0;t=e+292|0;u=e+372|0;v=e+288|0;w=e+16|0;x=e;y=e+352|0;z=e+332|0;A=e+32|0;a:do if((b|0)==-99){if((Au(790037,A)|0)!=268){if(!(k[30835]|0))k[30835]=431;Bx(f,790042,80)|0;i[f+79>>0]=0;tk(f);B=-1;r=e;return B|0}C=k[A>>2]|0;D=k[30821]|0;do if((D|0)==(k[30822]|0)){E=k[30820]|0;if(!E){k[30822]=100;F=cy(34400)|0}else{k[30822]=D<<1;F=fy(E,D*688|0)|0}if(!F){k[30835]=113;G=-1;break a}else{k[30820]=F;H=k[30821]|0;break}}else H=D;while(0);k[30821]=H+1;if((H|0)>-1?(D=k[30820]|0,k[D+(H*344|0)>>2]=0-C,k[D+(H*344|0)+4>>2]=0,k[D+(H*344|0)+8>>2]=0,E=k[30828]|0,k[D+(H*344|0)+52>>2]=k[E+(C*124|0)+84>>2],k[D+(H*344|0)+56>>2]=k[E+(C*124|0)+88>>2],I=k[E+(C*124|0)+92>>2]|0,k[D+(H*344|0)+60>>2]=I,(I|0)>0):0){J=0;do{k[D+(H*344|0)+64+(J<<2)>>2]=k[E+(C*124|0)+96+(J<<2)>>2];J=J+1|0}while((J|0)!=(I|0));G=H}else G=H}else G=b;while(0);b=Iu(260,0,G)|0;G=k[30821]|0;do if((G|0)==(k[30822]|0)){H=k[30820]|0;if(!H){k[30822]=100;K=cy(34400)|0}else{k[30822]=G<<1;K=fy(H,G*688|0)|0}if(!K){k[30835]=113;L=-1;break}else{k[30820]=K;M=k[30821]|0;N=23;break}}else{M=G;N=23}while(0);if((N|0)==23){k[30821]=M+1;L=M}if((L|b|0)<0){B=-1;r=e;return B|0}M=k[30810]|0;k[g>>2]=M;bl(M,o)|0;k[t>>2]=0;do if(Mo(k[g>>2]|0,790084,w,0,t)|0){k[t>>2]=0;if(Mo(k[g>>2]|0,790093,w,0,t)|0){p[x>>3]=0.0;p[w>>3]=0.0;break}if(Mo(k[g>>2]|0,790102,x,0,t)|0)p[x>>3]=0.0}else p[x>>3]=0.0;while(0);b:do switch(i[a>>0]|0){case 0:{k[n>>2]=1;O=1;break}case 91:{M=1;while(1){P=a+M|0;G=i[P>>0]|0;Q=G<<24>>24==0;if(G<<24>>24!=93&(Q^1))M=M+1|0;else break}k[h>>2]=M;if(Q){if(!(k[30835]|0))k[30835]=431;Bx(f,790151,80)|0;i[f+79>>0]=0;tk(f);B=-1;r=e;return B|0}i[P>>0]=0;Gh(a+1|0,n,u,q,s,y,z,123340)|0;if(i[u>>0]|0){Jl(k[g>>2]|0,k[s>>2]|0,u,k[q>>2]|0,123340)|0;bl(k[g>>2]|0,n)|0;O=1;break b}G=k[n>>2]|0;if(G){K=k[g>>2]|0;H=G+1|0;k[n>>2]=H;Fk(K,H,m,123340)|0;O=1;break b}if(k[30835]|0){B=-1;r=e;return B|0}k[30835]=431;Bx(f,790111,80)|0;i[f+79>>0]=0;tk(f);B=-1;r=e;return B|0}case 43:{H=ex(a)|0;K=H+1|0;k[n>>2]=K;if((H|0)>0){Fk(k[g>>2]|0,K,m,123340)|0;O=1;break b}if(!(k[30835]|0))k[30835]=431;Bx(f,790111,80)|0;i[f+79>>0]=0;tk(f);B=-1;r=e;return B|0}default:if(!(Hh(g,a,0,123340)|0)){bl(k[g>>2]|0,n)|0;O=0}else O=0}while(0);if(k[30835]|0){B=-1;r=e;return B|0}c:do if((k[n>>2]|0)==1){k[n>>2]=2;d:do if(!(Fk(k[g>>2]|0,2,m,123340)|0))do{if(((k[m>>2]|0)!=0?(k[t>>2]=0,(Jo(k[g>>2]|0,790194,u,0,t)|0)==0):0)?(Ck(u),(Ex(u,790202)|0)!=0):0)break d;a=(k[n>>2]|0)+1|0;k[n>>2]=a}while((Fk(k[g>>2]|0,a,m,123340)|0)==0);while(0);switch(k[30835]|0){case 0:{break c;break}case 107:break;default:{B=-1;r=e;return B|0}}Bx(f,790206,80)|0;i[f+79>>0]=0;tk(f);B=-1;r=e;return B|0}while(0);Tk(k[g>>2]|0,0,c,j,123340)|0;Tk(k[g>>2]|0,0,d,l,123340)|0;if(k[30835]|0){B=-1;r=e;return B|0}k[t>>2]=0;d=w+8|0;do if(Mo(k[g>>2]|0,790084,d,0,t)|0){k[t>>2]=0;if(Mo(k[g>>2]|0,790093,d,0,t)|0){p[x+8>>3]=0.0;p[d>>3]=0.0;break}c=x+8|0;if(Mo(k[g>>2]|0,790102,c,0,t)|0)p[c>>3]=0.0}else p[x+8>>3]=0.0;while(0);t=k[30821]|0;do if((t|0)==(k[30822]|0)){c=k[30820]|0;if(!c){k[30822]=100;R=cy(34400)|0}else{k[30822]=t<<1;R=fy(c,t*688|0)|0}if(!R){k[30835]=113;S=-1;break}else{k[30820]=R;T=k[30821]|0;N=74;break}}else{T=t;N=74}while(0);if((N|0)==74){k[30821]=T+1;if((T|0)>-1){t=k[30820]|0;R=t+(T*344|0)|0;k[t+(T*344|0)+8>>2]=2;k[t+(T*344|0)+16>>2]=b;k[R>>2]=1032;c=t+(T*344|0)+4|0;k[c>>2]=14;k[t+(T*344|0)+52>>2]=258;k[t+(T*344|0)+56>>2]=k[t+(b*344|0)+56>>2];f=t+(b*344|0)+60|0;k[t+(T*344|0)+60>>2]=k[f>>2];k[h>>2]=0;if((k[f>>2]|0)>0){n=0;do{k[t+(T*344|0)+64+(n<<2)>>2]=k[t+(b*344|0)+64+(n<<2)>>2];n=n+1|0;k[h>>2]=n}while((n|0)<(k[f>>2]|0))}k[t+(T*344|0)+12>>2]=L;k[t+(L*344|0)>>2]=-1e3;k[t+(L*344|0)+4>>2]=0;f=t+(L*344|0)+88|0;k[f>>2]=0;if(Ko(k[g>>2]|0,790243,v,0,123340)|0){B=-1;r=e;return B|0}k[t+(L*344|0)+56>>2]=k[v>>2];n=k[v>>2]|0;if(n){u=cy(n<<4)|0;k[f>>2]=u;if(!u){k[30835]=113;B=-1;r=e;return B|0}qm(k[g>>2]|0,k[j>>2]|0,1,0,1,0,n,((n|0)<0)<<31>>31,0.0,u,h,123340)|0;u=k[v>>2]|0;qm(k[g>>2]|0,k[l>>2]|0,1,0,1,0,u,((u|0)<0)<<31>>31,0.0,(k[f>>2]|0)+(u<<3)|0,h,123340)|0;if(k[30835]|0){dy(k[f>>2]|0);B=-1;r=e;return B|0}u=t+(L*344|0)+52|0;k[u>>2]=1;L=k[v>>2]|0;k[h>>2]=L;v=L;while(1){t=v+-1|0;if(!t){N=88;break}l=v+-2|0;n=k[f>>2]|0;if(+p[n+(l<<3)>>3]>=+p[n+(t<<3)>>3]){U=n;V=t;N=87;break}if(!(+p[n+(L+l<<3)>>3]>=+p[n+(L+t<<3)>>3]))v=t;else{U=n;V=t;N=87;break}}if((N|0)==87){k[h>>2]=V;k[u>>2]=0;X=U}else if((N|0)==88){k[h>>2]=0;X=k[f>>2]|0}Y=+p[d>>3]-+p[w>>3]+(+p[x+8>>3]-+p[x>>3]);x=L<<1;Z=+p[X>>3];if(+W(+(Y/(+p[X+(x+-1<<3)>>3]-Z)))>1.0e-12?(k[h>>2]=0,(L|0)>0):0){w=L<<1;L=(w|0)>1;p[X>>3]=Y+Z;if((x|0)>1){d=1;do{p[X+(d<<3)>>3]=Y+ +p[X+(d<<3)>>3];d=d+1|0}while((d|0)<(x|0))}k[h>>2]=L?w:1}}if((k[(k[30820]|0)+(b*344|0)>>2]|0)==-1e3){Sb[k[c>>2]&31](R);S=T}else S=T}else S=T}T=k[g>>2]|0;if(!O){Fh(T,123340)|0;B=S;r=e;return B|0}else{Fk(T,k[o>>2]|0,m,123340)|0;B=S;r=e;return B|0}return 0}function Ou(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0;e=r;r=r+432|0;f=e+344|0;g=e+340|0;h=e+336|0;j=e+332|0;l=e+256|0;m=e+328|0;n=e;a:do if((b|0)==-99){if((Au(789744,n)|0)!=268){if(!(k[30835]|0))k[30835]=431;Bx(f,789746,80)|0;i[f+79>>0]=0;tk(f);o=-1;r=e;return o|0}p=k[n>>2]|0;q=k[30821]|0;do if((q|0)==(k[30822]|0)){s=k[30820]|0;if(!s){k[30822]=100;t=cy(34400)|0}else{k[30822]=q<<1;t=fy(s,q*688|0)|0}if(!t){k[30835]=113;u=-1;break a}else{k[30820]=t;v=k[30821]|0;break}}else v=q;while(0);k[30821]=v+1;if((v|0)>-1?(q=k[30820]|0,k[q+(v*344|0)>>2]=0-p,k[q+(v*344|0)+4>>2]=0,k[q+(v*344|0)+8>>2]=0,s=k[30828]|0,k[q+(v*344|0)+52>>2]=k[s+(p*124|0)+84>>2],k[q+(v*344|0)+56>>2]=k[s+(p*124|0)+88>>2],w=k[s+(p*124|0)+92>>2]|0,k[q+(v*344|0)+60>>2]=w,(w|0)>0):0){x=0;do{k[q+(v*344|0)+64+(x<<2)>>2]=k[s+(p*124|0)+96+(x<<2)>>2];x=x+1|0}while((x|0)!=(w|0));u=v}else u=v}else u=b;while(0);b:do if((c|0)==-99){if((Au(789785,n)|0)!=268){if(!(k[30835]|0))k[30835]=431;Bx(f,789787,80)|0;i[f+79>>0]=0;tk(f);o=-1;r=e;return o|0}b=k[n>>2]|0;v=k[30821]|0;do if((v|0)==(k[30822]|0)){t=k[30820]|0;if(!t){k[30822]=100;y=cy(34400)|0}else{k[30822]=v<<1;y=fy(t,v*688|0)|0}if(!y){k[30835]=113;z=-1;break b}else{k[30820]=y;A=k[30821]|0;break}}else A=v;while(0);k[30821]=A+1;if((A|0)>-1?(v=k[30820]|0,k[v+(A*344|0)>>2]=0-b,k[v+(A*344|0)+4>>2]=0,k[v+(A*344|0)+8>>2]=0,t=k[30828]|0,k[v+(A*344|0)+52>>2]=k[t+(b*124|0)+84>>2],k[v+(A*344|0)+56>>2]=k[t+(b*124|0)+88>>2],w=k[t+(b*124|0)+92>>2]|0,k[v+(A*344|0)+60>>2]=w,(w|0)>0):0){x=0;do{k[v+(A*344|0)+64+(x<<2)>>2]=k[t+(b*124|0)+96+(x<<2)>>2];x=x+1|0}while((x|0)!=(w|0));z=A}else z=A}else z=c;while(0);c=Iu(260,0,u)|0;u=Iu(260,0,z)|0;z=k[30821]|0;do if((z|0)==(k[30822]|0)){A=k[30820]|0;if(!A){k[30822]=100;B=cy(34400)|0}else{k[30822]=z<<1;B=fy(A,z*688|0)|0}if(!B){k[30835]=113;C=-1;break}else{k[30820]=B;D=k[30821]|0;E=38;break}}else{D=z;E=38}while(0);if((E|0)==38){k[30821]=D+1;C=D}D=u|c;if((C|D|0)<0){o=-1;r=e;return o|0}do if((D|0)>=0){z=k[30820]|0;B=k[z+(c*344|0)+56>>2]|0;if((B|0)!=1?(A=k[z+(u*344|0)+56>>2]|0,(A|0)!=1):0){if(!((B|0)==(A|0)?(k[z+(c*344|0)+52>>2]|0)==(k[z+(u*344|0)+52>>2]|0):0))break;A=k[z+(c*344|0)+60>>2]|0;if((A|0)!=(k[z+(u*344|0)+60>>2]|0))break;if((A|0)>0){B=0;y=1;while(1){n=(k[z+(c*344|0)+64+(B<<2)>>2]|0)==(k[z+(u*344|0)+64+(B<<2)>>2]|0)?y:0;B=B+1|0;if((B|0)==(A|0)){F=n;break}else y=n}if(!F)break}}y=k[30821]|0;do if((y|0)==(k[30822]|0)){if(!z){k[30822]=100;G=cy(34400)|0}else{k[30822]=y<<1;G=fy(z,y*688|0)|0}if(G){k[30820]=G;H=k[30821]|0;I=G;break}k[30835]=113;o=-1;r=e;return o|0}else{H=y;I=z}while(0);k[30821]=H+1;if((H|0)<=-1){o=H;r=e;return o|0}z=I+(H*344|0)|0;k[I+(H*344|0)+8>>2]=3;k[I+(H*344|0)+12>>2]=C;k[I+(H*344|0)+16>>2]=c;k[I+(H*344|0)+20>>2]=u;k[z>>2]=1033;y=I+(H*344|0)+4|0;k[y>>2]=15;k[I+(H*344|0)+52>>2]=258;k[I+(H*344|0)+56>>2]=1;k[I+(H*344|0)+60>>2]=1;k[I+(H*344|0)+64>>2]=1;if((H|c|0)>=0?(k[I+(H*344|0)+56>>2]=k[I+(c*344|0)+56>>2],A=I+(c*344|0)+60|0,k[I+(H*344|0)+60>>2]=k[A>>2],(k[A>>2]|0)>0):0){B=0;do{k[I+(H*344|0)+64+(B<<2)>>2]=k[I+(c*344|0)+64+(B<<2)>>2];B=B+1|0}while((B|0)<(k[A>>2]|0))}A=k[I+(u*344|0)+56>>2]|0;if(!((H|u|0)<0?1:(k[I+(c*344|0)+56>>2]|0)>=(A|0))?(k[I+(H*344|0)+56>>2]=A,A=I+(u*344|0)+60|0,k[I+(H*344|0)+60>>2]=k[A>>2],(k[A>>2]|0)>0):0){B=0;do{k[I+(H*344|0)+64+(B<<2)>>2]=k[I+(u*344|0)+64+(B<<2)>>2];B=B+1|0}while((B|0)<(k[A>>2]|0))}k[I+(C*344|0)>>2]=-1e3;k[I+(C*344|0)+4>>2]=0;k[h>>2]=0;k[g>>2]=0;A=i[d>>0]|0;switch(A<<24>>24){case 0:{B=Ru(I+(c*344|0)|0)|0;k[g>>2]=B;n=Ru(I+(u*344|0)|0)|0;k[h>>2]=n;if((n|B|0)<0){if(!(k[30835]|0))k[30835]=431;Bx(f,789987,80)|0;i[f+79>>0]=0;tk(f);w=k[30821]|0;if(!w){o=-1;r=e;return o|0}k[30821]=w+-1;o=-1;r=e;return o|0}else{J=B;K=n}break}case 32:{n=d;while(1){B=n+1|0;w=i[B>>0]|0;if(w<<24>>24==32)n=B;else{L=B;M=w;E=66;break}}break}default:{L=d;M=A;E=66}}do if((E|0)==66){n=M;w=L;c:while(1){switch(n<<24>>24){case 0:{N=w;break c;break}case 44:case 32:{O=w;E=70;break c;break}default:{}}B=w+1|0;n=i[B>>0]|0;w=B}if((E|0)==70){i[O>>0]=0;N=O+1|0}d:while(1){switch(i[N>>0]|0){case 0:{E=73;break d;break}case 32:break;default:{P=N;E=77;break d}}N=N+1|0}if((E|0)==73){if(!(k[30835]|0))k[30835]=431;Bx(f,789879,80)|0;i[f+79>>0]=0;tk(f);w=k[30821]|0;if(!w){o=-1;r=e;return o|0}k[30821]=w+-1;o=-1;r=e;return o|0}else if((E|0)==77){Tk(k[30810]|0,0,L,g,123340)|0;Tk(k[30810]|0,0,P,h,123340)|0;if(!(k[30835]|0)){J=k[g>>2]|0;K=k[h>>2]|0;break}Bx(f,789939,80)|0;i[f+79>>0]=0;tk(f);w=k[30821]|0;if(!w){o=-1;r=e;return o|0}k[30821]=w+-1;o=-1;r=e;return o|0}}while(0);k[l>>2]=0;e:do if((J|0)>0&(K|0)>0){k[j>>2]=0;Ps(k[30810]|0,J,K,l+8|0,l+16|0,l+24|0,l+32|0,l+40|0,l+48|0,l+56|0,l+64|0,j)|0;A=k[j>>2]|0;switch(A|0){case 505:{k[l>>2]=0;break e;break}case 0:{k[l>>2]=1;break e;break}default:{k[30835]=A;A=k[30821]|0;if(!A){o=-1;r=e;return o|0}k[30821]=A+-1;o=-1;r=e;return o|0}}}while(0);zs(a,l,m,123340)|0;if(k[30835]|0){A=k[30821]|0;if(!A){o=-1;r=e;return o|0}k[30821]=A+-1;o=-1;r=e;return o|0}k[I+(C*344|0)+88>>2]=k[m>>2];A=k[30820]|0;if((k[A+(c*344|0)>>2]|0)!=-1e3){o=H;r=e;return o|0}if((k[A+(u*344|0)>>2]|0)!=-1e3){o=H;r=e;return o|0}Sb[k[y>>2]&31](z);o=H;r=e;return o|0}while(0);if(!(k[30835]|0))k[30835]=431;Bx(f,789826,80)|0;i[f+79>>0]=0;tk(f);o=-1;r=e;return o|0}function Pu(a){a=a|0;var b=0,c=0,d=0,e=0;if(k[30835]|0)return;b=k[30820]|0;c=b+(a*344|0)|0;if((k[c>>2]|0)<=0)return;d=k[b+(a*344|0)+8>>2]|0;while(1){if(!d)break;d=d+-1|0;Pu(k[b+(a*344|0)+12+(d<<2)>>2]|0);if(k[30835]|0){e=7;break}}if((e|0)==7)return;Sb[k[b+(a*344|0)+4>>2]&31](c);return}function Qu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0.0,h=0,j=0,l=0.0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0.0,D=0,E=0.0,F=0,G=0,H=0,I=0.0,J=0,K=0.0,L=0,M=0,N=0.0,O=0,P=0.0,Q=0,R=0,S=0.0,T=0,U=0.0,V=0,W=0,X=0.0,Y=0,Z=0.0;b=k[30820]|0;c=k[a+12>>2]|0;d=k[a+16>>2]|0;e=k[a+20>>2]|0;f=b+(d*344|0)|0;if((k[f>>2]|0)==-1e3){g=+p[b+(d*344|0)+88>>3];h=0}else{g=0.0;h=k[b+(d*344|0)+56>>2]|0}j=b+(e*344|0)|0;if((k[j>>2]|0)==-1e3){l=+p[b+(e*344|0)+88>>3];m=0}else{l=0.0;m=k[b+(e*344|0)+56>>2]|0}n=(h|0)==0;o=(m|0)==0;if(m|h){Su(a);if((k[30835]|0)==0?(q=k[30825]|0,r=a+56|0,s=k[r>>2]|0,(q|0)!=0):0){t=(h|0)>1;h=b+(d*344|0)+88|0;u=b+(d*344|0)+84|0;v=(m|0)>1;m=b+(e*344|0)+88|0;w=b+(e*344|0)+84|0;x=a+84|0;y=b+(c*344|0)+88|0;z=a+88|0;A=q;B=0;C=g;D=0;E=l;F=ha(s,q)|0;q=s;while(1){A=A+-1|0;if(!q){G=0;H=B;I=C;J=D;K=E;L=F}else{s=q;M=B;N=C;O=D;P=E;Q=F;while(1){s=s+-1|0;Q=Q+-1|0;if(!t)if(n){R=M;S=N}else{R=i[(k[u>>2]|0)+A>>0]|0;S=+p[(k[h>>2]|0)+(A<<3)>>3]}else{R=i[(k[u>>2]|0)+Q>>0]|0;S=+p[(k[h>>2]|0)+(Q<<3)>>3]}if(!v)if(o){T=O;U=P}else{T=i[(k[w>>2]|0)+A>>0]|0;U=+p[(k[m>>2]|0)+(A<<3)>>3]}else{T=i[(k[w>>2]|0)+Q>>0]|0;U=+p[(k[m>>2]|0)+(Q<<3)>>3]}i[(k[x>>2]|0)+Q>>0]=(T|R)<<24>>24!=0&1;if(!(i[(k[x>>2]|0)+Q>>0]|0)){V=(Fs(S,U,k[y>>2]|0)|0)!=0&1;i[(k[z>>2]|0)+Q>>0]=V}if(!s){W=R;X=S;Y=T;Z=U;break}else{M=R;N=S;O=T;P=U}}G=k[r>>2]|0;H=W;I=X;J=Y;K=Z;L=F-q|0}if(!A)break;else{B=H;C=I;D=J;E=K;F=L;q=G}}}}else{G=(Fs(g,l,k[b+(c*344|0)+88>>2]|0)|0)!=0&1;i[a+88>>0]=G;k[a>>2]=-1e3}if((k[f>>2]|0)>0)dy(k[b+(d*344|0)+88>>2]|0);if((k[j>>2]|0)<=0)return;dy(k[b+(e*344|0)+88>>2]|0);return}function Ru(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0;b=k[a+8>>2]|0;if(!b){c=k[a>>2]|0;if((c|0)>0|(c|0)==-1e3){d=0;e=0}else return k[(k[30827]|0)+((0-c|0)*244|0)+4>>2]|0}else if((b|0)>0){c=k[30820]|0;f=k[30827]|0;g=0;h=0;i=0;while(1){j=c+((k[a+12+(h<<2)>>2]|0)*344|0)|0;l=k[j>>2]|0;do if((l|0)>0){m=Ru(j)|0;if((m|0)<1){n=g;o=i-m|0;break}if(!i){n=m;o=1}else{n=g;o=((g|0)!=(m|0)&1)+i|0}}else if((l|0)!=-1e3){m=k[f+((0-l|0)*244|0)+4>>2]|0;if(!i){n=m;o=1}else{n=g;o=((g|0)!=(m|0)&1)+i|0}}else{n=g;o=i}while(0);h=h+1|0;if((h|0)>=(b|0)){d=n;e=o;break}else{g=n;i=o}}}else{d=0;e=0}return ((e|0)==1?d:0-e|0)|0}function Su(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0;b=k[a+52>>2]|0;if((b+-261|0)>>>0>=2){c=ha(k[30825]|0,k[a+56>>2]|0)|0;switch(b|0){case 260:{d=8;break}case 259:{d=4;break}case 258:{d=1;break}default:d=1}e=ey(d+1|0,c)|0;k[a+88>>2]=e;if(!e){k[30835]=113;return}else{f=e+(ha(d,c)|0)|0;k[a+84>>2]=f;return}}f=k[30825]|0;c=cy(f<<2)|0;d=a+88|0;k[d>>2]=c;if(!c){k[30835]=113;return}e=k[a+56>>2]|0;g=cy(ha(e+2|0,f)|0)|0;k[c>>2]=g;g=k[d>>2]|0;c=k[g>>2]|0;if(!c){k[30835]=113;dy(g);return}if((f|0)>1){h=e+1|0;i=(f|0)>2;k[g+4>>2]=c+h;if((f|0)>2){c=1;g=2;while(1){j=k[d>>2]|0;k[j+(g<<2)>>2]=(k[j+(c<<2)>>2]|0)+h;j=g+1|0;if((j|0)<(f|0)){l=g;g=j;c=l}else break}}m=i?f+-1|0:1}else m=0;if((b|0)==261){k[a+84>>2]=(k[(k[d>>2]|0)+(m<<2)>>2]|0)+(e+1);return}else{k[a+84>>2]=0;return}}function Tu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0.0,A=0,B=0,C=0;b=k[30820]|0;c=k[a+12>>2]|0;d=k[a+16>>2]|0;e=k[b+(c*344|0)+56>>2]|0;f=k[b+(c*344|0)+88>>2]|0;g=k[b+(c*344|0)+52>>2]|0;c=b+(d*344|0)|0;a:do if((k[c>>2]|0)!=-1e3){Su(a);h=k[b+(d*344|0)+88>>2]|0;if(!(k[30835]|0)){j=ha(k[a+56>>2]|0,k[30825]|0)|0;if(!e){if(!j)break;l=a+88|0;m=a+84|0;n=j;while(1){n=n+-1|0;i[(k[l>>2]|0)+n>>0]=0;i[(k[m>>2]|0)+n>>0]=0;if(!n)break a}}n=b+(d*344|0)+84|0;m=a+84|0;l=(e|0)>15&(g|0)!=0;o=f+((e<<1)+-1<<3)|0;q=e>>1;r=e+-1|0;s=a+88|0;t=j;u=-1;while(1){v=t;while(1){w=v+-1|0;if(!v)break a;x=i[(k[n>>2]|0)+w>>0]|0;i[(k[m>>2]|0)+w>>0]=x;if(!(x<<24>>24)){y=w;break}else v=w}z=+p[h+(y<<3)>>3];if(((u|0)>=0?!(z<+p[f+(u<<3)>>3]):0)?!(z>+p[f+(u+e<<3)>>3]):0)A=u;else B=27;b:do if((B|0)==27){B=0;if(!l){v=r;while(1){if(+p[f+(v<<3)>>3]<=z?+p[f+(v+e<<3)>>3]>=z:0){A=v;break b}w=v+-1|0;if(!v){A=w;break b}else v=w}}if(+p[f>>3]<=z?+p[o>>3]>=z:0){v=q;w=q;while(1){x=w>>((w|0)>1&1);if(+p[f+(v+e<<3)>>3]<z){if(!(+p[f+(v+1<<3)>>3]<=z)){A=-1;break b}v=x+v|0;w=x;continue}if(!(+p[f+(v<<3)>>3]>z)){A=v;break b}if(!(+p[f+(r+v<<3)>>3]>=z)){A=-1;break b}v=v-x|0;w=x}}else A=-1}while(0);i[(k[s>>2]|0)+y>>0]=A>>>31&255^1;t=y;u=A}}}else{z=+p[b+(d*344|0)+88>>3];c:do if((e|0)>15&(g|0)!=0)if(+p[f>>3]<=z?+p[f+((e<<1)+-1<<3)>>3]>=z:0){u=e>>1;t=e+-1|0;s=u;r=u;while(1){u=r>>((r|0)>1&1);if(+p[f+(s+e<<3)>>3]<z){if(!(+p[f+(s+1<<3)>>3]<=z)){C=-1;break c}s=u+s|0;r=u;continue}if(!(+p[f+(s<<3)>>3]>z)){C=s;break c}if(!(+p[f+(t+s<<3)>>3]>=z)){C=-1;break c}s=s-u|0;r=u}}else C=-1;else{r=e+-1|0;if(!e)C=r;else{s=r;while(1){if(+p[f+(s<<3)>>3]<=z?+p[f+(s+e<<3)>>3]>=z:0){C=s;break c}r=s+-1|0;if(!s){C=r;break}else s=r}}}while(0);i[a+88>>0]=C>>>31&255^1;k[a>>2]=-1e3}while(0);if((k[c>>2]|0)<=0)return;dy(k[b+(d*344|0)+88>>2]|0);return}function Uu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0.0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,Y=0.0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0.0,pa=0,qa=0,ra=0.0,sa=0,ta=0.0,ua=0,va=0,wa=0.0,xa=0,ya=0,za=0,Aa=0.0,Ba=0,Ca=0,Da=0,Ea=0.0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0.0,lb=0.0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0.0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0.0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0.0,Tb=0.0,Ub=0.0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0,gc=0,hc=0,ic=0,jc=0,kc=0.0,lc=0,mc=0;b=r;r=r+3056|0;c=b+2976|0;d=b+2920|0;e=b+2880|0;f=b;g=b+2960|0;h=a+8|0;j=k[h>>2]|0;a:do if(!j)l=1;else{m=j;n=1;while(1){o=m;while(1){o=o+-1|0;q=k[30820]|0;s=k[a+12+(o<<2)>>2]|0;t=q+(s*344|0)|0;k[d+(o<<2)>>2]=t;u=(k[t>>2]|0)!=-1e3;t=e+(o<<2)|0;k[t>>2]=u&1;if(u){v=o;w=q;x=s;y=t;break}switch(k[q+(s*344|0)+52>>2]|0){case 260:{p[f+(o*288|0)+32>>3]=+p[q+(s*344|0)+88>>3];break}case 259:{k[f+(o*288|0)+32>>2]=k[q+(s*344|0)+88>>2];break}case 258:{i[f+(o*288|0)+32>>0]=i[q+(s*344|0)+88>>0]|0;break}default:ux(f+(o*288|0)+32|0,q+(s*344|0)+88|0)|0}i[g+o>>0]=0;if(!o){l=n;break a}}k[y>>2]=k[w+(x*344|0)+56>>2];if(!v){l=0;break}else{m=v;n=0}}}while(0);v=k[a>>2]|0;b:do if((l|0)==0|((v|0)==1001|((k[h>>2]|0)==0|(v&-2|0)==1042))){Su(a);x=k[30825]|0;w=a+56|0;y=k[w>>2]|0;j=ha(y,x)|0;if(!(k[30835]|0))do switch(k[a>>2]|0){case 1035:{if(!x)break b;n=a+88|0;m=a+84|0;o=x;do{o=o+-1|0;k[(k[n>>2]|0)+(o<<2)>>2]=(k[30824]|0)+o;i[(k[m>>2]|0)+o>>0]=0}while((o|0)!=0);break}case 1036:{switch(k[a+52>>2]|0){case 259:{if(!x)break b;o=a+88|0;m=a+84|0;n=x;do{n=n+-1|0;k[(k[o>>2]|0)+(n<<2)>>2]=0;i[(k[m>>2]|0)+n>>0]=1}while((n|0)!=0);break}case 261:{if(!x)break b;n=a+88|0;m=a+84|0;o=x;do{o=o+-1|0;i[k[(k[n>>2]|0)+(o<<2)>>2]>>0]=0;i[(k[m>>2]|0)+o>>0]=1}while((o|0)!=0);break}default:break b}break}case 1001:{if(!j)break b;o=a+88|0;m=a+84|0;n=j;do{n=n+-1|0;z=+fv();p[(k[o>>2]|0)+(n<<3)>>3]=z;i[(k[m>>2]|0)+n>>0]=0}while((n|0)!=0);break}case 1042:{if(!j)break b;n=a+88|0;m=a+84|0;o=j;do{o=o+-1|0;z=+gv();p[(k[n>>2]|0)+(o<<3)>>3]=z;i[(k[m>>2]|0)+o>>0]=0}while((o|0)!=0);break}case 1043:{o=k[d>>2]|0;m=(k[o>>2]|0)==-1e3;n=(j|0)==0;if((k[o+52>>2]|0)==260){if(m){if(n)break b;s=f+32|0;q=a+84|0;t=a+88|0;u=j;while(1){u=u+-1|0;i[(k[q>>2]|0)+u>>0]=+p[s>>3]<0.0&1;if(!(i[(k[q>>2]|0)+u>>0]|0)){A=hv(+p[s>>3])|0;k[(k[t>>2]|0)+(u<<2)>>2]=A}if(!u)break b}}if(n)break b;u=a+84|0;t=a+88|0;s=j;q=o;while(1){A=s+-1|0;i[(k[u>>2]|0)+A>>0]=i[(k[q+84>>2]|0)+A>>0]|0;if(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3]<0.0)i[(k[u>>2]|0)+A>>0]=1;if(!(i[(k[u>>2]|0)+A>>0]|0)){B=hv(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3])|0;k[(k[t>>2]|0)+(A<<2)>>2]=B}if(!A)break b;s=A;q=k[d>>2]|0}}else{if(m){if(n)break b;q=f+32|0;s=a+84|0;t=a+88|0;u=j;while(1){u=u+-1|0;i[(k[s>>2]|0)+u>>0]=(k[q>>2]|0)>>>31;if(!(i[(k[s>>2]|0)+u>>0]|0)){A=hv(+(k[q>>2]|0))|0;k[(k[t>>2]|0)+(u<<2)>>2]=A}if(!u)break b}}if(n)break b;u=a+84|0;t=a+88|0;q=j;s=o;while(1){m=q+-1|0;i[(k[u>>2]|0)+m>>0]=i[(k[s+84>>2]|0)+m>>0]|0;if((k[(k[(k[d>>2]|0)+88>>2]|0)+(m<<2)>>2]|0)<0)i[(k[u>>2]|0)+m>>0]=1;if(!(i[(k[u>>2]|0)+m>>0]|0)){A=hv(+(k[(k[(k[d>>2]|0)+88>>2]|0)+(m<<2)>>2]|0))|0;k[(k[t>>2]|0)+(m<<2)>>2]=A}if(!m)break b;q=m;s=k[d>>2]|0}}break}case 1002:{s=k[d>>2]|0;q=ha(k[s+56>>2]|0,x)|0;switch(k[s+52>>2]|0){case 258:{if(!x)break b;t=a+88|0;u=a+84|0;o=x;n=q;while(1){o=o+-1|0;k[(k[t>>2]|0)+(o<<2)>>2]=0;i[(k[u>>2]|0)+o>>0]=1;m=k[d>>2]|0;A=k[m+56>>2]|0;if(!A)C=n;else{B=A;D=m;m=n;while(1){E=B+-1|0;F=m+-1|0;if(!(i[(k[D+84>>2]|0)+F>>0]|0)){G=(k[t>>2]|0)+(o<<2)|0;k[G>>2]=(k[G>>2]|0)+((i[(k[D+88>>2]|0)+F>>0]|0)!=0&1);i[(k[u>>2]|0)+o>>0]=0}if(!E)break;B=E;D=k[d>>2]|0;m=F}C=n-A|0}if(!o)break;else n=C}break}case 259:{if(!x)break b;n=a+88|0;o=a+84|0;u=x;t=q;while(1){u=u+-1|0;k[(k[n>>2]|0)+(u<<2)>>2]=0;i[(k[o>>2]|0)+u>>0]=1;m=k[d>>2]|0;D=k[m+56>>2]|0;if(!D)H=t;else{B=D;F=m;m=t;while(1){E=B+-1|0;G=m+-1|0;if(!(i[(k[F+84>>2]|0)+G>>0]|0)){I=(k[n>>2]|0)+(u<<2)|0;k[I>>2]=(k[I>>2]|0)+(k[(k[F+88>>2]|0)+(G<<2)>>2]|0);i[(k[o>>2]|0)+u>>0]=0}if(!E)break;B=E;F=k[d>>2]|0;m=G}H=t-D|0}if(!u)break;else t=H}break}case 260:{if(!x)break b;t=a+88|0;u=a+84|0;o=x;n=q;while(1){o=o+-1|0;p[(k[t>>2]|0)+(o<<3)>>3]=0.0;i[(k[u>>2]|0)+o>>0]=1;m=k[d>>2]|0;F=k[m+56>>2]|0;if(!F)J=n;else{B=F;A=m;m=n;while(1){G=B+-1|0;E=m+-1|0;if(!(i[(k[A+84>>2]|0)+E>>0]|0)){I=(k[t>>2]|0)+(o<<3)|0;p[I>>3]=+p[(k[A+88>>2]|0)+(E<<3)>>3]+ +p[I>>3];i[(k[u>>2]|0)+o>>0]=0}if(!G)break;B=G;A=k[d>>2]|0;m=E}J=n-F|0}if(!o)break;else n=J}break}default:{if(!x)break b;n=a+88|0;o=a+84|0;u=x;t=s;while(1){q=u+-1|0;m=k[(k[t+88>>2]|0)+(q<<2)>>2]|0;k[(k[n>>2]|0)+(q<<2)>>2]=0;i[(k[o>>2]|0)+q>>0]=0;A=m;c:while(1){switch(i[A>>0]|0){case 0:{break c;break}case 49:{m=(k[n>>2]|0)+(q<<2)|0;k[m>>2]=(k[m>>2]|0)+1;break}default:{}}A=A+1|0}if(!q)break b;u=q;t=k[d>>2]|0}}}break}case 1038:{t=k[d>>2]|0;u=k[t+56>>2]|0;n=ha(u,x)|0;switch(k[t+52>>2]|0){case 259:{if(!x)break b;o=a+88|0;s=a+84|0;A=x;F=u;m=t;B=n;while(1){D=A+-1|0;E=(k[o>>2]|0)+(D<<3)|0;p[E>>3]=0.0;G=m+84|0;I=m+88|0;z=0.0;K=0;L=B;M=F;d:while(1){N=L;O=M;while(1){P=O+-1|0;if(!O){Q=K;R=N;break d}S=N+-1|0;if(!(i[(k[G>>2]|0)+S>>0]|0)){T=S;U=P;break}else{N=S;O=P}}Y=+(k[(k[I>>2]|0)+(T<<2)>>2]|0)+z;p[E>>3]=Y;z=Y;K=K+1|0;L=T;M=U}M=(k[s>>2]|0)+D|0;if(!Q)i[M>>0]=1;else{i[M>>0]=0;M=(k[o>>2]|0)+(D<<3)|0;p[M>>3]=+p[M>>3]/+(Q|0)}if(!D)break b;M=k[d>>2]|0;A=D;F=k[M+56>>2]|0;m=M;B=R}break}case 260:{if(!x)break b;B=a+88|0;m=a+84|0;F=x;A=u;o=t;s=n;while(1){M=F+-1|0;L=(k[B>>2]|0)+(M<<3)|0;p[L>>3]=0.0;K=o+84|0;E=o+88|0;z=0.0;I=0;G=s;q=A;e:while(1){O=G;N=q;while(1){P=N+-1|0;if(!N){ia=I;ja=O;break e}S=O+-1|0;if(!(i[(k[K>>2]|0)+S>>0]|0)){ka=P;la=S;break}else{O=S;N=P}}Y=+p[(k[E>>2]|0)+(la<<3)>>3]+z;p[L>>3]=Y;z=Y;I=I+1|0;G=la;q=ka}q=(k[m>>2]|0)+M|0;if(!ia)i[q>>0]=1;else{i[q>>0]=0;q=(k[B>>2]|0)+(M<<3)|0;p[q>>3]=+p[q>>3]/+(ia|0)}if(!M)break b;q=k[d>>2]|0;F=M;A=k[q+56>>2]|0;o=q;s=ja}break}default:break b}break}case 1039:{s=k[d>>2]|0;o=k[s+56>>2]|0;A=ha(o,x)|0;switch(k[s+52>>2]|0){case 259:{if(!x)break b;F=a+84|0;B=a+88|0;m=x;n=s;t=o;u=A;while(1){q=n+84|0;G=n+88|0;I=0;L=u;E=t;z=0.0;f:while(1){K=L;D=E;while(1){N=D+-1|0;if(!D){ma=I;na=K;oa=z;break f}O=K+-1|0;if(!(i[(k[q>>2]|0)+O>>0]|0)){pa=N;qa=O;break}else{K=O;D=N}}I=I+1|0;L=qa;E=pa;z=z+ +(k[(k[G>>2]|0)+(qa<<2)>>2]|0)}G=m+-1|0;if((ma|0)>1){z=+(ma|0);Y=oa/z;E=n+88|0;L=t+na|0;I=t;ra=0.0;g:while(1){M=L;D=I;while(1){K=D+-1|0;if(!D){sa=M;ta=ra;break g}N=M+-1|0;if(!(i[(k[q>>2]|0)+N>>0]|0)){ua=K;va=N;break}else{M=N;D=K}}wa=+(k[(k[E>>2]|0)+(va<<2)>>2]|0)-Y;L=va;I=ua;ra=ra+wa*wa}i[(k[F>>2]|0)+G>>0]=0;ra=+X(+(ta/(z+-1.0)));p[(k[B>>2]|0)+(G<<3)>>3]=ra;xa=sa}else{i[(k[F>>2]|0)+G>>0]=0;p[(k[B>>2]|0)+(G<<3)>>3]=0.0;xa=na}if(!G)break b;I=k[d>>2]|0;m=G;n=I;t=k[I+56>>2]|0;u=xa}break}case 260:{if(!x)break b;u=a+84|0;t=a+88|0;n=x;m=s;B=o;F=A;while(1){I=m+84|0;L=m+88|0;E=0;q=F;D=B;ra=0.0;h:while(1){M=q;K=D;while(1){N=K+-1|0;if(!K){ya=E;za=M;Aa=ra;break h}O=M+-1|0;if(!(i[(k[I>>2]|0)+O>>0]|0)){Ba=N;Ca=O;break}else{M=O;K=N}}E=E+1|0;q=Ca;D=Ba;ra=ra+ +p[(k[L>>2]|0)+(Ca<<3)>>3]}L=n+-1|0;if((ya|0)>1){ra=+(ya|0);z=Aa/ra;D=m+88|0;q=B+za|0;E=B;Y=0.0;i:while(1){G=q;K=E;while(1){M=K+-1|0;if(!K){Da=G;Ea=Y;break i}N=G+-1|0;if(!(i[(k[I>>2]|0)+N>>0]|0)){Fa=M;Ga=N;break}else{G=N;K=M}}wa=+p[(k[D>>2]|0)+(Ga<<3)>>3]-z;q=Ga;E=Fa;Y=Y+wa*wa}i[(k[u>>2]|0)+L>>0]=0;Y=+X(+(Ea/(ra+-1.0)));p[(k[t>>2]|0)+(L<<3)>>3]=Y;Ha=Da}else{i[(k[u>>2]|0)+L>>0]=0;p[(k[t>>2]|0)+(L<<3)>>3]=0.0;Ha=za}if(!L)break b;E=k[d>>2]|0;n=L;m=E;B=k[E+56>>2]|0;F=Ha}break}default:break b}break}case 1037:{F=k[d>>2]|0;B=k[F+56>>2]|0;m=F+88|0;if((k[F+52>>2]|0)==259){n=k[m>>2]|0;t=k[F+84>>2]|0;u=cy(B<<2)|0;if(!u){k[30835]=431;Bx(c,790441,80)|0;i[c+79>>0]=0;tk(c);dy(k[a+88>>2]|0);break b}if((x|0)>0){A=(B|0)==0;o=u;s=a+84|0;E=a+88|0;q=n;n=0;D=t;while(1){if(A){Ia=q;Ja=u;Ka=D}else{t=D+B|0;I=B;K=q;G=u;M=D;while(1){I=I+-1|0;if(!(i[M>>0]|0)){k[G>>2]=k[K>>2];La=G+4|0}else La=G;if(!I){Ma=La;break}else{K=K+4|0;G=La;M=M+1|0}}Ia=q+(B<<2)|0;Ja=Ma;Ka=t}M=Ja-o|0;G=(k[s>>2]|0)+n|0;if((M|0)>0){i[G>>0]=0;K=Fu(u,M>>2)|0;k[(k[E>>2]|0)+(n<<2)>>2]=K}else{i[G>>0]=1;k[(k[E>>2]|0)+(n<<2)>>2]=0}n=n+1|0;if((n|0)==(x|0))break;else{q=Ia;D=Ka}}}dy(u);break b}else{D=k[m>>2]|0;q=k[F+84>>2]|0;n=cy(B<<3)|0;if(!n){k[30835]=431;Bx(c,790441,80)|0;i[c+79>>0]=0;tk(c);dy(k[a+88>>2]|0);break b}if((x|0)>0){E=(B|0)==0;s=n;o=a+84|0;A=a+88|0;G=D;D=0;K=q;while(1){if(E){Na=G;Oa=n;Pa=K}else{q=K+B|0;M=B;I=G;L=n;N=K;while(1){M=M+-1|0;if(!(i[N>>0]|0)){p[L>>3]=+p[I>>3];Qa=L+8|0}else Qa=L;if(!M){Ra=Qa;break}else{I=I+8|0;L=Qa;N=N+1|0}}Na=G+(B<<3)|0;Oa=Ra;Pa=q}N=Oa-s|0;L=(k[o>>2]|0)+D|0;if((N|0)>0){i[L>>0]=0;ra=+Gu(n,N>>3);p[(k[A>>2]|0)+(D<<3)>>3]=ra}else{i[L>>0]=1;p[(k[A>>2]|0)+(D<<3)>>3]=0.0}D=D+1|0;if((D|0)==(x|0))break;else{G=Na;K=Pa}}}dy(n);break b}break}case 1017:{K=k[d>>2]|0;G=(j|0)==0;if((k[K+52>>2]|0)==260){if(G)break b;D=a+88|0;A=a+84|0;o=j;s=K;while(1){B=o+-1|0;ra=+p[(k[s+88>>2]|0)+(B<<3)>>3];p[(k[D>>2]|0)+(B<<3)>>3]=ra>0.0?ra:-ra;i[(k[A>>2]|0)+B>>0]=i[(k[s+84>>2]|0)+B>>0]|0;if(!B)break b;o=B;s=k[d>>2]|0}}else{if(G)break b;s=a+88|0;o=a+84|0;A=j;D=K;while(1){n=A+-1|0;B=k[(k[D+88>>2]|0)+(n<<2)>>2]|0;k[(k[s>>2]|0)+(n<<2)>>2]=(B|0)>0?B:0-B|0;i[(k[o>>2]|0)+n>>0]=i[(k[D+84>>2]|0)+n>>0]|0;if(!n)break b;A=n;D=k[d>>2]|0}}break}case 1040:{D=k[d>>2]|0;A=(k[D+52>>2]|0)==261?1:k[D+56>>2]|0;if(!x)break b;D=a+84|0;o=a+88|0;s=(A|0)==0;K=x;G=ha(A,x)|0;while(1){K=K+-1|0;i[(k[D>>2]|0)+K>>0]=0;n=(k[o>>2]|0)+(K<<2)|0;k[n>>2]=0;if(s)Sa=G;else{B=k[(k[d>>2]|0)+84>>2]|0;E=A;F=0;m=G;while(1){E=E+-1|0;m=m+-1|0;if(!(i[B+m>>0]|0)){u=F+1|0;k[n>>2]=u;Ta=u}else Ta=F;if(!E)break;else F=Ta}Sa=G-A|0}if(!K)break;else G=Sa}break}case 1030:{G=k[d>>2]|0;K=(k[G+52>>2]|0)==261?x:j;if(!K)break b;A=a+88|0;s=a+84|0;o=K+-1|0;i[(k[A>>2]|0)+o>>0]=i[(k[G+84>>2]|0)+o>>0]|0;i[(k[s>>2]|0)+o>>0]=0;if(!o)break b;else Ua=o;do{Ua=Ua+-1|0;i[(k[A>>2]|0)+Ua>>0]=i[(k[(k[d>>2]|0)+84>>2]|0)+Ua>>0]|0;i[(k[s>>2]|0)+Ua>>0]=0}while((Ua|0)!=0);break}case 1031:{switch(k[a+52>>2]|0){case 258:{if(!x)break b;s=a+84|0;A=f+32|0;o=a+88|0;G=g+1|0;K=f+320|0;D=e+4|0;F=d+4|0;E=d+4|0;n=x;m=y;B=j;while(1){q=n+-1|0;if(!m)Va=B;else{u=m;L=B;do{L=L+-1|0;N=k[D>>2]|0;if((N|0)<=1){if(N){N=k[E>>2]|0;i[G>>0]=i[(k[N+84>>2]|0)+q>>0]|0;i[K>>0]=i[(k[N+88>>2]|0)+q>>0]|0}}else{N=k[F>>2]|0;i[G>>0]=i[(k[N+84>>2]|0)+L>>0]|0;i[K>>0]=i[(k[N+88>>2]|0)+L>>0]|0}N=k[e>>2]|0;do if((N|0)<=1)if(!N){Wa=i[g>>0]|0;break}else{I=k[d>>2]|0;M=i[(k[I+84>>2]|0)+q>>0]|0;i[g>>0]=M;i[A>>0]=i[(k[I+88>>2]|0)+q>>0]|0;Wa=M;break}else{M=k[d>>2]|0;I=i[(k[M+84>>2]|0)+L>>0]|0;i[g>>0]=I;i[A>>0]=i[(k[M+88>>2]|0)+L>>0]|0;Wa=I}while(0);u=u+-1|0;if(!(Wa<<24>>24)){i[(k[s>>2]|0)+L>>0]=0;i[(k[o>>2]|0)+L>>0]=i[A>>0]|0}else{i[(k[s>>2]|0)+L>>0]=i[G>>0]|0;i[(k[o>>2]|0)+L>>0]=i[K>>0]|0}}while((u|0)!=0);Va=B-m|0}if(!q)break b;n=q;m=k[w>>2]|0;B=Va}break}case 259:{if(!x)break b;B=a+84|0;m=f+32|0;n=a+88|0;K=g+1|0;o=f+320|0;G=e+4|0;s=d+4|0;A=d+4|0;F=x;E=y;D=j;while(1){u=F+-1|0;if(!E)Xa=D;else{L=E;N=D;do{N=N+-1|0;I=k[G>>2]|0;if((I|0)<=1){if(I){I=k[A>>2]|0;i[K>>0]=i[(k[I+84>>2]|0)+u>>0]|0;k[o>>2]=k[(k[I+88>>2]|0)+(u<<2)>>2]}}else{I=k[s>>2]|0;i[K>>0]=i[(k[I+84>>2]|0)+N>>0]|0;k[o>>2]=k[(k[I+88>>2]|0)+(N<<2)>>2]}I=k[e>>2]|0;do if((I|0)<=1)if(!I){Ya=i[g>>0]|0;break}else{M=k[d>>2]|0;t=i[(k[M+84>>2]|0)+u>>0]|0;i[g>>0]=t;k[m>>2]=k[(k[M+88>>2]|0)+(u<<2)>>2];Ya=t;break}else{t=k[d>>2]|0;M=i[(k[t+84>>2]|0)+N>>0]|0;i[g>>0]=M;k[m>>2]=k[(k[t+88>>2]|0)+(N<<2)>>2];Ya=M}while(0);L=L+-1|0;if(!(Ya<<24>>24)){i[(k[B>>2]|0)+N>>0]=0;k[(k[n>>2]|0)+(N<<2)>>2]=k[m>>2]}else{i[(k[B>>2]|0)+N>>0]=i[K>>0]|0;k[(k[n>>2]|0)+(N<<2)>>2]=k[o>>2]}}while((L|0)!=0);Xa=D-E|0}if(!u)break b;F=u;E=k[w>>2]|0;D=Xa}break}case 260:{if(!x)break b;D=a+84|0;E=f+32|0;F=a+88|0;o=g+1|0;n=f+320|0;K=e+4|0;B=d+4|0;m=d+4|0;s=x;A=y;G=j;while(1){L=s+-1|0;if(!A)Za=G;else{N=A;q=G;do{q=q+-1|0;I=k[K>>2]|0;if((I|0)<=1){if(I){I=k[m>>2]|0;i[o>>0]=i[(k[I+84>>2]|0)+L>>0]|0;p[n>>3]=+p[(k[I+88>>2]|0)+(L<<3)>>3]}}else{I=k[B>>2]|0;i[o>>0]=i[(k[I+84>>2]|0)+q>>0]|0;p[n>>3]=+p[(k[I+88>>2]|0)+(q<<3)>>3]}I=k[e>>2]|0;do if((I|0)<=1)if(!I){_a=i[g>>0]|0;break}else{M=k[d>>2]|0;t=i[(k[M+84>>2]|0)+L>>0]|0;i[g>>0]=t;p[E>>3]=+p[(k[M+88>>2]|0)+(L<<3)>>3];_a=t;break}else{t=k[d>>2]|0;M=i[(k[t+84>>2]|0)+q>>0]|0;i[g>>0]=M;p[E>>3]=+p[(k[t+88>>2]|0)+(q<<3)>>3];_a=M}while(0);N=N+-1|0;if(!(_a<<24>>24)){i[(k[D>>2]|0)+q>>0]=0;p[(k[F>>2]|0)+(q<<3)>>3]=+p[E>>3]}else{i[(k[D>>2]|0)+q>>0]=i[o>>0]|0;p[(k[F>>2]|0)+(q<<3)>>3]=+p[n>>3]}}while((N|0)!=0);Za=G-A|0}if(!L)break b;s=L;A=k[w>>2]|0;G=Za}break}case 261:{if(!x)break b;G=a+84|0;A=a+88|0;s=f+32|0;n=g+1|0;F=f+320|0;o=e+4|0;D=d+4|0;E=x;do{E=E+-1|0;if(k[o>>2]|0){B=k[D>>2]|0;i[n>>0]=i[(k[B+84>>2]|0)+E>>0]|0;ux(F,k[(k[B+88>>2]|0)+(E<<2)>>2]|0)|0}if(k[e>>2]|0){B=k[d>>2]|0;i[g>>0]=i[(k[B+84>>2]|0)+E>>0]|0;ux(s,k[(k[B+88>>2]|0)+(E<<2)>>2]|0)|0}if(!(i[g>>0]|0)){i[(k[G>>2]|0)+j>>0]=0;ux(k[(k[A>>2]|0)+(E<<2)>>2]|0,s)|0}else{i[(k[G>>2]|0)+E>>0]=i[n>>0]|0;ux(k[(k[A>>2]|0)+(E<<2)>>2]|0,F)|0}}while((E|0)!=0);break}default:break b}break}case 1004:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;if(!(n<<24>>24)){ra=+_(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3]));p[(k[F>>2]|0)+(A<<3)>>3]=ra}}while((A|0)!=0);break}case 1005:{if(!j)break b;A=a+84|0;F=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[A>>2]|0)+E>>0]=n;if(!(n<<24>>24)){ra=+Z(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3]));p[(k[F>>2]|0)+(E<<3)>>3]=ra}}while((E|0)!=0);break}case 1006:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;if(!(n<<24>>24)){ra=+$(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3]));p[(k[F>>2]|0)+(A<<3)>>3]=ra}}while((A|0)!=0);break}case 1007:{if(!j)break b;A=a+84|0;F=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[A>>2]|0)+E>>0]=n;do if(!(n<<24>>24)){ra=+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3];if(ra<-1.0|ra>1.0){p[(k[F>>2]|0)+(E<<3)>>3]=0.0;i[(k[A>>2]|0)+E>>0]=1;break}else{Y=+ba(+ra);p[(k[F>>2]|0)+(E<<3)>>3]=Y;break}}while(0)}while((E|0)!=0);break}case 1008:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;do if(!(n<<24>>24)){Y=+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3];if(Y<-1.0|Y>1.0){p[(k[F>>2]|0)+(A<<3)>>3]=0.0;i[(k[E>>2]|0)+A>>0]=1;break}else{ra=+aa(+Y);p[(k[F>>2]|0)+(A<<3)>>3]=ra;break}}while(0)}while((A|0)!=0);break}case 1009:{if(!j)break b;A=a+84|0;F=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[A>>2]|0)+E>>0]=n;if(!(n<<24>>24)){ra=+ca(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3]));p[(k[F>>2]|0)+(E<<3)>>3]=ra}}while((E|0)!=0);break}case 1010:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;if(!(n<<24>>24)){ra=+Nv(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3]);p[(k[F>>2]|0)+(A<<3)>>3]=ra}}while((A|0)!=0);break}case 1011:{if(!j)break b;A=a+84|0;F=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[A>>2]|0)+E>>0]=n;if(!(n<<24>>24)){ra=+Ev(+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3]);p[(k[F>>2]|0)+(E<<3)>>3]=ra}}while((E|0)!=0);break}case 1012:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;if(!(n<<24>>24)){ra=+Ov(+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3]);p[(k[F>>2]|0)+(A<<3)>>3]=ra}}while((A|0)!=0);break}case 1013:{if(!j)break b;A=a+84|0;F=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[A>>2]|0)+E>>0]=n;if(!(n<<24>>24)){ra=+ea(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3]));p[(k[F>>2]|0)+(E<<3)>>3]=ra}}while((E|0)!=0);break}case 1014:{if(!j)break b;E=a+84|0;F=a+88|0;A=j;do{A=A+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+A>>0]|0;i[(k[E>>2]|0)+A>>0]=n;do if(!(n<<24>>24)){ra=+p[(k[(k[d>>2]|0)+88>>2]|0)+(A<<3)>>3];if(!(ra<=0.0)){Y=+fa(+ra);p[(k[F>>2]|0)+(A<<3)>>3]=Y;break}else{p[(k[F>>2]|0)+(A<<3)>>3]=0.0;i[(k[E>>2]|0)+A>>0]=1;break}}while(0)}while((A|0)!=0);break}case 1015:{if(!j)break b;A=a+84|0;E=a+88|0;F=j;do{F=F+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+F>>0]|0;i[(k[A>>2]|0)+F>>0]=n;do if(!(n<<24>>24)){Y=+p[(k[(k[d>>2]|0)+88>>2]|0)+(F<<3)>>3];if(!(Y<=0.0)){ra=+Kv(Y);p[(k[E>>2]|0)+(F<<3)>>3]=ra;break}else{p[(k[E>>2]|0)+(F<<3)>>3]=0.0;i[(k[A>>2]|0)+F>>0]=1;break}}while(0)}while((F|0)!=0);break}case 1016:{if(!j)break b;F=a+84|0;A=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[F>>2]|0)+E>>0]=n;do if(!(n<<24>>24)){ra=+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3];if(ra<0.0){p[(k[A>>2]|0)+(E<<3)>>3]=0.0;i[(k[F>>2]|0)+E>>0]=1;break}else{Y=+X(+ra);p[(k[A>>2]|0)+(E<<3)>>3]=Y;break}}while(0)}while((E|0)!=0);break}case 1019:{if(!j)break b;E=a+84|0;A=a+88|0;F=j;do{F=F+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+F>>0]|0;i[(k[E>>2]|0)+F>>0]=n;if(!(n<<24>>24)){Y=+ga(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(F<<3)>>3]));p[(k[A>>2]|0)+(F<<3)>>3]=Y}}while((F|0)!=0);break}case 1020:{if(!j)break b;F=a+84|0;A=a+88|0;E=j;do{E=E+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+E>>0]|0;i[(k[F>>2]|0)+E>>0]=n;if(!(n<<24>>24)){Y=+V(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(E<<3)>>3]));p[(k[A>>2]|0)+(E<<3)>>3]=Y}}while((E|0)!=0);break}case 1021:{if(!j)break b;E=a+84|0;A=a+88|0;F=j;do{F=F+-1|0;n=i[(k[(k[d>>2]|0)+84>>2]|0)+F>>0]|0;i[(k[E>>2]|0)+F>>0]=n;if(!(n<<24>>24)){Y=+V(+(+p[(k[(k[d>>2]|0)+88>>2]|0)+(F<<3)>>3]+.5));p[(k[A>>2]|0)+(F<<3)>>3]=Y}}while((F|0)!=0);break}case 1018:{if(!x)break b;F=g+1|0;A=a+84|0;E=f+32|0;n=f+320|0;L=a+88|0;G=e+4|0;s=d+4|0;D=d+4|0;o=x;B=y;m=j;while(1){K=o+-1|0;if(!B)$a=m;else{N=B;q=m;do{q=q+-1|0;u=k[G>>2]|0;if((u|0)<=1){if(u){u=k[D>>2]|0;p[n>>3]=+p[(k[u+88>>2]|0)+(K<<3)>>3];i[F>>0]=i[(k[u+84>>2]|0)+K>>0]|0}}else{u=k[s>>2]|0;p[n>>3]=+p[(k[u+88>>2]|0)+(q<<3)>>3];i[F>>0]=i[(k[u+84>>2]|0)+q>>0]|0}u=k[e>>2]|0;do if((u|0)<=1)if(!u){ab=i[g>>0]|0;break}else{I=k[d>>2]|0;p[E>>3]=+p[(k[I+88>>2]|0)+(K<<3)>>3];M=i[(k[I+84>>2]|0)+K>>0]|0;i[g>>0]=M;ab=M;break}else{M=k[d>>2]|0;p[E>>3]=+p[(k[M+88>>2]|0)+(q<<3)>>3];I=i[(k[M+84>>2]|0)+q>>0]|0;i[g>>0]=I;ab=I}while(0);N=N+-1|0;u=(i[F>>0]|ab)<<24>>24!=0;i[(k[A>>2]|0)+q>>0]=u&1;if(!u){Y=+da(+(+p[E>>3]),+(+p[n>>3]));p[(k[L>>2]|0)+(q<<3)>>3]=Y}}while((N|0)!=0);$a=m-B|0}if(!K)break b;o=K;B=k[w>>2]|0;m=$a}break}case 1041:{if(!x)break b;m=g+1|0;B=g+2|0;o=g+3|0;L=a+84|0;n=f+32|0;E=f+320|0;A=f+608|0;F=f+896|0;s=a+88|0;D=e+12|0;G=d+12|0;N=e+8|0;q=d+8|0;u=e+4|0;I=d+4|0;M=d+4|0;t=d+8|0;O=d+12|0;P=x;S=y;bb=j;while(1){cb=P+-1|0;if(!S)db=bb;else{eb=S;fb=bb;do{fb=fb+-1|0;gb=k[D>>2]|0;if((gb|0)<=1){if(gb){gb=k[O>>2]|0;p[F>>3]=+p[(k[gb+88>>2]|0)+(cb<<3)>>3];i[o>>0]=i[(k[gb+84>>2]|0)+cb>>0]|0}}else{gb=k[G>>2]|0;p[F>>3]=+p[(k[gb+88>>2]|0)+(fb<<3)>>3];i[o>>0]=i[(k[gb+84>>2]|0)+fb>>0]|0}gb=k[N>>2]|0;if((gb|0)<=1){if(gb){gb=k[t>>2]|0;p[A>>3]=+p[(k[gb+88>>2]|0)+(cb<<3)>>3];i[B>>0]=i[(k[gb+84>>2]|0)+cb>>0]|0}}else{gb=k[q>>2]|0;p[A>>3]=+p[(k[gb+88>>2]|0)+(fb<<3)>>3];i[B>>0]=i[(k[gb+84>>2]|0)+fb>>0]|0}gb=k[u>>2]|0;if((gb|0)<=1){if(gb){gb=k[M>>2]|0;p[E>>3]=+p[(k[gb+88>>2]|0)+(cb<<3)>>3];i[m>>0]=i[(k[gb+84>>2]|0)+cb>>0]|0}}else{gb=k[I>>2]|0;p[E>>3]=+p[(k[gb+88>>2]|0)+(fb<<3)>>3];i[m>>0]=i[(k[gb+84>>2]|0)+fb>>0]|0}gb=k[e>>2]|0;do if((gb|0)<=1)if(!gb){hb=i[g>>0]|0;break}else{ib=k[d>>2]|0;p[n>>3]=+p[(k[ib+88>>2]|0)+(cb<<3)>>3];jb=i[(k[ib+84>>2]|0)+cb>>0]|0;i[g>>0]=jb;hb=jb;break}else{jb=k[d>>2]|0;p[n>>3]=+p[(k[jb+88>>2]|0)+(fb<<3)>>3];ib=i[(k[jb+84>>2]|0)+fb>>0]|0;i[g>>0]=ib;hb=ib}while(0);eb=eb+-1|0;if(!((hb|i[m>>0]|i[B>>0])<<24>>24)){gb=(i[o>>0]|0)!=0;i[(k[L>>2]|0)+fb>>0]=gb&1;if(!gb){Y=+p[n>>3];ra=+p[E>>3];z=+p[A>>3];wa=+p[F>>3];kb=+p[12420];if(kb==0.0){p[12420]=.017453292519943295;lb=.017453292519943295}else lb=kb;kb=+_(+((z-Y)*lb*.5));Y=+_(+((wa-ra)*lb*.5));z=Y*Y+kb*(kb*(+Z(+(ra*lb))*+Z(+(wa*lb))));wa=z<0.0?0.0:z;z=wa>1.0?1.0:wa;wa=+da(+(+X(+z)),+(+X(+(1.0-z))))*2.0/lb;p[(k[s>>2]|0)+(fb<<3)>>3]=wa}}else i[(k[L>>2]|0)+fb>>0]=1}while((eb|0)!=0);db=bb-S|0}if(!cb)break b;P=cb;S=k[w>>2]|0;bb=db}break}case 1022:{bb=k[d>>2]|0;S=ha(k[bb+56>>2]|0,x)|0;switch(k[a+52>>2]|0){case 259:{if(!x)break b;P=a+84|0;L=a+88|0;s=x;F=S;A=0;while(1){E=s+-1|0;i[(k[P>>2]|0)+E>>0]=1;n=k[d>>2]|0;o=n;B=F;m=A;I=k[n+56>>2]|0;n=1;j:while(1){M=o+84|0;u=B;q=I;while(1){t=q+-1|0;if(!q){mb=u;nb=m;break j}N=u+-1|0;if(!(i[(k[M>>2]|0)+N>>0]|0)){ob=t;pb=N;break}else{u=N;q=t}}q=k[(k[o+88>>2]|0)+(pb<<2)>>2]|0;i[(k[P>>2]|0)+E>>0]=0;o=k[d>>2]|0;B=pb;m=(n|0)==0&(m|0)<(q|0)?m:q;I=ob;n=0}k[(k[L>>2]|0)+(E<<2)>>2]=nb;if(!E)break;else{s=E;F=mb;A=nb}}break}case 260:{if(!x)break b;A=a+84|0;F=a+88|0;s=x;L=S;wa=0.0;while(1){P=s+-1|0;i[(k[A>>2]|0)+P>>0]=1;n=k[d>>2]|0;I=n;m=L;z=wa;B=k[n+56>>2]|0;n=1;k:while(1){o=I+84|0;cb=m;q=B;while(1){u=q+-1|0;if(!q){qb=cb;rb=z;break k}M=cb+-1|0;if(!(i[(k[o>>2]|0)+M>>0]|0)){sb=u;tb=M;break}else{cb=M;q=u}}ra=+p[(k[I+88>>2]|0)+(tb<<3)>>3];i[(k[A>>2]|0)+P>>0]=0;I=k[d>>2]|0;m=tb;z=(n|0)==0&z<ra?z:ra;B=sb;n=0}p[(k[F>>2]|0)+(P<<3)>>3]=rb;if(!P)break;else{s=P;L=qb;wa=rb}}break}case 262:{if(!x)break b;L=a+88|0;s=x;F=bb;while(1){A=s+-1|0;S=49;n=k[(k[F+88>>2]|0)+(A<<2)>>2]|0;l:while(1){switch(i[n>>0]|0){case 0:{ub=S;break l;break}case 48:{vb=48;break}default:vb=S}S=vb;n=n+1|0}i[k[(k[L>>2]|0)+(A<<2)>>2]>>0]=ub;i[(k[(k[L>>2]|0)+(A<<2)>>2]|0)+1>>0]=0;if(!A)break b;s=A;F=k[d>>2]|0}break}default:break b}break}case 1023:{switch(k[a+52>>2]|0){case 259:{if(!x)break b;F=g+1|0;s=a+84|0;L=f+32|0;bb=f+320|0;n=a+88|0;S=e+4|0;P=d+4|0;B=d+4|0;m=x;I=y;E=j;while(1){q=m+-1|0;if(!I)wb=E;else{cb=I;o=E;do{o=o+-1|0;u=k[S>>2]|0;if((u|0)<=1){if(u){u=k[B>>2]|0;k[bb>>2]=k[(k[u+88>>2]|0)+(q<<2)>>2];i[F>>0]=i[(k[u+84>>2]|0)+q>>0]|0}}else{u=k[P>>2]|0;k[bb>>2]=k[(k[u+88>>2]|0)+(o<<2)>>2];i[F>>0]=i[(k[u+84>>2]|0)+o>>0]|0}u=k[e>>2]|0;do if((u|0)<=1)if(!u){xb=i[g>>0]|0;break}else{M=k[d>>2]|0;k[L>>2]=k[(k[M+88>>2]|0)+(q<<2)>>2];t=i[(k[M+84>>2]|0)+q>>0]|0;i[g>>0]=t;xb=t;break}else{t=k[d>>2]|0;k[L>>2]=k[(k[t+88>>2]|0)+(o<<2)>>2];M=i[(k[t+84>>2]|0)+o>>0]|0;i[g>>0]=M;xb=M}while(0);cb=cb+-1|0;u=(i[F>>0]|0)==0;M=(k[s>>2]|0)+o|0;do if(!(xb<<24>>24)){i[M>>0]=0;t=k[L>>2]|0;if(u){N=k[bb>>2]|0;k[(k[n>>2]|0)+(o<<2)>>2]=(t|0)<(N|0)?t:N;break}else{k[(k[n>>2]|0)+(o<<2)>>2]=t;break}}else if(u){i[M>>0]=0;k[(k[n>>2]|0)+(o<<2)>>2]=k[bb>>2];break}else{i[M>>0]=1;k[(k[n>>2]|0)+(o<<2)>>2]=0;break}while(0)}while((cb|0)!=0);wb=E-I|0}if(!q)break b;m=q;I=k[w>>2]|0;E=wb}break}case 260:{if(!x)break b;E=g+1|0;I=a+84|0;m=f+32|0;n=f+320|0;bb=a+88|0;L=e+4|0;s=d+4|0;F=d+4|0;P=x;B=y;S=j;while(1){cb=P+-1|0;if(!B)yb=S;else{o=B;A=S;do{A=A+-1|0;M=k[L>>2]|0;if((M|0)<=1){if(M){M=k[F>>2]|0;p[n>>3]=+p[(k[M+88>>2]|0)+(cb<<3)>>3];i[E>>0]=i[(k[M+84>>2]|0)+cb>>0]|0}}else{M=k[s>>2]|0;p[n>>3]=+p[(k[M+88>>2]|0)+(A<<3)>>3];i[E>>0]=i[(k[M+84>>2]|0)+A>>0]|0}M=k[e>>2]|0;do if((M|0)<=1)if(!M){zb=i[g>>0]|0;break}else{u=k[d>>2]|0;p[m>>3]=+p[(k[u+88>>2]|0)+(cb<<3)>>3];t=i[(k[u+84>>2]|0)+cb>>0]|0;i[g>>0]=t;zb=t;break}else{t=k[d>>2]|0;p[m>>3]=+p[(k[t+88>>2]|0)+(A<<3)>>3];u=i[(k[t+84>>2]|0)+A>>0]|0;i[g>>0]=u;zb=u}while(0);o=o+-1|0;M=(i[E>>0]|0)==0;u=(k[I>>2]|0)+A|0;do if(!(zb<<24>>24)){i[u>>0]=0;wa=+p[m>>3];if(M){z=+p[n>>3];p[(k[bb>>2]|0)+(A<<3)>>3]=wa<z?wa:z;break}else{p[(k[bb>>2]|0)+(A<<3)>>3]=wa;break}}else if(M){i[u>>0]=0;p[(k[bb>>2]|0)+(A<<3)>>3]=+p[n>>3];break}else{i[u>>0]=1;p[(k[bb>>2]|0)+(A<<3)>>3]=0.0;break}while(0)}while((o|0)!=0);yb=S-B|0}if(!cb)break b;P=cb;B=k[w>>2]|0;S=yb}break}default:break b}break}case 1024:{S=k[d>>2]|0;B=ha(k[S+56>>2]|0,x)|0;switch(k[a+52>>2]|0){case 259:{if(!x)break b;P=a+84|0;bb=a+88|0;n=x;m=B;I=0;while(1){E=n+-1|0;i[(k[P>>2]|0)+E>>0]=1;s=k[d>>2]|0;F=s;L=m;o=I;A=k[s+56>>2]|0;s=1;m:while(1){q=F+84|0;u=L;M=A;while(1){t=M+-1|0;if(!M){Ab=u;Bb=o;break m}N=u+-1|0;if(!(i[(k[q>>2]|0)+N>>0]|0)){Cb=t;Db=N;break}else{u=N;M=t}}M=k[(k[F+88>>2]|0)+(Db<<2)>>2]|0;i[(k[P>>2]|0)+E>>0]=0;F=k[d>>2]|0;L=Db;o=(s|0)==0&(o|0)>(M|0)?o:M;A=Cb;s=0}k[(k[bb>>2]|0)+(E<<2)>>2]=Bb;if(!E)break;else{n=E;m=Ab;I=Bb}}break}case 260:{if(!x)break b;I=a+84|0;m=a+88|0;n=x;bb=B;wa=0.0;while(1){P=n+-1|0;i[(k[I>>2]|0)+P>>0]=1;s=k[d>>2]|0;A=s;o=bb;z=wa;L=k[s+56>>2]|0;s=1;n:while(1){F=A+84|0;cb=o;M=L;while(1){u=M+-1|0;if(!M){Eb=cb;Fb=z;break n}q=cb+-1|0;if(!(i[(k[F>>2]|0)+q>>0]|0)){Gb=u;Hb=q;break}else{cb=q;M=u}}ra=+p[(k[A+88>>2]|0)+(Hb<<3)>>3];i[(k[I>>2]|0)+P>>0]=0;A=k[d>>2]|0;o=Hb;z=(s|0)==0&z>ra?z:ra;L=Gb;s=0}p[(k[m>>2]|0)+(P<<3)>>3]=Fb;if(!P)break;else{n=P;bb=Eb;wa=Fb}}break}case 262:{if(!x)break b;bb=a+88|0;n=x;m=S;while(1){I=n+-1|0;B=48;s=k[(k[m+88>>2]|0)+(I<<2)>>2]|0;o:while(1){switch(i[s>>0]|0){case 0:{Ib=B;break o;break}case 49:{Jb=49;break}default:Jb=B}B=Jb;s=s+1|0}i[k[(k[bb>>2]|0)+(I<<2)>>2]>>0]=Ib;i[(k[(k[bb>>2]|0)+(I<<2)>>2]|0)+1>>0]=0;if(!I)break b;n=I;m=k[d>>2]|0}break}default:break b}break}case 1025:{switch(k[a+52>>2]|0){case 259:{if(!x)break b;m=g+1|0;n=a+84|0;bb=f+32|0;S=f+320|0;s=a+88|0;B=e+4|0;P=d+4|0;L=d+4|0;o=x;A=y;E=j;while(1){M=o+-1|0;if(!A)Kb=E;else{cb=A;F=E;do{F=F+-1|0;u=k[B>>2]|0;if((u|0)<=1){if(u){u=k[L>>2]|0;k[S>>2]=k[(k[u+88>>2]|0)+(M<<2)>>2];i[m>>0]=i[(k[u+84>>2]|0)+M>>0]|0}}else{u=k[P>>2]|0;k[S>>2]=k[(k[u+88>>2]|0)+(F<<2)>>2];i[m>>0]=i[(k[u+84>>2]|0)+F>>0]|0}u=k[e>>2]|0;do if((u|0)<=1)if(!u){Lb=i[g>>0]|0;break}else{q=k[d>>2]|0;k[bb>>2]=k[(k[q+88>>2]|0)+(M<<2)>>2];t=i[(k[q+84>>2]|0)+M>>0]|0;i[g>>0]=t;Lb=t;break}else{t=k[d>>2]|0;k[bb>>2]=k[(k[t+88>>2]|0)+(F<<2)>>2];q=i[(k[t+84>>2]|0)+F>>0]|0;i[g>>0]=q;Lb=q}while(0);cb=cb+-1|0;u=(i[m>>0]|0)==0;q=(k[n>>2]|0)+F|0;do if(!(Lb<<24>>24)){i[q>>0]=0;t=k[bb>>2]|0;if(u){N=k[S>>2]|0;k[(k[s>>2]|0)+(F<<2)>>2]=(t|0)>(N|0)?t:N;break}else{k[(k[s>>2]|0)+(F<<2)>>2]=t;break}}else if(u){i[q>>0]=0;k[(k[s>>2]|0)+(F<<2)>>2]=k[S>>2];break}else{i[q>>0]=1;k[(k[s>>2]|0)+(F<<2)>>2]=0;break}while(0)}while((cb|0)!=0);Kb=E-A|0}if(!M)break b;o=M;A=k[w>>2]|0;E=Kb}break}case 260:{if(!x)break b;E=g+1|0;A=a+84|0;o=f+32|0;s=f+320|0;S=a+88|0;bb=e+4|0;n=d+4|0;m=d+4|0;P=x;L=y;B=j;while(1){cb=P+-1|0;if(!L)Mb=B;else{F=L;I=B;do{I=I+-1|0;q=k[bb>>2]|0;if((q|0)<=1){if(q){q=k[m>>2]|0;p[s>>3]=+p[(k[q+88>>2]|0)+(cb<<3)>>3];i[E>>0]=i[(k[q+84>>2]|0)+cb>>0]|0}}else{q=k[n>>2]|0;p[s>>3]=+p[(k[q+88>>2]|0)+(I<<3)>>3];i[E>>0]=i[(k[q+84>>2]|0)+I>>0]|0}q=k[e>>2]|0;do if((q|0)<=1)if(!q){Nb=i[g>>0]|0;break}else{u=k[d>>2]|0;p[o>>3]=+p[(k[u+88>>2]|0)+(cb<<3)>>3];t=i[(k[u+84>>2]|0)+cb>>0]|0;i[g>>0]=t;Nb=t;break}else{t=k[d>>2]|0;p[o>>3]=+p[(k[t+88>>2]|0)+(I<<3)>>3];u=i[(k[t+84>>2]|0)+I>>0]|0;i[g>>0]=u;Nb=u}while(0);F=F+-1|0;q=(i[E>>0]|0)==0;u=(k[A>>2]|0)+I|0;do if(!(Nb<<24>>24)){i[u>>0]=0;wa=+p[o>>3];if(q){z=+p[s>>3];p[(k[S>>2]|0)+(I<<3)>>3]=wa>z?wa:z;break}else{p[(k[S>>2]|0)+(I<<3)>>3]=wa;break}}else if(q){i[u>>0]=0;p[(k[S>>2]|0)+(I<<3)>>3]=+p[s>>3];break}else{i[u>>0]=1;p[(k[S>>2]|0)+(I<<3)>>3]=0.0;break}while(0)}while((F|0)!=0);Mb=B-L|0}if(!cb)break b;P=cb;L=k[w>>2]|0;B=Mb}break}default:break b}break}case 1026:{if(!x)break b;B=g+1|0;L=g+2|0;P=a+84|0;S=f+32|0;s=f+320|0;o=f+608|0;A=a+88|0;E=e+8|0;n=d+8|0;m=e+4|0;bb=d+4|0;F=d+4|0;I=d+8|0;M=x;u=y;q=j;while(1){t=M+-1|0;if(!u)Ob=q;else{N=u;G=q;do{G=G+-1|0;O=k[E>>2]|0;if((O|0)<=1){if(O){O=k[I>>2]|0;p[o>>3]=+p[(k[O+88>>2]|0)+(t<<3)>>3];i[L>>0]=i[(k[O+84>>2]|0)+t>>0]|0}}else{O=k[n>>2]|0;p[o>>3]=+p[(k[O+88>>2]|0)+(G<<3)>>3];i[L>>0]=i[(k[O+84>>2]|0)+G>>0]|0}O=k[m>>2]|0;if((O|0)<=1){if(O){O=k[F>>2]|0;p[s>>3]=+p[(k[O+88>>2]|0)+(t<<3)>>3];i[B>>0]=i[(k[O+84>>2]|0)+t>>0]|0}}else{O=k[bb>>2]|0;p[s>>3]=+p[(k[O+88>>2]|0)+(G<<3)>>3];i[B>>0]=i[(k[O+84>>2]|0)+G>>0]|0}O=k[e>>2]|0;do if((O|0)<=1)if(!O){Pb=i[g>>0]|0;break}else{D=k[d>>2]|0;p[S>>3]=+p[(k[D+88>>2]|0)+(t<<3)>>3];eb=i[(k[D+84>>2]|0)+t>>0]|0;i[g>>0]=eb;Pb=eb;break}else{eb=k[d>>2]|0;p[S>>3]=+p[(k[eb+88>>2]|0)+(G<<3)>>3];D=i[(k[eb+84>>2]|0)+G>>0]|0;i[g>>0]=D;Pb=D}while(0);N=N+-1|0;if(!((Pb|i[B>>0])<<24>>24)){O=(i[L>>0]|0)!=0;i[(k[P>>2]|0)+G>>0]=O&1;if(!O){O=+W(+(+p[S>>3]-+p[s>>3]))<+p[o>>3]&1;i[(k[A>>2]|0)+G>>0]=O}}else i[(k[P>>2]|0)+G>>0]=1}while((N|0)!=0);Ob=q-u|0}if(!t)break b;M=t;u=k[w>>2]|0;q=Ob}break}case 1027:{if(!x)break b;q=g+1|0;u=g+2|0;M=g+3|0;P=g+4|0;A=a+84|0;o=f+32|0;s=f+320|0;S=f+608|0;L=f+896|0;B=f+1184|0;bb=a+88|0;F=x;m=y;n=j;while(1){I=F+-1|0;if(!m)Qb=n;else{E=m;N=n;do{N=N+-1|0;G=4;while(1){cb=k[e+(G<<2)>>2]|0;if((cb|0)<=1){if(cb){cb=k[d+(G<<2)>>2]|0;p[f+(G*288|0)+32>>3]=+p[(k[cb+88>>2]|0)+(I<<3)>>3];i[g+G>>0]=i[(k[cb+84>>2]|0)+I>>0]|0}}else{cb=k[d+(G<<2)>>2]|0;p[f+(G*288|0)+32>>3]=+p[(k[cb+88>>2]|0)+(N<<3)>>3];i[g+G>>0]=i[(k[cb+84>>2]|0)+N>>0]|0}if(!G)break;else G=G+-1|0}E=E+-1|0;if(!((i[g>>0]|i[q>>0]|i[u>>0]|i[M>>0])<<24>>24)){G=(i[P>>0]|0)!=0;i[(k[A>>2]|0)+N>>0]=G&1;if(!G){wa=+p[S>>3];z=+p[L>>3]-+p[o>>3];ra=+p[B>>3]-+p[s>>3];i[(k[bb>>2]|0)+N>>0]=z*z+ra*ra<=wa*wa&1}}else i[(k[A>>2]|0)+N>>0]=1}while((E|0)!=0);Qb=n-m|0}if(!I)break b;F=I;m=k[w>>2]|0;n=Qb}break}case 1028:{if(!x)break b;n=g+1|0;m=g+2|0;F=g+3|0;A=g+4|0;bb=g+5|0;s=g+6|0;B=a+84|0;o=f+32|0;L=f+320|0;S=f+608|0;P=f+896|0;M=f+1184|0;u=f+1472|0;q=f+1760|0;E=a+88|0;N=x;t=y;G=j;while(1){cb=N+-1|0;if(!t)Rb=G;else{O=t;D=G;do{D=D+-1|0;eb=6;while(1){fb=k[e+(eb<<2)>>2]|0;if((fb|0)<=1){if(fb){fb=k[d+(eb<<2)>>2]|0;p[f+(eb*288|0)+32>>3]=+p[(k[fb+88>>2]|0)+(cb<<3)>>3];i[g+eb>>0]=i[(k[fb+84>>2]|0)+cb>>0]|0}}else{fb=k[d+(eb<<2)>>2]|0;p[f+(eb*288|0)+32>>3]=+p[(k[fb+88>>2]|0)+(D<<3)>>3];i[g+eb>>0]=i[(k[fb+84>>2]|0)+D>>0]|0}if(!eb)break;else eb=eb+-1|0}O=O+-1|0;if(!((i[g>>0]|i[n>>0]|i[m>>0]|i[F>>0]|i[A>>0]|i[bb>>0])<<24>>24)){eb=(i[s>>0]|0)!=0;i[(k[B>>2]|0)+D>>0]=eb&1;if(!eb){wa=+p[S>>3];ra=+p[P>>3];z=+p[M>>3]/180.0*3.141592653589793;kb=+p[u>>3]-+p[o>>3];Y=+p[q>>3]-+p[L>>3];Sb=+Z(+z);Tb=+_(+z);z=kb*Sb+Y*Tb;Ub=Y*Sb-kb*Tb;i[(k[E>>2]|0)+D>>0]=(!(Ub<=ra*.5)|(!(Ub>=ra*-.5)|(!(z>=wa*-.5)|!(z<=wa*.5))))&1^1}}else i[(k[B>>2]|0)+D>>0]=1}while((O|0)!=0);Rb=G-t|0}if(!cb)break b;N=cb;t=k[w>>2]|0;G=Rb}break}case 1029:{if(!x)break b;G=g+1|0;t=g+2|0;N=g+3|0;B=g+4|0;E=g+5|0;L=g+6|0;q=a+84|0;o=f+32|0;u=f+320|0;M=f+608|0;P=f+896|0;S=f+1184|0;s=f+1472|0;bb=f+1760|0;A=a+88|0;F=x;m=y;n=j;while(1){O=F+-1|0;if(!m)Vb=n;else{D=m;I=n;do{I=I+-1|0;eb=6;while(1){fb=k[e+(eb<<2)>>2]|0;if((fb|0)<=1){if(fb){fb=k[d+(eb<<2)>>2]|0;p[f+(eb*288|0)+32>>3]=+p[(k[fb+88>>2]|0)+(O<<3)>>3];i[g+eb>>0]=i[(k[fb+84>>2]|0)+O>>0]|0}}else{fb=k[d+(eb<<2)>>2]|0;p[f+(eb*288|0)+32>>3]=+p[(k[fb+88>>2]|0)+(I<<3)>>3];i[g+eb>>0]=i[(k[fb+84>>2]|0)+I>>0]|0}if(!eb)break;else eb=eb+-1|0}D=D+-1|0;if(!((i[g>>0]|i[G>>0]|i[t>>0]|i[N>>0]|i[B>>0]|i[E>>0])<<24>>24)){eb=(i[L>>0]|0)!=0;i[(k[q>>2]|0)+I>>0]=eb&1;if(!eb){wa=+p[S>>3]/180.0*3.141592653589793;z=+p[s>>3]-+p[o>>3];ra=+p[bb>>3]-+p[u>>3];Ub=+Z(+wa);Tb=+_(+wa);wa=(z*Ub+ra*Tb)/+p[M>>3];kb=(ra*Ub-z*Tb)/+p[P>>3];i[(k[A>>2]|0)+I>>0]=wa*wa+kb*kb<=1.0&1}}else i[(k[q>>2]|0)+I>>0]=1}while((D|0)!=0);Vb=n-m|0}if(!O)break b;F=O;m=k[w>>2]|0;n=Vb}break}case 1034:{switch(k[a+52>>2]|0){case 258:{if(!x)break b;n=e+8|0;m=d+8|0;F=f+608|0;q=g+2|0;A=a+84|0;P=f+320|0;M=a+88|0;u=g+1|0;bb=f+32|0;o=e+4|0;s=d+4|0;S=d+4|0;L=x;E=y;B=j;while(1){N=L+-1|0;if(!E)Wb=B;else{t=E;G=B;do{t=t+-1|0;G=G+-1|0;D=k[n>>2]|0;if((D|0)<=1){if(D){D=k[m>>2]|0;i[F>>0]=i[(k[D+88>>2]|0)+N>>0]|0;i[q>>0]=i[(k[D+84>>2]|0)+N>>0]|0}}else{D=k[m>>2]|0;i[F>>0]=i[(k[D+88>>2]|0)+G>>0]|0;i[q>>0]=i[(k[D+84>>2]|0)+G>>0]|0}D=k[o>>2]|0;if((D|0)<=1){if(D){D=k[S>>2]|0;i[P>>0]=i[(k[D+88>>2]|0)+N>>0]|0;i[u>>0]=i[(k[D+84>>2]|0)+N>>0]|0}}else{D=k[s>>2]|0;i[P>>0]=i[(k[D+88>>2]|0)+G>>0]|0;i[u>>0]=i[(k[D+84>>2]|0)+G>>0]|0}D=k[e>>2]|0;if((D|0)<=1){if(D){D=k[d>>2]|0;i[bb>>0]=i[(k[D+88>>2]|0)+N>>0]|0;i[g>>0]=i[(k[D+84>>2]|0)+N>>0]|0}}else{D=k[d>>2]|0;i[bb>>0]=i[(k[D+88>>2]|0)+G>>0]|0;i[g>>0]=i[(k[D+84>>2]|0)+G>>0]|0}D=i[q>>0]|0;i[(k[A>>2]|0)+G>>0]=D;do if(!(D<<24>>24))if(!(i[F>>0]|0)){i[(k[M>>2]|0)+G>>0]=i[P>>0]|0;i[(k[A>>2]|0)+G>>0]=i[u>>0]|0;break}else{i[(k[M>>2]|0)+G>>0]=i[bb>>0]|0;i[(k[A>>2]|0)+G>>0]=i[g>>0]|0;break}while(0)}while((t|0)!=0);Wb=B-E|0}if(!N)break b;L=N;E=k[w>>2]|0;B=Wb}break}case 259:{if(!x)break b;B=e+8|0;E=d+8|0;L=f+608|0;A=g+2|0;bb=a+84|0;M=f+320|0;u=a+88|0;P=g+1|0;F=f+32|0;q=e+4|0;s=d+4|0;S=d+4|0;o=x;m=y;n=j;while(1){t=o+-1|0;if(!m)Xb=n;else{G=m;O=n;do{G=G+-1|0;O=O+-1|0;D=k[B>>2]|0;if((D|0)<=1){if(D){D=k[E>>2]|0;i[L>>0]=i[(k[D+88>>2]|0)+t>>0]|0;i[A>>0]=i[(k[D+84>>2]|0)+t>>0]|0}}else{D=k[E>>2]|0;i[L>>0]=i[(k[D+88>>2]|0)+O>>0]|0;i[A>>0]=i[(k[D+84>>2]|0)+O>>0]|0}D=k[q>>2]|0;if((D|0)<=1){if(D){D=k[S>>2]|0;k[M>>2]=k[(k[D+88>>2]|0)+(t<<2)>>2];i[P>>0]=i[(k[D+84>>2]|0)+t>>0]|0}}else{D=k[s>>2]|0;k[M>>2]=k[(k[D+88>>2]|0)+(O<<2)>>2];i[P>>0]=i[(k[D+84>>2]|0)+O>>0]|0}D=k[e>>2]|0;if((D|0)<=1){if(D){D=k[d>>2]|0;k[F>>2]=k[(k[D+88>>2]|0)+(t<<2)>>2];i[g>>0]=i[(k[D+84>>2]|0)+t>>0]|0}}else{D=k[d>>2]|0;k[F>>2]=k[(k[D+88>>2]|0)+(O<<2)>>2];i[g>>0]=i[(k[D+84>>2]|0)+O>>0]|0}D=i[A>>0]|0;i[(k[bb>>2]|0)+O>>0]=D;do if(!(D<<24>>24))if(!(i[L>>0]|0)){k[(k[u>>2]|0)+(O<<2)>>2]=k[M>>2];i[(k[bb>>2]|0)+O>>0]=i[P>>0]|0;break}else{k[(k[u>>2]|0)+(O<<2)>>2]=k[F>>2];i[(k[bb>>2]|0)+O>>0]=i[g>>0]|0;break}while(0)}while((G|0)!=0);Xb=n-m|0}if(!t)break b;o=t;m=k[w>>2]|0;n=Xb}break}case 260:{if(!x)break b;n=e+8|0;m=d+8|0;o=f+608|0;bb=g+2|0;F=a+84|0;u=f+320|0;P=a+88|0;M=g+1|0;L=f+32|0;A=e+4|0;s=d+4|0;S=d+4|0;q=x;E=y;B=j;while(1){G=q+-1|0;if(!E)Yb=B;else{O=E;N=B;do{O=O+-1|0;N=N+-1|0;D=k[n>>2]|0;if((D|0)<=1){if(D){D=k[m>>2]|0;i[o>>0]=i[(k[D+88>>2]|0)+G>>0]|0;i[bb>>0]=i[(k[D+84>>2]|0)+G>>0]|0}}else{D=k[m>>2]|0;i[o>>0]=i[(k[D+88>>2]|0)+N>>0]|0;i[bb>>0]=i[(k[D+84>>2]|0)+N>>0]|0}D=k[A>>2]|0;if((D|0)<=1){if(D){D=k[S>>2]|0;p[u>>3]=+p[(k[D+88>>2]|0)+(G<<3)>>3];i[M>>0]=i[(k[D+84>>2]|0)+G>>0]|0}}else{D=k[s>>2]|0;p[u>>3]=+p[(k[D+88>>2]|0)+(N<<3)>>3];i[M>>0]=i[(k[D+84>>2]|0)+N>>0]|0}D=k[e>>2]|0;if((D|0)<=1){if(D){D=k[d>>2]|0;p[L>>3]=+p[(k[D+88>>2]|0)+(G<<3)>>3];i[g>>0]=i[(k[D+84>>2]|0)+G>>0]|0}}else{D=k[d>>2]|0;p[L>>3]=+p[(k[D+88>>2]|0)+(N<<3)>>3];i[g>>0]=i[(k[D+84>>2]|0)+N>>0]|0}D=i[bb>>0]|0;i[(k[F>>2]|0)+N>>0]=D;do if(!(D<<24>>24))if(!(i[o>>0]|0)){p[(k[P>>2]|0)+(N<<3)>>3]=+p[u>>3];i[(k[F>>2]|0)+N>>0]=i[M>>0]|0;break}else{p[(k[P>>2]|0)+(N<<3)>>3]=+p[L>>3];i[(k[F>>2]|0)+N>>0]=i[g>>0]|0;break}while(0)}while((O|0)!=0);Yb=B-E|0}if(!G)break b;q=G;E=k[w>>2]|0;B=Yb}break}case 261:{if(!x)break b;B=e+8|0;E=d+8|0;q=f+608|0;F=g+2|0;L=a+84|0;P=a+88|0;M=f+320|0;u=g+1|0;o=f+32|0;bb=e+4|0;s=d+4|0;S=x;do{S=S+-1|0;if(k[B>>2]|0){A=k[E>>2]|0;i[q>>0]=i[(k[A+88>>2]|0)+S>>0]|0;i[F>>0]=i[(k[A+84>>2]|0)+S>>0]|0}if(k[bb>>2]|0){ux(M,k[(k[(k[s>>2]|0)+88>>2]|0)+(S<<2)>>2]|0)|0;i[u>>0]=i[(k[(k[s>>2]|0)+84>>2]|0)+S>>0]|0}if(k[e>>2]|0){ux(o,k[(k[(k[d>>2]|0)+88>>2]|0)+(S<<2)>>2]|0)|0;i[g>>0]=i[(k[(k[d>>2]|0)+84>>2]|0)+S>>0]|0}A=i[F>>0]|0;i[(k[L>>2]|0)+S>>0]=A;do if(!(A<<24>>24)){m=k[(k[P>>2]|0)+(S<<2)>>2]|0;if(!(i[q>>0]|0)){ux(m,M)|0;i[(k[L>>2]|0)+S>>0]=i[u>>0]|0;break}else{ux(m,o)|0;i[(k[L>>2]|0)+S>>0]=i[g>>0]|0;break}}else i[k[(k[P>>2]|0)+(S<<2)>>2]>>0]=0;while(0)}while((S|0)!=0);break}default:break b}break}case 1044:{S=k[d>>2]|0;P=(k[S>>2]|0)==-1e3;L=d+4|0;o=k[L>>2]|0;u=(k[o>>2]|0)==-1e3;M=d+8|0;q=(k[k[M>>2]>>2]|0)==-1e3;if(!x)break b;F=a+88|0;s=a+84|0;bb=x;E=o;o=k[S+56>>2]|0;while(1){S=bb+-1|0;B=E+88|0;if(u){Zb=B;_b=0}else{Zb=(k[B>>2]|0)+(S<<2)|0;_b=(i[(k[E+84>>2]|0)+S>>0]|0)!=0&1}B=k[Zb>>2]|0;A=k[d>>2]|0;G=A+88|0;if(P)if(!o){$b=xx(G)|0;ac=G;bc=_b}else{$b=o;ac=G;bc=_b}else{$b=o;ac=k[(k[G>>2]|0)+(S<<2)>>2]|0;bc=(i[(k[A+84>>2]|0)+S>>0]|0)==0?_b:1}if(q){cc=y;dc=bc}else{A=k[M>>2]|0;cc=k[(k[A+88>>2]|0)+(S<<2)>>2]|0;dc=(i[(k[A+84>>2]|0)+S>>0]|0)==0?bc:1}i[k[(k[F>>2]|0)+(S<<2)>>2]>>0]=0;A=(B|0)==0;G=A?1:dc;if(!G){m=k[(k[F>>2]|0)+(S<<2)>>2]|0;if(!$b)ec=xx(ac)|0;else ec=$b;if((B|0)<0)break;do if(!(A|(ec|0)<(B|0)))if((cc+B|0)>(ec|0)){n=ec-B+1|0;py(m|0,ac+(B+-1)|0,n|0)|0;sy(m+n|0,0,cc-n|0)|0;break}else{py(m|0,ac+(B+-1)|0,cc|0)|0;break}else sy(m|0,0,cc|0)|0;while(0);i[m+cc>>0]=0}i[(k[s>>2]|0)+S>>0]=G;if(!S)break b;bb=S;E=k[L>>2]|0;o=$b}if(!(k[30835]|0))k[30835]=431;Bx(c,790404,80)|0;i[c+79>>0]=0;tk(c);break b;break}case 1045:{o=k[d>>2]|0;L=(k[o>>2]|0)==-1e3;E=d+4|0;bb=(k[k[E>>2]>>2]|0)==-1e3;if(!x)break b;s=a+88|0;F=a+84|0;M=x;q=o;while(1){o=M+-1|0;P=q+88|0;if(L){fc=P;gc=0}else{fc=k[(k[P>>2]|0)+(o<<2)>>2]|0;gc=(i[(k[q+84>>2]|0)+o>>0]|0)!=0&1}P=k[E>>2]|0;u=P+88|0;if(!bb)if(!(i[(k[P+84>>2]|0)+o>>0]|0)){hc=k[(k[u>>2]|0)+(o<<2)>>2]|0;ic=780}else{k[(k[s>>2]|0)+(o<<2)>>2]=0;jc=1}else{hc=u;ic=780}do if((ic|0)==780){ic=0;u=(k[s>>2]|0)+(o<<2)|0;k[u>>2]=0;if(!gc){P=Ex(fc,hc)|0;if(!P){k[u>>2]=0;jc=1;break}else{k[u>>2]=1-fc+P;jc=0;break}}else jc=gc}while(0);i[(k[F>>2]|0)+o>>0]=jc;if(!o)break b;M=o;q=k[d>>2]|0}break}default:break b}while(0)}else{p:do switch(v|0){case 1002:{switch(k[(k[d>>2]|0)+52>>2]|0){case 258:{k[a+88>>2]=(i[f+32>>0]|0)!=0&1;break p;break}case 259:{k[a+88>>2]=k[f+32>>2];break p;break}case 260:{p[a+88>>3]=+p[f+32>>3];break p;break}case 262:{ux(a+88|0,f+32|0)|0;break p;break}default:break p}break}case 1038:{switch(k[(k[d>>2]|0)+52>>2]|0){case 259:{p[a+88>>3]=+(k[f+32>>2]|0);break p;break}case 260:{p[a+88>>3]=+p[f+32>>3];break p;break}default:break p}break}case 1039:{p[a+88>>3]=0.0;break}case 1037:{x=k[(k[d>>2]|0)+52>>2]|0;if((x|0)==258){k[a+88>>2]=(i[f+32>>0]|0)!=0&1;break p}y=f+32|0;if((x|0)==259){k[a+88>>2]=k[y>>2];break p}else{p[a+88>>3]=+p[y>>3];break p}break}case 1043:{y=f+32|0;if((k[(k[d>>2]|0)+52>>2]|0)==260){x=hv(+p[y>>3])|0;k[a+88>>2]=x;break p}else{x=hv(+(k[y>>2]|0))|0;k[a+88>>2]=x;break p}break}case 1017:{x=f+32|0;if((k[(k[d>>2]|0)+52>>2]|0)==260){kb=+p[x>>3];p[a+88>>3]=kb>0.0?kb:-kb;break p}else{y=k[x>>2]|0;k[a+88>>2]=(y|0)>0?y:0-y|0;break p}break}case 1040:{k[a+88>>2]=1;break}case 1030:{i[a+88>>0]=0;break}case 1031:{switch(k[a+52>>2]|0){case 258:{i[a+88>>0]=i[f+32>>0]|0;break p;break}case 259:{k[a+88>>2]=k[f+32>>2];break p;break}case 260:{p[a+88>>3]=+p[f+32>>3];break p;break}case 261:{ux(a+88|0,f+32|0)|0;break p;break}default:break p}break}case 1004:{kb=+_(+(+p[f+32>>3]));p[a+88>>3]=kb;break}case 1005:{kb=+Z(+(+p[f+32>>3]));p[a+88>>3]=kb;break}case 1006:{kb=+$(+(+p[f+32>>3]));p[a+88>>3]=kb;break}case 1007:{kb=+p[f+32>>3];if(!(kb<-1.0|kb>1.0)){wa=+ba(+kb);p[a+88>>3]=wa;break p}if(!(k[30835]|0))k[30835]=431;Bx(c,790250,80)|0;i[c+79>>0]=0;tk(c);break}case 1008:{wa=+p[f+32>>3];if(!(wa<-1.0|wa>1.0)){kb=+aa(+wa);p[a+88>>3]=kb;break p}if(!(k[30835]|0))k[30835]=431;Bx(c,790282,80)|0;i[c+79>>0]=0;tk(c);break}case 1009:{kb=+ca(+(+p[f+32>>3]));p[a+88>>3]=kb;break}case 1010:{kb=+Nv(+p[f+32>>3]);p[a+88>>3]=kb;break}case 1011:{kb=+Ev(+p[f+32>>3]);p[a+88>>3]=kb;break}case 1012:{kb=+Ov(+p[f+32>>3]);p[a+88>>3]=kb;break}case 1013:{kb=+ea(+(+p[f+32>>3]));p[a+88>>3]=kb;break}case 1014:{kb=+p[f+32>>3];if(!(kb<=0.0)){wa=+fa(+kb);p[a+88>>3]=wa;break p}if(!(k[30835]|0))k[30835]=431;Bx(c,790314,80)|0;i[c+79>>0]=0;tk(c);break}case 1015:{wa=+p[f+32>>3];if(!(wa<=0.0)){kb=+Kv(wa);p[a+88>>3]=kb;break p}if(!(k[30835]|0))k[30835]=431;Bx(c,790343,80)|0;i[c+79>>0]=0;tk(c);break}case 1016:{kb=+p[f+32>>3];if(!(kb<0.0)){wa=+X(+kb);p[a+88>>3]=wa;break p}if(!(k[30835]|0))k[30835]=431;Bx(c,790374,80)|0;i[c+79>>0]=0;tk(c);break}case 1019:{wa=+ga(+(+p[f+32>>3]));p[a+88>>3]=wa;break}case 1020:{wa=+V(+(+p[f+32>>3]));p[a+88>>3]=wa;break}case 1021:{wa=+V(+(+p[f+32>>3]+.5));p[a+88>>3]=wa;break}case 1018:{wa=+da(+(+p[f+32>>3]),+(+p[f+320>>3]));p[a+88>>3]=wa;break}case 1041:{wa=+p[f+32>>3];kb=+p[f+320>>3];Tb=+p[f+608>>3];z=+p[f+896>>3];Ub=+p[12420];if(Ub==0.0){p[12420]=.017453292519943295;kc=.017453292519943295}else kc=Ub;Ub=+_(+((Tb-wa)*kc*.5));wa=+_(+((z-kb)*kc*.5));Tb=wa*wa+Ub*(Ub*(+Z(+(kb*kc))*+Z(+(z*kc))));z=Tb<0.0?0.0:Tb;Tb=z>1.0?1.0:z;z=+da(+(+X(+Tb)),+(+X(+(1.0-Tb))))*2.0/kc;p[a+88>>3]=z;ic=80;break}case 1022:{ic=80;break}case 1023:{switch(k[a+52>>2]|0){case 260:{z=+p[f+32>>3];Tb=+p[f+320>>3];p[a+88>>3]=z<Tb?z:Tb;break p;break}case 259:{y=k[f+32>>2]|0;x=k[f+320>>2]|0;k[a+88>>2]=(y|0)<(x|0)?y:x;break p;break}default:break p}break}case 1024:{switch(k[a+52>>2]|0){case 260:{p[a+88>>3]=+p[f+32>>3];break p;break}case 259:{k[a+88>>2]=k[f+32>>2];break p;break}case 262:{ux(a+88|0,f+32|0)|0;break p;break}default:break p}break}case 1025:{switch(k[a+52>>2]|0){case 260:{Tb=+p[f+32>>3];z=+p[f+320>>3];p[a+88>>3]=Tb>z?Tb:z;break p;break}case 259:{x=k[f+32>>2]|0;y=k[f+320>>2]|0;k[a+88>>2]=(x|0)>(y|0)?x:y;break p;break}default:break p}break}case 1026:{y=+W(+(+p[f+32>>3]-+p[f+320>>3]))<+p[f+608>>3]&1;i[a+88>>0]=y;break}case 1027:{z=+p[f+608>>3];Tb=+p[f+896>>3]-+p[f+32>>3];kb=+p[f+1184>>3]-+p[f+320>>3];i[a+88>>0]=Tb*Tb+kb*kb<=z*z&1;break}case 1028:{z=+p[f+608>>3];kb=+p[f+896>>3];Tb=+p[f+1184>>3]/180.0*3.141592653589793;Ub=+p[f+1472>>3]-+p[f+32>>3];wa=+p[f+1760>>3]-+p[f+320>>3];ra=+Z(+Tb);Sb=+_(+Tb);Tb=Ub*ra+wa*Sb;Y=wa*ra-Ub*Sb;i[a+88>>0]=(!(Y<=kb*.5)|(!(Y>=kb*-.5)|(!(Tb>=z*-.5)|!(Tb<=z*.5))))&1^1;break}case 1029:{z=+p[f+1184>>3]/180.0*3.141592653589793;Tb=+p[f+1472>>3]-+p[f+32>>3];kb=+p[f+1760>>3]-+p[f+320>>3];Y=+Z(+z);Sb=+_(+z);z=(Tb*Y+kb*Sb)/+p[f+608>>3];Ub=(kb*Y-Tb*Sb)/+p[f+896>>3];i[a+88>>0]=z*z+Ub*Ub<=1.0&1;break}case 1034:{switch(k[a+52>>2]|0){case 258:{i[a+88>>0]=(i[f+608>>0]|0)!=0?i[f+32>>0]|0:i[f+320>>0]|0;break p;break}case 259:{k[a+88>>2]=(i[f+608>>0]|0)!=0?k[f+32>>2]|0:k[f+320>>2]|0;break p;break}case 260:{p[a+88>>3]=(i[f+608>>0]|0)!=0?+p[f+32>>3]:+p[f+320>>3];break p;break}case 261:{ux(a+88|0,((i[f+608>>0]|0)!=0?f:f+288|0)+32|0)|0;break p;break}default:break p}break}case 1044:{y=a+88|0;x=k[a+56>>2]|0;w=f+32|0;j=k[f>>2]|0;q=k[f+320>>2]|0;if(!j)lc=xx(w)|0;else lc=j;if((q|0)<0){if(!(k[30835]|0))k[30835]=431;Bx(c,790404,80)|0;i[c+79>>0]=0;tk(c);break p}do if(!((q|0)==0|(lc|0)<(q|0)))if((q+x|0)>(lc|0)){j=lc-q+1|0;py(y|0,w+(q+-1)|0,j|0)|0;sy(y+j|0,0,x-j|0)|0;break}else{py(y|0,w+(q+-1)|0,x|0)|0;break}else sy(y|0,0,x|0)|0;while(0);i[y+x>>0]=0;break}case 1045:{q=f+32|0;w=Ex(q,f+320|0)|0;if(!w){k[a+88>>2]=0;break p}else{k[a+88>>2]=(0-q|1)+w;break p}break}default:{}}while(0);q:do if((ic|0)==80)switch(k[a+52>>2]|0){case 260:{p[a+88>>3]=+p[f+32>>3];break q;break}case 259:{k[a+88>>2]=k[f+32>>2];break q;break}case 262:{ux(a+88|0,f+32|0)|0;break q;break}default:break q}while(0);k[a>>2]=-1e3}while(0);a=k[h>>2]|0;if(!a){r=b;return}else mc=a;do{mc=mc+-1|0;a=k[d+(mc<<2)>>2]|0;if((k[a>>2]|0)>0)dy(k[a+88>>2]|0)}while((mc|0)!=0);r=b;return}function Vu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0;b=r;r=r+144|0;c=b+64|0;d=b+40|0;e=b+20|0;f=b;g=k[30820]|0;h=k[a+12>>2]|0;j=k[a+8>>2]|0;l=j+-1|0;m=(l|0)==0;a:do if(m)n=1;else{o=j+-2|0;q=1;s=l;while(1){t=o;u=s;while(1){v=k[a+12+(u<<2)>>2]|0;w=g+(v*344|0)|0;k[d+(t<<2)>>2]=w;x=(k[w>>2]|0)==-1e3;k[e+(t<<2)>>2]=x&1;if(!x){y=t;break}k[f+(t<<2)>>2]=k[g+(v*344|0)+88>>2];if(!t){n=q;break a}else{v=t;t=t+-1|0;u=v}}if(!y){n=0;break}else{o=y+-1|0;q=0;s=y}}}while(0);y=a+52|0;switch(k[y>>2]|0){case 260:{z=8;break}case 259:{z=4;break}case 258:{z=1;break}default:z=0}Su(a);b:do if(!(k[30835]|0)){s=g+(h*344|0)+60|0;q=k[s>>2]|0;do if(n){if((q|0)!=(l|0)){if((l|0)!=1)break;o=k[f>>2]|0;if((o|0)>=1?(u=g+(h*344|0)+56|0,(o|0)<=(k[g+(h*344|0)+64+(q+-1<<2)>>2]|0)):0){t=a+56|0;v=k[t>>2]|0;x=ha(v,o+-1|0)|0;o=(k[30825]|0)>0;if(((k[y>>2]|0)+-261|0)>>>0>=2){if(!o)break b;w=a+84|0;A=g+(h*344|0)+84|0;B=a+88|0;C=g+(h*344|0)+88|0;D=v;v=x;E=0;while(1){F=(k[w>>2]|0)+(ha(D,E)|0)|0;py(F|0,(k[A>>2]|0)+v|0,D|0)|0;F=ha(E,z)|0;G=k[t>>2]|0;H=(k[B>>2]|0)+(ha(F,G)|0)|0;py(H|0,(k[C>>2]|0)+(ha(v,z)|0)|0,ha(G,z)|0)|0;G=E+1|0;if((G|0)>=(k[30825]|0))break b;D=k[t>>2]|0;v=(k[u>>2]|0)+v|0;E=G}}if(!o)break b;E=a+84|0;v=a+88|0;D=g+(h*344|0)+88|0;C=g+(h*344|0)+84|0;B=x;A=0;while(1){w=k[E>>2]|0;if(w)i[w+A>>0]=i[(k[C>>2]|0)+A>>0]|0;w=k[t>>2]|0;G=(k[k[v>>2]>>2]|0)+(ha(w+1|0,A)|0)|0;py(G|0,(k[k[D>>2]>>2]|0)+B|0,w|0)|0;i[(k[(k[v>>2]|0)+(A<<2)>>2]|0)+(k[t>>2]|0)>>0]=0;A=A+1|0;if((A|0)>=(k[30825]|0))break b;else B=B+1+(k[u>>2]|0)|0}}k[30835]=431;Bx(c,790646,80)|0;i[c+79>>0]=0;tk(c);dy(k[a+88>>2]|0);break b}c:do if(!m){u=j+-2|0;B=0;A=l;while(1){t=k[f+(u<<2)>>2]|0;if((t|0)<1){I=B;J=A;break}v=k[g+(h*344|0)+64+(u<<2)>>2]|0;if((t|0)>(v|0)){I=B;J=A;break}D=t+-1+(ha(v,B)|0)|0;if(!u){K=D;break c}else{v=u;u=u+-1|0;B=D;A=v}}if((J|0)<1)K=I;else{k[30835]=431;Bx(c,790646,80)|0;i[c+79>>0]=0;tk(c);dy(k[a+88>>2]|0);break b}}else K=0;while(0);if((k[30825]|0)<=0)break b;A=g+(h*344|0)+84|0;B=a+84|0;u=g+(h*344|0)+88|0;v=a+88|0;D=g+(h*344|0)+56|0;t=K;C=0;while(1){switch(k[y>>2]|0){case 261:{i[(k[B>>2]|0)+C>>0]=i[(k[A>>2]|0)+C>>0]|0;L=25;break}case 262:{L=29;break}default:{i[(k[B>>2]|0)+C>>0]=i[(k[A>>2]|0)+t>>0]|0;L=25}}d:do if((L|0)==25){L=0;switch(k[y>>2]|0){case 260:{p[(k[v>>2]|0)+(C<<3)>>3]=+p[(k[u>>2]|0)+(t<<3)>>3];break d;break}case 259:{k[(k[v>>2]|0)+(C<<2)>>2]=k[(k[u>>2]|0)+(t<<2)>>2];break d;break}case 258:{i[(k[v>>2]|0)+C>>0]=i[(k[u>>2]|0)+t>>0]|0;break d;break}default:{L=29;break d}}}while(0);if((L|0)==29){L=0;i[k[(k[v>>2]|0)+(C<<2)>>2]>>0]=i[(k[k[u>>2]>>2]|0)+(t+C)>>0]|0;i[(k[(k[v>>2]|0)+(C<<2)>>2]|0)+1>>0]=0}C=C+1|0;if((C|0)>=(k[30825]|0))break b;else t=(k[D>>2]|0)+t|0}}while(0);t=g+(h*344|0)+56|0;D=(k[30825]|0)>0;if((q|0)!=(l|0)){if(!D)break;C=k[d>>2]|0;v=C+84|0;u=C+88|0;C=c+79|0;A=a+88|0;B=a+56|0;E=a+84|0;x=g+(h*344|0)+88|0;o=g+(h*344|0)+84|0;w=0;while(1){if(i[(k[v>>2]|0)+w>>0]|0)break;G=k[(k[u>>2]|0)+(w<<2)>>2]|0;k[f>>2]=G;do if((G|0)>=1?(G|0)<=(k[g+(h*344|0)+64+((k[s>>2]|0)+-1<<2)>>2]|0):0){H=k[B>>2]|0;F=ha(H,G+-1|0)|0;M=k[t>>2]|0;if(((k[y>>2]|0)+-261|0)>>>0>=2){N=(ha(M,w)|0)+F|0;O=(k[E>>2]|0)+(ha(H,w)|0)|0;py(O|0,(k[o>>2]|0)+N|0,H|0)|0;O=ha(w,z)|0;P=k[B>>2]|0;Q=(k[A>>2]|0)+(ha(O,P)|0)|0;py(Q|0,(k[x>>2]|0)+(ha(N,z)|0)|0,ha(P,z)|0)|0;break}P=(ha(M+1|0,w)|0)+F|0;F=k[E>>2]|0;if(!F)R=H;else{i[F+w>>0]=i[(k[o>>2]|0)+w>>0]|0;R=k[B>>2]|0}F=(k[k[A>>2]>>2]|0)+(ha(R+1|0,w)|0)|0;py(F|0,(k[k[x>>2]>>2]|0)+P|0,R|0)|0;i[(k[(k[A>>2]|0)+(w<<2)>>2]|0)+(k[B>>2]|0)>>0]=0}else L=81;while(0);if((L|0)==81){L=0;if(!(k[30835]|0))k[30835]=431;Bx(c,790646,80)|0;i[C>>0]=0;tk(c);dy(k[A>>2]|0)}w=w+1|0;if((w|0)>=(k[30825]|0))break b}if(!(k[30835]|0))k[30835]=431;Bx(c,790665,80)|0;i[C>>0]=0;tk(c);dy(k[A>>2]|0);break}if(D){w=(j|0)>1;B=j+-2|0;x=g+(h*344|0)+84|0;o=a+84|0;E=g+(h*344|0)+88|0;s=a+88|0;u=c+79|0;v=0;do{e:do if(w){q=0;while(1){if(!(k[e+(q<<2)>>2]|0)){G=k[d+(q<<2)>>2]|0;if(i[(k[G+84>>2]|0)+v>>0]|0)break;k[f+(q<<2)>>2]=k[(k[G+88>>2]|0)+(v<<2)>>2]}q=q+1|0;if((q|0)>=(l|0))break e}if(!(k[30835]|0))k[30835]=431;Bx(c,790665,80)|0;i[u>>0]=0;tk(c);dy(k[s>>2]|0)}while(0);if(k[30835]|0)break b;f:do if(!m){q=B;G=0;P=l;while(1){F=k[f+(q<<2)>>2]|0;if((F|0)<1){S=G;T=P;break}H=k[g+(h*344|0)+64+(q<<2)>>2]|0;if((F|0)>(H|0)){S=G;T=P;break}M=F+-1+(ha(H,G)|0)|0;if(!q){U=M;L=65;break f}else{H=q;q=q+-1|0;G=M;P=H}}if((T|0)<1){U=S;L=65}else{k[30835]=431;Bx(c,790646,80)|0;i[u>>0]=0;tk(c);dy(k[s>>2]|0)}}else{U=0;L=65}while(0);g:do if((L|0)==65){L=0;P=(ha(k[t>>2]|0,v)|0)+U|0;switch(k[y>>2]|0){case 261:{i[(k[o>>2]|0)+v>>0]=i[(k[x>>2]|0)+v>>0]|0;L=68;break}case 262:break;default:{i[(k[o>>2]|0)+v>>0]=i[(k[x>>2]|0)+P>>0]|0;L=68}}h:do if((L|0)==68){L=0;switch(k[y>>2]|0){case 260:{p[(k[s>>2]|0)+(v<<3)>>3]=+p[(k[E>>2]|0)+(P<<3)>>3];break g;break}case 259:{k[(k[s>>2]|0)+(v<<2)>>2]=k[(k[E>>2]|0)+(P<<2)>>2];break g;break}case 258:{i[(k[s>>2]|0)+v>>0]=i[(k[E>>2]|0)+P>>0]|0;break g;break}default:break h}}while(0);i[k[(k[s>>2]|0)+(v<<2)>>2]>>0]=i[(k[k[E>>2]>>2]|0)+(P+v)>>0]|0;i[(k[(k[s>>2]|0)+(v<<2)>>2]|0)+1>>0]=0}while(0);v=v+1|0}while((v|0)<(k[30825]|0))}}while(0);do if((k[g+(h*344|0)>>2]|0)>0){y=g+(h*344|0)+88|0;if(((k[g+(h*344|0)+52>>2]|0)+-261|0)>>>0<2){dy(k[k[y>>2]>>2]|0);break}else{dy(k[y>>2]|0);break}}while(0);if((j|0)>1)V=0;else{r=b;return}do{j=k[d+(V<<2)>>2]|0;if((k[j>>2]|0)>0)dy(k[j+88>>2]|0);V=V+1|0}while((V|0)!=(l|0));r=b;return}function Wu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;b=k[30820]|0;c=k[a+12>>2]|0;d=k[a+16>>2]|0;e=b+(c*344|0)|0;f=(k[e>>2]|0)==-1e3;g=b+(d*344|0)|0;h=(k[g>>2]|0)==-1e3;j=b+(c*344|0)+88|0;c=f?j:0;l=b+(d*344|0)+88|0;d=h?l:0;a:do if(!(f&h)){Su(a);if(!(k[30835]|0)){b=k[30825]|0;switch(k[a>>2]|0){case 279:case 280:case 281:case 282:case 278:case 277:{if(!b)break a;m=a+84|0;n=a+88|0;o=b;p=c;q=d;while(1){o=o+-1|0;if(f)r=p;else r=k[(k[j>>2]|0)+(o<<2)>>2]|0;if(h)s=q;else s=k[(k[l>>2]|0)+(o<<2)>>2]|0;t=k[a>>2]|0;switch(t|0){case 277:{u=($u(r,s)|0)<<24>>24==0&1;i[(k[n>>2]|0)+o>>0]=u;break}case 278:{u=$u(r,s)|0;i[(k[n>>2]|0)+o>>0]=u;break}case 279:case 280:case 281:case 282:{u=av(r,t,s)|0;i[(k[n>>2]|0)+o>>0]=u;break}default:{}}i[(k[m>>2]|0)+o>>0]=0;if(!o)break a;else{p=r;q=s}}break}case 43:case 38:case 124:{if(!b)break a;q=a+88|0;p=b;o=c;m=d;while(1){p=p+-1|0;if(f)v=o;else v=k[(k[j>>2]|0)+(p<<2)>>2]|0;if(h)w=m;else w=k[(k[l>>2]|0)+(p<<2)>>2]|0;n=k[a>>2]|0;do if((n|0)!=124){u=k[(k[q>>2]|0)+(p<<2)>>2]|0;if((n|0)==38){cv(u,v,w);break}else{ux(u,v)|0;qx(k[(k[q>>2]|0)+(p<<2)>>2]|0,w)|0;break}}else bv(k[(k[q>>2]|0)+(p<<2)>>2]|0,v,w);while(0);if(!p)break;else{o=v;m=w}}break}case 289:{m=k[l>>2]|0;if((b|0)>0){o=a+88|0;p=a+84|0;q=0;n=m;while(1){u=0;t=k[(k[j>>2]|0)+(q<<2)>>2]|0;b:while(1){switch(i[t>>0]|0){case 0:{x=u;break b;break}case 49:{y=u+1|0;break}default:y=u}u=y;t=t+1|0}t=x+n|0;k[(k[o>>2]|0)+(q<<2)>>2]=t;i[(k[p>>2]|0)+q>>0]=0;q=q+1|0;if((q|0)==(b|0)){z=t;break}else n=t}}else z=m;k[l>>2]=z;break a;break}default:break a}}}else{n=k[a>>2]|0;c:do switch(n|0){case 277:{b=($u(c,d)|0)<<24>>24==0&1;i[a+88>>0]=b;break}case 278:{b=$u(c,d)|0;i[a+88>>0]=b;break}case 279:case 280:case 281:case 282:{b=av(c,n,d)|0;i[a+88>>0]=b;break}case 124:{bv(a+88|0,c,d);break}case 38:{cv(a+88|0,c,d);break}case 43:{b=a+88|0;ux(b,c)|0;qx(b,d)|0;break}case 289:{b=a+88|0;k[b>>2]=0;q=0;p=c;while(1){switch(i[p>>0]|0){case 0:{break c;break}case 49:{o=q+1|0;k[b>>2]=o;A=o;break}default:A=q}q=A;p=p+1|0}break}default:{}}while(0);k[a>>2]=-1e3}while(0);if((k[e>>2]|0)>0){dy(k[k[j>>2]>>2]|0);dy(k[j>>2]|0)}if((k[g>>2]|0)<=0)return;dy(k[k[l>>2]>>2]|0);dy(k[l>>2]|0);return}function Xu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0;b=k[30820]|0;c=k[a+12>>2]|0;d=k[a+16>>2]|0;e=b+(c*344|0)|0;f=(k[e>>2]|0)==-1e3;g=b+(d*344|0)|0;h=(k[g>>2]|0)==-1e3;j=b+(c*344|0)+88|0;l=f?j:0;m=b+(d*344|0)+88|0;n=h?m:0;a:do if(!(f&h)){Su(a);if(!(k[30835]|0)){o=k[30825]|0;switch(k[a>>2]|0){case 278:case 277:{p=a+84|0;q=b+(d*344|0)+84|0;r=b+(c*344|0)+84|0;s=a+88|0;t=0;u=0;v=o;w=l;x=n;while(1){if(f){y=u;z=v;while(1){A=z+-1|0;if(!z)break a;if(h)B=y;else B=i[(k[q>>2]|0)+A>>0]|0;i[(k[p>>2]|0)+A>>0]=(B|t)<<24>>24!=0&1;if(!(i[(k[p>>2]|0)+A>>0]|0)){C=A;D=t;E=B;break}else{y=B;z=A}}}else{z=u;y=v;while(1){A=y+-1|0;if(!y)break a;F=i[(k[r>>2]|0)+A>>0]|0;if(h)G=z;else G=i[(k[q>>2]|0)+A>>0]|0;i[(k[p>>2]|0)+A>>0]=(G|F)<<24>>24!=0&1;if(!(i[(k[p>>2]|0)+A>>0]|0)){C=A;D=F;E=G;break}else{z=G;y=A}}}if(f)H=w;else H=k[(k[j>>2]|0)+(C<<2)>>2]|0;if(h)I=x;else I=k[(k[m>>2]|0)+(C<<2)>>2]|0;y=i[H>>0]|0;z=i[I>>0]|0;if(y<<24>>24>=z<<24>>24)if(y<<24>>24>z<<24>>24)J=1;else J=tx(H,I)|0;else J=-1;i[(k[s>>2]|0)+C>>0]=((J|0)==0^(k[a>>2]|0)!=278)&1;t=D;u=E;v=C;w=H;x=I}break}case 281:case 282:{x=a+84|0;w=b+(d*344|0)+84|0;v=b+(c*344|0)+84|0;u=a+88|0;t=0;s=0;p=o;q=l;r=n;while(1){z=t;y=s;A=p;while(1){F=A+-1|0;if(!A)break a;if(f)K=z;else K=i[(k[v>>2]|0)+F>>0]|0;if(h)L=y;else L=i[(k[w>>2]|0)+F>>0]|0;i[(k[x>>2]|0)+F>>0]=(L|K)<<24>>24!=0&1;if(!(i[(k[x>>2]|0)+F>>0]|0)){M=F;N=K;O=L;break}else{z=K;y=L;A=F}}if(f)P=q;else P=k[(k[j>>2]|0)+(M<<2)>>2]|0;if(h)Q=r;else Q=k[(k[m>>2]|0)+(M<<2)>>2]|0;A=i[P>>0]|0;y=i[Q>>0]|0;if(A<<24>>24>=y<<24>>24)if(A<<24>>24>y<<24>>24)R=1;else R=tx(P,Q)|0;else R=-1;i[(k[u>>2]|0)+M>>0]=((k[a>>2]|0)==282?(R|0)>0:(R|0)<0)&1;t=N;s=O;p=M;q=P;r=Q}break}case 280:case 279:{r=a+84|0;q=b+(d*344|0)+84|0;p=b+(c*344|0)+84|0;s=a+88|0;t=0;u=0;x=o;w=l;v=n;while(1){y=t;A=u;z=x;while(1){F=z+-1|0;if(!z)break a;if(f)S=y;else S=i[(k[p>>2]|0)+F>>0]|0;if(h)T=A;else T=i[(k[q>>2]|0)+F>>0]|0;i[(k[r>>2]|0)+F>>0]=(T|S)<<24>>24!=0&1;if(!(i[(k[r>>2]|0)+F>>0]|0)){U=F;V=S;W=T;break}else{y=S;A=T;z=F}}if(f)X=w;else X=k[(k[j>>2]|0)+(U<<2)>>2]|0;if(h)Y=v;else Y=k[(k[m>>2]|0)+(U<<2)>>2]|0;z=i[X>>0]|0;A=i[Y>>0]|0;if(z<<24>>24>=A<<24>>24)if(z<<24>>24>A<<24>>24)Z=1;else Z=tx(X,Y)|0;else Z=-1;i[(k[s>>2]|0)+U>>0]=((k[a>>2]|0)==279?(Z|0)>-1:(Z|0)<1)&1;t=V;u=W;x=U;w=X;v=Y}break}case 43:{v=a+84|0;w=b+(d*344|0)+84|0;x=b+(c*344|0)+84|0;u=a+88|0;t=0;s=0;r=o;q=l;p=n;while(1){A=t;z=s;y=r;while(1){F=y+-1|0;if(!y)break a;if(f)_=A;else _=i[(k[x>>2]|0)+F>>0]|0;if(h)$=z;else $=i[(k[w>>2]|0)+F>>0]|0;i[(k[v>>2]|0)+F>>0]=($|_)<<24>>24!=0&1;if(!(i[(k[v>>2]|0)+F>>0]|0)){aa=F;ba=_;ca=$;break}else{A=_;z=$;y=F}}if(f)da=q;else da=k[(k[j>>2]|0)+(aa<<2)>>2]|0;if(h)ea=p;else ea=k[(k[m>>2]|0)+(aa<<2)>>2]|0;ux(k[(k[u>>2]|0)+(aa<<2)>>2]|0,da)|0;qx(k[(k[u>>2]|0)+(aa<<2)>>2]|0,ea)|0;t=ba;s=ca;r=aa;q=da;p=ea}break}default:break a}}}else{p=k[a>>2]|0;switch(p|0){case 278:case 277:{if((i[l>>0]|0)==(i[n>>0]|0))fa=(tx(l,n)|0)==0;else fa=0;i[a+88>>0]=(fa^(p|0)!=278)&1;break}case 282:{p=i[l>>0]|0;q=i[n>>0]|0;if(p<<24>>24>=q<<24>>24)if(p<<24>>24>q<<24>>24)ga=1;else ga=(tx(l,n)|0)>0&1;else ga=0;i[a+88>>0]=ga;break}case 281:{q=i[l>>0]|0;p=i[n>>0]|0;if(q<<24>>24>=p<<24>>24)if(q<<24>>24>p<<24>>24)ha=0;else ha=(tx(l,n)|0)>>>31&255;else ha=1;i[a+88>>0]=ha;break}case 279:{p=i[l>>0]|0;q=i[n>>0]|0;if(p<<24>>24>=q<<24>>24)if(p<<24>>24>q<<24>>24)ia=1;else ia=(tx(l,n)|0)>>>31&255^1;else ia=0;i[a+88>>0]=ia;break}case 280:{q=i[l>>0]|0;p=i[n>>0]|0;if(q<<24>>24>=p<<24>>24)if(q<<24>>24>p<<24>>24)ja=0;else ja=(tx(l,n)|0)<1&1;else ja=1;i[a+88>>0]=ja;break}case 43:{p=a+88|0;ux(p,l)|0;qx(p,n)|0;break}default:{}}k[a>>2]=-1e3}while(0);if((k[e>>2]|0)>0){dy(k[k[j>>2]>>2]|0);dy(k[j>>2]|0)}if((k[g>>2]|0)<=0)return;dy(k[k[m>>2]>>2]|0);dy(k[m>>2]|0);return}function Yu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0;b=k[30820]|0;c=k[a+12>>2]|0;d=k[a+16>>2]|0;e=b+(c*344|0)|0;if((k[e>>2]|0)==-1e3){f=i[b+(c*344|0)+88>>0]|0;g=0}else{f=0;g=k[b+(c*344|0)+56>>2]|0}h=b+(d*344|0)|0;if((k[h>>2]|0)==-1e3){j=i[b+(d*344|0)+88>>0]|0;l=0}else{j=0;l=k[b+(d*344|0)+56>>2]|0}m=(g|0)==0;n=(l|0)==0;o=k[a>>2]|0;do if(l|g){p=k[30825]|0;q=a+56|0;r=k[q>>2]|0;s=ha(r,p)|0;Su(a);t=(k[30835]|0)!=0;if((o|0)==289){if(t)break;u=b+(d*344|0)+88|0;v=k[u>>2]|0;if((s|0)>0){w=b+(c*344|0)+84|0;x=b+(c*344|0)+88|0;y=a+88|0;z=a+84|0;A=0;B=v;while(1){if(!(i[(k[w>>2]|0)+A>>0]|0))C=(i[(k[x>>2]|0)+A>>0]|0)+B|0;else C=B;k[(k[y>>2]|0)+(A<<2)>>2]=C;i[(k[z>>2]|0)+A>>0]=0;A=A+1|0;if((A|0)==(s|0)){D=C;break}else B=C}}else D=v;k[u>>2]=D;break}if(!t){if((k[a>>2]|0)==289){B=b+(d*344|0)+88|0;A=k[B>>2]|0;if((s|0)>0){z=b+(c*344|0)+84|0;y=b+(c*344|0)+88|0;x=a+88|0;w=a+84|0;E=0;F=A;while(1){if(!(i[(k[z>>2]|0)+E>>0]|0))G=(i[(k[y>>2]|0)+E>>0]|0)+F|0;else G=F;k[(k[x>>2]|0)+(E<<2)>>2]=G;i[(k[w>>2]|0)+E>>0]=0;E=E+1|0;if((E|0)==(s|0)){H=G;break}else F=G}}else H=A;k[B>>2]=H}if(p){F=(g|0)>1;E=b+(c*344|0)+88|0;w=b+(c*344|0)+84|0;x=(l|0)>1;y=b+(d*344|0)+88|0;z=b+(d*344|0)+84|0;t=a+84|0;u=a+88|0;v=p;I=s;J=r;K=0;L=0;M=f;N=j;while(1){v=v+-1|0;if(!J){O=I;P=K;Q=L;R=M;S=N}else{T=J;U=I;V=K;W=L;X=M;Y=N;while(1){T=T+-1|0;U=U+-1|0;if(!F)if(m){Z=V;_=X}else{Z=i[(k[w>>2]|0)+v>>0]|0;_=i[(k[E>>2]|0)+v>>0]|0}else{Z=i[(k[w>>2]|0)+U>>0]|0;_=i[(k[E>>2]|0)+U>>0]|0}if(!x)if(n){$=W;aa=Y}else{$=i[(k[z>>2]|0)+v>>0]|0;aa=i[(k[y>>2]|0)+v>>0]|0}else{$=i[(k[z>>2]|0)+U>>0]|0;aa=i[(k[y>>2]|0)+U>>0]|0}ba=Z<<24>>24==0;ca=$<<24>>24!=0;i[(k[t>>2]|0)+U>>0]=(ca|ba^1)&1;a:do switch(k[a>>2]|0){case 275:{if(!(($|Z)<<24>>24)){i[(k[u>>2]|0)+U>>0]=(aa|_)<<24>>24!=0&1;da=Z;break a}if(aa<<24>>24==0|(ba|ca)?_<<24>>24==0|($<<24>>24==0|Z<<24>>24!=0):0){da=Z;break a}i[(k[u>>2]|0)+U>>0]=1;i[(k[t>>2]|0)+U>>0]=0;da=Z;break}case 276:{if(!(($|Z)<<24>>24)){i[(k[u>>2]|0)+U>>0]=_<<24>>24!=0&aa<<24>>24!=0&1;da=Z;break a}if(ba|($|aa)<<24>>24!=0){if(!ba){da=Z;break a}if(_<<24>>24!=0|$<<24>>24==0){da=0;break a}}i[(k[u>>2]|0)+U>>0]=0;i[(k[t>>2]|0)+U>>0]=0;da=Z;break}case 278:{i[(k[u>>2]|0)+U>>0]=((aa|_)<<24>>24==0|_<<24>>24!=0&aa<<24>>24!=0)&1;da=Z;break}case 277:{ea=_<<24>>24==0;fa=aa<<24>>24!=0;i[(k[u>>2]|0)+U>>0]=(ea&fa|(ea|fa)^1)&1;da=Z;break}default:da=Z}while(0);if(!T){ga=da;ia=$;ja=_;ka=aa;break}else{V=da;W=$;X=_;Y=aa}}O=I-J|0;P=ga;Q=ia;R=ja;S=ka}if(!v)break;else{I=O;J=k[q>>2]|0;K=P;L=Q;M=R;N=S}}}}}else{switch(o|0){case 275:{i[a+88>>0]=f<<24>>24==0?j<<24>>24!=0&1:1;break}case 276:{i[a+88>>0]=f<<24>>24==0?0:j<<24>>24!=0&1;break}case 278:{N=f<<24>>24==0;M=j<<24>>24==0;i[a+88>>0]=N|M?(N?M&1:0):1;break}case 277:{M=f<<24>>24==0;N=j<<24>>24!=0;i[a+88>>0]=M|N?(M?N&1:0):1;break}case 289:{k[a+88>>2]=f<<24>>24;break}default:{}}k[a>>2]=-1e3}while(0);if((k[e>>2]|0)>0)dy(k[b+(c*344|0)+88>>2]|0);if((k[h>>2]|0)<=0)return;dy(k[b+(d*344|0)+88>>2]|0);return}function Zu(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0;b=r;r=r+80|0;c=b;d=k[30820]|0;e=k[a+12>>2]|0;f=k[a+16>>2]|0;g=d+(e*344|0)|0;if((k[g>>2]|0)==-1e3){h=k[d+(e*344|0)+88>>2]|0;j=0}else{h=0;j=k[d+(e*344|0)+56>>2]|0}l=d+(f*344|0)|0;if((k[l>>2]|0)==-1e3){m=k[d+(f*344|0)+88>>2]|0;n=0}else{m=0;n=k[d+(f*344|0)+56>>2]|0}o=(j|0)==0;p=(n|0)==0;q=k[a>>2]|0;do if(n|j){s=k[30825]|0;t=a+56|0;u=k[t>>2]|0;v=ha(u,s)|0;Su(a);if((q&-2|0)==288){if(k[30835]|0)break;w=d+(f*344|0)+88|0;x=k[w>>2]|0;y=d+(f*344|0)+84|0;z=k[y>>2]|0;A=(v|0)>0;if((k[a>>2]|0)==289)if(A){B=d+(e*344|0)+84|0;C=d+(e*344|0)+88|0;D=a+88|0;E=a+84|0;F=0;G=x;while(1){if(!(i[(k[B>>2]|0)+F>>0]|0))H=(k[(k[C>>2]|0)+(F<<2)>>2]|0)+G|0;else H=G;k[(k[D>>2]|0)+(F<<2)>>2]=H;i[(k[E>>2]|0)+F>>0]=0;F=F+1|0;if((F|0)==(v|0)){I=H;J=z;break}else G=H}}else{I=x;J=z}else if(A){G=d+(e*344|0)+88|0;F=d+(e*344|0)+84|0;E=a+88|0;D=a+84|0;C=k[F>>2]|0;B=0;K=x;L=z;while(1){M=k[(k[G>>2]|0)+(B<<2)>>2]|0;if((L|0)!=0|(i[C+B>>0]|0)!=0){k[(k[E>>2]|0)+(B<<2)>>2]=0;i[(k[D>>2]|0)+B>>0]=1}else{k[(k[E>>2]|0)+(B<<2)>>2]=M-K;i[(k[D>>2]|0)+B>>0]=0}C=k[F>>2]|0;N=i[C+B>>0]|0;B=B+1|0;if((B|0)==(v|0)){I=M;J=N;break}else{K=M;L=N}}}else{I=x;J=z}k[w>>2]=I;k[y>>2]=J;break}if((s|0)!=0&(k[30835]|0)==0){L=(j|0)>1;K=d+(e*344|0)+88|0;B=d+(e*344|0)+84|0;C=(n|0)>1;F=d+(f*344|0)+88|0;D=d+(f*344|0)+84|0;E=a+84|0;G=a+88|0;A=s;N=v;M=u;O=0;P=0;Q=h;R=m;while(1){A=A+-1|0;a:do if(!M){S=0;T=N;U=O;V=P;W=Q;X=R}else{Z=M;_=N;$=O;aa=P;ba=Q;ca=R;while(1){Z=Z+-1|0;da=_+-1|0;if(!L)if(o){ea=$;fa=ba}else{ea=i[(k[B>>2]|0)+A>>0]|0;fa=k[(k[K>>2]|0)+(A<<2)>>2]|0}else{ea=i[(k[B>>2]|0)+da>>0]|0;fa=k[(k[K>>2]|0)+(da<<2)>>2]|0}if(!C)if(p){ga=aa;ia=ca}else{ga=i[(k[D>>2]|0)+A>>0]|0;ia=k[(k[F>>2]|0)+(A<<2)>>2]|0}else{ga=i[(k[D>>2]|0)+da>>0]|0;ia=k[(k[F>>2]|0)+(da<<2)>>2]|0}i[(k[E>>2]|0)+da>>0]=(ga|ea)<<24>>24!=0&1;b:do switch(k[a>>2]|0){case 278:case 126:{i[(k[G>>2]|0)+da>>0]=(fa|0)==(ia|0)&1;ja=ia;break}case 277:{i[(k[G>>2]|0)+da>>0]=(fa|0)!=(ia|0)&1;ja=ia;break}case 282:{i[(k[G>>2]|0)+da>>0]=(fa|0)>(ia|0)&1;ja=ia;break}case 281:{i[(k[G>>2]|0)+da>>0]=(fa|0)<(ia|0)&1;ja=ia;break}case 280:{i[(k[G>>2]|0)+da>>0]=(fa|0)<=(ia|0)&1;ja=ia;break}case 279:{i[(k[G>>2]|0)+da>>0]=(fa|0)>=(ia|0)&1;ja=ia;break}case 43:{k[(k[G>>2]|0)+(da<<2)>>2]=ia+fa;ja=ia;break}case 45:{k[(k[G>>2]|0)+(da<<2)>>2]=fa-ia;ja=ia;break}case 42:{ka=ha(ia,fa)|0;k[(k[G>>2]|0)+(da<<2)>>2]=ka;ja=ia;break}case 37:{if(!ia){k[(k[G>>2]|0)+(da<<2)>>2]=0;i[(k[E>>2]|0)+da>>0]=1;ja=0;break b}else{k[(k[G>>2]|0)+(da<<2)>>2]=(fa|0)%(ia|0)|0;ja=ia;break b}break}case 47:{if(!ia){k[(k[G>>2]|0)+(da<<2)>>2]=0;i[(k[E>>2]|0)+da>>0]=1;ja=0;break b}else{k[(k[G>>2]|0)+(da<<2)>>2]=(fa|0)/(ia|0)|0;ja=ia;break b}break}case 283:{ka=~~+Y(+(+(fa|0)),+(+(ia|0)));k[(k[G>>2]|0)+(da<<2)>>2]=ka;ja=ia;break}default:ja=ia}while(0);ka=k[30835]|0;if(!((Z|0)!=0&(ka|0)==0)){S=ka;T=da;U=ea;V=ga;W=fa;X=ja;break a}else{_=da;$=ea;aa=ga;ba=fa;ca=ja}}}while(0);if(!((A|0)!=0&(S|0)==0))break;else{N=T;M=k[t>>2]|0;O=U;P=V;Q=W;R=X}}}}else{c:do switch(q|0){case 278:case 126:{i[a+88>>0]=(h|0)==(m|0)&1;break}case 277:{i[a+88>>0]=(h|0)!=(m|0)&1;break}case 282:{i[a+88>>0]=(h|0)>(m|0)&1;break}case 281:{i[a+88>>0]=(h|0)<(m|0)&1;break}case 280:{i[a+88>>0]=(h|0)<=(m|0)&1;break}case 279:{i[a+88>>0]=(h|0)>=(m|0)&1;break}case 43:{k[a+88>>2]=m+h;break}case 45:{k[a+88>>2]=h-m;break}case 42:{R=ha(m,h)|0;k[a+88>>2]=R;break}case 37:{if(m){k[a+88>>2]=(h|0)%(m|0)|0;break c}if(!(k[30835]|0))k[30835]=431;Bx(c,790748,80)|0;i[c+79>>0]=0;tk(c);break}case 47:{if(m){k[a+88>>2]=(h|0)/(m|0)|0;break c}if(!(k[30835]|0))k[30835]=431;Bx(c,790748,80)|0;i[c+79>>0]=0;tk(c);break}case 283:{R=~~+Y(+(+(h|0)),+(+(m|0)));k[a+88>>2]=R;break}case 289:{k[a+88>>2]=h;break}case 288:{k[a+88>>2]=0;break}default:{}}while(0);k[a>>2]=-1e3}while(0);if((k[g>>2]|0)>0)dy(k[d+(e*344|0)+88>>2]|0);if((k[l>>2]|0)<=0){r=b;return}dy(k[d+(f*344|0)+88>>2]|0);r=b;return}function _u(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0.0,j=0,l=0,m=0.0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0.0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0.0,I=0.0,J=0.0,K=0,L=0,M=0,N=0.0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,X=0,Z=0.0,_=0.0,$=0,aa=0,ba=0,ca=0,da=0.0,ea=0.0,fa=0,ga=0,ia=0.0,ja=0,ka=0.0,la=0,ma=0.0;b=r;r=r+80|0;c=b;d=k[30820]|0;e=k[a+12>>2]|0;f=k[a+16>>2]|0;g=d+(e*344|0)|0;if((k[g>>2]|0)==-1e3){h=+p[d+(e*344|0)+88>>3];j=0}else{h=0.0;j=k[d+(e*344|0)+56>>2]|0}l=d+(f*344|0)|0;if((k[l>>2]|0)==-1e3){m=+p[d+(f*344|0)+88>>3];n=0}else{m=0.0;n=k[d+(f*344|0)+56>>2]|0}o=(j|0)==0;q=(n|0)==0;s=k[a>>2]|0;do if(n|j){t=k[30825]|0;u=a+56|0;v=k[u>>2]|0;w=ha(v,t)|0;Su(a);if((s&-2|0)==288){if(k[30835]|0)break;x=d+(f*344|0)+88|0;y=+p[x>>3];z=d+(f*344|0)+84|0;A=k[z>>2]|0;B=(w|0)>0;if((k[a>>2]|0)==289)if(B){C=d+(e*344|0)+84|0;D=d+(e*344|0)+88|0;E=a+88|0;F=a+84|0;G=0;H=y;while(1){if(!(i[(k[C>>2]|0)+G>>0]|0))I=H+ +p[(k[D>>2]|0)+(G<<3)>>3];else I=H;p[(k[E>>2]|0)+(G<<3)>>3]=I;i[(k[F>>2]|0)+G>>0]=0;G=G+1|0;if((G|0)==(w|0)){J=I;K=A;break}else H=I}}else{J=y;K=A}else if(B){G=d+(e*344|0)+88|0;F=d+(e*344|0)+84|0;E=a+88|0;D=a+84|0;C=k[F>>2]|0;L=0;H=y;M=A;while(1){N=+p[(k[G>>2]|0)+(L<<3)>>3];if((M|0)!=0|(i[C+L>>0]|0)!=0){p[(k[E>>2]|0)+(L<<3)>>3]=0.0;i[(k[D>>2]|0)+L>>0]=1}else{p[(k[E>>2]|0)+(L<<3)>>3]=N-H;i[(k[D>>2]|0)+L>>0]=0}C=k[F>>2]|0;O=i[C+L>>0]|0;L=L+1|0;if((L|0)==(w|0)){J=N;K=O;break}else{H=N;M=O}}}else{J=y;K=A}p[x>>3]=J;k[z>>2]=K;break}if((t|0)!=0&(k[30835]|0)==0){M=(j|0)>1;L=d+(e*344|0)+88|0;C=d+(e*344|0)+84|0;F=(n|0)>1;D=d+(f*344|0)+88|0;E=d+(f*344|0)+84|0;G=a+84|0;B=a+88|0;O=t;P=w;Q=v;R=0;S=0;H=h;N=m;while(1){O=O+-1|0;a:do if(!Q){T=0;U=P;V=R;X=S;Z=H;_=N}else{$=Q;aa=P;ba=R;ca=S;da=H;ea=N;while(1){$=$+-1|0;fa=aa+-1|0;if(!M)if(o){ga=ba;ia=da}else{ga=i[(k[C>>2]|0)+O>>0]|0;ia=+p[(k[L>>2]|0)+(O<<3)>>3]}else{ga=i[(k[C>>2]|0)+fa>>0]|0;ia=+p[(k[L>>2]|0)+(fa<<3)>>3]}if(!F)if(q){ja=ca;ka=ea}else{ja=i[(k[E>>2]|0)+O>>0]|0;ka=+p[(k[D>>2]|0)+(O<<3)>>3]}else{ja=i[(k[E>>2]|0)+fa>>0]|0;ka=+p[(k[D>>2]|0)+(fa<<3)>>3]}i[(k[G>>2]|0)+fa>>0]=(ja|ga)<<24>>24!=0&1;b:do switch(k[a>>2]|0){case 126:{la=+W(+(ia-ka))<1.0e-07&1;i[(k[B>>2]|0)+fa>>0]=la;break}case 278:{i[(k[B>>2]|0)+fa>>0]=ia==ka&1;break}case 277:{i[(k[B>>2]|0)+fa>>0]=ia!=ka&1;break}case 282:{i[(k[B>>2]|0)+fa>>0]=ia>ka&1;break}case 281:{i[(k[B>>2]|0)+fa>>0]=ia<ka&1;break}case 280:{i[(k[B>>2]|0)+fa>>0]=ia<=ka&1;break}case 279:{i[(k[B>>2]|0)+fa>>0]=ia>=ka&1;break}case 43:{p[(k[B>>2]|0)+(fa<<3)>>3]=ia+ka;break}case 45:{p[(k[B>>2]|0)+(fa<<3)>>3]=ia-ka;break}case 42:{p[(k[B>>2]|0)+(fa<<3)>>3]=ia*ka;break}case 37:{if(ka!=0.0){p[(k[B>>2]|0)+(fa<<3)>>3]=ia-ka*+(~~(ia/ka)|0);break b}else{p[(k[B>>2]|0)+(fa<<3)>>3]=0.0;i[(k[G>>2]|0)+fa>>0]=1;break b}break}case 47:{if(ka!=0.0){p[(k[B>>2]|0)+(fa<<3)>>3]=ia/ka;break b}else{p[(k[B>>2]|0)+(fa<<3)>>3]=0.0;i[(k[G>>2]|0)+fa>>0]=1;break b}break}case 283:{ma=+Y(+ia,+ka);p[(k[B>>2]|0)+(fa<<3)>>3]=ma;break}default:{}}while(0);la=k[30835]|0;if(!(($|0)!=0&(la|0)==0)){T=la;U=fa;V=ga;X=ja;Z=ia;_=ka;break a}else{aa=fa;ba=ga;ca=ja;da=ia;ea=ka}}}while(0);if(!((O|0)!=0&(T|0)==0))break;else{P=U;Q=k[u>>2]|0;R=V;S=X;H=Z;N=_}}}}else{c:do switch(s|0){case 126:{S=+W(+(h-m))<1.0e-07&1;i[a+88>>0]=S;break}case 278:{i[a+88>>0]=h==m&1;break}case 277:{i[a+88>>0]=h!=m&1;break}case 282:{i[a+88>>0]=h>m&1;break}case 281:{i[a+88>>0]=h<m&1;break}case 280:{i[a+88>>0]=h<=m&1;break}case 279:{i[a+88>>0]=h>=m&1;break}case 43:{p[a+88>>3]=h+m;break}case 45:{p[a+88>>3]=h-m;break}case 42:{p[a+88>>3]=h*m;break}case 37:{if(m!=0.0){p[a+88>>3]=h-m*+(~~(h/m)|0);break c}if(!(k[30835]|0))k[30835]=431;Bx(c,790748,80)|0;i[c+79>>0]=0;tk(c);break}case 47:{if(m!=0.0){p[a+88>>3]=h/m;break c}if(!(k[30835]|0))k[30835]=431;Bx(c,790748,80)|0;i[c+79>>0]=0;tk(c);break}case 283:{N=+Y(+h,+m);p[a+88>>3]=N;break}case 289:{p[a+88>>3]=h;break}case 288:{p[a+88>>3]=0.0;break}default:{}}while(0);k[a>>2]=-1e3}while(0);if((k[g>>2]|0)>0)dy(k[d+(e*344|0)+88>>2]|0);if((k[l>>2]|0)<=0){r=b;return}dy(k[d+(f*344|0)+88>>2]|0);r=b;return}function $u(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0;c=r;r=r+256|0;d=c;e=xx(a)|0;f=xx(b)|0;if((e|0)>=(f|0))if((f|0)<(e|0)){if((e|0)==(f|0))g=0;else{h=e-f|0;sy(d|0,48,h|0)|0;g=h}if(!f)j=g;else{h=b;k=g;l=f;while(1){l=l+-1|0;i[d+k>>0]=i[h>>0]|0;if(!l)break;else{h=h+1|0;k=k+1|0}}j=f+g|0}i[d+j>>0]=0;m=a;n=d}else{m=a;n=b}else{if((f|0)==(e|0))o=0;else{j=f-e|0;sy(d|0,48,j|0)|0;o=j}if(!e)p=o;else{j=a;a=o;f=e;while(1){f=f+-1|0;i[d+a>>0]=i[j>>0]|0;if(!f)break;else{j=j+1|0;a=a+1|0}}p=e+o|0}i[d+p>>0]=0;m=d;n=b}while(1){b=i[m>>0]|0;if(!(b<<24>>24)){q=1;s=18;break}d=i[n>>0]|0;if(b<<24>>24==48&d<<24>>24==49){q=0;s=18;break}if(b<<24>>24==49&d<<24>>24==48){q=0;s=18;break}else{m=m+1|0;n=n+1|0}}if((s|0)==18){r=c;return q|0}return 0}function av(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0;d=r;r=r+256|0;e=d;f=xx(a)|0;g=xx(c)|0;if((f|0)>=(g|0))if((g|0)<(f|0)){if((f|0)==(g|0))h=0;else{j=f-g|0;sy(e|0,48,j|0)|0;h=j}if(!g)k=h;else{j=c;l=h;m=g;while(1){m=m+-1|0;i[e+l>>0]=i[j>>0]|0;if(!m)break;else{j=j+1|0;l=l+1|0}}k=g+h|0}i[e+k>>0]=0;n=a;o=e;p=f}else{n=a;o=c;p=f}else{if((g|0)==(f|0))q=0;else{k=g-f|0;sy(e|0,48,k|0)|0;q=k}if(!f)s=q;else{k=a;a=q;h=f;while(1){h=h+-1|0;i[e+a>>0]=i[k>>0]|0;if(!h)break;else{k=k+1|0;a=a+1|0}}s=f+q|0}i[e+s>>0]=0;n=e;o=c;p=g}a:do if(!p){t=0;u=0}else{g=p;c=1;e=0;s=0;while(1){q=g;b:while(1){q=q+-1|0;f=i[n+q>>0]|0;a=i[o+q>>0]|0;switch(f<<24>>24){case 88:case 120:break;default:switch(a<<24>>24){case 88:case 120:break;default:{v=q;w=f;x=a;break b}}}if(!q){t=e;u=s;break a}}q=(w<<24>>24==49?c:0)+e|0;a=(x<<24>>24==49?c:0)+s|0;if(!v){t=q;u=a;break}else{g=v;c=c<<1;e=q;s=a}}}while(0);switch(b|0){case 281:{y=(t|0)<(u|0)&1;r=d;return y|0}case 280:{y=(t|0)<=(u|0)&1;r=d;return y|0}case 282:{y=(t|0)>(u|0)&1;r=d;return y|0}case 279:{y=(t|0)>=(u|0)&1;r=d;return y|0}default:{y=0;r=d;return y|0}}return 0}function bv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;d=r;r=r+256|0;e=d;f=xx(b)|0;g=xx(c)|0;if((f|0)>=(g|0))if((g|0)<(f|0)){if((f|0)==(g|0))h=0;else{j=f-g|0;sy(e|0,48,j|0)|0;h=j}if(!g)k=h;else{j=c;l=h;m=g;while(1){m=m+-1|0;i[e+l>>0]=i[j>>0]|0;if(!m)break;else{j=j+1|0;l=l+1|0}}k=g+h|0}i[e+k>>0]=0;n=b;o=e}else{n=b;o=c}else{if((g|0)==(f|0))p=0;else{k=g-f|0;sy(e|0,48,k|0)|0;p=k}if(!f)q=p;else{k=b;b=p;g=f;while(1){g=g+-1|0;i[e+b>>0]=i[k>>0]|0;if(!g)break;else{k=k+1|0;b=b+1|0}}q=f+p|0}i[e+q>>0]=0;n=e;o=c}c=i[n>>0]|0;if(!(c<<24>>24)){s=a;i[s>>0]=0;r=d;return}else{t=a;u=n;v=o;w=c}while(1){u=u+1|0;do if(w<<24>>24!=49?(c=i[v>>0]|0,c<<24>>24!=49):0)if(w<<24>>24==48|c<<24>>24==48){i[t>>0]=48;break}else{i[t>>0]=120;break}else x=18;while(0);v=v+1|0;if((x|0)==18){x=0;i[t>>0]=49}c=t+1|0;w=i[u>>0]|0;if(!(w<<24>>24)){s=c;break}else t=c}i[s>>0]=0;r=d;return}function cv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0;d=r;r=r+256|0;e=d;f=xx(b)|0;g=xx(c)|0;if((f|0)>=(g|0))if((g|0)<(f|0)){if((f|0)==(g|0))h=0;else{j=f-g|0;sy(e|0,48,j|0)|0;h=j}if(!g)k=h;else{j=c;l=h;m=g;while(1){m=m+-1|0;i[e+l>>0]=i[j>>0]|0;if(!m)break;else{j=j+1|0;l=l+1|0}}k=g+h|0}i[e+k>>0]=0;n=b;o=e}else{n=b;o=c}else{if((g|0)==(f|0))p=0;else{k=g-f|0;sy(e|0,48,k|0)|0;p=k}if(!f)q=p;else{k=b;b=p;g=f;while(1){g=g+-1|0;i[e+b>>0]=i[k>>0]|0;if(!g)break;else{k=k+1|0;b=b+1|0}}q=f+p|0}i[e+q>>0]=0;n=e;o=c}c=i[n>>0]|0;if(!(c<<24>>24)){s=a;i[s>>0]=0;r=d;return}else{t=a;u=n;v=o;w=c}while(1){u=u+1|0;do if(w<<24>>24!=120?(c=i[v>>0]|0,c<<24>>24!=120):0)if(w<<24>>24==49&c<<24>>24==49){i[t>>0]=49;break}else{i[t>>0]=48;break}else x=18;while(0);v=v+1|0;if((x|0)==18){x=0;i[t>>0]=120}c=t+1|0;w=i[u>>0]|0;if(!(w<<24>>24)){s=c;break}else t=c}i[s>>0]=0;r=d;return}function dv(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,q=0,r=0,s=0,t=0,u=0,v=0;b=k[30820]|0;c=k[a+12>>2]|0;d=b+(c*344|0)|0;a:do if((k[d>>2]|0)!=-1e3){Su(a);if(!(k[30835]|0)){e=k[a+52>>2]|0;if((e|0)!=262){f=k[30825]|0;if((e|0)==261)g=f;else g=ha(k[a+56>>2]|0,f)|0;if(g){f=b+(c*344|0)+84|0;e=a+84|0;h=g;do{h=h+-1|0;i[(k[e>>2]|0)+h>>0]=i[(k[f>>2]|0)+h>>0]|0}while((h|0)!=0)}}h=k[30825]|0;f=ha(k[a+56>>2]|0,h)|0;switch(k[a>>2]|0){case 258:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{if(!f)break a;e=b+(c*344|0)+88|0;j=a+88|0;l=f;do{l=l+-1|0;i[(k[j>>2]|0)+l>>0]=+p[(k[e>>2]|0)+(l<<3)>>3]!=0.0&1}while((l|0)!=0);break}case 259:{if(!f)break a;l=b+(c*344|0)+88|0;e=a+88|0;j=f;do{j=j+-1|0;i[(k[e>>2]|0)+j>>0]=(k[(k[l>>2]|0)+(j<<2)>>2]|0)!=0&1}while((j|0)!=0);break}default:break a}break}case 285:case 260:{switch(k[b+(c*344|0)+52>>2]|0){case 259:{if(!f)break a;j=k[b+(c*344|0)+88>>2]|0;l=k[a+88>>2]|0;e=f;do{e=e+-1|0;p[l+(e<<3)>>3]=+(k[j+(e<<2)>>2]|0)}while((e|0)!=0);break}case 258:{if(!f)break a;e=k[b+(c*344|0)+88>>2]|0;j=k[a+88>>2]|0;l=f;do{l=l+-1|0;p[j+(l<<3)>>3]=(i[e+l>>0]|0)!=0?1.0:0.0}while((l|0)!=0);break}default:break a}break}case 286:case 259:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{if(!f)break a;l=k[b+(c*344|0)+88>>2]|0;e=k[a+88>>2]|0;j=f;do{j=j+-1|0;k[e+(j<<2)>>2]=~~+p[l+(j<<3)>>3]}while((j|0)!=0);break}case 258:{if(!f)break a;j=k[b+(c*344|0)+88>>2]|0;l=k[a+88>>2]|0;e=f;do{e=e+-1|0;k[l+(e<<2)>>2]=(i[j+e>>0]|0)!=0&1}while((e|0)!=0);break}default:break a}break}case 287:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{if(!f)break a;e=k[b+(c*344|0)+88>>2]|0;j=k[a+88>>2]|0;l=f;do{l=l+-1|0;p[j+(l<<3)>>3]=-+p[e+(l<<3)>>3]}while((l|0)!=0);break}case 259:{if(!f)break a;l=k[b+(c*344|0)+88>>2]|0;e=k[a+88>>2]|0;j=f;do{j=j+-1|0;k[e+(j<<2)>>2]=0-(k[l+(j<<2)>>2]|0)}while((j|0)!=0);break}default:break a}break}case 284:{switch(k[b+(c*344|0)+52>>2]|0){case 258:{if(!f)break a;j=b+(c*344|0)+88|0;l=a+88|0;e=f;while(1){e=e+-1|0;i[(k[l>>2]|0)+e>>0]=(i[(k[j>>2]|0)+e>>0]|0)==0&1;if(!e)break a}break}case 262:break;default:break a}if(!h)break a;e=a+88|0;j=b+(c*344|0)+88|0;l=h;do{l=l+-1|0;f=k[(k[e>>2]|0)+(l<<2)>>2]|0;m=k[(k[j>>2]|0)+(l<<2)>>2]|0;n=xx(m)|0;if(!n)o=f;else{q=m;m=f;r=n;while(1){r=r+-1|0;s=i[q>>0]|0;if(s<<24>>24==49)t=48;else t=s<<24>>24==48?49:s<<24>>24;i[m>>0]=t;if(!r)break;else{q=q+1|0;m=m+1|0}}o=f+n|0}i[o>>0]=0}while((l|0)!=0);break}default:break a}}}else{b:do switch(k[a>>2]|0){case 285:case 260:{switch(k[b+(c*344|0)+52>>2]|0){case 259:{p[a+88>>3]=+(k[b+(c*344|0)+88>>2]|0);break b;break}case 258:{p[a+88>>3]=(i[b+(c*344|0)+88>>0]|0)!=0?1.0:0.0;break b;break}default:break b}break}case 286:case 259:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{k[a+88>>2]=~~+p[b+(c*344|0)+88>>3];break b;break}case 258:{k[a+88>>2]=(i[b+(c*344|0)+88>>0]|0)!=0&1;break b;break}default:break b}break}case 258:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{i[a+88>>0]=+p[b+(c*344|0)+88>>3]!=0.0&1;break b;break}case 259:{i[a+88>>0]=(k[b+(c*344|0)+88>>2]|0)!=0&1;break b;break}default:break b}break}case 287:{switch(k[b+(c*344|0)+52>>2]|0){case 260:{p[a+88>>3]=-+p[b+(c*344|0)+88>>3];break b;break}case 259:{k[a+88>>2]=0-(k[b+(c*344|0)+88>>2]|0);break b;break}default:break b}break}case 284:{switch(k[b+(c*344|0)+52>>2]|0){case 258:{i[a+88>>0]=(i[b+(c*344|0)+88>>0]|0)==0&1;break b;break}case 262:break;default:break b}l=a+88|0;j=b+(c*344|0)+88|0;e=xx(j)|0;if(!e)u=l;else{h=j;j=l;m=e;while(1){m=m+-1|0;q=i[h>>0]|0;if(q<<24>>24==49)v=48;else v=q<<24>>24==48?49:q<<24>>24;i[j>>0]=v;if(!m)break;else{h=h+1|0;j=j+1|0}}u=l+e|0}i[u>>0]=0;break}default:{}}while(0);k[a>>2]=-1e3}while(0);if((k[d>>2]|0)<=0)return;dy(k[b+(c*344|0)+88>>2]|0);return}function ev(a){a=a|0;fw(a);return}function fv(){return +(+(gw()|0)*4.656612873077393e-10)}function gv(){var a=0.0,b=0.0,c=0.0,d=0.0;if(!(k[31160]|0)){a=+(gw()|0)*4.656612873077393e-10;b=+X(+(+fa(+a)*-2.0));a=+(gw()|0)*4.656612873077393e-10*6.283185307179586;c=b*+Z(+a);p[12421]=c;k[31160]=1;d=b*+_(+a);return +d}else{k[31160]=0;d=+p[12421];return +d}return +(0.0)}function hv(a){a=+a;var b=0.0,c=0,d=0;b=a<0.0?0.0:a;if(!(b<15.0)){c=iv(b)|0;return c|0}a=+ea(+-b);d=0;b=1.0;while(1){b=b*(+(gw()|0)*4.656612873077393e-10);if(!(b>a)){c=d;break}else d=d+1|0}return c|0}function iv(a){a=+a;var b=0.0,c=0.0,d=0,e=0.0,f=0.0,g=0.0,h=0;if(+p[12422]!=a){b=3.141592653589793/+X(+(a*3.0));p[12423]=b;p[12424]=b*a;c=+fa(+(.767-3.36/a))-a-+fa(+b);p[12425]=c;p[12422]=a}c=+fa(+a);while(1){a=+(gw()|0)*4.656612873077393e-10;b=+p[12424]-+fa(+((1.0-a)/a));a=b/+p[12423];d=~~+V(+(a+.5));if((d|0)<0)continue;b=+(gw()|0)*4.656612873077393e-10;e=+p[12424]-a*+p[12423];a=+ea(+e)+1.0;f=e+ +fa(+(b/(a*a)));if((d|0)>254){a=+(d+1|0);g=1.0/(a*12.0)+((a+-.5)*+fa(+a)-a+.9189385332046727)}else g=+p[99408+(d<<3)>>3];if(f<=+p[12425]+ +(d|0)*c-g){h=d;break}}return h|0}function jv(a){a=a|0;return (a+-97|0)>>>0<26|0}function kv(a){a=a|0;return ((a|0)==32|(a+-9|0)>>>0<5)&1|0}function lv(a){a=a|0;return (a+-65|0)>>>0<26|0}function mv(a){a=a|0;var b=0;b=(lv(a)|0)==0;return (b?a:a|32)|0}function nv(a){a=a|0;var b=0;b=(jv(a)|0)==0;return (b?a:a&95)|0}function ov(a){a=a|0;var b=0;b=Hx(k[a>>2]|0)|0;dy(a);return b|0}function pv(a){a=a|0;var b=0,c=0,d=0,e=0;b=r;r=r+16|0;c=b+8|0;d=uv(a,589824,b)|0;do if((d|0)>=0){a=ey(1,2072)|0;if(!a){k[c>>2]=d;cb(6,c|0)|0;e=0;break}else{k[a>>2]=d;e=a;break}}else e=0;while(0);r=b;return e|0}function qv(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;b=r;r=r+16|0;c=b;d=a+8|0;e=k[d>>2]|0;f=a+12|0;do if((e|0)>=(k[f>>2]|0)){k[c>>2]=k[a>>2];k[c+4>>2]=a+24;k[c+8>>2]=2048;g=Bb(220,c|0)|0;if((g|0)>=1){k[f>>2]=g;k[d>>2]=0;h=0;i=6;break}if((g|0)<0&(g|0)!=-2){j=rv()|0;k[j>>2]=0-g;l=0}else l=0}else{h=e;i=6}while(0);if((i|0)==6){k[d>>2]=(m[h+8+(a+24)>>1]|0)+h;k[a+4>>2]=k[h+4+(a+24)>>2];l=a+24+h|0}r=b;return l|0}function rv(){var a=0,b=0;if(!(k[31161]|0))a=124932;else{b=(hb()|0)+60|0;a=k[b>>2]|0}return a|0}function sv(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0;b=0;while(1){if((l[790763+b>>0]|0)==(a|0)){c=b;d=2;break}b=b+1|0;if((b|0)==87){e=87;f=790851;d=5;break}}if((d|0)==2)if(!c)g=790851;else{e=c;f=790851;d=5}if((d|0)==5)while(1){d=0;c=f;while(1){b=c+1|0;if(!(i[c>>0]|0)){h=b;break}else c=b}e=e+-1|0;if(!e){g=h;break}else{f=h;d=5}}return g|0}function tv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0;d=r;r=r+192|0;e=d+152|0;f=d+136|0;g=d+120|0;h=d+104|0;i=d+96|0;j=d+80|0;l=d+64|0;m=d+48|0;n=d+32|0;o=d+16|0;p=d;q=d+176|0;s=d+168|0;k[q>>2]=c;c=(k[q>>2]|0)+(4-1)&~(4-1);t=k[c>>2]|0;k[q>>2]=c+4;c=(b|0)==4?t|32768:t;a:do switch(b|0){case 14:{k[p>>2]=a;k[p+4>>2]=14;k[p+8>>2]=c;u=zv(eb(221,p|0)|0)|0;break}case 9:{k[o>>2]=a;k[o+4>>2]=16;k[o+8>>2]=s;t=eb(221,o|0)|0;switch(t|0){case -22:{k[n>>2]=a;k[n+4>>2]=9;k[n+8>>2]=c;u=eb(221,n|0)|0;break a;break}case 0:{q=k[s+4>>2]|0;u=(k[s>>2]|0)==2?0-q|0:q;break a;break}default:{u=zv(t)|0;break a}}break}case 1030:{k[m>>2]=a;k[m+4>>2]=1030;k[m+8>>2]=c;t=eb(221,m|0)|0;if((t|0)!=-22){if((t|0)>-1){k[l>>2]=t;k[l+4>>2]=2;k[l+8>>2]=1;eb(221,l|0)|0}u=zv(t)|0;break a}k[j>>2]=a;k[j+4>>2]=1030;k[j+8>>2]=0;t=eb(221,j|0)|0;if((t|0)==-22){k[h>>2]=a;k[h+4>>2]=0;k[h+8>>2]=c;q=eb(221,h|0)|0;if((q|0)>-1){k[g>>2]=q;k[g+4>>2]=2;k[g+8>>2]=1;eb(221,g|0)|0}u=zv(q)|0;break a}else{if((t|0)>-1){k[i>>2]=t;cb(6,i|0)|0}u=zv(-22)|0;break a}break}default:if((b+-12|0)>>>0<5){k[f>>2]=a;k[f+4>>2]=b;k[f+8>>2]=c;u=zv(eb(221,f|0)|0)|0;break a}else{k[e>>2]=a;k[e+4>>2]=b;k[e+8>>2]=c;u=zv(eb(221,e|0)|0)|0;break a}}while(0);r=d;return u|0}function uv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0;d=r;r=r+32|0;e=d;f=d+16|0;k[f>>2]=c;c=(k[f>>2]|0)+(4-1)&~(4-1);g=k[c>>2]|0;k[f>>2]=c+4;k[e>>2]=a;k[e+4>>2]=b|32768;k[e+8>>2]=g;g=zv(yb(5,e|0)|0)|0;r=d;return g|0}function vv(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0.0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,X=0,Y=0,Z=0.0,_=0,$=0.0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ia=0,ja=0,ka=0,la=0.0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0.0,ua=0,va=0.0,wa=0.0,xa=0,ya=0.0,za=0,Aa=0.0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0.0,La=0,Ma=0,Na=0,Oa=0,Pa=0.0,Qa=0,Ra=0,Sa=0,Ta=0.0,Ua=0.0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0,yb=0,zb=0,Ab=0,Bb=0,Cb=0,Db=0,Eb=0,Fb=0,Gb=0,Hb=0,Ib=0,Jb=0,Kb=0,Lb=0,Mb=0,Nb=0,Ob=0,Pb=0,Qb=0,Rb=0,Sb=0,Tb=0,Ub=0,Vb=0,Wb=0,Xb=0,Yb=0,Zb=0,_b=0,$b=0,ac=0,bc=0,cc=0,dc=0,ec=0,fc=0.0,gc=0,hc=0,ic=0,jc=0,kc=0,lc=0,mc=0,nc=0,oc=0,pc=0,qc=0,rc=0,sc=0,tc=0,uc=0,vc=0,wc=0,xc=0,yc=0,zc=0,Ac=0,Bc=0,Cc=0,Dc=0,Ec=0,Fc=0,Gc=0,Hc=0,Ic=0,Jc=0,Kc=0,Lc=0,Mc=0,Nc=0,Oc=0,Pc=0,Qc=0,Rc=0,Sc=0,Tc=0,Uc=0,Vc=0,Wc=0,Xc=0,Yc=0,Zc=0,_c=0,$c=0,ad=0,bd=0,cd=0,dd=0,ed=0.0,fd=0.0,gd=0.0,hd=0.0,id=0.0,jd=0.0,kd=0.0,ld=0,md=0,nd=0.0,od=0,pd=0.0;d=r;r=r+512|0;e=d;switch(b|0){case 0:{f=24;g=-149;h=4;break}case 1:{f=53;g=-1074;h=4;break}case 2:{f=53;g=-1074;h=4;break}default:j=0.0}a:do if((h|0)==4){b=a+4|0;m=a+100|0;do{n=k[b>>2]|0;if(n>>>0<(k[m>>2]|0)>>>0){k[b>>2]=n+1;o=l[n>>0]|0}else o=yv(a)|0}while((kv(o)|0)!=0);p=o;b:do switch(p|0){case 43:case 45:{n=1-(((p|0)==45&1)<<1)|0;q=k[b>>2]|0;if(q>>>0<(k[m>>2]|0)>>>0){k[b>>2]=q+1;s=l[q>>0]|0;t=n;break b}else{s=yv(a)|0;t=n;break b}break}default:{s=p;t=1}}while(0);n=s;q=0;while(1){if((n|32|0)!=(i[792655+q>>0]|0)){u=n;v=q;break}do if(q>>>0<7){w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;x=l[w>>0]|0;break}else{x=yv(a)|0;break}}else x=n;while(0);w=q+1|0;if(w>>>0<8){n=x;q=w}else{u=x;v=w;break}}c:do switch(v|0){case 8:break;case 3:{h=23;break}default:{q=(c|0)!=0;if(q&v>>>0>3)if((v|0)==8)break c;else{h=23;break c}d:do if(!v){n=u;w=0;while(1){if((n|32|0)!=(i[795583+w>>0]|0)){y=n;z=w;break d}do if(w>>>0<2){C=k[b>>2]|0;if(C>>>0<(k[m>>2]|0)>>>0){k[b>>2]=C+1;D=l[C>>0]|0;break}else{D=yv(a)|0;break}}else D=n;while(0);C=w+1|0;if(C>>>0<3){n=D;w=C}else{y=D;z=C;break}}}else{y=u;z=v}while(0);switch(z|0){case 3:{w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;E=l[w>>0]|0}else E=yv(a)|0;if((E|0)==40)F=1;else{if(!(k[m>>2]|0)){j=A;break a}k[b>>2]=(k[b>>2]|0)+-1;j=A;break a}while(1){w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;G=l[w>>0]|0}else G=yv(a)|0;if(!((G+-48|0)>>>0<10|(G+-65|0)>>>0<26)?!((G|0)==95|(G+-97|0)>>>0<26):0){H=G;I=F;break}F=F+1|0}if((H|0)==41){j=A;break a}w=(k[m>>2]|0)==0;if(!w)k[b>>2]=(k[b>>2]|0)+-1;if(!q){n=rv()|0;k[n>>2]=22;xv(a,0);j=0.0;break a}if(!I){j=A;break a}else J=I;while(1){J=J+-1|0;if(!w)k[b>>2]=(k[b>>2]|0)+-1;if(!J){j=A;break a}}break}case 0:{do if((y|0)==48){w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;K=l[w>>0]|0}else K=yv(a)|0;if((K|32|0)!=120){if(!(k[m>>2]|0)){M=48;break}k[b>>2]=(k[b>>2]|0)+-1;M=48;break}w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;N=l[w>>0]|0;O=0}else{N=yv(a)|0;O=0}e:while(1){switch(N|0){case 46:{P=O;h=74;break e;break}case 48:break;default:{Q=0;R=0;S=0;T=0;U=N;V=O;X=0;Y=0;Z=1.0;_=0;$=0.0;break e}}w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;N=l[w>>0]|0;O=1;continue}else{N=yv(a)|0;O=1;continue}}if((h|0)==74){w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;aa=l[w>>0]|0}else aa=yv(a)|0;if((aa|0)==48){w=0;q=0;while(1){n=k[b>>2]|0;if(n>>>0<(k[m>>2]|0)>>>0){k[b>>2]=n+1;ba=l[n>>0]|0}else ba=yv(a)|0;n=ky(w|0,q|0,-1,-1)|0;C=L;if((ba|0)==48){w=n;q=C}else{Q=0;R=0;S=n;T=C;U=ba;V=1;X=1;Y=0;Z=1.0;_=0;$=0.0;break}}}else{Q=0;R=0;S=0;T=0;U=aa;V=P;X=1;Y=0;Z=1.0;_=0;$=0.0}}while(1){q=U+-48|0;w=U|32;if(q>>>0>=10){C=(U|0)==46;if(!(C|(w+-97|0)>>>0<6)){ca=R;da=S;ea=Q;fa=T;ga=U;ia=V;ja=X;ka=_;la=$;break}if(C)if(!X){ma=R;na=Q;oa=R;pa=Q;qa=V;ra=1;sa=Y;ta=Z;ua=_;va=$}else{ca=R;da=S;ea=Q;fa=T;ga=46;ia=V;ja=X;ka=_;la=$;break}else h=86}else h=86;if((h|0)==86){h=0;C=(U|0)>57?w+-87|0:q;do if(!((Q|0)<0|(Q|0)==0&R>>>0<8)){if((Q|0)<0|(Q|0)==0&R>>>0<14){wa=Z*.0625;xa=Y;ya=wa;za=_;Aa=$+wa*+(C|0);break}if((Y|0)!=0|(C|0)==0){xa=Y;ya=Z;za=_;Aa=$}else{xa=1;ya=Z;za=_;Aa=$+Z*.5}}else{xa=Y;ya=Z;za=C+(_<<4)|0;Aa=$}while(0);C=ky(R|0,Q|0,1,0)|0;ma=S;na=T;oa=C;pa=L;qa=1;ra=X;sa=xa;ta=ya;ua=za;va=Aa}C=k[b>>2]|0;if(C>>>0<(k[m>>2]|0)>>>0){k[b>>2]=C+1;Q=pa;R=oa;S=ma;T=na;U=l[C>>0]|0;V=qa;X=ra;Y=sa;Z=ta;_=ua;$=va;continue}else{Q=pa;R=oa;S=ma;T=na;U=yv(a)|0;V=qa;X=ra;Y=sa;Z=ta;_=ua;$=va;continue}}if(!ia){C=(k[m>>2]|0)==0;if(!C)k[b>>2]=(k[b>>2]|0)+-1;if(c){if(!C?(C=k[b>>2]|0,k[b>>2]=C+-1,(ja|0)!=0):0)k[b>>2]=C+-2}else xv(a,0);j=+(t|0)*0.0;break a}C=(ja|0)==0;q=C?ca:da;w=C?ea:fa;if((ea|0)<0|(ea|0)==0&ca>>>0<8){C=ca;n=ea;Ba=ka;while(1){Ca=Ba<<4;C=ky(C|0,n|0,1,0)|0;n=L;if(!((n|0)<0|(n|0)==0&C>>>0<8)){Da=Ca;break}else Ba=Ca}}else Da=ka;if((ga|32|0)==112){Ba=Px(a,c)|0;C=L;if((Ba|0)==0&(C|0)==-2147483648){if(!c){xv(a,0);j=0.0;break a}if(!(k[m>>2]|0)){Ea=0;Fa=0}else{k[b>>2]=(k[b>>2]|0)+-1;Ea=0;Fa=0}}else{Ea=Ba;Fa=C}}else if(!(k[m>>2]|0)){Ea=0;Fa=0}else{k[b>>2]=(k[b>>2]|0)+-1;Ea=0;Fa=0}C=ty(q|0,w|0,2)|0;Ba=ky(C|0,L|0,-32,-1)|0;C=ky(Ba|0,L|0,Ea|0,Fa|0)|0;Ba=L;if(!Da){j=+(t|0)*0.0;break a}if((Ba|0)>0|(Ba|0)==0&C>>>0>(0-g|0)>>>0){n=rv()|0;k[n>>2]=34;j=+(t|0)*1797693134862315708145274.0e284*1797693134862315708145274.0e284;break a}n=g+-106|0;Ca=((n|0)<0)<<31>>31;if((Ba|0)<(Ca|0)|(Ba|0)==(Ca|0)&C>>>0<n>>>0){n=rv()|0;k[n>>2]=34;j=+(t|0)*2.2250738585072014e-308*2.2250738585072014e-308;break a}if((Da|0)>-1){n=C;Ca=Ba;Ga=Da;wa=la;while(1){Ha=!(wa>=.5);Ia=Ha&1|Ga<<1;Ja=Ia^1;Ka=wa+(Ha?wa:wa+-1.0);Ha=ky(n|0,Ca|0,-1,-1)|0;La=L;if((Ia|0)>-1){n=Ha;Ca=La;Ga=Ja;wa=Ka}else{Ma=Ha;Na=La;Oa=Ja;Pa=Ka;break}}}else{Ma=C;Na=Ba;Oa=Da;Pa=la}Ga=jy(32,0,g|0,((g|0)<0)<<31>>31|0)|0;Ca=ky(Ma|0,Na|0,Ga|0,L|0)|0;Ga=L;if(0>(Ga|0)|0==(Ga|0)&f>>>0>Ca>>>0)if((Ca|0)<0){Qa=0;h=127}else{Ra=Ca;h=125}else{Ra=f;h=125}if((h|0)==125)if((Ra|0)<53){Qa=Ra;h=127}else{Sa=Ra;Ta=+(t|0);Ua=0.0}if((h|0)==127){wa=+(t|0);Sa=Qa;Ta=wa;Ua=+Dv(+Lv(1.0,84-Qa|0),wa)}Ca=(Oa&1|0)==0&(Pa!=0.0&(Sa|0)<32);wa=Ta*(Ca?0.0:Pa)+(Ua+Ta*+(((Ca&1)+Oa|0)>>>0))-Ua;if(!(wa!=0.0)){Ca=rv()|0;k[Ca>>2]=34}j=+Mv(wa,Ma);break a}else M=y;while(0);Ca=g+f|0;Ga=0-Ca|0;n=M;w=0;f:while(1){switch(n|0){case 46:{Va=w;h=138;break f;break}case 48:break;default:{Wa=n;Xa=0;Ya=0;Za=w;_a=0;break f}}q=k[b>>2]|0;if(q>>>0<(k[m>>2]|0)>>>0){k[b>>2]=q+1;n=l[q>>0]|0;w=1;continue}else{n=yv(a)|0;w=1;continue}}if((h|0)==138){w=k[b>>2]|0;if(w>>>0<(k[m>>2]|0)>>>0){k[b>>2]=w+1;$a=l[w>>0]|0}else $a=yv(a)|0;if(($a|0)==48){w=0;n=0;while(1){q=ky(w|0,n|0,-1,-1)|0;Ja=L;La=k[b>>2]|0;if(La>>>0<(k[m>>2]|0)>>>0){k[b>>2]=La+1;ab=l[La>>0]|0}else ab=yv(a)|0;if((ab|0)==48){w=q;n=Ja}else{Wa=ab;Xa=q;Ya=Ja;Za=1;_a=1;break}}}else{Wa=$a;Xa=0;Ya=0;Za=Va;_a=1}}k[e>>2]=0;n=Wa+-48|0;w=(Wa|0)==46;g:do if(w|n>>>0<10){Ja=e+496|0;q=Wa;La=0;Ha=0;Ia=w;bb=n;cb=Xa;db=Ya;eb=Za;fb=_a;gb=0;hb=0;ib=0;h:while(1){do if(Ia)if(!fb){jb=La;kb=Ha;lb=La;mb=Ha;nb=eb;ob=1;pb=gb;qb=hb;rb=ib}else{sb=cb;tb=db;ub=La;vb=Ha;wb=eb;xb=gb;yb=hb;zb=ib;break h}else{Ab=ky(La|0,Ha|0,1,0)|0;Bb=L;Cb=(q|0)!=48;if((hb|0)>=125){if(!Cb){jb=cb;kb=db;lb=Ab;mb=Bb;nb=eb;ob=fb;pb=gb;qb=hb;rb=ib;break}k[Ja>>2]=k[Ja>>2]|1;jb=cb;kb=db;lb=Ab;mb=Bb;nb=eb;ob=fb;pb=gb;qb=hb;rb=ib;break}Db=e+(hb<<2)|0;if(!gb)Eb=bb;else Eb=q+-48+((k[Db>>2]|0)*10|0)|0;k[Db>>2]=Eb;Db=gb+1|0;Fb=(Db|0)==9;jb=cb;kb=db;lb=Ab;mb=Bb;nb=1;ob=fb;pb=Fb?0:Db;qb=(Fb&1)+hb|0;rb=Cb?Ab:ib}while(0);Ab=k[b>>2]|0;if(Ab>>>0<(k[m>>2]|0)>>>0){k[b>>2]=Ab+1;Gb=l[Ab>>0]|0}else Gb=yv(a)|0;bb=Gb+-48|0;Ia=(Gb|0)==46;if(!(Ia|bb>>>0<10)){Hb=Gb;Ib=lb;Jb=jb;Kb=mb;Lb=kb;Mb=nb;Nb=ob;Ob=pb;Pb=qb;Qb=rb;h=161;break g}else{q=Gb;La=lb;Ha=mb;cb=jb;db=kb;eb=nb;fb=ob;gb=pb;hb=qb;ib=rb}}Rb=ub;Sb=vb;Tb=sb;Ub=tb;Vb=(wb|0)!=0;Wb=xb;Xb=yb;Yb=zb;h=169}else{Hb=Wa;Ib=0;Jb=Xa;Kb=0;Lb=Ya;Mb=Za;Nb=_a;Ob=0;Pb=0;Qb=0;h=161}while(0);do if((h|0)==161){n=(Nb|0)==0;w=n?Ib:Jb;ib=n?Kb:Lb;n=(Mb|0)!=0;if(!((Hb|32|0)==101&n))if((Hb|0)>-1){Rb=Ib;Sb=Kb;Tb=w;Ub=ib;Vb=n;Wb=Ob;Xb=Pb;Yb=Qb;h=169;break}else{Zb=Ib;_b=Kb;$b=n;ac=w;bc=ib;cc=Ob;dc=Pb;ec=Qb;h=171;break}n=Px(a,c)|0;hb=L;if((n|0)==0&(hb|0)==-2147483648){if(!c){xv(a,0);fc=0.0;break}if(!(k[m>>2]|0)){gc=0;hc=0}else{k[b>>2]=(k[b>>2]|0)+-1;gc=0;hc=0}}else{gc=n;hc=hb}hb=ky(gc|0,hc|0,w|0,ib|0)|0;ic=hb;jc=Ib;kc=L;lc=Kb;mc=Ob;nc=Pb;oc=Qb;h=173}while(0);if((h|0)==169)if(k[m>>2]|0){k[b>>2]=(k[b>>2]|0)+-1;if(Vb){ic=Tb;jc=Rb;kc=Ub;lc=Sb;mc=Wb;nc=Xb;oc=Yb;h=173}else h=172}else{Zb=Rb;_b=Sb;$b=Vb;ac=Tb;bc=Ub;cc=Wb;dc=Xb;ec=Yb;h=171}if((h|0)==171)if($b){ic=ac;jc=Zb;kc=bc;lc=_b;mc=cc;nc=dc;oc=ec;h=173}else h=172;do if((h|0)==172){hb=rv()|0;k[hb>>2]=22;xv(a,0);fc=0.0}else if((h|0)==173){hb=k[e>>2]|0;if(!hb){fc=+(t|0)*0.0;break}if(((lc|0)<0|(lc|0)==0&jc>>>0<10)&((ic|0)==(jc|0)&(kc|0)==(lc|0))?f>>>0>30|(hb>>>f|0)==0:0){fc=+(t|0)*+(hb>>>0);break}hb=(g|0)/-2|0;ib=((hb|0)<0)<<31>>31;if((kc|0)>(ib|0)|(kc|0)==(ib|0)&ic>>>0>hb>>>0){hb=rv()|0;k[hb>>2]=34;fc=+(t|0)*1797693134862315708145274.0e284*1797693134862315708145274.0e284;break}hb=g+-106|0;ib=((hb|0)<0)<<31>>31;if((kc|0)<(ib|0)|(kc|0)==(ib|0)&ic>>>0<hb>>>0){hb=rv()|0;k[hb>>2]=34;fc=+(t|0)*2.2250738585072014e-308*2.2250738585072014e-308;break}if(!mc)pc=nc;else{if((mc|0)<9){hb=e+(nc<<2)|0;ib=k[hb>>2]|0;w=mc;while(1){n=ib*10|0;w=w+1|0;if((w|0)==9){qc=n;break}else ib=n}k[hb>>2]=qc}pc=nc+1|0}if((oc|0)<9?(oc|0)<=(ic|0)&(ic|0)<18:0){if((ic|0)==9){fc=+(t|0)*+((k[e>>2]|0)>>>0);break}if((ic|0)<9){fc=+(t|0)*+((k[e>>2]|0)>>>0)/+(k[124936+(8-ic<<2)>>2]|0);break}ib=f+27+(ha(ic,-3)|0)|0;w=k[e>>2]|0;if((ib|0)>30|(w>>>ib|0)==0){fc=+(t|0)*+(w>>>0)*+(k[124936+(ic+-10<<2)>>2]|0);break}}w=(ic|0)%9|0;if(!w){rc=0;sc=0;tc=ic;uc=pc}else{ib=(ic|0)>-1?w:w+9|0;w=k[124936+(8-ib<<2)>>2]|0;if(pc){n=1e9/(w|0)|0;gb=0;fb=0;eb=0;db=ic;while(1){cb=e+(eb<<2)|0;Ha=k[cb>>2]|0;La=((Ha>>>0)/(w>>>0)|0)+fb|0;k[cb>>2]=La;cb=ha((Ha>>>0)%(w>>>0)|0,n)|0;Ha=(eb|0)==(gb|0)&(La|0)==0;eb=eb+1|0;La=Ha?db+-9|0:db;q=Ha?eb&127:gb;if((eb|0)==(pc|0)){vc=cb;wc=q;xc=La;break}else{gb=q;fb=cb;db=La}}if(!vc){yc=wc;zc=xc;Ac=pc}else{k[e+(pc<<2)>>2]=vc;yc=wc;zc=xc;Ac=pc+1|0}}else{yc=0;zc=ic;Ac=0}rc=yc;sc=0;tc=9-ib+zc|0;uc=Ac}i:while(1){db=(tc|0)<18;fb=(tc|0)==18;gb=e+(rc<<2)|0;eb=sc;n=uc;while(1){if(!db){if(!fb){Bc=rc;Cc=eb;Dc=tc;Ec=n;break i}if((k[gb>>2]|0)>>>0>=9007199){Bc=rc;Cc=eb;Dc=18;Ec=n;break i}}w=0;hb=n+127|0;La=n;while(1){cb=hb&127;q=e+(cb<<2)|0;Ha=ty(k[q>>2]|0,0,29)|0;bb=ky(Ha|0,L|0,w|0,0)|0;Ha=L;if(Ha>>>0>0|(Ha|0)==0&bb>>>0>1e9){Ia=zy(bb|0,Ha|0,1e9,0)|0;Ja=Ay(bb|0,Ha|0,1e9,0)|0;Fc=Ja;Gc=Ia}else{Fc=bb;Gc=0}k[q>>2]=Fc;q=(cb|0)==(rc|0);bb=(cb|0)!=(La+127&127|0)|q?La:(Fc|0)==0?cb:La;if(q){Hc=Gc;Ic=bb;break}else{w=Gc;hb=cb+-1|0;La=bb}}La=eb+-29|0;if(!Hc){eb=La;n=Ic}else{Jc=La;Kc=Hc;Lc=Ic;break}}n=rc+127&127;if((n|0)==(Lc|0)){eb=Lc+127&127;gb=e+((Lc+126&127)<<2)|0;k[gb>>2]=k[gb>>2]|k[e+(eb<<2)>>2];Mc=eb}else Mc=Lc;k[e+(n<<2)>>2]=Kc;rc=n;sc=Jc;tc=tc+9|0;uc=Mc}j:while(1){Nc=Ec+1&127;ib=e+((Ec+127&127)<<2)|0;n=Bc;eb=Cc;gb=Dc;while(1){fb=(gb|0)==18;db=(gb|0)>27?9:1;La=fb^1;Oc=n;Pc=eb;while(1){Qc=Oc&127;Rc=(Qc|0)==(Ec|0);do if(!Rc){hb=k[e+(Qc<<2)>>2]|0;if(hb>>>0<9007199){h=219;break}if(hb>>>0>9007199)break;hb=Oc+1&127;if((hb|0)==(Ec|0)){h=219;break}w=k[e+(hb<<2)>>2]|0;if(w>>>0<254740991){h=219;break}if(!(w>>>0>254740991|La)){Sc=Qc;Tc=Oc;Uc=Pc;Vc=Ec;break j}}else h=219;while(0);if((h|0)==219?(h=0,fb):0){h=220;break j}w=Pc+db|0;if((Oc|0)==(Ec|0)){Oc=Ec;Pc=w}else{Wc=w;Xc=Oc;break}}fb=(1<<db)+-1|0;La=1e9>>>db;w=Xc;hb=0;bb=Xc;cb=gb;while(1){q=e+(bb<<2)|0;Ia=k[q>>2]|0;Ja=(Ia>>>db)+hb|0;k[q>>2]=Ja;q=ha(Ia&fb,La)|0;Ia=(bb|0)==(w|0)&(Ja|0)==0;bb=bb+1&127;Ja=Ia?cb+-9|0:cb;Ha=Ia?bb:w;if((bb|0)==(Ec|0)){Yc=q;Zc=Ha;_c=Ja;break}else{w=Ha;hb=q;cb=Ja}}if(!Yc){n=Zc;eb=Wc;gb=_c;continue}if((Nc|0)!=(Zc|0)){$c=Wc;ad=Yc;bd=Zc;cd=_c;break}k[ib>>2]=k[ib>>2]|1;n=Zc;eb=Wc;gb=_c}k[e+(Ec<<2)>>2]=ad;Bc=bd;Cc=$c;Dc=cd;Ec=Nc}if((h|0)==220)if(Rc){k[e+(Nc+-1<<2)>>2]=0;Sc=Ec;Tc=Oc;Uc=Pc;Vc=Nc}else{Sc=Qc;Tc=Oc;Uc=Pc;Vc=Ec}wa=+((k[e+(Sc<<2)>>2]|0)>>>0);gb=Tc+1&127;if((gb|0)==(Vc|0)){eb=Tc+2&127;k[e+(eb+-1<<2)>>2]=0;dd=eb}else dd=Vc;Ka=+(t|0);ed=Ka*(wa*1.0e9+ +((k[e+(gb<<2)>>2]|0)>>>0));gb=Uc+53|0;eb=gb-g|0;n=(eb|0)<(f|0);ib=n&1;cb=n?((eb|0)<0?0:eb):f;if((cb|0)<53){wa=+Dv(+Lv(1.0,105-cb|0),ed);fd=+Hv(ed,+Lv(1.0,53-cb|0));gd=wa;hd=fd;id=wa+(ed-fd)}else{gd=0.0;hd=0.0;id=ed}hb=Tc+2&127;do if((hb|0)==(dd|0))jd=hd;else{w=k[e+(hb<<2)>>2]|0;do if(w>>>0>=5e8){if(w>>>0>5e8){kd=Ka*.75+hd;break}if((Tc+3&127|0)==(dd|0)){kd=Ka*.5+hd;break}else{kd=Ka*.75+hd;break}}else{if((w|0)==0?(Tc+3&127|0)==(dd|0):0){kd=hd;break}kd=Ka*.25+hd}while(0);if((53-cb|0)<=1){jd=kd;break}if(+Hv(kd,1.0)!=0.0){jd=kd;break}jd=kd+1.0}while(0);Ka=id+jd-gd;do if((gb&2147483647|0)>(-2-Ca|0)){if(!(+W(+Ka)>=9007199254740992.0)){ld=ib;md=Uc;nd=Ka}else{ld=n&(cb|0)==(eb|0)?0:ib;md=Uc+1|0;nd=Ka*.5}if((md+50|0)<=(Ga|0)?!(jd!=0.0&(ld|0)!=0):0){od=md;pd=nd;break}hb=rv()|0;k[hb>>2]=34;od=md;pd=nd}else{od=Uc;pd=Ka}while(0);fc=+Mv(pd,od)}while(0);j=fc;break a;break}default:{if(k[m>>2]|0)k[b>>2]=(k[b>>2]|0)+-1;Ga=rv()|0;k[Ga>>2]=22;xv(a,0);j=0.0;break a}}}}while(0);if((h|0)==23){Ga=(k[m>>2]|0)==0;if(!Ga)k[b>>2]=(k[b>>2]|0)+-1;if((c|0)!=0&v>>>0>3){Ca=v;do{if(!Ga)k[b>>2]=(k[b>>2]|0)+-1;Ca=Ca+-1|0}while(Ca>>>0>3)}}j=+(t|0)*B}while(0);r=d;return +j}
function Gt(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0;f=a+b|0;b=a+3|0;g=l[a+2>>0]|0;h=0;m=(l[a>>0]|0)<<8|(l[a+1>>0]|0);a=8;while(1){if((h|0)>=(d|0)){n=b;break}o=a+-4|0;if((a|0)<4){p=(((o|0)>-8?a+7|0:3)-a|0)>>>3;q=p<<3;r=b;s=g;t=o;while(1){u=l[r>>0]|0|s<<8;if((t|0)<-8){r=r+1|0;s=u;t=t+8|0}else{v=u;break}}w=b+(p+1)|0;x=v;y=a+4+q|0}else{w=b;x=g;y=o}t=x>>>y;s=t+-1|0;r=(1<<y)+-1|0;u=r&x;z=h+e|0;A=(z|0)>(d|0)?d:z;a:do if((s|0)<0)if((h|0)<(A|0)){B=m&65535;C=(z|0)>(d|0)?d:z;D=h;do{j[c+(D<<1)>>1]=B;D=D+1|0}while((D|0)!=(C|0));E=w;F=u;G=C;H=m;I=y}else{E=w;F=u;G=h;H=m;I=y}else{C=(h|0)<(A|0);if((s|0)==14){if(!C){E=w;F=u;G=h;H=m;I=y;break}D=16-y|0;B=D+-8|0;J=(B|0)>-1;K=(y|0)>0;L=8-y|0;M=L>>>3;N=L-(M<<3)|0;L=(z|0)>(d|0)?d:z;O=M+1|0;M=w;P=u;Q=h;R=m;while(1){S=P<<D;if(J){T=M;U=S;V=B;while(1){W=(l[T>>0]|0)<<V|U;V=V+-8|0;if((V|0)<=-1){X=W;break}else{T=T+1|0;U=W}}Y=M+O|0;Z=X;_=N}else{Y=M;Z=S;_=D}if(K){U=l[Y>>0]|0;$=Y+1|0;aa=U&r;ba=U>>>(8-_|0)|Z}else{$=Y;aa=0;ba=Z}U=(ba<<31>>31^ba>>>1)+R|0;j[c+(Q<<1)>>1]=U;T=U&65535;Q=Q+1|0;if((Q|0)==(L|0)){E=$;F=aa;G=L;H=T;I=y;break a}else{M=$;P=aa;R=T}}}if(C){R=8-t|0;P=t+7|0;M=(z|0)>(d|0)?d:z;L=w;Q=u;K=h;D=m;N=y;while(1){if(!Q){O=L;B=N;while(1){J=B+8|0;T=O+1|0;U=i[O>>0]|0;if(!(U<<24>>24)){O=T;B=J}else{ca=J;da=T;ea=U;break}}fa=da;ga=ea&255;ha=ca}else{fa=L;ga=Q;ha=N}B=k[122e3+(ga<<2)>>2]|0;O=ha-B|0;S=ha+-1-O|0;U=1<<S^ga;T=S-s|0;if((T|0)<0){S=B-t|0;J=(P+((S|0)>-8?S:-8)-B|0)>>>3;S=J<<3;V=fa;W=U;ia=T;while(1){ja=l[V>>0]|0|W<<8;if((ia|0)<-8){V=V+1|0;W=ja;ia=ia+8|0}else{ka=ja;break}}la=fa+(J+1)|0;ma=ka;na=R+B+S|0}else{la=fa;ma=U;na=T}ia=ma>>>na|O<<s;W=(1<<na)+-1&ma;V=(ia<<31>>31^ia>>>1)+D|0;j[c+(K<<1)>>1]=V;ia=V&65535;K=K+1|0;if((K|0)==(M|0)){E=la;F=W;G=M;H=ia;I=na;break}else{L=la;Q=W;D=ia;N=na}}}else{E=w;F=u;G=h;H=m;I=y}}while(0);if(E>>>0>f>>>0){oa=31;break}else{b=E;g=F;h=G;m=H;a=I}}if((oa|0)==31){tk(782917);pa=1;return pa|0}if(n>>>0>=f>>>0){pa=0;return pa|0}tk(782972);pa=0;return pa|0}function Ht(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,fa=0,ga=0,ha=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0;f=a+b|0;b=a+2|0;g=l[a+1>>0]|0;h=0;j=l[a>>0]|0;a=8;while(1){if((h|0)>=(d|0)){m=b;break}n=a+-3|0;if((a|0)<3){o=(((n|0)>-8?a+7|0:2)-a|0)>>>3;p=o<<3;q=b;r=g;s=n;while(1){t=l[q>>0]|0|r<<8;if((s|0)<-8){q=q+1|0;r=t;s=s+8|0}else{u=t;break}}v=b+(o+1)|0;w=u;x=a+5+p|0}else{v=b;w=g;x=n}s=w>>>x;r=s+-1|0;q=(1<<x)+-1|0;t=q&w;y=h+e|0;z=(y|0)>(d|0)?d:y;a:do if((r|0)<0)if((h|0)<(z|0)){A=(y|0)>(d|0)?d:y;sy(c+h|0,j&255|0,~h-~A|0)|0;B=v;C=t;D=A;E=j;F=x}else{B=v;C=t;D=h;E=j;F=x}else{A=(h|0)<(z|0);if((r|0)==6){if(!A){B=v;C=t;D=h;E=j;F=x;break}G=8-x|0;H=G+-8|0;I=(H|0)>-1;J=(x|0)>0;K=0-x|0;L=K>>>3;M=K-(L<<3)|0;K=(y|0)>(d|0)?d:y;N=L+1|0;L=v;O=t;P=h;Q=j;while(1){R=O<<G;if(I){S=L;T=R;U=H;while(1){V=(l[S>>0]|0)<<U|T;U=U+-8|0;if((U|0)<=-1){W=V;break}else{S=S+1|0;T=V}}X=L+N|0;Y=W;Z=M}else{X=L;Y=R;Z=G}if(J){T=l[X>>0]|0;_=X+1|0;$=T&q;aa=T>>>(8-Z|0)|Y}else{_=X;$=0;aa=Y}T=(aa<<31>>31^aa>>>1)+Q|0;i[c+P>>0]=T;S=T&255;P=P+1|0;if((P|0)==(K|0)){B=_;C=$;D=K;E=S;F=x;break a}else{L=_;O=$;Q=S}}}if(A){Q=8-s|0;O=s+7|0;L=(y|0)>(d|0)?d:y;K=v;P=t;J=h;G=j;M=x;while(1){if(!P){N=K;H=M;while(1){I=H+8|0;S=N+1|0;T=i[N>>0]|0;if(!(T<<24>>24)){N=S;H=I}else{ba=I;ca=S;da=T;break}}ea=ca;fa=da&255;ga=ba}else{ea=K;fa=P;ga=M}H=k[122e3+(fa<<2)>>2]|0;N=ga-H|0;R=ga+-1-N|0;T=1<<R^fa;S=R-r|0;if((S|0)<0){R=H-s|0;I=(O+((R|0)>-8?R:-8)-H|0)>>>3;R=I<<3;U=ea;V=T;ha=S;while(1){ia=l[U>>0]|0|V<<8;if((ha|0)<-8){U=U+1|0;V=ia;ha=ha+8|0}else{ja=ia;break}}ka=ea+(I+1)|0;la=ja;ma=Q+H+R|0}else{ka=ea;la=T;ma=S}ha=la>>>ma|N<<r;V=(1<<ma)+-1&la;U=(ha<<31>>31^ha>>>1)+G|0;i[c+J>>0]=U;ha=U&255;J=J+1|0;if((J|0)==(L|0)){B=ka;C=V;D=L;E=ha;F=ma;break}else{K=ka;P=V;G=ha;M=ma}}}else{B=v;C=t;D=h;E=j;F=x}}while(0);if(B>>>0>f>>>0){na=30;break}else{b=B;g=C;h=D;j=E;a=F}}if((na|0)==30){tk(782917);oa=1;return oa|0}if(m>>>0>=f>>>0){oa=0;return oa|0}tk(782972);oa=0;return oa|0}function It(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,l=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;if((d|0)<1){e=0;return e|0}j[c+4>>1]=-100;j[c+2>>1]=7;j[c>>1]=0;j[c+10>>1]=0;j[c+12>>1]=0;f=d+b|0;d=f+-1|0;g=k[a+(b+-1<<2)>>2]|0;h=1;i=-2-b|0;l=b+4096|0;n=b;b=0;o=8;p=(g|0)<0?0:g;while(1){g=(n|0)<(d|0);a:do if(g){q=g;r=l;s=i;t=n;u=p;v=n;while(1){w=k[a+(t<<2)>>2]|0;x=(w|0)<0?0:w;w=t+1|0;if((x|0)!=(u|0))if(!u){y=x;z=w}else{A=q;B=r;C=s;D=t;E=x;F=u;G=v;break a}else{y=u;z=v}x=s+-1|0;H=r+1|0;I=(w|0)<(d|0);if(I){q=I;r=H;s=x;t=w;u=y;v=z}else{J=I;K=x;L=H;M=w;N=y;O=y;P=z;Q=6;break}}}else{J=g;K=i;L=l;M=n;N=b;O=p;P=n;Q=6}while(0);if((Q|0)==6){Q=0;A=J;B=L;C=K;D=M;E=N;F=O;G=(O|0)==0?f:P}g=D-G|0;v=g+1|0;u=G-n|0;t=(F|0)>0;do if(t){s=F-h|0;if((F|0)!=(h|0)){if((((s|0)>-1?s:0-s|0)|0)>4095){j[c+(o+-1<<1)>>1]=F&4095|4096;j[c+(o<<1)>>1]=(F|0)/4096|0;R=F;S=o+2|0;Q=16;break}if((s|0)<0){r=12288-s|0;j[c+(o+-1<<1)>>1]=r;T=r}else{r=s+8192|0;j[c+(o+-1<<1)>>1]=r;T=r}r=o+1|0;if((D|0)==(G|0)&(G|0)==(n|0)){j[c+(o+-1<<1)>>1]=T|16384;U=F;V=r}else{R=F;S=r;Q=16}}else{R=F;S=o;Q=16}}else{R=h;S=o;Q=16}while(0);do if((Q|0)==16){Q=0;if((u|0)>0){r=n+-1-G|0;s=((G+4095+((r|0)>-4096?r:-4096)-n|0)>>>0)/4095|0;r=u;q=S;while(1){w=(r|0)>4095;j[c+(q+-1<<1)>>1]=w?4095:r&65535;if(w){r=r+-4095|0;q=q+1|0}else break}q=S+1+s|0;if(t&(D|0)==(G|0)){r=c+(q+-2<<1)|0;j[r>>1]=(m[r>>1]|0)+20481;U=R;V=q;break}else W=q}else W=S;if((g|0)>-1){q=G+C|0;r=((((q|0)>-4096?q:-4096)+B-G|0)>>>0)/4095|0;q=v;w=W;while(1){H=(q|0)>4095;j[c+(w+-1<<1)>>1]=H?20479:q+16384&65535;if(H){q=q+-4095|0;w=w+1|0}else break}U=R;V=W+1+r|0}else{U=R;V=W}}while(0);if(A){h=U;i=C+-1|0;l=B+1|0;n=D+1|0;b=E;o=V;p=E}else{X=V;break}}V=X+-1|0;j[c+6>>1]=(V|0)%32768|0;j[c+8>>1]=(V|0)/32768|0;e=V;return e|0}function Jt(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0;e=j[a+4>>1]|0;if(e<<16>>16>0){f=4;g=e<<16>>16}else{f=(j[a+2>>1]|0)+1|0;g=(j[a+8>>1]<<15)+(j[a+6>>1]|0)|0}if((d|0)<1|(g|0)<1){h=0;return h|0}e=b+-1+d|0;a:do if((f|0)>(g|0))i=1;else{l=0-d-b|0;m=f;n=1;o=1;p=0;q=1;while(1){if(!p){r=j[a+(m+-1<<1)>>1]|0;s=r<<16>>16;t=s&4095;b:do if(r<<16>>16<-4095){u=n;v=o;w=0;x=q}else{switch((s|0)/4096|0|0){case 5:case 4:case 0:{y=t+q|0;z=y+-1|0;A=(q|0)>(b|0)?q:b;B=(z|0)<(e|0)?z:e;C=B-A|0;if((C|0)<=-1){u=n;v=o;w=0;x=y;break b}D=n+1+C|0;E=C+n|0;C=s&-4096;F=(n|0)>(E|0);if((C|0)==16384){if(F){u=D;v=o;w=0;x=y;break b}else G=n;while(1){k[c+(G+-1<<2)>>2]=o;if((G|0)<(E|0))G=G+1|0;else{u=D;v=o;w=0;x=y;break b}}}if(!F){H=0-q-(r&4095)|0;I=n+-1-A-((l|0)>(H|0)?l:H)|0;sy(c+(n+-1<<2)|0,0,((n|0)>(I|0)?n:I)+1-n<<2|0)|0}if(!((C|0)==20480&(B|0)==(z|0))){u=D;v=o;w=0;x=y;break b}k[c+(E+-1<<2)>>2]=o;u=D;v=o;w=0;x=y;break b;break}case 1:{u=n;v=j[a+(m<<1)>>1]<<12|t;w=1;x=q;break b;break}case 2:{u=n;v=t+o|0;w=0;x=q;break b;break}case 3:{u=n;v=o-t|0;w=0;x=q;break b;break}case 6:{J=t+o|0;break}case 7:{J=o-t|0;break}default:{u=n;v=o;w=0;x=q;break b}}if((q|0)<(b|0)|(q|0)>(e|0))K=n;else{k[c+(n+-1<<2)>>2]=J;K=n+1|0}u=K;v=J;w=0;x=q+1|0}while(0);if((x|0)>(e|0)){i=u;break a}else{L=u;M=v;N=w;O=x}}else{L=n;M=o;N=0;O=q}if((m|0)<(g|0)){m=m+1|0;n=L;o=M;p=N;q=O}else{i=L;break}}}while(0);if((i|0)>(d|0)){h=d;return h|0}sy(c+(i+-1<<2)|0,0,d+1-i<<2|0)|0;h=d;return h|0}function Kt(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0;h=r;r=r+16|0;m=h;n=h+12|0;o=k[g>>2]|0;if((o|0)>0){p=o;r=h;return p|0}o=(c|0)>(b|0)?c:b;q=~~(+fa(+(+(o|0)))/.6931471805599453+.5);s=((o|0)>(1<<q|0)&1)+q|0;q=cy(((o+1|0)/2|0)<<2)|0;if(!q){tk(783036);k[g>>2]=413;p=413;r=h;return p|0}if((s|0)>0){o=b<<1;t=b+2|0;u=0;v=-2;w=-4;x=1;y=c;z=b;A=1;B=2;C=0;while(1){D=(y|0)%2|0;E=(z|0)%2|0;F=y-D|0;if((F|0)>0){G=z-E|0;H=(E|0)==0;I=C^1;a:do if((G|0)<=0){if(!H){J=0;do{K=ha(J,b)|0;M=a+(K+b<<2)|0;N=k[M>>2]|0;O=a+(K<<2)|0;K=k[O>>2]|0;P=K+N<<I;Q=N-K<<I;k[M>>2]=((Q|0)>-1?A:0)+Q&v;k[O>>2]=((P|0)>-1?B:x)+P&w;J=J+2|0}while((J|0)<(F|0))}}else{J=z+-1-E&-2;if(H){P=0;while(1){O=ha(P,b)|0;Q=0;M=O;K=O+b|0;while(1){O=a+(K+1<<2)|0;N=k[O>>2]|0;R=a+(K<<2)|0;S=k[R>>2]|0;T=S+N|0;U=a+(M+1<<2)|0;V=k[U>>2]|0;W=a+(M<<2)|0;X=k[W>>2]|0;Y=T+V+X>>C;Z=T-V-X>>C;T=N-S|0;S=T+V-X>>C;k[O>>2]=T-V+X>>C;k[R>>2]=((Z|0)>-1?A:0)+Z&v;k[U>>2]=((S|0)>-1?A:0)+S&v;k[W>>2]=((Y|0)>-1?B:x)+Y&w;Q=Q+2|0;if((Q|0)>=(G|0))break;else{M=M+2|0;K=K+2|0}}P=P+2|0;if((P|0)>=(F|0))break a}}P=0;K=t+J|0;M=J+2|0;while(1){Q=ha(P,b)|0;Y=0;W=Q;S=Q+b|0;while(1){Q=a+(S+1<<2)|0;U=k[Q>>2]|0;Z=a+(S<<2)|0;R=k[Z>>2]|0;X=R+U|0;V=a+(W+1<<2)|0;T=k[V>>2]|0;O=a+(W<<2)|0;N=k[O>>2]|0;_=X+T+N>>C;$=X-T-N>>C;X=U-R|0;R=X+T-N>>C;k[Q>>2]=X-T+N>>C;k[Z>>2]=(($|0)>-1?A:0)+$&v;k[V>>2]=((R|0)>-1?A:0)+R&v;k[O>>2]=((_|0)>-1?B:x)+_&w;Y=Y+2|0;if((Y|0)>=(G|0))break;else{W=W+2|0;S=S+2|0}}S=a+(K<<2)|0;W=k[S>>2]|0;Y=a+(M<<2)|0;_=k[Y>>2]|0;O=_+W<<I;R=W-_<<I;k[S>>2]=((R|0)>-1?A:0)+R&v;k[Y>>2]=((O|0)>-1?B:x)+O&w;P=P+2|0;if((P|0)>=(F|0))break;else{K=K+o|0;M=M+o|0}}}while(0);aa=y+1-D&-2}else aa=0;if(D){F=ha(aa,b)|0;I=z-E|0;if((I|0)>0){G=C^1;H=z+-1|0;M=0;K=F;while(1){P=a+(K+1<<2)|0;J=k[P>>2]|0;O=a+(K<<2)|0;Y=k[O>>2]|0;R=Y+J<<G;S=J-Y<<G;k[P>>2]=((S|0)>-1?A:0)+S&v;k[O>>2]=((R|0)>-1?B:x)+R&w;M=M+2|0;if((M|0)>=(I|0))break;else K=K+2|0}ba=(H-E+2&-2)+F|0}else ba=F;if(E){K=a+(ba<<2)|0;I=k[K>>2]<<2-C;k[K>>2]=((I|0)>-1?B:x)+I&w}}if((y|0)>0){I=(z|0)>1;K=(z|0)>2;M=((z+2147483645|0)>>>1<<2)+8|0;G=0;do{D=ha(G,b)|0;R=a+(D<<2)|0;O=a+(D+1<<2)|0;if(I){D=1;S=O;P=q;while(1){k[P>>2]=k[S>>2];D=D+2|0;if((D|0)>=(z|0))break;else{S=S+8|0;P=P+4|0}}if(K){P=R;S=2;D=O;while(1){P=P+8|0;k[D>>2]=k[P>>2];S=S+2|0;if((S|0)>=(z|0))break;else D=D+4|0}ca=1;da=R+M|0;ea=q}else{ca=1;da=O;ea=q}while(1){k[da>>2]=k[ea>>2];ca=ca+2|0;if((ca|0)>=(z|0))break;else{da=da+4|0;ea=ea+4|0}}}G=G+1|0}while((G|0)!=(y|0))}if((z|0)>0){G=(y|0)>1;M=(y|0)>2;K=ha(((y+2147483645|0)>>>1<<2)+8|0,b)|0;I=0;do{E=a+(I<<2)|0;F=a+(I+b<<2)|0;if(G){H=1;O=F;R=q;while(1){k[R>>2]=k[O>>2];H=H+2|0;if((H|0)>=(y|0))break;else{O=O+(o<<2)|0;R=R+4|0}}if(M){R=E;O=2;H=F;while(1){R=R+(o<<2)|0;k[H>>2]=k[R>>2];O=O+2|0;if((O|0)>=(y|0))break;else H=H+(b<<2)|0}ga=1;ia=E+K|0;ja=q}else{ga=1;ia=F;ja=q}while(1){k[ia>>2]=k[ja>>2];ga=ga+2|0;if((ga|0)>=(y|0))break;else{ia=ia+(b<<2)|0;ja=ja+4|0}}}I=I+1|0}while((I|0)!=(z|0))}I=B<<1;u=u+1|0;if((u|0)==(s|0))break;else{K=B;M=w;w=w<<1;x=I+-1|0;y=y+1>>1;z=z+1>>1;B=I;C=1;A=K;v=M}}}dy(q);if((d|0)>=2){q=(d+1|0)/2|0;v=q+-1|0;A=ha(c,b)|0;C=a+(A+-1<<2)|0;if((A|0)<1)ka=A;else{B=1-q|0;q=a;do{z=k[q>>2]|0;k[q>>2]=(((z|0)>0?v:B)+z|0)/(d|0)|0;q=q+4|0}while(q>>>0<=C>>>0);ka=A}}else ka=ha(c,b)|0;k[30756]=k[f>>2];k[f>>2]=0;k[30757]=0;A=k[30756]|0;if((A|0)<2){la=0;ma=A}else{A=j[163161]|0;i[e>>0]=A;i[e+1>>0]=A>>8;A=(k[30757]|0)+2|0;k[30757]=A;la=A;ma=k[30756]|0}if((la|0)<(ma|0)){i[e+la>>0]=c>>>24;A=(k[30757]|0)+1|0;k[30757]=A;na=A;oa=k[30756]|0}else{na=la;oa=ma}if((na|0)<(oa|0)){i[e+na>>0]=c>>>16;ma=(k[30757]|0)+1|0;k[30757]=ma;pa=ma;qa=k[30756]|0}else{pa=na;qa=oa}if((pa|0)<(qa|0)){i[e+pa>>0]=c>>>8;oa=(k[30757]|0)+1|0;k[30757]=oa;ra=oa;sa=k[30756]|0}else{ra=pa;sa=qa}if((ra|0)<(sa|0)){i[e+ra>>0]=c;qa=(k[30757]|0)+1|0;k[30757]=qa;ta=qa;ua=k[30756]|0}else{ta=ra;ua=sa}if((ta|0)<(ua|0)){i[e+ta>>0]=b>>>24;sa=(k[30757]|0)+1|0;k[30757]=sa;va=sa;wa=k[30756]|0}else{va=ta;wa=ua}if((va|0)<(wa|0)){i[e+va>>0]=b>>>16;ua=(k[30757]|0)+1|0;k[30757]=ua;xa=ua;ya=k[30756]|0}else{xa=va;ya=wa}if((xa|0)<(ya|0)){i[e+xa>>0]=b>>>8;wa=(k[30757]|0)+1|0;k[30757]=wa;za=wa;Aa=k[30756]|0}else{za=xa;Aa=ya}if((za|0)<(Aa|0)){i[e+za>>0]=b;ya=(k[30757]|0)+1|0;k[30757]=ya;Ba=ya;Ca=k[30756]|0}else{Ba=za;Ca=Aa}if((Ba|0)<(Ca|0)){i[e+Ba>>0]=d>>>24;Aa=(k[30757]|0)+1|0;k[30757]=Aa;Da=Aa;Ea=k[30756]|0}else{Da=Ba;Ea=Ca}if((Da|0)<(Ea|0)){i[e+Da>>0]=d>>>16;Ca=(k[30757]|0)+1|0;k[30757]=Ca;Fa=Ca;Ga=k[30756]|0}else{Fa=Da;Ga=Ea}if((Fa|0)<(Ga|0)){i[e+Fa>>0]=d>>>8;Ea=(k[30757]|0)+1|0;k[30757]=Ea;Ha=Ea;Ia=k[30756]|0}else{Ha=Fa;Ia=Ga}if((Ha|0)<(Ia|0)){i[e+Ha>>0]=d;k[30757]=(k[30757]|0)+1}d=k[a>>2]|0;Mt(e,d,((d|0)<0)<<31>>31);k[a>>2]=0;d=ey(1,(ka+7|0)/8|0)|0;b:do if(!d){tk(783064);Ja=413}else{do if((ka|0)>0){Ha=0;Ia=8;Ga=0;Fa=0;while(1){Ea=a+(Ga<<2)|0;do if((Ha|0)>0){Da=d+Fa|0;i[Da>>0]=(l[Da>>0]|0)<<1;Ka=Ia+-1|0}else{if((Ha|0)>=0){Ka=Ia;break}Da=d+Fa|0;i[Da>>0]=(l[Da>>0]|0)<<1|1;k[Ea>>2]=0-Ha;Ka=Ia+-1|0}while(0);Ea=(Ka|0)==0;Da=(Ea&1)+Fa|0;Ca=Ea?8:Ka;Ea=Ga+1|0;if((Ea|0)==(ka|0)){La=Ca;Ma=Da;break}Ha=k[a+(Ea<<2)>>2]|0;Ia=Ca;Ga=Ea;Fa=Da}if((La|0)==8)Na=Ma;else{Fa=d+Ma|0;i[Fa>>0]=(l[Fa>>0]|0)<<La;Na=Ma+1|0};k[m>>2]=0;k[m+4>>2]=0;k[m+8>>2]=0;Fa=(c+1|0)/2|0;Ga=(b+1|0)/2|0;Ia=0;Ha=0;Da=0;while(1){Ea=m+(((Ha|0)>=(Ga|0)&1)+((Da|0)>=(Fa|0)&1)<<2)|0;Ca=k[a+(Ia<<2)>>2]|0;if((k[Ea>>2]|0)<(Ca|0))k[Ea>>2]=Ca;Ca=Ha+1|0;Ea=(Ca|0)<(b|0);Ia=Ia+1|0;if((Ia|0)==(ka|0))break;else{Ha=Ea?Ca:0;Da=(Ea&1^1)+Da|0}}Da=k[m>>2]|0;i[n>>0]=0;if((Da|0)>0){Oa=Da;Pa=0}else{Qa=0;Ra=Fa;Sa=Ga;Ta=Na;break}while(1){Da=Oa>>1;Ha=Pa+1<<24>>24;if((Da|0)>0){Oa=Da;Pa=Ha}else{Ua=Da;Va=Ha;break}}k[m>>2]=Ua;i[n>>0]=Va;Qa=Va;Ra=Fa;Sa=Ga;Ta=Na}else{k[m>>2]=0;k[m+4>>2]=0;k[m+8>>2]=0;i[n>>0]=0;Qa=0;Ra=(c+1|0)/2|0;Sa=(b+1|0)/2|0;Ta=0}while(0);Ha=n+1|0;i[Ha>>0]=0;Da=m+4|0;Ia=k[Da>>2]|0;if((Ia|0)>0){Ea=Ia;Ia=0;while(1){Ca=Ea>>1;Ba=Ia+1<<24>>24;if((Ca|0)>0){Ea=Ca;Ia=Ba}else{Wa=Ca;Xa=Ba;break}}k[Da>>2]=Wa;i[Ha>>0]=Xa;Ya=Xa}else Ya=0;Ia=n+2|0;i[Ia>>0]=0;Ea=m+8|0;Ba=k[Ea>>2]|0;if((Ba|0)>0){Ca=Ba;Ba=0;while(1){Aa=Ca>>1;za=Ba+1<<24>>24;if((Aa|0)>0){Ca=Aa;Ba=za}else{Za=Aa;_a=za;break}}k[Ea>>2]=Za;i[Ia>>0]=_a;$a=_a}else $a=0;Ba=k[30757]|0;if((Ba+3|0)>(k[30756]|0)){k[f>>2]=Ba;tk(783092);Ja=413;break}Ca=e+Ba|0;i[Ca>>0]=i[n>>0]|0;i[Ca+1>>0]=i[n+1>>0]|0;i[Ca+2>>0]=i[n+2>>0]|0;k[30757]=(k[30757]|0)+3;k[30758]=0;k[30759]=8;Ca=99096;k[Ca>>2]=0;k[Ca+4>>2]=0;Ca=Nt(e,a,b,Ra,Sa,Qa&255)|0;do if(!Ca){Ba=(b|0)/2|0;Ha=Ya&255;Da=Nt(e,a+(Sa<<2)|0,b,Ra,Ba,Ha)|0;if(Da){ab=Da;break}Da=ha(Ra,b)|0;za=(c|0)/2|0;Aa=Nt(e,a+(Da<<2)|0,b,za,Sa,Ha)|0;if(Aa){ab=Aa;break}ab=Nt(e,a+(Da+Sa<<2)|0,b,za,Ba,$a&255)|0}else ab=Ca;while(0);Ca=k[30758]<<4;k[30758]=Ca;Ia=k[30759]|0;Ea=Ia+-4|0;k[30759]=Ea;if((Ea|0)<1){i[e+(k[30757]|0)>>0]=Ca>>4-Ia;Ia=k[30757]|0;if((Ia|0)<(k[30756]|0))k[30757]=Ia+1;Ia=(k[30759]|0)+8|0;k[30759]=Ia;bb=Ia}else bb=Ea;Ea=99096;Ia=ky(k[Ea>>2]|0,k[Ea+4>>2]|0,4,0)|0;Ea=99096;k[Ea>>2]=Ia;k[Ea+4>>2]=L;if((bb|0)<8){i[e+(k[30757]|0)>>0]=k[30758]<<bb;Ea=k[30757]|0;if((Ea|0)<(k[30756]|0))k[30757]=Ea+1;Ea=k[30759]|0;Ia=99096;Ca=ky(k[Ia>>2]|0,k[Ia+4>>2]|0,Ea|0,((Ea|0)<0)<<31>>31|0)|0;Ea=99096;k[Ea>>2]=Ca;k[Ea+4>>2]=L}do if((Ta|0)>0){Ea=k[30757]|0;if((Ea+Ta|0)>(k[30756]|0)){dy(d);k[f>>2]=k[30757];tk(783092);Ja=413;break b}else{py(e+Ea|0,d|0,Ta|0)|0;k[30757]=(k[30757]|0)+Ta;break}}while(0);dy(d);k[f>>2]=k[30757];if((k[30757]|0)<(k[30756]|0)){Ja=ab;break}tk(783092);Ja=413}while(0);k[g>>2]=Ja;p=Ja;r=h;return p|0}function Lt(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0;h=r;r=r+32|0;m=h;n=h+24|0;o=k[g>>2]|0;if((o|0)>0){p=o;r=h;return p|0}o=(c|0)>(b|0)?c:b;q=~~(+fa(+(+(o|0)))/.6931471805599453+.5);s=((o|0)>(1<<q|0)&1)+q|0;q=cy(((o+1|0)/2|0)<<3)|0;if(!q){tk(783124);k[g>>2]=413;p=413;r=h;return p|0}if((s|0)>0){o=b<<1;t=b+2|0;u=1;v=0;w=-2;x=-1;y=2;z=1;A=0;B=0;C=-4;D=-1;E=0;F=c;G=b;H=0;while(1){I=(F|0)%2|0;J=(G|0)%2|0;K=F-I|0;if((K|0)>0){M=G-J|0;N=(M|0)>0;O=(J|0)==0;P=H^1;Q=G+-1-J&-2;R=0;S=t+Q|0;T=Q+2|0;while(1){Q=ha(R,b)|0;U=Q+b|0;if(N){V=0;W=Q;X=U;while(1){Y=a+(X+1<<3)|0;Z=Y;_=k[Z>>2]|0;$=k[Z+4>>2]|0;Z=a+(X<<3)|0;aa=Z;ba=k[aa>>2]|0;ca=k[aa+4>>2]|0;aa=ky(ba|0,ca|0,_|0,$|0)|0;da=L;ea=a+((W|1)<<3)|0;ga=ea;ia=k[ga>>2]|0;ja=k[ga+4>>2]|0;ga=ky(aa|0,da|0,ia|0,ja|0)|0;ka=a+(W<<3)|0;la=ka;ma=k[la>>2]|0;na=k[la+4>>2]|0;la=ky(ga|0,L|0,ma|0,na|0)|0;ga=my(la|0,L|0,H|0)|0;la=L;oa=jy(aa|0,da|0,ia|0,ja|0)|0;da=jy(oa|0,L|0,ma|0,na|0)|0;oa=my(da|0,L|0,H|0)|0;da=L;aa=jy(_|0,$|0,ba|0,ca|0)|0;ca=L;ba=ky(aa|0,ca|0,ia|0,ja|0)|0;$=jy(ba|0,L|0,ma|0,na|0)|0;ba=my($|0,L|0,H|0)|0;$=L;_=jy(aa|0,ca|0,ia|0,ja|0)|0;ja=ky(_|0,L|0,ma|0,na|0)|0;na=my(ja|0,L|0,H|0)|0;ja=Y;k[ja>>2]=na;k[ja+4>>2]=L;ja=(da|0)>-1|(da|0)==-1&oa>>>0>4294967295;na=ky((ja?u:0)|0,(ja?v:0)|0,oa|0,da|0)|0;da=Z;k[da>>2]=na&w;k[da+4>>2]=L&x;da=($|0)>-1|($|0)==-1&ba>>>0>4294967295;na=ky((da?u:0)|0,(da?v:0)|0,ba|0,$|0)|0;$=ea;k[$>>2]=na&w;k[$+4>>2]=L&x;$=(la|0)>-1|(la|0)==-1&ga>>>0>4294967295;na=ky(($?y:z)|0,($?A:B)|0,ga|0,la|0)|0;la=ka;k[la>>2]=na&C;k[la+4>>2]=L&D;V=V+2|0;if((V|0)>=(M|0)){pa=T;qa=S;break}else{W=W+2|0;X=X+2|0}}}else{pa=Q;qa=U}if(!O){X=a+(qa<<3)|0;W=X;V=k[W>>2]|0;la=k[W+4>>2]|0;W=a+(pa<<3)|0;na=W;ka=k[na>>2]|0;ga=k[na+4>>2]|0;na=ky(ka|0,ga|0,V|0,la|0)|0;$=ty(na|0,L|0,P|0)|0;na=L;ea=jy(V|0,la|0,ka|0,ga|0)|0;ga=ty(ea|0,L|0,P|0)|0;ea=L;ka=(ea|0)>-1|(ea|0)==-1&ga>>>0>4294967295;la=ky((ka?u:0)|0,(ka?v:0)|0,ga|0,ea|0)|0;ea=X;k[ea>>2]=la&w;k[ea+4>>2]=L&x;ea=(na|0)>-1|(na|0)==-1&$>>>0>4294967295;la=ky((ea?y:z)|0,(ea?A:B)|0,$|0,na|0)|0;na=W;k[na>>2]=la&C;k[na+4>>2]=L&D}R=R+2|0;if((R|0)>=(K|0))break;else{S=S+o|0;T=T+o|0}}ra=F+1-I&-2}else ra=0;if(I){T=ha(ra,b)|0;S=G-J|0;if((S|0)>0){K=H^1;R=G+-1|0;P=0;O=T;while(1){M=a+(O+1<<3)|0;N=M;na=k[N>>2]|0;la=k[N+4>>2]|0;N=a+(O<<3)|0;W=N;$=k[W>>2]|0;ea=k[W+4>>2]|0;W=ky($|0,ea|0,na|0,la|0)|0;X=ty(W|0,L|0,K|0)|0;W=L;ga=jy(na|0,la|0,$|0,ea|0)|0;ea=ty(ga|0,L|0,K|0)|0;ga=L;$=(ga|0)>-1|(ga|0)==-1&ea>>>0>4294967295;la=ky(($?u:0)|0,($?v:0)|0,ea|0,ga|0)|0;ga=M;k[ga>>2]=la&w;k[ga+4>>2]=L&x;ga=(W|0)>-1|(W|0)==-1&X>>>0>4294967295;la=ky((ga?y:z)|0,(ga?A:B)|0,X|0,W|0)|0;W=N;k[W>>2]=la&C;k[W+4>>2]=L&D;P=P+2|0;if((P|0)>=(S|0))break;else O=O+2|0}sa=(R-J+2&-2)+T|0}else sa=T;if(J){O=a+(sa<<3)|0;S=O;P=ty(k[S>>2]|0,k[S+4>>2]|0,2-H|0)|0;S=L;K=(S|0)>-1|(S|0)==-1&P>>>0>4294967295;I=ky((K?y:z)|0,(K?A:B)|0,P|0,S|0)|0;S=O;k[S>>2]=I&C;k[S+4>>2]=L&D}}if((F|0)>0){S=(G|0)>1;I=(G|0)>2;O=((G+1073741821|0)>>>1<<3)+16|0;P=0;do{K=ha(P,b)|0;W=a+(K<<3)|0;la=a+(K+1<<3)|0;if(S){K=1;N=la;X=q;while(1){ga=N;M=k[ga+4>>2]|0;ea=X;k[ea>>2]=k[ga>>2];k[ea+4>>2]=M;K=K+2|0;if((K|0)>=(G|0))break;else{N=N+16|0;X=X+8|0}}if(I){X=W;N=2;K=la;while(1){X=X+16|0;U=X;Q=k[U+4>>2]|0;M=K;k[M>>2]=k[U>>2];k[M+4>>2]=Q;N=N+2|0;if((N|0)>=(G|0))break;else K=K+8|0}ta=1;ua=W+O|0;va=q}else{ta=1;ua=la;va=q}while(1){K=va;N=k[K+4>>2]|0;X=ua;k[X>>2]=k[K>>2];k[X+4>>2]=N;ta=ta+2|0;if((ta|0)>=(G|0))break;else{ua=ua+8|0;va=va+8|0}}}P=P+1|0}while((P|0)!=(F|0))}if((G|0)>0){P=(F|0)>1;O=(F|0)>2;I=ha(((F+1073741821|0)>>>1<<3)+16|0,b)|0;S=0;do{J=a+(S<<3)|0;T=a+(S+b<<3)|0;if(P){R=1;la=T;W=q;while(1){N=la;X=k[N+4>>2]|0;K=W;k[K>>2]=k[N>>2];k[K+4>>2]=X;R=R+2|0;if((R|0)>=(F|0))break;else{la=la+(o<<3)|0;W=W+8|0}}if(O){W=J;la=2;R=T;while(1){W=W+(o<<3)|0;X=W;K=k[X+4>>2]|0;N=R;k[N>>2]=k[X>>2];k[N+4>>2]=K;la=la+2|0;if((la|0)>=(F|0))break;else R=R+(b<<3)|0}wa=1;xa=J+I|0;ya=q}else{wa=1;xa=T;ya=q}while(1){R=ya;la=k[R+4>>2]|0;W=xa;k[W>>2]=k[R>>2];k[W+4>>2]=la;wa=wa+2|0;if((wa|0)>=(F|0))break;else{xa=xa+(b<<3)|0;ya=ya+8|0}}}S=S+1|0}while((S|0)!=(G|0))}S=ty(C|0,D|0,1)|0;I=L;O=ty(y|0,A|0,1)|0;P=L;z=ky(O|0,P|0,-1,-1)|0;E=E+1|0;if((E|0)==(s|0))break;else{T=D;J=C;la=A;W=y;y=O;A=P;B=L;C=S;D=I;F=F+1>>1;G=G+1>>1;H=1;x=T;w=J;v=la;u=W}}}dy(q);if((d|0)>=2){q=((d+1|0)/2|0)+-1|0;u=((q|0)<0)<<31>>31;v=((d|0)<0)<<31>>31;w=ha(c,b)|0;x=a+(w+-1<<3)|0;if((w|0)<1)za=w;else{H=jy(0,0,q|0,u|0)|0;G=L;F=a;do{D=F;C=k[D>>2]|0;B=k[D+4>>2]|0;D=(B|0)>0|(B|0)==0&C>>>0>0;A=ky((D?q:H)|0,(D?u:G)|0,C|0,B|0)|0;B=wy(A|0,L|0,d|0,v|0)|0;A=F;k[A>>2]=B;k[A+4>>2]=L;F=F+8|0}while(F>>>0<=x>>>0);za=w}}else za=ha(c,b)|0;k[30756]=k[f>>2];k[f>>2]=0;k[30757]=0;w=k[30756]|0;if((w|0)<2){Aa=0;Ba=w}else{w=j[163161]|0;i[e>>0]=w;i[e+1>>0]=w>>8;w=(k[30757]|0)+2|0;k[30757]=w;Aa=w;Ba=k[30756]|0}if((Aa|0)<(Ba|0)){i[e+Aa>>0]=c>>>24;w=(k[30757]|0)+1|0;k[30757]=w;Ca=w;Da=k[30756]|0}else{Ca=Aa;Da=Ba}if((Ca|0)<(Da|0)){i[e+Ca>>0]=c>>>16;Ba=(k[30757]|0)+1|0;k[30757]=Ba;Ea=Ba;Fa=k[30756]|0}else{Ea=Ca;Fa=Da}if((Ea|0)<(Fa|0)){i[e+Ea>>0]=c>>>8;Da=(k[30757]|0)+1|0;k[30757]=Da;Ga=Da;Ha=k[30756]|0}else{Ga=Ea;Ha=Fa}if((Ga|0)<(Ha|0)){i[e+Ga>>0]=c;Fa=(k[30757]|0)+1|0;k[30757]=Fa;Ia=Fa;Ja=k[30756]|0}else{Ia=Ga;Ja=Ha}if((Ia|0)<(Ja|0)){i[e+Ia>>0]=b>>>24;Ha=(k[30757]|0)+1|0;k[30757]=Ha;Ka=Ha;La=k[30756]|0}else{Ka=Ia;La=Ja}if((Ka|0)<(La|0)){i[e+Ka>>0]=b>>>16;Ja=(k[30757]|0)+1|0;k[30757]=Ja;Ma=Ja;Na=k[30756]|0}else{Ma=Ka;Na=La}if((Ma|0)<(Na|0)){i[e+Ma>>0]=b>>>8;La=(k[30757]|0)+1|0;k[30757]=La;Oa=La;Pa=k[30756]|0}else{Oa=Ma;Pa=Na}if((Oa|0)<(Pa|0)){i[e+Oa>>0]=b;Na=(k[30757]|0)+1|0;k[30757]=Na;Qa=Na;Ra=k[30756]|0}else{Qa=Oa;Ra=Pa}if((Qa|0)<(Ra|0)){i[e+Qa>>0]=d>>>24;Pa=(k[30757]|0)+1|0;k[30757]=Pa;Sa=Pa;Ta=k[30756]|0}else{Sa=Qa;Ta=Ra}if((Sa|0)<(Ta|0)){i[e+Sa>>0]=d>>>16;Ra=(k[30757]|0)+1|0;k[30757]=Ra;Ua=Ra;Va=k[30756]|0}else{Ua=Sa;Va=Ta}if((Ua|0)<(Va|0)){i[e+Ua>>0]=d>>>8;Ta=(k[30757]|0)+1|0;k[30757]=Ta;Wa=Ta;Xa=k[30756]|0}else{Wa=Ua;Xa=Va}if((Wa|0)<(Xa|0)){i[e+Wa>>0]=d;k[30757]=(k[30757]|0)+1}d=a;Mt(e,k[d>>2]|0,k[d+4>>2]|0);d=a;k[d>>2]=0;k[d+4>>2]=0;d=ey(1,(za+7|0)/8|0)|0;a:do if(!d){tk(783154);Ya=413}else{do if((za|0)>0){Wa=0;Xa=0;Va=8;Ua=0;Ta=0;while(1){Sa=a+(Ua<<3)|0;do if((Wa|0)>0|(Wa|0)==0&Xa>>>0>0){Ra=d+Ta|0;i[Ra>>0]=(l[Ra>>0]|0)<<1;Za=Va+-1|0}else{if((Wa|0)>=0){Za=Va;break}Ra=d+Ta|0;i[Ra>>0]=(l[Ra>>0]|0)<<1|1;Ra=jy(0,0,Xa|0,Wa|0)|0;Qa=Sa;k[Qa>>2]=Ra;k[Qa+4>>2]=L;Za=Va+-1|0}while(0);Sa=(Za|0)==0;Qa=(Sa&1)+Ta|0;Ra=Sa?8:Za;Sa=Ua+1|0;if((Sa|0)==(za|0)){_a=Ra;$a=Qa;break}Pa=a+(Sa<<3)|0;Wa=k[Pa+4>>2]|0;Xa=k[Pa>>2]|0;Va=Ra;Ua=Sa;Ta=Qa}if((_a|0)==8)ab=$a;else{Ta=d+$a|0;i[Ta>>0]=(l[Ta>>0]|0)<<_a;ab=$a+1|0};k[m>>2]=0;k[m+4>>2]=0;k[m+8>>2]=0;k[m+12>>2]=0;k[m+16>>2]=0;k[m+20>>2]=0;Ta=(c+1|0)/2|0;Ua=(b+1|0)/2|0;Va=0;Xa=0;Wa=0;while(1){Qa=m+(((Xa|0)>=(Ua|0)&1)+((Wa|0)>=(Ta|0)&1)<<3)|0;Sa=Qa;Ra=k[Sa+4>>2]|0;Pa=a+(Va<<3)|0;Oa=k[Pa>>2]|0;Na=k[Pa+4>>2]|0;if((Ra|0)<(Na|0)|((Ra|0)==(Na|0)?(k[Sa>>2]|0)>>>0<Oa>>>0:0)){Sa=Qa;k[Sa>>2]=Oa;k[Sa+4>>2]=Na}Na=Xa+1|0;Sa=(Na|0)<(b|0);Va=Va+1|0;if((Va|0)==(za|0))break;else{Xa=Sa?Na:0;Wa=(Sa&1^1)+Wa|0}}Wa=m;Xa=k[Wa>>2]|0;Va=k[Wa+4>>2]|0;i[n>>0]=0;if((Va|0)>0|(Va|0)==0&Xa>>>0>0){bb=Xa;cb=Va;db=0}else{eb=0;fb=Ta;gb=Ua;hb=ab;break}while(1){Va=my(bb|0,cb|0,1)|0;Xa=L;Wa=db+1<<24>>24;if((Xa|0)>0|(Xa|0)==0&Va>>>0>0){bb=Va;cb=Xa;db=Wa}else{ib=Wa;jb=Va;kb=Xa;break}}Xa=m;k[Xa>>2]=jb;k[Xa+4>>2]=kb;i[n>>0]=ib;eb=ib;fb=Ta;gb=Ua;hb=ab}else{k[m>>2]=0;k[m+4>>2]=0;k[m+8>>2]=0;k[m+12>>2]=0;k[m+16>>2]=0;k[m+20>>2]=0;i[n>>0]=0;eb=0;fb=(c+1|0)/2|0;gb=(b+1|0)/2|0;hb=0}while(0);Xa=n+1|0;i[Xa>>0]=0;Va=m+8|0;Wa=Va;Sa=k[Wa>>2]|0;Na=k[Wa+4>>2]|0;if((Na|0)>0|(Na|0)==0&Sa>>>0>0){Wa=Sa;Sa=Na;Na=0;while(1){Oa=my(Wa|0,Sa|0,1)|0;Qa=L;Ra=Na+1<<24>>24;if((Qa|0)>0|(Qa|0)==0&Oa>>>0>0){Wa=Oa;Sa=Qa;Na=Ra}else{lb=Ra;mb=Oa;nb=Qa;break}}Na=Va;k[Na>>2]=mb;k[Na+4>>2]=nb;i[Xa>>0]=lb;ob=lb}else ob=0;Na=n+2|0;i[Na>>0]=0;Sa=m+16|0;Wa=Sa;Qa=k[Wa>>2]|0;Oa=k[Wa+4>>2]|0;if((Oa|0)>0|(Oa|0)==0&Qa>>>0>0){Wa=Qa;Qa=Oa;Oa=0;while(1){Ra=my(Wa|0,Qa|0,1)|0;Pa=L;Ma=Oa+1<<24>>24;if((Pa|0)>0|(Pa|0)==0&Ra>>>0>0){Wa=Ra;Qa=Pa;Oa=Ma}else{pb=Ma;qb=Ra;rb=Pa;break}}Oa=Sa;k[Oa>>2]=qb;k[Oa+4>>2]=rb;i[Na>>0]=pb;sb=pb}else sb=0;Oa=k[30757]|0;if((Oa+3|0)>(k[30756]|0)){k[f>>2]=Oa;tk(783092);Ya=413;break}Qa=e+Oa|0;i[Qa>>0]=i[n>>0]|0;i[Qa+1>>0]=i[n+1>>0]|0;i[Qa+2>>0]=i[n+2>>0]|0;k[30757]=(k[30757]|0)+3;k[30758]=0;k[30759]=8;Qa=99096;k[Qa>>2]=0;k[Qa+4>>2]=0;Qa=Ot(e,a,b,fb,gb,eb&255)|0;do if(!Qa){Oa=(b|0)/2|0;Wa=ob&255;Xa=Ot(e,a+(gb<<3)|0,b,fb,Oa,Wa)|0;if(Xa){tb=Xa;break}Xa=ha(fb,b)|0;Va=(c|0)/2|0;Pa=Ot(e,a+(Xa<<3)|0,b,Va,gb,Wa)|0;if(Pa){tb=Pa;break}tb=Ot(e,a+(Xa+gb<<3)|0,b,Va,Oa,sb&255)|0}else tb=Qa;while(0);Qa=k[30758]<<4;k[30758]=Qa;Na=k[30759]|0;Sa=Na+-4|0;k[30759]=Sa;if((Sa|0)<1){i[e+(k[30757]|0)>>0]=Qa>>4-Na;Na=k[30757]|0;if((Na|0)<(k[30756]|0))k[30757]=Na+1;Na=(k[30759]|0)+8|0;k[30759]=Na;ub=Na}else ub=Sa;Sa=99096;Na=ky(k[Sa>>2]|0,k[Sa+4>>2]|0,4,0)|0;Sa=99096;k[Sa>>2]=Na;k[Sa+4>>2]=L;if((ub|0)<8){i[e+(k[30757]|0)>>0]=k[30758]<<ub;Sa=k[30757]|0;if((Sa|0)<(k[30756]|0))k[30757]=Sa+1;Sa=k[30759]|0;Na=99096;Qa=ky(k[Na>>2]|0,k[Na+4>>2]|0,Sa|0,((Sa|0)<0)<<31>>31|0)|0;Sa=99096;k[Sa>>2]=Qa;k[Sa+4>>2]=L}do if((hb|0)>0){Sa=k[30757]|0;if((Sa+hb|0)>(k[30756]|0)){dy(d);k[f>>2]=k[30757];tk(783092);Ya=413;break a}else{py(e+Sa|0,d|0,hb|0)|0;k[30757]=(k[30757]|0)+hb;break}}while(0);dy(d);k[f>>2]=k[30757];if((k[30757]|0)<(k[30756]|0)){Ya=tb;break}tk(783184);Ya=413}while(0);k[g>>2]=Ya;p=Ya;r=h;return p|0}function Mt(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;d=ny(b|0,c|0,8)|0;e=ny(b|0,c|0,16)|0;f=ny(b|0,c|0,24)|0;g=ny(b|0,c|0,40)|0;h=ny(b|0,c|0,48)|0;j=k[30757]|0;l=k[30756]|0;if((j|0)<(l|0)){m=ny(b|0,c|0,56)|0;i[a+j>>0]=m;m=(k[30757]|0)+1|0;k[30757]=m;n=m;o=k[30756]|0}else{n=j;o=l}if((n|0)<(o|0)){i[a+n>>0]=h;h=(k[30757]|0)+1|0;k[30757]=h;p=h;q=k[30756]|0}else{p=n;q=o}if((p|0)<(q|0)){i[a+p>>0]=g;g=(k[30757]|0)+1|0;k[30757]=g;r=g;s=k[30756]|0}else{r=p;s=q}if((r|0)<(s|0)){i[a+r>>0]=c;c=(k[30757]|0)+1|0;k[30757]=c;t=c;u=k[30756]|0}else{t=r;u=s}if((t|0)<(u|0)){i[a+t>>0]=f;f=(k[30757]|0)+1|0;k[30757]=f;v=f;w=k[30756]|0}else{v=t;w=u}if((v|0)<(w|0)){i[a+v>>0]=e;e=(k[30757]|0)+1|0;k[30757]=e;x=e;y=k[30756]|0}else{x=v;y=w}if((x|0)<(y|0)){i[a+x>>0]=d;d=(k[30757]|0)+1|0;k[30757]=d;z=d;A=k[30756]|0}else{z=x;A=y}if((z|0)>=(A|0))return;i[a+z>>0]=b;k[30757]=(k[30757]|0)+1;return}function Nt(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;g=(d|0)>(e|0)?d:e;h=~~(+fa(+(+(g|0)))/.6931471805599453+.5);j=((g|0)>(1<<h|0)&1)+h|0;h=d+1|0;g=e+1|0;m=ha((g|0)/2|0,(h|0)/2|0)|0;n=(m+1|0)/2|0;o=cy(n<<1)|0;p=cy(n)|0;if((o|0)==0|(p|0)==0){tk(783254);q=413;return q|0}if((f|0)>0){r=h>>1;h=g>>1;g=ha(h,r)|0;s=(g|0)>0;t=(j|0)>1;u=f;do{f=u;u=u+-1|0;k[30760]=0;k[30761]=0;St(b,c,d,e,o,u);a:do if(s){v=0;w=0;while(1){x=i[o+w>>0]|0;if(x<<24>>24!=0?(y=x&255,x=k[30761]|0,z=k[30760]|k[123048+(y<<2)>>2]<<x,k[30760]=z,A=(k[123112+(y<<2)>>2]|0)+x|0,k[30761]=A,(A|0)>7):0){i[p+v>>0]=z;x=v+1|0;if((x|0)>=(n|0))break;k[30760]=z>>8;k[30761]=A+-8;B=x}else B=v;w=w+1|0;if((w|0)>=(g|0)){C=B;D=11;break a}else v=B}v=k[30758]<<4;k[30758]=v;w=k[30759]|0;x=w+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=v>>4-w;w=k[30757]|0;if((w|0)<(k[30756]|0))k[30757]=w+1;k[30759]=(k[30759]|0)+8}w=99096;v=ky(k[w>>2]|0,k[w+4>>2]|0,4,0)|0;w=99096;k[w>>2]=v;k[w+4>>2]=L;St(b,c,d,e,o,u);Qt(a,m,o)}else{C=0;D=11}while(0);b:do if((D|0)==11){D=0;c:do if(t){w=C;v=1;x=r;A=h;d:while(1){Rt(o,A,x,A,o);x=x+1>>1;A=A+1>>1;z=ha(A,x)|0;if((z|0)>0){y=w;E=0;while(1){F=i[o+E>>0]|0;if(F<<24>>24!=0?(G=F&255,F=k[30761]|0,H=k[30760]|k[123048+(G<<2)>>2]<<F,k[30760]=H,I=(k[123112+(G<<2)>>2]|0)+F|0,k[30761]=I,(I|0)>7):0){i[p+y>>0]=H;F=y+1|0;if((F|0)>=(n|0))break d;k[30760]=H>>8;k[30761]=I+-8;J=F}else J=y;E=E+1|0;if((E|0)>=(z|0)){K=J;break}else y=J}}else K=w;v=v+1|0;if((v|0)>=(j|0)){M=K;break c}else w=K}w=k[30758]<<4;k[30758]=w;v=k[30759]|0;x=v+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=w>>4-v;v=k[30757]|0;if((v|0)<(k[30756]|0))k[30757]=v+1;k[30759]=(k[30759]|0)+8}v=99096;w=ky(k[v>>2]|0,k[v+4>>2]|0,4,0)|0;v=99096;k[v>>2]=w;k[v+4>>2]=L;St(b,c,d,e,o,u);Qt(a,m,o);break b}else M=C;while(0);v=k[30758]<<4|15;k[30758]=v;w=k[30759]|0;x=w+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=v>>4-w;w=k[30757]|0;if((w|0)<(k[30756]|0))k[30757]=w+1;w=(k[30759]|0)+8|0;k[30759]=w;N=w}else N=x;x=99096;w=ky(k[x>>2]|0,k[x+4>>2]|0,4,0)|0;x=L;v=99096;k[v>>2]=w;k[v+4>>2]=x;v=k[30761]|0;A=(v|0)>0;if(!M)if(A){y=k[30760]&(1<<v)+-1&k[123176+(v<<2)>>2]|k[30758]<<v;k[30758]=y;z=N-v|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=y>>0-z;z=k[30757]|0;if((z|0)<(k[30756]|0))k[30757]=z+1;k[30759]=(k[30759]|0)+8;z=99096;O=k[z>>2]|0;P=k[z+4>>2]|0}else{O=w;P=x}z=ky(O|0,P|0,v|0,((v|0)<0)<<31>>31|0)|0;y=99096;k[y>>2]=z;k[y+4>>2]=L;break}else{y=k[30758]<<6|62;k[30758]=y;z=N+-6|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=y>>6-N;y=k[30757]|0;if((y|0)<(k[30756]|0))k[30757]=y+1;k[30759]=(k[30759]|0)+8;y=99096;Q=k[y>>2]|0;R=k[y+4>>2]|0}else{Q=w;R=x}y=ky(Q|0,R|0,6,0)|0;z=99096;k[z>>2]=y;k[z+4>>2]=L;break}if(A){A=k[30760]&(1<<v)+-1&k[123176+(v<<2)>>2]|k[30758]<<v;k[30758]=A;z=N-v|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=A>>0-z;A=k[30757]|0;if((A|0)<(k[30756]|0))k[30757]=A+1;A=(k[30759]|0)+8|0;k[30759]=A;y=99096;S=k[y>>2]|0;T=k[y+4>>2]|0;U=A}else{S=w;T=x;U=z}z=ky(S|0,T|0,v|0,((v|0)<0)<<31>>31|0)|0;v=L;A=99096;k[A>>2]=z;k[A+4>>2]=v;V=z;W=v;X=U}else{V=w;W=x;X=N}if((M|0)>0){x=X;w=V;v=W;z=M;while(1){A=z;z=z+-1|0;y=k[30758]<<8|(l[p+z>>0]|0);k[30758]=y;E=x+-8|0;k[30759]=E;if((E|0)<1){i[a+(k[30757]|0)>>0]=y>>8-x;y=k[30757]|0;if((y|0)<(k[30756]|0))k[30757]=y+1;y=(k[30759]|0)+8|0;k[30759]=y;F=99096;Y=k[F>>2]|0;Z=k[F+4>>2]|0;_=y}else{Y=w;Z=v;_=E}w=ky(Y|0,Z|0,8,0)|0;v=L;E=99096;k[E>>2]=w;k[E+4>>2]=v;if((A|0)<=1)break;else x=_}}}while(0)}while((f|0)>1)}dy(p);dy(o);q=0;return q|0}function Ot(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0;g=(d|0)>(e|0)?d:e;h=~~(+fa(+(+(g|0)))/.6931471805599453+.5);j=((g|0)>(1<<h|0)&1)+h|0;h=d+1|0;g=e+1|0;m=ha((g|0)/2|0,(h|0)/2|0)|0;n=(m+1|0)/2|0;o=cy(n<<1)|0;p=cy(n)|0;if((o|0)==0|(p|0)==0){tk(783218);q=413;return q|0}if((f|0)>0){r=h>>1;h=g>>1;g=ha(h,r)|0;s=(g|0)>0;t=(j|0)>1;u=f;do{f=u;u=u+-1|0;k[30760]=0;k[30761]=0;Pt(b,c,d,e,o,u);a:do if(s){v=0;w=0;while(1){x=i[o+w>>0]|0;if(x<<24>>24!=0?(y=x&255,x=k[30761]|0,z=k[30760]|k[123048+(y<<2)>>2]<<x,k[30760]=z,A=(k[123112+(y<<2)>>2]|0)+x|0,k[30761]=A,(A|0)>7):0){i[p+v>>0]=z;x=v+1|0;if((x|0)>=(n|0))break;k[30760]=z>>8;k[30761]=A+-8;B=x}else B=v;w=w+1|0;if((w|0)>=(g|0)){C=B;D=11;break a}else v=B}v=k[30758]<<4;k[30758]=v;w=k[30759]|0;x=w+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=v>>4-w;w=k[30757]|0;if((w|0)<(k[30756]|0))k[30757]=w+1;k[30759]=(k[30759]|0)+8}w=99096;v=ky(k[w>>2]|0,k[w+4>>2]|0,4,0)|0;w=99096;k[w>>2]=v;k[w+4>>2]=L;Pt(b,c,d,e,o,u);Qt(a,m,o)}else{C=0;D=11}while(0);b:do if((D|0)==11){D=0;c:do if(t){w=C;v=1;x=r;A=h;d:while(1){Rt(o,A,x,A,o);x=x+1>>1;A=A+1>>1;z=ha(A,x)|0;if((z|0)>0){y=w;E=0;while(1){F=i[o+E>>0]|0;if(F<<24>>24!=0?(G=F&255,F=k[30761]|0,H=k[30760]|k[123048+(G<<2)>>2]<<F,k[30760]=H,I=(k[123112+(G<<2)>>2]|0)+F|0,k[30761]=I,(I|0)>7):0){i[p+y>>0]=H;F=y+1|0;if((F|0)>=(n|0))break d;k[30760]=H>>8;k[30761]=I+-8;J=F}else J=y;E=E+1|0;if((E|0)>=(z|0)){K=J;break}else y=J}}else K=w;v=v+1|0;if((v|0)>=(j|0)){M=K;break c}else w=K}w=k[30758]<<4;k[30758]=w;v=k[30759]|0;x=v+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=w>>4-v;v=k[30757]|0;if((v|0)<(k[30756]|0))k[30757]=v+1;k[30759]=(k[30759]|0)+8}v=99096;w=ky(k[v>>2]|0,k[v+4>>2]|0,4,0)|0;v=99096;k[v>>2]=w;k[v+4>>2]=L;Pt(b,c,d,e,o,u);Qt(a,m,o);break b}else M=C;while(0);v=k[30758]<<4|15;k[30758]=v;w=k[30759]|0;x=w+-4|0;k[30759]=x;if((x|0)<1){i[a+(k[30757]|0)>>0]=v>>4-w;w=k[30757]|0;if((w|0)<(k[30756]|0))k[30757]=w+1;w=(k[30759]|0)+8|0;k[30759]=w;N=w}else N=x;x=99096;w=ky(k[x>>2]|0,k[x+4>>2]|0,4,0)|0;x=L;v=99096;k[v>>2]=w;k[v+4>>2]=x;v=k[30761]|0;A=(v|0)>0;if(!M)if(A){y=k[30760]&(1<<v)+-1&k[123176+(v<<2)>>2]|k[30758]<<v;k[30758]=y;z=N-v|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=y>>0-z;z=k[30757]|0;if((z|0)<(k[30756]|0))k[30757]=z+1;k[30759]=(k[30759]|0)+8;z=99096;O=k[z>>2]|0;P=k[z+4>>2]|0}else{O=w;P=x}z=ky(O|0,P|0,v|0,((v|0)<0)<<31>>31|0)|0;y=99096;k[y>>2]=z;k[y+4>>2]=L;break}else{y=k[30758]<<6|62;k[30758]=y;z=N+-6|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=y>>6-N;y=k[30757]|0;if((y|0)<(k[30756]|0))k[30757]=y+1;k[30759]=(k[30759]|0)+8;y=99096;Q=k[y>>2]|0;R=k[y+4>>2]|0}else{Q=w;R=x}y=ky(Q|0,R|0,6,0)|0;z=99096;k[z>>2]=y;k[z+4>>2]=L;break}if(A){A=k[30760]&(1<<v)+-1&k[123176+(v<<2)>>2]|k[30758]<<v;k[30758]=A;z=N-v|0;k[30759]=z;if((z|0)<1){i[a+(k[30757]|0)>>0]=A>>0-z;A=k[30757]|0;if((A|0)<(k[30756]|0))k[30757]=A+1;A=(k[30759]|0)+8|0;k[30759]=A;y=99096;S=k[y>>2]|0;T=k[y+4>>2]|0;U=A}else{S=w;T=x;U=z}z=ky(S|0,T|0,v|0,((v|0)<0)<<31>>31|0)|0;v=L;A=99096;k[A>>2]=z;k[A+4>>2]=v;V=z;W=v;X=U}else{V=w;W=x;X=N}if((M|0)>0){x=X;w=V;v=W;z=M;while(1){A=z;z=z+-1|0;y=k[30758]<<8|(l[p+z>>0]|0);k[30758]=y;E=x+-8|0;k[30759]=E;if((E|0)<1){i[a+(k[30757]|0)>>0]=y>>8-x;y=k[30757]|0;if((y|0)<(k[30756]|0))k[30757]=y+1;y=(k[30759]|0)+8|0;k[30759]=y;F=99096;Y=k[F>>2]|0;Z=k[F+4>>2]|0;_=y}else{Y=w;Z=v;_=E}w=ky(Y|0,Z|0,8,0)|0;v=L;E=99096;k[E>>2]=w;k[E+4>>2]=v;if((A|0)<=1)break;else x=_}}}while(0)}while((f|0)>1)}dy(p);dy(o);q=0;return q|0}function Pt(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0;g=ty(1,0,f|0)|0;h=L;j=ty(g|0,h|0,1)|0;l=L;m=ty(g|0,h|0,2)|0;n=L;o=ty(g|0,h|0,3)|0;p=L;q=c+-1|0;if((c|0)>1){r=d+-1|0;s=(d|0)>1;t=(d+-2|0)>>>1;u=t<<1;v=u+2|0;w=b<<1;x=t+1|0;t=0;y=u+b+2|0;u=v;z=0;while(1){A=ha(t,b)|0;B=A+b|0;if(s){C=0;D=z;E=A;F=B;while(1){G=a+(F+1<<3)|0;H=k[G>>2]&g;I=k[G+4>>2]&h;G=a+(F<<3)|0;J=ty(k[G>>2]|0,k[G+4>>2]|0,1)|0;G=L&l|I;I=a+((E|1)<<3)|0;K=ty(k[I>>2]|0,k[I+4>>2]|0,2)|0;I=G|L&n;G=a+(E<<3)|0;M=ty(k[G>>2]|0,k[G+4>>2]|0,3)|0;G=my(J&j|H|K&m|M&o|0,I|L&p|0,f|0)|0;i[e+D>>0]=G;C=C+2|0;if((C|0)>=(r|0))break;else{D=D+1|0;E=E+2|0;F=F+2|0}}N=v;O=x+z|0;P=u;Q=y}else{N=0;O=z;P=A;Q=B}if((N|0)<(d|0)){F=a+(Q<<3)|0;E=ty(k[F>>2]|0,k[F+4>>2]|0,1)|0;F=L&l;D=a+(P<<3)|0;C=ty(k[D>>2]|0,k[D+4>>2]|0,3)|0;D=my(C&o|E&j|0,L&p|F|0,f|0)|0;i[e+O>>0]=D;R=O+1|0}else R=O;t=t+2|0;if((t|0)>=(q|0)){S=R;break}else{y=y+w|0;u=u+w|0;z=R}}T=c&-2;U=S}else{T=0;U=0}if((T|0)>=(c|0))return;c=ha(T,b)|0;b=d+-1|0;if((d|0)>1){T=(d+-2|0)>>>1;S=T<<1;R=c+S|0;z=U+T|0;T=0;w=U;u=c;while(1){y=a+(u+1<<3)|0;q=ty(k[y>>2]|0,k[y+4>>2]|0,2)|0;y=L&n;t=a+(u<<3)|0;O=ty(k[t>>2]|0,k[t+4>>2]|0,3)|0;t=my(O&o|q&m|0,L&p|y|0,f|0)|0;i[e+w>>0]=t;T=T+2|0;if((T|0)>=(b|0))break;else{w=w+1|0;u=u+2|0}}V=S+2|0;W=z+1|0;X=R+2|0}else{V=0;W=U;X=c}if((V|0)>=(d|0))return;d=a+(X<<3)|0;X=ty(k[d>>2]|0,k[d+4>>2]|0,3)|0;d=my(X&o|0,L&p|0,f|0)|0;i[e+W>>0]=d;return}function Qt(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0;if((b|0)==1){d=k[30758]<<4|(l[c>>0]|0)&15;k[30758]=d;e=k[30759]|0;f=e+-4|0;k[30759]=f;if((f|0)<1){i[a+(k[30757]|0)>>0]=d>>4-e;e=k[30757]|0;if((e|0)<(k[30756]|0))k[30757]=e+1;k[30759]=(k[30759]|0)+8}e=99096;d=ky(k[e>>2]|0,k[e+4>>2]|0,4,0)|0;e=99096;k[e>>2]=d;k[e+4>>2]=L;return}e=k[30759]|0;if((e|0)<5){d=k[30758]<<4|(l[c>>0]|0)&15;k[30758]=d;f=e+-4|0;k[30759]=f;if((f|0)<1){i[a+(k[30757]|0)>>0]=d>>4-e;d=k[30757]|0;if((d|0)<(k[30756]|0))k[30757]=d+1;d=(k[30759]|0)+8|0;k[30759]=d;g=d}else g=f;f=99096;d=ky(k[f>>2]|0,k[f+4>>2]|0,4,0)|0;f=L;h=99096;k[h>>2]=d;k[h+4>>2]=f;if((b|0)==2){h=k[30758]<<4|(l[c+1>>0]|0)&15;k[30758]=h;j=g+-4|0;k[30759]=j;if((j|0)<1){i[a+(k[30757]|0)>>0]=h>>4-g;h=k[30757]|0;if((h|0)<(k[30756]|0))k[30757]=h+1;k[30759]=(k[30759]|0)+8;h=99096;m=k[h>>2]|0;n=k[h+4>>2]|0}else{m=d;n=f}f=ky(m|0,n|0,4,0)|0;n=99096;k[n>>2]=f;k[n+4>>2]=L;return}else{o=g;p=1}}else{o=e;p=0}e=8-o|0;g=b-p|0;n=(g|0)/2|0;if((o|0)==8){k[30758]=0;if((g|0)>1){o=(n|0)>1?n:1;f=o<<1;m=k[30757]|0;d=0;h=p;while(1){i[a+m>>0]=(l[c+(h+1)>>0]|0)&15|(l[c+h>>0]|0)<<4;m=(k[30757]|0)+1|0;k[30757]=m;d=d+1|0;if((d|0)>=(n|0))break;else h=h+2|0}q=o;r=p|f}else{q=0;r=p}}else if((g|0)>1){g=(n|0)>1?n:1;f=g<<1;o=k[30757]|0;h=0;d=p;while(1){m=(l[c+d>>0]|0)<<4&240|k[30758]<<8|(l[c+(d+1)>>0]|0)&15;k[30758]=m;i[a+o>>0]=m>>e;o=(k[30757]|0)+1|0;k[30757]=o;h=h+1|0;if((h|0)>=(n|0))break;else d=d+2|0}q=g;r=p|f}else{q=0;r=p}p=(q<<3)+-8|0;q=99096;f=ky(p|0,((p|0)<0)<<31>>31|0,k[q>>2]|0,k[q+4>>2]|0)|0;q=L;p=99096;k[p>>2]=f;k[p+4>>2]=q;if((r|0)==(b|0))return;r=k[30758]<<4|(l[c+(b+-1)>>0]|0)&15;k[30758]=r;b=k[30759]|0;c=b+-4|0;k[30759]=c;if((c|0)<1){i[a+(k[30757]|0)>>0]=r>>4-b;b=k[30757]|0;if((b|0)<(k[30756]|0))k[30757]=b+1;k[30759]=(k[30759]|0)+8;b=99096;s=k[b>>2]|0;t=k[b+4>>2]|0}else{s=f;t=q}q=ky(s|0,t|0,4,0)|0;t=99096;k[t>>2]=q;k[t+4>>2]=L;return}function Rt(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,k=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;f=c+-1|0;if((c|0)>1){g=d+-1|0;h=(d|0)>1;j=(d+-2|0)>>>1;k=j<<1;l=k+2|0;m=b<<1;n=j+1|0;j=0;o=k+b+2|0;k=l;p=0;while(1){q=ha(j,b)|0;r=q+b|0;if(h){s=0;t=p;u=q;v=r;while(1){i[e+t>>0]=((i[a+v>>0]|0)!=0&1)<<1|(i[a+(v+1)>>0]|0)!=0|((i[a+(u|1)>>0]|0)!=0&1)<<2|((i[a+u>>0]|0)!=0&1)<<3;s=s+2|0;if((s|0)>=(g|0))break;else{t=t+1|0;u=u+2|0;v=v+2|0}}w=l;x=n+p|0;y=k;z=o}else{w=0;x=p;y=q;z=r}if((w|0)<(d|0)){i[e+x>>0]=((i[a+y>>0]|0)!=0&1)<<3|((i[a+z>>0]|0)!=0&1)<<1;A=x+1|0}else A=x;j=j+2|0;if((j|0)>=(f|0)){B=A;break}else{o=o+m|0;k=k+m|0;p=A}}C=c&-2;D=B}else{C=0;D=0}if((C|0)>=(c|0))return;c=ha(C,b)|0;b=d+-1|0;if((d|0)>1){C=(d+-2|0)>>>1;B=C<<1;A=c+B|0;p=D+C|0;C=0;m=D;k=c;while(1){i[e+m>>0]=((i[a+k>>0]|0)!=0&1)<<3|((i[a+(k+1)>>0]|0)!=0&1)<<2;C=C+2|0;if((C|0)>=(b|0))break;else{m=m+1|0;k=k+2|0}}E=B+2|0;F=p+1|0;G=A+2|0}else{E=0;F=D;G=c}if((E|0)>=(d|0))return;i[e+F>>0]=((i[a+G>>0]|0)!=0&1)<<3;return}function St(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0;g=1<<f;h=g<<1;j=g<<2;l=g<<3;m=c+-1|0;if((c|0)>1){n=d+-1|0;o=(d|0)>1;p=(d+-2|0)>>>1;q=p<<1;r=q+2|0;s=b<<1;t=p+1|0;p=0;u=q+b+2|0;q=r;v=0;while(1){w=ha(p,b)|0;x=w+b|0;if(o){y=0;z=v;A=w;B=x;while(1){i[e+z>>0]=(k[a+(B<<2)>>2]<<1&h|k[a+(B+1<<2)>>2]&g|k[a+((A|1)<<2)>>2]<<2&j|k[a+(A<<2)>>2]<<3&l)>>f;y=y+2|0;if((y|0)>=(n|0))break;else{z=z+1|0;A=A+2|0;B=B+2|0}}C=r;D=t+v|0;E=q;F=u}else{C=0;D=v;E=w;F=x}if((C|0)<(d|0)){i[e+D>>0]=(k[a+(E<<2)>>2]<<3&l|k[a+(F<<2)>>2]<<1&h)>>f;G=D+1|0}else G=D;p=p+2|0;if((p|0)>=(m|0)){H=G;break}else{u=u+s|0;q=q+s|0;v=G}}I=c&-2;J=H}else{I=0;J=0}if((I|0)>=(c|0))return;c=ha(I,b)|0;b=d+-1|0;if((d|0)>1){I=(d+-2|0)>>>1;H=I<<1;G=c+H|0;v=J+I|0;I=0;s=J;q=c;while(1){i[e+s>>0]=(k[a+(q<<2)>>2]<<3&l|k[a+(q+1<<2)>>2]<<2&j)>>f;I=I+2|0;if((I|0)>=(b|0))break;else{s=s+1|0;q=q+2|0}}K=H+2|0;L=v+1|0;M=G+2|0}else{K=0;L=J;M=c}if((K|0)>=(d|0))return;i[e+L>>0]=(k[a+(M<<2)>>2]<<3&l)>>f;return}function Tt(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,L=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0;h=r;r=r+16|0;m=h;n=k[g>>2]|0;if((n|0)>0){o=n;r=h;return o|0}k[30803]=0;j[m>>1]=l[a>>0]|l[a+1>>0]<<8;k[30803]=2;if(lx(m,783288,2)|0){tk(783290);k[g>>2]=414;o=414;r=h;return o|0}m=i[a+2>>0]|0;k[30803]=3;n=i[a+3>>0]|0;k[30803]=4;p=i[a+4>>0]|0;k[30803]=5;q=i[a+5>>0]|0;k[30803]=6;k[e>>2]=((n&255|(m&255)<<8)<<8|p&255)<<8|q&255;q=i[a+6>>0]|0;k[30803]=7;p=i[a+7>>0]|0;k[30803]=8;m=i[a+8>>0]|0;k[30803]=9;n=i[a+9>>0]|0;k[30803]=10;k[d>>2]=((p&255|(q&255)<<8)<<8|m&255)<<8|n&255;n=i[a+10>>0]|0;k[30803]=11;m=i[a+11>>0]|0;k[30803]=12;q=i[a+12>>0]|0;k[30803]=13;k[f>>2]=((m&255|(n&255)<<8)<<8|q&255)<<8|(l[a+13>>0]|0);k[30803]=18;q=i[a+18>>0]|0;k[30803]=19;n=i[a+19>>0]|0;k[30803]=20;m=i[a+20>>0]|0;k[30803]=21;p=i[a+21>>0]|0;k[30803]=22;s=(((q&255)<<8|n&255)<<8|m&255)<<8|p&255;p=i[a+22>>0]|0;m=i[a+23>>0]|0;n=i[a+24>>0]|0;k[30803]=25;q=k[e>>2]|0;t=k[d>>2]|0;u=ha(t,q)|0;v=(q+1|0)/2|0;w=(t+1|0)/2|0;x=(u|0)>0;if(x)sy(c|0,0,u<<2|0)|0;k[30804]=0;y=Vt(a,c,t,v,w,p&255)|0;do if(!y){p=(t|0)/2|0;z=m&255;A=Vt(a,c+(w<<2)|0,t,v,p,z)|0;if(!A){B=ha(v,t)|0;C=(q|0)/2|0;D=Vt(a,c+(B<<2)|0,t,C,w,z)|0;if(!D){z=Vt(a,c+(B+w<<2)|0,t,C,p,n&255)|0;if(!z){p=k[30804]|0;C=k[30805]|0;if((p|0)<4){B=k[30803]|0;E=l[a+B>>0]|0|C<<8;k[30805]=E;k[30803]=B+1;B=p+8|0;k[30804]=B;F=B;G=E}else{F=p;G=C}C=F+-4|0;k[30804]=C;if(G>>C&15){tk(783306);H=414;break}k[30804]=0;if(x){C=0;do{p=c+(C<<2)|0;if(k[p>>2]|0){E=k[30804]|0;if(!E){B=k[30803]|0;I=l[a+B>>0]|0;k[30805]=I;k[30803]=B+1;k[30804]=8;J=7;K=I}else{J=E+-1|0;K=k[30805]|0}k[30804]=J;if(1<<J&K)k[p>>2]=0-(k[p>>2]|0)}C=C+1|0}while((C|0)!=(u|0))}k[c>>2]=s;k[g>>2]=0;C=k[f>>2]|0;if((C|0)>=2?(p=ha(k[e>>2]|0,k[d>>2]|0)|0,E=c+(p+-1<<2)|0,(p|0)>=1):0){p=c;do{I=ha(k[p>>2]|0,C)|0;k[p>>2]=I;p=p+4|0}while(p>>>0<=E>>>0);L=k[f>>2]|0}else L=C;E=k[e>>2]|0;p=k[d>>2]|0;I=(E|0)>(p|0)?E:p;B=~~(+fa(+(+(I|0)))/.6931471805599453+.5);M=((I|0)>(1<<B|0)&1)+B|0;B=cy(((I+1|0)/2|0)<<2)|0;if(!B){tk(783337);N=414}else{I=M+-1|0;O=1<<I;P=O<<1;Q=0-O|0;R=P>>1;S=k[c>>2]|0;k[c>>2]=S+(O<<2>>1)+(S>>31)&Q<<2;if((M|0)>0){S=L>>1;T=p<<1;U=0-S|0;V=0-p|0;W=0-T|0;X=p+2|0;Y=(b|0)==0|(S|0)<1;Z=O;_=P;P=1<<M;M=I;I=Q;$=Q<<1;Q=R+-1|0;aa=E;E=1;ba=p;ca=1;da=O;O=R;R=1;while(1){ea=O;ga=da>>1;ia=P>>1;ja=E<<1;ka=ca<<1;la=(aa|0)>(ia|0);ma=aa-(la?ia:0)|0;na=(la^1)<<31>>31;la=na+ja|0;oa=(ba|0)>(ia|0);pa=ba-(oa?ia:0)|0;qa=(oa^1)<<31>>31;oa=qa+ka|0;ra=(M|0)==0;sa=ra?2:R;ta=Q;ua=ra?0:ga+-1|0;ra=(la|0)>0;if(ra){va=oa+1>>1;wa=(va|0)<(oa|0);xa=va+-1|0;ya=(va|0)>0;za=xa<<1;Aa=(oa|0)>1;Ba=0;do{Ca=ha(Ba,p)|0;if(wa){Da=va;Ea=c+(Ca+va<<2)|0;Fa=B;while(1){k[Fa>>2]=k[Ea>>2];Da=Da+1|0;if((Da|0)==(oa|0))break;else{Ea=Ea+4|0;Fa=Fa+4|0}}}if(ya){Fa=xa;Ea=c+(Ca+za<<2)|0;Da=c+(Ca+xa<<2)|0;while(1){k[Ea>>2]=k[Da>>2];if((Fa|0)>0){Fa=Fa+-1|0;Ea=Ea+-8|0;Da=Da+-4|0}else break}}if(Aa){Da=1;Ea=c+(Ca+1<<2)|0;Fa=B;while(1){k[Ea>>2]=k[Fa>>2];Da=Da+2|0;if((Da|0)>=(oa|0))break;else{Ea=Ea+8|0;Fa=Fa+4|0}}}Ba=Ba+1|0}while((Ba|0)!=(la|0))}Ba=(oa|0)>0;if(Ba){Aa=la+1>>1;xa=(Aa|0)<(la|0);za=ha(Aa,p)|0;ya=Aa+-1|0;va=ha(ya,p)|0;wa=(Aa|0)>0;Fa=va<<1;Ea=(la|0)>1;Da=0;do{if(xa){Ca=Aa;Ga=c+(Da+za<<2)|0;Ha=B;while(1){k[Ha>>2]=k[Ga>>2];Ca=Ca+1|0;if((Ca|0)==(la|0))break;else{Ga=Ga+(p<<2)|0;Ha=Ha+4|0}}}if(wa){Ha=ya;Ga=c+(Da+Fa<<2)|0;Ca=c+(Da+va<<2)|0;while(1){k[Ga>>2]=k[Ca>>2];if((Ha|0)>0){Ha=Ha+-1|0;Ga=Ga+(W<<2)|0;Ca=Ca+(V<<2)|0}else break}}if(Ea){Ca=1;Ga=c+(Da+p<<2)|0;Ha=B;while(1){k[Ga>>2]=k[Ha>>2];Ca=Ca+2|0;if((Ca|0)>=(la|0))break;else{Ga=Ga+(T<<2)|0;Ha=Ha+4|0}}}Da=Da+1|0}while((Da|0)!=(oa|0))}do if(!Y){Da=la+-2|0;Ea=(Da|0)>2;if(Ea){va=2;do{Fa=ha(va,p)|0;if(Ba){ya=0;wa=Fa;za=Fa+p|0;while(1){Fa=k[c+(wa-T<<2)>>2]|0;Aa=k[c+(wa<<2)>>2]|0;xa=k[c+(wa+T<<2)>>2]|0;Ha=xa-Fa|0;Ga=xa-Aa|0;xa=Aa-Fa|0;Fa=(Ga|0)<(xa|0)?Ga:xa;Aa=(Fa|0)>0?Fa<<2:0;Fa=(Ga|0)>(xa|0)?Ga:xa;xa=(Fa|0)<0?Fa<<2:0;if((xa|0)<(Aa|0)){Fa=(Ha|0)<(Aa|0)?Ha:Aa;Aa=c+(za<<2)|0;Ha=k[Aa>>2]|0;Ga=((Fa|0)>(xa|0)?Fa:xa)-(Ha<<3)|0;xa=((Ga|0)>-1?Ga:Ga+7|0)>>3;Ga=(xa|0)<(S|0)?xa:S;k[Aa>>2]=((Ga|0)>(U|0)?Ga:U)+Ha}ya=ya+2|0;if((ya|0)>=(oa|0))break;else{wa=wa+2|0;za=za+2|0}}}va=va+2|0}while((va|0)<(Da|0))}if(ra){va=oa+-2|0;za=(va|0)>2;wa=0;do{if(za){ya=ha(wa,p)|0;Ha=ya+2|0;Ga=k[c+(ya<<2)>>2]|0;ya=k[c+(Ha<<2)>>2]|0;Aa=2;xa=Ha;while(1){Ha=xa;xa=xa+2|0;Fa=k[c+(xa<<2)>>2]|0;Ca=Fa-Ga|0;Ia=Fa-ya|0;Ja=ya-Ga|0;Ka=(Ia|0)<(Ja|0)?Ia:Ja;La=(Ka|0)>0?Ka<<2:0;Ka=(Ia|0)>(Ja|0)?Ia:Ja;Ja=(Ka|0)<0?Ka<<2:0;if((Ja|0)<(La|0)){Ka=(Ca|0)<(La|0)?Ca:La;La=c+(Ha+1<<2)|0;Ha=k[La>>2]|0;Ca=((Ka|0)>(Ja|0)?Ka:Ja)-(Ha<<3)|0;Ja=((Ca|0)>-1?Ca:Ca+7|0)>>3;Ca=(Ja|0)<(S|0)?Ja:S;k[La>>2]=((Ca|0)>(U|0)?Ca:U)+Ha}Aa=Aa+2|0;if((Aa|0)>=(va|0))break;else{Ha=ya;ya=Fa;Ga=Ha}}}wa=wa+2|0}while((wa|0)<(la|0))}if(!Ea)break;wa=oa+-2|0;va=(wa|0)>2;za=2;do{Ga=(ha(za,p)|0)+2|0;if(va){ya=2;Aa=Ga;xa=Ga+p|0;while(1){Ga=Aa-T|0;Ha=k[c+(Ga+-2<<2)>>2]|0;Fa=Aa+T|0;Ca=k[c+(Fa+-2<<2)>>2]|0;La=k[c+(Ga+2<<2)>>2]|0;Ga=k[c+(Fa+2<<2)>>2]|0;Fa=k[c+(Aa<<2)>>2]|0;Ja=Ha-Ca-La+Ga|0;Ka=k[c+(xa<<2)>>2]<<1;Ia=k[c+(Aa+1<<2)>>2]<<1;Ma=Ga-Fa|0;Ga=((Ma|0)>0?Ma:0)-Ka|0;Na=Fa-Ca|0;Ca=((Na|0)>0?Na:0)+Ka|0;Oa=((Ga|0)<(Ca|0)?Ga:Ca)-Ia|0;Ca=Fa-La|0;La=((Ca|0)>0?Ca:0)-Ka|0;Ga=Ha-Fa|0;Fa=((Ga|0)>0?Ga:0)+Ka|0;Ha=((La|0)<(Fa|0)?La:Fa)+Ia|0;Fa=((Oa|0)<(Ha|0)?Oa:Ha)<<4;Ha=((Ma|0)<0?Ma:0)-Ka|0;Ma=((Na|0)<0?Na:0)+Ka|0;Na=((Ha|0)>(Ma|0)?Ha:Ma)-Ia|0;Ma=((Ca|0)<0?Ca:0)-Ka|0;Ca=((Ga|0)<0?Ga:0)+Ka|0;Ka=((Ma|0)>(Ca|0)?Ma:Ca)+Ia|0;Ia=((Na|0)>(Ka|0)?Na:Ka)<<4;if((Ia|0)<(Fa|0)){Ka=(Ja|0)<(Fa|0)?Ja:Fa;Fa=c+(xa+1<<2)|0;Ja=k[Fa>>2]|0;Na=((Ka|0)>(Ia|0)?Ka:Ia)-(Ja<<6)|0;Ia=((Na|0)>-1?Na:Na+63|0)>>6;Na=(Ia|0)<(S|0)?Ia:S;k[Fa>>2]=((Na|0)>(U|0)?Na:U)+Ja}ya=ya+2|0;if((ya|0)>=(wa|0))break;else{Aa=Aa+2|0;xa=xa+2|0}}}za=za+2|0}while((za|0)<(Da|0))}while(0);ra=(la|0)%2|0;Ba=(oa|0)%2|0;Da=la-ra|0;if((Da|0)>0){za=oa-Ba|0;wa=(za|0)>0;va=(Ba|0)==0;Ea=ka+-1+qa-Ba&-2;xa=0;Aa=X+Ea|0;ya=Ea+2|0;while(1){Ea=ha(xa,p)|0;Ja=Ea+p|0;if(wa){Na=0;Fa=Ea;Ia=Ja;while(1){Ka=c+(Fa<<2)|0;Ca=k[Ka>>2]|0;Ma=c+(Ia<<2)|0;Ga=k[Ma>>2]|0;Ha=c+((Fa|1)<<2)|0;Oa=k[Ha>>2]|0;La=c+(Ia+1<<2)|0;Pa=k[La>>2]|0;Qa=((Ga|0)>-1?ea:ta)+Ga&$;Ga=((Oa|0)>-1?ea:ta)+Oa&$;Oa=((Pa|0)>-1?ga:ua)+Pa&I;Pa=Oa&Z;Ra=0-Pa|0;Sa=((Qa|0)>-1?Ra:Pa)+Qa|0;Qa=((Ga|0)>-1?Ra:Pa)+Ga|0;Ga=(Sa^Oa^Qa)&_;if((Ca|0)>-1)Ta=Pa+Ca-Ga|0;else Ta=((Pa|0)==0?Ga:Pa-Ga|0)+Ca|0;Ca=Ta+Sa|0;Ga=Qa+Oa|0;k[La>>2]=Ca+Ga>>sa;k[Ma>>2]=Ca-Ga>>sa;Ga=Ta-Sa|0;k[Ha>>2]=Qa-Oa+Ga>>sa;k[Ka>>2]=Oa-Qa+Ga>>sa;Na=Na+2|0;if((Na|0)>=(za|0)){Ua=ya;Va=Aa;break}else{Fa=Fa+2|0;Ia=Ia+2|0}}}else{Ua=Ea;Va=Ja}if(!va){Ia=c+(Ua<<2)|0;Fa=k[Ia>>2]|0;Na=c+(Va<<2)|0;Ga=k[Na>>2]|0;Qa=((Ga|0)>-1?ea:ta)+Ga&$;Ga=Qa&_;Oa=((Fa|0)>-1?0-Ga|0:Ga)+Fa|0;k[Na>>2]=Oa+Qa>>sa;k[Ia>>2]=Oa-Qa>>sa}xa=xa+2|0;if((xa|0)>=(Da|0))break;else{Aa=Aa+T|0;ya=ya+T|0}}Wa=(ja|1)+na-ra&-2}else Wa=0;do if(ra){ya=ha(Wa,p)|0;Aa=oa-Ba|0;if((Aa|0)>0){Da=ka+-1+qa|0;xa=0;va=ya;while(1){za=c+(va<<2)|0;wa=k[za>>2]|0;Qa=c+(va+1<<2)|0;Oa=k[Qa>>2]|0;Ia=((Oa|0)>-1?ea:ta)+Oa&$;Oa=Ia&_;Na=((wa|0)>-1?0-Oa|0:Oa)+wa|0;k[Qa>>2]=Na+Ia>>sa;k[za>>2]=Na-Ia>>sa;xa=xa+2|0;if((xa|0)>=(Aa|0))break;else va=va+2|0}Xa=(Da-Ba+2&-2)+ya|0}else Xa=ya;if(!Ba)break;va=c+(Xa<<2)|0;k[va>>2]=k[va>>2]>>sa}while(0);if((M|0)>0){Ba=I;ta=Z;Z=Z>>1;P=ia;M=M+-1|0;I=I>>1;Q=ua;aa=ma;E=la;ba=pa;ca=oa;da=ga;O=ga;R=sa;$=Ba;_=ta}else break}}dy(B);N=0}k[g>>2]=N;o=N;r=h;return o|0}else H=z}else H=D}else H=A}else H=y;while(0);k[c>>2]=s;k[g>>2]=H;o=H;r=h;return o|0}function Ut(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,m=0,n=0,o=0,p=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0,S=0,T=0,U=0,V=0,W=0,X=0,Y=0,Z=0,_=0,$=0,aa=0,ba=0,ca=0,da=0,ea=0,ga=0,ia=0,ja=0,ka=0,la=0,ma=0,na=0,oa=0,pa=0,qa=0,ra=0,sa=0,ta=0,ua=0,va=0,wa=0,xa=0,ya=0,za=0,Aa=0,Ba=0,Ca=0,Da=0,Ea=0,Fa=0,Ga=0,Ha=0,Ia=0,Ja=0,Ka=0,La=0,Ma=0,Na=0,Oa=0,Pa=0,Qa=0,Ra=0,Sa=0,Ta=0,Ua=0,Va=0,Wa=0,Xa=0,Ya=0,Za=0,_a=0,$a=0,ab=0,bb=0,cb=0,db=0,eb=0,fb=0,gb=0,hb=0,ib=0,jb=0,kb=0,lb=0,mb=0,nb=0,ob=0,pb=0,qb=0,rb=0,sb=0,tb=0,ub=0,vb=0,wb=0,xb=0;h=r;r=r+16|0;m=h;n=k[g>>2]|0;if((n|0)>0){o=n;r=h;return o|0}k[30803]=0;j[m>>1]=l[a>>0]|l[a+1>>0]<<8;k[30803]=2;if(lx(m,783288,2)|0){tk(783290);k[g>>2]=414;o=414;r=h;return o|0}m=i[a+2>>0]|0;k[30803]=3;n=i[a+3>>0]|0;k[30803]=4;p=i[a+4>>0]|0;k[30803]=5;q=i[a+5>>0]|0;k[30803]=6;k[e>>2]=((n&255|(m&255)<<8)<<8|p&255)<<8|q&255;q=i[a+6>>0]|0;k[30803]=7;p=i[a+7>>0]|0;k[30803]=8;m=i[a+8>>0]|0;k[30803]=9;n=i[a+9>>0]|0;k[30803]=10;k[d>>2]=((p&255|(q&255)<<8)<<8|m&255)<<8|n&255;n=i[a+10>>0]|0;k[30803]=11;m=i[a+11>>0]|0;k[30803]=12;q=i[a+12>>0]|0;k[30803]=13;p=i[a+13>>0]|0;k[30803]=14;k[f>>2]=((m&255|(n&255)<<8)<<8|q&255)<<8|p&255;p=i[a+14>>0]|0;k[30803]=15;q=i[a+15>>0]|0;k[30803]=16;n=i[a+16>>0]|0;k[30803]=17;m=i[a+17>>0]|0;k[30803]=18;s=i[a+18>>0]|0;k[30803]=19;t=i[a+19>>0]|0;k[30803]=20;u=i[a+20>>0]|0;k[30803]=21;v=i[a+21>>0]|0;k[30803]=22;w=ty(p&255|0,0,8)|0;p=ty(q&255|w|0,L|0,8)|0;w=ty(p|n&255|0,L|0,8)|0;n=ty(w|m&255|0,L|0,8)|0;m=ty(n|s&255|0,L|0,8)|0;s=ty(m|t&255|0,L|0,8)|0;t=ty(s|u&255|0,L|0,8)|0;u=L;s=t|v&255;v=i[a+22>>0]|0;t=i[a+23>>0]|0;m=i[a+24>>0]|0;k[30803]=25;n=k[e>>2]|0;w=k[d>>2]|0;p=ha(w,n)|0;q=(n+1|0)/2|0;x=(w+1|0)/2|0;y=(p|0)>0;if(y)sy(c|0,0,p<<3|0)|0;k[30804]=0;z=Wt(a,c,w,q,x,v&255)|0;do if(!z){v=(w|0)/2|0;A=t&255;B=Wt(a,c+(x<<3)|0,w,q,v,A)|0;if(!B){C=ha(q,w)|0;D=(n|0)/2|0;E=Wt(a,c+(C<<3)|0,w,D,x,A)|0;if(!E){A=Wt(a,c+(C+x<<3)|0,w,D,v,m&255)|0;if(!A){v=k[30804]|0;D=k[30805]|0;if((v|0)<4){C=k[30803]|0;F=l[a+C>>0]|0|D<<8;k[30805]=F;k[30803]=C+1;C=v+8|0;k[30804]=C;G=C;H=F}else{G=v;H=D}D=G+-4|0;k[30804]=D;if(H>>D&15){tk(783363);I=414;break}k[30804]=0;if(y){D=H;v=0;F=0;while(1){C=c+(F<<3)|0;J=C;K=k[J>>2]|0;M=k[J+4>>2]|0;if(!((K|0)==0&(M|0)==0)){if(!v){J=k[30803]|0;N=l[a+J>>0]|0;k[30805]=N;k[30803]=J+1;k[30804]=8;O=7;P=N}else{O=v+-1|0;P=D}k[30804]=O;if(!(1<<O&P)){Q=P;R=O}else{N=jy(0,0,K|0,M|0)|0;M=C;k[M>>2]=N;k[M+4>>2]=L;Q=P;R=O}}else{Q=D;R=v}F=F+1|0;if((F|0)==(p|0))break;else{D=Q;v=R}}}v=c;k[v>>2]=s;k[v+4>>2]=u;k[g>>2]=0;v=k[f>>2]|0;if((v|0)>=2){D=k[d>>2]|0;F=k[e>>2]|0;M=((v|0)<0)<<31>>31;N=ha(F,D)|0;C=c+(N+-1<<3)|0;if((N|0)>=1?(K=yy(s|0,u|0,v|0,M|0)|0,J=c,k[J>>2]=K,k[J+4>>2]=L,(N|0)>=2):0){N=c+8|0;do{J=N;K=yy(k[J>>2]|0,k[J+4>>2]|0,v|0,M|0)|0;J=N;k[J>>2]=K;k[J+4>>2]=L;N=N+8|0}while(N>>>0<=C>>>0);S=F;T=D}else{S=F;T=D}}else{S=k[e>>2]|0;T=k[d>>2]|0}C=(S|0)>(T|0)?S:T;N=~~(+fa(+(+(C|0)))/.6931471805599453+.5);M=((C|0)>(1<<N|0)&1)+N|0;N=cy(((C+1|0)/2|0)<<3)|0;if(!N){tk(783396);U=414}else{C=M+-1|0;J=ty(1,0,C|0)|0;K=L;V=ty(J|0,K|0,1)|0;W=L;X=ty(J|0,K|0,2)|0;Y=L;Z=jy(0,0,J|0,K|0)|0;_=L;$=ty(Z|0,_|0,2)|0;aa=L;ba=my(V|0,W|0,1)|0;ca=L;da=my(X|0,Y|0,1)|0;Y=L;X=c;ea=k[X>>2]|0;ga=k[X+4>>2]|0;X=my(ea|0,ga|0,63)|0;ia=L;ja=ky(ea|0,ga|0,da|0,Y|0)|0;Y=ky(ja|0,L|0,X|0,ia|0)|0;ia=c;k[ia>>2]=Y&$;k[ia+4>>2]=L&aa;if((M|0)>0){aa=ky(ba|0,ca|0,-1,-1)|0;ia=L;$=ty(Z|0,_|0,1)|0;Y=L;X=v>>1;ja=((X|0)<0)<<31>>31;da=T<<1;ga=jy(0,0,X|0,ja|0)|0;ea=L;ka=0-T|0;la=0-da|0;ma=T+2|0;na=(b|0)==0|(X|0)<1;oa=Z;Z=_;_=J;pa=K;qa=V;V=W;W=J;J=K;K=ba;ba=aa;aa=ca;ca=ia;ia=$;$=Y;Y=1<<M;M=C;C=S;ra=1;sa=T;ta=1;ua=1;while(1){va=K;wa=my(W|0,J|0,1)|0;xa=aa;ya=L;za=ky(wa|0,ya|0,-1,-1)|0;Aa=Y>>1;Ba=ra<<1;Ca=ta<<1;Da=(C|0)>(Aa|0);Ea=C-(Da?Aa:0)|0;Fa=(Da^1)<<31>>31;Da=Fa+Ba|0;Ga=(sa|0)>(Aa|0);Ha=sa-(Ga?Aa:0)|0;Ia=(Ga^1)<<31>>31;Ga=Ia+Ca|0;Ja=(M|0)==0;Ka=Ja?2:ua;La=ba;Ma=Ja?0:za;za=ca;Na=Ja?0:L;Ja=(Da|0)>0;if(Ja){Oa=Ga+1>>1;Pa=(Oa|0)<(Ga|0);Qa=Oa+-1|0;Ra=(Oa|0)>0;Sa=Qa<<1;Ta=(Ga|0)>1;Ua=0;do{Va=ha(Ua,T)|0;if(Pa){Wa=Oa;Xa=c+(Va+Oa<<3)|0;Ya=N;while(1){Za=Xa;_a=k[Za+4>>2]|0;$a=Ya;k[$a>>2]=k[Za>>2];k[$a+4>>2]=_a;Wa=Wa+1|0;if((Wa|0)==(Ga|0))break;else{Xa=Xa+8|0;Ya=Ya+8|0}}}if(Ra){Ya=Qa;Xa=c+(Va+Sa<<3)|0;Wa=c+(Va+Qa<<3)|0;while(1){_a=Wa;$a=k[_a+4>>2]|0;Za=Xa;k[Za>>2]=k[_a>>2];k[Za+4>>2]=$a;if((Ya|0)>0){Ya=Ya+-1|0;Xa=Xa+-16|0;Wa=Wa+-8|0}else break}}if(Ta){Wa=1;Xa=c+(Va+1<<3)|0;Ya=N;while(1){$a=Ya;Za=k[$a+4>>2]|0;_a=Xa;k[_a>>2]=k[$a>>2];k[_a+4>>2]=Za;Wa=Wa+2|0;if((Wa|0)>=(Ga|0))break;else{Xa=Xa+16|0;Ya=Ya+8|0}}}Ua=Ua+1|0}while((Ua|0)!=(Da|0))}Ua=(Ga|0)>0;if(Ua){Ta=Da+1>>1;Qa=(Ta|0)<(Da|0);Sa=ha(Ta,T)|0;Ra=Ta+-1|0;Oa=ha(Ra,T)|0;Pa=(Ta|0)>0;Ya=Oa<<1;Xa=(Da|0)>1;Wa=0;do{if(Qa){Va=Ta;Za=c+(Wa+Sa<<3)|0;_a=N;while(1){$a=Za;ab=k[$a+4>>2]|0;bb=_a;k[bb>>2]=k[$a>>2];k[bb+4>>2]=ab;Va=Va+1|0;if((Va|0)==(Da|0))break;else{Za=Za+(T<<3)|0;_a=_a+8|0}}}if(Pa){_a=Ra;Za=c+(Wa+Ya<<3)|0;Va=c+(Wa+Oa<<3)|0;while(1){ab=Va;bb=k[ab+4>>2]|0;$a=Za;k[$a>>2]=k[ab>>2];k[$a+4>>2]=bb;if((_a|0)>0){_a=_a+-1|0;Za=Za+(la<<3)|0;Va=Va+(ka<<3)|0}else break}}if(Xa){Va=1;Za=c+(Wa+T<<3)|0;_a=N;while(1){bb=_a;$a=k[bb+4>>2]|0;ab=Za;k[ab>>2]=k[bb>>2];k[ab+4>>2]=$a;Va=Va+2|0;if((Va|0)>=(Da|0))break;else{Za=Za+(da<<3)|0;_a=_a+8|0}}}Wa=Wa+1|0}while((Wa|0)!=(Ga|0))}do if(!na){Wa=Da+-2|0;Xa=(Wa|0)>2;if(Xa){Oa=2;do{Ya=ha(Oa,T)|0;if(Ua){Ra=0;Pa=Ya;Sa=Ya+T|0;while(1){Ya=c+(Pa-da<<3)|0;Ta=k[Ya>>2]|0;Qa=k[Ya+4>>2]|0;Ya=c+(Pa<<3)|0;_a=k[Ya>>2]|0;Za=k[Ya+4>>2]|0;Ya=c+(Pa+da<<3)|0;Va=k[Ya>>2]|0;$a=k[Ya+4>>2]|0;Ya=jy(Va|0,$a|0,Ta|0,Qa|0)|0;ab=L;bb=jy(Va|0,$a|0,_a|0,Za|0)|0;$a=L;Va=jy(_a|0,Za|0,Ta|0,Qa|0)|0;Qa=L;Ta=($a|0)<(Qa|0)|($a|0)==(Qa|0)&bb>>>0<Va>>>0;Za=Ta?bb:Va;_a=Ta?$a:Qa;Ta=(_a|0)>0|(_a|0)==0&Za>>>0>0;cb=ty(Za|0,_a|0,2)|0;_a=Ta?cb:0;cb=Ta?L:0;Ta=($a|0)>(Qa|0)|($a|0)==(Qa|0)&bb>>>0>Va>>>0;Za=Ta?$a:Qa;Qa=(Za|0)<0;$a=ty((Ta?bb:Va)|0,Za|0,2)|0;Za=Qa?$a:0;$a=Qa?L:0;if(($a|0)<(cb|0)|($a|0)==(cb|0)&Za>>>0<_a>>>0){Qa=(ab|0)<(cb|0)|(ab|0)==(cb|0)&Ya>>>0<_a>>>0;Va=Qa?Ya:_a;_a=Qa?ab:cb;cb=(_a|0)>($a|0)|(_a|0)==($a|0)&Va>>>0>Za>>>0;ab=c+(Sa<<3)|0;Qa=ab;Ya=k[Qa>>2]|0;bb=k[Qa+4>>2]|0;Qa=ty(Ya|0,bb|0,3)|0;Ta=jy((cb?Va:Za)|0,(cb?_a:$a)|0,Qa|0,L|0)|0;Qa=L;$a=(Qa|0)>-1|(Qa|0)==-1&Ta>>>0>4294967295;_a=ky(Ta|0,Qa|0,7,0)|0;cb=my(($a?Ta:_a)|0,($a?Qa:L)|0,3)|0;Qa=L;$a=(Qa|0)<(ja|0)|(Qa|0)==(ja|0)&cb>>>0<X>>>0;_a=$a?cb:X;cb=$a?Qa:ja;Qa=(cb|0)>(ea|0)|(cb|0)==(ea|0)&_a>>>0>ga>>>0;$a=ky((Qa?_a:ga)|0,(Qa?cb:ea)|0,Ya|0,bb|0)|0;bb=ab;k[bb>>2]=$a;k[bb+4>>2]=L}Ra=Ra+2|0;if((Ra|0)>=(Ga|0))break;else{Pa=Pa+2|0;Sa=Sa+2|0}}}Oa=Oa+2|0}while((Oa|0)<(Wa|0))}if(Ja){Oa=Ga+-2|0;Sa=(Oa|0)>2;Pa=0;do{if(Sa){Ra=ha(Pa,T)|0;bb=Ra+2|0;$a=c+(Ra<<3)|0;Ra=c+(bb<<3)|0;ab=k[$a>>2]|0;Ya=k[$a+4>>2]|0;$a=k[Ra>>2]|0;cb=k[Ra+4>>2]|0;Ra=2;Qa=bb;while(1){bb=Qa;Qa=Qa+2|0;_a=c+(Qa<<3)|0;Ta=k[_a>>2]|0;Za=k[_a+4>>2]|0;_a=jy(Ta|0,Za|0,ab|0,Ya|0)|0;Va=L;db=jy(Ta|0,Za|0,$a|0,cb|0)|0;eb=L;fb=jy($a|0,cb|0,ab|0,Ya|0)|0;gb=L;hb=(eb|0)<(gb|0)|(eb|0)==(gb|0)&db>>>0<fb>>>0;ib=hb?db:fb;jb=hb?eb:gb;hb=(jb|0)>0|(jb|0)==0&ib>>>0>0;kb=ty(ib|0,jb|0,2)|0;jb=hb?kb:0;kb=hb?L:0;hb=(eb|0)>(gb|0)|(eb|0)==(gb|0)&db>>>0>fb>>>0;ib=hb?eb:gb;gb=(ib|0)<0;eb=ty((hb?db:fb)|0,ib|0,2)|0;ib=gb?eb:0;eb=gb?L:0;if((eb|0)<(kb|0)|(eb|0)==(kb|0)&ib>>>0<jb>>>0){gb=(Va|0)<(kb|0)|(Va|0)==(kb|0)&_a>>>0<jb>>>0;fb=gb?_a:jb;jb=gb?Va:kb;kb=(jb|0)>(eb|0)|(jb|0)==(eb|0)&fb>>>0>ib>>>0;Va=c+(bb+1<<3)|0;bb=Va;gb=k[bb>>2]|0;_a=k[bb+4>>2]|0;bb=ty(gb|0,_a|0,3)|0;db=jy((kb?fb:ib)|0,(kb?jb:eb)|0,bb|0,L|0)|0;bb=L;eb=(bb|0)>-1|(bb|0)==-1&db>>>0>4294967295;jb=ky(db|0,bb|0,7,0)|0;kb=my((eb?db:jb)|0,(eb?bb:L)|0,3)|0;bb=L;eb=(bb|0)<(ja|0)|(bb|0)==(ja|0)&kb>>>0<X>>>0;jb=eb?kb:X;kb=eb?bb:ja;bb=(kb|0)>(ea|0)|(kb|0)==(ea|0)&jb>>>0>ga>>>0;eb=ky((bb?jb:ga)|0,(bb?kb:ea)|0,gb|0,_a|0)|0;_a=Va;k[_a>>2]=eb;k[_a+4>>2]=L}Ra=Ra+2|0;if((Ra|0)>=(Oa|0))break;else{_a=cb;eb=$a;$a=Ta;cb=Za;Ya=_a;ab=eb}}}Pa=Pa+2|0}while((Pa|0)<(Da|0))}if(!Xa)break;Pa=Ga+-2|0;Oa=(Pa|0)>2;Sa=2;do{ab=(ha(Sa,T)|0)+2|0;if(Oa){Ya=2;cb=ab;$a=ab+T|0;while(1){ab=cb-da|0;Ra=c+(ab+-2<<3)|0;Qa=k[Ra>>2]|0;eb=k[Ra+4>>2]|0;Ra=cb+da|0;_a=c+(Ra+-2<<3)|0;Za=k[_a>>2]|0;Ta=k[_a+4>>2]|0;_a=c+(ab+2<<3)|0;ab=k[_a>>2]|0;Va=k[_a+4>>2]|0;_a=c+(Ra+2<<3)|0;Ra=k[_a>>2]|0;gb=k[_a+4>>2]|0;_a=c+(cb<<3)|0;kb=k[_a>>2]|0;bb=k[_a+4>>2]|0;_a=jy(Qa|0,eb|0,Za|0,Ta|0)|0;jb=jy(_a|0,L|0,ab|0,Va|0)|0;_a=ky(jb|0,L|0,Ra|0,gb|0)|0;jb=L;db=c+($a<<3)|0;ib=ty(k[db>>2]|0,k[db+4>>2]|0,1)|0;db=L;fb=c+(cb+1<<3)|0;hb=ty(k[fb>>2]|0,k[fb+4>>2]|0,1)|0;fb=L;lb=jy(Ra|0,gb|0,kb|0,bb|0)|0;gb=L;Ra=(gb|0)>0|(gb|0)==0&lb>>>0>0;mb=jy((Ra?lb:0)|0,(Ra?gb:0)|0,ib|0,db|0)|0;Ra=L;nb=jy(kb|0,bb|0,Za|0,Ta|0)|0;Ta=L;Za=(Ta|0)>0|(Ta|0)==0&nb>>>0>0;ob=ky((Za?nb:0)|0,(Za?Ta:0)|0,ib|0,db|0)|0;Za=L;pb=(Ra|0)<(Za|0)|(Ra|0)==(Za|0)&mb>>>0<ob>>>0;qb=jy((pb?mb:ob)|0,(pb?Ra:Za)|0,hb|0,fb|0)|0;Za=L;Ra=jy(kb|0,bb|0,ab|0,Va|0)|0;Va=L;ab=(Va|0)>0|(Va|0)==0&Ra>>>0>0;pb=jy((ab?Ra:0)|0,(ab?Va:0)|0,ib|0,db|0)|0;ab=L;ob=jy(Qa|0,eb|0,kb|0,bb|0)|0;bb=L;kb=(bb|0)>0|(bb|0)==0&ob>>>0>0;eb=ky((kb?ob:0)|0,(kb?bb:0)|0,ib|0,db|0)|0;kb=L;Qa=(ab|0)<(kb|0)|(ab|0)==(kb|0)&pb>>>0<eb>>>0;mb=ky((Qa?pb:eb)|0,(Qa?ab:kb)|0,hb|0,fb|0)|0;kb=L;ab=(Za|0)<(kb|0)|(Za|0)==(kb|0)&qb>>>0<mb>>>0;Qa=ty((ab?qb:mb)|0,(ab?Za:kb)|0,4)|0;kb=L;Za=(gb|0)<0;ab=jy((Za?lb:0)|0,(Za?gb:0)|0,ib|0,db|0)|0;gb=L;Za=(Ta|0)<0;lb=ky((Za?nb:0)|0,(Za?Ta:0)|0,ib|0,db|0)|0;Ta=L;Za=(gb|0)>(Ta|0)|(gb|0)==(Ta|0)&ab>>>0>lb>>>0;nb=jy((Za?ab:lb)|0,(Za?gb:Ta)|0,hb|0,fb|0)|0;Ta=L;gb=(Va|0)<0;Za=jy((gb?Ra:0)|0,(gb?Va:0)|0,ib|0,db|0)|0;Va=L;gb=(bb|0)<0;Ra=ky((gb?ob:0)|0,(gb?bb:0)|0,ib|0,db|0)|0;db=L;ib=(Va|0)>(db|0)|(Va|0)==(db|0)&Za>>>0>Ra>>>0;bb=ky((ib?Za:Ra)|0,(ib?Va:db)|0,hb|0,fb|0)|0;fb=L;hb=(Ta|0)>(fb|0)|(Ta|0)==(fb|0)&nb>>>0>bb>>>0;db=ty((hb?nb:bb)|0,(hb?Ta:fb)|0,4)|0;fb=L;if((fb|0)<(kb|0)|(fb|0)==(kb|0)&db>>>0<Qa>>>0){Ta=(jb|0)<(kb|0)|(jb|0)==(kb|0)&_a>>>0<Qa>>>0;hb=Ta?_a:Qa;Qa=Ta?jb:kb;kb=(Qa|0)>(fb|0)|(Qa|0)==(fb|0)&hb>>>0>db>>>0;jb=c+($a+1<<3)|0;Ta=jb;_a=k[Ta>>2]|0;bb=k[Ta+4>>2]|0;Ta=ty(_a|0,bb|0,6)|0;nb=jy((kb?hb:db)|0,(kb?Qa:fb)|0,Ta|0,L|0)|0;Ta=L;fb=(Ta|0)>-1|(Ta|0)==-1&nb>>>0>4294967295;Qa=ky(nb|0,Ta|0,63,0)|0;kb=my((fb?nb:Qa)|0,(fb?Ta:L)|0,6)|0;Ta=L;fb=(Ta|0)<(ja|0)|(Ta|0)==(ja|0)&kb>>>0<X>>>0;Qa=fb?kb:X;kb=fb?Ta:ja;Ta=(kb|0)>(ea|0)|(kb|0)==(ea|0)&Qa>>>0>ga>>>0;fb=ky((Ta?Qa:ga)|0,(Ta?kb:ea)|0,_a|0,bb|0)|0;bb=jb;k[bb>>2]=fb;k[bb+4>>2]=L}Ya=Ya+2|0;if((Ya|0)>=(Pa|0))break;else{cb=cb+2|0;$a=$a+2|0}}}Sa=Sa+2|0}while((Sa|0)<(Wa|0))}while(0);Ja=(Da|0)%2|0;Ua=(Ga|0)%2|0;Wa=Da-Ja|0;if((Wa|0)>0){Sa=Ga-Ua|0;Pa=(Sa|0)>0;Oa=(Ua|0)==0;Xa=Ca+-1+Ia-Ua&-2;$a=0;cb=ma+Xa|0;Ya=Xa+2|0;while(1){Xa=ha($a,T)|0;bb=Xa+T|0;if(Pa){fb=0;jb=Xa;_a=bb;while(1){kb=c+(jb<<3)|0;Ta=kb;Qa=k[Ta>>2]|0;nb=k[Ta+4>>2]|0;Ta=c+(_a<<3)|0;db=Ta;hb=k[db>>2]|0;Va=k[db+4>>2]|0;db=c+((jb|1)<<3)|0;ib=db;Ra=k[ib>>2]|0;Za=k[ib+4>>2]|0;ib=c+(_a+1<<3)|0;gb=ib;ob=k[gb>>2]|0;lb=k[gb+4>>2]|0;gb=(Va|0)>-1|(Va|0)==-1&hb>>>0>4294967295;ab=ky((gb?va:La)|0,(gb?xa:za)|0,hb|0,Va|0)|0;Va=ab&ia;ab=L&$;hb=(Za|0)>-1|(Za|0)==-1&Ra>>>0>4294967295;gb=ky((hb?va:La)|0,(hb?xa:za)|0,Ra|0,Za|0)|0;Za=gb&ia;gb=L&$;Ra=(lb|0)>-1|(lb|0)==-1&ob>>>0>4294967295;hb=ky((Ra?wa:Ma)|0,(Ra?ya:Na)|0,ob|0,lb|0)|0;lb=hb&oa;hb=L&Z;ob=lb&_;Ra=hb&pa;mb=(ab|0)>-1|(ab|0)==-1&Va>>>0>4294967295;qb=jy(0,0,ob|0,Ra|0)|0;eb=L;pb=ky((mb?qb:ob)|0,(mb?eb:Ra)|0,Va|0,ab|0)|0;ab=L;Va=(gb|0)>-1|(gb|0)==-1&Za>>>0>4294967295;mb=ky((Va?qb:ob)|0,(Va?eb:Ra)|0,Za|0,gb|0)|0;gb=L;Za=(pb^lb^mb)&qa;eb=(ab^hb^gb)&V;if((nb|0)>-1|(nb|0)==-1&Qa>>>0>4294967295){Va=ky(ob|0,Ra|0,Qa|0,nb|0)|0;qb=jy(Va|0,L|0,Za|0,eb|0)|0;rb=qb;sb=L}else{qb=(ob|0)==0&(Ra|0)==0;Va=jy(ob|0,Ra|0,Za|0,eb|0)|0;Ra=ky((qb?Za:Va)|0,(qb?eb:L)|0,Qa|0,nb|0)|0;rb=Ra;sb=L}Ra=ky(rb|0,sb|0,pb|0,ab|0)|0;nb=L;Qa=ky(mb|0,gb|0,lb|0,hb|0)|0;eb=L;qb=ky(Ra|0,nb|0,Qa|0,eb|0)|0;Va=my(qb|0,L|0,Ka|0)|0;qb=ib;k[qb>>2]=Va;k[qb+4>>2]=L;qb=jy(Ra|0,nb|0,Qa|0,eb|0)|0;eb=my(qb|0,L|0,Ka|0)|0;qb=Ta;k[qb>>2]=eb;k[qb+4>>2]=L;qb=jy(rb|0,sb|0,pb|0,ab|0)|0;ab=L;pb=jy(mb|0,gb|0,lb|0,hb|0)|0;eb=ky(pb|0,L|0,qb|0,ab|0)|0;pb=my(eb|0,L|0,Ka|0)|0;eb=db;k[eb>>2]=pb;k[eb+4>>2]=L;eb=jy(lb|0,hb|0,mb|0,gb|0)|0;gb=ky(eb|0,L|0,qb|0,ab|0)|0;ab=my(gb|0,L|0,Ka|0)|0;gb=kb;k[gb>>2]=ab;k[gb+4>>2]=L;fb=fb+2|0;if((fb|0)>=(Sa|0)){tb=Ya;ub=cb;break}else{jb=jb+2|0;_a=_a+2|0}}}else{tb=Xa;ub=bb}if(!Oa){_a=c+(tb<<3)|0;jb=_a;fb=k[jb>>2]|0;gb=k[jb+4>>2]|0;jb=c+(ub<<3)|0;ab=jb;kb=k[ab>>2]|0;qb=k[ab+4>>2]|0;ab=(qb|0)>-1|(qb|0)==-1&kb>>>0>4294967295;eb=ky((ab?va:La)|0,(ab?xa:za)|0,kb|0,qb|0)|0;qb=eb&ia;eb=L&$;kb=qb&qa;ab=eb&V;mb=(gb|0)>-1|(gb|0)==-1&fb>>>0>4294967295;hb=jy(0,0,kb|0,ab|0)|0;lb=ky((mb?hb:kb)|0,(mb?L:ab)|0,fb|0,gb|0)|0;gb=L;fb=ky(lb|0,gb|0,qb|0,eb|0)|0;ab=my(fb|0,L|0,Ka|0)|0;fb=jb;k[fb>>2]=ab;k[fb+4>>2]=L;fb=jy(lb|0,gb|0,qb|0,eb|0)|0;eb=my(fb|0,L|0,Ka|0)|0;fb=_a;k[fb>>2]=eb;k[fb+4>>2]=L}$a=$a+2|0;if(($a|0)>=(Wa|0))break;else{cb=cb+da|0;Ya=Ya+da|0}}vb=(Ba|1)+Fa-Ja&-2}else vb=0;do if(Ja){Ya=ha(vb,T)|0;cb=Ga-Ua|0;if((cb|0)>0){Wa=Ca+-1+Ia|0;$a=0;Oa=Ya;while(1){Sa=c+(Oa<<3)|0;Pa=Sa;fb=k[Pa>>2]|0;eb=k[Pa+4>>2]|0;Pa=c+(Oa+1<<3)|0;_a=Pa;qb=k[_a>>2]|0;gb=k[_a+4>>2]|0;_a=(gb|0)>-1|(gb|0)==-1&qb>>>0>4294967295;lb=ky((_a?va:La)|0,(_a?xa:za)|0,qb|0,gb|0)|0;gb=lb&ia;lb=L&$;qb=gb&qa;_a=lb&V;ab=(eb|0)>-1|(eb|0)==-1&fb>>>0>4294967295;jb=jy(0,0,qb|0,_a|0)|0;mb=ky((ab?jb:qb)|0,(ab?L:_a)|0,fb|0,eb|0)|0;eb=L;fb=ky(mb|0,eb|0,gb|0,lb|0)|0;_a=my(fb|0,L|0,Ka|0)|0;fb=Pa;k[fb>>2]=_a;k[fb+4>>2]=L;fb=jy(mb|0,eb|0,gb|0,lb|0)|0;lb=my(fb|0,L|0,Ka|0)|0;fb=Sa;k[fb>>2]=lb;k[fb+4>>2]=L;$a=$a+2|0;if(($a|0)>=(cb|0))break;else Oa=Oa+2|0}wb=(Wa-Ua+2&-2)+Ya|0}else wb=Ya;if(!Ua)break;Oa=c+(wb<<3)|0;cb=Oa;$a=my(k[cb>>2]|0,k[cb+4>>2]|0,Ka|0)|0;cb=Oa;k[cb>>2]=$a;k[cb+4>>2]=L}while(0);Ua=my(_|0,pa|0,1)|0;za=L;xa=my(oa|0,Z|0,1)|0;if((M|0)>0){La=Z;va=oa;Ia=pa;Ca=_;oa=xa;Z=L;_=Ua;pa=za;W=wa;J=ya;K=wa;ba=Ma;aa=ya;ca=Na;Y=Aa;M=M+-1|0;C=Ea;ra=Da;sa=Ha;ta=Ga;ua=Ka;$=La;ia=va;V=Ia;qa=Ca}else break}}dy(N);U=0}k[g>>2]=U;qa=ha(k[d>>2]|0,k[e>>2]|0)|0;if((qa|0)>0)xb=0;else{o=U;r=h;return o|0}do{k[c+(xb<<2)>>2]=k[c+(xb<<3)>>2];xb=xb+1|0}while((xb|0)!=(qa|0));o=k[g>>2]|0;r=h;return o|0}else I=A}else I=E}else I=B}else I=z;while(0);z=c;k[z>>2]=s;k[z+4>>2]=u;k[g>>2]=I;o=I;r=h;return o|0}function Vt(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0;g=(d|0)>(e|0)?d:e;h=~~(+fa(+(+(g|0)))/.6931471805599453+.5);j=((g|0)>(1<<h|0)&1)+h|0;h=ha((e+1|0)/2|0,(d+1|0)/2|0)|0;g=cy(h)|0;if(!g){tk(783492);m=414;return m|0}n=f+-1|0;a:do if((f|0)>0){o=1<<j;b:do if((j|0)>1){p=n;while(1){q=k[30804]|0;r=k[30805]|0;if((q|0)<4){s=k[30803]|0;t=l[a+s>>0]|0|r<<8;k[30805]=t;k[30803]=s+1;s=q+8|0;k[30804]=s;u=s;v=t}else{u=q;v=r}r=u+-4|0;k[30804]=r;switch(v>>r&15|0){case 0:{Xt(a,h,g);$t(g,d,e,b,c,p);break}case 15:{r=(Zt(a)|0)&255;i[g>>0]=r;r=o;q=1;t=d;s=e;w=1;x=1;while(1){y=r>>1;z=(t|0)>(y|0);w=((z^1)<<31>>31)+(w<<1)|0;A=(s|0)>(y|0);x=((A^1)<<31>>31)+(x<<1)|0;_t(a,g,w,x,g);q=q+1|0;if((q|0)==(j|0))break;else{r=y;t=t-(z?y:0)|0;s=s-(A?y:0)|0}}$t(g,d,e,b,c,p);break}default:break b}if((p|0)>0)p=p+-1|0;else break a}}else{p=n;while(1){s=k[30804]|0;t=k[30805]|0;if((s|0)<4){r=k[30803]|0;q=l[a+r>>0]|0|t<<8;k[30805]=q;k[30803]=r+1;r=s+8|0;k[30804]=r;B=r;C=q}else{B=s;C=t}t=B+-4|0;k[30804]=t;switch(C>>t&15|0){case 0:{Xt(a,h,g);$t(g,d,e,b,c,p);break}case 15:{t=(Zt(a)|0)&255;i[g>>0]=t;$t(g,d,e,b,c,p);break}default:break b}if((p|0)>0)p=p+-1|0;else break a}}while(0);tk(783526);m=414;return m|0}while(0);dy(g);m=0;return m|0}function Wt(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0;g=(d|0)>(e|0)?d:e;h=~~(+fa(+(+(g|0)))/.6931471805599453+.5);j=((g|0)>(1<<h|0)&1)+h|0;h=ha((e+1|0)/2|0,(d+1|0)/2|0)|0;g=cy(h)|0;if(!g){tk(783424);m=414;return m|0}a:do if((f|0)>0){n=1<<j;o=(j|0)>1;p=f;b:while(1){q=p;p=p+-1|0;r=k[30804]|0;s=k[30805]|0;if((r|0)<4){t=k[30803]|0;u=l[a+t>>0]|0|s<<8;k[30805]=u;k[30803]=t+1;t=r+8|0;k[30804]=t;v=t;w=u}else{v=r;w=s}s=v+-4|0;k[30804]=s;switch(w>>s&15|0){case 0:{Xt(a,h,g);Yt(g,d,e,b,c,p);break}case 15:{s=(Zt(a)|0)&255;i[g>>0]=s;if(o){s=n;r=1;u=d;t=e;x=1;y=1;while(1){z=s>>1;A=(u|0)>(z|0);x=((A^1)<<31>>31)+(x<<1)|0;B=(t|0)>(z|0);y=((B^1)<<31>>31)+(y<<1)|0;_t(a,g,x,y,g);r=r+1|0;if((r|0)==(j|0))break;else{s=z;u=u-(A?z:0)|0;t=t-(B?z:0)|0}}}Yt(g,d,e,b,c,p);break}default:break b}if((q|0)<=1)break a}tk(783460);m=414;return m|0}while(0);dy(g);m=0;return m|0}function Xt(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0;d=k[30804]|0;if((b|0)==1){e=k[30805]|0;if((d|0)<4){f=k[30803]|0;g=l[a+f>>0]|0|e<<8;k[30805]=g;k[30803]=f+1;f=d+8|0;k[30804]=f;h=f;j=g}else{h=d;j=e}e=h+-4|0;k[30804]=e;i[c>>0]=j>>e&15;return}if((d|0)!=8){e=d+4|0;if(d){j=(b|0)/2|0;if((b|0)>1){h=(j|0)>1;g=0;f=0;while(1){m=k[30803]|0;n=l[a+m>>0]|0|k[30805]<<8;k[30805]=n;k[30803]=m+1;i[c+f>>0]=n>>e&15;i[c+(f|1)>>0]=k[30805]>>d&15;g=g+1|0;if((g|0)>=(j|0))break;else f=f+2|0}o=h?j:1}else o=0}else p=10}else{k[30803]=(k[30803]|0)+-1;k[30804]=0;p=10}if((p|0)==10){p=(b|0)/2|0;if((b|0)>1){j=(p|0)>1;h=0;f=0;while(1){g=k[30803]|0;d=i[a+g>>0]|0;k[30805]=d&255|k[30805]<<8;k[30803]=g+1;i[c+f>>0]=(d&255)>>>4;i[c+(f|1)>>0]=k[30805]&15;h=h+1|0;if((h|0)>=(p|0))break;else f=f+2|0}o=j?p:1}else o=0}if((o<<1|0)==(b|0))return;o=k[30804]|0;p=k[30805]|0;if((o|0)<4){j=k[30803]|0;f=l[a+j>>0]|0|p<<8;k[30805]=f;k[30803]=j+1;j=o+8|0;k[30804]=j;q=j;r=f}else{q=o;r=p}p=q+-4|0;k[30804]=p;i[c+(b+-1)>>0]=r>>p&15;return}function Yt(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0;g=ty(1,0,f|0)|0;f=L;h=b+-1|0;if((b|0)>1){i=c+-1|0;j=(c|0)>1;m=e+1|0;n=(c+-2|0)>>>1;o=(n<<1)+2|0;p=n+1|0;n=e<<1;q=0;r=o;s=0;while(1){t=ha(q,e)|0;if(j){u=0;v=s;w=t;while(1){do switch(l[a+v>>0]|0|0){case 15:{x=w+e|0;y=d+(x+1<<3)|0;z=y;A=k[z+4>>2]|f;B=y;k[B>>2]=k[z>>2]|g;k[B+4>>2]=A;A=d+(x<<3)|0;x=A;B=k[x+4>>2]|f;z=A;k[z>>2]=k[x>>2]|g;k[z+4>>2]=B;B=d+((w|1)<<3)|0;z=B;x=k[z+4>>2]|f;A=B;k[A>>2]=k[z>>2]|g;k[A+4>>2]=x;x=d+(w<<3)|0;A=x;z=k[A+4>>2]|f;B=x;k[B>>2]=k[A>>2]|g;k[B+4>>2]=z;break}case 1:{z=d+(m+w<<3)|0;B=z;A=k[B+4>>2]|f;x=z;k[x>>2]=k[B>>2]|g;k[x+4>>2]=A;break}case 2:{A=d+(w+e<<3)|0;x=A;B=k[x+4>>2]|f;z=A;k[z>>2]=k[x>>2]|g;k[z+4>>2]=B;break}case 3:{B=w+e|0;z=d+(B+1<<3)|0;x=z;A=k[x+4>>2]|f;y=z;k[y>>2]=k[x>>2]|g;k[y+4>>2]=A;A=d+(B<<3)|0;B=A;y=k[B+4>>2]|f;x=A;k[x>>2]=k[B>>2]|g;k[x+4>>2]=y;break}case 4:{y=d+((w|1)<<3)|0;x=y;B=k[x+4>>2]|f;A=y;k[A>>2]=k[x>>2]|g;k[A+4>>2]=B;break}case 5:{B=d+(m+w<<3)|0;A=B;x=k[A+4>>2]|f;y=B;k[y>>2]=k[A>>2]|g;k[y+4>>2]=x;x=d+((w|1)<<3)|0;y=x;A=k[y+4>>2]|f;B=x;k[B>>2]=k[y>>2]|g;k[B+4>>2]=A;break}case 6:{A=d+(w+e<<3)|0;B=A;y=k[B+4>>2]|f;x=A;k[x>>2]=k[B>>2]|g;k[x+4>>2]=y;y=d+((w|1)<<3)|0;x=y;B=k[x+4>>2]|f;A=y;k[A>>2]=k[x>>2]|g;k[A+4>>2]=B;break}case 7:{B=w+e|0;A=d+(B+1<<3)|0;x=A;y=k[x+4>>2]|f;z=A;k[z>>2]=k[x>>2]|g;k[z+4>>2]=y;y=d+(B<<3)|0;B=y;z=k[B+4>>2]|f;x=y;k[x>>2]=k[B>>2]|g;k[x+4>>2]=z;z=d+((w|1)<<3)|0;x=z;B=k[x+4>>2]|f;y=z;k[y>>2]=k[x>>2]|g;k[y+4>>2]=B;break}case 8:{B=d+(w<<3)|0;y=B;x=k[y+4>>2]|f;z=B;k[z>>2]=k[y>>2]|g;k[z+4>>2]=x;break}case 9:{x=d+(m+w<<3)|0;z=x;y=k[z+4>>2]|f;B=x;k[B>>2]=k[z>>2]|g;k[B+4>>2]=y;y=d+(w<<3)|0;B=y;z=k[B+4>>2]|f;x=y;k[x>>2]=k[B>>2]|g;k[x+4>>2]=z;break}case 10:{z=d+(w+e<<3)|0;x=z;B=k[x+4>>2]|f;y=z;k[y>>2]=k[x>>2]|g;k[y+4>>2]=B;B=d+(w<<3)|0;y=B;x=k[y+4>>2]|f;z=B;k[z>>2]=k[y>>2]|g;k[z+4>>2]=x;break}case 11:{x=w+e|0;z=d+(x+1<<3)|0;y=z;B=k[y+4>>2]|f;A=z;k[A>>2]=k[y>>2]|g;k[A+4>>2]=B;B=d+(x<<3)|0;x=B;A=k[x+4>>2]|f;y=B;k[y>>2]=k[x>>2]|g;k[y+4>>2]=A;A=d+(w<<3)|0;y=A;x=k[y+4>>2]|f;B=A;k[B>>2]=k[y>>2]|g;k[B+4>>2]=x;break}case 12:{x=d+((w|1)<<3)|0;B=x;y=k[B+4>>2]|f;A=x;k[A>>2]=k[B>>2]|g;k[A+4>>2]=y;y=d+(w<<3)|0;A=y;B=k[A+4>>2]|f;x=y;k[x>>2]=k[A>>2]|g;k[x+4>>2]=B;break}case 13:{B=d+(m+w<<3)|0;x=B;A=k[x+4>>2]|f;y=B;k[y>>2]=k[x>>2]|g;k[y+4>>2]=A;A=d+((w|1)<<3)|0;y=A;x=k[y+4>>2]|f;B=A;k[B>>2]=k[y>>2]|g;k[B+4>>2]=x;x=d+(w<<3)|0;B=x;y=k[B+4>>2]|f;A=x;k[A>>2]=k[B>>2]|g;k[A+4>>2]=y;break}case 14:{y=d+(w+e<<3)|0;A=y;B=k[A+4>>2]|f;x=y;k[x>>2]=k[A>>2]|g;k[x+4>>2]=B;B=d+((w|1)<<3)|0;x=B;A=k[x+4>>2]|f;y=B;k[y>>2]=k[x>>2]|g;k[y+4>>2]=A;A=d+(w<<3)|0;y=A;x=k[y+4>>2]|f;B=A;k[B>>2]=k[y>>2]|g;k[B+4>>2]=x;break}default:{}}while(0);u=u+2|0;if((u|0)>=(i|0))break;else{v=v+1|0;w=w+2|0}}C=o;D=p+s|0;E=r}else{C=0;D=s;E=t}if((C|0)<(c|0)){do switch(l[a+D>>0]|0|0){case 15:{w=d+(E+e<<3)|0;v=w;u=k[v+4>>2]|f;x=w;k[x>>2]=k[v>>2]|g;k[x+4>>2]=u;u=d+(E<<3)|0;x=u;v=k[x+4>>2]|f;w=u;k[w>>2]=k[x>>2]|g;k[w+4>>2]=v;break}case 14:{v=d+(E+e<<3)|0;w=v;x=k[w+4>>2]|f;u=v;k[u>>2]=k[w>>2]|g;k[u+4>>2]=x;x=d+(E<<3)|0;u=x;w=k[u+4>>2]|f;v=x;k[v>>2]=k[u>>2]|g;k[v+4>>2]=w;break}case 2:{w=d+(E+e<<3)|0;v=w;u=k[v+4>>2]|f;x=w;k[x>>2]=k[v>>2]|g;k[x+4>>2]=u;break}case 3:{u=d+(E+e<<3)|0;x=u;v=k[x+4>>2]|f;w=u;k[w>>2]=k[x>>2]|g;k[w+4>>2]=v;break}case 13:{v=d+(E<<3)|0;w=v;x=k[w+4>>2]|f;u=v;k[u>>2]=k[w>>2]|g;k[u+4>>2]=x;break}case 12:{x=d+(E<<3)|0;u=x;w=k[u+4>>2]|f;v=x;k[v>>2]=k[u>>2]|g;k[v+4>>2]=w;break}case 6:{w=d+(E+e<<3)|0;v=w;u=k[v+4>>2]|f;x=w;k[x>>2]=k[v>>2]|g;k[x+4>>2]=u;break}case 7:{u=d+(E+e<<3)|0;x=u;v=k[x+4>>2]|f;w=u;k[w>>2]=k[x>>2]|g;k[w+4>>2]=v;break}case 8:{v=d+(E<<3)|0;w=v;x=k[w+4>>2]|f;u=v;k[u>>2]=k[w>>2]|g;k[u+4>>2]=x;break}case 9:{x=d+(E<<3)|0;u=x;w=k[u+4>>2]|f;v=x;k[v>>2]=k[u>>2]|g;k[v+4>>2]=w;break}case 10:{w=d+(E+e<<3)|0;v=w;u=k[v+4>>2]|f;x=w;k[x>>2]=k[v>>2]|g;k[x+4>>2]=u;u=d+(E<<3)|0;x=u;v=k[x+4>>2]|f;w=u;k[w>>2]=k[x>>2]|g;k[w+4>>2]=v;break}case 11:{v=d+(E+e<<3)|0;w=v;x=k[w+4>>2]|f;u=v;k[u>>2]=k[w>>2]|g;k[u+4>>2]=x;x=d+(E<<3)|0;u=x;w=k[u+4>>2]|f;v=x;k[v>>2]=k[u>>2]|g;k[v+4>>2]=w;break}default:{}}while(0);F=D+1|0}else F=D;q=q+2|0;if((q|0)>=(h|0)){G=F;break}else{r=r+n|0;s=F}}H=b&-2;I=G}else{H=0;I=0}if((H|0)>=(b|0))return;b=ha(H,e)|0;e=c+-1|0;if((c|0)>1){H=(c+-2|0)>>>1;G=H<<1;F=I+H|0;H=b+G|0;s=0;n=I;r=b;while(1){do switch(l[a+n>>0]|0|0){case 15:{h=d+(r+1<<3)|0;q=h;D=k[q+4>>2]|f;E=h;k[E>>2]=k[q>>2]|g;k[E+4>>2]=D;D=d+(r<<3)|0;E=D;q=k[E+4>>2]|f;h=D;k[h>>2]=k[E>>2]|g;k[h+4>>2]=q;break}case 14:{q=d+(r+1<<3)|0;h=q;E=k[h+4>>2]|f;D=q;k[D>>2]=k[h>>2]|g;k[D+4>>2]=E;E=d+(r<<3)|0;D=E;h=k[D+4>>2]|f;q=E;k[q>>2]=k[D>>2]|g;k[q+4>>2]=h;break}case 13:{h=d+(r+1<<3)|0;q=h;D=k[q+4>>2]|f;E=h;k[E>>2]=k[q>>2]|g;k[E+4>>2]=D;D=d+(r<<3)|0;E=D;q=k[E+4>>2]|f;h=D;k[h>>2]=k[E>>2]|g;k[h+4>>2]=q;break}case 12:{q=d+(r+1<<3)|0;h=q;E=k[h+4>>2]|f;D=q;k[D>>2]=k[h>>2]|g;k[D+4>>2]=E;E=d+(r<<3)|0;D=E;h=k[D+4>>2]|f;q=E;k[q>>2]=k[D>>2]|g;k[q+4>>2]=h;break}case 4:{h=d+(r+1<<3)|0;q=h;D=k[q+4>>2]|f;E=h;k[E>>2]=k[q>>2]|g;k[E+4>>2]=D;break}case 5:{D=d+(r+1<<3)|0;E=D;q=k[E+4>>2]|f;h=D;k[h>>2]=k[E>>2]|g;k[h+4>>2]=q;break}case 6:{q=d+(r+1<<3)|0;h=q;E=k[h+4>>2]|f;D=q;k[D>>2]=k[h>>2]|g;k[D+4>>2]=E;break}case 7:{E=d+(r+1<<3)|0;D=E;h=k[D+4>>2]|f;q=E;k[q>>2]=k[D>>2]|g;k[q+4>>2]=h;break}case 8:{h=d+(r<<3)|0;q=h;D=k[q+4>>2]|f;E=h;k[E>>2]=k[q>>2]|g;k[E+4>>2]=D;break}case 9:{D=d+(r<<3)|0;E=D;q=k[E+4>>2]|f;h=D;k[h>>2]=k[E>>2]|g;k[h+4>>2]=q;break}case 10:{q=d+(r<<3)|0;h=q;E=k[h+4>>2]|f;D=q;k[D>>2]=k[h>>2]|g;k[D+4>>2]=E;break}case 11:{E=d+(r<<3)|0;D=E;h=k[D+4>>2]|f;q=E;k[q>>2]=k[D>>2]|g;k[q+4>>2]=h;break}default:{}}while(0);s=s+2|0;if((s|0)>=(e|0))break;else{n=n+1|0;r=r+2|0}}J=G+2|0;K=F+1|0;M=H+2|0}else{J=0;K=I;M=b}if((J|0)>=(c|0))return;switch(l[a+K>>0]|0|0){case 15:{K=d+(M<<3)|0;a=K;c=k[a+4>>2]|f;J=K;k[J>>2]=k[a>>2]|g;k[J+4>>2]=c;return}case 14:{c=d+(M<<3)|0;J=c;a=k[J+4>>2]|f;K=c;k[K>>2]=k[J>>2]|g;k[K+4>>2]=a;return}case 13:{a=d+(M<<3)|0;K=a;J=k[K+4>>2]|f;c=a;k[c>>2]=k[K>>2]|g;k[c+4>>2]=J;return}case 12:{J=d+(M<<3)|0;c=J;K=k[c+4>>2]|f;a=J;k[a>>2]=k[c>>2]|g;k[a+4>>2]=K;return}case 11:{K=d+(M<<3)|0;a=K;c=k[a+4>>2]|f;J=K;k[J>>2]=k[a>>2]|g;k[J+4>>2]=c;return}case 10:{c=d+(M<<3)|0;J=c;a=k[J+4>>2]|f;K=c;k[K>>2]=k[J>>2]|g;k[K+4>>2]=a;return}case 9:{a=d+(M<<3)|0;K=a;J=k[K+4>>2]|f;c=a;k[c>>2]=k[K>>2]|g;k[c+4>>2]=J;return}case 8:{J=d+(M<<3)|0;M=J;d=k[M+4>>2]|f;f=J;k[f>>2]=k[M>>2]|g;k[f+4>>2]=d;return}default:return}}function Zt(a){a=a|0;var b=0,c=0,d=0,e=0,f=0,g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0;b=k[30804]|0;c=k[30805]|0;if((b|0)<3){d=k[30803]|0;e=l[a+d>>0]|0|c<<8;k[30805]=e;k[30803]=d+1;d=b+8|0;k[30804]=d;f=e;g=d}else{f=c;g=b}b=g+-3|0;k[30804]=b;c=f>>b&7;if(c>>>0<4){h=1<<c;return h|0}if(!b){b=k[30803]|0;d=l[a+b>>0]|0;k[30805]=d;k[30803]=b+1;k[30804]=8;i=d;j=7}else{i=f;j=g+-4|0}k[30804]=j;g=i>>>j&1|c<<1;a:do if(g>>>0<13)switch(g|0){case 9:{h=5;return h|0}case 10:{h=10;return h|0}case 11:{h=12;return h|0}case 12:{h=15;return h|0}case 8:{h=3;return h|0}default:break a}while(0);if(!j){c=k[30803]|0;f=l[a+c>>0]|0;k[30805]=f;k[30803]=c+1;k[30804]=8;m=f;n=7}else{m=i;n=j+-1|0}k[30804]=n;j=m>>>n&1|g<<1;switch(j|0){case 30:{h=13;return h|0}case 27:{h=7;return h|0}case 28:{h=9;return h|0}case 29:{h=11;return h|0}case 26:{h=6;return h|0}default:{if(!n){g=k[30803]|0;i=l[a+g>>0]|0;k[30805]=i;k[30803]=g+1;k[30804]=8;o=7;p=i}else{o=n+-1|0;p=m}k[30804]=o;h=(p>>>o&1|j<<1|0)==62?0:14;return h|0}}return 0}function _t(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,j=0,k=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0;f=(c+1|0)/2|0;g=d+1|0;h=(g|0)/2|0;j=(ha(h,f)|0)+-1|0;if((c|0)>0){if((d|0)>0){k=-2-h+((g|0)<2?h+1|0:2)|0;g=f;f=j;while(1){j=g;g=g+-1|0;m=h;n=f;o=(ha(g,d)|0)+h<<1;while(1){o=o+-2|0;i[e+o>>0]=i[b+n>>0]|0;if((m|0)<=1)break;else{m=m+-1|0;n=n+-1|0}}if((j|0)<=1)break;else f=k+f|0}}f=c+-1|0;if((c|0)>1){k=d+-1|0;b=(d|0)>1;h=d+-2&-2;g=h+2|0;n=d<<1;m=0;o=d+2+h|0;h=g;while(1){p=ha(m,d)|0;q=p+d|0;a:do if(b){r=0;s=p;t=q;while(1){u=e+s|0;do switch(l[u>>0]|0){case 0:{i[e+(t+1)>>0]=0;i[e+t>>0]=0;i[e+(s|1)>>0]=0;i[u>>0]=0;break}case 1:{i[e+(t+1)>>0]=1;i[e+t>>0]=0;i[e+(s|1)>>0]=0;i[u>>0]=0;break}case 2:{i[e+(t+1)>>0]=0;i[e+t>>0]=1;i[e+(s|1)>>0]=0;i[u>>0]=0;break}case 3:{i[e+(t+1)>>0]=1;i[e+t>>0]=1;i[e+(s|1)>>0]=0;i[u>>0]=0;break}case 4:{i[e+(t+1)>>0]=0;i[e+t>>0]=0;i[e+(s|1)>>0]=1;i[u>>0]=0;break}case 5:{i[e+(t+1)>>0]=1;i[e+t>>0]=0;i[e+(s|1)>>0]=1;i[u>>0]=0;break}case 6:{i[e+(t+1)>>0]=0;i[e+t>>0]=1;i[e+(s|1)>>0]=1;i[u>>0]=0;break}case 7:{i[e+(t+1)>>0]=1;i[e+t>>0]=1;i[e+(s|1)>>0]=1;i[u>>0]=0;break}case 8:{i[e+(t+1)>>0]=0;i[e+t>>0]=0;i[e+(s|1)>>0]=0;i[u>>0]=1;break}case 9:{i[e+(t+1)>>0]=1;i[e+t>>0]=0;i[e+(s|1)>>0]=0;i[u>>0]=1;break}case 10:{i[e+(t+1)>>0]=0;i[e+t>>0]=1;i[e+(s|1)>>0]=0;i[u>>0]=1;break}case 11:{i[e+(t+1)>>0]=1;i[e+t>>0]=1;i[e+(s|1)>>0]=0;i[u>>0]=1;break}case 12:{i[e+(t+1)>>0]=0;i[e+t>>0]=0;i[e+(s|1)>>0]=1;i[u>>0]=1;break}case 13:{i[e+(t+1)>>0]=1;i[e+t>>0]=0;i[e+(s|1)>>0]=1;i[u>>0]=1;break}case 14:{i[e+(t+1)>>0]=0;i[e+t>>0]=1;i[e+(s|1)>>0]=1;i[u>>0]=1;break}case 15:{i[e+(t+1)>>0]=1;i[e+t>>0]=1;i[e+(s|1)>>0]=1;i[u>>0]=1;break}default:{}}while(0);r=r+2|0;if((r|0)>=(k|0)){v=g;w=h;x=o;break a}else{s=s+2|0;t=t+2|0}}}else{v=0;w=p;x=q}while(0);if((v|0)<(d|0)){q=e+w|0;i[e+x>>0]=(l[q>>0]|0)>>>1&1;i[q>>0]=(l[q>>0]|0)>>>3&1}m=m+2|0;if((m|0)>=(f|0))break;else{o=o+n|0;h=h+n|0}}y=c&-2}else y=0}else y=0;if((y|0)<(c|0)){n=ha(y,d)|0;y=d+-1|0;if((d|0)>1){h=0;o=n;while(1){f=e+o|0;m=l[f>>0]|0;i[e+(o|1)>>0]=m>>>2&1;i[f>>0]=m>>>3&1;h=h+2|0;if((h|0)>=(y|0))break;else o=o+2|0}o=d&-2;z=o;A=o+n|0}else{z=0;A=n}if((z|0)<(d|0)){z=e+A|0;i[z>>0]=(l[z>>0]|0)>>>3&1}}z=ha(d,c)|0;if((z|0)>0)B=z;else return;do{z=B;B=B+-1|0;c=e+B|0;if(i[c>>0]|0){d=(Zt(a)|0)&255;i[c>>0]=d}}while((z|0)>1);return}function $t(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0;g=1<<f;f=b+-1|0;if((b|0)>1){h=c+-1|0;i=(c|0)>1;j=e+1|0;m=(c+-2|0)>>>1;n=(m<<1)+2|0;o=m+1|0;m=e<<1;p=0;q=n;r=0;while(1){s=ha(p,e)|0;if(i){t=0;u=r;v=s;while(1){do switch(l[a+u>>0]|0|0){case 15:{w=v+e|0;x=d+(w+1<<2)|0;k[x>>2]=k[x>>2]|g;x=d+(w<<2)|0;k[x>>2]=k[x>>2]|g;x=d+((v|1)<<2)|0;k[x>>2]=k[x>>2]|g;x=d+(v<<2)|0;k[x>>2]=k[x>>2]|g;break}case 1:{x=d+(j+v<<2)|0;k[x>>2]=k[x>>2]|g;break}case 2:{x=d+(v+e<<2)|0;k[x>>2]=k[x>>2]|g;break}case 3:{x=v+e|0;w=d+(x+1<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(x<<2)|0;k[w>>2]=k[w>>2]|g;break}case 4:{w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;break}case 5:{w=d+(j+v<<2)|0;k[w>>2]=k[w>>2]|g;w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;break}case 6:{w=d+(v+e<<2)|0;k[w>>2]=k[w>>2]|g;w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;break}case 7:{w=v+e|0;x=d+(w+1<<2)|0;k[x>>2]=k[x>>2]|g;x=d+(w<<2)|0;k[x>>2]=k[x>>2]|g;x=d+((v|1)<<2)|0;k[x>>2]=k[x>>2]|g;break}case 8:{x=d+(v<<2)|0;k[x>>2]=k[x>>2]|g;break}case 9:{x=d+(j+v<<2)|0;k[x>>2]=k[x>>2]|g;x=d+(v<<2)|0;k[x>>2]=k[x>>2]|g;break}case 10:{x=d+(v+e<<2)|0;k[x>>2]=k[x>>2]|g;x=d+(v<<2)|0;k[x>>2]=k[x>>2]|g;break}case 11:{x=v+e|0;w=d+(x+1<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(x<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(v<<2)|0;k[w>>2]=k[w>>2]|g;break}case 12:{w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(v<<2)|0;k[w>>2]=k[w>>2]|g;break}case 13:{w=d+(j+v<<2)|0;k[w>>2]=k[w>>2]|g;w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(v<<2)|0;k[w>>2]=k[w>>2]|g;break}case 14:{w=d+(v+e<<2)|0;k[w>>2]=k[w>>2]|g;w=d+((v|1)<<2)|0;k[w>>2]=k[w>>2]|g;w=d+(v<<2)|0;k[w>>2]=k[w>>2]|g;break}default:{}}while(0);t=t+2|0;if((t|0)>=(h|0))break;else{u=u+1|0;v=v+2|0}}y=n;z=o+r|0;A=q}else{y=0;z=r;A=s}if((y|0)<(c|0)){do switch(l[a+z>>0]|0|0){case 15:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 14:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 2:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;break}case 3:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;break}case 13:{v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 12:{v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 6:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;break}case 7:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;break}case 8:{v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 9:{v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 10:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}case 11:{v=d+(A+e<<2)|0;k[v>>2]=k[v>>2]|g;v=d+(A<<2)|0;k[v>>2]=k[v>>2]|g;break}default:{}}while(0);B=z+1|0}else B=z;p=p+2|0;if((p|0)>=(f|0)){C=B;break}else{q=q+m|0;r=B}}D=b&-2;E=C}else{D=0;E=0}if((D|0)>=(b|0))return;b=ha(D,e)|0;e=c+-1|0;if((c|0)>1){D=(c+-2|0)>>>1;C=D<<1;B=E+D|0;D=b+C|0;r=0;m=E;q=b;while(1){do switch(l[a+m>>0]|0|0){case 15:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 14:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 13:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 12:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 4:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;break}case 5:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;break}case 6:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;break}case 7:{f=d+(q+1<<2)|0;k[f>>2]=k[f>>2]|g;break}case 8:{f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 9:{f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 10:{f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}case 11:{f=d+(q<<2)|0;k[f>>2]=k[f>>2]|g;break}default:{}}while(0);r=r+2|0;if((r|0)>=(e|0))break;else{m=m+1|0;q=q+2|0}}F=C+2|0;G=B+1|0;H=D+2|0}else{F=0;G=E;H=b}if((F|0)>=(c|0))return;switch(l[a+G>>0]|0|0){case 15:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 14:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 13:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 12:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 11:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 10:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 9:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}case 8:{G=d+(H<<2)|0;k[G>>2]=k[G>>2]|g;return}default:return}}function au(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}if((c|0)<0){k[e>>2]=304;g=304;return g|0}f=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((f|0)==(k[i+68>>2]|0))j=i;else{Fk(a,f+1|0,0,e)|0;j=k[h>>2]|0}f=wy(b|0,c|0,2880,0)|0;i=k[j+64>>2]|0;if(!((i|0)>=0?(f|0)==(k[j+1248+(i<<2)>>2]|0):0))bu(a,f,d,e)|0;d=k[e>>2]|0;if((d|0)>=1){g=d;return g|0}e=(k[h>>2]|0)+48|0;k[e>>2]=b;k[e+4>>2]=c;g=d;return g|0}function bu(a,b,c,d){a=a|0;b=b|0;c=c|0;d=d|0;var e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;e=k[a>>2]|0;f=a+4|0;g=k[f>>2]|0;if((e|0)==(k[g+68>>2]|0))h=g;else{Fk(a,e+1|0,0,d)|0;h=k[f>>2]|0}e=39;while(1){a=k[h+1568+(e<<2)>>2]|0;if((k[h+1248+(a<<2)>>2]|0)==(b|0)){i=a;j=e;l=5;break}a=e+-1|0;if((e|0)>0)e=a;else{m=a;n=e;l=8;break}}a:do if((l|0)==5){k[h+64>>2]=i;o=i;p=h;q=j}else if((l|0)==8){e=yy(b|0,((b|0)<0)<<31>>31|0,2880,0)|0;a=L;if((c|0)==0?(g=h+32|0,r=k[g+4>>2]|0,!((a|0)<(r|0)|((a|0)==(r|0)?e>>>0<(k[g>>2]|0)>>>0:0))):0){k[d>>2]=107;s=107;return s|0}g=k[h+1568>>2]|0;if((g|0)<0){k[d>>2]=103;s=103;return s|0}if(!(k[h+1408+(g<<2)>>2]|0))t=h;else{du(h,g,d)|0;t=k[f>>2]|0}r=t+24|0;u=k[r+4>>2]|0;if((a|0)<(u|0)|((a|0)==(u|0)?e>>>0<(k[r>>2]|0)>>>0:0)){r=t+56|0;if((k[r>>2]|0)==(e|0)?(k[r+4>>2]|0)==(a|0):0)v=t;else{fi(t,e,a)|0;v=k[f>>2]|0}Wh(v,2880,(k[v+1244>>2]|0)+(g*2880|0)|0,d)|0;r=ky(e|0,a|0,2880,0)|0;u=k[f>>2]|0;w=u+56|0;k[w>>2]=r;k[w+4>>2]=L;x=u}else{u=(k[t+1244>>2]|0)+(g*2880|0)|0;if((k[t+72>>2]|0)==1)sy(u|0,32,2880)|0;else sy(u|0,0,2880)|0;u=k[f>>2]|0;w=u+32|0;r=w;y=k[r>>2]|0;z=k[r+4>>2]|0;r=ky(e|0,a|0,2880,0)|0;a=L;e=(z|0)>(a|0)|(z|0)==(a|0)&y>>>0>r>>>0;A=w;k[A>>2]=e?y:r;k[A+4>>2]=e?z:a;k[u+1408+(g<<2)>>2]=1;x=u}k[x+1248+(g<<2)>>2]=b;k[x+64>>2]=g;if((n|0)<1){u=0;while(1){if((k[x+1568+(u<<2)>>2]|0)==(g|0)){o=g;p=x;q=u;break a}a=u+1|0;if((a|0)<40)u=a;else{o=g;p=x;q=a;break}}}else{o=g;p=x;q=m}}while(0);m=q+1|0;if((m|0)<40){x=q;q=m;while(1){k[p+1568+(x<<2)>>2]=k[p+1568+(q<<2)>>2];m=q+1|0;if((m|0)==40)break;else{n=q;q=m;x=n}}}k[p+1724>>2]=o;s=k[d>>2]|0;return s|0}function cu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}f=k[a>>2]|0;h=a+4|0;if((f|0)!=(k[(k[h>>2]|0)+68>>2]|0))Fk(a,f+1|0,0,e)|0;if((c|0)>0|(c|0)==0&b>>>0>2147483647){tk(783556);k[e>>2]=106;g=106;return g|0}f=k[h>>2]|0;if((k[f+64>>2]|0)<0){i=f+48|0;j=wy(k[i>>2]|0,k[i+4>>2]|0,2880,0)|0;bu(a,j,0,e)|0;l=k[h>>2]|0}else l=f;a:do if((c|0)>0|(c|0)==0&b>>>0>8639){f=k[l+64>>2]|0;j=l+48|0;i=k[j>>2]|0;m=k[j+4>>2]|0;j=k[l+1248+(f<<2)>>2]|0;n=ky(b|0,c|0,-1,-1)|0;o=ky(n|0,L|0,i|0,m|0)|0;n=wy(o|0,L|0,2880,0)|0;o=yy(j|0,((j|0)<0)<<31>>31|0,-2880,-1)|0;p=ky(i|0,m|0,o|0,L|0)|0;o=2880-p|0;if(!o){q=l;r=i;s=m;t=d;u=b}else{py((k[l+1244>>2]|0)+(p+(f*2880|0))|0,d|0,o|0)|0;p=ky(o|0,((o|0)<0)<<31>>31|0,i|0,m|0)|0;m=k[h>>2]|0;k[m+1408+(f<<2)>>2]=1;q=m;r=p;s=L;t=d+o|0;u=b-o|0}o=q;p=0;while(1){m=k[o+1248+(p<<2)>>2]|0;if((m|0)<(j|0)|(m|0)>(n|0))v=o;else{if(!(k[o+1408+(p<<2)>>2]|0))w=o;else{du(o,p,e)|0;w=k[h>>2]|0}k[w+1248+(p<<2)>>2]=-1;v=w}p=p+1|0;if((p|0)==40){x=v;break}else o=v}o=x+56|0;if((k[o>>2]|0)==(r|0)?(k[o+4>>2]|0)==(s|0):0)y=x;else{fi(x,r,s)|0;y=k[h>>2]|0}o=u+-1|0;p=o-((o|0)%2880|0)|0;gi(y,p,t,e)|0;o=u-p|0;j=t+p|0;m=ky(p|0,((p|0)<0)<<31>>31|0,r|0,s|0)|0;p=L;i=k[h>>2]|0;z=i+56|0;k[z>>2]=m;k[z+4>>2]=p;z=i+24|0;A=z;B=k[A+4>>2]|0;do if(!((p|0)<(B|0)|((p|0)==(B|0)?m>>>0<(k[A>>2]|0)>>>0:0))){C=z;k[C>>2]=m;k[C+4>>2]=p;C=f*2880|0;D=(k[i+1244>>2]|0)+C|0;if((k[i+72>>2]|0)==1){sy(D|0,32,2880)|0;E=C;break}else{sy(D|0,0,2880)|0;E=C;break}}else{C=f*2880|0;Wh(i,2880,(k[i+1244>>2]|0)+C|0,e)|0;D=(k[h>>2]|0)+56|0;F=D;G=ky(k[F>>2]|0,k[F+4>>2]|0,2880,0)|0;F=D;k[F>>2]=G;k[F+4>>2]=L;E=C}while(0);py((k[(k[h>>2]|0)+1244>>2]|0)+E|0,j|0,o|0)|0;i=k[h>>2]|0;k[i+1408+(f<<2)>>2]=1;k[i+1248+(f<<2)>>2]=n;z=i+32|0;A=z;B=k[A>>2]|0;C=k[A+4>>2]|0;A=n+1|0;F=yy(A|0,((A|0)<0)<<31>>31|0,2880,0)|0;A=L;G=(C|0)>(A|0)|(C|0)==(A|0)&B>>>0>F>>>0;D=z;k[D>>2]=G?B:F;k[D+4>>2]=G?C:A;A=ky(m|0,p|0,o|0,((o|0)<0)<<31>>31|0)|0;C=i+48|0;k[C>>2]=A;k[C+4>>2]=L}else{C=l+48|0;A=k[C>>2]|0;i=k[C+4>>2]|0;C=k[l+1248+(k[l+64>>2]<<2)>>2]|0;G=yy(C|0,((C|0)<0)<<31>>31|0,-2880,-1)|0;C=ky(A|0,i|0,G|0,L|0)|0;G=C;i=d;A=2880-C|0;C=b;while(1){D=i;F=C;while(1){if(!F)break a;B=(F|0)<(A|0)?F:A;z=k[h>>2]|0;py((k[z+1244>>2]|0)+(((k[z+64>>2]|0)*2880|0)+G)|0,D|0,B|0)|0;z=F-B|0;H=D+B|0;I=k[h>>2]|0;J=I+48|0;K=J;M=ky(k[K>>2]|0,k[K+4>>2]|0,B|0,((B|0)<0)<<31>>31|0)|0;K=J;k[K>>2]=M;k[K+4>>2]=L;k[I+1408+(k[I+64>>2]<<2)>>2]=1;if((F|0)==(B|0)){D=H;F=z}else{N=z;O=H;P=J;break}}F=P;D=wy(k[F>>2]|0,k[F+4>>2]|0,2880,0)|0;bu(a,D,1,e)|0;G=0;i=O;A=2880;C=N}}while(0);g=k[e>>2]|0;return g|0}function du(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0;if(!(k[a+76>>2]|0)){tk(783616);if((k[a+4>>2]|0)==8)tk(783657);k[a+1408+(b<<2)>>2]=0;k[c>>2]=112;d=112;return d|0}e=a+1248+(b<<2)|0;f=k[e>>2]|0;g=yy(f|0,((f|0)<0)<<31>>31|0,2880,0)|0;f=L;h=a+24|0;i=h;j=k[i>>2]|0;l=k[i+4>>2]|0;i=a+56|0;m=i;n=k[m>>2]|0;o=k[m+4>>2]|0;if((f|0)>(l|0)|(f|0)==(l|0)&g>>>0>j>>>0){if((n|0)==(j|0)&(o|0)==(l|0)){p=j;q=l}else{fi(a,j,l)|0;l=h;p=k[l>>2]|0;q=k[l+4>>2]|0}if((b|0)==40){r=p;s=q}else{l=a+1244|0;j=p;p=q;while(1){q=wy(j|0,p|0,2880,0)|0;m=b;t=0;u=k[e>>2]|0;while(1){v=k[a+1248+(t<<2)>>2]|0;w=(v|0)>=(q|0)&(v|0)<(u|0);x=w?t:m;y=w?v:u;t=t+1|0;if((t|0)==40){z=x;A=y;break}else{m=x;u=y}}u=yy(A|0,((A|0)<0)<<31>>31|0,2880,0)|0;m=L;if((m|0)>(p|0)|(m|0)==(p|0)&u>>>0>j>>>0){t=jy(u|0,m|0,j|0,p|0)|0;q=wy(t|0,L|0,2880,0)|0;a:do if((q|0)>0){t=0;do{if(k[c>>2]|0)break a;gi(a,2880,783709,c)|0;t=t+1|0}while((t|0)<(q|0))}while(0);q=h;k[q>>2]=u;k[q+4>>2]=m}gi(a,2880,(k[l>>2]|0)+(z*2880|0)|0,c)|0;k[a+1408+(z<<2)>>2]=0;q=h;t=ky(k[q>>2]|0,k[q+4>>2]|0,2880,0)|0;q=L;y=h;k[y>>2]=t;k[y+4>>2]=q;if((z|0)==(b|0)){r=t;s=q;break}else{j=t;p=q}}}p=i;k[p>>2]=r;k[p+4>>2]=s}else{if(!((n|0)==(g|0)&(o|0)==(f|0)))fi(a,g,f)|0;gi(a,2880,(k[a+1244>>2]|0)+(b*2880|0)|0,c)|0;o=ky(g|0,f|0,2880,0)|0;n=L;s=i;k[s>>2]=o;k[s+4>>2]=n;s=h;if((g|0)==(k[s>>2]|0)?(f|0)==(k[s+4>>2]|0):0){s=h;k[s>>2]=o;k[s+4>>2]=n}k[a+1408+(b<<2)>>2]=0}d=k[c>>2]|0;return d|0}function eu(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0,N=0,O=0,P=0,Q=0,R=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}g=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((g|0)==(k[j+68>>2]|0))l=j;else{Fk(a,g+1|0,0,f)|0;l=k[i>>2]|0}g=k[l+64>>2]|0;if((g|0)<0){j=l+48|0;m=wy(k[j>>2]|0,k[j+4>>2]|0,2880,0)|0;bu(a,m,0,f)|0;m=k[i>>2]|0;n=m;o=k[m+64>>2]|0}else{n=l;o=g}g=k[n+1248+(o<<2)>>2]|0;l=n+48|0;m=k[l>>2]|0;j=k[l+4>>2]|0;l=yy(g|0,((g|0)<0)<<31>>31|0,-2880,-1)|0;p=ky(m|0,j|0,l|0,L|0)|0;l=2880-p|0;j=(k[n+1244>>2]|0)+(p+(o*2880|0))|0;p=(l|0)>(b|0)?b:l;py(j|0,e|0,p|0)|0;n=e+p|0;m=(p|0)<(b|0);if((c|0)>1){q=2880-d|0;r=m;s=p;t=n;u=o;v=e;e=1;w=j;j=l;l=g;while(1){if(r){k[(k[i>>2]|0)+1408+(u<<2)>>2]=1;x=l+1|0;bu(a,x,1,f)|0;y=k[i>>2]|0;z=k[y+64>>2]|0;A=k[y+1244>>2]|0;y=z*2880|0;B=b-s|0;py(A+y|0,t|0,B|0)|0;C=z;D=v+b|0;E=A+(B+d+y)|0;F=q-B|0;G=x}else{x=s+d|0;C=u;D=t;E=w+x|0;F=j-x|0;G=l}if((F|0)<1){k[(k[i>>2]|0)+1408+(C<<2)>>2]=1;x=((2880-F|0)/2880|0)+G|0;bu(a,x,1,f)|0;B=k[i>>2]|0;y=k[B+64>>2]|0;A=(0-F|0)%2880|0;H=y;I=(k[B+1244>>2]|0)+((y*2880|0)+A)|0;J=2880-A|0;K=x}else{H=C;I=E;J=F;K=G}e=e+1|0;x=(J|0)>(b|0)?b:J;py(I|0,D|0,x|0)|0;A=D+x|0;y=(x|0)<(b|0);if((e|0)==(c|0)){M=x;N=A;O=y;P=H;Q=K;break}else{r=y;s=x;t=A;u=H;v=D;w=I;j=J;l=K}}}else{M=p;N=n;O=m;P=o;Q=g}if(O){k[(k[i>>2]|0)+1408+(P<<2)>>2]=1;bu(a,Q+1|0,1,f)|0;Q=k[i>>2]|0;a=k[Q+64>>2]|0;py((k[Q+1244>>2]|0)+(a*2880|0)|0,N|0,b-M|0)|0;R=a}else R=P;P=k[i>>2]|0;k[P+1408+(R<<2)>>2]=1;R=P+48|0;P=R;i=k[P>>2]|0;a=k[P+4>>2]|0;P=ha(c,b)|0;b=ha(c+-1|0,d)|0;d=ky(b|0,((b|0)<0)<<31>>31|0,P|0,((P|0)<0)<<31>>31|0)|0;P=ky(d|0,L|0,i|0,a|0)|0;a=R;k[a>>2]=P;k[a+4>>2]=L;h=k[f>>2]|0;return h|0}function fu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0;f=k[e>>2]|0;if((f|0)>0){g=f;return g|0}f=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((f|0)==(k[i+68>>2]|0))j=i;else{Fk(a,f+1|0,0,e)|0;j=k[h>>2]|0}a:do if((c|0)>0|(c|0)==0&b>>>0>8639){f=j+48|0;i=k[f>>2]|0;l=k[f+4>>2]|0;f=wy(i|0,l|0,2880,0)|0;m=ky(i|0,l|0,b|0,c|0)|0;n=L;o=ky(m|0,n|0,-1,-1)|0;p=wy(o|0,L|0,2880,0)|0;o=j;q=0;while(1){if((k[o+1408+(q<<2)>>2]|0)!=0?(r=k[o+1248+(q<<2)>>2]|0,!((r|0)<(f|0)|(r|0)>(p|0))):0){du(o,q,e)|0;s=k[h>>2]|0}else s=o;q=q+1|0;if((q|0)==40){t=s;break}else o=s}o=t+56|0;if((k[o>>2]|0)==(i|0)?(k[o+4>>2]|0)==(l|0):0)u=t;else{fi(t,i,l)|0;u=k[h>>2]|0}Wh(u,b,d,e)|0;o=(k[h>>2]|0)+56|0;k[o>>2]=m;k[o+4>>2]=n}else{o=k[j+64>>2]|0;if((o|0)<0){q=j+48|0;p=wy(k[q>>2]|0,k[q+4>>2]|0,2880,0)|0;bu(a,p,0,e)|0;p=k[h>>2]|0;v=p;w=k[p+64>>2]|0}else{v=j;w=o}o=v+48|0;p=k[o>>2]|0;q=k[o+4>>2]|0;o=k[v+1248+(w<<2)>>2]|0;f=yy(o|0,((o|0)<0)<<31>>31|0,-2880,-1)|0;o=ky(p|0,q|0,f|0,L|0)|0;f=o;q=d;p=2880-o|0;o=b;while(1){r=q;x=o;while(1){if(!x)break a;y=(x|0)<(p|0)?x:p;z=k[h>>2]|0;py(r|0,(k[z+1244>>2]|0)+(((k[z+64>>2]|0)*2880|0)+f)|0,y|0)|0;z=x-y|0;A=r+y|0;B=(k[h>>2]|0)+48|0;C=B;D=ky(k[C>>2]|0,k[C+4>>2]|0,y|0,((y|0)<0)<<31>>31|0)|0;C=L;E=B;k[E>>2]=D;k[E+4>>2]=C;if((x|0)==(y|0)){r=A;x=z}else{F=z;G=A;H=D;I=C;break}}x=wy(H|0,I|0,2880,0)|0;bu(a,x,0,e)|0;f=0;q=G;p=2880;o=F}}while(0);g=k[e>>2]|0;return g|0}function gu(a,b,c,d,e,f){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;var g=0,h=0,i=0,j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0,H=0,I=0,J=0,K=0,M=0;g=k[f>>2]|0;if((g|0)>0){h=g;return h|0}g=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((g|0)==(k[j+68>>2]|0))l=j;else{Fk(a,g+1|0,0,f)|0;l=k[i>>2]|0}g=k[l+64>>2]|0;if((g|0)<0){j=l+48|0;m=wy(k[j>>2]|0,k[j+4>>2]|0,2880,0)|0;bu(a,m,0,f)|0;m=k[i>>2]|0;n=m;o=k[m+64>>2]|0}else{n=l;o=g}g=k[n+1248+(o<<2)>>2]|0;l=n+48|0;m=k[l>>2]|0;j=k[l+4>>2]|0;l=yy(g|0,((g|0)<0)<<31>>31|0,-2880,-1)|0;p=ky(m|0,j|0,l|0,L|0)|0;l=2880-p|0;j=(k[n+1244>>2]|0)+(p+(o*2880|0))|0;o=(l|0)>(b|0)?b:l;py(e|0,j|0,o|0)|0;p=e+o|0;n=(o|0)<(b|0);if((c|0)>1){m=2880-d|0;q=n;r=o;s=p;t=e;e=1;u=j;j=l;l=g;while(1){if(q){v=l+1|0;bu(a,v,0,f)|0;w=k[i>>2]|0;x=k[w+1244>>2]|0;y=(k[w+64>>2]|0)*2880|0;w=b-r|0;py(s|0,x+y|0,w|0)|0;z=t+b|0;A=x+(w+d+y)|0;B=m-w|0;C=v}else{v=r+d|0;z=s;A=u+v|0;B=j-v|0;C=l}v=B+-1|0;if(v>>>0>2879){if((B|0)<1){D=(0-B|0)%2880|0;E=((2880-B|0)/2880|0)+C|0}else{D=2880-((B|0)%2880|0)|0;E=C-((v|0)/2880|0)|0}bu(a,E,0,f)|0;v=k[i>>2]|0;F=(k[v+1244>>2]|0)+(((k[v+64>>2]|0)*2880|0)+D)|0;G=2880-D|0;H=E}else{F=A;G=B;H=C}e=e+1|0;v=(G|0)>(b|0)?b:G;py(z|0,F|0,v|0)|0;w=z+v|0;y=(v|0)<(b|0);if((e|0)==(c|0)){I=v;J=w;K=y;M=H;break}else{q=y;r=v;s=w;t=z;u=F;j=G;l=H}}}else{I=o;J=p;K=n;M=g}if(K){bu(a,M+1|0,0,f)|0;M=k[i>>2]|0;py(J|0,(k[M+1244>>2]|0)+((k[M+64>>2]|0)*2880|0)|0,b-I|0)|0}I=(k[i>>2]|0)+48|0;i=I;M=k[i>>2]|0;J=k[i+4>>2]|0;i=ha(c,b)|0;b=ha(c+-1|0,d)|0;d=ky(b|0,((b|0)<0)<<31>>31|0,i|0,((i|0)<0)<<31>>31|0)|0;i=ky(d|0,L|0,M|0,J|0)|0;J=I;k[J>>2]=i;k[J+4>>2]=L;h=k[f>>2]|0;return h|0}function hu(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0;d=a+4|0;if(!b){b=0;do{a=k[d>>2]|0;if((k[a+1248+(b<<2)>>2]|0)>-1?(k[a+1408+(b<<2)>>2]|0)!=0:0)du(a,b,c)|0;b=b+1|0}while((b|0)!=40)}else{b=0;do{a=k[d>>2]|0;if((k[a+1248+(b<<2)>>2]|0)>-1?(k[a+1408+(b<<2)>>2]|0)!=0:0)du(a,b,c)|0;k[(k[d>>2]|0)+1248+(b<<2)>>2]=-1;b=b+1|0}while((b|0)!=40)}if((k[c>>2]|0)==112){e=112;return e|0}ei(k[d>>2]|0)|0;e=k[c>>2]|0;return e|0}function iu(a,b){a=a|0;b=b|0;var c=0,d=0,e=0,f=0,g=0,h=0,i=0;c=k[a+4>>2]|0;a=c+24|0;d=0;do{e=c+1248+(d<<2)|0;f=k[e>>2]|0;g=yy(f|0,((f|0)<0)<<31>>31|0,2880,0)|0;f=L;h=a;i=k[h+4>>2]|0;if(!((f|0)<(i|0)|((f|0)==(i|0)?g>>>0<(k[h>>2]|0)>>>0:0)))k[e>>2]=-1;d=d+1|0}while((d|0)!=40);return k[b>>2]|0}function ju(a,b,c){a=a|0;b=b|0;c=c|0;var d=0,e=0,f=0,g=0,h=0,i=0,j=0,l=0;d=r;r=r+16|0;e=d;f=k[a>>2]|0;g=a+4|0;h=k[g>>2]|0;if((f|0)==(k[h+68>>2]|0)){i=h+120|0;if(((k[i>>2]|0)==-1?(k[i+4>>2]|0)==-1:0)?(Vk(a,c)|0)>0:0){j=k[c>>2]|0;r=d;return j|0}}else Fk(a,f+1|0,0,c)|0;f=k[g>>2]|0;if(!(k[f+72>>2]|0)){Xk(a,2,e,0,0,c)|0;l=112320/((k[e>>2]|0)/10|0|0)|0}else{e=f+952|0;f=k[e>>2]|0;a=k[e+4>>2]|0;e=(a|0)<0|(a|0)==0&f>>>0<1;g=wy(112320,0,(e?1:f)|0,(e?0:a)|0)|0;k[b>>2]=g;l=(g|0)<1?1:g}k[b>>2]=l;j=k[c>>2]|0;r=d;return j|0}function ku(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0,o=0,p=0,q=0,r=0,s=0,t=0;j=k[i>>2]|0;if((g|0)<0|(g|0)==0&f>>>0<1|(j|0)>0){l=j;return l|0}if((c|0)<0|(c|0)==0&b>>>0<1){k[i>>2]=307;l=307;return l|0}if((e|0)<0|(e|0)==0&d>>>0<1){k[i>>2]=308;l=308;return l|0}j=k[a>>2]|0;m=a+4|0;n=k[m>>2]|0;if((j|0)==(k[n+68>>2]|0))o=n;else{Fk(a,j+1|0,0,i)|0;o=k[m>>2]|0}j=ky(d|0,e|0,-2,-1)|0;n=ky(j|0,L|0,f|0,g|0)|0;j=o+952|0;p=k[j>>2]|0;q=k[j+4>>2]|0;j=wy(n|0,L|0,p|0,q|0)|0;n=ky(j|0,L|0,b|0,c|0)|0;j=L;r=o+944|0;s=k[r+4>>2]|0;if((j|0)>(s|0)|((j|0)==(s|0)?n>>>0>(k[r>>2]|0)>>>0:0)){tk(786589);k[i>>2]=307;l=307;return l|0}r=o+120|0;n=k[r>>2]|0;s=k[r+4>>2]|0;r=ky(b|0,c|0,-1,-1)|0;c=yy(p|0,q|0,r|0,L|0)|0;r=ky(c|0,L|0,d|0,e|0)|0;e=ky(r|0,L|0,n|0,s|0)|0;s=L;n=ky(e|0,s|0,-1,-1)|0;r=L;do if((k[i>>2]|0)<=0){if((s|0)<0|(s|0)==0&e>>>0<1){k[i>>2]=304;break}d=k[a>>2]|0;if((d|0)==(k[o+68>>2]|0))t=o;else{Fk(a,d+1|0,0,i)|0;t=k[m>>2]|0}d=wy(n|0,r|0,2880,0)|0;c=k[t+64>>2]|0;if(!((c|0)>=0?(d|0)==(k[t+1248+(c<<2)>>2]|0):0))bu(a,d,0,i)|0;if((k[i>>2]|0)<1){d=(k[m>>2]|0)+48|0;k[d>>2]=n;k[d+4>>2]=r}}while(0);fu(a,f,g,h,i)|0;l=k[i>>2]|0;return l|0}function lu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0;if((e|0)!=1){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,1,d,e+-1|0,f,g)|0;m=k[g>>2]|0;return m|0}if((d|0)>=8640){e=a+4|0;l=(k[e>>2]|0)+48|0;h=l;i=k[h>>2]|0;j=k[h+4>>2]|0;h=l;k[h>>2]=b;k[h+4>>2]=c;fu(a,d,((d|0)<0)<<31>>31,f,g)|0;h=(k[e>>2]|0)+48|0;k[h>>2]=i;k[h+4>>2]=j;m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}j=k[a>>2]|0;h=a+4|0;i=k[h>>2]|0;if((j|0)==(k[i+68>>2]|0))n=i;else{Fk(a,j+1|0,0,g)|0;n=k[h>>2]|0}j=wy(b|0,c|0,2880,0)|0;i=k[n+64>>2]|0;if(!((i|0)>=0?(j|0)==(k[n+1248+(i<<2)>>2]|0):0))bu(a,j,0,g)|0;if((k[g>>2]|0)<1){j=(k[h>>2]|0)+48|0;k[j>>2]=b;k[j+4>>2]=c}}while(0);fu(a,d,((d|0)<0)<<31>>31,f,g)|0;m=k[g>>2]|0;return m|0}function mu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0;if((e|0)!=2){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,2,d,e+-2|0,f,g)|0;Ks(f,d);m=k[g>>2]|0;return m|0}e=d<<1;if((e|0)>=8640){l=a+4|0;h=(k[l>>2]|0)+48|0;i=h;j=k[i>>2]|0;n=k[i+4>>2]|0;i=h;k[i>>2]=b;k[i+4>>2]=c;fu(a,e,((e|0)<0)<<31>>31,f,g)|0;i=(k[l>>2]|0)+48|0;k[i>>2]=j;k[i+4>>2]=n;Ks(f,d);m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}n=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((n|0)==(k[j+68>>2]|0))o=j;else{Fk(a,n+1|0,0,g)|0;o=k[i>>2]|0}n=wy(b|0,c|0,2880,0)|0;j=k[o+64>>2]|0;if(!((j|0)>=0?(n|0)==(k[o+1248+(j<<2)>>2]|0):0))bu(a,n,0,g)|0;if((k[g>>2]|0)<1){n=(k[i>>2]|0)+48|0;k[n>>2]=b;k[n+4>>2]=c}}while(0);fu(a,e,((e|0)<0)<<31>>31,f,g)|0;Ks(f,d);m=k[g>>2]|0;return m|0}function nu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0;if((e|0)!=4){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,4,d,e+-4|0,f,g)|0;Ls(f,d);m=k[g>>2]|0;return m|0}e=d<<2;if((e|0)>=8640){l=a+4|0;h=(k[l>>2]|0)+48|0;i=h;j=k[i>>2]|0;n=k[i+4>>2]|0;i=h;k[i>>2]=b;k[i+4>>2]=c;fu(a,e,((e|0)<0)<<31>>31,f,g)|0;i=(k[l>>2]|0)+48|0;k[i>>2]=j;k[i+4>>2]=n;Ls(f,d);m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}n=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((n|0)==(k[j+68>>2]|0))o=j;else{Fk(a,n+1|0,0,g)|0;o=k[i>>2]|0}n=wy(b|0,c|0,2880,0)|0;j=k[o+64>>2]|0;if(!((j|0)>=0?(n|0)==(k[o+1248+(j<<2)>>2]|0):0))bu(a,n,0,g)|0;if((k[g>>2]|0)<1){n=(k[i>>2]|0)+48|0;k[n>>2]=b;k[n+4>>2]=c}}while(0);fu(a,e,((e|0)<0)<<31>>31,f,g)|0;Ls(f,d);m=k[g>>2]|0;return m|0}function ou(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0;if((e|0)!=8){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,8,d,e+-8|0,f,g)|0;Ms(f,d);m=k[g>>2]|0;return m|0}e=d<<3;if((e|0)>=8640){l=a+4|0;h=(k[l>>2]|0)+48|0;i=h;j=k[i>>2]|0;n=k[i+4>>2]|0;i=h;k[i>>2]=b;k[i+4>>2]=c;fu(a,e,((e|0)<0)<<31>>31,f,g)|0;i=(k[l>>2]|0)+48|0;k[i>>2]=j;k[i+4>>2]=n;Ms(f,d);m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}n=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((n|0)==(k[j+68>>2]|0))o=j;else{Fk(a,n+1|0,0,g)|0;o=k[i>>2]|0}n=wy(b|0,c|0,2880,0)|0;j=k[o+64>>2]|0;if(!((j|0)>=0?(n|0)==(k[o+1248+(j<<2)>>2]|0):0))bu(a,n,0,g)|0;if((k[g>>2]|0)<1){n=(k[i>>2]|0)+48|0;k[n>>2]=b;k[n+4>>2]=c}}while(0);fu(a,e,((e|0)<0)<<31>>31,f,g)|0;Ms(f,d);m=k[g>>2]|0;return m|0}function pu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0;if((e|0)!=4){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,4,d,e+-4|0,f,g)|0;Ls(f,d);m=k[g>>2]|0;return m|0}e=d<<2;if((e|0)>=8640){l=a+4|0;h=(k[l>>2]|0)+48|0;i=h;j=k[i>>2]|0;n=k[i+4>>2]|0;i=h;k[i>>2]=b;k[i+4>>2]=c;fu(a,e,((e|0)<0)<<31>>31,f,g)|0;i=(k[l>>2]|0)+48|0;k[i>>2]=j;k[i+4>>2]=n;Ls(f,d);m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}n=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((n|0)==(k[j+68>>2]|0))o=j;else{Fk(a,n+1|0,0,g)|0;o=k[i>>2]|0}n=wy(b|0,c|0,2880,0)|0;j=k[o+64>>2]|0;if(!((j|0)>=0?(n|0)==(k[o+1248+(j<<2)>>2]|0):0))bu(a,n,0,g)|0;if((k[g>>2]|0)<1){n=(k[i>>2]|0)+48|0;k[n>>2]=b;k[n+4>>2]=c}}while(0);fu(a,e,((e|0)<0)<<31>>31,f,g)|0;Ls(f,d);m=k[g>>2]|0;return m|0}function qu(a,b,c,d,e,f,g){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;var h=0,i=0,j=0,l=0,m=0,n=0,o=0;if((e|0)!=8){do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}h=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((h|0)==(k[j+68>>2]|0))l=j;else{Fk(a,h+1|0,0,g)|0;l=k[i>>2]|0}h=wy(b|0,c|0,2880,0)|0;j=k[l+64>>2]|0;if(!((j|0)>=0?(h|0)==(k[l+1248+(j<<2)>>2]|0):0))bu(a,h,0,g)|0;if((k[g>>2]|0)<1){h=(k[i>>2]|0)+48|0;k[h>>2]=b;k[h+4>>2]=c}}while(0);gu(a,8,d,e+-8|0,f,g)|0;Ms(f,d);m=k[g>>2]|0;return m|0}e=d<<3;if((e|0)>=8640){l=a+4|0;h=(k[l>>2]|0)+48|0;i=h;j=k[i>>2]|0;n=k[i+4>>2]|0;i=h;k[i>>2]=b;k[i+4>>2]=c;fu(a,e,((e|0)<0)<<31>>31,f,g)|0;i=(k[l>>2]|0)+48|0;k[i>>2]=j;k[i+4>>2]=n;Ms(f,d);m=k[g>>2]|0;return m|0}do if((k[g>>2]|0)<=0){if((c|0)<0){k[g>>2]=304;break}n=k[a>>2]|0;i=a+4|0;j=k[i>>2]|0;if((n|0)==(k[j+68>>2]|0))o=j;else{Fk(a,n+1|0,0,g)|0;o=k[i>>2]|0}n=wy(b|0,c|0,2880,0)|0;j=k[o+64>>2]|0;if(!((j|0)>=0?(n|0)==(k[o+1248+(j<<2)>>2]|0):0))bu(a,n,0,g)|0;if((k[g>>2]|0)<1){n=(k[i>>2]|0)+48|0;k[n>>2]=b;k[n+4>>2]=c}}while(0);fu(a,e,((e|0)<0)<<31>>31,f,g)|0;Ms(f,d);m=k[g>>2]|0;return m|0}function ru(a,b,c,d,e,f,g,h,i){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;f=f|0;g=g|0;h=h|0;i=i|0;var j=0,l=0,m=0,n=0,o=0,q=0,s=0,t=0,u=0,v=0,w=0,x=0,y=0,z=0,A=0,B=0,C=0,D=0,E=0,F=0,G=0;j=r;r=r+96|0;l=j;m=j+8|0;n=k[i>>2]|0;if((g|0)<0|(g|0)==0&f>>>0<1|(n|0)>0){o=n;r=j;return o|0}if((c|0)<0|(c|0)==0&b>>>0<1){k[i>>2]=307;o=307;r=j;return o|0}if((e|0)<0|(e|0)==0&d>>>0<1){k[i>>2]=308;o=308;r=j;return o|0}n=k[a>>2]|0;q=a+4|0;s=k[q>>2]|0;if((n|0)==(k[s+68>>2]|0)){if((k[s+120+4>>2]|0)<0)Vk(a,i)|0}else Fk(a,n+1|0,0,i)|0;n=ky(d|0,e|0,-2,-1)|0;s=ky(n|0,L|0,f|0,g|0)|0;n=k[q>>2]|0;t=n+952|0;u=k[t>>2]|0;v=k[t+4>>2]|0;t=wy(s|0,L|0,u|0,v|0)|0;s=ky(t|0,L|0,b|0,c|0)|0;t=L;w=n+944|0;x=w;y=k[x>>2]|0;z=k[x+4>>2]|0;do if((t|0)>(z|0)|(t|0)==(z|0)&s>>>0>y>>>0){if((k[n+40>>2]|0)!=0?(x=n+976|0,A=k[x+4>>2]|0,!((A|0)>0|(A|0)==0&(k[x>>2]|0)>>>0>0)):0){x=jy(s|0,t|0,y|0,z|0)|0;A=yy(x|0,L|0,u|0,v|0)|0;x=n+968|0;B=x;C=ky(k[B>>2]|0,k[B+4>>2]|0,A|0,L|0)|0;A=x;k[A>>2]=C;k[A+4>>2]=L;A=w;k[A>>2]=s;k[A+4>>2]=t;D=n;E=u;F=v;break}A=jy(s|0,t|0,y|0,z|0)|0;C=L;if((Nj(a,y,z,A,C,i)|0)<=0){x=k[q>>2]|0;B=x+952|0;D=x;E=k[B>>2]|0;F=k[B+4>>2]|0;break}p[l>>3]=+(A>>>0)+4294967296.0*+(C|0);Xw(m,786632,l)|0;tk(m);o=k[i>>2]|0;r=j;return o|0}else{D=n;E=u;F=v}while(0);v=D+120|0;u=k[v>>2]|0;n=k[v+4>>2]|0;v=ky(b|0,c|0,-1,-1)|0;c=yy(E|0,F|0,v|0,L|0)|0;v=L;F=ky(u|0,n|0,d|0,e|0)|0;e=ky(F|0,L|0,c|0,v|0)|0;v=L;c=ky(e|0,v|0,-1,-1)|0;F=L;do if((k[i>>2]|0)<=0){if((v|0)<0|(v|0)==0&e>>>0<1){k[i>>2]=304;break}d=k[a>>2]|0;if((d|0)==(k[D+68>>2]|0))G=D;else{Fk(a,d+1|0,0,i)|0;G=k[q>>2]|0}d=wy(c|0,F|0,2880,0)|0;n=k[G+64>>2]|0;if(!((n|0)>=0?(d|0)==(k[G+1248+(n<<2)>>2]|0):0))bu(a,d,1,i)|0;if((k[i>>2]|0)<1){d=(k[q>>2]|0)+48|0;k[d>>2]=c;k[d+4>>2]=F}}while(0);cu(a,f,g,h,i)|0;o=k[i>>2]|0;r=j;return o|0}function su(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0;if((c|0)==1){cu(a,b,((b|0)<0)<<31>>31,d,e)|0;f=k[e>>2]|0;return f|0}else{eu(a,1,b,c+-1|0,d,e)|0;f=k[e>>2]|0;return f|0}return 0}function tu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;Ks(d,b);if((c|0)==2){f=b<<1;cu(a,f,((f|0)<0)<<31>>31,d,e)|0;g=k[e>>2]|0;return g|0}else{eu(a,2,b,c+-2|0,d,e)|0;g=k[e>>2]|0;return g|0}return 0}function uu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;Ls(d,b);if((c|0)==4){f=b<<2;cu(a,f,((f|0)<0)<<31>>31,d,e)|0;g=k[e>>2]|0;return g|0}else{eu(a,4,b,c+-4|0,d,e)|0;g=k[e>>2]|0;return g|0}return 0}function vu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;Ms(d,b);if((c|0)==8){f=b<<3;cu(a,f,((f|0)<0)<<31>>31,d,e)|0;g=k[e>>2]|0;return g|0}else{eu(a,8,b,c+-8|0,d,e)|0;g=k[e>>2]|0;return g|0}return 0}function wu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;Ls(d,b);if((c|0)==4){f=b<<2;cu(a,f,((f|0)<0)<<31>>31,d,e)|0;g=k[e>>2]|0;return g|0}else{eu(a,4,b,c+-4|0,d,e)|0;g=k[e>>2]|0;return g|0}return 0}function xu(a,b,c,d,e){a=a|0;b=b|0;c=c|0;d=d|0;e=e|0;var f=0,g=0;Ms(d,b);if((c|0)==8){f=b<<3;cu(a,f,((f|0)<0)<<31>>31,d,e)|0;g=k[e>>2]|0;return g|0}else{eu(a,8,b,c+-8|0,d,e)|0;g=k[e>>2]|0;return g|0}return 0}

// EMSCRIPTEN_END_FUNCS
var Ob=[Ny,ck,ik,Zp,_p,Ny,Ny,Ny];var Pb=[Oy,_x,qw,pw,ow,rw,Bx,nh,Di,qi,vi,wi,Ai,ti,Ui,ij,jj,kj,Vi,Wi,dj,ej,aj,$i,Bi,yj,Dj,Ej,Fj,rj,lj,nj,mj,sj,oj,pj,qj,Ei,Ji,Ki,Li,bj,Xw,Ax,Ux,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy,Oy];var Qb=[Py,rf,sf,uf,vf,wf,xf,yf,zf,Cf,Df,Ef,Ff,Gf,Hf,If,Jf,Kf,Lf,Mf,Nf,Of,Pf,Qf,Rf,Sf,Tf,Uf,Vf,Wf,Xf,eg,fg,gg,hg,ig,jg,kg,lg,nf,of,pf,qf,Af,Bf,Yf,Zf,_f,$f,ag,bg,cg,dg,Py,Py,Py,Py,Py,Py,Py,Py,Py,Py,Py];var Rb=[Qy,li,pi,Mi,Qi,tj,xj,Qy];var Sb=[Ry,dy,tk,Gj,Hu,Ju,dv,Wu,Xu,Yu,Zu,_u,Vu,Uu,Tu,Qu,Qx,Rx,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry,Ry];var Tb=[Sy,oh];var Ub=[Ty,nw,kv,wx,ww,Nx,mi,ni,oi,ui,yi,zi,Ni,Oi,Pi,cj,gj,_i,hj,uj,vj,wj,Bj,Cj,Hi,Ii,Uy,xx,Aw,Ty,Ty,Ty];var Vb=[Vy,xh];var Wb=[Wy,Ww,Hw,Hj,rh,Ij,bw,Wy];var Xb=[Xy,Gg,Hg,Ig,vh,Fx,Nc,Dw,Fc,Ah,si,xi,Ri,fj,Si,zj,Aj,Fi,Gi,fy,Yy,Ex,rx,ux,Zw,tx,kk,zt,Xy,Xy,Xy,Xy];var Yb=[Zy,lk];return{_bitshift64Ashr:my,_reproject:sc,_bitshift64Lshr:ny,_gzread:Qg,_listhdu:tc,_closeFITSFile:Cc,_openFITSMem:yc,_zscale:qc,_initwcs:hc,_ffgky:Io,_htons:Yv,_wcssys:lc,_pix2wcsstr:jc,_bitshift64Shl:ty,_i64Add:ky,_gzwrite:Vg,_saostrtod:oc,_vcat:uc,_wcsinfo:ic,_saodtype:pc,_memset:sy,_vls:vc,_wcsunits:mc,_memcpy:py,_llvm_bswap_i32:ry,_gzclose:Kg,_filterTableToImage:Ac,_getImageToArray:zc,_maxFITSMemory:Dc,_tanhdr:rc,_i64Subtract:jy,_ffghdn:bl,_ffgerr:sk,_ntohs:_v,_htonl:Xv,_realloc:fy,_getHeaderToString:Bc,_ffmahd:Fk,_wcs2pixstr:kc,_gzopen:Lg,_openFITSFile:xc,_ffghdt:Zk,_gzseek:Og,_fflush:zw,_ffmnhd:Jl,___errno_location:rv,_testSetjmp:oy,_saveSetjmp:ly,_free:dy,_memmove:qy,_malloc:cy,_reg2wcsstr:nc,runPostSets:iy,_emscripten_replace_memory:Nb,stackAlloc:Zb,stackSave:_b,stackRestore:$b,establishStackSpace:ac,setThrew:bc,setTempRet0:ec,getTempRet0:fc,dynCall_iiiiiiii:Cy,dynCall_iiii:Dy,dynCall_iddiii:Ey,dynCall_i:Fy,dynCall_vi:Gy,dynCall_vii:Hy,dynCall_ii:Iy,dynCall_viiiiiidii:Jy,dynCall_iiiii:Ky,dynCall_iii:Ly,dynCall_iiiiii:My}})


// EMSCRIPTEN_END_ASM
(Module.asmGlobalArg,Module.asmLibraryArg,buffer);var _reproject=Module["_reproject"]=asm["_reproject"];var _gzread=Module["_gzread"]=asm["_gzread"];var _listhdu=Module["_listhdu"]=asm["_listhdu"];var _closeFITSFile=Module["_closeFITSFile"]=asm["_closeFITSFile"];var _openFITSMem=Module["_openFITSMem"]=asm["_openFITSMem"];var _zscale=Module["_zscale"]=asm["_zscale"];var _initwcs=Module["_initwcs"]=asm["_initwcs"];var _bitshift64Lshr=Module["_bitshift64Lshr"]=asm["_bitshift64Lshr"];var _ffghdt=Module["_ffghdt"]=asm["_ffghdt"];var _wcssys=Module["_wcssys"]=asm["_wcssys"];var _pix2wcsstr=Module["_pix2wcsstr"]=asm["_pix2wcsstr"];var _bitshift64Shl=Module["_bitshift64Shl"]=asm["_bitshift64Shl"];var _gzwrite=Module["_gzwrite"]=asm["_gzwrite"];var _saostrtod=Module["_saostrtod"]=asm["_saostrtod"];var _vcat=Module["_vcat"]=asm["_vcat"];var _wcsinfo=Module["_wcsinfo"]=asm["_wcsinfo"];var _bitshift64Ashr=Module["_bitshift64Ashr"]=asm["_bitshift64Ashr"];var _memset=Module["_memset"]=asm["_memset"];var _vls=Module["_vls"]=asm["_vls"];var _wcsunits=Module["_wcsunits"]=asm["_wcsunits"];var _memcpy=Module["_memcpy"]=asm["_memcpy"];var ___errno_location=Module["___errno_location"]=asm["___errno_location"];var _gzclose=Module["_gzclose"]=asm["_gzclose"];var _filterTableToImage=Module["_filterTableToImage"]=asm["_filterTableToImage"];var _getImageToArray=Module["_getImageToArray"]=asm["_getImageToArray"];var _testSetjmp=Module["_testSetjmp"]=asm["_testSetjmp"];var _tanhdr=Module["_tanhdr"]=asm["_tanhdr"];var _i64Subtract=Module["_i64Subtract"]=asm["_i64Subtract"];var _ffghdn=Module["_ffghdn"]=asm["_ffghdn"];var _ffgerr=Module["_ffgerr"]=asm["_ffgerr"];var _ntohs=Module["_ntohs"]=asm["_ntohs"];var _htonl=Module["_htonl"]=asm["_htonl"];var _realloc=Module["_realloc"]=asm["_realloc"];var _i64Add=Module["_i64Add"]=asm["_i64Add"];var _ffmahd=Module["_ffmahd"]=asm["_ffmahd"];var _openFITSFile=Module["_openFITSFile"]=asm["_openFITSFile"];var _maxFITSMemory=Module["_maxFITSMemory"]=asm["_maxFITSMemory"];var _gzopen=Module["_gzopen"]=asm["_gzopen"];var _ffgky=Module["_ffgky"]=asm["_ffgky"];var _htons=Module["_htons"]=asm["_htons"];var _gzseek=Module["_gzseek"]=asm["_gzseek"];var _fflush=Module["_fflush"]=asm["_fflush"];var _ffmnhd=Module["_ffmnhd"]=asm["_ffmnhd"];var _llvm_bswap_i32=Module["_llvm_bswap_i32"]=asm["_llvm_bswap_i32"];var _saodtype=Module["_saodtype"]=asm["_saodtype"];var _getHeaderToString=Module["_getHeaderToString"]=asm["_getHeaderToString"];var _saveSetjmp=Module["_saveSetjmp"]=asm["_saveSetjmp"];var _free=Module["_free"]=asm["_free"];var runPostSets=Module["runPostSets"]=asm["runPostSets"];var _wcs2pixstr=Module["_wcs2pixstr"]=asm["_wcs2pixstr"];var _memmove=Module["_memmove"]=asm["_memmove"];var _malloc=Module["_malloc"]=asm["_malloc"];var _reg2wcsstr=Module["_reg2wcsstr"]=asm["_reg2wcsstr"];var _emscripten_replace_memory=Module["_emscripten_replace_memory"]=asm["_emscripten_replace_memory"];var dynCall_iiiiiiii=Module["dynCall_iiiiiiii"]=asm["dynCall_iiiiiiii"];var dynCall_iiii=Module["dynCall_iiii"]=asm["dynCall_iiii"];var dynCall_iddiii=Module["dynCall_iddiii"]=asm["dynCall_iddiii"];var dynCall_i=Module["dynCall_i"]=asm["dynCall_i"];var dynCall_vi=Module["dynCall_vi"]=asm["dynCall_vi"];var dynCall_vii=Module["dynCall_vii"]=asm["dynCall_vii"];var dynCall_ii=Module["dynCall_ii"]=asm["dynCall_ii"];var dynCall_viiiiiidii=Module["dynCall_viiiiiidii"]=asm["dynCall_viiiiiidii"];var dynCall_iiiii=Module["dynCall_iiiii"]=asm["dynCall_iiiii"];var dynCall_iii=Module["dynCall_iii"]=asm["dynCall_iii"];var dynCall_iiiiii=Module["dynCall_iiiiii"]=asm["dynCall_iiiiii"];Runtime.stackAlloc=asm["stackAlloc"];Runtime.stackSave=asm["stackSave"];Runtime.stackRestore=asm["stackRestore"];Runtime.establishStackSpace=asm["establishStackSpace"];Runtime.setTempRet0=asm["setTempRet0"];Runtime.getTempRet0=asm["getTempRet0"];function ExitStatus(status){this.name="ExitStatus";this.message="Program terminated with exit("+status+")";this.status=status}ExitStatus.prototype=new Error;ExitStatus.prototype.constructor=ExitStatus;var initialStackTop;var preloadStartTime=null;var calledMain=false;dependenciesFulfilled=function runCaller(){if(!Module["calledRun"])run();if(!Module["calledRun"])dependenciesFulfilled=runCaller};Module["callMain"]=Module.callMain=function callMain(args){assert(runDependencies==0,"cannot call main when async dependencies remain! (listen on __ATMAIN__)");assert(__ATPRERUN__.length==0,"cannot call main when preRun functions remain to be called");args=args||[];ensureInitRuntime();var argc=args.length+1;function pad(){for(var i=0;i<4-1;i++){argv.push(0)}}var argv=[allocate(intArrayFromString(Module["thisProgram"]),"i8",ALLOC_NORMAL)];pad();for(var i=0;i<argc-1;i=i+1){argv.push(allocate(intArrayFromString(args[i]),"i8",ALLOC_NORMAL));pad()}argv.push(0);argv=allocate(argv,"i32",ALLOC_NORMAL);try{var ret=Module["_main"](argc,argv,0);exit(ret,true)}catch(e){if(e instanceof ExitStatus){return}else if(e=="SimulateInfiniteLoop"){Module["noExitRuntime"]=true;return}else{if(e&&typeof e==="object"&&e.stack)Module.printErr("exception thrown: "+[e,e.stack]);throw e}}finally{calledMain=true}};function run(args){args=args||Module["arguments"];if(preloadStartTime===null)preloadStartTime=Date.now();if(runDependencies>0){return}preRun();if(runDependencies>0)return;if(Module["calledRun"])return;function doRun(){if(Module["calledRun"])return;Module["calledRun"]=true;if(ABORT)return;ensureInitRuntime();preMain();if(Module["onRuntimeInitialized"])Module["onRuntimeInitialized"]();if(Module["_main"]&&shouldRunNow)Module["callMain"](args);postRun()}if(Module["setStatus"]){Module["setStatus"]("Running...");setTimeout((function(){setTimeout((function(){Module["setStatus"]("")}),1);doRun()}),1)}else{doRun()}}Module["run"]=Module.run=run;function exit(status,implicit){if(implicit&&Module["noExitRuntime"]){return}if(Module["noExitRuntime"]){}else{ABORT=true;EXITSTATUS=status;STACKTOP=initialStackTop;exitRuntime();if(Module["onExit"])Module["onExit"](status)}if(ENVIRONMENT_IS_NODE){process["stdout"]["once"]("drain",(function(){process["exit"](status)}));console.log(" ");setTimeout((function(){process["exit"](status)}),500)}else if(ENVIRONMENT_IS_SHELL&&typeof quit==="function"){quit(status)}throw new ExitStatus(status)}Module["exit"]=Module.exit=exit;var abortDecorators=[];function abort(what){if(what!==undefined){Module.print(what);Module.printErr(what);what=JSON.stringify(what)}else{what=""}ABORT=true;EXITSTATUS=1;var extra="\nIf this abort() is unexpected, build with -s ASSERTIONS=1 which can give more information.";var output="abort("+what+") at "+stackTrace()+extra;if(abortDecorators){abortDecorators.forEach((function(decorator){output=decorator(output,what)}))}throw output}Module["abort"]=Module.abort=abort;if(Module["preInit"]){if(typeof Module["preInit"]=="function")Module["preInit"]=[Module["preInit"]];while(Module["preInit"].length>0){Module["preInit"].pop()()}}var shouldRunNow=true;if(Module["noInitialRun"]){shouldRunNow=false}run();Module["print"]=(function(text){console.log(text)});Module["rootdir"]="/";Module["vfile"]=(function(filename,buf){var size;try{FS.unlink(Module["rootdir"]+filename)}catch(ignore){}FS.createDataFile(Module["rootdir"],filename,buf,true,true);if(buf.length!==undefined){size=buf.length}else if(buf.byteLength!==undefined){size=buf.byteLength}else if(buf.size!==undefined){size=buf.size}else{size=-1}return{path:filename,size:size}});Module["vunlink"]=(function(filename){try{FS.unlink(Module["rootdir"]+filename)}catch(ignore){}});Module["arrfile"]=(function(filename,arr){try{FS.unlink(Module["rootdir"]+filename)}catch(ignore){}FS.createDataFile(Module["rootdir"],filename,arr,true,true);return{path:filename,size:arr.byteLength}});Module["gzcompress"]=(function(data){var gzFile=ccall("gzopen","number",["string","string"],["output.gz","wb"]);var buffer=_malloc(data.length);HEAPU8.set(data,buffer);ccall("gzwrite","number",["number","number","number"],[gzFile,buffer,data.length]);ccall("gzclose","number",["number"],[gzFile]);_free(buffer);var ret=new Uint8Array(FS.root.contents["output.gz"].contents);FS.unlink("output.gz");return ret});Module["gzdecompress"]=(function(data){var BUFSIZE=1024*1024;FS.createDataFile(Module["rootdir"],"input.gz",data,true,true);var gzFile=ccall("gzopen","number",["string","string"],["input.gz","rb"]);var buffer=_malloc(BUFSIZE);var chunks=[];var total=0;var len;var i,ret,curr;while(true){len=ccall("gzread","number",["number","number","number"],[gzFile,buffer,BUFSIZE]);if(len<=0){break}chunks.push(new Uint8Array(len));chunks[chunks.length-1].set(HEAPU8.subarray(buffer,buffer+len));total+=len}ccall("gzclose","number",["number"],[gzFile]);FS.unlink("input.gz");_free(buffer);ret=new Uint8Array(total);curr=0;for(i=0;i<chunks.length;i++){ret.set(chunks[i],curr);curr+=chunks[i].length}return ret});Module["getFITSImage"]=(function(fits,hdu,options,handler){var i,ofptr,hptr,status,datalen,extnum,extname;var buf,bufptr,buflen,bufptr2,slice;var filter=null;var fptr=fits.fptr;var cens=[0,0];var dims=[0,0];var bin=1;if(hdu.fits){Module["cleanupFITSFile"](hdu.fits,false)}hdu.type=hdu.type||0;hptr=_malloc(4);ccall("ffghdn",null,["number","number"],[fptr,hptr]);extnum=getValue(hptr,"i32")-1;_free(hptr);hptr=_malloc(86);setValue(hptr+82,0,"i32");ccall("ffgky",null,["number","number","string","number","number","number"],[fptr,16,"EXTNAME",hptr,0,hptr+82]);status=getValue(hptr+82,"i32");if(status===0){extname=Pointer_stringify(hptr).replace(/^'/,"").replace(/'$/,"").trim()}else{extname=""}_free(hptr);switch(hdu.type){case 0:hdu.imtab="image";ofptr=fptr;break;case 1:case 2:hdu.imtab="table";hdu.table={};if(options&&options.table){if(options.table.filter){filter=options.table.filter}if(options.table.cx){cens[0]=options.table.cx}if(options.table.cy){cens[1]=options.table.cy}if(options.table.nx){dims[0]=options.table.nx}if(options.table.ny){dims[1]=options.table.ny}if(options.table.bin){bin=options.table.bin}}hptr=_malloc(28);setValue(hptr,dims[0],"i32");setValue(hptr+4,dims[1],"i32");setValue(hptr+8,cens[0],"double");setValue(hptr+16,cens[1],"double");setValue(hptr+24,0,"i32");ofptr=ccall("filterTableToImage","number",["number","string","number","number","number","number","number"],[fptr,filter,0,hptr,hptr+8,bin,hptr+24]);hdu.table.nx=getValue(hptr,"i32");hdu.table.ny=getValue(hptr+4,"i32");hdu.table.cx=getValue(hptr+8,"double");hdu.table.cy=getValue(hptr+16,"double");hdu.table.bin=bin;hdu.table.filter=filter;status=getValue(hptr+24,"i32");_free(hptr);Module["errchk"](status);break}hptr=_malloc(24);setValue(hptr,0,"i32");setValue(hptr+4,0,"i32");setValue(hptr+20,0,"i32");slice=options.slice||"";bufptr=ccall("getImageToArray","number",["number","number","number","string","number","number","number","number"],[ofptr,hptr,0,slice,hptr+8,hptr+12,hptr+16,hptr+20]);hdu.naxis1=getValue(hptr+8,"i32");hdu.naxis2=getValue(hptr+12,"i32");hdu.bitpix=getValue(hptr+16,"i32");status=getValue(hptr+20,"i32");_free(hptr);Module["errchk"](status);if(!bufptr){Module["error"]("image is too large (max is JS9.globalOpts.maxMemory)")}datalen=hdu.naxis1*hdu.naxis2;switch(hdu.bitpix){case 8:hdu.image=HEAPU8.subarray(bufptr,bufptr+datalen);break;case 16:hdu.image=HEAP16.subarray(bufptr/2,bufptr/2+datalen);break;case-16:hdu.image=HEAPU16.subarray(bufptr/2,bufptr/2+datalen);break;case 32:hdu.image=HEAP32.subarray(bufptr/4,bufptr/4+datalen);break;case-32:hdu.image=HEAPF32.subarray(bufptr/4,bufptr/4+datalen);break;case-64:hdu.image=HEAPF64.subarray(bufptr/8,bufptr/8+datalen);break}hptr=_malloc(16);setValue(hptr+12,0,"i32");ccall("getHeaderToString",null,["number","number","number","number"],[ofptr,hptr,hptr+8,hptr+12]);hdu.ncard=getValue(hptr+8,"i32");bufptr2=getValue(hptr,"*");buf=HEAPU8.subarray(bufptr2,bufptr2+hdu.ncard*80);buflen=buf.byteLength;hdu.cardstr="";for(i=0;i<buflen;i++){hdu.cardstr+=String.fromCharCode(buf[i])}status=getValue(hptr+12,"i32");_free(hptr);Module["errchk"](status);if(ofptr&&ofptr!==fptr){hptr=_malloc(4);setValue(hptr,0,"i32");ccall("closeFITSFile",null,["number","number"],[ofptr,hptr]);status=getValue(hptr,"i32");_free(hptr);Module["errchk"](status)}if(options.filename){hdu.filename=options.filename}hdu.fits={fptr:fptr,vfile:hdu.vfile,heap:bufptr,cardstr:hdu.cardstr,extnum:extnum,extname:extname};if(handler){handler(hdu,options)}else{Module["error"]("no handler specified for this FITS file")}});Module["handleFITSFile"]=(function(fits,options,handler){var fptr,hptr,fitsname,status;var fileReader,arr;var hdu={};options=options||{};handler=handler||Module["options"].handler;if(fits instanceof Blob){fileReader=new FileReader;fileReader.onload=(function(){if(!options.filename&&fits.name){options.filename=fits.name}if(options.filename){fitsname=options.filename.replace(/^\.\.*/,"X").replace(/\//g,"__");hdu.vfile=fitsname.replace(/\[.*\]/g,"")}else{fitsname="myblob.gz";hdu.vfile=fitsname}try{FS.unlink(Module["rootdir"]+hdu.vfile)}catch(ignore){}arr=new Uint8Array(this.result);try{FS.createDataFile(Module["rootdir"],hdu.vfile,arr,true,true)}catch(e){Module["error"]("can't create virtual file: "+hdu.vfile)}hptr=_malloc(8);setValue(hptr+4,0,"i32");fptr=ccall("openFITSFile","number",["string","string","number","number"],[fitsname,options.extlist,hptr,hptr+4]);hdu.type=getValue(hptr,"i32");status=getValue(hptr+4,"i32");_free(hptr);Module["errchk"](status);hptr=_malloc(4);ccall("ffghdn",null,["number","number"],[fptr,hptr]);hdu.extnum=getValue(hptr,"i32")-1;_free(hptr);Module["getFITSImage"]({fptr:fptr},hdu,options,handler)});fileReader.onerror=(function(e){Module["error"]("fileReader could not read blob as a FITS file")});fileReader.onabort=(function(e){Module["error"]("fileReader did not read blob as a FITS file")});fileReader.readAsArrayBuffer(fits)}else if(typeof fits==="string"){if(options.fptr){fptr=options.fptr;if(options.vfile){hdu.vfile=options.vfile}if(options.extname){hptr=_malloc(4);setValue(hptr,0,"i32");ccall("ffmnhd",null,["number","number","string","number","number"],[fptr,-1,options.extname,0,hptr]);status=getValue(hptr,"i32");_free(hptr);Module["errchk"](status);hptr=_malloc(8);setValue(hptr+4,0,"i32");ccall("ffghdt",null,["number","number","number"],[fptr,hptr,hptr+4]);hdu.type=getValue(hptr,"i32");status=getValue(hptr+4,"i32");_free(hptr);Module["errchk"](status)}else if(options.extnum!==undefined){hptr=_malloc(8);setValue(hptr+4,0,"i32");ccall("ffmahd",null,["number","number","number","number"],[fptr,options.extnum+1,hptr,hptr+4]);hdu.type=getValue(hptr,"i32");status=getValue(hptr+4,"i32");_free(hptr);Module["errchk"](status)}else if(options.slice!==undefined){hptr=_malloc(8);setValue(hptr+4,0,"i32");ccall("ffghdt",null,["number","number","number"],[fptr,hptr,hptr+4]);hdu.type=getValue(hptr,"i32");_free(hptr);Module["errchk"](status)}else{Module["error"]("missing extname/extnum/slice for FITS file")}}else{if(!fits){Module["error"]("FITS file name not specified")}hdu.vfile=fits;hptr=_malloc(8);setValue(hptr+4,0,"i32");fptr=ccall("openFITSFile","number",["string","string","number","number"],[fits,options.extlist,hptr,hptr+4]);hdu.type=getValue(hptr,"i32");status=getValue(hptr+4,"i32");_free(hptr);Module["errchk"](status)}Module["getFITSImage"]({fptr:fptr},hdu,options,handler)}else{Module["error"]("invalid fits input for handleFITSFile")}});Module["cleanupFITSFile"]=(function(fits,all){var hptr;if(!fits){return}if(fits.heap){Module._free(fits.heap);fits.heap=null}if(fits.cardstr){Module._free(fits.cardstr);fits.cardstr=null}if(all){if(fits.fptr){hptr=_malloc(4);setValue(hptr,0,"i32");ccall("closeFITSFile",null,["number","number"],[fits.fptr,hptr]);_free(hptr);fits.fptr=null}if(fits.vfile){try{FS.unlink(Module["rootdir"]+fits.vfile)}catch(ignore){}}}});Module["maxFITSMemory"]=(function(bytes){bytes=bytes||0;return ccall("maxFITSMemory","number",["number"],[bytes])});Module["errchk"]=(function(status){var i,c,hptr,bytes;var hlen=82;var s="ERROR from cfitsio.js: ";if(status){hptr=_malloc(hlen);ccall("ffgerr",null,["number","number"],[status,hptr]);bytes=HEAPU8.subarray(hptr,hptr+hlen);for(i=0;i<hlen;i++){if(bytes[i]===0){break}c=String.fromCharCode(bytes[i]);s+=c}_free(hptr);Module["error"](s)}});Module["error"]=(function(s,e){if(Module["options"].error){Module["options"].error(s,e,true)}else{throw new Error(s)}});Module["options"]={"library":"cfitsio"}





  return {
      initwcs:  Module.cwrap('initwcs', 'number', ['string', 'number']),
      listhdu:  Module.cwrap('listhdu', 'string', ['string', 'string']),
      wcsinfo:  Module.cwrap('wcsinfo', 'string', ['number']),
      wcssys:  Module.cwrap('wcssys', 'string', ['number', 'string']),
      wcsunits:  Module.cwrap('wcsunits', 'string', ['number', 'string']),
      pix2wcs: Module.cwrap('pix2wcsstr', 'string', ['number', 'number', 'number']),
      wcs2pix: Module.cwrap('wcs2pixstr', 'string', ['number', 'number', 'number']),
      reg2wcs: Module.cwrap('reg2wcsstr', 'string', ['number', 'string']),
      saostrtod: Module.cwrap('saostrtod', 'number', ['string']),
      saodtype:  Module.cwrap('saodtype', 'number'),
      arrfile: Module["arrfile"],
      vfile: Module["vfile"],
      vunlink: Module["vunlink"],
      vheap: Module["HEAPU8"],
      vmalloc: Module["_malloc"],
      vmemcpy: Module["writeArrayToMemory"],
      vfree: Module["_free"],
      gzopen: Module.cwrap('gzopen', 'number', ['string', 'string']),
      gzread: Module.cwrap('gzread', 'number', ['number', 'number', 'number']),
      gzwrite: Module.cwrap('gzwrite', 'number', ['number', 'number', 'number']),
      gzclose: Module.cwrap('gzclose', 'number', ['number']),
      gzseek: Module.cwrap('gzseek', 'number', ['number', 'number', 'number']),
      compress: Module['gzcompress'],
      decompress: Module['gzdecompress'],
      handleFITSFile: Module["handleFITSFile"],
      cleanupFITSFile: Module["cleanupFITSFile"],
      getFITSImage: Module["getFITSImage"],
      maxFITSMemory: Module["maxFITSMemory"],
      zscale: Module.cwrap('zscale', 'string', ['number', 'number', 'number', 'number', 'number', 'number', 'number']),
      tanhdr: Module.cwrap('tanhdr', 'string', ['string', 'string', 'string']),
      reproject: Module.cwrap('reproject', 'string', ['string', 'string', 'string', 'string']),
      vls: Module.cwrap('vls', 'int', ['string']),
      vcat: Module.cwrap('vcat', 'string', ['string', 'number']),
      options: Module["options"]
  };
})();

/*
 * arrive.js
 * v2.3.0
 * https://github.com/uzairfarooq/arrive
 * MIT licensed
 *
 * Copyright (c) 2014-2015 Uzair Farooq
 */

var Arrive = (function(window, $, undefined) {

  "use strict";

  if(!window.MutationObserver || typeof HTMLElement === 'undefined'){
    return; //for unsupported browsers
  }

  var arriveUniqueId = 0;

  var utils = (function() {
    var matches = HTMLElement.prototype.matches || HTMLElement.prototype.webkitMatchesSelector || HTMLElement.prototype.mozMatchesSelector
                  || HTMLElement.prototype.msMatchesSelector;

    return {
      matchesSelector: function(elem, selector) {
        return elem instanceof HTMLElement && matches.call(elem, selector);
      }, 
      // to enable function overloading - By John Resig (MIT Licensed)
      addMethod: function (object, name, fn) {
        var old = object[ name ];
        object[ name ] = function(){
          if ( fn.length == arguments.length )
            return fn.apply( this, arguments );
          else if ( typeof old == 'function' )
            return old.apply( this, arguments );
        };
      },
      callCallbacks: function(callbacksToBeCalled) {
        for (var i = 0, cb; cb = callbacksToBeCalled[i]; i++) {
          cb.callback.call(cb.elem);
        }
      },
      // traverse through all descendants of a node to check if event should be fired for any descendant
      checkChildNodesRecursively: function(nodes, registrationData, matchFunc, callbacksToBeCalled) {
        // check each new node if it matches the selector
        for (var i=0, node; node = nodes[i]; i++) {
          if (matchFunc(node, registrationData, callbacksToBeCalled)) {
            callbacksToBeCalled.push({ callback: registrationData.callback, elem: node });
          }

          if (node.childNodes.length > 0) {
            utils.checkChildNodesRecursively(node.childNodes, registrationData, matchFunc, callbacksToBeCalled);
          }
        }
      },
      mergeArrays: function(firstArr, secondArr){
        // Overwrites default options with user-defined options.
        var options = {},
            attrName;
        for (attrName in firstArr) {
          options[attrName] = firstArr[attrName];
        }
        for (attrName in secondArr) {
          options[attrName] = secondArr[attrName];
        }
        return options;
      },
      toElementsArray: function (elements) {
        // check if object is an array (or array like object)
        // Note: window object has .length property but it's not array of elements so don't consider it an array
        if (typeof elements !== "undefined" && (typeof elements.length !== "number" || elements === window)) {
          elements = [elements];
        }
        return elements;
      }
    };
  })();


  // Class to maintain state of all registered events of a single type
  var EventsBucket = (function() {
    var EventsBucket = function() {
      // holds all the events

      this._eventsBucket    = [];
      // function to be called while adding an event, the function should do the event initialization/registration
      this._beforeAdding    = null;
      // function to be called while removing an event, the function should do the event destruction
      this._beforeRemoving  = null;
    };

    EventsBucket.prototype.addEvent = function(target, selector, options, callback) {
      var newEvent = {
        target:             target, 
        selector:           selector, 
        options:            options, 
        callback:           callback, 
        firedElems:         []
      };

      if (this._beforeAdding) {
        this._beforeAdding(newEvent);
      }
      
      this._eventsBucket.push(newEvent);
      return newEvent;
    };

    EventsBucket.prototype.removeEvent = function(compareFunction) {
      for (var i=this._eventsBucket.length - 1, registeredEvent; registeredEvent = this._eventsBucket[i]; i--) {
        if (compareFunction(registeredEvent)) {
          if (this._beforeRemoving) {
              this._beforeRemoving(registeredEvent);
          }
          this._eventsBucket.splice(i, 1);
        }
      }
    };

    EventsBucket.prototype.beforeAdding = function(beforeAdding) {
      this._beforeAdding = beforeAdding;
    };

    EventsBucket.prototype.beforeRemoving = function(beforeRemoving) {
      this._beforeRemoving = beforeRemoving;
    };

    return EventsBucket;
  })();


  /**
   * @constructor
   * General class for binding/unbinding arrive and leave events
   */
  var MutationEvents = function(getObserverConfig, onMutation) {
    var eventsBucket    = new EventsBucket(), 
        me              = this;

    var defaultOptions = {
      fireOnAttributesModification: false
    };

    // actual event registration before adding it to bucket
    eventsBucket.beforeAdding(function(registrationData) {
      var 
        target    = registrationData.target, 
        selector  = registrationData.selector, 
        callback  = registrationData.callback, 
        observer;

      // mutation observer does not work on window or document
      if (target === window.document || target === window)
        target = document.getElementsByTagName("html")[0];

      // Create an observer instance
      observer = new MutationObserver(function(e) {
        onMutation.call(this, e, registrationData);
      });

      var config = getObserverConfig(registrationData.options);
      
      observer.observe(target, config);

      registrationData.observer = observer;
      registrationData.me = me;
    });

    // cleanup/unregister before removing an event
    eventsBucket.beforeRemoving(function (eventData) {
      eventData.observer.disconnect();
    });

    this.bindEvent = function(selector, options, callback) {
      options = utils.mergeArrays(defaultOptions, options);

      var elements = utils.toElementsArray(this);

      for (var i = 0; i < elements.length; i++) {
        eventsBucket.addEvent(elements[i], selector, options, callback);
      }
    };

    this.unbindEvent = function() {
      var elements = utils.toElementsArray(this);
      eventsBucket.removeEvent(function(eventObj) {
        for (var i = 0; i < elements.length; i++) {
          if (this === undefined || eventObj.target === elements[i]) {
            return true;
          }
        }
        return false;
      });
    };

    this.unbindEventWithSelectorOrCallback = function(selector) {
      var elements = utils.toElementsArray(this),
          callback = selector, 
          compareFunction;

      if (typeof selector === "function") {
        compareFunction = function(eventObj) {
          for (var i = 0; i < elements.length; i++) {
            if ((this === undefined || eventObj.target === elements[i]) && eventObj.callback === callback) {
              return true;
            }
          }
          return false;
        };
      }
      else {
        compareFunction = function(eventObj) {
          for (var i = 0; i < elements.length; i++) {
            if ((this === undefined || eventObj.target === elements[i]) && eventObj.selector === selector) {
              return true;
            }
          }
          return false;
        };
      }
      eventsBucket.removeEvent(compareFunction);
    };

    this.unbindEventWithSelectorAndCallback = function(selector, callback) {
      var elements = utils.toElementsArray(this);
      eventsBucket.removeEvent(function(eventObj) {
          for (var i = 0; i < elements.length; i++) {
            if ((this === undefined || eventObj.target === elements[i]) && eventObj.selector === selector && eventObj.callback === callback) {
              return true;
            }
          }
          return false;
      });
    };

    return this;
  };


  /**
   * @constructor
   * Processes 'arrive' events
   */
  var ArriveEvents = function() {
    var mutationEvents,
        me = this;

    // Default options for 'arrive' event
    var arriveDefaultOptions = {
      fireOnAttributesModification: false,
      onceOnly: false,
      existing: false
    };

    function getArriveObserverConfig(options) {
      var config = {
        attributes: false,
        childList: true,
        subtree: true
      };

      if (options.fireOnAttributesModification) {
        config.attributes = true;
      }

      return config;
    }

    function onArriveMutation(mutations, registrationData) {
      mutations.forEach(function( mutation ) {
        var newNodes    = mutation.addedNodes,
            targetNode = mutation.target,
            callbacksToBeCalled = [];

        // If new nodes are added
        if( newNodes !== null && newNodes.length > 0 ) {
          utils.checkChildNodesRecursively(newNodes, registrationData, nodeMatchFunc, callbacksToBeCalled);
        }
        else if (mutation.type === "attributes") {
          if (nodeMatchFunc(targetNode, registrationData, callbacksToBeCalled)) {
            callbacksToBeCalled.push({ callback: registrationData.callback, elem: node });
          }
        }

        utils.callCallbacks(callbacksToBeCalled);
      });
    }

    function nodeMatchFunc(node, registrationData, callbacksToBeCalled) {
      // check a single node to see if it matches the selector
      if (utils.matchesSelector(node, registrationData.selector)) {
        if(node._id === undefined) {
          node._id = arriveUniqueId++;
        }
        // make sure the arrive event is not already fired for the element
        if (registrationData.firedElems.indexOf(node._id) == -1) {

          if (registrationData.options.onceOnly) {
            if (registrationData.firedElems.length === 0) {
              // On first callback, unbind event.
              registrationData.me.unbindEventWithSelectorAndCallback.call(
                  registrationData.target, registrationData.selector, registrationData.callback);
            } else {
              // Ignore multiple mutations which may have been queued before the event was unbound.
              return;
            }
          }

          registrationData.firedElems.push(node._id);
          callbacksToBeCalled.push({ callback: registrationData.callback, elem: node });
        }
      }
    }

    arriveEvents = new MutationEvents(getArriveObserverConfig, onArriveMutation);

    var mutationBindEvent = arriveEvents.bindEvent;

    // override bindEvent function
    arriveEvents.bindEvent = function(selector, options, callback) {

      if (typeof callback === "undefined") {
        callback = options;
        options = arriveDefaultOptions;
      } else {
        options = utils.mergeArrays(arriveDefaultOptions, options);
      }

      var elements = utils.toElementsArray(this);

      if (options.existing) {
        var existing = [];

        for (var i = 0; i < elements.length; i++) {
          var nodes = elements[i].querySelectorAll(selector);
          for (var j = 0; j < nodes.length; j++) {
            existing.push({ callback: callback, elem: nodes[j] });
          }
        }

        // no need to bind event if the callback has to be fired only once and we have already found the element
        if (options.onceOnly && existing.length) {
          return callback.call(existing[0].elem);
        }

        utils.callCallbacks(existing)
      }

      mutationBindEvent.call(this, selector, options, callback);
    };

    return arriveEvents;
  };


  /**
   * @constructor
   * Processes 'leave' events
   */
  var LeaveEvents = function() {
    var mutationEvents,
        me = this;

    // Default options for 'leave' event
    var leaveDefaultOptions = {};

    function getLeaveObserverConfig(options) {
      var config = {
        childList: true,
        subtree: true
      };

      return config;
    }

    function onLeaveMutation(mutations, registrationData) {
      mutations.forEach(function( mutation ) {
        var removedNodes  = mutation.removedNodes,
            targetNode   = mutation.target,
            callbacksToBeCalled = [];

        if( removedNodes !== null && removedNodes.length > 0 ) {
          utils.checkChildNodesRecursively(removedNodes, registrationData, nodeMatchFunc, callbacksToBeCalled);
        }

        utils.callCallbacks(callbacksToBeCalled);
      });
    }

    function nodeMatchFunc(node, registrationData) {
      return utils.matchesSelector(node, registrationData.selector);
    }

    leaveEvents = new MutationEvents(getLeaveObserverConfig, onLeaveMutation);

    var mutationBindEvent = leaveEvents.bindEvent;

    // override bindEvent function
    leaveEvents.bindEvent = function(selector, options, callback) {

      if (typeof callback === "undefined") {
        callback = options;
        options = leaveDefaultOptions;
      } else {
        options = utils.mergeArrays(leaveDefaultOptions, options);
      }

      mutationBindEvent.call(this, selector, options, callback);
    };

    return leaveEvents;
  };


  var arriveEvents = new ArriveEvents(),
      leaveEvents  = new LeaveEvents();

  function exposeUnbindApi(eventObj, exposeTo, funcName) {
    // expose unbind function with function overriding
    utils.addMethod(exposeTo, funcName, eventObj.unbindEvent);
    utils.addMethod(exposeTo, funcName, eventObj.unbindEventWithSelectorOrCallback);
    utils.addMethod(exposeTo, funcName, eventObj.unbindEventWithSelectorAndCallback);
  }

  /*** expose APIs ***/
  function exposeApi(exposeTo) {
    exposeTo.arrive = arriveEvents.bindEvent;
    exposeUnbindApi(arriveEvents, exposeTo, "unbindArrive");

    exposeTo.leave = leaveEvents.bindEvent;
    exposeUnbindApi(leaveEvents, exposeTo, "unbindLeave");
  }

  if ($) {
    exposeApi($.fn);
  }
  exposeApi(HTMLElement.prototype);
  exposeApi(NodeList.prototype);
  exposeApi(HTMLCollection.prototype);
  exposeApi(HTMLDocument.prototype);
  exposeApi(Window.prototype);

  var Arrive = {};
  // expose functions to unbind all arrive/leave events
  exposeUnbindApi(arriveEvents, Arrive, "unbindAllArrive");
  exposeUnbindApi(leaveEvents, Arrive, "unbindAllLeave");

  return Arrive;

})(this, typeof jQuery === 'undefined' ? null : jQuery, undefined);
/*!
 * jQuery Double Tap Plugin.
 *
 * Copyright (c) 2010 Raul Sanchez (http://www.appcropolis.com)
 *
 * Dual licensed under the MIT and GPL licenses:
 * http://www.opensource.org/licenses/mit-license.php
 * http://www.gnu.org/licenses/gpl.html
 */
 
(function($){
    // Determine if we on iPhone or iPad
    var isiOS = false;
    var agent = navigator.userAgent.toLowerCase();
    if(agent.indexOf('iphone') >= 0 || agent.indexOf('ipad') >= 0){
           isiOS = true;
    }
 
    $.fn.doubletap = function(onDoubleTapCallback, onTapCallback, delay){
        var eventName, action;
        delay = delay == null? 500 : delay;
        eventName = isiOS == true? 'touchend' : 'click';
 
        $(this).bind(eventName, function(event){
            var now = new Date().getTime();
            var lastTouch = $(this).data('lastTouch') || now + 1 /** the first time this will make delta a negative number */;
            var delta = now - lastTouch;
            clearTimeout(action);
            if(delta<500 && delta>0){
                if(onDoubleTapCallback != null && typeof onDoubleTapCallback == 'function'){
                    onDoubleTapCallback(event);
                }
            }else{
                $(this).data('lastTouch', now);
                action = setTimeout(function(evt){
                    if(onTapCallback != null && typeof onTapCallback == 'function'){
                        onTapCallback(evt);
                    }
                    clearTimeout(action);   // clear the timeout
                }, delay, [event]);
            }
            $(this).data('lastTouch', now);
        });
    };
})(jQuery);
/*
Axis Labels Plugin for flot.
http://github.com/markrcote/flot-axislabels

Original code is Copyright (c) 2010 Xuan Luo.
Original code was released under the GPLv3 license by Xuan Luo, September 2010.
Original code was rereleased under the MIT license by Xuan Luo, April 2012.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

(function ($) {
    var options = {
      axisLabels: {
        show: true
      }
    };

    function canvasSupported() {
        return !!document.createElement('canvas').getContext;
    }

    function canvasTextSupported() {
        if (!canvasSupported()) {
            return false;
        }
        var dummy_canvas = document.createElement('canvas');
        var context = dummy_canvas.getContext('2d');
        return typeof context.fillText == 'function';
    }

    function css3TransitionSupported() {
        var div = document.createElement('div');
        return typeof div.style.MozTransition != 'undefined'    // Gecko
            || typeof div.style.OTransition != 'undefined'      // Opera
            || typeof div.style.webkitTransition != 'undefined' // WebKit
            || typeof div.style.transition != 'undefined';
    }


    function AxisLabel(axisName, position, padding, plot, opts) {
        this.axisName = axisName;
        this.position = position;
        this.padding = padding;
        this.plot = plot;
        this.opts = opts;
        this.width = 0;
        this.height = 0;
    }

    AxisLabel.prototype.cleanup = function() {
    };


    CanvasAxisLabel.prototype = new AxisLabel();
    CanvasAxisLabel.prototype.constructor = CanvasAxisLabel;
    function CanvasAxisLabel(axisName, position, padding, plot, opts) {
        AxisLabel.prototype.constructor.call(this, axisName, position, padding,
                                             plot, opts);
    }

    CanvasAxisLabel.prototype.calculateSize = function() {
        if (!this.opts.axisLabelFontSizePixels)
            this.opts.axisLabelFontSizePixels = 14;
        if (!this.opts.axisLabelFontFamily)
            this.opts.axisLabelFontFamily = 'sans-serif';

        var textWidth = this.opts.axisLabelFontSizePixels + this.padding;
        var textHeight = this.opts.axisLabelFontSizePixels + this.padding;
        if (this.position == 'left' || this.position == 'right') {
            this.width = this.opts.axisLabelFontSizePixels + this.padding;
            this.height = 0;
        } else {
            this.width = 0;
            this.height = this.opts.axisLabelFontSizePixels + this.padding;
        }
    };

    CanvasAxisLabel.prototype.draw = function(box) {
        if (!this.opts.axisLabelColour)
            this.opts.axisLabelColour = 'black';
        var ctx = this.plot.getCanvas().getContext('2d');
        ctx.save();
        ctx.font = this.opts.axisLabelFontSizePixels + 'px ' +
            this.opts.axisLabelFontFamily;
        ctx.fillStyle = this.opts.axisLabelColour;
        var width = ctx.measureText(this.opts.axisLabel).width;
        var height = this.opts.axisLabelFontSizePixels;
        var x, y, angle = 0;
        if (this.position == 'top') {
            x = box.left + box.width/2 - width/2;
            y = box.top + height*0.72;
        } else if (this.position == 'bottom') {
            x = box.left + box.width/2 - width/2;
            y = box.top + box.height - height*0.72;
        } else if (this.position == 'left') {
            x = box.left + height*0.72;
            y = box.height/2 + box.top + width/2;
            angle = -Math.PI/2;
        } else if (this.position == 'right') {
            x = box.left + box.width - height*0.72;
            y = box.height/2 + box.top - width/2;
            angle = Math.PI/2;
        }
        ctx.translate(x, y);
        ctx.rotate(angle);
        ctx.fillText(this.opts.axisLabel, 0, 0);
        ctx.restore();
    };


    HtmlAxisLabel.prototype = new AxisLabel();
    HtmlAxisLabel.prototype.constructor = HtmlAxisLabel;
    function HtmlAxisLabel(axisName, position, padding, plot, opts) {
        AxisLabel.prototype.constructor.call(this, axisName, position,
                                             padding, plot, opts);
        this.elem = null;
    }

    HtmlAxisLabel.prototype.calculateSize = function() {
        var elem = $('<div class="axisLabels" style="position:absolute;">' +
                     this.opts.axisLabel + '</div>');
        this.plot.getPlaceholder().append(elem);
        // store height and width of label itself, for use in draw()
        this.labelWidth = elem.outerWidth(true);
        this.labelHeight = elem.outerHeight(true);
        elem.remove();

        this.width = this.height = 0;
        if (this.position == 'left' || this.position == 'right') {
            this.width = this.labelWidth + this.padding;
        } else {
            this.height = this.labelHeight + this.padding;
        }
    };

    HtmlAxisLabel.prototype.cleanup = function() {
        if (this.elem) {
            this.elem.remove();
        }
    };

    HtmlAxisLabel.prototype.draw = function(box) {
        this.plot.getPlaceholder().find('#' + this.axisName + 'Label').remove();
        this.elem = $('<div id="' + this.axisName +
                      'Label" " class="axisLabels" style="position:absolute;">'
                      + this.opts.axisLabel + '</div>');
        this.plot.getPlaceholder().append(this.elem);
        if (this.position == 'top') {
            this.elem.css('left', box.left + box.width/2 - this.labelWidth/2 +
                          'px');
            this.elem.css('top', box.top + 'px');
        } else if (this.position == 'bottom') {
            this.elem.css('left', box.left + box.width/2 - this.labelWidth/2 +
                          'px');
            this.elem.css('top', box.top + box.height - this.labelHeight +
                          'px');
        } else if (this.position == 'left') {
            this.elem.css('top', box.top + box.height/2 - this.labelHeight/2 +
                          'px');
            this.elem.css('left', box.left + 'px');
        } else if (this.position == 'right') {
            this.elem.css('top', box.top + box.height/2 - this.labelHeight/2 +
                          'px');
            this.elem.css('left', box.left + box.width - this.labelWidth +
                          'px');
        }
    };


    CssTransformAxisLabel.prototype = new HtmlAxisLabel();
    CssTransformAxisLabel.prototype.constructor = CssTransformAxisLabel;
    function CssTransformAxisLabel(axisName, position, padding, plot, opts) {
        HtmlAxisLabel.prototype.constructor.call(this, axisName, position,
                                                 padding, plot, opts);
    }

    CssTransformAxisLabel.prototype.calculateSize = function() {
        HtmlAxisLabel.prototype.calculateSize.call(this);
        this.width = this.height = 0;
        if (this.position == 'left' || this.position == 'right') {
            this.width = this.labelHeight + this.padding;
        } else {
            this.height = this.labelHeight + this.padding;
        }
    };

    CssTransformAxisLabel.prototype.transforms = function(degrees, x, y) {
        var stransforms = {
            '-moz-transform': '',
            '-webkit-transform': '',
            '-o-transform': '',
            '-ms-transform': ''
        };
        if (x != 0 || y != 0) {
            var stdTranslate = ' translate(' + x + 'px, ' + y + 'px)';
            stransforms['-moz-transform'] += stdTranslate;
            stransforms['-webkit-transform'] += stdTranslate;
            stransforms['-o-transform'] += stdTranslate;
            stransforms['-ms-transform'] += stdTranslate;
        }
        if (degrees != 0) {
            var rotation = degrees / 90;
            var stdRotate = ' rotate(' + degrees + 'deg)';
            stransforms['-moz-transform'] += stdRotate;
            stransforms['-webkit-transform'] += stdRotate;
            stransforms['-o-transform'] += stdRotate;
            stransforms['-ms-transform'] += stdRotate;
        }
        var s = 'top: 0; left: 0; ';
        for (var prop in stransforms) {
            if (stransforms[prop]) {
                s += prop + ':' + stransforms[prop] + ';';
            }
        }
        s += ';';
        return s;
    };

    CssTransformAxisLabel.prototype.calculateOffsets = function(box) {
        var offsets = { x: 0, y: 0, degrees: 0 };
        if (this.position == 'bottom') {
            offsets.x = box.left + box.width/2 - this.labelWidth/2;
            offsets.y = box.top + box.height - this.labelHeight;
        } else if (this.position == 'top') {
            offsets.x = box.left + box.width/2 - this.labelWidth/2;
            offsets.y = box.top;
        } else if (this.position == 'left') {
            offsets.degrees = -90;
            offsets.x = box.left - this.labelWidth/2 + this.labelHeight/2;
            offsets.y = box.height/2 + box.top;
        } else if (this.position == 'right') {
            offsets.degrees = 90;
            offsets.x = box.left + box.width - this.labelWidth/2
                        - this.labelHeight/2;
            offsets.y = box.height/2 + box.top;
        }
        offsets.x = Math.round(offsets.x);
        offsets.y = Math.round(offsets.y);

        return offsets;
    };

    CssTransformAxisLabel.prototype.draw = function(box) {
        this.plot.getPlaceholder().find("." + this.axisName + "Label").remove();
        var offsets = this.calculateOffsets(box);
        this.elem = $('<div class="axisLabels ' + this.axisName +
                      'Label" style="position:absolute; ' +
                      this.transforms(offsets.degrees, offsets.x, offsets.y) +
                      '">' + this.opts.axisLabel + '</div>');
        this.plot.getPlaceholder().append(this.elem);
    };


    IeTransformAxisLabel.prototype = new CssTransformAxisLabel();
    IeTransformAxisLabel.prototype.constructor = IeTransformAxisLabel;
    function IeTransformAxisLabel(axisName, position, padding, plot, opts) {
        CssTransformAxisLabel.prototype.constructor.call(this, axisName,
                                                         position, padding,
                                                         plot, opts);
        this.requiresResize = false;
    }

    IeTransformAxisLabel.prototype.transforms = function(degrees, x, y) {
        // I didn't feel like learning the crazy Matrix stuff, so this uses
        // a combination of the rotation transform and CSS positioning.
        var s = '';
        if (degrees != 0) {
            var rotation = degrees/90;
            while (rotation < 0) {
                rotation += 4;
            }
            s += ' filter: progid:DXImageTransform.Microsoft.BasicImage(rotation=' + rotation + '); ';
            // see below
            this.requiresResize = (this.position == 'right');
        }
        if (x != 0) {
            s += 'left: ' + x + 'px; ';
        }
        if (y != 0) {
            s += 'top: ' + y + 'px; ';
        }
        return s;
    };

    IeTransformAxisLabel.prototype.calculateOffsets = function(box) {
        var offsets = CssTransformAxisLabel.prototype.calculateOffsets.call(
                          this, box);
        // adjust some values to take into account differences between
        // CSS and IE rotations.
        if (this.position == 'top') {
            // FIXME: not sure why, but placing this exactly at the top causes
            // the top axis label to flip to the bottom...
            offsets.y = box.top + 1;
        } else if (this.position == 'left') {
            offsets.x = box.left;
            offsets.y = box.height/2 + box.top - this.labelWidth/2;
        } else if (this.position == 'right') {
            offsets.x = box.left + box.width - this.labelHeight;
            offsets.y = box.height/2 + box.top - this.labelWidth/2;
        }
        return offsets;
    };

    IeTransformAxisLabel.prototype.draw = function(box) {
        CssTransformAxisLabel.prototype.draw.call(this, box);
        if (this.requiresResize) {
            this.elem = this.plot.getPlaceholder().find("." + this.axisName +
                                                        "Label");
            // Since we used CSS positioning instead of transforms for
            // translating the element, and since the positioning is done
            // before any rotations, we have to reset the width and height
            // in case the browser wrapped the text (specifically for the
            // y2axis).
            this.elem.css('width', this.labelWidth);
            this.elem.css('height', this.labelHeight);
        }
    };


    function init(plot) {
        plot.hooks.processOptions.push(function (plot, options) {

            if (!options.axisLabels.show)
                return;

            // This is kind of a hack. There are no hooks in Flot between
            // the creation and measuring of the ticks (setTicks, measureTickLabels
            // in setupGrid() ) and the drawing of the ticks and plot box
            // (insertAxisLabels in setupGrid() ).
            //
            // Therefore, we use a trick where we run the draw routine twice:
            // the first time to get the tick measurements, so that we can change
            // them, and then have it draw it again.
            var secondPass = false;

            var axisLabels = {};
            var axisOffsetCounts = { left: 0, right: 0, top: 0, bottom: 0 };

            var defaultPadding = 2;  // padding between axis and tick labels
            plot.hooks.draw.push(function (plot, ctx) {
                var hasAxisLabels = false;
                if (!secondPass) {
                    // MEASURE AND SET OPTIONS
                    $.each(plot.getAxes(), function(axisName, axis) {
                        var opts = axis.options // Flot 0.7
                            || plot.getOptions()[axisName]; // Flot 0.6

                        // Handle redraws initiated outside of this plug-in.
                        if (axisName in axisLabels) {
                            axis.labelHeight = axis.labelHeight -
                                axisLabels[axisName].height;
                            axis.labelWidth = axis.labelWidth -
                                axisLabels[axisName].width;
                            opts.labelHeight = axis.labelHeight;
                            opts.labelWidth = axis.labelWidth;
                            axisLabels[axisName].cleanup();
                            delete axisLabels[axisName];
                        }

                        if (!opts || !opts.axisLabel || !axis.show)
                            return;

                        hasAxisLabels = true;
                        var renderer = null;

                        if (!opts.axisLabelUseHtml &&
                            navigator.appName == 'Microsoft Internet Explorer') {
                            var ua = navigator.userAgent;
                            var re  = new RegExp("MSIE ([0-9]{1,}[\.0-9]{0,})");
                            if (re.exec(ua) != null) {
                                rv = parseFloat(RegExp.$1);
                            }
                            if (rv >= 9 && !opts.axisLabelUseCanvas && !opts.axisLabelUseHtml) {
                                renderer = CssTransformAxisLabel;
                            } else if (!opts.axisLabelUseCanvas && !opts.axisLabelUseHtml) {
                                renderer = IeTransformAxisLabel;
                            } else if (opts.axisLabelUseCanvas) {
                                renderer = CanvasAxisLabel;
                            } else {
                                renderer = HtmlAxisLabel;
                            }
                        } else {
                            if (opts.axisLabelUseHtml || (!css3TransitionSupported() && !canvasTextSupported()) && !opts.axisLabelUseCanvas) {
                                renderer = HtmlAxisLabel;
                            } else if (opts.axisLabelUseCanvas || !css3TransitionSupported()) {
                                renderer = CanvasAxisLabel;
                            } else {
                                renderer = CssTransformAxisLabel;
                            }
                        }

                        var padding = opts.axisLabelPadding === undefined ?
                                      defaultPadding : opts.axisLabelPadding;

                        axisLabels[axisName] = new renderer(axisName,
                                                            axis.position, padding,
                                                            plot, opts);

                        // flot interprets axis.labelHeight and .labelWidth as
                        // the height and width of the tick labels. We increase
                        // these values to make room for the axis label and
                        // padding.

                        axisLabels[axisName].calculateSize();

                        // AxisLabel.height and .width are the size of the
                        // axis label and padding.
                        // Just set opts here because axis will be sorted out on
                        // the redraw.

                        opts.labelHeight = axis.labelHeight +
                            axisLabels[axisName].height;
                        opts.labelWidth = axis.labelWidth +
                            axisLabels[axisName].width;
                    });

                    // If there are axis labels, re-draw with new label widths and
                    // heights.

                    if (hasAxisLabels) {
                        secondPass = true;
                        plot.setupGrid();
                        plot.draw();
                    }
                } else {
                    secondPass = false;
                    // DRAW
                    $.each(plot.getAxes(), function(axisName, axis) {
                        var opts = axis.options // Flot 0.7
                            || plot.getOptions()[axisName]; // Flot 0.6
                        if (!opts || !opts.axisLabel || !axis.show)
                            return;

                        axisLabels[axisName].draw(axis.box);
                    });
                }
            });
        });
    }


    $.plot.plugins.push({
        init: init,
        options: options,
        name: 'axisLabels',
        version: '2.0'
    });
})(jQuery);
/**
 * Copyright (c) 2011-2014 Felix Gnass
 * Licensed under the MIT license
 * http://spin.js.org/
 *
 * Example:
    var opts = {
      lines: 12             // The number of lines to draw
    , length: 7             // The length of each line
    , width: 5              // The line thickness
    , radius: 10            // The radius of the inner circle
    , scale: 1.0            // Scales overall size of the spinner
    , corners: 1            // Roundness (0..1)
    , color: '#000'         // #rgb or #rrggbb
    , opacity: 1/4          // Opacity of the lines
    , rotate: 0             // Rotation offset
    , direction: 1          // 1: clockwise, -1: counterclockwise
    , speed: 1              // Rounds per second
    , trail: 100            // Afterglow percentage
    , fps: 20               // Frames per second when using setTimeout()
    , zIndex: 2e9           // Use a high z-index by default
    , className: 'spinner'  // CSS class to assign to the element
    , top: '50%'            // center vertically
    , left: '50%'           // center horizontally
    , shadow: false         // Whether to render a shadow
    , hwaccel: false        // Whether to use hardware acceleration (might be buggy)
    , position: 'absolute'  // Element positioning
    }
    var target = document.getElementById('foo')
    var spinner = new Spinner(opts).spin(target)
 */
;(function (root, factory) {

  /* CommonJS */
  if (typeof module == 'object' && module.exports) module.exports = factory()

  /* AMD module */
  else if (typeof define == 'function' && define.amd) define(factory)

  /* Browser global */
  else root.Spinner = factory()
}(this, function () {
  "use strict"

  var prefixes = ['webkit', 'Moz', 'ms', 'O'] /* Vendor prefixes */
    , animations = {} /* Animation rules keyed by their name */
    , useCssAnimations /* Whether to use CSS animations or setTimeout */
    , sheet /* A stylesheet to hold the @keyframe or VML rules. */

  /**
   * Utility function to create elements. If no tag name is given,
   * a DIV is created. Optionally properties can be passed.
   */
  function createEl (tag, prop) {
    var el = document.createElement(tag || 'div')
      , n

    for (n in prop) el[n] = prop[n]
    return el
  }

  /**
   * Appends children and returns the parent.
   */
  function ins (parent /* child1, child2, ...*/) {
    for (var i = 1, n = arguments.length; i < n; i++) {
      parent.appendChild(arguments[i])
    }

    return parent
  }

  /**
   * Creates an opacity keyframe animation rule and returns its name.
   * Since most mobile Webkits have timing issues with animation-delay,
   * we create separate rules for each line/segment.
   */
  function addAnimation (alpha, trail, i, lines) {
    var name = ['opacity', trail, ~~(alpha * 100), i, lines].join('-')
      , start = 0.01 + i/lines * 100
      , z = Math.max(1 - (1-alpha) / trail * (100-start), alpha)
      , prefix = useCssAnimations.substring(0, useCssAnimations.indexOf('Animation')).toLowerCase()
      , pre = prefix && '-' + prefix + '-' || ''

    if (!animations[name]) {
      sheet.insertRule(
        '@' + pre + 'keyframes ' + name + '{' +
        '0%{opacity:' + z + '}' +
        start + '%{opacity:' + alpha + '}' +
        (start+0.01) + '%{opacity:1}' +
        (start+trail) % 100 + '%{opacity:' + alpha + '}' +
        '100%{opacity:' + z + '}' +
        '}', sheet.cssRules.length)

      animations[name] = 1
    }

    return name
  }

  /**
   * Tries various vendor prefixes and returns the first supported property.
   */
  function vendor (el, prop) {
    var s = el.style
      , pp
      , i

    prop = prop.charAt(0).toUpperCase() + prop.slice(1)
    if (s[prop] !== undefined) return prop
    for (i = 0; i < prefixes.length; i++) {
      pp = prefixes[i]+prop
      if (s[pp] !== undefined) return pp
    }
  }

  /**
   * Sets multiple style properties at once.
   */
  function css (el, prop) {
    for (var n in prop) {
      el.style[vendor(el, n) || n] = prop[n]
    }

    return el
  }

  /**
   * Fills in default values.
   */
  function merge (obj) {
    for (var i = 1; i < arguments.length; i++) {
      var def = arguments[i]
      for (var n in def) {
        if (obj[n] === undefined) obj[n] = def[n]
      }
    }
    return obj
  }

  /**
   * Returns the line color from the given string or array.
   */
  function getColor (color, idx) {
    return typeof color == 'string' ? color : color[idx % color.length]
  }

  // Built-in defaults

  var defaults = {
    lines: 12             // The number of lines to draw
  , length: 7             // The length of each line
  , width: 5              // The line thickness
  , radius: 10            // The radius of the inner circle
  , scale: 1.0            // Scales overall size of the spinner
  , corners: 1            // Roundness (0..1)
  , color: '#000'         // #rgb or #rrggbb
  , opacity: 1/4          // Opacity of the lines
  , rotate: 0             // Rotation offset
  , direction: 1          // 1: clockwise, -1: counterclockwise
  , speed: 1              // Rounds per second
  , trail: 100            // Afterglow percentage
  , fps: 20               // Frames per second when using setTimeout()
  , zIndex: 2e9           // Use a high z-index by default
  , className: 'spinner'  // CSS class to assign to the element
  , top: '50%'            // center vertically
  , left: '50%'           // center horizontally
  , shadow: false         // Whether to render a shadow
  , hwaccel: false        // Whether to use hardware acceleration (might be buggy)
  , position: 'absolute'  // Element positioning
  }

  /** The constructor */
  function Spinner (o) {
    this.opts = merge(o || {}, Spinner.defaults, defaults)
  }

  // Global defaults that override the built-ins:
  Spinner.defaults = {}

  merge(Spinner.prototype, {
    /**
     * Adds the spinner to the given target element. If this instance is already
     * spinning, it is automatically removed from its previous target b calling
     * stop() internally.
     */
    spin: function (target) {
      this.stop()

      var self = this
        , o = self.opts
        , el = self.el = createEl(null, {className: o.className})

      css(el, {
        position: o.position
      , width: 0
      , zIndex: o.zIndex
      , left: o.left
      , top: o.top
      })

      if (target) {
        target.insertBefore(el, target.firstChild || null)
      }

      el.setAttribute('role', 'progressbar')
      self.lines(el, self.opts)

      if (!useCssAnimations) {
        // No CSS animation support, use setTimeout() instead
        var i = 0
          , start = (o.lines - 1) * (1 - o.direction) / 2
          , alpha
          , fps = o.fps
          , f = fps / o.speed
          , ostep = (1 - o.opacity) / (f * o.trail / 100)
          , astep = f / o.lines

        ;(function anim () {
          i++
          for (var j = 0; j < o.lines; j++) {
            alpha = Math.max(1 - (i + (o.lines - j) * astep) % f * ostep, o.opacity)

            self.opacity(el, j * o.direction + start, alpha, o)
          }
          self.timeout = self.el && setTimeout(anim, ~~(1000 / fps))
        })()
      }
      return self
    }

    /**
     * Stops and removes the Spinner.
     */
  , stop: function () {
      var el = this.el
      if (el) {
        clearTimeout(this.timeout)
        if (el.parentNode) el.parentNode.removeChild(el)
        this.el = undefined
      }
      return this
    }

    /**
     * Internal method that draws the individual lines. Will be overwritten
     * in VML fallback mode below.
     */
  , lines: function (el, o) {
      var i = 0
        , start = (o.lines - 1) * (1 - o.direction) / 2
        , seg

      function fill (color, shadow) {
        return css(createEl(), {
          position: 'absolute'
        , width: o.scale * (o.length + o.width) + 'px'
        , height: o.scale * o.width + 'px'
        , background: color
        , boxShadow: shadow
        , transformOrigin: 'left'
        , transform: 'rotate(' + ~~(360/o.lines*i + o.rotate) + 'deg) translate(' + o.scale*o.radius + 'px' + ',0)'
        , borderRadius: (o.corners * o.scale * o.width >> 1) + 'px'
        })
      }

      for (; i < o.lines; i++) {
        seg = css(createEl(), {
          position: 'absolute'
        , top: 1 + ~(o.scale * o.width / 2) + 'px'
        , transform: o.hwaccel ? 'translate3d(0,0,0)' : ''
        , opacity: o.opacity
        , animation: useCssAnimations && addAnimation(o.opacity, o.trail, start + i * o.direction, o.lines) + ' ' + 1 / o.speed + 's linear infinite'
        })

        if (o.shadow) ins(seg, css(fill('#000', '0 0 4px #000'), {top: '2px'}))
        ins(el, ins(seg, fill(getColor(o.color, i), '0 0 1px rgba(0,0,0,.1)')))
      }
      return el
    }

    /**
     * Internal method that adjusts the opacity of a single line.
     * Will be overwritten in VML fallback mode below.
     */
  , opacity: function (el, i, val) {
      if (i < el.childNodes.length) el.childNodes[i].style.opacity = val
    }

  })


  function initVML () {

    /* Utility function to create a VML tag */
    function vml (tag, attr) {
      return createEl('<' + tag + ' xmlns="urn:schemas-microsoft.com:vml" class="spin-vml">', attr)
    }

    // No CSS transforms but VML support, add a CSS rule for VML elements:
    sheet.addRule('.spin-vml', 'behavior:url(#default#VML)')

    Spinner.prototype.lines = function (el, o) {
      var r = o.scale * (o.length + o.width)
        , s = o.scale * 2 * r

      function grp () {
        return css(
          vml('group', {
            coordsize: s + ' ' + s
          , coordorigin: -r + ' ' + -r
          })
        , { width: s, height: s }
        )
      }

      var margin = -(o.width + o.length) * o.scale * 2 + 'px'
        , g = css(grp(), {position: 'absolute', top: margin, left: margin})
        , i

      function seg (i, dx, filter) {
        ins(
          g
        , ins(
            css(grp(), {rotation: 360 / o.lines * i + 'deg', left: ~~dx})
          , ins(
              css(
                vml('roundrect', {arcsize: o.corners})
              , { width: r
                , height: o.scale * o.width
                , left: o.scale * o.radius
                , top: -o.scale * o.width >> 1
                , filter: filter
                }
              )
            , vml('fill', {color: getColor(o.color, i), opacity: o.opacity})
            , vml('stroke', {opacity: 0}) // transparent stroke to fix color bleeding upon opacity change
            )
          )
        )
      }

      if (o.shadow)
        for (i = 1; i <= o.lines; i++) {
          seg(i, -2, 'progid:DXImageTransform.Microsoft.Blur(pixelradius=2,makeshadow=1,shadowopacity=.3)')
        }

      for (i = 1; i <= o.lines; i++) seg(i)
      return ins(el, g)
    }

    Spinner.prototype.opacity = function (el, i, val, o) {
      var c = el.firstChild
      o = o.shadow && o.lines || 0
      if (c && i + o < c.childNodes.length) {
        c = c.childNodes[i + o]; c = c && c.firstChild; c = c && c.firstChild
        if (c) c.opacity = val
      }
    }
  }

  if (typeof document !== 'undefined') {
    sheet = (function () {
      var el = createEl('style', {type : 'text/css'})
      ins(document.getElementsByTagName('head')[0], el)
      return el.sheet || el.styleSheet
    }())

    var probe = css(createEl('group'), {behavior: 'url(#default#VML)'})

    if (!vendor(probe, 'transform') && probe.adj) initVML()
    else useCssAnimations = vendor(probe, 'animation')
  }

  return Spinner

}));
/**
 * Copyright Marc J. Schmidt. See the LICENSE file at the top-level
 * directory of this distribution and at
 * https://github.com/marcj/css-element-queries/blob/master/LICENSE.
 */
;
(function() {
    /**
     *
     * @type {Function}
     * @constructor
     */
    var ElementQueries = this.ElementQueries = function() {

        this.withTracking = false;
        var elements = [];

        /**
         *
         * @param element
         * @returns {Number}
         */
        function getEmSize(element) {
            if (!element) {
                element = document.documentElement;
            }
            var fontSize = getComputedStyle(element, 'fontSize');
            return parseFloat(fontSize) || 16;
        }

        /**
         *
         * @copyright https://github.com/Mr0grog/element-query/blob/master/LICENSE
         *
         * @param {HTMLElement} element
         * @param {*} value
         * @returns {*}
         */
        function convertToPx(element, value) {
            var units = value.replace(/[0-9]*/, '');
            value = parseFloat(value);
            switch (units) {
                case "px":
                    return value;
                case "em":
                    return value * getEmSize(element);
                case "rem":
                    return value * getEmSize();
                // Viewport units!
                // According to http://quirksmode.org/mobile/tableViewport.html
                // documentElement.clientWidth/Height gets us the most reliable info
                case "vw":
                    return value * document.documentElement.clientWidth / 100;
                case "vh":
                    return value * document.documentElement.clientHeight / 100;
                case "vmin":
                case "vmax":
                    var vw = document.documentElement.clientWidth / 100;
                    var vh = document.documentElement.clientHeight / 100;
                    var chooser = Math[units === "vmin" ? "min" : "max"];
                    return value * chooser(vw, vh);
                default:
                    return value;
                // for now, not supporting physical units (since they are just a set number of px)
                // or ex/ch (getting accurate measurements is hard)
            }
        }

        /**
         *
         * @param {HTMLElement} element
         * @constructor
         */
        function SetupInformation(element) {
            this.element = element;
            this.options = {};
            var key, option, width = 0, height = 0, value, actualValue, attrValues, attrValue, attrName;

            /**
             * @param {Object} option {mode: 'min|max', property: 'width|height', value: '123px'}
             */
            this.addOption = function(option) {
                var idx = [option.mode, option.property, option.value].join(',');
                this.options[idx] = option;
            };

            var attributes = ['min-width', 'min-height', 'max-width', 'max-height'];

            /**
             * Extracts the computed width/height and sets to min/max- attribute.
             */
            this.call = function() {
                // extract current dimensions
                width = this.element.offsetWidth;
                height = this.element.offsetHeight;

                attrValues = {};

                for (key in this.options) {
                    if (!this.options.hasOwnProperty(key)){
                        continue;
                    }
                    option = this.options[key];

                    value = convertToPx(this.element, option.value);

                    actualValue = option.property == 'width' ? width : height;
                    attrName = option.mode + '-' + option.property;
                    attrValue = '';

                    if (option.mode == 'min' && actualValue >= value) {
                        attrValue += option.value;
                    }

                    if (option.mode == 'max' && actualValue <= value) {
                        attrValue += option.value;
                    }

                    if (!attrValues[attrName]) attrValues[attrName] = '';
                    if (attrValue && -1 === (' '+attrValues[attrName]+' ').indexOf(' ' + attrValue + ' ')) {
                        attrValues[attrName] += ' ' + attrValue;
                    }
                }

                for (var k in attributes) {
                    if (attrValues[attributes[k]]) {
                        this.element.setAttribute(attributes[k], attrValues[attributes[k]].substr(1));
                    } else {
                        this.element.removeAttribute(attributes[k]);
                    }
                }
            };
        }

        /**
         * @param {HTMLElement} element
         * @param {Object}      options
         */
        function setupElement(element, options) {
            if (element.elementQueriesSetupInformation) {
                element.elementQueriesSetupInformation.addOption(options);
            } else {
                element.elementQueriesSetupInformation = new SetupInformation(element);
                element.elementQueriesSetupInformation.addOption(options);
                element.elementQueriesSensor = new ResizeSensor(element, function() {
                    element.elementQueriesSetupInformation.call();
                });
            }
            element.elementQueriesSetupInformation.call();

            if (ElementQueries.instance.withTracking && elements.indexOf(element) < 0) {
                elements.push(element);
            }
        }

        /**
         * @param {String} selector
         * @param {String} mode min|max
         * @param {String} property width|height
         * @param {String} value
         */
        function queueQuery(selector, mode, property, value) {
            var query;
            if (document.querySelectorAll) query = document.querySelectorAll.bind(document);
            if (!query && 'undefined' !== typeof $$) query = $$;
            if (!query && 'undefined' !== typeof jQuery) query = jQuery;

            if (!query) {
                throw 'No document.querySelectorAll, jQuery or Mootools\'s $$ found.';
            }

            var elements = query(selector);
            for (var i = 0, j = elements.length; i < j; i++) {
                setupElement(elements[i], {
                    mode: mode,
                    property: property,
                    value: value
                });
            }
        }

        var regex = /,?([^,\n]*?)\[[\s\t]*?(min|max)-(width|height)[\s\t]*?[~$\^]?=[\s\t]*?"([^"]*?)"[\s\t]*?]([^\n\s\{]*?)/mgi;

        /**
         * @param {String} css
         */
        function extractQuery(css) {
            var match;
            var smatch;
            css = css.replace(/'/g, '"');
            while (null !== (match = regex.exec(css))) {
                if (5 < match.length) {
                    smatch = match[1] || match[5] || smatch;
                    queueQuery(smatch, match[2], match[3], match[4]);
                }
            }
        }

        /**
         * @param {CssRule[]|String} rules
         */
        function readRules(rules) {
            var selector = '';
            if (!rules) {
                return;
            }
            if ('string' === typeof rules) {
                rules = rules.toLowerCase();
                if (-1 !== rules.indexOf('min-width') || -1 !== rules.indexOf('max-width')) {
                    extractQuery(rules);
                }
            } else {
                for (var i = 0, j = rules.length; i < j; i++) {
                    if (1 === rules[i].type) {
                        selector = rules[i].selectorText || rules[i].cssText;
                        if (-1 !== selector.indexOf('min-height') || -1 !== selector.indexOf('max-height')) {
                            extractQuery(selector);
                        }else if(-1 !== selector.indexOf('min-width') || -1 !== selector.indexOf('max-width')) {
                            extractQuery(selector);
                        }
                    } else if (4 === rules[i].type) {
                        readRules(rules[i].cssRules || rules[i].rules);
                    }
                }
            }
        }

        /**
         * Searches all css rules and setups the event listener to all elements with element query rules..
         *
         * @param {Boolean} withTracking allows and requires you to use detach, since we store internally all used elements
         *                               (no garbage collection possible if you don not call .detach() first)
         */
        this.init = function(withTracking) {
            this.withTracking = withTracking;
            for (var i = 0, j = document.styleSheets.length; i < j; i++) {
                try {
                    readRules(document.styleSheets[i].cssText || document.styleSheets[i].cssRules || document.styleSheets[i].rules);
                } catch(e) {
                    if (e.name !== 'SecurityError') {
                        throw e;
                    }
                }
            }
        };

        /**
         *
         * @param {Boolean} withTracking allows and requires you to use detach, since we store internally all used elements
         *                               (no garbage collection possible if you don not call .detach() first)
         */
        this.update = function(withTracking) {
            this.withTracking = withTracking;
            this.init();
        };

        this.detach = function() {
            if (!this.withTracking) {
                throw 'withTracking is not enabled. We can not detach elements since we don not store it.' +
                'Use ElementQueries.withTracking = true; before domready.';
            }

            var element;
            while (element = elements.pop()) {
                ElementQueries.detach(element);
            }

            elements = [];
        };
    };

    /**
     *
     * @param {Boolean} withTracking allows and requires you to use detach, since we store internally all used elements
     *                               (no garbage collection possible if you don not call .detach() first)
     */
    ElementQueries.update = function(withTracking) {
        ElementQueries.instance.update(withTracking);
    };

    /**
     * Removes all sensor and elementquery information from the element.
     *
     * @param {HTMLElement} element
     */
    ElementQueries.detach = function(element) {
        if (element.elementQueriesSetupInformation) {
            element.elementQueriesSensor.detach();
            delete element.elementQueriesSetupInformation;
            delete element.elementQueriesSensor;
            console.log('detached');
        } else {
            console.log('detached already', element);
        }
    };

    ElementQueries.withTracking = false;

    ElementQueries.init = function() {
        if (!ElementQueries.instance) {
            ElementQueries.instance = new ElementQueries();
        }

        ElementQueries.instance.init(ElementQueries.withTracking);
    };

    var domLoaded = function (callback) {
        /* Internet Explorer */
        /*@cc_on
        @if (@_win32 || @_win64)
            document.write('<script id="ieScriptLoad" defer src="//:"><\/script>');
            document.getElementById('ieScriptLoad').onreadystatechange = function() {
                if (this.readyState == 'complete') {
                    callback();
                }
            };
        @end @*/
        /* Mozilla, Chrome, Opera */
        if (document.addEventListener) {
            document.addEventListener('DOMContentLoaded', callback, false);
        }
        /* Safari, iCab, Konqueror */
        if (/KHTML|WebKit|iCab/i.test(navigator.userAgent)) {
            var DOMLoadTimer = setInterval(function () {
                if (/loaded|complete/i.test(document.readyState)) {
                    callback();
                    clearInterval(DOMLoadTimer);
                }
            }, 10);
        }
        /* Other web browsers */
        window.onload = callback;
    };

    if (window.addEventListener) {
        window.addEventListener('load', ElementQueries.init, false);
    } else {
        window.attachEvent('onload', ElementQueries.init);
    }
    domLoaded(ElementQueries.init);

})();
/**
 * Copyright Marc J. Schmidt. See the LICENSE file at the top-level
 * directory of this distribution and at
 * https://github.com/marcj/css-element-queries/blob/master/LICENSE.
 */
;
(function() {

    /**
     * Class for dimension change detection.
     *
     * @param {Element|Element[]|Elements|jQuery} element
     * @param {Function} callback
     *
     * @constructor
     */
    this.ResizeSensor = function(element, callback) {
        /**
         *
         * @constructor
         */
        function EventQueue() {
            this.q = [];
            this.add = function(ev) {
                this.q.push(ev);
            };

            var i, j;
            this.call = function() {
                for (i = 0, j = this.q.length; i < j; i++) {
                    this.q[i].call();
                }
            };
        }

        /**
         * @param {HTMLElement} element
         * @param {String}      prop
         * @returns {String|Number}
         */
        function getComputedStyle(element, prop) {
            if (element.currentStyle) {
                return element.currentStyle[prop];
            } else if (window.getComputedStyle) {
                return window.getComputedStyle(element, null).getPropertyValue(prop);
            } else {
                return element.style[prop];
            }
        }

        /**
         *
         * @param {HTMLElement} element
         * @param {Function}    resized
         */
        function attachResizeEvent(element, resized) {
            if (!element.resizedAttached) {
                element.resizedAttached = new EventQueue();
                element.resizedAttached.add(resized);
            } else if (element.resizedAttached) {
                element.resizedAttached.add(resized);
                return;
            }

            element.resizeSensor = document.createElement('div');
            element.resizeSensor.className = 'resize-sensor';
            var style = 'position: absolute; left: 0; top: 0; right: 0; bottom: 0; overflow: scroll; z-index: -1; visibility: hidden;';
            var styleChild = 'position: absolute; left: 0; top: 0;';

            element.resizeSensor.style.cssText = style;
            element.resizeSensor.innerHTML =
                '<div class="resize-sensor-expand" style="' + style + '">' +
                    '<div style="' + styleChild + '"></div>' +
                '</div>' +
                '<div class="resize-sensor-shrink" style="' + style + '">' +
                    '<div style="' + styleChild + ' width: 200%; height: 200%"></div>' +
                '</div>';
            element.appendChild(element.resizeSensor);

            if (!{fixed: 1, absolute: 1}[getComputedStyle(element, 'position')]) {
                element.style.position = 'relative';
            }

            var expand = element.resizeSensor.childNodes[0];
            var expandChild = expand.childNodes[0];
            var shrink = element.resizeSensor.childNodes[1];
            var shrinkChild = shrink.childNodes[0];

            var lastWidth, lastHeight;

            var reset = function() {
                expandChild.style.width = expand.offsetWidth + 10 + 'px';
                expandChild.style.height = expand.offsetHeight + 10 + 'px';
                expand.scrollLeft = expand.scrollWidth;
                expand.scrollTop = expand.scrollHeight;
                shrink.scrollLeft = shrink.scrollWidth;
                shrink.scrollTop = shrink.scrollHeight;
                lastWidth = element.offsetWidth;
                lastHeight = element.offsetHeight;
            };

            reset();

            var changed = function() {
                if (element.resizedAttached) {
                    element.resizedAttached.call();
                }
            };

            var addEvent = function(el, name, cb) {
                if (el.attachEvent) {
                    el.attachEvent('on' + name, cb);
                } else {
                    el.addEventListener(name, cb);
                }
            };
            
            var onScroll = function() {
              if (element.offsetWidth != lastWidth || element.offsetHeight != lastHeight) {
                  changed();
              }
              reset();
            }

            addEvent(expand, 'scroll', onScroll);
            addEvent(shrink, 'scroll', onScroll);
        }

        var elementType = Object.prototype.toString.call(element);
        var isCollectionTyped = ('[object Array]' === elementType
            || ('[object NodeList]' === elementType)
            || ('[object HTMLCollection]' === elementType)
            || ('undefined' !== typeof jQuery && element instanceof jQuery) //jquery
            || ('undefined' !== typeof Elements && element instanceof Elements) //mootools
        );

        if (isCollectionTyped) {
            var i = 0, j = element.length;
            for (; i < j; i++) {
                attachResizeEvent(element[i], callback);
            }
        } else {
            attachResizeEvent(element, callback);
        }

        this.detach = function() {
            if (isCollectionTyped) {
                var i = 0, j = element.length;
                for (; i < j; i++) {
                    ResizeSensor.detach(element[i]);
                }
            } else {
                ResizeSensor.detach(element);
            }
        };
    };

    this.ResizeSensor.detach = function(element) {
        if (element.resizeSensor) {
            element.removeChild(element.resizeSensor);
            delete element.resizeSensor;
            delete element.resizedAttached;
        }
    };

})();
// fast gausian blur
// taken from: blog.ivank.net/fastest-gaussian-blur.html
// (and http://elynxsdk.free.fr/ext-docs/Blur/Fast_box_blur.pdf)
// modified by: Eric Mandel (2/16/2016):
//   -- to use floating point data arrays
//   -- to ignore NaN values by converting them to 0
var gaussBlur = (function(){

function gaussBlur (scl, tcl, w, h, r) {
    var bxs = boxesForGauss(r, 3);
    boxBlur (scl, tcl, w, h, (bxs[0]-1)/2);
    boxBlur (tcl, scl, w, h, (bxs[1]-1)/2);
    boxBlur (scl, tcl, w, h, (bxs[2]-1)/2);
}

function boxesForGauss(sigma, n)  // standard deviation, number of boxes
{
    var wIdeal = Math.sqrt((12*sigma*sigma/n)+1);  // Ideal averaging filter width 
    var wl = Math.floor(wIdeal);  if(wl%2==0) wl--;
    var wu = wl+2;
				
    var mIdeal = (12*sigma*sigma - n*wl*wl - 4*n*wl - 3*n)/(-4*wl - 4);
    var m = Math.round(mIdeal);
    // var sigmaActual = Math.sqrt( (m*wl*wl + (n-m)*wu*wu - n)/12 );
				
    var sizes = [];  for(var i=0; i<n; i++) sizes.push(i<m?wl:wu);
    return sizes;
}

function boxBlur (scl, tcl, w, h, r) {
    for(var i=0; i<scl.length; i++){
	if( scl[i] === scl[i] ){
	    tcl[i] = scl[i];
	} else {
	    tcl[i] = 0;
	}
    }
    boxBlurH(tcl, scl, w, h, r);
    boxBlurT(scl, tcl, w, h, r);
}

function boxBlurH (scl, tcl, w, h, r) {
    var iarr = 1 / (r+r+1);
    for(var i=0; i<h; i++) {
        var ti = i*w, li = ti, ri = ti+r;
        var fv = scl[ti], lv = scl[ti+w-1], val = (r+1)*fv;
        for(var j=0; j<r; j++) val += scl[ti+j];
        for(var j=0  ; j<=r ; j++) { val += scl[ri++] - fv       ;   tcl[ti++] = val*iarr; }
        for(var j=r+1; j<w-r; j++) { val += scl[ri++] - scl[li++];   tcl[ti++] = val*iarr; }
        for(var j=w-r; j<w  ; j++) { val += lv        - scl[li++];   tcl[ti++] = val*iarr; }
    }
}

function boxBlurT (scl, tcl, w, h, r) {
    var iarr = 1 / (r+r+1);
    for(var i=0; i<w; i++) {
        var ti = i, li = ti, ri = ti+r*w;
        var fv = scl[ti], lv = scl[ti+w*(h-1)], val = (r+1)*fv;
        for(var j=0; j<r; j++) val += scl[ti+j*w];
        for(var j=0  ; j<=r ; j++) { val += scl[ri] - fv     ;  tcl[ti] = val*iarr;  ri+=w; ti+=w; }
        for(var j=r+1; j<h-r; j++) { val += scl[ri] - scl[li];  tcl[ti] = val*iarr;  li+=w; ri+=w; ti+=w; }
        for(var j=h-r; j<h  ; j++) { val += lv      - scl[li];  tcl[ti] = val*iarr;  li+=w; ti+=w; }
    }
}

return gaussBlur;

}());


// Image Processing Filters adapted from:
//   https://code.msdn.microsoft.com/Image-Filters-in-Javascript-0853f6c1
// and:
//   https://github.com/kig/canvasfilters
// which look to be adapted from:
//   http://www.html5rocks.com/en/tutorials/canvas/imagefilters/

/*jslint plusplus: true, vars: true, white: true  */

/*global Float32Array */

var ImageFilters;
ImageFilters = (function(){
"use strict";

    function normalize(array){
	var i, sum;
	sum = array.reduce( function (a, b){ return a+b; } );
	for(i=0; i<array.length; ++i){
	    array[i] /= sum;
	}
    }
    
    function convolve(ctx, pixels, weights, opaque, norm){
	var i, x, y, sx, sy, scx, scy, dstOff, cx, cy, srcOff, wt, r, g, b, a;
	var side = Math.round(Math.sqrt(weights.length));
	var halfSide = Math.floor(side/2);
	var sw = pixels.width;
	var sh = pixels.height;
	var swh = sw * sh * 4;
	// pad output by the convolution matrix
	var w = sw;
	var h = sh;
	var dst = pixels.data;
	// src data is a copy of the input (which we then return in place)
	var src = ctx.createImageData(pixels.width, pixels.height);
	for(i=0; i<swh; i++){
	    src[i] = dst[i];
	}
	// weights array is required
	if( Object.prototype.toString.call(weights) !== '[object Array]' ) {
	    return pixels;
	}
	// default is opaque
	if( opaque === undefined ){
	    opaque = true;
	}
	var alphaFac = opaque ? 1 : 0;
	// default is not to normalize
	if( norm ){
	    normalize(weights);
	}
	// go through the destination image pixels
	for (y = 0; y < h; y++){
	    for (x = 0; x < w; x++){
		sy = y;
		sx = x;
		dstOff = (y*w+x)*4;
		// calculate the weighed sum of the source image pixels that
		// fall under the convolution matrix
		r = 0; g = 0; b = 0; a = 0;
		for (cy = 0; cy < side; cy++){
		    for (cx = 0; cx < side; cx++){
			// scy = sy + cy - halfSide;
			// scx = sx + cx - halfSide;
			// if (scy >= 0 && scy < sh && scx >= 0 && scx < sw){
			scy = Math.min(sh-1, Math.max(0, sy + cy - halfSide));
			scx = Math.min(sw-1, Math.max(0, sx + cx - halfSide));
			srcOff = (scy * sw + scx)*4;
			wt = weights[cy * side + cx];
			r += src[srcOff] * wt;
			g += src[srcOff+1] * wt;
			b += src[srcOff+2] * wt;
			a += src[srcOff+3] * wt;
			// }
		    }
		}
		dst[dstOff] = r;
		dst[dstOff+1] = g;
		dst[dstOff+2] = b;
		dst[dstOff+3] = a + alphaFac*(255-a);
	    }	
	}
	return pixels;
    }
    
    function convolveFloat32(ctx, pixels, weights, opaque){
	var x, y, sx, sy, scx, scy, dstOff, cx, cy, srcOff, wt, r, g, b, a;
        var side = Math.round(Math.sqrt(weights.length));
        var halfSide = Math.floor(side/2);
	
        var src = pixels.data;
        var sw = pixels.width;
        var sh = pixels.height;
	
        var w = sw;
        var h = sh;
        var output = {
            width: w, height: h, data: new Float32Array(w*h*4)
        };
        var dst = output.data;
        var alphaFac = opaque ? 1 : 0;
        for (y = 0; y < h; y++){
            for (x = 0; x < w; x++){
		sy = y;
		sx = x;
		dstOff = (y * w + x) * 4;
		r = 0; g = 0; b = 0; a = 0;
		for (cy = 0; cy < side; cy++){
		    for (cx = 0; cx < side; cx++){
			scy = Math.min(sh-1, Math.max(0, sy + cy - halfSide));
			scx = Math.min(sw-1, Math.max(0, sx + cx - halfSide));
			srcOff = (scy*sw + scx)*4;
			wt = weights[cy*side + cx];
			r += src[srcOff] * wt;
			g += src[srcOff+1] * wt;
			b += src[srcOff+2] * wt;
			a += src[srcOff+3] * wt;
		    }
		}
		dst[dstOff] = r;
		dst[dstOff+1] = g;
		dst[dstOff+2] = b;
		dst[dstOff+3] = a + alphaFac*(255-a);
            }
        }
        return output.data;
    }
    
    function luminance(ctx, imageData){
	var i, v;
	var data = imageData.data;
	for(i = 0; i < data.length; i += 4){
	    // CIE luminance for the RGB
	    // Human eye is bad at seeing red and blue, so de-emphasize them.
	    v = 0.2126*data[i] + 0.7152*data[i+1] + 0.0722*data[i+2];
	    data[i] = data[i+1] = data[i+2] = v;
	}
	return imageData;
    }
    
    function greyscale(ctx, imageData){
	var i, v;
	var data = imageData.data;
	for(i = 0; i < data.length; i += 4){
	    v = 0.299*data[i] + 0.587*data[i+1] + 0.114*data[i+2];
	    data[i] = data[i+1] = data[i+2] = v;
	}
	return imageData;
    }
    
    function greyscaleAvg(ctx, imageData){
	var i, v;
	var data = imageData.data;
	for(i = 0; i < data.length; i += 4){
	    v = (data[i] + data[i+1] + data[i+2])/3;
	    data[i] = data[i+1] = data[i+2] = v;
	}
	return imageData;
    }
    
    function duotone(ctx, imageData, which) {
	var i, r;
	var data = imageData.data;
	switch(which){
	case "r":
	    which = 0;
	    break;
	case "g":
	    which = 1;
	    break;
	case "b":
	    which = 2;
	    break;
	default:
	    which = 0;
	    break;
	}
	for(i = 0; i < data.length; i += 4) {
	    // average the other two colors for this pixel
	    r = (data[i] + data[i+1] + data[i+2] - data[i+which])/2;
	    data[i+which] = r;
	}
	return imageData;
    }

    function noise(ctx, imageData, lower, upper){
	var i, radius, mod;
	var data = imageData.data;
	if(arguments.length < 3){
	    lower = -30;
	    upper = 30;
	}
	radius = upper - lower;
	for(i = 0; i < data.length; i += 4){
	    mod = Math.floor((Math.random() * radius) - upper);
	    data[i] += mod;
	    data[i+1] += mod;
	    data[i+2] += mod;
	}
	return imageData;
    }
    
    function invert(ctx, imageData){
	var i;
	var data = imageData.data;
	for(i = 0; i < data.length; i += 4){
	    data[i] = 255 - data[i];
	    data[i+1] = 255 - data[i+1];
	    data[i+2] = 255 - data[i+2];
	}
	return imageData;
    }
    
    function pixelate(ctx, imageData, radius){
	var r, g, b, idx, odx, ody;
	var x, y, nx, ny, delta;
	var width = imageData.width;
	var height = imageData.height;
	var data = imageData.data;
	if(arguments.length < 3){
	    radius = 2;
	}
	delta = (2 * radius);
	for(y = radius; y < height - radius; y += delta){
	    for(x = radius; x < width - radius; x += delta){
		idx = ((width * y) + x) * 4;
		r = data[idx];
		g = data[idx + 1];
		b = data[idx + 2];
		for(ny = -radius; ny < radius; ++ny){
		    ody = (width*(y + ny));
		    for(nx = -radius; nx < radius; ++nx){
			odx = (ody + (x + nx)) * 4;
			data[odx] = r; 
			data[odx+1] = g;
			data[odx+2] = b; 
		    }
		}
	    }
	}
	return imageData;
    }
    
    function gamma(ctx, imageData, amount){
	var i;
	var data = imageData.data;
	if(arguments.length < 3){
	    amount = 0.2;
	}
	for(i = 0; i < data.length; i += 4){
	    data[i] = 255 * Math.pow(data[i] / 255, amount); 
	    data[i+1] = 255 * Math.pow(data[i+1] / 255, amount);
	    data[i+2] = 255 * Math.pow(data[i+2] / 255, amount);
	}
	return imageData;
    }
    
    function brighten(ctx, imageData, a){
	var i;
	var data = imageData.data;
	a = a || 10;
	for(i = 0; i < data.length; i += 4){
	    data[i] += a;
	    data[i+1] += a;
	    data[i+2] += a;
	}
	return imageData;
    }
    
    function sepia(ctx, imageData){
	var i, r, g, b;
	var data = imageData.data;
	for(i = 0; i < data.length; i += 4){
	    r = data[i];
	    g = data[i+1];
	    b = data[i+2];
	    data[i] = (r * 0.393) + (g * 0.769) + (b * 0.189);
	    data[i+1] = (r * 0.349) + (g * 0.686) + (b * 0.168);
	    data[i+2] = (r * 0.272) + (g * 0.534) + (b * 0.131);
	}
	return imageData;
    }
    
    function contrast(ctx, imageData, amount){
	var i;
	var data = imageData.data;
	if(arguments.length < 3){
	    amount = 2;
	}
	if(amount < -100){
	    amount = -100;
	} else if(amount > 100){
	    amount = 100;
	}
	for(i = 0; i < data.length; i += 4){
	    data[i] = ((((data[i] / 255) - 0.5) * amount) + 0.5) * 255;
	    data[i+1] = ((((data[i+1] / 255) - 0.5) * amount) + 0.5) * 255;
	    data[i+2] = ((((data[i+2] / 255) - 0.5) * amount) + 0.5) * 255;
	}
	return imageData;
    }
    
    function threshold(ctx, imageData, thresh, low, high){
	var i, r, g, b, val;
	var data = imageData.data;
	if(arguments.length < 3){
	    thresh = 10;
	    low = 0;
	    high = 255;
	} else if(arguments.length < 5){
	    low = 0;
	    high = 255;
	}
	for(i = 0; i < data.length-1; i += 4){
	    r = data[i];
	    g = data[i+1];
	    b = data[i+2];
	    val = (0.2126*r + 0.7152*g + 0.0722*b >= thresh) ? high : low;
	    data[i] = data[i+1] = data[i+2] = val;
	}
	return imageData;
    }

    function medianFilter(ctx, imageData){
	return convolve(ctx, imageData,
			[1,0,0,0,1,0,1,0,1,0,1,1,1,1,1,0,1,0,1,0,1,0,0,0,1],
			true, true);
    }
    
    function sobel(ctx, imageData, colorize){
	var i, v, h;
	var vertical, horizontal;
	var data = imageData.data;
	imageData = greyscaleAvg(ctx, imageData);
	horizontal = convolveFloat32(ctx, imageData, 
				     [ -1, 0, 1, -2, 0, 2, -1, 0, 1 ]);
	vertical   = convolveFloat32(ctx, imageData,
				     [ -1, -2, -1, 0, 0, 0, 1, 2, 1 ]);
	if(colorize){
	    for (i=0; i<data.length; i+=4){
		v = Math.abs(vertical[i]);
		h = Math.abs(horizontal[i]);
		data[i] = v;
		data[i+1] = h;
		data[i+2] = (v+h)/4;
	    }
	} else {
	    for (i=0; i<data.length; i+=4){
		v = Math.abs(vertical[i]);
		h = Math.abs(horizontal[i]);
		data[i] = v;
		data[i+1] = h;
		data[i+2] = (v+h)/2;
	    }
	    greyscale(ctx, imageData);
	}
	return imageData;
    }
    
    function gaussBlur5(ctx, imageData){
	return convolve(ctx, imageData, [ 
	    0.0030, 0.0133, 0.0219, 0.0133, 0.0030,
	    0.0133, 0.0596, 0.0983, 0.0596, 0.0133,
	    0.0219, 0.0983, 0.1621, 0.0983, 0.0219,
	    0.0133, 0.0596, 0.0983, 0.0596, 0.0133,
	    0.0030, 0.0133, 0.0219, 0.0133, 0.0030 ]);
    }
    
    function posterize(ctx, imageData, colors){
	var i, r, g, b;
	var levels = [];
	var level = 1;
	var data = imageData.data;
	if(arguments.length < 3){
	    colors = 40;
	}
	for(i = 0; i < 256; ++i){
	    if( i < (colors * level) ){
		levels[i] = colors * (level - 1);
	    } else {
		levels[i] = (colors * level);
		++level;
	    }
	}
	for(i = 0; i < data.length; i += 4){
	    r = data[i];
	    g = data[i+1];
	    b = data[i+2];
	    data[i] = levels[r];
	    data[i+1] = levels[g];
	    data[i+2] = levels[b];
	}
	return imageData;
    }

    function scatter(ctx, imageData, radius){
	var x, y, swap, nswaps, delta, total;
	var px1, py1, px2, py2;
	var width = imageData.width;
	var height = imageData.height;
	var data = imageData.data;
	var swapPixel = function (x1, y1, x2, y2){
	    var temp = [];
	    var w1 = ((width * y1) + x1) * 4;
	    var w2 = ((width * y2) + x2) * 4;
	    temp[0] = data[w1];
	    temp[1] = data[w1+1];
	    temp[2] = data[w1+2];
	    temp[3] = data[w1+3];
	    data[w1]   = data[w2];
	    data[w1+1] = data[w2+1];
	    data[w1+2] = data[w2+2];
	    data[w1+3] = data[w2+3];
	    data[w2]   = temp[0];
	    data[w2+1] = temp[1];
	    data[w2+2] = temp[2];
	    data[w2+3] = temp[3];
	};
	if(arguments.length < 3){
	    radius = 5;
	}
	delta = 2 * radius;
	total = (radius - 1) * radius;
	for(y = radius; y < height - radius; y += delta){
	    for(x = radius; x < width - radius; x += delta){
		swap = false;
		nswaps = 0;
		while(!swap){
		    px1 = Math.floor((Math.random() * delta) - radius);
		    py1 = Math.floor((Math.random() * delta) - radius);
		    px2 = Math.floor((Math.random() * delta) - radius);
		    py2 = Math.floor((Math.random() * delta) - radius);
		    swapPixel(px1 + x, py1 + y, px2 + x, py2 + y);
		    ++nswaps;
		    if(nswaps > total){
			swap = true;
		    }
		}
	    }
	}
	return imageData;
    }
    
    function solarize(ctx, imageData, threshold){
	var i, r, g, b, intensity;
	var data = imageData.data;
	if(arguments.length < 3){
	    threshold = 50;
	}
	for(i = 0; i < data.length; ++i){
	    r = data[i];
	    g = data[i+1];
	    b = data[i+2];
	    intensity = (r + g + b)/3.0;
	    if(intensity < threshold){
		data[i] = 255 - r;
		data[i+1] = 255 - g;
		data[i+2] = 255 - b;
	    }
	}
	return imageData;
    }
	
    function edgeDetect(ctx, imageData){
	return convolve(ctx, imageData, 
			 [ -1, -1, -1, -1, 8, -1, -1, -1, -1 ]);
    }
	
    function sharpen(ctx, imageData, amount){
	if(arguments.length < 3){
	    amount = 20;
	}
	return convolve(ctx, imageData,
			[ 0, -3,  0, -3, amount, -3, 0, -3, 0 ], 
			true, true);
    }
	
    function blur(ctx, imageData){
	return convolve(ctx, imageData,
			[ 1, 2, 1, 2, 1, 2, 1, 2, 1 ], 
			true, true);
    }
	
    function emboss(ctx, imageData, amount){
	if(arguments.length < 3){
	    amount = 95;
	}
	return convolve(ctx, imageData,
			 [ -18, -9, 9, -9, 100 - amount, 9, 0, 9, 18 ], 
			 true, true);
    }
	
    function lighten(ctx, imageData, amount){
	if(arguments.length < 3 || amount < 0){
	    amount = 12/9;
	}
	return convolve(ctx, imageData, [ 0, 0, 0, 0, amount, 0, 0, 0, 0 ]);
    }
	
    function darken(ctx, imageData, amount){
	if(arguments.length < 3 || amount > 1){
	    amount = 6/9;
	}
	return convolve(ctx, imageData, [ 0, 0, 0, 0, amount, 0, 0, 0, 0 ]);
    }
	
    return {
	convolve : convolve,
	luminance : luminance,
	greyscale : greyscale,
	greyscaleAvg : greyscaleAvg,
	brighten : brighten,
	noise : noise,
	duotone : duotone,
	invert : invert,
	pixelate : pixelate,
	sobel : sobel,
	medianFilter : medianFilter,
	sepia : sepia,
	contrast : contrast,
	threshold : threshold,
	gamma : gamma,
	gaussBlur5 : gaussBlur5,
	posterize : posterize,
	scatter : scatter,
	solarize : solarize,
	sharpen : sharpen,
	edgeDetect : edgeDetect,
	blur : blur,
	emboss : emboss,
	lighten : lighten,
	darken : darken
    };
}());
/*!
 * jQuery UI Touch Punch 0.2.3
 *
 * Copyright 2011–2014, Dave Furfero
 * Dual licensed under the MIT or GPL Version 2 licenses.
 *
 * Depends:
 *  jquery.ui.widget.js
 *  jquery.ui.mouse.js
 */
(function ($) {

  // Detect touch support
  $.support.touch = 'ontouchend' in document;

  // Ignore browsers without touch support
  if (!$.support.touch) {
    return;
  }

  var mouseProto = $.ui.mouse.prototype,
      _mouseInit = mouseProto._mouseInit,
      _mouseDestroy = mouseProto._mouseDestroy,
      touchHandled;

  /**
   * Simulate a mouse event based on a corresponding touch event
   * @param {Object} event A touch event
   * @param {String} simulatedType The corresponding mouse event
   */
  function simulateMouseEvent (event, simulatedType) {

    // Ignore multi-touch events
    if (event.originalEvent.touches.length > 1) {
      return;
    }

    event.preventDefault();

    var touch = event.originalEvent.changedTouches[0],
        simulatedEvent = document.createEvent('MouseEvents');
    
    // Initialize the simulated mouse event using the touch event's coordinates
    simulatedEvent.initMouseEvent(
      simulatedType,    // type
      true,             // bubbles                    
      true,             // cancelable                 
      window,           // view                       
      1,                // detail                     
      touch.screenX,    // screenX                    
      touch.screenY,    // screenY                    
      touch.clientX,    // clientX                    
      touch.clientY,    // clientY                    
      false,            // ctrlKey                    
      false,            // altKey                     
      false,            // shiftKey                   
      false,            // metaKey                    
      0,                // button                     
      null              // relatedTarget              
    );

    // Dispatch the simulated event to the target element
    event.target.dispatchEvent(simulatedEvent);
  }

  /**
   * Handle the jQuery UI widget's touchstart events
   * @param {Object} event The widget element's touchstart event
   */
  mouseProto._touchStart = function (event) {

    var self = this;

    // Ignore the event if another widget is already being handled
    if (touchHandled || !self._mouseCapture(event.originalEvent.changedTouches[0])) {
      return;
    }

    // Set the flag to prevent other widgets from inheriting the touch event
    touchHandled = true;

    // Track movement to determine if interaction was a click
    self._touchMoved = false;

    // Simulate the mouseover event
    simulateMouseEvent(event, 'mouseover');

    // Simulate the mousemove event
    simulateMouseEvent(event, 'mousemove');

    // Simulate the mousedown event
    simulateMouseEvent(event, 'mousedown');
  };

  /**
   * Handle the jQuery UI widget's touchmove events
   * @param {Object} event The document's touchmove event
   */
  mouseProto._touchMove = function (event) {

    // Ignore event if not handled
    if (!touchHandled) {
      return;
    }

    // Interaction was not a click
    this._touchMoved = true;

    // Simulate the mousemove event
    simulateMouseEvent(event, 'mousemove');
  };

  /**
   * Handle the jQuery UI widget's touchend events
   * @param {Object} event The document's touchend event
   */
  mouseProto._touchEnd = function (event) {

    // Ignore event if not handled
    if (!touchHandled) {
      return;
    }

    // Simulate the mouseup event
    simulateMouseEvent(event, 'mouseup');

    // Simulate the mouseout event
    simulateMouseEvent(event, 'mouseout');

    // If the touch interaction did not move, it should trigger a click
    if (!this._touchMoved) {

      // Simulate the click event
      simulateMouseEvent(event, 'click');
    }

    // Unset the flag to allow other widgets to inherit the touch event
    touchHandled = false;
  };

  /**
   * A duck punch of the $.ui.mouse _mouseInit method to support touch events.
   * This method extends the widget with bound touch event handlers that
   * translate touch events to mouse events and pass them to the widget's
   * original mouse event handling methods.
   */
  mouseProto._mouseInit = function () {
    
    var self = this;

    // Delegate the touch handlers to the widget's element
    self.element.bind({
      touchstart: $.proxy(self, '_touchStart'),
      touchmove: $.proxy(self, '_touchMove'),
      touchend: $.proxy(self, '_touchEnd')
    });

    // Call the original $.ui.mouse init method
    _mouseInit.call(self);
  };

  /**
   * Remove the touch event handlers
   */
  mouseProto._mouseDestroy = function () {
    
    var self = this;

    // Delegate the touch handlers to the widget's element
    self.element.unbind({
      touchstart: $.proxy(self, '_touchStart'),
      touchmove: $.proxy(self, '_touchMove'),
      touchend: $.proxy(self, '_touchEnd')
    });

    // Call the original $.ui.mouse destroy method
    _mouseDestroy.call(self);
  };

})(jQuery);